
/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

// Allow for Node
declare var global: any;

/**
 * The ObjectData type is the same as `{ [index: string]: any; }`.
 * This is to ease writing `someDataObject.anyPropertyName`,
 * when dealing with arbitrary JavaScript Objects used as model data.
 * @category Type
 */
export interface ObjectData {
  [index: string]: any;
}

// Establish the root object, `window` (`self`) in the browser, `global`
// on the server, or `this` in some virtual machines. We use `self`
// instead of `window` for `WebWorker` support.
/** @hidden @internal */
export const root: Window = typeof self === 'object' && self['self'] === self && self ||
          typeof global === 'object' && global['global'] === global && global ||
          typeof window === 'object' && window['window'] === window && window || {};

// For DOM-less environments
if (root.requestAnimationFrame === undefined) root.requestAnimationFrame = root.setImmediate;

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
* This file defines the static Util class for
* internal error checking functions and other handy functions and constants.
*/

// import { root } from './root';


/**
 * @hidden @internal
 * @unrestricted
 */
export class Util {
  /** @hidden @internal */ public static className = 'Util';
  /** @hidden @internal @const */
  public static MTOP = 1;
  /** @hidden @internal @const */
  public static MLEFT = 2;
  /** @hidden @internal @const */
  public static MRIGHT = 4;
  /** @hidden @internal @const */
  public static MBOTTOM = 8;

  // needed for event.detail bug (broken in IE10 and 11):
  // http://connect.microsoft.com/IE/feedback/details/789773/ie-increments-event-detail-on-clicks-over-different-locations
  /** @hidden @internal @const */
  public static isIE9 = (root.navigator !== undefined && root.navigator.userAgent.indexOf('MSIE 9.0') > 0);
  /** @hidden @internal @const */
  public static isIE10 = (root.navigator !== undefined && root.navigator.userAgent.indexOf('MSIE 10.0') > 0);
  /** @hidden @internal @const */
  public static isIE11 = (root.navigator !== undefined && root.navigator.userAgent.indexOf('Trident/7') > 0);
  /** @hidden @internal @const */
  public static isEdge = (root.navigator !== undefined && root.navigator.userAgent.indexOf('Edge/') > 0);

  // Mac but not iOS
  /** @hidden @internal @const */
  public static isMac = (root.navigator !== undefined && root.navigator.platform !== undefined && root.navigator.platform.toUpperCase().indexOf('MAC') >= 0);
  // all iOS
  /** @hidden @internal @const */
  public static isIOS = (root.navigator !== undefined && root.navigator.platform !== undefined && root.navigator.platform.match(/(iPhone|iPod|iPad)/i) !== null);

  /*
  The debounce function must be invoked only once per unique debouncable action to encapsulate a timer
  This would work, because `debounce` is called only one time:

      // var timeout is set up only once
      var myEfficientFn = debounce(function() {
        // All the taxing stuff you do
      }, 250);

      root.addEventListener('resize', function() {
          // Immediate stuff
          myEfficientFn();
        }
      );

  This would not work, because `debounce` would be called each time:

      // var timeout is set up on every single resize
      root.addEventListener('resize', function() {
          // Immediate stuff
          debounce(function() {
            // All the taxing stuff you do
          }, 250)();
        }
      );
  */
  /**
   * @hidden @internal
   * `func` will not be invoked until it stops being called for `wait` milliseconds
   * @param {Function} func
   * @param {number} wait
   * @param {boolean} immediate trigger the function on the leading edge of the wait time instead of the trailing
   */
  public static debounce(func: Function, wait: number, immediate: boolean): ((a: Event) => void) {
    let timeout = -1;
    return function (this: any) {
      const context = this;
      const args = arguments;
      if (timeout !== -1) Util.clearTimeout(timeout);
      timeout = Util.setTimeout(function () {
        timeout = -1;
        if (!immediate) func.apply(context, args);
      }, wait);
      if (immediate && !timeout) func.apply(context, args);
    };
  }

  /**
   * @hidden @internal
   * @param {function()} func
   * @param {number} delay milliseconds
   * @return {number}
   */
  public static setTimeout(func: () => void, delay: number): number {
    return root.setTimeout(func, delay);
  }

  /**
   * @hidden @internal
   * @param {number} id
   */
  public static clearTimeout(id: number): void {
    root.clearTimeout(id);
  }

  /**
   * @hidden @internal
   * @param {string} tag
   * @return {Element}
   */
  public static createElement(tag: string): HTMLElement {
    return root.document.createElement(tag);
  }

  /**
   * @hidden @internal
   * @param {string} msg
   */
  public static throwError(msg: string): never {
    // if (Debug && Debug.assertions && root.alert) root.alert(msg);
    throw new Error(msg);
  }

  /**
   * @hidden @internal
   * @param {Object} obj an object that might no longer be modifiable.
   * @param {*=} val the attempted new value for the property.
   */
  public static throwFrozen(obj: Object, val?: any): void /* never */ {
    let s = 'The object is frozen, so its properties cannot be set: ' + obj.toString();
    if (val !== undefined) s += '  to value: ' + val;
    Util.throwError(s);
  }

  /**
   * @hidden @internal
   * @param {*} val
   * @param {Function} type a class.
   * @param {(string|Function)=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static checkClass(val: any, type: Function, name?: string | Function, member?: string): void {
    if (!(val instanceof type)) {
      let msg = Util.getTypeName(name);
      if (member !== undefined) msg += '.' + member;
      Util.throwTypeError(val, type, msg);
    }
    // end checkClass
  }

  /**
   * @hidden @internal
   * @param {*} val
   * @param {string} type a primitive type: 'number', 'string', 'boolean'.
   * @param {string|Function=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static checkPrimitive(val: any, type: string, name?: string | Function, member?: string): void {
    if (typeof val !== type) {
      let msg = Util.getTypeName(name);
      if (member !== undefined) msg += '.' + member;
      Util.throwTypeError(val, type, msg);
    }
    // end checkPrimitive
  }

  /**
   * @hidden @internal
   * @param {number} val
   * @param {string|function()=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static checkRealNumber(val: any, name?: string | Function, member?: string): void {
    if ((typeof val !== 'number') || !isFinite(val)) {
      let msg = Util.getTypeName(name);
      if (member !== undefined) msg += '.' + member;
      Util.throwError(msg + ' must be a real number type, and not NaN or Infinity: ' + val);
    }
    // end checkRealNumber
  }

  /**
   * @hidden @internal
   * @param {*} val
   * @param {string|Function} type a human-readable string description of the types that VAL isn't, or a class function.
   * @param {string|Function=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static throwTypeError(val: any, type: string | Function, name?: string | Function, member?: string): void /* never */ {
    const typestr = Util.getTypeName(type);
    let msg = Util.getTypeName(name);
    if (member !== undefined) msg += '.' + member;
    if (typeof val === 'string') {
      Util.throwError(msg + ' value is not an instance of ' + typestr + ': ' + '\"' + val + '\"');
    } else {
      Util.throwError(msg + ' value is not an instance of ' + typestr + ': ' + val);
    }
    // end throwTypeError
  }

  /**
   * @hidden @internal
   * @param {*} val
   * @param {string} range a human-readable description of the value range that VAL isn't in.
   * @param {string|function()=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static throwRangeError(val: any, range: string, name?: string | (() => void) | Function, member?: string): void /* never */ {
    let msg = Util.getTypeName(name);
    if (member !== undefined) msg += '.' + member;
    Util.throwError(msg + ' is not in the range ' + range + ': ' + val);
    // end throwRangeError
  }

  /**
   * @hidden @internal
   * @param {function()} type
   */
  public static throwConstructorArgumentsError(type: any): void /* never */ {
    const typename = (typeof type['className'] === 'string') ? type['className'] : '';
    Util.throwError(typename + ' constructor cannot take any arguments.');
    // end throwConstructorArgumentsError
  }

  /**
   * @hidden @internal
   * @param {*} coll
   */
  public static throwCollectionModifiedError(coll: any): void /* never */ {
    Util.throwError('Collection was modified during iteration: ' + coll.toString() + '\n  Perhaps you should iterate over a copy of the collection,' +
      '\n  or you could collect items to be removed from the collection after the iteration.');
      // end throwCollectionModifiedError
  }

  /**
   * @hidden @internal
   * @param {*} obj
   * @param {EnumValue} arg
   */
  public static throwEnumPropertyValueError(obj: any, arg: any): void /* never */ { // arg is any to avoid import statement
    Util.throwError('No property to set for this enum value: ' + arg + ' on ' + obj.toString());
    // end throwEnumPropertyValueError
  }

  /**
   * @hidden @internal
   * This should always produce some sort of logging message, if possible.
   * @param {string} msg
   */
  public static trace(msg: string): void {
    if (root.console) {
      root.console.log(msg);
    }
    // end trace
  }

  /**
   * @hidden @internal
   * Store all deprecated messages emitted so far, to only emit them once.
   */
  public static deprecatedSet: { [index: string]: boolean } = {};

  /**
   * @hidden @internal
   * This should always produce some sort of logging message, if possible.
   * @param {string} msg a description of a feature that may no longer be available in a future version
   * @param {string} version a string describing the future version where the feature might not be present
   */
  public static deprecated(msg: string, version: string): void {
    if (Util.deprecatedSet[msg] === true) return;
    Util.deprecatedSet[msg] = true;
    if (root.console) {
      root.console.log(msg + ' is deprecated in ' + version + ', see the GoJS change log for more information.');
    }
  }
  /**
   * @hidden @internal
   * Like deprecated but with custom messages and option to show only once.
   * This should always produce some sort of logging message, if possible.
   * @param {string} msg a description of a feature that may no longer be available in a future version
   * @param {boolean} showEveryTime true if you want the message to warn every time
   */
  public static warn(msg: string, showEveryTime: boolean): void {
    if (showEveryTime !== true) {
      if (Util.deprecatedSet[msg] === true) return;
      Util.deprecatedSet[msg] = true;
    }
    if (root.console) {
      root.console.log(msg);
    }
  }

  public static collectionWarning: string =
  'Warning: List/Map/Set constructors no longer take an argument that enforces type.' +
  'Instead they take an optional collection of Values to add to the collection. See 2.0 changelog for details.';

  /**
   * @hidden @internal
   * We should never use x instanceof Object, to avoid cross-frame metadata problems.
   */
  public static isObject(val: Object): boolean {
    return (typeof val === 'object' && val !== null);
  }

  // Handling array-like objects

  /**
   * @hidden @internal
   * We should never use x instanceof Array, to avoid cross-frame metadata problems.
   * This is more general than Array.isArray, because it handles XML NodeList and HTML Collection.
   */
  public static isArray(val: any): val is Array<any> {
    return (Array.isArray(val) ||
      (root.NodeList && val instanceof root.NodeList) || (root.HTMLCollection && val instanceof root.HTMLCollection));
  }

  /**
   * @hidden @internal
   */
  public static isModifiableArray(val: Array<any>): boolean {
    return Array.isArray(val);
  }

  /**
   * @hidden @internal
   */
  public static checkArray(val: Array<any>, classfunc: string | Function, prop: string): void {
    if (!Util.isArray(val)) {
      Util.throwTypeError(val, 'Array or NodeList or HTMLCollection', classfunc, prop);
    }
    // end checkArray
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  public static arrayLength(arr: Array<any>): number {
    return arr.length;
  }

  /**
   * @hidden @internal
   * @return {Array}
   */
  public static copyArray(arr: Array<any>): Array<any> {
    return Array.prototype.slice.call(arr);
  }

  /**
   * @hidden @internal
   */
  public static arrayElt<T>(arr: Array<T>, idx: number): T {
    return arr[idx];  // [] works on NodeList or HTMLCollection
  }

  /**
   * @hidden @internal
   */
  public static arraySetElt<T>(arr: Array<T>, idx: number, val: T): void {
    if (Array.isArray(arr)) {
      arr[idx] = val;
    } else {
      Util.throwError('Cannot replace an object in an HTMLCollection or NodeList at ' + idx);
    }
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  public static arrayIndexOf(arr: any, obj: any): number {
    if (Array.isArray(arr)) {
      return arr.indexOf(obj);
    } else {
      const len = arr.length;
      for (let i = 0; i < len; i++) {
        if (arr[i] === obj) return i;
      }
      return -1;
    }
  }

  /**
   * @hidden @internal
   */
  public static arrayInsertAt(arr: any, idx: number, obj: Object): void {
    if (Array.isArray(arr)) {
      if (idx >= arr.length) {  // add to end
        arr.push(obj);
      } else {
        arr.splice(idx, 0, obj);
      }
    } else {
      Util.throwError('Cannot insert an object into an HTMLCollection or NodeList: ' + obj + ' at ' + idx);
    }
  }

  /**
   * @hidden @internal
   */
  public static arrayRemoveAt(arr: any, idx: number): void {
    if (Array.isArray(arr)) {
      if (idx >= arr.length) {  // remove from end
        arr.pop();
      } else {
        arr.splice(idx, 1);
      }
    } else {
      Util.throwError('Cannot remove an object from an HTMLCollection or NodeList at ' + idx);
    }
  }

  // Support for caching commonly used classes now resides in the classes themselves

  /** @hidden @internal @const */
  public static _CachedArrays: Array<Array<any>> = ([]);
  /**
   * @hidden @internal
   * @return {Array}
   */
  public static tempArray<T>(): Array<T> {
    const temp = Util._CachedArrays.pop();
    if (temp === undefined) return [];
    return temp as Array<T>;
  }

  /**
   * @hidden @internal
   * @param {Array} a
   */
  public static freeArray(a: Array<any>): void {
    a.length = 0;  // clear any references to objects
    Util._CachedArrays.push(a);
  }

  /** @hidden @internal @const */
  public static EmptyArray = Object.freeze([]) as ReadonlyArray<never>;

  /**
   * @hidden @internal
   * @return {*} a type description or constructor
   */
  public static getTypeName(type: any): string {
    if (type === null) return '*';
    if (typeof type === 'string') return type;
    if (typeof type === 'function') {
      if (typeof type['className'] === 'string') return type['className'];
    }
    return '';
  }


  /**
   * @hidden @internal
   * @param {function()|Object} x
   * @return {string}
   */
  public static getClassName(x: Function | Object): string {
    if (typeof x === 'function') {
      if (x['className']) return x['className'];
      if (x['name']) return x['name'];
      // see if the function is named by writing to a string and using
      const str = x.toString();
      const paren = str.indexOf('(');
      const name = str.substring(9, paren).trim();
      if (name !== '') {
        x['_className'] = name;  // remember it to avoid future calls to toString()
        return name;
      }
    } else if (Util.isObject(x) && x.constructor) {
      return Util.getClassName(x.constructor);
    }
    return typeof x;
  }

  /**
   * @hidden @internal
   * This static function can be used to convert an object to a string,
   * looking for commonly defined data properties, such as "text", "name", "key", or "id".
   * If none are found, this just calls toString() on it.
   * @param {*=} val
   * @return {string}
   */
  public static toString(val?: any): string {
    let v = val;
    if (Util.isObject(val)) {
      if (val['text']) v = val['text'];
      else if (val['name']) v = val['name'];
      else if (val['key'] !== undefined) v = val['key'];
      else if (val['id'] !== undefined) v = val['id'];
      else if (val.constructor === Object) {
        if (val['Text']) v = val['Text'];
        else if (val['Name']) v = val['Name'];
        else if (val['Key'] !== undefined) v = val['Key'];
        else if (val['Id'] !== undefined) v = val['Id'];
        else if (val['ID'] !== undefined) v = val['ID'];
      }
    }
    if (v === undefined) return 'undefined';
    if (v === null) return 'null';
    return v.toString();
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {string} pname
   */
  public static hasDefinedProperty(obj: Object, pname: string): boolean {
    if (obj.hasOwnProperty(pname)) return true;
    let proto = Object.getPrototypeOf(obj);
    while (proto && proto !== Function) {
      if (proto.hasOwnProperty(pname)) return true;
      const props = proto._ownProperties;
      if (props && props[pname]) return true;
      proto = Object.getPrototypeOf(proto);
    }
    return false;
  }

  /**
   * @hidden @internal
   * @param {Function} classfunc
   * @param {string} propname 'propertyName'
   * @param {function()} getter
   * @param {function(*)} setter
   */
public static defineProperty(classfunc: Function, propname: string, getter: (() => any), setter: ((a: any) => void)): void {
  const desc = {
    'get': getter,
    'set': setter
    // , 'enumerable': true
  };
  Object.defineProperty(classfunc.prototype, propname, desc);
  /*const propdesc = Object.getOwnPropertyDescriptor(classfunc.prototype, propname);
  if (propname && propdesc) {
    Object.defineProperty(classfunc.prototype, propname, propdesc);
  }*/
}


  /**
   * @hidden @internal
   */
  public static clearDiagramsTraversed(): void {
    Util._diagramsTraversed = [];
  }

  /**
   * @hidden @internal
   * Return a list of all diagrams, or create one if none exists
   * @return {Array.<Diagram>}
   */
  public static getDiagramsTraversed(): Array<any> {
    const diags = Util._diagramsTraversed;
    const l = diags.length;
    if (l === 0) {
      // traverse and make list
      const cans = root.document.getElementsByTagName('canvas');
      const ll = cans.length;
      for (let i = 0; i < ll; i++) {
        const can = cans[i];
        // this would be one step fewer if we had a reference to the diagram on the canvas
        if (can.parentElement && can.parentElement._diagram) diags.push(can.parentElement._diagram);
      }
    }
    return diags;
  }
  /** @hidden @internal */
  public static _diagramsTraversed: Array<any> = ([]);

  /**
   * @hidden @internal
   * @param {string} pt
   */
  public static rc4Encrypt(pt: string): string {
    const s: Array<number> = [];
    let i = 0;
    for (i = 0; i < 256; i++) {
      s[i] = i;
    }
    let j = 0;
    let x = 0;
    for (i = 0; i < 256; i++) {
      j = (j + s[i] + 119) % 256; // 119 is arbitrary
      x = s[i];
      s[i] = s[j];
      s[j] = x;
    }
    i = 0;
    j = 0;
    let ct = '';
    for (let y = 0; y < pt.length; y++) {
      i = (i + 1) % 256;
      j = (j + s[i]) % 256;
      x = s[i];
      s[i] = s[j];
      s[j] = x;
      ct += String.fromCharCode(pt.charCodeAt(y) ^ s[(s[i] + s[j]) % 256]);
    }
    return ct;
  }

  /**
   * @hidden @internal
   * @param {string} data to decode
   */
  public static hexDecode(data: string): string {
    const b16Digits = '0123456789abcdef';
    const /** @type {*} */ b16Map = [] as any; // more like a map than an array, but rewrite carefully
    for (let i = 0; i < 256; i++) {
      b16Map[b16Digits.charAt(i >> 4) + b16Digits.charAt(i & 15)] = String.fromCharCode(i);
    }

    if (data.length % 2) data = '0' + data;

    const result = [] as Array<string>;
    let j = 0;
    for (let i = 0; i < data.length; i += 2) {
      result[j++] = b16Map[data.substr(i, 2)];
    }

    const k = result.join('');
    if (k === '') return '0';
    return k;
  }

  /**
   * @hidden @internal
   * @param {string} str string to decrypt
   */
  public static decrypt(str: string): string {
    return Util.rc4Encrypt(Util.hexDecode(str));
  }

  /**
   * @hidden @internal
   */
  public static color1 = '@COLOR1';
  /**
   * @hidden @internal
   */
  public static color2 = '@COLOR2';
  /**
   * @hidden @internal
   */
  public static wmq2 = '32ab5ff3b26f42dc0ed90f224d2913b5'; // 2019
} // end Util static object
Util['className'] = 'Util';
(Util as any)['Dx'] = Util.wmq2;




/**
 * @static
 */
(Util as any)['adym'] = 'gojs.net';
/**
 * @static
 */
(Util as any)['vfo'] = '28e647fdb37b1a8853835e27'; // version info

// #end Util
Util['className'] = 'Util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

// import { Set } from './collections';
// import { Util } from './util';

/**
 * @hidden
 * Undocumented
 * @unrestricted
 */
export class EnumValue {
  /** @hidden @internal */ public static className = 'EnumValue';
  /** @hidden @internal */ private _class: Function;
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _value: number;

  /**
   * @param {function()} classfunc
   * @param {string} name
   * @param {number} val
   */
  constructor(classfunc: Function, name: string, val: number) {
    Set.uniqueHash(this);
    this._class = classfunc;
    this._name = name;
    this._value = val;
  }

  /**
   * @hidden @internal
   * @this {EnumValue}
   * @return {string}
   */
  public toString() {
    return 'EnumValue.' + this._name;
  }

  /**
   * @hidden
   */
  /** @export */ get classType(): Function {
    return this._class;
  }

  /**
   * @hidden
   */
  /** @export */ get name(): string {
    return this._name;
  }

  /**
   * @hidden @internal
   */
  /** @export */ get value(): number {
    return this._value;
  }

  /**
   * @hidden @internal
   * This static function looks for an EnumValue with the given name for the given type.
   * Name comparison is case sensitive.
   * This returns null if no EnumValue is found.
   * @param {function()} classfunc
   * @param {string} name
   * @return {EnumValue}
   */
  public static findName(classfunc: Function, name: string) {
    if (name === undefined || name === null || name === '') return null;  // if NAME is null or an empty string
    return classfunc[name];
  }

  // Support for EnumValues
  /**
   * @hidden @internal
   * @param {*} val
   * @param {*} type a class.
   * @param {string|function()=} name "Class.property" or "Class.method:parameter" or a class function
   * @param {string=} member property name or method name
   */
  public static checkEnumValue(val: any, type: any, name?: string | Function, member?: string) {
    if (val.classType !== type) {
      let msg = Util.getTypeName(name);
      if (member !== undefined) msg += '.' + member;
      Util.throwTypeError(val, 'a constant of class ' + typeof type['className'] === 'function' ? type['className'] : '', msg);
    }
    // end checkEnumValue
  }

} // #end EnumValue
EnumValue['className'] = 'EnumValue';




/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/**
 * @hidden @internal
 * @unrestricted
 */
export class StringBuilder {
  /** @hidden @internal */ public static className = 'StringBuilder';
  /** @hidden @internal */ private _buffer: Array<string>;

  constructor() {
    this._buffer = [];
  }

  /**
   * @hidden @internal
   * @this {StringBuilder}
   * @return {string}
   */
  public toString(): string {
    return this._buffer.join('');
  }

  /**
   * @hidden @internal
   * @this {StringBuilder}
   * @param {string} str
   */
  public add(str: string) {
    if (str !== '') this._buffer.push(str);
  }

} // #end StringBuilder
StringBuilder['className'] = 'StringBuilder';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/**
 * @hidden @internal
 * Just a typed Object on which there are variable numbers of named properties that can have arbitrary values.
 * @unrestricted
 */
export class PropertyCollection {
  /** @hidden @internal */ public static className = 'PropertyCollection';
  [index: string]: any;
} // #end PropertyCollection
PropertyCollection['className'] = 'PropertyCollection';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
 * This file defines the Debug static class. It is not included in builds, but an extern is
 */

// import { root } from './root';

/**
 * @hidden @internal
 * Static Debug object containing convenience functions. This code is for
 * debugging only and is not supported.
 */
export let Debug: any = {
  /**
   * Determines whether or not default debug information is displayed
   */
  boundsInfoEnabled: false,
  drawLocationsEnabled: false,
  drawClipEnabled: false,
  drawLinkpointsEnabled: false,
  nodeInspector: false,  // when you click on a canvas it will return a list of all objects at that location
  handleMessages: false, // tells you whether an event was handled or not

  /**
   * For convenience one can set the default diagram that the Debug object acts
   * upon. Otherwise, the diagram will have to be given as an argument for each
   * function of Debug.
   */
  defaultDiagram: null,

  trace: function (msg: string) {
    if (root.console) {
      root.console.log(msg);
    }
  },

  drawDiagramBounds(ctx: any, t: any, pixelratio: any, docb: any) {
    // origin
    ctx.strokeStyle = 'red';
    ctx.fillStyle = 'red';
    ctx.font = '8px sans-serif';
    ctx.beginPath();
    ctx.moveTo(-10, 0);
    ctx.lineTo(10, 0);
    ctx.moveTo(0, -10);
    ctx.lineTo(0, 10);
    ctx.stroke();

    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.scale(pixelratio, pixelratio);
    ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

    ctx.lineWidth = 2;
    // Top left of document bounds
    ctx.beginPath();
    ctx.moveTo(docb.left, docb.top + 20);
    ctx.lineTo(docb.left, docb.top);
    ctx.lineTo(docb.left + 20, docb.top);
    // Bottom right of document bounds
    ctx.moveTo(docb.right, docb.bottom - 20);
    ctx.lineTo(docb.right, docb.bottom);
    ctx.lineTo(docb.right - 20, docb.bottom);
    ctx.stroke();
    ctx.fillText('DB: ' + Math.round(docb.x) + ', ' + Math.round(docb.y) + ', ' + Math.round(docb.width) + ', ' + Math.round(docb.height),
                  docb.left, docb.top - 5);

  },

  // This finds any read-only properties and attaches setters which throw an error. Only used in Debug library, called from index.
  attachReadonlyWarnings: function (classes: any) {
    for (const x in classes) {
      if (x === 'licenseKey') continue;
      const class1 = classes[x];
      if (class1.prototype === undefined) continue; // strings
      const properties = Object.getOwnPropertyNames(class1.prototype);
      for (let i = 0; i < properties.length; i++) {
        const testprop = Object.getOwnPropertyDescriptor(class1.prototype, properties[i]);
        if (testprop.get !== undefined && testprop.set === undefined) {
          Object.defineProperty(class1.prototype, properties[i], { set: function (value) {
            throw new Error('Property ' + properties[i] + ' of ' + x + ' is read-only.'); } });
        }
      }
    }
  }

}; // end debug


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
* This file defines the Iterable interfaces.
*/

////////////////
// Interfaces //
////////////////
/**
 * This interface is implemented by the {@link List}, {@link Set}, and {@link Map}
 * classes; it provides the {@link #iterator} read-only property that returns an {@link Iterator}.
 *
 * Typical usage is:
 * ```js
 * var it = anIterableCollection.iterator;
 * while (it.next()) {
 *   var item = it.value;
 * }
 * ```
 * @interface
 * @template T
 * @category Collection
 */
export interface Iterable<T> {
  /**
   * Gets an {@link Iterator} that can iterate over the items in the collection.
   *
   * Typical usage is:
   * ```js
   *  var it = anIterableCollection.iterator;
   *  while (it.next()) {
   *    var item = it.value;
   *  }
   * ```
   * @return {Iterator.<T>}
   */
  iterator: Iterator<T>;

  /**
   * Returns the first item in the list, or null if there is none.
   * @return {T|null} This returns null if there are no items in the list.
   */
  first(): T | null;

  /**
   * This read-only property is the number of elements in the collection.
   */
  readonly count: number;
} // #end Iterable

/**
 * This interface defines properties and methods for iterating over a collection;
 * it provides the {@link #next} predicate and the {@link #value} read-only property.
 * Some Iterators also provide `key` property values along with each `value`.
 *
 * Typical usage is:
 * ```js
 *  var it = anIterableCollection.iterator;
 *  while (it.next()) {
 *    var item = it.value;
 *  }
 * ```
 *
 * Many iterators will signal an error if {@link #next} is called
 * after the underlying collection has been modified.
 *
 * To avoid confusion when dealing with {@link Iterable}s,
 * iterators implement the {@link Iterable#iterator} property
 * by just returning themselves.
 * @interface
 * @template T
 * @extends {Iterable.<T>}
 * @category Collection
 */
export interface Iterator<T> extends Iterable<T> {

  /**
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   * @return {Iterator.<T>}
   */
  iterator: Iterator<T>;

  /**
   * Call this method to advance the iterator to the next item in the collection.
   * This should be called before accessing any {@link #value}.
   * @this {Iterator}
   * @return {boolean} whether another item is available; when true the value of {@link #value} will be that item.
   */
  next(): boolean;

  /**
   * @hidden
   * Call this method to advance the iterator to the next item in the collection.
   * This should be called before accessing any {@link #value}.
   * @this {Iterator}
   * @return {boolean} whether another item is available; when true the value of {@link #value} will be that item.
   */
  hasNext(): boolean;

  /**
   * Advance if needed to the first item in the collection and return it, or return null if there is none.
   *
   * Caution: this returns a *key/value pair*, not a *value*, for {@link Map} iterators.
   * @this {Iterator}
   * @return {T|null}
   * @since 1.1
   */
  first(): T | null;

  /**
   * Start this iterator all over again.
   * @this {Iterator}
   */
  reset(): void;

  /**
   * This is true if any invocation of the given predicate on items in the collection is true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns true, this returns true.
   * Otherwise this returns false.
   * For an empty collection this returns false.
   *
   * This automatically {@link #reset}'s itself when it is called.
   * @expose
   * @this {Iterator}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if any predicate call is true; false otherwise.
   * @since 1.4
   */
  any(pred: (x: T) => boolean): boolean;

  /**
   * This is true if all invocations of the given predicate on items in the collection are true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns false, this returns false.
   * Otherwise this returns true.
   * For an empty collection this returns true.
   *
   * This automatically {@link #reset}'s itself when it is called.
   * @expose
   * @this {Iterator}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if all predicate calls are true; false otherwise.
   * @since 1.4
   */
  all(pred: (x: T) => boolean): boolean;

  /**
   * Call the given function on each item in the collection.
   *
   * This automatically {@link #reset}'s itself when it is called.
   * @expose
   * @this {Iterator}
   * @param {function(T)} func This function must not modify the collection.
   * @return {Iterator} this iterator itself
   * @since 1.4
   */
  each(func: (x: T) => void): void;

  /**
   * @hidden
   * Call the given function on each item in the collection and present the results in an iterator.
   *
   * This automatically {@link #reset}'s itself when it is called.
   * @expose
   * @this {Iterator}
   * @param {function(T)} func This function must not modify the collection.
   * @return {Iterator} this
   */
  map<S>(func: (x: T) => S): Iterator<S>;

  /**
   * @hidden
   * Call the given predicate on each item in the collection and for each item that it returns true, present the item in an iterator.
   *
   * This automatically {@link #reset}'s itself when it is called.
   * @expose
   * @this {Iterator}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {Iterator.<T>} this
   */
  filter(pred: (x: T) => boolean): Iterator<T>;

  /**
   * Gets the current item in the collection, assuming {@link #next} has just returned true.
   */
  readonly value: T;

  /**
   * Gets the current index to the item in the collection, assuming {@link #next} has just returned true.
   */
  readonly key: any;

  /**
   * This read-only property is the total number of items in the iterated collection.
   */
  readonly count: number;
} // #end Iterator


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

// import { BrushLike } from './brush';
// import { GraphObject } from './graphObject';
// import { SGradient } from './svgContext';

/*
ISurface
IContext
*/

/**
 * @hidden
 * Undocumented interface used for both the SVGSurface and the CanvasSurface, which are undocumented classes.
 */
export interface ISurface {
  domElement: Element;
  context: IContext;
  width: number;
  height: number;
  ownerDocument: Document;
  resize(pixelWidth: number, pixelHeight: number, width: number, height: number): boolean;
  // Only used on SVG context
  elementFinished: ((a: GraphObject, b: SVGElement) => void) | null;

  getBoundingClientRect(): ClientRect | DOMRect;
  focus(): void;
  dispose(): void;
  style: CSSStyleDeclaration;
} // #end ISurface

/**
 * @hidden
 * Undocumented interface used for both the SVGContext and the CanvasSurfaceContext, which are undocumented classes.
 */
export interface IContext {
  fillStyle: string | CanvasGradient | CanvasPattern | SGradient;
  font: string;
  globalAlpha: number;
  lineCap: string;
  lineDashOffset: number;
  lineJoin: string;
  lineWidth: number;
  miterLimit: number;
  shadowBlur: number;
  shadowColor: string;
  shadowOffsetX: number;
  shadowOffsetY: number;
  strokeStyle: string | CanvasGradient | CanvasPattern | SGradient;
  textAlign: string;
  imageSmoothingEnabled: boolean;
  clipInsteadOfFill: boolean;

  setImageSmoothingEnabled(smooth: boolean): void;
  arc(x: number, y: number, radius: number, startAngle: number, endAngle: number, counterclockwise: boolean, lx?: number, ly?: number): void;
  beginPath(): void;
  bezierCurveTo(a: number, b: number, c: number, d: number, e: number, f: number): void;
  clearRect(x: number, y: number, w: number, h: number): void;
  clip(): void;
  closePath(): void;
  createLinearGradient(aX0: number, aY0: number, aX1: number, aY1: number): CanvasGradient | SGradient;
  createPattern(image: HTMLCanvasElement | HTMLImageElement, repetition: string): CanvasPattern | string;
  createRadialGradient(aX0: number, aY0: number, aR0: number, aX1: number, aY1: number, aR1: number): CanvasGradient | SGradient;
  drawImage(src: HTMLCanvasElement | HTMLImageElement | HTMLVideoElement,
            sx: number, sy: number, sw?: number, sh?: number, dx?: number, dy?: number, dw?: number, dh?: number): void;
  fill(): void;
  fillRect(x: number, y: number, w: number, h: number): void;
  fillText(str: string, x: number, y: number): void;
  getImageData(x: number, y: number, w: number, h: number): ImageData;
  lineTo(x: number, y: number): void;
  measureText(text: string): TextMetrics;
  moveTo(x: number, y: number): void;
  quadraticCurveTo(a: number, b: number, c: number, d: number): void;
  rect(x: number, y: number, w: number, h: number): void;
  restore(): void;
  rotate(angle: number): void;
  save(): void;
  setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void;
  scale(x: number, y: number): void;
  stroke(): void;
  transform(a: number, b: number, c: number, d: number, e: number, f: number): void;
  translate(x: number, y: number): void;
  fillContext(brush: BrushLike): void;
  strokeContext(): void;
  shadowsSet(x: number, y: number, blur: number): void;
  shadowsOff(): void;
  shadowsOn(): void;
  enableDash(strokeDashArray: Array<number>, strokeDashOffset: number): void;
  disableDash(): void;
  clearContextCache(clearFont: boolean): void;
} // #end IContext


// Type definitions for go

// import { Diagram } from "./diagram";

declare global {
  interface Element {
    [index: string]: any;
    /** @hidden @internal */ _diagram: Diagram;
    /** @hidden @internal */ _scrollHorizontal: boolean;
    /** @hidden @internal */ _scrollVertical: boolean;

    // Picture.ts:
    /** @hidden @internal */ _hasLoaded: boolean;
    /** @hidden @internal */ _hasError: Event | boolean;
    /** @hidden @internal */ _isLoaded: boolean;
    /** @hidden @internal */ _savedSuccessEvent: any;
    /** @hidden @internal */ _goJSAttachedEvents: any;
    /** @hidden @internal */ complete: boolean;
  }

  interface Window {
    [index: string]: any;
  }

  /** @hidden @internal */
  interface Function {
    [index: string]: any; // needed for "this.constructor.className"
  }

  /** @hidden @internal */
  interface Document {
    [index: string]: any;
  }

  /** @hidden @internal */
  interface StyleSheet {
    [index: string]: any;
  }

  /** @hidden @internal */
  interface EventTarget {
    _diagram: Diagram;
    scrollTop: number;
    scrollLeft: number;
  }

  /** @hidden @internal */
  interface Event {
    wheelDeltaX: number;
    wheelDeltaY: number;
    wheelDelta: number;
    readonly clientX: number;
    readonly clientY: number;
    readonly path: any;
    composedPath(): Array<EventTarget>;
  }

  /** @hidden @internal */
  interface CSSStyleDeclaration {
    [index: string]: any;
    MozUserSelect: string;
  }

} // #end global



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
* This file defines List, Set, and Map classes,
* their internal ...Iterator classes,
* the internal KeyValuePair class used by Map and Set, and
* the internal MapKeySet class for treating a Map as a Set of keys.
*/

// import { Iterable, Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { ObjectData } from './root';
// import { Util } from './util';

///////////////////////
// Special Iterators //
///////////////////////

/**
 * @hidden @internal
 * @template T
 * @implements {Iterator.<T>}
 * @unrestricted
 */
export class EmptyIterator<T> implements Iterator<T> {
  /** @hidden @internal */ public static className = 'EmptyIterator';
  constructor() { }

  /**
   * @hidden @internal
   * @return {number} int if {@link #next} returned true, this is the iterator's only index, zero.
   */
  public key: number;

  /**
   * @hidden @internal
   * @return {T}
   */
  public value: T;

  /**
   * @hidden @internal
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): EmptyIterator<T> { return this; }

  /**
   * @hidden @internal
   * @export
   * @this {EmptyIterator}
   */
  public reset(): void { }

  /**
   * @hidden @internal
   * @export
   * @this {EmptyIterator}
   * @return {boolean}
   */
  public next(): boolean { return false; }

  /**
   * @hidden @internal
   * @export
   * @this {EmptyIterator}
   * @return {boolean}
   */
  public hasNext(): boolean { return false; }

  /**
   * @hidden @internal
   * @export
   * @this {EmptyIterator}
   * @return {T|null}
   */
  public first(): T | null { return null; }

  /**
   * @hidden @internal
   * @expose
   * @this {EmptyIterator}
   * @return {boolean}
   */
  public any(): boolean { return false; }

  /**
   * @hidden @internal
   * @expose
   * @this {EmptyIterator}
   * @return {boolean}
   */
  public all(): boolean { return true; }

  /**
   * @hidden @internal
   * @expose
   * @this {EmptyIterator}
   * @return {EmptyIterator.<T>} this
   */
  public each(): EmptyIterator<T> { return this; }

  /**
   * @hidden @internal
   * @expose
   * @this {EmptyIterator}
   * @return {EmptyIterator.<T>} this
   */
  public map<S>(func: ((a: T) => S)): EmptyIterator<S> { return this as any; }

  /**
   * @hidden @internal
   * @expose
   * @this {EmptyIterator}
   * @return {EmptyIterator.<T>} this
   */
  public filter(): EmptyIterator<T> { return this; }

  /**
   * @hidden @internal
   */
  /** @export */ get count(): number { return 0; }

  /**
   * @hidden @internal
   * @this {EmptyIterator}
   */
  public _dispose(): void { }

  /**
   * @hidden @internal
   * @this {EmptyIterator}
   * @return {string}
   */
  public toString(): string { return 'EmptyIterator'; }

  public static instance: EmptyIterator<any> = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    EmptyIterator.instance = new EmptyIterator<any>();
  }
} // #end EmptyIterator
EmptyIterator['className'] = 'EmptyIterator';
EmptyIterator.initStatic();

/**
 * @hidden @internal
 * @template T
 * @implements {Iterator.<T>}
 * @unrestricted
 */
export class SingletonIterator<T> implements Iterator<T> {
  /** @hidden @internal */ public static className = 'SingletonIterator';
  /**
   * @param {T} val
   */
  constructor(val: T) {
    this.key = -1;
    this.value = val;
  }

  /**
   * @hidden @internal
   * Gets the only index in the collection, zero.
   * @return {number} int if {@link #next} returned true, this is the iterator's only index, zero.
   */
  public key: number;

  /**
   * @hidden @internal
   * Gets the only item in the collection.
   * @return {T}
   */
  public value: T;

  /**
   * @hidden @internal
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): SingletonIterator<T> { return this; }

  /**
   * @hidden @internal
   * @export
   * @this {SingletonIterator}
   */
  public reset(): void { this.key = -1; }

  /**
   * @hidden @internal
   * @export
   * @this {SingletonIterator}
   * @return {boolean}
   */
  public next(): boolean {
    if (this.key === -1) {
      this.key = 0;  // like a ListIterator?
      return true;
    } else {
      return false;
    }
  }

  /**
   * @hidden @internal
   * @export
   * @this {SingletonIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden @internal
   * @export
   * @this {SingletonIterator}
   * @return {T}
   */
  public first(): T {
    this.key = 0;
    return this.value;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {SingletonIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: T) => boolean)): boolean {
    this.key = -1;
    return pred(this.value);
  }

  /**
   * @hidden @internal
   * @expose
   * @this {SingletonIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: T) => boolean)): boolean {
    this.key = -1;
    return pred(this.value);
  }

  /**
   * @hidden @internal
   * @expose
   * @this {SingletonIterator}
   * @param {function(T)} func
   * @return {SingletonIterator.<T>} this
   */
  public each(func: ((a: T) => void)): SingletonIterator<T> {
    this.key = -1;
    func(this.value);
    return this;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {SingletonIterator}
   * @param {function(T):*} func
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: T) => S)): Iterator<S> {
    return new SingletonIterator(func(this.value));
  }

  /**
   * @hidden @internal
   * @expose
   * @this {SingletonIterator}
   * @param {function(T):boolean} pred
   * @return {Iterator.<T>}
   */
  public filter(pred: ((a: T) => boolean)): Iterator<T> {
    if (pred(this.value)) {
      return new SingletonIterator(this.value);
    } else {
      return EmptyIterator.instance;
    }
  }

  /**
   * @hidden @internal
   */
  /** @export */ get count(): number { return 1; }

  /**
   * @hidden @internal
   * @this {SingletonIterator}
   */
  public _dispose(): void { this.value = null; }

  /**
   * @hidden @internal
   * @this {SingletonIterator}
   * @return {string}
   */
  public toString(): string { return 'SingletonIterator(' + this.value + ')'; }

} // #end SingletonIterator
SingletonIterator['className'] = 'SingletonIterator';


//////////
// List //
//////////

/**
 * @hidden @internal
 * @template T
 * @implements {Iterator.<T>}
 * @unrestricted
 */
export class ListIterator<T> implements Iterator<T> {
  /** @hidden @internal */ public static className = 'ListIterator';
  private _list: List<T>;
  private _predicate: ((a: T) => boolean) | null;
  private _state: number;
  private _index: number;

  /**
   * @param {List.<T>} l
   */
  constructor(l: List<T>) {
    this._list = l;
    this._predicate = null;

    // reset:
    l._iterator = null;  // protect against recursion or multiple use
    this._state = l._changed;
    this._index = -1;
  }

  /**
   * @hidden @internal
   * Gets the current index into the list.
   * @return {number} int if {@link #next} returned true, this is the iterator's current index.
   */
  public key: number;

  /**
   * @hidden @internal
   * Gets the current item in the list.
   * @return {T}
   */
  public value: T;

  /**
   * @hidden @internal
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): ListIterator<T> { return this; }

  /**
   * @hidden @internal
   * Only used for filtering forward iterations.
   */
  /** @export */ get predicate(): ((x: T) => boolean) | null { return this._predicate; }
  /** @export */ set predicate(value: ((x: T) => boolean) | null) { this._predicate = value; }

  /**
   * @hidden @internal
   * @export
   * @this {ListIterator}
   */
  public reset(): void {
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    this._state = l._changed;
    this._index = -1;
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIterator}
   * @return {boolean}
   */
  public next(): boolean {
    const l = this._list;
    if (l._changed !== this._state) {
      if (this.key < 0) return false;
      Util.throwCollectionModifiedError(l);
    }
    const arr = l._arr;
    const len = arr.length;
    let idx = ++this._index;

    const pred = this._predicate;
    if (pred !== null) {
      while (idx < len) {  // try to find value that satisfies predicate
        const val = arr[idx];
        if (pred(val)) {
          this._index = idx;
          this.key = idx;
          this.value = val;
          return true;
        }
        idx++;
      }
    } else {  // no predicate, use all values
      if (idx < len) {
        this.key = idx;
        this.value = arr[idx];
        return true;
      }
      this._dispose();  // only dispose when there's no predicate
    }
    return false;
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIterator}
   * @return {T|null}
   */
  public first(): T | null {
    const l = this._list;
    this._state = l._changed;
    this._index = 0;
    const arr = l._arr;
    const len = arr.length;
    const pred = this._predicate;
    if (pred !== null) {
      let idx = 0;
      while (idx < len) {  // try to find value that satisfies predicate
        const val = arr[idx];
        if (pred(val)) {
          this._index = idx;
          this.key = idx;
          this.value = val;
          return val;
        }
        idx++;
      }
      return null;
    } else if (len > 0) {
      const val = arr[0];
      this.key = 0;
      this.value = val;
      return val;
    } else {
      return null;
    }
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: T) => boolean)): boolean {
    // reset
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    const state = l._changed;
    this._index = -1;

    const arr = l._arr;
    const len = arr.length;
    const inpred = this._predicate;
    for (let i = 0; i < len; i++) {
      const item = arr[i];
      if (inpred !== null && !inpred(item)) continue;
      if (pred(item)) return true;
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return false;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: T) => boolean)): boolean {
    // reset
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    const state = l._changed;
    this._index = -1;

    const arr = l._arr;
    const len = arr.length;
    const inpred = this._predicate;
    for (let i = 0; i < len; i++) {
      const item = arr[i];
      if (inpred !== null && !inpred(item)) continue;
      if (!pred(item)) return false;
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return true;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIterator}
   * @param {function(T)} func
   * @return {ListIterator.<T>} this
   */
  public each(func: ((a: T) => void)): ListIterator<T> {
    // reset
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    const state = l._changed;
    this._index = -1;

    const arr = l._arr;
    const len = arr.length;
    const inpred = this._predicate;
    for (let i = 0; i < len; i++) {
      const item = arr[i];
      if (inpred !== null && !inpred(item)) continue;
      func(item);
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return this;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIterator}
   * @param {function(T):*} func
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: T) => S)): Iterator<S> {
    // reset
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    const state = l._changed;
    this._index = -1;

    const newarr = [];
    const arr = l._arr;
    const len = arr.length;
    const inpred = this._predicate;
    for (let i = 0; i < len; i++) {
      const item = arr[i];
      if (inpred !== null && !inpred(item)) continue;
      newarr.push(func(item));
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    const newlist = new List<S>();  // unknown no element type
    newlist._arr = newarr;
    newlist._modify();
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIterator}
   * @param {function(T):boolean} pred
   * @return {Iterator.<T>}
   */
  public filter(pred: ((a: T) => boolean)): Iterator<T> {
    // reset
    const l = this._list;
    l._iterator = null;  // protect against recursion or multiple use
    const state = l._changed;
    this._index = -1;

    const newarr = [];
    const arr = l._arr;
    const len = arr.length;
    const inpred = this._predicate;
    for (let i = 0; i < len; i++) {
      const item = arr[i];
      if (inpred !== null && !inpred(item)) continue;
      if (pred(item)) newarr.push(item);
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    const newlist = new List<T>();
    newlist._arr = newarr;
    newlist._modify();
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   */
  /** @export */ get count(): number {
    const pred = this._predicate;
    if (pred !== null) {
      // if there's a predicate, need to test each item to see if it counts
      let count = 0;
      const arr = this._list._arr;
      const len = arr.length;
      for (let i = 0; i < len; i++) {
        if (pred(arr[i])) count++;
      }
      return count;
    } else {
      return this._list._arr.length;
    }
  }

  /**
   * @hidden @internal
   * @this {ListIterator}
   */
  public _dispose(): void {
    this.key = -1;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
    this._predicate = null;
    this._list._iterator = this;  // return iterator for re-use
  }

  /**
   * @hidden @internal
   * @this {ListIterator}
   * @return {string}
   */
  public toString(): string {
    return 'ListIterator@' + this._index + '/' + this._list.count;
  }

} // #end ListIterator
ListIterator['className'] = 'ListIterator';

/**
 * @hidden @internal
 * @template T
 * @implements {Iterator.<T>}
 * @unrestricted
 */
export class ListIteratorBackwards<T> implements Iterator<T> {
  /** @hidden @internal */ public static className = 'ListIteratorBackwards';
  private _list: List<T>;
  private _state: number;
  private _index: number;

  /**
   * @param {List.<T>} l
   */
  constructor(l: List<T>) {
    this._list = l;
    // reset:
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    this._state = l._changed;
    this._index = l._arr.length;
  }

  /**
   * @hidden @internal
   * Gets the current index into the list.
   * @return {number} int if {@link #next} returned true, this is the iterator's current index.
   */
  public key: number;

  /**
   * @hidden @internal
   * Gets the current item in the list.
   * @return {T} if {@link #next} returned true, this is the iterator's current item in the list.
   */
  public value: T;

  /**
   * @hidden @internal
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): ListIteratorBackwards<T> { return this; }

  /**
   * @hidden @internal
   * @export
   * @this {ListIteratorBackwards}
   */
  public reset(): void {
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    this._state = l._changed;
    this._index = l._arr.length;
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIteratorBackwards}
   * @return {boolean}
   */
  public next(): boolean {
    const l = this._list;
    if (l._changed !== this._state) {
      if (this.key < 0) return false;
      Util.throwCollectionModifiedError(l);
    }
    const idx = --this._index;
    if (idx >= 0) {
      this.key = idx;
      this.value = l._arr[idx];
      return true;
    } else {
      this._dispose();
      return false;
    }
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIteratorBackwards}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden @internal
   * @export
   * @this {ListIteratorBackwards}
   * @return {T|null}
   */
  public first(): T | null {
    const l = this._list;
    this._state = l._changed;
    const arr = l._arr;
    const idx = arr.length - 1;
    this._index = idx;
    if (idx >= 0) {
      const val = arr[idx];
      this.key = idx;
      this.value = val;
      return val;
    } else {
      return null;
    }
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIteratorBackwards}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: T) => boolean)): boolean {
    // reset
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    const state = l._changed;
    const arr = l._arr;
    const len = arr.length;
    this._index = len;

    for (let i = len - 1; i >= 0; i--) {
      const item = arr[i];
      if (pred(item)) return true;
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return false;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIteratorBackwards}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: T) => boolean)): boolean {
    // reset
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    const state = l._changed;
    const arr = l._arr;
    const len = arr.length;
    this._index = len;

    for (let i = len - 1; i >= 0; i--) {
      const item = arr[i];
      if (!pred(item)) return false;
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return true;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIteratorBackwards}
   * @param {function(T)} func
   * @return {ListIteratorBackwards.<T>} this
   */
  public each(func: ((a: T) => void)): ListIteratorBackwards<T> {
    // reset
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    const state = l._changed;
    const arr = l._arr;
    const len = arr.length;
    this._index = len;

    for (let i = len - 1; i >= 0; i--) {
      const item = arr[i];
      func(item);
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    return this;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIteratorBackwards}
   * @param {function(T):*} func
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: T) => S)): Iterator<S> {
    // reset
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    const state = l._changed;
    const newarr = [];
    const arr = l._arr;
    const len = arr.length;
    this._index = len;
    for (let i = len - 1; i >= 0; i--) {
      const item = arr[i];
      newarr.push(func(item));
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    const newlist = new List<S>();  // unknown no element type
    newlist._arr = newarr;
    newlist._modify();
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {ListIteratorBackwards}
   * @param {function(T):boolean} pred
   * @return {Iterator.<T>}
   */
  public filter(pred: ((a: T) => boolean)): Iterator<T> {
    // reset
    const l = this._list;
    l._iteratorBackwards = null;  // protect against recursion or multiple use
    const state = l._changed;
    const newarr = [];
    const arr = l._arr;
    const len = arr.length;
    this._index = len;
    for (let i = len - 1; i >= 0; i--) {
      const item = arr[i];
      if (pred(item)) newarr.push(item);
      if (l._changed !== state) Util.throwCollectionModifiedError(l);
    }
    const newlist = new List<T>();
    newlist._arr = newarr;
    newlist._modify();
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   */
  /** @export */ get count(): number { return this._list._arr.length; }

  /**
   * @hidden @internal
   * @this {ListIteratorBackwards}
   */
  public _dispose(): void {
    this.key = -1;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
    this._list._iteratorBackwards = this;  // return iterator for re-use
  }

  /**
   * @hidden @internal
   * @this {ListIteratorBackwards}
   * @return {string}
   */
  public toString(): string {
    return 'ListIteratorBackwards(' + this._index + '/' + this._list.count + ')';
  }

} // #end ListIteratorBackwards
ListIteratorBackwards['className'] = 'ListIteratorBackwards';

/**
 * <span style="color: red; font-weight: bold;">NOTE: For 2.0 the {@link #constructor} argument has changed.
 * List now optionally accepts a collection, and only checks types in TypeScript.</span>
 *
 * An ordered iterable collection.
 * In TypeScript it is a generic class that enforces at compile-time the type of elements that may be added to the List.
 *
 * An example usage:
 * ```js
 *   var list = new go.List();  // or in TypeScript: new go.List<go.Point>();
 *   list.add(new go.Point(0, 0));
 *   list.add(new go.Point(20, 10));
 *   list.add(new go.Point(10, 20));
 *   // now list.length === 3
 *   // and list.elt(1) instanceof go.Point
 * ```
 *
 * You can iterate over the items in a List:
 * ```js
 *   var it = aList.iterator;
 *   while (it.next()) {
 *     console.log("#" + it.key + " is " + it.value);
 *   }
 * ```
 * Or:
 * ```js
 *   aList.each(function(val) {
 *     console.log(val);
 *   });
 * ```
 * The key will range from zero to {@link #count}-1.
 *
 * For convenience this **GoJS** List class has synonyms for the following methods and property:
 *   - **get(idx)**: {@link #elt}
 *   - **set(idx,val)**: {@link #setElt}
 *   - **has(val)**: {@link #contains}
 *   - **delete(val)**: {@link #remove}
 *   - **clear()**: {@link #clear}
 *   - **size**: {@link #count}
 *
 * The constructor now takes an optional Iterable or Array argument that provides the initial elements for the new List.
 *
 * Note that GoJS iteration is quite different than ES6 iteration, so that functionality has not been made somewhat compatible.
 * These collection classes were defined in GoJS before the ES6 collection classes were proposed.
 * @template T
 * @implements {Iterable.<T>}
 * @unrestricted
 * @category Collection
 */
export class List<T> implements Iterable<T> {
  /** @hidden @internal */ public static className = 'List';
  /** @hidden @internal */ public _isFrozen: boolean;
  /** @hidden @internal */ public _arr: Array<T>;
  /** @hidden @internal */ public _changed: number;
  /** @hidden @internal */ public _iterator: ListIterator<T>;
  /** @hidden @internal */ public _iteratorBackwards: ListIteratorBackwards<T>;

  /**
   * There are two possible constructors:
   *
   * `new go.List()`, for JavaScript
   *
   * `new go.List<type>()` for TypeScript, to enforce type checking.
   *
   * Typical usage would be something like:
   * ```js
   *   var list = new go.List();  // keep a list of GraphObjects
   * ```
   * @param {Iterable.<T>|Array.<T>=} coll an optional collection of items to add.
   */
  constructor(coll?: Iterable<T> | Array<T>) {
    Set.uniqueHash(this);
    this._isFrozen = false;
    this._arr = [];
    this._changed = 0;
    this._iterator = null;
    this._iteratorBackwards = null;

    if (coll !== undefined) {
      if (typeof coll === 'function' || typeof coll === 'string') {
        Util.warn(Util.collectionWarning, true);
      } else {
        this.addAll(coll);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {List}
   */
  public _modify(): void {
    let c = this._changed;
    c++;
    if (c > 999999999) c = 0;
    this._changed = c;
  }

  /**
   * @hidden @internal
   * Make this List 'frozen' or immutable.
   * Future attempts to change its state, for example by adding or replacing an element,
   * may result in an exception.
   * @export
   * @this {List}
   * @return {List.<T>}
   */
  public freeze(): List<T> {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @export
   * @this {List}
   * @return {List.<T>}
   */
  public thaw(): List<T> {
    this._isFrozen = false;
    return this;
  }

  /**
   * @hidden
   * @this {List}
   * @return {string}
   */
  public toString(): string {
    return 'List()#' + Set.hashId(this);
  }

  /**
   * Adds a given value to the end of the List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {*} val
   * @return {List.<T>} This modified List.
   */
  public add(val: T): List<T> {
    if (val === null) return this;
    if (this._isFrozen) Util.throwFrozen(this, val);
    const arr = this._arr;
    arr.push(val);
    this._modify();
    return this;
  }

  /**
   * Adds a given value to the end of the List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {*} val
   */
  public push(val: T): void {
    this.add(val);
  }

  /**
   * Adds all of the values of a collection to the end of this List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {Iterable.<T>|Array.<T>} coll the collection of items to add.
   * @return {List.<T>} This modified List.
   */
  public addAll(coll: Iterable<T> | Array<T>): List<T> {
    if (coll === null) return this;
    if (this._isFrozen) Util.throwFrozen(this);
    const arr = this._arr;
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        const val = Util.arrayElt<T>(coll, i);
        arr.push(val);
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const val = it.value;
        arr.push(val);
      }
    }
    this._modify();
    return this;
  }

  /**
   * Clears the List.
   * This sets the {@link #count} to zero.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   */
  public clear(): void {
    if (this._isFrozen) Util.throwFrozen(this);
    this._arr.length = 0;
    this._modify();
  }

  /**
   * Returns whether the given value is in this List.
   * @export
   * @this {List}
   * @param {T} val The value to check.
   * @return {boolean} Whether or not the value is contained within the List.
   */
  public contains(val: T): boolean {
    if (val === null) return false;
    const arr = this._arr;
    return arr.indexOf(val) !== -1;
  }

  /**
   * Returns whether the given value is in this List.
   * @export
   * @this {List}
   * @param {T} val The value to check.
   * @return {boolean} Whether or not the value is contained within the List.
   */
  public has(val: T): boolean {
    return this.contains(val);
  }

  /**
   * Returns the index of the given value if it is in this List.
   * @export
   * @this {List}
   * @param {T} val The value to check.
   * @return {number} int returns -1 if the value is not in this list.
   */
  public indexOf(val: T): number {
    if (val === null) return -1;
    const arr = this._arr;
    return arr.indexOf(val);
  }

  /**
   * Returns the element at the given index.
   * @export
   * @this {List}
   * @param {number} i int The index of the element to return.
   * @return {T} the value at the given index.
   */
  public elt(i: number): T {
    if (Debug) Util.checkRealNumber(i, List, 'elt:i');
    const arr = this._arr;
    if (i < 0 || i >= arr.length) Util.throwRangeError(i, '0 <= i < length', List, 'elt:i');
    return arr[i];
  }

  /**
   * Returns the element at the given index.
   * @export
   * @this {List}
   * @param {number} i int The index of the element to return.
   * @return {T} the value at the given index.
   */
  public get(i: number): T {
    return this.elt(i);
  }

  /**
   * Set the element at the given index to a given value.
   * @export
   * @this {List}
   * @param {number} i int The index of the element to set.
   * @param {T} val The value to set at the index.
   */
  public setElt(i: number, val: T): void {
    if (Debug) {
      Util.checkRealNumber(i, List, 'setElt:i');
    }
    const arr = this._arr;
    if (i < 0 || i >= arr.length) Util.throwRangeError(i, '0 <= i < length', List, 'setElt:i');
    if (this._isFrozen) Util.throwFrozen(this, i);
    arr[i] = val;
  }

  /**
   * Set the element at the given index to a given value.
   * @export
   * @this {List}
   * @param {number} i int The index of the element to set.
   * @param {T} val The value to set at the index.
   */
  public set(i: number, val: T): void {
    this.setElt(i, val);
  }

  /**
   * Returns the first item in the list, or null if there is none.
   * @export
   * @this {List}
   * @return {T|null} This returns null if there are no items in the list.
   */
  public first(): T | null {
    const arr = this._arr;
    if (arr.length === 0) return null;
    return arr[0];
  }

  /**
   * Returns the last item in the list, or null if these is none.
   * @export
   * @this {List}
   * @return {T|null} This returns null if there are no items in the list.
   * @since 1.5
   */
  public last(): T | null {
    const arr = this._arr;
    const len = arr.length;
    if (len > 0) return arr[len - 1];
    return null;
  }

  /**
   * Returns the last item in the list and removes it from the list, or just return null if these is none.
   * Use {@link #add} to push an item onto the end of the list.
   * Use {@link #last} to get the last item without reducing the length of the list.
   * @export
   * @this {List}
   * @return {T|null} This returns null if there are no items in the list.
   * @since 1.5
   */
  public pop(): T | null {
    if (this._isFrozen) Util.throwFrozen(this);
    const arr = this._arr;
    if (arr.length > 0) return arr.pop();
    return null;
  }

  /**
   * This is true if any invocation of the given predicate on items in the collection is true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns true, this returns true.
   * Otherwise this returns false.
   * For an empty collection this returns false.
   * @expose
   * @this {List}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if any predicate call is true; false otherwise.
   * @since 1.4
   */
  public any(pred: ((a: T) => boolean)): boolean {
    const arr = this._arr;
    const state = this._changed;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      if (pred(arr[i])) return true;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
    }
    return false;
  }

  /**
   * This is true if all invocations of the given predicate on items in the collection are true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns false, this returns false.
   * Otherwise this returns true.
   * For an empty collection this returns true.
   * @expose
   * @this {List}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if all predicate calls are true; false otherwise.
   * @since 1.4
   */
  public all(pred: ((a: T) => boolean)): boolean {
    const arr = this._arr;
    const state = this._changed;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      if (!pred(arr[i])) return false;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
    }
    return true;
  }

  /**
   * Call the given function on each item in the collection.
   * @expose
   * @this {List}
   * @param {function(T)} func This function must not modify the collection.
   * @return {List.<T>} This List itself
   * @since 1.4
   */
  public each(func: ((a: T) => void)): List<T> {
    const arr = this._arr;
    const state = this._changed;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      func(arr[i]);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
    }
    return this;
  }

  /**
   * @hidden
   * Call the given function on each item in the collection and collect the results in a new List.
   *
   * Unlike {@link Iterator#map}, this returns a List, not an Iterator.
   * @expose
   * @this {List}
   * @param {function(T):*} func This function must not modify the collection.
   * @return {List.<S>}
   */
  public map<S>(func: ((a: T) => S)): List<S> {
    const newlist = new List<S>();  // unknown element type
    const newarr = [];
    const arr = this._arr;
    const state = this._changed;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      newarr.push(func(arr[i]));
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
    }
    newlist._arr = newarr;
    newlist._modify();
    return newlist;
  }

  /**
   * @hidden
   * Call the given predicate on each item in the collection and for each item that it returns true, collect the item in a new List.
   *
   * Unlike {@link Iterator#filter}, this returns a List, not an Iterator.
   * @expose
   * @this {List}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {List.<T>}
   */
  public filter(pred: ((a: T) => boolean)): List<T> {
    const newlist = new List<T>();
    const newarr = [];
    const arr = this._arr;
    const state = this._changed;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      const val = arr[i];
      if (pred(val)) newarr.push(val);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
    }
    newlist._arr = newarr;
    newlist._modify();
    return newlist;
  }

  /**
   * Insert a value before the index i.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {number} i int The index to insert before.
   * @param {T} val The value to insert.
   */
  public insertAt(i: number, val: T): void {
    if (Debug) {
      Util.checkRealNumber(i, List, 'insertAt:i');
    }
    if (i < 0) Util.throwRangeError(i, '>= 0', List, 'insertAt:i');
    if (this._isFrozen) Util.throwFrozen(this, i);
    const arr = this._arr;
    if (i >= arr.length) { // special case, add to end
      arr.push(val);
    } else {
      arr.splice(i, 0, val);
    }
    this._modify();
  }

  /**
   * Removes a given value (if found) from the List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {T} val The value to remove.
   * @return {boolean} true if the value was found and removed, false otherwise.
   */
  public remove(val: T): boolean {
    if (val === null) return false;
    if (this._isFrozen) Util.throwFrozen(this, val);
    const arr = this._arr;
    const idx = arr.indexOf(val);
    if (idx === -1) return false;
    // optimization for last element (8x faster)
    if (idx === arr.length - 1) {
      arr.pop();
    } else {
      arr.splice(idx, 1);
    }
    this._modify();
    return true;
  }

  /**
   * Removes a given value (if found) from the List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {T} val The value to remove.
   * @return {boolean} true if the value was found and removed, false otherwise.
   */
  public delete(val: T): boolean {
    return this.remove(val);
  }

  /**
   * Removes a value at a given index from the List.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {number} i int The index to remove.
   */
  public removeAt(i: number): void {
    if (Debug) Util.checkRealNumber(i, List, 'removeAt:i');
    const arr = this._arr;
    if (i < 0 || i >= arr.length) Util.throwRangeError(i, '0 <= i < length', List, 'removeAt:i');
    if (this._isFrozen) Util.throwFrozen(this, i);
    // optimization for last element (8x faster)
    if (i === arr.length - 1) {
      arr.pop();
    } else {
      arr.splice(i, 1);
    }
    this._modify();
  }

  /**
   * Removes a range of values from the List, given both the starting and the ending zero-based indexes.
   * For example,
   * ```js
   * list.removeRange(2, 4)
   * ```
   * will remove elements 2, 3, and 4 from the list.
   * If there were two or fewer elements in the list to begin with, the list is unchanged.
   * If *from* is greater than *to*, the list is unchanged.
   * If *from* is greater than or equal to the length, the list is unchanged.
   * If *to* is less than zero, the list is unchanged.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {List}
   * @param {number} from int The starting index of the range to remove, inclusive; negative values are treated as zero
   * @param {number} to int The ending index of the range to remove, inclusive; values greater than the length of the list are treated as referring to the last element
   * @return {List.<T>} This modified List
   */
  public removeRange(from: number, to: number): List<T> {
    if (Debug) {
      Util.checkRealNumber(from, List, 'removeRange:from');
      Util.checkRealNumber(to, List, 'removeRange:to');
    }
    const arr = this._arr;
    const len = arr.length;
    if (from < 0) from = 0;
    else if (from >= len) return this;
    if (to < 0) return this;
    else if (to >= len) to = len - 1;
    if (from > to) return this;
    if (this._isFrozen) Util.throwFrozen(this);

    let i = from;
    let j = to + 1;
    while (j < len) {
      arr[i++] = arr[j++];
    }
    arr.length = len - (to - from + 1);
    this._modify();
    return this;
  }

  /**
   * Makes a shallow copy of this List.
   * The values are not copied,
   * so if they are objects they may continue to be shared with the original List.
   * @expose
   * @this {List}
   * @return {List.<T>} The new List with the same elements.
   */
  public copy(): List<T> {
    const newlist = new List<T>();  /* T copy */
    const arr = this._arr;
    if (arr.length > 0) {
      newlist._arr = Array.prototype.slice.call(arr);
    }
    return newlist;
  }

  /**
   * Produces a JavaScript Array from the contents of this List.
   * @export
   * @this {List}
   * @return {Array.<T>} A copy of the List in Array form.
   */
  public toArray(): Array<T> {
    const arr = this._arr;
    const l = this.count;
    const newarr = new Array(l);
    for (let i = 0; i < l; i++) {
      newarr[i] = arr[i];
    }
    return newarr;
  }

  /**
   * @hidden @internal
   * Undocumented.
   */
  /** @export */ get _dataArray(): Array<T> { return this._arr; }

  /**
   * Converts the List to a {@link Set}.
   * The count of the resulting Set may be less than the count of this List
   * if any duplicates were removed.
   * @export
   * @this {List}
   * @return {Set.<T>} A copy of the contents of this List,
   * but with duplicates removed and ordering lost.
   */
  public toSet(): Set<T> {
    const newset = new Set<T>();
    const arr = this._arr;
    const l = this.count;
    for (let i = 0; i < l; i++) {
      newset.add(arr[i]);
    }
    return newset;
  }

  /**
   * Sort the List according to a comparison function.
   * @export
   * @this {List}
   * @param {function(T,T):number} sortfunc  This function is passed two items in the list.
   * It should return zero if they are equal,
   * less than zero if the first value should come before the second value,
   * or greater than zero if the first value should come after the second value.
   * @return {List.<T>} This modified List.
   */
  public sort(sortfunc: ((a: T, b: T) => number)): List<T> {
    if (Debug) Util.checkPrimitive(sortfunc, 'function', List, 'sort:sortfunc');
    if (this._isFrozen) Util.throwFrozen(this);
    this._arr.sort(sortfunc);
    this._modify();
    return this;
  }

  // Sorts a range of consecutive elements in this List based on the given comparison function.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {List}
   * @param {function(*,*):number} sortfunc  This function is passed two elements in the list.
   * It should return zero if they are equal,
   * less than zero if the first value should come before the second value,
   * or greater than zero if the first value should come after the second value.
   * @param {number=} from int The optional index at which to start the sort, including that element;
   * default to zero, the first element of the list.
   * @param {number=} to int The optional index at which to end the sort, excluding that element;
   * defaults to the end of the list.
   * @return {List.<T>} This modified List.
   */
  public sortRange(sortfunc: ((a: T, b: T) => number), from?: number, to?: number): List<T> {
    const arr = this._arr;
    const len = arr.length;
    if (from === undefined) from = 0;
    if (to === undefined) to = len;
    if (Debug) {
      Util.checkPrimitive(sortfunc, 'function', List, 'sortRange:sortfunc');
      Util.checkRealNumber(from, List, 'sortRange:from');
      Util.checkRealNumber(to, List, 'sortRange:to');
    }
    if (this._isFrozen) Util.throwFrozen(this);

    const num = to - from;
    if (num <= 1) return this;

    // FROM can't be at or beyond the last item, because NUM >= 2
    if (from < 0 || from >= len - 1) Util.throwRangeError(from, '0 <= from < length', List, 'sortRange:from');

    if (num === 2) {  // handle special case of just two items in the range
      const first = arr[from];
      const second = arr[from + 1];
      if (sortfunc(first, second) > 0) {
        arr[from] = second;
        arr[from + 1] = first;
        this._modify();
      }
      return this;
    }

    // now NUM >= 3
    if (from === 0) {
      if (to >= len) {  // sort the whole array
        arr.sort(sortfunc);
      } else {  // not the end of the array
        const temp = arr.slice(0, to);
        temp.sort(sortfunc);
        for (let i = 0; i < to; i++) arr[i] = temp[i];
      }
    } else {
      if (to >= len) { // not the beginning of the array, but all the rest
        const temp = arr.slice(from);
        temp.sort(sortfunc);
        for (let i = from; i < len; i++) arr[i] = temp[i - from];
      } else {  // neither beginning nor the end of the array
        const temp = arr.slice(from, to);
        temp.sort(sortfunc);
        for (let i = from; i < to; i++) arr[i] = temp[i - from];
      }
    }

    this._modify();
    return this;
  }

  /**
   * Reverse the order of items in this List.
   * @export
   * @this {List}
   * @return {List.<T>} This modified List.
   */
  public reverse(): List<T> {
    if (this._isFrozen) Util.throwFrozen(this);
    this._arr.reverse();
    this._modify();
    return this;
  }

  // ?? lastIndexOf(val, fromIndex)

  /**
   * This read-only property is the length of the List.
   */
  /** @export */ get count(): number { return this._arr.length; }

  /**
   * This read-only property is the length of the List.
   */
  /** @export */ get size(): number { return this._arr.length; }

  /**
   * This read-only property is the length of the List, a synonym for the {@link #count} property.
   */
  /** @export */ get length(): number { return this._arr.length; }

  /**
   * Gets an object that you can use for iterating over the List.
   * The key will be an integer from zero to the count-1.
   * The value will be the item at that index in the list.
   * Typical usage:
   * ```js
   *   var it = aList.iterator;
   *   while (it.next()) {
   *     . . . "index: " + it.key + " value: " + it.value . . .
   *   }
   * ```
   */
  /** @export */ get iterator(): Iterator<T> {
    if (this._arr.length <= 0) return EmptyIterator.instance;
    const it = this._iterator;
    if (it !== null) {  // reuse a single instance of an iterator, assuming single-threaded
      it.reset();
      return it;
    }
    return new ListIterator(this);
  }

  /**
   * Gets an object that you can use for iterating over the List in backwards order.
   * The key will be an integer from count-1 to zero.
   * The value will be the item at that index in the list.
   * The list is not modified by traversing in reverse order.
   * Typical usage:
   * ```js
   *   var it = aList.iteratorBackwards;
   *   while (it.next()) {
   *     . . . 'key: ' + it.key + ' value: ' + it.value . . .
   *   }
   * ```
   */
  /** @export */ get iteratorBackwards(): Iterator<T> {
    if (this._arr.length <= 0) return EmptyIterator.instance;
    const it = this._iteratorBackwards;
    if (it !== null) {  // reuse a single instance of an iterator, assuming single-threaded
      it.reset();
      return it;
    }
    return new ListIteratorBackwards(this);
  }

} // #end List
List['className'] = 'List';

/////////
// Set //
/////////

/**
 * @hidden @internal
 * @template T
 * @implements {Iterator.<T>}
 * @unrestricted
 */
export class SetIterator<T> implements Iterator<T> {
  /** @hidden @internal */ public static className = 'SetIterator';
  private _set: Set<T>;
  private _state: number;
  private _kvp: KeyValuePair<T, T>;

  /**
   * @param {Set.<T>} s
   */
  constructor(s: Set<T>) {
    this._set = s;
    // reset:
    s._iterator = null;  // protect against recursion or multiple use
    this._state = s._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * Gets the current key used by a value in the set.
   * @return {T} if {@link #next} returned true, this is the iterator's current key in the Set.
   */
  public key: T;

  /**
   * @hidden
   * Gets the current item in the set.
   * @return {T} if {@link #next} returned true, this is the iterator's current item in the Set.
   */
  public value: T;

  /**
   * @hidden
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): SetIterator<T> { return this; }

  /**
   * @hidden
   * @export
   * @this {SetIterator}
   */
  public reset(): void {
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    this._state = s._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * @export
   * @this {SetIterator}
   * @return {boolean}
   */
  public next(): boolean {
    const s = this._set;
    if (s._changed !== this._state) {
      if (this.key === null) return false;
      Util.throwCollectionModifiedError(s);
    }
    let kvp = this._kvp;
    if (kvp === null) {
      kvp = s._first;
    } else {
      kvp = kvp._next;
    }
    if (kvp !== null) {
      this._kvp = kvp;
      this.value = kvp.value;
      this.key = kvp.key;
      return true;
    }
    this._dispose();
    return false;
  }

  /**
   * @hidden
   * @export
   * @this {SetIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden
   * @export
   * @this {SetIterator}
   * @return {T|null}
   */
  public first(): T | null {
    const s = this._set;
    this._state = s._changed;
    const kvp = s._first;
    if (kvp !== null) {
      this._kvp = kvp;
      const val = kvp.value;
      this.key = kvp.key;
      this.value = val;
      return val;
    } else {
      return null;
    }
  }

  /**
   * @hidden
   * @expose
   * @this {SetIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: T) => boolean)): boolean {
    // reset
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    const state = s._changed;
    this._kvp = null;

    let kvp = s._first;
    while (kvp !== null) {
      if (pred(kvp.value)) return true;
      if (s._changed !== state) Util.throwCollectionModifiedError(s);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * @hidden
   * @expose
   * @this {SetIterator}
   * @param {function(T):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: T) => boolean)): boolean {
    // reset
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    const state = s._changed;
    this._kvp = null;

    let kvp = s._first;
    while (kvp !== null) {
      if (!pred(kvp.value)) return false;
      if (s._changed !== state) Util.throwCollectionModifiedError(s);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * @hidden
   * @expose
   * @this {SetIterator}
   * @param {function(T)} func
   * @return {SetIterator.<T>} this
   */
  public each(func: ((a: T) => void)): SetIterator<T> {
    // reset
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    const state = s._changed;
    this._kvp = null;

    let kvp = s._first;
    while (kvp !== null) {
      func(kvp.value);
      if (s._changed !== state) Util.throwCollectionModifiedError(s);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden
   * @expose
   * @this {SetIterator}
   * @param {function(T):*} func This function must not modify the collection.
   * @return {Iterator.<T>}
   */
  public map<S>(func: ((a: T) => S)): Iterator<S> {
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    const newlist = new List<S>();  // unknown element type
    const state = s._changed;
    let kvp = s._first;
    while (kvp !== null) {
      newlist.add(func(kvp.value));
      if (s._changed !== state) Util.throwCollectionModifiedError(s);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   * @expose
   * @this {SetIterator}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {Iterator.<T>}
   */
  public filter(pred: ((a: T) => boolean)): Iterator<T> {
    const s = this._set;
    s._iterator = null;  // protect against recursion or multiple use
    const newlist = new List<T>();
    const state = s._changed;
    let kvp = s._first;
    while (kvp !== null) {
      const val = kvp.value;
      if (pred(val)) newlist.add(val);
      if (s._changed !== state) Util.throwCollectionModifiedError(s);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   */
  /** @export */ get count(): number { return this._set._count; }

  /**
   * @hidden @internal
   * @this {SetIterator}
   */
  public _dispose(): void {
    this.key = null;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
    this._set._iterator = this;  // return iterator for re-use
  }

  /**
   * @hidden
   * @this {SetIterator}
   * @return {string}
   */
  public toString(): string {
    if (this._kvp !== null) {
      return 'SetIterator@' + this._kvp.value;
    } else {
      return 'SetIterator';
    }
  }

} // #end SetIterator
SetIterator['className'] = 'SetIterator';

/**
 * <span style="color: red; font-weight: bold;">NOTE: For 2.0 the {@link #constructor} argument has changed.
 * Set now optionally accepts a collection, and only checks types in TypeScript.</span>
 *
 * An unordered iterable collection that cannot contain two instances of the same value.
 * In TypeScript it is a generic class that enforces at compile-time the type of elements that may be added to the Set.
 *
 * An example usage:
 * ```js
 *   var set = new go.Set();  // In TypeScript: new go.Set<string>();
 *   set.add("orange");
 *   set.add("apple");
 *   set.add("orange");
 *   // now set.count === 2
 *   // and set.contains("orange") === true
 *   // and set.contains("banana") === false
 * ```
 *
 * You can iterate over the items in a Set:
 * ```js
 *   var it = aSet.iterator;
 *   while (it.next()) {
 *     . . . it.value . . .
 *   }
 * ```
 * Or:
 * ```js
 *   aSet.each(function(val) {
 *       . . . val . . .
 *     });
 * ```
 *
 * Although not precisely implementing the features of the EcmaScript 6 **Set** class,
 * this **GoJS** Set class has synonyms for the following methods and property:
 *   - **add(val)**: {@link #add}
 *   - **delete(val)**: {@link #remove}
 *   - **has(val)**: {@link #contains}
 *   - **clear()**: {@link #clear}
 *   - **size**: {@link #count}
 *
 * The constructor now takes an optional Iterable or Array argument that provides the initial elements for the new Set.
 *
 * Note that GoJS iteration is quite different than ES6 iteration, so that functionality has not been made somewhat compatible.
 * These collection classes were defined in GoJS before the ES6 collection classes were proposed.
 * @template T
 * @implements {Iterable.<T>}
 * @unrestricted
 * @category Collection
 */
export class Set<T> implements Iterable<T> {
  /** @hidden @internal */ public static className = 'Set';
  /** @hidden @internal */ protected _isFrozen: boolean;
  /** @hidden @internal */ private _dict: { [index: string]: KeyValuePair<T, T> };
  /** @hidden @internal */ public _count: number;
  /** @hidden @internal */ public _iterator: SetIterator<T>;
  /** @hidden @internal */ public _changed: number;
  /** @hidden @internal */ public _first: KeyValuePair<T, T>;
  /** @hidden @internal */ private _last: KeyValuePair<T, T>;

  /**
   * There are two possible constructors:
   *
   * `new go.Set()`, for JavaScript
   *
   * `new go.Set<T>()` for TypeScript
   *
   * In TypeScript, the optional generic argument describes the type of values
   * that this Set may hold.
   *
   * For example, the expression:
   * ```js
   * // TypeScript:
   * new go.Set<go.Point>()
   * ```
   *
   * Creates a new Set that may only contain {@link Point}s.
   *
   * @param {Iterable.<T>|Array.<T>=} coll an optional collection of items to add.
   */
  constructor(coll?: Iterable<T> | Array<T>) {
    Set.uniqueHash(this);
    this._isFrozen = false;

    this._dict = {};
    this._count = 0;
    this._iterator = null;
    this._changed = 0;
    this._first = null;
    this._last = null;

    if (coll !== undefined) {
      if (typeof coll === 'function' || typeof coll === 'string') {
        Util.warn(Util.collectionWarning, true);
      } else {
        this.addAll(coll);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Set}
   */
  public _modify(): void {
    let c = this._changed;
    c++;
    if (c > 999999999) c = 0;
    this._changed = c;
  }

  /**
   * @hidden @internal
   * Make this Set 'frozen' or immutable.
   * Future attempts to change its state, for example by adding or removing an element,
   * may result in an exception.
   * @export
   * @this {Set}
   * @return {Set.<T>}
   */
  public freeze(): Set<T> {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @export
   * @this {Set}
   * @return {Set.<T>}
   */
  public thaw(): Set<T> {
    this._isFrozen = false;
    return this;
  }

  /**
   * @hidden
   * @this {Set}
   * @return {string}
   */
  public toString(): string {
    return 'Set()#' + Set.hashId(this);
  }

  /**
   * Adds a given value to the Set, if not already present.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {T} val The value to add to the Set; must not be null.
   * @return {Set.<T>} This modified Set.
   */
  public add(val: T): Set<T> {
    if (val === null) return this;
    if (this._isFrozen) Util.throwFrozen(this, val);
    let realkey: any = val;
    if (Util.isObject(val)) {
      realkey = Set.hashIdUnique(val);
    }
    const kvp = this._dict[realkey];
    if (kvp === undefined) {  // not found
      this._count++;
      const newkvp = new KeyValuePair<T, T>(val, val);
      this._dict[realkey] = newkvp;
      const last = this._last;
      if (last === null) {
        this._first = newkvp;
        this._last = newkvp;
      } else {
        newkvp._prev = last;
        last._next = newkvp;
        this._last = newkvp;
      }
      this._modify();
    }
    return this;
  }

  /**
   * Adds all of the values of a collection to this Set.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {Iterable.<T>|Array.<T>} coll the collection of items to add.
   * @return {Set.<T>} This modified Set.
   */
  public addAll(coll: Iterable<T> | Array<T>): Set<T> {
    if (coll === null) return this;
    if (this._isFrozen) Util.throwFrozen(this);
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        this.add(Util.arrayElt(coll, i));
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        this.add(it.value);
      }
    }
    return this;
  }

  /**
   * Returns whether the given value is in this Set.
   * @export
   * @this {Set}
   * @param {T} val The value to check.
   * @return {boolean} Whether or not the value is contained within the Set.
   */
  public contains(val: T): boolean {
    if (val === null) return false;
    let realkey: any = val;
    if (Util.isObject(val)) {
      realkey = Set.hashId(val);
      if (realkey === undefined) return false;
    }
    const kvp = this._dict[realkey];
    return (kvp !== undefined);
  }

  /**
   * Returns whether the given value is in this Set.
   * @export
   * @this {Set}
   * @param {T} val The value to check.
   * @return {boolean} Whether or not the value is contained within the Set.
   */
  public has(val: T): boolean {
    return this.contains(val);
  }

  /**
   * Returns true if all of the values of a given collection are in this Set.
   * @export
   * @this {Set}
   * @param {Iterable.<T>} coll the collection of items to check for.
   * @return {boolean}
   */
  public containsAll(coll: Iterable<T>): boolean {
    if (coll === null) return true;
    const it = coll.iterator;
    while (it.next()) {
      if (!this.contains(it.value)) return false;
    }
    return true;
  }

  /**
   * Returns true if any of the values of a given collection are in this Set.
   * @export
   * @this {Set}
   * @param {Iterable.<T>} coll the collection of items to check for.
   * @return {boolean}
   */
  public containsAny(coll: Iterable<T>): boolean {
    if (coll === null) return true;
    const it = coll.iterator;
    while (it.next()) {
      if (this.contains(it.value)) return true;
    }
    return false;
  }

  /**
   * Returns the first item in the collection, or null if there is none.
   * @export
   * @this {Set}
   * @return {T|null} This returns null if there are no items in the collection.
   */
  public first(): T | null {
    const kvp = this._first;
    if (kvp === null) return null;
    return kvp.value;
  }

  /**
   * This is true if any invocation of the given predicate on items in the collection is true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns true, this returns true.
   * Otherwise this returns false.
   * For an empty collection this returns false.
   * @expose
   * @this {Set}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if any predicate call is true; false otherwise.
   * @since 1.4
   */
  public any(pred: ((a: T) => boolean)): boolean {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      if (pred(kvp.value)) return true;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * This is true if all invocations of the given predicate on items in the collection are true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns false, this returns false.
   * Otherwise this returns true.
   * For an empty collection this returns true.
   * @expose
   * @this {Set}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if all predicate calls are true; false otherwise.
   * @since 1.4
   */
  public all(pred: ((a: T) => boolean)): boolean {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      if (!pred(kvp.value)) return false;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * Call the given function on each item in the collection.
   * @expose
   * @this {Set}
   * @param {function(T)} func This function must not modify the collection.
   * @return {Set.<T>} This Set itself.
   * @since 1.4
   */
  public each(func: ((a: T) => void)): Set<T> {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      func(kvp.value);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden
   * Call the given function on each item in the collection and collect the results in a new Set.
   *
   * Unlike {@link Iterator#map}, this returns a Set, not an Iterator.
   * @expose
   * @this {Set}
   * @param {function(T):*} func This function must not modify the collection.
   * @return {Set.<S>}
   */
  public map<S>(func: ((a: T) => S)): Set<S> {
    const newset = new Set<S>();  // unknown element type
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      newset.add(func(kvp.value));
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return newset;
  }

  /**
   * @hidden
   * Call the given predicate on each item in the collection and for each item that it returns true, collect the item in a new Set.
   *
   * Unlike {@link Iterator#filter}, this returns a Set, not an Iterator.
   * @expose
   * @this {Set}
   * @param {function(T):boolean} pred This function must not have any side-effects.
   * @return {Set.<T>}
   */
  public filter(pred: ((a: T) => boolean)): Set<T> {
    const newset = new Set<T>();
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      const val = kvp.value;
      if (pred(val)) newset.add(val);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return newset;
  }

  /**
   * Removes a value (if found) from the Set.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {T} val The value to remove.
   * @return {boolean} true if the value was found and removed, false otherwise.
   */
  public remove(val: T): boolean {
    if (val === null) return false;
    if (this._isFrozen) Util.throwFrozen(this, val);
    let realkey: any = val;
    if (Util.isObject(val)) {
      realkey = Set.hashId(val);
      if (realkey === undefined) return false;
    }
    const kvp = this._dict[realkey];
    if (kvp === undefined) {  // not found
      return false;
    } else {
      const next = kvp._next;
      const prev = kvp._prev;
      if (next !== null) next._prev = prev;
      if (prev !== null) prev._next = next;
      if (this._first === kvp) this._first = next;
      if (this._last === kvp) this._last = prev;
      delete this._dict[realkey];
      this._count--;
      this._modify();
      return true;
    }
  }

  /**
   * Removes a value (if found) from the Set.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {T} val The value to insert.
   * @return {boolean} true if the value was found and removed, false otherwise.
   */
  public delete(val: T): boolean {
    return this.remove(val);
  }

  /**
   * Removes all of the values of a collection from this Set.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {Iterable.<T>|Array.<T>} coll the collection of items to remove.
   * @return {Set.<T>} This modified Set.
   */
  public removeAll(coll: Iterable<T> | Array<T>): Set<T> {
    if (coll === null) return this;
    if (this._isFrozen) Util.throwFrozen(this);
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        this.remove(Util.arrayElt(coll, i));
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        this.remove(it.value);
      }
    }
    return this;
  }

  /**
   * Removes from this Set all items that are not in the given collection.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   * @param {Iterable.<T>} coll the collection of items that should be kept in this Set.
   * @return {Set.<T>} This modified Set.
   */
  public retainAll(coll: Iterable<T>): Set<T> {
    if (coll === null) return this;
    if (this.count === 0) return this;
    if (this._isFrozen) Util.throwFrozen(this);
    // It would be more efficient not to use a Set if there are few items in COLL.
    const keep = new Set();
    keep.addAll(coll);
    // It would be more efficient to go through kvp chain,
    // rather than rehashing values and accumulating another Array.
    const remove = [];
    const it = this.iterator;
    while (it.next()) {
      const val = it.value;
      if (!keep.contains(val)) remove.push(val);
    }
    this.removeAll(remove);
    return this;
  }

  /**
   * Clears the Set.
   * This sets the {@link #count} to zero.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Set}
   */
  public clear(): void {
    if (this._isFrozen) Util.throwFrozen(this);
    this._dict = {};
    this._count = 0;
    if (this._iterator !== null) this._iterator.reset();
    this._first = null;
    this._last = null;
    this._modify();
  }

  /**
   * Makes a shallow copy of this Set.
   * The values are not copied,
   * so if they are objects they may continue to be shared with the original Set.
   * @expose
   * @this {Set}
   * @return {Set.<T>} The new Set with the same elements.
   */
  public copy(): Set<T> {
    const newset = new Set<T>();
    const dict = this._dict;
    for (const a in dict) {
      newset.add(dict[a].value);
    }
    return newset;
  }

  /**
   * Produces a JavaScript Array from the contents of this Set.
   * @export
   * @this {Set}
   * @return {Array.<T>} A copy of the Set in Array form.
   */
  public toArray(): Array<T> {
    const newarr = new Array(this._count);
    const dict = this._dict;
    let i = 0;
    for (const a in dict) {
      newarr[i] = dict[a].value;
      i++;
    }
    return newarr;
  }

  /**
   * Converts the Set to a {@link List}.
   * Because there is no ordering within a Set,
   * the values in the List may be in any order.
   * @export
   * @this {Set}
   * @return {List.<T>} A copy of the contents of this Set in List form.
   */
  public toList(): List<T> {
    const newlist = new List<T>();
    const dict = this._dict;
    for (const a in dict) {
      newlist.add(dict[a].value);
    }
    return newlist;
  }

  /**
   * This read-only property is the number of elements in the Set.
   */
  /** @export */ get count(): number { return this._count; }

  /**
   * This read-only property is the number of elements in the Set.
   */
  /** @export */ get size(): number { return this._count; }

  /**
   * Gets an object that you can use for iterating over the Set.
   * The value will be a member of the Set.
   * Typical usage:
   * ```js
   *   var it = aSet.iterator;
   *   while (it.next()) {
   *     . . . " value: " + it.value . . .
   *   }
   * ```
   */
  /** @export */ get iterator(): Iterator<T> {
    if (this._count <= 0) return EmptyIterator.instance;
    const it = this._iterator;
    if (it !== null) {  // reuse a single instance of an iterator, assuming single-threaded
      it.reset();
      return it;
    }
    return new SetIterator(this);
  }

  /** @hidden @internal */
  public static _ObjectCounter = 1;  // used by various objects for identity of reference

  /**
   * @hidden
   * Undocumented.
   * Set the unique hash ID for an object.
   * This should be called at the beginning of each constructor that does not inherit from another class.
   */
  public static uniqueHash(obj: ObjectData) {
    obj['__gohashid'] = Set._ObjectCounter++;
  }

  /**
   * @hidden
   * Undocumented.
   * Get the unique hash ID for an object, making it if necessary.
   */
  public static hashIdUnique(obj: ObjectData): number {
    let id = obj['__gohashid'];
    if (id === undefined) {
      id = Set._ObjectCounter++;
      obj['__gohashid'] = id;
    }
    return id;
  }

  /**
   * @hidden
   * Undocumented.
   * Get the unique hash ID for an object; may return undefined.
   */
  public static hashId(obj: ObjectData): number | undefined {
    return obj['__gohashid'];
  }

} // #end Set
Set['className'] = 'Set';
Set['uniqueHash'] = Set.uniqueHash;
Set['hashIdUnique'] = Set.hashIdUnique;
Set['hashId'] = Set.hashId;

///////////////
// MapKeySet //
///////////////

/**
 * @hidden @internal
 * @template K,V
 * @implements {Iterator.<K>}
 * @unrestricted
 */
export class MapKeySetIterator<K, V> implements Iterator<K> {
  /** @hidden @internal */ public static className = 'MapKeySetIterator';
  private _map: Map<K, V>;
  private _state: number;
  private _kvp: KeyValuePair<K, V>;

  /**
   * @param {Map.<K,V>} m
   */
  constructor(m: Map<K, V>) {
    this._map = m;
    // reset:
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * Gets the current key in the map.
   * @return {K}
   */
  public key: K;

  /**
   * @hidden
   * Gets the current key in the map.
   * @return {K} if {@link #next} returned true, this is the iterator's current key in the map.
   */
  public value: K;

  /**
   * @hidden
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): MapKeySetIterator<K, V> { return this; }

  /**
   * @hidden
   * @export
   * @this {MapKeySetIterator}
   */
  public reset(): void {
    const m = this._map;
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * @export
   * @this {MapKeySetIterator}
   * @return {boolean}
   */
  public next(): boolean {
    const m = this._map;
    if (m._changed !== this._state) {
      if (this.key === null) return false;
      Util.throwCollectionModifiedError(m);
    }
    let kvp = this._kvp;
    if (kvp === null) {
      kvp = m._first;
    } else {
      kvp = kvp._next;
    }
    if (kvp !== null) {
      this._kvp = kvp;
      const val = kvp.key;
      this.key = val;
      this.value = val;
      return true;
    }
    this._dispose();
    return false;
  }

  /**
   * @hidden
   * @export
   * @this {MapKeySetIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden
   * @export
   * @this {MapKeySetIterator}
   * @return {K|null}
   */
  public first(): K | null {
    const m = this._map;
    this._state = m._changed;
    const kvp = m._first;
    if (kvp !== null) {
      this._kvp = kvp;
      const val = kvp.key;
      this.key = val;
      this.value = val;
      return val;
    } else {
      return null;
    }
  }

  /**
   * @hidden
   * @expose
   * @this {MapKeySetIterator}
   * @param {function(K):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: K) => boolean)): boolean {
    // reset
    const m = this._map;
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (pred(kvp.key)) return true;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * @hidden
   * @expose
   * @this {MapKeySetIterator}
   * @param {function(K):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: K) => boolean)): boolean {
    // reset
    const m = this._map;
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (!pred(kvp.key)) return false;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * @hidden
   * @expose
   * @this {MapKeySetIterator}
   * @param {function(K)} func
   * @return {MapKeySetIterator.<K,V>} this
   */
  public each(func: ((a: K) => void)): MapKeySetIterator<K, V> {
    // reset
    const m = this._map;
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      func(kvp.key);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden
   *
   * @expose
   * @this {MapKeySetIterator}
   * @param {function(K):*} func This function must not modify the collection.
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: K) => S)): Iterator<S> {
    // reset
    const m = this._map;
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<S>();  // unknown element type
    let kvp = m._first;
    while (kvp !== null) {
      newlist.add(func(kvp.key));
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   *
   * @expose
   * @this {MapKeySetIterator}
   * @param {function(K):boolean} pred This function must not have any side-effects.
   * @return {Iterator.<K>}
   */
  public filter(pred: ((a: K) => boolean)): Iterator<K> {
    // reset
    const m = this._map;
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<K>();
    let kvp = m._first;
    while (kvp !== null) {
      const val = kvp.key;
      if (pred(val)) newlist.add(val);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   */
  /** @export */ get count(): number { return this._map._count; }

  /**
   * @hidden @internal
   * @this {MapKeySetIterator}
   */
  public _dispose(): void {
    this.key = null;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
  }

  /**
   * @hidden
   * @this {MapKeySetIterator}
   * @return {string}
   */
  public toString(): string {
    if (this._kvp !== null) {
      return 'MapKeySetIterator@' + this._kvp.value;
    } else {
      return 'MapKeySetIterator';
    }
  }

} // #end MapKeySetIterator
MapKeySetIterator['className'] = 'MapKeySetIterator';

/**
 * @hidden @internal
 * A read-only Set view of a Map's keys.
 * @template K,V
 * @extends {Set.<K>}
 * @unrestricted
 */
export class MapKeySet<K, V> extends Set<K> {
  /** @hidden @internal */ public static className = 'MapKeySet';
  private _map: Map<K, V>;

  /**
   * @param {Map.<K,V>} map
   */
  constructor(map: Map<K, V>) {
    super();
    Set.uniqueHash(this);
    this._isFrozen = true;
    this._map = map;
  }

  /**
   * @hidden @internal
   * A MapKeySet is read-only, so the concept of freezing and thawing is not applicable.
   * @export
   * @this {MapKeySet}
   * @return {MapKeySet.<K,V>}
   */
  public freeze(): MapKeySet<K, V> { return this; }

  /**
   * @hidden @internal
   * A MapKeySet is read-only, so the concept of freezing and thawing is not applicable.
   * @export
   * @this {MapKeySet}
   * @return {MapKeySet.<K,V>}
   */
  public thaw(): MapKeySet<K, V> { return this; }

  /**
   * @hidden @internal
   * @this {MapKeySet}
   * @return {string}
   */
  public toString(): string {
    return 'MapKeySet(' + this._map.toString() + ')';
  }

  /**
   * @hidden @internal
   * A MapKeySet is read-only.
   * @export
   * @this {MapKeySet}
   */
  public add(val: K): MapKeySet<K, V> {
    Util.throwError('This Set is read-only: ' + this.toString());
    return this;
  }

  /**
   * @hidden @internal
   * Returns whether the given value is in this MapKeySet.
   * @export
   * @this {MapKeySet}
   * @param {K} val The value to check.
   * @return {boolean} Whether or not the value is contained within the MapKeySet.
   */
  public contains(val: K): boolean {
    return this._map.contains(val);
  }

  /**
   * @hidden @internal
   * Returns whether the given value is in this MapKeySet.
   * @export
   * @this {MapKeySet}
   * @param {K} val The value to check.
   * @return {boolean} Whether or not the value is contained within the MapKeySet.
   */
  public has(val: K): boolean {
    return this.contains(val);
  }

  /**
   * @hidden @internal
   * A MapKeySet is read-only.
   * @export
   * @this {MapKeySet}
   */
  public remove(val: K): boolean {
    Util.throwError('This Set is read-only: ' + this.toString());
    return false;
  }

  /**
   * @hidden @internal
   * A MapKeySet is read-only.
   * @export
   * @this {MapKeySet}
   */
  public delete(val: K): boolean {
    return this.remove(val);
  }

  /**
   * @hidden @internal
   * A MapKeySet is read-only.
   * @export
   * @this {MapKeySet}
   */
  public clear(): void {
    Util.throwError('This Set is read-only: ' + this.toString());
  }

  /**
   * @hidden @internal
   * Returns the first item in the collection, or null if there is none.
   * @export
   * @this {MapKeySet}
   * @return {K|null} This returns null if there are no items in the collection.
   */
  public first(): K | null {
    const m = this._map;
    const kvp = m._first;
    if (kvp !== null) {
      return kvp.key;
    } else {
      return null;
    }
  }

  /**
   * @hidden @internal
   * This is true if any invocation of the given predicate on items in the collection is true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns true, this returns true.
   * Otherwise this returns false.
   * For an empty collection this returns false.
   * @expose
   * @this {MapKeySet}
   * @param {function(K):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if any predicate call is true; false otherwise.
   */
  public any(pred: ((a: K) => boolean)): boolean {
    let kvp = this._map._first;
    while (kvp !== null) {
      if (pred(kvp.key)) return true;
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * @hidden @internal
   * This is true if all invocations of the given predicate on items in the collection are true.
   *
   * Call the given predicate on each item in the collection.
   * As soon as a call returns false, this returns false.
   * Otherwise this returns true.
   * For an empty collection this returns true.
   * @expose
   * @this {MapKeySet}
   * @param {function(K):boolean} pred This function must not have any side-effects.
   * @return {boolean} True if all predicate calls are true; false otherwise.
   */
  public all(pred: ((a: K) => boolean)): boolean {
    let kvp = this._map._first;
    while (kvp !== null) {
      if (!pred(kvp.key)) return false;
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * @hidden @internal
   * Call the given function on each item in the collection.
   * @expose
   * @this {MapKeySet}
   * @param {function(K)} func This function must not modify the collection.
   * @return {MapKeySet.<K,V>} This Set itself
   */
  public each(func: ((a: K) => void)): MapKeySet<K, V> {
    let kvp = this._map._first;
    while (kvp !== null) {
      func(kvp.key);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapKeySet}
   * @param {function(K):*} func This function must not modify the collection.
   * @return {Set.<S>}
   */
  public map<S>(func: ((a: K) => S)): Set<S> {
    const newset = new Set<S>();  // unknown element type
    let kvp = this._map._first;
    while (kvp !== null) {
      newset.add(func(kvp.key));
      kvp = kvp._next;
    }
    return newset;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapKeySet}
   * @param {function(K):boolean} pred This function must not have any side-effects.
   * @return {Set.<K>}
   */
  public filter(pred: ((a: K) => boolean)): Set<K> {
    const newset = new Set<K>();
    let kvp = this._map._first;
    while (kvp !== null) {
      const val = kvp.key;
      if (pred(val)) newset.add(val);
      kvp = kvp._next;
    }
    return newset;
  }

  /**
   * @hidden @internal
   * This just creates another MapKeySet for the same {@link Map}.
   * @this {MapKeySet}
   */
  public copy(): MapKeySet<K, V> {
    return new MapKeySet(this._map);
  }

  /**
   * @hidden @internal
   * Makes a shallow copy of this Set.
   * The values are not copied,
   * so if they are objects they may continue to be shared with the original MapKeySet.
   * @export
   * @this {MapKeySet}
   * @return {Set.<K>} The new Set.
   */
  public toSet(): Set<K> {
    const newset = new Set<K>(); /* K copy */
    const dict = this._map._dict;
    for (const a in dict) {
      newset.add(dict[a].key);
    }
    return newset;
  }

  /**
   * @hidden @internal
   * Produces a JavaScript Array from the contents of this Set.
   * @export
   * @this {MapKeySet}
   * @return {Array.<K>} A copy of the Set in Array form.
   */
  public toArray(): Array<K> {
    const dict = this._map._dict;
    const newarr = new Array(this._map._count);
    let i = 0;
    for (const a in dict) {
      newarr[i] = (dict[a].key);
      i++;
    }
    return newarr;
  }

  /**
   * @hidden @internal
   * Converts the MapKeySet to a {@link List}.
   * @export
   * @this {MapKeySet}
   * @return {List.<K>} A copy of the contents of this Set in List form.
   */
  public toList(): List<K> {
    const newlist = new List<K>();  /* K copy */
    const dict = this._map._dict;
    for (const a in dict) {
      newlist.add(dict[a].key);
    }
    return newlist;
  }

  /**
   * @hidden @internal
   * This read-only property is the number of elements in the MapKeySet.
   */
  /** @export */ get count(): number { return this._map._count; }

  /**
   * @hidden @internal
   * Gets the number of elements in the MapKeySet.
   */
  /** @export */ get size(): number { return this._map._count; }

  /**
   * @hidden @internal
   * Gets an object that you can use for iterating over the MapKeySet.
   * The value will be a member of the MapKeySet.
   * Typical usage:
   * ```js
   *   var it = aSet.iterator;
   *   while (it.next()) {
   *     . . . " value: " + it.value . . .
   *   }
   * ```
   */
  /** @export */ get iterator(): Iterator<K> {
    if (this._map._count <= 0) return EmptyIterator.instance;
    return new MapKeySetIterator(this._map);
  }

} // #end MapKeySet
MapKeySet['className'] = 'MapKeySet';

/**
 * @hidden @internal
 * @template K,V
 * @implements {Iterator.<V>}
 * @unrestricted
 */
export class MapValueSetIterator<K, V> implements Iterator<V> {
  /** @hidden @internal */ public static className = 'MapValueSetIterator';
  private _map: Map<K, V>;
  private _state: number;
  private _kvp: KeyValuePair<K, V>;

  /**
   * @param {Map.<K,V>} m
   */
  constructor(m: Map<K, V>) {
    this._map = m;
    // reset:
    m._iteratorValues = null;  // protect against recursion or multiple use
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden @internal
   * Gets the current key used by a value in the set.
   * @return {K} if {@link #next} returned true, this is the iterator's current key in the Set.
   */
  public key: K;

  /**
   * @hidden @internal
   * Gets the current item in the set.
   * @return {V} if {@link #next} returned true, this is the iterator's current item in the Set.
   */
  public value: V;

  /**
   * @hidden @internal
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): MapValueSetIterator<K, V> { return this; }

  /**
   * @hidden @internal
   * @export
   * @this {MapValueSetIterator}
   */
  public reset(): void {
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden @internal
   * @export
   * @this {MapValueSetIterator}
   * @return {boolean}
   */
  public next(): boolean {
    const m = this._map;
    if (m._changed !== this._state) {
      if (this.key === null) return false;
      Util.throwCollectionModifiedError(m);
    }
    let kvp = this._kvp;
    if (kvp === null) {
      kvp = m._first;
    } else {
      kvp = kvp._next;
    }
    if (kvp !== null) {
      this._kvp = kvp;
      this.value = kvp.value;
      this.key = kvp.key;
      return true;
    }
    this._dispose();
    return false;
  }

  /**
   * @hidden @internal
   * @export
   * @this {MapValueSetIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden @internal
   * @export
   * @this {MapValueSetIterator}
   * @return {V|null}
   */
  public first(): V | null {
    const m = this._map;
    this._state = m._changed;
    const kvp = m._first;
    if (kvp !== null) {
      this._kvp = kvp;
      const val = kvp.value;
      this.key = kvp.key;
      this.value = val;
      return val;
    } else {
      return null;
    }
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapValueSetIterator}
   * @param {function(V):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: V) => boolean)): boolean {
    // reset
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (pred(kvp.value)) return true;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapValueSetIterator}
   * @param {function(V):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: V) => boolean)): boolean {
    // reset
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (!pred(kvp.value)) return false;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapValueSetIterator}
   * @param {function(V)} func
   * @return {MapValueSetIterator.<K,V>} this
   */
  public each(func: ((a: V) => void)): MapValueSetIterator<K, V> {
    // reset
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      func(kvp.value);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapValueSetIterator}
   * @param {function(V):*} func This function must not modify the collection.
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: V) => S)): Iterator<S> {
    // reset
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<S>();  // unknown element type
    let kvp = m._first;
    while (kvp !== null) {
      newlist.add(func(kvp.value));
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {MapValueSetIterator}
   * @param {function(V):boolean} pred This function must not have any side-effects.
   * @return {Iterator.<V>}
   */
  public filter(pred: ((a: V) => boolean)): Iterator<V> {
    // reset
    const m = this._map;
    m._iteratorValues = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<V>();
    let kvp = m._first;
    while (kvp !== null) {
      const val = kvp.value;
      if (pred(val)) newlist.add(val);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden @internal
   */
  /** @export */ get count(): number { return this._map._count; }

  /**
   * @hidden @internal
   * @this {MapValueSetIterator}
   */
  public _dispose(): void {
    this.key = null;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
    this._map._iteratorValues = this;  // return iterator for re-use
  }

  /**
   * @hidden @internal
   * @this {MapValueSetIterator}
   * @return {string}
   */
  public toString(): string {
    if (this._kvp !== null) {
      return 'MapValueSetIterator@' + this._kvp.value;
    } else {
      return 'MapValueSetIterator';
    }
  }

} // #end MapValueSetIterator
MapValueSetIterator['className'] = 'MapValueSetIterator';

//////////////////
// KeyValuePair //
//////////////////

/**
 * @hidden
 * Undocumented
 * @template K,V
 * @unrestricted
 */
export class KeyValuePair<K, V> {
  /** @hidden @internal */ public static className = 'KeyValuePair';
  /**
   * @hidden @internal
   */
  public _next: KeyValuePair<K, V>;
  /**
   * @hidden @internal
   */
  public _prev: KeyValuePair<K, V>;

  /**
   * @param {K} k
   * @param {V} v
   */
  constructor(k: K, v: V) {
    this.key = k;
    this.value = v;
    this._next = null;
    this._prev = null;
  }

  /**
   * @hidden
   * @this {KeyValuePair}
   * @return {string}
   */
  public toString(): string {
    return '{' + this.key + ':' + this.value + '}';
  }

  /**
   * @hidden
   * Gets a key for a value in a {@link Map}.
   * @return {K} the key uniquely identifying a {@link #value} in a {@link Map}.
   */
  public key: K;

  /**
   * @hidden
   * Gets a value in a {@link Map}.
   * @return {V} a value in a {@link Map} corresponding to a {@link #key}.
   */
  public value: V;

} // #end KeyValuePair
KeyValuePair['className'] = 'KeyValuePair';

/////////
// Map //
/////////


/**
 * @hidden
 * Undocumented.
 * Iterating over Maps is very similar to an Iterator<KeyValuePair<K, V>>,
 * but not exactly, because the type of the value property is V, not KeyValuePair<K, V>.
 */
export interface IMapIterator<K, T> {
  /** @hidden */
  iterator: IMapIterator<K, T>;
  /** @hidden */
  next(): boolean;
  /** @hidden */
  first(): KeyValuePair<K, T> | null;
  /** @hidden */
  reset(): void;
  /** @hidden */
  any(pred: (x: KeyValuePair<K, T>) => boolean): boolean;
  /** @hidden */
  all(pred: (x: KeyValuePair<K, T>) => boolean): boolean;
  /** @hidden */
  each(func: (x: KeyValuePair<K, T>) => void): IMapIterator<K, T>;
  /** @hidden */
  map<S>(func: (x: KeyValuePair<K, T>) => S): Iterator<S>;
  /** @hidden */
  filter(pred: (x: KeyValuePair<K, T>) => boolean): Iterator<KeyValuePair<K, T>>;
  /** @hidden */
  readonly key: K;
  /** @hidden */
  readonly value: T;
  /** @hidden */
  readonly count: number;
}

/**
 * @hidden @internal
 * Undocumented
 * @template K,V
 * @implements {Iterator.<V>}
 * @unrestricted
 */
export class MapIterator<K, V> implements IMapIterator<K, V> {
  /** @hidden @internal */ public static className = 'MapIterator';
  /** @hidden @internal */ private _map: Map<K, V>;
  /** @hidden @internal */ private _state: number;
  /** @hidden @internal */ public _kvp: KeyValuePair<K, V>;

  /**
   * @param {Map.<K,V>} m
   */
  constructor(m: Map<K, V>) {
    this._map = m;
    // reset:
    m._iterator = null;  // protect against recursion or multiple use
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * Gets the current key for a value in the map.
   * @return {K} if {@link #next} returned true, this is the iterator's current key in the map.
   */
  public key: K;

  /**
   * @hidden
   * Gets the current item in the map.
   * @return {V} if {@link #next} returned true, this is the iterator's current value in the map.
   */
  public value: V;

  /**
   * @hidden
   * Returns itself, which is convenient for code that expects an {@link Iterable}
   * instead of an {@link Iterator}.
   */
  /** @export */ get iterator(): IMapIterator<K, V> { return this; }

  /**
   * @hidden
   * @export
   * @this {MapIterator}
   */
  public reset(): void {
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    this._state = m._changed;
    this._kvp = null;
  }

  /**
   * @hidden
   * @export
   * @this {MapIterator}
   * @return {boolean}
   */
  public next(): boolean {
    const m = this._map;
    if (m._changed !== this._state) {
      if (this.key === null) return false;
      Util.throwCollectionModifiedError(m);
    }
    let kvp = this._kvp;
    if (kvp === null) {
      kvp = m._first;
    } else {
      kvp = kvp._next;
    }
    if (kvp !== null) {
      this._kvp = kvp;
      this.key = kvp.key;
      this.value = kvp.value;
      return true;
    }
    this._dispose();
    return false;
  }

  /**
   * @hidden
   * @export
   * @this {MapIterator}
   * @return {boolean}
   */
  public hasNext(): boolean {
    return this.next();
  }

  /**
   * @hidden
   * @export
   * @this {MapIterator}
   * @return {KeyValuePair.<K,V>|null}
   */
  public first(): KeyValuePair<K, V> | null {
    const m = this._map;
    this._state = m._changed;
    const kvp = m._first;
    if (kvp !== null) {
      this._kvp = kvp;
      this.key = kvp.key;
      this.value = kvp.value;
      return kvp;
    } else {
      return null;
    }
  }

  /**
   * @hidden
   * @expose
   * @this {MapIterator}
   * @param {function(KeyValuePair.<K,V>):boolean} pred
   * @return {boolean}
   */
  public any(pred: ((a: KeyValuePair<K, V>) => boolean)): boolean {
    // reset
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (pred(kvp)) return true;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * @hidden
   * @expose
   * @this {MapIterator}
   * @param {function(KeyValuePair.<K,V>):boolean} pred
   * @return {boolean}
   */
  public all(pred: ((a: KeyValuePair<K, V>) => boolean)): boolean {
    // reset
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      if (!pred(kvp)) return false;
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * @hidden
   * @expose
   * @this {MapIterator}
   * @param {function(KeyValuePair.<K,V>)} func
   * @return {MapIterator.<K,V>} this
   */
  public each(func: ((a: KeyValuePair<K, V>) => void)): IMapIterator<K, V> {
    // reset
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;

    let kvp = m._first;
    while (kvp !== null) {
      func(kvp);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden
   * @expose
   * @this {MapIterator}
   * @param {function(KeyValuePair.<K,V>):*} func This function must not modify the collection.
   * @return {Iterator.<S>}
   */
  public map<S>(func: ((a: KeyValuePair<K, V>) => S)): Iterator<S> {
    // reset
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<S>();  // unknown element type
    let kvp = m._first;
    while (kvp !== null) {
      newlist.add(func(kvp));
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   * @expose
   * @this {MapIterator}
   * @param {function(KeyValuePair.<K,V>):boolean} pred This function must not have any side-effects.
   * @return {Iterator.<KeyValuePair.<K,V>>}
   */
  public filter(pred: ((a: KeyValuePair<K, V>) => boolean)): Iterator<KeyValuePair<K, V>> {
    // reset
    const m = this._map;
    m._iterator = null;  // protect against recursion or multiple use
    const state = m._changed;
    this._kvp = null;
    const newlist = new List<KeyValuePair<K, V>>();
    let kvp = m._first;
    while (kvp !== null) {
      if (pred(kvp)) newlist.add(kvp);
      if (m._changed !== state) Util.throwCollectionModifiedError(m);
      kvp = kvp._next;
    }
    return newlist.iterator;
  }

  /**
   * @hidden
   */
  /** @export */ get count(): number { return this._map._count; }

  /**
   * @hidden @internal
   * @this {MapIterator}
   */
  public _dispose(): void {
    this.key = null;  // release any references
    this.value = null;
    this._state = -1;  // disallow continued use by next()
    this._map._iterator = this;  // return iterator for re-use
  }

  /**
   * @hidden
   * @this {MapIterator}
   * @return {string}
   */
  public toString(): string {
    if (this._kvp !== null) {
      return 'MapIterator@' + this._kvp;
    } else {
      return 'MapIterator';
    }
  }

} // #end MapIterator
MapIterator['className'] = 'MapIterator';

/**
 * <span style="color: red; font-weight: bold;">NOTE: For 2.0 the {@link #constructor} arguments have changed.
 * Map now optionally accepts a collection, and only checks types in TypeScript.</span>
 *
 * An unordered iterable collection of key/value pairs that cannot contain two instances of the same key.
 * In TypeScript it is a generic class that enforces at compile-time the type of the key and the type of the associated value.
 *
 * To create a Map:
 * ```js
 *   var map = new go.Map(); // In TypeScript: new go.Map<string, number>();
 *   map.add("one", 1);
 *   map.add("two", 2);
 *   map.add("three", 3);
 *   // now map.count === 3
 *   // and map.getValue("two") === 2
 *   // and map.contains("zero") === false
 * ```
 *
 * You can iterate over the key/value pairs in a Map:
 * ```js
 *   var it = aMap.iterator;
 *   while (it.next()) {
 *     console.log(it.key + ": " + it.value);
 *   }
 * ```
 * Or:
 * ```js
 *   aMap.each(function(kvp) {
 *     console.log(kvp.key + ": " + kvp.value);
 *   });
 * ```
 * But note that there is no guaranteed ordering amongst the key/value pairs.
 *
 * Call {@link #toKeySet} to get a read-only {@link Set} that holds all of the keys of a Map.
 * Iterating over that Set will produce values that are the keys in the Map.
 *
 * Although not precisely implementing the features and semantics of the EcmaScript 6 **Map** class,
 * this **GoJS** Map class has synonyms for the following methods and property:
 *   - **get(key)**: {@link #getValue}, but returns null instead of undefined when key is not present
 *   - **set(key,val)**: {@link #add}
 *   - **has(key)**: {@link #contains}
 *   - **delete(key)**: {@link #remove}
 *   - **clear()**: {@link #clear}
 *   - **size**: {@link #count}
 *
 * The constructor now takes an optional Iterable or Array argument that provides the initial entries for the new Map.
 *
 * Note that GoJS iteration is quite different than ES6 iteration, so that functionality has not been made somewhat compatible.
 * These collection classes were defined in GoJS before the ES6 collection classes were proposed.
 * @template K, V
 * @unrestricted
 * @category Collection
 */
export class Map<K, V> /* implements Iterable<KeyValuePair<K,V>> */ {
  /** @hidden @internal */ public static className = 'Map';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ public _dict: { [index: string]: KeyValuePair<K, V> };
  /** @hidden @internal */ public _count: number;
  /** @hidden @internal */ public _iterator: MapIterator<K, V>;
  /** @hidden @internal */ public _iteratorValues: MapValueSetIterator<K, V>;
  /** @hidden @internal */ public _changed: number;
  /** @hidden @internal */ public _first: KeyValuePair<K, V>;
  /** @hidden @internal */ private _last: KeyValuePair<K, V>;

  /**
   * There are two possible constructors:
   *
   * `new go.Map()`, for JavaScript
   *
   * `new go.Map<K, V>()` for TypeScript
   *
   * In TypeScript, the two optional generic arguments describe the types of keys
   * and the types of values that this Map may hold.
   *
   * For example, the expression:
   * ```js
   * // TypeScript:
   * new go.Map<string, go.Point>()
   * ```
   * produces a Map that has keys that must be strings and whose associated values
   * must be {@link Point}s.
   * @param {Iterable.<KeyValuePair.<K,V>>|Array.<KeyValuePair.<K,V>>=} coll an optional collection of keys/values to add, or an Array of { key: ..., value: ... } objects.
   * Note that the key/value pairs are objects with "key" and "value" properties, not Arrays of length 2.
   */
  constructor(coll?: Iterator<KeyValuePair<K, V>> | Array<KeyValuePair<K, V>>) {
    Set.uniqueHash(this);
    this._isFrozen = false;

    this._dict = {};

    this._count = 0;
    this._iterator = null;
    this._iteratorValues = null;
    this._changed = 0;
    this._first = null;
    this._last = null;

    if (coll !== undefined) {
      if (typeof coll === 'function' || typeof coll === 'string') {
        Util.warn(Util.collectionWarning, true);
      } else {
        this.addAll(coll);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Map}
   */
  public _modify(): void {
    let c = this._changed;
    c++;
    if (c > 999999999) c = 0;
    this._changed = c;
  }

  /**
   * @hidden @internal
   * Make this Map 'frozen' or immutable.
   * Future attempts to change its state,
   * for example by adding a key/value pair or modifying the value of an existing key,
   * may result in an exception.
   * @export
   * @this {Map}
   * @return {Map.<K,V>}
   */
  public freeze(): Map<K, V> {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @export
   * @this {Map}
   * @return {Map.<K,V>}
   */
  public thaw(): Map<K, V> {
    this._isFrozen = false;
    return this;
  }

  /**
   * @hidden
   * @this {Map}
   * @return {string}
   */
  public toString(): string {
    return 'Map()#' + Set.hashId(this);
  }

  /**
   * Adds a key-value association to the Map, or replaces the value associated with the key
   * if the key was already present in the map.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   * @param {K} key The key or index for storing the value in the Map.
   * @param {V} val The value to add to the Map, associated with the key.
   * @return {Map.<K,V>} This modified Map.
   */
  public add(key: K, val: V): Map<K, V> {
    if (this._isFrozen) Util.throwFrozen(this, key);
    let realkey: any = key;
    if (Util.isObject(key)) {
      realkey = Set.hashIdUnique(key);
    }
    const kvp = this._dict[realkey];
    if (kvp === undefined) {  // not found
      this._count++;
      const newkvp = new KeyValuePair<K, V>(key, val); /* K,V KeyValuePair<> */
      this._dict[realkey] = newkvp;
      const last = this._last;
      if (last === null) {
        this._first = newkvp;
        this._last = newkvp;
      } else {
        newkvp._prev = last;
        last._next = newkvp;
        this._last = newkvp;
      }
      this._modify();
    } else {
      // don't _modify() when replacing value for existing key
      kvp.value = val;
    }
    return this;
  }

  /**
   * Adds a key-value association to the Map, or replaces the value associated with the key
   * if the key was already present in the map.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   * @param {K} key The key or index for storing the value in the Map.
   * @param {V} val The value to add to the Map, associated with the key.
   * @return {Map.<K,V>} This modified Map.
   */
  public set(key: K, val: V): Map<K, V> {
    return this.add(key, val);
  }

  /**
   * Adds all of the key-value pairs of another Map to this Map.
   * If a key is already present in this Map,
   * its value is replaced with the corresponding value from the given map.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   * @param {Iterable.<KeyValuePair.<K,V>>|Array.<KeyValuePair.<K,V>>} coll the collection of keys/values to add, or an Array of { key: ..., value: ... } objects.
   * @return {Map.<K,V>} This modified Map.
   */
  public addAll(coll: Iterator<KeyValuePair<K, V>> | Array<KeyValuePair<K, V>>): Map<K, V> {
    if (coll === null) return this;
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        const kvp = Util.arrayElt(coll, i);
        this.add(kvp.key, kvp.value);
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const kvp = it.value;
        this.add(kvp.key, kvp.value);
      }
    }
    return this;
  }

  /**
   * Returns the first key/value pair in the collection, or null if there is none.
   * @export
   * @this {Map}
   * @return {KeyValuePair.<K,V>} This returns null if there are no items in the collection.
   * @since 1.4
   */
  public first(): KeyValuePair<K, V> {
    return this._first;
  }

  /**
   * This is true if any invocation of the given predicate on items in the collection is true.
   *
   * Call the given predicate on each key/value pair in the collection.
   * As soon as a call returns true, this returns true.
   * Otherwise this returns false.
   * For an empty collection this returns false.
   * @expose
   * @this {Map}
   * @param {function(KeyValuePair.<K,V>):boolean} pred The argument to the predicate will be an object with both "key" and "value" properties.
   * This function must not have any side-effects.
   * @return {boolean} True if any predicate call is true; false otherwise.
   * @since 1.4
   */
  public any(pred: ((a: KeyValuePair<K, V>) => boolean)): boolean {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      if (pred(kvp)) return true;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return false;
  }

  /**
   * This is true if all invocations of the given predicate on items in the collection are true.
   *
   * Call the given predicate on each key/value pair in the collection.
   * As soon as a call returns false, this returns false.
   * Otherwise this returns true.
   * For an empty collection this returns true.
   * @expose
   * @this {Map}
   * @param {function(KeyValuePair.<K,V>):boolean} pred The argument to the predicate will be an object with both "key" and "value" properties.
   * This function must not have any side-effects.
   * @return {boolean} True if all predicate calls are true; false otherwise.
   * @since 1.4
   */
  public all(pred: ((a: KeyValuePair<K, V>) => boolean)): boolean {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      if (!pred(kvp)) return false;
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return true;
  }

  /**
   * Call the given function on each key/value pair in the collection.
   * @expose
   * @this {Map}
   * @param {function(KeyValuePair.<K,V>)} func The argument to the function will be an object with both "key" and "value" properties.
   * This function must not modify the collection.
   * @return {Map.<K,V>} This Map itself
   * @since 1.4
   */
  public each(func: ((a: KeyValuePair<K, V>) => void)): Map<K, V> {
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      func(kvp);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return this;
  }

  /**
   * @hidden
   * Call the given function on each key-value pair in the collection and associate the key with the result of the function in a new Map.
   *
   * Unlike {@link Iterator#map}, this returns a Map, not an Iterator.
   * @expose
   * @this {Map}
   * @param {function(KeyValuePair.<K,V>):*} func The argument to the function will be an object with both "key" and "value" properties.
   * This function must not modify the collection.
   * @return {Map.<K,S>} a new Map with the same keys but values produced by the function
   */
  public map<S>(func: ((a: KeyValuePair<K, V>) => S)): Map<K, S> {
    const newmap = new Map<K, S>();  // unknown value type
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      newmap.add(kvp.key, func(kvp));
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return newmap;
  }

  /**
   * @hidden
   * Call the given predicate on each key-value pair in the collection and for each pair that it returns true, add the key-value association in a new Map.
   *
   * Unlike {@link Iterator#filter}, this returns a Map, not an Iterator.
   * @expose
   * @this {Map}
   * @param {function(KeyValuePair.<K,V>):boolean} pred This function must not have any side-effects.
   * @return {Map.<K,V>}
   */
  public filter(pred: ((a: KeyValuePair<K, V>) => boolean)): Map<K, V> {
    const newmap = new Map<K, V>();
    const state = this._changed;
    let kvp = this._first;
    while (kvp !== null) {
      if (pred(kvp)) newmap.add(kvp.key, kvp.value);
      if (this._changed !== state) Util.throwCollectionModifiedError(this);
      kvp = kvp._next;
    }
    return newmap;
  }

  /**
   * Returns whether the given key is in this Map.
   * @export
   * @this {Map}
   * @param {K} key The key to look up in the Map.
   * @return {boolean} Whether or not the key is contained within the Map.
   */
  public contains(key: K): boolean {
    let realkey: any = key;
    if (Util.isObject(key)) {
      realkey = Set.hashId(key);
      if (realkey === undefined) return false;
    }
    const kvp = this._dict[realkey];
    return (kvp !== undefined);
  }

  /**
   * Returns whether the given key is in this Map.
   * @export
   * @this {Map}
   * @param {K} key The key to look up in the Map.
   * @return {boolean} Whether or not the key is contained within the Map.
   */
  public has(key: K): boolean {
    return this.contains(key);
  }

  /**
   * Returns the value associated with a key.
   * @export
   * @this {Map}
   * @param {K} key The key to look up in the Map.
   * @return {V|null} The value associated with the given key, or null if not present in the Map.
   */
  public getValue(key: K): V | null {
    let realkey: any = key;
    if (Util.isObject(key)) {
      realkey = Set.hashId(key);
      if (realkey === undefined) return null;
    }
    const kvp = this._dict[realkey];
    if (kvp === undefined) return null;
    return kvp.value;
  }

  /**
   * Returns the value associated with a key.
   * @export
   * @this {Map}
   * @param {K} key The key to look up in the Map.
   * @return {V|null} The value associated with the given key, or null if not present in the Map.
   */
  public get(key: K): V | null {
    return this.getValue(key);
  }

  /**
   * Removes a key (if found) from the Map.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   * @param {K} key The key to remove.
   * @return {boolean} true if the key was found and removed, false otherwise.
   */
  public remove(key: K): boolean {
    if (key === null) return false;
    if (this._isFrozen) Util.throwFrozen(this, key);
    let realkey: any = key;
    if (Util.isObject(key)) {
      realkey = Set.hashId(key);
      if (realkey === undefined) return false;
    }
    const kvp = this._dict[realkey];
    if (kvp === undefined) {  // not found
      return false;
    } else {
      const next = kvp._next;
      const prev = kvp._prev;
      if (next !== null) next._prev = prev;
      if (prev !== null) prev._next = next;
      if (this._first === kvp) this._first = next;
      if (this._last === kvp) this._last = prev;
      delete this._dict[realkey];
      this._count--;
      this._modify();
      return true;
    }
  }

  /**
   * Removes a key (if found) from the Map.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   * @param {K} key The key to insert.
   * @return {boolean} true if the key was found and removed, false otherwise.
   */
  public delete(key: K): boolean {
    return this.remove(key);
  }

  /**
   * Clears the Map, removing all key-value associations.
   * This sets the {@link #count} to zero.
   *
   * Be careful not to call this method while iterating over the collection.
   * @export
   * @this {Map}
   */
  public clear(): void {
    if (this._isFrozen) Util.throwFrozen(this);
    this._dict = {};
    this._count = 0;
    if (this._iterator !== null) this._iterator.reset();
    if (this._iteratorValues !== null) this._iteratorValues.reset();
    this._first = null;
    this._last = null;
    this._modify();
  }

  /**
   * Makes a shallow copy of this Map.
   * The keys and their values are not copied,
   * so if they are objects they may continue to be shared with the original Map.
   * @expose
   * @this {Map}
   * @return {Map.<K,V>} The new Map with copies of the same entries.
   */
  public copy(): Map<K, V> {
    const newmap = new Map<K, V>();  /* K,V copy */
    const dict = this._dict;
    for (const rk in dict) {
      const kvp = dict[rk];
      newmap.add(kvp.key, kvp.value);
    }
    return newmap;
  }

  /**
   * Produces a JavaScript Array of key/value pair objects from the contents of this Map.
   * @export
   * @this {Map}
   * @return {Array.<KeyValuePair.<K,V>>} A copy of the Map in Array form,
   * each element being an Object with 'key' and 'value' properties.
   */
  public toArray(): Array<KeyValuePair<K, V>> {
    const dict = this._dict;
    const newarr = new Array(this._count);
    let i = 0;
    for (const rk in dict) {
      const kvp = dict[rk];
      newarr[i] = (new KeyValuePair(kvp.key, kvp.value));
      i++;
    }
    return newarr;
  }

  /**
   * Produces a {@link Set} that provides a read-only view onto the keys of this Map.
   * The collection of keys is not copied.
   * @export
   * @this {Map}
   * @return {Set.<K>}
   */
  public toKeySet(): Set<K> {
    return new MapKeySet(this);
  }

  /**
   * This read-only property is the number of associations in the Map.
   */
  /** @export */ get count(): number { return this._count; }

  /**
   * This read-only property is the number of associations in the Map.
   */
  /** @export */ get size(): number { return this._count; }

  /**
   * Gets an object that you can use for iterating over the key-value pairs of a Map.
   * Typical usage:
   * ```js
   *   var it = aMap.iterator;
   *   while (it.next()) {
   *     console.log("the key: " + it.key + " has value: " + it.value);
   *   }
   * ```
   */
  /** @export */ get iterator(): IMapIterator<K, V> {
    if (this.count <= 0) return (EmptyIterator.instance as any);
    const it = this._iterator;
    if (it !== null) {  // reuse a single instance of an iterator, assuming single-threaded
      it.reset();
      return it;
    }
    return new MapIterator(this);
  }

  /**
   * Gets an object that you can use for iterating over the keys of a Map.
   * Typical usage:
   * ```js
   *   var it = aMap.iteratorKeys;
   *   while (it.next()) {
   *     console.log("key: " + it.value);
   *   }
   * ```
   * @since 1.4
   */
  /** @export */ get iteratorKeys(): Iterator<K> {
    if (this.count <= 0) return EmptyIterator.instance;
    return new MapKeySetIterator(this);
  }

  /**
   * Gets an object that you can use for iterating over the values of a Map.
   * Typical usage:
   * ```js
   *   var it = aMap.iteratorValues;
   *   while (it.next()) {
   *     console.log("value: " + it.value);
   *   }
   * ```
   * @since 1.4
   */
  /** @export */ get iteratorValues(): Iterator<V> {
    if (this.count <= 0) return EmptyIterator.instance;
    const it = this._iteratorValues;
    if (it !== null) {  // reuse a single instance of an iterator, assuming single-threaded
      it.reset();
      return it;
    }
    return new MapValueSetIterator(this);
  }

} // #end Map
Map['className'] = 'Map';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Point
*/



/**
 * A Point represents an x- and y-coordinate pair in two-dimensional space.
 *
 * Use the static functions {@link Point.parse} and {@link Point.stringify} to convert to and from
 * a standard string representation that is independent of the current locale.
 *
 * When an instance of this class is the value of a property of a {@link GraphObject} class or {@link Diagram}
 * or {@link CommandHandler} or a {@link Tool} class, you should treat the object
 * as if it were frozen or read-only -- you cannot modify its properties.
 * This allows the property to return a value without allocating a new instance.
 * If you need to do your own calculations with the value, call {@link #copy} to make
 * a new instance with the same values that you can modify.
 *
 * Many methods modify the object's properties and then return a reference to "this" object.
 * The only instance method to allocate a new object is the {@link #copy} method.
 * The static {@link Point.parse} method also allocates a new object.
 *
 * The "Debug" implementation of this class is significantly slower than the "Release" implementation,
 * mostly due to additional error checking.
 *
 * You cannot inherit from this class.
 * @category Geometry
 */
export class Point {
  /** @hidden @internal */ public static className = 'Point';
  public x: number;
  public y: number;
  /** @hidden @internal */ private _isFrozen: boolean;

  /**
   * The default constructor produces the Point(0,0).
   * This constructor may take either zero arguments or two arguments.
   * @final
   * @struct
   * @param {number=} x The x value.
   * @param {number=} y The y value.
   */
  constructor(x?: number, y?: number) {
    if (x === undefined) {
      this.x = 0;
      this.y = 0;
    } else if (typeof x === 'number' && typeof y === 'number') {
      this.x = x;
      this.y = y;
    } else {
      Util.throwError('Invalid arguments to Point constructor: ' + x + ', ' + y);
    }
    this._isFrozen = false;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Point with those of another Point.
   * This can modify a frozen object.
   * @this {Point}
   * @param {Point} p the other Point from which to copy property values.
   * @return {Point} this Point
   */
  public assign(p: Point): Point {
    this.x = p.x;
    this.y = p.y;
    return this;
  }

  /**
   * @hidden @internal
   * Modify this Point with new X and Y values.
   * @this {Point}
   * @param {number} x
   * @param {number} y
   */
  public assignTo(x: number, y: number): Point {
    this.x = x;
    this.y = y;
    return this;
  }

  /**
   * Modify this Point with new X and Y values.
   * @export
   * @this {Point}
   * @param {number} x
   * @param {number} y
   * @return {Point} this.
   */
  public setTo(x: number, y: number): Point {
    if (Debug) {
      Util.checkPrimitive(x, 'number', Point, 'setTo:x');
      Util.checkPrimitive(y, 'number', Point, 'setTo:y');
      this.checkFrozen();
    }
    this.x = x;
    this.y = y;
    return this;
  }

  /**
   * Modify this Point so that its X and Y values are the same as the given Point.
   * @export
   * @this {Point}
   * @param {Point} p the given Point.
   * @return {Point} this.
   */
  public set(p: Point): Point {
    if (Debug) {
      Util.checkClass(p, Point, Point, 'set:p');
      this.checkFrozen();
    }
    this.x = p.x;
    this.y = p.y;
    return this;
  }

  /**
   * Create a copy of this Point, with the same values.
   * @expose
   * @this {Point}
   * @return {Point}
   */
  public copy(): Point {
    const copy = new Point();
    copy.x = this.x;
    copy.y = this.y;
    return copy;
  }

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Point}
   * @return {Point} this.
   */
  public constant(): Point {
    this._isFrozen = true;
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * If this object is 'constant', just return it.
   * Otherwise make a copy of this object and freeze it ({@link #frozen}).
   * @this {Point}
   * @return {Point}
   */
  public copyFrozen(): Point {
    if (this._isFrozen || Object.isFrozen(this)) return this;
    return this.copy().freeze();
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Point}
   * @return {Point} this.
   */
  public freeze(): Point {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Point}
   * @return {Point} this.
   */
  public thaw(): Point {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    return this;
  }

  /**
   * When about to modify this object, check that this object is thawed.
   * @hidden @internal
   * @this {Point}
   * @param {number=} val
   */
  public checkFrozen(val?: number): void {
    if (Debug) {
      if (this._isFrozen) {
        let s = 'The Point is frozen, so its properties cannot be set: ' + this.toString();
        if (val !== undefined) s += '  to value: ' + val;
        Util.throwError(s);
      }
    }
  }

  /**
   * This static function can be used to read in a {@link Point} from a string that was produced by {@link Point.stringify}.
   *
   * `go.Point.parse("1 2")` produces the Point `new go.Point(1, 2)`.
   * @param {string} str
   * @return {Point}
   */
  public static parse(str: string): Point {
    if (typeof str === 'string') {
      const split = str.split(' ');
      let i = 0;
      let x = 0.0;
      while (split[i] === '') i++;
      let s = split[i++];
      if (s) x = parseFloat(s);
      let y = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) y = parseFloat(s);
      return new Point(x, y);
    } else {
      return new Point();
    }
  }

  /**
   * This static function can be used to write out a {@link Point} as a string that can be read by {@link Point.parse}.
   *
   * `go.Point.stringify(new go.Point(1, 2))` produces the string "1 2".
   * @param {Point} val
   * @return {string}
   */
  public static stringify(val: Point): string {
    if (Debug) Util.checkClass(val, Point);
    return val.x.toString() + ' ' + val.y.toString();
  }

  /**
   * @hidden @internal
   * @this {Point}
   * @return {string} A string containing this Point's information.
   */
  public toString(): string {
    return 'Point(' + this.x + ',' + this.y + ')';
  }

  /**
   * Indicates whether the given Point is equal to this Point.
   * @export
   * @this {Point}
   * @param {Point} p The Point to compare to the current Point.
   * @return {boolean} True if the two Points have identical X and Y values,
   * false otherwise.
   * @see {@link #equalTo}
   */
  public equals(p: Point): boolean {
    if (!(p instanceof Point)) return false;
    return (this.x === p.x && this.y === p.y);
  }

  /**
   * Indicates whether the given point (x, y) is equal to this Point.
   * @export
   * @this {Point}
   * @param {number} x
   * @param {number} y
   * @return {boolean} True if the two Points have identical X and Y values,
   * false otherwise.
   * @see {@link #equals}
   */
  public equalTo(x: number, y: number): boolean {
    return (this.x === x && this.y === y);
  }

  //    * Indicates whether the given Point is nearly equal to this Point.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Point}
   * @param {Point} p The Point to compare to the current Point.
   * @return {boolean} True if the two Points have X and Y values that
   * are equal with a tolerance of 0.5, false otherwise.
   */
  public equalsApprox(p: Point): boolean {
    return Geo.isApprox(this.x, p.x) &&
      Geo.isApprox(this.y, p.y);
  }

  /**
   * Modify this point so that is the sum of the current Point and the
   * x and y co-ordinates of the given Point.
   * @export
   * @this {Point}
   * @param {Point} p The Point to add to this Point.
   * @return {Point} this.
   * @see {@link #offset}
   * @see {@link #subtract}
   */
  public add(p: Point): Point {
    if (Debug) {
      Util.checkClass(p, Point, Point, 'add:p');
      this.checkFrozen();
    }
    this.x += p.x;
    this.y += p.y;
    return this;
  }

  /**
   * Modify this point so that is the difference of this Point and the
   * x and y co-ordinates of the given Point.
   * @export
   * @this {Point}
   * @param {Point} p The Point to subtract from the current Point.
   * @return {Point} this.
   * @see {@link #offset}
   * @see {@link #add}
   */
  public subtract(p: Point): Point {
    if (Debug) {
      Util.checkClass(p, Point, Point, 'subtract:p');
      this.checkFrozen();
    }
    this.x -= p.x;
    this.y -= p.y;
    return this;
  }

  /**
   * Modify this point by shifting its values with the given DX and DY offsets.
   * @export
   * @this {Point}
   * @param {number} dx
   * @param {number} dy
   * @return {Point} this.
   * @see {@link #add}
   * @see {@link #subtract}
   */
  public offset(dx: number, dy: number): Point {
    if (Debug) {
      Util.checkRealNumber(dx, Point, 'offset:dx');
      Util.checkRealNumber(dy, Point, 'offset:dy');
      this.checkFrozen();
    }
    this.x += dx;
    this.y += dy;
    return this;
  }

  /**
   * Modify this Point so that has been rotated about the origin by the given angle.
   * @export
   * @this {Point}
   * @param {number} angle an angle in degrees.
   * @return {Point} this.
   */
  public rotate(angle: number): Point {
    if (Debug) {
      Util.checkRealNumber(angle, Point, 'rotate:angle');
      this.checkFrozen();
    }
    if (angle === 0) return this;
    const px = this.x;
    const py = this.y;
    if (px === 0 && py === 0) return this;
    // optimize common cases to avoid trig function calls
    let c = 0.0;
    let s = 0.0;
    if (angle >= 360) angle -= 360; else if (angle < 0) angle += 360;
    if (angle === 90) {
      c = 0; s = 1;
    } else if (angle === 180) {
      c = -1; s = 0;
    } else if (angle === 270) {
      c = 0; s = -1;
    } else {
      const rad = angle * Math.PI / 180;
      c = Math.cos(rad); s = Math.sin(rad);
    }
    this.x = c * px - s * py;
    this.y = s * px + c * py;
    return this;
  }

  /**
   * Modify this Point so that its X and Y values have been scaled by given factors along the X and Y axes.
   * @export
   * @this {Point}
   * @param {number} sx
   * @param {number} sy
   * @return {Point} this.
   */
  public scale(sx: number, sy: number): Point {
    if (Debug) {
      Util.checkRealNumber(sx, Point, 'scale:sx');
      Util.checkRealNumber(sy, Point, 'scale:sy');
      this.checkFrozen();
    }
    this.x *= sx;
    this.y *= sy;
    return this;
  }

  /**
   * Returns the square of the distance from this Point to a given Point.
   * @export
   * @this {Point}
   * @param {Point} p the other Point to measure to.
   * @return {number} the square of the euclidean distance.
   * @see {@link #distanceSquared}
   */
  public distanceSquaredPoint(p: Point): number {
    if (Debug) Util.checkClass(p, Point, Point, 'distanceSquaredPoint:p');
    const dx = p.x - this.x;
    const dy = p.y - this.y;
    return dx * dx + dy * dy;
  }

  /**
   * Returns the square of the distance from this point to a given point (px, py).
   * @export
   * @this {Point}
   * @param {number} px
   * @param {number} py
   * @return {number} the square of the euclidean distance.
   * @see {@link #distanceSquaredPoint}
   */
  public distanceSquared(px: number, py: number): number {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'distanceSquared:px');
      Util.checkRealNumber(py, Point, 'distanceSquared:py');
    }
    const dx = px - this.x;
    const dy = py - this.y;
    return dx * dx + dy * dy;
  }

  /**
   * Modify this Point so that its X and Y values have been normalized to a unit length.
   * However, if this Point is the origin (zero, zero), its length remains zero.
   * @export
   * @this {Point}
   * @return {Point} this.
   */
  public normalize(): Point {
    if (Debug) this.checkFrozen();
    const px = this.x;
    const py = this.y;
    const len = Math.sqrt(px * px + py * py);
    if (len > 0) {
      this.x = px / len;
      this.y = py / len;
    }
    return this;
  }

  /**
   * Compute the angle from this Point to a given Point.
   * However, if the given Point is the same as this Point, the direction is zero.
   * @export
   * @this {Point}
   * @param {Point} p the other Point to which to measure the relative angle.
   * @return {number} the angle, in degrees, of the line from this Point to the given point.
   * @see {@link #direction}
   */
  public directionPoint(p: Point): number {
    if (Debug) Util.checkClass(p, Point, Point, 'directionPoint:p');
    return this._angleOf(p.x - this.x, p.y - this.y);
  }

  /**
   * Compute the angle from this Point to a given (px,py) point.
   * However, if the point is the same as this Point, the direction is zero.
   * @export
   * @this {Point}
   * @param {number} px
   * @param {number} py
   * @return {number} the angle, in degrees, of the line from this Point to the given point.
   * @see {@link #directionPoint}
   */
  public direction(px: number, py: number): number {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'direction:px');
      Util.checkRealNumber(py, Point, 'direction:py');
    }
    return this._angleOf(px - this.x, py - this.y);
  }

  /**
   * @hidden @internal
   * @this {Point}
   * @param {number} x
   * @param {number} y
   * @return {number}
   */
  private _angleOf(x: number, y: number): number {
    if (x === 0) {
      if (y > 0) {
        return 90;
      } else if (y < 0) {
        return 270;
      } else {
        return 0;
      }
    } else if (y === 0) {
      if (x > 0) {
        return 0;
      } else { // x < 0
        return 180;
      }
    } else {
      if (isNaN(x) || isNaN(y)) return 0;
      let a = Math.atan(Math.abs(y / x)) * 180 / Math.PI;
      if (x < 0) {
        if (y < 0) {
          a += 180;
        } else {
          a = 180 - a;
        }
      } else if (y < 0) {
        a = 360 - a;
      }
      return a;
    }
  }

  /**
   * Modify this point to be the closest point to this point that is on a finite line segment.
   * @export
   * @this {Point}
   * @param {number} px one end of the finite line segment
   * @param {number} py one end of the finite line segment
   * @param {number} qx the other end of the finite line segment
   * @param {number} qy the other end of the finite line segment
   * @return {Point} this modified Point
   * @since 1.3
   */
  public projectOntoLineSegment(px: number, py: number, qx: number, qy: number): Point {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'projectOntoLineSegment:px');
      Util.checkRealNumber(py, Point, 'projectOntoLineSegment:py');
      Util.checkRealNumber(qx, Point, 'projectOntoLineSegment:qx');
      Util.checkRealNumber(qy, Point, 'projectOntoLineSegment:qy');
    }
    Geo.nearestPointOnLine(px, py, qx, qy, this.x, this.y, this);
    return this;
  }

  /**
   * Modify this point to be the closest point to this point that is on a finite line segment.
   * @export
   * @this {Point}
   * @param {Point} p one end of the finite line segment
   * @param {Point} q the other end of the finite line segment
   * @return {Point} this modified Point
   * @since 1.3
   */
  public projectOntoLineSegmentPoint(p: Point, q: Point): Point {
    if (Debug) {
      Util.checkClass(p, Point, Point, 'projectOntoLineSegmentPoint:p');
      Util.checkClass(q, Point, Point, 'projectOntoLineSegmentPoint:q');
    }
    Geo.nearestPointOnLine(p.x, p.y, q.x, q.y, this.x, this.y, this);
    return this;
  }

  /**
   * Modify this point to be at the nearest point on an infinite grid,
   * given the grid's origin and size of each grid cell.
   * @export
   * @this {Point}
   * @param {number} originx the top-left point of one of the grid cells
   * @param {number} originy the top-left point of one of the grid cells
   * @param {number} cellwidth the size of each grid cell -- must be a real number larger than zero
   * @param {number} cellheight the size of each grid cell -- must be a real number larger than zero
   * @return {Point} this modified Point
   * @since 1.3
   */
  public snapToGrid(originx: number, originy: number, cellwidth: number, cellheight: number): Point {
    if (Debug) {
      Util.checkRealNumber(originx, Point, 'snapToGrid:originx');
      Util.checkRealNumber(originy, Point, 'snapToGrid:originy');
      Util.checkRealNumber(cellwidth, Point, 'snapToGrid:cellwidth');
      Util.checkRealNumber(cellheight, Point, 'snapToGrid:cellheight');
    }
    Geo.findNearestInfiniteGridPoint(this.x, this.y, originx, originy, cellwidth, cellheight, this);
    return this;
  }

  /**
   * Modify this point to be at the nearest point on an infinite grid,
   * given the grid's origin and size of each grid cell.
   * @export
   * @this {Point}
   * @param {Point} origin the top-left point of one of the grid cells
   * @param {Size} cellsize the size of each grid cell
   * @return {Point} this modified Point
   * @since 1.3
   */
  public snapToGridPoint(origin: Point, cellsize: Size): Point {
    if (Debug) {
      Util.checkClass(origin, Point, Point, 'snapToGridPoint:p');
      Util.checkClass(cellsize, Size, Point, 'snapToGridPoint:q');
    }
    Geo.findNearestInfiniteGridPoint(this.x, this.y, origin.x, origin.y, cellsize.width, cellsize.height, this);
    return this;
  }

  /**
   * Modify this Point so that its X and Y values correspond to a particular {@link Spot}
   * in a given Rect.
   *
   * The result is meaningless if {@link Spot#isNoSpot} is true for the given Spot.
   * @export
   * @this {Point}
   * @param {Rect} r the Rect for which we are finding the point.
   * @param {Spot} spot the Spot; {@link Spot#isSpot} must be true for this Spot.
   * @return {Point} this.
   * @see {@link #setSpot}
   */
  public setRectSpot(r: Rect, spot: Spot): Point {
    if (Debug) {
      Util.checkClass(r, Rect, Point, 'setRectSpot:r');
      Util.checkClass(spot, Spot, Point, 'setRectSpot:spot');
      this.checkFrozen();
    }
    this.x = r.x + spot.x * r.width + spot.offsetX;
    this.y = r.y + spot.y * r.height + spot.offsetY;
    return this;
  }

  /**
   * Modify this Point so that its X and Y values correspond to a particular {@link Spot}
   * in a given rectangle.
   *
   * The result is meaningless if {@link Spot#isNoSpot} is true for the given Spot.
   * @export
   * @this {Point}
   * @param {number} x The X coordinate of the Rect for which we are finding the point.
   * @param {number} y The Y coordinate of the Rect for which we are finding the point.
   * @param {number} w The Width of the Rect for which we are finding the point.
   * @param {number} h The Height of the Rect for which we are finding the point.
   * @param {Spot} spot the Spot; {@link Spot#isSpot} must be true for this Spot.
   * @return {Point} this the point of the spot in the rectangle.
   * @see {@link #setRectSpot}
   */
  public setSpot(x: number, y: number, w: number, h: number, spot: Spot): Point {
    if (Debug) {
      Util.checkRealNumber(x, Point, 'setSpot:x');
      Util.checkRealNumber(y, Point, 'setSpot:y');
      Util.checkRealNumber(w, Point, 'setSpot:w');
      Util.checkRealNumber(h, Point, 'setSpot:h');
      if (w < 0 || h < 0) {
        Util.throwError('Point.setSpot:Width and height cannot be negative');
      }
      Util.checkClass(spot, Spot, Point, 'setSpot:spot');
      this.checkFrozen();
    }
    this.x = x + spot.x * w + spot.offsetX;
    this.y = y + spot.y * h + spot.offsetY;
    return this;
  }

  /**
   * @hidden @internal
   * Modify this Point so that its X and Y values have been multiplied with the given Transform.
   * @this {Point}
   * @param {Transform} t the given Transform.
   * @return {Point} this transformed Point.
   */
  public transform(t: Transform): Point {
    if (Debug) Util.checkClass(t, Transform, Point, 'transform:t');
    t.transformPoint(this);
    return this;
  }

  /**
   * @hidden @internal
   * Modify this Point so that its X and Y values have been multiplied
   * with the inverse of the given Transform.
   * @this {Point}
   * @param {Transform} t the given Transform, which is not modified.
   * @return {Point} this transformed Point.
   */
  public transformInverted(t: Transform): Point {
    if (Debug) Util.checkClass(t, Transform, Point, 'transformInverted:t');
    t.invertedTransformPoint(this);
    return this;
  }

  /**
   * This static function returns the square of the distance from the point P
   * to the finite line segment from point A to point B.
   * @param {number} px
   * @param {number} py
   * @param {number} ax
   * @param {number} ay
   * @param {number} bx
   * @param {number} by
   * @return {number} the euclidean distance.
   */
  public static distanceLineSegmentSquared(px: number, py: number, ax: number, ay: number, bx: number, by: number): number {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'distanceLineSegmentSquared:px');
      Util.checkRealNumber(py, Point, 'distanceLineSegmentSquared:py');
      Util.checkRealNumber(ax, Point, 'distanceLineSegmentSquared:ax');
      Util.checkRealNumber(ay, Point, 'distanceLineSegmentSquared:ay');
      Util.checkRealNumber(bx, Point, 'distanceLineSegmentSquared:bx');
      Util.checkRealNumber(by, Point, 'distanceLineSegmentSquared:by');
    }
    let ux = bx - ax;
    let uy = by - ay;
    const length = ux * ux + uy * uy;
    const vx = ax - px;
    const vy = ay - py;
    const dot = - vx * ux - vy * uy;
    if (dot <= 0 || dot >= length) {  // beyond the segment
      ux = bx - px;
      uy = by - py;
      return Math.min(vx * vx + vy * vy, ux * ux + uy * uy);
    } else {
      const det = ux * vy - uy * vx;
      return (det * det) / length;
    }
  }

  /**
   * This static function returns the square of the distance from the point P to the point Q.
   * @param {number} px
   * @param {number} py
   * @param {number} qx
   * @param {number} qy
   * @return {number}
   */
  public static distanceSquared(px: number, py: number, qx: number, qy: number): number {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'distanceSquared:px');
      Util.checkRealNumber(py, Point, 'distanceSquared:py');
      Util.checkRealNumber(qx, Point, 'distanceSquared:qx');
      Util.checkRealNumber(qy, Point, 'distanceSquared:qy');
    }
    const dx = qx - px;
    const dy = qy - py;
    return dx * dx + dy * dy;
  }

  /**
   * This static function returns the angle in degrees of the line from point P to point Q.
   * @param {number} px
   * @param {number} py
   * @param {number} qx
   * @param {number} qy
   * @return {number}
   */
  public static direction(px: number, py: number, qx: number, qy: number): number {
    if (Debug) {
      Util.checkRealNumber(px, Point, 'direction:px');
      Util.checkRealNumber(py, Point, 'direction:py');
      Util.checkRealNumber(qx, Point, 'direction:qx');
      Util.checkRealNumber(qy, Point, 'direction:qy');
    }
    const x = qx - px;
    const y = qy - py;
    if (x === 0) {
      if (y > 0) {
        return 90;
      } else if (y < 0) {
        return 270;
      } else {
        return 0;
      }
    } else if (y === 0) {
      if (x > 0) {
        return 0;
      } else { // x < 0
        return 180;
      }
    } else {
      if (isNaN(x) || isNaN(y)) return 0;
      let a = Math.atan(Math.abs(y / x)) * 180 / Math.PI;
      if (x < 0) {
        if (y < 0) {
          a += 180;
        } else {
          a = 180 - a;
        }
      } else if (y < 0) {
        a = 360 - a;
      }
      return a;
    }
  }



  /**
   * True if this Point has X and Y values that are real numbers and not infinity.
   * @export
   * @this {Point}
   * @return {boolean}
   */
  public isReal(): boolean {
    return isFinite(this.x) && isFinite(this.y);
  }

  // Used to be on Geo:
  /** @hidden @internal @const */
  public static Origin: Point = null;
  /** @hidden @internal @const */
  public static InfiniteTopLeft: Point = null;
  /** @hidden @internal @const */
  public static InfiniteBottomRight: Point = null;
  /** @hidden @internal @const */
  public static SixPoint: Point = null;
  /** @hidden @internal @const */
  public static NoPoint: Point = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Point['Origin'] = Point.Origin = new Point(0, 0).constant();
    Point['InfiniteTopLeft'] = Point.InfiniteTopLeft = new Point(-Infinity, -Infinity).constant();
    Point['InfiniteBottomRight'] = Point.InfiniteBottomRight = new Point(Infinity, Infinity).constant();
    Point['SixPoint'] = Point.SixPoint = new Point(6, 6).constant();
    Point['NoPoint'] = Point.NoPoint = new Point(NaN, NaN).constant();
    Point['parse'] = Point.parse;
    Point['stringify'] = Point.stringify;
    Point['distanceLineSegmentSquared'] = Point.distanceLineSegmentSquared;
    Point['distanceSquared'] = Point.distanceSquared;
    Point['direction'] = Point.direction;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedPoints: Array<Point> = ([]);

  /**
   * @hidden
   * @expose
   * @return {Point}
   */
  public static alloc(): Point {
    const temp = Point._CachedPoints.pop();
    if (temp === undefined) return new Point();
    return temp;
  }

  /**
   * @hidden
   * @expose
   * @param {number} x
   * @param {number} y
   * @return {Point}
   */
  public static allocAt(x: number, y: number): Point {
    const temp = Point._CachedPoints.pop();
    if (temp === undefined) return new Point(x, y);
    temp.x = x;
    temp.y = y;
    return temp;
  }

  /**
   * @hidden
   * @expose
   * @param {Point} temp
   */
  public static free(temp: Point): void {
    Point._CachedPoints.push(temp);
  }

} // #end Point
Point['className'] = 'Point';

Point['parse'] = Point.parse;
Point['stringify'] = Point.stringify;
Point['distanceLineSegmentSquared'] = Point.distanceLineSegmentSquared;
Point['distanceSquared'] = Point.distanceSquared;
Point['direction'] = Point.direction;
Point.initStatic();

// import { Debug } from './debug';
// import { Geo } from './geo';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Transform } from './transform';
// import { Util } from './util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Size
*/

// import { Debug } from './debug';
// import { Geo } from './geo';
// import { Util } from './util';

/**
 * A Size describes a width and a height in two-dimensional coordinates.
 * The width and height must both be non-negative.
 *
 * Use the static functions {@link Size.parse} and {@link Size.stringify} to convert to and from
 * a standard string representation that is independent of the current locale.
 *
 * When an instance of this class is the value of a property of a {@link GraphObject} class or {@link Diagram}
 * or {@link CommandHandler} or a {@link Tool} class, you should treat the object
 * as if it were frozen or read-only -- you cannot modify its properties.
 * This allows the property to return a value without allocating a new instance.
 * If you need to do your own calculations with the value, call {@link #copy} to make
 * a new instance with the same values that you can modify.
 *
 * Many methods modify the object's properties and then return a reference to "this" object.
 * The only instance method to allocate a new object is the {@link #copy} method.
 * The static {@link Size.parse} method also allocates a new object.
 *
 * The "Debug" implementation of this class is significantly slower than the "Release" implementation,
 * mostly due to additional error checking.
 *
 * You cannot inherit from this class.
 * @category Geometry
 */
export class Size {
  /** @hidden @internal */ public static className = 'Size';
  public width: number;
  public height: number;
  /** @hidden @internal */ private _isFrozen: boolean;

  /**
   * The default constructor produces the Size(0,0).
   * This constructor may take either zero arguments or two arguments.
   * @final
   * @struct
   * @param {number=} w The initial width (must be non-negative).
   * @param {number=} h The initial height (must be non-negative).
   */
  constructor(w?: number, h?: number) {
    if (w === undefined) {
      this.width = 0;
      this.height = 0;
    } else if (typeof w === 'number' && (w >= 0 || isNaN(w)) &&
               typeof h === 'number' && (h >= 0 || isNaN(h))) {
      this.width = w;
      this.height = h;
    } else {
      Util.throwError('Invalid arguments to Size constructor: ' + w + ', ' + h);
    }
    this._isFrozen = false;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Size with those of another Size.
   * This can modify a frozen object.
   * @this {Size}
   * @param {Size} s the other Size from which to copy property values.
   * @return {Size} this Size
   */
  public assign(s: Size): Size {
    this.width = s.width;
    this.height = s.height;
    return this;
  }

  /**
   * @hidden @internal
   * Modify this Size with new Width and Height values.
   * This can modify a frozen object.
   * @this {Size}
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {Size} this.
   */
  public assignTo(w: number, h: number): Size {
    this.width = w;
    this.height = h;
    return this;
  }

  /**
   * Modify this Size with new Width and Height values.
   * @export
   * @this {Size}
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {Size} this.
   */
  public setTo(w: number, h: number): Size {
    if (Debug) {
      Util.checkPrimitive(w, 'number', Size, 'setTo:w');
      Util.checkPrimitive(h, 'number', Size, 'setTo:h');
      if (w < 0) Util.throwRangeError(w, '>= 0', Size, 'setTo:w');
      if (h < 0) Util.throwRangeError(h, '>= 0', Size, 'setTo:h');
      this.checkFrozen();
    }
    this.width = w;
    this.height = h;
    return this;
  }

  /**
   * Modify this Size so that its Width and Height values are the same as the given Size.
   * @export
   * @this {Size}
   * @param {Size} s the given Size.
   * @return {Size} this.
   */
  public set(s: Size): Size {
    if (Debug) {
      Util.checkClass(s, Size, Size, 'set:s');
      this.checkFrozen();
    }
    this.width = s.width;
    this.height = s.height;
    return this;
  }

  /**
   * Create a copy of this Size, with the same values.
   * @expose
   * @this {Size}
   * @return {Size}
   */
  public copy(): Size {
    const copy = new Size();
    copy.width = this.width;
    copy.height = this.height;
    return copy;
  }

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Size}
   * @return {Size}
   */
  public constant(): Size {
    this._isFrozen = true;
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * If this object is 'constant', just return it.
   * Otherwise make a copy of this object and freeze it ({@link #frozen}).
   * @this {Size}
   * @return {Size}
   */
  public copyFrozen(): Size {
    if (this._isFrozen || Object.isFrozen(this)) return this;
    return this.copy().freeze();
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Size}
   * @return {Size}
   */
  public freeze(): Size {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Size}
   * @return {Size}
   */
  public thaw(): Size {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    return this;
  }

  /**
   * When about to modify this object, check that this object is thawed.
   * @hidden @internal
   * @this {Size}
   * @param {number=} val
   */
  public checkFrozen(val?: number): void {
    if (Debug) {
      if (this._isFrozen) {
        let s = 'The Size is frozen, so its properties cannot be set: ' + this.toString();
        if (val !== undefined) s += '  to value: ' + val;
        Util.throwError(s);
      }
    }
  }

  /**
   * This static function can be used to read in a {@link Size} from a string that was produced by {@link Size.stringify}.
   *
   * `go.Size.parse("1 2")` produces the Size `new go.Size(1, 2)`.
   * @param {string} str
   * @return {Size}
   */
  public static parse(str: string): Size {
    if (typeof str === 'string') {
      const split = str.split(' ');
      let i = 0;
      let w = 0.0;
      while (split[i] === '') i++;
      let s = split[i++];
      if (s) w = parseFloat(s);
      let h = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) h = parseFloat(s);
      return new Size(w, h);
    } else {
      return new Size();
    }
  }

  /**
   * This static function can be used to write out a {@link Size} as a string that can be read by {@link Size.parse}.
   *
   * `go.Size.stringify(new go.Size(1, 2))` produces the string "1 2".
   * @param {Size} val
   * @return {string}
   */
  public static stringify(val: Size): string {
    if (Debug) Util.checkClass(val, Size);
    return val.width.toString() + ' ' + val.height.toString();
  }

  /**
   * @hidden @internal
   * @this {Size}
   * @return {string} A string containing this Size's information.
   */
  public toString(): string {
    return 'Size(' + this.width + ',' + this.height + ')';
  }

  /**
   * Indicates whether the given Size is equal to the current Size.
   * @export
   * @this {Size}
   * @param {Size} s The Size to compare to the current Size.
   * @return {boolean} True if the Sizes have the same width and height,
   * false otherwise.
   * @see {@link #equalTo}
   */
  public equals(s: Size): boolean {
    if (!(s instanceof Size)) return false;
    return (this.width === s.width && this.height === s.height);
  }

  /**
   * Indicates whether the given size is equal to this Size.
   * @export
   * @this {Size}
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {boolean} True if the Sizes have the same width and height,
   * false otherwise.
   * @see {@link #equals}
   */
  public equalTo(w: number, h: number): boolean {
    return (this.width === w && this.height === h);
  }

   //    * Indicates whether the given Size is nearly equal to this Size.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Size}
   * @param {Size} s The Size to compare to the current Size.
   * @return {boolean} True if the two Sizes have Width and Height values that
   * are equal with a tolerance of 0.5, false otherwise.
   */
  public equalsApprox(s: Size): boolean {
    return Geo.isApprox(this.width, s.width) &&
      Geo.isApprox(this.height, s.height);
  }



  /**
   * True if this Size has Width and Height values that are real numbers and not infinity.
   * @export
   * @this {Size}
   * @return {boolean}
   */
  public isReal(): boolean {
    return isFinite(this.width) && isFinite(this.height);
  }


  /** @hidden @internal @const */
  public static ZeroSize: Size;
  /** @hidden @internal @const */
  public static OneSize: Size;
  /** @hidden @internal @const */
  public static SixSize: Size;
  /** @hidden @internal @const */
  public static EightSize: Size;
  /** @hidden @internal @const */
  public static TenSize: Size;
  /** @hidden @internal @const */
  public static InfiniteSize: Size;
  /** @hidden @internal @const */
  public static NoSize: Size;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Size['ZeroSize'] = Size.ZeroSize = new Size(0, 0).constant();
    Size['OneSize'] = Size.OneSize = new Size(1, 1).constant();
    Size['SixSize'] = Size.SixSize = new Size(6, 6).constant();
    Size['EightSize'] = Size.EightSize = new Size(8, 8).constant();
    Size['TenSize'] = Size.TenSize = new Size(10, 10).constant();
    Size['InfiniteSize'] = Size.InfiniteSize = new Size(Infinity, Infinity).constant();
    Size['NoSize'] = Size.NoSize = new Size(NaN, NaN).constant();
    Size['parse'] = Size.parse;
    Size['stringify'] = Size.stringify;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedSizes: Array<Size> = [];

  /**
   * @hidden
   * @return {Size}
   */
  public static alloc(): Size {
    const temp = Size._CachedSizes.pop();
    if (temp === undefined) return new Size();
    return temp;
  }

  /**
   * @hidden
   * @param {Size} temp
   */
  public static free(temp: Size): void {
    Size._CachedSizes.push(temp);
  }

} // #end Size
Size['className'] = 'Size';

Size['parse'] = Size.parse;
Size['stringify'] = Size.stringify;
Size.initStatic();


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Rect
*/

// import { Debug } from './debug';
// import { Geo } from './geo';
// import { Margin } from './margin';
// import { Point } from './point';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * A Rect describes a rectangular two-dimensional area as a top-left point (x and y values)
 * and a size (width and height values).
 *
 * Use the static functions {@link Rect.parse} and {@link Rect.stringify} to convert to and from
 * a standard string representation that is independent of the current locale.
 *
 * When an instance of this class is the value of a property of a {@link GraphObject} class or {@link Diagram}
 * or {@link CommandHandler} or a {@link Tool} class, you should treat the object
 * as if it were frozen or read-only -- you cannot modify its properties.
 * This allows the property to return a value without allocating a new instance.
 * If you need to do your own calculations with the value, call {@link #copy} to make
 * a new instance with the same values that you can modify.
 *
 * Many methods modify the object's properties and then return a reference to "this" object.
 * The only instance method to allocate a new object is the {@link #copy} method.
 * The static {@link Rect.parse} method also allocates a new object.
 * The {@link #center}, {@link #position}, and {@link #size} properties all allocate and return a new object.
 *
 * The "Debug" implementation of this class is significantly slower than the "Release" implementation,
 * mostly due to additional error checking.
 *
 * You cannot inherit from this class.
 * @category Geometry
 */
export class Rect {
  /** @hidden @internal */ public static className = 'Rect';
  public x: number;
  public y: number;
  public width: number;
  public height: number;
  /** @hidden @internal */ private _isFrozen: boolean;

  /**
   * There are four constructors: (), (Point, Point), (Point, Size),
   * and four numbers for (x, y, width, height).
   * The default constructor (no argument) results in a Rect(0,0,0,0).
   * @final
   * @struct
   * @param {Point|number=} x Top-left Point, or x value.
   * @param {Point|Size|number=} y Bottom-right Point or Size or y value.
   * @param {number=} w Width to be used if x,y are specified;
   * must be non-negative.
   * @param {number=} h Height to be used if x,y are specified;
   * must be non-negative.
   */
  constructor(x?: Point | number, y?: Point | Size | number, w?: number, h?: number) {
    if (x === undefined) {  // zero args
      this.x = 0;
      this.y = 0;
      this.width = 0;
      this.height = 0;
    } else {
      if (x instanceof Point) {
        const xx = x.x;
        const xy = x.y;
        if (y instanceof Point) {  // Point, Point
          const yx = y.x;
          const yy = y.y;
          this.x = Math.min(xx, yx);
          this.y = Math.min(xy, yy);
          this.width = Math.abs(xx - yx);
          this.height = Math.abs(xy - yy);
        } else if (y instanceof Size) {  // Point, Size
          this.x = xx;
          this.y = xy;
          this.width = y.width;
          this.height = y.height;
        } else {
          Util.throwError('Incorrect arguments supplied to Rect constructor');
        }
      } else if (typeof x === 'number' &&
        typeof y === 'number' &&
        typeof w === 'number' && (w >= 0 || isNaN(w)) &&
        typeof h === 'number' && (h >= 0 || isNaN(h))) {  // x, y, width, height
        this.x = x;
        this.y = y;
        this.width = w;
        this.height = h;
      } else {
        Util.throwError('Invalid arguments to Rect constructor: ' + x + ', ' + y + ', ' + w + ', ' + h);
      }
    }
    this._isFrozen = false;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Rect with those of another Rect.
   * This can modify a frozen object.
   * @this {Rect}
   * @param {Rect} r the other Rect from which to copy property values.
   * @return {Rect} this Rect
   */
  public assign(r: Rect): Rect {
    this.x = r.x;
    this.y = r.y;
    this.width = r.width;
    this.height = r.height;
    return this;
  }

  /**
   * @hidden @internal
   * Modify this Rect with new X, Y, Width, and Height values.
   * This can modify a frozen object.
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {Rect} this.
   */
  public assignTo(x: number, y: number, w: number, h: number): Rect {
    this.x = x;
    this.y = y;
    this.width = w;
    this.height = h;
    return this;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Rect with given values.
   * This can modify a frozen object.
   * @this {Rect}
   * @param {number} w
   * @param {number} h
   * @return {Rect} this Rect
   */
  public assignWidthHeight(w: number, h: number): Rect {
    this.width = w;
    this.height = h;
    return this;
  }

  /**
   * Modify this Rect with new X, Y, Width, and Height values.
   * @export
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {Rect} this.
   */
  public setTo(x: number, y: number, w: number, h: number): Rect {
    if (Debug) {
      Util.checkPrimitive(x, 'number', Rect, 'setTo:x');
      Util.checkPrimitive(y, 'number', Rect, 'setTo:y');
      Util.checkPrimitive(w, 'number', Rect, 'setTo:w');
      Util.checkPrimitive(h, 'number', Rect, 'setTo:h');
      if (w < 0) Util.throwRangeError(w, '>= 0', Rect, 'setTo:w');
      if (h < 0) Util.throwRangeError(h, '>= 0', Rect, 'setTo:h');
      this.checkFrozen();
    }
    this.x = x;
    this.y = y;
    this.width = w;
    this.height = h;
    return this;
  }

  /**
   * Modify this Rect so that its X, Y, Width, and Height values are the same as the given Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r the given Rect.
   * @return {Rect} this.
   */
  public set(r: Rect): Rect {
    if (Debug) {
      Util.checkClass(r, Rect, Rect, 'set:r');
      this.checkFrozen();
    }
    this.x = r.x;
    this.y = r.y;
    this.width = r.width;
    this.height = r.height;
    return this;
  }

  /**
   * Modify this Rect so that its X and Y values are the same as the given Point.
   * @export
   * @this {Rect}
   * @param {Point} p the given Point.
   * @return {Rect} this.
   */
  public setPoint(p: Point): Rect {
    if (Debug) {
      Util.checkClass(p, Point, Rect, 'setPoint:p');
      this.checkFrozen();
    }
    this.x = p.x;
    this.y = p.y;
    return this;
  }

  /**
   * Modify this Rect so that its Width and Height values are the same as the given Size.
   * @export
   * @this {Rect}
   * @param {Size} s the given Size.
   * @return {Rect} this.
   */
  public setSize(s: Size): Rect {
    if (Debug) {
      Util.checkClass(s, Size, Rect, 'setSize:s');
      this.checkFrozen();
    }
    this.width = s.width;
    this.height = s.height;
    return this;
  }

  /**
   * Create a copy of this Rect, with the same values.
   * @expose
   * @this {Rect}
   * @return {Rect}
   */
  public copy(): Rect {
    const copy = new Rect();
    copy.x = this.x;
    copy.y = this.y;
    copy.width = this.width;
    copy.height = this.height;
    return copy;
  }

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Rect}
   * @return {Rect}
   */
  public constant(): Rect {
    this._isFrozen = true;
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * If this object is 'constant', just return it.
   * Otherwise make a copy of this object and freeze it ({@link #frozen}).
   * @this {Rect}
   * @return {Rect}
   */
  public copyFrozen(): Rect {
    if (this._isFrozen || Object.isFrozen(this)) return this;
    return this.copy().freeze();
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Rect}
   * @return {Rect}
   */
  public freeze(): Rect {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Rect}
   * @return {Rect}
   */
  public thaw(): Rect {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    return this;
  }

  /**
   * When about to modify this object, check that this object is thawed.
   * @hidden @internal
   * @this {Rect}
   * @param {*=} val
   */
  public checkFrozen(val?: any): void {
    if (Debug) {
      if (this._isFrozen) {
        let s = 'The Rect is frozen, so its properties cannot be set: ' + this.toString();
        if (val !== undefined) s += '  to value: ' + val;
        Util.throwError(s);
      }
    }
  }

  /**
   * This static function can be used to read in a {@link Rect} from a string that was produced by {@link Rect.stringify}.
   *
   * `go.Rect.parse("1 2 3 4")` produces the Rect `new go.Rect(1, 2, 3, 4)`.
   * @param {string} str
   * @return {Rect}
   */
  public static parse(str: string): Rect {
    if (typeof str === 'string') {
      const split = str.split(' ');
      let i = 0;
      let x = 0.0;
      while (split[i] === '') i++;
      let s = split[i++];
      if (s) x = parseFloat(s);
      let y = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) y = parseFloat(s);
      let w = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) w = parseFloat(s);
      let h = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) h = parseFloat(s);
      return new Rect(x, y, w, h);
    } else {
      return new Rect();
    }
  }

  /**
   * This static function can be used to write out a {@link Rect} as a string that can be read by {@link Rect.parse}.
   *
   * `go.Rect.stringify(new go.Rect(1, 2, 3, 4))` produces the string "1 2 3 4".
   * @param {Rect} val
   * @return {string}
   */
  public static stringify(val: Rect): string {
    if (Debug) Util.checkClass(val, Rect);
    return val.x.toString() + ' ' + val.y.toString() + ' ' + val.width.toString() + ' ' + val.height.toString();
  }

  /**
   * @hidden @internal
   * @this {Rect}
   * @return {string} A string containing Rect's information.
   */
  public toString(): string {
    return 'Rect(' + this.x + ',' + this.y + ',' + this.width + ',' + this.height + ')';
  }

  /**
   * Indicates whether the given Rect is equal to the current Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r The rectangle to compare to the current rectangle.
   * @return {boolean} True if the Rects are equivalent in x, y, width,
   * and height.
   * @see {@link #equalTo}
   */
  public equals(r: Rect): boolean {
    if (!(r instanceof Rect)) return false;
    return (this.x === r.x && this.y === r.y &&
      this.width === r.width && this.height === r.height);
  }

  /**
   * Indicates whether the given Rect is equal to the current Rect.
   * @export
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w the width.
   * @param {number} h the height.
   * @return {boolean} True if the Rects are equivalent in x, y, width, and height.
   * @see {@link #equals}
   */
  public equalTo(x: number, y: number, w: number, h: number): boolean {
    return (this.x === x && this.y === y &&
      this.width === w && this.height === h);
  }

  //    * Indicates whether the given Rect is nearly equal to this Rect.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Rect}
   * @param {Rect} r The Rect to compare to the current Rect.
   * @return {boolean} True if the two Rects have respective X, Y, Width,
   * and Height values within 0.5, false otherwise.
   */
  public equalsApprox(r: Rect): boolean {
    return Geo.isApprox(this.x, r.x) &&
      Geo.isApprox(this.y, r.y) &&
      Geo.isApprox(this.width, r.width) &&
      Geo.isApprox(this.height, r.height);
  }

  /**
   * @hidden @internal
   * Indicates whether the given Rect is very nearly equal to this Rect.
   * @this {Rect}
   * @param {Rect} r The Rect to compare to the current Rect.
   * @return {boolean} True if the two Rects have respective X, Y, Width,
   * and Height values within 0.00000005, false otherwise.
   */
  public equalsApproxClose(r: Rect): boolean {
    return Geo.isApproxEqual(this.x, r.x) &&
      Geo.isApproxEqual(this.y, r.y) &&
      Geo.isApproxEqual(this.width, r.width) &&
      Geo.isApproxEqual(this.height, r.height);
  }

  /**
   * Indicates whether this Rect contains the given Point.
   * @export
   * @this {Rect}
   * @param {Point} p The Point to check.
   * @return {boolean} True if the Point is contained within this Rect,
   * false otherwise.
   * @see {@link #containsRect}
   * @see {@link #contains}
   */
  public containsPoint(p: Point): boolean {
    if (Debug) Util.checkClass(p, Point, Rect, 'containsPoint:p');
    return ((this.x <= p.x) && ((this.x + this.width) >= p.x) &&
      (this.y <= p.y) && ((this.y + this.height) >= p.y));
  }

  /**
   * Indicates whether this Rect contains the given Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r The Rect to check.
   * @return {boolean} True if the Rect is contained within this Rect,
   * false otherwise.
   * @see {@link #containsPoint}
   * @see {@link #contains}
   */
  public containsRect(r: Rect): boolean {
    if (Debug) Util.checkClass(r, Rect, Rect, 'containsRect:r');
    return (this.x <= r.x && r.x + r.width <= this.x + this.width &&
      this.y <= r.y && r.y + r.height <= this.y + this.height);
  }

  /**
   * Indicates whether this Rect contains the given Point/Rect.
   * @export
   * @this {Rect}
   * @param {number} x The X coordinate of the Point or Rect to include in the new bounds.
   * @param {number} y The Y coordinate of the Point or Rect to include in the new bounds.
   * @param {number=} w The Width of the Rect to include in the new bounds, defaults to zero.
   * @param {number=} h The Height of the Rect to include in the new bounds, defaults to zero.
   * @return {boolean} True if the Point/Rect is contained within this Rect,
   * false otherwise.
   * @see {@link #containsRect}
   * @see {@link #containsPoint}
   */
  public contains(x: number, y: number, w?: number, h?: number): boolean {
    if (Debug) {
      Util.checkRealNumber(x, Rect, 'contains:x');
      Util.checkRealNumber(y, Rect, 'contains:y');
      if (w === undefined) {
        w = 0;
      } else {
        Util.checkRealNumber(w, Rect, 'contains:w');
      }
      if (h === undefined) {
        h = 0;
      } else {
        Util.checkRealNumber(h, Rect, 'contains:h');
      }
      if (w < 0 || h < 0) {
        Util.throwError('Rect.contains:Width and height cannot be negative');
      }
    } else {
      if (w === undefined) w = 0;
      if (h === undefined) h = 0;
    }
    return (this.x <= x && x + w <= this.x + this.width &&
      this.y <= y && y + h <= this.y + this.height);
  }

  /**
   * @hidden @internal
   * Modify this Rect so that its X, Y, Width, and Height values are all zero.
   * @this {Rect}
   */
  public reset(): void {
    if (Debug) this.checkFrozen();
    this.x = 0;
    this.y = 0;
    this.width = 0;
    this.height = 0;
  }

  /**
   * Modify this Rect by shifting its values with the given DX and DY offsets.
   * @export
   * @this {Rect}
   * @param {number} dx
   * @param {number} dy
   * @return {Rect} this.
   */
  public offset(dx: number, dy: number): Rect {
    if (Debug) {
      Util.checkRealNumber(dx, Rect, 'offset:dx');
      Util.checkRealNumber(dy, Rect, 'offset:dy');
      this.checkFrozen();
    }
    this.x += dx;
    this.y += dy;
    return this;
  }

  /**
   * Modify this Rect so that its width and height are changed on all four sides,
   * equally on the left and right sides, and equally on the top and bottom sides.
   * When the arguments are negative, this operation deflates this Rect, but not beyond zero.
   * @export
   * @this {Rect}
   * @param {number} w The additional width on each side, left and right; may be negative.
   * @param {number} h The additional height on each side, top and bottom; may be negative.
   * @return {Rect} this.
   * @see {@link #grow}
   * @see {@link #addMargin}
   * @see {@link #subtractMargin}
   */
  public inflate(w: number, h: number): Rect {
    if (Debug) {
      Util.checkRealNumber(w, Rect, 'inflate:w');
      Util.checkRealNumber(h, Rect, 'inflate:h');
    }
    return this._grow(h, w, h, w);
  }

  /**
   * Modify this Rect by adding the given Margin to each side of the Rect.
   * @export
   * @this {Rect}
   * @param {Margin} m The Margin to add to the Rect.
   * @return {Rect} this bigger Rect.
   * @see {@link #subtractMargin}
   * @see {@link #grow}
   * @see {@link #inflate}
   */
  public addMargin(m: Margin): Rect {
    if (Debug) Util.checkClass(m, Margin, Rect, 'addMargin:m');
    return this._grow(m.top, m.right, m.bottom, m.left);
  }

  /**
   * Modify this Rect by subtracting the given Margin from each side of the Rect.
   * @export
   * @this {Rect}
   * @param {Margin} m The Margin to subtract from the Rect.
   * @return {Rect} this smaller Rect.
   * @see {@link #addMargin}
   * @see {@link #grow}
   * @see {@link #inflate}
   */
  public subtractMargin(m: Margin): Rect {
    if (Debug) Util.checkClass(m, Margin, Rect, 'subtractMargin:m');
    return this._grow(-m.top, -m.right, -m.bottom, -m.left);
  }

  /**
   * Modifies this Rect by adding some distance to each side of the Rect.
   * @export
   * @this {Rect}
   * @param {number} t the amount to move the top side upwards; may be negative.
   * @param {number} r the amount to move the right side rightwards; may be negative.
   * @param {number} b the amount to move the bottom side downwards; may be negative.
   * @param {number} l the amount to move the left side leftwards; may be negative.
   * @return {Rect} this modified Rect.
   * @see {@link #inflate}
   * @see {@link #addMargin}
   * @see {@link #subtractMargin}
   */
  public grow(t: number, r: number, b: number, l: number): Rect {
    if (Debug) {
      Util.checkRealNumber(t, Rect, 'grow:t');
      Util.checkRealNumber(r, Rect, 'grow:r');
      Util.checkRealNumber(b, Rect, 'grow:b');
      Util.checkRealNumber(l, Rect, 'grow:l');
    }
    return this._grow(t, r, b, l);
  }

  /**
   * @hidden @internal
   * Modifies this Rect by adding some distance to each side of the Rect.
   * @this {Rect}
   * @param {number} t the amount to move the top side upwards; may be negative.
   * @param {number} r the amount to move the right side rightwards; may be negative.
   * @param {number} b the amount to move the bottom side downwards; may be negative.
   * @param {number} l the amount to move the left side leftwards; may be negative.
   * @return {Rect} this modified Rect.
   */
  private _grow(t: number, r: number, b: number, l: number): Rect {
    if (Debug) this.checkFrozen();
    const w = this.width;
    if (r + l <= -w) {
      this.x += w / 2;
      this.width = 0;
    } else {
      this.x -= l;
      this.width += r + l;
    }
    const h = this.height;
    if (t + b <= -h) {
      this.y += h / 2;
      this.height = 0;
    } else {
      this.y -= t;
      this.height += t + b;
    }
    return this;
  }

  /**
   * Modify this Rect so that it is the intersection of this Rect and the given Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r Rect to intersect with.
   * @return {Rect} this.
   * @see {@link #intersect}
   * @see {@link #intersects}
   */
  public intersectRect(r: Rect): Rect {
    if (Debug) Util.checkClass(r, Rect, Rect, 'intersectRect:r');
    return this._intersect(r.x, r.y, r.width, r.height);
  }

  /**
   * Modify this Rect so that it is the intersection of this Rect and the rectangle
   * defined by x, y, w, h.
   * @export
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {Rect} this.
   * @see {@link #intersectRect}
   * @see {@link #intersects}
   */
  public intersect(x: number, y: number, w: number, h: number): Rect {
    if (Debug) {
      Util.checkRealNumber(x, Rect, 'intersect:x');
      Util.checkRealNumber(y, Rect, 'intersect:y');
      Util.checkRealNumber(w, Rect, 'intersect:w');
      Util.checkRealNumber(h, Rect, 'intersect:h');
      if (w < 0 || h < 0) {
        Util.throwError('Rect.intersect:Width and height cannot be negative');
      }
    }
    return this._intersect(x, y, w, h);
  }

  /**
   * @hidden @internal
   * Modifies this Rect so that it is the intersection of this Rect and the given Rect.
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {Rect}
   */
  private _intersect(x: number, y: number, w: number, h: number): Rect {
    if (Debug) this.checkFrozen();
    // ??? should this returns a special Empty value if there's no intersection
    const maxx = Math.max(this.x, x);
    const maxy = Math.max(this.y, y);
    const minr = Math.min(this.x + this.width, x + w);
    const minb = Math.min(this.y + this.height, y + h);
    this.x = maxx;
    this.y = maxy;
    this.width = Math.max(0, minr - maxx);
    this.height = Math.max(0, minb - maxy);
    return this;
  }

  /**
   * Determine if a given Rect is partly or wholly inside of this Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r Rect to test intersection with.
   * @return {boolean} true if there is an intersection.
   * @see {@link #intersects}
   * @see {@link #intersectRect}
   */
  public intersectsRect(r: Rect): boolean {
    if (Debug) Util.checkClass(r, Rect, Rect, 'intersectsRect:r');
    return this.intersects(r.x, r.y, r.width, r.height);
  }

  /**
   * Determine if this Rect partly or wholly overlaps the rectangle
   * defined by x, y, w, h.
   * @export
   * @this {Rect}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {boolean} true if there is any overlap.
   * @see {@link #intersectsRect}
   * @see {@link #intersect}
   */
  public intersects(x: number, y: number, w: number, h: number): boolean {
    if (Debug) {
      Util.checkRealNumber(x, Rect, 'intersects:x');
      Util.checkRealNumber(y, Rect, 'intersects:y');
      Util.checkRealNumber(x, Rect, 'intersects:w');
      Util.checkRealNumber(y, Rect, 'intersects:h');
      if (w < 0 || h < 0) {
        Util.throwError('Rect.intersects:Width and height cannot be negative');
      }
    }

    let tw = this.width;
    const tx = this.x;
    if (tw !== Infinity && w !== Infinity) {
      tw += tx;
      w += x;
      if (isNaN(w) || isNaN(tw) || tx > w || x > tw) return false;
    }
    let th = this.height;
    const ty = this.y;
    if (th !== Infinity && h !== Infinity) {
      th += ty;
      h += y;
      if (isNaN(h) || isNaN(th) || ty > h || y > th) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * Faster with added-in rect expansion
   * Assumes no NaN or Infinity values
   * @this {Rect}
   * @param {Rect} r Rect to test intersection with.
   * @param {number} expansion to expand r on all sides. Can be zero.
   * @return {boolean} true if there is an intersection.
   */
  public intersectsRectPlus(r: Rect, expansion: number): boolean {
    let tw = this.width;
    let rw = r.width + expansion + expansion;
    const tx = this.x;
    const rx = r.x - expansion;
    tw += tx;
    rw += rx;
    if (tx > rw || rx > tw) return false;

    let th = this.height;
    let rh = r.height + expansion + expansion;
    const ty = this.y;
    const ry = r.y - expansion;
    th += ty;
    rh += ry;
    if (ty > rh || ry > th) return false;
    return true;
  }

  /**
   * Modify this Rect to be exactly big enough to contain both the original Rect and the given Point.
   * @export
   * @this {Rect}
   * @param {Point} p The Point to include in the new bounds.
   * @return {Rect} this.
   * @see {@link #unionRect}
   * @see {@link #union}
   */
  public unionPoint(p: Point): Rect {
    if (Debug) Util.checkClass(p, Point, Rect, 'unionPoint:p');
    return this.unionNoCheck(p.x, p.y, 0, 0);
  }

  /**
   * Modify this Rect to be exactly big enough to contain this Rect and the given Rect.
   * @export
   * @this {Rect}
   * @param {Rect} r The Rect to include in the new bounds.
   * @return {Rect} this.
   * @see {@link #unionPoint}
   * @see {@link #union}
   */
  public unionRect(r: Rect): Rect {
    if (Debug) Util.checkClass(r, Rect, Rect, 'unionRect:r');
    return this.unionNoCheck(r.x, r.y, r.width, r.height);
  }

  /**
   * Modify this Rect to be exactly big enough to contain both the original Rect and the given rectangular area.
   * @export
   * @this {Rect}
   * @param {number} x The X coordinate of the Point or Rect to include in the new bounds.
   * @param {number} y The Y coordinate of the Point or Rect to include in the new bounds.
   * @param {number=} w The Width of the Rect to include in the new bounds, defaults to zero.
   * @param {number=} h The Height of the Rect to include in the new bounds, defaults to zero.
   * @return {Rect} this.
   * @see {@link #unionRect}
   * @see {@link #unionPoint}
   */
  public union(x: number, y: number, w?: number, h?: number): Rect {
    if (Debug) {
      Util.checkRealNumber(x, Rect, 'union:x');
      Util.checkRealNumber(y, Rect, 'union:y');
      if (w === undefined) {
        w = 0;
      } else {
        Util.checkRealNumber(w, Rect, 'union:w');
      }
      if (h === undefined) {
        h = 0;
      } else {
        Util.checkRealNumber(h, Rect, 'union:h');
      }
      if (w < 0 || h < 0) {
        Util.throwError('Rect.union:Width and height cannot be negative');
      }
      this.checkFrozen();
    } else {
      if (w === undefined) w = 0;
      if (h === undefined) h = 0;
    }
    return this.unionNoCheck(x, y, w, h);
  }

  /**
   * @hidden @internal
   * Modify this Rect to be exactly big enough to contain both the original Rect and the given rectangular area.
   * @this {Rect}
   * @param {number} x The X coordinate of the Point or Rect to include in the new bounds.
   * @param {number} y The Y coordinate of the Point or Rect to include in the new bounds.
   * @param {number} w The Width of the Rect to include in the new bounds, defaults to zero.
   * @param {number} h The Height of the Rect to include in the new bounds, defaults to zero.
   * @return {Rect} this.
   */
  public unionNoCheck(x: number, y: number, w: number, h: number): Rect {
    const minx = Math.min(this.x, x);
    const miny = Math.min(this.y, y);
    const maxr = Math.max(this.x + this.width, x + w);
    const maxb = Math.max(this.y + this.height, y + h);
    this.x = minx;
    this.y = miny;
    this.width = maxr - minx;
    this.height = maxb - miny;
    return this;
  }

  /**
   * Modify this Rect so that a given Spot is at a given (x,y) point using this Rect's size.
   * Return this rectangle for which the spot is at that point, without modifying the size.
   *
   * The result is meaningless if {@link Spot#isNoSpot} is true.
   * @export
   * @this {Rect}
   * @param {number} x the point where the spot should be.
   * @param {number} y the point where the spot should be.
   * @param {Spot} spot a Spot; {@link Spot#isSpot} must be true.
   * @return {Rect} this.
   */
  public setSpot(x: number, y: number, spot: Spot): Rect {
    if (Debug) {
      Util.checkRealNumber(x, Rect, 'setSpot:x');
      Util.checkRealNumber(y, Rect, 'setSpot:y');
      Util.checkClass(spot, Spot, Rect, 'setSpot:spot');
      this.checkFrozen();
    }
    this.x = x - spot.offsetX - spot.x * this.width;
    this.y = y - spot.offsetY - spot.y * this.height;
    return this;
  }

  /**
   * This static function indicates whether a Rect contains the given Point/Rect.
   * @param {number} rx The X coordinate of a Rect.
   * @param {number} ry The Y coordinate of a Rect.
   * @param {number} rw The Width of a Rect.
   * @param {number} rh The Height of a Rect.
   * @param {number} x The X coordinate of the Point or Rect that might be in the bounds (RX, RY, RW, RH).
   * @param {number} y The Y coordinate of the Point or Rect that might be in the bounds (RX, RY, RW, RH).
   * @param {number=} w The Width of the Rect to include in the new bounds, defaults to zero.
   * @param {number=} h The Height of the Rect to include in the new bounds, defaults to zero.
   * @return {boolean} True if the Point/Rect is contained within this Rect,
   * false otherwise.
   */
  public static contains(rx: number, ry: number, rw: number, rh: number, x: number, y: number, w?: number, h?: number): boolean {
    if (Debug) {
      Util.checkRealNumber(rx, Rect, 'contains:rx');
      Util.checkRealNumber(ry, Rect, 'contains:ry');
      Util.checkRealNumber(rw, Rect, 'contains:rw');
      Util.checkRealNumber(rh, Rect, 'contains:rh');
      Util.checkRealNumber(x, Rect, 'contains:x');
      Util.checkRealNumber(y, Rect, 'contains:y');
      if (w === undefined) {
        w = 0;
      } else {
        Util.checkRealNumber(w, Rect, 'contains:w');
      }
      if (h === undefined) {
        h = 0;
      } else {
        Util.checkRealNumber(h, Rect, 'contains:h');
      }
      if (rw < 0 || rh < 0 || w < 0 || h < 0) {
        Util.throwError('Rect.contains:Width and height cannot be negative');
      }
    } else {
      if (w === undefined) w = 0;
      if (h === undefined) h = 0;
    }
    return (rx <= x && x + w <= rx + rw &&
      ry <= y && y + h <= ry + rh);
  }

  /**
   * This static function indicates whether a Rect partly or wholly overlaps the given Rect.
   * @param {number} rx The X coordinate of a Rect.
   * @param {number} ry The Y coordinate of a Rect.
   * @param {number} rw The Width of a Rect.
   * @param {number} rh The Height of a Rect.
   * @param {number} x The X coordinate of the Point or Rect that might overlap the bounds (RX, RY, RW, RH).
   * @param {number} y The Y coordinate of the Point or Rect that might overlap the bounds (RX, RY, RW, RH).
   * @param {number} w
   * @param {number} h
   * @return {boolean} true if there is any overlap.
   */
  public static intersects(rx: number, ry: number, rw: number, rh: number, x: number, y: number, w: number, h: number): boolean {
    if (Debug) {
      Util.checkRealNumber(rx, Rect, 'intersects:rx');
      Util.checkRealNumber(ry, Rect, 'intersects:ry');
      Util.checkRealNumber(rw, Rect, 'intersects:rw');
      Util.checkRealNumber(rh, Rect, 'intersects:rh');
      Util.checkRealNumber(x, Rect, 'intersects:x');
      Util.checkRealNumber(y, Rect, 'intersects:y');
      Util.checkRealNumber(w, Rect, 'intersects:w');
      Util.checkRealNumber(h, Rect, 'intersects:h');
      if (rw < 0 || rh < 0 || w < 0 || h < 0) {
        Util.throwError('Rect.intersects:Width and height cannot be negative');
      }
    }

    let tw = rw;
    let uw = w;
    const tx = rx;
    const ux = x;
    tw += tx;
    uw += ux;
    if (tx > uw || ux > tw) return false;

    let th = rh;
    let uh = h;
    const ty = ry;
    const uy = y;
    th += ty;
    uh += uy;
    if (ty > uh || uy > th) return false;
    return true;
  }



  /**
   * Gets or sets the leftmost value of the Rect.
   * This is the same as the X value.
   */
  /** @export */ get left(): number { return this.x; }
  /** @export */ set left(value) {
    if (Debug) {
      Util.checkPrimitive(value, 'number', Rect, 'left');
      this.checkFrozen(value);
    }
    this.x = value;
  }

  /**
   * Gets or sets the topmost value of the Rect.
   * This is the same as the Y value.
   */
  /** @export */ get top(): number { return this.y; }
  /** @export */ set top(value) {
    if (Debug) {
      Util.checkPrimitive(value, 'number', Rect, 'top');
      this.checkFrozen(value);
    }
    this.y = value;
  }

  /**
   * Gets or sets the x-axis value of the right of the Rect.
   * This is equal to the sum of the x value and the width.
   */
  /** @export */ get right(): number { return this.x + this.width; }
  /** @export */ set right(value) {
    if (Debug) {
      Util.checkRealNumber(value, Rect, 'right');
      this.checkFrozen(value);
    }
    this.x += (value - (this.x + this.width));
  }

  /**
   * Gets or sets the y-axis value of the bottom of the Rect.
   * This is equal to the sum of the y value and the height.
   */
  /** @export */ get bottom(): number { return this.y + this.height; }
  /** @export */ set bottom(value) {
    if (Debug) {
      Util.checkRealNumber(value, Rect, 'top');
      this.checkFrozen(value);
    }
    this.y += (value - (this.y + this.height));
  }

  /**
   * Gets or sets the x- and y-axis position of the Rect as a Point.
   */
  /** @export */ get position(): Point { return new Point(this.x, this.y); }
  /** @export */ set position(value) {
    if (Debug) {
      Util.checkClass(value, Point, Rect, 'position');
      this.checkFrozen(value);
    }
    this.x = value.x;
    this.y = value.y;
  }

  /**
   * Gets or sets the width and height of the Rect as a Size.
   */
  /** @export */ get size(): Size { return new Size(this.width, this.height); }
  /** @export */ set size(value) {
    if (Debug) {
      Util.checkClass(value, Size, Rect, 'size');
      this.checkFrozen(value);
    }
    this.width = value.width;
    this.height = value.height;
  }

  /**
   * Gets or sets the Point at the center of this Rect.
   * Setting this property just shifts the X and Y values.
   */
  /** @export */ get center(): Point { return new Point(this.x + this.width / 2, this.y + this.height / 2); }
  /** @export */ set center(value) {
    if (Debug) {
      Util.checkClass(value, Point, Rect, 'center');
      this.checkFrozen(value);
    }
    this.x = value.x - this.width / 2;
    this.y = value.y - this.height / 2;
  }

  /**
   * Gets or sets the horizontal center X coordinate of the Rect.
   */
  /** @export */ get centerX(): number { return this.x + this.width / 2; }
  /** @export */ set centerX(value) {
    if (Debug) {
      Util.checkRealNumber(value, Rect, 'centerX');
      this.checkFrozen(value);
    }
    this.x = value - this.width / 2;
  }

  /**
   * Gets or sets the vertical center Y coordinate of the Rect.
   */
  /** @export */ get centerY(): number { return this.y + this.height / 2; }
  /** @export */ set centerY(value) {
    if (Debug) {
      Util.checkRealNumber(value, Rect, 'centerY');
      this.checkFrozen(value);
    }
    this.y = value - this.height / 2;
  }

  /**
   * True if this Rect has X, Y, Width, and Height values that are real numbers and not infinity.
   * @export
   * @this {Rect}
   * @return {boolean}
   */
  public isReal(): boolean {
    return isFinite(this.x) && isFinite(this.y) && isFinite(this.width) && isFinite(this.height);
  }

  /**
   * True if this Rect has a Width and Height of zero.
   * @export
   * @this {Rect}
   * @return {boolean}
   */
  public isEmpty(): boolean {
    return this.width === 0 && this.height === 0;
  }

  /** @hidden @internal @const */
  public static ZeroRect: Rect = null;
  /** @hidden @internal @const */
  public static NoRect: Rect = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Rect['ZeroRect'] = Rect.ZeroRect = new Rect(0, 0, 0, 0).constant();
    Rect['NoRect'] = Rect.NoRect = new Rect(NaN, NaN, NaN, NaN).constant();
    Rect['parse'] = Rect.parse;
    Rect['stringify'] = Rect.stringify;
    Rect['contains'] = Rect.contains;
    Rect['intersects'] = Rect.intersects;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedRects: Array<Rect> = ([]);

  /**
   * @hidden
   * @return {Rect}
   */
  public static alloc(): Rect {
    const temp = Rect._CachedRects.pop();
    if (temp === undefined) return new Rect();
    return temp;
  }

  /**
   * @hidden
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {Rect}
   */
  public static allocAt(x: number, y: number, w: number, h: number): Rect {
    const temp = Rect._CachedRects.pop();
    if (temp === undefined) return new Rect(x, y, w, h);
    return temp.assignTo(x, y, w, h);
  }

  /**
   * @hidden
   * @param {Rect} temp
   */
  public static free(temp: Rect): void {
    Rect._CachedRects.push(temp);
  }

} // #end Rect
Rect['className'] = 'Rect';

Rect['parse'] = Rect.parse;
Rect['stringify'] = Rect.stringify;
Rect['contains'] = Rect.contains;
Rect['intersects'] = Rect.intersects;
Rect.initStatic();


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Margin
*/

// import { Debug } from './debug';
// import { Geo } from './geo';
// import { Util } from './util';

/**
 * The MarginLike type is the same as `Margin | number`.
 * A number in place of a Margin object is treated as a uniform Margin with that thickness.
 * @category Type
 */
export type MarginLike = Margin | number;

/**
 * A Margin represents a band of space outside or inside a rectangular area,
 * with possibly different values on each of the four sides.
 *
 * Example uses include {@link GraphObject#margin}, {@link Panel#padding}, and {@link Diagram#padding}.
 *
 * Use the static functions {@link Margin.parse} and {@link Margin.stringify} to convert to and from
 * a standard string representation that is independent of the current locale.
 *
 * When an instance of this class is the value of a property of a {@link GraphObject} class or {@link Diagram}
 * or {@link CommandHandler} or a {@link Tool} class, you should treat the object
 * as if it were frozen or read-only -- you cannot modify its properties.
 * This allows the property to return a value without allocating a new instance.
 * If you need to do your own calculations with the value, call {@link #copy} to make
 * a new instance with the same values that you can modify.
 *
 * Many methods modify the object's properties and then return a reference to "this" object.
 * The only instance method to allocate a new object is the {@link #copy} method.
 * The static {@link Margin.parse} method also allocates a new object.
 *
 * The "Debug" implementation of this class is significantly slower than the "Release" implementation,
 * mostly due to additional error checking.
 *
 * You cannot inherit from this class.
 * @category Geometry
 */
export class Margin {
  /** @hidden @internal */ public static className = 'Margin';
  public top: number;
  public right: number;
  public bottom: number;
  public left: number;
  /** @hidden @internal */ private _isFrozen: boolean;

  /**
   * If zero arguments are supplied, zero is used for all four sides.
   * If one argument is supplied, that number is used for all four sides.
   * If two arguments are supplied, the top and bottom sides get the first value,
   * and the left and right sides get the second value.
   * Otherwise there must be four arguments, in the order: top, right, bottom, left.
   * @final
   * @struct
   * @param {number=} t the margin for the top side;
   *                    if not supplied, all sides are zero.
   * @param {number=} r the margin for the right side;
   *                    if not supplied, all sides have the value of the first argument.
   * @param {number=} b the margin for all bottom side;
   *                    if not supplied, the top and bottom get the value of the first argument,
   *                    and the right and left sides get the value of the second argument.
   * @param {number=} l the margin for the left side;
   *                    must be supplied if the third argument was supplied.
   */
  constructor(t?: number, r?: number, b?: number, l?: number) {
    if (t === undefined) {  // zero args
      this.top = 0;
      this.right = 0;
      this.bottom = 0;
      this.left = 0;
    } else {
      if (r === undefined) {  // one arg
        r = t;
        b = t;
        l = t;
        this.top = t;
        this.right = r;
        this.bottom = b;
        this.left = l;
      } else if (b === undefined) {  // two args
        b = t;
        l = r;
        this.top = t;
        this.right = r;
        this.bottom = b;
        this.left = l;
      } else if (l !== undefined) { // else four args
        this.top = t;
        this.right = r;
        this.bottom = b;
        this.left = l;
      } else {
        Util.throwError('Invalid arguments to Margin constructor: ' + t + ', ' + r + ', ' + b + ', ' + l);
      }
    }
    this._isFrozen = false;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Margin with those of another Margin.
   * This can modify a frozen object.
   * @this {Margin}
   * @param {Margin} m the other Margin from which to copy property values.
   * @return {Margin} this Margin
   */
  public assign(m: Margin): Margin {
    this.top = m.top;
    this.right = m.right;
    this.bottom = m.bottom;
    this.left = m.left;
    return this;
  }

  /**
   * Modify this Margin with new Top, Right, Bottom, and Left values.
   * @export
   * @this {Margin}
   * @param {number} t top.
   * @param {number} r right.
   * @param {number} b bottom.
   * @param {number} l left.
   * @return {Margin} this.
   */
  public setTo(t: number, r: number, b: number, l: number): Margin {
    if (Debug) {
      Util.checkPrimitive(t, 'number', Margin, 'setTo:t');
      Util.checkPrimitive(r, 'number', Margin, 'setTo:r');
      Util.checkPrimitive(b, 'number', Margin, 'setTo:b');
      Util.checkPrimitive(l, 'number', Margin, 'setTo:l');
      this.checkFrozen();
    }
    this.top = t;
    this.right = r;
    this.bottom = b;
    this.left = l;
    return this;
  }

  /**
   * Modify this Margin so that its Top, Right, Bottom, and Left values
   * are the same as the given Margin.
   * @export
   * @this {Margin}
   * @param {Margin} m the given Margin.
   * @return {Margin} this.
   */
  public set(m: Margin): Margin {
    if (Debug) {
      Util.checkClass(m, Margin, Margin, 'assign:m');
      this.checkFrozen();
    }
    this.top = m.top;
    this.right = m.right;
    this.bottom = m.bottom;
    this.left = m.left;
    return this;
  }

  /**
   * Create a copy of this Margin, with the same values.
   * @expose
   * @this {Margin}
   * @return {Margin}
   */
  public copy(): Margin {
    const copy = new Margin();
    copy.top = this.top;
    copy.right = this.right;
    copy.bottom = this.bottom;
    copy.left = this.left;
    return copy;
  }

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Margin}
   * @return {Margin}
   */
  public constant(): Margin {
    this._isFrozen = true;
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * If this object is 'constant', just return it.
   * Otherwise make a copy of this object and freeze it ({@link #frozen}).
   * @this {Margin}
   * @return {Margin}
   */
  public copyFrozen(): Margin {
    if (this._isFrozen || Object.isFrozen(this)) return this;
    return this.copy().freeze();
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Margin}
   * @return {Margin}
   */
  public freeze(): Margin {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Margin}
   * @return {Margin}
   */
  public thaw(): Margin {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    return this;
  }

  /**
   * @hidden @internal
   * When about to modify this object, check that this object is thawed.
   * @this {Margin}
   * @param {number=} val
   */
  private checkFrozen(val?: number): void {
    if (Debug) {
      if (this._isFrozen) {
        let s = 'The Margin is frozen, so its properties cannot be set: ' + this.toString();
        if (val !== undefined) s += '  to value: ' + val;
        Util.throwError(s);
      }
    }
  }

  /**
   * This static function can be used to read in a {@link Margin} from a string that was produced by {@link Margin.stringify}.
   *
   * `go.Margin.parse("1 2 3 4")` produces the Margin `new go.Margin(1, 2, 3, 4)`.
   * @param {string} str
   * @return {Margin}
   */
  public static parse(str: string): Margin {
    if (typeof str === 'string') {
      const split = str.split(' ');
      let i = 0;
      let t = NaN;
      while (split[i] === '') i++;
      let s = split[i++];
      if (s) t = parseFloat(s);
      if (isNaN(t)) return new Margin();
      let r = NaN;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) r = parseFloat(s);
      if (isNaN(r)) return new Margin(t);
      let b = NaN;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) b = parseFloat(s);
      if (isNaN(b)) return new Margin(t, r);
      let l = NaN;
      while (split[i] === '') i++;
      s = split[i++];
      if (s) l = parseFloat(s);
      if (isNaN(l)) return new Margin(t, r);
      return new Margin(t, r, b, l);
    } else {
      return new Margin();
    }
  }

  /**
   * This static function can be used to write out a {@link Margin} as a string that can be read by {@link Margin.parse}.
   *
   * `go.Margin.stringify(new go.Margin(1, 2, 3, 4))` produces the string "1 2 3 4".
   * @param {Margin} val
   * @return {string}
   */
  public static stringify(val: Margin): string {
    if (Debug) Util.checkClass(val, Margin);
    return val.top.toString() + ' ' + val.right.toString() + ' ' + val.bottom.toString() + ' ' + val.left.toString();
  }

  /**
   * @hidden @internal
   * @this {Margin}
   * @return {string} A string containing this Margin's information.
   */
  public toString(): string {
    return 'Margin(' + this.top + ',' + this.right + ',' + this.bottom + ',' + this.left + ')';
  }

  /**
   * Indicates whether the given Margin is equal to this Margin.
   * @export
   * @this {Margin}
   * @param {Margin} m The Margin to compare to this Margin.
   * @return {boolean} True if the two Margins have identical Top and Right and Bottom and Left values,
   * false otherwise.
   * @see {@link #equalTo}
   */
  public equals(m: Margin): boolean {
    if (!(m instanceof Margin)) return false;
    return (this.top === m.top && this.right === m.right &&
      this.bottom === m.bottom && this.left === m.left);
  }

  /**
   * Indicates whether the given margin is equal to this Margin.
   * @export
   * @this {Margin}
   * @param {number} t top.
   * @param {number} r right.
   * @param {number} b bottom.
   * @param {number} l left.
   * @return {boolean} True if the two Margins have identical Top and Right and Bottom and Left values,
   * false otherwise.
   * @see {@link #equals}
   */
  public equalTo(t: number, r: number, b: number, l: number): boolean {
    return (this.top === t && this.right === r &&
      this.bottom === b && this.left === l);
  }

  // * Indicates whether the given Margin is nearly equal to this Margin.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Margin}
   * @param {Margin} m The Margin to compare to the current Margin.
   * @return {boolean} True if the two Margins have Top, Right,
   * Bottom and Left values within 0.5, false otherwise.
   */
  public equalsApprox(m: Margin): boolean {
    return Geo.isApprox(this.top, m.top) &&
      Geo.isApprox(this.right, m.right) &&
      Geo.isApprox(this.bottom, m.bottom) &&
      Geo.isApprox(this.left, m.left);
  }



  /**
   * True if this Margin has values that are real numbers and not infinity.
   * @export
   * @this {Margin}
   * @return {boolean}
   */
  public isReal(): boolean {
    return isFinite(this.top) && isFinite(this.right) && isFinite(this.bottom) && isFinite(this.left);
  }

  /** @hidden @internal @const */
  public static ZeroMargin: Margin = null;
  /** @hidden @internal @const */
  public static TwoMargin: Margin = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Margin['ZeroMargin'] = Margin.ZeroMargin = new Margin(0, 0, 0, 0).constant();
    Margin['TwoMargin'] = Margin.TwoMargin = new Margin(2, 2, 2, 2).constant();
    Margin['parse'] = Margin.parse;
    Margin['stringify'] = Margin.stringify;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedMargins: Array<Margin> = ([]);

  /**
   * @hidden
   * @return {Margin}
   */
  public static alloc(): Margin {
    const temp = Margin._CachedMargins.pop();
    if (temp === undefined) return new Margin();
    return temp;
  }

  /**
   * @hidden
   * @param {Margin} temp
   */
  public static free(temp: Margin): void {
    Margin._CachedMargins.push(temp);
  }

} // #end Margin
Margin['className'] = 'Margin';

Margin['parse'] = Margin.parse;
Margin['stringify'] = Margin.stringify;
Margin.initStatic();


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Spot
*/

// import { Debug } from './debug';
// import { Util } from './util';

/** @hidden @internal */
const Util2 = {
  /** @hidden @internal @const */
  MTOP: 1,
  /** @hidden @internal @const */
  MLEFT: 2,
  /** @hidden @internal @const */
  MRIGHT: 4,
  /** @hidden @internal @const */
  MBOTTOM: 8
};

/**
 * A Spot represents a relative point from (0,0) to (1,1) within the bounds of
 * a rectangular area plus an absolute offset.
 *
 * Use the static functions {@link Spot.parse} and {@link Spot.stringify} to convert to and from
 * a standard string representation that is independent of the current locale.
 *
 * When an instance of this class is the value of a property of a {@link GraphObject} class or {@link Diagram}
 * or {@link CommandHandler} or a {@link Tool} class, you should treat the object
 * as if it were frozen or read-only -- you cannot modify its properties.
 * This allows the property to return a value without allocating a new instance.
 * If you need to do your own calculations with the value, call {@link #copy} to make
 * a new instance with the same values that you can modify.
 *
 * Many methods modify the object's properties and then return a reference to "this" object.
 * The only instance method to allocate a new object is the {@link #copy} method.
 * The static {@link Spot.parse} method also allocates a new object.
 *
 * The "Debug" implementation of this class is significantly slower than the "Release" implementation,
 * mostly due to additional error checking.
 *
 * You cannot inherit from this class.
 * @category Geometry
 */
export class Spot {
  /** @hidden @internal */ public static className = 'Spot';
  public x: number;
  public y: number;
  public offsetX: number;
  public offsetY: number;
  /** @hidden @internal */ private _isFrozen: boolean;

  /**
   * The default constructor produces the Spot(0, 0, 0, 0), at the top-left corner.
   * @final
   * @struct
   * @param {number=} x The x value of the Spot. Default is zero.
   * @param {number=} y The y value of the Spot. Default is zero.
   * @param {number=} offx (Optional) The absolute x offset. Default is zero.
   * @param {number=} offy (Optional) The absolute y offset. Default is zero.
   */
  constructor(x?: number, y?: number, offx?: number, offy?: number) {
    if (x === undefined) {  // zero args
      this.x = 0;
      this.y = 0;
      this.offsetX = 0;
      this.offsetY = 0;
    } else {  // one or more args
      if (y === undefined) y = 0;
      if (offx === undefined) offx = 0;
      if (offy === undefined) offy = 0;
      this.x = x;
      this.y = y;
      this.offsetX = offx;
      this.offsetY = offy;
    }
    this._isFrozen = false;
  }

  /**
   * @hidden @internal
   * Replace the properties of this Spot with those of another Spot.
   * This can modify a frozen object.
   * @this {Spot}
   * @param {Spot} s the other Spot from which to copy property values.
   * @return {Spot} this Spot
   */
  public assign(s: Spot): Spot {
    this.x = s.x;
    this.y = s.y;
    this.offsetX = s.offsetX;
    this.offsetY = s.offsetY;
    return this;
  }

  /**
   * Modify this Spot with new X, Y, OffsetX, and OffsetY values.
   * @export
   * @this {Spot}
   * @param {number} x
   * @param {number} y
   * @param {number} offx
   * @param {number} offy
   * @return {Spot} this.
   */
  public setTo(x: number, y: number, offx: number, offy: number): Spot {
    if (Debug) {
      this._checkLocation(x, 'setTo:x');
      this._checkLocation(y, 'setTo:y');
      this._checkOffset(offx, 'setTo:offx');
      this._checkOffset(offy, 'setTo:offy');
      this.checkFrozen();
    }
    this.x = x;
    this.y = y;
    this.offsetX = offx;
    this.offsetY = offy;
    return this;
  }

  /**
   * Modify this Spot so that its X, Y, OffsetX, and OffsetY values are the same as the given Spot.
   * @export
   * @this {Spot}
   * @param {Spot} s the given Spot.
   * @return {Spot} this.
   */
  public set(s: Spot): Spot {
    if (Debug) {
      Util.checkClass(s, Spot, Spot, 'set:s');
      this.checkFrozen();
    }
    this.x = s.x;
    this.y = s.y;
    this.offsetX = s.offsetX;
    this.offsetY = s.offsetY;
    return this;
  }

  /**
   * Create a copy of this Spot, with the same values.
   * @expose
   * @this {Spot}
   * @return {Spot}
   */
  public copy(): Spot {
    const copy = new Spot();
    copy.x = this.x;
    copy.y = this.y;
    copy.offsetX = this.offsetX;
    copy.offsetY = this.offsetY;
    return copy;
  }

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Spot}
   * @return {Spot}
   */
  public constant(): Spot {
    this._isFrozen = true;
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * If this object is 'constant', just return it.
   * Otherwise make a copy of this object and freeze it ({@link #frozen}).
   * @this {Spot}
   * @return {Spot}
   */
  public copyFrozen(): Spot {
    if (this._isFrozen || Object.isFrozen(this)) return this;
    return this.copy().freeze();
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Spot}
   * @return {Spot}
   */
  public freeze(): Spot {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Spot}
   * @return {Spot}
   */
  public thaw(): Spot {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    return this;
  }

  /**
   * When about to modify this object, check that this object is thawed.
   * @hidden @internal
   * @this {Spot}
   * @param {number=} val
   */
  public checkFrozen(val?: number): void {
    if (Debug) {
      if (this._isFrozen) {
        let s = 'The Spot is frozen, so its properties cannot be set: ' + this.toString();
        if (val !== undefined) s += '  to value: ' + val;
        Util.throwError(s);
      }
    }
  }

  /**
   * @hidden @internal
   * @param {number} cat
   * @return {Spot}
   */
  private _special(cat: number): Spot {
    this.x = NaN;
    this.y = NaN;
    this.offsetX = cat;
    return this;
  }

  /**
   * @hidden @internal
   * @param {number} v
   * @param {string} name
   */
  private _checkLocation(v: number, name: string): void {
    if (isNaN(v) || v > 1 || v < 0) {
      Util.throwRangeError(v, '0 <= ' + name + ' <= 1', Spot, name);
    }
  }

  /**
   * @hidden @internal
   * @param {number} v
   * @param {string} name
   */
  private _checkOffset(v: number, name: string): void {
    if (isNaN(v) || (v === Infinity) || (v === -Infinity)) {
      Util.throwRangeError(v, 'real number, not NaN or Infinity', Spot, name);
    }
  }

  /**
   * This static function can be used to read in a {@link Spot} from a string that was produced by {@link Spot.stringify}.
   *
   * `go.Spot.parse("0 1 2 3")` produces the Spot `new go.Spot(0, 1, 2, 3)`.
   * @param {string} str
   * @return {Spot}
   */
  public static parse(str: string): Spot {
    if (typeof str === 'string') {
      str = str.trim();
      // handle all no-spot cases
      if (str === 'None') return Spot.None;
      if (str === 'TopLeft') return Spot.TopLeft;
      if (str === 'Top' || str === 'TopCenter' || str === 'MiddleTop') return Spot.TopCenter;
      if (str === 'TopRight') return Spot.TopRight;
      if (str === 'Left' || str === 'LeftCenter' || str === 'MiddleLeft') return Spot.LeftCenter;
      if (str === 'Center') return Spot.Center;
      if (str === 'Right' || str === 'RightCenter' || str === 'MiddleRight') return Spot.RightCenter;
      if (str === 'BottomLeft') return Spot.BottomLeft;
      if (str === 'Bottom' || str === 'BottomCenter' || str === 'MiddleBottom') return Spot.BottomCenter;
      if (str === 'BottomRight') return Spot.BottomRight;
      if (str === 'TopSide') return Spot.TopSide;
      if (str === 'LeftSide') return Spot.LeftSide;
      if (str === 'RightSide') return Spot.RightSide;
      if (str === 'BottomSide') return Spot.BottomSide;
      if (str === 'TopBottomSides') return Spot.TopBottomSides;
      if (str === 'LeftRightSides') return Spot.LeftRightSides;
      if (str === 'TopLeftSides') return Spot.TopLeftSides;
      if (str === 'TopRightSides') return Spot.TopRightSides;
      if (str === 'BottomLeftSides') return Spot.BottomLeftSides;
      if (str === 'BottomRightSides') return Spot.BottomRightSides;
      if (str === 'NotTopSide') return Spot.NotTopSide;
      if (str === 'NotLeftSide') return Spot.NotLeftSide;
      if (str === 'NotRightSide') return Spot.NotRightSide;
      if (str === 'NotBottomSide') return Spot.NotBottomSide;
      if (str === 'AllSides') return Spot.AllSides;
      if (str === 'Default') return Spot.Default;

      // otherwise assume isSpot is true, and it has X,Y,OffsetX,OffsetY values
      const split = str.split(' ');
      let i = 0;
      let x = 0.0;
      while (split[i] === '') i++;
      let s = split[i++];
      if (s !== undefined && s.length > 0) x = parseFloat(s);
      let y = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s !== undefined && s.length > 0) y = parseFloat(s);
      let ox = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s !== undefined && s.length > 0) ox = parseFloat(s);
      let oy = 0.0;
      while (split[i] === '') i++;
      s = split[i++];
      if (s !== undefined && s.length > 0) oy = parseFloat(s);
      return new Spot(x, y, ox, oy);
    } else {
      return new Spot();
    }
  }

  /**
   * This static function can be used to write out a {@link Spot} as a string that can be read by {@link Spot.parse}.
   *
   * `go.Spot.stringify(new go.Spot(0, 1, 2, 3))` produces the string "0 1 2 3".
   * @param {Spot} val
   * @return {string}
   */
  public static stringify(val: Spot): string {
    if (Debug) Util.checkClass(val, Spot);
    if (val.isSpot()) {
      return val.x.toString() + ' ' + val.y.toString() + ' ' + val.offsetX.toString() + ' ' + val.offsetY.toString();
    } else {
      return val.toString();  // no-spot values
    }
  }

  /**
   * @hidden @internal
   * @this {Spot}
   * @return {string}
   */
  public toString(): string {
    if (this.isSpot()) {
      if (this.offsetX === 0 && this.offsetY === 0) {
        return 'Spot(' + this.x + ',' + this.y + ')';
      } else {
        return 'Spot(' + this.x + ',' + this.y + ',' + this.offsetX + ',' + this.offsetY + ')';
      }
    } else {
      // just generate no-spot values
      if (this.equals(Spot.None)) return 'None';
      if (this.equals(Spot.TopLeft)) return 'TopLeft';
      if (this.equals(Spot.TopCenter)) return 'Top';
      if (this.equals(Spot.TopRight)) return 'TopRight';
      if (this.equals(Spot.LeftCenter)) return 'Left';
      if (this.equals(Spot.Center)) return 'Center';
      if (this.equals(Spot.RightCenter)) return 'Right';
      if (this.equals(Spot.BottomLeft)) return 'BottomLeft';
      if (this.equals(Spot.BottomCenter)) return 'Bottom';
      if (this.equals(Spot.BottomRight)) return 'BottomRight';
      if (this.equals(Spot.TopSide)) return 'TopSide';
      if (this.equals(Spot.LeftSide)) return 'LeftSide';
      if (this.equals(Spot.RightSide)) return 'RightSide';
      if (this.equals(Spot.BottomSide)) return 'BottomSide';
      if (this.equals(Spot.TopBottomSides)) return 'TopBottomSides';
      if (this.equals(Spot.LeftRightSides)) return 'LeftRightSides';
      if (this.equals(Spot.TopLeftSides)) return 'TopLeftSides';
      if (this.equals(Spot.TopRightSides)) return 'TopRightSides';
      if (this.equals(Spot.BottomLeftSides)) return 'BottomLeftSides';
      if (this.equals(Spot.BottomRightSides)) return 'BottomRightSides';
      if (this.equals(Spot.NotTopSide)) return 'NotTopSide';
      if (this.equals(Spot.NotLeftSide)) return 'NotLeftSide';
      if (this.equals(Spot.NotRightSide)) return 'NotRightSide';
      if (this.equals(Spot.NotBottomSide)) return 'NotBottomSide';
      if (this.equals(Spot.AllSides)) return 'AllSides';
      if (this.equals(Spot.Default)) return 'Default';
      return 'None';
    }
  }

  /**
   * Two spots are equal if all four property values are the same.
   * @export
   * @this {Spot}
   * @param {Spot} spot The Spot to compare to the current Spot.
   * @return {boolean} True if the two spots are equal, false otherwise.
   */
  public equals(spot: Spot): boolean {
    if (!(spot instanceof Spot)) return false;
    return (
      (this.x === spot.x || (isNaN(this.x) && isNaN(spot.x))) &&
      (this.y === spot.y || (isNaN(this.y) && isNaN(spot.y))) &&
      this.offsetX === spot.offsetX && this.offsetY === spot.offsetY);
  }

  /**
   * Return a new spot that is opposite this spot.
   * The X and Y values will be an equal distance
   * away from the center on the other side of the center.
   * The OffsetX and OffsetY values are also negated.
   *
   * The result is meaningless if {@link #isNoSpot} is true.
   * @export
   * @this {Spot}
   * @return {Spot}
   */
  public opposite(): Spot {
    return new Spot(0.5 - (this.x - 0.5), 0.5 - (this.y - 0.5), -this.offsetX, -this.offsetY);
  }

  /**
   * This predicate is true if this Spot is a side that
   * includes the side(s) given by the argument Spot.
   * @export
   * @this {Spot}
   * @param {Spot} side
   * @return {boolean} false if either this Spot or the argument Spot is not a "Side";
   * true if the side(s) that this Spot represents are a superset or the same set of
   * sides represented by the argument Spot.
   */
  public includesSide(side: Spot): boolean {
    if (!this.isSide()) return false;
    if (!side.isSide()) {
      if (side.equals(Spot.Left)) side = Spot.LeftSide;
      else if (side.equals(Spot.Right)) side = Spot.RightSide;
      else if (side.equals(Spot.Top)) side = Spot.TopSide;
      else if (side.equals(Spot.Bottom)) side = Spot.BottomSide;
      else return false;
    }
    const sides = this.offsetY;
    const s = side.offsetY;
    return ((sides & s) === s);
  }



  /**
   * True if this is a specific spot, not a side nor {@link Spot.None}.
   * @export
   * @this {Spot}
   * @return {boolean}
   */
  public isSpot(): boolean {  // represents a spot, not None nor a side
    return !isNaN(this.x) && !isNaN(this.y);
  }

  /**
   * True if this is an unspecific special spot, such as {@link Spot.None}
   * or one of the sides.
   * @export
   * @this {Spot}
   * @return {boolean}
   */
  public isNoSpot(): boolean {  // could be === None or could be a side
    return isNaN(this.x) || isNaN(this.y);
  }

  /**
   * True if this is a special spot referring to one (or more) of the sides.
   * This is false if the spot is {@link Spot.None}.
   * @export
   * @this {Spot}
   * @return {boolean}
   */
  public isSide(): boolean {
    return isNaN(this.x) && isNaN(this.y) && this.offsetX === 1 && this.offsetY !== 0;
  }

  /**
   * True if this is a special spot referring to no particular spot or side.
   * @export
   * @this {Spot}
   * @return {boolean}
   */
  public isNone(): boolean {
    return isNaN(this.x) && isNaN(this.y) && this.offsetX === 0 && this.offsetY === 0;
  }

  /**
   * True if this is a special spot indicating that the real spot value will come from another source.
   * @export
   * @this {Spot}
   * @return {boolean}
   */
  public isDefault(): boolean {
    return isNaN(this.x) && isNaN(this.y) && this.offsetX === -1 && this.offsetY === 0;
  }

  /**
   * Use this Spot value to indicate no particular spot --
   * code looking for a particular point on an element will need to do their
   * own calculations to determine the desired point depending on the
   * circumstances.
   * @constant
   */
  public static None: Spot;

  /**
   * Use this value to indicate that the real spot value is inherited from elsewhere.
   * @constant
   */
  public static Default: Spot;

  // Specific spots:

  /**
   * The specific point at the top-left corner of the bounding rectangle.
   * @constant
   */
  public static TopLeft: Spot;

  /**
   * The specific point at the center of the top side of the bounding rectangle.
   * @constant
   */
  public static TopCenter: Spot;

  /**
   * The specific point at the top-right corner of the bounding rectangle.
   * @constant
   */
  public static TopRight: Spot;

  /**
   * The specific point at the middle of the left side of bounding rectangle.
   * @constant
   */
  public static LeftCenter: Spot;

  /**
   * The specific point at the very center of the bounding rectangle.
   * @constant
   */
  public static Center: Spot;

  /**
   * The specific point at the middle of the right side of bounding rectangle.
   * @constant
   */
  public static RightCenter: Spot;

  /**
   * The specific point at the bottom-left corner of the bounding rectangle.
   * @constant
   */
  public static BottomLeft: Spot;

  /**
   * The specific point at the middle of the bottom side of bounding rectangle.
   * @constant
   */
  public static BottomCenter: Spot;

  /**
   * The specific point at the bottom-right corner of the bounding rectangle.
   * @constant
   */
  public static BottomRight: Spot;

  /**
   * A synonym for {@link Spot.TopCenter}.
   * @constant
   */
  public static MiddleTop: Spot;

  /**
   * A synonym for {@link Spot.LeftCenter}.
   * @constant
   */
  public static MiddleLeft: Spot;

  /**
   * A synonym for {@link Spot.RightCenter}.
   * @constant
   */
  public static MiddleRight: Spot;

  /**
   * A synonym for {@link Spot.BottomCenter}.
   * @constant
   */
  public static MiddleBottom: Spot;

  /**
   * A synonym for {@link Spot.TopCenter}.
   * @constant
   */
  public static Top: Spot;

  /**
   * A synonym for {@link Spot.LeftCenter}.
   * @constant
   */
  public static Left: Spot;

  /**
   * A synonym for {@link Spot.RightCenter}.
   * @constant
   */
  public static Right: Spot;

  /**
   * A synonym for {@link Spot.BottomCenter}.
   * @constant
   */
  public static Bottom: Spot;

  // Sides:

  /**
   * The set of points at the top side of the bounding rectangle.
   * @constant
   */
  public static TopSide: Spot;

  /**
   * The set of points at the left side of the bounding rectangle.
   * @constant
   */
  public static LeftSide: Spot;

  /**
   * The set of points at the right side of the bounding rectangle.
   * @constant
   */
  public static RightSide: Spot;

  /**
   * The set of points at the bottom side of the bounding rectangle.
   * @constant
   */
  public static BottomSide: Spot;

  /**
   * The set of points at the top or bottom sides of the bounding rectangle.
   * @constant
   */
  public static TopBottomSides: Spot;

  /**
   * The set of points at the left or right sides of the bounding rectangle.
   * @constant
   */
  public static LeftRightSides: Spot;

  /**
   * The set of points at the top or left sides of the bounding rectangle.
   * @constant
   */
  public static TopLeftSides: Spot;

  /**
   * The set of points at the top or right sides of the bounding rectangle.
   * @constant
   */
  public static TopRightSides: Spot;

  /**
   * The set of points at the left or bottom sides of the bounding rectangle.
   * @constant
   */
  public static BottomLeftSides: Spot;

  /**
   * The set of points at the right or bottom sides of the bounding rectangle.
   * @constant
   */
  public static BottomRightSides: Spot;

  /**
   * The set of points on all sides of the bounding rectangle except top side.
   * @constant
   */
  public static NotTopSide: Spot;

  /**
   * The set of points on all sides of the bounding rectangle except left side.
   * @constant
   */
  public static NotLeftSide: Spot;

  /**
   * The set of points on all sides of the bounding rectangle except right side.
   * @constant
   */
  public static NotRightSide: Spot;

  /**
   * The set of points on all sides of bounding rectangle except bottom side.
   * @constant
   */
  public static NotBottomSide: Spot;

  /**
   * The set of points on all sides of the bounding rectangle.
   * @constant
   */
  public static AllSides: Spot;

  /** @hidden @internal @const */
  public static GeneratorEllipseSpot1: Spot;
  /** @hidden @internal @const */
  public static GeneratorEllipseSpot2: Spot;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {

    Spot['None'] = Spot.None = new Spot(0, 0, 0, 0)._special(0).constant();
    Spot['Default'] = Spot.Default = new Spot(0, 0, -1, 0)._special(-1).constant();
    Spot['TopLeft'] = Spot.TopLeft = new Spot(0.0, 0.0, 0, 0).constant();
    Spot['TopCenter'] = Spot.TopCenter = new Spot(0.5, 0.0, 0, 0).constant();
    Spot['TopRight'] = Spot.TopRight = new Spot(1.0, 0.0, 0, 0).constant();
    Spot['LeftCenter'] = Spot.LeftCenter = new Spot(0.0, 0.5, 0, 0).constant();
    Spot['Center'] = Spot.Center = new Spot(0.5, 0.5, 0, 0).constant();
    Spot['RightCenter'] = Spot.RightCenter = new Spot(1.0, 0.5, 0, 0).constant();
    Spot['BottomLeft'] = Spot.BottomLeft = new Spot(0.0, 1.0, 0, 0).constant();
    Spot['BottomCenter'] = Spot.BottomCenter = new Spot(0.5, 1.0, 0, 0).constant();
    Spot['BottomRight'] = Spot.BottomRight = new Spot(1.0, 1.0, 0, 0).constant();
    Spot['MiddleTop'] = Spot.MiddleTop = Spot.TopCenter;
    Spot['MiddleLeft'] = Spot.MiddleLeft = Spot.LeftCenter;
    Spot['MiddleRight'] = Spot.MiddleRight = Spot.RightCenter;
    Spot['MiddleBottom'] = Spot.MiddleBottom = Spot.BottomCenter;
    Spot['Top'] = Spot.Top = Spot.TopCenter;
    Spot['Left'] = Spot.Left = Spot.LeftCenter;
    Spot['Right'] = Spot.Right = Spot.RightCenter;
    Spot['Bottom'] = Spot.Bottom = Spot.BottomCenter;
    Spot['TopSide'] = Spot.TopSide = new Spot(0, 0, 1, Util2.MTOP)._special(1).constant();
    Spot['LeftSide'] = Spot.LeftSide = new Spot(0, 0, 1, Util2.MLEFT)._special(1).constant();
    Spot['RightSide'] = Spot.RightSide = new Spot(0, 0, 1, Util2.MRIGHT)._special(1).constant();
    Spot['BottomSide'] = Spot.BottomSide = new Spot(0, 0, 1, Util2.MBOTTOM)._special(1).constant();
    Spot['TopBottomSides'] = Spot.TopBottomSides = new Spot(0, 0, 1, Util2.MTOP | Util2.MBOTTOM)._special(1).constant();
    Spot['LeftRightSides'] = Spot.LeftRightSides = new Spot(0, 0, 1, Util2.MLEFT | Util2.MRIGHT)._special(1).constant();
    Spot['TopLeftSides'] = Spot.TopLeftSides = new Spot(0, 0, 1, Util2.MTOP | Util2.MLEFT)._special(1).constant();
    Spot['TopRightSides'] = Spot.TopRightSides = new Spot(0, 0, 1, Util2.MTOP | Util2.MRIGHT)._special(1).constant();
    Spot['BottomLeftSides'] = Spot.BottomLeftSides = new Spot(0, 0, 1, Util2.MBOTTOM | Util2.MLEFT)._special(1).constant();
    Spot['BottomRightSides'] = Spot.BottomRightSides = new Spot(0, 0, 1, Util2.MBOTTOM | Util2.MRIGHT)._special(1).constant();
    Spot['NotTopSide'] = Spot.NotTopSide = new Spot(0, 0, 1, Util2.MLEFT | Util2.MRIGHT | Util2.MBOTTOM)._special(1).constant();
    Spot['NotLeftSide'] = Spot.NotLeftSide = new Spot(0, 0, 1, Util2.MTOP | Util2.MRIGHT | Util2.MBOTTOM)._special(1).constant();
    Spot['NotRightSide'] = Spot.NotRightSide = new Spot(0, 0, 1, Util2.MTOP | Util2.MLEFT | Util2.MBOTTOM)._special(1).constant();
    Spot['NotBottomSide'] = Spot.NotBottomSide = new Spot(0, 0, 1, Util2.MTOP | Util2.MLEFT | Util2.MRIGHT)._special(1).constant();
    Spot['AllSides'] = Spot.AllSides = new Spot(0, 0, 1, Util2.MTOP | Util2.MLEFT | Util2.MRIGHT | Util2.MBOTTOM)._special(1).constant();

    Spot.GeneratorEllipseSpot1 = new Spot(0.156, 0.156).constant();
    Spot.GeneratorEllipseSpot2 = new Spot(0.844, 0.844).constant();
    Spot['parse'] = Spot.parse;
    Spot['stringify'] = Spot.stringify;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedSpots: Array<Spot> = [];

  /**
   * @hidden
   * @return {Spot}
   */
  public static alloc(): Spot {
    const temp = Spot._CachedSpots.pop();
    if (temp === undefined) return new Spot();
    return temp;
  }

  /**
   * @hidden
   * @param {Spot} temp
   */
  public static free(temp: Spot): void {
    Spot._CachedSpots.push(temp);
  }

} // #end Spot
Spot['className'] = 'Spot';

Spot['parse'] = Spot.parse;
Spot['stringify'] = Spot.stringify;
Spot.initStatic();


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Transform (Internal)
*/

// import { Debug } from './debug';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Util } from './util';

/**
 * @hidden @internal
 * The Transform class holds a transformation matrix.
 */
export class Transform {
  /** @hidden @internal */ public static className = 'Transform';
  /** @hidden @internal */ public m11: number;
  /** @hidden @internal */ public m12: number;
  /** @hidden @internal */ public m21: number;
  /** @hidden @internal */ public m22: number;
  /** @hidden @internal */ public dx: number;
  /** @hidden @internal */ public dy: number;

  /**
   * @final
   * @struct
   */
  constructor() {
    this.m11 = 1;
    this.m12 = 0;
    this.m21 = 0;
    this.m22 = 1;
    this.dx = 0;
    this.dy = 0;
  }

  /**
   * @hidden @internal
   * Replace the transformation matrix of this Transform with those of another Transform.
   * @export
   * @this {Transform}
   * @param {Transform} t the other Transform from which to copy the transformation matrix.
   * @return {Transform} this.
   */
  public set(t: Transform): Transform {
    this.m11 = t.m11;
    this.m12 = t.m12;
    this.m21 = t.m21;
    this.m22 = t.m22;
    this.dx = t.dx;
    this.dy = t.dy;
    return this;
  }

  /**
   * @hidden @internal
   * Create a copy of this Transform, with the same values.
   * @expose
   * @this {Transform}
   * @return {Transform}
   */
  public copy(): Transform {
    const copy = new Transform();
    copy.m11 = this.m11;
    copy.m12 = this.m12;
    copy.m21 = this.m21;
    copy.m22 = this.m22;
    copy.dx = this.dx;
    copy.dy = this.dy;
    return copy;
  }

  /**
   * @hidden @internal
   * @this {Transform}
   * @return {string}
   */
  public toString(): string {
    return 'Transform(' + this.m11 + ',' + this.m12 + ',' + this.m21 +
      ',' + this.m22 + ',' + this.dx + ',' + this.dy + ')';
  }

  /**
   * @hidden @internal
   * Indicates whether the given Transform is equal to the current Transform.
   * @export
   * @this {Transform}
   * @param {Transform} t The Transform to compare to the current Transform.
   * @return {boolean} True if the two Transforms have identical matrixes,
   * false otherwise.
   */
  public equals(t: Transform): boolean {
    if (!(t instanceof Transform)) return false;
    return (this.m11 === t.m11 && this.m12 === t.m12 && this.m21 === t.m21 &&
      this.m22 === t.m22 && this.dx === t.dx && this.dy === t.dy);
  }

  /**
   * @hidden @internal
   * @export
   * @this {Transform}
   * @return {boolean} True if this transformation is the identity transformation.
   */
  public isIdentity(): boolean {
    return (this.dx === 0 && this.dy === 0 &&
      this.m11 === 1 && this.m12 === 0 &&
      this.m21 === 0 && this.m22 === 1);
  }

  /**
   * @hidden @internal
   * Resets the Transform's matrix to an identity matrix.
   * @export
   * @this {Transform}
   * @return {Transform} this.
   */
  public reset(): Transform {
    this.m11 = 1;
    this.m12 = 0;
    this.m21 = 0;
    this.m22 = 1;
    this.dx = 0;
    this.dy = 0;
    return this;
  }

  /**
   * @hidden @internal
   * Modifies this Transform to be the product of this Transform with a given one.
   * @export
   * @this {Transform}
   * @param {Transform} matrix
   * @return {Transform} this.
   */
  public multiply(matrix: Transform): Transform {
    const m11 = this.m11 * matrix.m11 + this.m21 * matrix.m12;
    const m12 = this.m12 * matrix.m11 + this.m22 * matrix.m12;
    const m21 = this.m11 * matrix.m21 + this.m21 * matrix.m22;
    const m22 = this.m12 * matrix.m21 + this.m22 * matrix.m22;
    this.dx = this.m11 * matrix.dx + this.m21 * matrix.dy + this.dx;
    this.dy = this.m12 * matrix.dx + this.m22 * matrix.dy + this.dy;
    this.m11 = m11;
    this.m12 = m12;
    this.m21 = m21;
    this.m22 = m22;
    return this;
  }

  /**
   * @hidden @internal
   * Modifies this Transform to be the product of this Transform with the inverse of a given one.
   * @export
   * @this {Transform}
   * @param {Transform} t
   * @return {Transform} this.
   */
  public multiplyInverted(t: Transform): Transform {
    // invert
    const d = 1 / (t.m11 * t.m22 - t.m12 * t.m21);
    const m0 = t.m22 * d;
    const m1 = -t.m12 * d;
    const m2 = -t.m21 * d;
    const m3 = t.m11 * d;
    const m4 = d * (t.m21 * t.dy - t.m22 * t.dx);
    const m5 = d * (t.m12 * t.dx - t.m11 * t.dy);

    // multiply
    const m11 = this.m11 * m0 + this.m21 * m1;
    const m12 = this.m12 * m0 + this.m22 * m1;
    const m21 = this.m11 * m2 + this.m21 * m3;
    const m22 = this.m12 * m2 + this.m22 * m3;
    this.dx = this.m11 * m4 + this.m21 * m5 + this.dx;
    this.dy = this.m12 * m4 + this.m22 * m5 + this.dy;
    this.m11 = m11;
    this.m12 = m12;
    this.m21 = m21;
    this.m22 = m22;
    return this;
  }

  /**
   * @hidden @internal
   * Inverts this Transform matrix, modifying the matrix.
   * @export
   * @this {Transform}
   * @return {Transform} this.
   */
  public invert(): Transform {
    const d = 1 / (this.m11 * this.m22 - this.m12 * this.m21);
    const m0 = this.m22 * d;
    const m1 = -this.m12 * d;
    const m2 = -this.m21 * d;
    const m3 = this.m11 * d;
    const m4 = d * (this.m21 * this.dy - this.m22 * this.dx);
    const m5 = d * (this.m12 * this.dx - this.m11 * this.dy);
    this.m11 = m0;
    this.m12 = m1;
    this.m21 = m2;
    this.m22 = m3;
    this.dx = m4;
    this.dy = m5;
    return this;
  }

  /**
   * @hidden @internal
   * Rotates the Transform's matrix clockwise by a given angle.
   * This modifies this Transform.
   * @export
   * @this {Transform}
   * @param {number} angle The angle to rotate by.
   * @param {number} rx The X point to rotate about.
   * @param {number} ry The Y point to rotate about.
   * @return {Transform} this.
   */
  public rotate(angle: number, rx: number, ry: number): Transform {
    if (Debug) {
      Util.checkRealNumber(angle, Transform, 'rotate:angle');
      Util.checkRealNumber(rx, Transform, 'rotate:rx');
      Util.checkRealNumber(ry, Transform, 'rotate:ry');
    }
    if (angle >= 360) angle -= 360; else if (angle < 0) angle += 360;
    if (angle === 0) return this;
    this.translate(rx, ry);
    // optimize common cases to avoid trig function calls
    let c = 0.0;
    let s = 0.0;
    if (angle === 90) {
      c = 0; s = 1;
    } else if (angle === 180) {
      c = -1; s = 0;
    } else if (angle === 270) {
      c = 0; s = -1;
    } else {
      const rad = angle * Math.PI / 180;
      c = Math.cos(rad); s = Math.sin(rad);
    }
    const m11 = this.m11 * c + this.m21 * s;
    const m12 = this.m12 * c + this.m22 * s;
    const m21 = this.m11 * -s + this.m21 * c;
    const m22 = this.m12 * -s + this.m22 * c;
    this.m11 = m11;
    this.m12 = m12;
    this.m21 = m21;
    this.m22 = m22;
    this.translate(-rx, -ry);
    return this;
  }

  /**
   * @hidden @internal
   * Translates this Transform's matrix by x and y.
   * This modifies this Transform.
   * @export
   * @this {Transform}
   * @param {number} x The x-axis translation.
   * @param {number} y The y-axis translation.
   * @return {Transform} this.
   */
  public translate(x: number, y: number): Transform {
    if (Debug) {
      Util.checkRealNumber(x, Transform, 'translate:x');
      Util.checkRealNumber(y, Transform, 'translate:y');
    }
    this.dx += this.m11 * x + this.m21 * y;
    this.dy += this.m12 * x + this.m22 * y;
    return this;
  }

  /**
   * @hidden @internal
   * Scales this Transform's matrix by sx and sy.
   * This modifies this Transform.
   * @export
   * @this {Transform}
   * @param {number} sx A number, that may be negative,
   * but probably should not be zero.
   * @param {number=} sy If not supplied, defaults to the first argument.
   * @return {Transform} this.
   */
  public scale(sx: number, sy?: number): Transform {
    if (sy === undefined) sy = sx;
    if (Debug) {
      Util.checkRealNumber(sx, Transform, 'translate:sx');
      Util.checkRealNumber(sy, Transform, 'translate:sy');
    }
    this.m11 *= sx;
    this.m12 *= sx;
    this.m21 *= sy;
    this.m22 *= sy;
    return this;
  }

  /**
   * @hidden @internal
   * Computes the angle at which this matrix is rotated.
   * @this {Transform}
   * @return {number} angle in degrees, 0 <= a < 360
   */
  public getAngle(): number {
    if (this.m11 === 1 && this.m12 === 0) return 0;
    let atan = Math.atan2(this.m12, this.m11) * 180 / Math.PI;
    if (atan < 0) atan += 360;
    return atan;
  }

  /**
   * @hidden @internal
   * Modifies the given Point with this Transform.
   * @export
   * @this {Transform}
   * @param {Point} p The point to transform with this Transform.
   * @return {Point} the point P.
   */
  public transformPoint(p: Point): Point {
    const x = p.x;
    const y = p.y;
    return p.assignTo(x * this.m11 + y * this.m21 + this.dx, x * this.m12 + y * this.m22 + this.dy);
  }

  /**
   * @hidden @internal
   * Modifies the given Point with the inverse of this Transform.
   * @export
   * @this {Transform}
   * @param {Point} p The point to transform with the inverse of this Transform.
   * @return {Point} the point P.
   */
  public invertedTransformPoint(p: Point): Point {
    const d = 1 / (this.m11 * this.m22 - this.m12 * this.m21);
    const m0 = this.m22 * d;
    const m1 = -this.m12 * d;
    const m2 = -this.m21 * d;
    const m3 = this.m11 * d;
    const m4 = d * (this.m21 * this.dy - this.m22 * this.dx);
    const m5 = d * (this.m12 * this.dx - this.m11 * this.dy);
    const x = p.x;
    const y = p.y;
    return p.assignTo(x * m0 + y * m2 + m4, x * m1 + y * m3 + m5);
  }

  /**
   * @hidden @internal
   * Modifies a rectangle (X, Y, W, H) with the bounds of the transformed Rect in
   * the transformed coordinate system.
   * @export
   * @this {Transform}
   * @param {Rect} rect the Rect to modify with the resulting bounds.
   * @return {Rect} the argument Rect
   */
  public transformRect(rect: Rect): Rect {
    const x = rect.x;
    const y = rect.y;
    const r = x + rect.width;
    const b = y + rect.height;

    const m0 = this.m11;
    const m1 = this.m12;
    const m2 = this.m21;
    const m3 = this.m22;
    const m4 = this.dx;
    const m5 = this.dy;

    const ltx = x * m0 + y * m2 + m4;
    const lty = x * m1 + y * m3 + m5;

    const rtx = r * m0 + y * m2 + m4;
    const rty = r * m1 + y * m3 + m5;

    const lbx = x * m0 + b * m2 + m4;
    const lby = x * m1 + b * m3 + m5;

    const rbx = r * m0 + b * m2 + m4;
    const rby = r * m1 + b * m3 + m5;

    let minx = ltx;
    let maxx = ltx;
    let miny = lty;
    let maxy = lty;

    minx = Math.min(minx, rtx);
    maxx = Math.max(maxx, rtx);
    miny = Math.min(miny, rty);
    maxy = Math.max(maxy, rty);

    minx = Math.min(minx, lbx);
    maxx = Math.max(maxx, lbx);
    miny = Math.min(miny, lby);
    maxy = Math.max(maxy, lby);

    minx = Math.min(minx, rbx);
    maxx = Math.max(maxx, rbx);
    miny = Math.min(miny, rby);
    maxy = Math.max(maxy, rby);

    rect.assignTo(minx, miny, maxx - minx, maxy - miny);
    return rect;
  }

  // Used to be on Util:
  /** @hidden @internal @const */
  public static _CachedTransforms: Array<Transform> = ([]);

  /**
   * @hidden @internal
   * @return {Transform}
   */
  public static alloc(): Transform {
    const temp = Transform._CachedTransforms.pop();
    if (temp === undefined) return new Transform();
    return temp;
  }

  /**
   * @hidden @internal
   * @param {Transform} temp
   */
  public static free(temp: Transform): void {
    Transform._CachedTransforms.push(temp);
  }

  /**
   * @hidden @internal
   */
  public static wmq3 = '54a702f3e53909c447824c6706603faf4c'; // 'Not for distribution '
} // #end Transform
Transform['className'] = 'Transform';
(Transform as any)['xF'] = Transform.wmq3;


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
* This file defines the Geo static class
*/

// import { Point } from './point';
// import { PropertyCollection } from './propertyCollection';
// import { Rect } from './rect';

/**
 * @hidden @internal
 * Static Geo object for convenience functions
 */
export let Geo = {
  /* @hidden @internal */
  wmq4: '7da71ca0ad381e90', // gojs.net
  // constant used to approximate a symmetrical arc using a cubic Bezier curve
  /** @hidden @internal @const */
  KAPPA: 4 * ((Math.sqrt(2) - 1) / 3),

  /** @hidden @internal */
  Sqrts: null as Array<number>,

  /**
   * @hidden @internal
   * Geo.sqrt is just used by Layouts, where use of Math.sqrt is too slow and the precision of the result is not so important
   * Use Math.sqrt instead when needing good precision.
   * @param {number} n
   * @return {number}
   */
  sqrt: function (n: number): number {
    if (n <= 0) return 0;
    let a = Geo.Sqrts;
    if (a === null) {
      a = [];
      for (let i = 0; i <= 2000; i++) {
        a[i] = Math.sqrt(i);
      }
      Geo.Sqrts = a;
    }
    if (n < 1) {
      const x = 1 / n;
      if (x <= 2000) {
        return 1 / (a[x | 0]);
      } else {
        return Math.sqrt(n);
      }
    } else {
      if (n <= 2000) {
        return a[n | 0];
      } else {
        return Math.sqrt(n);
      }
    }
  },

  /**
   * @hidden @internal
   * True if the difference between X and Y is less than 0.5.
   * @param {number} x
   * @param {number} y
   * @return {boolean}
   */
  isApprox: function (x: number, y: number): boolean {
    const d = x - y;
    return d < 0.5 && d > -0.5;
  },

  /**
   * @hidden @internal
   * True if the difference between X and Y is less than 0.00000005.
   * @param {number} x
   * @param {number} y
   * @return {boolean}
   */
  isApproxEqual: function (x: number, y: number): boolean {
    const d = x - y;
    return d < 0.00000005 && d > -0.00000005;
  },

  /**
   * @hidden @internal
   * See if P is near the line A - B.
   * @param {number} ax
   * @param {number} ay
   * @param {number} bx
   * @param {number} by
   * @param {number} fuzz A double describing the resolution, with a minimum of 0.000001.
   * @param {number} px
   * @param {number} py
   * @return {boolean}
   */
  lineContainsPoint: function (ax: number, ay: number, bx: number, by: number, fuzz: number, px: number, py: number): boolean {
    if (fuzz <= 0) fuzz = 0.000001;
    let maxx = 0.0;
    let minx = 0.0;
    let maxy = 0.0;
    let miny = 0.0;
    if (ax < bx) {
      minx = ax;
      maxx = bx;
    } else {
      minx = bx;
      maxx = ax;
    }
    if (ay < by) {
      miny = ay;
      maxy = by;
    } else {
      miny = by;
      maxy = ay;
    }

    // do simple vertical and horizontal cases first,
    // without fuzz beyond the endpoints

    // ??? the added fuzz here fixes mitering errors on corners.
    // but the older version is being used because we do not want fuzz on endpoints.
    // in the future we should distinguish between endpoint line segments and non-enpoint segments.
    /*
    if (ax === bx) {
      return (miny - fuzz <= py && py <= maxy + fuzz &&
              ax - fuzz <= px && px <= ax + fuzz);
    }
    if (ay === by) {
      return (minx - fuzz <= px && px <= maxx + fuzz &&
              ay - fuzz <= py && py <= ay + fuzz);
    }*/

    // do simple vertical and horizontal cases first,
    // without fuzz beyond the endpoints
    if (ax === bx) {
      return (miny <= py && py <= maxy && ax - fuzz <= px && px <= ax + fuzz);
    }
    if (ay === by) {
      return (minx <= px && px <= maxx && ay - fuzz <= py && py <= ay + fuzz);
    }

    // if we're in the x-range, including fuzz,
    const xrangeHigh = maxx + fuzz;
    const xrangeLow = minx - fuzz;
    if ((xrangeLow <= px) && (px <= xrangeHigh)) {

      // and if we're in the y-range
      const yrangeHigh = maxy + fuzz;
      const yrangeLow = miny - fuzz;
      if ((yrangeLow <= py) && (py <= yrangeHigh)) {

        // see if we should compute the X coordinate or Y coordinate
        if (xrangeHigh - xrangeLow > yrangeHigh - yrangeLow) {
          if (ax - bx > fuzz || bx - ax > fuzz) {
            const slope = (by - ay) / (bx - ax);
            const guessY = (slope * (px - ax) + ay);

            if ((guessY - fuzz <= py) && (py <= guessY + fuzz)) {
              return true;
            }
          } else {
            return true;
          }
        } else {
          if (ay - by > fuzz || by - ay > fuzz) {
            const slope = (bx - ax) / (by - ay);
            const guessX = (slope * (py - ay) + ax);

            if ((guessX - fuzz <= px) && (px <= guessX + fuzz)) {
              return true;
            }
          } else {
            return true;
          }
        }
      }
    }
    return false;
  }, // end lineContainsPoint

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon
   * @param {number} px
   * @param {number} py
   * @param {number} thickness
   * @return {boolean}
   */
  bezierContainsPoint: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number,
                                 ex: number, ey: number, epsilon: number, px: number, py: number, thickness: number): boolean {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, thickness, c1x, c1y) ||
      !Geo.lineContainsPoint(sx, sy, ex, ey, thickness, c2x, c2y)) {
      const a1x = (sx + c1x) / 2;
      const a1y = (sy + c1y) / 2;
      const a2x = (c1x + c2x) / 2;
      const a2y = (c1y + c2y) / 2;
      const a3x = (c2x + ex) / 2;
      const a3y = (c2y + ey) / 2;

      const vx = (a1x + a2x) / 2;
      const vy = (a1y + a2y) / 2;
      const wx = (a2x + a3x) / 2;
      const wy = (a2y + a3y) / 2;
      const mx = (vx + wx) / 2;
      const my = (vy + wy) / 2;
      return Geo.bezierContainsPoint(sx, sy, a1x, a1y, vx, vy, mx, my, epsilon, px, py, thickness) ||
        Geo.bezierContainsPoint(mx, my, wx, wy, a3x, a3y, ex, ey, epsilon, px, py, thickness);
    } else {
      return Geo.lineContainsPoint(sx, sy, ex, ey, thickness, px, py);
    }
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {Point} midp  Modified to hold the result.
   * @return {Point}
   */
  bezierMidPoint: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number, ex: number, ey: number, midp: Point): Point {
    const m1x = (sx + c1x) / 2;
    const m1y = (sy + c1y) / 2;
    const m2x = (c1x + c2x) / 2;
    const m2y = (c1y + c2y) / 2;
    const m3x = (c2x + ex) / 2;
    const m3y = (c2y + ey) / 2;

    const vx = (m1x + m2x) / 2;
    const vy = (m1y + m2y) / 2;
    const wx = (m2x + m3x) / 2;
    const wy = (m2y + m3y) / 2;

    midp.assignTo((vx + wx) / 2, (vy + wy) / 2);
    return midp;
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @return {number}
   */
  bezierMidAngle: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number, ex: number, ey: number): number {
    const m1x = (sx + c1x) / 2;
    const m1y = (sy + c1y) / 2;
    const m2x = (c1x + c2x) / 2;
    const m2y = (c1y + c2y) / 2;
    const m3x = (c2x + ex) / 2;
    const m3y = (c2y + ey) / 2;

    const vx = (m1x + m2x) / 2;
    const vy = (m1y + m2y) / 2;
    const wx = (m2x + m3x) / 2;
    const wy = (m2y + m3y) / 2;

    return Point.direction(vx, vy, wx, wy);
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon too slow at 0.1, places that call bumped values to 0.5
   * @param {Rect} result a Rect that is modified to hold the union of the points on the line.
   * @return {Rect}
   */
  bezierBounds: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number, ex: number, ey: number, epsilon: number, result: Rect): Rect {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c1x, c1y) ||
      !Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c2x, c2y)) {
      const a1x = (sx + c1x) / 2;
      const a1y = (sy + c1y) / 2;
      const a2x = (c1x + c2x) / 2;
      const a2y = (c1y + c2y) / 2;
      const a3x = (c2x + ex) / 2;
      const a3y = (c2y + ey) / 2;

      const vx = (a1x + a2x) / 2;
      const vy = (a1y + a2y) / 2;
      const wx = (a2x + a3x) / 2;
      const wy = (a2y + a3y) / 2;
      const mx = (vx + wx) / 2;
      const my = (vy + wy) / 2;

      Geo.bezierBounds(sx, sy, a1x, a1y, vx, vy, mx, my, epsilon, result);
      Geo.bezierBounds(mx, my, wx, wy, a3x, a3y, ex, ey, epsilon, result);
    } else {
      result.unionNoCheck(sx, sy, 0, 0);
      result.unionNoCheck(ex, ey, 0, 0);
    }
    return result;
  },

  // return a bunch of line segments describing the bezier curve
  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon
   * @param {Array.<number>} result an Array that holds line segments, alternating X and Y, roughly describing the bezier.
   * @return {Array.<number>}
   */
  bezierLines: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number,
                         ex: number, ey: number, epsilon: number, result: Array<number>): Array<number> {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c1x, c1y) ||
      !Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c2x, c2y)) {
      const a1x = (sx + c1x) / 2;
      const a1y = (sy + c1y) / 2;
      const a2x = (c1x + c2x) / 2;
      const a2y = (c1y + c2y) / 2;
      const a3x = (c2x + ex) / 2;
      const a3y = (c2y + ey) / 2;

      const vx = (a1x + a2x) / 2;
      const vy = (a1y + a2y) / 2;
      const wx = (a2x + a3x) / 2;
      const wy = (a2y + a3y) / 2;
      const mx = (vx + wx) / 2;
      const my = (vy + wy) / 2;

      Geo.bezierLines(sx, sy, a1x, a1y, vx, vy, mx, my, epsilon, result);
      Geo.bezierLines(mx, my, wx, wy, a3x, a3y, ex, ey, epsilon, result);
    } else {
      if (result.length === 0) { result.push(sx); result.push(sy); }
      result.push(ex); result.push(ey);
    }
    return result;
  },

  //  hidden internal
  //  * param {number} startx
  //  * param {number} starty
  //  * param {number} c1x
  //  * param {number} c1y
  //  * param {number} c2x
  //  * param {number} c2y
  //  * param {number} endx
  //  * param {number} endy
  //  * param {number} fraction
  //  * param {Point} curve1cp1 out result
  //  * param {Point} curve1cp2 out result
  //  * param {Point} midpoint out result
  //  * param {Point} curve2cp1 out result
  //  * param {Point} curve2cp2 out result
  //
  // breakUpBezier: function (startx: number, starty: number, c1x: number, c1y: number, c2x: number, c2y: number, endx: number, endy: number,
  //                          fraction: number, curve1cp1: Point, curve1cp2: Point, midpoint: Point, curve2cp1: Point, curve2cp2: Point) {
  //   const fo = 1 - fraction;
  //   const so = fraction;
  //   const m1x = (startx * fo + c1x * so);
  //   const m1y = (starty * fo + c1y * so);
  //   const m2x = (c1x * fo + c2x * so);
  //   const m2y = (c1y * fo + c2y * so);
  //   const m3x = (c2x * fo + endx * so);
  //   const m3y = (c2y * fo + endy * so);
  //   const m12x = (m1x * fo + m2x * so);
  //   const m12y = (m1y * fo + m2y * so);
  //   const m23x = (m2x * fo + m3x * so);
  //   const m23y = (m2y * fo + m3y * so);
  //   const m123x = (m12x * fo + m23x * so);
  //   const m123y = (m12y * fo + m23y * so);
  //   curve1cp1.assignTo(m1x, m1y);
  //   curve1cp2.assignTo(m12x, m12y);
  //   midpoint.assignTo(m123x, m123y);
  //   curve2cp1.assignTo(m23x, m23y);
  //   curve2cp2.assignTo(m3x, m3y);
  // },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} cx
   * @param {number} cy
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon
   * @param {number} px
   * @param {number} py
   * @param {number} thickness
   * @return {boolean}
   */
  quadraticBezierContainsPoint: function (sx: number, sy: number, cx: number, cy: number, ex: number, ey: number,
                                          epsilon: number, px: number, py: number, thickness: number): boolean {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, thickness, cx, cy)) {
      const a1x = (sx + cx) / 2;
      const a1y = (sy + cy) / 2;
      const a2x = (cx + ex) / 2;
      const a2y = (cy + ey) / 2;

      const mx = (a1x + a2x) / 2;
      const my = (a1y + a2y) / 2;

      return Geo.quadraticBezierContainsPoint(sx, sy, a1x, a1y, mx, my, epsilon, px, py, thickness) ||
        Geo.quadraticBezierContainsPoint(mx, my, a2x, a2y, ex, ey, epsilon, px, py, thickness);
    } else {
      return Geo.lineContainsPoint(sx, sy, ex, ey, thickness, px, py);
    }
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} cx
   * @param {number} cy
   * @param {number} ex
   * @param {number} ey
   * @param {Point} midp  Modified to hold the result.
   * @return {Point}
   */
  quadraticBezierMidPoint: function (sx: number, sy: number, cx: number, cy: number, ex: number, ey: number, midp: Point): Point {
    const m1x = (sx + cx) / 2;
    const m1y = (sy + cy) / 2;
    const m2x = (cx + ex) / 2;
    const m2y = (cy + ey) / 2;

    midp.assignTo((m1x + m2x) / 2, (m1y + m2y) / 2);
    return midp;
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} cx
   * @param {number} cy
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon
   * @param {Rect} result a Rect that is modified to hold the union of the points on the line.
   * @return {Rect}
   */
  quadraticBezierBounds: function (sx: number, sy: number, cx: number, cy: number, ex: number, ey: number, epsilon: number, result: Rect): Rect {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, cx, cy)) {
      const a1x = (sx + cx) / 2;
      const a1y = (sy + cy) / 2;
      const a2x = (cx + ex) / 2;
      const a2y = (cy + ey) / 2;

      const mx = (a1x + a2x) / 2;
      const my = (a1y + a2y) / 2;

      Geo.quadraticBezierBounds(sx, sy, a1x, a1y, mx, my, epsilon, result);
      Geo.quadraticBezierBounds(mx, my, a2x, a2y, ex, ey, epsilon, result);
    } else {
      result.unionNoCheck(sx, sy, 0, 0);
      result.unionNoCheck(ex, ey, 0, 0);
    }
    return result;
  },

  // return a bunch of line segments describing the bezier curve
  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} cx
   * @param {number} cy
   * @param {number} ex
   * @param {number} ey
   * @param {number} epsilon
   * @param {Array.<number>} result an Array that holds line segments, alternating X and Y, roughly describing the bezier.
   * @return {Array.<number>}
   */
  quadraticBezierLines: function (sx: number, sy: number, cx: number, cy: number, ex: number, ey: number, epsilon: number, result: Array<number>): Array<number> {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, cx, cy)) {
      const a1x = (sx + cx) / 2;
      const a1y = (sy + cy) / 2;
      const a2x = (cx + ex) / 2;
      const a2y = (cy + ey) / 2;

      const mx = (a1x + a2x) / 2;
      const my = (a1y + a2y) / 2;

      Geo.quadraticBezierLines(sx, sy, a1x, a1y, mx, my, epsilon, result);
      Geo.quadraticBezierLines(mx, my, a2x, a2y, ex, ey, epsilon, result);
    } else {
      if (result.length === 0) { result.push(sx); result.push(sy); }
      result.push(ex); result.push(ey);
    }
    return result;
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {number} epsilon if (epsilon <= 0) epsilon = 0.000001;
   * @param {Point} R a Point that is modified to hold the union of the points on the line.
   * @return {boolean}
   */
  bezierNearestIntersectionOnLine: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number,
                                             ex: number, ey: number, p1x: number, p1y: number, p2x: number, p2y: number, epsilon: number, R: Point): boolean {
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c1x, c1y) ||
      !Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c2x, c2y)) {
      const a1x = (sx + c1x) / 2;
      const a1y = (sy + c1y) / 2;
      const a2x = (c1x + c2x) / 2;
      const a2y = (c1y + c2y) / 2;
      const a3x = (c2x + ex) / 2;
      const a3y = (c2y + ey) / 2;
      const vx = (a1x + a2x) / 2;
      const vy = (a1y + a2y) / 2;
      const wx = (a2x + a3x) / 2;
      const wy = (a2y + a3y) / 2;
      const mx = (vx + wx) / 2;
      const my = (vy + wy) / 2;

      let dist = (p2x - p1x) * (p2x - p1x) + (p2y - p1y) * (p2y - p1y);
      let fnd = false;

      if (Geo.bezierNearestIntersectionOnLine(sx, sy, a1x, a1y, vx, vy, mx, my, p1x, p1y, p2x, p2y, epsilon, R)) {
        const tempdist = (R.x - p1x) * (R.x - p1x) + (R.y - p1y) * (R.y - p1y);
        if (tempdist < dist) {
          dist = tempdist;
          fnd = true;
        }
      }
      const Rx = R.x;
      const Ry = R.y;
      if (Geo.bezierNearestIntersectionOnLine(mx, my, wx, wy, a3x, a3y, ex, ey, p1x, p1y, p2x, p2y, epsilon, R)) {
        const tempdist = (R.x - p1x) * (R.x - p1x) + (R.y - p1y) * (R.y - p1y);
        if (tempdist < dist) {
          dist = tempdist;
          fnd = true;
        } else {
          R.assignTo(Rx, Ry);
        }
      }
      return fnd;
    } else {
      const d = (sx - ex) * (p1y - p2y) - (sy - ey) * (p1x - p2x);
      if (d === 0) { return false; }
      const xint = ((sx * ey - sy * ex) * (p1x - p2x) - (sx - ex) * (p1x * p2y - p1y * p2x)) / d;
      const yint = ((sx * ey - sy * ex) * (p1y - p2y) - (sy - ey) * (p1x * p2y - p1y * p2x)) / d;

      if ((p1x > p2x ? p1x - p2x : p2x - p1x) < (p1y > p2y ? p1y - p2y : p2y - p1y)) {
        let min = 0.0;
        let max = 0.0;
        if (sy < ey) { min = sy; max = ey; } else { min = ey; max = sy; }
        if (yint < min || yint > max) { return false; }
      } else {
        let min = 0.0;
        let max = 0.0;
        if (sx < ex) { min = sx; max = ex; } else { min = ex; max = sx; }
        if (xint < min || xint > max) { return false; }
      }
      R.assignTo(xint, yint);
      return true;
    }
  },

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   * @param {number} c1x
   * @param {number} c1y
   * @param {number} c2x
   * @param {number} c2y
   * @param {number} ex
   * @param {number} ey
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {number} epsilon if (epsilon <= 0) epsilon = .000001;
   * @return {number}
   */
  bezierWinding: function (sx: number, sy: number, c1x: number, c1y: number, c2x: number, c2y: number,
                           ex: number, ey: number, p1x: number, p1y: number, p2x: number, p2y: number, epsilon: number): number {
    let res = 0;
    if (!Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c1x, c1y) ||
      !Geo.lineContainsPoint(sx, sy, ex, ey, epsilon, c2x, c2y)) {

      const a1x = (sx + c1x) / 2;
      const a1y = (sy + c1y) / 2;
      const a2x = (c1x + c2x) / 2;
      const a2y = (c1y + c2y) / 2;
      const a3x = (c2x + ex) / 2;
      const a3y = (c2y + ey) / 2;
      const vx = (a1x + a2x) / 2;
      const vy = (a1y + a2y) / 2;
      const wx = (a2x + a3x) / 2;
      const wy = (a2y + a3y) / 2;
      const mx = (vx + wx) / 2;
      const my = (vy + wy) / 2;

      res += Geo.bezierWinding(sx, sy, a1x, a1y, vx, vy, mx, my, p1x, p1y, p2x, p2y, epsilon);
      res += Geo.bezierWinding(mx, my, wx, wy, a3x, a3y, ex, ey, p1x, p1y, p2x, p2y, epsilon);
    } else {

      const d = (sx - ex) * (p1y - p2y) - (sy - ey) * (p1x - p2x);
      if (d === 0) { return res; }
      const xint = ((sx * ey - sy * ex) * (p1x - p2x) - (sx - ex) * (p1x * p2y - p1y * p2x)) / d;
      const yint = ((sx * ey - sy * ex) * (p1y - p2y) - (sy - ey) * (p1x * p2y - p1y * p2x)) / d;

      if (xint >= p2x) { return res; } // has to be on correct side

      if ((p1x > p2x ? p1x - p2x : p2x - p1x) < (p1y > p2y ? p1y - p2y : p2y - p1y)) {
        let min = 0.0;
        let max = 0.0;
        if (sy < ey) { min = sy; max = ey; } else { min = ey; max = sy; }
        if (yint < min || yint > max) { return res; }
      } else {
        let min = 0.0;
        let max = 0.0;
        if (sx < ex) { min = sx; max = ex; } else { min = ex; max = sx; }
        if (xint < min || xint > max) { return res; }
      }

      if (d > 0) {
        res++;
      } else if (d < 0) {
        res--;
      }
    }
    return res;
  },

  /**
   * @hidden @internal
   * Return a point on a straight line segment that is closest to a given point.
   * @param {number} Ax One end of the line.
   * @param {number} Ay
   * @param {number} Bx The other end of the line.
   * @param {number} By
   * @param {number} Px The point to be closest to.
   * @param {number} Py
   * @param {Point} result Modified to be a point that is on the finite length straight line segment from A-B.
   * @return {boolean} true if the point P is on on a perpendicular line to the line segment;
   *    false if the point P is beyond either end of the line segment.
   *    When this returns false, the RESULT will be either A or B.
   */
  nearestPointOnLine: function (Ax: number, Ay: number, Bx: number, By: number, Px: number, Py: number, result: Point): boolean {
    // A and B are the endpoints of this segment
    // P is the point we want to get closest to

    // handle vertical and horizontal lines specially
    if (Geo.isApproxEqual(Ax, Bx)) { // was ==
      let min = 0.0;
      let max = 0.0;
      if (Ay < By) {
        min = Ay;
        max = By;
      } else {
        min = By;
        max = Ay;
      }
      const newp = Py;
      if (newp < min) {
        result.assignTo(Ax, min);
        return false;
      }
      if (newp > max) {
        result.assignTo(Ax, max);
        return false;
      }
      result.assignTo(Ax, newp);
      return true;
    } else if (Geo.isApproxEqual(Ay, By)) {
      let min = 0.0;
      let max = 0.0;
      if (Ax < Bx) {
        min = Ax;
        max = Bx;
      } else {
        min = Bx;
        max = Ax;
      }
      const newp = Px;
      if (newp < min) {
        result.assignTo(min, Ay);
        return false;
      }
      if (newp > max) {
        result.assignTo(max, Ay);
        return false;
      }
      result.assignTo(newp, Ay);
      return true;
    } else {
      // ought to take sqrt to get real length, but don't bother...
      const L = (Bx - Ax) * (Bx - Ax) + (By - Ay) * (By - Ay);
      // ought to be dividing By L^2, but didn't bother to sqrt!
      const Q = ((Ax - Px) * (Ax - Bx) + (Ay - Py) * (Ay - By)) / L;

      if (Q < -0.000005) {
        result.assignTo(Ax, Ay);
        return false;
      } else if (Q > 1.000005) {
        result.assignTo(Bx, By);
        return false;
      } else {
        // OK to use point on line between A and B
        const x = Ax + Q * (Bx - Ax);
        const y = Ay + Q * (By - Ay);
        result.assignTo(x, y);
        return true;
      }
    }
  },

  /**
   * @hidden @internal
   * Find the intersection point of the finite line segment A-B and the
   * infinite line P-Q
   * that is closest to point P.
   * result is out variable fixed
   * @param {number} Ax
   * @param {number} Ay
   * @param {number} Bx
   * @param {number} By
   * @param {number} Px
   * @param {number} Py
   * @param {number} Qx
   * @param {number} Qy
   * @param {Point} result
   * @return {boolean}
   */
  nearestIntersectionOnLine: function (Ax: number, Ay: number, Bx: number, By: number, Px: number, Py: number, Qx: number, Qy: number, result: Point): boolean {
    if (Geo.isApprox(Ax, Bx) && Geo.isApprox(Ay, By)) {
      result.assignTo(Ax, Ay);
      return false;
    }

    if (Geo.isApproxEqual(Px, Qx)) {  // P-Q line is vertical
      if (Geo.isApproxEqual(Ax, Bx)) {  // parallel vertical lines: just return closest point
        Geo.nearestPointOnLine(Ax, Ay, Bx, By, Px, Py, result);
        return false;
      } else {  // line segment A-B is not vertical, but P-Q is
        const M = (By - Ay) / (Bx - Ax);
        const y = M * (Px - Ax) + Ay;
        return Geo.nearestPointOnLine(Ax, Ay, Bx, By, Px, y, result);
      }
    } else {
      const S = (Qy - Py) / (Qx - Px);
      if (Geo.isApproxEqual(Ax, Bx)) {  // line segment A-B is vertical, but P-Q is not
        const y = S * (Ax - Px) + Py;
        let min = 0.0;
        let max = 0.0;
        if (Ay < By) {
          min = Ay; max = By;
        } else {
          min = By; max = Ay;
        }
        if (y < min) {
          result.assignTo(Ax, min);
          return false;
        }
        if (y > max) {
          result.assignTo(Ax, max);
          return false;
        }
        result.assignTo(Ax, y);
        return true;
      } else {  // neither line is vertical
        const M = (By - Ay) / (Bx - Ax);
        if (Geo.isApproxEqual(S, M)) {  // parallel lines--easiest to just return closest point
          Geo.nearestPointOnLine(Ax, Ay, Bx, By, Px, Py, result);
          return false;
        } else {
          const x = (M * Ax - S * Px + Py - Ay) / (M - S);
          if (Geo.isApproxEqual(M, 0)) {  // line segment A-B is horizontal
            let min = 0.0;
            let max = 0.0;
            if (Ax < Bx) {
              min = Ax; max = Bx;
            } else {
              min = Bx; max = Ax;
            }
            if (x < min) {
              result.assignTo(min, Ay);
              return false;
            }
            if (x > max) {
              result.assignTo(max, Ay);
              return false;
            }
            result.assignTo(x, Ay);
            return true;
          } else {
            const y = M * (x - Ax) + Ay;
            return Geo.nearestPointOnLine(Ax, Ay, Bx, By, x, y, result);
          }
        }
      }
    }
  },

  /**
   * @hidden @internal
   * Returns true if line(line1, line2) intersects with line
   * segment(seg1, seg2) and modifies the intersection argument with the result
   * @param {Point} line1
   * @param {Point} line2
   * @param {Point} seg1
   * @param {Point} seg2
   * @param {Point} intersection modified to hold result.
   * @return {boolean}
   */
  lineIntersectsWithSegment: function (line1: Point, line2: Point, seg1: Point, seg2: Point, intersection: Point): boolean {
    return Geo.nearestIntersectionOnLine(seg1.x, seg1.y, seg2.x, seg2.y, line1.x, line1.y, line2.x, line2.y, intersection);
  },

  /**
   * @hidden @internal
   * Returns the closest intersection between the infinite line L1-L2
   * and the ellipse in the rectangle S-E, and accounting for a strokeWidth SW.
   * @param {number} l1x
   * @param {number} l1y
   * @param {number} l2x
   * @param {number} l2y
   * @param {number} sx
   * @param {number} sy
   * @param {number} ex
   * @param {number} ey
   * @param {Point} result
   * @param {number} sw
   * @return {Point}
   */
  getNearestLineEllipseIntersection: function (l1x: number, l1y: number, l2x: number, l2y: number, sx: number, sy: number,
                                               ex: number, ey: number, result: Point, sw: number): Point {
    let dist2 = Infinity;
    /**
     * @hidden @internal
     * @param {number} x1
     * @param {number} y1
     */
    function dist(x1: number, y1: number) {
      const d = (x1 - l1x) * (x1 - l1x) + (y1 - l1y) * (y1 - l1y);
      if (d < dist2) { dist2 = d; result.assignTo(x1, y1); }
    }

    dist(result.x, result.y);

    // Now, check for the actual intersections
    let rectL = 0.0;
    let rectR = 0.0;
    let rectT = 0.0;
    let rectB = 0.0;
    if (sx < ex) { rectL = sx; rectR = ex; } else { rectL = ex; rectR = sx; }
    if (sy < ey) { rectT = sx; rectB = ex; } else { rectT = ex; rectB = sx; }
    const rectW = rectR - rectL; const rectH = rectB - rectT;
    const ax = rectW / 2 + sw;
    const by = rectH / 2 + sw;
    const centerX = (sx + ex) / 2;
    const centerY = (sy + ey) / 2; // center assuming 2ax*2bx is the size
    // No space is occupied if radii are 0
    if (ax === 0 || by === 0) return result;
    if ((l2x > l1x ? l2x - l1x : l1x - l2x) < 0.5) { // If it's a vertical line
      const disc = 1 - (l2x - centerX) * (l2x - centerX) / (ax * ax);
      if (disc < 0) return result;
      const disc2 = Math.sqrt(disc);
      const y1 = by * disc2 + centerY;
      const y2 = -by * disc2 + centerY;
      dist(l2x, y1);
      dist(l2x, y2);
    } else { // Else we need to quadratic formula
      const slope = (l2y - l1y) / (l2x - l1x);
      const a = 1 / (ax * ax) + (slope * slope) / (by * by);
      const b = 2 * slope * (l1y - slope * l1x) /
        (by * by) - 2 * slope * centerY /
        (by * by) - 2 * centerX / (ax * ax);
      const c = 2 * slope * l1x * centerY /
        (by * by) - 2 * l1y * centerY / (by * by) +
        centerY * centerY / (by * by) + centerX *
        centerX / (ax * ax) - 1 + (l1y - slope * l1x) *
        (l1y - slope * l1x) / (by * by);

      // Quadratic Formula to find intersections:
      const discriminant = b * b - 4 * a * c;
      if (discriminant < 0) {
        return result; // No intersections
      }
      const disc2 = Math.sqrt(discriminant);
      const intX1 = (-b + disc2) / (2 * a);
      dist(intX1, slope * intX1 - slope * l1x + l1y);
      const intX2 = (-b - disc2) / (2 * a);
      dist(intX2, slope * intX2 - slope * l1x + l1y);
    }
    return result;
  },

  // uses no points except the result point
  /**
   * @hidden @internal
   * Find the closest point of a rectangle to a given point that is on a line from that point.
   * @param {number} l
   * @param {number} t
   * @param {number} r
   * @param {number} b
   * @param {number} Cx
   * @param {number} Cy
   * @param {number} Dx
   * @param {number} Dy
   * @param {Point} result
   * @return {boolean} true if the infinite line does intersect with the rectangle; false otherwise.
   */
  getNearestIntersectionPoint: function (l: number, t: number, r: number, b: number, Cx: number, Cy: number, Dx: number, Dy: number, result: Point): boolean {
    let closestdist = 10e20;
    let closestx = l;
    let closesty = t;
    if (Geo.nearestIntersectionOnLine(l, t, l, b, Cx, Cy, Dx, Dy, result)) {
      // calculate the (non-square-rooted) distance from C to result
      const dist = (result.x - Cx) * (result.x - Cx) + (result.y - Cy) * (result.y - Cy);
      if (dist < closestdist) {
        closestdist = dist;
        closestx = result.x;
        closesty = result.y;
      }
    }

    if (Geo.nearestIntersectionOnLine(r, t, r, b, Cx, Cy, Dx, Dy, result)) {
      const dist = (result.x - Cx) * (result.x - Cx) + (result.y - Cy) * (result.y - Cy);
      if (dist < closestdist) {
        closestdist = dist;
        closestx = result.x;
        closesty = result.y;
      }
    }

    if (Geo.nearestIntersectionOnLine(l, t, r, t, Cx, Cy, Dx, Dy, result)) {
      const dist = (result.x - Cx) * (result.x - Cx) + (result.y - Cy) * (result.y - Cy);
      if (dist < closestdist) {
        closestdist = dist;
        closestx = result.x;
        closesty = result.y;
      }
    }

    if (Geo.nearestIntersectionOnLine(l, b, r, b, Cx, Cy, Dx, Dy, result)) {
      const dist = (result.x - Cx) * (result.x - Cx) + (result.y - Cy) * (result.y - Cy);
      if (dist < closestdist) {
        closestdist = dist;
        closestx = result.x;
        closesty = result.y;
      }
    }

    result.assignTo(closestx, closesty);
    return (closestdist < 10e20);
  },

  /**
   * @hidden @internal
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {number} q1x
   * @param {number} q1y
   * @param {number} q2x
   * @param {number} q2y
   * @param {Point} result out
   * @return {Point} the result argument
   */
  getIntersection: function (p1x: number, p1y: number, p2x: number, p2y: number, q1x: number, q1y: number, q2x: number, q2y: number, result: Point): Point {
    const dx1 = p1x - p2x;
    const dx2 = q1x - q2x;
    let x = 0.0;
    let y = 0.0;

    if (dx1 === 0 || dx2 === 0) {
      if (dx1 === 0) {
        const m2 = (q1y - q2y) / dx2;
        const b2 = q1y - m2 * q1x;
        x = p1x;
        y = m2 * x + b2;
      } else {
        const m1 = (p1y - p2y) / dx1;
        const b1 = p1y - m1 * p1x;
        x = q1x;
        y = m1 * x + b1;
      }
    } else {
      const m1 = (p1y - p2y) / dx1;
      const m2 = (q1y - q2y) / dx2;
      const b1 = p1y - m1 * p1x;
      const b2 = q1y - m2 * q1x;

      x = (b2 - b1) / (m1 - m2);
      y = m1 * x + b1;
    }

    result.assignTo(x, y);
    return result;
  },

  /**
   * @hidden @internal
   * @param {Rect} rect
   * @param {Point} p1
   * @param {Point} p2
   * @return {boolean}
   */
  intersectsLineSegment: function (rect: Rect, p1: Point, p2: Point): boolean {
    const p1x = p1.x;
    const p1y = p1.y;
    const p2x = p2.x;
    const p2y = p2.y;
    const l = rect.left;
    const r = rect.right;
    const t = rect.top;
    const b = rect.bottom;
    if (p1x === p2x) {
      let min = 0.0;
      let max = 0.0;
      if (p1y < p2y) { min = p1y; max = p2y; } else { min = p2y; max = p1y; }
      return (l <= p1x && p1x <= r && min <= b && max >= t);
    }
    if (p1y === p2y) {
      let min = 0.0;
      let max = 0.0;
      if (p1x < p2x) { min = p1x; max = p2x; } else { min = p2x; max = p1x; }
      return (t <= p1y && p1y <= b && min <= r && max >= l);
    }
    if (rect.containsPoint(p1)) return true;
    if (rect.containsPoint(p2)) return true;
    if (Geo.intersectingLineSegments(l, t, r, t, p1x, p1y, p2x, p2y)) return true;
    if (Geo.intersectingLineSegments(r, t, r, b, p1x, p1y, p2x, p2y)) return true;
    if (Geo.intersectingLineSegments(r, b, l, b, p1x, p1y, p2x, p2y)) return true;
    if (Geo.intersectingLineSegments(l, b, l, t, p1x, p1y, p2x, p2y)) return true;
    return false;
  },

  /**
   * @hidden @internal
   * This predicate returns true if two finite straight line segments intersect.
   * @param {number} a1x
   * @param {number} a1y
   * @param {number} a2x
   * @param {number} a2y
   * @param {number} b1x
   * @param {number} b1y
   * @param {number} b2x
   * @param {number} b2y
   * @return {boolean}
   */
  intersectingLineSegments: function (a1x: number, a1y: number, a2x: number, a2y: number, b1x: number, b1y: number, b2x: number, b2y: number): boolean {
    return ((Geo.comparePointWithLine(a1x, a1y, a2x, a2y, b1x, b1y) * Geo.comparePointWithLine(a1x, a1y, a2x, a2y, b2x, b2y) <= 0) &&
      (Geo.comparePointWithLine(b1x, b1y, b2x, b2y, a1x, a1y) * Geo.comparePointWithLine(b1x, b1y, b2x, b2y, a2x, a2y) <= 0));
  },

  /**
   * @hidden @internal
   * @param {number} a1x
   * @param {number} a1y
   * @param {number} a2x
   * @param {number} a2y
   * @param {number} p1x
   * @param {number} p1y
   * @return {number} int -1, 0, or 1.
   */
  comparePointWithLine: function (a1x: number, a1y: number, a2x: number, a2y: number, p1x: number, p1y: number): number {
    const x2 = a2x - a1x;
    const y2 = a2y - a1y;
    let px = p1x - a1x;
    let py = p1y - a1y;
    let ccw = px * y2 - py * x2;
    if (ccw === 0) {
      ccw = px * x2 + py * y2;
      if (ccw > 0) {
        px -= x2;
        py -= y2;
        ccw = px * x2 + py * y2;
        if (ccw < 0) ccw = 0;
      }
    }
    return (ccw < 0) ? -1 : ((ccw > 0) ? 1 : 0);
  },

  /**
   * @hidden @internal
   * @param {number} a
   * @return {number}
   */
  normalizeAngle: function (a: number): number {
    if (a < 0) a += 360;
    if (a >= 360) a -= 360;
    return a;
  },

  /**
   * returns an array that represents the bezier curves needed to draw or hit-test an arc
   * @param {number} x
   * @param {number} y
   * @param {number} radius
   * @param {number} startAngle
   * @param {number} endAngle
   * @param {boolean} anglesInRadians
   * @return {Array.<Array.<number>>}
   */
  createArc: function (x: number, y: number, radius: number, startAngle: number, endAngle: number, anglesInRadians: boolean): Array<Array<number>> {
    const pi = Math.PI;
    if (!anglesInRadians) {
      // convert these to radians
      startAngle = startAngle * (pi / 180);
      endAngle = endAngle * (pi / 180);
    }

    const sign = (startAngle > endAngle) ? -1 : 1;
    const epsilon = 0.00001;
    const twoPI = pi * 2;

    // Compute the sequence of arc curves, up to PI/2 at a time.  Total arc angle
    // is less than 2PI.
    const curves: Array<Array<number>> = [];
    const piOverTwo = pi / 2.0;

    let a1 = startAngle;
    let totalAngle = Math.min(twoPI, Math.abs(endAngle - startAngle));

    // or === 0. This is to check for degenerate arcs where there is no sweep
    if (totalAngle < epsilon) {
      const a2 = a1 + sign * Math.min(totalAngle, piOverTwo);
      const p1x = x + radius * Math.cos(a1);
      const p1y = y + radius * Math.sin(a1);
      const p2x = x + radius * Math.cos(a2);
      const p2y = y + radius * Math.sin(a2);
      const pmidx = (p1x + p2x) / 2;
      const pmidy = (p1y + p2y) / 2;
      curves.push([
        p1x, p1y,
        pmidx, pmidy,
        pmidx, pmidy,
        p2x, p2y
      ]);
      return curves;
    }

    while (totalAngle > epsilon) {
      const a2 = a1 + sign * Math.min(totalAngle, piOverTwo);
      curves.push(Geo.createSmallArc(radius, a1, a2, x, y));
      totalAngle -= Math.abs(a2 - a1);
      a1 = a2;
    }

    return curves;
  },

  /**
   * Cubic bezier approximation of a circular arc centered at the origin,
   * from (radians) a1 to a2, where a2-a1 < pi/2.  The arc's radius is r.
   * @param {number} r
   * @param {number} a1
   * @param {number} a2
   * @param {number} x
   * @param {number} y
   * @return {Array.<number>}
   */
  createSmallArc: function (r: number, a1: number, a2: number, x: number, y: number): Array<number> {
    // Compute all four points for an arc that subtends the same total angle
    // but is centered on the X-axis
    const a = (a2 - a1) / 2.0;

    const x4 = r * Math.cos(a);
    const y4 = r * Math.sin(a);
    const x1 = x4;
    const y1 = -y4;

    // http://hansmuller-flex.blogspot.com/2011/10/more-about-approximating-circular-arcs.html
    // http://itc.ktu.lt/itc354/Riskus354.pdf
    const q1 = x1 * x1 + y1 * y1;
    const q2 = q1 + x1 * x4 + y1 * y4;
    const k2 = 4 / 3 * (Math.sqrt(2 * q1 * q2) - q2) / (x1 * y4 - y1 * x4);
    const x2 = x1 - k2 * y1;
    const y2 = y1 + k2 * x1;
    const x3 = x2;
    const y3 = -y2;

    // Find the arc points actual locations by computing x1,y1 and x4,y4
    // and rotating the control points by a + a1
    const ar = a + a1;
    const cosAr = Math.cos(ar);
    const sinAr = Math.sin(ar);
    return [
      x + r * Math.cos(a1),
      y + r * Math.sin(a1),
      x + x2 * cosAr - y2 * sinAr,
      y + x2 * sinAr + y2 * cosAr,
      x + x3 * cosAr - y3 * sinAr,
      y + x3 * sinAr + y3 * cosAr,
      x + r * Math.cos(a2),
      y + r * Math.sin(a2)
    ];
  },

  /**
   * @hidden @internal
   * This static function finds the nearest point on an infinite grid given the grids origin
   * and the size of the cells.
   * @param {number} x The X point to look near.
   * @param {number} y The Y point to look near.
   * @param {number} originX The X-coordinate of the origin of the grid.
   * @param {number} originY The Y-coordinate of the origin of the grid.
   * @param {number} cellwidth The width of the grid cell.
   * @param {number} cellheight The height of the grid cell.
   * @param {Point} result a Point that is modified and returned
   * @return {Point} The nearest grid point.
   */
  findNearestInfiniteGridPoint: function (x: number, y: number, originX: number, originY: number, cellwidth: number, cellheight: number, result: Point): Point {
    // the grid space looks like the figure below.
    //  +               +
    //   A(ax, ay)
    //
    //        *(x,y)
    //  +               +

    const gridW = cellwidth;
    const gridH = cellheight;

    const ax = (Math.floor((x - originX) / gridW)) * gridW + originX;
    const ay = (Math.floor((y - originY) / gridH)) * gridH + originY;

    let bestx = ax;
    if ((ax + gridW - x) < gridW / 2) {
      bestx = ax + gridW;
    }
    let besty = ay;
    if ((ay + gridH - y) < gridH / 2) {
      besty = ay + gridH;
    }

    result.assignTo(bestx, besty);
    return result;
  },

  /**
   * @hidden @internal
   * Compute the greatest common divisor of two positive integers
   * @param {number} x
   * @param {number} y
   * @return {number}
   */
  gcd: function (x: number, y: number): number {
    let a = Math.max(x, y);
    let b = Math.min(x, y);
    let c = 1;
    let result = 1;
    do {
      c = a % b;
      result = b;
      a = b;
      b = c;
    } while (c > 0);
    return result;
  },

  // The ellipse is (x0/e0)^2 + (x1/e1)^2 = 1.  The query point is (y0,y1).
  // The function returns the distance from the query point to the ellipse.
  /**
   * @hidden @internal
   * @param {number} e0
   * @param {number} e1
   * @param {number} p0
   * @param {number} p1
   * @return {number}
   */
  distancePointEllipse: function (e0: number, e1: number, p0: number, p1: number): number { // e[2], p[2], x[2]
    // Determine reflections for p to the first quadrant.
    const reflect0 = (p0 < 0);
    const reflect1 = (p1 < 0);
    let j = 0;

    // Determine the axis order for decreasing extents.
    let permute0 = 0;
    let permute1 = 0;
    if (e0 < e1) {
      permute0 = 1; permute1 = 0;
    } else {
      permute0 = 0; permute1 = 1;
    }

    let locE0 = 0.0;
    let locE1 = 0.0;
    let locY0 = 0.0;
    let locY1 = 0.0;
    j = permute0;
    locE0 = j === 0 ? e0 : e1;
    locY0 = j === 0 ? p0 : p1;
    if (j === 0 ? reflect0 : reflect1) {
      locY0 = -locY0;
    }
    j = permute1;
    locE1 = j === 0 ? e0 : e1;
    locY1 = j === 0 ? p0 : p1;
    if (j === 0 ? reflect0 : reflect1) {
      locY1 = -locY1;
    }

    const distance = Geo.distancePointEllipseHelper(locE0, locE1, locY0, locY1, 0, 0);
    return distance;
  },

  // The ellipse is (x0/e0)^2 + (x1/e1)^2 = 1 with e0 >= e1.  The query point is
  // (y0,y1) with y0 >= 0 and y1 >= 0.  The function returns the distance from
  // the query point to the ellipse.  It also computes the ellipse point (x0,x1)
  // in the first quadrant that is closest to (y0,y1).
  /**
   * @hidden @internal
   * @param {number} e0
   * @param {number} e1
   * @param {number} y0
   * @param {number} y1
   * @param {number} x0
   * @param {number} x1
   * @return {number}
   */
  distancePointEllipseHelper: function (e0: number, e1: number, y0: number, y1: number, x0: number, x1: number): number { // e[2], y[2], x[2]
    let distance = 0.0;
    if (y1 > 0) {
      if (y0 > 0) {
        // Bisect to compute the root of F(t) for t >= -e1*e1.
        const esqr0 = e0 * e0;
        const esqr1 = e1 * e1;
        const ey0 = e0 * y0;
        const ey1 = e1 * y1;
        let t0 = -esqr1 + ey1;
        let t1 = -esqr1 + Math.sqrt(ey0 * ey0 + ey1 * ey1);
        let t = t0;
        const imax = 9999999999; // ???
        for (let i = 0; i < imax; ++i) {
          t = (0.5) * (t0 + t1);
          if (t === t0 || t === t1) {
            break;
          }

          const r0 = ey0 / (t + esqr0);
          const r1 = ey1 / (t + esqr1);
          const f = r0 * r0 + r1 * r1 - 1;
          if (f > 0) {
            t0 = t;
          } else if (f < 0) {
            t1 = t;
          } else {
            break;
          }
        }

        x0 = esqr0 * y0 / (t + esqr0);
        x1 = esqr1 * y1 / (t + esqr1);
        const d0 = x0 - y0;
        const d1 = x1 - y1;
        distance = Math.sqrt(d0 * d0 + d1 * d1);
      } else { // y0 === 0
        x0 = 0;
        x1 = e1;
        distance = Math.abs(y1 - e1);
      }
    } else { // y1 === 0
      const denom0 = e0 * e0 - e1 * e1;
      const e0y0 = e0 * y0;
      if (e0y0 < denom0) {
        // y0 is inside the subinterval.
        const x0de0 = e0y0 / denom0;
        const x0de0sqr = x0de0 * x0de0;
        x0 = e0 * x0de0;
        x1 = e1 * Math.sqrt(Math.abs(1 - x0de0sqr));
        const d0 = x0 - y0;
        distance = Math.sqrt(d0 * d0 + x1 * x1);
      } else {
        // y0 is outside the subinterval.  The closest ellipse point has
        // x1 === 0 and is on the domain-boundary interval (x0/e0)^2 = 1.
        x0 = e0;
        x1 = 0;
        distance = Math.abs(y0 - e0);
      }
    }
    return distance;
  },

  /**
   * @hidden @internal
   */
  GeometryGenerators: new PropertyCollection(),

  /**
   * @hidden @internal
   */
  ArrowheadGeometryStrings: new PropertyCollection(),

  /**
   * @hidden @internal
   */
  CachedGeometries: new PropertyCollection(),
  /**
   * @hidden @internal
   */
  CachedGeometryCount: 0


}; // end Geo static object
(Geo as any)['za'] = Geo.wmq4;


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Geometry
PathFigure
PathSegment
*/

// import { EnumValue } from './enumValue';
// import { Util } from './util';

/**
 * The Geometry class is used to define the "shape" of a {@link Shape}.
 * A Geometry can be simple straight lines, rectangles, or ellipses.
 * A Geometry can also be an arbitrarily complex path, consisting of a list of {@link PathFigure}s.
 *
 * A Geometry must not be modified once it has been used by a {@link Shape}.
 * However, a Geometry may be shared by multiple Shapes.
 *
 * It is commonplace to create Geometries using geometry path string syntax:
 * <a href="../../intro/geometry.html">Geometry Path Strings</a>.
 * However it is much more efficient to create Geometries programmatically.
 * One way to do that is illustrated by several of the samples that evaluate JavaScript such as:
 * ```js
 * new go.Geometry()
 *     .add(new go.PathFigure(p.x, p.y)
 *          .add(new go.PathSegment(go.PathSegment.Arc, -sweep/2, sweep, 0, 0, radius+layerThickness, radius+layerThickness))
 *          .add(new go.PathSegment(go.PathSegment.Line, q.x, q.y))
 *          .add(new go.PathSegment(go.PathSegment.Arc, sweep/2, -sweep, 0, 0, radius, radius).close()));
 * ```
 * See samples that make use of Geometries in the <a href="../../samples/index.html#geometries">samples index</a>.
 * @unrestricted
 * @category Geometry
 */
export class Geometry {
  /** @hidden @internal */ public static className = 'Geometry';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ private _figuresChanged: number;
  /** @hidden @internal */ private _computedBounds: Rect;
  /** @hidden @internal */ private _invalidBounds: boolean;
  /** @hidden @internal */ private _allPathsSegments: Array<Array<number>> | null;
  /** @hidden @internal */ private _allPathsLengths: Array<Array<number>> | null;
  /** @hidden @internal */ private _allPathsTotalLength: number;
  /** @hidden @internal */ private _spot1: Spot;
  /** @hidden @internal */ private _spot2: Spot;
  /** @hidden @internal */ private _defaultStretch: EnumValue;
  /** @hidden @internal */ private _type: EnumValue;
  /** @hidden @internal */ private _startX: number;
  /** @hidden @internal */ private _startY: number;
  /** @hidden @internal */ private _endX: number;
  /** @hidden @internal */ private _endY: number;
  /** @hidden @internal */ private _figures: List<PathFigure>;
  /** @hidden @internal */ public _inputWidth: number;
  /** @hidden @internal */ public _inputHeight: number;

  /**
   * @param {?=} type
   * Construct an empty Geometry.
   * The geometry type must be one of the following values:
   * {@link Geometry.Line}, {@link Geometry.Ellipse}, {@link Geometry.Rectangle}, {@link Geometry.Path}.
   */
  constructor(type?: EnumValue) {
    if (Debug && arguments.length > 1) {
      Util.throwError('Geometry constructor can take at most one optional argument, the Geometry type.');
    }
    Set.uniqueHash(this);
    this._isFrozen = false;
    if (type === undefined) {
      type = Geometry.Path;
    } else {
      if (Debug) EnumValue.checkEnumValue(type, Geometry, Geometry, 'constructor:type');
    }
    this._type = type;

    // for Line, Rectangle, Ellipse:
    this._startX = 0;
    this._startY = 0;
    this._endX = 0;
    this._endY = 0;
    // for Path:
    this._figures = new List<PathFigure>();
    this._figuresChanged = this._figures._changed;

    this._computedBounds = new Rect().freeze();
    this._invalidBounds = true;
    this._allPathsSegments = null;
    this._allPathsLengths = null;
    this._allPathsTotalLength = NaN;

    this._spot1 = Spot.TopLeft;
    this._spot2 = Spot.BottomRight;

    // When a geometry is created the input width and height are saved,
    // to ensure that computed bounds are at least that size
    // That way, a 2D horizontal line with a demanded height of "50" will still be 50 tall.
    this._inputWidth = NaN;
    this._inputHeight = NaN;

    // For some figures this is Uniform
    this._defaultStretch = GraphObject.Fill;
  }

  /**
   * Create a copy of this Geometry, with the same values and figures.
   * @expose
   * @this {Geometry}
   * @return {Geometry}
   */
  public copy(): Geometry {
    const copy = new Geometry();
    copy._type = this._type;
    copy._startX = this._startX;
    copy._startY = this._startY;
    copy._endX = this._endX;
    copy._endY = this._endY;
    const oldfigs = this._figures._arr;
    const numfigs = oldfigs.length;
    const newfigs = copy._figures;
    for (let i = 0; i < numfigs; i++) {
      const oldfig = oldfigs[i];
      const newfig = oldfig.copy();
      newfigs.add(newfig);
    }
    copy._figuresChanged = this._figuresChanged;
    copy._computedBounds.assign(this._computedBounds);
    copy._invalidBounds = this._invalidBounds;
    copy._allPathsSegments = this._allPathsSegments;
    copy._allPathsLengths = this._allPathsLengths;
    copy._allPathsTotalLength = this._allPathsTotalLength;
    copy._spot1 = this._spot1.copyFrozen();
    copy._spot2 = this._spot2.copyFrozen();
    copy._inputWidth = this._inputWidth;
    copy._inputHeight = this._inputHeight;
    copy._defaultStretch = this._defaultStretch;
    return copy;
  }

  /**
   * For drawing a simple straight line;
   * a value for {@link Geometry#type}.
   * The line goes from the point
   * ({@link #startX}, {@link #startY}) to the point ({@link #endX}, {@link #endY}).
   * @constant
   */
  public static Line = new EnumValue(Geometry, 'Line', 0);

  /**
   * For drawing a rectangle;
   * a value for {@link Geometry#type}.
   * The rectangle goes from the point
   * ({@link #startX}, {@link #startY}) to the point ({@link #endX}, {@link #endY}).
   * @constant
   */
  public static Rectangle = new EnumValue(Geometry, 'Rectangle', 1);

  /**
   * For drawing an ellipse fitting within a rectangle;
   * a value for {@link Geometry#type}.
   * The ellipse fits within the rectangle that goes from the point
   * ({@link #startX}, {@link #startY}) to the point ({@link #endX}, {@link #endY}).
   * @constant
   */
  public static Ellipse = new EnumValue(Geometry, 'Ellipse', 2);

  /**
   * For drawing a complex path made of a list of {@link PathFigure}s;
   * a value for {@link Geometry#type}.
   * @constant
   */
  public static Path = new EnumValue(Geometry, 'Path', 3);

  /**
   * @hidden @internal
   * Make this object 'constant' and 'frozen'.
   * Future attempts to change its state may result in an exception.
   * This constant object may be shared freely.
   * @this {Geometry}
   * @return {Geometry}
   */
  public constant(): Geometry {
    this.freeze();
    Object.freeze(this);
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Geometry}
   * @return {Geometry}
   */
  public freeze(): Geometry {
    this._isFrozen = true;
    const figs = this.figures;
    figs.freeze();
    const arr = figs._arr;
    const numfigs = arr.length;
    for (let i = 0; i < numfigs; i++) {
      const fig = arr[i];
      fig.freeze();
    }
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * Caution: this object might be shared, which would mean that any changes
   * to this object would need to update all users of this object.
   * @this {Geometry}
   * @return {Geometry}
   */
  public thaw(): Geometry {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    const figs = this.figures;
    figs.thaw();
    const arr = figs._arr;
    const numfigs = arr.length;
    for (let i = 0; i < numfigs; i++) {
      const fig = arr[i];
      fig.thaw();
    }
    return this;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Geometry}
   * @param {Geometry} g
   * @return {boolean}
   */
  public equalsApprox(g: Geometry): boolean {
    if (!(g instanceof Geometry)) return false;
    if (this.type !== g.type) {
      if (this.type === Geometry.Line && g.type === Geometry.Path) {
        return this.equalsApproxLine(g);
      } else if (g.type === Geometry.Line && this.type === Geometry.Path) {
        return g.equalsApproxLine(this);
      }
      return false;
    }
    if (this.type === Geometry.Path) {
      const figs1 = this.figures._arr;
      const figs2 = g.figures._arr;
      const numfigs = figs1.length;
      if (numfigs !== figs2.length) return false;
      for (let i = 0; i < numfigs; i++) {
        const fig1 = figs1[i];
        const fig2 = figs2[i];
        if (!fig1.equalsApprox(fig2)) return false;
      }
      return true;
    } else {
      // don't bother with spot1, spot2, defaultStretch
      return Geo.isApprox(this.startX, g.startX) &&
        Geo.isApprox(this.startY, g.startY) &&
        Geo.isApprox(this.endX, g.endX) &&
        Geo.isApprox(this.endY, g.endY);
    }
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   * @param {Geometry} g
   * @return {boolean}
   */
  public equalsApproxLine(g: Geometry): boolean {
    if (this.type !== Geometry.Line) return false;
    if (g.type !== Geometry.Path) return false;
    if (g.figures.count === 1) {
      const fig = g.figures.elt(0);
      if (fig.segments.count === 1 &&
        Geo.isApprox(this.startX, fig.startX) &&
        Geo.isApprox(this.startY, fig.startY)) {
        const seg = fig.segments.elt(0);
        if (seg.type === PathSegment.Line &&
          Geo.isApprox(this.endX, seg.endX) &&
          Geo.isApprox(this.endY, seg.endY)) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * This static function can be used to write out a {@link Geometry} as a string
   * that can be read by {@link Geometry.parse}.
   *
   * The string produced by this method is a superset of the SVG path
   * string rules that contains some additional GoJS-specific tokens.
   * See the <a href="../../intro/pictures.html">Introduction page on Geometry Parsing</a> for more details.
   * @param {Geometry} val
   * @return {string}
   * @since 1.1
   */
  public static stringify(val: Geometry): string {
    return val.toString();
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Geometry}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Geometry) {
      this.type = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   * @param {number=} precision Optional: the amount of decimal places to round numbers to for debugging purposes,
   * leaves number unchanged by default.
   * @return {string}
   */
  public toString(precision?: number): string {
    if (precision === undefined) precision = -1;
    let r;
    switch (this.type) {
      case Geometry.Line:
        if (precision < 0) {
          return 'M' + this.startX.toString() + ' ' + this.startY.toString() +
            'L' + this.endX.toString() + ' ' + this.endY.toString();
        } else {
          return 'M' + this.startX.toFixed(precision) + ' ' + this.startY.toFixed(precision) +
            'L' + this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
      case Geometry.Rectangle:
        r = new Rect(this.startX, this.startY, 0, 0);
        r.union(this.endX, this.endY, 0, 0);
        if (precision < 0) {
          return 'M' + r.x.toString() + ' ' + r.y.toString() +
            'H' + r.right.toString() +
            'V' + r.bottom.toString() +
            'H' + r.left.toString() +
            'z';
        } else {
          return 'M' + r.x.toFixed(precision) + ' ' + r.y.toFixed(precision) +
            'H' + r.right.toFixed(precision) +
            'V' + r.bottom.toFixed(precision) +
            'H' + r.left.toFixed(precision) +
            'z';
        }
      case Geometry.Ellipse:
        r = new Rect(this.startX, this.startY, 0, 0);
        r.union(this.endX, this.endY, 0, 0);
        if (precision < 0) {
          const p1 = r.left.toString() + ' ' + (r.y + r.height / 2).toString();
          const p2 = r.right.toString() + ' ' + (r.y + r.height / 2).toString();
          return 'M' + p1 +
            'A' + (r.width / 2).toString() + ' ' + (r.height / 2).toString() + ' 0 0 1 ' + p2 +
            'A' + (r.width / 2).toString() + ' ' + (r.height / 2).toString() + ' 0 0 1 ' + p1;
        } else {
          const p1 = r.left.toFixed(precision) + ' ' + (r.y + r.height / 2).toFixed(precision);
          const p2 = r.right.toFixed(precision) + ' ' + (r.y + r.height / 2).toFixed(precision);
          return 'M' + p1 +
            'A' + (r.width / 2).toFixed(precision) + ' ' + (r.height / 2).toFixed(precision) + ' 0 0 1 ' + p2 +
            'A' + (r.width / 2).toFixed(precision) + ' ' + (r.height / 2).toFixed(precision) + ' 0 0 1 ' + p1;
        }
      case Geometry.Path:
        let str = '';
        const figs = this.figures._arr;
        const numfigs = figs.length;
        for (let i = 0; i < numfigs; i++) {
          const fig = figs[i];
          if (i > 0) str += ' x ';
          if (fig.isFilled) str += 'F ';
          str += fig.toString(precision);
        }
        return str;
      default:
        return this.type.toString();
    } // end type switch
  }

  /**
   * Given a SVG or GoJS path string, returns a congruent path string with each pathfigure filled.
   * For instance, "M0 0 L22 22 L33 0" would become "F M0 0 L22 22 L33 0".
   * @param {string} str
   * @return {string}
   * @since 1.1
   */
  public static fillPath(str: string): string {
    if (typeof str !== 'string') Util.throwTypeError(str, 'string', Geometry, 'fillPath:str');
    // split string by any x commands
    const paths = str.split(/[Xx]/);
    const l = paths.length;
    let r = '';
    for (let i = 0; i < l; i++) {
      const p = paths[i];
      if (p.match(/[Ff]/) !== null) {
        if (i === 0) {
          r += p;
        } else {
          r += 'X' + ((p[0] === ' ') ? '' : ' ') + p;
        }
        continue;
      }
      r += (i === 0 ? '' : 'X ') + 'F' + ((p[0] === ' ') ? '' : ' ') + p;
    }
    return r;
  }

  /**
   * Produce a {@link Geometry} from a string that uses an SVG-like compact path geometry syntax.
   * The syntax accepts all SVG Path Syntax (<a href="https://www.w3.org/TR/SVG/paths.html" target="_blank">SVG Path Syntax (w3.org)</a>),
   * as well as three GoJS-specific tokens.
   * Specifically, the following tokens are allowed:
   *   - `M (x,y)+ ` - Move commands
   *   - `L (x,y)+, H (x)+, V (y)+` Lines commands, including horizontal and vertical lines
   *   - `C (x1 y1 x2 y2 x y)+, S (x2 y2 x y)+` Cubic bezier curves
   *   - `Q (x1 y1 x y)+, T (x y)+` Quadratic bezier curves
   *   - `A (rx ry x-axis-rotation large-arc-flag clockwise-flag x y)+` Arcs (following SVG arc conventions)
   *   - `Z` Denotes that the previous subpath is closed
   *
   * Additionally there are some tokens specific to GoJS:
   *   - `B (startAngle, sweepAngle, centerX, centerY, radius)+` Arcs following GoJS canvas arc conventions
   *   - `X` Used before M-commands to denote separate PathFigures instead of a subpath
   *   - `F` Denotes whether the current PathFigure is filled (true if F is present)
   *   - `U` Denotes that the PathFigure is not shadowed
   *
   * See the <a href="../../intro/geometry.html">Introduction page on Geometry Parsing</a> for more details.
   * @param {string} str
   * @param {boolean=} filled whether figures should be filled.
   * If true, all PathFigures in the string will be filled regardless of the presence
   * of an "F" command or not.
   * If false, all PathFigures will determine their own filled state by the presence of an "F" command or not.
   * Default is false.
   * @return {Geometry}
   */
  public static parse(str: string, filled?: boolean): Geometry {
    if (filled === undefined) filled = false;
    if (typeof str !== 'string') Util.throwTypeError(str, 'string', Geometry, 'parse:str');
    // ??? check floating points
    str = str.replace(/,/gm, ' '); // get rid of all commas
    str = str.replace(/([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])/gm, '$1 $2'); // separate commands from commands
    str = str.replace(/([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])/gm, '$1 $2'); // separate commands from commands
    str = str.replace(/([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])([^\s])/gm, '$1 $2'); // separate commands from points
    str = str.replace(/([^\s])([UuBbMmZzLlHhVvCcSsQqTtAaFfXx])/gm, '$1 $2'); // separate commands from points
    str = str.replace(/([0-9])([+\-])/gm, '$1 $2'); // separate digits when no comma
    str = str.replace(/([Aa](\s+[0-9]+){3})\s+([01])\s*([01])/gm, '$1 $3 $4 '); // shorthand elliptical arc path syntax
    str = str.replace(/[\s\r\t\n]+/gm, ' '); // compress spaces
    str = str.replace(/^\s+|\s+$/g, ''); // trim

    const tokens = str.split(' ');

    // deal with decimal numbers and multiple '.'s without spaces, eg. 1.752.827.53
    for (let j = 0; j < tokens.length; j++) {
      const token = tokens[j];
      if (token.match(/(\.[0-9]*)(\.)/gm) !== null) {
        const split = Util.tempArray<string>();
        let num = '';
        let decConsumed = false;
        for (let k = 0; k < token.length; k++) {
          const char = token[k];
          if (char === '.' && !decConsumed) {
            decConsumed = true;
            num += char;
          } else if (char === '.') {
            split.push(num);
            num = '.';
          } else {
            num += char;
          }
        }
        split.push(num);
        // tokens.splice(j, 1, ...split);
        // no spread expressions in ES6, don't want polyfill code to be inserted by compiler either, instead do:
        tokens.splice(j, 1);
        for (let ii = 0; ii < split.length; ii++) {
          tokens.splice(j + ii, 0, split[ii]);
        }

        j += split.length - 1;
        Util.freeArray(split);
      }
    }

    let i = -1;
    let command = '';
    let previousCommand = '';
    let start = new Point(0, 0);
    let control = new Point(0, 0);
    let current = new Point(0, 0);

    // things done often:
    // returns true if the next item is a command
    /** @hidden @internal */
    function isCommandOrEnd(): boolean {
      if (i >= length - 1) return true;
      return tokens[i + 1].match(/[UuBbMmZzLlHhVvCcSsQqTtAaFfXx]/) !== null;
    }

    /** @hidden @internal */
    function getToken(): string {
      i++;
      return tokens[i];
    }

    /** @hidden @internal */
    function getPoint(): Point {
      const pt = new Point(parseFloat(getToken()), parseFloat(getToken()));
      // make it absolute
      if (command === command.toLowerCase()) {
        pt.x = current.x + pt.x;
        pt.y = current.y + pt.y;
      }
      return pt;
    }

    /** @hidden @internal */
    function getAsCurrentPoint(): Point {
      current = getPoint();
      return current;
    }

    /** @hidden @internal */
    function getAsControlPoint(): Point {
      control = getPoint();
      return control;
    }

    /** @hidden @internal */
    function getReflectedControlPoint(): Point {
      const prev = previousCommand.toLowerCase();
      if (!(prev === 'c' || prev === 's' || prev === 'q' || prev === 't')) {
        return current;
      }
      // reflect point
      return new Point(2 * current.x - control.x, 2 * current.y - control.y);
    }

    const length = tokens.length;
    const ctx = StreamGeometryContext.tempStreamGeometryContext(null);

    let fCommand = false;
    let uCommand = false;
    let newfig = true;
    let p: Point = null;
    while (!(i >= length - 1)) {
      previousCommand = command;
      command = getToken();
      if (command === '') continue;

      switch (command.toUpperCase()) {
        case 'X':
          newfig = true;
          fCommand = false;
          uCommand = false;
          break;
        case 'M':
          p = getAsCurrentPoint();
          if (ctx._fig === null || newfig === true) {
            // uCommand means no shadow, so !uCommand for isShadowed
            ctx.beginFigure(p.x, p.y, fCommand, false, !uCommand);
            newfig = false;
          } else {
            ctx.moveTo(p.x, p.y);
          }
          // multiple M params mean subsequent line segments:
          start = current;
          while (!isCommandOrEnd()) {
            p = getAsCurrentPoint();
            ctx.lineTo(p.x, p.y);
          }
          break;
        case 'L':
          while (!isCommandOrEnd()) {
            p = getAsCurrentPoint();
            ctx.lineTo(p.x, p.y);
          }
          break;
        case 'H':
          while (!isCommandOrEnd()) {
            const newP = new Point((command === command.toLowerCase() ? current.x : 0) + parseFloat(getToken()), current.y);
            current = newP;
            ctx.lineTo(current.x, current.y);
          }
          break;
        case 'V':
          while (!isCommandOrEnd()) {
            const newP = new Point(current.x, (command === command.toLowerCase() ? current.y : 0) + parseFloat(getToken()));
            current = newP;
            ctx.lineTo(current.x, current.y);
          }
          break;
        case 'C':
          while (!isCommandOrEnd()) {
            const p1 = getPoint();
            const cntrl = getAsControlPoint();
            const cp = getAsCurrentPoint();
            ctx.bezierTo(p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
          }
          break;
        case 'S':
          while (!isCommandOrEnd()) {
            const p1 = getReflectedControlPoint();
            const cntrl = getAsControlPoint();
            const cp = getAsCurrentPoint();
            ctx.bezierTo(p1.x, p1.y, cntrl.x, cntrl.y, cp.x, cp.y);
          }
          break;
        case 'Q':
          while (!isCommandOrEnd()) {
            const cntrl = getAsControlPoint();
            const cp = getAsCurrentPoint();
            ctx.quadraticBezierTo(cntrl.x, cntrl.y, cp.x, cp.y);
          }
          break;
        case 'T':
          while (!isCommandOrEnd()) {
            const cntrl = getReflectedControlPoint();
            control = cntrl;
            const cp = getAsCurrentPoint();
            ctx.quadraticBezierTo(cntrl.x, cntrl.y, cp.x, cp.y);
          }
          break;
        case 'B':
          while (!isCommandOrEnd()) {
            // 5 or 7 arguments
            // startAngle, sweepAngle, centerX, centerY, radiusX, radiusY)
            const startAngle = parseFloat(getToken());
            const sweepAngle = parseFloat(getToken());
            let centerX = parseFloat(getToken());
            let centerY = parseFloat(getToken());
            const radiusX = parseFloat(getToken());
            let radiusY = radiusX;
            let close = false;
            if (!isCommandOrEnd()) {
              radiusY = parseFloat(getToken());
              if (!isCommandOrEnd()) close = parseFloat(getToken()) !== 0;
            }

            // make it absolute
            if (command === command.toLowerCase()) {
              centerX += current.x;
              centerY += current.y;
            }

            ctx.arcTo(startAngle, sweepAngle, centerX, centerY, radiusX, radiusY, close);
          }
          break;
        case 'A': // elliptical arc
          while (!isCommandOrEnd()) {
            // rx, ry is scalar, not a literal point
            const rx = Math.abs(parseFloat(getToken()));
            const ry = Math.abs(parseFloat(getToken()));
            const xAxisRotation = parseFloat(getToken());
            const largeArcFlag = !!parseFloat(getToken());
            const clockwiseFlag = !!parseFloat(getToken());
            const cp = getAsCurrentPoint();

            ctx.svgArcTo(rx, ry, xAxisRotation, largeArcFlag, clockwiseFlag, cp.x, cp.y);
          }
          break;
        case 'Z':
          /* original*/
          // ctx.closePath();
          // current = start;

          // !original: closes path
          // const fig = ctx.geo.figures._arr[ctx.geo.figures.length - 1];
          ctx.closeLast();
          current = start;
          break;
        case 'F': {
          let nextToken = '';
          let j = 1;
          while (tokens[i + j]) {
            if (tokens[i + j].match(/[Uu]/) !== null) { j++; continue; }
            if (tokens[i + j].match(/[UuBbMmZzLlHhVvCcSsQqTtAaFfXx]/) === null) { j++; continue; }
            nextToken = tokens[i + j];
            break;
          }
          if (nextToken.match(/[Mm]/)) {
            fCommand = true;
          } else {
            ctx.fillLast();
          }
          break;
        }
        case 'U': {
          let nextToken = '';
          let j = 1;
          while (tokens[i + j]) {
            if (tokens[i + j].match(/[Ff]/) !== null) { j++; continue; }
            if (tokens[i + j].match(/[UuBbMmZzLlHhVvCcSsQqTtAaFfXx]/) === null) { j++; continue; }
            nextToken = tokens[i + j];
            break;
          }
          if (nextToken.match(/[Mm]/)) {
            uCommand = true;
          } else {
            ctx.setShadow(false);
          }

          break;
        }
        default:
          // the 1 and 2 of F1/F2 currently fall here
          // Util.throwError('Unknown geometry command: ' + command);
          break;
      }
    } // end switch

    const geo = ctx.geo;
    StreamGeometryContext.freeStreamGeometryContext(ctx);
    if (filled) {
      const it = geo.figures.iterator;
      while (it.next()) {
        const fig = it.value;
        fig.isFilled = true;
      }
    }
    return geo;
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   * @param {Array.<Array.<number>>} curves
   * @param {Transform} tt
   */
  public static transformCurves(curves: Array<Array<number>>, tt: Transform): void {
    const l = curves.length;
    const p = Point.alloc();
    for (let i = 0; i < l; i++) {
      const c = curves[i];
      p.x = c[0];
      p.y = c[1];
      tt.transformPoint(p);
      c[0] = p.x;
      c[1] = p.y;

      p.x = c[2];
      p.y = c[3];
      tt.transformPoint(p);
      c[2] = p.x;
      c[3] = p.y;

      p.x = c[4];
      p.y = c[5];
      tt.transformPoint(p);
      c[4] = p.x;
      c[5] = p.y;

      p.x = c[6];
      p.y = c[7];
      tt.transformPoint(p);
      c[6] = p.x;
      c[7] = p.y;
    }
    Point.free(p);
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   * @return {boolean}
   */
  private isInvalidBounds(): boolean {
    if (this._invalidBounds) return true;
    const figs = this.figures;
    if (this._figuresChanged !== figs._changed) return true;
    // for each PathFigure:
    const arr = this.figures._arr;
    const l = arr.length;
    for (let i = 0; i < l; i++) {
      const fig = arr[i];
      if (fig.isInvalidBounds()) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   */
  private computeBounds(): void {
    this._invalidBounds = false;
    this._allPathsSegments = null;  // invalidate _allPathsSegments
    this._allPathsLengths = null;
    this._allPathsTotalLength = NaN;
    const figs = this.figures;
    this._figuresChanged = figs._changed;
    // for each PathFigure:
    const arr = this.figures._arr;
    const l = arr.length;
    for (let i = 0; i < l; i++) {
      const fig = arr[i];
      fig.validateBounds();
    }

    const b = this._computedBounds;
    b.thaw();
    if (isNaN(this._inputWidth) || isNaN(this._inputHeight)) {
      b.assignTo(0, 0, 0, 0);
    } else {
      b.assignTo(0, 0, this._inputWidth, this._inputHeight);
    }

    this.computeBoundsInternal(b, false);
    b.unionNoCheck(0, 0, 0, 0);  // always include the origin
    b.freeze();
  }

  /**
   * Computes the Geometry's bounds without adding an origin point, and returns those bounds as a rect.
   * This method does not modify the Geometry or its bounds.
   * @export
   * @this {Geometry}
   * @return {Rect}
   * @since 1.1
   */
  public computeBoundsWithoutOrigin(): Rect {
    const b = new Rect();
    this.computeBoundsInternal(b, true);
    return b;
  }

  /**
   * @hidden @internal
   * Modifies a given rectangle to be the bounds
   * @this {Geometry}
   * @param {Rect} b the rectangle to modify.
   * @param {boolean} set whether or not the first point is set or unioned.
   * Use true only if you don't want the existing rectangle (ie 0,0,0,0) to be included.
   */
  private computeBoundsInternal(b: Rect, set: boolean): void {
    switch (this.type) {
      case Geometry.Line:
      case Geometry.Rectangle:
      case Geometry.Ellipse:
        if (set) b.assignTo(this._startX, this._startY, 0, 0);
        else b.unionNoCheck(this._startX, this._startY, 0, 0);
        b.unionNoCheck(this._endX, this._endY, 0, 0);
        break;
      case Geometry.Path:
        const figs = this.figures;
        // for each PathFigure:
        const farr = figs._arr;
        const numfigs = figs.length;
        for (let i = 0; i < numfigs; i++) {
          const fig = farr[i];
          if (set && i === 0) b.assignTo(fig.startX, fig.startY, 0, 0);
          else b.unionNoCheck(fig.startX, fig.startY, 0, 0);
          // for each segment of the PathFigure:
          const segments = fig.segments;
          const sarr = segments._arr;
          const numsegs = sarr.length;
          let lastpointX = fig.startX;
          let lastpointY = fig.startY;
          for (let j = 0; j < numsegs; j++) {
            const seg = sarr[j];
            switch (seg.type) {
              case PathSegment.Line:
              case PathSegment.Move:
                lastpointX = seg.endX;
                lastpointY = seg.endY;
                b.unionNoCheck(lastpointX, lastpointY, 0, 0);
                break;
              case PathSegment.Bezier:
                Geo.bezierBounds(lastpointX, lastpointY, seg.point1X, seg.point1Y,
                  seg.point2X, seg.point2Y, seg.endX, seg.endY, 0.5, b); // ?? epsilon
                lastpointX = seg.endX;
                lastpointY = seg.endY;
                break;
              case PathSegment.QuadraticBezier:
                Geo.quadraticBezierBounds(lastpointX, lastpointY, seg.point1X, seg.point1Y,
                  seg.endX, seg.endY, 0.5, b); // ?? epsilon
                lastpointX = seg.endX;
                lastpointY = seg.endY;
                break;
              case PathSegment.Arc:
              case PathSegment.SvgArc:
                const curves = (seg.type === PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, lastpointX, lastpointY);
                const clen = curves.length;
                if (clen === 0) {
                  lastpointX = seg.centerX;
                  lastpointY = seg.centerY;
                  b.unionNoCheck(lastpointX, lastpointY, 0, 0);
                  break;
                }
                let c: Array<number> = null;
                for (let k = 0; k < clen; k++) {
                  c = curves[k];
                  Geo.bezierBounds(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], 0.5, b);
                }
                if (c !== null) {
                  lastpointX = c[6];
                  lastpointY = c[7];
                }
                break;
              default:
                Util.throwError('Unknown Segment type: ' + seg.type);
                break;
            }
          } // end segments
        } // end figures
        break;
      default:
        Util.throwError('Unknown Geometry type: ' + this.type);
        break;
    }
  }

  /**
   * @hidden @internal
   * Returns a copy of this Geometry that has been scaled by an x and y value
   * @this {Geometry}
   * @param {number} width
   * @param {number} height
   * @return {Geometry}
   */
  public computeScaledGeometry(width: number, height: number): Geometry {
    const result = this.copy();
    const bounds = this.bounds;
    let x = width / bounds.width;
    let y = height / bounds.height;
    // bounds might be zero or input might be Infinity, we don't want this
    if (!isFinite(x)) x = 1;
    if (!isFinite(y)) y = 1;
    if (x !== 1 || y !== 1) result.scale(x, y);
    return result;
  }

  /**
   * Normalizes the Geometry points in place by ensuring the top-left bounds of the geometry lines up with (0, 0),
   * returning the {@link Point} (x, y) amount it was shifted.
   * After this method is called there will be no negative X or Y value for the Geometry's bounds,
   * and no empty space at the top-left of the Geometry.
   * @export
   * @this {Geometry}
   * @return {Point}
   * @since 1.1
   */
  public normalize(): Point {
    if (this._isFrozen) Util.throwFrozen(this);
    const r = this.computeBoundsWithoutOrigin();
    this.offset(-r.x, -r.y);
    return new Point(-r.x, -r.y);
  }

  /**
   * Offsets the Geometry in place by a given (x, y) amount
   * @export
   * @this {Geometry}
   * @param {number} x The x-axis offset factor.
   * @param {number} y The y-axis offset factor.
   * @return {Geometry} this
   * @since 1.1
   */
  public offset(x: number, y: number): Geometry {
    if (this._isFrozen) Util.throwFrozen(this);
    if (Debug) {
      Util.checkRealNumber(x, Geometry, 'offset');
      Util.checkRealNumber(y, Geometry, 'offset');
    }
    this.transform(1, 0, 0, 1, x, y);
    return this;
  }

  /**
   * Scales the Geometry in place by a given (x, y) scale factor.
   *
   * If you want to flip a Geometry horizontally, call `geo.scale(-1, 1)`.
   * If you want to flip a Geometry vertically, call `geo.scale(1, -1)`.
   * @export
   * @this {Geometry}
   * @param {number} x The x-axis scale factor.
   * @param {number} y The y-axis scale factor.
   * @return {Geometry} this
   * @since 1.1
   */
  public scale(x: number, y: number): Geometry {
    if (this._isFrozen) Util.throwFrozen(this);
    if (Debug) {
      Util.checkRealNumber(x, Geometry, 'scale:x');
      Util.checkRealNumber(y, Geometry, 'scale:y');
      if (x === 0) Util.throwRangeError(x, 'scale must be non-zero', Geometry, 'scale:x');
      if (y === 0) Util.throwRangeError(y, 'scale must be non-zero', Geometry, 'scale:y');
    }
    this.transform(x, 0, 0, y, 0, 0);
    return this;
  }

  /**
   * Rotates the Geometry in place by a given angle, with optional x and y values to rotate the geometry about.
   * If no x and y value are given, (0, 0) is used as the rotation point.
   * @export
   * @this {Geometry}
   * @param {number} angle The angle to rotate by.
   * @param {number=} x The optional X point to rotate the geometry about. If no point is given, this value is 0.
   * @param {number=} y The optional Y point to rotate the geometry about. If no point is given, this value is 0.
   * @return {Geometry} this
   * @since 1.1
   */
  public rotate(angle: number, x?: number, y?: number): Geometry {
    if (this._isFrozen) Util.throwFrozen(this);
    if (x === undefined) x = 0;
    if (y === undefined) y = 0;
    if (Debug) {
      Util.checkRealNumber(angle, Geometry, 'rotate:angle');
      Util.checkRealNumber(x, Geometry, 'rotate:x');
      Util.checkRealNumber(y, Geometry, 'rotate:y');
    }
    const t = Transform.alloc();
    t.reset();
    t.rotate(angle, x, y);
    this.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);
    Transform.free(t);
    return this;
  }

  /**
   * @hidden @internal
   * Transforms the Geometry in place by a given transform
   * @export
   * @this {Geometry}
   * @param {number} m11
   * @param {number} m12
   * @param {number} m21
   * @param {number} m22
   * @param {number} dx
   * @param {number} dy
   * @return {Geometry} this
   */
  private transform(m11: number, m12: number, m21: number, m22: number, dx: number, dy: number): Geometry {
    // x,y
    /* Transforming a point:
    x = x * m11 + y * m21 + dx;
    y = x * m12 + y * m22 + dy;
    */
    let x = 0.0;
    let y = 0.0;
    switch (this.type) {
      case Geometry.Line:
      case Geometry.Rectangle:
      case Geometry.Ellipse:
        x = this._startX;
        y = this._startY;
        this._startX = x * m11 + y * m21 + dx;
        this._startY = x * m12 + y * m22 + dy;

        x = this._endX;
        y = this._endY;
        this._endX = x * m11 + y * m21 + dx;
        this._endY = x * m12 + y * m22 + dy;
        break;
      case Geometry.Path:
        const figs = this.figures;
        // for each PathFigure:
        const farr = figs._arr;
        const numfigs = farr.length;
        for (let i = 0; i < numfigs; i++) {
          const fig = farr[i];
          x = fig.startX;
          y = fig.startY;
          fig.startX = x * m11 + y * m21 + dx;
          fig.startY = x * m12 + y * m22 + dy;
          // for each segment of the PathFigure:
          const segments = fig.segments;
          const sarr = segments._arr;
          const numsegs = sarr.length;
          for (let j = 0; j < numsegs; j++) {
            const seg = sarr[j];
            switch (seg.type) {
              case PathSegment.Line:
              case PathSegment.Move:
                x = seg.endX;
                y = seg.endY;
                seg.endX = x * m11 + y * m21 + dx;
                seg.endY = x * m12 + y * m22 + dy;
                break;
              case PathSegment.Bezier:
                x = seg.point1X;
                y = seg.point1Y;
                seg.point1X = x * m11 + y * m21 + dx;
                seg.point1Y = x * m12 + y * m22 + dy;

                x = seg.point2X;
                y = seg.point2Y;
                seg.point2X = x * m11 + y * m21 + dx;
                seg.point2Y = x * m12 + y * m22 + dy;

                x = seg.endX;
                y = seg.endY;
                seg.endX = x * m11 + y * m21 + dx;
                seg.endY = x * m12 + y * m22 + dy;
                break;
              case PathSegment.QuadraticBezier:
                x = seg.point1X;
                y = seg.point1Y;
                seg.point1X = x * m11 + y * m21 + dx;
                seg.point1Y = x * m12 + y * m22 + dy;

                x = seg.endX;
                y = seg.endY;
                seg.endX = x * m11 + y * m21 + dx;
                seg.endY = x * m12 + y * m22 + dy;
                break;
              case PathSegment.Arc:
                x = seg.centerX;
                y = seg.centerY;
                seg.centerX = x * m11 + y * m21 + dx;
                seg.centerY = x * m12 + y * m22 + dy;
                // angles:
                if (m12 !== 0) {  // assume no skew
                  let angle = Math.atan2(m12, m11) * 180 / Math.PI;
                  if (angle < 0) angle += 360;
                  seg.startAngle += angle;
                }
                if (m11 < 0) { seg.startAngle = 180 - seg.startAngle; seg.sweepAngle = -seg.sweepAngle; }
                if (m22 < 0) { seg.startAngle = -seg.startAngle; seg.sweepAngle = -seg.sweepAngle; }
                // scalar:
                seg.radiusX *= Math.sqrt((m11 * m11) + (m21 * m21));
                if (seg.radiusY !== undefined) {
                  seg.radiusY *= Math.sqrt((m12 * m12) + (m22 * m22));
                }
                break;
              case PathSegment.SvgArc:
                x = seg.endX;
                y = seg.endY;
                seg.endX = x * m11 + y * m21 + dx;
                seg.endY = x * m12 + y * m22 + dy;
                // angle:
                if (m12 !== 0) {  // assume no skew
                  let angle = Math.atan2(m12, m11) * 180 / Math.PI;
                  if (angle < 0) angle += 360;
                  seg.xAxisRotation += angle;
                }
                if (m11 < 0) { seg.xAxisRotation = 180 - seg.xAxisRotation; seg.isClockwiseArc = !seg.isClockwiseArc; }
                if (m22 < 0) { seg.xAxisRotation = -seg.xAxisRotation; seg.isClockwiseArc = !seg.isClockwiseArc; }
                // scalar:
                seg.radiusX *= Math.sqrt((m11 * m11) + (m21 * m21));
                seg.radiusY *= Math.sqrt((m12 * m12) + (m22 * m22));
                break;
              default:
                Util.throwError('Unknown Segment type: ' + seg.type);
                break;
            }
          } // end segment
        } // end path figures
        break;
    }
    this._invalidBounds = true;
    return this;
  }

  //
  //  * hidden internal
  //  * @this {Geometry}
  //  * @param {Geometry} geom  The Geometry to be modified and returned.
  //  * @param {Transform} tt  The Transform to be applied to the points of the Rectangle.
  //  * @return {Geometry}
  //
  // private _turnEllipseIntoBezierPath(geom: Geometry, tt: Transform): Geometry {
  //   var x1 = this._startX;
  //   var y1 = this._startY;
  //   var x2 = this._endX;
  //   var y2 = this._endY;

  //   var rx = Math.abs(x2 - x1) / 2;
  //   var ry = Math.abs(y2 - y1) / 2;
  //   var cx = Math.min(x1, x2) + rx;
  //   var cy = Math.min(y1, y2) + ry;

  //   var temp = Point.alloc();

  //   temp.x = cx;
  //   temp.y = cy - ry;
  //   tt.transformPoint(temp);
  //   var f = new PathFigure(temp.x, temp.y);

  //   var p1 = new PathSegment(PathSegment.Bezier);
  //   temp.x = cx + (Geo.KAPPA * rx);
  //   temp.y = cy - ry;
  //   tt.transformPoint(temp);
  //   p1.point1X = temp.x;
  //   p1.point1Y = temp.y;
  //   temp.x = cx + rx;
  //   temp.y = cy - (Geo.KAPPA * ry);
  //   tt.transformPoint(temp);
  //   p1.point2X = temp.x;
  //   p1.point2Y = temp.y;
  //   temp.x = cx + rx;
  //   temp.y = cy;
  //   tt.transformPoint(temp);
  //   p1.endX = temp.x;
  //   p1.endY = temp.y;
  //   f.segments.add(p1);

  //   var p2 = new PathSegment(PathSegment.Bezier);
  //   temp.x = cx + rx;
  //   temp.y = cy + (Geo.KAPPA * ry);
  //   tt.transformPoint(temp);
  //   p2.point1X = temp.x;
  //   p2.point1Y = temp.y;
  //   temp.x = cx + (Geo.KAPPA * rx);
  //   temp.y = cy + ry;
  //   tt.transformPoint(temp);
  //   p2.point2X = temp.x;
  //   p2.point2Y = temp.y;
  //   temp.x = cx;
  //   temp.y = cy + ry;
  //   tt.transformPoint(temp);
  //   p2.endX = temp.x;
  //   p2.endY = temp.y;
  //   f.segments.add(p2);

  //   var p3 = new PathSegment(PathSegment.Bezier);
  //   temp.x = cx - (Geo.KAPPA * rx);
  //   temp.y = cy + ry;
  //   tt.transformPoint(temp);
  //   p3.point1X = temp.x;
  //   p3.point1Y = temp.y;
  //   temp.x = cx - rx;
  //   temp.y = cy + (Geo.KAPPA * ry);
  //   tt.transformPoint(temp);
  //   p3.point2X = temp.x;
  //   p3.point2Y = temp.y;
  //   temp.x = cx - rx;
  //   temp.y = cy;
  //   tt.transformPoint(temp);
  //   p3.endX = temp.x;
  //   p3.endY = temp.y;
  //   f.segments.add(p3);

  //   var p4 = new PathSegment(PathSegment.Bezier);
  //   temp.x = cx - rx;
  //   temp.y = cy - (Geo.KAPPA * ry);
  //   tt.transformPoint(temp);
  //   p4.point1X = temp.x;
  //   p4.point1Y = temp.y;
  //   temp.x = cx - (Geo.KAPPA * rx);
  //   temp.y = cy - ry;
  //   tt.transformPoint(temp);
  //   p4.point2X = temp.x;
  //   p4.point2Y = temp.y;
  //   temp.x = cx;
  //   temp.y = cy - ry;
  //   tt.transformPoint(temp);
  //   p4.endX = temp.x;
  //   p4.endY = temp.y;
  //   f.segments.add(p4);

  //   Point.free(temp);

  //   geom.type = Geometry.Path;
  //   geom.figures.add(f);
  //   return geom;
  // }

  //
  //  * hidden @internal
  //  * this {Geometry}
  //  * param {Geometry} geom  The Geometry to be modified and returned.
  //  * param {Transform} tt  The Transform to be applied to the points of the Rectangle.
  //  * return {Geometry}
  //
  // private _turnRectangleIntoPath(geom: Geometry, tt: Transform): Geometry {
  //   var x1 = this._startX;
  //   var y1 = this._startY;
  //   var x2 = this._endX;
  //   var y2 = this._endY;

  //   var x = Math.min(x1, x2);
  //   var y = Math.min(y1, y2);
  //   var w = Math.abs(x2 - x1);
  //   var h = Math.abs(y2 - y1);

  //   var temp = Point.alloc();

  //   temp.x = x;
  //   temp.y = y;
  //   tt.transformPoint(temp);
  //   var fig = new PathFigure(temp.x, temp.y);

  //   temp.x = x + w;
  //   temp.y = y;
  //   tt.transformPoint(temp);
  //   fig.segments.add(new PathSegment(PathSegment.Line, temp.x, temp.y));

  //   temp.x = x + w;
  //   temp.y = y + h;
  //   tt.transformPoint(temp);
  //   fig.segments.add(new PathSegment(PathSegment.Line, temp.x, temp.y));

  //   temp.x = x;
  //   temp.y = y + h;
  //   tt.transformPoint(temp);
  //   fig.segments.add(new PathSegment(PathSegment.Line, temp.x, temp.y).close());

  //   Point.free(temp);

  //   geom.type = Geometry.Path;
  //   geom.figures.add(fig);
  //   return geom;
  // }

  /**
   * @hidden @internal
   * Returns true if the Geometry contains the point
   * @this {Geometry}
   * @param {Point} p
   * @param {number} sw
   * @param {boolean} swOnFilled specifies whether or not the stroke width is greater than 1.
   * @param {boolean=} precise whether or not to modify the strokewidth when checking unfilled paths.
   * @return {boolean}
   */
  public containsPoint(p: Point, sw: number, swOnFilled: boolean, precise?: boolean): boolean {
    const px = p.x;
    const py = p.y;
    const p0x = this.bounds.x - 20;
    const p0y = py;

    let wind = 0;
    let lx = 0.0;
    let ly = 0.0;

    // to avoid declaring same variables many times
    let crosses = 0;
    let ex = 0.0;
    let ey = 0.0;
    const figures = this.figures._arr;
    const l = figures.length;
    for (let i = 0; i < l; i++) {
      const fig = figures[i];

      // Determine containsPoint if the figure is not filled
      if (!fig.isFilled) {
        // See shape containsPointProtected for a similar + 2
        if (fig.containsPoint(px, py, precise ? sw : sw + 2)) return true;
        continue;
      } else if (swOnFilled && fig.containsPoint(px, py, sw)) {
        return true; // may not be necessary
      }

      // Determine containsPoint if the figure is filled

      const segs = fig.segments;
      lx = fig.startX;
      ly = fig.startY;

      // Keep track of the last move command issued for when a segment is closed
      // as this makes an implicit line to the last move location
      let lmX = lx;
      let lmY = ly;

      const sarr = segs._arr;
      for (let j = 0; j <= segs.length; j++) {
        let seg: PathSegment;
        let type: EnumValue;
        if (j !== segs.length) {
          seg = sarr[j];
          type = seg.type;
          ex = seg.endX;
          ey = seg.endY;
        } else {
          type = PathSegment.Line;
          ex = lmX;
          ey = lmY;
        }

        switch (type) {
          case PathSegment.Move: {
            // move command mid-figure means a new subpath, since we are concerned with filling
            // there is always an implicit close
            const lw = this._lineWind(px, py, p0x, p0y, lx, ly, lmX, lmY);
            if (isNaN(lw)) return true; // right on the line
            wind += lw;
            lmX = ex;
            lmY = ey;

            // this was put in, but incorrectly.
            // Move commands must still remain the same winding number as they represent a subpath.
            // if (wind !== 0) return true;
            // wind = 0;
            break;
          }
          case PathSegment.Line: {
            const lw = this._lineWind(px, py, p0x, p0y, lx, ly, ex, ey);
            if (isNaN(lw)) return true; // right on the line
            wind += lw;
            break;
          }
          case PathSegment.Bezier:
            crosses = Geo.bezierWinding(lx, ly, seg.point1X, seg.point1Y,
              seg.point2X, seg.point2Y, ex, ey, p0x, p0y, px, py, 0.5);
            wind += crosses;
            break;
          case PathSegment.QuadraticBezier: {
            const p1x = (lx + 2 * seg.point1X) / 3;
            const p1y = (ly + 2 * seg.point1Y) / 3;
            const p2x = (seg.point1X * 2 + ex) / 3;
            const p2y = (seg.point1Y * 2 + ey) / 3;
            crosses = Geo.bezierWinding(lx, ly, p1x, p1y, p2x, p2y, ex, ey, p0x, p0y, px, py, 0.5);
            wind += crosses;
            break;
          }
          case PathSegment.Arc:
          case PathSegment.SvgArc: {
            const curves = (seg.type === PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, lx, ly);
            const clen = curves.length;
            if (clen === 0) {
              const lw = this._lineWind(px, py, p0x, p0y, lx, ly, seg.centerX, seg.centerY);
              if (isNaN(lw)) return true; // right on the line
              wind += lw;
              break;
            }
            let c: Array<number> = null;
            for (let k = 0; k < clen; k++) {
              c = curves[k];
              if (k === 0) {
                const lw = this._lineWind(px, py, p0x, p0y, lx, ly, c[0], c[1]);
                if (isNaN(lw)) return true; // right on the line
                wind += lw;
              }
              crosses = Geo.bezierWinding(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], p0x, p0y, px, py, 0.5);
              wind += crosses;
            }
            if (c !== null) {
              ex = c[6];
              ey = c[7];
            }
            break;
          }
          default:
            Util.throwError('Unknown Segment type: ' + seg.type);
            break;
        }

        lx = ex;
        ly = ey;
      } // end segment for loop

      if (wind !== 0) return true;
      wind = 0;
    }
    return wind !== 0;
  }

  /**
   * @hidden @internal
   * Returns NaN if the geometry's unfilled path contains the point.
   * @this {Geometry}
   * @param {number} px
   * @param {number} py
   * @param {number} p0x
   * @param {number} p0y
   * @param {number} lx
   * @param {number} ly
   * @param {number} ex
   * @param {number} ey
   * @return {number}
   */
  private _lineWind(px: number, py: number, p0x: number, p0y: number, lx: number, ly: number, ex: number, ey: number): number {
    if (Geo.lineContainsPoint(lx, ly, ex, ey, .05, px, py)) return NaN;
    const d = (px - p0x) * (ly - ey);
    if (d === 0) return 0;  // horizontal line
    // The x/y intersect of the line
    const xint = ((px * p0y - py * p0x) * (lx - ex) - (px - p0x) * (lx * ey - ly * ex)) / d;
    const yint = ((px * p0y - py * p0x) * (ly - ey)) / d;

    if (xint >= px) return 0; // the line is to the right

    if ((lx > ex ? lx - ex : ex - lx) < (ly > ey ? ly - ey : ey - ly)) {
      if (ly < ey) {
        if (yint < ly || yint > ey) return 0;
      } else {
        if (yint < ey || yint > ly) return 0;
      }
    } else {
      // make sure the x intersect is in the line segment
      if (lx < ex) {
        if (xint < lx || xint > ex) return 0;
      } else {
        if (xint < ex || xint > lx) return 0;
      }
    }

    if (d > 0) return 1;
    return -1;
  }

  /**
   * @hidden @internal
   * Returns true if the geometry's unfilled path contains the point.
   * @this {Geometry}
   * @param {number} px
   * @param {number} py
   * @param {number} strokeWidth
   * @return {boolean}
   */
  public pathContainsPoint(px: number, py: number, strokeWidth: number): boolean {
    const arr = this.figures._arr;
    const numfigs = arr.length;
    for (let i = 0; i < numfigs; i++) {
      const fig = arr[i];
      if (fig.containsPoint(px, py, strokeWidth)) { return true; }
    }
    return false;
  }

  // * Returns the point at the fractional distance (0-1) along the path.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Geometry}
   * @param {number} fraction A fractional amount between 0 and 1, inclusive.
   * @param {Point=} result an optional Point that is modified and returned.
   * @return {Point}
   */
  public getPointAlongPath(fraction: number, result?: Point): Point {
    if (fraction < 0) fraction = 0.0;
    else if (fraction > 1) fraction = 1.0;
    if (result === undefined) result = new Point();
    if (this.type === Geometry.Line) {
      result.assignTo(this.startX + fraction * (this.endX - this.startX),
        this.startY + fraction * (this.endY - this.startY));
      return result;
    }

    // build allPaths array if needed
    const allPaths = this.flattenedSegments;
    const allDists = this.flattenedLengths;
    const totalDist = this.flattenedTotalLength;

    const l = allPaths.length;
    const fractionalDist = totalDist * fraction;
    let currentDist = 0.0;

    for (let i = 0; i < l; i++) {
      const dists = allDists[i];
      const ll = dists.length;
      for (let j = 0; j < ll; j++) {
        const dist = dists[j];

        if (currentDist + dist >= fractionalDist) {
          const offsetDist = fractionalDist - currentDist;
          const offsetBetweenPoints = offsetDist / dist;
          const paths = allPaths[i];
          const lastptx = paths[j * 2];
          const lastpty = paths[j * 2 + 1];
          const ptx = paths[j * 2 + 2];
          const pty = paths[j * 2 + 3];
          // now we have fraction between lastpt and pt
          result.assignTo(lastptx + (ptx - lastptx) * offsetBetweenPoints,
            lastpty + (pty - lastpty) * offsetBetweenPoints);
          return result;
        }
        currentDist += dist;
      }
    }

    // only if there was no path on the Geometry, shouldn't be hit
    result.assignTo(NaN, NaN);
    return result;
  }

  /**
   * @hidden @internal
   * Returns the angle at the fractional distance (0-1) along the path.
   * @this {Geometry}
   * @param {number} fraction A fractional amount between 0 and 1, inclusive.
   * @return {number}
   */
  public getAngleAlongPath(fraction: number): number {
    if (fraction < 0) fraction = 0.0;
    else if (fraction > 1) fraction = 1.0;
    if (this.type === Geometry.Line) {
      return Math.atan2(this.endY - this.startY, this.endX - this.startX) * 180 / Math.PI;
    }

    // build allPaths array if needed
    const allPaths = this.flattenedSegments;
    const allDists = this.flattenedLengths;
    const totalDist = this.flattenedTotalLength;

    const l = allPaths.length;
    const fractionalDist = totalDist * fraction;
    let currentDist = 0.0;

    for (let i = 0; i < l; i++) {
      const dists = allDists[i];
      const ll = dists.length;
      for (let j = 0; j < ll; j++) {
        const dist = dists[j];

        if (currentDist + dist >= fractionalDist) {
          const paths = allPaths[i];
          const lastptx = paths[j * 2];
          const lastpty = paths[j * 2 + 1];
          const ptx = paths[j * 2 + 2];
          const pty = paths[j * 2 + 3];
          // Handle orthogonal cases first
          if (Math.abs(ptx - lastptx) < 1 && Math.abs(pty - lastpty) < 1) {
            return 0;
          } else if (Math.abs(ptx - lastptx) < 1) {
            return pty - lastpty >= 0 ? 90 : 270;
          } else if (Math.abs(pty - lastpty) < 1) {
            return ptx - lastptx >= 0 ? 0 : 180;
          } else { // not a simple orthogonal
            return Math.atan2(pty - lastpty, ptx - lastptx) * 180 / Math.PI;
          }
        }
        currentDist += dist;
      }
    }

    // only if there was no path on the Geometry, shouldn't be hit
    return NaN;
  }

  //    * Returns the fractional distance (0-1) along the path for a nearby point
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Geometry}
   * @param {Point} pt A Point near the this Geometry.
   * @return {number}
   */
  public getFractionForPoint(pt: Point): number {
    if (this.type === Geometry.Line) {
      const sx = this.startX;
      const sy = this.startY;
      const ex = this.endX;
      const ey = this.endY;
      if (sx === ex && sy === ey) {
        return 0; // just a point, return 0
      } else {
        const px = pt.x;
        const py = pt.y;
        let min = 0;
        let max = 0;
        if (sx === ex) { // vertical
          if (sy < ey) {
            min = sy;
            max = ey;
          } else {
            min = ey;
            max = sy;
          }
          if (py <= min) {
            return min === sy ? 0 : 1;
          } else if (py >= max) {
            return max === sy ? 0 : 1;
          } else {
            return Math.abs(py - sy) / (max - min);
          }
        } else if (sy === ey) { // horizontal
          if (sx < ex) {
            min = sx;
            max = ex;
          } else {
            min = ex;
            max = sx;
          }
          if (px <= min) {
            return min === sx ? 0 : 1;
          } else if (px >= max) {
            return max === sx ? 0 : 1;
          } else {
            return Math.abs(px - sx) / (max - min);
          }
        } else {
          const l = (ex - sx) * (ex - sx) + (ey - sy) * (ey - sy); // length of segment squared
          return ((px - sx) * (px - sx) + (py - sy) * (py - sy)) / l; // don't need to sqrt here since we didn't above
        }
      }
    } else if (this.type === Geometry.Rectangle) {
      const sx = this.startX;
      const sy = this.startY;
      const ex = this.endX;
      const ey = this.endY;
      if (sx === ex && sy === ey) {
        return 0; // just a point, return 0
      } else {
        const width = ex - sx;
        const height = ey - sy;
        const perim = width * 2 + height * 2;
        let px = pt.x;
        let py = pt.y;
        // clamp x and y to bounds of rectangle
        px = Math.min(Math.max(px, sx), ex);
        py = Math.min(Math.max(py, sy), ey);
        const dl = Math.abs(px - sx);
        const dr = Math.abs(px - ex);
        const dt = Math.abs(py - sy);
        const db = Math.abs(py - ey);
        const minDist = Math.min(dl, dr, dt, db);
        if (minDist === dt) return px / perim;
        if (minDist === dr) return (width + py) / perim;
        if (minDist === db) return (width * 2 + height - px) / perim;
        if (minDist === dl) return (width * 2 + height * 2 - py) / perim;
      }
    } else {
      // build allPaths array
      const allPaths = this.flattenedSegments;
      const allDists = this.flattenedLengths;
      const totalDist = this.flattenedTotalLength;

      const temp = Point.alloc();
      let bestDist = Infinity;
      let bestCurrent = 0.0;
      let currentDist = 0.0;
      const l = allPaths.length;
      let lastptx = 0.0;
      let lastpty = 0.0;
      for (let i = 0; i < l; i++) {
        const path = allPaths[i];
        const dists = allDists[i];
        const ll = path.length;
        for (let j = 0; j < ll; j += 2) {
          const ptx = path[j];
          const pty = path[j + 1];
          if (j === 0) { // its a new path because a moveTo occurred, so set the lastPt and move on
            lastptx = ptx; lastpty = pty;
            continue;
          }
          // find distance from midpoint to given point
          Geo.nearestPointOnLine(lastptx, lastpty, ptx, pty, pt.x, pt.y, temp);
          const dist = (temp.x - pt.x) * (temp.x - pt.x) + (temp.y - pt.y) * (temp.y - pt.y);
          if (dist < bestDist) {
            bestDist = dist;
            bestCurrent = currentDist;
            bestCurrent += Math.sqrt((temp.x - lastptx) * (temp.x - lastptx) + (temp.y - lastpty) * (temp.y - lastpty));
          }
          currentDist += dists[(j - 2) / 2];
          lastptx = ptx; lastpty = pty;
        }
      }
      Point.free(temp);

      const frac = bestCurrent / totalDist;
      if (frac < 0) return 0.0;
      else if (frac > 1) return 1.0;
      return frac;
    }
    return 0;
  }

  /**
   * @hidden @internal
   * Flattened segments represents a series of points making up a Geometry.
   * The inner arrays consist of a sequence of PathSegment coordinates making up a PathFigure,
   * while the outer array represents the whole Geometry of these PathFigures.
   */
  /** @export */ get flattenedSegments(): Array<Array<number>> {
    this._maybeBuildAllPaths();
    return this._allPathsSegments;
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   */
  private _maybeBuildAllPaths() {
    if (this._allPathsSegments === null) {
      const allPaths: Array<Array<number>> = this._allPathsSegments = [];
      const allDists: Array<Array<number>> = this._allPathsLengths = [];
      this._buildAllPaths(allPaths, allDists);
    }
  }

  /**
   * @hidden @internal
   * Flattened lengths represents a series of lengths of the paths making up a Geometry.
   * The inner arrays consist of a sequence of PathSegment lengths making up a PathFigure,
   * while the outer array represents the whole Geometry of these PathFigures.
   */
  /** @export */ get flattenedLengths(): Array<Array<number>> {
    this._maybeBuildAllPaths();
    return this._allPathsLengths;
  }

  /**
   * @hidden @internal
   * This read-only property returns the approximate length of the path.
   * This is the overall distance of all the PathSegments that make up the geometry.
   */
  /** @export */ get flattenedTotalLength(): number {
    let len = this._allPathsTotalLength;
    if (isNaN(len)) {
      if (this.type === Geometry.Line) {
        const w = Math.abs(this.endX - this.startX);
        const h = Math.abs(this.endY - this.startY);
        len = Math.sqrt(w * w + h * h);
      } else if (this.type === Geometry.Rectangle) {
        const w = Math.abs(this.endX - this.startX);
        const h = Math.abs(this.endY - this.startY);
        len = w * 2 + h * 2;
      } else {
        const allDists = this.flattenedLengths;
        const numpaths = allDists.length;
        len = 0.0;
        for (let i = 0; i < numpaths; i++) {
          const dists = allDists[i];
          const ll = dists.length;
          for (let j = 0; j < ll; j++) {
            len += dists[j];
          }
        }
      }
      this._allPathsTotalLength = len;
    }
    return len;
  }

  /**
   * @hidden @internal
   * @this {Geometry}
   * @param {Array.<number>} path
   * @return {Array.<number>}
   */
  private distancesForPath(path: Array<number>): Array<number> {
    const currDist = [];
    let lastptx = 0.0;
    let lastpty = 0.0;
    const ll = path.length;
    for (let j = 0; j < ll; j += 2) {
      const ptx = path[j];
      const pty = path[j + 1];
      if (j === 0) { // its a new path because a moveTo occurred, so set the lastPt and move on
        lastptx = ptx;
        lastpty = pty;
        continue;
      }
      // distance between pt and lastpt
      const dist = Math.sqrt(Point.distanceSquared(lastptx, lastpty, ptx, pty));
      currDist.push(dist);
      lastptx = ptx;
      lastpty = pty;
    }
    return currDist;
  }

  /**
   * Builds an array of segments that make up the geometry.
   * @hidden @internal
   * @this {Geometry}
   * @param {Array.<Array.<number>>} allPaths an array to modify, holding Arrays of numbers
   * @param {Array.<Array.<number>>} allDists an array to modify, holding Arrays of numbers
   */
  private _buildAllPaths(allPaths: Array<Array<number>>, allDists: Array<Array<number>>): void {
    let currPath: Array<number> = [];
    const currDist: Array<number> = [];
    if (this.type === Geometry.Line) {
      currPath.push(this.startX); currPath.push(this.startY);
      currPath.push(this.endX); currPath.push(this.endY);
      allPaths.push(currPath);
      currDist.push(Math.sqrt((this.startX - this.endX) * (this.startX - this.endX) +
        (this.startY - this.endY) * (this.startY - this.endY)));
      allDists.push(currDist);
    } else if (this.type === Geometry.Rectangle) {
      currPath.push(this.startX); currPath.push(this.startY);
      currPath.push(this.endX); currPath.push(this.startY);
      currPath.push(this.endX); currPath.push(this.endY);
      currPath.push(this.startX); currPath.push(this.endY);
      currPath.push(this.startX); currPath.push(this.startY);
      allPaths.push(currPath);
      currDist.push(Math.abs(this.startX - this.endX));
      currDist.push(Math.abs(this.startY - this.endY));
      currDist.push(Math.abs(this.startX - this.endX));
      currDist.push(Math.abs(this.startY - this.endY));
      allDists.push(currDist);
    } else if (this.type === Geometry.Ellipse) {
      // need to get a set of Bezier curves from which we can call Geo.bezierLines
      const ellfig = new PathFigure();
      ellfig.startX = this.endX;
      ellfig.startY = (this.startY + this.endY) / 2;

      const ellseg = new PathSegment(PathSegment.Arc);
      ellseg.startAngle = 0;
      ellseg.sweepAngle = 360;
      ellseg.centerX = (this.startX + this.endX) / 2;
      ellseg.centerY = (this.startY + this.endY) / 2;
      ellseg.radiusX = Math.abs(this.startX - this.endX) / 2;
      ellseg.radiusY = Math.abs(this.startY - this.endY) / 2;
      ellfig.add(ellseg);

      const curves = ellseg.buildBeziers(ellfig);
      const cl = curves.length;
      if (cl === 0) {
        currPath.push(ellseg.centerX); currPath.push(ellseg.centerY);
      } else {
        let lastpointX = ellfig.startX;
        let lastpointY = ellfig.startY;
        for (let cli = 0; cli < cl; cli++) {
          const cu = curves[cli];
          Geo.bezierLines(lastpointX, lastpointY,
            cu[2], cu[3],
            cu[4], cu[5],
            cu[6], cu[7],
            0.5, currPath);
          lastpointX = cu[6];
          lastpointY = cu[7];
        }
      }
      allPaths.push(currPath);
      allDists.push(this.distancesForPath(currPath));
    } else {
      const fit = this.figures.iterator;
      while (fit.next()) {
        const fig = fit.value;
        currPath = [];
        currPath.push(fig.startX); currPath.push(fig.startY);

        // lastPoint only needed for drawing dash approximations
        let lastpointX = fig.startX;
        let lastpointY = fig.startY;
        let lastSubX = lastpointX;
        let lastSubY = lastpointY;

        // for each segment of the PathFigure:
        const segments = fig.segments._arr;
        const sl = segments.length;
        for (let j = 0; j < sl; j++) {
          const seg = segments[j];
          switch (seg.type) {
            case PathSegment.Move:
              if (currPath.length >= 4) {
                allPaths.push(currPath); // multiple moves in a row shouldn't be included
                allDists.push(this.distancesForPath(currPath));
              }
              currPath = [];  // restart
              currPath.push(seg.endX); currPath.push(seg.endY);

              lastpointX = seg.endX;
              lastpointY = seg.endY;
              lastSubX = lastpointX;
              lastSubY = lastpointY;
              break;
            case PathSegment.Line:
              currPath.push(seg.endX); currPath.push(seg.endY);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.Bezier:
              Geo.bezierLines(lastpointX, lastpointY,
                seg.point1X, seg.point1Y,
                seg.point2X, seg.point2Y,
                seg.endX, seg.endY,
                0.5, currPath);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.QuadraticBezier:
              Geo.quadraticBezierLines(lastpointX, lastpointY,
                seg.point1X, seg.point1Y,
                seg.endX, seg.endY,
                0.5, currPath);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.Arc: {
              const curves = seg.buildBeziers(fig);
              const cl = curves.length;
              if (cl === 0) {
                currPath.push(seg.centerX); currPath.push(seg.centerY);
                lastpointX = seg.centerX;
                lastpointY = seg.centerY;
                break;
              }
              for (let cli = 0; cli < cl; cli++) {
                const cu = curves[cli];
                Geo.bezierLines(lastpointX, lastpointY,
                  cu[2], cu[3],
                  cu[4], cu[5],
                  cu[6], cu[7],
                  0.5, currPath);
                lastpointX = cu[6];
                lastpointY = cu[7];
              }
              break;
            }
            case PathSegment.SvgArc: {
              const curves = seg.buildSVGBeziers(fig, lastpointX, lastpointY);
              const cl = curves.length;
              if (cl === 0) {
                currPath.push(seg.centerX); currPath.push(seg.centerY);
                lastpointX = seg.centerX;
                lastpointY = seg.centerY;
                break;
              }
              for (let cli = 0; cli < cl; cli++) {
                const cu = curves[cli];
                Geo.bezierLines(lastpointX, lastpointY,
                  cu[2], cu[3],
                  cu[4], cu[5],
                  cu[6], cu[7],
                  0.5, currPath);
                lastpointX = cu[6];
                lastpointY = cu[7];
              }
              break;
            }
            default:
              Util.throwError('Segment not of valid type: ' + seg.type);
          }
          if (seg.isClosed) {
            currPath.push(lastSubX); currPath.push(lastSubY);
          }
        } // end segment iteration
        if (currPath.length >= 4) {
          allPaths.push(currPath);
          allDists.push(this.distancesForPath(currPath));
        }
      }
    }
  }

  /**
   * Gets or sets the type of the Geometry.
   * The default type is {@link Geometry.Path}.
   * Other permissible values are {@link Geometry.Line}, {@link Geometry.Ellipse},
   * or {@link Geometry.Rectangle}.
   */
  /** @export */ get type(): EnumValue { return this._type; }
  /** @export */ set type(value) {
    const old = this._type;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Geometry, Geometry, 'type');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._type = value;
      this._invalidBounds = true;
    }
  }

  // simple Geometry specific:
  /**
   * Gets or sets the starting X coordinate of the Geometry if it is of type
   * {@link Geometry.Line|Line}, {@link Geometry.Rectangle|Rectangle}, or {@link Geometry.Ellipse|Ellipse}.
   * The initial value is zero.
   */
  /** @export */ get startX(): number { return this._startX; }
  /** @export */ set startX(value) {
    const old = this._startX;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Geometry, 'startX');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._startX = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Gets or sets the starting Y coordinate of the Geometry if it is of type
   * {@link Geometry.Line|Line}, {@link Geometry.Rectangle|Rectangle}, or {@link Geometry.Ellipse|Ellipse}.
   * The initial value is zero.
   */
  /** @export */ get startY(): number { return this._startY; }
  /** @export */ set startY(value) {
    const old = this._startY;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Geometry, 'startY');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._startY = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Gets or sets the ending X coordinate of the Geometry if it is of type
   * {@link Geometry.Line|Line}, {@link Geometry.Rectangle|Rectangle}, or {@link Geometry.Ellipse|Ellipse}.
   * The initial value is zero.
   */
  /** @export */ get endX(): number { return this._endX; }
  /** @export */ set endX(value) {
    const old = this._endX;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Geometry, 'endX');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._endX = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Gets or sets the ending Y coordinate of the Geometry if it is of type
   * {@link Geometry.Line|Line}, {@link Geometry.Rectangle|Rectangle}, or {@link Geometry.Ellipse|Ellipse}.
   * The initial value is zero.
   */
  /** @export */ get endY(): number { return this._endY; }
  /** @export */ set endY(value) {
    const old = this._endY;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Geometry, 'endY');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._endY = value;
      this._invalidBounds = true;
    }
  }

  // All Geometry:
  // A collection of figures that describes the content of the path.
  /**
   * Gets or sets the {@link List} of {@link PathFigure}s
   * that describes the content of the
   * path for Geometries of type {@link Geometry.Path|Path}.
   */
  /** @export */ get figures(): List<PathFigure> { return this._figures; }
  /** @export */ set figures(value) {
    const old = this._figures;
    if (old !== value) {
      if (Debug) Util.checkClass(value, List, Geometry, 'figures');
      if (this._isFrozen) Util.throwFrozen(this, value);
      this._figures = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Add a {@link PathFigure} to the {@link #figures} list.
   * @export
   * @this {Geometry}
   * @param {PathFigure} figure a newly allocated unshared PathFigure that will become owned by this Geometry
   * @return {Geometry} this
   * @since 1.5
   */
  public add(figure: PathFigure): Geometry {
    this._figures.add(figure);
    return this;
  }

  // Set {@link #spot1} and {@link #spot2} without allocating new {@link Spot}s.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Geometry}
   * @param {number} f1x spot1.x
   * @param {number} f1y spot1.y
   * @param {number} f2x spot2.x
   * @param {number} f2y spot2.y
   * @param {number=} o1x offset1.x, default is zero
   * @param {number=} o1y offset1.y, default is zero
   * @param {number=} o2x offset2.x, default is zero
   * @param {number=} o2y offset2.y, default is zero
   * @return {Geometry} this
   */
  public setSpots(f1x: number, f1y: number, f2x: number, f2y: number, o1x?: number, o1y?: number, o2x?: number, o2y?: number): Geometry {
    if (this._isFrozen) Util.throwFrozen(this);
    this._spot1 = new Spot(f1x, f1y, o1x, o1y).freeze();
    this._spot2 = new Spot(f2x, f2y, o2x, o2y).freeze();
    return this;
  }

  /**
   * Gets or sets the spot an "Auto" {@link Panel} will use for the top-left corner of any panel content
   * when the {@link Shape#spot1} value is {@link Spot.Default}.
   * The default value is {@link Spot.TopLeft}, at the top-left point of the bounds of the Shape.
   */
  /** @export */ get spot1(): Spot { return this._spot1; }
  /** @export */ set spot1(value) {
    if (Debug) Util.checkClass(value, Spot, Geometry, 'spot1');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._spot1 = value.copyFrozen();
  }

  /**
   * Gets or sets the spot an "Auto" {@link Panel} will use for the bottom-right corner of any panel content
   * when the {@link Shape#spot2} value is {@link Spot.Default}.
   * The default value is {@link Spot.BottomRight}, at the bottom-right point of the bounds of the Shape.
   */
  /** @export */ get spot2(): Spot { return this._spot2; }
  /** @export */ set spot2(value) {
    if (Debug) Util.checkClass(value, Spot, Geometry, 'spot2');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._spot2 = value.copyFrozen();
  }

  /**
   * Gets or sets the {@link Shape#geometryStretch} value the {@link Shape} should use by default
   * when the {@link Shape#geometryStretch} value is {@link GraphObject.None}.
   * The default value is {@link GraphObject.Fill}.
   * Some figure generators return a Geometry with this property set to {@link GraphObject.Uniform},
   * in order to preserve its aspect ratio when used by a Shape that may have different sizes.
   * @since 1.5
   */
  /** @export */ get defaultStretch(): EnumValue { return this._defaultStretch; }
  /** @export */ set defaultStretch(value) {
    if (Debug) EnumValue.checkEnumValue(value, GraphObject, Geometry, 'stretch');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._defaultStretch = value;
  }

  /**
   * This read-only property returns a rectangle that contains all points within the Geometry.
   * The result will always contain the origin (0, 0).
   */
  /** @export */ get bounds(): Rect {
    if (this.isInvalidBounds()) {
      this.computeBounds();
    }
    return this._computedBounds;
  }

} // #end Geometry
Geometry['className'] = 'Geometry';
Geometry['stringify'] = Geometry.stringify;
Geometry['fillPath'] = Geometry.fillPath;
Geometry['parse'] = Geometry.parse;

Geometry['Line'] = Geometry.Line;
Geometry['Rectangle'] = Geometry.Rectangle;
Geometry['Ellipse'] = Geometry.Ellipse;
Geometry['Path'] = Geometry.Path;

/**
 * A PathFigure represents a section of a {@link Geometry}.
 * It is a single connected series of
 * two-dimensional geometric {@link PathSegment}s.
 *
 * The {@link Geometry#figures} property is a list of PathFigures.
 *
 * PathFigures are drawn with the {@link Shape#stroke} and other stroke properties.
 * Filled PathFigures are painted with the {@link Shape#fill}.
 *
 * A PathFigure must not be modified once its containing {@link Geometry}
 * has been assigned to a {@link Shape}.
 * @unrestricted
 * @category Geometry
 */
export class PathFigure {
  /** @hidden @internal */ public static className = 'PathFigure';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ private _startX: number;
  /** @hidden @internal */ private _startY: number;
  /** @hidden @internal */ private _segments: List<PathSegment>;
  /** @hidden @internal */ private _segmentsChanged: number;
  /** @hidden @internal */ private _isFilled: boolean;
  /** @hidden @internal */ private _isShadowed: boolean;
  /** @hidden @internal */ public _invalidBounds: boolean;

  /**
   * Constructs an empty figure.
   * The optional arguments specify the starting point of the figure.
   * You'll want to add a new instance of a PathFigure to the
   * {@link Geometry#figures} list of a {@link Geometry}.
   * @param {number=} sx optional: the X coordinate of the start point (default is zero).
   * @param {number=} sy optional: the Y coordinate of the start point (default is zero).
   * @param {boolean=} filled optional: whether the figure is filled (default is true).
   * @param {boolean=} shadowed optional: whether the figure may be drawn with a shadow (default is true).
   */
  constructor(sx?: number, sy?: number, filled?: boolean, shadowed?: boolean) {
    Set.uniqueHash(this);
    this._isFrozen = false;
    if (filled === undefined) filled = true;
    this._isFilled = filled;
    if (shadowed === undefined) shadowed = true;
    this._isShadowed = shadowed;
    if (sx !== undefined) {
      if (Debug) Util.checkRealNumber(sx, PathFigure, 'sx');
      this._startX = sx;
    } else {
      this._startX = 0;
    }
    if (sy !== undefined) {
      if (Debug) Util.checkRealNumber(sy, PathFigure, 'sy');
      this._startY = sy;
    } else {
      this._startY = 0;
    }
    this._segments = new List<PathSegment>();

    this._segmentsChanged = this._segments._changed;
    this._invalidBounds = true;
  }

  /**
   * Create a copy of this PathFigure, with the same values and segments.
   * @expose
   * @this {PathFigure}
   * @return {PathFigure}
   */
  public copy(): PathFigure {
    const copy = new PathFigure();
    copy._isFilled = this._isFilled;
    copy._isShadowed = this._isShadowed;
    copy._startX = this._startX;
    copy._startY = this._startY;
    const oldsegs = this._segments._arr;
    const numsegs = oldsegs.length;
    const newsegs = copy._segments;
    for (let i = 0; i < numsegs; i++) {
      const oldseg = oldsegs[i];
      const newseg = oldseg.copy();
      newsegs.add(newseg);
    }
    copy._segmentsChanged = this._segmentsChanged;
    copy._invalidBounds = this._invalidBounds;
    return copy;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {PathFigure}
   * @param {PathFigure} f
   * @return {boolean}
   */
  public equalsApprox(f: PathFigure): boolean {
    if (!(f instanceof PathFigure)) return false;
    // ??? if (this.isFilled !== f.isFilled) return false;
    // do not check isShadowed
    if (!Geo.isApprox(this.startX, f.startX)) return false;
    if (!Geo.isApprox(this.startY, f.startY)) return false;
    const segs1 = this.segments._arr;
    const segs2 = f.segments._arr;
    const numsegs = segs1.length;
    if (numsegs !== segs2.length) return false;
    for (let i = 0; i < numsegs; i++) {
      const seg1 = segs1[i];
      const seg2 = segs2[i];
      if (!seg1.equalsApprox(seg2)) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * @this {PathFigure}
   * @param {number=} precision Optional: the amount of decimal places to round numbers to for debugging purposes,
   * leaves number unchanged by default.
   * @return {string}
   */
  public toString(precision?: number): string {
    if (precision === undefined) precision = -1;
    let str = '';
    if (precision < 0) {
      str = 'M' + this.startX.toString() + ' ' + this.startY.toString();
    } else {
      str = 'M' + this.startX.toFixed(precision) + ' ' + this.startY.toFixed(precision);
    }
    const segs = this.segments._arr;
    const numsegs = segs.length;
    for (let i = 0; i < numsegs; i++) {
      const seg = segs[i];
      str += ' ' + seg.toString(precision);
    }
    return str;
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {PathFigure}
   * @return {PathFigure}
   */
  public freeze(): PathFigure {
    this._isFrozen = true;
    const segs = this.segments;
    segs.freeze();
    const arr = segs._arr;
    const numsegs = segs.length;
    for (let i = 0; i < numsegs; i++) {
      const seg = arr[i];
      seg.freeze();
    }
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {PathFigure}
   * @return {PathFigure}
   */
  public thaw(): PathFigure {
    this._isFrozen = false;
    const segs = this.segments;
    segs.thaw();
    const arr = segs._arr;
    const numsegs = arr.length;
    for (let i = 0; i < numsegs; i++) {
      const seg = arr[i];
      seg.thaw();
    }
    return this;
  }

  /**
   * @hidden @internal
   * @this {PathFigure}
   * @return {boolean}
   */
  public isInvalidBounds(): boolean {
    if (this._invalidBounds) return true;
    const segs = this.segments;
    if (this._segmentsChanged !== segs._changed) return true;
    const arr = segs._arr;
    const sl = arr.length;
    for (let j = 0; j < sl; j++) {
      const seg = arr[j];
      if (seg._invalidBounds) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {PathFigure}
   */
  public validateBounds(): void {
    this._invalidBounds = false;
    const segs = this.segments;
    this._segmentsChanged = segs._changed;
    const arr = segs._arr;
    const sl = arr.length;
    for (let j = 0; j < sl; j++) {
      const seg = arr[j];
      seg._invalidBounds = false;
      seg._beziers = null;
    }
  }

  /**
   * Gets or sets whether this PathFigure is drawn filled.
   * The default value is true.
   */
  /** @export */ get isFilled(): boolean { return this._isFilled; }
  /** @export */ set isFilled(value) {
    if (Debug) Util.checkPrimitive(value, 'boolean', PathFigure, 'isFilled');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._isFilled = value;
  }

  /**
   * Gets or sets whether this PathFigure will render a shadow if one is defined.
   * This is used for turning off shadows on specific subpaths.
   * The default value is true.
   */
  /** @export */ get isShadowed(): boolean { return this._isShadowed; }
  /** @export */ set isShadowed(value) {
    if (Debug) Util.checkPrimitive(value, 'boolean', PathFigure, 'isShadowed');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._isShadowed = value;
  }

  /**
   * Gets or sets the starting point X coordinate of the PathFigure.
   * The default value is zero.
   */
  /** @export */ get startX(): number { return this._startX; }
  /** @export */ set startX(value) {
    if (Debug) Util.checkRealNumber(value, PathFigure, 'startX');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._startX = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the starting point Y coordinate of the PathFigure.
   * The default value is zero.
   */
  /** @export */ get startY(): number { return this._startY; }
  /** @export */ set startY(value) {
    if (Debug) Util.checkRealNumber(value, PathFigure, 'startY');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._startY = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the {@link List} of {@link PathSegment}s that define this PathFigure.
   */
  /** @export */ get segments(): List<PathSegment> { return this._segments; }
  /** @export */ set segments(value) {
    if (Debug) Util.checkClass(value, List, PathFigure, 'segments');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._segments = value;
    this._invalidBounds = true;
  }

  /**
   * Add a {@link PathSegment} to the {@link #segments} list.
   * @export
   * @this {PathFigure}
   * @param {PathSegment} segment a newly allocated unshared PathSegment that will become owned by this PathFigure
   * @return {PathFigure}
   * @since 1.5
   */
  public add(segment: PathSegment): PathFigure {
    this._segments.add(segment);
    return this;
  }

  /**
   * @hidden @internal
   * @this {PathFigure}
   * @param {number} px
   * @param {number} py
   * @param {number} strokeWidth
   * @return {boolean}
   */
  public containsPoint(px: number, py: number, strokeWidth: number): boolean {
    const fig = this;
    let lastX = fig.startX;
    let lastY = fig.startY;
    // The last move command must be recorded to deal with closed subpaths
    let lmX = lastX;
    let lmY = lastY;
    const segments = fig.segments._arr;
    const numsegs = segments.length;
    for (let j = 0; j < numsegs; j++) {
      const seg = segments[j];
      switch (seg.type) {
        case PathSegment.Move:
          lmX = seg.endX;
          lmY = seg.endY;
          lastX = seg.endX;
          lastY = seg.endY;
          break;
        case PathSegment.Line:
          if (Geo.lineContainsPoint(lastX, lastY,
            seg.endX, seg.endY,
            strokeWidth, px, py)) {
            return true;
          }
          lastX = seg.endX;
          lastY = seg.endY;
          break;
        case PathSegment.Bezier:
          if (Geo.bezierContainsPoint(lastX, lastY,
            seg.point1X, seg.point1Y,
            seg.point2X, seg.point2Y,
            seg.endX, seg.endY,
            0.5, px, py, strokeWidth)) {
            return true;
          }
          lastX = seg.endX;
          lastY = seg.endY;
          break;
        case PathSegment.QuadraticBezier:
          if (Geo.quadraticBezierContainsPoint(lastX, lastY,
            seg.point1X, seg.point1Y,
            seg.endX, seg.endY,
            0.5, px, py, strokeWidth)) {
            return true;
          }
          lastX = seg.endX;
          lastY = seg.endY;
          break;
        case PathSegment.Arc:
        case PathSegment.SvgArc:
          const curves = (seg.type === PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, lastX, lastY);
          const clen = curves.length;
          if (clen === 0) {
            if (Geo.lineContainsPoint(lastX, lastY,
              seg.centerX, seg.centerY,
              strokeWidth, px, py)) {
              return true;
            }
            lastX = seg.centerX;
            lastY = seg.centerY;
            break;
          }
          let c: Array<number> = null;
          for (let i = 0; i < clen; i++) {
            c = curves[i];
            if (i === 0 && Geo.lineContainsPoint(lastX, lastY, c[0], c[1], strokeWidth, px, py)) {
              return true;
            }
            if (Geo.bezierContainsPoint(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], 0.5, px, py, strokeWidth)) {
              return true;
            }
          }
          if (c !== null) {
            lastX = c[6];
            lastY = c[7];
          }
          break;
        default:
          Util.throwError('Unknown Segment type: ' + seg.type);
      } // end switch
      if (seg.isClosed && (lastX !== lmX || lastY !== lmY) &&
        Geo.lineContainsPoint(lastX, lastY, lmX, lmY, strokeWidth, px, py)) {
        return true;
      }

    }
    return false;
  }

} // #end PathFigure
PathFigure['className'] = 'PathFigure';

/**
 * A PathSegment represents a straight line or curved segment of a path between
 * two or more points that are part of a {@link PathFigure}.
 * The very first point is given by the {@link PathFigure#startX} and {@link PathFigure#startY} properties.
 * After the first segment the beginning point of the next segment is the same as the end point of the previous segment.
 * For most types of PathSegments the end point of the segment is given by {@link #endX} and {@link #endY}.
 *
 * A PathSegment must not be modified once its containing {@link PathFigure}'s
 * {@link Geometry} has been assigned to a {@link Shape}.
 * @unrestricted
 * @category Geometry
 */
export class PathSegment {
  /** @hidden @internal */ public static className = 'PathSegment';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ private _largeArcFlag: boolean;
  /** @hidden @internal */ private _clockwiseFlag: boolean;
  /** @hidden @internal */ private _isClosed: boolean;
  /** @hidden @internal */ private _type: EnumValue;
  /** @hidden @internal */ private _endX: number;
  /** @hidden @internal */ private _endY: number;
  /** @hidden @internal */ private _point1X: number;
  /** @hidden @internal */ private _point1Y: number;
  /** @hidden @internal */ private _point2X: number;
  /** @hidden @internal */ private _point2Y: number;
  /** @hidden @internal */ public _invalidBounds: boolean;
  /** @hidden @internal */ public _beziers: Array<Array<number>> | null;

  /**
   * Constructs a segment that goes nowhere unless you specify some Points.
   *
   * The very first point of a path is specified in the containing {@link PathFigure} by
   * its {@link PathFigure#startX} and {@link PathFigure#startY} properties.
   *
   * The segment type must be one of the following values:
   * {@link PathSegment.Line}, {@link PathSegment.Bezier},
   * {@link PathSegment.QuadraticBezier}, {@link PathSegment.Arc}, {@link PathSegment.SvgArc}.
   * You will want to add a new instance of a PathSegment to the
   * {@link PathFigure#segments} list of a {@link PathFigure}.
   * @param {EnumValue=} type if not supplied, the default PathSegment type is {@link PathSegment.Line}.
   *   But if the type is supplied, one must also provide the endpoint X and Y values,
   *   either as arguments in this constructor or by setting the {@link #endX} and {@link #endY} properties.
   *   If the type is QuadraticBezier, the X1 and Y1 control point values must both be supplied.
   *   If the type is Bezier, X1, Y1, X2, and Y2 control point values must all be supplied.
   *
   * If the type is {@link PathSegment.Line|Line} it needs the following arguments:
   *
   * `(go.PathSegment.Line, ex, ey)`
   *   - **ex, ey** describe the end point
   *
   * If the type is {@link PathSegment.QuadraticBezier|QuadraticBezier} it needs the following arguments:
   *
   * `(go.PathSegment.QuadraticBezier, ex, ey, x1, y1)`
   *   - **ex, ey** describe the end point
   *   - **x1, y1** describe the only control point
   *
   * If the type is {@link PathSegment.Bezier|Bezier} it needs the following arguments:
   *
   * `(go.PathSegment.Bezier, ex, ey, x1, y1, x2, y2)`
   *   - **ex, ey** describe the end point
   *   - **x1, y1** describe the first control point
   *   - **x2, y2** describe the second control point
   *
   * If the type is {@link PathSegment.Arc|Arc} it needs the following arguments:
   *
   * `(go.PathSegment.Arc, startAngle, sweepAngle, centerX, centerY, radiusX, radiusY)`
   *   - **startAngle** describes the start angle, in degrees
   *   - **sweepAngle** describes the sweep angle, in degrees
   *   - **centerX, centerY** describe the center point
   *   - **radiusX, radiusY** describe the radiusX and radiusY
   *
   * If the type is {@link PathSegment.SvgArc|SvgArc} it needs the following arguments:
   *
   * `(go.PathSegment.SvgArc, ex, ey, radiusX, radiusY, xAxisRotation, largeArcFlag, clockwiseFlag)`
   *
   * They are in the same order as arcs in SVG path strings, except the endpoint x and y values come first, not last.
   *   - **ex, ey** describe the endpoint
   *   - **radiusX, radiusY** describe the radius
   *   - **xAxisRotation** describes the {@link #xAxisRotation} (number in degrees)
   *   - **largeArcFlag** describes the {@link #isLargeArc} (true or false)
   *   - **clockwiseFlag** describes the {@link #isClockwiseArc} (true or false).
   *
   * If the type is {@link PathSegment.Move|Move} it needs the following arguments:
   *
   * `(go.PathSegment.Move, ex, ey)`
   *   - **ex, ey** describe the next starting point
   *
   * @param {number=} ex optional: the X coordinate of the end point, or the startAngle of an Arc.
   * @param {number=} ey optional: the Y coordinate of the end point, or the sweepAngle of an Arc.
   * @param {number=} x1 optional: the X coordinate of the first bezier control point, or the centerX of an Arc, or the radiusX of an SvgArc.
   * @param {number=} y1 optional: the Y coordinate of the first bezier control point, or the centerY of an Arc, or the radiusY of an SvgArc.
   * @param {number=} x2 optional: the X coordinate of the second cubic bezier control point, or the radiusX of an Arc, or the xAxisRotation of an SvgArc.
   * @param {number|boolean=} y2 optional: the Y coordinate of the second cubic bezier control point, or the radiusY of an Arc, or whether this is the larger arc of an SvgArc.
   * @param {boolean=} clockwise optional: whether an SvgArc goes clockwise or counterclockwise.
   */
  constructor(type?: EnumValue, ex?: number, ey?: number, x1?: number, y1?: number, x2?: number, y2?: number | boolean, clockwise?: boolean) {
    Set.uniqueHash(this);
    this._isFrozen = false;
    if (type === undefined) {
      type = PathSegment.Line;
    } else {
      if (Debug) EnumValue.checkEnumValue(type, PathSegment, PathSegment, 'constructor:type');
    }

    this._type = type;

    if (ex !== undefined) {
      if (Debug) Util.checkRealNumber(ex, PathSegment, 'ex');
      this._endX = ex;  // either endX or startAngle
    } else {
      this._endX = 0;
    }
    if (ey !== undefined) {
      if (Debug) Util.checkRealNumber(ey, PathSegment, 'ey');
      this._endY = ey;  // either endY or sweepAngle
    } else {
      this._endY = 0;
    }

    if (x1 === undefined) x1 = 0;
    if (y1 === undefined) y1 = 0;
    if (x2 === undefined) x2 = 0;
    if (y2 === undefined) y2 = 0;  // might be number or boolean

    // new PathSegment(PathSegment.SvgArc, endX, endY, radiusX, radiusY, xAxisRotation, largeArcFlag, clockwiseFlag)
    if (type === PathSegment.SvgArc) {
      let val = x2;
      val = val % 360;
      if (val < 0) val += 360;
      this._point1X = val;  // xAxisRotation
      this._point1Y = 0;  // not used by SvgArc

      if (Debug) Util.checkRealNumber(x1, PathSegment, 'x1');
      this._point2X = Math.max(x1, 0);  // radiusX
      if (Debug) Util.checkRealNumber(y1, PathSegment, 'y1');
      this._point2Y = Math.max(y1, 0);  // radiusY

      if (typeof y2 === 'boolean') {
        this._largeArcFlag = !!y2; // large arc flag
      } else {
        this._largeArcFlag = false;
      }

      this._clockwiseFlag = !!clockwise; // clockwise flag
    } else {
      // new PathSegment(PathSegment.Line,            endX,       endY)
      // new PathSegment(PathSegment.QuadraticBezier, endX,       endY,  controlX,  controlY)
      // new PathSegment(PathSegment.Bezier,          endX,       endY, control1X, control1Y, control2X, control2Y)
      // new PathSegment(PathSegment.Arc,       startAngle, sweepAngle,   centerX,   centerY,   radiusX,   radiusY)
      if (Debug) Util.checkRealNumber(x1, PathSegment, 'x1');
      this._point1X = x1;  // either control1X or centerX
      if (Debug) Util.checkRealNumber(y1, PathSegment, 'y1');
      this._point1Y = y1;  // either control1Y or centerY

      if (Debug) Util.checkRealNumber(x2, PathSegment, 'x2');
      if (type === PathSegment.Arc) x2 = Math.max(x2, 0);
      this._point2X = x2;  // either control2X or radiusX
      if (typeof y2 === 'number') {
        if (type === PathSegment.Arc) y2 = Math.max(y2, 0);
        this._point2Y = y2;  // either control2Y or radiusY
      } else {
        this._point2Y = 0;
      }

      // not used by these types of segments, but need to be set for uniformity:
      this._largeArcFlag = false;
      this._clockwiseFlag = false;
    }

    this._isClosed = false;

    this._invalidBounds = true;
    this._beziers = null;
  }

  /**
   * Create a copy of this PathSegment, of the same type and with the same point values.
   * @expose
   * @this {PathSegment}
   * @return {PathSegment}
   */
  public copy(): PathSegment {
    const copy = new PathSegment();
    copy._type = this._type;
    copy._endX = this._endX;
    copy._endY = this._endY;
    copy._point1X = this._point1X;
    copy._point1Y = this._point1Y;
    copy._point2X = this._point2X;
    copy._point2Y = this._point2Y;
    copy._largeArcFlag = this._largeArcFlag;
    copy._clockwiseFlag = this._clockwiseFlag;
    copy._isClosed = this._isClosed;
    copy._invalidBounds = this._invalidBounds;
    // don't copy _beziers
    return copy;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {PathSegment}
   * @param {PathSegment} s
   * @return {boolean}
   */
  public equalsApprox(s: PathSegment): boolean {
    if (!(s instanceof PathSegment)) return false;
    if (this.type !== s.type) return false;
    if (this.isClosed !== s.isClosed) return false;
    switch (this.type) {
      case PathSegment.Move:
      case PathSegment.Line:
        return Geo.isApprox(this.endX, s.endX) &&
          Geo.isApprox(this.endY, s.endY);
      case PathSegment.Bezier:
        return Geo.isApprox(this.endX, s.endX) &&
          Geo.isApprox(this.endY, s.endY) &&
          Geo.isApprox(this.point1X, s.point1X) &&
          Geo.isApprox(this.point1Y, s.point1Y) &&
          Geo.isApprox(this.point2X, s.point2X) &&
          Geo.isApprox(this.point2Y, s.point2Y);
      case PathSegment.QuadraticBezier:
        return Geo.isApprox(this.endX, s.endX) &&
          Geo.isApprox(this.endY, s.endY) &&
          Geo.isApprox(this.point1X, s.point1X) &&
          Geo.isApprox(this.point1Y, s.point1Y);
      case PathSegment.Arc:
        return Geo.isApprox(this.startAngle, s.startAngle) &&
          Geo.isApprox(this.sweepAngle, s.sweepAngle) &&
          Geo.isApprox(this.centerX, s.centerX) &&
          Geo.isApprox(this.centerY, s.centerY) &&
          Geo.isApprox(this.radiusX, s.radiusX) &&
          Geo.isApprox(this.radiusY, s.radiusY);
      case PathSegment.SvgArc:
        return this.isClockwiseArc === s.isClockwiseArc &&
          this.isLargeArc === s.isLargeArc &&
          Geo.isApprox(this.xAxisRotation, s.xAxisRotation) &&
          Geo.isApprox(this.endX, s.endX) &&
          Geo.isApprox(this.endY, s.endY) &&
          Geo.isApprox(this.radiusX, s.radiusX) &&
          Geo.isApprox(this.radiusY, s.radiusY);
      default:
        return false;
    }
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {PathSegment}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === PathSegment) {
      this.type = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {PathSegment}
   * @param {number=} precision Optional: the amount of decimal places to round numbers to for debugging purposes,
   * leaves number unchanged by default.
   * @return {string}
   */
  public toString(precision?: number): string {
    if (precision === undefined) precision = -1;
    let str = '';
    switch (this.type) {
      case PathSegment.Move:
        if (precision < 0) {
          str = 'M' + this.endX.toString() + ' ' + this.endY.toString();
        } else {
          str = 'M' + this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
        break;
      case PathSegment.Line:
        if (precision < 0) {
          str = 'L' + this.endX.toString() + ' ' + this.endY.toString();
        } else {
          str = 'L' + this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
        break;
      case PathSegment.Bezier:
        if (precision < 0) {
          str = 'C' + this.point1X.toString() + ' ' + this.point1Y.toString() + ' ' +
            this.point2X.toString() + ' ' + this.point2Y.toString() + ' ' +
            this.endX.toString() + ' ' + this.endY.toString();
        } else {
          str = 'C' + this.point1X.toFixed(precision) + ' ' + this.point1Y.toFixed(precision) + ' ' +
            this.point2X.toFixed(precision) + ' ' + this.point2Y.toFixed(precision) + ' ' +
            this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
        break;
      case PathSegment.QuadraticBezier:
        if (precision < 0) {
          str = 'Q' + this.point1X.toString() + ' ' + this.point1Y.toString() + ' ' +
            this.endX.toString() + ' ' + this.endY.toString();
        } else {
          str = 'Q' + this.point1X.toFixed(precision) + ' ' + this.point1Y.toFixed(precision) + ' ' +
            this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
        break;
      case PathSegment.Arc:
        if (precision < 0) {
          str = 'B' + this.startAngle.toString() + ' ' + this.sweepAngle.toString() + ' ' +
            this.centerX.toString() + ' ' + this.centerY.toString() + ' ' +
            this.radiusX.toString() + ' ' + this.radiusY.toString();
        } else {
          str = 'B' + this.startAngle.toFixed(precision) + ' ' + this.sweepAngle.toFixed(precision) + ' ' +
            this.centerX.toFixed(precision) + ' ' + this.centerY.toFixed(precision) + ' ' +
            this.radiusX.toFixed(precision) + ' ' + this.radiusY.toFixed(precision);
        }
        break;
      case PathSegment.SvgArc:
        if (precision < 0) {
          str = 'A' +
            this.radiusX.toString() + ' ' + this.radiusY.toString() + ' ' +
            this.xAxisRotation.toString() + ' ' +
            (this.isLargeArc ? 1 : 0) + ' ' +
            (this.isClockwiseArc ? 1 : 0) + ' ' +
            this.endX.toString() + ' ' + this.endY.toString();
        } else {
          str = 'A' +
            this.radiusX.toFixed(precision) + ' ' + this.radiusY.toFixed(precision) + ' ' +
            this.xAxisRotation.toFixed(precision) + ' ' +
            (this.isLargeArc ? 1 : 0) + ' ' +
            (this.isClockwiseArc ? 1 : 0) + ' ' +
            this.endX.toFixed(precision) + ' ' + this.endY.toFixed(precision);
        }
        break;
      default:
        str = this.type.toString();
    }
    return str + (this._isClosed ? 'z' : '');
  }

  /**
   * For beginning a new subpath, a value for {@link PathSegment#type}.
   * @constant
   */
  public static Move = new EnumValue(PathSegment, 'Move', 0);

  /**
   * For drawing a straight line segment, a value for {@link PathSegment#type}.
   * @constant
   */
  public static Line = new EnumValue(PathSegment, 'Line', 1);

  /**
   * For drawing a cubic bezier segment, a value for {@link PathSegment#type}.
   * @constant
   */
  public static Bezier = new EnumValue(PathSegment, 'Bezier', 2);

  /**
   * For drawing a quadratic bezier segment, a value for {@link PathSegment#type}.
   * @constant
   */
  public static QuadraticBezier = new EnumValue(PathSegment, 'QuadraticBezier', 3);

  /**
   * For drawing an arc segment, a value for {@link PathSegment#type}.
   * @constant
   */
  public static Arc = new EnumValue(PathSegment, 'Arc', 4);

  /**
   * For drawing an SVG arc segment, a value for {@link PathSegment#type}.
   *
   * See: <a href="https://www.w3.org/TR/SVG/paths.html#PathDataEllipticalArcCommands">SVG Arc specification (w3.org)</a>
   * @constant
   * @since 1.1
   */
  public static SvgArc = new EnumValue(PathSegment, 'SvgArc', 4);

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {PathSegment}
   * @return {PathSegment}
   */
  public freeze(): PathSegment {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {PathSegment}
   * @return {PathSegment}
   */
  public thaw(): PathSegment {
    this._isFrozen = false;
    return this;
  }

  /**
   * Closes the path after this PathSegment
   * @this {PathSegment}
   * @return {PathSegment} returns this PathSegment.
   */
  public close(): PathSegment {
    this._isClosed = true;
    return this;
  }

  /**
   * @hidden @internal
   * Constructs an array of bezier curves that represent an Arc
   * @this {PathSegment}
   * @param {PathFigure} fig
   * @return {Array.<Array.<number>>} of arc descriptor objects.
   */
  public buildBeziers(fig: PathFigure): Array<Array<number>> {
    if (this._beziers !== null && fig._invalidBounds === false) return this._beziers;
    const rx = this.radiusX;
    let ry = this.radiusY;
    if (ry === undefined) ry = rx;

    if (rx === 0 || ry === 0) {
      this._beziers = [];
      return this._beziers;
    }

    const centerx = this._point1X;
    const centery = this._point1Y;
    const curves = Geo.createArc(0, 0,
      (rx < ry) ? rx : ry, // use the smaller radius
      this.startAngle,
      (this.startAngle + this.sweepAngle),
      false);

    if (rx !== ry) {
      // we need to normalize all curves before continuing based on x1/y1 equalling 0/0
      const tt = Transform.alloc();
      tt.reset();
      if (rx < ry) { // rx is smaller, scale up y
        tt.scale(1, ry / rx);
      } else { // ry is smaller, scale up x
        tt.scale(rx / ry, 1);
      }
      Geometry.transformCurves(curves, tt);
      Transform.free(tt);
    }

    const l = curves.length;

    for (let i = 0; i < l; i++) {
      const c = curves[i];
      c[0] += centerx;
      c[1] += centery;
      c[2] += centerx;
      c[3] += centery;
      c[4] += centerx;
      c[5] += centery;
      c[6] += centerx;
      c[7] += centery;
    }

    this._beziers = curves;
    return this._beziers;
  }

  /**
   * @hidden @internal
   * Constructs an array of bezier curves that represent an Arc
   * @this {PathSegment}
   * @param {PathFigure} fig
   * @param {number} lpx
   * @param {number} lpy
   * @return {Array.<Array.<number>>} of arc descriptor objects.
   */
  public buildSVGBeziers(fig: PathFigure, lpx: number, lpy: number): Array<Array<number>> {
    if (this._beziers !== null && fig._invalidBounds === false) return this._beziers;
    // otherwise make a new list

    // REQS:
    const currx = lpx;
    const curry = lpy;
    let rx = this._point2X;
    let ry = this._point2Y;

    // Simple way to avoid degenerate geometries
    // Curve approximates a straight line
    if (rx === 0) rx = 0.0001;
    if (ry === 0) ry = 0.0001;

    // pay attention to which is used below
    const xAxisRotationRadians = this._point1X * (Math.PI / 180.0);
    const largeArcFlag = this._largeArcFlag;
    const clockwiseFlag = this._clockwiseFlag;
    const cpx = this._endX;
    const cpy = this._endY;
    // rx, ry xAxisRotation, largeArcFlag, clockwiseFlag, cpx, cp.y

    // Conversion from endpoint to center parameterization
    // https://www.w3.org/TR/SVG11/implnote.html#ArcImplementationNotes
    // x1', y1'
    const cos = Math.cos(xAxisRotationRadians);
    const sin = Math.sin(xAxisRotationRadians);
    const currpx = cos * (currx - cpx) / 2.0 + sin * (curry - cpy) / 2.0;
    const currpy = -sin * (currx - cpx) / 2.0 + cos * (curry - cpy) / 2.0;
    // adjust radii
    const l = (currpx * currpx) / (rx * rx) + (currpy * currpy) / (ry * ry);
    if (l > 1) {
      rx *= Math.sqrt(l);
      ry *= Math.sqrt(l);
    }
    // cx', cy'
    let s = (largeArcFlag === clockwiseFlag ? -1 : 1) * Math.sqrt(
      ((rx * rx * ry * ry) - (rx * rx * currpy * currpy) - (ry * ry * currpx * currpx)) /
      (rx * rx * currpy * currpy + ry * ry * currpx * currpx)
    );
    if (isNaN(s)) s = 0;
    let cppx = s * rx * currpy / ry;
    let cppy = s * -ry * currpx / rx;
    if (isNaN(cppx)) cppx = 0;
    if (isNaN(cppy)) cppy = 0;
    // cx, cy
    const centpx = (currx + cpx) / 2.0 + cos * cppx - sin * cppy;
    const centpy = (curry + cpy) / 2.0 + sin * cppx + cos * cppy;

    // helper functions:

    // vector magnitude
    /**
     * @hidden @internal
     * @param {number} x
     * @param {number} y
     * @return {number}
     */
    const M = function (x: number, y: number): number { return Math.sqrt(x * x + y * y); };

    // ratio between two vectors
    /**
     * @hidden @internal
     * @param {number} ux1
     * @param {number} uy1
     * @param {number} vx1
     * @param {number} vy1
     * @return {number}
     */
    const R = function (ux1: number, uy1: number, vx1: number, vy1: number): number { return (ux1 * vx1 + uy1 * vy1) / (M(ux1, uy1) * M(vx1, vy1)); };

    // angle between two vectors
    /**
     * @hidden @internal
     * @param {number} ux1
     * @param {number} uy1
     * @param {number} vx1
     * @param {number} vy1
     * @return {number}
     */
    const A = function (ux1: number, uy1: number, vx1: number, vy1: number): number { return (ux1 * vy1 < uy1 * vx1 ? -1 : 1) * Math.acos(R(ux1, uy1, vx1, vy1)); };

    // initial angle
    const a1 = A(1, 0, (currpx - cppx) / rx, (currpy - cppy) / ry);
    // angle delta
    const ux = (currpx - cppx) / rx;
    const uy = (currpy - cppy) / ry;
    const vx = (-currpx - cppx) / rx;
    const vy = (-currpy - cppy) / ry;
    let ad = A(ux, uy, vx, vy);
    const rat = R(ux, uy, vx, vy);
    if (rat <= -1) ad = Math.PI;
    else if (rat >= 1) ad = 0;

    if (!clockwiseFlag && ad > 0) ad = ad - 2 * Math.PI;
    if (clockwiseFlag && ad < 0) ad = ad + 2 * Math.PI;

    /*
    // for markers
    var dir = 1 - clockwiseFlag ? 1.0 : -1.0;
    var ah = a1 + dir * (ad / 2.0);
    var halfWayx = centpx + rx * Math.cos(ah);
    var halfWayy = centpy + ry * Math.sin(ah);
    */

    const rr = (rx > ry) ? rx : ry;
    const sx = rx > ry ? 1 : rx / ry;
    const sy = rx > ry ? ry / rx : 1;

    // x, y, radius, startAngle, endAngle
    const curves = Geo.createArc(0, 0, rr, a1, a1 + ad, true);

    const tt = Transform.alloc();
    tt.reset();
    tt.translate(centpx, centpy);
    tt.rotate(this._point1X, 0, 0); // rotate uses degrees not radians
    tt.scale(sx, sy);
    Geometry.transformCurves(curves, tt);
    Transform.free(tt);

    /* If we wanted to make path segments instead
    for (var ii = 0; ii < curves.length; ii++) {
      var c = curves[ii];
      ctx.bezierTo(c[2], c[3], c[4], c[5], c[6], c[7]);
    }
    */

    /* for drawing maybe
    ctx.translate(centpx, centpy);
    ctx.rotate(xAxisRotationRadians);
    ctx.scale(sx, sy);
    ctx.arc(0, 0, r, a1, a1 + ad, 1 - clockwiseFlag);
    ctx.scale(1/sx, 1/sy);
    ctx.rotate(-xAxisRotationRadians);
    ctx.translate(-centpx, -centpy);
    */

    this._beziers = curves;
    return this._beziers;
  }

  /**
   * Gets or sets whether the path is closed after this PathSegment.
   * Default value is false.
   */
  /** @export */ get isClosed(): boolean { return this._isClosed; }
  /** @export */ set isClosed(value) {
    if (this._isClosed !== value) {
      this._isClosed = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Gets or sets the type of the PathSegment.
   * The value must be one of the following:
   * {@link PathSegment.Line}, {@link PathSegment.Bezier},
   * {@link PathSegment.QuadraticBezier}, {@link PathSegment.Arc},
   * {@link PathSegment.Move}, {@link PathSegment.Arc}, and {@link PathSegment.SvgArc}.
   */
  /** @export */ get type(): EnumValue { return this._type; }
  /** @export */ set type(value: EnumValue) {
    if (Debug) EnumValue.checkEnumValue(value, PathSegment, PathSegment, 'type');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._type = value;
    this._invalidBounds = true;
  }

  // all PathSegments:
  /**
   * Gets or sets the X coordinate of the end point for all types of PathSegment except {@link PathSegment.Arc|Arc}.
   * The default value is zero.
   */
  /** @export */ get endX(): number { return this._endX; }
  /** @export */ set endX(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'endX');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._endX = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the Y coordinate of the end point for all types of PathSegment except {@link PathSegment.Arc|Arc}.
   * The default value is zero.
   */
  /** @export */ get endY(): number { return this._endY; }
  /** @export */ set endY(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'endY');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._endY = value;
    this._invalidBounds = true;
  }

  // Bezier PathSegments:
  /**
   * Gets or sets the X value of the first control point for a PathSegment
   * of type {@link PathSegment.Bezier|Bezier} or {@link PathSegment.QuadraticBezier|QuadraticBezier}.
   */
  /** @export */ get point1X(): number { return this._point1X; }
  /** @export */ set point1X(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'point1X');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point1X = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the Y value of the first control point for a PathSegment
   * of type {@link PathSegment.Bezier|Bezier} or {@link PathSegment.QuadraticBezier|QuadraticBezier}.
   */
  /** @export */ get point1Y(): number { return this._point1Y; }
  /** @export */ set point1Y(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'point1Y');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point1Y = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the X value of the second control point for a PathSegment
   * of type cubic {@link PathSegment.Bezier|Bezier}.
   */
  /** @export */ get point2X(): number { return this._point2X; }
  /** @export */ set point2X(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'point2X');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point2X = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the Y value of the second control point for a PathSegment
   * of type cubic {@link PathSegment.Bezier|Bezier}.
   */
  /** @export */ get point2Y(): number { return this._point2Y; }
  /** @export */ set point2Y(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'point2Y');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point2Y = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the center X value of the Arc for a PathSegment of type {@link PathSegment.Arc|Arc}.
   */
  /** @export */ get centerX(): number { return this._point1X; }
  /** @export */ set centerX(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'centerX');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point1X = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the center Y value of the Arc for a PathSegment of type {@link PathSegment.Arc|Arc}.
   */
  /** @export */ get centerY(): number { return this._point1Y; }
  /** @export */ set centerY(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'centerY');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point1Y = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the X value of the radius for a PathSegment of type {@link PathSegment.Arc|Arc} or {@link PathSegment.SvgArc|SvgArc}.
   * Value must be a positive number.
   */
  /** @export */ get radiusX(): number { return this._point2X; }
  /** @export */ set radiusX(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'radiusX');
    if (value < 0) Util.throwRangeError(value, '>= zero', PathSegment, 'radiusX');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point2X = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the Y value of the radius for a PathSegment of type {@link PathSegment.Arc|Arc} or {@link PathSegment.SvgArc|SvgArc}.
   * Value must be a positive number.
   */
  /** @export */ get radiusY(): number { return this._point2Y; }
  /** @export */ set radiusY(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'radiusY');
    if (value < 0) Util.throwRangeError(value, '>= zero', PathSegment, 'radiusY');
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point2Y = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the starting angle for a PathSegment of type {@link PathSegment.Arc|Arc}.
   * Value must within the range: (0 <= value < 360).
   */
  /** @export */ get startAngle(): number { return this._endX; }
  /** @export */ set startAngle(value: number) {
    if (this._endX !== value) {
      if (this._isFrozen) Util.throwFrozen(this, value);
      if (Debug) Util.checkRealNumber(value, PathSegment, 'startAngle');
      value = value % 360;
      if (value < 0) value += 360;
      this._endX = value;
      this._invalidBounds = true;
    }
  }

  /**
   * Gets or sets the length of angle in degrees, or amount of arc to "sweep" for a PathSegment of type {@link PathSegment.Arc|Arc}.
   * Must be between -360 and 360, inclusive.
   */
  /** @export */ get sweepAngle(): number { return this._endY; }
  /** @export */ set sweepAngle(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'sweepAngle');
    if (this._isFrozen) Util.throwFrozen(this, value);
    if (value > 360) value = 360;
    if (value < -360) value = -360;
    this._endY = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the clockwise-flag for a PathSegment of type {@link PathSegment.SvgArc|SvgArc}.
   * SVG Arcs specify a radius and an endpoint, and are always a portion of an ellipse.
   * The parameters allow for two potential ellipses and four potential arcs.
   * A clockwise-flag set to true will use one of the two possible positive-angle arcs,
   * and false will use one of the two negative-angle arcs. Which arc is chosen (small or large)
   * depends on the value of {@link #isLargeArc}.
   * For more information see the visual examples in the
   * <a href="https://www.w3.org/TR/SVG/paths.html#PathDataEllipticalArcCommands">SVG Arc specification (w3.org)</a>
   */
  /** @export */ get isClockwiseArc(): boolean { return this._clockwiseFlag; }
  /** @export */ set isClockwiseArc(value: boolean) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._clockwiseFlag = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the large-arc-flag for a PathSegment of type {@link PathSegment.SvgArc|SvgArc}.
   * SVG Arcs specify a radius and an endpoint, and are always a portion of an ellipse.
   * The parameters allow for two potential ellipses and four potential arcs.
   * A large-arc-flag set to true will choose the larger of the two arc sweeps.
   * Which way the arc sweeps (positive angle or negative angle) depends on the value of {@link #isClockwiseArc}
   * For more information see the visual examples in the
   * <a href="https://www.w3.org/TR/SVG/paths.html#PathDataEllipticalArcCommands">SVG Arc specification (w3.org)</a>
   */
  /** @export */ get isLargeArc(): boolean { return this._largeArcFlag; }
  /** @export */ set isLargeArc(value: boolean) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._largeArcFlag = value;
    this._invalidBounds = true;
  }

  /**
   * Gets or sets the X-axis rotation for a PathSegment of type {@link PathSegment.SvgArc|SvgArc}.
   * X-axis rotation is used to rotate the ellipse that the arc is created from,
   * and must be between 0 and 360 degrees.
   * Default is 0.
   */
  /** @export */ get xAxisRotation(): number { return this._point1X; }
  /** @export */ set xAxisRotation(value: number) {
    if (Debug) Util.checkRealNumber(value, PathSegment, 'xAxisRotation');
    value = value % 360;
    if (value < 0) value += 360;
    if (this._isFrozen) Util.throwFrozen(this, value);
    this._point1X = value;
    this._invalidBounds = true;
  }
} // #end PathSegment
PathSegment['className'] = 'PathSegment';

PathSegment['Move'] = PathSegment.Move;
PathSegment['Line'] = PathSegment.Line;
PathSegment['Bezier'] = PathSegment.Bezier;
PathSegment['QuadraticBezier'] = PathSegment.QuadraticBezier;
PathSegment['Arc'] = PathSegment.Arc;
PathSegment['SvgArc'] = PathSegment.SvgArc;

// import { List, Set } from './collections';
// import { Debug } from './debug';
// import { StreamGeometryContext } from './figures';
// import { Geo } from './geo';
// import { GraphObject } from './graphObject';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { Transform } from './transform';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
InputEvent
DiagramEvent
*/



/**
 * An InputEvent represents a mouse or keyboard or touch event.
 * The principal properties hold information about a particular input event.
 * These properties include the {@link #documentPoint} at which a mouse event
 * occurred in document coordinates,
 * the corresponding point in view/element coordinates, {@link #viewPoint},
 * the {@link #key} for keyboard events,
 * and the {@link #modifiers} and {@link #button} at the time.
 * Additional descriptive properties include {@link #clickCount}, {@link #delta},
 * {@link #timestamp}, and the source event {@link #event} (if available).
 *
 * Many of its properties are provided for convenient access to the state of the input event,
 * such as asking whether the {@link #control} key was held down at the time,
 * or the {@link #targetObject} (a {@link GraphObject}) that the mouse was over .
 *
 * When real events fire on the Diagram, InputEvents are created automatically set update the value of {@link Diagram#lastInput}.
 * These events set the value of {@link #event} with the backing browser-defined Event,
 * which may be a MouseEvent, KeyboardEvent, PointerEvent, TouchEvent, and so on.
 *
 * InputEvents backed by MouseEvents set both {@link #button}, the button that caused the action,
 * and {@link #buttons}, the set of buttons currently pressed after the action has happened.
 * By default a user-created InputEvent sets {@link #button} and {@link #buttons} as if the event was a left-click.
 *
 * You can create InputEvents and set the value of {@link Diagram#lastInput} in order to simulate user actions in tools.
 * This can be useful for testing. See the <a href="../../extensions/Robot.html">Robot extension sample</a> for
 * an example of creating InputEvents to simulate user input.
 * @unrestricted
 */
export class InputEvent {
  /** @hidden @internal */ public static className = 'InputEvent';
  /** @hidden @internal */ private _diagram: Diagram;
  /** @hidden @internal */ private _viewPoint: Point;
  /** @hidden @internal */ private _documentPoint: Point;
  /** @hidden @internal */ private _modifiers: number;
  /** @hidden @internal */ private _button: number;
  /** @hidden @internal */ private _buttons: number;
  /** @hidden @internal */ private _key: string;
  /** @hidden @internal */ private _down: boolean;
  /** @hidden @internal */ private _up: boolean;
  /** @hidden @internal */ private _clickCount: number;
  /** @hidden @internal */ private _delta: number;
  /** @hidden @internal */ private _isMultiTouch: boolean;
  /** @hidden @internal */ private _handled: boolean;
  /** @hidden @internal */ private _bubbles: boolean;
  /** @hidden @internal */ private _event: Event | null;
  /** @hidden @internal */ private _timestamp: number;
  /** @hidden @internal */ private _targetDiagram: Diagram | null;
  /** @hidden @internal */ private _targetObject: GraphObject | null;

  /**
   * The InputEvent class constructor produces an empty InputEvent.
   * You should never need to call this constructor.
   * But if you do, you must first set {@link #diagram} to the Diagram before any other use.
   */
  constructor() {
    this._diagram = null;
    this._viewPoint = new Point(0, 0).freeze();  // keep own instance of a Point
    this._documentPoint = new Point(0, 0).freeze();  // keep own instance of a Point
    this._modifiers = 0;
    this._button = 0;
    this._buttons = 1;
    this._key = '';
    this._down = false;
    this._up = false;
    this._clickCount = 0;
    this._delta = 0;
    this._isMultiTouch = false;
    this._handled = false;
    this._bubbles = false;
    this._event = null;
    this._timestamp = 0;
    this._targetDiagram = null;
    this._targetObject = null;
  }

  /**
   * Make a copy of this InputEvent.
   * @expose
   * @this {InputEvent}
   * @return {InputEvent}
   */
  public copy(): InputEvent {
    const copy = new InputEvent();
    return this.clone(copy);
  }

  /**
   * Clone this InputEvent's state onto a given InputEvent
   * @hidden @internal
   * @this {InputEvent}
   * @param {InputEvent} copy
   * @return {InputEvent}
   */
  public clone(copy: InputEvent): InputEvent {
    copy._diagram = this._diagram;
    // keep own instance of a Point
    copy._viewPoint.assign(this.viewPoint);
    copy._documentPoint.assign(this.documentPoint);
    copy._modifiers = this._modifiers;
    copy._button = this._button;
    copy._buttons = this._buttons;
    copy._key = this._key;
    copy._down = this._down;
    copy._up = this._up;
    copy._clickCount = this._clickCount;
    copy._delta = this._delta;
    copy._isMultiTouch = this._isMultiTouch;
    copy._handled = this._handled;
    copy._bubbles = this._bubbles;
    copy._event = this._event;
    copy._timestamp = this._timestamp;
    copy._targetDiagram = this._targetDiagram;
    copy._targetObject = this._targetObject;
    return copy;
  }

  /**
   * @hidden @internal
   * @this {InputEvent}
   * @return {string}
   */
  public toString(): string {
    let s = '^';
    if (this.modifiers !== 0) s += 'M:' + this.modifiers;
    if (this.button !== 0) s += 'B:' + this.button;
    if (this.key !== '') s += 'K:' + this.key;
    if (this.clickCount !== 0) s += 'C:' + this.clickCount;
    if (this.delta !== 0) s += 'D:' + this.delta;
    if (this.handled) s += 'h';
    if (this.bubbles) s += 'b';
    if (this.documentPoint !== null) s += '@' + this.documentPoint.toString();
    return s;
  }

  /**
   * Gets the source diagram associated with the event.
   */
  /** @export */ get diagram(): Diagram { return this._diagram; }
  /** @export */ set diagram(value: Diagram) { this._diagram = value; }

  /**
   * Gets or sets the point at which this input event occurred.
   * The Point is in view coordinates within the viewport, not in document coordinates.
   * This should be valid for mouse events.
   * For keyboard events, this is the last available mouse point.
   * @see {@link #documentPoint}
   * @see {@link Diagram#transformViewToDoc}
   * @see {@link Diagram#transformDocToView}
   */
  /** @export */ get viewPoint(): Point { return this._viewPoint; }
  /** @export */ set viewPoint(value: Point) {
    Util.checkClass(value, Point, InputEvent, 'viewPoint');
    this._viewPoint.assign(value);  // keep own instance of a Point
  }

  /**
   * Gets or sets the point at which this input event occurred, in document coordinates.
   * The Point is in document coordinates, the same as the {@link GraphObject#position}
   * for {@link Part}s in the diagram.
   * This should be valid for mouse events.
   * For keyboard events, this is the last available mouse point.
   * @see {@link #viewPoint}
   * @see {@link Diagram#transformViewToDoc}
   * @see {@link Diagram#transformDocToView}
   */
  /** @export */ get documentPoint(): Point { return this._documentPoint; }
  /** @export */ set documentPoint(value: Point) {
    Util.checkClass(value, Point, InputEvent, 'documentPoint');
    this._documentPoint.assign(value);  // keep own instance of a Point
  }

  /**
   * @hidden @internal
   * @export
   * @this {InputEvent}
   * @param {number} n the touch number, starting from 0. Typical values are 0 or 1 for two-touch input.
   * @param {Point} p a point to be modified
   * @return {Point}
   */
  public getMultiTouchViewPoint(n: number, p: Point): Point {
    const diagram = this.diagram;
    if (diagram === null) return p;

    diagram._getMultiTouchViewpoint(this.event as TouchEvent, n, p);
    return p;
  }

  /**
   * @hidden @internal
   * @export
   * @this {InputEvent}
   * @param {number} n the touch number, starting from 0. Typical values are 0 or 1 for two-touch input.
   * @param {Point} p a point to be modified
   * @return {Point}
   */
  public getMultiTouchDocumentPoint(n: number, p: Point): Point {
    const diagram = this.diagram;
    if (diagram === null) return p;

    diagram._getMultiTouchViewpoint(this.event as TouchEvent, n, p);
    p.assign(diagram.transformViewToDoc(p));
    return p;
  }

  /**
   * Gets or sets the modifier keys that were used with the mouse or keyboard event.
   * The number will be a combination of flags representing Control, Shift, Alt or Meta.
   * @see {@link #control}
   * @see {@link #shift}
   * @see {@link #alt}
   * @see {@link #meta}
   */
  /** @export */ get modifiers(): number { return this._modifiers; }
  /** @export */ set modifiers(value: number) { this._modifiers = value; }

  /**
   * Gets or sets the numerical value representing the mouse button that caused this event.
   *
   * Only one mouse button can cause an event at a time,
   * but the set of all buttons currently pressed is kept in {@link #buttons}.
   *
   * This property describes nothing during a `mousemove` event,
   * since no button press causes the event. Instead, use the convenience properties
   * {@link #left}, {@link #middle}, or {@link #right}, or use the value
   * of {@link #buttons} to determine which mouse buttons are
   * held during `mousemove` events.
   *
   * Common values for this property:
   *   - 0: left mouse button
   *   - 1: middle mouse button
   *   - 2: right mouse button
   *
   * Other values are possible if the mouse has additional buttons.
   *
   * If there is no associated {@link #event}, setting this also
   * sets the {@link #buttons} flags to only this button.
   *
   * This property is valid if this is a mouse event.
   * @see {@link #buttons}
   * @see {@link #left}
   * @see {@link #middle}
   * @see {@link #right}
   */
  /** @export */ get button(): number { return this._button; }
  /** @export */ set button(value: number) {
    this._button = value;
    // If this is a simulated InputEvent, also set buttons flag:
    if (this.event !== null) return;
    switch (value) {
      case 0: this.buttons = 1; return;
      case 1: this.buttons = 4; return;
      case 2: this.buttons = 2; return;
    }
  }

  /**
   * Gets or sets the set of buttons that are currently being held down.
   * If this is a `mouseup` event, this set of buttons does not include
   * the button that was just released, which will be the value of {@link #button}.
   *
   * Common values for this property:
   *   - 0: not holding down any button
   *   - 1: left mouse button
   *   - 2: right mouse button
   *   - 3: both left and right mouse buttons
   *   - 4: middle mouse button
   *   - 5: middle and left mouse buttons
   *   - 6: middle and right mouse buttons
   *   - 7: all three common mouse buttons
   *
   * Other values are possible if the mouse has additional buttons.
   *
   * This property is valid if this is a mouse event.
   * @see {@link #button}
   * @see {@link #left}
   * @see {@link #middle}
   * @see {@link #right}
   */
  /** @export */ get buttons(): number { return this._buttons; }
  /** @export */ set buttons(value: number) { this._buttons = value; }

  /**
   * Gets or sets the key pressed or released as this event.
   * This property is valid if this is a keyboard event.
   */
  /** @export */ get key(): string { return this._key; }
  /** @export */ set key(value: string) { this._key = value; }

  /**
   * Gets or sets whether the InputEvent represents a mouse-down or a key-down event.
   * The default value is false.
   * @since 1.1
   */
  /** @export */ get down(): boolean { return this._down; }
  /** @export */ set down(value: boolean) { this._down = value; }

  /**
   * Gets or sets whether the InputEvent represents a mouse-up or a key-up event.
   * The default value is false.
   * @since 1.1
   */
  /** @export */ get up(): boolean { return this._up; }
  /** @export */ set up(value: boolean) { this._up = value; }

  /**
   * Gets or sets whether this event represents a click or a double-click.
   * It is zero if not a click; one if a single-click; two if a double-click.
   * This property is valid if this is a mouse event.
   */
  /** @export */ get clickCount(): number { return this._clickCount; }
  /** @export */ set clickCount(value: number) { this._clickCount = value; }

  /**
   * Gets or sets the amount of change associated with a mouse-wheel rotation.
   * It is an abstract number, either positive or negative.
   * This property is valid if this is a mouse-wheel event.
   */
  /** @export */ get delta(): number { return this._delta; }
  /** @export */ set delta(value: number) { this._delta = value; }

  /**
   * This property is true when the InputEvent is caused by a touch event that registered more than one touch.
   * @see {@link #isTouchEvent}
   * @since 1.5
   */
  /** @export */ get isMultiTouch(): boolean { return this._isMultiTouch; }
  /** @export */ set isMultiTouch(value: boolean) { this._isMultiTouch = value; }

  /**
   * Gets or sets whether an InputEvent that applies to a {@link GraphObject} and bubbles
   * up the chain of containing {@link Panel}s is stopped from continuing up the chain.
   *
   * Some event handlers may set this to true to avoid getting any behavior from the containing Panels.
   * The default value is false.
   */
  /** @export */ get handled(): boolean { return this._handled; }
  /** @export */ set handled(value: boolean) { this._handled = value; }

  //    * Gets or sets whether the underlying {@link #event} is prevented from bubbling up
  //  * the hierarchy of HTML elements outside of the Diagram and whether any default action is canceled.
  //  *
  //  * Some event handlers may set this to true if the event should continue beyond the diagram's DIV
  //  * and result in any default behaviors by the user-agent.
  //  * The default value is false.
  /**
   * @hidden
   * Undocumented
   * @see {@link #handled}
   */
  /** @export */ get bubbles(): boolean { return this._bubbles; }
  /** @export */ set bubbles(value: boolean) { this._bubbles = value; }

  /**
   * Gets or sets the platform's user-agent-supplied event for this event.
   * It may be null if no underlying event exists.
   * @see {@link #timestamp}
   */
  /** @export */ get event(): Event | null { return this._event; }
  /** @export */ set event(value: Event | null) { this._event = value; }

  /**
   * This read-only property is true when the InputEvent is caused by a touch event.
   * @see {@link #isMultiTouch}
   * @since 1.5
   */
  /** @export */ get isTouchEvent(): boolean {
    let te = root['TouchEvent'];
    const evt = this.event;
    if (te && evt instanceof te) return true;
    te = root['PointerEvent'];
    return te && evt instanceof te && ((evt as PointerEvent).pointerType === 'touch' || (evt as PointerEvent).pointerType === 'pen');
  }

  /**
   * Gets or sets the time at which the event occurred, in milliseconds.
   */
  /** @export */ get timestamp(): number { return this._timestamp; }
  /** @export */ set timestamp(value: number) { this._timestamp = value; }

  /**
   * Gets or sets the diagram associated with the canvas that the event is currently targeting.
   */
  /** @export */ get targetDiagram(): Diagram | null { return this._targetDiagram; }
  /** @export */ set targetDiagram(value: Diagram | null) { this._targetDiagram = value; }

  /**
   * Gets or sets the {@link GraphObject} that is at the current mouse point, if any.
   *
   * For those events that are bubbled up the chain of parent {@link Panel}s,
   * this property provides access to the original {@link GraphObject} where the input event occurred.
   * @see {@link #handled}
   */
  /** @export */ get targetObject(): GraphObject | null { return this._targetObject; }
  /** @export */ set targetObject(value: GraphObject | null) { this._targetObject = value; }

  // Convenience properties

  /**
   * Gets or sets whether the control key is being held down.
   * This is true if the key was pressed at the time of the event.
   * @see {@link #modifiers}
   * @see {@link #shift}
   * @see {@link #alt}
   * @see {@link #meta}
   */
  /** @export */ get control(): boolean { return (this.modifiers & 1) !== 0; }
  /** @export */ set control(value: boolean) { this.modifiers = (value ? this.modifiers | 1 : this.modifiers & ~1); }

  /**
   * Gets or sets whether the shift key is being held down.
   * This is true if the key was pressed at the time of the event.
   * @see {@link #modifiers}
   * @see {@link #control}
   * @see {@link #alt}
   * @see {@link #meta}
   */
  /** @export */ get shift(): boolean { return (this.modifiers & 4) !== 0; }
  /** @export */ set shift(value: boolean) { this.modifiers = (value ? this.modifiers | 4 : this.modifiers & ~4); }

  /**
   * Gets or sets whether the alt key is being held down.
   * This is true if the key was pressed at the time of the event.
   * @see {@link #modifiers}
   * @see {@link #control}
   * @see {@link #shift}
   * @see {@link #meta}
   */
  /** @export */ get alt(): boolean { return (this.modifiers & 2) !== 0; }
  /** @export */ set alt(value: boolean) { this.modifiers = (value ? this.modifiers | 2 : this.modifiers & ~2); }

  /**
   * Gets or sets whether the meta key is being held down.
   * This is true if the key was pressed at the time of the event.
   * @see {@link #modifiers}
   * @see {@link #control}
   * @see {@link #shift}
   * @see {@link #alt}
   */
  /** @export */ get meta(): boolean { return (this.modifiers & 8) !== 0; }
  /** @export */ set meta(value: boolean) { this.modifiers = (value ? this.modifiers | 8 : this.modifiers & ~8); }

  /**
   * Gets or sets whether the logical left mouse button is being held down.
   * This is true if the button was pressed at the time of the event.
   *
   * If this InputEvent has a {@link #event} of type `MouseEvent` with `e.type` of "mouseup" or "mousedown",
   * this property uses the value of {@link #button}. Otherwise, it uses the value of {@link #buttons}.
   *
   * When setting, this sets the value of {@link #buttons}.
   * @see {@link #button}
   * @see {@link #middle}
   * @see {@link #right}
   */
  /** @export */ get left(): boolean {
    const e = this.event;
    if (e !== null) {
      if (e.type === 'mousedown' || e.type === 'mouseup' || e.type === 'pointerdown' || e.type === 'pointerup') return this.button === 0;
    }
    return (this.buttons & 1) !== 0;
  }
  /** @export */ set left(value: boolean) {
    if (value) this.buttons |= 1;
    else this.buttons &= ~1;
  }

  /**
   * Gets or sets whether the logical right mouse button is being held down.
   * This is true if the button was pressed at the time of the event.
   * @see {@link #button}
   * @see {@link #left}
   * @see {@link #middle}
   */
  /** @export */ get right(): boolean {
    const e = this.event;
    if (e !== null) {
      if (e.type === 'mousedown' || e.type === 'mouseup' || e.type === 'pointerdown' || e.type === 'pointerup') return this.button === 2;
    }
    return (this.buttons & 2) !== 0;
  }
  /** @export */ set right(value: boolean) {
    if (value) this.buttons |= 2;
    else this.buttons &= ~2;
  }

  /**
   * Gets or sets whether the logical middle mouse button is being held down.
   * This is true if the button was pressed at the time of the event.
   * @see {@link #button}
   * @see {@link #left}
   * @see {@link #right}
   */
  /** @export */ get middle(): boolean {
    const e = this.event;
    if (e !== null) {
      if (e.type === 'mousedown' || e.type === 'mouseup' || e.type === 'pointerdown' || e.type === 'pointerup') return this.button === 1;
    }
    return (this.buttons & 4) !== 0;
  }
  /** @export */ set middle(value: boolean) {
    if (value) this.buttons |= 4;
    else this.buttons &= ~4;
  }
} // #end InputEvent
InputEvent['className'] = 'InputEvent';

/**
 * A DiagramEvent represents a more abstract event than an {@link InputEvent}.
 * They are raised on the {@link Diagram} class.
 * One can receive such events by registering a DiagramEvent listener on a Diagram
 * by calling {@link Diagram#addDiagramListener}.
 * The listener function, when called, will be passed an instance of a {@link DiagramEvent}.
 * Use the {@link #name} property to decide what kind of diagram event it is.
 * The {@link #diagram} property refers to the Diagram, and you can get additional information
 * from that, such as the {@link Diagram#lastInput}, which in turn provides information
 * such as {@link InputEvent#documentPoint} that may be relevant for that kind of DiagramEvent.
 *
 * The {@link #subject} and {@link #parameter} properties optionally provide additional information
 * about the diagram event.  The subject could be a collection of {@link Part}s or it could be
 * an individual object such as a {@link Link} or a {@link GraphObject} within a {@link Node}.
 * Everything depends on the kind of diagram event that it is.
 *
 * Some DiagramEvents such as "ObjectSingleClicked" or "BackgroundDoubleClicked" are normally
 * associated with {@link InputEvent}s.
 * Some DiagramEvents such as "SelectionMoved" or "PartRotated" are associated with the
 * results of {@link Tool}-handled gestures or {@link CommandHandler} actions.
 * Some DiagramEvents are not necessarily associated with any input events at all,
 * such as "ViewportBoundsChanged", which can happen due to programmatic
 * changes to the {@link Diagram#position} and {@link Diagram#scale} properties.
 *
 * DiagramEvents that occur during a transaction may be called before the state of the whole diagram has settled down.
 * This often means that such events occur before a layout, so nodes may not have their final positions,
 * links may not have their final routes, and the {@link Diagram#documentBounds} and {@link Diagram#viewportBounds}
 * may not yet have been updated.
 * Such events may make additional changes to the diagram, which may in turn result in additional side-effects.
 *
 * DiagramEvents that occur outside of a transaction require you to start and commit a transaction around any side-effects that you want to do.
 * However, some DiagramEvents do not allow you to make any changes to the Diagram or Model.
 *
 * Currently defined diagram event names include:
 *   - **"AnimationStarting"**, a {@link AnimationManager#defaultAnimation} is about to start;<br/>
 *     do not modify the Diagram or its Model in the event listener.
 *   - **"AnimationFinished"**, a {@link AnimationManager#defaultAnimation} just completed;<br/>
 *     do not modify the Diagram or its Model in the event listener.
 *   - **"BackgroundSingleClicked"**, a click that was not on any Part;<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"BackgroundDoubleClicked"**, a double-click that was not on any Part;<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"BackgroundContextClicked"**, a context-click that was not on any Part;<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"ChangingSelection"**, an operation is about to change the {@link Diagram#selection} collection,<br/>
 *     which is also the value of the {@link DiagramEvent#subject};<br/>
 *     do not make any changes to the selection or the diagram in the event listener;
 *     note that just setting {@link Part#isSelected} will not raise this event, but tools and commands will.
 *   - **"ChangedSelection"**, an operation has just changed the {@link Diagram#selection} collection,<br/>
 *     which is also the value of the {@link DiagramEvent#subject};<br/>
 *     do not make any changes to the selection or the diagram in the event listener;
 *     note that just setting {@link Part#isSelected} will not raise this event, but tools and commands will.
 *   - **"ClipboardChanged"**, Parts have been copied to the clipboard by {@link CommandHandler#copySelection};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Part}s;<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"ClipboardPasted"**, Parts have been copied from the clipboard into the Diagram by {@link CommandHandler#pasteSelection};<br/>
 *     the {@link DiagramEvent#subject} is the {@link Diagram#selection},<br/>
 *     and this is called within a transaction.
 *   - **"DocumentBoundsChanged"**, the area of the diagram's Parts, {@link Diagram#documentBounds}, has changed;<br/>
 *     the {@link DiagramEvent#parameter} is the old {@link Rect}
 *   - **"ExternalObjectsDropped"**, Parts have been copied into the Diagram by drag-and-drop from outside of the Diagram;<br/>
 *     the {@link DiagramEvent#subject} is the set of {@link Part}s that were dropped (which is also the {@link Diagram#selection}),
 *     the {@link DiagramEvent#parameter} is the source {@link Diagram},
 *     and this is called within a transaction.
 *   - **"GainedFocus"**, the diagram has gained keyboard focus, such as after a call to {@link Diagram#focus}.
 *   - **"InitialLayoutCompleted"**, the whole diagram layout has updated for the first time since a major change to the Diagram, such as replacing the Model;<br/>
 *     if you make any changes, you do not need to perform a transaction.
 *   - **"LayoutCompleted"**, the whole diagram layout has just been updated;<br/>
 *     if you make any changes, you do not need to perform a transaction.
 *   - **"LinkDrawn"**, the user has just created a new Link using {@link LinkingTool};<br/>
 *     the {@link DiagramEvent#subject} is the new {@link Link},<br/>
 *     and this is called within a transaction.
 *   - **"LinkRelinked"**, the user has just reconnected an existing Link using {@link RelinkingTool} or {@link DraggingTool};<br/>
 *     the {@link DiagramEvent#subject} is the modified {@link Link},<br/>
 *     the {@link DiagramEvent#parameter} is the {@link GraphObject} port that the link was disconnected from,<br/>
 *     and this is called within a transaction.
 *   - **"LinkReshaped"**, the user has just rerouted an existing Link using {@link LinkReshapingTool};<br/>
 *     the {@link DiagramEvent#subject} is the modified {@link Link},<br/>
 *     the {@link DiagramEvent#parameter} is the List of Points of the link's original route,
 *     and this is called within a transaction.
 *   - **"LostFocus"**, the diagram has lost keyboard focus, a.k.a. "blur".
 *   - **"Modified"**, the {@link Diagram#isModified} property has been set to a new value --
 *     useful for marking a window as having been modified since the last save;<br/>
 *     do not modify the Diagram or its Model in the event listener.
 *   - **"ObjectSingleClicked"**, a click that occurred on a GraphObject;<br/>
 *     the {@link DiagramEvent#subject} is the {@link GraphObject};<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"ObjectDoubleClicked"**, a double-click that occurred on a GraphObject;<br/>
 *     the {@link DiagramEvent#subject} is the {@link GraphObject};<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"ObjectContextClicked"**, a context-click that occurred on a GraphObject;<br/>
 *     the {@link DiagramEvent#subject} is the {@link GraphObject};<br/>
 *     if you make any changes, start and commit your own transaction.
 *   - **"PartCreated"**, the user inserted a new Part by {@link ClickCreatingTool};<br/>
 *     the {@link DiagramEvent#subject} is the new {@link Part},<br/>
 *     and this is called within a transaction.
 *   - **"PartResized"**, the user has changed the size of a GraphObject by {@link ResizingTool};<br/>
 *     the {@link DiagramEvent#subject} is the {@link GraphObject},<br/>
 *     the {@link DiagramEvent#parameter} is the original Size,<br/>
 *     and this is called within a transaction.
 *   - **"PartRotated"**, the user has changed the angle of a GraphObject by {@link RotatingTool};<br/>
 *     the {@link DiagramEvent#subject} is the {@link GraphObject},<br/>
 *     the {@link DiagramEvent#parameter} is the original angle in degrees,<br/>
 *     and this is called within a transaction.
 *   - **"SelectionMoved"**, the user has moved selected Parts by {@link DraggingTool};<br/>
 *     the {@link DiagramEvent#subject} is a Set of the moved Parts,<br/>
 *     and this is called within a transaction.
 *   - **"SelectionCopied"**, the user has copied selected Parts by {@link DraggingTool};<br/>
 *     the {@link DiagramEvent#subject} is Set of the newly copied Parts,<br/>
 *     and this is called within a transaction.
 *   - **"SelectionDeleted"**, the user has deleted selected Parts by {@link CommandHandler#deleteSelection};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Part}s that were deleted,<br/>
 *     and this is called within a transaction.
 *   - **"SelectionDeleting"**, the user is about to delete selected Parts by {@link CommandHandler#deleteSelection};<br/>
 *     the {@link DiagramEvent#subject} is the {@link Diagram#selection} collection of Parts to be deleted,<br/>
 *     and this is called within a transaction.
 *   - **"SelectionGrouped"**, the user has made a new Group out of the selected Parts by {@link CommandHandler#groupSelection};<br/>
 *     the {@link DiagramEvent#subject} is the new {@link Group},<br/>
 *     and this is called within a transaction.
 *   - **"SelectionUngrouped"**, the user has removed a selected Group but kept its members by {@link CommandHandler#ungroupSelection};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Group}s that were ungrouped,<br/>
 *     the {@link DiagramEvent#parameter} is the collection of former member {@link Part}s that were ungrouped,<br/>
 *     and this is called within a transaction.
 *   - **"SubGraphCollapsed"**, the user has collapsed selected Groups by {@link CommandHandler#collapseSubGraph};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Group}s that were collapsed,<br/>
 *     and this is called within a transaction.
 *   - **"SubGraphExpanded"**, the user has expanded selected Groups by {@link CommandHandler#expandSubGraph};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Group}s that were expanded,<br/>
 *     and this is called within a transaction.
 *   - **"TextEdited"**, the user has changed the string value of a TextBlock by {@link TextEditingTool};<br/>
 *     the {@link DiagramEvent#subject} is the edited {@link TextBlock},<br/>
 *     the {@link DiagramEvent#parameter} is the original string,<br/>
 *     and this is called within a transaction.
 *   - **"TreeCollapsed"**, the user has collapsed selected Nodes with subtrees by {@link CommandHandler#collapseTree};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Node}s that were collapsed,<br/>
 *     and this is called within a transaction.
 *   - **"TreeExpanded"**, the user has expanded selected Nodes with subtrees by {@link CommandHandler#expandTree};<br/>
 *     the {@link DiagramEvent#subject} is the collection of {@link Node}s that were expanded,<br/>
 *     and this is called within a transaction.
 *   - **"ViewportBoundsChanged"**, the visible area of the Diagram, {@link Diagram#viewportBounds}, has changed;<br/>
 *     the {@link DiagramEvent#subject} is an object whose "scale" property is the old {@link Diagram#scale} value,
 *     whose "position" property is the old {@link Diagram#position} value,
 *     and whose "bounds" property is the old {@link Diagram#viewportBounds} value;
 *     the {@link DiagramEvent#parameter} is also the old viewportBounds {@link Rect}.
 * @unrestricted
 */
export class DiagramEvent {
  /** @hidden @internal */ public static className = 'DiagramEvent';
  /** @hidden @internal */ private _diagram: Diagram;
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _subject: any;
  /** @hidden @internal */ private _parameter: any;

  /**
   * The DiagramEvent class constructor produces an empty DiagramEvent.
   * You should never need to call this constructor.
   * But if you do, you must first set {@link #diagram} to the Diagram before any other use.
   */
  constructor() {
    this._diagram = null;
    this._name = '';
    this._subject = null;
    this._parameter = null;
  }

  /**
   * @hidden @internal
   * Make a copy of this DiagramEvent.
   * @expose
   * @this {DiagramEvent}
   * @return {DiagramEvent}
   */
  public copy(): DiagramEvent {
    const copy = new DiagramEvent();
    copy._diagram = this._diagram;
    copy._name = this._name;
    copy._subject = this._subject;
    copy._parameter = this._parameter;
    return copy;
  }

  /**
   * @hidden @internal
   * @this {DiagramEvent}
   * @return {string}
   */
  public toString(): string {
    let s = '*' + this.name;
    if (this.subject !== null) s += ':' + this.subject.toString();
    if (this.parameter !== null) s += '(' + this.parameter.toString() + ')';
    return s;
  }

  /**
   * Gets the diagram associated with the event.
   */
  /** @export */ get diagram(): Diagram { return this._diagram; }
  /** @export */ set diagram(value: Diagram) { this._diagram = value; }

  /**
   * Gets or sets the name of the kind of diagram event that this represents.
   * This property should always be set to one of the recognized list of names,
   * as listed in the documentation for {@link DiagramEvent}.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) { this._name = value; }

  /**
   * Gets or sets an optional object that is the subject of the diagram event.
   * This property defaults to null.
   */
  /** @export */ get subject(): any { return this._subject; }
  /** @export */ set subject(value: any) { this._subject = value; }

  /**
   * Gets or sets an optional object that describes the change to the subject of the diagram event.
   * This property defaults to null.
   */
  /** @export */ get parameter(): any { return this._parameter; }
  /** @export */ set parameter(value: any) { this._parameter = value; }
} // #end DiagramEvent
DiagramEvent['className'] = 'DiagramEvent';

/**
 * @hidden
 * The signature for a function acting as a handler for DiagramEvents.
 */
export type DiagramEventHandler = (e: DiagramEvent) => void;

/**
 * @hidden
 * (Undocumented, internal interface)
 */
export type DiagramEventName =
  'AnimationStarting' |
  'AnimationFinished' |
  'BackgroundSingleClicked' |
  'BackgroundDoubleClicked' |
  'BackgroundContextClicked' |
  'ChangingSelection' |
  'ChangedSelection' |
  'ClipboardChanged' |
  'ClipboardPasted' |
  'DocumentBoundsChanged' |
  'ExternalObjectsDropped' |
  'GainedFocus' |
  'InitialLayoutCompleted' |
  'LayoutCompleted' |
  'LinkDrawn' |
  'LinkRelinked' |
  'LinkReshaped' |
  'LostFocus' |
  'Modified' |
  'ObjectSingleClicked' |
  'ObjectDoubleClicked' |
  'ObjectContextClicked' |
  'PartCreated' |
  'PartResized' |
  'PartRotated' |
  'SelectionMoved' |
  'SelectionCopied' |
  'SelectionDeleted' |
  'SelectionDeleting' |
  'SelectionGrouped' |
  'SelectionUngrouped' |
  'SubGraphCollapsed' |
  'SubGraphExpanded' |
  'TextEdited' |
  'TreeCollapsed' |
  'TreeExpanded' |
  'ViewportBoundsChanged' |
  'InvalidateDraw'  // undocumented
 ;
/**
 * @hidden
 * (Undocumented, internal interface)
 */
export interface DiagramEventsInterface {
  AnimationStarting?: DiagramEventHandler;
  AnimationFinished?: DiagramEventHandler;
  BackgroundSingleClicked?: DiagramEventHandler;
  BackgroundDoubleClicked?: DiagramEventHandler;
  BackgroundContextClicked?: DiagramEventHandler;
  ChangingSelection?: DiagramEventHandler;
  ChangedSelection?: DiagramEventHandler;
  ClipboardChanged?: DiagramEventHandler;
  ClipboardPasted?: DiagramEventHandler;
  DocumentBoundsChanged?: DiagramEventHandler;
  ExternalObjectsDropped?: DiagramEventHandler;
  GainedFocus?: DiagramEventHandler;
  InitialLayoutCompleted?: DiagramEventHandler;
  LayoutCompleted?: DiagramEventHandler;
  LinkDrawn?: DiagramEventHandler;
  LinkRelinked?: DiagramEventHandler;
  LinkReshaped?: DiagramEventHandler;
  LostFocus?: DiagramEventHandler;
  Modified?: DiagramEventHandler;
  ObjectSingleClicked?: DiagramEventHandler;
  ObjectDoubleClicked?: DiagramEventHandler;
  ObjectContextClicked?: DiagramEventHandler;
  PartCreated?: DiagramEventHandler;
  PartResized?: DiagramEventHandler;
  PartRotated?: DiagramEventHandler;
  SelectionMoved?: DiagramEventHandler;
  SelectionCopied?: DiagramEventHandler;
  SelectionDeleted?: DiagramEventHandler;
  SelectionDeleting?: DiagramEventHandler;
  SelectionGrouped?: DiagramEventHandler;
  SelectionUngrouped?: DiagramEventHandler;
  SubGraphCollapsed?: DiagramEventHandler;
  SubGraphExpanded?: DiagramEventHandler;
  TextEdited?: DiagramEventHandler;
  TreeCollapsed?: DiagramEventHandler;
  TreeExpanded?: DiagramEventHandler;
  ViewportBoundsChanged?: DiagramEventHandler;
 }

// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { Point } from './point';
// import { root } from './root';
// import { Util } from './util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
ChangedEvent
*/

// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Model } from './model';
// import { ObjectData } from './root';
// import { Util } from './util';

/**
 * A ChangedEvent represents a change to an object, typically a {@link GraphObject},
 * but also for model data, a {@link Model}, or a {@link Diagram}.
 * The most common case is for remembering the name of a property
 * and the before-and-after values for that property.
 *
 * You can listen for changed events on the model using {@link Model#addChangedListener}
 * or {@link Diagram#addModelChangedListener}, and on the Diagram using {@link Diagram#addChangedListener}.
 *
 * There are four kinds of changes, represented by enumerated values:
 * {@link ChangedEvent.Property} (the most common), {@link ChangedEvent.Insert} and {@link ChangedEvent.Remove}
 * (to represent inserting or removing objects from collections),
 * and {@link ChangedEvent.Transaction} (to notify about beginning or ending transactions or undo or redo).
 *
 * The most common kind of ChangedEvent is a Property change.
 * The name of the property is given by {@link #propertyName}.
 * The modified object is given by {@link #object}.
 * Use the {@link #oldValue} and {@link #newValue} properties for the before and after property values.
 *
 * For an Insert ChangedEvent, the modified collection (often an Array) is a property value on the {@link #object}.
 * The {@link #propertyName} helps distinguish between different collections on the object.
 * Use the {@link #newValue} property to indicate the value that was inserted.
 * Use the {@link #newParam} property to indicate where or how, such as an array index or dictionary key.
 *
 * For a Remove ChangedEvent, the modified collection is a property value on the {@link #object}.
 * The {@link #propertyName} helps distinguish between different collections on the object.
 * Use the {@link #oldValue} property to indicate the value that was removed.
 * Use the {@link #oldParam} property to indicate where or how, such as an array index or dictionary key.
 *
 * Transaction ChangedEvents are generated by the {@link UndoManager}.
 * The {@link #propertyName} names the nature of the ChangedEvent.
 * For the very first transaction, the property name is "StartingFirstTransaction".
 * This ChangedEvent precedes a ChangedEvent whose property name is "StartedTransaction",
 * which occurs for every top-level transaction.
 *
 * When ending a transaction, there is first a ChangedEvent whose name is "ComittingTransaction".
 * This is followed by one with either "CommittedTransaction" or "RolledBackTransaction",
 * depending on how the transaction is ending.
 * The {@link #oldValue} provides the transaction name and the {@link #object} is the {@link Transaction} being finished.
 * (Note that the Transaction value may be null if no Transaction is available at that time,
 * perhaps because there were no changes made during the transaction.)
 * That Transaction can be scanned to look for ChangedEvents that you may wish to record in a database,
 * all within a single database transaction.
 *
 * There are also Transaction ChangedEvents corresponding to "StartingUndo", "FinishedUndo",
 * "StartingRedo", and "FinishedRedo".
 * The {@link #object} property provides the {@link Transaction} that is about-to-be or just-was undone or redone.
 *
 * Non-Transaction ChangedEvents are remembered by the {@link UndoManager}, if {@link UndoManager#isEnabled},
 * and held in the {@link UndoManager#history} as {@link Transaction}s which hold lists of ChangedEvents.
 * That is why ChangedEvent implements {@link #undo} and {@link #redo} of the change that it remembers.
 *
 * When the ChangedEvent represents a change to a {@link Model}, the value of {@link #model} is non-null
 * and the value of {@link #diagram} is meaningless.
 * If the change is a structural change to the model,
 * the value of {@link #modelChange} indicates the kind of change.
 * Currently defined model changed event names include:
 *   - **"nodeDataArray"**, after the model's {@link Model#nodeDataArray} is replaced, inserted into or removed from
 *     (setting {@link Model#nodeDataArray} or calling {@link Model#addNodeData} or {@link Model#removeNodeData})
 *   - **"nodeKey"**, after changing a node data's unique key ({@link Model#setKeyForNodeData})
 *   - **"nodeCategory"**, after changing a node data's category ({@link Model#setCategoryForNodeData})
 *   - **"linkDataArray"**, after the model's {@link GraphLinksModel#linkDataArray} is replaced, inserted into or removed from
 *     (setting {@link GraphLinksModel#linkDataArray} or calling {@link GraphLinksModel#addLinkData}
 *     or {@link GraphLinksModel#removeLinkData})
 *   - **"linkKey"**, after changing a link data's unique key ({@link GraphLinksModel#setKeyForLinkData})
 *   - **"linkCategory"**, after changing a link data's category ({@link GraphLinksModel#setCategoryForLinkData})
 *   - **"linkFromKey"**, after changing a link data's "from" node key ({@link GraphLinksModel#setFromKeyForLinkData})
 *   - **"linkToKey"**, after changing a link data's "to" node key ({@link GraphLinksModel#setToKeyForLinkData})
 *   - **"linkFromPortId"**, after changing a link data's "from" port identifier string ({@link GraphLinksModel#setFromPortIdForLinkData})
 *   - **"linkToPortId"**, after changing a link data's "to" port identifier string ({@link GraphLinksModel#setToPortIdForLinkData})
 *   - **"linkLabelKeys"**, after replacing, inserting into, or removing from a link data's array of keys to label nodes
 *     (calling {@link GraphLinksModel#setLabelKeysForLinkData}, {@link GraphLinksModel#addLabelKeyForLinkData},
 *     or {@link GraphLinksModel#removeLabelKeyForLinkData})
 *   - **"nodeGroupKey"**, after changing a node data's key for a containing group data ({@link GraphLinksModel#setGroupKeyForNodeData})
 *   - **"nodeParentKey"**, after changing a node data's "parent" node key ({@link TreeModel#setParentKeyForNodeData})
 *   - **"parentLinkCategory"**, after changing a node data's "parent" link's category({@link TreeModel#setParentLinkCategoryForNodeData})
 *   - other names are for internal implementation use only, only on Transaction ChangedEvents
 *
 * The value of {@link ChangedEvent#propertyName} indicates the actual name of the property that was modified.
 * {@link ChangedEvent#modelChange} is a non-empty string only when there is a known structural change to the model,
 * not just the setting of some property on some object.
 *
 * When the ChangedEvent represents a change to a {@link Diagram} or a {@link GraphObject} within a diagram,
 * the value of {@link #diagram} is non-null and
 * the values of {@link #model} and {@link #modelChange} are meaningless.
 *
 * Please note that ChangedEvents can be raised for many different causes.
 * You may not be interested in changes to temporary objects -- in that case ignore the ChangedEvent when
 * {@link Model#skipsUndoManager} or {@link Diagram#skipsUndoManager} is true.
 * @unrestricted
 * @category Model
 */
export class ChangedEvent {
  /** @hidden @internal */ public static className = 'ChangedEvent';
  /** @hidden @internal */ private _change: EnumValue;
  /** @hidden @internal */ private _modelChange: string;
  /** @hidden @internal */ private _propertyName: string | ((a: ObjectData, b: any) => any);
  /** @hidden @internal */ private _model: Model | null;
  /** @hidden @internal */ private _diagram: Diagram | null;
  /** @hidden @internal */ private _object: ObjectData | null;
  /** @hidden @internal */ private _oldValue: any;
  /** @hidden @internal */ private _oldParam: any;
  /** @hidden @internal */ private _newValue: any;
  /** @hidden @internal */ private _newParam: any;

  /**
   * The ChangedEvent class constructor produces an empty ChangedEvent object.
   */
  constructor() {
    this._change = ChangedEvent.Property;
    this._modelChange = '';
    this._propertyName = '';
    this._model = null;
    this._diagram = null;
    this._object = null;
    this._oldValue = null;
    this._oldParam = null;
    this._newValue = null;
    this._newParam = null;
  }

  /**
   * For informational events, such as transactions and undo/redo operations,
   * and used as the value for {@link ChangedEvent#change}.
   * The {@link ChangedEvent#object} refers to the {@link Transaction} affected, if any.
   * The {@link ChangedEvent#propertyName} distinguishes the different transaction or undo or redo stages.
   * The {@link ChangedEvent#oldValue} may provide the transaction name, if available, as given to {@link UndoManager#commitTransaction}.
   * @constant
   */
  public static Transaction = new EnumValue(ChangedEvent, 'Transaction', -1);

  /**
   * For simple property changes,
   * and used as the value for {@link ChangedEvent#change}.
   * The name of the property is given by {@link ChangedEvent#propertyName}.
   * The modified object is given by {@link ChangedEvent#object}.
   * Use the {@link ChangedEvent#oldValue} and {@link ChangedEvent#newValue} properties for the previous and next property values.
   *
   * For model changes, the {@link ChangedEvent#modelChange} may be non-empty, indicating a structural change to the model.
   * @constant
   */
  public static Property = new EnumValue(ChangedEvent, 'Property', 0);

  /**
   * For inserting into collections,
   * and used as the value for {@link ChangedEvent#change}.
   * The modified object is given by {@link ChangedEvent#object}.
   * Use the optional {@link ChangedEvent#propertyName} to distinguish between different collections on the object.
   * Use the {@link ChangedEvent#newValue} property to indicate the value that was inserted.
   * Use the optional {@link ChangedEvent#newParam} property to indicate where or how, such as an array index or dictionary key.
   * @constant
   */
  public static Insert = new EnumValue(ChangedEvent, 'Insert', 1);

  /**
   * For removing from collections,
   * and used as the value for {@link ChangedEvent#change}.
   * The modified object is given by {@link ChangedEvent#object}.
   * Use the optional {@link ChangedEvent#propertyName} to distinguish between different collections on the object.
   * Use the {@link ChangedEvent#oldValue} property to indicate the value that was removed.
   * Use the optional {@link ChangedEvent#oldParam} property to indicate where or how, such as an array index or dictionary key.
   * @constant
   */
  public static Remove = new EnumValue(ChangedEvent, 'Remove', 2);

  /**
   * Forget any object references that this ChangedEvent may have.
   * @export
   * @this {ChangedEvent}
   */
  public clear(): void {
    this._model = null;
    this._diagram = null;
    this._object = null;
    this._oldValue = null;
    this._oldParam = null;
    this._newValue = null;
    this._newParam = null;
  }

  /**
   * Make a copy of this ChangedEvent.
   * ChangedEvents are copied when the {@link UndoManager} adds to a {@link Transaction}.
   * @this {ChangedEvent}
   * @return {ChangedEvent}
   */
  public copy(): ChangedEvent {
    const copy = new ChangedEvent();
    copy._change = this._change;
    copy._modelChange = this._modelChange;
    copy._propertyName = this._propertyName;
    copy._model = this._model;
    copy._diagram = this._diagram;
    copy._object = this._object;
    let v = this._oldValue;
    if (Util.isObject(v) && typeof v.copyFrozen === 'function') {
      copy._oldValue = v.copyFrozen();
    } else {
      copy._oldValue = v;
    }
    v = this._oldParam;
    if (Util.isObject(v) && typeof v.copyFrozen === 'function') {
      copy._oldParam = v.copyFrozen();
    } else {
      copy._oldParam = v;
    }
    v = this._newValue;
    if (Util.isObject(v) && typeof v.copyFrozen === 'function') {
      copy._newValue = v.copyFrozen();
    } else {
      copy._newValue = v;
    }
    v = this._newParam;
    if (Util.isObject(v) && typeof v.copyFrozen === 'function') {
      copy._newParam = v.copyFrozen();
    } else {
      copy._newParam = v;
    }
    return copy;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {ChangedEvent}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === ChangedEvent) {
      this.change = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {ChangedEvent}
   * @return {string}
   */
  public toString(): string {
    let s = '';
    if (this.change === ChangedEvent.Transaction) {
      s += '* ';
    } else if (this.change === ChangedEvent.Property) {
      s += (this.model !== null ? '!m' : '!d');
    } else {
      s += (this.model !== null ? '!m' : '!d') + this.change;
    }
    if (this.propertyName && typeof this.propertyName === 'string') s += ' ' + this.propertyName;
    if (this.modelChange && this.modelChange !== this.propertyName) s += ' ' + this.modelChange;
    s += ': ';
    if (this.change === ChangedEvent.Transaction) {
      if (this.oldValue !== null) s += ' ' + this.oldValue;
    } else {
      if (this.object !== null) s += Util.toString(this.object);
      if (this.oldValue !== null) s += '  old: ' + Util.toString(this.oldValue);
      if (this.oldParam !== null) s += ' ' + this.oldParam;
      if (this.newValue !== null) s += '  new: ' + Util.toString(this.newValue);
      if (this.newParam !== null) s += ' ' + this.newParam;
    }
    return s;
  }

  /**
   * This is a convenient method to get the right value, depending on the value of undo,
   * when implementing a state change as part of an undo or a redo.
   * @export
   * @this {ChangedEvent}
   * @param {boolean} undo If true, returns the oldValue, otherwise returns the newValue.
   * @return {*} Either the oldValue or the newValue.
   */
  public getValue(undo: boolean): any {
    if (undo) {
      return this.oldValue;
    } else {
      return this.newValue;
    }
  }

  /**
   * This is a convenient method to get the right parameter value, depending on the value of undo,
   * when implementing a state change as part of an undo or a redo.
   * @export
   * @this {ChangedEvent}
   * @param {boolean} undo If true, returns the oldParam, otherwise returns the newParam.
   * @return {*} Either the oldParam or the newParam.
   */
  public getParam(undo: boolean): any {
    if (undo) {
      return this.oldParam;
    } else {
      return this.newParam;
    }
  }

  /**
   * This predicate returns true if you can call undo().
   * @export
   * @this {ChangedEvent}
   * @return {boolean} True if ready for undo() to be called.
   */
  public canUndo(): boolean {
    if (this.model !== null) return true;
    if (this.diagram !== null) return true;
    return false;
  }

  /**
   * Reverse the effects of this object change.
   * canUndo() must be true for this method to have any effect.
   * @export
   * @this {ChangedEvent}
   */
  public undo(): void {
    if (this.canUndo()) {
      if (this.model !== null) this.model.changeState(this, true);
      else if (this.diagram !== null) this.diagram.changeState(this, true);
    }
  }

  /**
   * This predicate returns true if you can call redo().
   * @export
   * @this {ChangedEvent}
   * @return {boolean} True if ready for redo() to be called.
   */
  public canRedo(): boolean {
    if (this.model !== null) return true;
    if (this.diagram !== null) return true;
    return false;
  }

  /**
   * Re-perform this object change after an undo().
   * canRedo() must be true for this method to have any effect.
   * @export
   * @this {ChangedEvent}
   */
  public redo(): void {
    if (this.canRedo()) {
      if (this.model !== null) this.model.changeState(this, false);
      else if (this.diagram !== null) this.diagram.changeState(this, false);
    }
  }

  /**
   * Gets or sets the {@link Model} or {@link TreeModel} or {@link GraphLinksModel} that was modified.
   * When this property is non-null, the {@link #diagram} property will be null.
   * However this property and the {@link #diagram} property may both be null simultaneously,
   * when no particular model or diagram applies.
   */
  /** @export */ get model(): Model | null { return this._model; }
  /** @export */ set model(value: Model | null) { this._model = value; }

  /**
   * Gets or sets the {@link Diagram} that was modified.
   * When this property is non-null, the {@link #model} property will be null.
   * However this property and the {@link #model} property may both be null simultaneously,
   * when no particular model or diagram applies.
   */
  /** @export */ get diagram(): Diagram | null { return this._diagram; }
  /** @export */ set diagram(value: Diagram | null) { this._diagram = value; }

  /**
   * Gets or sets the nature of change that occurred.
   * The default is {@link ChangedEvent.Property}.
   * Other values are {@link ChangedEvent.Insert}, {@link ChangedEvent.Remove},
   * and {@link ChangedEvent.Transaction}.
   */
  /** @export */ get change(): EnumValue { return this._change; }
  /** @export */ set change(value: EnumValue) {
    if (Debug) EnumValue.checkEnumValue(value, ChangedEvent, ChangedEvent, 'change');
    this._change = value;
  }

  /**
   * Gets or sets the name of the model change, reflecting a change to
   * model data in addition to a change to the model itself.
   *
   * The default is an empty string, which indicates that this is just
   * a regular change to some object's state, probably its property.
   * For a list of possible model change names, see the documentation for {@link ChangedEvent}.
   * The names are compared in a case-sensitive manner.
   */
  /** @export */ get modelChange(): string { return this._modelChange; }
  /** @export */ set modelChange(value: string) {
    if (Debug) Util.checkPrimitive(value, 'string', ChangedEvent, 'modelChange');
    this._modelChange = value;
  }

  /**
   * Gets or sets the name of the property change.
   * The default is an empty string, which is not a valid property name.
   * This property can be useful even when the type of change is
   * not {@link ChangedEvent.Property}, because it can help identify
   * the collection in the {@link #object} that was modified
   * (for {@link ChangedEvent.Insert} or {@link ChangedEvent.Remove})
   * or the stage of the current transaction (for {@link ChangedEvent.Transaction}).
   */
  /** @export */ get propertyName(): string | ((a: ObjectData, b: any) => any) { return this._propertyName; }
  /** @export */ set propertyName(value: string | ((a: ObjectData, b: any) => any)) {
    if (Debug && typeof value !== 'string') Util.checkPrimitive(value, 'function', ChangedEvent, 'propertyName');
    this._propertyName = value;
  }

  /**
   * This read-only property is true when this ChangedEvent is of type {@link ChangedEvent.Transaction} and represents the end of a transactional change.
   * It is implemented as:
   * ```js
   * return (this.change === ChangedEvent.Transaction &&
   *         (this.propertyName === "CommittedTransaction" ||
   *          this.propertyName === "FinishedUndo" ||
   *          this.propertyName === "FinishedRedo"));
   * ```
   */
  /** @export */ get isTransactionFinished(): boolean {
    return (this._change === ChangedEvent.Transaction &&
      (this._propertyName === 'CommittedTransaction' || this._propertyName === 'FinishedUndo' || this._propertyName === 'FinishedRedo'));
  }

  /**
   * Gets or sets the Object that was modified.
   * The default is null.
   *
   * For {@link ChangedEvent.Transaction} changes, this may be the {@link Transaction}.
   */
  /** @export */ get object(): ObjectData | null { return this._object; }
  /** @export */ set object(value: ObjectData | null) { this._object = value; }

  /**
   * Gets or sets the previous or old value that the property had.
   * The default is null.
   */
  /** @export */ get oldValue(): any { return this._oldValue; }
  /** @export */ set oldValue(value: any) { this._oldValue = value; }

  /**
   * Gets or sets an optional value associated with the old value.
   * Most properties do not require any parameter to describe the change.
   * This is typically a value that helps distinguish the old value, such as an index into an array.
   * It is null if it is not used.
   * The default is null.
   */
  /** @export */ get oldParam(): any { return this._oldParam; }
  /** @export */ set oldParam(value: any) { this._oldParam = value; }

  /**
   * Gets or sets the next or current value that the property has.
   * The default is null.
   */
  /** @export */ get newValue(): any { return this._newValue; }
  /** @export */ set newValue(value: any) { this._newValue = value; }

  /**
   * Gets or sets an optional value associated with the new value.
   * Most properties do not require any parameter to describe the change.
   * This is typically a value that helps distinguish the new value, such as an index into an array.
   * It is null if it is not used.
   * The default is null.
   */
  /** @export */ get newParam(): any { return this._newParam; }
  /** @export */ set newParam(value: any) { this._newParam = value; }

} // end class
ChangedEvent['className'] = 'ChangedEvent';

/**
 * @hidden
 * The signature for a function acting as a handler for ChangedEvents.
 */
export type ChangedEventHandler = (e: ChangedEvent) => void;

ChangedEvent['Transaction'] = ChangedEvent.Transaction;
ChangedEvent['Property'] = ChangedEvent.Property;
ChangedEvent['Insert'] = ChangedEvent.Insert;
ChangedEvent['Remove'] = ChangedEvent.Remove;

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Transaction
UndoManager
*/



/**
 * A Transaction holds a list of {@link ChangedEvent}s collected during a transaction,
 * as the value of the read-only {@link #changes} property.
 *
 * Start a transaction by calling {@link UndoManager#startTransaction}
 * (or {@link Model#startTransaction} or {@link Diagram#startTransaction}, which call that method).
 * Be sure to finish a transaction with a matching call to {@link UndoManager#commitTransaction}
 * (or {@link Model#commitTransaction} or {@link Diagram#commitTransaction}),
 * or a call to {@link UndoManager#rollbackTransaction}
 * (or the same named methods on {@link Model} or {@link Diagram}).
 *
 * If you are performing multiple or repeated changes to a model or diagram,
 * surround all of the code with calls to start and commit the transaction --
 * do not perform repeated calls to start-commit-start-commit.
 * Typically each change caused by the user, such as a button click or a change of focus or a mouse drag,
 * should perform one transaction in which all changes are made.
 * All of the predefined commands and tools perform transactions.
 *
 * Undoing or redoing a transaction is done by calling {@link UndoManager#undo} or {@link UndoManager#redo}.
 * Those methods call the {@link #undo} or {@link #redo} methods here.
 *
 * The {@link UndoManager} holds a list of Transactions in its {@link UndoManager#history}.
 * @unrestricted
 * @category Model
 */
export class Transaction {
  /** @hidden @internal */ public static className = 'Transaction';
  /** @hidden @internal */ private _changes: List<ChangedEvent>;
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _isComplete: boolean;

  /**
   * Construct an object holding an empty list of {@link ChangedEvent}s
   * and no {@link #name}.
   */
  constructor() {
    this._changes = new List<ChangedEvent>().freeze();
    this._name = '';
    this._isComplete = false;
  }

  /**
   * @hidden @internal
   * @this {Transaction}
   * @param {number=} details Optional: If greater than zero, this also includes information about each {@link ChangedEvent}.
   * @return {string}
   */
  public toString(details?: number): string {
    let s = 'Transaction: ' + this.name + ' ' + this.changes.count.toString() + (this.isComplete ? '' : ', incomplete');
    if (details !== undefined && details > 0) {
      const len = this.changes.count;
      for (let i = 0; i < len; i++) {
        const c = this.changes.elt(i);
        if (c !== null) s += '\n  ' + c.toString();
      }
    }
    return s;
  }

  /**
   * Clear all of the saved changes.
   * @export
   * @this {Transaction}
   */
  public clear(): void {
    const chs = this.changes;
    chs.thaw();
    for (let i = chs.count - 1; i >= 0; i--) {
      const c = chs.elt(i);
      if (c !== null) c.clear();
    }
    chs.clear();
    chs.freeze();
  }

  /**
   * This predicate returns true if you can call {@link #undo},
   * namely when {@link #isComplete} is true.
   * @export
   * @this {Transaction}
   * @return {boolean} true if ready for {@link #undo} to be called.
   */
  public canUndo(): boolean {
    return this.isComplete;
  }

  /**
   * Undo all of the changes, in reverse order.
   * This is normally called only by the {@link UndoManager}.
   * {@link #canUndo} must be true for this method to have any effect.
   * @export
   * @this {Transaction}
   */
  public undo(): void {
    if (!this.canUndo()) return;
    for (let i = this.changes.count - 1; i >= 0; i--) {
      const c = this.changes.elt(i);
      if (c !== null) c.undo();
    }
  }

  /**
   * This predicate returns true if you can call {@link #redo},
   * namely when {@link #isComplete} is true.
   * @export
   * @this {Transaction}
   * @return {boolean} true if ready for {@link #redo} to be called.
   */
  public canRedo(): boolean {
    return this.isComplete;
  }

  /**
   * Re-perform these changes after an {@link #undo}.
   * This is normally called only by the {@link UndoManager}.
   * {@link #canRedo} must be true for this method to have any effect.
   * @export
   * @this {Transaction}
   */
  public redo(): void {
    if (!this.canRedo()) return;
    const len = this.changes.count;
    for (let i = 0; i < len; i++) {
      const c = this.changes.elt(i);
      if (c !== null) c.redo();
    }
  }

  /**
   * @hidden @internal
   * Partial undo, to implement rollback for nested rollbackTransaction() calls.
   * @this {Transaction}
   * @param {number} start
   */
  public rollbackTo(start: number): void {
    const chs = this.changes;
    for (let i = chs.count - 1; i >= start; i--) {
      const c = chs.elt(i);
      if (c !== null) c.undo();
      chs.thaw();  // each undo might freeze the list of changes
      chs.removeAt(i);
    }
    chs.freeze();
  }

  /**
   * This read-only property returns the list of {@link ChangedEvent}s.
   * The changes are stored in order of occurrence.
   *
   * You should not modify this list.
   */
  /** @export */ get changes(): List<ChangedEvent> { return this._changes; }

  /**
   * Gets or sets the transaction name for this collection of changes.
   * This is set by a top-level call to {@link UndoManager#commitTransaction}.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) { this._name = value; }

  /**
   * Gets or sets whether we can add more {@link ChangedEvent}s to this list of changes.
   * This is initially false.
   * It is set to true by {@link UndoManager#commitTransaction} and {@link UndoManager#rollbackTransaction}.
   */
  /** @export */ get isComplete(): boolean { return this._isComplete; }
  /** @export */ set isComplete(value: boolean) { this._isComplete = value; }
} // #end Transaction
Transaction['className'] = 'Transaction';


/**
 * An UndoManager observes and records model and diagram changes in transactions and
 * supports undo/redo operations.
 * You will need to set the {@link #isEnabled} property to true in order for
 * the UndoManager to record changes and for users to perform an undo or a redo.
 *
 * Typically an operation will call {@link #startTransaction},
 * make some changes to the {@link Model} and/or {@link Diagram},
 * and then call {@link #commitTransaction}.
 * Any {@link ChangedEvent}s that occur will be recorded in a
 * {@link Transaction} object.
 * If for some reason you do not wish to complete the transaction
 * successfully, you can call {@link #rollbackTransaction} instead
 * of {@link #commitTransaction}.
 *
 * For convenience the {@link Diagram#commit} and {@link Model#commit} methods
 * execute a function within a transaction and then perform a commit,
 * or else a rollback upon an error.
 *
 * The {@link #history} property is a list of {@link Transaction}s.
 * {@link #commitTransaction} will add the {@link #currentTransaction}
 * to the {@link #history} list.
 * {@link #rollbackTransaction} will undo the changes remembered in the
 * {@link #currentTransaction} and then discard it, without changing the {@link #history}.
 * You can limit how many transactions are remembered in the history
 * by setting {@link #maxHistoryLength}.
 *
 * Transactions may be nested.
 * Be sure to call either {@link #commitTransaction} or {@link #rollbackTransaction}
 * for each call to {@link #startTransaction}.
 * Avoid repeated start-commit-start-commit calls as a result of a user's actions.
 * Instead, start, make all changes, and then commit.
 *
 * If you want to restore the diagram to the state before the latest complete
 * transaction, call {@link #undo}.
 * Call {@link #redo} to change the diagram to a later state.
 * If after some number of undo's you start a transaction,
 * all of the history after the current state is discarded,
 * and a new transaction may be recorded.
 * You cannot undo or redo during a transaction.
 *
 * Initially each {@link Model} has its own UndoManager.
 * UndoManagers may be shared by multiple Models by replacing
 * the standard {@link Model#undoManager} created by the model constructor.
 *
 * There are several informational properties:
 *   - {@link #isInTransaction} is true when a top-level transaction has been started that has not yet been committed or rolled-back.
 *   - {@link #currentTransaction} holds the flattened list of all {@link ChangedEvent}s
 *     that have happened within the current transaction.
 *   - {@link #transactionLevel} indicates the current depth of nesting.
 *   - {@link #nestedTransactionNames} holds the stack of transaction names supplied to
 *     {@link #startTransaction} calls.
 *   - {@link #history} holds only complete top-level transactions.
 *   - {@link #isUndoingRedoing} is true during a call to {@link #undo} or {@link #redo}.
 *   - {@link #historyIndex} indicates which {@link Transaction} in the {@link #history} is the next to be "undone";
 *     this is decremented by each undo and incremented by each redo.
 *   - {@link #transactionToUndo} and {@link #transactionToRedo} indicate which {@link Transaction} may be undone or redone next, if any.
 *   - {@link #models} returns an iterator over all of the {@link Model}s that this UndoManager is handling.
 *
 * A transaction may not be ongoing when replacing a {@link Diagram#model},
 * because it would not make sense to be replacing the UndoManager (the {@link Model#undoManager})
 * while changes are being recorded.
 *
 * Replacing a {@link Diagram#model} copies certain properties from the old UndoManager to the new one,
 * including {@link #isEnabled} and {@link #maxHistoryLength}.
 * @unrestricted
 * @category Model
 */
export class UndoManager {
  /** @hidden @internal */ public static className = 'UndoManager';
  /** @hidden @internal */ private _models: Set<Model>;
  /** @hidden @internal */ private _isEnabled: boolean;
  /** @hidden @internal */ private _history: List<Transaction>;
  /** @hidden @internal */ private _historyIndex: number;
  /** @hidden @internal */ private _maxHistoryLength: number;
  /** @hidden @internal */ private _currentTransaction: Transaction | null;
  /** @hidden @internal */ private _transactionLevel: number;
  /** @hidden @internal */ private _checksTransactionLevel: boolean;
  /** @hidden @internal */ private _nestedTransactionNames: List<string>;
  /** @hidden @internal */ private _nestedTransactionStarts: List<number>;
  /** @hidden @internal */ private _firstTransaction: boolean;
  /** @hidden @internal */ private _isUndoingRedoing: boolean;
  /** @hidden @internal */ public _justDiscarded: boolean;

  /**
   * The constructor produces an empty UndoManager
   * with no transaction history.
   */
  constructor() {
    this._models = new Set<Model>();
    this._isEnabled = false;
    this._history = new List<Transaction>().freeze();
    this._historyIndex = -1;
    this._maxHistoryLength = 999;
    this._isUndoingRedoing = false;
    this._currentTransaction = null;
    this._transactionLevel = 0;
    this._checksTransactionLevel = false;
    if (Debug) this._checksTransactionLevel = true;
    this._nestedTransactionNames = new List<string>().freeze();
    this._nestedTransactionStarts = new List<number>();
    this._firstTransaction = true;
    this._justDiscarded = false;
  }

  /**
   * @hidden @internal
   * @this {UndoManager}
   * @param {number=} details Optional: If greater than zero, this includes descriptions of each {@link Transaction} in its {@link #history}.
   * @return {string}
   */
  public toString(details?: number): string {
    let s = 'UndoManager ' + this.historyIndex + '<' +
      this.history.count + '<=' + this.maxHistoryLength;
    s += '[';
    let len = this.nestedTransactionNames.count;
    for (let i = 0; i < len; i++) {
      if (i > 0) s += ' ';
      s += this.nestedTransactionNames.elt(i);
    }
    s += ']';
    if (details !== undefined && details > 0) {
      len = this.history.count;
      for (let i = 0; i < len; i++) {
        s += '\n ' + this.history.elt(i).toString(details - 1);
      }
    }
    return s;
  }

  /**
   * Clear all of the {@link Transaction}s and clear all other state,
   * including any ongoing transaction without rolling back.
   * However, this maintains its references to its {@link Model}s.
   *
   * You should not call this method during a transaction.
   * @export
   * @this {UndoManager}
   */
  public clear(): void {
    const hist = this.history;
    hist.thaw();
    for (let i = hist.count - 1; i >= 0; i--) {
      const c = hist.elt(i);
      if (c !== null) c.clear();
    }
    hist.clear();
    this._historyIndex = -1;
    hist.freeze();
    this._isUndoingRedoing = false;
    this._currentTransaction = null;
    this._transactionLevel = 0;
    this._nestedTransactionNames.thaw();
    this._nestedTransactionNames.clear();
    this._nestedTransactionNames.freeze();
    this._nestedTransactionStarts.clear();
  }

  /**
   * @hidden
   * @expose
   * Undocumented.
   * Copy persistent properties from an old UndoManager to this new one.
   * This is called by the {@link Diagram#model} property setter.
   * @this {UndoManager}
   * @param old
   */
  public copyProperties(old: UndoManager) {
    this.isEnabled = old.isEnabled;
    this.maxHistoryLength = old.maxHistoryLength;
    this.checksTransactionLevel = old.checksTransactionLevel;
  }

  /**
   * Make sure this UndoManager knows about a {@link Model} for which
   * it may receive {@link ChangedEvent}s when the given {@link Model} is changed.
   * The model will also receive notifications about transactions and undo or redo operations.
   *
   * You should not call this method during a transaction.
   * @export
   * @this {UndoManager}
   * @param {Model} model A {@link Model} that this UndoManager is managing.
   * @see {@link #models}
   * @see {@link #removeModel}
   */
  public addModel(model: Model): void {
    this._models.add(model);
  }

  /**
   * Inform this UndoManager that it will no longer be receiving {@link ChangedEvent}s
   * when the given {@link Model} is changed.
   * The model will no longer receive notifications about transactions and undo or redo operations.
   *
   * You should not call this method during a transaction.
   * If you call this method between transactions when there is a transaction history,
   * you should be careful that there are no {@link ChangedEvent}s referring to that model in any {@link Transaction}s.
   * @export
   * @this {UndoManager}
   * @param {Model} model A {@link Model} that this UndoManager should no longer manage.
   * @see {@link #models}
   * @see {@link #addModel}
   */
  public removeModel(model: Model): void {
    this._models.remove(model);
  }

  /**
   * Begin a transaction, where the changes are held by a {@link Transaction} object
   * as the value of {@link #currentTransaction}.
   * You must call either {@link #commitTransaction} or {@link #rollbackTransaction} afterwards.
   *
   * For convenience, this method is called by {@link Model#startTransaction} and {@link Diagram#startTransaction}.
   *
   * Transactions can be nested.
   * Starting or ending a nested transaction will return false.
   * Nested transactions will share the same {@link Transaction} list of {@link ChangedEvent}s.
   *
   * Starting a transaction will not necessarily cause {@link #currentTransaction} to be non-null.
   * A {@link Transaction} object is usually only created by {@link #handleChanged} when a {@link ChangedEvent} first occurs.
   * @export
   * @this {UndoManager}
   * @param {string=} tname a short string describing the transaction, pushed onto the {@link #nestedTransactionNames} stack.
   * @return {boolean} true if starting a top-level transaction.
   * @see {@link #commitTransaction}
   * @see {@link #rollbackTransaction}
   */
  public startTransaction(tname?: string): boolean {
    if (tname === undefined) tname = '';
    if (tname === null) tname = '';
    if (this.isUndoingRedoing) return false;
    if (this._firstTransaction === true) {
      this._firstTransaction = false;
      // do initial layout, pretend we are within a transaction, even if isEnabled is false
      this._transactionLevel++;
      this._raiseChanged('StartingFirstTransaction', tname, this.currentTransaction);
      if (this._transactionLevel > 0) this._transactionLevel--;
    }
    if (this.isEnabled) {
      this._nestedTransactionNames.thaw();
      this._nestedTransactionNames.add(tname);
      this._nestedTransactionNames.freeze();
      if (this.currentTransaction === null) {
        this._nestedTransactionStarts.add(0);
      } else {
        this._nestedTransactionStarts.add(this.currentTransaction.changes.count);
      }
    }
    this._transactionLevel++;
    const result = (this.transactionLevel === 1);
    if (result) {
      this._raiseChanged('StartedTransaction', tname, this.currentTransaction);
    }
    return result;
  }

  /**
   * Commit the current transaction started by a call to {@link #startTransaction}.
   *
   * For convenience, this method is called by {@link Model#commitTransaction} and {@link Diagram#commitTransaction}.
   *
   * If this call stops a top-level transaction,
   * we mark the {@link #currentTransaction} as complete ({@link Transaction#isComplete}),
   * we add the {@link Transaction} to the {@link #history} list,
   * and we return true.
   * Committing a transaction when there have been some undos without corresponding
   * redos will throw away the {@link Transaction}s holding changes that happened
   * after the current state, before adding the new {@link Transaction} to the
   * {@link #history} list.
   * @export
   * @this {UndoManager}
   * @param {string=} tname a short string describing the transaction;
   *        this is recorded as the {@link Transaction#name} and need not be the same as the string passed to {@link #startTransaction}.
   *        If the value is an empty string or not supplied, this will use the name given to {@link #startTransaction}.
   * @return {boolean} true if ending a top-level transaction.
   * @see {@link #startTransaction}
   * @see {@link #rollbackTransaction}
   */
  public commitTransaction(tname?: string): boolean {
    if (tname === undefined) tname = '';
    return this._endTransaction(true, tname);
  }

  /**
   * Rollback the current transaction started by a call to {@link #startTransaction}, undoing any changes.
   *
   * For convenience, this method is called by {@link Model#rollbackTransaction} and {@link Diagram#rollbackTransaction}.
   *
   * This undoes and then discards the changes in the {@link #currentTransaction}.
   * You must have started a transaction previously.
   * @export
   * @this {UndoManager}
   * @return {boolean} true if ending a top-level transaction.
   * @see {@link #startTransaction}
   * @see {@link #commitTransaction}
   */
  public rollbackTransaction(): boolean {
    return this._endTransaction(false, '');
  }

  /**
   * @hidden @internal
   * @param {boolean} commit
   * @param {string} tname a short string describing the transaction.
   * @return {boolean}
   */
  private _endTransaction(commit: boolean, tname: string): boolean {
    if (this.isUndoingRedoing) return false;
    if (this.checksTransactionLevel && this.transactionLevel < 1) {
      Util.trace('Ending transaction without having started a transaction: ' + tname);
    }
    const toplevel = (this.transactionLevel === 1);
    // before decrementing the transaction level, notify diagrams so that they can update,
    // especially any invalid layouts
    if (toplevel && commit) {
      this._raiseChanged('CommittingTransaction', tname, this.currentTransaction);
    }
    let start = 0;
    // decrement the transaction level, but not below zero
    if (this.transactionLevel > 0) {
      this._transactionLevel--;
      if (this.isEnabled) {
        const numnames = this._nestedTransactionNames.count;
        if (numnames > 0) {
          if (tname === '') tname = this._nestedTransactionNames.elt(0);
          this._nestedTransactionNames.thaw();
          this._nestedTransactionNames.removeAt(numnames - 1);
          this._nestedTransactionNames.freeze();
        }
        const numstarts = this._nestedTransactionStarts.count;
        if (numstarts > 0) {
          start = this._nestedTransactionStarts.elt(numstarts - 1);
          this._nestedTransactionStarts.removeAt(numstarts - 1);
        }
      }
    }
    let curr = this.currentTransaction;
    if (toplevel) {
      if (commit) {
        this._justDiscarded = false;
        if (this.isEnabled) {
          // finish the current transaction
          if (curr !== null) {
            const cedit = curr;
            cedit.isComplete = true;
            cedit.name = tname;
            const hist = this.history;
            hist.thaw();
            // throw away any compound edits following the current index
            for (let i = hist.count - 1; i > this.historyIndex; i--) {
              const t = hist.elt(i);
              if (t !== null) t.clear();
              hist.removeAt(i);
              this._justDiscarded = true;
            }
            // if there is a limit, maybe throw away the oldest transaction(s)
            const max = this.maxHistoryLength;
            if (max >= 0) {  // < 0 means no limit
              if (max === 0) {
                hist.clear();
              } else if (hist.count >= max) {  // prune history
                const t = hist.elt(0);
                if (t !== null) t.clear();
                hist.removeAt(0);
                this._historyIndex--;
              }
            }
            if (max !== 0) {  // == 0 means don't remember Transaction
              // add to History list
              hist.add(cedit);
              this._historyIndex++;
            }
            hist.freeze();
            curr = cedit;
          }
        }
        this._raiseChanged('CommittedTransaction', tname, curr);
      } else {  // !commit
        this._isUndoingRedoing = true;
        try {
          if (this.isEnabled && curr !== null) {
            // rollback the current transaction by undoing all of its changes
            curr.isComplete = true;
            curr.undo();
          }
        } finally {
          this._raiseChanged('RolledBackTransaction', tname, curr);
          this._isUndoingRedoing = false;
        }
        // now we can throw away all those undone changes
        if (curr !== null) curr.clear();
      }
      this._currentTransaction = null;
      return true;
    } else {  // !toplevel
      if (this.isEnabled) {
        // rollback the current transaction by undoing all of its edits,
        // but not of parent transactions
        if (!commit && curr !== null) {
          curr.rollbackTo(start);
        }
      }
      return false;
    }
  }

  /**
   * This predicate returns true if you can call {@link #undo}.
   * This will return false if {@link #isEnabled} is false (as it is by default),
   * if any transaction is ongoing, or
   * if there is no {@link #transactionToUndo} that can be undone.
   * @export
   * @this {UndoManager}
   * @return {boolean} true if ready for {@link #undo} to be called.
   */
  public canUndo(): boolean {
    if (!this.isEnabled) return false;
    if (this.transactionLevel > 0) return false;
    const curr = this.transactionToUndo;
    if (curr !== null && curr.canUndo()) return true;
    return false;
  }

  /**
   * Reverse the effects of the {@link #transactionToUndo}.
   * {@link #canUndo} must be true for this method to have any effect.
   *
   * This is called by {@link CommandHandler#undo}.
   *
   * This will raise a "StartingUndo" {@link ChangedEvent} of type {@link ChangedEvent.Transaction},
   * perform the {@link Transaction#undo} on the {@link #transactionToUndo}, and then
   * raise a "FinishedUndo" {@link ChangedEvent} of type {@link ChangedEvent.Transaction}.
   * The two ChangedEvents are to let model listeners know that an undo is about to take place
   * and that it just finished.
   * {@link #isUndoingRedoing} will temporarily be set to true during this operation.
   * @export
   * @this {UndoManager}
   */
  public undo(): void {
    if (!this.canUndo()) return;
    const tx = this.transactionToUndo;
    try {
      this._isUndoingRedoing = true;
      this._raiseChanged('StartingUndo', 'Undo', tx);
      this._historyIndex--;
      tx.undo();
    } catch (ex) {
      Util.trace('undo error: ' + ex.toString());
    } finally {
      this._raiseChanged('FinishedUndo', 'Undo', tx);
      this._isUndoingRedoing = false;
    }
  }

  /**
   * This predicate returns true if you can call {@link #redo}.
   * This will return false if {@link #isEnabled} is false (as it is by default),
   * if any transaction is ongoing, or
   * if there is no {@link #transactionToRedo} that can be redone.
   * @this {UndoManager}
   * @return {boolean} true if ready for {@link #redo} to be called.
   */
  public canRedo(): boolean {
    if (!this.isEnabled) return false;
    if (this.transactionLevel > 0) return false;
    const curr = this.transactionToRedo;
    if (curr !== null && curr.canRedo()) return true;
    return false;
  }

  /**
   * After an {@link #undo}, re-perform the changes in {@link #transactionToRedo}.
   * {@link #canRedo} must be true for this method to have any effect.
   *
   * This is called by {@link CommandHandler#redo}.
   *
   * This will raise a "StartingRedo" {@link ChangedEvent} of type {@link ChangedEvent.Transaction},
   * perform the {@link Transaction#redo} on the {@link #transactionToRedo}, and then
   * raise a "FinishedRedo" {@link ChangedEvent} of type {@link ChangedEvent.Transaction}.
   * The two ChangedEvents are to let model listeners know that a redo is about to take place
   * and that it just finished.
   * {@link #isUndoingRedoing} will temporarily be set to true during this operation.
   * @export
   * @this {UndoManager}
   */
  public redo(): void {
    if (!this.canRedo()) return;
    const tx = this.transactionToRedo;
    try {
      this._isUndoingRedoing = true;
      this._raiseChanged('StartingRedo', 'Redo', tx);
      this._historyIndex++;
      tx.redo();
    } catch (ex) {
      Util.trace('redo error: ' + ex.toString());
    } finally {
      this._raiseChanged('FinishedRedo', 'Redo', tx);
      this._isUndoingRedoing = false;
    }
  }

  /**
   * @hidden @internal
   * An internal method for raising Transaction ChangedEvents.
   * @this {UndoManager}
   * @param {string} ch the transaction type.
   * @param {string} tname the Transaction name, or "Undo" or "Redo", or the property name for Property changes.
   * @param {Transaction=} val the Transaction itself, if available.
   */
  private _raiseChanged(ch: string, tname: string, val?: Transaction): void {
    if (val === undefined) val = null;
    const e = new ChangedEvent();
    e.change = ChangedEvent.Transaction;
    e.propertyName = ch;
    e.object = val;  // might be null
    e.oldValue = tname;
    const mit = this.models;
    while (mit.next()) {
      const m = mit.value;
      e.model = m;
      m.callChangedListeners(e);
    }
  }

  /**
   * Maybe record a {@link ChangedEvent} in the {@link #currentTransaction}.
   * This calls {@link #skipsEvent} to see if this should ignore the change.
   * If {@link #skipsEvent} returns false, this creates a copy of the {@link ChangedEvent}
   * and adds it to the {@link #currentTransaction}.
   * If there is no {@link #currentTransaction}, this first creates and remembers it.
   *
   * This method always ignores all changes while performing
   * an {@link #undo} or {@link #redo}.
   * This method is also a no-op if {@link #isEnabled} is false.
   * @export
   * @this {UndoManager}
   * @param {ChangedEvent} e a {@link ChangedEvent}.
   */
  public handleChanged(e: ChangedEvent): void {
    if (!this.isEnabled) return;
    if (this.isUndoingRedoing) return;
    if (!this.skipsEvent(e)) {
      let curr = this.currentTransaction;
      if (curr === null) {
        curr = new Transaction();
        this._currentTransaction = curr;
      }
      const copy = e.copy();
      const chs = curr.changes;
      chs.thaw();
      chs.add(copy);
      chs.freeze();
      if (this.checksTransactionLevel && this.transactionLevel <= 0 && !this._firstTransaction) {
        const diagram = e.diagram;
        if (diagram !== null && diagram.isInitialLayoutCompleted === false) return;
        Util.trace('Change not within a transaction: ' + copy.toString());
      }
    }
  }

  /**
   * This predicate is called by {@link #handleChanged} to decide if a {@link ChangedEvent}
   * is not interesting enough to be remembered.
   *
   * Transactional events (of change type {@link ChangedEvent.Transaction}) are always skipped.
   * Changed events for {@link GraphObject}s that are in {@link Layer#isTemporary} layers are also skipped.
   *
   * Sometimes changed events do not even get to {@link #handleChanged} because
   * {@link Model#skipsUndoManager} or {@link Diagram#skipsUndoManager} is true.
   * @expose
   * @this {UndoManager}
   * @param {ChangedEvent} e the {@link ChangedEvent} received by {@link #handleChanged}.
   * @return {boolean} true to not record the change.
   */
  public skipsEvent(e: ChangedEvent): boolean {
    if (e === null) return true;
    if (e.change.value < 0) return true;

    const obj = e.object as any; // obj instanceof GraphObject (remove dependency)
    if (obj.layer !== undefined) {
      const lay = obj.layer;
      if (lay !== null) {
        if (lay.isTemporary) return true;
      }
    } else /*if (obj instanceof Layer) */{
      if (obj.isTemporary) return true;
    }

    return false;
  }

  /**
   * This read-only property returns an iterator for all of the {@link Model}s that this UndoManager is handling.
   * @see {@link #addModel}
   * @see {@link #removeModel}
   */
  /** @export */ get models(): Iterator<Model> { return this._models.iterator; }

  /**
   * Gets or sets whether this UndoManager records any changes.
   * The default value is false -- you need to set this to true if
   * you want the user to be able to undo or redo.
   *
   * You can temporarily turn off recording by setting {@link Diagram#skipsUndoManager}
   * and {@link Model#skipsUndoManager} to true.
   */
  /** @export */ get isEnabled(): boolean { return this._isEnabled; }
  /** @export */ set isEnabled(value: boolean) { this._isEnabled = value; }

  /**
   * This read-only property returns the {@link Transaction} in the {@link #history} to be undone next.
   * The value may be null if the UndoManager is not ready to perform an undo.
   * @see {@link #transactionToRedo}
   */
  /** @export */ get transactionToUndo(): Transaction | null {
    if (this.historyIndex >= 0 &&
      this.historyIndex <= this.history.count - 1) {
      return this.history.elt(this.historyIndex);
    }
    return null;
  }

  /**
   * This read-only property returns the {@link Transaction} in the {@link #history} to be redone next.
   * The value may be null if the UndoManager is not ready to perform a redo.
   * @see {@link #transactionToUndo}
   */
  /** @export */ get transactionToRedo(): Transaction | null {
    if (this.historyIndex < this.history.count - 1) {
      return this.history.elt(this.historyIndex + 1);
    }
    return null;
  }

  /**
   * This read-only property is true during a call to {@link #undo} or {@link #redo}.
   */
  /** @export */ get isUndoingRedoing(): boolean { return this._isUndoingRedoing; }

  /**
   * This read-only property returns the whole history, a list of all of the {@link Transaction}s,
   * each representing a transaction with some number of {@link ChangedEvent}s.
   *
   * You should not modify this List.
   */
  /** @export */ get history(): List<Transaction> { return this._history; }

  /**
   * Gets or sets the maximum number of transactions that this undo manager will remember.
   * When a transaction is committed and the number exceeds this value,
   * the UndoManager will discard the oldest transaction(s) in order to meet this limit.
   * The initial value is 999.
   * Any new value must be an integer.
   * A negative value is treated as if there were no limit.
   * A zero value will not remember any Transactions in the {@link #history},
   * but will allow commits and rollbacks to occur normally,
   * including raising "Transaction" type {@link ChangedEvent}s.
   *
   * This property is useful in helping limit the memory consumption of typical applications.
   * But this does not limit the number of {@link ChangedEvent}s that are recorded,
   * because there may be an unlimited number of those within each {@link Transaction}.
   * Decreasing this value will not necessarily remove any existing {@link Transaction}s
   * if there currently exist more in {@link #history} than the new value permits.
   */
  /** @export */ get maxHistoryLength(): number { return this._maxHistoryLength; }
  /** @export */ set maxHistoryLength(value: number) { this._maxHistoryLength = value; }

  /**
   * This read-only property returns the index into {@link #history} for the current undoable {@link Transaction}.
   * The value is -1 if there is no undoable {@link Transaction} to be undone.
   */
  /** @export */ get historyIndex(): number { return this._historyIndex; }

  /**
   * This read-only property returns the current {@link Transaction} for recording additional model change events.
   * This is initialized and augmented by {@link #handleChanged}
   * before it is added to {@link #history} by a top-level call
   * to {@link #commitTransaction}.
   * The value will be null between transactions.
   */
  /** @export */ get currentTransaction(): Transaction | null { return this._currentTransaction; }

  /**
   * This read-only property returns the current transaction level.
   * The value is zero when there is no ongoing transaction.
   * The initial value is zero.
   * {@link #startTransaction} will increment this value;
   * {@link #commitTransaction} or {@link #rollbackTransaction} will decrement it.
   * When this value is greater than zero, {@link #canUndo}
   * and {@link #canRedo} will be false, because
   * additional logically related model change events may occur.
   */
  /** @export */ get transactionLevel(): number { return this._transactionLevel; }

  /**
   * This read-only property is true after the first call to {@link #startTransaction}
   * and before a corresponding call to {@link #commitTransaction} or {@link #rollbackTransaction}.
   *
   * During a transaction {@link #canUndo} and {@link #canRedo} will be false.
   * {@link #currentTransaction} may be non-null if any {@link ChangedEvent}s were recorded.
   */
  /** @export */ get isInTransaction(): boolean { return this._transactionLevel > 0; }

  /**
   * @hidden @internal
   * Gets or sets whether the system should issue warnings if the {@link #transactionLevel}
   * appears to be inconsistent with current behavior.
   * The default value is false.
   */
  /** @export */ get checksTransactionLevel(): boolean { return this._checksTransactionLevel; }
  /** @export */ set checksTransactionLevel(value: boolean) { this._checksTransactionLevel = value; }

  /**
   * This read-only property returns a stack of ongoing transaction names.
   * The outermost transaction name will be the first item in the list.
   * The last one will be the name of the most recent (nested) call
   * to {@link #startTransaction}.
   *
   * You should not modify this List.
   */
  /** @export */ get nestedTransactionNames(): List<string> { return this._nestedTransactionNames; }

} // #end UndoManager
UndoManager['className'] = 'UndoManager';

// import { ChangedEvent } from './changedEvent';
// import { List, Set } from './collections';
// import { Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// // import { GraphObject } from './graphObject';
// import { Model } from './model';
// import { Util } from './util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Tool
*/

/**
 * Tools handle mouse, keyboard, and touch events.
 * The currently running tool, {@link Diagram#currentTool}, receives all input events from the {@link Diagram}
 * via canonicalized {@link InputEvent}s.
 * <p class="boxrun">
 * For more discussion, see <a href="../../intro/tools.html">Introduction to Tools</a>.
 * See samples that make use of tools in the <a href="../../samples/index.html#tools">samples index</a>.
 *
 * Most tools are "mode-less" tools that are managed by the {@link ToolManager},
 * which chooses the current tool based on the kind and position of the mouse event and the parts in the diagram.
 * The {@link ToolManager} has properties holding instances of most of the pre-defined {@link Tool} classes.
 * These classes include:
 *   - In the {@link ToolManager#mouseDownTools} {@link List}:
 *     - {@link ActionTool}
 *     - {@link RelinkingTool}
 *     - {@link LinkReshapingTool}
 *     - {@link RotatingTool}
 *     - {@link ResizingTool}
 *   - In the {@link ToolManager#mouseMoveTools} {@link List}:
 *     - {@link LinkingTool}
 *     - {@link DraggingTool}
 *     - {@link DragSelectingTool}
 *     - {@link PanningTool}
 *   - In the {@link ToolManager#mouseUpTools} {@link List}:
 *     - {@link ContextMenuTool}
 *     - {@link TextEditingTool}
 *     - {@link ClickCreatingTool}
 *     - {@link ClickSelectingTool}
 *
 * The ToolManager chooses a tool to run as the diagram's current tool by finding in its lists of tools the first tool
 * whose {@link #canStart} method returns true.  The ToolManager then sets {@link Diagram#currentTool} to be that tool.
 *
 * A tool is in the "running" state when it is the value of {@link Diagram#currentTool}.
 * The {@link Diagram#currentTool} property setter will call {@link #doStop} on the old tool
 * and then call {@link #doStart} on the new tool.
 *
 * A tool can then go into the "active" state once it decides it can actually do something.
 * This happens with a call to {@link #doActivate}, normally called by the ToolManager.
 * Later it is deactivated ({@link #doDeactivate}) and then stopped.
 * {@link #isActive} should be true when the tool is "active".
 * Often tools should ignore certain common events, such as calls to {@link #doMouseMove},
 * unless the tool {@link #isActive}.
 *
 * You can prevent a "mode-less" tool (i.e. one managed by the {@link ToolManager})
 * from being started by the ToolManager by setting {@link #isEnabled} to false.
 *
 * You can also go into a particular "mode" by setting {@link Diagram#currentTool} explicitly,
 * thereby circumventing the normal operation of the {@link ToolManager}.
 * This ignores the {@link #isEnabled} property and does not call the {@link #canStart} predicate.
 * The behavior will depend on the tool -- not all of the predefined tools support operating as a "modal" tool.
 *
 * Tools cannot be shared amongst multiple Diagrams.
 *
 * If you define a Tool subclass, you may override any of the methods whose names start with "do"
 * and any other methods that are documented to be overridable, such as {@link #canStart}.
 * However you must seriously consider calling the base method in order to gets its default behavior.
 * There may be situations where not calling the base method may cause subtle bugs.
 * But that depends on the method and the tool.
 * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call the base method.
 * @unrestricted
 * @category Tool
 */
export abstract class Tool {
  /** @hidden @internal */ public static className = 'Tool';
  /** @hidden @internal */ private _diagram: Diagram;
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _isEnabled: boolean;
  /** @hidden @internal */ private _isActive: boolean;
  /** @hidden @internal */ private _transactionResult: string | null;
  /** @hidden @internal */ private _waitAfterTimer: number;
  /** @hidden @internal */ private _toolEvent: InputEvent;

  /**
   * Don't construct this directly -- this is an abstract class.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(Tool);
    Set.uniqueHash(this);

    this._diagram = Diagram.getLastDiagram();  // called by Diagram property setters and ToolManager methods
    this._name = '';
    this._isEnabled = true;
    this._isActive = false;

    this._transactionResult = null;

    // We don't really need one per tool, but this stops a larger number
    // of events getting created for standardWaitAfter
    this._toolEvent = new InputEvent();

    this._waitAfterTimer = -1;
  }

  /**
   * This read-only property returns the {@link Diagram} that owns this tool and
   * for which this tool is handling input events.
   */
  /** @export */ get diagram(): Diagram { return this._diagram; }
  /** @export */ set diagram(val: Diagram) { if (val instanceof Diagram) this._diagram = val; }

  /**
   * @hidden @internal
   * @this {Tool}
   * @return {string}
   */
  public toString(): string {
    if (this.name !== '') {
      return this.name + ' Tool';
    } else {
      return Util.getClassName(this.constructor);
    }
  }

  /**
   * The diagram asks each tool to update any adornments the tool might
   * use for a given part.
   * If the tool uses its own tool handles,
   * this should display them or hide them as appropriate.
   * Typically this should only show them if the part is selected.
   *
   * By default this method does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {Part} part
   */
  public updateAdornments(part: Part): void { }

  // States and state transitions

  /**
   * This predicate is used by the {@link ToolManager} to decide if this tool can be started mode-lessly
   * by mouse and touch events.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * By default this method returns {@link #isEnabled}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @return {boolean} true if {@link #isEnabled} is true and
   * if the {@link Diagram#toolManager} can make this tool the {@link Diagram#currentTool} and
   * then call the {@link #doStart} method.
   */
  public canStart(): boolean {
    return this.isEnabled;
  }

  /**
   * The Diagram calls this method when this tool becomes the current tool; you should not call this method.
   * Tool implementations should perform their per-use initialization here, such
   * as setting up internal data structures, or capturing the mouse.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * You should not call this method -- only the {@link Diagram#currentTool} property setter should call this method.
   *
   * By default this method does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * If you override this method, it is commonplace to also override {@link #doStop} to clean up whatever you set up in this method.
   * @expose
   * @this {Tool}
   */
  public doStart(): void { }

  /**
   * The Diagram calls this method after setting {@link Diagram#currentTool},
   * to make the new tool active.
   * This should set {@link #isActive} to true.
   * Overrides of this method might call {@link #startTransaction},
   * if this tool's activity involves modification of the model.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * You should call this method only after setting {@link Diagram#currentTool} to the Tool that you want to activate.
   *
   * By default this only sets {@link #isActive} to true.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * If you override this method, it is commonplace to also override {@link #doDeactivate} to clean up whatever you set up in this method.
   * @expose
   * @this {Tool}
   */
  public doActivate(): void {
    this.isActive = true;
  }

  /**
   * The Diagram calls this method on the old tool when
   * {@link Diagram#currentTool} is set to a new tool.
   * This needs to set {@link #isActive} to false.
   * Overrides of this method might call {@link #stopTransaction},
   * if this tool's activity involves modification of the model.
   *
   * You should have no reason to call this method, because it is automatically called by the
   * {@link Diagram#currentTool} property setter on the old tool.
   *
   * By default this only sets {@link #isActive} to false.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * It is commonplace to override this method in order to clean up whatever you have set up in an override of {@link #doActivate}.
   * @expose
   * @this {Tool}
   */
  public doDeactivate(): void {
    this.isActive = false;
  }

  /**
   * The Diagram calls this method when this tool stops being the current tool; you should not call this method.
   * Tool implementations should perform their per-use cleanup here,
   * such as releasing mouse capture.
   *
   * You should not call this method -- only the {@link Diagram#currentTool} property setter should call this method.
   * If you want to stop a tool unexpectedly, you should call {@link #doCancel}.
   * If your implementation of a tool wants to stop itself, you should call {@link #stopTool}.
   *
   * By default this method does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * It is commonplace to override this method in order to clean up whatever you have set up in an override of {@link #doStart}.
   * @expose
   * @this {Tool}
   */
  public doStop(): void { }

  /**
   * The diagram will call this method when the user wishes to cancel the
   * current tool's operation.
   * Typically this is called when the user hits the ESCAPE key.
   * This should restore the original state of what was modified by this tool, and then it should call {@link #stopTool}.
   * This method is not responsible for cleaning up any side-effects that should be performed
   * by {@link #doDeactivate} and/or {@link #doStop}, which will always be called whether the tool stops normally or abnormally.
   *
   * By default this method just sets {@link #transactionResult} to null and calls {@link #stopTool}.
   * You will want to override this method even in tools that call {@link #startTransaction} and {@link #stopTransaction},
   * because the {@link UndoManager} might not be enabled.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doCancel(): void {
    this.transactionResult = null;
    this.stopTool();
  }

  /**
   * If the {@link Diagram#currentTool} is this tool,
   * stop this tool and start the {@link Diagram#defaultTool}
   * by making it be the new current tool.
   * The implementation of various tool methods can call this method to stop the current tool.
   * This will call {@link #doStop} -- you should not call that method directly.
   *
   * If you want to stop the current tool and have it restore the original state, call {@link #doCancel}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public stopTool(): void {
    const diagram = this.diagram;
    if (diagram.currentTool === this) {
      // this will call doStop on the current tool,
      // and then make the defaultTool current and call doStart on it
      diagram.currentTool = null;  // NOT diagram.defaultTool, otherwise currentTool side-effects might not happen
      diagram.currentCursor = '';
    }
  }

  // Input events

  /**
   * The diagram will call this method upon a mouse down event.
   * This is normally overridden for mouse-down tools;
   * it is not called for mouse-move or mouse-up tools.
   * However it may also be called when the tool is run in a modal fashion,
   * when code explicitly sets the diagram's {@link Diagram#currentTool}.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * By default this method checks {@link #isActive}; if that is false it calls {@link #canStart}.
   * If that in turn is true, this calls {@link #doActivate}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doMouseDown(): void {
    if (!this.isActive && this.canStart()) {
      this.doActivate();
    }
  }

  /**
   * The diagram will call this method upon a mouse move event.
   * This is normally overridden for mouse-move tools;
   * it is not called for mouse-up tools.
   * However it may also be called when the tool is run in a modal fashion,
   * when code explicitly sets the diagram's {@link Diagram#currentTool}.
   * An override of this method usually does nothing when {@link #isActive} is false.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * By default this method does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doMouseMove(): void { }

  /**
   * The diagram will call this method upon a mouse up event.
   * This is normally overridden for mouse-up tools.
   * An override of this method usually does nothing when {@link #isActive} is false,
   * except for calling {@link #stopTool}.
   * Tools normally stop upon a mouse up, by calling {@link #stopTool}.
   * If you want to handle multiple mouse down-up gestures in one tool activation,
   * you will need to override this method to only stop the tool when you want.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * By default this method just calls {@link #stopTool}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doMouseUp(): void {
    this.stopTool();
  }

  /**
   * The diagram will call this method as the mouse wheel is rotated.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the mouse
   * event and input state.
   *
   * By default this method does nothing.
   * (But the {@link ToolManager#doMouseWheel} override will call {@link Tool#standardMouseWheel}.)
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doMouseWheel(): void { }

  /**
   * Called by {@link ToolManager#doMouseDown} and {@link ToolManager#doMouseMove},
   * this method determines whether or not to allow pinch zooming from a multi-touch event.
   * By default this predicate just returns true.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @return {boolean}
   * @see {@link #standardPinchZoomStart}
   * @see {@link #standardPinchZoomMove}
   * @since 1.5
   */
  public canStartMultiTouch(): boolean {
    return true;
  }

  /**
   * Initiates pinch-zooming on multi-touch devices.
   *
   * This is called by {@link ToolManager#doMouseDown} if the {@link Diagram#lastInput} has {@link InputEvent#isMultiTouch}
   * set to true and {@link #canStartMultiTouch} returns true.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @see {@link #canStartMultiTouch}
   * @see {@link #standardPinchZoomMove}
   * @since 1.5
   */
  public standardPinchZoomStart(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;

    const first = e.getMultiTouchViewPoint(0, Point.allocAt(NaN, NaN));
    const second = e.getMultiTouchViewPoint(1, Point.allocAt(NaN, NaN));
    if (!first.isReal() || !second.isReal()) {
      Point.free(first);
      Point.free(second);
      return;
    }

    // event.isMultiTouch
    // call doCancel to stop the rest of typical toolManager.domousedown/domousemove behavior
    this.doCancel();
    if (diagram.getInputOption('hasGestureZoom')) {
      diagram._origScale = diagram.scale;
      const x = second.x - first.x;
      const y = second.y - first.y;
      const dist = Math.sqrt((x * x) + (y * y));
      diagram._startDist = dist;
      e.bubbles = false;
    }
    Point.free(first);
    Point.free(second);
  }

  /**
   * Continues pinch-zooming (started by {@link #standardPinchZoomStart} on multi-touch devices.
   *
   * This is called by {@link ToolManager#doMouseMove} if the {@link Diagram#lastInput} has {@link InputEvent#isMultiTouch}
   * set to true and {@link #canStartMultiTouch} returns true.
   * By default this calls {@link #doCancel} in order to cancel the regular tool behavior caused by the multitouch events.
   * This then calculates the appropriate zoom level and calls {@link CommandHandler#canResetZoom} to decide
   * whether to call {@link CommandHandler#resetZoom} to actually set {@link Diagram#scale}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @see {@link #canStartMultiTouch}
   * @see {@link #standardPinchZoomStart}
   * @since 1.5
   */
  public standardPinchZoomMove(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;

    const first = e.getMultiTouchViewPoint(0, Point.allocAt(NaN, NaN));
    const second = e.getMultiTouchViewPoint(1, Point.allocAt(NaN, NaN));
    if (!first.isReal() || !second.isReal()) {
      Point.free(first);
      Point.free(second);
      return;
    }

    // event.isMultiTouch
    // call doCancel to stop the rest of typical toolManager.domousedown/domousemove behavior
    this.doCancel();

    if (diagram.getInputOption('hasGestureZoom')) {
      const x = second.x - first.x;
      const y = second.y - first.y;
      const dist = Math.sqrt((x * x) + (y * y));
      const scale = dist / diagram._startDist;
      const center = new Point((Math.min(second.x, first.x) + Math.max(second.x, first.x)) / 2,
        (Math.min(second.y, first.y) + Math.max(second.y, first.y)) / 2);

      const newscale = diagram._origScale * scale;
      const command = diagram.commandHandler;
      if (newscale !== diagram.scale && command.canResetZoom(newscale)) {
        const oldzoom = diagram.zoomPoint;
        diagram.zoomPoint = center;
        command.resetZoom(newscale);
        diagram.zoomPoint = oldzoom;
      }
      e.bubbles = false;
    }
    Point.free(first);
    Point.free(second);
  }

  /**
   * The diagram will call this method upon a key down event.
   * By default this just calls {@link #doCancel} if the key is the ESCAPE key.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the key.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doKeyDown(): void {
    const diagram = this.diagram;
    if (diagram.lastInput.key === 'Esc') {
      this.doCancel();
    }
  }

  /**
   * The diagram will call this method upon a key up event.
   * Implementations of this method can look at {@link Diagram#lastInput} to get the key.
   *
   * By default this method does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public doKeyUp(): void { }

  // Transactions

  /**
   * Call {@link Diagram#startTransaction} with the given transaction name.
   * This always sets {@link #transactionResult} to null.
   *
   * This is normally called in an override of {@link #doActivate}, if the tool modifies the model,
   * along with a call to {@link #stopTransaction} in an override of {@link #doDeactivate}.
   * Alternatively, you can surround a block of code that sets the {@link #transactionResult}
   * with calls to startTransaction and stopTransaction.
   * @export
   * @this {Tool}
   * @param {string=} tname a string describing the transaction.
   * @return {boolean} the value of the call to {@link Diagram#startTransaction}.
   */
  public startTransaction(tname?: string): boolean {
    if (tname === undefined) tname = this.name;
    this.transactionResult = null;
    const diagram = this.diagram;
    return diagram.startTransaction(tname);
  }

  /**
   * If {@link #transactionResult} is null, call {@link Diagram#rollbackTransaction},
   * otherwise call {@link Diagram#commitTransaction}.
   *
   * This is normally called in an override of {@link #doDeactivate},
   * if {@link #startTransaction} was called in {@link #doActivate}.
   * Alternatively, you can surround a block of code that sets the {@link #transactionResult}
   * with calls to startTransaction and stopTransaction.
   * @export
   * @this {Tool}
   * @return {boolean} the result of the call to rollback or commit the transaction.
   */
  public stopTransaction(): boolean {
    const diagram = this.diagram;
    if (this.transactionResult === null) {
      return diagram.rollbackTransaction();
    } else {
      return diagram.commitTransaction(this.transactionResult);
    }
  }

  // Useful methods

  /**
   * Implement the standard behavior for selecting parts with the mouse,
   * depending on the control and shift modifier keys.
   *
   * Control-clicking on a part will select it if it wasn't already,
   * and will deselect if it had been selected.
   * Shift-clicking on a part will add it to the selection (if it wasn't already).
   * Otherwise, clicking on a part will select it (if it wasn't already).
   *
   * Note that there are restrictions on selection.
   * For example, a part cannot be selected in this manner if {@link Part#selectable} is false,
   * or if {@link Diagram#maxSelectionCount} would be exceeded.
   *
   * A left click in the background of the diagram with no modifier keys clears the selection.
   *
   * This method does not implement any click event behavior -- that is implemented by {@link #standardMouseClick}.
   *
   * The {@link ClickSelectingTool} calls this method in its override of {@link #doMouseUp} in order to change the selection.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public standardMouseSelect(): void {
    const diagram = this.diagram;
    if (!diagram.allowSelect) return;
    const e = diagram.lastInput;
    const curobj = diagram.findPartAt(e.documentPoint, false);  // to select containing Group if Part.canSelect() is false
    if (curobj !== null) {
      if (Util.isMac ? e.meta : e.control) {  // toggle the part's selection
        diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
        let part = curobj;
        while (part !== null && !part.canSelect()) part = part.containingGroup;
        if (part !== null) part.isSelected = !part.isSelected;
        diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      } else if (e.shift) {  // add the part to the selection
        if (!curobj.isSelected) {
          diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
          let part = curobj;
          while (part !== null && !part.canSelect()) part = part.containingGroup;
          if (part !== null) part.isSelected = true;
          diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
        }
      } else {
        if (!curobj.isSelected) {
          let part = curobj;
          while (part !== null && !part.canSelect()) part = part.containingGroup;
          if (part !== null) diagram.select(part);  // also raises ChangingSelection/Finished
        }
      }
    } else if (e.left && !(Util.isMac ? e.meta : e.control) && !e.shift) {
      // left click on background with no modifier: clear selection
      diagram.clearSelection();  // also raises ChangingSelection/Finished
    }
  }

  /**
   * Implement the standard behavior for mouse clicks,
   * searching for and calling click handler functions on {@link GraphObject}s
   * or on {@link Diagram},
   * and raising the corresponding {@link DiagramEvent}.
   *
   * A click on a {@link GraphObject} of the diagram will raise one of the following {@link DiagramEvent}s:
   * "ObjectSingleClicked", "ObjectDoubleClicked", or "ObjectContextClicked".
   * This will also look at the corresponding click property:
   * {@link GraphObject#click}, {@link GraphObject#doubleClick}, or {@link GraphObject#contextClick}.
   * If the value is a function, this will call it, passing the current {@link InputEvent}
   * and the GraphObject.
   * If the value is null, it tries looking at the parent {@link GraphObject#panel},
   * and so on, walking up the visual tree until it finds the appropriate function to call.
   * After calling the click function, if the value of {@link InputEvent#handled} is false,
   * this method will continue walking up the visual tree looking for more click functions
   * to call.
   * Once it has looked at the top-level object (a {@link Part})
   * for a click function, this method stops.
   *
   * A click in the background of the diagram will raise one of the following {@link DiagramEvent}s:
   * "BackgroundSingleClicked", "BackgroundDoubleClicked", or "BackgroundContextClicked".
   * This will also look at the corresponding click property:
   * {@link Diagram#click}, {@link Diagram#doubleClick}, or {@link Diagram#contextClick}.
   * If the value is a function, this will call it, passing the current {@link InputEvent}.
   *
   * This method is not responsible for selecting or deselecting any parts.
   * Call {@link #standardMouseSelect} for that functionality.
   *
   * Note that this calls {@link GraphObject#isEnabledObject} on the target object;
   * if it returns false, no click action will occur.
   *
   * The {@link ClickSelectingTool} calls this method in its override of {@link #doMouseUp} in order to raise "click" events.
   * Note that by default GraphObjects in Layers that are {@link Layer#isTemporary} will not be "clicked".
   * To change that behavior it is easiest to set {@link GraphObject#isActionable} to true on those objects for which you wish to handle "click" events.
   * Then the {@link ActionTool}'s {@link #doMouseUp} override will raise the standard "click" events.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {function(GraphObject):GraphObject | null=} navig An optional custom navigation
   * function to find target objects.
   * @param {function(GraphObject):boolean | null=} pred An optional custom predicate
   * function to find target objects. No value means that only objects in layers holding permanent objects.
   * @return {boolean} true if {@link InputEvent#handled} had been set to true on the {@link Diagram#lastInput}.
   */
  public standardMouseClick<T extends GraphObject>(navig?: (a: GraphObject) => (T | null), pred?: (a: T) => boolean): boolean {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = function (x) { return !x.layer.isTemporary; };

    const diagram = this.diagram;

    const e = diagram.lastInput;
    const curobj = diagram.findObjectAt(e.documentPoint, navig, pred);
    e.targetObject = curobj;

    this._standardClick(curobj, e, diagram);
    return e.handled;
  }

  /**
   * @hidden @internal
   * @this {Tool}
   * @param {GraphObject} curobj
   * @param {InputEvent} e
   * @param {Diagram} diagram
   */
  protected _standardClick(curobj: GraphObject, e: InputEvent, diagram: Diagram): void {
    e.handled = false;

    // no clicks for disabled objects!
    if (curobj !== null && !curobj.isEnabledObject()) return;

    // decide whether single-click, double-click, or context-click
    let kind = 0;
    if (e.left) {
      if (e.clickCount === 1) {
        kind = 1;
      } else if (e.clickCount === 2) {
        kind = 2;
      } else { // assume more than a double-click is just a rapid single click
        kind = 1;
      }
    } else if (e.right) {
      if (e.clickCount === 1) {
        kind = 3;
      }
    }

    // raise DiagramEvent
    let eventkind: DiagramEventName = 'ObjectSingleClicked';
    if (curobj !== null) {
      switch (kind) {
        case 1: eventkind = 'ObjectSingleClicked'; break;
        case 2: eventkind = 'ObjectDoubleClicked'; break;
        case 3: eventkind = 'ObjectContextClicked'; break;
      }
      if (kind !== 0) diagram.raiseDiagramEvent(eventkind, curobj);
    } else {
      switch (kind) {
        case 1: eventkind = 'BackgroundSingleClicked'; break;
        case 2: eventkind = 'BackgroundDoubleClicked'; break;
        case 3: eventkind = 'BackgroundContextClicked'; break;
      }
      if (kind !== 0) diagram.raiseDiagramEvent(eventkind);
    }

    // call any event handlers, bubbling up the chain of GraphObjects
    if (curobj !== null) {
      while (curobj !== null) {
        let func: (e: InputEvent, obj: GraphObject) => void = null;
        switch (kind) {
          case 1: func = curobj.click; break;
          case 2: func = curobj.doubleClick ? curobj.doubleClick : curobj.click; break;
          case 3: func = curobj.contextClick; break;
        }
        if (func !== null) {
          func(e, curobj);  // pass the InputEvent and the bubbled GraphObject
          // (the original GraphObject getting the event is in InputEvent.targetObject)
          if (e.handled) break;
        }
        curobj = curobj.panel;
      }
    } else {
      let dfunc: (e: InputEvent) => void = null;
      switch (kind) {
        case 1: dfunc = diagram.click; break;
        case 2: dfunc = diagram.doubleClick ? diagram.doubleClick : diagram.click; break;
        case 3: dfunc = diagram.contextClick; break;
      }
      if (dfunc !== null) {
        dfunc(e);  // pass the InputEvent; there's no current object or e.targetObject
      }
    }
  }

  /**
   * Implement the standard behavior for mouse enter, over, and leave events,
   * where the mouse is moving but no button is pressed.
   * This should be called by mouse move event handlers when wanting to
   * detect and invoke mouse enter/over/leave event handlers.
   *
   * The {@link GraphObject#mouseEnter} property provides a function to call
   * when the mouse first enters an object or any of its contained objects
   * (if the object is actually a {@link Panel}).
   *
   * The {@link GraphObject#mouseLeave} property provides a function to call
   * when the mouse leaves an object and all of its contained objects
   * (if the object is actually a {@link Panel}).
   *
   * The {@link GraphObject#mouseOver} property and {@link Diagram#mouseOver}
   * properties provide functions to call when the mouse moves but stays
   * within the same {@link GraphObject} or when the mouse moves in the
   * background of the {@link Diagram}.
   *
   * This method is also responsible for updating the {@link Diagram#currentCursor}
   * according to the value of {@link GraphObject#cursor} and {@link Diagram#defaultCursor}.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public standardMouseOver(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    // Will stop some flickering effects while animations set up
    if (diagram.animationManager._pendingAnimation === true) return;

    const oldskip = diagram.skipsUndoManager;
    diagram.skipsUndoManager = true;

    let currobj = diagram.viewportBounds.containsPoint(e.documentPoint) ? diagram.findObjectAt(e.documentPoint, null, null) : null;
    e.targetObject = currobj;

    let anyfunc = false;

    // if we've changed current objects, call mouseLeave and mouseEnter handlers
    if (currobj !== diagram._currentEnterLeaveObject) {
      // save old and new objects
      let prevobj = diagram._currentEnterLeaveObject;
      const previousEnterLeaveObject = prevobj;
      diagram._currentEnterLeaveObject = currobj;

      // needed to cleanup tooltips:
      this.doCurrentObjectChanged(prevobj, currobj);

      // look for mouseLeave event handler for objects that we left
      e.handled = false;
      while (prevobj !== null) {
        const func = prevobj.mouseLeave;
        if (func !== null) {
          // if the new/current object is inside the previous Panel,
          // we haven't left that Panel
          if (currobj === prevobj) break;
          if (currobj !== null && currobj.isContainedBy(prevobj)) break;
          func(e, prevobj, currobj);
          anyfunc = true;
          if (e.handled) break;
        }
        prevobj = prevobj.panel;
      }
      prevobj = previousEnterLeaveObject;

      // look for mouseEnter event handler for objects that we entered
      // ??? should this call event handler in reverse order: from outside in?
      e.handled = false;
      while (currobj !== null) {
        const func = currobj.mouseEnter;
        if (func !== null) {
          // if the old/previous object is inside the current Panel,
          // we haven't entered this Panel
          if (prevobj === currobj) break;
          if (prevobj !== null && prevobj.isContainedBy(currobj)) break;
          func(e, currobj, prevobj);
          anyfunc = true;
          if (e.handled) break;
        }
        currobj = currobj.panel;
      }
      currobj = diagram._currentEnterLeaveObject;
    }

    if (currobj !== null) {  // now inside some GraphObject
      // look for a cursor -- anything that isn't an empty string
      let obj = currobj;
      let curs = '';  // but default to an empty string
      while (obj !== null) {
        curs = obj.cursor;
        if (curs !== '') break;
        obj = obj.panel;
      }
      diagram.currentCursor = curs;

      // look for mouseOver event handler for current object that we're over
      e.handled = false;
      obj = currobj;
      while (obj !== null) {
        const func2 = obj.mouseOver;
        if (func2 !== null) {
          func2(e, obj);
          anyfunc = true;
          if (e.handled) break;
        }
        obj = obj.panel;
      }
    } else {  // not now in any GraphObject
      diagram.currentCursor = '';  // use Diagram.defaultCursor
      // maybe execute a mouseOver event handler
      const func3 = diagram.mouseOver;
      if (func3 !== null) {
        func3(e);
        anyfunc = true;
      }
    }

    if (anyfunc) diagram.requestUpdate();
    diagram.skipsUndoManager = oldskip;
  }

  /**
   * @hidden @internal
   * This is called by {@link #standardMouseOver} when the current object changes.
   * By default this does nothing.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {GraphObject} prev
   * @param {GraphObject} curr
   */
  public doCurrentObjectChanged(prev: GraphObject, curr: GraphObject): void {
  }

  /**
   * Implement the standard behavior for mouse wheel events.
   * {@link ToolManager#doMouseWheel} calls this method.
   *
   * Turning the mouse wheel if {@link Diagram#allowVerticalScroll} is true
   * causes the diagram to scroll up or down.
   * If {@link InputEvent#shift} and {@link Diagram#allowHorizontalScroll} are true,
   * the diagram scrolls left or right.
   *
   * If {@link InputEvent#control} and {@link Diagram#allowZoom} are true,
   * turning the mouse wheel changes the diagram's scale,
   * zooming in or out while trying to keep the point in the model
   * at the same point as the mouse.
   *
   * The value of {@link ToolManager#mouseWheelBehavior} affects what operations might occur
   * upon mouse wheel events.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public standardMouseWheel(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    const delta = e.delta;
    if (delta === 0) return;
    if (!diagram.documentBounds.isReal()) return;
    const cmd = diagram.commandHandler;
    const wheelBehavior = diagram.toolManager.mouseWheelBehavior;
    if (cmd !== null &&
      (((wheelBehavior === ToolManager.WheelZoom && !e.shift) ||
      (wheelBehavior === ToolManager.WheelScroll && e.control)) &&
      (delta > 0 ? cmd.canIncreaseZoom() : cmd.canDecreaseZoom()))) {
      // zoom in or out at current mouse point
      const oldzoom = diagram.zoomPoint;
      diagram.zoomPoint = e.viewPoint;
      if (delta > 0) {
        cmd.increaseZoom();
      } else {
        cmd.decreaseZoom();
      }
      diagram.zoomPoint = oldzoom;
      e.bubbles = false;
    } else if ((wheelBehavior === ToolManager.WheelZoom && e.shift) ||
      (wheelBehavior === ToolManager.WheelScroll && !e.control)) {
      const oldpos = diagram.position.copy();
      let total = (delta > 0) ? delta : -delta;
      const wheelEvent = e.event as WheelEvent;
      let deltaMode = wheelEvent.deltaMode;
      let deltaX = wheelEvent.deltaX;
      let deltaY = wheelEvent.deltaY;
      let scrollType: ('pixel' | 'line' | 'page' | 'document') = 'pixel';

      if (Util.isIE10 || Util.isIE11 || Util.isEdge) {
        deltaMode = 1;
        if (deltaX > 0) deltaX = 3;
        if (deltaX < 0) deltaX = -3;
        if (deltaY > 0) deltaY = 3;
        if (deltaY < 0) deltaY = -3;
      }

      // if the event has delta values defined, use those
      // https://developer.mozilla.org/en-US/docs/Web/Events/wheel
      if (deltaMode !== undefined && deltaX !== undefined && deltaY !== undefined && (deltaX !== 0 || deltaY !== 0) && !e.shift) {
        switch (deltaMode) {
          case 0: scrollType = 'pixel'; break;
          case 1: scrollType = 'line'; break;
          case 2: scrollType = 'page'; break;
          default: scrollType = 'pixel'; break;
        }
        if (deltaX !== 0 && diagram.allowHorizontalScroll) {
          if (deltaX > 0) diagram.scroll(scrollType, 'left', -deltaX);
          else diagram.scroll(scrollType, 'right', deltaX);
        }
        if (deltaY !== 0 && diagram.allowVerticalScroll) {
          if (deltaY > 0) diagram.scroll(scrollType, 'up', -deltaY);
          else diagram.scroll(scrollType, 'down', deltaY);
        }
      } else {
        // 48 pixels is normal (3 lines, 16*3)
        // delta is normally -1 or 1
        if (!e.shift && diagram.allowVerticalScroll) {
          // scroll up or down
          const c = diagram.scrollVerticalLineChange;
          total = (total * 3) * c;
          if (delta > 0) diagram.scroll('pixel', 'up', total);
          else diagram.scroll('pixel', 'down', total);
        } else if (e.shift && diagram.allowHorizontalScroll) {
          // scroll left or right
          const c = diagram.scrollHorizontalLineChange;
          total = (total * 3) * c;
          if (delta > 0) diagram.scroll('pixel', 'left', total);
          else diagram.scroll('pixel', 'right', total);
        }
      }

      if (!diagram.position.equals(oldpos)) {
        e.bubbles = false;
      }
    }
  }

  /**
   * This is called to start a new timer to call {@link #doWaitAfter} after a given delay.
   * It first cancels any previously running "WaitAfter" timer, by calling {@link #cancelWaitAfter}.
   *
   * This is normally used to implement mouse hover and mouse hold events.
   * If the mouse has moved, it must not have moved beyond the distance as determined by
   * {@link Tool#isBeyondDragSize} for it be considered "stationary".
   * So the regular {@link ToolManager#doMouseMove} implementation only calls
   * this method when the mouse has moved beyond the drag size.
   *
   * This method is rarely overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {number} delay The delay, in milliseconds.
   * @param {InputEvent=} event An optional event that caused this timer. Defaults to {@link Diagram#lastInput}.
   * This gets passed on to {@link #doWaitAfter}.
   */
  public standardWaitAfter(delay: number, event?: InputEvent): void {
    if (Debug) Util.checkPrimitive(delay, 'number', Tool, 'standardWaitAfter:delay');
    if (event === undefined) event = this.diagram.lastInput;
    this.cancelWaitAfter();
    const tool = this;  // closure needed as arg to setTimeout
    const e = event.clone(this._toolEvent); // so we aren't just passing a reference to diagram.lastInput
    this._waitAfterTimer = Util.setTimeout(function () { tool.doWaitAfter(e); }, delay);
  }

  /**
   * This is called to cancel any running "WaitAfter" timer.
   *
   * This is called when a tool is stopped.
   *
   * This method is rarely overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   */
  public cancelWaitAfter(): void {
    if (this._waitAfterTimer !== -1) {
      Util.clearTimeout(this._waitAfterTimer);
    }
    this._waitAfterTimer = -1;
  }

  /**
   * This is called a certain delay after a call to {@link #standardWaitAfter} if there
   * has not been any call to {@link #cancelWaitAfter}.
   * The {@link ToolManager} overrides this method in order to implement support for mouse-hover behavior and tooltips.
   *
   * By default this does nothing.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {InputEvent} event The event that caused {@link #standardWaitAfter}.
   */
  public doWaitAfter(event: InputEvent): void {
  }

  /**
   * This convenience function finds the front-most {@link GraphObject}
   * that is at a given point and that is an element of an {@link Adornment}
   * that is of a given category.
   * The tool handle must be an immediate element of the Adornment,
   * not a GraphObject that is nested within Panels within the Adornment.
   *
   * This method is very infrequently overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {Point} p a Point in document coordinates.
   * @param {string} category the required {@link Part#category} of the {@link Adornment}.
   * @return {GraphObject}
   */
  public findToolHandleAt(p: Point, category: string): GraphObject | null {
    const diagram = this.diagram;
    const element = diagram.findObjectAt(p,
      function (x) {
        while (x !== null && !(x.panel instanceof Adornment)) x = x.panel;
        return x;
      });
    if (element === null) return null;
    if (element.part.category === category) return element;
    return null;
  }

  /**
   * Return true when the last mouse point is far enough away from the first mouse down point
   * to constitute a drag operation instead of just a potential click.
   *
   * This uses the value of {@link ToolManager#dragSize}.
   * On touch devices the value is automatically increased to accommodate the unavoidable movement of fingers.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Tool}
   * @param {Point=} first Point in view coordinates, defaults to {@link Diagram#firstInput}'s {@link InputEvent#viewPoint}.
   * @param {Point=} last Point in view coordinates, defaults to {@link Diagram#lastInput}'s {@link InputEvent#viewPoint}.
   * @return {boolean}
   */
  public isBeyondDragSize(first?: Point, last?: Point): boolean {
    const diagram = this.diagram;
    // Get both of the points in screen/view coordinates
    if (first === undefined) first = diagram.firstInput.viewPoint;
    if (last === undefined) last = diagram.lastInput.viewPoint;
    const dsize = diagram.toolManager.dragSize;
    let dw = dsize.width;
    let dh = dsize.height;
    if (diagram.firstInput.isTouchEvent) {
      dw += 6;
      dh += 6;
    }
    return (Math.abs(last.x - first.x) > dw ||
      Math.abs(last.y - first.y) > dh);
  }

  /**
   * Gets or sets the name of this tool.
   * The default name is an empty string,
   * but the constructor for each instance of a subclass of Tool will initialize it appropriately.
   * For example, the name of the {@link DragSelectingTool} is "DragSelecting".
   *
   * This name is sometimes used by tools that use {@link Adornment}s as the {@link Part#category} for their Adornments.
   * It is also sometimes used by tools that conduct transactions as the transaction name.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) {
    Util.checkPrimitive(value, 'string', Tool, 'name');
    this._name = value;
  }

  /**
   * Gets or sets whether this tool can be started by a mouse event.
   *
   * Set this to false to prevent {@link #canStart} from returning true.
   * Setting this property to false should prevent this tool from being used in a mode-less fashion
   * by the {@link ToolManager} with a mouse down/move/up event.
   * However, even when this property is false, this tool can still be used in a modal fashion:
   * it can still be started by explicitly setting the
   * {@link Diagram#currentTool} property to this tool.
   *
   * The default value is true.
   */
  /** @export */ get isEnabled(): boolean { return this._isEnabled; }
  /** @export */ set isEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Tool, 'isEnabled');
    this._isEnabled = value;
  }

  /**
   * Gets or sets whether this tool is started and is actively doing something.
   *
   * You can set this to true after your tool is started (i.e. when it is the
   * {@link Diagram#currentTool} and {@link #doStart}
   * had been called), but when it is not yet in a state
   * that it is actually "doing" something, because it is waiting for the right
   * circumstances.  This is typically only important when the tool is used in
   * a modal fashion.
   *
   * The default value is false.
   * This is normally set by {@link #doActivate} and {@link #doDeactivate}.
   */
  /** @export */ get isActive(): boolean { return this._isActive; }
  /** @export */ set isActive(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Tool, 'isActive');
    this._isActive = value;
  }

  /**
   * Gets or sets the name of the transaction to be committed by {@link #stopTransaction}
   *
   * If null, the transaction will be rolled back.
   *
   * If this is non-null at the time of a call to {@link #stopTransaction},
   * it calls {@link Diagram#commitTransaction} with this transaction name;
   * if this is null at that time, it calls {@link Diagram#rollbackTransaction}.
   *
   * The default value is null; {@link #startTransaction} will also set this to null.
   * Because a value of null when {@link #stopTransaction} is called will rollback the transaction,
   * it is important that your code sets this property to a non-null value when it thinks it has succeeded.
   *
   * This property exists so that no matter what execution path occurs to end the usage of a tool,
   * any ongoing transaction can be properly committed or rolled-back.
   * Many tools call {@link #startTransaction} and {@link #stopTransaction}; thus they set this property
   * for their transaction to be committed.
   * {@link #doCancel} also sets this property to null.
   */
  /** @export */ get transactionResult(): string | null { return this._transactionResult; }
  /** @export */ set transactionResult(value: string | null) {
    if (value !== null) Util.checkPrimitive(value, 'string', Tool, 'transactionResult');
    this._transactionResult = value;
  }

} // end class Tool
Tool['className'] = 'Tool';

// import { Set } from './collections';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { DiagramEventName, InputEvent  } from './inputEvent';
// import { Adornment, Part } from './parts'; // maybe don't need Node and Link if we redefine the args in the linking tools
// import { Point } from './point';
// import { ToolManager } from './toolManager';
// import { Util } from './util';




/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
ToolManager
*/

// import { EnumValue } from './enumValue';
// import { Tool } from './tool';

/**
 * This special {@link Tool} is responsible for managing all of the Diagram's
 * mode-less tools, which you can access as the {@link Diagram#toolManager}.
 *
 * Mode-less tools are tools that are present in one of the following lists:
 * {@link #mouseDownTools}, {@link #mouseMoveTools}, or {@link #mouseUpTools}.
 * This ToolManager tool is normally the {@link Diagram#defaultTool},
 * so it is also usually the {@link Diagram#currentTool} when the user is doing "nothing".
 *
 * When this tool is running as the current tool, it handles mouse-down,
 * mouse-move, and mouse-up events and the corresponding touch events.
 * For each event it iterates over each of the tools in the corresponding list,
 * calling the tool's {@link Tool#canStart} predicate.
 * If that predicate returns true, it starts that tool by making it the
 * diagram's current tool.
 * It then activates the tool and passes on the event to the tool by calling
 * the corresponding method (either {@link Tool#doMouseDown},
 * {@link Tool#doMouseMove}, or {@link Tool#doMouseUp}).
 *
 * Because this tool is typically the one running as the diagram's current tool
 * when the user isn't "doing" anything, this tool can also handle other events,
 * such as mouse wheel events and keyboard commands.
 *
 * Keyboard events are just passed on to the {@link Diagram#commandHandler}'s
 * {@link CommandHandler#doKeyDown} or {@link CommandHandler#doKeyUp} method.
 *
 * This tool also is responsible for showing tooltips.
 * Tooltip {@link Adornment}s may be declared as any {@link GraphObject#toolTip},
 * or as the {@link Diagram#toolTip} if the mouse or finger remains motionless in the background of the diagram.
 * You can set {@link #toolTipDuration} to control how long the tooltip remains visible after being motionless.
 *
 * This tool does not utilize any tool handles.
 * This tool does not conduct any transactions.
 * But of course some of the tools that the ToolManager starts can show tool handles and/or
 * conduct their own transactions.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ToolManager extends Tool {
  /** @hidden @internal */ public static className = 'ToolManager';
  /** @hidden @internal */ private _mouseDownTools: List<Tool>;
  /** @hidden @internal */ private _mouseMoveTools: List<Tool>;
  /** @hidden @internal */ private _mouseUpTools: List<Tool>;
  /** @hidden @internal */ private _hoverDelay: number;
  /** @hidden @internal */ private _holdDelay: number;
  /** @hidden @internal */ private _dragSize: Size;
  /** @hidden @internal */ private _toolTipDuration: number;
  /** @hidden @internal */ private _mouseWheelBehavior: EnumValue;
  /** @hidden @internal */ private _gestureBehavior: EnumValue;
  /** @hidden @internal */ private _currentToolTip: Adornment | HTMLInfo | null;
  /** @hidden @internal */ private _currentToolTipPart: Part | null;
  /** @hidden @internal */ private _toolTipTimer: number;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link Diagram#toolManager}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "ToolManager".
   *
   * The constructor produces a ToolManager that manages no tools.
   * Call {@link #initializeStandardTools} to create various tools,
   * initialize the tool properties such as {@link #draggingTool},
   * and initialize the three mouse tool lists with those newly created tools.
   */
  constructor() {
    super();
    this.name = 'ToolManager';
    this._mouseDownTools = new List<Tool>();
    this._mouseMoveTools = new List<Tool>();
    this._mouseUpTools = new List<Tool>();
    this._hoverDelay = 850;
    this._holdDelay = 850;
    this._dragSize = new Size(2, 2).constant();
    this._toolTipDuration = 5000;

    this._mouseWheelBehavior = ToolManager.WheelScroll;
    this._gestureBehavior = ToolManager.GestureZoom;

    // transient properties
    this._currentToolTip = null;
    this._currentToolTipPart = null;  // needed when tooltip is for an element that is removed from the Part
    this._toolTipTimer = -1;
  }

  /**
   * This default value for {@link #mouseWheelBehavior} indicates that mouse wheel events scroll the diagram.
   * @constant
   */
  public static WheelScroll = new EnumValue(ToolManager, 'WheelScroll', 0);

  /**
   * This value for {@link #mouseWheelBehavior} indicates that the mouse wheel events change the scale of the diagram.
   * @constant
   */
  public static WheelZoom = new EnumValue(ToolManager, 'WheelZoom', 1);

  /**
   * This value for {@link #mouseWheelBehavior} indicates that the mouse wheel events are ignored,
   * although scrolling or zooming by other means may still be allowed.
   * @constant
   * @since 1.2
   */
  public static WheelNone = new EnumValue(ToolManager, 'WheelNone', 2);

  /**
   * This value for {@link #gestureBehavior} indicates that the pointer/touch pinch gestures
   * on the canvas intend to zoom the Diagram.
   * @constant
   * @since 1.5
   */
  public static GestureZoom = new EnumValue(ToolManager, 'GestureZoom', 3);

  /**
   * This value for {@link #gestureBehavior} indicates that the pointer/touch pinch gestures
   * on the canvas intend to have no effect on the Diagram, but also no effect on the page.
   * @constant
   * @since 1.5
   */
  public static GestureCancel = new EnumValue(ToolManager, 'GestureCancel', 4);

  /**
   * This value for {@link #gestureBehavior} indicates that the pointer/touch pinch gestures
   * on the canvas intend to have no effect on the Diagram, but will not be prevented,
   * and may bubble up the page to have other effects (such as zooming the page).
   * @constant
   * @since 1.5
   */
  public static GestureNone = new EnumValue(ToolManager, 'GestureNone', 5);

  /**
   * Gets or sets the ToolManager's mouse wheel behavior. Allowed values are {@link ToolManager.WheelScroll} and
   * {@link ToolManager.WheelZoom} and {@link ToolManager.WheelNone}.
   *
   * The default value is {@link ToolManager.WheelScroll}.
   */
  /** @export */ get mouseWheelBehavior(): EnumValue { return this._mouseWheelBehavior; }
  /** @export */ set mouseWheelBehavior(value: EnumValue) {
    EnumValue.checkEnumValue(value, ToolManager, ToolManager, 'mouseWheelBehavior');
    this._mouseWheelBehavior = value;
  }

  /**
   * Gets or sets the ToolManager's default gestureBehavior behavior. Allowed values are {@link ToolManager.GestureZoom} and
   * {@link ToolManager.GestureCancel} and {@link ToolManager.GestureNone}.
   *
   * The default value is {@link ToolManager.GestureZoom}.
   * @since 1.5
   */
  /** @export */ get gestureBehavior(): EnumValue { return this._gestureBehavior; }
  /** @export */ set gestureBehavior(value: EnumValue) {
    EnumValue.checkEnumValue(value, ToolManager, ToolManager, 'gestureBehavior');
    this._gestureBehavior = value;
  }

  /**
   * Initialize the three mouse tool lists with instances of the standard tools.
   * This adds new instances of tools to the following three mouse tool lists:
   * {@link #mouseDownTools}, {@link #mouseMoveTools}, or {@link #mouseUpTools}.
   * This also sets the various tool properties of this ToolManager
   * to those newly created tools.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public initializeStandardTools(): void {
    // For each of these lists, the tools need to be assigned in a specific order,
    // so that they are added to the list in the correct order.
/*
    // mouse-down tools:
    this.actionTool = new ActionTool();
    this.relinkingTool = new RelinkingTool();
    this.linkReshapingTool = new LinkReshapingTool();
    this.rotatingTool = new RotatingTool();
    this.resizingTool = new ResizingTool();

    // mouse-move tools:
    this.linkingTool = new LinkingTool();
    this.draggingTool = new DraggingTool();
    this.dragSelectingTool = new DragSelectingTool();
    this.panningTool = new PanningTool();

    // mouse-up tools:
    this.contextMenuTool = new ContextMenuTool();
    this.textEditingTool = new TextEditingTool();
    this.clickCreatingTool = new ClickCreatingTool();
    this.clickSelectingTool = new ClickSelectingTool();
    */
  }

  /**
   * @hidden @internal
   * Update any tooltip.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   * @param {Part} part
   */
  public updateAdornments(part: Part): void {
    const tt = this.currentToolTip;
    if (tt instanceof Adornment && this._currentToolTipPart === part) {  // part might be null
      const ttobj = tt.adornedObject;
      // update if adornedObject is still within the part
      if (part !== null ? ttobj.part === part : ttobj === null) {
        this.showToolTip(tt, ttobj);
      } else {
        this.hideToolTip();
      }
    }
  }

  /**
   * Iterate over the {@link #mouseDownTools} list and start the first tool
   * for which its {@link Tool#canStart} predicate returns true.
   *
   * Starting a tool replaces the {@link Diagram#currentTool} with the new tool.
   * Successfully doing so also activates the new tool by calling {@link Tool#doActivate}
   * and passes on the mouse-down event to it by calling {@link Tool#doMouseDown}.
   *
   * Not finding any startable tools causes this tool manager to activate,
   * thereby enabling the mouse-move and mouse-up behaviors and starts
   * detection of a mouse-hold event after {@link #holdDelay} milliseconds.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doMouseDown(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;

    // must happen on the first touch
    if (e.isTouchEvent && this.gestureBehavior === ToolManager.GestureCancel) e.bubbles = false;

    if (e.isMultiTouch) {
      this.cancelWaitAfter(); // second touch should stop context click starting from the first touch
      if (this.gestureBehavior === ToolManager.GestureNone) {
        e.bubbles = true;
        return;
      }

      if (this.gestureBehavior === ToolManager.GestureCancel) {
        return;
      }

      const canstart = diagram.currentTool.canStartMultiTouch();
      if (canstart) {
        // This will call doCancel to stop the rest of toolManager.domousedown/domousemove
        diagram.currentTool.standardPinchZoomStart();
        return;
      }
    }

    // check to see if there is an ongoing transaction
    const mgr = diagram.undoManager;
    if (Debug && mgr.checksTransactionLevel && mgr.transactionLevel !== 0) {
      Util.trace('WARNING: In ToolManager.doMouseDown: UndoManager.transactionLevel is not zero');
    }

    const len = this.mouseDownTools.length;
    for (let i = 0; i < len; i++) {
      const tool = this.mouseDownTools.elt(i);
      tool.diagram = this.diagram;
      if (tool.canStart()) {  // check if applicable
        diagram.doFocus();
        diagram.currentTool = tool;  // calls tool.doStart()
        // if tool hasn't stopped itself already, pass on the mouse event
        if (diagram.currentTool === tool) {
          if (!tool.isActive) tool.doActivate();
          tool.doMouseDown();
        }
        return;
      }
    }

    // Default behavior of scroll wheel click
    if (diagram.lastInput.button === 1) {
      if (this.mouseWheelBehavior === ToolManager.WheelScroll) {
        this.mouseWheelBehavior = ToolManager.WheelZoom;
      } else if (this.mouseWheelBehavior === ToolManager.WheelZoom) {
        this.mouseWheelBehavior = ToolManager.WheelScroll;
      }
    }

    this.doActivate();  // enable doMouseMove and doMouseUp

    this.standardWaitAfter(this.holdDelay, e);  // check for mouseHold event
  }

  /**
   * Iterate over the {@link #mouseMoveTools} list and start the first tool
   * for which its {@link Tool#canStart} predicate returns true.
   *
   * Starting a tool replaces the {@link Diagram#currentTool} with the new tool.
   * Successfully doing so also activates the new tool by calling {@link Tool#doActivate}
   * and passes on the mouse-move event to it by calling {@link Tool#doMouseMove}.
   *
   * If no tool is found and activated, this:
   *   - Calls {@link Tool#standardMouseOver} to invoke any mouse-enter, mouse-over, or mouse-leave functions.
   *   - if the mouse has moved sufficiently, such that {@link Tool#isBeyondDragSize} is true,
   *     this also calls {@link Tool#standardWaitAfter} for a possible call to
   *     {@link #doWaitAfter} after {@link #hoverDelay} milliseconds.
   *   - Sets the {@link Diagram#lastInput} to bubble.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doMouseMove(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;

    if (e.isMultiTouch) {

      if (this.gestureBehavior === ToolManager.GestureNone) {
        e.bubbles = true;
        return;
      }

      if (this.gestureBehavior === ToolManager.GestureCancel) {
        return;
      }

      const canstart = diagram.currentTool.canStartMultiTouch();
      if (canstart) {
        // This will call doCancel to stop the rest of toolManager.domousedown/domousemove
        diagram.currentTool.standardPinchZoomMove();
        return;
      }
    }

    if (this.isActive) {  // gotta handle a mouse-down first
      const len = this.mouseMoveTools.length;
      for (let i = 0; i < len; i++) {
        const tool = this.mouseMoveTools.elt(i);
        tool.diagram = this.diagram;
        if (tool.canStart()) {  // check if applicable
          diagram.doFocus();
          diagram.currentTool = tool;  // calls tool.doStart()
          // if tool hasn't stopped itself already, pass on the mouse event
          if (diagram.currentTool === tool) {
            if (!tool.isActive) tool.doActivate();
            tool.doMouseMove();
          }
          return;  // don't do mouse-over events
        }
      }
    }

    this.standardMouseActivity(diagram);

    // no tool started, let the event bubble? (Added for 1.6.5, amended for pointers in 1.8.6)
    // testing for event.cancelable added for 1.6.13
    // some non-cancellable touch-move events are emitted, and we should not let those bubble
    // 1.6.14: but we do want to let regular mouse-move events bubble, such as when the mouse is simply over a Diagram
    // So if this is a mousemove, or if its non-cancelable (touch, maybe pointer), let it bubble.
    const evt = e.event;
    if (evt !== null && (evt.type === 'mousemove' || evt.type === 'pointermove' || !evt.cancelable)) e.bubbles = true;
  }

  /**
   * @hidden @internal
   * This function calls standardMouseOver for mouse events
   * And standardWaitAfter for hover and hold events
   * @this {ToolManager}
   * @param {Diagram} diagram
   */
  public standardMouseActivity(diagram: Diagram): void {
    this.standardMouseOver();

    if (this.isBeyondDragSize()) {
      // check for mouseHover or mouseHold event
      this.standardWaitAfter((this.isActive ? this.holdDelay : this.hoverDelay), diagram.lastInput);
    }
  }

  /**
   * @hidden @internal
   * This is called by {@link #standardMouseOver} when the current object changes.
   * @this {ToolManager}
   * @param {GraphObject} prev
   * @param {GraphObject} curr
   */
  public doCurrentObjectChanged(prev: GraphObject, curr: GraphObject): void {
    const tt = this.currentToolTip;
    if (tt === null) return;
    // don't remove tooltip if we're now over any part of it
    if (curr !== null && tt instanceof Adornment && (curr === tt || curr.isContainedBy(tt))) return;
    this.hideToolTip();
  }

  /**
   * Implement the standard behavior for when the mouse has not moved for a period of time.
   * This is due to an expired timer started by calling {@link Tool#standardWaitAfter}.
   *
   * This calls {@link #doMouseHover} and, if not mouse-down, {@link #doToolTip}.
   * Afterwards, if the event that called {@link Tool#standardWaitAfter} was a touch event,
   * this simulates a right click, enabling context menu functionality and other actions on touch devices.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   * @param {InputEvent} event The event that caused {@link #standardWaitAfter}. In this override,
   * if it is a touch event and the {@link Diagram#lastInput} event is not handled ({@link InputEvent#handled} set to false),
   * then this method will simulate a right click.
   */
  public doWaitAfter(event: InputEvent): void {
    // It's possible for this to fire after a Diagram has been unloaded
    const diagram = this.diagram;
    if (!diagram._canvas) return;

    this.doMouseHover();
    if (!this.isActive) {  // not when mouse down
      this.doToolTip();
    }

    // A long-tap touch event simulates a right click.
    // If someone handles the last event during mouseHover, don't simulate
    if (event.isTouchEvent && !diagram.lastInput.handled) {
      const evt2 = event.copy();
      evt2.button = 2;
      evt2.buttons = 2;
      diagram.lastInput = evt2;
      diagram._tapHeld = true;
      diagram.doMouseUp();
      return;
    }
  }

  /**
   * Implement the standard behavior for mouse hover and mouse hold events,
   * called by {@link #doWaitAfter} when the mouse has not moved for a period of time.
   *
   * If there has been no mouse down, the timer is set for {@link #hoverDelay} milliseconds.
   * If it executes, it calls any {@link GraphObject#mouseHover} function on the object
   * at the mouse or on any of its containing panels,
   * or it calls any {@link Diagram#mouseHover} function for a background mouse-hover event.
   *
   * If there had been a mouse down, the timer is set for {@link #holdDelay} milliseconds.
   * If it executes, it calls any {@link GraphObject#mouseHold} function on the object
   * at the mouse or on any of its containing panels,
   * or it calls any {@link Diagram#mouseHold} function for a background mouse-held-down event.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doMouseHover(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    if (e.targetObject === null) {
      // try again, just in case it had been off by a pixel or two
      e.targetObject = diagram.findObjectAt(e.documentPoint, null, null);
    }
    let currobj = e.targetObject;

    // this.isActive should be true if there had been a mouse down
    // and no mouse-down tool is running

    if (currobj !== null) {  // now inside some GraphObject
      e.handled = false;
      // look for an event handler for current object that we're over
      while (currobj !== null) {
        const func = (this.isActive ? currobj.mouseHold : currobj.mouseHover);
        if (func !== null) {
          func(e, currobj);
          if (e.handled) break;
        }
        currobj = currobj.panel;
      }
    } else {  // not now in any GraphObject
      // maybe execute an event handler;
      const func2 = (this.isActive ? diagram.mouseHold : diagram.mouseHover);
      if (func2 !== null) {
        func2(e);
      }
    }
  }

  /**
   * Implement the standard behavior for tooltips,
   * called by {@link #doWaitAfter} when the mouse has not moved for a period of time.
   *
   * This looks for a {@link GraphObject} at the latest mouse point.
   * If it finds an object, it checks for a {@link GraphObject#toolTip}.
   * If it has none, this method searches up the visual tree for a containing
   * {@link Panel} that does have a tooltip.
   *
   * If it didn't find any object, this looks for a {@link Diagram#toolTip}.
   *
   * If it eventually finds a tooltip, this calls {@link #showToolTip}.
   * Otherwise this calls {@link #hideToolTip}.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doToolTip(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    if (e.targetObject === null) {
      // try again, just in case it had been off by a pixel or two
      e.targetObject = diagram.findObjectAt(e.documentPoint, null, null);
    }
    let currobj = e.targetObject;

    // this.isActive should be true if there had been a mouse down
    // and no mouse-down tool is running

    if (currobj !== null) {  // now inside some GraphObject
      // don't cancel tooltip if the mouse is now over the tooltip itself
      const currtt = this.currentToolTip;
      if (currtt instanceof Adornment && (currobj === currtt || currobj.isContainedBy(currtt))) return;
      // walk up the visual tree, looking for a tooltip
      while (currobj !== null) {
        const tt = currobj.toolTip;
        if (tt !== null) {
          this.showToolTip(tt, currobj);
          return;
        }
        currobj = currobj.panel;
      }
      // no tooltip found: cancel any current tooltip
      this.hideToolTip();
    } else {  // not now in any GraphObject
      const tt = diagram.toolTip;
      if (tt !== null) {
        this.showToolTip(tt, null);
      } else {
        this.hideToolTip();
      }
    }
  }

  /**
   * Show a tooltip Adornment or HTMLInfo.
   * This is called by {@link #doToolTip} once that method has found a tooltip to display.
   *
   * This calls {@link #positionToolTip} to make it easier to customize how the tooltip
   * is positioned relative to the object with the tooltip.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   * @param {Adornment|HTMLInfo} tooltip
   * @param {GraphObject} obj The GraphObject getting the tooltip; this is null if the tooltip is being shown for the diagram background.
   */
  public showToolTip(tooltip: Adornment | HTMLInfo, obj: GraphObject): void {
    if (Debug && !((tooltip instanceof Adornment) || (tooltip instanceof HTMLInfo))) {
      Util.throwError('showToolTip:tooltip must be an Adornment or HTMLInfo.');
    }

    if (obj !== null) Util.checkClass(obj, GraphObject, ToolManager, 'showToolTip:obj');
    const diagram = this.diagram;

    // hide any different existing tooltip
    if (tooltip !== this.currentToolTip) {
      this.hideToolTip();
    }

    if (tooltip instanceof Adornment) {
      // always use Tool layer
      tooltip.layerName = 'Tool';
      tooltip.selectable = false;
      tooltip.scale = 1 / diagram.scale;
      tooltip.category = 'ToolTip';
      if (tooltip.placeholder !== null) tooltip.placeholder.scale = diagram.scale;

      // show
      const olddiag = tooltip.diagram;
      if (olddiag !== null && olddiag !== diagram) olddiag.remove(tooltip);
      diagram.add(tooltip);

      // maybe data-bind
      if (obj !== null) {
        // this will set tooltip.data automatically
        tooltip.adornedObject = obj;
      } else {  // for diagram background, bind to model
        tooltip.data = diagram.model;
      }

      // position inside the viewport
      tooltip.ensureBounds();
      this.positionToolTip(tooltip, obj);
    } else if (tooltip instanceof HTMLInfo) {
      if (tooltip !== this.currentToolTip) tooltip.show(obj, diagram, this);
    }

    // remember
    this.currentToolTip = tooltip;

    // cancel any previous toolTip timer before starting a new one
    if (this._toolTipTimer !== -1) {
      Util.clearTimeout(this._toolTipTimer);
      this._toolTipTimer = -1;
    }
    // schedule hiding the toolTip after toolTipDuration milliseconds
    const duration = this.toolTipDuration;
    if (duration > 0 && duration !== Infinity) {
      const tool = this;
      this._toolTipTimer = Util.setTimeout(function () { tool.hideToolTip(); }, duration);
    }
  }

  /**
   * This is called by {@link #showToolTip} to position the part within the viewport.
   * It normally goes just below the cursor.
   * But if the mouse is too close to the right edge or the bottom edge of the viewport,
   * it is positioned left and/or above the cursor.
   *
   * This method only operates if the tooltip, an {@link Adornment}, does not have a {@link Adornment#placeholder}.
   * When there is a {@link Placeholder} in the tooltip, that Adornment is automatically positioned
   * so that the Placeholder is positioned at the adorned object, the second argument to this method.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * If you override this method to position the tooltip,
   * the tooltip has already been measured but not arranged,
   * so you can use its {@link GraphObject#measuredBounds} width and height
   * but not its {@link GraphObject#actualBounds}.
   * @expose
   * @this {ToolManager}
   * @param {Adornment} tooltip
   * @param {GraphObject} obj The {@link GraphObject} getting the tooltip,
   * or null if the tooltip is for the diagram background.
   */
  public positionToolTip(tooltip: Adornment, obj: GraphObject): void {
    if (tooltip.placeholder !== null) return;

    const diagram = this.diagram;

    const p = diagram.lastInput.documentPoint.copy();
    const ttb = tooltip.measuredBounds;
    const viewb = diagram.viewportBounds;
    // when touch event -- shift towards the left, so it's not obscured by the finger
    if (diagram.lastInput.isTouchEvent) {
      p.x -= ttb.width;
    }
    // if extends too far to the right -- shift left
    if (p.x + ttb.width > viewb.right) {
      p.x -= ttb.width + 5 / diagram.scale;
    }
    // but don't go beyond the left edge of the viewport
    if (p.x < viewb.x) {
      p.x = viewb.x;
    }
    // if extends too far down -- shift up
    if (p.y + 20 / diagram.scale + ttb.height > viewb.bottom) {
      p.y -= ttb.height + 5 / diagram.scale;
    } else {
      p.y += 20 / diagram.scale;
    }
    // but don't go beyond the top edge of the viewport
    if (p.y < viewb.y) {
      p.y = viewb.y;
    }
    tooltip.position = p;
  }

  /**
   * Hide any tooltip.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public hideToolTip(): void {
    if (this._toolTipTimer !== -1) {
      Util.clearTimeout(this._toolTipTimer);
      this._toolTipTimer = -1;
    }
    const diagram = this.diagram;
    const tt = this.currentToolTip;
    if (tt === null) return;
    if (tt instanceof Adornment) {
      diagram.remove(tt);
      if (this._currentToolTipPart !== null) this._currentToolTipPart.removeAdornment(tt.category);
      tt.data = null;
      tt.adornedObject = null;
    } else if (tt instanceof HTMLInfo) {
      if (tt.hide !== null) tt.hide(diagram, this);
    }
    this.currentToolTip = null;
  }

  /**
   * This read-only property returns the currently showing tooltip, or null if there is none.
   */
  /** @export */ get currentToolTip(): Adornment | HTMLInfo | null { return this._currentToolTip; }
  /** @export */ set currentToolTip(value: Adornment | HTMLInfo | null) {
    if (Debug && value !== null && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
      Util.throwError('ToolManager.currentToolTip must be an Adornment or HTMLInfo.');
    }
    this._currentToolTip = value;
    this._currentToolTipPart = (value !== null && value instanceof Adornment) ? value.adornedPart : null;
  }

  /**
   * Iterate over the {@link #mouseUpTools} list and start the first tool
   * for which its {@link Tool#canStart} predicate returns true.
   *
   * Starting a tool replaces the {@link Diagram#currentTool} with the new tool.
   * Successfully doing so also activates the new tool by calling {@link Tool#doActivate}
   * and passes on the mouse-up event to it by calling {@link Tool#doMouseUp}.
   *
   * If no startable tool is found it deactivates this tool manager,
   * to get ready for a mouse-down and ignore mouse-move and mouse-up events.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doMouseUp(): void {
    this.cancelWaitAfter();
    const diagram = this.diagram;

    if (this.isActive) {  // gotta handle a mouse-down first
      const len = this.mouseUpTools.length;
      for (let i = 0; i < len; i++) {
        const tool = this.mouseUpTools.elt(i);
        tool.diagram = this.diagram;
        if (tool.canStart()) {  // check if applicable
          diagram.doFocus();
          diagram.currentTool = tool;  // calls tool.doStart()
          // if tool hasn't stopped itself already, pass on the mouse event
          if (diagram.currentTool === tool) {
            if (!tool.isActive) tool.doActivate();
            tool.doMouseUp();
          }
          return;
        }
      }
    }
    // No tool, focus the diagram
    diagram.doFocus();
    this.doDeactivate();  // in case no tool is found that canStart
  }

  /**
   * The diagram will call this method as the mouse wheel is rotated.
   *
   * By default this just calls {@link Tool#standardMouseWheel}.
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doMouseWheel(): void {
    this.standardMouseWheel();
  }

  /**
   * This just calls {@link CommandHandler#doKeyDown} on the diagram's {@link Diagram#commandHandler}.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doKeyDown(): void {
    const diagram = this.diagram;
    if (diagram.commandHandler !== null) diagram.commandHandler.doKeyDown();
  }

  /**
   * This just calls {@link CommandHandler#doKeyUp} on the diagram's {@link Diagram#commandHandler}.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   */
  public doKeyUp(): void {
    const diagram = this.diagram;
    if (diagram.commandHandler !== null) diagram.commandHandler.doKeyUp();
  }

  /**
   * Find a mouse tool of a given name.
   * This searches the {@link #mouseDownTools}, {@link #mouseMoveTools}, and {@link #mouseUpTools} lists.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   * @param {string} name the type of tool, such as "Dragging" or "ClickSelecting".
   * @return {Tool} a {@link Tool} whose {@link Tool#name} exactly matches the given name,
   * or null if no such tool is found in any of the three lists.
   */
  public findTool(name: string): Tool | null {
    Util.checkPrimitive(name, 'string', ToolManager, 'findTool:name');

    let len = this.mouseDownTools.length;
    for (let i = 0; i < len; i++) {
      const tool = this.mouseDownTools.elt(i);
      if (tool.name === name) return tool;
    }

    len = this.mouseMoveTools.length;
    for (let i = 0; i < len; i++) {
      const tool = this.mouseMoveTools.elt(i);
      if (tool.name === name) return tool;
    }

    len = this.mouseUpTools.length;
    for (let i = 0; i < len; i++) {
      const tool = this.mouseUpTools.elt(i);
      if (tool.name === name) return tool;
    }
    return null;
  }

  /**
   * Replace a mouse tool of a given name with a new tool, or remove an existing tool (if the *newtool* is null).
   * This searches the {@link #mouseDownTools}, {@link #mouseMoveTools},
   * and {@link #mouseUpTools} lists.
   * The new tool is inserted into the same list in which the same-named tool is found,
   * at the same position as the old tool.
   * However, if no existing tool with the given name is present, this does **not** add the new tool to any list, since it cannot know where it should be added.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ToolManager}
   * @param {string} name the type of tool, such as "Dragging" or "ClickSelecting".
   * @param {Tool} newtool If null, any tool that the search finds will just be removed
   * from the list in which it was found.
   * @return {Tool} the old tool that was replaced by the new one; this is null if none was found and the new tool was not added to any mouse tool list
   */
  public replaceTool(name: string, newtool: Tool): Tool {
    Util.checkPrimitive(name, 'string', ToolManager, 'replaceTool:name');
    if (newtool !== null) {
      Util.checkClass(newtool, Tool, ToolManager, 'replaceTool:newtool');
      newtool.diagram = this.diagram;
    }

    let len = this.mouseDownTools.length;
    for (let i = 0; i < len; i++) {
      const oldtool = this.mouseDownTools.elt(i);
      if (oldtool.name === name) {
        if (newtool !== null) {
          this.mouseDownTools.setElt(i, newtool);
        } else {
          this.mouseDownTools.removeAt(i);
        }
        return oldtool;
      }
    }

    len = this.mouseMoveTools.length;
    for (let i = 0; i < len; i++) {
      const oldtool = this.mouseMoveTools.elt(i);
      if (oldtool.name === name) {
        if (newtool !== null) {
          this.mouseMoveTools.setElt(i, newtool);
        } else {
          this.mouseMoveTools.removeAt(i);
        }
        return oldtool;
      }
    }

    len = this.mouseUpTools.length;
    for (let i = 0; i < len; i++) {
      const oldtool = this.mouseUpTools.elt(i);
      if (oldtool.name === name) {
        if (newtool !== null) {
          this.mouseUpTools.setElt(i, newtool);
        } else {
          this.mouseUpTools.removeAt(i);
        }
        return oldtool;
      }
    }
    return null;  // not found, thus newtool was not added to any list
  }

  // Since 2.1
  /**
   * @hidden
   * (Undocumented)
   * @export
   * @this {ToolManager}
   * @param {string} name
   * @param {Tool} newtool
   * @param {List.<Tool>} list
   */
  public replaceStandardTool(name: string, newtool: Tool, list: List<Tool>): void {
    Util.checkPrimitive(name, 'string', ToolManager, 'replaceStandardTool:name');
    Util.checkClass(list, List, ToolManager, 'replaceStandardTool:list');

    if (newtool !== null) {
      Util.checkClass(newtool, Tool, ToolManager, 'replaceStandardTool:newtool');
      newtool.name = name;
      newtool.diagram = this.diagram;
    }

    if (this.findTool(name)) {  // found an existing tool, regardless of list
      this.replaceTool(name, newtool);  // newtool may be null
    } else if (newtool !== null) {
      list.add(newtool);
    }
  }

  /**
   * This read-only property returns the list of {@link Tool}s that might be started upon a mouse or finger press event.
   * When the ToolManager handles a mouse-down or touch-down event in {@link #doMouseDown},
   * it searches this list in order, starting the first tool for which
   * {@link Tool#canStart} returns true.
   *
   * This list may be modified, but it must not be modified while any tool
   * is handling events.
   *
   * {@link #initializeStandardTools} installs the following tools, in order:
   *   - {@link #actionTool}, an {@link ActionTool}
   *   - {@link #relinkingTool}, a {@link RelinkingTool}
   *   - {@link #linkReshapingTool}, a {@link LinkReshapingTool}
   *   - {@link #rotatingTool}, a {@link RotatingTool}
   *   - {@link #resizingTool}, a {@link ResizingTool}
   */
  /** @export */ get mouseDownTools(): List<Tool> { return this._mouseDownTools; }

  /**
   * This read-only property returns the list of {@link Tool}s that might be started upon a mouse or finger move event.
   * When the ToolManager handles a mouse-move or touch-move event in {@link #doMouseMove},
   * it searches this list in order, starting the first tool for which
   * {@link Tool#canStart} returns true.
   *
   * This list may be modified, but it must not be modified while any tool
   * is handling events.
   *
   * {@link #initializeStandardTools} installs the following tools, in order:
   *   - {@link #linkingTool}, a {@link LinkingTool}
   *   - {@link #draggingTool}, a {@link DraggingTool}
   *   - {@link #dragSelectingTool}, a {@link DragSelectingTool}
   *   - {@link #panningTool}, a {@link PanningTool}
   */
  /** @export */ get mouseMoveTools(): List<Tool> { return this._mouseMoveTools; }

  /**
   * This read-only property returns the list of {@link Tool}s that might be started upon a mouse or finger up event.
   * When the ToolManager handles a mouse-up or touch-up event in {@link #doMouseUp},
   * it searches this list in order, starting the first tool for which
   * {@link Tool#canStart} returns true.
   *
   * This list may be modified, but it must not be modified while any tool
   * is handling events.
   *
   * {@link #initializeStandardTools} installs the following tools, in order:
   *   - {@link #contextMenuTool}, a {@link ContextMenuTool}
   *   - {@link #textEditingTool}, a {@link TextEditingTool}
   *   - {@link #clickCreatingTool}, a {@link ClickCreatingTool}
   *   - {@link #clickSelectingTool}, a {@link ClickSelectingTool}
   */
  /** @export */ get mouseUpTools(): List<Tool> { return this._mouseUpTools; }

  /**
   * Gets or sets the time between when the mouse stops moving and a hover event,
   * in milliseconds. This value affects the delay before {@link GraphObject#toolTip}s are shown.
   * The default value is 850 milliseconds.
   */
  /** @export */ get hoverDelay(): number { return this._hoverDelay; }
  /** @export */ set hoverDelay(value: number) {
    Util.checkPrimitive(value, 'number', ToolManager, 'hoverDelay');
    this._hoverDelay = value;
  }

  /**
   * Gets or sets the time between when the mouse stops moving and a hold event,
   * in milliseconds.
   * The default value is 850 milliseconds.
   */
  /** @export */ get holdDelay(): number { return this._holdDelay; }
  /** @export */ set holdDelay(value: number) {
    Util.checkPrimitive(value, 'number', ToolManager, 'holdDelay');
    this._holdDelay = value;
  }

  /**
   * Gets or sets the distance in view coordinates within which a mouse down-and-up is considered a click and beyond which a mouse movement is considered a drag.
   * The default value is 2 pixels horizontally and vertically for mouse events, and increases by 6 pixels for touch events.
   * This value is used by {@link Tool#isBeyondDragSize}.
   * @since 1.2
   */
  /** @export */ get dragSize(): Size { return this._dragSize; }
  /** @export */ set dragSize(value: Size) {
    Util.checkClass(value, Size, ToolManager, 'dragSize');
    this._dragSize = value.copyFrozen();
  }

  /**
   * Gets or sets how long a tool tip is visible.
   * The default value is 5000 milliseconds.
   *
   * This is used by {@link #showToolTip} to determine how long to wait before calling {@link #hideToolTip}.
   */
  /** @export */ get toolTipDuration(): number { return this._toolTipDuration; }
  /** @export */ set toolTipDuration(value: number) {
    Util.checkPrimitive(value, 'number', ToolManager, 'toolTipDuration');
    this._toolTipDuration = value;
  }

  // Standard mode-less tools

  /**
   * Gets or sets the mode-less {@link ActionTool}, normally one of the {@link #mouseDownTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get actionTool(): ActionTool { return this.findTool('Action') as ActionTool; }
  // %DOC% /** @export */ set actionTool(value: ActionTool) { }

  /**
   * Gets or sets the mode-less {@link RelinkingTool}, normally one of the {@link #mouseDownTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get relinkingTool(): RelinkingTool { return this.findTool('Relinking') as RelinkingTool; }
  // %DOC% /** @export */ set relinkingTool(value: RelinkingTool) { }

  /**
   * Gets or sets the mode-less {@link LinkReshapingTool}, normally one of the {@link #mouseDownTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get linkReshapingTool(): LinkReshapingTool { return this.findTool('LinkReshaping') as LinkReshapingTool; }
  // %DOC% /** @export */ set linkReshapingTool(value: LinkReshapingTool) { }

  /**
   * Gets or sets the mode-less {@link ResizingTool}, normally one of the {@link #mouseDownTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get resizingTool(): ResizingTool { return this.findTool('Resizing') as ResizingTool; }
  // %DOC% /** @export */ set resizingTool(value: ResizingTool) { }

  /**
   * Gets or sets the mode-less {@link RotatingTool}, normally one of the {@link #mouseDownTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get rotatingTool(): Tool { return this.findTool('Rotating') as RotatingTool; }
  // %DOC% /** @export */ set rotatingTool(value: Tool) { }

  /**
   * Gets or sets the mode-less {@link LinkingTool}, normally one of the {@link #mouseMoveTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get linkingTool(): LinkingTool { return this.findTool('Linking') as LinkingTool; }
  // %DOC% /** @export */ set linkingTool(value: LinkingTool) { }

  /**
   * Gets or sets the mode-less {@link DraggingTool}, normally one of the {@link #mouseMoveTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get draggingTool(): DraggingTool { return this.findTool('Dragging') as DraggingTool; }
  // %DOC% /** @export */ set draggingTool(value: DraggingTool) { }

  /**
   * Gets or sets the mode-less {@link DragSelectingTool}, normally one of the {@link #mouseMoveTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get dragSelectingTool(): DragSelectingTool { return this.findTool('DragSelecting') as DragSelectingTool; }
  // %DOC% /** @export */ set dragSelectingTool(value: DragSelectingTool) { }

  /**
   * Gets or sets the mode-less {@link PanningTool}, normally one of the {@link #mouseMoveTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get panningTool(): PanningTool { return this.findTool('Panning') as PanningTool; }
  // %DOC% /** @export */ set panningTool(value: PanningTool) { }

  /**
   * Gets or sets the mode-less {@link ContextMenuTool}, normally one of the {@link #mouseUpTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get contextMenuTool(): ContextMenuTool { return this.findTool('ContextMenu') as ContextMenuTool; }
  // %DOC% /** @export */ set contextMenuTool(value: ContextMenuTool) { }

  /**
   * Gets or sets the mode-less {@link TextEditingTool}, normally one of the {@link #mouseUpTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get textEditingTool(): TextEditingTool { return this.findTool('TextEditing') as TextEditingTool; }
  // %DOC% /** @export */ set textEditingTool(value: TextEditingTool) { }

  /**
   * Gets or sets the mode-less {@link ClickCreatingTool}, normally one of the {@link #mouseUpTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get clickCreatingTool(): ClickCreatingTool { return this.findTool('ClickCreating') as ClickCreatingTool; }
  // %DOC% /** @export */ set clickCreatingTool(value: ClickCreatingTool) { }

  /**
   * Gets or sets the mode-less {@link ClickSelectingTool}, normally one of the {@link #mouseUpTools}.
   *
   * You can disable this tool by setting its {@link Tool#isEnabled} property to false.
   */
  // %DOC% get clickSelectingTool(): ClickSelectingTool { return this.findTool('ClickSelecting') as ClickSelectingTool; }
  // %DOC% /** @export */ set clickSelectingTool(value: ClickSelectingTool) { }

} // #end ToolManager
ToolManager['className'] = 'ToolManager';

ToolManager['WheelScroll'] = ToolManager.WheelScroll;
ToolManager['WheelZoom'] = ToolManager.WheelZoom;
ToolManager['WheelNone'] = ToolManager.WheelNone;
ToolManager['GestureZoom'] = ToolManager.GestureZoom;
ToolManager['GestureCancel'] = ToolManager.GestureCancel;
ToolManager['GestureNone'] = ToolManager.GestureNone;

// import { List } from './collections';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { HTMLInfo } from './htmlInfo';
// import { InputEvent } from './inputEvent';
// import { Adornment, Part } from './parts';
// import { Size } from './size';
// import { Util } from './util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
DraggingTool
*/

// import { List, Map, Set } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Diagram, DraggingInfo, DraggingOptions } from './diagram';
// import { GraphObject } from './graphObject';
// import { InputEvent } from './inputEvent';
// import { RelinkingTool } from './linkingTools';
// import { Panel } from './panel';
// import { Node } from './parts';
// import { Adornment, Part } from './parts';
// import { Link } from './parts';
// import { Group } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Tool } from './tool';
// import { Util } from './util';

/**
 * The DraggingTool is used to move or copy selected parts with the mouse.
 * This sets the {@link Part#location} property; you may want to save the location to the model
 * by using a TwoWay {@link Binding} on the "location" property in your Parts/Nodes/Groups templates.
 *
 * Dragging the selection moves parts for which {@link Part#canMove} is true.
 * If the user holds down the Control key (Option key on Mac), this tool will make a copy of the parts being dragged,
 * for those parts for which {@link Part#canCopy} is true.
 *
 * When the drag starts it calls {@link #computeEffectiveCollection} to find the actual collection
 * of {@link Part}s to be dragged.
 * Normally this collection includes not only the {@link Diagram#selection}, but also parts that belong
 * to those selected parts, such as members of groups.
 * If {@link #dragsTree} is true, the effective collection also includes all of the nodes and links that
 * constitute the subtree starting from selected nodes.
 * The result of {@link #computeEffectiveCollection} is not a {@link Set} but a {@link Map}
 * which remembers the original {@link Part#location} for all of the dragged parts.
 * This map is saved as the value of {@link #draggedParts}.
 *
 * During the drag if the user holds down the Control/Option key this tool makes a copy of the {@link #draggedParts}
 * and proceeds to drag it around.
 * (It only copies the {@link Diagram#selection}, not the whole effective collection,
 * if {@link #copiesEffectiveCollection} is false.)
 * The collection of copied parts is held by {@link #copiedParts}.
 * It too is a {@link Map} remembering the original locations of the parts.
 * {@link #copiedParts} will be null when this tool is moving (not copying) at the moment.
 *
 * Each Part's movement is limited by the {@link Diagram#computeMove} method.
 * By default it limits the {@link Part#location} to be within the bounds given by {@link Part#minLocation} and {@link Part#maxLocation}.
 * (Those default to minus Infinity to plus Infinity.)
 * As a further convenience, the value of NaN in minLocation and maxLocation cause {@link Diagram#computeMove} to use the
 * part's current location.
 * So, for example, an easy way to declare that the user may only drag a node horizontally is to just set:
 * ```js
 * $(go.Node,
 *   . . .
 *   { minLocation: new go.Point(-Infinity, NaN), maxLocation: new go.Point(Infinity, NaN) },
 *   . . .
 * )
 * ```
 *
 * If you set {@link #isGridSnapEnabled} to true, dragged or copied parts will be snapped to points on a grid.
 * The snapping occurs continuously during a drag unless you set {@link #isGridSnapRealtime} to false.
 * Normally the grid points come from the {@link Diagram#grid}, even if that grid is not {@link GraphObject#visible}.
 * However you can override those grid's properties for the snapping grid cell size and offset
 * by setting the properties here: {@link #gridSnapCellSize} and {@link #gridSnapOrigin}.
 * This computes the point to snap to for each dragged part.
 * The resulting point is used as the new {@link Part#location}.
 *
 * For the most general control over where a part may be dragged, either set the {@link Part#dragComputation} property
 * or override {@link Diagram#computeMove}.
 * For the common case of wanting to keep member nodes within the {@link Group} that they are members of,
 * you can do something like:
 * ```js
 *  // this is a Part.dragComputation function for limiting where a Node may be dragged
 *   function stayInGroup(part, pt, gridpt) {
 *     // don't constrain top-level nodes
 *     var grp = part.containingGroup;
 *     if (grp === null) return pt;
 *     // try to stay within the background Shape of the Group
 *     var back = grp.resizeObject;
 *     if (back === null) return pt;
 *     // allow dragging a Node out of a Group if the Shift key is down
 *     if (part.diagram.lastInput.shift) return pt;
 *     var p1 = back.getDocumentPoint(go.Spot.TopLeft);
 *     var p2 = back.getDocumentPoint(go.Spot.BottomRight);
 *     var b = part.actualBounds;
 *     var loc = part.location;
 *     // find the padding inside the group's placeholder that is around the member parts
 *     var m = grp.placeholder.padding;
 *     // now limit the location appropriately
 *     var x = Math.max(p1.x + m.left, Math.min(pt.x, p2.x - m.right - b.width - 1)) + (loc.x-b.x);
 *     var y = Math.max(p1.y + m.top, Math.min(pt.y, p2.y - m.bottom - b.height - 1)) + (loc.y-b.y);
 *     return new go.Point(x, y);
 *   }
 * ```
 * Note that this expects there to be a "SHAPE" object within the Group's visual tree that delimits
 * where the part may be dragged within the group.
 * This also expects that {@link Group#computesBoundsIncludingLinks} is false.
 * Then in your node template(s), just set:
 * ```js
 * $(go.Node,
 *   . . .,
 *   { dragComputation: stayInGroup },
 *   . . .
 * )
 * ```
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 * If the drag is successful, it raises the "SelectionMoved" or "SelectionCopied" {@link DiagramEvent}
 * and produces a "Move" or a "Copy" transaction.
 *
 * If you want to programmatically start a new user's dragging of a particular existing node,
 * you can make sure that node is selected, set the {@link #currentPart} property, and then start and activate the tool.
 * ```js
 *   var node = ...;
 *   myDiagram.select(node);        // in this case the only selected node
 *   var tool = myDiagram.toolManager.draggingTool;
 *   tool.currentPart = node;       // the DraggingTool will not call standardMouseSelect
 *   myDiagram.currentTool = tool;  // starts the DraggingTool
 *   tool.doActivate();             // activates the DraggingTool
 * ```
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class DraggingTool extends Tool {
  /** @hidden @internal */ public static className = 'DraggingTool';
  /** @hidden @internal */ private _isCopyEnabled: boolean;
  /** @hidden @internal */ private _copiesEffectiveCollection: boolean;
  /** @hidden @internal */ private _relinkingTool: RelinkingTool;
  /** @hidden @internal */ private _currentPart: Part | null;
  /** @hidden @internal */ private _draggedParts: Map<Part, DraggingInfo> | null;
  /** @hidden @internal */ private _copiedParts: Map<Part, DraggingInfo> | null;
  /** @hidden @internal */ private _isDragOutStarted: boolean;
  /** @hidden @internal */ private _startPanelPoint: Point;
  /** @hidden @internal */ private _startPoint: Point;
  /** @hidden @internal */ private _isComplexRoutingRealtime: boolean;
  /** @hidden @internal */ private _delay: number;
  /** @hidden @internal */ private _dragOverStack: Array<GraphObject>;
  /** @hidden @internal */ private _emptySet: Set<Part>;
  /** @hidden @internal */ private _dragOptions: DraggingOptions;
  /** @hidden @internal */ public _dropped: boolean;
  /** @hidden @internal */ private _copyCursor: string;
  /** @hidden @internal */ private _moveCursor: string;
  /** @hidden @internal */ private _nodropCursor: string;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#draggingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "Dragging".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(DraggingTool);

    this.name = 'Dragging';
    this._isCopyEnabled = true;
    this._copiesEffectiveCollection = true;
    this._relinkingTool = null;
    this._currentPart = null;
    this._draggedParts = null;
    this._copiedParts = null;
    this._isDragOutStarted = false;
    this._dropped = false;
    this._startPanelPoint = new Point(NaN, NaN);
    this._startPoint = new Point(); // the mouse point from which parts start to move
    this._isComplexRoutingRealtime = true;
    this._delay = 100;
    this._dragOverStack = [];
    this._emptySet = new Set<Part>().freeze();
    this._dragOptions = new DraggingOptions();
    this._copyCursor = 'copy';
    this._moveCursor = '';
    this._nodropCursor = 'no-drop';
  }

  /**
   * Gets or sets whether for any internal copying operation is permitted by control-drag-and-drop.
   * This property affects the behavior of {@link #mayCopy},
   * but does not affect whether copied objects may be dropped into this diagram from a different diagram.
   *
   * The default value is true.
   * Setting this property does not raise any events.
   * @since 1.4
   */
  /** @export */ get isCopyEnabled(): boolean { return this._isCopyEnabled; }
  /** @export */ set isCopyEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'isCopyEnabled');
    this._isCopyEnabled = value;
  }

  /**
   * Gets or sets whether for a copying operation the extended selection
   * is copied or only the selected parts.
   * The default value is true.
   * Setting this property does not raise any events.
   *
   * The {@link CommandHandler#copiesConnectedLinks} property serves a similar role for the {@link CommandHandler#copySelection} command,
   * when the user types control-C to copy the currently selected parts.
   */
  /** @export */ get copiesEffectiveCollection(): boolean { return this._copiesEffectiveCollection; }
  /** @export */ set copiesEffectiveCollection(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'copiesEffectiveCollection');
    this._copiesEffectiveCollection = value;
  }

  /**
   * Gets or sets the {@link DraggingTool}'s {@link DraggingOptions} instance, which controls several dragging properties.
   *
   * Several DraggingTool properties are just convenience properties:
   *   - {@link isGridSnapEnabled}
   *   - {@link isGridSnapRealtime}
   *   - {@link gridSnapCellSize}
   *   - {@link gridSnapCellSpot}
   *   - {@link gridSnapOrigin}
   *   - {@link dragsLink}
   *   - {@link dragsTree}
   *
   * Setting any of these properties really sets the corresponding dragOptions property.
   *
   * Setting this property does not raise any events.
   * @since 2.0
   */
  /** @export */ get dragOptions(): DraggingOptions {
    return this._dragOptions;
  }
  /** @export */ set dragOptions(value: DraggingOptions) {
    Util.checkClass(value, DraggingOptions, DraggingTool, 'dragOptions');
    this._dragOptions = value;
  }

  /**
   * Gets or sets whether the {@link DraggingTool} snaps objects to grid points.
   * Whether the snapping movement of the dragged parts occurs during the drag or only upon a drop
   * is determined by the value of {@link #isGridSnapRealtime}.
   *
   * This property does not affect dragging disconnected links,
   * but those links to respect the {@link Part#dragComputation}, which can be used to snap them.
   *
   * By default this property is false.
   * Setting this property does not raise any events.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   */
  /** @export */ get isGridSnapEnabled(): boolean {
    return this.dragOptions.isGridSnapEnabled;
  }
  /** @export */ set isGridSnapEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'isGridSnapEnabled');
    this.dragOptions.isGridSnapEnabled = value;
  }

  /**
   * Gets or sets whether link routing takes some short-cuts during dragging.
   * When false Links whose routing is AvoidsNodes are not routed to avoid Nodes, in order to improve dragging performance.
   * The default value is true.
   * @since 1.4
   */
  /** @export */ get isComplexRoutingRealtime(): boolean { return this._isComplexRoutingRealtime; }
  /** @export */ set isComplexRoutingRealtime(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'isComplexRoutingRealtime');
    this._isComplexRoutingRealtime = value;
  }

  /**
   * Gets or sets whether the {@link DraggingTool} snaps objects to grid points during the drag.
   * This property is ignored unless {@link #isGridSnapEnabled} is true.
   * By default this property is true; when false parts are only snapped to grid locations upon the drop (i.e. mouse-up).
   * Setting this property does not raise any events.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   * @since 1.1
   */
  /** @export */ get isGridSnapRealtime(): boolean {
    return this.dragOptions.isGridSnapRealtime;
  }
  /** @export */ set isGridSnapRealtime(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'isGridSnapRealtime');
    this.dragOptions.isGridSnapRealtime = value;
  }

  /**
   * Gets or sets the size of the grid cell used when snapping during a drag
   * if the value of {@link #isGridSnapEnabled} is true.
   * By default this property is the Size(NaN, NaN), which causes this tool to use the {@link Panel#gridCellSize} value of the {@link Diagram#grid}.
   * Setting this property does not raise any events.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   */
  /** @export */ get gridSnapCellSize(): Size {
    return this.dragOptions.gridSnapCellSize;
  }
  /** @export */ set gridSnapCellSize(value: Size) {
    Util.checkClass(value, Size, DraggingTool, 'gridSnapCellSize');
    if (this.diagram === null) return;
    const old = this.dragOptions.gridSnapCellSize;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this.dragOptions.gridSnapCellSize = value;
    }
  }

  /**
   * Gets or sets the {@link Spot} that specifies what point in the grid cell dragged parts snap to,
   * if the value of {@link #isGridSnapEnabled} is true.
   * By default this property is {@link Spot.TopLeft}:
   * node locations will snap exactly to the grid point.
   * Setting this property does not raise any events.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   */
  /** @export */ get gridSnapCellSpot(): Spot {
    return this.dragOptions.gridSnapCellSpot;
  }
  /** @export */ set gridSnapCellSpot(value: Spot) {
    Util.checkClass(value, Spot, DraggingTool, 'gridSnapCellSpot');
    const old = this.dragOptions.gridSnapCellSpot;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this.dragOptions.gridSnapCellSpot = value;
    }
  }

  /**
   * Gets or sets the snapping grid's origin point, in document coordinates,
   * if the value of {@link #isGridSnapEnabled} is true.
   * By default this property is the Point(NaN, NaN),
   * which causes this tool to use the {@link Panel#gridOrigin} value from the {@link Diagram#grid}.
   * Setting this property does not raise any events.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   */
  /** @export */ get gridSnapOrigin(): Point {
    return this.dragOptions.gridSnapOrigin;
  }
  /** @export */ set gridSnapOrigin(value: Point) {
    Util.checkClass(value, Point, DraggingTool, 'gridSnapOrigin');
    const old = this.dragOptions.gridSnapOrigin;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this.dragOptions.gridSnapOrigin = value;
    }
  }

  /**
   * Gets or sets whether the user can drag a single {@link Link},
   * disconnecting it from its connected nodes and possibly connecting it
   * to valid ports when the link is dropped.
   * The default value is false.
   * Setting this property does not raise any events.
   *
   * In order to avoid too many cases of having both ends of a dragged Link connect to the same node (if allowed),
   * it is commonplace to decrease the {@link LinkingBaseTool#portGravity} to a smaller value such as 10 or 20.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   * @since 1.3
   */
  /** @export */ get dragsLink(): boolean {
    return this.dragOptions.dragsLink;
  }
  /** @export */ set dragsLink(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'dragsLink');
    this.dragOptions.dragsLink = value;
  }

  /**
   * Gets or sets whether moving or copying a node also includes all of the
   * node's tree children and their descendants, along with the links to those additional nodes.
   * The default value is false.
   * Setting this property does not raise any events.
   *
   * The {@link CommandHandler#copiesTree} property serves a similar role for the {@link CommandHandler#copySelection} command,
   * when the user types control-C to copy the currently selected parts.
   *
   * This property is a convenience getter/setter, and sets a value on {@link dragOptions}.
   */
  /** @export */ get dragsTree(): boolean {
    return this.dragOptions.dragsTree;
  }
  /** @export */ set dragsTree(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DraggingTool, 'dragsTree');
    this.dragOptions.dragsTree = value;
  }

  /** @hidden @internal */
  /** @export */ get copyCursor() { return this._copyCursor; }
  /** @export */ set copyCursor(c: string) { this._copyCursor = c; }
  /** @hidden @internal */
  /** @export */ get moveCursor() { return this._moveCursor; }
  /** @export */ set moveCursor(c: string) { this._moveCursor = c; }
  /** @hidden @internal */
  /** @export */ get nodropCursor() { return this._nodropCursor; }
  /** @export */ set nodropCursor(c: string) { this._nodropCursor = c; }

  /**
   * Gets the {@link Part} found at the mouse point.
   * This is normally set by a call to {@link #standardMouseSelect}.
   */
  /** @export */ get currentPart(): Part | null { return this._currentPart; }
  /** @export */ set currentPart(value: Part | null) {
    if (value !== null) Util.checkClass(value, Part, DraggingTool, 'currentPart');
    this._currentPart = value;
  }

  /**
   * Gets the collection of {@link Part}s that this tool has copied.
   * The value is a {@link Map} mapping {@link Part}s to DraggingInfo Objects
   * that have a "point" property remembering the original location of that Part.
   * The value is null when moving instead of copying.
   *
   * {@link #draggedParts} provides the map of Parts that are being moved and from which this collection was copied.
   */
  /** @export */ get copiedParts(): Map<Part, DraggingInfo> | null { return this._copiedParts; }
  /** @export */ set copiedParts(value: Map<Part, DraggingInfo> | null) { this._copiedParts = value; }

  /**
   * Gets the collection of {@link Part}s being moved.
   * The value is a {@link Map} mapping {@link Part}s to DraggingInfo Objects
   * that have a "point" property remembering the original location of that Part.
   *
   * {@link #copiedParts} provides the map of Parts that have been copied during a copying operation, if any.
   */
  /** @export */ get draggedParts(): Map<Part, DraggingInfo> | null { return this._draggedParts; }
  /** @export */ set draggedParts(value: Map<Part, DraggingInfo> | null) { this._draggedParts = value; }

  /**
   * @hidden
   * This read-only property returns a {@link Set} that holds all of the Parts that are currently being dragged for either copying or moving.
   * @return {Set.<Part>} Returns the {@link Map#toKeySet} of either {@link #copiedParts} or {@link #draggedParts}, or else an empty Set.
   */
  /** @export */ get draggingParts(): Set<Part> {
    if (this.copiedParts !== null) return this.copiedParts.toKeySet();
    if (this.draggedParts !== null) return this.draggedParts.toKeySet();
    return this._emptySet;
  }

  /**
   * @hidden @internal
   * Gets or sets the {@link Link} that may be dragged and reconnected
   * when {@link DraggingTool#dragsLink} is true.
   *
   * This property is set during {@link DraggingTool#doActivate}.
   * The default value is null.
   * Setting this property does not raise any events.
   */
  /** @export */ get draggedLink(): Link {
    return this.diagram.draggedLink;
  }
  /** @export */ set draggedLink(value: Link) {
    if (value !== null) Util.checkClass(value, Link, DraggingTool, 'draggedLink');
    this.diagram.draggedLink = value;
  }

  /**
   * @hidden @internal
   * Gets or sets whether a potentially external drag-and-drop has been started by this tool.
   *
   * This property is set to true when starting a drag-and-drop in {@link DraggintTool#doDragOut}.
   *
   * The default value is false.
   * Setting this property does not raise any events.
   */
  /** @export */ get isDragOutStarted(): boolean { return this._isDragOutStarted; }
  /** @export */ set isDragOutStarted(value: boolean) { this._isDragOutStarted = value; }

  /**
   * Gets or sets the mouse point from which parts start to move.
   * The value is a Point in document coordinates.
   * This property is normally set to the diagram's mouse-down point in {@link #doActivate},
   * but may be set to a different point if parts are being copied from a different control.
   * Setting this property does not raise any events.
   */
  /** @export */ get startPoint(): Point { return this._startPoint; }
  /** @export */ set startPoint(value: Point) {
    Util.checkClass(value, Point, DraggingTool, 'startPoint');
    const old = this._startPoint;
    if (!old.equals(value)) {
      this._startPoint.assign(value);
    }
  }

  /**
   * On touch gestures only, this property gets or sets the time in milliseconds
   * for which the mouse must be stationary before this tool can be started.
   * The default value is 100 milliseconds.
   * Setting this property does not raise any events.
   */
  /** @export */ get delay(): number { return this._delay; }
  /** @export */ set delay(value: number) {
    Util.checkPrimitive(value, 'number', DraggingTool, 'delay');
    this._delay = value;
  }

  /**
   * This tool can run if the diagram allows selection and moves/copies/dragging-out,
   * if the mouse has moved far enough away to be a drag and not a click,
   * and if {@link #findDraggablePart} has found a selectable part at the mouse-down point.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;

    const diagram = this.diagram;
    // allow dragging out of Palettes and other isReadOnly Diagrams,
    // but don't allow this tool to start otherwise
    if (diagram.isReadOnly && !diagram.allowDragOut) return false;
    if (!diagram.allowMove && !diagram.allowCopy && !diagram.allowDragOut) return false;
    if (!diagram.allowSelect) return false;

    // require left button & that it has moved far enough away from the mouse down point,
    // so it isn't a click
    const e = diagram.lastInput;
    if (!e.left) return false;
    // don't include the following check when this tool is running modally
    if (diagram.currentTool !== this) {
      if (!this.isBeyondDragSize()) return false;
      if (e.isTouchEvent) {
        // must wait for "delay" milliseconds before that tool can run
        if (e.timestamp - diagram.firstInput.timestamp < this._delay) return false;
      }
    }

    // need to be over a part that is movable or copyable
    const part = this.findDraggablePart();
    return (part !== null);
  }

  /**
   * Return the selectable and movable/copyable {@link Part} at the mouse-down point.
   * This is called by {@link #canStart} to decide if this tool is ready to run.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @return {Part}
   */
  public findDraggablePart(): Part | null {
    const diagram = this.diagram;
    // to select containing Group if Part.canSelect() is false, allow nonselectable
    let part = diagram.findPartAt(diagram.firstInput.documentPoint, false);
    if (part === null) return null;
    while (part !== null && !part.canSelect()) part = part.containingGroup;
    if (part !== null && (part.canMove() || part.canCopy())) return part;
    return null;
  }

  /**
   * This override prevents the Control modifier unselecting an already selected part.
   * This also remembers the selectable {@link #currentPart} at the current mouse point.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   */
  public standardMouseSelect(): void {
    const diagram = this.diagram;
    if (!diagram.allowSelect) return;
    // to select containing Group if Part.canSelect() is false, allow nonselectable
    let part = diagram.findPartAt(diagram.firstInput.documentPoint, false);
    if (part === null) return;
    while (part !== null && !part.canSelect()) part = part.containingGroup;
    this.currentPart = part;

    // change behavior of StandardMouseSelect because we don't want the Control modifier
    // to unselect an already selected object.
    if (this.currentPart !== null && !this.currentPart.isSelected) {
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      const e = diagram.lastInput;
      if (!(Util.isMac ? e.meta : e.control) && !e.shift) {
        diagram.internalClearSelection();  // don't raise ChangingSelection/Finished
      }
      this.currentPart.isSelected = true;
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
    }
  }

  /**
   * Start the dragging operation.
   * This calls {@link #computeEffectiveCollection} and saves the result as {@link #draggedParts}.
   *
   * This starts a "Drag" transaction.
   * Depending on what happens, the transaction may be finished as a "Move" or a "Copy"
   * transaction, or it may be rolled-back if the tool is cancelled.
   *
   * Normally when this method is called the value of {@link #currentPart} will be null,
   * in which case this will call {@link Tool#standardMouseSelect} which will set {@link #currentPart}.
   * But if when this method is called the value of {@link #currentPart} has already been set
   * because the programmer wants the user to start dragging that Part,
   * then this method will not need to call {@link Tool#standardMouseSelect} because
   * the Part(s) to be selected and dragged have already been determined by the caller.
   * @this {DraggingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this.currentPart === null) this.standardMouseSelect();
    const mainnode = this.currentPart;
    if (mainnode === null || (!mainnode.canMove() && !mainnode.canCopy())) return;

    DraggingTool._source = null; // this used to be in Diagram's mouseDown
    this.isActive = true;
    this._startPanelPoint.set(diagram.position);
    this._checkForDraggedLink(diagram.selection);
    this._dragOverStack.length = 0;
    const coll = this.computeEffectiveCollection(diagram.selection, this.dragOptions);
    this.draggedParts = coll;

    diagram.skipsBoundsUpdates = true;
    if (diagram.getRenderingHint('temporaryPixelRatio') === true && diagram.avgSpf > 30) diagram.enableTempPixelRatio();

    diagram._suspendRouting(this.draggedParts);
    this.startTransaction('Drag');
    this.startPoint = diagram.firstInput.documentPoint;
    diagram.isMouseCaptured = true;
    if (diagram.allowDragOut) {
      this.doDragOut();
    }
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {Iterable.<Part>} parts
   */
  private _checkForDraggedLink(parts: Iterable<Part>): void {
    if (this.dragsLink) {
      const diagram = this.diagram;
      if (!diagram.allowRelink) return;
      const isGLModel = diagram.model.isGraphLinksModel();
      if (isGLModel && parts.count === 1 && parts.first() instanceof Link) {
        this.draggedLink = parts.first() as Link;
        if (this.draggedLink.canRelinkFrom() && this.draggedLink.canRelinkTo()) {
          this.draggedLink.clearAdornments();
        }
        this._relinkingTool = diagram.toolManager.findTool('Relinking') as RelinkingTool;
        if (this._relinkingTool === null) {
          this._relinkingTool = new RelinkingTool();
          this._relinkingTool.diagram = diagram;
        }
      } else {
        this.draggedLink = null;
        this._relinkingTool = null;
      }
    }
  }

  /**
   * This just calls {@link CommandHandler#computeEffectiveCollection}. The implementation of this method was moved to CommandHandler for 2.0.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @param {Iterable.<Part>} parts A {@link Set} or {@link List} of {@link Part}s.
   * @param {DraggingOptions} options Potential options for the collection computation
   * @return {Map.<Part,DraggingInfo>}  a Map mapping {@link Part}s to DraggingInfo Objects
   *         that have a "point" property remembering the original location of that Part.
   */
  public computeEffectiveCollection(parts: Iterable<Part>, options: DraggingOptions): Map<Part, DraggingInfo> {
    return this.diagram.commandHandler.computeEffectiveCollection(parts, options);
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {Point=} p
   * @return {DraggingInfo}
   */
  private _newDraggingInfo(p?: Point): DraggingInfo {
    if (p === undefined) return new DraggingInfo(Point.Origin);
    // When snapping, we should round so link routing stays suspended:
    // a.location.x = 99.999999, b.location.x = 200.000001 would get snapped differently when moving half the grid cell width
    if (this.isGridSnapEnabled) {
      return new DraggingInfo(new Point(Math.round(p.x), Math.round(p.y)));
    }
    return new DraggingInfo(p.copy());
  }

  /**
   * Stop the dragging operation by stopping the transaction and cleaning up any temporary state.
   * @this {DraggingTool}
   */
  public doDeactivate(): void {
    this.isActive = false;
    const diagram = this.diagram;
    diagram.stopAutoScroll();
    this._stopEnterLeave();
    diagram._resumeRouting(this.draggedParts);
    this.currentPart = null;
    this.draggedParts = null;
    this.isDragOutStarted = false;
    this._dropped = false;
    DraggingTool.cleanUpDraggingTool();
    this._cleanUpDraggedLink();
    this._startPanelPoint.assignTo(NaN, NaN);
    /** @hidden @internal */
    DraggingTool._currentDiagram = null;
    /** @hidden @internal */
    DraggingTool._source = null;
    this.removeCopies();
    diagram.isMouseCaptured = false;
    diagram.currentCursor = '';  // back to defaultCursor

    diagram.skipsBoundsUpdates = false;
    this.stopTransaction();
    diagram.disableTempPixelRatio(true); // will no-op if not turned on
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private _stopEnterLeave(): void {
    const diagram = this.diagram;
    const oldskip = diagram.skipsUndoManager;
    diagram.skipsUndoManager = true;
    const e = diagram.lastInput;
    this.enterDragOverStack(e, null);
    diagram.skipsUndoManager = oldskip;
    this._dragOverStack.length = 0;
  }

  /**
   * @hidden @internal
   * Called by doSimulatedDragEnter to restore state in the source diagram without stopping the DraggingTool.
   * @this {DraggingTool}
   */
  public _sourceCancel(): void {
    this.removeCopies();
    this._restoreOriginalLocations();
    const diagram = this.diagram;
    if (this._startPanelPoint.isReal()) {
      diagram.position = this._startPanelPoint;
    }
    diagram.stopAutoScroll();
  }

  /**
   * Abort any dragging operation.
   * @this {DraggingTool}
   */
  public doCancel(): void {
    this.removeCopies();
    this._restoreOriginalLocations();
    const diagram = this.diagram;
    if (this._startPanelPoint.isReal()) {
      diagram.position = this._startPanelPoint;
    }
    this.stopTool();
  }

  // DURING A DRAG, either moving or copying:

  /**
   * Handle switching between copying and moving modes as the Control/Option key is pressed or released.
   * @this {DraggingTool}
   */
  public doKeyDown(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    if (!this.isActive) return;
    if (e.key === 'Esc') {
      this.doCancel();
    } else {
      this.doMouseMove();
    }
  }

  /**
   * Handle switching between copying and moving modes as the Control/Option key is pressed or released.
   * @this {DraggingTool}
   */
  public doKeyUp(): void {
    if (!this.isActive) return;
    this.doMouseMove();
  }

  // DURING A DRAG: copying parts

  /**
   * @hidden @internal
   * used in DraggingTool
   * @this {Diagram}
   * @param {Set.<Part>|Map.<Part,Part>} nodes a collection of {@link Part}s
   * @param {Rect} result
   * @return {Rect}
   */
  private computeLocations(nodes: Set<Part> | Map<Part, Part>, result: Rect): Rect {
    let minx = Infinity;
    let miny = Infinity;
    let maxx = -Infinity;
    let maxy = -Infinity;
    const itr = nodes.iterator;
    while (itr.next()) {
      const n = itr.value;
      if (!n.isNodeLike()) continue; // only nodes
      if (!n.isVisible()) continue; // don't count collapsed nodes, etc
      const l = n.location;
      const x = l.x;
      const y = l.y;
      if (isNaN(x) || isNaN(y)) continue;
      if (x < minx) minx = x;
      if (y < miny) miny = y;
      if (x > maxx) maxx = x;
      if (y > maxy) maxy = y;
    }
    if (minx === Infinity) {
      result.assignTo(0, 0, 0, 0);
    } else {
      result.assignTo(minx, miny, maxx - minx, maxy - miny);
    }
    return result;
  }

  /**
   * @hidden @internal
   * When !undoable, calls Diagram.copyParts with Layer = 'Tool'
   * @this {DraggingTool}
   * @param {boolean} undoable
   */
  private addCopies(undoable: boolean) {
    if (this.copiedParts !== null) return;
    const diagram = this.diagram;
    // allow temporary copies even if diagram or model is read only
    if (undoable && (diagram.isReadOnly || diagram.isModelReadOnly)) return;
    if (this.draggedParts === null) return;
    const mgr = diagram.undoManager;
    if (mgr.isEnabled && mgr.isInTransaction) {
      // don't bother rolling back if there's nothing to roll back
      if (mgr.currentTransaction !== null && mgr.currentTransaction.changes.count > 0) {
        // undo all changes and restart the transaction
        diagram.undoManager.rollbackTransaction();
        diagram.startTransaction('Drag');
      }
    } else {  // if no UndoManager recording, just move the dragged Parts back
      this._restoreOriginalLocations();
    }
    diagram.skipsUndoManager = !undoable;
    diagram.partManager.addsToTemporaryLayer = !undoable; // temp copies go in Tool layer

    this.startPoint = diagram.firstInput.documentPoint;

    const srccoll = this.copiesEffectiveCollection ? this.draggedParts.toKeySet() : diagram.selection;
    const copieddata = diagram.copyParts(srccoll, diagram, true);

    // Copy the locations over separately, in case they are not in the data
    const it = copieddata.iterator;
    while (it.next()) {
      const v = it.value;
      const k = it.key;
      v.location = k.location;
    }

    const bounds = Rect.alloc();
    this.computeLocations(copieddata, bounds);
    // var midx = bounds.x + bounds.width / 2;
    // var midy = bounds.y + bounds.height / 2;
    Rect.free(bounds);

    const copiedparts = new Map<Part, DraggingInfo>();
    const nit = this.draggedParts.iterator;
    while (nit.next()) {
      const n = nit.key;
      if (n.isNodeLike() && n.canCopy()) {
        const c = copieddata.getValue(n);
        if (c === null) continue;
        // someday we will want to be more careful than just duplicating the location
        // this causes the copied node's location to exist at the mouse point:
        // var l = n.location;
        // c.location = new Point(loc.x - (midx - l.x), loc.y - (midy - l.y));
        c.ensureBounds();
        copiedparts.add(c, this._newDraggingInfo(c.location));
      }
    }
    const lit = copieddata.iterator;
    while (lit.next()) {
      const p = lit.value;
      if (p instanceof Link) {
        if (p.canCopy()) {
          copiedparts.add(p, this._newDraggingInfo());
        }
      }
    }
    this.copiedParts = copiedparts;

    this._checkForDraggedLink(copiedparts.toKeySet());
    if (this.draggedLink !== null) {
      const route = this.draggedLink;
      const b = route.routeBounds;
      route.movePoints(this.startPoint.x - (b.x + b.width / 2),
        this.startPoint.y - (b.y + b.height / 2));
    }

  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private removeCopies(): void {
    const diagram = this.diagram;
    if (this.copiedParts !== null) {
      diagram.removeParts(this.copiedParts.toKeySet(), false);
      this.copiedParts = null;
      // this fixes up the dragging of non-copied links when their link points
      // have been restored due to a rollbackTransaction:
      if (this.draggedParts !== null) {
        const it = this.draggedParts.iterator;
        while (it.next()) {
          const part = it.key;
          if (part instanceof Link) {
            const info = it.value;
            info.point = new Point(0, 0);
          }
        }
      }
    }
    diagram.skipsUndoManager = false;
    diagram.partManager.addsToTemporaryLayer = false;
    this.startPoint = diagram.firstInput.documentPoint;
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private _cleanUpDraggedLink(): void {
    if (this.draggedLink !== null) {
      if (this.dragsLink && this._relinkingTool !== null) {
        this._relinkingTool.stopDraggingMouseMove();
      }
      this.draggedLink = null;
      this._relinkingTool = null;
    }
  }

  // DURING A DRAG: moving parts (perhaps copied parts)

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {Map.<Part,DraggingInfo>} parts
   * @param {boolean} check
   */
  private moveCollection(parts: Map<Part, DraggingInfo>, check: boolean): void {
    const diagram = this.diagram;
    const oldpos = this.startPoint;
    const newpos = Point.alloc();
    newpos.assign(diagram.lastInput.documentPoint);
    this.moveParts(parts, newpos.subtract(oldpos), check);
    Point.free(newpos);
    // If the temp pixel ratio is not on, perhaps turn it on
    if (diagram.getRenderingHint('temporaryPixelRatio') === true && diagram._tempPixelRatio === null && diagram.avgSpf > 30) {
      diagram.enableTempPixelRatio();
      diagram.redrawLater(); // but also stop it in the near future
    }

  }

  /**
   * Move a collection Map of {@link Part}s by a given offset.
   *
   * If *check* is true this respects the {@link Part#canMove} predicate for {@link Node}s or simple {@link Part}s
   * when this is the {@link Diagram#currentTool}.
   * It also respects {@link #isGridSnapEnabled} in order to try to automatically snap part locations to a grid.
   * And it also uses the {@link Part#dragComputation} function, if any, to determine the new location for each part.
   *
   * The first argument is a {@link Map} as produced by {@link #computeEffectiveCollection},
   * not a {@link List} or {@link Set} or {@link Iterator} of Parts.
   * Call {@link Diagram#moveParts} if you want to move a simple collection of Parts without having to create the argument Map.
   * @expose
   * @this {DraggingTool}
   * @param {Map.<Part,DraggingInfo>} parts  a Map mapping {@link Part}s to DraggingInfo Objects
   *        that have a "point" property remembering the original location of that Part.
   * @param {Point} offset The offset, before snapping, to move parts.
   * This offset reflects the total amount moved during tool operation,
   * based on original Part locations remembered when the DraggingTool activated.
   * @param {boolean} check  Whether to check {@link Part#canMove} on each part.
   * @since 1.1
   */
  public moveParts(parts: Map<Part, DraggingInfo>, offset: Point, check: boolean): void {
    const diagram = this.diagram;
    if (diagram !== null) diagram.moveParts2(parts, offset, this.dragOptions, check);
  }

  // move everything back to their original locations, respecting no rerouting of links in DraggedParts
  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private _restoreOriginalLocations(): void {
    if (this.draggedParts === null) return;
    const diagram = this.diagram;
    let kvpit = this.draggedParts.iterator;
    while (kvpit.next()) {
      const n = kvpit.key;
      if (n.isNodeLike()) {
        n.location = kvpit.value.point;
      }
    }
    kvpit = this.draggedParts.iterator;
    while (kvpit.next()) {
      const p = kvpit.key;
      if (p instanceof Link) {
        if (p.suspendsRouting) {
          const oldlinkoff = kvpit.value.point;
          this.draggedParts.add(p, this._newDraggingInfo());
          p.movePoints(-oldlinkoff.x, -oldlinkoff.y);
        }
      }
    }
    diagram.maybeUpdate();  // needed to update link routes before skipping undo manager
  }

  // DURING A DRAG: affecting stationary parts

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {GraphObject} obj
   * @return {boolean}
   */
  private _isDraggedObject(obj: GraphObject): boolean {
    if (obj === null) return true;
    const part = obj.part;
    if (part === null) return true;
    if (part instanceof Adornment) return true;
    if (part.layer.isTemporary) return true;
    if (this.draggedParts && this.draggedParts.contains(part)) return true;
    if (this.copiedParts && this.copiedParts.contains(part)) return true;
    return false;
  }

  /**
   * @hidden @internal
   * Affect some stationary unselected parts that are under a given point, and consider auto-scrolling.
   *
   * Look for a {@link GraphObject} at the given point.
   * If it is a different object, this calls {@link GraphObject#mouseDragLeave} (if it is a function) on the previous object
   * and {@link GraphObject#mouseDragEnter} (if it is a function) on the new object.
   * If there is no object at the given point, this calls {@link Diagram#mouseDragOver} if that property is a function.
   *
   * This method also performs auto-scrolling, if {@link Diagram#allowScroll} is true.
   * @this {DraggingTool}
   * @param {Point} pt a Point in document coordinates.
   */
  private internalDragOver(pt: Point): void {
    const diagram = this.diagram;

    if (this.dragsLink) {
      // disconnect the link from any ports
      if (this.draggedLink !== null) {
        this.draggedLink.fromNode = null;
        this.draggedLink.toNode = null;
      }
      // then see if either or both ends are over any valid ports
      this._updateDraggedLinkOverPorts(false);
    }

    const tool = this;
    const currobj = diagram.findObjectAtNoTemp(pt, null,
      function (o) { return !tool._isDraggedObject(o); });
    const e = diagram.lastInput;
    e.targetObject = currobj;

    const oldskip2 = diagram.skipsUndoManager;
    let anyfunc = false;
    try {
      diagram.skipsUndoManager = true;

      // call mouseDragLeave and mouseDragEnter handlers
      anyfunc = this.enterDragOverStack(e, currobj);
      // what if a mouseDragEnter cancelled the tool?
      if (!this.isActive && DraggingTool._source === null) return;

      // not now in any GraphObject
      // there is no GrahpObject.mouseDragOver, so we don't need to check for its existence here, at least for now
      if (currobj === null || diagram.handlesDragDropForTopLevelParts) {
        const func = diagram.mouseDragOver;
        if (func !== null) {
          func(e);
          anyfunc = true;
        }
      }
      // what if a mouseDragOver cancelled the tool?
      if (!this.isActive && DraggingTool._source === null) return;

      this.doDragOver(pt, currobj);
      // what if an override cancelled the tool?
      if (!this.isActive && DraggingTool._source === null) return;
    } finally {
      diagram.skipsUndoManager = oldskip2;
      if (anyfunc) diagram.maybeUpdate();
    }

    // only autoscroll this diagram if there's actually somewhere the dragged parts can be dropped
    if (!diagram.isReadOnly && (diagram.allowMove || diagram.allowCopy) &&
        (diagram.allowHorizontalScroll || diagram.allowVerticalScroll)) {
      diagram.doAutoScroll(e.viewPoint);
    }
  }

  /**
   * @hidden @internal
   * Walk up the visual tree or containing Groups and push them on the stack, outermost first.
   * @this {DraggingTool}
   * @param {InputEvent} e  the current InputEvent
   * @param {GraphObject} curr  the GraphObject at the current mouse point, might be null
   * @return {boolean} true if any event handler was called
   */
  private enterDragOverStack(e: InputEvent, curr: GraphObject): boolean {
    let anyfunc = false;  // was any event handler called?
    let len = this._dragOverStack.length;
    // remember the previously entered object, if any
    const prev = len > 0 ? this._dragOverStack[0] : null;
    // still in same object, or still in background?  no-op!
    if (curr === prev) return false;

    // iterate over all old elements, starting inside and going outward
    e.handled = false;
    for (let i = 0; i < len; i++) {
      const elt = this._dragOverStack[i];
      const func = elt.mouseDragLeave;
      if (func !== null) {
        func(e, elt, curr);
        anyfunc = true;
        if (e.handled) break;  // if HANDLED, stop calling event handlers
      }
    }
    this._dragOverStack.length = 0;  // clear stack

    // what if a mouseDragLeave cancelled the tool?
    if (!this.isActive && DraggingTool._source === null) return anyfunc;
    if (curr === null) return anyfunc;

    e.handled = false;
    // now maybe enter some elements
    // first push them onto the stack
    let obj = curr;
    while (obj !== null) {
      this._dragOverStack.push(obj);
      obj = this.findPanelOrGroup(obj);
    }
    // then iterate over all new elements, starting inside and going outward
    len = this._dragOverStack.length;
    for (let i = 0; i < len; i++) {
      const elt = this._dragOverStack[i];
      const func = elt.mouseDragEnter;
      if (func !== null) {
        func(e, elt, prev);
        anyfunc = true;
        if (e.handled) break;  // if HANDLED, stop calling event handlers
      }
    }

    return anyfunc;  // true if some event handler was called
  }

  /**
   * @hidden @internal
   * Support bubbling of drag-and-drop events from member Parts to containing Groups.
   * @this {DraggingTool}
   * @param {GraphObject} obj
   * @param {GraphObject} panel  normally a Panel, might be a Group
   * @return {boolean}
   */
  public isInPanelOrGroup(obj: GraphObject, panel: GraphObject): boolean {
    if (obj === null) return false;
    if (obj === panel) return true;
    if (obj.isContainedBy(panel)) return true;
    if (panel instanceof Group && !(obj instanceof Group)) {
      if (panel.handlesDragDropForMembers && obj.part.isMemberOf(panel)) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * Support bubbling of drag-and-drop events from member Parts to containing Groups.
   * @this {DraggingTool}
   * @param {GraphObject} obj
   * @return {Panel} either a simple Panel or a Group
   */
  private findPanelOrGroup(obj: GraphObject): Panel | null {
    const pan = obj.panel;
    if (pan !== null) {
      return pan;
    } else if (obj instanceof Part && !(obj instanceof Group)) {
      const group = obj.containingGroup;
      if (group !== null && group.handlesDragDropForMembers) {
        return group;
      }
    }
    return null;
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {Point} p
   * @param {boolean} toend
   * @return {GraphObject}
   */
  private _findClosestValidPort(p: Point, toend: boolean): GraphObject {
    const diagram = this.diagram;
    const tool = this._relinkingTool;
    if (tool === null) return null;
    const gravity = tool.portGravity;
    const nearports = diagram.findObjectsNear(p, gravity,
      function (obj) { return tool.findValidLinkablePort(obj, toend); });
    const temp = Point.alloc();
    let bestDist = Infinity;
    let bestPort: GraphObject = null;
    for (const it = nearports.iterator; it.next(); ) {
      const port = it.value;
      const node = port.part;
      if (node === null) continue;
      const toPoint = port.getDocumentPoint(Spot.Center, temp);  // ?? assumes center point of port
      const dist = p.distanceSquaredPoint(toPoint);
      if (dist < bestDist) {
        bestPort = port;
        bestDist = dist;
      }
    }
    Point.free(temp);
    return bestPort;
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   * @param {boolean} commit
   */
  private _updateDraggedLinkOverPorts(commit: boolean): void {
    const link = this.draggedLink;
    if (link !== null) {
      if (link.pointsCount < 2) return;
      const diagram = this.diagram;
      if (diagram.isReadOnly) return;
      const tool = this._relinkingTool;
      if (tool === null) return;
      // see if either (or both) ends of the link are over a valid port
      let maybefromport: GraphObject = null;
      let maybefromnode: Node = null;
      if (link.fromNode === null) {
        maybefromport = this._findClosestValidPort(link.getPoint(0), false);
        if (maybefromport !== null) maybefromnode = maybefromport.part as Node;
      }
      let maybetoport: GraphObject = null;
      let maybetonode: Node = null;
      if (link.toNode === null) {
        maybetoport = this._findClosestValidPort(link.getPoint(link.pointsCount - 1), true);
        if (maybetoport !== null) maybetonode = maybetoport.part as Node;
      }
      const ok = tool.isValidLink(maybefromnode, maybefromport, maybetonode, maybetoport);
      if (ok) {
        if (commit) {
          link.defaultFromPoint = link.getPoint(0);
          link.defaultToPoint = link.getPoint(link.pointsCount - 1);
          link.suspendsRouting = false;
          link.fromNode = maybefromnode;
          if (maybefromport !== null) link.fromPortId = maybefromport.portId;
          link.toNode = maybetonode;
          if (maybetoport !== null) link.toPortId = maybetoport.portId;

          if (link.fromPort !== diagram.oldFromPort) {
            diagram.raiseDiagramEvent('LinkRelinked', link, diagram.oldFromPort);
          }
          if (link.toPort !== diagram.oldToPort) {
            diagram.raiseDiagramEvent('LinkRelinked', link, diagram.oldToPort);
          }
        } else {
          tool.doDraggingMouseMove(maybefromnode, maybefromport, maybetonode, maybetoport);
        }
      } else {
        tool.doDraggingMouseMove(null, null, null, null);
      }
    }
  }

  /**
   * Perform any additional side-effects during a drag, whether an internal move or copy or an external drag,
   * that may affect the existing non-moved object(s).
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @param {Point} pt a Point in document coordinates.
   * @param {GraphObject} obj the {@link GraphObject} at the point,
   * excluding what is being dragged or temporary objects;
   * the argument may be null if the drag is occurring in the background of the diagram.
   * Use {@link GraphObject#part} to get the {@link Node} or {@link Part} at the root of
   * the visual tree of the stationary object.
   * @since 1.1
   */
  public doDragOver(pt: Point, obj: GraphObject | null): void {
  }

  // UPON A DROP:

  /**
   * @hidden @internal
   * Perform any additional side-effects after a drop.
   *
   * This looks for a {@link GraphObject} at the given point.
   * If it finds one, this calls its {@link GraphObject#mouseDrop} value if it is a function.
   * Otherwise this calls the {@link Diagram#mouseDrop} value if it is a function.
   *
   * This also reroutes "AvoidsNodes" links that intersect with the area(s) where the selection is dropped.
   * @this {DraggingTool}
   * @param {Point} pt a Point in document coordinates.
   */
  private internalDropOnto(pt: Point): void {
    const diagram = this.diagram;

    if (this.dragsLink) this._updateDraggedLinkOverPorts(true);

    // pretend the mouse has moved out of the current object, to clean up any side-effects
    this._stopEnterLeave();

    const tool = this;
    const currobj = diagram.findObjectAtNoTemp(pt, null,
      function (o) { return !tool._isDraggedObject(o); });
    const e = diagram.lastInput;
    e.targetObject = currobj;

    if (currobj !== null) {
      e.handled = false;
      let obj = currobj;
      // look for mouseDrop event handlers up chain of panels
      while (obj !== null) {
        const func = obj.mouseDrop;
        if (func !== null) {
          func(e, obj);
          if (e.handled) break;
        }
        this.maybeDiagramMouseDrop(e, obj); // potentially trigger the diagram's mouseDrop
        obj = this.findPanelOrGroup(obj);
      }
    } else {
      // maybe execute a mouseDrop event handler for the Diagram background
      const func2 = diagram.mouseDrop;
      if (func2 !== null) {
        func2(e);
      }
    }
    // what if a mouseDrop cancelled the tool?
    if (!this.isActive && DraggingTool._source === null) return;

    // in case doDropOnto shifts nodes in a manner other than normal, make sure any connected links are rerouted
    const nitr = (this.copiedParts || this.draggedParts).iterator;
    while (nitr.next()) {
      const n = nitr.key;
      if (n instanceof Node) {
        // what about non-visible links within collapsed groups or collapsed subtrees
        n.linksConnected.each(function (l) { l.suspendsRouting = false; });
      }
    }

    this.doDropOnto(pt, currobj);
    // what if an override cancelled the tool?
    if (!this.isActive && DraggingTool._source === null) return;

    // after dropping a Node, see if it overlaps with any AvoidsNodes link that might need to be re-routed
    const temp = Rect.alloc();
    const itr = diagram.selection.iterator;
    while (itr.next()) {
      const part = itr.value;
      if (part instanceof Node) {
        this.invalidateAvoidsNodesLinks(diagram, part, temp);
      }
    }
    Rect.free(temp);
  }

  /**
   * @hidden @internal
   * If the object is a top-level part that isn't a group, maybe run the diagram's mouseDrop.
   * Called only by DraggingTool.internalDropOnto.
   * @this {DraggingTool}
   * @param {InputEvent} e
   * @param {GraphObject} obj
   */
  private maybeDiagramMouseDrop(e: InputEvent, obj: GraphObject) {
    const diagram = this.diagram;
    // if Diagram.handlesDragDropForTopLevelParts, maybe do the diagram's mouseDrop
    if (obj instanceof Part && obj.containingGroup === null &&
        !(obj instanceof Group) && diagram.handlesDragDropForTopLevelParts) {
      // maybe execute a mouseDrop event handler for the Diagram background
      const func2 = diagram.mouseDrop;
      if (func2 !== null) {
        func2(e);
      }
    }
  }

  /**
   * @hidden @internal
   * Called only by DraggingTool.internalDropOnto
   * @this {DraggingTool}
   * @param {Diagram} diagram
   * @param {Node} node
   * @param {Part} region re-used rect
   */
  private invalidateAvoidsNodesLinks(diagram: Diagram, node: Node, region: Rect): void {
    let inView = false;
    node.getAvoidableRect(region);
    if (diagram.viewportBounds.containsRect(region)) {
      inView = true;
    }
    // If findObjectsIn was faster, we should use that instead
    const links = diagram.links;
    const parts = this.copiedParts || this.draggedParts;
    while (links.next()) {
      const link = links.value;
      if (inView && !link.getInView()) continue;
      if (parts.contains(link) && parts.contains(node)) continue;
      if (!link.isMemberOf(node) && link.isAvoiding && link.actualBounds.intersectsRectPlus(region, 0)) {
        link.invalidateRoute();
      }
    }
  }

  /**
   * Perform any additional side-effects after a drop, whether an internal move or copy or an external drop,
   * that may affect the existing non-moved object(s).
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @param {Point} pt a Point in document coordinates.
   * @param {GraphObject} obj the {@link GraphObject} where the drop occurred,
   * excluding what was dropped or temporary objects;
   * the argument may be null if the drop occurred in the background of the diagram.
   * Use {@link GraphObject#part} to get the {@link Node} or {@link Part} at the root of
   * the visual tree of the stationary object.
   * @since 1.1
   */
  public doDropOnto(pt: Point, obj: GraphObject | null): void {
  }

  // Internal mouse drag and drop

  /**
   * Move the {@link #draggedParts} (or if copying, the {@link #copiedParts}) to follow the current mouse point.
   *
   * If this creates any temporary parts, by default it adds them to the Tool layer.
   *
   * This calls {@link #doDragOver} for any side-effects on stationary parts.
   * @this {DraggingTool}
   */
  public doMouseMove(): void {
    if (!this.isActive) return;
    const diagram = this.diagram;

    const e = diagram.lastInput;
    if (this.simulatedMouseMove(e.event, null, e.targetDiagram || null)) return;


    const mainnode = this.currentPart;

    if (mainnode !== null && this.draggedParts !== null) {
      // handle copying by making copy immediately, and dragging it
      if (this.mayCopy()) {
        diagram.currentCursor = this.copyCursor;
        this.addCopies(false);  // false means it doesn't get recorded in UndoManager
        diagram._suspendRouting(this.copiedParts);
        this.moveCollection(this.copiedParts, false);  // don't check for Part.canMove()
        diagram._resumeRouting(this.copiedParts);
      } else if (this.mayMove()) {
        const curs = diagram.currentCursor;
        if (curs === this.copyCursor || curs === this.nodropCursor) diagram.currentCursor = this.moveCursor;
        this.removeCopies();  // get rid of any copies, in case COPYING had been true before
        this.moveCollection(this.draggedParts, true);
      } else if (this.mayDragOut()) {  // can't copy or move within the diagram, but may copy out
        diagram.currentCursor = this.nodropCursor;
        this.addCopies(false);  // false means it doesn't get recorded in UndoManager
        this.moveCollection(this.copiedParts, false);  // don't check for Part.canMove()
      } else {
        this.removeCopies();
      }

      this.internalDragOver(diagram.lastInput.documentPoint);
    }
  }

  /**
   * On a mouse-up finish moving or copying the effective selection.
   *
   * This calls {@link #doDropOnto} for any side-effects on stationary parts.
   *
   * This also updates the diagram's bounds,
   * raises a "SelectionCopied" or "SelectionMoved" DiagramEvent, and stops this tool.
   *
   * This method also raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed in a "Drag" transaction, but the "ChangedSelection" event is raised outside the transaction.
   * @this {DraggingTool}
   */
  public doMouseUp(): void {
    if (!this.isActive) return;
    const diagram = this.diagram;

    const e = diagram.lastInput;
    if (this.simulatedMouseUp(e.event, null, e.documentPoint, e.targetDiagram)) return;

    let maybeUpdateGroups = false;

    const copying = this.mayCopy();
    if (copying && this.copiedParts !== null) {
      this.removeCopies();
      this.addCopies(true);  // new collection of this.copiedParts
      diagram._suspendRouting(this.copiedParts);
      this.moveCollection(this.copiedParts, false);  // don't check for Part.canMove()
      diagram._resumeRouting(this.copiedParts);
      if (this.copiedParts !== null) {
        diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
        diagram.internalClearSelection();
        this.copiedParts.iteratorKeys.each((p: Part) => {
          p.isSelected = true;
        });
      }
    } else {
      maybeUpdateGroups = true;
      this.removeCopies();  // get rid of any copies, in case COPYING had been true before
      // one last dragOver event, this time not skipping the undo manager
      if (this.mayMove()) {
        this.moveCollection(this.draggedParts, true);
        this.internalDragOver(diagram.lastInput.documentPoint);
      }
    }
    this._dropped = true;
    // consider side-effects on the "target" object, if any
    this.internalDropOnto(diagram.lastInput.documentPoint);

    if (this.isActive) {
      const rescoll = (copying ? this.copiedParts.toKeySet() : this.draggedParts.toKeySet());
      this.copiedParts = null;
      if (maybeUpdateGroups) this._maybeUpdateContainingGroups();

      diagram.invalidateDocumentBounds();
      // call this before any DiagramEvent listener executes
      diagram._resumeRouting(this.draggedParts);
      // set the EditResult before raising event, in case it changes the result or cancels the tool
      this.transactionResult = (copying ? 'Copy' : 'Move');
      diagram.raiseDiagramEvent(copying ? 'SelectionCopied' : 'SelectionMoved', rescoll);
    }
    this.stopTool();
    if (copying) diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
  }


  //  * Called in mouseMove/touchMove events, returns true or false depending on
  //  * whether or not the mouse move is over another diagram (dragging tool target)
  //  * that could recieve a simulated mouse move.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {DraggingTool}
   * @param {Event} e
   * @param {Point} modelpt
   * @param {Diagram} overdiag the diagram attached to the target of the mouse move event, if any.
   * @return {boolean}
   */
  public simulatedMouseMove(e: Event | Touch | null, modelpt: Point, overdiag: Diagram | null): boolean {
    if (DraggingTool._source === null) return false;
    const originDiagram = DraggingTool._source.diagram;

    // Dragging tool is active but we still may
    // wish to return false (go on normally) if the mouse is still in initial diagram
    if (!(overdiag instanceof Diagram)) overdiag = null;
    const olddiag = DraggingTool._currentDiagram;

    // changed current diagram
    if (overdiag !== olddiag) {
      // leaving a different diagram

      if (olddiag !== null && olddiag !== originDiagram) {
        olddiag.stopAutoScroll();
        DraggingTool._source.isDragOutStarted = false;
        const olddiagtool = olddiag.toolManager.findTool('Dragging') as DraggingTool;
        if (olddiagtool !== null) olddiagtool.doSimulatedDragLeave();
      }
      DraggingTool._currentDiagram = overdiag;
      // entering a different diagram

      if (overdiag !== null && overdiag !== originDiagram) {
        DraggingTool._source._sourceCancel();
        const overdiagtool = overdiag.toolManager.findTool('Dragging') as DraggingTool;
        if (overdiagtool !== null) {
          overdiagtool.rememberDraggingTool();
          overdiagtool.doSimulatedDragEnter();
        }
      }
    }

    if (overdiag === null || overdiag === originDiagram ||
      // If the diagram we're over does not allow drop, then treat it as if it were empty space
      !overdiag.allowDrop || overdiag.isReadOnly || !overdiag.allowInsert) {
      return false;
    }

    const target = overdiag.toolManager.findTool('Dragging') as DraggingTool;
    if (target !== null) {

      let docpos = modelpt;
      if (e !== null) {
        if ((e as any)['targetTouches'] !== undefined) {
          if ((e as TouchEvent).targetTouches.length > 0) e = (e as TouchEvent).targetTouches[0];
          else if ((e as TouchEvent).changedTouches.length > 0) e = (e as TouchEvent).changedTouches[0];
        }
        docpos = overdiag.getMouse(e);
      } else if (docpos === null) {
        docpos = new Point();
      }
      overdiag.lastInput.documentPoint = docpos;
      overdiag.lastInput.viewPoint = overdiag.transformDocToView(docpos);
      overdiag.lastInput.down = false;
      overdiag.lastInput.up = false;
      /*
      let docpos = modelpt;

      if (e !== null) {
        if (e['targetTouches'] !== undefined) {
          if ((e as TouchEvent).targetTouches.length > 0) e = (e as TouchEvent).targetTouches[0];
          else if ((e as TouchEvent).changedTouches.length > 0) e = (e as TouchEvent).changedTouches[0];
        }
        overdiag._setDocView(e, overdiag.lastInput);
      } else if (docpos === null) {
        docpos = new Point();
        overdiag.lastInput.documentPoint = docpos;
        overdiag.lastInput.viewPoint = overdiag.transformDocToView(docpos);
      }

      overdiag.lastInput.down = false;
      overdiag.lastInput.up = false;
      */
      target.doSimulatedDragOver();
    }
    return true;
  }

  //  * Called by mouseUp/touchEnd, returns true if the action occurs
  //  * over a different (but valid) Diagram (dragging tool target)
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {DraggingTool}
   * @param {Event} e
   * @param {Diagram} other
   * @param {Point} modelpt
   * @param {Diagram} curdiag
   * @return {boolean}
   */
  public simulatedMouseUp(e: Event | Touch | null, other: Diagram | null, modelpt: Point, curdiag: Diagram | null): boolean {
    if (DraggingTool._source === null) return false;
    if (curdiag === null) curdiag = other;
    const olddiag = DraggingTool._currentDiagram;
    const originDiagram = DraggingTool._source.diagram;
    if (curdiag === null) {
      DraggingTool._source.doCancel();  // ?? generalize dropping onto other Controls
      return true;
    }
    if (curdiag !== olddiag) {
      // leaving a different diagram
      const olddiagtool = olddiag.toolManager.findTool('Dragging') as DraggingTool;
      if (olddiag !== null && olddiag !== originDiagram && olddiagtool !== null) {
        olddiag.stopAutoScroll();
        DraggingTool._source.isDragOutStarted = false;
        olddiagtool.doSimulatedDragLeave();
        return false; // back to no longer simulating
      }
      DraggingTool._currentDiagram = curdiag;
      // entering a different diagram
      const curdiagtool = curdiag.toolManager.findTool('Dragging') as DraggingTool;
      if (curdiagtool !== null) {
        DraggingTool._source._sourceCancel();
        curdiagtool.rememberDraggingTool();
        curdiagtool.doSimulatedDragEnter();
      }
    }

    if (curdiag !== this.diagram) {
      let docpos = modelpt;
      if (e !== null) {
        if ((e as any)['targetTouches'] !== undefined) {
          if ((e as TouchEvent).targetTouches.length > 0) e = (e as TouchEvent).targetTouches[0];
          else if ((e as TouchEvent).changedTouches.length > 0) e = (e as TouchEvent).changedTouches[0];
        }
        // curdiag._setDocView(e, curdiag.lastInput);
        docpos = curdiag.getMouse(e);
      } else if (docpos === null) {
        docpos = new Point();
      }
      curdiag.lastInput.documentPoint = docpos;
      curdiag.lastInput.viewPoint = curdiag.transformDocToView(docpos);
      curdiag.lastInput.down = false;
      curdiag.lastInput.up = true;

      const target = curdiag.toolManager.findTool('Dragging') as DraggingTool;
      if (target !== null) target.doSimulatedDrop();

      const dsource = DraggingTool._source;
      if (dsource !== null) {
        const copying = dsource.mayCopy();
        dsource.transactionResult = (copying ? 'Copy' : 'Move');
        dsource.stopTool();
      }
      return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private _maybeUpdateContainingGroups(): void {
    if (this.draggedParts === null) return;
    const it = this.draggedParts.iterator;
    while (it.next()) {
      const part = it.key;
      if (part instanceof Node) {
        const sg = part.containingGroup;
        if (sg !== null && sg.placeholder !== null && !this.draggedParts.contains(sg)) {
          // used to also check sg.computesBoundsAfterDrag,
          // but only happens on mouseup, so why not invalidate them all for sure one last time?
          sg.placeholder.invalidateMeasure();
        }
      }
    }
  }

  /**
   * This predicate is true when the diagram allows objects to be copied and inserted,
   * and some object in the selection is copyable,
   * and the user is holding down the Control key (Option key on Mac).
   *
   * This method may be overridden, although in most cases it is easiest to set {@link Part#copyable}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @return {boolean}
   */
  public mayCopy(): boolean {
    if (!this.isCopyEnabled) return false;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly || !diagram.allowInsert || !diagram.allowCopy) return false;
    if (!(Util.isMac ? diagram.lastInput.alt : diagram.lastInput.control)) return false;
    const it = diagram.selection.iterator;
    while (it.next()) {
      const part = it.value;
      if (part.isNodeLike()) {
        const n = part;
        if (n.canCopy()) {
          return true;
        }
      }
    }
    if (this.draggedLink !== null && this.dragsLink && this.draggedLink.canCopy()) return true;
    return false;
  }

  /**
   * @hidden @internal
   * This predicate is true when the diagram allows objects to be copied out of the diagram,
   * and some object in the selection is copyable.
   *
   * This method may be overridden, although in most cases it is easiest to set {@link Part#copyable}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @return {boolean}
   */
  public mayDragOut(): boolean {
    if (!this.isCopyEnabled) return false;
    const diagram = this.diagram;
    if (!diagram.allowDragOut || !diagram.allowCopy) return false;
    // if allowMove is false, we should enable drag-out temporary copies (Palette sets allowMove to false)
    if (diagram.allowMove) return false;
    const it = diagram.selection.iterator;
    while (it.next()) {
      const part = it.value;
      if (part.isNodeLike()) {
        const n = part;
        if (n.canCopy()) {
          return true;
        }
      }
    }
    if (this.draggedLink !== null && this.dragsLink && this.draggedLink.canCopy()) return true;
    return false;
  }

  /**
   * This predicate is true when the diagram allows objects to be moved,
   * and some object in the selection is movable.
   *
   * This method may be overridden, although in most cases it is easiest to set {@link Part#movable}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DraggingTool}
   * @return {boolean}
   */
  public mayMove(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || !diagram.allowMove) return false; // isModelReadOnly does not stop the diagram from moving an object
    const it = diagram.selection.iterator;
    while (it.next()) {
      const part = it.value;
      if (part.isNodeLike()) {
        const n = part;
        if (n.canMove()) {
          return true;
        }
      }
    }
    if (this.draggedLink !== null && this.dragsLink && this.draggedLink.canMove()) return true;
    return false;
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  public rememberDraggingTool(): void {
    if (!DraggingTool._draggedOverTools.contains(this)) {
      DraggingTool._draggedOverTools.add(this);
    }
  }

  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  public static cleanUpDraggingTool(): void {
    // hack to clean up temporary nodes created by DragOver in other Diagrams
    if (DraggingTool._draggedOverTools.count > 0) {
      const tools = DraggingTool._draggedOverTools;
      const l = tools.length;
      for (let i = 0; i < l; i++) {
        const tool = tools.elt(i);
        tool._cleanUpDraggedLink();
        tool.removeCopies();
        tool._stopEnterLeave();
        tool.diagram.stopAutoScroll();  // make sure there isn't any timer going
      }
      tools.clear();
    }
  }

  /**
   * @hidden @internal
   * @expose ??
   * @this {DraggingTool}
   */
  public computeBorder(g: Group, savedBounds: Rect, result: Rect): Rect {
    if (!this._dropped && this.draggedParts !== null && !this.draggedParts.contains(g)) {
      return result.assign(savedBounds);
    }
    return null;
  }

  // Simulated drag-and-drop

  // in the target DraggingTool:

  /**
   * @hidden @internal
   * Return the DraggingTool of the Diagram that is the source of an external drag-and-drop.
   * @export
   * @this {DraggingTool}
   * @return {DraggingTool} often null
   */
  public getDraggingSource(): DraggingTool {
    return DraggingTool._source as DraggingTool;
  }

  /**
   * @hidden @internal
   * @expose
   * @this {DraggingTool}
   * @return {boolean}
   */
  public mayDragIn(): boolean {
    const diagram = this.diagram;  // this diagram must allow insert
    if (!diagram.allowDrop || diagram.isReadOnly || diagram.isModelReadOnly || !diagram.allowInsert) return false;
    // this model must support the source model's dataFormat
    const draggingTool = DraggingTool._source;
    if (draggingTool === null || draggingTool.diagram.model.dataFormat !== diagram.model.dataFormat) return false;
    return true;
  }

  /**
   * @hidden @internal
   * Called when entering a target diagram.
   * @expose
   * @this {DraggingTool}
   */
  public doSimulatedDragEnter(): void {
    if (!this.mayDragIn()) return;
    const diagram = this.diagram; // don't have to check for null because mayDragIn() does
    diagram.animationManager.stopAnimation();
    diagram.doUpdate();
    diagram.animationManager.stopAnimation(); // necessary if the diagram is brand new, but maybe unnecessary otherwise?
    const draggingTool = DraggingTool._source;
    if (draggingTool !== null) {
      draggingTool.diagram.currentCursor = this.copyCursor;
      draggingTool.diagram.skipsBoundsUpdates = false;
    }
  }

  /**
   * @hidden @internal
   * Called when leaving a target diagram.
   * @expose
   * @this {DraggingTool}
   */
  public doSimulatedDragLeave(): void {
    const draggingTool = DraggingTool._source as DraggingTool;
    if (draggingTool !== null) {  // pretend we have just left the source diagram
      draggingTool.doSimulatedDragOut();
    }
    this.doCancel();
  }

  /**
   * @hidden @internal
   * Called when moving the mouse over a target Diagram, after a call to doSimulatedDragEnter.
   * @expose
   * @this {DraggingTool}
   */
  public doSimulatedDragOver(): void {
    const diagram = this.diagram;

    const src = DraggingTool._source as DraggingTool;
    if (src !== null && src.draggedParts !== null) {
      // support copying only
      if (!this.mayDragIn()) return;
      diagram.currentCursor = this.copyCursor;
      this.addCopiesSim(src.draggedParts.toKeySet(), false, diagram.firstInput);
      this.moveCollection(this.copiedParts, false);  // don't check for Part.canMove()

      this.internalDragOver(diagram.lastInput.documentPoint);
    }
  }

  /**
   * @hidden @internal
   * @expose
   * @this {DraggingTool}
   */
  public doSimulatedDrop(): void {
    const diagram = this.diagram;

    const src = DraggingTool._source as DraggingTool;
    if (src !== null) {
      const sourceDiagram = src.diagram;
      // always tell source Diagram/DraggingTool that the drop happened, even if disallowed
      src._dropped = true;

      // get rid of any temporary objects before turning on any UndoManager
      this.removeCopies();

      if (!this.mayDragIn()) return;

      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      this.startTransaction('Drop');
      this.addCopiesSim(src.draggedParts.toKeySet(), true, diagram.lastInput);
      this.moveCollection(this.copiedParts, false);  // don't check for Part.canMove()
      if (this.copiedParts !== null) {
        diagram.internalClearSelection();
        this.copiedParts.iteratorKeys.each((p: Part) => {
          p.isSelected = true;
        });
      }

      // consider side-effects on the "target" object, if any
      this.internalDropOnto(diagram.lastInput.documentPoint);
      diagram.invalidateDocumentBounds();

      let newparts = diagram.selection;
      if (this.copiedParts !== null) { // success
        this.transactionResult = 'ExternalCopy';
      } else {  // failure -- canceled -- provide an empty Set
        newparts = new Set<Part>();
      }
      // cleanup
      this.copiedParts = null;
      diagram.doFocus();  // new for 1.5

      diagram.raiseDiagramEvent('ExternalObjectsDropped', newparts, sourceDiagram);
      this.stopTransaction();
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
    }
  }

  /**
   * @hidden @internal
   * Nearly identical to addCopies
   * @this {DraggingTool}
   * @param {Set.<Part>} originals a collection of {@link Part}s
   * @param {boolean} undoable
   * @param {InputEvent} startEvt
   */
  private addCopiesSim(originals: Set<Part>, undoable: boolean, startEvt: InputEvent): void {
    if (this.copiedParts !== null) return;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return;
    diagram.skipsUndoManager = !undoable;
    diagram.partManager.addsToTemporaryLayer = !undoable;

    this.startPoint = startEvt.documentPoint; // firstInput if dragging over, lastInput if dropping;
    const copieddata = diagram.copyParts(originals, diagram, true);

    const bounds = Rect.alloc();
    this.computeLocations(originals, bounds);
    const midx = bounds.x + bounds.width / 2;
    const midy = bounds.y + bounds.height / 2;
    Rect.free(bounds);
    const loc = this._startPoint;

    const copiedparts = new Map<Part, DraggingInfo>();
    const temp = Point.alloc();
    const nit = originals.iterator;
    while (nit.next()) {
      const n = nit.value;

      const c = copieddata.getValue(n);
      if (n.isNodeLike() && n.canCopy()) {
        const nloc = n.location;
        temp.assignTo(loc.x - (midx - nloc.x), loc.y - (midy - nloc.y));
        c.location = temp; // ?? does this do anything at all? try commenting it out
        c.ensureBounds();
        copiedparts.add(c, this._newDraggingInfo(temp));
      } else if (c instanceof Link && n.canCopy()) {
        // ?? If you wanted to move links across Diagram bounds too you could
        // but if the templates in each Diagram are different or the node sizes are different the results may be bad
        c.movePoints(loc.x - midx, loc.y - midy);
        copiedparts.add(c, this._newDraggingInfo());
      }
    }
    Point.free(temp);

    this.copiedParts = copiedparts;

    this._checkForDraggedLink(copiedparts.toKeySet());
    if (this.draggedLink !== null) {
      const route = this.draggedLink;
      const b = route.routeBounds;
      route.movePoints(this.startPoint.x - (b.x + b.width / 2),
        this.startPoint.y - (b.y + b.height / 2));
    }
  }

  // simulated drag-and-drop

  // in the source DraggingTool:
  /**
   * @hidden @internal
   * @this {DraggingTool}
   */
  private doDragOut(): void {
    this.isDragOutStarted = true;
    this._dropped = false;
    DraggingTool._source = this;
    DraggingTool._currentDiagram = this.diagram;
    this.doSimulatedDragOut();
  }

  /**
   * @hidden @internal
   * Called on the original Diagram when the mouse is not over a different diagram.
   * @expose
   * @this {DraggingTool}
   */
  public doSimulatedDragOut(): void {
    const diagram = this.diagram;
    diagram.skipsBoundsUpdates = false;
    if (!this.mayCopy() && !this.mayMove()) {
      diagram.currentCursor = this.nodropCursor;
    } else {
      diagram.currentCursor = '';
    }
  }


  /**
   * This method computes the new location for a {@link Node} or simple {@link Part},
   * given a new desired location and an optional Map of dragged parts,
   * taking any grid-snapping into consideration,
   * any {@link Part#dragComputation} function,
   * and any {@link Part#minLocation} and {@link Part#maxLocation}.
   *
   * As of 2.0, this just calls {@link Diagram#computeMove} and remains for compatibility.
   *
   * This method may be overridden, but should usually be overridden on Diagram.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Diagram}
   * @param {Part} n  the Node or simple Part that is being moved
   * @param {Point} newloc the proposed new location
   * @param {Map.<Part,DraggingInfo>=} draggedparts  an optional Map mapping {@link Part}s to DraggingInfo Objects
   *        that have a "point" property remembering the original location of that Part.
   * @param {Point=} result  an optional Point that is modified and returned
   * @return {Point}  the possibly grid-snapped computed Point that is within the minimum and maximum permitted locations
   * @since 1.1
   */
  public computeMove(n: Part, newloc: Point, draggedparts?: Map<Part, DraggingInfo> | null, result?: Point): Point {
    const diagram = this.diagram;
    if (diagram !== null) return diagram.computeMove(n, newloc, this.dragOptions, result);
    return new Point();
  }

  // Static
  /** @hidden @internal */
  public static _draggedOverTools: List<DraggingTool> = null;

  // the DraggingTool that is the source of a simulated drag-and-drop
  /** @hidden @internal */
  public static _source: DraggingTool = null; // DraggingTool
  /** @hidden @internal */
  public static _currentDiagram: Diagram = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    DraggingTool._draggedOverTools = new List<DraggingTool>();
  }

} // #end DraggingTool
DraggingTool['className'] = 'DraggingTool';
DraggingTool.initStatic();

// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'draggingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): DraggingTool {
    return this.findTool('Dragging') as DraggingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: DraggingTool) { this.replaceStandardTool('Dragging', value, this.mouseMoveTools); }
);

/**
 * @hidden @internal
 * @this {ToolManager}
 */
ToolManager.prototype.doCancel = function(): void {
  if (DraggingTool._source !== null) {
    DraggingTool._source.doCancel();
  }
  Tool.prototype.doCancel.call(this);
};


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
LinkingBaseTool
LinkingTool
RelinkingTool
*/

// import { Map, Set } from './collections';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { Adornment, Part } from './parts';
// import { Node } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData } from './root';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Tool } from './tool';
// import { Util } from './util';


/**
 * This abstract class is the base class for the {@link LinkingTool} and {@link RelinkingTool} classes.
 *
 * This class includes properties for defining and accessing any temporary nodes and temporary link
 * that are used during any linking operation, as well as access to the existing diagram's nodes and link
 * (if any) that are involved with the linking operation.
 * <p class="boxread">
 * For a general discussion of link routing, see:
 * <a href="../../intro/links.html">Introduction to Links</a>,
 * <a href="../../intro/linkLabels.html">Introduction to Link Labels</a>, and
 * <a href="../../intro/connectionPoints.html">Introduction to Link Connection Points</a>.
 * For customizing the linking tools, see <a href="../../intro/tools.html#LinkingAndRelinkingTools">Introduction to the Linking Tools</a>.
 * For customizing the reshaping of Links, see <a href="../../intro/tools.html#LinkReshapingTool">Introduction to the LinkReshapingTool</a>.
 * For a general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export abstract class LinkingBaseTool extends Tool {
  /** @hidden @internal */ public static className = 'LinkingBaseTool';
  /** @hidden @internal */ private _portGravity: number;
  /** @hidden @internal */ private _isUnconnectedLinkValid: boolean;
  /** @hidden @internal */ private _cursor: string;
  /** @hidden @internal */ private _temporaryLink: Link;
  /** @hidden @internal */ private _temporaryFromNode: Node;
  /** @hidden @internal */ private _temporaryFromPort: GraphObject;
  /** @hidden @internal */ private _temporaryToNode: Node;
  /** @hidden @internal */ private _temporaryToPort: GraphObject;
  /** @hidden @internal */ private _originalLink: Link | null;
  /** @hidden @internal */ private _originalFromNode: Node | null;
  /** @hidden @internal */ private _originalFromPort: GraphObject | null;
  /** @hidden @internal */ private _originalToNode: Node | null;
  /** @hidden @internal */ private _originalToPort: GraphObject | null;
  /** @hidden @internal */ private _isForwards: boolean;
  /** @hidden @internal */ private _validPortsCache: Map<GraphObject, boolean>;
  /** @hidden @internal */ private _targetPort: GraphObject | null;
  /** @hidden @internal */ private _linkValidation: ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null;
  /** @hidden @internal */ private _portTargeted: ((node: Node, port: GraphObject, tempNode: Node, tempPort: GraphObject, toEnd: boolean) => void) | null;

  /**
   * Don't construct this directly -- this is an abstract class.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(LinkingBaseTool);
    super();

    this._portGravity = 100.0;
    this._isUnconnectedLinkValid = false;
    this._cursor = 'pointer';

    const link = new Link();
    const path = new Shape();
    path.isPanelMain = true;
    path.stroke = 'blue';
    link.add(path);
    const arrow = new Shape();
    arrow.toArrow = 'Standard';
    arrow.fill = 'blue';
    arrow.stroke = 'blue';
    link.add(arrow);
    link.layerName = 'Tool';
    this._temporaryLink = link;

    const fromNode = new Node();
    const fromPort = new Shape();
    fromPort.portId = '';
    fromPort.figure = 'Rectangle';
    fromPort.fill = null;
    fromPort.stroke = 'magenta';
    fromPort.strokeWidth = 2;
    fromPort.desiredSize = Size.OneSize;
    fromNode.add(fromPort);
    fromNode.selectable = false;
    fromNode.layerName = 'Tool';
    this._temporaryFromNode = fromNode;
    this._temporaryFromPort = fromPort;

    const toNode = new Node();
    const toPort = new Shape();
    toPort.portId = '';
    toPort.figure = 'Rectangle';
    toPort.fill = null;
    toPort.stroke = 'magenta';
    toPort.strokeWidth = 2;
    toPort.desiredSize = Size.OneSize;
    toNode.add(toPort);
    toNode.selectable = false;
    toNode.layerName = 'Tool';
    this._temporaryToNode = toNode;
    this._temporaryToPort = toPort;

    // transient state
    this._originalLink = null;
    this._originalFromNode = null;
    this._originalFromPort = null;
    this._originalToNode = null;
    this._originalToPort = null;

    this._isForwards = true;
    this._validPortsCache = new Map<GraphObject, boolean>();
    this._targetPort = null;
    this._linkValidation = null;
    this._portTargeted = null;
  }

  /**
   * @hidden @internal
   * Clear out any temporary state.
   * @this {LinkingBaseTool}
   */
  public doStop(): void {
    const diagram = this.diagram;
    diagram.stopAutoScroll();
    this.originalLink = null;
    this.originalFromNode = null;
    this.originalFromPort = null;
    this.originalToNode = null;
    this.originalToPort = null;
    this.validPortsCache.clear();
    this.targetPort = null;
  }

  /**
   * Gets or sets the distance at which link snapping occurs.
   * The default value is 100.0.
   * The value must be non-negative.
   * Setting this property does not raise any events.
   */
  /** @export */ get portGravity(): number { return this._portGravity; }
  /** @export */ set portGravity(value: number) {
    Util.checkPrimitive(value, 'number', LinkingBaseTool, 'portGravity');
    if (value >= 0) {
      this._portGravity = value;
    }
  }

  /**
   * Gets or sets whether it is valid to have partly or completely unconnected links.
   *
   * The default value is false -- the user cannot draw or reconnect a link to "nowhere".
   * Setting this property does not raise any events.
   * @since 1.3
   */
  /** @export */ get isUnconnectedLinkValid(): boolean { return this._isUnconnectedLinkValid; }
  /** @export */ set isUnconnectedLinkValid(value: boolean) {
    Util.checkPrimitive(value, 'boolean', LinkingBaseTool, 'isUnconnectedLinkValid');
    this._isUnconnectedLinkValid = value;
  }

  /** @hidden @internal */
  /** @export */ get linkingCursor() { return this._cursor; }
  /** @export */ set linkingCursor(c: string) { this._cursor = c; }

  /**
   * Gets or sets the temporary {@link Link} that is shown while the user is drawing or reconnecting a link.
   * Setting this property does not raise any events.
   */
  /** @export */ get temporaryLink(): Link { return this._temporaryLink; }
  /** @export */ set temporaryLink(value: Link) {
    Util.checkClass(value, Link, LinkingBaseTool, 'temporaryLink');
    this._temporaryLink = value;
  }

  /**
   * Gets or sets the temporary {@link Node} at the "from" end of the {@link #temporaryLink}
   * while the user is drawing or reconnecting a link.
   * Setting this property does not raise any events.
   */
  /** @export */ get temporaryFromNode(): Node { return this._temporaryFromNode; }
  /** @export */ set temporaryFromNode(value: Node) {
    Util.checkClass(value, Node, LinkingBaseTool, 'temporaryFromNode');
    this._temporaryFromNode = value;
  }

  /**
   * Gets or sets the {@link GraphObject} that is the port at the "from" end of the {@link #temporaryLink}
   * while the user is drawing or reconnecting a link.
   * Setting this property does not raise any events.
   */
  /** @export */ get temporaryFromPort(): GraphObject { return this._temporaryFromPort; }
  /** @export */ set temporaryFromPort(value: GraphObject) {
    Util.checkClass(value, GraphObject, LinkingBaseTool, 'temporaryFromPort');
    this._temporaryFromPort = value;
  }

  /**
   * Gets or sets the temporary {@link Node} at the "to" end of the {@link #temporaryLink}
   * while the user is drawing or reconnecting a link.
   * Setting this property does not raise any events.
   */
  /** @export */ get temporaryToNode(): Node { return this._temporaryToNode; }
  /** @export */ set temporaryToNode(value: Node) {
    Util.checkClass(value, Node, LinkingBaseTool, 'temporaryToNode');
    this._temporaryToNode = value;
  }

  /**
   * Gets or sets the {@link GraphObject} that is the port at the "to" end of the {@link #temporaryLink}
   * while the user is drawing or reconnecting a link.
   * Setting this property does not raise any events.
   */
  /** @export */ get temporaryToPort(): GraphObject { return this._temporaryToPort; }
  /** @export */ set temporaryToPort(value: GraphObject) {
    Util.checkClass(value, GraphObject, LinkingBaseTool, 'temporaryToPort');
    this._temporaryToPort = value;
  }

  /**
   * Gets or sets the original {@link Link} being reconnected by the {@link RelinkingTool}.
   * Setting this property does not raise any events.
   */
  /** @export */ get originalLink(): Link | null { return this._originalLink; }
  /** @export */ set originalLink(value: Link | null) {
    if (value !== null) Util.checkClass(value, Link, LinkingBaseTool, 'originalLink');
    this._originalLink = value;
  }

  /**
   * Gets or sets the original {@link Node} from which the new link is being drawn
   * or from which the {@link #originalLink} was connected when being relinked.
   * Setting this property does not raise any events.
   */
  /** @export */ get originalFromNode(): Node | null { return this._originalFromNode; }
  /** @export */ set originalFromNode(value: Node | null) {
    if (value !== null) Util.checkClass(value, Node, LinkingBaseTool, 'originalFromNode');
    this._originalFromNode = value;
  }

  /**
   * Gets or sets the {@link GraphObject} that is the port in the {@link #originalFromNode}.
   * Setting this property does not raise any events.
   */
  /** @export */ get originalFromPort(): GraphObject | null { return this._originalFromPort; }
  /** @export */ set originalFromPort(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkingBaseTool, 'originalFromPort');
    this._originalFromPort = value;
  }

  /**
   * Gets or sets the original {@link Node} to which the new link is being drawn
   * or to which the {@link #originalLink} was connected when being relinked.
   * Setting this property does not raise any events.
   */
  /** @export */ get originalToNode(): Node | null { return this._originalToNode; }
  /** @export */ set originalToNode(value: Node | null) {
    if (value !== null) Util.checkClass(value, Node, LinkingBaseTool, 'originalToNode');
    this._originalToNode = value;
  }

  /**
   * Gets or sets the {@link GraphObject} that is the port in the {@link #originalToNode}.
   * Setting this property does not raise any events.
   */
  /** @export */ get originalToPort(): GraphObject | null { return this._originalToPort; }
  /** @export */ set originalToPort(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkingBaseTool, 'originalToPort');
    this._originalToPort = value;
  }

  /**
   * Gets whether the linking operation is in the forwards direction,
   * connecting from the "From" port to the "To" port.
   */
  /** @export */ get isForwards(): boolean { return this._isForwards; }
  /** @export */ set isForwards(value: boolean) {
    Util.checkPrimitive(value, 'boolean', LinkingBaseTool, 'isForwards');
    this._isForwards = value;
  }

  /**
   * @hidden @internal
   * This read-only property returns the Map used to keep track of ports for which a link may be valid.
   * This remembers the results of calls to {@link #isValidLink}.
   * The values are either true or false if known, or null if the validity is untested.
   */
  /** @export */ get validPortsCache(): Map<GraphObject, boolean> { return this._validPortsCache; }

  /**
   * Gets or sets a proposed {@link GraphObject} port for connecting a link.
   * Whether this is a "to" port or a "from" port depends on the direction
   * ({@link #isForwards}) in which the link is being drawn or reconnected.
   *
   * This is set when the mouse is being dragged and when a mouse-up event occurs
   * with the result of a call to {@link #findTargetPort}.
   * Setting this property does not raise any events.
   */
  /** @export */ get targetPort(): GraphObject | null { return this._targetPort; }
  /** @export */ set targetPort(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkingBaseTool, 'targetPort');
    this._targetPort = value;
  }

  /**
   * Make a temporary port look and act like a real one.
   *
   * This is called by {@link #doMouseMove} as the tool finds new valid target ports.
   * If {@link #findTargetPort} returns null, it calls {@link #setNoTargetPortProperties} instead.
   *
   * If the {@link #portTargeted} property is a function, that function is called.
   * It is easier to customize the behavior by setting that functional property than it is to override
   * this method and the {@link #setNoTargetPortProperties} method.
   * But you may want to call this method to get the standard behavior for dynamically adapting
   * the temporary node/port to "act like" the target port.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} realnode
   * @param {GraphObject} realport
   * @param {Node} tempnode
   * @param {GraphObject} tempport
   * @param {boolean} toend
   * @since 1.2
   */
  protected copyPortProperties(realnode: Node | null, realport: GraphObject | null, tempnode: Node, tempport: GraphObject, toend: boolean): void {
    if (realnode === null || realport === null || tempnode === null || tempport === null) return;
    const scale = realport.getDocumentScale();
    const tempsize = Size.alloc();
    tempsize.width = realport.naturalBounds.width * scale;
    tempsize.height = realport.naturalBounds.height * scale;
    tempport.desiredSize = tempsize;
    Size.free(tempsize);
    if (toend) {
      tempport.toSpot = realport.toSpot;
      tempport.toEndSegmentLength = realport.toEndSegmentLength;
    } else {
      tempport.fromSpot = realport.fromSpot;
      tempport.fromEndSegmentLength = realport.fromEndSegmentLength;
    }
    tempnode.locationSpot = Spot.Center;
    const temploc = Point.alloc();
    tempnode.location = realport.getDocumentPoint(Spot.Center, temploc);
    Point.free(temploc);
    tempport.angle = realport.getDocumentAngle();
    if (this.portTargeted !== null) {
      this.portTargeted(realnode, realport, tempnode, tempport, toend);
    }
  }

  /**
   * Reset a temporary port's properties to neutral values when there is no target port.
   *
   * This is called by {@link #doMouseMove} when the tool finds no new valid target port.
   * If {@link #findTargetPort} returns non-null, it calls {@link #copyPortProperties} instead.
   *
   * If the {@link #portTargeted} property is a function, that function is called with null values as the first two arguments.
   * It is easier to customize the behavior by setting that functional property than it is to override
   * this method and the {@link #copyPortProperties} method.
   * But you may want to call this method to get the standard behavior for dynamically adapting
   * the temporary node/port to "act like" it is not connecting with any target port.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} tempnode
   * @param {GraphObject} tempport
   * @param {boolean} toend
   * @since 1.2
   */
  protected setNoTargetPortProperties(tempnode: Node, tempport: GraphObject, toend: boolean): void {
    if (tempport !== null) {
      tempport.desiredSize = Size.OneSize;
      tempport.fromSpot = Spot.None;
      tempport.toSpot = Spot.None;
    }
    if (tempnode !== null) {
      tempnode.location = this.diagram.lastInput.documentPoint;
    }
    if (this.portTargeted !== null) {
      this.portTargeted(null, null, tempnode, tempport, toend);
    }
  }

  /**
   * @hidden @internal
   * Call {@link #doMouseMove} in order to update any temporary nodes.
   * @this {LinkingBaseTool}
   */
  public doMouseDown(): void {
    if (this.isActive) {
      this.doMouseMove();
    }
  }

  /**
   * Mouse movement results in a temporary node moving to where a valid target port is located,
   * or to where the mouse is if there is no valid target port nearby.
   *
   * This calls {@link #findTargetPort} to update the {@link #targetPort}
   * given the new mouse point.
   * If a valid target port is found this moves the temporary node/port and
   * makes them appear like the target node/port by calling {@link #copyPortProperties}.
   * If no valid target port is found, this moves the temporary node to
   * where the mouse currently is and removes any node/port appearance,
   * by calling {@link #setNoTargetPortProperties}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {LinkingBaseTool}
   */
  public doMouseMove(): void {
    if (this.isActive) {
      const diagram = this.diagram;
      this.targetPort = this.findTargetPort(this.isForwards);
      if (this.targetPort !== null && this.targetPort.part instanceof Node) {
        const targetnode = this.targetPort.part;
        if (this.isForwards) {
          this.copyPortProperties(targetnode, this.targetPort, this.temporaryToNode, this.temporaryToPort, true);
        } else {
          this.copyPortProperties(targetnode, this.targetPort, this.temporaryFromNode, this.temporaryFromPort, false);
        }
      } else {
        // found no potential port
        if (this.isForwards) {
          this.setNoTargetPortProperties(this.temporaryToNode, this.temporaryToPort, true);
        } else {
          this.setNoTargetPortProperties(this.temporaryFromNode, this.temporaryFromPort, false);
        }
      }

      if (diagram.allowHorizontalScroll || diagram.allowVerticalScroll) {
        diagram.doAutoScroll(diagram.lastInput.viewPoint);
      }
    }
  }

  /**
   * @hidden @internal
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {GraphObject} obj
   * @param {boolean} toend
   * @return {GraphObject}
   */
  public findValidLinkablePort(obj: GraphObject, toend: boolean): GraphObject | null {
    if (obj === null) return null;
    const part = obj.part;
    if (!(part instanceof Node)) return null;
    while (obj !== null) {
      const allow = (toend ? obj.toLinkable : obj.fromLinkable);
      if (allow === true && (obj.portId !== null || obj instanceof Node) &&
        (toend ? this.isValidTo(part, obj) : this.isValidFrom(part, obj))) return obj;
      if (allow === false) return null;
      // if ALLOW is null, continue searching up the parent chain
      obj = obj.panel;
    }
    return null;
  }

  /**
   * Find a port with which the user could complete a valid link.
   *
   * This finds objects near to the current mouse point for which a valid link connection is possible.
   *
   * For each port element found, this calls {@link #isValidLink} to find out if a link between
   * the original node/port and the found node/port would be valid.
   * The result is saved in a cache for faster decisions later during operation of this tool.
   * The closest valid port is returned.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {boolean} toend true if looking for a "to" port.
   * @return {GraphObject} a valid port, or null if no such port is near the current mouse point
   * within the {@link #portGravity} distance.
   */
  public findTargetPort(toend: boolean): GraphObject | null {
    const diagram = this.diagram;
    const p = diagram.lastInput.documentPoint;
    let gravity = this.portGravity;
    if (gravity <= 0) gravity = 0.1;
    const tool = this;
    const nearports = diagram.findObjectsNear(p, gravity,
      function (x) { return tool.findValidLinkablePort(x, toend); },
      null, true); // ??? temporarily true?
    let bestDist = Infinity;
    let bestPort: GraphObject = null;
    const it = nearports.iterator;
    while (it.next()) {
      const port = it.value;
      const part = port.part;
      if (!(part instanceof Node)) continue;
      const toPoint = port.getDocumentPoint(Spot.Center, Point.alloc());  // ?? assumes center point of port
      const dx = p.x - toPoint.x;
      const dy = p.y - toPoint.y;
      Point.free(toPoint);
      const dist = dx * dx + dy * dy;  // don't bother taking sqrt
      if (dist < bestDist) {  // closest so far
        // check cache of isValidLink calls
        const valid = this.validPortsCache.getValue(port);  // true, false, or null
        if (valid !== null) {
          // known to be either valid or invalid
          if (valid) {  // known to be a valid port for a link
            bestPort = port;
            bestDist = dist;
          } // else known not valid: don't need to call isValidLink again
        } else {  // but not if cached, try isValidLink in the appropriate direction
          if ((toend && this.isValidLink(this.originalFromNode, this.originalFromPort, part, port)) ||
            (!toend && this.isValidLink(part, port, this.originalToNode, this.originalToPort))) {
            // now known valid, remember in cache
            this.validPortsCache.add(port, true);
            bestPort = port;
            bestDist = dist;
          } else {
            // now known not valid, remember in cache
            this.validPortsCache.add(port, false);
          }
        }
      }
    }
    if (bestPort !== null) {
      const tpart = bestPort.part;
      if (tpart instanceof Node) {
        if (tpart.layer === null || tpart.layer.allowLink) {
          return bestPort;
        }
      }
    }
    return null;
  }

  /**
   * This predicate is true if it is permissible to connect a link from a given node/port.
   * This is called by {@link #isValidLink}.
   * <p class="boxread">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} fromnode
   * @param {GraphObject} fromport
   * @return {boolean}
   * False if the node is in a {@link Layer} that does not {@link Layer#allowLink}.
   * False if the port's {@link GraphObject#fromLinkable} is either false or null.
   * False if the number of links connected to the port would exceed the port's {@link GraphObject#fromMaxLinks}.
   * Otherwise true.
   */
  public isValidFrom(fromnode: Node | null, fromport: GraphObject | null): boolean {
    if (fromnode === null || fromport === null) {
      return this.isUnconnectedLinkValid;
    }
    if (this.diagram.currentTool === this) {
      if (fromnode.layer !== null && !fromnode.layer.allowLink) return false;
      if (fromport.fromLinkable !== true) return false;  // false or null value means not linkable
    }
    const maxlinks = fromport.fromMaxLinks;
    if (maxlinks < Infinity) {
      // ?? wrong number, because RelinkingTool doesn't temporarily disconnect?
      if (this.originalLink !== null && fromnode === this.originalFromNode && fromport === this.originalFromPort) return true;
      let fromportid = fromport.portId;
      if (fromportid === null) fromportid = '';
      if (fromnode.findLinksOutOf(fromportid).count >= maxlinks) return false;
    }
    return true;
  }

  /**
   * This predicate is true if it is permissible to connect a link to a given node/port.
   * This is called by {@link #isValidLink}.
   * <p class="boxread">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} tonode
   * @param {GraphObject} toport
   * @return {boolean}
   * False if the node is in a {@link Layer} that does not {@link Layer#allowLink}.
   * False if the port's {@link GraphObject#toLinkable} is either false or null.
   * False if the number of links connected from the port would exceed the port's {@link GraphObject#toMaxLinks}.
   * Otherwise true.
   */
  public isValidTo(tonode: Node | null, toport: GraphObject | null): boolean {
    if (tonode === null || toport === null) {
      return this.isUnconnectedLinkValid;
    }
    if (this.diagram.currentTool === this) {
      if (tonode.layer !== null && !tonode.layer.allowLink) return false;
      if (toport.toLinkable !== true) return false;  // false or null value means not linkable
    }
    const maxlinks = toport.toMaxLinks;
    if (maxlinks < Infinity) {
      // ?? wrong number, because RelinkingTool doesn't temporarily disconnect?
      if (this.originalLink !== null && tonode === this.originalToNode && toport === this.originalToPort) return true;
      let toportid = toport.portId;
      if (toportid === null) toportid = '';
      if (tonode.findLinksInto(toportid).count >= maxlinks) return false;
    }
    return true;
  }

  /**
   * This predicate is true if both argument ports are in the same {@link Node}.
   * This is called by {@link #isValidLink}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * It is rare that you will want to override this method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {GraphObject} fromport
   * @param {GraphObject} toport
   * @return {boolean}
   */
  public isInSameNode(fromport: GraphObject | null, toport: GraphObject | null): boolean {
    if (fromport === null || toport === null) return false;
    if (fromport === toport) return true;
    const fromnode = fromport.part;
    const tonode = toport.part;
    return (fromnode !== null && fromnode === tonode);
  }

  /**
   * This predicate is true if there is a link in the diagram going from the given port to the given port.
   * This is called by {@link #isValidLink}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * It is rare that you will want to override this method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {GraphObject} fromport
   * @param {GraphObject} toport
   * @return {boolean}
   */
  public isLinked(fromport: GraphObject | null, toport: GraphObject | null): boolean {
    if (fromport === null || toport === null) return false;
    const fpart = fromport.part;
    if (!(fpart instanceof Node)) return false;
    let fromportid = fromport.portId;
    if (fromportid === null) fromportid = '';
    const tpart = toport.part;
    if (!(tpart instanceof Node)) return false;
    let toportid = toport.portId;
    if (toportid === null) toportid = '';
    const it = tpart.findLinksInto(toportid);
    while (it.next()) {
      const link = it.value;
      if (link.fromNode === fpart && link.fromPortId === fromportid) return true;
    }
    return false;
  }

  /**
   * This predicate should be true when it is logically valid to connect a new link from
   * one node/port to another node/port.
   * When this is called by the {@link RelinkingTool}, that tool sets {@link #originalLink}
   * to be the {@link Link} being reconnected.
   * <p class="boxread">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * This method may be overridden, although it is usually much easier to just set {@link #linkValidation}
   * or {@link Node#linkValidation} in order to add some application-specific link validation.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} fromnode the "from" {@link Node}.
   * @param {GraphObject} fromport the "from" {@link GraphObject} port.
   * @param {Node} tonode the "to" {@link Node}.
   * @param {GraphObject} toport the "to" {@link GraphObject} port.
   * @return {boolean}
   * False if {@link #isValidFrom} is false for the "from" node/port.
   * False if {@link #isValidTo} is false for the "to" node/port.
   * False if {@link #isInSameNode} is true unless {@link GraphObject#fromLinkableSelfNode}
   * and {@link GraphObject#toLinkableSelfNode} are true for the two ports.
   * False if {@link #isLinked} is true unless {@link GraphObject#fromLinkableDuplicates}
   * and {@link GraphObject#toLinkableDuplicates} are true for the two ports.
   * False if trying to link to the link's own label node(s).
   * If {@link Node#linkValidation} is a predicate on either or both of the "from" node and the "to" node,
   * the predicate is called and if it returns false, this predicate returns false.
   * If {@link #linkValidation} is a predicate and if it returns false, this predicate returns false.
   * Otherwise this predicate is true.
   */
  public isValidLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): boolean {
    // allow partly connected or unconnected links
    if (!this.isValidFrom(fromnode, fromport)) return false;
    if (!this.isValidTo(tonode, toport)) return false;

    if (fromport !== null && toport !== null) {
      if (!(fromport.fromLinkableSelfNode && toport.toLinkableSelfNode) &&
        this.isInSameNode(fromport, toport)) return false;
      if (!(fromport.fromLinkableDuplicates && toport.toLinkableDuplicates) &&
        this.isLinked(fromport, toport)) return false;
    }

    // disallow linking to the link's own label(s)
    if (this.originalLink !== null) {
      if (fromnode !== null && this.isLabelDependentOnLink(fromnode, this.originalLink)) return false;
      if (tonode !== null && this.isLabelDependentOnLink(tonode, this.originalLink)) return false;
    }

    // either both nodes are in model or neither are
    if (fromnode !== null && tonode !== null) {
      if (fromnode.data === null && tonode.data !== null) return false;
      if (fromnode.data !== null && tonode.data === null) return false;
    }

    // check Diagram.validCycle
    if (!this.isValidCycle(fromnode, tonode, this.originalLink)) return false;
    let pred;
    if (fromnode !== null) {
      pred = fromnode.linkValidation;
      if (pred !== null && !pred(fromnode, fromport, tonode, toport, this.originalLink)) return false;
    }
    if (tonode !== null) {
      pred = tonode.linkValidation;
      if (pred !== null && !pred(fromnode, fromport, tonode, toport, this.originalLink)) return false;
    }
    pred = this.linkValidation;
    if (pred !== null) {
      return pred(fromnode, fromport, tonode, toport, this.originalLink);
    } else {
      return true;
    }
  }

  /**
   * @hidden @internal
   * False if the node is not a link label node.
   * True if the node is a label on the link (i.e. if the link owns the node).
   * Otherwise true if the label node is on a link whose from or to node is dependent on the link.
   * This is called by {@link #isValidLink}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * It is rare that you will want to override this method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} node
   * @param {Link} link
   * @return {boolean}
   */
  protected isLabelDependentOnLink(node: Node, link: Link): boolean {
    if (node === null) return false;
    const l = node.labeledLink;
    if (l === null) return false;
    if (l === link) return true;
    const labelnodes = new Set<Node>();
    labelnodes.add(node);
    return this._isLinkDependent(l, link, labelnodes);
  }

  /**
   * @hidden @internal
   * False if the node is not a link label node.
   * True if the node is a label on the link (i.e. if the link owns the node).
   * Otherwise true if the label node is on a link whose from or to node is dependent on the link.
   * @this {LinkingBaseTool}
   * @param {Link} l
   * @param {Link} link
   * @param {Set.<Node>} labelnodes
   * @return {boolean}
   */
  private _isLinkDependent(l: Link, link: Link, labelnodes: Set<Node>): boolean {
    if (l === link) return true;
    const from = l.fromNode;
    if (from !== null && from.isLinkLabel) {
      labelnodes.add(from);
      if (this._isLinkDependent(from.labeledLink, link, labelnodes)) return true;
    }
    const to = l.toNode;
    if (to !== null && to.isLinkLabel) {
      labelnodes.add(to);
      if (this._isLinkDependent(to.labeledLink, link, labelnodes)) return true;
    }
    return false;
  }

  /**
   * Checks whether a proposed link would be valid according to {@link Diagram#validCycle}.
   * This does not distinguish between different ports on a node, so this method does not need to take port arguments.
   * This is called by {@link #isValidLink}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * It is rare that you will want to override this method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingBaseTool}
   * @param {Node} from
   * @param {Node} to
   * @param {Link} ignore may be null; this is useful during relinking to ignore the {@link #originalLink}
   * @return {boolean}
   * @since 1.5
   */
  public isValidCycle(from: Node | null, to: Node | null, ignore?: Link | null): boolean {
    if (ignore === undefined) ignore = null;
    if (from === null || to === null) {
      return this.isUnconnectedLinkValid;
    }

    const vc = this.diagram.validCycle;
    if (vc === Diagram.CycleAll) return true;
    else if (vc === Diagram.CycleDestinationTree) {
      const link = ignore || this.temporaryLink;
      if (link !== null && !link.isTreeLink) return true;
      const lit = to.linksConnected;
      while (lit.next()) {
        const tl = lit.value;
        if (tl === ignore) continue;
        if (!tl.isTreeLink) continue;
        if (tl.toNode === to) return false;
      }
      return !this._MakesDirectedCycleFast(from, to, ignore, true);
    } else if (vc === Diagram.CycleSourceTree) {
      const link = ignore || this.temporaryLink;
      if (link !== null && !link.isTreeLink) return true;
      const lit = from.linksConnected;
      while (lit.next()) {
        const fl = lit.value;
        if (fl === ignore) continue;
        if (!fl.isTreeLink) continue;
        if (fl.fromNode === from) return false;
      }
      return !this._MakesDirectedCycleFast(from, to, ignore, true);
    } else if (vc === Diagram.CycleNotDirected) {
      return !this._MakesDirectedCycle(from, to, ignore);
    } else if (vc === Diagram.CycleNotDirectedFast) {
      return !this._MakesDirectedCycleFast(from, to, ignore, false);
    } else if (vc === Diagram.CycleNotUndirected) {
      return !this._MakesUndirectedCycle(from, to, ignore);
    }
    return true;
  }

  /**
   * @hidden @internal
   * @this {LinkingBaseTool}
   * @param {Node} a
   * @param {Node} b
   * @param {Link} ignore
   * @param {boolean} tree
   * @return {boolean}
   */
  public _MakesDirectedCycleFast(a: Node, b: Node, ignore: Link, tree: boolean): boolean {
    if (a === b) return true;
    if (a === null) return false;
    if (b === null) return false;
    const lit = a.linksConnected;
    while (lit.next()) {
      const l = lit.value;
      if (l === ignore) continue;
      if (tree && !l.isTreeLink) continue;
      if (l.toNode !== a) continue;
      const n = l.fromNode;
      if (n === a) continue;
      if (this._MakesDirectedCycleFast(n, b, ignore, tree)) {
        return true;
      }
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {LinkingBaseTool}
   * @param {Node} a
   * @param {Node} b
   * @param {Link} ignore
   * @return {boolean}
   */
  private _MakesDirectedCycle(a: Node, b: Node, ignore: Link): boolean {
    if (a === b) return true;
    const seen = new Set<Node>();
    seen.add(b);
    return this._MakesDirectedCycle1(seen, a, b, ignore);
  }

  /**
   * @hidden @internal
   * @this {LinkingBaseTool}
   * @param {Set.<Node>} seen
   * @param {Node} a
   * @param {Node} b
   * @param {Link} ignore
   * @return {boolean}
   */
  private _MakesDirectedCycle1(seen: Set<Node>, a: Node, b: Node, ignore: Link): boolean {
    if (a === b) return true;
    if (a === null) return false;
    if (b === null) return false;
    if (seen.contains(a)) return false;
    seen.add(a);
    const lit = a.linksConnected;
    while (lit.next()) {
      const l = lit.value;
      if (l === ignore) continue;
      if (l.toNode !== a) continue;
      const n = l.fromNode;
      if (n === a) continue;
      if (this._MakesDirectedCycle1(seen, n, b, ignore)) {
        return true;
      }
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {LinkingBaseTool}
   * @param {Node} a
   * @param {Node} b
   * @param {Link} ignore
   * @return {boolean}
   */
  private _MakesUndirectedCycle(a: Node, b: Node, ignore: Link): boolean {
    if (a === b) return true;
    const seen = new Set<Node>();
    seen.add(b);
    return this._MakesUndirectedCycle1(seen, a, b, ignore);
  }

  /**
   * @hidden @internal
   * @this {LinkingBaseTool}
   * @param {Set.<Node>} seen
   * @param {Node} a
   * @param {Node} b
   * @param {Link} ignore
   * @return {boolean}
   */
  private _MakesUndirectedCycle1(seen: Set<Node>, a: Node, b: Node, ignore: Link): boolean {
    if (a === b) return true;
    if (a === null) return false;
    if (b === null) return false;
    if (seen.contains(a)) return false;
    seen.add(a);
    const lit = a.linksConnected;
    while (lit.next()) {
      const l = lit.value;
      if (l === ignore) continue;
      const from = l.fromNode;
      const to = l.toNode;
      const n = (from === a ? to : from);
      if (n === a) continue;
      if (this._MakesUndirectedCycle1(seen, n, b, ignore)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Gets or sets a predicate that determines whether or not a new link between two ports would be valid.
   * This predicate is called in addition to the normal link checking performed by {@link #isValidLink}
   * and any {@link Node#linkValidation} predicates on the "from" and "to" nodes.
   * When relinking, the {@link Link} being considered for reconnection is passed as the fifth argument.
   * The default predicate is null, which is equivalent to simply returning true.
   * <p class="boxread">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * The function, if supplied, must not have any side-effects.
   */
  /** @export */ get linkValidation(): ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null { return this._linkValidation; }
  /** @export */ set linkValidation(value: ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null) {
    if (value !== null) Util.checkPrimitive(value, 'function', LinkingBaseTool, 'linkValidation');
    this._linkValidation = value;
  }

  /**
   * Gets or sets a function that is called as the tool targets the nearest valid port.
   * The first two arguments specify the port by providing the {@link Node} that it is in
   * and the {@link GraphObject} that is the actual port object.
   * The next two arguments are the temporary port that has been moved and styled to be like the valid port.
   * These values will be either the {@link #temporaryToNode} and {@link #temporaryToPort}
   * or the {@link #temporaryFromNode} and {@link #temporaryFromPort}, depending on
   * which end of the temporary link is being updated.
   * The fifth argument is true if the target port represents a potential "to" end of a link;
   * it is false if it is for the "from" end of a link.
   *
   * When there is no valid port within the {@link #portGravity}, the first two arguments are null.
   *
   * The function, if supplied, must not add or remove any links or nodes or port objects,
   * nor may it change the validity of any potential link connection.
   * @since 1.2
   */
  /** @export */ get portTargeted(): ((node: Node, port: GraphObject, tempNode: Node, tempPort: GraphObject, toEnd: boolean) => void) | null { return this._portTargeted; }
  /** @export */ set portTargeted(value: ((node: Node, port: GraphObject, tempNode: Node, tempPort: GraphObject, toEnd: boolean) => void) | null) {
    if (value !== null) Util.checkPrimitive(value, 'function', LinkingBaseTool, 'portTargeted');
    this._portTargeted = value;
  }
} // #end LinkingBaseTool
LinkingBaseTool['className'] = 'LinkingBaseTool';

/*
One way to minify property names with the TypeScript build.
Requires you to also modify the gulpfile so that we stop @export -ing getters and settesrs.

LinkingBaseTool.prototype['portGravity'] = LinkingBaseTool.prototype.portGravity;
LinkingBaseTool.prototype['isUnconnectedLinkValid'] = LinkingBaseTool.prototype.isUnconnectedLinkValid;
LinkingBaseTool.prototype['temporaryLink'] = LinkingBaseTool.prototype.temporaryLink;
LinkingBaseTool.prototype['temporaryFromNode'] = LinkingBaseTool.prototype.temporaryFromNode;
LinkingBaseTool.prototype['temporaryFromPort'] = LinkingBaseTool.prototype.temporaryFromPort;
LinkingBaseTool.prototype['temporaryToNode'] = LinkingBaseTool.prototype.temporaryToNode;
LinkingBaseTool.prototype['temporaryToPort'] = LinkingBaseTool.prototype.temporaryToPort;
LinkingBaseTool.prototype['originalLink'] = LinkingBaseTool.prototype.originalLink;
LinkingBaseTool.prototype['originalFromNode'] = LinkingBaseTool.prototype.originalFromNode;
LinkingBaseTool.prototype['originalFromPort'] = LinkingBaseTool.prototype.originalFromPort;
LinkingBaseTool.prototype['originalToNode'] = LinkingBaseTool.prototype.originalToNode;
LinkingBaseTool.prototype['originalToPort'] = LinkingBaseTool.prototype.originalToPort;
LinkingBaseTool.prototype['isForwards'] = LinkingBaseTool.prototype.isForwards;
// LinkingBaseTool.prototype['validPortsCache'] = LinkingBaseTool.prototype.validPortsCache;
LinkingBaseTool.prototype['targetPort'] = LinkingBaseTool.prototype.targetPort;
*/


/**
 * The LinkingTool lets a user draw a new {@link Link} between two ports,
 * using a mouse-drag operation.
 *
 * By default an instance of this tool is installed as a mouse-move tool in the
 * {@link Diagram#toolManager} as the {@link ToolManager#linkingTool}.
 * However this tool may be used modally, as described below.
 *
 * {@link #canStart} calls {@link #findLinkablePort} to find a valid "port" element
 * from which (or to which) the user may interactively draw a new link.
 * {@link #doActivate} sets up a temporary link and two temporary nodes,
 * one at the start port and one following the mouse.
 * <p class="boxread">
 * For a general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 *
 * This tool conducts a transaction while the tool is active.
 * A successful linking will result in a "LinkDrawn" {@link DiagramEvent} and a "Linking" transaction.
 *
 * If you want to programmatically start a new user mouse-gesture to draw a new link
 * from a given {@link GraphObject} that either is a "port" or may be within
 * the visual tree of a "port", set the {@link #startObject} property
 * to let {@link #findLinkablePort} find the real "port" element.
 * Then start and activate this tool:
 * ```js
 *   var tool = myDiagram.toolManager.linkingTool;
 *   tool.startObject = ...;
 *   myDiagram.currentTool = tool;
 *   tool.doActivate();
 * ```
 * @extends LinkingBaseTool
 * @unrestricted
 * @category Tool
 */
export class LinkingTool extends LinkingBaseTool {
  /** @hidden @internal */ public static className = 'LinkingTool';
  /** @hidden @internal */ private _archetypeLinkData: ObjectData | null;
  /** @hidden @internal */ private _archetypeLabelNodeData: ObjectData | null;
  /** @hidden @internal */ private _direction: EnumValue;
  /** @hidden @internal */ private _startObject: GraphObject | null;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#linkingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "Linking".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(LinkingTool);
    super();

    this.name = 'Linking';

    this._archetypeLinkData = {};
    this._archetypeLabelNodeData = null;
    this._direction = LinkingTool.Either;
    this._startObject = null;
  }

  /**
   * This value for {@link LinkingTool#direction} indicates that users may draw new links in either direction.
   * @constant
   */
  public static Either = new EnumValue(LinkingTool, 'Either', 0);

  /**
   * This value for {@link LinkingTool#direction} indicates that users may draw new links forwards only
   * (i.e. from "from" node to "to" node).
   * @constant
   */
  public static ForwardsOnly = new EnumValue(LinkingTool, 'ForwardsOnly', 0);

  /**
   * This value for {@link LinkingTool#direction} indicates that users may draw new links backwards only
   * (i.e. from "to" node to "from" node).
   * @constant
   */
  public static BackwardsOnly = new EnumValue(LinkingTool, 'BackwardsOnly', 0);

  /**
   * Gets or sets a data object that is copied by {@link #insertLink}
   * and added to the {@link GraphLinksModel} when creating a new {@link Link}.
   * The default value is an empty Object, which will be copied.
   * The value must be an Object for the linking operation to succeed.
   * A null value will cause {@link #insertLink} to fail.
   *
   * The copied link data's {@link GraphLinksModel#linkFromKeyProperty}
   * and {@link GraphLinksModel#linkToKeyProperty} properties are set to
   * the corresponding node's data's key values.
   * If the ports have {@link GraphObject#portId} values that are not null,
   * the link data's {@link GraphLinksModel#linkFromPortIdProperty}
   * and {@link GraphLinksModel#linkToPortIdProperty} properties are also set.
   *
   * Setting this property does not raise any events.
   * This property is ignored if the {@link Diagram#model} is not a {@link GraphLinksModel}.
   */
  /** @export */ get archetypeLinkData(): ObjectData | null { return this._archetypeLinkData; }
  /** @export */ set archetypeLinkData(value: ObjectData | null) {
    if (value !== null) Util.checkPrimitive(value, 'object', LinkingTool, 'archetypeLinkData');
    if (value instanceof GraphObject) Util.checkClass(value, Link, LinkingTool, 'archetypeLinkData');
    this._archetypeLinkData = value;
  }

  /**
   * Gets or sets an optional node data object representing a link label, that is copied by {@link #insertLink}
   * and added to the {@link GraphLinksModel} when creating a new {@link Link}.
   *
   * The default value is null, which causes no such label node data to be added to the model
   * along with the new link data.
   *
   * {@link #insertLink} calls {@link GraphLinksModel#addLabelKeyForLinkData}
   * to associate the label node data with the link data.
   *
   * Setting this property does not raise any events.
   * This property is ignored if the {@link Diagram#model} is not a {@link GraphLinksModel}.
   */
  /** @export */ get archetypeLabelNodeData(): ObjectData | null { return this._archetypeLabelNodeData; }
  /** @export */ set archetypeLabelNodeData(value: ObjectData | null) {
    if (value !== null) Util.checkPrimitive(value, 'object', LinkingTool, 'archetypeLabelNodeData');
    if (value instanceof GraphObject) Util.checkClass(value, Node, LinkingTool, 'archetypeLabelNodeData');
    this._archetypeLabelNodeData = value;
  }

  /**
   * Gets or sets the direction in which new links may be drawn.
   * Possible values are {@link LinkingTool.ForwardsOnly}, {@link LinkingTool.BackwardsOnly}, or {@link LinkingTool.Either}.
   * This defaults to {@link LinkingTool.Either}.
   * Setting this property does not raise any events.
   */
  /** @export */ get direction(): EnumValue { return this._direction; }
  /** @export */ set direction(value: EnumValue) {
    EnumValue.checkEnumValue(value, LinkingTool, LinkingTool, 'direction');
    this._direction = value;
  }

  /**
   * Gets or sets the {@link GraphObject} at which {@link #findLinkablePort} should start its search.
   * The default value is null.
   * Setting this property does not raise any events.
   *
   * If you want to explicitly start a new user mouse-gesture to draw a new link
   * from a given {@link GraphObject} that may be a "port" object or may be within
   * the visual tree of a "port" object, set this property to that object
   * to let {@link #findLinkablePort} find the real "port" object.
   * Then start and activate this tool:
   * ```js
   *   var tool = myDiagram.toolManager.linkingTool;
   *   tool.startObject = ...;
   *   myDiagram.currentTool = tool;
   *   tool.doActivate();
   * ```
   */
  /** @export */ get startObject(): GraphObject | null { return this._startObject; }
  /** @export */ set startObject(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkingTool, 'startObject');
    this._startObject = value;
  }

  /**
   * This tool can run when the diagram allows linking, the model is modifiable,
   * the left-button mouse drag has moved far enough away to not be a click, and
   * when {@link #findLinkablePort} has returned a valid port
   * or when {@link #startObject} is a valid port.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {LinkingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowLink) return false;
    const model = diagram.model;
    if (!model.isLinkModel()) return false;
    // require left button & that it has moved far enough away from the mouse down point, so it isn't a click
    if (!diagram.lastInput.left) return false;
    // don't include the following check when this tool is running modally
    if (diagram.currentTool !== this) {
      if (!this.isBeyondDragSize()) return false;
    }
    const port = this.findLinkablePort();
    return (port !== null);
  }

  /**
   * Return the {@link GraphObject} at the mouse-down point,
   * if it is part of a node and if it is valid to link with it.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingTool}
   * @return {GraphObject}
   * If the {@link #direction} is {@link LinkingTool.Either} or {@link LinkingTool.ForwardsOnly},
   * this checks the element and its parent {@link Node} by calling {@link LinkingBaseTool#isValidFrom}.
   * If the {@link #direction} is {@link LinkingTool.Either} or {@link LinkingTool.BackwardsOnly},
   * this checks the element and its parent {@link Node} by calling {@link LinkingBaseTool#isValidTo}.
   * In either case finding a matching port will return that port and set {@link LinkingBaseTool#isForwards} appropriately.
   * Otherwise this will return null.
   */
  public findLinkablePort(): GraphObject | null {
    const diagram = this.diagram;
    let obj = this.startObject;
    if (obj === null) {
      obj = diagram.findObjectAt(diagram.firstInput.documentPoint, null, null);
    }
    if (obj === null) return null;
    const part = obj.part;
    if (!(part instanceof Node)) return null;
    // don't search for valid ports "underneath" the object at the current mouse point;
    // only search up the parent tree of elements for one that isValidFrom
    const dir = this.direction;
    if (dir === LinkingTool.Either || dir === LinkingTool.ForwardsOnly) {
      const port = this.findValidLinkablePort(obj, false);
      if (port !== null) {
        this.isForwards = true;
        return port;
      }
    }
    if (dir === LinkingTool.Either || dir === LinkingTool.BackwardsOnly) {
      const port = this.findValidLinkablePort(obj, true);
      if (port !== null) {
        this.isForwards = false;
        return port;
      }
    }
    return null;
  }

  /**
   * Start the linking operation.
   *
   * If the {@link #startObject} is already set, it uses that object to find the starting port.
   * If it is not set, this calls {@link #findLinkablePort} and remembers it as the starting port.
   *
   * It then starts a transaction, captures the mouse, and changes the cursor.
   * Next it initializes and adds the {@link LinkingBaseTool#temporaryFromNode},
   * {@link LinkingBaseTool#temporaryToNode}, and {@link LinkingBaseTool#temporaryLink} to the diagram.
   * The temporary nodes that are positioned and sized to be like the real {@link LinkingBaseTool#originalFromPort}
   * and {@link LinkingBaseTool#originalToPort} ports.
   * The temporary link connects the two temporary ports, of course.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {LinkingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    const startPort = this.findLinkablePort();
    if (startPort === null) return;

    this.startTransaction(this.name);

    diagram.isMouseCaptured = true;
    diagram.currentCursor = this.linkingCursor;

    if (this.isForwards) {
      if (this.temporaryToNode !== null && !this.temporaryToNode.location.isReal()) {
        this.temporaryToNode.location = diagram.lastInput.documentPoint;
      }
      this.originalFromPort = startPort;
      const node = this.originalFromPort.part;
      if (node instanceof Node) this.originalFromNode = node;
      this.copyPortProperties(this.originalFromNode, this.originalFromPort, this.temporaryFromNode, this.temporaryFromPort, false);
    } else {
      if (this.temporaryFromNode !== null && !this.temporaryFromNode.location.isReal()) {
        this.temporaryFromNode.location = diagram.lastInput.documentPoint;
      }
      this.originalToPort = startPort;
      const node = this.originalToPort.part;
      if (node instanceof Node) this.originalToNode = node;
      this.copyPortProperties(this.originalToNode, this.originalToPort, this.temporaryToNode, this.temporaryToPort, true);
    }

    diagram.add(this.temporaryFromNode);
    diagram.add(this.temporaryToNode);

    if (this.temporaryLink !== null) {
      if (this.temporaryFromNode !== null) {
        this.temporaryLink.fromNode = this.temporaryFromNode;
      }
      if (this.temporaryToNode !== null) {
        this.temporaryLink.toNode = this.temporaryToNode;
      }
      this.temporaryLink.isTreeLink = this.isNewTreeLink();
      this.temporaryLink.invalidateRoute();
      diagram.add(this.temporaryLink);
    }

    this.isActive = true;
  }

  /**
   * Finishing the linking operation stops the transaction, releases the mouse, and resets the cursor.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {LinkingTool}
   */
  public doDeactivate(): void {
    this.isActive = false;
    const diagram = this.diagram;
    diagram.remove(this.temporaryLink);
    diagram.remove(this.temporaryFromNode);
    diagram.remove(this.temporaryToNode);
    diagram.isMouseCaptured = false;
    diagram.currentCursor = '';
    this.stopTransaction();
  }

  /**
   * @hidden @internal
   * Clean up tool state.
   * @this {LinkingTool}
   */
  public doStop(): void {
    super.doStop();
    this.startObject = null;
  }

  /**
   * A mouse-up ends the linking operation; if there is a valid {@link #targetPort} nearby,
   * this adds a new {@link Link} by calling {@link #insertLink}.
   *
   * If there is a new link, it is selected and the "LinkDrawn" {@link DiagramEvent} is raised
   * with the new link as the {@link DiagramEvent#subject}.
   * If there is no new link, this calls {@link #doNoLink}.
   * In any case this stops the tool.
   *
   * This method also raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed in a "Linking" transaction, but the "ChangedSelection" event is raised outside the transaction.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * It is usually easier to override {@link #insertLink} or just set {@link #archetypeLinkData}.
   * It is also common to implement a "LinkDrawn" {@link DiagramEvent} listener on the {@link Diagram}.
   * @this {LinkingTool}
   */
  public doMouseUp(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      this.transactionResult = null;

      let fromnode: Node = null;
      let fromport: GraphObject = null;
      let tonode: Node = null;
      let toport: GraphObject = null;
      let link: Link = null;

      try {
        this.targetPort = this.findTargetPort(this.isForwards);
        const targetport = this.targetPort;
        if (targetport !== null) {
          const tpart = targetport.part;
          if (tpart instanceof Node) {
            if (this.isForwards) {
              if (this.originalFromNode !== null) {
                fromnode = this.originalFromNode;
                fromport = this.originalFromPort;
              }
              tonode = tpart;
              toport = targetport;
            } else {
              fromnode = tpart;
              fromport = targetport;
              if (this.originalToNode !== null) {
                tonode = this.originalToNode;
                toport = this.originalToPort;
              }
            }
          }
        } else {  // not connecting to a port; set FROMNODE or TONODE, but not both
          if (this.isForwards) {
            if (this.originalFromNode !== null && this.isUnconnectedLinkValid) {
              fromnode = this.originalFromNode;
              fromport = this.originalFromPort;
            }
          } else {
            if (this.originalToNode !== null && this.isUnconnectedLinkValid) {
              tonode = this.originalToNode;
              toport = this.originalToPort;
            }
          }
        }
        if (fromnode !== null || tonode !== null) {
          link = this.insertLink(fromnode, fromport, tonode, toport);
          if (link !== null) {
            if (targetport === null) {
              // need to tell Link that it ought to end at Diagram.lastInput.documentPoint
              if (this.isForwards) {
                link.defaultToPoint = diagram.lastInput.documentPoint;
              } else {
                link.defaultFromPoint = diagram.lastInput.documentPoint;
              }
            }
            if (diagram.allowSelect) {
              diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
              link.isSelected = true;
            }
            // set the TransactionResult before raising event, in case it changes the result or cancels the tool
            this.transactionResult = this.name;
            diagram.raiseDiagramEvent('LinkDrawn', link);
          } else {  // insertLink returned null
            diagram.model.clearUnresolvedReferences();
            this.doNoLink(fromnode, fromport, tonode, toport);
          }
        } else {  // no fromnode or tonode
          if (this.isForwards) {
            this.doNoLink(this.originalFromNode, this.originalFromPort, null, null);
          } else {
            this.doNoLink(null, null, this.originalToNode, this.originalToPort);
          }
        }
      } finally {
        this.stopTool();
        if (link && diagram.allowSelect) diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      }
    }
  }

  /**
   * @hidden @internal
   * Note: this does not make an actual copy of a link template and bind the data to it,
   * so any data binding of Link.isTreeLink will not take effect to produce a more accurate answer.
   * @expose
   * @this {LinkingTool}
   * @return {boolean} true iff a newly drawn Link will have Link.isTreeLink true
   */
  protected isNewTreeLink(): boolean {
    const data = this.archetypeLinkData;
    if (data === null) return true;
    if (data instanceof Link) return data.isTreeLink;
    const diagram = this.diagram;
    if (diagram === null) return true;
    const cat = diagram.partManager.getLinkCategoryForData(data);
    const template = diagram.partManager.findLinkTemplateForCategory(cat);
    if (template !== null) return template.isTreeLink;
    return true;
  }

  /**
   * Make a copy of the {@link #archetypeLinkData}, set its node and port properties, and add it to the {@link GraphLinksModel}.
   * If the model is a {@link TreeModel}, set the parent key of the child's node data object.
   *
   * For GraphLinksModels, if {@link #archetypeLabelNodeData} is non-null, this method also adds that node data as an initial label node
   * for the new link data.
   *
   * For TreeModels, if {@link Diagram#isTreePathToChildren} is false, the roles of the *fromnode* and *tonode* arguments are exchanged.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Usually though it is easiest to just set {@link #archetypeLinkData} to control what kind of link is created.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingTool}
   * @param {Node} fromnode
   * @param {GraphObject} fromport the from-port, or null to use the node itself
   * @param {Node} tonode
   * @param {GraphObject} toport the to-port, or null to use the node itself
   * @return {Link} the newly created Link, or null if it failed.
   * @see {@link #doNoLink}
   */
  public insertLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): Link | null {
    return this.diagram.partManager.insertLink(fromnode, fromport, tonode, toport);
  }

  /**
   * This method is called upon a mouse up when no new link is drawn,
   * either because no valid {@link LinkingBaseTool#targetPort} was found or because
   * {@link #insertLink} returned null.
   *
   * This method may be overridden.  By default this method does nothing.
   * If you want to successfully perform any side-effects, you will need to set {@link Tool#transactionResult} to a string;
   * otherwise this tool's transaction will be rolled-back.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkingTool}
   * @param {Node} fromnode
   * @param {GraphObject} fromport the from-port, or null to use the node itself
   * @param {Node} tonode
   * @param {GraphObject} toport the to-port, or null to use the node itself
   * @since 1.7
   */
  public doNoLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): void {
  }

} // #end LinkingTool
LinkingTool['className'] = 'LinkingTool';

LinkingTool['Either'] = LinkingTool.Either;
LinkingTool['ForwardsOnly'] = LinkingTool.ForwardsOnly;
LinkingTool['BackwardsOnly'] = LinkingTool.BackwardsOnly;

/**
 * The RelinkingTool allows the user to reconnect an existing {@link Link}
 * if the {@link Link#relinkableTo} and/or {@link Link#relinkableFrom} properties are true.
 * <p class="boxread">
 * For a general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 *
 * By default an instance of this tool is installed as a mouse-down tool in the
 * {@link Diagram#toolManager} as the {@link ToolManager#relinkingTool}.
 *
 * This tool makes use of two {@link Adornment}s,
 * each including a relink handle (potentially one for each end of the link),
 * shown when a link is selected.
 *
 * This tool conducts a transaction while the tool is active.
 * A successful relinking will result in a "LinkRelinked" {@link DiagramEvent} and a "Relinking" transaction.
 *
 * If you want to programmatically start a new user's relinking of a {@link Link},
 * you can set the {@link #handle} property to the specific "RelinkFrom" or "RelinkTo" handle and
 * then start and activate the tool.
 * ```js
 *   var tool = myDiagram.toolManager.relinkingTool;
 *   tool.originalLink = ...;   // specify which Link to have the user reconnect
 *   tool.isForwards = true;    // specify which end of the Link to reconnect
 *   myDiagram.currentTool = tool;  // starts the RelinkingTool
 *   tool.doActivate();             // activates the RelinkingTool
 * ```
 * @extends LinkingBaseTool
 * @unrestricted
 * @category Tool
 */
export class RelinkingTool extends LinkingBaseTool {
  /** @hidden @internal */ public static className = 'RelinkingTool';
  /** @hidden @internal */ private _fromHandleArchetype: GraphObject | null;
  /** @hidden @internal */ private _toHandleArchetype: GraphObject | null;
  /** @hidden @internal */ private _handle: GraphObject | null;
  /** @hidden @internal */ private _originalLinkBounds: Rect;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#relinkingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "Relinking".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(RelinkingTool);
    super();

    this.name = 'Relinking';

    let h = new Shape();
    h.figure = 'Diamond';
    h.desiredSize = Size.EightSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    h.cursor = this.linkingCursor;
    h.segmentIndex = 0;
    this._fromHandleArchetype = h;

    h = new Shape();
    h.figure = 'Diamond';
    h.desiredSize = Size.EightSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    h.cursor = this.linkingCursor;
    h.segmentIndex = -1;
    this._toHandleArchetype = h;

    // transient state
    this._handle = null;
    this._originalLinkBounds = new Rect();
  }

  /**
   * Show an {@link Adornment} for each end of the {@link Link} that the user may reconnect.
   * The Adornment may have a copy of {@link #fromHandleArchetype} and/or {@link #toHandleArchetype}.
   *
   * This method may be overridden.
   * @this {RelinkingTool}
   * @param {Part} part
   */
  public updateAdornments(part: Part): void {
    if (part === null || !(part instanceof Link)) return;  // tool only applies to Links
    // show handles if link is selected, remove them if no longer selected
    let category = 'RelinkFrom';
    let adornment: Adornment = null;
    if (part.isSelected && !this.diagram.isReadOnly) {
      const selelt = part.selectionObject;
      if (selelt !== null && part.canRelinkFrom() && part.actualBounds.isReal() && part.isVisible() && selelt.actualBounds.isReal() && selelt.isVisibleObject()) {
        adornment = part.findAdornment(category);
        if (adornment === null) {
          adornment = this.makeAdornment(selelt, false);
          // don't set the location, measureLink does that
          part.addAdornment(category, adornment);
        }
      }
    }
    if (adornment === null) part.removeAdornment(category);
    category = 'RelinkTo';
    adornment = null;
    if (part.isSelected && !this.diagram.isReadOnly) {
      const selelt = part.selectionObject;
      if (selelt !== null && part.canRelinkTo() && part.actualBounds.isReal() && part.isVisible() && selelt.actualBounds.isReal() && selelt.isVisibleObject()) {
        adornment = part.findAdornment(category);
        if (adornment === null) {
          adornment = this.makeAdornment(selelt, true);
          // don't set the location, measureLink does that
          part.addAdornment(category, adornment);
        } else {
          adornment.invalidateMeasure();
        }
      }
    }
    if (adornment === null) part.removeAdornment(category);
  }

  //  * Create and return an {@link Adornment} that lets the user grab a handle to reconnect one end of the selected link.
  //  *
  //  * This method may be overridden, but we recommend that you call this base method.
  //  * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {RelinkingTool}
   * @param {GraphObject} selelt the {@link GraphObject} of the {@link Link} being relinked.
   * @param {boolean} toend
   * @return {Adornment}
   */
  public makeAdornment(selelt: GraphObject, toend: boolean): Adornment {
    const adornment = new Adornment();
    adornment.type = Panel.Link;
    const h = (toend ? this.toHandleArchetype : this.fromHandleArchetype);
    if (h !== null) {
      // add a single handle for relinking at one end
      adornment.add(h.copy());
    }
    adornment.adornedObject = selelt;
    return adornment;
  }

  /**
   * Gets or sets a small GraphObject that is copied as a relinking handle for the selected link path
   * at the "from" end of the link.
   * By default this is a {@link Shape} that is a small blue diamond.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handle to be larger green triangles:
   * ```js
   * myDiagram.toolManager.relinkingTool.toHandleArchetype =
   *   $(go.Shape, "Triangle",
   *     { width: 10, height: 10, fill: "limegreen", segmentIndex: 0 });
   * ```
   * @see {@link #toHandleArchetype}
   */
  /** @export */ get fromHandleArchetype(): GraphObject | null { return this._fromHandleArchetype; }
  /** @export */ set fromHandleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, RelinkingTool, 'fromHandleArchetype');
    this._fromHandleArchetype = value;
  }

  /**
   * Gets or sets a small GraphObject that is copied as a relinking handle for the selected link path
   * at the "to" end of the link.
   * By default this is a {@link Shape} that is a small blue diamond.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handle to be larger orange triangles:
   * ```js
   *   myDiagram.toolManager.relinkingTool.toHandleArchetype =
   *   $(go.Shape, "Triangle",
   *     { width: 10, height: 10, fill: "orange", segmentIndex: -1 });
   * ```
   * @see {@link #fromHandleArchetype}
   */
  /** @export */ get toHandleArchetype(): GraphObject | null { return this._toHandleArchetype; }
  /** @export */ set toHandleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, RelinkingTool, 'toHandleArchetype');
    this._toHandleArchetype = value;
  }

  /**
   * Returns the {@link GraphObject} that is the tool handle being dragged by the user.
   * This will be contained by an {@link Adornment} whose category is "RelinkFrom" or "RelinkTo".
   * Its {@link Adornment#adornedPart} is the same as the {@link #originalLink}.
   *
   * This property is also settable, but should only be set either within an override of {@link #doActivate}
   * or prior to calling {@link #doActivate}.
   */
  /** @export */ get handle(): GraphObject | null { return this._handle; }
  /** @export */ set handle(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, RelinkingTool, 'handle');
      if (!(value.part instanceof Adornment)) throw new Error('new handle is not in an Adornment: ' + value);
    }
    this._handle = value;
  }

  /**
   * This tool can run when the diagram allows relinking, the model is modifiable,
   * and there is a relink handle at the mouse-down point.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {RelinkingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowRelink) return false;
    const model = diagram.model;
    if (!model.isLinkModel()) return false;
    if (!diagram.lastInput.left) return false;
    let h = this.findToolHandleAt(diagram.firstInput.documentPoint, 'RelinkFrom');
    if (h === null) h = this.findToolHandleAt(diagram.firstInput.documentPoint, 'RelinkTo');
    return (h !== null);
  }

  /**
   * Start the relinking operation.
   *
   * Find the relink handle by calling {@link Tool#findToolHandleAt}
   * looking for either the "RelinkFrom" adornment or the "RelinkTo" adornment,
   * saving the result in {@link #handle}.
   *
   * Normally when this method is called the value of {@link LinkingBaseTool#originalLink} and {@link #handle} will be null,
   * resulting in a call to {@link Tool#findToolHandleAt} to find a "RelinkFrom" or "RelinkTo" tool handle,
   * which is then remembered as the value of {@link #handle}.
   * If when this method is called the value of {@link #handle} is already set,
   * then there is no need to call {@link Tool#findToolHandleAt},
   * because the programmer has already set up which relinking handle they want the user to be relinking.
   * Finding a handle is not necessary if {@link LinkingBaseTool#originalLink} and
   * {@link #isForwards} have been set before calling this method.
   *
   * This starts a transaction, captures the mouse, and sets the cursor.
   *
   * If {@link LinkingBaseTool#originalLink} or {@link #handle} was not set beforehand,
   * the value of {@link #isForwards} is set depending on the category of the relink handle found.
   * The {@link LinkingBaseTool#originalLink} property and various
   * "Original..." port and node properties are set too.
   * The temporary nodes and temporary link are also initialized.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {RelinkingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this.originalLink === null) {
      let h = this.handle;
      if (h === null) {
        h = this.findToolHandleAt(diagram.firstInput.documentPoint, 'RelinkFrom');
        if (h === null) h = this.findToolHandleAt(diagram.firstInput.documentPoint, 'RelinkTo');
      }
      if (h === null) return;

      const part = h.part;
      if (!(part instanceof Adornment)) return;
      if (!(part.adornedPart instanceof Link)) return;
      this.handle = h;

      // "forwards" means the user started dragging the RelinkHandle at the "To" end of the link
      this.isForwards = (part === null || part.category === 'RelinkTo');

      this.originalLink = part.adornedPart;
    }

    this.startTransaction(this.name);

    diagram.isMouseCaptured = true;
    diagram.currentCursor = this.linkingCursor;
    this.originalFromPort = this.originalLink.fromPort;
    this.originalFromNode = this.originalLink.fromNode;
    this.originalToPort = this.originalLink.toPort;
    this.originalToNode = this.originalLink.toNode;
    this._originalLinkBounds.set(this.originalLink.actualBounds);

    // in case either end is not connected to a node/port, initialize the temporary node/port
    if (this.originalLink !== null && this.originalLink.pointsCount > 0) {
      if (this.originalLink.fromNode === null) {
        if (this.temporaryFromPort !== null) {
          this.temporaryFromPort.desiredSize = Size.ZeroSize;
        }
        if (this.temporaryFromNode !== null) {
          this.temporaryFromNode.location = this.originalLink.getPoint(0);
        }
      }
      if (this.originalLink.toNode === null) {
        if (this.temporaryToPort !== null) {
          this.temporaryToPort.desiredSize = Size.ZeroSize;
        }
        if (this.temporaryToNode !== null) {
          this.temporaryToNode.location = this.originalLink.getPoint(this.originalLink.pointsCount - 1);
        }
      }
    }

    // initialize temporary nodes/ports
    this.copyPortProperties(this.originalFromNode, this.originalFromPort, this.temporaryFromNode, this.temporaryFromPort, false);
    this.copyPortProperties(this.originalToNode, this.originalToPort, this.temporaryToNode, this.temporaryToPort, true);

    diagram.add(this.temporaryFromNode);
    diagram.add(this.temporaryToNode);

    if (this.temporaryLink !== null) {
      if (this.temporaryFromNode !== null) {
        this.temporaryLink.fromNode = this.temporaryFromNode;
      }
      if (this.temporaryToNode !== null) {
        this.temporaryLink.toNode = this.temporaryToNode;
      }
      this.copyLinkProperties(this.originalLink, this.temporaryLink);
      this.temporaryLink.invalidateRoute();
      diagram.add(this.temporaryLink);
    }

    this.isActive = true;
  }

  /**
   * Make a temporary link look and act like the real Link being relinked.
   * By default this method copies many of the routing-oriented properties
   * from the {@link LinkingBaseTool#originalLink} to the {@link LinkingBaseTool#temporaryLink}.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RelinkingTool}
   * @param {Link} reallink
   * @param {Link} templink
   * @since 1.3
   */
  protected copyLinkProperties(reallink: Link | null, templink: Link): void {
    if (reallink === null || templink === null) return;
    templink.adjusting = reallink.adjusting;
    templink.corner = reallink.corner;
    let curve = reallink.curve;
    if (curve === Link.JumpOver || curve === Link.JumpGap) curve = Link.None;  // don't copy "jumping"
    templink.curve = curve;
    templink.curviness = reallink.curviness;
    templink.isTreeLink = reallink.isTreeLink;
    templink.points = reallink.points;
    templink.routing = reallink.routing;
    templink.smoothness = reallink.smoothness;
    templink.fromSpot = reallink.fromSpot;
    templink.fromEndSegmentLength = reallink.fromEndSegmentLength;
    templink.fromShortLength = reallink.fromShortLength;
    templink.toSpot = reallink.toSpot;
    templink.toEndSegmentLength = reallink.toEndSegmentLength;
    templink.toShortLength = reallink.toShortLength;
  }

  /**
   * Finishing the linking operation stops the transaction, releases the mouse, and resets the cursor.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * @this {RelinkingTool}
   */
  public doDeactivate(): void {
    this.isActive = false;
    const diagram = this.diagram;
    diagram.remove(this.temporaryLink);
    diagram.remove(this.temporaryFromNode);
    diagram.remove(this.temporaryToNode);
    diagram.isMouseCaptured = false;
    diagram.currentCursor = '';
    this.stopTransaction();
  }

  /**
   * @hidden @internal
   * Clean up tool state.
   * @this {RelinkingTool}
   */
  public doStop(): void {
    super.doStop();
    this.handle = null;
  }

  /**
   * A mouse-up ends the relinking operation; if there is a valid {@link #targetPort} nearby,
   * this modifies the old link to connect with the target port.
   *
   * A successful relinking calls {@link #reconnectLink} to actually change the link.
   * The "LinkRelinked" {@link DiagramEvent} is raised with the link as the {@link DiagramEvent#subject}
   * and with the now-disconnected original port as the {@link DiagramEvent#parameter}.
   * If the link was not reconnected, this calls {@link #doNoRelink}.
   * In any case this stops the tool.
   *
   * A failure to find a valid target port results in no changes and no DiagramEvent.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * You might find it easier to override {@link #reconnectLink}.
   * It is actually most common to implement a "LinkRelinked" {@link DiagramEvent} listener on the {@link Diagram}.
   * @this {RelinkingTool}
   */
  public doMouseUp(): void {
    if (this.isActive) {
      const diagram = this.diagram;
      this.transactionResult = null;

      let newfromnode = this.originalFromNode;
      let newfromport = this.originalFromPort;
      let newtonode = this.originalToNode;
      let newtoport = this.originalToPort;

      let origlink = this.originalLink;
      this.targetPort = this.findTargetPort(this.isForwards);
      if (this.targetPort !== null) {
        const tpart = this.targetPort.part;
        if (tpart instanceof Node) {
          if (this.isForwards) {
            newtonode = tpart;
            newtoport = this.targetPort;
          } else {
            newfromnode = tpart;
            newfromport = this.targetPort;
          }
        }
      } else { // no target port found, disconnect if allowed
        if (this.isUnconnectedLinkValid) {
          if (this.isForwards) {
            newtonode = null;
            newtoport = null;
          } else {
            newfromnode = null;
            newfromport = null;
          }
        } else {
          origlink = null;
          // don't set transactionResult and let any changes be rolled back
        }
      }

      if (origlink !== null) {
        this.reconnectLink(origlink, (this.isForwards ? newtonode : newfromnode),
          (this.isForwards ? newtoport : newfromport), this.isForwards);

        if (this.targetPort === null) {
          // need to tell Link that it ought to end at last mouse point
          if (this.isForwards) {
            origlink.defaultToPoint = diagram.lastInput.documentPoint;
          } else {
            origlink.defaultFromPoint = diagram.lastInput.documentPoint;
          }
          origlink.invalidateRoute();
        }
        // select the link
        if (diagram.allowSelect) origlink.isSelected = true;
        // set the TransactionResult before raising event, in case it changes the result or cancels
        this.transactionResult = this.name;
        diagram.raiseDiagramEvent('LinkRelinked', origlink, (this.isForwards ? this.originalToPort : this.originalFromPort));
      } else {
        this.doNoRelink(this.originalLink, this.isForwards);
      }

      // fix any jump-overs
      this.originalLink.invalidateOtherJumpOvers(this._originalLinkBounds);
    }
    this.stopTool();
  }

  /**
   * Modify an existing {@link Link} to connect to a new node and port.
   *
   * This method may be overridden, but we recommend that you call this base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RelinkingTool}
   * @param {Link} existinglink
   * @param {Node} newnode the {@link Node} to connect to or from.
   * @param {GraphObject} newport the {@link GraphObject} port to connect to or from.
   * @param {boolean} toend If true, this modifies the link's "to" node and port; otherwise it modifies the "from" node and port.
   * @return {boolean} true if successful.
   * @see {@link #doNoRelink}
   */
  public reconnectLink(existinglink: Link, newnode: Node | null, newport: GraphObject | null, toend: boolean): boolean {
    const newportid = ((newport !== null && newport.portId !== null) ? newport.portId : '');
    if (toend) {
      existinglink.toNode = newnode;
      existinglink.toPortId = newportid;
    } else {
      existinglink.fromNode = newnode;
      existinglink.fromPortId = newportid;
    }
    return true;
  }

  /**
   * This method is called upon a mouse up when {@link #reconnectLink} is not called.
   *
   * This method may be overridden.  By default this method does nothing.
   * If you want to successfully perform any side-effects, you will need to set {@link Tool#transactionResult} to a string;
   * otherwise this tool's transaction will be rolled-back.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RelinkingTool}
   * @param {Link} existinglink
   * @param {boolean} toend If true, the user was trying to modify the link's "to" node and port.
   * @since 1.7
   */
  public doNoRelink(existinglink: Link, toend: boolean): void {
  }

  /**
   * @hidden @internal
   * @this {RelinkingTool}
   * @param {Node} fromnode
   * @param {GraphObject} fromport
   * @param {Node} tonode
   * @param {GraphObject} toport
   */
  public doDraggingMouseMove(fromnode: Node, fromport: GraphObject, tonode: Node, toport: GraphObject): void {
    if (fromnode !== null) {
      this.copyPortProperties(fromnode, fromport, this.temporaryFromNode, this.temporaryFromPort, false);
      this.diagram.add(this.temporaryFromNode);
    } else {
      this.diagram.remove(this.temporaryFromNode);
    }
    if (tonode !== null) {
      this.copyPortProperties(tonode, toport, this.temporaryToNode, this.temporaryToPort, true);
      this.diagram.add(this.temporaryToNode);
    } else {
      this.diagram.remove(this.temporaryToNode);
    }
  }

  /**
   * @hidden @internal
   * @this {RelinkingTool}
   */
  public stopDraggingMouseMove(): void {
    this.diagram.remove(this.temporaryFromNode);
    this.diagram.remove(this.temporaryToNode);
  }

} // #end RelinkingTool
RelinkingTool['className'] = 'RelinkingTool';


// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'linkingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): LinkingTool {
    return this.findTool('Linking') as LinkingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: LinkingTool) { this.replaceStandardTool('Linking', value, this.mouseMoveTools); }
);

Util.defineProperty(ToolManager, 'relinkingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): RelinkingTool {
    return this.findTool('Relinking') as RelinkingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: RelinkingTool) { this.replaceStandardTool('Relinking', value, this.mouseDownTools); }
);


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
LinkReshapingTool
*/

// import { List } from './collections';
// import { EnumValue } from './enumValue';
// import { Geo } from './geo';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { Adornment, Part } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Tool } from './tool';
// import { Util } from './util';

 //
 // If you want to programmatically start a user's reshaping of an existing selected link,
 // you can set the {@link #handle} property to the specific reshape handle and then start and activate the tool.
 // ```js
 //   var link = ...;
 //   myDiagram.select(link);
 //   var tool = myDiagram.toolManager.linkReshapingTool;
 //   // specify which reshape handle of the "Reshaping" Adornment of the Link
 //   // to be reshaped; for this example, choose the second handle
 //   tool.handle = link.findAdornment("LinkReshaping").elt(1);
 //   myDiagram.currentTool = tool;  // starts the LinkReshapingTool
 //   tool.doActivate();             // activates the LinkReshapingTool
 // ```

/**
 * The LinkReshapingTool is used to interactively change the route of a {@link Link}
 * by setting its {@link Link#points} list.
 * You may want to save the route to the model by using a TwoWay {@link Binding} on the "points" property of the Link.
 *
 * This tool makes use of an {@link Adornment}, shown when the adorned {@link Link} is selected,
 * that includes some number of reshape handles.
 * This tool conducts a transaction while the tool is active.
 * A successful reshaping will result in a "LinkReshaped" {@link DiagramEvent} and a "LinkReshaping" transaction.
 *
 * <p class="boxread">
 * For a general discussion of link routing, see:
 * <a href="../../intro/links.html">Introduction to Links</a>,
 * <a href="../../intro/linkLabels.html">Introduction to Link Labels</a>, and
 * <a href="../../intro/connectionPoints.html">Introduction to Link Connection Points</a>.
 * For customizing the linking tools, see <a href="../../intro/tools.html#LinkingAndRelinkingTools">Introduction to the Linking Tools</a>.
 * For a general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class LinkReshapingTool extends Tool {
  /** @hidden @internal */ public static className = 'LinkReshapingTool';
  /** @hidden @internal */ private _handleArchetype: GraphObject | null;
  /** @hidden @internal */ private _midHandleArchetype: GraphObject | null;
  /** @hidden @internal */ private _resegmentingDistance: number;
  /** @hidden @internal */ private _handle: GraphObject | null;
  /** @hidden @internal */ private _adornedLink: Link | null;
  /** @hidden @internal */ private _originalPoint: Point;
  /** @hidden @internal */ private _originalPoints: List<Point>;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#linkReshapingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "LinkReshaping".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(LinkReshapingTool);
    super();

    this.name = 'LinkReshaping';

    let h = new Shape();
    h.figure = 'Rectangle';
    h.desiredSize = Size.SixSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    this._handleArchetype = h;

    h = new Shape();
    h.figure = 'Diamond';
    h.desiredSize = Size.EightSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    h.cursor = 'move';
    this._midHandleArchetype = h;

    this._resegmentingDistance = 3;

    // internal state
    this._handle = null;
    this._adornedLink = null;

    this._originalPoint = new Point();
    this._originalPoints = new List<Point>();  // in case the tool is cancelled and the UndoManager is not enabled
  }

  /**
   * Disallow dragging.
   * @constant
   */
  public static None = new EnumValue(LinkReshapingTool, 'None', 0);

  /**
   * Allow only horizontal (left-and-right) dragging.
   * @constant
   */
  public static Horizontal = new EnumValue(LinkReshapingTool, 'Horizontal', 1);

  /**
   * Allow only vertical (up-and-down) dragging.
   * @constant
   */
  public static Vertical = new EnumValue(LinkReshapingTool, 'Vertical', 2);

  /**
   * Allow dragging in any direction.
   * @constant
   */
  public static All = new EnumValue(LinkReshapingTool, 'All', 3);

  /**
   * Get the permitted reshaping behavior for a particular reshape handle.
   * @export
   * @this {LinkReshapingTool}
   * @param {GraphObject} obj a reshape handle in the "LinkReshaping" Adornment.
   * @return {EnumValue} one of LinkReshapingTool.All, .Vertical, .Horizontal, or .None
   * @since 1.3
   */
  protected getReshapingBehavior(obj: GraphObject): EnumValue {
    if (!obj || !obj._reshapingBehavior || obj._reshapingBehavior.value === 0) return LinkReshapingTool.None;
    return obj._reshapingBehavior;
  }

  /**
   * Set the permitted reshaping behavior for a particular reshape handle.
   * @export
   * @this {LinkReshapingTool}
   * @param {GraphObject} obj a reshape handle in the "LinkReshaping" Adornment.
   * @param {EnumValue} behavior one of LinkReshapingTool.All, .Vertical, .Horizontal, or .None
   * @since 1.3
   */
  protected setReshapingBehavior(obj: GraphObject, behavior: EnumValue) {
    Util.checkClass(obj, GraphObject, LinkReshapingTool, 'setReshapingBehavior:obj');
    EnumValue.checkEnumValue(behavior, LinkReshapingTool, LinkReshapingTool, 'setReshapingBehavior:behavior');
    obj._reshapingBehavior = behavior;
  }

  /**
   * Show an {@link Adornment} with reshape handles at each of the interesting points of the link's route,
   * if the link is selected and visible and if {@link Part#canReshape} is true.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkReshapingTool}
   * @param {Part} part
   */
  public updateAdornments(part: Part): void {
    if (part === null || !(part instanceof Link)) return;  // tool only applies to Links
    let adornment: Adornment = null;
    if (part.isSelected && !this.diagram.isReadOnly) {
      const pathshape = part.path;
      if (pathshape !== null && part.canReshape() && part.actualBounds.isReal() && part.isVisible() && pathshape.actualBounds.isReal() && pathshape.isVisibleObject()) {
        adornment = part.findAdornment(this.name);
        if (adornment === null || adornment._oldPointsCount !== part.pointsCount || adornment._wasResegmentable !== part.resegmentable) {
          adornment = this.makeAdornment(pathshape);
          if (adornment !== null) {
            adornment._oldPointsCount = part.pointsCount;
            adornment._wasResegmentable = part.resegmentable;
            part.addAdornment(this.name, adornment);
          }
        }
      }
    }
    if (adornment === null) part.removeAdornment(this.name);
  } // end updateAdornments

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {LinkReshapingTool}
   * @param {GraphObject} pathshape
   * @return {Adornment}
   */
  public makeAdornment(pathshape: GraphObject): Adornment {
    const link = pathshape.part as Link;
    const pts = link.points;
    const numpts = link.pointsCount;
    const ortho = link.isOrthogonal;
    let adornment: Adornment = null;
    if (pts !== null && numpts > 1) {
      adornment = new Adornment();
      adornment.type = Panel.Link;

      const firstindex = link.firstPickIndex;
      const lastindex = link.lastPickIndex;
      const orthadjust = (ortho ? 1 : 0);
      if (link.resegmentable && link.computeCurve() !== Link.Bezier) {
        for (let i = firstindex + orthadjust; i < lastindex - orthadjust; i++) {
          const h = this.makeResegmentHandle(pathshape, i);
          // needs to be a GraphObject so we can set its Cursor
          if (h !== null) {
            // identify this particular handle within the LinkPanel
            h.segmentIndex = i;
            h.segmentFraction = 0.5;
            h.fromMaxLinks = 999; // set this unused property to easily identify that we have a resegmenting handle
            adornment.add(h);
          }
        }
      }
      // don't bother creating handles for firstindex or lastindex
      for (let i = firstindex + 1; i < lastindex; i++) {
        const h = this.makeHandle(pathshape, i);
        // needs to be a GraphObject so we can set its Cursor
        if (h !== null) {
          // identify this particular handle within the LinkPanel
          h.segmentIndex = i;
          // now determines its reshape behavior and cursor, depending on whether Orthogonal et al.
          if (i === firstindex) {
            // default LinkReshapingBehavior.None
          } else if (i === firstindex + 1 && ortho) {
            const a = link.getPoint(firstindex);
            let b = link.getPoint(firstindex + 1);
            if (Geo.isApprox(a.x, b.x) && Geo.isApprox(a.y, b.y)) {
              b = link.getPoint(firstindex - 1);
              if (Geo.isApprox(a.x, b.x)) {
                this.setReshapingBehavior(h, LinkReshapingTool.Vertical);
                h.cursor = 'n-resize';
              } else if (Geo.isApprox(a.y, b.y)) {
                this.setReshapingBehavior(h, LinkReshapingTool.Horizontal);
                h.cursor = 'w-resize';
              }
            } else if (Geo.isApprox(a.x, b.x)) {
              this.setReshapingBehavior(h, LinkReshapingTool.Vertical);
              h.cursor = 'n-resize';
            } else if (Geo.isApprox(a.y, b.y)) {
              this.setReshapingBehavior(h, LinkReshapingTool.Horizontal);
              h.cursor = 'w-resize';
            }
          } else if (i === lastindex - 1 && ortho) {
            let a = link.getPoint(lastindex - 1);
            const b = link.getPoint(lastindex);
            if (Geo.isApprox(a.x, b.x) && Geo.isApprox(a.y, b.y)) {
              a = link.getPoint(lastindex + 1);
              if (Geo.isApprox(a.x, b.x)) {
                this.setReshapingBehavior(h, LinkReshapingTool.Vertical);
                h.cursor = 'n-resize';
              } else if (Geo.isApprox(a.y, b.y)) {
                this.setReshapingBehavior(h, LinkReshapingTool.Horizontal);
                h.cursor = 'w-resize';
              }
            } else if (Geo.isApprox(a.x, b.x)) {
              this.setReshapingBehavior(h, LinkReshapingTool.Vertical);
              h.cursor = 'n-resize';
            } else if (Geo.isApprox(a.y, b.y)) {
              this.setReshapingBehavior(h, LinkReshapingTool.Horizontal);
              h.cursor = 'w-resize';
            }
          } else if (i === lastindex) {
            // default LinkReshapingBehavior.None
          } else {
            this.setReshapingBehavior(h, LinkReshapingTool.All);
            h.cursor = 'move';
          }
          adornment.add(h);
        }
      }
      adornment.adornedObject = pathshape;
    }
    return adornment;
  }

  //    * Create and return a {@link GraphObject} that the user can "grab" to relink the selected Link.
  //  * By default this returns a copy of {@link #handleArchetype}, a {@link Shape} that is a small blue rectangle.
  //  * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {LinkReshapingTool}
   * @param {GraphObject} pathshape the {@link GraphObject} path of the link being reshaped.
   * @param {number} idx the index of the route point to be moved.
   * @return {GraphObject}
   */
  public makeHandle(pathshape: GraphObject, idx: number): GraphObject {
    const h = this.handleArchetype;
    if (h === null) return null;
    return h.copy();
  }

  /**
   * Gets or sets a small GraphObject that is copied as a reshape handle at each movable point in the selected link's route.
   * By default this is a {@link Shape} that is a small blue rectangle.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handles to be larger yellow circles:
   * ```js
   *   myDiagram.toolManager.linkReshapingTool.handleArchetype =
   *     $(go.Shape, "Circle",
   *       { width: 10, height: 10, fill: "yellow" });
   * ```
   * @see {@link #midHandleArchetype}
   */
  /** @export */ get handleArchetype(): GraphObject | null { return this._handleArchetype; }
  /** @export */ set handleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkReshapingTool, 'handleArchetype');
    this._handleArchetype = value;
  }

  //    * Create and return a {@link GraphObject} that the user can "grab" to add a segment to the selected Link.
  //  * By default this returns a copy of {@link #midHandleArchetype}, a {@link Shape} that is a small blue diamond.
  //  * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {LinkReshapingTool}
   * @param {GraphObject} pathshape the {@link GraphObject} path of the link being reshaped.
   * @param {number} idx the index of the route point to be moved.
   * @return {GraphObject}
   */
  public makeResegmentHandle(pathshape: GraphObject, idx: number): GraphObject {
    const h = this.midHandleArchetype;
    if (h === null) return null;
    return h.copy();
  }

  /**
   * Gets or sets a small GraphObject that is copied as a resegment handle at each mid-point in the selected Link's route.
   * By default this is a {@link Shape} that is a small blue diamond.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handle to be larger cyan triangles:
   * ```js
   *   myDiagram.toolManager.linkReshapingTool.midHandleArchetype =
   *     $(go.Shape, "Triangle", { width: 10, height: 10, fill: "cyan" });
   * ```
   * @see {@link #handleArchetype}
   * @since 1.2
   */
  /** @export */ get midHandleArchetype(): GraphObject | null { return this._midHandleArchetype; }
  /** @export */ set midHandleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, LinkReshapingTool, 'midHandleArchetype');
    this._midHandleArchetype = value;
  }

  // This property is also settable, but should only be set either within an override of {@link #doActivate}
  // or prior to calling {@link #doActivate}.

  /**
   * Returns the {@link GraphObject} that is the tool handle being dragged by the user.
   * This will be contained by an {@link Adornment} whose category is "LinkReshaping".
   * Its {@link Adornment#adornedPart} is the same as the {@link #adornedLink}.
   * This is normally set by {@link #doActivate}, remembering the result of the call to {@link Tool#findToolHandleAt}.
   */
  /** @export */ get handle(): GraphObject | null { return this._handle; }
  /** @export */ set handle(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, LinkReshapingTool, 'handle');
      if (!(value.part instanceof Adornment)) throw new Error('new handle is not in an Adornment: ' + value);
    }
    this._handle = value;
  }

  /**
   * This read-only property returns the {@link Link} that is being routed manually.
   */
  /** @export */ get adornedLink(): Link | null { return this._adornedLink; }

  /**
   * This tool may run when there is a mouse-down event on a reshape handle.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkReshapingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;

    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (!diagram.allowReshape) return false;
    if (!diagram.lastInput.left) return false;
    const h = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
    return (h !== null);
  }

  /**
   * Start reshaping, if {@link #findToolHandleAt} finds a reshape handle at the mouse down point.
   *
   * If successful this sets {@link #handle} to be the reshape handle that it finds
   * and {@link #adornedLink} to be the {@link Link} being routed.
   * It also remembers the original link route (a list of Points) in case this tool is cancelled.
   * And it starts a transaction.
   * @this {LinkReshapingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this.handle === null) {
      this.handle = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
    }
    if (this.handle === null) return;
    const part = (this.handle.part as Adornment).adornedPart;
    if (!(part instanceof Link)) return;
    this._adornedLink = part;
    diagram.isMouseCaptured = true;

    this.startTransaction(this.name);

    if (part.resegmentable && this.handle.fromMaxLinks === 999) {  // a hack to identify resegmenting handles
      const pts = part.points.copy();
      const hpt = this.getResegmentingPoint();
      pts.insertAt(this.handle.segmentIndex + 1, hpt);
      if (part.isOrthogonal) {
        pts.insertAt(this.handle.segmentIndex + 1, hpt);
      }
      part.points = pts;
      part.invalidateAdornments();
      part.updateAdornments();  // make sure that the new reshaping handle can be found
      this.handle = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
      if (this.handle === null) {
        this.doDeactivate();  // need to rollback the transaction and not set .isActive
        return;
      }
    }

    this._originalPoint = part.getPoint(this.handle.segmentIndex);
    this._originalPoints = part.points.copy();
    this.isActive = true;
  }

  /**
   * This stops the current reshaping operation with the link route shaped the way it is.
   * @this {LinkReshapingTool}
   */
  public doDeactivate(): void {
    this.stopTransaction();

    this.handle = null;
    this._adornedLink = null;
    const diagram = this.diagram;
    diagram.isMouseCaptured = false;
    this.isActive = false;
  }

  /**
   * Restore the link route to be the original points and stop this tool.
   * @this {LinkReshapingTool}
   */
  public doCancel(): void {
    const link = this.adornedLink;
    if (link !== null) {
      // explicitly restore the original route, in case !UndoManager.isEnabled
      link.points = this._originalPoints;
    }
    this.stopTool();
  }

  // * If the handle clicked during activation was a resegment handle, return the point at which to start a new segment.
  // * By default this returns the center of the handle that was clicked.
  // * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {LinkReshapingTool}
   * @return {Point} the point where resegmenting begins
   */
  protected getResegmentingPoint(): Point {
    return this.handle.getDocumentPoint(Spot.Center);
  }

  /**
   * Call {@link #reshape} with a new point determined by the mouse
   * to change the route of the {@link #adornedLink}.
   * @this {LinkReshapingTool}
   */
  public doMouseMove(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      const newpt = this.computeReshape(diagram.lastInput.documentPoint);
      this.reshape(newpt);
    }
  }

  /**
   * Reshape the route with a point based on the most recent mouse point by calling {@link #reshape},
   * and then raise a "LinkReshaped" {@link DiagramEvent} before stopping this tool.
   * @this {LinkReshapingTool}
   */
  public doMouseUp(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      const newpt = this.computeReshape(diagram.lastInput.documentPoint);
      this.reshape(newpt);

      const link = this.adornedLink;
      if (link !== null && link.resegmentable) {
        const idx = this.handle.segmentIndex;
        const a = link.getPoint(idx - 1);
        const b = link.getPoint(idx);
        const c = link.getPoint(idx + 1);
        // see if we should delete a segment because both adjacent segments are co-linear
        if (link.isOrthogonal) {
          if (idx > link.firstPickIndex + 1 && idx < link.lastPickIndex - 1) {
            let q = link.getPoint(idx - 2);
            if (this.isWithinResegmentingDistance(a, b) &&
              (this.isInLineOrtho(q, a, b, c, true) || this.isInLineOrtho(q, a, b, c, false))) {
              const pts = link.points.copy();
              // re-align remaining two points Q and C to maintain orthogonality
              if (this.isInLineOrtho(q, a, b, c, true)) {
                pts.setElt(idx - 2, new Point(q.x, (c.y + q.y) / 2));
                pts.setElt(idx + 1, new Point(c.x, (c.y + q.y) / 2));
              } else {
                pts.setElt(idx - 2, new Point((c.x + q.x) / 2, q.y));
                pts.setElt(idx + 1, new Point((c.x + q.x) / 2, c.y));
              }
              // remove points A and B
              pts.removeAt(idx);  // order matters!
              pts.removeAt(idx - 1);
              link.points = pts;
              link.invalidateAdornments();
            } else {
              q = link.getPoint(idx + 2);
              if (this.isWithinResegmentingDistance(b, c) &&
                (this.isInLineOrtho(a, b, c, q, true) || this.isInLineOrtho(a, b, c, q, false))) {
                const pts = link.points.copy();
                // re-align remaining two points A and Q to maintain orthogonality
                if (this.isInLineOrtho(a, b, c, q, true)) {
                  pts.setElt(idx - 1, new Point(a.x, (a.y + q.y) / 2));
                  pts.setElt(idx + 2, new Point(q.x, (a.y + q.y) / 2));
                } else {
                  pts.setElt(idx - 1, new Point((a.x + q.x) / 2, a.y));
                  pts.setElt(idx + 2, new Point((a.x + q.x) / 2, q.y));
                }
                // remove points B and C
                pts.removeAt(idx + 1);  // order matters!
                pts.removeAt(idx);
                link.points = pts;
                link.invalidateAdornments();
              }
            }
          }
        } else {
          const q = Point.alloc();  // the closest point to B on the line between A and C
          if (Geo.nearestPointOnLine(a.x, a.y, c.x, c.y, b.x, b.y, q)) {
            // if B is within 3 units of the line from A to C, and if Q is between A and C,
            // remove that point from the route
            if (q.distanceSquaredPoint(b) < this.resegmentingDistance * this.resegmentingDistance) {
              const pts = link.points.copy();
              pts.removeAt(idx);
              link.points = pts;
              link.invalidateAdornments();
            }
          }
          Point.free(q);
        }
      }

      diagram.invalidateDocumentBounds();
      // set the EditResult before raising event, in case it changes the result or cancels the tool
      this.transactionResult = this.name;
      diagram.raiseDiagramEvent('LinkReshaped', this.adornedLink, this._originalPoints);
    }
    this.stopTool();
  }

  /**
   * @hidden @internal
   * @this {LinkReshapingTool}
   * @param {Point} p1
   * @param {Point} p2
   * @return {boolean}
   */
  protected isWithinResegmentingDistance(p1: Point, p2: Point): boolean {
    return Math.abs(p1.x - p2.x) < this.resegmentingDistance && Math.abs(p1.y - p2.y) < this.resegmentingDistance;
  }

  /**
   * @hidden @internal
   * @this {LinkReshapingTool}
   * @param {Point} a
   * @param {Point} b
   * @param {Point} c
   * @param {Point} d
   * @param {boolean} horiz
   * @return {boolean}
   */
  protected isInLineOrtho(a: Point, b: Point, c: Point, d: Point, horiz: boolean): boolean {
    if (horiz) {
      return (Math.abs(a.y - b.y) < this.resegmentingDistance &&
        Math.abs(b.y - c.y) < this.resegmentingDistance &&
        Math.abs(c.y - d.y) < this.resegmentingDistance);
    } else {
      return (Math.abs(a.x - b.x) < this.resegmentingDistance &&
        Math.abs(b.x - c.x) < this.resegmentingDistance &&
        Math.abs(c.x - d.x) < this.resegmentingDistance);
    }
  }

  /**
   * @hidden @internal
   * Gets or sets the distance from a straight line that controls whether a
   * {@link Link#resegmentable} Link segment is kept as a separate segment or
   * removed to form a straight segment of the adjacent points.
   *
   * The default value is 3.
   */
  /** @export */ get resegmentingDistance(): number { return this._resegmentingDistance; }
  /** @export */ set resegmentingDistance(value: number) {
    Util.checkPrimitive(value, 'number', LinkReshapingTool, 'resegmentingDistance');
    this._resegmentingDistance = value;
  }

  /**
   * Change the route of the {@link #adornedLink} by moving the point corresponding to the current
   * {@link #handle} to be at the given {@link Point}.
   * This is called by {@link #doMouseMove} and {@link #doMouseUp} with the result of calling
   * {@link #computeReshape} to constrain the input point.
   *
   * This method respects the orthogonality of the link, if necessary limiting movement along only one axis.
   * To maintain orthogonality it may need to modify more than one point in the route.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkReshapingTool}
   * @param {Point} newPoint the value of the call to {@link #computeReshape}.
   */
  public reshape(newPoint: Point): void {
    const link = this.adornedLink;
    link.startRoute();
    let index = this.handle.segmentIndex;
    const behavior = this.getReshapingBehavior(this.handle);
    if (link.isOrthogonal) {  // need to adjust adjacent points as well
      if (index === link.firstPickIndex + 1) {
        const midfirst = link.firstPickIndex + 1;
        if (behavior === LinkReshapingTool.Vertical) {
          // move segment vertically
          link.setPointAt(midfirst, link.getPoint(midfirst - 1).x, newPoint.y);
          link.setPointAt(midfirst + 1, link.getPoint(midfirst + 2).x, newPoint.y);
        } else if (behavior === LinkReshapingTool.Horizontal) {
          // move segment horizontally
          link.setPointAt(midfirst, newPoint.x, link.getPoint(midfirst - 1).y);
          link.setPointAt(midfirst + 1, newPoint.x, link.getPoint(midfirst + 2).y);
        }
      } else if (index === link.lastPickIndex - 1) {
        const midlast = link.lastPickIndex - 1;
        if (behavior === LinkReshapingTool.Vertical) {
          // move segment vertically
          link.setPointAt(midlast - 1, link.getPoint(midlast - 2).x, newPoint.y);
          link.setPointAt(midlast, link.getPoint(midlast + 1).x, newPoint.y);
        } else if (behavior === LinkReshapingTool.Horizontal) {
          // move segment horizontally
          link.setPointAt(midlast - 1, newPoint.x, link.getPoint(midlast - 2).y);
          link.setPointAt(midlast, newPoint.x, link.getPoint(midlast + 1).y);
        }
      } else {
        // can move anywhere, but need to keep adjacent segments orthogonal
        let i = index;
        const oldpt = link.getPoint(i);
        const before = link.getPoint(i - 1);
        const after = link.getPoint(i + 1);
        if (Geo.isApprox(before.x, oldpt.x) && Geo.isApprox(oldpt.y, after.y)) {
          if (Geo.isApprox(before.x, link.getPoint(i - 2).x) && !Geo.isApprox(before.y, link.getPoint(i - 2).y)) {
            link.insertPointAt(i, newPoint.x, before.y);
            index++;
            i++;
          } else link.setPointAt(i - 1, newPoint.x, before.y);
          if (Geo.isApprox(after.y, link.getPoint(i + 2).y) && !Geo.isApprox(after.x, link.getPoint(i + 2).x)) {
            link.insertPointAt(i + 1, after.x, newPoint.y);
          } else link.setPointAt(i + 1, after.x, newPoint.y);
        } else if (Geo.isApprox(before.y, oldpt.y) && Geo.isApprox(oldpt.x, after.x)) {
          if (Geo.isApprox(before.y, link.getPoint(i - 2).y) && !Geo.isApprox(before.x, link.getPoint(i - 2).x)) {
            link.insertPointAt(i, before.x, newPoint.y);
            index++;
            i++;
          } else link.setPointAt(i - 1, before.x, newPoint.y);
          if (Geo.isApprox(after.x, link.getPoint(i + 2).x) && !Geo.isApprox(after.y, link.getPoint(i + 2).y)) {
            link.insertPointAt(i + 1, newPoint.x, after.y);
          } else link.setPointAt(i + 1, newPoint.x, after.y);
        } else if (Geo.isApprox(before.x, oldpt.x) && Geo.isApprox(oldpt.x, after.x)) {
          if (Geo.isApprox(before.x, link.getPoint(i - 2).x) && !Geo.isApprox(before.y, link.getPoint(i - 2).y)) {
            link.insertPointAt(i, newPoint.x, before.y);
            index++;
            i++;
          } else link.setPointAt(i - 1, newPoint.x, before.y);
          if (Geo.isApprox(after.x, link.getPoint(i + 2).x) && !Geo.isApprox(after.y, link.getPoint(i + 2).y)) {
            link.insertPointAt(i + 1, newPoint.x, after.y);
          } else link.setPointAt(i + 1, newPoint.x, after.y);
        } else if (Geo.isApprox(before.y, oldpt.y) && Geo.isApprox(oldpt.y, after.y)) {
          if (Geo.isApprox(before.y, link.getPoint(i - 2).y) && !Geo.isApprox(before.x, link.getPoint(i - 2).x)) {
            link.insertPointAt(i, before.x, newPoint.y);
            index++;
            i++;
          } else link.setPointAt(i - 1, before.x, newPoint.y);
          if (Geo.isApprox(after.y, link.getPoint(i + 2).y) && !Geo.isApprox(after.x, link.getPoint(i + 2).x)) {
            link.insertPointAt(i + 1, after.x, newPoint.y);
          } else link.setPointAt(i + 1, after.x, newPoint.y);
        }
        link.setPointAt(index, newPoint.x, newPoint.y);
      }
    } else {  // no Orthogonal constraints, just set the new point
      link.setPointAt(index, newPoint.x, newPoint.y);
      // also adjust the end point if there is no spot for the port
      let fromNode = link.fromNode;
      let fromPort = link.fromPort;
      if (fromNode !== null) {
        const visfromnode = fromNode.findVisibleNode();
        if (visfromnode !== null && visfromnode !== fromNode) {
          fromNode = visfromnode;
          fromPort = fromNode.port;  // use default port of Group
        }
      }
      if (index === 1 && link.computeSpot(true, fromPort).isNoSpot()) {
        const ctr = fromPort.getDocumentPoint(Spot.Center, Point.alloc());
        const endpt = link.getLinkPointFromPoint(fromNode, fromPort, ctr, newPoint, true, Point.alloc());
        link.setPointAt(0, endpt.x, endpt.y);
        Point.free(ctr);
        Point.free(endpt);
      }
      let toNode = link.toNode;
      let toPort = link.toPort;
      if (toNode !== null) {
        const vistonode = toNode.findVisibleNode();
        if (vistonode !== null && vistonode !== toNode) {
          toNode = vistonode;
          toPort = toNode.port;  // use default port of Group
        }
      }
      if (index === link.pointsCount - 2 && link.computeSpot(false, toPort).isNoSpot()) {
        const ctr = toPort.getDocumentPoint(Spot.Center, Point.alloc());
        const endpt = link.getLinkPointFromPoint(toNode, toPort, ctr, newPoint, false, Point.alloc());
        link.setPointAt(link.pointsCount - 1, endpt.x, endpt.y);
        Point.free(ctr);
        Point.free(endpt);
      }
    }
    link.commitRoute();
  }

  /**
   * This is called by {@link #doMouseMove} and {@link #doMouseUp} to limit the input point
   * before calling {@link #reshape}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LinkReshapingTool}
   * @param {Point} p the point where the handle is being dragged.
   * @return {Point} Either the same Point p or one constrained by the reshape behavior to be moved
   * only vertically or only horizontally.
   */
  public computeReshape(p: Point): Point {
    const link = this.adornedLink;
    const index = this.handle.segmentIndex;
    switch (this.getReshapingBehavior(this.handle)) {
      case LinkReshapingTool.All:
        return p;  // no constraints
      case LinkReshapingTool.Vertical: {
        const oldpt = link.getPoint(index);
        return new Point(oldpt.x, p.y);
      }
      case LinkReshapingTool.Horizontal: {
        const oldpt = link.getPoint(index);
        return new Point(p.x, oldpt.y);
      }
      default:
      case LinkReshapingTool.None:
        return link.getPoint(index);  // can't move -- return old point
    }
  }

  /**
   * This read-only property returns the {@link Point} that was the original location of the handle that is being dragged to reshape the Link.
   * @since 1.1
   */
  /** @export */ get originalPoint(): Point { return this._originalPoint; }

  /**
   * This read-only property returns the {@link List} of {@link Point}s that was the original route of the Link that is being reshaped.
   * This List should not be modified; its value is indeterminate until a reshaping has been activated.
   * @since 1.1
   */
  /** @export */ get originalPoints(): List<Point> { return this._originalPoints; }

} // end class
LinkReshapingTool['className'] = 'LinkReshapingTool';

LinkReshapingTool['None'] = LinkReshapingTool.None;
LinkReshapingTool['Horizontal'] = LinkReshapingTool.Horizontal;
LinkReshapingTool['Vertical'] = LinkReshapingTool.Vertical;
LinkReshapingTool['All'] = LinkReshapingTool.All;


// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'linkReshapingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): LinkReshapingTool {
    return this.findTool('LinkReshaping') as LinkReshapingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: LinkReshapingTool) { this.replaceStandardTool('LinkReshaping', value, this.mouseDownTools); }
);

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
ResizingTool
*/

// import { List } from './collections';
// import { Geo } from './geo';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { Adornment, Group, Part, Placeholder } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Tool } from './tool';
// import { Util } from './util';

/**
 * The ResizingTool is used to interactively change the size of a {@link GraphObject}
 * in the selected {@link Part} or {@link Node} by setting its {@link GraphObject#desiredSize} property.
 * You may want to save the size to the model by using a TwoWay {@link Binding} on the "desiredSize" property
 * of the GraphObject that is named by {@link Part#resizeObjectName}.
 * This tool does not operate on {@link Link}s.
 *
 * You can limit the permitted minimum and maximum dimensions by setting
 * {@link #minSize} and {@link #maxSize}.
 * The resizing will also respect the {@link GraphObject#minSize} and
 * {@link GraphObject#maxSize} properties.
 * Width or height values that are NaN do not constrain the resizing.
 * Override {@link #computeMinSize} and/or {@link #computeMaxSize} to change this behavior.
 *
 * You can also limit the width and/or height to be multiples of a particular size by
 * setting {@link Part#resizeCellSize}.
 * If either or both of these values are NaN, as they are by default,
 * it will get the values from this tool's {@link #cellSize}.
 * Finally it will consider the {@link Diagram#grid}'s {@link Panel#gridCellSize}
 * if {@link #isGridSnapEnabled} is true.
 * Override {@link #computeCellSize} to change this behavior.
 *
 * Pressing the Shift key or resizing a {@link Shape} with a {@link Shape#geometryStretch} of
 * {@link GraphObject.Uniform} or {@link GraphObject.UniformToFill} will maintain the aspect ratio during the resize.
 * Override {@link #computeReshape} to change this behavior.
 *
 * This tool makes use of an {@link Adornment}, shown when the {@link Part} or {@link Node} is selected,
 * that includes some number of resize handles.
 * The resize handles are normally copies of {@link ResizingTool#handleArchetype},
 * unless you specify a custom resize Adornment by setting {@link Part#resizeAdornmentTemplate}.
 * The resize Adornment is normally a "Spot" Panel with eight resize handles,
 * each with {@link GraphObject#alignment} set to one of the eight standard {@link Spot} values --
 * the four corners and the four side middles.
 * The {@link GraphObject#alignment} is what identifies and distinguishes each of the handles and
 * the behavior when the user drags the handle.
 *
 * This tool conducts a transaction while the tool is active.
 * A successful resizing will result in a "PartResized" {@link DiagramEvent} and a "Resizing" transaction.
 * <p class="boxread">
 * For a general discussion of the sizing of objects, see: <a href="../../intro/sizing.html">Introduction to the sizing of GraphObjects</a>.
 * For customizing the ResizingTool, see <a href="../../intro/tools.html#ResizingTool">Introduction to the ResizingTool</a>.
 *
 * If you want to programmatically start a user's resizing of the {@link Part#resizeObject} of an existing selected node,
 * you can set the {@link #handle} property to the specific resize handle and then start and activate the tool.
 * ```js
 *   var node = ...;
 *   myDiagram.select(node);
 *   var adorn = node.findAdornment("Resizing");
 *   var tool = myDiagram.toolManager.resizingTool;
 *   // specify which resize handle of the "Resizing" Adornment of the selected node
 *   tool.handle = adorn.elt(...);
 *   myDiagram.currentTool = tool;  // starts the ResizingTool
 *   tool.doActivate();             // activates the ResizingTool
 * ```
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ResizingTool extends Tool {
  /** @hidden @internal */ public static className = 'ResizingTool';
  /** @hidden @internal */ private _minSize: Size;
  /** @hidden @internal */ private _maxSize: Size;
  /** @hidden @internal */ private _cellSize: Size;
  /** @hidden @internal */ private _isGridSnapEnabled: boolean;
  /** @hidden @internal */ private _adornedObject: GraphObject | null;
  /** @hidden @internal */ private _handleArchetype: GraphObject | null;
  /** @hidden @internal */ private _handle: GraphObject | null;
  /** @hidden @internal */ private _originalPoint: Point;
  /** @hidden @internal */ private _originalDesiredSize: Size;
  /** @hidden @internal */ private _originalLocation: Point;
  /** @hidden @internal */ private _effectiveMinSize: Size;
  /** @hidden @internal */ private _effectiveMaxSize: Size;
  /** @hidden @internal */ private _effectiveCellSize: Size;
  /** @hidden @internal */ private _oldAnimation: boolean;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#resizingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "Resizing".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ResizingTool);
    super();

    this.name = 'Resizing';

    // external properties
    this._minSize = new Size(1, 1).freeze();
    this._maxSize = new Size(9999, 9999).freeze();
    this._cellSize = new Size(NaN, NaN).freeze();
    this._isGridSnapEnabled = false;

    // internal state
    this._adornedObject = null;

    const h = new Shape();
    h.alignmentFocus = Spot.Center;
    h.figure = 'Rectangle';
    h.desiredSize = Size.SixSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    h.strokeWidth = 1;
    h.cursor = 'pointer';
    this._handleArchetype = h;

    this._handle = null;
    this._originalPoint = new Point();
    this._originalDesiredSize = new Size();
    this._originalLocation = new Point();
    this._effectiveMinSize = new Size(0, 0);
    this._effectiveMaxSize = new Size(Infinity, Infinity);
    this._effectiveCellSize = new Size(1, 1);

    this._oldAnimation = true;
  }

  /**
   * Show an {@link Adornment} with the resize handles at points along the edge of the bounds of the
   * selected {@link Part}'s {@link Part#resizeObject}.
   *
   * First this finds the object in the visual tree of the Part that should
   * get the resize adornment and that the user will be able to resize interactively.
   * It finds the object that has the {@link Part#resizeObjectName} property of the Part.
   * If the {@link Part#resizeObjectName} property is an empty string, as it is by default,
   * it uses the whole part.
   *
   * It then builds the adornment, associating it with the chosen resize object.
   * If {@link Part#resizeAdornmentTemplate} is non-null, it is copied.
   * Otherwise it constructs a new Adornment with a {@link Placeholder} and eight copies of {@link #handleArchetype},
   * four at the corners and four at the middle of each side.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @param {Part} part
   */
  public updateAdornments(part: Part): void {
    if (part === null || part instanceof Link) return;  // can't resize links
    if (part.isSelected && !this.diagram.isReadOnly) {
      const resizeObj = part.resizeObject;
      let adornment = part.findAdornment(this.name);
      if (resizeObj !== null && part.canResize() && part.actualBounds.isReal() && part.isVisible() && resizeObj.actualBounds.isReal() && resizeObj.isVisibleObject()) {
        if (adornment === null || adornment.adornedObject !== resizeObj) {
          adornment = this.makeAdornment(resizeObj);
        }
        if (adornment !== null) {
          const angle = resizeObj.getDocumentAngle();
          // ?? Ideal if this doesn't happen often (like all the time while dragging)
          // but needs to happen often enough to fix resize handles.
          if (part.getInView()) this.updateResizeHandles(adornment, angle);  // fix up any cursors to point the right way
          part.addAdornment(this.name, adornment);
          return;
        }
      }
    }
    part.removeAdornment(this.name);
  }

  //  * Create and return an {@link Adornment} that lets the user grab a handle to resize the selected part.
  //  * The Adornment will be a "Spot" Panel with a {@link Placeholder} and eight copies of {@link #handleArchetype}.
  //  * Each handle will have one of the standard Spots as their {@link GraphObject#alignment}.
  //  *
  //  * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {ResizingTool}
   * @param {GraphObject} resizeObj the {@link GraphObject} being resized.
   * @return {Adornment}
   */
  public makeAdornment(resizeObj: GraphObject): Adornment {
    let adornment: Adornment = null;
    const template = resizeObj.part.resizeAdornmentTemplate;
    if (template === null) {
      adornment = new Adornment();
      adornment.type = Panel.Spot;
      adornment.locationSpot = Spot.Center;

      const surrogate = new Placeholder();
      surrogate.isPanelMain = true;
      adornment.add(surrogate);

      // add the standard eight handles: four at the corners
      adornment.add(this.makeHandle(resizeObj, Spot.TopLeft));
      adornment.add(this.makeHandle(resizeObj, Spot.TopRight));
      adornment.add(this.makeHandle(resizeObj, Spot.BottomRight));
      adornment.add(this.makeHandle(resizeObj, Spot.BottomLeft));
      // and four at the middle of each side
      adornment.add(this.makeHandle(resizeObj, Spot.MiddleTop));
      adornment.add(this.makeHandle(resizeObj, Spot.MiddleRight));
      adornment.add(this.makeHandle(resizeObj, Spot.MiddleBottom));
      adornment.add(this.makeHandle(resizeObj, Spot.MiddleLeft));
    } else {
      template.freezeBindings();
      adornment = template.copy();
      if (adornment === null) return null;
    }
    adornment.adornedObject = resizeObj;
    return adornment;
  }

  /**
   * @hidden @internal
   * Create and return a {@link GraphObject} that the user can "grab" to resize an object in the selected part.
   * By default this returns a {@link Shape} that is a small blue rectangle.
   *
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @param {GraphObject} resizeObj the {@link GraphObject} to be resized, which may be the same as the whole Part.
   * @param {Spot} spot
   * @return {GraphObject}
   */
  public makeHandle(resizeObj: GraphObject, spot: Spot): GraphObject {
    const h = this.handleArchetype;
    if (h === null) return null;
    const c = h.copy();
    c.alignment = spot;
    return c;
  }

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {ResizingTool}
   * @param {GraphObject} elt
   * @param {number} angle
   */
  public updateResizeHandles(elt: GraphObject, angle: number) {
    if (elt === null) return;
    if (!elt.alignment.isDefault() && (elt.cursor === 'pointer' || elt.cursor.indexOf('resize') > 0)) {
      this.setResizeCursor(elt, angle);
    } else if (elt instanceof Panel) {
      const it = elt.elements;
      while (it.next()) {
        const e = it.value;
        this.updateResizeHandles(e, angle);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {ResizingTool}
   * @param {GraphObject} h
   * @param {number} angle
   */
  private setResizeCursor(h: GraphObject, angle: number) {
    let spot = h.alignment;
    if (spot.isNoSpot()) spot = Spot.Center;
    let a = angle;
    if (spot.x <= 0) {  // left
      if (spot.y <= 0) {  // top-left
        a += 225;
      } else if (spot.y >= 1) {  // bottom-left
        a += 135;
      } else {  // middle-left
        a += 180;
      }
    } else if (spot.x >= 1) {  // right
      if (spot.y <= 0) {  // top-right
        a += 315;
      } else if (spot.y >= 1) {  // bottom-right
        a += 45;
      } else {  // middle-right
        // a += 0;
      }
    } else {  // middle-X
      if (spot.y <= 0) {  // top-middle
        a += 270;
      } else if (spot.y >= 1) {  // bottom-middle
        a += 90;
      } else {
        // handle is in the middle-middle -- don't do anything
        return;
      }
    }
    if (a < 0) a += 360;
    else if (a >= 360) a -= 360;
    if (a < 22.5) {
      h.cursor = 'e-resize';
    } else if (a < 67.5) {
      h.cursor = 'se-resize';
    } else if (a < 112.5) {
      h.cursor = 's-resize';
    } else if (a < 157.5) {
      h.cursor = 'sw-resize';
    } else if (a < 202.5) {
      h.cursor = 'w-resize';
    } else if (a < 247.5) {
      h.cursor = 'nw-resize';
    } else if (a < 292.5) {
      h.cursor = 'n-resize';
    } else if (a < 337.5) {
      h.cursor = 'ne-resize';
    } else {
      h.cursor = 'e-resize';
    }
  }

  /**
   * Gets or sets a small GraphObject that is copied as a resizing handle for the selected part.
   * By default this is a {@link Shape} that is a small blue rectangle.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handle to be larger yellow circles:
   * ```js
   *   myDiagram.toolManager.resizingTool.handleArchetype =
   *     $(go.Shape, "Circle",
   *       { width: 10, height: 10, fill: "yellow" });
   * ```
   *
   * This property is ignored when a custom resizing Adornment is specified as the {@link Part#resizeAdornmentTemplate}.
   * That property is normally null, in which case this tool will automatically construct Adornments
   * holding eight copies of this handle archetype, each with a {@link GraphObject#alignment} being one of the
   * standard eight {@link Spot}s.
   */
  /** @export */ get handleArchetype(): GraphObject | null { return this._handleArchetype; }
  /** @export */ set handleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, ResizingTool, 'handleArchetype');
    this._handleArchetype = value;
  }

  /**
   * Returns the {@link GraphObject} that is the tool handle being dragged by the user.
   * This will be contained by an {@link Adornment} whose category is "ResizingTool".
   * Its {@link Adornment#adornedObject} is the same as the {@link #adornedObject}.
   * This is normally set by {@link #doActivate}, remembering the result of the call to {@link Tool#findToolHandleAt}.
   *
   * This property is also settable, but should only be set either within an override of {@link #doActivate}
   * or prior to calling {@link #doActivate}.
   */
  /** @export */ get handle(): GraphObject | null { return this._handle; }
  /** @export */ set handle(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, ResizingTool, 'handle');
      if (!(value.part instanceof Adornment)) throw new Error('new handle is not in an Adornment: ' + value);
    }
    this._handle = value;
  }

  /**
   * Gets the {@link GraphObject} that is being resized.
   * This may be the same object as the selected {@link Part} or it may be contained within that Part.
   *
   * This property is also settable, but should only be set when overriding functions
   * in ResizingTool, and not during normal operation.
   */
  /** @export */ get adornedObject(): GraphObject | null { return this._adornedObject; }
  /** @export */ set adornedObject(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, ResizingTool, 'handle');
      if (value.part instanceof Adornment) throw new Error('new handle must not be in an Adornment: ' + value);
    }
    this._adornedObject = value;
  }

  /**
   * This tool may run when there is a mouse-down event on a resize handle,
   * the diagram is not read-only and it allows resizing,
   * the left mouse button is being used,
   * and this tool's adornment's resize handle is at the current mouse point.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (!diagram.allowResize) return false;
    if (!diagram.lastInput.left) return false;

    const h = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
    return (h !== null);
  }

  /**
   * Find the {@link #handle}, remember the object's original bounds,
   * save the results of calling {@link #computeMinSize}, {@link #computeMaxSize}, and {@link #computeCellSize},
   * capture the mouse, and start a transaction.
   *
   * Normally when this method is called the value of {@link #handle} will be null,
   * resulting in a call to {@link Tool#findToolHandleAt} to find a "Resizing" tool handle,
   * which is then remembered as the value of {@link #handle}.
   * If when this method is called the value of {@link #handle} is already set,
   * then there is no need to call {@link Tool#findToolHandleAt},
   * because the programmer has already set up which resize handle they want the user to be resizing.
   * @this {ResizingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this.handle === null) {
      this.handle = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
    }
    if (this.handle === null) return;
    this.adornedObject = (this.handle.part as Adornment).adornedObject;
    if (this.adornedObject === null) return;

    this._originalPoint.set(this.adornedObject.getDocumentPoint(this.handle.alignment.opposite()));
    // remember these properties so they can be restored if the tool is cancelled and the UndoManager is not enabled
    this._originalLocation.set(this.adornedObject.part.location);
    this._originalDesiredSize.set(this.adornedObject.desiredSize);
    // these are just called once
    this._effectiveCellSize = this.computeCellSize();
    this._effectiveMinSize = this.computeMinSize();
    this._effectiveMaxSize = this.computeMaxSize();
    diagram.isMouseCaptured = true;
    this._oldAnimation = diagram.animationManager.isEnabled;
    diagram.animationManager.isEnabled = false;
    this.startTransaction(this.name);
    this.isActive = true;
  }

  /**
   * Stop the current transaction, forget the {@link #handle} and {@link #adornedObject}, and release the mouse.
   * @this {ResizingTool}
   */
  public doDeactivate(): void {
    const diagram = this.diagram;

    this.stopTransaction();
    this.handle = null;
    this._adornedObject = null;
    diagram.isMouseCaptured = false;
    this.isActive = false;
    diagram.animationManager.isEnabled = this._oldAnimation;
  }

  /**
   * Restore the original size of the {@link GraphObject}.
   * @this {ResizingTool}
   */
  public doCancel(): void {
    // explicitly restore these values in case !UndoManager.isEnabled
    if (this.adornedObject !== null) {
      this.adornedObject.desiredSize = this.originalDesiredSize;
      this.adornedObject.part.location = this.originalLocation;
    }
    this.stopTool();
  }

  /**
   * Call {@link #resize} with a new size determined by the current mouse point.
   * This determines the new bounds by calling {@link #computeResize}.
   *
   * When this calls {@link #computeResize} it passes as the reshape argument the result of calling {@link #computeReshape}.
   * The min, max, and cell arguments will be the saved results of {@link #doActivate}'s calls to {@link #computeMinSize}, {@link #computeMaxSize}, and {@link #computeCellSize}.
   * @this {ResizingTool}
   */
  public doMouseMove(): void {
    const diagram = this.diagram;
    if (!this.isActive) return;
    const minsz = this._effectiveMinSize;
    const maxsz = this._effectiveMaxSize;
    const cellsz = this._effectiveCellSize;
    const docpt = diagram.lastInput.documentPoint;
    const locpt = this.adornedObject.getLocalPoint(docpt, Point.alloc());
    const reshape = this.computeReshape();
    const newr = this.computeResize(locpt, this.handle.alignment, minsz, maxsz, cellsz, reshape);
    this.resize(newr);
    // do it immediately
    diagram.maybeUpdate();
    Point.free(locpt);
  }

  /**
   * Call {@link #resize} with the final bounds based on the most recent mouse point,
   * commit the transaction, and raise the "PartResized" {@link DiagramEvent}.
   * This determines the new bounds by calling {@link #computeResize}.
   *
   * When this calls {@link #computeResize} it passes as the reshape argument the result of calling {@link #computeReshape}.
   * @this {ResizingTool}
   */
  public doMouseUp(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      const minsz = this._effectiveMinSize;
      const maxsz = this._effectiveMaxSize;
      const cellsz = this._effectiveCellSize;
      const locpt = this.adornedObject.getLocalPoint(diagram.lastInput.documentPoint, Point.alloc());
      const reshape = this.computeReshape();
      const newr = this.computeResize(locpt, this.handle.alignment, minsz, maxsz, cellsz, reshape);
      this.resize(newr);
      Point.free(locpt);

      diagram.invalidateDocumentBounds();
      // set the TransactionResult before raising event, in case it changes the result or cancels the tool
      this.transactionResult = this.name;
      diagram.raiseDiagramEvent('PartResized', this.adornedObject, this.originalDesiredSize);
    }
    this.stopTool();
  }

  /**
   * Change the size of the selected part's {@link Part#resizeObject} to have the given bounds.
   * This modifies its {@link GraphObject#desiredSize} and maybe its {@link Part#location}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @param {Rect} newr a Rectangle in the {@link #adornedObject}'s local coordinates, not in document coordinates
   */
  public resize(newr: Rect): void {
    const diagram = this.diagram;
    const obj = this.adornedObject;
    const part = obj.part;

    obj.desiredSize = newr.size;

    part.ensureBounds();
    const pt = this.adornedObject.getDocumentPoint(this.handle.alignment.opposite());
    if (part instanceof Group) {
      const l = new List<Part>();
      l.add(part);
      diagram.moveParts(l, this._originalPoint.copy().subtract(pt), true);
    } else {
      part.location = part.location.copy().subtract(pt).add(this._originalPoint);
    }
  }

  /**
   * Given a {@link Spot} in the original bounds of the object being resized and a new {@link Point},
   * compute the new {@link Rect}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @param {Point} newPoint a Point in local coordinates.
   * @param {Spot} spot the alignment spot of the handle being dragged.
   * @param {Size} min the result of the call to {@link #computeMinSize}.
   * @param {Size} max the result of the call to {@link #computeMaxSize}.
   * @param {Size} cell the result of the call to {@link #computeCellSize}.
   * @param {boolean} reshape true if the new size may change the aspect ratio from that of the natural bounds of the {@link #adornedObject}.
   * @return {Rect} a Rectangle in the {@link #adornedObject}'s local coordinates, not in document coordinates
   */
  public computeResize(newPoint: Point, spot: Spot, min: Size, max: Size, cell: Size, reshape: boolean): Rect {
    if (spot.isNoSpot()) spot = Spot.Center;

    const origRect = this.adornedObject.naturalBounds;
    const left = origRect.x;
    const top = origRect.y;
    const right = origRect.x + origRect.width;
    const bottom = origRect.y + origRect.height;

    let oldratio = 1.0;
    if (!reshape) {
      let oldw = origRect.width;
      let oldh = origRect.height;
      if (oldw <= 0) oldw = 1;
      if (oldh <= 0) oldh = 1;
      oldratio = oldh / oldw;
    }

    const gridpt = Point.alloc();
    Geo.findNearestInfiniteGridPoint(newPoint.x, newPoint.y, left, top, cell.width, cell.height, gridpt);

    const newRect = origRect.copy();
    if (spot.x <= 0) {  // left
      if (spot.y <= 0) {  // top-left
        newRect.x = Math.max(gridpt.x, right - max.width);
        newRect.x = Math.min(newRect.x, right - min.width);
        newRect.width = Math.max(right - newRect.x, min.width);

        newRect.y = Math.max(gridpt.y, bottom - max.height);
        newRect.y = Math.min(newRect.y, bottom - min.height);
        newRect.height = Math.max(bottom - newRect.y, min.height);

        if (!reshape) {
          const newratio = newRect.height / newRect.width;
          if (newratio >= 1) {  // h > w, constrain by width
            newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
            newRect.width = newRect.height / oldratio;
          } else {
            newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
            newRect.height = oldratio * newRect.width;
          }
          newRect.x = right - newRect.width;
          newRect.y = bottom - newRect.height;
        }
      } else if (spot.y >= 1) {  // bottom-left
        newRect.x = Math.max(gridpt.x, right - max.width);
        newRect.x = Math.min(newRect.x, right - min.width);
        newRect.width = Math.max(right - newRect.x, min.width);

        newRect.height = Math.max(Math.min(gridpt.y - top, max.height), min.height);

        if (!reshape) {
          const newratio = newRect.height / newRect.width;
          if (newratio >= 1) {  // h > w, constrain by width
            newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
            newRect.width = newRect.height / oldratio;
          } else {
            newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
            newRect.height = oldratio * newRect.width;
          }
          newRect.x = right - newRect.width;
        }
      } else {  // middle-left
        newRect.x = Math.max(gridpt.x, right - max.width);
        newRect.x = Math.min(newRect.x, right - min.width);
        newRect.width = right - newRect.x;

        if (!reshape) {
          newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
          newRect.width = newRect.height / oldratio;
          newRect.y = top + 0.5 * ((bottom - top) - newRect.height);
        }
      }
    } else if (spot.x >= 1) {  // right
      if (spot.y <= 0) {  // top-right
        newRect.width = Math.max(Math.min(gridpt.x - left, max.width), min.width);

        newRect.y = Math.max(gridpt.y, bottom - max.height);
        newRect.y = Math.min(newRect.y, bottom - min.height);
        newRect.height = Math.max(bottom - newRect.y, min.height);

        if (!reshape) {
          const newratio = newRect.height / newRect.width;
          if (newratio >= 1) {  // h > w, constrain by width
            newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
            newRect.width = newRect.height / oldratio;
          } else {
            newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
            newRect.height = oldratio * newRect.width;
          }
          newRect.y = bottom - newRect.height;
        }
      } else if (spot.y >= 1) {  // bottom-right
        newRect.width = Math.max(Math.min(gridpt.x - left, max.width), min.width);
        newRect.height = Math.max(Math.min(gridpt.y - top, max.height), min.height);

        if (!reshape) {
          const newratio = newRect.height / newRect.width;
          if (newratio >= 1) { // h > w, constrain by width
            newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
            newRect.width = newRect.height / oldratio;
          } else {
            newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
            newRect.height = oldratio * newRect.width;
          }
        }
      } else {  // middle-right
        newRect.width = Math.max(Math.min(gridpt.x - left, max.width), min.width);

        if (!reshape) {
          newRect.height = Math.max(Math.min(oldratio * newRect.width, max.height), min.height);
          newRect.width = newRect.height / oldratio;
          newRect.y = top + 0.5 * ((bottom - top) - newRect.height);
        }
      }
    } else {  // middle-X
      if (spot.y <= 0) {  // top-middle
        newRect.y = Math.max(gridpt.y, bottom - max.height);
        newRect.y = Math.min(newRect.y, bottom - min.height);
        newRect.height = bottom - newRect.y;

        if (!reshape) {
          newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
          newRect.height = oldratio * newRect.width;
          newRect.x = left + 0.5 * ((right - left) - newRect.width);
        }
      } else if (spot.y >= 1) {  // bottom-middle
        newRect.height = Math.max(Math.min(gridpt.y - top, max.height), min.height);

        if (!reshape) {
          newRect.width = Math.max(Math.min(newRect.height / oldratio, max.width), min.width);
          newRect.height = oldratio * newRect.width;
          newRect.x = left + 0.5 * ((right - left) - newRect.width);
        }
      } else {
        // handle is in the middle-middle -- don't do anything
      }
    }

    Point.free(gridpt);
    return newRect;
  }

  /**
   * Decide whether to allow arbitrary reshaping or whether to keep the same aspect ratio of the object being resized.
   * If the {@link #adornedObject} is a {@link Shape},
   * then if the {@link Shape#geometryStretch} is {@link GraphObject.Uniform},
   * this method will return false to restrict reshaping to maintain the object's current ratio of height to width.
   * Also, if the user is holding down the Shift key, this method will return false.
   *
   * This is called on each mouse-move and on mouse-up; the result is passed to the call to {@link #resize}.
   * This permits the user to change the behavior dynamically during resizing.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * For example, to always keep the object's original aspect ratio, override this method to return false.
   * When calling {@link GraphObject.make} to initialize a {@link Diagram}:
   * ```js
   * $(go.Diagram, . . .,
   *   {
   *     "resizingTool.computeReshape": function() { return false; },
   *     . . . })
   * ```
   * Or when overriding the method dynamically:
   * ```js
   *   myDiagram.toolManager.resizingTool.computeReshape = function() { return false; }
   * ```
   * Your override might want to look at the `this.adornedObject.part.data` properties to decide whether to allow reshaping.
   * @expose
   * @this {ResizingTool}
   * @return {boolean} true to allow any aspect ratio; false to preserve the {@link #adornedObject}'s height/width ratio
   * @since 1.7
   */
  public computeReshape(): boolean {
    let stretch = GraphObject.None;
    if (this.adornedObject instanceof Shape) {
      const shape = this.adornedObject;
      stretch = shape.getGeometryStretch();
    }
    return !(stretch === GraphObject.Uniform || this.diagram.lastInput.shift);
  }

  /**
   * The effective minimum resizing size is the maximum of {@link #minSize} and the
   * {@link #adornedObject}'s {@link GraphObject#minSize}.
   *
   * This is called once when the tool is activated.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @return {Size}
   */
  public computeMinSize(): Size {
    const msz = this.adornedObject.minSize.copy();
    const sz = this.minSize;
    if (!isNaN(sz.width) && sz.width > msz.width) msz.width = sz.width;
    if (!isNaN(sz.height) && sz.height > msz.height) msz.height = sz.height;
    return msz;
  }

  /**
   * The effective maximum resizing size is the minimum of the {@link #maxSize} and the
   * {@link #adornedObject}'s {@link GraphObject#maxSize}.
   *
   * This is called once when the tool is activated.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @return {Size}
   */
  public computeMaxSize(): Size {
    const msz = this.adornedObject.maxSize.copy();
    const sz = this.maxSize;
    if (!isNaN(sz.width) && sz.width < msz.width) msz.width = sz.width;
    if (!isNaN(sz.height) && sz.height < msz.height) msz.height = sz.height;
    return msz;
  }

  /**
   * The size should be a multiple of the value returned by this method.
   *
   * This is called once when the tool is activated.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ResizingTool}
   * @return {Size}
   */
  public computeCellSize(): Size {
    const csz = new Size(NaN, NaN);
    const part = this.adornedObject.part;
    if (part !== null) {  // any information on the Part takes precedence
      const psz = part.resizeCellSize;
      if (!isNaN(psz.width) && psz.width > 0) csz.width = psz.width;
      if (!isNaN(psz.height) && psz.height > 0) csz.height = psz.height;
    }

    // skip looking for snappers

    let sz = this.cellSize; // look at this ResizingTool's cell property
    if (isNaN(csz.width) && !isNaN(sz.width) && sz.width > 0) csz.width = sz.width;
    if (isNaN(csz.height) && !isNaN(sz.height) && sz.height > 0) csz.height = sz.height;

    // maybe need to look at the diagram's background grid
    const diagram = this.diagram;
    if ((isNaN(csz.width) || isNaN(csz.height)) && diagram) {
      const backgrid = diagram.grid;
      if (backgrid !== null && backgrid.visible && this.isGridSnapEnabled) {
        sz = backgrid.gridCellSize;
        if (isNaN(csz.width) && !isNaN(sz.width) && sz.width > 0) csz.width = sz.width;
        if (isNaN(csz.height) && !isNaN(sz.height) && sz.height > 0) csz.height = sz.height;
      }
    }

    // finally, disallow NaN or zero or Infinity: default to 1
    if (isNaN(csz.width) || csz.width === 0 || csz.width === Infinity) csz.width = 1;
    if (isNaN(csz.height) || csz.height === 0 || csz.height === Infinity) csz.height = 1;
    return csz;
  }

  /**
   * Gets or sets the minimum size to which the user can resize.
   * The effective minimum size is the maximum of this value and the {@link GraphObject#minSize},
   * independently in each direction.
   *
   * The default value is `go.Size(1, 1)`.
   * Any new value must be of type Size; `NaN` width or height values are treated as zero.
   * Setting this property does not raise any events.
   */
  /** @export */ get minSize(): Size { return this._minSize; }
  /** @export */ set minSize(value: Size) {
    Util.checkClass(value, Size, ResizingTool, 'minSize');
    const old = this._minSize;
    if (!old.equals(value)) {
      let w = value.width;
      if (isNaN(w)) w = 0;
      let h = value.height;
      if (isNaN(h)) h = 0;
      this._minSize.assignTo(w, h);
    }
  }

  /**
   * Gets or sets the maximum size to which the user can resize.
   * The effective maximum size is the minimum of this value and the {@link GraphObject#maxSize},
   * independently in each direction.
   *
   * The default value is `go.Size(9999, 9999)`.
   * Any new value must be of type Size; `NaN` width or height values are treated as Infinity.
   * Setting this property does not raise any events.
   */
  /** @export */ get maxSize(): Size { return this._maxSize; }
  /** @export */ set maxSize(value: Size) {
    Util.checkClass(value, Size, ResizingTool, 'maxSize');
    const old = this._maxSize;
    if (!old.equals(value)) {
      let w = value.width;
      if (isNaN(w)) w = Infinity;
      let h = value.height;
      if (isNaN(h)) h = Infinity;
      this._maxSize.assignTo(w, h);
    }
  }

  /**
   * Gets or sets the width and height multiples with which the user must resize.
   * The effective cell size is computed by first looking at the {@link Adornment#adornedPart}'s {@link Part#resizeCellSize}.
   * If either or both of its width and height are `NaN`, it will use this property, `cellSize`.
   * If either or both of this property's width and height are `NaN`, it will
   * consider the {@link Diagram#grid}'s {@link Panel#gridCellSize}.
   *
   * The default value is `go.Size(NaN, NaN)`.
   * Setting this property does not raise any events.
   */
  /** @export */ get cellSize(): Size { return this._cellSize; }
  /** @export */ set cellSize(value: Size) {
    Util.checkClass(value, Size, ResizingTool, 'cellSize');
    const old = this._cellSize;
    if (!old.equals(value)) {
      this._cellSize.assign(value);
    }
  }

  /**
   * Gets or sets whether the {@link ResizingTool} snaps object sizes to the diagram's background grid during the resize.
   * By default this property is false.
   * Setting this property does not raise any events.
   */
  /** @export */ get isGridSnapEnabled(): boolean { return this._isGridSnapEnabled; }
  /** @export */ set isGridSnapEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', ResizingTool, 'isGridSnapEnabled');
    this._isGridSnapEnabled = value;
  }

  /**
   * This read-only property returns the {@link Size} that was the original value of the {@link GraphObject#desiredSize}
   * of the element that is being resized.
   * @since 1.1
   */
  /** @export */ get originalDesiredSize(): Size { return this._originalDesiredSize; }

  /**
   * This read-only property returns the {@link Point} that was the original value of the {@link Part#location}
   * of the Part that is being resized.
   * @since 1.1
   */
  /** @export */ get originalLocation(): Point { return this._originalLocation; }

} // end class
ResizingTool['className'] = 'ResizingTool';


// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'resizingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): ResizingTool {
    return this.findTool('Resizing') as ResizingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: ResizingTool) { this.replaceStandardTool('Resizing', value, this.mouseDownTools); }
);

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
RotatingTool
*/

// import { Debug } from './debug';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { Adornment, Part } from './parts';
// import { Point } from './point';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Tool } from './tool';
// import { Util } from './util';

/**
 * The RotatingTool is used to interactively change the {@link GraphObject#angle} of a {@link GraphObject}
 * by setting its {@link GraphObject#angle} property.
 * You may want to save the angle to the model by using a TwoWay {@link Binding} on the "angle" property
 * of the GraphObject that is named by {@link Part#rotateObjectName}.
 *
 * This tool allows the user to rotate the {@link Part#rotateObject} of the selected {@link Part}.
 * The {@link Part} must be {@link Part#rotatable}, which is false by default.
 * Normally this works with {@link Part}s or {@link Node}s; it does not make sense for whole {@link Link}s
 * or {@link Link#path}s, so if you want to rotate a label on a Link, make sure to name that label
 * and refer to it as the {@link Part#rotateObjectName}.
 *
 * You can control the point about which the object is rotated by setting {@link Part#rotationSpot}.
 * The rotation spot can be computed dynamically by overriding {@link #computeRotationPoint}.
 * The value of that method call is saved as {@link #rotationPoint}.
 *
 * You can limit the permitted angles by setting {@link #snapAngleMultiple} and {@link #snapAngleEpsilon}.
 * For example, if you want to permit only angles that are multiples of 90 degrees,
 * set {@link #snapAngleMultiple} to 90 and {@link #snapAngleEpsilon} to 45.
 * Pressing the Shift key during rotation ignores these two properties.
 *
 * This tool makes use of an {@link Adornment} that includes a rotation handle.
 * It is shown when the selected {@link Part} is {@link Part#rotatable}.
 * You can control its direction relative to the {@link #rotationPoint}
 * by setting {@link #handleAngle} to a multiple of 90 degrees,
 * and its distance from the rotation point by setting {@link #handleDistance}.
 * The rotate handle is normally a copy of {@link RotatingTool#handleArchetype}.
 * unless you specify a custom rotate Adornment by setting {@link Part#rotateAdornmentTemplate}.
 *
 * This tool conducts a transaction while the tool is active.
 * A successful rotation will result in a "PartRotated" {@link DiagramEvent} and a "Rotating" transaction.
 * <p class="boxread">
 * For customizing the RotatingTool, see <a href="../../intro/tools.html#RotatingTool">Introduction to the RotatingTool</a>.
 *
 * If you want to programmatically start a user's rotating of the {@link Part#rotateObject} of an existing selected node,
 * you can set the {@link #handle} property to the rotate handle and then start and activate the tool.
 * ```js
 *   var node = ...;
 *   myDiagram.select(node);
 *   var adorn = node.findAdornment("Rotating");
 *   var tool = myDiagram.toolManager.rotatingTool;
 *   // specify the rotation handle of the "Rotating" Adornment of the selected node
 *   tool.handle = adorn.elt(0);
 *   myDiagram.currentTool = tool;  // starts the RotatingTool
 *   tool.doActivate();             // activates the RotatingTool
 * ```
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class RotatingTool extends Tool {
  /** @hidden @internal */ public static className = 'RotatingTool';
  /** @hidden @internal */ private _snapAngleMultiple: number;
  /** @hidden @internal */ private _snapAngleEpsilon: number;
  /** @hidden @internal */ private _adornedObject: GraphObject | null;
  /** @hidden @internal */ private _handleArchetype: GraphObject | null;
  /** @hidden @internal */ private _handle: GraphObject | null;
  /** @hidden @internal */ private _originalAngle: number;
  /** @hidden @internal */ private _rotationPoint: Point;
  /** @hidden @internal */ private _originalLocation: Point;
  /** @hidden @internal */ private _handleAngle: number;
  /** @hidden @internal */ private _handleDistance: number;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#rotatingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "Rotating".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(RotatingTool);
    super();

    this.name = 'Rotating';

    // external properties
    this._snapAngleMultiple = 45;
    this._snapAngleEpsilon = 2;
    this._originalLocation = new Point();

    // internal state
    this._adornedObject = null;

    const h = new Shape();
    h.figure = 'Ellipse';
    h.desiredSize = Size.EightSize;
    h.fill = 'lightblue';
    h.stroke = 'dodgerblue';
    h.strokeWidth = 1;
    h.cursor = 'pointer';
    this._handleArchetype = h;

    this._handle = null;
    this._originalAngle = 0;  // in case the tool is cancelled and the UndoManager is not enabled
    this._rotationPoint = new Point(NaN, NaN);
    this._handleAngle = 0;
    this._handleDistance = 50;
  }

  /**
   * Show an {@link Adornment} with a rotate handle at a point to the side of the adorned object
   * if the part is selected and visible and if {@link Part#canRotate}() is true.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RotatingTool}
   * @param {Part} part the part.
   */
  public updateAdornments(part: Part): void {
    if (part === null) return;
    if (part.isLinkLike()) {  // this tool never applies to whole Links or Link main elements
      const rotateObj = part.rotateObject;
      if (rotateObj === part || rotateObj === part.path || rotateObj.isPanelMain) return;
    }
    if (part.isSelected && !this.diagram.isReadOnly) {
      const rotateObj = part.rotateObject;
      if (rotateObj !== null && part.canRotate() && part.actualBounds.isReal() && part.isVisible() && rotateObj.actualBounds.isReal() && rotateObj.isVisibleObject()) {
        let adornment = part.findAdornment(this.name);
        if (adornment === null || adornment.adornedObject !== rotateObj) {
          adornment = this.makeAdornment(rotateObj);
        }
        if (adornment !== null) {
          adornment.angle = rotateObj.getDocumentAngle();
          if (adornment.placeholder === null) {
            adornment.location = this.computeAdornmentLocation(rotateObj);
          }
          part.addAdornment(this.name, adornment);
          return;
        }
      }
    }
    part.removeAdornment(this.name);
  }

  //  * Create and return an {@link Adornment} that lets the user grab a handle to rotate an object in the selected part.
  //  * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {RotatingTool}
   * @param {GraphObject} rotateObj the {@link GraphObject} being rotated.
   * @return {Adornment}
   */
  public makeAdornment(rotateObj: GraphObject): Adornment {
    let adornment: Adornment = null;
    const template = rotateObj.part.rotateAdornmentTemplate;
    if (template === null) {
      adornment = new Adornment();
      adornment.type = Panel.Position;
      adornment.locationSpot = Spot.Center;
      const h = this.handleArchetype;
      if (h !== null) {
        // add a single handle for rotation
        adornment.add(h.copy());
      }
    } else {
      template.freezeBindings();
      adornment = template.copy();
      if (adornment === null) return null;
    }
    adornment.adornedObject = rotateObj;
    return adornment;
  }

  /**
   * Gets or sets a small GraphObject that is copied as a rotation handle for the selected part.
   * By default this is a {@link Shape} that is a small yellow circle.
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the default handle to be green "X":
   * ```js
   *   myDiagram.toolManager.rotatingTool.handleArchetype =
   *     $(go.Shape, "XLine",
   *       { width: 8, height: 8, stroke: "green", fill: "transparent" });
   * ```
   *
   * This property is ignored when a custom rotating Adornment is specified as the {@link Part#rotateAdornmentTemplate}.
   */
  /** @export */ get handleArchetype(): GraphObject | null { return this._handleArchetype; }
  /** @export */ set handleArchetype(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, RotatingTool, 'handleArchetype');
    this._handleArchetype = value;
  }

  /**
   * This read-only property returns the {@link GraphObject} that is the tool handle being dragged by the user.
   * This will be contained by an {@link Adornment} whose category is "RotatingTool".
   * Its {@link Adornment#adornedObject} is the same as the {@link #adornedObject}.
   *
   * This property is also settable, but should only be set either within an override of {@link #doActivate}
   * or prior to calling {@link #doActivate}.
   */
  /** @export */ get handle(): GraphObject | null { return this._handle; }
  /** @export */ set handle(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, RotatingTool, 'handle');
      if (!(value.part instanceof Adornment)) throw new Error('new handle is not in an Adornment: ' + value);
    }
    this._handle = value;
  }

  /**
   * Gets the {@link GraphObject} that is being rotated.
   * This may be the same object as the selected {@link Part} or it may be contained within that Part.
   *
   * This property is also settable, but should only be set when overriding functions
   * in RotatingTool, and not during normal operation.
   */
  /** @export */ get adornedObject(): GraphObject | null { return this._adornedObject; }
  /** @export */ set adornedObject(value: GraphObject | null) {
    if (value !== null) {
      Util.checkClass(value, GraphObject, RotatingTool, 'handle');
      if (value.part instanceof Adornment) throw new Error('new handle must not be in an Adornment: ' + value);
    }
    this._adornedObject = value;
  }

  /**
   * This tool may run when there is a mouse-down event on a rotate handle,
   * the diagram is not read-only and it allows rotation,
   * the left mouse button is being used,
   * and this tool's adornment's rotate handle is at the current mouse point.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RotatingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (!diagram.allowRotate) return false;
    if (!diagram.lastInput.left) return false;

    const h = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
    return (h !== null);
  }

  /**
   * Capture the mouse, remember the original {@link GraphObject#angle},
   * and start a transaction.
   *
   * Normally when this method is called the value of {@link #handle} will be null,
   * resulting in a call to {@link Tool#findToolHandleAt} to find a "Rotating" tool handle,
   * which is then remembered as the value of {@link #handle}.
   * If when this method is called the value of {@link #handle} is already set,
   * then there is no need to call {@link Tool#findToolHandleAt},
   * because the programmer has already set up which rotate handle they want the user to be rotating.
   * @this {RotatingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this.adornedObject === null) {
      if (this.handle === null) {
        this.handle = this.findToolHandleAt(diagram.firstInput.documentPoint, this.name);
      }
      if (this.handle === null) return;
      this.adornedObject = (this.handle.part as Adornment).adornedObject;
    }
    if (this.adornedObject === null) return;

    // remember the angle in case the UndoManager is not enabled and the tool is cancelled
    this._originalAngle = this.adornedObject.angle;
    this._rotationPoint = this.computeRotationPoint(this.adornedObject);
    this._originalLocation = this.adornedObject.part.location.copy();

    diagram.isMouseCaptured = true;
    diagram.delaysLayout = true;
    this.startTransaction(this.name);
    this.isActive = true;
  }

  /**
   * Compute the Point about which things should rotate.
   * This is called when this tool is activated and sometimes when updating the Adornment.
   * The value of the call is remembered as {@link #rotationPoint}.
   * @expose
   * @this {RotatingTool}
   * @param {GraphObject} obj  the {@link Part#rotateObject}
   * @return {Point}
   */
  public computeRotationPoint(obj: GraphObject): Point {
    const part = obj.part;
    const locationObject = part.locationObject;
    if (part.rotationSpot.isSpot()) {
      return obj.getDocumentPoint(part.rotationSpot);
    } if (obj === part || obj === locationObject) {
      return locationObject.getDocumentPoint(part.locationSpot);
    } else {
      return obj.getDocumentPoint(Spot.Center);
    }
  }

  /**
   * @hidden @internal
   * Compute the location for the Adornment, assuming the Adornment only holds a rotate handle.
   * This is only called when the Adornment does not have a Placeholder.
   *
   * Only supports {@link #handleAngle} as multiples of 90 degrees.
   * Behavior when a multiple of 45 degrees (but not 90) is non-optimal.
   * @expose
   * @this {RotatingTool}
   * @param {GraphObject} obj  the {@link Part#rotateObject}
   * @return {Point}
   */
  public computeAdornmentLocation(obj: GraphObject): Point {
    let p = this.rotationPoint;
    if (!p.isReal()) p = this.computeRotationPoint(obj);
    const q = obj.getLocalPoint(p);
    let a = this.handleAngle;
    if (a < 0) a += 360;
    else if (a >= 360) a -= 360;
    a = Math.round(Math.round(a / 45) * 45);
    const d = this.handleDistance;
    if (a === 0) {
      q.x = obj.naturalBounds.width + d;
    } else if (a === 45) {
      q.x = obj.naturalBounds.width + d;
      q.y = obj.naturalBounds.height + d;
    } else if (a === 90) {
      q.y = obj.naturalBounds.height + d;
    } else if (a === 135) {
      q.x = -d;
      q.y = obj.naturalBounds.height + d;
    } else if (a === 180) {
      q.x = -d;
    } else if (a === 225) {
      q.x = -d;
      q.y = -d;
    } else if (a === 270) {
      q.y = -d;
    } else if (a === 315) {
      q.x = obj.naturalBounds.width + d;
      q.y = -d;
    }
    return obj.getDocumentPoint(q);
  }

  /**
   * Stop the current transaction, forget the {@link #handle} and {@link #adornedObject}, and release the mouse.
   * @this {RotatingTool}
   */
  public doDeactivate(): void {
    const diagram = this.diagram;

    this.stopTransaction();
    this.handle = null;
    this._adornedObject = null;
    this._rotationPoint = new Point(NaN, NaN);
    diagram.isMouseCaptured = false;
    this.isActive = false;
  }

  /**
   * Restore the original {@link GraphObject#angle} of the adorned object.
   * @this {RotatingTool}
   */
  public doCancel(): void {
    const diagram = this.diagram;
    diagram.delaysLayout = false;
    // explicitly restore this property in case !UndoManager.isEnabled
    this.rotate(this.originalAngle);
    this.stopTool();
  }

  /**
   * Call {@link #rotate} with a new angle determined by the current mouse point.
   * This determines the new angle by calling {@link #computeRotate}.
   * @this {RotatingTool}
   */
  public doMouseMove(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      const newangle = this.computeRotate(diagram.lastInput.documentPoint);
      this.rotate(newangle);
    }
  }

  /**
   * Call {@link #rotate} with an angle based on the most recent mouse point,
   * commit the transaction, and raise the "PartRotated" {@link DiagramEvent}.
   * @this {RotatingTool}
   */
  public doMouseUp(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      diagram.delaysLayout = false;
      const newangle = this.computeRotate(diagram.lastInput.documentPoint);
      this.rotate(newangle);

      diagram.invalidateDocumentBounds();
      // set the TransactionResult before raising event, in case it changes the result or cancels the tool
      this.transactionResult = this.name;
      diagram.raiseDiagramEvent('PartRotated', this.adornedObject, this.originalAngle);
    }
    this.stopTool();
  }

  /**
   * Change the angle of the selected part's {@link Part#rotateObject}.
   * This modifies its {@link GraphObject#angle}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RotatingTool}
   * @param {number} newangle in degrees.
   */
  public rotate(newangle: number): void {
    if (Debug) Util.checkRealNumber(newangle, RotatingTool, 'rotate:newangle');
    const obj = this.adornedObject;
    if (obj !== null) {
      obj.angle = newangle;

      const part = obj.part;
      part.ensureBounds();
      const locobj = part.locationObject;
      const rotobj = part.rotateObject;

      if (locobj === rotobj || locobj.isContainedBy(rotobj)) {
        const pt = this._originalLocation.copy();
        part.location = pt.subtract(this.rotationPoint).rotate(newangle - this.originalAngle).add(this.rotationPoint);
      }
    }
  }

  /**
   * Compute the new angle given a point.
   * If the Shift key is pressed, this method ignores the {@link #snapAngleMultiple} and {@link #snapAngleEpsilon}.
   *
   * If the angle is close (by {@link #snapAngleEpsilon} degrees)
   * to a multiple of {@link #snapAngleMultiple} degrees,
   * make it exactly that multiple.
   *
   * This method is called by both {@link #doMouseMove} and {@link #doMouseUp}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {RotatingTool}
   * @param {Point} newPoint in document coordinates.
   * @return {number} the new angle, in degrees.
   */
  public computeRotate(newPoint: Point): number {
    // the apparent angle from RotationPoint to newPoint
    let a = this.rotationPoint.directionPoint(newPoint) - this.handleAngle;

    // account for the angle of the rotateObject's panel, if it exists
    const parent = this.adornedObject.panel;
    if (parent !== null) {
      a -= parent.getDocumentAngle();
    }
    if (a >= 360) a -= 360;
    else if (a < 0) a += 360;

    const interval = Math.min(Math.abs(this.snapAngleMultiple), 180);
    const epsilon = Math.min(Math.abs(this.snapAngleEpsilon), interval / 2);
    // if it's close to a multiple of INTERVAL degrees, make it exactly so
    if (!this.diagram.lastInput.shift && interval > 0 && epsilon > 0) {
      if (a % interval < epsilon) {
        a = Math.floor(a / interval) * interval;
      } else if (a % interval > interval - epsilon) {
        a = (Math.floor(a / interval) + 1) * interval;
      }
    }
    if (a >= 360) a -= 360;
    else if (a < 0) a += 360;
    return a;
  }

  /**
   * Gets or sets the preferred angles for the selected object.
   *
   * The default value is 45 degrees, meaning that angles that are multiples
   * of 45 degrees are automatically preferred, if the actual angle is
   * close to that multiple.
   * The closeness is determined by the {@link #snapAngleEpsilon} property.
   * A value of zero for {@link #snapAngleMultiple} results in no snapping at all.
   * Setting this property does not raise any events.
   */
  /** @export */ get snapAngleMultiple(): number { return this._snapAngleMultiple; }
  /** @export */ set snapAngleMultiple(value: number) {
    Util.checkPrimitive(value, 'number', RotatingTool, 'snapAngleMultiple');
    this._snapAngleMultiple = value;
  }

  /**
   * Gets or sets the the closeness to a desired angle at which the angle is "snapped to".
   *
   * The default value is 2 degrees, meaning that any angle within 2 degrees
   * of a multiple of the {@link #snapAngleMultiple} automatically
   * snaps to that multiple.
   * Values are limited to half of the {@link #snapAngleMultiple};
   * such values restrict user selected angles only to exact multiples of
   * {@link #snapAngleMultiple} -- no other angles between them.
   * Setting this property does not raise any events.
   */
  /** @export */ get snapAngleEpsilon(): number { return this._snapAngleEpsilon; }
  /** @export */ set snapAngleEpsilon(value: number) {
    Util.checkPrimitive(value, 'number', RotatingTool, 'snapAngleEpsilon');
    this._snapAngleEpsilon = value;
  }

  /**
   * This read-only property returns the angle that was the original value of the {@link GraphObject#angle}
   * of the GraphObject that is being rotated.
   * The value is invalid when this tool is not active.
   * @since 1.1
   */
  /** @export */ get originalAngle(): number { return this._originalAngle; }

  /**
   * This read-only property returns the value returned by the call to {@link #computeRotationPoint}.
   * The value is invalid when this tool is not active.
   * @since 2.0
   */
  /** @export */ get rotationPoint(): Point { return this._rotationPoint; }

  /**
   * Gets or sets the spot to locate the Adornment for the rotation handle when it does not have a Placeholder.
   * This assumes the Adornment's location will be at the center of the rotation handle.
   *
   * By default this is zero degrees, which locates the Adornment
   * so that it is placed to the right of an un-rotated object.
   * You can set this to a multiple of 90 degrees to position the rotation handle differently
   * with respect to the {@link #rotationPoint}.
   * @since 2.0
   */
  /** @export */ get handleAngle(): number { return this._handleAngle; }
  /** @export */ set handleAngle(value: number) {
    Util.checkPrimitive(value, 'number', RotatingTool, 'handleAngle');
    this._handleAngle = value;
  }

  /**
   * Gets or sets the spot to locate the Adornment at a particular distance from the {@link #adornedObject}.
   * This assumes the Adornment's location will be at the center of the rotation handle.
   *
   * By default this is 50, which locates the Adornment
   * so that it is placed to the right of an un-rotated object.
   * @since 2.0
   */
  /** @export */ get handleDistance(): number { return this._handleDistance; }
  /** @export */ set handleDistance(value: number) {
    Util.checkPrimitive(value, 'number', RotatingTool, 'handleDistance');
    this._handleDistance = value;
  }

} // end class
RotatingTool['className'] = 'RotatingTool';

// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'rotatingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): RotatingTool {
    return this.findTool('Rotating') as RotatingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: RotatingTool) { this.replaceStandardTool('Rotating', value, this.mouseDownTools); }
);


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
ClickSelectingTool
ActionTool
ClickCreatingTool
DragSelectingTool
PanningTool
*/

// import { List } from './collections';
// import { GraphObject } from './graphObject';
// import { Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData, root } from './root';
// import { Shape } from './shape';
// import { Tool } from './tool';
// import { Util } from './util';


/**
 * The ClickSelectingTool selects and deselects {@link Part}s when there is a click.
 * It does this by calling {@link Tool#standardMouseSelect}.
 * It is also responsible for handling and dispatching click events on {@link GraphObject}s
 * by calling {@link Tool#standardMouseClick}.
 *
 * Note that this tool avoids raising click events on objects that are in temporary layers.
 * This is to prevent parts such as selection adornments from interfering with clicking on selected nodes or links.
 * (Adornments are in the "Adornment" {@link Layer}, which {@link Layer#isTemporary}.)
 * However this means that if you add a {@link GraphObject#click} event handler on a GraphObject in an Adornment,
 * it will not be called.
 * You can get it to be called by setting {@link GraphObject#isActionable} to true on that object in the adornment.
 *
 * This tool is a standard mouse-up tool, the {@link ToolManager#clickSelectingTool}.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 * This tool does not modify the model or conduct any transaction.
 *
 * An example customization of this tool is shown in the <a href="../../extensions/TreeMap.html">Tree Map</a> sample,
 * where the {@link Tool#standardMouseSelect} method is overridden to permit the user to cycle through
 * the chain of containing groups, changing the selection on each click to the next containing group.
 *
 * If you want to programmatically select a Part, you can set {@link Part#isSelected} or call {@link Diagram#select}.
 * If you want to know which Part is at a particular point, you can call {@link Diagram#findPartAt}.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ClickSelectingTool extends Tool {
  /** @hidden @internal */ public static className = 'ClickSelectingTool';
  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#clickSelectingTool}.
   *
   * The {@link Tool#name} of this tool is "ClickSelecting".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ClickSelectingTool);
    this.name = 'ClickSelecting';
  }

  /**
   * This tool can run whenever a click occurs.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ClickSelectingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    // don't check for  !diagram.allowSelect  any more,
    // to always support click events on GraphObjects
    if (this.isBeyondDragSize()) return false;
    return true;
  }

  /**
   * Upon a click, this calls {@link Tool#standardMouseSelect} to change the {@link Diagram#selection} collection,
   * then calls {@link Tool#standardMouseClick} to perform the normal click behaviors,
   * and then stops this tool.
   * @this {ClickSelectingTool}
   */
  public doMouseUp(): void {
    if (this.isActive) {
      this.standardMouseSelect();
      const handled = this.standardMouseClick();
      if (!handled) {
        if (this.diagram.lastInput.isTouchEvent) {
          this.diagram.toolManager.doToolTip();
        }
      }
    }
    this.stopTool();
  }

} // #end ClickSelectingTool
ClickSelectingTool['className'] = 'ClickSelectingTool';

/**
 * The ActionTool is responsible for handling and dispatching mouse events on {@link GraphObject}s
 * that have {@link GraphObject#isActionable} set to true.
 * This is how one implements "controls", such as buttons or sliders or knobs, as {@link GraphObject}s
 * that can be inside {@link Part}s without interfering with the standard tool behaviors.
 *
 * This tool allows individual GraphObjects (usually Panels) to handle mouse-down-move-up events without
 * having to define new {@link Tool}s.
 * It does this by calling {@link GraphObject#actionDown}, {@link GraphObject#actionMove}, {@link GraphObject#actionUp},
 * and {@link GraphObject#actionCancel} on objects that have {@link GraphObject#isActionable} set to true.
 *
 * This tool is a standard mouse-down tool, the {@link ToolManager#actionTool}.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 * This tool does not modify the model or conduct any transaction, although the actions that this invokes may do so.
 *
 * It would be very unusual to want to customize this tool.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ActionTool extends Tool {
  /** @hidden @internal */ public static className = 'ActionTool';
  /** @hidden @internal */ private _actObj: GraphObject | null;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#actionTool}.
   *
   * The {@link Tool#name} of this tool is "Action".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ActionTool);
    this.name = 'Action';
    this._actObj = null;
  }

  /**
   * This tool can run when there is a mouse-down on an object with {@link GraphObject#isActionable} true
   * or if the object is within a {@link Panel} that "isActionable".
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ActionTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    // don't check for !diagram.allowSelect
    const e = diagram.lastInput;
    const curobj = diagram.findObjectAt(e.documentPoint,
      function (x) {
        while (x.panel !== null) {
          if (x.isActionable) return x;
          x = x.panel;
        }
        return x;
      });
    if (curobj !== null) {
      if (!curobj.isActionable) return false;
      this._actObj = curobj;
      // ActionTool still needs the plain currentObject in addition
      diagram._currentEnterLeaveObject = diagram.findObjectAt(e.documentPoint, null, null);
      return true;
    }
    return false;
  }

  /**
   * If there is a {@link GraphObject} found with {@link GraphObject#isActionable} set to true,
   * call that object's {@link GraphObject#actionDown} event, if it exists.
   * @this {ActionTool}
   */
  public doMouseDown(): void {
    if (!this.isActive) {
      if (this.canStart()) {
        this.doActivate();
      }
    } else {
      const diagram = this.diagram;
      const e = diagram.lastInput;
      const curobj = this._actObj;
      if (curobj === null) return;
      e.targetObject = curobj;
      if (curobj.actionDown !== null) {
        curobj.actionDown(e, curobj);
      }
    }
  }

  /**
   * If this tool is active call {@link GraphObject#actionMove}, if it exists, on the active object.
   * @this {ActionTool}
   */
  public doMouseMove(): void {
    if (this.isActive) {
      const diagram = this.diagram;
      const e = diagram.lastInput;
      const curobj = this._actObj;
      if (curobj === null) return;
      e.targetObject = curobj;
      if (curobj.actionMove !== null) {
        curobj.actionMove(e, curobj);
      }
    }
  }

  /**
   * Calls the {@link GraphObject#actionUp} event if defined, then effectively calls
   * {@link Tool#standardMouseClick} to perform the normal click behaviors,
   * and then stops this tool.
   * @this {ActionTool}
   */
  public doMouseUp(): void {
    if (this.isActive) {
      const diagram = this.diagram;
      const e = diagram.lastInput;
      const curobj = this._actObj;
      if (curobj === null) return;
      e.targetObject = curobj;
      if (curobj.actionUp !== null) {
        curobj.actionUp(e, curobj);
      }
      // Do a "click" too, if one is defined and the mouseup is in the original object
      this.standardMouseClick(
        (x: GraphObject): GraphObject => {
          while (x.panel !== null) {
            if (x.isActionable && x === curobj) return x;
            x = x.panel;
          }
          return x;
        },
        (x: GraphObject): boolean => x === curobj
      );
    }
    this.stopTool();
  }

  /**
   * Call the {@link GraphObject#actionCancel} event if defined on the current object.
   * @this {ActionTool}
   */
  public doCancel(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    const curobj = this._actObj;
    if (curobj === null) return;
    e.targetObject = curobj;
    if (curobj.actionCancel !== null) {
      curobj.actionCancel(e, curobj);
    }
    this.stopTool();
  }

  /**
   * @hidden @internal
   * @this {ActionTool}
   */
  public doStop(): void {
    this._actObj = null;
  }

} // #end ActionTool
ActionTool['className'] = 'ActionTool';

/**
 * The ClickCreatingTool lets the user create a node by clicking where they want the new node to be.
 * By default a double-click is required to start this tool;
 * set {@link #isDoubleClick} to false if you want a single-click to create a node.
 *
 * This tool is a standard mouse-up tool, the {@link ToolManager#clickCreatingTool}.
 * However, it will not be able to start running unless you have set the
 * {@link #archetypeNodeData} property to an object that can be copied and added to the diagram's model.
 *
 * If you set {@link #isGridSnapEnabled} to true, newly created parts will be snapped to points on a grid.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 * This tool does conduct a transaction when inserting the new node.
 *
 * If you want to programmatically create a Part, you can call {@link #insertPart}.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ClickCreatingTool extends Tool {
  /** @hidden @internal */ public static className = 'ClickCreatingTool';
  /** @hidden @internal */ private _archetypeNodeData: ObjectData | null;
  /** @hidden @internal */ private _isDoubleClick: boolean;
  /** @hidden @internal */ private _isGridSnapEnabled: boolean;
  /** @hidden @internal */ private _firstPoint: Point;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#clickCreatingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "ClickCreating".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ClickCreatingTool);
    this.name = 'ClickCreating';
    this._archetypeNodeData = null;
    this._isDoubleClick = true;
    this._isGridSnapEnabled = false;
    this._firstPoint = new Point(0, 0);
  }

  /**
   * This tool can run when the diagram is not read-only and supports creating new nodes,
   * and when there has been a click (or double-click if {@link #isDoubleClick} is true)
   * in the background of the diagram (not on a {@link Part}),
   * and {@link #archetypeNodeData} is an object that can be copied and added to the model.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ClickCreatingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;

    // gotta have some node data that can be copied
    if (this.archetypeNodeData === null) return false;

    const diagram = this.diagram;

    // heed IsReadOnly & AllowInsert
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowInsert) return false;

    // only works with the left button
    if (!diagram.lastInput.left) return false;

    // the mouse down point needs to be near the mouse up point
    if (this.isBeyondDragSize()) return false;

    // maybe requires double-click; otherwise avoid accidental double-create
    if (this.isDoubleClick) {
      if (diagram.lastInput.clickCount === 1) this._firstPoint = diagram.lastInput.viewPoint.copy();
      if (diagram.lastInput.clickCount !== 2) return false;
      if (this.isBeyondDragSize(this._firstPoint)) return false;
    } else {
      if (diagram.lastInput.clickCount !== 1) return false;
    }

    // don't include the following check when this tool is running modally
    if (diagram.currentTool !== this) {
      // only operates in the background, not on some Part
      const part = diagram.findPartAt(diagram.lastInput.documentPoint, true);
      if (part !== null) return false;
    }

    return true;
  }

  /**
   * Upon a click, call {@link #insertPart} and stop this tool.
   * @this {ClickCreatingTool}
   */
  public doMouseUp(): void {
    const diagram = this.diagram;
    if (this.isActive) {
      this.insertPart(diagram.lastInput.documentPoint);
    }
    this.stopTool();
  }

  /**
   * Create a node by adding a copy of the {@link #archetypeNodeData} object
   * to the diagram's model, assign its {@link Part#location}
   * to be the given point, and select the new part.
   *
   * This starts and commits a transaction and also raises the "PartCreated" {@link DiagramEvent}.
   * The actual part that is added to the diagram may be a {@link Part}, a {@link Node},
   * or even a {@link Group}, depending on the properties of the {@link #archetypeNodeData}
   * and the type of the template that is copied to create the part.
   *
   * This method also raises the "ChangingSelection" and "ChangedSelection" diagram events outside the transaction.
   *
   * This method may be overridden, although it is usually much easier to just set {@link #archetypeNodeData}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ClickCreatingTool}
   * @param {Point} loc a Point in document coordinates.
   * @return {Part} the newly created Part, or null if it failed.
   */
  public insertPart(loc: Point): Part {
    const diagram = this.diagram;
    const arch = this.archetypeNodeData;
    if (arch === null) return null;
    let part: Part = null;

    try {
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      this.startTransaction(this.name);
      if (arch instanceof Part) {
        if (arch.isNodeLike()) {
          arch.freezeBindings();
          part = arch.copy();
          if (part !== null) {
            diagram.add(part);
          }
        }
      } else if (arch !== null) {
        const data = diagram.model.copyNodeData(arch);
        if (Util.isObject(data)) {
          diagram.model.addNodeData(data);
          part = diagram.findPartForData(data);
        }
      }
      if (part !== null) {
        // The following code should really call Part.move or Diagram.moveParts
        // To make sure groups with nodes that have been added have their children snap-or-not correctly.
        const snappt = Point.allocAt(loc.x, loc.y);
        if (this.isGridSnapEnabled) {
          this.diagram.snapTo(part, loc, snappt);
        }

        part.location = snappt;
        if (diagram.allowSelect) {
          diagram.internalClearSelection();
          part.isSelected = true;  // don't raise selection events
        }
        Point.free(snappt);
      }

      diagram.invalidateDocumentBounds();
      // set the TransactionResult before raising event, in case it changes the result or cancels the tool
      this.transactionResult = this.name;
      diagram.raiseDiagramEvent('PartCreated', part);
    } finally {
      this.stopTransaction();
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
    }
    return part;
  }

  /**
   * Gets or sets a data object that will be copied and added to the diagram's model each time this tool executes.
   * The default value is null.
   * The value must be non-null for this tool to be able to run.
   * Setting this property does not raise any events.
   */
  /** @export */ get archetypeNodeData(): ObjectData | null { return this._archetypeNodeData; }
  /** @export */ set archetypeNodeData(value: ObjectData | null) {
    if (value !== null) Util.checkPrimitive(value, 'object', ClickCreatingTool, 'archetypeNodeData');
    this._archetypeNodeData = value;
  }

  /**
   * Gets or sets whether a double click rather than a single-click is required
   * to insert a new {@link Part} at the mouse-up point.
   * The default value is true -- only a double-click will cause a node to be created.
   * Setting this property does not raise any events.
   */
  /** @export */ get isDoubleClick(): boolean { return this._isDoubleClick; }
  /** @export */ set isDoubleClick(value: boolean) {
    Util.checkPrimitive(value, 'boolean', ClickCreatingTool, 'isDoubleClick');
    this._isDoubleClick = value;
  }

  /**
   * Gets or sets whether the {@link ClickCreatingTool} snaps object sizes to the diagram's background grid during creation.
   * By default this property is false.
   * Setting this property does not raise any events.
   *
   * @since 2.0
   */
  /** @export */ get isGridSnapEnabled(): boolean { return this._isGridSnapEnabled; }
  /** @export */ set isGridSnapEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', ClickCreatingTool, 'isGridSnapEnabled');
    this._isGridSnapEnabled = value;
  }

} // #end ClickCreatingTool
ClickCreatingTool['className'] = 'ClickCreatingTool';


/**
 * The DragSelectingTool lets the user select multiple parts within a rectangular area drawn by the user.
 * There is a temporary part, the {@link #box},
 * that shows the current area encompassed between the mouse-down
 * point and the current mouse point.
 * The default drag selection box is a magenta rectangle.
 * You can change the {@link #box} to customize its appearance -- see its documentation for an example.
 *
 * This tool is a standard mouse-move tool, the {@link ToolManager#dragSelectingTool}.
 * However this cannot start running unless there has been a motionless delay
 * after the mouse-down event of at least {@link #delay} milliseconds.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles,
 * but it does temporarily add the {@link #box} part to the diagram.
 * This tool does not modify the model or conduct any transaction.
 *
 * Selection occurs on a mouse-up when it calls {@link #selectInRect}
 * with the value of {@link #computeBoxBounds}.
 * Selectable parts are selected when their bounds fall entirely within the rectangle,
 * unless {@link #isPartialInclusion} is set to true.
 * <p class="boxread">
 * For customizing the DragSelectingTool, see <a href="../../intro/tools.html#DragSelectingTool">Introduction to the DragSelectingTool</a>.
 *
 * If you implement your own drag-in-the-background-to-do-something tool, you may need to disable
 * this tool or insert your new tool in the {@link ToolManager#mouseMoveTools} list before this tool,
 * in order for your tool to run.  There are examples of such tools defined in the extensions directory:
 * <a href="../../extensions/RealtimeDragSelecting.html">Realtime Drag Selecting Tool</a>,
 * <a href="../../extensions/DragCreating.html">Drag Creating Tool</a>, and
 * <a href="../../extensions/DragZooming.html">Drag Zooming Tool</a>.
 *
 * If you want to programmatically select some Parts in a rectangular area,
 * you can call {@link #selectInRect}.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class DragSelectingTool extends Tool {
  /** @hidden @internal */ public static className = 'DragSelectingTool';
  /** @hidden @internal */ private _delay: number;
  /** @hidden @internal */ private _isPartialInclusion: boolean;
  /** @hidden @internal */ private _box: Part | null;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#dragSelectingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "DragSelecting".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(DragSelectingTool);
    this.name = 'DragSelecting';
    this._delay = 175;
    this._isPartialInclusion = false;

    const b = new Part();
    b.layerName = 'Tool';
    b.selectable = false;
    const r = new Shape();
    r.name = 'SHAPE';
    r.figure = 'Rectangle';
    r.fill = null;
    r.stroke = 'magenta';
    b.add(r);
    this._box = b;
  }

  /**
   * This tool can run when the diagram allows selection,
   * there has been delay of at least {@link #delay} milliseconds
   * after the mouse-down before a mouse-move,
   * there has been a mouse-drag far enough away not to be a click,
   * and there is no selectable part at the mouse-down point.
   *
   * The delay required to start this tool enables both this tool and
   * the {@link ToolManager#panningTool} to co-exist as mode-less mouse-move tools.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DragSelectingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (!diagram.allowSelect) return false;
    const e = diagram.lastInput;
    // require left button & that it has moved far enough away from the mouse down point, so it isn't a click
    if (!e.left) return false;
    // don't include the following checks when this tool is running modally
    if (diagram.currentTool !== this) {
      if (!this.isBeyondDragSize()) return false;
      // must wait for "delay" milliseconds before that tool can run
      if (e.timestamp - diagram.firstInput.timestamp < this.delay) return false;
      // don't start if we're over a selectable part
      if (diagram.findPartAt(e.documentPoint, true) !== null) return false;
    }
    return true;
  }

  /**
   * Capture the mouse and show the {@link #box}.
   * @this {DragSelectingTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    this.isActive = true;
    diagram.isMouseCaptured = true;
    diagram.skipsUndoManager = true;
    diagram.add(this.box);
    this.doMouseMove();
  }

  /**
   * Release the mouse and remove any {@link #box}.
   * @this {DragSelectingTool}
   */
  public doDeactivate(): void {
    const diagram = this.diagram;
    diagram.stopAutoScroll();
    diagram.remove(this.box);
    diagram.skipsUndoManager = false;
    diagram.isMouseCaptured = false;
    this.isActive = false;
  }

  /**
   * Update the {@link #box}'s position and size according to the value
   * of {@link #computeBoxBounds}.
   * @this {DragSelectingTool}
   */
  public doMouseMove(): void {
    const diagram = this.diagram;
    if (this.isActive && this.box !== null) {
      const r = this.computeBoxBounds();
      let shape = this.box.findObject('SHAPE');
      if (shape === null) shape = this.box.findMainElement();
      const sz = Size.alloc().assignTo(r.width, r.height);
      const pt = Point.allocAt(r.x, r.y);
      shape.desiredSize = sz;
      this.box.position = pt;
      Size.free(sz);
      Point.free(pt);

      if (diagram.allowHorizontalScroll || diagram.allowVerticalScroll) {
        diagram.doAutoScroll(diagram.lastInput.viewPoint);
      }
    }
  }

  /**
   * Call {@link #selectInRect} with the value of a call to {@link #computeBoxBounds}.
   *
   * This method changes the cursor to "wait" and
   * raises the "ChangingSelection" {@link DiagramEvent} before calling {@link #selectInRect},
   * and raises the "ChangedSelection" DiagramEvent afterward.
   * @this {DragSelectingTool}
   */
  public doMouseUp(): void {
    if (this.isActive) {
      const diagram = this.diagram;
      diagram.remove(this.box);
      try {
        diagram.currentCursor = 'wait';
        diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
        this.selectInRect(this.computeBoxBounds());
        diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      } finally {
        diagram.currentCursor = '';
      }
    }
    this.stopTool();
  }

  /**
   * This just returns a {@link Rect} stretching from the mouse-down point to the current mouse point.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DragSelectingTool}
   * @return {Rect} a {@link Rect} in document coordinates.
   */
  public computeBoxBounds(): Rect {
    const diagram = this.diagram;
    return new Rect(diagram.firstInput.documentPoint, diagram.lastInput.documentPoint);
  }

  /**
   * This method is called to select some parts within the area of a given rectangle.
   *
   * The normal behavior is to set the diagram's selection collection to only those parts
   * in the given rectangle according to the {@link #isPartialInclusion} policy.
   * However, if the Shift key modifier is used, no parts are deselected --
   * this adds to the selection the parts in the rectangle not already selected.
   * If the Control key (Command on Mac) modifier is used, this toggles the selectedness of the parts in the rectangle.
   * If the Control key (Command on Mac) and Shift key modifiers are both used, this deselects the parts in the rectangle.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {DragSelectingTool}
   * @param {Rect} r a rectangular bounds in document coordinates.
   */
  public selectInRect(r: Rect): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;

  // ?? Generalize what object in a Part must be within or intersect rect R
  // ?? For example, to require containing or intersecting with the Part.selectionObject:
  // var parts = diagram.findObjectsIn(r,
  //  function(p) {
  //    var part = p.part;
  //    var selobj = part.selectionObject;
  //    if (p === selobj || p.isContainedBy(selobj) || p instanceof go.Link) return part;
  //    return null;
  //  },
  //  function(p) { return p.canSelect(); },
  //  this.isPartialInclusion);

    const parts = diagram.findPartsIn(r, this.isPartialInclusion);
    if (Util.isMac ? e.meta : e.control) {  // toggle or deselect
      if (e.shift) {  // deselect only
        const it = parts.iterator;
        while (it.next()) {
          const p = it.value;
          if (p.isSelected) p.isSelected = false;
        }
      } else {  // toggle selectedness of parts
        const it = parts.iterator;
        while (it.next()) {
          const tp = it.value;
          tp.isSelected = !tp.isSelected;
        }
      }
    } else if (e.shift) {  // extend selection only
      const it = parts.iterator;
      while (it.next()) {
        const ep = it.value;
        if (!ep.isSelected) ep.isSelected = true;
      }
    } else {  // select parts, and unselect all other previously selected parts
      // this tries to avoid deselecting and then reselecting any Part
      const tounselect = new List<Part>();
      const sit = diagram.selection.iterator;
      while (sit.next()) {
        const sp = sit.value;
        if (!parts.contains(sp)) tounselect.add(sp);
      }
      const uit = tounselect.iterator;
      while (uit.next()) {
        const up = uit.value;
        up.isSelected = false;
      }
      const it = parts.iterator;
      while (it.next()) {
        const ps = it.value;
        if (!ps.isSelected) ps.isSelected = true;
      }
    }
  }

  /**
   * Gets or sets the time in milliseconds for which the mouse must be stationary
   * before this tool can be started.
   * The default value is 175 milliseconds.
   * Setting this property does not raise any events.
   */
  /** @export */ get delay(): number { return this._delay; }
  /** @export */ set delay(value: number) {
    Util.checkPrimitive(value, 'number', DragSelectingTool, 'delay');
    this._delay = value;
  }

  /**
   * Gets or sets whether a selectable {@link Part} may be only partly
   * or must be completely enclosed by the rectangle given to {@link #selectInRect}.
   * The default value is false: parts must be completely inside the rectangle.
   * Setting this property does not raise any events.
   */
  /** @export */ get isPartialInclusion(): boolean { return this._isPartialInclusion; }
  /** @export */ set isPartialInclusion(value: boolean) {
    Util.checkPrimitive(value, 'boolean', DragSelectingTool, 'isPartialInclusion');
    this._isPartialInclusion = value;
  }

  /**
   * Gets or sets the {@link Part} used as the "rubber-band selection box"
   * that is stretched to follow the mouse, as feedback for what area will
   * be passed to {@link #selectInRect} upon a mouse-up.
   *
   * Initially this is a {@link Part} containing only a simple magenta rectangular {@link Shape}.
   * The object to be resized during dragging should be named "SHAPE".
   * Setting this property does not raise any events.
   *
   * Here is an example of changing the selection box to be a thicker bright green rectangle:
   * ```js
   *   myDiagram.toolManager.dragSelectingTool.box =
   *     $(go.Part,
   *       { layerName: "Tool", selectable: false },
   *       $(go.Shape,
   *         { name: "SHAPE", fill: null, stroke: "chartreuse", strokeWidth: 3 }));
   * ```
   * Note that the Part should be put into a {@link Layer} that {@link Layer#isTemporary}.
   *
   * Modifying this property while this tool {@link Tool#isActive} might have no effect.
   */
  /** @export */ get box(): Part | null { return this._box; }
  /** @export */ set box(value: Part | null) {
    if (value !== null) Util.checkClass(value, Part, DragSelectingTool, 'box');
    this._box = value;
  }
} // #end DragSelectingTool
DragSelectingTool['className'] = 'DragSelectingTool';

/**
 * The PanningTool supports manual panning, where the user can shift the
 * {@link Diagram#position} by dragging the mouse.
 *
 * This tool is a standard mouse-move tool, the {@link ToolManager#panningTool}.
 * Although the {@link ToolManager#dragSelectingTool} takes precedence over this tool,
 * the {@link DragSelectingTool} only starts if there is a delay after a mouse-down event.
 * If it does not start, then this PanningTool gets a chance to start.
 *
 * This tool does not utilize any {@link Adornment}s or tool handles.
 * This tool does not modify the model or conduct any transaction.
 *
 * If you want to programmatically "pan" the diagram, you can just set {@link Diagram#position}
 * or call methods such as {@link Diagram#scroll}, {@link Diagram#scrollToRect}, or {@link Diagram#centerRect}.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class PanningTool extends Tool {
  /** @hidden @internal */ public static className = 'PanningTool';
  /** @hidden @internal */ private _originalPosition: Point;
  /** @hidden @internal */ private _tempPosition: Point;
  /** @hidden @internal */ private _bubbles: boolean;
  /** @hidden @internal */ private _stopToolOnPan: (() => void);

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#panningTool}.
   *
   * The {@link Tool#name} of this tool is "Panning".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(PanningTool);
    this.name = 'Panning';
    this._originalPosition = new Point();
    this._tempPosition = new Point();
    this._bubbles = false;
    const tool = this;
    this._stopToolOnPan = function () {
      const diagram = tool.diagram;
      if (diagram !== null) diagram.removeEventListener(root.document, 'scroll', tool._stopToolOnPan, false);
      tool.stopTool();
    };
  }

  /**
   * This tool can run when the diagram allows scrolling and
   * the mouse has been dragged with the left button far enough away
   * from the mouse-down point to avoid being a click.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {PanningTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (!diagram.allowHorizontalScroll && !diagram.allowVerticalScroll) return false;
    // require left button & that it has moved far enough away from the mouse down point, so it isn't a click
    if (!diagram.lastInput.left) return false;
    // don't include the following check when this tool is running modally
    if (diagram.currentTool !== this) {
      // mouse needs to have moved from the mouse-down point
      if (!this.isBeyondDragSize()) return false;
    }
    return true;
  }

  /**
   * Capture the mouse, change the diagram cursor to "move", and remember the {@link Diagram#position}.
   * @this {PanningTool}
   */
  public doActivate(): void {
    const diagram = this.diagram;
    if (this._bubbles) {
      diagram.lastInput.bubbles = true;
      // attach event to root
      diagram.addEventListener(root.document, 'scroll', this._stopToolOnPan, false);
    } else {
      diagram.currentCursor = 'move';
      diagram.isMouseCaptured = true;
      this._originalPosition.assign(diagram.position);
    }
    this.isActive = true;
  }

  /**
   * Release the mouse and restore the default diagram cursor.
   * @this {PanningTool}
   */
  public doDeactivate(): void {
    const diagram = this.diagram;
    diagram.currentCursor = '';
    diagram.isMouseCaptured = false;
    this.isActive = false;
  }

  /**
   * Restore the {@link Diagram#position} to what it was when this tool activated.
   * @this {PanningTool}
   */
  public doCancel(): void {
    const diagram = this.diagram;
    diagram.position = this._originalPosition;
    diagram.isMouseCaptured = false;
    this.stopTool();
  }

  /**
   * Modify the {@link Diagram#position} according to how much the mouse has moved.
   * @this {PanningTool}
   */
  public doMouseMove(): void {
    this.move();
  }

  /**
   * Modify the {@link Diagram#position} according to how much the mouse has moved.
   * @this {PanningTool}
   */
  public doMouseUp(): void {
    this.move();
    this.stopTool();
  }

  /**
   * @hidden @internal
   * @this {PanningTool}
   */
  private move(): void {
    const diagram = this.diagram;
    if (this.isActive && diagram) {
      if (this._bubbles) {
        diagram.lastInput.bubbles = true;
        return;
      }
      const pos = diagram.position;
      const first = diagram.firstInput.documentPoint;
      const last = diagram.lastInput.documentPoint;
      let x = pos.x + first.x - last.x;
      let y = pos.y + first.y - last.y;
      if (!diagram.allowHorizontalScroll) x = pos.x;
      if (!diagram.allowVerticalScroll) y = pos.y;
      diagram.position = this._tempPosition.assignTo(x, y);
      // no maybeUpdate necessary, is called by invalidateViewport from position setter
    }
  }

  /**
   * Gets or sets whether panning actions will allow
   * events to bubble **instead** of panning in the diagram.
   *
   * Set this to true to allow mobile devices to scroll the page with panning gestures on the diagram.
   * Otherwise, the panning events will be captured and used to pan inside of the diagram.
   *
   * The default value is false.
   */
  /** @export */ get bubbles(): boolean { return this._bubbles; }
  /** @export */ set bubbles(value: boolean) {
    Util.checkPrimitive(value, 'boolean', PanningTool, 'bubbles');
    this._bubbles = value;
  }

  /**
   * This read-only property returns the {@link Point} that was the original value of {@link Diagram#position} when the panning operation started.
   * @since 1.1
   */
  /** @export */ get originalPosition(): Point { return this._originalPosition; }

} // #end PanningTool
PanningTool['className'] = 'PanningTool';

// import { Size } from './size';
// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'clickCreatingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): ClickCreatingTool {
    return this.findTool('ClickCreating') as ClickCreatingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: ClickCreatingTool) { this.replaceStandardTool('ClickCreating', value, this.mouseUpTools); }
);

Util.defineProperty(ToolManager, 'clickSelectingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): ClickSelectingTool {
    return this.findTool('ClickSelecting') as ClickSelectingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: ClickSelectingTool) { this.replaceStandardTool('ClickSelecting', value, this.mouseUpTools); }
);

Util.defineProperty(ToolManager, 'panningTool',
  /** @this {ToolManager} */
  function(this: ToolManager): PanningTool {
    return this.findTool('Panning') as PanningTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: PanningTool) { this.replaceStandardTool('Panning', value, this.mouseMoveTools); }
);

Util.defineProperty(ToolManager, 'dragSelectingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): DragSelectingTool {
    return this.findTool('DragSelecting') as DragSelectingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: DragSelectingTool) { this.replaceStandardTool('DragSelecting', value, this.mouseMoveTools); }
);

Util.defineProperty(ToolManager, 'actionTool',
  /** @this {ToolManager} */
  function(this: ToolManager): ActionTool {
    return this.findTool('Action') as ActionTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: ActionTool) { this.replaceStandardTool('Action', value, this.mouseDownTools); }
);

// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { Tool } from './tool';
// import { Util } from './util';

/**
 * HTMLInfo is used to show and hide custom HTML page elements, such as a context menu, tooltip, or text editor made of HTML.
 *
 * Properties that can be set to an HTMLInfo include:
 *   - {@link TextEditingTool#defaultTextEditor}
 *   - {@link TextBlock#textEditor}
 *   - {@link GraphObject#contextMenu}
 *   - {@link Diagram#contextMenu}
 *   - {@link GraphObject#toolTip}
 *   - {@link Diagram#toolTip}
 *
 * When a context menu is set to an instance of HTMLInfo,
 * {@link ContextMenuTool#showContextMenu} and {@link ContextMenuTool#hideContextMenu}
 * call {@link #show} and {@link #hide} respectively. You may define {@link #mainElement}
 * instead of {@link #hide} in order to automatically use a default hide method.
 *
 * When a tooltip is set to an instance of HTMLInfo,
 * {@link ToolManager#showToolTip} and {@link ToolManager#hideToolTip}
 * call {@link #show} and {@link #hide} respectively.
 *
 * When a text editor is set to an instance of HTMLInfo,
 * {@link TextEditingTool#doActivate} calls {@link #show} and {@link TextEditingTool#doDeactivate} calls {@link #hide}.
 *
 * For HTMLInfo to work, you must define {@link #show}
 * and either {@link #hide} or {@link #mainElement}.
 * Typical usage will also stop the {@link ContextMenuTool} once the desired context action occurs,
 * typically by calling `diagram.currentTool.stopTool();`.
 *
 * Example usage of HTMLInfo can be found in the
 * <a href="../../samples/customContextMenu.html">Custom Context Menu</a> and
 * <a href="../../samples/htmlLightBoxContextMenu.html">HTML LightBox Context Menu</a> samples, the
 * <a href="../../samples/customTextEditingTool.html">Custom TextEditingTool sample</a>, and the
 * <a href="../../extensions/textEditor.html">Text Editor implementation extension</a>.
 *
 * Here is the outline for typical usage of HTMLInfo as a context menu:
 * ```js
 * // Assign an HTMLInfo to the Diagram:
 * myDiagram.contextMenu = $(go.HTMLInfo, {
 *   show: showContextMenu,
 *   hide: hideContextMenu
 * });
 *
 * function showContextMenu(obj, diagram, tool) {
 *   // Show the context menu HTML element:
 *   SomeDOMElement.style.display = "block";
 *
 *   // Also show relevant buttons given the current state
 *   // and the GraphObject obj; if null, the context menu is for the whole Diagram
 * }
 *
 * function hideContextMenu() {
 *   SomeDOMElement.style.display = "none";
 * }
 *
 * function buttonClick() {
 *   // do some action when a context menu button is clicked
 *
 *   // then:
 *   myDiagram.currentTool.stopTool();
 * }
 * ```
 *
 * By default, {@link TextEditingTool#defaultTextEditor} is an instance of HTMLInfo.
 * You can see its default implementation details <a href="../../extensions/textEditor.html">here</a>.
 * @unrestricted
 * @since 1.7
 * @category Tool
 */
export class HTMLInfo {
  /** @hidden @internal */ public static className = 'HTMLInfo';
  /** @hidden @internal */ private _show: ((a: GraphObject, b: Diagram, c: Tool) => void) | null;
  /** @hidden @internal */ private _hide: ((a: Diagram, b: Tool) => void) | null;
  /** @hidden @internal */ private _mainElement: HTMLElement | null;
  /** @hidden @internal */ private _valueFunction: (() => void) | null;

  /**
   * A newly constructed HTMLInfo can be assigned as a value of {@link TextEditingTool#defaultTextEditor}, {@link TextBlock#textEditor}
   * {@link GraphObject#contextMenu}, {@link Diagram#contextMenu}, {@link GraphObject#toolTip}, or {@link Diagram#toolTip}.
   */
  constructor() {
    this._show = null;
    this._hide = null;
    this._mainElement = null;

    this._valueFunction = null;
  }

  /**
   * Gets or sets the primary HTML Element that represents this HTMLInfo.
   * In a context menu, this would be the outermost HTML element, the one which typically
   * shows and hides. If this is set and {@link #hide} is not, HTMLInfo will automatically execute:
   * ```js
   *   tool.mainElement.style.display = "none";
   * ```
   *
   * when {@link #hide} would typically be called.
   *
   * This is set only for convenience; the default value for this property is null.
   * @see {@link #hide}
   */
  /** @export */ get mainElement(): HTMLElement | null { return this._mainElement; }
  /** @export */ set mainElement(value: HTMLElement | null) {
    if (value !== null) Util.checkClass(value, HTMLElement, HTMLInfo, 'mainElement');
    this._mainElement = value;
  }

  /**
   * Gets or sets the function to call when an HTMLInfo is to be shown,
   * such as when used as the {@link GraphObject#contextMenu} or {@link Diagram#toolTip} or {@link TextBlock#textEditor}.
   *
   * If this is called by:
   *   - {@link ContextMenuTool#showContextMenu}, the first argument is the {@link GraphObject} for which the context menu is being shown, or null for the diagram background.
   *   - {@link ToolManager#showToolTip}, the first argument is the {@link GraphObject} for which the tooltip is being shown, or null for the diagram background.
   *   - {@link TextEditingTool#doActivate}, the first argument is the {@link TextBlock} for which the text editor is being shown.
   *
   * If you need access to any bound data for the first argument, if it is non-null, you can get it via:
   * `obj.part.data`
   * and then you can look at any of the properties you have put on that data.
   *
   * When used as a context menu, typically shown elements, such as buttons, should call
   * `diagram.currentTool.stopTool();` when their action is completed.
   */
  /** @export */ get show(): ((a: GraphObject, b: Diagram, c: Tool) => void) | null { return this._show; }
  /** @export */ set show(value: ((a: GraphObject, b: Diagram, c: Tool) => void) | null) {
    const old = this._show;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', HTMLInfo, 'show');
      this._show = value;
    }
  }

  /**
   * Gets or sets the function to call when an HTMLInfo is to be hidden.
   * The function should "hide" the HTMLInfo, either by removing any traces of it or otherwise
   * returning the page state to one where the HTMLInfo content is no longer considered active.
   *
   * Setting this is not strictly necessary, if no action is to be taken when hiding the HTMLInfo.
   *
   * If this is called by:
   *   - {@link ContextMenuTool#hideContextMenu}, the second argument is the {@link ContextMenuTool}.
   *   - {@link ToolManager#hideToolTip}, the second argument is the {@link ToolManager}.
   *   - {@link TextEditingTool#doDeactivate}, the second argument is the {@link TextEditingTool}.
   *
   * @see {@link #mainElement}
   */
  /** @export */ get hide(): ((a: Diagram, b: Tool) => void) | null { return this._hide; }
  /** @export */ set hide(value: ((a: Diagram, b: Tool) => void) | null) {
    const old = this._hide;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', HTMLInfo, 'hide');
      this._hide = value;
    }
  }

  /**
   * Gets or sets a function that returns the primary value associated with this HTMLInfo,
   * such as the string value of a text editor, which would be solicited by the {@link TextEditingTool}.
   *
   * This typically returns a string.
   */
  /** @export */ get valueFunction(): (() => any) | null { return this._valueFunction; }
  /** @export */ set valueFunction(value: (() => any) | null) {
    this._valueFunction = value;
  }

} // #end HTMLInfo
HTMLInfo['className'] = 'HTMLInfo';


// import { List } from './collections';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { HTMLInfo } from './htmlInfo';
// import { Adornment, Part } from './parts';
// import { Point } from './point';
// import { root } from './root';
// import { ActionTool } from './selectingTools';
// import { Tool } from './tool';
// import { ToolManager } from './toolManager';
// import { Util } from './util';

/**
 * @hidden @internal
 * This is only used when defining the default context menu
 * @unrestricted
 */
export class ContextMenuButtonInfo {
  /** @hidden @internal */ public static className = 'ContextMenuButtonInfo';
  /** @hidden @internal */ public text: string;
  /** @hidden @internal */ public command: ((a: Diagram) => void);
  /** @hidden @internal */ public visible: ((a: Diagram) => boolean) | null;

  /**
   * @param {string} text
   * @param {function(Diagram)} command
   * @param {function(Diagram):boolean|null} visible
   */
  constructor(text: string, command: ((a: Diagram) => void), visible: ((a: Diagram) => boolean) | null) {
    this.text = text;
    this.command = command;
    this.visible = visible;
  }
} // #end ContextMenuButtonInfo
ContextMenuButtonInfo['className'] = 'ContextMenuButtonInfo';


/**
 * The ContextMenuTool is used to create and show a context menu.
 * It automatically disables any browser context menu.
 *
 * Define context menus on individual {@link GraphObject}s by setting {@link GraphObject#contextMenu}.
 * Define a context menu for the diagram background by setting {@link Diagram#contextMenu}.
 *
 * This tool is a standard mouse-down tool, the {@link ToolManager#contextMenuTool}.
 *
 * This tool does not utilize any tool handles.
 * This tool does not modify the model or conduct any transaction,
 * although any code invoked by context menu commands might do so.
 *
 * There are examples of customizing this tool in the
 * <a href="../../samples/customContextMenu.html">Custom Context Menu</a> and
 * <a href="../../samples/htmlLightBoxContextMenu.html">HTML LightBox Context Menu</a> samples.
 *
 * If you want to programmatically show a context menu for a particular GraphObject or for the
 * whole diagram, call {@link CommandHandler#showContextMenu}.
 * That command method is also invoked by the Menu key on the keyboard.
 *
 * Normally this shows a context menu (if available) on a right-mouse-up event.
 * If you want it to happen on a right-mouse-down event, you'll need to move this tool
 * from the {@link ToolManager#mouseUpTools} list to the {@link ToolManager#mouseDownTools} list:
 * ```js
 * myDiagram.toolManager.mouseDownTools.add(myDiagram.toolManager.replaceTool("ContextMenu", null));
 * ```
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class ContextMenuTool extends Tool {
  /** @hidden @internal */ public static className = 'ContextMenuTool';
  /** @hidden @internal */ private _currentContextMenu: Adornment | HTMLInfo | null;
  /** @hidden @internal */ private _currentContextMenuPart: Part | null;
  /** @hidden @internal */ private _currentObject: GraphObject | null;
  /** @hidden @internal */ private _mouseDownPoint: Point;
  /** @hidden @internal */ private _defaultButtons: List<ContextMenuButtonInfo> | null;
  /** @hidden @internal */ private _toolStopper: (() => void);
  /** @hidden @internal */ private _isOpening: boolean;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#contextMenuTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "ContextMenu".
   */
  constructor() {
    super();
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ContextMenuTool);
    this.name = 'ContextMenu';
    this._currentContextMenu = null;
    this._currentContextMenuPart = null;  // needed when contextmenu is for an element that is removed from the Part
    this._currentObject = null;
    this._mouseDownPoint = new Point();
    this._defaultButtons = null;
    this._isOpening = false;

    const tool = this;
    this._toolStopper = function () { tool.stopTool(); };
  }

  /**
   * @hidden @internal
   * Creates the default context menu for touch systems
   * Called on the first showDefaultContextMenu
   * @this {ContextMenuTool}
   */
  private createDefaultContextMenu(): void {
    // HTML state for the default touch context menu
    const menu = new HTMLInfo();
    menu.show = function (obj, d, tool2) { (tool2 as ContextMenuTool).showDefaultContextMenu(); };
    menu.hide = function (d, tool2) { (tool2 as ContextMenuTool).hideDefaultContextMenu(); };
    ContextMenuTool._defaultTouchContextMenu = menu;
    const tool = this;
    this._toolStopper = function () { tool.stopTool(); };

    const cmLight = Util.createElement('div');
    const cmDark = Util.createElement('div');

    cmLight.style.cssText =
      'top: 0px;' +
      'z-index:10002;' +
      'position: fixed;' +
      'display: none;' +
      'text-align: center;' +
      'left: 25%;' +
      'width: 50%;' +
      'background-color: #F5F5F5;' +
      'padding: 16px;' +
      'border: 16px solid #444;' +
      'border-radius: 10px;' +
      'margin-top: 10px';

    cmDark.style.cssText =
      'z-index:10001;' +
      'position: fixed;' +
      'display: none;' +
      'top: 0;' +
      'left: 0;' +
      'width: 100%;' +
      'height: 100%;' +
      'background-color: black;' +
      'opacity: 0.8;';

    const ns = Util.createElement('style') as any;
    root.document.getElementsByTagName('head')[0].appendChild(ns);

    ns.sheet.insertRule('.goCXul { list-style: none; }', 0);
    ns.sheet.insertRule('.goCXli {' +
      'font:700 1.5em Helvetica, Arial, sans-serif;' +
      'position: relative;' +
      'min-width: 60px; }', 0);
    ns.sheet.insertRule('.goCXa {' +
      'color: #444;' +
      'display: inline-block;' +
      'padding: 4px;' +
      'text-decoration: none;' +
      'margin: 2px;' +
      'border: 1px solid gray;' +
      'border-radius: 10px; }', 0);

    // We don't want the div acting as a context menu to have a (browser) context menu!
    const diagram = this.diagram;
    if (diagram !== null) {
      diagram.addEventListener(cmLight, 'contextmenu', ContextMenuTool.preventDefault, false);
      diagram.addEventListener(cmLight, 'selectstart', ContextMenuTool.preventDefault, false);
      diagram.addEventListener(cmDark, 'contextmenu', ContextMenuTool.preventDefault, false);
    }

    cmLight.className = 'goCXforeground';
    cmDark.className = 'goCXbackground';


    if (root.document.body) {
      root.document.body.appendChild(cmLight);
      root.document.body.appendChild(cmDark);
    }

    ContextMenuTool.cmLight = cmLight;
    ContextMenuTool.cmDark = cmDark;
    ContextMenuTool.defaultContextMenuSetup = true;
  }

  // Used to create the context menu and associated CSS in the DOM only once
  /** @hidden @internal */
  private static _defaultTouchContextMenu: HTMLInfo | Adornment | null = null;
  /** @hidden @internal */
  private static defaultContextMenuSetup = false;
  /** @hidden @internal */
  private static cmDark: HTMLElement = null;
  /** @hidden @internal */
  private static cmLight: HTMLElement = null;

  /**
   * @hidden @internal
   * @this {ContextMenuTool}
   * @param {Event} e
   */
  public static preventDefault(e: Event): boolean { e.preventDefault(); return false; }

  /**
   * Return true if it's a mouse right click that hasn't moved {@link Tool#isBeyondDragSize}
   * and that is on a {@link GraphObject} with a {@link GraphObject#contextMenu}.
   * This is also true if the mouse right click is in the diagram background
   * and the diagram's {@link Diagram#contextMenu} is non-null.
   *
   * On touch devices, a special default context menu will appear even if no object with a context menu is found.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (this.isBeyondDragSize()) return false;
    if (!diagram.lastInput.right) return false;
    if ((diagram.lastInput.isTouchEvent && this.defaultTouchContextMenu !== null) ||
      this.findObjectWithContextMenu() !== null) return true;
    return false;
  }

  /**
   * @hidden @internal
   * @this {ContextMenuTool}
   */
  public doStart(): void {
    const diagram = this.diagram;
    this._mouseDownPoint.set(diagram.firstInput.documentPoint);
  }

  /**
   * @hidden @internal
   * Make sure any context menu has been removed.
   * @this {ContextMenuTool}
   */
  public doStop(): void {
    this.hideContextMenu();
    this.currentObject = null;
  }

  /**
   * Find a {@link GraphObject} at the current mouse point with a {@link GraphObject#contextMenu},
   * or return the {@link Diagram} if there is a {@link Diagram#contextMenu}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   * @param {GraphObject|Diagram=} obj Optional GraphObject with which to start searching for a context menu.
   * If null, the Diagram will be used. If no argument is specified, this method will look for an object at the current mouse point.
   * @return {GraphObject|Diagram|null} something with a `contextMenu`,
   * or null if nothing can be found with a context menu at the current mouse point.
   */
  public findObjectWithContextMenu(obj?: GraphObject | Diagram): GraphObject | Diagram | null {
    if (obj === undefined) obj = null;
    const diagram = this.diagram;
    const e = diagram.lastInput;
    let currobj: GraphObject = null;
    if (obj instanceof Diagram) {
      // assume no GraphObject
    } else if (obj instanceof GraphObject) {
      currobj = obj;
    } else {
      currobj = diagram.findObjectAt(e.documentPoint, null,
        function (x) { return !x.layer.isTemporary; });
    }

    if (currobj !== null) {  // now inside some GraphObject
      let nextobj = currobj;
      while (nextobj !== null) {
        if (nextobj.contextMenu !== null) return nextobj;
        nextobj = nextobj.panel;
      }
      // No context menu found, return the part in case the default menu wants it
      if (diagram.lastInput.isTouchEvent && this.defaultTouchContextMenu) {
        return currobj.part;
      }
    } else {  // not now in any GraphObject
      if (diagram.contextMenu !== null) return diagram;
    }
    return null;
  }

  /**
   * Do nothing, activation is special and relies on doMouseUp
   * @this {ContextMenuTool}
   */
  public doActivate(): void {
  }

  /**
   * Activate this but also execute the normal behavior that would happen on a mouse-up
   * if this tool is in the {@link ToolManager#mouseDownTools} list.
   * Perform {@link ActionTool#doMouseDown} if a ContextMenuButton was hit.
   * @this {ContextMenuTool}
   */
  public doMouseDown(): void {
    super.doMouseDown();
    if (this.isActive && this.currentContextMenu instanceof Adornment) {
      const actionTool = this.diagram.toolManager.findTool('Action') as ActionTool;
      if (actionTool !== null && actionTool.canStart()) {
        actionTool.doActivate();
        actionTool.doMouseDown();
        actionTool.doDeactivate();
      }
    }
    if (this.diagram.toolManager.mouseDownTools.contains(this)) {
      this.doContextClick();
    }
  }

  /**
   * If there is something found by {@link #findObjectWithContextMenu},
   * call {@link #showContextMenu} with that object's {@link GraphObject#contextMenu}
   * or the {@link Diagram#contextMenu}.
   *
   * Once a context menu is being shown,
   * if a click occurs on a part of the context menu, call {@link Tool#standardMouseClick}.
   * Otherwise if the click occurs elsewhere, just stop this tool.
   * Unlike most tools, the first mouse-up should not stop this tool.
   * @this {ContextMenuTool}
   */
  public doMouseUp(): void {
    if (this.isActive && this.currentContextMenu instanceof Adornment) {
      const actionTool = this.diagram.toolManager.findTool('Action') as ActionTool;
      if (actionTool !== null && actionTool.canStart()) {
        actionTool.doActivate();
        actionTool.doCancel();
        actionTool.doDeactivate();
      }
    }
    this.doContextClick();
  }

  /**
   * @hidden @internal
   * @this {ContextMenuTool}
   */
  public doContextClick(): void {
    const diagram = this.diagram;
    if (!this.isActive) {
      if (this.canStart()) {
        this.openMenu(true);
        if (!this.isActive) {
          this.stopTool();
        }
      }
    } else {  // isActive
      const menu = this.currentContextMenu;
      if (menu === null) return;
      if (menu instanceof HTMLInfo) {
        // do anything else?
        // maybe don't  stop the tool for either HTMLInfo or adornment
        // if some ContextMenuTool bool is set
      } else { // menu is an Adornment
        const currobj = diagram.findObjectAt(diagram.lastInput.documentPoint, null, null);
        if (currobj !== null && currobj.isContainedBy(menu)) {
          // click on a context menu button
          this.standardMouseClick(null, null);
        }
      }

      this.stopTool();
      // maybe start another context menu
      if (this.canStart()) {
        diagram.currentTool = this;
        this.doMouseUp();
      }
    }
  }

  /**
   * Called by doContextClick and CommandHandler.showContextMenu
   * @hidden @internal
   * @this {ContextMenuTool}
   * @param {boolean} mouseUp wether or not this was called by a real mouse(up) event
   * @param {Diagram|GraphObject=} cmo
   */
  public openMenu(mouseUp: boolean, cmo?: Diagram | GraphObject | null): void {
    if (cmo === undefined) cmo = null;
    if (this._isOpening) return; // In case someone is opening a context menu on click, which will infinite loop on standardMouseClick
    this._isOpening = true;
    // like ClickSelectingTool:
    if (mouseUp) this.standardMouseSelect();
    const handled = this.standardMouseClick();
    this._isOpening = false;
    if (!handled) {
      // defacto doActivate
      this.isActive = true;
      const defaultCx = ContextMenuTool._defaultTouchContextMenu;
      if (cmo === null) cmo = this.findObjectWithContextMenu();
      if (cmo !== null) {  // might be a GraphObject or the Diagram
        const cm = cmo.contextMenu;
        if (cm !== null) {
          this.currentObject = cmo instanceof GraphObject ? cmo : null;
          this.showContextMenu(cm, this.currentObject);
        } else if (defaultCx !== null) {
          this.showContextMenu(defaultCx, this.currentObject);
        }
        // now currentContextMenu ought to be non-null
      } else if (defaultCx !== null) {
        // no object found, is the default touch context menu enabled?
        this.showContextMenu(defaultCx, null);
      }
    }
  }

  /**
   * Handle mouse-enter, mouse-over, and mouse-leave events, as well as tooltips.
   * @this {ContextMenuTool}
   */
  public doMouseMove(): void {
    // ToolManager calls standardMouseActivity
    // Note that the ToolManager will not be .isActive!
    const tool = this.diagram.toolManager.findTool('Action') as ActionTool;
    if (tool !== null) tool.doMouseMove();
    if (this.isActive) {
      this.diagram.toolManager.doMouseMove();
    }
  }

  /**
   * Show an {@link Adornment} or {@link HTMLInfo} as a context menu.
   *
   * This method is called by the context click ({@link Tool#doMouseDown}) and {@link CommandHandler#showContextMenu}.
   * If you want to programmatically show a context menu for a particular GraphObject or for the
   * whole diagram, do not call this method, which only does one small piece of the process of
   * bringing up a context menu.
   * Instead call {@link CommandHandler#showContextMenu}, which will start this tool
   * and eventually call this method and handle additional input events.
   *
   * For {@link Adornment} context menus: If the object's containing {@link Part} is data-bound,
   * this sets the contextmenu's `Part.data` to the same value.
   * The {@link Adornment#adornedObject} property is set to the {@link GraphObject} for which the menu is being shown.
   *
   * This method sets the {@link #currentContextMenu}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   * @param {Adornment|HTMLInfo} contextmenu
   * @param {GraphObject} obj the {@link GraphObject} for which the context menu is being shown;
   *   this is null if the contextmenu is being shown for the diagram background.
   */
  public showContextMenu(contextmenu: Adornment | HTMLInfo, obj: GraphObject | null): void {
    if (Debug && !((contextmenu instanceof Adornment) || (contextmenu instanceof HTMLInfo))) {
      Util.throwError('showContextMenu:contextMenu must be an Adornment or HTMLInfo.');
    }
    if (obj !== null) Util.checkClass(obj, GraphObject, ContextMenuTool, 'showContextMenu:obj');
    const diagram = this.diagram;

    // hide any other existing context menu
    if (contextmenu !== this.currentContextMenu) {
      this.hideContextMenu();
    }

    if (contextmenu instanceof Adornment) {
      // always use Tool layer
      contextmenu.layerName = 'Tool';
      contextmenu.selectable = false;
      contextmenu.scale = 1 / diagram.scale;
      contextmenu.category = this.name;
      if (contextmenu.placeholder !== null) contextmenu.placeholder.scale = diagram.scale;

      // show
      const olddiag = contextmenu.diagram;
      if (olddiag !== null && olddiag !== diagram) olddiag.remove(contextmenu);
      diagram.add(contextmenu);

      // maybe data-bind
      if (obj !== null) {
        // this will set contextmenu.data automatically
        contextmenu.adornedObject = obj;
      } else {  // for diagram background, bind to model
        contextmenu.data = diagram.model;
      }

      // position inside the viewport
      contextmenu.ensureBounds();
      this.positionContextMenu(contextmenu, obj);
    } else if (contextmenu instanceof HTMLInfo) {
      contextmenu.show(obj, diagram, this);
    }

    // remember
    this.currentContextMenu = contextmenu;
  }

  /**
   * This is called by {@link #showContextMenu} to position the context menu within the viewport.
   * It normally goes just below the cursor.
   * But if the mouse is too close to the right edge or the bottom edge of the viewport,
   * it is positioned left and/or above the cursor.
   *
   * This method only operates if the context menu, an {@link Adornment}, does not have a {@link Adornment#placeholder}.
   * When there is a {@link Placeholder} in the context menu, that Adornment is automatically positioned
   * so that the Placeholder is positioned at the adorned object, the second argument to this method.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * If you override this method to position the context menu,
   * the context menu has already been measured but not arranged,
   * so you can use its {@link GraphObject#measuredBounds} width and height
   * but not its {@link GraphObject#actualBounds}.
   * @expose
   * @this {ContextMenuTool}
   * @param {Adornment} contextmenu
   * @param {GraphObject} obj The {@link GraphObject} getting the context menu,
   * or null if the context menu is for the diagram background.
   */
  public positionContextMenu(contextmenu: Adornment, obj: GraphObject | null): void {
    if (contextmenu.placeholder !== null) return;

    const diagram = this.diagram;

    const p = diagram.lastInput.documentPoint.copy();
    const ttb = contextmenu.measuredBounds;
    const viewb = diagram.viewportBounds;
    // when touch event -- shift towards the left, so it's not obscured by the finger
    if (diagram.lastInput.isTouchEvent) {
      p.x -= ttb.width;
    }
    // if extends too far to the right -- shift left
    if (p.x + ttb.width > viewb.right) {
      p.x -= ttb.width + 5 / diagram.scale;
    }
    // but don't go beyond the left edge of the viewport
    if (p.x < viewb.x) {
      p.x = viewb.x;
    }
    // if extends too far down -- shift up
    if (p.y + ttb.height > viewb.bottom) {
      p.y -= ttb.height + 5 / diagram.scale;
    }
    // but don't go beyond the top edge of the viewport
    if (p.y < viewb.y) {
      p.y = viewb.y;
    }
    contextmenu.position = p;
  }

  /**
   * Hide any context menu.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   */
  public hideContextMenu(): void {
    const diagram = this.diagram;
    const current = this.currentContextMenu;
    if (current === null) return;

    if (current instanceof Adornment) {
      diagram.remove(current);
      if (this._currentContextMenuPart !== null) this._currentContextMenuPart.removeAdornment(current.category);
      current.data = null;
      current.adornedObject = null;
    } else if (current instanceof HTMLInfo) {
      if (current.hide !== null) {
        current.hide(diagram, this);
      } else if (current.mainElement !== null) {
        current.mainElement.style.display = 'none';
      }
    }
    this.currentContextMenu = null;
    this.standardMouseOver();
  }

  /**
   * @hidden @internal
   * @this {ContextMenuTool}
   * @return {List.<ContextMenuButtonInfo>} of objects describing button functionality,
   * each with the properties "text", "command", and "visible".
   */
  private initializeDefaultButtons(): List<ContextMenuButtonInfo> {
    const tool = this;
    const buttons = new List<ContextMenuButtonInfo>();

    buttons.add(new ContextMenuButtonInfo('Copy',
      function (diagram) { diagram.commandHandler.copySelection(); },
      function (diagram) { return diagram.commandHandler.canCopySelection(); }));
    buttons.add(new ContextMenuButtonInfo('Cut',
      function (diagram) { diagram.commandHandler.cutSelection(); },
      function (diagram) { return diagram.commandHandler.canCutSelection(); }));
    buttons.add(new ContextMenuButtonInfo('Delete',
      function (diagram) { diagram.commandHandler.deleteSelection(); },
      function (diagram) { return diagram.commandHandler.canDeleteSelection(); }));
    buttons.add(new ContextMenuButtonInfo('Paste',
      function (diagram) { diagram.commandHandler.pasteSelection(tool.mouseDownPoint); },
      function (diagram) { return diagram.commandHandler.canPasteSelection(tool.mouseDownPoint); }));
    buttons.add(new ContextMenuButtonInfo('Select All',
      function (diagram) { diagram.commandHandler.selectAll(); },
      function (diagram) { return diagram.commandHandler.canSelectAll(); }));

    buttons.add(new ContextMenuButtonInfo('Undo',
      function (diagram) { diagram.commandHandler.undo(); },
      function (diagram) { return diagram.commandHandler.canUndo(); }));
    buttons.add(new ContextMenuButtonInfo('Redo',
      function (diagram) { diagram.commandHandler.redo(); },
      function (diagram) { return diagram.commandHandler.canRedo(); }));

    buttons.add(new ContextMenuButtonInfo('Scroll To Part',
      function (diagram) { diagram.commandHandler.scrollToPart(); },
      function (diagram) { return diagram.commandHandler.canScrollToPart(); }));
    buttons.add(new ContextMenuButtonInfo('Zoom To Fit',
      function (diagram) { diagram.commandHandler.zoomToFit(); },
      function (diagram) { return diagram.commandHandler.canZoomToFit(); }));
    buttons.add(new ContextMenuButtonInfo('Reset Zoom',
      function (diagram) { diagram.commandHandler.resetZoom(); },
      function (diagram) { return diagram.commandHandler.canResetZoom(); }));

    buttons.add(new ContextMenuButtonInfo('Group Selection',
      function (diagram) { diagram.commandHandler.groupSelection(); },
      function (diagram) { return diagram.commandHandler.canGroupSelection(); }));
    buttons.add(new ContextMenuButtonInfo('Ungroup Selection',
      function (diagram) { diagram.commandHandler.ungroupSelection(); },
      function (diagram) { return diagram.commandHandler.canUngroupSelection(); }));
    buttons.add(new ContextMenuButtonInfo('Edit Text',
      function (diagram) { diagram.commandHandler.editTextBlock(); },
      function (diagram) { return diagram.commandHandler.canEditTextBlock(); }));

    return buttons;
  }

  /**
   * This is the {@link HTMLInfo#show} method for the {@link #defaultTouchContextMenu}.
   *
   * If the object's containing {@link Part} is data-bound,
   * set the contextmenu's `Part.data` to the same value.
   * The {@link Adornment#adornedObject} property is set to the {@link GraphObject}
   * for which the menu is being shown.
   *
   * The menu carries some default CSS styling and uses the following CSS classes:
   *
   * * `goCXforeground` for the DIV containing the buttons
   * * `goCXbackground` for the darker DIV behind the context menu
   * * `goCXul` for the HTML `ul` items
   * * `goCXli` for the HTML `li` tag items
   * * `goCXa` for the HTML `a` tag items
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a>
   * for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   */
  public showDefaultContextMenu(): void {
    const diagram = this.diagram;

    if (this._defaultButtons === null) {
      this._defaultButtons = this.initializeDefaultButtons();
    }

    // Show only the relevant buttons
    ContextMenuTool.cmLight.innerHTML = ''; // empty?

    ContextMenuTool.cmDark.addEventListener('click', this._toolStopper, false);

    const tool = this;
    const ul = Util.createElement('ul');
    ul.className = 'goCXul';
    ContextMenuTool.cmLight.appendChild(ul);
    ul.innerHTML = '';

    const itr = this._defaultButtons.iterator;
    while (itr.next()) {
      const button = itr.value;
      const command = button.command;
      const visible = button.visible;

      if (!(typeof command === 'function')) continue;
      if (typeof visible === 'function' && !visible(diagram)) continue;
      const li = Util.createElement('li');
      li.className = 'goCXli';
      const ahref = Util.createElement('a') as HTMLAnchorElement;
      ahref.className = 'goCXa';
      ahref.href = '#';
      ahref._command = button.command;
      ahref.addEventListener('click', function (e: Event) {
        this._command(diagram);
        tool.stopTool();
        e.preventDefault();
        return false;
      }, false);
      ahref.textContent = button.text;
      li.appendChild(ahref);
      ul.appendChild(li);
    }

    ContextMenuTool.cmLight.style.display = 'block';
    ContextMenuTool.cmDark.style.display = 'block';
  }

  /**
   * This is the {@link HTMLInfo#hide} method for the {@link #defaultTouchContextMenu}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a>
   * for how to override methods and how to call this base method.
   * @expose
   * @this {ContextMenuTool}
   */
  public hideDefaultContextMenu(): void {
    if (this.currentContextMenu === null) return;
    if (this.currentContextMenu !== ContextMenuTool._defaultTouchContextMenu) return;
    ContextMenuTool.cmLight.style.display = 'none';
    ContextMenuTool.cmDark.style.display = 'none';
    const diagram = this.diagram;
    if (diagram !== null) diagram.removeEventListener(ContextMenuTool.cmDark, 'click', this._toolStopper, false);
    this.currentContextMenu = null;
  }

  /**
   * Gets or sets the currently showing context menu, or null if there is none.
   * This is typically only set in {@link #showContextMenu} and not by the user.
   * It is also typically set to null in {@link #hideContextMenu}.
   */
  /** @export */ get currentContextMenu(): Adornment | HTMLInfo | null { return this._currentContextMenu; }
  /** @export */ set currentContextMenu(value: Adornment | HTMLInfo | null) {
    if (Debug && value !== null && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
      Util.throwError('ContextMenuTool.currentContextMenu must be an Adornment or HTMLInfo.');
    }
    this._currentContextMenu = value;
    this._currentContextMenuPart = (value instanceof Adornment) ? value.adornedPart : null;
  }

  /**
   * Gets or sets the HTMLInfo or Adornment that acts as the default touch context menu.
   * On touch devices, this context menu will appear even there is no context menu defined.
   *
   * You can disable this functionality by setting this property to null.
   *
   * By default shows a set of HTML elements acting as a context menu.
   * @since 1.7
   */
  /** @export */ get defaultTouchContextMenu(): Adornment | HTMLInfo | null {
    if (ContextMenuTool.defaultContextMenuSetup === false && ContextMenuTool._defaultTouchContextMenu === null && Diagram.isUsingDOM()) {
      this.createDefaultContextMenu();
    }
    return ContextMenuTool._defaultTouchContextMenu;
  }
  /** @export */ set defaultTouchContextMenu(value: Adornment | HTMLInfo | null) {
    if (Debug && value !== null && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
      Util.throwError('ContextMenuTool.defaultTouchContextMenu must be an Adornment or HTMLInfo.');
    }
    if (value === null) { // Explicitly setting the default context menu to null will stop one from ever being made.
      ContextMenuTool.defaultContextMenuSetup = true;
    }
    ContextMenuTool._defaultTouchContextMenu = value;
  }

  /**
   * Gets or sets the {@link GraphObject} found at the mouse point that has a context menu.
   * This property remembers the result returned by {@link #findObjectWithContextMenu} if it is a GraphObject.
   * This value is passed to {@link #showContextMenu} as the second argument.
   * The value will be null if the context menu is for the diagram rather than for a particular GraphObject.
   * @since 1.4
   */
  /** @export */ get currentObject(): GraphObject | null { return this._currentObject; }
  /** @export */ set currentObject(value: GraphObject | null) {
    if (value !== null) Util.checkClass(value, GraphObject, ContextMenuTool, 'currentObject');
    this._currentObject = value;
  }

  /**
   * This read-only property returns the original mouse-down point in document coordinates.
   */
  /** @export */ get mouseDownPoint(): Point { return this._mouseDownPoint; }

} // #end ContextMenuTool
ContextMenuTool['className'] = 'ContextMenuTool';

Util.defineProperty(ToolManager, 'contextMenuTool',
  /** @this {ToolManager} */
  function(this: ToolManager): ContextMenuTool {
    return this.findTool('ContextMenu') as ContextMenuTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: ContextMenuTool) { this.replaceStandardTool('ContextMenu', value, this.mouseUpTools); }
);


// #end (\w+)[\r\n]
/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
TextEditingTool
*/

// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { HTMLInfo } from './htmlInfo';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { TextBlock } from './textBlock';
// import { Tool } from './tool';
// import { Util } from './util';

/**
 * The TextEditingTool is used to let the user interactively edit text in place.
 * This sets the {@link TextBlock#text} property; you may want to save the changed text to the model
 * by using a TwoWay {@link Binding} on the "text" property of editable TextBlocks.
 *
 * Typically this is used by setting the {@link TextBlock#editable} property to true
 * on a particular {@link TextBlock} in a part.
 * When the part is selected and the user clicks on the {@link TextBlock}
 * or invokes the {@link CommandHandler#editTextBlock} command,
 * this tool is started and it uses an HTMLTextArea to perform in-place text editing.
 * (For more details see the description for {@link TextEditingTool#doActivate}.)
 *
 * The {@link TextBlock} is accessible as the {@link TextEditingTool#textBlock} property.
 * The text editor is accessible as the {@link TextEditingTool#currentTextEditor} property.
 * From the text editor control one can access the {@link TextBlock} being edited via the
 * 'textEditingTool' property to get to this tool, from which one can use
 * the {@link TextEditingTool#textBlock} property.
 *
 * You can disable mouse clicking from starting this text editing tool
 * by setting {@link Tool#isEnabled} to false.
 * You can disable the F2 key from starting this text editing tool
 * by making sure {@link Part#canEdit} returns false,
 * by either setting {@link Diagram#allowTextEdit} to false
 * or by setting {@link Part#textEditable} to false.
 *
 * If you want to programmatically start the user editing a particular TextBlock,
 * call {@link CommandHandler#editTextBlock}.
 * That command method is also invoked by the F2 key on the keyboard.
 * <p class="boxread">
 * For a general discussion of text editing validation, see: <a href="../../intro/validation.html#TextEditingValidation">Introduction to Text Validation</a>.
 * For customizing the TextEditingTool, read about {@link HTMLInfo} and see <a href="../../intro/HTMLInteraction.html#TextEditors">Introduction to Text Editors</a>.
 * @extends Tool
 * @unrestricted
 * @category Tool
 */
export class TextEditingTool extends Tool {
  /** @hidden @internal */ public static className = 'TextEditingTool';
  /** @hidden @internal */ private _temporaryTextBlock: TextBlock;
  /** @hidden @internal */ private _textBlock: TextBlock;
  /** @hidden @internal */ private _starting: EnumValue;
  /** @hidden @internal */ private _textValidation: ((aTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null;
  /** @hidden @internal */ private _state: EnumValue;
  /** @hidden @internal */ private _minimumEditorScale: number;
  /** @hidden @internal */ private _selectsTextOnActivate: boolean;
  /** @hidden @internal */ private _currentTextEditor: HTMLInfo | null;
  /** @hidden @internal */ private _defaultTextEditor: HTMLInfo;
  /** @hidden @internal */ private _defaultTextEditorInput: HTMLElement;

  /**
   * You do not normally need to create an instance of this tool
   * because one already exists as the {@link ToolManager#textEditingTool}, which you can modify.
   *
   * The {@link Tool#name} of this tool is "TextEditing".
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(TextEditingTool);
    super();

    this.name = 'TextEditing';

    this._temporaryTextBlock = new TextBlock();
    this._textBlock = null;
    this._starting = TextEditingTool.SingleClickSelected;
    this._textValidation = null;

    // internal no-property holder of TextEditingState
    this._state = TextEditingTool.StateNone;

    this._minimumEditorScale = 1;
    this._selectsTextOnActivate = true;

    this._currentTextEditor = null;
    this._defaultTextEditor = new HTMLInfo();

    // Internal only:
    this._defaultTextEditorInput = null;
    this.setupDefaultEditor(this._defaultTextEditor);
  }

  /**
   * A possible value for {@link TextEditingTool#acceptText}, the text editing control has lost focus.
   * @constant
   */
  public static LostFocus = new EnumValue(TextEditingTool, 'LostFocus', 0);

  /**
   * A possible value for {@link TextEditingTool#acceptText}, the user has clicked somewhere else in the diagram.
   * @constant
   */
  public static MouseDown = new EnumValue(TextEditingTool, 'MouseDown', 1);

  /**
   * A possible value for {@link TextEditingTool#acceptText}, the user has typed TAB.
   * @constant
   */
  public static Tab = new EnumValue(TextEditingTool, 'Tab', 2);

  /**
   * A possible value for {@link TextEditingTool#acceptText}, the user has typed ENTER.
   * @constant
   */
  public static Enter = new EnumValue(TextEditingTool, 'Enter', 3);

  /**
   * A possible value for {@link TextEditingTool#starting},
   * A single click on a {@link TextBlock} with {@link TextBlock#editable} property set to true
   * will start in-place editing.
   * @constant
   */
  public static SingleClick = new EnumValue(TextEditingTool, 'SingleClick', 0);

  /**
   * A possible value for {@link TextEditingTool#starting},
   * A single click on a {@link TextBlock} with {@link TextBlock#editable} property set to true
   * will start in-place editing, but only if the {@link Part} that the {@link TextBlock} is in is already selected.
   * @constant
   */
  public static SingleClickSelected = new EnumValue(TextEditingTool, 'SingleClickSelected', 1);

  /**
   * A possible value for {@link TextEditingTool#starting},
   * A double click on a {@link TextBlock} with {@link TextBlock#editable} property set to true
   * will start in-place editing.
   * @constant
   */
  public static DoubleClick = new EnumValue(TextEditingTool, 'DoubleClick', 2);

  /**
   * A possible value for {@link #state},
   * this is the starting value before the tool is activated.
   * @constant
   */
  public static StateNone = new EnumValue(TextEditingTool, 'StateNone', 0);

  /**
   * A possible value for {@link #state},
   * {@link TextEditingTool#doActivate} sets the value of {@link #state} to this.
   * @constant
   */
  public static StateActive = new EnumValue(TextEditingTool, 'StateActive', 1);

  /**
   * A possible value for {@link #state},
   * The default text editor receiving focus sets the value of {@link #state} to this.
   * @constant
   */
  public static StateEditing = new EnumValue(TextEditingTool, 'StateEditing', 2);

  /**
   * A possible value for {@link #state},
   * TextEditingTool.acceptText sets the value of {@link #state} to this.
   * @constant
   */
  public static StateValidating = new EnumValue(TextEditingTool, 'StateValidating', 3);

  /**
   * A possible value for {@link #state},
   * If validation fails, TextEditingTool.acceptText sets the value of {@link #state} to this.
   * @constant
   */
  public static StateInvalid = new EnumValue(TextEditingTool, 'StateInvalid', 4);

  /**
   * A possible value for {@link #state},
   * Once accepted and the tool begins the "TextEditing" transaction, the value of {@link #state} is set to this.
   * @constant
   */
  public static StateValidated = new EnumValue(TextEditingTool, 'StateValidated', 5);

  /**
   * @hidden @internal
   * @this {TextEditingTool}
   * @param {HTMLInfo} htmlinfo
   */
  private setupDefaultEditor(htmlinfo: HTMLInfo): void {
    if (!Diagram.isUsingDOM()) return; // DOM-less environments don't have text editors
    const textarea = Util.createElement('textarea') as HTMLTextAreaElement;
    this._defaultTextEditorInput = textarea;
    const tool = this;

    textarea.addEventListener('input', function (e: Event) {
      if (tool.textBlock === null) return;
      const tempText = tool.measureTemporaryTextBlock(this.value);
      const scale = this.textScale;
      this.style.width = 20 + tempText.measuredBounds.width * scale + 'px';
      this.rows = tempText.lineCount;
    }, false);

    textarea.addEventListener('keydown', function (e: KeyboardEvent) {
      if (tool.textBlock === null) return;
      const keynum = e.which;
      if (keynum === 13) { // Enter
        if (tool.textBlock.isMultiline === false) e.preventDefault();
        tool.acceptText(TextEditingTool.Enter);
        return;
      } else if (keynum === 9) { // Tab
        tool.acceptText(TextEditingTool.Tab);
        e.preventDefault();
        return;
      } else if (keynum === 27) { // Esc
        tool.doCancel();
        if (tool.diagram !== null) tool.diagram.doFocus();
      }
    }, false);

    textarea.addEventListener('focus', function (e: Event) {
      tool.doFocusInternal(e);
    }, false);

    // Disallow blur.
    // If the textEditingTool blurs and the text is not valid,
    // we do not want focus taken off the element just because a user clicked elsewhere.
    textarea.addEventListener('blur', function (e: Event) {
      tool.doBlurInternal(e);
    }, false);

    /** @hidden @internal */
    htmlinfo.valueFunction = function () { return textarea.value; };

    htmlinfo.mainElement = textarea; // a useful reference for customizers

    // used to be in doActivate
    /**
     * @hidden @internal
     * @param {GraphObject} textblock
     * @param {Diagram} diagram
     * @param {Tool} tool2
     */
    htmlinfo.show = function (textblock: GraphObject, diagram: Diagram, tool2: Tool) {
      if (!(textblock instanceof TextBlock && tool2 instanceof TextEditingTool)) return;

      // This is called during validation, if validation failed:
      if (tool2.state === TextEditingTool.StateInvalid) {
        textarea.style.border = '3px solid red';
        textarea.focus();
        return;
      }

      const loc = textblock.getDocumentPoint(Spot.Center);
      const pos = diagram.position;
      const sc = diagram.scale;
      let textscale = textblock.getDocumentScale() * sc;
      if (textscale < tool2.minimumEditorScale) textscale = tool2.minimumEditorScale;
      // Add slightly more width/height to stop scrollbars and line wrapping on some browsers
      // +6 is firefox minimum, otherwise lines will be wrapped improperly
      const textwidth = (textblock.naturalBounds.width * textscale) + 6;
      const textheight = (textblock.naturalBounds.height * textscale) + 2;
      const left = (loc.x - pos.x) * sc;
      const top = (loc.y - pos.y) * sc;

      textarea.value = textblock.text;
      // the only way you can mix font and fontSize is if the font inherits and the fontSize overrides
      // in the future maybe have textarea contained in its own div
      diagram.div.style['font'] = textblock.font;

      const paddingsize = 1;
      textarea.style['position'] = 'absolute';
      textarea.style['zIndex'] = '100';
      textarea.style['font'] = 'inherit';
      textarea.style['fontSize'] = (textscale * 100) + '%';
      textarea.style['lineHeight'] = 'normal';
      textarea.style['width'] = (textwidth) + 'px';
      textarea.style['left'] = ((left - (textwidth / 2) | 0) - paddingsize) + 'px';
      textarea.style['top'] = ((top - (textheight / 2) | 0) - paddingsize) + 'px';
      textarea.style['textAlign'] = textblock.textAlign;
      textarea.style['margin'] = '0';
      textarea.style['padding'] = paddingsize + 'px';
      textarea.style['border'] = '0';
      textarea.style['outline'] = 'none';
      textarea.style['whiteSpace'] = 'pre-wrap';
      textarea.style['overflow'] = 'hidden'; // for proper IE wrap
      textarea.rows = textblock.lineCount;
      textarea.textScale = textscale;
      textarea.className = 'goTXarea';

      // Show:
      diagram.div.appendChild(textarea);

      // After adding, focus:
      textarea['focus']();
      if (tool2.selectsTextOnActivate) {
        textarea['select']();
        textarea['setSelectionRange'](0, 9999);
      }
    };

    /**
     * @hidden @internal
     * @param {Diagram} diagram
     * @param {Tool} tool2
     */
    htmlinfo.hide = function (diagram: Diagram, tool2: Tool) {
      diagram.div.removeChild(textarea);
    };
  }

  /**
   * Gets or sets the {@link TextBlock} that is being edited.
   * This property is initially null and is set in {@link TextEditingTool#doActivate}
   * as the {@link TextBlock} at the mouse click point.
   * However, if you set this property beforehand, {@link TextEditingTool#doActivate}
   * will not set it, and this tool will edit the given {@link TextBlock}.
   */
  /** @export */ get textBlock(): TextBlock | null { return this._textBlock; }
  /** @export */ set textBlock(value: TextBlock | null) {
    if (value !== null) Util.checkClass(value, TextBlock, TextEditingTool, 'textBlock');
    this._textBlock = value;
  }

  /**
   * Gets or sets the {@link HTMLInfo} that is editing the text.
   */
  /** @export */ get currentTextEditor(): HTMLInfo | null { return this._currentTextEditor; }
  /** @export */ set currentTextEditor(value: HTMLInfo | null) { this._currentTextEditor = value; }

  /**
   * Gets or sets the default {@link HTMLInfo} that edits the text.
   *
   * When {@link TextEditingTool#doActivate} is called, the {@link #currentTextEditor} is set to this value by default.
   * If a {@link TextBlock#textEditor} is specified on the TextBlock, that editor is used instead.
   *
   * This tool will call {@link HTMLInfo#show} during {@link #doActivate},
   * and {@link HTMLInfo#hide} during {@link #doDeactivate}.
   *
   * By default the value is an {@link HTMLInfo}, and the {@link HTMLInfo#mainElement} is an `HTMLTextArea`.
   * You can see the default implementation details <a href="../../extensions/textEditor.html">here</a>.
   *
   * For typical operation, HTMLInfo implementations should have a way of calling {@link TextEditingTool#acceptText}.
   */
  /** @export */ get defaultTextEditor(): HTMLInfo { return this._defaultTextEditor; }
  /** @export */ set defaultTextEditor(value: HTMLInfo) {
    if (Debug && !(value instanceof HTMLInfo)) {
      Util.throwError('TextEditingTool.defaultTextEditor must be an HTMLInfo.');
    }
    this._defaultTextEditor = value;
  }

  /**
   * Gets or sets how user gestures can start in-place editing of text.
   *
   * Possible values are {@link TextEditingTool.SingleClickSelected}, {@link TextEditingTool.SingleClick}, and {@link TextEditingTool.DoubleClick}.
   *
   * The default is {@link TextEditingTool.SingleClickSelected}
   */
  /** @export */ get starting(): EnumValue { return this._starting; }
  /** @export */ set starting(value: EnumValue) {
    EnumValue.checkEnumValue(value, TextEditingTool, TextEditingTool, 'starting');
    this._starting = value;
  }

  /**
   * This may run when there is a mouse-click on a {@link TextBlock} for which the
   * {@link TextBlock#editable} property is true in a {@link Part}
   * that {@link Part#isSelected}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TextEditingTool}
   * @return {boolean}
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;
    const diagram = this.diagram;
    if (diagram === null || diagram.isReadOnly) return false;
    // only works with the left button
    if (!diagram.lastInput.left) return false;

    // the mouse down point needs to be near the mouse up point
    if (this.isBeyondDragSize()) return false;

    const p = diagram.lastInput.documentPoint;
    const tb = diagram.findObjectAt(p);
    if (tb === null || !(tb instanceof TextBlock)) return false;
    if (!tb.editable || !tb.part.canEdit()) return false;

    const part = tb.part;
    if (part === null) return false;
    if (this.starting === TextEditingTool.SingleClickSelected && !part.isSelected) return false;
    if (this.starting === TextEditingTool.DoubleClick && diagram.lastInput.clickCount < 2) return false;

    return true;
  }

  /**
   * This calls {@link TextEditingTool#doActivate} if there is a
   * {@link #textBlock} set. {@link #doActivate} attempts to set {@link #textBlock} if it is null.
   * @this {TextEditingTool}
   */
  public doStart(): void {
    if (!this.isActive && this.textBlock !== null) {
      this.doActivate();
    }
  }

  /**
   * Start editing the text for a {@link #textBlock}.
   *
   * If {@link TextEditingTool#textBlock} is not already specified, this looks for one at the current
   * mouse point. If none is found, this method does nothing.
   *
   * This method sets {@link #currentTextEditor}.
   * If {@link TextBlock#textEditor} is defined on the {@link TextBlock} it will use that as the value.
   * By default, it uses the value of {@link #defaultTextEditor}, which is an {@link HTMLInfo}
   * showing an HTML textarea, with the CSS class `goTXarea`.
   *
   * If the {@link #currentTextEditor} is an {@link HTMLInfo}, this method calls {@link HTMLInfo#show} on that instance.
   *
   * This sets {@link Tool#isActive} to true.
   * Custom text editors should call {@link TextEditingTool#acceptText} to finish the edit
   * by modifying the {@link TextBlock} and committing the edit transaction.
   * Or call {@link TextEditingTool#doCancel} to abort the edit.
   * @this {TextEditingTool}
   */
  public doActivate(): void {
    if (this.isActive) return;
    const diagram = this.diagram;
    if (diagram === null) return;
    let textblock = this.textBlock as any;

    // there should be a TextBlock; find it if not given
    if (textblock === null) {
      textblock = diagram.findObjectAt(diagram.lastInput.documentPoint);
    }
    if (textblock === null || !(textblock instanceof TextBlock)) return;

    this.textBlock = textblock;
    // not needed if activated the normal way because the tool checks beforehand ?
    // if (!textblock.editable) return;  //?? able to operate on all TextBlocks
    const part = textblock.part;
    if (part === null) return;  // ?? don't check for IsSelected here, just in CanStart?

    // start editing
    this.isActive = true;
    this._state = TextEditingTool.StateActive;

    // make sure there's a TextEditor
    let editor = this.defaultTextEditor;
    if (textblock.textEditor !== null) editor = textblock.textEditor;

    // updates the size of the textarea
    this._temporaryTextBlock = this.textBlock.copy();

    // Make sure the TextBlock (and editor) are visible in the viewport
    // If the editor itself is not visible, rely on the browser to scroll it
    const docb = new Rect(this.textBlock.getDocumentPoint(Spot.TopLeft),
      this.textBlock.getDocumentPoint(Spot.BottomRight));
    diagram.scrollToRect(docb);

    // Additional setup done every time for the default editor only
    editor.show(textblock, diagram, this);

    this.currentTextEditor = editor;
  }

  /**
   * Abort any text editing operation.
   * @this {Tool}
   */
  public doCancel(): void {
    this.stopTool();
  }

  /**
   * A click (mouse up) calls {@link TextEditingTool#doActivate} if this tool is not already active
   * and if {@link TextEditingTool#canStart} returns true.
   * @this {TextEditingTool}
   */
  public doMouseUp(): void {
    if (!this.isActive && this.canStart()) {
      this.doActivate();
    }
  }

  /**
   * This calls {@link #acceptText} with the reason {@link TextEditingTool.MouseDown},
   * if this tool {@link Tool#isActive}.
   * @this {TextEditingTool}
   */
  public doMouseDown(): void {
    if (this.isActive) {
      this.acceptText(TextEditingTool.MouseDown);
    }
  }

  /**
   * Finish editing by trying to accept the new text.
   *
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TextEditingTool}
   * @param {EnumValue} reason The reason must be either {@link TextEditingTool.LostFocus},
   * {@link TextEditingTool.MouseDown}, {@link TextEditingTool.Tab}, or {@link TextEditingTool.Enter}.
   */
  public acceptText(reason: EnumValue): void {
    switch (reason) {
      case TextEditingTool.MouseDown:
        if (this._state === TextEditingTool.StateValidated) {
          if (this.currentTextEditor instanceof HTMLElement) this.currentTextEditor.focus();
        } else if (this._state === TextEditingTool.StateActive ||
          this._state === TextEditingTool.StateInvalid ||
          this._state === TextEditingTool.StateEditing) {
          this._state = TextEditingTool.StateValidating;
          this.doAcceptText();
        }
        break;
      case TextEditingTool.LostFocus:
      case TextEditingTool.Enter:
      case TextEditingTool.Tab:
        if (TextEditingTool.Enter === reason && this.textBlock.isMultiline === true) return;
        if (this._state === TextEditingTool.StateActive ||
          this._state === TextEditingTool.StateInvalid ||
          this._state === TextEditingTool.StateEditing) {
          this._state = TextEditingTool.StateValidating;
          this.doAcceptText();
        }
        break;
    }
  }

  // Call textValidation
  /**
   * @hidden @internal
   * @this {TextEditingTool}
   */
  private doAcceptText(): void {
    const tb = this.textBlock;
    const diagram = this.diagram;
    const editor = this.currentTextEditor;
    if (tb !== null && editor !== null) {
      const oldstring = tb.text;
      let newstring = '';
      if (editor.valueFunction !== null) newstring = editor.valueFunction();

      if (!this.isValidText(tb, oldstring, newstring)) {
        this._state = TextEditingTool.StateInvalid;
        this.doError(oldstring, newstring);
        // call show, with the knowledge that this.state changed
        editor.show(tb, diagram, this);
        return;
      }

      this.startTransaction(this.name);
      this._state = TextEditingTool.StateValidated;

      // set the TransactionResult before raising events, in case they change the result or cancel the tool
      this.transactionResult = this.name;
      // update the TextBlock.text
      tb.text = newstring;
      // raise events
      this.doSuccess(oldstring, newstring);
      if (diagram !== null) diagram.raiseDiagramEvent('TextEdited', tb, oldstring);
      this.stopTransaction();

      // all done with this tool
      this.stopTool();
      if (diagram !== null) diagram.doFocus();
    }
  }

  // Since 2.1
  /**
   * @hidden
   * Call the {@link #textBlock}'s {@link TextBlock#errorFunction}, if there is one.
   * This is called only when the {@link #isValidText} method returned false.
   * The value of {@link #state} will be StateInvalid.
   * When this method returns, the text editor will be shown again.
   * @param oldstring
   * @param newstring
   */
  public doError(oldstring: string, newstring: string) {
    const tb = this.textBlock;
    if (tb.errorFunction !== null) tb.errorFunction(this, oldstring, newstring);
  }

  // Since 2.1
  /**
   * @hidden
   * Call the {@link #textBlock}'s {@link TextBlock#textEdited} event handler, if there is one.
   * This is called just after the {@link TextBlock.text} has been set to the new string value.
   * When this method returns, this tool raises the "TextEdited" {@link DiagramEvent}
   * and commits the transaction.
   * @param oldstring
   * @param newstring
   */
  public doSuccess(oldstring: string, newstring: string) {
    const tb = this.textBlock;
    if (tb.textEdited !== null) tb.textEdited(tb, oldstring, newstring);
  }

  /**
   * Release the mouse.
   *
   * If the {@link #currentTextEditor} is an {@link HTMLInfo}, this calls {@link HTMLInfo#hide}.
   * @this {TextEditingTool}
   */
  public doDeactivate(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    this._state = TextEditingTool.StateNone;

    this.textBlock = null;
    if (this.currentTextEditor !== null) {
      const input = this.currentTextEditor;
      input.hide(diagram, this);
    }

    this.isActive = false;
  }

  /**
   * @hidden @internal
   * @this {TextEditingTool}
   * @param {Event} e
   */
  private doFocusInternal(e: Event): void {
    if (this.currentTextEditor === null || this.state === TextEditingTool.StateNone) return;
    const textarea = this._defaultTextEditorInput;

    // handle focus:
    if (this._state === TextEditingTool.StateActive) {
      this._state = TextEditingTool.StateEditing;
    }

    if (typeof textarea['select'] === 'function' && this.selectsTextOnActivate) {
      textarea['select']();
      textarea['setSelectionRange'](0, 9999);
    }
  }

  /**
   * @hidden @internal
   * @this {TextEditingTool}
   * @param {Event} e
   */
  private doBlurInternal(e: Event): void {
    if (this.currentTextEditor === null || this.state === TextEditingTool.StateNone) return;
    const textarea = this._defaultTextEditorInput;

    if (typeof textarea['focus'] === 'function') textarea['focus']();

    if (typeof textarea['select'] === 'function' && this.selectsTextOnActivate) {
      textarea['select']();
      textarea['setSelectionRange'](0, 9999);
    }
  }

  /**
   * This predicate checks any {@link TextBlock#textValidation} predicate and
   * this tool's {@link #textValidation} predicate to make sure the {@link TextBlock#text}
   * property may be set to the new string.
   *
   * This method may be overridden, although usually it is sufficient to set {@link #textValidation}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TextEditingTool}
   * @param {TextBlock} textblock the {@link TextBlock} that is being edited.
   * @param {string} oldstr the previous string value.
   * @param {string} newstr the proposed new string value.
   * @return {boolean} true if the new string is valid for the given TextBlock.
   */
  public isValidText(textblock: TextBlock, oldstr: string, newstr: string): boolean {
    Util.checkClass(textblock, TextBlock, TextEditingTool, 'isValidText:textblock');
    // Check both this.textValidation and textBlock.textValidation
    const func = this.textValidation;
    if (func !== null && !func(textblock, oldstr, newstr)) return false;
    const textfunc = textblock.textValidation;
    if (textfunc !== null && !textfunc(textblock, oldstr, newstr)) return false;
    return true;
  }

  /**
   * Gets or sets the predicate that determines whether or not a string of text is valid.
   * If this is non-null, this predicate is called in addition to any {@link TextBlock#textValidation} predicate.
   * See {@link #isValidText} for more details.
   * The default predicate is null, which is equivalent to simply returning true.
   *
   * The function, if supplied, must not have any side-effects.
   */
  /** @export */ get textValidation(): ((aTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null { return this._textValidation; }
  /** @export */ set textValidation(value: ((aTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null) {
    if (value !== null) Util.checkPrimitive(value, 'function', TextEditingTool, 'textValidation');
    this._textValidation = value;
  }

  /**
   * @hidden @internal
   * Gets or sets the minimum scale that the TextEditingTool's default editor will use when constructing itself.
   * This is useful for stopping the editor from becoming too small when the diagram is zoomed out and editing is activated.
   * The default is 1.
   * @since 1.5
   */
  /** @export */ get minimumEditorScale(): number { return this._minimumEditorScale; }
  /** @export */ set minimumEditorScale(value: number) {
    if (value !== null) Util.checkPrimitive(value, 'number', TextEditingTool, 'minimumEditorScale');
    this._minimumEditorScale = value;
  }

  /**
   * Gets or sets whether to select (highlight) the editable text when the TextEditingTool is activated.
   * The default is true.
   * @since 1.5
   */
  /** @export */ get selectsTextOnActivate(): boolean { return this._selectsTextOnActivate; }
  /** @export */ set selectsTextOnActivate(value: boolean) {
    if (value !== null) Util.checkPrimitive(value, 'boolean', TextEditingTool, 'selectsTextOnActivate');
    this._selectsTextOnActivate = value;
  }

  /**
   * Gets or sets the state of the TextEditingTool.
   * The only accepted values are listed as constant properties of TextEditingTool, including:
   *   - {@link TextEditingTool.StateNone}
   *   - {@link TextEditingTool.StateActive}
   *   - {@link TextEditingTool.StateEditing}
   *   - {@link TextEditingTool.StateValidating}
   *   - {@link TextEditingTool.StateValidated}
   *
   * The starting value value is {@link TextEditingTool.StateNone},
   * {@link TextEditingTool#doActivate} sets the value to {@link TextEditingTool.StateActive}.
   * The default text editor receiving focus sets the value to {@link TextEditingTool.StateEditing}.
   * {@link TextEditingTool#acceptText} sets the value to {@link TextEditingTool.StateValidating}.
   * Once accepted and the tool begins the "TextEditing" transaction, the value is set to {@link TextEditingTool.StateValidated}.
   * @since 1.7
   */
  /** @export */ get state(): EnumValue { return this._state; }
  /** @export */ set state(value: EnumValue) {
    const old = this._state;
    if (old !== value) {
      EnumValue.checkEnumValue(value, TextEditingTool, TextEditingTool, 'starting');
      this._state = value;
    }
  }

  /**
   * This method returns a temporary {@link TextBlock} used for measuring text during editing.
   * The TextBlock.text is set to the parameter's value,
   * and the TextBlock is measured with the last available width of the {@link #textBlock}.
   *
   * Text editors can use the {@link GraphObject#measuredBounds} and {@link TextBlock#lineCount}
   * to determine a reasonable size for their text areas.
   * @export
   * @this {TextEditingTool}
   * @param {string} text the text to measure
   * @return {TextBlock}
   * @since 1.7
   */
  public measureTemporaryTextBlock(text: string): TextBlock {
    const tempText = this._temporaryTextBlock;
    tempText.text = text;
    tempText.measure(this.textBlock._lastAvailableWidth, Infinity);  // ??? maintain aspect ratio
    return tempText;
  }

} // end class
TextEditingTool['className'] = 'TextEditingTool';

TextEditingTool['LostFocus'] = TextEditingTool.LostFocus;
TextEditingTool['MouseDown'] = TextEditingTool.MouseDown;
TextEditingTool['Tab'] = TextEditingTool.Tab;
TextEditingTool['Enter'] = TextEditingTool.Enter;
TextEditingTool['SingleClick'] = TextEditingTool.SingleClick;
TextEditingTool['SingleClickSelected'] = TextEditingTool.SingleClickSelected;
TextEditingTool['DoubleClick'] = TextEditingTool.DoubleClick;
TextEditingTool['StateNone'] = TextEditingTool.StateNone;
TextEditingTool['StateActive'] = TextEditingTool.StateActive;
TextEditingTool['StateEditing'] = TextEditingTool.StateEditing;
TextEditingTool['StateValidating'] = TextEditingTool.StateValidating;
TextEditingTool['StateInvalid'] = TextEditingTool.StateInvalid;
TextEditingTool['StateValidated'] = TextEditingTool.StateValidated;


// import { ToolManager } from './toolManager';

Util.defineProperty(ToolManager, 'textEditingTool',
  /** @this {ToolManager} */
  function(this: ToolManager): TextEditingTool {
    return this.findTool('TextEditing') as TextEditingTool;
  },
  /** @this {ToolManager} */
  function(this: ToolManager, value: TextEditingTool) { this.replaceStandardTool('TextEditing', value, this.mouseUpTools); }
);

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
AnimationManager
*/


// import { EnumValue } from './enumValue';

/**
 * An EasingFunction describes how much to modify a scalar value based on
 * the current time, the start value, the change in value, and the duration.
 * @category Type
 */
export type EasingFunction = ((currentTime: number, startValue: number, byValue: number, duration: number) => number);

/**
 * AnimationManager handles animations in a {@link Diagram}. Each Diagram has one, {@link Diagram#animationManager}.
 * Setting the Model, performing a Layout, Group expansion and Tree expansion automatically start animations through
 * the {@link #defaultAnimation}. Animations can be manually started by creating {@link Animation}s, which are associated with an
 * AnimationManager.
 *
 * Animation is enabled by default, setting the {@link #isEnabled} property to false will turn off animations for a Diagram.
 *
 * When the {@link #defaultAnimation} begins it raises the `"AnimationStarting"` Diagram event,
 * upon completion it raises the `"AnimationFinished"` Diagram event.
 *
 * The {@link #defaultAnimation}, if running, will stop if a new transaction is started, if an undo or redo is called,
 * if a layout is invalidated, or if a model is replaced.
 * When an {@link Animation} is stopped, the Diagram immediately finishes the animation and draws the final state.
 * Animations can be stopped programatically with the methods {@link AnimationManager#stopAnimation} or {@link Animation#stop}.
 * @since 1.4
 * @unrestricted
 */
export class AnimationManager {
  /** @hidden @internal */ public static className = 'AnimationManager';
  /** @hidden @internal */ private _diagram: Diagram;

  /** @hidden @internal */ public _isAnimating: boolean;
  /** @hidden @internal */ private _isEnabled: boolean;
  /** @hidden @internal */ private _isInitial: boolean;
  /** @hidden @internal */ public _animationReasons: Set<string>;
  /** @hidden @internal */ private _duration: number;
  /** @hidden @internal */ private _undoskip: boolean;
  /** @hidden @internal */ private _bindskip: boolean;
  /** @hidden @internal */ private _boundsskip: EnumValue;
  /** @hidden @internal */ private _isTicking: boolean;
  /** @hidden @internal */ public _pendingAnimation: boolean;
  /** @hidden @internal */ public _pendingAutomaticAnimation: boolean;
  /** @hidden @internal */ public _initialAnimationOngoing: boolean;

  /** @hidden @internal */ private _defaultAnimation: Animation;
  /** @hidden @internal */ private _activeAnimations: Set<Animation>;
  /** @hidden @internal */ private _toAddSoon: Set<Animation>;
  /** @hidden @internal */ public _useInitialPositioning: boolean;

  /**
   * You do not normally need to create an instance of this class because one already exists as the {@link Diagram#animationManager}, which you can modify.
   */
  constructor() {
    if (!AnimationManager._hasInit) {
      AnimationManager.initStatic();
      AnimationManager._hasInit = true;
    }

    this._diagram = Diagram.getLastDiagram();  // replaced by setDiagram, called by Diagram
    this._isAnimating = false;
    this._isTicking = false;
    this._pendingAnimation = false;
    this._pendingAutomaticAnimation = false;
    this._initialAnimationOngoing = false;
    this._useInitialPositioning = true;

    this._isEnabled = true;
    this._isInitial = true;
    this._duration = 600;

    // temporary diagram state that is saved/restored
    this._undoskip = false;
    this._bindskip = false;
    this._boundsskip = Diagram.DocumentScroll;

    this._animationReasons = new Set<string>();
    this._defaultAnimation = new Animation();
    this._defaultAnimation._animationManager = this; // necessary for addToAnimation?
    this._activeAnimations = new Set<Animation>();
    this._toAddSoon = new Set<Animation>();
  }

  /**
   * @hidden @internal
   */
  public setDiagram(val: Diagram): void { this._diagram = val; }

  /**
   * @hidden @internal
   */
  public getDiagram(): Diagram { return this._diagram; }

  // easing args are: currentTime, startValue, byValue (change in value), duration
  // There are many online examples of easing functions, such as https://easings.net/en, and http://gizma.com/easing/


  /**
   * This read-only property returns a Set of reasons the default animation may start.
   * This set can be queried in {@link #canStart} to turn off specific default animations.
   *
   * These are the possible reasons GoJS will begin an animation:
   * ```md
   *   Called by CommandHandler:
   *     "Collapse SubGraph"
   *     "Expand SubGraph"
   *     "Collapse Tree"
   *     "Expand Tree"
   *     "Scroll To Part"
   *     "Zoom To Fit"
   *   Called by Diagram:
   *     "Model"
   *     "Layout"
   *   Called by AnimationTriggers:
   *     "Trigger"
   * ```
   * @return {Set.<string>}
   * @see {@link #canStart}
   * @since 2.1
   */
  /** @export */ get animationReasons(): Set<string> { return this._animationReasons; }

  /**
   * This method is passed the reason the animation is to begin,
   * and must return true or false based on whether or not the animation is to be allowed.
   * Returning true means the animation will occur, returning false will stop the animation's setup.
   *
   * By default, this method always returns true.
   *
   * The reasons GoJS will begin an animation are collected in the set {@link #animationReasons},
   * and can be queried in this method to conditionally allow animations, for instance:
   * ```js
   * // disallow expand/collapse animations, but allow all others
   * myDiagram.animationManager.canStart = function(reason) {
   *   if (this.animationReasons.contains("Expand Tree")) return false;
   *   return true;
   * }
   * ```
   * @this {AnimationManager}
   * @param {string} reason Reason for starting the animation
   * @return {boolean}
   * @see {@link #animationReasons}
   * @since 2.1
   */
  public canStart(reason: string): boolean {
    return true;
  }


  //
  //  Typical usage:
  //
  //  ```js
  //  myDiagram.startTransaction();
  //  myDiagram.animationManager.animateTransaction();
  //  node1.position = new go.Point(newValueX, newValueY);
  //  // On committing, node1 will animate from its current position to the new one:
  //  myDiagram.commitTransaction();
  //
  //  ```
  //  param {Array<string>=} graphObjectProperties An array of GraphObject properties to animate. By default, `"position"` is animated.
  //  Possible properties are `"position"`, `"scale"`, `"opacity"`, `"angle"`, and those defined with {@link AnimationManager.defineAnimationEffect}.
  //  You can animate {@link GraphObject#name}d shapes by passing in their name and the property, separated with a period: `"myGraphObject.angle"`
  //  param {Array<string>=} diagramProperties An array of Diagram to animate. By default, these properties are `"position"` and `"scale"`.
  //
  //  More properties can be supported by defining them with {@link AnimationManager.defineAnimationEffect}.

  // *
  //  * Activates the {@link #defaultAnimation}. During a transaction, this begins listening for changes to animate at the end of the transaction.
  //  * By default, this listens for {@link Diagram#position}, {@link Diagram#scale}, and Part positions.
  //  * This should be called after {@link Diagram#startTransaction}. The animation begins when {@link Diagram#commitTransaction} is called.
  //  *
  //  * Calling this method stops any currently running {@link #defaultAnimation}.
  //  *
  //  * since 2.1
  //  *
  // public animateTransaction(graphObjectProperties?: Array<string>, diagramProperties?: Array<string>): void {
  //   this.prepareAutomaticAnimation('Custom');
  // }

  /**
   * @hidden @internal
   * Begins listening for changes to {@link Diagram#position}, {@link Diagram#scale},
   * and Part positions, and animates those automatically in the near-future.
   *
   * This is called after {@link Diagram#startTransaction} and the animation begins when {@link Diagram#commitTransaction} is called.
   *
   * Calling this method stops any currently running animation.
   *
   * @param {string} reason The reason the animation is to begin. This is added to a list of reasons, which informs whether or not the animation manager can start.
   */
  public prepareAutomaticAnimation(reason: string): void {
    if (this.prepareAnimation(reason)) {
      this._pendingAutomaticAnimation = true;
    }
  }

  /**
   * @hidden @internal
   * Begins listening for changes to {@link Diagram#position}, {@link Diagram#scale},
   * and Part positions, and animates those automatically in the near-future.
   *
   * This is called after {@link Diagram#startTransaction} and the animation begins when {@link Diagram#commitTransaction} is called.
   *
   * Calling this method stops any currently running animation.
   *
   * @param {string} reason The reason the animation is to begin. This is added to a list of reasons, which informs whether or not the animation manager can start.
   */
  public prepareAnimation(reason: string): boolean {
    if (!this._isEnabled) return false;
    // If we do this, then we cannot have animations that are started in onInitialLayoutCompleted
    // so we should avoid this and only check during beginAnimation
    // if (!this._isInitial && !this._diagram.isInitialLayoutCompleted) return false;
    if (!this.canStart(reason)) return false;
    this._animationReasons.add(reason);

    if (this.defaultAnimation.isAnimating) {
      this.stopAnimation();
    }
    this._pendingAnimation = true;
    return true;
  }


  /**
   * @hidden @internal
   * Called by Diagram in doModelChanged, or CommandHandler in zoomToFit
   * @this {AnimationManager}
   */
  public beginAnimation(): void {
    if (!this._isEnabled) return;
    if (!this._pendingAnimation) return;

    const da = this._defaultAnimation;
    const diagram = this._diagram;
    const firstAnimation = this._animationReasons.contains('Model');
    if (firstAnimation) {
      this._initialAnimationOngoing = true;
      // possibly here, fix diagram position ???
      // const states = (da._objectsToAnimate.get(diagram));
      // if (states !== null && states.start['position:diagram']) {
      //   console.log('first:');
      //   console.log(states.start['position:diagram'].x);
      //   console.log(states.end['position:diagram'].x);
      // }
    }
    // No-op if initial (from 'Model' reason), or no-op if nothing animating
    if ((firstAnimation && !this._isInitial) ||
      da._objectsToAnimate.count === 0) {
      this._animationReasons.clear();
      this._pendingAnimation = false;
      this._pendingAutomaticAnimation = false;
      da._objectsToAnimate.clear();
      da.clearState();
      this._initialAnimationOngoing = false;
      diagram.invalidateDraw();
      // // update links: ???
      // const links = this._diagram.links;
      // while (links.next()) {
      //   const link = links.value;
      //   link.restoreSavedRoute();
      // }
      // this._diagram.loadedRouteLinks.clear();
      return;
    }
    this._animationReasons.clear();

    // Cancel queued updates in case something else caused one in the meantime
    diagram._updateQueued = false;

    const self = this;
    root.requestAnimationFrame(function () {
      // maybe something cancelled it or started it in the meantime
      if (self._pendingAnimation === false || da._isAnimating) return;
      if (diagram.getRenderingHint('temporaryPixelRatio')) diagram.enableTempPixelRatio();
      diagram.maybeResize(); // what if the size has changed since pending animation started?
      self._pendingAnimation = false;
      self._pendingAutomaticAnimation = false;



      if (diagram._savedPosition.equals(diagram._position)) { // positions match, keep initial... positioning during animation
        self._useInitialPositioning = true;
        // Perhaps: Remove the position:diagram from: da._objectsToAnimate.get(diagram);
      } else { // position has changed, do not use "initial..." position options during the animation
        self._useInitialPositioning = false;
      }

      da.start();
      self.skipState(da);
      diagram.invalidateDocumentBounds(); // ?? necessary?
      da.modifyObjects(0 /*da._currentDuration*/); // ??? should this arg be zero? Does this cause animation flickering?
      // on the first frame its possible that stuff was modified
      // during pendingAnimation and invalid state may still be true, so update
      diagram.doUpdate(true);
      self.restoreState();

    });
  }

  /**
   * @hidden @internal
   * Called by Diagram and Part
   */
  public addToAnimation(obj: GraphObject, propname: string, start: any, end: any, cosmetic: boolean): void {
    // In default animations, only allow disconnected links to animate position
    if (obj instanceof Link && (obj.fromNode !== null || obj.toNode !== null))  {
      return;
    }
    this._defaultAnimation.add(obj, propname, start, end, cosmetic);
  }

  /**
   * @hidden @internal
   * Called by Part and Node isVisible
   */
  public objRemainsVisible(obj: GraphObject): boolean {
    return this._defaultAnimation.objRemainsVisible(obj);
  }

  /**
   * @hidden @internal
   * Called by Placeholder measure
   */
  public isAnimatingPosition(obj: GraphObject): boolean {
    return this._defaultAnimation.isAnimatingPosition(obj);
  }

  /**
   * @hidden @internal
   * Called by {@link Animation#start}.
   * @this {AnimationManager}
   */
  public animateObjects(animation: Animation): void {
    const animations = this._activeAnimations;
    const AM = this;
    if (this._defaultAnimation === animation) this._diagram.raiseDiagramEvent('AnimationStarting', this);
    // If this is the first animation, start stepping
    /**
     * @hidden @internal
     */
    function step() {
      // These may have been added during an animation,
      // but we do not want to modify the animations list while iterating over it
      if (AM._toAddSoon.count > 0) {
        animations.addAll(AM._toAddSoon);
        AM._toAddSoon.clear();
        AM._isAnimating = true;
      }
      if (AM._isAnimating === false) return; // for this to work "isAnimating" must mean "is there at least one animation ongoing"
      if (animations.count === 0) return;
      const complete = Util.tempArray<Animation>();
      const itr = animations.iterator;
      while (itr.next()) {
        const anim = itr.value;
        if (anim._isAnimating === false) {
          complete.push(anim);
          continue;
        }
        if (!anim.objectsHaveDiagram()) {
          // set this animation as inactive until the next transaction
          anim._isSuspended = true;
        } else {
          anim.tick(false); // For each (active) animation
        }
      }

      for (let i = 0; i < complete.length; i++) {
        AM.doCompletion(complete[i]);
      }
      Util.freeArray(complete);
      root.requestAnimationFrame(step);
    }
    if (!this._isAnimating) {
      this._isAnimating = true;
      animations.add(animation);
      root.requestAnimationFrame(function () { step(); });
    } else {
      this.addSoon(animation);
    }
  }


  /**
   * @hidden @internal
   * You cannot add an animation while iterating through the list, so this defers them
   * @this {AnimationManager}
   */
  private addSoon(animation: Animation): void {
    this._toAddSoon.add(animation);
  }

  /**
   * @hidden @internal
   * Called by Diagram, at the end of a transaction remove the animation suspended flags.
   * @this {AnimationManager}
   */
  public clearSuspendedAnimations(): void {
    const itr = this._activeAnimations.iterator;
    while (itr.next()) {
      itr.value._isSuspended = false;
    }
  }

  /**
   * @hidden @internal
   * Skip all diagram state so that animation changes do not end up in the undo manager or have other side effects
   * @this {AnimationManager}
   */
  public skipState(animation: Animation): void {
    if (this._isTicking) return;
    const diagram = this._diagram;
    this._undoskip = diagram.skipsUndoManager;
    this._bindskip = diagram.skipsModelSourceBindings;
    diagram.skipsUndoManager = true;
    diagram.skipsModelSourceBindings = true;
    if (animation.isViewportUnconstrained) {
      diagram.scrollMode = Diagram.InfiniteScroll;
      this._boundsskip = diagram.scrollMode;
    }
    this._isTicking = true;
  }

  /**
   * @hidden @internal
   * Called after skipState
   * @this {AnimationManager}
   */
  public restoreState(): void {
    const diagram = this._diagram;
    diagram.skipsUndoManager = this._undoskip;
    diagram.skipsModelSourceBindings = this._bindskip;
    diagram.scrollMode = this._boundsskip;
    this._isTicking = false;
  }

  /**
   * Stops the {@link #defaultAnimation} and updates the Diagram to its final state.
   *
   * If the argument is true, this stops all running animations.
   * If an Animation was about to begin, it will be cancelled.
   *
   * If the {@link AnimationManager#defaultAnimation} is running, this will raise the `"AnimationFinished"` Diagram event.
   * @export
   * @this {AnimationManager}
   * @param {boolean=} stopsAllAnimations Whether to stop all animations, instead of just the {@link #defaultAnimation}. Default false.
   */
  public stopAnimation(stopsAllAnimations?: boolean): void {
    // If an animation was about to begin, stop it
    const da = this._defaultAnimation;
    if (this._pendingAnimation === true) {
      this._pendingAnimation = false;
      this._pendingAutomaticAnimation = false;
      this._initialAnimationOngoing = false;
      this._animationReasons.clear();
      // pending animation might have stopped an update, update soon
      if (da.isAnythingAnimating()) this._diagram.requestUpdate();
    }

    // If no animation or not enabled, quit now
    if (!this._isAnimating || !this._isEnabled) {
      da._objectsToAnimate.clear();
      da.clearState();
      return;
    }

    da.doCompletion(true);
    da._objectsToAnimate.clear();
    da.clearState();
    // but maybe stop all animations?
    if (stopsAllAnimations === true) {
      const animations = this._activeAnimations.toArray();
      for (let i = 0; i < animations.length; i++) {
        animations[i].doCompletion(true);
      }
    }

  }

  /**
   * @hidden @internal
   */
  public recomputeOnCompletion(obj: GraphObject): void {
    //     5/13/19 seemingly not needed
    // this._defaultAnimation.recomputeOnCompletion(obj);
  }

  /**
   * @hidden @internal
   * This is called by Animation's doCompletion.
   *
   * This raises the `"AnimationFinished"` Diagram event. diagram.raiseDiagramEvent('AnimationFinished', this);
   * @this {AnimationManager}
   */
  public doCompletion(animation: Animation): void {
    if (animation === this.defaultAnimation) {
      this.defaultAnimation._objectsToAnimate.clear();
      this._diagram.raiseDiagramEvent('AnimationFinished', this.defaultAnimation);
    }
    this._activeAnimations.remove(animation);
    if (this._activeAnimations.count === 0) {
      this._isAnimating = false;
      this._diagram.requestUpdate();
    }
  }

  /**
   * @hidden @internal
   * Called only by Node's expandTree and Group's expandSubGraph
   * @this {AnimationManager}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addExpandedNode(node: Node, rootNode: Node): void {
    if (!this._pendingAutomaticAnimation) return;
    if (!this._animationReasons.contains('Expand Tree') && !this._animationReasons.contains('Expand SubGraph')) return;
    this._defaultAnimation.addExpandedNode(node, rootNode);
  }

  /**
   * @hidden @internal
   * Called only by Node's collapseTree and Group's collapseSubGraph
   * @this {AnimationManager}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addCollapsedNode(node: Node, rootNode: Node): void {
    if (!this._pendingAutomaticAnimation) return;
    if (!this._animationReasons.contains('Collapse Tree') && !this._animationReasons.contains('Collapse SubGraph')) return;
    this._defaultAnimation.addCollapsedNode(node, rootNode);
    this._defaultAnimation.add2(rootNode, 'position', rootNode.position, rootNode.position);
  }

  /**
   * @hidden @internal
   * Called by Diagram when the position changes during pending animation.
   * @this {AnimationManager}
   * @param {Point} oldpos
   * @param {Point} newpos
   */
  public addDiagramPos(oldpos: Point, newpos: Point): void {
    if (!this._pendingAutomaticAnimation || oldpos.equals(newpos) /*|| (!this._diagram.isInitialLayoutCompleted)*/) return;
    if ((!this._diagram.isInitialLayoutCompleted)) oldpos = newpos.copy(); // ??? lazy
    this._defaultAnimation.add2(this._diagram, 'position', oldpos, newpos);
  }

  /**
   * @hidden @internal
   * Called by Diagram when the position changes during pending animation.
   * @this {AnimationManager}
   * @param {number} oldscale
   * @param {number} newscale
   */
  public addDiagramScale(oldscale: number, newscale: number): void {
    if (!this._pendingAutomaticAnimation) return;
    this._defaultAnimation.add(this._diagram, 'scale', oldscale, newscale);
  }

  /**
   * Gets or sets whether this AnimationManager operates.
   *
   * The default value is true.
   * Setting this to false does not stop an animation, it only stops future animations.
   * To stop any ongoing animation, use {@link #stopAnimation}.
   * Setting this property does not raise any events.
   */
  /** @export */ get isEnabled(): boolean { return this._isEnabled; }
  /** @export */ set isEnabled(value: boolean) {
    Util.checkPrimitive(value, 'boolean', AnimationManager, 'isEnabled');
    this._isEnabled = value;
  }

  /**
   * Gets or sets the default duration, in milliseconds, used as the duration for animations that have their {@link Animation#duration} set to `NaN`.
   *
   * Typically these values are short. The default value is 600 milliseconds.
   * The value must be a number greater than or equal to 1.
   * Setting this property does not raise any events.
   *
   * @see {@link Animation#duration}
   */
  /** @export */ get duration(): number { return this._duration; }
  /** @export */ set duration(value: number) {
    Util.checkPrimitive(value, 'number', AnimationManager, 'duration');
    if (value < 1) Util.throwRangeError(value, '>= 1', AnimationManager, 'duration');
    this._duration = value;
  }

  /**
   * This read-only property is true when the animation manger is currently animating.
   *
   * This value cannot be set, but animation can be stopped by calling {@link #stopAnimation},
   * and it can be prevented by setting {@link #isEnabled}.
   */
  /** @export */ get isAnimating(): boolean { return this._isAnimating; }

  /**
   * This read-only property is true when the animation manger is in the middle of an animation tick.
   * Animation only operates on GraphObjects during ticks, but code outside of AnimationManager's control may execute between ticks.
   *
   * isTicking can only be true when {@link #isAnimating} is also true.
   */
  /** @export */ get isTicking(): boolean { return this._isTicking; }

  /**
   * Gets or sets whether an animation is performed on an initial layout.
   *
   * The default value is true.
   * Changing the value does not affect any ongoing animation.
   * Setting this property does not raise any events.
   * @since 1.6
   */
  /** @export */ get isInitial(): boolean { return this._isInitial; }
  /** @export */ set isInitial(value: boolean) {
    Util.checkPrimitive(value, 'boolean', AnimationManager, 'isInitial');
    this._isInitial = value;
  }

  /**
   * This read-only property gets the {@link Animation} that carries out default GoJS animations.
   * This animation is usually only referenced to modify default animation properties,
   * such as the {@link Animation#easing} or {@link Animation#duration}.
   *
   * You should not add anything to or start the default animation, GoJS does so automatically, internally.
   * When the default animation begins it raises the `"AnimationStarting"` Diagram event,
   * upon completion it raises the `"AnimationFinished"` Diagram event.
   *
   * See the <a href="../../intro/animation.html">Introduction Page on Animations</a> for more detail.
   * @since 2.1
   */
  /** @export */ get defaultAnimation(): Animation { return this._defaultAnimation; }

  /**
   * Gets the set of currently animating {@link Animation}s being managed by this AnimationManager.
   * @since 2.1
   */
  /** @export */ get activeAnimations(): Set<Animation> { return this._activeAnimations; }


  /** @hidden @internal */
  public static AnimationProperties: Map<string, (obj: any /*Diagram | GraphObject*/, startValue: any, endValue: any,
                                                  easing: EasingFunction, currentTime: number,
                                                  duration: number, animation?: Animation) => void> = null;




  /** @hidden @internal */
  private static _hasInit: boolean = false;

  /**
   * Defines a new named effect to be used in animation, along with a function that tells the AnimationManager how to modify that property.
   *
   * Effect names do not need to reflect GraphObject properties, and you can define an effect with a function that modifies several properties for convenience.
   *
   * For example, one could define an animation effect named `"moveAndSpin"` which modifies the object's `position` and `angle`.
   *
   * Most commonly, an effect is defined with one GraphObject property in mind to be animated,
   * and the function uses the start and end values, an easing function, and the times to determine a new value for each tick of animation.
   * Here is an example for animating the fill of GraphObjects:
   *
   * ```js
   * // This presumes the object to be animated is a Shape
   * go.AnimationManager.defineAnimationEffect('fill', function(obj, startValue, endValue, easing, currentTime, duration, animation) {
   *   var hueValue = easing(currentTime, startValue, endValue - startValue, duration);
   *   obj.fill = 'hsl(' + hueValue + ', 100%, 80%)';
   * });
   * ```
   *
   * @param {string} effectName Named effect to animate
   * @param {Function} animationFunction Function
   * that transforms the property values. It takes the animated object, start value, end value,
   * easing function (the {@link Animation#easing}), current time, duration, and animation state. It should modify one or more properties on the object.
   * @since 2.1
   */
  public static defineAnimationEffect(effectName: string, animationFunction: (obj: Diagram | GraphObject, startValue: any, endValue: any,
                                                                              easing: EasingFunction, currentTime: number,
                                                                              duration: number, animation: Animation) => void): void {
      if (!AnimationManager._hasInit) {
        AnimationManager.initStatic();
        AnimationManager._hasInit = true;
      }
      AnimationManager.AnimationProperties.add(effectName, animationFunction);
  }


  /**
   * @hidden @internal
   */
  public static staticEaseColor(starts: any, ends: any, currentTime: number, duration: number, easing: Function): any {
    if (starts === null) starts = 'rgba(0,0,0,0)';
    if (ends === null) ends = 'rgba(0,0,0,0)';
    Brush._colorToRgba(starts);
    Brush._rgbaToHsla();
    const startHue = Brush._hslNums.n0;
    const startSat = Brush._hslNums.n1;
    const startLux = Brush._hslNums.n2;
    const startAlpha = Brush._hslNums.n3;

    Brush._colorToRgba(ends);
    Brush._rgbaToHsla();
    const endHue = Brush._hslNums.n0;
    const endSat = Brush._hslNums.n1;
    const endLux = Brush._hslNums.n2;
    const endAlpha = Brush._hslNums.n3;

    const hueValue = easing(currentTime, startHue, endHue - startHue, duration);
    const satValue = easing(currentTime, startSat, endSat - startSat, duration);
    const luxValue = easing(currentTime, startLux, endLux - startLux, duration);
    const alphaValue = easing(currentTime, startAlpha, endAlpha - startAlpha, duration);
    return 'hsla(' + hueValue + ', ' + satValue + '%, ' + luxValue + '%, ' + alphaValue + ')';
  }

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    const map = new Map<string, (obj: any /* Diagram | GraphObject*/, startValue: any, endValue: any,
                                 easing: EasingFunction, currentTime: number,
                                 duration: number, animation?: Animation) => void>();

    map.add('position:diagram', function (obj: Diagram, starts: Point, ends: Point, easing: Function, currentTime: number, duration: number) {
       obj.position = new Point(
         easing(currentTime, starts.x, ends.x - starts.x, duration),
         easing(currentTime, starts.y, ends.y - starts.y, duration));
    });

    map.add('position', function (obj: GraphObject, starts: Point, ends: Point, easing: Function, currentTime: number, duration: number) {
      if (currentTime !== duration) {
        obj.setPosInplace(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration), false);
      } else {
        // last frame: be sure to set the position using the normal setter
        obj.position = new Point(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration));
      }
    });
    map.add('location', function (obj: Part, starts: Point, ends: Point, easing: Function, currentTime: number, duration: number) {
      if (currentTime !== duration) {
        obj.setPosInplace(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration), true);
      } else {
        // last frame: be sure to set the location using the normal setter
        obj.location = new Point(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration));
      }
    });
    // ? same as position
    map.add('position:placeholder', function (obj: GraphObject, starts: Point, ends: Point, easing: Function, currentTime: number, duration: number) {
      if (currentTime !== duration) {
        obj.setPosInplace(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration), false);
      } else {
        // last frame: be sure to set the position using the normal setter
        obj.position = new Point(
          easing(currentTime, starts.x, ends.x - starts.x, duration),
          easing(currentTime, starts.y, ends.y - starts.y, duration));
      }
    });
    map.add('position:node', function (obj: Part, starts: Point, ends: GraphObject, easing: Function, currentTime: number, duration: number) {
      // move an object from a position to a node's position
      const nab = obj.actualBounds;
      const rab = ends.actualBounds;
      const ex = rab.x + (rab.width / 2) - (nab.width / 2);
      const ey = rab.y + (rab.height / 2) - (nab.height / 2);

      if (currentTime !== duration) {
        obj.setPosInplace(
          easing(currentTime, starts.x, ex - starts.x, duration),
          easing(currentTime, starts.y, ey - starts.y, duration), false);
      } else {
        // last frame: be sure to set the position using the normal setter
        obj.position = new Point(
          easing(currentTime, starts.x, ex - starts.x, duration),
          easing(currentTime, starts.y, ey - starts.y, duration));
      }
    });
    map.add('desiredSize', function (obj: GraphObject, starts: any, ends: any, easing: Function, currentTime: number, duration: number) {
      (obj as any).desiredSize = new Rect(
        easing(currentTime, starts.width, ends.width - starts.width, duration),
        easing(currentTime, starts.height, ends.height - starts.height, duration)
      );
    });
    map.add('width', function (obj: GraphObject, starts: any, ends: any, easing: Function, currentTime: number, duration: number) {
      (obj as any).width = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('height', function (obj: GraphObject, starts: any, ends: any, easing: Function, currentTime: number, duration: number) {
      (obj as any).height = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('fill', function (obj: GraphObject, starts: any, ends: any, easing: Function, currentTime: number, duration: number) {
      (obj as any).fill = AnimationManager.staticEaseColor(starts, ends, currentTime, duration, easing);
    });
    map.add('stroke', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      (obj as any).stroke = AnimationManager.staticEaseColor(starts, ends, currentTime, duration, easing);
    });
    map.add('strokeWidth', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      (obj as any).strokeWidth = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('strokeDashOffset', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      (obj as any).strokeDashOffset = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('background', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      (obj as any).background = AnimationManager.staticEaseColor(starts, ends, currentTime, duration, easing);
    });
    map.add('areaBackground', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      (obj as any).areaBackground = AnimationManager.staticEaseColor(starts, ends, currentTime, duration, easing);
    });
    map.add('opacity', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      obj.opacity = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('scale', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      obj.scale = easing(currentTime, starts, ends - starts, duration);
    });
    map.add('angle', function (obj: GraphObject, starts: number, ends: number, easing: Function, currentTime: number, duration: number) {
      obj.angle = easing(currentTime, starts, ends - starts, duration);
    });

    AnimationManager.AnimationProperties = map;
  }


} // #end AnimationManager
AnimationManager['className'] = 'AnimationManager';
AnimationManager['defineAnimationEffect'] = AnimationManager.defineAnimationEffect;






/**
 * Animations are used to animate GraphObject and Diagram properties.
 *
 * This class is useful for creating manual animations.
 * If you wish to animate particular properties on a GraphObject every time their value changes,
 * you may want to use {@link AnimationTrigger}s instead, which automatically create and start Animations.
 *
 * The method {@link #add} specifies which objects and which animation effects/properties to animate, plus start and end values for this property.
 * As objects are added to an Animation, the Animation infers which {@link Diagram} and {@link AnimationManager} is relevant.
 *
 * Animations are started by calling {@link #start}, and stopped when the {@link #duration} is reached, or when {@link #stop} is called,
 * or stopped when {@link AnimationManager#stopAnimation} is called with `true` as its argument.
 *
 * Unlike the {@link AnimationManager#defaultAnimation}, Animations can be started any time,
 * and do not stop automatically when a new transaction begins. They can also run indefinitely,
 * when {@link runCount} is set to `Infinity`.
 *
 *
 * @since 2.1
 * @unrestricted
 */
export class Animation {
  /** @hidden @internal */ public static className = 'Animation';
  /** @hidden @internal */ private _easing: EasingFunction;
  /** @hidden @internal */ public _isAnimating: boolean;
  /** @hidden @internal */ private _finished: ((animation: Animation) => void) | null;
  /** @hidden @internal */ public _duration: number;
  /** @hidden @internal */ public _currentDuration: number;
  /** @hidden @internal */ private _originPoint: Point;
  /** @hidden @internal */ private _temporaryParts: Set<Part>;
  /** @hidden @internal */ private _objectsToRecompute: Set<GraphObject>;
  /** @hidden @internal */ public _objectsToAnimate: Map<GraphObject | Diagram, AnimationState>;
  /** @hidden @internal */ public _objectsState: Map<GraphObject | Diagram, ObjectData>;
  /** @hidden @internal */ private _start: number;
  /** @hidden @internal */ private _finish: number;
  /** @hidden @internal */ private _reversible: boolean; // option
  /** @hidden @internal */ private _isReversing: boolean; // state
  /** @hidden @internal */ private _runCount: number;
  /** @hidden @internal */ private _repeats: number;
  /** @hidden @internal */ public _isSuspended: boolean;
  /** @hidden @internal */ public _trigger: AnimationTrigger;
  /** @hidden @internal */ public _triggerTarget: GraphObject;

  /** @hidden @internal */ public _animationManager: AnimationManager | null;
  /** @hidden @internal */ private _diagram: Diagram | null;
  /** @hidden @internal */ private _isViewportUnconstrained: boolean;
  /** @hidden @internal */ private _debugCurrentTime: any;

  /**
   *
   */
  constructor() {
    this._easing = Animation.EaseInOutQuad;
    this._diagram = null;
    this._animationManager = null;
    this._trigger = null; // used for Immediate AnimationTriggers only
    this._triggerTarget = null; // used for Immediate AnimationTriggers only

    this._isViewportUnconstrained = false;
    this._isAnimating = false;
    this._isSuspended = false;
    this._start = 0;
    this._finish = 0;
    this._reversible = false; // option
    this._isReversing = false; // state
    this._runCount = 1; // option
    this._repeats = 0; // state
    this._duration = NaN; // option
    this._currentDuration = NaN; // state
    this._debugCurrentTime = 0; // Debug only, used with suspend/advanceTo

    this._finished = null;
    this._originPoint = Point.Origin;

    this._objectsToAnimate = new Map<GraphObject | Diagram, AnimationState>();
    this._objectsState = new Map<GraphObject | Diagram, ObjectData>();
    this._temporaryParts = new Set<Part>();
    this._objectsToRecompute = new Set<GraphObject>();
  }


  /**
   * @hidden
   * @expose
   * Undocumented. Debug only.
   * @this {Animation}
   */
  public suspend(): void {
    this._isSuspended = true;
  }

  /**
   * @hidden
   * @expose
   * Undocumented. Debug only.
   * @this {Animation}
   */
  public advanceTo(duration: number, resume?: boolean): void {
    if (resume) this._isSuspended = false;
    this._debugCurrentTime = duration;
    this.tick(true); // force a tick even if suspended
    this._diagram.redraw(); // force redraw immediately
  }

  /**
   * @hidden @internal
   * Occurs during doCompletion
   * doCompletion also sets:
   *  this._isAnimating = false;
   * beginAnimation also sets, if aborting the isInitial animation:
   *  this._objectsToAnimate.clear();
   * @this {Animation}
   */
  public clearState(): void {
    // don't clear this._objectsToAnimate
    this._objectsState.clear();
    this._isReversing = false; // state
    this._repeats = 0; // state
    this._currentDuration = NaN; // state
    if (this._temporaryParts.count > 0) this._temporaryParts.clear();
    if (this._objectsToRecompute.count > 0) this._objectsToRecompute.clear();
  }

  /**
   * @hidden @internal
   * True if an object or the Diagram position or scale are added to this animation.
   * @this {Animation}
   */
  public isAnythingAnimating() {
    return (this._objectsToAnimate.count > 0);
  }

  /**
   * Start this animation.
   *
   * This adds the Animation to its {@link AnimationManager}'s list of active animations.
   * The AnimationManager is inferred from the list of objects to be animted, by inspecting their Diagram.
   *
   * This does nothing if there are no objects to animate.
   * @this {Animation}
   */
  public start(): void {
    if (this._objectsToAnimate.count === 0) return;
    if (this._isAnimating) return;
    let diagram: Diagram | null = this._diagram;
    // Finish setting up any animation state before animation begins
    const itr = this._objectsToAnimate.iterator;
    while (itr.next()) {
      const settings = itr.value;
      const ends = settings.end;
      const key = itr.key;
      //  If not visible, call findVisibleNode, if its a group, use the placeholder
      //  This used to use the value of ends['position:placeholder'], which was the saved
      //  root group placeholder from addCollapsedNode.
      //  Finding out the root dynamically may help complex situations where
      //  multiple nested groups are collapsed within one transaction.
      if (ends['position:placeholder']) {
        const obj = key as Node;
        const group = obj.findVisibleNode();
        if (group instanceof Group && group.placeholder !== null) {
          const place = group.placeholder;
          const origin = place.getDocumentPoint(Spot.TopLeft);
          const padding = place.padding as Margin;
          origin.x += padding.left;
          origin.y += padding.top;
          ends['position:placeholder'] = origin;
        }
      }
      if (diagram === null) {
        if (key instanceof Diagram) diagram = key;
        else if (key instanceof GraphObject) diagram = key.diagram;
      }
    }
    // Debug:
    // console.log('start');
    // itr.reset();
    // while (itr.next()) {
    //   const settings = itr.value;
    //   const starts = settings.start;
    //   const ends = settings.end;
    //   const key = itr.key;
    //   let startstring = '';
    //   for (const k in starts) {
    //     if (k === 'scale') continue;
    //     startstring += `${k}, ${starts[k]}`;
    //   }
    //   let endstring = '';
    //   for (const k in ends) {
    //     if (k === 'scale') continue;
    //     endstring += `${k}, ${ends[k]}`;
    //   }
    //   console.log(`${key.toString()} with ${startstring} and ${endstring}`);
    // }
    // End Debug
    if (diagram !== null) {
      this._diagram = diagram;
      this._animationManager = diagram.animationManager;
    } else {
      return; // no object to animate had a Diagram, quit
    }
    this._currentDuration = isNaN(this._duration) ? this._animationManager.duration : this._duration;

    this._animationManager.skipState(this);
    this._temporaryParts.each(function(part) {
      part.data = null; // So removing the original part does't remove this part too
      diagram.add(part);
    });
    this._animationManager.restoreState();

    this._isAnimating = true;
    // end setup state, ready to animate
    this._start = +new Date();
    this._finish = this._start + this._currentDuration;
    this._animationManager.animateObjects(this);
  }

  // ??? Closure bug: If I do not @export this, it won't appear at all in minified library
  /**
   * @export
   * Add a temporary Part to this animation.
   * This part will be added to the Diagram when the animation is started,
   * and removed from the Diagram when the animation completes.
   * This is intended to be used with {@link #add}, to animate properties of this Part or its elements.
   *
   * The temporary part added is typically either a {@link GraphObject#copy} of an existing Part,
   * which is to be deleted and requires a copy for animated effects, or else a wholly new temporary Part,
   * constructed in memory for the purpose of creating some effect.
   *
   *
   * @this {Animation}
   * @param {Part} part A part to add to the Diagram at the start of the animation and remove at the end.
   *                    This is typically either a copied Part already in the Diagram, to animate its deletion,
   *                    or a Part created programmatically to be used for some effect.
   * @param {Diagram=} diagram The Diagram to add the temporary part to, and remove it from, at the start and end of animation, respectively.
   */
  public addTemporaryPart(part: Part, diagram: Diagram): void {
    if (!part.isNodeLike()) return; // right now: no adornments or links are animated during removal
    if (Debug) {
      if (diagram === undefined) Util.throwError('addTemporaryPart: Required Diagram argument missing');
      if (part.diagram === diagram) Util.throwError('addTemporaryPart: Part already in Diagram, did you mean to pass in a copy?');
      if (this._diagram !== null && this._diagram !== diagram) {
        Util.throwError('addTemporaryPart: A different Diagram is already associated with this Animation: ' + this._diagram.toString());
      }
    }
    // don't add them until the animation starts, during that time call diagram.add(part)
    this._temporaryParts.add(part);
    this._diagram = diagram;
  }

  /**
   * Add an object (GraphObject or Diagram) and effect name, with specified start and end values, to this Animation.
   *
   * @this {Animation}
   * @param {GraphObject|Diagram} obj GraphObject or Diagram to animate.
   * @param {string} effectName Animation effect name, such as `"scale"` to change GraphObject.scale.
   * By default the supported properties are, for GraphObjects:
   * * `"position"`
   * * `"location"` (on Parts)
   * * `"scale"`
   * * `"opacity"`
   * * `"angle"`
   * * `"desiredSize"`
   * * `"width"`
   * * `"height"`
   * * `"background"`
   * * `"areaBackground"`
   * * `"fill"` (on Shapes)
   * * `"strokeWidth"` (on Shapes)
   * * `"strokeDashOffset"` (on Shapes)
   * * `"stroke"` (on Shapes, TextBlocks)
   *
   * For Diagrams:
   * * `"position"`
   * * `"scale"`
   * * `"opacity"`
   *
   * More properties can be supported by defining new effects with {@link AnimationManager.defineAnimationEffect}.
   * @param {*} startValue The starting value for the animated property. Often this is the current value of the property.
   * @param {*} endValue The ending value for the animated property. Even if the animation is just cosmetic, this must be a valid value for the property.
   * For instance, for {@link GraphObject#scale}, you cannot animate to 0, as this is an invalid scale value.
   * Instead you would animate to a very small (but still valid) value, such as 0.001.
   * @param {boolean=} cosmetic Determines if the animation should revert the property value to the start value at the end of animation.
   * Default false. This is commonly used when animating opacity or scale of "disappearing" nodes during collapse.
   * Even though the node may appear to go to scale 0.001, the programmer usually wants the scale to be reflect its prior value, once hidden.
   * @export
   */
  public add(obj: GraphObject | Diagram, effectName: string, startValue: any, endValue: any, cosmetic?: boolean): void {
    if (effectName === 'position' && startValue.equals(endValue)) return;
    if (this._diagram === null) { // determine diagram and animation manager automatically
      if (obj instanceof Diagram) this._diagram = obj;
      else if (obj instanceof GraphObject && obj.diagram !== null) this._diagram = obj.diagram;
    }
    if (obj instanceof Part) {
      if (!obj.isAnimated) return; // formerly we didn't animate Group positions, which now causes problems.  if (obj instanceof Group && effectName === 'position') return;
    }
    this.add2(obj, effectName, startValue, endValue, cosmetic);
  }

  /**
   * @hidden @internal
   * Used internally to skip checks
   */
  public add2(obj: GraphObject | Diagram, effectName: string, startValue: any, endValue: any, cosmetic?: boolean): void {

    const objmap = this._objectsToAnimate;
    let animationState: AnimationState;
    let mystart: ObjectData;
    let myend: ObjectData;
    if (obj instanceof Diagram && effectName === 'position') effectName = 'position:diagram';
    if (objmap.contains(obj)) {
      animationState = objmap.getValue(obj);
      mystart = animationState.start;
      myend = animationState.end;
      // do not replace starting value if it exists
      if (mystart[effectName] === undefined) mystart[effectName] = this.copyValue(startValue);
      myend[effectName] = this.copyValue(endValue);
    } else {
      mystart = {};
      myend = {};
      mystart[effectName] = this.copyValue(startValue);
      myend[effectName] = this.copyValue(endValue);

      // modify states: ??? do this better
      const p = mystart['position'];
      if (obj instanceof GraphObject && p instanceof Point && !p.isReal() && obj.diagram.animationManager._animationReasons.contains('Expand SubGraph')) {
        p.assign(myend['position']);
      }

      animationState = new AnimationState(mystart, myend, cosmetic);
      objmap.add(obj, animationState);
    }
    // might start at (NaN, NaN)
    const pointProp = mystart[effectName];
    if (pointProp instanceof Point && !pointProp.isReal()) pointProp.assign(this._originPoint);

    // cosmetic collapse: save location to restore:
    if (cosmetic && effectName.indexOf('position:') === 0 && obj instanceof Part) {
      animationState.final['location'] = this.copyValue(obj.location);
    } else if (cosmetic) {
      animationState.final[effectName] = this.copyValue(startValue);
    }
  }

  /**
   * @hidden @internal
   * @this {Animation}
   * @param {*} val
   * @return {*}
   */
  private copyValue(val: any): any {
    if (val instanceof Point) return val.copy();
    if (val instanceof Size) return val.copy();
    // ?? other types, such as Rect, Margin, Spot, Brush, Geometry
    return val;
  }

  /**
   * @hidden @internal
   * @this {Animation}
   * @param {GraphObject} obj
   */
  public markVisible(obj: GraphObject): void {
    const objmap = this._objectsToAnimate;
    if (objmap.contains(obj)) {
      const animationState = objmap.getValue(obj);
      animationState.remainsVisible = true;
    }
  }

  /**
   * @hidden @internal
   * @this {Animation}
   * @param {GraphObject} obj
   * @return {boolean}
   */
  public objRemainsVisible(obj: GraphObject): boolean {
    if (!this._isAnimating) return false;
    const animationState = this._objectsToAnimate.getValue(obj);
    return (animationState !== null && animationState.remainsVisible);
  }

  /**
   * @hidden @internal
   * Checks for position or location.
   * This is used in Placeholder to determine if the group is being moved by the animation manager.
   * If not, it may get moved by other forces, like member parts shifting their actualBounds.
   * @this {Animation}
   * @param {GraphObject} obj
   * @return {boolean}
   */
  public isAnimatingPosition(obj: GraphObject): boolean {
    if (!this._isAnimating) return false;
    const animationState = this._objectsToAnimate.getValue(obj);
    return (animationState !== null && (animationState.start['position'] || animationState.start['location']));
  }

  /**
   * True if any of the objects to animate has a non-null diagram
   * @hidden @internal
   */
  public objectsHaveDiagram() {
    if (this._temporaryParts.count > 0) return true; // deleted parts have no diagram, but we must still animate
    const itr = this._objectsToAnimate.iterator;
    while (itr.next()) {
      const obj = itr.key;
      if ((obj instanceof GraphObject && obj.diagram !== null) || obj instanceof Diagram) {
        return true;
      }
    }
    return false;
  }

  /**
   * @hidden @internal
   */
  public tick(force: boolean) {
    if (this._isSuspended && !force) return;
    const AM = this._animationManager;
    if (this._isAnimating === false) return;
    const time = +new Date();
    let currentTime = time > this._finish ? this._currentDuration : (time - this._start);
    if (force) {
      currentTime = this._debugCurrentTime;
      if (currentTime < this._currentDuration) {
        // Reset start/finish on every forced tick except final ones
        this._start = +new Date() - currentTime; // pretend we started now - currentTime ago
        this._finish = this._start + this._currentDuration;
      } else {
        currentTime = this._currentDuration; // dont allow animations longer than currentDuration
      }
    }
    AM.skipState(this);
    this.modifyObjects(currentTime);
    // if (this._currentOnChange) this._currentOnChange();
    this._diagram.doUpdate();
    AM.restoreState();
    if (time > this._finish) {
      if (this._reversible && !this._isReversing) {
        // now animate in reverse:
        this._start = +new Date();
        this._finish = this._start + this._currentDuration;
        this._isReversing = true;
      } else {
        this.doCompletion(false);
      }
    }
  }

  /**
   * @hidden @internal
   * Called during each animation tick, and at the start and end of Animation.
   * @this {Animation}
   * @param {number} currentTime
   */
  public modifyObjects(currentTime: number): void {
    const duration = this._currentDuration;
    const itr = this._objectsToAnimate.iterator;
    const reverse = this._isReversing;
    while (itr.next()) {
      const obj = itr.key;
      if (obj instanceof GraphObject && obj.diagram === null) { // ?? efficient only in the uncommon case
        continue;
      }
      const settings = itr.value;
      const starts = !reverse ? settings.start : settings.end;
      const ends = !reverse ? settings.end : settings.start;
      const modfunctions = AnimationManager.AnimationProperties;
      // for every setting, see if there's a modfunction for it
      for (const x in ends) {
        // If there's a position:placeholder or position:node set in the animation,
        // then disregard 'position' during the animation loop.
        // The information for 'position' may still be used in the final placement
        if ((x === 'position') && (ends['position:placeholder'] || ends['position:node'])) {
          continue;
        }
        if (modfunctions.get(x) !== null) {
          modfunctions.get(x)(obj, starts[x], ends[x], this._easing, currentTime, duration, this);
        }
      }
    }
  }

  /**
   * Stops a running Animation and updates the animating objects to their final state.
   *
   * If an animation was about to begin, it is cancelled.
   * @export
   * @this {Animation}
   */
  public stop(): void {
    // If no animation or not enabled, quit now
    if (!this._isAnimating) return;
    // Otherwise tick for the last time and do completion
    this.doCompletion(true);
  }

  /**
   * @hidden @internal
   */
  public recomputeOnCompletion(obj: GraphObject): void {
    this._objectsToRecompute.add(obj);
  }

  /**
   * @hidden @internal
   * When animation finishes this calls the finished callback, clears state,
   * notifies the AnimationManager to remove this Animation from its list of current animations, and updates the Diagram.
   *
   * @this {Animation}
   */
  public doCompletion(stopped: boolean): void {
    if (this._triggerTarget !== null) {
      // remove the trigger from the list
      this._triggerTarget.removeTrigger(this._trigger);
    }
    if (!this._isAnimating) return; // ??? check stopped
    const diagram = this._diagram;
    const AM = this._animationManager;
    AM._initialAnimationOngoing = false;
    // the step loop calls AM.doCompletion(this)
    this._isAnimating = false;
    this._isSuspended = false;

    AM.skipState(this);
    const parts = this._objectsToAnimate;

    // Remove any fake parts used for deletion effects
    const itr = this._temporaryParts.iterator;
    while (itr.next()) {
      diagram.remove(itr.value);
    }
    const itr2 = this._objectsToRecompute.iterator;
    while (itr2.next()) {
      itr2.value.invalidateMeasure();
    }

    const reverse = this._reversible;
    const mitr = parts.iterator;
    const modfunctions = AnimationManager.AnimationProperties;
    while (mitr.next()) {
      const obj = mitr.key;
      const settings = mitr.value;
      const starts = (!reverse) ? settings.start : settings.end;
      const ends = (!reverse) ? settings.end : settings.start;
      const final = settings.final; // ??? but in reverse?
      // for every setting, see if there's a modfunction for it
      for (const x in ends) {
        if (modfunctions.get(x) !== null) {
          let func = x;
          if (settings.cosmetic && (func === 'position:node' || func === 'position:placeholder')) func = 'position';
          modfunctions.get(func)(obj, starts[x],
            // For the final tick, if a "final" value was defined, use that.
            // Otherwise, if cosmetic, reset it to "start" (not cosmetic, set it to "end")
            final[x] !== undefined ? final[x] : (settings.cosmetic ? starts[x] : ends[x]),
            this._easing, this._currentDuration, this._currentDuration, this);
        }
      }
      // ?? there might be a more normal way to do this:
      // If its a cosmetic collapse, restore locations:
      if (settings.cosmetic && final['location'] !== undefined && obj instanceof Part) {
        obj.location = final['location'];
      }
      // When objects artifically remain visible their adornments

      // are re-created as they move and need to be cleared again
      if (settings.remainsVisible && obj instanceof Part) obj.madeVisible(false);
    }
    // end modifyObjects

    // ??? update links on arbitrary animations? OR should this only be a function of default animation?
    // necessary for default animation, see flowchart
    if (AM.defaultAnimation === this) {
      const links = this._diagram.links;
      while (links.next()) {
        const link = links.value;
        link.restoreSavedRoute();
      }
    }

    diagram.loadedRouteLinks.clear();
    diagram.disableTempPixelRatio(false);
    diagram.invalidateDocumentBounds();
    diagram.invalidateDraw(); // necessary to notify overviews
    diagram.maybeUpdate(); // update right away to ensure undo manager skips (must be done before restoreState)
    // not clear this is still needed:
    // diagram.updateBounds(); // update bounds manually because skipState will stop it from happening in doUpdate
    AM.restoreState();

    this._repeats++;
    if (!stopped && this._runCount > this._repeats) {
      this._isReversing = false;
      this.start();
      return;
    }

    if (this._finished) this._finished(this);
    this.clearState();
    diagram.requestUpdate();
  }

  /**
   * @hidden @internal
   * Called only by Node's expandTree and Group's expandSubGraph
   * @this {Animation}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addExpandedNode(node: Node, rootNode: Node): void {
    const rab = rootNode.actualBounds;
    let place = null;
    if (rootNode instanceof Group) {
      place = rootNode.placeholder;
    }
    if (place !== null) {
      const origin = place.getDocumentPoint(Spot.TopLeft);
      const padding = place.padding as Margin;
      origin.x += padding.left;
      origin.y += padding.top;
      this.add(node, 'position', origin, node.position, false);
    } else {
      // the default, align centers
      this.add(node, 'position',
        new Point(rab.x + (rab.width / 2), rab.y + (rab.height / 2)),
        node.position,
        false
      );
    }

    this.add(node, 'scale', 0.01, node.scale, false); // not cosmetic
    // this.add(node, 'opacity', 0.01, node.opacity, false);
    if (node instanceof Group) this.expandVisible(node, rootNode);
  }

  /**
   * @hidden @internal
   * When you expand a Tree that contains a group,
   * you must animate the expand of that group's soon-to-be-visible members
   * @this {Animation}
   * @param {Group} group
   * @param {Node} rootNode
   */
  private expandVisible(group: Group, rootNode: Node): void {
    const cit = group.memberParts;
    while (cit.next()) {
      const member = cit.value;
      if (member instanceof Node) this.addExpandedNode(member, rootNode);
    }
  }

  /**
   * @hidden @internal
   * Called only by Node's collapseTree and Group's collapseSubGraph
   * @this {Animation}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addCollapsedNode(node: Node, rootNode: Node): void {
    if (!node.isVisible()) return;
    let place = null;
    if (rootNode instanceof Group) {
      place = rootNode.placeholder;
    }

    if (place !== null) {
      this.add(node, 'position:placeholder', node.position, place, true);
    } else { // the default, align centers
      this.add(node, 'position:node', node.position, rootNode, true);
    }

    this.add(node, 'scale', node.scale, 0.01, true);
    // this.add(node, 'opacity', node.opacity, 0.01, true);
    this.markVisible(node);
    if (node instanceof Group) this.collapseVisible(node, rootNode);
  }

  /**
   * @hidden @internal
   * When you collapse a Tree that contains a group,
   * you must animate the collapse of that group's soon-to-be-invisible members
   * @this {Animation}
   * @param {Group} group
   * @param {Node} rootNode
   */
  private collapseVisible(group: Group, rootNode: Node): void {
    const cit = group.memberParts;
    while (cit.next()) {
      const member = cit.value;
      if (member instanceof Node) this.addCollapsedNode(member, rootNode);
    }
  }

  /**
   * Gets or sets the duration for animations, in milliseconds.
   *
   * The default value is `NaN`, which means it inherits the default value from the {@link AnimationManager#duration},
   * which defaults to 600 milliseconds.
   *
   * The value must be a number greater than or equal to 1, or `NaN`.
   * Setting this property does not raise any events.
   */
  /** @export */ get duration(): number { return this._duration; }
  /** @export */ set duration(value: number) {
    Util.checkPrimitive(value, 'number', Animation, 'duration');
    if (value < 1) Util.throwRangeError(value, '>= 1', Animation, 'duration');
    this._duration = value;
  }

  /**
   * Gets or sets whether this Animation will repeat its animation in reverse at the end of the duration. Default false.
   *
   * A reversible Animation, if stopped early, will end at its original state.
   *
   * Setting this to true doubles the effective {@link #duration} of the Animation.
   */
  /** @export */ get reversible(): boolean { return this._reversible; }
  /** @export */ set reversible(value: boolean) {
    this._reversible = value;
  }

  /**
   * Gets or sets whether this Animation should be repeat, and how many times. The default is 1, which means the animation does not repeat.
   *
   * This can be set to any non-zero positive integer, or `Infinity`. Setting this to `Infinity` will repeat an animation forever.
   *
   * @see {@link #reversible}
   */
  /** @export */ get runCount(): number { return this._runCount; }
  /** @export */ set runCount(value: number) {
    if (value > 0) {
      this._runCount = value;
    } else {
      Util.throwError('Animation.runCount value must be a positive integer.');
    }
  }

  /**
   * Gets or sets the function to execute when the user Animation finishes.
   *
   * By default this property is null.
   */
  /** @export */ get finished(): ((animation: Animation) => void) | null { return this._finished; }
  /** @export */ set finished(value: ((animation: Animation) => void) | null) {
    const old = this._finished;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'click');
      this._finished = value;
    }
  }

  /**
   * Gets or sets the easing function this Animation will use to modify default properties.
   *
   * Pre-defined animatable values are processed by passing scalars into this easing function.
   *
   * The default value is {@link Animation.EaseInOutQuad}.
   *
   *
   * The value can be an arbitrary easing function, or one of the six provided:
   * {@link Animation.EaseLinear}, {@link Animation.EaseInOutQuad}, {@link Animation.EaseInQuad},
   * {@link Animation.EaseOutQuad}, {@link Animation.EaseInExpo}, {@link Animation.EaseOutExpo}.
   *
   */
  /** @export */ get easing(): EasingFunction { return this._easing; }
  /** @export */ set easing(value: EasingFunction) {
    this._easing = value;
  }

  /**
   * Gets or sets whether this Animation should allow an unconstrained viewport during the runtime of the animation.
   */
  /** @export */ get isViewportUnconstrained(): boolean { return this._isViewportUnconstrained; }
  /** @export */ set isViewportUnconstrained(value: boolean) {
    this._isViewportUnconstrained = value;
  }

  /**
   * This read-only property is true when the Animation is currently running.
   *
   * This value cannot be set, but Animation can be stopped by calling {@link #stop}.
   */
  /** @export */ get isAnimating(): boolean { return this._isAnimating; }

  /**
   * Gets the {@link ObjectData} assocaited with this GraphObject or Diagram.
   * If no state exists, this creates and returns a new {@link ObjectData}.
   *
   * This can be used to store temporary information per animated object during the course of an animation.
   * This state is cleared at the end of an animation.
   * @export
   */
  public getTemporaryState(obj: GraphObject | Diagram): ObjectData {
    let state = this._objectsState.get(obj);
    if (state === null) {
      state = { };
      this._objectsState.add(obj, state);
    }
    return state;
  }


  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * @constant
   */
  public static EaseLinear: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    return byValue * currentTime / duration + startValue;
  };

  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * This is the default value for {@link Animation#easing}.
   * @constant
   */
  public static EaseInOutQuad: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    currentTime /= (duration / 2);
    if (currentTime < 1) return byValue / 2 * currentTime * currentTime + startValue;
    return -byValue / 2 * ((--currentTime) * (currentTime - 2) - 1) + startValue;
  };

  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * @constant
   */
  public static EaseInQuad: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    return byValue * (currentTime /= duration) * currentTime + startValue;
  };

  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * @constant
   */
  public static EaseOutQuad: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    return -byValue * (currentTime /= duration) * (currentTime - 2) + startValue;
  };

  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * @constant
   */
  public static EaseInExpo: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    return (currentTime === 0) ? startValue : byValue * Math.pow(2, 10 * (currentTime / duration - 1)) + startValue;
  };

  /**
   * Built-in static function for computing interpolated values. Can be used as a value for {@link Animation#easing}.
   * @constant
   */
  public static EaseOutExpo: EasingFunction = function(currentTime: number, startValue: number, byValue: number, duration: number): number {
    return (currentTime === duration) ? startValue + byValue : byValue * (-Math.pow(2, -10 * currentTime / duration) + 1) + startValue;
  };

} // #end Animation
Animation['className'] = 'Animation';

Animation['EaseLinear'] = Animation.EaseLinear;
Animation['EaseInOutQuad'] = Animation.EaseInOutQuad;
Animation['EaseInQuad'] = Animation.EaseInQuad;
Animation['EaseOutQuad'] = Animation.EaseOutQuad;
Animation['EaseInExpo'] = Animation.EaseInExpo;
Animation['EaseOutExpo'] = Animation.EaseOutExpo;


// * AnimationState collect information for each GraphObject or Diagram being animated.
// *
// * The method {@link Animation#add} specifies which objects and which properties to animate, plus start and end values for this property.
// * As objects are added to an Animation, the Animation infers which Diagram and AnimationManager is relevant.
// *
// * Animations are started by calling {@link #start}, and stopped when the {@link #duration} is reached, or when {@link #stop} is called,
// * or stopped when {@link AnimationManager#stopAnimation} is called with `true` as its argument.
// * @since 2.1

/**
 * @hidden @internal
 * Undocumented
 *
 * @unrestricted
 */
export class AnimationState {
  /** @hidden @internal */ public static className = 'AnimationState';
  public start: ObjectData;
  public end: ObjectData;
  public final: ObjectData;
  public cosmetic: boolean;
  public remainsVisible: boolean;

  /**
   * @param {ObjectData} start
   * @param {ObjectData} end
   * @param {boolean} cosmetic true if the change is cosmetic (reverts to start after animation completes)
   */
  constructor(start: ObjectData, end: ObjectData, cosmetic: boolean) {
    this.start = start;
    this.end = end;
    this.final = {};
    this.cosmetic = cosmetic;
    this.remainsVisible = false;
  }
} // #end AnimationState
AnimationState['className'] = 'AnimationState';



/**
 * An AnimationTrigger describes how to automatically animate a property on a {@link GraphObject}
 * when it changes value.
 * The target property name is a string, and all name matching is case-sensitive.
 *
 * Triggers will be shared by all copies of the template's GraphObjects.
 * You can include AnimationTriggers in your templates just like {@link Binding}s are included:
 *
 * ```js
 * $(go.Panel, "Vertical",
 *   // This trigger uses the default value of AnimationTrigger.startCondition:
 *   // If a transaction is ongoing and Panel.position is changed, this trigger will animate
 *   // all changes to Panel.position at the end of the next transaction, in one bundled Animation.
 *   // If no transaction is ongoing, then it will animate this value immediately.
 *   new go.AnimationTrigger("position"),
 *   {
 *     // ...
 *   },
 *   $(go.Shape,
 *     // Animate all changes to Shape.opacity immediately
 *     new go.AnimationTrigger("opacity", null, startCondition: go.AnimationTrigger.Immediate),
 *     {
 *      // ...
 *     }
 *   )
 * ```
 * When the {@link #startCondition} is {@link AnimationTrigger.Default},
 * GoJS will attempt to {@link AnimationTrigger.Bundled} or {@link AnimationTrigger.Immediate} based on the state of the transaction.
 * If no transaction is ongoing, this trigger will treat the default as using {@link AnimationTrigger.Immediate}.
 * Otherwise it will work as {@link AnimationTrigger.Bundled}.
 *
 * When the {@link #startCondition} is {@link AnimationTrigger.Bundled},
 * the AnimationManager will use the default animation to prepare a single Animation that begins when the current transaction has ended.
 * This animation may be canceled if a new transaction is started.
 *
 * When the {@link #startCondition} is {@link AnimationTrigger.Immediate},
 * a new animation will be created for every instance of the property changed, and started immediately,
 * and run until completion. This may be useful for cosmetic changes, such as animating
 * the opacity or color of an object on mouseEnter or mouseLeave.
 *
 * You can only specify properties that exist on the GraphObject, and are also registered with {@link AnimationManager.defineAnimationEffect}.
 * By default these properties are:
 *
 * * `"position"`
 * * `"location"` (on Parts)
 * * `"scale"`
 * * `"opacity"`
 * * `"angle"`
 * * `"desiredSize"`
 * * `"width"`
 * * `"height"`
 * * `"background"`
 * * `"areaBackground"`
 * * `"fill"` (on Shapes)
 * * `"strokeWidth"` (on Shapes)
 * * `"strokeDashOffset"` (on Shapes)
 * * `"stroke"` (on Shapes, TextBlocks)
 *
 * Examples of defining additional animation properties are given in the <a href="../../intro/animation.html">Introduction Page on Animations</a>.
 *
 * @since 2.1
 * @unrestricted
 */
export class AnimationTrigger {
  /** @hidden @internal */ public static className = 'AnimationTrigger';
  /** @hidden @internal */ public _targetObject: GraphObject;
  /** @hidden @internal */ private _propertyName: string;
  /** @hidden @internal */ private _startCondition: EnumValue;
  /** @hidden @internal */ private _animationSettings: { duration?: number, finished?: (animation: Animation) => void, easing?: EasingFunction };

  /**
   * This constructor creates an AnimationTrigger. These are typically constructed within Part templates. Using {@link GraphObject.make} it might look like:
   *
   * ```js
   *  var $ = go.GraphObject.make;
   *
   *    // ...
   *    $(go.Shape,
   *      // Animate all changes to Shape.opacity immediately
   *      new go.AnimationTrigger("opacity", null, startCondition: go.AnimationTrigger.Immediate),
   *      {
   *       // ...
   *      }
   *    )
   * ```
   * @param {string} propertyName A string naming the target property to animate.
   *   This should not be the empty string.
   * @param {Object=} animationSettings An optional Object describing properties to set on animations created by this AnimationTrigger.
   *   See the {@link #animationSettings} property for detail.
   *   If set this also defaults the {@link #startCondition} to {@link AnimationTrigger.Immediate}.
   * @param {EnumValue=} startCondition An optional EnumValue to set the {@link #startCondition} property.
   */
  constructor(propertyName: string, animationSettings?: { duration?: number, finished?: (animation: Animation) => void, easing?: EasingFunction }, startCondition?: EnumValue) {
    this._targetObject = null;
    this._propertyName = propertyName;
    this._startCondition = startCondition || AnimationTrigger.Default;
    this._animationSettings = null;
    if (animationSettings !== undefined) {
      this._animationSettings = animationSettings;
      if (startCondition === undefined) this._startCondition = AnimationTrigger.Immediate;
    }
  }

  /**
   * Create a copy of this AnimationTrigger, with the same property values.
   * @expose
   * @this {AnimationTrigger}
   * @return {AnimationTrigger}
   */
  public copy(): AnimationTrigger {
    const copy = new AnimationTrigger(this._propertyName);
    copy._startCondition = this._startCondition;
    const settings = this._animationSettings; // copy animationSettings instead of just referencing the object?
    if (settings !== null) {
      const copysettings: ObjectData = {};
      if (settings.duration !== undefined) copysettings.duation = settings.duration;
      if (settings.finished !== undefined) copysettings.duation = settings.finished;
      if (settings.easing !== undefined) copysettings.duation = settings.easing;
      copy._animationSettings = copysettings;
    }
    return copy;
  }
  /**
   * Gets or sets the name of the property to animate on the target {@link GraphObject}.
   * The default value is set during constructor initalization.
   *
   * You can only specify properties that exist on the GraphObject, and are also registered with {@link AnimationManager.defineAnimationEffect}.
   * By default these properties are the same as the list of possible Animation effects:
   *
   * * `"position"`
   * * `"location"` (on Parts)
   * * `"scale"`
   * * `"opacity"`
   * * `"angle"`
   * * `"desiredSize"`
   * * `"width"`
   * * `"height"`
   * * `"background"`
   * * `"areaBackground"`
   * * `"fill"` (on Shapes)
   * * `"strokeWidth"` (on Shapes)
   * * `"strokeDashOffset"` (on Shapes)
   * * `"stroke"` (on Shapes, TextBlocks)
   *
   * Examples of defining additional properties by adding animation effects are given in the <a href="../../intro/animation.html">Introduction Page on Animations</a>.
   */
  /** @export */ get propertyName(): string { return this._propertyName; }
  /** @export */ set propertyName(value: string) {
    this._propertyName = value;
  }

  /**
   * Gets or sets the settings that this trigger should set on any {@link Animation}s it creates
   * if the {@link #startCondition} is {@link AnimationTrigger.Immediate}. Immediate triggers create a new Animation with each triggering,
   * and apply these settings to that Animation.
   *
   * This can be set to an object with a subset of possible {@link Animation} settings. The default value is `null`, which keeps default Animation settings.
   *
   * Since a {@link #startCondition} of {@link AnimationTrigger.Bundled} uses the default animation,
   * you must set the properties of {@link AnimationManager#defaultAnimation}, and not this property, to modify the animation settings.
   *
   * To set default settings for *all* created Animations, you can modify the settings on {@link AnimationManager} instead, such as {@link AnimationManager#duration}.
   *
   * Possible properties to set in this object are:
   *
   * * **duration**: a `number`, corresponding to {@link Animation#duration}.
   * * **finished**: a `Function`, corresponding to {@link Animation#finished}.
   * * **easing**: an `EasingFunction`, corresponding to {@link Animation#easing}.
   */
  /** @export */ get animationSettings(): { duration?: number, finished?: (animation: Animation) => void, easing?: EasingFunction } { return this._animationSettings; }
  /** @export */ set animationSettings(value: { duration?: number, finished?: (animation: Animation) => void, easing?: EasingFunction }) {
    this._animationSettings = value;
  }

  /**
   * Used in GraphObject.raiseChangedEvent
   * @hidden @internal
   */
  public applySettings(animation: Animation) {
    const settings = this._animationSettings;
    if (settings === null) return;
    if (settings.duration) animation.duration = settings.duration;
    if (settings.finished) animation.finished = settings.finished;
    if (settings.easing) animation.easing = settings.easing;
  }

  /**
   * Gets or sets the starting condition for this trigger.
   *
   * AnimationTriggers can invoke an animation immediately,
   * starting a new animation with each property of each GraphObject that has been modified,
   * or they can (more efficiently) be bundled together into the default animation ({@link AnimationManager#defaultAnimation})
   * and begin only one animation, at the end of the next transaction.
   *
   * It is useful for the startCondition to be {@link AnimationTrigger.Immediate} when changing GraphObject properties
   * on {@link GraphObject#mouseEnter} or {@link GraphObject#mouseLeave}.
   * It is useful for the startCondition  to be {@link AnimationTrigger.Bundled} when changing several GraphObject properties together,
   * such as when highlighting multiple parts, on selection changes, and during transactions.
   *
   * These behaviors can be set with the values {@link AnimationTrigger.Immediate} and {@link AnimationTrigger.Bundled}, respectively.
   * The default value, {@link AnimationTrigger.Default}, attempts to infer which is best:
   * It will start immediately if there is no ongoing transaction
   * or if {@link Diagram#skipsUndoManager} is true.
   */
  /** @export */ get startCondition(): EnumValue { return this._startCondition; }
  /** @export */ set startCondition(value: EnumValue) {
    EnumValue.checkEnumValue(value, Brush, Brush, 'type');
    if (Debug) EnumValue.checkEnumValue(value, AnimationTrigger, AnimationTrigger, 'startCondition');
    this._startCondition = value;
  }

  /**
   * Used as a value for {@link #startCondition}.
   * GoJS will attempt to {@link AnimationTrigger.Bundled} or {@link AnimationTrigger.Immediate} based on the state of the transaction.
   * If no transaction is ongoing, this trigger will be treated as using {@link AnimationTrigger.Immediate},
   * otherwise it will work as {@link AnimationTrigger.Bundled}.
   * @constant
   */
  public static Default = new EnumValue(AnimationTrigger, 'Default', 1);

  /**
   * Used as a value for {@link #startCondition}.
   * A new animation will be created for every instance of the property changed, and started immediately,
   * and run until completion. This may be useful for cosmetic changes, such as animating
   * the opacity or color of an object on mouseEnter or mouseLeave.
   * However, using {@link AnimationTrigger.Bundled} may be more efficient, as it will create fewer
   * @constant
   */
  public static Immediate = new EnumValue(AnimationTrigger, 'Immediate', 2);

  /**
   * Used as a value for {@link #startCondition}.
   * The AnimationManager will use the default animation to prepare a single Animation that begins when the current transaction has ended.
   * This animation may be canceled if a new transaction is started.
   * @constant
   */
  public static Bundled = new EnumValue(AnimationTrigger, 'Bundled', 3);

} // #end AnimationTrigger
AnimationTrigger['className'] = 'AnimationTrigger';
AnimationTrigger['Default'] = AnimationTrigger.Default;
AnimationTrigger['Immediate'] = AnimationTrigger.Immediate;
AnimationTrigger['Bundled'] = AnimationTrigger.Bundled;

// import { Brush } from './brush';
// import { Map, Set } from './collections';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Group, Link, Node, Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData, root } from './root';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Util } from './util';






/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Layer
*/


/**
 * Layers are how named collections of {@link Part}s are drawn in front or behind other collections of Parts in a {@link Diagram}.
 * Layers can only contain {@link Part}s, such as {@link Node}s and {@link Link}s. They cannot hold {@link GraphObject}s directly.
 *
 * You put a Part into a Layer by assigning {@link Part#layerName} with the name of the Layer.
 * You can use data binding to initialize and remember a Part's layer's name.
 * You can change a Part's layer by modifying its {@link Part#layerName}, which changes its {@link Part#layer}.
 *
 * Each Diagram starts off with the following list of Layers:
 * "Grid", "Background", "" (the default layer), "Foreground", "Adornment", "Tool".
 * Parts are normally put in the default layer.
 * The "Grid", "Adornment", and "Tool" layers are considered {@link #isTemporary}.
 * Changes to objects in temporary layers are not recorded by the {@link UndoManager}.
 * Parts in temporary layers are not selected and are not considered to be part of the document.
 * Objects in temporary layers do not receive click events unless you set their {@link GraphObject#isActionable} to true.
 * The "Grid" layer is the furthest back; it also contains "temporary" parts that cannot be selected.
 * Furthermore the "Grid" layer has {@link #pickable} set to false so that mouse or touch events
 * and calls to the "find..." methods do not even consider any parts in that layer.
 *
 * Layers have many properties that control what actions users are permitted to perform involving the parts in the layer.
 * These properties are very much like the similarly named properties on {@link Diagram}.
 * <h3>Z-ordering</h3>
 *
 * Layers are drawn and presented in order.
 * You can add your own layers by calling {@link Diagram#addLayerBefore} or {@link Diagram#addLayerAfter}
 * to insert a new layer at a particular place in the Z-order, or to re-order existing layers.
 * Use {@link Diagram#findLayer} to get the Layer with a particular name.
 * {@link Part}s can be individually z-ordered within a layer by setting {@link Part#zOrder}.
 * @unrestricted
 */
export class Layer {
  /** @hidden @internal */ public static className = 'Layer';
  /** @hidden @internal */ private _diagram: Diagram | null;
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _opacity: number;
  /** @hidden @internal */ private _isTemporary: boolean;
  /** @hidden @internal */ private _allowCopy: boolean;
  /** @hidden @internal */ private _allowDelete: boolean;
  /** @hidden @internal */ private _allowTextEdit: boolean;
  /** @hidden @internal */ private _allowGroup: boolean;
  /** @hidden @internal */ private _allowUngroup: boolean;
  /** @hidden @internal */ private _allowLink: boolean;
  /** @hidden @internal */ private _allowRelink: boolean;
  /** @hidden @internal */ private _allowMove: boolean;
  /** @hidden @internal */ private _allowReshape: boolean;
  /** @hidden @internal */ private _allowResize: boolean;
  /** @hidden @internal */ private _allowRotate: boolean;
  /** @hidden @internal */ private _allowSelect: boolean;
  /** @hidden @internal */ private _visible: boolean;
  /** @hidden @internal */ private _pickable: boolean;
  /** @hidden @internal */ private _isBoundsIncluded: boolean;
  /** @hidden @internal */ private _orthoLinks: Array<Link>;
  /** @hidden @internal */ public _parts: List<Part>;

  /**
   * This constructs an empty Layer; you should set the {@link #name} before adding the Layer to a Diagram.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(Layer);
    Set.uniqueHash(this);

    this._diagram = null;
    this._parts = new List<Part>();
    this._name = '';
    this._opacity = 1;
    this._isTemporary = false;

    this._allowCopy = true;
    this._allowDelete = true;
    this._allowTextEdit = true;
    this._allowGroup = true;
    this._allowUngroup = true;
    this._allowLink = true;
    this._allowRelink = true;
    this._allowMove = true;
    this._allowReshape = true;
    this._allowResize = true;
    this._allowRotate = true;
    this._allowSelect = true;
    this._visible = true;
    this._pickable = true;

    this._isBoundsIncluded = false;

    this._orthoLinks = [];  // Orthogonal Links that were drawn
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {Diagram} val
   */
  public setDiagram(val: Diagram) {
    this._diagram = val;
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number=} details int Optional: If greater than zero, this includes descriptions of each part in the layer.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    const str = 'Layer \"' + this.name + '\"';
    if (details <= 0) return str;
    // if details >= 1, talk about the parts in the layer
    let numparts = 0;
    let numnodes = 0;
    let numgroups = 0;
    let numlinks = 0;
    let numadorns = 0;
    const it = this._parts.iterator;
    while (it.next()) {
      const p = it.value;
      // gotta check most specific classes first...
      if (p instanceof Group) numgroups++;
      else if (p instanceof Node) numnodes++;
      else if (p instanceof Link) numlinks++;
      else if (p instanceof Adornment) numadorns++;
      else numparts++;
    }
    let msg = '';
    if (numparts > 0) msg += numparts + ' Parts ';
    if (numnodes > 0) msg += numnodes + ' Nodes ';
    if (numgroups > 0) msg += numgroups + ' Groups ';
    if (numlinks > 0) msg += numlinks + ' Links ';
    if (numadorns > 0) msg += numadorns + ' Adornments ';
    if (details > 1) {
      const pit = this._parts.iterator;
      while (pit.next()) {
        const part = pit.value;
        msg += '\n    ' + part.toString();
        const data = part.data;
        if (data !== null && Set.hashId(data)) msg += ' #' + Set.hashId(data);
        // ?? move into Part
        if (part instanceof Node) {
          msg += ' ' + Util.toString(data);
        } else if (part instanceof Link) {
          msg += ' ' + Util.toString(part.fromNode) + ' ' + Util.toString(part.toNode);
        }
      }
    }
    return str + ' ' + this._parts.count + ': ' + msg;
  }

  /**
   * Find the front-most {@link GraphObject} in this layer
   * at the given point in document coordinates.
   *
   * If {@link #visible} is false, this method will not find any objects in this layer.
   * However, {@link #opacity} does not affect this method.
   * @export
   * @this {Layer}
   * @param {Point} p A Point in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @return {GraphObject} The first GraphObject in the Z-order, or else null.
   */
  public findObjectAt<T extends GraphObject>(p: Point,
                                             navig?: ((a: GraphObject) => (T | null)) | null,
                                             pred?: ((a: T) => boolean) | null): (T | null) {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (this._pickable === false) return null;
    if (Debug && !p.isReal()) Util.throwError('findObjectAt: Point must have a real value, not: ' + p.toString());

    // if p is in the viewportBounds we can just look at viewportObjects
    let inView = false;
    if (this.diagram !== null && this.diagram.viewportBounds.containsPoint(p)) {
      inView = true;
    }

    const mult = Point.alloc();
    const parts = this._parts._arr;
    const len = parts.length;
    for (let i = len; i--; ) {
      const part = parts[i];
      if (inView === true && part.getInView() === false) continue;
      if (!part.isVisible()) continue;
      mult.assign(p);
      mult.transformInverted(part.allTransforms);
      let obj = part.findObjectAt(mult, navig, pred);
      if (obj !== null) {
        if (navig !== null) obj = navig(obj);
        if (obj !== null && (pred === null || pred(obj))) {
          Point.free(mult);
          return obj;
        }
      }
    }
    Point.free(mult);
    return null;
  }

  /**
   * Return a collection of the {@link GraphObject}s of this layer
   * at the given point in document coordinates.
   *
   * If {@link #visible} is false, this method will not find any objects in this layer.
   * However, {@link #opacity} does not affect this method.
   * @export
   * @this {Layer}
   * @param {Point} p A Point in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects that will contain all GraphObjects
   * located at Point p, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   */
  public findObjectsAt<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(p: Point,
                                                                                   navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                   pred?: ((a: T) => boolean) | null,
                                                                                   coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    if (this._pickable === false) return coll;
    if (Debug && !p.isReal()) Util.throwError('findObjectsAt: Point must have a real value, not: ' + p.toString());

    // if p is in the viewportBounds we can just look at viewportObjects
    let inView = false;
    if (this.diagram !== null && this.diagram.viewportBounds.containsPoint(p)) {
      inView = true;
    }

    const mult = Point.alloc();
    const parts = this._parts._arr;
    const len = parts.length;
    for (let i = len; i--; ) {
      const obj = parts[i];
      if (inView === true && obj.getInView() === false) continue;
      if (!obj.isVisible()) continue;
      mult.assign(p);
      mult.transformInverted(obj.allTransforms);
      let graphobject = obj as any as T;
      if (obj.findObjectsAt(mult, navig, pred, coll)) {
        if (navig !== null) graphobject = navig(graphobject);
        if (graphobject !== null && (pred === null || pred(graphobject))) {
          coll.add(graphobject);
        }
      }
    }
    Point.free(mult);
    return coll;
  }

  /**
   * Returns a collection of all {@link GraphObject}s that are inside or that intersect
   * a given {@link Rect} in document coordinates.
   *
   * If {@link #visible} is false, this method will not find any objects in this layer.
   * However, {@link #opacity} does not affect this method.
   * @export
   * @this {Layer}
   * @param {Rect} r A Rect in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {boolean=} partialInclusion Whether an object can match if it merely intersects the rectangular area (true) or
   * if it must be entirely inside the rectangular area (false).  The default value is false.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects that will contain all GraphObjects
   * located in or near Rect r, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   */
  public findObjectsIn<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(r: Rect,
                                                                                   navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                   pred?: ((a: T) => boolean) | null,
                                                                                   partialInclusion?: boolean,
                                                                                   coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = false;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    if (this._pickable === false) return coll;
    if (Debug && !r.isReal()) Util.throwError('findObjectsIn: Rect must have a real value, not: ' + r.toString());

    // if r is in the viewportBounds we can just look at viewportObjects
    let inView = false;
    if (this.diagram !== null && this.diagram.viewportBounds.containsRect(r)) {
      inView = true;
    }

    const parts = this._parts._arr;
    const len = parts.length;
    for (let i = len; i--; ) {
      const obj = parts[i];
      if (inView === true && obj.getInView() === false) continue;
      if (!obj.isVisible()) continue;
      let graphobject = obj as any as T;
      if (obj.findObjectsIn(r, navig, pred, partialInclusion, coll)) {
        if (navig !== null) graphobject = navig(graphobject);
        if (graphobject !== null && (pred === null || pred(graphobject))) {
          coll.add(graphobject);
        }
      }
    }

    return coll;
  }

  /**
   * Returns a collection of all {@link GraphObject}s that are within a certain distance
   * of a given point in document coordinates.
   *
   * If {@link #visible} is false, this method will not find any objects in this layer.
   * However, {@link #opacity} does not affect this method.
   * @export
   * @this {Layer}
   * @param {Point} p A Point in document coordinates.
   * @param {number} dist The distance from the point.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {*=} partialInclusion Whether an object can match if it merely intersects the circular area (true) or
   * if it must be entirely inside the circular area (false).  The default value is true.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects that will contain all GraphObjects
   * located at Point p, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   */
  public findObjectsNear<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(p: Point,
                                                                                     dist: number,
                                                                                     navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                     pred?: ((a: T) => boolean) | null,
                                                                                     partialInclusion?: boolean | S,
                                                                                     coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = true;
    if (partialInclusion !== false && partialInclusion !== true) {  // compatibility
      if (partialInclusion instanceof List || partialInclusion instanceof Set) {
        coll = partialInclusion as S;
      }
      partialInclusion = true;
    }
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    if (this._pickable === false) return coll;
    if (Debug && !p.isReal()) Util.throwError('findObjectsNear: Point must have a real value, not: ' + p.toString());

    // if p is in the viewportBounds we can just look at viewportObjects
    // ??? but here maybe we want to test objects even if offscreen because they might be "near"?
    let inView = false;
    if (this.diagram !== null && this.diagram.viewportBounds.containsPoint(p)) {
      inView = true;
    }

    const mult = Point.alloc();
    const distp = Point.alloc();
    const parts = this._parts._arr;
    const len = parts.length;
    for (let i = len; i--; ) {
      const obj = parts[i];
      if (inView === true && obj.getInView() === false) continue;
      if (!obj.isVisible()) continue;
      mult.assign(p);
      mult.transformInverted(obj.allTransforms);
      distp.assignTo(p.x + dist, p.y);
      distp.transformInverted(obj.allTransforms);
      let graphobject = obj as any as T;
      if (obj.findObjectsNear(mult, distp, navig, pred, partialInclusion, coll)) {
        if (navig !== null) graphobject = navig(graphobject);
        if (graphobject !== null && (pred === null || pred(graphobject))) {
          coll.add(graphobject);
        }
      }
    }
    Point.free(mult);
    Point.free(distp);
    return coll;
  }

  /**
   * @hidden @internal
   * Called from Link
   * @this {Layer}
   * @return {Array.<Part>}
   */
  public getOrthoLinksArray(): Array<Part> {
    return this._parts._arr;  // ?? cannot return _orthoLinks because of missing offscreen links
  }

  /**
   * @hidden @internal
   * Sets flags to draw only the parts in the exclusive set or map.
   * @this {Layer}
   * @param {Diagram} diagram
   * @param {Rect=} viewRect an optional alternative viewport rectangle.
   */
  public setInView(diagram: Diagram, viewRect?: Rect) {
    if (!this.visible) return;

    let vpb: Rect;
    if (viewRect === undefined) {
      vpb = diagram.viewportBounds;
    } else {
      vpb = viewRect;
    }

    const parts = this._parts._arr;
    const len = parts.length;
    // Separated in case they modify the parts list
    const adornValidate = Util.tempArray<Part>();
    const adornUpdate = Util.tempArray<Part>();
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      p._layerIndex = i;
      if (p instanceof Link) {
        if (p.validRoute === false) continue;
      }

      if (p instanceof Adornment) {
        if (p.adornedPart !== null) continue;
      }

      const ab = p.actualBounds;
      if (ab.intersectsRectPlus(vpb, 10)) { // ?? 10 is an arbitrary amount to cover all adornments
        p.setInView(true);
        adornValidate.push(p);
      } else {
        p.setInView(false);
        if (p.adornments !== null && p.adornments.count > 0) {
          adornUpdate.push(p);
        }
      }
    }

    for (let i = 0; i < adornValidate.length; i++) {
      const p = adornValidate[i];
      p.validateAdornments();
      const ads = p.adornments;
      while (ads.next()) {
        const ad = ads.value;
        ad.measure(Infinity, Infinity);
        ad.arrange();
        ad.setInView(true);
      }
    }
    for (let i = 0; i < adornUpdate.length; i++) {
      const p = adornUpdate[i];
      // used to be p.validateAdornments();
      p.updateAdornments();
      p.setInvalidAdornments(true);
    }
    Util.freeArray(adornValidate);
    Util.freeArray(adornUpdate);


  }

  /**
   * @hidden @internal
   */
  private setOpacity(context: IContext): number {
    let old = 1;
    if (this._opacity !== 1) {
      old = context.globalAlpha;
      context.globalAlpha = old * this._opacity;
    }
    return old;
  }

  /**
   * @hidden @internal
   * Draws each Part in this layer.
   * @this {Layer}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {boolean=} showTemporary true if everything should be drawn, false if temporary objects should be omitted.
   */
  public draw(context: IContext, diagram: Diagram, showTemporary?: boolean) {
    if (!this.visible || this._opacity === 0) return;
    if (showTemporary === undefined) showTemporary = true;
    if (!showTemporary && this.isTemporary) return;
    const parts = this._parts._arr;
    const len = parts.length;
    if (len === 0) return;
    const old = this.setOpacity(context);

    const ortho = this._orthoLinks;
    ortho.length = 0;
    const scale = diagram.scale;
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      if (!p.getInView() && !p.hasHTMLElement()) continue;
      if (p instanceof Link) {
        if (p.isOrthogonal) ortho.push(p);
        if (p.validRoute === false) continue;
      }
      const ab = p.actualBounds;
      if (ab.width * scale > diagram._minDrawingLength || ab.height * scale > diagram._minDrawingLength) {
        p.draw(context, diagram);
      } else {
        this.drawSimple(context, p);
      }
    }

    context.globalAlpha = old;
  }

  /**
   * @hidden @internal
   * Draws each Part in this layer.
   * @this {Layer}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {Set.<Part>} exclusion
   */
  public drawExclusion(context: IContext, diagram: Diagram, exclusion: Set<Part>) {
    if (!this.visible || this._opacity === 0) return;
    const parts = this._parts._arr;
    const len = parts.length;
    if (len === 0) return;
    const old = this.setOpacity(context);

    const ortho = this._orthoLinks;
    ortho.length = 0;
    const scale = diagram.scale;
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      if (!p.getInView() && !p.hasHTMLElement()) continue;
      if (exclusion.contains(p)) continue;
      if (p instanceof Link) {
        if (p.isOrthogonal) ortho.push(p);
        if (p.validRoute === false) continue;
      }
      const ab = p.actualBounds;
      if (ab.width * scale > diagram._minDrawingLength || ab.height * scale > diagram._minDrawingLength) {
        p.draw(context, diagram);
      } else {
        this.drawSimple(context, p);
      }
    }

    context.globalAlpha = old;
  }

  /**
   * @hidden @internal
   * Draws only the parts in the exclusive set or map.
   * @this {Layer}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {Set.<Part>} exclusive
   * @param {boolean=} showTemporary show temporary layers or not
   */
  public drawOnly(context: IContext, diagram: Diagram, exclusive: Set<Part>, showTemporary?: boolean) {
    if (!this.visible || this._opacity === 0) return;
    if (showTemporary === undefined) showTemporary = true;
    if (!showTemporary && this.isTemporary) return;
    const parts = this._parts._arr;
    const len = parts.length;
    if (len === 0) return;
    const old = this.setOpacity(context);

    const ortho = this._orthoLinks;
    ortho.length = 0;
    const scale = diagram.scale;
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      if (!p.getInView() && !p.hasHTMLElement()) continue;
      if (!exclusive.contains(p)) continue;
      if (p instanceof Link) {
        if (p.isOrthogonal) ortho.push(p);
        if (p.validRoute === false) continue;
      }
      const ab = p.actualBounds;
      if (ab.width * scale > diagram._minDrawingLength || ab.height * scale > diagram._minDrawingLength) {
        p.draw(context, diagram);
      } else {
        this.drawSimple(context, p);
      }
    }

    context.globalAlpha = old;
  }

  /**
   * @hidden @internal
   * Draws each Part in this layer that intersects a given rect.
   * No invalidation information is changed with the layer's diagram.
   * This is only used by overviews
   * @this {Layer}
   * @param {IContext} context
   * @param {Rect} allowedBounds
   * @param {Diagram} diagram
   * @param {boolean=} showTemporary
   */
  public drawAllowed(context: IContext, allowedBounds: Rect, diagram: Diagram, showTemporary?: boolean) {
    if (!this.visible || this._opacity === 0) return;
    if (showTemporary === undefined) showTemporary = true;
    if (!showTemporary && this.isTemporary) return;
    const old = this.setOpacity(context);
    const scale = diagram.scale;

    const parts = this._parts._arr;
    const len = parts.length;
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      const ab = p.actualBounds;
      if (ab.intersectsRect(allowedBounds)) {
        if (ab.width * scale > diagram._minDrawingLength || ab.height * scale > diagram._minDrawingLength) {
          p.draw(context, diagram);
        } else {
          this.drawSimple(context, p);
        }
      }
    }

    context.globalAlpha = old;
  }

  /**
   * @hidden @internal
   * Draws only the parts that intersect with the given rectangles
   * @this {Layer}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {Array.<Rect>} rects
   */
  public drawIntersection(context: IContext, diagram: Diagram, rects: Array<Rect>) {
    if (!this.visible || this._opacity === 0) return;
    const old = this.setOpacity(context);

    const ortho = this._orthoLinks;
    ortho.length = 0;
    const scale = diagram.scale;
    const parts = this._parts._arr;
    const len = parts.length;
    const rlen = rects.length;
    for (let i = 0; i < len; i++) {
      const p = parts[i];
      if (!p.getInView() && !p.hasHTMLElement()) continue;
      if (p instanceof Link) {
        if (p.isOrthogonal) ortho.push(p);
        if (p.validRoute === false) continue;
      }
      const ab = p.getDrawnSize(p.actualBounds);
      if (this.intersectsAnyRect(ab, rects, rlen, scale)) {
        if (ab.width * scale > diagram._minDrawingLength || ab.height * scale > diagram._minDrawingLength) {
          p.draw(context, diagram);
        } else {
          this.drawSimple(context, p);
        }
      }
    }

    context.globalAlpha = old;
  }

  /**
   * @hidden @internal
   * Simplistic drawing (gray rectangles) for objects smaller than one pixel
   * @this {Layer}
   * @param {IContext} context
   * @param {Part} p
   */
  private drawSimple(context: IContext, p: Part) {
    const ab = p.actualBounds;
    const nb = p.naturalBounds;
    if (ab.width === 0 || ab.height === 0 || isNaN(ab.x) || isNaN(ab.y)) return;
    if (!p.isVisible()) return;
    const tt = p.transform;
    // areaBackground is in the parent's transform space
    if (p.areaBackground !== null) {
      p.setBrush(context, p.areaBackground, true, true, nb, ab);
      context.fillRect(ab.x, ab.y, ab.width, ab.height);
    }

    if (p.areaBackground === null && p.background === null) {
      p.setBrush(context, 'rgba(0,0,0,0.3)', true, false, nb, ab);
      context.fillRect(ab.x, ab.y, ab.width, ab.height);
    }

    // background is in local space
    if (p.background !== null) {
      context.transform(tt.m11, tt.m12, tt.m21, tt.m22, tt.dx, tt.dy); // checks for identity
      p.setBrush(context, p.background, true, false, nb, ab);
      context.fillRect(0, 0, nb.width / 2, nb.height / 2);
      // for when not saving+clipping+restoring
      if (!tt.isIdentity()) {
        const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
        context.transform(
          tt.m22 * d,
          -tt.m12 * d,
          -tt.m21 * d,
          tt.m11 * d,
          d * (tt.m21 * tt.dy - tt.m22 * tt.dx),
          d * (tt.m12 * tt.dx - tt.m11 * tt.dy));
      }
    }
  }

  /**
   * @hidden @internal
   * Draws only the parts that intersect with the given rectangles
   * @this {Layer}
   * @param {Rect} ab
   * @param {Array.<Rect>} rects
   * @param {number} l int
   * @param {number} scale
   */
  private intersectsAnyRect(ab: Rect, rects: Array<Rect>, l: number, scale: number) {
    // The size of the rectangles is slightly expanded due to the need to clip to integer pixels
    // See the comment in (Diagram.draw) that references (Layer.intersectsAnyRect)
    const posoff = 2 / scale;
    const sizeoff = 4 / scale;
    for (let i = 0; i < l; i++) {
      const r = rects[i];
      if (r.width === 0 || r.height === 0) continue;
      if (ab.intersects(r.x - posoff, r.y - posoff, r.width + sizeoff, r.height + sizeoff)) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * Call this method to notify about a property having changed value.
   * @this {Layer}
   * @param {string} propertyname the name of the property.
   * @param {*} oldval the previous or old value for the property.
   * @param {*} newval the next or new value for the property.
   * @param {*=} oldparam an optional value additionally describing the old value.
   * @param {*=} newparam an optional value additionally describing the new value.
   */
  private raiseChanged(propertyname: string, oldval: any, newval: any, oldparam?: any, newparam?: any) {
    const diagram = this.diagram;
    if (diagram !== null) {
      diagram.raiseChangedEvent(ChangedEvent.Property, propertyname, this, oldval, newval, oldparam, newparam);
    }
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number} index int
   * @param {Part} part
   * @param {boolean} switchingLayers
   * @return {number} int
   */
  public internalAdd(index: number, part: Part, switchingLayers: boolean): number {
    const parts = this._parts;
    // set back-pointer to this Layer
    part.setLayer(this);

    // update internal data structures
    if (index >= parts.count) {  // assume add at end
      index = parts.count;
    } else {
      const p = parts.elt(index);
      if (p === part) return -1;  // already there
    }
    parts.insertAt(index, part);
    part.internalAddProtected(switchingLayers);

    const diagram = this.diagram;
    if (diagram !== null) {
      if (!switchingLayers) {
        diagram.internalAdd(part);
      } else {
        diagram.invalidateDraw(); // full redraw, ?? using AB causes undo/redo issues
      }
    }

    this.sortZOrderIndividual(index, (part));
    return index;
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number} index int
   * @param {Part} part
   * @param {boolean} switchingLayers
   * @return {number} int
   */
  public internalRemove(index: number, part: Part, switchingLayers: boolean): number {
    // A part might have changed layers outside of the undo manager
    // Make sure it is removed from its current layer
    if (!switchingLayers && part.layer !== this && part.layer !== null) {
      return part.layer.internalRemove(index, part, switchingLayers);
    }

    // confirm that the given Part is in fact at the index
    const parts = this._parts;
    if (index < 0 || index >= parts.length) {  // unknown index
      index = parts.indexOf(part);
      if (index < 0) return -1;
    } else {  // known index, but check to be sure
      const p = parts.elt(index);
      if (p !== part) {  // not there?
        index = parts.indexOf(part);
        if (index < 0) return -1;  // really not there
      }
    }

    // update internal data structures
    part.internalRemoveProtected(switchingLayers);
    parts.removeAt(index);

    const diagram = this.diagram;
    if (diagram !== null) {
      if (!switchingLayers) {
        diagram.internalRemove(part);
      } else {
        diagram.invalidateDraw(); // full redraw, ?? using AB causes undo/redo issues
      }
    }
    // remove back-pointer to its Layer
    part.setLayer(null);

    return index;
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number} i can be -1 if the index of N in this layer is not known
   * @param {Part} p
   */
  public sortZOrderIndividual(i: number, p: Part) {
    i = this.moveToZOrder(i, p);
    if (p instanceof Group) this.ensureGroupsBehind(i, p);
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number} i
   * @param {Part} p
   */
  private moveToZOrder(i: number, p: Part) {
    const zo = p.zOrder;
    if (isNaN(zo)) return i;
    // move node to an index consistent with its ZOrder

    const list = this._parts;
    const count = list.count;

    if (count <= 1) return i;
    if (i < 0) i = list.indexOf(p);
    if (i < 0) return -1;

    let j = i - 1;
    // skip over previous nodes with undetermined ZOrder
    let jzo = NaN;
    while (j >= 0) {
      jzo = list.elt(j).zOrder;
      if (!isNaN(jzo)) break;
      j--;
    }

    let k = i + 1;
    // skip over following nodes with undetermined ZOrder
    let kzo = NaN;
    while (k < count) {
      kzo = list.elt(k).zOrder;
      if (!isNaN(kzo)) break;
      k++;
    }

    // which direction to go?
    if (!isNaN(jzo) && jzo > zo) {
      // gotta move backward
      while (true) {
        if (j === -1 || jzo <= zo) {
          j++;  // insert after first node found with ZOrder <= p.zOrder
          if (j === i) return i;
          list.removeAt(i);
          list.insertAt(j, p);
          return j;
        }
        // skip over nodes with undetermined ZOrder
        jzo = NaN;
        while (--j >= 0) {
          jzo = list.elt(j).zOrder;
          if (!isNaN(jzo)) break;
        }
      }
    } else if (!isNaN(kzo) && kzo < zo) {
      // gotta move forward
      while (true) {
        if (k === count || kzo >= zo) {
          k--;  // insert before first node found with ZOrder >= p.zOrder
          if (k === i) return i;
          list.removeAt(i);
          list.insertAt(k, p);
          return k;
        }
        // skip over nodes with undetermined ZOrder
        kzo = NaN;
        while (++k < count) {
          kzo = (list.elt(k)).zOrder;
          if (!isNaN(kzo)) break;
        }
      }
    }  // otherwise it's ZOrder is consistent with the adjacent nodes (skipping NaN's)

    return i;
  }

  /**
   * @hidden @internal
   * @this {Layer}
   * @param {number} idx
   * @param {Group} g
   */
  private ensureGroupsBehind(idx: number, g: Group) {
    // now ensure a Group is behind its members, unless its ZOrder is specified
    if (g === null || !isNaN(g.zOrder)) return;
    this.moveGroupBehind(idx, g);
    const cg = g.containingGroup; // foreach (Group cg in g.ContainingGroups) ?
    if (cg !== null) {
      this.sortZOrderIndividual(-1, cg);
    }
  }

  /**
   * @hidden @internal
   * Make sure the given Group is behind all of its immediate members.
   * Assumes the Group is also in this Layer.
   * @this {Layer}
   * @param {number} idx index of sg
   * @param {Group} sg
   */
  private moveGroupBehind(idx: number, sg: Group) {
    if (sg.memberParts.count === 0) return;  // and has at least one member
    let min = -1;  // index of first member of SG
    const arr = this._parts._arr;
    const len = arr.length;
    for (let i = 0; i < len; i++) {
      const p = arr[i];
      if (p === sg) {
        idx = i;  // found index of SG
        if (min >= 0) break;
      }
      if (min < 0 && p.containingGroup === sg) {
        min = i;  // found index of first member of SG
        if (idx >= 0) break;
      }
    }
    // If no members of the group are found in this layer, do nothing
    if (min < 0) return;
    // if some member is behind the group, change index of SG to be at MIN
    if (min < idx) {
      const parts = this._parts;
      parts.removeAt(idx);
      parts.insertAt(min, sg);
    }
  }

  /**
   * @hidden @internal
   * Remove all Parts from this layer.
   * @this {Layer}
   */
  public clear() {
    const parts = this._parts.toArray();
    const len = parts.length;
    for (let i = 0; i < len; i++) {
      parts[i].setInView(false);
      this.internalRemove(-1, parts[i], false);
    }
    this._orthoLinks.length = 0;
  }

  /**
   * This read-only property returns an iterator for this Layer's {@link Part}s.
   * The Parts can be {@link Node}s, {@link Link}s, {@link Group}s, {@link Adornment}s, or simple {@link Part}s.
   */
  /** @export */ get parts(): Iterator<Part> { return this._parts.iterator; }

  /**
   * This read-only property returns a backwards iterator for this Layer's {@link Part}s,
   * for iterating over the parts in reverse order.
   * The Parts can be {@link Node}s, {@link Link}s, {@link Group}s, {@link Adornment}s, or simple {@link Part}s.
   */
  /** @export */ get partsBackwards(): Iterator<Part> { return this._parts.iteratorBackwards; }

  /**
   * This read-only property returns the {@link Diagram} that is using this Layer.
   */
  /** @export */ get diagram(): Diagram | null { return this._diagram; }

  /**
   * Gets or sets the name for this layer.
   * The initial value is an empty string, which is also the name of the default layer.
   * The name should be unique among the diagram's {@link Diagram#layers}.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) {
    Util.checkPrimitive(value, 'string', Layer, 'name');
    const old = this._name;
    if (old !== value) {
      const d = this.diagram;
      if (d !== null) {
        // don't allow changing name of default layer
        if (old === '') Util.throwError('Cannot rename default Layer to: ' + value);
        const lit = d.layers;
        while (lit.next()) {
          const lay = lit.value;
          if (lay.name === value) {
            Util.throwError('Layer.name is already present in this diagram: ' + value);
          }
        }
      }
      this._name = value;
      this.raiseChanged('name', old, value);
      // Change the name of all parts in this layer to match
      const it = this._parts.iterator;
      while (it.next()) {
        const part = it.value;
        part.layerName = this._name;
      }
    }
  }

  /**
   * Gets or sets the opacity for all parts in this layer.
   * The value must be between 0.0 (fully transparent) and 1.0 (no additional transparency).
   * This value is multiplicative with any existing transparency,
   * for instance from a {@link Brush} or image transparency.
   * The default value is 1.
   *
   * This property, unlike {@link #visible}, does not change whether any objects are found by the "find..." methods.
   * @since 1.1
   * @see {@link GraphObject#opacity}
   * @see {@link Diagram#opacity}
   */
  /** @export */ get opacity(): number { return this._opacity; }
  /** @export */ set opacity(value: number) {
    const old = this._opacity;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Layer, 'opacity');
      if (value < 0 || value > 1) Util.throwRangeError(value, '0 <= value <= 1', Layer, 'opacity');
      this._opacity = value;
      this.raiseChanged('opacity', old, value);
      // and make sure everything is re-drawn
      const diagram = this.diagram;
      if (diagram !== null) {
        diagram.invalidateDraw(); // full redraw needed
      }
    }
  }

  /**
   * Gets or sets whether the objects in this layer are considered temporary.
   *
   * Parts in temporary layers are not selectable,
   * and changes to Parts in temporary layers are not recorded in the {@link UndoManager}.
   * Objects in temporary layers do not receive click events unless you set their {@link GraphObject#isActionable} to true.
   *
   * Temporary layers are excluded from bounds calculations, with the exception of the "Tool" layer,
   * so that temporary objects created while dragging are included in the bounds.
   *
   * Default value is false.
   */
  /** @export */ get isTemporary(): boolean { return this._isTemporary; }
  /** @export */ set isTemporary(value: boolean) {
    const old = this._isTemporary;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'isTemporary');
      this._isTemporary = value;
      this.raiseChanged('isTemporary', old, value);
    }
  }

  /**
   * Gets or sets whether the user may view any of the objects in this layer.
   *
   * The default value is true -- all visible Parts are drawn.
   * When this property is false, all of the "find..." methods will fail to find parts that are in this layer.
   * @see {@link GraphObject#visible}
   */
  /** @export */ get visible(): boolean { return this._visible; }
  /** @export */ set visible(value: boolean) {
    const old = this._visible;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'visible');
      this._visible = value;
      this.raiseChanged('visible', old, value);
      // update adornments for all parts in this layer
      const it = this._parts.iterator;
      while (it.next()) {
        const part = it.value;
        part.madeVisible(value);
      }
      // and make sure everything is re-drawn
      const diagram = this.diagram;
      if (diagram !== null) {
        diagram.invalidateDraw(); // full redraw needed
      }
    }
  }

  /**
   * Gets or sets whether methods such as {@link #findObjectAt} find any of the objects in this layer.
   *
   * The default value is true.
   * When this property is false, all of the "find..." methods will fail to find parts that are in this layer.
   *
   * Note that setting pickable to false does not prevent users from selecting nodes.
   * It does prevent them from selecting nodes by clicking on them,
   * but does not prevent selection through other mechanisms such as the {@link DragSelectingTool} or
   * {@link CommandHandler#selectAll} or calls to {@link Diagram#select}.
   *
   * You can control whether individual GraphObjects are "hittable" by setting {@link GraphObject#pickable}.
   * @since 1.2
   * @see {@link GraphObject#pickable}
   */
  /** @export */ get pickable(): boolean { return this._pickable; }
  /** @export */ set pickable(value: boolean) {
    const old = this._pickable;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'pickable');
      this._pickable = value;
      this.raiseChanged('pickable', old, value);
    }
  }

  /**
   * @hidden @internal
   * Predicate which determines whether or not a layer is included in the documentBounds
   * Default value is null, which resolves to the predicate: function() { return !this.isTemporary; }
   * @since 1.4
   */
  /** @export */ get isBoundsIncluded(): boolean { return this._isBoundsIncluded; }
  /** @export */ set isBoundsIncluded(value: boolean) {
    const old = this._isBoundsIncluded;
    if (old !== value) {
      this._isBoundsIncluded = value;
      if (this.diagram !== null) this.diagram.invalidateDocumentBounds();
    }
  }

  // Permissions

  /**
   * Gets or sets whether the user may copy objects in this layer.
   * The initial value is true.
   */
  /** @export */ get allowCopy(): boolean { return this._allowCopy; }
  /** @export */ set allowCopy(value: boolean) {
    const old = this._allowCopy;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowCopy');
      this._allowCopy = value;
      this.raiseChanged('allowCopy', old, value);
    }
  }

  /**
   * Gets or sets whether the user may delete objects in this layer.
   * The initial value is true.
   */
  /** @export */ get allowDelete(): boolean { return this._allowDelete; }
  /** @export */ set allowDelete(value: boolean) {
    const old = this._allowDelete;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowDelete');
      this._allowDelete = value;
      this.raiseChanged('allowDelete', old, value);
    }
  }

  /**
   * Gets or sets whether the user may do in-place text editing in this layer.
   * The initial value is true.
   */
  /** @export */ get allowTextEdit(): boolean { return this._allowTextEdit; }
  /** @export */ set allowTextEdit(value: boolean) {
    const old = this._allowTextEdit;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowTextEdit');
      this._allowTextEdit = value;
      this.raiseChanged('allowTextEdit', old, value);
    }
  }

  /**
   * Gets or sets whether the user may group parts together in this layer.
   * The initial value is true.
   */
  /** @export */ get allowGroup(): boolean { return this._allowGroup; }
  /** @export */ set allowGroup(value: boolean) {
    const old = this._allowGroup;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowGroup');
      this._allowGroup = value;
      this.raiseChanged('allowGroup', old, value);
    }
  }

  /**
   * Gets or sets whether the user may ungroup existing groups in this layer.
   * The initial value is true.
   */
  /** @export */ get allowUngroup(): boolean { return this._allowUngroup; }
  /** @export */ set allowUngroup(value: boolean) {
    const old = this._allowUngroup;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowUngroup');
      this._allowUngroup = value;
      this.raiseChanged('allowUngroup', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw new links in this layer.
   * The initial value is true.
   */
  /** @export */ get allowLink(): boolean { return this._allowLink; }
  /** @export */ set allowLink(value: boolean) {
    const old = this._allowLink;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowLink');
      this._allowLink = value;
      this.raiseChanged('allowLink', old, value);
    }
  }

  /**
   * Gets or sets whether the user may reconnect existing links in this layer.
   * The initial value is true.
   */
  /** @export */ get allowRelink(): boolean { return this._allowRelink; }
  /** @export */ set allowRelink(value: boolean) {
    const old = this._allowRelink;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowRelink');
      this._allowRelink = value;
      this.raiseChanged('allowRelink', old, value);
    }
  }

  /**
   * Gets or sets whether the user may move objects in this layer.
   * The initial value is true.
   */
  /** @export */ get allowMove(): boolean { return this._allowMove; }
  /** @export */ set allowMove(value: boolean) {
    const old = this._allowMove;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowMove');
      this._allowMove = value;
      this.raiseChanged('allowMove', old, value);
    }
  }

  /**
   * Gets or sets whether the user may reshape parts in this layer.
   * The initial value is true.
   */
  /** @export */ get allowReshape(): boolean { return this._allowReshape; }
  /** @export */ set allowReshape(value: boolean) {
    const old = this._allowReshape;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowReshape');
      this._allowReshape = value;
      this.raiseChanged('allowReshape', old, value);
    }
  }

  /**
   * Gets or sets whether the user may resize parts in this layer.
   * The initial value is true.
   */
  /** @export */ get allowResize(): boolean { return this._allowResize; }
  /** @export */ set allowResize(value: boolean) {
    const old = this._allowResize;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowResize');
      this._allowResize = value;
      this.raiseChanged('allowResize', old, value);
    }
  }

  /**
   * Gets or sets whether the user may rotate parts in this layer.
   * The initial value is true.
   */
  /** @export */ get allowRotate(): boolean { return this._allowRotate; }
  /** @export */ set allowRotate(value: boolean) {
    const old = this._allowRotate;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowRotate');
      this._allowRotate = value;
      this.raiseChanged('allowRotate', old, value);
    }
  }

  /**
   * Gets or sets whether the user may select objects in this layer.
   * The initial value is true.
   */
  /** @export */ get allowSelect(): boolean { return this._allowSelect; }
  /** @export */ set allowSelect(value: boolean) {
    const old = this._allowSelect;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Layer, 'allowSelect');
      this._allowSelect = value;
      this.raiseChanged('allowSelect', old, value);
    }
  }

} // end class
Layer['className'] = 'Layer';

// import { ChangedEvent } from './changedEvent';
// import { List, Set } from './collections';
// import { Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';
// import { Adornment, Group, Part } from './parts';
// import { Node } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { IContext } from './surfaceInterfaces';
// import { Util } from './util';

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Diagram
*/

// import { BrushLike } from './brush';
// import { KeyValuePair, List, Map, Set } from './collections';
// import { EnumValue } from './enumValue';
// import { Geo } from './geo';
// import { ObjectData, root } from './root';
// import { Transform } from './transform';
// import { Util } from './util';

/**
 * A Diagram is associated with an HTML DIV element. Constructing a Diagram creates
 * an HTML Canvas element which it places inside of the given DIV element, in addition to several helper DIVs.
 * **GoJS** will manage the contents of this DIV -- you should not modify the contents of the DIV,
 * although you may style the given DIV (background, border, etc) and position and size it as needed.
 *
 * Minimal Diagram construction looks like this. HTML:
 * ```html
 * <div id="myDiagramDiv" style="border: solid 1px black; width:400px; height:400px"></div>
 * ```
 *
 * JavaScript:
 * ```js
 * var $ = go.GraphObject.make;  // for conciseness
 *
 * myDiagram = $(go.Diagram, "myDiagramDiv",  // create a Diagram for the DIV HTML element
 *               {
 *                 "undoManager.isEnabled": true  // enable undo & redo
 *               });
 * ```
 *
 * The diagram will draw onto an HTML Canvas element, created inside the Diagram DIV.
 *
 * Each Diagram holds a set of {@link Layer}s each of which holds some number of {@link Part}s
 * such as {@link Node}s and {@link Link}s.
 * Each {@link Part} consists of {@link GraphObject}s such as {@link TextBlock}s and {@link Shape}s
 * and {@link Panel}s holding yet more GraphObjects.
 *
 * A Diagram and its Parts provide the visual representation of a {@link Model} that holds JavaScript
 * data objects for the nodes and the links.
 * The model provides the way to recognize the relationships between the data.
 * <p class="boxrun">
 * Two Diagrams can display and manipulate the same Model. (<a href="../../samples/updateDemo.html">Example</a>)
 *
 * A diagram will automatically create {@link Node}s and {@link Link}s corresponding to the model data.
 * The diagram has a number of named templates it uses to create the actual parts:
 * {@link #nodeTemplateMap}, {@link #groupTemplateMap}, and {@link #linkTemplateMap}.
 * Each template may have some data {@link Binding}s that set the part's {@link GraphObject}s' properties
 * based on the value of properties of the data.
 *
 * A simple Node template and Model data (both nodes and links) may look like this:
 *
 * ```js
 * var $ = go.GraphObject.make;  // for conciseness
 *
 * // define a simple Node template
 * myDiagram.nodeTemplate =
 *   $(go.Node, "Auto",  // the Shape will go around the TextBlock
 *     $(go.Shape, "RoundedRectangle",
 *       // Shape.fill is bound to Node.data.color
 *       new go.Binding("fill", "color")),
 *     $(go.TextBlock,
 *       { margin: 3 },  // some room around the text
 *       // TextBlock.text is bound to Node.data.key
 *       new go.Binding("text", "key"))
 *   );
 *
 * // create the model data that will be represented by Nodes and Links
 * myDiagram.model = new go.GraphLinksModel(
 * [
 *   { key: "Alpha", color: "lightblue" },
 *   { key: "Beta", color: "orange" },
 *   { key: "Gamma", color: "lightgreen" },
 *   { key: "Delta", color: "pink" }
 * ],
 * [
 *   { from: "Alpha", to: "Beta" },
 *   { from: "Alpha", to: "Gamma" },
 *   { from: "Beta", to: "Beta" },
 *   { from: "Gamma", to: "Delta" },
 *   { from: "Delta", to: "Alpha" }
 * ]);
 * ```
 * <p class="boxrun">
 * The above code is used to make the <a href="../../samples/minimal.html">Minimal sample</a>, a simple example of
 * creating a Diagram and setting its model.
 * <p class="boxread">
 * Read about models on the <a href="../../intro/usingModels.html">Using Models</a> page in the introduction.
 * A diagram is responsible for scrolling ({@link #position}) and zooming ({@link #scale}) all of the parts that it shows.
 * Each {@link Part} occupies some area given by its {@link GraphObject#actualBounds}.
 *
 * The union of all of the parts' bounds constitutes the {@link #documentBounds}.
 * The document bounds determines the area that the diagram can be scrolled to.
 * There are several properties that you can set, such as {@link #initialContentAlignment},
 * that control the initial size and position of the diagram contents.
 *
 * At any later time you can also explicitly set the {@link #position} and/or {@link #scale} to
 * get the appearance that you want.  But you may find it easier to call methods to get the desired effect.
 * For example, if you want to make a particular Node be centered in the viewport,
 * call either {@link #centerRect} or {@link #scrollToRect} with the Node's {@link GraphObject#actualBounds},
 * depending on whether or not you want the view to be scrolled if the node is already in view.
 * <p class="boxread">
 * Read in the Introduction about <a href="../../intro/viewport.html">Viewports</a>
 * and the <a href="../../intro/initialView.html">Initial Viewport</a>.
 * You can have the diagram perform automatic layouts of its nodes and links by setting
 * {@link #layout} to an instance of the {@link Layout} subclass of your choice.
 * The default {@link #layout} is an instance of the {@link Layout} base class that ignores links and
 * only positions {@link Node}s that do not have a location.
 * This default layout will allow you to programmatically position nodes (including by loading
 * from a database) and will also allow the user to manually position nodes using the {@link DraggingTool}.
 *
 * If you do supply a particular layout as the {@link #layout}, you can control which {@link Part}s it operates
 * on by setting {@link Part#isLayoutPositioned}.
 * Normally, of course, it works on all top-level nodes and links.
 * The layout is performed both after the model is first loaded as well as after any part is added or removed
 * or changes visibility or size.
 * You can disable the initial layout by setting {@link Layout#isInitial} to false.
 * You can disable later automatic layouts by setting {@link Layout#isOngoing} to false.
 * <p class="boxread">
 * See the <a href="../../intro/layouts.html">Layouts</a> page in the Introduction for a summary of layout behavior.
 *
 * A diagram maintains a collection of selected parts, the {@link Diagram#selection}.
 * To select a Part you set its {@link Part#isSelected} property to true.
 *
 * There are many properties, named "allow...", that control what operations the user
 * may perform on the parts in the diagram.  These correspond to the same named
 * properties on {@link Layer} that govern the behavior for those parts in a particular layer.
 * Furthermore for some of these properties there are corresponding properties on
 * {@link Part}, named "...able", that govern the behavior for that individual part.
 * For example, the {@link #allowCopy} property corresponds to {@link Layer#allowCopy} and
 * to the property {@link Part#copyable}.
 * The {@link Part#canCopy} predicate is false if any of these properties is false.
 * <p class="boxread">
 * See the <a href="../../intro/permissions.html">Permissions</a> page for a more thorough discussion.
 *
 * The {@link #commandHandler} implements various standard commands,
 * such as the {@link CommandHandler#deleteSelection} method and the
 * {@link CommandHandler#canDeleteSelection} predicate.
 * <p class="boxread">
 * See the <a href="../../intro/commands.html">Commands</a> page for a listing of keyboard commands and
 * the use of commands in general.
 *
 * The diagram supports modular behavior for mouse events by implementing "tools".
 * All mouse and keyboard events are represented by {@link InputEvent}s and redirected
 * to the {@link #currentTool}.
 * The default tool is an instance of {@link ToolManager} which keeps three lists of mode-less tools:
 * {@link ToolManager#mouseDownTools}, {@link ToolManager#mouseMoveTools}, and {@link ToolManager#mouseUpTools}.
 * The ToolManager searches these lists when a mouse event happens to find the first tool that can run.
 * It then makes that tool the new {@link #currentTool}, where it can continue to process input events.
 * When the tool is done, it stops itself, causing the {@link #defaultTool} to be the new {@link #currentTool}.
 *
 * Mouse-down tools include:
 *   - {@link ToolManager#actionTool}, to support objects like "buttons"
 *   - {@link ToolManager#relinkingTool}, to reconnect an existing link
 *   - {@link ToolManager#linkReshapingTool}, to modify the route of an existing link
 *   - {@link ToolManager#rotatingTool}, to change the angle of an object
 *   - {@link ToolManager#resizingTool}, to change the size of an object
 *
 * Mouse-move tools include:
 *   - {@link ToolManager#linkingTool}, to draw a new link
 *   - {@link ToolManager#draggingTool}, to move or copy the selection
 *   - {@link ToolManager#dragSelectingTool}, to select parts within a rectangular area
 *   - {@link ToolManager#panningTool}, to pan the diagram
 *
 * Mouse-up tools include:
 *   - {@link ToolManager#contextMenuTool}, to manage context menus
 *   - {@link ToolManager#textEditingTool}, to support in-place text editing
 *   - {@link ToolManager#clickCreatingTool}, to create new parts where the user clicks
 *   - {@link ToolManager#clickSelectingTool}, to select parts
 *
 * You can also run a tool in a modal fashion by explicitly setting {@link #currentTool}.
 * That tool will keep running until some code replaces the {@link #currentTool}.
 * This normally happens when the current tool calls {@link Tool#stopTool}, such as on a mouse-up event.
 * <p class="boxread">
 * See the <a href="../../intro/tools.html">Tools</a> page for a listing of predefined tools and how they operate.
 *
 * A diagram raises various {@link DiagramEvent}s when interesting things happen that may have affected the whole diagram.
 * See the documentation for {@link DiagramEvent} for a complete listing.
 *
 * When you need to display multiple Models, but not at the same time, you can do so by using only one Diagram and setting the {@link #model} to a different one.
 * You can also have two Diagrams share a DIV by swapping the {@link #div} to `null` on one Diagram and setting it on the other.
 * When permanently removing a Diagram,t o clear any memory used,
 * set the {@link #div} to `null` and remove all references to the Diagram. These scenarios are discussed more on the
 * <a href="../../intro/replacingDeleting.html">Replacing Diagrams and Models</a> intro page.
 * @unrestricted
 * @category Diagram
 */
export class Diagram {
  /** @hidden @internal */ public static className = 'Diagram';
  /** @hidden @internal */ private _animationManager: AnimationManager;
  /** @hidden @internal */ private _scrollBarBreadth: number;
  /** @hidden @internal */ private _scrollBarBreadthSet: boolean; // whether setScrollWidth has ever been called
  /** @hidden @internal */ private _rtlScrollType: string;
  /** @hidden @internal */ private _renderingHints: PropertyCollection;
  /** @hidden @internal */ private _inputOptions: PropertyCollection;
  /** @hidden @internal */ private _initialPosition: Point;
  /** @hidden @internal */ private _initialScale: number;
  /** @hidden @internal */ private _minScale: number;
  /** @hidden @internal */ private _maxScale: number;
  /** @hidden @internal */ private _zoomPoint: Point;
  /** @hidden @internal */ private _fixedBounds: Rect;
  /** @hidden @internal */ private _scrollMargin: Margin;
  /** @hidden @internal */ private _scrollMode: EnumValue;
  /** @hidden @internal */ private _scrollsPageOnFocus: boolean;
  /** @hidden @internal */ private _positionComputation: ((thisDiagram: Diagram, newPosition: Point) => Point) | null;
  /** @hidden @internal */ private _scaleComputation: ((thisDiagram: Diagram, newScale: number) => number) | null;
  /** @hidden @internal */ private _autoScale: EnumValue;
  /** @hidden @internal */ private _contentAlignment: Spot;
  /** @hidden @internal */ private _initialAutoScale: EnumValue;
  /** @hidden @internal */ private _initialContentAlignment: Spot;
  /** @hidden @internal */ private _initialDocumentSpot: Spot;
  /** @hidden @internal */ private _initialViewportSpot: Spot;
  /** @hidden @internal */ private _invalidCanvasSize: boolean;
  /** @hidden @internal */ private _invalidParts: Set<Part>;
  /** @hidden @internal */ private _autoScrollInterval: number;
  /** @hidden @internal */ private _autoScrollTimer: number;
  /** @hidden @internal */ private _autoScrollRegion: Margin;
  /** @hidden @internal */ private _doNonRealtimeLayouts: boolean;
  /** @hidden @internal */ private _firstUpdate: boolean;
  /** @hidden @internal */ private _changedListeners: List<((e: ChangedEvent) => void)>;
  /** @hidden @internal */ private _modelChangedListeners: List<((e: ChangedEvent) => void)>;
  /** @hidden @internal */ private _diagramListeners: Map<string, List<((e: DiagramEvent) => void)>>;
  /** @hidden @internal */ private _diagramEventNames: Map<string, string>;
  /** @hidden @internal */ private _allowsPointers: boolean;
  /** @hidden @internal */ private _isTreePathToChildren: boolean;
  /** @hidden @internal */ private _treeCollapsePolicy: EnumValue;
  /** @hidden @internal */ private _isCollapsingExpanding: boolean;
  /** @hidden @internal */ private _validCycle: EnumValue;
  /** @hidden @internal */ private _toolManager: ToolManager;
  /** @hidden @internal */ private _defaultTool: Tool;
  /** @hidden @internal */ private _currentTool: Tool;
  /** @hidden @internal */ private _currentCursor: string;
  /** @hidden @internal */ private _defaultCursor: string;
  /** @hidden @internal */ private _click: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _doubleClick: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _contextClick: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseOver: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseHover: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseHold: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseDragOver: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseDrop: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _handlesDragDropForTopLevelParts: boolean;
  /** @hidden @internal */ private _mouseEnter: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _mouseLeave: ((e: InputEvent) => void) | null;
  /** @hidden @internal */ private _toolTip: Adornment | HTMLInfo | null;
  /** @hidden @internal */ private _contextMenu: Adornment | HTMLInfo | null;
  /** @hidden @internal */ private _mouseOverQueued: boolean;
  /** @hidden @internal */ private _commandHandler: ICommandHandler;
  /** @hidden @internal */ private _skipsUndoManager: boolean;
  /** @hidden @internal */ private _skipsModelSourceBindings: boolean;
  /** @hidden @internal */ private _skipsBoundsUpdates: boolean;
  /** @hidden @internal */ private _delaysLayout: boolean;
  /** @hidden @internal */ private _isUpdatingModel: boolean;
  /** @hidden @internal */ private _model: Model;
  /** @hidden @internal */ private _modelBindingHandler: ((e: ChangedEvent) => void);
  /** @hidden @internal */ private _modelChangedHandler: ((e: ChangedEvent) => void);
  /** @hidden @internal */ private _wasModified: boolean;
  /** @hidden @internal */ private _historyIndex: number;
  /** @hidden @internal */ private _isReadOnly: boolean;
  /** @hidden @internal */ private _isEnabled: boolean;
  /** @hidden @internal */ private _allowClipboard: boolean;
  /** @hidden @internal */ private _allowCopy: boolean;
  /** @hidden @internal */ private _allowDelete: boolean;
  /** @hidden @internal */ private _allowDragOut: boolean;
  /** @hidden @internal */ private _allowDrop: boolean;
  /** @hidden @internal */ private _allowTextEdit: boolean;
  /** @hidden @internal */ private _allowGroup: boolean;
  /** @hidden @internal */ private _allowUngroup: boolean;
  /** @hidden @internal */ private _allowInsert: boolean;
  /** @hidden @internal */ private _allowLink: boolean;
  /** @hidden @internal */ private _allowRelink: boolean;
  /** @hidden @internal */ private _allowMove: boolean;
  /** @hidden @internal */ private _allowReshape: boolean;
  /** @hidden @internal */ private _allowResize: boolean;
  /** @hidden @internal */ private _allowRotate: boolean;
  /** @hidden @internal */ private _allowSelect: boolean;
  /** @hidden @internal */ private _allowUndo: boolean;
  /** @hidden @internal */ private _allowZoom: boolean;
  /** @hidden @internal */ private _isMouseCaptured: boolean;
  /** @hidden @internal */ protected _hasHorizontalScrollbar: boolean;
  /** @hidden @internal */ protected _hasVerticalScrollbar: boolean;
  /** @hidden @internal */ private _allowHorizontalScroll: boolean;
  /** @hidden @internal */ private _allowVerticalScroll: boolean;
  /** @hidden @internal */ private _scrollHorizontalLineChange: number;
  /** @hidden @internal */ private _scrollVerticalLineChange: number;
  /** @hidden @internal */ private _scrollBarMoving: boolean;
  /** @hidden @internal */ private _scrollBarsModified: boolean;
  /** @hidden @internal */ private _scrollTop: number;
  /** @hidden @internal */ private _scrollLeft: number;
  /** @hidden @internal */ private _padding: Margin;
  /** @hidden @internal */ private _selection: Set<Part>;
  /** @hidden @internal */ private _maxSelectionCount: number;
  /** @hidden @internal */ private _highlighteds: Set<Part>;
  /** @hidden @internal */ private _invalidScrollBars: boolean;
  /** @hidden @internal */ private _hasHScroll: boolean;
  /** @hidden @internal */ private _hasVScroll: boolean;
  /** @hidden @internal */ private _layout: Layout;
  /** @hidden @internal */ private _isInitialLayoutCompleted: boolean;
  /** @hidden @internal */ private _loadedRouteLinks: Set<Part>;
  /** @hidden @internal */ private _positions: PositionArray;
  /** @hidden @internal */ private _viewportBounds: Rect;
  /** @hidden @internal */ private _viewSize: Size;
  /** @hidden @internal */ private _documentBounds: Rect;
  /** @hidden @internal */ private _isVirtualized: boolean;
  /** @hidden @internal */ private _nodeTemplateMap: Map<string, Part>;
  /** @hidden @internal */ private _linkTemplateMap: Map<string, Link>;
  /** @hidden @internal */ private _groupTemplateMap: Map<string, Group>;
  /** @hidden @internal */ public _defaultNodeTemplate: Node;
  /** @hidden @internal */ public _defaultLinkTemplate: Link;
  /** @hidden @internal */ private _nodeSelectionAdornmentTemplate: Adornment;
  /** @hidden @internal */ private _groupSelectionAdornmentTemplate: Adornment;
  /** @hidden @internal */ private _linkSelectionAdornmentTemplate: Adornment;
  /** @hidden @internal */ private _grid: Panel;
  /** @hidden @internal */ private _gridWasVisible: boolean;
  /** @hidden @internal */ private _autoScrollPoint: Point;
  /** @hidden @internal */ private _isModified: boolean;
  /** @hidden @internal */ public _renderers: Map<string, ISurface>;
  /** @hidden @internal */ private _firstInput: InputEvent;
  /** @hidden @internal */ private _previousInput: InputEvent;
  /** @hidden @internal */ private _lastInput: InputEvent;
  /** @hidden @internal */ private _partManager: PartManager;
  /** @hidden @internal */ private _defaultScale: number;
  /** @hidden @internal */ protected _isAligning: boolean;
  /** @hidden @internal */ protected _transform: Transform;
  /** @hidden @internal */ protected _invalidDraw: boolean;
  /** @hidden @internal */ public _updateQueued: boolean;
  /** @hidden @internal */ protected _viewportBoundsChangedInfo: { scale: number, position: Point, bounds: Rect, isScroll: boolean };
  /** @hidden @internal */ protected _isUpdating: boolean;
  /** @hidden @internal */ public _defaultGroupTemplate: Group;
  /** @hidden @internal */ public _picturesMap: Map<string, Array<Picture>>;
  /** @hidden @internal */ public _anyLayoutInvalid: boolean;
  /** @hidden @internal */ public _topLevelGroups: Set<Group>;
  /** @hidden @internal */ public _invalidDocBounds: boolean;
  /** @hidden @internal */ public _invalidViewportObjects: boolean;
  /** @hidden @internal */ public _invalidLinkBundles: Set<LinkBundle>;
  /** @hidden @internal */ public _drawTainted: boolean;
  /** @hidden @internal */ public _isDrawingForeign: boolean;
  /** @hidden @internal */ public _picturesToInvalidate: List<Picture>;
  /** @hidden @internal */ public _layers: List<Layer>;
  /** @hidden @internal */ public _delayedGeometries: Map<Link, Rect>;
  /** @hidden @internal */ public _routingFromPort: Shape | null;
  /** @hidden @internal */ public _routingFromNode: Node | null;
  /** @hidden @internal */ public _routingToPort: Shape | null;
  /** @hidden @internal */ public _routingToNode: Node | null;
  /** @hidden @internal */ public _position: Point;
  /** @hidden @internal */ public _scale: number;
  /** @hidden @internal */ public _isRTL: boolean;
  /** @hidden @internal */ public _canvasWidth: number;
  /** @hidden @internal */ public _canvasHeight: number;
  /** @hidden @internal */ public _canvas: CanvasSurface | SVGSurface;
  /** @hidden @internal */ public _tapHeld: boolean;
  /** @hidden @internal */ public _pointers: { [index: string]: PointerEvent };
  /** @hidden @internal */ public _pointers2: Array<PointerEvent>;
  /** @hidden @internal */ public _isMouseOverDiagram: boolean;
  /** @hidden @internal */ public _scrollParent: HTMLElement | null;
  /** @hidden @internal */ public _scrollHelper: HTMLElement | null;
  /** @hidden @internal */ public _drawImage: Function;
  /** @hidden @internal */ public _drawCollectionCanvas: CanvasSurface;
  /** @hidden @internal */ public _drawCollectionCtx: CanvasSurfaceContext;
  /** @hidden @internal */ public _pixelRatio: number;
  /** @hidden @internal */ public _tempPixelRatio: number | null;
  /** @hidden @internal */ public _minDrawingLength: number;
  /** @hidden @internal */ public _lastTouchCount: number;
  /** @hidden @internal */ public _lastTouchStamp: number;
  /** @hidden @internal */ public _lastTouchpt: Point;
  /** @hidden @internal */ public _touchDoubleClickDelay: number;
  /** @hidden @internal */ public _simulatedDetected: Event;
  /** @hidden @internal */ public _isTouchOngoing: boolean;
  /** @hidden @internal */ public touchStart: EventListener | null;
  /** @hidden @internal */ public touchMove: EventListener | null;
  /** @hidden @internal */ public touchEnd: EventListener | null;
  /** @hidden @internal */ public windowResize: ((e: Event) => void) | null;
  /** @hidden @internal */ public mouseMove: ((e: MouseEvent) => void) | null;
  /** @hidden @internal */ public mouseDown: ((e: MouseEvent) => void) | null;
  /** @hidden @internal */ public mouseUp: ((e: MouseEvent) => void) | null;
  /** @hidden @internal */ public mouseWheel: ((e: WheelEvent) => void) | null;
  /** @hidden @internal */ public mouseOut: ((e: MouseEvent) => void) | null;
  /** @hidden @internal */ public pointerDown: ((e: PointerEvent) => void) | null;
  /** @hidden @internal */ public pointerMove: ((e: PointerEvent) => void) | null;
  /** @hidden @internal */ public pointerUp: ((e: PointerEvent) => void) | null;
  /** @hidden @internal */ public pointerOut: ((e: PointerEvent) => void) | null;
  /** @hidden @internal */ public redrawLater: Function;
  /** @hidden @internal */ public preventDefault: ((e: Event) => void) | null;
  /** @hidden @internal */ public _ctx: CanvasSurfaceContext | SVGContext;
  /** @hidden @internal */ public _div: HTMLDivElement | null;
  /** @hidden @internal */ public _helper: DiagramHelper;
  /** @hidden @internal */ public _origScale: number;
  /** @hidden @internal */ public _startDist: number;
  /** @hidden @internal */ public _currentEnterLeaveObject: GraphObject | null;
  /** @hidden @internal */ public _opacity: number;

  /** @hidden @internal */ private static _last: Diagram = null;

  /** @hidden @internal */ private static _partManagerMap = new Map<string, typeof PartManager>();


  /** @hidden @internal */ private static scrollOuter: HTMLElement = null;
  /** @hidden @internal */ private static scrollInner: HTMLElement = null;

  // Formerly on DraggingTool:
  /** @hidden @internal */ private _suspendedRouting: boolean;
  // Formerly on DraggingTool & used by DraggingTool:
  /** @hidden @internal */ private _draggedLink: Link;
  /** @hidden @internal */ public _defaultDragOptions: DraggingOptions;

  /** @hidden @internal */ private _oldFromPort: GraphObject;
  /** @hidden @internal */ private _oldToPort: GraphObject;
  /** @hidden @internal */ private _movingAvg: number;
  /** @hidden @internal */ private _movingAvgArray: Array<number>;
  /** @hidden @internal */ private _movingAvgFrame: number;
  /** @hidden @internal */ public _svg: boolean; /* NYI, temp for now */

  /** @hidden @internal */ private _focusObject: GraphObject | null;
  /** @hidden @internal */ private _focusViewportPoint: Point;
  /** @hidden @internal */ public _savedPosition: Point;


  /**
   * Construct an empty Diagram for a particular DIV HTML element.
   *
   * You will normally initialize properties of the Diagram that control its appearance and behavior.
   * These properties include:
   *   - templates (such as {@link #nodeTemplate} or {@link #linkSelectionAdornmentTemplate}),
   *   - an automatic layout (if desired, {@link #layout}),
   *   - positioning or scaling (such as {@link #initialContentAlignment} or {@link #initialAutoScale}),
   *   - permissions (such as {@link #isReadOnly} or {@link #allowDelete}), or
   *   - Tool or CommandHandler customization (such as {@link CommandHandler#archetypeGroupData} or
   *     {@link Tool#isEnabled}): `diagram.toolManager.dragSelectingTool.isEnabled = false;`.
   *
   * Then you will need to construct a Model (usually a {@link GraphLinksModel}) for the Diagram,
   * initialize its data by setting its {@link Model#nodeDataArray} and other properties,
   * and then set the diagram's {@link #model}.
   *
   * Finally, if you want to disassociate the Diagram from the HTML Div element, set {@link Diagram#div} to null.
   * If you remove a part of the HTML DOM containing a Div with a Diagram, you will need to
   * set {@link #div} to null in order for the page to recover the memory.
   *
   * It is commonplace to use the static function {@link GraphObject.make} to build a Diagram:
   * ```js
   * var $ = go.GraphObject.make;
   *
   * var diagram =
   *   $(go.Diagram, "myDiagramDiv",
   *     {
   *       allowZoom: false,
   *       "animationManager.isEnabled": false,  // turn off automatic animations
   *       "grid.visible": true,  // display a background grid for the whole diagram
   *       "grid.gridCellSize": new go.Size(20, 20),
   *       // allow double-click in background to create a new node
   *       "clickCreatingTool.archetypeNodeData": { text: "Node" },
   *       // allow Ctrl-G to call the groupSelection command
   *       "commandHandler.archetypeGroupData":
   *         { text: "Group", isGroup: true, color: "blue" },
   *       "commandHandler.copiesTree": true,  // for the copy command
   *       "commandHandler.deletesTree": true, // for the delete command
   *       "toolManager.hoverDelay": 100,  // how quickly tooltips are shown
   *       // mouse wheel zooms instead of scrolls
   *       "toolManager.mouseWheelBehavior": go.ToolManager.WheelZoom,
   *       "draggingTool.dragsTree": true,  // dragging for both move and copy
   *       "draggingTool.isGridSnapEnabled": true,
   *       layout: $(go.TreeLayout,
   *                 { angle: 90, sorting: go.TreeLayout.SortingAscending }),
   *       "undoManager.isEnabled": true,  // enable undo & redo
   *       // a Changed listener on the Diagram.model
   *       "ModelChanged": function(e) { if (e.isTransactionFinished) saveModel(); }
   *     });
   * ```
   * @param {Element|string=} div A reference to a DIV HTML element or its ID as a string.
   * If no DIV is supplied one will be created in memory. The Diagram's {@link Diagram#div} property
   * can then be set later on.
   */
  constructor(div?: Element | string) {
    if (arguments.length > 1) {
      Util.throwError('Diagram constructor can only take one optional argument, the DIV HTML element or its id.');
    }
    if (!Diagram.wmq2) { // init
      Diagram.wmq3();
      Diagram.wmq2 = true;
    }
    Set.uniqueHash(this);
    Diagram.rememberDiagram(this);
    Util.clearDiagramsTraversed();

    // set to false at end of constructor
    this._isAligning = true;

    this._animationManager = new AnimationManager();
    this._animationManager.setDiagram(this);

    // for now, set scrollBarBreadth for every Diagram
    this._scrollBarBreadth = 17; // change when document is loaded, but 17 is the standard for most browsers
    this._scrollBarBreadthSet = false; // set to true in setScrollWidth
    this._isRTL = false;
    this._rtlScrollType = 'default';
    this._div = null; // div by ID
    const d = this;
    /** @hidden @internal */
    const setFunc = function() {
      d.removeEventListener(root.document, 'DOMContentLoaded', setFunc, false);
      d.setRTL();
    };

    if (Diagram.isUsingDOM()) {
      if (root.document.body !== null) {
        this.setRTL();
      } else {
        d.addEventListener(root.document, 'DOMContentLoaded', setFunc, false);
      }
    }

    this._layers = new List<Layer>();
    this._canvasWidth = 0;
    this._canvasHeight = 0;
    this._canvas = null; // (main) canvas, also used by toolManager.js
    this._renderers = new Map<string, ISurface>();

    this._ctx = null; // the 2D context of this._canvas, also used by textBlock.js


    this._renderingHints = null;
    this.resetRenderingHints();
    this._inputOptions = null;
    this.resetInputOptions();

    this._opacity = 1;
    // 0, 0 instead of NaN NaN for the position, so that the updateScroll() method
    // will consider the bounds real, and update the scrollbars,
    // potentially before any Diagram.position has been set
    this._position = new Point(NaN, NaN).freeze();  // also used by animationManager.js
    this._savedPosition = new Point(NaN, NaN);
    this._scale = 1;
    this._defaultScale = 1.00;
    this._initialPosition = new Point(NaN, NaN).freeze();
    this._initialScale = NaN;

    this._minScale = 0.0001;
    this._maxScale = 100;

    this._transform = new Transform();
    this._zoomPoint = new Point(NaN, NaN).freeze();
    this._fixedBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._scrollMargin = new Margin(0, 0, 0, 0).freeze();
    this._scrollMode = Diagram.DocumentScroll;
    this._scrollsPageOnFocus = false;
    this._positionComputation = null;
    this._scaleComputation = null;

    this._autoScale = Diagram.None;
    this._contentAlignment = Spot.Default;
    this._initialAutoScale = Diagram.None;
    this._initialContentAlignment = Spot.Default;
    this._initialDocumentSpot = Spot.TopLeft;
    this._initialViewportSpot = Spot.TopLeft;

    this._invalidDraw = true;
    this._invalidCanvasSize = false;
    this._invalidParts = new Set<Part>();
    this._delayedGeometries = new Map<Link, Rect>();  // used in link.js
    this._drawTainted = true;  // used in picture.js

    // private for autoScroll
    this._autoScrollInterval = 250;
    this._autoScrollTimer = -1;  // -1 for no timer, or the timer ID
    // public for autoScroll
    this._autoScrollRegion = new Margin(16, 16, 16, 16).freeze();

    // timers and animation
    this._updateQueued = false;
    this._doNonRealtimeLayouts = false;
    this._firstUpdate = true;

    // Events
    this._firstInput = new InputEvent();
    this._firstInput.diagram = this;
    this._previousInput = new InputEvent();
    this._previousInput.diagram = this;
    this._lastInput = new InputEvent();
    this._lastInput.diagram = this;
    this._changedListeners = null;
    this._modelChangedListeners = null;
    this._tapHeld = false;

    this._diagramListeners = null;
    this._diagramEventNames = null;
    this._allowsPointers = (root['PointerEvent'] && (Util.isIE10 || Util.isIE11 || Util.isEdge) && root.navigator && root.navigator.msPointerEnabled !== false);
    this._setupDiagramEvents();

    this._topLevelGroups = new Set<Group>();  // also used by part.js
    this._isTreePathToChildren = true;
    this._treeCollapsePolicy = Diagram.TreeParentCollapsed;
    this._isCollapsingExpanding = false;
    this._validCycle = Diagram.CycleAll;

    // PartManager
    this._partManager = null;
    Diagram.addPartManager('Model', PartManager);

    // Tools
    this._toolManager = null;
    this._defaultTool = null;
    this._currentTool = null;

    this._currentCursor = '';
    this._defaultCursor = 'auto';
    this._click = null;
    this._doubleClick = null;
    this._contextClick = null;
    this._mouseOver = null;
    this._mouseHover = null;
    this._mouseHold = null;
    this._mouseDragOver = null;
    this._mouseDrop = null;
    this._handlesDragDropForTopLevelParts = false;
    this._mouseEnter = null;
    this._mouseLeave = null;
    this._toolTip = null;
    this._contextMenu = null;

    this._mouseOverQueued = false;
    this._pointers = {};
    // should be <PointerEvent> but closure compiler does't know about it
    this._pointers2 = [null, null];  // also used by tool.js

    this._commandHandler = null;

    this._skipsUndoManager = false;
    this._skipsModelSourceBindings = false;
    this._skipsBoundsUpdates = false;
    this._delaysLayout = false;

    this._isUpdatingModel = true;
    this._isUpdating = false;
    this._isDrawingForeign = false; // used to stop drawing optimizations during makeImage/makeSvg, also used by picture.js

    this._model = null;

    const diag = this;
    this._modelBindingHandler =
      /** @hidden @internal @param {ChangedEvent} e */
      function(e: ChangedEvent) { diag.partManager.doModelDataChanged(e); };

    this._modelChangedHandler =
      /** @hidden @internal @param {ChangedEvent} e */
      function(e: ChangedEvent) { diag.partManager.doModelChanged(e); };

    this._wasModified = true;
    this._historyIndex = -2;  // cannot be -1 or 0 or other valid index

    this._picturesMap = new Map<string, Array<Picture>>();  // Array of Picture, also used by picture.js
    this._picturesToInvalidate = new List<Picture>();  // also used by picture.js

    this._isReadOnly = false;
    this._isEnabled = true;
    this._allowClipboard = true;
    this._allowCopy = true;
    this._allowDelete = true;
    this._allowDragOut = false;
    this._allowDrop = true;
    this._allowTextEdit = true;
    this._allowGroup = true;
    this._allowUngroup = true;
    this._allowInsert = true;
    this._allowLink = true;
    this._allowRelink = true;
    this._allowMove = true;
    this._allowReshape = true;
    this._allowResize = true;
    this._allowRotate = true;
    this._allowSelect = true;
    this._allowUndo = true;
    this._allowZoom = true;

    this._isMouseCaptured = false;
    this._isMouseOverDiagram = false;

    // Scrollbars
    this._hasHorizontalScrollbar = true;
    this._hasVerticalScrollbar = true;

    this._allowHorizontalScroll = true;
    this._allowVerticalScroll = true;

    // defaults for these?
    this._scrollHorizontalLineChange = 16;
    this._scrollVerticalLineChange = 16;
    this._scrollBarMoving = false;
    this._scrollBarsModified = false;
    this._scrollParent = null;
    this._scrollHelper = null;
    // record last known scrollTop and ScrollLeft
    // to know which scrollbar was moved
    this._scrollTop = 0;
    this._scrollLeft = 0;

    this._padding = new Margin(5).freeze();

    this._selection = new Set<Part>().freeze();
    this._maxSelectionCount = 999999999;
    this._highlighteds = new Set<Part>().freeze();

    // invalidation
    this._invalidDocBounds = true;  // also used by part.js
    this._invalidScrollBars = true;
    this._invalidViewportObjects = true;  // also used by part.js

    // Hidden private only for use in draw and onscroll
    this._hasHScroll = false;
    this._hasVScroll = false;

    this._layout = null;
    this._anyLayoutInvalid = true;  // also used by layout.js
    this._isInitialLayoutCompleted = false;  // also used by animationManager.js, layout.js, link.js, undoManager.js???

    this._loadedRouteLinks = new Set<Part>();
    this._invalidLinkBundles = new Set<LinkBundle>();  // also used by link.js
    this._positions = null;

    // For gesture events:
    this._origScale = 1.0;  // also used by tool.js
    this._startDist = 0.0;  // also used by tool.js

    // Internal so we don't have to allocate so many:
    this._viewportBoundsChangedInfo = { scale: 1.0, position: new Point(), bounds: new Rect(), isScroll: false };

    // 0, 0 instead of NaN NaN for the position, so that the updateScroll() method
    // will consider the bounds real, and update the scrollbars,
    // potentially before any Diagram.position has been set
    this._viewportBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._viewSize = new Size(NaN, NaN).freeze();
    this._documentBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._isVirtualized = false;
    this._drawImage = null;

    this._setupLayers();

    this._nodeTemplateMap = null;
    this._groupTemplateMap = null;
    this._linkTemplateMap = null;

    this._defaultNodeTemplate = null;
    this._defaultGroupTemplate = null;
    this._defaultLinkTemplate = null;

    this._nodeSelectionAdornmentTemplate = null;
    this._groupSelectionAdornmentTemplate = null;
    this._linkSelectionAdornmentTemplate = null;

    this._setupTemplates();

    this._grid = null;
    this._gridWasVisible = false;

    this._currentEnterLeaveObject = null;  // also used by selectingTool.js, tool.js

    // initialize PartManager via property setter
    this.partManager = new PartManager();
    // this.initializeStandardPartManagers();

    // setup the ToolManager via property setter
    this.toolManager = new ToolManager();
    this.toolManager.initializeStandardTools();
    this.defaultTool = this.toolManager;
    // initialize currentTool via property setter
    this.currentTool = this.defaultTool;

    // Moved from draggingtool:
    this._draggedLink = null;
    this._defaultDragOptions = new DraggingOptions();

    this._oldFromPort = null;
    this._oldToPort = null;
    this._suspendedRouting = false;

    // setup the CommandHandler via property setter
    this.commandHandler = Diagram.setupCommandHandler();
    // initialize model via property setter, after setting _modelChangedHandler
    // if GraphLinksModel is loaded, use that, otherwise basic Model
    this.model = Model.initDiagramModel();
    this._skipsUndoManager = true;
    // initialize with property setter
    this.layout = new Layout();
    this._skipsUndoManager = false;

    // for optimizations
    this._drawCollectionCanvas = null;
    this._drawCollectionCtx = null;
    this._pixelRatio = 1;  // also used by graphObject.js
    this._tempPixelRatio = null; // some operations can override the normal pixel ratio with a temporary value.
    this._minDrawingLength = 1;
    this._movingAvg = 0; // The amount of time Diagram.draw took to complete, when the temp pixel ratio was non-null.
    this._movingAvgArray = [0, 0, 0, 0, 0];
    this._movingAvgFrame = 0;

    this._lastTouchCount = 1;
    this._lastTouchStamp = 0; // used for both IE and android browser bugs
    this._lastTouchpt = new Point();
    this._touchDoubleClickDelay = 500;
    this._autoScrollPoint = new Point();

    // simulated mouse events by a touch interface that we want to record and cancel
    // Some mobile browsers (like Android Default browser)
    // insert a mousedown/mousemove/mouseup and we want to disallow that
    // but doing so gets in the way of using tablet styluses
    // to be used in conjunction with Util.isAndroid534
    this._simulatedDetected = null;
    this._isTouchOngoing = false;  // also used by graphObject.js

    this._routingFromPort = null;
    this._routingFromNode = null;
    this._routingToPort = null;
    this._routingToNode = null;

    this.touchStart = null;
    this.touchMove = null;
    this.touchEnd = null;
    this.windowResize = null;
    this.mouseMove = null;
    this.mouseDown = null;
    this.mouseUp = null;
    this.mouseWheel = null;
    this.mouseOut = null;
    this.pointerDown = null;
    this.pointerMove = null;
    this.pointerUp = null;
    this.pointerOut = null;
    this.redrawLater = null;
    this.preventDefault = null;
    this._isModified = false;
    this._helper = new DiagramHelper(this, div);  // also used by graphObject.js
    this._svg = false; // NYI: debug test SVG
    if (div !== undefined) {
      this._helper.setup(this, div);
    }
    this._isAligning = false;

    this._focusObject = null;
    this._focusViewportPoint = Point.NoPoint;
  }

  /** @hidden @internal */
  private static DOMValue = (root.document !== undefined);

  /**
   * @hidden
   * Undocumented
   * @export
   * This static function returns true if Diagrams are using a DOM, false otherwise.
   * This can be false if there is no root <code>document</code> in the JavaScript context,
   * or if the DOM was disabled through {@link #useDOM}.
   * @this {Diagram}
   */
  public static isUsingDOM() {
    return Diagram.DOMValue;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * This static function sets whether or not GoJS should use a DOM, if one exists.
   * Can be set to false to simulate a DOM-less environment.
   * @this {Diagram}
   * @param {boolean} value
   */
  public static useDOM(value: boolean) {
    Diagram.DOMValue = value ? (root.document !== undefined) : false;
  }

  /**
   * Removes all Parts from the Diagram, including unbound Parts except for the background grid,
   * and also clears out the {@link Model} and {@link UndoManager} and clipboard.
   * This operation is not undoable.
   *
   * Alternative actions are to replace the {@link #model} with a new {@link Model}
   * (probably a {@link GraphLinksModel} or a {@link TreeModel}),
   * or to set {@link Model#nodeDataArray} with an empty JavaScript Array
   * (and {@link GraphLinksModel#linkDataArray}).
   *
   * This does not remove any listeners from the diagram.
   * @this {Diagram}
   */
  public clear(): void {
    this.model.clear();
    Diagram.clearClipboard();
    this.clearAllParts(false);
    this._documentBounds = new Rect(NaN, NaN, NaN, NaN).freeze();  // ??? why is this here instead of in reset()
    this.invalidateDraw();
  }

  /**
   * @hidden @internal
   * Just remove all Parts from this Diagram without affecting the model,
   * and while still keeping the background grid and clipboard.
   * @this {Diagram}
   * @param {boolean} keepUnmodeleds
   */
  private clearAllParts(keepUnmodeleds: boolean): void {
    let gridpart: Part = null;
    if (this._grid !== null) {
      gridpart = this._grid.part;
    }
    this.animationManager.stopAnimation(true); // perhaps pass in (!keepUnmodeleds)?
    // do not call Picture.clearCache(); here, instead call Picture.clearCache(url) in forgetPicture
    const unmodeleds = [];
    const l = this._layers.length;
    for (let i = 0; i < l; i++) {
      const lay = this._layers._arr[i];
      if (keepUnmodeleds) {
        const pit = lay.parts;
        while (pit.next()) {
          const part = pit.value;
          if (part !== gridpart && part.data === null) unmodeleds.push(part);
        }
      }
      lay.clear();
    }
    this.partManager.clear();
    this._invalidParts.clear();
    this._delayedGeometries.clear();
    this._topLevelGroups.clear();
    this.clearSelection();
    this.clearHighlighteds();
    this._currentEnterLeaveObject = null;
    Util._CachedArrays = [];
    if (gridpart !== null) {
      this.add(gridpart);  // add this Part to the Diagram
      this.partManager.parts.remove(gridpart);  // don't show up in Diagram.parts collection
    }
    if (keepUnmodeleds) {
      for (let j = 0; j < unmodeleds.length; j++) {
        this.add(unmodeleds[j]);
      }
    }
  }

  // this is the internal static clipboard variable
  /** @hidden @internal */
  public static _clipboardParts: Iterable<Part> = null;
  /** @hidden @internal */
  public static _clipboardFormat = '';

  /**
   * @hidden @internal
   */
  public static clearClipboard(): void {
    Diagram._clipboardParts = null;
    Diagram._clipboardFormat = '';
  }

  /**
   * @hidden @internal
   */
  public static setupCommandHandler(): ICommandHandler {
    return null;
  }

  // * Call {@link #clear} and also restore the templates, layers, layout and various Diagram properties to their original state.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   */
  public reset(): void {
    this._isAligning = true;
    this.clear();
    this._animationManager.stopAnimation();
    this._animationManager = new AnimationManager();
    this._animationManager.setDiagram(this);
    this._layers = new List<Layer>();
    this.resetRenderingHints();
    this.resetInputOptions();
    this._position = new Point(NaN, NaN).freeze();
    this._scale = 1;
    this._initialPosition = new Point(NaN, NaN).freeze();
    this._initialScale = NaN;
    this._minScale = 0.0001;
    this._maxScale = 100;
    this._zoomPoint = new Point(NaN, NaN).freeze();
    this._fixedBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._scrollMargin = new Margin(0, 0, 0, 0).freeze();
    this._scrollMode = Diagram.DocumentScroll;
    this._scrollsPageOnFocus = false;
    this._positionComputation = null;
    this._scaleComputation = null;
    this._autoScale = Diagram.None;
    this._contentAlignment = Spot.Default;
    this._initialAutoScale = Diagram.None;
    this._initialContentAlignment = Spot.Default;
    this._initialDocumentSpot = Spot.TopLeft;
    this._initialViewportSpot = Spot.TopLeft;
    this._autoScrollInterval = 250;
    this._autoScrollRegion = new Margin(16, 16, 16, 16).freeze();
    this._isTreePathToChildren = true;
    this._treeCollapsePolicy = Diagram.TreeParentCollapsed;
    this._validCycle = Diagram.CycleAll;
    this._defaultCursor = 'auto';
    this._click = null;
    this._doubleClick = null;
    this._contextClick = null;
    this._mouseOver = null;
    this._mouseHover = null;
    this._mouseHold = null;
    this._mouseDragOver = null;
    this._mouseDrop = null;
    this._handlesDragDropForTopLevelParts = false;
    this._mouseEnter = null;
    this._mouseLeave = null;
    this._toolTip = null;
    this._contextMenu = null;
    this._isReadOnly = false;
    this._isEnabled = true;
    this._allowClipboard = true;
    this._allowCopy = true;
    this._allowDelete = true;
    this._allowDragOut = false;
    this._allowDrop = true;
    this._allowTextEdit = true;
    this._allowGroup = true;
    this._allowUngroup = true;
    this._allowInsert = true;
    this._allowLink = true;
    this._allowRelink = true;
    this._allowMove = true;
    this._allowReshape = true;
    this._allowResize = true;
    this._allowRotate = true;
    this._allowSelect = true;
    this._allowUndo = true;
    this._allowZoom = true;
    this._hasHorizontalScrollbar = true;
    this._hasVerticalScrollbar = true;
    this._allowHorizontalScroll = true;
    this._allowVerticalScroll = true;
    this._scrollHorizontalLineChange = 16;
    this._scrollVerticalLineChange = 16;
    this._padding = new Margin(5).freeze();
    this._maxSelectionCount = 999999999;
    this._layout = null;
    this._isVirtualized = false;

    this._pixelRatio = 1;
    this._tempPixelRatio = null;
    this._movingAvg = 0;
    this._movingAvgArray = [0, 0, 0, 0, 0];
    this._movingAvgFrame = 0;

    this._setupTemplates();
    this._grid = null;

    // initialize PartManager via property setter
    this.partManager = new PartManager();
    // setup the ToolManager via property setter
    this.toolManager = new ToolManager();
    this.toolManager.initializeStandardTools();
    // moved from dragging tool:
    this._draggedLink = null;
    this._oldFromPort = null;
    this._oldToPort = null;
    this._suspendedRouting = false;

    this._defaultDragOptions.reset();

    this._renderers = new Map<string, ISurface>();
    this._renderers.clear(); // Delete standard renderers, like SVGSurfance, if they exist

    this.defaultTool = this.toolManager;
    // initialize currentTool via property setter
    this.currentTool = this.defaultTool;
    // setup the CommandHandler via property setter
    this.commandHandler = Diagram.setupCommandHandler();

    this._skipsUndoManager = true;
    this._setupLayers();
    this.layout = new Layout();
    this._skipsUndoManager = false;
    this.model = Model.initDiagramModel();
    this.model.undoManager = new UndoManager();  // after setting .model

    this._isInitialLayoutCompleted = false;
    this._firstUpdate = true;
    this._updateQueued = false;

    this._isAligning = false;
    this.invalidateDraw();

    this._changedListeners = null;
    this._modelChangedListeners = null;
    this._setupDiagramEvents();
    this._currentCursor = '';
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   */
  private _setupTemplates(): void {
    // Node Templates
    this._nodeTemplateMap = new Map<string, Part>();

    // create the default Node template
    const archnode = new Node();
    const nodet = new TextBlock();
    nodet.bind(new Binding('text', '', Util.toString));
    archnode.add(nodet);
    this._defaultNodeTemplate = archnode;
    this._nodeTemplateMap.add('', archnode);

    // create the default Comment Node template
    const archcmnt = new Node();
    const nodec = new TextBlock();
    nodec.stroke = 'brown';
    nodec.bind(new Binding('text', '', Util.toString));
    archcmnt.add(nodec);
    this._nodeTemplateMap.add('Comment', archcmnt);

    // create the default Link Label Node template
    const archllab = new Node();
    archllab.selectable = false;
    archllab.avoidable = false;
    const nodel = new Shape();
    nodel.figure = 'Ellipse';
    nodel.fill = 'black';
    nodel.stroke = null;
    nodel.desiredSize = new Size(3, 3).constant();
    archllab.add(nodel);
    this._nodeTemplateMap.add('LinkLabel', archllab);

    // Group Templates
    this._groupTemplateMap = new Map<string, Group>();

    // create the default Group template
    const archgrp = new Group();
    archgrp.selectionObjectName = 'GROUPPANEL';
    archgrp.type = Panel.Vertical;
    const grpt = new TextBlock();
    grpt.font = 'bold 12pt sans-serif';
    grpt.bind(new Binding('text', '', Util.toString));
    archgrp.add(grpt);
    const grppan = new Panel(Panel.Auto);
    grppan.name = 'GROUPPANEL';
    const grpbord = new Shape();
    grpbord.figure = 'Rectangle';
    grpbord.fill = 'rgba(128,128,128,0.2)';
    grpbord.stroke = 'black';
    grppan.add(grpbord);
    const phold = new Placeholder();
    phold.padding = new Margin(5, 5, 5, 5).constant();
    grppan.add(phold);
    archgrp.add(grppan);
    this._defaultGroupTemplate = archgrp;
    this._groupTemplateMap.add('', archgrp);

    // Link Templates
    this._linkTemplateMap = new Map<string, Link>();

    // create the default Link template
    const archlink = new Link();
    const archpath = new Shape();
    archpath.isPanelMain = true;
    archlink.add(archpath);
    const archarrow = new Shape();
    archarrow.toArrow = 'Standard';
    archarrow.fill = 'black';
    archarrow.stroke = null;
    archarrow.strokeWidth = 0;
    archlink.add(archarrow);
    this._defaultLinkTemplate = archlink;
    this._linkTemplateMap.add('', archlink);

    // create the default Comment Link template
    const archcmntlink = new Link();
    const archcmntpath = new Shape();
    archcmntpath.isPanelMain = true;
    archcmntpath.stroke = 'brown';
    archcmntlink.add(archcmntpath);
    this._linkTemplateMap.add('Comment', archcmntlink);

    // Selection Adornments
    // create the default Adornment for selection
    let selad = new Adornment();
    selad.type = Panel.Auto;
    let seladhandle = new Shape();
    seladhandle.fill = null;
    seladhandle.stroke = 'dodgerblue';
    seladhandle.strokeWidth = 3;
    selad.add(seladhandle);
    const selplace = new Placeholder();
    selplace.margin = new Margin(1.5, 1.5, 1.5, 1.5).constant();
    selad.add(selplace);
    this._nodeSelectionAdornmentTemplate = selad;

    // reuse the default Node Adornment for selection
    this._groupSelectionAdornmentTemplate = selad;

    // create the default Link Adornment for selection
    selad = new Adornment();
    selad.type = Panel.Link;
    seladhandle = new Shape();
    seladhandle.isPanelMain = true;
    seladhandle.fill = null;
    seladhandle.stroke = 'dodgerblue';
    seladhandle.strokeWidth = 3;  // ?? zero to use selection object's strokeWidth is often not wide enough
    selad.add(seladhandle);
    this._linkSelectionAdornmentTemplate = selad;
  }

  /**
   * @hidden
   * @expose
   * Used in Diagram constructor setup, this computes the pixel width of the scrollbars
   * @param {HTMLElement=} elem
   * @this {Diagram}
   */
  public setRTL(elem?: HTMLElement): void {
    let appendage = elem === undefined ? this.div : elem;
    if (appendage === null) appendage = root.document.body; // if;no diagram.div yet
    const div = Util.createElement('div');
    div.dir = 'rtl';
    div.style.cssText = 'font-size: 14px; width: 1px; height: 1px; position: absolute; top: -1000px; overflow: scroll;';
    div.textContent = 'A';
    appendage.appendChild(div);
    let type = 'reverse';
    if (div.scrollLeft > 0) {
      type = 'default';
    } else {
      div.scrollLeft = 1;
      if (div.scrollLeft === 0) type = 'negative';
    }
    appendage.removeChild(div);
    this._rtlScrollType = type;
  }

  // * Computes the pixel width of the scrollbars
  /**
   * @hidden
   * Undocumented
   * @expose
   * @param {HTMLElement=} elem
   * @this {Diagram}
   */
  public setScrollWidth(elem?: HTMLElement): void {
    let appendage = elem === undefined ? this.div : elem;
    if (appendage === null) appendage = root.document.body; // if;no diagram.div yet
    let width = 0;
    if (Diagram.isUsingDOM()) {
      let inner = Diagram.scrollInner;
      let outer = Diagram.scrollOuter;
      if (inner === null) {
        Diagram.scrollInner = Util.createElement('p');
        inner = Diagram.scrollInner;
        inner.style.width = '100%';
        inner.style.height = '200px';
        inner.style.boxSizing = 'content-box';
        Diagram.scrollOuter = Util.createElement('div');
        outer = Diagram.scrollOuter;
        outer.style.position = 'absolute';
        outer.style.visibility = 'hidden';
        outer.style.width = '200px';
        outer.style.height = '150px';
        outer.style.boxSizing = 'content-box';
        outer.appendChild(inner);
      }
      outer.style.overflow = 'hidden';
      appendage.appendChild(outer);
      const w1 = inner.offsetWidth;
      outer.style.overflow = 'scroll';
      let w2 = inner.offsetWidth;
      if (w1 === w2) w2 = outer.clientWidth;
      appendage.removeChild(outer);
      width = w1 - w2;
      if (width === 0 && !Util.isIOS) width = 11; // for Mac hidden scrollbars
      // ?? try setting width to 0. Ideally we want to accomodate zero-sized scrollbars that fit "over" the canvas
    }
    this._scrollBarBreadth = width;
    this._scrollBarBreadthSet = true;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Diagram}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Diagram) {
      this.autoScale = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {number=} details int Optional: If greater than zero, this includes descriptions of each layer.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    let name = '';
    if (this.div && this.div.id) name = this.div.id;
    let str = 'Diagram \"' + name + '\"';
    if (details <= 0) return str;
    // if details >= 1, talk about the layers
    const lit = this._layers.iterator;
    while (lit.next()) {
      const layer = lit.value;
      str += '\n  ' + layer.toString(details - 1);
    }
    return str;
  }

  /**
   * @hidden @internal
   * @param diag
   */
  private static rememberDiagram(diag: Diagram) {
    Diagram._last = diag;
  }

  /**
   * @hidden @internal
   */
  public static getLastDiagram(): Diagram {
    return Diagram._last as Diagram;
  }

  /**
   * This static function gets the {@link Diagram} that is attached to an HTML DIV element.
   * @this {Diagram}
   * @param {Element|string} div
   * @return {Diagram} null if there is no {@link Diagram} associated with the given DIV,
   * or if the argument is not a DIV element nor a string naming such a DIV element in the HTML document.
   */
  public static fromDiv(div: Element | string): Diagram | null {
    let elt = div;
    if (typeof div === 'string') elt = root.document.getElementById(div);
    if (elt instanceof HTMLDivElement && elt._diagram instanceof Diagram) return elt._diagram;
    return null;
  }

  /**
   * Gets or sets the Diagram's HTMLDivElement, via an HTML Element ID.
   * This is typically set automatically when a Div is supplied as an argument to Diagram's constructor.
   *
   * Setting this property to a new value will clobber any HTML and
   * inner DOM elements inside of both the new and the old divs.
   * It will then populate the Div with the elements
   * (inner Divs, Canvases) needed for the Diagram to function.
   *
   * If you want to disassociate the Diagram from the HTML Div element, set {@link Diagram#div} to null.
   * If you remove a part of the HTML DOM containing a Div with a Diagram, you will need to
   * set {@link #div} to null in order for the page to recover the memory.
   *
   * You should not attempt to manually modify the contents of this Div.
   * Changing this property value does not raise a Changed event.
   */
  /** @export */ get div(): HTMLDivElement | null { return this._div; }
  /** @export */ set div(value: HTMLDivElement | null) {
    if (value !== null) Util.checkClass(value, HTMLDivElement, Diagram, 'div');
    if (this._div !== value) {
      Util.clearDiagramsTraversed();

      const olddiv = this._div;
      if (olddiv !== null) {
        (olddiv as any)._diagram = undefined;
        // Clobber any HTML inside the div
        // Also destroys all inner DOM objects, such as the Canvas and the four scroll-related divs
        olddiv.innerHTML = '';
        if (this._canvas !== null) {
          const canvasElt = this._canvas.domElement;
          this.removeEventListener(canvasElt, 'touchstart', this.touchStart, false);
          this.removeEventListener(canvasElt, 'touchmove', this.touchMove, false);
          this.removeEventListener(canvasElt, 'touchend', this.touchEnd, false);
          this._canvas.dispose();
        }
        // remove all possible tool timers
        const tools = this.toolManager;
        if (tools !== null) {
          tools.mouseDownTools.each(function(t) { t.cancelWaitAfter(); });
          tools.mouseMoveTools.each(function(t) { t.cancelWaitAfter(); });
          tools.mouseUpTools.each(function(t) { t.cancelWaitAfter(); });
        }
        tools.cancelWaitAfter();
        this.currentTool.doCancel();

        this._canvas = null;
        this._ctx = null;
        this.removeEventListener(root, 'resize', this.windowResize, false);
        // just in case a div was changed while captured:
        this.removeEventListener(root, 'mousemove', this.mouseMove, true);
        this.removeEventListener(root, 'mousedown', this.mouseDown, true);
        this.removeEventListener(root, 'mouseup', this.mouseUp, true);
        this.removeEventListener(root, 'wheel', this.mouseWheel, true);
        this.removeEventListener(root, 'mouseout', this.mouseOut, true);

        if (Diagram.getLastDiagram() === this) Diagram.rememberDiagram(null);
      } else {
        // old div was null
        this._isInitialLayoutCompleted = false;
      }
      this._div = null;  // not value, because setup will set ._div
      if (value !== null) {
        // if new DIV already attached to a Diagram, detach it first
        const olddiag = value._diagram;
        if (olddiag) {
          olddiag.div = null;
        }
        // attach this Diagram to the new DIV element
        this._helper.setup(this, value);

        this.redraw();
      }
    }
  }

  /**
   * @hidden
   * Undocumented, but may be useful for change detection calls in Angular.
   * @expose
   * @this {Diagram}
   */
  public addEventListener(DOMElement: Element | Window | Document, name: string, listener: any /* EventListenerOrEventListenerObject */, capture: boolean): void {
    DOMElement.addEventListener(name, listener, capture);
  }

  /**
   * @hidden
   * Undocumented, but may be useful for change detection calls in Angular.
   * @expose
   * @this {Diagram}
   */
  public removeEventListener(DOMElement: Element | Window | Document, name: string, listener: any /* EventListenerOrEventListenerObject */, capture: boolean): void {
    DOMElement.removeEventListener(name, listener, capture);
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   */
  public addListeners(): void {
    const diagram = this;
    const newCanvasElt = this._canvas.domElement;
    if (!(newCanvasElt instanceof HTMLCanvasElement)) return; // NYI: no events in SVGs

    if (diagram._allowsPointers) {
      this.addEventListener(newCanvasElt, 'pointerdown', diagram.pointerDown, false);
      this.addEventListener(newCanvasElt, 'pointermove', diagram.pointerMove, false);
      this.addEventListener(newCanvasElt, 'pointerup', diagram.pointerUp, false);
      this.addEventListener(newCanvasElt, 'pointerout', diagram.pointerOut, false);
    } else {
      this.addEventListener(newCanvasElt, 'touchstart', diagram.touchStart, false);
      this.addEventListener(newCanvasElt, 'touchmove', diagram.touchMove, false);
      this.addEventListener(newCanvasElt, 'touchend', diagram.touchEnd, false);
      this.addEventListener(newCanvasElt, 'mousemove', diagram.mouseMove, false);
      this.addEventListener(newCanvasElt, 'mousedown', diagram.mouseDown, false);
      this.addEventListener(newCanvasElt, 'mouseup', diagram.mouseUp, false);
      this.addEventListener(newCanvasElt, 'mouseout', diagram.mouseOut, false);
    }

    this.addEventListener(newCanvasElt, 'mouseenter', diagram.doMouseEnter, false);
    this.addEventListener(newCanvasElt, 'mouseleave', diagram.doMouseLeave, false);
    this.addEventListener(newCanvasElt, 'wheel', diagram.mouseWheel, false);
    this.addEventListener(newCanvasElt, 'keydown', diagram.keyDown, false);
    this.addEventListener(newCanvasElt, 'keyup', diagram.keyUp, false);
    this.addEventListener(newCanvasElt, 'blur', diagram.canvasBlur, false);
    this.addEventListener(newCanvasElt, 'focus', diagram.canvasFocus, false);

    // fixes a problem where double clicking causes text to get selected on the canvas
    this.addEventListener(newCanvasElt, 'selectstart', function(e: Event) { e.preventDefault(); return false; }, false);
    this.addEventListener(newCanvasElt, 'contextmenu', function(e: Event) { e.preventDefault(); return false; }, false);

    this.addEventListener(newCanvasElt, 'gesturestart', function(e: Event) {
      // Needed for iOS + Mac
      if (diagram.toolManager.gestureBehavior === ToolManager.GestureNone) {
        return;
      }
      if (diagram.toolManager.gestureBehavior === ToolManager.GestureCancel) {
        e.preventDefault();
        return;
      }
      if (diagram._isTouchOngoing && diagram.lastInput.handled) {
        return; // stop if the last touch event was handled
      }

      // Needed for Mac Trackpad two finger scaling
      e.preventDefault();
      diagram._origScale = diagram.scale;
      diagram.currentTool.doCancel();
    }, false);

    this.addEventListener(newCanvasElt, 'gesturechange', function(e: any /* b/c Event doesn't have scale/pageX/pageY */) {
      // Needed for Mac Trackpad two finger scaling
      if (diagram.toolManager.gestureBehavior === ToolManager.GestureNone) {
        return;
      }
      if (diagram.toolManager.gestureBehavior === ToolManager.GestureCancel) {
        e.preventDefault();
        return; // possibly not necessary
      }
      if (diagram._isTouchOngoing && diagram.lastInput.handled) {
        return; // stop if the last touch event was handled
      }
      e.preventDefault();
      const scale = e.scale;

      // Stops retro-active events. If the default is not prevented,
      // at the end of the gesture, a large number of gesturechange events happen all at once.
      if (diagram._origScale === null) return;

      const bbox = diagram._canvas.getBoundingClientRect();
      const pt = new Point((e.pageX - window.scrollX) - bbox.left * (diagram._canvasWidth / bbox.width),
        (e.pageY - window.scrollY) - bbox.top * (diagram._canvasHeight / bbox.height));
      const newscale = diagram._origScale * scale;
      const command = diagram.commandHandler;
      if (newscale !== diagram.scale && command.canResetZoom(newscale)) {
        const oldzoom = diagram.zoomPoint;
        diagram.zoomPoint = pt;
        command.resetZoom(newscale);
        diagram.zoomPoint = oldzoom;
      }
    }, false);

    this.addEventListener(root, 'resize', diagram.windowResize, false);

    // Potentially we could listen for visibility changes and:
    // request an update when the browser tab is visible again
    // cancel timers when the browser tab becomes hidden
    /*
    this.addEventListener(root.document, 'visibilitychange', function() {
      if (document.visibilityState === 'visible') diagram.requestUpdate();
    });
    */
  }

  /**
   * @hidden @internal
   * Sets tempPixelRatio to a non-null value, which overrides any computed pixel ratio
   * @this {Diagram}
   */
  public enableTempPixelRatio(): void {
    if (this._movingAvg > 30) this._tempPixelRatio = 1;
  }

  /**
   * @hidden @internal
   * @param {boolean} redraw true if the diagram should redraw soon
   * @this {Diagram}
   */
  public disableTempPixelRatio(redraw: boolean): void {
    if (this._tempPixelRatio === null) return; // might have never been enabled
    this._tempPixelRatio = null;
    if (redraw) {
      this.redrawLater();
    }
    this._movingAvg = 0;
    this._movingAvgArray = [0, 0, 0, 0, 0];
    this._movingAvgFrame = 0;
  }

  // Can be overriden, consider overriding on Overview?
  //  * Computes the device pixel ratio divided by the backing store pixel ratio
  //  * Used to set this._pixelRatio to a value other than 1.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Diagram}
   * @return {number}
   */
  public computePixelRatio(): number {
    if (this._tempPixelRatio !== null) return this._tempPixelRatio;
    const dpr = root.devicePixelRatio || 1;
    return dpr;
  }

  /**
   * @hidden @internal
   * Undocumented
   * @this {Diagram}
   * @return {number}
   */
  /* noexport */get avgSpf(): number {
    return this._movingAvg;
  }


  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doMouseMove(): void {
    this.currentTool.doMouseMove();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doMouseDown(): void {
    this.currentTool.doMouseDown();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doMouseUp(): void {
    this.currentTool.doMouseUp();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doMouseWheel(): void {
    this.currentTool.doMouseWheel();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doKeyDown(): void {
    this.currentTool.doKeyDown();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Diagram}
   */
  public doKeyUp(): void {
    this.currentTool.doKeyUp();
  }

  // * Focus the Diagram's canvas, allowing it to receive keyboard events.
  // * This is called upon activation of mouseDown or mouseMove tools, or on any mouseUp.
  // *
  // * This is normally implemented as a simple call to {@link #focus}.
  // * Override this if you do not wish to focus the canvas under certain circumstances.
  // * However the method should call {@link #focus} most of the time in order to get the normal behaviors.
  // *
  // * {@link #lastInput} can be used to inspect the mouseDown event that calls this method.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Diagram}
   */
  public doFocus(): void {
    this.focus();
  }

  /**
   * Explicitly bring HTML focus to the Diagram's canvas.
   * This is called by tools that may create other HTML elements such as TextEditingTool.
   *
   * If {@link #scrollsPageOnFocus} is false, this tries to keep the page at the same scroll position
   * that it had before calling {@link #focus}.
   * This method is not overridable.
   * @export
   * @this {Diagram}
   */
  public focus(): void {
    if (!this._canvas) return;
    if (this.scrollsPageOnFocus) {
      this._canvas.focus(); // will ultimately call canvasFocus
    } else {
      const x = root.scrollX || root.pageXOffset;
      const y = root.scrollY || root.pageYOffset;
      this._canvas.focus(); // will ultimately call canvasFocus
      root.scrollTo(x, y);
    }
  }
  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   */
  public canvasFocus(this: HTMLCanvasElement, e: Event): void {
    const diagram = this._diagram;
    diagram.raiseDiagramEvent('GainedFocus');
  }

  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   */
  public canvasBlur(this: HTMLCanvasElement, e: Event): void {
    const diagram = this._diagram;
    diagram.raiseDiagramEvent('LostFocus');
  }

  /**
   * @hidden @internal
   * Called when the window changes size. When this occurs, we might need to resize the div.
   * @this {Diagram}
   */
  public maybeResize(): void {
    const can = this._canvas;
    if (can === null) return;
    const div = this._div;
    if (div.clientWidth === 0 || div.clientHeight === 0) {
      return;
    }
    if (!this._scrollBarBreadthSet) this.setScrollWidth();
    const sw = (this._hasVScroll) ? this._scrollBarBreadth : 0;
    const sh = (this._hasHScroll) ? this._scrollBarBreadth : 0;

    // ??? ensure a call to this.updateCanvasSize(w,h) if it changes? When that happens, demand a redraw at a minimum?
    // ??? why does this happen here? this means we are calling computePixelRatio way too often!
    const old = this._pixelRatio;
    this._pixelRatio = this.computePixelRatio();
    if (this._pixelRatio !== old) {
      this._invalidCanvasSize = true;
      this.requestUpdate();
    }

    if (div.clientWidth !== this._canvasWidth + sw || div.clientHeight !== this._canvasHeight + sh) {
      // ?? consider stopping any animation when the canvas resizes
      this._invalidScrollBars = true;
      this._invalidDraw = true;
      const lay = this.layout;
      if (lay !== null && lay.isViewportSized && this.autoScale === Diagram.None) {
        // There might be some danger setting this so far from the source of the animation.
        this._doNonRealtimeLayouts = true;  // perform any automatic layouts.
        lay.invalidateLayout();
      }
      if (!this._isUpdating) this.requestUpdate();
    }
  }

  /**
   * @hidden @internal
   * Called by setup, initializes default layers
   * @this {Diagram}
   */
  private _setupLayers(): void {
    // three normal layers and two temporary layers

    let layer = new Layer();
    layer.name = 'Background';
    this.addLayer(layer);

    layer = new Layer();
    layer.name = '';
    this.addLayer(layer);

    layer = new Layer();
    layer.name = 'Foreground';
    this.addLayer(layer);

    layer = new Layer();
    layer.name = 'Adornment';
    layer.isTemporary = true;
    this.addLayer(layer);

    layer = new Layer();
    layer.name = 'Tool';
    layer.isTemporary = true;
    layer.isBoundsIncluded = true;
    this.addLayer(layer);

    layer = new Layer();
    layer.name = 'Grid';
    layer.allowSelect = false;
    layer.pickable = false;
    layer.isTemporary = true;
    this.addLayerBefore(layer, this.findLayer('Background'));
  }

  /**
   * @hidden @internal
   * Called by setup, initializes .grid
   * @this {Diagram}
   */
  private _setupBackgroundGrid(): void {
    // make the background Grid Panel
    this._grid = new Panel(Panel.Grid);
    this._grid.name = 'GRID';

    let hlines = new Shape();
    hlines.figure = 'LineH';
    hlines.stroke = 'lightgray';
    hlines.strokeWidth = 0.5;
    hlines.interval = 1;
    this._grid.add(hlines);

    hlines = new Shape();
    hlines.figure = 'LineH';
    hlines.stroke = 'gray';
    hlines.strokeWidth = 0.5;
    hlines.interval = 5;
    this._grid.add(hlines);

    hlines = new Shape();
    hlines.figure = 'LineH';
    hlines.stroke = 'gray';
    hlines.strokeWidth = 1;
    hlines.interval = 10;
    this._grid.add(hlines);

    let vlines = new Shape();
    vlines.figure = 'LineV';
    vlines.stroke = 'lightgray';
    vlines.strokeWidth = 0.5;
    vlines.interval = 1;
    this._grid.add(vlines);

    vlines = new Shape();
    vlines.figure = 'LineV';
    vlines.stroke = 'gray';
    vlines.strokeWidth = 0.5;
    vlines.interval = 5;
    this._grid.add(vlines);

    vlines = new Shape();
    vlines.figure = 'LineV';
    vlines.stroke = 'gray';
    vlines.strokeWidth = 1;
    vlines.interval = 10;
    this._grid.add(vlines);

    // create the Part holding the Grid Panel
    const gridpart = new Part();
    gridpart.add(this._grid);
    gridpart.layerName = 'Grid';  // goes in the "Grid" layer
    gridpart.zOrder = 0;  // to make it easier for other background parts to be behind the grid
    gridpart.isInDocumentBounds = false;  // never part of the document bounds
    gridpart.isAnimated = false;
    gridpart.pickable = false;
    gridpart.locationObjectName = 'GRID';
    this.add(gridpart);  // add this Part to the Diagram
    this.partManager.parts.remove(gridpart);  // don't show up in Diagram.parts collection

    // _updateBackgroundGrid will update the size and position based on the viewport
    this._grid.visible = false;  // grid is by default not visible
  }

  /**
   * @hidden @internal
   * Event that fires whenever a scrollbar is moved. Is a no-op unless the scrollbar
   * has been clicked by a user.
   * @this {HTMLDivElement}
   */
  public static _onscroll(this: HTMLElement): void {
    if (this._diagram._scrollBarsModified) {
      this._diagram._scrollBarsModified = false;
      return;
    }
    if (!this._diagram.isEnabled) {
      this._diagram.updateScroll();
      return;
    }
    this._diagram.diagramScroll(this);
  }

  /**
   * @hidden @internal
   * Used to differentiate scrollbars being clicked and scrollbars moving by other means
   * @this {HTMLDivElement}
   * @param {Event} e
   */
  public static _onscrollClick(this: any, e: any): void { // using "any" because this is used for both mouse down and touch start.
    if (!this._diagram.isEnabled) {
      this._diagram.updateScroll();
      return;
    }
    this._diagram._scrollTop = e.target.scrollTop;
    this._diagram._scrollLeft = e.target.scrollLeft;
  }

  /**
   * @hidden @internal
   * Called when a scrollbar moves
   * @export
   * @this {Diagram}
   * @param {HTMLDivElement} scrolldiv
   */
  public diagramScroll(scrolldiv: HTMLElement): void {
    if (this._canvas === null) return;
    this._scrollBarMoving = true;
    // this.scrollLeft, this.scrollTop only works if document is entirely within view

    let docB = this.documentBounds;
    let viewB = this.viewportBounds;
    const scrollMargin = this._scrollMargin;

    let dx = docB.x - scrollMargin.left;
    let dy = docB.y - scrollMargin.top;
    const dw = docB.width + scrollMargin.left + scrollMargin.right;
    const dh = docB.height + scrollMargin.top + scrollMargin.bottom;
    let dright = docB.right + scrollMargin.right;
    let dbottom = docB.bottom + scrollMargin.bottom;

    let vx = viewB.x;
    let vy = viewB.y;
    const vw = viewB.width;
    const vh = viewB.height;
    let vright = viewB.right;
    let vbottom = viewB.bottom;

    const sc = this.scale;
    let p;

    let value = scrolldiv.scrollLeft;
    if (this._isRTL) {
      switch (this._rtlScrollType) {
        case 'negative':
          value = value + scrolldiv.scrollWidth - scrolldiv.clientWidth;
          break;
        case 'reverse':
          value = scrolldiv.scrollWidth - value - scrolldiv.clientWidth;
          break;
      }
    }
    // or 'default'
    const scrolldivScrollLeft = value;

    // if the view is smaller than the Diagram
    if ((vw < dw) || (vh < dh)) {
      p = Point.allocAt(this.position.x, this.position.y);
      if (this.allowHorizontalScroll && this._scrollLeft !== scrolldivScrollLeft) {
        p.x = (scrolldivScrollLeft / sc + dx);
        this._scrollLeft = scrolldivScrollLeft;
      }
      if (this.allowVerticalScroll && this._scrollTop !== scrolldiv.scrollTop) {
        p.y = (scrolldiv.scrollTop / sc + dy);
        this._scrollTop = scrolldiv.scrollTop;
      }
      this.position = p;
      Point.free(p);

      this._scrollBarMoving = false;
      this._invalidScrollBars = false;
      return;
    }

    p = Point.alloc();
    // scrollHorizontal is something we added to the div to distinguish them
    if (scrolldiv._scrollHorizontal && this.allowHorizontalScroll) {
      if (dx < vx) {  // Left scrolling
        this.position = p.assignTo(scrolldivScrollLeft + dx, this.position.y);
      }

      if (dright > vright) {  // Right scrolling
        const sw = (this._scrollParent.scrollWidth - this._canvasWidth);
        this.position = p.assignTo(-sw + scrolldivScrollLeft - (this._canvasWidth / sc) + dright, this.position.y);
      }
    }

    // scrollVertical is something we added to the div to distinguish them
    if (scrolldiv._scrollVertical && this.allowVerticalScroll) {
      if (dy < vy) { // Top scrolling
        this.position = p.assignTo(this.position.x, scrolldiv.scrollTop + dy);
      }

      if (dbottom > vbottom) { // Bottom scrolling
        const sh = (this._scrollParent.scrollHeight - this._canvasHeight);
        this.position = p.assignTo(this.position.x, -sh + scrolldiv.scrollTop - (this._canvasHeight / sc) + dbottom);
      }
    }

    Point.free(p);
    this.invalidateViewportObjects();
    this._scrollBarMoving = false;
    this._invalidScrollBars = false;

    // now we need to see if a scrollbar is no longer necessary (view larger than Diagram)
    docB = this.documentBounds;
    viewB = this.viewportBounds;

    dright = docB.right;
    vright = viewB.right;
    dbottom = docB.bottom;
    vbottom = viewB.bottom;

    dx = docB.x;
    vx = viewB.x;
    dy = docB.y;
    vy = viewB.y;

    // no more horizontal bar:
    if ((vw >= dw) && (dx >= vx) && (dright <= vright)) {
      this._scrollHelper.style.width = '1px';
    }

    // no more vertical bar:
    if ((vh >= dh) && (dy >= vy) && (dbottom <= vbottom)) {
      this._scrollHelper.style.height = '1px';
    }
  }

  /**
   * This is called during a Diagram update to determine a new value for {@link #documentBounds}.
   * By default this computes the union of the bounds of all the visible
   * {@link GraphObject}s in this Diagram, unless {@link Diagram#fixedBounds} is set.
   *
   * To compute the bounds of a collection of Parts, call {@link #computePartsBounds}.
   * @expose
   * @this {Diagram}
   * @return {Rect} a Rect in document coordinates.
   */
  protected computeBounds(): Rect {
    if (this._invalidParts.count > 0) {
      this.validateParts();
    }
    return this.computeBoundsNoValidate();
  }

  /**
   * @hidden @internal
   * Always copies/creates new rect
   * @this {Diagram}
   * @return {Rect} a Rect in document coordinates.
   */
  private computeBoundsNoValidate(): Rect {
    if (this.fixedBounds.isReal()) {
      const fb = this.fixedBounds.copy();
      fb.addMargin(this._padding);
      return fb;
    }

    let b: Rect;
    let first = true;
    const lit = this._layers._arr;
    const l = lit.length;
    for (let i = 0; i < l; i++) {
      const layer = lit[i];
      if (!layer.visible) continue;
      if (layer.isTemporary && !layer.isBoundsIncluded) continue;
      const pit = layer._parts._arr;
      const ll = pit.length;
      for (let j = 0; j < ll; j++) {
        const part = pit[j];
        if (!part.isInDocumentBounds) continue;
        if (!part.isVisible()) continue;
        const ab = part.actualBounds;
        if (!ab.isReal()) continue;
        if (first) {
          first = false;
          b = ab.copy();
        } else {
          b.unionRect(ab);
        }
      }
    }
    // if no elements make an empty rect
    if (first) {
      b = new Rect(0, 0, 0, 0);
    }
    b.addMargin(this._padding);
    return b;
  }

  /**
   * Find the union of the {@link GraphObject#actualBounds} of all of the {@link Part}s in the given collection,
   * excluding Links unless the second argument is true.
   *
   * Unlike {@link #computeBounds}, this ignores the visibility of each Part and does not add any padding to the result.
   * @expose
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll an iterable collection or Array of {@link Part}s.
   * @param {boolean=} includeLinks defaults to false
   * @return {Rect} This returns the bounding area of the given Parts;
   * if there are no Parts in the collection, this returns a Rect with zero width and height and an X and Y that are NaN.
   * @since 1.1
   */
  public computePartsBounds(coll: Iterable<Part> | Array<Part>, includeLinks?: boolean): Rect {
    if (includeLinks === undefined) includeLinks = false;
    let b: Rect = null;
    if (Util.isArray(coll)) {
      for (let i = 0; i < coll.length; i++) {
        const part = coll[i];
        if (!includeLinks && part instanceof Link) continue;
        part.ensureBounds();
        if (b === null) {
          b = part.actualBounds.copy();
        } else {
          b.unionRect(part.actualBounds);
        }
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const part = it.value;
        if (!includeLinks && part instanceof Link) continue;
        part.ensureBounds();
        if (b === null) {
          b = part.actualBounds.copy();
        } else {
          b.unionRect(part.actualBounds);
        }
      }
    }

    if (b === null) return new Rect(NaN, NaN, 0, 0);
    return b;
  }

  /**
   * @hidden @internal
   * This method is called when the {@link #documentBounds}
   * or {@link #contentAlignment} properties change.
   * This may change the {@link Diagram#position} and/or {@link Diagram#scale} in order to
   * maintain the desired content alignment, depending on the {@link #autoScale} policy.
   * The {@link Diagram#scale} will not be greater than 1.
   * @param {boolean} initial whether or not the content alignment is occurring for the first time. Default is false.
   * @this {Diagram}
   */
  private onContentAlignmentChanged(initial: boolean): void {
    if (!initial && !this._isInitialLayoutCompleted) return;
    if (this._isAligning) return;
    if (this._canvas === null) return;
    // ??? commenting this out allows scale to work when
    //       contentAlignment: go.Spot.Center,
    //       autoScale: go.Diagram.Uniform,
    //     are set, and nodes with a position AnimationTrigger are moved immediately.
    // if (this.animationManager.isTicking) return; // ? was .isAnimating. But such a value might break animations done during initialLayoutCompleted

    // if (this.animationManager._initialAnimationFixedBounds && this.animationManager.isAnimating) return; // DEBUG_TEST
    if (!this.documentBounds.isReal()) return;
    if (initial) {

      const initialPosition = this.initialPosition;
      if (initialPosition.isReal()) {
        this.position = initialPosition;
        return; // Do no more content aligning
      }

      const dp = Point.alloc();
      dp.setRectSpot(this.documentBounds, this.initialDocumentSpot);
      // control/view coordinates
      const vp = this.viewportBounds;
      const v = Rect.allocAt(0, 0, vp.width, vp.height);
      const p = Point.alloc();
      p.setRectSpot(v, this.initialViewportSpot);
      p.assignTo(dp.x - p.x, dp.y - p.y);
      // potential first time the position gets set.
      // If there's no content alignment spot, this will be the final position
      this.position = p;
      Rect.free(v);
      Point.free(p);
      Point.free(dp);
      this.invalidateScroll();
      this.updateScrollbars(true);
    }

    this._isAligning = true;

    let stretch = this._autoScale;
    if (initial && this._initialAutoScale !== Diagram.None) stretch = this._initialAutoScale;
    const scale = stretch !== Diagram.None ? this._computeScaleForStretch(stretch) : this.scale;
    const oldvbounds = this.viewportBounds.copy();
    const viewwidth = this._canvasWidth / scale;
    const viewheight = this._canvasHeight / scale;

    let contentAlign = this._contentAlignment;
    // only if content alignment is not a spot do we allow initial to have any effect
    const ica = this._initialContentAlignment;
    if (initial && (!contentAlign.isSpot()) && (ica.isSpot() || ica.isDefault())) contentAlign = ica.isDefault() ? Spot.Center : ica;
    this._setupPos(this.documentBounds, viewwidth, viewheight, contentAlign, initial);
    const oldscale = this.scale;
    this.scale = scale;
    this._isAligning = false;

    const newvbounds = this.viewportBounds;
    if (!newvbounds.equalsApprox(oldvbounds)) {
      this.onViewportBoundsChanged(oldvbounds, newvbounds, oldscale, false);
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {EnumValue} policy Diagram enum (Diagram.None, .Uniform, or .UniformToFill)
   * @return {number}
   */
  private _computeScaleForStretch(policy: EnumValue): number {
    const defaultScale = this._defaultScale;
    if (this._canvas === null) return defaultScale;
    this.ensureDocumentBounds();

    // Use diagram Bounds to determine if we need to scale up or down.
    const db = this.documentBounds;
    if (!db.isReal()) return defaultScale; // It's possible the documentBounds have never been computed

    const dw = db.width;
    const dh = db.height;
    const vw = this._canvasWidth + (this._hasVScroll ? this._scrollBarBreadth : 0);
    const vh = this._canvasHeight + (this._hasHScroll ? this._scrollBarBreadth : 0);

    const widthscale = vw / dw;
    const heightscale = vh / dh;

    // A scale calculated by Uniform or UniformToFill cannot be greater than 1
    let scale;
    if (policy === Diagram.Uniform) {
      // uniform: expand until the larger side exists within the bounds
      scale = Math.min(heightscale, widthscale);
      if (scale > defaultScale) scale = defaultScale;
      if (scale < this.minScale) { scale = this.minScale; }
      if (scale > this.maxScale) { scale = this.maxScale; }
      return scale;
    } else if (policy === Diagram.UniformToFill) {
      scale = 0.0;
      // uniform to fill: expand until the smaller side exists within the bounds
      if (heightscale > widthscale) {
        scale = (vh - this._scrollBarBreadth) / dh; // width is larger so there may be a horizontal scroll
      } else {
        scale = (vw - this._scrollBarBreadth) / dw; // height is larger so there may be a vertical scroll
      }
      if (scale > defaultScale) scale = defaultScale;
      if (scale < this.minScale) { scale = this.minScale; }
      if (scale > this.maxScale) { scale = this.maxScale; }
      return scale;
    }
    // else Diagram.None
    return this.scale;
  }

  /**
   * Scales the Diagram to uniformly fit into the viewport. To have this done automatically,
   * set the Diagram's {@link #autoScale} to {@link Diagram.Uniform}.
   *
   * To Animate zoomToFit, use {@link CommandHandler#zoomToFit}.
   * @export
   * @see {@link CommandHandler#zoomToFit}
   */
  public zoomToFit(): void {
    const oldScroll = this._scrollMode;
    this._scrollMode = Diagram.DocumentScroll; // Simulate document scroll during ZoomToFit operations
    this.scale = this._computeScaleForStretch(Diagram.Uniform);
    // Force the position to move as if infinite scrolling did not exist:

    if (oldScroll !== Diagram.DocumentScroll) {
      this.onContentAlignmentChanged(false);
      this._setupPos(this.documentBounds, this._canvasWidth / this._scale, this._canvasHeight / this._scale, this._contentAlignment, false);
    }

    this._scrollMode = oldScroll;
  }

  /**
   * Modifies the {@link #scale} and {@link #position} of the Diagram
   * so that the viewport displays a given document-coordinates rectangle.
   * @export
   * @this {Diagram}
   * @param {Rect} r rectangular bounds in document coordinates.
   * @param {EnumValue=} scaling an optional value of either {@link Diagram.Uniform} (the default) or {@link Diagram.UniformToFill}.
   * @since 1.1
   */
  public zoomToRect(r: Rect, scaling?: EnumValue): void {
    if (scaling === undefined) scaling = Diagram.Uniform;
    const rw = r.width;
    const rh = r.height;
    if (rw === 0 || rh === 0 || (isNaN(rw) && isNaN(rh))) return;
    let sc = 1.0;
    if (scaling === Diagram.Uniform || scaling === Diagram.UniformToFill) {
      if (isNaN(rw)) {
        sc = (this.viewportBounds.height * this.scale) / rh;
      } else if (isNaN(rh)) {
        sc = (this.viewportBounds.width * this.scale) / rw;
      } else {
        const vw = this._canvasWidth;
        const vh = this._canvasHeight;
        if (scaling === Diagram.UniformToFill) {
          if (vh / rh > vw / rw) {
            sc = (vh - (this._hasHScroll ? this._scrollBarBreadth : 0)) / rh;  // might be a horizontal scrollbar
          } else {
            sc = (vw - (this._hasVScroll ? this._scrollBarBreadth : 0)) / rw;  // might be a vertical scrollbar
          }
        } else {
          sc = Math.min(vh / rh, vw / rw);
        }
      }
    }

    this.scale = sc;
    this.position = new Point(r.x, r.y);
  }

  /**
   * @hidden @internal
   */
  /* noexport */get isAligning(): boolean { return this._isAligning; }
  set isAligning(value: boolean) { this._isAligning = value; }

  /**
   * Aligns the Diagram's {@link #position} based on a desired document {@link Spot} and viewport {@link Spot}.
   * @param {Spot} documentspot
   * @param {Spot} viewportspot
   * @export
   * @this {Diagram}
   */
  public alignDocument(documentspot: Spot, viewportspot: Spot): void {
    this.ensureDocumentBounds();
    const doc = this.documentBounds;
    const view = this.viewportBounds;
    this.position = new Point(doc.x + (documentspot.x * doc.width + documentspot.offsetX) - (viewportspot.x * view.width - viewportspot.offsetX),
      doc.y + (documentspot.y * doc.height + documentspot.offsetY) - (viewportspot.y * view.height - viewportspot.offsetY));
  }

  /**
   * @hidden
   * Undocumented.
   * Sets the GraphObject on which to focus the viewport.
   * @export
   * @this {Diagram}
   * @param {GraphObject | null} obj
   * @since 2.1
   */
  public focusObject(obj: GraphObject | null): void {
    if (obj instanceof GraphObject) {
      this._focusObject = obj;
      const pt = Point.alloc();
      this._focusViewportPoint = this.transformDocToView(obj.getDocumentPoint(Spot.TopLeft, pt));
      Point.free(pt);
    } else {
      this._focusObject = null;
      this._focusViewportPoint = Point.NoPoint;
    }
  }

  /**
   * @hidden @internal
   * Code to keep the Diagram constrained
   * @this {Diagram}
   * @param {Rect} bounds typically documentBounds.
   * @param {number} viewwidth
   * @param {number} viewheight
   * @param {Spot} contentAlignment
   * @param {boolean} initial
   */
  private _setupPos(bounds: Rect, viewwidth: number, viewheight: number, contentAlignment: Spot, initial: boolean): void {
    const pos = Point.allocAt(this._position.x, this._position.y);
    let px = pos.x;
    let py = pos.y;

    if (this._focusObject !== null) {
      let objpt = Point.alloc();
      objpt = this._focusObject.getDocumentPoint(Spot.TopLeft, objpt);
      px = objpt.x - this._focusViewportPoint.x / this.scale;
      py = objpt.y - this._focusViewportPoint.y / this.scale;
      contentAlignment = Spot.None;
      Point.free(objpt);
    }

    if (initial || this.scrollMode === Diagram.DocumentScroll) { // as opposed to Diagram.InfiniteScroll
      // This is where content alignment, if a real spot, is taken care of
      if (contentAlignment.isSpot()) {
        // constrain position so that there edge of the Diagram cannot go past the edge of the view
        if (viewwidth > bounds.width) {
          px = bounds.x + (contentAlignment.x * bounds.width + contentAlignment.offsetX) - (contentAlignment.x * viewwidth - contentAlignment.offsetX);
        }
        if (viewheight > bounds.height) {
          py = bounds.y + (contentAlignment.y * bounds.height + contentAlignment.offsetY) - (contentAlignment.y * viewheight - contentAlignment.offsetY);
        }
      }

      const marg = this._scrollMargin;
      // constrain position so that there edge of the Diagram cannot go past the edge of the view
      const scrollmargWidth = viewwidth - bounds.width;
      if (viewwidth < bounds.width + marg.left + marg.right) {
        let x = px + viewwidth / 2;
        x = Math.min(x, bounds.right + Math.max(scrollmargWidth, marg.right) - viewwidth / 2);
        x = Math.max(x, bounds.left - Math.max(scrollmargWidth, marg.left) + viewwidth / 2);
        px = x - viewwidth / 2;
      } else { // contained entirely within view
        if (px > bounds.left) px = bounds.left;
        else if (px < bounds.right - viewwidth) px = bounds.right - viewwidth;
      }
      const scrollmargHeight = viewheight - bounds.height;
      if (viewheight < bounds.height + marg.top + marg.bottom) {
        let y = py + viewheight / 2;
        y = Math.min(y, bounds.bottom + Math.max(scrollmargHeight, marg.bottom) - viewheight / 2);
        y = Math.max(y, bounds.top - Math.max(scrollmargHeight, marg.top) + viewheight / 2);
        py = y - viewheight / 2;
      } else { // contained entirely within view
        if (py > bounds.top) py = bounds.top;
        else if (py < bounds.bottom - viewheight) py = bounds.bottom - viewheight;
      }
    }

    // end constrain code
    pos.x = isFinite(px) ? px : -this._padding.left;
    pos.y = isFinite(py) ? py : -this._padding.top;

    if (this.positionComputation !== null) {
      const pc = this.positionComputation(this, pos);
      pos.x = pc.x;
      pos.y = pc.y;
    }
    if (this._animationManager._pendingAnimation) {
      this._animationManager.addDiagramPos(this._position, pos);
    }
    this._position.assignTo(pos.x, pos.y);

    Point.free(pos);
  }

  /**
   * This convenience function finds the front-most {@link Part}
   * that is at a given point that might be selectable and that is not in a temporary layer.
   *
   * This just calls {@link #findObjectAt} with appropriate arguments,
   * but ignoring Layers that are {@link Layer#isTemporary}.
   * @export
   * @this {Diagram}
   * @param {Point} p a Point in document coordinates.
   * @param {boolean=} selectable Whether to only consider parts that are {@link Part#selectable}. The default is true.
   * @return {Part}
   * @see {@link #findObjectAt}
   * @see {@link #findObjectsAt}
   * @see {@link #findObjectsIn}
   * @see {@link #findObjectsNear}
   * @see {@link #findPartsAt}
   * @see {@link #findPartsIn}
   * @see {@link #findPartsNear}
   */
  public findPartAt(p: Point, selectable?: boolean): Part | null {
    if (selectable === undefined) selectable = true;
    if (selectable) {
      const obj = this.findObjectAtNoTemp(p,
        function(x) { return x.part; },
        function(x) { return (x as Part).canSelect(); });
      if (obj instanceof Part) return obj;
    } else {
      const obj = this.findObjectAtNoTemp(p,
        function(x) { return x.part; });
      if (obj instanceof Part) return obj;
    }
    return null;
  }

  // In the future we might add an optional final argument
  // that serves as a predicate to determine which layers to act on

  /**
   * Find the front-most {@link GraphObject} at the given point in document coordinates.
   *
   * If {@link Layer#visible} is false, this method will not find any objects in that layer.
   * However, {@link Layer#opacity} does not affect this method.
   *
   * Example usage:
   * ```js
   * // Returns the top-most object that is a selectable Part, or null if there isn't one
   * myDiagram.findObjectAt(
   *   myDiagram.lastInput.documentPoint,
   *   // Navigation function
   *   function(x) { return x.part; },
   *   // Because of the navigation function, x will always be a Part.
   *   function(x) { return x.canSelect(); }
   * );
   * ```
   * @export
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @return {GraphObject} The first GraphObject returned by the *navig* function
   * and satisfying the *pred* function that is at the point *p*, in Z-order from front to back,
   * or else null if nothing is found.
   * @see {@link #findObjectsAt}
   * @see {@link #findObjectsIn}
   * @see {@link #findObjectsNear}
   * @see {@link #findPartAt}
   */
  public findObjectAt<T extends GraphObject>(p: Point,
                                             navig?: ((a: GraphObject) => (T | null)) | null,
                                             pred?: ((a: T) => boolean) | null): (T | null) {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible) continue;
      const obj = lay.findObjectAt(p, navig, pred);
      if (obj !== null) return obj;
    }
    return null;
  }

  /**
   * @hidden @internal
   * like findObjectAt but excludes temporary layers
   * Find the front-most {@link GraphObject}
   * at the given point in document coordinates.
   *
   * If {@link Layer#visible} is false, this method will not find any objects in that layer.
   * However, {@link Layer#opacity} does not affect this method.
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @return {GraphObject} The first GraphObject returned by the *navig* function
   * and satisfying the *pred* function that is at the point *p*, in Z-order from front to back,
   * or else null if nothing is found.
   */
  public findObjectAtNoTemp<T extends GraphObject>(p: Point,
                                                   navig?: ((a: GraphObject) => (T | null)) | null,
                                                   pred?: ((a: T) => boolean) | null): (T | null) {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible || lay.isTemporary) continue;
      const obj = lay.findObjectAt(p, navig, pred);
      if (obj !== null) return obj;
    }
    return null;
  }

  /**
   * This convenience function finds all {@link Part}s that are at a point in document coordinates
   * and that are not in temporary layers.
   *
   * @export
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {boolean=} selectable Whether to only consider parts that are {@link Part#selectable}. The default is true.
   * @param {List.<Part>|Set.<Part>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<Part>}
   * @see {@link #findPartAt}
   * @see {@link #findPartsIn}
   * @see {@link #findPartsNear}
   * @see {@link #findObjectsAt}
   * @since 2.0
   */
  public findPartsAt<T extends Part, S extends List<T> | Set<T> = Set<T>>(p: Point, selectable?: boolean, coll?: S): S {
    if (selectable === undefined) selectable = true;
    return this.findObjectsAtNoTemp(p,
      function(x) { return x.part; },
      selectable ? function(part) { return part instanceof Part && part.canSelect(); } : null,
      coll as any // we know our predicate limits collection to Parts
    );
  }

  /**
   * @hidden @internal
   * Like findObjectsAt but excludes temporary layers
   * @this {Diagram}
   * @param {Point} p
   * @param {function(GraphObject):GraphObject | null=} navig
   * @param {function(GraphObject):boolean | null=} pred
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll
   * @return {Iterable.<GraphObject>}
   */
  public findObjectsAtNoTemp<T extends GraphObject, S extends (List<T> | Set<T>) = Set<T>>(p: Point,
                                                                                           navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                           pred?: ((a: T) => boolean) | null,
                                                                                           coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible || lay.isTemporary) continue;
      lay.findObjectsAt(p, navig, pred, coll);
    }
    return coll;
  }

  /**
   * Return a collection of the {@link GraphObject}s at the given point in document coordinates.
   *
   * If {@link Layer#visible} is false, this method will not find any objects in that layer.
   * However, {@link Layer#opacity} does not affect this method.
   *
   * Example usage:
   * ```js
   * // Returns the Nodes that are at a given point, overlapping each other
   * myDiagram.findObjectsAt(somePoint,
   *   // Navigation function -- only return Nodes
   *   function(x) { var p = x.part; return (p instanceof go.Node) ? p : null; }
   * );
   * ```
   * @export
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects returned by the *navig* function
   * and satisfying the *pred* that are located at the point *p*, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   * @see {@link #findObjectAt}
   * @see {@link #findObjectsIn}
   * @see {@link #findObjectsNear}
   * @see {@link #findPartsAt}
   */
  public findObjectsAt<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(p: Point,
                                                                                   navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                   pred?: ((a: T) => boolean) | null,
                                                                                   coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible) continue;
      lay.findObjectsAt(p, navig, pred, coll);
    }
    return coll;
  }

  /**
   * This convenience function finds {@link Part}s that are inside or that intersect
   * a given {@link Rect} in document coordinates.
   *
   * This just calls {@link #findObjectsIn} with appropriate arguments,
   * but ignoring Layers that are {@link Layer#isTemporary}.
   * @export
   * @this {Diagram}
   * @param {Rect} r a Rect in document coordinates.
   * @param {boolean=} partialInclusion Whether a Part can match if it merely intersects the rectangular area (true) or
   * if it must be entirely inside the rectangular area (false). The default value is false.
   * @param {boolean=} selectable Whether to only consider parts that are {@link Part#selectable}. The default is true.
   * @param {List.<Part>|Set.<Part>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<Part>}
   * @see {@link #findPartAt}
   * @see {@link #findPartsAt}
   * @see {@link #findPartsNear}
   * @see {@link #findObjectsIn}
   * @since 2.0
   */
  public findPartsIn<T extends Part, S extends List<T> | Set<T> = Set<T>>(r: Rect, partialInclusion?: boolean, selectable?: boolean, coll?: S): S {
    if (partialInclusion === undefined) partialInclusion = false;
    if (selectable === undefined) selectable = true;
    return this.findObjectsInNoTemp(r,
      null,
      function(part) { return part instanceof Part && (!selectable || part.canSelect()); },
      partialInclusion,
      coll as any // we know our predicate limits collection to Parts
    );
  }

  /**
   * Returns a collection of all {@link GraphObject}s that are inside or that intersect
   * a given {@link Rect} in document coordinates.
   *
   * If {@link Layer#visible} is false, this method will not find any objects in that layer.
   * However, {@link Layer#opacity} does not affect this method.
   *
   * Example usage:
   * ```js
   * // Returns the Links that intersect a given rectangle and have a certain data property
   * myDiagram.findObjectsIn(someRect,
   *   // Navigation function -- only return Links
   *   function(x) { var p = x.part; return (p instanceof go.Link) ? p : null; },
   *   // Predicate that always receives a Link, due to above navigation function
   *   function(link) { return link.data.someProp > 17; },
   *   // the links may only partly overlap the given rectangle
   *   true
   * );
   * ```
   * @export
   * @this {Diagram}
   * @param {Rect} r A Rect in document coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {boolean=} partialInclusion Whether an object can match if it merely intersects the rectangular area (true) or
   * if it must be entirely inside the rectangular area (false). The default value is false.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects returned by the *navig* function
   * and satisfying the *pred* function that are within the rectangle *r*, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   * @see {@link #findObjectAt}
   * @see {@link #findObjectsAt}
   * @see {@link #findObjectsNear}
   * @see {@link #findPartsIn}
   */
  public findObjectsIn<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(r: Rect,
                                                                                   navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                   pred?: ((a: T) => boolean) | null,
                                                                                   partialInclusion?: boolean,
                                                                                   coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = false;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible) continue;
      lay.findObjectsIn(r, navig, pred, partialInclusion, coll);
    }
    return coll;
  }

  /**
   * @hidden @internal
   * Like findObjectsIn but excludes temporary layers
   * @this {Diagram}
   * @param {Rect} r
   * @param {function(GraphObject):GraphObject | null=} navig
   * @param {function(GraphObject):boolean | null=} pred
   * @param {boolean=} partialInclusion
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll
   * @return {Iterable.<GraphObject>}
   */
  public findObjectsInNoTemp<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(r: Rect,
                                                                                         navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                         pred?: ((a: T) => boolean) | null,
                                                                                         partialInclusion?: boolean,
                                                                                         coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = false;
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible || lay.isTemporary) continue;
      lay.findObjectsIn(r, navig, pred, partialInclusion, coll);
    }
    return coll;
  }

  /**
   * This convenience function finds {@link Part}s that are within a certain distance
   * of a given point in document coordinates.
   *
   * This just calls {@link #findObjectsNear} with appropriate arguments,
   * but ignoring Layers that are {@link Layer#isTemporary}.
   * @export
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {number} dist The distance from the point.
   * @param {boolean=} partialInclusion Whether an object can match if it merely intersects the circular area (true) or
   * if it must be entirely inside the circular area (false). The default is true.
   * @param {boolean=} selectable Whether to only consider parts that are {@link Part#selectable}. The default is true.
   * @param {List.<Part>|Set.<Part>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<Part>}
   * @see {@link #findPartAt}
   * @see {@link #findPartsAt}
   * @see {@link #findPartsIn}
   * @see {@link #findObjectsNear}
   * @since 2.0
   */
  public findPartsNear<T extends Part, S extends List<T> | Set<T> = Set<T>>(p: Point, dist: number, partialInclusion?: boolean, selectable?: boolean, coll?: S): S {
    if (partialInclusion === undefined) partialInclusion = true;
    if (selectable === undefined) selectable = true;
    return this.findObjectsNearNoTemp(p,
      dist,
      null,
      function(part) { return part instanceof Part && (!selectable || part.canSelect()); },
      partialInclusion,
      coll as any // we know our predicate limits collection to Parts
    );
  }

  /**
   * Returns a collection of all {@link GraphObject}s that are within a certain distance
   * of a given point in document coordinates.
   *
   * If {@link Layer#visible} is false, this method will not find any objects in that layer.
   * However, {@link Layer#opacity} does not affect this method.
   *
   * Example usage:
   * ```js
   * // Returns the Nodes that intersect a given circular area and have a certain data property
   * myDiagram.findObjectsNear(somePoint,
   *   // The circular area is centered at somePoint and has radius 100
   *   100,
   *   // Navigation function -- only return Nodes
   *   function(x) { var p = x.part; return (p instanceof go.Node) ? p : null; },
   *   // Predicate that always receives a Node, due to above navigation function
   *   function(node) { return node.data.someProp > 17; },
   *   // the nodes may only partly overlap the given circular area
   *   true
   * );
   * ```
   * @export
   * @this {Diagram}
   * @param {Point} p A Point in document coordinates.
   * @param {number} dist The distance from the point.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * If this function returns null, the given GraphObject will not be included in the results.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {*=} partialInclusion Whether an object can match if it merely intersects the circular area (true) or
   * if it must be entirely inside the circular area (false).  The default value is true.
   * The default is true.
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll An optional collection (List or Set) to add the results to.
   * @return {Iterable.<GraphObject>} a collection of GraphObjects returned by the *navig* function
   * and satisfying the *pred* that are located near the point *p*, or else an empty collection.
   * If a List or Set was passed in, it is returned.
   * @see {@link #findObjectAt}
   * @see {@link #findObjectsAt}
   * @see {@link #findObjectsIn}
   * @see {@link #findPartsNear}
   */
  public findObjectsNear<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(p: Point,
                                                                                     dist: number,
                                                                                     navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                     pred?: ((a: T) => boolean) | null,
                                                                                     partialInclusion?: boolean | S,
                                                                                     coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = true;
    if (partialInclusion !== false && partialInclusion !== true) {  // compatibility
      if (partialInclusion instanceof List || partialInclusion instanceof Set) {
        coll = partialInclusion as S;
      }
      partialInclusion = true;
    }
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible) continue;
      lay.findObjectsNear(p, dist, navig, pred, partialInclusion, coll);
    }
    return coll;
  }

  /**
   * @hidden @internal
   * Like findObjectsNear but excludes temporary layers
   * @this {Diagram}
   * @param {Point} p
   * @param {number} dist
   * @param {function(GraphObject):GraphObject | null=} navig
   * @param {function(GraphObject):boolean | null=} pred
   * @param {*=} partialInclusion
   * @param {List.<GraphObject>|Set.<GraphObject>=} coll
   * @return {Iterable.<GraphObject>}
   */
  public findObjectsNearNoTemp<T extends GraphObject, S extends List<T> | Set<T> = Set<T>>(p: Point,
                                                                                           dist: number,
                                                                                           navig?: ((a: GraphObject) => (T | null)) | null,
                                                                                           pred?: ((a: T) => boolean) | null,
                                                                                           partialInclusion?: boolean | S,
                                                                                           coll?: S): S {
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (partialInclusion === undefined) partialInclusion = true;
    if (partialInclusion !== false && partialInclusion !== true) {  // compatibility
      if (partialInclusion instanceof List || partialInclusion instanceof Set) {
        coll = partialInclusion as S;
      }
      partialInclusion = true;
    }
    if (!(coll instanceof List) && !(coll instanceof Set)) {
      coll = new Set<T>() as S;
    }
    this.validateParts();
    const lit = this._layers.iteratorBackwards;
    while (lit.next()) {
      const lay = lit.value;
      if (!lay.visible || lay.isTemporary) continue;
      lay.findObjectsNear(p, dist, navig, pred, partialInclusion, coll);
    }
    return coll;
  }

  // Events:

  /**
   * @hidden @internal
   * Sometimes you may want to pass additional events from the Diagram div onto GoJS
   * such as the dragover event, where you may wish to do:
   * ```js
   *    myDiagram.acceptEvent(event);
   *    myDiagram.doMouseMove();
   * ```
   * so that dragover events are treated like normal mouse-moves (for instance for highlighting purposes).
   *
   * Common code that occurs on every event. Some methods may go on to change
   * the properties from these defaults (ie, mouseWheel changes delta)
   * @expose
   * @param {Event} event
   * @return {InputEvent}
   */
  public acceptEvent(event: Event): InputEvent {
    return this._commonEvent(event, (event instanceof MouseEvent));
  }

  /**
   * @hidden @internal
   * Common code that occurs on every event. Some methods may go on to change
   * the properties from these defaults (ie, mouseWheel changes delta)
   * @param {Event} e
   * @param {boolean} mouse
   * @return {InputEvent} n
   */
  public _commonEvent(e: Event, mouse: boolean): InputEvent {
    const n = this._previousInput;
    this._previousInput = this._lastInput;
    this._lastInput = n;

    n.diagram = this;
    n.event = e;
    if (mouse) {
      this._setDocView(e, n);
    } else {
      n.viewPoint = this._previousInput.viewPoint;
      n.documentPoint = this._previousInput.documentPoint;
    }
    let mod = 0;
    // this is to allow the same code for MouseEvent, KeyboardEvent, and other unpredictable types.
    const evt = e as any;
    if (evt.ctrlKey) mod += 1;
    if (evt.altKey) mod += 2;
    if (evt.shiftKey) mod += 4;
    if (evt.metaKey) mod += 8;
    n.modifiers = mod;
    n.button = evt.button;
    // record the current state of the buttons, if available
    // if not available, this keeps the value of buttons from _previousInput
    if (evt.buttons !== undefined && !Util.isIE9) n.buttons = evt.buttons;  // don't set if evt.buttons is zero -- needed for IE9
    // Place this elsewhere?
    // CTRL left-click on a Mac is equivalent to a right-click
    if (Util.isMac && evt.button === 0 && evt.ctrlKey) {
      n.button = 2;
    }
    n.down = false;
    n.up = false;
    n.clickCount = 1;
    n.delta = 0;
    n.handled = false;
    n.bubbles = false;
    n.timestamp = e.timeStamp;
    n.isMultiTouch = false;
    n.targetDiagram = this._findTargetDiagram(e);
    n.targetObject = null;
    return n;
  }

  /**
   * @hidden @internal
   * @param {Event} e
   * @return {Diagram}
   */
  public _findTargetDiagram(e: Event): Diagram | null {
    let diag: Diagram = e.target._diagram;
    if (!diag) {
      let path = e.path;
      if (!path && (typeof e.composedPath === 'function')) path = e.composedPath();
      if (path && path[0]) diag = path[0]._diagram;
    }
    if (diag) return diag;
    return null;  // don't return undefined
  }

  /**
   * @hidden @internal
   * @param {TouchEvent | PointerEvent} e original event
   * @param {Event} evt the event (if a pointerEvent) or an event in the touch list if a touchEvent
   * @param {boolean} isMulti whether or not its multi touch
   * @return {InputEvent} n the event to setup
   * @this {Diagram}
   */
  public _commonTouchStart(e: TouchEvent | PointerEvent, evt: Event | Touch, isMulti: boolean): InputEvent {
    const n = this._commonTouchSetup(e,
      true,
      false,
      false,
      true,
      isMulti);
    // bubbles arg: single finger touch start must bubble to allow canceling of pinches
    // If you wanted to zoom the page instead of the Diagram, etc

    this._setDocView(evt, n);

    n.targetDiagram = this._findTargetDiagram(e);
    n.targetObject = null;
    if (!isMulti) n.clone(this._firstInput);
    // DraggingTool._source = null;
    return n;
  }

  /**
   * @hidden @internal
   * @param {TouchEvent | PointerEvent} e original event
   * @param {Event} evt the event (if a pointerEvent) or an event in the touch list if a touchEvent
   * @param {boolean} isMulti whether or not its multi touch
   * @return {InputEvent} n the event to setup
   * @this {Diagram}
   */
  public _commonTouchMove(e: TouchEvent | PointerEvent, evt: Event | Touch, isMulti: boolean): InputEvent {
    let targ: Element = null;
    const n = this._commonTouchSetup(e,
      false,
      false,
      false,
      false,
      isMulti);

    if (evt !== null) {
      targ = root.document.elementFromPoint(evt.clientX, evt.clientY);
      let diag: Diagram;
      let mouseevent;
      if (targ && targ._diagram) {
        mouseevent = evt;
        diag = targ._diagram;
      } else {
        mouseevent = ((e as any)['targetTouches'] !== undefined) ? (e as TouchEvent).targetTouches[0] : e;
        diag = this;
      }
      n.targetDiagram = diag;
      this._setDocView(mouseevent, n);
    } else if (this._previousInput !== null) {
      n.documentPoint = this._previousInput.documentPoint;
      n.viewPoint = this._previousInput.viewPoint;
      n.targetDiagram = this._previousInput.targetDiagram;
    } else if (this._firstInput !== null) {
      n.documentPoint = this._firstInput.documentPoint;
      n.viewPoint = this._firstInput.viewPoint;
      n.targetDiagram = this._firstInput.targetDiagram;
    }

    n.targetObject = null;
    return n;
  }

  /**
   * @hidden @internal
   * @param {TouchEvent|PointerEvent} e event
   * @param {boolean} down
   * @param {boolean} up
   * @param {boolean} handled
   * @param {boolean} bubbles
   * @param {boolean} isMulti
   * @return {InputEvent}
   * @this {Diagram}
   */
  public _commonTouchSetup(e: TouchEvent | PointerEvent, down: boolean, up: boolean, handled: boolean, bubbles: boolean, isMulti: boolean): InputEvent {
    const n = this._previousInput;
    this._previousInput = this._lastInput;
    this._lastInput = n;
    n.diagram = this;
    n.clickCount = 1;
    n.delta = 0;
    let mod = 0;
    if (e.ctrlKey) mod += 1;
    if (e.altKey) mod += 2;
    if (e.shiftKey) mod += 4;
    if (e.metaKey) mod += 8;
    n.modifiers = mod;
    n.button = 0;
    n.buttons = 1;
    n.event = e;
    n.timestamp = e.timeStamp;
    if (this._allowsPointers && e instanceof root['PointerEvent'] && (e as PointerEvent).pointerType !== 'touch') {
      n.button = (e as PointerEvent).button;
      // don't set if (e as PointerEvent).buttons is zero -- needed for IE9
      if ((e as PointerEvent).buttons !== undefined && !Util.isIE9) n.buttons = (e as PointerEvent).buttons;
      if (Util.isMac && (e as PointerEvent).button === 0 && (e as PointerEvent).ctrlKey) {
        n.button = 2;
      }
    }
    n.down = down;
    n.up = up;
    n.handled = handled;
    n.bubbles = bubbles;
    n.isMultiTouch = isMulti;
    return n;
  }

  /**
   * @hidden @internal
   * Determines whether or not to prevent the event from bubbling
   * @this {Diagram}
   * @param {InputEvent} n
   * @param {Event} e
   * @return {boolean}
   */
  public _maybeHandled(n: InputEvent, e: Event): boolean {
    if (!n.bubbles) {
      if (Debug && Debug.handleMessages) Util.trace('handled ' + e.type + ' ' + this.currentTool.name + ' ' + n.toString());
      if (e.stopPropagation !== undefined) {
        e.stopPropagation();
      }
      if (e.cancelable !== false) e.preventDefault();
      e.cancelBubble = true;
      return false;
    } else {
      if (Debug && Debug.handleMessages) Util.trace('NOT handled ' + e.type + ' ' + n.toString());
      return true;
    }
  }

  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   * @return {boolean}
   */
  public keyDown(this: HTMLCanvasElement, e: KeyboardEvent): boolean {
    const diagram = this._diagram;
    if (!this._diagram.isEnabled) return false;
    const n = diagram._commonEvent(e, false);
    n.key = String.fromCharCode((e as KeyboardEvent).which);
    n.down = true;

    switch ((e as KeyboardEvent).which) {
      case 8:
        n.key = 'Backspace'; break;
      case 33:
        n.key = 'PageUp'; break;
      case 34:
        n.key = 'PageDown'; break;
      case 35:
        n.key = 'End'; break;
      case 36:
        n.key = 'Home'; break;
      case 37:
        n.key = 'Left'; break;
      case 38:
        n.key = 'Up'; break;
      case 39:
        n.key = 'Right'; break;
      case 40:
        n.key = 'Down'; break;
      case 45:
        n.key = 'Insert'; break;
      case 46:
        n.key = 'Del'; break;
      case 48:
        n.key = '0'; break;
      case 187: // the plus/equals key chrome/ie
      case 61: // the plus/equals key firefox
      case 107: //  in keypad
        n.key = 'Add'; break;
      case 189: // the minus/underscore key chrome/ie
      case 173: // the minus/underscore key firefox
      case 109: //  in keypad
        n.key = 'Subtract'; break;
      case 27:
        n.key = 'Esc'; break;
    }

    diagram.doKeyDown();

    return diagram._maybeHandled(n, e);
  }

  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   * @return {boolean}
   */
  public keyUp(this: HTMLCanvasElement, e: KeyboardEvent): boolean {
    const diagram = this._diagram;
    if (!diagram.isEnabled) return false;
    const n = diagram._commonEvent(e, false);
    n.key = String.fromCharCode((e as KeyboardEvent).which);
    n.up = true;
    // We must check to see if 'which' is one of the special keys. Otherwise
    // PageUp will parse as "!", etc.
    switch ((e as KeyboardEvent).which) {
      case 8:
        n.key = 'Backspace'; break;
      case 33:
        n.key = 'PageUp'; break;
      case 34:
        n.key = 'PageDown'; break;
      case 35:
        n.key = 'End'; break;
      case 36:
        n.key = 'Home'; break;
      case 37:
        n.key = 'Left'; break;
      case 38:
        n.key = 'Up'; break;
      case 39:
        n.key = 'Right'; break;
      case 40:
        n.key = 'Down'; break;
      case 45:
        n.key = 'Insert'; break;
      case 46:
        n.key = 'Del'; break;
      case 93: // context menu button
        // emits an additional event (mousedown) unless the keyup event is handled
        // We do not want this additional event (will be prevented since bubbles = false)
        break;
    }
    diagram.doKeyUp();
    return diagram._maybeHandled(n, e);
  }

  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   * @return {boolean}
   */
  public doMouseEnter(this: HTMLCanvasElement, e: Event): boolean {
    const diagram = this._diagram;
    if (!diagram.isEnabled) return false;
    const n = diagram._commonEvent(e, true);
    if (diagram.mouseEnter !== null) diagram.mouseEnter(n);
    return diagram._maybeHandled(n, e);
  }

  /**
   * @hidden @internal
   * @this {HTMLCanvasElement}
   * @param {Event} e
   * @return {boolean}
   */
  public doMouseLeave(this: HTMLCanvasElement, e: Event): boolean {
    const diagram = this._diagram;
    if (!diagram.isEnabled) return false;
    const n = diagram._commonEvent(e, true);
    if (diagram.mouseLeave !== null) diagram.mouseLeave(n);
    return diagram._maybeHandled(n, e);
  }
  /**
   * @hidden @internal
   * Returns a mouse position relative to the canvas.
   * Does not account for margin/width/padding on the Canvas,
   * better to put those attributes on a div that surrounds the Diagram.
   * @expose
   * @this {Diagram}
   * @param {Event} e
   * @return {Point}
   */
  public getMouse(e: Event | Touch): Point {
    const can = this._canvas;
    if (can === null) return new Point(0, 0);
    const bbox = can.getBoundingClientRect();
    const mx = e.clientX - bbox.left * (this._canvasWidth / bbox.width);
    const my = e.clientY - bbox.top * (this._canvasHeight / bbox.height);
    if (this._transform !== null) {
      return new Point(mx, my).transformInverted(this._transform);
    }
    return new Point(mx, my);
  }

  /**
   * @hidden @internal
   * Called as an alternative to getMouse
   * Given an Event and an InputEvent, set the .documentPoint and .viewportPoint without allocating anything
   * @this {Diagram}
   * @param {Event} e
   * @param {InputEvent} n
   */
  public _setDocView(e: Event | Touch, n: InputEvent): void {
    // avoids creation of new points by not calling getMouse/transformPoint
    const can = this._canvas;
    const canwidth = this._canvasWidth;
    const canheight = this._canvasHeight;
    let mx = 0.0;
    let my = 0.0;
    if (can !== null) {
      const bbox = can.getBoundingClientRect();
      mx = e.clientX - bbox.left * (canwidth / bbox.width);
      my = e.clientY - bbox.top * (canheight / bbox.height);
    }
    n.viewPoint.assignTo(mx, my);
    if (this._transform !== null) {
      const pt = Point.allocAt(mx, my);
      this._transform.invertedTransformPoint(pt);
      n.documentPoint.assign(pt);
      Point.free(pt);
    } else {
      n.documentPoint.assignTo(mx, my);
    }
  }

  /**
   * @hidden @internal
   * Given an Event and an index (that is either a touch in a touchlist or a pointer in a pointer list),
   * set a given Point to the appropriate viewpoint.
   * @this {Diagram}
   * @param {Event} e
   * @param {number} n
   * @param {Point} p
   */
  public _getMultiTouchViewpoint(e: TouchEvent, n: number, p: Point): void {
    let evt = null;
    if (e.targetTouches !== undefined) {
      if (e.targetTouches.length < 2) return;
      evt = e.targetTouches[n];
    } else if (this._pointers2[0] !== null) {
      evt = this._pointers2[n];
    } else {
      return;
    }

    // avoids creation of new points by not calling getMouse/transformPoint
    const can = this._canvas;
    if (can !== null) {
      const bbox = can.getBoundingClientRect();
      p.assignTo(
        evt.clientX - bbox.left * (this._canvasWidth / bbox.width),
        evt.clientY - bbox.top * (this._canvasHeight / bbox.height));
    }
  }

  /**
   * @hidden
   * Requests that the Diagram updates its {@link #documentBounds} in the near-future.
   * @export
   * @this {Diagram}
   */
  public invalidateDocumentBounds(): void {
    if (this._invalidDocBounds) return;
    this._invalidDocBounds = true;
    this.requestUpdate(true);
  }

  /**
   * @hidden @internal
   * Call {@link #computeBounds} and set {@link #documentBounds}
   * @this {Diagram}
   */
  public updateBounds(): void {
    if (!this._isUpdating) this.validateParts();
    this.ensureDocumentBounds();
  }

  /**
   * @hidden
   * Undocumented
   *
   * Invalidates all non-layout diagram state and forces an immediate redraw.
   * Because this can be very inefficent, to discourage its use, it remains an undocumented part of the API.
   * @export
   * @this {Diagram}
   */
  public redraw(): void {
    if (this._isAligning || this._isUpdating) return;
    this.invalidateDraw();
    this.invalidateScroll();
    this.invalidateViewportObjects();
    this.invalidateDocumentBounds();
    // do it immediately
    this.maybeUpdate();
  }

  /**
   * @hidden @internal
   * @export
   * @this {Diagram}
   * @return {boolean}
   */
  public isUpdateRequested(): boolean {
    return this._updateQueued;
  }

  /**
   * Updates the diagram immediately, then resets initialization flags so that
   * actions taken in the argument function will be considered part of Diagram initialization,
   * and will participate in initial layouts, {@link #initialAutoScale}, {@link #initialContentAlignment}, etc.
   *
   * This is useful in situations where you do not wish for the first content added to the diagram to be considered
   * the "initial" content, such as with a Node that represents a "Loading" bar.
   * @export
   * @this {Diagram}
   * @param {function()|null=} func an optional function of actions to perform as part of another diagram initialization.
   * @since 1.1
   */
  public delayInitialization(func?: (() => void) | null): void {
    if (func === undefined) func = null;
    // if we begin animation in this delayInit, then we want to reset that too
    const AM = this.animationManager;
    const old = AM.isEnabled;
    AM.stopAnimation();
    AM.isEnabled = false;
    this.doUpdate();
    this._isInitialLayoutCompleted = false;
    AM.isEnabled = old;

    if (func !== null) {
      Util.setTimeout(function() {
        AM.prepareAutomaticAnimation('Model'); // Simulate model load animation
        func();
      }, 1);
    }
  }

  /**
   * Usage of this method is uncommon and may affect performance,
   * for efficiency do not call this method unless you have a well-defined need.
   * Normally, GoJS updates the diagram automatically, and completeing a transaction ensures an immediate update.
   *
   * The most common reason to call this method when the HTML Div has changed size but the window has not changed size,
   * and the Diagram needs to be notified of this DOM change. See an example of resizing diagrams
   * <a href="../../intro/resizing.html">here</a>.
   *
   * Requests that in the near-future the diagram makes sure all GraphObjects are arranged,
   * recomputes the document bounds, updates the scrollbars, and redraws the viewport.
   *
   * @param {boolean=} alwaysQueueUpdate If true the Diagram will queue another update,
   * even if an update is already occurring. The default value is false.
   * Side effects in an "InitialLayoutCompleted" DiagramEvent listener might necessitate setting this parameter.
   * @since 1.6
   * @export
   * @this {Diagram}
   */
  public requestUpdate(alwaysQueueUpdate?: boolean): void {
    if (alwaysQueueUpdate === undefined) alwaysQueueUpdate = false;
    if (this._updateQueued === true) return;
    if (this._isAligning || (alwaysQueueUpdate === false && this._isUpdating)) return;
    this._updateQueued = true;
    const dia = this;
    root.requestAnimationFrame(
      function() {
        // Are we still queued?
        if (!dia._updateQueued) { return; } // no need!
        dia.maybeUpdate();
      });
  }

  /**
   * A special update called when only the viewport has been modified
   * @hidden @internal
   * @this {Diagram}
   * @param {Rect} oldv old viewport bounds.
   * @param {Rect} newv new viewport bounds.
   */
  private updateViewport(oldv: Rect, newv: Rect): void {
    if (this._isUpdating) return;
    this._updateQueued = false;
    if (this._div === null) return;
    this._isUpdating = true;
    this.clearDelayedGeometries();

    // make sure documentBounds are sane
    // Note that we do not compute the bounds constantly because we only want them to update on a transaction end
    if (!this.documentBounds.isReal()) this.setDocumentBounds(this.computeBounds());

    const can = this._canvas;
    if (can === null) return;
    if (can instanceof SVGSurface) return; // NYI: query this._svg instead?
    const pixelratio = this._pixelRatio;
    const width = this._canvasWidth * pixelratio;
    const height = this._canvasHeight * pixelratio;
    const scale = this.scale * pixelratio;

    // Difference between old and new positions, scaled to screen pixels and rounded to the nearest pixel
    // outermost math.round for floating point issues only
    const x = Math.round(Math.round(newv.x * scale) - Math.round(oldv.x * scale));
    const y = Math.round(Math.round(newv.y * scale) - Math.round(oldv.y * scale));

    // Draw the rect onto an in-memory canvas
    const tempCan = this._drawCollectionCanvas;
    const tempCtx = this._drawCollectionCtx;

    if (tempCan.width !== width) tempCan.width = width;
    if (tempCan.height !== height) tempCan.height = height;
    tempCtx.clearRect(0, 0, width, height);

    // 190x70 for wmq size
    const wmwidth = 190 * pixelratio;
    const wmheight = 70 * pixelratio;

    // Positive if scrolling rightward/downward
    // Zero if scrolling leftward/upward
    const sx = Math.max(x, 0);
    const sy = Math.max(y, 0);
    const sw = Math.floor(width - sx);
    const sh = Math.floor(height - sy);
    tempCtx.drawImage(can.domElement,
      sx, sy, sw, sh, // Source
      0, 0, sw, sh);  // Destination
    this._helper.wmq(this) && tempCtx.clearRect(0, 0, wmwidth, wmheight);

    // We need the rects in both document and viewport coordinates
    const rects = Util.tempArray<Rect>();
    const viewRects = Util.tempArray<Rect>();
    const abx = Math.abs(x);
    const aby = Math.abs(y);
    let temp = (sx === 0) ? 0 : width - abx;
    const p1 = Point.allocAt(temp, 0);
    const p2 = Point.allocAt(abx + temp, height);
    viewRects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));

    const t = this._transform;
    t.reset();
    t.scale(pixelratio, pixelratio);
    if (this._scale !== 1) { t.scale(this._scale); }
    const pos = this._position;
    if ((pos.x !== 0 || pos.y !== 0) && isFinite(pos.x) && isFinite(pos.y)) {
      t.translate(-pos.x, -pos.y);
    }
    p1.transformInverted(t);
    p2.transformInverted(t);
    rects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));

    temp = (sy === 0) ? 0 : height - aby;
    p1.assignTo(0, temp);
    p2.assignTo(width, aby + temp);
    viewRects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));

    p1.transformInverted(t);
    p2.transformInverted(t);
    rects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));

    if (this._helper.wmq(this)) {
      const tempx = ((x > 0) ? 0 : - x);
      const tempy = ((y > 0) ? 0 : - y);
      p1.assignTo(tempx, tempy);
      p2.assignTo(wmwidth + tempx, wmheight + tempy); // displace by wmq size
      viewRects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));
      p1.transformInverted(t);
      p2.transformInverted(t);
      rects.push(new Rect(Math.min(p1.x, p2.x), Math.min(p1.y, p2.y), Math.abs(p1.x - p2.x), Math.abs(p1.y - p2.y)));
    }

    Point.free(p1);
    Point.free(p2);
    this.updateScrollbars(false);
    this.redrawViewport(rects, viewRects, x, y);
    Util.freeArray(rects);
    Util.freeArray(viewRects);
    this._isUpdating = false;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   */
  public maybeUpdate(): void {
    // unlike update, this method tests several conditions before it does anything
    // this should almost always be used internally instead of update
    if (this._firstUpdate && !this._updateQueued) return;
    // This flag stops hard (non-queued) updates during initialization from prematurely updating the diagram
    if (this._firstUpdate) this._firstUpdate = false;
    this.doUpdate();
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {boolean} validateparts
   */
  public updateScrollbars(validateparts: boolean): void {
    // skip while animating (as of 1.6) (only default animation as of 2.1)
    //   (this._isInitialLayoutCompleted && this.animationManager._pendingAnimation)
    //   is a compromise: don't modify scrollbars during collapse/expand animations, but we do want them to update
    //   for the initial animation
    // DEBUG_TEST or can this one permanently change?
    // if (this._isInitialLayoutCompleted && this.animationManager._pendingAnimation) return;
    if (this.animationManager.defaultAnimation.isAnimating) return;

    if (!this._isAligning && this._invalidScrollBars) {
      if (!this.updateScroll()) {
        if (validateparts) this.validateParts(); // A Canvas resize may mean that parts have moved during an update

        // This replaces the call that used to be in onViewportBoundsChanged
        // It is no longer in onViewportBoundsChanged because that created redundancy with the position setter
        // Since both onContentAlignmentChanged and the position setter call _setupPos()
        this.onContentAlignmentChanged(false);
      }
    }
  }

  /**
   * @hidden @internal
   * Ensure that all GraphObjects are arranged, update the document bounds and the scrollbars,
   * and draw all objects in the viewport.
   *
   * **Caution:** this method can be very expensive in time and space.
   * Avoid calling this method except in infrequent circumstances where redrawing immediately is necessary.
   * Under most circumstances you should allow the transaction to perform its normal updating process.
   * Calling {@link #requestUpdate} is the usually the right thing to do.
   *
   * @this {Diagram}
   */
  public doUpdate(avoidDraw?: boolean): void {
    if (this._isUpdating) return;
    this._updateQueued = false;
    if (this._div === null && !this._viewSize.isReal()) return;
    this._isUpdating = true;
    const AM = this.animationManager;

    const pics = this._picturesToInvalidate;
    if (!AM.isTicking && pics.length !== 0) {
      const arr = pics._arr;
      const len = arr.length;
      for (let i = 0; i < len; i++) {
        const pic = arr[i];
        pic.setInvalidMeasure(false);
        pic.invalidateMeasure();
      }
      pics.clear();
    }

    const bundles = this._invalidLinkBundles;
    if (bundles.count > 0) {
      bundles.each(function(b) { b.validate(); });
      bundles.clear();
    }

    let oldskip = false;
    let wasanimating = false; // isAnimating might have changed!
    if (AM.isAnimating) {
      wasanimating = true;
      oldskip = this.skipsUndoManager;
      this.skipsUndoManager = true;
    }
    if (!AM._pendingAnimation) {
      // It's possible the viewportbounds changed. Possible to optimzie this in the future too
      // might invalidate scrollbars and resize the diagram,
      // potentially need to do this before and after a layout
      this.maybeResize();
    }

    this.updateScrollbars(false);

    if (this._grid !== null) {
      if (this._grid.visible && !this._gridWasVisible) {
        this._updateBackgroundGrid();
        this._gridWasVisible = true;
      }
      if (!this._grid.visible && this._gridWasVisible) {
        this._gridWasVisible = false;
      }
    }

    // ?? is this necessary? May validate parts that then get invalidated by layout right afterwards.
    this.validateParts();

    let layoutCompleted = false;
    // by default redraws without performing any layouts, except needing initial layout
    if (!this._isInitialLayoutCompleted || this._anyLayoutInvalid) {
      if (!this._isInitialLayoutCompleted) {
        this.startTransaction('Initial Layout');
        if (AM.isEnabled === false) AM.stopAnimation();
        this.updateAllLayouts(false); // perform all invalid layouts right now
      } else {
        this.updateAllLayouts(!this._doNonRealtimeLayouts);
      }
      layoutCompleted = true;
    }
    this._doNonRealtimeLayouts = false;

    this.validateParts();
    // check isAnimating too, in case the doUpdate happens outside of a tick
    // ?? But this won't allow onContentAlignmentChanged during an animation, which means autoScale/etc won't work during an animation
    //    so changed back to just checking _skipsBoundsUpdates

    // ?? commenting out the if statement fixes updating of position in autoScale: true animations, but not scale
    if (!this._skipsBoundsUpdates) this.updateBounds();

    // But this is better before "if (layoutCompleted)" because
    // scrollbars need to be updated to correctly calculate
    // onInitialLayoutCompleted -> onContentAlignmentChanged -> _computeScaleForStretch
    this.updateScrollbars(true);

    let wasInitLayoutCompleted = false;
    if (layoutCompleted) {
      // this is the first time we've computed the diagram bounds since finishing a layout;
      // only raise the Diagram.InitialLayoutCompleted and Diagram.LayoutCompleted events after computing new diagram bounds
      if (!this._isInitialLayoutCompleted) {
        this._isInitialLayoutCompleted = true;
        wasInitLayoutCompleted = true;
        this.onInitialLayoutCompleted();
      }
      this.onLayoutCompleted();
    } else if (AM._initialAnimationOngoing && AM._useInitialPositioning) {
      // New in 2.1:
      // During the initial animation, if the position had not been set to a new value
      // We constantly re-align to the initial... position properties.


      // This replicates much of onLayoutCompleted
      // @@@ should this go in onContentAlignmentChanged(true) also?
      if (this._initialAutoScale !== Diagram.None) {
        this.scale = this._computeScaleForStretch(this._initialAutoScale);
      } else if (this._autoScale !== Diagram.None) {
        this.scale = this._computeScaleForStretch(this._autoScale);
      } else {
        const scale = this.initialScale;
        if (isFinite(scale) && scale > 0) {
          this.scale = scale;
        }
      }
      this.onContentAlignmentChanged(true);
    }

    this.validateParts();

    // inital layouts can only complete if updates are unpaused
    if (layoutCompleted && wasInitLayoutCompleted) {
      // We cannot call requestUpdate update here,
      // it will ruin apps that use delayInitialization
      // such as the virtualized samples
      this.commitTransaction('Initial Layout');
      if (!this.skipsUndoManager) this.undoManager.clear();
      const diagram = this;
      Util.setTimeout(function() {
        diagram.isModified = false;  // do this outside of any transaction
      }, 1);
    }

    // all the links that have had simplified geometries must invalidate their neighboring link geometries
    // real geometries will be constructed during Link's draw
    this.clearDelayedGeometries();

    if (!avoidDraw) this.draw(this._ctx);

    if (wasanimating) {
      this.skipsUndoManager = oldskip;
    }

    this._isUpdating = false;
  }

  /**
   * @hidden @internal
   */
  /* noexport */get isInitialLayoutCompleted(): boolean { return this._isInitialLayoutCompleted; }

  /**
   * @hidden @internal
   * This overridable method is called after the first layout has been performed
   * and the {@link #documentBounds} has been computed.
   * This uses the {@link #autoScale}, {@link #initialDocumentSpot}, and {@link #initialViewportSpot}
   * properties to set the {@link #scale} and {@link #position}
   * properties for the initial appearance of the Diagram.
   * It then raises the "InitialLayoutCompleted" {@link DiagramEvent}.
   * It also clears out the {@link UndoManager}'s recorded changes, if any.
   * @this {Diagram}
   */
  private onInitialLayoutCompleted(): void {
    // After the first layout this needs to happen so that getInView are correct on all nodes
    // It's not guaranteed to happen in Part.notifyDiagram because vpb might have been NaN at the time (no position set yet)
    const layers = this._layers._arr;
    this.setInView(layers, layers.length, this);

    if (this._initialAutoScale !== Diagram.None) {
      this.scale = this._computeScaleForStretch(this._initialAutoScale);
    } else if (this._autoScale !== Diagram.None) {
      this.scale = this._computeScaleForStretch(this._autoScale);
    } else {
      const scale = this.initialScale;
      if (isFinite(scale) && scale > 0) {
        this.scale = scale;
      }
    }

    this.onContentAlignmentChanged(true);
    // Keep the position on initial layout completed in order to compare it at the start of initial animation
    // If they differ, the
    this._savedPosition.assign(this._position);
    this.raiseDiagramEvent('InitialLayoutCompleted');
    this._updateBackgroundGrid();
  }

  /**
   * @hidden @internal
   * This overridable method is called after each layout has been performed
   * and the {@link #documentBounds} has been computed.
   * This raises the "LayoutCompleted" {@link DiagramEvent}.
   * @this {Diagram}
   */
  private onLayoutCompleted(): void {
    this.raiseDiagramEvent('LayoutCompleted');
  }

  /**
   * @hidden @internal
   * @param {Part} part
   * @this {Diagram}
   */
  public invalidatePart(part: Part): void {
    this._invalidParts.add(part);
  }

  /**
   * @hidden @internal
   * Validate all of the invalid Parts.
   * For efficiency, some parts can only be measured when other parts are valid.
   * For example, a Link should not be measured until the two connected Nodes have been arranged.
   * Measuring and arranging some parts will cause other parts to become invalid.
   * For example, measuring and arranging a Link will result in routing that may reposition label nodes.
   * @this {Diagram}
   */
  public validateParts(): void {
    if (!this._isUpdating && this.animationManager.isTicking) return; // was isAnimating, breaks
    if (this._invalidParts.count === 0) return;
    // only update the objects that are in the measure and arrange sets.
    for (let i = 0; i < 23; i++) {
      const validatings = this._invalidParts.iterator;
      if (validatings === null || this._invalidParts.count === 0) break;
      // accumulate a new set of invalid parts, if there are side-effects of measuring/arranging any part
      // or if we have to wait with measuring/arranging because of unmeasured/unarranged dependencies
      this._invalidParts = new Set<Part>();
      this.validate(validatings, this._invalidParts); // also calls measure/arrange on each
      if (Debug && i === 22) Util.trace('failure to validate parts'); // needs a more descriptive message, only useful to us as is
    }
    this.nodes.each(function(n) {
      if (!(n instanceof Group)) return;
      n.setHasInvalidatedConnectedRoutes(false);
    });
  }

  /**
   * @hidden @internal
   * One pass of the validation loop.
   * For Parts that probably should not be measured/arranged yet due to dependencies on other Parts,
   * instead of calling Part.measure and arrange, just add the Part to the DELAY hashset,
   * which will be passed in as the INVALIDS hashset in the next iteration.
   * @param {Iterator.<Part>} invalids
   * @param {Set.<Part>} delay
   * @this {Diagram}
   */
  private validate(invalids: Iterator<Part>, delay: Set<Part>): void { // make sure everything is measured and arranged
    // first do all atomic (non-Group) Nodes and all instances of simple Parts
    invalids.reset();
    while (invalids.next()) {
      const n = invalids.value;
      // measure only Node and Part, not Adornment nor Link nor Group
      if (!n.isNodeLike() || n instanceof Group) continue;
      // cannot skip invisible nodes, need to validate so that link routes are preserved.
      // Otherwise when the nodes become visible, links might lose their custom routes
      if (n.isReadyToMeasureArrange()) {
        n.measure(Infinity, Infinity);
        n.arrange();
      } else {
        delay.add(n); // try again later
      }
    }

    // do a depth first measure/arrange of all the Nodes/Links that are in Groups
    invalids.reset();
    while (invalids.next()) {
      const part = invalids.value;
      if (part instanceof Group) {
        if (!part.isVisible()) continue;
        this.measureArrangeGroup(part);
      }
    }

    // then do all Links, which might invalidate label nodes
    invalids.reset();
    while (invalids.next()) {
      const part = invalids.value;
      if (!(part instanceof Link)) continue;
      const l = part;
      if (!l.isVisible()) continue;
      if (l.isReadyToMeasureArrange()) {
        l.measure(Infinity, Infinity);
        l.arrange();
      } else {
        delay.add(l);
      }
    }

    // then do all Adornments
    invalids.reset();
    while (invalids.next()) {
      const a = invalids.value;
      if (!(a instanceof Adornment)) continue;
      if (!a.isVisible()) continue;
      if (a.isReadyToMeasureArrange()) {
        a.measure(Infinity, Infinity);
        a.arrange();
      } else {
        delay.add(a);
      }
    }
  }

  /**
   * @hidden @internal
   * Recursively walk through group and measure/arrange all its member parts
   * @this {Diagram}
   * @param {Group} group
   */
  private measureArrangeGroup(group: Group): void {
    const links = Util.tempArray<Link>();
    const selflinks = Util.tempArray<Link>();
    const itr = group.memberParts;
    while (itr.next()) {
      const m = itr.value;
      if (!m.isVisible()) continue;
      if (m instanceof Group) {
        if (m.getInvalidMeasure() || m.getInvalidArrange() || m.getInvalidPosition()) {
          this.measureArrangeGroup(m);
        }
      } else if (m instanceof Link) {
        // only measure/arrange Links that connect with the group itself
        // *after* this group has been measured and arranged;
        // note that Placeholder.computeMemberBounds always ignores such Links
        if (m.fromNode === group || m.toNode === group) {
          selflinks.push(m);
        } else {
          links.push(m);
        }
      } else { // It's a Node or Part
        m.measure(Infinity, Infinity);
        m.arrange();
      }
    }

    let l = links.length;
    for (let i = 0; i < l; i++) {
      const link = links[i];
      link.measure(Infinity, Infinity);
      link.arrange();
    }
    Util.freeArray(links);

    group.measure(Infinity, Infinity);
    group.arrange();

    l = selflinks.length;
    for (let i = 0; i < l; i++) {
      const link = selflinks[i];
      link.measure(Infinity, Infinity);
      link.arrange();
    }
    Util.freeArray(selflinks);
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Array.<Layer>} layers
   * @param {number} l int
   * @param {Diagram} diagram
   * @param {Rect=} viewport
   */
  private setInView(layers: Array<Layer>, l: number, diagram: Diagram, viewport?: Rect): void {
    if (!this._invalidViewportObjects && !this.animationManager.isAnimating) return;
    for (let i = 0; i < l; i++) {
      layers[i].setInView(diagram, viewport);
    }
  }

  /**
   * @hidden @internal
   * Draws the contents of all layers onto this diagram's HTML canvas element.
   * @this {Diagram}
   * @param {IContext} ctx Special context to draw upon.
   * @param {Set.<Part>=} exclusion a set of objects to exclude from drawing.
   */
  protected draw(ctx: IContext, exclusion?: Set<Part>): void {
    if (exclusion === undefined) exclusion = null;
    if (this._div === null) return; // no div, no drawing
    const can = this._canvas;
    if (can === null) { Util.throwError('No canvas specified'); }

    const AM = this.animationManager;
    if (AM._pendingAnimation) {
      // A pending animation may have started without the source explicitly giving it a way to begin
      // (As is the case when changes are made outside of a transaction)
      return;
    }

    // Avoid flicker. During animations, only draw on ticks
    if (AM.isAnimating && !AM.isTicking) return;

    const time1 = new Date();
    this.validateCanvasSize();

    if (this._div.style.opacity === '0') return;  // ??? don't need to draw anything

    const foreign = (ctx !== this._ctx);

    const layers = this._layers._arr;
    let l = layers.length;
    const diagram = this;
    this.setInView(layers, l, diagram);

    if (foreign) {
      ctx.clearContextCache(true);
      this.invalidateViewportObjects();
    } else {
      if (!this._invalidDraw && exclusion === null && !AM.isAnimating) return;
    }

    const pos = this._position;
    const scale = this._scale;
    const roundedx = Math.round(pos.x * scale) / scale;
    const roundedy = Math.round(pos.y * scale) / scale;

    const t = this._transform;
    t.reset();
    if (scale !== 1) t.scale(scale);
    if (pos.x !== 0 || pos.y !== 0) t.translate(-roundedx, -roundedy);

    const pixelratio = this._pixelRatio;
    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.scale(pixelratio, pixelratio);
    ctx.clearRect(0, 0, this._canvasWidth, this._canvasHeight);


    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.scale(pixelratio, pixelratio);
    ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

    if (Debug && Debug.boundsInfoEnabled && Debug.drawBoundsInfo) Debug.drawBoundsInfo(this, ctx);

    this.commitRenderingHints(ctx);
    ctx.globalAlpha = this._opacity;

    let func: (lay: Layer) => void;
    if (exclusion !== null) {
      func = function(lay: Layer) { lay.drawExclusion(ctx, diagram, exclusion); };
    } else { // redraw everything
      func = function(lay: Layer) { lay.draw(ctx, diagram); };
    }

    l = layers.length;
    for (let i = 0; i < l; i++) {
      // This fixes occasional compound error drawing issues resulting from transformations
      // and their inverse being applied in some text-drawing situations
      ctx.setTransform(1, 0, 0, 1, 0, 0);
      ctx.scale(pixelratio, pixelratio);
      ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);
      func(layers[i]);
    }

    // wmq
    if (this._helper) {
      this._helper.wmq(this) && this._drawImage();
    } else {
      // this.getMouse = function () { return new Point(0, 0); };
    }

    if (Debug && (Debug.drawLocationsEnabled || Debug.boundsInfoEnabled) && Debug.drawBoundsInfoLocations) { Debug.drawBoundsInfoLocations(ctx, this, t); }
    // Debug:
    // if (Debug) Debug.drawDiagramBounds(ctx, t, pixelratio, this._documentBounds);


    if (foreign) {
      this._ctx.clearContextCache(true);
      this.invalidateViewportObjects();
    } else { // validate
      this._invalidViewportObjects = false;
      this._invalidDraw = false;
    }

    // You could log: (time, this._tempPixelRatio === null, this._movingAvg);
    const time2 = new Date();
    const time = (+time2 - (+time1));
    if (this._tempPixelRatio === null) {
      const _movingAvgArray = this._movingAvgArray;
      _movingAvgArray[this._movingAvgFrame] = time;
      this._movingAvgFrame = (this._movingAvgFrame + 1) % _movingAvgArray.length;
      let sum = 0;
      for (let i = 0; i < this._movingAvgArray.length; i++) {
        sum += this._movingAvgArray[i];
      }
      this._movingAvg = sum / _movingAvgArray.length;
    }

  }

  /**
   * @hidden @internal
   * Uses a saved image to to redraw the viewport at a different location.
   * After drawing the image, this draws the contents that intersects with a given set of rects.
   * This is used for scrolling optimizations where two or three regions need to be redrawn.
   * @this {Diagram}
   * @param {Array.<Rect>} rects clipping region rects in document coordinates.
   * @param {Array.<Rect>} viewRects precise (integer coordinate) clipping region rects in viewport coordinates.
   * @param {number} x coordinates for the saved image.
   * @param {number} y
   */
  private redrawViewport(rects: Array<Rect>, viewRects: Array<Rect>, x: number, y: number): void {
    if (this._div === null) { Util.throwError('No div specified'); }
    const can = this._canvas;
    if (can === null) { Util.throwError('No canvas specified'); }

    if (this.animationManager._pendingAnimation) {
      // A pending animation may have started without the source explicitly giving it a way to begin
      // (As is the case when changes are made outside of a transaction)
      // ? I don't think this line is needed anymore:
      //     if (this.undoManager.transactionLevel === 0) this.requestUpdate(true);
      return;
    }

    const ctx = this._ctx;
    if (!this._invalidDraw) return;
    this.validateCanvasSize();
    const pixelratio = this._pixelRatio;
    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.clearRect(0, 0, this._canvasWidth * pixelratio, this._canvasHeight * pixelratio);

    ctx.drawImage(this._drawCollectionCanvas.domElement,
      (x > 0) ? 0 : Math.round(-x),
      (y > 0) ? 0 : Math.round(-y));

    const pos = this._position;
    const scale = this._scale;
    const roundedx = Math.round(pos.x * scale) / scale;
    const roundedy = Math.round(pos.y * scale) / scale;
    const t = this._transform;
    t.reset();
    if (scale !== 1) t.scale(scale);
    if (pos.x !== 0 || pos.y !== 0) t.translate(-roundedx, -roundedy);

    // Before transform, make clipping region with viewRects:
    ctx.save();
    ctx.beginPath();
    let l = viewRects.length;
    for (let i = 0; i < l; i++) {
      const rect = viewRects[i];
      if (rect.width === 0 || rect.height === 0) continue;
      // modifying the clipping region like this means we need to intersect a larger area
      // (see Layer.intersectsAnyRect) ??
      ctx.rect(Math.floor(rect.x), Math.floor(rect.y), Math.ceil(rect.width), Math.ceil(rect.height));
    }
    ctx.clip();
    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.scale(pixelratio, pixelratio);
    ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

    if (Debug && Debug.boundsInfoEnabled && Debug.drawBoundsInfo) Debug.drawBoundsInfo(this, ctx);

    const layers = this._layers._arr;
    l = layers.length;
    this.setInView(layers, l, this);

    this.commitRenderingHints(ctx);
    ctx.globalAlpha = this._opacity;

    for (let i = 0; i < l; i++) {
      layers[i].drawIntersection(ctx, this, rects);
    }

    ctx.restore();
    ctx.clearContextCache(true);

    if (Debug && (Debug.drawLocationsEnabled || Debug.boundsInfoEnabled) && Debug.drawBoundsInfoLocations) {
      Debug.drawBoundsInfoLocations(ctx, this, t);
    }

    // wmq
    if (this._helper) {
      this._helper.wmq(this) && this._drawImage();
    } else {
      // this.getMouse = function () { return new Point(0, 0); };
    }

    this._invalidViewportObjects = false;
    this._invalidDraw = false;
    this.redrawLater();
  }

  /**
   * @hidden @internal
   * Draws the contents of all layers onto a foreign context
   * @this {Diagram}
   * @param {IContext} ctx Special context to draw upon.
   * @param {Margin} padding Padding transform MUST be taken care of here, but padding size must be taken care of before.
   * @param {Size} size specific drawing size (typically the size of ctx).
   * @param {number} scale specific drawing scale.
   * @param {Point} position specific drawing position.
   * @param {Iterable.<Part>|null=} parts a set of parts to draw instead of everything.
   * @param {string|null=} background a background color.
   * @param {boolean=} showTemporary
   * @param {boolean=} showGrid
   */
  private drawForeign(ctx: IContext, padding: Margin, size: Size, scale: number, position: Point,
                      parts?: Iterable<Part> | null, background?: string | null, showTemporary?: boolean, showGrid?: boolean): void {
    if (this._div === null) return;
    if (this._canvas === null) { Util.throwError('No canvas specified'); }
    if (parts === undefined) parts = null;
    if (background === undefined) background = null;
    if (showTemporary === undefined) showTemporary = false;
    if (showGrid === undefined) showGrid = false;
    this.validateCanvasSize();
    this._ctx.clearContextCache(true);
    this.invalidateViewportObjects();
    this._isDrawingForeign = true;
    const savedScale = this._scale;
    this._scale = scale; // necessary at least so shadows scale correctly.
    const diagram = this;
    const layers = this._layers._arr;
    const l = layers.length;
    try {
      const viewRect = new Rect(position.x, position.y, size.width / scale, size.height / scale);
      const viewRectPlusPadding = viewRect.copy();
      viewRectPlusPadding.addMargin(padding);
      this._updateBackgroundGrid(viewRectPlusPadding);
      this.validateParts();

      this.setInView(layers, l, diagram, viewRect);

      const pixelratio = this._pixelRatio;
      ctx.setTransform(1, 0, 0, 1, 0, 0);
      ctx.scale(pixelratio, pixelratio);
      ctx.clearRect(0, 0, size.width, size.height);
      if (background !== null && background !== '') {
        ctx.fillStyle = background;
        ctx.fillRect(0, 0, size.width, size.height);
      }

      const t = Transform.alloc();
      t.reset();
      t.translate(padding.left, padding.top);
      t.scale(scale);
      if (position.x !== 0 || position.y !== 0) {
        t.translate(-position.x, -position.y);
      }
      ctx.setTransform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

      Transform.free(t);

      this.commitRenderingHints(ctx);
      ctx.globalAlpha = this._opacity;
      let func: (lay: Layer) => void;
      if (parts !== null) {
        const myset = new Set<Part>();
        const itr = parts.iterator;
        itr.reset();
        while (itr.next()) {
          const part = itr.value;
          if (showGrid === false && part.layer.name === 'Grid') continue;
          if (part !== null) myset.add(part);
        }
        func = function(lay: Layer) { lay.drawOnly(ctx, diagram, myset, showTemporary); };
      } else if (!showTemporary && showGrid) {
        const gp = diagram.grid.part;
        const gridlayer = gp.layer;

        func = function(lay: Layer) {
          if (lay === gridlayer) {
            gp.draw(ctx, diagram);
          } else {
            lay.draw(ctx, diagram, showTemporary);
          }
        };

      } else {
        func = function(lay: Layer) {
          lay.draw(ctx, diagram, showTemporary);
        };
      }

      for (let i = 0; i < l; i++) {
        func(layers[i]);
      }

      this._isDrawingForeign = false;
      // wmq
      if (this._helper) {
        this._helper.wmq(this) && this._drawImage();
      } else {
        // this.getMouse = function () { return new Point(0, 0); };
      }
    } finally {
      this._scale = savedScale;
      this._ctx.clearContextCache(true);
      this.invalidateViewportObjects();
      this.setInView(layers, l, diagram);
      this._updateBackgroundGrid(); // return to normal
    }
  }

  // Get value of a given rendering hint
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   * @param {string} name
   * @return {*}
   */
  public getRenderingHint(name: string): any {
    return this._renderingHints[name]; // may be undefined
  }

  // Set a rendering hint and draw immediately
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   * @param {string} name
   * @param {*} val
   */
  public setRenderingHint(name: string, val: any): void {
    if (name === 'minDrawingLength') this._minDrawingLength = val;
    this._renderingHints[name] = val;
    this.redraw();
  }

  /**
   * @hidden @internal
   * Set back to defaults
   * @export
   * @this {Diagram}
   */
  public resetRenderingHints(): void {
    this._renderingHints = new PropertyCollection();
    this._renderingHints['drawShadows'] = true;
    this._renderingHints['textGreeking'] = true;
    // turn off for iOS devices. Turn off for IE10/11 because the clipping region + shadows breaks IE.
    this._renderingHints['viewportOptimizations'] = (Util.isIOS || Util.isIE10 || Util.isIE11) ? false : true;
    // This also has diagram.partManager.nodes.count > 50 as a condition in dragging tool
    this._renderingHints['temporaryPixelRatio'] = true;
    this._renderingHints['pictureRatioOptimization'] = true;

    this._renderingHints['minDrawingLength'] = 1;
    this._minDrawingLength = 1;
  }

  /**
   * @hidden @internal
   * Used at draw time
   * @this {Diagram}
   * @param {IContext} ctx
   */
  private commitRenderingHints(ctx: IContext): void {
    const r = this._renderingHints;
    if (r === null) return;
    if (r['imageSmoothingEnabled'] !== undefined) {
      const smooth = !!(r['imageSmoothingEnabled']);
      ctx.setImageSmoothingEnabled(smooth);
    }

    const font = r['defaultFont'];
    if (font !== undefined && font !== null) {
      ctx.font = font;
    }
  }

  // * Get value of a given input option
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   * @param {string} name
   * @return {*}
   */
  public getInputOption(name: string): any {
    return this._inputOptions[name]; // may be undefined
  }

  // * Current input options:
  // * extraTouchArea: A number describing the width and height to increase small touch objects, default 10
  // * extraTouchThreshold: A number describing the real-pixel size an object must be to increase its touch area, default 10
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   * @param {string} name
   * @param {*} val
   */
  public setInputOption(name: string, val: any): void {
    this._inputOptions[name] = val;
  }

  /**
   * @hidden @internal
   * Set back to defaults
   * @export
   * @this {Diagram}
   */
  private resetInputOptions(): void {
    this._inputOptions = new PropertyCollection();
    this._inputOptions['extraTouchArea'] = 10;
    this._inputOptions['extraTouchThreshold'] = 10;
    // For disabling gestures to allow them to bubble instead
    this._inputOptions['hasGestureZoom'] = true;
  }

  /**
   * This method sets a collection of properties according to the property/value pairs that have been set on the given Object,
   * in the same manner as {@link GraphObject.make} does when constructing a Diagram with an argument that is a simple JavaScript Object.
   *
   * You can set properties on an object that is the value of a property of the Diagram,
   * or on the {@link Diagram#toolManager}, by using a *subpropname.property* syntax for the property name.
   * At the current time only a single dot is permitted in the property "name".
   *
   * The property name may also be the name of a {@link DiagramEvent}, in which case this calls {@link #addDiagramListener} with that DiagramEvent name.
   * ```js
   * aDiagram.setProperties({
   *   allowDelete: false,
   *   // display a background grid for the whole diagram
   *   "grid.visible": true,
   *   "grid.gridCellSize": new go.Size(20, 20),
   *   "animationManager.isEnabled": false,  // turn off automatic animations
   *   // specify a group membership validation predicate
   *   "commandHandler.memberValidation": function(group, part) { return ...; },
   *   "commandHandler.copiesTree": true,  // for the copy command
   *   // mouse wheel zooms instead of scrolls
   *   "toolManager.mouseWheelBehavior": go.ToolManager.WheelZoom,
   *   "draggingTool.dragsTree": true,  // dragging for both move and copy
   *   "draggingTool.isGridSnapEnabled": true,
   *   layout: $(go.TreeLayout),
   *   // add a DiagramEvent listener
   *   "ExternalObjectsDropped": function(e) { e.subject.each(function(part) { ... }); }
   * });
   * ```
   * @export
   * @this {Diagram}
   * @param {Object} props a plain JavaScript object with various property values to be set on this Diagram or on a part of this Diagram.
   * @since 1.5
   */
  public setProperties(props: ObjectData): void {
    Diagram.setCheckedProperties(this, props);
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {Object} props ideally a PropertyCollection
   */
  public static setCheckedProperties(obj: any, props: ObjectData): void {
    // now actually set all of the properties, with some error checking
    // and special cases for modifying existing objects
    const inpanel = obj instanceof Panel;
    const indiagram = obj instanceof Diagram;
    for (const prop in props) {
      if (prop === '') Util.throwError('Setting properties requires non-empty property names');
      let target = obj;
      let property = prop;
      if (inpanel || indiagram) {
        // support setting { "OBJNAME.prop": somevalue } to set "prop" on element named "OBJNAME"
        const dotindex = prop.indexOf('.');
        if (dotindex > 0) {
          const objname = prop.substring(0, dotindex);
          if (inpanel) {  // if OBJ is a Panel
            target = obj.findObject(objname);
          } else {  // must be a Diagram
            target = obj[objname];
            if (target === undefined || target === null) {
              target = obj.toolManager[objname];
            }
          }
          if (Util.isObject(target)) {
            property = prop.substr(dotindex + 1);
          } else {
            Util.throwError('Unable to find object named: ' + objname + ' in ' + obj.toString() +
              ' when trying to set property: ' + prop);
          }
        }
      }
      if (property[0] !== '_' && !Util.hasDefinedProperty(target, property)) {
        if (indiagram && property === 'ModelChanged') {
          obj.addModelChangedListener(props[property]);
          continue;  // go on to next property
        } else if (indiagram && property === 'Changed') {
          obj.addChangedListener(props[property]);
          continue;  // go on to next property
        } else if (indiagram && Util.hasDefinedProperty(obj.toolManager, property)) {
          target = obj.toolManager;
        } else if (indiagram && obj.canonicalizeDiagramEventName(property)) {
          // OBJ is a Diagram and the property name is actually a DiagramEvent name,
          // so the property value should be a listener for that DiagramEvent
          obj.addDiagramListener(property, props[property]);
          continue;  // go on to next property
        } else if (obj instanceof Model && property === 'Changed') {
          obj.addChangedListener(props[property]);
          continue;  // go on to next property
        } else {
          Util.throwError('Trying to set undefined property \"' + property + '\" on object: ' + target.toString());
        }
      }
      const val = props[prop];
      target[property] = val;
      // record _propnames as properties to be copied by GraphObject.cloneProtected
      if (property[0] === '_' && target instanceof GraphObject) {
        target.addCopyProperty(property);
      }
    }
  }

  /**
   * Used at draw time
   * @hidden @internal
   * @export
   * @this {Diagram}
   */
  private clearDelayedGeometries(): void {
    // ??? This is so full geometries don't get created during dragging tool operation:
    if (this.undoManager.transactionLevel !== 0) return;
    if (this._delayedGeometries.count === 0) return;
    while (this._delayedGeometries.count > 0) {
      const delayeds = this._delayedGeometries;
      // replace this Map with an empty one, in case it is modified during the iteration
      this._delayedGeometries = new Map<Link, Rect>();
      const it = delayeds.iterator;
      while (it.next()) {
        const link = it.key;
        // Right now this is also at the end of arrangeLink (good or bad?)
        link.invalidateOtherJumpOvers(it.value); // no other conditions?
        link.invalidateGeometry(); // also needs to invalidate itself to make a non-simplified geometry
      }
    }
    this.invalidateDraw();
  }

  /**
   * Causes the diagram to redraw eventually. Use Diagram.redraw for immediate redrawing.
   * @hidden @internal
   * @this {Diagram}
   * @param {Rect=} rect1 An optional area to invalidate in document coordinates.
   */
  public invalidateDraw(rect1?: Rect): void {
    if (rect1 === undefined) rect1 = null;
    // only invalidate draw and request an update if the rect is in the viewportBounds
    if (rect1 === null) {
      this._invalidDraw = true;
      this.requestUpdate();
    } else {
      const v = this.viewportBounds;
      if ((rect1 !== null && rect1.isReal() && v.intersectsRect(rect1))) {
        this._invalidDraw = true;
        this.requestUpdate();
      }
    }

    this.raiseDiagramEvent('InvalidateDraw');
  }

  /**
   * @hidden @internal
   * Causes the diagram to redraw.
   * @this {Diagram}
   * @param {Rect=} oldv old viewport bounds.
   * @param {Rect=} newv new viewport bounds.
   */
  protected invalidateViewport(oldv?: Rect, newv?: Rect): void {
    // only invalidate draw and request an update if the rect is in the viewportBounds
    if (this._invalidDraw === true) return;
    this._invalidDraw = true;
    const tempPixelRatio = this.getRenderingHint('temporaryPixelRatio') === true;

    if (this.getRenderingHint('viewportOptimizations') === true &&
      this.scrollMode !== Diagram.InfiniteScroll &&
      this._scrollMargin.equalTo(0, 0, 0, 0) &&
      newv.width === oldv.width &&
      newv.height === oldv.height) {
      // Old viewport optimization: copy view partially onto a second canvas
      const scale = this.scale;
      const maxx = Math.max(oldv.x, newv.x);
      const maxy = Math.max(oldv.y, newv.y);
      const minr = Math.min(oldv.x + oldv.width, newv.x + newv.width);
      const minb = Math.min(oldv.y + oldv.height, newv.y + newv.height);
      // intersection of oldv and newv
      const r = Rect.allocAt(maxx, maxy, Math.max(0, minr - maxx) * scale, Math.max(0, minb - maxy) * scale);

      // if intersection exists
      if (!this._svg && r.width > 0 && r.height > 0) {
        this.updateViewport(/*r, */oldv, newv);
      } else { // if it does not exist, request a normal update
        this.maybeUpdate();
      }
      Rect.free(r);

      // New viewport optimization: set the pixel ratio to 1 temporarily
      if (tempPixelRatio) {
        this.enableTempPixelRatio();
        this.maybeUpdate();
        this.disableTempPixelRatio(true);
      }
    } else if (tempPixelRatio) {
      this.enableTempPixelRatio();
      this.maybeUpdate();
      this.disableTempPixelRatio(true);
    } else {
      this.maybeUpdate();
    }
  }

  /**
   * @hidden @internal
   * Causes the diagram to update the scrollbars.
   * @this {Diagram}
   */
  protected invalidateScroll(): void {
    if (this._invalidScrollBars === false) {
      this._invalidScrollBars = true;
    }
  }

  /**
   * @hidden @internal
   * Causes the diagram to redraw.
   * @this {Diagram}
   */
  public invalidateViewportObjects(): void {
    if (this._invalidViewportObjects === false) {
      this._invalidViewportObjects = true;
    }
  }

  /**
   * @hidden @internal
   * Flag is called instead of updateCanvasSize, this is called in draw methods
   * @this {Diagram}
   */
  protected validateCanvasSize(): void {
    if (this._invalidCanvasSize === false) return;
    this._invalidCanvasSize = false;
    this.updateCanvasSize(this._canvasWidth, this._canvasHeight);
  }

  /**
   * @hidden @internal
   * The only method that should be resizing the canvas
   * @this {Diagram}
   * @param {number} width
   * @param {number} height
   */
  public updateCanvasSize(width: number, height: number): void {
    const pixelratio = this._pixelRatio;
    const pixelWidth = width * pixelratio;
    const pixelHeight = height * pixelratio;
    if (this._canvas.resize(pixelWidth, pixelHeight, width, height)) {
      this._invalidDraw = true;
      this._ctx.clearContextCache(true);
    }
  }

  /**
   * @hidden @internal
   * Method for resizing the canvas as well as the scrollbar helper divs.
   * @this {Diagram}
   * @return {boolean} false if the canvas changed size, true if the canvas did not change size
   */
  public updateScroll(): boolean {
    const can = this._canvas;
    if (can === null) return true;
    const div = this._div;
    const oldw = this._canvasWidth;
    const oldh = this._canvasHeight;

    const oldvpb = this._viewportBounds.copy();
    if (isNaN(oldvpb.width) || isNaN(oldvpb.height)) return true; // Don't check pos

    let scrollbarsChanged = false;
    // check and see if the canvas is the same size as its immediate parent in the DOM (this._div)
    // if not, resize the canvas
    let sw = this._hasVScroll ? this._scrollBarBreadth : 0;
    let sh = this._hasHScroll ? this._scrollBarBreadth : 0;

    let cw = div.clientWidth || oldw + sw;
    let ch = div.clientHeight || oldh + sh;

    if (cw !== oldw + sw || ch !== oldh + sh) {
      // if the canvas changes size, we'll need a full repaint
      this._hasVScroll = false;
      this._hasHScroll = false;
      sw = 0;
      sh = 0;
      this._canvasWidth = cw;
      this._canvasHeight = ch;
      this._invalidCanvasSize = true;
      scrollbarsChanged = true;
    }
    // no scrollbars, no need:
    if (!scrollbarsChanged && !this._hasVScroll && !this._hasHScroll && !this._hasHorizontalScrollbar && !this._hasVerticalScrollbar) return true;

    this._invalidScrollBars = false;

    // The following only happens if we need to create or reposition the scrollbars.
    // Because of resize, scale change, diagram position change, docbounds change, etc.
    // This code does not fire if a user is moving the scrollbar, that's diagramScroll.
    const viewportBounds = this.viewportBounds; // since it is generated each time
    const docb = this.documentBounds;
    let scrollLeft = 0;
    let scrollRight = 0;
    let scrollTop = 0;
    let scrollBottom = 0;
    const vw = viewportBounds.width;
    const vh = viewportBounds.height;
    const scrollMargin = this._scrollMargin;
    if (!this.contentAlignment.isSpot()) {
      scrollLeft = scrollMargin.left;
      scrollRight = scrollMargin.right;
      scrollTop = scrollMargin.top;
      scrollBottom = scrollMargin.bottom;
    } else {
      if (docb.width > vw) {
        scrollLeft = scrollMargin.left;
        scrollRight = scrollMargin.right;
      }
      if (docb.height > vh) {
        scrollTop = scrollMargin.top;
        scrollBottom = scrollMargin.bottom;
      }
    }

    const dw = docb.width + scrollLeft + scrollRight;
    const dh = docb.height + scrollTop + scrollBottom;

    // for when document is left of viewport
    const dx = docb.x - scrollLeft;
    const vx = viewportBounds.x;
    // for when document is right of viewport
    const dright = docb.right + scrollRight;
    let vright = viewportBounds.right + sw;
    // for when document is above
    const dy = docb.y - scrollTop;
    const vy = viewportBounds.y;
    // for when document is below
    const dbottom = docb.bottom + scrollBottom;
    let vbottom = viewportBounds.bottom + sh;

    let helperwidth = '1px';
    let helperheight = '1px';

    const scale = this.scale;
    const oldscale = scale;

    // If the client width/height is larger than the document bounds,
    // we don't need to do any of this because there cannot be scrollbars
    // (except if infinite scrolling)
    let horizScrollPossible = !(dw < vw + sw);
    let vertScrollPossible = !(dh < vh + sh);
    if (this.scrollMode === Diagram.DocumentScroll && (horizScrollPossible || vertScrollPossible)) {
      if (horizScrollPossible && this.hasHorizontalScrollbar && this.allowHorizontalScroll) { // horizontal scrollbar
        let hw = 1;
        if (dx + 1 < vx) { // diagram is out of the views bounds on the left
          hw = Math.max((vx - dx) * scale + this._canvasWidth, hw);
        }
        if (dright > vright + 1) { // diagram is out of the views bounds on the right
          hw = Math.max((dright - vright) * scale + this._canvasWidth, hw);
        }
        if (vw + sw + 1 < dw) { // smaller view than the diagram
          hw = Math.max((dw - vw) * scale + this._canvasWidth, hw);
        }
        helperwidth = hw.toString() + 'px';
      }

      if (vertScrollPossible && this.hasVerticalScrollbar && this.allowVerticalScroll) { // vertical scrollbar
        let hh = 1;
        if (dy + 1 < vy) { // diagram is out of the views bounds on the top
          hh = Math.max((vy - dy) * scale + this._canvasHeight, hh);
        }
        if (dbottom > vbottom + 1) { // diagram is out of the views bounds on the bottom
          hh = Math.max((dbottom - vbottom) * scale + this._canvasHeight, hh);
        }
        if (vh + sh + 1 < dh) { // smaller view than the diagram
          hh = Math.max(((dh - vh) * scale) + this._canvasHeight, hh);
        }
        helperheight = hh.toString() + 'px';
      }
    }
    let newHScroll = (helperwidth !== '1px');
    let newVScroll = (helperheight !== '1px');

    // The existence of one scrollbar can make the viewport smaller, so it may require the existence of another
    if (!(newHScroll && newVScroll) && (newHScroll || newVScroll)) {
      horizScrollPossible = !(dw < vw + sw);
      vertScrollPossible = !(dh < vh + sh);
      if (newVScroll) vright -= this._scrollBarBreadth;
      if (newHScroll) vbottom -= this._scrollBarBreadth;
      if (horizScrollPossible && this.hasHorizontalScrollbar && this.allowHorizontalScroll) { // horizontal scrollbar
        let hw = 1;
        if (dx + 1 < vx) { // diagram is out of the views bounds on the left
          hw = Math.max((vx - dx) * scale + this._canvasWidth, hw);
        }
        if (dright > vright + 1) { // diagram is out of the views bounds on the right
          hw = Math.max((dright - vright) * scale + this._canvasWidth, hw);
        }
        if (vw + 1 < dw) { // smaller view than the diagram
          hw = Math.max((dw - vw) * scale + this._canvasWidth, hw);
        }
        helperwidth = hw.toString() + 'px';
      }
      newHScroll = (helperwidth !== '1px');

      // if different than before
      let tempCH = this._canvasHeight;
      if (newHScroll !== this._hasHScroll) {
        tempCH = newHScroll ? this._canvasHeight - this._scrollBarBreadth : this._canvasHeight + this._scrollBarBreadth;
      }

      if (vertScrollPossible && this.hasVerticalScrollbar && this.allowVerticalScroll) { // vertical scrollbar
        let hh = 1;
        if (dy + 1 < vy) { // diagram is out of the views bounds on the top
          hh = Math.max((vy - dy) * scale + tempCH, hh);
        }
        if (dbottom > vbottom + 1) { // diagram is out of the views bounds on the bottom
          hh = Math.max((dbottom - vbottom) * scale + tempCH, hh);
        }
        if (vh + 1 < dh) { // smaller view than the diagram
          hh = Math.max(((dh - vh) * scale) + tempCH, hh);
        }
        helperheight = hh.toString() + 'px';
      }
      newVScroll = (helperheight !== '1px');
    }

    // If physically touching scrollbar, don't continue unless a scrollbar must appear or disappear
    if (this._scrollBarMoving && !(newHScroll !== this._hasHScroll || newVScroll !== this._hasVScroll)) {
      if (oldw !== this._canvasWidth || oldh !== this._canvasHeight) this.maybeUpdate();
      return false;
    }

    // if different than before
    if (newHScroll !== this._hasHScroll) {
      if (helperwidth === '1px') { // there is no more scrollbar (but once was)
        this._canvasHeight = this._canvasHeight + this._scrollBarBreadth;
      } else { // there is now a scrollbar
        this._canvasHeight = Math.max(this._canvasHeight - this._scrollBarBreadth, 1);
      }
      scrollbarsChanged = true; // this._canvasHeight changed
    }
    // Must happen every time (scrollbar might already exist and get bigger/smaller)
    this._hasHScroll = newHScroll;
    this._scrollHelper.style.width = helperwidth;

    // if different than before
    if (newVScroll !== this._hasVScroll) {
      if (helperheight === '1px') { // there is no more scrollbar (but once was)
        this._canvasWidth = this._canvasWidth + this._scrollBarBreadth;
      } else { // there is now a scrollbar
        this._canvasWidth = Math.max(this._canvasWidth - this._scrollBarBreadth, 1);
      }
      scrollbarsChanged = true; // this._canvasWidth changed

      // Vertical bar only: Accommodate RTL webpages
      if (this._isRTL) {
        const p = Point.alloc();
        if (newVScroll) {
          can.style.left = this._scrollBarBreadth + 'px';
          this.position = p.assignTo(this._position.x + (this._scrollBarBreadth / this.scale), this._position.y);
        } else {
          can.style.left = '0px';
          this.position = p.assignTo(this._position.x - (this._scrollBarBreadth / this.scale), this._position.y);
        }
        Point.free(p);
      }
    }
    // Must happen every time (scrollbar might already exist and get bigger/smaller)
    this._hasVScroll = newVScroll;
    this._scrollHelper.style.height = helperheight;

    this._scrollBarsModified = true;
    if (scrollbarsChanged) {
      this._invalidCanvasSize = true;
    }

    // Set scrollbar positions
    const scrollParent = this._scrollParent;
    let value = scrollParent.scrollLeft;
    if (this.hasHorizontalScrollbar && this.allowHorizontalScroll) { // horizontal scrollbar
      if (vw + 1 < dw) { // smaller view than the diagram
        value = (this.position.x - dx) * scale;
      } else if (dx + 1 < vx) { // diagram is out of the views bounds on the left
        value = scrollParent.scrollWidth - scrollParent.clientWidth;
      } else if (dright > vright + 1) { // diagram is out of the views bounds on the right
        value = this.position.x * scale;
      }
    }

    // make sure value is adjusted if RTL
    if (this._isRTL) {
      switch (this._rtlScrollType) {
        case 'negative':
          value = -(scrollParent.scrollWidth - value - scrollParent.clientWidth);
          break;
        case 'reverse':
          value = scrollParent.scrollWidth - value - scrollParent.clientWidth;
          break;
      }
    }
    scrollParent.scrollLeft = value;

    if (this.hasVerticalScrollbar && this.allowVerticalScroll) { // vertical scrollbar
      if (vh + 1 < dh) { // smaller view than the diagram
        scrollParent.scrollTop = (this.position.y - dy) * scale;
      } else if (dy + 1 < vy) { // diagram is out of the views bounds on the top
        scrollParent.scrollTop = scrollParent.scrollHeight - scrollParent.clientHeight;
      } else if (dbottom > vbottom + 1) { // diagram is out of the views bounds on the bottom
        scrollParent.scrollTop = this.position.y * scale;
      }
    }

    // scroll parents must be sized with the canvas to keep the right form
    cw = this._canvasWidth;
    ch = this._canvasHeight;

    scrollParent.style.width = (cw + (this._hasVScroll ? this._scrollBarBreadth : 0)) + 'px';
    scrollParent.style.height = (ch + (this._hasHScroll ? this._scrollBarBreadth : 0)) + 'px';

    // Only if a canvas has changed size
    // Or if animation is pending, since we have all the info to ensure proper scrollbars pre-animation
    // and don't want to wait for animation to complete
    if (oldw !== cw || oldh !== ch || this.animationManager._pendingAnimation) {
      this.onViewportBoundsChanged(oldvpb, this.viewportBounds, oldscale, scrollbarsChanged);
      return false; // returns false if the canvas size has changed
    }
    return true; // returns true if the canvas size has not changed
  }

  /**
   * Adds a {@link Part} to the {@link Layer} that matches the Part's {@link Part#layerName},
   * or else the default layer, which is named with the empty string.
   *
   * Normally parts added to a diagram are top-level parts.
   * If you want nodes to be members of a {@link Group}, in addition to calling this method
   * call {@link Group#addMembers} or set each {@link Part#containingGroup}.
   * @export
   * @this {Diagram}
   * @param {Part} part
   * @see {@link #remove}
   */
  public add(part: Part): void {
    Util.checkClass(part, Part, Diagram, 'add:part');
    const d = part.diagram;
    if (d === this) return;
    if (d !== null) {
      Util.throwError('Cannot add part ' + part.toString() + ' to ' + this.toString() +
        '. It is already a part of ' + d.toString());
    }
    let layer = this.findLayer(part.layerName);
    if (layer === null) layer = this.findLayer('');
    if (layer === null) Util.throwError('Cannot add a Part when unable find a Layer named \"' + part.layerName + '\" and there is no default Layer');
    if (part.layer === layer) return; // should be a no-op

    // add to layer; Layer.internalAdd will call Diagram.internalAdd for local bookkeeping
    const index = layer.internalAdd(99999999, part, part.diagram === this);  // just swapping layers?
    if (index >= 0) this.raiseChangedEvent(ChangedEvent.Insert, 'parts', layer, null, part, null, index);

    if (!layer.isTemporary) this.invalidateDocumentBounds();
    // called after adding the part to a layer
    part.invalidateLayout(Part.LayoutAdded);

    const func = part.layerChanged;
    if (func !== null) func(part, null, layer);
  }

  /**
   * @hidden @internal
   * This is called from {@link Layer#internalAdd} for bookkeeping.
   * @this {Diagram}
   * @param {Part} part
   */
  public internalAdd(part: Part): void {
    // Maintain the collection of Nodes (and Groups) and Links, but not Adornments
    this.partManager.internalAdd(part);
    const dia = this;
    part.walkPanels(function(p) {
      dia.rememberPictures(p);
    });

    // might be null (Adornment with none) or undefined (regular Node)
    if (part instanceof Adornment || part instanceof Group && part.placeholder !== null) {
      part.invalidateMeasure();
    }

    const data = part.data;
    if (data !== null) {
      part.walkPanels(function(p) { dia.partManager.rememberItemArrayPanel(p); });
    }

    if (part.getInvalidArrange() === true || part.getInvalidPosition() === true) {
      this.invalidatePart(part);
    }
    // always notify to ensure that the part is book-kept in the layer too. Important for copying.
    part.notifyDiagram(true, this);

    // Update the document bounds
    if (part.canIncludeBounds()) {
      if (part.actualBounds.isReal()) this.invalidateDraw(part.getDrawnSize(part.actualBounds));
      this.invalidateDocumentBounds();
    } else if (part.isVisible()) {
      if (part.actualBounds.isReal()) this.invalidateDraw(part.getDrawnSize(part.actualBounds));
    }
    this.requestUpdate();
  }

  /**
   * @hidden @internal
   * This is called from {@link Layer#internalRemove} for bookkeeping.
   * @this {Diagram}
   * @param {Part} part
   */
  public internalRemove(part: Part): void {
    part.clearAdornments();

    this.partManager.internalRemove(part);

    const dia = this;

    const data = part.data;
    if (data !== null) {
      part.walkPanels(function(p) { dia.partManager.forgetItemArrayPanel(p, dia); });
    }

    this._invalidParts.remove(part);
    // Update the document bounds
    if (part.canIncludeBounds()) {
      if (part.actualBounds.isReal()) this.invalidateDraw(part.getDrawnSize(part.actualBounds));
      this.invalidateDocumentBounds();
    } else if (part.isVisible()) {
      if (part.actualBounds.isReal()) this.invalidateDraw(part.getDrawnSize(part.actualBounds));
    }
    this.requestUpdate();
  }

  /**
   * Removes a {@link Part} from its {@link Layer}, provided the Layer is in this Diagram.
   * Removing a {@link Node} will also remove any Links that are connected with it.
   * Removing a {@link Group} will also remove all of its members.
   * Removing a {@link Link} will also remove all of its label Nodes, if it has any.
   * @export
   * @this {Diagram}
   * @param {Part} part
   * @see {@link #add}
   */
  public remove(part: Part): void {
    Util.checkClass(part, Part, Diagram, 'remove:part');
    this.doRemove(part, true);
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Part} part
   * @param {boolean} cascade
   */
  public doRemove(part: Part, cascade: boolean): void {
    const layer = part.layer;
    if (layer === null) return;
    if (layer.diagram !== this) return;

    // make sure it's no longer selected
    part.isSelected = false;
    part.isHighlighted = false;

    // call this before removing from any layer
    part.invalidateLayout(Part.LayoutRemoved);

    // delete links connected to nodes, members of groups, labelnodes of links
    if (cascade) part.deleteCascade();

    // remove from layer; Layer.internalRemove will call Diagram.internalRemove for local bookkeeping
    const index = layer.internalRemove(-1, part, false);
    if (index >= 0) this.raiseChangedEvent(ChangedEvent.Remove, 'parts', layer, part, null, index, null);

    const func = part.layerChanged;
    if (func !== null) func(part, layer, null);
  }

  /**
   * This method removes from this Diagram all of the {@link Part}s in a collection.
   * Removing a {@link Node} will also remove any Links that are connected with it.
   * Removing a {@link Group} will also remove all of its members.
   * Removing a {@link Link} will also remove all of its label Nodes, if it has any.
   *
   * This does not perform a transaction nor does it raise a {@link DiagramEvent}.
   * Call {@link CommandHandler#deleteSelection}, which calls this method,
   * if you want to delete all selected Parts.
   * The {@link CommandHandler#deleteSelection} command may delete other Parts as well, depending on {@link CommandHandler#deletesTree}.
   *
   * At this time there is no "addParts" method -- just call {@link Diagram#add} on each Part.
   * @export
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll A {@link List} or {@link Set} or {@link Iterator} or Array of {@link Part}s.
   * @param {boolean} check Whether to check {@link Part#canDelete} on each part.
   * @since 1.3
   */
  public removeParts(coll: Iterable<Part> | Array<Part>, check: boolean): void {
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        const part = Util.arrayElt(coll, i);
        if (check && !part.canDelete()) continue;
        this.remove(part);
      }
    } else {
      // always make a copy of the incoming collection, in case it is modified by any removals
      const collcopy = new Set<Part>();
      collcopy.addAll(coll);
      const it = collcopy.iterator;
      while (it.next()) {
        const part = it.value;
        if (check && !part.canDelete()) continue;
        this.remove(part);
      }
    }
  }

  /**
   * Make a copy of a collection of {@link Part}s and return them in a {@link Map} mapping each original Part to its copy.
   * It may optionally add them to a given {@link Diagram}.
   * Copying a {@link Group} will also copy its member Nodes and Links.
   * Copying a {@link Link} will also copy any label Nodes that it owns.
   *
   * This does not perform a transaction nor does it raise a {@link DiagramEvent}.
   * Call {@link CommandHandler#copySelection}, which calls this method,
   * if you want to copy all selected Parts into the clipboard.
   * The {@link CommandHandler#copySelection} command may also copy additional Parts as well, depending on {@link CommandHandler#copiesTree}.
   * @export
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll A {@link List} or a {@link Set} or {@link Iterator} of {@link Part}s, or an Array of {@link Part}s.
   * @param {Diagram} diagram The destination diagram; if null, the copied parts are not added to this diagram.
   * @param {boolean} check Whether to check {@link Part#canCopy} on each part.
   * @return {Map.<Part,Part>}
   * @since 1.3
   */
  public copyParts(coll: Iterable<Part> | Array<Part>, diagram: Diagram | null, check: boolean): Map<Part, Part> {
    return this.partManager.copyParts(coll, diagram, check);
  }

  /**
   * Move a collection of {@link Part}s in this Diagram by a given offset.
   * Moving a {@link Group} will also move its member Nodes and Links.
   * Moving with a zero X and a zero Y offset is potentially useful in order to snap Parts to the grid if {@link DraggingTool#isGridSnapEnabled} is true.
   *
   * This does not perform a transaction nor does it raise a {@link DiagramEvent}.
   *
   * @export
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll A {@link List} or a {@link Set} or {@link Iterator} of {@link Part}s, or an Array of {@link Part}s,
   * or null to move all of the Parts in this Diagram.
   * @param {Point} offset the amount to move each Part, in document coordinates.
   * @param {boolean} check Whether to check {@link Part#canMove} on each part.
   * @param {DraggingOptions=} dragOptions Optional dragging options. By default this uses the settings from the Diagram's DraggingTool.
   * @since 1.3
   */
  public moveParts(coll: Iterable<Part> | Array<Part>, offset: Point, check: boolean, dragOptions?: DraggingOptions): void {
    if (dragOptions === undefined) dragOptions = this.getDragOptions();
    Util.checkClass(offset, Point, Diagram, 'moveParts:offset');
    const mgr = this.toolManager;
    if (mgr === null) return;
    const map = new Map<Part, DraggingInfo>();
    if (coll !== null) {
      if (Util.isArray(coll)) {
        for (let i = 0; i < coll.length; i++) {
          this.gatherMembers(map, coll[i], check, dragOptions);
        }
      } else {
        const it = coll.iterator;
        while (it.next()) {
          this.gatherMembers(map, it.value, check, dragOptions);
        }
      }
    } else {
      let it = this.parts;
      while (it.next()) {
        const sp = it.value;
        this.gatherMembers(map, sp, check, dragOptions);
      }
      it = this.nodes;
      while (it.next()) {
        const n = it.value;
        this.gatherMembers(map, n, check, dragOptions);
      }
      it = this.links;
      while (it.next()) {
        const l = it.value;
        this.gatherMembers(map, l, check, dragOptions);
      }
    }
    this.moveParts2(map, offset, dragOptions, check);
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Map.<Part,DraggingInfo>} map
   * @param {Part} p
   * @param {boolean} dragging whether or not the dragging tool is running
   * @param {DraggingOptions=} dragOptions
   * @param {boolean=} groupsAlwaysMove
   */
  public gatherMembers(map: Map<Part, DraggingInfo>, p: Part, dragging: boolean, dragOptions?: DraggingOptions, groupsAlwaysMove?: boolean): void {
    if (map.contains(p)) return;
    if (groupsAlwaysMove === undefined) groupsAlwaysMove = false;
    // only check canMove or canCopy when dragging interactively, because this is the Diagram.currentTool
    // also, dragging a Group can move immovable nodes: do not check canMove/canCopy if groupsAlwaysMove is true (default) when gathering from a Group.
    if (dragging && !groupsAlwaysMove && !p.canMove() && !p.canCopy()) return;
    if (dragOptions === undefined) dragOptions = this.getDragOptions();
    if (p instanceof Node) {
      map.add(p, this._newDraggingInfo(dragOptions, p, p.location));
      if (p instanceof Group) {
        const mit = p.memberParts;
        while (mit.next()) {
          const m = mit.value;
          this.gatherMembers(map, m, dragging, dragOptions, dragOptions.groupsAlwaysMove);
        }
      }
      const lit = p.linksConnected;
      while (lit.next()) {
        const l = lit.value;
        if (map.contains(l)) continue;
        const from = l.fromNode;
        const to = l.toNode;
        if (from !== null && map.contains(from) &&
          to !== null && map.contains(to)) {
          this.gatherMembers(map, l, dragging, dragOptions);
        }
      }
      if (dragOptions.dragsTree) {
        const it = p.findTreeChildrenNodes();
        while (it.next()) {
          const c = it.value;
          this.gatherMembers(map, c, dragging, dragOptions);
        }
      }
    } else if (p instanceof Link) {
      map.add(p, this._newDraggingInfo(dragOptions, p));
      const lit2 = p.labelNodes;
      while (lit2.next()) {
        const lab = lit2.value;
        this.gatherMembers(map, lab, dragging, dragOptions);
      }
    } else if (!(p instanceof Adornment)) {
      map.add(p, this._newDraggingInfo(dragOptions, p, p.location));
    }
  }

  /**
   * @hidden @internal
   * Called by Diagram.moveParts, DraggingTool.moveParts, and CommandHandler.pasteSelection
   * This implementation used to be DraggingTool.moveParts, which now just calls this.
   * @this {Diagram}
   * @param {Map.<Part,DraggingInfo>} parts a Map mapping {@link Part}s to DraggingInfo Objects
   *        that have a "point" property remembering the original location of that Part.
   * @param {Point} offset
   * @param {DraggingOptions} dragOptions
   * @param {boolean} check Whether to check {@link Part#canMove} on each part.
   */
  public moveParts2(parts: Map<Part, DraggingInfo>, offset: Point, dragOptions: DraggingOptions, check: boolean): void {
    if (parts === null) return;
    Util.checkClass(parts, Map, Diagram, 'moveParts:parts');
    if (parts.count === 0) return;
    const temp = Point.alloc();
    const off = Point.alloc();
    off.assign(offset);
    if (isNaN(off.x)) off.x = 0;
    if (isNaN(off.y)) off.y = 0;
    // can't optimize a zero offset move, due to possible need to snap the nodes

    const wassuspended = this._suspendedRouting;
    if (!wassuspended) this._suspendRouting(parts);

    const nestednodes = Util.tempArray<DraggingNodeInfoPair>();
    const linkkvps = Util.tempArray<KeyValuePair<Part, DraggingInfo>>();
    const it = parts.iterator;
    let newloc = Point.alloc(); // re-used for each part
    while (it.next()) {
      const n = it.key;
      const v = it.value;
      if (n.isNodeLike()) {
        const ginfo = this._findContainingInfo(n, parts);
        if (ginfo !== null) {
          nestednodes.push(new DraggingNodeInfoPair(n, v, ginfo));
        } else if (!check || n.canMove()) {  // only check canMove() if the user is dragging
          const oldloc = v.point;
          temp.assign(oldloc);
          this.computeMove(n, temp.add(off), dragOptions, newloc); // modifies newloc
          n.location = newloc; // location setter copies newloc
          if (v.shifted === undefined) v.shifted = new Point(); // In case they pass in a plain object
          v.shifted.assign(newloc.subtract(oldloc));  // ... which is saved in the NodeInfoPair
        }
      } else {
        const l = it.key;
        if (l instanceof Link) {
          linkkvps.push((it as any)._kvp);
        }
      }
    }
    Point.free(newloc);

    const nnlength = nestednodes.length;
    for (let i = 0; i < nnlength; i++) {
      const nip = nestednodes[i] as DraggingNodeInfoPair;
      const oldloc = nip.info.point;
      temp.assign(oldloc);
      if (nip.groupInfo.shifted === undefined) nip.groupInfo.shifted = new Point(); // In case they pass in a plain object
      nip.node.location = temp.add(nip.groupInfo.shifted);
    }

    const fromoff = Point.alloc();
    const tooff = Point.alloc();
    const kvplength = linkkvps.length;
    for (let i = 0; i < kvplength; i++) {
      const kvp = linkkvps[i] as KeyValuePair<Part, DraggingInfo>;
      const link = kvp.key;
      if (!(link instanceof Link)) continue;
      if (link.suspendsRouting) {  // need to move the route explicitly
        link._savedRoute = null; // if moved during animation, don't let animation save route ?? only do when !AM.isTicking?
        const from = link.fromNode;
        const to = link.toNode;
        if (this.draggedLink !== null && dragOptions.dragsLink) {
          const oldlinkoff = kvp.value.point;
          // Pre 2.0, which did not allow dragged links to use dragComputation:
          if (link.dragComputation === null) {
            parts.add(link, this._newDraggingInfo(dragOptions, link, off));
            const dx = off.x - oldlinkoff.x;
            const dy = off.y - oldlinkoff.y;
            link.movePoints(dx, dy);
          } else { // new in 2.0:
            const oldLoc = Point.allocAt(0, 0);
            const pt0 = link.getPoint(0);
            if (pt0 && pt0.isReal()) oldLoc.assign(pt0);
            newloc = Point.alloc().assign(oldLoc).add(off);
            // Instead of this.computeMove(link, newloc, dragOptions, parts, newOffset);
            // Because we only want to use dragComputation, and not snap by default
            let snappt = newloc;
            if (dragOptions.isGridSnapEnabled && (dragOptions.isGridSnapRealtime || this.lastInput.up)) {
              snappt = Point.alloc();
              this.snapTo(link, newloc, snappt, dragOptions);
            }
            newloc.assign(link.dragComputation(link, newloc, snappt)).subtract(oldLoc); // newloc becomes the new offset
            parts.add(link, this._newDraggingInfo(dragOptions, link, newloc));
            const dx = newloc.x - oldlinkoff.x;
            const dy = newloc.y - oldlinkoff.y;
            link.movePoints(dx, dy);
            Point.free(oldLoc);
            Point.free(newloc);
            if (snappt !== newloc) Point.free(snappt);
          }
        } else {
          if (from !== null) {
            fromoff.assign(from.location);
            const info = parts.getValue(from);
            if (info !== null) {
              fromoff.subtract(info.point);
            }
          }
          if (to !== null) {
            tooff.assign(to.location);
            const info = parts.getValue(to);
            if (info !== null) {
              tooff.subtract(info.point);
            }
          }
          if (from !== null && to !== null) {  // connected on both ends
            // if both ends moved the same amount, can keep the original route
            if (fromoff.equalsApprox(tooff)) {
              const oldlinkoff = kvp.value.point;
              const linkoff = temp;
              linkoff.assign(fromoff);
              linkoff.subtract(oldlinkoff);
              parts.add(link, this._newDraggingInfo(dragOptions, link, fromoff));
              link.movePoints(linkoff.x, linkoff.y);
            } else {  // don't try to maintain shape of path
              link.suspendsRouting = false;
              link.invalidateRoute();
            }
          } else {  // only one or neither end is connected
            const oldlinkoff = kvp.value.point;
            let o: Point;
            if (from !== null) o = fromoff;
            else if (to !== null) o = tooff;
            else o = off;
            parts.add(link, this._newDraggingInfo(dragOptions, link, o));
            const dx = o.x - oldlinkoff.x;
            const dy = o.y - oldlinkoff.y;
            link.movePoints(dx, dy);
          }
        }
      } else {  // move explicitly if partly or fully disconnected
        if (link.fromNode === null || link.toNode === null) {
          const oldlinkoff = kvp.value.point;
          parts.add(link, this._newDraggingInfo(dragOptions, link, off));
          const dx = off.x - oldlinkoff.x;
          const dy = off.y - oldlinkoff.y;
          link.movePoints(dx, dy);
        }
      }
    }
    Point.free(temp);
    Point.free(off);
    Point.free(fromoff);
    Point.free(tooff);
    Util.freeArray(nestednodes);
    Util.freeArray(linkkvps);
    if (!wassuspended) {
      // validate parts is needed so
      // moved nodes update before resuming routing
      this.validateParts();
      this._resumeRouting(parts);
    }
  }

  /**
   * You do not need to call or orverride this method, but you may want to override {@link DraggingTool#computeMove}, which calls this method.
   *
   * This method computes the new location for a {@link Node} or simple {@link Part},
   * given a new desired location and an optional Map of dragged parts,
   * taking any grid-snapping into consideration,
   * any {@link Part#dragComputation} function,
   * and any {@link Part#minLocation} and {@link Part#maxLocation}.
   *
   * @expose
   * @this {Diagram}
   * @param {Part} n the Node or simple Part that is being moved
   * @param {Point} newloc the proposed new location
   * @param {DraggingOptions} dragOptions the proposed new location
   * @param {Point=} result an optional Point that is modified and returned
   * @return {Point} the possibly grid-snapped computed Point that is within the minimum and maximum permitted locations
   * @since 2.0
   */
  public computeMove(n: Part, newloc: Point, dragOptions: DraggingOptions, result?: Point): Point {
    if (result === undefined) result = new Point();
    result.assign(newloc);
    if (n === null) return result;

    let snappt = newloc;
    // ??? is this necessary? Probably only needed in _newDraggingInfo
    // only consider diagram snapping if this.isGridSnapEnabled
    // and either the snapping is realtime or when a mouse-up occurs
    // const nodeInGroup = (!dragOptions.groupsSnapMembers && n.containingGroup !== null);
    const snap = dragOptions.isGridSnapEnabled;
    if (snap /*&& !nodeInGroup*/ && (dragOptions.isGridSnapRealtime || this.lastInput.up)) {
      snappt = Point.alloc();
      this.snapTo(n, newloc, snappt, dragOptions);
    }

    // consider if the part has a special drag computation method defined
    const pt = (n.dragComputation !== null) ? n.dragComputation(n, newloc, snappt) : snappt;
    // interpret NaN in minLocation or maxLocation to continue using the old location X or Y
    const min = n.minLocation;
    let minx = min.x;
    // We are rounding these because _newDraggingInfo also rounds snapped location points.
    if (isNaN(minx)) minx = snap ? Math.round(n.location.x) : n.location.x;
    let miny = min.y;
    if (isNaN(miny)) miny = snap ? Math.round(n.location.y) : n.location.y;
    const max = n.maxLocation;
    let maxx = max.x;
    if (isNaN(maxx)) maxx = snap ? Math.round(n.location.x) : n.location.x;
    let maxy = max.y;
    if (isNaN(maxy)) maxy = snap ? Math.round(n.location.y) : n.location.y;
    result.assignTo(Math.max(minx, Math.min(pt.x, maxx)), Math.max(miny, Math.min(pt.y, maxy)));
    if (snappt !== newloc) Point.free(snappt);
    return result;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @return {DraggingOptions}
   */
  private getDragOptions(): DraggingOptions {
    const tool = this.toolManager.findTool('Dragging') as any;
    if (tool !== null) {
      return tool['dragOptions'] as DraggingOptions;
    } else {
      return this._defaultDragOptions;
    }
  }

  /**
   * @hidden @internal
   * (usually) during a drag: while moving parts, snapping to either nodes or the diagram's grid
   * @this {Diagram}
   * @param {Part} moving
   * @param {Point} pt
   * @param {Point} result
   * @param {DraggingOptions=} dragOptions
   * @return {Point}
   */
  public snapTo(moving: Part, pt: Point, result: Point, dragOptions?: DraggingOptions): Point {
    if (dragOptions === undefined) dragOptions = this.getDragOptions();
    result.assign(pt);
    if (moving === null) return result;
    const grid = this.grid;
    const cell = dragOptions.gridSnapCellSize;
    let cellw = cell.width;
    let cellh = cell.height;
    const origin = dragOptions.gridSnapOrigin;
    let origx = origin.x;
    let origy = origin.y;
    const spot = dragOptions.gridSnapCellSpot;
    if (grid !== null) {
      const gcell = grid.gridCellSize;
      if (isNaN(cellw)) cellw = gcell.width;
      if (isNaN(cellh)) cellh = gcell.height;
      const gorigin = grid.gridOrigin;
      if (isNaN(origx)) origx = gorigin.x;
      if (isNaN(origy)) origy = gorigin.y;
    }
    const gridpt = Point.allocAt(0, 0);
    gridpt.setSpot(0, 0, cellw, cellh, spot);
    Geo.findNearestInfiniteGridPoint(pt.x, pt.y, origx + gridpt.x, origy + gridpt.y, cellw, cellh, result);
    Point.free(gridpt);
    return result;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Map.<Part,DraggingInfo>} parts
   */
  public _suspendRouting(parts: Map<Part, DraggingInfo>): void {
    if (parts === null) return;
    this._suspendedRouting = true;
    const it = parts.iterator;
    while (it.next()) {
      const part = it.key;
      if (part instanceof Link) {
        part.suspendsRouting = true;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Map.<Part,DraggingInfo>} parts
   */
  public _resumeRouting(parts: Map<Part, DraggingInfo>): void {
    if (parts === null) return;
    const it = parts.iterator;
    while (it.next()) {
      const part = it.key;
      if (part instanceof Link) {
        part.suspendsRouting = false;
        // only need to re-route if the routing style is AvoidsNodes
        if (part.crossesOverNode()) part.invalidateRoute();
      }
    }
    this._suspendedRouting = false;
  }

  /**
   * @hidden @internal
   * Gets or sets the {@link Link} that may be dragged and reconnected
   * when {@link DraggingTool#dragsLink} is true.
   *
   * This property is set during {@link DraggingTool#doActivate}.
   * The default value is null.
   * Setting this property does not raise any events.
   */
  /** @export */ get draggedLink(): Link { return this._draggedLink; }
  /** @export */ set draggedLink(value: Link) {
    const old = this._draggedLink;
    if (old !== value) {
      this._draggedLink = value;
      if (value !== null) {
        this._oldFromPort = value.fromPort;
        this._oldToPort = value.toPort;
      }
    }
  }

  // Necessary to allow dragging tool to compile separately
  /** @hidden @internal */
  /* noexport */get oldFromPort() { return this._oldFromPort; }
  set oldFromPort(value) { this._oldFromPort = value; }
  /** @hidden @internal */
  /* noexport */get oldToPort() { return this._oldToPort; }
  set oldToPort(value) { this._oldToPort = value; }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Part} n
   * @param {Map.<Part,DraggingInfo>} parts
   * @return DraggingInfo
   */
  private _findContainingInfo(n: Part, parts: Map<Part, DraggingInfo>): DraggingInfo {
    const g = n.containingGroup;
    if (g !== null) {
      let info = this._findContainingInfo(g, parts);
      if (info !== null) return info;
      info = parts.getValue(g);
      if (info !== null) return info;
    }
    return null;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {DraggingOptions} dragOptions
   * @param {Part} part
   * @param {Point=} p
   * @return {DraggingInfo}
   */
  private _newDraggingInfo(dragOptions: DraggingOptions, part: Part, p?: Point): DraggingInfo {
    if (p === undefined) return new DraggingInfo(Point.Origin); // ?? this doesn't snap? Should it?

    // Only snap if grid snap is enabled and, if groupsSnapMembers, this part is not in a group
    let snap = dragOptions.isGridSnapEnabled;
    if (!dragOptions.groupsSnapMembers && part.containingGroup !== null) snap = false;

    // When snapping, we should round so link routing stays suspended:
    // a.location.x = 99.999999, b.location.x = 200.000001 would get snapped differently when moving half the grid cell width
    if (snap) {
      return new DraggingInfo(new Point(Math.round(p.x), Math.round(p.y)));
    }
    return new DraggingInfo(p.copy());
  }

  /**
   * @hidden @internal
   * called when adding/inserting layers
   * @this {Diagram}
   * @param {Layer} layer
   * @param {Layer} existingLayer
   */
  private _checkLayer(layer: Layer, existingLayer: Layer): void {
    Util.checkClass(layer, Layer, Diagram, 'addLayer:layer');
    if (layer.diagram !== null && layer.diagram !== this) Util.throwError('Cannot share a Layer with another Diagram: ' + layer + ' of ' + layer.diagram);

    if (existingLayer === null) {
      if (layer.diagram !== null) Util.throwError('Cannot add an existing Layer to this Diagram again: ' + layer);
    } else {
      Util.checkClass(existingLayer, Layer, Diagram, 'addLayer:existingLayer');
      if (existingLayer.diagram !== this) Util.throwError('Existing Layer must be in this Diagram: ' + existingLayer + ' not in ' + existingLayer.diagram);
      if (layer === existingLayer) Util.throwError('Cannot move a Layer before or after itself: ' + layer);
    }

    // now check for duplicate layer name
    if (layer.diagram === this) return;  // if already in this Diagram, it must be OK
    const name = layer.name;
    const layers = this._layers;
    const num = layers.count;
    for (let i = 0; i < num; i++) {
      if (layers.elt(i).name === name) {
        Util.throwError('Cannot add Layer with the name \'' + name + '\'; a Layer with the same name is already present in this Diagram.');
      }
    }
  }

  /**
   * Adds a new {@link Layer} to the list of layers.
   * If {@link Layer#isTemporary} is false, the layer is added after all existing non-temporary layers.
   * If {@link Layer#isTemporary} is true, the layer is added as the very last layer.
   * @export
   * @this {Diagram}
   * @param {Layer} layer The new Layer to add.  It is an error if the Layer already belongs to a Diagram.
   * @see {@link #addLayerBefore}
   * @see {@link #addLayerAfter}
   * @see {@link #removeLayer}
   * @see {@link #findLayer}
   */
  public addLayer(layer: Layer): void {
    this._checkLayer(layer, null);
    layer.setDiagram(this);
    const layers = this._layers;
    let i = layers.count - 1;
    if (!layer.isTemporary) {
      while (i >= 0 && layers.elt(i).isTemporary) i--;
    }
    layers.insertAt(i + 1, layer);
    if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Insert, 'layers', this, null, layer, null, i + 1);
    this.invalidateDraw();
    this.invalidateDocumentBounds();
  }

  /**
   * Adds a layer to the list of layers before a specified layer.
   * This method can also re-order layers.
   * @export
   * @this {Diagram}
   * @param {Layer} layer the new Layer to add or existing Layer to move in Z-order.
   * @param {Layer} existingLayer the other Layer in this Diagram which should come just after the new or moved layer.
   * @see {@link #addLayer}
   * @see {@link #addLayerAfter}
   * @see {@link #removeLayer}
   */
  public addLayerBefore(layer: Layer, existingLayer: Layer): void {
    this._checkLayer(layer, existingLayer);
    layer.setDiagram(this);
    const layers = this._layers;
    const idx = layers.indexOf(layer);
    if (idx >= 0) {
      layers.remove(layer);
      if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Remove, 'layers', this, layer, null, idx, null);
    }
    const num = layers.count;
    let i;
    for (i = 0; i < num; i++) {
      if (layers.elt(i) === existingLayer) {
        layers.insertAt(i, layer);
        break;
      }
    }
    if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Insert, 'layers', this, null, layer, null, i);
    this.invalidateDraw();
    if (idx < 0) this.invalidateDocumentBounds();
  }

  /**
   * Adds a layer to the list of layers after a specified layer.
   * This method can also re-order layers.
   * @export
   * @this {Diagram}
   * @param {Layer} layer the new Layer to add or existing Layer to move in Z-order.
   * @param {Layer} existingLayer the other Layer in this Diagram which should come just before the new or moved layer.
   * @see {@link #addLayer}
   * @see {@link #addLayerBefore}
   * @see {@link #removeLayer}
   */
  public addLayerAfter(layer: Layer, existingLayer: Layer): void {
    this._checkLayer(layer, existingLayer);
    layer.setDiagram(this);
    const layers = this._layers;
    const idx = layers.indexOf(layer);
    if (idx >= 0) {
      layers.remove(layer);
      if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Remove, 'layers', this, layer, null, idx, null);
    }
    const num = layers.count;
    let i;
    for (i = 0; i < num; i++) {
      if (layers.elt(i) === existingLayer) {
        layers.insertAt(i + 1, layer);
        break;
      }
    }
    if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Insert, 'layers', this, null, layer, null, i + 1);
    this.invalidateDraw();
    if (idx < 0) this.invalidateDocumentBounds();
  }

  /**
   * @hidden @internal
   * Adds a layer to the list of layers at a specified index
   * only used in undo and redo
   * @this {Diagram}
   * @param {Layer} layer The Layer to add.
   * @param {number} index int index to insert at.
   */
  private addLayerAt(layer: Layer, index: number): void {
    layer.setDiagram(this);
    this._layers.insertAt(index, layer);
  }

  /**
   * @hidden @internal
   * only used in undo and redo
   * @this {Diagram}
   * @param {number} index int
   */
  private removeLayerAt(index: number): void {
    this._layers.removeAt(index);
  }

  /**
   * Removes the given layer from the list of layers.
   *
   * Removing a layer does not remove the Parts in the layer. Instead, those Parts are placed into the default layer.
   * To remove all Parts in a layer you can call {@link Diagram#removeParts} with {@link Layer#parts} as the argument.
   *
   * You cannot remove the default layer, the one named with the empty string.
   * @export
   * @this {Diagram}
   * @param {Layer} layer
   * @see {@link #addLayer}
   * @see {@link #addLayerBefore}
   * @see {@link #addLayerAfter}
   * @see {@link #findLayer}
   */
  public removeLayer(layer: Layer): void {
    Util.checkClass(layer, Layer, Diagram, 'removeLayer:layer');
    if (layer.diagram !== this) Util.throwError('Cannot remove a Layer from another Diagram: ' + layer + ' of ' + layer.diagram);
    if (layer.name === '') return;
    const layers = this._layers;
    const index = layers.indexOf(layer);
    if (layers.remove(layer)) {
      const oldparts = layer._parts.copy();
      const pit = oldparts.iterator;
      while (pit.next()) {
        const part = pit.value;
        const oldlayername = part.layerName;
        if (oldlayername !== layer.name) {
          part.layerName = oldlayername;  // maybe switch to the originally intended layer
        } else {
          part.layerName = '';  // change to default layer
        }
      }
      if (this._model !== null) this.raiseChangedEvent(ChangedEvent.Remove, 'layers', this, layer, null, index, null);
      this.invalidateDraw();
      this.invalidateDocumentBounds();
    }
  }

  /**
   * Finds a layer with a given name.
   * @export
   * @this {Diagram}
   * @param {string} name
   * @return {Layer} a Layer with the given name, or null if no such layer was found.
   * @see {@link #addLayerBefore}
   * @see {@link #addLayerAfter}
   * @see {@link #removeLayer}
   */
  public findLayer(name: string): Layer | null {
    const it = this.layers;
    while (it.next()) {
      const layer = it.value;
      if (layer.name === name) return layer;
    }
    return null;
  }

  // For Model ChangedEvents, for convenience when the Diagram.model might be replaced

  /**
   * Register an event handler on this Diagram's {@link Diagram#model} that is called when there is a {@link ChangedEvent}
   * on the {@link Model}, not in this diagram.
   * Be sure to call {@link #removeModelChangedListener} when you are done with the diagram.
   *
   * This is convenient when the {@link Diagram#model} may be replaced.
   * Using this method to register a Model Changed listener is more convenient than calling {@link Model#addChangedListener} directly
   * because when this diagram's Model is replaced, one does not need to call {@link Model#removeChangedListener} on the old Model
   * and then {@link Model#addChangedListener} again on the new Model.
   *
   * You can establish Model Changed listeners when you create a Diagram with {@link GraphObject.make}.  For example:
   * ```js
   * $(go.Diagram, . . .,
   *    {
   *        "ModelChanged": function(e) { if (e.isTransactionFinished) saveModel(); }
   *        . . .
   *    })
   * ```
   *
   * Do not add or remove Changed listeners during the execution of a Changed listener.
   * @export
   * @this {Diagram}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #removeModelChangedListener}
   * @since 1.6
   */
  public addModelChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Diagram, 'addModelChangedListener:listener');
    if (this._modelChangedListeners === null) {
      this._modelChangedListeners = new List<((a: ChangedEvent) => void)>();
    }
    this._modelChangedListeners.add(listener);
    this.model.addChangedListener(listener);
  }

  /**
   * Unregister a {@link ChangedEvent} handler from this Diagram's {@link Diagram#model}.
   * @export
   * @this {Diagram}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #addModelChangedListener}
   * @since 1.6
   */
  public removeModelChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Diagram, 'removeModelChangedListener:listener');
    if (this._modelChangedListeners !== null) {
      this._modelChangedListeners.remove(listener);
      if (this._modelChangedListeners.count === 0) this._modelChangedListeners = null;
    }
    this.model.removeChangedListener(listener);
  }

  // For Diagram ChangedEvents and notification:

  /**
   * Register an event handler that is called when there is a {@link ChangedEvent} because this Diagram
   * or one of its Parts has changed, but not because the Model or any model data has changed.
   *
   * It is unusual to listen for Diagram ChangedEvents --
   * it is far more common to listen for specific {@link DiagramEvent}s by calling {@link #addDiagramListener},
   * or to listen for Model ChangedEvents (i.e. changes to the model) by calling {@link #addModelChangedListener}.
   *
   * Do not add or remove Changed listeners during the execution of a Changed listener.
   * @export
   * @this {Diagram}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #removeChangedListener}
   */
  public addChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Diagram, 'addChangedListener:listener');
    if (this._changedListeners === null) {
      this._changedListeners = new List<((a: ChangedEvent) => void)>();
    }
    this._changedListeners.add(listener);
  }

  /**
   * Unregister a {@link ChangedEvent} handler.
   * @export
   * @this {Diagram}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #addChangedListener}
   */
  public removeChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Diagram, 'removeChangedListener:listener');
    if (this._changedListeners !== null) {
      this._changedListeners.remove(listener);
      if (this._changedListeners.count === 0) this._changedListeners = null;
    }
  }

  /**
   * @hidden @internal
   * Notify any ChangedEvent listeners.
   * This is called by {@link #raiseChangedEvent}.
   * Each listener function registered with calls to {@link #addChangedListener} is called with the given event.
   * Unless {@link #skipsUndoManager} is true, this then calls {@link UndoManager#handleChanged}
   * so that the UndoManager gets the opportunity to record the change.
   * @this {Diagram}
   * @param {ChangedEvent} e
   */
  private callChangedListeners(e: ChangedEvent): void {
    if (!this.skipsUndoManager && !this.model.skipsUndoManager) {
      this.model.undoManager.handleChanged(e);
    }

    if (e.change !== ChangedEvent.Transaction) {
      this.isModified = true;
    }

    if (this._changedListeners !== null) {
      const list = this._changedListeners;
      const len = list.length;
      for (let i = 0; i < len; i++) {
        const func = list.elt(i);
        func(e);
      }
    }
  }

  /**
   * @hidden @internal
   * Call this method to notify that the diagram or its objects have changed.
   * This constructs a {@link ChangedEvent} and calls the Diagram listeners.
   * @export
   * @this {Diagram}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string|function(Object):*} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}, {@link Diagram}, or a {@link Layer}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raiseChangedEvent(change: EnumValue, propertyname: string | ((a: ObjectData, b: any) => any), obj: ObjectData,
                           oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    // ChangedEvent property setters will check argument types
    if (oldparam === undefined) oldparam = null;
    if (newparam === undefined) newparam = null;

    const e = new ChangedEvent();  // ?? cache event objects
    e.diagram = this;
    e.change = change;
    e.propertyName = propertyname;
    e.object = obj;
    e.oldValue = oldval;
    e.oldParam = oldparam;
    e.newValue = newval;
    e.newParam = newparam;

    this.callChangedListeners(e);
  }

  /**
   * @hidden @internal
   * Call this method to notify about a diagram property having changed value.
   * @export
   * @this {Diagram}
   * @param {string|function(*)} propertyname the name of the property.
   * @param {*} oldval the previous or old value for the property.
   * @param {*} newval the next or new value for the property.
   * @param {*=} oldparam an optional value additionally describing the old value.
   * @param {*=} newparam an optional value additionally describing the new value.
   */
  protected raiseChanged(propertyname: string | ((a: any) => void), oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    this.raiseChangedEvent(ChangedEvent.Property, propertyname, this, oldval, newval, oldparam, newparam);
  }

  /**
   * This read-only property returns the {@link AnimationManager} for this Diagram.
   * @since 1.4
   */
  /** @export */ get animationManager(): AnimationManager { return this._animationManager; }

  // for undo/redo:

  /**
   * This read-only property returns the {@link UndoManager} for this Diagram, which actually belongs to the {@link #model}.
   *
   * The default UndoManager has its {@link UndoManager#isEnabled} property set to false.
   * If you want users to undo and redo, you should set that property to true once you have initialized the Diagram or its Model.
   *
   * Note that the UndoManager might be shared with other Diagrams that are showing the same Model.
   * The UndoManager might also be shared with other Models too.
   */
  /** @export */ get undoManager(): UndoManager { return this._model.undoManager; }

  /**
   * Gets or sets whether ChangedEvents are not recorded by the UndoManager.
   * The initial and normal value is false.
   * WARNING: while this property is true do not perform any changes that cause any previous transactions
   * to become impossible to undo.
   *
   * While this property is true, changing the Diagram or any GraphObject does not call {@link UndoManager#handleChanged}.
   * Even when this property is true,
   * transactions (such as calls to {@link #startTransaction}) and
   * undo/redo (such as calls to {@link CommandHandler#undo})
   * are still delegated to the {@link #undoManager}.
   *
   * You should set this to true only temporarily, and you should remember its previous value
   * before setting this to true.
   * When finishing the period for which you want the UndoManager to be disabled,
   * you should set this back to the remembered value it had before it was set to true.
   *
   * For more permanent disabling of the {@link UndoManager}, set {@link UndoManager#isEnabled} to false.
   *
   * Setting this property also sets {@link Model#skipsUndoManager} to the same value.
   * Setting this property does not notify about any changed event.
   */
  /** @export */ get skipsUndoManager(): boolean { return this._skipsUndoManager; }
  /** @export */ set skipsUndoManager(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Diagram, 'skipsUndoManager');
    this._skipsUndoManager = value;
    this._model.skipsUndoManager = value;
  }

  // * When set to true Diagram layouts will not get invalidated.
  // * Used by some tools stop real-time layouts from occuring during their operation.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get delaysLayout(): boolean { return this._delaysLayout; }
  /** @export */ set delaysLayout(value: boolean) {
    this._delaysLayout = value;
  }

  /**
   * @hidden @internal
   * This is called during an undo or redo to modify the diagram or its objects.
   * @this {Diagram}
   * @param {ChangedEvent} e This describes the change that needs to be done.
   * @param {boolean} undo If true, this method should restore the older state, otherwise the newer state.
   */
  public changeState(e: ChangedEvent, undo: boolean): void {
    if (e === null) return;
    if (e.diagram !== this) return;  // ignore changes to Model
    const oldskips = this.skipsModelSourceBindings;
    try {
      this.skipsModelSourceBindings = true;
      const chg = e.change;
      // closure will whine if you re-define vars within if statements
      let rcd: RowColumnDefinition;
      if (chg === ChangedEvent.Property) {
        const obj = e.object as any;
        const pn = e.propertyName;
        const v = e.getValue(undo);
        Model.safePropertySet(obj, pn, v);
        if (obj instanceof GraphObject) {
          const part = obj.part;
          if (part !== null) part.invalidateAdornments();
        }
        this.isModified = true;
      } else if (chg === ChangedEvent.Insert) {
        const container = e.object;
        const idx = e.newParam;
        const obj = e.newValue;
        if (container instanceof Panel) {
          if (typeof idx === 'number' && obj instanceof GraphObject) {
            if (undo) {
              container.internalRemove(idx);
            } else {
              container.insertAt(idx, obj);
            }
            const part = container.part;
            if (part !== null) part.invalidateAdornments();
          } else if (typeof idx === 'number' && obj instanceof RowColumnDefinition) {
            if (undo) {
              if (obj.isRow) {
                container.removeRowDefinition(idx);
              } else {
                container.removeColumnDefinition(idx);
              }
            } else {
              if (obj.isRow) {
                rcd = container.getRowDefinition(obj.index);
              } else {
                rcd = container.getColumnDefinition(obj.index);
              }
              // now can initialize the RowColumnDefinition properties
              rcd.copyFrom(obj);
            }
          }
        } else if (container instanceof Layer) {
          const layerSwap = (e.oldParam === true); // oldParam is true if a part is changing layers instead of being deleted
          if (typeof idx === 'number' && obj instanceof Part) {
            if (undo) {
              // undoing insertion of a selected or highlighted Part deselects/unhighlights it,
              // so that after undo there are no non-present parts that are selected/highlighted
              obj.isSelected = false;
              obj.isHighlighted = false;
              obj.invalidateAdornments();
              container.internalRemove(layerSwap ? idx : -1, obj, layerSwap);
            } else {
              container.internalAdd(idx, obj, layerSwap);
            }
          }
        } else if (container instanceof Diagram) {
          if (typeof idx === 'number' && obj instanceof Layer) {
            if (undo) {
              this.removeLayerAt(idx);
            } else {
              this.addLayerAt(obj, idx);
            }
          }
        } else {
          Util.throwError('unknown ChangedEvent.Insert object: ' + e.toString());
        }
        this.isModified = true;
      } else if (chg === ChangedEvent.Remove) {
        const container = e.object;
        const idx = e.oldParam;
        const obj = e.oldValue;
        if (container instanceof Panel) {
          if (typeof idx === 'number' && obj instanceof GraphObject) {
            if (undo) {
              container.insertAt(idx, obj);
            } else {
              container.internalRemove(idx);
            }
          } else if (typeof idx === 'number' && obj instanceof RowColumnDefinition) {
            if (undo) {
              if (obj.isRow) {
                rcd = container.getRowDefinition(obj.index);
              } else {
                rcd = container.getColumnDefinition(obj.index);
              }
              // now can initialize the RowColumnDefinition properties
              rcd.copyFrom(obj);
            } else {
              if (obj.isRow) {
                container.removeRowDefinition(idx);
              } else {
                container.removeColumnDefinition(idx);
              }
            }
          }
        } else if (container instanceof Layer) {
          const layerSwap = (e.newParam === true); // newParam is true if a part is changing layers instead of being deleted
          if (typeof idx === 'number' && obj instanceof Part) {
            if (undo) {
              if (container._parts.indexOf(obj) < 0) container.internalAdd(idx, obj, layerSwap);
            } else {
              obj.isSelected = false;
              obj.isHighlighted = false;
              obj.invalidateAdornments();
              container.internalRemove(layerSwap ? idx : -1, obj, layerSwap);
            }
          }
        } else if (container instanceof Diagram) {
          if (typeof idx === 'number' && obj instanceof Layer) {
            if (undo) {
              this.addLayerAt(obj, idx);
            } else {
              this.removeLayerAt(idx);
            }
          }
        } else {
          Util.throwError('unknown ChangedEvent.Remove object: ' + e.toString());
        }
        this.isModified = true;
      } else if (chg === ChangedEvent.Transaction) {
        // don't do anything
      } else {
        Util.throwError('unknown ChangedEvent: ' + e.toString());
      }
    } finally {
      this.skipsModelSourceBindings = oldskips;
    }
  }

  /**
   * Begin a transaction, where the changes are held by a {@link Transaction} object
   * in the {@link UndoManager}.
   * This just calls {@link UndoManager#startTransaction}.
   * @export
   * @this {Diagram}
   * @param {string=} tname a descriptive name for the transaction.
   * @return {boolean} the value returned by {@link UndoManager#startTransaction}.
   * @see {@link #commit}
   */
  public startTransaction(tname?: string): boolean {
    return this.undoManager.startTransaction(tname);
  }

  /**
   * Commit the changes of the current transaction.
   * This just calls {@link UndoManager#commitTransaction}.
   * @export
   * @this {Diagram}
   * @param {string=} tname a descriptive name for the transaction.
   * @return {boolean} the value returned by {@link UndoManager#commitTransaction}.
   */
  public commitTransaction(tname?: string): boolean {
    return this.undoManager.commitTransaction(tname);
  }

  /**
   * Rollback the current transaction, undoing any recorded changes.
   * This just calls {@link UndoManager#rollbackTransaction}.
   * @export
   * @this {Diagram}
   * @return {boolean} the value returned by {@link UndoManager#rollbackTransaction}.
   */
  public rollbackTransaction(): boolean {
    return this.undoManager.rollbackTransaction();
  }

  /**
   * Starts a new transaction, calls the provided function, and commits the transaction.
   * Code is called within a try-finally statement.
   * If the function does not return normally, this rolls back the transaction rather than committing it.
   * Example usage:
   * ```js
   * myDiagram.commit(d => d.remove(somePart), "Remove Part");
   * ```
   * @export
   * @this {Diagram}
   * @param {Function} func the function to call as the transaction body
   * @param {(string|null)=} tname a descriptive name for the transaction, or null to temporarily set {@link #skipsUndoManager} to true;
   *        if no string transaction name is given, an empty string is used as the transaction name
   * @since 1.8
   */
  public commit(func: (d: Diagram) => void, tname?: string | null): void {
    let tn = tname;
    if (tn === undefined) tn = '';
    const oldskips = this.skipsUndoManager;
    if (tn === null) {
      this.skipsUndoManager = true;
      tn = '';
    }
    this.undoManager.startTransaction(tn);
    let success = false;
    try {
      func(this);
      success = true;
    } finally {
      if (success) {
        this.undoManager.commitTransaction(tn);
      } else {
        this.undoManager.rollbackTransaction();
      }
      this.skipsUndoManager = oldskips;
    }
  }


  /**
   * Update all of the data-bound properties of {@link Node}s and {@link Link}s in this diagram,
   * without having to call {@link Model#setDataProperty}.
   * This copies/converts model data properties to set properties on {@link Part}s.
   * This method does not conduct a transaction, so you need to start and commit one yourself.
   *
   * It is better to call {@link Model#setDataProperty} to modify data properties,
   * because that will both record changes for undo/redo and will update all bindings
   * that make depend on that property.
   * Simply modifying the data and calling an "update..." method will not be able to record
   * the previous value(s) of properties in the model data to support undo.
   *
   * If you know which model data objects have been modified,
   * it will be more efficient to update only the {@link Part}s that need it
   * by calling {@link Panel#updateTargetBindings}.
   *
   * To update relationships between nodes, call {@link #updateAllRelationshipsFromData}.
   * @export
   * @this {Diagram}
   * @param {string=} srcprop An optional source data property name:
   *   when provided, only evaluates those {@link Binding}s that use that particular property;
   *   when not provided or when it is the empty string, all bindings are evaluated.
   * @see {@link #updateAllRelationshipsFromData}
   */
  public updateAllTargetBindings(srcprop?: string): void {
    this.partManager.updateAllTargetBindings(srcprop);
  }

  /**
   * Add or remove any nodes or links according to additional or missing data objects in the model
   * and update all of the references to nodes, in case they had been modified in the model without
   * properly notifying the model by calling {@link Model#addNodeData} or
   * {@link GraphLinksModel#removeLinkData} or
   * {@link GraphLinksModel#setGroupKeyForNodeData} or
   * {@link GraphLinksModel#setToKeyForLinkData} or other similar methods.
   * This method does not conduct a transaction, so you need to start and commit one yourself.
   *
   * It is better to call {@link Model#addNodeData}, {@link Model#removeNodeData},
   * {@link GraphLinksModel#addLinkData}, {@link GraphLinksModel#removeLinkData},
   * {@link Model#setDataProperty}, and other model methods to add/remove/modify data,
   * because those methods will both record changes for undo/redo and will update all bindings
   * that make depend on that property.
   * Simply modifying the data and calling an "update..." method will not be able to record
   * the previous value(s) of properties in the model data to support undo.
   *
   * This only adds, removes, or updates the relationships between nodes and links,
   * to have them reflect what is now declared in the model data.
   * If you know which model data objects have been modified,
   * it will be more efficient to update only the {@link Part}s that need it
   * by calling {@link Part#updateRelationshipsFromData}.
   *
   * To update {@link GraphObject} properties that are data bound, call {@link #updateAllTargetBindings}.
   * @export
   * @this {Diagram}
   * @see {@link #updateAllTargetBindings}
   * @since 1.5
   */
  public updateAllRelationshipsFromData(): void {
    this.partManager.updateAllRelationshipsFromData();
  }

  /**
   * @hidden @internal
   * This method is called when {@link Diagram#scale} changes.
   * This may change the {@link Diagram#position} in order to maintain the desired content alignment.
   * This method may be called during initialization even if the old and new scale values are the same.
   * This method calls {@link Diagram#onViewportBoundsChanged}.
   * @param {number} oldscale
   * @param {number} newscale
   */
  private onScaleChanged(oldscale: number, newscale: number): void {
    // if (this.isAnimating) return;
    const AM = this.animationManager;
    if (this._isAligning || this._isUpdating) {
      this._scale = newscale;
      AM.addDiagramScale(oldscale, this._scale);
      return;
    }
    this._isAligning = true;
    const can = this._canvas;
    if (can === null) {
      this._scale = newscale;
      return;
    }



    const oldvbounds = this.viewportBounds.copy();
    const viewwidth = this._canvasWidth;
    const viewheight = this._canvasHeight;
    oldvbounds.width = this._canvasWidth / oldscale;
    oldvbounds.height = this._canvasHeight / oldscale;

    // in element coordinates, not document coordinates
    let zx = this.zoomPoint.x;
    let zy = this.zoomPoint.y;

    // if the zoom point is NaN, default it according to the ...ContentAlignment properties
    const calign = this.contentAlignment;
    if (isNaN(zx)) {
      if (calign.isSide()) {
        if (calign.includesSide(Spot.LeftSide)) zx = 0;
        else if (calign.includesSide(Spot.RightSide)) zx = viewwidth - 1;
      } else if (calign.isSpot()) {
        zx = calign.x * (viewwidth - 1);
      } else {  // None or Default
        zx = viewwidth / 2;
      }
    }
    if (isNaN(zy)) {
      if (calign.isSide()) {
        if (calign.includesSide(Spot.TopSide)) zy = 0;
        else if (calign.includesSide(Spot.BottomSide)) zy = viewheight - 1;
      } else if (calign.isSpot()) {
        zy = calign.y * (viewheight - 1);
      } else {  // None or Default
        zy = viewheight / 2;
      }
    }

    if (this.scaleComputation !== null && !this.animationManager.isAnimating) {
      newscale = this.scaleComputation(this, newscale); // ?? possibly force another scale computation at the end of animation
    }
    if (newscale < this.minScale) { newscale = this.minScale; }
    if (newscale > this.maxScale) { newscale = this.maxScale; }

    // must be done while isAligning still true to stop extra onViewportBoundsChanged calls
    const p = Point.allocAt(this._position.x + zx / oldscale - zx / newscale, this._position.y + zy / oldscale - zy / newscale);
    this.position = p;
    Point.free(p);

    this._scale = newscale; // must be set before onViewportBoundsChanged is called
    this.onViewportBoundsChanged(oldvbounds, this.viewportBounds, oldscale, false);
    this._isAligning = false;

    // not needed? requestUpdate later on will ensure that it happens then
    // requestUpdate->doupdate->updateScrollbars x2
    // update: needs to stay in, breaks zoomToFit() (write test)
    this.onContentAlignmentChanged(false);

    AM.addDiagramScale(oldscale, this._scale);

    this.invalidateDraw();
    this.invalidateScroll();
  }

  /**
   * @hidden @internal
   * This method is called when the {@link #position} or {@link #scale} change, or when update notices the canvas has changed size.
   * @this {Diagram}
   * @param {Rect} oldvbounds a Rect in document coordinates.
   * @param {Rect} newvbounds a Rect in document coordinates.
   * The X or Y may have changed because of a change to {@link #position}.
   * The Width or Height may have changed either because this Diagram has changed size
   * or because the {@link #scale} has changed value.
   * @param {number} oldscale
   * @param {boolean} isScroll true if the viewport change was only caused by a scrollbar changing size.
   */
  protected onViewportBoundsChanged(oldvbounds: Rect, newvbounds: Rect, oldscale: number, isScroll: boolean): void {
    if (oldvbounds.equals(newvbounds)) return; // no op if the bounds didn't actually change
    if (isScroll === undefined) isScroll = false;
    // check isScroll because onViewportBoundsChanged gets called while validating scroll
    if (!isScroll) this.invalidateScroll();
    this.invalidateViewportObjects();
    const lay = this.layout;
    if ((lay !== null && lay.isViewportSized && this.autoScale === Diagram.None && !isScroll) &&
      (oldvbounds.width !== newvbounds.width || oldvbounds.height !== newvbounds.height)) {
      lay.invalidateLayout();
    }

    // This is necessary because the VPB will change (such as on keypress) without the mouse moving
    // Chrome accomodates this automatically by firing a mouseMove event, other browsers do not
    // maybe the mouse is hovering over a different object (change tooltips, etc)
    const mgr = this.currentTool;
    if (this._isMouseOverDiagram === true && mgr instanceof ToolManager) {
      // adjust the mouse point for the new position/scale
      this.lastInput.documentPoint = this.transformViewToDoc(this.lastInput.viewPoint);
      mgr.standardMouseActivity(this);
    }
    if (!this._isAligning) this.invalidateViewport(oldvbounds, newvbounds);

    this._updateBackgroundGrid();

    // For now, if the scrollbar caused the viewportBounds change, we pass an empty object instead of null
    this._viewportBoundsChangedInfo.scale = oldscale;
    this._viewportBoundsChangedInfo.position.x = oldvbounds.x;
    this._viewportBoundsChangedInfo.position.y = oldvbounds.y;
    this._viewportBoundsChangedInfo.bounds.assign(oldvbounds);
    this._viewportBoundsChangedInfo.isScroll = isScroll;
    this.raiseDiagramEvent('ViewportBoundsChanged', this._viewportBoundsChangedInfo, oldvbounds);

    if (this.isVirtualized) {
      this.links.each(function(l) {
        if (l.isAvoiding && l.actualBounds.intersectsRect(newvbounds)) {
          l.invalidateRoute();
        }
      });
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Rect=} viewRect optional viewportBounds replacement.
   */
  public _updateBackgroundGrid(viewRect?: Rect): void {
    if (viewRect === undefined) viewRect = null;
    const grid = this._grid;
    if (grid === null || !grid.visible) return;
    const tile = Size.alloc();
    grid.computeTileSize(tile);
    let width = 0.0;
    let height = 0.0;
    let vx = 0.0;
    let vy = 0.0;
    if (viewRect !== null) {
      width = viewRect.width;
      height = viewRect.height;
      vx = viewRect.x;
      vy = viewRect.y;
    } else {
      const bounds = Rect.alloc();
      const vb = this.viewportBounds;
      bounds.assignTo(vb.x, vb.y, vb.width, vb.height);
      if (!bounds.isReal()) { Rect.free(bounds); return; }
      width = bounds.width;
      height = bounds.height;
      vx = bounds.x;
      vy = bounds.y;
      Rect.free(bounds);
    }
    grid.width = width + tile.width * 2;
    grid.height = height + tile.height * 2;
    const gridpt = Point.alloc();
    Geo.findNearestInfiniteGridPoint(vx, vy, 0, 0, tile.width, tile.height, gridpt);
    gridpt.offset(-tile.width, -tile.height);
    Size.free(tile);
    grid.part.location = gridpt;
    Point.free(gridpt);
  }

  // Selection

  /**
   * Deselect all selected {@link Part}s.
   * This removes all parts from the {@link #selection} collection.
   * This method raises the "ChangingSelection" and "ChangedSelection" Diagram events.
   * @export
   * @this {Diagram}
   * @see {@link #select}
   * @see {@link #selectCollection}
   */
  public clearSelection(): void {
    const anySel = this.selection.count > 0;
    if (anySel) this.raiseDiagramEvent('ChangingSelection', this.selection);
    this.internalClearSelection();
    if (anySel) this.raiseDiagramEvent('ChangedSelection', this.selection);
  }

  /**
   * @hidden @internal
   * Deselect all selected {@link Part}s,
   * but don't raise "ChangingSelection" or "ChangedSelection" Diagram events.
   * @this {Diagram}
   */
  public internalClearSelection(): void {
    const coll = this.selection;
    if (coll.count > 0) {
      const arr = coll.toArray();
      const len = arr.length;
      for (let i = 0; i < len; i++) {
        const part = arr[i];
        part.isSelected = false;
      }
      coll.thaw();
      coll.clear();
      coll.freeze();
    }
  }

  /**
   * Make the given object the only selected object.
   * This method raises the "ChangingSelection" and "ChangedSelection" Diagram events.
   * @export
   * @this {Diagram}
   * @param {Part} part a Part that is already in a layer of this Diagram.
   * If the value is null, this does nothing.
   * @see {@link #selectCollection}
   * @see {@link #clearSelection}
   */
  public select(part: Part | null): void {
    if (part === null) return;
    Util.checkClass(part, Part, Diagram, 'select:part');
    if (part.layer.diagram === this) {
      if (!part.isSelected || this.selection.count > 1) {
        this.raiseDiagramEvent('ChangingSelection', this.selection);
        this.internalClearSelection();  // don't raise ChangingSelection/Finished
        part.isSelected = true;
        this.raiseDiagramEvent('ChangedSelection', this.selection);
      }
    }
  }

  /**
   * Select all of the {@link Part}s supplied in the given collection, and deselect all other Parts.
   * This method raises the "ChangingSelection" and "ChangedSelection" Diagram events.
   * @export
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll a {@link List} or {@link Set} or {@link Iterator} or Array, of {@link Part}s to be selected.
   * @see {@link #select}
   * @see {@link #clearSelection}
   */
  public selectCollection(coll: Iterable<Part> | Array<Part>): void {
    this.raiseDiagramEvent('ChangingSelection', this.selection);
    this.internalClearSelection();  // ??? ought not to de-select things in the given collection
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        const obj = Util.arrayElt(coll, i);
        if (!(obj instanceof Part)) Util.throwError('Diagram.selectCollection given something that is not a Part: ' + obj);
        obj.isSelected = true;
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const obj = it.value;
        if (!(obj instanceof Part)) Util.throwError('Diagram.selectCollection given something that is not a Part: ' + obj);
        obj.isSelected = true;
      }
    }
    this.raiseDiagramEvent('ChangedSelection', this.selection);
  }

  /**
   * Remove highlights from all {@link Part}s.
   * This removes all parts from the {@link #highlighteds} collection.
   * @export
   * @this {Diagram}
   * @see {@link #highlight}
   * @see {@link #highlightCollection}
   * @see {@link Part#isHighlighted}
   * @since 1.4
   */
  public clearHighlighteds(): void {
    const coll = this.highlighteds;
    if (coll.count > 0) {
      const arr = coll.toArray();
      const len = arr.length;
      for (let i = 0; i < len; i++) {
        const part = arr[i];
        part.isHighlighted = false;
      }
      coll.thaw();
      coll.clear();
      coll.freeze();
    }
  }

  /**
   * Make the given part the only highlighted part.
   * @export
   * @this {Diagram}
   * @param {Part} part a Part that is already in a layer of this Diagram.
   * If the value is null, this does nothing.
   * @see {@link Part#isHighlighted}
   * @see {@link #highlightCollection}
   * @see {@link #clearHighlighteds}
   * @since 1.4
   */
  public highlight(part: Part | null): void {
    if (part !== null && part.layer.diagram === this) {
      Util.checkClass(part, Part, Diagram, 'highlight:part');
      if (!part.isHighlighted || this.highlighteds.count > 1) {
        this.clearHighlighteds();
        part.isHighlighted = true;
      }
    }
  }

  /**
   * Highlight all of the {@link Part}s supplied in the given collection, and clear all other highlighted Parts.
   * @export
   * @this {Diagram}
   * @param {Iterable.<Part>|Array.<Part>} coll a {@link List} or {@link Set} or {@link Iterator} or Array, of {@link Part}s to be highlighted.
   * @see {@link Part#isHighlighted}
   * @see {@link #highlight}
   * @see {@link #clearHighlighteds}
   * @since 1.4
   */
  public highlightCollection(coll: Iterable<Part> | Array<Part>): void {
    const collset = new Set<Part>().addAll(coll);  // a Set of Parts to highlight
    const collclear = this.highlighteds.copy().removeAll(collset);  // currently highlighted Parts that are not in COLL
    const hit = collclear.iterator;
    while (hit.next()) {
      const part = hit.value;
      part.isHighlighted = false;
    }
    const it = collset.iterator;
    while (it.next()) {
      const obj = it.value;
      if (!(obj instanceof Part)) Util.throwError('Diagram.highlightCollection given something that is not a Part: ' + obj);
      obj.isHighlighted = true;  // some of these Parts might already have been highlighted -- no-op is OK
    }
  }

  /**
   * Scrolling function used by primarily by {@link #commandHandler}'s {@link CommandHandler#doKeyDown}.
   * @export
   * @this {Diagram}
   * @param {string} unit A string representing the unit of the scroll operation. Can only be 'pixel', 'line', 'page', or 'document'.
   * @param {string} dir The direction of the scroll operation. Can only be 'up', 'down', 'left', or 'right'.
   * @param {number=} dist An optional distance multiplier, for multiple pixels, lines, or pages. The default value is 1.
   *                       This argument is ignored when the unit is 'document'.
   * @see {@link #scrollToRect}
   * @see {@link #centerRect}
   */
  public scroll(unit: ('pixel' | 'line' | 'page' | 'document'),
                dir: ('up' | 'down' | 'left' | 'right'),
                dist?: number): void {
    if (dist === undefined) dist = 1;
    const vertical = (dir === 'up' || dir === 'down');

    let scrollPixels = 0.0;
    if (unit === 'pixel') {
      scrollPixels = dist;
    } else if (unit === 'line') {
      scrollPixels = dist * (vertical ? this.scrollVerticalLineChange : this.scrollHorizontalLineChange);
    } else if (unit === 'page') {
      // We want the page's length in pixels minue one line
      // But we don't want accidentally scroll backwards
      let viewport = vertical ? this.viewportBounds.height : this.viewportBounds.width;
      viewport *= this.scale;
      if (viewport !== 0) {
        const length = Math.max(viewport - (vertical ? this.scrollVerticalLineChange : this.scrollHorizontalLineChange), 0);
        scrollPixels = dist * length;
      }
    } else if (unit === 'document') {
      const b = this.documentBounds;
      const v = this.viewportBounds;
      const pt = Point.alloc() as Point;
      if (dir === 'up') {
        this.position = pt.assignTo(v.x, b.y);
      } else if (dir === 'left') {
        this.position = pt.assignTo(b.x, v.y);
      } else if (dir === 'down') {
        this.position = pt.assignTo(v.x, b.bottom - v.height);
      } else if (dir === 'right') {
        this.position = pt.assignTo(b.right - v.width, v.y);
      }
      Point.free(pt);
      return;
    } else {
      Util.throwError('scrolling unit must be \'pixel\', \'line\', \'page\', or \'document\', not: ' + unit);
    }

    scrollPixels /= this.scale;

    const p = this.position.copy();
    if (dir === 'up') {
      p.y = this.position.y - scrollPixels;
    } else if (dir === 'down') {
      p.y = this.position.y + scrollPixels;
    } else if (dir === 'left') {
      p.x = this.position.x - scrollPixels;
    } else if (dir === 'right') {
      p.x = this.position.x + scrollPixels;
    } else {
      Util.throwError('scrolling direction must be \'up\', \'down\', \'left\', or \'right\', not: ' + dir);
    }
    this.position = p;
  }

  /**
   * Modifies the {@link #position} to show a given Rect of the Diagram by centering the
   * viewport on that Rect. Does nothing if the Rect is already in view.
   *
   * See also {@link #centerRect}
   * @export
   * @this {Diagram}
   * @param {Rect} r
   * @see {@link #centerRect}
   * @see {@link #scroll}
   */
  public scrollToRect(r: Rect): void {
    const vb = this.viewportBounds;
    if (vb.containsRect(r)) return; // do nothing if R is already in view
    const midpt = r.center;
    midpt.x -= vb.width / 2;
    midpt.y -= vb.height / 2;
    this.position = midpt;
  }

  /**
   * Modifies the {@link #position} to show a given Rect of the Diagram by centering the
   * viewport on that Rect.
   * @export
   * @this {Diagram}
   * @param {Rect} r
   * @see {@link #scrollToRect}
   * @see {@link #scroll}
   */
  public centerRect(r: Rect): void {
    const vb = this.viewportBounds;
    const midpt = r.center;
    midpt.x -= vb.width / 2;
    midpt.y -= vb.height / 2;
    this.position = midpt;
  }

  /**
   * Given a Point in document coorindates, return a new Point in viewport coordinates.
   * @export
   * @this {Diagram}
   * @param {Point} p
   * @return {Point} The given Point converted into View coordinates.
   * @see {@link #transformViewToDoc}
   * @see {@link GraphObject#getDocumentPoint}
   */
  public transformDocToView(p: Point): Point {
    const t = this._transform;
    t.reset();
    if (this._scale !== 1) { t.scale(this._scale); }
    const pos = this._position;
    if ((pos.x !== 0 || pos.y !== 0) && isFinite(pos.x) && isFinite(pos.y)) {
      t.translate(-pos.x, -pos.y);
    }
    return p.copy().transform(this._transform);
  }

  /**
   * @hidden @internal
   * Given a rect[0,1,2,3] in document coordinates, return a new rect in view coordinates.
   * @export
   * @this {Diagram}
   * @param {Rect} rect
   * @return {Rect}
   */
  public transformRectDocToView(rect: Rect): Rect {
    const t = this._transform;
    const x = rect.x;
    const y = rect.y;
    const r = x + rect.width;
    const b = y + rect.height;
    const m0 = t.m11;
    const m1 = t.m12;
    const m2 = t.m21;
    const m3 = t.m22;
    const m4 = t.dx;
    const m5 = t.dy;
    const ltx = x * m0 + y * m2 + m4;
    const lty = x * m1 + y * m3 + m5;
    const rtx = r * m0 + y * m2 + m4;
    const rty = r * m1 + y * m3 + m5;
    const lbx = x * m0 + b * m2 + m4;
    const lby = x * m1 + b * m3 + m5;
    const rbx = r * m0 + b * m2 + m4;
    const rby = r * m1 + b * m3 + m5;
    let minx = ltx;
    let maxx = ltx;
    let miny = lty;
    let maxy = lty;

    minx = Math.min(minx, rtx);
    maxx = Math.max(maxx, rtx);
    miny = Math.min(miny, rty);
    maxy = Math.max(maxy, rty);
    minx = Math.min(minx, lbx);
    maxx = Math.max(maxx, lbx);
    miny = Math.min(miny, lby);
    maxy = Math.max(maxy, lby);
    minx = Math.min(minx, rbx);
    maxx = Math.max(maxx, rbx);
    miny = Math.min(miny, rby);
    maxy = Math.max(maxy, rby);

    return new Rect(minx, miny, maxx - minx, maxy - miny);
  }

  /**
   * Given a point in viewport coordinates, return a new Point in document coordinates.
   * @export
   * @this {Diagram}
   * @param {Point} p
   * @return {Point} The given point converted into Document coordinates.
   * @see {@link #transformDocToView}
   * @see {@link GraphObject#getDocumentPoint}
   */
  public transformViewToDoc(p: Point): Point {
    // ?? We might optimize this in the future instead of (re)constructing that transform every time. Same for doctoview
    const t = this._transform;
    t.reset();
    if (this._scale !== 1) { t.scale(this._scale); }
    const pos = this._position;
    if ((pos.x !== 0 || pos.y !== 0) && isFinite(pos.x) && isFinite(pos.y)) {
      t.translate(-pos.x, -pos.y);
    }
    return p.copy().transformInverted(this._transform);
  }

  /**
   * The default autoScale type, used as the value of {@link Diagram#autoScale}:
   * The Diagram does not attempt to scale so that its documentBounds would fit the view.
   * @constant
   */
  public static None = new EnumValue(Diagram, 'None', 0);

  /**
   * Diagrams with this autoScale type, used as the value of {@link Diagram#autoScale},
   * are scaled uniformly until the whole documentBounds fits in the view.
   * @constant
   */
  public static Uniform = new EnumValue(Diagram, 'Uniform', 1);

  /**
   * Diagrams with this autoScale type, used as the value of {@link Diagram#autoScale},
   * are scaled until the documentBounds fits in the view in one direction while
   * a scrollbar is still needed in the other direction.
   * @constant
   */
  public static UniformToFill = new EnumValue(Diagram, 'UniformToFill', 2);

  /**
   * This value for {@link Diagram#validCycle} states that
   * there are no restrictions on making cycles of links.
   * @constant
   */
  public static CycleAll = new EnumValue(Diagram, 'CycleAll', 10);

  /**
   * This value for {@link Diagram#validCycle} states that
   * a valid link from a node will not produce a directed cycle in the graph.
   * @constant
   */
  public static CycleNotDirected = new EnumValue(Diagram, 'CycleNotDirected', 11);

  // * This value for {@link Diagram#validCycle} states that
  // * a valid link from a node will not produce a directed cycle in the graph,
  // * assuming there are no directed cycles anywhere accessible from either node.

  /**
   * @hidden @internal
   * Undocumented
   * @constant
   */
  public static CycleNotDirectedFast = new EnumValue(Diagram, 'CycleNotDirectedFast', 12);

  /**
   * This value for {@link Diagram#validCycle} states that
   * a valid link from a node will not produce an undirected cycle in the graph.
   * @constant
   */
  public static CycleNotUndirected = new EnumValue(Diagram, 'CycleNotUndirected', 13);

  /**
   * This value for {@link Diagram#validCycle} states that
   * any number of destination links may go out of a node, but at most one
   * source link may come into a node, and there are no directed cycles.
   *
   * This value assumes that the graph does not already have any cycles in it,
   * or else the behavior may be indeterminate.
   * @constant
   */
  public static CycleDestinationTree = new EnumValue(Diagram, 'CycleDestinationTree', 14);

  /**
   * This value for {@link Diagram#validCycle} states that
   * any number of source links may come into a node, but at most one
   * destination link may go out of a node, and there are no directed cycles.
   *
   * This value assumes that the graph does not already have any cycles in it,
   * or else the behavior may be indeterminate.
   * @constant
   */
  public static CycleSourceTree = new EnumValue(Diagram, 'CycleSourceTree', 15);

  /**
   * This value for {@link Diagram#scrollMode} states that
   * the viewport constrains scrolling to the Diagram document bounds.
   * @constant
   */
  public static DocumentScroll = new EnumValue(Diagram, 'DocumentScroll', 1);

  /**
   * This value for {@link Diagram#scrollMode} states that
   * the viewport does not constrain scrolling to the Diagram document bounds.
   * @constant
   */
  public static InfiniteScroll = new EnumValue(Diagram, 'InfiniteScroll', 2);

  /**
   * @hidden @internal
   * This value for {@link Diagram#treeCollapsePolicy} states that only the {@link Node#findTreeParentNode}'s
   * {@link Node#isTreeExpanded} property determines whether a "child" node is visible.
   * @constant
   */
  public static TreeParentCollapsed = new EnumValue(Diagram, 'TreeParentCollapsed', 21);

  /**
   * @hidden @internal
   * This value for {@link Diagram#treeCollapsePolicy} states that if all of the {@link Node#findNodesInto}
   * or {@link Node#findNodesOutOf}, depending on {@link #isTreePathToChildren} being true or false,
   * are not {@link Node#isTreeExpanded}, the "child" node will be not visible.
   * @constant
   */
  public static AllParentsCollapsed = new EnumValue(Diagram, 'AllParentsCollapsed', 22);

  /**
   * @hidden @internal
   * This value for {@link Diagram#treeCollapsePolicy} states that if any of the {@link Node#findNodesInto}
   * or {@link Node#findNodesOutOf}, depending on {@link #isTreePathToChildren} being true or false,
   * is not {@link Node#isTreeExpanded}, the "child" node will be not visible.
   * @constant
   */
  public static AnyParentsCollapsed = new EnumValue(Diagram, 'AnyParentsCollapsed', 23);

  /**
   * Gets or sets the opacity for all parts in this diagram.
   * The value must be between 0.0 (fully transparent) and 1.0 (no additional transparency).
   * This value is multiplicative with any existing transparency,
   * for instance from a {@link Brush} or image transparency.
   * The default value is 1.
   * @since 2.1
   * @see {@link Layer#opacity}
   * @see {@link GraphObject#opacity}
   */
  /** @export */ get opacity(): number { return this._opacity; }
  /** @export */ set opacity(value: number) {
    const old = this._opacity;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Diagram, 'opacity');
      if (value < 0 || value > 1) Util.throwRangeError(value, '0 <= value <= 1', Diagram, 'opacity');
      this._opacity = value;
      this.raiseChanged('opacity', old, value);
      // and make sure everything is re-drawn
      this.invalidateDraw(); // full redraw needed
    }
  }

  /**
   * Gets or sets what kinds of graphs this diagram allows the user to draw.
   * By default this property is {@link Diagram.CycleAll} -- all kinds of cycles are permitted.
   * Common values include {@link Diagram.CycleDestinationTree} and {@link Diagram.CycleNotDirected}.
   */
  /** @export */ get validCycle(): EnumValue { return this._validCycle; }
  /** @export */ set validCycle(value: EnumValue) {
    const old = this._validCycle;
    if (old !== value) {
      EnumValue.checkEnumValue(value, Diagram, Diagram, 'validCycle');
      this._validCycle = value;
      this.raiseChanged('validCycle', old, value);
    }
  }

  /**
   * This read-only property returns an iterator for this Diagram's Layers.
   * @see {@link #addLayer}
   * @see {@link #removeLayer}
   */
  /** @export */ get layers(): Iterator<Layer> { return this._layers.iterator; }

  // Permissions

  // ??? focusable?

  /**
   * Gets or sets whether the Diagram's {@link Diagram#model} is {@link Model#isReadOnly}.
   * @see {@link #isReadOnly}
   * @see {@link #isEnabled}
   */
  /** @export */ get isModelReadOnly(): boolean {
    const model = this._model;
    if (model === null) return false;
    return model.isReadOnly;
  }
  /** @export */ set isModelReadOnly(value: boolean) {
    const model = this._model;
    if (model !== null) model.isReadOnly = value;
  }

  /**
   * Gets or sets whether the Diagram may be modified by the user,
   * while still allowing the user to scroll, zoom, and select.
   * The initial value is false.
   * @see {@link #isModelReadOnly}
   * @see {@link #isEnabled}
   */
  /** @export */ get isReadOnly(): boolean { return this._isReadOnly; }
  /** @export */ set isReadOnly(value: boolean) {
    const old = this._isReadOnly;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'isReadOnly');
      this._isReadOnly = value;
      this.raiseChanged('isReadOnly', old, value);
    }
  }

  /**
   * Gets or sets whether the user may interact with the Diagram.
   * @see {@link #isReadOnly}
   * @see {@link #isModelReadOnly}
   */
  /** @export */ get isEnabled(): boolean { return this._isEnabled; }
  /** @export */ set isEnabled(value: boolean) {
    const old = this._isEnabled;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'isEnabled');
      this._isEnabled = value;
      this.raiseChanged('isEnabled', old, value);
    }
  }

  /**
   * Gets or sets whether the user may copy to or paste parts from the internal clipboard.
   * This allows use of {@link CommandHandler#cutSelection},
   * {@link CommandHandler#copySelection} and {@link CommandHandler#pasteSelection}.
   * The initial value is true.
   */
  /** @export */ get allowClipboard(): boolean { return this._allowClipboard; }
  /** @export */ set allowClipboard(value: boolean) {
    const old = this._allowClipboard;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowClipboard');
      this._allowClipboard = value;
      this.raiseChanged('allowClipboard', old, value);
    }
  }

  /**
   * Gets or sets whether the user may copy objects.
   * The initial value is true.
   */
  /** @export */ get allowCopy(): boolean { return this._allowCopy; }
  /** @export */ set allowCopy(value: boolean) {
    const old = this._allowCopy;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowCopy');
      this._allowCopy = value;
      this.raiseChanged('allowCopy', old, value);
    }
  }

  /**
   * Gets or sets whether the user may delete objects from the Diagram.
   * The initial value is true.
   */
  /** @export */ get allowDelete(): boolean { return this._allowDelete; }
  /** @export */ set allowDelete(value: boolean) {
    const old = this._allowDelete;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowDelete');
      this._allowDelete = value;
      this.raiseChanged('allowDelete', old, value);
    }
  }

  /**
   * Gets or sets whether the user may start a drag-and-drop in this Diagram,
   * possibly dropping in a different element.
   * The initial value is false.
   */
  /** @export */ get allowDragOut(): boolean { return this._allowDragOut; }
  /** @export */ set allowDragOut(value: boolean) {
    const old = this._allowDragOut;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowDragOut');
      this._allowDragOut = value;
      this.raiseChanged('allowDragOut', old, value);
    }
  }

  /**
   * Gets or sets whether the user may end a drag-and-drop operation in this Diagram.
   * This is typically set to true when a Diagram is used with a {@link Palette}.
   *
   * The initial value is true.
   */
  /** @export */ get allowDrop(): boolean { return this._allowDrop; }
  /** @export */ set allowDrop(value: boolean) {
    const old = this._allowDrop;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowDrop');
      this._allowDrop = value;
      this.raiseChanged('allowDrop', old, value);
    }
  }

  /**
   * Gets or sets whether the user may do in-place text editing.
   * The initial value is true.
   */
  /** @export */ get allowTextEdit(): boolean { return this._allowTextEdit; }
  /** @export */ set allowTextEdit(value: boolean) {
    const old = this._allowTextEdit;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowTextEdit');
      this._allowTextEdit = value;
      this.raiseChanged('allowTextEdit', old, value);
    }
  }

  /**
   * Gets or sets whether the user may group parts together.
   * The initial value is true.
   */
  /** @export */ get allowGroup(): boolean { return this._allowGroup; }
  /** @export */ set allowGroup(value: boolean) {
    const old = this._allowGroup;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowGroup');
      this._allowGroup = value;
      this.raiseChanged('allowGroup', old, value);
    }
  }

  /**
   * Gets or sets whether the user may ungroup existing groups.
   * The initial value is true.
   */
  /** @export */ get allowUngroup(): boolean { return this._allowUngroup; }
  /** @export */ set allowUngroup(value: boolean) {
    const old = this._allowUngroup;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowUngroup');
      this._allowUngroup = value;
      this.raiseChanged('allowUngroup', old, value);
    }
  }

  /**
   * Gets or sets whether the user may add parts to the Diagram.
   * The initial value is true.
   */
  /** @export */ get allowInsert(): boolean { return this._allowInsert; }
  /** @export */ set allowInsert(value: boolean) {
    const old = this._allowInsert;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowInsert');
      this._allowInsert = value;
      this.raiseChanged('allowInsert', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw new links.
   * The initial value is true.
   */
  /** @export */ get allowLink(): boolean { return this._allowLink; }
  /** @export */ set allowLink(value: boolean) {
    const old = this._allowLink;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowLink');
      this._allowLink = value;
      this.raiseChanged('allowLink', old, value);
    }
  }

  /**
   * Gets or sets whether the user may reconnect existing links.
   * The initial value is true.
   */
  /** @export */ get allowRelink(): boolean { return this._allowRelink; }
  /** @export */ set allowRelink(value: boolean) {
    const old = this._allowRelink;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowRelink');
      this._allowRelink = value;
      this.raiseChanged('allowRelink', old, value);
    }
  }

  /**
   * Gets or sets whether the user may move objects.
   * The initial value is true.
   */
  /** @export */ get allowMove(): boolean { return this._allowMove; }
  /** @export */ set allowMove(value: boolean) {
    const old = this._allowMove;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowMove');
      this._allowMove = value;
      this.raiseChanged('allowMove', old, value);
    }
  }

  /**
   * Gets or sets whether the user may reshape parts.
   * The initial value is true.
   */
  /** @export */ get allowReshape(): boolean { return this._allowReshape; }
  /** @export */ set allowReshape(value: boolean) {
    const old = this._allowReshape;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowReshape');
      this._allowReshape = value;
      this.raiseChanged('allowReshape', old, value);
    }
  }

  /**
   * Gets or sets whether the user may resize parts.
   * The initial value is true.
   */
  /** @export */ get allowResize(): boolean { return this._allowResize; }
  /** @export */ set allowResize(value: boolean) {
    const old = this._allowResize;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowResize');
      this._allowResize = value;
      this.raiseChanged('allowResize', old, value);
    }
  }

  /**
   * Gets or sets whether the user may rotate parts.
   * The initial value is true.
   */
  /** @export */ get allowRotate(): boolean { return this._allowRotate; }
  /** @export */ set allowRotate(value: boolean) {
    const old = this._allowRotate;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowRotate');
      this._allowRotate = value;
      this.raiseChanged('allowRotate', old, value);
    }
  }

  /**
   * Gets or sets whether the user may select objects.
   * The initial value is true.
   */
  /** @export */ get allowSelect(): boolean { return this._allowSelect; }
  /** @export */ set allowSelect(value: boolean) {
    const old = this._allowSelect;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowSelect');
      this._allowSelect = value;
      this.raiseChanged('allowSelect', old, value);
    }
  }

  /**
   * Gets or sets whether the user may undo or redo any changes.
   * The initial value is true.
   */
  /** @export */ get allowUndo(): boolean { return this._allowUndo; }
  /** @export */ set allowUndo(value: boolean) {
    const old = this._allowUndo;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowUndo');
      this._allowUndo = value;
      this.raiseChanged('allowUndo', old, value);
    }
  }

  /**
   * Gets or sets whether the user may zoom into or out of the Diagram.
   * The initial value is true.
   */
  /** @export */ get allowZoom(): boolean { return this._allowZoom; }
  /** @export */ set allowZoom(value: boolean) {
    const old = this._allowZoom;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowZoom');
      this._allowZoom = value;
      this.raiseChanged('allowZoom', old, value);
    }
  }

  /**
   * Gets or sets whether the Diagram has a vertical Scrollbar.
   *
   * To enable or disable scrolling itself, use {@link #allowHorizontalScroll} and {@link #allowVerticalScroll}.
   *
   * Adding or removing a scrollbar modifies the diagram's viewport.
   *
   * The initial value is true.
   * @see {@link #allowVerticalScroll}
   * @see {@link #hasHorizontalScrollbar}
   */
  /** @export */ get hasVerticalScrollbar(): boolean { return this._hasVerticalScrollbar; }
  /** @export */ set hasVerticalScrollbar(value: boolean) {
    const old = this._hasVerticalScrollbar;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'hasVerticalScrollbar');
      this._hasVerticalScrollbar = value;
      this.invalidateScroll();
      this.invalidateDraw(); // not strictly needed, since a transaction will do this
      this.raiseChanged('hasVerticalScrollbar', old, value);
      this.onContentAlignmentChanged(false);
    }
  }

  /**
   * Gets or sets whether the Diagram has a horizontal Scrollbar.
   *
   * To enable or disable scrolling itself, use {@link #allowHorizontalScroll} and {@link #allowVerticalScroll}.
   *
   * Adding or removing a scrollbar modifies the diagram's viewport.
   *
   * The initial value is true.
   * @see {@link #allowHorizontalScroll}
   * @see {@link #hasVerticalScrollbar}
   */
  /** @export */ get hasHorizontalScrollbar(): boolean { return this._hasHorizontalScrollbar; }
  /** @export */ set hasHorizontalScrollbar(value: boolean) {
    const old = this._hasHorizontalScrollbar;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'hasHorizontalScrollbar');
      this._hasHorizontalScrollbar = value;
      this.invalidateScroll();
      this.invalidateDraw(); // not strictly needed, since a transaction will do this
      this.raiseChanged('hasHorizontalScrollbar', old, value);
      this.onContentAlignmentChanged(false);
    }
  }

  /**
   * Gets or sets whether the user is allowed to use the horizontal scrollbar.
   * The initial value is true.
   * @see {@link #hasHorizontalScrollbar}
   * @see {@link #allowVerticalScroll}
   */
  /** @export */ get allowHorizontalScroll(): boolean { return this._allowHorizontalScroll; }
  /** @export */ set allowHorizontalScroll(value: boolean) {
    const old = this._allowHorizontalScroll;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowHorizontalScroll');
      this._allowHorizontalScroll = value;
      this.raiseChanged('allowHorizontalScroll', old, value);
      this.onContentAlignmentChanged(false);
    }
  }

  /**
   * Gets or sets whether the user is allowed to use the vertical scrollbar.
   * The initial value is true.
   * @see {@link #hasVerticalScrollbar}
   * @see {@link #allowHorizontalScroll}
   */
  /** @export */ get allowVerticalScroll(): boolean { return this._allowVerticalScroll; }
  /** @export */ set allowVerticalScroll(value: boolean) {
    const old = this._allowVerticalScroll;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'allowVerticalScroll');
      this._allowVerticalScroll = value;
      this.raiseChanged('allowVerticalScroll', old, value);
      this.onContentAlignmentChanged(false);
    }
  }

  /**
   * Gets or sets the distance in screen pixels that the horizontal scrollbar will scroll
   * when scrolling by a line.
   *
   * The default value is 16.
   * @see {@link #scrollVerticalLineChange}
   */
  /** @export */ get scrollHorizontalLineChange(): number { return this._scrollHorizontalLineChange; }
  /** @export */ set scrollHorizontalLineChange(value: number) {
    const old = this._scrollHorizontalLineChange;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Diagram, 'scrollHorizontalLineChange');
      if (value < 0) Util.throwRangeError(value, '>= 0', Diagram, 'scrollHorizontalLineChange');
      this._scrollHorizontalLineChange = value;
      this.raiseChanged('scrollHorizontalLineChange', old, value);
    }
  }

  /**
   * Gets or sets the distance in screen pixels that the vertical scrollbar will scroll
   * when scrolling by a line.
   *
   * The default value is 16.
   * @see {@link #scrollHorizontalLineChange}
   */
  /** @export */ get scrollVerticalLineChange(): number { return this._scrollVerticalLineChange; }
  /** @export */ set scrollVerticalLineChange(value: number) {
    const old = this._scrollVerticalLineChange;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Diagram, 'scrollVerticalLineChange');
      if (value < 0) Util.throwRangeError(value, '>= 0', Diagram, 'scrollVerticalLineChange');
      this._scrollVerticalLineChange = value;
      this.raiseChanged('scrollVerticalLineChange', old, value);
    }
  }

  // Input

  /**
   * Gets or sets the last {@link InputEvent} that occurred.
   *
   * This property is useful in tools and real-time operations for determining
   * where the mouse pointer was most recently located.
   *
   * Setting this property does not notify about any changed event.
   * @see {@link #firstInput}
   */
  /** @export */ get lastInput(): InputEvent { return this._lastInput; }
  /** @export */ set lastInput(value: InputEvent) {
    if (Debug) Util.checkClass(value, InputEvent, Diagram, 'lastInput');
    this._lastInput = value;
  }

  /**
   * Gets or sets the most recent mouse-down {@link InputEvent} that occurred.
   *
   * Setting this property does not notify about any changed event.
   * @see {@link #lastInput}
   */
  /** @export */ get firstInput(): InputEvent { return this._firstInput; }
  /** @export */ set firstInput(value: InputEvent) {
    if (Debug) Util.checkClass(value, InputEvent, Diagram, 'firstInput');
    this._firstInput = value;
  }

  /**
   * Gets or sets the current cursor for the Diagram, overriding the {@link #defaultCursor}.
   *
   * Valid CSS cursors are accepted,
   * such as "auto", "default", "none", "context-menu", "help", "pointer", "progress", "wait", etc.
   *
   * It is possible to use custom cursors with the syntax "url(path_to_image), default".
   * A fallback (like default here) is necessary for a custom cursor to work.
   *
   * To read more about cursor syntax, go to:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/cursor">CSS cursors (mozilla.org)</a>.
   *
   * If the specified cursor is not accepted by the platform, GoJS will append
   * `-webkit-` or `-moz-` prefixes.
   *
   * Setting this property does not notify about any changed event.
   * Setting this value to the empty string ('') returns the Diagram's cursor to the {@link #defaultCursor}.
   * @see {@link #defaultCursor}
   * @see {@link GraphObject#cursor}
   */
  /** @export */ get currentCursor(): string { return this._currentCursor; }
  /** @export */ set currentCursor(value: string) {
    if (value === '') value = this._defaultCursor;
    const old = this._currentCursor;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', Diagram, 'currentCursor');
      const canvas = this._canvas;
      const div = this._div;
      if (canvas === null) return;
      this._currentCursor = value;
      const canvasOld = canvas.style.cursor;
      // set element's style.cursor
      canvas.style.cursor = value;
      div.style.cursor = value;
      // if the cursor wasn't set, check for a webkit fallback
      if (canvas.style.cursor === canvasOld) {
        canvas.style.cursor = '-webkit-' + value;
        div.style.cursor = '-webkit-' + value;
        // then for moz fallback
        if (canvas.style.cursor === canvasOld) {
          canvas.style.cursor = '-moz-' + value;
          div.style.cursor = '-moz-' + value;
          // then go back to what was set if neither were found
          if (canvas.style.cursor === canvasOld) {
            canvas.style.cursor = value;
            div.style.cursor = value;
          }
        }
      }
    }
  }

  /**
   * Gets or sets the cursor to be used for the Diagram
   * when no {@link GraphObject} specifies a different cursor.
   *
   * Valid CSS cursors are accepted,
   * such as "auto", "default", "none", "context-menu", "help", "pointer", "progress", "wait", etc.
   *
   * It is possible to use custom cursors with the syntax "url(path_to_image), default".
   * A fallback (like default here) is necessary for a custom cursor to work.
   *
   * To read more about cursor syntax, go to:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/cursor">CSS cursors (mozilla.org)</a>.
   * The default value is "auto".
   * @see {@link #currentCursor}
   * @see {@link GraphObject#cursor}
   */
  /** @export */ get defaultCursor(): string { return this._defaultCursor; }
  /** @export */ set defaultCursor(value: string) {
    if (value === '') value = 'auto';
    const old = this._defaultCursor;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', Diagram, 'defaultCursor');
      this._defaultCursor = value;
      this.raiseChanged('defaultCursor', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user single-primary-clicks
   * on the background of the Diagram.
   * This typically involves a mouse-down followed by a prompt mouse-up
   * at approximately the same position using the left (primary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on no object.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name "BackgroundSingleClicked".
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #doubleClick}
   * @see {@link #contextClick}
   * @see {@link GraphObject#click}
   */
  /** @export */ get click(): ((e: InputEvent) => void) | null { return this._click; }
  /** @export */ set click(value: ((e: InputEvent) => void) | null) {
    const old = this._click;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'click');
      this._click = value;
      this.raiseChanged('click', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user double-primary-clicks
   * on the background of the Diagram.
   * This typically involves a mouse-down/up/down/up in rapid succession
   * at approximately the same position using the left (primary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on no object.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name "BackgroundDoubleClicked".
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #click}
   * @see {@link #contextClick}
   * @see {@link GraphObject#doubleClick}
   */
  /** @export */ get doubleClick(): ((e: InputEvent) => void) | null { return this._doubleClick; }
  /** @export */ set doubleClick(value: ((e: InputEvent) => void) | null) {
    const old = this._doubleClick;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'doubleClick');
      this._doubleClick = value;
      this.raiseChanged('doubleClick', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user single-secondary-clicks
   * on the background of the Diagram.
   * This typically involves a mouse-down followed by a prompt mouse-up
   * at approximately the same position using the right (secondary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on no object.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name "BackgroundContextClicked".
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #click}
   * @see {@link #doubleClick}
   * @see {@link GraphObject#contextClick}
   */
  /** @export */ get contextClick(): ((e: InputEvent) => void) | null { return this._contextClick; }
  /** @export */ set contextClick(value: ((e: InputEvent) => void) | null) {
    const old = this._contextClick;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'contextClick');
      this._contextClick = value;
      this.raiseChanged('contextClick', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse in
   * the background of the Diagram without holding down any buttons,
   * not over any {@link GraphObject}s.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * After calling this function the diagram will be updated immediately.
   * @see {@link #mouseHover}
   * @see {@link GraphObject#mouseOver}
   */
  /** @export */ get mouseOver(): ((e: InputEvent) => void) | null { return this._mouseOver; }
  /** @export */ set mouseOver(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseOver;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseOver');
      this._mouseOver = value;
      this.raiseChanged('mouseOver', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user holds the mouse stationary in
   * the background of the Diagram without holding down any buttons,
   * not over any {@link GraphObject}s.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #mouseOver}
   * @see {@link GraphObject#mouseHover}
   * @see {@link ToolManager#doMouseHover}
   */
  /** @export */ get mouseHover(): ((e: InputEvent) => void) | null { return this._mouseHover; }
  /** @export */ set mouseHover(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseHover;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseHover');
      this._mouseHover = value;
      this.raiseChanged('mouseHover', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user holds the mouse stationary in
   * the background of the Diagram while holding down a button,
   * not over any {@link GraphObject}s.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link GraphObject#mouseHold}
   * @seev ToolManager#doMouseHover
   */
  /** @export */ get mouseHold(): ((e: InputEvent) => void) | null { return this._mouseHold; }
  /** @export */ set mouseHold(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseHold;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseHold');
      this._mouseHold = value;
      this.raiseChanged('mouseHold', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user is dragging the selection in
   * the background of the Diagram during a {@link DraggingTool} drag-and-drop,
   * not over any {@link GraphObject}s.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * It is called within the transaction performed by the {@link DraggingTool}.
   * By default this property is null.
   *
   * Note that for a drag-and-drop that originates in a different diagram, the target diagram's
   * selection collection will not be the parts that are being dragged.
   * Instead the temporary parts being dragged can be found as the source diagram's {@link DraggingTool#copiedParts}.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * After calling this function the diagram will be updated immediately.
   *
   * For example, if you want to prevent the user from dropping Parts into the background of the diagram,
   * and want to provide feedback about that during a drag:
   * ```js
   *   myDiagram.mouseDragOver = function(e) {
   *     myDiagram.currentCursor = "no-drop";
   *   }
   * ```
   * @see {@link #mouseDrop}
   * @see {@link GraphObject#mouseDragEnter}
   * @see {@link GraphObject#mouseDragLeave}
   */
  /** @export */ get mouseDragOver(): ((e: InputEvent) => void) | null { return this._mouseDragOver; }
  /** @export */ set mouseDragOver(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseDragOver;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseDragOver');
      this._mouseDragOver = value;
      this.raiseChanged('mouseDragOver', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user drops the selection in
   * the background of the Diagram at the end of a {@link DraggingTool} drag-and-drop,
   * not onto any {@link GraphObject}s.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * It is called within the transaction performed by the {@link DraggingTool}.
   * By default this property is null.
   *
   * For example, if you want to prevent the user from dropping Parts into the background of the diagram:
   * ```js
   *   myDiagram.mouseDrop = function(e) {
   *     myDiagram.currentTool.doCancel();
   *   }
   * ```
   * @see {@link #mouseDragOver}
   * @see {@link GraphObject#mouseDrop}
   */
  /** @export */ get mouseDrop(): ((e: InputEvent) => void) | null { return this._mouseDrop; }
  /** @export */ set mouseDrop(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseDrop;
    if (old !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseDrop');
      this._mouseDrop = value;
      this.raiseChanged('mouseDrop', old, value);
    }
  }

  /**
   * Gets or sets whether drag-and-drop events may be bubbled up to the diagram if not handled by a part.
   * The default value is false -- each Node or Link that in the diagram needs to define
   * its own {@link GraphObject#mouseDragEnter}, {@link GraphObject#mouseDragLeave}, and {@link GraphObject#mouseDrop}
   * event handlers if you want dragging/dropping on a part to act as if the user were acting on the diagram.
   *
   * If set to true, this will call {@link Diagram#mouseDragOver} during a drag, even while dragging over top-level parts,
   * and {@link Diagram#mouseDrop} will be called even when dropping onto parts.
   *
   * This property will have no impact while dropping on a Group.
   * The Group's mouseDrop and {@link Group#handlesDragDropForMembers} should be set if desired.
   *
   * GraphObjects do not have a mouseDragOver property, so if this is set to true, the Diagram's mouseDragOver will
   * always be called, even when dragging over a part.
   * @see {@link #mouseDragOver}
   * @see {@link #mouseDrop}
   * @see {@link Group#handlesDragDropForMembers}
   * @see {@link GraphObject#mouseDrop}
   * @since 2.0
   */
  /** @export */ get handlesDragDropForTopLevelParts(): boolean { return this._handlesDragDropForTopLevelParts; }
  /** @export */ set handlesDragDropForTopLevelParts(value: boolean) {
    const old = this._handlesDragDropForTopLevelParts;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'handlesDragDropForTopLevelParts');
      this._handlesDragDropForTopLevelParts = value;
      this.raiseChanged('handlesDragDropForTopLevelParts', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the mouse enters the Diagram.
   * (When the browser's <a href="https://developer.mozilla.org/en-US/docs/Web/Events/mouseenter">mouseEnter</a> event fires on the Diagram canvas.)
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #mouseLeave}
   * @see {@link GraphObject#mouseEnter}
   * @since 2.0
   */
  /** @export */ get mouseEnter(): ((e: InputEvent) => void) | null { return this._mouseEnter; }
  /** @export */ set mouseEnter(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseEnter;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseEnter');
      this._mouseEnter = value;
      this.raiseChanged('mouseEnter', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the mouse leaves the Diagram.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link #startTransaction} and {@link #commitTransaction}.
   * @see {@link #mouseEnter}
   * @see {@link GraphObject#mouseLeave}
   * @since 2.0
   */
  /** @export */ get mouseLeave(): ((e: InputEvent) => void) | null { return this._mouseLeave; }
  /** @export */ set mouseLeave(value: ((e: InputEvent) => void) | null) {
    const old = this._mouseLeave;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'mouseLeave');
      this._mouseLeave = value;
      this.raiseChanged('mouseLeave', old, value);
    }
  }

  /**
   * This {@link Adornment} or {@link HTMLInfo} is shown when the mouse stays motionless in the background.
   * The default value is null, which means no tooltip is shown.
   *
   * Here is a simple example:
   * ```js
   *  diagram.toolTip =
   *    $(go.Adornment, "Auto",
   *      $(go.Shape, { fill: "#CCFFCC" }),
   *      $(go.TextBlock, { margin: 4 },
   *        "This diagram lets you control the world.")
   *    );
   * ```
   * @see {@link GraphObject#toolTip}
   * @see {@link ToolManager#doToolTip}
   */
  /** @export */ get toolTip(): Adornment | HTMLInfo | null { return this._toolTip; }
  /** @export */ set toolTip(value: Adornment | HTMLInfo | null) {
    const old = this._toolTip;
    if (old !== value) {
      if (Debug && value !== null && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
        Util.throwError('Diagram.toolTip must be an Adornment or HTMLInfo.');
      }
      this._toolTip = value;
      this.raiseChanged('toolTip', old, value);
    }
  }

  /**
   * This {@link Adornment} or {@link HTMLInfo} is shown when the use context clicks in the background.
   * The default value is null, which means no context menu is shown.
   * On touch devices, a special default context menu will appear even there is no context menu defined.
   * See {@link ContextMenuTool#defaultTouchContextMenu} for details.
   *
   * ```js
   *  diagram.contextMenu =
   *    $("ContextMenu",
   *      $("ContextMenuButton",
   *        $(go.TextBlock, "Undo"),
   *        { click: function(e, obj) { e.diagram.commandHandler.undo(); } },
   *        new go.Binding("visible", "", function(o) {
   *            return o.diagram.commandHandler.canUndo();
   *          }).ofObject()),
   *      $("ContextMenuButton",
   *        $(go.TextBlock, "Redo"),
   *        { click: function(e, obj) { e.diagram.commandHandler.redo(); } },
   *        new go.Binding("visible", "", function(o) {
   *            return o.diagram.commandHandler.canRedo();
   *          }).ofObject())
   *    );
   * ```
   * @see {@link GraphObject#contextMenu}
   * @see {@link ContextMenuTool}
   */
  /** @export */ get contextMenu(): Adornment | HTMLInfo | null { return this._contextMenu; }
  /** @export */ set contextMenu(value: Adornment | HTMLInfo | null) {
    const old = this._contextMenu;
    if (old !== value) {
      if (Debug && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
        Util.throwError('Diagram.contextMenu must be an Adornment or HTMLInfo.');
      }
      this._contextMenu = value;
      this.raiseChanged('contextMenu', old, value);
    }
  }

  /**
   * Gets or sets the {@link CommandHandler} for this Diagram.
   *
   * This is set to a new instance of {@link CommandHandler} on Diagram instantiation.
   *
   * Setting this property does not notify about any changed event.
   * The value cannot be null and must not be shared with other Diagrams.
   */
  /** @export */ get commandHandler(): ICommandHandler { return this._commandHandler; }
  /** @export */ set commandHandler(value: ICommandHandler) {
    const old = this._commandHandler;
    if (old !== value) {
      // Util.checkClass(value, CommandHandler, Diagram, 'commandHandler');
      this._commandHandler = value;
      value.setDiagram(this);
    }
  }

  /**
   * Gets or sets the {@link ToolManager} for this Diagram.
   * This tool is used for mode-less operation.
   * It is responsible for choosing a particular tool to run as the {@link #currentTool}.
   *
   * This tool is normally also the {@link #defaultTool}.
   * If you don't want the ToolManager to run at all,
   * replace the {@link #defaultTool} with your own tool.
   *
   * Setting this property does not notify about any changed event.
   * The value cannot be null and must not be shared with other Diagrams.
   * If you set this property, you will probably also want to set {@link #defaultTool}.
   * @see {@link #defaultTool}
   */
  /** @export */ get toolManager(): ToolManager { return this._toolManager; }
  /** @export */ set toolManager(value: ToolManager) {
    const old = this._toolManager;
    if (old !== value) {
      Util.checkClass(value, ToolManager, Diagram, 'toolManager');
      this._toolManager = value;
      value.diagram = this;
    }
  }

  /**
   * Gets or sets the default tool for this Diagram that becomes the current tool when the current tool stops.
   * Initially this value is the same tool as {@link #toolManager}, which is an instance of {@link ToolManager}.
   *
   * Setting this property also sets the {@link #currentTool} if the old default tool is the currently running tool.
   *
   * Setting this property does not notify about any changed event.
   * The value cannot be null and must not be shared with other Diagrams.
   * @see {@link #currentTool}
   * @see {@link #toolManager}
   */
  /** @export */ get defaultTool(): Tool { return this._defaultTool; }
  /** @export */ set defaultTool(value: Tool) {
    const old = this._defaultTool;
    if (old !== value) {
      Util.checkClass(value, Tool, Diagram, 'defaultTool');
      this._defaultTool = value;
      value.diagram = this;
      if (this.currentTool === old) {  // maybe replace the currentTool too
        this.currentTool = value;
      }
    }
  }

  /**
   * Gets or sets the current tool for this Diagram that handles all input events.
   * This value is frequently replaced by the {@link #toolManager} as different tools run.
   *
   * Each Diagram has a number of tools that define its behavior when responding to mouse events.
   * These include {@link ClickSelectingTool}, {@link DraggingTool}, {@link DragSelectingTool}, {@link LinkingTool}, and {@link ResizingTool}, among others.
   *
   * Initially this is set to the value of {@link #defaultTool}.
   * Setting this to a null value is treated as if it were set to the {@link #defaultTool},
   * because there should always be a currently running tool, except when the diagram is being initialized.
   *
   * A {@link ToolManager} is the default tool used by a Diagram - it chooses to run one of the other tools
   * depending on the circumstances.
   *
   * Setting this property to a new tool stops the previous current tool
   *
   * Setting this property does not notify about any changed event.
   * @see {@link #defaultTool}
   * @see {@link #toolManager}
   */
  /** @export */ get currentTool(): Tool { return this._currentTool; }
  /** @export */ set currentTool(value: Tool) {
    const old = this._currentTool;
    if (old !== null) {
      if (old.isActive) old.doDeactivate();
      old.cancelWaitAfter();
      old.doStop();
    }
    // assume null value for new tool means use the defaultTool instead
    if (value === null) value = this.defaultTool;
    if (value !== null) {
      Util.checkClass(value, Tool, Diagram, 'currentTool');
      this._currentTool = value;
      value.diagram = this;
      value.doStart();
      // but not doActivate, since that might depend on conditions
    }
  }

  // Selection

  /**
   * This read-only property returns the read-only collection of selected objects.
   *
   * Do not modify this collection.
   * If you want to select or deselect a particular object in a Diagram,
   * set the {@link Part#isSelected} property.
   * If you want to deselect all objects, call {@link #clearSelection}.
   * If you want to deselect all objects and select a single object, call {@link #select}.
   *
   * You can limit how many objects the user can select by setting {@link #maxSelectionCount}.
   */
  /** @export */ get selection(): Set<Part> { return this._selection; }

  /**
   * Gets or sets the maximum number of selected objects.
   * The default value is a large positive integer.
   * Values must be non-negative.
   * Decreasing this value may cause objects to be removed from {@link #selection}
   * in order to meet the new lower limit.
   */
  /** @export */ get maxSelectionCount(): number { return this._maxSelectionCount; }
  /** @export */ set maxSelectionCount(value: number) {
    const old = this._maxSelectionCount;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Diagram, 'maxSelectionCount');
      if (value >= 0 && !isNaN(value)) {
        this._maxSelectionCount = value;
        this.raiseChanged('maxSelectionCount', old, value);
        if (!this.undoManager.isUndoingRedoing) {
          const excess = this.selection.count - value;
          if (excess > 0) {
            this.raiseDiagramEvent('ChangingSelection', this.selection);
            const all = this.selection.toArray();
            for (let i = 0; i < excess; i++) {
              all[i].isSelected = false;
            }
            this.raiseDiagramEvent('ChangedSelection', this.selection);
          }
        }
      } else {
        Util.throwRangeError(value, '>= 0', Diagram, 'maxSelectionCount');
      }
    }
  }

  /**
   * Gets or sets the default selection {@link Adornment} template, used to adorn selected Parts other than Groups or Links.
   *
   * Each {@link Node} or simple {@link Part} can have its own {@link Part#selectionAdornmentTemplate},
   * which if non-null will take precedence over this Diagram property.
   *
   * This Adornment must not be in the visual tree of any Diagram.
   */
  /** @export */ get nodeSelectionAdornmentTemplate(): Adornment { return this._nodeSelectionAdornmentTemplate; }
  /** @export */ set nodeSelectionAdornmentTemplate(value: Adornment) {
    const old = this._nodeSelectionAdornmentTemplate;
    if (old !== value) {
      Util.checkClass(value, Adornment, Diagram, 'nodeSelectionAdornmentTemplate');
      this._nodeSelectionAdornmentTemplate = value;
      this.raiseChanged('nodeSelectionAdornmentTemplate', old, value);
    }
  }

  /**
   * Gets or sets the default selection {@link Adornment} template, used to adorn selected Groups.
   *
   * Each {@link Group} can have its own {@link Part#selectionAdornmentTemplate}, which if non-null will take precedence over this Diagram property.
   *
   * This Adornment must not be in the visual tree of any Diagram.
   */
  /** @export */ get groupSelectionAdornmentTemplate(): Adornment { return this._groupSelectionAdornmentTemplate; }
  /** @export */ set groupSelectionAdornmentTemplate(value: Adornment) {
    const old = this._groupSelectionAdornmentTemplate;
    if (old !== value) {
      Util.checkClass(value, Adornment, Diagram, 'groupSelectionAdornmentTemplate');
      this._groupSelectionAdornmentTemplate = value;
      this.raiseChanged('groupSelectionAdornmentTemplate', old, value);
    }
  }

  /**
   * Gets or sets the default selection {@link Adornment} template, used to adorn selected Links.
   *
   * Each {@link Link} can have its own {@link Part#selectionAdornmentTemplate}, which if non-null will take precedence over this Diagram property.
   *
   * This Adornment must not be in the visual tree of any Diagram.
   */
  /** @export */ get linkSelectionAdornmentTemplate(): Adornment { return this._linkSelectionAdornmentTemplate; }
  /** @export */ set linkSelectionAdornmentTemplate(value: Adornment) {
    const old = this._linkSelectionAdornmentTemplate;
    if (old !== value) {
      Util.checkClass(value, Adornment, Diagram, 'linkSelectionAdornmentTemplate');
      this._linkSelectionAdornmentTemplate = value;
      this.raiseChanged('linkSelectionAdornmentTemplate', old, value);
    }
  }

  /**
   * This read-only property returns the read-only collection of highlighted parts.
   *
   * Do not modify this collection.
   * If you want to highlight or remove the highlight for a particular Part in a Diagram,
   * set the {@link Part#isHighlighted} property.
   * If you want to remove all highlights, call {@link #clearHighlighteds}.
   * If you want to removal all highlights and highlight a single object, call {@link #highlight}.
   */
  /** @export */ get highlighteds(): Set<Part> { return this._highlighteds; }

  /**
   * Gets or sets whether this Diagram's state has been modified.
   * Setting this property does not notify about any changed event,
   * but it does raise the "Modified" {@link DiagramEvent},
   * although perhaps not immediately.
   *
   * Returns true if the Diagram has been changed,
   * if the {@link #undoManager} has recorded any changes, or
   * if an undo has been performed without a corresponding redo.
   *
   * Replacing the {@link #model} automatically sets this property to false after the initial layout has completed.
   * The "Modified" DiagramEvent is also raised when an undo or a redo has finished.
   * A "Modified" DiagramEvent listener must not modify this Diagram or its Model.
   */
  /** @export */ get isModified(): boolean {
    const m = this.undoManager;
    if (!m.isEnabled) return this._isModified;
    if (m.currentTransaction !== null) return true;
    return this._isModified && this._historyIndex !== m.historyIndex;
  }
  /** @export */ set isModified(value: boolean) {
    const old = this._isModified;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'isModified');
      this._isModified = value;
      const m = this.undoManager;
      if (!value && m.isEnabled) {  // remember current UndoManager.historyIndex
        this._historyIndex = m.historyIndex;
      }
      if (!value) this.onModifiedChanged();  // if changing to true, wait until end of transaction
    }
  }

  /**
   * @hidden @internal
   * Called by isModified setter and doModelChanged
   * @this {Diagram}
   */
  private onModifiedChanged(): void {
    const mod = this.isModified;  // not _isModified !
    if (this._wasModified !== mod) {
      this._wasModified = mod;
      this.raiseDiagramEvent('Modified');
    }
  }

  /**
   * Gets or sets the {@link Model} holding data corresponding to the
   * data-bound nodes and links of this Diagram.
   *
   * Replacing this value causes all of the bound Nodes and Links to be deleted and re-created
   * from the new model data.
   *
   * Models may be shared by multiple Diagrams. One common approach is to have two
   * Diagrams displaying the same Model but using different templates
   * (see {@link #nodeTemplate}, {@link #nodeTemplateMap}, and the associated link and group properties) and sometimes even different Layouts.
   *
   * Setting this property does not notify about any changed event; the new value must not be null.
   * Typically a new Model will have its own {@link UndoManager}, thereby replacing the Diagram's
   * current UndoManager.
   *
   * Replacing or re-setting the model will re-initialize the Diagram, taking in to account
   * {@link #initialPosition}, {@link #initialScale}, {@link #initialAutoScale}, and {@link #initialContentAlignment}.
   * It will also set {@link #isModified} to false.
   *
   * The default behavior when replacing the model is to copy a few {@link UndoManager} properties to the
   * new UndoManager, including {@link UndoManager#isEnabled} and {@link UndoManager#maxHistoryLength}.
   *
   * It is an error to replace the Diagram.model while a transaction is in progress.
   */
  /** @export */ get model(): Model { return this._model; }
  /** @export */ set model(value: Model) {
    const old = this._model;
    if (old !== value) {
      Util.checkClass(value, Model, Diagram, 'model');  // null is not allowed

      this.currentTool.doCancel();
      if (old !== null && old.undoManager !== value.undoManager && old.undoManager.isInTransaction) {
        Util.throwError('Do not replace a Diagram.model while a transaction is in progress.');
      }


      this.animationManager.stopAnimation(true); // Stop all animations
      this.clearAllParts(true); // remove all modeled Parts from this Diagram

      // reset some internal state
      this._isInitialLayoutCompleted = false;
      this._firstUpdate = true;
      this._historyIndex = -2;  // reset to a value that cannot match any valid history index
      this._updateQueued = false;
      // need to make layouts invalid without updating
      const oldUpdating = this._isUpdating;
      this._isUpdating = true;

      // Only necessary when there won't be an animation due to layout
      this.animationManager.prepareAutomaticAnimation('Model');

      if (old !== null) {
        if (this._modelChangedListeners !== null) {  // remove any ModelChanged listeners, to avoid GC retention
          this._modelChangedListeners.each(function(listen) { old.removeChangedListener(listen); });
        }
        old.removeChangedListener(this._modelChangedHandler);
        // no longer needed due to clearAllParts():
        // this.partManager.removeAllModeledParts();
      }
      this._model = value;
      this.partManager = this.lookupPartManager(this._model.constructor.type);
      // add partial listener first, in case of need to update data bindings while creating Parts
      value.addChangedListener(this._modelBindingHandler);
      this.partManager.addAllModeledParts();
      // remove temporary partial listener
      value.removeChangedListener(this._modelBindingHandler);
      // add real listener afterwards, to avoid unnecessary churn during the construction of Parts
      value.addChangedListener(this._modelChangedHandler);
      if (this._modelChangedListeners !== null) {  // restore any ModelChanged listeners
        this._modelChangedListeners.each(function(listen) { value.addChangedListener(listen); });
      }
      this._isUpdating = oldUpdating;
      if (!this._isAligning) this.invalidateDraw(); // calls requestUpdate

      // if the old UndoManager was enabled, the new one is too
      if (old !== null) {
        value.undoManager.copyProperties(old.undoManager);
      }
    }
  }

  /**
   * @hidden @internal
   * Add a PartManager to the map such that a model can use the proper PartManager when
   * performing various operations on parts.
   * @param {string} key a string representing the type of model, should match {@link Model#type}
   * @param {function} pm the type of PartManager that will be created when looking up from the map
   * @since 2.0
   */
  public static addPartManager(key: string, pm: typeof PartManager): void {
    Diagram._partManagerMap.add(key, pm);
  }

  /**
   * Gets or sets the license key.
   * @since 2.0
   */
/** @export */   public static get licenseKey(): string {
    return Diagram._licenseKey.last();
  }
/** @export */   public /** @export */ static set licenseKey(value: string) {
    Diagram._licenseKey.add(value);
  }

  /**
   * Gets the current GoJS version.
   * @since 2.0
   */
/** @export */   public static get version(): string {
    return Diagram._version;
  }
  /**
   * @hidden @internal
   */
  public static _licenseKey: List<string> = new List<string>();
  /**
   * @hidden @internal
   */
  public static _version: string = '2.1.0-alpha4';

  /**
   * @hidden @internal
   * Create a new instance of the PartManager from the map that corresponds with the passed model type.
   * @this {Diagram}
   * @param {string} key a string representing the type of model, should match {@link Model#type}
   * @return {PartManager} an instance of the PartManager for the given {@link Model#type}
   */
  private lookupPartManager(key: string): PartManager {
    const pm = Diagram._partManagerMap.get(key);
    if (pm !== null) return new pm();
    return new PartManager();
  }

  /** @hidden @internal */
  /* noexport */get isUpdatingModel(): boolean { return this._isUpdatingModel; }
  set isUpdatingModel(value: boolean) { this._isUpdatingModel = value; }

  /** @hidden @internal */
  /* noexport */get loadedRouteLinks(): Set<Part> { return this._loadedRouteLinks; }

  /**
   * @hidden @internal
   * This function is called by PartManager when the Diagram's {@link #model} has changed.
   * @this {Diagram}
   * @param {ChangedEvent} e
   */
  public doModelChanged(e: ChangedEvent): void {
    if (e.model !== this.model) return;

    const chg = e.change;
    const xname = e.propertyName as string;
    if (chg === ChangedEvent.Transaction && xname[0] === 'S') {
      if (xname === 'StartingFirstTransaction') {
        // make sure all mouse tools are for this Diagram
        const dia = this;
        const mgr = this.toolManager;
        mgr.mouseDownTools.each(function(t) { t.diagram = dia; });
        mgr.mouseMoveTools.each(function(t) { t.diagram = dia; });
        mgr.mouseUpTools.each(function(t) { t.diagram = dia; });
        if (!this._isUpdating && !this._isInitialLayoutCompleted) {
          this._doNonRealtimeLayouts = true;  // perform any automatic layouts
          if (this._firstUpdate) this._updateQueued = true;
        }
      } else if (xname === 'StartingUndo' || xname === 'StartingRedo') {
        // Undo/redo started, stop any running animation:
        const AM = this.animationManager;
        if (AM.defaultAnimation.isAnimating && !this.skipsUndoManager) {
          AM.stopAnimation();
        }
        this.raiseDiagramEvent('ChangingSelection', this.selection);
      } else if (xname === 'StartedTransaction') {
        // New transaction started, stop any running animation:
        const AM = this.animationManager;
        if (AM.defaultAnimation.isAnimating && !this.skipsUndoManager) {
          AM.stopAnimation();
        }

      }
      return;
    }

    if (!this.isUpdatingModel) return;
    this.isUpdatingModel = false;  // now Model is updating Diagram, not vice-versa
    try {
      if (e.modelChange === '' && chg === ChangedEvent.Transaction) {
        // 'Start...' transaction events handled above

        if (xname === 'FinishedUndo' || xname === 'FinishedRedo') {
          // At the end of an undo/redo we want to validate all parts but not have any side effects occur,
          // such as parts invalidating layouts. In the future we will do away with this in favor of having
          // nothing invalid (therefore nothing to validate) by the time Finished Undo/Redo occur.
          this.raiseDiagramEvent('ChangedSelection', this.selection);
          this.validateParts();
        }

        const AM = this.animationManager;
        if (xname === 'RolledBackTransaction') { // cancel any animation pending
          AM.stopAnimation();
        }

        this._doNonRealtimeLayouts = true;  // perform any automatic layouts
        this.maybeUpdate();
        // Animate if no transaction or top-level
        if (this.undoManager.transactionLevel === 0 || this.undoManager.transactionLevel === 1) {
          AM.beginAnimation();
        }

        if (xname === 'CommittedTransaction' && this.undoManager._justDiscarded) {
          // limit index to a valid value within the new history
          this._historyIndex = Math.min(this._historyIndex, this.undoManager.historyIndex - 1);
        }
        const finishedTransaction = e.isTransactionFinished;
        if (finishedTransaction) {
          this.onModifiedChanged();
          this.loadedRouteLinks.clear();
          this.animationManager.clearSuspendedAnimations();
        }
        // Parts may have moved or been removed, so we need to give tooltips and other objects
        // that are sensitive to a mouse-over a chance to update themselves.
        // This must be done after the transaction is all done, as if a separate mouse event.
        // Should this only be done when the real mouse position is over this Diagram?
        // (Tool.standardMouseOver uses lastInput, which may contain out-of-date event info.)
        if (!this._mouseOverQueued && finishedTransaction) {
          this._mouseOverQueued = true;
          const diag = this;
          Util.setTimeout(function() {
            diag.currentTool.standardMouseOver();
            diag._mouseOverQueued = false;
          }, 10);
        }
        // call Diagram.checkProperties at each transaction, to look for unexpectedly set property names
        // var msg = this.checkProperties(); if (msg !== '') Util.trace(msg);  // for debugging only!

      } // end of ChangedEvent.Transaction
    } finally {
      this.isUpdatingModel = true;
    }
  }

  /**
   * @hidden @internal
   */
  /** @export */ get skipsModelSourceBindings(): boolean { return this._skipsModelSourceBindings; }
  /** @export */ set skipsModelSourceBindings(value: boolean) { this._skipsModelSourceBindings = value; }

  /**
   * @hidden @internal
   * Used in the DraggingTool and AnimationManager to stop real-time updates of the bounds.
   */
  /* noexport */get skipsBoundsUpdates(): boolean { return this._skipsBoundsUpdates; }
  set skipsBoundsUpdates(value: boolean) { this._skipsBoundsUpdates = value; }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Panel} pan
   */
  public rememberPictures(pan: Panel): void {
    const elems = pan._elements._arr;
    const l = elems.length;
    for (let i = 0; i < l; i++) {
      this.rememberPicture(elems[i]);
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {GraphObject} p
   */
  public rememberPicture(p: GraphObject): void {
    if (!(p instanceof Picture)) return;
    const elem = p.element;
    if (elem === null) return;
    if (!(elem instanceof HTMLImageElement)) return; // could be a HTMLCanvasElement or HTMLVideoElement
    p.checkLoadStatus();
    const url = elem.getAttribute('src');
    const pictures = this._picturesMap.getValue(url);
    if (pictures === null) {
      // if the array is unknown, remember the array and the one Picture
      const pics: Array<Picture> = [];
      pics.push(p);
      this._picturesMap.add(url, pics);
    } else {
      // is the Picture already known for the array, don't need to do anything
      for (let i = 0; i < pictures.length; i++) {
        if (pictures[i] === p) return;
      }
      pictures.push(p); // otherwise add to the collection
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Panel} pan
   */
  public forgetPictures(pan: Panel): void {
    const elems = pan._elements._arr;
    const l = elems.length;
    for (let i = 0; i < l; i++) {
      this.forgetPicture(elems[i]);
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {GraphObject} p
   */
  public forgetPicture(p: GraphObject): void {
    if (!(p instanceof Picture)) return;
    const elem = p.element;
    if (elem === null) return;
    if (!(elem instanceof HTMLImageElement)) return; // could be a HTMLCanvasElement or HTMLVideoElement
    const url = elem.getAttribute('src');

    const pictures = this._picturesMap.getValue(url);
    if (pictures !== null) {
      for (let i = 0; i < pictures.length; i++) {
        if (pictures[i] === p) {
          pictures.splice(i, 1);
          if (pictures.length === 0) {
            this._picturesMap.remove(url);
            Picture.clearCache(url);
          }
          return;
        }
      }
    }
  }

  /**
   * Remove all of the {@link Part}s created from model data
   * and then create them again.
   * This must be called after modifying or replacing any of the template maps
   * such as {@link #nodeTemplateMap}.
   * This re-selects all of the new Parts that were created from data of the original selected Parts.
   *
   * If you modify a template Map, there is no notification that the map has changed.
   * You will need to call rebuildParts explicitly.
   * If you are replacing the {@link #nodeTemplate} or the {@link #nodeTemplateMap} or
   * the corresponding properties for Groups or Links,
   * the Diagram property setters will automatically call rebuildParts.
   *
   * It is extremely wasteful to call this method after making some model data changes that you want to
   * be reflected in the diagram.  Instead, it is better call {@link Model#setDataProperty},
   * {@link Model#addNodeData}, {@link Model#removeNodeData}, or other model methods.
   * Not only do those methods update efficiently, they also preserve unbound state and support undo/redo.
   * @export
   * @this {Diagram}
   */
  public rebuildParts(): void {
    this.partManager.rebuildParts();
  }

  /**
   * @hidden @internal
   * Called only by Node's collapseTree and Group's collapseSubGraph
   * @this {Diagram}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addCollapsedNode(node: Node, rootNode: Node): void {
    this._animationManager.addCollapsedNode(node, rootNode);
  }

  /**
   * @hidden @internal
   * Called only by Node's expandTree and Group's expandSubGraph
   * @this {Diagram}
   * @param {Node} node
   * @param {Node} rootNode
   */
  public addExpandedNode(node: Node, rootNode: Node): void {
    this._animationManager.addExpandedNode(node, rootNode);
  }

  /**
   * Look for a {@link Part} or {@link Node} or {@link Group} corresponding to a model's data object's unique key.
   * This will find a {@link Link} if the model is a {@link GraphLinksModel} that is maintaining a key on the link data objects.
   * @export
   * @this {Diagram}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Part} null if a data with that key cannot be found in the model,
   * or if a corresponding {@link Part} cannot be found in the Diagram.
   * This will not return a {@link Link} unless the model is a {@link GraphLinksModel} and
   * {@link GraphLinksModel#linkKeyProperty} has been set.
   * If the same key is used for both a node data object and a link data object, this will return a Node.
   */
  public findPartForKey(key: Key): Part | null {
    return this.partManager.findPartForKey(key);
  }

  /**
   * Look for a {@link Node} or {@link Group} corresponding to a model's node data object's unique key.
   * @export
   * @this {Diagram}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Node} null if a node data with that key cannot be found in the model,
   * or if a corresponding {@link Node} or {@link Group} cannot be found in the Diagram,
   * or if what is found is just a {@link Part}.
   */
  public findNodeForKey(key: Key): Node | null {
    return this.partManager.findNodeForKey(key);
  }

  /**
   * Look for a {@link Link} corresponding to a model's link data object's unique key.
   * @export
   * @this {Diagram}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Link} null if a link data with that key cannot be found in the model,
   * or if a corresponding {@link Link} cannot be found in the Diagram,
   * or if the model is not a {@link GraphLinksModel} with {@link GraphLinksModel#linkKeyProperty} set to a non-empty string.
   * @since 2.1
   */
  public findLinkForKey(key: Key): Link | null {
    return this.partManager.findLinkForKey(key);
  }

  /**
   * Look for a {@link Part}, {@link Node}, {@link Group}, or {@link Link} corresponding to a {@link Model}'s data object.
   * We recommend that you call {@link #findNodeForData} or {@link #findLinkForData} if you are looking for a {@link Node} or a {@link Link}.
   * @export
   * @this {Diagram}
   * @param {Object} data a JavaScript object matched by reference identity
   * @return {Part} an existing {@link Part} in this Diagram that was
   * created because its `Part.data` was the data in the Diagram's {@link Model}.
   */
  public findPartForData(data: ObjectData): Part | null {
    return this.partManager.findPartForData(data);
  }

  /**
   * Look for a {@link Node} or {@link Group}  corresponding to a model's node data object.
   * @export
   * @this {Diagram}
   * @param {Object} nodedata a JavaScript object matched by reference identity;
   *   use {@link #findNodesByExample} if you want to find those Nodes whose data matches an example data object
   * @return {Node} an existing {@link Node} or {@link Group} in this Diagram that was
   * created because its `Part.data` was the node data in the Diagram's {@link Model}.
   * This will be null if there is no such part or if it's just a {@link Part} or {@link Link}.
   */
  public findNodeForData(nodedata: ObjectData): Node | null {
    return this.partManager.findNodeForData(nodedata);
  }

  /**
   * Look for a {@link Link} corresponding to a {@link GraphLinksModel}'s link data object.
   * @export
   * @this {Diagram}
   * @param {Object} linkdata a JavaScript object matched by reference identity;
   *   use {@link #findLinksByExample} if you want to find those Links whose data matches an example data object
   * @return {Link} an existing {@link Link} in this Diagram that was
   * created because its `Part.data` was the link data in the Diagram's {@link Model}.
   */
  public findLinkForData(linkdata: ObjectData): Link | null {
    return this.partManager.findLinkForData(linkdata);
  }

  /**
   * Search for Nodes or Groups by matching the Node data with example data holding values, RegExps, or predicates.
   *
   * For example, calling this method with an argument object:
   * `{ sex: "M", name: /^Alex/i, age: function(n) { return n >= 18; } }`
   * will return an iterator of Nodes whose `Node.data` is a JavaScript object whose:
   *   - sex is "M" (a case-sensitive string comparison), and
   *   - name starts with the string "Alex" (using a case-insensitive match), and
   *   - age is greater than or equal to 18
   *
   * Here is how an example value can match the corresponding data value:
   *   - A string, number, or boolean is compared with the === operator.
   *   - A function is treated as a predicate and applied to the data value.
   *   - A regular expression (RexExp) is matched against the toString() of the data value.
   *     Common expressions include:
   *     - `/abc/` matches any string that includes exactly the substring "abc"
   *     - `/abc/i` matches any string that includes the substring "abc", but uses a case-insensitive comparison
   *     - `/^no/i` matches any string that starts with "no", ignoring case
   *     - `/ism$` matches any string that ends with "ism" exactly
   *     - `/(green|red) apple/` matches any string that includes either "green apple" or "red apple"
   *
   *     For more details read <a href="https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions">Regular Expressions (mozilla.org)</a>.
   *   - An Array requires the data value to also be an Array of equal or greater length.
   *     Each example array item that is not undefined is matched with the corresponding data array item.
   *   - An Object is recursively matched with the data value, which must also be an Object.
   *
   * All properties given by the argument example data must be present on the node data,
   * unless the example property value is undefined.
   * All other data properties are ignored when matching.
   *
   * When multiple argument objects are given, if any of the objects match the node's data,
   * the node is included in the results.
   * @export
   * @this {Diagram}
   * @param {...Object} examples one or more JavaScript Objects whose properties are either predicates to be
   *     applied or RegExps to be tested or values to be compared to the corresponding data property value
   * @return {Iterator.<Node>}
   * @see {@link #findLinksByExample}
   * @since 1.5
   */
  public findNodesByExample(...examples: Array<ObjectData>): Iterator<Node> {
    return this.partManager.findNodesByExample(...examples);
  }

  /**
   * Search for Links by matching the Link data with example data holding values, RegExps, or predicates.
   *
   * See the documentation of {@link #findNodesByExample} for how the example
   * data can match data of bound Parts.
   * @export
   * @this {Diagram}
   * @param {...Object} examples one or more JavaScript Objects whose properties are either predicates to be
   *     applied or RegExps to be tested or values to be compared to the corresponding data property value
   * @return {Iterator.<Link>}
   * @see {@link #findNodesByExample}
   * @since 1.5
   */
  public findLinksByExample(...examples: Array<ObjectData>): Iterator<Link> {
    return this.partManager.findLinksByExample(...examples);
  }

  /**
   * Gets or sets the default {@link Node} template used as the archetype
   * for node data that is added to the {@link #model}.
   * Setting this property just modifies the {@link #nodeTemplateMap}
   * by replacing the entry named with the empty string.
   *
   * The value must not be null.
   * The template may be either a {@link Node} or a simple {@link Part},
   * but not a {@link Link} or a {@link Group}.
   *
   * This Part must not be in the visual tree of any Diagram.
   */
  /** @export */ get nodeTemplate(): Part { return this._nodeTemplateMap.getValue(''); }
  /** @export */ set nodeTemplate(value: Part) {
    const old = this._nodeTemplateMap.getValue('');
    if (old !== value) {
      Util.checkClass(value, Part, Diagram, 'nodeTemplate');
      this._nodeTemplateMap.add('', value);
      this.raiseChanged('nodeTemplate', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets a {@link Map} mapping template names to {@link Part}s.
   * These nodes are copied for each node data that is added to the {@link #model}.
   *
   * The new value must not be null and must contain {@link Node}s or simple {@link Part}s.
   * These Parts must not be in the visual tree of any Diagram.
   * Replacing this Map will automatically call {@link #rebuildParts}.
   *
   * If you modify this {@link Map}, by replacing a {@link Node} or by adding or
   * removing a map entry, you need to explicitly call {@link #rebuildParts} afterwards.
   * Any new map values must not be {@link Link}s or {@link Group}s.
   *
   * If you want to create {@link Group}s, use {@link #groupTemplateMap} instead.
   */
  /** @export */ get nodeTemplateMap(): Map<string, Part> { return this._nodeTemplateMap; }
  /** @export */ set nodeTemplateMap(value: Map<string, Part>) {
    const old = this._nodeTemplateMap;
    if (old !== value) {
      Util.checkClass(value, Map, Diagram, 'nodeTemplateMap');
      this._nodeTemplateMap = value;
      this.raiseChanged('nodeTemplateMap', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets the default {@link Group} template used as the archetype
   * for group data that is added to the {@link #model}.
   *
   * Setting this property just modifies the {@link #groupTemplateMap}
   * by replacing the entry named with the empty string.
   * The value must not be null and must be a {@link Group}, not a {@link Node} or simple {@link Part}.
   * This Part must not be in the visual tree of any Diagram.
   */
  /** @export */ get groupTemplate(): Group { return this._groupTemplateMap.getValue(''); }
  /** @export */ set groupTemplate(value: Group) {
    const old = this._groupTemplateMap.getValue('');
    if (old !== value) {
      Util.checkClass(value, Group, Diagram, 'groupTemplate');
      this._groupTemplateMap.add('', value);
      this.raiseChanged('groupTemplate', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets a {@link Map} mapping template names to {@link Group}s.
   * These groups are copied for each group data that is added to the {@link #model}.
   *
   * The new value must not be null, nor may it contain a {@link Node} or {@link Link} or simple {@link Part}.
   * The Links must not be in the visual tree of any Diagram.
   * Replacing this Map will automatically call {@link #rebuildParts}.
   *
   * If you modify this {@link Map}, by replacing a {@link Group} in it or by adding or
   * removing a map entry, you need to explicitly call {@link #rebuildParts} afterwards.
   */
  /** @export */ get groupTemplateMap(): Map<string, Group> { return this._groupTemplateMap; }
  /** @export */ set groupTemplateMap(value: Map<string, Group>) {
    const old = this._groupTemplateMap;
    if (old !== value) {
      Util.checkClass(value, Map, Diagram, 'groupTemplateMap');
      this._groupTemplateMap = value;
      this.raiseChanged('groupTemplateMap', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets the default {@link Link} template used as the archetype
   * for link data that is added to the {@link #model}.
   *
   * Setting this property just modifies the {@link #linkTemplateMap}
   * by replacing the entry named with the empty string.
   * The value must not be null and must be a {@link Link}, not a {@link Node} or simple {@link Part}.
   * This Link must not be in the visual tree of any Diagram.
   */
  /** @export */ get linkTemplate(): Link { return this._linkTemplateMap.getValue(''); }
  /** @export */ set linkTemplate(value: Link) {
    const old = this._linkTemplateMap.getValue('');
    if (old !== value) {
      Util.checkClass(value, Link, Diagram, 'linkTemplate');
      this._linkTemplateMap.add('', value);
      this.raiseChanged('linkTemplate', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets a {@link Map} mapping template names to {@link Link}s.
   * These links are copied for each link data that is added to the {@link #model}.
   *
   * The new value must not be null and must contain only {@link Link}s, not {@link Node}s or simple {@link Part}s.
   * The Links must not be in the visual tree of any Diagram.
   * Replacing this Map will automatically call {@link #rebuildParts}.
   *
   * If you modify this {@link Map}, by replacing a {@link Link} in it or by adding or
   * removing a map entry, you need to explicitly call {@link #rebuildParts} afterwards.
   */
  /** @export */ get linkTemplateMap(): Map<string, Link> { return this._linkTemplateMap; }
  /** @export */ set linkTemplateMap(value: Map<string, Link>) {
    const old = this._linkTemplateMap;
    if (old !== value) {
      Util.checkClass(value, Map, Diagram, 'linkTemplateMap');
      this._linkTemplateMap = value;
      this.raiseChanged('linkTemplateMap', old, value);
      if (!this.undoManager.isUndoingRedoing) this.rebuildParts();
    }
  }

  /**
   * Gets or sets whether mouse events initiated within the Diagram will be captured.
   * The initial value is true.
   * Setting this property does not notify about any changed event.
   */
  /** @export */ get isMouseCaptured(): boolean { return this._isMouseCaptured; }
  /** @export */ set isMouseCaptured(value: boolean) {
    const can = this._canvas;
    if (can === null) return;
    const canvasElt = can.domElement;
    if (canvasElt instanceof SVGElement) return; // NYI: no events yet
    if (value) {
      this.lastInput.bubbles = false;
      if (this._allowsPointers) {
        this.removeEventListener(canvasElt, 'pointermove', this.pointerMove, false);
        this.removeEventListener(canvasElt, 'pointerdown', this.pointerDown, false);
        this.removeEventListener(canvasElt, 'pointerup', this.pointerUp, false);
        this.removeEventListener(canvasElt, 'pointerout', this.pointerOut, false);
        this.addEventListener(root, 'pointermove', this.pointerMove, true);
        this.addEventListener(root, 'pointerdown', this.pointerDown, true);
        this.addEventListener(root, 'pointerup', this.pointerUp, true);
        this.addEventListener(root, 'pointerout', this.pointerOut, true);
      } else {
        this.removeEventListener(canvasElt, 'mousemove', this.mouseMove, false);
        this.removeEventListener(canvasElt, 'mousedown', this.mouseDown, false);
        this.removeEventListener(canvasElt, 'mouseup', this.mouseUp, false);
        this.removeEventListener(canvasElt, 'mouseout', this.mouseOut, false);
        this.addEventListener(root, 'mousemove', this.mouseMove, true);
        this.addEventListener(root, 'mousedown', this.mouseDown, true);
        this.addEventListener(root, 'mouseup', this.mouseUp, true);
        this.addEventListener(root, 'mouseout', this.mouseOut, true);
      }

      this.removeEventListener(canvasElt, 'wheel', this.mouseWheel, false);
      this.addEventListener(root, 'wheel', this.mouseWheel, true);
      this.addEventListener(root, 'selectstart', this.preventDefault, false);
    } else {
      if (this._allowsPointers) {
        this.removeEventListener(root, 'pointermove', this.pointerMove, true);
        this.removeEventListener(root, 'pointerdown', this.pointerDown, true);
        this.removeEventListener(root, 'pointerup', this.pointerUp, true);
        this.removeEventListener(root, 'pointerout', this.pointerOut, true);
        this.addEventListener(canvasElt, 'pointermove', this.pointerMove, false);
        this.addEventListener(canvasElt, 'pointerdown', this.pointerDown, false);
        this.addEventListener(canvasElt, 'pointerup', this.pointerUp, false);
        this.addEventListener(canvasElt, 'pointerout', this.pointerOut, false);
      } else {
        this.removeEventListener(root, 'mousemove', this.mouseMove, true);
        this.removeEventListener(root, 'mousedown', this.mouseDown, true);
        this.removeEventListener(root, 'mouseup', this.mouseUp, true);
        this.removeEventListener(root, 'mouseout', this.mouseOut, true);
        this.addEventListener(canvasElt, 'mousemove', this.mouseMove, false);
        this.addEventListener(canvasElt, 'mousedown', this.mouseDown, false);
        this.addEventListener(canvasElt, 'mouseup', this.mouseUp, false);
        this.addEventListener(canvasElt, 'mouseout', this.mouseOut, false);
      }

      this.removeEventListener(root, 'wheel', this.mouseWheel, true);
      this.removeEventListener(root, 'selectstart', this.preventDefault, false);
      this.addEventListener(canvasElt, 'wheel', this.mouseWheel, false);
    }

    this._isMouseCaptured = value;
  }

  /**
   * Gets or sets the coordinates of this Diagram in the viewport.
   * Value must be of type {@link Point} in document coordinates.
   * The default is Point(NaN, NaN), but is typically set to a real value when a Diagram is initialized.
   *
   * Scrolling and panning the Diagram modify the Diagram's position.
   *
   * Setting this property does not notify about any changed event.
   * However you can listen with {@link #addDiagramListener} for a {@link DiagramEvent}
   * with the name "ViewportBoundsChanged".
   *
   * The {@link #viewportBounds} x and y values are always the same as the Diagram's position values.
   *
   * If you set this property any replacement of the {@link #model} will result in a layout
   * and a computation of new {@link #documentBounds}, which in turn may cause the diagram to be scrolled
   * and zoomed, depending on various Diagram properties named "initial...".
   * You may want to set {@link #initialPosition} instead of setting this property around the time
   * that you are loading a model.
   */
  /** @export */ get position(): Point { return this._position; }
  /** @export */ set position(value: Point) {
    const old = Point.alloc().assign(this._position);
    if (!old.equals(value)) {
      Util.checkClass(value, Point, Diagram, 'position');
      const oldvbounds = this.viewportBounds.copy(); // copy it before position is set
      this._position.assign(value);
      // setup pos calls this in the if statement
      this.animationManager.addDiagramPos(old, this._position);

      if (!this._isAligning && !(this._canvas === null && !this._viewSize.isReal())) {
        this._isAligning = true;
        const scale = this.scale;
        const viewwidth = this._canvasWidth / scale;
        const viewheight = this._canvasHeight / scale;
        this._setupPos(this._documentBounds, viewwidth, viewheight, this._contentAlignment, false);
        this._isAligning = false;
      }

      // If its aligning (such as setting scale), onViewportBoundsChanged will get called
      if (!this._isAligning) {
        const newvbounds = this.viewportBounds;
        this.onViewportBoundsChanged(oldvbounds, newvbounds, this._scale, false);
      }
    }
    Point.free(old);
  }

  /**
   * Gets or sets the initial coordinates of this Diagram in the viewport, eventually setting the {@link #position}.
   * This value is relevant on initialization of a {@link #model} or if {@link #delayInitialization} is called.
   * Value must be of type Point in document coordinates.
   * The default is Point(NaN, NaN).
   *
   * Setting this property has the same effect as implementing
   * an "InitialLayoutCompleted" DiagramEvent listener that sets {@link #position}.
   *
   * Setting this property does not notify about any changed event.
   * @see {@link #initialDocumentSpot}
   * @see {@link #initialViewportSpot}
   * @since 1.1
   */
  /** @export */ get initialPosition(): Point { return this._initialPosition; }
  /** @export */ set initialPosition(value: Point) {
    const old = this._initialPosition;
    if (!old.equals(value)) {
      Util.checkClass(value, Point, Diagram, 'initialPosition');
      this._initialPosition = value.copyFrozen();
    }
  }

  /**
   * Gets or sets the initial scale of this Diagram in the viewport, eventually setting the {@link #scale}.
   * This value is relevant on initialization of a {@link #model} or if {@link #delayInitialization} is called.
   * The default is NaN.
   *
   * Setting this property has the same effect as implementing
   * an "InitialLayoutCompleted" DiagramEvent listener that sets {@link #scale}.
   * @since 1.1
   */
  /** @export */ get initialScale(): number { return this._initialScale; }
  /** @export */ set initialScale(value: number) {
    const old = this._initialScale;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', Diagram, 'initialScale');
      this._initialScale = value;
    }
  }

  /**
   * Gets or sets a {@link Panel} of type {@link Panel.Grid} acting as the background grid
   * extending across the whole viewport of this diagram.
   */
  /** @export */ get grid(): Panel {
    if (this._grid === null) this._setupBackgroundGrid();
    return this._grid;
  }
  /** @export */ set grid(value: Panel) {
    let old = this._grid;
    if (old !== value) {
      if (old === null) {
        this._setupBackgroundGrid();
        old = this._grid;
      }
      Util.checkClass(value, Panel, Diagram, 'grid');
      if (value.type !== Panel.Grid) Util.throwError('Diagram.grid must be a Panel of type Panel.Grid');
      const oldpart = old.panel;
      if (oldpart !== null) oldpart.remove(old);
      this._grid = value;
      value.name = 'GRID';
      if (oldpart !== null) oldpart.add(value);
      this._updateBackgroundGrid();
      this.invalidateDraw();
      this.raiseChanged('grid', old, value);
    }
  }

  /**
   * This read-only property returns the bounds of the portion of the Diagram in document coordinates that is viewable from its HTML Canvas.
   * Typically when the viewport bounds are smaller than the {@link #documentBounds}, the user can scroll or pan the view.
   *
   * The x and y coordinates are equal to the {@link #position} of the Diagram,
   * and the width and height are equal to the Diagram's canvas width and height,
   * divided by the {@link #scale}.
   */
  /** @export */ get viewportBounds(): Rect {
    const vb = this._viewportBounds;
    const pos = this._position;
    const sc = this._scale;
    if (this._canvas === null) {
      if (this._viewSize.isReal()) vb.assignTo(pos.x, pos.y, this._canvasWidth / sc, this._canvasHeight / sc);
      return vb;
    }
    vb.assignTo(pos.x, pos.y, Math.max(this._canvasWidth, 0) / sc, Math.max(this._canvasHeight, 0) / sc);
    return vb;
  }

  /**
   * Gets or sets a fixed bounding rectangle to be returned by {@link #viewportBounds}
   * when {@link #div} is null.
   * By default this is (NaN, NaN), and it is not typically set except in DOM-less environments
   * where there will not be a Diagram DIV. Normally, the viewportBounds is sized by the DIV.
   *
   * See the intro page on <a href="../../intro/nodeScript.html">GoJS within Node</a> for a usage example.
   *
   * @since 2.0
   */
  /** @export */ get viewSize(): Size { return this._viewSize; }
  /** @export */ set viewSize(value: Size) {
    const old = this.viewSize;
    if (!old.equals(value)) {
      Util.checkClass(value, Size, Diagram, 'viewSize');
      value = value.copyFrozen();
      this._viewSize = value;
      this._canvasWidth = value.width;
      this._canvasHeight = value.height;
      this.invalidateDocumentBounds();
      this.raiseChanged('viewSize', old, value);
    }
  }

  /**
   * Gets or sets a fixed bounding rectangle to be returned by {@link #documentBounds}
   * and {@link #computeBounds}.
   * By default this has NaN values, meaning that {@link #computeBounds}
   * will compute the union of all of the parts in the Diagram to determine the {@link #documentBounds}.
   * If all x/y/width/height values are real numbers, this value is used as the {@link #documentBounds}.
   */
  /** @export */ get fixedBounds(): Rect { return this._fixedBounds; }
  /** @export */ set fixedBounds(value: Rect) {
    const old = this._fixedBounds;
    if (!old.equals(value)) {
      Util.checkClass(value, Rect, Diagram, 'fixedBounds');
      if (Debug && value.width === Infinity || value.width === -Infinity ||
        value.height === Infinity || value.height === -Infinity) {
        Util.throwError('fixedBounds width/height must not be Infinity');
      }
      value = value.copyFrozen();
      this._fixedBounds = value;
      this.invalidateDocumentBounds();
      this.raiseChanged('fixedBounds', old, value);
    }
  }

  /**
   * Gets or sets a scrollable area in document coordinates that surrounds the document bounds, allowing the user to scroll into empty space.
   *
   * The margin is only effective in each direction when the document bounds plus margin is greater than the viewport bounds.
   *
   * The default value is a margin of 0, all around the edge of the document.
   * @since 1.5
   */
  /** @export */ get scrollMargin(): MarginLike { return this._scrollMargin; }
  /** @export */ set scrollMargin(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Diagram, 'scrollMargin');
    }
    const old = this._scrollMargin;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._scrollMargin = value;
      this.raiseChanged('scrollMargin', old, value);
      this.redraw();
    }
  }

  /**
   * Gets or sets the scrollMode of the Diagram, allowing the user to either
   * scroll to document bound borders with {@link Diagram.DocumentScroll},
   * or scroll endlessly with {@link Diagram.InfiniteScroll}.
   *
   * The default value is {@link Diagram.DocumentScroll}.
   * Changing this property value does not raise a Changed event.
   * @since 1.5
   */
  /** @export */ get scrollMode(): EnumValue { return this._scrollMode; }
  /** @export */ set scrollMode(value: EnumValue) {
    const old = this._scrollMode;
    if (old !== value) {
      EnumValue.checkEnumValue(value, Diagram, Diagram, 'scrollMode');
      this._scrollMode = value;
      if (value === Diagram.DocumentScroll) this.onContentAlignmentChanged(false);
      this.raiseChanged('scrollMode', old, value);
      this.redraw(); // ensures scrollbars dis/appear
    }
  }

  /**
   * Gets or sets whether the page may be scrolled when the diagram receives focus.
   * This happens in some browsers when the top-left corner of the diagram's HTMLDivElement
   * is scrolled out of view, the diagram does not have focus, and the user clicks in the diagram.
   *
   * The default value is false.
   * @since 1.8
   */
  /** @export */ get scrollsPageOnFocus(): boolean { return this._scrollsPageOnFocus; }
  /** @export */ set scrollsPageOnFocus(value: boolean) {
    const old = this._scrollsPageOnFocus;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'scrollsPageOnFocus');
      this._scrollsPageOnFocus = value;
      this.raiseChanged('scrollsPageOnFocus', old, value);
    }
  }

  /**
   * Gets or sets the function used to determine the position that this Diagram can be scrolled or moved to.
   *
   * By default this function is null and the Diagram's position is bound only by the document bounds.
   *
   * When this property is set the function is given a reference to the diagram and the proposed new position Point.
   * The function must return a new point.
   *
   * An example that disallows decimal position values:
   * ```js
   *   function computeIntegralPosition(diagram, pt) {
   *     return new go.Point(Math.floor(pt.x), Math.floor(pt.y));
   *   }
   * ```
   *
   * The function, if supplied, must not have any side-effects.
   * @since 1.5
   */
  /** @export */ get positionComputation(): ((thisDiagram: Diagram, newPosition: Point) => Point) | null { return this._positionComputation; }
  /** @export */ set positionComputation(value: ((thisDiagram: Diagram, newPosition: Point) => Point) | null) {
    const old = this._positionComputation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'positionComputation');
      this._positionComputation = value;
      this.onContentAlignmentChanged(false);
      this.raiseChanged('positionComputation', old, value);
    }
  }

  /**
   * Gets or sets the function used to determine valid scale values for this Diagram.
   * @since 1.5
   */
  /** @export */ get scaleComputation(): ((thisDiagram: Diagram, newScale: number) => number) | null { return this._scaleComputation; }
  /** @export */ set scaleComputation(value: ((thisDiagram: Diagram, newScale: number) => number) | null) {
    const old = this._scaleComputation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Diagram, 'scaleComputation');
      this._scaleComputation = value;
      this.onScaleChanged(this.scale, this.scale);
      this.raiseChanged('scaleComputation', old, value);
    }
  }

  /**
   * This read-only property returns the bounds of the diagram's contents, in document coordinates.
   *
   * This is normally computed and set by {@link #computeBounds} during Diagram updates
   * that can occur for any number of relevant reasons, such as a {@link Part} changing size.
   *
   * The Diagram's documentBounds can have an unvarying specific value
   * by setting the {@link #fixedBounds} property.
   *
   * If the documentBounds are larger than the {@link #viewportBounds},
   * scrollbars will appear on desktop browsers. You can disable scrolling with the
   * {@link #allowHorizontalScroll} and {@link #allowVerticalScroll} properties, and
   * you can disable scrollbars themselves with the {@link #hasHorizontalScrollbar}
   * and {@link #hasVerticalScrollbar} properties.
   */
  /** @export */ get documentBounds(): Rect {
    return this._documentBounds;
  }

  /**
   * @hidden @internal
   * Essentially the documentBounds setter (but that's a readonly property). Does appropriate side effects.
   * @this {Diagram}
   * @param {Rect} val
   */
  public setDocumentBounds(val: Rect): void {
    this._invalidDocBounds = false; // now they're valid
    const old = this._documentBounds;
    if (!old.equals(val)) {
      val = val.copyFrozen();
      this._documentBounds = val;
      this.onContentAlignmentChanged(false);
      this.raiseDiagramEvent('DocumentBoundsChanged', null, old.copy());
      this.invalidateScroll();
    }
  }

  // Since 2.1
  /**
   * @hidden
   * Undocumented
   * @this {Diagram}
   */
  public ensureDocumentBounds(): void {
    if (this._invalidDocBounds) {
      this.setDocumentBounds(this.computeBounds());
    }
  }

  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get isVirtualized(): boolean { return this._isVirtualized; }
  /** @export */ set isVirtualized(value: boolean) {
    const old = this._isVirtualized;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'isVirtualized');
      this._isVirtualized = value;
      this.raiseChanged('isVirtualized', old, value);
    }
  }

  /**
   * Gets or sets the scale transform of this Diagram.
   * Value must be a positive number.
   * The default value is 1.
   * Any new value will be coerced to be between {@link #minScale} and {@link #maxScale}.
   *
   * Scale can automatically be set by the {@link #autoScale} property.
   * There are also {@link #initialScale} and {@link #initialAutoScale} for
   * setting the scale on (re)initialization of a Diagram.
   *
   * Setting this property does not notify about any changed event.
   * However you can listen with {@link #addDiagramListener} for a {@link DiagramEvent}
   * with the name "ViewportBoundsChanged".
   *
   * If you set this property any replacement of the {@link #model} will result in a layout
   * and a computation of new {@link #documentBounds}, which in turn may cause the diagram to be scrolled
   * and zoomed, depending on various Diagram properties named "initial...".
   * You may want to set {@link #initialScale} instead of setting this property around the time
   * that you are loading a model.
   */
  /** @export */ get scale(): number { return this._scale; }
  /** @export */ set scale(value: number) {
    const old = this._scale;
    Util.checkRealNumber(value, Diagram, 'scale');
    if (old !== value) {
      this.onScaleChanged(old, value); // sets this._scale
    }
  }


  /**
   * Gets or sets the {@link Diagram#scale} set by {@link CommandHandler#resetZoom}
   * and when computing stretch values, such as when {@link #autoScale} or
   * {@link #initialAutoScale} are set, or when {@link #zoomToFit} is called.
   *
   * The default value is 1.0.
   * The value must be a number larger than 0.
   * Setting this property does not raise any events.
   * @since 2.0
   */
  /** @export */ get defaultScale(): number { return this._defaultScale; }
  /** @export */ set defaultScale(value: number) {
    if (Debug) Util.checkRealNumber(value, Diagram, 'defaultScale');
    if (Debug && !(value > 0)) Util.throwError('defaultScale must be larger than zero, not: ' + value);
    this._defaultScale = value;
  }

  /**
   * Gets or sets the autoScale behavior of the Diagram, controlling whether or not the
   * Diagram's bounds automatically scale to fit the view.
   *
   * The only accepted values are the constant properties of Diagram,
   * {@link Diagram.None}, {@link Diagram.Uniform}, or {@link Diagram.UniformToFill}.
   * Setting this will change the Diagram's {@link Diagram#scale} and {@link Diagram#position}, if appropriate.
   *
   * The default value is {@link Diagram.None} - the scale and position are not automatically adjusted
   * according to the area covered by the document.
   * When the value is not None, any value for {@link #initialAutoScale} or {@link #initialScale} is ignored.
   * <p class="box">
   * When autoScale is set to a non-{@link Diagram.None} value,
   * the user will not be able to zoom, and setting {@link #scale} will do nothing.
   * If you only want to scale automatically on initialization, use {@link #initialAutoScale}.
   *
   * Setting this property to {@link Diagram.Uniform} is basically the same as calling {@link #zoomToFit}
   * all the time, or just disabling interactive zooming.
   *
   * Note that depending on the values of {@link #maxScale} and {@link #minScale}, the actual value for {@link #scale}
   * might be limited.
   */
  /** @export */ get autoScale(): EnumValue { return this._autoScale; }
  /** @export */ set autoScale(value: EnumValue) {
    const old = this._autoScale;
    if (old !== value) {
      EnumValue.checkEnumValue(value, Diagram, Diagram, 'autoScale');
      this._autoScale = value;
      this.raiseChanged('autoScale', old, value);
      if (value !== Diagram.None) {
        this.onContentAlignmentChanged(false);
      }
    }
  }

  /**
   * Gets or sets how the scale of the diagram is automatically
   * set at the time of the "InitialLayoutCompleted" DiagramEvent, after the model has been replaced.
   *
   * The only accepted values are listed as constant properties of Diagram,
   * such as {@link Diagram.None}, {@link Diagram.Uniform}, or {@link Diagram.UniformToFill}.
   * Setting this will change the Diagram's {@link Diagram#scale} and {@link Diagram#position}, if appropriate.
   *
   * If you want to always automatically scale the Diagram, set {@link #autoScale} instead.
   * If you want to set the scale to a specific value on initialization (each time the model is replaced),
   * set {@link #initialScale}.
   *
   * The default value is {@link Diagram.None} -- the scale and position are not automatically adjusted
   * according to the area covered by the document.
   *
   * Setting this property to {@link Diagram.Uniform} is basically the same as calling {@link #zoomToFit}
   * in an "InitialLayoutCompleted" {@link DiagramEvent} listener.
   *
   * Note that depending on the values of {@link #maxScale} and {@link #minScale}, the actual value for {@link #scale}
   * might be limited.
   */
  /** @export */ get initialAutoScale(): EnumValue { return this._initialAutoScale; }
  /** @export */ set initialAutoScale(value: EnumValue) {
    const old = this._initialAutoScale;
    if (old !== value) {
      EnumValue.checkEnumValue(value, Diagram, Diagram, 'initialAutoScale');
      this._initialAutoScale = value;
      this.raiseChanged('initialAutoScale', old, value);
    }
  }

  /**
   * Gets or sets the spot in the viewport that should be coincident with the
   * {@link #initialDocumentSpot} of the document when the document is first initialized.
   * The default value is {@link Spot.TopLeft}.
   *
   * If you set this, often you will also want to set {@link #initialDocumentSpot}.
   * If you set {@link #initialPosition}, it will take precedence over this property.
   *
   * Setting this property and {@link #initialDocumentSpot} has the same effect as implementing
   * an "InitialLayoutCompleted" DiagramEvent listener that calls {@link #alignDocument}.
   *
   * @see {@link #initialDocumentSpot}
   * @see {@link #initialContentAlignment}
   */
  /** @export */ get initialViewportSpot(): Spot { return this._initialViewportSpot; }
  /** @export */ set initialViewportSpot(value: Spot) {
    const old = this._initialViewportSpot;
    if (old !== value) {
      Util.checkClass(value, Spot, Diagram, 'initialViewportSpot');
      if (!value.isSpot()) Util.throwError('initialViewportSpot must be a specific Spot: ' + value);
      this._initialViewportSpot = value;
      this.raiseChanged('initialViewportSpot', old, value);
    }
  }

  /**
   * Gets or sets the spot in the document's area that should be coincident with the
   * {@link #initialViewportSpot} of the viewport when the document is first initialized.
   * The default value is {@link Spot.TopLeft}.
   *
   * If you set this, often you will also want to set {@link #initialViewportSpot}.
   * If you set {@link #initialPosition}, it will take precedence over this property.
   *
   * Setting this property and {@link #initialViewportSpot} has the same effect as implementing
   * an "InitialLayoutCompleted" DiagramEvent listener that calls {@link #alignDocument}.
   *
   * @see {@link #initialViewportSpot}
   * @see {@link #initialPosition}
   * @see {@link #initialContentAlignment}
   */
  /** @export */ get initialDocumentSpot(): Spot { return this._initialDocumentSpot; }
  /** @export */ set initialDocumentSpot(value: Spot) {
    const old = this._initialDocumentSpot;
    if (old !== value) {
      Util.checkClass(value, Spot, Diagram, 'initialDocumentSpot');
      if (!value.isSpot()) Util.throwError('initialViewportSpot must be a specific Spot: ' + value);
      this._initialDocumentSpot = value;
      this.raiseChanged('initialDocumentSpot', old, value);
    }
  }

  /**
   * Gets or sets the smallest value greater than zero that {@link #scale} may take.
   * This property is only used to limit the range of new values of {@link #scale}.
   *
   * The default value is 0.0001.
   * Values must be larger than zero and not greater than one.
   * Setting this to a value that is greater than the current {@link #scale}
   * will cause the current diagram scale to be set to this new value.
   */
  /** @export */ get minScale(): number { return this._minScale; }
  /** @export */ set minScale(value: number) {
    Util.checkRealNumber(value, Diagram, 'minScale');
    const old = this._minScale;
    if (old !== value) {
      if (value > 0) {
        this._minScale = value;
        this.raiseChanged('minScale', old, value);
        if (value > this.scale) this.scale = value;
      } else {
        Util.throwRangeError(value, '> 0', Diagram, 'minScale');
      }
    }
  }

  /**
   * Gets or sets the largest value that {@link #scale} may take.
   * This property is only used to limit the range of new values of {@link #scale}.
   *
   * The default value is 100.0.
   * Values must be no less than one.
   * Setting this to a value that is less than the current {@link #scale}
   * will cause the current diagram scale to be set to this new value.
   */
  /** @export */ get maxScale(): number { return this._maxScale; }
  /** @export */ set maxScale(value: number) {
    Util.checkRealNumber(value, Diagram, 'maxScale');
    const old = this._maxScale;
    if (old !== value) {
      if (value > 0) {
        this._maxScale = value;
        this.raiseChanged('maxScale', old, value);
        if (value < this.scale) this.scale = value;
      } else {
        Util.throwRangeError(value, '> 0', Diagram, 'maxScale');
      }
    }
  }

  /**
   * Gets or sets the zoom point of this Diagram, in viewport coordinates.
   * This is used by {@link Tool#standardMouseWheel} and scale-setting commands to control where to zoom in or out.
   *
   * Typical usage is to remember the value of this property and
   * to set this property to some point within the viewport (between zero and the canvas width and height).
   * This is commonly accomplished by using the {@link InputEvent#viewPoint} of {@link Diagram#lastInput}.
   * Then one changes the {@link #scale} somehow, perhaps by executing one of the {@link CommandHandler} commands,
   * or by rotating the mouse wheel, or just by setting the {@link Diagram#scale} property.
   * Finally one restores the original value of this property.
   *
   * The default value is Point(NaN, NaN).
   * Value must be of type {@link Point}, in element coordinates, not in document coordinates.
   * Setting this property does not notify about any changed event.
   * @since 1.2
   */
  /** @export */ get zoomPoint(): Point { return this._zoomPoint; }
  /** @export */ set zoomPoint(value: Point) {
    const old = this._zoomPoint;
    if (!old.equals(value)) {
      Util.checkClass(value, Point, Diagram, 'zoomPoint');
      value = value.copyFrozen();
      this._zoomPoint = value;
    }
  }

  /**
   * Gets or sets the content alignment {@link Spot} of this Diagram, to be used in determining
   * how parts are positioned when the {@link #viewportBounds} width or height is smaller than the {@link #documentBounds}.
   *
   * For instance a spot of {@link Spot.Center} would ensure that the Diagram's
   * contents are always centered in the viewport.
   *
   * If you want the content to be aligned only initially, use {@link #initialContentAlignment} instead.
   *
   * The default value is {@link Spot.Default}, which causes no automatic scrolling or positioning.
   * When the value is not Default, any value for {@link #initialContentAlignment} or {@link #initialPosition} is ignored.
   *
   * Setting this property has the same effect as implementing
   * a "LayoutCompleted" DiagramEvent listener that scrolls the viewport to align the content.
   *
   */
  /** @export */ get contentAlignment(): Spot { return this._contentAlignment; }
  /** @export */ set contentAlignment(value: Spot) {
    const old = this._contentAlignment;
    if (!old.equals(value)) {
      Util.checkClass(value, Spot, Diagram, 'contentAlignment');
      value = value.copyFrozen();
      this._contentAlignment = value;
      this.raiseChanged('contentAlignment', old, value);
      this.onContentAlignmentChanged(false);
    }
  }

  /**
   * Gets or sets the initial content alignment {@link Spot} of this Diagram, to be used in determining
   * how parts are positioned initially relative to the viewport, when the {@link #viewportBounds} width or height is smaller than the {@link #documentBounds}.
   *
   * For instance a spot of {@link Spot.Center} would ensure that the Diagram's contents are initially centered in the viewport.
   *
   * To initially align the document when the documentBounds are larger than the viewport, use {@link #initialDocumentSpot} and {@link #initialViewportSpot}.
   *
   * If you want the content to be constantly aligned with a spot, use {@link #contentAlignment} instead.
   *
   * The default value is {@link Spot.Default}, which causes no automatic scrolling or positioning.
   *
   * Setting this property has the same effect as implementing
   * an "InitialLayoutCompleted" DiagramEvent listener that scrolls the viewport to align the content.
   *
   * @see {@link #initialDocumentSpot}
   * @see {@link #initialViewportSpot}
   */
  /** @export */ get initialContentAlignment(): Spot { return this._initialContentAlignment; }
  /** @export */ set initialContentAlignment(value: Spot) {
    const old = this._initialContentAlignment;
    if (!old.equals(value)) {
      Util.checkClass(value, Spot, Diagram, 'initialContentAlignment');
      value = value.copyFrozen();
      this._initialContentAlignment = value;
      this.raiseChanged('initialContentAlignment', old, value);
    }
  }

  /**
   * Gets or sets the Margin that describes the Diagram's padding,
   * which controls how much extra space in document coordinates there is around the area occupied by the document.
   * This keeps nodes from butting up against the side of the diagram (unless scrolled).
   *
   * The default value is a margin of 5, all around the edge of the document.
   */
  /** @export */ get padding(): MarginLike { return this._padding; }
  /** @export */ set padding(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Diagram, 'padding');
    }
    const old = this._padding;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._padding = value;
      this.invalidateDocumentBounds();
      this.raiseChanged('padding', old, value);
    }
  }

  /**
   * @hidden @internal
   * Gets or sets the {@link PartManager} responsible for creating and updating
   * {@link Part}s for all of the data managed by the {@link Model}.
   *
   * The default value is an instance of {@link PartManager}.
   * There must always be a non-null value for this property.
   * {@link PartManager}s cannot be shared by Diagrams.
   */
  /** @export */ get partManager(): PartManager { return this._partManager; }
  /** @export */ set partManager(value: PartManager) {
    const old = this._partManager;
    if (old !== value) {
      Util.checkClass(value, PartManager, Diagram, 'partManager');
      if (value.diagram !== null) Util.throwError('Cannot share PartManagers between Diagrams: ' + value.toString());
      if (old !== null) old.setDiagram(null);
      this._partManager = value;
      value.setDiagram(this);
    }
  }

  /**
   * This read-only property returns an iterator of all {@link Node}s and {@link Group}s in the Diagram.
   *
   * This includes both data-bound and unbound nodes,
   * and both top-level nodes and nodes inside {@link Group}s.
   * All of the simple {@link Part}s are accessible via the {@link #parts} property.
   * @see {@link #findTopLevelGroups}
   * @see {@link #findTreeRoots}
   */
  /** @export */ get nodes(): Iterator<Node> { return this.partManager.nodes.iterator; }

  /**
   * This read-only property returns an iterator of all {@link Link}s in the Diagram.
   *
   * This includes both data-bound and unbound links,
   * and both top-level links and links inside {@link Group}s.
   */
  /** @export */ get links(): Iterator<Link> { return this.partManager.links.iterator; }

  /**
   * This read-only property returns an iterator of all {@link Part}s in the Diagram
   * that are not {@link Node}s or {@link Link}s or {@link Adornment}s.
   *
   * This includes both data-bound and unbound parts,
   * and both top-level parts and parts inside {@link Group}s.
   * Use the {@link #nodes} or {@link #links} properties for getting the
   * collection of all {@link Node}s or {@link Link}s in the diagram.
   */
  /** @export */ get parts(): Iterator<Part> { return this.partManager.parts.iterator; }

  /**
   * @hidden @internal
   * Returns an iterator of all {@link Node}s and {@link Link}s that are top-level,
   * including top-level {@link Group}s, but not including any simple {@link Part}s or {@link Adornment}s.
   *
   * This is useful for operations on the diagram's top-level graph, such as in implementations of {@link Layout#doLayout}.
   * @export
   * @this {Diagram}
   * @return {Iterator.<Part>}
   */
  public findTopLevelNodesAndLinks(): Iterator<Part> {
    const graph = new Set<Part>();
    const nit = this.nodes;
    while (nit.next()) {
      const n = nit.value;
      if (n.isTopLevel) graph.add(n);
    }
    const lit = this.links;
    while (lit.next()) {
      const l = lit.value;
      if (l.isTopLevel) graph.add(l);
    }
    return graph.iterator;
  }

  /**
   * Returns an iterator of all {@link Group}s that are at top-level,
   * in other words that are not themselves inside other Groups.
   *
   * This is useful for when you want to traverse the diagram's graph by recursing into Groups.
   * @export
   * @this {Diagram}
   * @return {Iterator.<Group>}
   * @since 1.2
   */
  public findTopLevelGroups(): Iterator<Group> {
    return this._topLevelGroups.iterator;
  }

  /**
   * Gets or sets the {@link Layout} used to position all of the top-level nodes and links in this Diagram.
   * By default this property is an instance of a simple {@link Layout}
   * that assigns positions to all parts that need it.
   * The value cannot be null and must not be shared with other Diagrams.
   */
  /** @export */ get layout(): Layout { return this._layout; }
  /** @export */ set layout(value: Layout) {
    const old = this._layout;
    if (old !== value) {
      Util.checkClass(value, Layout, Diagram, 'layout');
      this._layout = value;
      value.diagram = this;
      value.group = null;
      this._anyLayoutInvalid = true;
      this.raiseChanged('layout', old, value);
      this.requestUpdate();
    }
  }

  /**
   * Perform all invalid layouts.
   * If the optional argument is true,
   * this will perform all of the layouts ({@link Diagram#layout} and all {@link Group#layout}s),
   * not just the invalid ones.
   *
   * Under normal circumstances you should not need to call this method,
   * because layouts will be performed automatically after they become invalid.
   * However you may have disabled automatic layouts by setting {@link Layout#isInitial}
   * and/or {@link Layout#isOngoing} to false, or by restricting a Part's {@link Part#layoutConditions}.
   * If that is the case you might call this method
   * (perhaps due to a user command) to perform the layout at a time of your choosing.
   * @export
   * @this {Diagram}
   * @param {boolean=} invalidateAll If true, this will explicitly set {@link Layout#isValidLayout} to false on each {@link Layout} in the diagram.
   */
  public layoutDiagram(invalidateAll?: boolean): void {
    this.validateParts();
    if (invalidateAll) this.invalidateAllLayouts(true);
    this._doNonRealtimeLayouts = true;  // perform any automatic layouts (pass false to updateAllLayouts)
    this.doUpdate(); // in order to ultimately call updateAllLayouts(false)
  }

  /**
   * @hidden @internal
   * Called by layoutDiagram, addModeledParts, addModeledLinks
   * @this {Diagram}
   * @param {boolean} unconditional
   */
  public invalidateAllLayouts(unconditional: boolean): void {
    const git = this._topLevelGroups.iterator;
    while (git.next()) {
      const g = git.value;
      this.invalidateGroupLayouts(g, unconditional);
    }
    if (this.layout !== null) {
      if (unconditional) {
        this.layout.isValidLayout = false;
      } else {
        this.layout.invalidateLayout();
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Group} grp
   * @param {boolean} unconditional
   */
  private invalidateGroupLayouts(grp: Group, unconditional: boolean): void {
    if (grp === null) return;
    const git = grp._nestedGroups.iterator;
    while (git.next()) {
      const g = git.value;
      this.invalidateGroupLayouts(g, unconditional);
    }
    if (grp.layout !== null) {
      if (unconditional) {
        grp.layout.isValidLayout = false;
      } else {
        grp.layout.invalidateLayout();
      }
    }
  }

  /**
   * @hidden @internal
   * Layout all invalid layouts in this Diagram. Should only be called within doUpdate, since it resets _anyLayoutInvalid.
   * @this {Diagram}
   * @param {boolean} checkRealtime if true, only do layouts with isRealtime = true
   */
  private updateAllLayouts(checkRealtime: boolean): void {
    if (!this._anyLayoutInvalid) return;
    if (this._delaysLayout) return;

    // temporarily set this to true to allow layout implementations to modify model data
    // and have such modifications automatically update target bindings
    const oldUpdatingModel = this.isUpdatingModel;
    this.isUpdatingModel = true;
    const reason = 'Layout';
    const tl = this.undoManager.transactionLevel;
    const lay = this.layout;
    try {
      if (tl === 0) this.startTransaction(reason);

      // mechanism for animating layouts that do not happen inside a transaction
      const AM = this.animationManager;

      if (tl <= 1 && !AM.isAnimating && !AM._pendingAnimation) {
        // Don't prepare animations during realtime layout updates
        if (!checkRealtime) AM.prepareAutomaticAnimation(reason);
      }

      this._anyLayoutInvalid = false;  // reset flag for optimization
      const git = this._topLevelGroups.iterator;
      while (git.next()) {
        const g = git.value;
        this.updateGroupLayouts(g, checkRealtime, tl);
      }

      if (!lay.isValidLayout) {
        // realtime layouts should still happen if this Layout transaction is a top-level transaction
        // (if there's no other transaction ongoing)
        if ((!checkRealtime || lay.isRealtime || lay.isRealtime === null) || tl === 0) {
          lay.doLayout(this);
          this.validateParts();  // might invalidate layout!
          lay.isValidLayout = true;  // make sure it's now valid
        } else {
          this._anyLayoutInvalid = true;
        }
      }
    } finally {
      if (tl === 0) this.commitTransaction(reason);
      this._anyLayoutInvalid = !lay.isValidLayout;  // ??? wrong, but for now needed to avoid duplicate Transaction bug
      this.isUpdatingModel = oldUpdatingModel;
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Group} grp
   * @param {boolean} checkRealtime
   * @param {number} tl
   */
  private updateGroupLayouts(grp: Group, checkRealtime: boolean, tl: number): void {
    if (grp === null) return;

    const git = grp._nestedGroups.iterator;
    while (git.next()) {
      const g = git.value;
      this.updateGroupLayouts(g, checkRealtime, tl);
    }

    const lay = grp.layout;
    if (lay !== null && !lay.isValidLayout) {
      if ((!checkRealtime || lay.isRealtime) || tl === 0) {
        grp.needsLocation = !grp.location.isReal();  // hack for Layout.needsLocation
        lay.doLayout(grp);
        grp.invalidateLayout(Part.LayoutGroupLayout);  // maybe invalidate the parent Layout
        // now make sure all of GRP's members have valid size and position
        this.measureArrangeGroup(grp);  // might invalidate layout!
        lay.isValidLayout = true;  // make sure it's now valid
      } else {
        this._anyLayoutInvalid = true;
      }
    }
  }

  /**
   * Gets or sets whether the Diagram tree structure is defined by
   * links going from the parent node to their children, or vice-versa.
   * By default this property is true: links go from the parent node to the child node.
   */
  /** @export */ get isTreePathToChildren(): boolean { return this._isTreePathToChildren; }
  /** @export */ set isTreePathToChildren(value: boolean) {
    const old = this._isTreePathToChildren;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Diagram, 'isTreePathToChildren');
      this._isTreePathToChildren = value;
      this.raiseChanged('isTreePathToChildren', old, value);
      // need to update Node.isTreeLeaf on every Node
      if (!this.undoManager.isUndoingRedoing) {
        const nit = this.nodes;
        while (nit.next()) {
          const node = nit.value;
          node.updateTreeLeaf();
        }
      }
    }
  }

  /**
   * Returns an iterator of all top-level {@link Node}s that have no tree parents.
   *
   * This is useful for when you want to traverse the diagram's graph by starting at the root of each tree,
   * assuming that the diagram consists of one tree or a forest of trees.
   * @export
   * @this {Diagram}
   * @return {Iterator.<Node>}
   * @since 1.2
   */
  public findTreeRoots(): Iterator<Node> {
    // this could be implemented more efficiently
    const roots = new List<Node>();
    const it = this.nodes;
    while (it.next()) {
      const node = it.value;
      if (node.isTopLevel && node.findTreeParentLink() === null) {
        roots.add(node);
      }
    }
    return roots.iterator;
  }

  /**
   * @hidden @internal
   * Gets or sets the conditions under which "child" nodes become not visible.
   * By default this property is {@link Diagram.TreeParentCollapsed}: only the true tree parent node's
   * {@link Node#isTreeExpanded} property controls the visibility of a child node.
   *
   * You can set this to {@link Diagram.AllParentsCollapsed}, to require all of a node's
   * "parents" to be collapsed in order for the node to become hidden.
   * Caution: cycles in the graph structure may cause all of the nodes in the cycle to become hidden.
   */
  /** @export */ get treeCollapsePolicy(): EnumValue { return this._treeCollapsePolicy; }
  /** @export */ set treeCollapsePolicy(value: EnumValue) {
    const old = this._treeCollapsePolicy;
    if (old !== value) {
      if (value !== Diagram.TreeParentCollapsed && value !== Diagram.AllParentsCollapsed && value !== Diagram.AnyParentsCollapsed) {
        Util.throwError('Unknown Diagram.treeCollapsePolicy: ' + value);
      }
      this._treeCollapsePolicy = value;
      this.raiseChanged('treeCollapsePolicy', old, value);
    }
  }

  /**
   * @hidden @internal
   */
  /* noexport */get isCollapsingExpanding(): boolean { return this._isCollapsingExpanding; }
  set isCollapsingExpanding(value: boolean) { this._isCollapsingExpanding = value; }

  /**
   * @hidden @internal
   * @this {Diagram}
   */
  private _setupDiagramEvents(): void {
    const handlers = new Map<string, List<((a: DiagramEvent) => void)>>();
    const names = new Map<string, string>();
    // store the same List of event handler functions for both the regular name and the downcased name
    // also be able to canonicalize the name by mapping the downcased name to the "normally" case name
    /**
     * @hidden @internal
     * @param {string} n
     */
    function defineListener(n: string): void {
      const lname = n.toLowerCase();
      const l = new List<((a: DiagramEvent) => void)>();
      handlers.add(n, l);
      handlers.add(lname, l);
      names.add(n, n);
      names.add(lname, n);
    }
    defineListener('AnimationStarting');
    defineListener('AnimationFinished');
    defineListener('BackgroundSingleClicked');
    defineListener('BackgroundDoubleClicked');
    defineListener('BackgroundContextClicked');
    defineListener('ClipboardChanged');
    defineListener('ClipboardPasted');
    defineListener('DocumentBoundsChanged');
    defineListener('ExternalObjectsDropped');
    defineListener('GainedFocus');
    defineListener('InitialLayoutCompleted');
    defineListener('LayoutCompleted');
    defineListener('LinkDrawn');
    defineListener('LinkRelinked');
    defineListener('LinkReshaped');
    defineListener('LostFocus');
    defineListener('Modified');
    defineListener('ObjectSingleClicked');
    defineListener('ObjectDoubleClicked');
    defineListener('ObjectContextClicked');
    defineListener('PartCreated');
    // defineListener('PartReshaped');
    defineListener('PartResized');
    defineListener('PartRotated');
    defineListener('SelectionMoved');
    defineListener('SelectionCopied');
    defineListener('SelectionDeleting');
    defineListener('SelectionDeleted');
    defineListener('SelectionGrouped');
    defineListener('SelectionUngrouped');
    defineListener('ChangingSelection');
    defineListener('ChangedSelection');
    defineListener('SubGraphCollapsed');
    defineListener('SubGraphExpanded');
    defineListener('TextEdited');
    defineListener('TreeCollapsed');
    defineListener('TreeExpanded');
    defineListener('ViewportBoundsChanged');
    defineListener('InvalidateDraw');  // undocumented

    this._diagramListeners = handlers;

    this._diagramEventNames = names;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {string} name
   * @return {string|null} may return null if the name is unknown.
   */
  private canonicalizeDiagramEventName(name: string): string | null {
    // avoid converting to lower case unnecessarily
    const canon = this._diagramEventNames.getValue(name);
    if (canon !== null) return canon;
    return this._diagramEventNames.getValue(name.toLowerCase());
  }

  /**
   * @hidden @internal
   * This throws an error if the {@link DiagramEvent} name is unknown.
   * @this {Diagram}
   * @param {string} name
   * @return {List.<function(DiagramEvent)>}  a List of event handler functions.
   */
  private getDiagramListeners(name: string): List<((a: DiagramEvent) => void)> {
    let list = this._diagramListeners.getValue(name);
    if (list !== null) return list;
    const lname = name.toLowerCase();
    list = this._diagramListeners.getValue(lname);
    if (list !== null) return list;
    Util.throwError('Unknown DiagramEvent name: ' + name);
    return null;
  }

  /**
   * Register an event handler that is called when there is a {@link DiagramEvent} of a given name.
   *
   * See the {@link DiagramEvent} documentation for a complete listing of diagram event names and their purposes.
   * @export
   * @this {Diagram}
   * @param {string} name the name is normally capitalized, but this method uses case-insensitive comparison.
   * @param {function(DiagramEvent)} listener a function that takes a {@link DiagramEvent} as its argument.
   * @see {@link #removeDiagramListener}
   */
  public addDiagramListener(name: DiagramEventName, listener: ((e: DiagramEvent) => void)): void {
    Util.checkPrimitive(name, 'string', Diagram, 'addDiagramListener:name');
    Util.checkPrimitive(listener, 'function', Diagram, 'addDiagramListener:listener');
    const list = this.getDiagramListeners(name);
    if (list !== null) list.add(listener);
  }

  /**
   * Unregister a {@link DiagramEvent} handler.
   *
   * See the documentation for {@link DiagramEvent} for a complete listing of diagram event names and their purposes.
   * @export
   * @this {Diagram}
   * @param {string} name the name is normally capitalized, but this method uses case-insensitive comparison.
   * @param {function(DiagramEvent)} listener a function that takes a {@link DiagramEvent} as its argument.
   * @see {@link #addDiagramListener}
   */
  public removeDiagramListener(name: DiagramEventName, listener: ((e: DiagramEvent) => void)): void {
    Util.checkPrimitive(name, 'string', Diagram, 'removeDiagramListener:name');
    Util.checkPrimitive(listener, 'function', Diagram, 'addDiagramListener:listener');
    const list = this.getDiagramListeners(name);
    if (list !== null) list.remove(listener);
  }

  // * Notify any {@link DiagramEvent} listeners by calling all event handlers registered by {@link #addDiagramListener}.

  /**
   * @hidden
   * Undocumented.
   * @export
   * @this {Diagram}
   * @param {string} name the name is normally capitalized, but this method uses case-insensitive comparison.
   * @param {Object=} obj an optional subject of the event.
   * @param {*=} param an optional parameter describing the change to the subject of the event.
   * @see {@link #addDiagramListener}
   * @see {@link #removeDiagramListener}
   */
  public raiseDiagramEvent(name: DiagramEventName, obj?: ObjectData, param?: any) {
    if (Debug) Util.checkPrimitive(name, 'string', Diagram, 'raiseDiagramEvent:name');
    const list = this.getDiagramListeners(name);
    const e = new DiagramEvent();
    e.diagram = this;
    const canonName = this.canonicalizeDiagramEventName(name);  // make sure the name is cased correctly
    if (canonName !== null) e.name = canonName;
    if (obj !== undefined) e.subject = obj;
    if (param !== undefined) e.parameter = param;
    const l = list.length;
    if (l === 1) {
      const func = list.elt(0);
      func(e);
    } else if (l !== 0) {
      // copy list to avoid problems when listeners add/remove listeners
      const arr = list.toArray();
      for (let i = 0; i < l; i++) {
        const func = arr[i];
        func(e);
      }
    }
  }

  // AvoidsNodes support

  /**
   * @hidden @internal
   * instead of Link.maybeAvoidsNodes having to know about diagram state
   * like animation manager status or the current tool,
   * it calls this instead.
   * @this {Diagram}
   * Called by Link.maybeAvoidsNodes.
   * @return {boolean}
   */
  public canAvoidNodes(): boolean {
    if (this.animationManager.isAnimating) return false;
    const current = this.currentTool as any;
    // suspend must be true
    if (current === this.toolManager.findTool('Dragging')) {
      return (!this._suspendedRouting || current['isComplexRoutingRealtime']);
    }
    // otherwise we are OK to avoid nodes
    return true;
  }

  /**
   * @hidden @internal
   * Returns true if there are no avoidable nodes within or intersecting the given rectangular area.
   * @export
   * @this {Diagram}
   * @param {Rect} r a Rect in document coordinates
   * @param {Node=} skip an optional Node to be skipped
   * @return {boolean}
   */
  public isUnoccupied(r: Rect, skip?: Node): boolean {
    if (skip === undefined) skip = null;
    const positions = this.getPositions(false, null, skip);
    return positions.isUnoccupied(r.x, r.y, r.width, r.height);
  }

  /**
   * @hidden @internal
   * @expose
   * @return {Rect} in document coordinates
   */
  public computeOccupiedArea(): Rect {
    if (this.isVirtualized) return this.viewportBounds.copy();
    if (this._invalidDocBounds) {
      // ?? This method occurs inside of validation,
      //    and computing the bounds in the middle of validation is a problem
      //    So instead we avoid validating further.
      //    This could be optimized so that addOrthoPoints/getPositions only occurs after all objects are normally validated
      return this.computeBoundsNoValidate();
    } else {
      return this.documentBounds.copy();
    }
  }

  /**
   * @hidden @internal
   * Internal, used by Link.addOrthoPoints to get and initialize a PositionArray
   * representing the occupied areas of this Diagram.
   * @this {Diagram}
   * @param {boolean} clearunoccupied
   * @param {Group} group limit avoidable nodes to members of this Group
   * @param {Node} skip a Node (presumably an avoidable node) whose bounds are not declared "occupied"
   * @return {PositionArray}
   */
  public getPositions(clearunoccupied: boolean, group: Group, skip: Node): PositionArray {
    if (this._positions === null) {
      this._positions = new PositionArray();
    }
    const subgraph = group;
    if (this._positions.invalid || this._positions.group !== subgraph || this._positions.skip !== skip) {
      if (subgraph === null) {
        const docbounds = this.computeOccupiedArea();
        docbounds.inflate(100, 100);
        // make sure the array is big enough, and set all cells to MAX
        this._positions.initialize(docbounds);
        const tempRect = Rect.alloc();
        // now for each object that matters, set the corresponding cell to OCCUPIED
        const nit = this.nodes;
        while (nit.next()) {
          const node = nit.value;
          const layer = node.layer;
          if (layer === null || !layer.visible || layer.isTemporary) continue;
          this.getPositions1(node, skip, tempRect);
        }
        Rect.free(tempRect);
      } else {
        if (subgraph.memberParts.count > 0) {
          const sgbounds = this.computePartsBounds(subgraph.memberParts, false);
          sgbounds.inflate(20, 20);
          this._positions.initialize(sgbounds);
        } // otherwise there are no member Nodes and don't need to call initialize
        const tempRect = Rect.alloc();
        const mit = subgraph.memberParts;
        while (mit.next()) {
          const n = mit.value;
          if (n instanceof Node) {
            this.getPositions1(n, skip, tempRect);
          }
        }
        Rect.free(tempRect);
      }
      this._positions.group = subgraph;
      this._positions.skip = skip;
      this._positions.invalid = false;
    } else if (clearunoccupied) {
      // assume the OCCUPIED cells should be left alone, but clear out
      // any earlier distance calculations
      this._positions.clearAllUnoccupied();
    }
    return this._positions;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Node} node
   * @param {Node} skip
   * @param {Rect} temp
   */
  private getPositions1(node: Node, skip: Node, temp: Rect): void {
    if (node === skip) return;
    if (node.isVisible() && node.canAvoid()) {
      const rect = node.getAvoidableRect(temp);
      const cellx = this._positions.cellWidth;
      const celly = this._positions.cellHeight;
      const endx = rect.x + rect.width;
      const endy = rect.y + rect.height;
      for (let i = rect.x; i < endx; i += cellx) {
        for (let j = rect.y; j < endy; j += celly) {
          this._positions.setOccupied(i, j);
        }
        this._positions.setOccupied(i, endy);
      }
      for (let j = rect.y; j < endy; j += celly) {
        this._positions.setOccupied(endx, j);
      }
      this._positions.setOccupied(endx, endy);
    } else {
      if (node instanceof Group) {
        const it = node.memberParts;
        while (it.next()) {
          const o = it.value;
          if (o instanceof Node) {
            this.getPositions1(o, skip, temp);
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Called by Node and Part
   * @this {Diagram}
   * @param {Node=} node
   */
  public invalidatePositionArray(node?: Node): void {
    if (this._positions !== null && !this._positions.invalid) {
      if (node === undefined) node = null;
      if (node === null || node.canAvoid()) {
        this._positions.invalid = true;
      }
    }
  }

  // common methods for simulated drag-and-drop


  // Autoscrolling

  /**
   * @hidden @internal
   * Gets or sets number of milliseconds between autoscroll events.
   * The default value is 250.
   */
  /** @export */ get autoScrollInterval(): number { return this._autoScrollInterval; }
  /** @export */ set autoScrollInterval(value: number) {
    const old = this._autoScrollInterval;
    Util.checkRealNumber(value, Diagram, 'scale');
    if (old !== value) {
      this._autoScrollInterval = value;
      this.raiseChanged('autoScrollInterval', old, value);
    }
  }

  /**
   * Gets or sets the Margin that describes the Diagram's autoScrollRegion. The default value is a Margin of 16 on all sides.
   * When the mouse drag point is within this region on the left or right sides,
   * the view will automatically scroll horizontally in that direction.  When the point is within
   * the region on the top or bottom, the view will automatically scroll
   * vertically in that direction.  You can specify a distance of zero to disable autoscrolling
   * in a direction; a value of 0,0,0,0 turns off autoscrolling altogether.
   */
  /** @export */ get autoScrollRegion(): MarginLike { return this._autoScrollRegion; }
  /** @export */ set autoScrollRegion(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Diagram, 'autoScrollRegion');
    }
    const old = this._autoScrollRegion;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._autoScrollRegion = value;
      this.invalidateDocumentBounds();
      this.raiseChanged('autoScrollRegion', old, value);
    }
  }

  // * Called by DraggingTool and LinkingTool

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   * @param {Point} viewPt
   */
  public doAutoScroll(viewPt: Point): void {
    this._autoScrollPoint.assign(viewPt);
    const newdocpos = this.computeAutoScrollPosition(this._autoScrollPoint);
    if (!newdocpos.equalsApprox(this.position)) {
      this.startAutoScroll();
    } else {
      this.stopAutoScroll();
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   */
  public startAutoScroll(): void {
    if (this._autoScrollTimer !== -1) return;
    const interval = this._autoScrollInterval;
    // starts automatically
    const diagram = this;
    this._autoScrollTimer = Util.setTimeout(function() {
      if (diagram._autoScrollTimer !== -1) {
        diagram.stopAutoScroll();
        const e = diagram.lastInput.event;
        if (e === null) return;

        const newdocpos = diagram.computeAutoScrollPosition(diagram._autoScrollPoint);
        if (!newdocpos.equalsApprox(diagram.position)) {
          diagram.position = newdocpos;
          // adjust due to new diagram position
          diagram.lastInput.documentPoint = diagram.transformViewToDoc(diagram._autoScrollPoint);
          // need to fake a mouse move to keep dragging while autoscrolling
          // ?? moved to dragging tool, does not seem necessary:
          /* const target = diagram._findTargetDiagram(e);
             if (!diagram.simulatedMouseMove(e, null, target)) */
          diagram.doMouseMove();

          diagram._invalidDocBounds = true;
          // Usually we just pass in diagram.computeBounds, but here we want to include all bounds-areas
          // explored by auto-scroll, so that there are no jarring scrolls as users explore more area
          diagram.setDocumentBounds(diagram.documentBounds.copy().unionRect(diagram.computeBounds()));
          diagram._invalidDraw = true;
          diagram.maybeUpdate(); // DoUpdateDiagramBounds //?? synchronous
          diagram.startAutoScroll();
        }
      }
    }, interval);
  }

  // * Stop any ongoing auto-scroll action.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Diagram}
   */
  public stopAutoScroll(): void {
    if (this._autoScrollTimer === -1) return;
    Util.clearTimeout(this._autoScrollTimer); // stops
    this._autoScrollTimer = -1;
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Point} viewPnt
   * @return {Point}
   */
  public computeAutoScrollPosition(viewPnt: Point): Point {
    let docpos = this.position;
    const margin = this._autoScrollRegion;
    // no autoscroll region? just return current Position
    if (margin.top <= 0 && margin.left <= 0 && margin.right <= 0 && margin.bottom <= 0) return docpos;
    // get the visible part of the panel
    const vp = this.viewportBounds;
    const scale = this.scale;
    const dispRect = Rect.allocAt(0, 0, vp.width * scale, vp.height * scale);
    // get the current panel position, in element coordinates
    const viewpos = Point.allocAt(0, 0);
    // and compute the new position if viewPnt is near any of the four sides, within the margin
    if (viewPnt.x >= dispRect.x && viewPnt.x < dispRect.x + margin.left) {
      let deltaX = Math.max(this.scrollHorizontalLineChange, 1);
      deltaX = deltaX | 0;
      viewpos.x -= deltaX;
      if (viewPnt.x < dispRect.x + margin.left / 2) {
        viewpos.x -= deltaX;
      }
      if (viewPnt.x < dispRect.x + margin.left / 4) {
        viewpos.x -= 4 * deltaX;
      }
    } else if (viewPnt.x <= dispRect.x + dispRect.width && viewPnt.x > dispRect.x + dispRect.width - margin.right) {
      let deltaX = Math.max(this.scrollHorizontalLineChange, 1);
      deltaX = deltaX | 0;
      viewpos.x += deltaX;
      if (viewPnt.x > dispRect.x + dispRect.width - margin.right / 2) {
        viewpos.x += deltaX;
      }
      if (viewPnt.x > dispRect.x + dispRect.width - margin.right / 4) {
        viewpos.x += 4 * deltaX;
      }
    }
    if (viewPnt.y >= dispRect.y && viewPnt.y < dispRect.y + margin.top) {
      let deltaY = Math.max(this.scrollVerticalLineChange, 1);
      deltaY = deltaY | 0;
      viewpos.y -= deltaY;
      if (viewPnt.y < dispRect.y + margin.top / 2) {
        viewpos.y -= deltaY;
      }
      if (viewPnt.y < dispRect.y + margin.top / 4) {
        viewpos.y -= 4 * deltaY;
      }
    } else if (viewPnt.y <= dispRect.y + dispRect.height && viewPnt.y > dispRect.y + dispRect.height - margin.bottom) {
      let deltaY = Math.max(this.scrollVerticalLineChange, 1);
      deltaY = deltaY | 0;
      viewpos.y += deltaY;
      if (viewPnt.y > dispRect.y + dispRect.height - margin.bottom / 2) {
        viewpos.y += deltaY;
      }
      if (viewPnt.y > dispRect.y + dispRect.height - margin.bottom / 4) {
        viewpos.y += 4 * deltaY;
      }
    }
    // don't bother changing DOCPOS if VIEWPOS hasn't changed
    // Also avoids differences due to round-off errors!
    if (!viewpos.equalsApprox(Point.Origin)) {
      // but return a Point in document coordinates
      docpos = new Point(docpos.x + viewpos.x / scale, docpos.y + viewpos.y / scale);
    }
    Rect.free(dispRect);
    Point.free(viewpos);
    return docpos;
  }

  /**
   * Create an SVGElement that contains a SVG rendering of the current Diagram.
   *
   * By default this method returns a snapshot of the visible diagram, but optional arguments give more options.
   * <p class="boxread">
   * See the page on <a href="../../intro/makingSVG.html">Making SVG</a> for usage examples.
   * See the <a href="../../samples/minimalSVG.html">Minimal SVG Download sample</a>,
   * which also demonstrates downloading an SVG file without involving a web server.
   *
   * See {@link #makeImageData} for an explanation of possible options that are shared by both methods.
   * Additional SVG-specific options for this method:
   *   - **elementFinished:**
   *     A function with two arguments, GraphObject and SVGElement.
   *     As the SVG elements are created representing each graph object, this function is called on them,
   *     allowing you to modify the SVG as it is being built, to assign stylesheets, IDs, etc. Example:
   *     ```js
   *     elementFinished: function(graphObject, SVGElement) {
   *       // set something on every SVG element that represents a GoJS TextBlock
   *       if (graphObject instanceof go.TextBlock) SVGElement.setAttribute(...);
   *     }
   *     ```
   *
   * At the current time methods such as {@link Diagram#makeImage},
   * {@link Diagram#makeImageData} and {@link Diagram#makeSvg} do not work on {@link Overview}s.
   * @export
   * @this {Diagram}
   * @param {any=} options a JavaScript object detailing optional arguments for SVG creation.
   * @return {SVGElement}
   * @see {@link #makeImage}
   */
  public makeSvg(options?: SvgRendererOptions): SVGElement {
    return null; // defined in svgContext
  }

  /**
   * @hidden @internal
   * @export
   * @this {Diagram}
   */
  public makeSVG(options?: SvgRendererOptions): SVGElement {
    return null; // defined in svgContext
  }

  /**
   * @hidden
   * Add a renderer to the Diagram. This property is only used when building GoJS from source.
   *
   * When building from source, to include SVG rendering functionality for {@link Diagram#makeSvg},
   * you must add the SVGSurface renderer with the name 'SVG': `myDiagram.addRenderer('SVG', new SVGSurface(myDiagram, document));`
   *
   * @export
   * @param {string} name Name of the renderer
   * @param {ISurface} surface Rendering class, either SVGSurface or CanvasSurface.
   * @this {Diagram}
   */
  public addRenderer(name: string, surface: ISurface) {
    this._renderers.add(name, surface);
  }

  /**
   * @hidden @internal
   *
   * Remove a renderer.
   * @param {string} name
   * @this {Diagram}
   */
  public removeRenderer(name: string) {
    this._renderers.remove(name);
  }

  /**
   * @hidden @internal
   * Potentially set up a callback for makeImage/Data or makeSVG
   * Only public because makeSVG is defined within svgContext.ts
   * @this {Diagram}
   */
  public setupTimers(funcToCall: Function, options: ImageRendererOptions) {
    const callbackFunc = options['callback'];
    let allLoaded = true;
    const pics = this._picturesMap;
    const itr = pics.iterator;
    while (itr.next()) {
      if (!itr.value[0]._loaded) {
        allLoaded = false;
        break;
      }
    }
    if (typeof callbackFunc !== 'function' || allLoaded) return funcToCall(options, callbackFunc, this);
    const self = this;
    function tick() {
      const time = +new Date();
      allLoaded = true;
      itr.reset();
      while (itr.next()) {
        if (!itr.value[0]._loaded) {
          allLoaded = false;
          break;
        }
      }

      if (allLoaded || (time - d1) > callbackTimeout) {
        // either no more loading or timed out
        funcToCall(options, callbackFunc, self);
        return;
      }
      // not done, request again
      root.requestAnimationFrame(tick);
    }

    // Some images still need loading, wait on drawing
    const callbackTimeout = options['callbackTimeout'] || 300;
    const d1 = +new Date();
    root.requestAnimationFrame(function() { tick(); });
    return null;
  }

  /**
   *
   * Create an HTMLImageElement that contains a bitmap of the current Diagram.
   * This method is just a convenience function that creates an image,
   * sets its source to the returned string of {@link #makeImageData},
   * and returns a reference to that Image.
   *
   * See {@link #makeImageData} for a complete explanation of possible options.
   *
   * By default this method returns a snapshot of the visible diagram, but optional arguments give more options.
   *
   * At the current time methods such as {@link Diagram#makeImage},
   * {@link Diagram#makeImageData} and {@link Diagram#makeSvg} do not work on {@link Overview}s.
   * @export
   * @this {Diagram}
   * @param {any=} options a JavaScript object detailing optional arguments for image creation, to be passed to {@link #makeImageData}.
   * @return {HTMLImageElement | null} An HTML Image element, or null if a callback is specified, or null if there is no DOM.
   * @see {@link #makeImageData}
   * @see {@link #makeSvg}
   */
  public makeImage(options?: ImageRendererOptions): HTMLImageElement | null {
    if (!Diagram.isUsingDOM()) return null;
    if (options === undefined) options = new PropertyCollection() as any;
    options['returnType'] = 'Image';
    return this.makeImageData(options) as HTMLImageElement | null;
  }

  /**
   * Create a bitmap of the current Diagram encoded as a base64 string, or returned as an ImageData object.
   * This method uses the toDataURL method of the HTMLCanvasElement to create the data URL,
   * or the getImageData method of the Canvas Context.
   * Unlike toDataURL, this method will not throw an error if cross-domain images
   * were drawn on the canvas, instead it will return a data URL of a bitmap with those images omitted.
   *
   * A simple example:
   * ```js
   * myDiagram.makeImageData({
   *   scale: 1.5,
   *   size: new go.Size(100,100)
   * });
   * ```
   * <p class="boxread">
   * See the page on <a href="../../intro/makingImages.html">Making Images</a> for more usage examples.
   *
   * At the current time methods such as {@link Diagram#makeImage},
   * {@link Diagram#makeImageData} and {@link Diagram#makeSvg} do not work on {@link Overview}s.
   * @export
   * @this {Diagram}
   * @param {any=} options a JavaScript object detailing optional arguments for image creation.
   * Rendering options for both images and SVG:
   *   - **size:**
   *     The size of the created image, as a {@link Size}, limited by the **maxSize** property.
   *     If no **scale** or **position** is specified then the diagram will be scaled to fit the given size.
   *     If you set a size, you should also set a **position**. If you are scaling the diagram, you may also want to scale the size.
   *   - **scale:**
   *     The scale of the diagram. If **scale** is specified and **size** is not, the resulting image will be sized to uniformly
   *     fit the space needed for the given scale.
   *     Can be constrained by the **maxSize** property. A scale value of `NaN` will
   *     automatically scale ot fit within the maxSize, but may be smaller, with a maximum computed scale of 1.
   *   - **maxSize:**
   *     The maximum size of the created image, as a {@link Size}. The default value is `(Infinity, Infinity)` for SVG and `(2000, 2000)` for images.
   *     This is typically used when **scale** is specified.
   *   - **position:**
   *     The position of the diagram, as a {@link Point}.
   *     By default this is the position of {@link Diagram#documentBounds} with the {@link Diagram#padding} removed.
   *     If a specific parts collection is used, by default this is the top-left diagram position of their collective bounds.
   *     If you set a `position`, you should also set a `size`.
   *   - **parts:**
   *     An iterator of GraphObjects, typically Parts, such as one from {@link Diagram#selection} or {@link Layer#parts}.
   *     If GraphObjects are specified their containing Part will be drawn. By default all Parts are drawn except temporary parts (see **showTemporary**).
   *   - **padding:**
   *     A {@link Margin} (or number) to pad the image with. If a size is specified,
   *     the padding will not increase the image size, it will only offset the Diagram contents within the image.
   *     The default value is a padding of `1`.
   *   - **background:**
   *     A valid CSS color to replace the default (transparent) canvas background. Any padding area is also colored.
   *   - **showTemporary:**
   *     A boolean value, defaulting to false,
   *     that determines whether or not temporary objects such as adornments are included in the image.
   *   - **showGrid:**
   *     A boolean value, defaulting to the value of **showTemporary**,
   *     that determines whether or not the Grid Layer (containing {@link Diagram#grid}) is included in the image
   *     regardless of the value of **showTemporary**.
   *     This is useful if you want to include the grid but not adornments, or vice versa.
   *   - **document:**
   *     An HTML Document, defaulting to `window.document` (or the root object in other contexts)
   *     This may be useful to set if you intend your Image or SVG to be opened in a new window.
   *
   * Additional image-specific arguments (not for SVG):
   *   - **type:**
   *     The optional MIME type of the image. Valid values are typically `"image/png"` and `"image/jpeg"`.
   *     Some browsers allow `"image/webp"`. The default value is `"image/png"`, and unrecognized values will defer to the default.
   *   - **returnType:**
   *     The optional return type of the image data. Valid values are `"ImageData"`, `"string"`, and `"blob"`.
   *     The `"string"` option returns a base64 string representation of the image.
   *     The `"ImageData"` option returns an `ImageData` object representation of the image.
   *     The `"Image"` option returns an `HTMLImageElement` using `ImageData` as the `HTMLImageElement.src`.
   *     The `"blob"` option requires that the **callback** property is also defined.
   *     The default value is `"string"`, and unrecognized values will return a string.
   *   - **callback:**
   *     The function to call when an image is finished creation. It has one argument, which is of the type specified by the value of the **returnType**.
   *     If provided, call the callback when finished instead of returning immediately. This can be useful if you need to wait for image assets to load.
   *     This also respects the **callbackTimeout**.
   *     This argument is necessary if the **returnType** is `"blob"`, however a callback can be used with any **returnType**.
   *     See the <a href="../../samples/minimalBlob.html">Minimal Image Blob Download sample</a> for an example usage,
   *     which also demonstrates downloading an image file without involving a web server.
   *   - **callbackTimeout:**
   *     If a **callback** is specified, the additional amount of time in milliseconds a call will wait before completeing. Right now, it will only wait if
   *     image assets in the Diagram are not yet loaded. Default is 300 (milliseconds).
   *   - **details:**
   *     The optional details to pass to the HTMLCanvasElement's toDataURL function.
   *     If the type is `"image/jpeg"` then this can be a number from `0` to `1`, inclusive, describing the desired jpeg quality.
   *
   * @return {ImageData|string|null} An ImageData, or a base64-encoded string describing an image, or null if a callback is specified.
   * @see {@link #makeImage}
   */
  public makeImageData(options?: ImageRendererOptions): HTMLImageElement | ImageData | string | null {
    if (options === undefined) options = new PropertyCollection() as any;
    return this.setupTimers(this.makeImageData2, options);
  }

  /**
   * @hidden @internal
   * @param options
   * @param callbackFunc
   */
  public makeImageData2(options: ImageRendererOptions,
                        callbackFunc: ((result: any) => void) | null, diagram: Diagram): HTMLImageElement | ImageData | string | null {
    const context = diagram.drawData(options, 'canvas', null) as CanvasSurfaceContext;
    if (context === null) return null;
    const canvas = context._backingContext.canvas;
    let returnValue: HTMLImageElement | ImageData | string | null = null;
    if (canvas !== null) {
      let returnType = options['returnType'];
      if (returnType === undefined) returnType = 'string';
      else returnType = returnType.toLowerCase();
      switch (returnType) {
        case 'imagedata':
          returnValue = context.getImageData(0, 0, canvas.width, canvas.height);
          break;
        case 'image':
          const doc = options['document'] || document;
          const img = doc.createElement('img');
          img.src = canvas.toDataURL(options['type'], options['details']);
          returnValue = img;
          break;
        case 'blob': {
          const canv = canvas;
          if (typeof callbackFunc !== 'function') {
            Util.throwError('Error: Diagram.makeImageData called with "returnType: toBlob", but no required "callback" function property defined.');
            returnValue = '';
          }
          if (typeof canv['toBlob'] === 'function') {
            canv['toBlob'](callbackFunc, options['type'], options['details']);
            return 'toBlob';
          } else if (typeof canv['msToBlob'] === 'function') {
            callbackFunc(canv['msToBlob']());
            return 'msToBlob';
          } else {
            callbackFunc(null);
            return null;
          }
        }
        case 'string':
        default:
          returnValue = canvas.toDataURL(options['type'], options['details']);
      }
    }
    if (typeof callbackFunc === 'function') {
      callbackFunc(returnValue);
      return null;
    }
    return returnValue;
  }

  /** @hidden @internal */
  private static _warnedAboutMakeImageSize = false;

  /**
   * @hidden @internal
   * @return {IContext} called by makeSvg, or makeImage/data
   * @param {Object} properties
   * @param {string} contextType
   * @param {SVGSurface|null} svgSurface
   */
  public drawData(properties: ObjectData, contextType: string, svgSurface: SVGSurface | null): IContext {
    this.animationManager.stopAnimation();
    this.maybeUpdate();
    const can = this._canvas;
    if (can === null) return null;
    if (typeof properties !== 'object') Util.throwError('properties argument must be an Object.');
    let defaultMaxSize = false;

    let size = properties['size'] || null;
    let scale = properties['scale'] || null;
    if (properties['scale'] !== undefined && isNaN(properties['scale'])) scale = 'NaN'; // This is a string to allow "if (scale)"s below to return true

    let maxSize = properties['maxSize'];
    if (properties['maxSize'] === undefined) {
      defaultMaxSize = true;
      if (contextType === 'SVG') {
        maxSize = new Size(Infinity, Infinity);
      } else {
        maxSize = new Size(2000, 2000);
      }
    }
    const position = properties['position'] || null;
    const parts = properties['parts'] || null;
    let padding = properties['padding'] === undefined ? 1 : properties['padding'];  // allow zero
    const background = properties['background'] || null;
    let omitTemporary = properties['omitTemporary'];
    if (omitTemporary === undefined) omitTemporary = true;
    const doc = properties['document'] || document;
    const elementFinished = properties['elementFinished'] || null;

    let showTemporary = properties['showTemporary'];
    // If showTemporary is not defined it defaults to omitTemporary, which is typically false
    if (showTemporary === undefined) showTemporary = !omitTemporary;
    let showGrid = properties['showGrid'];
    if (showGrid === undefined) showGrid = showTemporary;

    if (size !== null && isNaN(size.width) && isNaN(size.height)) size = null;
    if (typeof padding === 'number') {
      padding = new Margin(padding);
    } else if (!(padding instanceof Margin)) { Util.throwError('MakeImage padding must be a Margin or a number.'); }
    // make sure padding is non-negative
    padding.left = Math.max(padding.left, 0);
    padding.right = Math.max(padding.right, 0);
    padding.top = Math.max(padding.top, 0);
    padding.bottom = Math.max(padding.bottom, 0);

    this._ctx.clearContextCache(true);
    const canvas = new CanvasSurface(null, doc);
    const ctx = canvas.context;
    if (!(size || scale || parts || position)) {
      canvas.width = this._canvasWidth + Math.ceil(padding.left + padding.right);
      canvas.height = this._canvasHeight + Math.ceil(padding.top + padding.bottom);
      if (contextType === 'SVG') {
        if (svgSurface === null) return null;
        svgSurface.resize(canvas.width, canvas.height, canvas.width, canvas.height);
        svgSurface.ownerDocument = doc;
        svgSurface.elementFinished = elementFinished;
        this.drawForeign(svgSurface.context, padding, new Size(canvas.width, canvas.height), this._scale, this._position, parts, background, showTemporary, showGrid);
        return svgSurface.context;
      }
      this._drawTainted = false;
      this.drawForeign(ctx, padding, new Size(canvas.width, canvas.height), this._scale, this._position, parts, background, showTemporary, showGrid);
      this._drawTainted = true;
      return canvas.context;
    }

    const defaultScale = this._defaultScale;
    const db = this.documentBounds.copy();
    db.subtractMargin(this._padding);
    // if showTemporary is true, union in the temporary layers: (adornments, etc)
    if (showTemporary) {
      const lit = this._layers._arr;
      const l = lit.length;
      for (let i = 0; i < l; i++) {
        const layer = lit[i];
        if (!layer.visible) continue;
        if (!layer.isTemporary) continue; // ONLY temporary layers
        const pit = layer._parts._arr;
        const ll = pit.length;
        for (let j = 0; j < ll; j++) {
          const part = pit[j];
          if (!part.isInDocumentBounds) continue;
          if (!part.isVisible()) continue;
          const ab = part.actualBounds;
          if (!ab.isReal()) continue;
          db.unionRect(ab);
        }
      }
    }
    let pos = new Point(db.x, db.y); // default position

    if (parts !== null) {
      let b: Rect;
      let first = true;
      const pit2 = parts.iterator;
      pit2.reset();
      while (pit2.next()) {
        const obj = pit2.value;
        if (!(obj instanceof Part)) continue;
        const layer = obj.layer;
        if (layer !== null && !layer.visible) continue;
        if (layer !== null && !showTemporary && layer.isTemporary) continue;
        if (!obj.isVisible()) continue;
        const ab = obj.actualBounds;
        if (!ab.isReal()) continue;
        if (first) {
          first = false;
          b = ab.copy();
        } else {
          b.unionRect(ab);
        }
      }
      // if no elements make an empty rect
      if (first) {
        b = new Rect(0, 0, 0, 0);
      }
      // replace document bounds with the smaller bounds from parts list
      db.width = b.width;
      db.height = b.height;
      pos.x = b.x;
      pos.y = b.y;
    }

    // replace position if given by options, which also replaces scale if scale is not given
    if (position !== null && position.isReal()) {
      pos = position;
      if (!scale) scale = defaultScale;
    }

    let pw = 0.0;
    let ph = 0.0;
    if (padding !== null) {
      pw = (padding.left + padding.right);
      ph = (padding.top + padding.bottom);
    }

    let sw = 0.0;
    let sh = 0.0;
    if (size !== null) {
      sw = size.width;
      sh = size.height;
      if (isFinite(sw)) sw = Math.max(0, sw - pw);
      if (isFinite(sh)) sh = Math.max(0, sh - ph);
    }

    let width = 0.0;
    let height = 0.0;
    if (size !== null && scale !== null) {
      // Special case that only operates when no size is given (only maxSize), so set scale to defaultScale
      if (scale === 'NaN') scale = defaultScale;

      // size and scale
      if (size.isReal()) {
        width = sw;
        height = sh;
      } else if (isNaN(sh)) {
        width = sw;
        height = db.height * scale;
      } else {
        width = db.width * scale;
        height = sh;
      }
    } else if (size !== null) {
      // size and no scale = auto scale
      if (size.isReal()) {
        scale = Math.min(sw / db.width, sh / db.height);
        width = sw;
        height = sh;
      } else if (isNaN(sh)) {
        scale = (sw / db.width);
        width = sw;
        height = db.height * scale;
      } else {
        scale = (sh / db.height);
        width = db.width * scale;
        height = sh;
      }
    } else if (scale !== null) {
      // no size, only scale
      if (scale === 'NaN' && maxSize.isReal()) {
        // Special case - fit to Max Size
        scale = Math.min((maxSize.width - pw) / db.width, (maxSize.height - ph) / db.height);
        if (scale > defaultScale) {
          scale = defaultScale;
          width = db.width;
          height = db.height;
        } else {
          width = maxSize.width;
          height = maxSize.height;
        }

      } else {
        width = db.width * scale;
        height = db.height * scale;
      }
    } else {
      scale = defaultScale;
      width = db.width;
      height = db.height;
    }

    if (padding !== null) {
      width += pw;
      height += ph;
    } else {
      padding = new Margin(0);
    }

    if (maxSize !== null) {
      let mw = maxSize.width;
      let mh = maxSize.height;
      if (contextType !== 'SVG' && defaultMaxSize && !Diagram._warnedAboutMakeImageSize) {
        if (Debug && ((width > mw) || (height > mh))) {
          Util.trace('Diagram.makeImage(data): Diagram width or height is larger than the default max size. (' +
            Math.ceil(width) + 'x' + Math.ceil(height) + ' vs 2000x2000) Consider increasing the max size.');
          Diagram._warnedAboutMakeImageSize = true;
        }
      }
      if (isNaN(mw)) mw = 2000;
      if (isNaN(mh)) mh = 2000;

      if (isFinite(mw)) width = Math.min(width, mw);
      if (isFinite(mh)) height = Math.min(height, mh);
    }

    canvas.width = Math.ceil(width);
    canvas.height = Math.ceil(height);
    if (contextType === 'SVG') {
      if (svgSurface === null) return null;
      svgSurface.resize(canvas.width, canvas.height, canvas.width, canvas.height);
      svgSurface.ownerDocument = doc;
      svgSurface.elementFinished = elementFinished;
      this.drawForeign(svgSurface.context, padding, new Size(Math.ceil(width), Math.ceil(height)), scale, pos, parts, background, showTemporary, showGrid);
      return svgSurface.context;
    }
    this._drawTainted = false;
    this.drawForeign(ctx, padding, new Size(Math.ceil(width), Math.ceil(height)), scale, pos, parts, background, showTemporary, showGrid);
    this._drawTainted = true;
    return canvas.context;
  }

  /**
   * This static function declares that a class (constructor function) derives from another class --
   * **_but please note that most classes do not support inheritance_**.
   * Do not call this function when your class is defined using an ES2015 or TypeScript "class" declaration.
   *
   * Because you must not modify the prototypes for the **GoJS** classes,
   * in order to override methods you need to define a new class that inherits from a predefined class.
   * You can then modify the prototype of your derived class.
   *
   * Typical usage is:
   * ```js
   * function BaseClass() {
   *   this._total = 0;
   * }
   *
   * public increment(x) {
   *   this._total += x;
   * }
   *
   * function DerivedClass() {
   *   this._anotherProperty = "";
   * }
   * go.Diagram.inherit(DerivedClass, BaseClass);
   *
   * DerivedClass.prototype.someMethod = ...;
   * ```
   *
   * **_Note that most classes do not support inheritance._**
   * Currently you can only inherit from {@link Layout}, {@link Tool}, {@link CommandHandler}, and {@link Link} or their subclasses.
   * When you override a method, you must strongly consider calling the base method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call a base method.
   *
   * The call to `Diagram.inherit` should be made immediately after defining the subclass constructor function
   * and before defining any new methods or overriding any base class methods.
   * You must not call this static function more than once on a derived class, or at all on a class defined using an ES2015 or TypeScript class declaration.
   *
   * The need for subclassing is greatly diminished by the presence of a number of properties that have functional values.
   * Setting such a property to a function will cause that function to be called as if it were
   * an event handler for that particular object.
   * Example properties include: {@link GraphObject#click}, {@link GraphObject#mouseEnter}, {@link Part#layerChanged},
   * {@link Node#treeExpandedChanged}, {@link LinkingBaseTool#linkValidation}, {@link CommandHandler#memberValidation},
   * {@link TextEditingTool#textValidation}.
   * @param {Function} derivedclass
   * @param {Function} baseclass
   */
  public static inherit(derivedclass: Function, baseclass: Function): void {
    if (Object.getPrototypeOf(derivedclass).prototype) {
      throw new Error('Used go.Diagram.inherit defining already defined class \n' + derivedclass);
    }
    Util.checkPrimitive(derivedclass, 'function', Diagram, 'inherit');
    Util.checkPrimitive(baseclass, 'function', Diagram, 'inherit');
    /**
     * @hidden @internal
     * @constructor
     */
    const classy: any = function subproto() { };
    classy.prototype = baseclass.prototype;
    derivedclass.prototype = new classy();
    derivedclass.prototype.constructor = derivedclass;
  }

  /**
   * @hidden @internal
   */
  public static wmq: any = null;
  /**
   * @hidden @internal
   */
  public static wmq1 = '5da73c80a36455d5038e4972187c3cae51fd22'; // 2.1 evaluation
  /**
   * @hidden @internal
   */
  public static wmq2 = false;
  /**
   * @hidden @internal
   */
  public static wmq3 = function(): void {
    if (!Diagram.isUsingDOM()) return null;
    const nc = root.document.createElement('canvas');
    const nctx = nc.getContext('2d') as any;
    const fstyle = Util.decrypt('7ca11abfd022028846');
    nctx[fstyle] = Util.decrypt('398c3597c01238');
    const draw = '7ca11abfd7330390';
    const draw2 = [Diagram.wmq1,
    ((Util as any)['Dx'] + '4ae6247590da4bb21c324ba3a84e385776'),
    ((Transform as any)['xF'] + 'fb236cdfda5de14c134ba1a95a2d4c7cc6f93c1387' /*'or production use'*/),
    (Geo as any)['za']];
    const x = 10;
    const y = 0;
    for (let i = 1; i < 5; i++) {
      nctx[Util.decrypt(draw)](Util.decrypt(draw2[i - 1]), x, (i * 15) + y);
    }
    // x++;y--;
    nctx[fstyle] = Util.decrypt('39f046ebb36e4b');
    for (let i = 1; i < 5; i++) {
      nctx[Util.decrypt(draw)](Util.decrypt(draw2[i - 1]), x, (i * 15) + y);
    }

    if (draw2.length !== 4 || draw2[0][0] !== '5' || draw2[3][0] !== '7') { // test by changing 5 to 6
      // TODO find something odd/annoying to no-tamper
      // Diagram.prototype._setDocView = function(e: Event | Touch, n: InputEvent): void  {
      //   // avoids creation of new points by not calling getMouse/transformPoint
      //   const can = this._canvas;
      //   const canwidth = this._canvasWidth;
      //   const canheight = this._canvasHeight;
      //   let mx = 0.0;
      //   let my = 0.0;
      //   if (can !== null) {
      //     const bbox = can.getBoundingClientRect();
      //     mx = e.clientX - bbox.left * (canwidth / bbox.width);
      //     my = e.clientY - bbox.top * (canheight / bbox.height);
      //   }
      //   n.viewPoint.assignTo(mx, my);
      //   if (this._transform !== null) {
      //     const pt = Point.allocAt(mx, my + 20);
      //     this._transform.invertedTransformPoint(pt);
      //     n.documentPoint.assign(pt);
      //     Point.free(pt);
      //   } else {
      //     n.documentPoint.assignTo(mx, my + 20);
      //   }
      // };
    }

    Diagram.wmq = nc;
  };

} // #end Diagram
Diagram['className'] = 'Diagram';
Diagram['fromDiv'] = Diagram.fromDiv;
Diagram['inherit'] = Diagram.inherit;

Diagram['None'] = Diagram.None;
Diagram['Uniform'] = Diagram.Uniform;
Diagram['UniformToFill'] = Diagram.UniformToFill;
Diagram['CycleAll'] = Diagram.CycleAll;
Diagram['CycleNotDirected'] = Diagram.CycleNotDirected;
Diagram['CycleNotDirectedFast'] = Diagram.CycleNotDirectedFast;
Diagram['CycleNotUndirected'] = Diagram.CycleNotUndirected;
Diagram['CycleDestinationTree'] = Diagram.CycleDestinationTree;
Diagram['CycleSourceTree'] = Diagram.CycleSourceTree;
Diagram['DocumentScroll'] = Diagram.DocumentScroll;
Diagram['InfiniteScroll'] = Diagram.InfiniteScroll;
Diagram['TreeParentCollapsed'] = Diagram.TreeParentCollapsed;
Diagram['AllParentsCollapsed'] = Diagram.AllParentsCollapsed;
Diagram['AnyParentsCollapsed'] = Diagram.AnyParentsCollapsed;



// wmq
/**
 * @hidden @internal
 * @unrestricted
 */
export class DiagramHelper {
  /** @hidden @internal */ public static className = 'DiagramHelper';
  /** @hidden @internal */ private _t: any;
  /** @hidden @internal */ private _tk: any;
  /** @hidden @internal */ private _tk2: any;
  /** @hidden @internal */ private _e: any;

  /**
   * @param {Diagram=} dia
   * @param {Element|string=} div
   */
  constructor(dia?: Diagram, div?: Element | string) {
    // this._t = true;
    // this._f = false;
    this._t = null;
    this._tk = '63ad05bbe23a1786468a4c741b6d2';
    this._tk2 = 'zz@orderNum';
    // wmq
    // root.location.origin or root.document.URL
    // this one uses root.location.href
    // we could strip it: .replace(/.*?:\/\//g, "").replace(Util.decrypt('6dbf01fd'),'')
    // don't even make the property until this very moment
    // To debug while testing go.licenseKey, comment out:
    if (this._tk === this['_tk']) { this._tk2 = true; this._e = this._tk2; return; }
    this._e = null; // instead of calling his.trywmq(dia, false);
  }

  /**
   * @hidden @internal
   * @this {DiagramHelper}
   * @param {Diagram} dia
   * @param {boolean} secondTime
   */
  private trywmq(dia: Diagram, secondTime: boolean) {
    let n: any = 'f';
    const b = root[Util.decrypt('76a715b2f73f148a')][Util.decrypt('72ba13b5')];
    const decrypt = Util.decrypt;
    this._e = true;
    if (!Diagram.isUsingDOM()) return;
    // if no license key present, try to find gojs.net:
    let licenseKey = (Diagram as any)[decrypt('76a115b6ed251eaf4692')];
    if (!licenseKey) {
      const l1 = b[decrypt('76ad18b4f73e')];
      for (let i = b[decrypt('73a612b6fb191d')](decrypt('35e7')) + 2; i < l1; i++) {
        n += b[i];
      }
      const n2 = n[decrypt('73a612b6fb191d')](decrypt('7da71ca0ad381e90')); // index of the address
      // must be found and also before the first '/'
      this._e = !(n2 >= 0 && n2 < n[decrypt('73a612b6fb191d')](decrypt('35')));
      return; // good domain (gojs.net) or not, no license key so we must finish
    }
    const keys = Diagram._licenseKey.iterator;
    // look at Diagram.licenseKey if it exists
    while (keys.next()) {
      licenseKey = keys.value;
      const keyArray = decrypt(licenseKey).split(decrypt('39e9')); // [ordernum,version,domain,email,product,date]
      if (keyArray.length < 6) return; // bad key assume all are bad
      const keyVersion = decrypt(keyArray[1]).split('.');
      // check product !== "gojs"
      if (keyArray[4] !== '7da71ca0') return; // bad key assume all are bad

      const fileVersion = decrypt((Util as any)[decrypt('6cae19')]).split('.'); // decrypt(Util.vfo) = "x.y.z"
      // if (fileVersion[0] === '28e647fdb37b1a8853835e27') fileVersion = '9.8.7'.split('.'); // To debug while testing, because its 28e647fdb37b1a8853835e27 before building
      if (!(keyVersion[0] > fileVersion[0] || (keyVersion[0] === fileVersion[0] && keyVersion[1] >= fileVersion[1]))) {
        continue; // bad version, try next key
      }
      // version is acceptable, now check domain
      // domain: decrypt(keyArray[2]);
      const l1 = b[decrypt('76ad18b4f73e')];
      const io = '73a612b6fb191d'; // indexOf
      for (let i = b[decrypt(io)](decrypt('35e7')) + 2; i < l1; i++) {
        n += b[i];
      }
      let nn1 = n[decrypt(io)](decrypt(keyArray[2])); // index of the address
      // if not there, consider looking for loc host under some additional domain conditions
      // if not licensed to gojs.net, check and allow localhost
      if (nn1 < 0 && decrypt(keyArray[2]) !== decrypt('7da71ca0ad381e90')/* gojs.net*/) nn1 = n[decrypt(io)](decrypt('76a715b2ef3e149757')/*localhost*/);
      // key present and version number OK, allow localhost:
      if (nn1 < 0) nn1 = n[decrypt(io)](decrypt('76a715b2ef3e149757'));

      // key present and version number OK, allow about:blank too:
      if (nn1 < 0) nn1 = b[decrypt(io)](decrypt('7baa19a6f76c1988428554'));

      // found the domain in the key, must be found and also before the first '/' or no existing '/'
      this._e = !(nn1 >= 0 && (nn1 < n[decrypt(io)](decrypt('35')) /* indexOf('/') */) || n[decrypt(io)](decrypt('35')) === -1 /* no / in url */  );
      if (!this._e) return; // good version + good domain, we can finish
      // otherwise we have good version and might have good CSS, so keep going

      // new since 1.8:
      // look for gsfBackground and background-color
      const color = decrypt(keyArray[2]);
      if (color[0] !== '#') return; // not a hex color, so quit (show wmq)

      const tempdiv = root.document['createElement']('div');

      let onlyDigits = keyArray[0].replace(/[A-Za-z]/g, '');
      while (onlyDigits.length < 4) { onlyDigits += '9'; }
      onlyDigits = onlyDigits.substr(onlyDigits.length - 4);
      // classname with no leading period
      let classname = '';
      const first = ['gsh', 'gsf'];
      const second = ['Header', 'Background', 'Display', 'Feedback'];
      classname += first[parseInt(onlyDigits.substr(0, 1), 10) % 2];
      classname += second[parseInt(onlyDigits.substr(0, 1), 10) % 4];

      tempdiv[decrypt('79a417a0f0181a8946')] = classname;
      if (root.document[decrypt('78a712aa')]) { // if root.document.body exists
        // document body appendchild
        root.document[decrypt('78a712aa')][decrypt('7bb806b6ed32388c4a875b')](tempdiv);
        // computed style
        const test = root['getComputedStyle'](tempdiv)['getPropertyValue'](decrypt('78a704b7e62456904c9b12701b6532a8')) as any;
        // documeny body removechild
        root.document[decrypt('78a712aa')][decrypt('68ad1bbcf533388c4a875b')](tempdiv);
        if (!test) return; // if null or undefined we're done (bad)
        // Util.color1 and color2 must be numerical values

        if (test.indexOf(parseInt(color[1] + color[2], 16)) !== -1 && test.indexOf(parseInt(color[3] + color[4], 16)) !== -1) {
          this._e = false;
          // good! quit:
          return;
        } else if (Util.isIE9 || Util.isIE10 || Util.isIE11 || Util.isEdge) {
          // Maybe high contrast mode ruined the color, check for just the css class
          classname = '.' + classname;
          for (let i = 0; i < document.styleSheets.length; i++) {
            const rules = document.styleSheets[i]['rules'] || document.styleSheets[i]['cssRules'];
            for (const x in rules) {
              if (classname === rules[x].selectorText) {
                // compare selector text vs class name including '.'
                this._e = false;
                break; // found class, good
              }
            }
          }
          // didn't find class, bad
        }
        // bad, continue
      } else { // if root.document.body does not exist
        this._e = null;
        if (secondTime) this._e = false;
      }
    }

  }

  /**
   * @hidden @internal
   * @this {DiagramHelper}
   * @param {Diagram} dia
   * @return {boolean}
   */
  public wmq(dia: Diagram): boolean {
    dia._ctx.setTransform(dia._pixelRatio, 0, 0, dia._pixelRatio, 0, 0);
    if (this._e === null) this.trywmq(dia, true);

    if (0 < this._e && this !== this._t) return true;
    return false;
  }

  /**
   * @hidden @internal
   * Happens the first time a Div is attached to the Diagram. Adds a canvas to the specified div and will
   * maintain the canvas' size to always be that of the div.
   * This is called automatically if a div is supplied in the Diagram's constructor.
   * Otherise its called during the _div setup.
   * @this {DiagramHelper}
   * @param {Diagram} dia A diagram.
   * @param {Object|string} div A reference to a div or its ID as a string.
   */
  public setup(dia: Diagram, div: ObjectData | string): void {
    if (!Diagram.isUsingDOM()) return; // DOM-less
    if (div === undefined || div === null) {
      Util.throwError('Diagram setup requires an argument DIV.');
    }
    if (dia._div !== null) {
      Util.throwError('Diagram has already completed setup.');
    }

    if (typeof div === 'string') {
      dia._div = root.document.getElementById(div) as HTMLDivElement;
    } else if (div instanceof HTMLDivElement) {
      dia._div = div;
    } else {
      Util.throwError('No DIV or DIV id supplied: ' + div);
    }

    if (dia._div === null) {
      Util.throwError('Invalid DIV id; could not get element with id: ' + div);
    }

    if (dia._div._diagram !== undefined) {
      Util.throwError('Invalid div id; div already has a Diagram associated with it.');
    }

    // We cannot let the div position remain static. If it is, make it relative:
    if (root.getComputedStyle(dia._div, null).position === 'static') {
      dia._div.style.position = 'relative';
    }
    // avoid highlighting upon touch in Android
    dia._div.style['-webkit-tap-highlight-color'] = 'rgba(255, 255, 255, 0)';
    // avoid letting users pan over the Diagram in IE10
    dia._div.style['-ms-touch-action'] = 'none';

    // Clobber any HTML inside the div
    dia._div.innerHTML = '';
    (dia._div as any)._diagram = dia;

    // Create a new canvas
    const svg = dia._svg;

    const newCanvas = svg ? new SVGSurface(dia) : new CanvasSurface(dia);

    if (newCanvas.style !== undefined) {
      newCanvas.style.position = 'absolute';
      newCanvas.style.top = '0px';
      newCanvas.style.left = '0px';
      const direction = root.getComputedStyle(dia._div, null).getPropertyValue('direction');
      if (direction === 'rtl') dia._isRTL = true;
      newCanvas.style.zIndex = '2';
      newCanvas.style.userSelect = 'none';
      newCanvas.style.webkitUserSelect = 'none';
      newCanvas.style.MozUserSelect = 'none'; // capital M is intentional
    }

    // Add the canvas to this class and append it to the div
    dia._canvas = newCanvas;
    dia._ctx = newCanvas.context;
    const ctx = dia._ctx;

    dia._pixelRatio = dia.computePixelRatio();
    dia._canvasWidth = dia._div.clientWidth || 1;
    dia._canvasHeight = dia._div.clientHeight || 1;
    dia.updateCanvasSize(dia._canvasWidth, dia._canvasHeight);
    const diagram = dia; // closure used for all events in constructor

    // wmq
    const ctx2 = ctx as CanvasSurfaceContext; // NYI: change this when it can accept SVG
    dia._drawImage = (ctx2._backingContext as any)[Util.decrypt('7eba17a4ca3b1a8346')][Util.decrypt('78a118b7')](ctx2._backingContext, Diagram.wmq, 4, 4);
    dia._div.insertBefore(newCanvas.domElement, dia._div.firstChild);

    // Init a temp canvas
    const can = new CanvasSurface(null);
    can.width = 1;
    can.height = 1;
    dia._drawCollectionCanvas = can;
    dia._drawCollectionCtx = can.context;

    // ****************************************
    // DOM scrollbars
    // ****************************************
    if (Diagram.isUsingDOM()) {
      const scrollParent = Util.createElement('div');
      const scrollHelper = Util.createElement('div');

      scrollParent.style.position = 'absolute';
      scrollParent.style.overflow = 'auto';
      scrollParent.style.width = dia._canvasWidth + 'px';
      scrollParent.style.height = dia._canvasHeight + 'px';
      scrollParent.style.zIndex = '1';

      scrollHelper.style.position = 'absolute';
      scrollHelper.style.width = '1px';
      scrollHelper.style.height = '1px';

      dia._div.appendChild(scrollParent);
      scrollParent.appendChild(scrollHelper);

      scrollParent.onscroll = Diagram._onscroll as any;
      scrollParent.onmousedown = Diagram._onscrollClick;
      scrollParent.ontouchstart = Diagram._onscrollClick;
      (scrollParent as any)._diagram = dia; // needed for _onscroll
      scrollParent._scrollHorizontal = true;
      scrollParent._scrollVertical = true;

      // attach these to the diagram so we can keep track
      dia._scrollParent = scrollParent;
      dia._scrollHelper = scrollHelper;
    }
    // ****************************************
    // Event handlers
    // ****************************************

    // debounce functions. Invoke Util.debounce only once, when defining the function
    /**
     * @hidden @internal
     * Calls invalidate draw after 250ms of not being called
     */
    dia.redrawLater = Util.debounce(function() {
      dia._tempPixelRatio = null;
      dia.invalidateDraw();
    }, 300, false);

    /**
     * @hidden @internal
     */
    dia.windowResize = Util.debounce(function() {
      dia.maybeResize();
      // don't force a layout here. Can't assume that the diagram has changed size, might be a static size!
    }, 250, false);

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {Event} e
     */
    dia.preventDefault = function(e: Event) { e.preventDefault(); return false; };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {MouseEvent} e
     */
    dia.mouseMove = function(e: MouseEvent) {
      if (!diagram.isEnabled) return;
      dia._isMouseOverDiagram = true;
      const n = diagram._commonEvent(e, true);

      diagram.doMouseMove();
      if (diagram.currentTool.isBeyondDragSize()) {
        diagram._lastTouchCount = 0;
      }
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {MouseEvent} e
     */
    dia.mouseDown = function(e: MouseEvent) {
      if (!diagram.isEnabled) return;
      diagram._isMouseOverDiagram = true;

      // This could be a simulated right click from a long tap.
      // We handle that separately, so cancel the event.
      // on Surface Pro 4 it simulates e.button === 0, other devices simulated e.button === 2
      if (diagram._isTouchOngoing) {
        e.preventDefault();
        return;
      }

      const n = diagram._commonEvent(e, true);
      n.down = true;
      n.clickCount = e.detail;
      // IE10+IE11 have a bug with event.detail:
      // http://connect.microsoft.com/IE/feedback/details/789773/ie-increments-event-detail-on-clicks-over-different-locations
      // this is a workaround for that bug.
      if (Util.isIE10 || Util.isIE11) {
        // potential double click, make sure its close in time and space
        if (e.timeStamp - diagram._lastTouchStamp < diagram._touchDoubleClickDelay &&
          !diagram.currentTool.isBeyondDragSize()) {
          diagram._lastTouchCount++;
        } else {
          diagram._lastTouchCount = 1;
        }
        diagram._lastTouchStamp = e.timeStamp;
        n.clickCount = diagram._lastTouchCount;
      }

      // DraggingTool._source = null;
      n.clone(diagram.firstInput);
      diagram.doMouseDown();

      // we do not want to call maybeHandled on a mouseDown because it may break mouseUp events
      // But we do want to stop the page from initiating autoscroll if we middle mouse click
      if (e.button === 1) { e.preventDefault(); return; }
      diagram._maybeHandled(n, e); // this breaks keyboard focus,  unless something else calls focus()

      // if (Debug && Debug.nodeInspector) {
      // root.nodeInspector = diagram.findObjectsAt(n.documentPoint);
      // Util.trace(root.nodeInspector)
      // }
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {MouseEvent} e
     */
    dia.mouseUp = function(e: MouseEvent): void {
      if (!diagram.isEnabled) return;

      // This could be a simulated right click from a long tap.
      // We handle that separately, so cancel the event.
      // on Surface Pro 4 it simulates e.button === 0, other devices simulated e.button === 2
      if (diagram._isTouchOngoing && e.button === 2) {
        e.preventDefault();
        return;
      }
      // Surface:
      if (diagram._isTouchOngoing && e.button === 0) {
        diagram._isTouchOngoing = false;
      }
      if (diagram._tapHeld) { e.preventDefault(); return; }

      dia._isMouseOverDiagram = true;

      const n = diagram._commonEvent(e, true);
      n.up = true;
      n.clickCount = e.detail;

      // IE10+IE11 have a bug with event.detail:
      // http://connect.microsoft.com/IE/feedback/details/789773/ie-increments-event-detail-on-clicks-over-different-locations
      // this is a workaround for that bug.
      if (Util.isIE10 || Util.isIE11) {
        n.clickCount = diagram._lastTouchCount;
      }

      n.bubbles = e.bubbles;
      n.targetDiagram = diagram._findTargetDiagram(e);
      diagram.doMouseUp();
      diagram.stopAutoScroll();
      // We do want to call maybeHandled because if the window is uncapturing the event
      // both a window mouseup and a canvas mouseup will fire. This stops that from happening.
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {WheelEvent} e
     */
    dia.mouseWheel = function(e: WheelEvent): void {
      if (!diagram.isEnabled) return;

      const n = diagram._commonEvent(e, true);
      n.bubbles = true; // allow mouse wheel events to be handled by browser scrollbars
      // normalize wheel events, converting to just +1 or -1
      let deltaX = 0;
      let deltaY = 0;
      n.delta = 0;
      if (e.deltaX !== undefined) {
        if (e.deltaX !== 0) deltaX = e.deltaX > 0 ? 1 : -1;
        if (e.deltaY !== 0) deltaY = e.deltaY > 0 ? 1 : -1;
        n.delta = (Math.abs(e.deltaX) > Math.abs(e.deltaY)) ? -deltaX : -deltaY;
      } else if (e.wheelDeltaX !== undefined) {
        if (e.wheelDeltaX !== 0) deltaX = e.wheelDeltaX > 0 ? -1 : 1;
        if (e.wheelDeltaY !== 0) deltaY = e.wheelDeltaY > 0 ? -1 : 1;
        n.delta = (Math.abs(e.wheelDeltaX) > Math.abs(e.wheelDeltaY)) ? -deltaX : -deltaY;
      } else if (e.wheelDelta !== undefined) {
        if (e.wheelDelta !== 0) n.delta = e.wheelDelta > 0 ? 1 : -1;
      }

      diagram.doMouseWheel();
      // The rest of this event gets cancelled sometimes in Tool's standardMouseWheel
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {MouseEvent} e
     */
    dia.mouseOut = function(e: MouseEvent): void {
      if (!diagram.isEnabled) return;
      dia._isMouseOverDiagram = false;
      diagram._commonEvent(e, true);
      const tool = diagram.currentTool;
      tool.cancelWaitAfter(); // cancel any timer
      tool.standardMouseOver(); // make sure mouseout happens on any objects, clear _currentEnterLeaveObject
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {TouchEvent} e2
     */
    dia.touchStart = function(e2: Event): void {
      if (!diagram.isEnabled) return;
      diagram._tapHeld = false;
      diagram._isTouchOngoing = true;
      const e = e2 as TouchEvent;
      const n = diagram._commonTouchStart(e, e.targetTouches[0], e.touches.length > 1);

      let evt: Touch = null;
      if (e.targetTouches.length > 0) evt = e.targetTouches[0];
      else if (e.changedTouches.length > 0) evt = e.changedTouches[0];
      if (evt !== null) {
        const firstx = evt.screenX;
        const firsty = evt.screenY;
        const last = diagram._lastTouchpt;
        // potential double click, make sure its close in time and space
        if (e.timeStamp - diagram._lastTouchStamp < diagram._touchDoubleClickDelay &&
          // Special isBeyondDragSize
          !(Math.abs(last.x - firstx) > 25 || Math.abs(last.y - firsty) > 25)) {
          diagram._lastTouchCount++;
        } else {
          diagram._lastTouchCount = 1;
        }
        n.clickCount = diagram._lastTouchCount;
        diagram._lastTouchStamp = e.timeStamp;
        diagram._lastTouchpt.assignTo(firstx, firsty);
      }

      diagram.doMouseDown();
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {TouchEvent} e2
     */
    dia.touchMove = function(e2: Event): void { // Event or TouchEvent
      if (!diagram.isEnabled) return;
      const e = e2 as TouchEvent;

      let evt: Touch = null;
      if (e.targetTouches.length > 0) evt = e.targetTouches[0];
      else if (e.changedTouches.length > 0) evt = e.changedTouches[0];
      const n = diagram._commonTouchMove(e, evt, e.touches.length > 1);

      diagram.doMouseMove();
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {Event} e2
     */
    dia.touchEnd = function(e2: Event): void { // Event or TouchEvent
      if (!diagram.isEnabled) return;
      const e = e2 as TouchEvent;
      if (diagram._tapHeld) { e.preventDefault(); return; }

      // not currently handling more than one touch on end
      if (e.touches.length > 1) {
        return;
      }

      let targ: Element = null;
      let evt: Touch = null;
      if (e.targetTouches.length > 0) evt = e.targetTouches[0];
      else if (e.changedTouches.length > 0) evt = e.changedTouches[0];
      const n = diagram._commonTouchSetup(e,
        false,
        true,
        false,
        false, // Might cancel double-click zoom on some browsers
        false);

      if (evt !== null) {
        targ = root.document.elementFromPoint(evt.clientX, evt.clientY);
        if (targ !== null && targ._diagram instanceof Diagram && targ._diagram !== diagram) {
          targ._diagram._setDocView(evt, n);
        }
        diagram._setDocView(evt, n);
        n.clickCount = diagram._lastTouchCount;
      }

      if (targ === null) {
        n.targetDiagram = diagram._findTargetDiagram(e);
      } else if (targ._diagram) {
        n.targetDiagram = targ._diagram;
      } else {
        n.targetDiagram = null;
      }
      n.targetObject = null;

      diagram.doMouseUp();
      diagram._maybeHandled(n, e);
      diagram._isTouchOngoing = false;
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {PointerEvent} e
     */
    dia.pointerDown = function(e: PointerEvent): void {
      if (!diagram.isEnabled) return;
      dia._isMouseOverDiagram = true;
      const pointers = diagram._pointers;

      if (pointers[e.pointerId] === undefined) {
        pointers[e.pointerId] = e;
      }

      // using the primary input, deal with the first two fingers
      const pointers2 = diagram._pointers2;

      let isMulti = false;
      if (pointers2[0] !== null && pointers2[0].pointerId === e.pointerId) {
        pointers2[0] = e;
      } else if (pointers2[1] !== null && pointers2[1].pointerId === e.pointerId) {
        pointers2[1] = e;
        isMulti = true;
      } else if (pointers2[0] === null) {
        // allow pointerMove moves without the pointerDown first:
        pointers2[0] = e;
      } else if (pointers2[1] === null) {
        pointers2[1] = e;
        isMulti = true;
      } else {
        // Don't deal with 3rd+ fingers
        e.preventDefault();
        return;
      }

      if (e.pointerType === 'touch' || e.pointerType === 'pen') {
        diagram._tapHeld = false;
        diagram._isTouchOngoing = true;
      }

      // We do not want to set the last input with the second pointer
      // so commonTouchStart looks at isMulti
      const n = diagram._commonTouchStart(e, e, isMulti);
      const isTouchLike = (e.pointerType === 'touch' || e.pointerType === 'pen');
      const last = diagram._lastTouchpt;
      const dragSize = isTouchLike ? 25 : 10;
      // potential double click, make sure its close in time and space
      if (e.timeStamp - diagram._lastTouchStamp < diagram._touchDoubleClickDelay &&
        // Special isBeyondDragSize
        !(Math.abs(last.x - e.screenX) > dragSize || Math.abs(last.y - e.screenY) > dragSize)) {
        diagram._lastTouchCount++;
      } else {
        diagram._lastTouchCount = 1;
      }
      n.clickCount = diagram._lastTouchCount;
      diagram._lastTouchStamp = e.timeStamp;
      diagram._lastTouchpt.setTo(e.screenX, e.screenY);

      diagram.doMouseDown();
      // we do not want to call maybeHandled on a mouseDown because it may break mouseUp events
      // But we do want to stop the page from initiating autoscroll if we middle mouse click
      if (e.button === 1) { e.preventDefault(); return; }
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {PointerEvent} e
     */
    dia.pointerMove = function(e: PointerEvent): void {
      if (!diagram.isEnabled) return;
      dia._isMouseOverDiagram = true;
      const pointers2 = diagram._pointers2;

      if (pointers2[0] !== null && pointers2[0].pointerId === e.pointerId) {
        pointers2[0] = e;
      } else if (pointers2[1] !== null && pointers2[1].pointerId === e.pointerId) {
        pointers2[1] = e;
        return; // We do not want to set the last input or call doMouseMove with the second pointer.
      } else if (pointers2[0] === null) {
        // allow pointerMove moves without the pointerDown first:
        pointers2[0] = e;
      } else {
        return; // disallow 3rd, 4th, etc
      }

      if (pointers2[0].pointerId !== e.pointerId) return;

      const n = diagram._commonTouchMove(e, e, pointers2[1] !== null /* isMulti or not*/);
      n.targetDiagram = diagram._findTargetDiagram(e);
      diagram.doMouseMove();
      diagram._maybeHandled(n, e);
    };

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {PointerEvent} e
     */
    dia.pointerUp = function(e: PointerEvent) {
      if (!diagram.isEnabled) return;
      dia._isMouseOverDiagram = true;
      const isTouchLike = (e.pointerType === 'touch' || e.pointerType === 'pen');

      const pointers = diagram._pointers;
      if (isTouchLike) { // is this block necessary?
        if (diagram._tapHeld) {
          delete pointers[e.pointerId]; // set undefined instead?
          e.preventDefault();
          return;
        }
      }
      const pointers2 = diagram._pointers2;

      if (pointers2[0] !== null && pointers2[0].pointerId === e.pointerId) {
        pointers2[0] = null;
      } else if (pointers2[1] !== null && pointers2[1].pointerId === e.pointerId) {
        pointers2[1] = null;
        return; // Only looking at the original pointer, no subsequent ones
      } else {
        return; // Only looking at the original pointer, no subsequent ones
      }

      const n = diagram._commonTouchSetup(e,
        false,
        true,
        false,
        true, // false for touch up, but true for pointer up. Seems OK on Surface regardless of pointer type.
        false);

      const targ = root.document.elementFromPoint(e.clientX, e.clientY);
      if (targ !== null && targ._diagram instanceof Diagram && targ._diagram !== diagram) {
        targ._diagram._setDocView(e, n);
      }
      diagram._setDocView(e, n);

      n.clickCount = diagram._lastTouchCount;

      if (targ === null) {
        n.targetDiagram = diagram._findTargetDiagram(e);
      } else if (targ._diagram) {
        n.targetDiagram = targ._diagram;
      } else {
        n.targetDiagram = null;
      }
      n.targetObject = null;

      diagram.doMouseUp();
      diagram._maybeHandled(n, e);
      if (isTouchLike) diagram._isTouchOngoing = false;
    }; // end pointerUp

    /**
     * @hidden @internal
     * @this {HTMLCanvasElement}
     * @param {Event} e
     */
    dia.pointerOut = function(e: PointerEvent) {
      if (!diagram.isEnabled) return;

      dia._isMouseOverDiagram = false;
      const pointers = diagram._pointers;
      if (pointers[e.pointerId]) {
        delete pointers[e.pointerId];
      }
      const pointers2 = diagram._pointers2;
      if (pointers2[0] !== null && pointers2[0].pointerId === e.pointerId) {
        pointers2[0] = null;
      }
      if (pointers2[1] !== null && pointers2[1].pointerId === e.pointerId) {
        pointers2[1] = null;
      }

      // don't do normal mouse-out stuff on touch-like events, otherwise this will remove tooltips
      if (e.pointerType === 'touch' || e.pointerType === 'pen') return;
      const tool = diagram.currentTool;
      tool.cancelWaitAfter();
      tool.standardMouseOver();
    }; // end pointerOut

    ctx.clearContextCache(true);
    dia.addListeners();
  }

} // #end DiagramHelper
DiagramHelper['className'] = 'DiagramHelper';

/**
 * This is used in dragging methods to remember initial locations, for relative positioning and for cancellation.
 * You do not typically need to create these unless you are overriding tools.
 *
 * {@link CommandHandler#computeEffectiveCollection} returns a Map of Parts and DraggingInfo,
 * which is used as an argument for methods like {@link Diagram#computeMove}.
 *
 * {@link Diagram#moveParts} constructs a  Map of Parts and DraggingInfo to record each Part's location offset during moves.
 *
 * @unrestricted
 * @since 2.0
 */
export class DraggingInfo {
  /** @hidden @internal */ public static className = 'DraggingInfo';
  /** @hidden @internal */ private _point: Point;
  /** @hidden @internal */ private _shifted: Point;

  /**
   * Optionally takes a Point that sets {@link #point}.
   * @param {Point=} pt
   */
  constructor(pt?: Point) {
    this._point = (pt === undefined) ? new Point() : pt;
    this._shifted = new Point(); // not Point.Origin because we constantly re-use the same point
  }

  /**
   * Gets or sets the initial location for a dragged object.
   * This is most commonly the Part.location of the dragged object, or a rounded Part.location if snapping.
   * @export
   */
  /** @export */ get point(): Point { return this._point; }
  /** @export */ set point(value: Point) { this._point = value; }

  /**
   * @hidden @internal
   * Records the amount shifted during a drag. Initially this value is `Point(0, 0)`.
   * {@link Diagram#moveParts} modifies this value to reflect the currently-shifted amount.
   */
  /** @export */ get shifted(): Point { return this._shifted; }
  /** @export */ set shifted(value: Point) { this._shifted = value; }

} // #end DraggingInfo
DraggingInfo['className'] = 'DraggingInfo';

/**
 * @hidden @internal
 * @unrestricted
 */
export class DraggingNodeInfoPair {
  /** @hidden @internal */ public static className = 'DraggingNodeInfoPair';
  /** @hidden @internal */ public node: Part;
  /** @hidden @internal */ public info: DraggingInfo;
  /** @hidden @internal */ public groupInfo: DraggingInfo | null;

  /**
   * @param {Part} node
   * @param {DraggingInfo} info
   * @param {DraggingInfo} groupinfo
   */
  constructor(node: Part, info: DraggingInfo, groupinfo: DraggingInfo) {
    this.node = node;
    this.info = info;
    this.groupInfo = groupinfo;
  }
} // #end DraggingNodeInfoPair
DraggingNodeInfoPair['className'] = 'DraggingNodeInfoPair';

/**
 * New in 2.0, this class contains options for dragging and moving parts.
 *
 * This is used by {@link Diagram#moveParts}, {@link Diagram#computeMove}, and {@link DraggingTool#computeMove}.
 *
 * Both the {@link Diagram} and its {@link DraggingTool} have a DraggingOptions instance.
 * Typically, only the DraggingTool one is used, unless the Diagram has no DraggingTool defined.
 *
 * In typical {@link DraggingTool} operation you do not need to create or modify a DraggingOptions instance.
 * Setting properties on the DraggingTool such as {@link DraggingTool#isGridSnapEnabled}
 * sets the value on the {@link DraggingTool#dragOptions} instance for you.
 *
 * @unrestricted
 * @see {@link DraggingTool#dragOptions}
 * @since 2.0
 */
export class DraggingOptions {
  /**
   * Whether or not the dragged parts snap to the {@link Diagram#grid}.
   * By default this property is false.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#isGridSnapEnabled}.
   */
  public isGridSnapEnabled: boolean;
  /**
   * If dragging with the {@link DraggingTool}, whether or not Parts snap to a grid (if applicable) in real-time,
   * or only at the end of the dragging transaction.
   * By default this property is true.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#isGridSnapRealtime}.
   */
  public isGridSnapRealtime: boolean;
  /**
   * Gets or sets the size of the grid cell used when snapping during a drag.
   * By default this property is the Size(NaN, NaN), which causes this tool to use the {@link Panel#gridCellSize} value of the {@link Diagram#grid}.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#gridSnapCellSize}.
   */
  public gridSnapCellSize: Size;
  /**
   * Gets or sets the {@link Spot} that specifies what point in the grid cell dragged parts snap to,
   * if the value of {@link #isGridSnapEnabled} is true. By default this property is {@link Spot.TopLeft}.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#gridSnapCellSpot}.
   */
  public gridSnapCellSpot: Spot;
  /**
   * Gets or sets the snapping grid's origin point, in document coordinates,
   * if the value of {@link #isGridSnapEnabled} is true. By default this property is the Point(NaN, NaN).
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#gridSnapOrigin}.
   */
  public gridSnapOrigin: Point;
  /**
   * Gets or sets whether the user can move a single {@link Link},
   * disconnecting it from connected nodes and possibly connecting it to valid ports.
   * By default this property is false.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#dragsLink}.
   */
  public dragsLink: boolean;
  /**
   * Gets or sets whether moving or copying a node also includes all of the
   * node's tree children and their descendants, along with the links to those additional nodes.
   * By default this property is false.
   *
   * In typical {@link DraggingTool} operation this is set via {@link DraggingTool#dragsTree}.
   */
  public dragsTree: boolean;
  /**
   * Determines whether or not Groups will snap their member Parts in addition to themselves when moving.
   * By default this property is false.
   */
  public groupsSnapMembers: boolean;
  /**
   * Determines whether or not Groups will ensure their member Parts always move,
   * even if those member parts have {@link Part#copyable} and {@link Part#movable} set to false.
   * By default this property is true.
   */
  public groupsAlwaysMove: boolean;
  // Maybe linksAlwaysMove some day for link labels


  constructor() {
    this.reset();
  }

  /**
   * @hidden @internal
   * Resets the properties of a DraggingOptions instance.
   * @this {DraggingOptions}
   */
  public reset() {
    this.isGridSnapEnabled = false;
    this.isGridSnapRealtime = true;
    this.gridSnapCellSize = new Size(NaN, NaN).freeze();
    this.gridSnapCellSpot = Spot.TopLeft;
    this.gridSnapOrigin = new Point(NaN, NaN).freeze();
    this.dragsLink = false;
    this.dragsTree = false;
    this.groupsSnapMembers = false; // new for 2.0
    this.groupsAlwaysMove = true; // new for 2.0.
  }
}


/**
 * Used for the options arguments to Diagram.makeImage, Diagram.makeImageData, and Diagram.makeSvg.
 * @hidden
 */
export interface DiagramRendererOptions {
  size?: Size;
  scale?: number;
  maxSize?: Size;
  position?: Point;
  parts?: Iterable<Part>;
  padding?: MarginLike;
  background?: BrushLike;
  showTemporary?: boolean;
  showGrid?: boolean;
  document?: HTMLDocument;
}

/**
 * Used for the options argument to Diagram.makeSvg.
 * @hidden
 */
export interface SvgRendererOptions extends DiagramRendererOptions {
  elementFinished?: ((graphobj: GraphObject, svgelt: SVGElement) => void) | null;
}

/**
 * Used for the options argument to Diagram.makeImage and Diagram.makeImageData.
 * @hidden
 */
export interface ImageRendererOptions extends DiagramRendererOptions {
  type?: string;
  returnType?: string;
  callback?: ((result: any) => void) | null;
  callbackTimeout?: number;
  details?: ObjectData;
}


// import { CanvasSurface, CanvasSurfaceContext } from './abstractCanvas';
// import { AnimationManager } from './animationManager';
// import { ChangedEvent } from './changedEvent';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { ICommandHandler } from './commandHandler';
// import { Debug } from './debug';
// import { } from './globals';
// import { GraphObject } from './graphObject';
// import { HTMLInfo } from './htmlInfo';
// import { DiagramEvent, DiagramEventName, InputEvent } from './inputEvent';
// import { Layer } from './layer';
// import { Layout } from './layout';
// import { Margin, MarginLike } from './margin';
// import { Binding, Key, Model } from './model';
// import { Panel } from './panel';
// import { PartManager } from './partManager';
// import { Adornment, Group, Part, Placeholder } from './parts';
// import { Node } from './parts';
// import { Link, LinkBundle, PositionArray } from './parts';
// import { Picture } from './picture';
// import { Point } from './point';
// import { PropertyCollection } from './propertyCollection';
// import { Rect } from './rect';
// import { RowColumnDefinition } from './rowColumnDefinition';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { IContext, ISurface } from './surfaceInterfaces';
// import { SVGContext, SVGSurface } from './svgContext';
// import { TextBlock } from './textBlock';
// import { Tool } from './tool';
// import { ToolManager } from './toolManager';
// import { UndoManager } from './undoManager';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Palette
*/

// import { Diagram } from './diagram';
// import { GridLayout } from './gridLayout';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * Palette extends the Diagram class to allow objects to be dragged and placed onto other Diagrams.
 * Its {@link Diagram#layout} is a {@link GridLayout}.
 * The Palette is {@link Diagram#isReadOnly} but to support drag-and-drop its {@link Diagram#allowDragOut} is true.
 * <p class="box">
 * For more discussion, see <a href="../../intro/palette.html">Introduction to Palettes</a>.
 * See samples that make use of Palettes in the <a href="../../samples/index.html#palette">samples index</a>.
 *
 * You can control the order of Parts in the palette in several manners:
 *   - If you want to keep the order given by `myPalette.model.nodeDataArray`,
 *     just set `myPalette.layout.sorting = go.GridLayout.Forward` ({@link GridLayout#sorting}).
 *   - If you want to sort based on some text string associated with each Part, just bind {@link Part#text}.
 *   - If you want to sort more generally, set `myPalette.layout.comparer` ({@link GridLayout#comparer}).
 * @extends Diagram
 * @unrestricted
 */
export class Palette extends Diagram {
  /** @hidden @internal */ public static className = 'Palette';
  /**
   * @param {HTMLDivElement|string} div A reference to a div or its ID as a string.
   */
  constructor(div?: Element | string) {
    if (arguments.length > 1) {
      Util.throwError('Palette constructor can only take one optional argument, the DIV HTML element or its id.');
    }
    super(div);

    this.allowDragOut = true;
    this.allowMove = false;
    this.isReadOnly = true;
    this.contentAlignment = Spot.TopCenter;

    this.layout = new GridLayout();
  }
} // #end Palette
Palette['className'] = 'Palette';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Overview
*/

// import { Diagram, DraggingInfo } from './diagram';
// import { DraggingTool } from './draggingTool';

/**
 * An Overview is a {@link Diagram} that displays all of a different diagram,
 * with a rectangular box showing the viewport displayed by that other diagram.
 * <p class="box">
 * For more discussion, see <a href="../../intro/overview.html">Introduction to Overviews</a>.
 * See samples that make use of Overviews in the <a href="../../samples/index.html#overview">samples index</a>.
 *
 * All you need to do is set {@link Overview#observed}.
 * For example:
 *
 * ```js
 *   var myDiagram = new go.Diagram("myDiagramDIV");
 *   . . . other initialization . . .
 *
 *   // create and initialize the Overview:
 *   new go.Overview("myOverviewDIV").observed = myDiagram;
 * ```
 *
 * The Overview draws what the observed Diagram is displaying,
 * so setting or modifying any diagram templates or template Maps has no effect.
 *
 * Animations are not shown in an Overview.
 *
 * At the current time methods such as {@link Diagram#makeImage},
 * {@link Diagram#makeImageData} and {@link Diagram#makeSvg} do not work on Overviews.
 * @extends Diagram
 * @unrestricted
 */
export class Overview extends Diagram {
  /** @hidden @internal */ public static className = 'Overview';
  /** @hidden @internal */ private _observed: Diagram | null;
  /** @hidden @internal */ private _drawsTemporaryLayers: boolean;
  /** @hidden @internal */ private _box: Part;
  /** @hidden @internal */ private _tempCanvas: CanvasSurface;
  /** @hidden @internal */ private _tempCtx: CanvasSurfaceContext;
  /** @hidden @internal */ private observedPanelViewportBoundsChanged: ((a: DiagramEvent) => void);
  /** @hidden @internal */ private observedPanelDocumentBoundsChanged: ((a: DiagramEvent) => void);
  /** @hidden @internal */ private observedPanelInvalidateDraw: ((a: DiagramEvent) => void);
  /** @hidden @internal */ private observedPanelAnimationFinished: ((a: DiagramEvent) => void);

  /**
   * @param {Element|string} div A reference to a div or its ID as a string.
   */
  constructor(div?: Element | string) {
    if (arguments.length > 1) {
      Util.throwError('Overview constructor can only take one optional argument, the DIV HTML element or its id.');
    }
    super(div);

    this.animationManager.isEnabled = false;

    this._isAligning = true;
    this._observed = null;

    this._hasHorizontalScrollbar = false;
    this._hasVerticalScrollbar = false;
    this._drawsTemporaryLayers = true;
    // don't draw shadows on overviews by default
    this.setRenderingHint('drawShadows', false);

    const box = new Part();
    const s = new Shape();
    s.stroke = 'magenta';
    s.strokeWidth = 2;
    s.fill = 'transparent';
    s.name = 'BOXSHAPE';
    box.selectable = true;
    box.selectionAdorned = false;
    box.selectionObjectName = 'BOXSHAPE';
    box.locationObjectName = 'BOXSHAPE';
    box.resizeObjectName = 'BOXSHAPE';
    box.cursor = 'move';
    box.add(s);
    this._box = box;

    this.allowCopy = false;
    this.allowDelete = false;
    this.allowSelect = true;
    this.autoScrollRegion = new Margin(0, 0, 0, 0);
    this._position.assignTo(0, 0);

    this._tempCanvas = new CanvasSurface(null);

    this._tempCtx = this._tempCanvas.context;

    this.toolManager.replaceStandardTool('Dragging', new OverviewDraggingTool(), this.toolManager.mouseMoveTools);

    const that = this;

    this.click = function () {
      const observed = that._observed;
      if (observed === null) return;
      const vpb = observed.viewportBounds;
      const pt = that.lastInput.documentPoint;
      observed.position = new Point(pt.x - vpb.width / 2, pt.y - vpb.height / 2);
    };

    this.observedPanelViewportBoundsChanged = function (e) {
      that.invalidateDocumentBounds();
      that.updateBox();
    };

    this.observedPanelDocumentBoundsChanged = function (e) {
      const observed = that._observed;
      if (observed === null) return;
      that.invalidateDocumentBounds();
      that.invalidateDraw();
    };

    this.observedPanelInvalidateDraw = function (e) {
      that.invalidateDraw();
    };

    this.observedPanelAnimationFinished = function (e) {
      const observed = that._observed;
      if (observed === null) return;
      that.updateBox();
    };

    this.autoScale = Diagram.Uniform;
    this._isAligning = false;
  }

  /**
   * @hidden @internal
   * Uses draw2 instead of draw
   * @this {Overview}
   */
  private requestCachedUpdate() {
    if (this._isAligning || this._isUpdating) return;
    if (this._updateQueued === false) {
      this._updateQueued = true;
      const dia = this;
      root.requestAnimationFrame(
        function () {
          // Are we still queued?
          if (!dia._updateQueued) { return; } // no need!
          dia.update2();
        });
    }
  }

  /**
   * @hidden @internal
   * Overviews use a lame pixel ratio for performance reasons
   * @this {Overview}
   * @return {number}
   */
  public computePixelRatio(): number {
    return 1;
  }

  /**
   * @hidden @internal
   * Ensure that all GraphObjects are arranged, update the document bounds and the scrollbars,
   * and draw all objects in the viewport.
   *
   * **Caution:** this method can be very expensive in time and space.
   * Avoid calling this method except in infrequent circumstances where redrawing immediately is necessary.
   * Under most circumstances you should allow the transaction to perform its normal updating process.
   * Calling {@link #requestUpdate} is the usually the right thing to do.
   * @this {Overview}
   */
  public update2() {
    if (this._isUpdating) return;
    this._updateQueued = false;
    if (this._div === null) return;
    this._isUpdating = true;

    this.validateParts();

    // make sure documentBounds are sane
    // Note that we do not compute the bounds constantly because we only want them to update on a transaction end
    if (!this.documentBounds.isReal()) this.setDocumentBounds(this.computeBounds());

    this.draw2();
    this._isUpdating = false;
  }

  /**
   * @hidden @internal
   * Draw a scaled version of the observed diagram
   * @this {Overview}
   */
  protected draw() {
    if (this._div === null) { Util.throwError('No div specified'); }
    if (this._canvas === null) { Util.throwError('No canvas specified'); }
    if (this._canvas instanceof SVGSurface) return; // NYI
    this.box.validateAdornments();  // ensure any potential box adornments are updated

    if (!this._invalidDraw) return;
    const observed = this._observed;
    if (observed === null) return;
    if (observed.animationManager.isAnimating) return;

    this.validateCanvasSize();

    const grid = observed.grid;
    if (grid !== null && grid.visible && !(isNaN(grid.width) || isNaN(grid.height))) {
      // was observed._updateBackgroundGrid(this.viewportBounds), but then grid only renders with size of overview area
      // It needs to consider both viewports
      const rect = Rect.alloc().assign(this.viewportBounds).unionRect(observed.viewportBounds);
      observed._updateBackgroundGrid(rect);
      Rect.free(rect);
      observed.validateParts();
    }

    const can = this._canvas;
    const ctx = this._ctx;
    const tempcan = this._tempCanvas;
    const tempctx = this._tempCtx;
    // ?? ideally we set this much more rarely and clear the tempcan a different way:
    tempcan.width = can.width;
    tempcan.height = can.height;

    // ?? cached font needs to be reset in case model or something changes. Might be a more efficient place to put this
    ctx.clearContextCache(true);

    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.clearRect(0, 0, can.width, can.height);

    const t = this._transform;
    t.reset();
    if (this.scale !== 1) { t.scale(this.scale); }
    if (this.position.x !== 0 || this.position.y !== 0) {
      t.translate(-this.position.x, -this.position.y);
    }
    // respect pixel ratio in case it's been overridden
    ctx.scale(this._pixelRatio, this._pixelRatio);
    ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

    const showTemporary = this._drawsTemporaryLayers;
    const vb = this.viewportBounds;
    const oblayers = observed._layers._arr;
    const l = oblayers.length;
    for (let i = 0; i < l; i++) {
      oblayers[i].drawAllowed(ctx, vb, this, showTemporary);
    }

    // Once the observed is finished, draw it to the temp canvas:
    tempctx.drawImage(can.domElement, 0, 0);
    if (Debug && Debug.boundsInfoEnabled) {
      tempctx.fillStyle = 'red';
      tempctx.fillRect(0, tempcan.height / 2, tempcan.width, 4);
    }

    const layers = this._layers._arr;
    const ll = layers.length;
    for (let i = 0; i < ll; i++) {
      layers[i].draw(ctx, this);
    }

    if (Debug && (Debug.drawLocationsEnabled || Debug.boundsInfoEnabled) &&  Debug.drawBoundsInfoLocations) {
      Debug.drawBoundsInfoLocations(ctx, this, t);
    }

    // validate
    this._invalidViewportObjects = false;
    this._invalidDraw = false;
  }

  /**
   * @hidden @internal
   * Draw a scaled version of the observed diagram
   * @this {Overview}
   */
  private draw2() {
    if (this._div === null) { Util.throwError('No div specified'); }
    if (this._canvas === null) { Util.throwError('No canvas specified'); }
    this.box.validateAdornments();  // ensure any potential box adornments are updated

    if (!this._invalidDraw) return;
    const observed = this._observed;
    if (observed === null) return;
    if (observed.animationManager.isAnimating) return;
    if (observed.animationManager._pendingAnimation) return;

    const ctx = this._ctx;
    const tempcan = this._tempCanvas;

    ctx.setTransform(1, 0, 0, 1, 0, 0);
    ctx.clearRect(0, 0, this._canvas.width, this._canvas.height);

    // Recall the scene from the temp canvas instead of iterating through all observed objects
    ctx.drawImage(tempcan.domElement, 0, 0);

    const t = this._transform;
    t.reset();
    if (this.scale !== 1) { t.scale(this.scale); }
    if (this.position.x !== 0 || this.position.y !== 0) {
      t.translate(-this.position.x, -this.position.y);
    }
    // respect pixel ratio in case it's been overridden
    ctx.scale(this._pixelRatio, this._pixelRatio);
    ctx.transform(t.m11, t.m12, t.m21, t.m22, t.dx, t.dy);

    const layers = this._layers._arr;
    const ll = layers.length;
    for (let i = 0; i < ll; i++) {
      layers[i].draw(ctx, this);
    }

    // validate
    this._invalidViewportObjects = false;
    this._invalidDraw = false;
  }

  /**
   * Gets or sets the {@link Diagram} for which this Overview is
   * displaying a model and showing its viewport into that model.
   *
   * The value must be null or another Diagram, but may not be an Overview.
   */
  /** @export */ get observed(): Diagram | null { return this._observed; }
  /** @export */ set observed(value: Diagram | null) {
    const old = this._observed;
    if (value !== null) Util.checkClass(value, Diagram, Overview, 'observed');
    if (value instanceof Overview) Util.throwError('Overview.observed Diagram may not be an Overview itself: ' + value);
    if (old !== value) {
      this.unbindFromDiagram(old);
      this._observed = value;
      this.bindToDiagram(value);
      this.invalidateDocumentBounds();
      this.raiseChanged('observed', old, value);
    }
  }

  /**
   * Gets or sets the rectangular {@link Part} that represents the
   * viewport of the {@link #observed} {@link Diagram}.
   * By default the part contains only a magenta {@link Shape}.
   * The box's {@link Part#selectionObject} is what is resized by the Overview to the apparent size of the observed diagram's viewport.
   */
  /** @export */ get box(): Part { return this._box; }
  /** @export */ set box(value: Part) {
    const old = this._box;
    if (old !== value) {
      this._box = value;
      this.remove(old);
      this.add(this._box);
      this.updateBox();

      this.raiseChanged('box', old, value);
    }
  }

  /**
   * Gets or sets whether this overview draws the temporary layers of the observed Diagram.
   * The default value is true.
   * @since 1.2
   */
  /** @export */ get drawsTemporaryLayers(): boolean { return this._drawsTemporaryLayers; }
  /** @export */ set drawsTemporaryLayers(value: boolean) {
    const old = this._drawsTemporaryLayers;
    if (old !== value) {
      this._drawsTemporaryLayers = value;
      this.redraw();
    }
  }

  /**
   * @hidden @internal
   * @this {Overview}
   * @param {Diagram} diagram
   */
  private bindToDiagram(diagram: Diagram) {
    if (diagram === null) return;
    diagram.addDiagramListener('ViewportBoundsChanged', this.observedPanelViewportBoundsChanged);
    diagram.addDiagramListener('DocumentBoundsChanged', this.observedPanelDocumentBoundsChanged);
    diagram.addDiagramListener('InvalidateDraw', this.observedPanelInvalidateDraw);
    diagram.addDiagramListener('AnimationFinished', this.observedPanelAnimationFinished);
    this.add(this.box);
    this.updateBox();
  }

  /**
   * @hidden @internal
   * @this {Overview}
   * @param {Diagram} diagram
   */
  private unbindFromDiagram(diagram: Diagram) {
    if (diagram === null) return;
    this.remove(this.box);
    diagram.removeDiagramListener('ViewportBoundsChanged', this.observedPanelViewportBoundsChanged);
    diagram.removeDiagramListener('DocumentBoundsChanged', this.observedPanelDocumentBoundsChanged);
    diagram.removeDiagramListener('InvalidateDraw', this.observedPanelInvalidateDraw);
    diagram.removeDiagramListener('AnimationFinished', this.observedPanelAnimationFinished);
  }

  /**
   * @hidden @internal
   * update this.box.location
   * @this {Overview}
   */
  private updateBox() {
    const box = this.box;
    if (box === null) return;
    const observed = this._observed;
    if (observed === null) return;
    this._invalidDraw = true;
    const viewportbounds = observed.viewportBounds;
    const selObj = box.selectionObject;
    const dessz = Size.alloc();
    dessz.assignTo(viewportbounds.width, viewportbounds.height);
    selObj.desiredSize = dessz;
    Size.free(dessz);
    const sw = 2 / this.scale;
    if (selObj instanceof Shape) {
      selObj.strokeWidth = sw;
    }
    box.location = new Point(viewportbounds.x - sw / 2, viewportbounds.y - sw / 2);
    box.isSelected = true;  // the box should always be selected, allows for custom tool adornments
  }

  /**
   * @hidden @internal
   * @this {Overview}
   * @return {Rect}
   */
  protected computeBounds(): Rect {
    const observed = this._observed;
    if (observed === null) return Rect.ZeroRect;
    const db = observed.documentBounds.copy();
    // if the viewport is larger, include it in the bounds
    db.unionRect(observed.viewportBounds);
    return db;
  }

  /**
   * @hidden @internal
   * Causes the diagram to redraw.
   * @this {Overview}
   * @param {Rect=} oldv old viewport bounds.
   * @param {Rect=} newv new viewport bounds.
   */
  protected invalidateViewport(oldv?: Rect, newv?: Rect) {
    // only invalidate draw and request an update if the rect is in the viewportBounds
    if (this._invalidDraw === true) return;
    this._invalidDraw = true;
    this.requestCachedUpdate();
  }

  /**
   * @hidden @internal
   * This method is called when the {@link #position} or {@link #scale} change.
   * @this {Overview}
   * @param {Rect} oldvbounds a Rect in document coordinates.
   * @param {Rect} newvbounds a Rect in document coordinates.
   * The X or Y may have changed because of a change to {@link #position}.
   * The Width or Height may have changed either because this Diagram has changed size
   * or because the {@link #scale} has changed value.
   * @param {number} oldscale
   * @param {boolean} isScroll true if the viewport change was only caused by a scrollbar changing size.
   */
  protected onViewportBoundsChanged(oldvbounds: Rect, newvbounds: Rect, oldscale: number, isScroll: boolean) {
    if (this._isAligning) return;
    this.invalidateViewportObjects();
    this.invalidateDraw();
    this.invalidateScroll();
    this.invalidateDocumentBounds();
    this.updateBox();
    this._viewportBoundsChangedInfo.scale = oldscale;
    this._viewportBoundsChangedInfo.position.x = oldvbounds.x;
    this._viewportBoundsChangedInfo.position.y = oldvbounds.y;
    this._viewportBoundsChangedInfo.bounds.assign(oldvbounds);
    this._viewportBoundsChangedInfo.isScroll = isScroll;
    this.raiseDiagramEvent('ViewportBoundsChanged', this._viewportBoundsChangedInfo, oldvbounds);
  }

} // #end Overview
Overview['className'] = 'Overview';

/**
 * @hidden @internal
 * @extends DraggingTool
 * @unrestricted
 * @category Tool
 */
export class OverviewDraggingTool extends DraggingTool {
  /** @hidden @internal */ public static className = 'OverviewDraggingTool';
  /** @hidden @internal */ private _offset: Point;

  constructor() {
    super();
    this._offset = null;
  }

  /**
   * @hidden @internal
   */
  public canStart(): boolean {
    if (!this.isEnabled) return false;

    const overview = this.diagram as Overview;
    if (overview === null) return false;
    if (!overview.allowMove) return false;
    if (!overview.allowSelect) return false;

    const observed = overview.observed;  // no observed Diagram?
    if (observed === null) return false;

    // require left button & that it has moved far enough away from the mouse down point,
    // so it isn't a click
    const e = overview.lastInput;
    if (!e.left) return false;
    // don't include the following check when this tool is running modally
    if (overview.currentTool !== this) {
      if (!this.isBeyondDragSize()) return false;
      if (e.isTouchEvent) {
        // must wait for "delay" milliseconds before that tool can run
        if (e.timestamp - overview.firstInput.timestamp < this.delay) return false;
      }
    }

    // need to be over a part that is movable
    const part = this.findDraggablePart();
    if (part === null) {  // dragging outside of box
      const vpb = observed.viewportBounds;
      this._offset = new Point(vpb.width / 2, vpb.height / 2);
      // just go there directly and start dragging anyway
      const pos = overview.firstInput.documentPoint;
      observed.position = new Point(pos.x - this._offset.x, pos.y - this._offset.y);
    }
    return true;
  }

  /**
   * @hidden @internal
   */
  public doActivate(): void {
    this._offset = null;
    super.doActivate();
  }

  /**
   * @hidden @internal
   * @param {Map.<Part,DraggingInfo>} parts  a Map mapping {@link Part}s to DraggingInfo Objects
   *        that have a "point" property remembering the original location of that Part.
   * @param {Point} offset
   * @param {boolean} check  Whether to check {@link Part#canMove} on each part.
   */
  public moveParts(parts: Map<Part, DraggingInfo>, offset: Point, check: boolean): void {
    const overview = this.diagram as Overview;
    const observed = overview.observed;
    if (observed === null) return;
    const box = overview.box;
    if (box === null) return;
    if (this._offset === null) {
      const pos = overview.firstInput.documentPoint;
      const loc = box.location;
      this._offset = new Point(pos.x - loc.x, pos.y - loc.y);
    }
    const p = overview.lastInput.documentPoint;
    observed.position = new Point(p.x - this._offset.x, p.y - this._offset.y);
  }

} // #end OverviewDraggingTool
OverviewDraggingTool['className'] = 'OverviewDraggingTool';

// import { CanvasSurface, CanvasSurfaceContext } from './abstractCanvas';
// import { Map } from './collections';
// import { Debug } from './debug';
// import { DiagramEvent } from './inputEvent';
// import { Margin } from './margin';
// import { Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { root } from './root';
// import { Shape } from './shape';
// import { Size } from './size';
// import { SVGSurface } from './svgContext';
// import { Util } from './util';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
CommandHandler
*/


// import { List, Map, Set } from './collections';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { ContextMenuTool } from './contextMenuTool';
// import { Debug } from './debug';
// import { Diagram, DraggingInfo, DraggingOptions } from './diagram';
// import { GraphObject } from './graphObject';
// import { Adornment, Group, Part } from './parts';
// import { Node } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData } from './root';
// import { Spot } from './spot';
// import { TextBlock } from './textBlock';
// import { TextEditingTool } from './textEditingTool';
// import { ToolManager } from './toolManager';
// import { Util } from './util';

/** @hidden @internal */
export interface ICommandHandler {
  // not exposed:
  setDiagram(val: Diagram): void;
  copiesClipboardData: boolean;
  findUnnestedNodes(coll: Iterable<Part>): Set<Part>;

  // exposed:
  readonly diagram: Diagram;
  doKeyDown(): void;
  doKeyUp(): void;
  stopCommand(): void;
  canStopCommand(): boolean;
  selectAll(): void;
  canSelectAll(): boolean;
  deleteSelection(): void;
  canDeleteSelection(): boolean;
  copySelection(): void;
  canCopySelection(): boolean;
  cutSelection(): void;
  canCutSelection(): boolean;
  copyToClipboard(coll: Iterable<Part>): void;
  pasteFromClipboard(): Set<Part>;
  pasteSelection(pos?: Point): void;
  canPasteSelection(pos?: Point): boolean;
  undo(): void;
  canUndo(): boolean;
  redo(): void;
  canRedo(): boolean;
  decreaseZoom(factor?: number): void;
  canDecreaseZoom(factor?: number): boolean;
  increaseZoom(factor?: number): void;
  canIncreaseZoom(factor?: number): boolean;
  resetZoom(newscale?: number): void;
  canResetZoom(newscale?: number): boolean;
  zoomToFit(): void;
  canZoomToFit(): boolean;
  scrollToPart(part?: Part): void;
  canScrollToPart(part?: Part): boolean;
  collapseTree(node?: Node): void;
  canCollapseTree(node?: Node): boolean;
  expandTree(node?: Node): void;
  canExpandTree(node?: Node): boolean;
  groupSelection(): void;
  canGroupSelection(): boolean;
  isValidMember(group: Group, part: Part): boolean;
  ungroupSelection(group?: Group): void;
  canUngroupSelection(group?: Group): boolean;
  addTopLevelParts(coll: Iterable<Part>, check?: boolean): boolean;
  collapseSubGraph(group?: Group): void;
  canCollapseSubGraph(group?: Group): boolean;
  expandSubGraph(group?: Group): void;
  canExpandSubGraph(group?: Group): boolean;
  editTextBlock(textblock?: TextBlock): void;
  canEditTextBlock(textblock?: TextBlock): boolean;
  showContextMenu(obj?: GraphObject | Diagram): void;
  canShowContextMenu(obj?: GraphObject | Diagram): boolean;
  copiesConnectedLinks: boolean;
  deletesConnectedLinks: boolean;
  copiesTree: boolean;
  deletesTree: boolean;
  copiesParentKey: boolean;
  copiesGroupKey: boolean;
  archetypeGroupData: ObjectData;
  memberValidation: ((aGroup: Group, somePart: Part) => boolean) | null;
  defaultScale: number;
  zoomFactor: number;
  computeEffectiveCollection(parts: Iterable<Part>, options: DraggingOptions): Map<Part, DraggingInfo>;
}


/**
 * The {@link Diagram#commandHandler} implements various
 * commands such as {@link CommandHandler#deleteSelection} or {@link CommandHandler#redo}.
 * The CommandHandler includes keyboard event handling to interpret
 * key presses as commands.
 *
 * CommandHandlers cannot be shared amongst multiple Diagrams.
 *
 * You may define a CommandHandler subclass and override methods.
 * However you must seriously consider calling the base method in order to get its default behavior.
 * There may be situations where not calling the base method may cause subtle bugs,
 * but that depends on the method.
 * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call a base method.
 *
 * There is an example custom CommandHandler in the extensions directory: <a href="../../extensions/DrawCommandHandler.js">DrawCommandHandler.js</a>,
 * which provides alignment commands and additional behaviors for the arrow keys.
 * <p class="boxread">
 * For additional discussion, please read the <a href="../../intro/commands.html">Introduction page on Commands</a>.
 * <h3>Keyboard Shortcuts</h3>
 *
 * The CommandHandler implements the following command bindings for keyboard input in {@link #doKeyDown}:
 *   - `Ctrl-X` & `Shift-Del` invoke {@link #cutSelection}
 *   - `Ctrl-C` & `Ctrl-Insert` invoke {@link #copySelection}
 *   - `Ctrl-V` & `Shift-Insert` invoke {@link #pasteSelection}
 *   - `Del` & `Backspace` invoke {@link #deleteSelection}
 *   - `Ctrl-A` invokes {@link #selectAll}
 *   - `Ctrl-Z` & `Alt-Backspace` invoke {@link #undo}
 *   - `Ctrl-Y` & `Alt-Shift-Backspace` invoke {@link #redo}
 *   - `Up` & `Down` & `Left` & `Right` (arrow keys) call {@link Diagram#scroll}
 *   - `PageUp` & `PageDown` call {@link Diagram#scroll}
 *   - `Home` & `End` call {@link Diagram#scroll}
 *   - `Space` invokes {@link #scrollToPart}
 *   - `Ctrl-- & Keypad--` (minus) invoke {@link #decreaseZoom}
 *   - `Ctrl-+ & Keypad-+` (plus) invoke {@link #increaseZoom}
 *   - `Ctrl-0` invokes {@link #resetZoom}
 *   - `Shift-Z` invokes {@link #zoomToFit}; repeat to return to the original scale and position
 *   - `Ctrl-G` invokes {@link #groupSelection}
 *   - `Ctrl-Shift-G` invokes {@link #ungroupSelection}
 *   - `F2` invokes {@link #editTextBlock}
 *   - `Menu Key` invokes {@link #showContextMenu}
 *   - `Esc` invokes {@link #stopCommand}
 *
 * On a Macintosh the Command key is used as the modifier instead of the Control key.
 *
 * On touch devices there is a default context menu that shows many commonly-used commands
 * when you hold a finger down on the diagram.
 * @unrestricted
 */
export class CommandHandler implements ICommandHandler {
  /** @hidden @internal */ public static className = 'CommandHandler';
  /** @hidden @internal */ private _diagram: Diagram;
  /** @hidden @internal */ private _copiesClipboardData: boolean;
  /** @hidden @internal */ private _copiesConnectedLinks: boolean;
  /** @hidden @internal */ private _deletesConnectedLinks: boolean;
  /** @hidden @internal */ private _copiesTree: boolean;
  /** @hidden @internal */ private _deletesTree: boolean;
  /** @hidden @internal */ private _copiesParentKey: boolean;
  /** @hidden @internal */ private _copiesGroupKey: boolean;
  /** @hidden @internal */ private _archetypeGroupData: ObjectData;
  /** @hidden @internal */ private _memberValidation: ((aGroup: Group, somePart: Part) => boolean) | null;
  /** @hidden @internal */ private _zoomFactor: number;
  /** @hidden @internal */ private _lastScale: number;
  /** @hidden @internal */ private _lastPosition: Point;
  /** @hidden @internal */ private _zoomScale: number;
  /** @hidden @internal */ private _zoomBounds: Rect;
  /** @hidden @internal */ private _scrollToPartIterator: Iterator<Part> | null;
  /** @hidden @internal */ private _scrollToPartPause: number;

  /**
   * The constructor produces a CommandHandler with the default key bindings.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(CommandHandler);
    Set.uniqueHash(this);

    this._diagram = Diagram.getLastDiagram();  // replaced by setDiagram, called by Diagram.commandHandler setter
    this._copiesClipboardData = true;
    this._copiesConnectedLinks = true;
    this._deletesConnectedLinks = true;
    this._copiesTree = false;
    this._deletesTree = false;
    this._copiesParentKey = false;
    this._copiesGroupKey = false;
    this._archetypeGroupData = null;
    this._memberValidation = null;
    this._zoomFactor = 1.05;
    this._lastScale = NaN;
    this._lastPosition = null;
    this._zoomScale = NaN;
    this._zoomBounds = Rect.ZeroRect;
    this._scrollToPartIterator = null;
    this._scrollToPartPause = 200;
  }

  /**
   * @hidden @internal
   * @this {CommandHandler}
   * @return {string} A human-readable string.
   */
  public toString(): string {
    return 'CommandHandler';
  }

  /**
   * This read-only property returns the {@link Diagram} that is using this CommandHandler,
   * after {@link Diagram#commandHandler} has been set to this object.
   */
  /** @export */ get diagram(): Diagram { return this._diagram; }

  /**
   * @hidden @internal
   * @this {CommandHandler}
   * @param {Diagram} val
   */
  public setDiagram(val: Diagram): void { this._diagram = val; }

  /**
   * This is called by tools to handle keyboard commands.
   * For most commands, this calls the "can..." predicate; if that returns true it calls the command method.
   * If GoJS handles a key-down event as a keyboard command, the underlying event will not bubble.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * There is an example custom CommandHandler in the extensions directory: <a href="../../extensions/DrawCommandHandler.js">DrawCommandHandler.js</a>,
   * which implements additional behaviors for the arrow keys by overriding this method.
   * For additional discussion, please read the <a href="../../intro/commands.html">Introduction page on Commands</a>.
   * @expose
   * @this {CommandHandler}
   */
  public doKeyDown(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    // The meta (Command) key substitutes for "control" for Mac commands
    const control = Util.isMac ? e.meta : e.control;
    const shift = e.shift;
    const alt = e.alt;
    const key = e.key;

    if (control && (key === 'C' || key === 'Insert')) {
      if (this.canCopySelection()) this.copySelection();
    } else if ((control && key === 'X') || (shift && key === 'Del')) {
      if (this.canCutSelection()) this.cutSelection();
    } else if ((control && key === 'V') || (shift && key === 'Insert')) {
      if (this.canPasteSelection()) this.pasteSelection();
    } else if ((control && key === 'Y') || (alt && shift && key === 'Backspace')) {
      if (this.canRedo()) this.redo();
    } else if ((control && key === 'Z') || (alt && key === 'Backspace')) {
      if (this.canUndo()) this.undo();
    } else if (key === 'Del' || key === 'Backspace') {
      if (this.canDeleteSelection()) this.deleteSelection();
    } else if (control && key === 'A') {
      if (this.canSelectAll()) this.selectAll();
    } else if (key === 'Esc') {
      if (this.canStopCommand()) this.stopCommand();
    } else if (key === 'Up') {
      if (diagram.allowVerticalScroll) {
        if (control) {
          diagram.scroll('pixel', 'up');
        } else {
          diagram.scroll('line', 'up');
        }
      }
    } else if (key === 'Down') {
      if (diagram.allowVerticalScroll) {
        if (control) {
          diagram.scroll('pixel', 'down');
        } else {
          diagram.scroll('line', 'down');
        }
      }
    } else if (key === 'Left') {
      if (diagram.allowHorizontalScroll) {
        if (control) {
          diagram.scroll('pixel', 'left');
        } else {
          diagram.scroll('line', 'left');
        }
      }
    } else if (key === 'Right') {
      if (diagram.allowHorizontalScroll) {
        if (control) {
          diagram.scroll('pixel', 'right');
        } else {
          diagram.scroll('line', 'right');
        }
      }
    } else if (key === 'PageUp') {
      if (shift && diagram.allowHorizontalScroll) {
        diagram.scroll('page', 'left');
      } else if (diagram.allowVerticalScroll) {
        diagram.scroll('page', 'up');
      }
    } else if (key === 'PageDown') {
      if (shift && diagram.allowHorizontalScroll) {
        diagram.scroll('page', 'right');
      } else if (diagram.allowVerticalScroll) {
        diagram.scroll('page', 'down');
      }
    } else if (key === 'Home') {
      if (control && diagram.allowVerticalScroll) {
        diagram.scroll('document', 'up');
      } else if (!control && diagram.allowHorizontalScroll) {
        diagram.scroll('document', 'left');
      }
    } else if (key === 'End') {
      if (control && diagram.allowVerticalScroll) {
        diagram.scroll('document', 'down');
      } else if (!control && diagram.allowHorizontalScroll) {
        diagram.scroll('document', 'right');
      }
    } else if (key === ' ') {  // Space
      if (this.canScrollToPart()) this.scrollToPart();
    } else if (key === 'Subtract') {  // captures both - and keypad -
      if (this.canDecreaseZoom()) this.decreaseZoom();
    } else if (key === 'Add') {  // captures both + and keypad +
      if (this.canIncreaseZoom()) this.increaseZoom();
    } else if (control && key === '0') {
      if (this.canResetZoom()) this.resetZoom();
    } else if (shift && key === 'Z') {
      if (this.canZoomToFit()) this.zoomToFit();
    } else if (control && !shift && key === 'G') {
      if (this.canGroupSelection()) this.groupSelection();
    } else if (control && shift && key === 'G') {
      if (this.canUngroupSelection()) this.ungroupSelection();
    } else if (e.event && (e.event as KeyboardEvent).which === 113) { // F2
      if (this.canEditTextBlock()) this.editTextBlock();
    } else if (e.event && (e.event as KeyboardEvent).which === 93) { // Context Menu button
      if (this.canShowContextMenu()) this.showContextMenu();
    } else {
      e.bubbles = true;  // pass on other keys up to the browser
    }
  }

  /**
   * This is called by tools to handle keyboard commands.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   */
  public doKeyUp(): void {
    const diagram = this.diagram;
    const e = diagram.lastInput;
    e.bubbles = true;  // pass on other keys up to the browser
  }

  /**
   * This command cancels the operation of the current tool.
   * This is typically called when the user presses ESCAPE.
   *
   * If the current tool is a {@link ToolManager}, this clears the diagram's selection.
   * This then calls {@link Tool#doCancel} on the current tool.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canStopCommand}
   */
  public stopCommand(): void {
    const diagram = this.diagram;
    const tool = diagram.currentTool;
    if (tool instanceof ToolManager && diagram.allowSelect) {
      diagram.clearSelection();  // raises ChangingSelection/Finished
    }
    if (tool !== null) {
      tool.doCancel();
    }
  }

  /**
   * This predicate controls whether the user may stop the current tool.
   * This just returns true.
   *
   * This method may be overridden, but probably should not be overridden.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean} true.
   * @see {@link #stopCommand}
   */
  public canStopCommand(): boolean {
    return true;
  }

  /**
   * This command selects all of the selectable {@link Part}s in the diagram by setting {@link Part#isSelected} to true on each one.
   * This is normally invoked by the `Ctrl-A` keyboard shortcut.
   *
   * This method raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * This ignores all parts in temporary layers.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canSelectAll}
   */
  public selectAll(): void {
    const diagram = this.diagram;
    // since every object on the screen will probably need a redraw,
    // might as well skip viewport optimizations:
    diagram.invalidateDraw();
    try {
      diagram.currentCursor = 'wait';
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      const pit = diagram.parts;
      while (pit.next()) {
        const part = pit.value;
        part.isSelected = true;
      }
      const nit = diagram.nodes;
      while (nit.next()) {
        const node = nit.value;
        node.isSelected = true;
      }
      const lit = diagram.links;
      while (lit.next()) {
        const link = lit.value;
        link.isSelected = true;
      }
    } finally {
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      diagram.currentCursor = '';
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #selectAll} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean} This returns true if {@link Diagram#allowSelect} is true.
   * @see {@link #selectAll}
   */
  public canSelectAll(): boolean {
    const diagram = this.diagram;
    return diagram.allowSelect;
  }

  /**
   * This command deletes the currently selected parts from the diagram.
   * This is normally invoked by the `Del` keyboard shortcut.
   *
   * This will first start a "Delete" transaction, then
   * raise the "SelectionDeleting" {@link DiagramEvent},
   * call {@link Diagram#removeParts} on a perhaps extended collection of selected Parts,
   * raise the "SelectionDeleted" diagram event,
   * and finally commit the transaction.
   *
   * Because this command changes the selection,
   * this method also raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed within a transaction, but the selection events are raised outside the transaction.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canDeleteSelection}
   */
  public deleteSelection(): void {
    const diagram = this.diagram;
    try {
      diagram.currentCursor = 'wait';
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      diagram.startTransaction('Delete');
      diagram.raiseDiagramEvent('SelectionDeleting', diagram.selection);
      // collect all of the parts to delete
      const allobjs = new Set<Part>();
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        Part.gatherCollection(allobjs, part, true, this.deletesTree ? Infinity : 0, this.deletesConnectedLinks ? null : false,
          function (p) { return p.canDelete(); });
      }
      // remove all of the deletable parts
      diagram.removeParts(allobjs, true);
      diagram.raiseDiagramEvent('SelectionDeleted', allobjs);
    } finally {
      diagram.commitTransaction('Delete');
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      diagram.currentCursor = '';
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #deleteSelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean} This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowDelete} is true, and
   * if there are some selected {@link Part}s.
   * @see {@link #deleteSelection}
   */
  public canDeleteSelection(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowDelete) return false;
    if (diagram.selection.count === 0) return false;
    return true;
  }

  // clipboard commands

  /**
   * This command copies the currently selected parts, {@link Diagram#selection}, from the {@link Diagram} into the clipboard.
   * This is normally invoked by the `Ctrl-C` keyboard shortcut.
   *
   * This makes a copy of the current selection by calling {@link #copyToClipboard}.
   * This also raises the "ClipboardChanged" diagram event.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canCopySelection}
   */
  public copySelection(): void {
    const diagram = this.diagram;
    try {
      // collect all of the parts to copy
      const allobjs = new Set<Part>();
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        Part.gatherCollection(allobjs, part, true, this.copiesTree ? Infinity : 0, this.copiesConnectedLinks,
          function (p) { return p.canCopy(); });
      }
      // copy to clipboard
      this.copyToClipboard(allobjs);
    } finally {
      // empty
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #copySelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true:
   * if {@link Diagram#allowCopy} is true,
   * if {@link Diagram#allowClipboard} is true, and
   * if there are some selected {@link Part}s.
   * @see {@link #copySelection}
   */
  public canCopySelection(): boolean {
    const diagram = this.diagram;
    if (!diagram.allowCopy || !diagram.allowClipboard) return false;
    if (diagram.selection.count === 0) return false;
    return true;
  }

  /**
   * This command executes a {@link #copySelection} followed by a {@link #deleteSelection}.
   * This is normally invoked by the `Ctrl-X` keyboard shortcut.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canCutSelection}
   */
  public cutSelection(): void {
    this.copySelection();
    this.deleteSelection();
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #cutSelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowCopy} is true,
   * if {@link Diagram#allowDelete} is true,
   * if {@link Diagram#allowClipboard} is true, and
   * if there are some selected {@link Part}s.
   * @see {@link #cutSelection}
   */
  public canCutSelection(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowCopy || !diagram.allowDelete || !diagram.allowClipboard) return false;
    if (diagram.selection.count === 0) return false;
    return true;
  }

  /**
   * Make a copy of the given collection of {@link Part}s
   * and stores it in a static variable acting as the clipboard.
   *
   * The clipboard is initially null.
   * It can hold a collection of copied {@link Part}s.
   * It also remembers the {@link Model#dataFormat} of the diagram
   * from which the parts were copied.
   *
   * This calls {@link Diagram#copyParts} in order to make a copy of the Parts for the clipboard.
   * The values of {@link #copiesParentKey} and {@link #copiesGroupKey} affect whether
   * a copied node data remembers its tree parent node (if in a {@link TreeModel}) or
   * its containing group (if in a {@link GraphLinksModel}).
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Iterable.<Part>} coll A collection of {@link Part}s.
   * If the value is null, the clipboard is cleared of all data.
   * @see {@link #pasteFromClipboard}
   */
  public copyToClipboard(coll: Iterable<Part>): void {
    const diagram = this.diagram;
    let clip: List<Part> = null;
    if (coll === null) {
      Diagram.clearClipboard();
    } else {
      const model = diagram.model;
      let oldcopytree = false;
      let oldcopygroup = false;
      let dict: Map<Part, Part> = null;
      try {
        if (model.isTreeModel()) {
          oldcopytree = model.copiesParentKeyOfNodeData;
          model.copiesParentKeyOfNodeData = this.copiesParentKey;
        }
        if (model.isGraphLinksModel() || model.isGraphModel()) {
          oldcopygroup = model.copiesGroupKeyOfNodeData;
          model.copiesGroupKeyOfNodeData = this.copiesGroupKey;
        }
        dict = diagram.copyParts(coll, null, true);
      } finally {
        if (model.isTreeModel()) {
          model.copiesParentKeyOfNodeData = oldcopytree;
        }
        if (model.isGraphLinksModel() || model.isGraphModel()) {
          model.copiesGroupKeyOfNodeData = oldcopygroup;
        }
        clip = new List<Part>();
        clip.addAll(dict as any); // JStoTS: Map not simple Iterable
        Diagram._clipboardParts = clip;
        Diagram._clipboardFormat = diagram.model.dataFormat;
      }
    }
    diagram.raiseDiagramEvent('ClipboardChanged', clip);
  }

  /**
   * If the clipboard holds a collection of {@link Part}s,
   * and if the {@link Model#dataFormat} matches that stored in the clipboard,
   * this makes a copy of the clipboard's parts and adds the copies to this {@link Diagram}.
   *
   * This calls {@link Diagram#copyParts} in order to make a copy of the Parts in the clipboard and add them to this diagram.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {Set.<Part>} a collection of the newly pasted {@link Part}s,
   * or an empty Set if there was no data in the clipboard.
   * @see {@link #copyToClipboard}
   */
  public pasteFromClipboard(): Set<Part> {
    const coll = new Set<Part>();
    const clip = Diagram._clipboardParts;
    if (clip === null) return coll;
    const diagram = this.diagram;
    if (Diagram._clipboardFormat !== diagram.model.dataFormat) return coll;

    const model = diagram.model;
    let oldcopytree = false;
    let oldcopygroup = false;
    let map: Map<Part, Part> = null;
    // ?? this is not the only thing needed for copy/paste to work
    //    this._suspendRouting(clip);
    try {
      if (model.isTreeModel()) {
        oldcopytree = model.copiesParentKeyOfNodeData;
        model.copiesParentKeyOfNodeData = this.copiesParentKey;
      }
      if (model.isGraphLinksModel() || model.isGraphModel()) {
        oldcopygroup = model.copiesGroupKeyOfNodeData;
        model.copiesGroupKeyOfNodeData = this.copiesGroupKey;
      }
      map = diagram.copyParts(clip, diagram, false);
    } finally {
      if (model.isTreeModel()) {
        model.copiesParentKeyOfNodeData = oldcopytree;
      }
      if (model.isGraphLinksModel() || model.isGraphModel()) {
        model.copiesGroupKeyOfNodeData = oldcopygroup;
      }
      // try to assign a location or position to each of the newly copied Parts
      const it = map.iterator;
      while (it.next()) {
        const copy = it.value;
        const orig = it.key;
        if (!copy.location.isReal()) {
          if (orig.location.isReal()) copy.location = orig.location;
          else if (!copy.position.isReal() && orig.position.isReal()) copy.position = orig.position;
        }
        coll.add(copy);
      }
    }
    // ?? this is not the only thing needed for copy/paste to work
    //    this._resumeRouting(clip);
    return coll;
  }

  /**
   * This command copies the contents of the clipboard into this diagram and makes those new parts the new selection.
   * This is normally invoked by the `Ctrl-V` keyboard shortcut.
   *
   * This calls {@link #pasteFromClipboard} to add copies of {@link Part}s
   * into this diagram, and then selects all of the newly created parts.
   * This also raises the "ClipboardPasted" diagram event.
   * This method raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed in a transaction, but the selection events are raised outside the transaction.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Point=} pos Point at which to center the newly pasted parts; if not present the parts are not moved.
   * @see {@link #canPasteSelection}
   */
  public pasteSelection(pos?: Point): void {
    if (pos === undefined) pos = null;
    const diagram = this.diagram;
    try {
      diagram.currentCursor = 'wait';
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      diagram.startTransaction('Paste');
      const copies = this.pasteFromClipboard();
      // now update the Diagram.selection
      if (copies.count > 0) diagram.internalClearSelection();  // don't raise ChangingSelection/Finished
      const it = copies.iterator;
      while (it.next()) {
        const part = it.value;
        part.isSelected = true;
      }
      if (pos !== null) {
        const b = diagram.computePartsBounds(diagram.selection);
        if (b.isReal()) {
          const coll = this.computeEffectiveCollection(diagram.selection, diagram._defaultDragOptions);
          diagram.moveParts2(coll, new Point(pos.x - b.centerX, pos.y - b.centerY), diagram._defaultDragOptions, false);
        }
      }
      diagram.raiseDiagramEvent('ClipboardPasted', copies);
    } finally {
      diagram.commitTransaction('Paste');
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      diagram.currentCursor = '';
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #pasteSelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Point=} pos Point at which to center the newly pasted parts; if not present the parts would not be moved.
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowInsert} is true,
   * if {@link Diagram#allowClipboard} is true, and
   * if the clipboard has parts in it.
   * @see {@link #pasteSelection}
   */
  public canPasteSelection(pos?: Point): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowInsert || !diagram.allowClipboard) return false;
    if (Diagram._clipboardParts === null || Diagram._clipboardParts.count === 0) return false;
    if (Diagram._clipboardFormat !== diagram.model.dataFormat) return false;
    return true;
  }

  // undo commands

  /**
   * This command calls {@link UndoManager#undo}.
   * This is normally invoked by the `Ctrl-Z` keyboard shortcut.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canUndo}
   */
  public undo(): void {
    const diagram = this.diagram;
    diagram.undoManager.undo();
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #undo} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowUndo} is true, and
   * if the {@link UndoManager#canUndo} predicate returns true.
   * @see {@link #undo}
   */
  public canUndo(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    return (diagram.allowUndo && diagram.undoManager.canUndo());
  }

  /**
   * This command calls {@link UndoManager#redo}.
   * This is normally invoked by the `Ctrl-Y` keyboard shortcut.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canRedo}
   */
  public redo(): void {
    const diagram = this.diagram;
    diagram.undoManager.redo();
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #redo} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowUndo} is true, and
   * if the {@link UndoManager#canRedo} predicate returns true.
   * @see {@link #redo}
   */
  public canRedo(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    return (diagram.allowUndo && diagram.undoManager.canRedo());
  }

  // zooming commands

  /**
   * This command decreases the {@link Diagram#scale} by a given factor.
   * This is normally invoked by the `Ctrl--` and `Keypad--` keyboard shortcuts.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} factor This defaults to 1/{@link #zoomFactor}.  The value should be less than one and greater than zero.
   * @see {@link #canDecreaseZoom}
   */
  public decreaseZoom(factor?: number): void {
    if (factor === undefined) factor = 1.0 / this.zoomFactor;
    Util.checkRealNumber(factor, CommandHandler, 'decreaseZoom:factor');
    const diagram = this.diagram;
    if (diagram.autoScale !== Diagram.None) return;
    const newscale = diagram.scale * factor;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return;
    diagram.scale = newscale;
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #decreaseZoom} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} factor This defaults to 1/{@link #zoomFactor}.  The value should be less than one and greater than zero.
   * @return {boolean}
   * This returns true if {@link Diagram#allowZoom} is true
   * and if the new scale is within the range of {@link Diagram#minScale} and {@link Diagram#maxScale}.
   * @see {@link #decreaseZoom}
   */
  public canDecreaseZoom(factor?: number): boolean {
    if (factor === undefined) factor = 1.0 / this.zoomFactor;
    Util.checkRealNumber(factor, CommandHandler, 'canDecreaseZoom:factor');
    const diagram = this.diagram;
    if (diagram.autoScale !== Diagram.None) return false;
    const newscale = diagram.scale * factor;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return false;
    return diagram.allowZoom;
  }

  /**
   * This command increases the {@link Diagram#scale} by a given factor.
   * This is normally invoked by the `Ctrl-+` and `Keypad-+` keyboard shortcuts.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} factor This defaults to {@link #zoomFactor}.  The value should be greater than one.
   * @see {@link #canIncreaseZoom}
   */
  public increaseZoom(factor?: number): void {
    if (factor === undefined) factor = this.zoomFactor;
    Util.checkRealNumber(factor, CommandHandler, 'increaseZoom:factor');
    const diagram = this.diagram;
    if (diagram.autoScale !== Diagram.None) return;
    const newscale = diagram.scale * factor;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return;
    diagram.scale = newscale;
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #increaseZoom} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} factor This defaults to {@link #zoomFactor}.  The value should be greater than one.
   * @return {boolean}
   * This returns true if {@link Diagram#allowZoom} is true
   * and if the new scale is within the range of {@link Diagram#minScale} and {@link Diagram#maxScale}.
   * @see {@link #increaseZoom}
   */
  public canIncreaseZoom(factor?: number): boolean {
    if (factor === undefined) factor = this.zoomFactor;
    Util.checkRealNumber(factor, CommandHandler, 'canIncreaseZoom:factor');
    const diagram = this.diagram;
    if (diagram.autoScale !== Diagram.None) return false;
    const newscale = diagram.scale * factor;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return false;
    return diagram.allowZoom;
  }

  /**
   * This command sets the {@link Diagram#scale} to a new scale value, by default 1.
   * This is normally invoked by the `Ctrl-0` keyboard shortcut.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} newscale This defaults to {@link #defaultScale}, which is normally 1.0.  The value should be greater than zero.
   * @see {@link #canResetZoom}
   */
  public resetZoom(newscale?: number): void {
    if (newscale === undefined) newscale = this.defaultScale;
    Util.checkRealNumber(newscale, CommandHandler, 'resetZoom:newscale');
    const diagram = this.diagram;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return;
    diagram.scale = newscale;
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #resetZoom} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {number=} newscale This defaults to {@link #defaultScale}, which is normally 1.0.  The value should be greater than zero.
   * @return {boolean}
   * This returns true if {@link Diagram#allowZoom} is true.
   * and if the new scale is within the range of {@link Diagram#minScale} and {@link Diagram#maxScale}.
   * @see {@link #resetZoom}
   */
  public canResetZoom(newscale?: number): boolean {
    if (newscale === undefined) newscale = this.defaultScale;
    Util.checkRealNumber(newscale, CommandHandler, 'canResetZoom:newscale');
    const diagram = this.diagram;
    if (newscale < diagram.minScale || newscale > diagram.maxScale) return false;
    return diagram.allowZoom;
  }

  /**
   * This command changes the {@link Diagram#scale} so that the {@link Diagram#documentBounds} fits within the viewport.
   * If this command had been called before without any other zooming since then,
   * the original Diagram scale and position are restored.
   * This is normally invoked by the `Shift-Z` keyboard shortcut.
   *
   * As of 2.1, this animates zooming by default. {@link Diagram#zoomToFit} does not animate.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canZoomToFit}
   * @since 1.1
   */
  public zoomToFit(): void {
    const diagram = this.diagram;
    const AM = diagram.animationManager;
    AM.stopAnimation();
    diagram.redraw(); // was diagram.ensureDocumentBounds();, changed to accomodate multiple zoom to fits in a row with animation
    const position = diagram.position;
    const scale = diagram.scale;
    AM.prepareAutomaticAnimation('Zoom To Fit');
    // don't restore original viewport if the scale or the document bounds have changed
    if (scale === this._zoomScale && !isNaN(this._lastScale) && diagram.documentBounds.equals(this._zoomBounds)) {
      // restore the original viewport
      diagram.scale = this._lastScale;
      diagram.position = this._lastPosition;
      this._zoomScale = NaN;
      this._zoomBounds = Rect.ZeroRect;
    } else {  // zoom to fit, after remembering original viewport and document bounds
      this._lastScale = scale;
      this._lastPosition = position.copy();
      diagram.zoomToFit();
      this._zoomScale = diagram.scale;
      this._zoomBounds = diagram.documentBounds.copy();
    }
    AM.beginAnimation();
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #zoomToFit} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true if {@link Diagram#allowZoom} is true.
   * @see {@link #zoomToFit}
   * @since 1.1
   */
  public canZoomToFit(): boolean {
    const diagram = this.diagram;
    return diagram.allowZoom;
  }

  /**
   * This command scrolls the diagram to make a highlighted or selected Part visible in the viewport.
   * Call this command repeatedly to cycle through the {@link Diagram#highlighteds} collection,
   * if there are any Parts in that collection, or else in the {@link Diagram#selection} collection,
   * scrolling to each one in turn by calling {@link Diagram#centerRect}.
   *
   * This is normally invoked by the `Space` keyboard shortcut.
   * If there is no argument and there is no highlighted or selected Part, this command does nothing.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Part=} part This defaults to the first highlighted Part of {@link Diagram#highlighteds},
   *                     or, if there are no highlighted Parts, the first selected Part.
   * @see {@link #canScrollToPart}
   * @since 1.6
   */
  public scrollToPart(part?: Part): void {
    if (part === undefined) part = null;
    if (part !== null) Util.checkClass(part, Part, CommandHandler, 'part');
    const diagram = this.diagram;
    diagram.ensureDocumentBounds();
    if (part === null) {
      try {
        if (this._scrollToPartIterator !== null) {
          if (this._scrollToPartIterator.next()) {  // get the next highlighted/selected part
            part = this._scrollToPartIterator.value;
          } else {
            this._scrollToPartIterator = null;  // finished iteration
          }
        }
      } catch (ex) {  // error if collection changed since last call: start over again
        this._scrollToPartIterator = null;
      }
      if (part === null) {  // either because no more in collection or because of error
        // start over again to cycle back through
        if (diagram.highlighteds.count > 0) {
          this._scrollToPartIterator = diagram.highlighteds.iterator;
        } else if (diagram.selection.count > 0) {
          this._scrollToPartIterator = diagram.selection.iterator;
        }
        if (this._scrollToPartIterator !== null && this._scrollToPartIterator.next()) {
          part = this._scrollToPartIterator.value;
        }
      }
    }
    if (part !== null) {
      const AM = diagram.animationManager;
      AM.prepareAutomaticAnimation('Scroll To Part'); // Doesn't need to call beginAnimation because there's a transaction
      const pause = this.scrollToPartPause;
      if (pause > 0) {
        const show: Array<Part> = this.followCollapsedPart(part, [part]);

        if (show.length === 1) {
          // simple case:
          diagram.startTransaction();
          diagram.centerRect(part.actualBounds);
          diagram.commitTransaction('Scroll To Part');
        } else {
          const expander = function () {
            diagram.startTransaction();
            let p = show.pop();
            while (show.length > 0 && (p instanceof Node && p.isTreeExpanded &&
                                       (!(p instanceof Group) || p.isSubGraphExpanded))) {
              p = show.pop();
            }
            if (show.length > 0) {
              if (p instanceof Part) diagram.scrollToRect(p.actualBounds);
              if (p instanceof Node && !p.isTreeExpanded) p.isTreeExpanded = true;
              if (p instanceof Group && !p.isSubGraphExpanded) p.isSubGraphExpanded = true;
            } else {
              if (p instanceof Part) diagram.centerRect(p.actualBounds);
              diagram.removeDiagramListener('LayoutCompleted', handler);
            }
            diagram.commitTransaction('Scroll To Part');
          };

          const handler = function () {
            Util.setTimeout(expander, (AM.isEnabled ? AM.duration : 0) + pause);
          };

          diagram.addDiagramListener('LayoutCompleted', handler);
          expander();
        }

      } else {
        const oldPos = diagram.position.copy();
        diagram.centerRect(part.actualBounds);
        if (oldPos.equalsApprox(diagram.position)) AM.stopAnimation();
      }
    }
  }

  /**
   * @hidden @internal
   * Return an Array of Parts that need to be expanded in order for the given PART to become visible.
   * @this {CommandHandler}
   * @param {Part} part
   * @param {Array.<Part>} found
   * @return {Array.<Part>}
   */
  private followCollapsedPart(part: Part, found: Array<Part>): Array<Part> {
    // stop recursion when Part can already be seen
    if (part.isVisible()) return found;
    // can't handle Parts whose .visible is set to false
    // ??? if (!part.visible || part.layer === null || !part.layer.visible) return [part];
    if (part instanceof Adornment) {
      this.followCollapsedPart(part.adornedPart, found);
    } else if (part instanceof Link) {
      const from = part.fromNode;
      if (from !== null) this.followCollapsedPart(from, found);
      const to = part.toNode;
      if (to !== null) this.followCollapsedPart(to, found);
    } else {  // Group, Node or plain Part
      if (part instanceof Node) {
        const link = part.labeledLink;
        if (link !== null) {
          this.followCollapsedPart(link, found);
        }
        const parent = part.findTreeParentNode();
        if (parent !== null) {
          if (!parent.isTreeExpanded && !parent.wasTreeExpanded) {
            found.push(parent);
          }
          this.followCollapsedPart(parent, found);
        }
      }
      const container = part.containingGroup;
      if (container !== null) {
        if (!container.isSubGraphExpanded && !container.wasSubGraphExpanded) {
          found.push(container);
        }
        this.followCollapsedPart(container, found);
      }
    }
    return found;
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #scrollToPart} command.
   * This returns false if there is no argument Part and there are no selected Parts.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Part=} part This defaults to the first selected Part of {@link Diagram#selection}
   * @return {boolean}
   * This returns true if {@link Diagram#allowHorizontalScroll} and {@link Diagram#allowVerticalScroll} are true.
   * @see {@link #scrollToPart}
   * @since 1.6
   */
  public canScrollToPart(part?: Part): boolean {
    if (part === undefined) part = null;
    if (part !== null && !(part instanceof Part)) return false;
    const diagram = this.diagram;
    if (diagram.selection.count === 0 && diagram.highlighteds.count === 0) return false;
    return diagram.allowHorizontalScroll && diagram.allowVerticalScroll;
  }

  /**
   * This command collapses all expanded selected {@link Node}s.
   * This operation is performed within a "Collapse Tree" transaction.
   * Just before the end of the transaction this raises the "TreeCollapsed" {@link DiagramEvent},
   * with a collection of collapsed {@link Node}s as the subject.
   * This currently has no default keyboard shortcut.
   *
   * This calls {@link Node#collapseTree} to perform the collapse, which will set {@link Node#isTreeExpanded} to false.
   * You may want to save the collapsed/expanded state to the model by using a TwoWay {@link Binding} on the "isTreeExpanded" property of your Nodes,
   * and perhaps also on the {@link Node#wasTreeExpanded} property.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Node=} node if supplied, ignore the selection and collapse this particular Node subtree.
   * @see {@link #canCollapseTree}
   */
  public collapseTree(node?: Node): void {
    if (node === undefined) node = null;
    const diagram = this.diagram;
    const reason = 'Collapse Tree';
    try {
      diagram.startTransaction(reason);
      const AM = diagram.animationManager;
      AM.prepareAutomaticAnimation(reason);
      const nodes = new List<Node>();
      if (node !== null && node.isTreeExpanded) {
        node.collapseTree();
        nodes.add(node);
      } else if (node === null) {
        const it = diagram.selection.iterator;
        while (it.next()) {
          const part = it.value;
          if (part instanceof Node) {
            if (part.isTreeExpanded) {
              part.collapseTree();
              nodes.add(part);
            }
          }
        }
        // ?? also change selection
      }
      diagram.raiseDiagramEvent('TreeCollapsed', nodes);
    } finally {
      diagram.commitTransaction(reason);
    }
  }

  /**
   * This predicate controls whether the user can collapse expanded subtrees of {@link Node}s.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Node=} node if supplied, ignore the selection and consider collapsing this particular Node.
   * @return {boolean}
   * @see {@link #collapseTree}
   */
  public canCollapseTree(node?: Node): boolean {
    if (node === undefined) node = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (node !== null) {
      if (!(node instanceof Node)) return false;
      if (!node.isTreeExpanded) return false;  // already collapsed
      if (node.findTreeChildrenLinks().count > 0) return true;
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part instanceof Node) {
          if (!part.isTreeExpanded) continue;  // not this Node -- it's already collapsed
          if (part.findTreeChildrenLinks().count > 0) return true;
        }
      }
    }
    return false;
  }

  /**
   * This command expands all collapsed selected {@link Node}s.
   * This operation is performed within an "Expand Tree" transaction.
   * Just before the end of the transaction this raises the "TreeExpanded" {@link DiagramEvent},
   * with a collection of expanded {@link Node}s as the subject.
   * This currently has no default keyboard shortcut.
   *
   * This calls {@link Node#expandTree} to perform the expand, which will set {@link Node#isTreeExpanded} to true.
   * You may want to save the collapsed/expanded state to the model by using a TwoWay {@link Binding} on the "isTreeExpanded" property of your Nodes,
   * and perhaps also on the {@link Node#wasTreeExpanded} property.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Node=} node if supplied, ignore the selection and collapse this particular Node subtree.
   * @see {@link #canExpandTree}
   */
  public expandTree(node?: Node): void {
    if (node === undefined) node = null;
    const diagram = this.diagram;
    const reason = 'Expand Tree';
    try {
      diagram.startTransaction(reason);
      const AM = diagram.animationManager;
      AM.prepareAutomaticAnimation(reason);
      const nodes = new List<Node>();
      if (node !== null && !node.isTreeExpanded) {
        node.expandTree();
        nodes.add(node);
      } else if (node === null) {
        const it = diagram.selection.iterator;
        while (it.next()) {
          const part = it.value;
          if (part instanceof Node) {
            if (!part.isTreeExpanded) {
              part.expandTree();
              nodes.add(part);
            }
          }
        }
        // ?? also change selection
      }
      diagram.raiseDiagramEvent('TreeExpanded', nodes);
    } finally {
      diagram.commitTransaction(reason);
    }
  }

  /**
   * This predicate controls whether the user can expand collapsed subtrees of {@link Node}s.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Node=} node if supplied, ignore the selection and consider expanding this particular Node.
   * @return {boolean}
   * @see {@link #expandTree}
   */
  public canExpandTree(node?: Node): boolean {
    if (node === undefined) node = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (node !== null) {
      if (!(node instanceof Node)) return false;
      if (node.isTreeExpanded) return false;  // already expanded
      if (node.findTreeChildrenLinks().count > 0) return true;
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part instanceof Node) {
          if (part.isTreeExpanded) continue;  // not this Node -- it's already expanded
          if (part.findTreeChildrenLinks().count > 0) return true;
        }
      }
    }
    return false;
  }

  // grouping commands

  /**
   * This command adds a copy of {@link #archetypeGroupData} to the diagram's model
   * to create a new {@link Group} and then adds the selected {@link Part}s to that new group.
   * This is normally invoked by the `Ctrl-G` keyboard shortcut.
   *
   * This creates a new {@link Group} by adding a copy of the {@link #archetypeGroupData} to the model.
   * Each of the selected parts for which {@link Part#canGroup} is true
   * and for which {@link #isValidMember} is true
   * is made a member of that new group.
   * If all of the selected groupable parts were members of a pre-existing group,
   * the new group also becomes a member of that pre-existing group,
   * if {@link #isValidMember} is true for that existing group with the new group.
   * The new group becomes the only selected part.
   * This raises the "SelectionGrouped" diagram event.
   * This method also raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed in a "Group" transaction, but the selection events are raised outside the transaction.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @see {@link #canGroupSelection}
   */
  public groupSelection(): void {
    const diagram = this.diagram;
    const model = diagram.model;
    if (!model.isGroupModel()) return;
    const arch = this.archetypeGroupData;
    if (arch === null) return;
    let newgroup: Group = null;
    try {
      diagram.currentCursor = 'wait';
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      diagram.startTransaction('Group');

      // get all of the selected parts that can be grouped
      const selparts = new List<Part>();
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part.isNodeLike() && part.canGroup()) selparts.add(part);
      }
      // keep selected parts that are not part of any other selected parts
      const parts = new List<Part>();
      const sit = selparts.iterator;
      while (sit.next()) {
        const n = sit.value;
        let mem = false;
        const sit2 = selparts.iterator;
        while (sit2.next()) {
          const m = sit2.value;
          if (n.isMemberOf(m)) {
            mem = true;
            break;
          }
        }
        if (!mem) {
          parts.add(n);
        }
      }
      // find lowest common containing group for selected parts
      if (parts.count > 0) {
        let firstsg = parts.first().containingGroup;
        if (firstsg !== null) {
          while (firstsg !== null) {  // stop going up containingGroup chain when at top-level
            // are any NODES not a member of the group FIRSTSG?
            let notmem = false;
            const pit = parts.iterator;
            while (pit.next()) {
              const p = pit.value;
              if (!p.isMemberOf(firstsg)) {
                notmem = true;
                break;
              }
            }
            if (notmem) {
              firstsg = firstsg.containingGroup;  // try next Group up the chain
            } else {
              break;  // done!
            }
          }
        }

        // create new group
        if (arch instanceof Group) {
          arch.freezeBindings();
          newgroup = arch.copy();
          if (newgroup !== null) {
            diagram.add(newgroup);
          }
        } else if (model.isGroupForData(arch)) {
          const newgroupdata = model.copyNodeData(arch);
          if (Util.isObject(newgroupdata)) {
            // add it to the model and diagram
            model.addNodeData(newgroupdata);
            // find the corresponding Group node
            newgroup = diagram.findNodeForData(newgroupdata) as Group;
          }
        }

        if (newgroup !== null) {
          if (firstsg !== null && this.isValidMember(firstsg, newgroup)) {
            // make it a member of an existing group
            newgroup.containingGroup = firstsg;
          }
          // update all of the NODES to be members of the new group
          const nit = parts.iterator;
          while (nit.next()) {
            const n1 = nit.value;
            if (this.isValidMember(newgroup, n1)) {
              n1.containingGroup = newgroup;
            }
          }
          // make the new group the only selected part without raising events
          diagram.internalClearSelection();
          newgroup.isSelected = true;
        }
      }

      diagram.raiseDiagramEvent('SelectionGrouped', newgroup);
    } finally {
      diagram.commitTransaction('Group');
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      diagram.currentCursor = '';
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #groupSelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @return {boolean}
   * This returns true:
   * if the {@link #archetypeGroupData} is not null,
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowInsert} is true,
   * if {@link Diagram#allowGroup} is true,
   * if {@link Diagram#model} is a {@link GraphLinksModel}, and
   * if there is any selected {@link Part} that can be {@link Part#canGroup}ed.
   * @see {@link #groupSelection}
   */
  public canGroupSelection(): boolean {
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowInsert || !diagram.allowGroup) return false;
    const model = diagram.model;
    if (!model.isGroupModel()) return false;
    if (this.archetypeGroupData === null) return false;
    const it = diagram.selection.iterator;
    while (it.next()) {
      const part = it.value;
      if (part.isNodeLike() && part.canGroup()) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {CommandHandler}
   * @param {Iterable.<Part>} coll a collection of Parts
   * @return {Set.<Part>} of non-Links.
   */
  public findUnnestedNodes(coll: Iterable<Part>): Set<Part> {
    // find nodes
    const selnodes = Util.tempArray<Part>();
    const itr = coll.iterator;
    while (itr.next()) {
      const n = itr.value;
      if (n instanceof Link) continue;
      selnodes.push(n);
    }

    // filter out nodes that are members (perhaps indirectly) of other selected nodes
    // return new HashSet<Node>(selnodes.Where(n => !selnodes.Any(m => n.IsContainedBy(m))));

    const set = new Set<Part>();
    const l = selnodes.length;
    for (let i = 0; i < l; i++) {
      const cur = selnodes[i];
      // filter out the "bad" nodes that are members (perhaps indirectly) of other selected nodes
      let good = true;
      for (let j = 0; j < l; j++) {
        if (cur.isMemberOf(selnodes[j])) {
          good = false;
          break;
        }
      }
      if (good) {
        set.add(cur);
      }
    }
    Util.freeArray(selnodes);
    return set;
  }

  /**
   * This predicate is called to determine whether a Node may be added as a member of a Group.
   * This always checks to make sure no group might become a member of itself, either directly or indirectly.
   * If the Group has a {@link Group#memberValidation} predicate and if it returns false, this method returns false.
   * If this CommandHandler has a {@link #memberValidation} predicate and if it returns false, this method returns false.
   * Otherwise this will return true.
   * <p class="box">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group} group this may be null if the node is being added as a top-level node.
   * @param {Part} part a Part, usually a Node, possibly another Group, but not a Link or an Adornment.
   * @return {boolean} true if OK to add the node to the group.
   */
  public isValidMember(group: Group, part: Part): boolean {
    if (part === null) return false;
    if (group === part) return false;  // can't be a member of itself!
    if (part instanceof Link) return false;
    let pred;
    if (group !== null) {
      if (group === part || group.isMemberOf(part)) return false;  // do not allow cycles!
      pred = group.memberValidation;
      if (pred !== null && !pred(group, part)) return false;
      if (group.data === null && part.data !== null) return false;  // either both must be in model,
      if (group.data !== null && part.data === null) return false;  // or neither are in model
    }
    pred = this.memberValidation;
    if (pred !== null) {
      return pred(group, part);
    } else {
      return true;
    }
  }

  /**
   * This command removes selected groups from the diagram without removing their members from the diagram.
   * This is normally invoked by the `Ctrl-Shift-G` keyboard shortcut.
   *
   * For the given group, or if not supplied, each selected {@link Group} that is {@link Group#ungroupable},
   * expand the subgraph and change all of its member parts to be members of the group that the selected group node is in.
   * (If the selected group is a top-level node, i.e. not a member of any group node, its members become top-level parts too.)
   * All of those selected groups are deleted.
   * All of the reparented member parts are selected.
   *
   * This raises the "SelectionUngrouped" diagram event.
   * This method also raises the "ChangingSelection" and "ChangedSelection" diagram events.
   * Changes are performed in an "Ungroup" transaction, but the selection events are raised outside the transaction.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and consider ungrouping this particular Group.
   * @see {@link #canUngroupSelection}
   */
  public ungroupSelection(group?: Group): void {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    const model = diagram.model;
    if (!model.isGroupModel()) return;
    try {
      diagram.currentCursor = 'wait';
      diagram.raiseDiagramEvent('ChangingSelection', diagram.selection);
      diagram.startTransaction('Ungroup');

      // find all selected groups that are ungroupable
      const groups = new List<Group>();
      if (group !== null) {
        groups.add(group);
      } else {
        const it = diagram.selection.iterator;
        while (it.next()) {
          const part = it.value;
          if (part instanceof Group) {
            if (part.canUngroup()) groups.add(part);
          }
        }
      }
      const members = new List<Part>();
      if (groups.count > 0) {
        // forget selection without raising events, so we can select members as they are promoted from out of each Group
        diagram.internalClearSelection();
        const git = groups.iterator;
        while (git.next()) {
          const groupnode = git.value;
          groupnode.expandSubGraph();
          // iterate over the member parts and reassign their group key to be SG's original
          // containing group key, if any
          const parent = groupnode.containingGroup;
          const pkey = (parent !== null && parent.data !== null) ? model.getKeyForNodeData(parent.data) : undefined;
          members.addAll(groupnode.memberParts);
          const mit = members.iterator;
          while (mit.next()) {
            const member = mit.value;
            member.isSelected = true;
            if (member instanceof Link) continue;
            const memdata = member.data;
            if (memdata !== null) {
              model.setGroupForData(memdata, pkey);
            } else {
              member.containingGroup = parent;
            }
          }
          // then delete the empty group -- include deleting any connected links
          diagram.remove(groupnode);
        }
      }

      diagram.raiseDiagramEvent('SelectionUngrouped', groups, members);
    } finally {
      diagram.commitTransaction('Ungroup');
      diagram.raiseDiagramEvent('ChangedSelection', diagram.selection);
      diagram.currentCursor = '';
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #ungroupSelection} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and consider ungrouping this particular Group.
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowDelete} is true,
   * if {@link Diagram#allowUngroup} is true,
   * if {@link Diagram#model} is a {@link GraphLinksModel}, and
   * if there are any selected {@link Group}s that are {@link Group#ungroupable}.
   * @see {@link #ungroupSelection}
   */
  public canUngroupSelection(group?: Group): boolean {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowDelete || !diagram.allowUngroup) return false;
    const model = diagram.model;
    if (!model.isGroupModel()) return false;
    if (group !== null) {
      if (!(group instanceof Group)) return false;
      if (group.canUngroup()) return true;
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part instanceof Group) {
          if (part.canUngroup()) return true;
        }
      }
    }
    return false;
  }

  // ?? these non-command methods really ought to be in a PartManager-like class,
  // to avoid confusion with commands implemented by methods

  /**
   * Make sure all of the unnested {@link Part}s in the given collection
   * are removed from any containing {@link Group}s.
   *
   * This sets {@link Part#containingGroup} to null on each Part that is
   * not a member of another Part in the argument collection.
   * If the check argument to this method is supplied and true,
   * this will call {@link #isValidMember} on each part, passing null as the first argument.
   *
   * This functionality is similar to {@link #ungroupSelection},
   * except that this is not a command (there is no transaction and this does not raise a {@link DiagramEvent})
   * and the parts are necessarily becoming top-level parts
   * (whereas ungrouping would add them to the Group containing the Group being ungrouped).
   *
   * This function is typically called in a {@link Diagram#mouseDrop} event handler
   * in order to remove the selected Parts from whatever {@link Group} they had been in.
   *
   * If you want to add Parts to be members of a Group, call {@link Group#addMembers}.
   * If you want to remove Parts completely from a Diagram, call {@link Diagram#removeParts}.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Iterable.<Part>} coll a collection of Parts.
   * @param {boolean=} check whether to call {@link #isValidMember} to confirm that changing the Part to be a top-level Part is valid.
   * @return {boolean} true if all non-Links were changed to be top-level Parts in this Diagram; false if some Parts or Nodes were not able to be added.
   */
  public addTopLevelParts(coll: Iterable<Part>, check?: boolean): boolean {
    // get all of the simple Parts and Nodes or Groups, but not Links,
    // that are not members of other parts in the given collection
    const unnested = this.findUnnestedNodes(coll);
    let ok = true;
    const it = unnested.iterator;
    while (it.next()) {
      const node = it.value;
      if (node.containingGroup !== null) {
        if (!check || this.isValidMember(null, node)) {
          node.containingGroup = null;
        } else {
          ok = false;
        }
      }
      // do NOT call Diagram.add, to allow this to work with Parts that are not currently in a Diagram
    }
    return ok;
  }

  /**
   * This command collapses all expanded selected {@link Group}s.
   * This operation is performed within a "Collapse SubGraph" transaction.
   * Just before the end of the transaction this raises the "SubGraphCollapsed" {@link DiagramEvent},
   * with a collection of collapsed {@link Group}s as the subject.
   * This currently has no default keyboard shortcut.
   *
   * This calls {@link Group#collapseSubGraph} to perform the collapse, which will set {@link Group#isSubGraphExpanded} to false.
   * You may want to save the collapsed/expanded state to the model by using a TwoWay {@link Binding} on the "isSubGraphExpanded" property of your Groups,
   * and perhaps also on the {@link Group#wasSubGraphExpanded} property.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and collapse this particular Group.
   * @see {@link #canCollapseSubGraph}
   */
  public collapseSubGraph(group?: Group): void {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    const reason = 'Collapse SubGraph';
    try {
      diagram.startTransaction(reason);
      const AM = diagram.animationManager;
      AM.prepareAutomaticAnimation(reason);
      const groups = new List<Group>();
      if (group !== null && group.isSubGraphExpanded) {
        group.collapseSubGraph();
        groups.add(group);
      } else if (group === null) {
        const it = diagram.selection.iterator;
        while (it.next()) {
          const part = it.value;
          if (part instanceof Group) {
            if (part.isSubGraphExpanded) {
              part.collapseSubGraph();
              groups.add(part);
            }
          }
        }
        // ?? also change selection
      }
      diagram.raiseDiagramEvent('SubGraphCollapsed', groups);
    } finally {
      diagram.commitTransaction(reason);
    }
  }

  /**
   * This predicate controls whether the user can collapse expanded {@link Group}s.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and consider collapsing this particular Group.
   * @return {boolean}
   * @see {@link #collapseSubGraph}
   */
  public canCollapseSubGraph(group?: Group): boolean {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (group !== null) {
      if (!(group instanceof Group)) return false;
      if (!group.isSubGraphExpanded) return false;  // already collapsed
      return true;  // even if there are zero members
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part instanceof Group) {
          if (part.isSubGraphExpanded) return true;  // even if there are zero members
          // not this Group -- it's already collapsed
        }
      }
    }
    return false;
  }

  /**
   * This command expands all collapsed selected {@link Group}s.
   * This operation is performed within an "Expand SubGraph" transaction.
   * Just before the end of the transaction this raises the "SubGraphExpanded" {@link DiagramEvent},
   * with a collection of expanded {@link Group}s as the subject.
   * This currently has no default keyboard shortcut.
   *
   * This calls {@link Group#expandSubGraph} to perform the collapse, which will set {@link Group#isSubGraphExpanded} to true.
   * You may want to save the collapsed/expanded state to the model by using a TwoWay {@link Binding} on the "isSubGraphExpanded" property of your Groups,
   * and perhaps also on the {@link Group#wasSubGraphExpanded} property.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and expand this particular Group.
   * @see {@link #canExpandSubGraph}
   */
  public expandSubGraph(group?: Group): void {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    const reason = 'Expand SubGraph';
    try {
      diagram.startTransaction(reason);
      const AM = diagram.animationManager;
      AM.prepareAutomaticAnimation(reason);
      const groups = new List<Group>();
      if (group !== null && !group.isSubGraphExpanded) {
        group.expandSubGraph();
        groups.add(group);
      } else if (group === null) {
        const it = diagram.selection.iterator;
        while (it.next()) {
          const part = it.value;
          if (part instanceof Group) {
            if (!part.isSubGraphExpanded) {
              part.expandSubGraph();
              groups.add(part);
            }
          }
        }
        // ?? also change selection
      }
      diagram.raiseDiagramEvent('SubGraphExpanded', groups);
    } finally {
      diagram.commitTransaction(reason);
    }
  }

  /**
   * This predicate controls whether the user can expand collapsed {@link Group}s.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {Group=} group if supplied, ignore the selection and consider expanding this particular Group.
   * @return {boolean}
   * @see {@link #expandSubGraph}
   */
  public canExpandSubGraph(group?: Group): boolean {
    if (group === undefined) group = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly) return false;
    if (group !== null) {
      if (!(group instanceof Group)) return false;
      if (group.isSubGraphExpanded) return false;  // already expanded
      return true;  // even if there are zero members
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part instanceof Group) {
          if (!part.isSubGraphExpanded) return true;  // even if there are zero members
          // not this Group -- it's already expanded
        }
      }
    }
    return false;
  }

  // editing commands

  /**
   * This command starts in-place editing of a {@link TextBlock} in the selected {@link Part}.
   * This is normally invoked by the `F2` keyboard shortcut.
   *
   * This starts the {@link TextEditingTool} to have the user enter or modify the text string and finally set the {@link TextBlock#text}.
   * You may want to save the new string to the model by using a TwoWay {@link Binding} on the "text" property of your TextBlock.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {TextBlock=} textblock the {@link TextBlock} to start editing.
   * @see {@link #canEditTextBlock}
   */
  public editTextBlock(textblock?: TextBlock): void {
    if (textblock === undefined) textblock = null;
    if (textblock !== null) Util.checkClass(textblock, TextBlock, CommandHandler, 'editTextBlock');
    const diagram = this.diagram;
    const tool = diagram.toolManager.findTool('TextEditing') as TextEditingTool;
    if (tool === null) return;

    if (textblock === null) {  // need to find a TextBlock to edit
      // find the first eligible node
      let firstpart: Part = null;
      const itr = diagram.selection.iterator;
      while (itr.next()) {
        const part = itr.value;
        if (part.canEdit()) {
          firstpart = part;
          break;
        }
      }
      if (firstpart === null) return;
      textblock = firstpart.findInVisualTree(function (x) { return (x instanceof TextBlock && x.editable); }) as TextBlock;
    }
    if (textblock !== null) {
      diagram.currentTool = null;
      tool.textBlock = textblock;
      diagram.currentTool = tool;
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #editTextBlock} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {TextBlock=} textblock the {@link TextBlock} to consider editing.
   * @return {boolean}
   * This returns true:
   * if the diagram is not {@link Diagram#isReadOnly},
   * if {@link Diagram#allowTextEdit} is true,
   * if there is a {@link ToolManager#textEditingTool}, and
   * if there is any selected {@link Part} for which {@link Part#canEdit} is true.
   * @see {@link #editTextBlock}
   */
  public canEditTextBlock(textblock?: TextBlock): boolean {
    if (textblock === undefined) textblock = null;
    const diagram = this.diagram;
    if (diagram.isReadOnly || diagram.isModelReadOnly) return false;
    if (!diagram.allowTextEdit) return false;
    if (diagram.toolManager.findTool('TextEditing') === null) return false;
    if (textblock !== null) {
      if (!(textblock instanceof TextBlock)) return false;
      const part = textblock.part;
      if (part !== null && part.canEdit()) return true;
    } else {
      const it = diagram.selection.iterator;
      while (it.next()) {
        const part = it.value;
        if (part.canEdit()) {
          textblock = part.findInVisualTree(function (x) { return (x instanceof TextBlock && x.editable); }) as TextBlock;
          if (textblock !== null) return true;
        }
      }
    }
    return false;
  }

  /**
   * This command opens the context menu for a selected Part or given GraphObject, or else for the whole Diagram.
   * This is normally invoked by the `Menu` keyboard shortcut.
   *
   * The given GraphObject must have a {@link GraphObject#contextMenu}
   * defined in order to show anything.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {GraphObject|Diagram=} obj a GraphObject or Diagram with a contextMenu defined.
   * If none is given, this method will use the first selected object, or else the Diagram.
   * The method will simulate a right-button click at the middle of the GraphObject or,
   * if a Diagram, at the current mouse position if it is in the viewport.
   * @see {@link #canShowContextMenu}
   */
  public showContextMenu(obj?: GraphObject | Diagram): void {
    const diagram = this.diagram;
    const tool = diagram.toolManager.findTool('ContextMenu') as ContextMenuTool;
    if (tool === null) return;

    if (obj === undefined) {
      if (diagram.selection.count > 0) {
        obj = diagram.selection.first();
      } else {
        obj = diagram;
      }
    }

    const ctxObj = tool.findObjectWithContextMenu(obj);
    if (ctxObj !== null) {
      const input = diagram.lastInput;

      let pt: Point = null;
      if (ctxObj instanceof GraphObject) {
        pt = ctxObj.getDocumentPoint(Spot.Center);
      } else { // Diagram
        if (!diagram.viewportBounds.containsPoint(input.documentPoint)) {
          const vpb = diagram.viewportBounds;
          pt = new Point(vpb.x + (vpb.width / 2), vpb.y + (vpb.height / 2));
        } else {
          // mouse is still over diagram, leave lastInput as-is and show context menu there
        }
      }

      if (pt !== null) {
        input.viewPoint = diagram.transformDocToView(pt);
        input.documentPoint = pt;
        input.left = false;  // because InputEvent.left is true by default
        input.right = true;
        input.up = true;
      }
      diagram.currentTool = tool;
      tool.openMenu(false, ctxObj);
    }
  }

  /**
   * This predicate controls whether or not the user can invoke the {@link #showContextMenu} command.
   *
   * This method may be overridden, but you should consider calling this base method in order to get all of its functionality.
   * This method must not have any side-effects.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CommandHandler}
   * @param {GraphObject|Diagram=} obj a GraphObject or Diagram with a contextMenu defined.
   * If none is given, this method will use the first selected object, or else the Diagram.
   * @return {boolean}
   * @see {@link #showContextMenu}
   */
  public canShowContextMenu(obj?: GraphObject | Diagram): boolean {
    const diagram = this.diagram;
    const tool = diagram.toolManager.findTool('ContextMenu') as ContextMenuTool;
    if (tool === null) return false;

    if (obj === undefined) {
      if (diagram.selection.count > 0) {
        obj = diagram.selection.first();
      } else {
        obj = diagram;
      }
    }

    const contextobj = tool.findObjectWithContextMenu(obj);
    if (contextobj === null) return false;
    return true;
  }

  // state

  /**
   * @hidden @internal
   * Gets or sets whether {@link #copyToClipboard} should also copy any bound data objects.
   * The default value is true.
   * Setting this property does not raise any events.
   */
  /** @export */ get copiesClipboardData(): boolean { return this._copiesClipboardData; }
  /** @export */ set copiesClipboardData(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'copiesClipboardData');
    this._copiesClipboardData = value;
  }

  /**
   * Gets or sets whether {@link #copySelection} should also copy Links that connect with selected Nodes.
   * The default value is true.
   * Setting this property does not raise any events.
   *
   * The {@link DraggingTool#copiesEffectiveCollection} property serves a similar role for the {@link DraggingTool}
   * when the user holds down the control key to modify the drag into a copy operation.
   * @since 1.5
   */
  /** @export */ get copiesConnectedLinks(): boolean { return this._copiesConnectedLinks; }
  /** @export */ set copiesConnectedLinks(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'copiesConnectedLinks');
    this._copiesConnectedLinks = value;
  }

  /**
   * Gets or sets whether {@link #deleteSelection} should also delete links that are connected to nodes that are deleted.
   * The default value is true.
   * Setting this property does not raise any events.
   * @since 1.6
   */
  /** @export */ get deletesConnectedLinks(): boolean { return this._deletesConnectedLinks; }
  /** @export */ set deletesConnectedLinks(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'deletesConnectedLinks');
    this._deletesConnectedLinks = value;
  }

  /**
   * Gets or sets whether {@link #copySelection} should also copy subtrees.
   * The default value is false.
   * Setting this property does not raise any events.
   *
   * The {@link DraggingTool#dragsTree} property serves a similar role for the {@link DraggingTool}
   * for both moving and copying operations.
   */
  /** @export */ get copiesTree(): boolean { return this._copiesTree; }
  /** @export */ set copiesTree(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'copiesTree');
    this._copiesTree = value;
  }

  /**
   * Gets or sets whether {@link #deleteSelection} should also delete subtrees.
   * The default value is false.
   * Setting this property does not raise any events.
   */
  /** @export */ get deletesTree(): boolean { return this._deletesTree; }
  /** @export */ set deletesTree(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'deletesTree');
    this._deletesTree = value;
  }

  /**
   * Gets or sets whether {@link #copySelection} and {@link #copyToClipboard} copy the node data property
   * whose value is the tree-parent node data's key.
   *
   * Set this property to true if you want a copy/paste of a node to automatically have the new node be a tree-child of the original tree-parent node.
   * Caution: this only has an effect if the Diagram's {@link Diagram#model} is a {@link TreeModel}.
   *
   * The default value is false.
   * @since 1.3
   */
  /** @export */ get copiesParentKey(): boolean { return this._copiesParentKey; }
  /** @export */ set copiesParentKey(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'copiesParentKey');
    this._copiesParentKey = value;
  }

  /**
   * Gets or sets whether {@link #copySelection} and {@link #copyToClipboard} copy the node data property
   * whose value is the containing group data's key.
   *
   * Set this property to true if you want a copy/paste of a node to automatically have the new node be a member of the original group.
   * Caution: this only has an effect if the Diagram's {@link Diagram#model} is a {@link GraphLinksModel}.
   *
   * The default value is false.
   * @since 1.3
   */
  /** @export */ get copiesGroupKey(): boolean { return this._copiesGroupKey; }
  /** @export */ set copiesGroupKey(value: boolean) {
    Util.checkPrimitive(value, 'boolean', CommandHandler, 'copiesGroupKey');
    this._copiesGroupKey = value;
  }

  /**
   * Gets or sets a data object that is copied by {@link #groupSelection}
   * when creating a new Group.
   *
   * The default value is null.
   * If you set this to an Object, be sure that {@link GraphLinksModel#isGroupForNodeData} is true for that object.
   * Setting this property does not raise any events.
   */
  /** @export */ get archetypeGroupData(): ObjectData { return this._archetypeGroupData; }
  /** @export */ set archetypeGroupData(value: ObjectData) {
    if (value !== null) Util.checkPrimitive(value, 'object', CommandHandler, 'archetypeGroupData');
    const diagram = this.diagram;
    if (Debug) {
      const model = diagram.model;
      if (model.isGroupModel() && !(value instanceof Group) && !model.isGroupForData(value)) {
        Util.throwError('CommandHandler.archetypeGroupData must be either a Group or a data object for which GraphLinksModel.isGroupForNodeData is true: ' + value);
      }
    }
    this._archetypeGroupData = value;
  }

  /**
   * Gets or sets the predicate that determines whether or not a node may become a member of a group.
   * This predicate is called in addition to any existing group's {@link Group#memberValidation} predicate.
   * The default predicate is null, which is equivalent to simply returning true.
   * The predicate may be called passing null as the first argument (the Group) --
   * this asks whether it is OK to make the second argument (the Part, but not a Link) a top-level Part of the diagram.
   * <p class="box">
   * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
   *
   * The function, if supplied, must not have any side-effects.
   */
  /** @export */ get memberValidation(): ((aGroup: Group, somePart: Part) => boolean) | null { return this._memberValidation; }
  /** @export */ set memberValidation(value: ((aGroup: Group, somePart: Part) => boolean) | null) {
    if (value !== null) Util.checkPrimitive(value, 'function', CommandHandler, 'memberValidation');
    this._memberValidation = value;
  }

  /**
   * Deprecated in favor of {@link Diagram#defaultScale}.
   *
   * Gets or sets the {@link Diagram#scale} set by {@link #resetZoom}.
   *
   * The default value is 1.0.
   * The value must be a number larger than 0.
   * Setting this property does not raise any events.
   * @since 1.3
   */
  /** @export */ get defaultScale(): number { return this.diagram.defaultScale; }
  /** @export */ set defaultScale(value: number) {
    this.diagram.defaultScale = value;
  }

  /**
   * Gets or sets the amount by which {@link #decreaseZoom} and {@link #increaseZoom} change
   * the {@link Diagram#scale}.
   *
   * The default value is 1.05 (5%).
   * The value must be a number larger than 1.0.
   * Setting this property does not raise any events.
   * @since 1.1
   */
  /** @export */ get zoomFactor(): number { return this._zoomFactor; }
  /** @export */ set zoomFactor(value: number) {
    Util.checkRealNumber(value, CommandHandler, 'zoomFactor');
    if (!(value > 1.0)) Util.throwError('zoomFactor must be larger than 1.0, not: ' + value);
    this._zoomFactor = value;
  }

  /**
   * @hidden @internal
   * Gets or sets the delay used between each expansion step when {@link #scrollToPart}
   * needs to expand subtrees or subgraphs in order to make the given Part visible.
   *
   * The default value is 200 -- there is a pause after each tree or subgraph expansion
   * when trying to scroll to a hidden Part.
   * Setting this property does not raise any events.
   */
  /** @export */ get scrollToPartPause(): number { return this._scrollToPartPause; }
  /** @export */ set scrollToPartPause(value: number) {
    Util.checkRealNumber(value, CommandHandler, 'scrollToPartPause');
    this._scrollToPartPause = value;
  }

  /**
   * Find the actual collection of nodes and links to be moved or copied,
   * given an initial collection.
   * This includes links that connected at both ends to nodes being moved or copied,
   * members of {@link Group}s, and if {@link DraggingTool#dragsTree} is true,
   * this includes nodes and links that are "tree" descendants from selected nodes.
   *
   * Note that this does not return a simple collection of {@link Part}s,
   * but a {@link Map} associating a chosen {@link Part}
   * with an Object holding its original location Points as the value of the "point" property.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * This was moved to CommandHandler from {@link DraggingTool} in 2.0.
   * @expose
   * @this {CommandHandler}
   * @param {Iterable.<Part>} parts A {@link Set} or {@link List} of {@link Part}s.
   * @param {DraggingOptions=} options Potential options for the collection computation. If not specified, this uses the DraggingTool's drag options.
   * @return {Map.<Part,DraggingInfo>}  a Map mapping {@link Part}s to DraggingInfo Objects
   *         that have a "point" property remembering the original location of that Part.
   * @since 2.0
   */
  public computeEffectiveCollection(parts: Iterable<Part>, options?: DraggingOptions): Map<Part, DraggingInfo> {
    const diagram = this.diagram;
    const dragtool = diagram.toolManager.findTool('Dragging') as any;
    const dragging = (diagram.currentTool === dragtool);
    if (options === undefined) options = dragging ? dragtool.dragOptions : diagram._defaultDragOptions;
    const map = new Map<Part, DraggingInfo>();
    if (parts === null) return map;
    let it = parts.iterator;
    while (it.next()) {
      const p = it.value;
      diagram.gatherMembers(map, p, dragging, options);
    }
    if (diagram.draggedLink !== null && options.dragsLink) return map; // to avoid DraggingTool dependency. Unwise?
    // ignore any links that are connected to nodes not in the effective collection
    it = parts.iterator;
    while (it.next()) {
      const part = it.value;
      if (part instanceof Link) {
        const from = part.fromNode;
        if (from !== null && !map.contains(from)) {
          map.remove(part);
        } else {
          const to = part.toNode;
          if (to !== null && !map.contains(to)) {
            map.remove(part);
          }
        }
      }
    }
    return map;
  }

} // #end CommandHandler
CommandHandler['className'] = 'CommandHandler';

/**
 * @hidden @internal
 * @suppress {duplicate}
 */
Diagram.setupCommandHandler = function() {
  return new CommandHandler();
};



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
GraphObject
*/

// import { List, Map, Set } from './collections';
// import { Debug } from './debug';
// import { EnumValue } from './enumValue';
// import { Util } from './util';

/**
 * @hidden
 * A type that describes a constructor.  Used in the declaration of GraphObject.make.
 */
export type ConstructorType<T extends new(...args: Array<any>) => InstanceType<T>> = new(...args: Array<any>) => InstanceType<T>;


/**
 * @hidden
 * A type that declares which types may be passed as arguments, depending on the type of instance being built.
 * Used in the declaration of GraphObject.make.
 */
export type MakeAllow<CT extends ConstructorType<CT>, C, E> = (InstanceType<CT> extends C ? E : never);

/**
 * This is the abstract base class for all graphical objects.
 * Classes inheriting from GraphObject include:
 * {@link Shape}, {@link TextBlock}, {@link Picture}, and {@link Panel}.
 * From the Panel class the {@link Part} class is derived, from which the
 * {@link Node} and {@link Link} classes derive.
 *
 * It is very common to make use of the static function {@link GraphObject.make} in order to build up
 * a visual tree of GraphObjects.  You can see many examples of this throughout the
 * Introduction, starting at <a href="../../intro/buildingObjects.html">Building Objects</a>,
 * and the Samples, starting with <a href="../../samples/minimal.html">Minimal Sample</a>.
 *
 * Since GraphObject is an abstract class, programmers do not create GraphObjects themselves,
 * but this class defines many properties used by all kinds of GraphObjects.
 *
 * The only visual properties on GraphObject are {@link #background} and {@link #areaBackground}.
 * However one can control whether the GraphObject is drawn at all by setting {@link #visible},
 * or by setting {@link #opacity} to zero if you still want the GraphObject to occupy space.
 * Also, if you want to control whether any mouse or touch events "see" the GraphObject,
 * you can set {@link #pickable} to false.
 *
 * For more information about specifying how things get drawn, see the properties on the
 * {@link Shape}, {@link TextBlock}, and {@link Picture} classes.
 * <h3>GraphObject Sizing</h3>
 *
 * GraphObject defines most of the properties that cause objects to size themselves differently.
 * The most prominent ones include:
 *   - The {@link #desiredSize}, {@link #minSize}, and {@link #maxSize} properties are used to
 *     explicitly set or limit the size of visual elements. {@link #width} and {@link #height} are
 *     convenience properties that set the {@link #desiredSize} width and height, respectively.
 *   - The {@link #angle} and {@link #scale} properties are used to transform visual elements.
 *   - The {@link #stretch} property determines how a GraphObject will fill its visual space,
 *     contextually granted to it by its containing {@link Panel}. Top-level ({@link Part})
 *     GraphObjects are not affected by this property because they are always granted infinite space.
 *
 * All GraphObjects in a Diagram are measured and then arranged by their containing {@link Panel}s in a tree-like fashion.
 * After measuring and arranging, a GraphObject will have valid values for the read-only
 * properties {@link #naturalBounds}, {@link #measuredBounds}, and {@link #actualBounds}.
 *
 *   - The {@link #naturalBounds} of a GraphObject describe its local size,
 *     without any transformations ({@link #scale}, {@link #angle}) affecting it.
 *   - The {@link #measuredBounds} of a GraphObject describe its size within its containing Panel.
 *   - The {@link #actualBounds} of a GraphObject describe its position and given size inside of its panel.
 *     This size may be smaller than {@link #measuredBounds}, for instance if a GraphObject with a large {@link #desiredSize}
 *     is placed in a {@link Panel} of a smaller {@link #desiredSize}. Smaller {@link #actualBounds} than {@link #measuredBounds}
 *     typically means an object will be cropped.
 *
 * <p class="boxread">
 * See <a href="../../intro/sizing.html">the Introduction page on sizing</a>
 * for usage information and examples.
 * <h3>GraphObject Size and Position within Panel</h3>
 * Several GraphObject properties guide the containing {@link Panel} for how to size and position the object within the panel.
 *   - The {@link #alignment} specifies where the object should be relative to some area of the panel.
 *     For example, an alignment value of {@link Spot.BottomRight} means that the GraphObject should be at the bottom-right corner of the panel.
 *   - The {@link #alignmentFocus} specifies precisely which point of the GraphObject should be aligned at the {@link #alignment} spot.
 *   - The {@link #column} and {@link #row} properties are only used by {@link Panel.Table} panels, to indicate where the GraphObject should be.
 *   - The {@link #columnSpan} and {@link #rowSpan} properties tell the {@link Panel.Table} panel how large the GraphObject should be.
 *   - The {@link #isPanelMain} property indicates to some kinds of {@link Panel}s that the GraphObject is the "primary" object
 *     that other panel children should be measured with or positioned in.
 *   - The {@link #margin} property tells the containing {@link Panel} how much extra space to put around this GraphObject.
 *   - The {@link #position} property is used to determine the relative position of GraphObjects when they are elements of a {@link Panel.Position} panel.
 *
 * <p class="boxread">
 * See <a href="../../intro/panels.html">the Introduction page on Panels</a>
 * and <a href="../../intro/tablePanels.html">Table Panels</a> for an overview of the capabilities.
 * <h3>Top-level GraphObjects are Parts</h3>
 *
 * A {@link Part} is a derived class of GraphObject representing a top-level object.
 * All top-level GraphObjects must be Parts, and Node, Link, Group, and Adornment derive from Part.
 * The position of a Part determines the point of the Part's top-left corner in document coordinates.
 * See also {@link Part#location}, which supports an way to specify the position based on a different
 * spot of a different element within the Part.
 *
 * There are several read-only properties that help navigate up the visual tree.
 *   - {@link #panel} returns the {@link Panel} that directly contains this GraphObject
 *   - {@link #part} returns the {@link Part} that this GraphObject is in, perhaps via intervening Panels;
 *     this is frequently used in order to get to the model data, {@link Panel#data}
 *   - {@link #layer} returns the {@link Layer} that this GraphObject's Part is in
 *   - {@link #diagram} returns the {@link Diagram} that this GraphObject's Part's Layer is in
 *
 * <p class="boxrun">
 * See <a href="../../samples/visualTree.html">the Visual Tree sample</a>
 * for a diagram displaying the visual tree of a simple diagram.
 * <h3>User Interaction</h3>
 *
 * GraphObjects have several properties enabling dynamic customizable interaction.
 * There are several definable functions that execute on input events: {@link #mouseDragEnter},
 * {@link #mouseDragLeave}, {@link #mouseDrop}, {@link #mouseEnter}, {@link #mouseHold},
 * {@link #mouseHover}, {@link #mouseLeave}, and {@link #mouseOver}.
 * For example, you could define mouse enter-and-leave event handlers to modify the appearance of a link
 * as the mouse passes over it:
 * ```js
 * myDiagram.linkTemplate =
 *   $(go.Link,
 *     $(go.Shape,
 *       { strokeWidth: 2, stroke: "gray" },  // default color is "gray"
 *       { // here E is the InputEvent and OBJ is this Shape
 *         mouseEnter: function(e, obj) { obj.strokeWidth = 4; obj.stroke = "dodgerblue"; },
 *         mouseLeave: function(e, obj) { obj.strokeWidth = 2; obj.stroke = "gray"; }
 *       }));
 * ```
 *
 * There are {@link #click}, {@link #doubleClick}, and {@link #contextClick} functions
 * that execute when a user appropriately clicks the GraphObject.
 * These click functions are called with the {@link InputEvent} as the first argument
 * and this GraphObject as the second argument.
 * For example, you could define a click event handler on a Node that goes to another page:
 * ```js
 * myDiagram.nodeTemplate =
 *   $(go.Node, "Auto",
 *     $(go.Shape, "RoundedRectangle",
 *       new go.Binding("fill", "color")),
 *     $(go.TextBlock,
 *       { name: "TB", margin: 3 },
 *       new go.Binding("text", "key")),
 *     { // second arg will be this GraphObject, which in this case is the Node itself:
 *       click: function(e, node) {
 *         window.open("https://en.wikipedia.org/Wiki/" + node.data.key);
 *       }
 *     });
 * ```
 *
 * Note: you may prefer defining {@link DiagramEvent} listeners on the {@link Diagram}
 * rather than on individual GraphObjects.  DiagramEvents also include more general events that
 * do not necessarily correspond to input events.
 *
 * The properties {@link #actionCancel}, {@link #actionDown}, {@link #actionMove},
 * and {@link #actionUp} define functions to execute when the GraphObject's {@link #isActionable} property
 * is set to true (default false). See the {@link ActionTool} for more detail.
 * <p class="boxread">
 * See <a href="../../intro/events.html">the Introduction page on Events</a> for a more general discussion.
 * <h3>GraphObjects as Ports</h3>
 *
 * In GoJS, {@link Link}s can only connect to elements within a {@link Node}
 * that are specified as "ports", and by default the only port is the Node itself.
 * Setting the {@link #portId} of a GraphObject inside a Node allows that object to act as a port.
 * Note: the only kind of model that can save which port a link is connected with, i.e. portIds that are not an empty string,
 * is a {@link GraphLinksModel} whose {@link GraphLinksModel#linkFromPortIdProperty} and
 * {@link GraphLinksModel#linkToPortIdProperty} have been set to name properties on the link data objects.
 *
 * GraphObjects have several properties that are only relevant when they are acting as ports.
 * These port-related properties are:
 *   - {@link #portId}, which must be set to a string that is unique within the {@link Node},
 *     in order for this GraphObject to be treated as a "port", rather than the whole node
 *   - {@link #fromSpot} and {@link #toSpot}, where a link should connect with this port
 *   - {@link #fromEndSegmentLength} and {@link #toEndSegmentLength}, the length of the link segment adjacent to this port
 *   - {@link #fromShortLength} and {@link #toShortLength}, the distance the link should terminate before touching this port
 *   - {@link #fromLinkable} and {@link #toLinkable}, whether the user may draw links connecting with this port
 *   - {@link #fromLinkableDuplicates} and {@link #toLinkableDuplicates}, whether the user may draw multiple links between the same pair of ports
 *   - {@link #fromLinkableSelfNode} and {@link #toLinkableSelfNode}, whether the user may draw a link between ports on the same node
 *   - {@link #fromMaxLinks} and {@link #toMaxLinks}, to limit the number of links connecting with this port in a particular direction
 *
 * <p class="boxread">
 * See <a href="../../intro/ports.html">the Introduction page on ports</a>
 * and <a href="../../intro/links.html">link routing</a>
 * and <a href="../../intro/connectionPoints.html">link connection points</a>
 * for port usage information and examples.
 * <h3>GraphObjects as labels on a Link</h3>
 *
 * GraphObjects can also be used as "labels" on a {@link Link}.
 * In addition to the {@link #alignmentFocus} property, these properties direct a Link Panel
 * to position a "label" at a particular point along the route of the link, in a particular manner:
 *   - {@link #segmentIndex}, which segment the label should be on
 *   - {@link #segmentFraction}, how far along the segment the label should be
 *   - {@link #segmentOffset}, where the label should be positioned relative to the segment
 *   - {@link #segmentOrientation}, how the label should be rotated relative to the angle of the segment
 *
 * <p class="boxread">
 * See <a href="../../intro/linkLabels.html">the Introduction page on link labels</a>
 * for examples of how to make use of labels on Links.
 * <h3>Interactive Behavior</h3>
 *
 * There are several properties that specify fairly high-level interactive behavior:
 *   - {@link #cursor}, a CSS string specifying a cursor
 *   - {@link #contextMenu}, an {@link Adornment}
 *   - {@link #toolTip}, an {@link Adornment}
 *
 * <p class="boxread">
 * For more information, please read <a href="../../intro/contextMenus.html">the Introduction page about Context Menus</a>
 * and <a href="../../intro/toolTips.html">the page about ToolTips</a>.
 * <p class="boxrun">
 * Also see <a href="../../samples/basic.html">the Basic sample</a>
 * for examples of how to show context menus and tooltips.
 * @unrestricted
 */
export abstract class GraphObject {
  /** @hidden @internal */ public static className = 'GraphObject';
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _portId: string | null;
  /** @hidden @internal */ private _row: number;
  /** @hidden @internal */ private _column: number;
  /** @hidden @internal */ private _spanAllocation: ((a: GraphObject, b: RowColumnDefinition, c: number) => number) | null;
  /** @hidden @internal */ private _cursor: string;
  /** @hidden @internal */ private _eventHandlers: GraphObjectEventHandlers | null;
  /** @hidden @internal */ private _copyProperties: Array<string> | null;
  /** @hidden @internal */ protected _linkSettings: LinkSettings | null;
  /** @hidden @internal */ protected _position: Point;
  /** @hidden @internal */ protected _measuredBounds: Rect;
  /** @hidden @internal */ public _reshapingBehavior: EnumValue;
  /** @hidden @internal */ public _flags: number;
  /** @hidden @internal */ protected _opacity: number;
  /** @hidden @internal */ public _panel: Panel | null;
  /** @hidden @internal */ protected _background: BrushLike;
  /** @hidden @internal */ protected _areaBackground: BrushLike;
  /** @hidden @internal */ protected _desiredSize: Size;
  /** @hidden @internal */ protected _minSize: Size;
  /** @hidden @internal */ protected _maxSize: Size;
  /** @hidden @internal */ public _transform: Transform;
  /** @hidden @internal */ public _allTransforms: Transform;
  /** @hidden @internal */ public _localTransform: Transform;
  /** @hidden @internal */ public _documentScale: number;
  /** @hidden @internal */ public _scale: number;
  /** @hidden @internal */ public _angle: number;
  /** @hidden @internal */ protected _stretch: EnumValue;
  /** @hidden @internal */ protected _margin: Margin;
  /** @hidden @internal */ public _actualBounds: Rect;
  /** @hidden @internal */ protected _naturalBounds: Rect;
  /** @hidden @internal */ public _portInfo: Knot | null;
  /** @hidden @internal */ protected _alignment: Spot;
  /** @hidden @internal */ protected _alignmentFocus: Spot;
  /** @hidden @internal */ protected _rowSpan: number;
  /** @hidden @internal */ protected _columnSpan: number;
  /** @hidden @internal */ private _segmentIndex: number;
  /** @hidden @internal */ protected _segmentFraction: number;
  /** @hidden @internal */ protected _segmentOffset: Point;
  /** @hidden @internal */ protected _segmentOrientation: EnumValue;
  /** @hidden @internal */ public _binds: List<Binding> | null;
  /** @hidden @internal */ public _bindingId: number;
  /** @hidden @internal */ public _srcbinds: List<Binding> | null;
  /** @hidden @internal */ public _clipRect: Rect;
  /** @hidden @internal */ protected _shadowVisible: boolean | null;
  /** @hidden @internal */ public _part: Part | null;
  /** @hidden @internal */ private _svg: SVGElement | null;
  /** @hidden @internal */ public _animationTriggers: Map<string, AnimationTrigger> | null;
  /** @hidden @internal */ private _runningTriggers: Map<AnimationTrigger, Animation> | null;

  /**
   * This is an abstract class, so you should not use this constructor.
   */
  constructor() {
    Set.uniqueHash(this);

    // BIT FLAGS:
    // GraphObject:
    //    0x00000001 this._visible = true;
    //    0x00000002 this._pickable = true;
    //    0x00000004 this._fromLinkableDuplicates = false;
    //    0x00000008 this._fromLinkableSelfNode = false;
    //    0x00000010 this._toLinkableDuplicates = false;
    //    0x00000020 this._toLinkableSelfNode = false;
    //    0x00000040 this._isPanelMain = false;
    //    0x00000080 this._isActionable = false;
    //    0x00000100 this._clipFlag = false; // internal flag set by panels. Tells whether or not object ought to be clipped
    //    0x00000200 this._alreadyShadowed = false; // Internal flag that stops the propogation of shadows if set to true
    //    0x00000400 this._bindsTwoWay = false; // Used in graphObject and panel
    //    0x00000800 this._invalidTransform = true;
    //    0x00001000 this._invalidAllTransforms = true;
    //    0x00002000 this._invalidMeasure = true;
    //    0x00004000 this._invalidArrange = true;
    // Panel:
    //    0x00100000 this._isAtomic = false;
    //    0x00200000 this._isClipping = false;
    //    0x00400000 this._isEnabled = true;
    //    0x00800000 this._hasNamedPorts = false;
    //    0x01000000 this._hasUplevelBindings = false;
    //    0x02000000 this._isOpposite = false;

    // 0x00000001|0x00000002|0x00000800|0x00001000|0x00002000|0x00004000|0x00400000
    this._flags = 4225027;
    this._opacity = 1;

    this._panel = null;
    this._name = '';

    this._background = null;
    this._areaBackground = null;

    this._position = new Point(NaN, NaN).freeze();
    this._desiredSize = Size.NoSize;
    this._minSize = Size.ZeroSize;
    this._maxSize = Size.InfiniteSize;

    this._transform = new Transform();
    this._allTransforms = new Transform();
    this._localTransform = new Transform();
    this._documentScale = 1; // valid as long as allTransforms is valid

    this._scale = 1;
    this._angle = 0;
    this._stretch = GraphObject.Default;

    this._margin = Margin.ZeroMargin;
    this._measuredBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._actualBounds = new Rect(NaN, NaN, NaN, NaN).freeze();
    this._naturalBounds = new Rect(0, 0, NaN, NaN).freeze();

    this._portInfo = null;

    this._portId = null;

    // In both GraphObject and Link:
    this._linkSettings = null;

    // for most Panel types:
    this._alignment = Spot.Default;
    this._alignmentFocus = Spot.Default;

    // for Table Panel:
    this._row = 0;
    this._rowSpan = 1;
    this._column = 0;
    this._columnSpan = 1;
    this._spanAllocation = null;

    // for Link Panel:
    this._segmentIndex = -Infinity;
    this._segmentFraction = 0;
    this._segmentOffset = Point.Origin;
    this._segmentOrientation = Link.None;

    // event handling

    this._cursor = '';

    // Virtualized attributes
    this._eventHandlers = null;

    this._binds = null;
    this._bindingId = -1;
    this._srcbinds = null;  // ?? could remember locally in Panel.freezeBindings, rather than here on GraphObject

    this._copyProperties = null;

    this._clipRect = null;
    this._shadowVisible = null;

    // This works because we test for reshapingBehavior.value which is 0 for both GraphOBject.None and LinkReshapingTool.None
    // This allows us to not reference LinkReshapingTool from GraphObject
    this._reshapingBehavior = GraphObject.None;

    // cache
    this._part = null;
    this._animationTriggers = null;
    this._runningTriggers = null;
  }

  /**
   * Copies properties from this object to the given object, which must be of the same class.
   * This is called by {@link #copy}.
   * This method may be overridden.
   *
   * For every property that you add to a subclass of a GraphObject-inheriting class,
   * in this method you should assign its value to the copied object.
   * For performance reasons you should set all of the same properties to <code>this</code>
   * that the constructor does, and in the same order.
   *
   * For example, let us define a custom {@link Link} class and add two properties:
   *
   * ```js
   * function CustomLink() {
   *   go.Link.call(this);
   *   this._someNewProperty = 17;
   *   this._someNewProperty2 = [];  // an Array
   * }
   * go.Diagram.inherit(CustomLink, go.Link);
   *
   * CustomLink.prototype.cloneProtected = function() {
   *   // Always call the base method in an override
   *   // In TypeScript you would use: super.cloneProtected(copy);
   *   go.Link.prototype.cloneProtected.call(this, copy);
   *   // assign every new property to the copy:
   *   copy._someNewProperty = this._someNewProperty;
   *   copy._someNewProperty2 = this._someNewProperty2.slice(0);  // make a copy of the Array
   * }
   * ```
   *
   * This ensures that copies of GraphObjects and their subclasses are faithful reproductions.
   * Consider for properties that are references to objects whether the reference should be shared
   * or whether that object value should be copied as well, resulting in a less shallow copy.
   * This is demonstrated above by making a copy of the property value that is an Array,
   * so that modifications to the Array will not be shared by copies of the CustomLink.
   * Further copies of the Array items might be warranted, depending on their purpose.
   *
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {GraphObject}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    // BIT FLAGS:
    //     0x00000001 copy._visible = this._visible;
    //     0x00000002 copy._pickable = this._pickable;
    //     0x00000004 copy._fromLinkableDuplicates = this._fromLinkableDuplicates;
    //     0x00000008 copy._fromLinkableSelfNode = this._fromLinkableSelfNode;
    //     0x00000010 copy._toLinkableDuplicates = this._toLinkableDuplicates;
    //     0x00000020 copy._toLinkableSelfNode = this._toLinkableSelfNode;
    //     0x00000040 copy._isPanelMain = this._isPanelMain;
    //     0x00000080 copy._isActionable = this._isActionable;
    //     0x00000100 copy._clipFlag = this._clipFlag;
    //     0x00000200 copy._alreadyShadowed = this._alreadyShadowed;
    //     0x00000400 copy._bindsTwoWay = this._bindsTwoWay;

    //   DONT COPY - Default true:
    //     0x00000800 copy._invalidTransform = this._invalidTransform;
    //   DONT COPY - Default true:
    //     0x00001000 copy._invalidAllTransforms = this._invalidAllTransforms;

    //     0x00002000 copy._invalidMeasure = this._invalidMeasure;
    //     0x00004000 copy._invalidArrange = this._invalidArrange;
    // Panel:
    //     0x00100000 copy._isAtomic = this._isAtomic;
    //     0x00200000 copy._isClipping = this._isClipping;
    //     0x00400000 copy._isEnabled = this._isEnabled;
    //     0x00800000 copy._hasNamedPorts = this._hasNamedPorts;
    //     0x01000000 copy._hasUplevelBindings= this._hasUplevelBindings;
    //     0x02000000 copy._isOpposite= this._isOpposite;

    copy._flags = this._flags | 0x00000800 | 0x00001000;
    copy._opacity = this._opacity;

    // new owner needs to set this backpointer: _panel
    copy._name = this._name;

    copy._background = this._background;  // Brushes can be shared
    copy._areaBackground = this._areaBackground;
    copy._animationTriggers = this._animationTriggers;
    // Not copying:
    // copy._runningTriggers = this._runningTriggers;

    copy._position.assign(this._position);  // always copy into local storage
    copy._desiredSize = this._desiredSize.copyFrozen();
    copy._minSize = this._minSize.copyFrozen();  // may be shared if the Spot is constant
    copy._maxSize = this._maxSize.copyFrozen();

    /* Not copying:
    copy._transform = this._transform;
    copy._allTransforms = this._allTransforms;
    */
    copy._localTransform = this._localTransform.copy();
    // don't copy: _documentScale, valid when allTransforms is valid

    copy._scale = this._scale;
    copy._angle = this._angle;
    copy._stretch = this._stretch;

    copy._margin = this._margin.copyFrozen();
    copy._measuredBounds.assign(this._measuredBounds);
    copy._actualBounds.assign(this._actualBounds);
    copy._naturalBounds.assign(this._naturalBounds);

    // don't copy Knot _portInfo

    copy._portId = this._portId;

    if (this._linkSettings !== null) {
      copy._linkSettings = this._linkSettings.copy();
    }

    // for most Panel types:
    copy._alignment = this._alignment.copyFrozen();
    copy._alignmentFocus = this._alignmentFocus.copyFrozen();

    // for Table Panel:
    copy._row = this._row;
    copy._rowSpan = this._rowSpan;
    copy._column = this._column;
    copy._columnSpan = this._columnSpan;
    copy._spanAllocation = this._spanAllocation;

    // for Link Panel:
    copy._segmentIndex = this._segmentIndex;
    copy._segmentFraction = this._segmentFraction;
    copy._segmentOffset = this._segmentOffset.copyFrozen();
    copy._segmentOrientation = this._segmentOrientation;

    // event handling; the functions are shared
    copy._cursor = this._cursor;

    // Virtualized attributes:
    if (this._eventHandlers !== null) {
      copy._eventHandlers = this._eventHandlers.copy();
    }

    // local bindings are shared
    copy._binds = this._binds;
    copy._bindingId = this._bindingId;
    // don't need to copy _srcbinds

    // copy those properties named in _copyProperties
    if (this._copyProperties !== null) {
      copy._copyProperties = Util.copyArray(this._copyProperties);
      // the named properties are only copied after the call to cloneProtected, below
    }

    if (this._clipRect !== null) {
      copy._clipRect = this._clipRect.copy();
    }
    copy._shadowVisible = this._shadowVisible;
    // cannot copy cached _part
  }

  /**
   * @hidden @internal
   * Add a property to be copied by cloneProtected.
   * Only called for properties that are not predefined, but are ad-hoc additions in particular apps.
   * Should only be used when this object may be copied, and certainly when it is part of a template.
   * @export
   * @this {GraphObject}
   * @param {string} propname
   */
  public addCopyProperty(propname: string): void {
    let pnames = this._copyProperties;
    if (!Util.isArray(pnames)) {
      pnames = [];
      this._copyProperties = pnames;
    } else {  // see if already present
      for (let i = 0; i < pnames.length; i++) {
        if (pnames[i] === propname) return;
      }
    }
    pnames.push(propname);
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can displace this object in the visual tree.
   * @this {GraphObject}
   * @param {GraphObject} copy  This might be a copy of this object,
   * but more likely is a separately created object of the same class with completely different internals.
   */
  protected displaceProtected(copy: this): void {
    copy._portInfo = null;  // don't copy Knot
    copy._part = null;  // recompute this cached reference
    copy.invalidateMeasure();
  }

  /**
   * @hidden @internal
   * Creates a shallow copy of this GraphObject and returns it.
   * This calls {@link #cloneProtected} with a new object of the same class as this object.
   * This is called by {@link #copy}.
   * @this {GraphObject}
   * @return {GraphObject} the newly copied object.
   */
  public clone(): this {
    const copy = new (this.constructor as any)();
    this.cloneProtected(copy);
    // copy those properties named in _copyProperties
    if (this._copyProperties !== null) {
      for (let i = 0; i < this._copyProperties.length; i++) {
        const pname = this._copyProperties[i];
        const that: any = this;
        copy[pname] = that[pname];
      }
    }
    return copy;
  }

  /**
   * Creates a deep copy of this GraphObject and returns it.
   * This method is the same as a clone for simple GraphObjects
   * such as {@link Shape}, {@link TextBlock}, and {@link Picture}.
   * For {@link Panel} this method copies the visual tree of GraphObjects that it contains.
   * @expose
   * @this {GraphObject}
   * @return {GraphObject}
   */
  public copy(): this {
    // no children, shallow clone is equivalent
    return this.clone();
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {GraphObject}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Link) {
      if (arg.name.indexOf('Orient') === 0) {
        this.segmentOrientation = arg;
      } else {
        Util.throwError('Unknown Link enum value for GraphObject.segmentOrientation property: ' + arg);
      }
    } else if (arg.classType === GraphObject) {
      this.stretch = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {string}
   */
  public toString(): string {
    return Util.getClassName(this.constructor) + '#' + Set.hashId(this);
  }

  /**
   * GraphObjects with this enumeration as the value of {@link GraphObject#stretch}
   * are not automatically scaled to fit in the given bounds;
   * there may be clipping in one or both directions if the available dimensions are too small.
   * @constant
   */
  public static None: EnumValue;

  /**
   * GraphObjects with this enumeration as the value of {@link GraphObject#stretch}
   * are stretched depending on the context they are used. For instance a 'Circle' figure might
   * be uniformly stretched whereas an 'Ellipse' figure might be non-uniformly stretched.
   * @constant
   */
  public static Default: EnumValue;

  // "any" type Because GraphObject.Vertical and Panel.Vertical mean different things
  // Same with horizontal
  /**
   * GraphObjects with this enumeration as the value of {@link GraphObject#stretch}
   * are scaled as much as possible in the y-axis. In another context, can be used as a value of PanelLayout, so type is "any".
   * @constant
   */
  public static Vertical: any; // EnumValue | PanelLayout;

  /**
   * GraphObjects with this enumeration as the value of {@link GraphObject#stretch}
   * are scaled as much as possible in the x-axis. In another context, can be used as a value of PanelLayout, so type is "any".
   * @constant
   */
  public static Horizontal: any; // EnumValue | PanelLayout;

  /**
   * GraphObjects with this enumeration as the value of {@link GraphObject#stretch}
   * are scaled in both directions so as to fit exactly in the given bounds;
   * there is no clipping but the aspect ratio may change, causing the object to appear stretched.
   * @constant
   */
  public static Fill: EnumValue;

  /**
   * Pictures with this enumeration as the value of {@link Picture#imageStretch} are drawn with equal
   * scale in both directions to fit the arranged (actual) bounds;
   * Panels of type Viewbox with this as the value of {@link Panel#viewboxStretch}
   * are scaled equally in both directions to fit in the given bounds.
   * @constant
   */
  public static Uniform: EnumValue;

  /**
   * Pictures with this enumeration as the value of {@link Picture#imageStretch} are drawn with equal
   * scale in both directions to fit the larger side of the image bounds;
   * Panels of type Viewbox with this as the value of {@link Panel#viewboxStretch}
   * are scaled equally in both directions to fit the larger side of the element's bounds.
   * there may be clipping in one dimension.
   * @constant
   */
  public static UniformToFill: EnumValue;

  /**
   * GraphObjects with this enumeration as the value of {@link Picture#flip} or {@link TextBlock#flip}
   * are drawn upside-down, with Y coordinate points increasing upwards.
   * @constant
   * @since 2.0
   */
  public static FlipVertical: EnumValue;

  /**
   * GraphObjects with this enumeration as the value of {@link Picture#flip} or {@link TextBlock#flip}
   * are drawn mirror-image, with X coordinate points increasing towards the left.
   * @constant
   * @since 2.0
   */
  public static FlipHorizontal: EnumValue;

  /**
   * GraphObjects with this enumeration as the value of {@link Picture#flip} or {@link TextBlock#flip}
   * are drawn with both X and Y coordinates reversed.
   * @constant
   * @since 2.0
   */
  public static FlipBoth: EnumValue;

  /**
   * @hidden @internal
   * Set up the GraphObjectEventHandlers subproperty, only if needed.
   * @this {GraphObject}
   */
  private initEventHandlers(): void {
    if (this._eventHandlers === null) {
      this._eventHandlers = new GraphObjectEventHandlers();
    }
  }

  /**
   * @hidden @internal
   * Set up the LinkSettings subproperty for objects acting as ports, only if needed.
   * Notice that the default values for general GraphObjects are different than those for Links.
   * @this {GraphObject}
   */
  protected initLinkSettings(): void {
    if (this._linkSettings === null) {
      const ls = new LinkSettings();
      ls._fromSpot = Spot.None;
      ls._toSpot = Spot.None;
      ls._fromEndSegmentLength = 10;
      ls._toEndSegmentLength = 10;
      ls._fromShortLength = 0;
      ls._toShortLength = 0;
      this._linkSettings = ls;
    }
  }

  /**
   * @hidden @internal
   * Call this method to notify that some state has changed.
   * This just calls the {@link #part}'s {@link Part#raiseChangedEvent} method.
   *
   * A simpler version of this method is {@link #raiseChanged}.
   * @export
   * @this {GraphObject}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raiseChangedEvent(change: EnumValue, propertyname: string, obj: ObjectData, oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    const part = this.part;
    if (part !== null) {
      part.raisePartChangedEvent(change, propertyname, obj, oldval, newval, oldparam, newparam);
      if (obj === this && change === ChangedEvent.Property) {
        if (this.getBindsTwoWay()) this.updateSourceBindings(part, propertyname);
      }

      const diagram = this.diagram;
      // No triggers during initial layout, no triggers during undo/redo
      if (this._animationTriggers !== null && diagram !== null &&
          diagram.isInitialLayoutCompleted && !diagram.undoManager.isUndoingRedoing && diagram.currentTool === diagram.toolManager) {
        const trigger = this._animationTriggers.get(propertyname);
        if (trigger !== null && !diagram.animationManager.isTicking) {
          if (this._runningTriggers === null) this._runningTriggers = new Map<AnimationTrigger, Animation>();
          let immediate = (diagram.undoManager.transactionLevel === 0); // Default conditions
          if (trigger.startCondition === AnimationTrigger.Immediate) {
            immediate = true;
          } else if (trigger.startCondition === AnimationTrigger.Bundled) {
            immediate = false;
          }
          if (immediate) {
            const animation = new Animation();
            trigger.applySettings(animation);
            const oldAnimation = this._runningTriggers.get(trigger);
            if (oldAnimation !== null) oldAnimation.stop(); // ?? perhaps transition some other way, or "re-route" the ongoing animation
            this._runningTriggers.add(trigger, animation);
            // Save the trigger and graphobject so that when the animation stops,
            // it can remove the animation from this._runningTriggers by calling removeTrigger
            // ?? might be a more efficient way to deal with this bookkeeping
            animation._triggerTarget = this;
            animation._trigger = trigger;
            animation.add(this, propertyname, oldval, newval);
            animation.start();
          } else { // transaction
            diagram.animationManager.prepareAnimation('Trigger');
            const animation = diagram.animationManager.defaultAnimation;
            animation.add(this, propertyname, oldval, newval);
            // will get started automatically at end of next transaction ?
          }

        }
      }

      // for now, only propagate Part property changes down to item panel bindings, when Binding.ofObject("/")
      if (this instanceof Panel && obj === part && part.getUplevelBindings() && part.data !== null) {
        const a = this._elements._arr;
        const l = a.length;
        for (let i = 0; i < l; i++) {
          const child = a[i];
          if (child instanceof Panel) {
            child.walkPanels(function (p) {
              if (p.data !== null && p.getUplevelBindings()) {
                p.updateTargetBindings(propertyname);
              }
            });
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Called by Animations triggered by an Immediate AnimationTrigger when they stop
   */
  public removeTrigger(trigger: AnimationTrigger): void {
    this._runningTriggers.remove(trigger);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {Part} part  provides context
   * @param {string} propertyname
   */
  protected updateSourceBindings(part: Part, propertyname: string): void {
    const binder = this.findTemplateBinder();
    if (binder !== null) {
      const it = this._binds.iterator;
      while (it.next()) {
        const bind = it.value;

        // normally bound to DATA, but use a named GraphObject if sourceName
        const srcname = bind.sourceName;
        let srcelt = null;
        if (srcname !== null) {
          srcelt = bind.findSourceGraphObject(binder, this);
          if (srcelt === null) continue;
          // maybe transfer from this GraphObject to source GraphObject
          bind.updateSource(this, srcelt, propertyname, null);
        } else if (bind.isToModel) {
          const diagram = part.diagram;
          if (diagram !== null && !diagram.skipsModelSourceBindings) {
            // maybe transfer from this GraphObject to source modelData
            bind.updateSource(this, diagram.model.modelData, propertyname, binder);
          }
        } else {
          const data = binder.data;
          if (data === null) continue;
          // maybe transfer from this GraphObject to data
          const diagram = part.diagram;
          if (diagram !== null && !diagram.skipsModelSourceBindings) {
            bind.updateSource(this, data, propertyname, binder);
          }
        }

        // maybe transfer from this named GraphObject to another GraphObject
        if (srcelt === this) {
          const dstid = bind.targetId;
          const dstelt = binder.findId(dstid);
          if (dstelt !== null) {
            bind.updateTarget(dstelt, srcelt, propertyname);
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Used by data binding to find the unique object in a visual tree corresponding to a particular object in the template.
   * This could cache references to reduce searching overhead.
   * @this {GraphObject}
   * @param {number} id int
   * @return {GraphObject}
   */
  public findId(id: number): GraphObject | null {
    if (this._bindingId === id) return this;
    return null;
  }

  /**
   * @hidden @internal
   * Call this method to notify about a property having changed value.
   * This just calls the {@link #part}'s {@link Part#raiseChangedEvent} method.
   * @export
   * @this {GraphObject}
   * @param {string} propertyname the name of the property.
   * @param {*} oldval the previous or old value for the property.
   * @param {*} newval the next or new value for the property.
   */
  protected raiseChanged(propertyname: string, oldval: any, newval: any): void {
    this.raiseChangedEvent(ChangedEvent.Property, propertyname, this, oldval, newval);
  }

  /**
   * @hidden @internal
   * Shared code among each measureProtected routine. Sets measuredBounds.
   * @this {GraphObject}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public _measureRect(x: number, y: number, w: number, h: number): void {
    const rect = this._measuredBounds;
    const tt = this._localTransform;
    tt.reset();
    this._setTransformAngle(tt, x, y, w, h);
    this._localTransform = tt;
    rect.assignTo(x, y, w, h);
    if (!tt.isIdentity()) {
      tt.transformRect(rect);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {Rect} r A Rect in container coordinates.
   * @param {boolean} partialInclusion True if objects are to be included when they
   * intersect.
   * @param {Transform} transform
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectIn(r: Rect, partialInclusion: boolean, transform: Transform): boolean {
    if (this.pickable === false) return false;
    transform.multiply(this.transform);
    return partialInclusion ? this.intersectsRect(r, transform) : this.containedInRect(r, transform);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {Point} p A Point in container coordinates.
   * @param {Point} distp A Point representing the distance from the point p.
   * @param {boolean} partialInclusion
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectNear(p: Point, distp: Point, partialInclusion: boolean): boolean {
    if (this.pickable === false) return false;
    const ns = this.naturalBounds;
    const dist = p.distanceSquaredPoint(distp);
    if (partialInclusion) {
      return (Point.distanceLineSegmentSquared(p.x, p.y, 0, 0, 0, ns.height) <= dist) ||
        (Point.distanceLineSegmentSquared(p.x, p.y, 0, ns.height, ns.width, ns.height) <= dist) ||
        (Point.distanceLineSegmentSquared(p.x, p.y, ns.width, ns.height, ns.width, 0) <= dist) ||
        (Point.distanceLineSegmentSquared(p.x, p.y, ns.width, 0, 0, 0) <= dist);
    } else {
      return p.distanceSquared(0, 0) <= dist &&
        p.distanceSquared(0, ns.height) <= dist &&
        p.distanceSquared(ns.width, 0) <= dist &&
        p.distanceSquared(ns.width, ns.height) <= dist;
    }
  }

  /**
   * @hidden @internal
   * Determine if an object's transform should or should not affect its children
   * This is only meaningful on Panel right now, and only for TableRow/TableColumn
   * If this method returns true then the transform should not be applied to children,
   * and calculations made with the object in question might need to substitute actual bounds for natural bounds,
   * to simulate the transformation that is not occurring
   *
   * When going up or down the tree of transforms you must skip their transforms for child objects
   * In other words in the visual tree: Table Panel -> TableRow -> Child
   * The child's transform goes Table Panel -> Child
   * TableRows and TableColumns still have their own transforms only for their own hit testing/drawing their own backgrounds
   * But otherwise they do not participate in the nested transform tree.
   * @this {GraphObject}
   * @return {boolean}
   */
  public trueTransform(): boolean {
    return true;
  }

  /**
   * @hidden @internal
   * Determine if a given point is inside and on this GraphObject.
   *
   * This method tries to return true for points near a stroke or near or
   * inside a possibly filled object such as an ellipse or a polygon.
   * This method ignores any drop shadow, but normally includes the
   * width of any line/stroke.
   *
   * The default behavior of this method is to return true if the
   * point p is within this object's Bounds.
   * If this object's width or height are zero,
   * the given point p has to be exactly on for this predicate to return true.
   *
   * However, some classes that are inherently somewhat "one dimensional",
   * may intentionally be more forgiving by supporting some margin nearby
   * where a point can be considered to be "inside" the object.
   * @export
   * @this {GraphObject}
   * @param {Point} p A Point in container coordinates.
   * @return {boolean} true if the argument p is considered to be "in"
   * this object.
   */
  public containsPoint(p: Point): boolean {
    if (Debug) Util.checkClass(p, Point, GraphObject, 'containsPoint:p');

    // Make sure the point is in its actualBounds in case it was clipped
    const parentsCoords = Point.alloc();
    parentsCoords.assign(p);
    this.transform.transformPoint(parentsCoords);
    const ab = this.actualBounds;
    if (!ab.isReal()) { Point.free(parentsCoords); return false; }
    const diagram = this.diagram;
    if (diagram !== null && diagram._isTouchOngoing) {
      const threshold = diagram.getInputOption('extraTouchThreshold');
      const area = diagram.getInputOption('extraTouchArea');
      const halfarea = area / 2;
      const nb = this.naturalBounds;
      const docscale = this.getDocumentScale() * diagram.scale;
      const sc = 1 / docscale;
      // if its in the bounds, iterate over this panel's collection in reverse order, calling containsPoint
      if (nb.width * docscale < threshold && nb.height * docscale < threshold) {
        const result = (Rect.contains(
          ab.x - (halfarea * sc), ab.y - (halfarea * sc),
          ab.width + (area * sc), ab.height + (area * sc),
          parentsCoords.x, parentsCoords.y));
        Point.free(parentsCoords);
        return result;
      }
    }

    // if its an adornment, it can skip this first check
    let returnval = false;
    if (this instanceof Adornment || (this instanceof Shape) ?
      Rect.contains(ab.x - 5, ab.y - 5, ab.width + 10, ab.height + 10, parentsCoords.x, parentsCoords.y) :
      ab.containsPoint(parentsCoords)) {
      if (this._clipRect && !this._clipRect.containsPoint(parentsCoords)) returnval = false;
      else if (this._areaBackground !== null && ab.containsPoint(parentsCoords)) returnval = true;
      else if (this._background !== null && this._naturalBounds.containsPoint(p)) returnval = true;
      else returnval = this.containsPointProtected(p);
    }
    Point.free(parentsCoords);
    return returnval;
  }

  /**
   * @hidden @internal
   * Determine if a point is within local bounds.
   * The default checks to see if the contained point is within the measured
   * bounds. Most classes override this functionality.
   * @this {GraphObject}
   * @param {Point} p A Point in local coordinates.
   * @param {boolean=} _precise optional argument that forces the strokewidth to stay small.
   * @return {boolean} true if the argument p is considered to be "in"
   * this object.
   */
  protected containsPointProtected(p: Point, _precise?: boolean): boolean {
    // this is set up so that the point coming in should be in local
    // co-ordinates and already transformed, so naturalBounds must be used instead
    // of measuredSize.
    const sz = this.naturalBounds;
    return Rect.contains(0, 0, sz.width, sz.height, p.x, p.y);
  }

  // ??? not used anywhere except test system

  /**
   * @hidden @internal
   * Determine if a given Rect is entirely contained within this GraphObject.
   * @export
   * @this {GraphObject}
   * @param {Rect} r A Rect in container coordinates.
   * @return {boolean} true if the argument r is considered to be "in" this object.
   */
  protected containsRect(r: Rect): boolean {
    if (Debug) Util.checkClass(r, Rect, GraphObject, 'containsRect:r');

    if (this.angle === 0) {
      return this.actualBounds.containsRect(r);
    }

    // See if all of the 4 r points are inside the container
    const ns = this.naturalBounds;
    const nb = Rect.allocAt(0, 0, ns.width, ns.height);
    const at = this.transform;
    let result = false;
    const tp = Point.allocAt(r.x, r.y);
    if (nb.containsPoint(at.invertedTransformPoint(tp))) {
      tp.assignTo(r.x, r.bottom);
      if (nb.containsPoint(at.invertedTransformPoint(tp))) {
        tp.assignTo(r.right, r.bottom);
        if (nb.containsPoint(at.invertedTransformPoint(tp))) {
          tp.assignTo(r.right, r.y);
          if (nb.containsPoint(at.invertedTransformPoint(tp))) {
            result = true;
          }
        }
      }
    }
    Point.free(tp);
    Rect.free(nb);
    return result;
  }

  /**
   * @hidden @internal
   * Determine if a given Rect is large enough to entirely contain this
   * GraphObject.
   * @export
   * @this {GraphObject}
   * @param {Rect} r A Rect in container coordinates.
   * @param {Transform=} transform A transform to use if something other than the GraphObject's own transform is desired.
   * @return {boolean} true if rect r contains the entire object.
   */
  protected containedInRect(r: Rect, transform?: Transform): boolean {
    if (Debug) Util.checkClass(r, Rect, GraphObject, 'containedInRect:r');

    // using its own transform means we can just check the actualBounds
    if (transform === undefined) {
      return r.containsRect(this.actualBounds);
    }

    // see if all of the four points are inside the rect
    const ns = this.naturalBounds;
    const at = transform;
    let result = false;
    const tp = Point.allocAt(0, 0);
    if (r.containsPoint(at.transformPoint(tp))) {
      tp.assignTo(0, ns.height);
      if (r.containsPoint(at.transformPoint(tp))) {
        tp.assignTo(ns.width, ns.height);
        if (r.containsPoint(at.transformPoint(tp))) {
          tp.assignTo(ns.width, 0);
          if (r.containsPoint(at.transformPoint(tp))) {
            result = true;
          }
        }
      }
    }
    Point.free(tp);
    return result;
  }

  /**
   * @hidden @internal
   * Determines if a given rectangle intersects the GraphObject's drawn bounds.
   * @export
   * @this {GraphObject}
   * @param {Rect} r a rectangle in container coordinates.
   * @param {Transform=} transform
   * @return {boolean}
   */
  protected intersectsRect(r: Rect, transform?: Transform): boolean {
    if (Debug) Util.checkClass(r, Rect, GraphObject, 'intersectsRect:r');

    if (transform === undefined) {
      transform = this.transform;
      // We can do this if there is no transform
      if (this.angle === 0) {
        return r.intersectsRect(this.actualBounds);
      }
    }

    // see if any of the four points are inside r
    // ??? possibly reundant because of intersectsLineSegment tests?
    const ns = this.naturalBounds;
    const at = transform;
    const r1 = Point.allocAt(0, 0);
    const r2 = Point.allocAt(0, ns.height);
    const r3 = Point.allocAt(ns.width, ns.height);
    const r4 = Point.allocAt(ns.width, 0);

    let result = false;
    if (r.containsPoint(at.transformPoint(r1)) ||
      r.containsPoint(at.transformPoint(r2)) ||
      r.containsPoint(at.transformPoint(r3)) ||
      r.containsPoint(at.transformPoint(r4))) {
      result = true;
    } else {
      // See if any of the 4 r points are inside the container
      const nb = Rect.allocAt(0, 0, ns.width, ns.height);
      const tp = Point.allocAt(r.x, r.y);
      if (nb.containsPoint(at.invertedTransformPoint(tp))) {
        result = true;
      } else {
        tp.assignTo(r.x, r.bottom);
        if (nb.containsPoint(at.invertedTransformPoint(tp))) {
          result = true;
        } else {
          tp.assignTo(r.right, r.bottom);
          if (nb.containsPoint(at.invertedTransformPoint(tp))) {
            result = true;
          } else {
            tp.assignTo(r.right, r.y);
            if (nb.containsPoint(at.invertedTransformPoint(tp))) {
              result = true;
            }
          }
        }
      }
      Point.free(tp);
      Rect.free(nb);

      if (!result) {
        // ??? see if any of 4 lines intersect the rect. Is there no better way?
        if (Geo.intersectsLineSegment(r, r1, r2) ||
          Geo.intersectsLineSegment(r, r2, r3) ||
          Geo.intersectsLineSegment(r, r3, r4) ||
          Geo.intersectsLineSegment(r, r4, r1)) {
          result = true;
        }
      }
    }

    Point.free(r1);
    Point.free(r2);
    Point.free(r3);
    Point.free(r4);
    return result;
  }

  /**
   * Returns the Point in document coordinates for a given Spot in this object's bounds
   * or for a Point in local coordinates.
   *
   * For example, for an instance of a Node like this:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node, "Auto",
   *     $(go.Shape, "RoundedRectangle",
   *       new go.Binding("fill", "color")),
   *     $(go.TextBlock,
   *       { name: "TB", margin: 3 },
   *       new go.Binding("text", "key")));
   * ```
   * where the Node is positioned at `100, 200`,
   * ```js
   *   node.findObject("TB").getDocumentPoint(go.Spot.Center)
   * ```
   * could return a Point that is approximately at `122, 213`.
   * @export
   * @this {GraphObject}
   * @param {Spot|Point} local a real Spot describing a relative location in or near this GraphObject, or a real Point in local coordinates.
   * @param {Point=} result an optional Point that is modified and returned.
   * @return {Point} in document coordinates.
   * @see {@link #getDocumentBounds}
   * @see {@link #getLocalPoint}
   */
  public getDocumentPoint(local: Spot | Point, result?: Point): Point {
    if (result === undefined) result = new Point();
    if (local instanceof Spot) {
      const s = local;
      if (Debug && s.isNoSpot()) Util.throwError('getDocumentPoint:s Spot must be specific: ' + s.toString());
      const ns = this.naturalBounds;
      result.assignTo((s.x * ns.width) + s.offsetX, (s.y * ns.height) + s.offsetY);
      this.allTransforms.transformPoint(result);
      return result;
    } else {
      result.set(local);  // may check that LOCAL is a Point
      this.allTransforms.transformPoint(result);
      return result;
    }
  }

  /**
   * Returns the Rect in document coordinates for this object's bounds.
   * If this GraphObject is a Part, the rect will be identical to its {@link #actualBounds}.
   * @export
   * @this {GraphObject}
   * @param {Rect=} result an optional Rect that is modified and returned.
   * @return {Rect} in document coordinates.
   * @see {@link #getDocumentPoint}
   * @since 2.0
   */
  public getDocumentBounds(result?: Rect): Rect {
    if (result === undefined) result = new Rect();
    const ns = this.naturalBounds;
    const t = this.allTransforms;
    const p = Point.allocAt(0, 0).transform(t); // Top left spot
    result.assignTo(p.x, p.y, 0, 0);
    p.assignTo(ns.width, 0).transform(t); // Top right
    result.unionNoCheck(p.x, p.y, 0, 0);
    p.assignTo(ns.width, ns.height).transform(t); // Bottom right
    result.unionNoCheck(p.x, p.y, 0, 0);
    p.assignTo(0, ns.height).transform(t); // bottom left
    result.unionNoCheck(p.x, p.y, 0, 0);
    Point.free(p);
    return result;
  }

  /**
   * Returns the effective angle that the object is drawn at, in document coordinates, normalized to between 0 and 360.
   *
   * Basically this adds together all of the rotation declared by this {@link #angle} and the angles of all of its containing Panels, including the Part.
   * @export
   * @this {GraphObject}
   * @return {number}
   */
  public getDocumentAngle(): number {
    return this.allTransforms.getAngle();
  }

  /**
   * Returns the total scale that the object is drawn at, in document coordinates.
   *
   * Basically this multiplies together this {@link #scale} with the scales of all of its containing Panels, including the Part.
   * @export
   * @this {GraphObject}
   * @return {number}
   */
  public getDocumentScale(): number {
    if (this.getInvalidAllTransforms() === false) return this._documentScale;
    const scale = this._scale;
    if (this.panel !== null) return scale * this.panel.getDocumentScale();
    return scale;
  }

  /**
   * Given a Point in document coordinates, returns a new Point in local coordinates.
   *
   * For example, if you have a mouse event whose {@link InputEvent#documentPoint} is at `122, 213`,
   * and if you have a Node whose {@link #position} is at `100, 200`,
   * `node.getLocalPoint(e.documentPoint)` could return a Point that is at `22, 13`.
   * For a GraphObject within the Node named "TB",
   * ```js
   *   node.findObject("TB").getLocalPoint(e.documentPoint)
   * ```
   * could return a Point that is at `15.7, 6.7`,
   * if that "TB" object is positioned somewhat inside the bounds of the Node.
   * @export
   * @this {GraphObject}
   * @param {Point} p a real Point in document coordinates.
   * @param {Point=} result an optional Point that is modified and returned.
   * @return {Point} The corresponding Point in local coordinates.
   * @see {@link #getDocumentPoint}
   */
  public getLocalPoint(p: Point, result?: Point): Point {
    if (result === undefined) result = new Point();
    result.assign(p);
    this.allTransforms.invertedTransformPoint(result);
    return result;
  }

  /**
   * @hidden @internal
   * Find the closest point in this object to the given point p1 that is on the
   * line p1-p2.
   * @export
   * @this {GraphObject}
   * @param {Point} p1 the point we are looking to be closest to, on the line formed with p2.
   * @param {Point} p2 forms a line with p1.
   * @param {Point} result (reference) the point of this object that is closest to p1 and that is on the infinite line from p1 to p2.
   * @return {boolean} true if the infinite line does intersect with this object; false otherwise.
   */
  protected getNearestIntersectionPoint(p1: Point, p2: Point, result: Point): boolean {
    return this.getNearestIntersectionPoint2(p1.x, p1.y, p2.x, p2.y, result);
  }

  // getNearestIntersectionPoint without using Points or calling transform methods
  /**
   * @hidden @internal
   * Find the closest point in this object to the given point p1 that is on the
   * line p1-p2.
   * @this {GraphObject}
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {Point} result (reference) the point of this object that is closest to p1 and that is on the infinite line from p1 to p2.
   * @return {boolean} true if the infinite line does intersect with this object; false otherwise.
   */
  public getNearestIntersectionPoint2(p1x: number, p1y: number, p2x: number, p2y: number, result: Point): boolean {
    const tt = this.transform;
    const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
    const m0 = tt.m22 * d;
    const m1 = -tt.m12 * d;
    const m2 = -tt.m21 * d;
    const m3 = tt.m11 * d;
    const m4 = d * (tt.m21 * tt.dy - tt.m22 * tt.dx);
    const m5 = d * (tt.m12 * tt.dx - tt.m11 * tt.dy);

    if (this.areaBackground !== null) {
      // find intersection with transformed bounds
      const ab = this.actualBounds;
      return Geo.getNearestIntersectionPoint(ab.left, ab.top, ab.right, ab.bottom, p1x, p1y, p2x, p2y, result);
    } else {
      // Inverted transform point
      const p12x = p1x * m0 + p1y * m2 + m4;
      const p12y = p1x * m1 + p1y * m3 + m5;

      const p22x = p2x * m0 + p2y * m2 + m4;
      const p22y = p2x * m1 + p2y * m3 + m5;
      result.assignTo(0, 0);
      const nb = this.naturalBounds;
      const found = Geo.getNearestIntersectionPoint(0, 0, nb.width, nb.height, p12x, p12y, p22x, p22y, result);
      result.transform(tt);
      return found;
    }
  }

  /**
   * @hidden @internal
   * Given the available size and desired size, measure this GraphObject and
   * determine its expected drawing size. Sets the measuredBounds of the object.
   * This sets {@link #measuredBounds}.
   * @this {GraphObject}
   * @param {number} width The available width in panel coordinates.
   * @param {number} height The available height in panel coordinates.
   * @param {number=} minw
   * @param {number=} minh
   */
  public measure(width: number, height: number, minw?: number, minh?: number): void {
    // don't measure unless invalid
    if (this.getInvalidMeasure() === false) return;
    const marg = this._margin;
    const margw = marg.right + marg.left;
    const margh = marg.top + marg.bottom;
    width = Math.max(width - margw, 0);
    height = Math.max(height - margh, 0);
    minw = minw || 0;
    minh = minh || 0;
    minw = Math.max(minw - margw, 0);
    minh = Math.max(minh - margh, 0);

    const ang = this.angle;
    let tempw = 0.0;
    const des = this.desiredSize;
    // desiredSize for shapes represents just the geometry size
    // If desired size dictates the available width/height make sure its really desired+sw
    let sw = 0.0;
    if (this instanceof Shape) sw = this.strokeWidth;

    // A desiredSize always overrides any fill
    if (ang === 90 || ang === 270) {
      width = isFinite(des.height) ? des.height + sw : width;
      height = isFinite(des.width) ? des.width + sw : height;
    } else {
      width = isFinite(des.width) ? des.width + sw : width;
      height = isFinite(des.height) ? des.height + sw : height;
    }

    let tempminw = minw || 0;
    let tempminh = minh || 0;
    const ispanel = (this instanceof Panel);
    const stretch = this.getEffectiveStretch(true);
    switch (stretch) {
      case GraphObject.None:
        // We don't care if its not stretching
        tempminw = 0;
        tempminh = 0;
        if (ispanel) {
          width = Infinity;
          height = Infinity;
        }
        break;
      case GraphObject.Fill:
        if (isFinite(width) && width > minw) tempminw = width;
        if (isFinite(height) && height > minh) tempminh = height;
        break;
      case GraphObject.Horizontal:
        if (isFinite(width) && width > minw) tempminw = width;
        // We don't care if its not stretching
        tempminh = 0;
        if (ispanel) height = Infinity;
        break;
      case GraphObject.Vertical:
        if (isFinite(height) && height > minh) tempminh = height;
        // We don't care if its not stretching
        tempminw = 0;
        if (ispanel) width = Infinity;
        break;
    }
    const max = this.maxSize;
    const min = this.minSize;
    if (tempminw > max.width && min.width < max.width) tempminw = max.width;
    if (tempminh > max.height && min.height < max.height) tempminh = max.height;
    minw = Math.max(tempminw / this.scale, min.width); // scale is needed here but not in subclass' measureProtected
    minh = Math.max(tempminh / this.scale, min.height);
    // If the projected minimum is more than the maximum, bring it down to the smallest minimum possible
    if (max.width < minw) minw = Math.min(min.width, minw);
    if (max.height < minh) minh = Math.min(min.height, minh);
    width = Math.min(max.width, width);
    height = Math.min(max.height, height);
    width = Math.max(minw, width);
    height = Math.max(minh, height);

    // ??? minimal fix for right angles only:
    if (ang === 90 || ang === 270) {
      tempw = width;
      width = height;
      height = tempw;
      tempw = minw;
      minw = minh;
      minh = tempw;
    }

    this._measuredBounds.thaw();
    this.measureProtected(width, height, minw, minh);

    this._measuredBounds.freeze();
    if (!this._measuredBounds.isReal()) {
      Util.throwError('Non-real measuredBounds has been set. Object ' +
        this + ', measuredBounds: ' + this._measuredBounds.toString());
    }
    this.setInvalidMeasure(false);
  }

  /**
   * @hidden @internal
   * Given the available size and desired size, measure this GraphObject and
   * determine its expected drawing size. Sets the measuredBounds of the object.
   * This sets {@link #measuredBounds}.
   * @this {GraphObject}
   * @param {number} width The available width in panel coordinates.
   * @param {number} height The available height in panel coordinates.
   * @param {number} minw
   * @param {number} minh
   */
  protected measureProtected(width: number, height: number, minw: number, minh: number): void {
    // override me!
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public isPart(): this is Part {
    return false;
  }

  /**
   * @hidden @internal
   * Arranges the GraphObject onto its parent.
   * The passed-in numbers typically account for {@link GraphObject#margin} and other offsets.
   * This sets {@link #actualBounds}.
   * @this {GraphObject}
   * @param {number} fx The final x value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fy The final y value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fw The final width value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fh The final height value of actualBounds that the Panel computes for the GraphObject.
   * @param {Rect=} clipRect an optional area to constrain this actualBounds to when hit testing/drawing. Used with Fixed/Table panels
   * element, provided as a Rect.
   */
  public arrange(fx: number, fy: number, fw: number, fh: number, clipRect?: Rect): void {
    this._invalidateTransform();  // re-arranging or setting a new position requires invalidating the transform

    const oldActual = Rect.alloc();
    oldActual.assign(this._actualBounds);

    this._actualBounds.thaw();
    if (this.getInvalidArrange() === false) { // if the arrange is valid just update the actualBounds
      this._actualBounds.assignTo(fx, fy, fw, fh);
    } else {
      this.arrangeProtected(fx, fy, fw, fh);
    }
    this._actualBounds.freeze();

    if (clipRect === undefined) {
      this._clipRect = null;
    } else {
      this._clipRect = clipRect;
    }
    let clipFlag = false;
    if (clipRect !== undefined) {
      clipFlag = true;
    } else {
      let panel = this.panel;
      if (panel !== null && (panel.type === Panel.TableRow || panel.type === Panel.TableColumn)) panel = panel.panel;
      if (panel !== null) {
        const nb = panel._naturalBounds;
        let m = this.measuredBounds;
        // If there's an areaBackground then the drawn space is potentially
        // larger than the measured Bounds, so we have to clip more
        if (this.areaBackground !== null) m = this._actualBounds;
        const height = fy + m.height;
        const width = fx + m.width;
        clipFlag = !(
          0 <= fx + 0.05 &&
          width <= nb.width + 0.05 &&
          0 <= fy + 0.05 &&
          height <= nb.height + 0.05);
        if (this instanceof TextBlock) {
          const n = this.naturalBounds;
          if (this._maxPosition > n.height || this._maxLineWidth > n.width) {
            clipFlag = true;
          }
        }
      }
    }
    if (clipFlag) {
      this._flags |= 0x00000100;
    } else {
      this._flags &= ~0x00000100;
    }

    if (!this._actualBounds.isReal()) {
      Util.throwError('Non-real actualBounds has been set. Object ' +
        this + ', actualBounds: ' + this._actualBounds.toString());
    }

    this.onBoundsChanged(oldActual, this._actualBounds);
    this.setInvalidArrange(false);
    Rect.free(oldActual);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    //  Util.throwError('arrangeProtected not implemented for this class');
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected commonArrange(fx: number, fy: number, fw: number, fh: number): void {
    this._actualBounds.assignTo(fx, fy, fw, fh);
    if (this.desiredSize.isReal()) return;

    const mb = this._measuredBounds;
    const marg = this._margin;
    const margw = marg.right + marg.left;
    const margh = marg.top + marg.bottom;
    const mbwidth = mb.width + margw;
    const mbheight = mb.height + margh;
    fw += margw;
    fh += margh;

    let effectivestretch = this.getEffectiveStretch(true);
    if (mbwidth === fw && mbheight === fh) effectivestretch = GraphObject.None;
    switch (effectivestretch) {
      case GraphObject.None: // if fw or fh are smaller than our size, we have to remeasure
        if (mbwidth > fw || mbheight > fh) {
          this.setInvalidMeasure(true);
          this.measure(mbwidth > fw ? fw : mbwidth, mbheight > fh ? fh : mbheight, 0, 0);
        }
        break;
      case GraphObject.Fill: // fill the space completely, ignoring desired aspect ratio
        this.setInvalidMeasure(true);
        this.measure(fw, fh, 0, 0);
        break;
      case GraphObject.Horizontal:
        this.setInvalidMeasure(true);
        this.measure(fw, mbheight, 0, 0);
        break;
      case GraphObject.Vertical:
        this.setInvalidMeasure(true);
        this.measure(mbwidth, fh, 0, 0);
        break;
    } // end stretch switch
  }

  /**
   * @hidden @internal
   * Called at the end of GraphObject.arrange, Part.arrange, and Part.validatePosition (used in lieu of arrange)
   * @this {GraphObject}
   * @param {Rect} oldb
   * @param {Rect} newb
   */
  protected onBoundsChanged(oldb: Rect, newb: Rect): void {
    // we should check to see if oldb equals newb
    // the rest can be skipped if that is true
    const part = this.part;
    if (part === null) return;
    const diagram = part.diagram;
    if (diagram === null) return;

    if (part.selectionObject === this ||
      part.resizeObject === this ||
      part.rotateObject === this) part.setNotifyLater(true);

    this.invalidateDraw();

    if (oldb.equalsApproxClose(newb)) return;
    // do these if bounds really are different
    part.invalidateContainingGroup();
    this._notifyPorts(part);
  }

  /**
   * @hidden @internal
   * See also on Panel
   * ??? This sorta duplicates functionality of _invalidatePortLinkRoutes -- need to rearchitect and reimplement
   * @this {GraphObject}
   * @param {Part} part
   */
  public _notifyPorts(part: Part): void {
    if (this.portId === null) return;
    part.setNotifyLater(true);
    if (part instanceof Node) {
      part.invalidatePortLinks(this);
    }
  }

  /**
   * Gets or sets whether or not this GraphObject will be shadowed inside a Part that has {@link Part#isShadowed} set to true.
   *
   * The default is null, which means this GraphObject will obey the default shadow rules (see {@link Part#isShadowed}).
   *
   * A value of true or false will ensure that this part is shadowed or not regardless of the default shadow rules,
   * but this GraphObject's shadowed status will not affect other GraphObjects in the Part.
   *
   * Typically this property does not need to be set, but you may need to set this value to false
   * on GraphObjects inside a Part that you do not wish to be shadowed.
   * @see {@link Part#isShadowed}
   * @since 1.6
   */
  /** @export */ get shadowVisible(): boolean | null { return this._shadowVisible; }
  /** @export */ set shadowVisible(value: boolean | null) {
    const old = this._shadowVisible;
    if (old !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'boolean', GraphObject, 'shadowVisible');
      this._shadowVisible = value;
      this.invalidateDraw();
      this.raiseChanged('shadowVisible', old, value);
    }
  }


/**
 * @hidden @internal
 * see if something is in bounds before drawing
 * @this {GraphObject}
 * @param {Diagram} diagram
 * @param {number} extraSpace additional space to add to bounds
 * @return {boolean}
 */
private notInBounds(diagram: Diagram, extraSpace: number): boolean {
  const ns = this.naturalBounds;
  const t = this._allTransforms;
  const tm11 = t.m11;
  const tm21 = t.m21;
  const tdx = t.dx;
  const tm12 = t.m12;
  const tm22 = t.m22;
  const tdy = t.dy;

  let px = 0;
  let py = 0;
  // transform top left spot:
  let tx = px;
  let ty = py;
  px = tx * tm11 + ty * tm21 + tdx;
  py = tx * tm12 + ty * tm22 + tdy;
  let resultx = px;
  let resulty = py;
  let resultwidth = 0;
  let resultheight = 0;

  // transform top right spot:
  tx = ns.width + extraSpace;
  ty = 0;
  px = tx * tm11 + ty * tm21 + tdx;
  py = tx * tm12 + ty * tm22 + tdy;
  let minx = Math.min(resultx, px);
  let miny = Math.min(resulty, py);
  resultwidth = Math.max(resultx + resultwidth, px) - minx;
  resultheight = Math.max(resulty + resultheight, py) - miny;
  resultx = minx;
  resulty = miny;

  // Bottom right
  tx = ns.width + extraSpace;
  ty = ns.height + extraSpace;
  px = tx * tm11 + ty * tm21 + tdx;
  py = tx * tm12 + ty * tm22 + tdy;
  minx = Math.min(resultx, px);
  miny = Math.min(resulty, py);
  resultwidth = Math.max(resultx + resultwidth, px) - minx;
  resultheight = Math.max(resulty + resultheight, py) - miny;
  resultx = minx;
  resulty = miny;

  // bottom left
  tx = 0;
  ty = ns.height + extraSpace;
  px = tx * tm11 + ty * tm21 + tdx;
  py = tx * tm12 + ty * tm22 + tdy;
  minx = Math.min(resultx, px);
  miny = Math.min(resulty, py);
  resultwidth = Math.max(resultx + resultwidth, px) - minx;
  resultheight = Math.max(resulty + resultheight, py) - miny;
  resultx = minx;
  resulty = miny;

  const vpb = diagram.viewportBounds;
  const x = vpb.x;
  const y = vpb.y;
  const w = vpb.width;
  const h = vpb.height;

  // intersects:
  if (resultx > w + x || x > resultwidth + resultx) return true;
  if (resulty > h + y || y > resultheight + resulty) return true;
  return false;
  /*
  if (!intersects) {
    // if (window.declined === undefined) window.declined = 0;
    // window.declined++;
    return true;
  }
  // if (window.inclined === undefined) window.inclined = 0;
  // window.inclined++;
  return false;
  */
}

  /**
   * @hidden @internal
   * The draw function. Handles translation and rotation, then calls the
   * internal draw for the specific GraphObject.
   * @this {GraphObject}
   * @param {IContext} context a canvas drawing context.
   * @param {Diagram} diagram
   */
  public draw(context: IContext, diagram: Diagram): void {
    if (!this.visible) return;
    const tableRowOrCol = (this instanceof Panel && (this.type === Panel.TableRow || this.type === Panel.TableColumn));
    const ab = this._actualBounds;
    if (!tableRowOrCol && (ab.width === 0 || ab.height === 0 || isNaN(ab.x) || isNaN(ab.y))) {
      return;
    }
    const opacity = this.opacity;
    if (opacity === 0) return;
    let old = 1.0;
    if (opacity !== 1) {
      old = context.globalAlpha;
      context.globalAlpha = old * opacity;
    }

    const svg = this.drawForSVG(context, diagram);
    if (svg) {
      if (opacity !== 1) context.globalAlpha = old;
      return;
    }

    if (tableRowOrCol) {
      this.drawRowColumn(context, diagram);
      if (opacity !== 1) context.globalAlpha = old;
      return;
    }

    // Create geometries that have been invalidated due to delay
    if (this instanceof Link) {
      this.makeShapeGeometry(false);
    }

    if (Debug && Debug.boundsInfoEnabled && Debug.drawBoundsInfoActual) { Debug.drawBoundsInfoActual(context, this); }

    const tt = this.transform;
    const myPanel = this.panel;
    // construct the allTransforms for this object
    if (this.getInvalidAllTransforms() === true) { this._validateTransforms(); }

    const part = this.part;
    let isShadowed = false;
    let extraSpace = 0;
    if (part && diagram.getRenderingHint('drawShadows')) {
      isShadowed = part.isShadowed;
      if (isShadowed) {
        const shad = part.shadowOffset;
        extraSpace = Math.max(shad.y, shad.x) * diagram.scale * diagram._pixelRatio;
      }
    }

    // cannot do when drawing to foreign context, viewport may differ
    if (!diagram._isDrawingForeign && this.isPart() && this.notInBounds(diagram, extraSpace)) {
      if (opacity !== 1) context.globalAlpha = old;
      return;
    }

    let clip = (this._flags & 0x00000100) !== 0; // clipFlag;
    if (context.clipInsteadOfFill) clip = false;
    // done before potential clip to reduce cache clearing due to save and restore
    if (this instanceof TextBlock) {
      context.font = this.font;
    }
    if (clip) {
      if (Debug && Debug.drawClipEnabled) Util.trace('clip' + this.toString());
      const nb2 = myPanel.trueTransform() ? myPanel.naturalBounds : myPanel.actualBounds;
      let cx;
      let cy;
      let cwidth;
      let cheight;
      if (this._clipRect !== null) {
        const cr = this._clipRect;
        cx = cr.x;
        cy = cr.y;
        cwidth = cr.width;
        cheight = cr.height;
      } else {
        cx = Math.max(ab.x, nb2.x);
        cy = Math.max(ab.y, nb2.y);
        cwidth = Math.min(ab.right, nb2.right) - cx;
        cheight = Math.min(ab.bottom, nb2.bottom) - cy;
      }
      // ?? this second set of conditions causes issues if the natural bounds does not reflect the true size if it were unclipped
      // this can happen in Table panels if the text is given a clip rectangle and an offset
      if (cx > ab.width + ab.x || ab.x > nb2.width + nb2.x /*||
      cy > ab.height + ab.y || ab.y > nb2.height + nb2.y*/) {
        if (opacity !== 1) context.globalAlpha = old;
        return;
      }

      // if (Debug && Debug.drawClipEnabled) Debug.drawClip(context, cx, cy, cwidth, cheight);

      context.save();
      context.beginPath();
      context.rect(cx, cy, cwidth, cheight);
      context.clip();
    }

    if (this.isPart()) {
      if (!part.isVisible()) { // ??? can't this be optimized with setInView???
        if (opacity !== 1) context.globalAlpha = old;
        return;
      }
      if (isShadowed) {
        const shad = part.shadowOffset;
        context.shadowsSet(
          shad.x * diagram.scale * diagram._pixelRatio,
          shad.y * diagram.scale * diagram._pixelRatio,
          part.shadowBlur);
          context.shadowsOn();
          context.shadowColor = part.shadowColor;
      }
    }

    // If explicitly turned on or off
    if (this.shadowVisible === true) {
      context.shadowsOn();
    } else if (this.shadowVisible === false) {
      context.shadowsOff();
    }
    const nb = this.naturalBounds;
    // areaBackground is in the parent's transform space
    if (this._areaBackground !== null) {
      this.setBrush(context, this._areaBackground, true, true, nb, ab);
      if (this._areaBackground instanceof Brush && this._areaBackground.type === Brush.Radial) {
        context.beginPath();
        context.rect(ab.x, ab.y, ab.width, ab.height);
        context.fillContext(this._areaBackground);
      } else {
        context.fillRect(ab.x, ab.y, ab.width, ab.height);
      }
    }

    context.transform(tt.m11, tt.m12, tt.m21, tt.m22, tt.dx, tt.dy); // checks for identity

    // Consider not drawing background shadows:
    if (isShadowed &&
      ((myPanel !== null && myPanel.getAlreadyShadowed()) ||
      (myPanel !== null && (myPanel.type === Panel.Auto || myPanel.type === Panel.Spot) && myPanel.findMainElement() !== this))) {
      if (this.shadowVisible === null) context.shadowsOff();
    }

    // background is in local space
    if (this._background !== null) {
      let ns = this.naturalBounds;
      let nx = 0.0;
      let ny = 0.0;
      let nw = ns.width;
      let nh = ns.height;
      let sw = 0.0;
      if (this instanceof Shape) {
        ns = this._geometry.bounds;
        nx = ns.x;
        ny = ns.y;
        nw = ns.width;
        nh = ns.height;
        sw = this.strokeWidth;
      }
      this.setBrush(context, this._background, true, false, nb, ab);
      if (this._background instanceof Brush && this._background.type === Brush.Radial) {
        context.beginPath();
        context.rect(nx - (sw / 2), ny - (sw / 2), nw + sw, nh + sw);
        context.fillContext(this._background);
      } else {
        context.fillRect(nx - (sw / 2), ny - (sw / 2), nw + sw, nh + sw);
      }
    }

    if (Debug && Debug.boundsInfoEnabled && Debug.drawBoundsInfoNatural) { Debug.drawBoundsInfoNatural(context, this); }

    // conditions for stopping drawing + propagation of shadows
    // 1. background exists (background-drawn shadows happens above this code)
    // 2. has panel and is already shadowed
    // 3. has auto/spot panel and this isn't the main element
    if (isShadowed &&
      (this._background !== null || this._areaBackground !== null ||
        (myPanel !== null && myPanel.getAlreadyShadowed()) ||
        (myPanel !== null && (myPanel.type === Panel.Auto || myPanel.type === Panel.Spot) && myPanel.findMainElement() !== this))) {
      this.setAlreadyShadowed(true); // no more shadows
      if (this.shadowVisible === null) context.shadowsOff();
    } else { // continue with shadows
      this.setAlreadyShadowed(false);
    }

    this.drawProtected(context, diagram);

    // Reset things for the next sibling
    if (isShadowed && this.getAlreadyShadowed() === true) {
      context.shadowsOn();
    }

    // When done drawing the whole part, reset any shadow:
    if (this.isPart() && isShadowed) {
      context.shadowsOff();
      // ??? reset diagram._shadow... too?
    }

    if (clip) {
      context.restore();
      if (this instanceof Panel) {
        context.clearContextCache(true);
      } else {
        context.clearContextCache(false);
      }
    } else if (!tt.isIdentity()) {
      // for when not saving+clipping+restoring
      const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
      context.transform(
        tt.m22 * d,
        -tt.m12 * d,
        -tt.m21 * d,
        tt.m11 * d,
        d * (tt.m21 * tt.dy - tt.m22 * tt.dx),
        d * (tt.m12 * tt.dx - tt.m11 * tt.dy));
    }

    if (opacity !== 1) context.globalAlpha = old;
  } // end draw

  /**
   * @hidden @internal
   * The draw function when rendering to an SVG context.
   * @this {GraphObject}
   * @param {IContext} context Special SVG drawing context
   * @param {Diagram} diagram
   * @return {boolean} true if implemented
   */
  public drawForSVG(context: IContext, diagram: Diagram): boolean {
    return false;
  } // end drawForSVG

  // ??? could redo now that transforms are on the table for TableRow/Column
  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {IContext} context
   * @param {Diagram} diagram
   */
  public drawRowColumn(context: IContext, diagram: Diagram): void {
    const ab = this._actualBounds;
    const nb = this._naturalBounds;
    // areaBackground is in the parent's transform space
    if (this._areaBackground !== null) {
      this.setBrush(context, this._areaBackground, true, true, nb, ab);
      if (this._areaBackground instanceof Brush && this._areaBackground.type === Brush.Radial) {
        context.beginPath();
        context.rect(ab.x, ab.y, ab.width, ab.height);
        context.fillContext(this._areaBackground);
      } else {
        context.fillRect(ab.x, ab.y, ab.width, ab.height);
      }
    }

    // background is in local space
    if (this._background !== null) {
      this.setBrush(context, this._background, true, false, nb, ab);
      if (this._background instanceof Brush && this._background.type === Brush.Radial) {
        context.beginPath();
        context.rect(ab.x, ab.y, ab.width, ab.height);
        context.fillContext(this._background);
      } else {
        context.fillRect(ab.x, ab.y, ab.width, ab.height);
      }
    }
    this.drawProtected(context, diagram);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {IContext} context
   * @param {Diagram} diagram
   */
  public drawProtected(context: IContext, diagram: Diagram): void {
    // do nothing
  }

  /**
   * @hidden @internal
   * Static function for drawing dashed lines on a context
   * @this {GraphObject}
   * @param {IContext} context
   * @param {number} x
   * @param {number} y
   * @param {number} x2
   * @param {number} y2
   * @param {Array.<number>} strokeDashArray
   * @param {number} offset
   */
  public static drawDash(context: IContext, x: number, y: number, x2: number, y2: number, strokeDashArray: Array<number>, offset: number): void {
    let dashLength = 0.001;
    const dashCount = strokeDashArray.length;
    context.moveTo(x, y);
    let dx = (x2 - x);
    const dy = (y2 - y);
    if (dx === 0) dx = 0.001;
    const slope = dy / dx;
    let distRemaining = Math.sqrt(dx * dx + dy * dy);
    let dashIndex = 0;
    let draw = true;
    dashLength = strokeDashArray[dashIndex % dashCount];

    let first = (offset === 0) ? false : true; // first line is different if there is an offset
    while (distRemaining >= 0.1) {
      if (first) {
        dashLength = strokeDashArray[dashIndex++ % dashCount];
        dashLength -= offset;
        while (dashLength < 0) {
          dashLength += strokeDashArray[dashIndex++ % dashCount];
          draw = !draw;
        }
        first = false;
      } else {
        dashLength = strokeDashArray[dashIndex++ % dashCount];
      }
      if (dashLength > distRemaining) dashLength = distRemaining;
      let xStep = Math.sqrt(dashLength * dashLength / (1 + slope * slope));
      if (dx < 0) xStep = -xStep;
      x += xStep;
      y += slope * xStep;
      if (draw) context.lineTo(x, y);
      else context.moveTo(x, y);
      distRemaining -= dashLength;
      draw = !draw;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {IContext} context
   * @param {Brush|string} brush
   * @param {boolean} fill
   * @param {boolean} parent coordinate system. Local by default.
   * @param {Rect} nb naturalBounds
   * @param {Rect} ab actualBounds
   */
  public setBrush(context: IContext, brush: BrushLike, fill: boolean, parent: boolean, nb: Rect, ab: Rect): void {
    if (brush !== null) {
      let w = 1.0;
      let h = 1.0;
      if (typeof brush === 'string') {
        if (fill) {
          context.fillStyle = brush;
        } else {
          context.strokeStyle = brush;
        }
      } else if (brush.type === Brush.Solid) {
        if (fill) {
          context.fillStyle = brush.color;
        } else {
          context.strokeStyle = brush.color;
        }
      } else { // its a Brush
        let grad: CanvasPattern | CanvasGradient | SGradient | string;
        w = nb.width;
        h = nb.height;
        if (parent) { // parent coordinates, like for areaBackground
          w = ab.width;
          h = ab.height;
        }
        const canvasContext = (context instanceof CanvasSurfaceContext);
        // We can only use the last constructed brush if the width and height are the same
        if (canvasContext && brush._constructed && (brush.type === Brush.Pattern ||
          (brush._constructedWidth === w && brush._constructedHeight === h))) {
          grad = brush._constructed;
        } else { // setup for both Linear and Radial
          let x0 = 0.0;
          let y0 = 0.0;
          let x1 = 0.0;
          let y1 = 0.0;
          let r0 = 0.0;
          let r1 = 0.0;
          let x = 0.0;
          let y = 0.0;
          if (parent) { // parent coordinates, like for areaBackground
            x = ab.x;
            y = ab.y;
          }

          x0 = brush.start.x * w + brush.start.offsetX;
          y0 = brush.start.y * h + brush.start.offsetY;
          x1 = brush.end.x * w + brush.end.offsetX;
          y1 = brush.end.y * h + brush.end.offsetY;

          // Since we are working in the panel's coordinates we must shift the values by
          // the actualBounds x/y
          x0 += x;
          x1 += x;
          y0 += y;
          y1 += y;

          // Specific setup
          if (brush.type === Brush.Linear) {
            grad = context.createLinearGradient(x0, y0, x1, y1);
          } else if (brush.type === Brush.Radial) {
            r1 = (isNaN(brush.endRadius)) ? Math.max(w, h) / 2 : brush.endRadius;
            if (isNaN(brush.startRadius)) {
              // if there's no startRadius, assume there's no endRadius either
              r0 = 0;
              r1 = Math.max(w, h) / 2;
            } else {
              r0 = brush.startRadius;
            }

            grad = context.createRadialGradient(x0, y0, r0, x1, y1, r1);
          } else if (brush.type === Brush.Pattern) {
            try {
              grad = context.createPattern(brush.pattern, 'repeat');
            } catch (ex) {
              // pattern stays unconstructed, can try again
              grad = null;
            }
          } else {
            Util.throwTypeError(brush.type, 'Brush type');
          }

          if (brush.type !== Brush.Pattern) {
            const stops = brush.colorStops;
            if (stops !== null) {
              const it = stops.iterator;
              while (it.next()) {
                (grad as any).addColorStop(it.key, it.value);
              }
            }
          }
          if (canvasContext) {
            brush._constructed = grad as any; // in SVG it may be a string, but this is canvas-only
            if (grad !== null) {
              brush._constructedWidth = w;
              brush._constructedHeight = h;
            }

            if (grad === null && brush.type === Brush.Pattern) {
              // Set width to -1 to act as a flag to "try again" only once.
              // This might happen because the image has not yet loaded, or has an error in loading
              if (brush._constructedWidth !== -1) {
                brush._constructedWidth = -1;
                const diagram = this.diagram;
                if (diagram !== null && brush._constructedWidth === -1) Util.setTimeout(function () { diagram.redraw(); }, 600);
              }
            }
          }
        } // end construction

        if (fill) {
          context.fillStyle = grad;
        } else {
          context.strokeStyle = grad;
        }
      }
    }
  } // end setBrush

  /**
   * This predicate is true if this object is an element, perhaps indirectly,
   * of the given panel.
   *
   * For example, if this GraphObject is inside a Part but is not itself the Part,
   * `obj.isContainedBy(obj.part)` should be true.
   * @export
   * @this {GraphObject}
   * @param {GraphObject} panel
   * @return {boolean} true if this object is contained by the given panel,
   * or if it is contained by another panel that is contained by the given panel,
   * to any depth; false if the argument is null or is not a {@link Panel}.
   */
  public isContainedBy(panel: GraphObject): boolean {
    if (panel instanceof Panel) return this._isContainedBy(this, panel);
    return false;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {GraphObject} obj
   * @param {Panel} panel
   * @return {boolean}
   */
  private _isContainedBy(obj: GraphObject, panel: Panel): boolean {
    if (obj === panel || panel === null) return false;
    let pan = obj.panel;
    while (pan !== null) {
      if (pan === panel) return true;
      pan = pan.panel;
    }
    return false;
  }

  /**
   * This predicate is true if this object is {@link #visible}
   * and each of its visual containing panels are also visible.
   * This ignores the actual location or appearance (except visibility)
   * of the panel that this object is part of, as well as ignoring all
   * properties of the {@link Layer} or {@link Diagram}.
   *
   * For {@link Part}s, you can call the {@link Part#isVisible} predicate to determine
   * if not only the Part is visible but also any containing {@link Group} or {@link Link} or {@link Layer}.
   * @export
   * @this {GraphObject}
   * @return {boolean}
   */
  public isVisibleObject(): boolean {
    if (!this.visible) return false;
    const p = this.panel;
    if (p !== null) return p.isVisibleObject();
    return true;
  }

  /**
   * This predicate is false if this object is inside any Panel that is not {@link Panel#isEnabled},
   * or if this is itself a disabled panel.
   * This ignores the {@link #visible} and {@link #pickable} properties.
   * @export
   * @this {GraphObject}
   * @return {boolean}
   * @see {@link #enabledChanged}
   * @see {@link Panel#isEnabled}
   * @since 1.7
   */
  public isEnabledObject(): boolean {
    let pan = (this instanceof Panel) ? this : this.panel;
    while (pan !== null && pan.isEnabled) {
      pan = pan.panel;
    }
    return (pan === null);  // couldn't find any containing Panel !isEnabled
  }

  /**
   * Gets or sets the function to execute when some containing Panel changes the value of {@link Panel#isEnabled}.
   * It is typically used to modify the appearance of the object.
   * This function must not change the value of any panel {@link Panel#isEnabled}.
   *
   * If this property value is a function, it is called with two arguments,
   * this {@link GraphObject} and the new value.
   * By default this property is null -- no function is called.
   * @see {@link #isEnabledObject}
   * @see {@link Panel#isEnabled}
   * @since 1.7
   */
  /** @export */ get enabledChanged(): ((thisObj: GraphObject, enabled: boolean) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._enabledChanged : null; }
  /** @export */ set enabledChanged(value: ((thisObj: GraphObject, enabled: boolean) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._enabledChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'enabledChanged');
      this._eventHandlers._enabledChanged = value;
      this.raiseChanged('enabledChanged', old, value);
    }
  }

  /**
   * @hidden @internal
   * Only called when transform or allTransforms is asked for and invalidTransform is true.
   * @this {GraphObject}
   */
  public _validateTransforms(): void {
    if (this.getInvalidTransform() === true) { // if invalid, make it valid
      const tt = this._transform;
      tt.reset();

      if (!this._actualBounds.isReal() || !this._measuredBounds.isReal()) {
        this.setInvalidTransform(false);
        return;
      }
      // measuredBounds offset is to move an object if its been rotated. For example, a square
      // rotated 45 degrees will need to be "pushed down" to fit where it once was
      tt.translate(this._actualBounds.x - this._measuredBounds.x, this._actualBounds.y - this._measuredBounds.y);

      if (this.scale !== 1 || this.angle !== 0) {
        const ns = this.naturalBounds;
        this._setTransformAngle(tt, ns.x, ns.y, ns.width, ns.height);
      }

      // the normal transform is now valid
      this.setInvalidTransform(false);
      this.setInvalidAllTransforms(true);
    }

    if (this.getInvalidAllTransforms() === true) { // if invalid, make it valid

      // if this is the topmost object:
      const panel = this.panel;
      if (panel === null) {
        this._allTransforms.set(this._transform);
        this._documentScale = this.scale;
        this.setInvalidAllTransforms(false);
        // else if object one above has a valid allTransforms:
      } else if (panel.allTransforms !== null) { // second condition?
        const t2 = this._allTransforms;
        t2.reset();
        if (panel.trueTransform()) {
          t2.multiply(panel._allTransforms);
        } else if (panel.panel !== null) {
          t2.multiply(panel.panel._allTransforms);
        }
        t2.multiply(this._transform);
        this._documentScale = this.scale * panel._documentScale;
        this.setInvalidAllTransforms(false);
      }

    }
  }

  /**
   * @hidden @internal
   * Shared code used by measureRect and validateTransforms.
   * @this {GraphObject}
   * @param {Transform} tt
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public _setTransformAngle(tt: Transform, x: number, y: number, w: number, h: number): void {
    if (this.scale !== 1) {
      tt.scale(this.scale);
    }
    if (this._angle !== 0) {
      let rspot = Spot.Center;
      if (this.isPart() && this.locationSpot.isSpot()) {
        rspot = (this).locationSpot;
      }

      const pt = Point.alloc();
      if (this instanceof Part && this.locationObject !== this) {
        const obj = this.locationObject;
        // if obj.transform were ready at this time we could use that
        // but it isn't because obj has no actualBounds
        // so instead we can localTransform + offset by measuredBounds
        const ns = obj.naturalBounds;
        pt.setSpot(ns.x, ns.y, ns.width, ns.height, rspot);
        obj._localTransform.transformPoint(pt);
        pt.offset(-obj.measuredBounds.x, -obj.measuredBounds.y);

        let parent = obj.panel;
        while (parent !== null && parent !== this) {
          parent._localTransform.transformPoint(pt);
          pt.offset(-parent.measuredBounds.x, -parent.measuredBounds.y);
          parent = parent.panel;
        }
      } else {
        pt.setSpot(x, y, w, h, rspot);
      }
      tt.rotate(this._angle, pt.x, pt.y);
      Point.free(pt);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean=} parent flag, meaning the parent invalidated a child. Stops the child from trying to invalidate the parent.
   */
  public invalidateMeasure(parent?: boolean): void {
    if (parent === undefined) parent = false;
    if (this.getInvalidMeasure() === true) return;
    this.setInvalidMeasure(true);
    this.setInvalidArrange(true);

    const panel = this.panel;
    if (panel !== null && !parent) {
      panel.invalidateMeasure();
    }
  }

  /**
   * @hidden @internal
   * Special invalidateMeasure for invalidating panels and their children only
   * This does not notify the diagram nor any parent panels
   * @this {GraphObject}
   */
  public invalidateMeasureChildren(): void {
    if (this.getInvalidMeasure() === true) return;
    this.setInvalidMeasure(true);
    this.setInvalidArrange(true);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean=} parent flag, meaning the parent invalidated a child. Stops the child from trying to invalidate the parent.
   */
  public invalidateArrange(parent?: boolean): void {
    if (this.getInvalidArrange() === false) {
      const panel = this.panel;
      if (panel !== null) {
        if (!parent) panel.invalidateMeasure(); // invalidate parents
      } else if (this.isPart()) {
        const diagram = this.diagram;
        if (diagram !== null) {
          diagram.invalidatePart(this);
          if (this instanceof Node) {
            this.invalidateConnectedLinks();
          }
          diagram.requestUpdate();
        }
      }
      this.setInvalidArrange(true);
    }
  }

  /**
   * @hidden @internal
   * Invalidates an Object's transform and then recurses through the visual tree to invalidate the allTransforms for each object in collection.
   * @this {GraphObject}
   */
  public _invalidateTransform(): void {
    if (this.getInvalidTransform() === false) { // add || invalidAllTransforms?
      this.setInvalidTransform(true);
      this.setInvalidAllTransforms(true);
    }
  }

  /**
   * @hidden @internal
   * Recurses through the visual tree to invalidate the allTransforms for each GraphObject.
   * @this {GraphObject}
   */
  public invalidateAllTransforms(): void {
    this.setInvalidAllTransforms(true);
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   */
  public invalidateDraw(): void {
    const part = this.part;
    if (part !== null) {
      part.invalidateDraw();
    }
  }

  /**
   * @hidden @internal
   * Computes the stretch for a GraphObject based on its stretch value and its containing panel's (default) stretch values
   * Auto panels and Table panels have different default rules for their children
   * @this {GraphObject}
   * @param {boolean} localcoords true if the stretch is in the local coordinate space, false if in the parent's coordiante space
   * @return {EnumValue}
   */
  public getEffectiveStretch(localcoords: boolean): EnumValue {
    const effectivestretch = this.stretch;
    const pan = this.panel;
    if (pan !== null && pan.type === Panel.Table) {
      return this.getEffectiveTableStretch(pan.getRowDefinition(this.row), pan.getColumnDefinition(this.column), localcoords);
    }
    // Auto panels have special stretch rules
    // The main object will be filled even if the stretch value is None,
    // but it will be filled to a different area
    if (pan !== null && pan.type === Panel.Auto) {
      if (pan.findMainElement() === this) return this.amendStretch(GraphObject.Fill, localcoords);
    }
    if (effectivestretch === GraphObject.Default) {
      if (pan !== null) {
        if (pan.type === Panel.Spot) {
          if (pan.findMainElement() === this) return this.amendStretch(GraphObject.Fill, localcoords);
        }

        const str = pan.defaultStretch;
        if (str === GraphObject.Default) return this.amendStretch(GraphObject.None, localcoords);
        return this.amendStretch(str, localcoords);
      }
      // Everything else is none by default
      return this.amendStretch(GraphObject.None, localcoords);
    }
    return this.amendStretch(effectivestretch, localcoords);
  }

  /**
   * @hidden @internal
   * Only ever called from Table's measure and arrange
   * @this {GraphObject}
   * @param {RowColumnDefinition} row
   * @param {RowColumnDefinition} col
   * @param {boolean} localcoords true if the stretch is in the local coordinate space, false if in the parent's coordiante space
   * @return {EnumValue}
   */
  public getEffectiveTableStretch(row: RowColumnDefinition, col: RowColumnDefinition, localcoords: boolean): EnumValue {
    const effectivestretch = this.stretch;
    if (effectivestretch !== GraphObject.Default) return this.amendStretch(effectivestretch, localcoords);
    // which directions are we stretching?
    let horizontal = null;
    let vertical = null;
    switch (row.stretch) {
      case GraphObject.Default:
      case GraphObject.Horizontal: break;
      case GraphObject.Vertical: vertical = true; break;
      case GraphObject.Fill: vertical = true; break;
    }
    switch (col.stretch) {
      case GraphObject.Default:
      case GraphObject.Vertical: break;
      case GraphObject.Horizontal: horizontal = true; break;
      case GraphObject.Fill: horizontal = true; break;
    }

    const str = this.panel.defaultStretch;
    if (horizontal === null) {
      horizontal = (str === GraphObject.Horizontal || str === GraphObject.Fill);
    }
    if (vertical === null) {
      vertical = (str === GraphObject.Vertical || str === GraphObject.Fill);
    }
    if (horizontal === true && vertical === true) return this.amendStretch(GraphObject.Fill, localcoords);
    if (horizontal === true) return this.amendStretch(GraphObject.Horizontal, localcoords);
    if (vertical === true) return this.amendStretch(GraphObject.Vertical, localcoords);
    // Everything else is none by default
    return this.amendStretch(GraphObject.None, localcoords);
  }

  /**
   * @hidden @internal
   * Stop stretching in some directions if there is a desired width or height that would block it
   * @this {GraphObject}
   * @param {EnumValue} effectivestretch
   * @param {boolean} self
   * @return {EnumValue}
   */
  private amendStretch(effectivestretch: EnumValue, self: boolean): EnumValue {
    if (self) return effectivestretch;
    if (effectivestretch === GraphObject.None) return GraphObject.None;
    const ds = this.desiredSize;
    if (ds.isReal()) return GraphObject.None;
    const realwidth = !isNaN(ds.width);
    const realheight = !isNaN(ds.height);
    const angle = this.angle;
    if (realwidth) {
      if (angle !== 90 && angle !== 270) {
        if (effectivestretch === GraphObject.Horizontal) return GraphObject.None;
        if (effectivestretch === GraphObject.Fill) return GraphObject.Vertical;
      } else {
        if (effectivestretch === GraphObject.Vertical) return GraphObject.None;
        if (effectivestretch === GraphObject.Fill) return GraphObject.Horizontal;
      }
    }

    if (realheight) {
      if (angle !== 90 && angle !== 270) {
        if (effectivestretch === GraphObject.Vertical) return GraphObject.None;
        if (effectivestretch === GraphObject.Fill) return GraphObject.Horizontal;
      } else {
        if (effectivestretch === GraphObject.Horizontal) return GraphObject.None;
        if (effectivestretch === GraphObject.Fill) return GraphObject.Vertical;
      }
    }
    return effectivestretch;
  }

  /**
   * Gets or sets the orientation of a GraphObject that is in a {@link Link} or {@link Panel.Graduated}.
   * This controls the automatic rotation of the object by the Link Panel or Graduated Panel.
   * The only accepted values are the {@link Link} "Orient..." values of Link
   * and the default value: {@link Link.None}.
   *
   * When the value is {@link Link.None}, the {@link #angle} of this object is unchanged as the link is routed.
   * Setting this to a value of {@link Link.OrientAlong} will cause routing to set the {@link #angle}
   * to be the angle of the segment that this object is on.
   * Other values compute the angle somewhat differently.
   * If the value is changed back to {@link Link.None}, the {@link #angle} of this object is set to zero.
   *
   * Note that when this property is not {@link Link.None}, this property takes precedence
   * over any setting or binding of the {@link #angle} property.
   * Changes to the angle caused by orientation might not result in Changed events,
   * and any original value for the angle may be lost.
   *
   * For examples of how to use this property, see <a href="../../intro/linkLabels.html">Link Labels</a>.
   * @see {@link #segmentFraction}
   * @see {@link #segmentIndex}
   * @see {@link #segmentOffset}
   */
  /** @export */ get segmentOrientation(): EnumValue { return this._segmentOrientation; }
  /** @export */ set segmentOrientation(value: EnumValue) {
    const old = this._segmentOrientation;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Link, GraphObject, 'segmentOrientation');
      this._segmentOrientation = value;
      this.invalidateMeasure();
      this.raiseChanged('segmentOrientation', old, value);
      if (value === Link.None) {
        this.angle = 0;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   */
  protected segmentOrientationAlong(): void {
    this.segmentOrientation = Link.OrientAlong;
  }

  /**
   * Gets or sets the segment index of a GraphObject that is in a {@link Link}.
   * Non-negative numbers count up from zero, which is the first segment, at the "from" end of the Link.
   * Negative numbers count segments from the "to" end of the Link, where -1 means the last segment
   * and -2 means the next-to-last segment.
   * The value should be an integer.
   *
   * If you do not set this property, the Link will choose a place that is approximately at the
   * mid-point of the link's route.
   *
   * For examples of how to use this property, see <a href="../../intro/linkLabels.html">Link Labels</a>.
   * @see {@link #segmentFraction}
   * @see {@link #segmentOffset}
   * @see {@link #segmentOrientation}
   */
  /** @export */ get segmentIndex(): number { return this._segmentIndex; }
  /** @export */ set segmentIndex(value: number) {
    if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'segmentIndex');
    value = Math.round(value);  // only integers
    const old = this._segmentIndex;
    if (old !== value) {
      this._segmentIndex = value;
      this.invalidateMeasure();
      this.raiseChanged('segmentIndex', old, value);
    }
  }

  /**
   * Gets or sets the fractional distance along a segment of a GraphObject that is in a {@link Link}.
   * The value should be between zero and one, where zero is at the point at the start of the segment,
   * and where one is at the point at the end of the segment.
   * The default value is zero.
   *
   * For examples of how to use this property, see <a href="../../intro/linkLabels.html">Link Labels</a>.
   * @see {@link #segmentIndex}
   * @see {@link #segmentOffset}
   * @see {@link #segmentOrientation}
   */
  /** @export */ get segmentFraction(): number { return this._segmentFraction; }
  /** @export */ set segmentFraction(value: number) {
    if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'segmentFraction');
    // Force it to be between 0 and 1. Maybe throw error instead?
    if (isNaN(value)) value = 0;
    else if (value < 0) value = 0;
    else if (value > 1) value = 1;
    const old = this._segmentFraction;
    if (old !== value) {
      this._segmentFraction = value; // only integers
      this.invalidateMeasure();
      this.raiseChanged('segmentFraction', old, value);
    }
  }

  /**
   * Gets or sets the offset of a GraphObject that is in a {@link Link} from a point on a segment
   * or in a {@link Panel.Graduated} from a point along the main element.
   * The X component of the Point indicates the distance along the route,
   * with positive values going further toward the "to" end of the link or panel.
   * The Y component of the Point indicates the distance away from the route,
   * with positive values towards the right as seen when facing further towards the "to" end of the link or panel.
   * The value defaults to the Point (0, 0).
   * You cannot modify the x or y of the value of this property --
   * if you want to change the segmentOffset you must set this property to a different Point.
   *
   * For labels that are near either end of a link, it may be convenient to set the segmentOffset
   * to Point(NaN, NaN). This causes the offset to be half the width and half the height of the label object.
   *
   * For examples of how to use this property, see <a href="../../intro/linkLabels.html">Link Labels</a>.
   * @see {@link #segmentFraction}
   * @see {@link #segmentIndex}
   * @see {@link #segmentOrientation}
   */
  /** @export */ get segmentOffset(): Point { return this._segmentOffset; }
  /** @export */ set segmentOffset(value: Point) {
    const old = this._segmentOffset;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Point, GraphObject, 'segmentOffset');
      value = value.copyFrozen();
      this._segmentOffset = value;
      this.invalidateMeasure();
      this.raiseChanged('segmentOffset', old, value);
    }
  }

  // ??? "Objects with an angle..."
  /**
   * Gets or sets the stretch of the GraphObject.
   * This controls whether the width and/or height of this object automatically adjusts to fill
   * the area allotted by the containing Panel.
   *
   * The only accepted values are listed as constant properties of GraphObject,
   * such as {@link GraphObject.None}, {@link GraphObject.Fill}, {@link GraphObject.Horizontal}, or {@link GraphObject.Vertical}.
   * The default value is {@link GraphObject.Default}, which allows the Panel to decide how to treat this object, depending on the type of Panel.
   *
   * Objects with an {@link #angle} that are stretched may look incorrect unless the angle is a multiple of 90.
   *
   * Stretch will have have different effects based upon the Panel containing this object. Elements of:
   * - Auto panels will not stretch, except the main element growing to fill the panel or being made uniform
   * - Horizontal panels will only stretch vertically
   * - Vertical panels will only stretch horizontally
   * - Spot panels will stretch to the size of the main element
   * - Table panels will stretch to the size of their cell, defined by their row and column, which is usually determined by other GraphObjects in that cell that are not stretching
   * - Grid panels, Link panels, and Graduated panels will not stretch
   * @see {@link Panel#defaultStretch}
   * @see {@link #desiredSize}
   * @see {@link #minSize}
   * @see {@link #maxSize}
   * @see {@link #measuredBounds}
   * @see {@link #actualBounds}
   * @see {@link #scale}
   * @see {@link Picture#imageStretch}
   */
  /** @export */ get stretch(): EnumValue { return this._stretch; }
  /** @export */ set stretch(value: EnumValue) {
    const old = this._stretch;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, GraphObject, GraphObject, 'stretch');
      this._stretch = value;
      this.invalidateMeasure();
      this.raiseChanged('stretch', old, value);
    }
  }

  /**
   * Gets or sets the name for this object.
   * The default value is the empty string.
   * The name should be unique within a {@link Panel}, although if it isn't,
   * it reduces the usefulness of methods such as {@link Panel#findObject}.
   *
   * You must not modify the name of a GraphObject once it is in the visual tree of a Part.
   *
   * This is frequently needed to identify a particular GraphObject in the visual tree of a Part,
   * for example as the value of the {@link Part#locationObjectName} or
   * {@link Part#selectionObjectName} properties.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) {
    const old = this._name;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', GraphObject, 'name');
      this._name = value;
      if (this.part !== null) this.part.clearNameMap();
      this.raiseChanged('name', old, value);
    }
  }

  /**
   * Gets or sets the multiplicative opacity for this GraphObject and (if a Panel) all elements.
   * The value must be between 0.0 (fully transparent) and 1.0 (no additional transparency).
   *
   * Unlike {@link #visible}, Opacity only affects drawing, it does not cause objects to be resized or remeasured.
   * Opacity settings do not change the shape of the object or exclude it from object-picking
   * (does not change whether any objects are found by the "find..." methods).
   *
   * This value is multiplicative with any existing transparency,
   * for instance from {@link Layer#opacity} or a GraphObject's opacity higher in the visual tree,
   * or from a {@link Brush} or image transparency.
   * The default value is 1.
   * @see {@link #visible}
   * @see {@link #pickable}
   * @see {@link Layer#opacity}
   * @see {@link Diagram#opacity}
   * @since 1.4
   */
  /** @export */ get opacity(): number { return this._opacity; }
  /** @export */ set opacity(value: number) {
    const old = this._opacity;
    if (old !== value) {
      Util.checkPrimitive(value, 'number', GraphObject, 'opacity');
      if (value < 0 || value > 1) Util.throwRangeError(value, '0 <= value <= 1', GraphObject, 'opacity');
      this._opacity = value;
      this.raiseChanged('opacity', old, value);
      // and make sure everything is re-drawn
      const diagram = this.diagram;
      const part = this.part;
      if (diagram !== null && part !== null) {
        diagram.invalidateDraw(part.getDrawnSize(part.actualBounds));
      }
    }
  }

  /**
   * Gets or sets whether a GraphObject is visible.
   * The default value is true.
   * A not visible object takes no space in the Panel that it is in.
   * Toggling visibility may cause elements in the visual tree to re-measure and re-arrange.
   * Making a Panel not visible causes all of its elements not to be seen or receive input events.
   * Changing a Panel to become visible causes all of its elements to be seen and be active,
   * unless those elements are themselves not visible.
   *
   * This object does not get any mouse/touch events if it is not {@link #visible}
   * or if it is not {@link #pickable}.
   *
   * One can have a visible {@link Shape} that is not drawn by setting its {@link Shape#fill}
   * and {@link Shape#stroke} to null or to `"transparent"`.
   * Similarly, one can set {@link TextBlock#stroke} to null or to `"transparent"`.
   * It is also possible make a GraphObjects transparent by setting {@link GraphObject#opacity} to 0.
   * Finally, one can make a whole {@link Layer}-full of Parts invisible by setting {@link Layer#visible} to false.
   *
   * Use the {@link #isVisibleObject} predicate to see if this GraphObject is visible and is inside a {@link Panel}
   * that is {@link #isVisibleObject}, and so forth up the chain of panels until reaching the {@link Part}.
   *
   * For {@link Part}s, you can call the {@link Part#isVisible} predicate to determine
   * if not only the Part is visible but also any containing {@link Group} or {@link Link},
   * and whether the {@link Layer} it is in is visible.
   * @see {@link #opacity}
   * @see {@link #pickable}
   * @see {@link Layer#visible}
   */
  /** @export */ get visible(): boolean { return (this._flags & 0x00000001) !== 0; }
  /** @export */ set visible(value: boolean) {
    const old = (this._flags & 0x00000001) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'visible');
      this._flags = this._flags ^ 0x00000001;
      this.raiseChanged('visible', old, value);
      const panel = this.panel;
      if (panel !== null) {
        panel.invalidateMeasure();
      } else {
        if (this.isPart()) {
          this.madeVisible(value);
        }
      }
      this.invalidateDraw();
      this._invalidateLinkRoutesForContainedPorts();
    }
  }

  /**
   * Gets or sets whether or not this GraphObject can be chosen by visual "find" or "hit-test" methods such
   * as {@link Diagram#findObjectAt}.
   *
   * This object does not get any mouse/touch events if it is not {@link #visible}
   * or if it is not {@link #pickable}.
   *
   * The default value is true -- mouse events on this object will be noticed.
   * If this value is false and this object is a {@link Panel}, not only is this Panel not "hittable",
   * but all of the elements inside the Panel will be ignored.
   * @see {@link #visible}
   * @see {@link #opacity}
   * @see {@link Layer#pickable}
   * @see {@link Panel#isEnabled}
   * @since 1.2
   */
  /** @export */ get pickable(): boolean { return (this._flags & 0x00000002) !== 0; }
  /** @export */ set pickable(value: boolean) {
    const old = (this._flags & 0x00000002) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'pickable');
      this._flags = this._flags ^ 0x00000002;
      this.raiseChanged('pickable', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw duplicate Links from this port.
   * This property is used by {@link LinkingBaseTool#isValidLink}.
   * The default value is false.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #fromLinkable}
   * @see {@link #fromLinkableSelfNode}
   * @see {@link #toLinkableDuplicates}
   */
  /** @export */ get fromLinkableDuplicates(): boolean { return (this._flags & 0x00000004) !== 0; }
  /** @export */ set fromLinkableDuplicates(value: boolean) {
    const old = (this._flags & 0x00000004) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'fromLinkableDuplicates');
      this._flags = this._flags ^ 0x00000004;
      this.raiseChanged('fromLinkableDuplicates', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw Links that connect from this port's Node.
   * This property is used by {@link LinkingBaseTool#isValidLink}.
   * The default value is false.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #fromLinkable}
   * @see {@link #fromLinkableDuplicates}
   * @see {@link #toLinkableSelfNode}
   */
  /** @export */ get fromLinkableSelfNode(): boolean { return (this._flags & 0x00000008) !== 0; }
  /** @export */ set fromLinkableSelfNode(value: boolean) {
    const old = (this._flags & 0x00000008) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'fromLinkableSelfNode');
      this._flags = this._flags ^ 0x00000008;
      this.raiseChanged('fromLinkableSelfNode', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw duplicate Links to this port.
   * This property is used by {@link LinkingBaseTool#isValidLink}.
   * The default value is false.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #toLinkable}
   * @see {@link #toLinkableSelfNode}
   * @see {@link #fromLinkableDuplicates}
   */
  /** @export */ get toLinkableDuplicates(): boolean { return (this._flags & 0x00000010) !== 0; }
  /** @export */ set toLinkableDuplicates(value: boolean) {
    const old = (this._flags & 0x00000010) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'toLinkableDuplicates');
      this._flags = this._flags ^ 0x00000010;
      this.raiseChanged('toLinkableDuplicates', old, value);
    }
  }

  /**
   * Gets or sets whether the user may draw Links that connect to this port's Node.
   * This property is used by {@link LinkingBaseTool#isValidLink}.
   * The default value is false.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #toLinkable}
   * @see {@link #toLinkableDuplicates}
   * @see {@link #fromLinkableSelfNode}
   */
  /** @export */ get toLinkableSelfNode(): boolean { return (this._flags & 0x00000020) !== 0; }
  /** @export */ set toLinkableSelfNode(value: boolean) {
    const old = (this._flags & 0x00000020) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'toLinkableSelfNode');
      this._flags = this._flags ^ 0x00000020;
      this.raiseChanged('toLinkableSelfNode', old, value);
    }
  }

  /**
   * Gets or sets whether a GraphObject is the "main" object for some types of {@link Panel}.
   * Panels that use a "main" object include {@link Panel.Auto}, {@link Panel.Spot}, and {@link Panel.Link}.
   *
   * Panels that use a "main" object will use the first object that has this property set to true,
   * or else just the first object, if none have the property set.
   *
   * Do not modify this property once this object is an element of a panel.
   */
  /** @export */ get isPanelMain(): boolean { return (this._flags & 0x00000040) !== 0; }
  /** @export */ set isPanelMain(value: boolean) {
    const old = (this._flags & 0x00000040) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'isPanelMain');
      this._flags = this._flags ^ 0x00000040;
      this.invalidateMeasure();
      this.raiseChanged('isPanelMain', old, value);
    }
  }

  /**
   * This property determines whether or not this GraphObject's events occur
   * before all other events, including selection. This enables the {@link #actionDown},
   * {@link #actionMove}, {@link #actionUp}, and {@link #actionCancel} events,
   * which are all handled by the {@link ActionTool}.
   *
   * This object does not get any mouse/touch events if it is not {@link #visible}
   * or if it is not {@link #pickable}.
   *
   * This property is infrequently used -- typically only when implementing objects
   * that act as buttons or knobs or sliders.
   * The default value is false.
   * @see {@link #actionDown}
   * @see {@link #actionMove}
   * @see {@link #actionUp}
   * @see {@link #actionCancel}
   * @see {@link #pickable}
   * @see {@link Panel#isEnabled}
   */
  /** @export */ get isActionable(): boolean { return (this._flags & 0x00000080) !== 0; }
  /** @export */ set isActionable(value: boolean) {
    const old = (this._flags & 0x00000080) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', GraphObject, 'isActionable');
      this._flags = this._flags ^ 0x00000080;
      this.raiseChanged('isActionable', old, value);
    }
  }

  /**
   * Gets or sets the areaBackground Brush of this GraphObject.
   * The areaBackground fills the rectangle described by this GraphObject's containing panel's coordinates.
   * If the object is rotated, the area background will expand to fill the entire measured bounds of the object,
   * without rotating the brush.
   *
   * The value may be either a {@link Brush} object or a string that is a CSS color.
   * The default value is null -- no area background is drawn.
   * More information about the syntax of CSS color strings is available at:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color">CSS colors (mozilla.org)</a>.
   * @see {@link #background}
   * @see {@link Shape#fill}
   */
  /** @export */ get areaBackground(): BrushLike { return this._areaBackground; }
  /** @export */ set areaBackground(value: BrushLike) {
    const old = this._areaBackground;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'GraphObject.areaBackground');
      if (value instanceof Brush) value.freeze();
      this._areaBackground = value;
      this.invalidateDraw();
      this.raiseChanged('areaBackground', old, value);
    }
  }

  /**
   * Gets or sets the background Brush of this GraphObject,
   * filling the rectangle of this object's local coordinate space.
   * If the object is rotated, the background will rotate with it.
   *
   * The value may be either a {@link Brush} object or a string that is a CSS color.
   * The default value is null -- no background is drawn.
   * More information about the syntax of CSS color strings is available at:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color">CSS colors (mozilla.org)</a>.
   * @see {@link #areaBackground}
   * @see {@link Shape#fill}
   */
  /** @export */ get background(): BrushLike { return this._background; }
  /** @export */ set background(value: BrushLike) {
    const old = this._background;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'GraphObject.background');
      if (value instanceof Brush) value.freeze();
      this._background = value;
      this.invalidateDraw();
      this.raiseChanged('background', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getAlreadyShadowed(): boolean {
    return (this._flags & 0x00000200) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setAlreadyShadowed(value: boolean): void {
    if (value) {
      this._flags |= 0x00000200;
    } else {
      this._flags &= ~0x00000200;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getBindsTwoWay(): boolean {
    return (this._flags & 0x00000400) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setBindsTwoWay(value: boolean): void {
    if (value) {
      this._flags |= 0x00000400;
    } else {
      this._flags &= ~0x00000400;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getInvalidTransform(): boolean {
    return (this._flags & 0x00000800) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setInvalidTransform(value: boolean): void {
    if (value) {
      this._flags |= 0x00000800;
    } else {
      this._flags &= ~0x00000800;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getInvalidAllTransforms(): boolean {
    return (this._flags & 0x00001000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setInvalidAllTransforms(value: boolean): void {
    if (value) {
      this._flags |= 0x00001000;
    } else {
      this._flags &= ~0x00001000;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getInvalidMeasure(): boolean {
    return (this._flags & 0x00002000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setInvalidMeasure(value: boolean): void {
    if (value) {
      this._flags |= 0x00002000;
    } else {
      this._flags &= ~0x00002000;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @return {boolean}
   */
  public getInvalidArrange(): boolean {
    return (this._flags & 0x00004000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {boolean} value
   */
  public setInvalidArrange(value: boolean): void {
    if (value) {
      this._flags |= 0x00004000;
    } else {
      this._flags &= ~0x00004000;
    }
  }

  /**
   * This read-only property returns the {@link Part} containing this object, if any.
   * The Part will be the root GraphObject in this GraphObject's visual tree.
   *
   * It is common to refer to the containing Part of a GraphObject
   * in order to refer to the {@link Panel#data} to which it is bound.
   *
   * This property is not settable.
   * If you want this GraphObject to belong to a Part, you will need to add it to a Part,
   * or else add it to some visual tree structure that is added to a Part using {@link Panel#add}.
   *
   * Note that for objects such as buttons that are in {@link Adornment}s such as tooltips or context menus,
   * this property will return that Adornment, not the Node or Link that is adorned.
   *
   * If you want to find a {@link Group} that contains a Part, use the {@link Part#containingGroup} property:
   * `someObj.part.containingGroup`
   */
  /** @export */ get part(): Part | null {
    if (this.isPart()) return this;
    if (this._part !== null) return this._part;
    let p = this as GraphObject;
    p = p.panel;
    while (p) {
      if (p instanceof Part) {
        this._part = p;
        return p;
      }
      p = p.panel;
    }
    return null;
  }

  /**
   * @hidden @internal
   *
   * The root SVG element representing this object
   */
  /** @export */ get svg(): SVGElement { return this._svg; }
  /** @export */ set svg(value: SVGElement) {
    this._svg = value;
  }

  /**
   * This read-only property returns the GraphObject's containing {@link Panel}, or null if this object is not in a Panel.
   *
   * Although {@link Part} inherits from this class, a Part will never belong to a Panel,
   * so this property will always be null for every {@link Node} or {@link Link}.
   *
   * This property is not settable.
   * Instead, call {@link Panel#add} in order to put a GraphObject in a Panel.
   */
  /** @export */ get panel(): Panel | null { return this._panel; }

  /**
   * @hidden @internal
   * Setter for read-only property panel
   * @this {GraphObject}
   * @param {Panel} value
   */
  public setPanel(value: Panel): void {
    this._panel = value;
  }

  /**
   * This read-only property returns the GraphObject's containing {@link Layer}, if there is any.
   * A plain GraphObject cannot belong directly to a Layer -- only a {@link Part} can belong directly to a Layer.
   *
   * This property is not settable.
   * Normally one changes which Layer that a GraphObject is in by setting {@link Part#layerName}.
   * Adding a Part to a Diagram will automatically add that Part to a Layer in that Diagram based on the layerName.
   */
  /** @export */ get layer(): Layer | null {
    const gp = this.part;
    if (gp !== null) return gp.layer;
    return null;
  }

  /**
   * This read-only property returns the {@link Diagram} that this GraphObject is in, if it is.
   *
   * This property is not settable.
   * Although you cannot add any plain GraphObject to a Diagram, you can call {@link Diagram#add}
   * to add a {@link Part} to a Diagram.
   */
  /** @export */ get diagram(): Diagram | null {
    const gp = this.part;
    if (gp !== null) return gp.diagram;
    return null;
  }

  // Measurement and arrangement:

  /**
   * Gets or sets the position of this GraphObject in container coordinates
   * (either a {@link Panel} or the document).
   * Value must be of type Point.
   * You cannot modify the x or y of the value of this property --
   * if you want to change the position you must set this property to a different Point.
   * Default is Point(NaN, NaN).
   *
   * For {@link Part}s, see also {@link Part#location}.
   */
  /** @export */ get position(): Point { return this._position; }
  /** @export */ set position(value: Point) {
    if (Debug) Util.checkClass(value, Point, GraphObject, 'position');
    const nx = value.x;
    const ny = value.y;
    const old = this._position;
    const ox = old.x;
    const oy = old.y;
    // if old and new are not equal and if both old and new are not NaN
    if ((ox !== nx && (!isNaN(ox) || !isNaN(nx))) ||
      (oy !== ny && (!isNaN(oy) || !isNaN(ny)))) {
      value = value.copyFrozen();
      // NOTE: Sometimes this setter can be bypassed in:
      //       * Part.determinePosLoc (only Parts, during arranging)
      //       * GraphObject.setPosInplace (during animation)
      if (this.setPos(value, old)) this.raiseChanged('position', old, value.copy());
    } else {
      this.setPositionFlag();
    }
  }

  /**
   * @hidden @internal
   * Only used in Parts
   * @this {GraphObject}
   */
  protected setPositionFlag(): void {
    // no-op
  }

  /**
   * @hidden @internal
   * This is not called during animation, instead, it calls setPosInplace
   * @this {GraphObject}
   * @param {Point} _val
   * @param {Point} _old
   * @return {boolean} true if successful, false if a no-op (only true calls raisechanged)
   */
  protected setPos(_val: Point, _old: Point): boolean {
    this._position = _val;
    this.invalidateArrange();
    return true;
  }

  /**
   * @hidden @internal
   * Undocumented. Used during animation only
   * @this {GraphObject}
   * @param {number} x
   * @param {number} y
   */
  public setPosInplace(x: number, y: number, setloc: boolean /* unused */): void {
    if (this._position.x === x && this._position.y === y) return;
    this._position.assignTo(x, y);
    this._invalidateTransform(); // required because arrange will not happen
  }

  /**
   * This read-only property returns the bounds of this GraphObject in container coordinates. This means that
   * the actualBounds are in the coordinate space of the GraphObject's {@link Panel},
   * unless this is a {@link Part}, in which case they are in
   * the {@link Diagram}'s coordinate system.
   *
   * You must not modify any of the properties of the {@link Rect} that is the value of this property.
   *
   * If this GraphObject is a Part,
   * then the x and y values of the actualBounds are identical to that Part's {@link #position},
   * and the width and height values of the actualBounds represent the rectangular space occupied
   * by the Part in {@link Diagram#documentBounds} coordinates.
   *
   * If this GraphObject is not a top-level object (not a {@link Part}), then the actualBounds
   * x and y values represent that GraphObject's position within its Panel. In a Panel of type {@link Panel.Position}
   * this is identical to the GraphObject's {@link #position}, but in other cases it is dependent on
   * the unique workings of each Panel type. The actualBounds width and height
   * of a GraphObject are the final size after the {@link #scale} and {@link #angle} are applied.
   *
   * It is possible for a GraphObject (be it an GraphObject or a Panel containing several more GraphObjects)
   * to have no containing Part, in which case these GraphObjects cannot possibly be in a Diagram.
   * These GraphObjects are unlikely to have real-number values for their actualBounds, as they may
   * never have had the chance to be measured and arranged.
   *
   * As with all read-only properties, using this property as a binding source is unlikely to be useful.
   * @see {@link #measuredBounds}
   * @see {@link #desiredSize}
   */
  /** @export */ get actualBounds(): Rect { return this._actualBounds; }

  /**
   * Gets or sets the scale transform of this GraphObject.
   * Value must be a number; larger values will make this object appear bigger.
   * Default is 1.
   * @see {@link #angle}
   * @see {@link #stretch}
   */
  /** @export */ get scale(): number { return this._scale; }
  /** @export */ set scale(value: number) {
    const old = this._scale;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, GraphObject, 'scale');
      if (value <= 0) Util.throwError('GraphObject.scale for ' + this + ' must be greater than zero, not: ' + value);
      this._scale = value;
      this.invalidateMeasure();
      this.raiseChanged('scale', old, value);
    }
  }

  /**
   * Gets or sets the angle transform, in degrees, of this GraphObject.
   * Value must be a number. If the value is not between (0 &lt;= value &lt; 360),
   * it will be normalized to be in that range.
   * Zero is along the positive X-axis (rightwards); 90 is along the positive Y-axis (downwards).
   * Default is 0.
   * @see {@link #scale}
   * @see {@link #stretch}
   */
  /** @export */ get angle(): number { return this._angle; }
  /** @export */ set angle(value: number) {
    const old = this._angle;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, GraphObject, 'angle');
      value = value % 360;  // canonicalize the angle 0 &lt;= value &lt; 360
      if (value < 0) value += 360;
      if (old === value) return;
      this._angle = value;
      this._invalidateLinkRoutesForContainedPorts();
      this.invalidateMeasure();
      this.raiseChanged('angle', old, value);
    }
  }

  /**
   * Gets or sets the desired size of this GraphObject in local coordinates.
   * Value must be of type {@link Size}.
   * Default is Size(NaN, NaN).
   * You cannot modify the width or height of the value of this property --
   * if you want to change the desiredSize you must set this property to a different Size.
   *
   * Getting or setting {@link #width} or {@link #height} is equivalent to getting or setting the
   * width or height of this property.
   *
   * The size does not include any transformation due to {@link #scale} or {@link #angle},
   * nor any pen thickness due to {@link Shape#strokeWidth} if this is a {@link Shape}.
   * If there is a containing {@link Panel} the Panel will determine the actual size.
   * If the desiredSize is greater than the allowed size that the GraphObject's Panel determines,
   * then the GraphObject may be visually clipped. If the desiredSize does not meet the constraints
   * of {@link #minSize} and {@link #maxSize}, the GraphObject will be resized to meet them.
   * @see {@link #minSize}
   * @see {@link #maxSize}
   * @see {@link #naturalBounds}
   * @see {@link #measuredBounds}
   * @see {@link #actualBounds}
   */
  /** @export */ get desiredSize(): Size { return this._desiredSize; }
  /** @export */ set desiredSize(value: Size) {
    if (Debug) Util.checkClass(value, Size, GraphObject, 'desiredSize');
    const nw = value.width;
    const nh = value.height;
    const old = this._desiredSize;
    const ow = old.width;
    const oh = old.height;
    // if old and new are not equal and if both old and new are not NaN
    if ((ow !== nw && (!isNaN(ow) || !isNaN(nw))) ||
      (oh !== nh && (!isNaN(oh) || !isNaN(nh)))) {
      value = value.copyFrozen();
      this._desiredSize = value;
      this.invalidateMeasure();
      // let Geometry be reset
      if (this instanceof Shape) {
        this.invalidateGeometry();
      }
      this.raiseChanged('desiredSize', old, value);
      // changing the desiredSize also changes the width and the height
      if (this.getBindsTwoWay()) {
        const part = this.part;
        if (part !== null) {
          this.updateSourceBindings(part, 'width');
          this.updateSourceBindings(part, 'height');
        }
      }
    }
  }

  /**
   * Gets or sets the desired width of this GraphObject in local coordinates.
   * This just gets or sets the width component of the {@link #desiredSize}.
   * Default is NaN.
   *
   * Size can also be constrained by setting {@link #minSize} and {@link #maxSize}.
   *
   * The width does not include any transformation due to {@link #scale} or {@link #angle},
   * nor any pen thickness due to {@link Shape#strokeWidth} if this is a {@link Shape}.
   * If there is a containing {@link Panel} the Panel will determine the actual size.
   */
  /** @export */ get width(): number { return this._desiredSize.width; }
  /** @export */ set width(value: number) {
    const old = this._desiredSize.width;
    if (old !== value && (!isNaN(old) || !isNaN(value))) {  // no-op if both old and value are NaN
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'width');
      const oldsz = this._desiredSize;
      const newsz = new Size(value, this._desiredSize.height).freeze();
      this._desiredSize = newsz;
      this.invalidateMeasure();
      // let Geometry be reset
      if (this instanceof Shape) {
        this.invalidateGeometry();
      }
      this.raiseChanged('desiredSize', oldsz, newsz);
      // changing the desiredSize also changes the width
      if (this.getBindsTwoWay()) {
        const part = this.part;
        if (part !== null) {
          this.updateSourceBindings(part, 'width');
        }
      }
    }
  }

  /**
   * Gets or sets the desired height of this GraphObject in local coordinates.
   * This just gets or sets the height component of the {@link #desiredSize}.
   * Default is NaN.
   *
   * Size can also be constrained by setting {@link #minSize} and {@link #maxSize}.
   *
   * The height does not include any transformation due to {@link #scale} or {@link #angle},
   * nor any pen thickness due to {@link Shape#strokeWidth} if this is a {@link Shape}.
   * If there is a containing {@link Panel} the Panel will determine the actual size.
   */
  /** @export */ get height(): number { return this._desiredSize.height; }
  /** @export */ set height(value: number) {
    const old = this._desiredSize.height;
    if (old !== value && (!isNaN(old) || !isNaN(value))) {  // no-op if both old and value are NaN
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'height');
      const oldsz = this._desiredSize;
      const newsz = new Size(this._desiredSize.width, value).freeze();
      this._desiredSize = newsz;
      this.invalidateMeasure();
      // let Geometry be reset
      if (this instanceof Shape) {
        this.invalidateGeometry();
      }
      this.raiseChanged('desiredSize', oldsz, newsz);
      // changing the desiredSize also changes the height
      if (this.getBindsTwoWay()) {
        const part = this.part;
        if (part !== null) {
          this.updateSourceBindings(part, 'height');
        }
      }
    }
  }

  /**
   * Gets or sets the minimum size of this GraphObject in container coordinates
   * (either a {@link Panel} or the document).
   * Any new value must be of type Size; NaN values are treated as 0.
   *
   * You cannot modify the width or height of the value of this property --
   * if you want to change the minSize you must set this property to a different Size.
   * The default value is zero by zero.
   * A containing Panel will determine the actual size of this object.
   * @see {@link #maxSize}
   * @see {@link #desiredSize}
   */
  /** @export */ get minSize(): Size { return this._minSize; }
  /** @export */ set minSize(value: Size) {
    const old = this._minSize;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Size, GraphObject, 'minSize');
      value = value.copy();
      if (isNaN(value.width)) value.width = 0;
      if (isNaN(value.height)) value.height = 0;
      value.freeze();
      this._minSize = value;
      this.invalidateMeasure();
      this.raiseChanged('minSize', old, value);
    }
  }

  /**
   * Gets or sets the maximum size of this GraphObject in container coordinates
   * (either a {@link Panel} or the document).
   * Any new value must be of type Size; NaN values are treated as Infinity. If you want no maximum width or height, use NaN or Infinity.
   *
   * You cannot modify the width or height of the value of this property --
   * if you want to change the maxSize you must set this property to a different Size.
   * The default value is Infinity by Infinity.
   * A containing Panel will determine the actual size of this object.
   * @see {@link #minSize}
   * @see {@link #desiredSize}
   */
  /** @export */ get maxSize(): Size { return this._maxSize; }
  /** @export */ set maxSize(value: Size) {
    const old = this._maxSize;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Size, GraphObject, 'maxSize');
      value = value.copy();
      if (isNaN(value.width)) value.width = Infinity;
      if (isNaN(value.height)) value.height = Infinity;
      value.freeze();
      this._maxSize = value;
      this.invalidateMeasure();
      this.raiseChanged('maxSize', old, value);
    }
  }

  /**
   * This read-only property returns the measuredBounds of the GraphObject in container coordinates
   * (either a {@link Panel} or the document).
   * This describes the transformed bounds with margins excluded.
   *
   * You must not modify any of the properties of the {@link Rect} that is the value of this property.
   *
   * As with all read-only properties, using this property as a binding source is unlikely to be useful.
   * @see {@link #naturalBounds}
   * @see {@link #desiredSize}
   * @see {@link #actualBounds}
   */
  /** @export */ get measuredBounds(): Rect { return this._measuredBounds; }

  /**
   * This read-only property returns the natural bounding rectangle of this GraphObject in local coordinates,
   * before any transformation by {@link #scale} or {@link #angle},
   * and before any resizing due to {@link #minSize} or {@link #maxSize} or {@link #stretch}.
   * Defaults to unknown (NaN,NaN).
   *
   * You must not modify any of the properties of the {@link Rect} that is the value of this property.
   *
   * The value can only be changed by changing properties of the particular GraphObject,
   * such as {@link GraphObject#desiredSize}, {@link Shape#geometry}, or {@link TextBlock#font}.
   *
   * As with all read-only properties, using this property as a binding source is unlikely to be useful.
   * @see {@link #desiredSize}
   * @see {@link #measuredBounds}
   * @see {@link #actualBounds}
   */
  /** @export */ get naturalBounds(): Rect { return this._naturalBounds; }

  /**
   * Gets or sets the size of empty area around this GraphObject, as a {@link Margin},
   * in the containing {@link Panel} coordinates.
   *
   * Negative values are permitted but may cause overlaps with adjacent
   * objects in a {@link Panel}.
   * You cannot modify the top or left or right or bottom of the value of this property --
   * if you want to change the margin you must set this property to a different Margin.
   * Default margin is Margin(0,0,0,0).
   *
   * The property setter accepts a number instead of a Margin object: providing a
   * number N will result in using a Margin(N, N, N, N).
   * The property getter will always return a Margin.
   * @see {@link #desiredSize}
   * @see {@link #measuredBounds}
   * @see {@link #actualBounds}
   * @see {@link Panel#padding}
   */
  /** @export */ get margin(): MarginLike { return this._margin; }
  /** @export */ set margin(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      if (Debug) Util.checkClass(value, Margin, GraphObject, 'margin');
    }
    const old = this._margin;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._margin = value;
      this.invalidateMeasure();
      this.raiseChanged('margin', old, value);
    }
  }

  /**
   * @hidden @internal
   * This read-only property returns the GraphObject's transformation matrix
   */
  /* noexport */get transform(): Transform {
    if (this.getInvalidTransform() === true) { this._validateTransforms(); }
    return this._transform;
  }

  /**
   * @hidden @internal
   * This read-only property returns a transformation matrix with respect to document coordinates describing
   * all transforms operated on the object's coordinates, including the
   * GraphObject's own transform. Modified by changes to the position, scale,
   * and angle properties.
   */
  /* noexport */get allTransforms(): Transform {
    if (this.getInvalidAllTransforms() === true) { this._validateTransforms(); }
    return this._allTransforms;
  }

  // maybe don't need
  // *
  //  * hidden internal
  //  * Shared
  //  *
  // /* no export */get animationTriggers(): Map<string, AnimationTrigger> | null { return this._animationTriggers; }
  // /* no export *//** @export */ set animationTriggers(value: Map<string, AnimationTrigger> | null) {
  //   const old = this._animationTriggers;
  //   if (old !== value) {
  //     this._animationTriggers = value;
  //   }
  // }


  /**********************************************************
  Panel-specific properties:
  ***********************************************************/

  /**
   * Gets or sets the alignment {@link Spot} of this GraphObject used in {@link Panel} layouts,
   * to determine where in the area allocated by the panel this object should be placed.
   *
   * The default value is {@link Spot.Default}, which lets the Panel determine the Spot using
   * {@link Panel#defaultAlignment}. If that property is also {@link Spot.Default},
   * then the alignment spot will be different depending on the Panel type.
   *
   * The {@link #alignmentFocus} is often used along with this property to specify
   * where this object should be positioned in a Panel.
   *
   * A {@link Spot.Default} is equivalent to Spot.Center in Spot, Auto, Horizontal, and Vertical panels.
   * For examples of alignments in different panels, see the <a href="../../intro/panels.html">Introduction page on Panels</a>.
   *
   * @see {@link #alignmentFocus}
   * @see {@link Panel#defaultAlignment}
   */
  /** @export */ get alignment(): Spot { return this._alignment; }
  /** @export */ set alignment(value: Spot) {
    const old = this._alignment;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, GraphObject, 'alignment');
      if (value.isNoSpot() && !value.isDefault()) {
        Util.throwError('GraphObject.alignment for ' + this + ' must be a real Spot or Spot.Default, not: ' + value);
      }
      value = value.copyFrozen();
      this._alignment = value;
      this.invalidateArrange();
      this.raiseChanged('alignment', old, value);
    }
  }

  /**
   * Gets or sets the column of this GraphObject if it is in a Table {@link Panel}.
   * The value must be a small non-negative integer. The default is 0.
   */
  /** @export */ get column(): number { return this._column; }
  /** @export */ set column(value: number) {
    if (Debug) Util.checkRealNumber(value, GraphObject, 'column');
    value = Math.round(value);
    const old = this._column;
    if (old !== value) {
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'column');
      this._column = value;
      this.invalidateMeasure();
      this.raiseChanged('column', old, value);
    }
  }

  /**
   * Gets or sets the number of columns spanned by this GraphObject if it is in a Table {@link Panel}.
   * The value must be a small positive integer. The default is 1.
   */
  /** @export */ get columnSpan(): number { return this._columnSpan; }
  /** @export */ set columnSpan(value: number) {
    if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'columnSpan');
    value = Math.round(value);
    const old = this._columnSpan;
    if (old !== value) {
      if (value < 1) Util.throwRangeError(value, '>= 1', GraphObject, 'columnSpan');
      this._columnSpan = value;
      this.invalidateMeasure();
      this.raiseChanged('columnSpan', old, value);
    }
  }

  /**
   * Gets or sets the row of this GraphObject if it is in a Table {@link Panel}.
   * The value must be a small non-negative integer. The default is 0.
   */
  /** @export */ get row(): number { return this._row; }
  /** @export */ set row(value: number) {
    if (Debug) Util.checkRealNumber(value, GraphObject, 'row');
    value = Math.round(value);
    const old = this._row;
    if (old !== value) {
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'row');
      this._row = value;
      this.invalidateMeasure();
      this.raiseChanged('row', old, value);
    }
  }

  /**
   * Gets or sets the number of rows spanned by this GraphObject if it is in a Table {@link Panel}.
   * The value must be a small positive integer. The default is 1.
   */
  /** @export */ get rowSpan(): number { return this._rowSpan; }
  /** @export */ set rowSpan(value: number) {
    if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'rowSpan');
    value = Math.round(value);
    const old = this._rowSpan;
    if (old !== value) {
      if (value < 1) Util.throwRangeError(value, '>= 1', GraphObject, 'rowSpan');
      this._rowSpan = value;
      this.invalidateMeasure();
      this.raiseChanged('rowSpan', old, value);
    }
  }

  //  * Gets or sets the function that determines what a GraphObject with a
  //  * {@link #rowSpan} or {@link #columnSpan} greater than 1 should do with its additional space,
  //  * if it is larger than all rows or columns it spans.
  //  *
  //  * If this property value is a function, the function is called with three arguments:
  //  *   - This {@link GraphObject}.
  //  *   - The {@link RowColumnDefinition} to consider granting width or height to.
  //  *   - The available width or height remaining (as a number).
  //  *
  //  * The function must return an amount of space to allocate for the given {@link RowColumnDefinition}.
  //  * For any given GraphObject, it should allocate no greater than the total available space
  //  * divided however the author chooses among all rows and columns spanned.
  //  * ```js
  //  * spanAllocation: function(obj, rowcoldef, available) {
  //  *   // place all extra span into row or column 2:
  //  *   if (rowcoldef.index === 2) return available;
  //  *   // give no extra span to any other row or column
  //  *   return 0;
  //  * },
  //  * ```
  //  *
  //  * By default this value is null, and by default objects that span multiple rows or columns
  //  * attempt to place any additional width or height in the last row or column that they cover.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get spanAllocation(): ((a: GraphObject, b: RowColumnDefinition, c: number) => number) | null { return this._spanAllocation; }
  /** @export */ set spanAllocation(value: ((a: GraphObject, b: RowColumnDefinition, c: number) => number) | null) {
    const old = this._spanAllocation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'spanAllocation');
      this._spanAllocation = value;
      this.invalidateMeasure();
      this.raiseChanged('spanAllocation', old, value);
    }
  }

  /**
   * Gets or sets the spot on this GraphObject to be used as the alignment point
   * in Spot and Fixed {@link Panel}s.
   * Value must be of the Spot.
   *
   * The default value is {@link Spot.Default}, which means that the Panel type can decide the effective alignment spot.
   *
   * The {@link #alignment} is often used along with this property to specify
   * where this object should be positioned in a Panel.
   *
   * For {@link Panel.Graduated}, the alignmentFocus spot determines the spot on a child element to be aligned with some
   * point along the main element.
   *
   * When you want a link label Node to be positioned by its location spot rather than by this alignmentFocus spot,
   * you can set this property to {@link Spot.None}, only on {@link Node}s.
   * @see {@link Panel#alignmentFocusName}
   *
   * For examples of alignments in different panels, see the <a href="../../intro/panels.html">Introduction page on Panels</a>.
   *
   * WARNING: Since 2.0, for Spot Panels, the offsetX/offsetY of {@link #alignmentFocus} has been reversed.
   * The offsetX/Y now describes offset distance from the alignmentFocus point to the alignment point, rather than the opposite.
   * This is what it has always described when using {@link #alignmentFocus} with Link Labels.
   */
  /** @export */ get alignmentFocus(): Spot { return this._alignmentFocus; }
  /** @export */ set alignmentFocus(value: Spot) {
    const old = this._alignmentFocus;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, GraphObject, 'alignmentFocus');
      if (Debug && value.isNoSpot() && !value.isDefault() &&
        !(value.isNone() && this instanceof Node)) {  // allow Spot.None for aligning label Nodes to locationSpot
        Util.throwError('GraphObject.alignmentFocus must be a real Spot or Spot.Default, not: ' + value);
      }
      value = value.copyFrozen();
      this._alignmentFocus = value;
      // ?? changing this to invalidateArrange broke tests
      this.invalidateMeasure();
      this.raiseChanged('alignmentFocus', old, value);
    }
  }

  /**********************************************************
  Port-specific properties:
  ***********************************************************/

  /**
   * Gets or sets an identifier for an object acting as a port on a Node.
   * The default value is null -- this object is not a port.
   *
   * A value that is the empty string is used by convention to mean the primary
   * (and usually only) port of the node.
   *
   * If a Node has no named ports, then the Node itself is the sole port.
   *
   * Note: the only kind of model that can save port information, i.e. portIds that are not an empty string,
   * for links is a {@link GraphLinksModel} whose {@link GraphLinksModel#linkFromPortIdProperty} and
   * {@link GraphLinksModel#linkToPortIdProperty} have been set to name properties on the link data objects.
   *
   * The value should be unique within the {@link Node}.
   * You must not modify this property once this GraphObject is in the visual tree of a Node.
   * <p class="boxread">
   * See <a href="../../intro/ports.html">the Introduction page on ports</a>
   * for usage information and examples.
   * @see {@link #fromLinkable}
   * @see {@link #toLinkable}
   * @see {@link #fromSpot}
   * @see {@link #toSpot}
   * @see {@link Link#fromSpot}
   * @see {@link Link#toSpot}
   */
  /** @export */ get portId(): string { return this._portId; }
  /** @export */ set portId(value: string) {
    const old = this._portId;
    if (old !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'string', GraphObject, 'portId');
      let part = this.part;
      if (part !== null && !(part instanceof Node)) {
        Util.throwError('Cannot set portID on a Link: ' + value);
        part = null;
      }
      const node = part as Node;
      if (old !== null && node !== null) {
        node.internalRemovePort(this);
      }
      this._portId = value;
      if (value !== null && node !== null) {
        node.hasNamedPorts = true;
        node.internalAddPort(this);
      }
      this.raiseChanged('portId', old, value);
    }
  }

  /**
   * @hidden @internal
   * Called by port-specific property setters
   * ??? This sorta duplicates functionality of _notifyPorts -- need to rearchitect and reimplement
   * @this {GraphObject}
   */
  private _invalidatePortLinkRoutes(): void {
    // no-op if isUndoingRedoing checked in invalidateConnectedLinks
    const part = this.part;
    if (part instanceof Node) {
      if (this.portId !== null || this === part.port) {
        const diagram = part.diagram;
        if (diagram !== null && !diagram.undoManager.isUndoingRedoing) {
          part.invalidatePortLinks(this);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Invalidate all link routes connected with port elements inside this element, including this itself.
   * @this {GraphObject}
   */
  private _invalidateLinkRoutesForContainedPorts(): void {
    const diagram = this.diagram;
    if (diagram === null || diagram.undoManager.isUndoingRedoing) return;
    if (this instanceof Panel) {
      if (this instanceof Node) {
        this.invalidateConnectedLinks();
      } else {
        this.walkVisualTreeFrom(this, function (obj) {
          obj._invalidatePortLinkRoutes();
        });
      }
    } else {
      this._invalidatePortLinkRoutes();
    }
  }

  /// for links coming into a port:

  /**
   * Gets or sets where a link should connect to this port.
   * The default value is {@link Spot.None}, meaning that the link routing
   * must consider the shape of the port and connect to the closest point.
   *
   * The value of {@link Link#toSpot}, if not {@link Spot.Default}, takes precedence over the value at this port
   * when determining the route of the link.
   * A number of the predefined {@link Layout}s automatically set {@link Link#fromSpot} and {@link Link#toSpot},
   * thereby causing this property and {@link #fromSpot} on the port element to be ignored.
   * Depending on the layout, you may be able to disable that behavior, such as by setting {@link ForceDirectedLayout#setsPortSpots},
   * {@link TreeLayout#setsPortSpot}, {@link TreeLayout#setsChildPortSpot}, or {@link LayeredDigraphLayout#setsPortSpots} to false.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link Link#toSpot}
   * @see {@link Link#computeSpot}
   * @see {@link #fromSpot}
   * @see {@link #portId}
   */
  /** @export */ get toSpot(): Spot { return this._linkSettings !== null ? this._linkSettings._toSpot : Spot.None; }
  /** @export */ set toSpot(value: Spot) {
    this.initLinkSettings();
    const old = this._linkSettings._toSpot;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, GraphObject, 'toSpot');
      value = value.copyFrozen();
      this._linkSettings._toSpot = value;
      this.raiseChanged('toSpot', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets the length of the last segment of a link going to this port.
   * This value is used when the computed "to spot" is not {@link Spot.None}.
   * The default value is 10.
   *
   * The value of {@link Link#toEndSegmentLength}, if not NaN, takes precedence over the value at this port
   * when determining the route of the link.
   * This value also limits how short the {@link Link#toShortLength} may be drawn.
   *
   * For examples of how to use this property, see <a href="../../intro/links.html#EndSegmentLengths">Link End Segment Lengths</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link Link#toEndSegmentLength}
   * @see {@link Link#computeEndSegmentLength}
   * @see {@link #fromEndSegmentLength}
   * @see {@link #portId}
   */
  /** @export */ get toEndSegmentLength(): number { return this._linkSettings !== null ? this._linkSettings._toEndSegmentLength : 10; }
  /** @export */ set toEndSegmentLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._toEndSegmentLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'toEndSegmentLength');
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'toEndSegmentLength');
      this._linkSettings._toEndSegmentLength = value;
      this.raiseChanged('toEndSegmentLength', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets how far the end segment of a link going to this port stops short of the actual port.
   * Positive values are limited by the {@link #toEndSegmentLength} or {@link Link#toEndSegmentLength}.
   * Negative values cause the link to extend into the port.
   * The default value is zero.
   *
   * This property is useful when you have a thick link and a pointy arrowhead.
   * Normally the link Shape extends all the way to the end of the arrowhead.
   * If the link Shape is wide, its edges will be seen behind the arrowhead.
   * By setting this property to a small positive value, the link Shape can end within the
   * body of the arrowhead, leaving only the point of the arrowhead visible at the end of the link.
   *
   * A negative value for this property can also be useful when you want the link Shape to continue
   * into the port, perhaps because a portion of the port is transparent and you want the link to
   * appear to connect visually with a different point on the node.
   *
   * The value of {@link Link#toShortLength}, if not NaN, takes precedence over the value at this port
   * when determining the route of the link.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #fromShortLength}
   * @see {@link Link#toShortLength}
   * @see {@link #portId}
   */
  /** @export */ get toShortLength(): number { return this._linkSettings !== null ? this._linkSettings._toShortLength : 0; }
  /** @export */ set toShortLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._toShortLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'toShortLength');
      this._linkSettings._toShortLength = value;
      this.raiseChanged('toShortLength', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets whether the user may draw Links to this port.
   * This property is used by {@link LinkingBaseTool#isValidTo}.
   *
   * The default value is null, which indicates that the real value is inherited from
   * the parent {@link Panel}, or false if there is no containing panel.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node,
   * or unless you are disabling the "linkability" of a particular GraphObject
   * inside a Panel whose toLinkable has been set or bound to true.
   * @see {@link #fromLinkable}
   * @see {@link #toMaxLinks}
   * @see {@link #portId}
   * @see {@link #cursor}
   */
  /** @export */ get toLinkable(): boolean | null { return this._linkSettings !== null ? this._linkSettings._toLinkable : null; }
  /** @export */ set toLinkable(value: boolean | null) {
    this.initLinkSettings();
    const old = this._linkSettings._toLinkable;
    if (old !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'boolean', GraphObject, 'toLinkable');
      this._linkSettings._toLinkable = value;
      this.raiseChanged('toLinkable', old, value);
    }
  }

  /**
   * Gets or sets the maximum number of links that may go into this port.
   * This property is used by {@link LinkingBaseTool#isValidTo}.
   *
   * The value must be non-negative.
   * The default value is Infinity.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #toLinkable}
   * @see {@link #fromMaxLinks}
   * @see {@link #portId}
   */
  /** @export */ get toMaxLinks(): number { return this._linkSettings !== null ? this._linkSettings._toMaxLinks : Infinity; }
  /** @export */ set toMaxLinks(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._toMaxLinks;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'toMaxLinks');
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'toMaxLinks');
      this._linkSettings._toMaxLinks = value;
      this.raiseChanged('toMaxLinks', old, value);
    }
  }

  /// for links going out of a port:

  /**
   * Gets or sets where a link should connect from this port.
   * The default value is {@link Spot.None}, meaning that the link routing
   * must consider the shape of the port and connect at the closest point.
   *
   * The value of {@link Link#fromSpot}, if not {@link Spot.Default}, takes precedence over the value at this port
   * when determining the route of the link.
   * A number of the predefined {@link Layout}s automatically set {@link Link#fromSpot} and {@link Link#toSpot},
   * thereby causing this property and {@link #toSpot} on the port element to be ignored.
   * Depending on the layout, you may be able to disable that behavior, such as by setting {@link ForceDirectedLayout#setsPortSpots},
   * {@link TreeLayout#setsPortSpot}, {@link TreeLayout#setsChildPortSpot}, or {@link LayeredDigraphLayout#setsPortSpots} to false.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link Link#fromSpot}
   * @see {@link Link#computeSpot}
   * @see {@link #toSpot}
   * @see {@link #portId}
   */
  /** @export */ get fromSpot(): Spot { return this._linkSettings !== null ? this._linkSettings._fromSpot : Spot.None; }
  /** @export */ set fromSpot(value: Spot) {
    this.initLinkSettings();
    const old = this._linkSettings._fromSpot;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, GraphObject, 'fromSpot');
      value = value.copyFrozen();
      this._linkSettings._fromSpot = value;
      this.raiseChanged('fromSpot', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets the length of the first segment of a link coming from this port.
   * This value is used when the computed "from spot" is not {@link Spot.None}.
   * The default value is 10.
   * This value also limits how short the {@link Link#fromShortLength} may be drawn.
   *
   * The value of {@link Link#fromEndSegmentLength}, if not NaN, takes precedence over the value at this port
   * when determining the route of the link.
   *
   * For examples of how to use this property, see <a href="../../intro/links.html#EndSegmentLengths">Link End Segment Lengths</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link Link#fromEndSegmentLength}
   * @see {@link Link#computeEndSegmentLength}
   * @see {@link #toEndSegmentLength}
   * @see {@link #portId}
   */
  /** @export */ get fromEndSegmentLength(): number { return this._linkSettings !== null ? this._linkSettings._fromEndSegmentLength : 10; }
  /** @export */ set fromEndSegmentLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._fromEndSegmentLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'fromEndSegmentLength');
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'fromEndSegmentLength');
      this._linkSettings._fromEndSegmentLength = value;
      this.raiseChanged('fromEndSegmentLength', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets how far the end segment of a link coming from this port stops short of the actual port.
   * Positive values are limited by the {@link #fromEndSegmentLength} or {@link Link#fromEndSegmentLength}.
   * Negative values cause the link to extend into the port.
   * The default value is zero.
   *
   * This property is useful when you have a thick link and a pointy arrowhead.
   * Normally the link Shape extends all the way to the end of the arrowhead.
   * If the link Shape is wide, its edges will be seen behind the arrowhead.
   * By setting this property to a small positive value, the link Shape can end within the
   * body of the arrowhead, leaving only the point of the arrowhead visible at the end of the link.
   *
   * A negative value for this property can also be useful when you want the link Shape to continue
   * into the port, perhaps because a portion of the port is transparent and you want the link to
   * appear to connect visually with a different point on the node.
   *
   * The value of {@link Link#fromShortLength}, if not NaN, takes precedence over the value at this port
   * when determining the route of the link.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #toShortLength}
   * @see {@link Link#fromShortLength}
   * @see {@link #portId}
   */
  /** @export */ get fromShortLength(): number { return this._linkSettings !== null ? this._linkSettings._fromShortLength : 0; }
  /** @export */ set fromShortLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._fromShortLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'fromShortLength');
      this._linkSettings._fromShortLength = value;
      this.raiseChanged('fromShortLength', old, value);
      this._invalidatePortLinkRoutes();
    }
  }

  /**
   * Gets or sets whether the user may draw Links from this port.
   * This property is used by {@link LinkingBaseTool#isValidFrom}.
   *
   * The default value is null, which indicates that the real value is inherited from
   * the parent {@link Panel}, or false if there is no containing panel.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node,
   * or unless you are disabling the "linkability" of a particular GraphObject
   * inside a Panel whose fromLinkable has been set or bound to true.
   * @see {@link #toLinkable}
   * @see {@link #fromMaxLinks}
   * @see {@link #portId}
   * @see {@link #cursor}
   */
  /** @export */ get fromLinkable(): boolean | null { return this._linkSettings !== null ? this._linkSettings._fromLinkable : null; }
  /** @export */ set fromLinkable(value: boolean | null) {
    this.initLinkSettings();
    const old = this._linkSettings._fromLinkable;
    if (old !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'boolean', GraphObject, 'fromLinkable');
      this._linkSettings._fromLinkable = value;
      this.raiseChanged('fromLinkable', old, value);
    }
  }

  /**
   * Gets or sets the maximum number of links that may come out of this port.
   * This property is used by {@link LinkingBaseTool#isValidFrom}.
   *
   * The value must be non-negative.
   * The default value is Infinity.
   *
   * You must set this property on a GraphObject whose {@link #portId} is non-null,
   * unless the whole {@link Node} is acting as a single port,
   * in which case this property should be set on the Node.
   * @see {@link #fromLinkable}
   * @see {@link #toMaxLinks}
   * @see {@link #portId}
   */
  /** @export */ get fromMaxLinks(): number { return this._linkSettings !== null ? this._linkSettings._fromMaxLinks : Infinity; }
  /** @export */ set fromMaxLinks(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._fromMaxLinks;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', GraphObject, 'fromMaxLinks');
      if (value < 0) Util.throwRangeError(value, '>= 0', GraphObject, 'fromMaxLinks');
      this._linkSettings._fromMaxLinks = value;
      this.raiseChanged('fromMaxLinks', old, value);
    }
  }

  /**********************************************************
  Event Handling:
  ***********************************************************/

  /**
   * Gets or sets the mouse cursor to use when the mouse is over this object with no mouse buttons pressed.
   * The value is null when no particular cursor is specified for this object;
   * the actual cursor is determined by any containing {@link Panel}.
   *
   * The default value is the empty string, which means the
   * current mouse cursor is determined by the Diagram.
   * Other strings should be valid CSS strings that specify a cursor.
   * This provides some more information about cursor syntax:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/cursor">CSS cursors (mozilla.org)</a>.
   * @see {@link Diagram#defaultCursor}
   * @see {@link Diagram#currentCursor}
   */
  /** @export */ get cursor(): string { return this._cursor; }
  /** @export */ set cursor(value: string) {
    const old = this._cursor;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', GraphObject, 'cursor');
      this._cursor = value;
      this.raiseChanged('cursor', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user single-primary-clicks on this object.
   * This typically involves a mouse-down followed by a prompt mouse-up
   * at approximately the same position using the left (primary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on a {@link GraphObject}.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name `"ObjectSingleClicked"`.
   *
   * If this property value is a function, it is called with an {@link InputEvent}
   * and this {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   *
   * From the second argument, *obj*, you can get to the Node or Link via the {@link #part} property.
   * From there you can access the bound data via the {@link Panel#data} property.
   * So from an event handler you can get the bound data by `obj.part.data`.
   *
   * By default this property is null.
   *
   * Objects in Layers that are {@link Layer#isTemporary} do not receive click events.
   * If you do want such objects to respond to clicks, set {@link #isActionable} to true.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   * <p class="boxrun">
   * An example of a click event handler is shown in the
   * <a href="../../samples/arrowheads.html">Arrowheads sample</a>.
   * @see {@link #doubleClick}
   * @see {@link #contextClick}
   * @see {@link Diagram#click}
   */
  /** @export */ get click(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._click : null; }
  /** @export */ set click(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._click;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'click');
      this._eventHandlers._click = value;
      this.raiseChanged('click', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user double-primary-clicks on this object.
   * This typically involves a mouse-down/up/down/up in rapid succession
   * at approximately the same position using the left (primary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on a {@link GraphObject}.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name `"ObjectDoubleClicked"`.
   *
   * If this property value is a function, it is called with an {@link InputEvent}
   * and this {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   *
   * From the second argument, *obj*, you can get to the Node or Link via the {@link #part} property.
   * From there you can access the bound data via the {@link Panel#data} property.
   * So from an event handler you can get the bound data by `obj.part.data`.
   *
   * By default this property is null.
   *
   * Objects in Layers that are {@link Layer#isTemporary} do not receive click events.
   * If you do want such objects to respond to clicks, set {@link #isActionable} to true.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   *
   * The <a href="../../samples/classHierarchy.html">Class Hierarchy sample</a>
   * demonstrates the definition of a double-click event handler that opens up
   * a web page with the documentation for that class:
   * ```js
   * diagram.nodeTemplate =
   *   $(go.Node, . . .,
   *     {
   *       doubleClick:  // here the second argument is this object, which is this Node
   *         function(e, node) { window.open("../api/symbols/" + node.data.key + ".html"); }
   *     },
   *     . . .
   *     );
   * ```
   * @see {@link #click}
   * @see {@link #contextClick}
   * @see {@link Diagram#doubleClick}
   */
  /** @export */ get doubleClick(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._doubleClick : null; }
  /** @export */ set doubleClick(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._doubleClick;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'doubleClick');
      this._eventHandlers._doubleClick = value;
      this.raiseChanged('doubleClick', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user single-secondary-clicks on this object.
   * This typically involves a mouse-down followed by a prompt mouse-up
   * at approximately the same position using the right (secondary) mouse button.
   * This property is used by the {@link ClickSelectingTool}
   * when the user clicks on a {@link GraphObject}.
   * The function is called in addition to the {@link DiagramEvent}
   * that is raised with the name `"ObjectContextClicked"`.
   *
   * If this property value is a function, it is called with an {@link InputEvent}
   * and this {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   *
   * From the second argument, *obj*, you can get to the Node or Link via the {@link #part} property.
   * From there you can access the bound data via the {@link Panel#data} property.
   * So from an event handler you can get the bound data by `obj.part.data`.
   *
   * By default this property is null.
   *
   * Objects in Layers that are {@link Layer#isTemporary} do not receive click events.
   * If you do want such objects to respond to clicks, set {@link #isActionable} to true.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   * @see {@link #click}
   * @see {@link #doubleClick}
   * @see {@link Diagram#contextClick}
   */
  /** @export */ get contextClick(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._contextClick : null; }
  /** @export */ set contextClick(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._contextClick;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'contextClick');
      this._eventHandlers._contextClick = value;
      this.raiseChanged('contextClick', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse
   * into this object without holding down any buttons.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent},
   * this {@link GraphObject} that the mouse is now in,
   * and any previous {@link GraphObject} that the mouse was in.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * By default this property is null.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * After calling this function the diagram will be updated immediately.
   *
   * For example, consider the situation where one wants to display buttons that the user can click
   * whenever the user passes the mouse over a node, and the buttons automatically disappear when the
   * mouse leaves the node.  This can be implemented by showing an Adornment holding the buttons.
   * ```js
   * var nodeContextMenu =
   *   $(go.Adornment, "Spot",
   *     { background: "transparent" },  // to help detect when the mouse leaves the area
   *     $(go.Placeholder),
   *     $(go.Panel, "Vertical",
   *       { alignment: go.Spot.Right, alignmentFocus: go.Spot.Left },
   *       $("Button",
   *         $(go.TextBlock, "Command 1"),
   *         {
   *           click: function(e, obj) {
   *             var node = obj.part.adornedPart;
   *             alert("Command 1 on " + node.data.text);
   *             node.removeAdornment("ContextMenuOver");
   *           }
   *         }),
   *       $("Button",
   *         $(go.TextBlock, "Command 2"),
   *         {
   *           click: function(e, obj) {
   *             var node = obj.part.adornedPart;
   *             alert("Command 2 on " + node.data.text);
   *             node.removeAdornment("ContextMenuOver");
   *           }
   *         })
   *       ));
   * ```
   * Then in the definition of the Node we can implement a mouseEnter event handler:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node,
   *     . . .
   *     {
   *       mouseEnter: function(e, node) {
   *         nodeContextMenu.adornedObject = node;
   *         nodeContextMenu.mouseLeave = function(ev, cm) {
   *           node.removeAdornment("ContextMenuOver");
   *         }
   *         node.addAdornment("ContextMenuOver", nodeContextMenu);
   *       }
   *     });
   * ```
   * Note how it automatically defines a {@link #mouseLeave} event handler too.
   * The context menu Adornment is removed either when the mouse leaves the area of the Adornment
   * or when the user executes a button click event handler.
   * @see {@link #mouseLeave}
   * @see {@link #mouseOver}
   * @see {@link #mouseHover}
   * @see {@link #mouseDragEnter}
   */
  /** @export */ get mouseEnter(): ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseEnter : null; }
  /** @export */ set mouseEnter(value: ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseEnter;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseEnter');
      this._eventHandlers._mouseEnter = value;
      this.raiseChanged('mouseEnter', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse
   * out of this object without holding down any buttons.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent},
   * this {@link GraphObject} that the mouse has left,
   * and any next {@link GraphObject} that the mouse is now in.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * By default this property is null.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * After calling this function the diagram will be updated immediately.
   *
   * For example, the <a href="../../samples/flowchart.html">Flow Chart sample</a>
   * automatically shows and hides the ports as the mouse passes over a node.
   * The node template includes the following settings:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node,
   *     . . .
   *     {
   *       . . .
   *       // handle mouse enter/leave events to show/hide the ports
   *       mouseEnter: function(e, obj) { showPorts(obj.part, true); },
   *       mouseLeave: function(e, obj) { showPorts(obj.part, false); }
   *       . . .
   *     });
   * ```
   * where the `showPorts` function is defined to set the {@link #visible}
   * property of each of the port elements of the node.
   * @see {@link #mouseEnter}
   * @see {@link #mouseOver}
   * @see {@link #mouseHover}
   * @see {@link #mouseDragLeave}
   */
  /** @export */ get mouseLeave(): ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseLeave : null; }
  /** @export */ set mouseLeave(value: ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseLeave;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseLeave');
      this._eventHandlers._mouseLeave = value;
      this.raiseChanged('mouseLeave', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse
   * over this object without holding down any buttons.
   * This property is used by the {@link ToolManager}.
   * This property is infrequently used -- it is more common to implement {@link #mouseEnter}
   * and {@link #mouseLeave} functions.
   *
   * If this property value is a function, it is called with an {@link InputEvent}
   * and this {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * By default this property is null.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * After calling this function the diagram will be updated immediately.
   * @see {@link #mouseHover}
   * @see {@link #mouseEnter}
   * @see {@link #mouseLeave}
   */
  /** @export */ get mouseOver(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseOver : null; }
  /** @export */ set mouseOver(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseOver;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseOver');
      this._eventHandlers._mouseOver = value;
      this.raiseChanged('mouseOver', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user holds the mouse still for a while
   * over this object without holding down any buttons.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   *
   * You can control how long the user must wait with a motionless mouse before
   * a "mouse hover" event occurs, by setting {@link ToolManager#hoverDelay}.
   * For example:
   * ```js
   * myDiagram =
   *   $(go.Diagram, "myDiagramDiv",
   *     { "toolManager.hoverDelay": 500 });  // 500 milliseconds
   * ```
   * or:
   * ```js
   * myDiagram.toolManager.hoverDelay = 500;  // 500 milliseconds
   * ```
   * @see {@link #mouseOver}
   * @see {@link #mouseEnter}
   * @see {@link #mouseLeave}
   */
  /** @export */ get mouseHover(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseHover : null; }
  /** @export */ set mouseHover(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseHover;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseHover');
      this._eventHandlers._mouseHover = value;
      this.raiseChanged('mouseHover', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user holds the mouse still for a while
   * over this object while holding down a button.
   * This property is used by the {@link ToolManager}.
   *
   * If this property value is a function, it is called with an {@link InputEvent}.
   * By default this property is null.
   *
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   *
   * You can control how long the user must wait during a drag with a motionless mouse before
   * a "mouse hold" event occurs, by setting {@link ToolManager#holdDelay}.
   * For example:
   * ```js
   * myDiagram =
   *   $(go.Diagram, "myDiagramDiv",
   *     { "toolManager.holdDelay": 500 });  // 500 milliseconds
   * ```
   * or:
   * ```js
   * myDiagram.toolManager.holdDelay = 500;  // 500 milliseconds
   * ```
   * @see {@link #mouseDragEnter}
   * @see {@link #mouseDragLeave}
   * @see {@link #mouseHover}
   */
  /** @export */ get mouseHold(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseHold : null; }
  /** @export */ set mouseHold(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseHold;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseHold');
      this._eventHandlers._mouseHold = value;
      this.raiseChanged('mouseHold', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse
   * into this stationary object during a {@link DraggingTool} drag;
   * this allows you to provide feedback during a drag based on where it might drop.
   *
   * If this property value is a function, it is called with an {@link InputEvent},
   * this {@link GraphObject}, and any previous {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * By default this property is null.
   *
   * Note that for a drag-and-drop that originates in a different diagram, the target diagram's
   * selection collection will not be the parts that are being dragged.
   * Instead the temporary parts being dragged can be found as the source diagram's {@link DraggingTool#copiedParts}.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function,
   * because the {@link DraggingTool} will be conducting one already.
   * After calling this function the diagram will be updated immediately.
   *
   * For an example of a mouseDragEnter event handler, see the node template in the
   * <a href="../../samples/orgChartEditor.html">Org Chart Editor sample</a>.
   * @see {@link #mouseDragLeave}
   * @see {@link #mouseHold}
   * @see {@link #mouseDrop}
   * @see {@link #mouseEnter}
   * @see {@link Group#handlesDragDropForMembers}
   */
  /** @export */ get mouseDragEnter(): ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null {
    return this._eventHandlers !== null ? this._eventHandlers._mouseDragEnter : null;
  }
  /** @export */ set mouseDragEnter(value: ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseDragEnter;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseDragEnter');
      this._eventHandlers._mouseDragEnter = value;
      this.raiseChanged('mouseDragEnter', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the user moves the mouse
   * out of this stationary object during a {@link DraggingTool} drag;
   * this allows you to provide feedback during a drag based on where it might drop.
   *
   * If this property value is a function, it is called with an {@link InputEvent},
   * this {@link GraphObject}, and any new {@link GraphObject} that the mouse is in.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * By default this property is null.
   *
   * Note that for a drag-and-drop that originates in a different diagram, the target diagram's
   * selection collection will not be the parts that are being dragged.
   * Instead the temporary parts being dragged can be found as the source diagram's {@link DraggingTool#copiedParts}.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function,
   * because the {@link DraggingTool} will be conducting one already.
   * After calling this function the diagram will be updated immediately.
   *
   * For an example of a mouseDragLeave event handler, see the node template in the
   * <a href="../../samples/orgChartEditor.html">Org Chart Editor sample</a>.
   * @see {@link #mouseDragEnter}
   * @see {@link #mouseHold}
   * @see {@link #mouseDrop}
   * @see {@link #mouseLeave}
   * @see {@link Group#handlesDragDropForMembers}
   */
  /** @export */ get mouseDragLeave(): ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null {
    return this._eventHandlers !== null ? this._eventHandlers._mouseDragLeave : null;
  }
  /** @export */ set mouseDragLeave(value: ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseDragLeave;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseDragLeave');
      this._eventHandlers._mouseDragLeave = value;
      this.raiseChanged('mouseDragLeave', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when a user drops the selection on this object
   * at the end of a {@link DraggingTool} drag;
   * this allows you to customize the behavior when a drop occurs on an object.
   *
   * If this property value is a function, it is called with an {@link InputEvent},
   * this {@link GraphObject}.
   * The {@link InputEvent#targetObject} provides the GraphObject that was found
   * at the mouse point before looking up the visual tree of {@link GraphObject#panel}s
   * to get to this object.
   * The function is called within the transaction performed by the {@link DraggingTool},
   * so you do not need to conduct one.
   * By default this property is null.
   *
   * For an example of a mouseDrop event handler, see the node template in the
   * <a href="../../samples/orgChartEditor.html">Org Chart Editor sample</a>.
   * @see {@link #mouseDragEnter}
   * @see {@link #mouseDragLeave}
   * @see {@link #mouseHold}
   * @see {@link Group#handlesDragDropForMembers}
   */
  /** @export */ get mouseDrop(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._mouseDrop : null; }
  /** @export */ set mouseDrop(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._mouseDrop;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'mouseDrop');
      this._eventHandlers._mouseDrop = value;
      this.raiseChanged('mouseDrop', old, value);
    }
  }

  /**
   * Gets or sets the function to execute on a mouse-down event when this GraphObject's {@link #isActionable}
   * is set to true.
   * This property is infrequently set.
   * By default this property is null.
   *
   * This functional property is only set on objects such as buttons, knobs, or sliders that want to handle all events,
   * in conjunction with {@link ActionTool}, pre-empting the normal tool mechanisms.
   *
   * The {@link ActionTool} does not conduct any transaction, so if this property has a value,
   * the function will not be called within a transaction.
   * @see {@link #actionMove}
   * @see {@link #actionUp}
   * @see {@link #actionCancel}
   */
  /** @export */ get actionDown(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._actionDown : null; }
  /** @export */ set actionDown(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._actionDown;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'actionDown');
      this._eventHandlers._actionDown = value;
      this.raiseChanged('actionDown', old, value);
    }
  }

  /**
   * Gets or sets the function to execute on a mouse-move event when this GraphObject's {@link #isActionable}
   * is set to true.
   * This property is infrequently set.
   * By default this property is null.
   *
   * This functional property is only set on objects such as buttons, knobs, or sliders that want to handle all events,
   * in conjunction with {@link ActionTool}, pre-empting the normal tool mechanisms.
   *
   * The {@link ActionTool} does not conduct any transaction, so if this property has a value,
   * the function will not be called within a transaction.
   * @see {@link #actionDown}
   * @see {@link #actionUp}
   * @see {@link #actionCancel}
   */
  /** @export */ get actionMove(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._actionMove : null; }
  /** @export */ set actionMove(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._actionMove;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'actionMove');
      this._eventHandlers._actionMove = value;
      this.raiseChanged('actionMove', old, value);
    }
  }

  /**
   * Gets or sets the function to execute on a mouse-up event when this GraphObject's {@link #isActionable}
   * is set to true.
   * This property is infrequently set.
   * By default this property is null.
   *
   * This functional property is only set on objects such as buttons, knobs, or sliders that want to handle all events,
   * in conjunction with {@link ActionTool}, pre-empting the normal tool mechanisms.
   *
   * The {@link ActionTool} does not conduct any transaction, so if this property has a value,
   * the function will not be called within a transaction.
   * If you do provide a function that makes changes to the diagram or to its model,
   * you should do so within a transaction -- call {@link Diagram#startTransaction} and
   * {@link Diagram#commitTransaction}.
   * @see {@link #actionDown}
   * @see {@link #actionMove}
   * @see {@link #actionCancel}
   */
  /** @export */ get actionUp(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._actionUp : null; }
  /** @export */ set actionUp(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._actionUp;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'actionUp');
      this._eventHandlers._actionUp = value;
      this.raiseChanged('actionUp', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when the {@link ActionTool} is cancelled and this GraphObject's {@link #isActionable}
   * is set to true.
   * This property is infrequently set.
   * By default this property is null.
   *
   * This functional property is only set on objects such as buttons, knobs, or sliders that want to handle all events,
   * in conjunction with {@link ActionTool}, pre-empting the normal tool mechanisms.
   *
   * The {@link ActionTool} does not conduct any transaction, so if this property has a value,
   * the function will not be called within a transaction.
   * @see {@link #actionDown}
   * @see {@link #actionMove}
   * @see {@link #actionUp}
   */
  /** @export */ get actionCancel(): ((e: InputEvent, thisObj: GraphObject) => void) | null { return this._eventHandlers !== null ? this._eventHandlers._actionCancel : null; }
  /** @export */ set actionCancel(value: ((e: InputEvent, thisObj: GraphObject) => void) | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._actionCancel;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphObject, 'actionCancel');
      this._eventHandlers._actionCancel = value;
      this.raiseChanged('actionCancel', old, value);
    }
  }

  /**
   * This {@link Adornment} or {@link HTMLInfo} is shown when the mouse hovers over this object.
   * The default value is null, which means no tooltip is shown.
   *
   * A typical tooltip is defined in the following manner, as taken from
   * the <a href="../../samples/kittenMonitor.html">Kitten Monitor sample</a>:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node,
   *     . . .
   *     { // this tooltip shows the name and picture of the kitten
   *       toolTip:
   *         $("ToolTip",
   *           $(go.Panel, "Vertical",
   *             $(go.Picture,
   *               new go.Binding("source", "src", function(s) { return "images/" + s + ".png"; })),
   *             $(go.TextBlock, { margin: 3 },
   *               new go.Binding("text", "key"))))
   *     });
   * ```
   * Note that this Adornment depends on having the same data binding as the adorned Part
   * (i.e. the same value for {@link Panel#data}).
   *
   * Tooltips are not copied by {@link #copy}, so that tooltips may be shared by all instances of a template.
   *
   * Tooltips are shown after a timed delay given by the {@link ToolManager#hoverDelay}.
   * You can change the delay time by:
   * ```js
   * myDiagram =
   *   $(go.Diagram, "myDiagramDiv",
   *     { "toolManager.hoverDelay": 500 });  // 500 milliseconds
   * ```
   * or:
   * ```js
   * myDiagram.toolManager.hoverDelay = 500;  // 500 milliseconds
   * ```
   *
   * Tooltips are normally positioned by {@link ToolManager#positionToolTip}.
   * However, if there is a {@link Placeholder} in the tooltip, the tooltip (i.e. an Adornment)
   * will be positioned so that the Placeholder is at the same position as this adorned GraphObject.
   *
   * Replacing this value will not modify or remove any existing tooltip that is being shown for this object.
   *
   * Read more about tooltips at <a href="../../intro/toolTips.html">ToolTips</a>.
   */
  /** @export */ get toolTip(): Adornment | HTMLInfo | null { return this._eventHandlers !== null ? this._eventHandlers._toolTip : null; }
  /** @export */ set toolTip(value: Adornment | HTMLInfo | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._toolTip;
    if (old !== value) {
      if (Debug && value !== null && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
        Util.throwError('GraphObject.toolTip must be an Adornment or HTMLInfo.');
      }
      this._eventHandlers._toolTip = value;
      this.raiseChanged('toolTip', old, value);
    }
  }

  /**
   * This {@link Adornment} or {@link HTMLInfo} is shown upon a context click on this object.
   * The default value is null, which means no context menu is shown.
   *
   * Changing this value will not modify or remove any existing menu that is being shown for this object.
   *
   * Context menus may also depend on having the same data binding as the adorned Part
   * (i.e. the same value for {@link Panel#data}).
   *
   * Context menus are not copied by {@link #copy}, so that context menus may be shared by all instances of a template.
   *
   * A typical context menu is implemented as an Adornment with several buttons in it.
   * For example, this context menu is defined in the
   * <a href="../../samples/dynamicPorts.html">Dynamic Port sample</a>:
   * ```js
   * var nodeMenu =  // context menu for each Node
   *  $("ContextMenu",
   *     $("ContextMenuButton",
   *       $(go.TextBlock, "Add top port"),
   *       { click: function(e, obj) { addPort("top"); } }),
   *     $("ContextMenuButton",
   *       $(go.TextBlock, "Add left port"),
   *       { click: function(e, obj) { addPort("left"); } }),
   *     $("ContextMenuButton",
   *       $(go.TextBlock, "Add right port"),
   *       { click: function(e, obj) { addPort("right"); } }),
   *     $("ContextMenuButton",
   *       $(go.TextBlock, "Add bottom port"),
   *       { click: function(e, obj) { addPort("bottom"); } }));
   * ```
   * and is used in the node template:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node, "Table",
   *     { . . .
   *       contextMenu: nodeMenu
   *     },
   *     . . .);
   * ```
   *
   * Context menus are normally positioned by {@link ContextMenuTool#positionContextMenu}.
   * However, if there is a {@link Placeholder} in the context menu, the context menu (i.e. an Adornment)
   * will be positioned so that the Placeholder is at the same position as this adorned GraphObject.
   *
   * The <a href="../../samples/basic.html">Basic sample</a> also shows how
   * to make context menu items invisible when the command is disabled.
   *
   * Replacing this value will not modify or remove any existing context menu that is being shown for this object.
   *
   * Read more about context menus at <a href="../../intro/contextMenus.html">Context Menus</a>.
   */
  /** @export */ get contextMenu(): Adornment | HTMLInfo | null { return this._eventHandlers !== null ? this._eventHandlers._contextMenu : null; }
  /** @export */ set contextMenu(value: Adornment | HTMLInfo | null) {
    this.initEventHandlers();
    const old = this._eventHandlers._contextMenu;
    if (old !== value) {
      if (Debug && !((value instanceof Adornment) || (value instanceof HTMLInfo))) {
        Util.throwError('GraphObject.contextMenu must be an Adornment or HTMLInfo.');
      }
      this._eventHandlers._contextMenu = value;
      this.raiseChanged('contextMenu', old, value);
    }
  }

  /**
   * Add a data-binding of a property on this GraphObject to a property on a data object.
   * It is unusual to call this method -- {@link GraphObject.make} will call this method for you
   * when you pass it an instance of {@link Binding}.
   *
   * Do not add, modify, or remove any Bindings after this object has been copied.
   *
   * Read more about {@link Binding}s at <a href="../../intro/dataBinding.html">the Introduction page about Data Bindings</a>.
   * @export
   * @this {GraphObject}
   * @param {Binding} binding
   */
  public bind(binding: Binding): void {
    binding._targetObject = this;  // remember so that Part.freezeBindings can generalize using name-lookup
    const binder = this.findTemplateBinder();
    if (binder !== null && binder.isTemplateBinderFrozen()) {
      Util.throwError('Cannot add a Binding to a template that has already been copied: ' + binding);
    }
    if (this._binds === null) this._binds = new List<Binding>();
    this._binds.add(binding);
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {GraphObject}
   * @return {Panel}
   */
  public findTemplateBinder(): Panel | null {
    let elt = this instanceof Panel ? this : this.panel;
    while (elt !== null) {
      if (elt.isTemplateBinder()) return elt;
      elt = elt.panel;
    }
    return null;
  }

  /**
   * This method sets a collection of properties according to the property/value pairs that have been set on the given Object,
   * in the same manner as {@link GraphObject.make} does when constructing a GraphObject with an argument that is a simple JavaScript Object.
   *
   * If this is a {@link Panel}, you can set properties on named elements within the panel by using a *name.property* syntax for the property name.
   * For example, if a {@link Node} has a {@link Picture} that is named "ICON" (because its {@link #name} property has been set to "ICON")
   * and a {@link TextBlock} whose name is "TB", one could set properties on the Node and on each of those named elements by:
   * ```js
   * aNode.setProperties({
   *     background: "red",
   *     "ICON.source": "https://www.example.com/images/alert.jpg",
   *     "TB.font": "bold 12pt sans-serif"
   * });
   * ```
   *
   * At the current time only a single dot is permitted in the property "name".
   * Note that the use of all-upper-case object names is simply a convention.
   *
   * Calling this method is much less efficient than setting properties directly.
   * @export
   * @this {GraphObject}
   * @param {Object} props a plain JavaScript object with various property values to be set on this GraphObject.
   * @since 1.5
   */
  public setProperties(props: ObjectData): void {
    Diagram.setCheckedProperties(this, props);
  }

  // Since 2.1
  /**
   * @hidden
   * Undocumented
   */
  public trigger(trigger: AnimationTrigger): void {
    trigger._targetObject = this;
    if (this._animationTriggers === null) this._animationTriggers = new Map<string, AnimationTrigger>();
    this._animationTriggers.add(trigger.propertyName, trigger);
  }


  /**
   * This static function builds an object given its class and additional arguments
   * providing initial properties or {@link GraphObject}s that become {@link Panel} elements.
   *
   * The first argument must be the class type or the name of a class or the name of a predefined kind of Panel.
   * This function will construct a new instance of that type and use the rest of the arguments to initialize the object.
   * The first argument cannot be a regular Object (such as a GraphObject) that you are trying to initialize;
   * for that you can call {@link #setProperties} or {@link Diagram#setProperties}, although that would
   * be less efficient than setting properties directly.
   *
   * If an initializer argument is an enumerated value, this tries to set the property that seems most appropriate.
   *
   * If an initializer argument is a string, this sets a particular property depending on the type of object being built.
   *   - If the object is a {@link TextBlock}, it sets {@link TextBlock#text}.
   *   - If the object is a {@link Shape}, it sets {@link Shape#figure}.
   *   - If the object is a {@link Picture}, it sets {@link Picture#source}.
   *   - If the object is a {@link Panel} (including {@link Part}, {@link Node}, or {@link Group}), it sets {@link Panel#type}.
   *
   * If an initializer argument is a particular kind of object, this can add that object to the object being built.
   *   - {@link GraphObject}s and {@link RowColumnDefinition}s can only be added as elements of {@link Panel}s.
   *   - {@link Binding}s can only be applied to {@link GraphObject}s and {@link RowColumnDefinition}s.
   *   - {@link PathFigure}s can only be added to {@link Geometry} objects.
   *   - {@link PathSegment}s can only be added to {@link PathFigure} objects.
   *   - Regular JavaScript Arrays provide a sequence of initializer arguments.
   *   - Regular JavaScript objects provide property/value pairs that are set on the object being built.
   *
   * When the initializer argument is a plain JavaScript Object, there are several ways that that object's properties are applied.
   * If the property name is a string with a period inside it, this has a special meaning if the object is a {@link Panel} or a {@link Diagram}.
   * At the current time only a single period separator is valid syntax for a property string, and it is valid only on Panels and Diagrams.
   *
   * For Panels, the substring before the period is used as the name passed to {@link Panel#findObject}
   * to get the actual object on which to set the property, which is the substring after the period.
   * This is normally useful only on the predefined Panels:
   *   - a **"Button"** has a {@link Shape} named "ButtonBorder" surrounding the content of the Panel.
   *   - a **"TreeExpanderButton"** has a "ButtonBorder" Shape and a "ButtonIcon" Shape that is the plus-or-minus sign.
   *   - a **"SubGraphExpanderButton"** has a "ButtonBorder" Shape and a "ButtonIcon" Shape that is the plus-or-minus sign.
   *   - a **"ContextMenuButton"** has a Shape named "ButtonBorder" surrounding the content of the Panel.
   *
   * But you can define your own names that GraphObject.make can build by calling the static function {@link GraphObject.defineBuilder}.
   *
   * For Diagrams, the substring before the period is used as the name of a property on the Diagram itself
   * to get the actual object on which to set the property.
   * As a special case, if such a property value does not exist on the Diagram, it looks on the {@link Diagram#toolManager}.
   * See some examples below.
   *
   * Also for Diagrams, and only for Diagrams, if the property name is the name of a {@link DiagramEvent},
   * the property value must be a DiagramEvent listener function, and {@link Diagram#addDiagramListener} is called
   * using that DiagramEvent name and that function.
   * Note that all DiagramEvent names are capitalized and do not contain any periods,
   * so there cannot be any name conflicts with any properties on {@link Diagram} or {@link ToolManager}.
   * Although you can register multiple listeners for the same DiagramEvent names, due to JavaScript limitations
   * those need to be declared using separate JavaScript objects, because JavaScript does not permit duplicate property names
   * in an Object literal.
   *
   * Furthermore for Diagrams, if the property name is `"Changed"` or `"ModelChanged"`,
   * the property value must be a ChangedEvent listener function, which is called with a {@link ChangedEvent} argument.
   * When the property name is `"Changed"`, it calls {@link Diagram#addChangedListener},
   * notifying about changes to the Diagram or its Layers or GraphObjects.
   * When the property name is `"ModelChanged"`, it calls {@link Model#addChangedListener} on the {@link Diagram#model},
   * resulting in notifications about changes to the Model or its data.
   * This is handy because the {@link Diagram#model} property setter will automatically call
   * {@link Model#removeChangedListener} on the old model, thereby avoiding any overhead if there are any
   * more changes to the old model and also avoiding a reference to the listener which might cause garbage collection retention.
   * It also will call {@link Model#addChangedListener} on the new model, helping implement the same behavior with the new model.
   *
   * If the property name is a number and if the object being constructed is a {@link Brush},
   * the number and value are added to the Brush by calling {@link Brush#addColorStop}.
   *
   * Otherwise the property name is used as a regular property name on the object being built.
   * This tries to do some property name and value checking:
   * when a property is not defined on the object being built, it will signal an error.
   * Many typos can be found this way that would be ignored by JavaScript code.
   *
   * If the property name begins with an underscore, this will not complain about the property being undefined.
   * Not only is that underscore property set on the object being built, but calls to {@link #copy}
   * will also copy the values of such named properties to the new objects.
   *
   * In the samples and in the intro pages this function is called using the alias `$`.
   * You can use a different short name if you would like to preserve the use of `$` for another JavaScript library.
   * ```js
   *  var $ = go.GraphObject.make;
   *
   *  var diagram =
   *    $(go.Diagram, "myDiagramDiv",
   *      {
   *         // don't initialize some properties until after a new model has been loaded
   *         "InitialLayoutCompleted": loadDiagramProperties,
   *        allowZoom: false,  // don't allow the user to change the diagram's scale
   *        "grid.visible": true,  // display a background grid for the whole diagram
   *        "grid.gridCellSize": new go.Size(20, 20),
   *        // allow double-click in background to create a new node
   *        "clickCreatingTool.archetypeNodeData": { text: "Node" },
   *        // allow Ctrl-G to call the groupSelection command
   *        "commandHandler.archetypeGroupData":
   *            { text: "Group", isGroup: true, color: "blue" },
   *        "toolManager.hoverDelay": 100,  // how quickly tooltips are shown
   *        // mouse wheel zooms instead of scrolls
   *        "toolManager.mouseWheelBehavior": go.ToolManager.WheelZoom,
   *        "commandHandler.copiesTree": true,  // for the copy command
   *        "commandHandler.deletesTree": true, // for the delete command
   *        "draggingTool.dragsTree": true,  // dragging for both move and copy
   *        "draggingTool.isGridSnapEnabled": true,
   *        layout: $(go.TreeLayout,
   *                  { angle: 90, sorting: go.TreeLayout.SortingAscending })
   *      });
   *
   *  diagram.nodeTemplate =
   *    $(go.Node, "Auto",  // or go.Panel.Auto
   *      new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify),
   *      $(go.Shape, "RoundedRectangle",
   *        {
   *          fill: $(go.Brush, "Linear", { 0: "#FEC901", 1: "#FEA200" }),
   *          stroke: "gray",
   *          strokeWidth: 2,
   *          strokeDashArray: [3, 3]
   *        }),
   *      $(go.TextBlock,
   *        { margin: 5, font: "bold 12pt sans-serif" },
   *        new go.Binding("text", "key"))
   *    );
   * ```
   * <p class="boxread">
   * See <a href="../../intro/buildingObjects.html">the Introduction page on building objects</a>
   * for usage information and examples of GraphObject.make.
   * @param {Function|string} type a class function or the name of a class in the `go` namespace,
   * or one of several predefined kinds of {@link Panel}s: `"Button"`, `"TreeExpanderButton"`,
   * `"SubGraphExpanderButton"`, or `"ContextMenuButton"`.
   * @param {...*} initializers zero or more values that initialize the new object,
   * typically an Object with properties whose values are set on the new object,
   * or a JavaScript Array with additional initializer arguments,
   * or a {@link GraphObject} that is added to a {@link Panel},
   * or a {@link Binding} for one of the new object's properties,
   * or a constant value as the initial value of a single property of the new object that
   * is recognized to take that value,
   * or a string that is used as the value of a commonly set property.
   * @return {Object}
   */
  public static make<T extends Adornment>(  // for specific named builders
    cls: ('ToolTip' | 'ContextMenu'),
    ...initializers: Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout |
                     Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout >>
  ): T;

  public static make<T extends Panel>(  // for specific named Panel builders
    cls: ('Button' | 'TreeExpanderButton' | 'SubGraphExpanderButton' | 'ContextMenuButton' | 'PanelExpanderButton' | 'CheckBoxButton' | 'CheckBox'),
    ...initializers: Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout |
                     Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout >>
  ): T;

  public static make<T extends GraphObject>(
    cls: string,  // for named Panel builders
    ...initializers: Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout |
                     Array<string | Partial<GraphObject> & { [p: string]: any } | Binding | EnumValue | RowColumnDefinition | PanelLayout >>
  ): T;

  public static make<CT extends ConstructorType<CT>>(
    cls: CT,
    ...initializers:
      Array<string |
            (Partial<InstanceType<CT>> &
             { [p: string]: any } &
             (InstanceType<CT> extends Diagram
               ? DiagramEventsInterface & { Changed?: ChangedEventHandler; ModelChanged?: ChangedEventHandler; }
               : {})) |
            MakeAllow<CT, GraphObject, Binding> |
            MakeAllow<CT, Panel, GraphObject> |
            MakeAllow<CT, Panel, RowColumnDefinition> |
            MakeAllow<CT, Panel, PanelLayout> |
            MakeAllow<CT, RowColumnDefinition, Binding> |
            MakeAllow<CT, Geometry, PathFigure> |
            MakeAllow<CT, PathFigure, PathSegment> |
            EnumValue |
            HTMLDivElement |
      Array<string |
            (Partial<InstanceType<CT>> &
             { [p: string]: any } &
             (InstanceType<CT> extends Diagram
              ? DiagramEventsInterface & { Changed?: ChangedEventHandler; ModelChanged?: ChangedEventHandler; }
              : {})) |
            MakeAllow<CT, GraphObject, Binding> |
            MakeAllow<CT, Panel, GraphObject> |
            MakeAllow<CT, Panel, RowColumnDefinition> |
            MakeAllow<CT, Panel, PanelLayout> |
            MakeAllow<CT, RowColumnDefinition, Binding> |
            MakeAllow<CT, Geometry, PathFigure> |
            MakeAllow<CT, PathFigure, PathSegment> |
            EnumValue >>
  ): InstanceType<CT>;

  public static make<T>(type: (new () => T) | string, ...initializers: Array<any>): T {
    let args = arguments as any;
    let obj: T = null;
    let classfunc = null;
    if (typeof type === 'function') {
      classfunc = type;
    } else if (typeof type === 'string') {
      const builder = GraphObject.Makers.getValue(type);
      if (typeof builder === 'function') {
        args = Util.copyArray(arguments as any);
        obj = builder(args);  // call the builder
        if (!Util.isObject(obj)) {
          Util.throwError('GraphObject.make invoked object builder "' + type + '", but it did not return an Object');
        }
      } else {
        classfunc = root['go'][type];
      }
    }
    if (obj === null) {
      if (classfunc === undefined || classfunc === null || !classfunc.constructor) {
        Util.throwError('GraphObject.make requires a class function or GoJS class name or name of an object builder, not: ' + type);
      }
      obj = new classfunc();
    }
    let firstarg = 1;
    // handle initialization of Diagram specially, using first arg as its DIV
    if (obj instanceof Diagram && args.length > 1) {
      const dia = obj;
      const arg = args[1];
      if (typeof arg === 'string' || arg instanceof HTMLDivElement) {
        dia._helper.setup(dia, arg);
        firstarg++;
      }
      // Used to have "else" throwing error, but div-less diagrams are now allowed
    }
    // process all additional arguments
    for (let i = firstarg; i < args.length; i++) {
      const arg = args[i];
      if (arg === undefined) {
        Util.throwError('Undefined value at argument ' + i + ' for object being constructed by GraphObject.make: ' + obj);
      } else {
        GraphObject.processMakeArg(obj, arg);
      }
    }
    return obj;
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {*} arg
   */
  private static processMakeArg(obj: ObjectData, arg: any): void {
    if (typeof arg === 'string') {
      GraphObject.processMakeString(obj, arg);
    } else if (arg instanceof GraphObject) {
      if (!(obj instanceof Panel)) {
        Util.throwError('A GraphObject can only be added to a Panel, not to: ' + obj);
      }
      const panel = obj as Panel;
      panel.add(arg);
    } else if (arg instanceof RowColumnDefinition) {
      // obj could be Panel or TableLayout
      // get the real RowColumnDefinition to be initialized
      let rcd: RowColumnDefinition;
      if (arg.isRow && typeof obj.getRowDefinition === 'function') {
        rcd = obj.getRowDefinition(arg.index);
      } else if (!arg.isRow && typeof obj.getColumnDefinition === 'function') {
        rcd = obj.getColumnDefinition(arg.index);
      }
      if (rcd instanceof RowColumnDefinition) {
        // now can initialize the RowColumnDefinition properties
        rcd.copyFrom(arg);
      } else {
        Util.throwError('A RowColumnDefinition can only be added to an object that implements getRowDefinition/getColumnDefinition, not to: ' + obj);
      }
    } else if (arg instanceof EnumValue) {
      GraphObject.processMakeEnum(obj, arg);
    } else if (arg instanceof PanelLayout) {
      const panel = obj as Panel;
      panel.type = arg;
    } else if (arg instanceof Binding) {
      if (obj instanceof GraphObject) {
        obj.bind(arg);
      } else if (obj instanceof RowColumnDefinition) {
        obj.bind(arg);
      } else {
        Util.throwError('A Binding can only be applied to a GraphObject or RowColumnDefinition, not to: ' + obj);
      }
    } else if (arg instanceof AnimationTrigger) {
      if (obj instanceof GraphObject) {
        obj.trigger(arg);
      } else {
        Util.throwError('An AnimationTrigger can only be applied to a GraphObject, not to: ' + obj);
      }
    } else if (arg instanceof PathFigure) {
      if (obj instanceof Geometry) {
        obj.figures.add(arg);
      } else {
        Util.throwError('A PathFigure can only be added to a Geometry, not to: ' + obj);
      }
    } else if (arg instanceof PathSegment) {
      if (obj instanceof PathFigure) {
        obj.segments.add(arg);
      } else {
        Util.throwError('A PathSegment can only be added to a PathFigure, not to: ' + obj);
      }
    } else if (arg instanceof Layout) {
      //    if (obj instanceof MultiLayout) {
      //      obj.add(layout);
      //    } else
      if (obj instanceof Diagram) {
        obj.layout = arg;
      } else if (obj instanceof Group) {
        obj.layout = arg;
      } else {
        Util.throwError('A Layout can only be assigned to a Diagram or a Group, not to: ' + obj);
      }
    } else if (Array.isArray(arg)) {
      for (let i = 0; i < arg.length; i++) {
        const subarg = arg[i];
        GraphObject.processMakeArg(obj, subarg);
      }
    } else if (typeof arg === 'object' && arg !== null) {  // plain JavaScript object: add its properties to the OBJ
      GraphObject.processMakeObject(obj, arg);
    } else {
      Util.throwError('Unknown initializer \"' + arg + '\" for object being constructed by GraphObject.make: ' + obj);
    }
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {string} arg
   */
  private static processMakeString(obj: ObjectData, arg: string): void {
    if (obj instanceof TextBlock) {
      obj.text = arg;
    } else if (obj instanceof Shape) {
      obj.figure = arg;
    } else if (obj instanceof Picture) {
      obj.source = arg;
    } else if (obj instanceof Panel) {
      const enm = Panel.Layouts.getValue(arg);
      if (enm !== null) {
        obj.type = enm;
      } else if (Debug) {
        Util.throwError('Unknown Panel type as an argument to GraphObject.make: ' + arg  + '. If building from source, you may need to call Panel.definePanelLayout.');
      }
    } else if (obj instanceof Brush) {
      const enm = EnumValue.findName(Brush, arg);
      if (enm !== null) {
        obj.type = enm;
      } else {
        Util.throwError('Unknown Brush type as an argument to GraphObject.make: ' + arg);
      }
    } else if (obj instanceof Geometry) {
      const enm = EnumValue.findName(Geometry, arg);
      if (enm !== null) {
        obj.type = enm;
      } else if (Debug) {
        Util.throwError('Unknown Geometry type as an argument to GraphObject.make: ' + arg);
      }
    } else if (obj instanceof PathSegment) {
      const enm = EnumValue.findName(PathSegment, arg);
      if (enm !== null) {
        obj.type = enm;
      } else if (Debug) {
        Util.throwError('Unknown PathSegment type as an argument to GraphObject.make: ' + arg);
      }
    } else if (Debug) {
      Util.throwError('Unable to use a string as an argument to GraphObject.make: ' + arg);
    }
  }

  /**
   * @hidden @internal
   * @param {*} obj
   * @param {EnumValue} arg
   */
  private static processMakeEnum(obj: any, arg: EnumValue): void {
    if (typeof obj.acceptEnum === 'function') {
      obj.acceptEnum(arg);
    } else {
      Util.throwEnumPropertyValueError(obj, arg);
    }
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {Object} arg
   */
  private static processMakeObject(obj: ObjectData, arg: ObjectData): void {
    if (obj instanceof Brush) {
      // strip out any colorStop definitions in ARG and add them explicitly to the Brush
      const props = new PropertyCollection();
      for (const p in arg) {
        const frac = parseFloat(p);
        if (!isNaN(frac)) {
          obj.addColorStop(frac, arg[p]);
        } else {
          props[p] = arg[p];
        }
      }
      Diagram.setCheckedProperties(obj, props);
    } else if (obj instanceof RowColumnDefinition) {
      if (arg['row'] !== undefined) {
        const idx = arg['row'];
        if (idx === undefined || idx === null || idx === Infinity || isNaN(idx) || idx < 0) {
          Util.throwError('Must specify non-negative integer row for RowColumnDefinition ' + arg + ', not: ' + idx);
        }
        obj.isRow = true;
        obj.index = idx;
      } else if (arg['column'] !== undefined) {
        const idx = arg['column'];
        if (idx === undefined || idx === null || idx === Infinity || isNaN(idx) || idx < 0) {
          Util.throwError('Must specify non-negative integer column for RowColumnDefinition ' + arg + ', not: ' + idx);
        }
        obj.isRow = false;
        obj.index = idx;
      }
      // remove any 'row' or 'column' property from the argument Object
      const props = new PropertyCollection();
      for (const p in arg) {
        if (p === 'row' || p === 'column') continue;
        props[p] = arg[p];
      }
      Diagram.setCheckedProperties(obj, props);
    } else {
      Diagram.setCheckedProperties(obj, arg);
    }
  }

  /**
   * @hidden @internal
   * This static function returns a Map of named GraphObject Makers.
   * @suppress {checkTypes} for in GraphObject.Makers, closure complains that its a struct
   * @return {Map.<string,function(Array.<*>)>}
   */
  public static getBuilders(): Map<string, ((a: Array<any>) => ObjectData)> {
    const map = new Map<string, ((a: Array<any>) => ObjectData)>();
    for (const name in GraphObject.Makers) {
      if (name === name.toLowerCase()) continue;
      const f = GraphObject.Makers.getValue(name);
      if (typeof f === 'function') map.add(name, f);
    }
    map.freeze();
    return map;
  }

  /**
   * This static function defines a named function that {@link GraphObject.make} can use to build objects.
   * Once this is called one can use the name as the first argument for {@link GraphObject.make}.
   * Names are case sensitive.
   *
   * The second argument must be a function that returns a newly created object, typically a {@link GraphObject}.
   * It is commonplace for that object to be a {@link Panel} holding a newly created visual tree of GraphObjects.
   * The function receives as its only argument an Array that is holds all of the arguments that are being
   * passed to {@link GraphObject.make}, which it may modify in order to change the arguments that GraphObject.make receives.
   *
   * Predefined builder names include: `"Button"`, `"TreeExpanderButton"`, `"SubGraphExpanderButton"`,
   * `"PanelExpanderButton"`, and `"ContextMenuButton"`.
   * The implementation of these builders is provided by <a href="../../extensions/Buttons.js">Buttons.js</a>
   * in the Extensions directory.
   * @param {string} name a capitalized name; must not be `""` or `"None"`
   * @param {function(Array.<*>):Object} func that takes an Array of `GraphObject.make` arguments and returns a new object
   * @since 1.5
   */
  public static defineBuilder(name: string, func: ((a: Array<any>) => ObjectData)): void {
    Util.checkPrimitive(name, 'string', GraphObject, 'defineBuilder:name');
    Util.checkPrimitive(func, 'function', GraphObject, 'defineBuilder:func');
    const low = name.toLowerCase();
    if (Debug) {
      if (name === '' || low === 'none' || name === low) Util.throwError('Shape.defineFigureGenerator name must not be empty or None or all-lower-case: ' + name);
    }
    GraphObject.Makers.add(name, func);
    // no support for case-insensitive names?
  }

  /**
   * This static function returns the first argument from the arguments array passed
   * to a {@link GraphObject.defineBuilder} function by {@link GraphObject.make}.
   * By default this requires the first argument to be a string,
   * but you can provide a predicate to determine whether the argument is suitable.
   * @param {Array.<*>} args the arguments Array passed to the builder function;
   *        this may be modified if an acceptable argument is found and returned
   * @param {*=} defval the default value to return if the argument is optional and not present as the first argument;
   *        otherwise throw an error when the argument is not there
   * @param {function(*):boolean|null=} pred a predicate to determine the acceptability of the argument;
   *        the default predicate checks whether the argument is a string
   * @return {*}
   * @since 1.5
   */
  public static takeBuilderArgument(args: Array<any>, defval?: any, pred?: ((a: any) => boolean) | null): any {
    if (pred === undefined) pred = null;
    const arg = args[1];
    if ((typeof pred === 'function') ? pred(arg) : (typeof arg === 'string')) {
      args.splice(1, 1);  // take out the argument from the ARGS list
      return arg;
    }
    if (defval === undefined) throw new Error('no ' + (typeof pred === 'function' ? 'satisfactory' : 'string') + ' argument for GraphObject builder ' + args[0]);
    return defval;
  }

  /**
   * @hidden @internal
   */
  public static Makers: Map<string, ((a: Array<any>) => any)> = null;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    GraphObject.Makers = new Map<string, ((a: Array<any>) => ObjectData)>();

    // A "Button" is a Panel that has a Shape surrounding some content
    // and that has mouseEnter/mouseLeave behavior to highlight the button.
    // The content of the button, whether a TextBlock or a Picture or a complicated Panel,
    // must be supplied by the caller.
    // The caller must also provide a click event handler.
    // Typical usage:
    //    $("Button",
    //      $(go.TextBlock, "Click me!"),  // the content is just the text label
    //      { click: function(e, obj) { alert("I was clicked"); } }
    //    )

    GraphObject.defineBuilder('Button', (args: any): Panel => {
      // default colors for 'Button' shape
      const buttonFillNormal = '#F5F5F5';
      const buttonStrokeNormal = '#BDBDBD';
      const buttonFillOver = '#E0E0E0';
      const buttonStrokeOver = '#9E9E9E';
      const buttonFillPressed = '#BDBDBD'; // set to null for no button pressed effects
      const buttonStrokePressed = '#9E9E9E';
      const buttonFillDisabled = '#E5E5E5';

      // padding inside the ButtonBorder to match sizing from previous versions
      const paddingHorizontal = 2.76142374915397;
      const paddingVertical = 2.761423749153969;

      const button =
        GraphObject.make(Panel, Panel.Auto,
          {
            isActionable: true,  // needed so that the ActionTool intercepts mouse events
            enabledChanged: (btn: GraphObject, enabled: boolean): void => {
              if (btn instanceof Panel) {
                const shape = btn.findObject('ButtonBorder') as Shape;
                if (shape !== null) {
                  shape.fill = enabled ? (btn as any)['_buttonFillNormal'] : (btn as any)['_buttonFillDisabled'];
                }
              }
            },
            cursor: 'pointer',
            // save these values for the mouseEnter and mouseLeave event handlers
            '_buttonFillNormal': buttonFillNormal,
            '_buttonStrokeNormal': buttonStrokeNormal,
            '_buttonFillOver': buttonFillOver,
            '_buttonStrokeOver': buttonStrokeOver,
            '_buttonFillPressed': buttonFillPressed,
            '_buttonStrokePressed': buttonStrokePressed,
            '_buttonFillDisabled': buttonFillDisabled
          },
          GraphObject.make(Shape,  // the border
            {
              name: 'ButtonBorder',
              figure: 'RoundedRectangle',
              spot1: new Spot(0, 0, paddingHorizontal, paddingVertical),
              spot2: new Spot(1, 1, -paddingHorizontal, -paddingVertical),
              parameter1: 2,
              parameter2: 2,
              fill: buttonFillNormal,
              stroke: buttonStrokeNormal
            }
          )
        ) as Panel;

      // There's no GraphObject inside the button shape -- it must be added as part of the button definition.
      // This way the object could be a TextBlock or a Shape or a Picture or arbitrarily complex Panel.

      // mouse-over behavior
      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       * @param {GraphObject} prev
       */
      button.mouseEnter = (e: InputEvent, btn: GraphObject, prev: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        const shape = btn.findObject('ButtonBorder');  // the border Shape
        if (shape instanceof Shape) {
          let brush = (btn as any)['_buttonFillOver'];
          (btn as any)['_buttonFillNormal'] = shape.fill;
          shape.fill = brush;
          brush = (btn as any)['_buttonStrokeOver'];
          (btn as any)['_buttonStrokeNormal'] = shape.stroke;
          shape.stroke = brush;
        }
      };

      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       * @param {GraphObject} prev
       */
      button.mouseLeave = (e: InputEvent, btn: GraphObject, prev: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        const shape = btn.findObject('ButtonBorder');  // the border Shape
        if (shape instanceof Shape) {
          shape.fill = (btn as any)['_buttonFillNormal'];
          shape.stroke = (btn as any)['_buttonStrokeNormal'];
        }
      };

      button.actionDown = (e: InputEvent, btn: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        if ((btn as any)['_buttonFillPressed'] === null) return;
        if (e.button !== 0) return;
        const shape = btn.findObject('ButtonBorder');  // the border Shape
        if (shape instanceof Shape) {
          const diagram = e.diagram;
          const oldskip = diagram.skipsUndoManager;
          diagram.skipsUndoManager = true;
          let brush = (btn as any)['_buttonFillPressed'];
          (btn as any)['_buttonFillOver'] = shape.fill;
          shape.fill = brush;
          brush = (btn as any)['_buttonStrokePressed'];
          (btn as any)['_buttonStrokeOver'] = shape.stroke;
          shape.stroke = brush;
          diagram.skipsUndoManager = oldskip;
        }
      };

      button.actionUp = (e: InputEvent, btn: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        if ((btn as any)['_buttonFillPressed'] === null) return;
        if (e.button !== 0) return;
        const shape = btn.findObject('ButtonBorder');  // the border Shape
        if (shape instanceof Shape) {
          const diagram = e.diagram;
          const oldskip = diagram.skipsUndoManager;
          diagram.skipsUndoManager = true;
          if (overButton(e, btn)) {
            shape.fill = (btn as any)['_buttonFillOver'];
            shape.stroke = (btn as any)['_buttonStrokeOver'];
          } else {
            shape.fill = (btn as any)['_buttonFillNormal'];
            shape.stroke = (btn as any)['_buttonStrokeNormal'];
          }
          diagram.skipsUndoManager = oldskip;
        }
      };

      button.actionCancel = (e: InputEvent, btn: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        if ((btn as any)['_buttonFillPressed'] === null) return;
        const shape = btn.findObject('ButtonBorder');  // the border Shape
        if (shape instanceof Shape) {
          const diagram = e.diagram;
          const oldskip = diagram.skipsUndoManager;
          diagram.skipsUndoManager = true;
          if (overButton(e, btn)) {
            shape.fill = (btn as any)['_buttonFillOver'];
            shape.stroke = (btn as any)['_buttonStrokeOver'];
          } else {
            shape.fill = (btn as any)['_buttonFillNormal'];
            shape.stroke = (btn as any)['_buttonStrokeNormal'];
          }
          diagram.skipsUndoManager = oldskip;
        }
      };

      button.actionMove = (e: InputEvent, btn: GraphObject): void => {
        if (!btn.isEnabledObject()) return;
        if (!(btn instanceof Panel)) return;
        if ((btn as any)['_buttonFillPressed'] === null) return;
        const diagram = e.diagram;
        if (diagram.firstInput.button !== 0) return;
        diagram.currentTool.standardMouseOver();
        if (overButton(e, btn)) {
          const shape = btn.findObject('ButtonBorder');
          if (shape instanceof Shape) {
            const oldskip = diagram.skipsUndoManager;
            diagram.skipsUndoManager = true;
            let brush = (btn as any)['_buttonFillPressed'];
            if (shape.fill !== brush) shape.fill = brush;
            brush = (btn as any)['_buttonStrokePressed'];
            if (shape.stroke !== brush) shape.stroke = brush;
            diagram.skipsUndoManager = oldskip;
          }
        }
      };

      const overButton = (e: InputEvent, btn: Panel): boolean => {
        const over = e.diagram.findObjectAt(
          e.documentPoint,
          (x: GraphObject): GraphObject => {
            while (x.panel !== null) {
              if (x.isActionable) return x;
              x = x.panel;
            }
            return x;
          },
          (x: GraphObject): boolean => x === btn
        );
        return over !== null;
      };

      return button;
    });

    // This is a complete Button that you can have in a Node template
    // to allow the user to collapse/expand the subtree beginning at that Node.
    GraphObject.defineBuilder('TreeExpanderButton', (args: any): Panel => {
      const button =
        GraphObject.make('Button',
          { // set these values for the isTreeExpanded binding conversion
            '_treeExpandedFigure': 'MinusLine',
            '_treeCollapsedFigure': 'PlusLine'
          },
          GraphObject.make(Shape,  // the icon
            {
              name: 'ButtonIcon',
              figure: 'MinusLine',  // default value for isTreeExpanded is true
              stroke: '#424242',
              strokeWidth: 2,
              desiredSize: Size.EightSize
            },
            // bind the Shape.figure to the Node.isTreeExpanded value using this converter:
            new Binding('figure', 'isTreeExpanded',
              (exp: boolean, shape: Shape): string => {
                const but = shape.panel;
                return exp ? (but as any)['_treeExpandedFigure'] : (but as any)['_treeCollapsedFigure'];
              }
            ).ofObject()
          ),
          // assume initially not visible because there are no links coming out
          { visible: false },
          // bind the button visibility to whether it's not a leaf node
          new Binding('visible', 'isTreeLeaf',
            (leaf: boolean): boolean => !leaf
          ).ofObject()
        ) as Panel;

      // tree expand/collapse behavior
      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       */
      button.click = (e: InputEvent, btn: GraphObject): void => {
        let node = btn.part;
        if (node instanceof Adornment) node = node.adornedPart;
        if (!(node instanceof Node)) return;
        const diagram = node.diagram;
        if (diagram === null) return;
        const cmd = diagram.commandHandler;
        if (node.isTreeExpanded) {
          if (!cmd.canCollapseTree(node)) return;
        } else {
          if (!cmd.canExpandTree(node)) return;
        }
        e.handled = true;
        if (node.isTreeExpanded) {
          cmd.collapseTree(node);
        } else {
          cmd.expandTree(node);
        }
      };

      return button;
    });

    // This is a complete Button that you can have in a Group template
    // to allow the user to collapse/expand the subgraph that the Group holds.
    GraphObject.defineBuilder('SubGraphExpanderButton', (args: any): Panel => {
      const button =
        GraphObject.make('Button',
          { // set these values for the isSubGraphExpanded binding conversion
            '_subGraphExpandedFigure': 'MinusLine',
            '_subGraphCollapsedFigure': 'PlusLine'
          },
          GraphObject.make(Shape,  // the icon
            {
              name: 'ButtonIcon',
              figure: 'MinusLine',  // default value for isSubGraphExpanded is true
              stroke: '#424242',
              strokeWidth: 2,
              desiredSize: Size.EightSize
            },
            // bind the Shape.figure to the Group.isSubGraphExpanded value using this converter:
            new Binding('figure', 'isSubGraphExpanded',
              (exp: boolean, shape: Shape): string => {
                const but = shape.panel;
                return exp ? (but as any)['_subGraphExpandedFigure'] : (but as any)['_subGraphCollapsedFigure'];
              }
            ).ofObject()
          )
        ) as Panel;

      // subgraph expand/collapse behavior
      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       */
      button.click = (e: InputEvent, btn: GraphObject): void => {
        let group = btn.part;
        if (group instanceof Adornment) group = group.adornedPart;
        if (!(group instanceof Group)) return;
        const diagram = group.diagram;
        if (diagram === null) return;
        const cmd = diagram.commandHandler;
        if (group.isSubGraphExpanded) {
          if (!cmd.canCollapseSubGraph(group)) return;
        } else {
          if (!cmd.canExpandSubGraph(group)) return;
        }
        e.handled = true;
        if (group.isSubGraphExpanded) {
          cmd.collapseSubGraph(group);
        } else {
          cmd.expandSubGraph(group);
        }
      };

      return button;
    });

    // This is just an "Auto" Adornment that can hold some contents within a light gray, shadowed box.

    // Typical usage:
    //   toolTip:
    //     $("ToolTip",
    //       $(go.TextBlock, . . .)
    //     )
    GraphObject.defineBuilder('ToolTip', (args: any): Adornment => {
      const ad = GraphObject.make(Adornment, Panel.Auto,
        {
          isShadowed: true,
          shadowColor: 'rgba(0, 0, 0, .4)',
          shadowOffset: new Point(0, 3),
          shadowBlur: 5
        },
        GraphObject.make(Shape,
          {
            name: 'Border',
            figure: 'RoundedRectangle',
            parameter1: 1,
            parameter2: 1,
            fill: '#F5F5F5',
            stroke: '#F0F0F0',
            spot1: new Spot(0, 0, 4, 6),
            spot2: new Spot(1, 1, -4, -4)
          }
        )
      );
      return ad;
    });

    // This is just a "Vertical" Adornment that can hold some "ContextMenuButton"s.

    // Typical usage:
    //   contextMenu:
    //     $("ContextMenu",
    //       $("ContextMenuButton",
    //         $(go.TextBlock, . . .),
    //         { click: . . .}
    //       ),
    //       $("ContextMenuButton", . . .)
    //     )
    GraphObject.defineBuilder('ContextMenu', (args: any): Adornment => {
      const ad = GraphObject.make(Adornment, Panel.Vertical,
        {
          background: '#F5F5F5',
          isShadowed: true,
          shadowColor: 'rgba(0, 0, 0, .4)',
          shadowOffset: new Point(0, 3),
          shadowBlur: 5
        },
        // don't set the background if the ContextMenu is adorning something and there's a Placeholder
        new Binding('background', '', (obj: Adornment) => {
          const part = obj.adornedPart;
          if (part !== null && obj.placeholder !== null) return null;
          return '#F5F5F5';
        })
      );
      return ad;
    });

    // This just holds the "ButtonBorder" Shape that acts as the border
    // around the button contents, which must be supplied by the caller.
    // The button contents are usually a TextBlock or Panel consisting of a Shape and a TextBlock.
    GraphObject.defineBuilder('ContextMenuButton', (args: any): Panel => {
      const button = GraphObject.make('Button') as Panel;
      button.stretch = GraphObject.Horizontal;
      const border = button.findObject('ButtonBorder');
      if (border instanceof Shape) {
        border.figure = 'Rectangle';
        border.strokeWidth = 0;
        border.spot1 = new Spot(0, 0, 2, 3);
        border.spot2 = new Spot(1, 1, -2, -2);
      }
      return button;
    });

    // This button is used to toggle the visibility of a GraphObject named
    // by the second argument to GraphObject.make.  If the second argument is not present
    // or if it is not a string, this assumes that the element name is "COLLAPSIBLE".
    // You can only control the visibility of one element in a Part at a time,
    // although that element might be an arbitrarily complex Panel.
    GraphObject.defineBuilder('PanelExpanderButton', (args: any): Panel => {
      const eltname = GraphObject.takeBuilderArgument(args, 'COLLAPSIBLE');

      const button: Panel =
        GraphObject.make('Button',
          { // set these values for the button's look
            '_buttonExpandedFigure': 'M0 0 M0 6 L4 2 8 6 M8 8',
            '_buttonCollapsedFigure': 'M0 0 M0 2 L4 6 8 2 M8 8',
            '_buttonFillNormal': 'rgba(0, 0, 0, 0)',
            '_buttonStrokeNormal': null,
            '_buttonFillOver': 'rgba(0, 0, 0, .2)',
            '_buttonStrokeOver': null,
            '_buttonFillPressed': 'rgba(0, 0, 0, .4)',
            '_buttonStrokePressed': null
          },
          GraphObject.make(Shape,
            { name: 'ButtonIcon', strokeWidth: 2 },
            new Binding('geometryString', 'visible',
              (vis: boolean): string => vis ? (button as any)['_buttonExpandedFigure'] : (button as any)['_buttonCollapsedFigure']
            ).ofObject(eltname)
          )
        ) as Panel;

      const border = button.findObject('ButtonBorder');
      if (border instanceof Shape) {
        border.stroke = null;
        border.fill = 'rgba(0, 0, 0, 0)';
      }

      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       */
      button.click = (e: InputEvent, btn: GraphObject): void => {
        const diagram = btn.diagram;
        if (diagram === null) return;
        if (diagram.isReadOnly) return;
        let elt = btn.findTemplateBinder();
        if (elt === null) elt = btn.part;
        if (elt !== null) {
          const pan = elt.findObject(eltname);
          if (pan !== null) {
            e.handled = true;
            diagram.startTransaction('Collapse/Expand Panel');
            pan.visible = !pan.visible;
            diagram.commitTransaction('Collapse/Expand Panel');
          }
        }
      };

      return button;
    });

    // Define a common checkbox button; the first argument is the name of the data property
    // to which the state of this checkbox is data bound.  If the first argument is not a string,
    // it raises an error.  If no data binding of the checked state is desired,
    // pass an empty string as the first argument.
    // Examples:
    // $("CheckBoxButton", "dataPropertyName", ...)
    // or:
    // $("CheckBoxButton", "", { "_doClick": function(e, obj) { alert("clicked!"); } })
    GraphObject.defineBuilder('CheckBoxButton', (args: any): Panel => {
      // process the one required string argument for this kind of button
      const propname = GraphObject.takeBuilderArgument(args);

      const button =
        GraphObject.make('Button',
          { desiredSize: new Size(14, 14) },
          GraphObject.make(Shape,
            {
              name: 'ButtonIcon',
              geometryString: 'M0 0 M0 8.85 L4.9 13.75 16.2 2.45 M16.2 16.2',  // a 'check' mark
              strokeWidth: 2,
              stretch: GraphObject.Fill,  // this Shape expands to fill the Button
              geometryStretch: GraphObject.Uniform,  // the check mark fills the Shape without distortion
              visible: false  // visible set to false: not checked, unless data.PROPNAME is true
            },
            // create a data Binding only if PROPNAME is supplied and not the empty string
            (propname !== '' ? new Binding('visible', propname).makeTwoWay() : [])
          )
        ) as Panel;

      /**
       * @hidden @internal
       * @param {InputEvent} e
       * @param {GraphObject} btn
       */
      button.click = (e: InputEvent, btn: GraphObject): void => {
        if (!(btn instanceof Panel)) return;
        const diagram = e.diagram;
        if (diagram === null || diagram.isReadOnly) return;
        if (propname !== '' && diagram.model.isReadOnly) return;
        e.handled = true;
        const shape = btn.findObject('ButtonIcon');
        diagram.startTransaction('checkbox');
        shape.visible = !shape.visible;  // this toggles data.checked due to TwoWay Binding
        // support extra side-effects without clobbering the click event handler:
        if (typeof (btn as any)['_doClick'] === 'function') (btn as any)['_doClick'](e, btn);
        diagram.commitTransaction('checkbox');
      };

      return button;
    });

    // This defines a whole check-box -- including both a "CheckBoxButton" and whatever you want as the check box label.
    // Note that mouseEnter/mouseLeave/click events apply to everything in the panel, not just in the "CheckBoxButton".
    // Examples:
    // $("CheckBox", "aBooleanDataProperty", $(go.TextBlock, "the checkbox label"))
    // or
    // $("CheckBox", "someProperty", $(go.TextBlock, "A choice"),
    //   { "_doClick": function(e, obj) { ... perform extra side-effects ... } })
    GraphObject.defineBuilder('CheckBox', (args: any): Panel => {
      // process the one required string argument for this kind of button
      const propname = GraphObject.takeBuilderArgument(args);

      const button =
        GraphObject.make('CheckBoxButton', propname,  // bound to this data property
          {
            name: 'Button',
            isActionable: false, // actionable is set on the whole horizontal panel
            margin: new Margin(0, 1, 0, 0)
          }
        ) as Panel;

      const box =
        GraphObject.make(Panel, 'Horizontal',
          button,
          {
            isActionable: true,
            cursor: button.cursor,
            margin: 1,
            // transfer CheckBoxButton properties over to this new CheckBox panel
            '_buttonFillNormal': (button as any)['_buttonFillNormal'],
            '_buttonStrokeNormal': (button as any)['_buttonStrokeNormal'],
            '_buttonFillOver': (button as any)['_buttonFillOver'],
            '_buttonStrokeOver': (button as any)['_buttonStrokeOver'],
            '_buttonFillPressed': (button as any)['_buttonFillPressed'],
            '_buttonStrokePressed': (button as any)['_buttonStrokePressed'],
            '_buttonFillDisabled': (button as any)['_buttonFillDisabled'],
            mouseEnter: button.mouseEnter,
            mouseLeave: button.mouseLeave,
            actionDown: button.actionDown,
            actionUp: button.actionUp,
            actionCancel: button.actionCancel,
            actionMove: button.actionMove,
            click: button.click,
            // also save original Button behavior, for potential use in a Panel.click event handler
            '_buttonClick': button.click
          }
        ) as Panel;
      // avoid potentially conflicting event handlers on the "CheckBoxButton"
      button.mouseEnter = null;
      button.mouseLeave = null;
      button.actionDown = null;
      button.actionUp = null;
      button.actionCancel = null;
      button.actionMove = null;
      button.click = null;
      return box;
    });

    GraphObject['None'] = GraphObject.None = new EnumValue(GraphObject, 'None', 0);
    GraphObject['Default'] = GraphObject.Default = new EnumValue(GraphObject, 'Default', 0);
    GraphObject['Vertical'] = GraphObject.Vertical = new EnumValue(GraphObject, 'Vertical', 4);
    GraphObject['Horizontal'] = GraphObject.Horizontal = new EnumValue(GraphObject, 'Horizontal', 5);
    GraphObject['Fill'] = GraphObject.Fill = new EnumValue(GraphObject, 'Fill', 3);
    GraphObject['Uniform'] = GraphObject.Uniform = new EnumValue(GraphObject, 'Uniform', 1);
    GraphObject['UniformToFill'] = GraphObject.UniformToFill = new EnumValue(GraphObject, 'UniformToFill', 2);
    GraphObject['FlipVertical'] = GraphObject.FlipVertical = new EnumValue(GraphObject, 'FlipVertical', 1);
    GraphObject['FlipHorizontal'] = GraphObject.FlipHorizontal = new EnumValue(GraphObject, 'FlipHorizontal', 2);
    GraphObject['FlipBoth'] = GraphObject.FlipBoth = new EnumValue(GraphObject, 'FlipBoth', 3);
    GraphObject['make'] = GraphObject.make;
    GraphObject['getBuilders'] = GraphObject.getBuilders;
    GraphObject['defineBuilder'] = GraphObject.defineBuilder;
    GraphObject['takeBuilderArgument'] = GraphObject.takeBuilderArgument;

  } // end initstatic

} // #end GraphObject
GraphObject['className'] = 'GraphObject';
GraphObject.initStatic();

/**
 * @hidden @internal
 * @unrestricted
 */
export class GraphObjectEventHandlers {
  /** @hidden @internal */ public static className = 'GraphObjectEventHandlers';
  /** @hidden @internal */ public _click: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _doubleClick: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _contextClick: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseEnter: ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseLeave: ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseOver: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseHover: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseHold: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseDragEnter: ((e: InputEvent, thisObj: GraphObject, prevObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseDragLeave: ((e: InputEvent, thisObj: GraphObject, nextObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _mouseDrop: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _actionDown: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _actionMove: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _actionUp: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _actionCancel: ((e: InputEvent, thisObj: GraphObject) => void) | null;
  /** @hidden @internal */ public _toolTip: Adornment | HTMLInfo | null;
  /** @hidden @internal */ public _contextMenu: Adornment | HTMLInfo | null;
  /** @hidden @internal */ public _enabledChanged: ((a: GraphObject, b: boolean) => void) | null;

  constructor() {
    this._click = null;
    this._doubleClick = null;
    this._contextClick = null;

    this._mouseEnter = null;
    this._mouseLeave = null;
    this._mouseOver = null;
    this._mouseHover = null;
    this._mouseHold = null;
    this._mouseDragEnter = null;
    this._mouseDragLeave = null;
    this._mouseDrop = null;

    this._actionDown = null;
    this._actionMove = null;
    this._actionUp = null;
    this._actionCancel = null;

    this._toolTip = null;
    this._contextMenu = null;

    this._enabledChanged = null;
  }

  /**
   * @hidden @internal
   * @this{GraphObjectEventHandlers}
   * @return {GraphObjectEventHandlers}
   */
  public copy(): GraphObjectEventHandlers {
    const copy = new GraphObjectEventHandlers();
    copy._click = this._click;
    copy._doubleClick = this._doubleClick;
    copy._contextClick = this._contextClick;

    copy._mouseEnter = this._mouseEnter;
    copy._mouseLeave = this._mouseLeave;
    copy._mouseOver = this._mouseOver;
    copy._mouseHover = this._mouseHover;
    copy._mouseHold = this._mouseHold;
    copy._mouseDragEnter = this._mouseDragEnter;
    copy._mouseDragLeave = this._mouseDragLeave;
    copy._mouseDrop = this._mouseDrop;

    copy._actionDown = this._actionDown;
    copy._actionMove = this._actionMove;
    copy._actionUp = this._actionUp;
    copy._actionCancel = this._actionCancel;

    copy._toolTip = this._toolTip;
    copy._contextMenu = this._contextMenu;

    copy._enabledChanged = this._enabledChanged;
    return copy;
  }

} // #end GraphObjectEventHandlers
GraphObjectEventHandlers['className'] = 'GraphObjectEventHandlers';


// import { CanvasSurfaceContext } from './abstractCanvas';
// import { Animation, AnimationTrigger } from './animationManager';
// import { Brush, BrushLike } from './brush';
// import { ChangedEvent, ChangedEventHandler } from './changedEvent';
// import { Diagram } from './diagram';
// import { Geo } from './geo';
// import { Geometry, PathFigure, PathSegment } from './geometry';
// import { HTMLInfo } from './htmlInfo';
// import { DiagramEventsInterface, InputEvent } from './inputEvent';
// import { Layer } from './layer';
// import { Layout } from './layout';
// import { Margin, MarginLike } from './margin';
// import { Binding } from './model';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Knot, Link, LinkSettings } from './parts';
// import { Node } from './parts';
// import { Adornment, Group, Part } from './parts';
// import { Picture } from './picture';
// import { Point } from './point';
// import { PropertyCollection } from './propertyCollection';
// import { Rect } from './rect';
// import { ObjectData, root } from './root';
// import { RowColumnDefinition } from './rowColumnDefinition';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { IContext } from './surfaceInterfaces';
// import { SGradient } from './svgContext';
// import { TextBlock } from './textBlock';
// import { Transform } from './transform';




/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
STransform
SGradient
SVGContext
SVGSurface
*/

// import { Brush, BrushLike } from './brush';
// import { Set } from './collections';
// import { Diagram, SvgRendererOptions } from './diagram';
// import { Geo } from './geo';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { Part } from './parts';
// import { PropertyCollection } from './propertyCollection';
// import { ObjectData, root } from './root';
// import { Shape } from './shape';
// import { IContext, ISurface } from './surfaceInterfaces';
// import { TextBlock } from './textBlock';
// import { Util } from './util';

// Simple transform
/**
 * @hidden @internal
 * @unrestricted
 */
export class STransform {
  /** @hidden @internal */ public static className = 'STransform';
  /** @hidden @internal */ public m: Array<number>;

  constructor() {
    /** @hidden @internal */
    this.m = [1, 0, 0, 1, 0, 0];
  }

  /**
   * @hidden @internal
   */
  public copy(): STransform {
    const s = new STransform();
    s.m[0] = this.m[0];
    s.m[1] = this.m[1];
    s.m[2] = this.m[2];
    s.m[3] = this.m[3];
    s.m[4] = this.m[4];
    s.m[5] = this.m[5];
    return s;
  }

  /**
   * @hidden @internal
   * @param {number} x
   * @param {number} y
   */
  public translate(x: number, y: number): void {
    this.m[4] += this.m[0] * x + this.m[2] * y;
    this.m[5] += this.m[1] * x + this.m[3] * y;
  }

  /**
   * @hidden @internal
   * @param {number} sx
   * @param {number} sy
   */
  public scale(sx: number, sy: number): void {
    this.m[0] *= sx;
    this.m[1] *= sx;
    this.m[2] *= sy;
    this.m[3] *= sy;
  }

} // #end STransform
STransform['className'] = 'STransform';

// Simple gradient
/**
 * @hidden
 * @unrestricted
 */
export class SGradient {
  /** @hidden @internal */ public static className = 'SGradient';
  /** @hidden @internal */ public type: string | SGradient;
  /** @hidden @internal */ public x1: number;
  /** @hidden @internal */ public y1: number;
  /** @hidden @internal */ public r1: number;
  /** @hidden @internal */ public x2: number;
  /** @hidden @internal */ public y2: number;
  /** @hidden @internal */ public r2: number;
  /** @hidden @internal */ public colors: Array<ObjectData>;
  /** @hidden @internal */ public pattern: any;

  /**
   * @param {string} aType
   */
  constructor(aType: string) {
    this.type = aType;
    this.x1 = 0;
    this.y1 = 0;
    this.r1 = 0;
    this.x2 = 0;
    this.y2 = 0;
    this.r2 = 0;
    this.colors = [];

    // only used in patterns
    this.pattern = null;
  }

  /**
   * @hidden @internal
   * @this {SGradient}
   * @param {boolean} offset
   * @param {string} color
   */
  public addColorStop(offset: boolean, color: string): void {
    this.colors.push({ offset: offset, color: color });
  }

} // #end SGradient
SGradient['className'] = 'SGradient';

/**
 * @hidden @internal
 * @unrestricted
 */
export class SVGSurface implements ISurface {
  /** @hidden @internal */ public static className = 'SVGSurface';
  /** @hidden @internal */ public domElement: SVGElement;
  /** @hidden @internal */ public context: SVGContext;
  /** @hidden @internal */ public ownerDocument: Document;
  /** @hidden @internal */ public elementFinished: ((a: GraphObject, b: SVGElement) => void) | null; // only used in SVG
  /** @hidden @internal */ private ns: string;

  /**
   * @param {Diagram} diagram can be null
   * @param {Document=} ownerDocument
   */
  constructor(diagram: Object | null, ownerDocument?: Document) {
    const doc = (ownerDocument === undefined) ? root.document : ownerDocument;
    this.ownerDocument = doc;
    this.ns = 'http://www.w3.org/2000/svg';
    const attrlink = 'http://www.w3.org/2000/xmlns/';
    if (doc !== undefined) {
      this.domElement = this.makeElement('svg', {
        'width': 1 + 'px',
        'height': 1 + 'px',
        'viewBox': '0 0 ' + 1 + ' ' + 1
      });
      this.domElement.setAttributeNS(attrlink, 'xmlns', this.ns);
      this.domElement.setAttributeNS(attrlink, 'xmlns:xlink', 'http://www.w3.org/1999/xlink');
    }
    this.elementFinished = null;
    this.context = new SVGContext(this);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurface}
   * @param {number} pixelWidth
   * @param {number} pixelHeight
   * @param {number} width
   * @param {number} height
   * @returns {boolean} true if the surface changed size
   */
  public resize(pixelWidth: number, pixelHeight: number, width: number, height: number) {
    if (this.width !== pixelWidth || this.height !== pixelHeight) {
      // this.width = pixelWidth;
      // this.height = pixelHeight;
      this.style.width = width + 'px';
      this.style.height = height + 'px';
      // 'width': width + 'px',
      // 'height': height + 'px',
      // 'viewBox': '0 0 ' + width + ' ' + height
      this.domElement.setAttributeNS(null, 'width', width + 'px');
      this.domElement.setAttributeNS(null, 'height', height + 'px');
      this.domElement.setAttributeNS(null, 'viewBox', '0 0 ' + width + ' ' + height);
      this.context._mainClip.firstElementChild.setAttributeNS(null, 'width', width + 'px');
      this.context._mainClip.firstElementChild.setAttributeNS(null, 'height', height + 'px');
      return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {SVGSurface}
   * @param {string} ln
   * @param {*} attr
   * @param {*=} txt
   * @return {SVGElement}
   */
  public makeElement(ln: string, attr: any, txt?: any): SVGElement {
    const el = this.ownerDocument.createElementNS(this.ns, ln);
    if (Util.isObject(attr)) {
      for (const k in attr) {
        el.setAttributeNS(k === 'href' ? 'http://www.w3.org/1999/xlink' : '', k, attr[k]);
      }
    }
    if (txt !== undefined) el.textContent = txt;
    return el as SVGElement;
  }

  /**
   * @hidden @internal
   * @this {SVGSurface}
   */
  public getBoundingClientRect(): ClientRect | DOMRect {
    return this.domElement.getBoundingClientRect();
  }

  /**
   * @hidden @internal
   * @this {SVGSurface}
   */
  public focus(): void {
    this.domElement.focus();
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  /** @export */ get width(): number { return this.domElement.width.baseVal.value; }
  /** @export */ set width(value: number) {
    this.domElement.width = value; // @@@@ breaking, use setattribute?
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  /** @export */ get height(): number { return this.domElement.height.baseVal.value; }
  /** @export */ set height(value: number) {
    this.domElement.height = value; // @@@@ breaking, use setattribute?
  }

  /**
   * @hidden @internal
   * @this {SVGSurface}
   */
  public dispose(): void {
    this.ownerDocument = null;
  }

  /**
   * @hidden @internal
   * @return {CSSStyleDeclaration}
   */
  /** @export */ get style(): CSSStyleDeclaration { return this.domElement.style; }

} // #end SVGSurface
SVGSurface['className'] = 'SVGSurface';

/**
 * @hidden @internal
 * @unrestricted
 */
export class SVGContext implements IContext {
  /** @hidden @internal */ public static className = 'SVGContext';
  /** @hidden @internal */ private _shadowOffsetX: number;
  /** @hidden @internal */ private _shadowOffsetY: number;
  /** @hidden @internal */ private _shadowBlur: number;

  /** @hidden @internal */ private path: any;
  /** @hidden @internal */ private currentTransform: any;

  /** @hidden @internal */ public fillStyle: string | SGradient;
  /** @hidden @internal */ public font: string;
  /** @hidden @internal */ public globalAlpha: number;
  /** @hidden @internal */ public lineCap: string;
  /** @hidden @internal */ public lineDashOffset: number;
  /** @hidden @internal */ public lineJoin: string;
  /** @hidden @internal */ public lineWidth: number;
  /** @hidden @internal */ public miterLimit: number;
  /** @hidden @internal */ public shadowBlur: number;
  /** @hidden @internal */ public shadowColor: string;
  /** @hidden @internal */ public shadowOffsetX: number;
  /** @hidden @internal */ public shadowOffsetY: number;
  /** @hidden @internal */ public strokeStyle: string | SGradient;
  /** @hidden @internal */ public textAlign: string;
  /** @hidden @internal */ public imageSmoothingEnabled: boolean;
  /** @hidden @internal */ public clipInsteadOfFill: boolean;
  /** @hidden @internal */ public lastCreatedElement: SVGElement;
  /** @hidden @internal */ private stack: Array<ObjectData>;
  /** @hidden @internal */ private groupStack: Array<SVGElement>;
  /** @hidden @internal */ public surface: SVGSurface;
  /** @hidden @internal */ public svg: SVGElement; // a reference to the surface one
  /** @hidden @internal */ public _mainClip: SVGElement | null; // referenced to resize the clipping area
  /** @hidden @internal */ private patterns: { [index: string]: string };
  /** @hidden @internal */ private _isDashing: boolean;
  /** @hidden @internal */ private _strokeDashArray: Array<number> | null;
  /** @hidden @internal */ private _strokeDashOffset: number;

  constructor(surface: SVGSurface) {
    this.surface = surface;

    this.svg = surface.domElement;    // initalzied in the Surface
    this.stack = [];
    this.groupStack = [];

    this.fillStyle = '#000000';
    this.font = '10px sans-serif';
    this.globalAlpha = 1;
    this.lineCap = 'butt';
    this.lineDashOffset = 0;
    this.lineJoin = 'miter';
    this.lineWidth = 1;
    this.miterLimit = 10;
    this.shadowBlur = 0;
    this.shadowColor = 'rgba(0, 0, 0, 0)';
    this.shadowOffsetX = 0;
    this.shadowOffsetY = 0;
    this.strokeStyle = '#000000';
    this.textAlign = 'start';
    this.clipInsteadOfFill = false;
    this._shadowOffsetX = 0;
    this._shadowOffsetY = 0;
    this._shadowBlur = 0;
    this.lastCreatedElement = null;

    this.path = [];           // list of SVG path commands
    this._isDashing = false;
    this._strokeDashArray = null;
    this._strokeDashOffset = 0;

    this.currentTransform = new STransform();

    this.newGroup(1, 0, 0, 1, 0, 0);
    const idnum = Set._ObjectCounter++;
    const clipattr = {
      'id': ('mainClip' + idnum)
    };
    const clipElem = this.makeElement('clipPath', clipattr);
    const attr = {
      'x': 0,
      'y': 0,
      'width': surface.width,
      'height': surface.height
    };

    clipElem.appendChild(this.makeElement('rect', attr));
    this._mainClip = clipElem;
    this.surface.domElement.appendChild(clipElem);
    this.groupStack[0].setAttributeNS(null, 'clip-path', 'url(#mainClip' + idnum + ')');
    this.patterns = {};
  }

  /**
   * @this {SVGContext}
   */
  public reset() {
    // Duplicates what happens in constructor ?
    this.stack = [];
    this.groupStack = [];
    this.fillStyle = '#000000';
    this.font = '10px sans-serif';
    this.globalAlpha = 1;
    this.lineCap = 'butt';
    this.lineDashOffset = 0;
    this.lineJoin = 'miter';
    this.lineWidth = 1;
    this.miterLimit = 10;
    this.shadowBlur = 0;
    this.shadowColor = 'rgba(0, 0, 0, 0)';
    this.shadowOffsetX = 0;
    this.shadowOffsetY = 0;
    this.strokeStyle = '#000000';
    this.textAlign = 'start';
    this.clipInsteadOfFill = false;
    this._shadowOffsetX = 0;
    this._shadowOffsetY = 0;
    this._shadowBlur = 0;

    this.lastCreatedElement = null;
    this.path = [];
    this.currentTransform = new STransform();
    this.newGroup(1, 0, 0, 1, 0, 0);
    const idnum = Set._ObjectCounter++;
    const clipattr = {
      'id': ('mainClip' + idnum)
    };

    const clipElem = this.makeElement('clipPath', clipattr);
    const attr = {
      'x': 0,
      'y': 0,
      'width': this.surface.width,
      'height': this.surface.height
    };

    clipElem.appendChild(this.makeElement('rect', attr));
    this._mainClip = clipElem;
    this.surface.domElement.appendChild(clipElem);
    this.groupStack[0].setAttributeNS(null, 'clip-path', 'url(#mainClip' + idnum + ')');
  }
  // Paths

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   * @param {number} rad
   * @param {number} startAngle
   * @param {number} endAngle
   * @param {boolean} counterclockwise
   * @param {number} lx last point in path
   * @param {number} ly last point in path
   */
  public arc(x: number, y: number, rad: number, startAngle: number, endAngle: number, counterclockwise: boolean, lx: number, ly: number): void {
    const tau = 2 * Math.PI;
    const epsilon = 1e-6;
    const tauEpsilon = tau - epsilon;
    const dx = rad * Math.cos(startAngle);
    const dy = rad * Math.sin(startAngle);
    const startX = x + dx;
    const startY = y + dy;
    const cw = (!!counterclockwise) ? 0 : 1;
    let delta = counterclockwise ? startAngle - endAngle : endAngle - startAngle;

    // (startX, startY) not coincident with the previous point? Line to (startX, startY)
    if (Math.abs(lx - startX) > epsilon || Math.abs(ly - startY) > epsilon) {
      this.path.push(['L', startX, + startY]);
    }

    if (delta < 0) delta = delta % tau + tau; // maybe flip direction

    if (delta > tauEpsilon) { // Complete circle, draw two arcs:
      this.path.push(['A', rad, rad, 0, 1, cw, (x - dx), (y - dy)]);
      this.path.push(['A', rad, rad, 0, 1, cw, startX, startY]);
    } else if (delta > epsilon) { // arc
      this.path.push(['A', rad, rad, 0, (+(delta >= Math.PI)), cw, (x + rad * Math.cos(endAngle)), (y + rad * Math.sin(endAngle))]);
    }
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public beginPath(): void {
    this.path = [];
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public bezierCurveTo(a: number, b: number, c: number, d: number, e: number, f: number): void {
    // this.path.push(['S', a, b, c, d]);
    this.path.push(['C', a, b, c, d, e, f]);
  }

  /**
   * @hidden @internal
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @this {SVGContext}
   */
  public clearRect(x: number, y: number, w: number, h: number): void {
    // no-op, SVG does not need to clear the drawing surface.
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public clip(): void {
    this.addPath('clipPath', this.path, this.currentTransform); // @@@ may be incorrectly sized
    const s = new STransform();
    this.addPath('clipPath', this.path, s);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public closePath(): void {
    this.path.push(['z']);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} aX0
   * @param {number} aY0
   * @param {number} aX1
   * @param {number} aY1
   * @return {SGradient}
   */
  public createLinearGradient(aX0: number, aY0: number, aX1: number, aY1: number): SGradient {
    const gradient = new SGradient('linear');
    gradient.x1 = aX0;
    gradient.y1 = aY0;
    gradient.x2 = aX1;
    gradient.y2 = aY1;
    return gradient;
  }

  /**
   * @hidden @internal
   * @param {HTMLCanvasElement|HTMLImageElement} image
   * @param {string} repetition
   * @this {SVGContext}
   * @return {CanvasPattern}
   */
  public createPattern(image: HTMLCanvasElement | HTMLImageElement, repetition: string): string {
    /*
      <pattern id="img1" patternUnits="userSpaceOnUse" width="100" height="100">
        <image xlink:href="wall.jpg" x="0" y="0" width="100" height="100" />
      </pattern>
    */
    // See if its in the list of patterns, { url : patternName }
    let href = '';
    if (image instanceof HTMLCanvasElement) href = image.toDataURL();
    if (image instanceof HTMLImageElement) href = image.getAttribute('src');
    const patterns = this.patterns;
    if (patterns[href]) {
      return 'url(#' + patterns[href] + ')';
    }
    const patternName = 'PATTERN' + Set._ObjectCounter++;
    const attr = {
      width: image.width,
      height: image.height,
      id: patternName,
      patternUnits: 'userSpaceOnUse'
    };
    const imgattr = {
      x: 0, y: 0, width: image.width, height: image.height,
      'href': href
    };
    const gradElem = this.makeElement('pattern', attr);
    gradElem.appendChild(this.makeElement('image', imgattr));
    this.svg.appendChild(gradElem);
    patterns[href] = patternName;
    return 'url(#' + patternName + ')';
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} aX0
   * @param {number} aY0
   * @param {number} aR0
   * @param {number} aX1
   * @param {number} aY1
   * @param {number} aR1
   * @return {SGradient}
   */
  public createRadialGradient(aX0: number, aY0: number, aR0: number, aX1: number, aY1: number, aR1: number): SGradient {
    const gradient = new SGradient('radial');
    gradient.x1 = aX0;
    gradient.y1 = aY0;
    gradient.r1 = aR0;
    gradient.x2 = aX1;
    gradient.y2 = aY1;
    gradient.r2 = aR1;
    return gradient;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} src
   * @param {number} sx
   * @param {number} sy
   * @param {number=} sw
   * @param {number=} sh
   * @param {number=} dx
   * @param {number=} dy
   * @param {number=} dw
   * @param {number=} dh
   */
  public drawImage(src: HTMLCanvasElement | HTMLImageElement | HTMLVideoElement, sx: number, sy: number, sw?: number, sh?: number,
                   dx?: number, dy?: number, dw?: number, dh?: number): void {
    let href = '';
    if (src instanceof HTMLCanvasElement) href = src.toDataURL();
    if (src instanceof HTMLImageElement) href = src.getAttribute('src');
    const nw = src instanceof HTMLImageElement ? src.naturalWidth : src.width;
    const nh = src instanceof HTMLImageElement ? src.naturalHeight : src.height;
    if (sw === undefined) {
      dx = sx;
      dy = sy;
      dw = sw = nw;
      dh = sh = nh;
    }
    // for closure:
    sw = sw || 0;
    sh = sh || 0;
    dx = dx || 0;
    dy = dy || 0;
    dw = dw || 0;
    dh = dh || 0;

    const attr = {
      'x': 0,
      'y': 0,
      'width': nw || sw, // if it didn't load, take source w/h
      'height': nh || sh, // if it didn't load, take source w/h
      'href': href,
      'preserveAspectRatio': 'xMidYMid slice'
    };

    if (!Geo.isApproxEqual(sw, dw) || !Geo.isApproxEqual(sh, dh)) {
      attr.preserveAspectRatio = 'none';
    }

    // TRANSFORMS:
    // matrix normal
    // ADDITIONAL TRANSFORMS: (for source rect positioning)
    // translate destx desty
    // scale (destw/sourcew, desth/sourceh)
    // translate -sx -sy
    let additionalTrans = '';

    const a = (dw / sw);
    const b = (dh / sh);
    if (dx !== 0 || dy !== 0) additionalTrans += ' translate(' + dx + ', ' + dy + ')';
    if (a !== 1 || b !== 1) additionalTrans += ' scale(' + a + ', ' + b + ')';
    if (sx !== 0 || sy !== 0) additionalTrans += ' translate(' + (-sx) + ', ' + (-sy) + ')';

    // determine if a clipping path is needed:
    if (sx !== 0 ||
      sy !== 0 ||
      sw !== nw ||
      sh !== nh) {
      const clipname = 'CLIP' + Set._ObjectCounter++;
      const clipattr = {
        'id': clipname
      };

      const clipElem = this.makeElement('clipPath', clipattr);
      const rectattr = {
        'x': sx,
        'y': sy,
        'width': sw,
        'height': sh
      };

      clipElem.appendChild(this.makeElement('rect', rectattr));
      this.svg.appendChild(clipElem);
      (attr as any)['clip-path'] = 'url(#' + clipname + ')';
    }

    this.processCommon('image', attr, this.currentTransform, additionalTrans);
    this.addElement('image', attr);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public fill(): void {
    this.addPath('fill', this.path, this.currentTransform);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {Brush|string} brush
   */
  public fillContext(brush: BrushLike): void {
    // does not need scaling because of how SVG handles brushes
    if (this.clipInsteadOfFill) {
      this.clip();
    } else {
      this.fill();
    }
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public fillRect(x: number, y: number, w: number, h: number): void {
    this.addRect('fill', [x, y, w, h], this.currentTransform);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {string} str
   * @param {number} x
   * @param {number} y
   */
  public fillText(str: string, x: number, y: number): void {
    this.addText('fill', [str, x, y], this.currentTransform);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   */
  public lineTo(x: number, y: number): void {
    this.path.push(['L', x, y]);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   */
  public moveTo(x: number, y: number): void {
    this.path.push(['M', x, y]);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   */
  public quadraticCurveTo(a: number, b: number, c: number, d: number): void {
    this.path.push(['Q', a, b, c, d]);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public rect(x: number, y: number, w: number, h: number): void {
    this.path.push(
      ['M', x, y],
      ['L', x + w, y],
      ['L', x + w, y + h],
      ['L', x, y + h],
      ['z']
    );
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public restore(): void {
    // pop currentTransform path, state (in that order)
    this.currentTransform = this.stack.pop();
    this.path = this.stack.pop();
    const state = this.stack.pop();
    this.fillStyle = state.fillStyle;
    this.font = state.font;
    this.globalAlpha = state.globalAlpha;
    this.lineCap = state.lineCap;
    this.lineDashOffset = state.lineDashOffset;
    this.lineJoin = state.lineJoin;
    this.lineWidth = state.lineWidth;
    this.miterLimit = state.miterLimit;
    this.shadowBlur = state.shadowBlur;
    this.shadowColor = state.shadowColor;
    this.shadowOffsetX = state.shadowOffsetX;
    this.shadowOffsetY = state.shadowOffsetY;
    this.strokeStyle = state.strokeStyle;
    this.textAlign = state.textAlign;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public save(): void {
    // Push state, path, currentTransform (in that order)
    const state = {
      fillStyle: this.fillStyle,
      font: this.font,
      globalAlpha: this.globalAlpha,
      lineCap: this.lineCap,
      lineDashOffset: this.lineDashOffset,
      lineJoin: this.lineJoin,
      lineWidth: this.lineWidth,
      miterLimit: this.miterLimit,
      shadowBlur: this.shadowBlur,
      shadowColor: this.shadowColor,
      shadowOffsetX: this.shadowOffsetX,
      shadowOffsetY: this.shadowOffsetY,
      strokeStyle: this.strokeStyle,
      textAlign: this.textAlign
    };

    this.stack.push(state);
    const path = [];
    for (let i = 0; i < this.path.length; i++) {
      path.push(this.path[i]);
    }
    this.stack.push(path);
    this.stack.push(this.currentTransform.copy());
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void {
    // only done because of diagram.draw
    // ignore identity
    if (a === 1 && b === 0 && c === 0 && d === 1 && e === 0 && f === 0) return;
    this.newGroup(a, b, c, d, e, f);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public setCurrentTransform(a: number, b: number, c: number, d: number, e: number, f: number): void {
    this.currentTransform.m = [a, b, c, d, e, f];
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   */
  public scale(x: number, y: number): void {
    // only time its used in GoJS is when drawing diagrams with a larger devicePixelRatio (not needed)
    // and when using .flip (needed)
    this.currentTransform.scale(x, y);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   */
  public translate(x: number, y: number): void {
    this.currentTransform.translate(x, y);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public transform(a: number, b: number, c: number, d: number, e: number, f: number): void {
    // not used by our code, but here for interface consistency between SVGContext and CanvasSurfaceContext
    // NYI
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public stroke(): void {
    this.addPath('stroke', this.path, this.currentTransform);
  }

  /**
   * @hidden @internal
   * only used in Shape so that Shapes set to clip won't draw their stroke
   * @this {SVGContext}
   */
  public strokeContext(): void {
    if (this.clipInsteadOfFill) return;
    this.stroke();
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public strokeRect(x: number, y: number, w: number, h: number): void {
    this.addRect('stroke', [x, y, w, h], this.currentTransform);
  }

  // --- SVG basics
  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {string} ln
   * @param {*} attr
   * @param {*=} txt
   * @return {SVGElement}
   */
  public makeElement(ln: string, attr: any, txt?: any): SVGElement {
    return this.surface.makeElement(ln, attr, txt);
  }


  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {string} ln
   * @param {*} attr
   * @param {*=} txt
   * @return {SVGElement}
   */
  public addElement(ln: string, attr: any, txt?: any): SVGElement {
    const elem = this.makeElement(ln, attr, txt);
    if (this.groupStack.length > 0) {
      const lastgroup = this.groupStack[this.groupStack.length - 1];
      lastgroup.appendChild(elem);
    } else {
      this.svg.appendChild(elem);
    }
    this.lastCreatedElement = elem;
    return elem;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} type 'fill' or 'stroke'
   * @param {*} attr attributes, such as {x, y, width, height}
   * @param {*} trans transform
   * @param {*=} additionalTrans extra transforms (used in images)
   */
  public processCommon(type: any, attr: any, trans: any, additionalTrans?: any): void {
    if (this.globalAlpha !== 1) attr.opacity = this.globalAlpha;
    if (type === 'fill') this.setFill(attr);
    else if (type === 'stroke') this.setStroke(attr);
    const m = trans.m;
    let transform = 'matrix(' + m[0] + ', ' + m[1] + ', ' + m[2] + ', ' + m[3] + ', ' + m[4] + ', ' + m[5] + ')';
    if (additionalTrans !== undefined) transform += additionalTrans;
    attr.transform = transform;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} attr
   */
  public setFill(attr: any): void {
    // WebKit refuses rgba() in SVG
    if (this.fillStyle instanceof SGradient) {
      attr.fill = this.addGradient(this.fillStyle);
    } else {
      if (/^rgba\(/.test(this.fillStyle)) {
        const re = /^\s*rgba\s*\(([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\)\s*$/i;
        const res = re.exec(this.fillStyle);
        attr.fill = 'rgb(' + res[1] + ',' + res[2] + ',' + res[3] + ')';
        attr['fill-opacity'] = res[4];
      }
      attr.fill = this.fillStyle;
    }

    attr.stroke = 'none';
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} attr
   */
  public setStroke(attr: any): void {
    attr.fill = 'none';
    // WebKit refuses rgba() in SVG
    if (this.strokeStyle instanceof SGradient) {
      attr.stroke = this.addGradient(this.strokeStyle);
    } else {
      if (/^rgba\(/.test(this.strokeStyle)) {
        const re = /^\s*rgba\s*\(([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\)\s*$/i;
        const res = re.exec(this.strokeStyle);
        attr.stroke = 'rgb(' + res[1] + ',' + res[2] + ',' + res[3] + ')';
        attr['stroke-opacity'] = res[4];
      }
      attr.stroke = this.strokeStyle;
    }

    attr['stroke-width'] = this.lineWidth;
    attr['stroke-linecap'] = this.lineCap;
    attr['stroke-linejoin'] = this.lineJoin;
    attr['stroke-miterlimit'] = this.miterLimit;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {SGradient} grad
   */
  public addGradient(grad: SGradient): string {
    const gradname = 'GRAD' + Set._ObjectCounter++;
    let gradElem: SVGElement;
    if (grad.type === 'linear') {
      const gradattr = {
        x1: grad.x1, x2: grad.x2,
        y1: grad.y1, y2: grad.y2,
        id: gradname,
        gradientUnits: 'userSpaceOnUse'
      };

      gradElem = this.makeElement('linearGradient', gradattr);
    } else if (grad.type === 'radial') {
      /*
        <radialGradient id="MyGradient" gradientUnits="userSpaceOnUse"
                        cx="400" cy="200" r="300" fx="400" fy="200">
          <stop offset="0%" stop-color="red" />
          <stop offset="100%" stop-color="blue" />
        </radialGradient>
      */

      const gradattr2 = {
        x1: grad.x1, x2: grad.x2, y1: grad.y1,
        y2: grad.y2, r1: grad.r1, r2: grad.r2,
        id: gradname
        // gradientUnits: 'userSpaceOnUse' // not for radial!
      };

      gradElem = this.makeElement('radialGradient', gradattr2);
    } else {
      throw new Error('invalid gradient');
    }

    const colors = grad.colors;
    const l = colors.length;
    const allstops = [];
    for (let i = 0; i < l; i++) {
      const c = colors[i];
      const color = c.color;
      const stopAttr = {
        offset: c.offset,
        'stop-color': color
      };

      if (/^rgba\(/.test(color)) {
        const re = /^\s*rgba\s*\(([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\s*,\s*([^,\s]+)\)\s*$/i;
        const res = re.exec(color);
        (stopAttr as any)['stop-color'] = 'rgb(' + res[1] + ',' + res[2] + ',' + res[3] + ')';
        (stopAttr as any)['stop-opacity'] = res[4];
      }

      allstops.push(stopAttr);
    }
    allstops.sort(function (a, b) {
      if (a.offset > b.offset) return 1;
      return -1;
    });
    for (let i = 0; i < l; i++) {
      gradElem.appendChild(this.makeElement('stop', allstops[i]));
    }

    this.svg.appendChild(gradElem);
    return 'url(#' + gradname + ')';
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} type
   * @param {*} prm
   * @param {*} trans
   */
  public addRect(type: any, prm: any, trans: any): void {
    const attr = { x: prm[0], y: prm[1], width: prm[2], height: prm[3] };

    this.processCommon(type, attr, trans);
    this.addElement('rect', attr);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} type
   * @param {*} prm
   * @param {*} trans
   */
  public addText(type: any, prm: any, trans: any): void {
    let anchor = this.textAlign;
    // SVG only accepts start | middle | end
    if (anchor === 'left') {
      anchor = 'start';
    } else if (anchor === 'right') {
      anchor = 'end';
    } else if (anchor === 'center') {
      anchor = 'middle';
    }
    const attr = {
      x: prm[1],
      y: prm[2],
      style: 'font: ' + this.font, // there isn't a font attribute
      'text-anchor': anchor
    };

    this.processCommon(type, attr, trans);
    this.addElement('text', attr, prm[0]);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {*} type
   * @param {*} path
   * @param {*} trans
   */
  public addPath(type: any, path: any, trans: any): void {
    const cmds = [];
    for (let i = 0; i < path.length; i++) {
      const subP = Util.copyArray(path[i]);
      const cmd = [subP.shift()];
      if (cmd[0] === 'A') {
        cmd.push(subP.shift() + ',' + subP.shift(),
          subP.shift(),
          subP.shift() + ',' + subP.shift(),
          subP.shift() + ',' + subP.shift()
        );
      } else {
        while (subP.length) {
          cmd.push(subP.shift() + ',' + subP.shift());
        }
      }
      cmds.push(cmd.join(' '));
    }
    const built = cmds.join(' ');
    const attr: ObjectData = { d: built };
    if (type === 'stroke' && this._isDashing) {
      attr['stroke-dasharray'] = this._strokeDashArray.toString();
      attr['stroke-dashoffset'] =  this._strokeDashOffset;
    }

    this.processCommon(type, attr, trans);
    if (type === 'clipPath') {
      const clipname = 'CLIP' + Set._ObjectCounter++;
      const clipattr = {
        id: clipname
      };

      const clipElem = this.makeElement('clipPath', clipattr);
      clipElem.appendChild(this.makeElement('path', attr));
      this.svg.appendChild(clipElem);

      if (this.groupStack.length > 0) {
        const lastgroup = this.groupStack[this.groupStack.length - 1];
        lastgroup.setAttributeNS(null, 'clip-path', 'url(#' + clipname + ')');
      }
    } else {
      this.addElement('path', attr);
    }
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public newGroup(a: number, b: number, c: number, d: number, e: number, f: number): void {
    const trans = new STransform();
    trans.m = [a, b, c, d, e, f];
    const attr = {};

    this.processCommon('g', attr, trans);
    const g = this.addElement('g', attr);
    // add new group to the stack
    this.groupStack.push(g);
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public endGroup(): SVGElement {
    return this.groupStack.pop();
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public dummyGroup(): SVGElement {
    this.newGroup(1, 0, 0, 1, 0, 0);
    return this.endGroup();
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public setShadow(): void {
    if (this.shadowOffsetX === 0 && this.shadowOffsetY === 0 && this.shadowBlur === 0) return;
    const id = 'SHADOW' + Set._ObjectCounter++;
    // width and height are approximations, may need to be larger
    // The SVG Filter defaults are: x="-10%" y="-10%" width="120%" height="120%"
    // which are not nearly large enough for many shadows
    const filter = this.addElement('filter', { id: id, x: '-100%', y: '-100%', width: '300%', height: '300%' }, null);
    let feGaussianBlur: SVGElement;
    let feFlood: SVGElement;
    let feComposite: SVGElement;
    let feOffset: SVGElement;
    let feMerge: SVGElement;

    /*
    <filter id="MyFilter" x="0" y="0" width="250%" height="250%">
        <feGaussianBlur in="SourceAlpha" stdDeviation="1" result="blur" />
        <feFlood in="blur" result="flood" flood-color="red" />
        <feComposite in="flood" in2="blur" operator="in" result="comp"/>
        <feOffset in="blur" dx="8" dy="8" result="offsetBlur" />
        <feMerge>
            <feMergeNode in="offsetBlur" />
            <feMergeNode in="SourceGraphic" />
        </feMerge>
    </filter>
    */

    feGaussianBlur = this.makeElement('feGaussianBlur', {
      'in': 'SourceAlpha',
      result: 'blur',
      stdDeviation: (this.shadowBlur / 2)
    });

    feFlood = this.makeElement('feFlood', {
      'in': 'blur',
      result: 'flood',
      'flood-color': this.shadowColor
    });
    feComposite = this.makeElement('feComposite', {
      'in': 'flood',
      in2: 'blur',
      operator: 'in',
      result: 'comp'
    });

    feOffset = this.makeElement('feOffset', {
      'in': 'comp',
      result: 'offsetBlur',
      dx: this.shadowOffsetX,
      dy: this.shadowOffsetY
    });

    feMerge = this.makeElement('feMerge', {});
    feMerge.appendChild(this.makeElement('feMergeNode', { 'in': 'offsetBlur' }));
    feMerge.appendChild(this.makeElement('feMergeNode', { 'in': 'SourceGraphic' }));

    filter.appendChild(feGaussianBlur);
    filter.appendChild(feFlood);
    filter.appendChild(feComposite);
    filter.appendChild(feOffset);
    filter.appendChild(feMerge);

    //  attr.filter = "url(#" + id + ")";

    if (this.groupStack.length > 0) {
      const lastgroup = this.groupStack[this.groupStack.length - 1];
      lastgroup.setAttributeNS(null, 'filter', 'url(#' + id + ')');
    }
  }

  /**
   * @hidden @internal
   * @param {number} x
   * @param {number} y
   * @param {number} blur
   * @this {SVGContext}
   */
  public shadowsSet(x: number, y: number, blur: number): void {
    this._shadowOffsetX = x;
    this._shadowOffsetY = y;
    this._shadowBlur = blur;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public shadowsOff(): void {
    this.shadowOffsetX = 0;
    this.shadowOffsetY = 0;
    this.shadowBlur = 0;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public shadowsOn(): void {
    this.shadowOffsetX = this._shadowOffsetX;
    this.shadowOffsetY = this._shadowOffsetY;
    this.shadowBlur = this._shadowBlur;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {Array.<number>} strokeDashArray
   * @param {number} strokeDashOffset
   */
  public enableDash(strokeDashArray: Array<number>, strokeDashOffset: number) {
    this._isDashing = true;
    // These will be used in addPath if the path is a stroke.
    this._strokeDashArray = strokeDashArray;
    this._strokeDashOffset = strokeDashOffset;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   */
  public disableDash(): void {
    this._isDashing = false;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {boolean} clearFont
   */
  public clearContextCache(clearFont: boolean): void {
    // unused
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {boolean} smooth
   */
  public setImageSmoothingEnabled(smooth: boolean): void {
    // unused
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} angle
   */
  public rotate(angle: number): void {
    // unused
  }

  // The following added for interface compatibility

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {ImageData}
   */
  public getImageData(x: number, y: number, w: number, h: number): ImageData {
    return null;
  }

  /**
   * @hidden @internal
   * @this {SVGContext}
   * @param {string} text
   * @return {TextMetrics}
   */
  public measureText(text: string): TextMetrics {
    return null;
  }


} // #end SVGContext
SVGContext['className'] = 'SVGContext';


/**
 * @hidden @internal
 * @export
 * @this {Diagram}
 * @suppress {duplicate}
 */
Diagram.prototype.makeSvg = function(this: Diagram, options?: SvgRendererOptions): SVGElement | null {
  // if (!this._renderers.contains('SVG')) {
  //   this.addRenderer('SVG', new SVGSurface(this, root.document));
  // }
  // One time SVG surface.
  const surface = new SVGSurface(this, root.document);
  if (options === undefined) options = new PropertyCollection() as any;

  const self = this;
  function svg2(opts: SvgRendererOptions, callbackFunc: ((result: any) => void) | null): SVGElement | null {
    const svgc = self.drawData(opts, 'SVG', surface) as SVGContext;
    const returnValue = (svgc !== null) ? svgc.svg : null;
    if (typeof callbackFunc === 'function') {
      callbackFunc(returnValue);
      return null;
    }
    return returnValue;
  }
  return this.setupTimers(svg2, options);
};

/**
 * @hidden @internal
 * @export
 * @this {Diagram}
 * @suppress {duplicate}
 */
Diagram.prototype.makeSVG = function(this: Diagram, options?: SvgRendererOptions): SVGElement | null {
  return this.makeSvg(options);
};


/**
 * @hidden @internal
 * The draw function when rendering to an SVG context.
 * @this {GraphObject}
 * @param {SVGContext} context Special SVG drawing context
 * @param {Diagram} diagram
 * @return {boolean}
 * @suppress {duplicate}
 */
GraphObject.prototype.drawForSVG = function(this: GraphObject, context: SVGContext, diagram: Diagram): boolean {
    if (!(context instanceof SVGContext)) return false;
    if (!this.visible) return true;
    let SVGElem: SVGElement = null;
    const LastSVGElem = context.lastCreatedElement;

    if (this instanceof Panel) {
      if (this.type === Panel.TableRow || this.type === Panel.TableColumn) {
        this.drawRowColumn(context, diagram);
        return true;
      }
    }

    const ab = this._actualBounds;
    if (ab.width === 0 || ab.height === 0 || isNaN(ab.x) || isNaN(ab.y)) return true;

    const tt = this.transform;
    const myPanel = this.panel;
    // construct the allTransforms for this object
    if (this.getInvalidAllTransforms() === true) { this._validateTransforms(); }

    const clip = (this._flags & 0x00000100) !== 0; // clipFlag;
    let clippingGroup = false;
    if (this instanceof TextBlock) {
      context.font = this.font;
    }
    if (clip) {
      const nb2 = myPanel.trueTransform() ? myPanel.naturalBounds : myPanel.actualBounds;
      let cx;
      let cy;
      let cwidth;
      let cheight;
      if (this._clipRect !== null) {
        const cr = this._clipRect;
        cx = cr.x;
        cy = cr.y;
        cwidth = cr.width;
        cheight = cr.height;
      } else {
        cx = Math.max(ab.x, nb2.x);
        cy = Math.max(ab.y, nb2.y);
        cwidth = Math.min(ab.right, nb2.right) - cx;
        cheight = Math.min(ab.bottom, nb2.bottom) - cy;
      }
      if (cx > ab.width + ab.x || ab.x > nb2.width + nb2.x) return true;
      if (cy > ab.height + ab.y || ab.y > nb2.height + nb2.y) return true;

      clippingGroup = true;
      context.newGroup(1, 0, 0, 1, 0, 0);
      context.save();
      context.beginPath();
      context.rect(cx, cy, cwidth, cheight);
      context.clip();
    }

    if (this.isPart()) {
      if (!this.isVisible()) return true;
    }

    context.setCurrentTransform(1, 0, 0, 1, 0, 0);

    // Both TextBlock and Shape may make multiple SVG elements that should be <g>rouped together.
    if ((this instanceof TextBlock && (this as TextBlock).lineCount > 1) ||
        (this instanceof Shape && this.geometry.figures.length > 1)) {
      context.newGroup(1, 0, 0, 1, 0, 0);
    }

    let shadowGroupMade = false;
    if (this.isPart() && (this as Part).isShadowed && diagram.getRenderingHint('drawShadows')) {
      const shad = ((this as Part)).shadowOffset;
      context.shadowsSet(
        shad.x * diagram.scale * diagram._pixelRatio,
        shad.y * diagram.scale * diagram._pixelRatio,
        (this as Part).shadowBlur);
      context.shadowsOn();
      context.shadowColor = (this as Part).shadowColor;
    }
    let isShadowed = false;
    if (this.part && diagram.getRenderingHint('drawShadows')) isShadowed = this.part.isShadowed;

    // If explicitly turned on or off
    if (this.shadowVisible === true) {
      context.shadowsOn();
      if (shadowGroupMade === false && isShadowed) { context.newGroup(1, 0, 0, 1, 0, 0); context.setShadow(); shadowGroupMade = true; }
    } else if (this.shadowVisible === false) {
      context.shadowsOff();
    }

    // areaBackground is in the parent's transform space
    const nb = this.naturalBounds;
    if (this.areaBackground !== null) {
      this.setBrush(context, this.areaBackground, true, true, nb, ab);
      if (shadowGroupMade === false && isShadowed) { context.newGroup(1, 0, 0, 1, 0, 0); context.setShadow(); shadowGroupMade = true; }
      if (this.areaBackground instanceof Brush && this.areaBackground.type === Brush.Radial) {
        context.beginPath();
        context.rect(ab.x, ab.y, ab.width, ab.height);
        context.fillContext(this.areaBackground);
      } else {
        context.fillRect(ab.x, ab.y, ab.width, ab.height);
      }
    }

    if (this instanceof Panel) {
      context.newGroup(tt.m11, tt.m12, tt.m21, tt.m22, tt.dx, tt.dy);
    } else {
      context.setCurrentTransform(tt.m11, tt.m12, tt.m21, tt.m22, tt.dx, tt.dy);
    }

    // background is in local space
    if (this.background !== null) {
      if (shadowGroupMade === false && isShadowed) { context.newGroup(1, 0, 0, 1, 0, 0); context.setShadow(); shadowGroupMade = true; }
      let ns = this.naturalBounds;
      let nx = 0.0;
      let ny = 0.0;
      let nw = ns.width;
      let nh = ns.height;
      let sw = 0.0;
      if (this instanceof Shape) {
        ns = this.geometry.bounds;
        nx = ns.x;
        ny = ns.y;
        nw = ns.width;
        nh = ns.height;
        sw = this.strokeWidth;
      }
      this.setBrush(context, this.background, true, false, nb, ab);
      if (this.background instanceof Brush && this.background.type === Brush.Radial) {
        context.beginPath();
        context.rect(nx - (
          sw / 2),
          ny - (sw / 2),
          nw + sw, nh + sw);
        context.fillContext(this.background);
      } else {
        context.fillRect(
          nx - (sw / 2),
          ny - (sw / 2),
          nw + sw, nh + sw);
      }
    }

    if (isShadowed &&
      (this.background !== null || this.areaBackground !== null ||
        (myPanel !== null && myPanel.getAlreadyShadowed()) ||
        (myPanel !== null && (myPanel.type === Panel.Auto || myPanel.type === Panel.Spot) && myPanel.findMainElement() !== this))) {
      this.setAlreadyShadowed(true); // no more shadows
      if (this.shadowVisible === null) context.shadowsOff();
    } else { // continue with shadows
      this.setAlreadyShadowed(false);
    }

    this.drawProtected(context, diagram);

    // Reset things for the next sibling
    if (isShadowed && this.getAlreadyShadowed() === true) {
      context.shadowsOn();
    }

    // When done drawing the whole part, reset any shadow:
    if (this.isPart() && isShadowed) {
      context.shadowsOff();
      // ??? reset diagram._shadow... too?
    }

    if (clip) {
      context.restore();
      if (clippingGroup) context.endGroup();
    }
    if (this instanceof Panel) {
      SVGElem = context.endGroup();
    }

    if (shadowGroupMade === true) {
      context.endGroup();
    }

    if ((this instanceof TextBlock && (this as TextBlock).lineCount > 1) ||
        (this instanceof Shape && this.geometry.figures.length > 1)) {
      SVGElem = context.endGroup();
    }

    if (context.surface.elementFinished !== null) {
      if (SVGElem === null) {
        // see if this GraphObject created an SVG element
        if (LastSVGElem === context.lastCreatedElement) {
          // if it did not, create an element
          SVGElem = context.dummyGroup();
        } else {
          // if it did, use the last element
          SVGElem = context.lastCreatedElement;
        }
      }
      context.surface.elementFinished(this, SVGElem);
    }
    this.svg = SVGElem;
    return true;
  }; // end drawForSVG

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
CanvasSurface
CanvasSurfaceContext
*/

// import { root } from './root';
// import { IContext, ISurface } from './surfaceInterfaces';
// import { Util } from './util';

/**
 * @hidden @internal
 * @unrestricted
 */
export class CanvasSurface implements ISurface {
  /** @hidden @internal */ public static className = 'CanvasSurface';
  /** @hidden @internal */ public domElement: HTMLCanvasElement;
  /** @hidden @internal */ public context: CanvasSurfaceContext;
  /** @hidden @internal */ public ownerDocument: Document;
  /** @hidden @internal */ public elementFinished: ((a: GraphObject, b: SVGElement) => void) | null; // only used in SVG

  /**
   * @param {Diagram} diagram can be null
   * @param {Document=} ownerDocument
   */
  constructor(diagram: Diagram | null, ownerDocument?: Document) {
    const doc = (ownerDocument === undefined) ? root.document : ownerDocument;
    this.ownerDocument = doc;
    this.elementFinished = null;
    const can = doc.createElement('canvas');
    can.tabIndex = 0; // the canvas must be allowed focus for focusout and blur events to work
    this.domElement = can;
    this.domElement.innerHTML = 'This text is displayed if your browser does not support the Canvas HTML element.';
    this.context = new CanvasSurfaceContext(can);
    can._diagram = diagram;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurface}
   * @param {number} pixelWidth
   * @param {number} pixelHeight
   * @param {number} width
   * @param {number} height
   * @returns {boolean} true if the surface changed size
   */
  public resize(pixelWidth: number, pixelHeight: number, width: number, height: number): boolean {
    if (this.width !== pixelWidth || this.height !== pixelHeight) {
      this.width = pixelWidth;
      this.height = pixelHeight;
      this.style.width = width + 'px';
      this.style.height = height + 'px';
      return true;
    }
    return false;
  }


  /**
   * @hidden @internal
   * @this {CanvasSurface}
   * @param {string} type
   * @param {number} details
   */
  public toDataURL(type: string, details: number): string {
    return this.domElement.toDataURL(type, details);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurface}
   */
  public getBoundingClientRect(): ClientRect | DOMRect {
    return this.domElement.getBoundingClientRect();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurface}
   */
  public focus(): void {
    this.domElement.focus();
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  /** @export */ get width(): number { return this.domElement.width; }
  /** @export */ set width(value: number) {
    this.domElement.width = value;
  }

  /**
   * @hidden @internal
   * @return {number}
   */
  /** @export */ get height(): number { return this.domElement.height; }
  /** @export */ set height(value: number) {
    this.domElement.height = value;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurface}
   */
  public dispose(): void {
    this.domElement._diagram = null;
    this.ownerDocument = null;
  }

  /**
   * @hidden @internal
   * @return {CSSStyleDeclaration}
   */
  /** @export */ get style(): CSSStyleDeclaration { return this.domElement.style; }

} // #end CanvasSurface
CanvasSurface['className'] = 'CanvasSurface';



/**
 * @hidden @internal
 * @unrestricted
 */
export class CanvasSurfaceContext implements IContext {
  /** @hidden @internal */ public static className = 'CanvasSurfaceContext';
  /** @hidden @internal */ public _backingContext: CanvasRenderingContext2D;
  /** @hidden @internal */ private _cachedFont: string;
  /** @hidden @internal */ private _cachedStrokeStyle: string | CanvasGradient | CanvasPattern;
  /** @hidden @internal */ private _cachedFillStyle: string | CanvasGradient | CanvasPattern;
  /** @hidden @internal */ private _clipInsteadOfFill: boolean;
  /** @hidden @internal */ private _shadowOffsetX: number;
  /** @hidden @internal */ private _shadowOffsetY: number;
  /** @hidden @internal */ private _shadowBlur: number;

  /**
   * @param {HTMLCanvasElement} DOMCanvas
   */
  constructor(DOMCanvas: HTMLCanvasElement) {
    if (!(DOMCanvas.getContext && DOMCanvas.getContext('2d'))) {
      Util.throwError('Browser does not support HTML Canvas Element');
    }
    this._backingContext = (DOMCanvas.getContext('2d'/*, { desynchronized: true }*/) as CanvasRenderingContext2D);

    this._cachedFont = '';
    this._cachedStrokeStyle = '';
    this._cachedFillStyle = '';
    this._clipInsteadOfFill = false;
    // used in drawing shadows
    this._shadowOffsetX = 0;
    this._shadowOffsetY = 0;
    this._shadowBlur = 0;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {boolean} smooth
   */
  public setImageSmoothingEnabled(smooth: boolean): void {
    this._backingContext.imageSmoothingEnabled = smooth;
  }

  /** @export */ get fillStyle() { return this._backingContext.fillStyle; }
  /** @export */ set fillStyle(value) {
    if (this._cachedFillStyle === value) return;
    this._backingContext.fillStyle = value;
    this._cachedFillStyle = value;
  }
  /** @export */ get font() { return this._backingContext.font; }
  /** @export */ set font(value) {
    if (this._cachedFont === value) return;
    this._backingContext.font = value;
    this._cachedFont = value;
  }
  /** @export */ get globalAlpha() { return this._backingContext.globalAlpha; }
  /** @export */ set globalAlpha(value) { this._backingContext.globalAlpha = value; }
  /** @export */ get lineCap() { return this._backingContext.lineCap; }
  /** @export */ set lineCap(value) { this._backingContext.lineCap = value; }
  /** @export */ get lineDashOffset() { return this._backingContext.lineDashOffset; }
  /** @export */ set lineDashOffset(value) { this._backingContext.lineDashOffset = value; }
  /** @export */ get lineJoin() { return this._backingContext.lineJoin; }
  /** @export */ set lineJoin(value) { this._backingContext.lineJoin = value; }
  /** @export */ get lineWidth() { return this._backingContext.lineWidth; }
  /** @export */ set lineWidth(value) { this._backingContext.lineWidth = value; }
  /** @export */ get miterLimit() { return this._backingContext.miterLimit; }
  /** @export */ set miterLimit(value) { this._backingContext.miterLimit = value; }
  /** @export */ get shadowBlur() { return this._backingContext.shadowBlur; }
  /** @export */ set shadowBlur(value) { this._backingContext.shadowBlur = value; }
  /** @export */ get shadowColor() { return this._backingContext.shadowColor; }
  /** @export */ set shadowColor(value) { this._backingContext.shadowColor = value; }
  /** @export */ get shadowOffsetX() { return this._backingContext.shadowOffsetX; }
  /** @export */ set shadowOffsetX(value) { this._backingContext.shadowOffsetX = value; }
  /** @export */ get shadowOffsetY() { return this._backingContext.shadowOffsetY; }
  /** @export */ set shadowOffsetY(value) { this._backingContext.shadowOffsetY = value; }
  /** @export */ get strokeStyle() { return this._backingContext.strokeStyle; }
  /** @export */ set strokeStyle(value) {
    if (this._cachedStrokeStyle === value) return;
    this._backingContext.strokeStyle = value;
    this._cachedStrokeStyle = value;
  }
  /** @export */ get textAlign() { return this._backingContext.textAlign; }
  /** @export */ set textAlign(value) { this._backingContext.textAlign = value; }
  /** @export */ get imageSmoothingEnabled() { return this._backingContext.imageSmoothingEnabled; }
  /** @export */ set imageSmoothingEnabled(value) { this._backingContext.imageSmoothingEnabled = value; }


  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} radius
   * @param {number} startAngle
   * @param {number} endAngle
   * @param {boolean} counterclockwise
   * @param {number} lx last point in path (only used in SVG)
   * @param {number} ly last point in path (only used in SVG)
   */
  public arc(x: number, y: number, radius: number, startAngle: number, endAngle: number, counterclockwise: boolean, lx: number, ly: number): void {
    this._backingContext.arc(x, y, radius, startAngle, endAngle, counterclockwise);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public beginPath(): void {
    this._backingContext.beginPath();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public bezierCurveTo(a: number, b: number, c: number, d: number, e: number, f: number): void {
    this._backingContext.bezierCurveTo(a, b, c, d, e, f);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public clearRect(x: number, y: number, w: number, h: number): void {
    this._backingContext.clearRect(x, y, w, h);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public clip(): void {
    this._backingContext.clip();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public closePath(): void {
    this._backingContext.closePath();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} aX0
   * @param {number} aY0
   * @param {number} aX1
   * @param {number} aY1
   * @return {CanvasGradient}
   */
  public createLinearGradient(aX0: number, aY0: number, aX1: number, aY1: number): CanvasGradient {
    return this._backingContext.createLinearGradient(aX0, aY0, aX1, aY1);
  }

  /**
   * @hidden @internal
   * @param {HTMLCanvasElement|HTMLImageElement} image
   * @param {string} repetition
   * @this {CanvasSurfaceContext}
   * @return {CanvasPattern}
   */
  public createPattern(image: HTMLCanvasElement | HTMLImageElement, repetition: string): CanvasPattern {
    return this._backingContext.createPattern(image, repetition);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} aX0
   * @param {number} aY0
   * @param {number} aR0
   * @param {number} aX1
   * @param {number} aY1
   * @param {number} aR1
   * @return {CanvasGradient}
   */
  public createRadialGradient(aX0: number, aY0: number, aR0: number, aX1: number, aY1: number, aR1: number): CanvasGradient {
    return this._backingContext.createRadialGradient(aX0, aY0, aR0, aX1, aY1, aR1);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} src
   * @param {number} sx
   * @param {number} sy
   * @param {number=} sw
   * @param {number=} sh
   * @param {number=} dx
   * @param {number=} dy
   * @param {number=} dw
   * @param {number=} dh
   */
  public drawImage(src: HTMLCanvasElement | HTMLImageElement | HTMLVideoElement,
                   sx: number, sy: number, sw?: number, sh?: number, dx?: number, dy?: number, dw?: number, dh?: number): void {
    if (sw === undefined) {
      this._backingContext.drawImage(src, sx, sy);
    } else {
      this._backingContext.drawImage(src, sx, sy, sw, sh, dx, dy, dw, dh);
    }
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public fill(): void {
    this._backingContext.fill();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public fillRect(x: number, y: number, w: number, h: number): void {
    this._backingContext.fillRect(x, y, w, h);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {string} str
   * @param {number} x
   * @param {number} y
   */
  public fillText(str: string, x: number, y: number): void {
    this._backingContext.fillText(str, x, y);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {ImageData}
   */
  public getImageData(x: number, y: number, w: number, h: number): ImageData {
    return this._backingContext.getImageData(x, y, w, h);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   */
  public lineTo(x: number, y: number): void {
    this._backingContext.lineTo(x, y);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {string} text
   * @return {TextMetrics}
   */
  public measureText(text: string): TextMetrics {
    return this._backingContext.measureText(text);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   */
  public moveTo(x: number, y: number): void {
    this._backingContext.moveTo(x, y);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   */
  public quadraticCurveTo(a: number, b: number, c: number, d: number): void {
    this._backingContext.quadraticCurveTo(a, b, c, d);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} w
   * @param {number} h
   */
  public rect(x: number, y: number, w: number, h: number): void {
    this._backingContext.rect(x, y, w, h);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public restore(): void {
    this._backingContext.restore();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} angle in radians
   */
  public rotate(angle: number): void {
    this._backingContext.rotate(angle);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public save(): void {
    this._backingContext.save();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public setTransform(a: number, b: number, c: number, d: number, e: number, f: number): void {
    this._backingContext.setTransform(a, b, c, d, e, f);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   */
  public scale(x: number, y: number): void {
    this._backingContext.scale(x, y);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public stroke(): void {
    this._backingContext.stroke();
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} a
   * @param {number} b
   * @param {number} c
   * @param {number} d
   * @param {number} e
   * @param {number} f
   */
  public transform(a: number, b: number, c: number, d: number, e: number, f: number): void {
    if (a === 1 && b === 0 && c === 0 && d === 1 && e === 0 && f === 0) return; // ignore identity
    this._backingContext.transform(a, b, c, d, e, f);
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   */
  public translate(x: number, y: number): void {
    this._backingContext.translate(x, y);
  }

  // More custom stuff:

  /**
   * @hidden @internal
   * This is only needed because of radial gradient brushes
   * Fills/strokes a radial brush on the context, scaling in X or Y axis appropriately to "fit" the radial brush
   * If the brush is not radial, proceeds to fill/stroke normally.
   * @this {CanvasSurfaceContext}
   * @param {Brush|string} brush
   */
  public fillContext(brush: BrushLike): void {
    // special scaling if radial brush only
    if ((brush instanceof Brush) && brush.type === Brush.Radial) {
      const w = brush._constructedWidth;
      const h = brush._constructedHeight;

      if (h > w) {
        this.scale((w / h), 1);
        this.translate((h - w) / 2, 0);
      } else if (w > h) {
        this.scale(1, (h / w));
        this.translate(0, (w - h) / 2);
      }
      this.fillOrClip();
      if (h > w) {
        this.translate(-(h - w) / 2, 0);
        this.scale(1 / (w / h), 1);
      } else if (w > h) {
        this.translate(0, -(w - h) / 2);
        this.scale(1, 1 / (h / w));
      }
      return;
    }
    this.fillOrClip();
  }

  /**
   * @hidden @internal
   * Only used in Shape so that Shapes set to clip won't draw their stroke
   * @this {CanvasSurfaceContext}
   */
  public strokeContext(): void {
    if (this._clipInsteadOfFill) return;
    this.stroke();
  }

  /**
   * @hidden @internal
   * Calls context.fill, or context.clip if appropriate
   * This can be replaced by an abstract context in the future
   * @this {CanvasSurfaceContext}
   */
  public fillOrClip(): void {
    if (this._clipInsteadOfFill) {
      this.clip();
    } else {
      this.fill();
    }
  }
  /** @export */ get clipInsteadOfFill(): boolean { return this._clipInsteadOfFill; }
  /** @export */ set clipInsteadOfFill(value: boolean) { this._clipInsteadOfFill = value; }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {number} x
   * @param {number} y
   * @param {number} blur
   */
  public shadowsSet(x: number, y: number, blur: number): void {
    this._shadowOffsetX = x;
    this._shadowOffsetY = y;
    this._shadowBlur = blur;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public shadowsOff(): void {
    this.shadowOffsetX = 0;
    this.shadowOffsetY = 0;
    this.shadowBlur = 0;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public shadowsOn(): void {
    this.shadowOffsetX = this._shadowOffsetX;
    this.shadowOffsetY = this._shadowOffsetY;
    this.shadowBlur = this._shadowBlur;
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   * @param {Array.<number>} strokeDashArray
   * @param {number} strokeDashOffset
   */
  public enableDash(strokeDashArray: Array<number>, strokeDashOffset: number) {
    const context = this._backingContext;
    if (context.setLineDash !== undefined) {
      context.setLineDash(strokeDashArray);
      context.lineDashOffset = strokeDashOffset;
    }
  }

  /**
   * @hidden @internal
   * @this {CanvasSurfaceContext}
   */
  public disableDash(): void {
    const context = this._backingContext;
    if (context.setLineDash !== undefined) {
      context.setLineDash(CanvasSurfaceContext.EmptyArray);
      context.lineDashOffset = 0;
    }
  }

  /**
   * @hidden @internal
   * Called by setup and whenever the canvas changes size. Changing the size of the canvas will destroy all canvas state.
   * @this {CanvasSurfaceContext}
   * @param {boolean} clearFont if true, clear the font (false is uncommon, a minor optimization)
   */
  public clearContextCache(clearFont: boolean): void {
    if (clearFont) this._cachedFont = '';
    this._cachedStrokeStyle = '';
    this._cachedFillStyle = '';
  }

  public static EmptyArray = Object.freeze([]) as Array<never>;
} // #end CanvasSurfaceContext
CanvasSurfaceContext['className'] = 'CanvasSurfaceContext';

// import { Brush, BrushLike } from './brush';
// import { Diagram } from './diagram';
// import { GraphObject } from './graphObject';





/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Brush
*/

// import { CanvasSurface, CanvasSurfaceContext } from './abstractCanvas';
// import { Map, Set } from './collections';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Spot } from './spot';
// import { SGradient } from './svgContext';
// import { Util } from './util';

/**
 * @hidden @internal
 * For storing the set of numbers that make up a color, eg. [R, G, B, A] or [L, a, b]
 * @unrestricted
 */
export class ColorNumbers {
  /** @hidden @internal */ public static className = 'ColorNumbers';
  /** @hidden @internal */ public n0: number;
  /** @hidden @internal */ public n1: number;
  /** @hidden @internal */ public n2: number;
  /** @hidden @internal */ public n3: number;

  constructor() {
    this.n0 = 0;
    this.n1 = 0;
    this.n2 = 0;
    this.n3 = 0;
  }

} // #end ColorNumbers
ColorNumbers['className'] = 'ColorNumbers';

/**
 * The BrushLike type is the same as `Brush | string | null`.
 * A string in place of a Brush object is treated as a Solid Brush of that color.
 * @category Type
 */
export type BrushLike = Brush | string | null;

/**
 * A Brush holds color information and describes how to draw the inside
 * of a {@link Shape} or the stroke of a shape or a {@link TextBlock} or the
 * background of any {@link GraphObject}.
 *
 * A Brush must not be modified once it has been assigned to a {@link GraphObject},
 * such as the {@link Shape#fill} or {@link TextBlock#stroke}
 * or {@link GraphObject#background}.
 * However, a Brush may be shared by multiple GraphObjects.
 * @unrestricted
 * @category Geometry
 */
export class Brush {
  /** @hidden @internal */ public static className = 'Brush';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ private _type: EnumValue;
  /** @hidden @internal */ private _color: string;
  /** @hidden @internal */ private _start: Spot;
  /** @hidden @internal */ private _end: Spot;
  /** @hidden @internal */ private _startRadius: number;
  /** @hidden @internal */ private _endRadius: number;
  /** @hidden @internal */ private _colorStops: Map<number, string> | null;
  /** @hidden @internal */ private _pattern: HTMLCanvasElement | HTMLImageElement | null;
  /** @hidden @internal */ public _constructed: CanvasGradient | SGradient | CanvasPattern | null;
  /** @hidden @internal */ public _constructedWidth: number;
  /** @hidden @internal */ public _constructedHeight: number;

  /**
   * Construct a Brush class that holds the given color information.
   * @param {EnumValue|string=} type Optional, one of the values {@link Brush.Solid}, {@link Brush.Linear}, {@link Brush.Radial}, {@link Brush.Pattern},
   * or a well-formed CSS string describing a solid color brush. No parameter
   * defaults to a {@link Brush.Solid} with a color description of 'black'.
   */
  constructor(type?: EnumValue | string) {
    if (Debug && arguments.length > 1) {
      Util.throwError('Brush constructor can take at most one optional argument, the Brush type.');
    }
    if (!Brush._hasInit) {
      Brush.initStatic();
      Brush._hasInit = true;
    }

    Set.uniqueHash(this);
    this._isFrozen = false;
    if (type === undefined) {
      this._type = Brush.Solid;
      this._color = 'black';
    } else if (typeof type === 'string') {
      this._type = Brush.Solid;
      if (Debug && !Brush.isValidColor(type)) Util.throwError('Color \"' + type + '\" is not a valid color string for Brush constructor');
      this._color = type;
    } else {
      if (Debug) EnumValue.checkEnumValue(type, Brush, Brush, 'constructor:type');
      this._type = type;
      this._color = 'black';
    }

    const t = this._type;
    if (t === Brush.Linear) {
      this._start = Spot.TopCenter;
      this._end = Spot.BottomCenter;
    } else if (t === Brush.Radial) {
      this._start = Spot.Center;
      this._end = Spot.Center;
    } else {
      this._start = Spot.None;
      this._end = Spot.None;
    }

    this._startRadius = 0;
    this._endRadius = NaN;

    this._colorStops = null;

    this._pattern = null;

    // private only caching of the brush
    // The first time a Brush is drawn it is constructed on the canvas
    // Gets set in GraphObject.setBrush and cleared whenever a Brush property is changed
    this._constructed = null;
    this._constructedWidth = 0;
    this._constructedHeight = 0;
  }

  /**
   * Create a copy of this Brush with the same values.
   * @this {Brush}
   * @return {Brush}
   */
  public copy(): Brush {
    const copy = new Brush();
    copy._type = this._type;
    copy._color = this._color;
    copy._start = this._start.copyFrozen();
    copy._end = this._end.copyFrozen();
    copy._startRadius = this._startRadius;
    copy._endRadius = this._endRadius;
    if (this._colorStops !== null) {
      copy._colorStops = this._colorStops.copy();
    }
    copy._pattern = this._pattern;
    // _constructed* are private cached properties
    return copy;
  }

  //
  //  * hidden internal
  //  * Make this object 'constant' and 'frozen'.
  //  * Future attempts to change its state may result in an exception.
  //  * This constant object may be shared freely.
  //  * this {Brush}
  //  * return {Brush}
  //
  // private constant(): Brush {
  //   this.freeze();
  //   Object.freeze(this);
  //   return this;
  // }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Brush}
   * @return {Brush}
   */
  public freeze(): Brush {
    this._isFrozen = true;
    if (this._colorStops !== null) this._colorStops.freeze();
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * Caution: this object might be shared, which would mean that any changes
   * to this object would need to update all users of this object.
   * @this {Brush}
   * @return {Brush}
   */
  public thaw(): Brush {
    if (Object.isFrozen(this)) Util.throwError('cannot thaw constant: ' + this);
    this._isFrozen = false;
    if (this._colorStops !== null) this._colorStops.thaw();
    return this;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Brush}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Brush) {
      this.type = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {Brush}
   * @return {string}
   */
  public toString(): string {
    let s = 'Brush(';
    if (this.type === Brush.Solid) {
      s += this.color;
    } else {
      if (this.type === Brush.Linear) s += 'Linear ';
      else if (this.type === Brush.Radial) s += 'Radial ';
      else if (this.type === Brush.Pattern) s += 'Pattern ';
      else s += '(unknown) ';
      s += this.start + ' ' + this.end;
      if (this.colorStops !== null) {
        const it = this.colorStops.iterator;
        while (it.next()) {
          const key = it.key;
          const val = it.value;
          s += ' ' + key + ':' + val;
        }
      }
    }
    s += ')';
    return s;
  }

  /**
   * Specify a particular color at a particular fraction of the distance.
   * If the {@link #type} is {@link Brush.Solid}, change the type to {@link Brush.Linear}.
   * You should have a color stop at 0 and a color stop at 1.
   * You should not have duplicate color stop values at the same fractional distance.
   * @export
   * @this {Brush}
   * @param {number} loc A number between 0 and 1 (inclusive).
   * @param {string} color A valid CSS color string.
   * @return {Brush} this Brush
   */
  public addColorStop(loc: number, color: string): Brush {
    if (this._isFrozen) Util.throwFrozen(this);
    if (typeof loc !== 'number' || !isFinite(loc) || loc > 1 || loc < 0) {
      Util.throwRangeError(loc, '0 <= loc <= 1', Brush, 'addColorStop:loc');
    }
    Util.checkPrimitive(color, 'string', Brush, 'addColorStop:color');
    if (Debug && !Brush.isValidColor(color)) Util.throwError('Color \"' + color + '\" is not a valid color string for Brush.addColorStop');
    if (this._colorStops === null) this._colorStops = new Map<number, string>();
    this._colorStops.add(loc, color);
    if (this._type === Brush.Solid) this.type = Brush.Linear;
    this._constructed = null;
    return this;
  }

  /**
   * Gets or sets the type of brush.
   * The default value is {@link Brush.Solid}.
   * The value must be one of: {@link Brush.Solid}, {@link Brush.Linear}, {@link Brush.Radial}, {@link Brush.Pattern}.
   * If the new value is a linear or radial brush type,
   * and if the {@link #start} or {@link #end} spots are not specific spots,
   * they are changed to be specific spots, depending on the type of brush.
   */
  /** @export */ get type(): EnumValue { return this._type; }
  /** @export */ set type(value: EnumValue) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    EnumValue.checkEnumValue(value, Brush, Brush, 'type');
    this._type = value;
    if (this.start.isNoSpot()) {
      if (value === Brush.Linear) this.start = Spot.TopCenter;
      else if (value === Brush.Radial) this.start = Spot.Center;
    }
    if (this.end.isNoSpot()) {
      if (value === Brush.Linear) this.end = Spot.BottomCenter;
      else if (value === Brush.Radial) this.end = Spot.Center;
    }
    this._constructed = null;
  }

  /**
   * Gets or sets the color of a solid Brush.
   * The default value is 'black'.
   * The value must be a valid CSS color string.
   */
  /** @export */ get color(): string { return this._color; }
  /** @export */ set color(value: string) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    if (Debug && !Brush.isValidColor(value)) Util.throwError('Color \"' + value + '\" is not a valid color string for Brush.color');
    this._color = value;
    this._constructed = null;
  }

  /**
   * Gets or sets the starting location for a linear or radial gradient.
   * A {@link Spot} value specifies a relative point in the object's {@link GraphObject#naturalBounds}.
   * The default value is {@link Spot.TopCenter} for linear gradients and {@link Spot.Center} for radial gradients.
   */
  /** @export */ get start(): Spot { return this._start; }
  /** @export */ set start(value: Spot) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    Util.checkClass(value, Spot, Brush, 'start');
    this._start = value.copyFrozen();
    this._constructed = null;
  }

  /**
   * Gets or sets the ending location for a linear or radial gradient.
   * A {@link Spot} value specifies a relative point in the object's {@link GraphObject#naturalBounds}.
   * The default value is {@link Spot.BottomCenter} for linear gradients and {@link Spot.Center} for radial gradients.
   */
  /** @export */ get end(): Spot { return this._end; }
  /** @export */ set end(value: Spot) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    Util.checkClass(value, Spot, Brush, 'end');
    this._end = value.copyFrozen();
    this._constructed = null;
  }

  /**
   * Gets or sets the radius of a radial brush at the start location.
   * The default value is 0.
   */
  /** @export */ get startRadius(): number { return this._startRadius; }
  /** @export */ set startRadius(value: number) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    Util.checkRealNumber(value, Brush, 'startRadius');
    if (value < 0) Util.throwRangeError(value, '>= zero', Brush, 'startRadius');
    this._startRadius = value;
    this._constructed = null;
  }

  /**
   * Gets or sets the radius of a radial brush at the end location.
   * The default value is NaN.
   */
  /** @export */ get endRadius(): number { return this._endRadius; }
  /** @export */ set endRadius(value: number) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    Util.checkRealNumber(value, Brush, 'endRadius');
    if (value < 0) Util.throwRangeError(value, '>= zero', Brush, 'endRadius');
    this._endRadius = value;
    this._constructed = null;
  }

  /**
   * Gets or sets a Map holding all of the color stops used in this gradient,
   * where the key is a number, the fractional distance between 0 and 1 (inclusive),
   * and where the corresponding value is a color string.
   *
   * Call {@link #addColorStop} in order to add color stops to this brush.
   * This property value may be null if no gradient stops have been defined.
   */
  /** @export */ get colorStops(): Map<number, string> | null { return this._colorStops; }
  /** @export */ set colorStops(value: Map<number, string> | null) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    if (Debug) Util.checkClass(value, Map, Brush, 'colorStops');
    this._colorStops = value;
    this._constructed = null;
  }

  /**
   * Gets or sets the pattern of a brush of type {@link Brush.Pattern}.
   */
  /** @export */ get pattern(): HTMLCanvasElement | HTMLImageElement | null { return this._pattern; }
  /** @export */ set pattern(value: HTMLCanvasElement | HTMLImageElement | null) {
    if (this._isFrozen) Util.throwFrozen(this, value);
    // NYI type checking
    this._pattern = value;
    this._constructed = null;
  }

  /**
   * This static function can be used to generate a random color.
   * @param {number=} min A number between 0 and 255, defaults to 128.
   * @param {number=} max A number between 0 and 255, defaults to 255.
   * @return {string} A color value in # hexadecimal format.
   */
  public static randomColor(min?: number, max?: number): string {
    if (min === undefined) min = 128;
    if (Debug) {
      Util.checkRealNumber(min, Brush, 'randomColor:min');
      if (min < 0 || min > 255) Util.throwRangeError(min, '0 <= min <= 255', Brush, 'randomColor:min');
    }
    if (max === undefined) max = Math.max(min, 255);
    if (Debug) {
      Util.checkRealNumber(max, Brush, 'randomColor:max');
      if (max < min || max > 255) Util.throwRangeError(max, 'min <= max <= 255', Brush, 'randomColor:max');
    }
    const range = Math.abs(max - min);
    let r = Math.floor(min + Math.random() * range).toString(16);
    let g = Math.floor(min + Math.random() * range).toString(16);
    let b = Math.floor(min + Math.random() * range).toString(16);
    if (r.length < 2) r = '0' + r;
    if (g.length < 2) g = '0' + g;
    if (b.length < 2) b = '0' + b;
    return '#' + r + g + b;
  }

  /**
   * This static function returns true if a given color string is well-formed for drawing.
   * @param {string} color A color string to validate.
   * @return {boolean}
   * @since 1.7
   */
  public static isValidColor(color: string): boolean {
    if (color === 'black') return true;  // most common case
    if (color === '') return false;
    if (Debug) Util.checkPrimitive(color, 'string', Brush, 'isValidColor');

    if (!Brush._hasInit) {
      Brush.initStatic();
      Brush._hasInit = true;
    }
    const ctx = Brush._sharedTempCtx;
    if (ctx === null) return true;
    ctx.fillStyle = '#000000';
    const black = ctx.fillStyle;
    ctx.fillStyle = color;  // if it modifies the fillStyle, it must be a valid color
    if (ctx.fillStyle !== black) return true;

    // if the COLOR is effectively equal to black, need to try it again with a non-black
    // previous fillStyle to see if the value changes
    ctx.fillStyle = '#FFFFFF';
    const white = ctx.fillStyle;
    ctx.fillStyle = color;
    return (ctx.fillStyle !== white);
  }

  // Color modification functions

  /**
   * This static function takes a color and lightens it by 20% in the Lab color space.
   * This is a convenience function which calls {@link Brush.lightenBy}.
   * @param {string} color A valid CSS color string.
   * @return {string} A CSS string for the lightened color in RGBA.
   * @since 1.7
   */
  public static lighten(color: string): string {
    return Brush.lightenBy(color);
  }

  /**
   * Modifies all colors within this Brush, lightening them by some fraction.
   * @export
   * @this {Brush}
   * @param {number=} fraction Fraction to lighten the colors by. Defaults to 0.2, must be between 0 and 1 (inclusive).
   * @param {EnumValue=} mode Color space to use for adjusting. Must be {@link Brush.Lab} or {@link Brush.HSL}, defaults to {@link Brush.Lab}.
   * @return {Brush} This Brush with modified color values.
   * @since 1.7
   */
  public lightenBy(fraction?: number, mode?: EnumValue): Brush {
    if (this._isFrozen) Util.throwFrozen(this);
    const frac = (fraction === undefined || !(typeof fraction === 'number')) ? 0.2 : fraction;
    const md = (mode === undefined) ? Brush.Lab : mode;

    if (this.type === Brush.Solid) {
      Brush._colorToRgba(this.color);
      this.color = Brush._shadeRgba(frac, md);
    } else if (this.type === Brush.Linear || this.type === Brush.Radial) {
      if (this.colorStops !== null) {
        const it = this.colorStops.iterator;
        while (it.next()) {
          Brush._colorToRgba(it.value);
          this.addColorStop(it.key, Brush._shadeRgba(frac, md));
        }
      }
    }
    return this;
  }

  /**
   * This static function takes a color and lightens it.
   * @param {string} color A valid CSS color string.
   * @param {number=} fraction Fraction to lighten the colors by. Defaults to 0.2, must be between 0 and 1 (inclusive).
   * @param  {EnumValue=} mode Color space to use for adjusting. Must be {@link Brush.Lab} or {@link Brush.HSL}, defaults to {@link Brush.Lab}.
   * @return {string} A CSS string for the lightened color in RGBA or HSLA.
   * @since 1.7
   */
  public static lightenBy(color: string, fraction?: number, mode?: EnumValue): string {
    const frac = (fraction === undefined || !(typeof fraction === 'number')) ? 0.2 : fraction;
    const md = (mode === undefined) ? Brush.Lab : mode;

    Brush._colorToRgba(color);
    return Brush._shadeRgba(frac, md);
  }

  /**
   * This static function takes a color and darkens it by 20% in the Lab color space.
   * This is a convenience function which calls {@link Brush.darkenBy}.
   * @param {string} color A valid CSS color string.
   * @return {string} A CSS string for the darkened color in RGBA.
   * @since 1.7
   */
  public static darken(color: string): string {
    return Brush.darkenBy(color);
  }

  /**
   * Modifies all colors within this Brush, darkening them by some fraction.
   * @export
   * @this {Brush}
   * @param {number=} fraction Fraction to darken the colors by. Defaults to 0.2, must be between 0 and 1 (inclusive).
   * @param {EnumValue=} mode Color space to use for adjusting. Must be {@link Brush.Lab} or {@link Brush.HSL}, defaults to {@link Brush.Lab}.
   * @return {Brush} This Brush with modified color values.
   * @since 1.7
   */
  public darkenBy(fraction?: number, mode?: EnumValue): Brush {
    if (this._isFrozen) Util.throwFrozen(this);
    const frac = (fraction === undefined || !(typeof fraction === 'number')) ? 0.2 : fraction;
    const md = (mode === undefined) ? Brush.Lab : mode;

    if (this.type === Brush.Solid) {
      Brush._colorToRgba(this.color);
      this.color = Brush._shadeRgba(-frac, md);
    } else if (this.type === Brush.Linear || this.type === Brush.Radial) {
      if (this.colorStops !== null) {
        const it = this.colorStops.iterator;
        while (it.next()) {
          Brush._colorToRgba(it.value);
          this.addColorStop(it.key, Brush._shadeRgba(-frac, md));
        }
      }
    }
    return this;
  }

  /**
   * This static function takes a color and darkens it.
   * @param {string} color A valid CSS color string
   * @param {number=} fraction Fraction to darken the color by. Defaults to 0.2, must be between 0 and 1 (inclusive).
   * @param {EnumValue=} mode Color space to use for adjusting. Must be {@link Brush.Lab} or {@link Brush.HSL}, defaults to {@link Brush.Lab}.
   * @return {string} A CSS string for the darkened color in RGBA or HSLA.
   * @since 1.7
   */
  public static darkenBy(color: string, fraction?: number, mode?: EnumValue): string {
    const frac = (fraction === undefined || !(typeof fraction === 'number')) ? 0.2 : fraction;
    const md = (mode === undefined) ? Brush.Lab : mode;

    Brush._colorToRgba(color);
    return Brush._shadeRgba(-frac, md);
  }

  /**
   * This static function takes two colors and mixes them together, using the (optionally) specified amount of the second color.
   * @param {string} color1 A valid CSS color string.
   * @param {string} color2 Another valid CSS color string to mix.
   * @param {number=} fraction Fraction specifying how much color2 to mix into color1. Defaults to .5, must be between 0 and 1 (inclusive).
   * @return {boolean}
   * @since 2.0
   */
  public static mix(color1: string, color2: string, fraction?: number): string {
    Brush._colorToRgba(color1);
    const rgb1r = Brush._rgbNums.n0;
    const rgb1g = Brush._rgbNums.n1;
    const rgb1b = Brush._rgbNums.n2;
    const rgb1a = Brush._rgbNums.n3;
    Brush._colorToRgba(color2);

    if (fraction === undefined) fraction = .5;

    const r = Math.round(((Brush._rgbNums.n0 - rgb1r) * fraction) + rgb1r);
    const g = Math.round(((Brush._rgbNums.n1 - rgb1g) * fraction) + rgb1g);
    const b = Math.round(((Brush._rgbNums.n2 - rgb1b) * fraction) + rgb1b);
    const a = Math.round(((Brush._rgbNums.n3 - rgb1a) * fraction) + rgb1a);

    return `rgba(${r}, ${g}, ${b}, ${a})`;
  }

  /**
   * This function determines whether this Brush is "dark."
   * @export
   * @this {Brush}
   * @return {boolean}
   * @since 2.0
   */
  public isDark(): boolean {
    if (this.type === Brush.Solid) {
      return Brush.isDark(this.color);
    } else if (this.type === Brush.Linear || this.type === Brush.Radial) {
      if (this.colorStops !== null) {
        const colStops = this.colorStops;
        // Radial, just take center color
        if (this.type === Brush.Radial) {
          return Brush.isDark(colStops.first().value);
        }
        // Linear
        // If a mid color is specified use it
        if (colStops.get(.5) !== null) {
          return Brush.isDark(colStops.get(.5));
        }
        // Handle simple case of two colors
        if (colStops.count === 2) {
          const cols = colStops.toArray();
          return Brush.isDark(Brush.mix(cols[0].value, cols[1].value));
        }
        // Otherwise find the colors closest to the middle and interpolate between them
        const it = colStops.iterator;
        let key1 = -1;
        let key2 = -1;
        let diff1 = 1;
        let diff2 = 1;
        while (it.next()) {
          const key = it.key;
          const diffMid = Math.abs(.5 - it.key);
          if (diff1 > diff2 && diffMid < diff1) {
            key1 = key;
            diff1 = diffMid;
          } else if (diff2 >= diff1 && diffMid < diff2) {
            key2 = key;
            diff2 = diffMid;
          }
        }
        if (key1 > key2) key1 = [key2, key2 = key1][0]; // one-liner, swap when needed
        const diff = key2 - key1;
        return Brush.isDark(Brush.mix(colStops.get(key1), colStops.get(key2), 1 - (diff1 / diff)));
      }
    }
    return false;
  }

  /**
   * This static function takes a color and determines whether it is "dark." Does not account for transparency.
   *
   * Example usage:
   * ```js
   * myDiagram.nodeTemplate =
   *   $(go.Node, "Auto",
   *     $(go.Shape, "RoundedRectangle", { strokeWidth: 0 },
   *       new go.Binding("fill", "color")),
   *     $(go.TextBlock,
   *       { margin: 8 },
   *       new go.Binding("stroke", "color", function (c) {
   *         // Dark nodes use white text, light nodes use black text
   *         return go.Brush.isDark(c) ? "white" : "black";
   *       }),
   *       new go.Binding("text", "key")
   *     )
   *   );
   * ```
   * @param {Brush|string} color A valid CSS color string or a Brush.
   * @return {boolean}
   * @since 2.0
   */
  public static isDark(color: BrushLike): boolean {
    if (color === null) return null;
    if (color instanceof Brush) {
      return color.isDark();
    }
    Brush._colorToRgba(color);
    // get perceived brightness of color: http://www.w3.org/TR/AERT#color-contrast
    const brightness = (299 * Brush._rgbNums.n0 + 587 * Brush._rgbNums.n1 + 114 * Brush._rgbNums.n2) / 1000;
    return brightness < 128;
  }

  /**
   * @hidden @internal
   * @param {number} fraction
   * @param {EnumValue} mode
   * @return {string}
   */
  public static _shadeRgba(fraction: number, mode: EnumValue): string {
    switch (mode) {
      case Brush.Lab:
        Brush._rgbToLab();
        Brush._labNums.n0 = Math.min(100, Math.max(0, Brush._labNums.n0 + 100 * fraction));
        Brush._labToRgb();
        return 'rgba(' + Brush._rgbNums.n0 + ', ' + Brush._rgbNums.n1 + ', ' + Brush._rgbNums.n2 + ', ' + Brush._rgbNums.n3 + ')';
      case Brush.HSL:
        Brush._rgbaToHsla();
        Brush._hslNums.n2 = Math.min(100, Math.max(0, Brush._hslNums.n2 + 100 * fraction));
        return 'hsla(' + Brush._hslNums.n0 + ', ' + Brush._hslNums.n1 + '%, ' + Brush._hslNums.n2 + '%, ' + Brush._hslNums.n3 + ')';
      default:
        Util.throwError('Unknown color space: ' + mode);
        return 'rgba(0, 0, 0, 1)';
    }
  }

  /**
   * @hidden @internal
   * This static function converts a CSS color string to an RGBA array.
   * @param {string} color The CSS color string
   */
  public static _colorToRgba(color: string): void {
    if (!Brush._hasInit) {
      Brush.initStatic();
      Brush._hasInit = true;
    }
    // This is a combination of determining color and also validating it
    const ctx = Brush._sharedTempCtx;
    if (ctx === null) return;
    ctx.clearRect(0, 0, 1, 1);
    ctx.fillStyle = '#000000';
    const black = ctx.fillStyle;
    ctx.fillStyle = color;
    if (ctx.fillStyle !== black) { // if it's not black it changed, so it's valid
      ctx.fillRect(0, 0, 1, 1);
      const data = ctx.getImageData(0, 0, 1, 1).data;
      Brush._rgbNums.n0 = data[0];
      Brush._rgbNums.n1 = data[1];
      Brush._rgbNums.n2 = data[2];
      Brush._rgbNums.n3 = data[3] / 255;
    } else {
      ctx.fillStyle = '#FFFFFF';
      const white = ctx.fillStyle;
      ctx.fillStyle = color;
      if (ctx.fillStyle === white) {
        if (Debug) Util.throwError('Color \"' + color + '\" is not a valid color string for RGBA color conversion');
      }
      // Set the RGBA to black since it's either actually black or it's invalid and we aren't in debug
      Brush._rgbNums.n0 = 0;
      Brush._rgbNums.n1 = 0;
      Brush._rgbNums.n2 = 0;
      Brush._rgbNums.n3 = 1;
    }
  }

  /**
   * @hidden @internal
   * This static function converts an RGBA color value to HSLA.
   * Assumes r, g, and b are contained in the set [0, 255] and a is in [0, 1].
   * Returns h [0, 360], s [0, 100], l [0, 100], and a [0, 1].
   */
  public static _rgbaToHsla(): void {
    const r = Brush._rgbNums.n0 / 255;
    const g = Brush._rgbNums.n1 / 255;
    const b = Brush._rgbNums.n2 / 255;
    const max = Math.max(r, g, b);
    const min = Math.min(r, g, b);
    const delta = max - min;
    let h;
    let s;
    let l;

    // Lightness
    l = (max + min) / 2;
    if (delta === 0) {
      h = s = 0; // achromatic
    } else {
      // Hue
      switch (max) {
        case r: h = (g - b) / delta % 6; break;
        case g: h = (b - r) / delta + 2; break;
        case b: h = (r - g) / delta + 4; break;
      }
      h *= 60;
      if (h < 0) h += 360;
      // Saturation
      s = delta / (1 - Math.abs(2 * l - 1));
    }

    Brush._hslNums.n0 = Math.round(h);
    Brush._hslNums.n1 = Math.round(s * 100);
    Brush._hslNums.n2 = Math.round(l * 100);
    Brush._hslNums.n3 = Brush._rgbNums.n3;
  }

  /**
   * @hidden @internal
   * This static function converts an HSLA color value to RGBA.
   * Assumes h [0, 360], s [0, 100], l [0, 100], and a [0, 1].
   * Returns r, g, and b in the set [0, 255] and a in the set [0, 1].
   */
  public static _hslaToRgba(): void {
    const h = Brush._hslNums.n0;
    const s = Brush._hslNums.n1 / 100;
    const l = Brush._hslNums.n2 / 100;
    let r;
    let g;
    let b;

    // Find chroma
    const c = (1 - Math.abs(2 * l - 1)) * s;
    // Match hue and chroma along RGB cube
    const h1 = h / 60;
    const x = c * (1 - Math.abs(h1 % 2 - 1));
    if (h1 >= 0 && h1 < 1) {
      r = c;
      g = x;
      b = 0;
    } else if (h1 >= 1 && h1 < 2) {
      r = x;
      g = c;
      b = 0;
    } else if (h1 >= 2 && h1 < 3) {
      r = 0;
      g = c;
      b = x;
    } else if (h1 >= 3 && h1 < 4) {
      r = 0;
      g = x;
      b = c;
    } else if (h1 >= 4 && h1 < 5) {
      r = x;
      g = 0;
      b = c;
    } else if (h1 >= 5 && h1 < 6) {
      r = c;
      g = 0;
      b = x;
    }
    // Match lightness
    const m = l - .5 * c;
    r += m;
    g += m;
    b += m;

    Brush._rgbNums.n0 = Math.round(r * 255);
    Brush._rgbNums.n1 = Math.round(g * 255);
    Brush._rgbNums.n2 = Math.round(b * 255);
    Brush._rgbNums.n3 = Brush._hslNums.n3;
  }

  /**
   * @hidden @internal
   * This static function converts an RGBA color to Lab.
   * Assumes r, g, and b are contained in the set [0, 255].
   */
  public static _rgbToLab(): void {
    Brush._rgbToXyz();
    Brush._xyzToLab();
  }

  /**
   * @hidden @internal
   * This static function converts a Lab color to RGB.
   */
  public static _labToRgb(): void {
    Brush._labToXyz();
    Brush._xyzToRgb();
  }

  /**
   * @hidden @internal
   * This static function performs inverse sRGB companding.
   * @param {number} rgbVal An R, G, or B value in the set [0, 255]
   * @return {number} Inverse companded RGB value
   */
  public static _rgbXyzInverseCompanding(rgbVal: number): number {
    rgbVal /= 255;
    if (rgbVal <= .04045) {
      return rgbVal / 12.92;
    }
    return Math.pow(((rgbVal + .055) / 1.055), 2.4);
  }

  /**
   * @hidden @internal
   * This static function performs sRGB companding.
   * @param {number} xyzVal An X, Y, or Z value
   * @return {number} Companded XYZ value
   */
  public static _xyzRgbCompanding(xyzVal: number): number {
    if (xyzVal <= .0031308) {
      return xyzVal * 12.92;
    }
    return 1.055 * Math.pow(xyzVal, 1 / 2.4) - .055;
  }

  /**
   * @hidden @internal
   * This static function converts an RGB color to XYZ.
   * This is an intermediate step in the RGB to Lab conversion.
   */
  public static _rgbToXyz(): void {
    // Inverse companding
    const x = 100 * Brush._rgbXyzInverseCompanding(Brush._rgbNums.n0);
    const y = 100 * Brush._rgbXyzInverseCompanding(Brush._rgbNums.n1);
    const z = 100 * Brush._rgbXyzInverseCompanding(Brush._rgbNums.n2);
    // Matrix multiplication
    Brush._xyzNums.n0 = 0.4124564 * x + 0.3575761 * y + 0.1804375 * z;
    Brush._xyzNums.n1 = 0.2126729 * x + 0.7151522 * y + 0.0721750 * z;
    Brush._xyzNums.n2 = 0.0193339 * x + 0.1191920 * y + 0.9503041 * z;
    Brush._xyzNums.n3 = Brush._rgbNums.n3;
  }

  /**
   * @hidden @internal
   * This static function converts an XYZ color to RGB.
   * This is an intermediate step in the Lab to RGB conversion.
   */
  public static _xyzToRgb(): void {
    // Matrix multiplication
    const r = 3.2404542 * Brush._xyzNums.n0 + -1.5371385 * Brush._xyzNums.n1 + -0.4985314 * Brush._xyzNums.n2;
    const g = -0.9692660 * Brush._xyzNums.n0 + 1.8760108 * Brush._xyzNums.n1 + 0.0415560 * Brush._xyzNums.n2;
    const b = 0.0556434 * Brush._xyzNums.n0 + -0.2040259 * Brush._xyzNums.n1 + 1.0572252 * Brush._xyzNums.n2;
    // Companding
    Brush._rgbNums.n0 = Brush._xyzRgbCompanding(r / 100) * 255;
    Brush._rgbNums.n1 = Brush._xyzRgbCompanding(g / 100) * 255;
    Brush._rgbNums.n2 = Brush._xyzRgbCompanding(b / 100) * 255;
    Brush._rgbNums.n3 = Brush._xyzNums.n3;
    // Sanitize the RGB values so they'll be valid for CSS
    Brush._rgbNums.n0 = Math.round(Brush._rgbNums.n0);
    if (Brush._rgbNums.n0 > 255) Brush._rgbNums.n0 = 255;
    else if (Brush._rgbNums.n0 < 0) Brush._rgbNums.n0 = 0;
    Brush._rgbNums.n1 = Math.round(Brush._rgbNums.n1);
    if (Brush._rgbNums.n1 > 255) Brush._rgbNums.n1 = 255;
    else if (Brush._rgbNums.n1 < 0) Brush._rgbNums.n1 = 0;
    Brush._rgbNums.n2 = Math.round(Brush._rgbNums.n2);
    if (Brush._rgbNums.n2 > 255) Brush._rgbNums.n2 = 255;
    else if (Brush._rgbNums.n2 < 0) Brush._rgbNums.n2 = 0;
  }

  /** @hidden @internal */ public static _LabEPSILON = 216 / 24389;
  /** @hidden @internal */ public static _LabKAPPA = 24389 / 27;
  /** @hidden @internal */ public static _xyzWhiteRef = [95.047, 100, 108.883]; // D65 Illuminant chromatic adaptation

  /**
   * @hidden @internal
   * @param {number} inputXyz
   * @return {number}
   */
  public static _xyzLabHelperFunction(inputXyz: number): number {
    if (inputXyz > Brush._LabEPSILON) {
      return Math.pow(inputXyz, 1 / 3);
    }
    return (Brush._LabKAPPA * inputXyz + 16) / 116;
  }

  /**
   * @hidden @internal
   */
  public static _xyzToLab(): void {
    const fx = Brush._xyzLabHelperFunction(Brush._xyzNums.n0 / Brush._xyzWhiteRef[0]);
    const fy = Brush._xyzLabHelperFunction(Brush._xyzNums.n1 / Brush._xyzWhiteRef[1]);
    const fz = Brush._xyzLabHelperFunction(Brush._xyzNums.n2 / Brush._xyzWhiteRef[2]);

    Brush._labNums.n0 = 116 * fy - 16;
    Brush._labNums.n1 = 500 * (fx - fy);
    Brush._labNums.n2 = 200 * (fy - fz);
    Brush._labNums.n3 = Brush._xyzNums.n3;
  }

  /**
   * @hidden @internal
   * @param {number} inputLab
   * @return {number}
   */
  public static _labXyzHelperFunction(inputLab: number): number {
    const inputCbd = inputLab * inputLab * inputLab;
    if (inputCbd > Brush._LabEPSILON) {
      return inputCbd;
    }
    return (116 * inputLab - 16) / Brush._LabKAPPA;
  }

  /**
   * @hidden @internal
   */
  public static _labToXyz(): void {
    const fy = (Brush._labNums.n0 + 16) / 116;
    const fx = (Brush._labNums.n1 / 500) + fy;
    const fz = fy - (Brush._labNums.n2 / 200);

    Brush._xyzNums.n0 = Brush._xyzWhiteRef[0] * Brush._labXyzHelperFunction(fx);
    Brush._xyzNums.n1 = Brush._xyzWhiteRef[1] * (Brush._labNums.n0 > Brush._LabKAPPA * Brush._LabEPSILON ? Math.pow(fy, 3) : Brush._labNums.n0 / Brush._LabKAPPA);
    Brush._xyzNums.n2 = Brush._xyzWhiteRef[2] * Brush._labXyzHelperFunction(fz);
    Brush._xyzNums.n3 = Brush._labNums.n3;
  }

  /**
   * @hidden @internal
   * @param {*} val expected to be either a color string or a Brush
   * @param {string} where
   */
  public static checkBrushString(val: any, where: string): void {
    if (typeof val === 'string') {
      if (Debug && !Brush.isValidColor(val)) Util.throwError('Color \"' + val + '\" is not a valid color string for ' + where);
    } else if (!(val instanceof Brush)) {
      Util.throwError('Value for ' + where + ' must be a color string or a Brush, not ' + val);
    }
    // end checkBrushString
  }

  /**
   * For simple, solid color brushes, used as the value for {@link Brush#type}.
   * @constant
   */
  public static Solid: EnumValue;

  /**
   * For linear gradient brushes, used as the value for {@link Brush#type}.
   * @constant
   */
  public static Linear: EnumValue;

  /**
   * For radial gradient brushes, used as the value for {@link Brush#type}.
   * @constant
   */
  public static Radial: EnumValue;

  /**
   * For pattern brushes, used as the value for {@link Brush#type}.
   * @constant
   */
  public static Pattern: EnumValue;

  /**
   * For lightening and darkening, used as a color-space value.
   * @constant
   */
  public static Lab: EnumValue;

  /**
   * For lightening and darkening, used as a color-space value.
   * @constant
   */
  public static HSL: EnumValue;

  // This static canvas is used for testing and converting colors
  /** @hidden @internal @const */
  public static _sharedTempCtx: CanvasSurfaceContext = null;

  // We only ever need to keep track of one color per color space, so preallocate them
  /** @hidden @internal */ public static _rgbNums: ColorNumbers = new ColorNumbers();
  /** @hidden @internal */ public static _hslNums: ColorNumbers = new ColorNumbers();
  /** @hidden @internal */ public static _xyzNums: ColorNumbers = new ColorNumbers();
  /** @hidden @internal */ public static _labNums: ColorNumbers = new ColorNumbers();

  /** @hidden @internal */
  private static _hasInit: boolean = false;
  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Brush._sharedTempCtx = Diagram.isUsingDOM() ? new CanvasSurface(null).context : null;
  }
} // #end Brush
Brush['className'] = 'Brush';

Brush['Solid'] = Brush.Solid = new EnumValue(Brush, 'Solid', 0);
Brush['Linear'] = Brush.Linear = new EnumValue(Brush, 'Linear', 1);
Brush['Radial'] = Brush.Radial = new EnumValue(Brush, 'Radial', 2);
Brush['Pattern'] = Brush.Pattern = new EnumValue(Brush, 'Pattern', 4);
Brush['Lab'] = Brush.Lab = new EnumValue(Brush, 'Lab', 5);
Brush['HSL'] = Brush.HSL = new EnumValue(Brush, 'HSL', 6);
Brush['randomColor'] = Brush.randomColor;
Brush['isValidColor'] = Brush.isValidColor;
Brush['lighten'] = Brush.lighten;
Brush['lightenBy'] = Brush.lightenBy;
Brush['darken'] = Brush.darken;
Brush['darkenBy'] = Brush.darkenBy;
Brush['mix'] = Brush.mix;
Brush['isDark'] = Brush.isDark;


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayout
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';

/**
 * This is the abstract base class for all Panel Layouts, which inform the possible Panel types.
 * It is possible to create your own Panel type by creating a subclass of PanelLayout,
 * though this is not common and not recommended for beginners.
 *
 * By default, GoJS has 12 panel types, each corresponding to a PanelLayout subclass:
 *   - `'Position', PanelLayoutPosition`
 *   - `'Horizontal', PanelLayoutHorizontal`
 *   - `'Vertical', PanelLayoutVertical`
 *   - `'Spot', PanelLayoutSpot`
 *   - `'Auto', PanelLayoutAuto`
 *   - `'Table', PanelLayoutTable`
 *   - `'Viewbox', PanelLayoutViewbox`
 *   - `'TableRow', PanelLayoutTableRow`
 *   - `'TableColumn', PanelLayoutTableColumn`
 *   - `'Link', PanelLayoutLink`
 *   - `'Grid', PanelLayoutGrid`
 *   - `'Graduated', PanelLayoutGraduated`
 *
 * These are included by default in builds of `go.js` and `go-debug.js`.
 * When building from source, you can optionally exclude all of them except `Position`,
 * `Vertical`, `Auto`, `Link`, and `Grid`.
 * This is demonstrated in `minimalSource` and `maximalSource`, in the `/projects` folder.
 *
 * Adding a new Layout is done by calling the static function, {@link Panel.definePanelLayout}:
 *
 * ```js
 * Panel.definePanelLayout('Table', new PanelLayoutTable());
 * ```
 *
 * Each PanelLayout defines a {@link #measure} and {@link #arrange} routine.
 * The measure routine must call {@link #measureElement} with each element of the Panel,
 * and the arrange routine must similarly call {@link #arrangeElement} with each element of the Panel.
 *
 * There is an example PanelLayout in the <a href="../../samples/panelLayout.html">PanelLayout sample</a>.
 * @since 2.0
 * @unrestricted
 */
export abstract class PanelLayout {
  /** @hidden @internal */ public static className = 'PanelLayout';
  public name: string;

  constructor() {
    this.name = 'Base';
  }

  /**
   * @hidden
   */
  /** @export */ get classType(): Function {
    return Panel;
  }

  /**
   * Given the available size, measure the Panel and
   * determine its expected drawing size. Sets the measuredBounds of the object.
   *
   * This must call {@link #measureElement} with each Panel element.
   *
   * This must also construct the union.width and union.height of the passed in union Rect argument.
   * This union must reflect the measured size of the panel.
   *
   * @expose
   * @this {PanelLayout}
   * @param {Panel} panel Panel which called this layout
   * @param {number} width expected width of the panel
   * @param {number} height expected width of the panel
   * @param {Array.<GraphObject>} elements Array of Panel elements
   * @param {Rect} union rectangle to contain the expected union bounds of every element in the Panel. Useful for arrange.
   * @param {number} minw minimum width of the panel
   * @param {number} minh minimum height of the panel
   */
  public measure(panel: Panel, width: number, height: number, elements: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    // override this in base classes. See the API for details
  }


  /**
   * Given the available size, measure one element of the Panel and
   * determine its expected drawing size. Sets the measuredBounds of the object.
   *
   * @export
   * @this {PanelLayout}
   * @param {GraphObject} obj Panel which called this layout
   * @param {number} width expected width of the GraphObject
   * @param {number} height expected width of the GraphObject
   * @param {number} minw minimum width of the GraphObject
   * @param {number} minh minimum height of the GraphObject
   */
  protected measureElement(obj: GraphObject, width: number, height: number, minw: number, minh: number) {
    obj.measure(width, height, minw, minh);
  }

  /**
   * Given the panel and its list of elements, arrange each element.
   *
   * This must call {@link #arrangeElement} with each Panel element, which will set that element's {@link GraphObject#actualBounds}.
   *
   * For arranging some elements, it is useful to know the total unioned area of every element.
   * This Rect can be used to right-align or center-align, etc, elements within an area.
   *
   * @expose
   * @this {PanelLayout}
   * @param {Panel} panel Panel which called this layout
   * @param {Array.<GraphObject>} elements Array of Panel elements
   * @param {Rect} union rectangle, if properly constructed in {@link #measure}, that contains the expected union bounds of every element in the Panel.
   */
  public arrange(panel: Panel, elements: Array<GraphObject>, union: Rect): void {
    // override this in base classes. See the API for details
  }


  /**
   * Arranges the GraphObject onto its parent Panel.
   * The passed-in numbers typically account for {@link GraphObject#margin} and other offsets.
   * This sets {@link GraphObject#actualBounds}.
   *
   * @export
   * @this {GraphObject}
   * @param {GraphObject} obj GraphObject to be arranged.
   * @param {number} fx The final x value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fy The final y value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fw The final width value of actualBounds that the Panel computes for the GraphObject.
   * @param {number} fh The final height value of actualBounds that the Panel computes for the GraphObject.
   * @param {Rect=} clipRect an optional area to constrain this actualBounds to when picking and drawing.
   * By default, this is only used with Fixed/Table panels element, provided as a Rect.
   */
  protected arrangeElement(obj: GraphObject, fx: number, fy: number, fw: number, fh: number, clipRect?: Rect): void {
    obj.arrange(fx, fy, fw, fh, clipRect);
  }


  /**
   * @hidden @internal
   * Undocumented
   */
  public draw(panel: Panel, context: IContext, diagram: Diagram): void {
    // override this in base classes. See the API for details
  }

}
PanelLayout['className'] = 'PanelLayout';

// import { Diagram } from './diagram';
// import { Rect } from './rect';
// import { IContext } from './surfaceInterfaces';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutPosition
*/

// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Rect } from './rect';
// import { Shape } from './shape';


/**
 * @hidden @internal
 * PanelLayoutPosition
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutPosition extends PanelLayout {
  constructor() {
    super();
    this.name = 'Position';
  }

  /**
   * @hidden @internal
   * Measure all elements to get the bounding box of the panel
   * @this {PanelLayoutPosition}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    const locObj = panel.getLocationObject();
    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (!child.visible && child !== locObj) continue;
      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(width, height, minw, minh);
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + margw, 0);
      const mheight = Math.max(m.height + margh, 0);

      let x = child.position.x;
      let y = child.position.y;
      if (!isFinite(x)) { x = 0; }
      if (!isFinite(y)) { y = 0; }
      if (child instanceof Shape) {
        if (child.isGeometryPositioned) {
          const sw = child.strokeWidth / 2;
          x -= sw;
          y -= sw;
        }
      }
      union.unionNoCheck(x, y, mwidth, mheight);
    }
  }

  /**
   * @hidden @internal
   * Arrange children according to their positions
   * @this {PanelLayoutPosition}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const padding = panel.padding as Margin;
    const ux = union.x - padding.left;
    const uy = union.y - padding.top;
    for (let i = 0; i < l; i++) {
      const child = children[i];
      const m = child.measuredBounds;
      const marg = child.margin as Margin;
      const x = child.position.x;
      const y = child.position.y;
      let arx = isNaN(x) ? -ux : x - ux;
      let ary = isNaN(y) ? -uy : y - uy;
      if (child instanceof Shape) {
        if (child.isGeometryPositioned) {
          const sw = child.strokeWidth / 2;
          arx -= sw;
          ary -= sw;
        }
      }
      if (child.visible) {
        child.arrange(arx + marg.left, ary + marg.top, m.width, m.height);
      }
    }
  }

}



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutHorizontal
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * @hidden @internal
 * PanelLayoutHorizontal
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutHorizontal extends PanelLayout {
  constructor() {
    super();
    this.name = 'Horizontal';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutHorizontal}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    const strchildren = Util.tempArray<GraphObject>();
    const locObj = panel.getLocationObject();
    // let lastChildStretch: GraphObject = null; // lastChildStretch code would enable the last child to stretch horizontally.
    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (!child.visible && child !== locObj) continue;
      const stretch = child.getEffectiveStretch(false);
      // if ((stretch === GraphObject.Fill || stretch === GraphObject.Horizontal) && i === l - 1) {
      //   lastChildStretch = child;
      //   continue;
      // }
      if (stretch !== GraphObject.None && stretch !== GraphObject.Horizontal) {
        strchildren.push(child);
        continue;
      }

      child.measure(Infinity, height, 0, minh); // measure children with infinite width
      const marg = child.margin as Margin;
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + marg.right + marg.left, 0);
      const mheight = Math.max(m.height + marg.top + marg.bottom, 0);

      union.width += mwidth;
      union.height = Math.max(union.height, mheight);
    }

    const len = strchildren.length;
    // modify height for stretch children
    if (panel.desiredSize.height) {
      height = Math.min(panel.desiredSize.height, panel.maxSize.height);
    } else if (union.height !== 0) {
      height = Math.min(union.height, panel.maxSize.height);
    }

    for (let i = 0; i < len; i++) {
      const child = strchildren[i];
      if (!child.visible && child !== locObj) continue;
      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(Infinity, height, 0, minh); // measure children with infinite width
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + margw, 0);
      const mheight = Math.max(m.height + margh, 0);

      union.width += mwidth;
      union.height = Math.max(union.height, mheight);
    }

    // if (lastChildStretch !== null) {
    //   const remainingWidth = width - union.width;
    //   const child = lastChildStretch;
    //   const marg = child.margin;
    //   const margw = marg.right + marg.left;
    //   const margh = marg.top + marg.bottom;
    //   child.measure(remainingWidth, height, 0, minh); // measure children with infinite width
    //   const m = child.measuredBounds;
    //   const mwidth = Math.max(m.width + margw, 0);
    //   const mheight = Math.max(m.height + margh, 0);

    //   union.width += mwidth;
    //   union.height = Math.max(union.height, mheight);
    // }

    Util.freeArray(strchildren);
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutHorizontal}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const padding = panel.padding as Margin;
    const pt = padding.top;
    const opposite = panel.isOpposite;
    let x = opposite ? union.width : padding.left; // x is incremented or decremented on each child
    for (let i = 0; i < l; i++) {
      const y = pt; // y must be reset each time
      const child = children[i];
      if (!child.visible) continue;
      const m = child.measuredBounds;
      const marg = child.margin as Margin;
      const margh = marg.top + marg.bottom;
      const padh = pt + padding.bottom;
      let height = m.height;
      const stretch = child.getEffectiveStretch(false);
      // child desired size!
      if (isNaN(child.desiredSize.height) && stretch === GraphObject.Fill || stretch === GraphObject.Vertical) {
        height = Math.max(union.height - margh - padh, 0);
      }
      const heighttotal = height + margh + padh; // total vertical space a child occupies

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;

      if (opposite) x -= (m.width + marg.left + marg.right);
      child.arrange(
        x + align.offsetX + marg.left,
        (y + align.offsetY + marg.top) + (union.height * align.y - (heighttotal * align.y)),
        m.width,
        height);
      if (!opposite) x += (m.width + marg.left + marg.right);
    }
  }
}

// import { Margin } from './margin';
// import { Rect } from './rect';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutVertical
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * @hidden @internal
 * PanelLayoutVertical
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutVertical extends PanelLayout {
  constructor() {
    super();
    this.name = 'Vertical';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutVertical}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    const strchildren = Util.tempArray<GraphObject>();
    const locObj = panel.getLocationObject();
    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (!child.visible && child !== locObj) continue;
      const stretch = child.getEffectiveStretch(false);
      if (stretch !== GraphObject.None && stretch !== GraphObject.Vertical) {
        strchildren.push(child);
        continue;
      }

      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(width, Infinity, minw, 0); // measure children with infinite height
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + margw, 0);
      const mheight = Math.max(m.height + margh, 0);

      union.assignWidthHeight(Math.max(union.width, mwidth), union.height + mheight);
    }

    const len = strchildren.length;
    if (len === 0) return;
    // modify width for stretch children
    if (panel.desiredSize.width) {
      width = Math.min(panel.desiredSize.width, panel.maxSize.width);
    } else if (union.width !== 0) {
      width = Math.min(union.width, panel.maxSize.width);
    }

    for (let i = 0; i < len; i++) {
      const child = strchildren[i];
      if (!child.visible && child !== locObj) continue;
      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(width, Infinity, minw, 0); // measure children with infinite height
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + margw, 0);
      const mheight = Math.max(m.height + margh, 0);

      union.width = Math.max(union.width, mwidth);
      union.height += mheight;
    }
    Util.freeArray(strchildren);
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutVertical}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const padding = panel.padding as Margin;
    const pl = padding.left;
    const opposite = panel.isOpposite;
    let y = opposite ? union.height : padding.top; // y is incremented or decremented on each child
    for (let i = 0; i < l; i++) {
      const x = pl; // x must be reset each time
      const child = children[i];
      if (!child.visible) continue;
      const m = child.measuredBounds;
      const marg = child.margin as Margin;
      const margw = marg.left + marg.right;
      const padw = pl + padding.right;
      let width = m.width;
      const stretch = child.getEffectiveStretch(false);
      // child desired size!
      if (isNaN(child.desiredSize.width) && stretch === GraphObject.Fill || stretch === GraphObject.Horizontal) {
        width = Math.max(union.width - margw - padw, 0);
      }
      const widthtotal = width + margw + padw; // total horizontal space a child occupies

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;

      if (opposite) y -= (m.height + marg.bottom + marg.top);
      child.arrange(
        (x + align.offsetX + marg.left) + (union.width * align.x - (widthtotal * align.x)),
        y + align.offsetY + marg.top,
        width,
        m.height);
      if (!opposite) y += (m.height + marg.bottom + marg.top);
    }
  }
}

// import { Margin } from './margin';
// import { Rect } from './rect';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutSpot
*/


// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Spot } from './spot';


/**
 * @hidden @internal
 * PanelLayoutSpot
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutSpot extends PanelLayout {
  constructor() {
    super();
    this.name = 'Spot';
  }

  /**
   * @hidden @internal
   * Elements measured according to alignment spots around a main element
   * @this {PanelLayoutSpot}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    const main = panel.findMainElement();
    // If the panel isn't stretch this is fine.
    // If it is stretch, we will have to remeasure the main object and recreate the union
    let marg = main.margin as Margin;
    let margw = 0.0;
    let margh = 0.0; // for children
    const mainmargw = marg.right + marg.left;
    const mainmargh = marg.top + marg.bottom;
    main.measure(width, height, minw, minh);
    let m = main.measuredBounds;
    let mainWidth = m.width;
    let mainHeight = m.height;
    let mwidth = Math.max(mainWidth + mainmargw, 0);
    let mheight = Math.max(mainHeight + mainmargh, 0);

    const isClip = panel.isClipping;

    // we want union.x to be the main object's top left corner, not where its margin is
    let mainRect = Rect.allocAt(-marg.left, -marg.top, mwidth, mheight);

    let first = true;
    const locObj = panel.getLocationObject();
    let stretch;
    for (let i = 0; i < l; i++) {
      // measure children
      const child = children[i];
      if (child === main) continue;
      if (!child.visible && child !== locObj) continue;
      marg = child.margin as Margin;
      margw = marg.right + marg.left;
      margh = marg.top + marg.bottom;


      stretch = child.getEffectiveStretch(false);
      switch (stretch) {
        case GraphObject.Fill:
          width = mainWidth;
          height = mainHeight;
          break;
        case GraphObject.Horizontal:
          width = mainWidth;
          break;
        case GraphObject.Vertical:
          height = mainHeight;
          break;
      }

      child.measure(width, height, 0, 0); // the minimum is not related to the panel for children of spot panels
      m = child.measuredBounds;
      mwidth = Math.max(m.width + margw, 0);
      mheight = Math.max(m.height + margh, 0);

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;
      let alignfocus = child.alignmentFocus;
      if (alignfocus.isDefault()) alignfocus = Spot.Center;

      let alignmentChild: GraphObject | null = null;
      if (child instanceof Panel && child._alignmentFocusName !== '') {
        child.arrange(0, 0, m.width, m.height); // sadly needed for transforms
        alignmentChild = child.findObject(child._alignmentFocusName);
        if (alignmentChild === child) alignmentChild = null;
      }

      let arrangedX;
      let arrangedY;
      if (alignmentChild !== null) {
        const m2 = alignmentChild.naturalBounds;
        const marg2 = alignmentChild.margin as Margin;
        const pt = Point.allocAt(
          (alignfocus.x * m2.width - alignfocus.offsetX) - marg2.left,
          (alignfocus.y * m2.height - alignfocus.offsetY) - marg2.top);

        // walk up the panel chain
        while (alignmentChild !== child) {
          alignmentChild.transform.transformPoint(pt);
          alignmentChild = alignmentChild.panel;
        }

        arrangedX = align.x * mainWidth + align.offsetX - pt.x;
        arrangedY = align.y * mainHeight + align.offsetY - pt.y;
        Point.free(pt);
      } else {
        arrangedX = align.x * mainWidth + align.offsetX - (alignfocus.x * m.width + alignfocus.offsetX) - marg.left;
        arrangedY = align.y * mainHeight + align.offsetY - (alignfocus.y * m.height + alignfocus.offsetY) - marg.top;
      }

      if (first) {
        first = false;
        union.assignTo(arrangedX, arrangedY, mwidth, mheight);
      } else {
        union.unionNoCheck(arrangedX, arrangedY, mwidth, mheight);
      }

    }

    if (first) { // no elements, just use the main object's bounds
      union.assign(mainRect);
    } else if (isClip) {
      union.intersect(mainRect.x, mainRect.y, mainRect.width, mainRect.height);
    } else {
      union.unionNoCheck(mainRect.x, mainRect.y, mainRect.width, mainRect.height);
    }

    Rect.free(mainRect);

    // We might need to resize the main object and redo the union if there's a stretch.
    // Otherwise we are done
    stretch = main.stretch;
    if (stretch === GraphObject.Default) stretch = main.getEffectiveStretch(false);
    switch (stretch) {
      case GraphObject.None:
        return; // we're done
      case GraphObject.Fill:
        if (!isFinite(width) && !isFinite(height)) return;
        break;
      case GraphObject.Horizontal:
        if (!isFinite(width)) return;
        break;
      case GraphObject.Vertical:
        if (!isFinite(height)) return;
        break;
    }

    m = main.measuredBounds;
    mainWidth = m.width;
    mainHeight = m.height;
    mwidth = Math.max(mainWidth + mainmargw, 0);
    mheight = Math.max(mainHeight + mainmargh, 0);
    marg = main.margin as Margin;

    // we want union.x to be the main object's top left corner, not where its margin is
    mainRect = Rect.allocAt(-marg.left, -marg.top, mwidth, mheight);

    //  Don't remeasure the children but do recreate the union
    for (let i = 0; i < l; i++) {
      // measure children
      const child = children[i];
      if (child === main) continue;
      if (!child.visible && child !== locObj) continue;
      marg = child.margin as Margin;
      margw = marg.right + marg.left;
      margh = marg.top + marg.bottom;
      m = child.measuredBounds;
      mwidth = Math.max(m.width + margw, 0);
      mheight = Math.max(m.height + margh, 0);

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;
      let alignfocus = child.alignmentFocus;
      if (alignfocus.isDefault()) alignfocus = Spot.Center;

      if (first) {
        first = false;
        union.assignTo(
          align.x * mainWidth + align.offsetX - (alignfocus.x * m.width + alignfocus.offsetX) - marg.left,
          align.y * mainHeight + align.offsetY - (alignfocus.y * m.height + alignfocus.offsetY) - marg.top,
          mwidth,
          mheight);
      } else {
        union.unionNoCheck(align.x * mainWidth + align.offsetX - (alignfocus.x * m.width + alignfocus.offsetX) - marg.left,
          align.y * mainHeight + align.offsetY - (alignfocus.y * m.height + alignfocus.offsetY) - marg.top,
          mwidth, mheight);
      }
    }

    if (first) { // no elements, just use the main object's bounds
      union.assign(mainRect);
    } else if (isClip) {
      union.intersect(mainRect.x, mainRect.y, mainRect.width, mainRect.height);
    } else {
      union.unionNoCheck(mainRect.x, mainRect.y, mainRect.width, mainRect.height);
    }
    Rect.free(mainRect);
  }

  /**
   * @hidden @internal
   * Arrange children in spot panel
   * @this {PanelLayoutSpot}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const main = panel.findMainElement();
    const size = main.measuredBounds;
    const mainWidth = size.width;
    const mainHeight = size.height;
    const padding = panel.padding as Margin;
    const pl = padding.left;
    const pt = padding.top;

    let arx = pl - union.x;
    let ary = pt - union.y;
    main.arrange(arx, ary, mainWidth, mainHeight);
    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (child === main) continue;
      const m = child.measuredBounds;
      const width = m.width;
      const height = m.height;

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;
      let alignFocus = child.alignmentFocus;
      if (alignFocus.isDefault()) alignFocus = Spot.Center;

      let alignmentChild = null;
      if (child instanceof Panel && child._alignmentFocusName !== '') {
        alignmentChild = child.findObject(child._alignmentFocusName);
        if (alignmentChild === child) alignmentChild = null;
      }

      if (alignmentChild !== null) {
        const m2 = alignmentChild.naturalBounds;
        const tpoint = Point.allocAt(
          (alignFocus.x * m2.width - alignFocus.offsetX),
          (alignFocus.y * m2.height - alignFocus.offsetY));

        // walk up the panel chain
        while (alignmentChild !== child) {
          alignmentChild.transform.transformPoint(tpoint);
          alignmentChild = alignmentChild.panel;
        }

        arx = align.x * mainWidth + align.offsetX - tpoint.x;
        ary = align.y * mainHeight + align.offsetY - tpoint.y;
        Point.free(tpoint);
      } else {
        arx = align.x * mainWidth + align.offsetX - (alignFocus.x * width + alignFocus.offsetX);
        ary = align.y * mainHeight + align.offsetY - (alignFocus.y * height + alignFocus.offsetY);
      }
      arx -= union.x;
      ary -= union.y;

      if (child.visible) {
        child.arrange(pl + arx, pt + ary, width, height);
      }
    }
  }

}



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutAuto
*/

// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';


/**
 * @hidden @internal
 * PanelLayoutAuto
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutAuto extends PanelLayout {
  constructor() {
    super();
    this.name = 'Auto';
  }

  /**
   * @hidden @internal
   * The main object is sized around all the other ones
   * @this {PanelLayoutAuto}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    const main = panel.findMainElement();

    // Main is only correctly measured after the others have been
    // but it must be dummy-measured first to get a basic size for spot1 and 2
    let marg = main.margin as Margin;
    let cw = width;
    let ch = height;
    const mainmargw = marg.right + marg.left;
    const mainmargh = marg.top + marg.bottom;
    main.measure(width, height, minw, minh);
    let m = main.measuredBounds;
    let sw = 0.0;
    let mainShape = null;
    if (main instanceof Shape) {
      mainShape = main;
      sw = mainShape.strokeWidth * mainShape.scale;
    }
    let mwidth = Math.max(m.width + mainmargw, 0);
    let mheight = Math.max(m.height + mainmargh, 0);
    // use the inner size for relevant width/height
    let spot1 = panel.computeSpot1(main);
    const p1x = spot1.x * mwidth + spot1.offsetX;
    const p1y = spot1.y * mheight + spot1.offsetY;
    let spot2 = panel.computeSpot2(main);
    const p2x = spot2.x * mwidth + spot2.offsetX;
    const p2y = spot2.y * mheight + spot2.offsetY;
    if (isFinite(width)) cw = Math.max(Math.abs(p1x - p2x) - sw, 0);
    if (isFinite(height)) ch = Math.max(Math.abs(p1y - p2y) - sw, 0);

    // measure the children
    const totalsize = Size.alloc();
    totalsize.assignTo(0, 0);
    const locObj = panel.getLocationObject();
    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (child === main) { continue; }
      if (!child.visible && child !== locObj) continue;
      marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(cw, ch, 0, 0);
      m = child.measuredBounds;
      mwidth = Math.max(m.width + margw, 0);
      mheight = Math.max(m.height + margh, 0);

      // Keep track of the size we need to display all children
      totalsize.assignTo(Math.max(mwidth, totalsize.width),
        Math.max(mheight, totalsize.height));
    }

    if (l === 1) {
      m = main.measuredBounds;
      union.width = mwidth;
      union.height = mheight;
      Size.free(totalsize);
      return;
    }

    spot1 = panel.computeSpot1(main);
    spot2 = panel.computeSpot2(main);
    let mainWidth = 0.0;
    let mainHeight = 0.0;
    if (spot2.x !== spot1.x && spot2.y !== spot1.y) {
      mainWidth = totalsize.width / Math.abs(spot2.x - spot1.x);
      mainHeight = totalsize.height / Math.abs(spot2.y - spot1.y);
    }
    Size.free(totalsize);
    sw = 0.0;
    if (mainShape !== null) {
      sw = mainShape.strokeWidth * mainShape.scale;
      if (mainShape.getGeometryStretch() === GraphObject.Uniform) {
        mainWidth = mainHeight = Math.max(mainWidth, mainHeight);
      }
    }
    mainWidth += Math.abs(spot1.offsetX) + Math.abs(spot2.offsetX) + sw;
    mainHeight += Math.abs(spot1.offsetY) + Math.abs(spot2.offsetY) + sw;

    let stretch = main.stretch;
    if (stretch === GraphObject.Default) stretch = main.getEffectiveStretch(false);
    switch (stretch) {
      case GraphObject.None: // main element will fill its contents
        minw = 0;
        minh = 0;
        break;
      case GraphObject.Fill: // main element will fill the panel
        if (isFinite(width)) mainWidth = width;
        if (isFinite(height)) mainHeight = height;
        break;
      case GraphObject.Horizontal:
        if (isFinite(width)) mainWidth = width;
        minh = 0;
        break;
      case GraphObject.Vertical:
        minw = 0;
        if (isFinite(height)) mainHeight = height;
        break;
    }

    main.invalidateMeasureChildren();
    main.measure(mainWidth, mainHeight, minw, minh);
    // The union width and height merely max the size of the main object and the size of all other objects.
    union.width = main.measuredBounds.width + mainmargw;
    union.height = main.measuredBounds.height + mainmargh;
  }

  /**
   * @hidden @internal
   * Arrange children in Auto panel
   * @this {PanelLayoutAuto}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const main = panel.findMainElement();
    const mainSize = main.measuredBounds;

    // outer is 0, 0, mainSize
    const inner = Rect.alloc();
    inner.assignTo(0, 0, 1, 1);

    let marg = main.margin as Margin;
    const mainOffsetX = marg.left;
    const mainOffsetY = marg.top;

    const padding = panel.padding as Margin;
    const pl = padding.left;
    const pt = padding.top;
    let arx = mainOffsetX;
    let ary = mainOffsetY;
    let arwidth = mainSize.width;
    let arheight = mainSize.height;
    main.arrange(pl + arx, pt + ary, arwidth, arheight);
    const spot1 = panel.computeSpot1(main);
    const spot2 = panel.computeSpot2(main);
    // setting rect spot inline
    const p1x = 0 + spot1.x * mainSize.width + spot1.offsetX;
    const p1y = 0 + spot1.y * mainSize.height + spot1.offsetY;
    const p2x = 0 + spot2.x * mainSize.width + spot2.offsetX;
    const p2y = 0 + spot2.y * mainSize.height + spot2.offsetY;
    inner.x = p1x;
    inner.y = p1y;
    inner.unionNoCheck(p2x, p2y, 0, 0);
    inner.x += mainOffsetX + pl;
    inner.y += mainOffsetY + pt;

    for (let i = 0; i < l; i++) {
      const child = children[i];
      if (child === main) continue;

      const m = child.measuredBounds;
      marg = child.margin as Margin;
      const mwidth = Math.max(m.width + marg.right + marg.left, 0);
      const mheight = Math.max(m.height + marg.top + marg.bottom, 0);

      let align = child.alignment;
      if (align.isDefault()) align = panel.defaultAlignment;
      if (!align.isSpot()) align = Spot.Center;

      arx = inner.width * align.x + align.offsetX - (mwidth * align.x) + marg.left + inner.x;
      ary = inner.height * align.y + align.offsetY - (mheight * align.y) + marg.top + inner.y;
      arwidth = inner.width;
      arheight = inner.height;

      if (child.visible) {
        if (Rect.contains(inner.x, inner.y, inner.width, inner.height, arx, ary, m.width, m.height)) {
          child.arrange(arx, ary, m.width, m.height);
        } else {
          child.arrange(arx, ary, m.width, m.height, new Rect(inner.x, inner.y, inner.width, inner.height));
        }
      }
    }
    Rect.free(inner);
  }

}



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutTable
*/

// import { Debug } from './debug';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Rect } from './rect';
// import { RowColumnDefinition } from './rowColumnDefinition';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Util } from './util';


/**
 * @hidden @internal
 * PanelLayoutTable
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutTable extends PanelLayout {
  constructor() {
    super();
    this.name = 'Table';
  }

    /* //???
    There's a bug with row/column measurements under these conditions:
    * Textblock in cell with Fill or Horizontal stretch
    * Column has objects of a small "hard" width, which suggest a small width to fill into
    * Textblock measures with lots of wrapping, making the cell/row height tall
    * Later, proportional extra space allows column to be wider
    * Textblock remeasures so that the row COULD be less tall, but its already too late

    The solution is probably to re-order measureTable so that
    allocating proportional extra comes before measuring the stretch objects, instead of last.
    That solution may cause panels to measure differently.
  */

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTable}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    let l = children.length;
    // "flatten" out the TableRow and TableColumns
    const flatChildren = Util.tempArray<GraphObject>();
    const tableRowsCols = Util.tempArray<Panel>();
    for (let i = 0; i < l; i++) {
      const child = children[i];
      const cpanel = child instanceof Panel ? child : null;
      if (cpanel !== null && (cpanel.type === Panel.TableRow || cpanel.type === Panel.TableColumn) && child.visible) {
        if (Debug) {
          if (cpanel.desiredSize.isReal()) {
            Util.throwError(cpanel.toString() + ' TableRow/TableColumn Panels cannot set a desiredSize: ' + cpanel.desiredSize.toString());
          }
          if (!cpanel.minSize.equals(Size.ZeroSize)) {
            Util.throwError(cpanel.toString() + ' TableRow/TableColumn Panels cannot set a minSize: ' + cpanel.minSize.toString());
          }
          if (!cpanel.maxSize.equals(Size.InfiniteSize)) {
            Util.throwError(cpanel.toString() + ' TableRow/TableColumn Panels cannot set a maxSize: ' + cpanel.maxSize.toString());
          }
        }
        tableRowsCols.push(cpanel);
        const childlist = cpanel._elements._arr;
        const ll = childlist.length;
        for (let j = 0; j < ll; j++) {
          const flatchild = childlist[j];
          if (cpanel.type === Panel.TableRow) {
            flatchild.row = child.row;
          } else if (cpanel.type === Panel.TableColumn) {
            flatchild.column = child.column;
          }
          flatChildren.push(flatchild);
        }
      } else {
        flatChildren.push(child);
      }
    }
    l = flatChildren.length;

    if (l === 0) {
      panel.getRowDefinition(0);
      panel.getColumnDefinition(0);
    }

    // Make the array that holds [rows][cols] of the table
    // saved (so no temp array) starts as an array of rows, will end up [row][col][cell]
    const rowcol = [] as Array<Array<Array<GraphObject>>>;
    for (let i = 0; i < l; i++) {
      const child = flatChildren[i];
      // ?? this could be optimized. Not all children have to be invalidated
      child.setInvalidMeasure(true);
      child.setInvalidArrange(true);

      if (!rowcol[child.row]) {
        // make new column for this row
        rowcol[child.row] = [];
      }
      if (!rowcol[child.row][child.column]) {
        // new list for this cell
        rowcol[child.row][child.column] = [];
      }
      rowcol[child.row][child.column].push(child); // push child into right cell
    }
    Util.freeArray(flatChildren);

    // Reset the row/col definitions because the ones from last measure are irrelevant
    const resetCols = Util.tempArray<boolean>(); // keep track of which columns we've already reset

    // Objects that span multiple columns and
    const spanners = Util.tempArray<GraphObject>();
    const nosize = Util.tempArray<GraphObject>();
    // These hashes are used to tally the number of rows and columns that do not have a size
    const nosizeCols: { [index: string]: number } = { 'count': 0 };

    const nosizeRows: { [index: string]: number } = { 'count': 0 };

    let colleft = width;
    let rowleft = height;

    let defs = panel._rowDefs;
    let def: RowColumnDefinition;
    l = defs.length;
    for (let i = 0; i < l; i++) {
      def = defs[i];
      if (def !== undefined) def.actual = 0;
    }

    defs = panel._colDefs;
    l = defs.length;
    for (let i = 0; i < l; i++) {
      def = defs[i];
      if (def !== undefined) def.actual = 0;
    }

    let lrow = rowcol.length; // number of rows
    let lcol = 0;
    for (let i = 0; i < lrow; i++) {
      if (!rowcol[i]) continue;
      lcol = Math.max(lcol, rowcol[i].length); // column length in this row
    }
    const idxrow = Math.min(panel.topIndex, lrow - 1);
    const idxcol = Math.min(panel.leftIndex, lcol - 1);

    // Go through each cell (first pass)
    let amt = 0.0;
    let stretch: EnumValue;
    lrow = rowcol.length; // number of rows
    const locObj = panel.getLocationObject();
    for (let i = 0; i < lrow; i++) {
      if (!rowcol[i]) continue;
      lcol = rowcol[i].length; // column length in this row
      const rowHerald = panel.getRowDefinition(i);
      rowHerald.actual = 0; // Reset rows (only on first pass)
      for (let j = 0; j < lcol; j++) {
        // foreach column j in row i...
        if (!rowcol[i][j]) continue;
        const colHerald = panel.getColumnDefinition(j);
        if (resetCols[j] === undefined) { // make sure we only reset these once
          colHerald.actual = 0;
          resetCols[j] = true;
        }

        const cell = rowcol[i][j];
        const len = cell.length;
        for (let k = 0; k < len; k++) {
          // foreach element in cell, measure
          const child = cell[k];
          if (!child.visible && child !== locObj) continue;

          // Skip children that span more than one row or column or do not have a set size
          const spanner = (child.rowSpan > 1 || child.columnSpan > 1);
          if (spanner) {
            if (!(i < idxrow) && !(j < idxcol)) spanners.push(child);
            // We used to not measure spanners twice, but now we do
            // The reason is that there may be a row whose size
            // is dictated by an object with columnSpan 2+ and vice versa

            // continue;
          }

          const marg = child.margin as Margin;
          const margw = marg.right + marg.left;
          const margh = marg.top + marg.bottom;

          stretch = child.getEffectiveTableStretch(rowHerald, colHerald, false);
          const dsize = child.desiredSize; // child desired size!
          const realwidth = !(isNaN(dsize.width));
          const realheight = !(isNaN(dsize.height));
          const realsize = realwidth && realheight;
          if (!spanner && stretch !== GraphObject.None && !realsize && (!(i < idxrow) && !(j < idxcol))) {

            if (nosizeCols[j] === undefined && (stretch === GraphObject.Fill || stretch === GraphObject.Horizontal)) {
              nosizeCols[j] = -1; nosizeCols.count++;
            }
            if (nosizeRows[i] === undefined && (stretch === GraphObject.Fill || stretch === GraphObject.Vertical)) {
              nosizeRows[i] = -1; nosizeRows.count++;
            }
            nosize.push(child);
            // we want the natural size of stretched objects
            // ?? Maybe there is a good way around doing this? It means we measure these objects twice
          }
          child.measure(Infinity, Infinity, 0, 0);

          // Objects preceding the topIndex and leftIndex should be measured, but considered no further
          if (i < idxrow) continue;
          if (j < idxcol) continue;

          const m = child.measuredBounds;
          const mwidth = Math.max(m.width + margw, 0);
          const mheight = Math.max(m.height + margh, 0);

          //  Make sure the heralds have the right layout size
          //    the row/column should use the largest meausured size of any
          //    GraphObject contained, constrained by mins and maxes
          if (child.rowSpan === 1 && (stretch === GraphObject.None || stretch === GraphObject.Horizontal)) {
            def = panel.getRowDefinition(i);
            const spacing = def.computeEffectiveSpacing();
            amt = Math.max(mheight - def.actual, 0);
            if (amt + spacing > rowleft) amt = Math.max(rowleft - spacing, 0);
            const includespacing = def.actual === 0;
            def.actual = def.actual + amt;
            rowleft = Math.max(rowleft - (amt + (includespacing ? spacing : 0)), 0);
          }

          if (child.columnSpan === 1 && (stretch === GraphObject.None || stretch === GraphObject.Vertical)) {
            def = panel.getColumnDefinition(j);
            const spacing = def.computeEffectiveSpacing();
            amt = Math.max(mwidth - def.actual, 0);
            if (amt + spacing > colleft) amt = Math.max(colleft - spacing, 0);
            const includespacing = def.actual === 0;
            def.actual = def.actual + amt;
            colleft = Math.max(colleft - (amt + (includespacing ? spacing : 0)), 0);
          }

          if (spanner) {
            child.invalidateMeasureChildren();
          }

        } // end cell
      } // end col
    } // end row

    Util.freeArray(resetCols);

    // For objects of no desired size we allocate what is left as we go,
    // or else what is already in the column
    let totalColWidth = 0.0;
    let totalRowHeight = 0.0;
    l = panel.columnCount;
    for (let i = 0; i < l; i++) {
      const coldef = panel._colDefs[i];
      if (coldef === undefined) continue;
      totalColWidth += coldef._actual;
      if (coldef._actual !== 0) totalColWidth += coldef.computeEffectiveSpacing();
    }
    l = panel.rowCount;
    for (let i = 0; i < l; i++) {
      const rowdef = panel._rowDefs[i];
      if (rowdef === undefined) continue;
      totalRowHeight += rowdef._actual;
      if (rowdef._actual !== 0) totalRowHeight += rowdef.computeEffectiveSpacing();
    }
    colleft = Math.max(width - totalColWidth, 0);
    rowleft = Math.max(height - totalRowHeight, 0);
    const originalrowleft = rowleft;
    const originalcolleft = colleft;

    // Determine column sizes for the yet-to-be-sized columns
    l = nosize.length;
    for (let i = 0; i < l; i++) {
      const child = nosize[i];
      const rowHerald = panel.getRowDefinition(child.row);
      const colHerald = panel.getColumnDefinition(child.column);
      // We want to gather the largest difference between desired and expected col/row sizes
      const mb = child.measuredBounds;
      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;

      if (colHerald.actual === 0 && nosizeCols[child.column] !== undefined) {
        nosizeCols[child.column] = Math.max(mb.width + margw, nosizeCols[child.column]);
      } else {
        nosizeCols[child.column] = null; // obey the column herald
      }
      if (rowHerald.actual === 0 && nosizeRows[child.row] !== undefined) {
        nosizeRows[child.row] = Math.max(mb.height + margh, nosizeRows[child.row]);
      } else {
        nosizeRows[child.row] = null; // obey the row herald
      }

    }
    // we now have the size that all these columns prefer to be
    // we also have the amount left over
    let desiredRowTotal = 0.0;
    let desiredColTotal = 0.0;
    let ii;
    for (ii in nosizeRows) { if (ii !== 'count') desiredRowTotal += nosizeRows[ii]; }
    for (ii in nosizeCols) { if (ii !== 'count') desiredColTotal += nosizeCols[ii]; }

    const allowedSize = Size.alloc(); // used in stretch and span loops

    // Deal with objects that have a stretch
    for (let i = 0; i < l; i++) {
      const child = nosize[i];
      if (!child.visible && child !== locObj) continue;
      const rowHerald = panel.getRowDefinition(child.row);
      const colHerald = panel.getColumnDefinition(child.column);

      let w = 0.0;
      if (isFinite(colHerald.width)) {
        w = colHerald.width;
      } else {
        if (isFinite(colleft) && nosizeCols[child.column] !== null) {
          if (desiredColTotal === 0) w = colHerald.actual + colleft;
          else w = /*colHerald.actual +*/ ((nosizeCols[child.column] / desiredColTotal) * originalcolleft);
        } else {
          // Only use colHerald.actual if it was nonzero before this loop
          if (nosizeCols[child.column] !== null) w = colleft;
          else w = colHerald.actual || colleft;
        }
        w = Math.max(0, w - colHerald.computeEffectiveSpacing());
      }
      let h = 0.0;
      if (isFinite(rowHerald.height)) {
        h = rowHerald.height;
      } else {
        if (isFinite(rowleft) && nosizeRows[child.row] !== null) {
          if (desiredRowTotal === 0) h = rowHerald.actual + rowleft;
          else h = /*rowHerald.actual +*/ ((nosizeRows[child.row] / desiredRowTotal) * originalrowleft);
        } else {
          // Only use rowHerald.actual if it was nonzero before this loop
          if (nosizeRows[child.row] !== null) h = rowleft;
          else h = rowHerald.actual || rowleft;
        }
        h = Math.max(0, h - rowHerald.computeEffectiveSpacing());
      }

      allowedSize.assignTo(
        Math.max(colHerald.minimum, Math.min(w, colHerald.maximum)),
        Math.max(rowHerald.minimum, Math.min(h, rowHerald.maximum)));

      // Which way do we care about fill:
      stretch = child.getEffectiveTableStretch(rowHerald, colHerald, false);
      // This used to set allowedSize height/width to Infinity,
      // but we can only set it to the current row/column space, plus rowleft/colleft values, at most.
      switch (stretch) {
        case GraphObject.Horizontal: // H stretch means it can be as large as its wants vertically
          allowedSize.height = Math.max(allowedSize.height, rowHerald.actual + rowleft);
          break;
        case GraphObject.Vertical: // vice versa
          allowedSize.width = Math.max(allowedSize.width, colHerald.actual + colleft);
          break;
      }

      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      // ??? here I should be optimizing, if the allowed size is equal to the child's already measured size then we can do nothing
      child.invalidateMeasureChildren();
      child.measure(allowedSize.width, allowedSize.height, colHerald.minimum, rowHerald.minimum);
      const m = child.measuredBounds;
      let mwidth = Math.max(m.width + margw, 0);
      let mheight = Math.max(m.height + margh, 0);
      if (isFinite(colleft)) mwidth = Math.min(mwidth, allowedSize.width);
      if (isFinite(rowleft)) mheight = Math.min(mheight, allowedSize.height);

      let oldAmount = 0.0;

      oldAmount = rowHerald.actual;
      rowHerald.actual = Math.max(rowHerald.actual, mheight);
      amt = rowHerald.actual - oldAmount;
      rowleft = Math.max(rowleft - amt, 0);

      oldAmount = colHerald.actual;
      colHerald.actual = Math.max(colHerald.actual, mwidth);
      amt = colHerald.actual - oldAmount;
      colleft = Math.max(colleft - amt, 0);
    } // end no fixed size objects
    Util.freeArray(nosize);

    // Go through each object that spans multiple rows or columns
    const additionalSpan = Size.alloc();
    const actualSizeRows = Util.tempArray<number>();
    const actualSizeColumns = Util.tempArray<number>();
    l = spanners.length;
    if (l !== 0) {
      // record the actual sizes of every row/column before measuring spanners
      // because they will change during the loop and we want to use their 'before' values
      for (let i = 0; i < lrow; i++) {
        if (!rowcol[i]) continue;
        lcol = rowcol[i].length; // column length in this row
        const rowHerald = panel.getRowDefinition(i);
        actualSizeRows[i] = rowHerald.actual;
        for (let j = 0; j < lcol; j++) {
          // foreach column j in row i...
          if (!rowcol[i][j]) continue;
          const colHerald = panel.getColumnDefinition(j);
          actualSizeColumns[j] = colHerald.actual;
        }
      }
    }
    for (let i = 0; i < l; i++) {
      const child = spanners[i];
      if (!child.visible && child !== locObj) continue;
      const rowHerald = panel.getRowDefinition(child.row);
      const colHerald = panel.getColumnDefinition(child.column);

      // If there's a set column width/height we don't care about the given width/height
      allowedSize.assignTo(
        Math.max(colHerald.minimum, Math.min(width, colHerald.maximum)),
        Math.max(rowHerald.minimum, Math.min(height, rowHerald.maximum)));

      // If it is a spanner and has a fill:
      stretch = child.getEffectiveTableStretch(rowHerald, colHerald, false);
      switch (stretch) {
        case GraphObject.Fill:
          if (actualSizeColumns[colHerald.index] !== 0) allowedSize.width = Math.min(allowedSize.width, actualSizeColumns[colHerald.index]);
          if (actualSizeRows[rowHerald.index] !== 0) allowedSize.height = Math.min(allowedSize.height, actualSizeRows[rowHerald.index]);
          break;
        case GraphObject.Horizontal:
          if (actualSizeColumns[colHerald.index] !== 0) allowedSize.width = Math.min(allowedSize.width, actualSizeColumns[colHerald.index]);
          break;
        case GraphObject.Vertical:
          if (actualSizeRows[rowHerald.index] !== 0) allowedSize.height = Math.min(allowedSize.height, actualSizeRows[rowHerald.index]);
          break;
      }
      // If there's a set column width/height we don't care about any of the above:
      if (isFinite(colHerald.width)) allowedSize.width = colHerald.width;
      if (isFinite(rowHerald.height)) allowedSize.height = rowHerald.height;

      // take into account rowSpan and columnSpan
      additionalSpan.assignTo(0, 0);
      for (let n = 1; n < child.rowSpan; n++) {
        if (child.row + n >= panel.rowCount) break; // if the row exists at all
        def = panel.getRowDefinition(child.row + n);
        amt = 0;
        if (stretch === GraphObject.Fill || stretch === GraphObject.Vertical) {
          amt = Math.max(def.minimum, (actualSizeRows[child.row + n] === 0) ? def.maximum : Math.min(actualSizeRows[child.row + n], def.maximum));
        } else {
          amt = Math.max(def.minimum, isNaN(def._size) ? def.maximum : Math.min(def._size, def.maximum));
        }
        additionalSpan.height += amt;
      }
      for (let n = 1; n < child.columnSpan; n++) {
        if (child.column + n >= panel.columnCount) break; // if the col exists at all
        def = panel.getColumnDefinition(child.column + n);
        amt = 0;
        if (stretch === GraphObject.Fill || stretch === GraphObject.Horizontal) {
          amt = Math.max(def.minimum, (actualSizeColumns[child.column + n] === 0) ? def.maximum : Math.min(actualSizeColumns[child.column + n], def.maximum));
        } else {
          amt = Math.max(def.minimum, isNaN(def._size) ? def.maximum : Math.min(def._size, def.maximum));
        }
        additionalSpan.width += amt;
      }
      allowedSize.width += additionalSpan.width;
      allowedSize.height += additionalSpan.height;

      const marg = child.margin as Margin;
      const margw = marg.right + marg.left;
      const margh = marg.top + marg.bottom;
      child.measure(allowedSize.width, allowedSize.height, minw, minh);
      const m = child.measuredBounds;
      const mwidth = Math.max(m.width + margw, 0);
      const mheight = Math.max(m.height + margh, 0);

      let totalRow = 0.0;
      for (let n = 0; n < child.rowSpan; n++) {
        if (child.row + n >= panel.rowCount) break; // if the row exists at all
        def = panel.getRowDefinition(child.row + n);
        totalRow += def.total || 0;
      }
      // def is the last row definition
      if (totalRow < mheight) {
        let roomLeft = mheight - totalRow;
        const roomLeftTotal = mheight - totalRow;

        // Pass each row to spanAllocation
        if (child.spanAllocation !== null) {
          const rules = child.spanAllocation;
          for (let n = 0; n < child.rowSpan; n++) {
            if (roomLeft <= 0) break;
            if (child.row + n >= panel.rowCount) break; // if the row exists at all
            def = panel.getRowDefinition(child.row + n);
            const act = def._actual || 0;
            const roomGiven = rules(child, def, roomLeftTotal);
            if (Debug && typeof roomGiven !== 'number') Util.throwError(child + ' spanAllocation does not return a number: ' + roomGiven);
            def.actual = Math.min(def.maximum, act + roomGiven);
            if (def._actual !== act) roomLeft -= def._actual - act;
          }
        }

        // Working backwards, add the extra to the first row that allows us to:
        while (roomLeft > 0) {
          const act = def._actual || 0;
          if (isNaN(def.height) && def.maximum > act) {
            def.actual = Math.min(def.maximum, act + roomLeft);
            if (def._actual !== act) roomLeft -= def._actual - act;
          }
          if (def.index === 0) break;
          def = panel.getRowDefinition(def.index - 1);
        }
      }

      let totalCol = 0.0;
      for (let n = 0; n < child.columnSpan; n++) {
        if (child.column + n >= panel.columnCount) break; // if the col exists at all
        def = panel.getColumnDefinition(child.column + n);
        totalCol += def.total || 0;
      }
      // def is the last col definition
      if (totalCol < mwidth) {
        let roomLeft = mwidth - totalCol;
        const roomLeftTotal = mwidth - totalCol;

        // Pass each column to spanAllocation
        if (child.spanAllocation !== null) {
          const rules = child.spanAllocation;
          for (let n = 0; n < child.columnSpan; n++) {
            if (roomLeft <= 0) break;
            if (child.column + n >= panel.columnCount) break; // if the col exists at all
            def = panel.getColumnDefinition(child.column + n);
            const act = def._actual || 0;
            const roomGiven = rules(child, def, roomLeftTotal);
            if (Debug && typeof roomGiven !== 'number') Util.throwError(child + ' spanAllocation does not return a number: ' + roomGiven);
            def.actual = Math.min(def.maximum, act + roomGiven);
            if (def._actual !== act) roomLeft -= def._actual - act;
          }
        }

        // Working backwards, add the extra to the first column that allows us to:
        while (roomLeft > 0) {
          const act = def._actual || 0;
          if (isNaN(def.width) && def.maximum > act) {
            def.actual = Math.min(def.maximum, act + roomLeft);
            if (def._actual !== act) roomLeft -= def._actual - act;
          }
          if (def.index === 0) break;
          def = panel.getColumnDefinition(def.index - 1);
        }
      }
    } // end spanning objects
    Util.freeArray(spanners);
    Size.free(additionalSpan);
    Size.free(allowedSize);
    if (actualSizeRows !== undefined) Util.freeArray(actualSizeRows);
    if (actualSizeColumns !== undefined) Util.freeArray(actualSizeColumns);

    // the rest of this propotionally resizes rows and columns, but we only ever size them up

    let functionalWidth = 0.0;
    let functionalHeight = 0.0;
    const des = panel.desiredSize;
    const max = panel.maxSize;
    stretch = panel.getEffectiveStretch(true);
    totalColWidth = 0;
    totalRowHeight = 0;

    // we cannot resize columns and rows with set width/height so we must exclude them
    let rigidwidth = 0.0;
    let rigidheight = 0.0;

    l = panel.columnCount;
    for (let i = 0; i < l; i++) {
      if (panel._colDefs[i] === undefined) continue;
      def = panel.getColumnDefinition(i);
      if (isFinite(def.width)) {
        rigidwidth += def.width;
        rigidwidth += def.computeEffectiveSpacing();
        continue;
      } else if (def.getEffectiveSizing() === RowColumnDefinition.None) {
        rigidwidth += def._actual;
        rigidwidth += def.computeEffectiveSpacing();
        continue;
      }
      if (def._actual !== 0) {
        totalColWidth += def._actual;
        totalColWidth += def.computeEffectiveSpacing();
      }
    }
    if (isFinite(des.width)) {
      functionalWidth = Math.min(des.width, max.width);
    } else {
      if (stretch !== GraphObject.None && isFinite(width)) functionalWidth = width;
      else functionalWidth = totalColWidth;
    }
    functionalWidth = Math.max(functionalWidth, panel.minSize.width);
    functionalWidth = Math.max(functionalWidth - rigidwidth, 0);
    const colProportion = totalColWidth === 0 ? 1 : Math.max(functionalWidth / totalColWidth, 1);

    for (let i = 0; i < l; i++) {
      if (panel._colDefs[i] === undefined) continue;
      def = panel.getColumnDefinition(i);
      if (!isFinite(def.width) && def.getEffectiveSizing() !== RowColumnDefinition.None) def.actual = def._actual * colProportion;
      def.position = union.width;
      if (def._actual !== 0) {
        union.width += def._actual;
        union.width += def.computeEffectiveSpacing();
      }
    }

    l = panel.rowCount;
    for (let i = 0; i < l; i++) {
      if (panel._rowDefs[i] === undefined) continue;
      def = panel.getRowDefinition(i);
      if (isFinite(def.height)) {
        rigidheight += def.height;
        rigidheight += def.computeEffectiveSpacing();
        continue;
      } else if (def.getEffectiveSizing() === RowColumnDefinition.None) {
        rigidheight += def._actual;
        rigidheight += def.computeEffectiveSpacing();
        continue;
      }
      if (def._actual !== 0) {
        totalRowHeight += def._actual;
        // no real amount, don't add spacing
        if (def._actual !== 0) totalRowHeight += def.computeEffectiveSpacing();
      }
    }
    if (isFinite(des.height)) {
      functionalHeight = Math.min(des.height, max.height);
    } else {
      if (stretch !== GraphObject.None && isFinite(height)) functionalHeight = height;
      else functionalHeight = totalRowHeight;
    }
    functionalHeight = Math.max(functionalHeight, panel.minSize.height);
    functionalHeight = Math.max(functionalHeight - rigidheight, 0);
    const rowProportion = totalRowHeight === 0 ? 1 : Math.max(functionalHeight / totalRowHeight, 1);

    for (let i = 0; i < l; i++) {
      if (panel._rowDefs[i] === undefined) continue;
      def = panel.getRowDefinition(i);
      if (!isFinite(def.height) && def.getEffectiveSizing() !== RowColumnDefinition.None) def.actual = def._actual * rowProportion;
      def.position = union.height;
      if (def._actual !== 0) {
        union.height += def._actual;
        // no real amount, don't add spacing
        if (def._actual !== 0) union.height += def.computeEffectiveSpacing();
      }
    }

    l = tableRowsCols.length;
    for (let i = 0; i < l; i++) {
      const trc = tableRowsCols[i];
      let w = 0.0;
      let h = 0.0;
      if (trc.type === Panel.TableRow) {
        w = union.width;
        def = panel.getRowDefinition(trc.row);
        h = def.actual;
      } else { // TableColumn
        def = panel.getColumnDefinition(trc.column);
        w = def.actual;
        h = union.height;
      }

      trc.measuredBounds.assignTo(0, 0, w, h);
      trc.setInvalidMeasure(false);

      if (!rowcol[trc.row]) {
        rowcol[trc.row] = []; // make new column for this row
      }
      if (!rowcol[trc.row][trc.column]) {
        rowcol[trc.row][trc.column] = []; // new list for this cell
      }
      rowcol[trc.row][trc.column].push(trc);
    }
    Util.freeArray(tableRowsCols);

    // save these for arrange (destroy them or not? Possibly needed for drawing spacers)
    panel._rowcol = rowcol;
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTable}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;
    const padding = panel.padding as Margin;
    const pl = padding.left;
    const pt = padding.top;

    const rowcol = panel._rowcol;
    let x = 0.0;
    let y = 0.0;

    const lrow = rowcol.length; // number of rows
    let lcol = 0;
    for (let i = 0; i < lrow; i++) {
      if (!rowcol[i]) continue;
      lcol = Math.max(lcol, rowcol[i].length); // column length in this row
    }

    // find the logical top index and left index
    let idxrow = Math.min(panel.topIndex, lrow - 1);
    while (idxrow !== lrow && (panel._rowDefs[idxrow] === undefined || panel._rowDefs[idxrow]._actual === 0)) idxrow++;
    idxrow = Math.min(idxrow, lrow - 1);
    const rowIndexOffset = -panel._rowDefs[idxrow].position;

    let idxcol = Math.min(panel.leftIndex, lcol - 1);
    while (idxcol !== lcol && (panel._colDefs[idxcol] === undefined || panel._colDefs[idxcol]._actual === 0)) idxcol++;
    idxcol = Math.min(idxcol, lcol - 1);
    const colIndexOffset = -panel._colDefs[idxcol].position;

    const additionalSpan = Size.alloc();
    // Find cell space and arrange objects:
    for (let i = 0; i < lrow; i++) {
      if (!rowcol[i]) continue;
      lcol = rowcol[i].length; // column length in this row
      const rowHerald = panel.getRowDefinition(i);
      y = rowHerald.position + rowIndexOffset + pt;
      if (rowHerald._actual !== 0) y += rowHerald.computeEffectiveSpacingTop();
      for (let j = 0; j < lcol; j++) {
        // foreach column j in row i...
        if (!rowcol[i][j]) continue;
        const colHerald = panel.getColumnDefinition(j);
        x = colHerald.position + colIndexOffset + pl;
        if (colHerald._actual !== 0) x += colHerald.computeEffectiveSpacingTop();
        const cell = rowcol[i][j];
        const len = cell.length;

        for (let k = 0; k < len; k++) {
          // foreach element in cell
          const child = cell[k];
          const m = child.measuredBounds;
          const cpanel = child instanceof Panel ? child : null;
          if (cpanel !== null && (cpanel.type === Panel.TableRow || cpanel.type === Panel.TableColumn)) {
            // TableRows/Columns bypass GraphObject.arrange, so they need to invalidate transforms here
            // and also do some onBoundsChanged work:
            child._invalidateTransform();
            child.actualBounds.thaw();
            const ab = child.actualBounds;
            const oldb = Rect.allocAt(ab.x, ab.y, ab.width, ab.height);
            ab.x = cpanel.type === Panel.TableRow ? pl : x;
            ab.y = cpanel.type === Panel.TableColumn ? pt : y;
            ab.width = m.width;
            ab.height = m.height;
            child.actualBounds.freeze();
            child.setInvalidArrange(false);

            if (oldb.equalsApproxClose(ab)) {
              Rect.free(oldb);
              continue;
            }
            const part = child.part;
            if (part !== null) {
              part.invalidateContainingGroup();
              child._notifyPorts(part);
            }
            Rect.free(oldb);
            continue;
          }

          // add to layoutWidth/Height any additional span
          additionalSpan.assignTo(0, 0);

          for (let n = 1; n < child.rowSpan; n++) {
            // if the row exists at all
            if (i + n >= panel.rowCount) break;
            const rh = panel.getRowDefinition(i + n);
            additionalSpan.height += rh.total;
          }

          for (let n = 1; n < child.columnSpan; n++) {
            // if the col exists at all
            if (j + n >= panel.columnCount) break;
            const ch = panel.getColumnDefinition(j + n);
            additionalSpan.width += ch.total;
          }

          // Construct containing rect (cell):

          // total width and height of the cell that an object could possibly be created in
          const colwidth = colHerald._actual + additionalSpan.width;
          const rowheight = rowHerald._actual + additionalSpan.height;

          // construct a rect that represents the total cell size allowed for this object
          let arx = x;
          let ary = y;
          const arwidth = colwidth;
          const arheight = rowheight;

          // Also keep them for clip values
          const cellx = x;
          const celly = y;
          let cellw = colwidth;
          let cellh = rowheight;
          // Ending rows/col might have actual spaces that are larger than the remaining space
          // Modify them for clipping regions
          if (x + colwidth > union.width) cellw = Math.max(union.width - x, 0);
          if (y + rowheight > union.height) cellh = Math.max(union.height - y, 0);

          // Construct alignment:
          let align = child.alignment;
          let alignx = 0.0;
          let aligny = 0.0;
          let alignoffsetX = 0.0;
          let alignoffsetY = 0.0;
          if (align.isDefault()) {
            align = panel.defaultAlignment;
            if (!align.isSpot()) align = Spot.Center;
            alignx = align.x;
            aligny = align.y;
            alignoffsetX = align.offsetX;
            alignoffsetY = align.offsetY;
            const ca = colHerald.alignment;
            const ra = rowHerald.alignment;
            if (ca.isSpot()) {
              alignx = ca.x;
              alignoffsetX = ca.offsetX;
            }
            if (ra.isSpot()) {
              aligny = ra.y;
              alignoffsetY = ra.offsetY;
            }
          } else {
            alignx = align.x;
            aligny = align.y;
            alignoffsetX = align.offsetX;
            alignoffsetY = align.offsetY;
          }

          // same as if (!align.isSpot()) align = Spot.Center;
          if (isNaN(alignx) || isNaN(aligny)) {
            alignx = 0.5;
            aligny = 0.5;
            alignoffsetX = 0;
            alignoffsetY = 0;
          }

          let width = m.width;
          let height = m.height;

          const marg = child.margin as Margin;
          const margw = marg.left + marg.right;
          const margh = marg.top + marg.bottom;
          const stretch = child.getEffectiveTableStretch(rowHerald, colHerald, false);
          // child desired size!
          if (isNaN(child.desiredSize.width) && (stretch === GraphObject.Fill || stretch === GraphObject.Horizontal)) {
            width = Math.max(colwidth - margw, 0);
          }
          if (isNaN(child.desiredSize.height) && (stretch === GraphObject.Fill || stretch === GraphObject.Vertical)) {
            height = Math.max(rowheight - margh, 0);
          }

          // min and max override any stretch values
          const max = child.maxSize;
          const min = child.minSize;
          width = Math.min(max.width, width);
          height = Math.min(max.height, height);
          width = Math.max(min.width, width);
          height = Math.max(min.height, height);

          const widthmarg = width + margw;
          const heightmarg = height + margh;

          arx += (arwidth * alignx) - (widthmarg * alignx) + alignoffsetX + marg.left;
          ary += (arheight * aligny) - (heightmarg * aligny) + alignoffsetY + marg.top;

          if (child.visible) {
            // m.width and height because we must test against its true, not expected, size
            if (Rect.contains(cellx, celly, cellw, cellh, arx, ary, m.width, m.height)) {
              child.arrange(arx, ary, width, height);
            } else {
              child.arrange(arx, ary, width, height, new Rect(cellx, celly, cellw, cellh));
            }
          }
        } // end cell
      } // end col
    } // end row
    Size.free(additionalSpan);

    // Finally, duplicate actualBounds/measuredBounds/allTransforms onto any TableRows any TableColumns
    for (let i = 0; i < l; i++) {
      const child = children[i];
      const cpanel = child instanceof Panel ? child : null;
      if (cpanel !== null && (cpanel.type === Panel.TableRow || cpanel.type === Panel.TableColumn)) {
        const ab = child.actualBounds;
        child.naturalBounds.thaw();
        child.naturalBounds.assignTo(0, 0, ab.width, ab.height);
        child.naturalBounds.freeze();
      }
    }
    // panel._rowcol is kept in case arrange is called without measure
  }

}


/**
 * @hidden @internal
 * PanelLayoutTableRow
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutTableRow extends PanelLayout {
  constructor() {
    super();
    this.name = 'TableRow';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTableRow}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {

  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTableRow}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {

  }
}

/**
 * @hidden @internal
 * PanelLayoutTableColumn
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutTableColumn extends PanelLayout {
  constructor() {
    super();
    this.name = 'TableColumn';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTableColumn}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {

  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutTableColumn}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {

  }
}



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutViewbox
*/

// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { Util } from './util';


/**
 * @hidden @internal
 * PanelLayoutViewbox
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutViewbox extends PanelLayout {
  constructor() {
    super();
    this.name = 'Viewbox';
  }

  /**
   * @hidden @internal
   * Element scaled to fit inside of panel
   * @this {PanelLayoutViewbox}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    if (l > 1) Util.throwError('Viewbox Panel cannot contain more than one GraphObject.');
    const child = children[0];
    // Always measure the child with Infinity and not the width and height given to Viewbox
    child._scale = 1;
    child.invalidateMeasureChildren();
    child.measure(Infinity, Infinity, minw, minh);
    let m = child.measuredBounds;
    const marg = child.margin as Margin;
    const margw = marg.right + marg.left;
    const margh = marg.top + marg.bottom;

    // If the width and height are infinte then we measure and quit
    if (isFinite(width) || isFinite(height)) {
      const oldscale = child.scale;
      const mwidth = m.width; // no margin here
      const mheight = m.height;
      const availWidth = Math.max(width - margw, 0);
      const availHeight = Math.max(height - margh, 0);
      let value = 1;
      if (panel.viewboxStretch === GraphObject.Uniform) {
        // Resized accoring to the larger side
        if (mwidth !== 0 && mheight !== 0) value = Math.min((availWidth / mwidth), (availHeight / mheight));
        if (value === 0) value = 0.0001;
        child._scale *= value;
      } else { /// UniformToFill
        // Resized accoring to the smaller side
        if (mwidth !== 0 && mheight !== 0) value = Math.max((availWidth / mwidth), (availHeight / mheight));
        if (value === 0) value = 0.0001;
        child._scale *= value;
      }
      // if the scale changed, time to remeasure
      if (oldscale !== child.scale) {
        child.setInvalidMeasure(true);
        child.measure(Infinity, Infinity, minw, minh);
      }
    }
    m = child.measuredBounds;
    union.width = isFinite(width) ? width : Math.max(m.width + margw, 0);
    union.height = isFinite(height) ? height : Math.max(m.height + margh, 0);
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutViewbox}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const child = children[0];
    const m = child.measuredBounds;
    const marg = child.margin as Margin;
    const margw = marg.right + marg.left;
    const margh = marg.top + marg.bottom;
    const mwidth = Math.max(m.width + margw, 0);
    const mheight = Math.max(m.height + margh, 0);

    let align = child.alignment;
    if (align.isDefault()) align = panel.defaultAlignment;
    if (!align.isSpot()) align = Spot.Center;
    child.arrange(
      union.width * align.x - (mwidth * align.x) + align.offsetX,
      union.height * align.y - (mheight * align.y) + align.offsetY,
      m.width,
      m.height);
  }
}


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutGrid
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';


/**
 * @hidden @internal
 * PanelLayoutGrid
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutGrid extends PanelLayout {
  constructor() {
    super();
    this.name = 'Grid';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutGrid}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {

  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutGrid}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {

  }

  /**
   * @hidden @internal
   *
   */
  public draw(panel: Panel, context: IContext, diagram: Diagram): void {
    let scale = panel.getDocumentScale() * diagram.scale;
    if (scale <= 0) scale = 1;
    const csz = panel.gridCellSize;
    const cw = csz.width;
    const ch = csz.height;
    const nb = panel.naturalBounds;
    const ab = panel.actualBounds;
    const gw = nb.width;
    const gh = nb.height;
    const numx = Math.ceil(gw / cw);
    const numy = Math.ceil(gh / ch);
    const cpt = panel.gridOrigin;

    context.save();
    context.beginPath();
    context.rect(0, 0, gw, gh);
    context.clip();

    const intsarr = this._collectIntervals(panel);  // ??? cache!
    const elems = panel._elements._arr;
    const len = elems.length;
    for (let c = 0; c < len; c++) {
      const shape = elems[c] as Shape;
      // don't need to check for Shapes because .insertAt now checks
      // if (!(shape instanceof Shape)) continue;
      if (!shape.visible) continue;
      // use the properties of shape to draw relevant lines
      const freq = shape.interval;
      if (cw * freq * scale < 2) continue; // don't draw if too small
      const opacity = shape.opacity;
      let old = 1.0;
      if (opacity !== 1) {
        if (opacity === 0) continue;
        old = context.globalAlpha;
        context.globalAlpha = old * opacity;
      }
      const ints = intsarr[c];
      let dash = false;
      const strokeDashArray = shape.strokeDashArray;
      if (strokeDashArray !== null) {
        dash = true;
        context.enableDash(strokeDashArray, shape.strokeDashOffset);
      }

      if (shape.figure === 'LineV' && shape.stroke !== null) {
        context.lineWidth = shape.strokeWidth;
        panel.setBrush(context, shape.stroke, false, false, nb, ab);
        context.beginPath();
        const start = Math.floor(-cpt.x / cw);
        for (let i = start; i <= start + numx; i++) {
          const x = i * cw + cpt.x;
          if (0 <= x && x <= gw && this._shouldDraw(i, freq, ints)) {
            context.moveTo(x, 0);
            context.lineTo(x, gh);
          }
        }
        context.stroke();
      } else if (shape.figure === 'LineH' && shape.stroke !== null) {
        context.lineWidth = shape.strokeWidth;
        panel.setBrush(context, shape.stroke, false, false, nb, ab);
        context.beginPath();
        const start = Math.floor(-cpt.y / ch);
        for (let i = start; i <= start + numy; i++) {
          const y = i * ch + cpt.y;
          if (0 <= y && y <= gh && this._shouldDraw(i, freq, ints)) {
            context.moveTo(0, y);
            context.lineTo(gw, y);
          }
        }
        context.stroke();
      } else if (shape.figure === 'BarV' && shape.fill !== null) {
        panel.setBrush(context, shape.fill, true, false, nb, ab);
        let barw = shape.width;
        if (isNaN(barw)) barw = cw;
        const start = Math.floor(-cpt.x / cw);
        for (let i = start; i <= start + numx; i++) {
          const x = i * cw + cpt.x;
          if (0 <= x && x <= gw && this._shouldDraw(i, freq, ints)) {
            context.fillRect(x, 0, barw, gh);
          }
        }
      } else if (shape.figure === 'BarH' && shape.fill !== null) {
        panel.setBrush(context, shape.fill, true, false, nb, ab);
        let barh = shape.height;
        if (isNaN(barh)) barh = ch;
        const start = Math.floor(-cpt.y / ch);
        for (let i = start; i <= start + numy; i++) {
          const y = i * ch + cpt.y;
          if (0 <= y && y <= gh && this._shouldDraw(i, freq, ints)) {
            context.fillRect(0, y, gw, barh);
          }
        }
      }

      // Set dash back when done
      if (dash) {
        context.disableDash();
      }

      if (opacity !== 1) context.globalAlpha = old;
    } // end draw loop

    context.restore();
    context.clearContextCache(false);
  }

  /**
   * @hidden @internal
   * See if the Nth line should be drawn on a grid panel, considering both its frequency
   * and whether any later (in Z order) lines would be drawn at this point
   * @this {Panel}
   * @param {number} n
   * @param {number} freq
   * @param {Array.<number>} ints
   * @return {boolean}
   */
  private _shouldDraw(n: number, freq: number, ints: Array<number>): boolean {
    // negative numerators don't matter when comparing with zero
    // assumes FREQ > 0
    if (n % freq !== 0) return false;
    const len = ints.length;
    for (let j = 0; j < len; j++) {
      // assumes interval > 0
      const fs = ints[j];
      if (n % fs === 0) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * For now, just assume LineH, LineV, BarH, and BarV Shape.figure values.
   * Unlike GoXam, no support for Shape.offset, just Shape.interval.
   * @this {Panel}
   * @return {Array.<Array.<number>>}
   */
  private _collectIntervals(panel: Panel): Array<Array<number>> {
    const list: Array<Array<number>> = [];  // Array of Array of number
    const paths = panel._elements._arr;
    const len = paths.length;
    for (let i = 0; i < len; i++) {
      const path = paths[i] as Shape;
      const ints: Array<number> = [];  // intervals for this element,
      list.push(ints);  // even if this Shape should be ignored (e.g., !visible)
      if (!path.visible) continue;
      const fig = path.figure;
      const vert = panel._isVertical(fig);
      // iterate over other shapes in the same orientation and collect their intervals
      for (let j = i + 1; j < len; j++) {
        const q = paths[j] as Shape;
        if (!q.visible) continue;
        const f = q.figure;
        if (panel._isVertical(f) !== vert) continue;  // skip Shapes that are not the same orientation
        const v = q.interval;
        if (v >= 2) ints.push(v);  // save interval info if it isn't every time
      }
    }
    return list;
  }


}


// import { Diagram } from './diagram';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { IContext } from './surfaceInterfaces';




/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutLink
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Spot } from './spot';

/**
 * @hidden @internal
 * PanelLayoutLink
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutLink extends PanelLayout {
  constructor() {
    super();
    this.name = 'Link';
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutLink}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const l = children.length;
    if (!(panel instanceof Adornment || panel instanceof Link)) return; // Because TS is not always smart
    // There are some methods we want to use the adorned part for if this is an adornment
    let maybelink = null;
    let asLink: Link = null;
    let asAdornment: Adornment = null;
    if (panel instanceof Link) {
      maybelink = panel as Link;
      asLink = panel as Link; // Tsickle: type Link
    }
    if (panel instanceof Adornment) {
      asAdornment = panel as Adornment; // Tsickle: type Adornment
      maybelink = asAdornment.adornedPart as Link;
    }

    if (!(maybelink instanceof Link)) return;
    const link = maybelink;

    // no child elements?
    if (l === 0) {
      const ns = panel.naturalBounds;
      ns.assignWidthHeight(0, 0);
      const mb = panel.measuredBounds;
      mb.assignTo(0, 0, 0, 0);
      return;
    }

    const stroke = (panel instanceof Adornment) ? null : maybelink.path; // may be null
    const routeBounds = maybelink.routeBounds;
    const linkBounds = panel._union;
    // We want to repurpose x/y
    linkBounds.assignTo(0, 0, routeBounds.width, routeBounds.height); // will include all decorations


    const pts = link.points;
    // Adornments such as ReshapeLink do not have a .pointsCount
    const numpts = maybelink.pointsCount;
    if (asAdornment !== null) { // needed for closure compiler only
      asAdornment.makeShapeGeometry(false);
    } else if (asLink !== null) {
      asLink.makeShapeGeometry(false);
    }

    const geometryWidth = routeBounds.width;  // Math.max(routeBounds.width, this.geometry.bounds.width);
    const geometryHeight = routeBounds.height; // Math.max(routeBounds.height, this.geometry.bounds.height);

    // Don't want to set property, or at least we want this to skip undo
    panel.location.assignTo(routeBounds.x, routeBounds.y);
    panel._childrenBounds.length = 0;
    if (stroke !== null) {
      this.measureLinkPath(panel, geometryWidth, geometryHeight, stroke);
      const mb = stroke.measuredBounds;
      linkBounds.unionRect(mb);
      panel._childrenBounds.push(mb);  // remember for Arrange
    }

    const et = Transform.alloc();
    const alignpt = Point.alloc();
    const offsetpt = Point.alloc();

    // measure all other children
    for (let i = 0; i < l; i++) {
      const e = children[i];
      if (e === stroke) continue;  // already measured the stroke, above

      let mb = e.measuredBounds;
      // Take all shapes that have panelMain = true and give them a specific, shared geometry
      if (e.isPanelMain && e instanceof Shape) {
        this.measureLinkPath(panel, geometryWidth, geometryHeight, e);
        mb = e.measuredBounds;
        linkBounds.unionRect(mb);
        panel._childrenBounds.push(mb);
        continue;
      }

      // if no route it's just a simple measurement
      if (numpts < 2) {
        e.measure(Infinity, Infinity, 0, 0);
        mb = e.measuredBounds;
        linkBounds.unionRect(mb);
        panel._childrenBounds.push(mb);
        continue;
      }

      // get the properties controlling where the "label" should be along the route
      const index = e.segmentIndex;
      const frac = e.segmentFraction;
      let align = e.alignmentFocus;
      if (align.isNoSpot()) align = Spot.Center;
      const orient = e.segmentOrientation;
      const offset = e.segmentOffset;

      // compute the point along the route at which the element must be positioned
      let itemX = 0.0; // local coordinates
      let itemY = 0.0;
      let labelangle = 0.0;
      let segangle = 0.0;
      if (index < -numpts || index >= numpts) {  // beyond range? assume at the MidPoint, with the MidAngle
        const mid = link.midPoint;  // in document coordinates
        segangle = link.midAngle;
        // maybe rotate the label
        if (orient !== Link.None) {
          labelangle = link.computeAngle(e, orient, segangle);
          e._angle = labelangle; // _angle because we don't want it to be in the undo manager
        }
        itemX = mid.x - routeBounds.x;  // local coordinates
        itemY = mid.y - routeBounds.y;
      } else {  // on a particular segment, given by Index, at a point given by Fraction
        // negative index means start from last point, going "backwards"
        let a: Point;
        let b: Point; // Points
        let idx = 0;
        if (index >= 0) {
          a = pts.elt(index);
          b = (index < numpts - 1) ? pts.elt(index + 1) : a;
        } else {
          idx = numpts + index;  // remember that index is negative here
          a = pts.elt(idx);
          b = (idx > 0) ? pts.elt(idx - 1) : a;
        }
        segangle = 0;
        // if both of these segment points are the same, check points on either side to determine a direction for the segment
        // this works similar to computeMidAngle
        if (a.equalsApprox(b)) {
          let c: Point; // previous point, if available; otherwise, just use a
          let d: Point; // next point, if available; otherwise, just use b
          if (index >= 0) {
            c = (index > 0) ? pts.elt(index - 1) : a;
            d = (index < numpts - 2) ? pts.elt(index + 2) : b;
          } else {
            c = (idx < numpts - 1) ? pts.elt(idx + 1) : a;
            d = (idx > 1) ? pts.elt(idx - 2) : b;
          }
          const d1 = c.distanceSquaredPoint(a);
          const d2 = b.distanceSquaredPoint(d);
          if (d1 > d2 + 10) {
            segangle = (index >= 0 ? c.directionPoint(a) : a.directionPoint(c));
          } else if (d2 > d1 + 10) {
            segangle = (index >= 0 ? b.directionPoint(d) : d.directionPoint(b));
          } else {
            segangle = (index >= 0 ? c.directionPoint(d) : d.directionPoint(c));
          }
        } else {
          segangle = (index >= 0 ? a.directionPoint(b) : b.directionPoint(a));
        }
        // maybe rotate the label
        if (orient !== Link.None) {
          labelangle = link.computeAngle(e, orient, segangle);
          e._angle = labelangle; // _angle because we don't want it to be in the undo manager
        }
        // compute the fractional point along the line, in local coordinates
        itemX = a.x + (b.x - a.x) * frac - routeBounds.x;
        itemY = a.y + (b.y - a.y) * frac - routeBounds.y;
      }

      e.measure(Infinity, Infinity, 0, 0);  // NB: e._angle may have changed
      mb = e.measuredBounds;
      const ns = e.naturalBounds;
      let sw = 0.0;
      if (e instanceof Shape) {
        sw = e.strokeWidth;
      }
      const ew = ns.width + sw;
      const eh = ns.height + sw;

      // now determine the label/arrowhead's bounds
      et.reset();
      et.translate(-mb.x, -mb.y);
      et.scale(e.scale, e.scale);
      et.rotate((orient === Link.None) ? e.angle : segangle, ew / 2, eh / 2);
      // take into account additional rotation due to +/-
      if (orient === Link.OrientPlus90 || orient === Link.OrientPlus90Upright) {
        et.rotate(90, ew / 2, eh / 2);
      }
      if (orient === Link.OrientMinus90 || orient === Link.OrientMinus90Upright) {
        et.rotate(-90, ew / 2, eh / 2);
      }
      if (orient === Link.OrientUpright45 && ((segangle > 45 && segangle < 135) || (segangle > 225 && segangle < 315))) {
        et.rotate(-segangle, ew / 2, eh / 2);
      }

      // find the transformed alignment point inside the element
      const nb = new Rect(0, 0, ew, eh);
      alignpt.setRectSpot(nb, align);
      et.transformPoint(alignpt);
      const dx = -alignpt.x + (sw / 2) * e.scale;
      const dy = -alignpt.y + (sw / 2) * e.scale;

      // maybe the alignment point is away from the line
      // if the offset is NaN, it becomes half the width and half the height of the label object
      // the + 3 is an arbitrary additional offset (exists in GoXam too)
      offsetpt.assign(offset);
      const nanx = isNaN(offset.x);
      const nany = isNaN(offset.y);
      if (nanx || nany) {
        const w2 = ew / 2 + 3;
        const h2 = eh / 2 + 3;
        // hard-coding 45 degree leeways assumes squarish nodes
        const downwards = segangle >= 45 && segangle <= 135;
        const upwards = segangle >= 225 && segangle <= 315;
        // specially handle cases when the label offset is NaN,NaN and is not being rotated
        // when the route is more vertical than horizontal, adjust for any non-square labels
        if (orient === Link.None && (downwards || upwards)) {
          // don't rotate; exchange x and y
          offsetpt.x = nany ? w2 : offset.y;
          offsetpt.y = nanx ? h2 : offset.x;
          if (downwards) {
            if (index >= 0) {
              // if (nanx) offsetpt.y = h2;
              // if (nany) offsetpt.x = w2;
            } else {
              if (nanx) offsetpt.y = -h2;
              // if (nany) offsetpt.x = w2;
            }
          } else if (upwards) {
            if (index >= 0) {
              if (nanx) offsetpt.y = -h2;
              if (nany) offsetpt.x = -w2;
            } else {
              // if (nanx) offsetpt.y = h2;
              if (nany) offsetpt.x = -w2;
            }
          }
        } else {
          if (nanx) {
            if (index >= 0) {
              offsetpt.x = w2;
            } else {
              offsetpt.x = -w2;
            }
          }
          if (nany) {
            offsetpt.y = -h2;
          }
          offsetpt.rotate(segangle);
        }
      } else {  // real values for offsetpt
        offsetpt.rotate(segangle);
      }
      itemX += offsetpt.x;
      itemY += offsetpt.y;

      // reuse the Rect NB to remember for arrangeProtected
      nb.set(mb);
      nb.assignTo(itemX + dx, itemY + dy, mb.width, mb.height);
      panel._childrenBounds.push(nb);
      linkBounds.unionRect(nb);
    } // done measuring children

    if (asLink !== null) { // Adornments such as ReshapeLink do not have labelNodes
      // LabelNodes are distinct top-level objects and as such don't have to obey any of the rules of the Link.
      const labels = asLink.labelNodes;
      while (labels.next()) {
        const lab = labels.value;
        lab.measure(Infinity, Infinity);
      }
    }

    panel._union = linkBounds;

    // The location must be modified by the linkBounds
    // which was the union of all the link shape's bounds during measure
    const loc = panel.location;
    loc.assignTo(loc.x + linkBounds.x, loc.y + linkBounds.y);

    // ignore any desired size
    union.assignWidthHeight(linkBounds.width || 0, linkBounds.height || 0);

    Transform.free(et);
    Point.free(alignpt);
    Point.free(offsetpt);
  }

  /**
   * @hidden @internal
   *
   * @this {PanelLayoutLink}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    const l = children.length;

    if (!(panel instanceof Adornment || panel instanceof Link)) return; // Because TS is not always smart
    // There are some methods we want to use the adorned part for if this is an adornment
    let maybelink: Link = null;
    let asLink: Link = null;
    let asAdornment: Adornment = null;
    if (panel instanceof Link) {
      maybelink = panel as Link;
      asLink = panel as Link; // Tsickle: type Link
    }
    if (panel instanceof Adornment) {
      asAdornment = panel as Adornment; // Tsickle: type Adornment
      maybelink = asAdornment.adornedPart as Link;
    }
    const that = panel; // Tsickle: type Adornment|Link
    const link: Link = maybelink;
    // First, we operate on the "path" object that was panelMain = true and a Shape
    const stroke = (panel instanceof Adornment) ? null : link.path; // may be null
    if (that._childrenBounds.length !== 0) {
      const childrenBounds = that._childrenBounds;
      let childidx = 0;
      if (stroke !== null) {
        if (childidx < that._childrenBounds.length) {
          const cb = childrenBounds[childidx];
          childidx++;
          stroke.arrange(cb.x - that._union.x, cb.y - that._union.y, cb.width, cb.height);
        }
      }

      for (let i = 0; i < l; i++) {
        const e = children[i];
        if (e === stroke) continue;
        if (childidx < that._childrenBounds.length) {
          const cb = childrenBounds[childidx];
          childidx++;
          // cb is in the parent's coordinate system and not local,
          // change it to local:
          e.arrange(cb.x - that._union.x, cb.y - that._union.y, cb.width, cb.height);
        }
      }
    }

    const pts = link.points;
    const numpts = pts.count;
    if (numpts >= 2 && that instanceof Link) {
      const lnit = that.labelNodes;
      while (lnit.next()) {
        const lab = lnit.value;
        that._arrangeLabelNode(lab, numpts, pts);
      }
    }
    // Different depending on Link or Adornment:
    if (asAdornment !== null) { // needed for closure compiler only
      asAdornment.makeShapeGeometry(false);
    } else if (asLink !== null) {
      asLink.makeShapeGeometry(false);
    }
  }


  /**
   * @hidden @internal
   * For Link Shapes only.
   * Create a Shape's measured bounds based on its geometry while making some assumptions:
   * - The geometry exists
   * - There is no stretch and width/height do not matter
   * These conditions should almost never be assumed, except when the shape is a main element of a Link.
   * This function replaces both measure and measureProtected.
   * @this {Panel}
   * @param {number} width
   * @param {number} height
   * @param {Shape} linkpath the Shape's part, which will always be a Link or Link Adornment
   */
  public measureLinkPath(panel: Part, width: number, height: number, linkpath: Shape): void {
    // don't measure unless invalid
    if (linkpath.getInvalidMeasure() === false) return;

    let logicalSW = linkpath.strokeWidth;
    if (logicalSW === 0 && panel instanceof Adornment && panel.type === Panel.Link) {
      if (panel.adornedObject instanceof Shape) {
        logicalSW = panel.adornedObject.strokeWidth;
      }
    }
    logicalSW = logicalSW * linkpath._scale;

    if (panel instanceof Link && panel._geometry !== null) {
      const geometrybounds = panel._geometry.bounds;
      linkpath._measureRect(geometrybounds.x - (logicalSW / 2), geometrybounds.y - (logicalSW / 2), geometrybounds.width + logicalSW, geometrybounds.height + logicalSW);
    } else if (panel instanceof Adornment && (panel.adornedPart as Link)._geometry !== null) {
      const geometrybounds = (panel.adornedPart as Link)._geometry.bounds;
      linkpath._measureRect(geometrybounds.x - (logicalSW / 2), geometrybounds.y - (logicalSW / 2), geometrybounds.width + logicalSW, geometrybounds.height + logicalSW);
    } else {
      linkpath._measureRect(-(logicalSW / 2), -(logicalSW / 2), width + logicalSW, height + logicalSW);
    }

    linkpath.setInvalidMeasure(false);
  }


}



// import { Adornment, Link, Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { Transform } from './transform';




/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PanelLayoutGraduated
*/

// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { PanelLayout } from './panelLayout';
// import { Spot } from './spot';
// import { Util } from './util';

/** @hidden @internal */
export interface GraduatedMark {
  pt: Point;
  angle: number;
  text?: string;
  labelAngle?: number;
  lineCount?: number;
  lines?: [number, number, number | null, number, Array<number> | null, Array<string> | null]; // @@@ ???
  actualBounds?: Rect;
  measuredBounds?: Rect;
  naturalBounds?: Rect;
}

/**
 * @hidden @internal
 * PanelLayoutGraduated
 * @extends PanelLayout
 * @unrestricted
 */
export class PanelLayoutGraduated extends PanelLayout {
  constructor() {
    super();
    this.name = 'Graduated';
  }

  /**
   * @hidden @internal
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {number} width
   * @param {number} height
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   * @param {number} minw
   * @param {number} minh
   */
  public measure(panel: Panel, width: number, height: number, children: Array<GraphObject>, union: Rect, minw: number, minh: number): void {
    const main = panel.findMainElement() as Shape;

    panel._graduatedBounds = [];

    const marg = main.margin as Margin;
    const mainmargw = marg.right + marg.left;
    const mainmargh = marg.top + marg.bottom;
    main.measure(width, height, minw, minh);
    const m = main.measuredBounds;
    const mainWidth = m.width;
    const mainHeight = m.height;
    const mwidth = Math.max(mainWidth + mainmargw, 0);
    const mheight = Math.max(mainHeight + mainmargh, 0);

    // ?? allocated rect but put into array
    const mainRect = new Rect(-marg.left, -marg.top, mwidth, mheight);
    panel._graduatedBounds.push(mainRect);
    union.assign(mainRect);

    // Determine placement of ticks/text
    this.determineGraduatedMarks(panel, main);
    const marks = panel._graduatedMarks;
    if (marks === null) return;
    const l = children.length;
    for (let c = 0; c < l; c++) {
      const elt = children[c];
      const markArr = marks[c];
      if (!elt.visible || elt === main || markArr.length === 0) continue;

      // measure the overall space taken by an element, then mark it as measured despite not calling measure()
      if (elt instanceof Shape) {
        this.measureGraduatedTicks(panel, elt, markArr, union);
      } else if (elt instanceof TextBlock) {
        this.measureGraduatedLabels(panel, elt, markArr, union);
      }
      elt.setInvalidMeasure(false);
    }
  }

  /**
   * @hidden @internal
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {Array.<GraphObject>} children
   * @param {Rect} union
   */
  public arrange(panel: Panel, children: Array<GraphObject>, union: Rect): void {
    if (panel._graduatedBounds === null) return; // ?? might want to do more than just return, throw an error, etc
    const main = panel.findMainElement();
    const marks = panel._graduatedMarks;
    if (marks === null) return;
    const graduatedBounds = panel._graduatedBounds;

    let bIdx = 0; // separate index since not all elements necessarily have bounds
    let cb = graduatedBounds[bIdx];
    bIdx++;
    if (main !== null) main.arrange(cb.x - union.x, cb.y - union.y, cb.width, cb.height);

    const l = children.length;
    for (let i = 0; i < l; i++) {
      const elt = children[i];
      const markArr = marks[i];

      if (!elt.visible || elt === main || markArr.length === 0) continue;

      cb = graduatedBounds[bIdx];
      bIdx++;
      elt.arrange(cb.x - union.x, cb.y - union.y, cb.width, cb.height);
    }
    panel._graduatedBounds = null;
  }



  /**
   * @hidden @internal
   * Use main element as the measure, then perform a union with ticks/text.
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {GraphObject} elt
   * @param {Array.<Array.<*>>} ticks
   * @param {Rect} union
   */
  public measureGraduatedTicks(panel: Panel, elt: GraphObject, ticks: Array<GraduatedMark>, union: Rect): void {
    // get four pts for the mark, then we can translate/rotate for the various places needed
    let align = elt.alignmentFocus;
    if (align.isNoSpot()) align = Spot.TopCenter;

    // using bounds at (0, 0) and angle of 0, determine four corners of bounding box
    const savedAngle = elt.angle;
    elt._angle = 0;
    elt.measure(Infinity, Infinity, 0, 0);
    elt._angle = savedAngle;
    const mb = elt.measuredBounds;
    const rectWidth = mb.width;
    const rectHeight = mb.height;

    // find the transformed alignment point inside the element
    const nb = Rect.allocAt(0, 0, rectWidth, rectHeight);
    const tempPt = Point.alloc();
    tempPt.setRectSpot(nb, align);
    Rect.free(nb);
    const rectX = -tempPt.x;
    const rectY = -tempPt.y;

    // for every tick, make sure it's included in the union
    const cb = new Rect(); // must allocate, pushed to _graduatedBounds
    const l = ticks.length;
    for (let i = 0; i < l; i++) {
      const ti = ticks[i];
      const x = ti.pt.x;
      const y = ti.pt.y;
      const angle = ti.angle;
      // for each of the four pts making up the mark, rotate and translate, then add to the union
      for (let j = 0; j < 4; j++) {

        switch (j) {
          case 0: tempPt.assignTo(rectX, rectY); break; // left, top
          case 1: tempPt.assignTo(rectX + rectWidth, rectY); break; // right, top
          case 2: tempPt.assignTo(rectX, rectY + rectHeight); break; // left, bottom
          case 3: tempPt.assignTo(rectX + rectWidth, rectY + rectHeight); break; // right, bottom
        }

        tempPt.rotate(angle + elt.angle);
        tempPt.offset(x, y);

        if (i === 0 && j === 0) {
          cb.assignTo(tempPt.x, tempPt.y, 0, 0); // create an initial Rect based on the very first point
        } else {
          cb.unionPoint(tempPt); // union with the Rect to expand bounds as we go
        }

        // restore to original
        tempPt.offset(-x, -y);
        tempPt.rotate(-angle - elt.angle);
      }
    }
    Point.free(tempPt);
    if (panel._graduatedBounds !== null) panel._graduatedBounds.push(cb);
    union.unionNoCheck(cb.x, cb.y, cb.width, cb.height);
  }

  /**
   * @hidden @internal
   * Use main element as the measure, then perform a union
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {TextBlock} elt
   * @param {Array.<Array.<*>>} labels
   * @param {Rect} union
   */
  public measureGraduatedLabels(panel: Panel, elt: TextBlock, labels: Array<GraduatedMark>, union: Rect): void {
    if (panel._temporaryTextBlock === null) panel._temporaryTextBlock = new TextBlock();
    const tempText = panel._temporaryTextBlock;
    this._copyGraduatedLabelProperties(elt, tempText);

    let align = elt.alignmentFocus;
    if (align.isNoSpot()) align = Spot.TopCenter;
    const orient = elt.segmentOrientation;
    const offset = elt.segmentOffset;

    // for every mark, make sure it's included in the union
    const cb = new Rect();
    let x = 0;
    let y = 0;
    let mainangle = 0;
    let labelangle = 0;
    const l = labels.length;
    // first pass gathers rects and saves measure info
    for (let i = 0; i < l; i++) {
      const li = labels[i];
      x = li.pt.x;
      y = li.pt.y;
      mainangle = li.angle;

      if (orient !== Link.None) {
        labelangle = Link.computeAngle(orient, mainangle);
        tempText._angle = labelangle;
      }
      tempText.text = li.text || '';
      tempText.measure(Infinity, Infinity, 0, 0);

      const mb = tempText.measuredBounds;
      const ns = tempText.naturalBounds;
      const tw = ns.width;
      const th = ns.height;

      // now determine the label bounds
      const et = Transform.alloc();
      et.reset();
      et.translate(-mb.x, -mb.y);
      et.scale(tempText.scale, tempText.scale);
      et.rotate((orient === Link.None) ? tempText.angle : mainangle, tw / 2, th / 2);
      // take into account additional rotation due to +/-
      if (orient === Link.OrientPlus90 || orient === Link.OrientPlus90Upright) {
        et.rotate(90, tw / 2, th / 2);
      }
      if (orient === Link.OrientMinus90 || orient === Link.OrientMinus90Upright) {
        et.rotate(-90, tw / 2, th / 2);
      }
      if (orient === Link.OrientUpright45 && ((mainangle > 45 && mainangle < 135) || (mainangle > 225 && mainangle < 315))) {
        et.rotate(-mainangle, tw / 2, th / 2);
      }

      // find the transformed alignment point inside the element
      const nb = Rect.allocAt(0, 0, tw, th);
      const alignpt = Point.alloc();
      alignpt.setRectSpot(nb, align);
      et.transformPoint(alignpt);
      const dx = -alignpt.x;
      const dy = -alignpt.y;

      // maybe the alignment point is away from the line
      const offsetpt = Point.alloc();
      offsetpt.assign(offset);
      if (isNaN(offsetpt.x)) {
        offsetpt.x = tw / 2 + 3;
      }
      if (isNaN(offsetpt.y)) {
        offsetpt.y = -(th / 2 + 3);
      }
      offsetpt.rotate(mainangle);

      x += offsetpt.x + dx;
      y += offsetpt.y + dy;

      // add additional info to the array needed during drawing
      const rectAb = new Rect(x, y, mb.width, mb.height);
      const rectMb = new Rect(mb.x, mb.y, mb.width, mb.height);
      const rectNb = new Rect(ns.x, ns.y, ns.width, ns.height);
      li.labelAngle = labelangle;
      li.lineCount = tempText.lineCount;
      li.lines = tempText.getMetrics();
      li.actualBounds = rectAb;
      li.measuredBounds = rectMb;
      li.naturalBounds = rectNb;

      // add to union
      if (i === 0) {
        cb.assign(rectAb); // create an initial Rect based on the first label
      } else {
        cb.unionRect(rectAb); // union with the Rect to expand bounds as we go
      }

      Point.free(offsetpt);
      Point.free(alignpt);
      Rect.free(nb);
      Transform.free(et);
    }
    if (panel._graduatedBounds !== null) panel._graduatedBounds.push(cb);
    union.unionNoCheck(cb.x, cb.y, cb.width, cb.height);
  }

  /**
   * @hidden @internal
   * Traverse main path and determine points and angles for where ticks/text will be placed.
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {Shape} main
   */
  public determineGraduatedMarks(panel: Panel, main: Shape): void {
    // Build paths for the main Geometry
    const geo = main.geometry;
    const sw = main.strokeWidth;
    // Get paths, distances, and total distance from the geometry
    const allPaths = geo.flattenedSegments;
    const allDists = geo.flattenedLengths;
    const totalDist = geo.flattenedTotalLength;

    const l = allPaths.length;
    let lastptx = 0.0;
    let lastpty = 0.0;
    const allAngles = Util.tempArray<Array<number>>();
    for (let i = 0; i < l; i++) {
      const path = allPaths[i];
      const currAngle = [];
      lastptx = 0;
      lastpty = 0;
      const ll = path.length;
      for (let j = 0; j < ll; j += 2) {
        const ptx = path[j];
        const pty = path[j + 1];
        if (j === 0) { // its a new path because a moveTo occurred, so set the lastPt and move on
          lastptx = ptx;
          lastpty = pty;
          continue;
        }
        let angle = Math.atan2(pty - lastpty, ptx - lastptx) * 180 / Math.PI;
        if (angle < 0) angle += 360;
        currAngle.push(angle);
        lastptx = ptx;
        lastpty = pty;
      }
      allAngles.push(currAngle);
    }

    // Loop through tick intervals and determine tick distances
    const intsarr = this._collectTickIntervals(panel);
    const elems = panel._elements._arr;
    const len = elems.length;
    let currDist = 0;
    let startDist = 0;
    let endDist = totalDist;
    panel._graduatedMarks = [];
    let markArr: Array<GraduatedMark>;

    for (let c = 0; c < len; c++) {
      const elt = elems[c] as Shape | TextBlock;
      markArr = [];
      if (!elt.visible || elt === main) {
        panel._graduatedMarks.push(markArr); // need to keep lists at same length
        continue;
      }
      const freq = elt.interval;
      const delta = panel.graduatedTickUnit;
      // don't draw if too close
      if (delta * freq * totalDist / panel.graduatedRange < 2) {
        panel._graduatedMarks.push(markArr); // need to keep lists at same length
        continue;
      }
      // keep track of which segment we're looking at and the distance we've covered
      let traveledDist = allDists[0][0];
      let currFigNum = 0;
      let currSegNum = 0;
      startDist = totalDist * elt.graduatedStart - .0001; // include slight tolerance for floating point math
      endDist = totalDist * elt.graduatedEnd + .0001;

      const tickDelta = delta * freq;
      let val = panel.graduatedTickBase;
      if (val < panel.graduatedMin) {
        let mod = (panel.graduatedMin - val) / tickDelta;
        mod = mod % 1 === 0 ? mod : Math.floor(mod + 1); // truncate result to int
        val += mod * tickDelta;
      } else if (val > panel.graduatedMin + tickDelta) {
        const mod = Math.floor((val - panel.graduatedMin) / tickDelta);
        val -= mod * tickDelta;
      }
      const ints = intsarr[c];
      while (val <= panel.graduatedMax) {
        if (this._shouldDrawTick(panel, (val - panel.graduatedTickBase), ints) &&
            (elt.graduatedSkip === null || !elt.graduatedSkip(val))) {
          currDist = (val - panel.graduatedMin) * totalDist / panel.graduatedRange;
          if (currDist > totalDist) currDist = totalDist; // in case of floating point errors going beyond end
          if (startDist <= currDist && currDist <= endDist) {
            let angle = allAngles[currFigNum][currSegNum];
            let segDist = allDists[currFigNum][currSegNum];
            // find which segment the tick falls in
            while (currFigNum < allDists.length) {
              while (currDist > traveledDist && currSegNum < allDists[currFigNum].length - 1) {
                currSegNum++;
                angle = allAngles[currFigNum][currSegNum];
                segDist = allDists[currFigNum][currSegNum];
                traveledDist += segDist;
              }
              if (currDist <= traveledDist) break;
              // move to next figure
              currFigNum++;
              currSegNum = 0;
              angle = allAngles[currFigNum][currSegNum];
              segDist = allDists[currFigNum][currSegNum];
              traveledDist += segDist;
            }

            const figPts = allPaths[currFigNum];
            const x = figPts[currSegNum * 2];
            const y = figPts[currSegNum * 2 + 1];
            const x2 = figPts[currSegNum * 2 + 2];
            const y2 = figPts[currSegNum * 2 + 3];

            const offsetBetweenPoints = (currDist - (traveledDist - segDist)) / segDist; // fractional distance along the current segment
            const pt = new Point(x + (x2 - x) * offsetBetweenPoints + sw / 2 - geo.bounds.x,
              y + (y2 - y) * offsetBetweenPoints + sw / 2 - geo.bounds.y);
            pt.scale(main.scale, main.scale);
            // if we're very close to the end of a segment, consider smoothing the angle
            let otherangle = angle;
            const figAngles = allAngles[currFigNum];
            if (offsetBetweenPoints < .0001) {
              if (currSegNum > 0) {
                // there is a prior segment we can look at
                otherangle = figAngles[currSegNum - 1];
              } else if (Geo.isApproxEqual(figPts[0], figPts[figPts.length - 2]) &&
                Geo.isApproxEqual(figPts[1], figPts[figPts.length - 1])) {
                // this is a closed figure and we can look at the last segment
                otherangle = figAngles[figAngles.length - 1];
              }
            } else if (offsetBetweenPoints > .9999) {
              if (currSegNum + 1 < figAngles.length) {
                // there is a next segment we can look at
                otherangle = figAngles[currSegNum + 1];
              } else if (Geo.isApproxEqual(figPts[0], figPts[figPts.length - 2]) &&
                Geo.isApproxEqual(figPts[1], figPts[figPts.length - 1])) {
                // this is a closed figure and we can look at the first segment
                otherangle = figAngles[0];
              }
            }
            if (angle !== otherangle) {
              const anglediff = Math.abs(angle - otherangle);
              if (anglediff > 180) {
                if (angle < otherangle) angle += 360;
                else otherangle += 360;
              }
              angle = (angle + otherangle) / 2 % 360;
            }

            // add to list of points/angles/text for drawing
            if (elt instanceof TextBlock) {
              let txt = '';
              if (elt.graduatedFunction !== null) {
                txt = elt.graduatedFunction(val);
                txt = (txt !== null && txt !== undefined) ? txt.toString() : '';
              } else {
                txt = (+val.toFixed(2)).toString(); // round to at most 2 decimals
              }
              if (txt !== '') {
                markArr.push({ pt, angle, text: txt });
              }
            } else {
              markArr.push({ pt, angle });
            }
          }
        }
        val += tickDelta;
      }
      panel._graduatedMarks.push(markArr);
    }
    Util.freeArray(allAngles);
  }

  /**
   * @hidden @internal
   * Get the various intervals for ticks/text being drawn along the main shape of a Graduated Panel.
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @return {Array.<Array.<number>>}
   */
  private _collectTickIntervals(panel: Panel): Array<Array<number>> {
    if (panel._graduatedIntervals === null) {
      const list: Array<Array<number>> = [];  // Array of Array of number
      const paths = panel._elements._arr;
      const len = paths.length;
      for (let i = 0; i < len; i++) {
        const path = paths[i] as Shape;
        const ints: Array<number> = [];  // intervals for this element,
        list.push(ints);  // even if this Shape should be ignored (e.g., !visible)
        if (!path.visible) continue;
        const intp = path.interval;
        // iterate over other shapes and collect their intervals
        for (let j = 0; j < len; j++) {
          const q = paths[j] as Shape;
          if (!q.visible || path === q) continue;
          // Shapes and TextBlocks are treated independently
          if ((path instanceof Shape && !(q instanceof Shape)) ||
            (path instanceof TextBlock && !(q instanceof TextBlock))) continue;
          const intq = q.interval;
          if (intq > intp) ints.push(intq);  // save interval info if it is less frequently drawn
        }
      }
      panel._graduatedIntervals = list;
    }
    return panel._graduatedIntervals;
  }


  /**
   * @hidden @internal
   * See if a tick should be drawn on a Graduated panel, considering its value,
   * and whether other ticks with greater intervals would be drawn.
   * @this {PanelLayoutGraduated}
   * @param {Panel} panel
   * @param {number} val
   * @param {Array.<number>} ints
   * @return {boolean}
   */
  private _shouldDrawTick(panel: Panel, val: number, ints: Array<number>): boolean {
    const len = ints.length;
    for (let j = 0; j < len; j++) {
      // assumes interval > 0
      const fs = ints[j] * panel.graduatedTickUnit;
      if (Geo.isApproxEqual(val % fs, 0)) return false; // approxEqual for floating point math
    }
    return true;
  }


  /**
   * @hidden @internal
   *
   */
  public draw(panel: Panel, context: IContext, diagram: Diagram): void {
    const olddraw = diagram._isDrawingForeign;
    diagram._isDrawingForeign = true;
    const nb = panel.naturalBounds;
    const gw = nb.width;
    const gh = nb.height;

    context.save();
    // clip outside the Panel's bounds; use a slight margin for cases where text is measured too small
    context.beginPath();
    context.rect(-1, -1, gw + 1, gh + 1);
    context.clip();

    const main = panel.findMainElement() as Shape;
    main.draw(context, diagram);

    let scale = panel.getDocumentScale() * diagram.scale;
    if (scale <= 0) scale = 1;
    const mainab = main.actualBounds;
    const elems = panel._elements._arr;
    const marks = panel._graduatedMarks;
    const len = elems.length;
    for (let c = 0; c < len; c++) {
      const elt = elems[c];
      const markArr = marks[c];
      const ml = markArr.length;

      if (!elt.visible || elt === main || markArr.length === 0) continue;

      if (elt instanceof Shape) {
        if (panel.graduatedTickUnit * elt.interval * main.geometry.flattenedTotalLength / panel.graduatedRange * scale < 2) continue; // don't draw if we're scaled out too much

        const mb = elt.measuredBounds;
        const sw = elt.strokeWidth * elt.scale;
        let align = elt.alignmentFocus;
        if (align.isNoSpot()) align = Spot.TopCenter;

        for (let i = 0; i < ml; i++) {
          const pt = markArr[i].pt;
          const angle = markArr[i].angle;

          this._setGraduatedTickTransform(elt, pt, mainab, angle, mb, sw, align);
          elt.draw(context, diagram);
          elt._transform.reset(); // clear out transform when done drawing so it doesn't affect picking
        }
      } else if (elt instanceof TextBlock) {
        if (panel._temporaryTextBlock === null) panel._temporaryTextBlock = new TextBlock();
        const tempText = panel._temporaryTextBlock;
        this._copyGraduatedLabelProperties(elt, tempText);

        for (let i = 0; i < ml; i++) {
          const textInfo = markArr[i];
          if (textInfo.actualBounds && textInfo.measuredBounds && textInfo.naturalBounds) { // if we didn't append additional measure info, we don't want to draw it
            tempText._restoreMeasurements(textInfo);
            // bounds info from measuring
            const ab = textInfo.actualBounds;
            tempText.arrange(ab.x, ab.y, ab.width, ab.height);
            // set up the transform for drawing since we don't measure
            this._setGraduatedLabelTransform(tempText, mainab, ab, textInfo.measuredBounds, textInfo.naturalBounds);
            tempText.draw(context, diagram);
          }
        }
      }
    }
    diagram._isDrawingForeign = olddraw;
    context.restore();
    context.clearContextCache(true);
  }

  /**
   * @hidden @internal
   * Set the transform such that when we draw, the context gets adjusted correctly for a particular tick
   *
   * @param {Shape} s
   * @param {Point} pt
   * @param {Rect} mainab
   * @param {number} angle
   * @param {Rect} mb Shape.measuredBounds
   * @param {number} sw Shape.strokeWidth
   * @param {Spot} align Shape.alignmentFocus
   */
  private _setGraduatedTickTransform(s: Shape, pt: Point, mainab: Rect, angle: number, mb: Rect, sw: number, align: Spot): void {
    const tt = s._transform;
    tt.reset();

    tt.translate(pt.x + mainab.x, pt.y + mainab.y);
    tt.rotate(angle + s.angle, 0, 0);
    tt.translate(-mb.width * align.x + align.offsetX + sw / 2, -mb.height * align.y + align.offsetY + sw / 2);
    tt.scale(s.scale, s.scale);

    s.setInvalidTransform(false);

    s._allTransforms.set(s._transform);
    s._documentScale = s.scale;
    s.setInvalidAllTransforms(false);
  }

  /**
   * @hidden @internal
   * Since we aren't measuring the temp text during drawing, we need to set the transform manually
   *
   * @param {GraphObject} tb
   * @param {Rect} mainab
   * @param {Rect} ab
   * @param {Rect} mb
   * @param {Rect} nb
   */
  private _setGraduatedLabelTransform(tb: GraphObject, mainab: Rect, ab: Rect, mb: Rect, nb: Rect): void {
    const tt = tb._transform;
    tt.reset();
    tt.translate(ab.x + mainab.x, ab.y + mainab.y);
    // this is to move an object if its been rotated. For example, a square
    // rotated 45 degrees will need to be "pushed down" to fit where it once was
    tt.translate(-mb.x, -mb.y);
    tb._setTransformAngle(tt, nb.x, nb.y, nb.width, nb.height);
    tb.setInvalidTransform(false);
    tb._allTransforms.set(tb._transform);
    tb._documentScale = tb.scale;
    tb.setInvalidAllTransforms(false);
  }

  /**
   * @hidden @internal
   * Copy measure/drawing properties from the Graduated Panel's label TextBlock to the temporary textblock.
   * We don't want everything from the cloneProtected functions since they would overwrite some saved info.
   *
   * @param {GraphObject} elt
   * @param {TextBlock} tempText
   */
  public _copyGraduatedLabelProperties(elt: GraphObject, tempText: TextBlock): void {
    tempText._copyGraduatedLabelPropertiesFrom(elt);
  }


}


// import { Diagram } from './diagram';
// import { Geo } from './geo';
// import { Margin } from './margin';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { IContext } from './surfaceInterfaces';
// import { TextBlock } from './textBlock';
// import { Transform } from './transform';





/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Panel
RowColumnDefinition
*/

// import { List, Map, Set } from './collections';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { Model } from './model';
// import { Util } from './util';

/**
 * A Panel is a {@link GraphObject} that holds other GraphObjects as its elements.
 * A Panel is responsible for sizing and positioning its elements.
 * The elements of a panel are drawn in the order in which they appear in the {@link #elements} collection.
 *
 * The {@link Part} class inherits from Panel; Part in turn is the base class of {@link Node} and {@link Link}.
 *
 * Every Panel has a {@link #type} and establishes its own coordinate system. The type of a Panel
 * determines how it will size and arrange its elements:
 *   - {@link Panel.Position} is used to arrange elements based on their absolute positions within the Panel's local coordinate system.
 *   - {@link Panel.Vertical} and {@link Panel.Horizontal} are used to create linear "stacks" of elements.
 *   - {@link Panel.Auto} is used to size the main element to fit around other elements in the Panel.
 *   - {@link Panel.Spot} is used to arrange elements based on the {@link Spot} properties {@link GraphObject#alignment}
 *     and {@link GraphObject#alignmentFocus}, relative to a main element of the panel.
 *     Spot panels can align relative to other elements by using {@link Panel#alignmentFocusName}.
 *   - {@link Panel.Table} is used to arrange elements into rows and columns, typically employing the different
 *     elements' {@link GraphObject#row}, {@link GraphObject#rowSpan}, {@link GraphObject#column},
 *     and {@link GraphObject#columnSpan} properties. This Panel type also makes use of {@link RowColumnDefinition}.
 *   - {@link Panel.TableRow} and {@link Panel.TableColumn} can only be used immediately within a {@link Panel.Table} Panel
 *     to organize a collection of elements as a row or as a column in a table.
 *   - {@link Panel.Viewbox} is used to automatically resize a single element to fit inside the panel's available area.
 *   - {@link Panel.Grid} is not used to house typical elements, but is used only to draw regular patterns of lines.
 *     The elements must be {@link Shape}s used to describe the repeating lines.
 *   - {@link Panel.Link} is only used by {@link Link} parts and Link Adornments.
 *   - {@link Panel.Graduated} is used to draw regular tick marks and text along the main {@link Shape} element.
 *
 * Using <a href="../../intro/buildingObjects.html">GraphObject.make</a>, the second argument can be used to declare the Panel type:
 * ```js
 * // Either:
 * $(go.Panel, go.Panel.Horizontal, ...
 * // Or:
 * $(go.Panel, "Horizontal", ...
 *
 * // Full example:
 * $(go.Panel, "Horizontal",
 *   { width: 60, height: 60 }, // panel properties
 *   // elements in the panel:
 *   $(go.Shape, "Rectangle", { stroke: "lime" }),
 *   $(go.TextBlock, "Some Text")
 * ) // end of panel definition
 * ```
 *
 * <p class="boxread">
 * For an overview of most Panel types, please read the <a href="../../intro/panels.html">Introduction page on Panels</a>.
 *
 * {@link Panel.Vertical} and {@link Panel.Horizontal} panels are frequently used to position two or more GraphObjects
 * vertically above each other or horizontally next to each other.  Use the {@link GraphObject#alignment} or
 * {@link GraphObject#stretch} properties on the individual elements to control their position and size.
 * Set {@link #isOpposite} to true if you want the elements arranged from right-to-left in Horizontal Panels
 * or from bottom-to-top in Vertical Panels.
 *
 * {@link Panel.Spot} and {@link Panel.Auto} panels have a "main" element, signified by the Panel's first element
 * with {@link GraphObject#isPanelMain} set to true.  If there is no such element, it uses the first element as the "main" one.
 * Use the {@link GraphObject#alignment} property to position elements with respect to the main element.
 * Use the {@link GraphObject#alignmentFocus} property to further specify the position within Spot Panels.
 * "Spot" and "Auto" Panels should have two or more elements in them.
 *
 * In {@link Panel.Table} panels you will want to set the {@link GraphObject#row} and {@link GraphObject#column}
 * properties on each element.  The {@link GraphObject#alignment} and {@link GraphObject#stretch} properties are also
 * useful when an element's table cell is larger than that element.
 * <p class="boxrun">Please read the <a href="../../intro/tablePanels.html">Introduction page on Table Panels</a>
 * for more examples and explanation.
 *
 * {@link Panel.TableRow} and {@link Panel.TableColumn} panels can only be used as elements within a {@link Panel.Table} Panel.
 * They are typically only used in item templates, e.g. for automatically creating rows in a Table Panel based on model data
 * provided in an {@link #itemArray}. You will still need to specify properties on the individual elements within a TableRow
 * or TableColumn as if they were immediate elements of the containing Table panel.
 * <p class="boxrun">For an example that uses TableRow Panels, see <a href="../../samples/records.html">Records sample</a>.
 *
 * {@link Panel.Grid} panels are often used for the Diagram's {@link Diagram#grid}.
 * ```js
 * $(go.Diagram, "myDiagramDiv",
 *   { . . .
 *     grid: $(go.Panel, "Grid",
 *             { gridCellSize: new go.Size(40, 40) },
 *             $(go.Shape, "LineH", { stroke: "lightgray" }),
 *             $(go.Shape, "LineV", { stroke: "lightgray" })
 *           ),
 *     . . .
 *   });
 * ```
 * Or to get a green bar effect:
 * ```js
 * $(go.Diagram, "myDiagramDiv",
 *   { . . .
 *     grid: $(go.Panel, "Grid",
 *             { gridCellSize: new go.Size(100, 100) },
 *             $(go.Shape, "BarH", { fill: "lightgreen", height: 50 })
 *           ),
 *     . . .
 *   });
 * ```
 * But Grid Panels can also be stand alone objects:
 * ```js
 * $(go.Node, go.Panel.Grid,
 *   { gridCellSize: new go.Size(6, 6), width: 60, height: 60 },
 *   $(go.Shape, "LineH", { stroke: "gray" }),
 *   $(go.Shape, "LineV", { stroke: "gray" }))
 * ```
 *
 * A Grid Panel's elements do not participate in object picking.
 * <p class="boxrun">Please read the <a href="../../intro/grids.html">Introduction page on Grid Patterns</a>
 * for more examples and explanation.
 *
 * {@link Panel.Graduated} panels, like Spot and Auto Panels have a "main" element. The other elements within a Graduated Panel
 * are used to define ticks and labels to draw along the main shape's path.
 * ```js
 * $(go.Part, "Graduated",
 *   $(go.Shape, { geometryString: "M0 0 H400" }),
 *   $(go.Shape, { geometryString: "M0 0 V10" }),
 *   // offset to display below ticks
 *   $(go.TextBlock, { segmentOffset: new go.Point(0, 12) })
 * );
 * ```
 *
 * Only the main shape of a Graduated Panel participates in object picking, but a `background`
 * can be set if the entire panel needs to be pickable.
 * You cannot set or bind the {@link Panel#itemArray} of a Graduated Panel.
 * Events on the tick Shapes and TextBlock labels of a Graduated Panel will be ignored.
 * Graduated Panel TextBlock labels cannot be edited.
 *
 * Rotating the main shape will not rotate the ticks, just as rotating a Spot Panel's main element
 * won't rotate its children. Rotation should generally be done at the Panel level. Another similarity
 * to Spot Panels is that resizing of a Graduated Panel should generally be done on the main shape.
 * <p class="boxrun">Please read the <a href="../../intro/graduatedPanels.html">Introduction page on Graduated Panels</a>
 * for more examples and explanation.
 * <h3>Changing and accessing elements of a Panel</h3>
 *
 * You can change the collection of {@link #elements} by calling {@link #add}, {@link #insertAt}, {@link #remove}, or {@link #removeAt}.
 * You can get direct access to a particular element by calling {@link #elt}.
 *
 * You can search the visual tree of a Panel for GraphObjects that given a {@link GraphObject#name} using {@link #findObject}.
 * <h3>Panel Size and Appearance</h3>
 *
 * Panels typically compute their own size based on their elements and Panel {@link #type},
 * but can also be sized by setting {@link GraphObject#desiredSize}, {@link GraphObject#minSize}, and {@link GraphObject#maxSize}.
 * Setting an explicit size on a Panel may cause nested elements of that panel to size themselves differently,
 * especially in the cases of nested elements having a {@link GraphObject#stretch} value or {@link TextBlock}'s having word wrap.
 *
 * Panels have no visual components of their own unless a {@link GraphObject#background} or {@link GraphObject#areaBackground} is specified
 * or separators are specified either as defaults for the whole Table Panel or on individual {@link RowColumnDefinition}s.
 * Panels can specify {@link #padding}, to make the Panel larger including its background. Setting a padding
 * when the Panel is constrained in size will reduce the total area that it has to arrange its elements. Setting a {@link #margin}
 * will not do this -- instead the Panel will expand in size.
 *
 * In addition to the {@link GraphObject} properties on elements that are only used by certain types of panels,
 * several Panel properties only apply to specific Panel types.
 *   - Panels of {@link #type} {@link Panel.Table} use the {@link #rowCount}, {@link #rowSizing}, {@link #columnCount},
 *     {@link #columnSizing}, {@link #leftIndex}, {@link #topIndex}, and all of the "default" separator properties.
 *   - Panels of {@link #type} {@link Panel.TableRow} and {@link Panel.TableColumn} do not act like regular GraphObjects,
 *     instead they are only to be used immediately within a {@link Panel.Table}. They are pass-through containers
 *     that hold elements for their parent table, and ignore their own scale and angle.
 *   - Panels of {@link #type} {@link Panel.Grid} use the {@link #gridCellSize} and {@link #gridOrigin} properties.
 *   - Panels of {@link #type} {@link Panel.Viewbox} use the {@link #viewboxStretch} property.
 *   - Panels of {@link #type} {@link Panel.Graduated} use the {@link #graduatedMin}, {@link #graduatedMax},
 *     {@link #graduatedTickUnit}, and {@link #graduatedTickBase} properties.
 *
 * <p class="boxread">
 * For live examples of all Panel types, see the <a href="../../intro/panels.html">Introduction page on Panels.</a>
 * <h3>Data Binding</h3>
 *
 * Panels also provide fundamental support for data binding.
 * When a diagram gets a new model or when a diagram's model is augmented with additional data,
 * the diagram automatically creates a new Node or Link whose {@link #data} property refers to the corresponding
 * node data or link data object.
 * <p class="boxread">
 * For more discussion of data binding, please read the <a href="../../intro/usingModels.html">Introduction page on Models</a>
 * and <a href="../../intro/dataBinding.html">Data Binding</a>.
 *
 * Panels provide support for automatically creating elements within the Panel based on items in a JavaScript Array.
 * This is achieved by setting or binding the {@link #itemArray} property, which acts in a manner similar
 * to the {@link Model#nodeDataArray} property.
 * You can supply an {@link #itemTemplate}, which must itself be a simple Panel, which is copied to create the element
 * in this container Panel corresponding to an item in the itemArray.
 * This property is analogous to the {@link Diagram#nodeTemplate} property, although for the diagram the
 * template must be a {@link Node}, {@link Group}, or simple {@link Part}.
 *
 * Much like the {@link Diagram#nodeTemplateMap}, Panel's {@link #itemTemplateMap} supports having multiple templates,
 * so that the actual structure of the element created for a data item can be chosen dynamically.
 * Just as the {@link Model#nodeCategoryProperty} determines which template in the {@link Diagram#nodeTemplateMap}
 * is copied to create a {@link Node}, the {@link #itemCategoryProperty} names the data property whose value
 * chooses the Panel in the itemTemplateMap to copy for the item.
 *
 * When binding the {@link #itemArray} property, it is commonplace to set {@link Model#copiesArrays} and
 * {@link Model#copiesArrayObjects} properties to true, so that when a node is copied, the item Array and its
 * contents are copied, not shared.  Or more generally, to customize the model's copying processes, you
 * can supply a custom {@link Model#copyNodeDataFunction}.
 * <p class="boxread">
 * For more discussion and examples of item arrays, please read the <a href="../../intro/itemArrays.html">Introduction page on Item Arrays</a>.
 * @extends GraphObject
 * @unrestricted
 */
export class Panel extends GraphObject {
  /** @hidden @internal */ public static className = 'Panel';
  /** @hidden @internal */ private _padding: Margin;
  /** @hidden @internal */ private _defaultAlignment: Spot;
  /** @hidden @internal */ private _defaultStretch: EnumValue;
  /** @hidden @internal */ private _viewboxStretch: EnumValue;
  /** @hidden @internal */ private _gridCellSize: Size;
  /** @hidden @internal */ private _gridOrigin: Point;

  // ??? consider moving to PanelLayoutGraduated
  /** @hidden @internal */ private _graduatedMin: number;
  /** @hidden @internal */ private _graduatedMax: number;
  /** @hidden @internal */ private _graduatedTickUnit: number;
  /** @hidden @internal */ private _graduatedTickBase: number;
  /** @hidden @internal */ public _graduatedMarks: Array<Array<GraduatedMark>> | null;
  /** @hidden @internal */ public _graduatedBounds: Array<Rect> | null;
  /** @hidden @internal */ public _temporaryTextBlock: TextBlock | null;

  /** @hidden @internal */ private _bindings: List<Binding> | null;
  /** @hidden @internal */ private _itemIndex: number;
  /** @hidden @internal */ private _itemTemplateMap: Map<string, Panel>;
  /** @hidden @internal */ private _itemCategoryProperty: string | ((a: any) => string);
  /** @hidden @internal */ private _itemElementMap: Map<ObjectData, Panel> | null;
  /** @hidden @internal */ private _tempMain: GraphObject | null;
  /** @hidden @internal */ public _rowcol: Array<Array<Array<GraphObject>>> | null;
  /** @hidden @internal */ public _alignmentFocusName: string;
  /** @hidden @internal */ public _graduatedIntervals: Array<Array<number>> | null;
  /** @hidden @internal */ private _type: PanelLayout;
  /** @hidden @internal */ public _elements: List<GraphObject>;
  /** @hidden @internal */ public _data: any | null;
  /** @hidden @internal */ public _itemArray: Array<any> | null;
  /** @hidden @internal */ public _union: Rect;

  // Table panel only:
  /** @hidden @internal */ private _defaultRowSeparatorDashArray: Array<number> | null | undefined;
  /** @hidden @internal */ private _defaultColumnSeparatorDashArray: Array<number> | null | undefined;
  /** @hidden @internal */ public _rowDefs: Array<RowColumnDefinition> | undefined;
  /** @hidden @internal */ public _colDefs: Array<RowColumnDefinition> | undefined;
  /** @hidden @internal */ private _rowSizing: EnumValue | undefined;
  /** @hidden @internal */ private _columnSizing: EnumValue | undefined;
  /** @hidden @internal */ private _topIndex: number | undefined;
  /** @hidden @internal */ private _leftIndex: number | undefined;
  /** @hidden @internal */ private _defaultSeparatorPadding: Margin | undefined;
  /** @hidden @internal */ private _defaultRowSeparatorStrokeWidth: number | undefined;
  /** @hidden @internal */ private _defaultRowSeparatorStroke: BrushLike | undefined;
  /** @hidden @internal */ private _defaultColumnSeparatorStrokeWidth: number | undefined;
  /** @hidden @internal */ private _defaultColumnSeparatorStroke: BrushLike | undefined;

  /**
   * Constructs an empty Panel of the given {@link #type}. Default type is {@link Panel.Position}.
   * The panel type must be one of the enumerated values defined on the {@link Panel} class, including:
   * {@link Panel.Position}, {@link Panel.Vertical}, {@link Panel.Horizontal}, {@link Panel.Table},
   * {@link Panel.Auto}, {@link Panel.Spot}, or {@link Panel.Viewbox}.
   * @param {EnumValue=} type If not supplied, the default Panel type is {@link Panel.Position}.
   */
  constructor(type?: PanelLayout) {
    super();
    if (type === undefined) {
      this._type = Panel.Position;
    } else {
      Util.checkClass(type, PanelLayout, Panel, 'type');
      this._type = type;
    }

    if (this._type === null) {
      Util.throwError('Panel type not specified or PanelLayout not loaded: ' + type);
    }

    // this._flags initialized in GraphObject constructor

    this._elements = new List<GraphObject>();
    this._padding = Margin.ZeroMargin;
    if (this._type === Panel.Grid) this.isAtomic = true;

    this._defaultAlignment = Spot.Default;
    this._defaultStretch = GraphObject.Default;

    // ******** Table Panels: ********
    if (this._type === Panel.Table) {
      this.setTablePanelDefaults();
    }

    // ******** Viewbox Panels: ********

    this._viewboxStretch = GraphObject.Uniform;  // for Viewbox Panels

    // ******** Grid Panels: ********

    this._gridCellSize = Size.TenSize;
    this._gridOrigin = Point.Origin;

    // ******** Graduated Panels: ********

    this._graduatedMin = 0;
    this._graduatedMax = 100;
    this._graduatedTickUnit = 10;
    this._graduatedTickBase = 0;
    this._graduatedMarks = null;
    this._graduatedBounds = null;
    this._graduatedIntervals = null;

    // for data bound elements
    this._data = null;
    this._bindings = null;  // all Bindings to _data for everything in this panel's visual tree
    this._itemIndex = NaN;  // for an element representing an item in the parent Panel's itemArray

    // for Panels bound to an array of item data
    // an Array-like object: must use Util methods for access!
    this._itemArray = null;  // this Panel's children correspond 1-to-1 to this array's data items
    this._itemTemplateMap = null;  // holds Panel templates for items
    this._itemCategoryProperty = 'category';  // if the item is an object, the name of the property
    this._itemElementMap = null;

    // Holds the "natural" size of the Panel
    this._union = new Rect(NaN, NaN, NaN, NaN);

    // keeps track of main object in Auto or Spot Panel
    this._tempMain = null;

    // temporary array of arrays used in Table panels only
    this._rowcol = null;

    // temporary TextBlock used to allow graduated panels to draw many labels without allocating too much memory
    this._temporaryTextBlock = null;

    this._alignmentFocusName = '';

    // To complete the set of properties that Panels have,
    // call setTablePanelDefaults() and set various "_" properties used by Buttons.
  }

  /**
   * @hidden @internal
   * @this {Panel}
   */
  private setTablePanelDefaults(): void {
    this._defaultSeparatorPadding = Margin.ZeroMargin;
    this._defaultRowSeparatorStrokeWidth = 1;
    this._defaultRowSeparatorStroke = null;
    this._defaultRowSeparatorDashArray = null;
    this._defaultColumnSeparatorStrokeWidth = 1;
    this._defaultColumnSeparatorStroke = null;
    this._defaultColumnSeparatorDashArray = null;
    this._rowDefs = [];
    this._colDefs = [];
    this._rowSizing = RowColumnDefinition.ProportionalExtra;
    this._columnSizing = RowColumnDefinition.ProportionalExtra;
    this._topIndex = 0;
    this._leftIndex = 0;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._type = this._type;

    // constructor will allocate a new List(GraphObject) for panelcopy._elements;
    // copy() will fill this list with new copies of objects
    copy._padding = this._padding.copyFrozen();
    // copied in GraphObject._flags:
    // panelcopy._isAtomic = this._isAtomic;
    // panelcopy._isClipping = this._isClipping;
    // panelcopy._isEnabled = this._isEnabled;
    // panelcopy._hasNamedPorts = this._hasNamedPorts;
    // panelcopy._hasUplevelBindings = this._hasUplevelBindings;
    // panelcopy._isOpposite = this._isOpposite;

    copy._defaultAlignment = this._defaultAlignment.copyFrozen();
    copy._defaultStretch = this._defaultStretch;

    if (copy._type === Panel.Table) {
      copy._defaultSeparatorPadding = this._defaultSeparatorPadding.copyFrozen();
      copy._defaultRowSeparatorStrokeWidth = this._defaultRowSeparatorStrokeWidth;
      copy._defaultRowSeparatorStroke = this._defaultRowSeparatorStroke;
      copy._defaultRowSeparatorDashArray = this._defaultRowSeparatorDashArray;
      copy._defaultColumnSeparatorStrokeWidth = this._defaultColumnSeparatorStrokeWidth;
      copy._defaultColumnSeparatorStroke = this._defaultColumnSeparatorStroke;
      copy._defaultColumnSeparatorDashArray = this._defaultColumnSeparatorDashArray;

      // copy RowColumnDefinitions for all known rows and columns
      let defs: Array<RowColumnDefinition> = [];
      if (this._rowDefs.length > 0) {
        const rd = this._rowDefs;
        const l = rd.length;
        for (let i = 0; i < l; i++) {
          if (rd[i] === undefined) continue;
          const defcopy = rd[i].copy();
          defcopy.setPanel(copy);
          defs[i] = defcopy;
        }
      }
      copy._rowDefs = defs;

      defs = [];
      if (this._colDefs.length > 0) {
        const cd = this._colDefs;
        const l = cd.length;
        for (let i = 0; i < l; i++) {
          if (cd[i] === undefined) continue;
          const defcopy = cd[i].copy();
          defcopy.setPanel(copy);
          defs[i] = defcopy;
        }
      }
      copy._colDefs = defs;

      copy._rowSizing = this._rowSizing;
      copy._columnSizing = this._columnSizing;
      copy._topIndex = this._topIndex;
      copy._leftIndex = this._leftIndex;
    }

    copy._viewboxStretch = this._viewboxStretch;

    copy._gridCellSize = this._gridCellSize.copyFrozen();
    copy._gridOrigin = this._gridOrigin.copyFrozen();

    copy._graduatedMin = this._graduatedMin;
    copy._graduatedMax = this._graduatedMax;
    copy._graduatedTickUnit = this._graduatedTickUnit;
    copy._graduatedTickBase = this._graduatedTickBase;
    copy._graduatedMarks = this._graduatedMarks;
    copy._graduatedIntervals = this._graduatedIntervals;

    // for data bound elements
    copy._data = this._data;  // share the data reference, if any
    copy._bindings = this._bindings;  // shared
    copy._itemIndex = this._itemIndex;

    // for Panels bound to an array of item data
    copy._itemArray = this._itemArray;  // share the array, if any
    copy._itemTemplateMap = this._itemTemplateMap;
    copy._itemCategoryProperty = this._itemCategoryProperty;
    // don't copy this._itemElementMap -- needs to be built for copied elements

    copy._union.assign(this._union);

    // _tempMain not copied, always re-computed before being accessed

    copy._alignmentFocusName = this._alignmentFocusName;

    if (this._rowcol !== null) copy._rowcol = this._rowcol;
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can displace this object in the visual tree.
   * @this {Panel}
   * @param {GraphObject} copy  This might be a copy of this object,
   * but more likely is a separately created object of the same class with completely different internals.
   */
  protected displaceProtected(copy: this): void {
    super.displaceProtected(copy);
    copy._elements = this._elements;  // transfer this Panel's list of elements
    const children = copy._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {  // and setup new back-pointers to that Panel
      const child = children[i];
      child._panel = copy;
    }
    copy._tempMain = null;
  }

  /**
   * Creates a deep copy of this Panel and returns it.
   * @this {Panel}
   * @return {Panel}
   */
  public copy(): this {
    const copy = super.copy() as this;
    if (copy !== null) {
      const children = this._elements._arr;
      const l = children.length;
      for (let i = 0; i < l; i++) {
        const child = children[i];
        const c = child.copy();
        copy.copyInsert(c); // sets copy as the panel
      }
      return copy;
    } else {
      return null;
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @return {string}
   */
  public toString(): string {
    return 'Panel(' + this.type + ')#' + Set.hashId(this);
  }

  /**
   * Gets or sets the type of the Panel, which controls how the Panel's elements are measured and arranged.
   * The only accepted values are listed as constant properties of Panel, including:
   *   - {@link Panel.Position}
   *   - {@link Panel.Vertical}
   *   - {@link Panel.Horizontal}
   *   - {@link Panel.Auto}
   *   - {@link Panel.Spot}
   *   - {@link Panel.Table}
   *   - {@link Panel.Viewbox}
   *   - {@link Panel.Link} (see also {@link Link}s, which are all Panels of type Link)
   *   - {@link Panel.TableRow}
   *   - {@link Panel.TableColumn}
   *   - {@link Panel.Grid}
   *   - {@link Panel.Graduated}
   *
   * The default value is {@link Panel.Position}.
   *
   * You must not modify this property once the Panel has been measured or the Panel has been added to another Panel.
   */
  /** @export */ get type(): PanelLayout { return this._type; }
  /** @export */ set type(value: PanelLayout) {
    const old = this._type;
    if (old !== value) {
      // if (Debug) EnumValue.checkEnumValue(value, Panel, Panel, 'type');
      // if (old === Panel.TableRow || old === Panel.TableColumn) Util.throwError('Cannot change Panel.type when it is already a TableRow or a TableColumn: ' + value);
      this._type = value;
      if (this._type === Panel.Grid) this.isAtomic = true;
      else if (this._type === Panel.Table) this.setTablePanelDefaults();
      this.invalidateMeasure();
      this.raiseChanged('type', old, value);
    }
  }

  /**
   * This read-only property returns an iterator over the collection of the {@link GraphObject}s that this panel manages.
   *
   * You can change the collection by calling {@link #add}, {@link #insertAt}, {@link #remove}, or {@link #removeAt}.
   *
   * You can also get direct access to individual elements by calling {@link #elt}.
   */
  /** @export */ get elements(): Iterator<GraphObject> { return this._elements.iterator; }

  /**
   * @hidden @internal
   * This read-only property returns the natural bounds of this Panel.
   */
  /** @export */ get naturalBounds(): Rect { return this._naturalBounds; }

  /**
   * Gets or sets the space between this Panel's border and its content.
   * Unlike {@link GraphObject#margin}, padding expands the area inside of the Panel's border.
   * If this Panel's size is unconstrained, this will increase the size of the panel.
   * If this Panel's size is constrained, this will decrease the total area
   * for the Panel elements to arrange themselves.
   *
   * Unlike margin, increases in size due to padding are visually covered by the
   * {@link GraphObject#background} and {@link GraphObject#areaBackground}.
   *
   * Padding cannot contain negative numbers.
   * The default value is a {@link Margin} of zero.
   * @see {@link GraphObject#margin}
   */
  /** @export */ get padding(): MarginLike { return this._padding; }
  /** @export */ set padding(value: MarginLike) {
    if (typeof value === 'number') {
      if (value < 0) Util.throwRangeError(value, '>= 0', Panel, 'padding');
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Panel, 'padding');
      if (value.left < 0) Util.throwRangeError(value.left, '>= 0', Panel, 'padding:value.left');
      if (value.right < 0) Util.throwRangeError(value.right, '>= 0', Panel, 'padding:value.right');
      if (value.top < 0) Util.throwRangeError(value.top, '>= 0', Panel, 'padding:value.top');
      if (value.bottom < 0) Util.throwRangeError(value.bottom, '>= 0', Panel, 'padding:value.bottom');
    }
    const old = this._padding;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._padding = value;
      this.invalidateMeasure();
      this.raiseChanged('padding', old, value);
    }
  }

  /**
   * Gets or sets the default alignment spot of this Panel, used as
   * the alignment for an element when its {@link GraphObject#alignment} value is {@link Spot.Default}.
   * The default value is {@link Spot.Default}, which is interpreted by the Panel in whatever manner seems reasonable,
   * depending on the Panel type.
   */
  /** @export */ get defaultAlignment(): Spot { return this._defaultAlignment; }
  /** @export */ set defaultAlignment(value: Spot) {
    const old = this._defaultAlignment;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, Panel, 'defaultAlignment');
      value = value.copyFrozen();
      this._defaultAlignment = value;
      this.invalidateMeasure();
      this.raiseChanged('defaultAlignment', old, value);
    }
  }

  /**
   * Gets or sets the default stretch of this Panel, used as
   * the stretch for an element when its {@link GraphObject#stretch} value is {@link GraphObject.Default}.
   * The default value is {@link GraphObject.Default}, which typically resolves to {@link GraphObject.None}.
   */
  /** @export */ get defaultStretch(): EnumValue { return this._defaultStretch; }
  /** @export */ set defaultStretch(value: EnumValue) {
    const old = this._defaultStretch;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, Panel, 'defaultStretch');
      this._defaultStretch = value;
      this.invalidateMeasure();
      this.raiseChanged('defaultStretch', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the additional padding for rows and columns.
   * Padding is applied both before and after a row or column's contents.
   * @see {@link RowColumnDefinition#separatorPadding}
   * @see {@link #defaultColumnSeparatorStrokeWidth}
   * @see {@link #defaultRowSeparatorStrokeWidth}
   * @since 1.2
   */
  /** @export */ get defaultSeparatorPadding(): MarginLike { return this._defaultSeparatorPadding === undefined ? Margin.ZeroMargin : this._defaultSeparatorPadding; }
  /** @export */ set defaultSeparatorPadding(value: MarginLike) {
    if (this._defaultSeparatorPadding === undefined) return;
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      if (Debug) Util.checkClass(value, Margin, Panel, 'defaultSeparatorPadding');
    }
    const old = this._defaultSeparatorPadding;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._defaultSeparatorPadding = value;
      this.invalidateMeasure();
      this.raiseChanged('defaultSeparatorPadding', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default stroke (color) for rows
   * provided a given row has a nonzero {@link RowColumnDefinition#separatorStrokeWidth}.
   * {@link RowColumnDefinition#separatorStroke} can override this default value.
   * The default value is null -- no line is drawn.
   * @TablePanel
   * @see {@link #defaultRowSeparatorStrokeWidth}
   * @see {@link #defaultRowSeparatorDashArray}
   * @see {@link RowColumnDefinition#separatorStroke}
   * @see {@link #defaultColumnSeparatorStroke}
   * @since 1.2
   */
  /** @export */ get defaultRowSeparatorStroke(): BrushLike { return this._defaultRowSeparatorStroke === undefined ? null : this._defaultRowSeparatorStroke; }
  /** @export */ set defaultRowSeparatorStroke(value: BrushLike) {
    const old = this._defaultRowSeparatorStroke;
    if (old !== value) {
      if (value === null || typeof value === 'string' || value instanceof Brush) {
        if (value instanceof Brush) value.freeze();
        this._defaultRowSeparatorStroke = value;
        this.invalidateDraw();
        this.raiseChanged('defaultRowSeparatorStroke', old, value);
      }
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default stroke width for a row's separator.
   * {@link RowColumnDefinition#separatorStrokeWidth} can override this default value.
   * The default value is 1.  Any new value must be a real, non-negative number.
   * @see {@link RowColumnDefinition#separatorStrokeWidth}
   * @see {@link #defaultColumnSeparatorStrokeWidth}
   * @since 1.2
   */
  /** @export */ get defaultRowSeparatorStrokeWidth(): number { return this._defaultRowSeparatorStrokeWidth === undefined ? 1 : this._defaultRowSeparatorStrokeWidth; }
  /** @export */ set defaultRowSeparatorStrokeWidth(value: number) {
    if (this._defaultRowSeparatorStrokeWidth === undefined) return;
    const old = this._defaultRowSeparatorStrokeWidth;
    if (old !== value && isFinite(value) && value >= 0) {
      this._defaultRowSeparatorStrokeWidth = value;
      this.invalidateMeasure();
      this.raiseChanged('defaultRowSeparatorStrokeWidth', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default dash array for a row's separator.
   * {@link RowColumnDefinition#separatorDashArray} can override this default value.
   *
   * Must be an array of positive numbers and zeroes,
   * or else null to indicate a solid line.
   *
   * For example, the array [5, 10] would create dashes of 5 pixels and spaces of 10 pixels.
   *
   * Setting an array with all zeroes will set the value to null.
   *
   * Default is null.
   * @see {@link RowColumnDefinition#separatorDashArray}
   * @see {@link #defaultColumnSeparatorDashArray}
   * @since 1.2
   */
  /** @export */ get defaultRowSeparatorDashArray(): Array<number> | null { return this._defaultRowSeparatorDashArray === undefined ? null : this._defaultRowSeparatorDashArray; }
  /** @export */ set defaultRowSeparatorDashArray(value: Array<number> | null) {
    if (this._defaultRowSeparatorDashArray === undefined) return;
    const old = this._defaultRowSeparatorDashArray;
    if (old !== value) {
      // make sure its an array of numbers?
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', Panel, 'defaultRowSeparatorDashArray:value');
      if (value !== null) {
        const l = value.length;
        let total = 0;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (typeof n !== 'number' || !(n >= 0) || !isFinite(n)) {
            Util.throwError('defaultRowSeparatorDashArray value ' + n + ' at index ' + i + ' must be a positive number or zero.');
          }
          total += n;
        }
        if (total === 0) {
          if (old === null) return; // no-op
          value = null; // never let defaultRowSeparatorDashArray be all zeroes, ie [0, 0]
        }
      }

      this._defaultRowSeparatorDashArray = value;
      this.invalidateDraw();
      this.raiseChanged('defaultRowSeparatorDashArray', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default stroke (color) for columns
   * provided a given column has a nonzero {@link RowColumnDefinition#separatorStrokeWidth}.
   * {@link RowColumnDefinition#separatorDashArray} can override this default value.
   * The default value is null -- no line is drawn.
   * @see {@link #defaultColumnSeparatorStrokeWidth}
   * @see {@link #defaultColumnSeparatorDashArray}
   * @see {@link RowColumnDefinition#separatorStroke}
   * @see {@link #defaultRowSeparatorStroke}
   * @since 1.2
   */
  /** @export */ get defaultColumnSeparatorStroke(): BrushLike { return this._defaultColumnSeparatorStroke === undefined ? null : this._defaultColumnSeparatorStroke; }
  /** @export */ set defaultColumnSeparatorStroke(value: BrushLike) {
    if (this._defaultColumnSeparatorStroke === undefined) return;
    const old = this._defaultColumnSeparatorStroke;
    if (old !== value) {
      if (value === null || typeof value === 'string' || value instanceof Brush) {
        if (value instanceof Brush) value.freeze();
        this._defaultColumnSeparatorStroke = value;
        this.invalidateDraw();
        this.raiseChanged('defaultColumnSeparatorStroke', old, value);
      }
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default stroke width for a column's separator.
   * {@link RowColumnDefinition#separatorStrokeWidth} can override this default value.
   * The default value is 1.  Any new value must be a real, non-negative number.
   * @see {@link RowColumnDefinition#separatorStrokeWidth}
   * @see {@link #defaultRowSeparatorStrokeWidth}
   * @since 1.2
   */
  /** @export */ get defaultColumnSeparatorStrokeWidth(): number { return this._defaultColumnSeparatorStrokeWidth === undefined ? 1 : this._defaultColumnSeparatorStrokeWidth; }
  /** @export */ set defaultColumnSeparatorStrokeWidth(value: number) {
    if (this._defaultColumnSeparatorStrokeWidth === undefined) return;
    const old = this._defaultColumnSeparatorStrokeWidth;
    if (old !== value && isFinite(value) && value >= 0) {
      this._defaultColumnSeparatorStrokeWidth = value;
      this.invalidateMeasure();
      this.raiseChanged('defaultColumnSeparatorStrokeWidth', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the default dash array for a column's separator.
   * {@link RowColumnDefinition#separatorStrokeWidth} can override this default value.
   *
   * Must be an array of positive numbers and zeroes,
   * or else null to indicate a solid line.
   *
   * For example, the array [5, 10] would create dashes of 5 pixels and spaces of 10 pixels.
   *
   * Setting an array with all zeroes will set the value to null.
   *
   * Default is null.
   * @see {@link RowColumnDefinition#separatorDashArray}
   * @see {@link #defaultRowSeparatorDashArray}
   * @since 1.2
   */
  /** @export */ get defaultColumnSeparatorDashArray(): Array<number> | null { return this._defaultColumnSeparatorDashArray === undefined ? null : this._defaultColumnSeparatorDashArray; }
  /** @export */ set defaultColumnSeparatorDashArray(value: Array<number> | null) {
    if (this._defaultColumnSeparatorDashArray === undefined) return;
    const old = this._defaultColumnSeparatorDashArray;
    if (old !== value) {
      // make sure its an array of numbers?
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', Panel, 'defaultColumnSeparatorDashArray:value');
      if (value !== null) {
        const l = value.length;
        let total = 0;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (typeof n !== 'number' || !(n >= 0) || !isFinite(n)) {
            Util.throwError('defaultColumnSeparatorDashArray value ' + n + ' at index ' + i + ' must be a positive number or zero.');
          }
          total += n;
        }
        if (total === 0) {
          if (old === null) return; // no-op
          value = null; // never let defaultColumnSeparatorDashArray be all zeroes, ie [0, 0]
        }
      }

      this._defaultColumnSeparatorDashArray = value;
      this.invalidateDraw();
      this.raiseChanged('defaultColumnSeparatorDashArray', old, value);
    }
  }

  /**
   * For {@link Panel.Viewbox|Viewbox} Panels: Gets or sets how the panel will resize its content.
   *
   * Possible values are {@link GraphObject.Uniform} and {@link GraphObject.UniformToFill}.
   * The default is {@link GraphObject.Uniform}.
   */
  /** @export */ get viewboxStretch(): EnumValue { return this._viewboxStretch; }
  /** @export */ set viewboxStretch(value: EnumValue) {
    const old = this._viewboxStretch;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, Panel, 'viewboxStretch');
      this._viewboxStretch = value;
      this.invalidateMeasure();
      this.raiseChanged('viewboxStretch', old, value);
    }
  }

  /**
   * For {@link Panel.Grid|Grid} Panels: Gets or sets the distance between lines.
   * The units are in local coordinates.
   * The default is 10x10.  Any new width or height must be a positive real number.
   */
  /** @export */ get gridCellSize(): Size { return this._gridCellSize; }
  /** @export */ set gridCellSize(value: Size) {
    const old = this._gridCellSize;
    if (!old.equals(value)) {
      Util.checkClass(value, Size, Panel, 'gridCellSize');
      if (!value.isReal() || value.width === 0 || value.height === 0) Util.throwError('Invalid Panel.gridCellSize: ' + value);
      this._gridCellSize = value.copyFrozen();
      const diagram = this.diagram;
      if (diagram !== null && this === diagram.grid) {
        diagram._updateBackgroundGrid();
      }
      this.invalidateDraw();
      this.raiseChanged('gridCellSize', old, value);
    }
  }

  /**
   * For {@link Panel.Grid|Grid} Panels: Gets or sets an origin point for the grid cells.
   * The units are in local coordinates.
   * The default is (0,0).  Any new value must use real numbers.
   */
  /** @export */ get gridOrigin(): Point { return this._gridOrigin; }
  /** @export */ set gridOrigin(value: Point) {
    const old = this._gridOrigin;
    if (!old.equals(value)) {
      Util.checkClass(value, Point, Panel, 'gridOrigin');
      if (!value.isReal()) Util.throwError('Invalid Panel.gridOrigin: ' + value);
      this._gridOrigin = value.copyFrozen();
      const diagram = this.diagram;
      if (diagram !== null && this === diagram.grid) {
        diagram._updateBackgroundGrid();
      }
      this.invalidateDraw();
      this.raiseChanged('gridOrigin', old, value);
    }
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Gets or sets the minimum value represented.
   * Must be less than {@link #graduatedMax}. The default is 0.
   * @since 1.7
   */
  /** @export */ get graduatedMin(): number { return this._graduatedMin; }
  /** @export */ set graduatedMin(value: number) {
    Util.checkRealNumber(value, Panel, 'graduatedMin');
    const old = this._graduatedMin;
    if (old !== value) {
      this._graduatedMin = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedMin', old, value);
      // changing the min also changes the range
      if (this.getBindsTwoWay()) {
        const part = this.part;
        if (part !== null) {
          this.updateSourceBindings(part, 'graduatedRange');
        }
      }
    }
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Gets or sets the maximum value represented.
   * Must be greater than {@link #graduatedMin}. The default is 100.
   * @since 1.7
   */
  /** @export */ get graduatedMax(): number { return this._graduatedMax; }
  /** @export */ set graduatedMax(value: number) {
    Util.checkRealNumber(value, Panel, 'graduatedMax');
    const old = this._graduatedMax;
    if (old !== value) {
      this._graduatedMax = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedMax', old, value);
      // changing the max also changes the range
      if (this.getBindsTwoWay()) {
        const part = this.part;
        if (part !== null) {
          this.updateSourceBindings(part, 'graduatedRange');
        }
      }
    }
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: This read-only property returns the range of values represented by the Panel.
   *
   * For example, a {@link #graduatedMin} of 25 and {@link #graduatedMax} of 75 would return 50.
   * @since 1.7
   */
  /** @export */ get graduatedRange(): number {
    return this.graduatedMax - this.graduatedMin;
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Gets or sets the difference between two consecutive values marked by ticks.
   * Must be positive. The default is 10.
   * @since 1.7
   */
  /** @export */ get graduatedTickUnit(): number { return this._graduatedTickUnit; }
  /** @export */ set graduatedTickUnit(value: number) {
    Util.checkRealNumber(value, Panel, 'graduatedTickUnit');
    const old = this._graduatedTickUnit;
    if (old !== value && value > 0) {
      this._graduatedTickUnit = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedTickUnit', old, value);
    }
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Gets or sets the base value which is marked with a tick.
   * The default is 0.
   * @since 1.7
   */
  /** @export */ get graduatedTickBase(): number { return this._graduatedTickBase; }
  /** @export */ set graduatedTickBase(value: number) {
    Util.checkRealNumber(value, Panel, 'graduatedTickBase');
    const old = this._graduatedTickBase;
    if (old !== value) {
      this._graduatedTickBase = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedTickBase', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Part} part
   */
  public _notifyPorts(part: Part): void {
    super._notifyPorts(part);
    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const child = children[i];
      child._notifyPorts(part);
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {IContext} context
   * @param {Diagram} diagram
   */
  public drawProtected(context: IContext, diagram: Diagram): void {
    if (this._type === Panel.Grid) {
      this._type.draw(this, context, diagram); // Draw grid
      return;
    }
    if (this._type === Panel.Graduated) {
      this._type.draw(this, context, diagram); // Draw graduated
      return;
    }
    if (this._type === Panel.Table) {
      this.drawProtectedSpacers(context, diagram); // todo, move to Table
    }

    const clip = this.isClipping;
    if (clip) {
      context.save();
      if (Debug && this.type !== Panel.Spot) Util.trace('Warning: Panel.isClipping set on non-Spot Panel: ' + this.toString());
    }
    const main = this.findMainElement();
    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const child = children[i];

      if (clip && child === main) {
        context.clipInsteadOfFill = true;
      }

      child.draw(context, diagram);

      if (clip && child === main) {
        context.clipInsteadOfFill = false;
      }

    }

    if (clip) {
      context.restore();
      context.clearContextCache(true); // panel element may have changed font
    }

    /*
    if (Debug && Debug.drawLinkpointsEnabled && this instanceof Link) {
      Debug.drawLinkpoints(context, diagram, this);
    }
    */
  }

  /**
   * @hidden @internal
   * Table Panel spacers
   * @this {Panel}
   * @param {IContext} context
   * @param {Diagram} diagram
   */
  protected drawProtectedSpacers(context: IContext, diagram: Diagram): void {
    context.lineCap = 'butt';
    this.drawProtectedBackgrounds(context, diagram, true, this._rowDefs, true);
    this.drawProtectedBackgrounds(context, diagram, false, this._colDefs, true);

    this.drawProtectedSpacers2(context, diagram, true, this._rowDefs);
    this.drawProtectedSpacers2(context, diagram, false, this._colDefs);

    this.drawProtectedBackgrounds(context, diagram, true, this._rowDefs, false);
    this.drawProtectedBackgrounds(context, diagram, false, this._colDefs, false);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {boolean} isRow
   * @param {Array.<RowColumnDefinition>} rowColDefs
   */
  protected drawProtectedSpacers2(context: IContext, diagram: Diagram, isRow: boolean, rowColDefs: Array<RowColumnDefinition>): void {
    const l = rowColDefs.length;
    const ab = this.actualBounds;
    const nb = this.naturalBounds;
    let first = true;
    for (let i = 0; i < l; i++) {
      const def = rowColDefs[i];
      if (def === undefined) continue;
      if (first) {
        // never draw the first row or column spacer
        first = false;
        continue;
      }
      if (def.actual === 0) continue;
      if (isRow) {
        if (def.position > ab.height) continue;
      } else {
        if (def.position > ab.width) continue;
      }
      let spacing = def.separatorStrokeWidth;
      if (isNaN(spacing)) spacing = isRow ? this._defaultRowSeparatorStrokeWidth : this._defaultColumnSeparatorStrokeWidth;
      let stroke = def.separatorStroke;
      if (stroke === null) stroke = isRow ? this._defaultRowSeparatorStroke : this._defaultColumnSeparatorStroke;
      if (spacing === 0 || stroke === null) continue;
      this.setBrush(context, stroke, false, false, nb, ab);

      let dash = false;
      let separatorDashArray = def.separatorDashArray;
      if (separatorDashArray === null) separatorDashArray = isRow ? this._defaultRowSeparatorDashArray : this._defaultColumnSeparatorDashArray;
      if (separatorDashArray !== null) {
        dash = true;
        context.enableDash(separatorDashArray, 0 /* there is no strokeDashOffset on Panel/RowColDef */);
      }

      context.beginPath();

      const bot = def.position + spacing; // "bottom" of the line to be drawn
      if (isRow) {
        if (bot > ab.height) spacing -= bot - ab.height;
      } else {
        if (bot > ab.width) spacing -= bot - ab.width;
      }
      let p = def.position + (spacing / 2); // middle of the line
      context.lineWidth = spacing;
      const padding = this._padding;

      if (isRow) {
        p += padding.top;
        const start = padding.left;
        const end = ab.width - padding.right;
        context.moveTo(start, p);
        context.lineTo(end, p);

      } else {
        p += padding.left;
        const start = padding.top;
        const end = ab.height - padding.bottom;
        context.moveTo(p, start);
        context.lineTo(p, end);
      }
      context.stroke();

      // Set dash back when done
      if (dash) {
        context.disableDash();
      }

    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {boolean} isRow
   * @param {Array.<RowColumnDefinition>} rowColDefs
   * @param {boolean} behind
   */
  protected drawProtectedBackgrounds(context: IContext, diagram: Diagram, isRow: boolean, rowColDefs: Array<RowColumnDefinition>, behind: boolean): void {
    const l = rowColDefs.length;
    const ab = this.actualBounds;
    const nb = this.naturalBounds;
    for (let i = 0; i < l; i++) {
      const def = rowColDefs[i];
      if (def === undefined) continue;
      if (def.background === null) continue;
      if (def.coversSeparators === behind) continue;
      if (def.actual === 0) continue;

      const panelActual = isRow ? ab.height : ab.width;
      if (def.position > panelActual) continue;
      let spacing = def.computeEffectiveSpacing();
      let sw = def.separatorStrokeWidth;
      if (isNaN(sw)) sw = isRow ? this._defaultRowSeparatorStrokeWidth : this._defaultColumnSeparatorStrokeWidth;
      let stroke = def.separatorStroke;
      if (stroke === null) stroke = isRow ? this._defaultRowSeparatorStroke : this._defaultColumnSeparatorStroke;
      if (stroke === null) sw = 0;

      spacing -= sw;
      const pos = def.position + sw;
      let amt = spacing + def.actual;
      if (pos + amt > panelActual) {
        amt = panelActual - pos;
      }
      if (amt <= 0) continue;
      const padding = this._padding;
      this.setBrush(context, def.background, true, false, nb, ab);
      if (isRow) {
        context.fillRect(padding.left, pos + padding.top, ab.width - (padding.left + padding.right), amt);
      } else {
        context.fillRect(pos + padding.left, padding.top, amt, ab.height - (padding.top + padding.bottom));
      }
    }
  }

  /**
   * @hidden @internal
   * Set the size arg to the size at which the grid pattern repeats.
   * @this {Panel}
   * @param {Size} size
   */
  public computeTileSize(size: Size): void {
    let hlcm = 1;
    let vlcm = 1;
    const elems = this._elements._arr;
    const len = elems.length;
    for (let c = 0; c < len; c++) {
      const shape = elems[c] as Shape;
      const interval = shape.interval;
      if (interval < 2) continue;
      if (this._isVertical(shape.figure)) {
        vlcm = (vlcm * interval) / Geo.gcd(vlcm, interval);
      } else {
        hlcm = (hlcm * interval) / Geo.gcd(hlcm, interval);
      }
    }
    const csz = this.gridCellSize;
    size.assignTo(vlcm * csz.width, hlcm * csz.height);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {string} fig a Shape.figure
   * @return {boolean}
   */
  public _isVertical(fig: string): boolean {
    return (fig === 'LineV' || fig === 'BarV');
  }

  // getNearestIntersectionPoint without using Points or calling transform methods
  /**
   * @hidden @internal
   * Find the closest point in this object to the given point p1 that is on the
   * line p1-p2.
   * @this {GraphObject}
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {Point} result (reference) the point of this object that is closest to p1 and that is on the infinite line from p1 to p2.
   * @return {boolean} true if the infinite line does intersect with this object; false otherwise.
   */
  public getNearestIntersectionPoint2(p1x: number, p1y: number, p2x: number, p2y: number, result: Point): boolean {
    const trueTransform = this.trueTransform();
    const tt = this.transform;
    const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
    let m0 = tt.m22 * d;
    let m1 = -tt.m12 * d;
    let m2 = -tt.m21 * d;
    let m3 = tt.m11 * d;
    let m4 = d * (tt.m21 * tt.dy - tt.m22 * tt.dx);
    let m5 = d * (tt.m12 * tt.dx - tt.m11 * tt.dy);

    if (this.areaBackground !== null) {
      // find intersection with transformed bounds
      const ab = this.actualBounds;
      return Geo.getNearestIntersectionPoint(ab.left, ab.top, ab.right, ab.bottom, p1x, p1y, p2x, p2y, result);
    } else if (this.background !== null) {
      // Inverted transform point
      const p12x = p1x * m0 + p1y * m2 + m4;
      const p12y = p1x * m1 + p1y * m3 + m5;

      const p22x = p2x * m0 + p2y * m2 + m4;
      const p22y = p2x * m1 + p2y * m3 + m5;
      result.assignTo(0, 0);
      const nb = this.naturalBounds;
      const found = Geo.getNearestIntersectionPoint(0, 0, nb.width, nb.height, p12x, p12y, p22x, p22y, result);
      result.transform(tt);
      return found;
    } else { // recurse on the children

      // If this = not a true transform, ignore it here
      if (!trueTransform) {
        m0 = 1;
        m1 = 0;
        m2 = 0;
        m3 = 1;
        m4 = 0;
        m5 = 0;
      }

      // Inverted transform point
      const q1x = p1x * m0 + p1y * m2 + m4;
      const q1y = p1x * m1 + p1y * m3 + m5;

      const q2x = p2x * m0 + p2y * m2 + m4;
      const q2y = p2x * m1 + p2y * m3 + m5;

      result.assignTo(q2x, q2y);
      let closestdist = (q2x - q1x) * (q2x - q1x) + (q2y - q1y) * (q2y - q1y);
      let found = false;
      const elems = this._elements._arr;
      const len = elems.length;
      const point = Point.alloc();
      // New for 1.7, intersect with the main element if isClipping
      let clipChild = null;
      let clipDist = Infinity;
      let point2 = null;
      if (this.isClipping) {
        point2 = Point.alloc();
        clipChild = this.findMainElement();
        found = clipChild.getNearestIntersectionPoint2(q1x, q1y, q2x, q2y, point2);
        if (found) {
          clipDist = (q1x - point2.x) * (q1x - point2.x) + (q1y - point2.y) * (q1y - point2.y);
        }
      }
      for (let i = 0; i < len; i++) {
        // This will return potential child points even if those points are clipped,
        // such as in a panel with a desired size that is too small to show all its contents
        // We may want to exclude such points in the future
        const child = elems[i];
        if (!child.visible) continue;
        if (child === clipChild) continue; // only gets set if isClipping
        if (!child.getNearestIntersectionPoint2(q1x, q1y, q2x, q2y, point)) continue;
        found = true;
        const dist = (q1x - point.x) * (q1x - point.x) + (q1y - point.y) * (q1y - point.y);
        if (dist < closestdist) {
          closestdist = dist;
          result.set(point);
        }
      }

      if (this.isClipping) {
        if (clipDist > closestdist) {
          result.set(point2);
        }
        Point.free(point2);
      }

      Point.free(point);
      if (trueTransform) result.transform(tt); // convert back to container coords.
      return found;
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {boolean=} parent flag, meaning the parent invalidated a child. Stops the child from trying to invalidate the parent.
   */
  public invalidateMeasure(parent?: boolean): void {
    // Added in 2.1, may stop re-measurement issues in panel designs that require arranging children during measure phase:
    if (this.getInvalidMeasure() === true) return;
    super.invalidateMeasure(parent);
    // Invalidate Children
    let mainchild = null;
    if (this._type === Panel.Auto || this._type === Panel.Link) {
      mainchild = this.findMainElement();
    }
    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const child = children[i];
      // must invalidate main objects:
      if (child === mainchild || child.isPanelMain) child.invalidateMeasure(true);

      // A real desired size means that the child does not need to be invalidated,
      // since its size cannot change
      if (child.desiredSize.isReal()) continue;
      // No stretch means that a parent's change of size will not affect a child,
      // so it does not need to be invalidated
      const stretch = child.getEffectiveStretch(false);
      if (!(child instanceof Placeholder) &&
        !(child instanceof Panel) &&
        !(child instanceof TextBlock) &&
        stretch === GraphObject.None) continue;

      // ??? This breaks resizing an Auto panel with a textblock with wrap: WrapFit. Fix + test before uncommenting
      /*
      // new in 1.7: If there's no stretch, don't invalidate a child's measure, invalidate just its arrange instead
      if (stretch === GraphObject.None) {
        // don't invalidate the measure, just its arrange
        child.invalidateArrange(true);
        continue;
      }
      */

      child.invalidateMeasure(true);
    }
  }

  /**
   * @hidden @internal
   * Special invalidateMeasure for invalidating panels and their children only
   * This does not notify the diagram nor any parent panels
   * @this {Panel}
   */
  public invalidateMeasureChildren(): void {
    if (this.getInvalidMeasure() === false) {
      this.setInvalidMeasure(true);
      this.setInvalidArrange(true);
      // Invalidate Children
      const children = this._elements._arr;
      const l = children.length;
      for (let i = 0; i < l; i++) {
        const child = children[i];
        // In the future we should only do this on specific children ??
        child.invalidateMeasureChildren();
      }
    }
  }

  /**
   * @hidden @internal
   * Invalidates an Object's transform and then recurses through the visual tree to invalidate the allTransforms for each object in collection.
   * @this {Panel}
   */
  public _invalidateTransform(): void {
    if (this.getInvalidTransform() === false) { // add || invalidAllTransforms?
      this.setInvalidTransform(true);
      this.setInvalidAllTransforms(true);
      const children = this._elements._arr;
      const l = children.length;
      for (let i = 0; i < l; i++) {
        const child = children[i];
        child.invalidateAllTransforms();
      }
    }
  }

  /**
   * @hidden @internal
   * Recurses through the visual tree to invalidate the allTransforms for each GraphObject.
   * @this {Panel}
   */
  public invalidateAllTransforms(): void {
    this.setInvalidAllTransforms(true);
    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const child = children[i];
      child.invalidateAllTransforms();
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {number} width
   * @param {number} height
   * @param {number} minw
   * @param {number} minh
   */
  protected measureProtected(width: number, height: number, minw: number, minh: number): void {
    const union = this._union;
    union.assignTo(0, 0, 0, 0);

    const des = this.desiredSize;
    const min = this.minSize;
    if (minw === undefined) {
      minw = min.width;
      minh = min.height;
    }
    minw = Math.max(minw, min.width);
    minh = Math.max(minh, min.height);
    const max = this.maxSize;
    if (!isNaN(des.width)) width = Math.min(des.width, max.width);
    if (!isNaN(des.height)) height = Math.min(des.height, max.height);
    // height and width are affected by desired size and min size, but not max size
    // max size affects child objects only after they have been measured
    width = Math.max(minw, width);
    height = Math.max(minh, height);

    // Take out all padding from the available width and height
    // But don't let them fall below zero
    const padding = this._padding;
    width = Math.max(width - padding.left - padding.right, 0);
    height = Math.max(height - padding.top - padding.bottom, 0);

    const children = this._elements._arr;
    if (children.length !== 0) {
      this._type.measure(this, width, height, children, union, minw, minh);
    }

    // constrain union to get true natural width/height
    let nw = union.width + padding.left + padding.right;
    let nh = union.height + padding.top + padding.bottom;
    if (isFinite(des.width)) { nw = des.width; }
    if (isFinite(des.height)) { nh = des.height; }
    nw = Math.min(max.width, nw);
    nh = Math.min(max.height, nh);
    nw = Math.max(min.width, nw);
    nh = Math.max(min.height, nh);
    nw = Math.max(minw, nw);
    nh = Math.max(minh, nh);
    union.assignWidthHeight(nw, nh);
    this.naturalBounds.assignWidthHeight(nw, nh);
    this._measureRect(0, 0, nw, nh);
  }  // end measure

  /**
   * Return an immediate child element whose {@link GraphObject#isPanelMain} is true,
   * or else just return the first child element.
   * @export
   * @this {Panel}
   * @return {GraphObject}  this may return null if there are no child elements
   * @since 1.5
   */
  public findMainElement(): GraphObject | null {
    if (this._tempMain === null) {
      const children = this._elements._arr;
      const l = children.length;
      if (l === 0) return null;
      for (let i = 0; i < l; i++) {
        const child = children[i];
        if (child.isPanelMain === true) {
          this._tempMain = child;
          return child;
        }
      }
      // didn't find one marked "isPanelMain", so return first one
      this._tempMain = children[0];
    }
    return this._tempMain;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @return {GraphObject} this may return null
   */
  public getLocationObject(): GraphObject {
    return (this.part !== null) ? this.part.locationObject : null;
  }

  /**
   * @hidden @internal
   * Causes a Panel to arrange it's children.
   * @this {Panel}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    const children = this._elements._arr;
    this.actualBounds.assignTo(fx, fy, fw, fh);
    if (children.length === 0) {
      return;
    }

    if (!this.desiredSize.isReal()) {
      let effectivestretch = this.getEffectiveStretch(true);
      const mb = this.measuredBounds;
      const mbwidth = mb.width;
      const mbheight = mb.height;
      const marg = this._margin;
      const margw = marg.left + marg.right;
      const margh = marg.top + marg.bottom;
      if (mbwidth === fw && mbheight === fh) effectivestretch = GraphObject.None;
      switch (effectivestretch) {
        case GraphObject.None: // if fw or fh are smaller than our size, we have to remeasure
          if (mbwidth > fw || mbheight > fh) {
            this.invalidateMeasure(); // test me with nested panels of varying desired size
            this.measure(mbwidth > fw ? fw : mbwidth, mbheight > fh ? fh : mbheight, 0, 0);
          }
          break;
        case GraphObject.Fill: // fill the space completely, ignoring desired aspect ratio
          this.invalidateMeasure(true);
          this.measure(fw + margw, fh + margh, 0, 0);
          break;
        case GraphObject.Horizontal:
          this.invalidateMeasure(true);
          this.measure(fw + margw, mbheight + margh, 0, 0);
          break;
        case GraphObject.Vertical:
          this.invalidateMeasure(true);
          this.measure(mbwidth + margw, fh + margh, 0, 0);
          break;
      } // end stretch switch
    }
    this._type.arrange(this, children, this._union);
  }   // end arrange


  /**
   * @hidden @internal
   * Search in elements for objects containing the given point
   * @this {Panel}
   * @param {Point} p a Point in local coordinates.
   * @return {boolean}
   */
  protected containsPointProtected(p: Point): boolean {
    // get the natural size because the point p is already transformed
    const sz = this.naturalBounds;
    const locObj = this.getLocationObject();

    // if its in the bounds, iterate over this panel's collection in reverse order, calling containsPoint
    if (Rect.contains(0, 0, sz.width, sz.height, p.x, p.y)) {
      const arr = this._elements._arr;
      const l = arr.length;
      const newp = Point.allocAt(0, 0);
      for (let i = l; i--; ) {
        const child = arr[i];
        if (!child.visible && child !== locObj) continue;
        newp.set(p).transformInverted(child.transform);
        if (child.containsPoint(newp)) {
          Point.free(newp);
          return true;
        }
      }
      Point.free(newp);

      if (this._background === null && this._areaBackground === null) return false;
      return true; // No elements hit, but there was a background

    }
    return false; // Outside of the panel's bounds, even if it is inside the panel's arranged area.
  }

  /**
   * @hidden @internal
   * Used by data binding to find the unique object in a visual tree corresponding to a particular object in the template.
   * This could cache references to reduce searching overhead.
   * @this {Panel}
   * @param {number} id int
   * @return {GraphObject}
   */
  public findId(id: number): GraphObject | null {
    if (this._bindingId === id) return this;
    const elts = this._elements._arr;
    const l = elts.length;
    for (let i = 0; i < l; i++) {
      const elt = elts[i];
      const found = elt.findId(id);
      if (found !== null) return found;
    }
    return null;
  }

  /**
   * @hidden @internal
   * Walk the visual tree, starting at a given GraphObject in the tree,
   * passing this "root" Panel as the first arg to the given function for each element in the tree.
   * @export
   * @this {Panel}
   * @param {GraphObject} obj the GraphObject from which to start walking the tree
   * @param {function(GraphObject, GraphObject)} func  first arg is this Panel; second arg is the current GraphObject in the tree walk
   */
  public walkVisualTreeFrom(obj: GraphObject, func: ((a: GraphObject, b: GraphObject) => void)): void {
    func(this, obj);
    if (obj instanceof Panel) {
      const a = obj._elements._arr;
      const l = a.length;
      for (let i = 0; i < l; i++) {
        this.walkVisualTreeFrom(a[i], func);
      }
    }
  }

  /**
   * @hidden @internal
   * Walk the visual tree, ignoring non-Panels.
   * @this {Panel}
   * @param {function(Panel)} func
   */
  public walkPanels(func: ((a: Panel) => void)): void {
    this.walkPanels1(this, func);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Panel} panel
   * @param {function(Panel)} func
   */
  private walkPanels1(panel: Panel, func: ((a: Panel) => void)): void {
    func(panel);
    const a = panel._elements._arr;
    const l = a.length;
    for (let i = 0; i < l; i++) {
      const p = a[i];
      if (p instanceof Panel) {
        this.walkPanels1(p, func);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {function(GraphObject)} func
   */
  private walkVisualTree(func: ((a: GraphObject) => void)): void {
    this._walkVisualTree1(this, func);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} g
   * @param {function(GraphObject)} func
   */
  private _walkVisualTree1(g: GraphObject, func: ((a: GraphObject) => void)): void {
    func(g);
    if (g instanceof Panel) {
      const a = g._elements._arr;
      const l = a.length;
      for (let i = 0; i < l; i++) {
        this._walkVisualTree1(a[i], func);
      }
    }
  }

  /**
   * @hidden @internal
   * @export
   * @this {Panel}
   * @param {function(GraphObject):boolean} pred
   * @return {GraphObject}
   */
  public findInVisualTree(pred: ((a: GraphObject) => boolean)): GraphObject | null {
    return this._findInVisualTree1(this, pred);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} g
   * @param {function(GraphObject):boolean} pred
   * @return {GraphObject}
   */
  private _findInVisualTree1(g: GraphObject, pred: ((a: GraphObject) => boolean)): GraphObject | null {
    if (pred(g)) return g;
    if (g instanceof Panel) {
      const a = g._elements._arr;
      const l = a.length;
      for (let i = 0; i < l; i++) {
        const found = this._findInVisualTree1(a[i], pred);
        if (found !== null) return found;
      }
    }
    return null;
  }

  /**
   * Search the visual tree starting at this Panel for a GraphObject
   * whose {@link GraphObject#name} is the given name.
   *
   * This does not recurse into the elements inside a Panel that holds elements for an {@link #itemArray}.
   * @export
   * @this {Panel}
   * @param {string} name The name to search for, using a case-sensitive string comparison.
   * @return {GraphObject} If no such named object can be found, this returns null.
   */
  public findObject(name: string): GraphObject | null {
    if (this.name === name) return this;

    const elements = this._elements._arr;
    let len = elements.length;
    if (this.isItemsPanel()) {
      // If it's a panel bound to an item Array,
      // don't recurse into those elements -- they're separate namespaces.
      // Exception: for some panel types the main element(s)
      // should be searched normally, because they aren't generated
      // from item template due to binding.
      len = this.keepFirstItemElements();
    }
    for (let i = 0; i < len; i++) {
      const elt = elements[i];
      if (elt instanceof Panel) {
        const found = elt.findObject(name);
        if (found !== null) return found;
      }
      if (elt.name === name) return elt;
    }
    return null;
  }

  /**
   * @hidden @internal
   * Finds the largest miter of all children
   * @this {Panel}
   * @return {number}
   */
  protected findMiter(): number {
    const elements = this._elements._arr;
    const l = elements.length;
    let max = 0.0;
    for (let i = 0; i < l; i++) {
      const elt = elements[i];
      if (elt instanceof Panel) {
        max = Math.max(max, elt.findMiter());
      } else if (elt instanceof Shape) {
        max = Math.max(max, elt.getMiterLength());
      }
    }
    return max;
  }

  /**
   * @hidden @internal
   * Determine if an object's transform should or should not affect its children
   * This is only meaningful on Panel right now, and only for TableRow/TableColumn
   * @this {Panel}
   * @return {boolean}
   */
  public trueTransform(): boolean {
    return !(this.type === Panel.TableRow || this.type === Panel.TableColumn);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Point} p A Point in local coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @return {GraphObject} The first GraphObject in the Z-order, or else null.
   */
  public findObjectAt<T extends GraphObject>(p: Point,
                                             navig: ((a: GraphObject) => (T | null)) | null,
                                             pred: ((a: T) => boolean) | null): (T | null) {
    if (this.pickable === false) return null;
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    if (this.getInvalidArrange()) return null;

    // get the natural size because the point p is already transformed
    const sz = this.naturalBounds;
    const sc = 1 / this.getDocumentScale();
    const tt = this.trueTransform();
    const self = tt ? p : Point.allocAt(p.x, p.y).transformInverted(this.transform);

    // if its in the bounds, iterate over this panel's collection in reverse order, calling containsPoint
    const diagram = this.diagram;
    let area = 10;
    let halfarea = 5;
    if (diagram !== null) {
      area = diagram.getInputOption('extraTouchArea');
      halfarea = area / 2;
    }
    if (Rect.contains(-(halfarea * sc), -(halfarea * sc), sz.width + (area * sc), sz.height + (area * sc), self.x, self.y)) {
      if (!this.isAtomic) {
        // first check if any of the panel's elements are at the given point
        const arr = this._elements._arr;
        const l = arr.length;
        const newp = Point.alloc();
        const clip = this.isClipping;
        const main = clip ? this.findMainElement() : null;
        if (clip) {
          if (main.trueTransform()) {
            newp.set(p).transformInverted(main.transform);
          } else {
            newp.set(p);
          }
          if (!main.containsPoint(newp)) {
            Point.free(newp);
            if (!tt) Point.free(self);
            return null;
          }
        }

        const locObj = this.getLocationObject();
        for (let i = l; i--; ) {
          const child = arr[i];
          if (!child.visible && child !== locObj) continue;

          if (child.trueTransform()) {
            newp.set(p).transformInverted(child.transform);
          } else {
            newp.set(p);
          }

          if (clip && child === main) {
            continue;
          }

          let obj: T = null;
          if (child instanceof Panel) {
            obj = child.findObjectAt(newp, navig, pred);
          } else {
            // because containsPoint does not test pickable
            if (child.pickable === true && child.containsPoint(newp)) obj = child as T;
          }

          if (obj !== null) {
            if (navig !== null) obj = navig(obj);
            if (obj !== null && (pred === null || pred(obj))) {
              Point.free(newp);
              if (!tt) Point.free(self);
              return obj;
            }
          }
        }
        Point.free(newp);
      }

      if (this.background === null && this.areaBackground === null) {
        if (!tt) Point.free(self);
        return null;
      }
      // there's a background: maybe can just return this panel
      const result = Rect.contains(0, 0, sz.width, sz.height, self.x, self.y) ? this as any as T : null; // No elements hit, but there was a background
      if (!tt) Point.free(self);
      return result;
    }
    if (!tt) Point.free(self);
    return null; // Outside of the panel's bounds, even if it is inside the panel's arranged area.
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Point} p A Point in local coordinates.
   * @param {function(GraphObject):GraphObject | null=} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null=} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {List.<GraphObject>|Set.<GraphObject>=} list An optional List (or Set) to add the results to.
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectsAt<T extends GraphObject>(p: Point,
                                              navig: ((a: GraphObject) => (T | null)) | null,
                                              pred: ((a: T) => boolean) | null,
                                              list: List<T> | Set<T>): boolean {
    if (this.pickable === false) return false;
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    // get the natural size because the point p is already transformed
    const sz = this.naturalBounds;
    const tt = this.trueTransform();
    const self = tt ? p : Point.allocAt(p.x, p.y).transformInverted(this.transform);

    // TableRow could be sized 0 height yet have objs inside that span 2+ rows
    const isTableRowOrCol = (this.type === Panel.TableRow || this.type === Panel.TableColumn);
    const insideRect = Rect.contains(0, 0, sz.width, sz.height, self.x, self.y);
    // if its in the bounds, iterate over this panel's collection in reverse order, calling containsPoint
    if (isTableRowOrCol || insideRect) {
      if (!this.isAtomic) {
        // first check for all of the panel's elements at the given point
        const arr = this._elements._arr;
        const l = arr.length;
        const newp = Point.alloc();
        const locObj = this.getLocationObject();
        for (let i = l; i--; ) {
          const child = arr[i];
          if (!child.visible && child !== locObj) continue;
          if (child.trueTransform()) {
            newp.set(p).transformInverted(child.transform);
          } else {
            newp.set(p);
          }

          let obj = child as T;
          const cpanel = child instanceof Panel ? child : null;
          if (cpanel !== null ? cpanel.findObjectsAt(newp, navig, pred, list) : obj.containsPoint(newp)) {
            if (obj.pickable === false) continue; // because containsPoint does not test pickable
            if (navig !== null) obj = navig(obj);
            if (obj !== null && (pred === null || pred(obj))) {
              list.add(obj);
            }
          }
        }
        Point.free(newp);
      }

      // No elements hit, see if there was a background
      if (!tt) Point.free(self);
      return (insideRect && (this.background !== null || this.areaBackground !== null));
    }
    if (!tt) Point.free(self);
    return false;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Rect} r A Rect in container coordinates.
   * @param {function(GraphObject):GraphObject | null} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {boolean} partialInclusion True if objects are to be included when they
   * intersect.
   * @param {List.<GraphObject>|Set.<GraphObject>} list A List (or Set) to add the results to.
   * @param {Transform=} txfm
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectsIn<T extends GraphObject>(r: Rect,
                                              navig: ((a: GraphObject) => (T | null)) | null,
                                              pred: ((a: T) => boolean) | null,
                                              partialInclusion: boolean,
                                              list: List<T> | Set<T>,
                                              txfm?: Transform): boolean {
    if (this.pickable === false) return false;
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    let transform = txfm;
    if (txfm === undefined) {
      transform = Transform.alloc();
      transform.reset();
    }

    // ??? how do we ensure this works when trueTransform is false?

    transform.multiply(this.transform);
    if (this.containedInRect(r, transform)) { // first 'yes'
      // add to the list "part" and everything inside of "part",
      // going down the object tree using a special function that only tests pred and nothing geometrical,c
      // because by everything inside is already completely within r
      this.addObjs(navig, pred, list);
      if (txfm === undefined) Transform.free(transform);
      return true;
    } else if (this.intersectsRect(r, transform)) { // first 'no'
      // 1. We have to go through all the children of "part", invoking "findObjectsIn" on each.
      // We do it regardless of partialInclusion or not, because "part" might have a child that is entirely included,
      // or it may not, and we have no way of knowing yet.
      if (!this.isAtomic) {
        const locObj = this.getLocationObject();
        const arr = this._elements._arr;
        const l = arr.length;
        for (let i = l; i--; ) {
          const child = arr[i];
          if (!child.visible && child !== locObj) continue;
          const cab = child.actualBounds;
          const ab = this.naturalBounds;
          if (cab.x > ab.width || cab.y > ab.height || cab.x + cab.width < 0 || cab.y + cab.height < 0) continue;

          let obj = child as T;
          const cpanel = child instanceof Panel ? child : null;
          const tt = Transform.alloc();
          tt.set(transform);
          if (cpanel !== null ?
            cpanel.findObjectsIn(r, navig, pred, partialInclusion, list, tt) :
            obj.findObjectIn(r, partialInclusion, tt)) {
            if (navig !== null) obj = navig(obj);
            if (obj !== null && (pred === null || pred(obj))) {
              list.add(obj);
            }
          }
          Transform.free(tt);
        }
      }
      if (txfm === undefined) Transform.free(transform);
      return partialInclusion;
    }
    // second 'no'
    // nothing added to the list, we are done.
    if (txfm === undefined) Transform.free(transform);
    return false;
  }

  /**
   * @hidden @internal
   * Used by Panel.findObjectsIn to place more objects into List that are wholly within the rect.
   * @this {Panel}
   * @param {function(GraphObject):GraphObject | null} navig
   * @param {function(GraphObject):boolean | null} pred
   * @param {List.<GraphObject>|Set.<GraphObject>} list
   */
  protected addObjs<T extends GraphObject>(navig: ((a: GraphObject) => (T | null)) | null,
                                           pred: ((a: T) => boolean) | null,
                                           list: List<T> | Set<T>): void {
    const arr = this._elements._arr;
    const l = arr.length;
    for (let i = l; i--; ) {
      const child = arr[i];
      if (!child.visible) continue;
      const cab = child.actualBounds;
      const ab = this.naturalBounds;
      if (cab.x > ab.width || cab.y > ab.height || cab.x + cab.width < 0 || cab.y + cab.height < 0) continue;
      if (child instanceof Panel) {
        child.addObjs(navig, pred, list);
      }
      let obj = child as T;
      if (navig !== null) obj = navig(obj);
      if (obj !== null && (pred === null || pred(obj))) {
        list.add(obj);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {Point} p A Point in container coordinates.
   * @param {Point} distp A Point representing the distance from the point p.
   * @param {function(GraphObject):GraphObject | null} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {boolean} partialInclusion True if objects are to be included when they intersect.
   * @param {List.<GraphObject>|Set.<GraphObject>} list An optional List (or Set) to add the results to.
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectsNear<T extends GraphObject>(p: Point,
                                                distp: Point,
                                                navig: ((a: GraphObject) => (T | null)) | null,
                                                pred: ((a: T) => boolean) | null,
                                                partialInclusion: boolean,
                                                list: List<T> | Set<T>): boolean {
    if (this.pickable === false) return false;
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;

    const ns = this.naturalBounds;
    const tt = this.trueTransform();
    const self = tt ? p : Point.allocAt(p.x, p.y).transformInverted(this.transform);
    const selfdist = tt ? distp : Point.allocAt(distp.x, distp.y).transformInverted(this.transform);

    const dist = self.distanceSquaredPoint(selfdist);
    const partial = (self.x > 0 && self.x < ns.width && self.y > 0 && self.y < ns.height) ||
      (Point.distanceLineSegmentSquared(self.x, self.y, 0, 0, 0, ns.height) <= dist) ||
      (Point.distanceLineSegmentSquared(self.x, self.y, 0, ns.height, ns.width, ns.height) <= dist) ||
      (Point.distanceLineSegmentSquared(self.x, self.y, ns.width, ns.height, ns.width, 0) <= dist) ||
      (Point.distanceLineSegmentSquared(self.x, self.y, ns.width, 0, 0, 0) <= dist);

    const total = self.distanceSquared(0, 0) <= dist &&
      self.distanceSquared(0, ns.height) <= dist &&
      self.distanceSquared(ns.width, 0) <= dist &&
      self.distanceSquared(ns.width, ns.height) <= dist;

    if (!tt) { Point.free(self); Point.free(selfdist); }

    if (partial) {
      if (!this.isAtomic) {
        const newp = Point.alloc();
        const newdistp = Point.alloc();
        const locObj = this.getLocationObject();
        const arr = this._elements._arr;
        const l = arr.length;
        for (let i = l; i--; ) {
          const child = arr[i];
          if (!child.visible && child !== locObj) continue;
          const cab = child.actualBounds;
          const ab = this.naturalBounds;
          if (tt && (cab.x > ab.width || cab.y > ab.height || cab.x + cab.width < 0 || cab.y + cab.height < 0)) continue;

          if (child.trueTransform()) {
            const ct = child.transform;
            newp.set(p).transformInverted(ct);
            newdistp.set(distp).transformInverted(ct);
          } else {
            newp.set(p);
            newdistp.set(distp);
          }

          let obj = child as T;
          const cpanel = child instanceof Panel ? child : null;
          if (cpanel !== null ? cpanel.findObjectsNear(newp, newdistp, navig, pred, partialInclusion, list) : obj.findObjectNear(newp, newdistp, partialInclusion)) {
            if (navig !== null) obj = navig(obj);
            if (obj !== null && (pred === null || pred(obj))) {
              list.add(obj);
            }
          }
        }
        Point.free(newp);
        Point.free(newdistp);
      }
      if (partialInclusion) {
        return partial;
      } else {
        return total;
      }
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} elt
   * @return {Spot}
   */
  public computeSpot1(elt: GraphObject): Spot {
    let spot = null;
    if (elt instanceof Shape) {
      spot = elt.spot1;
      if (spot === Spot.Default) spot = null;
      const geo = elt.geometry;
      if (geo !== null && spot === null) spot = geo.spot1;
    }
    if (spot === null) spot = Spot.TopLeft;
    return spot;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} elt
   * @return {Spot}
   */
  public computeSpot2(elt: GraphObject): Spot {
    let spot = null;
    if (elt instanceof Shape) {
      spot = elt.spot2;
      if (spot === Spot.Default) spot = null;
      const geo = elt.geometry;
      if (geo !== null && spot === null) spot = geo.spot2;
    }
    if (spot === null) spot = Spot.BottomRight;
    return spot;
  }

  // Adding

  /**
   * Adds a GraphObject to the end of this Panel's list of elements, visually in front of all of the other elements.
   *
   * If the element to be added is already in this Panel's list of elements,
   * the object is moved to the end of the list.
   * You cannot add a GraphObject to a Panel if that GraphObject is already in a different Panel.
   * @export
   * @this {Panel}
   * @param {GraphObject} element A GraphObject.
   */
  public add(element: GraphObject): void {
    Util.checkClass(element, GraphObject, Panel, 'add:element');
    this.insertAt(this._elements.count, element);
  }

  /**
   * Returns the GraphObject in this Panel's list of elements at the specified index.
   * @export
   * @this {Panel}
   * @param {number} idx
   * @return {GraphObject}
   */
  public elt(idx: number): GraphObject {
    return this._elements.elt(idx);
  }

  /**
   * Adds a GraphObject to the Panel's list of elements at the specified index.
   *
   * If the element to be added is already in this Panel's list of elements,
   * the object is moved to the specified index.
   * You cannot add a GraphObject to a Panel that if that GraphObject is already in a different Panel.
   * @export
   * @this {Panel}
   * @param {number} index
   * @param {GraphObject} element A GraphObject.
   */
  public insertAt(index: number, element: GraphObject): void {
    if (element instanceof Part) Util.throwError('Cannot add a Part to a Panel: ' + element + '; use a Panel instead');
    if (this === element || this.isContainedBy(element)) {
      if (this === element) Util.throwError('Cannot make a Panel contain itself: ' + this.toString());
      Util.throwError('Cannot make a Panel indirectly contain itself: ' + this.toString() + ' already contains ' + element.toString());
    }
    const eltpanel = element.panel;
    if (eltpanel !== null && eltpanel !== this) {
      Util.throwError('Cannot add a GraphObject that already belongs to another Panel to this Panel: ' +
        element.toString() + ', already contained by ' + eltpanel.toString() + ', cannot be shared by this Panel: ' + this.toString());
    }
    if (this._type === Panel.Grid && !(element instanceof Shape)) {
      Util.throwError('Can only add Shapes to a Grid Panel, not: ' + element);
    }
    if (this._type === Panel.Graduated && !(element instanceof Shape || element instanceof TextBlock)) {
      Util.throwError('Can only add Shapes or TextBlocks to a Graduated Panel, not: ' + element);
    }
    element.setPanel(this);
    element._part = null;

    if (this.itemArray !== null) {
      const itemdata = (element as any).data;
      if (itemdata !== null && typeof itemdata === 'object') {
        if (this._itemElementMap === null) this._itemElementMap = new Map<ObjectData, Panel>();
        this._itemElementMap.add(itemdata, element as Panel);  // keep track of existing items-elements
      }
    }

    const elements = this._elements;
    let wasRemoved = -1;
    if (eltpanel === this) {
      let idx = -1;
      const arr = this._elements._arr;
      const l = arr.length;
      for (let i = 0; i < l; i++) {
        if (arr[i] === element) { idx = i; break; }
      }
      if (idx !== -1) {
        if (idx === index) return; // no-op, insert in place
        if (idx + 1 >= elements.count && index >= elements.count) return; // no-op, insert at end
        elements.removeAt(idx);
        wasRemoved = idx;
      } else {
        // This should never happen, but if it does, we'd really like to know
        Util.throwError('element ' + element.toString() + ' has panel ' +
          eltpanel.toString() + ' but is not contained by it.');
      }
    }
    if (index < 0 || index > elements.count) index = elements.count;
    elements.insertAt(index, element);
    if (index === 0 || element.isPanelMain) this._tempMain = null; // reset flag
    if (!this.getInvalidMeasure()) {
      this.invalidateMeasure();
    }
    element.invalidateMeasure(false);
    if (element.portId !== null) this.hasNamedPorts = true;
    else if (element instanceof Panel) {
      if (element.hasNamedPorts === true) this.hasNamedPorts = true;
    }
    this._graduatedIntervals = null;
    const part = this.part;
    if (part !== null) {
      part.clearNameMap();
      part.resetMaxMiter();
      if (this.hasNamedPorts && part instanceof Node) {
        part.hasNamedPorts = true;
      }
      if (part.hasNamedPorts && part instanceof Node) part._ports = null;  // clear cache

      // don't do some side-effects during undo or redo:
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      if (wasRemoved !== -1) part.raiseChangedEvent(ChangedEvent.Remove, 'elements', this, element, null, wasRemoved, null);
      part.raiseChangedEvent(ChangedEvent.Insert, 'elements', this, null, element, null, index);
      // if this panel is disabled, make sure the element is disabled
      if (!this.isEnabledObject()) {
        this._walkIsEnabled(element, false);
      }
    }
  }

  // set to true if a graphobject with a port is added to this panel, or if a portId in this node changes
  /** @hidden @internal */
  /* noexport */get hasNamedPorts() { return (this._flags & 0x00800000) !== 0; }
  set hasNamedPorts(value) {
    const old = (this._flags & 0x00800000) !== 0;
    if (old !== value) {
      this._flags = this._flags ^ 0x00800000;
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   */
  public getUplevelBindings(): boolean {
    return (this._flags & 0x01000000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {boolean} value
   */
  protected setUplevelBindings(value: boolean) {
    if (value) {
      this._flags |= 0x01000000;
    } else {
      this._flags &= ~0x01000000;
    }
  }

  /**
   * @hidden @internal
   * Special insert for copying panels
   * @this {Panel}
   * @param {GraphObject} element A GraphObject.
   */
  protected copyInsert(element: GraphObject): void {
    element.setPanel(this);
    element._part = null;

    const elements = this._elements;
    const index = elements.count;
    elements.insertAt(index, element);

    // ??? why is this raiseChangeEvent being done at all?
    const part = this.part;
    if (part !== null) {
      part.clearNameMap();
      // dont need to resetMaxMiter, it isn't copied
      if (element.portId !== null && part instanceof Node) {
        part.hasNamedPorts = true;
      }
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      part.raiseChangedEvent(ChangedEvent.Insert, 'elements', this, null, element, null, index);
    }
  }

  // Removing

  /**
   * Removes a GraphObject from this Panel's list of elements.
   * @export
   * @this {Panel}
   * @param {GraphObject} element A GraphObject.
   */
  public remove(element: GraphObject): void {
    Util.checkClass(element, GraphObject, Panel, 'remove:element');
    const arr = this._elements._arr;
    const l = arr.length;
    let index = -1;
    for (let i = 0; i < l; i++) {
      if (arr[i] === element) { index = i; break; }
    }
    if (index !== -1) {
      this.internalRemove(index, true);
    }
  }

  /**
   * Removes an GraphObject from this Panel's list of elements at the specified index.
   * @export
   * @this {Panel}
   * @param {number} idx
   */
  public removeAt(idx: number): void {
    if (Debug) Util.checkRealNumber(idx, Panel, 'removeAt:idx');
    if (idx >= 0) {
      this.internalRemove(idx, true);
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {number} index
   * @param {boolean=} removelinks
   */
  public internalRemove(index: number, removelinks?: boolean): void {
    const elements = this._elements;
    const element = elements.elt(index);
    element._part = null;
    element.setPanel(null);

    if (this._itemElementMap !== null) {
      const itemdata = (element as any).data;
      if (typeof itemdata === 'object') {
        this._itemElementMap.remove(itemdata);
      }
    }

    elements.removeAt(index);
    this.setInvalidMeasure(false);
    this.invalidateMeasure();
    if (this._tempMain === element) this._tempMain = null;
    this._graduatedIntervals = null;

    const part = this.part;
    if (part !== null) {
      part.clearNameMap();
      part.resetMaxMiter();
      part.invalidateAdornments();  // in case an Adornment depends on the removed element
      if (part instanceof Node) {
        if (element instanceof Panel) {
          element.walkVisualTreeFrom(element,
            function (panel, x) {
              part.internalRemovePort(x, removelinks);
            });
        } else {
          part.internalRemovePort(element, removelinks);
        }
      }
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      part.raiseChangedEvent(ChangedEvent.Remove, 'elements', this, element, null, index, null);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: This read-only property returns the number of rows.
   * This value is only valid after the Panel has been measured.
   */
  /** @export */ get rowCount(): number { return this._rowDefs === undefined ? 0 : this._rowDefs.length; }  // ?? gotta keep check for undefined

  /**
   * For {@link Panel.Table|Table} Panels: Gets the {@link RowColumnDefinition} for a particular row.
   * If you ask for the definition of a row at or beyond the {@link #rowCount},
   * it will automatically create one and return it.
   *
   * If this Panel is not a Table Panel, this method returns null.
   * @expose
   * @this {Panel}
   * @param {number} idx the non-negative zero-based integer row index.
   * @return {RowColumnDefinition}
   */
  public getRowDefinition(idx: number): RowColumnDefinition {
    if (Debug) Util.checkRealNumber(idx, Panel, 'getRowDefinition:idx');
    if (idx < 0) Util.throwRangeError(idx, '>= 0', Panel, 'getRowDefinition:idx');
    idx = Math.round(idx);
    const defs = this._rowDefs;
    if (defs === undefined) return null;

    if (defs[idx] === undefined) {
      const d = new RowColumnDefinition();
      d.setPanel(this);
      d.isRow = true;
      d.index = idx;
      defs[idx] = d;
    }
    return defs[idx];
  }

  /**
   * For {@link Panel.Table|Table} Panels: Removes the {@link RowColumnDefinition} for a particular row.
   *
   * If this Panel is not a Table Panel, this method does nothing.
   * @export
   * @this {Panel}
   * @param {number} idx the non-negative zero-based integer row index.
   */
  public removeRowDefinition(idx: number): void {
    if (Debug) Util.checkRealNumber(idx, Panel, 'removeRowDefinition:idx');
    if (idx < 0) Util.throwRangeError(idx, '>= 0', Panel, 'removeRowDefinition:idx');
    idx = Math.round(idx);
    const defs = this._rowDefs;
    if (defs === undefined) return;
    this.raiseChangedEvent(ChangedEvent.Remove, 'coldefs', this, defs[idx], null, idx, null);
    if (defs[idx]) { delete defs[idx]; }
    this.invalidateMeasure();
  }

  /**
   * For {@link Panel.Table|Table} Panels: This read-only property returns the number of columns.
   * This value is only valid after the Panel has been measured.
   */
  /** @export */ get columnCount(): number { return this._colDefs === undefined ? 0 : this._colDefs.length; }  // ?? gotta keep check for undefined

  /**
   * For {@link Panel.Table|Table} Panels: Gets the {@link RowColumnDefinition} for a particular column.
   * If you ask for the definition of a column at or beyond the {@link #columnCount},
   * it will automatically create one and return it.
   *
   * If this Panel is not a Table Panel, this method returns null.
   * @expose
   * @this {Panel}
   * @param {number} idx the non-negative zero-based integer column index.
   * @return {RowColumnDefinition}
   */
  public getColumnDefinition(idx: number): RowColumnDefinition {
    if (Debug) Util.checkRealNumber(idx, Panel, 'getColumnDefinition:idx');
    if (idx < 0) Util.throwRangeError(idx, '>= 0', Panel, 'getColumnDefinition:idx');
    idx = Math.round(idx);
    const defs = this._colDefs;
    if (defs === undefined) return null;

    if (defs[idx] === undefined) {
      const d = new RowColumnDefinition();
      d.setPanel(this);
      d.isRow = false;
      d.index = idx;
      defs[idx] = d;
    }
    return defs[idx];
  }

  /**
   * For {@link Panel.Table|Table} Panels: Removes the {@link RowColumnDefinition} for a particular row.
   *
   * If this Panel is not a Table Panel, this method does nothing.
   * @export
   * @this {Panel}
   * @param {number} idx the non-negative zero-based integer row index.
   */
  public removeColumnDefinition(idx: number): void {
    if (Debug) Util.checkRealNumber(idx, Panel, 'removeColumnDefinition:idx');
    if (idx < 0) Util.throwRangeError(idx, '>= 0', Panel, 'removeColumnDefinition:idx');
    idx = Math.round(idx);
    const defs = this._colDefs;
    if (defs === undefined) return;
    this.raiseChangedEvent(ChangedEvent.Remove, 'coldefs', this, defs[idx], null, idx, null);
    if (defs[idx]) { delete defs[idx]; }
    this.invalidateMeasure();
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets how this Panel's rows deal with extra space.
   * Valid values are {@link RowColumnDefinition.ProportionalExtra} and {@link RowColumnDefinition.None}.
   * The default is {@link RowColumnDefinition.ProportionalExtra}.
   * @see {@link RowColumnDefinition#sizing}
   */
  /** @export */ get rowSizing(): EnumValue { return this._rowSizing === undefined ? RowColumnDefinition.ProportionalExtra : this._rowSizing; }
  /** @export */ set rowSizing(value: EnumValue) {
    if (this._rowSizing === undefined) return;
    const old = this._rowSizing;
    if (old !== value) {
      if (Debug && value !== RowColumnDefinition.ProportionalExtra && value !== RowColumnDefinition.None) {
        Util.throwError('Panel.rowSizing must be RowColumnDefinition.ProportionalExtra or RowColumnDefinition.None, not: ' + value);
      }
      this._rowSizing = value;
      this.invalidateMeasure();
      this.raiseChanged('rowSizing', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets how this Panel's columns deal with extra space.
   * Valid values are {@link RowColumnDefinition.ProportionalExtra} and {@link RowColumnDefinition.None}.
   * The default is {@link RowColumnDefinition.ProportionalExtra}.
   * @see {@link RowColumnDefinition#sizing}
   * @see {@link #rowSizing}
   */
  /** @export */ get columnSizing(): EnumValue { return this._columnSizing === undefined ? RowColumnDefinition.ProportionalExtra : this._columnSizing; }
  /** @export */ set columnSizing(value: EnumValue) {
    if (this._columnSizing === undefined) return;
    const old = this._columnSizing;
    if (old !== value) {
      if (Debug && value !== RowColumnDefinition.ProportionalExtra && value !== RowColumnDefinition.None) {
        Util.throwError('Panel.columnSizing must be RowColumnDefinition.ProportionalExtra or RowColumnDefinition.None, not: ' + value);
      }
      this._columnSizing = value;
      this.invalidateMeasure();
      this.raiseChanged('columnSizing', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the first row that this Panel displays.
   * The default value is 0.
   * @see {@link #leftIndex}
   */
  /** @export */ get topIndex(): number { return this._topIndex === undefined ? 0 : this._topIndex; }
  /** @export */ set topIndex(value: number) {
    if (this._topIndex === undefined) return;
    const old = this._topIndex;
    if (old !== value) {
      if (!isFinite(value) || value < 0) Util.throwError('Panel.topIndex must be greater than zero and a real number, not: ' + value);
      this._topIndex = value;
      this.invalidateMeasure();
      this.raiseChanged('topIndex', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Gets or sets the first column that this Panel displays.
   * The default value is 0.
   * @see {@link #topIndex}
   */
  /** @export */ get leftIndex(): number { return this._leftIndex === undefined ? 0 : this._leftIndex; }
  /** @export */ set leftIndex(value: number) {
    if (this._leftIndex === undefined) return;
    const old = this._leftIndex;
    if (old !== value) {
      if (!isFinite(value) || value < 0) Util.throwError('Panel.leftIndex must be greater than zero and a real number, not: ' + value);
      this._leftIndex = value;
      this.invalidateMeasure();
      this.raiseChanged('leftIndex', old, value);
    }
  }

  /**
   * For {@link Panel.Table|Table} Panels: Returns the row at a given y-coordinate in local coordinates.
   * Call {@link GraphObject#getLocalPoint} to convert a Point in document coordinates into
   * a Point in local coordinates.
   *
   * @export
   * @this {Panel}
   * @param {number} y
   * @return {number} a zero-based integer
   * @see {@link #findColumnForLocalX}
   * @since 1.2
   */
  public findRowForLocalY(y: number): number {
    if (y < 0) return -1;
    if (this.type !== Panel.Table) return -1;
    let total = 0.0;
    const it = this._rowDefs;
    const l = it.length;
    let i = this._topIndex;
    for ( ; i < l; i++) {
      const def = it[i];
      if (def === undefined) continue;
      total += def.total;
      if (y < total) {
        return i;
      }
    }
    return i;
  }

  /**
   * For {@link Panel.Table|Table} Panels: Returns the cell at a given x-coordinate in local coordinates.
   * Call {@link GraphObject#getLocalPoint} to convert a Point in document coordinates into
   * a Point in local coordinates.
   *
   * @export
   * @this {Panel}
   * @param {number} x
   * @return {number} a zero-based integer
   * @see {@link #findRowForLocalY}
   * @since 1.2
   */
  public findColumnForLocalX(x: number): number {
    if (x < 0) return -1;
    if (this.type !== Panel.Table) return -1;
    let total = 0.0;
    const it = this._colDefs;
    const l = it.length;
    let i = this._leftIndex;
    for ( ; i < l; i++) {
      const def = it[i];
      if (def === undefined) continue;
      total += def.total;
      if (x < total) {
        return i;
      }
    }
    return i;
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Returns the point that corresponds with a value, in the panel's coordinates.
   *
   * If the value provided is not within the {@link #graduatedMin} and {@link #graduatedMax},
   * it will be constrained to within those values.
   *
   * If this Panel is not a Graduated Panel, this method returns Point(NaN, NaN).
   * @export
   * @this {Panel}
   * @param {number} val a value between {@link #graduatedMin} and {@link #graduatedMax}
   * @param {Point=} result an optional Point that is modified and returned
   * @return {Point}
   * @see {@link #graduatedValueForPoint}
   * @since 1.7
   */
  public graduatedPointForValue(val: number, result?: Point): Point {
    if (result === undefined) result = new Point(NaN, NaN);
    if (this.type !== Panel.Graduated) {
      result.assignTo(NaN, NaN);
      return result;
    }
    // value needs to be within min and max
    val = Math.min(Math.max(val, this.graduatedMin), this.graduatedMax);
    const main = this.findMainElement() as Shape;
    main.geometry.getPointAlongPath((val - this.graduatedMin) / this.graduatedRange, result);
    // convert from geometry coordinates to panel coordinates
    return main.transform.transformPoint(result);
  }

  /**
   * For {@link Panel.Graduated|Graduated} Panels: Returns the value that corresponds with the given Point.
   * The Point must be in the panel's coordinates.
   * The value returned will be in the Graduated Panel's range.
   *
   * If this Panel is not a Graduated Panel, this method returns NaN.
   * @export
   * @this {Panel}
   * @param {Point} pt a Point in the Graduated Panel's coordinates
   * @return {number}
   * @see {@link #graduatedPointForValue}
   * @since 1.7
   */
  public graduatedValueForPoint(pt: Point): number {
    if (this.type !== Panel.Graduated) return NaN;
    const main = this.findMainElement() as Shape;
    // convert from panel coordinates to geometry coordinates
    main.transform.invertedTransformPoint(pt);
    return main.geometry.getFractionForPoint(pt) * this.graduatedRange + this.graduatedMin;
  }

  /**
   * Gets or sets the optional model data to which this panel is data-bound.
   * The data must be a JavaScript Object if this is a {@link Part}.
   * The data can be any JavaScript value if this is a {@link Panel} created for
   * an item in an Array that was data-bound by the {@link #itemArray} property.
   * The default value is null.
   *
   * Setting it to a new value automatically calls {@link #updateTargetBindings}
   * in order to assign new values to all of the data-bound GraphObject properties.
   *
   * Once you set this property you cannot add, remove, or modify any data bindings
   * on any of the GraphObjects in the visual tree of this Panel, including on
   * this panel itself.
   *
   * You should not modify this property on a {@link Part} that is created automatically
   * to represent model data, nor on a {@link Panel} that is created automatically for a data item
   * in the containing Panel's {@link Panel#itemArray}.
   * Call {@link Model#removeNodeData} and {@link Model#addNodeData} if you want to replace
   * this Part with another one, or call {@link Model#removeArrayItem} and {@link Model#insertArrayItem}
   * if you want to replace this Panel with another one.
   *
   * Although you might not be able to replace this data value if this Part was created
   * automatically by the Diagram, you can still modify that data object's properties.
   * Call the appropriate Model method for changing properties that affect the structure of the diagram.
   * Call {@link Model#setDataProperty} for changing other properties that may be the sources
   * of {@link Binding}s on {@link GraphObject} properties that are in the visual tree of this panel/part.
   */
  /** @export */ get data(): any | null { return this._data; }
  /** @export */ set data(value: any | null) {
    const old = this._data;
    if (old !== value) {
      const ispart = (this instanceof Part) && !(this instanceof Adornment);
      if (ispart) Util.checkPrimitive(value, 'object', Panel, 'data');
      this.freezeBindings();
      this._data = value;
      const diagram = this.diagram;
      if (diagram !== null) {
        if (ispart) {
          diagram.partManager.replaceBoundData(this, old, value);
        } else {
          const listcontainer = this.panel;
          if (listcontainer !== null && listcontainer._itemElementMap !== null) {
            if (old !== null) listcontainer._itemElementMap.remove(old);
            if (value !== null) listcontainer._itemElementMap.add(value, this);
          }
        }
      }
      this.raiseChanged('data', old, value);
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      if (value !== null) this.updateTargetBindings();
    }
  }

  /**
   * Gets the index of this Panel's data if it was created to represent an item in its containing Panel's {@link Panel#itemArray}.
   * The default value is NaN.
   *
   * This is only set internally by code such as {@link #rebuildItemElements} or {@link Model#insertArrayItem}
   * when building or shifting Panel representing items
   * in the Panel whose {@link Panel#itemArray} was set or bound to an Array of value.
   *
   * This property can be used in data bindings within the item template to produce values
   * that depend on its position in the item Array.
   * For example:
   * ```js
   *  $(go.Panel, // the item Panel
   *    . . .,
   *    new go.Binding("itemArray", "someProperty"),
   *    {
   *      itemTemplate:
   *        $(go.Panel,
   *          // set Panel.background to a color based on the Panel.itemIndex
   *          new go.Binding("background", "itemIndex",
   *                         // using this conversion function
   *                         function(i) { return (i%2 === 0) ? "lightgreen" : "lightyellow"; })
   *                   // bound to this Panel itself, not to the Panel.data item
   *                   .ofObject(),
   *          $(go.TextBlock,  // a trivial item template, just showing some text
   *            new go.Binding("text"))  // sets TextBlock.text = data.text
   *        )
   *    }
   *  )
   * ```
   *
   * The main element of a Spot or Auto or Link Panel, or the first TableRow or TableColumn element of a Table Panel
   * whose {@link #isPanelMain} property is true, will not have this property set to a number,
   * because it will not have been created by {@link #rebuildItemElements}.
   * @since 1.4
   */
  /** @export */ get itemIndex(): number { return this._itemIndex; }
  /** @export */ set itemIndex(value: number) {
    const old = this._itemIndex;
    if (old !== value) {
      this._itemIndex = value;
      this.raiseChanged('itemIndex', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @return {boolean}
   */
  public isTemplateBinder(): boolean {
    return this._bindings !== null;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @return {boolean}
   */
  public isTemplateBinderFrozen(): boolean {
    const allbinds = this._bindings;
    return allbinds !== null && allbinds._isFrozen;
  }

  // * Collect all of the Bindings in the visual tree, freeze them,
  // * and make sure there is a name association to the target object,
  // * so that they can be found in each copy of the visual tree.

  /**
   * @hidden
   * Undocumented
   * @this {Panel}
   */
  public freezeBindings(): void {
    let allbinds = this._bindings;
    if (allbinds === null) {
      if (this.data !== null) Util.throwError('Template cannot have .data be non-null: ' + this);
      this._bindings = allbinds = new List<Binding>();
    } else if (allbinds._isFrozen) {
      return;
    }

    const srcbinds = new List<GraphObject>();

    // collect all Bindings from the visual tree
    this.setUplevelBindings(false);
    this.walkVisualTreeFrom(this,
      function (rootPanel, obj) {
        const locals = obj._binds;
        if (locals !== null) {
          obj.setBindsTwoWay(false);
          const bit2 = locals.iterator;
          while (bit2.next()) {
            const b = bit2.value;
            // If two-way binding, remember that, so GraphObject.raiseChangedEvent
            // can find and call Binding.updateSource.
            if (b.mode === Binding.TwoWay) obj.setBindsTwoWay(true);
            // If binding element to element, record binding in source GraphObject,
            // so that GraphObject.raiseChangedEvent can find it and update this GraphObject.
            const srcname = b.sourceName;
            if (srcname !== null) {
              if (srcname === '/') {
                (rootPanel as Panel).setUplevelBindings(true);
              }
              const srcelt = b.findSourceGraphObject(rootPanel as Panel, obj);
              if (srcelt !== null) {
                // postpone adding Bindings to the source element, to avoid concurrent modification of Lists
                srcbinds.add(srcelt);
                // remember the Bindings to be added on the source element itself, in _srcbinds
                if (srcelt._srcbinds === null) srcelt._srcbinds = new List<Binding>();
                srcelt._srcbinds.add(b);
              }
            }
            allbinds.add(b);
          }
        }
        if (!(obj instanceof Panel)) return;
        if (obj.type === Panel.Table) {
          // handle RowColumnDefinitions, but don't need to record bindsTwoWay
          if (obj._rowDefs.length > 0) {
            const it = obj._rowDefs;
            const l = it.length;
            for (let i = 0; i < l; i++) {
              const def = it[i];
              if (def === undefined) continue;
              if (def._binds !== null) {
                const dit = def._binds.iterator;
                while (dit.next()) {
                  const rb = dit.value;
                  rb._targetObject = def;
                  rb._targetKind = 2;
                  rb._targetArg = def.index;
                  allbinds.add(rb);
                }
              }
            }
          }
          if (obj._colDefs.length > 0) {
            const it = obj._colDefs;
            const l = it.length;
            for (let i = 0; i < l; i++) {
              const def = it[i];
              if (def === undefined) continue;
              if (def._binds !== null) {
                const dit = def._binds.iterator;
                while (dit.next()) {
                  const cb = dit.value;
                  cb._targetObject = def;
                  cb._targetKind = 1;
                  cb._targetArg = def.index;
                  allbinds.add(cb);
                }
              }
            }
          }
        }
      });

    // Now iterate over all source elements (if any).
    // Make sure it's declared as TwoWay, to inform notification about side-effects via bindings,
    // and add all bindings using the source to the source's _binds list.
    const sbit = srcbinds.iterator;
    while (sbit.next()) {
      const srcelt = sbit.value;
      if (srcelt._srcbinds !== null) {
        srcelt.setBindsTwoWay(true);
        const sbbit = srcelt._srcbinds.iterator;
        while (sbbit.next()) {
          const sb = sbbit.value;
          if (srcelt._binds === null) srcelt._binds = new List<Binding>();
          srcelt._binds.add(sb);
        }
      }
      srcelt._srcbinds = null;
    }

    // make sure each Binding and corresponding target have an id
    let bit = allbinds.iterator;
    while (bit.next()) {
      const bind = bit.value;
      let target = bind._targetObject;
      if (target !== null) {  // not yet setup to search by name
        bind._targetObject = null;

        // handle "OBJNAME.prop" syntax, when the Binding is on a Panel,
        // assume "OBJNAME" is the name of the actual element within the Panel,
        // referenced by Binding as the object with the target property
        const prop = bind.targetProperty;
        const dotindex = prop.indexOf('.');
        if (dotindex > 0 && target instanceof Panel) {
          const objname = prop.substring(0, dotindex);
          const propname = prop.substr(dotindex + 1);
          const actual = target.findObject(objname);
          if (actual !== null) {
            target = actual;
            bind.targetProperty = propname;
          } else {
            Util.trace('Warning: unable to find GraphObject named \"' + objname + '\" for Binding: ' + bind.toString());
          }
        }

        // now set up search target object id
        if (target instanceof RowColumnDefinition) {
          const id = Set.hashId(target.panel);
          bind.targetId = (id === undefined) ? -1 : id;
          target.panel._bindingId = bind.targetId;
        } else if (target instanceof GraphObject) {
          const id = Set.hashId(target);
          bind.targetId = (id === undefined) ? -1 : id;
          target._bindingId = bind.targetId;
        } else {
          Util.throwError('Unknown type of binding target: ' + target);
        }
      }
    }
    allbinds.freeze();  // freeze the collection

    // We should try to measure and arrange all templates
    if (this instanceof Part) {
      if (this.isNodeLike()) { // ??? not Part?
        this.ensureBounds();
      }
      // warn about no elements or 1 element in an Auto or Spot Panel
      if (Debug && !Panel._warnedAboutNoElementsInTemplate) {
        this.walkVisualTreeFrom(this, function (rootObj, obj) {
          if (!(obj instanceof Panel)) return;
          if ((obj.type === Panel.Auto || obj.type === Panel.Spot || obj.type === Panel.Graduated) &&
            obj.elements.count <= 1 &&
            !(obj instanceof Part)) {
            let variable = false;
            if (obj.elements.count === 1) {
              // but OK to have only one element if itemArray is set or bound
              variable = (obj.itemArray !== null);
              if (!variable) {
                bit = allbinds.iterator;
                while (bit.next()) {
                  const bind = bit.value;
                  if (bind.targetProperty === 'itemArray') {
                    variable = true;
                    break;
                  }
                }
              }
            }
            if (!variable) {
              Util.trace('Auto, Spot, or Graduated Panel should not have zero or one elements: ' + obj.toString() + ' in ' + rootObj.toString());
              Panel._warnedAboutNoElementsInTemplate = true;
            }
          }
        });
      }
    }
  }

  /**
   * @hidden @internal
   */
  public static _warnedAboutNoElementsInTemplate = false;

  /**
   * @hidden
   * Undocumented.
   * Make a deep copy of this Panel and allow it to be a template.
   * This makes copies of Bindings, unlike the regular copy().
   * @export
   * @this {Panel}
   * @return {Panel}
   */
  public copyTemplate(): Panel {
    const copy = this.copy();
    copy.walkVisualTree(function(obj) {
      // discard frozen binding collections
      if (obj instanceof Panel) {
        obj._bindings = null;
        obj._data = null;  // don't set .data
      }
      // copy individual Bindings
      const oldbinds = obj._binds;
      if (oldbinds !== null) {
        // clear this list and replace with copied Bindings
        obj._binds = null;
        oldbinds.each(function(b) {
          obj.bind(b.copy());
        });
      }
      // copy individual Triggers
      const oldtriggers = obj._animationTriggers;
      if (oldtriggers !== null) {
        // clear this list and replace with copied Bindings
        obj._animationTriggers = null;
        oldtriggers.each(function(b) {
          obj.trigger(b.value.copy());
        });
      }
      // ??? Should this copy itemTemplateMaps?
    });
    return copy;
  }

  /**
   * Re-evaluate all data bindings on this panel,
   * in order to assign new property values to the {@link GraphObject}s in this visual tree
   * based on this object's {@link #data} property values.
   *
   * It is better to call {@link Model#setDataProperty} to modify data properties,
   * because that will both record changes for undo/redo and will update all bindings
   * that may depend on that property.
   *
   * This method does nothing if {@link #data} is null.
   * @export
   * @this {Panel}
   * @param {string=} srcprop An optional source data property name:
   *   when provided, only evaluates those {@link Binding}s that use that particular property;
   *   when not provided or when it is the empty string, all bindings are evaluated.
   * @see {@link Model#updateTargetBindings}
   */
  public updateTargetBindings(srcprop?: string): void {
    const allbinds = this._bindings;
    if (allbinds === null) return;
    if (srcprop === undefined) srcprop = '';
    const bit = allbinds.iterator;
    while (bit.next()) {
      const bind = bit.value;
      const sprop = bind.sourceProperty;
      // if the modified property and the Binding property are both specified but not the same, skip it
      if (srcprop !== '' && sprop !== '' && sprop !== srcprop) continue;
      const tprop = bind.targetProperty;
      const conv = bind.converter;
      if (conv === null && tprop === '') continue;  // can't replace GraphObject!

      // normally bound to DATA, but use a named GraphObject if sourceName
      let data = this.data;
      const srcnm = bind.sourceName;
      if (srcnm !== null) {
        // can't use Binding.findSourceGraphObject here
        if (srcnm === '') data = this;  // NOT this.part!
        else if (srcnm === '/') data = this;  // temporarily set to template binder; later use this.part
        else if (srcnm === '.') data = this;  // temporarily set to template binder; later use target ELT itself
        else if (srcnm === '..') data = this;  // temporarily set to template binder; later use target ELT.panel
        else data = this.findObject(srcnm);
      } else {
        const diagram = this.diagram;
        if (diagram !== null && bind.isToModel) data = diagram.model.modelData;
      }
      if (data === null) {
        if (Debug) Util.trace('Binding error: missing GraphObject named ' + srcnm + ' in ' + this.toString());
        continue;
      }

      // Find the data-bound object
      let elt: ObjectData = this;  // could be any number of types
      const id = bind.targetId;
      // Find the named object; use THIS if name is empty
      if (id !== -1) {
        elt = this.findId(id);
        if (elt === null) continue;  // couldn't find named object
      } else if (bind._targetObject !== null) {
        elt = bind._targetObject;  // not being used in a template
      }
      // special case: binding to itself
      if (srcnm === '/') data = elt.part;
      else if (srcnm === '.') data = elt;
      else if (srcnm === '..') data = elt.panel;

      // Find the target object, which might be a RowColumnDefinition instead
      const kind = bind._targetKind;
      if (kind !== 0) {
        if (!(elt instanceof Panel)) continue;  // and the object must be a Panel
        if (kind === 1) elt = elt.getColumnDefinition(bind._targetArg);
        else if (kind === 2) elt = elt.getRowDefinition(bind._targetArg);
      }

      if (elt === undefined) continue;
      bind.updateTarget(elt, data);
    }
  }

  /**
   * Gets or sets a JavaScript Array of values or objects, each of which will be represented
   * by a {@link Panel} as elements in this Panel.
   * Replacing this array results all of this panel's child objects being replaced with
   * a copy of the {@link Panel} found in {@link #itemTemplateMap} for each particular
   * item in the Array.
   *
   * Because the software does not receive any notifications when an Array is modified,
   * any insertions or removals or replacements of data in the Array will not be noticed
   * unless you call {@link Model#insertArrayItem} or {@link Model#removeArrayItem}.
   * You may also reset this property to its current value (the modified Array) or
   * call {@link #updateTargetBindings}, if there is a {@link Binding} whose target is this property.
   *
   * When binding this property, it is commonplace to set {@link Model#copiesArrays} and
   * {@link Model#copiesArrayObjects} properties to true, so that when a node is copied,
   * the item Array and its contents are copied, not shared.
   * Or more generally, to customize the model's copying processes, you can supply a custom
   * {@link Model#copyNodeDataFunction} and perhaps a {@link GraphLinksModel#copyLinkDataFunction}.
   *
   * Any JavaScript Object that is in this Array must only appear once in the array
   * and must not appear in any other {@link Panel#itemArray}s.
   * Use {@link #findItemPanelForData} to find the data-bound Panel created for an Object in this panel's item Array.
   *
   * Non-Object values in an item Array may appear multiple times.
   * An item Array may be shared by multiple Panels.
   *
   * Item Arrays should not be used with Grid Panels or Graduated Panels as they may not have nested Panels.
   */
  /** @export */ get itemArray(): Array<any> | null { return this._itemArray; }
  /** @export */ set itemArray(value: Array<any> | null) {
    const old = this._itemArray;
    if (old !== value || (value !== null && this.isItemArrayDifferent(value))) {
      if (Debug && value !== null && !Util.isArray(value)) Util.throwError('Panel.itemArray must be an Array-like object or null, not: ' + value);
      const diagram = this.diagram;
      if (old !== value) {
        if (diagram !== null && old !== null) diagram.partManager.forgetItemArrayPanel(this, diagram);
        this._itemArray = value;
        if (diagram !== null && value !== null) diagram.partManager.rememberItemArrayPanel(this);
      }
      this.raiseChanged('itemArray', old, value);
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      this.rebuildItemElements();
    }
  }

  /**
   * @hidden @internal
   * If any of the child Panels do not correspond with each of the given Array's items in order,
   * or if there isn't such a Panel for an Array item, return true.
   * @return {boolean}
   */
  private isItemArrayDifferent(arr: Array<any>): boolean {
    const children = this._elements._arr;
    const clen = children.length;
    const alen = arr.length;
    let ci = 0;

    // skip elements that are not modeled (i.e. when not a Panel or Panel.data is null)
    let c: GraphObject | null = null;
    while (ci < clen) {
      c = children[ci];
      if (!(c instanceof Panel) || c.data === null) {
        ci++;
        c = children[ci];
      } else {
        break;
      }
    }
    if (clen - ci !== alen) return true;  // easy case -- number of elements doesn't match number of items
    if (c === null) return alen > 0;  // no modeled child?  different if ALEN non-zero
    // now C will be a Panel whose Panel.data is non-null

    // rest of elements must be data-bound to the corresponding values in the itemArray
    let ai = 0;
    while (ci < clen && ai < alen) {
      c = children[ci];
      if (!(c instanceof Panel) || c.data !== arr[ai]) return true;
      ci++;
      ai++;
    }
    return false;  // all the same
  }

  /**
   * @hidden @internal
   * This should only be called when itemArrays are involved.
   * Returns how many of the first elements of the children are kept by rebuildItemElements.
   * @this {Panel}
   * @return {number}
   */
  public keepFirstItemElements(): number {
    if (this.type === Panel.Spot || this.type === Panel.Auto) {
      return Math.min(this._elements.length, 1);
    }
    if (this.type === Panel.Link) {
      const list = this._elements;
      const len = list.length;
      let i = 0;
      for (; i < len; i++) {
        const elt = list.elt(i);
        // a plain Shape will probably be the Link.path;
        // all generated items will be Panels
        if (!(elt instanceof Shape) || !elt.isPanelMain) {
          return i;
        }
      }
      return i;
    }
    if (this.type === Panel.Table && this._elements.length > 0) {
      const list = this._elements;
      const first = list.elt(0);
      if (first.isPanelMain && first instanceof Panel) {
        if (first.type === Panel.TableRow || first.type === Panel.TableColumn) return 1;
      }
    }
    return 0;
  }

  /**
   * Create and add new GraphObjects corresponding to and bound to the data in the {@link #itemArray},
   * after removing all existing elements from this Panel.
   * This method is automatically called when replacing the {@link #itemArray} value,
   * or when changing the value of {@link #itemTemplate} or {@link #itemTemplateMap}.
   *
   * This uses {@link #itemCategoryProperty} to determine the category for an item data.
   * That string is used to look up a template in {@link #itemTemplateMap}.
   * The resulting template (which is also a Panel) is copied, added to this panel,
   * and its {@link #itemIndex} is set to its index in that Array.
   * That new child Panel is then data-bound to that Array item by setting its {@link #data}.
   *
   * If {@link #itemArray} is null, this method just removes all elements from this panel.
   * Actually, if this Panel {@link #type} is "Spot", "Auto", or "Link", the very first element
   * is always kept by this method.
   * Also, if this Panel type is "Table", and if the first element is a "TableRow" or "TableColumn" Panel
   * whose {@link #isPanelMain} property is set to true, that first element will be kept too.
   * That is useful for defining literal TableRow headers in Table panels, when the header information
   * is not kept as the first item in the itemArray.
   *
   * It is wasteful to call this method after making some model data changes.
   * It is better to call {@link Model#setDataProperty}, {@link Model#addArrayItem},
   * {@link Model#insertArrayItem}, or {@link Model#removeArrayItem}, or other model methods.
   * Not only do those methods update efficiently, they also preserve unbound state and support undo/redo.
   * @export
   * @this {Panel}
   */
  public rebuildItemElements(): void {
    // remove all existing elements
    const start = this.keepFirstItemElements();  // maybe skip first element
    while (this._elements.length > start) {
      this.internalRemove(this._elements.length - 1, false);
    }

    const arr = this.itemArray;
    if (arr === null) return;

    // add an element for each new data in the itemArray
    const len = Util.arrayLength(arr);
    for (let i = 0; i < len; i++) {
      this.addItemElement(Util.arrayElt(arr, i), i);
    }
  }

  /**
   * Return the {@link Panel} that was made for a particular data object in this panel's {@link #itemArray}.
   * If this returns a Panel, its {@link #data} property will be the argument data object,
   * and its containing {@link GraphObject#panel} will be this panel.
   * @export
   * @this {Panel}
   * @param {Object} data must be an Object, not a string or a number or a boolean or a function
   * @return {Panel} or null if not found
   * @since 1.6
   */
  public findItemPanelForData(data: ObjectData): Panel | null {
    if (data === undefined || data === null) return null;
    if (this._itemElementMap === null) return null;
    Util.checkPrimitive(data, 'object', Panel, 'findItemPanelForData');
    return this._itemElementMap.getValue(data);
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {*} data
   * @param {number} idx
   * @return {Panel}
   */
  public addItemElement(data: any, idx: number): Panel {
    if (data === undefined || data === null) return null;
    if (idx < 0) return null;
    const cat = this.getCategoryForItemData(data, idx);
    const template = this.findTemplateForItemData(data, idx, cat);
    if (template !== null) {
      template.freezeBindings();
      const elt = template.copy();
      if (elt.getUplevelBindings()) {
        const binder = this.findTemplateBinder();
        if (binder !== null) binder.setUplevelBindings(true);
      }
      if (typeof data === 'object') {
        if (this._itemElementMap === null) this._itemElementMap = new Map<ObjectData, Panel>();
        this._itemElementMap.add(data, elt);  // keep track of existing items-elements
      }
      const rowcol = idx + this.keepFirstItemElements();  // maybe skip elements
      this.insertAt(rowcol, elt);
      elt._data = data;  // temporarily, in case there are any Bindings evaluated when setting row/column/itemIndex
      this.updateTableIndexesStartingAt(rowcol, idx);
      elt._data = null;  // don't call property setter
      elt.data = data;   // until now, and evaluate all Bindings
      return elt;
    }
    return null;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {number} idx
   */
  public removeItemElement(idx: number): void {
    if (idx < 0) return;
    // add or remove an element from that Panel, but Spot and Auto Panels treat first element specially
    const rowcol = idx + this.keepFirstItemElements();  // maybe skip elements
    this.internalRemove(rowcol, true);  // ??? but would like to be able to remove without disconnecting links
    this.updateTableIndexesStartingAt(rowcol, idx);  // update any row/column values starting with this item
  }

  /**
   * @hidden @internal
   * After removing or inserting an item into an items-array Panel,
   * we need to update the row or column of all TableRow or TableColumn Panels
   * that come at and after that index in this Panel.
   * @this {Panel}
   * @param {number} rowcol
   * @param {number} idx
   */
  private updateTableIndexesStartingAt(rowcol: number, idx: number): void {
    const elts = this._elements;
    let rc = rowcol;
    let i = idx;
    while (rc < elts.length) {
      const child = elts.elt(rc);
      if (child instanceof Panel) {  // ought to always be true
        child.internalSetAllRowColumn(rc, i);
      }
      rc++;
      i++;
    }
  }

  /**
   * @hidden @internal
   * Assume this is an item Panel generated for a data item in an itemsArray.
   * Depending on whether this Panel is of type TableRow or TableColumn,
   * make sure all of its elements have the given row or column index.
   * @this {Panel}
   * @param {number} rowcol
   * @param {number} idx
   */
  private internalSetAllRowColumn(rowcol: number, idx: number): void {
    if (this.type === Panel.TableRow) {
      this.row = rowcol;
    } else if (this.type === Panel.TableColumn) {
      this.column = rowcol;
    }
    this.itemIndex = idx;
  }

  // ??? NO: define a predicate on Panel that is true when it has data and is the root for a namespace
  //    But see also keepFirstItemElements()
  /**
   * @hidden @internal
   * @this {Panel}
   * @return {boolean}
   */
  public isItemsPanel(): boolean {
    return this._itemArray !== null || this._itemTemplateMap !== null;
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @return {boolean}
   */
  public hasHTMLElement(): boolean {
    return false;
    // NYI:
    // // ?? shortcuts? performance?
    // var children = this._elements._arr;
    // var l = children.length;
    // for (var i = 0; i < l; i++) {
    //   var child = children[i];
    //   if (child instanceof HTMLHost) return true;
    //   if (child instanceof Panel) {
    //     if (child.hasHTMLElement()) return true;
    //   }
    // }
    // return false;
  }

  /**
   * Gets or sets the default {@link Panel} template used as the archetype
   * for item data that are in {@link #itemArray}.
   *
   * Setting this property just modifies the {@link #itemTemplateMap}
   * by replacing the entry named with the empty string.
   * Any new value must be a {@link Panel} but not a {@link Part}.
   * By default this property is null.
   *
   * {@link GraphObject#copy} when copying a panel will share the {@link #itemTemplateMap}
   * between the original panel and the copied panel.
   */
  /** @export */ get itemTemplate(): Panel {
    if (this._itemTemplateMap === null) return null;
    return this._itemTemplateMap.getValue('');
  }
  /** @export */ set itemTemplate(value: Panel) {
    if (this._itemTemplateMap === null) {
      if (value === null) return;  // don't allocate a Map unnecessarily
      this._itemTemplateMap = new Map<string, Panel>();
    }
    const old = this._itemTemplateMap.getValue('');
    if (old !== value) {
      Util.checkClass(value, Panel, Panel, 'itemTemplate');
      if (value instanceof Part || value.isPanelMain) Util.throwError('Panel.itemTemplate must not be a Part or be Panel.isPanelMain: ' + value);
      this._itemTemplateMap.add('', value);
      this.raiseChanged('itemTemplate', old, value);
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      this.rebuildItemElements();
    }
  }

  /**
   * Gets or sets a {@link Map} mapping template names to {@link Panel}s.
   * One of these Panels is copied for each item data that is in the {@link #itemArray}.
   * Replacing this map will automatically rebuild all of the elements in this Panel.
   *
   * By default this property is null.
   * All values in the {@link Map} must be {@link Panel}s but not {@link Part}s.
   *
   * If you modify this {@link Map}, by replacing a {@link Panel} or by adding or
   * removing a map entry, you need to explicitly call {@link #rebuildItemElements} afterwards.
   *
   * {@link GraphObject#copy} when copying a panel will share the {@link #itemTemplateMap}
   * between the original panel and the copied panel.
   */
  /** @export */ get itemTemplateMap(): Map<string, Panel> | null { return this._itemTemplateMap; }
  /** @export */ set itemTemplateMap(value: Map<string, Panel>) {
    const old = this._itemTemplateMap;
    if (old !== value) {
      Util.checkClass(value, Map, Panel, 'itemTemplateMap');
      const it = value.iterator;
      while (it.next()) {
        const t = it.value;
        if (Debug && (t instanceof Part || t.isPanelMain)) Util.throwError('Template in Panel.itemTemplateMap must not be a Part or be Panel.isPanelMain: ' + t);
      }
      this._itemTemplateMap = value;
      this.raiseChanged('itemTemplateMap', old, value);
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      this.rebuildItemElements();
    }
  }

  /**
   * Gets or sets the name of the item data property that returns a string describing that data's category,
   * or a function that takes an item data object and returns that string;
   * the default value is the name 'category'.
   * This is used to distinguish between different
   * kinds of items in the {@link #itemArray}.
   *
   * The name must not be null.
   * If the value is an empty string, the category is assumed to be an empty string,
   * the default category name, for all item data objects.
   * You must not change this property when the {@link #itemArray} already has a value.
   */
  /** @export */ get itemCategoryProperty(): string | ((a: any) => string) { return this._itemCategoryProperty; }
  /** @export */ set itemCategoryProperty(value: string | ((a: any) => string)) {
    const old = this._itemCategoryProperty;
    if (old !== value) {
      if (typeof value !== 'string' && typeof value !== 'function') {
        Util.throwTypeError(value, 'string or function', Panel, 'itemCategoryProperty');
      }
      this._itemCategoryProperty = value;
      this.raiseChanged('itemCategoryProperty', old, value);
    }
  }

  /**
   * @hidden @internal
   * Find the category of a given item data, a string naming the template
   * in this {@link Panel}'s {@link Panel#itemTemplateMap} to use to
   * represent the item data.
   * By default this just returns the empty string, resulting in the use
   * of the default {@link #itemTemplate}.
   * @this {Panel}
   * @param {*} itemdata a value that is in the {@link Panel#itemArray} .
   * @param {number} idx the zero-based index of the data in the {@link Panel#itemArray}.
   * @return {string}
   * @see {@link #itemCategoryProperty}
   */
  private getCategoryForItemData(itemdata: any, idx: number): string {
    if (itemdata === null) return '';
    const cname = this._itemCategoryProperty;
    let cat = '';
    if (typeof cname === 'function') {
      cat = cname(itemdata);
    } else if (typeof cname === 'string' && typeof itemdata === 'object') {
      if (cname === '') return '';
      cat = Model.safePropertyValue(itemdata, cname);
    } else {  // give up
      return '';
    }
    if (cat === undefined) return '';
    if (typeof cat === 'string') return cat;
    Util.throwError('Panel.getCategoryForItemData found a non-string category for ' + itemdata + ': ' + cat);
    return '';
  }

  // ??? setCategoryForItemData

  /**
   * @hidden @internal
   */
  public static _warnedAboutNoItemTemplate = false;
  /**
   * @hidden @internal
   */
  public static _defaultItemTemplate: Panel = null;

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {*} data a data item in the itemArray, of any type.
   * @param {number} idx the zero-based index of the item to be created.
   * @param {string} cat the category.
   * @return {Panel}
   */
  private findTemplateForItemData(data: any, idx: number, cat: string): Panel {
    const map = this.itemTemplateMap;
    let template: Panel = null;
    if (map !== null) template = map.getValue(cat);
    if (template === null) {
      if (!Panel._warnedAboutNoItemTemplate) {
        Panel._warnedAboutNoItemTemplate = true;
        Util.trace('No item template Panel found for category \"' + cat + '\" on ' + this);
        Util.trace('  Using default item template.');

        // build the shared default item template
        const architem = new Panel();
        const itemtxt = new TextBlock();
        itemtxt.bind(new Binding('text', '', Util.toString));
        architem.add(itemtxt);
        Panel._defaultItemTemplate = architem;
      }
      template = Panel._defaultItemTemplate;
    }
    return template;
  }

  /**
   * @hidden @internal
   * Determines whether or not this Panel's elements participate in object picking
   * (findObjectAt, findObjectsIn, findObjectNear, etc)
   * This is set to true if the Panel is of type {@link Panel.Grid|Grid}.
   */
  /** @export */ get isAtomic(): boolean { return (this._flags & 0x00100000) !== 0; }
  /** @export */ set isAtomic(value: boolean) {
    const old = (this._flags & 0x00100000) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Panel, 'isAtomic');
      this._flags = this._flags ^ 0x00100000;
      this.raiseChanged('isAtomic', old, value);
    }
  }

  /**
   * For {@link Spot} Panels: Gets or sets whether this Panel's main element clips instead of fills.
   * The main element will not paint its stroke, if it has any.
   * This assumes that the main element is a Shape.
   *
   * When this property is `true`, the Spot panel will size itself to be the *intersection* of the main element bounds and
   * all other elements' bounds, rather than the union of these bounds.
   * @since 1.7
   */
  /** @export */ get isClipping(): boolean { return (this._flags & 0x00200000) !== 0; }
  /** @export */ set isClipping(value: boolean) {
    const old = (this._flags & 0x00200000) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Panel, 'isClipping');
      this._flags = this._flags ^ 0x00200000;
      this.invalidateMeasure();
      this.raiseChanged('isClipping', old, value);
    }
  }

  /**
   * For {@link Panel.Horizontal|Horizontal} and {@link Panel.Vertical|Vertical} Panels: gets or sets whether this Panel arranges its contents from the
   * typical side (left and top, respectively), or the opposite side (right and bottom, respectively).
   *
   * The default value is false.
   * @since 1.7
   */
  /** @export */ get isOpposite(): boolean { return (this._flags & 0x02000000) !== 0; }
  /** @export */ set isOpposite(value: boolean) {
    const old = (this._flags & 0x02000000) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Panel, 'isOpposite');
      this._flags = this._flags ^ 0x02000000;
      this.invalidateMeasure();
      this.raiseChanged('isOpposite', old, value);
    }
  }

  /**
   * Gets or sets whether this Panel or any GraphObject inside the panel actually responds to user click events.
   * It may be used as a Binding target.
   * See how this property is used in <a href="../../extensions/Buttons.js">Buttons.js</a>.
   *
   * This property does not have any effect on picking ("hit-testing") -- that behavior is implemented
   * by the {@link GraphObject#pickable} property.
   * When this property is false, non-click events may still occur on this panel or on objects within this panel.
   * This property is normally only used for Panels that are {@link GraphObject#isActionable}.
   *
   * Call {@link GraphObject#isEnabledObject} to decide whether a particular object can be clicked.
   * @see {@link GraphObject#isEnabledObject}
   * @see {@link GraphObject#enabledChanged}
   * @see {@link GraphObject#isActionable}
   * @see {@link GraphObject#pickable}
   * @since 1.7
   */
  /** @export */ get isEnabled(): boolean { return (this._flags & 0x00400000) !== 0; }
  /** @export */ set isEnabled(value: boolean) {
    const old = (this._flags & 0x00400000) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Panel, 'isEnabled');
      const withinenabled = (this.panel === null || this.panel.isEnabledObject());
      this._flags = this._flags ^ 0x00400000;
      this.raiseChanged('isEnabled', old, value);
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      if (withinenabled) {
        this._walkIsEnabled(this, value);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Panel}
   * @param {GraphObject} obj
   * @param {boolean} value
   */
  private _walkIsEnabled(obj: GraphObject, value: boolean): void {
    const func = obj.enabledChanged;
    if (func !== null) func(obj, value);
    if (obj instanceof Panel) {
      const a = obj._elements._arr;
      const l = a.length;
      for (let i = 0; i < l; i++) {
        const child = a[i];
        // if enabling, don't update Panels that are !isEnabled (or their children)
        if (value && child instanceof Panel && !child.isEnabled) continue;
        this._walkIsEnabled(child, value);
      }
    }
  }

  /**
   * For Panels which are elements of Spot Panels: Gets or sets the name
   * of this Panel's element that should be used as the alignment object instead of this Panel.
   *
   * This allows Spot Panels to align objects that are nested in the element tree of its own elements.
   * @since 1.7
   */
  /** @export */ get alignmentFocusName(): string { return this._alignmentFocusName; }
  /** @export */ set alignmentFocusName(value: string) {
    const old = this._alignmentFocusName;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Panel, 'alignmentFocusName');
      this._alignmentFocusName = value;
      this.invalidateMeasure();
      this.raiseChanged('alignmentFocusName', old, value);
    }
  }

  /**
   * @hidden @internal
   * Map of PanelLayouts, used when building from source.
   */
  public static Layouts: Map<string, PanelLayout> = new Map<string, PanelLayout>();

  /**
   * Register a PanelLayout. This is called when making new Panel types.
   * See the <a href="../../samples/panelLayout.html">PanelLayout sample</a> for an example.
   * @param {string} layoutName Panel name
   * @param {PanelLayout} layout instance of the PanelLayout
   */
  public static definePanelLayout(layoutName: string, layout: PanelLayout): void {
    Panel.Layouts.add(layoutName, layout);
  }

  /**
   * The default {@link #type}, arranges each element according to their {@link GraphObject#position}.
   * @constant
   */
/** @export */   public static get Position(): PanelLayout {
    return Panel.Layouts.getValue('Position');
  }

  /**
   * This value for {@link #type} lays out the elements horizontally with
   * their {@link GraphObject#alignment} property dictating their alignment on the Y-axis.
   * @constant
   */
/** @export */   public static get Horizontal(): PanelLayout {
    return Panel.Layouts.getValue('Horizontal');
  }

  /**
   * This value for {@link #type} lays out the elements vertically with
   * their {@link GraphObject#alignment} property dictating their alignment on the X-axis.
   * @constant
   */
/** @export */   public static get Vertical(): PanelLayout {
    return Panel.Layouts.getValue('Vertical');
  }

  /**
   * This value for {@link #type} arranges GraphObjects about a main element using the
   * {@link GraphObject#alignment} and {@link GraphObject#alignmentFocus} properties;
   * the main element is the first GraphObject with {@link GraphObject#isPanelMain} set to true,
   * or else the first GraphObject if none have that property set to true.
   * @constant
   */
/** @export */   public static get Spot(): PanelLayout {
    return Panel.Layouts.getValue('Spot');
  }

  /**
   * This value for {@link #type} resizes the main element to fit around the other elements;
   * the main element is the first GraphObject with {@link GraphObject#isPanelMain} set to true,
   * or else the first GraphObject if none have that property set to true.
   * @constant
   */
/** @export */   public static get Auto(): PanelLayout {
    return Panel.Layouts.getValue('Auto');
  }

  /**
   * This value for {@link #type} arranges GraphObjects into rows and columns;
   * set the {@link GraphObject#row} and {@link GraphObject#column} properties on each element.
   * @constant
   */
/** @export */   public static get Table(): PanelLayout {
    return Panel.Layouts.getValue('Table');
  }

  /**
   * This value for {@link #type} rescales a single GraphObject to fit inside the panel
   * depending on the element's {@link GraphObject#stretch} property.
   * @constant
   */
/** @export */   public static get Viewbox(): PanelLayout {
    return Panel.Layouts.getValue('Viewbox');
  }

  /**
   * Organizational Panel type that is only valid inside of a Table panel;
   * This Panel ignores its angle and scale, and does not  have a meaningful size on its own,
   * it is only an organizational container for other elements of a Panel.
   * @constant
   * @since 1.1
   */
/** @export */   public static get TableRow(): PanelLayout {
    return Panel.Layouts.getValue('TableRow');
  }

  /**
   * Organizational Panel type that is only valid inside of a Table panel;
   * This Panel ignores its angle and scale, and does not  have a meaningful size on its own,
   * it is only an organizational container for other elements of a Panel.
   * @constant
   * @since 1.1
   */
/** @export */   public static get TableColumn(): PanelLayout {
    return Panel.Layouts.getValue('TableColumn');
  }

  /**
   * This value for {@link #type} is used for {@link Link}s and adornments that act as {@link Link}s.
   * @constant
   */
/** @export */   public static get Link(): PanelLayout {
    return Panel.Layouts.getValue('Link');
  }

  /**
   * This value for {@link #type} is used to draw regular patterns of lines.
   * @constant
   */
/** @export */   public static get Grid(): PanelLayout {
    return Panel.Layouts.getValue('Grid');
  }

  /**
   * This value for {@link #type} is used to draw regular tick marks and labels along some shape.
   * The main element is the first GraphObject with {@link GraphObject#isPanelMain} set to true,
   * or else the first GraphObject if none have that property set to true.
   * @constant
   * @since 1.7
   */
/** @export */   public static get Graduated(): PanelLayout {
    return Panel.Layouts.getValue('Graduated');
  }

} // #end Panel
Panel['className'] = 'Panel';
Panel['definePanelLayout'] = Panel.definePanelLayout;

// import { PanelLayout } from './panelLayout';
// import { PanelLayoutAuto } from './panelLayoutAuto';
// import { GraduatedMark } from './panelLayoutGraduated';
// import { PanelLayoutLink } from './panelLayoutLink';
// import { PanelLayoutPosition } from './panelLayoutPosition';
// import { PanelLayoutVertical } from './panelLayoutVertical';
// // import { PanelLayoutGrid } from './panelLayoutGrid';
// // import { PanelLayoutHorizontal } from './panelLayoutHorizontal';
// // import { PanelLayoutSpot } from './panelLayoutSpot';
// // import { PanelLayoutTable, PanelLayoutTableColumn, PanelLayoutTableRow } from './panelLayoutTable';
// // import { PanelLayoutViewbox } from './panelLayoutViewbox';

// These PanelLayouts are essential to GoJs:
Panel.definePanelLayout('Position', new PanelLayoutPosition());
Panel.definePanelLayout('Vertical', new PanelLayoutVertical());
Panel.definePanelLayout('Auto', new PanelLayoutAuto());
Panel.definePanelLayout('Link', new PanelLayoutLink());
// By default, these are added to the library in index.ts. When building from source, you must add them manually.
// Panel.definePanelLayout('Horizontal', new PanelLayoutHorizontal());
// Panel.definePanelLayout('Spot', new PanelLayoutSpot());
// Panel.definePanelLayout('Table', new PanelLayoutTable());
// Panel.definePanelLayout('Viewbox', new PanelLayoutViewbox());
// Panel.definePanelLayout('TableRow', new PanelLayoutTableRow());
// Panel.definePanelLayout('TableColumn', new PanelLayoutTableColumn());
// Panel.definePanelLayout('Graduated', new PanelLayoutGraduated());
// Panel.definePanelLayout('Grid', new PanelLayoutGrid());

// import { Brush, BrushLike } from './brush';
// import { ChangedEvent } from './changedEvent';
// import { Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { Geo } from './geo';
// import { Margin, MarginLike } from './margin';
// import { Binding } from './model';
// import { Node } from './parts';
// import { Adornment, Part } from './parts';
// import { Placeholder } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData } from './root';
// import { RowColumnDefinition } from './rowColumnDefinition';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { IContext } from './surfaceInterfaces';
// import { TextBlock } from './textBlock';
// import { Transform } from './transform';


/*
Panel['Position'] = Panel.Position = new PanelLayoutPosition();
Panel['Horizontal'] = Panel.Horizontal = new PanelLayoutHorizontal();
Panel['Vertical'] = Panel.Vertical = new PanelLayoutVertical();
Panel['Spot'] = Panel.Spot = new PanelLayoutSpot();
Panel['Auto'] = Panel.Auto = new PanelLayoutAuto();
Panel['Table'] = Panel.Table = new PanelLayoutTable();
Panel['Viewbox'] = Panel.Viewbox = new PanelLayoutViewbox();
Panel['Link'] = Panel.Link = new PanelLayoutLink();
Panel['Grid'] = Panel.Grid = new PanelLayoutGrid();
Panel['TableRow'] = Panel.TableRow = new PanelLayoutTableRow();
Panel['TableColumn'] = Panel.TableColumn = new PanelLayoutTableColumn();
Panel['Graduated'] = Panel.Graduated = new PanelLayoutGraduated();
*/



// import { List, Set } from './collections';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { Util } from './util';


/**
 * The RowColumnDefinition class describes constraints on a row or a column
 * in a {@link Panel} of type {@link Panel.Table}.
 * It also provides information about the actual layout after the
 * Table Panel has been arranged.
 * @unrestricted
 */
export class RowColumnDefinition {
  /** @hidden @internal */ public static className = 'RowColumnDefinition';
  /** @hidden @internal */ private _panel: Panel | null;
  /** @hidden @internal */ private _isRow: boolean;
  /** @hidden @internal */ private _index: number;
  /** @hidden @internal */ private _minimum: number;
  /** @hidden @internal */ private _maximum: number;
  /** @hidden @internal */ private _alignment: Spot;
  /** @hidden @internal */ private _position: number;
  /** @hidden @internal */ private _sizing: EnumValue;
  /** @hidden @internal */ private _stretch: EnumValue;
  /** @hidden @internal */ private _separatorPadding: Margin;
  /** @hidden @internal */ private _separatorStroke: BrushLike;
  /** @hidden @internal */ private _separatorStrokeWidth: number;
  /** @hidden @internal */ private _separatorDashArray: Array<number> | null;
  /** @hidden @internal */ private _background: BrushLike;
  /** @hidden @internal */ private _coversSeparators: boolean;
  /** @hidden @internal */ public _binds: List<Binding> | null;
  /** @hidden @internal */ public _actual: number;
  /** @hidden @internal */ public _size: number;

  /**
   * You should not use this constructor, because calls to
   * {@link Panel#getRowDefinition} or {@link Panel#getColumnDefinition}
   * will automatically create and remember a RowColumnDefinition for you.
   */
  constructor() {
    Set.uniqueHash(this);
    this._panel = null;  // set only by Panel.getRow/ColumnDefinition
    this._isRow = true;  // set only by Panel.getRow/ColumnDefinition
    this._index = 0;  // set only by Panel.getRow/ColumnDefinition
    this._size = NaN;  // finite, >= 0
    this._minimum = 0;  // finite, >= 0
    this._maximum = Infinity;  // >= 0
    this._alignment = Spot.Default;
    this._actual = 0;  // valid after arrange
    this._position = 0;  // valid after arrange
    this._binds = null;
    this._sizing = RowColumnDefinition.Default;
    this._stretch = GraphObject.Default;

    this._separatorPadding = null;
    this._separatorStroke = null;
    this._separatorStrokeWidth = NaN;
    this._separatorDashArray = null;

    this._background = null;
    this._coversSeparators = false;
  }

  /**
   * @hidden @internal
   * Create a copy of this RowColumnDefinition, with the same values.
   * @this {RowColumnDefinition}
   * @return {RowColumnDefinition}
   */
  public copy(): RowColumnDefinition {
    const copy = new RowColumnDefinition();
    // new Panel needs to set _panel
    copy._isRow = this._isRow;
    copy._index = this._index;
    copy._size = this._size;
    copy._minimum = this._minimum;
    copy._maximum = this._maximum;
    copy._alignment = this._alignment;
    copy._actual = this._actual;
    copy._position = this._position;
    copy._stretch = this._stretch;
    copy._sizing = this._sizing;
    if (this._separatorPadding === null) copy._separatorPadding = null;
    else copy._separatorPadding = this._separatorPadding.copyFrozen();

    copy._separatorStroke = this._separatorStroke; // OK to share, should be frozen
    copy._separatorStrokeWidth = this._separatorStrokeWidth;
    copy._separatorDashArray = null;
    if (this._separatorDashArray !== null) copy.separatorDashArray = Util.copyArray(this.separatorDashArray);
    copy._background = this._background; // OK to share, should be frozen
    copy._coversSeparators = this._coversSeparators; // OK to share, should be frozen
    copy._binds = this._binds;  // share any bindings
    return copy;
  }

  /**
   * @hidden @internal
   * Modify this RowColumnDefinition by copying the settable properties from another RowColumnDefinition.
   * This is only called by GraphObject.make.
   * @this {RowColumnDefinition}
   * @param {RowColumnDefinition} rcd The RowColumnDefinition whose properties should be copied into this one.
   */
  public copyFrom(rcd: RowColumnDefinition) {
    Util.checkClass(rcd, RowColumnDefinition, RowColumnDefinition, 'copyFrom:pd');
    // use setters to call raiseChanged!
    if (rcd.isRow) {
      this.height = rcd.height;
    } else {
      this.width = rcd.width;
    }
    this.minimum = rcd.minimum;
    this.maximum = rcd.maximum;
    this.alignment = rcd.alignment;
    this.stretch = rcd.stretch;
    this.sizing = rcd.sizing;
    this._separatorPadding = rcd.separatorPadding === null ? null : (rcd.separatorPadding as Margin).copyFrozen();
    this.separatorStroke = rcd.separatorStroke; // OK to share, should be frozen
    this.separatorStrokeWidth = rcd.separatorStrokeWidth; // OK to share, should be frozen
    this._separatorDashArray = null;
    if (rcd.separatorDashArray) this._separatorDashArray = Util.copyArray(rcd.separatorDashArray);
    this.background = rcd.background; // OK to share, should be frozen
    this.coversSeparators = rcd.coversSeparators;

    this._binds = rcd._binds;  // share any bindings
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {RowColumnDefinition}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue) {
    if (arg.classType === RowColumnDefinition) {
      this.sizing = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * @hidden @internal
   * @this {RowColumnDefinition}
   * @return {string}
   */
  public toString(): string {
    return 'RowColumnDefinition ' +
      (this.isRow ? '(Row ' : '(Column ') +
      this.index +
      ') #' + Set.hashId(this);
  }

  /**
   * The default {@link #sizing}, which resolves to {@link RowColumnDefinition.None} or else
   * the Table Panel's rowSizing and columnSizing if present.
   * @constant
   */
  public static Default = new EnumValue(RowColumnDefinition, 'Default', 0);

  /**
   * The default {@link #sizing} if none is specified on the Table Panel's rowSizing and columnSizing.
   * @constant
   */
  public static None = new EnumValue(RowColumnDefinition, 'None', 1);

  /**
   * If a Table Panel is larger than all the rows then this {@link #sizing}
   * grants this row and any others with the same value the extra space, apportioned proportionally between them
   * @constant
   */
  public static ProportionalExtra = new EnumValue(RowColumnDefinition, 'ProportionalExtra', 2);

  /**
   * @hidden @internal
   * @this {RowColumnDefinition}
   * @param {Panel} val
   */
  public setPanel(val: Panel) {
    this._panel = val;
  }

  //    * Get the spacing at the top or left of the row or column.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {RowColumnDefinition}
   * @return {number}
   * @see {@link #computeEffectiveSpacing}
   */
  public computeEffectiveSpacingTop(): number {
    let sw = 0.0;
    let first = 0;
    const panel = this._panel;
    const isRow = this.isRow;
    if (panel !== null && panel.type === Panel.Table) {
      const l = isRow ? panel._rowDefs.length : panel._colDefs.length;
      for (let i = 0; i < l; i++) {
        const coldef = isRow ? panel._rowDefs[i] : panel._colDefs[i];
        if (coldef === undefined) continue;
        first = coldef.index;
        break;
      }
    }

    if (this.index !== first) {  // First rows/columns have no separator
      let sk = this.separatorStroke;
      if (sk === null && panel !== null) sk = this.isRow ? panel.defaultRowSeparatorStroke : panel.defaultColumnSeparatorStroke;
      if (sk !== null) {
        sw = this.separatorStrokeWidth;
        if (isNaN(sw)) {
          if (panel !== null) sw = this.isRow ? panel.defaultRowSeparatorStrokeWidth : panel.defaultColumnSeparatorStrokeWidth;
          else sw = 0;
        }
      }
    }
    let t = this._separatorPadding;
    if (t === null) {
      if (panel !== null) t = panel.defaultSeparatorPadding as Margin;
      else return sw;
    }
    return sw + (this.isRow ? t.top : t.left);
  }

  //    * Get the total spacing for a row or column.
  //  *
  //  * For a row, this would be the value of {@link Panel#defaultRowSeparatorStrokeWidth} or {@link RowColumnDefinition#separatorStrokeWidth},
  //  * plus the top and bottom values of {@link Panel#defaultSeparatorPadding} or {@link RowColumnDefinition#separatorPadding}.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {RowColumnDefinition}
   * @return {number}
   */
  public computeEffectiveSpacing(): number {
    let sw = 0.0;
    const panel = this._panel;
    let first = 0;
    const isRow = this.isRow;
    if (panel !== null && panel.type === Panel.Table) {
      const l = isRow ? panel._rowDefs.length : panel._colDefs.length;
      for (let i = 0; i < l; i++) {
        const coldef = isRow ? panel._rowDefs[i] : panel._colDefs[i];
        if (coldef === undefined) continue;
        first = coldef.index;
        break;
      }
    }

    if (this.index !== first) {  // First rows/columns have no separator
      let sk = this.separatorStroke;
      if (sk === null && panel !== null) sk = isRow ? panel.defaultRowSeparatorStroke : panel.defaultColumnSeparatorStroke;
      if (sk !== null) {
        sw = this.separatorStrokeWidth;
        if (isNaN(sw)) {
          if (panel !== null) sw = isRow ? panel.defaultRowSeparatorStrokeWidth : panel.defaultColumnSeparatorStrokeWidth;
          else sw = 0;
        }
      }
    }
    let t = this._separatorPadding;
    if (t === null) {
      if (panel !== null) t = panel.defaultSeparatorPadding as Margin;
      else return sw;
    }
    return sw + (this.isRow ? t.top + t.bottom : t.left + t.right);
  }

  /**
   * @hidden @internal
   * @this {RowColumnDefinition}
   * @param {string} pname names the property that was modified.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  private _raiseChanged(pname: string, oldval: any, newval: any, oldparam?: any, newparam?: any) {
    const panel = this._panel;
    if (panel === null) return;

    panel.raiseChangedEvent(ChangedEvent.Property, pname, this, oldval, newval, oldparam, newparam);
    if (this._binds !== null) {
      const diagram = panel.diagram;
      if (diagram !== null && !diagram.skipsModelSourceBindings) {  // assume no TwoWay Binding.ofObject
        const binder = panel.findTemplateBinder();
        if (binder !== null) {
          const data = binder.data;
          if (data !== null) {
            const it = this._binds.iterator;
            while (it.next()) {
              const bind = it.value;
              bind.updateSource(this, data, pname, binder);
              // ??? no support for TwoWay Binding.ofObject(...)
            }
          }
        }
      }
    }
  }

  /**
   * This read-only property returns the {@link Panel} that this row or column definition is in.
   */
  /** @export */ get panel(): Panel | null { return this._panel; }

  /**
   * This read-only property is true when this describes a row instead of a column in the {@link #panel}.
   * When this is true, the {@link #height}, {@link #minimum}, and {@link #maximum}
   * all describe the row height.
   * Otherwise {@link #width} and the other two properties describe the column width.
   */
  /** @export */ get isRow(): boolean { return this._isRow; }
  /** @export */ set isRow(value: boolean) { this._isRow = value; }

  /**
   * This read-only property returns which row or column this RowColumnDefinition describes in the {@link #panel}.
   * The value is a zero-based integer.
   * @see {@link Panel#findColumnForLocalX}
   * @see {@link Panel#findRowForLocalY}
   * @see {@link Panel#leftIndex}
   * @see {@link Panel#topIndex}
   * @see {@link #position}
   */
  /** @export */ get index(): number { return this._index; }
  /** @export */ set index(value: number) { this._index = value; }

  /**
   * Gets or sets the row height, in local coordinates.
   * This describes the row height if {@link #isRow} is true;
   * otherwise this property is meaningless.
   * The value must be non-negative and finite.
   *
   * The default value is NaN, which means this row will get a height
   * that is just big enough to hold all of the objects in the row.
   *
   * Setting this value to a number will mean that all of the objects
   * of this {@link Panel} in this row will be allocated that
   * amount of row height.
   * Whether an object in the row is actually arranged to have that height
   * depends on whether the {@link GraphObject#stretch} stretches vertically.
   * @see {@link #width}
   * @see {@link #minimum}
   * @see {@link #maximum}
   * @see {@link #sizing}
   */
  /** @export */ get height(): number { return this._size; }
  /** @export */ set height(value: number) {
    const old = this._size;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', RowColumnDefinition, 'height');
      if (value < 0) Util.throwRangeError(value, '>= 0', RowColumnDefinition, 'height');
      this._size = value;
      this.actual = this._actual;  // ensure between minimum and maximum
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('height', old, value);
    }
  }

  /**
   * Gets or sets the column width, in local coordinates.
   * The size describes the column width if {@link #isRow} is false;
   * otherwise this property is meaningless.
   * The value must be non-negative and finite.
   *
   * The default value is NaN, which means this column will get a width
   * that is just big enough to hold all of the objects in the column.
   *
   * Setting this value to a number will mean that all of the objects
   * of this {@link Panel} in this column will be allocated that
   * amount of column width.
   * Whether an object in the column is actually arranged to have that width
   * depends on whether the {@link GraphObject#stretch} stretches horizontally.
   * @see {@link #height}
   * @see {@link #minimum}
   * @see {@link #maximum}
   * @see {@link #sizing}
   */
  /** @export */ get width(): number { return this._size; }
  /** @export */ set width(value: number) {
    const old = this._size;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', RowColumnDefinition, 'width');
      if (value < 0) Util.throwRangeError(value, '>= 0', RowColumnDefinition, 'width');
      this._size = value;
      this.actual = this._actual;  // ensure between minimum and maximum
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('width', old, value);
    }
  }

  /**
   * Gets or sets the minimum row height or column width, in local coordinates.
   * The minimum describes the row height if {@link #isRow} is true;
   * otherwise this describes the column width.
   *
   * The value must be non-negative and finite.
   * The default value is zero.
   * The arranged height of all objects in this row,
   * or the arranged width of all objects in this column,
   * will be no less than this value.
   * @see {@link #width}
   * @see {@link #height}
   * @see {@link #maximum}
   * @see {@link #sizing}
   */
  /** @export */ get minimum(): number { return this._minimum; }
  /** @export */ set minimum(value: number) {
    const old = this._minimum;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', RowColumnDefinition, 'minimum');
      if (value < 0 || !isFinite(value)) Util.throwRangeError(value, '>= 0', RowColumnDefinition, 'minimum');
      this._minimum = value;
      this.actual = this._actual;  // ensure between minimum and maximum
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('minimum', old, value);
    }
  }

  /**
   * Gets or sets the maximum row height or column width, in local coordinates.
   * The maximum describes the row height if {@link #isRow} is true;
   * otherwise this describes the column width.
   *
   * The value must be non-negative.
   * The default value is Infinity.
   * The arranged height of all objects in this row,
   * or the arranged width of all objects in this column,
   * will be no greater than this value.
   * @see {@link #width}
   * @see {@link #height}
   * @see {@link #minimum}
   * @see {@link #sizing}
   */
  /** @export */ get maximum(): number { return this._maximum; }
  /** @export */ set maximum(value: number) {
    const old = this._maximum;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', RowColumnDefinition, 'maximum');
      if (value < 0) Util.throwRangeError(value, '>= 0', RowColumnDefinition, 'maximum');
      this._maximum = value;
      this.actual = this._actual;  // ensure between minimum and maximum
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('maximum', old, value);
    }
  }

  /**
   * Gets or sets a default alignment for elements that are in this row or column.
   * The value must be a {@link Spot}.
   * The default value is {@link Spot.Default}, so that this RowColumnDefinition
   * does not supply any alignment information for the row or column.
   *
   * When an element's {@link GraphObject#alignment} property is {@link Spot.Default},
   * it gets the horizontal alignment from the element's column's {@link RowColumnDefinition}
   * and the vertical alignment from the element's row's RowColumnDefinition.
   * When that {@link RowColumnDefinition#alignment} property is also Default,
   * it takes the value from the table panel's {@link Panel#defaultAlignment} property.
   * @see {@link GraphObject#alignment}
   * @see {@link Panel#defaultAlignment}
   */
  /** @export */ get alignment(): Spot { return this._alignment; }
  /** @export */ set alignment(value: Spot) {
    const old = this._alignment;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, RowColumnDefinition, 'alignment');
      this._alignment = value.copyFrozen();
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('alignment', old, value);
    }
  }

  /**
   * Gets or sets the default stretch for elements that are in this row or column.
   * The only accepted values are listed as constant properties of GraphObject:
   * {@link GraphObject.None}, {@link GraphObject.Fill}, {@link GraphObject.Horizontal}, {@link GraphObject.Vertical}, or {@link GraphObject.Default}.
   * The default value is {@link GraphObject.Default}, so that this RowColumnDefinition
   * does not supply any stretch information for the row or column.
   *
   * When an element's {@link GraphObject#stretch} property is {@link GraphObject.Default},
   * it gets the horizontal stretch from the element's column's {@link RowColumnDefinition}
   * and the vertical stretch from the element's row's RowColumnDefinition.
   * When that {@link RowColumnDefinition#stretch} property is also Default,
   * it takes the value from the table panel's {@link Panel#defaultStretch} property.
   * @see {@link GraphObject#stretch}
   * @see {@link Panel#defaultStretch}
   * @see {@link #sizing}
   * @see {@link #width}
   * @see {@link #height}
   */
  /** @export */ get stretch(): EnumValue { return this._stretch; }
  /** @export */ set stretch(value: EnumValue) {
    const old = this._stretch;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, GraphObject, RowColumnDefinition, 'stretch');
      this._stretch = value;
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('stretch', old, value);
    }
  }

  /**
   * Gets or sets the additional padding for a particular row or column, in local coordinates.
   * Padding is applied on two sides - before and after a row or column's contents.
   * The {@link #separatorStrokeWidth} comes before any padding {@link Margin#top} or {@link Margin#left}.
   *
   * The default value is null, so that this RowColumnDefinition
   * does not supply any padding information for the row or column.
   * There may still be some padding between rows and/or columns if {@link Panel#defaultSeparatorPadding} is non-zero.
   * @see {@link Panel#defaultSeparatorPadding}
   * @see {@link #separatorStroke}
   * @see {@link #separatorStrokeWidth}
   * @see {@link #separatorDashArray}
   * @see {@link #background}
   * @since 1.2
   */
  /** @export */ get separatorPadding(): MarginLike { return this._separatorPadding; }
  /** @export */ set separatorPadding(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else if (value !== null) {
      if (Debug) Util.checkClass(value, Margin, RowColumnDefinition, 'separatorPadding');
    }
    const old = this._separatorPadding;
    if (value === null || old === null || !old.equals(value)) {
      if (value !== null) value = value.copyFrozen();
      this._separatorPadding = value;
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('separatorPadding', old, value);
    }
  }

  /**
   * Gets or sets the stroke (color) for the separator line that is drawn before a particular row or column,
   * provided that row or column has a nonzero {@link #separatorStrokeWidth}.
   *
   * The default value is null, so that this RowColumnDefinition
   * does not specify any brush for the separator line to draw before the row or column.
   * The line may still be drawn if the value of {@link Panel#defaultRowSeparatorStroke} or
   * {@link Panel#defaultColumnSeparatorStroke} is non-null.
   *
   * However, if no stroke color is specified, then no stroke width will be drawn,
   * even if the {@link #separatorStrokeWidth} value is non-zero.
   * @see {@link Panel#defaultColumnSeparatorStroke}
   * @see {@link Panel#defaultRowSeparatorStroke}
   * @see {@link #separatorPadding}
   * @see {@link #separatorStrokeWidth}
   * @see {@link #separatorDashArray}
   * @see {@link #background}
   * @since 1.2
   */
  /** @export */ get separatorStroke(): BrushLike { return this._separatorStroke; }
  /** @export */ set separatorStroke(value: BrushLike) {
    const old = this._separatorStroke;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'RowColumnDefinition.separatorStroke');
      if (value instanceof Brush) value.freeze();
      this._separatorStroke = value;
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('separatorStroke', old, value);
    }
  }

  /**
   * Gets or sets the stroke width for a particular row or column's separator line, in local coordinates.
   *
   * The default value is NaN, so that this RowColumnDefinition
   * does not supply any stroke width information for what is drawn before the row or column.
   * The separator line may still be drawn if {@link Panel#defaultRowSeparatorStroke}
   * or {@link Panel#defaultColumnSeparatorStroke} is a real number (they default to 1).
   * @see {@link Panel#defaultColumnSeparatorStrokeWidth}
   * @see {@link Panel#defaultRowSeparatorStrokeWidth}
   * @see {@link #separatorPadding}
   * @see {@link #separatorStroke}
   * @see {@link #separatorDashArray}
   * @see {@link #background}
   * @since 1.2
   */
  /** @export */ get separatorStrokeWidth(): number { return this._separatorStrokeWidth; }
  /** @export */ set separatorStrokeWidth(value: number) {
    const old = this._separatorStrokeWidth;
    if (old !== value) {
      this._separatorStrokeWidth = value;
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('separatorStrokeWidth', old, value);
    }
  }

  /**
   * Gets or sets the dash array for dashing the separator line, provided this
   * row or column has a nonzero {@link RowColumnDefinition#separatorStrokeWidth} and non-null {@link RowColumnDefinition#separatorStroke}.
   *
   * Must be an array of positive numbers and zeroes,
   * or else null to indicate a solid line.
   *
   * For example, the array [5, 10] would create dashes of 5 pixels and spaces of 10 pixels.
   *
   * Setting an array with all zeroes will set the value to null.
   *
   * Default is null, so that this RowColumnDefinition
   * does not supply any stroke dash array information for what is drawn before the row or column.
   * The separator line may still be drawn using dashes if {@link Panel#defaultRowSeparatorDashArray}
   * or {@link Panel#defaultColumnSeparatorDashArray} is non-null.
   * @see {@link Panel#defaultColumnSeparatorDashArray}
   * @see {@link Panel#defaultRowSeparatorDashArray}
   * @see {@link #separatorPadding}
   * @see {@link #separatorStroke}
   * @see {@link #separatorStrokeWidth}
   * @see {@link #background}
   * @since 1.2
   */
  /** @export */ get separatorDashArray(): Array<number> { return this._separatorDashArray; }
  /** @export */ set separatorDashArray(value: Array<number>) {
    const old = this._separatorDashArray;
    if (old !== value) {
      // make sure its an array of numbers?
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', RowColumnDefinition, 'separatorDashArray:value');
      if (value !== null) {
        const l = value.length;
        let total = 0;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (typeof n !== 'number' || !(n >= 0) || !isFinite(n)) {
            Util.throwError('separatorDashArray value ' + n + ' at index ' + i + ' must be a positive number or zero.');
          }
          total += n;
        }
        if (total === 0) {
          if (old === null) return; // no-op
          value = null; // never let separatorDashArray be all zeroes, ie [0, 0]
        }
      }

      this._separatorDashArray = value;
      if (this.panel !== null) this.panel.invalidateDraw();
      this._raiseChanged('separatorDashArray', old, value);
    }
  }

  /**
   * Gets or sets the background color for a particular row or column,
   * which fills the entire span of the row or column, including any separatorPadding.
   *
   * The default value is null, which means nothing is drawn in the background of the row or column.
   * @see {@link GraphObject#background}
   * @since 1.2
   */
  /** @export */ get background(): BrushLike { return this._background; }
  /** @export */ set background(value: BrushLike) {
    const old = this._background;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'RowColumnDefinition.background');
      if (value instanceof Brush) value.freeze();
      this._background = value;
      if (this.panel !== null) this.panel.invalidateDraw();
      this._raiseChanged('background', old, value);

    }
  }

  /**
   * Determines whether or not the {@link #background}, if there is one, is in front of or behind the separators.
   *
   * The default value is false -- any background is drawn behind any separator lines.
   * @since 1.2
   */
  /** @export */ get coversSeparators(): boolean { return this._coversSeparators; }
  /** @export */ set coversSeparators(value: boolean) {
    const old = this._coversSeparators;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', RowColumnDefinition, 'coversSeparators');
      this._coversSeparators = value;
      if (this.panel !== null) this.panel.invalidateDraw();
      this._raiseChanged('coversSeparators', old, value);
    }
  }

  /**
   * Gets or sets how this row or column deals with a Table Panel's extra space.
   * The value must be one of: {@link RowColumnDefinition.None}, {@link RowColumnDefinition.ProportionalExtra},
   * or {@link RowColumnDefinition.Default}.
   * The default value is {@link RowColumnDefinition.Default}.
   * @see {@link Panel#columnSizing}
   * @see {@link Panel#rowSizing}
   * @see {@link #actual}
   * @see {@link #total}
   * @see {@link #height}
   * @see {@link #width}
   */
  /** @export */ get sizing(): EnumValue { return this._sizing; }
  /** @export */ set sizing(value: EnumValue) {
    const old = this._sizing;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, RowColumnDefinition, RowColumnDefinition, 'sizing');
      this._sizing = value;
      if (this.panel !== null) this.panel.invalidateMeasure();
      this._raiseChanged('sizing', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {RowColumnDefinition}
   * @return {EnumValue}
   */
  public getEffectiveSizing(): EnumValue {
    if (this.sizing === RowColumnDefinition.Default) {
      const p = this._panel;
      if (this.isRow) return p.rowSizing;
      else return p.columnSizing;
    }
    return this.sizing;
  }

  /**
   * This read-only property returns the usable row height or column width, after arrangement, in local coordinates,
   * that objects in this row or column can be arranged within.
   * This does not include {@link #separatorPadding} or {@link #separatorStrokeWidth}, as {@link #total} does.
   *
   * This value gives the row height if {@link #isRow} is true;
   * otherwise this gives the column width.
   * The value is meaningless until after the Table Panel using this
   * RowColumnDefinition has been arranged.
   * @see {@link #total}
   * @see {@link #height}
   * @see {@link #width}
   * @see {@link #sizing}
   */
  /** @export */ get actual(): number { return this._actual; }
  /** @export */ set actual(value: number) {
    if (!isNaN(this._size)) { // ignore any computed value if the width or height was specified
      this._actual = Math.max(Math.min(this._maximum, this._size), this._minimum);
    } else {
      this._actual = Math.max(Math.min(this._maximum, value), this._minimum);
    }
  }

  /**
   * This read-only property returns the total arranged row height or column width, after arrangement, in local coordinates.
   * This value gives the {@link #actual} size plus the {@link #separatorPadding} and {@link #separatorStrokeWidth}.
   *
   * This value gives the vertical space occupied by the row if {@link #isRow} is true;
   * otherwise this gives the horizontal space occupied by the column.
   * The value is meaningless until after the Table Panel using this
   * RowColumnDefinition has been arranged.
   * @see {@link #actual}
   * @see {@link #height}
   * @see {@link #width}
   * @see {@link #sizing}
   * @since 1.2
   */
  /** @export */ get total(): number { return this._actual + this.computeEffectiveSpacing(); }
  /** @export */ set total(value: number) {
    if (!isNaN(this._size)) { // ignore any computed value if the width or height was specified
      this._actual = Math.max(Math.min(this._maximum, this._size), this._minimum);
    } else {
      this._actual = Math.max(Math.min(this._maximum, value), this._minimum);
    }
    this._actual = Math.max(0, this._actual - this.computeEffectiveSpacing());
  }

  /**
   * This read-only property returns the actual arranged row or column starting position, after arrangement, in local coordinates.
   * This value gives the Y position if {@link #isRow} is true;
   * otherwise this gives the X position.
   * The value is meaningless until after the Table Panel using this
   * RowColumnDefinition has been arranged.
   * @see {@link Panel#findColumnForLocalX}
   * @see {@link Panel#findRowForLocalY}
   * @see {@link Panel#leftIndex}
   * @see {@link Panel#topIndex}
   * @see {@link #index}
   */
  /** @export */ get position(): number { return this._position; }
  /** @export */ set position(value: number) { this._position = value; }

  /**
   * Add a data-binding of a property on this object to a property on a data object.
   * @export
   * @this {RowColumnDefinition}
   * @param {Binding} binding
   */
  public bind(binding: Binding) {
    binding._targetObject = this;  // remember so that Panel.freezeBindings can generalize using name-lookup
    const panel = this.panel;
    if (panel !== null) {  // maybe in a visual tree already
      const binder = panel.findTemplateBinder();
      if (binder !== null && binder.isTemplateBinderFrozen()) {
        Util.throwError('Cannot add a Binding to a RowColumnDefinition that is already frozen: ' + binding + ' on ' + panel);
      }
    }
    // just store here for later collection by freezeBindings
    if (this._binds === null) this._binds = new List<Binding>();
    this._binds.add(binding);
  }

} // #end RowColumnDefinition
RowColumnDefinition['className'] = 'RowColumnDefinition';

RowColumnDefinition['Default'] = RowColumnDefinition.Default;
RowColumnDefinition['None'] = RowColumnDefinition.None;
RowColumnDefinition['ProportionalExtra'] = RowColumnDefinition.ProportionalExtra;


// import { Brush, BrushLike } from './brush';
// import { ChangedEvent } from './changedEvent';
// import { Debug } from './debug';
// import { Margin, MarginLike } from './margin';
// import { Binding } from './model';
// import { Panel } from './panel';
// import { Spot } from './spot';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Shape
*/

// import { Map, Set } from './collections';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { PropertyCollection } from './propertyCollection';
// import { Util } from './util';

/**
 * A Shape is a {@link GraphObject} that shows a geometric figure.
 * The {@link Geometry} determines what is drawn;
 * the properties {@link #fill} and {@link #stroke}
 * (and other stroke properties) determine how it is drawn.
 *
 * There are generally two types of shapes: Those that use a custom {@link Geometry} by setting
 * {@link Shape#geometry}, and those that receive an automatically generated Geometry using the value of
 * {@link #figure}, {@link #toArrow}, or {@link #fromArrow}. An explicitly set Geometry always supersedes
 * the figure and arrowhead properties.
 *
 * Some created Shapes:
 * ```js
 * var $ = go.GraphObject.make;  // for conciseness in defining GraphObjects
 *
 * // A shape with the figure set to RoundedRectangle:
 * $(go.Shape, { figure: "RoundedRectangle", fill: "lightgreen" })
 * // Alternatively:
 * $(go.Shape, "RoundedRectangle", { fill: "lightgreen" })
 *
 * // A shape with a custom geometry, using geometryString:
 * $(go.Shape,
 *     { geometry: go.Geometry.parse("M120 0 L80 80 0 50z") })
 *
 * // A shape with a custom geometry, using geometryString:
 * $(go.Shape,
 *    { geometryString: "F M120 0 L80 80 0 50z",
 *      fill: "lightgreen" })
 *
 * // A common link template, using two shapes,
 * // the first for the link path and the second for the arrowhead
 * myDiagram.linkTemplate =
 *   $(go.Link,
 *     // The first shape in a link is special, its geometry is set by the Link's routing,
 *     // so it does not need a geometry or figure set manually
 *     $(go.Shape,
 *       { strokeWidth: 2, stroke: 'gray' }),
 *     $(go.Shape,
 *       { toArrow: "Standard", fill: 'gray', stroke: null })
 *   );
 * ```
 * <p class="box">
 * You can see more custom geometry examples and read about geometryString
 * on the <a href="../../intro/geometry.html">Geometry Path Strings Introduction page.</a>
 *
 * When automatically generating a Shape {@link Geometry}, the value of {@link #toArrow} takes precedence,
 * then {@link #fromArrow}, then {@link #figure}. If the value of {@link #toArrow} or {@link #fromArrow} is "None"
 * then it is ignored, and the "None" value of {@link #figure} is identical to "Rectangle".
 *
 * All of the predefined figures are shown in the <a href="../../samples/shapes.html">Shapes sample</a>.
 * You can define your own named figures by calling the static function {@link Shape.defineFigureGenerator}.
 * Get a {@link Map} of named figures by calling the static function {@link Shape.getFigureGenerators}.
 *
 * All of the predefined arrowheads are shown in the <a href="../../samples/arrowheads.html">Arrowheads sample</a>.
 * You can define your own named arrowheads by calling the static function {@link Shape.defineArrowheadGeometry}.
 * Get a {@link Map} of named arrowheads by calling the static function {@link Shape.getArrowheadGeometries}.
 *
 * You can see a copy of all of the built-in arrowhead definitions in this file: <a href="../../extensions/Arrowheads.js">Arrowheads.js</a>.
 *
 * The Shape properties {@link #parameter1}, and {@link #parameter2} determine details of the
 * construction of some {@link #figure} geometries.
 * Specifically, they often set the {@link #spot1}, {@link #spot2} for the Shape.
 * These spots determine the "inner area" of an Auto panel when a Shape is the main object.
 * See the <a href="../../intro/panels.html">Auto Panels section of the Panels Introduction page</a> for more details.
 *
 * Shapes use their geometric bounds when determining hit testing,
 * but use rectangular bounds when participating in (panel) layouts.
 * @extends GraphObject
 * @unrestricted
 */
export class Shape extends GraphObject {
  /** @hidden @internal */ public static className = 'Shape';
  /** @hidden @internal */ private _suppliedGeometry: Geometry | null;
  /** @hidden @internal */ private _figure: string;
  /** @hidden @internal */ private _geometryStretch: EnumValue;
  /** @hidden @internal */ private _fill: BrushLike;
  /** @hidden @internal */ private _stroke: BrushLike;
  /** @hidden @internal */ private _strokeWidth: number;
  /** @hidden @internal */ private _strokeCap: ('butt' | 'round' | 'square');
  /** @hidden @internal */ private _strokeJoin: ('miter' | 'bevel' | 'round');
  /** @hidden @internal */ private _strokeMiterLimit: number;
  /** @hidden @internal */ private _strokeDashArray: Array<number> | null;
  /** @hidden @internal */ private _strokeDashOffset: number;
  /** @hidden @internal */ private _spot1: Spot;
  /** @hidden @internal */ private _spot2: Spot;
  /** @hidden @internal */ private _parameter1: number;
  /** @hidden @internal */ private _parameter2: number;
  /** @hidden @internal */ private _isGeometryPositioned: boolean;
  /** @hidden @internal */ private _pathPattern: GraphObject | null;
  /** @hidden @internal */ private _toArrow: string;
  /** @hidden @internal */ private _fromArrow: string;
  /** @hidden @internal */ private _interval: number;
  /** @hidden @internal */ private _graduatedStart: number;
  /** @hidden @internal */ private _graduatedEnd: number;
  /** @hidden @internal */ private _graduatedSkip: ((val: number) => boolean) | null;
  /** @hidden @internal */ public _geometry: Geometry | null;

  /**
   * A newly constructed {@link Shape} has a default {@link #figure} of "None",
   * which constructs a rectangular geometry, and is filled and stroked with a black brush.
   */
  constructor() {
    super();

    this._geometry = null;
    this._suppliedGeometry = null; // used when a custom geometry is set

    this._figure = 'None';
    this._geometryStretch = GraphObject.Default;
    this._fill = 'black';
    this._stroke = 'black'; // color string any well-formed CSS color
    this._strokeWidth = 1.0;  // float > 0
    this._strokeCap = 'butt';  // string 'butt' 'round' 'square'
    this._strokeJoin = 'miter';  // string 'miter' 'bevel' 'round'
    this._strokeMiterLimit = 10;  // float >= 1
    this._strokeDashArray = null;
    this._strokeDashOffset = 0;

    this._spot1 = Spot.Default;
    this._spot2 = Spot.Default;
    this._parameter1 = NaN;
    this._parameter2 = NaN;
    this._isGeometryPositioned = false;

    this._pathPattern = null;

    // For links
    this._toArrow = 'None';
    this._fromArrow = 'None';

    // for Grid/Graduated Panel:
    this._interval = 1;

    // for Graduated Panel:
    this._graduatedStart = 0;
    this._graduatedEnd = 1;
    this._graduatedSkip = null;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Shape.
   * @this {Shape}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._geometry = this._geometry;
    copy._figure = this._figure;

    copy._geometryStretch = this._geometryStretch;
    copy._suppliedGeometry = this._suppliedGeometry;

    copy._fill = this._fill;  // OK to share: should be frozen
    copy._stroke = this._stroke;  // OK to share: should be frozen
    copy._strokeWidth = this._strokeWidth;
    copy._strokeCap = this._strokeCap;
    copy._strokeJoin = this._strokeJoin;
    copy._strokeMiterLimit = this._strokeMiterLimit;
    if (this._strokeDashArray !== null) {
      copy._strokeDashArray = Util.copyArray(this._strokeDashArray);
    }
    copy._strokeDashOffset = this._strokeDashOffset;

    copy._spot1 = this._spot1.copyFrozen();
    copy._spot2 = this._spot2.copyFrozen();
    copy._parameter1 = this._parameter1;
    copy._parameter2 = this._parameter2;
    copy._isGeometryPositioned = this._isGeometryPositioned;

    copy._pathPattern = this._pathPattern;

    copy._toArrow = this._toArrow;
    copy._fromArrow = this._fromArrow;

    copy._interval = this._interval;

    copy._graduatedStart = this._graduatedStart;
    copy._graduatedEnd = this._graduatedEnd;
    copy._graduatedSkip = this._graduatedSkip;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Shape}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg === GraphObject.None || arg === GraphObject.Uniform || arg === GraphObject.UniformToFill || arg === GraphObject.Default) {
      this.geometryStretch = arg;
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * @hidden @internal
   * @this {Shape}
   * @return {string}
   */
  public toString(): string {
    return 'Shape(' + (this.figure !== 'None' ? this.figure : (this.toArrow !== 'None' ? this.toArrow : this.fromArrow)) + ')#' + Set.hashId(this);
  }

  /**
   * @hidden @internal
   * Draws a GraphObject repeatedly along the path.
   * @this {Shape}
   * @param {IContext} context A Canvas 2D Context for drawing.
   * @param {Diagram} diagram
   */
  private drawPathPattern(context: IContext, diagram: Diagram) {
    const s = this.pathPattern;
    s.measure(Infinity, Infinity);
    const mb = s.measuredBounds;
    s.arrange(0, 0, mb.width, mb.height);
    const geo = this.geometry;

    context.save();
    context.beginPath();
    const allpaths = Util.tempArray<number>();

    if (geo.type === Geometry.Line) {
      allpaths.push(geo.startX); allpaths.push(geo.startY);
      allpaths.push(geo.endX); allpaths.push(geo.endY);
      this.drawAlong(context, diagram, allpaths, s);
    } else if (geo.type === Geometry.Path) {
      const fit = geo.figures.iterator;
      while (fit.next()) {
        const fig = fit.value;
        allpaths.length = 0;
        allpaths.push(fig.startX); allpaths.push(fig.startY);

        // lastPoint only needed for drawing dash approximations
        let lastpointX = fig.startX;
        let lastpointY = fig.startY;
        let lastSubX = lastpointX;
        let lastSubY = lastpointY;

        // for each segment of the PathFigure:
        const segments = fig.segments._arr;
        const sl = segments.length;
        for (let j = 0; j < sl; j++) {
          const seg = segments[j];
          switch (seg.type) {
            case PathSegment.Move:
              this.drawAlong(context, diagram, allpaths, s);
              allpaths.length = 0;  // restart
              allpaths.push(seg.endX); allpaths.push(seg.endY);

              lastpointX = seg.endX;
              lastpointY = seg.endY;
              lastSubX = lastpointX;
              lastSubY = lastpointY;
              break;
            case PathSegment.Line:
              allpaths.push(seg.endX); allpaths.push(seg.endY);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.Bezier:
              Geo.bezierLines(lastpointX, lastpointY,
                seg.point1X, seg.point1Y,
                seg.point2X, seg.point2Y,
                seg.endX, seg.endY,
                0.5, allpaths);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.QuadraticBezier:
              Geo.quadraticBezierLines(lastpointX, lastpointY,
                seg.point1X, seg.point1Y,
                seg.endX, seg.endY,
                0.5, allpaths);
              lastpointX = seg.endX;
              lastpointY = seg.endY;
              break;
            case PathSegment.Arc: {
              const curves = seg.buildBeziers(fig);
              const cl = curves.length;
              if (cl === 0) {
                allpaths.push(seg.centerX); allpaths.push(seg.centerY);
                lastpointX = seg.centerX;
                lastpointY = seg.centerY;
                break;
              }
              for (let cli = 0; cli < cl; cli++) {
                const cu = curves[cli];
                Geo.bezierLines(lastpointX, lastpointY,
                  cu[2], cu[3],
                  cu[4], cu[5],
                  cu[6], cu[7],
                  0.5, allpaths);
                lastpointX = cu[6];
                lastpointY = cu[7];
              }
              break;
            }
            case PathSegment.SvgArc: {
              const curves = seg.buildSVGBeziers(fig, lastpointX, lastpointY);
              const cl = curves.length;
              if (cl === 0) {
                allpaths.push(seg.centerX); allpaths.push(seg.centerY);
                lastpointX = seg.centerX;
                lastpointY = seg.centerY;
                break;
              }
              for (let cli = 0; cli < cl; cli++) {
                const cu = curves[cli];
                Geo.bezierLines(lastpointX, lastpointY,
                  cu[2], cu[3],
                  cu[4], cu[5],
                  cu[6], cu[7],
                  0.5, allpaths);
                lastpointX = cu[6];
                lastpointY = cu[7];
              }
              break;
            }
            default:
              Util.throwError('Segment not of valid type: ' + seg.type);
          }
          if (seg.isClosed) {
            allpaths.push(lastSubX); allpaths.push(lastSubY);
            this.drawAlong(context, diagram, allpaths, s);
          }
        } // end segment iteration
        this.drawAlong(context, diagram, allpaths, s);
      }
    } else if (geo.type === Geometry.Rectangle) {
      allpaths.push(geo.startX); allpaths.push(geo.startY);
      allpaths.push(geo.endX); allpaths.push(geo.startY);
      allpaths.push(geo.endX); allpaths.push(geo.endY);
      allpaths.push(geo.startX); allpaths.push(geo.endY);
      allpaths.push(geo.startX); allpaths.push(geo.startY);
      this.drawAlong(context, diagram, allpaths, s);
    } else if (geo.type === Geometry.Ellipse) {
      // need to get a set of Bezier curves from which we can call Geo.bezierLines
      const ellfig = new PathFigure();
      ellfig.startX = geo.endX;
      ellfig.startY = (geo.startY + geo.endY) / 2;

      const ellseg = new PathSegment(PathSegment.Arc);
      ellseg.startAngle = 0;
      ellseg.sweepAngle = 360;
      ellseg.centerX = (geo.startX + geo.endX) / 2;
      ellseg.centerY = (geo.startY + geo.endY) / 2;
      ellseg.radiusX = Math.abs(geo.startX - geo.endX) / 2;
      ellseg.radiusY = Math.abs(geo.startY - geo.endY) / 2;
      ellfig.add(ellseg);

      const curves = ellseg.buildBeziers(ellfig);
      const cl = curves.length;
      if (cl === 0) {
        allpaths.push(ellseg.centerX); allpaths.push(ellseg.centerY);
      } else {
        let lastpointX = ellfig.startX;
        let lastpointY = ellfig.startY;
        for (let cli = 0; cli < cl; cli++) {
          const cu = curves[cli];
          Geo.bezierLines(lastpointX, lastpointY,
            cu[2], cu[3],
            cu[4], cu[5],
            cu[6], cu[7],
            0.5, allpaths);
          lastpointX = cu[6];
          lastpointY = cu[7];
        }
      }
      this.drawAlong(context, diagram, allpaths, s);
    }  // end Geometry.Ellipse

    Util.freeArray(allpaths);
    context.restore();
    context.clearContextCache(false);
  }

  /**
   * @hidden @internal
   * Return the interval distance by which drawing of the given object should be repeated.
   * If the OBJ is a Shape, the distance is reduced by the strokeWidth,
   * to allow for smooth lines crossing between drawings, especially diagonals.
   * @this {Shape}
   * @param {GraphObject} obj
   * @return {number}  will be at least 1
   */
  private computeDashLength(obj: GraphObject): number {
    let dashLength = obj.measuredBounds.width;
    if (obj instanceof Shape) dashLength -= obj.strokeWidth;
    if (dashLength < 1) dashLength = 1;
    return dashLength;
  }

  // This is no longer used to draw dashed curves, only path
  // This is a special implementation of drawDash used for drawing dashed curves
  // It is slower, but it takes an array of points instead of a single point and
  // faithfully dashes along the given chain of points,
  // carrying over dash remainders properly from one segment to the next.
  /**
   * @hidden @internal
   * @this {Shape}
   * @param {IContext} context
   * @param {Diagram} diagram
   * @param {Array.<number>} arr
   * @param {GraphObject} obj
   */
  private drawAlong(context: IContext, diagram: Diagram, arr: Array<number>, obj: GraphObject): void {
    const len = arr.length;
    if (len < 4) return;
    const mb = obj.measuredBounds;
    const objWidth = Math.max(1, mb.width);
    const objHeight = mb.height;

    let x = arr[0];
    let y = arr[1];
    let x2 = 0.0;
    let y2 = 0.0;

    let dx = 0.0;
    let dy = 0.0;
    let slope = 0.0;
    let angle = 0.0;
    let dist = 0.0;
    let distRemainingTotal = 0.0;
    // an array of every line segment's dx, slope and distance [dx, slope, dist, remain]
    const segData = Util.tempArray<Array<number>>();
    for (let i = 2; i < len; i += 2) {
      x2 = arr[i];
      y2 = arr[i + 1];
      dx = (x2 - x);
      dy = (y2 - y);
      if (dx === 0) dx = 0.001;
      slope = dy / dx;
      angle = Math.atan2(dy, dx);
      dist = Math.sqrt(dx * dx + dy * dy);
      // ?? wasteful array alloc:
      segData.push([dx, angle, slope, dist]);
      distRemainingTotal += dist;
      // for the next iteration:
      x = x2;
      y = y2;
    }

    // reset x/y to starting point values:
    x = arr[0];
    y = arr[1];

    const originalDashLength = this.computeDashLength(obj);
    // var dashIndex = 0;
    let dashLength = originalDashLength;

    // using midpoints, so we want to start in the middle of an object
    const offset = objWidth / 2;
    let first = (offset === 0) ? false : true; // first line is different if there is an offset

    let currSegNum = 0;
    // when currSeg changes the others must change with it to be consistent
    let currSeg = segData[currSegNum];
    dx = currSeg[0];
    angle = currSeg[1];
    slope = currSeg[2];
    dist = currSeg[3]; // Distance in the segment, also used as distance remaining in the segment

    // *********** Draw the rest

    // We dash a single segment for as long as possible until its distance runs out
    // then we create a "remainder" that is added to the start of the next segment before resuming dashing
    let remainder = 0.0; // amount of dash to add to next segment
    while (distRemainingTotal >= 0.1) {
      // If there's a remainder we don't increment the dash index and make a new dash length
      // the remainder is our last dash length and we must finish drawing that dash first
      if (remainder === 0) {
        if (first) {
          dashLength = originalDashLength;
          // dashIndex++; // explicitly incremented!
          // offset = offset % dashLength;
          dashLength -= offset;
          distRemainingTotal -= offset;
          first = false;
        } else {
          dashLength = originalDashLength;
          // dashIndex++; // explicitly incremented!
        }
        if (dashLength === 0) dashLength = 1;
      }

      // applicible to last dash in entire set:
      if (dashLength > distRemainingTotal) {
        dashLength = distRemainingTotal;
        Util.freeArray(segData);
        return; // we are done, we never draw the last pathPattern in the set
      }
      // applicible to last dash in segment (must save remainder)
      if (dashLength > dist) {
        remainder = dashLength - dist;
        dashLength = dist;
      } else {
        remainder = 0;
      }

      let xStep = Math.sqrt(dashLength * dashLength / (1 + slope * slope));
      if (dx < 0) xStep = -xStep;
      x += xStep;
      y += slope * xStep;

      // save:
      context.translate(x, y);
      context.rotate(angle);
      context.translate(-(objWidth / 2), -(objHeight / 2));

      if (remainder === 0) obj.drawProtected(context, diagram);

      // restore:
      context.translate((objWidth / 2), (objHeight / 2));
      context.rotate(-angle);
      context.translate(-x, -y);

      distRemainingTotal -= dashLength;
      dist -= dashLength;

      if (remainder !== 0) {
        // increment segment if possible
        currSegNum++;
        if (currSegNum === segData.length) { Util.freeArray(segData); return; }
        currSeg = segData[currSegNum];

        dx = currSeg[0];
        angle = currSeg[1];
        slope = currSeg[2];
        dist = currSeg[3]; // Also used as distance remaining
        dashLength = remainder;
      }
    }
    Util.freeArray(segData);
  }

  /**
   * @hidden @internal
   * Draw the shape to a specified context.
   * @this {Shape}
   * @param {IContext} context A Canvas 2D Context for drawing.
   * @param {Diagram} diagram
   */
  public drawProtected(context: IContext, diagram: Diagram): void {
    const stroke = this._stroke;
    const fill = this._fill;
    if (stroke === null && fill === null) return;
    const ab = this.actualBounds;
    const nb = this.naturalBounds;
    if (fill !== null) this.setBrush(context, fill, true, false, nb, ab);
    const part = this.part;
    let logicalSW = this._strokeWidth;
    if (stroke !== null && logicalSW === 0 && part !== null) {
      logicalSW = part.getLogicalStrokeWidth();
    }
    if (stroke !== null && logicalSW !== 0) {
      this.setBrush(context, stroke, false, false, nb, ab);
      context.lineWidth = logicalSW;
      context.lineJoin = this._strokeJoin;
      context.lineCap = this._strokeCap;
      context.miterLimit = this._strokeMiterLimit;
    }
    let isShadowed = false;
    if (part && diagram.getRenderingHint('drawShadows')) isShadowed = part.isShadowed;
    let hidestrokeshadow = true;
    // If the stroke exists and the fill does not, don't hide the stroke's shadow
    if (stroke !== null && (fill === null || fill === 'transparent')) hidestrokeshadow = false;

    let dash = false;
    const strokeDashArray = this.strokeDashArray;
    if (strokeDashArray !== null) {
      dash = true;
      context.enableDash(strokeDashArray, this._strokeDashOffset);
    }

    const g = this._geometry;
    if (g === null) return;
    if (g.type === Geometry.Line) {
      context.beginPath();
      context.moveTo(g.startX, g.startY);
      context.lineTo(g.endX, g.endY);
      if (fill !== null) context.fillContext(fill);
      if (logicalSW !== 0 && stroke !== null) {
        context.strokeContext();
      }
    } else if (g.type === Geometry.Rectangle) {
      const x1 = g.startX;
      const y1 = g.startY;
      const x2 = g.endX;
      const y2 = g.endY;
      const x = Math.min(x1, x2);
      const y = Math.min(y1, y2);
      const w = Math.abs(x2 - x1);
      const h = Math.abs(y2 - y1);
      context.beginPath();
      context.rect(x, y, w, h); // use path for both stroke and fill
      if (fill !== null) {
        context.fillContext(fill);
      }
      if (stroke !== null) {
        if (hidestrokeshadow && isShadowed) {
          context.shadowsOff();
        }
        if (logicalSW !== 0) {
          // only make second path if needed for IE bug, otherwise former context path will be stroked
          if (w === 0 || h === 0) {
            context.beginPath();
            context.rect(x, y, Math.max(w, 0.1), Math.max(h, 0.1)); // IE won't stroke paths of 0 width/height
          }
          context.strokeContext();
        }
        if (hidestrokeshadow && isShadowed) {
          context.shadowsOn();
        }
      }
    } else if (g.type === Geometry.Ellipse) {
      const x1 = g.startX;
      const y1 = g.startY;
      const x2 = g.endX;
      const y2 = g.endY;
      const rx = Math.abs(x2 - x1) / 2;
      const ry = Math.abs(y2 - y1) / 2;
      const cx = Math.min(x1, x2) + rx;
      const cy = Math.min(y1, y2) + ry;
      context.beginPath();
      context.moveTo(cx, cy - ry);
      context.bezierCurveTo(cx + (Geo.KAPPA * rx), cy - ry, cx + rx, cy - (Geo.KAPPA * ry), cx + rx, cy);
      context.bezierCurveTo(cx + rx, cy + (Geo.KAPPA * ry), cx + (Geo.KAPPA * rx), cy + ry, cx, cy + ry);
      context.bezierCurveTo(cx - (Geo.KAPPA * rx), cy + ry, cx - rx, cy + (Geo.KAPPA * ry), cx - rx, cy);
      context.bezierCurveTo(cx - rx, cy - (Geo.KAPPA * ry), cx - (Geo.KAPPA * rx), cy - ry, cx, cy - ry);
      context.closePath();

      if (fill !== null) context.fillContext(fill);

      if (logicalSW !== 0 && stroke !== null) {
        if (hidestrokeshadow && isShadowed) {
          context.shadowsOff();
          context.strokeContext();
          context.shadowsOn();
        } else {
          context.strokeContext();
        }
      }
    } else if (g.type === Geometry.Path) {
      const figs = g.figures;
      // for each PathFigure:
      const l = figs.length;
      for (let i = 0; i < l; i++) {
        const fig = figs._arr[i];
        context.beginPath();
        context.moveTo(fig.startX, fig.startY);
        // for each segment of the PathFigure:
        const segments = fig.segments._arr;
        const sl = segments.length;
        let lastSeg: PathSegment = null;
        for (let j = 0; j < sl; j++) {
          const seg = segments[j];
          switch (seg.type) {
            case PathSegment.Move:
              context.moveTo(seg.endX, seg.endY);
              break;
            case PathSegment.Line:
              context.lineTo(seg.endX, seg.endY);
              break;
            case PathSegment.Bezier:
              context.bezierCurveTo(seg.point1X, seg.point1Y,
                seg.point2X, seg.point2Y,
                seg.endX, seg.endY);
              break;
            case PathSegment.QuadraticBezier:
              context.quadraticCurveTo(seg.point1X, seg.point1Y,
                seg.endX, seg.endY);
              break;
            case PathSegment.Arc:
              if (seg.radiusX === seg.radiusY) {
                const lx = lastSeg !== null ? lastSeg.endX : fig.startX;
                const ly = lastSeg !== null ? lastSeg.endY : fig.startY;
                const pi180 = Math.PI / 180;
                context.arc(
                  seg.point1X,
                  seg.point1Y,
                  seg.radiusX, // x and y are the same
                  seg.startAngle * pi180,
                  (seg.startAngle + seg.sweepAngle) * pi180,
                  (seg.sweepAngle < 0),
                  lx, ly);
              } else {
                const curves = seg.buildBeziers(fig);
                const cl = curves.length;
                if (cl === 0) {
                  context.lineTo(seg.centerX, seg.centerY);
                  break;
                }
                for (let cli = 0; cli < cl; cli++) {
                  const cu = curves[cli];
                  if (cli === 0) context.lineTo(cu[0], cu[1]);
                  context.bezierCurveTo(cu[2], cu[3], cu[4], cu[5], cu[6], cu[7]);
                }
              }
              break;
            case PathSegment.SvgArc: {
              let lx = 0.0;
              let ly = 0.0;
              let curves: Array<Array<number>>;
              if (lastSeg !== null && lastSeg.type === PathSegment.Arc) {
                curves = lastSeg.buildBeziers(fig);
                const clen = curves.length;
                if (clen === 0) {
                  context.lineTo(seg.centerX, seg.centerY);
                  break;
                }
                const c = curves[clen - 1] || null;
                if (c !== null) {
                  lx = c[6];
                  ly = c[7];
                }
              } else {
                lx = lastSeg !== null ? lastSeg.endX : fig.startX;
                ly = lastSeg !== null ? lastSeg.endY : fig.startY;
              }
              curves = seg.buildSVGBeziers(fig, lx, ly);
              const cl = curves.length;
              if (cl === 0) {
                context.lineTo(seg.centerX, seg.centerY);
                break;
              }
              for (let cli = 0; cli < cl; cli++) {
                const cu = curves[cli];
                context.bezierCurveTo(cu[2], cu[3], cu[4], cu[5], cu[6], cu[7]);
              }
              break;
            }
            default:
              Util.throwError('Segment not of valid type: ' + seg.type);
          }
          if (seg.isClosed) context.closePath();
          lastSeg = seg;
        } // end segment iteration
        // Time to stroke+fill the PathFigure
        // ??? needs cleanup at some point
        if (isShadowed) {
          if (fig.isShadowed) {
            if (fig.isFilled === true && fill !== 'transparent' && fill !== null) {
              context.fillContext(fill);
            }
            if (logicalSW !== 0 && stroke !== null) {
              if (hidestrokeshadow) {
                context.shadowsOff();
              }
              context.strokeContext();

              if (hidestrokeshadow) {
                context.shadowsOn();
              }
            }
          } else { // no shadow on the PathFigure
            if (hidestrokeshadow) {
              context.shadowsOff();
            }
            if (fig.isFilled === true && fill !== 'transparent' && fill !== null) context.fillContext(fill);
            if (logicalSW !== 0 && stroke !== null) {
              context.strokeContext();
            }

            if (hidestrokeshadow) {
              context.shadowsOn();
            }
          }
        } else { // not shadowed
          if (fig.isFilled === true && fill !== null) {
            context.fillContext(fill);
          }
          if (logicalSW !== 0 && stroke !== null) {
            context.strokeContext();
          }
        }
      } // end path figures

    } // end geometry type switch

    // Set dash back when done
    if (dash) {
      context.disableDash();
    }

    if (this.pathPattern !== null) this.drawPathPattern(context, diagram);
  } // end draw

  /**
   * @hidden @internal
   * This override takes the {@link #strokeWidth} into account.
   * @export
   * @this {Shape}
   * @param {Spot|Point} local a Spot or a Point in local coordinates.
   * @param {Point=} result an optional Point that is modified and returned.
   * @return {Point} in document coordinates
   */
  public getDocumentPoint(local: Spot | Point, result?: Point): Point {
    if (result === undefined) result = new Point();
    if (local instanceof Spot) {
      const s = local;
      if (s.isNoSpot()) Util.throwError('getDocumentPoint Spot must be a real, specific Spot, not: ' + s.toString());
      const ns = this.naturalBounds;
      const sw = this.strokeWidth;
      result.assignTo((s.x * (ns.width + sw) - (sw / 2)) + ns.x + s.offsetX,
        (s.y * (ns.height + sw) - (sw / 2)) + ns.y + s.offsetY);
      this.allTransforms.transformPoint(result);
      return result;
    } else {
      result.set(local);  // may check that LOCAL is a Point
      this.allTransforms.transformPoint(result);
      return result;
    }
  }

  /**
   * @hidden @internal
   * Returns the Rect in document coordinates for this object's bounds.
   * If this GraphObject is a Part, the rect will be identical to its {@link #actualBounds}.
   * @export
   * @this {Shape}
   * @param {Rect=} result an optional Rect that is modified and returned.
   * @return {Rect} in document coordinates.
   * @see {@link #getDocumentPoint}
   * @since 2.0
   */
  public getDocumentBounds(result?: Rect): Rect {
    if (result === undefined) result = new Rect();
    const ns = this.naturalBounds;
    const t = this.allTransforms;
    // account for strokeWidth in bounds before transforming
    const bounds = Rect.allocAt(ns.x, ns.y, ns.width, ns.height);
    const sw = this.strokeWidth;
    bounds.inflate(sw / 2, sw / 2);

    const p = Point.allocAt(bounds.x, bounds.y).transform(t); // Top left spot
    result.assignTo(p.x, p.y, 0, 0);
    p.assignTo(bounds.right, bounds.y).transform(t); // Top right
    result.unionNoCheck(p.x, p.y, 0, 0);
    p.assignTo(bounds.right, bounds.bottom).transform(t); // Bottom right
    result.unionNoCheck(p.x, p.y, 0, 0);
    p.assignTo(bounds.x, bounds.bottom).transform(t); // bottom left
    result.unionNoCheck(p.x, p.y, 0, 0);
    Rect.free(bounds);
    Point.free(p);
    return result;
  }

  /**
   * @hidden @internal
   * For Paths, count intersections on the ray from outside bounding box to p.
   * @this {Shape}
   * @param {Point} p
   * @param {boolean=} precise optional argument that forces the strokewidth to stay small.
   * @return {boolean} true if p is in the geometry.
   */
  protected containsPointProtected(p: Point, precise?: boolean): boolean {
    const geom = this.geometry;
    if (geom === null) return false;
    if (this.fill === null && this.stroke === null) return false;
    const bounds = geom.bounds;
    let sw = (this.strokeWidth / 2);

    // If its a line or if there are no filled subpaths by definition
    if (geom.type === Geometry.Line && !precise) { sw += 2; }

    // First make sure the point [p] is even in [bounds + sw]
    const r = Rect.alloc();
    r.assign(bounds);
    r.inflate(sw + 2, sw + 2);
    if (!r.containsPoint(p)) {
      Rect.free(r);
      return false;
    }
    const swepsilon = sw + 0.0001;

    // Now decide what to do for each kind of Geometry
    if (geom.type === Geometry.Line) {
      if (this.stroke === null) return false;
      // don't allow extension over endpoints
      const dot1 = (geom.startX - geom.endX) * (p.x - geom.endX) + (geom.startY - geom.endY) * (p.y - geom.endY);
      const dot2 = (geom.endX - geom.startX) * (p.x - geom.startX) + (geom.endY - geom.startY) * (p.y - geom.startY);
      if (dot1 < 0 || dot2 < 0) { return false; }
      Rect.free(r);
      return Geo.lineContainsPoint(geom.startX, geom.startY, geom.endX, geom.endY, sw, p.x, p.y);
    } else if (geom.type === Geometry.Rectangle) {
      const x1 = geom.startX;
      const y1 = geom.startY;
      const x2 = geom.endX;
      const y2 = geom.endY;

      r.x = Math.min(x1, x2);
      r.y = Math.min(y1, y2);
      r.width = Math.abs(x2 - x1);
      r.height = Math.abs(y2 - y1);

      if (this.fill === null) {  // hollow rectangle?
        r.inflate(-swepsilon, -swepsilon);
        // if the point is inside the rectangle, false
        if (r.containsPoint(p)) { Rect.free(r); return false; }
        r.inflate(swepsilon, swepsilon);
      }
      if (this.stroke !== null) { // within the outside of the rectangle?
        r.inflate(sw, sw);
      }
      const ret = r.containsPoint(p);
      Rect.free(r);
      return ret;
    } else if (geom.type === Geometry.Ellipse) {
      const x1 = geom.startX;
      const y1 = geom.startY;
      const x2 = geom.endX;
      const y2 = geom.endY;

      let x = Math.min(x1, x2);
      let y = Math.min(y1, y2);
      const w = Math.abs(x2 - x1);
      const h = Math.abs(y2 - y1);

      let a = w / 2;
      let b = h / 2;
      x = p.x - (x + a);  // normalize X and Y
      y = p.y - (y + b);
      if (this.fill === null) {  // hollow ellipse?
        a -= swepsilon;
        b -= swepsilon;
        if (a <= 0 || b <= 0) { Rect.free(r); return false; }
        // if the point is inside the ellipse, false
        if ((x * x) / (a * a) + (y * y) / (b * b) <= 1) { Rect.free(r); return false; }
        a += swepsilon;
        b += swepsilon;
      }
      if (this.stroke !== null) { // within the outside of the ellipse?
        a += swepsilon;
        b += swepsilon;
      }
      Rect.free(r);
      if (a <= 0 || b <= 0) return false;
      return ((x * x) / (a * a) + (y * y) / (b * b) <= 1);
    } else if (geom.type === Geometry.Path) {
      Rect.free(r);
      // If there's no fill we can treat the entire geometry as if there are no filled subpaths
      if (this.fill === null) {
        return geom.pathContainsPoint(p.x, p.y, sw);
      }

      return geom.containsPoint(p, sw, this.strokeWidth > 1, precise);
    } else {
      Util.throwError('Unknown Geometry type: ' + geom.type);
      return false;
    }
  }

  /**
   * @hidden @internal
   * @this {Shape}
   * @param {number} width
   * @param {number} height
   * @param {number} minw
   * @param {number} minh
   */
  protected measureProtected(width: number, height: number, minw: number, minh: number): void {
    const des = this.desiredSize;
    const sw = this._strokeWidth;
    width = Math.max(width, 0);
    height = Math.max(height, 0);

    let cb: Rect;
    if (this._suppliedGeometry !== null) {
      cb = this.geometry.bounds;
    } else {
      const fig = this.figure;
      let preset = Shape.PresetGeometries[fig];
      if (preset === undefined) {
        let func = Geo.GeometryGenerators[fig];
        if (typeof func === 'string') func = Geo.GeometryGenerators[func];
        if (typeof func === 'function') {
          preset = func(null, 100, 100);
          Shape.PresetGeometries[fig] = preset;
        } else {
          Util.throwError('Unsupported Figure: ' + fig);
        }
      }
      cb = preset.bounds;
    }

    let geoWidth = cb.width;
    let geoHeight = cb.height;
    // These are the expected width and height
    let w = cb.width;
    let h = cb.height;

    // Determine if the shape is stretched and use the input width and height appropriately
    const effectivestretch = this.getEffectiveStretch(true);
    switch (effectivestretch) {
      case GraphObject.None:
        minw = 0;
        minh = 0;
        break;
      case GraphObject.Fill:
        w = Math.max(width - sw, 0);  // given width and height determine end size (not geometry size) so "- sw"
        h = Math.max(height - sw, 0);
        break;
      case GraphObject.Horizontal:
        w = Math.max(width - sw, 0);
        minh = 0;
        break;
      case GraphObject.Vertical:
        minw = 0;
        h = Math.max(height - sw, 0);
        break;
    } // end stretch switch
    // desired always overrides
    if (isFinite(des.width)) w = des.width;
    if (isFinite(des.height)) h = des.height;
    const max = this.maxSize;
    const min = this.minSize;
    minw = Math.max(minw - sw, min.width);
    minh = Math.max(minh - sw, min.height);
    w = Math.min(max.width, w);
    h = Math.min(max.height, h);
    w = isFinite(w) ? Math.max(minw, w) : Math.max(geoWidth, minw);
    h = isFinite(h) ? Math.max(minh, h) : Math.max(geoHeight, minh);

    // w,h cannot get larger or smaller
    // but the geometry can shrink or fill to fit w,h
    // by using geometryStretch
    const stretch = this.getGeometryStretch();
    switch (stretch) {
      case GraphObject.None:
        // Do nothing
        break;
      case GraphObject.Fill:
        geoWidth = w;
        geoHeight = h;
        break;
      case GraphObject.Uniform:
        // Find the smaller side and compute for it
        let ratio = Math.min(w / geoWidth, h / geoHeight);
        if (!isFinite(ratio)) ratio = 1;
        geoWidth = geoWidth * ratio;
        geoHeight = geoHeight * ratio;
        break;
      default:
        Util.throwError(stretch + ' is not a valid geometryStretch.');
    }

    if (this._suppliedGeometry !== null) {
      geoWidth = Math.max(geoWidth, 0.01);
      geoHeight = Math.max(geoHeight, 0.01);
      const originalGeo = this._suppliedGeometry !== null ? this._suppliedGeometry : this._geometry;
      const geom = originalGeo.computeScaledGeometry(geoWidth, geoHeight);
      if (Debug) geom.freeze();
      this._geometry = geom; // don't override the supplied geometry
    } else if (this._geometry === null ||
      !Geo.isApproxEqual(this._geometry._inputWidth, width - sw) ||
      !Geo.isApproxEqual(this._geometry._inputHeight, height - sw)) {
      this._geometry = Shape.makeGeometry(this, geoWidth, geoHeight);
    }

    cb = this._geometry.bounds;
    // variable width or height must measure by their geometry, all others we can use the truly determined width and height
    if (width === Infinity || height === Infinity) {
      this._measureRect(cb.x - (sw / 2), cb.y - (sw / 2), (width === 0 && geoWidth === 0) ? 0 : cb.width + sw, (height === 0 && geoHeight === 0) ? 0 : cb.height + sw);
    } else {
      this._measureRect(-(sw / 2), -(sw / 2), w + sw, h + sw);
      // If the shape area is larger or smaller than the geometric area, the line above will place the geometry
      // in the top-left corner of the shape instead of in its "center", as Picture currently does
      // consider changing in the future, perhaps allowing it to be specified with a geometryAlign property (+ imageAlign for Picture)
      // this._measureRect(((cb.width-w)/2)  -(sw / 2),((cb.height-h)/2) -(sw / 2), w + sw, h + sw);
    }
  }  // end measure

  /**
   * @hidden @internal
   * Some geometries have their own default stretch (like Circle, to keep its aspect ratio)
   * @this {Shape}
   * @return {EnumValue}
   */
  public getGeometryStretch(): EnumValue {
    const stretch = this.geometryStretch;
    if (this._suppliedGeometry !== null) {
      if (stretch === GraphObject.Default) return GraphObject.Fill;
      return stretch;
    } else {
      if (stretch === GraphObject.Default) {
        return Shape.PresetGeometries[this.figure].defaultStretch;
      }
      return stretch;
    }
  }

  /**
   * @hidden @internal
   * Arranges the GraphObject onto its panel.
   * @this {Shape}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    this.commonArrange(fx, fy, fw, fh);
  }

  /**
   * @hidden @internal
   * ??? needed for override purposes so long as GraphObject.getNearestIntersectionPoint is public
   * @export
   * @this {Shape}
   * @param {Point} p1 the point we are looking to be closest to, on the line formed with p2.
   * @param {Point} p2 forms a line with p1.
   * @param {Point} result (reference) the point of this object that is closest to p1 and that is on the infinite line from p1 to p2.
   * @return {boolean} true if the infinite line does intersect with this object; false otherwise.
   */
  protected getNearestIntersectionPoint(p1: Point, p2: Point, result: Point): boolean {
    return this.getNearestIntersectionPoint2(p1.x, p1.y, p2.x, p2.y, result);
  }

  /**
   * @hidden @internal
   * The closest intersection point of a shape with a line is the
   * closest such point for each of its segments.
   * This does not always(???) take into account any pen width.
   * @this {Shape}
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {Point} result (out)
   * @return {boolean}
   */
  public getNearestIntersectionPoint2(p1x: number, p1y: number, p2x: number, p2y: number, result: Point): boolean {
    const tt = this.transform;
    const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
    const m0 = tt.m22 * d;
    const m1 = -tt.m12 * d;
    const m2 = -tt.m21 * d;
    const m3 = tt.m11 * d;
    const m4 = d * (tt.m21 * tt.dy - tt.m22 * tt.dx);
    const m5 = d * (tt.m12 * tt.dx - tt.m11 * tt.dy);

    // Inverted transform point
    const q1x = p1x * m0 + p1y * m2 + m4;
    const q1y = p1x * m1 + p1y * m3 + m5;
    const q2x = p2x * m0 + p2y * m2 + m4;
    const q2y = p2x * m1 + p2y * m3 + m5;

    const sw2 = this._strokeWidth / 2;
    let geom = this._geometry;
    if (geom === null) {
      this.measure(Infinity, Infinity);
      geom = this._geometry;
    }
    const bounds = geom.bounds;
    let ret = false; // the actual value returned
    if (geom.type === Geometry.Line) {
      // shortcut instead of adding startPoint and endPoint to curFigure
      if (this.strokeWidth <= 1.5) {
        ret = Geo.nearestIntersectionOnLine(geom.startX, geom.startY, geom.endX, geom.endY, q1x, q1y, q2x, q2y, result);
      } else {
        let dx = 0.0;
        let dy = 0.0;
        if (geom.startX === geom.endX) {
          dx = sw2;
          dy = 0;
        } else {
          const m = (geom.endY - geom.startY) / (geom.endX - geom.startX);
          dy = sw2 / Math.sqrt(1 + m * m);
          dx = dy * m;
        }

        // ??? test dumb objects or arrays [x, y] vs points sometime
        const points = Util.tempArray<Point>();
        let p = new Point();
        if (Geo.nearestIntersectionOnLine(geom.startX + dx, geom.startY + dy, geom.endX + dx, geom.endY + dy, q1x, q1y, q2x, q2y, p)) { points.push(p); }
        p = new Point();
        if (Geo.nearestIntersectionOnLine(geom.startX - dx, geom.startY - dy, geom.endX - dx, geom.endY - dy, q1x, q1y, q2x, q2y, p)) { points.push(p); }
        p = new Point();
        if (Geo.nearestIntersectionOnLine(geom.startX + dx, geom.startY + dy, geom.startX - dx, geom.startY - dy, q1x, q1y, q2x, q2y, p)) { points.push(p); }
        p = new Point();
        if (Geo.nearestIntersectionOnLine(geom.endX + dx, geom.endY + dy, geom.endX - dx, geom.endY - dy, q1x, q1y, q2x, q2y, p)) { points.push(p); }
        const l = points.length;
        if (l === 0) { Util.freeArray(points); return false; }
        ret = true;
        let min = Infinity;
        for (let i = 0; i < l; i++) {
          const point = points[i];
          const dist = (point.x - q1x) * (point.x - q1x) + (point.y - q1y) * (point.y - q1y);
          if (dist < min) {
            min = dist;
            result.x = point.x;
            result.y = point.y;
          }
        }
        Util.freeArray(points);
      }

    } else if (geom.type === Geometry.Rectangle) {
      ret = Geo.getNearestIntersectionPoint(
        bounds.x - sw2, // l
        bounds.y - sw2, // t
        bounds.x + bounds.width + sw2, // r
        bounds.y + bounds.height + sw2, // b
        q1x, q1y, q2x, q2y, result);
    } else if (geom.type === Geometry.Ellipse) {
      const trect = Rect.allocAt(bounds.x, bounds.y, bounds.width, bounds.height).inflate(sw2, sw2);
      ret = this.nearestIntersectionOnEllipse(trect, q1x, q1y, q2x, q2y, result);
      Rect.free(trect);
    } else if (geom.type === Geometry.Path) {
      let sx = 0.0;
      let sy = 0.0;
      let ex = 0.0;
      let ey = 0.0;
      const intp = Point.alloc();
      const dx = q2x - q1x;
      const dy = q2y - q1y;
      let min = dx * dx + dy * dy;
      const epsilon = 0.6; // was 0.5 in 1.x
      result.x = q2x; result.y = q2y;
      for (let i = 0; i < geom.figures.count; i++) {
        const fig = geom.figures._arr[i];
        const segs = fig.segments;
        sx = fig.startX;
        sy = fig.startY;
        let lmX = sx;
        let lmY = sy;
        for (let j = 0; j < segs.count; j++) {
          const seg = segs._arr[j];
          const type = seg.type;
          ex = seg.endX;
          ey = seg.endY;
          let inter = false;
          switch (type) {
            case PathSegment.Move:
              lmX = ex;
              lmY = ey;
              break;
            case PathSegment.Line:
              inter = this._lineIntersection(sx, sy, ex, ey, q1x, q1y, q2x, q2y, intp);
              break;
            case PathSegment.Bezier: {
              // ?? potential optimization, see if rect of bezier intersects line
              inter = Geo.bezierNearestIntersectionOnLine(sx, sy,
                seg.point1X, // c1x
                seg.point1Y, // c1y
                seg.point2X, // c2x
                seg.point2Y, // c2y
                ex, ey, q1x, q1y, q2x, q2y, epsilon, intp);
              break;
            }
            case PathSegment.QuadraticBezier: {
              inter = Geo.bezierNearestIntersectionOnLine(sx, sy,
                // elevate degree
                (sx + 2 * seg.point1X) / 3, // c1x
                (sy + 2 * seg.point1Y) / 3, // c1y
                (seg.point1X * 2 + ex) / 3, // c2x
                (seg.point1X * 2 + ex) / 3, // c2y
                ex, ey, q1x, q1y, q2x, q2y, epsilon, intp);
              break;
            }
            case PathSegment.Arc:
            case PathSegment.SvgArc: {
              // We don't set inter for Arcs, instead we test the several beziers and single line point inline
              const curves = (seg.type === PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, sx, sy);
              const clen = curves.length;
              if (clen === 0) {
                inter = this._lineIntersection(sx, sy, seg.centerX, seg.centerY, q1x, q1y, q2x, q2y, intp);
                break;
              }
              let c: Array<number> = null;
              for (let k = 0; k < clen; k++) {
                c = curves[k];
                if (k === 0 && this._lineIntersection(sx, sy, c[0], c[1], q1x, q1y, q2x, q2y, intp)) {
                  const dist = this._testDist(q1x, q1y, intp, min, result);
                  if (dist < min) { min = dist; ret = true; }
                }
                if (Geo.bezierNearestIntersectionOnLine(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], q1x, q1y, q2x, q2y, epsilon, intp)) {
                  const dist = this._testDist(q1x, q1y, intp, min, result);
                  if (dist < min) { min = dist; ret = true; }
                }
              }
              // manually set this since Arc's endX isn't "normal"
              ex = c[6];
              ey = c[7];
              break;
            }
            default:
              Util.throwError('Unknown Segment type: ' + type);
              break;
          }

          sx = ex;
          sy = ey;

          if (inter) {
            const dist = this._testDist(q1x, q1y, intp, min, result);
            if (dist < min) { min = dist; ret = true; }
          }

          if (seg.isClosed) {
            // Check one last Line: from this endpoint (ex/ey) to the last move point (lmX/lmY)
            ex = lmX;
            ey = lmY;
            if (this._lineIntersection(sx, sy, ex, ey, q1x, q1y, q2x, q2y, intp)) {
              const dist = this._testDist(q1x, q1y, intp, min, result);
              if (dist < min) { min = dist; ret = true; }
            }
          }

        } // end segment for loop
      }

      // offset due to strokewidth
      let lx = p2x - p1x;
      let ly = p2y - p1y;
      const l = Math.sqrt(lx * lx + ly * ly);
      if (l !== 0) {
        lx /= l; ly /= l;
      }
      result.x -= lx * sw2;
      result.y -= ly * sw2;
      Point.free(intp);
    } else {
      Util.throwError('Unknown Geometry type: ' + geom.type);
    }
    if (!ret) { return false; }

    this.transform.transformPoint(result);
    return true;
  }

  /**
   * @hidden @internal
   * helper for getNearestIntersectionPoint
   * @this {Shape}
   * @param {number} q1x
   * @param {number} q1y
   * @param {Point} intp
   * @param {number} min
   * @param {Point} result
   * @return {number}
   */
  private _testDist(q1x: number, q1y: number, intp: Point, min: number, result: Point): number {
    const dx = intp.x - q1x;
    const dy = intp.y - q1y;
    const dist = dx * dx + dy * dy;
    if (dist < min) {
      result.x = intp.x; result.y = intp.y;
      return dist;
    }
    return min;
  }

  /**
   * @hidden @internal
   * helper for getNearestIntersectionPoint
   * @this {Shape}
   * @param {number} sx
   * @param {number} sy
   * @param {number} ex
   * @param {number} ey
   * @param {number} q1x
   * @param {number} q1y
   * @param {number} q2x
   * @param {number} q2y
   * @param {Point} intp
   * @return {boolean}
   */
  private _lineIntersection(sx: number, sy: number, ex: number, ey: number, q1x: number, q1y: number, q2x: number, q2y: number, intp: Point): boolean {
    let inter = false;
    const d = (q1x - q2x) * (sy - ey) - (q1y - q2y) * (sx - ex);
    if (d === 0) { return false; }
    intp.x = ((q1x * q2y - q1y * q2x) * (sx - ex) - (q1x - q2x) * (sx * ey - sy * ex)) / d;
    intp.y = ((q1x * q2y - q1y * q2x) * (sy - ey) - (q1y - q2y) * (sx * ey - sy * ex)) / d;
    if ((sx > ex ? sx - ex : ex - sx) < (sy > ey ? sy - ey : ey - sy)) {
      const qmin = sy < ey ? sy : ey;
      const qmax = sy < ey ? ey : sy;
      if ((intp.y > qmin || Geo.isApproxEqual(intp.y, qmin)) && (intp.y < qmax || Geo.isApproxEqual(intp.y, qmax))) { inter = true; }
    } else {
      const qmin = sx < ex ? sx : ex;
      const qmax = sx < ex ? ex : sx;
      if ((intp.x > qmin || Geo.isApproxEqual(intp.x, qmin)) && (intp.x < qmax || Geo.isApproxEqual(intp.x, qmax))) { inter = true; }
    }
    return inter;
  }

  /**
   * @hidden @internal
   * @this {GraphObject}
   * @param {Rect} r A Rect in container coordinates.
   * @param {boolean} partialInclusion True if objects are to be included when they
   * intersect.
   * @param {Transform} transform
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectInPath(r: Rect, partialInclusion: boolean, transform: Transform): boolean {
    if (this.pickable === false) return false;
    transform.multiply(this.transform);
    if (partialInclusion) {
      return this.intersectsRectPath(r, transform);
    } else {
      return this.containedInRect(r, transform);
    }
  }

  /**
   * @hidden @internal
   * Determine if a given Rect is large enough to entirely contain this Shape.
   * @export
   * @this {Shape}
   * @param {Rect} r A Rect in container coordinates.
   * @param {Transform=} transform A transform to use if something other than the Shape's own transform is desired.
   * @return {boolean} true if rect r contains the entire object.
   */
  protected containedInRect(r: Rect, transform?: Transform): boolean {
    // using its own transform means we can just check the actualBounds
    if (transform === undefined) {
      return r.containsRect(this.actualBounds);
    }

    // see if all of the four points are inside
    let geom = this._geometry;
    if (geom === null) {
      this.measure(Infinity, Infinity);
      geom = this._geometry;
    }
    const ns = geom.bounds;
    const sw = this.strokeWidth / 2;
    // just add instead of inflating so we don't have to make a new Rect

    let result = false;
    const pnt = Point.alloc();
    pnt.assignTo(ns.x - sw, ns.y - sw);
    if (r.containsPoint(transform.transformPoint(pnt))) {
      pnt.assignTo(ns.x - sw, ns.bottom + sw);
      if (r.containsPoint(transform.transformPoint(pnt))) {
        pnt.assignTo(ns.right + sw, ns.bottom + sw);
        if (r.containsPoint(transform.transformPoint(pnt))) {
          pnt.assignTo(ns.right + sw, ns.y - sw);
          if (r.containsPoint(transform.transformPoint(pnt))) {
            result = true;
          }
        }
      }
    }
    Point.free(pnt);
    return result;

    // NYI: Instead of above implementation (taken from GraphObject), use the points specific to the geometry
  }

  // ?? optimize this sometime

  /**
   * @hidden @internal
   * Determines if a given rectangle intersects the Shape's bounds.
   * @export
   * @this {GraphObject}
   * @param {Rect} rect a rectangle in container coordinates.
   * @param {Transform=} transform optional transform.
   * @return {boolean}
   */
  protected intersectsRect(rect: Rect, transform?: Transform): boolean {
    // rect completely encompasses the shape
    if (this.containedInRect(rect, transform)) return true;

    if (transform === undefined) {
      transform = this.transform;
      if (rect.containsRect(this.actualBounds)) { return true; }
    }

    const tt = Transform.alloc();
    tt.set(transform);
    tt.invert();

    const l = rect.left;
    const r = rect.right;
    const t = rect.top;
    const b = rect.bottom;
    const pnt = Point.alloc();

    pnt.assignTo(l, t); tt.transformPoint(pnt);
    if (this.containsPointProtected(pnt, true)) { Point.free(pnt); return true; }  // LC
    pnt.assignTo(r, t); tt.transformPoint(pnt);
    if (this.containsPointProtected(pnt, true)) { Point.free(pnt); return true; }
    pnt.assignTo(l, b); tt.transformPoint(pnt);
    if (this.containsPointProtected(pnt, true)) { Point.free(pnt); return true; }
    pnt.assignTo(r, b); tt.transformPoint(pnt);
    if (this.containsPointProtected(pnt, true)) { Point.free(pnt); return true; }

    const pnt2 = Point.alloc();
    const temp = Point.alloc();

    // _intersectsSegment needs container coordinates, not local
    tt.set(transform);
    tt.multiplyInverted(this.transform);
    tt.invert();
    pnt2.x = r; pnt2.y = t; pnt2.transform(tt);
    pnt.x = l; pnt.y = t; pnt.transform(tt);

    let result = false;
    if (this._intersectsSegment(pnt, pnt2, temp)) {    // CC
      result = true;
    } else {
      pnt.x = r; pnt.y = b; pnt.transform(tt);
      if (this._intersectsSegment(pnt, pnt2, temp)) {
        result = true;
      } else {
        pnt2.x = l; pnt2.y = b; pnt2.transform(tt);
        if (this._intersectsSegment(pnt, pnt2, temp)) {
          result = true;
        } else {
          pnt.x = l; pnt.y = t; pnt.transform(tt);
          if (this._intersectsSegment(pnt, pnt2, temp)) {
            result = true;
          }
        }
      }
    }
    Point.free(pnt);
    Transform.free(tt);
    Point.free(pnt2);
    Point.free(temp);
    return result;
  }

  /**
   * @hidden @internal
   * Helper for intersectsRect and intersectsRectPath.
   * Uses container coordinates.
   * @this {GraphObject}
   * @param {Point} p1
   * @param {Point} p2
   * @param {Point} temp
   * @return {boolean}
   */
  private _intersectsSegment(p1: Point, p2: Point, temp: Point): boolean {
    if (!this.getNearestIntersectionPoint(p1, p2, temp)) { return false; }
    const x1 = p1.x;
    const y1 = p1.y;
    const x2 = p2.x;
    const y2 = p2.y;
    const x = temp.x;
    const y = temp.y;
    if (x1 === x2) {
      let min = 0.0;
      let max = 0.0;
      if (y1 < y2) { min = y1; max = y2; } else { min = y2; max = y1; }
      return y >= min && y <= max;
    } else {
      let min = 0.0;
      let max = 0.0;
      if (x1 < x2) { min = x1; max = x2; } else { min = x2; max = x1; }
      return x >= min && x <= max;
    }
  }

  /**
   * @hidden @internal
   * Determines if a given rectangle intersects the Shape's bounds.
   * @this {GraphObject}
   * @param {Rect} rect a rectangle in container coordinates.
   * @param {Transform=} transform optional transform.
   * @return {boolean}
   */
  private intersectsRectPath(rect: Rect, transform?: Transform): boolean {
    // rect completely encompasses the shape
    if (this.containedInRect(rect, transform)) return true;

    if (transform === undefined) {
      transform = this.transform;
      if (rect.containsRect(this.actualBounds)) { return true; }
    }

    const l = rect.left;
    const r = rect.right;
    const t = rect.top;
    const b = rect.bottom;
    const pnt = Point.alloc();
    const pnt2 = Point.alloc();
    const temp = Point.alloc();
    const tt = Transform.alloc();
    // _intersectsSegment needs container coordinates, not local
    tt.set(transform);
    tt.multiplyInverted(this.transform);
    tt.invert();
    pnt2.x = r; pnt2.y = t; pnt2.transform(tt);
    pnt.x = l; pnt.y = t; pnt.transform(tt);

    let result = false;
    if (this._intersectsSegment(pnt, pnt2, temp)) {    // CC
      result = true;
    } else {
      pnt.x = r; pnt.y = b; pnt.transform(tt);
      if (this._intersectsSegment(pnt, pnt2, temp)) {
        result = true;
      } else {
        pnt2.x = l; pnt2.y = b; pnt2.transform(tt);
        if (this._intersectsSegment(pnt, pnt2, temp)) {
          result = true;
        } else {
          pnt.x = l; pnt.y = t; pnt.transform(tt);
          if (this._intersectsSegment(pnt, pnt2, temp)) {
            result = true;
          }
        }
      }
    }
    Transform.free(tt);
    Point.free(pnt);
    Point.free(pnt2);
    Point.free(temp);
    return result;
  }

  /**
   * @hidden @internal
   * @this {Shape}
   * @param {Point} p A Point in container coordinates.
   * @param {Point} distp A Point representing the distance from the point p.
   * @param {boolean} partialInclusion
   * @return {boolean}
   */
  public findObjectNear(p: Point, distp: Point, partialInclusion: boolean): boolean {
    // if it's filled, check to see if it contains the point
    if (partialInclusion && this.fill !== null && this.containsPointProtected(p, true)) {
      return true;
    }

    let dist2 = p.distanceSquaredPoint(distp);
    const dist2orig = dist2;

    if (this.strokeWidth > 1.5) {
      dist2 = this.strokeWidth / 2 + Math.sqrt(dist2);
      dist2 *= dist2;
    }
    let geo = this._geometry;
    if (geo === null) {
      this.measure(Infinity, Infinity);
      geo = this._geometry;
      if (geo === null) return false;
    }

    // If the circle completely encompasses the bounds of the Shape,
    // regardless of its size, then it is completely contained
    if (!partialInclusion) {
      const bounds = geo.bounds;
      const L = bounds.x;
      const T = bounds.y;
      const R = bounds.x + bounds.width;
      const B = bounds.y + bounds.height;
      if ((Point.distanceSquared(p.x, p.y, L, T) <= dist2) &&
        (Point.distanceSquared(p.x, p.y, R, T) <= dist2) &&
        (Point.distanceSquared(p.x, p.y, L, B) <= dist2) &&
        (Point.distanceSquared(p.x, p.y, R, B) <= dist2)) return true;
    }

    // returns true if a point is OUTSIDE the distance
    // if testCurve is true, we need to return false!
    /**
     * @hidden @internal
     * @param {Array.<number>} arr
     * @param {Point} p2
     */
    function testCurve(arr: Array<number>, p2: Point) {
      const l = arr.length;
      for (let i = 0; i < l; i += 2) {
        if (p2.distanceSquared(arr[i], arr[i + 1]) > dist2) return true;
      }
      return false;
    }

    // _getKeyFigsInPathGeometry
    const x1 = geo.startX;
    const y1 = geo.startY;
    const x2 = geo.endX;
    const y2 = geo.endY;
    if (geo.type === Geometry.Line) {
      const d = Point.distanceLineSegmentSquared(p.x, p.y, x1, y1, x2, y2);
      const dot1 = (x2 - x1) * (p.x - x1) + (y2 - y1) * (p.y - y1); // if either dot product is negative, point is past an endpoin
      const dot2 = (x1 - x2) * (p.x - x2) + (y1 - y2) * (p.y - y2);
      const thresh = (dot1 >= 0 && dot2 >= 0) ? dist2 : dist2orig; // thick stroke doesm't go past endpoints
      if (d <= thresh) { return true; }
      return false;
    } else if (geo.type === Geometry.Rectangle) {
      let hit = false;
      if (partialInclusion) {
        hit = Point.distanceLineSegmentSquared(p.x, p.y, x1, y1, x1, y2) <= dist2 ||
        Point.distanceLineSegmentSquared(p.x, p.y, x1, y1, x2, y1) <= dist2 ||
        Point.distanceLineSegmentSquared(p.x, p.y, x2, y1, x2, y2) <= dist2 ||
        Point.distanceLineSegmentSquared(p.x, p.y, x1, y2, x2, y2) <= dist2;
      }
      // !partialInclusion is already taken care of above
      return hit;
    } else if (geo.type === Geometry.Ellipse) {
      // find center
      const cx = (x1 + x2) / 2;
      const cy = (y1 + y2) / 2;

      // make center (0,0)
      const px = p.x - cx;
      const py = p.y - cy;

      const rx = Math.abs(x2 - x1) / 2;
      const ry = Math.abs(y2 - y1) / 2;

      if (rx === 0 || ry === 0) { // ???
        const d = Point.distanceLineSegmentSquared(p.x, p.y, x1, y1, x2, y2);
        if (d <= dist2) { return true; }
        return false;
      }

      if (partialInclusion) {
        const dist = Geo.distancePointEllipse(rx, ry, px, py);
        if (dist * dist <= dist2) { return true; }
        return false;
      } else {
        if (Point.distanceSquared(px, py, -rx, 0) >= dist2) return false;
        if (Point.distanceSquared(px, py, 0, -ry) >= dist2) return false;
        if (Point.distanceSquared(px, py, 0, ry) >= dist2) return false;
        if (Point.distanceSquared(px, py, rx, 0) >= dist2) return false;
        return true;
      }

    } else if (geo.type === Geometry.Path) {
      // check bounds
      const bounds = geo.bounds;
      const L = bounds.x;
      const T = bounds.y;
      const R = bounds.x + bounds.width;
      const B = bounds.y + bounds.height;
      if (p.x > R && p.x < L && p.y > B && p.y < T &&
        Point.distanceLineSegmentSquared(p.x, p.y, L, T, L, B) > dist2 &&
        Point.distanceLineSegmentSquared(p.x, p.y, L, T, R, T) > dist2 &&
        Point.distanceLineSegmentSquared(p.x, p.y, R, B, L, B) > dist2 &&
        Point.distanceLineSegmentSquared(p.x, p.y, R, B, R, T) > dist2
      ) { return false; }

      const tempsw = Math.sqrt(dist2);

      if (partialInclusion) {
        if (this.fill === null ? geo.pathContainsPoint(p.x, p.y, tempsw) : geo.containsPoint(p, tempsw, true)) {
          return true;
        }
      } else {
        // Path must be entirely contained within circle?
        const figs = geo.figures;
        for (let i = 0; i < figs.count; i++) {
          const fig = figs._arr[i];
          let lastX = fig.startX;
          let lastY = fig.startY;
          if (p.distanceSquared(lastX, lastY) > dist2) return false;
          const segments = fig.segments._arr;
          const numsegs = segments.length;
          for (let j = 0; j < numsegs; j++) {
            const seg = segments[j];
            switch (seg.type) {
              case PathSegment.Move:
              case PathSegment.Line:
                lastX = seg.endX;
                lastY = seg.endY;
                if (p.distanceSquared(lastX, lastY) > dist2) return false;
                break;
              case PathSegment.Bezier: {
                const res = Util.tempArray<number>();
                Geo.bezierLines(
                  lastX, lastY,
                  seg.point1X, seg.point1Y,
                  seg.point2X, seg.point2Y,
                  seg.endX, seg.endY, 0.8, res);
                const tested = testCurve(res, p);
                Util.freeArray(res);
                if (tested) return false;
                lastX = seg.endX;
                lastY = seg.endY;
                if (p.distanceSquared(lastX, lastY) > dist2) return false;
                break;
              }
              case PathSegment.QuadraticBezier: {
                const res = Util.tempArray<number>();
                Geo.quadraticBezierLines(
                  lastX, lastY,
                  seg.point1X, seg.point1Y,
                  seg.endX, seg.endY, 0.8, res);
                const tested = testCurve(res, p);
                Util.freeArray(res);
                if (tested) return false;
                lastX = seg.endX;
                lastY = seg.endY;
                if (p.distanceSquared(lastX, lastY) > dist2) return false;
                break;
              }
              case PathSegment.Arc:
              case PathSegment.SvgArc: {
                const curves = (seg.type === PathSegment.Arc) ? seg.buildBeziers(fig) : seg.buildSVGBeziers(fig, lastX, lastY);
                const clen = curves.length;
                if (clen === 0) {
                  lastX = seg.centerX;
                  lastY = seg.centerY;
                  if (p.distanceSquared(lastX, lastY) > dist2) return false;
                  break;
                }
                let c: Array<number> = null;
                const res = Util.tempArray<number>();
                for (let k = 0; k < clen; k++) {
                  c = curves[k];
                  res.length = 0;
                  Geo.bezierLines(c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7], 0.8, res);
                  if (testCurve(res, p)) { Util.freeArray(res); return false; }
                }
                Util.freeArray(res);
                if (c !== null) {
                  lastX = c[6];
                  lastY = c[7];
                }
                break;
              }
              default:
                Util.throwError('Unknown Segment type: ' + seg.type);
            } // end switch

          } // end segs
        }
        return true;
      }

    }

    return false;
  }

  /**
   * @hidden @internal
   * Find the intersection point of the elliptical path defined by rectangle
   * rect and an infinite line p1-p2 that is closest to point p1.
   * @this {Shape}
   * @param {Rect} rect
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {Point} result
   * @return {boolean}
   */
  private nearestIntersectionOnEllipse(rect: Rect, p1x: number, p1y: number, p2x: number, p2y: number, result: Point): boolean {
    if (rect.width === 0) {
      return Geo.nearestIntersectionOnLine(rect.x, rect.y,
        rect.x, rect.y + rect.height, p1x, p1y, p2x, p2y, result);
    } else if (rect.height === 0) {
      return Geo.nearestIntersectionOnLine(rect.x, rect.y,
        rect.x + rect.width, rect.y, p1x, p1y, p2x, p2y, result);
    } else {

      const Rx = rect.width / 2;
      const Ry = rect.height / 2;

      const Cx = rect.x + Rx;
      const Cy = rect.y + Ry;

      // see if the slope of P1-P2 is vertical or close to it (1 : 10000)
      let m = 9999;

      if (p1x !== p2x) {
        m = (p1y - p2y) / (p1x - p2x);
      }

      if (Math.abs(m) < 9999) {
        const b = (p1y - Cy) - m * (p1x - Cx);

        if ((Rx * Rx) * (m * m) + (Ry * Ry) - (b * b) < 0) {
          // ??? should this return the closest point,
          // even if there's no intersection
          result.x = NaN; result.y = NaN;
          return false;
        }

        const sqrt = Math.sqrt((Rx * Rx) * (m * m) + (Ry * Ry) - (b * b));

        const xplus = ((-((Rx * Rx) * m * b) + (Rx * Ry * sqrt)) /
          ((Ry * Ry) + ((Rx * Rx) * (m * m)))) + Cx;
        const xminus = ((-((Rx * Rx) * m * b) - (Rx * Ry * sqrt)) /
          ((Ry * Ry) + ((Rx * Rx) * (m * m)))) + Cx;

        const yplus = m * (xplus - Cx) + b + Cy;
        const yminus = m * (xminus - Cx) + b + Cy;

        const distplus = Math.abs((p1x - xplus) * (p1x - xplus)) +
          Math.abs((p1y - yplus) * (p1y - yplus));
        const distminus = Math.abs((p1x - xminus) * (p1x - xminus)) +
          Math.abs((p1y - yminus) * (p1y - yminus));

        if (distplus < distminus) {
          result.x = xplus;
          result.y = yplus;
        } else {
          result.x = xminus;
          result.y = yminus;
        }
      } else {
        const Ry2 = Ry * Ry;
        const Rx2 = Rx * Rx;
        const diff = p1x - Cx;
        const sqr = Ry2 - (Ry2 / Rx2) * (diff * diff);
        if (sqr < 0) {
          result.x = NaN; result.y = NaN;
          return false;
        }

        const sqrt = Math.sqrt(sqr);

        const yplus = Cy + sqrt;
        const yminus = Cy - sqrt;

        const distplus = Math.abs(yplus - p1y);
        const distminus = Math.abs(yminus - p1y);

        if (distplus < distminus) {
          result.x = p1x;
          result.y = yplus;
        } else {
          result.x = p1x;
          result.y = yminus;
        }
      }
      return true;
    }
  }

  /**
   * @hidden @internal
   * Returns miter length. May be used for drawing/drawnSize measurements.
   * @return {number}
   */
  public getMiterLength(): number {
    if (!this._suppliedGeometry !== null) {
      // its a preset figure, optimize some common ones
      switch (this._figure) {
        case 'None':
        case 'Square':
        case 'Ellipse':
        case 'Circle':
        case 'LineH':
        case 'LineV':
        case 'FramedRectangle':
        case 'RoundedRectangle':
        case 'Line1':
        case 'Line2':
        case 'Border':
        case 'Cube1':
        case 'Cube2':
        case 'Junction':
        case 'Cylinder1':
        case 'Cylinder2':
        case 'Cylinder3':
        case 'Cylinder4':
        case 'PlusLine':
        case 'XLine':
        case 'ThinCross':
        case 'ThickCross':
          return 0;
      }
    }

    return ((this._strokeWidth / 2) * this._strokeMiterLimit) * this.getDocumentScale();
  }

  /**
   * Gets or sets the Shape's {@link Geometry} that defines the Shape's figure.
   * Setting a geometry is not necessary if a {@link #figure} is specified,
   * as that will construct a geometry instead.
   *
   * Setting this geometry property will freeze the supplied Geometry.
   *
   * Setting this geometry property always overrides any set figure.
   * The default value is null.
   * @see {@link #geometryString}
   */
  /** @export */ get geometry(): Geometry | null {
    // do not create if it doesn't yet exist
    if (this._geometry !== null) return this._geometry;
    return this._suppliedGeometry;
  }
  /** @export */ set geometry(value: Geometry) {
    const old = this._geometry;
    if (old === value) return;
    if (value !== null) {
      if (Debug) Util.checkClass(value, Geometry, Shape, 'geometry');
      this._geometry = value.freeze();  // "in-use": don't allow modifications
      this._suppliedGeometry = this._geometry;
    } else {
      this._geometry = null;
      this._suppliedGeometry = null;
    }
    const p = this.part;
    if (p !== null) p.resetMaxMiter();
    this.invalidateMeasure();
    this.raiseChanged('geometry', old, value);
    // changing the geometry also changes the geometryString
    if (this.getBindsTwoWay()) {
      const part = this.part;
      if (part !== null) {
        this.updateSourceBindings(part, 'geometryString');
      }
    }
  }

  /**
   * When set, creates a {@link Geometry} and normalizes it from a given path string,
   * then sets the {@link #geometry} property on this Shape and
   * set the {@link GraphObject#position} to the amount computed by normalization.
   * The property getter simply returns the toString value of the {@link #geometry},
   * or the empty string if there is no Geometry value.
   *
   * See the <a href="../../intro/geometry.html">Geometry Path Strings Introduction page</a> for examples.
   * @since 1.1
   */
  /** @export */ get geometryString(): string {
    if (this.geometry === null) return '';
    return this.geometry.toString();
  }
  /** @export */ set geometryString(value: string) {
    const geo = Geometry.parse(value);
    const pos = geo.normalize();
    this.geometry = geo;
    const temp = Point.allocAt(-pos.x, -pos.y);
    this.position = temp;
    Point.free(temp);
  }

  /**
   * Gets or sets the whether the {@link GraphObject#position} of this shape denotes
   * the top-left corner of this shape in panel coordinates or the origin of this geometry's coordinate system.
   * Basically, this determines whether the {@link #strokeWidth} affects the rendered location.
   * A true value allows multiple shapes to be positioned precisely in a "Position" Panel independent of the stroke width.
   * The default is false.
   * @since 1.1
   */
  /** @export */ get isGeometryPositioned(): boolean { return this._isGeometryPositioned; }
  /** @export */ set isGeometryPositioned(value: boolean) {
    if (Debug) Util.checkPrimitive(value, 'boolean', Shape, 'isGeometryPositioned');
    const old = this._isGeometryPositioned;
    if (old !== value) {
      this._isGeometryPositioned = value;
      this.invalidateMeasure();
      this.raiseChanged('isGeometryPositioned', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Shape}
   */
  public invalidateGeometry(): void {
    // do not set _suppliedGeometry to null. It is set once in the geometry setter.
    // _geometry on the other hand is rebuilt all the time.
    this._geometry = null;
    // whatever calls invalidate geometry ought to also call invalidate measure if it needs it
  }

  /**
   * Gets or sets the {@link Brush} or string that describes how the geometry is filled when drawn.
   *
   * The default value is "black", causing the shape to be filled with solid black.
   * Any valid CSS string can specify a solid color, and the {@link Brush}
   * class can be used to specify a gradient or pattern.
   * A null fill will mean no fill is drawn and the filled portion
   * of the Shape will not be pickable.
   * A "transparent" fill is useful when wanting to allow a shape to be pickable
   * without obscuring any other objects behind it.
   * More information about the syntax of CSS color strings is available at:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color">CSS colors (mozilla.org)</a>.
   *
   * The geometry is filled before the {@link #stroke} is drawn.
   */
  /** @export */ get fill(): BrushLike { return this._fill; }
  /** @export */ set fill(value: BrushLike) {
    const old = this._fill;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'Shape.fill');
      if (value instanceof Brush) value.freeze();
      this._fill = value;
      this.invalidateDraw();
      this.raiseChanged('fill', old, value);
    }
  }

  /**
   * Gets or sets the {@link Brush} or string that describes how the geometry is drawn as if by a pen.
   *
   * The default value is "black", causing the shape to be outlined in black.
   * Any valid CSS string can specify a solid color, and the {@link Brush}
   * class can be used to specify a gradient or pattern.
   * A null stroke will mean no stroke is drawn.
   * A "transparent" stroke is useful when wanting to allow a shape to be pickable
   * without obscuring any other objects behind it.
   * More information about the syntax of CSS color strings is available at:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color">CSS colors (mozilla.org)</a>.
   *
   * The stroke is drawn after the geometry is filled with the {@link #fill} Brush.
   */
  /** @export */ get stroke(): BrushLike { return this._stroke; }
  /** @export */ set stroke(value: BrushLike) {
    const old = this._stroke;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'Shape.stroke');
      if (value instanceof Brush) value.freeze();
      this._stroke = value;
      this.invalidateDraw();
      this.raiseChanged('stroke', old, value);
    }
  }

  /**
   * Gets or sets the thickness of the stroke's pen.
   *
   * Value must be a real number greater than or equal to zero.
   * The default value is 1.0.
   * A value of zero will cause the stroke not to be drawn.
   *
   * The stroke width will affect the {@link GraphObject#measuredBounds} and {@link GraphObject#actualBounds} of this shape.
   * The stroke is drawn centered on the path of the geometry.
   */
  /** @export */ get strokeWidth(): number { return this._strokeWidth; }
  /** @export */ set strokeWidth(value: number) {
    const old = this._strokeWidth;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Shape, 'strokeWidth');
      // On setting, negative, infinite, and NaN values must be ignored
      if (value >= 0) {
        this._strokeWidth = value;
        this.invalidateMeasure();
        const p = this.part;
        if (p !== null) p.resetMaxMiter();
        this.raiseChanged('strokeWidth', old, value);
      } else {
        Util.throwRangeError(value, 'value >= 0', Shape, 'strokeWidth:value');
      }
    }
  }

  /**
   * Gets or sets the style for how the ends of the stroke's line are drawn.
   * The value must be one of "butt", "round", or "square". The default is "butt".
   *
   * For more information, see <a href="https://www.w3.org/TR/2dcontext/#dom-context-2d-linecap">Stroke Line Cap (w3.org)</a>.
   */
  /** @export */ get strokeCap(): ('butt' | 'round' | 'square') { return this._strokeCap; }
  /** @export */ set strokeCap(value: ('butt' | 'round' | 'square')) {
    const old = this._strokeCap;
    if (old !== value) {
      if (typeof value === 'string' &&
        (value === 'butt' || value === 'round' || value === 'square')) {
        this._strokeCap = value;
        this.invalidateDraw();
        this.raiseChanged('strokeCap', old, value);
      } else {
        Util.throwRangeError(value, '\"butt\", \"round\", or \"square\"', Shape, 'strokeCap');
      }
    }
  }

  /**
   * Gets or sets the type of corner that will be drawn for a stroke at the intersection of two straight segments of the geometry.
   * The value must be one of "miter", "bevel", or "round". The default is "miter".
   *
   * For more information, see <a href="https://www.w3.org/TR/2dcontext/#dom-context-2d-linejoin">Stroke Line Join (w3.org)</a>.
   */
  /** @export */ get strokeJoin(): ('miter' | 'bevel' | 'round') { return this._strokeJoin; }
  /** @export */ set strokeJoin(value: ('miter' | 'bevel' | 'round')) {
    const old = this._strokeJoin;
    if (old !== value) {
      if (typeof value === 'string' &&
        (value === 'miter' || value === 'bevel' || value === 'round')) {
        this._strokeJoin = value;
        this.invalidateDraw();
        this.raiseChanged('strokeJoin', old, value);
      } else {
        Util.throwRangeError(value, '\"miter\", \"bevel\", or \"round\"', Shape, 'strokeJoin');
      }
    }
  }

  /**
   * Gets or sets the style for the stroke's mitre limit ratio.
   * The value must be a real number greater than or equal to one.
   * The default is 10.0.
   *
   * For more information, see <a href="https://www.w3.org/TR/2dcontext/#dom-context-2d-miterlimit">Stroke Miter Limit (w3.org)</a>.
   */
  /** @export */ get strokeMiterLimit(): number { return this._strokeMiterLimit; }
  /** @export */ set strokeMiterLimit(value: number) {
    const old = this._strokeMiterLimit;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Shape, 'strokeMiterLimit');
      // On setting, zero, negative, infinite, and NaN values must be ignored
      if (value >= 1) {
        this._strokeMiterLimit = value;
        this.invalidateDraw();
        const p = this.part;
        if (p !== null) p.resetMaxMiter();
        this.raiseChanged('strokeMiterLimit', old, value);
      } else {
        if (Debug) Util.throwRangeError(value, 'value >= 1', Shape, 'strokeWidth:value');
      }
    }
  }

  /**
   * Gets or sets the dash array for creating dashed or dotted lines.
   * The value must be an array of positive numbers and zeroes,
   * or else null to indicate a solid line.
   * For example, the array [5, 10] would create dashes of 5 pixels and spaces of 10 pixels.
   * For more information, see <a href="https://www.w3.org/TR/2dcontext/#dom-context-2d-setlinedash">Stroke Line Dash Array (w3.org)</a>.
   *
   * The default value is null, resulting in a line without dashes or dots.
   * Setting an array with all zeroes will set the value to null.
   * @since 1.1
   */
  /** @export */ get strokeDashArray(): Array<number> | null { return this._strokeDashArray; }
  /** @export */ set strokeDashArray(value: Array<number> | null) {
    const old = this._strokeDashArray;
    if (old !== value) {
      // make sure its an array of numbers?
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', Shape, 'strokeDashArray:value');
      if (value !== null) {
        const l = value.length;
        let total = 0;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (Debug && typeof n !== 'number' || !(n >= 0) || !isFinite(n)) {
            Util.throwError('strokeDashArray:value ' + n + ' at index ' + i + ' must be a positive number or zero.');
          }
          total += n;
        }
        if (total === 0) {
          if (old === null) return; // no-op
          value = null; // never let strokeDashArray be all zeroes, ie [0, 0]
        }
      }

      this._strokeDashArray = value;
      this.invalidateDraw();
      this.raiseChanged('strokeDashArray', old, value);
    }
  }

  /**
   * Gets or sets the offset for dashed lines, used to start the drawing of the dash pattern with some space.
   * The value must be a real non-negative number. The default is zero.
   *
   * For more information, see <a href="https://www.w3.org/TR/2dcontext/#dom-context-2d-linedashoffset">Stroke Line Dash Offset (w3.org)</a>.
   * @since 1.1
   */
  /** @export */ get strokeDashOffset(): number { return this._strokeDashOffset; }
  /** @export */ set strokeDashOffset(value: number) {
    const old = this._strokeDashOffset;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, Shape, 'strokeDashOffset');
      // On setting, negative, infinite, and NaN values must be ignored
      if (value >= 0) {
        this._strokeDashOffset = value;
        this.invalidateDraw();
        this.raiseChanged('strokeDashOffset', old, value);
      }
    }
  }

  /**
   * Gets or sets the figure name, used to construct a {@link Geometry}.
   * The value must be a string. The default value is "None".
   *
   * The name can be any case but will always be canonicalized when set. For instance,
   * setting "roundedrectangle" will set the value of figure to "RoundedRectangle".
   * All of the predefined figures are shown in the <a href="../../samples/shapes.html">Shapes sample</a>.
   *
   * At most one of the following three properties may be set to a non-"None" value at the same time on the same shape:
   * {@link #figure}, {@link #toArrow}, {@link #fromArrow}.
   *
   * You can define your own named figures by calling the static function {@link Shape.defineFigureGenerator}.
   */
  /** @export */ get figure(): string { return this._figure; }
  /** @export */ set figure(value: string) {
    const old = this._figure;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Shape, 'figure');
      let cval = Geo.GeometryGenerators[value];
      if (typeof cval === 'function') {
        cval = value;  // VAL is already in canonical form
      } else {
        cval = Geo.GeometryGenerators[value.toLowerCase()]; // Canonicalize it
        if (!cval) Util.throwError('Unknown Shape.figure: ' + value);
      }
      if (old !== cval) {
        const p = this.part;
        if (p !== null) p.resetMaxMiter();
        this._figure = cval;
        this._suppliedGeometry = null;
        this.invalidateGeometry();
        this.invalidateMeasure();
        this.raiseChanged('figure', old, cval);
      }
    }
  }

  /**
   * Gets or sets the name of the kind of arrowhead that this shape should take
   * when this shape is an element of a {@link Link}.
   * Value must be a string.
   *
   * The default is "None", which means that this Shape is not an arrowhead, causing it to be the default Shape, a large filled Rectangle.
   * If you want to have an arrowhead Shape but sometimes not show an arrowhead, you can set or bind
   * the {@link GraphObject#visible} property, or you can set or bind this "toArrow" property to
   * be the empty string.
   * The arrowhead named "", an empty string, will display as nothing.
   *
   * The name can be any case but will always be canonicalized when set. For instance,
   * setting "opentriangle" will set the value of the arrowhead to "OpenTriangle".
   * All of the predefined arrowheads are shown in the <a href="../../samples/arrowheads.html">Arrowheads sample</a>.
   *
   * Setting this property may also set the {@link GraphObject#segmentIndex},
   * {@link GraphObject#segmentOrientation}, and {@link GraphObject#alignmentFocus} properties.
   * This shape should be an element of a {@link Link}.
   *
   * At most one of the following three properties may be set to a non-"None" value at the same time on the same shape:
   * {@link #figure}, {@link #toArrow}, {@link #fromArrow}.
   *
   * You can define your own named arrowheads by calling the static function {@link Shape.defineArrowheadGeometry}.
   *
   * You can see a copy of all of the built-in arrowhead definitions in this file: <a href="../../extensions/Arrowheads.js">Arrowheads.js</a>.
   */
  /** @export */ get toArrow(): string { return this._toArrow; }
  /** @export */ set toArrow(value: string) {
    const old = this._toArrow;
    if (value as any === true) value = 'Standard';
    else if (value as any === false) value = '';
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Shape, 'toArrow');
      const cval = Shape.canonicalizeArrowheadName(value);
      if (cval === null) {
        Util.throwError('Unknown Shape.toArrow: ' + value);
      } else if (old !== cval) {
        this._toArrow = cval;
        this._suppliedGeometry = null;
        this.invalidateGeometry();
        this.invalidateMeasure();
        this._setArrowheadPath(cval);
        this.raiseChanged('toArrow', old, cval);
      }
    }
  }

  /**
   * Gets or sets the name of the kind of arrowhead that this shape should take
   * when this shape is an element of a {@link Link}.
   * Value must be a string.
   * For bi-directional links the arrowhead name often starts with "Backward...".
   *
   * The default is "None", which means that this Shape is not an arrowhead, causing it to be the default Shape, a large filled Rectangle.
   * If you want to have an arrowhead Shape but sometimes not show an arrowhead, you can set or bind
   * the {@link GraphObject#visible} property, or you can set or bind this "toArrow" property to
   * be the empty string.
   * The arrowhead named "", an empty string, will display as nothing.
   *
   * The name can be any case but will always be canonicalized when set. For instance,
   * setting "opentriangle" will set the value of the arrowhead to "OpenTriangle".
   * All of the predefined arrowheads are shown in the <a href="../../samples/arrowheads.html">Arrowheads sample</a>.
   *
   * Setting this property may also set the {@link GraphObject#segmentIndex},
   * {@link GraphObject#segmentOrientation}, and {@link GraphObject#alignmentFocus} properties.
   * This shape should be an element of a {@link Link}.
   *
   * At most one of the following three properties may be set to a non-"None" value at the same time on the same shape:
   * {@link #figure}, {@link #toArrow}, {@link #fromArrow}.
   *
   * You can define your own named arrowheads by calling the static function {@link Shape.defineArrowheadGeometry}.
   *
   * You can see a copy of all of the built-in arrowhead definitions in this file: <a href="../../extensions/Arrowheads.js">Arrowheads.js</a>.
   */
  /** @export */ get fromArrow(): string { return this._fromArrow; }
  /** @export */ set fromArrow(value: string) {
    const old = this._fromArrow;
    if (value as any === true) value = 'Standard';
    else if (value as any === false) value = '';
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Shape, 'fromArrow');
      const cval = Shape.canonicalizeArrowheadName(value);
      if (cval === null) {
        Util.throwError('Unknown Shape.fromArrow: ' + value);
      } else if (old !== cval) {
        this._fromArrow = cval;
        this._suppliedGeometry = null;
        this.invalidateGeometry();
        this.invalidateMeasure();
        this._setArrowheadPath(cval);
        this.raiseChanged('fromArrow', old, cval);
      }
    }
  }

  /**
   * @hidden @internal
   * Some shared code. More could go here later but its not urgent.
   * @this {Shape}
   * @param {string} val canonicalized arrowhead name
   */
  private _setArrowheadPath(val: string): void {
    const diagram = this.diagram;
    if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
    this.segmentOrientationAlong();
    const alignmentSpot = Spot.MiddleRight;
    if (this._toArrow !== 'None') {
      this.segmentIndex = -1;
      this.alignmentFocus = alignmentSpot;
    } else if (this._fromArrow !== 'None') {
      this.segmentIndex = 0;
      this.alignmentFocus = new Spot(1 - alignmentSpot.x, alignmentSpot.y);  // Opposite horizontal alignment in Spot
    }
  }

  /**
   * Gets or sets the top-left Spot used by some {@link Panel}s for determining where in the shape other objects may be placed.
   * The value is normally {@link Spot.Default}, but you may want to set it to override the value that many {@link #figure}s use.
   */
  /** @export */ get spot1(): Spot { return this._spot1; }
  /** @export */ set spot1(value: Spot) {
    Util.checkClass(value, Spot, Shape, 'spot1');
    const old = this._spot1;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._spot1 = value;
      this.invalidateMeasure();
      this.raiseChanged('spot1', old, value);
    }
  }

  /**
   * Gets or sets the bottom-right Spot used by some {@link Panel}s for determining where in the shape other objects may be placed.
   * The value is normally {@link Spot.Default}, but you may want to set it to override the value that many {@link #figure}s use.
   */
  /** @export */ get spot2(): Spot { return this._spot2; }
  /** @export */ set spot2(value: Spot) {
    Util.checkClass(value, Spot, Shape, 'spot2');
    const old = this._spot2;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._spot2 = value;
      this.invalidateMeasure();
      this.raiseChanged('spot2', old, value);
    }
  }

  /**
   * Gets or sets a property for parameterizing the construction of a {@link Geometry} from a figure.
   * The meaning of this property depends on the particular figure.
   * The value must be a number; the default value is NaN.
   */
  /** @export */ get parameter1(): number { return this._parameter1; }
  /** @export */ set parameter1(value: number) {
    const old = this._parameter1;
    if (old !== value) {
      this._parameter1 = value;
      this.invalidateGeometry();
      this.invalidateMeasure();
      this.raiseChanged('parameter1', old, value);
    }
  }

  /**
   * Gets or sets a property for parameterizing the construction of a {@link Geometry} from a figure.
   * The meaning of this property depends on the particular figure.
   * The value must be a number; the default value is NaN.
   */
  /** @export */ get parameter2(): number { return this._parameter2; }
  /** @export */ set parameter2(value: number) {
    const old = this._parameter2;
    if (old !== value) {
      this._parameter2 = value;
      this.invalidateGeometry();
      this.invalidateMeasure();
      this.raiseChanged('parameter2', old, value);
    }
  }

  /**
   * This read-only property returns the natural bounds of this Shape as determined by its {@link #geometry}'s bounds.
   * The bounds will always include the (0,0) point.
   * If there is no geometry available, it returns a {@link Rect} with the {@link #desiredSize},
   * which may have NaN values for the width and height.
   */
  /** @export */ get naturalBounds(): Rect {
    if (this._geometry !== null) {
      const r = this._geometry.bounds;
      this._naturalBounds.assign(r);
      return this._naturalBounds;
    }
    const des = this.desiredSize;
    return new Rect(0, 0, des.width, des.height);
  }

  /**
   * Gets or sets a {@link GraphObject} that is drawn repeatedly along the path of the stroke of this shape.
   * This property may be set to a shared GraphObject; the GraphObject should not belong to any Panel.
   * Note that data bindings do <em>not</em> work in such shared GraphObjects, because they are not part of the visual tree.
   * The default value is null, causing no object to be drawn repeatedly.
   *
   * Typically the object is a small {@link Shape} or a {@link Picture}.
   * The larger the object is the worse the results will be, especially if the stroke has short segments or sharp curves.
   * The pathPattern object is not part of the measured bounds of the Shape, it is a cosmetic element only.
   *
   * The stroke is always drawn normally -- having a value for this property will draw the value along the stroke as well,
   * so it is commonplace to set the {@link #stroke} to "transparent" and the {@link #strokeWidth} to be as wide as
   * the height of the GraphObject being drawn along the stroke.
   *
   * Examples of path patterns can be seen in the <a href="../../samples/relationships.html">Relationships sample</a>.
   * @since 1.6
   */
  /** @export */ get pathPattern(): GraphObject | null { return this._pathPattern; }
  /** @export */ set pathPattern(value: GraphObject | null) {
    const old = this._pathPattern;
    if (old !== value) {
      if (Debug && value !== null) Util.checkClass(value, GraphObject, Shape, 'pathPattern');
      this._pathPattern = value;
      this.invalidateDraw();
      this.raiseChanged('pathPattern', old, value);
    }
  }

  // Later we could allow this to stretch Horizontally or Vertically ??
  /**
   * Gets or sets how the shape's geometry is proportionally created given its computed size.
   * Possible values are {@link GraphObject.None}, {@link GraphObject.Fill}, {@link GraphObject.Uniform}, and {@link GraphObject.Default}.
   * The default is {@link GraphObject.Default}, which resolves to {@link GraphObject.Fill} for most figures, though
   * some regular figures such as "Circle" and "Square" default to {@link GraphObject.Uniform}.
   */
  /** @export */ get geometryStretch(): EnumValue { return this._geometryStretch; }
  /** @export */ set geometryStretch(value: EnumValue) {
    const old = this._geometryStretch;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, Shape, 'geometryStretch');
      this._geometryStretch = value;
      this.raiseChanged('geometryStretch', old, value);
    }
  }

  /**
   * Gets or sets how frequently this shape should be drawn within a "Grid" or "Graduated" {@link Panel},
   * in multiples of the {@link Panel#gridCellSize} or {@link Panel#graduatedTickUnit}.
   * The default is 1.  Any new value must be a positive integer.
   */
  /** @export */ get interval(): number { return this._interval; }
  /** @export */ set interval(value: number) {
    const old = this._interval;
    if (Debug) Util.checkRealNumber(value, Shape, 'interval');
    value = Math.floor(value);
    if (old !== value) {
      // On setting, negative, infinite, and NaN values must be ignored
      if (value >= 0) {
        this._interval = value;
        // may need to update the Diagram's background grid
        const diagram = this.diagram;
        if (diagram !== null && this.panel === diagram.grid) {
          diagram._updateBackgroundGrid();
        }
        this.invalidateMeasure();
        const panel = this.panel;
        if (panel !== null) panel._graduatedIntervals = null;
        this.raiseChanged('interval', old, value);
      }
    }
  }

  /**
   * Gets or sets the fractional distance along the main shape of a "Graduated" {@link Panel} at which this kind of tick should start.
   * The default is 0. Any new value should range from 0 to 1.
   * @since 1.7
   */
  /** @export */ get graduatedStart(): number { return this._graduatedStart; }
  /** @export */ set graduatedStart(value: number) {
    const old = this._graduatedStart;
    if (Debug) Util.checkRealNumber(value, Shape, 'graduatedStart');
    if (old !== value) {
      if (value < 0) {
        value = 0;
      } else if (value > 1) {
        value = 1;
      }
      this._graduatedStart = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedStart', old, value);
    }
  }

  /**
   * Gets or sets the fractional distance along the main shape of a "Graduated" {@link Panel} at which this kind of tick should end.
   * The default is 1. Any new value should range from 0 to 1.
   * @since 1.7
   */
  /** @export */ get graduatedEnd(): number { return this._graduatedEnd; }
  /** @export */ set graduatedEnd(value: number) {
    const old = this._graduatedEnd;
    if (Debug) Util.checkRealNumber(value, Shape, 'graduatedEnd');
    if (old !== value) {
      if (value < 0) {
        value = 0;
      } else if (value > 1) {
        value = 1;
      }
      this._graduatedEnd = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedEnd', old, value);
    }
  }

  /**
   * Gets or sets the function to determine which values along a "Graduated" {@link Panel} will be skipped.
   * The default is null and doesn't skip any ticks.
   *
   * The function takes a number argument, a value between {@link Panel#graduatedMin} and {@link Panel#graduatedMax}.
   * The function will return a boolean, whether the tick will be skipped at the value of the argument.
   * @since 2.0
   */
  /** @export */ get graduatedSkip(): ((val: number) => boolean) | null { return this._graduatedSkip; }
  /** @export */ set graduatedSkip(value: ((val: number) => boolean) | null) {
    const old = this._graduatedSkip;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Shape, 'graduatedSkip');
      this._graduatedSkip = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedSkip', old, value);
    }
  }

  /**
   * @hidden @internal
   * @param {Shape} shape
   * @param {number} w
   * @param {number} h
   */
  public static makeGeometry(shape: Shape, w: number, h: number): Geometry | null {
    let geo: Geometry = null;
    // See if the shape has a toArrow, fromArrow, or Figure, in that order.
    if (shape.toArrow !== 'None') {
      geo = Shape.ArrowheadGeometries[shape.toArrow];
    } else if (shape.fromArrow !== 'None') {
      geo = Shape.ArrowheadGeometries[shape.fromArrow];
    } else { // normal figure
      // look up the geometry generator by name
      let func = Geo.GeometryGenerators[shape.figure];
      // maybe it found a name instead of a function -- look again
      if (typeof func === 'string') func = Geo.GeometryGenerators[func];
      if (func === undefined) Util.throwError('Unknown Shape.figure: ' + shape.figure);
      geo = func(shape, w, h);
      geo._inputWidth = w;
      geo._inputHeight = h;
    }
    if (geo === null) {
      const func = Geo.GeometryGenerators['Rectangle'];
      if (typeof func === 'function') geo = func(shape, w, h);
    }
    const epsilon = 0.00001;
    if (Debug && (geo.bounds.width > w + epsilon || geo.bounds.height > h + epsilon)) {
      Util.throwError('Geometry made with figure "' + shape.figure + '" has bounds ' + geo.bounds.toString() +
      ' that are too large for the given size (' + w + ',' + h + '). See documentation for Shape.defineFigureGenerator.');
    }

    return geo;
  }

  /**
   * This static function returns a read-only Map of named geometry generators.
   * The keys are figure names.
   * The values are either string synonyms for other figure names, or functions
   * that take a {@link Shape} and a width and a height and return a {@link Geometry}.
   *
   * The predefined shape figures can be seen in the <a href="../../samples/shapes.html">Shapes sample</a>.
   * @return {Map.<string,string|function(Shape, number, number):Geometry>}
   * @since 1.5
   */
  public static getFigureGenerators(): Map<string, string | ((a: Shape, b: number, c: number) => Geometry)> {
    const map = new Map<string, string | ((a: Shape, b: number, c: number) => Geometry)>();
    for (const name in Geo.GeometryGenerators) {
      if (name === name.toLowerCase()) continue;
      const f = Geo.GeometryGenerators[name];
      map.add(name, f);
    }
    map.freeze();
    return map;
  }

  /**
   * This static function defines a named figure geometry generator for Shapes.
   * Once this is called one can use the name as a value for {@link Shape#figure}.
   *
   * The first argument is the new figure name and must be a non-empty string that starts with a capital letter and that is not "None".
   *
   * If the second argument is a string this call defines a synonym for an existing figure generator.
   * Do not define cycles of synonyms -- the behavior will be undefined.
   *
   * If the second argument is a function,
   * the {@link Geometry} generator function's first argument is the {@link Shape} for which the function is producing a Geometry.
   * But this Shape argument may be null in some circumstances.
   * The second and third arguments are the desired width and height.
   * These will always be finite non-negative numbers.
   * The function may look at the {@link Shape#parameter1} and {@link Shape#parameter2} properties, which may be NaN,
   * to decide what geometry to create for the figure given its intended width and height.
   *
   * The function must return a Geometry; you may want to set {@link Geometry#spot1} and {@link Geometry#spot2} on it
   * to indicate where content should be placed within the figure when using an "Auto" Panel.
   * For some figures you may also want to set {@link Geometry#defaultStretch} to {@link GraphObject.Uniform}
   * in order to maintain the geometry's aspect ratio within the Shape.
   *
   * Generated figures must create a Geometry that is not larger than the supplied with and height. Doing so will signal an error.
   * @param {string} name a capitalized figure name ("RoundedRectangle"); must not be "" or "None"
   * @param {string|function(Shape,number,number):Geometry} func A function that takes (Shape,width,height) and returns a Geometry,
   *    or an existing figure generator name for which the new name will be a synonym.
   * @since 1.5
   */
  public static defineFigureGenerator(name: string, func: string | ((shape: Shape, width: number, height: number) => Geometry)): void {
    Util.checkPrimitive(name, 'string', Shape, 'defineFigureGenerator:name');
    if (typeof func === 'string') {
      if (Debug && (func === '' || !Geo.GeometryGenerators[func])) {
        Util.throwError('Shape.defineFigureGenerator synonym must not be empty or None or not a defined figure name: ' + func);
      }
    } else {
      Util.checkPrimitive(func, 'function', Shape, 'defineFigureGenerator:func');
    }
    const low = name.toLowerCase();
    if (Debug && (name === '' || name === low)) {
      Util.throwError('Shape.defineFigureGenerator name must not be empty or all-lower-case: ' + name);
    }
    const gens = Geo.GeometryGenerators;
    gens[name] = func;
    gens[low] = name;
  }

  /**
   * This static function returns a read-only Map of named arrowhead geometries.
   * The keys are arrowhead names.
   * The values are {@link Geometry} objects.
   *
   * The predefined arrowheads can be seen in the <a href="../../samples/arrowheads.html">Arrowheads sample</a>.
   * @return {Map.<string,Geometry>}
   * @since 1.5
   */
  public static getArrowheadGeometries(): Map<string, Geometry> {
    const map = new Map<string, Geometry>();
    // Each predefined arrowhead Geometry is now generated on demand;
    // so Shape.getArrowheadGeometries needs to fill out the table without overwriting any
    // geometries defined via Shape.defineArrowheadGeometry.
    for (const name in Geo.ArrowheadGeometryStrings) {
      if (Shape.ArrowheadGeometries[name] === undefined) {
        const geo = Geometry.parse(Geo.ArrowheadGeometryStrings[name], false);
        Shape.ArrowheadGeometries[name] = geo;
        // also remember mapping from lowercased name to canonical cased name
        const lower = name.toLowerCase();
        if (lower !== name) Shape.ArrowheadGeometries[lower] = name;
      }
    }
    for (const name in Shape.ArrowheadGeometries) {
      if (name === name.toLowerCase()) continue;
      const geo2 = Shape.ArrowheadGeometries[name];
      if (geo2 instanceof Geometry) map.add(name, geo2);
    }
    map.freeze();
    return map;
  }

  /**
   * This static function defines a named arrowhead geometry.
   * Once this is called one can use the name as a value for {@link Shape#toArrow} or {@link Shape#fromArrow}.
   *
   * The first argument is the new arrowhead name and must be a non-empty string that starts with a capital letter and that is not "None".
   *
   * If the second argument is a string, it is converted into a {@link Geometry} by calling
   * `go.Geometry.parse(pathstr, false)`,
   * which may throw an error if there are problems with the syntax of the string.
   * @param {string} name a capitalized arrowhead name ("OpenTriangle"); must not be "" or "None"
   * @param {Geometry|string} pathstr a {@link Geometry} or a Geometry path string, e.g. "m 0,0 l 8,4 -8,4"
   * @since 1.5
   */
  public static defineArrowheadGeometry(name: string, pathstr: Geometry | string): void {
    Util.checkPrimitive(name, 'string', Shape, 'defineArrowheadGeometry:name');
    let geo: Geometry = null;
    if (typeof pathstr === 'string') {
      Util.checkPrimitive(pathstr, 'string', Shape, 'defineArrowheadGeometry:pathstr');
      geo = Geometry.parse(pathstr, false);
    } else {
      Util.checkClass(pathstr, Geometry, Shape, 'defineArrowheadGeometry:pathstr');
      geo = pathstr;
    }
    const low = name.toLowerCase();
    if (Debug && name === '' || low === 'none' || name === low) Util.throwError('Shape.defineArrowheadGeometry name must not be empty or None or all-lower-case: ' + name);
    const gens = Shape.ArrowheadGeometries;
    gens[name] = geo;
    gens[low] = name;
  }

  /**
   * @hidden @internal
   */
  private static ArrowheadGeometries: PropertyCollection = new PropertyCollection();

  /**
   * @hidden @internal
   */
  private static PresetGeometries: PropertyCollection = new PropertyCollection();

  /**
   * @hidden @internal
   * @param {string} name a possible arrowhead name, in any case
   * @return {string|null}
   */
  private static canonicalizeArrowheadName(name: string): string | null {
    // see if the given name is either a canonical name in ArrowheadGeometries
    // (in which case the value will be an instance of a Geometry),
    // or if it is a lowercase name that maps to a canonical name
    let x = Shape.ArrowheadGeometries[name];
    if (x === undefined) {
      const lower = name.toLowerCase();
      if (lower === 'none') return 'None';
      x = Shape.ArrowheadGeometries[lower];
      if (x === undefined) {
        // find the canonical name for the given arrowhead name
        let canon = null;
        for (const cn in Geo.ArrowheadGeometryStrings) {
          if (cn.toLowerCase() === lower) {
            canon = cn;
            break;
          }
        }
        if (canon !== null) {
          // parse string from Geo.ArrowheadGeometryStrings and keep cached
          const geo = Geometry.parse(Geo.ArrowheadGeometryStrings[canon], false);
          Shape.ArrowheadGeometries[canon] = geo;
          // also remember mapping from lowercased name to canonical cased name
          if (lower !== canon) Shape.ArrowheadGeometries[lower] = canon;
          return canon;
        }
      }
    }
    if (typeof x === 'string') return x;
    if (x instanceof Geometry) return name;
    return null;
  }
} // end class
Shape['className'] = 'Shape';

Shape['getFigureGenerators'] = Shape.getFigureGenerators;
Shape['defineFigureGenerator'] = Shape.defineFigureGenerator;
Shape['getArrowheadGeometries'] = Shape.getArrowheadGeometries;
Shape['defineArrowheadGeometry'] = Shape.defineArrowheadGeometry;

// import { Brush, BrushLike } from './brush';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { Geo } from './geo';
// import { Geometry, PathFigure, PathSegment } from './geometry';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { IContext } from './surfaceInterfaces';
// import { Transform } from './transform';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

// import { CanvasSurface, CanvasSurfaceContext } from './abstractCanvas';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { PropertyCollection } from './propertyCollection';
// import { IContext } from './surfaceInterfaces';
// import { Util } from './util';

/**
 * A TextBlock is a {@link GraphObject} that displays a {@link #text} string in a given {@link #font}.
 *
 * The size and appearance of the text is specified by {@link #font},
 * which takes a well-formed CSS string as its value.
 * The order of the CSS properties given is important for cross-browser compatibility,
 * and should be given in this order:
 *
 * <em>"font-style font-variant font-weight font-size font-family"</em>
 *
 * For example, "Italic small-caps bold 32px Georgia, Serif" is a valid font string
 * using every CSS font property. Note that not all browsers may support every property.
 *
 * Text is drawn using the {@link #stroke} brush, which may be any CSS color string or a {@link Brush}.
 *
 * Some created TextBlocks:
 * ```js
 * var $ = go.GraphObject.make;  // for conciseness in defining GraphObjects
 *
 * // A TextBlock with text and stroke properties set:
 * $(go.TextBlock, { text: "Hello World", stroke: "gray" })
 * // Alternatively:
 * $(go.TextBlock, "Hello World", { stroke: "gray" })
 * ```
 *
 * TextBlocks typically receive a natural size based on their text and font strings,
 * but often a width is given in order to cause the text to wrap at a certain place.
 * In order for wrapping to occur, the {@link #wrap} property must not be {@link TextBlock.None}.
 *
 * TextBlocks can be edited by users using the {@link TextEditingTool}.
 * The {@link HTMLInfo} that a given TextBlock uses as its text editor can be customized
 * by setting the {@link #textEditor} property. For an example of custom text editing tool use,
 * see the <a href="../../samples/customTextEditingTool.html">Custom TextEditingTool Sample</a>.
 * <p class="boxread">
 * For examples of TextBlock possibilities and functionality,
 * see the <a href="../../intro/textBlocks.html">Introduction page on TextBlocks</a>.
 * @extends GraphObject
 * @unrestricted
 */
export class TextBlock extends GraphObject {
  /** @hidden @internal */ public static className = 'TextBlock';
  /** @hidden @internal */ private _measuredWidth: number | null;
  /** @hidden @internal */ private _measuredHeight: number | null;
  /** @hidden @internal */ private _editable: boolean;
  /** @hidden @internal */ private _textEditor: HTMLInfo | null;
  /** @hidden @internal */ private _textValidation: ((thisTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null;
  /** @hidden @internal */ private _textEdited: ((thisTextBlock: TextBlock, oldString: string, newString: string) => void) | null;
  /** @hidden @internal */ private _choices: Array<string> | null;
  /** @hidden @internal */ private _errorFunction: ((tool: TextEditingTool, oldString: string, newString: string) => void) | null;
  /** @hidden @internal */ private _text: string;
  /** @hidden @internal */ private _renderedText: string;
  /** @hidden @internal */ private _stroke: BrushLike;
  /** @hidden @internal */ private _font: string;
  /** @hidden @internal */ private _textAlign: ('start'| 'end' | 'left' | 'right' | 'center');
  /** @hidden @internal */ private _flip: EnumValue;
  /** @hidden @internal */ private _verticalAlignment: Spot;
  /** @hidden @internal */ private _isMultiline: boolean;
  /** @hidden @internal */ private _isUnderline: boolean;
  /** @hidden @internal */ private _isStrikethrough: boolean;
  /** @hidden @internal */ private _overflow: EnumValue;
  /** @hidden @internal */ private _wrap: EnumValue;
  /** @hidden @internal */ private _lineCount: number;
  /** @hidden @internal */ private _spacingAbove: number;
  /** @hidden @internal */ private _spacingBelow: number;
  /** @hidden @internal */ private _maxLines: number;
  /** @hidden @internal */ private _interval: number;
  /** @hidden @internal */ private _graduatedStart: number;
  /** @hidden @internal */ private _graduatedEnd: number;
  /** @hidden @internal */ private _graduatedFunction: ((val: number) => string) | null;
  /** @hidden @internal */ private _graduatedSkip: ((val: number) => boolean) | null;
  /** @hidden @internal */ public _maxPosition: number;
  /** @hidden @internal */ public _maxLineWidth: number;
  /** @hidden @internal */ public _lastAvailableWidth: number;
  /** @hidden @internal */ public _fontHeight: number | null;
  /** @hidden @internal */ public _numLines: number;
  /** @hidden @internal */ public _arrSize: Array<number> | null;
  /** @hidden @internal */ public _arrText: Array<string> | null;
  /** @hidden @internal */ public _firstLine: string;

  /**
   * A newly constructed {@link TextBlock} has no string to show; if it did,
   * it would draw the text, wrapping if needed, in the default font using a black stroke.
   */
  constructor() {
    super();
    if (!TextBlock._hasInit) {
      TextBlock.initStatic();
      TextBlock._hasInit = true;
    }
    this._text = '';
    this._renderedText = ''; // Only used if there's a single line of text, in lieu of TextBlockMetrics
    this._stroke = 'black';
    this._font = '13px sans-serif';
    this._textAlign = 'start';

    this._flip = GraphObject.None;
    this._verticalAlignment = Spot.Top;
    this._isMultiline = true;
    this._isUnderline = false;
    this._isStrikethrough = false;
    this._overflow = TextBlock.OverflowClip;

    // italics and bold are up to the user and we have no way of controlling them
    // (short of parsing the 'text' property)
    this._wrap = TextBlock.WrapDesiredSize;

    this._lineCount = 0;
    this._maxLineWidth = 0;
    // this is the absolute line position of the measured text
    // it is used to determine if we need to crop text
    this._maxPosition = 0;

    // Invalidated along with measure, cached from measure
    // ?? they may not be necessary
    this._measuredWidth = null;
    this._measuredHeight = null;

    this._editable = false;
    this._textEditor = null;
    this._textValidation = null;
    this._textEdited = null;
    this._choices = null;

    this._errorFunction = null;

    this._spacingAbove = 0;
    this._spacingBelow = 0;
    this._maxLines = Infinity;
    this._lastAvailableWidth = 0;
    this._fontHeight = null;
    this._numLines = 0;
    this._arrSize = null; // @@@ ???
    this._arrText = null; // @@@ ???
    this._firstLine = '';

    // for Graduated Panel:
    this._interval = 1;
    this._graduatedStart = 0;
    this._graduatedEnd = 1;
    this._graduatedFunction = null;
    this._graduatedSkip = null;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned TextBlock.
   * @this {TextBlock}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._text = this._text;
    copy._renderedText = this._renderedText;
    copy._stroke = this._stroke;
    copy._font = this._font;
    copy._textAlign = this._textAlign;
    copy._flip = this._flip;
    copy._verticalAlignment = this._verticalAlignment;
    copy._isMultiline = this._isMultiline;
    copy._isUnderline = this._isUnderline;
    copy._isStrikethrough = this._isStrikethrough;
    copy._overflow = this._overflow;
    copy._wrap = this._wrap;
    copy._lineCount = this._lineCount;
    copy._maxPosition = this._maxPosition;
    copy._maxLineWidth = this._maxLineWidth;
    copy._measuredWidth = this._measuredWidth;
    copy._measuredHeight = this._measuredHeight;

    copy._editable = this._editable;
    copy._textEditor = this._textEditor;
    copy._textValidation = this._textValidation;
    copy._textEdited = this._textEdited;
    copy._choices = this._choices;
    copy._errorFunction = this._errorFunction;
    copy._spacingAbove = this._spacingAbove;
    copy._spacingBelow = this._spacingBelow;
    copy._maxLines = this._maxLines;

    copy._fontHeight = this._fontHeight;
    copy._numLines = this._numLines;
    copy._arrSize = this._arrSize;
    copy._arrText = this._arrText;
    copy._firstLine = this._firstLine;
    copy._lastAvailableWidth = this._lastAvailableWidth;

    copy._interval = this._interval;
    copy._graduatedStart = this._graduatedStart;
    copy._graduatedEnd = this._graduatedEnd;
    copy._graduatedFunction = this._graduatedFunction;
    copy._graduatedSkip = this._graduatedSkip;
  }

  /**
   * @hidden @internal
   * Called by Panel for graduated Panel labels
   * @this {TextBlock}
   * @param {GraphObject} elt
   */
  public _copyGraduatedLabelPropertiesFrom(elt: GraphObject): void {
    // GraphObject properties needed for measure or arrange
    this._flags = elt._flags | 0x00000800 | 0x00001000;
    this._opacity = elt.opacity;
    this._background = elt.background;  // Brushes can be shared
    this._areaBackground = elt.areaBackground;
    this._desiredSize = elt.desiredSize.copyFrozen();
    this._minSize = elt.minSize.copyFrozen();  // may be shared if the Spot is constant
    this._maxSize = elt.maxSize.copyFrozen();
    this._localTransform = elt._localTransform.copy();
    this._scale = elt.scale;
    this._angle = elt.angle;
    this._stretch = elt.stretch;
    this._margin = (elt.margin as Margin).copyFrozen();
    this._alignment = elt.alignment.copyFrozen();
    this._alignmentFocus = elt.alignmentFocus.copyFrozen();
    this._segmentFraction = elt.segmentFraction;
    this._segmentOffset = elt.segmentOffset.copyFrozen();
    this._segmentOrientation = elt.segmentOrientation;

    if (elt._clipRect !== null) {
      this._clipRect = elt._clipRect.copy();
    }
    this._shadowVisible = elt.shadowVisible;

    // TextBlock properties needed for measure or arrange
    if (elt instanceof TextBlock) {
      this._text = elt._text;
      this._stroke = elt._stroke;
      this._font = elt._font;
      this._textAlign = elt._textAlign;
      this._flip = elt._flip;
      this._verticalAlignment = elt._verticalAlignment;
      this._isMultiline = elt._isMultiline;
      this._isUnderline = elt._isUnderline;
      this._isStrikethrough = elt._isStrikethrough;
      this._overflow = elt._overflow;
      this._wrap = elt._wrap;
      this._fontHeight = null; // need to remeasure stringHeight
      this._spacingAbove = elt._spacingAbove;
      this._spacingBelow = elt._spacingBelow;
      this._maxLines = elt._maxLines;
      this._lastAvailableWidth = elt._lastAvailableWidth;
      this._interval = elt._interval;
      this._graduatedStart = elt._graduatedStart;
      this._graduatedEnd = elt._graduatedEnd;
      this._graduatedFunction = elt._graduatedFunction;
      this._graduatedSkip = elt._graduatedSkip;
    }
  }

  /**
   * @hidden @internal
   * Called by Panel for graduated Panel labels
   * @this {TextBlock}
   */
  public _restoreMeasurements(labinfo: GraduatedMark) {
    // text info from measuring
    this._text = labinfo.text || '';
    this._angle = labinfo.labelAngle || 0;
    this._lineCount = labinfo.lineCount || 0;
    this.setMetrics(labinfo.lines);
    if (labinfo.naturalBounds) this._naturalBounds = labinfo.naturalBounds;
    // bounds info from measuring
    if (labinfo.actualBounds) {
      const ab = labinfo.actualBounds;
      this.arrange(ab.x, ab.y, ab.width, ab.height);
    }
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {TextBlock}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === TextBlock) {
      this.wrap = arg;
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @return {string}
   */
  public toString(): string {
    if (this._text.length > 22) {
      return 'TextBlock(\"' + this._text.substring(0, 20) + '\"...)';
    } else {
      return 'TextBlock(\"' + this._text + '\")';
    }
  }

  //    * This static function gets the current ellipsis character or string for
  //  * TextBlocks that have {@link TextBlock#overflow} set to {@link TextBlock.Ellipsis}.
  //  * @since 1.4
  /**
   * @hidden
   * Undocumented
   */
  public static getEllipsis(): string {
    return TextBlock._ellipsisCharacter;
  }

  //    * This static function sets the ellipsis character or string for all
  //  * TextBlocks that have {@link TextBlock#overflow} set to {@link TextBlock.Ellipsis}.
  //  *
  //  * The default is `"..."`, which works with all fonts.
  //  *
  //  * Note when using the ellipsis character, "&hellip;",  you will need to specify a compatible
  //  * character set for your page, like UTF-8.
  //  * @since 1.4
  /**
   * @hidden
   * Undocumented
   * @param {string} val
   */
  public static setEllipsis(val: string): void {
    TextBlock._ellipsisCharacter = val;
    TextBlock._fontEllipsisWidth = new PropertyCollection(); // clear the map of font strings to width values
    TextBlock._fontEllipsisWidthCount = 0;
  }

  /**
   * Gets the function that, given the TextBlock and numerical text height, computes the position to draw the baseline of a line of text in all TextBlocks.
   * By default this is null and default behavior returns `(textHeight * 0.75)`.
   *
   * Note: This affects drawing only, and does not change TextBlock measurement calculations.
   * @return {(textBlock: TextBlock, textHeight: number) => number | null}
   * @since 2.0
   */
  public static getBaseline(): ((textBlock: TextBlock, textHeight: number) => number) | null {
    return TextBlock.baselineComputation;
  }

  /**
   * Sets the function that, given the TextBlock and numerical text height, computes the position to draw the baseline of a line of text in all TextBlocks.
   *
   * Note: This affects drawing only, and does not change TextBlock measurement calculations.
   * @param {(textBlock: TextBlock, textHeight: number) => number | null} value
   * @since 2.0
   */
  public static setBaseline(value: ((textBlock: TextBlock, textHeight: number) => number) | null) {
    TextBlock.baselineComputation = value;
    const diags = Util.getDiagramsTraversed();
    const l = diags.length;
    for (let i = 0; i < l; i++) {
      diags[i].redraw();
    }
  }

  /**
   * @hidden @internal
   */
  public static baselineComputation: ((textBlock: TextBlock, textHeight: number) => number) | null = null;

  /**
   * Gets the function that, given the TextBlock and numerical text height, computes the position to draw the underline of a line of text in all TextBlocks.
   * By default this is null and default behavior returns `(textHeight * 0.75)`.
   *
   * Note: This affects drawing only, and does not change TextBlock measurement calculations.
   * @return {(textBlock: TextBlock, textHeight: number) => number | null}
   * @since 2.0
   */
  public static getUnderline(): ((textBlock: TextBlock, textHeight: number) => number) | null {
    return TextBlock.underlineComputation;
  }

  /**
   * Sets the function that, given the TextBlock and numerical text height, computes the position to draw the underline of a line of text in all TextBlocks.
   *
   * Note: This affects drawing only, and does not change TextBlock measurement calculations.
   * @param {(textBlock: TextBlock, textHeight: number) => number | null} value
   * @since 2.0
   */
  public static setUnderline(value: ((textBlock: TextBlock, textHeight: number) => number) | null) {
    TextBlock.underlineComputation = value;
    const diags = Util.getDiagramsTraversed();
    const l = diags.length;
    for (let i = 0; i < l; i++) {
      diags[i].redraw();
    }
  }

  /**
   * @hidden @internal
   */
  public static underlineComputation: ((textBlock: TextBlock, textHeight: number) => number) | null = null;

  /**
   * Used as a value for {@link TextBlock#wrap}, the TextBlock will not wrap its text.
   * @constant
   */
  public static None = new EnumValue(TextBlock, 'None', 0);

  /**
   * Used as a value for {@link TextBlock#wrap}, the TextBlock will wrap text, making the width of
   * the TextBlock equal to the width of the longest line.
   * @constant
   */
  public static WrapFit = new EnumValue(TextBlock, 'WrapFit', 1);

  /**
   * Used as the default value for {@link TextBlock#wrap}, the TextBlock will wrap text and the width of
   * the TextBlock will be the desiredSize's width, if any.
   * @constant
   */
  public static WrapDesiredSize = new EnumValue(TextBlock, 'WrapDesiredSize', 2);

  /**
   * Used a a value for {@link TextBlock#wrap}, the TextBlock will attempt to wrap at each character, allowing
   * breaks within "words."
   * @since 2.0
   * @constant
   */
  public static WrapBreakAll = new EnumValue(TextBlock, 'WrapBreakAll', 3);

  /**
   * Used as the default value for {@link TextBlock#overflow}: if the width is too small to display all text,
   * the TextBlock will clip.
   * @constant
   * @since 1.4
   */
  public static OverflowClip = new EnumValue(TextBlock, 'OverflowClip', 0);

  /**
   * Used as a value for {@link TextBlock#overflow}: if the width is too small to display all text,
   * the TextBlock will display an ellipsis.
   * @constant
   * @since 1.4
   */
  public static OverflowEllipsis = new EnumValue(TextBlock, 'OverflowEllipsis', 1);

  /**
   * @hidden @internal
   * Call GraphObject's invalidateMeasure and invalidate private cached data
   * @this {TextBlock}
   */
  public invalidateMeasure(): void {
    super.invalidateMeasure();
    this._measuredWidth = null;
    this._measuredHeight = null;
  }

  /**
   * Gets or sets the current font settings.
   * The font property must be a valid CSS string describing a font.
   * The font string can accept several CSS properties but they must be
   * in a specific order in order to render correctly across all browsers:
   *
   * `"font-style font-variant font-weight font-size font-family"`
   *
   * For example, `"Italic small-caps bold 32px Georgia, Serif"` is a valid font string
   * using every CSS font property. Not every browser can render every font option.
   * For more information about CSS font syntax, see <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/font">CSS fonts (mozilla.org)</a>.
   *
   * If your Node sizes depend on TextBlocks, it is best to ensure any custom fonts you are using are finished loading before you load your Diagram.
   * This will ensure nodes are sized appropriately for the initial Diagram layout.
   *
   * The default font is "13px sans-serif".
   */
  /** @export */ get font(): string { return this._font; }
  /** @export */ set font(value: string) {
    const old = this._font;
    if (old !== value) {
      if (Debug) {
        Util.checkPrimitive(value, 'string', TextBlock, 'font');
        if (!TextBlock.isValidFont(value)) Util.throwError('Not a valid font: \"' + value + '\"');
      }
      this._font = value;
      this._fontHeight = null;
      this.invalidateMeasure();
      this.raiseChanged('font', old, value);
    }
  }

  // * This static predicate returns true if the given font is a valid font.
  /**
   * @hidden
   * Undocumented
   * @param {string} font
   * @return {boolean}
   */
  public static isValidFont(font: string): boolean {
    // ensure sharedTempCtx has been initialized
    if (!TextBlock._hasInit) {
      TextBlock.initStatic();
      TextBlock._hasInit = true;
    }
    // first check for the cached case, already known valid
    if (TextBlock._sharedTempCtx === null) return true;
    const origfont = TextBlock._sharedTempCtx.font;
    if (font === origfont) return true;

    // assume most fonts are different than the default one
    const deffont = '10px sans-serif';  // and assume this is known to be valid
    if (font === deffont) return true;
    // need to see if the context's font actually changes when set to FONT
    TextBlock._sharedTempCtx.font = deffont;
    let oldfont = deffont;
    TextBlock._sharedTempCtx.font = font;
    let newfont = TextBlock._sharedTempCtx.font;
    if (newfont !== oldfont) {
      // restore cache state
      TextBlock._sharedTempCtx.font = origfont;
      return true;
    }

    // just in case the FONT is the same as '10px sans-serif' but is not string= to it,
    // e.g. '100% sans-serif',
    // set the context font to be something definitely different, assuming '19px serif' is valid
    TextBlock._sharedTempCtx.font = '19px serif';
    oldfont = TextBlock._sharedTempCtx.font;
    TextBlock._sharedTempCtx.font = font;
    newfont = TextBlock._sharedTempCtx.font;
    TextBlock._sharedTempCtx.font = origfont;  // restore cache state
    return (newfont !== oldfont);
  }

  /**
   * Gets or sets the TextBlock's text string. The default is an empty string.
   * The text of a TextBlock, along with the values of {@link #font}, {@link #wrap},
   * {@link #isMultiline} and sizing restrictions are what naturally determine
   * the size of the TextBlock.
   *
   * The text in textblocks can include manual line-breaks by using the character escape, `\n`.
   *
   * Leading and trailing whitespace is eliminated in each line of TextBlock text.
   *
   * If {@link #editable} is set to true, users can edit textblocks with the {@link TextEditingTool}.
   */
  /** @export */ get text(): string { return this._text; }
  /** @export */ set text(value: string) {
    const old = this._text;
    // convert the new value to be a string, if necessary
    if (value !== null && value !== undefined) {
      value = value.toString();
    } else {
      value = '';
    }
    if (old !== value) {
      this._text = value;
      this.invalidateMeasure();
      this.raiseChanged('text', old, value);
    }
  }

  /**
   * Gets or sets the alignment location in the TextBlock's given space.
   * The only possible values are `"start"`, `"end"`, `"left"`, `"right"`, and `"center"`.
   * Any other value is invalid.
   *
   * This property is most pertinent when the TextBlock has multiple lines of text,
   * or when the TextBlock is given a size that differs from the text's natural size (such as with {@link #desiredSize}).
   *
   * In left-to-right writing systems, `"start"` and `"left"` are synonymous, as are `"end"` and `"right"`.
   *
   * The default is `"start"`.
   * @see {@link #verticalAlignment}
   */
  /** @export */ get textAlign(): ('start'| 'end' | 'left' | 'right' | 'center') { return this._textAlign; }
  /** @export */ set textAlign(value: ('start'| 'end' | 'left' | 'right' | 'center')) {
    const old = this._textAlign;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', TextBlock, 'textAlign');
      if (value === 'start' || value === 'end' || value === 'left' ||
        value === 'right' || value === 'center') {
        this._textAlign = value;
        this.invalidateDraw();
        this.raiseChanged('textAlign', old, value);
      } else {
        if (Debug) Util.throwRangeError(value, '\"start\", \"end\", \"left\", \"right\", or \"center\"', TextBlock, 'textAlign');
      }
    }
  }

  /**
   * Gets or sets how the TextBlock is displayed: Either normally or with a Horizontal or Vertical flip or both.
   *
   * Possible values are {@link GraphObject.None}, {@link GraphObject.FlipHorizontal}, {@link GraphObject.FlipVertical}, or {@link GraphObject.FlipBoth}.
   * The default is {@link GraphObject.None}.
   * @since 2.0
   */
  /** @export */ get flip(): EnumValue { return this._flip; }
  /** @export */ set flip(value: EnumValue) {
    const old = this._flip;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, TextBlock, 'flip');
      this._flip = value;
      this.invalidateDraw();
      this.raiseChanged('flip', old, value);
    }
  }

  /**
   * Gets or sets the vertical alignment {@link Spot} of this TextBlock, used when
   * the TextBlock has more available vertical space than it needs to draw all lines.
   *
   * The default value is {@link Spot.Top}, which aligns the TextBlock to the top of its available space.
   *
   * The {@link #textAlign} is often used along with this property to specify
   * where the should be positioned in its available space.
   *
   * This does not affect TextBlock coordinates or bounds, it only affects where text is drawn within the given area.
   * @see {@link #textAlign}
   * @since 1.7
   */
  /** @export */ get verticalAlignment(): Spot { return this._verticalAlignment; }
  /** @export */ set verticalAlignment(value: Spot) {
    const old = this._verticalAlignment;
    if (!old.equals(value)) {
      if (Debug) {
        Util.checkClass(value, Spot, TextBlock, 'verticalAlignment');
        if (value.isNoSpot()) {
          Util.throwError('TextBlock.verticalAlignment for ' + this + ' must be a real Spot, not:' + value);
        }
      }
      value = value.copyFrozen();
      this._verticalAlignment = value;
      this.invalidateArrange();
      this.raiseChanged('verticalAlignment', old, value);
    }
  }

  /**
   * This read-only property returns the natural bounds of this TextBlock in local coordinates,
   * as determined by its {@link #font} and {@link #text} string, and optionally its {@link #desiredSize}.
   */
  /** @export */ get naturalBounds(): Rect {
    if (!this._naturalBounds.isReal()) {
      // Usually happens in measureProtected
      const sz = Size.alloc();
      this.measureString(this._text, 999999, sz);
      let w = sz.width;
      Size.free(sz);
      let h = this.computeHeight(w);
      const ds = this.desiredSize;
      if (!isNaN(ds.width)) w = ds.width;
      if (!isNaN(ds.height)) h = ds.height;
      const nb = this._naturalBounds;
      nb.assignWidthHeight(w, h);
    }
    return this._naturalBounds;
  }

  /**
   * Gets or sets whether or not the text displays multiple lines or embedded newlines.
   * If this is false, all characters including and after the first newline will be omitted.
   * The default is true.
   * @see {@link #maxLines}
   */
  /** @export */ get isMultiline(): boolean { return this._isMultiline; }
  /** @export */ set isMultiline(value: boolean) {
    const old = this._isMultiline;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', TextBlock, 'isMultiline');
      this._isMultiline = value;
      this.invalidateMeasure();
      this.raiseChanged('isMultiline', old, value);
    }
  }

  /**
   * Gets or sets whether or not the text is underlined.
   * The default is false.
   * @see {@link #isStrikethrough}
   * @since 1.2
   */
  /** @export */ get isUnderline(): boolean { return this._isUnderline; }
  /** @export */ set isUnderline(value: boolean) {
    const old = this._isUnderline;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', TextBlock, 'isUnderline');
      this._isUnderline = value;
      this.invalidateDraw();
      this.raiseChanged('isUnderline', old, value);
    }
  }

  /**
   * Gets or sets whether or not the text has a strikethrough line (line-through).
   * The default is false.
   * @see {@link #isUnderline}
   * @since 1.2
   */
  /** @export */ get isStrikethrough(): boolean { return this._isStrikethrough; }
  /** @export */ set isStrikethrough(value: boolean) {
    const old = this._isStrikethrough;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', TextBlock, 'isStrikethrough');
      this._isStrikethrough = value;
      this.invalidateDraw();
      this.raiseChanged('isStrikethrough', old, value);
    }
  }

  /**
   * Gets or sets whether the text should be wrapped if it is too long to fit on one line.
   *
   * Possible values are {@link TextBlock.WrapDesiredSize}, {@link TextBlock.WrapFit}, {@link TextBlock.WrapBreakAll},
   * and {@link TextBlock.None}.
   *
   * The default value is {@link TextBlock.WrapDesiredSize}.
   * @see {@link #overflow}
   */
  /** @export */ get wrap(): EnumValue { return this._wrap; }
  /** @export */ set wrap(value: EnumValue) {
    const old = this._wrap;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, TextBlock, TextBlock, 'wrap');
      this._wrap = value;
      this.invalidateMeasure();
      this.raiseChanged('wrap', old, value);
    }
  }

  /**
   * Gets or sets how text that is too long to display should be handled.
   *
   * Possible values are {@link TextBlock.OverflowClip} and {@link TextBlock.OverflowEllipsis}.
   * For OverflowEllipsis to work, you must constrain the available size of the TextBlock in some way,
   * such as setting {@link #wrap} to {@link TextBlock.None},
   * or limiting the number of lines with {@link #maxLines} or a height constraint.
   *
   * The default value is {@link TextBlock.OverflowClip}.
   * @see {@link #wrap}
   * @since 1.4
   */
  /** @export */ get overflow(): EnumValue { return this._overflow; }
  /** @export */ set overflow(value: EnumValue) {
    const old = this._overflow;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, TextBlock, TextBlock, 'overflow');
      this._overflow = value;
      this.invalidateMeasure();
      this.raiseChanged('overflow', old, value);
    }
  }

  /**
   * Gets or sets the {@link Brush} or string that describes the stroke (color) of the text that is drawn.
   *
   * The default value is `"black"`.
   * Any valid CSS string can specify a solid color, and the {@link Brush}
   * class can be used to specify a gradient or pattern.
   * More information about the syntax of CSS color strings is available at:
   * <a href="https://developer.mozilla.org/en-US/docs/Web/CSS/color">CSS colors (mozilla.org)</a>.
   */
  /** @export */ get stroke(): BrushLike { return this._stroke; }
  /** @export */ set stroke(value: BrushLike) {
    const old = this._stroke;
    if (old !== value) {
      if (value !== null) Brush.checkBrushString(value, 'TextBlock.stroke');
      if (value instanceof Brush) value.freeze();
      this._stroke = value;
      this.invalidateDraw();
      this.raiseChanged('stroke', old, value);
    }
  }

  /**
   * This read-only property returns the total number of lines in this TextBlock, including lines created
   * from embedded newlines (`\n`), {@link #wrap}ping, and {@link #maxLines}.
   *
   * This value may be meaningless before the TextBlock is measured.
   */
  /** @export */ get lineCount(): number { return this._lineCount; }

  /**
   * Gets or sets whether or not this TextBlock allows in-place editing of the {@link #text}
   * string by the user with the help of the {@link TextEditingTool}.
   * The default is false.
   *
   * See also {@link Part#textEditable}.
   */
  /** @export */ get editable(): boolean { return this._editable; }
  /** @export */ set editable(value: boolean) {
    const old = this._editable;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', TextBlock, 'editable');
      this._editable = value;
      this.raiseChanged('editable', old, value);
    }
  }

  /**
   * Gets or sets the HTMLInfo that this TextBlock uses as its text editor in the TextEditingTool.
   * If null, the TextBlock will use the default text editor of the {@link TextEditingTool}.
   * The default is null.
   * The value should be set to an instance of {@link HTMLInfo}.
   *
   * As of 2.0 setting this to an HTML Element is no longer supported.
   *
   * For example usage, see the <a href="../../samples/customTextEditingTool.html">Custom TextEditingTool Sample</a>.
   */
  /** @export */ get textEditor(): HTMLInfo | null { return this._textEditor; }
  /** @export */ set textEditor(value: HTMLInfo | null) {
    const old = this._textEditor;
    if (old !== value) {
      if (Debug && !(value instanceof HTMLInfo)) {
        Util.throwError('TextBlock.textEditor must be an HTMLInfo.');
      }
      this._textEditor = value;
      this.raiseChanged('textEditor', old, value);
    }
  }

  /**
   * Gets or sets the function to call if a text edit made with the {@link TextEditingTool} is invalid.
   * The default is null.
   */
  /** @export */ get errorFunction(): ((tool: TextEditingTool, oldString: string, newString: string) => void) | null { return this._errorFunction; }
  /** @export */ set errorFunction(value: ((tool: TextEditingTool, oldString: string, newString: string) => void) | null) {
    const old = this._errorFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', TextBlock, 'errorFunction');
      this._errorFunction = value;
      this.raiseChanged('errorFunction', old, value);
    }
  }

  /**
   * Gets or sets how frequently this text should be drawn within a "Graduated" {@link Panel},
   * in multiples of the {@link Panel#graduatedTickUnit}.
   * The default is 1.  Any new value must be a positive integer.
   * @since 1.7
   */
  /** @export */ get interval(): number { return this._interval; }
  /** @export */ set interval(value: number) {
    const old = this._interval;
    if (Debug) Util.checkRealNumber(value, TextBlock, 'interval');
    value = Math.floor(value);
    if (old !== value) {
      // On setting, negative, infinite, and NaN values must be ignored
      if (value >= 0) {
        this._interval = value;
        this.invalidateMeasure();
        const panel = this.panel;
        if (panel !== null) panel._graduatedIntervals = null;
        this.raiseChanged('interval', old, value);
      }
    }
  }

  /**
   * Gets or sets the fractional distance along the main shape of a "Graduated" {@link Panel} at which this text should start.
   * The default is 0; the value should range from 0 to 1.
   * @since 1.7
   */
  /** @export */ get graduatedStart(): number { return this._graduatedStart; }
  /** @export */ set graduatedStart(value: number) {
    const old = this._graduatedStart;
    if (Debug) Util.checkRealNumber(value, TextBlock, 'graduatedStart');
    if (old !== value) {
      if (value < 0) {
        value = 0;
      } else if (value > 1) {
        value = 1;
      }
      this._graduatedStart = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedStart', old, value);
    }
  }

  /**
   * Gets or sets the fractional distance along the main shape of a "Graduated" {@link Panel} at which this kind of tick text should end.
   * The default is 1; the value should range from 0 to 1.
   * @since 1.7
   */
  /** @export */ get graduatedEnd(): number { return this._graduatedEnd; }
  /** @export */ set graduatedEnd(value: number) {
    const old = this._graduatedEnd;
    if (Debug) Util.checkRealNumber(value, TextBlock, 'graduatedEnd');
    if (old !== value) {
      if (value < 0) {
        value = 0;
      } else if (value > 1) {
        value = 1;
      }
      this._graduatedEnd = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedEnd', old, value);
    }
  }

  /**
   * Gets or sets the function to convert from a value along a "Graduated" {@link Panel} to a string.
   * The default returns a string representing the value rounded to at most 2 decimals.
   *
   * The function takes a number argument, a value between {@link Panel#graduatedMin} and {@link Panel#graduatedMax}.
   * The function will return a string, the text that will appear at the value of the argument.
   * @since 1.7
   */
  /** @export */ get graduatedFunction(): ((val: number) => string) | null { return this._graduatedFunction; }
  /** @export */ set graduatedFunction(value: ((val: number) => string) | null) {
    const old = this._graduatedFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', TextBlock, 'graduatedFunction');
      this._graduatedFunction = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedFunction', old, value);
    }
  }

  /**
   * Gets or sets the function to determine which values along a "Graduated" {@link Panel} will be skipped.
   * The default is null and doesn't skip any text labels.
   *
   * The function takes a number argument, a value between {@link Panel#graduatedMin} and {@link Panel#graduatedMax}.
   * The function will return a boolean, whether the text label will be skipped at the value of the argument.
   * @since 2.0
   */
  /** @export */ get graduatedSkip(): ((val: number) => boolean) | null { return this._graduatedSkip; }
  /** @export */ set graduatedSkip(value: ((val: number) => boolean) | null) {
    const old = this._graduatedSkip;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', TextBlock, 'graduatedSkip');
      this._graduatedSkip = value;
      this.invalidateMeasure();
      this.raiseChanged('graduatedSkip', old, value);
    }
  }

  /**
   * @hidden @internal
   * Draw the TextBlock to a specified context.
   * @this {TextBlock}
   * @param {IContext} context A Canvas 2D Context for drawing.
   * @param {Diagram} diagram
   */
  public drawProtected(context: IContext, diagram: Diagram): void {
    if (this._stroke === null || this._text.length === 0 || this._font === null) return;
    const nb = this.naturalBounds;
    const ab = this.actualBounds;
    const maxw = nb.width;
    const maxh = nb.height;
    const lineHeight = this.getStringHeight();

    let align = context.textAlign = this._textAlign;
    // normalize values to left, right, center
    const isRTL = diagram._isRTL;
    if (align === 'start') { align = (isRTL ? 'right' : 'left');
    } else if (align === 'end') { align = (isRTL ? 'left' : 'right'); }
    // context.textBaseline = 'alphabetic'; // 'top' would make this cleaner but is not consistent across all browsers
    const isUnderline = this._isUnderline;
    const isStrikethrough = this._isStrikethrough;
    this.setBrush(context, this._stroke, true, false, nb, ab);
    if (isUnderline || isStrikethrough) {
      this.setBrush(context, this._stroke, false, false, nb, ab);
    }

    const x = 0.0;
    let y = 0.0;
    const width = maxw;

    let simple = false;
    const a = Point.allocAt(0, 0);
    this.allTransforms.transformPoint(a);
    const b = Point.allocAt(0, lineHeight);
    this.allTransforms.transformPoint(b);
    let distsq = a.distanceSquaredPoint(b);
    Point.free(a);
    Point.free(b);
    const sc = diagram.scale;
    distsq = distsq * sc * sc;
    if (distsq < 8) simple = true;
    // disable greeking unless its the context owned by the diagram
    if (diagram._ctx !== context) simple = false;
    // disable greeking if explicitly disabled:
    if (diagram.getRenderingHint('textGreeking') === false) simple = false;

    const spacingabove = this._spacingAbove;
    const spacingbelow = this._spacingBelow;

    switch (this.flip) {
      case GraphObject.None:
        break; // do nothing!
      case GraphObject.FlipHorizontal:
        context.translate(maxw, 0);
        context.scale(-1, 1);
        break;
      case GraphObject.FlipVertical:
        context.translate(0, maxh);
        context.scale(1, -1);
        break;
      case GraphObject.FlipBoth:
        context.translate(maxw, maxh);
        context.scale(-1, -1);
        break;
    }

    // We have arrSize and arrText
    // use _lineCount instead of _numLines
    // because this may be larger than the allowed number of lines
    const l = this._lineCount;
    const oneLineHeight = spacingabove + lineHeight + spacingbelow;
    const allLinesHeight = oneLineHeight * l;
    if (maxh > allLinesHeight) {
      const valign = this._verticalAlignment;
      y = (valign.y * maxh) - (valign.y * allLinesHeight) + valign.offsetY;
    }

    if (l === 1) {
      let sw = this._maxLineWidth;
      if (sw > width) sw = width;
      y += spacingabove;
      this.drawPart(this._firstLine, context, x, y, width, lineHeight, sw, simple, align, isUnderline, isStrikethrough);
    } else if (this._arrSize !== null && this._arrText !== null) {
      for (let i = 0; i < l; i++) {
        let sw = this._arrSize[i];
        if (sw > width) sw = width;
        y += spacingabove;
        this.drawPart(this._arrText[i], context, x, y, width, lineHeight, sw, simple, align, isUnderline, isStrikethrough);
        y += lineHeight + spacingbelow;
      }
    }

    switch (this.flip) {
      case GraphObject.None:
        break; // do nothing!
      case GraphObject.FlipHorizontal:
        context.scale(-1, 1);
        context.translate(-maxw, 0);
        break;
      case GraphObject.FlipVertical:
        context.scale(1, -1);
        context.translate(0, -maxh);
        break;
      case GraphObject.FlipBoth:
        context.scale(-1, -1);
        context.translate(-maxw, -maxh);
        break;
    }
  }

  /**
   * @hidden @internal
   * Draw one line of a string to a context
   * @param {string} line
   * @param {IContext} context
   * @param {number} rx
   * @param {number} ry
   * @param {number} rw
   * @param {number} rh line height
   * @param {number} stringwidth
   * @param {boolean} simple whether or not to draw a line (greeking) instead of text at small sizes.
   * @param {string} align left, right, or center
   * @param {boolean} isUnderline
   * @param {boolean} isStrikethrough
   */
  private drawPart(line: string, context: IContext, rx: number, ry: number, rw: number, rh: number, stringwidth: number,
                   simple: boolean, align: string, isUnderline: boolean, isStrikethrough: boolean): void {
    let offset = 0.0;

    if (simple) { // Greeking: just draw a line (rects are faster than paths so use a rect)
      // offsetting a line is different than offsetting text!
      if (align === 'left') {
        offset = 0.0;
      } else if (align === 'right') {
        offset = rw - stringwidth;
      } else if (align === 'center') {
        offset = (rw - stringwidth) / 2;
      }
      context.fillRect(rx + offset, ry + (rh * .25), stringwidth, 1);
      return;
    }

    // if the textBaseline was 'top' the y coordinate of fillText would have to be something like:
    // ry - (rh * .15));
    if (align === 'left') {
      offset = 0.0;
    } else if (align === 'right') {
      offset = rw;
    } else if (align === 'center') {
      offset = rw / 2;
    }

    const baselineY = (TextBlock.baselineComputation !== null) ? TextBlock.baselineComputation(this, rh) : (rh * .75);
    context.fillText(line, rx + offset, ry + baselineY);

    let lh = ((rh / 20) | 0);
    if (lh === 0) lh = 1;
    if (align === 'right') {
      offset -= stringwidth;
    } else if (align === 'center') {
      offset -= (stringwidth / 2);
    }
    if (isUnderline) {
      const underlineY = (TextBlock.underlineComputation !== null) ? TextBlock.underlineComputation(this, rh) : (rh * .80);
      context.beginPath();
      context.lineWidth = lh;
      context.moveTo(rx + offset, ry + underlineY);
      context.lineTo(rx + offset + stringwidth, ry + underlineY);
      context.stroke();
    }
    if (isStrikethrough) {
      context.beginPath();
      context.lineWidth = lh;
      let y = (ry + rh - (rh / 2.2)) | 0;
      if (lh % 2 !== 0) y += 0.5; // If the line height is odd, draw on halfpixel
      context.moveTo(rx + offset, y);
      context.lineTo(rx + offset + stringwidth, y);
      context.stroke();
    }
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   */
  protected resetMeasurements(): void {
    this._numLines = 0;
    this._maxLineWidth = 0;
    this._fontHeight = null;
    this._arrSize = null;
    this._arrText = null;
    this._firstLine = '';
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {number} width
   * @param {number} height
   * @param {number} minw
   * @param {number} minh
   */
  protected measureProtected(width: number, height: number, minw: number, minh: number): void {
    // used in text editing tool to determine potential sizes for the text area:
    this._lastAvailableWidth = width;
    const font = this._font;
    if (TextBlock._sharedTempCtx !== null) {
      if (TextBlock._sharedTempFont !== font) {
        TextBlock._sharedTempCtx.font = font;
        TextBlock._sharedTempFont = font;
      }
    }

    this.resetMeasurements();
    let w = 0.0;
    let h = 0.0;
    if (!isNaN(this.desiredSize.width)) {
      w = this.desiredSize.width;
    } else {
      w = this.computeWidth();
      // comparing local width to panel width so divide panel width by scale
      w = Math.min(w, width / this.scale);
      w = Math.max(8, w);
    }
    // measuring text needs to know this sooner rather than later
    // since the width affects the height
    if (this.panel !== null) {
      w = Math.min(w, this.panel.maxSize.width);
    }
    // This computes lines, an object that holds information about wrapping
    h = this.computeHeight(w);
    // If we wanted the max text size, it would be Size(lines._maxLineWidth, h)
    if (!isNaN(this.desiredSize.height)) {
      h = this.desiredSize.height;
    } else {
      // comparing local width to panel height so divide panel height by scale
      h = Math.min(h, height / this.scale);
    }

    this.multiLineEllipsis(width, h);

    if (this.wrap === TextBlock.WrapFit || isNaN(this.desiredSize.width)) {
      w = isNaN(width) ? this._maxLineWidth : Math.min(width, this._maxLineWidth);
      if (isNaN(this.desiredSize.width)) w = Math.max(8, w);
    }

    w = Math.max(minw, w);
    h = Math.max(minh, h);

    const nb = this._naturalBounds;
    nb.assignWidthHeight(w, h);
    this._measureRect(0, 0, w, h);
  } // end measure

  /**
   * @hidden @internal
   * Arranges the TextBlock onto its panel.
   * @this {TextBlock}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    this.commonArrange(fx, fy, fw, fh);
  }


  /**
   * @hidden @internal
   * add to arrText, arrSize
   * @this {TextBlock}
   */
  private addLine(text: string, size: number): void {
    if (this._firstLine === '') {
      this._firstLine = text;
      this._maxLineWidth = size;
    } else {
      if (this._arrText === null || this._arrSize === null) {
        this._arrText = [];
        this._arrSize = [];
        this._arrText.push(this._firstLine);
        this._arrSize.push(this._maxLineWidth);
      }
      this._arrText.push(text);
      this._arrSize.push(size);
    }
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {string} str discrete line of text to measure, possibly splitting it into multiple lines.
   * str has no embedded newlines.
   * @param {number} ww Wrapping Width, the maximum allowed width
   * @param {Size | null} sz return value
   */
  private measureString(str: string, ww: number, sz: Size | null): void {
    // ??? should work but untested. When would this predicate ever be true?
    // if (this.getInvalidMeasure() === false) return this.naturalBounds;
    str = str.trim();
    let maxLineWidth = 0.0;
    let tempWidth = 0.0;
    let lineWidth = 0.0;
    const font = this._font;
    const spacing = this._spacingAbove + this._spacingBelow;
    const fheight = Math.max(0, this.getStringHeight() + spacing);
    const ellipsisSize = (this._overflow === TextBlock.OverflowEllipsis) ? this.getEllipsisSize(font) : 0;

    // can't measure any more string
    if (this._lineCount >= this._maxLines) {
      if (sz !== null) sz.assignTo(0, fheight);
      return;
    }
    let line = str;
    if (this._wrap === TextBlock.None) {
      this._numLines = 1;
      tempWidth = this.getStringWidth(str);
      if (ellipsisSize === 0 || tempWidth <= ww) {
        this._maxLineWidth = Math.max(this._maxLineWidth, tempWidth);
        this.addLine(str, this._maxLineWidth);
        if (sz !== null) sz.assignTo(tempWidth, fheight);
        return;
      }

      // Else take off some of the line to make way for an ellipsis
      let part = this.getNextWord(line);
      line = line.substr(part.length);
      let nextword = this.getNextWord(line);
      tempWidth = this.getStringWidth(part + nextword);
      lineWidth = 0;

      while (nextword.length > 0 && tempWidth <= ww) {
        part += nextword;
        line = line.substr(nextword.length);
        nextword = this.getNextWord(line);
        lineWidth = tempWidth;
        tempWidth = this.getStringWidth((part + nextword).trim());
      }
      part += nextword.trim();

      // The entire tempWidth doesnt fit, time to see how much of it fits in the space less an ellipsis
      ww = Math.max(1, ww - ellipsisSize);

      // "part" is the whole of what can fit and, see how much of "nextword + ellipsis" fits
      while (this.getStringWidth(part) > ww && part.length > 1) {
        part = part.substr(0, part.length - 1);
      }

      part += TextBlock._ellipsisCharacter;
      lineWidth = this.getStringWidth(part);
      this._maxLineWidth = lineWidth;
      this.addLine(part, lineWidth);
      if (sz !== null) sz.assignTo(lineWidth, fheight);
      return;
    }
    // Wrap the line of text instead
    let numlines = 0;
    if (line.length === 0) {
      numlines = 1;
      this.addLine(line, 0);
    }
    while (line.length > 0) { // num of characters in line
      let part = this.getNextWord(line);

      line = line.substr(part.length);
      // if first part doesn't fit, just display as much as we can
      while (this.getStringWidth(part) > ww) {
        let l = 1;
        // get the biggest prefix of the part that will fit

        tempWidth = this.getStringWidth(part.substr(0, l));
        lineWidth = 0;
        while (tempWidth <= ww) {
          l++;
          lineWidth = tempWidth;
          tempWidth = this.getStringWidth(part.substr(0, l));
        }
        // push the valid line width determined by when the while loop ends
        if (numlines > 1 && this._arrSize !== null) {
          if (l === 1) {
            this._arrSize[this._lineCount + numlines] = tempWidth;
            maxLineWidth = Math.max(maxLineWidth, tempWidth);
          } else {
            this._arrSize[this._lineCount + numlines] = lineWidth;
            maxLineWidth = Math.max(maxLineWidth, lineWidth);
          }
        }


        l--;  // remember the number of characters that fit
        if (l < 1) l = 1;  // always take at least one character
        const prefix = part.substr(0, l);
        if (numlines > 1 && this._arrText !== null) {
          // prefix is now the part of this long word that fits
          this._arrText[this._lineCount + numlines] = prefix;
        }
        numlines++;
        part = part.substr(l); // now continue with the rest of this very long word
        // we want at least one additional line to exist so we can determine if ellipses are needed
        // we may also want the additional data later to expose in text metrics? If so comment out following "if"
        if (this._lineCount + numlines > this._maxLines) break;
      }
      // now PART fits in the WW width, see if we can add any more words that fit

      let nextword = this.getNextWord(line);

      tempWidth = this.getStringWidth(part + nextword);
      lineWidth = 0;

      while (nextword.length > 0 && tempWidth <= ww) {
        part += nextword;
        line = line.substr(nextword.length);
        nextword = this.getNextWord(line);
        lineWidth = tempWidth;
        tempWidth = this.getStringWidth((part + nextword).trim());
      }

      // Trim the spaces before and after part. lineWidth is already correct.
      part = part.trim();
      if (part === '') continue;

      // If a line ends with an invisible soft hypen after measuring, replace it with a real hyphen.
      if (part[part.length - 1] === '\u00AD') part = part.substring(0, part.length - 1) + '‐';

      // push the valid line width determined by when the while loop ends
      let width = 0;
      if (nextword.length === 0) {
        width = tempWidth;
        maxLineWidth = Math.max(maxLineWidth, tempWidth);
      } else {
        lineWidth = this.getStringWidth(part);
        width = lineWidth;
        maxLineWidth = Math.max(maxLineWidth, lineWidth);
      }
      this.addLine(part, width);
      numlines++;

      // we want at least one additional line to exist so we can determine if ellipses are needed
      // we may also want the additional data later to expose in text metrics? If so comment out following "if"
      if (this._lineCount + numlines > this._maxLines) break;
    }
    this._numLines = Math.min(this._maxLines, numlines);
    this._maxLineWidth = Math.max(this._maxLineWidth, maxLineWidth);
    if (sz !== null) sz.assignTo(this._maxLineWidth, fheight * this._numLines);
  }

  /**
   * @hidden @internal
   * Modifies "lines" object when multiline and ellipsis are enabled
   * @this {TextBlock}
   * @param {number} w
   * @param {number} h
   */
  private multiLineEllipsis(w: number, h: number): void {
    // If there's no wrap and its height restricted, put an ellipsis on the last real line
    if (this._maxLineWidth === 0 || this._arrSize === null || this._arrText === null /* formerly: lines._arrText.length === 1 */ ||
      this._overflow !== TextBlock.OverflowEllipsis) return;

    const font = this._font;
    const ellipsisSize = (this._overflow === TextBlock.OverflowEllipsis) ? this.getEllipsisSize(font) : 0;
    const spacing = this._spacingAbove + this._spacingBelow;
    const fheight = Math.max(0, this.getStringHeight() + spacing);
    // lastline is an index in the this.Array<_arrSize> array
    const lastline = Math.min(this.maxLines - 1, Math.max(Math.floor((h / fheight) + 0.01) - 1, 0));
    if (lastline + 1 >= this._arrText.length) return; // cant put ellipsis on the last line of wrapped text
    let part = this._arrText[lastline];

    // wrap width, see how much of it fits in the space less an ellipsis
    const ww = Math.max(1, w - ellipsisSize);

    // "part" is the whole of what can fit and, see how much of "nextword + ellipsis" fits
    while (this.getStringWidth(part) > ww && part.length > 1) {
      part = part.substr(0, part.length - 1);
    }

    part += TextBlock._ellipsisCharacter;
    const lineWidth = this.getStringWidth(part);
    this._arrText[lastline] = part;
    this._arrText = this._arrText.slice(0, lastline + 1);
    this._arrSize[lastline] = lineWidth;
    this._arrSize = this._arrSize.slice(0, lastline + 1);
    this._numLines = this._arrText.length;
    this._maxLineWidth = Math.max(this._maxLineWidth, lineWidth);
    this._lineCount = this._numLines; // ellipses reduce the line count since there's no more cropping of text
    if (this._lineCount === 1) {
      this._firstLine = this._arrText[0];
    }
  }

  // for manual wrapping
  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {string} str
   * @return {string}
   */
  private getNextWord(str: string): string {
    if (this._wrap === TextBlock.WrapBreakAll) return str.substr(0, 1);
    const len = str.length;
    let i = 0;
    const regex = TextBlock._whitespaceRegex;
    // scan characters until we get to whitespace again
    while (i < len && !regex.test(str.charAt(i))) i++;
    // include trailing whitespace
    while (i < len && regex.test(str.charAt(i))) i++;
    if (i >= len) {
      return str;
    } else {
      return str.substr(0, i);
    }
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {string} str
   * @return {number}
   */
  private getStringWidth(str: string): number {
    if (TextBlock._sharedTempCtx === null) {
      return str.length * 8; // ??? very arbitrary string width in DOM-less environment!
    }
    return TextBlock._sharedTempCtx.measureText(str).width;
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @return {number}
   */
  private getStringHeight(): number {
    if (this._fontHeight !== null) return this._fontHeight; // ?? is this a worthwhile optimization?
    const font = this._font;
    let m = 0.0;
    if (TextBlock._sharedTempCtx === null) {
      m = 16; // ??? very arbitrary font height when DOM-less
      this._fontHeight = m;
      return m;
    }

    if (TextBlock._fontHeightMap[font] !== undefined && TextBlock._fontHeightMapCount < 5000) {
      m = TextBlock._fontHeightMap[font];
    } else {
      const metrics = TextBlock._sharedTempCtx.measureText('M');
      // ?? SUPER ARBITRARY
      m = metrics.width * 1.3; // 30% more to get the titles of i and j
      TextBlock._fontHeightMap[font] = m;
      TextBlock._fontHeightMapCount++;
    }
    this._fontHeight = m;
    return m;
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {string} font
   * @return {number}
   */
  private getEllipsisSize(font: string): number {
    if (TextBlock._sharedTempCtx === null) return 6; // ??? arbitrary DOM-less value
    let m = 0.0;
    if (TextBlock._fontEllipsisWidth[font] !== undefined && TextBlock._fontEllipsisWidthCount < 5000) {
      m = TextBlock._fontEllipsisWidth[font];
    } else {
      const metrics = TextBlock._sharedTempCtx.measureText(TextBlock._ellipsisCharacter);
      m = metrics.width;
      TextBlock._fontEllipsisWidth[font] = m;
      TextBlock._fontEllipsisWidthCount++;
    }
    return m;
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @param {string} str MUST be normalized to only contain \n, and not \r or \r\n
   * @param {number} start
   * @return {number}
   */
  private findNextLineBreak(str: string, start: number): number {
    // must do this to str before calling this method:
    // str.replace(/\r\n/g, "\n").replace(/\r/g, "\n")
    return str.indexOf('\n', start);
  }

  /**
   * @hidden @internal
   * Computes the total height of the TextBlock (the height of all lines), and returns the total height.
   * @this {TextBlock}
   * @param {number} maxw
   * @return {number}
   */
  private computeHeight(maxw: number): number {
    // normalize: \r and \r\n become \n
    let str = this._text.replace(/\r\n/g, '\n').replace(/\r/g, '\n');
    const spacing = this._spacingAbove + this._spacingBelow;
    const fheight = Math.max(0, this.getStringHeight() + spacing);

    if (str.length === 0) {
      this._maxLineWidth = 0;
      this._lineCount = 1;
      return fheight;
    }

    if (!this.isMultiline) {
      const newline = this.findNextLineBreak(str, 0);
      if (newline >= 0) {
        str = str.substr(0, newline);
      }
    }

    let linePos = 0.0;
    this._lineCount = 0;
    let startingIndex = 0;
    let endingIndex = -1;
    let done = false;
    while (!done) {
      endingIndex = this.findNextLineBreak(str, startingIndex);

      // if there is no newline, then we use the rest of the string
      if (endingIndex === -1) {
        endingIndex = str.length;
        done = true;
      }

      if (startingIndex <= endingIndex) {  // we have a logical line
        const line = str.substr(startingIndex, endingIndex - startingIndex);
        if (this._wrap !== TextBlock.None) {
          this._numLines = 0;
          const sz = Size.alloc();
          this.measureString(line, maxw, sz);
          linePos += sz.height;
          Size.free(sz);
          this._lineCount += this._numLines;
        } else {
          // measureString must still be called to compute the lines data (arrSize, arrText, maxLineWidth etc)
          this.measureString(line, maxw, null);
          linePos += fheight;
          this._lineCount++;
        }
        if (this._lineCount === this._maxLines) {
          done = true;
        }
      }
      startingIndex = endingIndex + 1; // move to the next line
    }
    this._maxPosition = linePos;
    return linePos;
  }

  /**
   * @hidden @internal
   * @this {TextBlock}
   * @return {number}
   */
  private computeWidth(): number {
    // normalize: \r and \r\n become \n
    let str = this._text.replace(/\r\n/g, '\n').replace(/\r/g, '\n');
    if (str.length === 0) {
      return 0;
    }

    if (this.isMultiline) {
      let maxWidth = 0.0;
      let startingIndex = 0;
      let done = false;

      while (!done) {
        let endingIndex = this.findNextLineBreak(str, startingIndex);

        // if there is no newline, then we use the rest of the string
        if (endingIndex === -1) {
          endingIndex = str.length;
          done = true;
        }

        const line = str.substr(startingIndex, endingIndex - startingIndex).trim();
        maxWidth = Math.max(maxWidth, this.getStringWidth(line));

        // move to the next line
        startingIndex = endingIndex + 1;
      }
      return maxWidth;
    } else {
      const newline = this.findNextLineBreak(str, 0);
      if (newline >= 0) {
        str = str.substr(0, newline);
      }
      return this.getStringWidth(str);
    }
  }

  //
  //  * hidden internal
  //  * Gets the row with the largest number of columns for sizing a textArea
  //  * this {TextBlock}
  //  * return {number}
  //
  // private getMaxCols(): number {
  //   var lines = this._lines;
  //   var max = 1;
  //   var text = lines._arrText;
  //   if (text === undefined) return 1;
  //   var l = text.length;
  //   for (var i = 0; i < l; i++) {
  //     max = Math.max(text[i].length, max);
  //   }
  //   return max;
  // }

  /**
   * Gets or sets the predicate that determines whether or not a user-edited string of text is valid.
   * If this is non-null, the predicate is called in addition to any {@link TextEditingTool#textValidation} predicate.
   * See {@link TextEditingTool#isValidText} for more details.
   *
   * ```js
   * function(textBlock, oldString, newString)
   * ```
   *
   * The default predicate is null, which is equivalent to simply returning true.
   *
   * The function, if supplied, must not have any side-effects, and must return true or false.
   * @see {@link TextEditingTool#textValidation}
   */
  /** @export */ get textValidation(): ((thisTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null { return this._textValidation; }
  /** @export */ set textValidation(value: ((thisTextBlock: TextBlock, oldString: string, newString: string) => boolean) | null) {
    const old = this._textValidation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', TextBlock, 'textValidation');
      this._textValidation = value;
      this.raiseChanged('textValidation', old, value);
    }
  }

  /**
   * Gets or sets the function that is called after the TextBlock's text has been edited by the {@link TextEditingTool}.
   *   - The first argument is a reference to this TextBlock.
   *   - The second argument is the previous text, before editing.
   *   - The third argument is the current text, which is also TextBlock.text.
   *
   * ```js
   * function(textBlock, previousText, currentText)
   * ```
   *
   * The default value is null -- no function is called.
   * @since 1.7
   */
  /** @export */ get textEdited(): ((thisTextBlock: TextBlock, oldString: string, newString: string) => void) | null { return this._textEdited; }
  /** @export */ set textEdited(value: ((thisTextBlock: TextBlock, oldString: string, newString: string) => void) | null) {
    const old = this._textEdited;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', TextBlock, 'textEdited');
      this._textEdited = value;
      this.raiseChanged('textEdited', old, value);
    }
  }

  //  * Gets or sets additional spacing above each line of text.
  //  * The default is zero. The value may be negative.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get spacingAbove(): number { return this._spacingAbove; }
  /** @export */ set spacingAbove(value: number) {
    const old = this._spacingAbove;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', TextBlock, 'spacingAbove');
      this._spacingAbove = value;
      this.raiseChanged('spacingAbove', old, value);
    }
  }

  //    * Gets or sets additional spacing below each line of text.
  //  * The default is zero. The value may be negative.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get spacingBelow(): number { return this._spacingBelow; }
  /** @export */ set spacingBelow(value: number) {
    const old = this._spacingBelow;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', TextBlock, 'spacingBelow');
      this._spacingBelow = value;
      this.raiseChanged('spacingBelow', old, value);
    }
  }

  /**
   * Gets or sets the maximum number of lines that this TextBlock can display.
   * Value must be a greater than zero whole number or `Infinity`.
   * The default is `Infinity`.
   *
   * Modifying this value may modify the computed height of the TextBlock.
   * If maxLines is set, the value of {@link #lineCount} will never be larger than maxLines.
   * @see {@link #isMultiline}
   * @since 1.5
   */
  /** @export */ get maxLines(): number { return this._maxLines; }
  /** @export */ set maxLines(value: number) {
    const old = this._maxLines;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', TextBlock, 'maxLines');
      value = Math.floor(value);
      if (value <= 0) Util.throwRangeError(value, '> 0', TextBlock, 'maxLines');
      this._maxLines = value;
      this.raiseChanged('maxLines', old, value);
      this.invalidateMeasure();
    }
  }

  /**
   * @hidden @internal
   * Used only by graduated panels to record a bunch of properties, possibly to transfer to another text block.
   * ???
   */
  public getMetrics(): [number, number, number | null, number, Array<number> | null, Array<string> | null] {
    return [
      this._maxLineWidth, // number;
      this._lastAvailableWidth, // number
      this._fontHeight, // number | null
      this._numLines, // number
      this._arrSize, // Array<number> | null
      this._arrText // Array<string> | null
    ];
  }

  /**
   * @hidden @internal
   * Used only by graduated panels to record a bunch of properties, possibly to transfer to another text block.
   * ???
   */
  public setMetrics(lines: [number, number, number | null, number, Array<number> | null, Array<string> | null] | undefined): void {
    if (lines === undefined) return;
    this._maxLineWidth = lines[0];
    this._lastAvailableWidth = lines[1];
    this._fontHeight = lines[2];
    this._numLines = lines[3];
    this._arrSize = lines[4];
    this._arrText = lines[5];
  }

  /**
   * @hidden @internal
   * For compatibility this returns an object like TextMetrics, a pre 2.1 class
   */
  /** @export */ get metrics(): PropertyCollection {
    return {
      'arrSize': (this._arrSize !== null) ? this._arrSize : [this._maxLineWidth],
      'arrText': (this._arrText !== null) ? this._arrText : [this._firstLine],
      'maxLineWidth': this._maxLineWidth,
      'fontHeight': this._fontHeight
    };
  }

  /**
   * Gets or sets the an array of possible choices for a custom {@link TextEditingTool}.
   * The value must be an array of strings.
   *
   * The default value is null.
   * For example usage, see the <a href="../../samples/customTextEditingTool.html">Custom TextEditingTool Sample</a>.
   * @since 1.7
   */
  /** @export */ get choices(): Array<string> | null { return this._choices; }
  /** @export */ set choices(value: Array<string> | null) {
    const old = this._choices;
    if (old !== value) {
      if (Debug && value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', TextBlock, 'choices:value');
      this._choices = value;
      this.raiseChanged('choices', old, value);
    }
  }

  // Whitespace match for space, \u200B (zero-width space), and \u00AD (soft hyphen)
  /** @hidden @internal */
  public static _whitespaceRegex: RegExp = new RegExp('[ \u200B\u00AD]');

  // A record of the font height of each font that has been used so far
  /** @hidden @internal @type {Object.<string, number>} */
  public static _fontHeightMap: Record<string, number> = new PropertyCollection();
  /** @hidden @internal */
  public static _fontHeightMapCount = 0;
  /** @hidden @internal */
  public static _fontEllipsisWidth: Record<string, number> = new PropertyCollection();
  /** @hidden @internal */
  public static _fontEllipsisWidthCount = 0;
  /** @hidden @internal */
  public static _ellipsisCharacter = '...';
  /** @hidden @internal */
  public static _sharedTempFont = '';

  // This static canvas is used solely for measuring text
  /** @hidden @internal */
  private static _sharedTempCtx: CanvasSurfaceContext | null = null;
  /** @hidden @internal */
  private static _hasInit: boolean = false;

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    TextBlock._sharedTempCtx = Diagram.isUsingDOM() ? new CanvasSurface(null).context : null;
  }
} // #end TextBlock
TextBlock['className'] = 'TextBlock';

TextBlock['getEllipsis'] = TextBlock.getEllipsis;
TextBlock['setEllipsis'] = TextBlock.setEllipsis;
TextBlock['getBaseline'] = TextBlock.getBaseline;
TextBlock['setBaseline'] = TextBlock.setBaseline;
TextBlock['getUnderline'] = TextBlock.getUnderline;
TextBlock['setUnderline'] = TextBlock.setUnderline;
TextBlock['isValidFont'] = TextBlock.isValidFont;
TextBlock['None'] = TextBlock.None;
TextBlock['WrapFit'] = TextBlock.WrapFit;
TextBlock['WrapDesiredSize'] = TextBlock.WrapDesiredSize;
TextBlock['WrapBreakAll'] = TextBlock.WrapBreakAll;
TextBlock['OverflowClip'] = TextBlock.OverflowClip;
TextBlock['OverflowEllipsis'] = TextBlock.OverflowEllipsis;

/**
 * @hidden @internal
 * @unrestricted
 */
export class TextBlockMetrics {
  /** @hidden @internal */ public static className = 'TextBlockMetrics';
  public _arrSize: Array<number>;
  public _arrText: Array<string>;

  constructor() {
    this._arrSize = [];
    this._arrText = [];
  }

  /** @hidden @internal */
  public reset(): void {
    this._arrSize = [];
    this._arrText = [];
  }

  /**
   * @hidden @internal
   * @param {TextBlockMetrics} target
   */
  public copyFrom(target: TextBlockMetrics): void {
    this._arrSize = Util.copyArray(target._arrSize);
    this._arrText = Util.copyArray(target._arrText);
  }
} // #end TextBlockMetrics
TextBlockMetrics['className'] = 'TextBlockMetrics';

// import { Brush, BrushLike } from './brush';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { HTMLInfo } from './htmlInfo';
// import { Margin } from './margin';
// import { GraduatedMark } from './panelLayoutGraduated';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Spot } from './spot';
// import { TextEditingTool } from './textEditingTool';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Picture
*/

// import { Set } from './collections';
// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { PropertyCollection } from './propertyCollection';
// import { Util } from './util';

/**
 * A Picture is a {@link GraphObject} that shows an image, video-frame, or Canvas element.
 *
 * You can specify what to show by either setting the {@link #source} URL property
 * to a URL string or the {@link #element} property to an HTMLImageElement,
 * HTMLCanvasElement, or HTMLVideoElement.
 *
 * If a {@link #source} URL is set, the Picture will automatically create a corresponding
 * HTMLImageElement and retain a reference to it in memory. If multiple Pictures specify
 * the same {@link #source} URL then they will all refer to the same HTMLImageElement.
 *
 * Some created Pictures:
 * ```js
 * var $ = go.GraphObject.make;  // for conciseness in defining GraphObjects
 *
 * // A shape with the figure set to "example.png". It will show a gray area until the image is loaded:
 * $(go.Picture, { source: "example.png", background: "gray", width: 50, height: 50 })
 * // Alternatively:
 * $(go.Picture, "example.png", { background: "gray", width: 50, height: 50 })
 * ```
 *
 * If an element is not completely loaded during Diagram initialization, a redraw may occur,
 * and if an image's size is not known before loading, the containing Part of this Picture
 * may be resized, causing side effects such as layouts.
 * This can be avoided by knowing the size of the image beforehand,
 * and setting the Picture's {@link GraphObject#desiredSize}.
 *
 * With some images (notably sprite sheets) only a portion of the image is expected to be drawn.
 * The {@link #sourceRect} property allows the programmer to specify a rectangular area of
 * the source image that the Picture should display.
 *
 * The {@link #imageStretch} property allows an image to be resized inside of its bounding box.
 * This property does not change the size of the Picture element, it only resizes or re-scales
 * the image to fit (or not) in its bounds.
 * <p class="boxread">
 * For examples of sizing and {@link #imageStretch},
 * see the <a href="../../intro/pictures.html">Introduction page on Pictures</a>.
 *
 * The {@link #errorFunction} property allows one to set a function to call when a source fails to load.
 * This is useful in instances where images cannot be guaranteed to work, such as with
 * user specified input. The error function can set the {@link #source} to a known good value,
 * but care should be taken to avoid error infinite loops when doing so.
 * @extends GraphObject
 * @unrestricted
 */
export class Picture extends GraphObject {
  /** @hidden @internal */ public static className = 'Picture';
  /** @hidden @internal */ private _element: HTMLImageElement | HTMLVideoElement | HTMLCanvasElement | null;
  /** @hidden @internal */ private _source: string;
  /** @hidden @internal */ private _sourceRect: Rect;
  /** @hidden @internal */ private _imageStretch: EnumValue;
  /** @hidden @internal */ private _errorFunction: ((a: Picture, b: Event) => void) | null;
  /** @hidden @internal */ private _successFunction: ((a: Picture, b: Event) => void) | null;
  /** @hidden @internal */ private _imageAlignment: Spot;
  /** @hidden @internal */ private _flip: EnumValue;
  /** @hidden @internal */ private _tainted: boolean | null;
  /** @hidden @internal */ private _isCanvas: boolean;
  /** @hidden @internal */ private _firstError: boolean;
  /** @hidden @internal */ private _sourceCrossOrigin: ((a: Picture) => string) | null;
  /** @hidden @internal */ public _loaded: boolean; // public for Diagram internal use

  /**
   * The constructor creates a picture that shows nothing until the
   * {@link #source} or {@link #element} is specified. It is also common to specify the
   * {@link GraphObject#desiredSize} to make sure that the picture's size is known
   * before the image is loaded asynchronously, so that layouts do not need to be recomputed.
   */
  constructor() {
    super();
    this._element = null;
    this._source = '';
    this._sourceRect = Rect.NoRect;
    this._imageStretch = GraphObject.Fill;
    this._errorFunction = null;
    this._successFunction = null;
    this._imageAlignment = Spot.Center;
    this._flip = GraphObject.None;

    this._tainted = null;
    this._isCanvas = false;

    this._firstError = true;

    this._loaded = false;

    this._sourceCrossOrigin = null;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Picture.
   * @this {Picture}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy.element = this._element;  // shared
    copy._source = this._source;

    copy._sourceRect = this._sourceRect.copyFrozen();
    copy._imageStretch = this._imageStretch;
    copy._flip = this._flip;
    copy._errorFunction = this._errorFunction;
    copy._successFunction = this._successFunction;
    copy._imageAlignment = this._imageAlignment.copyFrozen();

    // _tainted and _isCanvas are transient

    copy._firstError = this._firstError;

    copy._sourceCrossOrigin = this._sourceCrossOrigin;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Picture}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg === GraphObject.None || arg === GraphObject.Uniform || arg === GraphObject.UniformToFill) {
      this.imageStretch = arg;
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * @hidden @internal
   * @this {Picture}
   * @return {string}
   */
  public toString(): string {
    return 'Picture(' + this.source + ')#' + Set.hashId(this);
  }

  //    * This static function clears the cache of loaded images, or removes the cached image for a particular URL.
  //  *
  //  * When an image is loaded by GoJS when specifying a Picture.source URL,
  //  * GoJS creates an Image element using that source and caches it.
  //  * That way, when many Picture elements refer to the same source,
  //  * they can refer to the same single object in memory.
  //  * @since 1.4
   /**
    * @hidden
    * Undocumented
    * @param {string=} url optional, clear only a specified URL from the image cache. If relative urls are used as Picture sources, this must be in the same format.
    */
  public static clearCache(url?: string): void {
    if (url === undefined) url = '';
    Util.checkPrimitive(url, 'string', Picture, 'clearCache:url');
    if (url !== '') {
      if (Picture._imageSourceMap[url]) {
        delete Picture._imageSourceMap[url];
        Picture._imageSourceMapCount--;
      }
    } else {
      Picture._imageSourceMap = new PropertyCollection();
      Picture._imageSourceMapCount = 0;
    }
  }

  /**
   * @hidden @internal
   * @this {Picture}
   */
  public checkLoadStatus(): void {
    const element = this._element;
    if (element === null) return;
    if (element._hasError instanceof Event) {
      if (this._errorFunction !== null) this._errorFunction(this, element._hasError);
    }

    if (element._isLoaded === true) {
      if (this._successFunction !== null) {
        this._successFunction(this, element._savedSuccessEvent);
      }
      if (this.diagram !== null) this.diagram._picturesToInvalidate.add(this);
    }
  }

  /**
   * Gets or sets the Picture's HTML element that provides some kind of visual image.
   *
   * The default value is null.
   * This can be a HTMLImageElement, HTMLVideoElement, or HTMLCanvasElement.
   * If an image, this element must have its source (src) attribute defined.
   * Setting this does not set the {@link Picture#source} attribute and that attribute may be unknowable.
   */
  /** @export */ get element(): HTMLImageElement | HTMLVideoElement | HTMLCanvasElement | null { return this._element; }
  /** @export */ set element(value: HTMLImageElement | HTMLVideoElement | HTMLCanvasElement | null) {
    const old = this._element;
    if (old !== value) {
      if (value !== null && !(value instanceof HTMLImageElement || value instanceof HTMLVideoElement || value instanceof HTMLCanvasElement)) {
        Util.throwError('Picture.element must be an instance of Image, Canvas, or Video, not: ' + value);
      }
      this._isCanvas = (value instanceof HTMLCanvasElement);
      this._element = value;
      if (value !== null) {
        if (value instanceof HTMLCanvasElement || value.complete === true) {
          if (value._hasError instanceof Event) {
            if (this._errorFunction !== null) this._errorFunction(this, value._hasError);
          }

          if (value._isLoaded === true && this._successFunction !== null) {
            this._successFunction(this, value._savedSuccessEvent);
          }

          value._isLoaded = true; // property we are attaching to images
          if (!this.desiredSize.isReal()) {
            this.setInvalidMeasure(false);
            this.invalidateMeasure();
          }
        } else {
          // the image is not yet loaded/error'd
          const img = value as HTMLImageElement;
          const self = this;
          if (!img._goJSAttachedEvents) {
            img.addEventListener('load', function (e) {
              Picture.imageLoaded(img, e);
              // In case the first use is on a template, which may also need to reload
              if (!self.desiredSize.isReal()) {
                self.setInvalidMeasure(false);
                self.invalidateMeasure();
              }
            });
            img.addEventListener('error', function (e) {
              Picture.imageError(img, e);
            });
            img._goJSAttachedEvents = true;
          }
        }
      }
      this.raiseChanged('element', old, value);
      this.invalidateDraw();
    }
  }

  /**
   * @hidden @internal
   * Check if an image will fail or succeed as a cross-origin resource.
   * Important when calling Diagram.makeImage
   * @this {Picture}
   */
  private _testCors(): void {
    if (this._element === null) {
      this._tainted = false;
      return;
    }
    const ctx = new CanvasSurface(null).context;
    ctx.drawImage(this._element, 0, 0);
    try {
      // dummy statement to make sure it stays after compilation
      if (ctx.getImageData(0, 0, 1, 1).data[3]) this._tainted = false;
      this._tainted = false;
    } catch (e) {
      this._tainted = true;
    }
  }

  /**
   * Gets or sets the Picture's source URL, which can be any valid image (png, jpg, gif, etc) URL.
   *
   * The default value is the empty string, which specifies no image source.
   * Setting this attribute creates an HTMLImageElement and sets the {@link Picture#element}
   * attribute to that element.
   * Setting the source of multiple Pictures to the same URL will cause only one HTMLImageElement
   * to be created and shared.
   * It is commonplace to either specify a constant URL or to data bind this property to some data property,
   * perhaps using a conversion function in order to produce a proper URL.
   *
   * To avoid remeasuring and rearranging Parts as images load asynchronously,
   * be sure to set the {@link GraphObject#desiredSize}
   * (or {@link GraphObject#width} and {@link GraphObject#height}) to fixed values
   *
   * For cross-browser support of SVG sources additional care is needed.
   * See the final section of the <a href="../../intro/pictures.html">Introduction page on Pictures</a>.
   */
  /** @export */ get source(): string { return this._source; }
  /** @export */ set source(value: string) {
    const old = this._source;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', Picture, 'source');
      this._source = value;
      let map = Picture._imageSourceMap;

      const diagram = this.diagram;

      let img: HTMLImageElement = null;
      if (map[value] !== undefined) {
        img = map[value];
      } else {
        // image doesn't exist in the map
        // ??? no-op if isUndoingRedoing
        // 30 is arbitrarily the max number allowed in the _imageSourceMap
        if (Picture._imageSourceMapCount > 30) {
          Picture.clearCache();
          map = Picture._imageSourceMap;
        }
        img = Util.createElement('img') as HTMLImageElement;
        const self = this;
        img.addEventListener('load', function (e) {
          Picture.imageLoaded(img, e);
          // In case the first use is on a template, which may also need to reload
          if (!self.desiredSize.isReal()) {
            self.setInvalidMeasure(false);
            self.invalidateMeasure();
          }
        });
        img.addEventListener('error', function (e) {
          Picture.imageError(img, e);
        });
        img._goJSAttachedEvents = true;

        const srcCross = this._sourceCrossOrigin;
        if (srcCross !== null) {
          img['crossOrigin'] = srcCross(this);
        }
        // Must be set after crossOrigin flag. See sourceCrossOrigin setter
        img.src = value;

        map[value] = img;

        Picture._imageSourceMapCount++;
      }
      if (diagram !== null) diagram.forgetPicture(this);
      this.element = img;
      if (diagram !== null) diagram.rememberPicture(this);
      if (img['__goCache'] === undefined) img['__goCache'] = new PictureCacheArray();

      // invalidate right away when a picture source changes, in case its already loaded
      this.invalidateMeasure();
      this.invalidateDraw();
      this.raiseChanged('source', old, value);
    }
  }

  /**
   * @hidden @internal
   * Notify every diagram that a contained Picture might have loaded
   * @param {HTMLImageElement} img
   * @param {Event} e
   */
  public static imageLoaded(img: HTMLImageElement, e: Event): void {
    img._isLoaded = true;
    img._hasError = false;

    let pic: Picture = null;
    const diags = Util.getDiagramsTraversed();
    const l = diags.length;
    for (let i = 0; i < l; i++) {
      const diagram = diags[i];
      const url = img.getAttribute('src');
      const pics = diagram._picturesMap.getValue(url);
      if (pics === null) continue;
      const l2 = pics.length;
      for (let j = 0; j < l2; j++) {
        pic = pics[j];
        // if desiredSize is real, the image's node might still need to be notified
        diagram._picturesToInvalidate.add(pic);
        diagram.requestUpdate();
        // if it was note already called
        if (img._savedSuccessEvent === undefined) {
          img._savedSuccessEvent = e;
          if (pic._successFunction !== null) pic._successFunction(pic, e);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Notify every diagram that a contained Picture might have error'd
   * @param {HTMLImageElement} img
   * @param {Event} e
   */
  public static imageError(img: HTMLImageElement, e: Event): void {
    img._hasError = e;

    let pic: Picture = null;
    const diags = Util.getDiagramsTraversed();
    const l = diags.length;
    for (let i = 0; i < l; i++) {
      const diagram = diags[i];
      const url = img.getAttribute('src');
      const pics = diagram._picturesMap.getValue(url);
      if (pics === null) continue;
      const l2 = pics.length;
      // might modify the picture map, so operate on an array copy
      const newarr = Util.tempArray<Picture>();
      for (let j = 0; j < l2; j++) {
        newarr.push(pics[j]);
      }
      for (let j = 0; j < l2; j++) {
        pic = newarr[j];
        if (pic._errorFunction !== null) pic._errorFunction(pic, e);
      }
      Util.freeArray(newarr);
    }
  }

  //

  /**
   * Attempts to reload a {@link Picture#source} image. This can be useful if the content on a server has changed, or was missing before.
   * If a new image is loaded, this Picture may remeasure and/or redraw.
   *
   * This should normally be called within a transaction.
   * @since 2.1
   */
  public reloadSource() {
    if (this.source === '') return;
    Picture.clearCache(this.source);
    const old = this.source;
    this.source = '';
    this.source = old;
  }

  /**
   * Redraws a Picture, which can be useful if the backing {@link #element} is an HTML Canvas or HTML Video that has changed.
   * This will not attempt to *reload* any image. If you need to do that, call {@link #reloadSource}.
   *
   * This does not need to be called within a transaction, and will not re-measure anything.
   * @since 2.1
   */
  public redraw() {
    this.invalidateDraw();
  }

  /**
   * Gets or sets a function that returns a value for image.crossOrigin.
   *
   * The default value is null, which will not set a value for image.crossOrigin.
   * Set this property to a function that returns "anonymous" or "use-credentials"
   * if you want images to complete a cross-origin request.
   * Example:
   * ```js
   *    $(go.Picture,
   *      { width: 64, height: 64 },
   *      { sourceCrossOrigin: function(pict) { return "use-credentials"; } },
   *      new go.Binding("source", "path"))
   * ```
   * @since 1.5
   */
  /** @export */ get sourceCrossOrigin(): ((a: Picture) => string) | null { return this._sourceCrossOrigin; }
  /** @export */ set sourceCrossOrigin(value: ((a: Picture) => string) | null) {
    const old = this._sourceCrossOrigin;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Picture, 'sourceCrossOrigin');
      this._sourceCrossOrigin = value;
      if (this.element !== null) {
        const fullsrc = this.element.getAttribute('src');
        if (value === null && typeof fullsrc === 'string') {
          this.element['crossOrigin'] = null;
        } else if (value !== null) {
          this.element['crossOrigin'] = value(this);
        }
        // fixes chrome and IE caching bugs
        // example: https://bugs.chromium.org/p/chromium/issues/detail?id=678265
        this.element.src = fullsrc;
      }
    }
  }

  /**
   * Gets or sets the rectangular area of the source image that this picture should display.
   * This is only common with sprite maps and image tables.
   *
   * The value must be of type {@link Rect}.
   * The default value is Rect(NaN, NaN, NaN, NaN), which means the whole source image should be used.
   */
  /** @export */ get sourceRect(): Rect { return this._sourceRect; }
  /** @export */ set sourceRect(value: Rect) {
    const old = this._sourceRect;
    if (!old.equals(value)) {
      Util.checkClass(value, Rect, Picture, 'sourceRect');
      value = value.copyFrozen();
      this._sourceRect = value;
      this.invalidateDraw();
      this.raiseChanged('sourceRect', old, value);
    }
  }

  /**
   * Gets or sets how the Picture's image is stretched within its bounding box.
   * Some images will be a different aspect ratio than their given size or {@link #desiredSize}.
   * This property will never change the size of the Picture itself,
   * only the size of the image that is drawn in the Picture's {@link #actualBounds}.
   *
   * Possible values are {@link GraphObject.Fill}, {@link GraphObject.Uniform},
   * {@link GraphObject.UniformToFill}, and {@link GraphObject.None}.
   * The default is {@link GraphObject.Fill}.
   * @see {@link #imageAlignment}
   */
  /** @export */ get imageStretch(): EnumValue { return this._imageStretch; }
  /** @export */ set imageStretch(value: EnumValue) {
    const old = this._imageStretch;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, Picture, 'imageStretch');
      this._imageStretch = value;
      this.invalidateDraw();
      this.raiseChanged('imageStretch', old, value);
    }
  }

  /**
   * Gets or sets how the Picture is displayed: Either normally or with a Horizontal or Vertical flip or both.
   *
   * Possible values are {@link GraphObject.None}, {@link GraphObject.FlipHorizontal}, {@link GraphObject.FlipVertical}, or {@link GraphObject.FlipBoth}.
   * The default is {@link GraphObject.None}.
   * @since 2.0
   */
  /** @export */ get flip(): EnumValue { return this._flip; }
  /** @export */ set flip(value: EnumValue) {
    const old = this._flip;
    if (old !== value) {
      EnumValue.checkEnumValue(value, GraphObject, Picture, 'flip');
      this._flip = value;
      this.invalidateDraw();
      this.raiseChanged('flip', old, value);
    }
  }

  /**
   * Gets or sets the Spot to align the source image to, when the source image
   * is smaller than the Picture. This is only relevant when
   * the {@link #imageStretch} property value is not go.GraphObject.Fill.
   *
   * This does not affect Picture coordinates or bounds, it only affects what is drawn within the given area.
   * The default value is {@link Spot.Center}.
   * @see {@link #imageStretch}
   * @since 1.7
   */
  /** @export */ get imageAlignment(): Spot { return this._imageAlignment; }
  /** @export */ set imageAlignment(value: Spot) {
    Util.checkClass(value, Spot, Picture, 'imageAlignment');
    const old = this._imageAlignment;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._imageAlignment = value;
      this.invalidateMeasure();
      this.raiseChanged('imageAlignment', old, value);
    }
  }

  /**
   * Gets or sets the function to call if an image fails to load.
   * The arguments to this function are this Picture and the HTMLImageElement's "error" (onerror) Event.
   * The default value is null, meaning that no specific action occurs when there is an error loading an image.
   * @see {@link #successFunction}
   */
  /** @export */ get errorFunction(): ((a: Picture, b: Event) => void) | null { return this._errorFunction; }
  /** @export */ set errorFunction(value: ((a: Picture, b: Event) => void) | null) {
    const old = this._errorFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Picture, 'errorFunction');
      this._errorFunction = value;
      this.raiseChanged('errorFunction', old, value);
    }
  }

  /**
   * Gets or sets the function to call when an image loads successfully.
   * The arguments to this function are this Picture and the HTMLImageElement's "load" Event.
   * The default value is null, meaning that no specific action occurs when an image finishes loading.
   * @see {@link #errorFunction}
   * @since 1.7
   */
  /** @export */ get successFunction(): ((a: Picture, b: Event) => void) | null { return this._successFunction; }
  /** @export */ set successFunction(value: ((a: Picture, b: Event) => void) | null) {
    const old = this._successFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Picture, 'successFunction');
      this._successFunction = value;
      this.raiseChanged('successFunction', old, value);
    }
  }

  /**
   * @hidden @internal
   * Draw the Picture to a specified context.
   * @this {Picture}
   * @param {IContext} context A Canvas 2D Context for drawing.
   * @param {Diagram} diagram
   */
  public drawProtected(context: IContext, diagram: Diagram): void {
    const elem = this._element;
    if (elem === null) return;
    const src = elem.getAttribute('src');
    if (elem instanceof HTMLImageElement && (src === null || src === '')) {
      Util.throwError('Element has no source ("src") attribute: ' + elem);
    }

    if (elem._hasError instanceof Event) return;
    // ?? not all elems will have this, and in SVG rendering we don't want to check it, would need "&& context instanceof CanvasSurfaceContext"
    // if (elem._isLoaded !== true) return;

    const siz = this.naturalBounds;
    // the source image size:
    let sx = 0.0;
    let sy = 0.0;
    const iscan = this._isCanvas;
    let sw = iscan ? +elem.width : elem.naturalWidth;
    let sh = iscan ? +elem.height : elem.naturalHeight;
    if (sw === undefined && elem.videoWidth) sw = elem.videoWidth;
    if (sh === undefined && elem.videoHeight) sh = elem.videoHeight;

    // In IE11 at least, its possible for a loaded SVG image element
    // to have no .width/height and no .naturalWidth/height
    // In those cases we must defer to the desiredSize, if there is any
    sw = sw || siz.width;
    sh = sh || siz.height;

    if (sw === 0 || sh === 0) return; // Image failed to load or is nonexistent
    const originalWidth = sw;
    const originalHeight = sh;

    if (this.sourceRect.isReal()) {
      sx = this._sourceRect.x;
      sy = this._sourceRect.y;
      sw = this._sourceRect.width;
      sh = this._sourceRect.height;
    }

    // ??? we shouldn't do this during draw time
    // these become the destination size:
    let nw = sw;
    let nh = sh;
    const stretch = this._imageStretch;
    const align = this._imageAlignment;
    switch (stretch) {
      case GraphObject.None: // Draw it as the natural size
        // we only want a slice of the source rect
        // that is as large as the natural bounds of the object
        if (this.sourceRect.isReal()) break; // ?? suspicious
        if (nw >= siz.width) sx = (sx + align.offsetX) + (nw * align.x - (siz.width * align.x));
        if (nh >= siz.height) sy = (sy + align.offsetY) + (nh * align.y - (siz.height * align.y));
        // pre image alignment:
        sw = Math.min(siz.width, nw);
        sh = Math.min(siz.height, nh);
        break;
      case GraphObject.Fill:
        // we want the entire object as the source rect:
        nw = siz.width;
        nh = siz.height;
        break;
      case GraphObject.Uniform:
      case GraphObject.UniformToFill:
        // we want the entire object as the source rect:
        let ratio = 0.0;
        if (stretch === GraphObject.Uniform) {// if its Uniform we want the larger side to fit
          ratio = Math.min(siz.height / nh, siz.width / nw);
          nw *= ratio;
          nh *= ratio;
        } else if (stretch === GraphObject.UniformToFill) {// if its UniformToFill we want the smaller side to fit
          ratio = Math.max(siz.height / nh, siz.width / nw);
          // uniform to fill is more complex because we are also cropping the image to avoid any need for clipping
          nw *= ratio;
          nh *= ratio;
          // we want the source element to be "cropped" here

          if (nw >= siz.width) sx = (((sx + align.offsetX) + (nw * align.x - (siz.width * align.x)) / nw) * sw);
          if (nh >= siz.height) sy = (((sy + align.offsetY) + (nh * align.y - (siz.height * align.y)) / nh) * sh);
          // pre image alignment:
          // sx += (((nw - siz.width)/ 2)/nw)*sw;
          // sy += (((nh - siz.height) / 2)/nh)*sh;
          sw *= 1 / ((nw) / siz.width);
          sh *= 1 / ((nh) / siz.height);
          // we want the resulting image to be exactly the width and height of the picture
          nw = siz.width;
          nh = siz.height;
        }
        break;
    }

    const pixelCount = sw * sh;
    const scale = this.getDocumentScale() * diagram.scale;
    const screenwidth = nw * scale;
    const screenheight = nh * scale;
    const finalPixelCount = screenwidth * screenheight;

    // squared ratios to avoid sqrt use
    const pixelratioSquared = pixelCount / finalPixelCount;

    const source = elem['__goCache'] as PictureCacheArray;
    let finalimg = null;
    const maxratio = Picture._maxPixelRatio;

    if (elem._isLoaded && source !== undefined && finalPixelCount > 4 && pixelratioSquared > (maxratio * maxratio)) {
      // ??? how do we decide what ratios to build and when?
      //     when to rebuild them? discard them?
      //     For now, we'll just make two scaled down by 4x and 16x
      if (source.array === null) {
        source.drawSmallerInstance(4, originalWidth, originalHeight, elem);
        source.drawSmallerInstance(16, originalWidth, originalHeight, elem);
      }

      // go through all looking for the one with the largest ratio smaller than pixelratioSquared
      const arr = source.array;
      const length = arr.length; // if none created, finalImage will stay null
      for (let i = 0; i < length; i++) {
        if (arr[i].ratio * arr[i].ratio < pixelratioSquared) {
          finalimg = arr[i];
        } else {
          break; // ??? assumes they are in order of ascending ratio
        }
      }
    }

    if (!diagram._drawTainted) {
      if (this._tainted === null) this._testCors();
      if (this._tainted) return;
    }

    let dx = 0;
    if (nw < siz.width) dx = (align.offsetX) + (siz.width * align.x - (nw * align.x));
    let dy = 0;
    if (nh < siz.height) dy = (align.offsetY) + (siz.height * align.y - (nh * align.y));

    switch (this.flip) {
      case GraphObject.None:
        break; // do nothing!
      case GraphObject.FlipHorizontal:
        context.translate(Math.min(siz.width, nw), 0);
        context.scale(-1, 1);
        break;
      case GraphObject.FlipVertical:
        context.translate(0, Math.min(siz.height, nh));
        context.scale(1, -1);
        break;
      case GraphObject.FlipBoth:
        context.translate(Math.min(siz.width, nw), Math.min(siz.height, nh));
        context.scale(-1, -1);
        break;
    }

    // ??? try-catch needed for some Firefox oddities with SVG images
    //     https://bugzilla.mozilla.org/show_bug.cgi?id=574330
    //     and because of this?
    //     https://bugzilla.mozilla.org/show_bug.cgi?id=1109693
    //     but try/catch may slow down rendering, better to find another way

    // optimization, so don't run code if drawForeign is in use
    if (diagram.getRenderingHint('pictureRatioOptimization') &&
      !diagram._isDrawingForeign &&
      source !== undefined &&
      finalimg !== null && finalimg.ratio !== 1) {
      context.save();
      const ratio = finalimg.ratio as number;
      try {
        context.drawImage(
          finalimg.source,
          sx / ratio, sy / ratio, Math.min(finalimg.source.width, sw / ratio), Math.min(finalimg.source.height, sh / ratio),
          // dx, dy, dw, dh
          dx, // destination x, typically zero unless the image is smaller than size
          dy, // destination y, (same)
          Math.min(siz.width, nw), Math.min(siz.height, nh));
      } catch (ex) {
        if (Debug && this._firstError) Util.trace(ex.toString());
        this._firstError = false;
      }
      context.restore(); // doesn't set fill/brush/font, so no clearContextCache call needed
    } else { // Normal:
      try {
        context.drawImage(
          elem,
          sx, sy, sw, sh,
          // dx, dy, dw, dh
          dx, // destination x, typically zero unless the image is smaller than size
          dy, // destination y, (same)
          Math.min(siz.width, nw), Math.min(siz.height, nh));
      } catch (ex) {
        if (Debug && this._firstError) Util.trace(ex.toString());
        this._firstError = false;
      }
    }

    switch (this.flip) {
      case GraphObject.None:
        break; // do nothing!
      case GraphObject.FlipHorizontal:
        context.scale(-1, 1);
        context.translate(-Math.min(siz.width, nw), 0);
        break;
      case GraphObject.FlipVertical:
        context.scale(1, -1);
        context.translate(0, -Math.min(siz.height, nh));
        break;
      case GraphObject.FlipBoth:
        context.scale(-1, -1);
        context.translate(-Math.min(siz.width, nw), -Math.min(siz.height, nh));
        break;
    }
  } // end draw

  /**
   * This read-only property returns the natural size of this picture as determined by its source's width and height.
   * The value is initially NaN x NaN until the picture has been measured.
   * It will use the {@link #element}'s `naturalWidth` and `naturalHeight` if available.
   */
  /** @export */ get naturalBounds(): Rect { return this._naturalBounds; }

  /**
   * @hidden @internal
   * @this {Picture}
   * @param {number} width
   * @param {number} height
   * @param {number=} minw
   * @param {number=} minh
   */
  protected measureProtected(width: number, height: number, minw?: number, minh?: number): void {
    const des = this.desiredSize;
    const stretch = this.getEffectiveStretch(true);
    const elem = this._element;
    const iscan = this._isCanvas;
    if (iscan || (!this._loaded && elem && elem.complete)) this._loaded = true;
    if (elem === null) {
      if (!isFinite(des.width)) width = 0;
      if (!isFinite(des.height)) height = 0;
    }

    if (!isFinite(des.width) && stretch !== GraphObject.Fill && stretch !== GraphObject.Horizontal) {
      if (elem !== null && this._loaded !== false) {
        if (this.sourceRect.isReal()) {
          width = this.sourceRect.width;
        } else {
          width = iscan ? +elem.width : elem.naturalWidth;
        }
      }
    } else {
      if (!isFinite(width)) {
        if (this.sourceRect.isReal()) {
          width = this.sourceRect.width;
        } else {
          width = iscan ? +elem.width : elem.naturalWidth;
        }
      }
      minw = 0;
    }

    if (!isFinite(des.height) && stretch !== GraphObject.Fill && stretch !== GraphObject.Vertical) {
      if (elem !== null && this._loaded !== false) {
        if (this.sourceRect.isReal()) {
          height = this.sourceRect.height;
        } else {
          height = iscan ? +elem.height : elem.naturalHeight;
        }
      }
    } else {
      if (!isFinite(height)) {
        if (this.sourceRect.isReal()) {
          height = this.sourceRect.height;
        } else {
          height = iscan ? +elem.height : elem.naturalHeight;
        }
      }
      minh = 0;
    }
    if (isFinite(des.width)) {
      width = des.width;
    }
    if (isFinite(des.height)) {
      height = des.height;
    }
    // ?? consider removing:
    const max = this.maxSize;
    const min = this.minSize;
    minw = Math.max(minw, min.width);
    minh = Math.max(minh, min.height);
    width = Math.min(max.width, width);
    height = Math.min(max.height, height);
    width = Math.max(minw, width);
    height = Math.max(minh, height);

    if (elem !== null && !elem.complete) {
      if (!isFinite(width)) width = 0;
      if (!isFinite(height)) height = 0;
    }

    const nb = this._naturalBounds;
    nb.assignWidthHeight(width, height);

    this._measureRect(0, 0, width, height);
  } // end measure

  /**
   * @hidden @internal
   * @this {Picture}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    this.commonArrange(fx, fy, fw, fh);
  }

  /** @hidden @internal @type {PropertyCollection} */
  public static _imageSourceMap: PropertyCollection = null;
  /** @hidden @internal */
  public static _imageSourceMapCount = 0;
  /** @hidden @internal */
  public static _maxPixelRatio = 4; // if an image reaches max pixel count and is to be drawn maxPixelRatio times smaller, cache it

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Picture._imageSourceMap = new PropertyCollection();
    Picture['clearCache'] = Picture.clearCache;
  }
} // #end Picture
Picture['className'] = 'Picture';

Picture.initStatic();

/**
 * @hidden @internal
 * @unrestricted
 */
export class PictureCacheArray {
  /** @hidden @internal */ public static className = 'PictureCacheArray';
  /** @hidden @internal */ public array: Array<PictureCacheInstance>;

  /**
   *
   */
  constructor() {
    // has an array of PictureCacheInstance objects, implicitly sorted by ratio size?
    this.array = null; // [new PictureCacheInstance(img, 1)];
  }

  /**
   * @hidden @internal
   * @this {PictureCacheArray}
   * @param {number} ratio perhaps 2,4,8,16?
   * @param {number} sw original image width
   * @param {number} sh original image height
   */
  public drawSmallerInstance(ratio: number, sw: number, sh: number, img: HTMLImageElement | HTMLCanvasElement | HTMLVideoElement): void {
    if (this.array === null) this.array = [];
    const can = new CanvasSurface(null);
    const ctx = can.context;
    const scale = 1 / ratio;
    can.width = sw / ratio;
    can.height = sh / ratio;
    if (can.width === 0 || can.height === 0) return; // Do not draw empty canvas
    const instance = new PictureCacheInstance(can.domElement, ratio);
    let source = img;
    let sourceratio = 1;
    if (this.array.length > 0) {
      const last = this.array[this.array.length - 1];
      source = last.source;
      sourceratio = last.ratio;
    }
    ctx.setTransform(scale * sourceratio, 0, 0, scale * sourceratio, 0, 0);
    ctx.drawImage(source, 0, 0);

    /*
    // For debugging, mark the smaller images
    ctx.setTransform(1,0,0,1,0,0);
    var color = 'red';
    if (ratio === 8) color = 'lime';
    if (ratio === 16) color = 'blue';
    ctx.fillStyle = color;
    ctx.fillRect(0,0,10,10);
    root.document.body.appendChild(can.domElement);
    */

    this.array.push(instance);
  }

} // #end PictureCacheArray
PictureCacheArray['className'] = 'PictureCacheArray';

/**
 * @hidden @internal
 * @unrestricted
 */
export class PictureCacheInstance {
  /** @hidden @internal */ public static className = 'PictureCacheInstance';
  /** @hidden @internal */ public source: HTMLImageElement | HTMLCanvasElement;
  /** @hidden @internal */ public ratio: number;

  /**
   * @param {HTMLImageElement|HTMLCanvasElement} source
   * @param {number} ratio comparison to the original
   */
  constructor(source: HTMLImageElement | HTMLCanvasElement, ratio: number) {
    this.source = source;
    this.ratio = ratio;
  }

} // #end PictureCacheInstance
PictureCacheInstance['className'] = 'PictureCacheInstance';

// import { CanvasSurface } from './abstractCanvas';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { IContext } from './surfaceInterfaces';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

// internal definitions of standard shapes


// import { Geo } from './geo';

/**
 * @hidden @internal
 * Used in many figure generators, for source compatibility with GoXam and GoDiagram.
 * Also used in Geometry.parse and in Link.makeGeometry.
 * @unrestricted
 */
export class StreamGeometryContext {
  /** @hidden @internal */ public static className = 'StreamGeometryContext';
  /** @hidden @internal */ public _fig: PathFigure;
  /** @hidden @internal */ public geo: Geometry;

  constructor() {
    this.geo = new Geometry();
    this._fig = null;
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {Geometry | null} geo
   */
  public reset(geo: Geometry | null): void {
    if (geo !== null) {
      geo.thaw();
      this.geo = geo;
      geo.figures.clear();
    } else {
      this.geo = new Geometry();
    }
    this._fig = null;
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} sx
   * @param {number} sy
   * @param {boolean} isFilled
   * @param {boolean=} isSubPath
   * @param {boolean=} isShadowed true if shadowed, false if not shadowed. Figures are shadowed by default.
   */
  public beginFigure(sx: number, sy: number, isFilled: boolean, isSubPath?: boolean, isShadowed?: boolean): void {
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    if (isSubPath !== undefined && isSubPath === true) {
      // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
      const seg = new PathSegment(PathSegment.Move);
      seg.endX = sx;
      seg.endY = sy;
      this._fig.segments.add(seg);
    } else {
      this._fig = new PathFigure();
      this._fig.startX = sx;
      this._fig.startY = sy;
      this._fig.isFilled = isFilled;
      this.geo.figures.add(this._fig);
    }
    if (isShadowed !== undefined) this._fig.isShadowed = isShadowed;
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   */
  public closeLast(): void {
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const len = this._fig.segments.length;
    if (len > 0) {
      this._fig.segments.elt(len - 1).close();
    }
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   */
  public fillLast(): void {
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const len = this._fig.segments.length;
    if (len > 0) {
      this._fig.isFilled = true;
    }
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {boolean} shadow whether or not the figure should be shadowed.
   */
  public setShadow(shadow: boolean): void {
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    this._fig.isShadowed = shadow;
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} px
   * @param {number} py
   * @param {boolean=} close
   */
  public moveTo(px: number, py: number, close?: boolean): void {
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.Move);
    seg.endX = px;
    seg.endY = py;
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} px
   * @param {number} py
   * @param {boolean=} close
   */
  public lineTo(px: number, py: number, close?: boolean): void {
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.Line);
    seg.endX = px;
    seg.endY = py;
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} p1x
   * @param {number} p1y
   * @param {number} p2x
   * @param {number} p2y
   * @param {number} ex
   * @param {number} ey
   * @param {boolean=} close
   */
  public bezierTo(p1x: number, p1y: number, p2x: number, p2y: number, ex: number, ey: number, close?: boolean): void {
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.Bezier);
    seg.point1X = p1x;
    seg.point1Y = p1y;
    seg.point2X = p2x;
    seg.point2Y = p2y;
    seg.endX = ex;
    seg.endY = ey;
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} px
   * @param {number} py
   * @param {number} ex
   * @param {number} ey
   * @param {boolean=} close
   */
  public quadraticBezierTo(px: number, py: number, ex: number, ey: number, close?: boolean): void {
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.QuadraticBezier);
    seg.point1X = px;
    seg.point1Y = py;
    seg.endX = ex;
    seg.endY = ey;
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} startAngle
   * @param {number} sweepAngle
   * @param {number} centerX
   * @param {number} centerY
   * @param {number} radiusX
   * @param {number=} radiusY
   * @param {boolean=} close
   */
  public arcTo(startAngle: number, sweepAngle: number, centerX: number, centerY: number, radiusX: number, radiusY?: number, close?: boolean): void {
    if (radiusY === undefined) radiusY = 0;
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.Arc);
    seg.startAngle = startAngle;
    seg.sweepAngle = sweepAngle;
    seg.centerX = centerX;
    seg.centerY = centerY;
    seg.radiusX = radiusX;
    if (radiusY !== 0) seg.radiusY = radiusY;
    else seg.radiusY = radiusX;
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /**
   * @hidden @internal
   * @this {StreamGeometryContext}
   * @param {number} rx
   * @param {number} ry
   * @param {number} xAxisRotation
   * @param {boolean} largeArcFlag
   * @param {boolean} sweepFlag
   * @param {number} cpx
   * @param {number} cpy
   * @param {boolean=} close
   */
  public svgArcTo(rx: number, ry: number, xAxisRotation: number, largeArcFlag: boolean, sweepFlag: boolean, cpx: number, cpy: number, close?: boolean): void {
    if (close === undefined) close = false;
    // if (Debug && this.geo === null) Util.throwError('StreamGeometryContext has been closed');
    // if (Debug && this._fig === null) Util.throwError('Need to call beginFigure first');
    const seg = new PathSegment(PathSegment.SvgArc, cpx, cpy, rx, ry, xAxisRotation, largeArcFlag, sweepFlag);
    if (close) seg.close();
    this._fig.segments.add(seg);
  }

  /** @hidden @internal */
  public static _CachedStreamGeometryContext: StreamGeometryContext = null;
  /**
   * @hidden @internal
   * @return {StreamGeometryContext}
   * @param {Geometry | null} geo
   */
  public static tempStreamGeometryContext(geo: Geometry | null): StreamGeometryContext {
    const sgc = StreamGeometryContext._CachedStreamGeometryContext;
    if (sgc !== null) {
      StreamGeometryContext._CachedStreamGeometryContext = null;
      sgc.reset(geo);
      return sgc;
    } else {
      const sg = new StreamGeometryContext();
      sg.reset(geo);
      return sg;
    }
  }

  /**
   * @hidden @internal
   * @param {StreamGeometryContext} temp
   */
  public static freeStreamGeometryContext(temp: StreamGeometryContext) {
    StreamGeometryContext._CachedStreamGeometryContext = temp;
  }

} // #end StreamGeometryContext
StreamGeometryContext['className'] = 'StreamGeometryContext';

// For version 2.0, only these need to be predefined:

/**
 * @hidden @internal
 * @param {string} name a capitalized figure name ("RoundedRectangle"); must not be "" or "None"
 * @param {function(Shape, number, number):Geometry|string} func A function that takes (Shape,width,height) and returns a Geometry,
 *    or an existing figure generator name for which the new name will be a synonym.
 */
function defineFigureGeneratorSimple(name: string, func: ((a: any /* so we don't import Shape */, b: number, c: number) => Geometry) | string): void {
  const low = name.toLowerCase();
  const gens = Geo.GeometryGenerators;
  gens[name] = func;
  gens[low] = name;
}



defineFigureGeneratorSimple('Rectangle', function(shape, w, h) {
  const cacheable = (Geo.CachedGeometryCount < 50);
  const name = 'r' + w + ',' + h;
  let geo = Geo.CachedGeometries[name];
  if (geo !== undefined) {
    return geo;
  } else {
    geo = new Geometry(Geometry.Rectangle);
    geo.endX = w;
    geo.endY = h;
    if (cacheable) {
      Geo.CachedGeometries[name] = geo;
      Geo.CachedGeometryCount++;
    }
    return geo;
  }
});

defineFigureGeneratorSimple('Square', function(shape, w, h) {
  const cacheable = (Geo.CachedGeometryCount < 50);
  const name = 's' + w + ',' + h;
  let geo = Geo.CachedGeometries[name];
  if (geo !== undefined) {
    return geo;
  } else {
    geo = new Geometry(Geometry.Rectangle);
    geo.endX = w;
    geo.endY = h;
    geo.defaultStretch = GraphObject.Uniform;
    if (cacheable) {
      Geo.CachedGeometries[name] = geo;
      Geo.CachedGeometryCount++;
    }
    return geo;
  }
});

defineFigureGeneratorSimple('RoundedRectangle', function(shape, w, h) {
  const cacheable = (Geo.CachedGeometryCount < 50);
  const name = 'rr' + w + ',' + h;
  let geo = Geo.CachedGeometries[name];
  if (geo !== undefined) {
    return geo;
  } else {
    let param1 = shape ? shape.parameter1 : NaN;
    if (isNaN(param1) || param1 <= 0) param1 = 5;  // default corner
    param1 = Math.min(param1, w / 3);
    param1 = Math.min(param1, h / 3);

    const cpOffset = param1 * Geo.KAPPA;
    geo = new Geometry()
      .add(new PathFigure(param1, 0, true)
        .add(new PathSegment(PathSegment.Line, w - param1, 0))
        .add(new PathSegment(PathSegment.Bezier, w, param1, w - cpOffset, 0, w, cpOffset))
        .add(new PathSegment(PathSegment.Line, w, h - param1))
        .add(new PathSegment(PathSegment.Bezier, w - param1, h, w, h - cpOffset, w - cpOffset, h))
        .add(new PathSegment(PathSegment.Line, param1, h))
        .add(new PathSegment(PathSegment.Bezier, 0, h - param1, cpOffset, h, 0, h - cpOffset))
        .add(new PathSegment(PathSegment.Line, 0, param1))
        .add(new PathSegment(PathSegment.Bezier, param1, 0, 0, cpOffset, cpOffset, 0).close()));
    if (cpOffset > 1) {
      geo.spot1 = new Spot(0, 0, cpOffset, cpOffset);
      geo.spot2 = new Spot(1, 1, -cpOffset, -cpOffset);
    }
    if (cacheable) {
      Geo.CachedGeometries[name] = geo;
      Geo.CachedGeometryCount++;
    }
    return geo;
  }
});

defineFigureGeneratorSimple('Border', 'RoundedRectangle');

defineFigureGeneratorSimple('Ellipse', function(shape, w, h) {
  const cacheable = (Geo.CachedGeometryCount < 50);
  const name = 'e' + w + ',' + h;
  let geo = Geo.CachedGeometries[name];
  if (geo !== undefined) {
    return geo;
  } else {
    geo = new Geometry(Geometry.Ellipse);
    geo.endX = w;
    geo.endY = h;
    geo.spot1 = Spot.GeneratorEllipseSpot1;
    geo.spot2 = Spot.GeneratorEllipseSpot2;
    if (cacheable) {
      Geo.CachedGeometries[name] = geo;
      Geo.CachedGeometryCount++;
    }
    return geo;
  }
});

defineFigureGeneratorSimple('Circle', function(shape, w, h) {
  const cacheable = (Geo.CachedGeometryCount < 50);
  const name = 'c' + w + ',' + h;
  let geo = Geo.CachedGeometries[name];
  if (geo !== undefined) {
    return geo;
  } else {
    geo = new Geometry(Geometry.Ellipse);
    geo.endX = w;
    geo.endY = h;
    geo.spot1 = Spot.GeneratorEllipseSpot1;
    geo.spot2 = Spot.GeneratorEllipseSpot2;
    geo.defaultStretch = GraphObject.Uniform;
    if (cacheable) {
      Geo.CachedGeometries[name] = geo;
      Geo.CachedGeometryCount++;
    }
    return geo;
  }
});

defineFigureGeneratorSimple('TriangleRight', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(0, 0)
      .add(new PathSegment(PathSegment.Line, w, 0.5 * h))
      .add(new PathSegment(PathSegment.Line, 0, h).close()))
    .setSpots(0, 0.25, 0.5, 0.75);
});

defineFigureGeneratorSimple('TriangleDown', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(0, 0)
      .add(new PathSegment(PathSegment.Line, w, 0))
      .add(new PathSegment(PathSegment.Line, 0.5 * w, h).close()))
    .setSpots(0.25, 0, 0.75, 0.5);
});

defineFigureGeneratorSimple('TriangleLeft', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(w, h)
      .add(new PathSegment(PathSegment.Line, 0, 0.5 * h))
      .add(new PathSegment(PathSegment.Line, w, 0).close()))
    .setSpots(0.5, 0.25, 1, 0.75);
});

defineFigureGeneratorSimple('TriangleUp', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(w, h)
      .add(new PathSegment(PathSegment.Line, 0, h))
      .add(new PathSegment(PathSegment.Line, 0.5 * w, 0).close()))
    .setSpots(0.25, 0.5, 0.75, 1);
});

defineFigureGeneratorSimple('Triangle', 'TriangleUp');

defineFigureGeneratorSimple('Diamond', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(0.5 * w, 0)
      .add(new PathSegment(PathSegment.Line, 0, 0.5 * h))
      .add(new PathSegment(PathSegment.Line, 0.5 * w, h))
      .add(new PathSegment(PathSegment.Line, w, 0.5 * h).close()))
    .setSpots(0.25, 0.25, 0.75, 0.75);
});

defineFigureGeneratorSimple('LineH', function(shape, w, h) {
  const geo = new Geometry(Geometry.Line);
  geo.startX = 0;
  geo.startY = h / 2;
  geo.endX = w;
  geo.endY = h / 2;
  return geo;
});

defineFigureGeneratorSimple('LineV', function(shape, w, h) {
  const geo = new Geometry(Geometry.Line);
  geo.startX = w / 2;
  geo.startY = 0;
  geo.endX = w / 2;
  geo.endY = h;
  return geo;
});

defineFigureGeneratorSimple('None', 'Rectangle');
defineFigureGeneratorSimple('BarH', 'Rectangle');
defineFigureGeneratorSimple('BarV', 'Rectangle');
defineFigureGeneratorSimple('MinusLine', 'LineH');

defineFigureGeneratorSimple('PlusLine', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(0, h / 2, false)
      .add(new PathSegment(PathSegment.Line, w, h / 2))
      .add(new PathSegment(PathSegment.Move, w / 2, 0))
      .add(new PathSegment(PathSegment.Line, w / 2, h)));
});

defineFigureGeneratorSimple('XLine', function(shape, w, h) {
  return new Geometry()
    .add(new PathFigure(0, h, false)
      .add(new PathSegment(PathSegment.Line, w, 0))
      .add(new PathSegment(PathSegment.Move, 0, 0))
      .add(new PathSegment(PathSegment.Line, w, h)));
});

// Arrowheads for links
/** @hidden @internal */
Geo.ArrowheadGeometryStrings = {
  '': '', // Empty name results in arrowhead instead of figure, but with an empty geometry
  'Standard': 'F1 m 0,0 l 8,4 -8,4 2,-4 z',
  'Backward': 'F1 m 8,0 l -2,4 2,4 -8,-4 z',
  'Triangle': 'F1 m 0,0 l 8,4.62 -8,4.62 z',
  'BackwardTriangle': 'F1 m 8,4 l 0,4 -8,-4 8,-4 0,4 z',
  'Boomerang': 'F1 m 0,0 l 8,4 -8,4 4,-4 -4,-4 z',
  'BackwardBoomerang': 'F1 m 8,0 l -8,4 8,4 -4,-4 4,-4 z',
  'SidewaysV': 'm 0,0 l 8,4 -8,4 0,-1 6,-3 -6,-3 0,-1 z',
  'BackwardV': 'm 8,0 l -8,4 8,4 0,-1 -6,-3 6,-3 0,-1 z',

  'OpenTriangle': 'm 0,0 l 8,4 -8,4',
  'BackwardOpenTriangle': 'm 8,0 l -8,4 8,4',
  'OpenTriangleLine': 'm 0,0 l 8,4 -8,4 m 8.5,0 l 0,-8',
  'BackwardOpenTriangleLine': 'm 8,0 l  -8,4 8,4 m -8.5,0 l 0,-8',

  'OpenTriangleTop': 'm 0,0 l 8,4 m 0,4',
  'BackwardOpenTriangleTop': 'm 8,0 l -8,4 m 0,4',
  'OpenTriangleBottom': 'm 0,8 l 8,-4',
  'BackwardOpenTriangleBottom': 'm 0,4 l 8,4',

  'HalfTriangleTop': 'F1 m 0,0 l 0,4 8,0 z m 0,8',
  'BackwardHalfTriangleTop': 'F1 m 8,0 l 0,4 -8,0 z m 0,8',
  'HalfTriangleBottom': 'F1 m 0,4 l 0,4 8,-4 z',
  'BackwardHalfTriangleBottom': 'F1 m 8,4 l 0,4 -8,-4 z',

  'ForwardSemiCircle': 'm 4,0 b 270 180 0 4 4',
  'BackwardSemiCircle': 'm 4,8 b 90 180 0 -4 4',

  'Feather': 'm 0,0 l 3,4 -3,4',
  'BackwardFeather': 'm 3,0 l -3,4 3,4',
  'DoubleFeathers': 'm 0,0 l 3,4 -3,4 m 3,-8 l 3,4 -3,4',
  'BackwardDoubleFeathers': 'm 3,0 l -3,4 3,4 m 3,-8 l -3,4 3,4',
  'TripleFeathers': 'm 0,0 l 3,4 -3,4 m 3,-8 l 3,4 -3,4 m 3,-8 l 3,4 -3,4',
  'BackwardTripleFeathers': 'm 3,0 l -3,4 3,4 m 3,-8 l -3,4 3,4 m 3,-8 l -3,4 3,4',

  'ForwardSlash': 'm 0,8 l 5,-8',
  'BackSlash': 'm 0,0 l 5,8',
  'DoubleForwardSlash': 'm 0,8 l 4,-8 m -2,8 l 4,-8',
  'DoubleBackSlash': 'm 0,0 l 4,8 m -2,-8 l 4,8',
  'TripleForwardSlash': 'm 0,8 l 4,-8 m -2,8 l 4,-8 m -2,8 l 4,-8',
  'TripleBackSlash': 'm 0,0 l 4,8 m -2,-8 l 4,8 m -2,-8 l 4,8',

  'Fork': 'm 0,4 l 8,0 m -8,0 l 8,-4 m -8,4 l 8,4',
  'BackwardFork': 'm 8,4 l -8,0 m 8,0 l -8,-4 m 8,4 l -8,4',
  'LineFork': 'm 0,0 l 0,8 m 0,-4 l 8,0 m -8,0 l 8,-4 m -8,4 l 8,4',
  'BackwardLineFork': 'm 8,4 l -8,0 m 8,0 l -8,-4 m 8,4 l -8,4 m 8,-8 l 0,8',
  'CircleFork': 'F1 m 6,4 b 0 360 -3 0 3 z m 0,0 l 6,0 m -6,0 l 6,-4 m -6,4 l 6,4',
  'BackwardCircleFork': 'F1 m 0,4 l 6,0 m -6,-4 l 6,4 m -6,4 l 6,-4 m 6,0 b 0 360 -3 0 3',
  'CircleLineFork': 'F1 m 6,4 b 0 360 -3 0 3 z m 1,-4 l 0,8 m 0,-4 l 6,0 m -6,0 l 6,-4 m -6,4 l 6,4',
  'BackwardCircleLineFork': 'F1 m 0,4 l 6,0 m -6,-4 l 6,4 m -6,4 l 6,-4 m 0,-4 l 0,8 m 7,-4 b 0 360 -3 0 3',

  'Circle': 'F1 m 8,4 b 0 360 -4 0 4 z',
  'Block': 'F1 m 0,0 l 0,8 8,0 0,-8 z',
  'StretchedDiamond': 'F1 m 0,3 l 5,-3 5,3 -5,3 -5,-3 z',
  'Diamond': 'F1 m 0,4 l 4,-4 4,4 -4,4 -4,-4 z',
  'Chevron': 'F1 m 0,0 l 5,0 3,4 -3,4 -5,0 3,-4 -3,-4 z',
  'StretchedChevron': 'F1 m 0,0 l 8,0 3,4 -3,4 -8,0 3,-4 -3,-4 z',

  'NormalArrow': 'F1 m 0,2 l 4,0 0,-2 4,4 -4,4 0,-2 -4,0 z',
  'X': 'm 0,0 l 8,8 m 0,-8 l -8,8',
  'TailedNormalArrow': 'F1 m 0,0 l 2,0 1,2 3,0 0,-2 2,4 -2,4 0,-2 -3,0 -1,2 -2,0 1,-4 -1,-4 z',
  'DoubleTriangle': 'F1 m 0,0 l 4,4 -4,4 0,-8 z  m 4,0 l 4,4 -4,4 0,-8 z',
  'BigEndArrow': 'F1 m 0,0 l 5,2 0,-2 3,4 -3,4 0,-2 -5,2 0,-8 z',
  'ConcaveTailArrow': 'F1 m 0,2 h 4 v -2 l 4,4 -4,4 v -2 h -4 l 2,-2 -2,-2 z',
  'RoundedTriangle': 'F1 m 0,1 a 1,1 0 0 1 1,-1 l 7,3 a 0.5,1 0 0 1 0,2 l -7,3 a 1,1 0 0 1 -1,-1 l 0,-6 z',
  'SimpleArrow': 'F1 m 1,2 l -1,-2 2,0 1,2 -1,2 -2,0 1,-2 5,0 0,-2 2,2 -2,2 0,-2 z',
  'AccelerationArrow': 'F1 m 0,0 l 0,8 0.2,0 0,-8 -0.2,0 z m 2,0 l 0,8 1,0 0,-8 -1,0 z m 3,0 l 2,0 2,4 -2,4 -2,0 0,-8 z',
  'BoxArrow': 'F1 m 0,0 l 4,0 0,2 2,0 0,-2 2,4 -2,4 0,-2 -2,0 0,2 -4,0 0,-8 z',
  'TriangleLine': 'F1 m 8,4 l -8,-4 0,8 8,-4 z m 0.5,4 l 0,-8',

  'CircleEndedArrow': 'F1 m 10,4 l -2,-3 0,2 -2,0 0,2 2,0 0,2 2,-3 z m -4,0 b 0 360 -3 0 3 z',

  'DynamicWidthArrow': 'F1 m 0,3 l 2,0 2,-1 2,-2 2,4 -2,4 -2,-2 -2,-1 -2,0 0,-2 z',
  'EquilibriumArrow': 'm 0,3 l 8,0 -3,-3 m 3,5 l -8,0 3,3',
  'FastForward': 'F1 m 0,0 l 3.5,4 0,-4 3.5,4 0,-4 1,0 0,8 -1,0 0,-4 -3.5,4 0,-4 -3.5,4 0,-8 z',
  'Kite': 'F1 m 0,4 l 2,-4 6,4 -6,4 -2,-4 z',
  'HalfArrowTop': 'F1 m 0,0 l 4,4 4,0 -8,-4 z m 0,8',
  'HalfArrowBottom': 'F1 m 0,8 l 4,-4 4,0 -8,4 z',
  'OpposingDirectionDoubleArrow': 'F1 m 0,4 l 2,-4 0,2 4,0 0,-2 2,4 -2,4 0,-2 -4,0 0,2 -2,-4 z',
  'PartialDoubleTriangle': 'F1 m 0,0 4,3 0,-3 4,4 -4,4 0,-3 -4,3 0,-8 z',
  'LineCircle': 'F1 m 0,0 l 0,8 m 7 -4 b 0 360 -3 0 3 z',
  'DoubleLineCircle': 'F1 m 0,0 l 0,8 m 2,-8 l 0,8 m 7 -4 b 0 360 -3 0 3 z',
  'TripleLineCircle': 'F1 m 0,0 l 0,8 m 2,-8 l 0,8 m 2,-8 l 0,8 m 7 -4 b 0 360 -3 0 3 z',
  'CircleLine': 'F1 m 6 4 b 0 360 -3 0 3 z m 1,-4 l 0,8',
  'DiamondCircle': 'F1 m 8,4 l -4,4 -4,-4 4,-4 4,4 m 8,0 b 0 360 -4 0 4 z',
  'PlusCircle': 'F1 m 8,4 b 0 360 -4 0 4 l -8 0 z m -4 -4 l 0 8',
  'OpenRightTriangleTop': 'm 8,0 l 0,4 -8,0 m 0,4',
  'OpenRightTriangleBottom': 'm 8,8 l 0,-4 -8,0',
  'Line': 'm 0,0 l 0,8',
  'DoubleLine': 'm 0,0 l 0,8 m 2,0 l 0,-8',
  'TripleLine': 'm 0,0 l 0,8 m 2,0 l 0,-8 m 2,0 l 0,8',
  'PentagonArrow': 'F1 m 8,4 l -4,-4 -4,0 0,8 4,0 4,-4 z'
}; // end Geo.ArrowheadGeometryStrings

// import { Geometry, PathFigure, PathSegment } from './geometry';
// import { GraphObject } from './graphObject';
// import { Spot } from './spot';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Part
Adornment
Node
Link
Group
Placeholder
*/

// import { EnumValue } from './enumValue';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';

/**
 * This is the base class for all user-manipulated top-level objects.
 * Because it inherits from {@link Panel}, it is automatically a visual container
 * of other {@link GraphObject}s.
 * Because it thus also inherits from {@link GraphObject}, it also has properties such as
 * {@link GraphObject#actualBounds}, {@link GraphObject#contextMenu}, and {@link GraphObject#visible}.
 *
 * If you just want an object that users can select and manipulate,
 * you can create an instance of this class.
 *
 * If you want an object that also supports being connected by links to other objects,
 * use the {@link Node} class, which inherits from Part.
 * Create those connections by using instances of the {@link Link} class.
 *
 * If you want a node that logically contains a subgraph of nodes and links,
 * use the {@link Group} class, which inherits from {@link Node}.
 *
 * If you want an object that decorates another Part, without having to modify that Part,
 * use the {@link Adornment} class.
 * Adornments do not support linking or grouping or being selected.
 *
 * You can construct a Part, add {@link GraphObject}s to it programmatically,
 * and then add the part to a diagram by calling {@link Diagram#add}.
 * However it is commonplace to add data to a model by setting its {@link Model#nodeDataArray} or calling {@link Model#addNodeData},
 * or for {@link Link}s, setting the {@link GraphLinksModel#linkDataArray} or calling {@link GraphLinksModel#addLinkData}.
 * Such actions will cause a diagram that is displaying the model to copy a template,
 * which is a Part that may have data {@link Binding}s, and add the new part to the diagram.
 * The {@link Panel#data} property will refer to that data object in the model.
 *
 * Some examples of adding Parts to a Diagram:
 * ```js
 * // A simple Part template
 * myDiagram.nodeTemplate =
 *   $(go.Part, "Horizontal",
 *     $(go.Shape, "Circle", { width: 20, height: 20 }),
 *     $(go.TextBlock, "Hello World")
 *   );
 * // Node templates can be either Nodes, or simple Parts
 * // (But not Groups, Adornments, or Links)
 *
 * // Adds copies of the nodeTemplate bound to the specified node data:
 * myDiagram.model.nodeDataArray =
 * [
 *   { key: "Alpha" },
 *   { key: "Beta" }
 * ];
 *
 * // Adds one copy of the nodeTemplate bound to the given node data:
 * myDiagram.model.addNodeData( { key: "Gamma" } );
 * ```
 * <p class="boxread">
 * See the <a href="../../intro/usingModels.html">Introduction on using Models</a>
 * for examples and more information.
 * <h3>Layers and Z-ordering</h3>
 *
 * Parts added to a Diagram exist in one of the Diagram's {@link Layer}s.
 * You can specify which layer the part should be in by setting {@link #layerName}.
 * Parts cannot be nested in the visual tree -- they cannot be added to other Parts of Panels.
 *
 * Parts can be individually z-ordered within a layer by setting {@link #zOrder}.
 * Parts within the same layer that have a higher zOrder number will be drawn above parts with a lower number.
 * <h3>Size and Position</h3>
 *
 * The size and position of a part are given by its {@link GraphObject#actualBounds}.
 * The size is determined by the {@link GraphObject}s that are elements inside this part.
 * You can change the position by setting {@link GraphObject#position} or {@link Part#location}.
 *
 * The "location" of a part is commonly the same as its "position".
 * The "position" is always the point that is at the top-left corner of the area occupied by the part.
 * But the "location" may be different from the "position" if you want to think of the part as being "at" a different spot in the part.
 * For example, you might want the "location" to be at the center of a {@link Picture} that has a {@link TextBlock} title of arbitrary size.
 * In this case you would set the {@link #locationSpot} to be {@link Spot.Center} and the {@link #locationObjectName}
 * to be the name of the Picture element in your Part.
 *
 * A part may be selected or de-selected by setting its {@link #isSelected} property.
 * This may also happen due to a call to {@link Diagram#select} or other operations that change the selection.
 * The user may change this property as part of the operation of the {@link ClickSelectingTool},
 * due to the user's mouse click, if the part is {@link #selectable}.
 * <h3>Ability Properties (Permissions)</h3>
 *
 * There are many properties named "...able", that control what operations the user may perform on this part.
 * These properties correspond to the similarly named properties on {@link Diagram} and {@link Layer} that govern the
 * behavior for all parts in all layers or for all parts in the given layer.
 * For example, the {@link Part#copyable} property corresponds to the properties {@link Diagram#allowCopy} and {@link Layer#allowCopy}.
 *
 * For each of these "ability" properties there is a corresponding "can..." predicate.
 * For example, the {@link Part#canCopy} predicate is false if any of the three previously named properties is false.
 * Commands and tools will normally call these predicates rather than just looking at Part properties.
 * <p class="boxread">
 * For more discussion about permissions, please read: <a href="../../intro/permissions.html">Permissions</a>.
 *
 * As previously mentioned, each {@link Diagram} supports the notion of selected parts.
 * One way of displaying that a part is selected is by modifying the part.
 * You can set the {@link #selectionChanged} property to be a function that is called when the value of {@link #isSelected} has changed;
 * it is passed the Part as the first argument.
 * The function can modify the color of one or more {@link GraphObject}s in the visual tree of that Part.
 * Or perhaps it could toggle the {@link GraphObject#visible} property of an object that is normally hidden when the part is not selected.
 *
 * The Part class also supports showing separate visual objects for a part when it gets selected.
 * These visuals are typically used to show that the part is selected ("selection handles")
 * or are used to allow the user to manipulate or modify the part with a tool ("tool handles").
 * These handles are instances of {@link Adornment}s.
 * The {@link #updateAdornments} method is responsible for showing or hiding adornments,
 * normally depending on whether the part is selected.
 *
 * When the {@link #selectionAdorned} property is true, a selected part automatically gets an {@link Adornment} created for it.
 * By default the selection adornment is just a simple blue box around the Part, and a blue shape following the route of a selected Link.
 * However you can set the {@link #selectionAdornmentTemplate} to an arbitrarily complex Adornment.
 * This way it can show more information or buttons for executing various commands when the user selects a Part.
 *
 * Tool handles are shown for those mode-less mouse-down tools that need it.
 * The process of updating adornments for a part will call {@link Tool#updateAdornments} on each tool in {@link ToolManager#mouseDownTools}.
 * Most tools might not need special tool handles.
 * But, for example, {@link ResizingTool} naturally will want to create an adornment with eight resize handles
 * positioned at the corners and at the middles of the sides of the selected node's visual element,
 * if the node has its {@link #canResize} function returning true.
 *
 * One may not always want the whole Part to get the selection handle or all tool handles.
 * Sometimes one wants to emphasize selection by highlighting a particular element within the part's visual tree.
 * This can be achieved by setting the {@link #selectionObjectName} property,
 * and making sure the desired element has the same {@link GraphObject#name} property value.
 * <p class="boxread">
 * For more discussion about selection, see <a href="../../intro/selection.html">Selection</a>.
 *
 * Similarly the {@link #resizeObjectName} and {@link #rotateObjectName} properties direct the
 * corresponding {@link ResizingTool} and {@link RotatingTool} to operate on the particular {@link GraphObject}
 * in the Part's visual tree with the given name.
 * That includes both providing tool handles and actually modifying properties on that object.
 *
 * Parts are not resizable or rotatable by default: you need to set {@link #resizable} and/or {@link #rotatable} to true.
 * <p class="boxread">
 * For more discussion about tools, see <a href="../../intro/tools.html">Tools</a>.
 *
 * A Part may be positioned (or a {@link Link} may be routed) by a {@link Layout}.
 * This will happen automatically if {@link Diagram#layout} or {@link Group#layout} are set.
 * The default {@link Diagram#layout} will position any nodes that were not given explicit positions or location.
 *
 * If you set {@link #isLayoutPositioned} to false, this part will not participate in any of the standard layouts,
 * so it will not be moved by a layout or affect other parts in a layout.
 * In order for the part to get a {@link #location} or position you will need to supply it explicitly.
 *
 * As parts are added to or removed from a diagram, the {@link Layout} responsible for positioning the part is invalidated.
 * This will cause the layout to be performed again in the near future, at the end of the transaction.
 * This automatic layout invalidation also occurs as parts change their visibility ({@link GraphObject#visible})
 * or their size ({@link GraphObject#actualBounds}).
 * If you do want there to be a {@link Diagram#layout} but you do not want an automatic layout to happen
 * after removing parts (for example), you can set {@link #layoutConditions} not to include the {@link Part.LayoutRemoved} flag.
 * In this particular case, you could set {@link #layoutConditions} to:
 * `go.Part.LayoutStandard & ~go.Part.LayoutRemoved`.
 * It may also reasonable for your application to set it to {@link Part.LayoutNone}.
 * Do not forget to consider applying the same conditions to links as well as to nodes and groups.
 *
 * If you want to save the locations/positions of the parts in a diagram, it is commonplace to
 * data bind the {@link #location} to a property on your node data with a TwoWay {@link Binding}
 * (call {@link Binding#makeTwoWay}). For example:
 * ```js
 * $(go.Part, "Horizontal",
 *   new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify),
 *   ...
 * ```
 *
 * Then as the nodes are moved, whether manually by the user or automatically by a {@link Layout},
 * the model data is automatically updated with the location.
 * <p class="boxread">
 * For more discussion about related topics, see <a href="../../intro/selection.html">Selection</a>,
 * <a href="../../intro/tools.html">Tools</a>, and
 * <a href="../../intro/permissions.html">Permissions</a>.
 *
 * Parts that are templates should have no relationships with other Parts.
 * Only real Parts that are in a Diagram can belong to Groups or have any Adornments.
 * Only real Nodes in a Diagram can be connected with Links.
 * @extends Panel
 * @unrestricted
 */
export class Part extends Panel {
  /** @hidden @internal */ public static className = 'Part';
  /** @hidden @internal */ private _layerChanged: ((thisPart: Part, oldLayer: Layer | null, newLayer: Layer | null) => void) | null;
  /** @hidden @internal */ private _selectionAdornmentTemplate: Adornment | null;
  /** @hidden @internal */ private _resizeAdornmentTemplate: Adornment | null;
  /** @hidden @internal */ private _rotateAdornmentTemplate: Adornment | null;
  /** @hidden @internal */ private _selectionObjectName: string;
  /** @hidden @internal */ private _selectionObject: GraphObject | null;
  /** @hidden @internal */ private _selectionChanged: ((thisPart: Part) => void) | null;
  /** @hidden @internal */ private _highlightedChanged: ((thisPart: Part) => void) | null;
  /** @hidden @internal */ private _adornments: Map<string, Adornment> | null;
  /** @hidden @internal */ private _resizeObjectName: string;
  /** @hidden @internal */ private _resizeCellSize: Size;
  /** @hidden @internal */ private _rotateObjectName: string;
  /** @hidden @internal */ private _text: string;
  /** @hidden @internal */ private _containingGroupChanged: ((thisPart: Part, oldGroup: Group | null, newGroup: Group | null) => void) | null;
  /** @hidden @internal */ private _layer: Layer | null;
  /** @hidden @internal */ protected _location: Point;
  /** @hidden @internal */ private _locationObjectName: string;
  /** @hidden @internal */ private _locationSpot: Spot;
  /** @hidden @internal */ private _rotationSpot: Spot;
  /** @hidden @internal */ private _minLocation: Point;
  /** @hidden @internal */ private _maxLocation: Point;
  /** @hidden @internal */ private _dragComputation: ((thisPart: Part, newLoc: Point, snappedLoc: Point) => Point) | null;
  /** @hidden @internal */ private _layoutConditions: number;
  /** @hidden @internal */ private _zOrder: number;
  /** @hidden @internal */ private _drawnRect: Rect;
  /** @hidden @internal */ private _nameMap: Map<string, GraphObject|null>;
  /** @hidden @internal */ private _maxMiter: number;
  /** @hidden @internal */ protected _containingGroup: Group | null;
  /** @hidden @internal */ protected _partFlags: number;
  /** @hidden @internal */ public _category: string;
  /** @hidden @internal */ public _layerName: string;
  /** @hidden @internal */ protected _locationObject: GraphObject | null;
  /** @hidden @internal */ private _shadowOffset: Point;
  /** @hidden @internal */ private _shadowColor: string;
  /** @hidden @internal */ private _shadowBlur: number;
  /** @hidden @internal */ public _layerIndex: number;

  /**
   * The constructor builds an empty Part.
   * @param {PanelLayout=} type if not supplied, the default Panel type is {@link Panel.Position}.
   */
  constructor(type?: PanelLayout) {
    super(type);

    // BIT FLAGS:

    // Part:
    // 0x00000001 this._copyable = true;
    // 0x00000002 this._deletable = true;
    // 0x00000004 this._textEditable = true;
    // 0x00000008 this._groupable = true;
    // 0x00000010 this._movable = true;
    // 0x00000020 this._selectionAdorned = true;
    // 0x00000040 this._isInDocumentBounds = true;
    // 0x00000080 this._isLayoutPositioned = true;
    // 0x00000100 this._selectable = true;
    // 0x00000200 this._reshapable = false;
    // 0x00000400 this._resizable = false;
    // 0x00000800 this._rotatable = false;
    // 0x00001000 this._isSelected = false;
    // 0x00002000 this._isShadowed = false;
    // 0x00004000 this._invalidAdornments = true;
    // 0x00008000 this._invalidPosition = true;
    // 0x00010000 this._notifyLater = false;
    // 0x00020000 this._inView = false;
    // 0x00040000 this._isAnimated = true;
    // 0x00080000 this._isHighlighted = false;
    // 0x00100000 this._commonParent = false;
    // 0x00200000 this._locSetLast = true;

    // 0x00000001|0x00000002|0x00000004|0x00000008|0x00000010|0x00000020|
    // 0x00000040|0x00000080|0x00000100|0x00004000|0x00008000|0x00040000|0x00200000
    this._partFlags = 2408959;

    this._category = '';
    this._layerName = '';
    this._layerChanged = null;

    this._selectionAdornmentTemplate = null;
    this._resizeAdornmentTemplate = null;
    this._rotateAdornmentTemplate = null;

    this._selectionObjectName = '';
    this._selectionObject = null;  // must be in visual tree
    this._selectionChanged = null;
    this._highlightedChanged = null;

    this._adornments = null;  // owned Adornments
    this._resizeObjectName = '';
    this._resizeCellSize = Size.NoSize;
    this._rotateObjectName = '';
    this._text = '';

    this._containingGroup = null;  // owning Group
    this._containingGroupChanged = null;
    this._layer = null;

    this._location = new Point(NaN, NaN).freeze();
    this._locationObjectName = '';
    this._locationObject = null;  // must be in visual tree
    this._locationSpot = Spot.TopLeft;
    this._rotationSpot = Spot.Default;
    this._minLocation = Point.InfiniteTopLeft;
    this._maxLocation = Point.InfiniteBottomRight;
    this._dragComputation = null;

    this._layoutConditions = Part.LayoutStandard;

    this._shadowOffset = Point.SixPoint;
    this._shadowColor = 'gray';
    this._shadowBlur = 4;

    // set only when drawn in the layer
    this._layerIndex = -1;
    this._zOrder = NaN;

    // used in getDrawnSize
    this._drawnRect = new Rect();

    this._nameMap = null;

    this._maxMiter = NaN;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Part.
   * @this {Part}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);

    // BIT FLAGS:
    //    0x00000001 copy._copyable = this._copyable;
    //    0x00000002 copy._deletable = this._deletable;
    //    0x00000004 copy._textEditable = this._textEditable;
    //    0x00000008 copy._groupable = this._groupable;
    //    0x00000010 copy._movable = this._movable;
    //    0x00000020 copy._selectionAdorned = this._selectionAdorned;
    //    0x00000040 copy._isInDocumentBounds = this._isInDocumentBounds;
    //    0x00000080 copy._isLayoutPositioned = this._isLayoutPositioned;
    //    0x00000100 copy._selectable = this._selectable;
    //    0x00000200 copy._reshapable = this._reshapable;
    //    0x00000400 copy._resizable = this._resizable;
    //    0x00000800 copy._rotatable = this._rotatable;

    //  DONT COPY - Default false:
    //    0x00001000 copy._isSelected = false;

    //    0x00002000 copy._isShadowed = this._isShadowed;

    //  DONT COPY any of these, use defaults:
    //    0x00004000 copy._invalidAdornments = true;
    //    0x00008000 copy._invalidPosition = true;
    //    0x00010000 copy._notifyLater = false;
    //    0x00020000 copy._inView = false;
    //    0x00100000 copy._commonParent = false;

    copy._partFlags = this._partFlags & (~0x00001000) | 0x00004000 | 0x00008000 & (~0x00010000) & (~0x00020000) & (~0x00100000);

    copy._category = this._category;
    copy._layerName = this._layerName;
    copy._layerChanged = this._layerChanged;

    copy._selectionAdornmentTemplate = this._selectionAdornmentTemplate;  // share templates
    copy._resizeAdornmentTemplate = this._resizeAdornmentTemplate;
    copy._rotateAdornmentTemplate = this._rotateAdornmentTemplate;

    copy._selectionObjectName = this._selectionObjectName;
    //  copy._selectionObject = null;  // cached
    copy._selectionChanged = this._selectionChanged;
    copy._highlightedChanged = this._highlightedChanged;

    copy._adornments = null;  // Parts need their own Adornments
    copy._resizeObjectName = this._resizeObjectName;
    copy._resizeCellSize = this._resizeCellSize.copyFrozen();
    copy._rotateObjectName = this._rotateObjectName;
    copy._rotationSpot = this._rotationSpot.copyFrozen();
    copy._text = this._text;

    // copy._containingGroup updated by copier
    copy._containingGroupChanged = this._containingGroupChanged;
    // copy._layer updated by copier

    copy._location.assign(this._location);
    copy._locationObjectName = this._locationObjectName;
    //  copy._locationObject = null;  // cached
    copy._locationSpot = this._locationSpot.copyFrozen();
    copy._minLocation = this._minLocation.copyFrozen();
    copy._maxLocation = this._maxLocation.copyFrozen();
    copy._dragComputation = this._dragComputation;

    copy._layoutConditions = this._layoutConditions;

    copy._shadowOffset = this._shadowOffset.copyFrozen();
    copy._shadowColor = this._shadowColor;
    copy._shadowBlur = this._shadowBlur;

    copy._zOrder = this._zOrder;

    // don't copy temporary properties
    // this._layerIndex
    // this._drawnRect
    // this._nameMap
    // this._maxMiter
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can get a new visual tree
   * @this {Part}
   * @param {GraphObject} copy  The original Part whose category is being changed by displacing its visual tree with this Part's visual tree.
   */
  protected displaceProtected(copy: this): void {
    super.displaceProtected(copy);
    // force these cached references to be recalculated
    copy.invalidateContainingGroup();
    copy._selectionObject = null;
    copy._locationObject = null;
    copy._nameMap = null;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @return {string}
   */
  public toString(): string {
    let s = Util.getClassName(this.constructor) + '#' + Set.hashId(this);
    if (this.data !== null) s += '(' + Util.toString(this.data) + ')';
    return s;
  }

  /**
   * This value may be used as the value of the {@link Part#layoutConditions} property to indicate that
   * no operation on this Part causes invalidation of the {@link Layout} responsible for this Part.
   * @constant
   */
  public static LayoutNone = 0;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Part is added to a Diagram or Group, it invalidates the {@link Layout} responsible for the Part.
   * @constant
   */
  public static LayoutAdded = 1;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Part is removed from a Diagram or Group, it invalidates the {@link Layout} responsible for the Part.
   * @constant
   */
  public static LayoutRemoved = 2;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Part's {@link GraphObject#visible} becomes true, it invalidates the {@link Layout} responsible for the Part.
   * @constant
   */
  public static LayoutShown = 4;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Part's {@link GraphObject#visible} becomes false, it invalidates the {@link Layout} responsible for the Part.
   * @constant
   */
  public static LayoutHidden = 8;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Node or simple Part's {@link GraphObject#actualBounds} changes size, it invalidates the {@link Layout} responsible for the Part;
   * this flag is ignored for Parts that are Links.
   * @constant
   */
  public static LayoutNodeSized = 16;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a {@link Group} has been laid out, it invalidates the {@link Layout} responsible for that Group;
   * this flag is ignored for Parts that are not Groups.
   * @constant
   */
  public static LayoutGroupLayout = 32;

  /**
   * This flag may be combined with other "Layout" flags as the value of the {@link Part#layoutConditions} property to indicate that
   * when a Node or simple Part's {@link #category} changes, it invalidates the {@link Layout} responsible for the Part;
   * this flag is ignored for Parts that are Links.
   * @constant
   */
  public static LayoutNodeReplaced = 64;

  /**
   * This is the default value for the {@link Part#layoutConditions} property, basically a combination of all of the conditions:
   * the {@link Layout} responsible for the Part is invalidated when the Part is added or removed or replaced
   * from the Diagram or Group, or when it changes visibility or size, or when a Group's layout has been performed.
   * @constant
   */
  public static LayoutStandard = Part.LayoutAdded | Part.LayoutRemoved | Part.LayoutShown | Part.LayoutHidden |
                                 Part.LayoutNodeSized | Part.LayoutGroupLayout | Part.LayoutNodeReplaced;

  /**
   * @hidden @internal
   * This value may be used as the value of {@link Part#layoutConditions} property to indicate that
   * the {@link Layout} responsible for the Part is invalidated when the Part is added or removed
   * from the Diagram or Group or when it changes visibility or size.
   * @constant
   */
  public static LayoutAll = 0xFFFFFF;

  /**
   * @hidden @internal
   * Call this method to notify that the some state has changed.
   * This just calls the {@link #diagram}'s {@link Diagram#raiseChangedEvent} method.
   * @this {Part}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raisePartChangedEvent(change: EnumValue, propertyname: string, obj: Object, oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    if (change === ChangedEvent.Insert && propertyname === 'elements') {
      if (newval instanceof Panel) {
        newval.walkPanels(function (p) {
          diagram.partManager.rememberItemArrayPanel(p);
          diagram.rememberPictures(p);
        });
      } else {
        diagram.rememberPicture(newval); // might be a Picture
      }
    } else if (change === ChangedEvent.Remove && propertyname === 'elements') {
      if (newval instanceof Panel) {
        newval.walkPanels(function (p) {
          diagram.partManager.forgetItemArrayPanel(p, diagram);
        });
      } else {
        diagram.forgetPicture(newval); // might be a Picture
      }
    }

    diagram.raiseChangedEvent(change, propertyname, obj, oldval, newval, oldparam, newparam);
  }

  /**
   * Re-evaluate all data bindings in this Part,
   * in order to assign new property values to the {@link GraphObject}s in this visual tree
   * based on this this object's {@link #data} property values.
   * This method does nothing if {@link #data} is null.
   *
   * It is better to call {@link Model#setDataProperty} to modify data properties,
   * because that will both record changes for undo/redo and will update all bindings
   * that make depend on that property.
   *
   * To update relationships between nodes, call {@link #updateRelationshipsFromData}.
   * @export
   * @this {Part}
   * @param {string=} srcprop An optional source data property name:
   *   when provided, only evaluates those {@link Binding}s that use that particular property;
   *   when not provided or when it is the empty string, all bindings are evaluated.
   * @see {@link #updateRelationshipsFromData}
   */
  public updateTargetBindings(srcprop?: string): void {
    super.updateTargetBindings(srcprop);
    if (this.data === null) return;
    const a = this._elements._arr;
    const l = a.length;
    for (let i = 0; i < l; i++) {
      const obj = a[i];
      if (obj instanceof Panel) {
        obj.walkPanels(function (p) { if (p.data !== null) p.updateTargetBindings(); });  // don't limit to SRCPROP
      }
    }
  }

  /**
   * Update all of the references to nodes in case they had been modified in the model without
   * properly notifying the model by calling {@link GraphLinksModel#setGroupKeyForNodeData} or
   * {@link GraphLinksModel#setToKeyForLinkData} or other similar methods.
   * This method does not conduct a transaction, so you need to start and commit one yourself.
   *
   * This only updates the relationships between nodes, to have them reflect what is now declared in the model data.
   * For example, in a GraphLinksModel if code has changed the value of the "to" property of a link data,
   * calling this method on the corresponding {@link Link} would cause the link to connect with
   * the {@link Node} whose data has the new key.
   *
   * To update {@link GraphObject} properties that are data bound, call {@link #updateTargetBindings}.
   * @expose
   * @this {Part}
   * @see {@link #updateTargetBindings}
   * @since 1.5
   */
  public updateRelationshipsFromData(): void {
    const data = this.data;
    if (data === null) return;
    const diagram = this.diagram;
    diagram.partManager.updateRelationshipsFromData(this);
  }


  /**
   * This read-only property returns the Part's Model data key if it is in a Diagram and is backed by Model data.
   * Otherwise this returns undefined.
   * @since 1.8
   * @see {@link Model#getKeyForNodeData}
   */
  /** @export */ get key(): Key {
    const diagram = this.diagram;
    if (diagram === null) return undefined;
    return diagram.model.getKeyForNodeData(this.data);
  }

  // Adornments

  /**
   * This read-only property returns an iterator over all of the {@link Adornment}s associated with this part.
   * After each call to the iterator's next() method that returns true,
   * the iterator's key will be the category and the iterator's value will
   * be an {@link Adornment}.
   *
   * Templates should not have any adornments.
   */
  /** @export */ get adornments(): Iterator<Adornment> {
    if (this._adornments === null) return EmptyIterator.instance;
    return this._adornments.iteratorValues;  // _adornments is a Map, want .each to operate on an Adornment, not a KVP
  }

  /**
   * Find an {@link Adornment} of a given category associated with this Part.
   *
   * Templates should not have any adornments.
   * @export
   * @this {Part}
   * @param {string} category
   * @return {Adornment} Returns null if no such Adornment was found.
   */
  public findAdornment(category: string): Adornment | null {
    if (Debug) Util.checkPrimitive(category, 'string', Part, 'findAdornment:category');
    const ads = this._adornments;
    if (ads === null) return null;
    return ads.getValue(category);
  }

  /**
   * Associate an {@link Adornment} with this Part, perhaps replacing any existing adornment of the same category.
   * Don't forget to set {@link Adornment#adornedObject} before calling this method.
   * This adds the Adornment to the {@link Layer} named by {@link #layerName}, normally "Adornment".
   *
   * Adornments are also data bound to the same data that this Part has, if any.
   * If the Adornment was already associated with a Part, it is unassociated with that old Part.
   *
   * This method should not be called on templates.
   * @export
   * @this {Part}
   * @param {string} category a string identifying the kind or role of the given adornment for this Part.
   * @param {Adornment} ad the new Adornment.
   */
  public addAdornment(category: string, ad: Adornment): void {
    if (ad === null) return;  // not adding or replacing anything
    if (Debug) {
      Util.checkPrimitive(category, 'string', Part, 'addAdornment:category');
      Util.checkClass(ad, Adornment, Part, 'addAdornment:ad');
    }
    let oldad: Adornment = null;
    let ads = this._adornments;
    if (ads !== null) oldad = ads.getValue(category);
    if (oldad !== ad) {
      if (oldad !== null) {
        const od = oldad.diagram;
        if (od !== null) od.remove(oldad);
      }
      if (ads === null) {
        ads = new Map<string, Adornment>();
        this._adornments = ads;
      }
      if (ad._category !== category) ad.category = category;
      ads.add(category, ad);
      const diagram = this.diagram;
      if (diagram !== null) {
        diagram.add(ad);
        // Adornments are data bound too, so that they can customized by the data
        const adObj = ad.adornedObject;
        if (adObj !== null) {
          const pan = adObj.findTemplateBinder();
          if (pan !== null) ad.data = pan.data;
        }
      }
    }
  }

  /**
   * Remove any {@link Adornment} of the given category that may be associated with this Part.
   * @export
   * @this {Part}
   * @param {string} category a string identifying the kind or role of the given adornment for this Part.
   */
  public removeAdornment(category: string): void {
    if (Debug) Util.checkPrimitive(category, 'string', Part, 'removeAdornment:category');
    const ads = this._adornments;
    if (ads === null) return;
    const oldad = ads.getValue(category);
    if (oldad !== null) {
      const od = oldad.diagram;
      if (od !== null) od.remove(oldad);
    }
    ads.remove(category);
    if (ads.count === 0) this._adornments = null;
  }

  /**
   * Remove all adornments associated with this part.
   * @export
   * @this {Part}
   */
  public clearAdornments(): void {
    const ads = this._adornments;
    if (ads === null) return;
    const list = Util.tempArray<string>();
    const it = ads.iterator;
    while (it.next()) {
      const name = it.key;
      list.push(name);
    }
    const l = list.length;
    for (let i = 0; i < l; i++) {
      this.removeAdornment(list[i]);  // remove Adornment from Layer and remove dictionary entry
    }
    Util.freeArray(list);
  }

  /**
   * This is responsible for creating any selection {@link Adornment} (if this Part {@link #isSelected}) and
   * any tool adornments for this part.
   * @expose
   * @this {Part}
   */
  public updateAdornments(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    let itr = this.adornments;
    while (itr.next()) {
      const adorn = itr.value;
      adorn.invalidateMeasure();
      if (adorn.placeholder) adorn.placeholder.invalidateMeasure();
    }
    this.updateSelectionAdornment(diagram);
    this.updateToolAdornments(diagram);
    // update all adornments
    itr = this.adornments;
    while (itr.next()) {
      const adorn = itr.value;
      adorn.updateTargetBindings();
    }
  }

  // Since 2.1
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Part}
   */
  public invalidateAdornments(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    diagram.invalidateViewportObjects();
    if (this.getInvalidAdornments() === true) return; // ? causes issues if this line is first
    this.setInvalidAdornments(true);
    diagram.requestUpdate();
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public validateAdornments(): void {
    if (this.getInvalidAdornments() === false) return;
    this.updateAdornments();
    this.setInvalidAdornments(false);
  }

  /**
   * @hidden @internal
   * This is responsible for creating or removing any selection {@link Adornment},
   * depending on whether {@link #isSelected}.
   * By default the adornment is a {@link Shape} that surrounds this part,
   * or that follows the {@link Link}'s route if this is a Link.
   * @this {Part}
   * @param {Diagram} diagram
   */
  private updateSelectionAdornment(diagram: Diagram): void {
    const category = 'Selection';
    if (this.isSelected && this.selectionAdorned) {
      const selelt = this.selectionObject;
      if (selelt !== null && this.actualBounds.isReal() && this.isVisible() &&
          selelt.isVisibleObject() && selelt.actualBounds.isReal()) {
        let adornment = this.findAdornment(category);
        if (adornment === null) {
          let template = this.selectionAdornmentTemplate;
          if (template === null) {
            if (this.isLinkLike()) {
              template = diagram.linkSelectionAdornmentTemplate;
            } else {
              if (this instanceof Group) {
                template = diagram.groupSelectionAdornmentTemplate;
              } else {
                template = diagram.nodeSelectionAdornmentTemplate;
              }
            }
          }
          if (!(template instanceof Adornment)) return;
          template.freezeBindings();
          adornment = template.copy();
          if (adornment !== null) {
            if (this.isLinkLike() && this.selectionObject === this.path) adornment.type = Panel.Link;
            adornment.adornedObject = selelt;
          }
        }
        if (adornment !== null) {
          // Links, including adornment links, set their location in their measure phase
          if (adornment.type === Panel.Link) {
            adornment.invalidateMeasure();
           } else {
            const temploc = Point.alloc();
            Point.free(temploc);
           }
          this.addAdornment(category, adornment);
          return;
        }
      }
    }
    this.removeAdornment(category);
  }

  /**
   * @hidden @internal
   * This is responsible for creating any tool {@link Adornment}s for this part.
   * This iterates over the mouse-down tools,
   * calling {@link Tool#updateAdornments} on each one.
   * @this {Part}
   * @param {Diagram} diagram
   */
  private updateToolAdornments(diagram: Diagram): void {
    const part = this;
    diagram.toolManager.mouseDownTools.each(function (t) {
      if (t.isEnabled) t.updateAdornments(part);
    });
    diagram.toolManager.updateAdornments(part);  // update tooltip
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Layer} value
   */
  public setLayer(value: Layer): void {
    this._layer = value;
  }

  /**
   * This read-only property returns the {@link Layer} that this Part is in.
   * The value is the Layer that is named with the value of {@link #layerName}.
   * If you want to change what Layer this Part is in, change the value of {@link #layerName} to refer to a different Layer.
   *
   * This will be null if it has not yet been added to a {@link Diagram},
   * or if it has already been removed from a Diagram.
   * @see {@link #layerName}
   */
  /** @export */ get layer(): Layer | null { return this._layer; }

  /**
   * This read-only property returns the {@link Diagram} that this Part is in.
   *
   * This will be null if it is not in a {@link Layer}.
   */
  /** @export */ get diagram(): Diagram | null {
    const layer = this._layer;
    if (layer !== null) return layer.diagram;
    return null;
  }

  /**
   * Gets or sets the layer name for this part.
   * The initial value is an empty string, which is the name of the default layer.
   * The value of this property determines the value of {@link #layer}.
   *
   * If this part is not yet in a {@link Diagram},
   * this value is used by {@link Diagram#add}
   * to determine which {@link Layer} this part should go in.
   * If no layer can be found with this name, it uses the default layer.
   *
   * Changing the value of this property while it is already in a layer
   * causes it to change layers if needed.
   * @see {@link #layerChanged}
   */
  /** @export */ get layerName(): string { return this._layerName; }
  /** @export */ set layerName(value: string) {
    const old = this._layerName;
    if (old === value) return;
    Util.checkPrimitive(value, 'string', Part, 'layerName');
    let diagram = this.diagram;
    // no-op if in Diagram and named layer does not exist;
    // also need to check addsToTemporaryLayer because if a selection binding changes the layerName during an ongoing drag,
    // it will pull it out of the tool layer causing unwanted bounds changes
    if (diagram !== null && (diagram.findLayer(value) === null || diagram.partManager.addsToTemporaryLayer)) return;
    this._layerName = value;
    if (diagram !== null) diagram.invalidateDocumentBounds();
    this.raiseChanged('layerName', old, value);

    // if already in a differently named layer, remove it,
    // find the right new layer, and add it
    // ??? no-op if isUndoingRedoing
    const oldlayer = this.layer;
    if (oldlayer === null || oldlayer.name === value) return;
    diagram = oldlayer.diagram;
    if (diagram === null) return;
    const newlayer = diagram.findLayer(value);
    if (newlayer === null || newlayer === oldlayer) return;
    let index = oldlayer.internalRemove(-1, this, true);
    // use newParam === true to indicate that it is changing layers, not being removed from the diagram
    if (index >= 0) diagram.raiseChangedEvent(ChangedEvent.Remove, 'parts', oldlayer, this, null, index, true);
    index = newlayer.internalAdd(99999999, this, true);
    // similar to the layer.visible setter, switching to in/visible layer
    if (oldlayer.visible !== newlayer.visible) this.madeVisible(newlayer.visible);
    // use oldParam === true to indicate that it is changing layers, not being added to the diagram
    if (index >= 0) diagram.raiseChangedEvent(ChangedEvent.Insert, 'parts', newlayer, null, this, true, index);
    const func = this.layerChanged;
    if (func === null) return;
    const oldUpdatingModel = diagram.isUpdatingModel;
    diagram.isUpdatingModel = true;
    func(this, oldlayer, newlayer);
    diagram.isUpdatingModel = oldUpdatingModel;
  }

  /**
   * Gets or sets the function to execute when this part changes layers.
   * It is typically used to modify the appearance of the part.
   * This function must not change the layer of this part by setting {@link #layerName}.
   *
   * If this property value is a function, it is called with three arguments,
   * this {@link Part}, the old {@link Layer} (may be null), and the new {@link Layer} (may be null).
   * By default this property is null -- no function is called.
   * @see {@link #layerName}
   */
  /** @export */ get layerChanged(): ((thisPart: Part, oldLayer: Layer | null, newLayer: Layer | null) => void) | null { return this._layerChanged; }
  /** @export */ set layerChanged(value: ((thisPart: Part, oldLayer: Layer | null, newLayer: Layer | null) => void) | null) {
    const old = this._layerChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Part, 'layerChanged');
      this._layerChanged = value;
      this.raiseChanged('layerChanged', old, value);
    }
  }

  /**
   * Gets or sets the Z-ordering position of this Part within its Layer.
   *
   * Within the same layer, nodes with larger zOrder values are placed in front of nodes with smaller zOrder values.
   * When the value is NaN the ordering is not specified.
   * The default value is NaN.
   *
   * When a {@link Group} has a zOrder value of NaN, it is automatically placed behind its member nodes and links
   * that also have no zOrder.
   * Such automatic ordering is not guaranteed if any nodes including the groups have a numeric zOrder.
   * If you do want to specify the zOrder of nodes, you should also specify the zOrder of their containing groups
   * unless those groups are in different layers.
   * @since 1.6
   */
  /** @export */ get zOrder(): number { return this._zOrder; }
  /** @export */ set zOrder(value: number) {
    const old = this._zOrder;
    if (old === value) return;
    Util.checkPrimitive(value, 'number', Part, 'zOrder');
    this._zOrder = value;
    const layer = this.layer;
    if (layer !== null) {
      layer.sortZOrderIndividual(-1, this);
    }
    this.raiseChanged('zOrder', old, value);
    // and make sure everything is re-drawn
    const diagram = this.diagram;
    if (diagram !== null) {
      diagram.invalidateDraw();
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  private sortZOrder(): void {
    const lay = this.layer;
    if (lay !== null) lay.sortZOrderIndividual(-1, this);
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  protected _invalidatePosition(): void {
    if (this.getInvalidPosition() !== false) return;
    this.setInvalidPosition(true);
    this._invalidateTransform(); // ?? is this required?
    const diagram = this.diagram;
    if (diagram !== null) {
      diagram.invalidatePart(this);
      // without this, location setter may not requestUpdate
      // position setter always updates because SetPos calls validatePosition after calling _invalidatePosition
      diagram.requestUpdate();
    }
  }

  /**
   * @hidden @internal
   * This is used in place of arrange to validate Parts when *only* their position/location has changed.
   * @this {Part}
   */
  protected validatePosition(): void {
    // once valid, assume location is most correct
    // do this before getInvalidPosition check, in case arrange validated the position last
    this._partFlags |= 0x00200000;  // _locSetLast = true

    if (this.getInvalidPosition() === false) return;

    const pos = this.position;
    const loc = this.location;
    if (!loc.isReal() || !pos.isReal()) {
      this.determinePosLoc(pos, loc);
    }

    const ab = this._actualBounds;
    const old = Rect.alloc().assign(ab);
    ab.thaw();
    ab.x = pos.x;
    ab.y = pos.y;
    ab.freeze();

    this.onBoundsChanged(old, ab);
    Rect.free(old);
    this.setInvalidPosition(false);
  }

  /**
   * This read-only property returns the {@link GraphObject} that determines the location of this Part.
   * The value will be in the visual tree of this Part and is usually named with
   * the value of {@link #locationObjectName}.
   */
  /** @export */ get locationObject(): GraphObject {
    if (this._locationObject === null) {
      const ename = this.locationObjectName;
      if (ename !== '') {
        const elt = this.findObject(ename);
        if (elt !== null) {  // found it: cache it
          this._locationObject = elt;
        } else {  // not yet found
          this._locationObject = this;
        }
      } else if (this instanceof Adornment) {
        if (this.type !== Panel.Link && this.placeholder !== null) {
          this._locationObject = this.placeholder;
        } else {
          this._locationObject = this;  // use main visual child instead
        }
      } else {  // always use main visual child
        this._locationObject = this;
      }
    }
    return this._locationObject.visible ? this._locationObject : this;
  }

  /**
   * Gets or sets the minimum location of this Part to which the user may drag using the {@link DraggingTool}.
   *
   * Value must be of type {@link Point}.
   * The initial value is (-Infinity, -Infinity), which imposes no position constraint.
   * A X value of NaN causes {@link Diagram#computeMove} to use the part's current location's X value as the minimum,
   * and similarly for NaN as the Y value.
   * @see {@link #location}
   * @see {@link #maxLocation}
   * @see {@link #dragComputation}
   * @see {@link #movable}
   */
  /** @export */ get minLocation(): Point { return this._minLocation; }
  /** @export */ set minLocation(value: Point) {
    const old = this._minLocation;
    if (old.equals(value)) return;
    if (Debug) Util.checkClass(value, Point, Part, 'minLocation');
    value = value.copyFrozen();
    this._minLocation = value;
    this.raiseChanged('minLocation', old, value);
  }

  /**
   * Gets or sets the maximum location of this Part to which the user may drag using the {@link DraggingTool}.
   *
   * Value must be of type {@link Point}.
   * The initial value is (Infinity, Infinity), which imposes no position constraint.
   * A X value of NaN causes {@link Diagram#computeMove} to use the part's current location's X value as the maximum,
   * and similarly for NaN as the Y value.
   * @see {@link #location}
   * @see {@link #minLocation}
   * @see {@link #dragComputation}
   * @see {@link #movable}
   */
  /** @export */ get maxLocation(): Point { return this._maxLocation; }
  /** @export */ set maxLocation(value: Point) {
    const old = this._maxLocation;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Point, Part, 'maxLocation');
      value = value.copyFrozen();
      this._maxLocation = value;
      this.raiseChanged('maxLocation', old, value);
    }
  }

  /**
   * Gets or sets the name of the {@link GraphObject} that provides the location of this Part.
   * This name determines the value of {@link #locationObject}.
   * The actual {@link #location} also depends on the {@link #locationSpot}.
   *
   * The initial value is an empty string, meaning the whole Part itself determines the location.
   * If you want to use a particular GraphObject in the visual tree of this Part,
   * set this property to be the {@link GraphObject#name} of the element that you want to be the {@link #locationObject}.
   * @see {@link #location}
   * @see {@link #locationSpot}
   */
  /** @export */ get locationObjectName(): string { return this._locationObjectName; }
  /** @export */ set locationObjectName(value: string) {
    const old = this._locationObjectName;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'locationObjectName');
      this._locationObjectName = value;
      this._locationObject = null;
      this.invalidateMeasure();
      this.raiseChanged('locationObjectName', old, value);
    }
  }

  /**
   * Gets or sets the location Spot of this Node, the spot on the {@link #locationObject}
   * that is used in positioning this part in the diagram.
   *
   * Value must be of the type {@link Spot}.
   * The initial value is `Spot.TopLeft`.
   * The value must be a specific spot -- i.e. one for which {@link Spot#isSpot} is true.
   *
   * It is commonplace to set this property to `Spot.Center`, so that the {@link #location}
   * has a value corresponding to the point at the center of this Part's {@link #locationObject} element.
   * But the {@link GraphObject#position} of a Part is always at the top-left corner
   * point of the {@link GraphObject#actualBounds}.
   * @see {@link #location}
   * @see {@link #locationObjectName}
   */
  /** @export */ get locationSpot(): Spot { return this._locationSpot; }
  /** @export */ set locationSpot(value: Spot) {
    const old = this._locationSpot;
    if (!old.equals(value)) {
      if (Debug) {
        Util.checkClass(value, Spot, Part, 'locationSpot');
        if (!value.isSpot()) Util.throwError('Part.locationSpot must be a specific Spot value, not: ' + value);
      }
      value = value.copyFrozen();
      this._locationSpot = value;
      this.invalidateMeasure();
      this.raiseChanged('locationSpot', old, value);
    }
  }

  /**
   * Move this part and any parts that are owned by this part to a new position.
   *
   * If this part is a {@link Group}, it also moves all of its members, recursively.
   * If this part is a {@link Link}, it also moves all of its label nodes.
   * @expose
   * @this {Part}
   * @param {Point} newpos a new {@link Point} in document coordinates.
   * @param {boolean=} useLocation true if you want to set the {@link #location} instead of the position. False by default.
   */
  public move(newpos: Point, useLocation?: boolean): void {
    if (useLocation === true) {
      this.location = newpos;
    } else { // false or undefined:
      this.position = newpos;
    }
  }

  /**
   * Move this part and any parts that are owned by this part to a new position.
   * This just calls {@link #move} without the caller having to allocate a new {@link Point}.
   * @export
   * @this {Part}
   * @param {number} newx a new X value in document coordinates.
   * @param {number} newy a new Y value in document coordinates.
   * @param {boolean=} useLocation true if you want to set the {@link #location} instead of the position. False by default.
   * @since 1.4
   */
  public moveTo(newx: number, newy: number, useLocation?: boolean): void {
    const p = Point.allocAt(newx, newy);
    this.move(p, useLocation);
    Point.free(p);
  }

  /**
   * This predicate is true if this Part can be seen.
   * Parts that can be seen can be manipulated by the user, can take space in the document, or can take part in a layout,
   * among many possibilities.
   * Note that the value of this predicate can often be false even while {@link GraphObject#visible} is true.
   *
   * A Part is not seen if it is not {@link GraphObject#visible} or if it is in a {@link Layer} that is not {@link Layer#visible}.
   *
   * If a Part is a member of a {@link Group} and the Group is not {@link Group#isSubGraphExpanded}, the part is not seen.
   * (The containing Group might still be visible.)
   *
   * If a {@link Node} is a "tree child" of a Node that is not {@link Node#isTreeExpanded}, the node is not seen.
   * (The parent Node might still be visible.)
   *
   * If a {@link Link} is connected to or from a {@link Node} that is not `isVisible()`
   * and is not a member of a {@link Group} that `isVisible()`, the link is not seen.
   *
   * If a {@link Node} is a "link label" of a {@link Link} and that Link is not `isVisible()`, the node is not seen.
   *
   * This is different from {@link GraphObject#isVisibleObject},
   * which ignores whether the {@link Layer} is visible and just checks {@link GraphObject#visible}
   * up the chain of containing {@link Panel}s.
   *
   * If you want to know whether a Part is in the Diagram's viewport, try:
   * ```js
   *   diagram.viewportBounds.containsRect(part.actualBounds);
   * ```
   *
   * or:
   * ```js
   *   diagram.viewportBounds.intersectsRect(part.actualBounds);
   * ```
   * @expose
   * @this {Part}
   * @return {boolean} true if {@link GraphObject#visible} is true and if {@link Layer#visible} is true.
   */
  public isVisible(): boolean {
    // GraphObject.visible must be true
    if (!this.visible) return false;
    // if there's a Layer, the layer had better be visible
    const layer = this.layer;
    if (layer !== null && !layer.visible) return false;

    // Objects that are animating are forced to be visible?
    const diagram = this.diagram;
    if (diagram !== null) {
      if (diagram.animationManager.objRemainsVisible(this)) return true;
    }

    // if this is inside a Group, it should not be seen if the group is collapsed or if the group is not seen
    const container = this.containingGroup;
    if (container !== null && (!container.isSubGraphExpanded || !container.isVisible())) return false;

    // overridden in Node and Link classes
    return true;
  }

  /**
   * @hidden @internal
   * Called when a part's visibility changes, from .visible or other indirect sources (like layer.visible)
   * Calling this with true does not guarantee that the object is visible. It may still be invisible for other reasons
   * Calling this with false does guarantee that the object is no longer visible.
   * @this {Part}
   * @param {boolean} value whether the part was just made visible (true) or just made invisible (false)
   */
  public madeVisible(value: boolean): void {
    const diagram = this.diagram;
    if (value) {
      this.invalidateLayout(Part.LayoutShown);  // no-op if isUndoingRedoing
      this.invalidateAdornments();  // don't call updateAdornments, to avoid creating offscreen Adornments
      // maybe add to invalid objects of diagram
      if (diagram !== null) diagram.invalidatePart(this);
    } else {
      this.invalidateLayout(Part.LayoutHidden);  // no-op if isUndoingRedoing
      this.clearAdornments();
    }
    this.invalidateContainingGroup();
    if (diagram !== null) {
      diagram.invalidateDocumentBounds();
      diagram.invalidateDraw(); // if still totally valid we still need a redraw (became invisible)
    }
  }

  /**
   * @hidden @internal
   * @export
   * @this {Part}
   * @param {string} name The name to search for, using a case-sensitive string comparison.
   * @return {GraphObject} If no such named object can be found, this returns null.
   */
  public findObject(name: string): GraphObject | null {
    if (this.name === name) return this;

    let map = this._nameMap;
    if (map === null) this._nameMap = map = new Map<string, GraphObject>();
    if (map.getValue(name) !== null) return map.getValue(name);

    const found = super.findObject(name);
    if (found !== null) {
      map.set(name, found);
      return found;
    }
    map.set(name, null);
    return null;
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public clearNameMap(): void {
    this._nameMap = null;
  }

  // * Gets a point on a GraphObject relative to this Part
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Part}
   * @param {GraphObject} elt
   * @param {Spot} spot
   * @param {Point=} result an optional Point that is modified and returned
   * @return {Point}
   */
  public getRelativePoint(elt: GraphObject, spot: Spot, result?: Point): Point {
    if (result === undefined) result = new Point();
    const s = spot.isNoSpot() ? Spot.Center : spot;
    const sz = elt.naturalBounds;
    result.assignTo(sz.width * s.x + s.offsetX, sz.height * s.y + s.offsetY);
    if (elt === null || elt === this) return result;

    elt.transform.transformPoint(result);
    let parent = elt.panel;
    while (parent !== null && parent !== this) {
      parent.transform.transformPoint(result);
      parent = parent.panel;
    }
    // definitely needed to get the correct offset
    this._localTransform.transformPoint(result);
    // local transform means we need this extra bit typically included in the normal transform
    result.offset(-this._measuredBounds.x, -this._measuredBounds.y);
    return result;
  }

  /**
   * Returns the Rect in document coordinates for this object's bounds.
   * If this GraphObject is a Part, the rect will be identical to its {@link #actualBounds}.
   * @export
   * @this {Part}
   * @param {Rect=} result an optional Rect that is modified and returned.
   * @return {Rect} in document coordinates.
   * @see {@link #getDocumentPoint}
   * @since 2.0
   */
  public getDocumentBounds(result?: Rect): Rect {
    if (result === undefined) result = new Rect();
    return result.assign(this.actualBounds);
  }

  /**
   * Measures if needed to make sure the {@link GraphObject#measuredBounds} and {@link GraphObject#naturalBounds} are all real numbers,
   * primarily to get the actual width and height.
   * {@link GraphObject#actualBounds} will get a real width and height, but the x and y values may continue to be `NaN`
   * if they were that way beforehand.
   *
   * This is sometimes necessary to call when defining custom layouts or implementing virtualization,
   * so that it can work with the actual size of the nodes.
   *
   * For efficiency, do not call this method unnecessarily.
   * @export
   * @this {Part}
   * @since 1.6
   */
  public ensureBounds(): void {
    if (this.getInvalidMeasure() === true) {
      if (this instanceof Group) {
        this.memberParts.each(function(p) {
          p.ensureBounds();
        });
      }
      this.measure(Infinity, Infinity);
    }
    this.arrange();
  }

  /**
   * @hidden @internal
   * Find the difference between position and location. This can only be called if the Part is arranged.
   * @param {Point} result a Point that is modified and returned
   * @this {Part}
   */
  protected determineOffset(result: Point): Point {
    const s = this.locationSpot;
    const obj = this.locationObject;
    if (s.isNoSpot()) Util.throwError('determineOffset: Part\'s locationSpot must be real: ' + s.toString());

    // if the object is a GraphObject and has a location object other than itself
    // then we need to find the appropriate spot location by transforming backwards to this part
    const ns = obj.naturalBounds;
    const sw2 = (obj instanceof Shape) ? obj.strokeWidth : 0.0;

    result.setSpot(0, 0, ns.width + sw2, ns.height + sw2, s); // scale is taken care of by localTransform below

    if (obj !== this) {
      result.offset(-sw2 / 2, -sw2 / 2);
      obj.transform.transformPoint(result);

      let parent = obj.panel;
      while (parent !== null && parent !== this) {
        parent.transform.transformPoint(result);
        parent = parent.panel;
      }
    }
    // definitely needed to get the correct offset
    this._localTransform.transformPoint(result);
    // local transform means we need this extra bit typically included in the normal transform
    result.offset(-this._measuredBounds.x, -this._measuredBounds.y);
    return result;
  }

  /**
   * @hidden @internal
   * Takes a rect (usually a part's actualBounds) and expands it according to the shadow size
   * @this {Part}
   * @param {Rect} rect the actual bounds of an object.
   * @param {Rect=} result optionally, modify this Rect and return it
   * @return {Rect} a rectangle in document coordinates.
   */
  public getDrawnSize(rect: Rect, result?: Rect): Rect {
    const r = (result === undefined) ? this._drawnRect : result;

    // miter amounts
    const m = this.getMaxMiter();
    const m2 = m * 2;

    if (!this.isShadowed) {
      r.assignTo(rect.x - 1 - m, rect.y - 1 - m, rect.width + 2 + m2, rect.height + 2 + m2);
      return r;
    }

    let x = rect.x;
    let y = rect.y;
    let w = rect.width;
    let h = rect.height;
    const blur = this.shadowBlur;
    const off = this.shadowOffset;
    w += blur;
    h += blur;
    x -= blur / 2;
    y -= blur / 2;
    if (off.x > 0) {
      w += off.x;
    } else {
      x += off.x;
      w -= off.x;
    }
    if (off.y > 0) {
      h += off.y;
    } else {
      y += off.y;
      h -= off.y;
    }
    r.assignTo(x - 1, y - 1, w + 2, h + 2);
    return r;
  }

  /**
   * @hidden @internal
   * Exists so that Shape does not need to import Adornment
   * @this {Part}
   * @return {number}
   */
  public getLogicalStrokeWidth(): number {
    if (this.type === Panel.Link && this instanceof Adornment) {
      // only if the adorned object is a shape, and only if its the selection adornment
      if (this.category === 'Selection' && this.adornedObject instanceof Shape && this.adornedPart.findMainElement() === this.adornedObject) {
        const shape = this.adornedObject;
        return shape.strokeWidth;
      }
    }
    return 0;
  }

  /**
   * @hidden @internal
   * The maxMiter is the maximum possible miter value of all Shapes in a panel
   * @this {Part}
   * @return {number}
   */
  public getMaxMiter(): number {
    if (isNaN(this._maxMiter)) {
      this._maxMiter = this.findMiter();
    }
    return this._maxMiter;
  }

  /**
   * @hidden @internal
   * Called when a shape changes within a panel (added, removed, strokeWidth, strokeMiterLimit changed)
   * @this {Part}
   */
  public resetMaxMiter(): void {
    this._maxMiter = NaN;
  }

  /**
   * @hidden @internal
   * Arrange a graphObject onto its diagram.
   * @this {Part}
   */
  public arrange(): void {
    if (this.getInvalidArrange() === false) {
      // This is uncommon but needed because cloned objects
      // can start with invalidArrange = false
      // Which means that location will not be set normally
      // (during an invalid arrage) and must be set here
      this.validatePosition();
      return;
    }

    const act = this._actualBounds;
    const oldActual = Rect.alloc();
    oldActual.assign(act);

    // Part 1: Determine bounds
    act.thaw();
    // don't modify inView if the bounds do not change
    const oldInView = this.getInView();
    this.arrangeProtected(0, 0, this._measuredBounds.width, this._measuredBounds.height);

    // Part 2: Determine position/location
    const pos = this.position;
    this.determinePosLoc(pos, this.location);

    // Part 3: set actualBounds x/y
    // This is not done sooner because the arrange step may set location/position
    // (As it does in the case of Link)
    // And because the locationObject must be arranged
    act.x = pos.x;
    act.y = pos.y;
    act.freeze();

    this.onBoundsChanged(oldActual, act);
    this.setInvalidArrange(false);
    if (oldActual.equals(act)) {
      this.setInView(oldInView);
    } else {
      // maybe invalidate Layout if Node or Part (but not Link or Adornment) changed size
      if (this.isNodeLike() &&
        (!Geo.isApprox(oldActual.width, act.width) || !Geo.isApprox(oldActual.height, act.height))) {
        if (this._layerIndex >= 0) {  // not if this is a newly added Node
          this.invalidateLayout(Part.LayoutNodeSized);  // no-op if isUndoingRedoing
        }
      }
    }
    Rect.free(oldActual);
    this.setInvalidPosition(false);
  }

  /**
   * @hidden @internal
   * Called at the end of GraphObject.arrange, Part.arrange, and Part.validatePosition (used in lieu of arrange)
   * @this {Part}
   * @param {Rect} oldb
   * @param {Rect} newb
   */
  protected onBoundsChanged(oldb: Rect, newb: Rect): void {
    // we should check to see if oldb equals newb
    // the rest can be skipped if that is true
    const diagram = this.diagram;
    if (diagram === null) return;

    let inside = false;
    // only do if the document bounds are valid.
    if (diagram._invalidDocBounds === false && oldb.isReal()) {
      const docb = Rect.alloc();
      docb.assign(diagram.documentBounds);
      docb.subtractMargin(diagram.padding as Margin);

      // are both the oldb and newb inside the docb by at least a pixel?
      if (oldb.x > docb.x && oldb.y > docb.y && oldb.right < docb.right && oldb.bottom < docb.bottom &&
          newb.x > docb.x && newb.y > docb.y && newb.right < docb.right && newb.bottom < docb.bottom) {
        inside = true;
      }
      Rect.free(docb);
    }

    if (this.getNotifyLater() === true || !oldb.equals(newb)) {
      this.notifyDiagram(inside, diagram);
    }

    diagram.invalidateDraw();
    if (oldb.equalsApproxClose(newb)) return;
    // if bounds really are different:

    if (this instanceof Node && !diagram.undoManager.isUndoingRedoing) {
      this.invalidateConnectedLinks(); // called when it validates and definitely changes size
    }

    this.invalidateContainingGroup();
  }

  /**
   * Gets or sets the position of this part in document coordinates,
   * based on the {@link #locationSpot} in this part's {@link #locationObject}.
   *
   * Value must be of type {@link Point}.
   * The initial value is Point(NaN, NaN).
   * It is commonplace to data bind this property to some property on your model node data.
   *
   * The value is related to the {@link GraphObject#position}.
   * For Parts, both are in document coordinates; setting one property will set the other property.
   * By default both will have the same value.
   * However, by setting either or both of {@link #locationSpot} and {@link #locationObjectName},
   * the location will be determined by a spot in the {@link #locationObject},
   * a {@link GraphObject} that is in the visual tree of this Part.
   * The {@link GraphObject#position} will always refer to the point at the top-left corner of the whole part.
   *
   * The {@link #minLocation} and {@link #maxLocation} limit the location of a part, not its position.
   * Grid snapping will normally locate the location to be on grid points.
   * @see {@link #locationObjectName}
   * @see {@link #locationSpot}
   * @see {@link #movable}
   * @see {@link #maxLocation}
   * @see {@link #minLocation}
   * @see {@link #dragComputation}
   */
  /** @export */ get location(): Point { return this._location; }
  /** @export */ set location(value: Point) {
    if (Debug) Util.checkClass(value, Point, Part, 'location');
    const nx = value.x;
    const ny = value.y;
    const old = this._location;
    const ox = old.x;
    const oy = old.y;
    // if old and new are not equal and if both old and new are not NaN
    if ((ox !== nx && (!isNaN(ox) || !isNaN(nx))) ||
        (oy !== ny && (!isNaN(oy) || !isNaN(ny)))) {
      value = value.copyFrozen();
      if (this.setLoc(value, old)) this.raiseChanged('location', old.copy(), value);
    }
  }

  /**
   * This is not called during animation, instead, it calls setPosInplace
   * @hidden @internal
   * @this {Part}
   * @param {Point} value
   * @param {Point} old
   * @return {boolean} true if successful, false if a no-op (only true calls raisechanged)
   */
  protected setLoc(value: Point, old: Point): boolean {
    if (this.isLinkLike()) return false;
    this._location = value;
    this._partFlags |= 0x00200000;  // _locSetLast = true
    if (this.getInvalidArrange() === false) {
      this._invalidatePosition(); // Instead of arrange
      const pos = this._position;
      if (pos.isReal()) {
        const oldpos = pos.copy();
        pos.assignTo(pos.x + (value.x - old.x), pos.y + (value.y - old.y));
        this.animatePosLoc(this.diagram, pos, oldpos);
        this.raiseChanged('position', oldpos, pos);
      }

    }
    return true;
  }

  /**
   * This is not called during animation, instead, it calls setPosInplace
   * @hidden @internal
   * @this {Part}
   * @param {Point} value
   * @param {Point} old
   * @return {boolean} true if successful, false if a no-op (only true calls raiseChanged)
   */
  protected setPos(value: Point, old: Point): boolean {
    if (this.isLinkLike()) return false;
    if (!value.isReal()) return false; // no-op if you try to set position to NaN on a Part (since it has location)
    const diagram = this.diagram;
    if (diagram !== null) {
      this.animatePosLoc(diagram, value, old);
      // undo/redo will sometimes set the location and then the position, which will offset the loc here unnessecarily
      if (diagram.undoManager.isUndoingRedoing === true) return true;
    }

    this._position = value;
    this._partFlags &= ~0x00200000; // _locSetLast = false
    const loc = this._location;
    if (loc.isReal()) {
      const oldloc = loc.copy();
      loc.assignTo(loc.x + (value.x - old.x), loc.y + (value.y - old.y));
      this.raiseChanged('location', oldloc, loc);
    }

    if (this.getInvalidPosition() === false && this.getInvalidArrange() === false) {
      this._invalidatePosition(); // Instead of arrange

      // This is needed for viewport optimizations
      // If the position is not validated right away,
      // it might get validated after the Diagram.position has changed, creating a mismatch
      this.validatePosition();
    }

    return true;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Diagram} diagram
   * @param {Point} value a position (not location)
   * @param {Point} old the old position
   */
  protected animatePosLoc(diagram: Diagram, value: Point, old: Point): void {
    if (diagram === null || (this instanceof Adornment)) return;
    const AM = diagram.animationManager;
    if (AM._pendingAutomaticAnimation) AM.addToAnimation(this, 'position', old.copy(), value.copy(), false);
  }

  /**
   * Used during animation only
   * @hidden @internal
   * @this {Part}
   * @param {number} x
   * @param {number} y
   */
  public setPosInplace(x: number, y: number, setloc: boolean): void {
    const loc = this._location;
    const pos = this._position;
    if (setloc) {
      if (loc.x === x && loc.y === y) return;
      if (this.getInvalidPosition() || this.getInvalidArrange()) {
        pos.assignTo(NaN, NaN);
      } else {
        pos.assignTo(pos.x + x - loc.x, pos.y + y - loc.y);
      }
      loc.assignTo(x, y);
    } else {
      if (pos.x === x && pos.y === y) return;
      if (this.getInvalidPosition() || this.getInvalidArrange()) {
        loc.assignTo(NaN, NaN);
      } else {
        loc.assignTo(loc.x + x - pos.x, loc.y + y - pos.y);
      }
      pos.assignTo(x, y);
    }
    this._invalidatePosition();
  }

  /**
   * @hidden @internal
   * Sets the flag for locLastSet, even if the position setter is a no-op
   * Even though its a no-op, we need to make sure the locLastSet ultimately can be set back to true,
   * so invalidatePosition is called even though the values may all be valid
   * @this {GraphObject}
   */
  protected setPositionFlag(): void {
    this._partFlags &= ~0x00200000; // _locSetLast = false
    this._invalidatePosition();
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Point} pos
   * @param {Point} loc
   */
  protected determinePosLoc(pos: Point, loc: Point): void {
    // if no position/location exist we can default to NaN
    const off = this.determineOffset(Point.alloc());
    const diagram = this.diagram;
    const realloc = loc.isReal();
    const realpos = pos.isReal();
    if (realloc && realpos) {
      // decide by whichever was set last
      if ((this._partFlags & 0x00200000) !== 0) { // if _locSetLast
        this.useLoc(pos, loc, diagram, off);
      } else {
        this.usePos(pos, loc, diagram, off);
      }
    } else if (realloc) {
      this.useLoc(pos, loc, diagram, off);
    } else if (realpos) {
      this.usePos(pos, loc, diagram, off);
    }
    this._partFlags |= 0x00200000;  // _locSetLast = true - once valid, assume location is most correct
    Point.free(off);
    this._invalidateTransform(); // ??2 is this required?
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Point} pos
   * @param {Point} loc
   * @param {Diagram} diagram
   * @param {Point} off
   */
  protected useLoc(pos: Point, loc: Point, diagram: Diagram, off: Point): void {
    const oldx = pos.x;
    const oldy = pos.y;
    // We have the offset, fix position
    pos.assignTo(loc.x - off.x, loc.y - off.y);
    if (diagram !== null) {
      const AM = diagram.animationManager;
      const animating = AM.isAnimating;
      if (!animating && AM._pendingAutomaticAnimation && !(this instanceof Adornment)) {
        AM.addToAnimation(this, 'position', new Point(oldx, oldy), pos, false);
      }
      if (!animating && !(pos.x === oldx && pos.y === oldy)) {
        // ??? This position change is not recorded in the undo manager because we do not want
        //     the position setter (during undo) to also (incorrectly) set the location.
        //     However we may need to record this in the future if the undo manager is to never need
        //     validation after undoing.
        const skips = diagram.skipsUndoManager;
        diagram.skipsUndoManager = true;
        this.raiseChanged('position', new Point(oldx, oldy), pos);
        diagram.skipsUndoManager = skips;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Point} pos
   * @param {Point} loc
   * @param {Diagram} diagram
   * @param {Point} off
   */
  protected usePos(pos: Point, loc: Point, diagram: Diagram, off: Point): void {
    const oldloc = loc.copy();
    loc.assignTo(pos.x + off.x, pos.y + off.y);
    if (!loc.equals(oldloc) && diagram !== null) {
      // ??? see comment on this.raiseChanged('position'... above
      const skips = diagram.skipsUndoManager;
      diagram.skipsUndoManager = true;
      this.raiseChanged('location', oldloc, loc);
      diagram.skipsUndoManager = skips;
    }
  }

  /**
   * @hidden @internal
   * After this is called you must invalidate the draw
   * @this {Part}
   * @param {boolean} inside whether or not the part is inside of the document bounds.
   * @param {Diagram} diagram
   */
  public notifyDiagram(inside: boolean, diagram: Diagram): void {
    this.setNotifyLater(false);

    const node = this; // for closure compiler
    if (node instanceof Node) {
      diagram.invalidatePositionArray(node);
    }

    if (!this.layer.isTemporary && !inside) diagram.invalidateDocumentBounds();

    const ab = this._actualBounds;
    const vpb = diagram.viewportBounds;
    if (vpb.isReal()) {
      if (this.getInView()) { // was it in before?
        // ??? 10 is an arbitrary amount to cover all adornments (used in layer too)
        if (!ab.intersectsRectPlus(vpb, 10)) { // no longer in the viewport
          this.setInView(false);
          this.updateAdornments();
        } else {
          this.updateAdornments();
        }
      } else { // see if the object is in viewport now
        if (ab.intersectsRect(vpb)) {
          this.setInView(true);
          this.updateAdornments();
        } else {
          this.invalidateAdornments();
        }
      }

    } else {
      diagram._invalidViewportObjects = true; // make sure that this is invalid
    }
  }

  /**
   * @hidden @internal
   * @virtual
   * @this {Part}
   * @return {boolean}
   */
  public isReadyToMeasureArrange(): boolean {
    return true;
  }

  /**
   * @hidden @internal
   * @virtual
   * @this {Part}
   * @return {boolean}
   */
  public isNodeLike(): boolean {
    return true;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @return {boolean}
   */
  public isLinkLike(): this is Link {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @return {boolean}
   */
  public isPart(): this is Part {
    return true;
  }

  /**
   * Gets or sets the category of this part,
   * typically used to distinguish different kinds of nodes or links.
   *
   * The initial value is an empty string, which is the default category.
   * Any new value must be a string.
   * This should not be set in templates.
   *
   * When building Parts for node data or link data in a model,
   * the {@link Diagram} will call {@link Model#getCategoryForNodeData} or
   * {@link GraphLinksModel#getCategoryForLinkData} to get the category string for the data object.
   * The diagram uses this value to look up a template in {@link Diagram#nodeTemplateMap},
   * {@link Diagram#linkTemplateMap} or {@link Diagram#groupTemplateMap}.
   * That template is copied to create the actual Part that is added to the diagram.
   * The diagram will set this property to remember the category it used.
   *
   * Note that the class of the new Part must be the same as the class of the original Part.
   * For example, a Node cannot be replaced by a simple Part or vice-versa.
   * Nor can a Link be replaced by a subclass of Link or vice-versa.
   *
   * To change the category for a Part created for model data, call {@link Model#setCategoryForNodeData} or
   * {@link GraphLinksModel#setCategoryForLinkData}.
   *
   * This property is also used to distinguish {@link Adornment}s on a Part.
   * In this scenario you create the Adornment, often indirectly by specifying a template,
   * and set this property explicitly.
   * For example, {@link ResizingTool#updateAdornments} creates a resizing Adornment from the
   * {@link Part#resizeAdornmentTemplate} and sets its category to be "Resizing".
   * Changing the category of an existing Adornment will update any adorned part's association.
   */
  /** @export */ get category(): string { return this._category; }
  /** @export */ set category(value: string) {
    const old = this._category;
    if (old === value) return;
    Util.checkPrimitive(value, 'string', Part, 'category');
    const diagram = this.diagram;
    const data = this.data;
    let oldclone: Part = null;
    if (diagram !== null && data !== null && !(this instanceof Adornment)) {
      const mgr = diagram.model.undoManager;
      if (mgr.isEnabled && !mgr.isUndoingRedoing) {
        oldclone = this.clone();
        oldclone._elements.addAll(this._elements);  // shallow copy of Part and its elements
      }
    }
    this._category = value;
    this.raiseChanged('category', old, value);
    if (diagram !== null && data !== null && !(this instanceof Adornment)) {
      const model = diagram.model;
      // don't have all these side-effects during undo or redo
      if (model.undoManager.isUndoingRedoing) return;
      if (this.isLinkLike()) {
        // update model
        diagram.partManager.setLinkCategoryForData(data, value);
        // replace visual tree of Link
        const template = diagram.partManager.findLinkTemplateForCategory(value);
        if (template !== null) {
          template.freezeBindings();
          const link = template.copy();
          if (link !== null) {
            this._displaceWith(link, old, value);
          }
        }
      } else {  // Node or simple Part
        // update model
        if (model !== null) {
          model.setCategoryForNodeData(data, value);
        }
        // replace visual tree of Node or Part
        const template = diagram.partManager.findTemplateForNodeData(data, value);
        if (template !== null) {
          template.freezeBindings();
          const part = (template.copy() as Part);
          if (part !== null && !(part instanceof Link)) {
            // keep the location the same, to avoid unexpected animation from origin after re-layout
            const curloc = this.location.copy();
            this._displaceWith(part, old, value);
            if (!this.location.isReal()) this.location = curloc;
          }
        }
      }

      if (oldclone !== null) {
        const newclone = this.clone();
        newclone._elements.addAll(this._elements);
        this.raiseChanged('self', oldclone, newclone);  // only used for undo/redo
      }
    } else if (this instanceof Adornment) { // this instanceof Adornment
      const p = this.adornedPart;
      if (p !== null) {
        const ads = p._adornments;
        // don't bother calling removeAdornment,
        // since we don't want it to be removed from the layer unnecessarily
        if (ads !== null) ads.remove(old);
        p.addAdornment(this.category, this);
      }
    }
  }

  /**
   * @hidden @internal
   * Just for internal use when displacing a Part due to a change in category, only for undo/redo.
   */
  /** @export */ get self() { return this; }
  /** @export */ set self(clone: Part) {
    this._displaceWith(clone, this.category, clone.category);
  }

  /** @hidden @internal */
  public static _warnedAboutDisplacingDifferentClass = false;

  /**
   * @hidden @internal
   * @virtual
   * @this {Part} the Part whose category is changing
   * @param {Part} part a Part copy of the new template specified by the new category.
   * @param {string} oldcat the old category.
   * @param {string} newcat the new category.
   */
  protected _displaceWith(part: Part, oldcat: string, newcat: string): void {
    if (part.constructor !== this.constructor) {
      if (!Part._warnedAboutDisplacingDifferentClass) {
        Part._warnedAboutDisplacingDifferentClass = true;
        Util.trace('Should not change the class of the Part when changing category from \"' + oldcat + '\" to \"' + newcat + '\"');
        Util.trace('  Old class: ' + Util.getClassName(this.constructor) + ', new class: ' + Util.getClassName(part.constructor) + ', part: ' + this.toString());
      }
    }

    this.clearAdornments();
    const olddata = this.data;
    const oldlayername = this.layerName;
    const oldisselected = this.isSelected;
    const oldishighlighted = this.isHighlighted;
    let oldistreeleaf = true;
    let oldistreeexpanded = true;
    let oldwastreeexpanded = false;
    if (this instanceof Node) {
      oldistreeleaf = this.isTreeLeaf;
      oldistreeexpanded = this.isTreeExpanded;
      oldwastreeexpanded = this.wasTreeExpanded;
    }

    // Fix up references from the (old) Part to any of its visual elements.
    // Note that this process does not recurse through the visual tree -- it's only a top-level displacement.
    part.displaceProtected(this);
    // Panel._elements list is copied before cloneProtected so that Group._placeholder can be initialized

    // this keeps valid all references to this Part that are in the Diagram/Layer or in other Parts
    part.cloneProtected(this);
    this._category = newcat;
    this.invalidateMeasure();
    this.invalidateDraw();

    const diagram = this.diagram;
    let oldskip = true;
    if (diagram !== null) {
      oldskip = diagram.skipsUndoManager;
      diagram.skipsUndoManager = true;
    }

    this._data = olddata;  // use same data
    // keep some "transient" properties, don't copy them from the new template:
    this._partFlags = oldisselected ? (this._partFlags | 0x00001000) : (this._partFlags & ~0x00001000);
    this._partFlags = oldishighlighted ? (this._partFlags | 0x00080000) : (this._partFlags & ~0x00080000);
    if (this instanceof Node) {
      this._nodeFlags = oldistreeleaf ? (this._nodeFlags | 0x00000004) : (this._nodeFlags & ~0x00000004);
      this._nodeFlags = oldistreeexpanded ? (this._nodeFlags | 0x00000001) : (this._nodeFlags & ~0x00000001);
      this._nodeFlags = oldwastreeexpanded ? (this._nodeFlags | 0x00000002) : (this._nodeFlags & ~0x00000002);
      // don't do this for Group properties such as isSubGraphExpanded
    }

    if (olddata !== null) this.updateTargetBindings();

    // change layers, if needed
    const newlayername = this.layerName;
    if (newlayername !== oldlayername) {
      this._layerName = oldlayername;
      this.layerName = newlayername;
    }

    if (diagram !== null) {
      diagram.skipsUndoManager = oldskip;
    }

    if (this.isNodeLike()) {
      this.invalidateLayout(Part.LayoutNodeReplaced);  // no-op if isUndoingRedoing
    }
  }

  // User permissions

  /**
   * This predicate returns true if {@link #copyable} is true,
   * if the layer's {@link Layer#allowCopy} is true, and
   * if the diagram's {@link Diagram#allowCopy} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may copy this part.
   */
  public canCopy(): boolean {
    if (!this.copyable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowCopy) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowCopy) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #deletable} is true,
   * if the layer's {@link Layer#allowDelete} is true, and
   * if the diagram's {@link Diagram#allowDelete} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may delete this part.
   */
  public canDelete(): boolean {
    if (!this.deletable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowDelete) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowDelete) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #textEditable} is true,
   * if the layer's {@link Layer#allowTextEdit} is true, and
   * if the diagram's {@link Diagram#allowTextEdit} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may edit this part.
   */
  public canEdit(): boolean {
    if (!this.textEditable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowTextEdit) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowTextEdit) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #groupable} is true,
   * if the layer's {@link Layer#allowGroup} is true, and
   * if the diagram's {@link Diagram#allowGroup} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may group this part.
   */
  public canGroup(): boolean {
    if (!this.groupable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowGroup) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowGroup) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #movable} is true,
   * if the layer's {@link Layer#allowMove} is true, and
   * if the diagram's {@link Diagram#allowMove} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may move this part.
   */
  public canMove(): boolean {
    if (!this.movable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowMove) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowMove) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #reshapable} is true,
   * if the layer's {@link Layer#allowReshape} is true, and
   * if the diagram's {@link Diagram#allowReshape} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may reshape this part.
   */
  public canReshape(): boolean {
    if (!this.reshapable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowReshape) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowReshape) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #resizable} is true,
   * if the layer's {@link Layer#allowResize} is true, and
   * if the diagram's {@link Diagram#allowResize} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may resize this part.
   */
  public canResize(): boolean {
    if (!this.resizable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowResize) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowResize) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #rotatable} is true,
   * if the layer's {@link Layer#allowRotate} is true, and
   * if the diagram's {@link Diagram#allowRotate} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may rotate this part.
   */
  public canRotate(): boolean {
    if (!this.rotatable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowRotate) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowRotate) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #selectable} is true,
   * if the layer's {@link Layer#allowSelect} is true, and
   * if the diagram's {@link Diagram#allowSelect} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean} true if the user may select this part.
   */
  public canSelect(): boolean {
    if (!this.selectable) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowSelect) return false;
    const diagram = layer.diagram;
    if (diagram === null) return true;
    if (!diagram.allowSelect) return false;
    return true;
  }

  // Permission properties

  /**
   * Gets or sets whether the user may copy this part.
   * The initial value is true.
   * @see {@link #canCopy}
   */
  /** @export */ get copyable(): boolean { return (this._partFlags & 0x00000001) !== 0; }
  /** @export */ set copyable(value: boolean) {
    const old = (this._partFlags & 0x00000001) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'copyable');
      this._partFlags = this._partFlags ^ 0x00000001;
      this.raiseChanged('copyable', old, value);
    }
  }

  /**
   * Gets or sets whether the user may delete this part.
   * The initial value is true.
   * @see {@link #canDelete}
   */
  /** @export */ get deletable(): boolean { return (this._partFlags & 0x00000002) !== 0; }
  /** @export */ set deletable(value: boolean) {
    const old = (this._partFlags & 0x00000002) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'deletable');
      this._partFlags = this._partFlags ^ 0x00000002;
      this.raiseChanged('deletable', old, value);
    }
  }

  /**
   * Gets or sets whether the user may do in-place text editing on {@link TextBlock}s in this part
   * that have {@link TextBlock#editable} set to true.
   * The initial value is true.
   * @see {@link #canEdit}
   * @see {@link TextBlock#editable}
   */
  /** @export */ get textEditable(): boolean { return (this._partFlags & 0x00000004) !== 0; }
  /** @export */ set textEditable(value: boolean) {
    const old = (this._partFlags & 0x00000004) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'textEditable');
      this._partFlags = this._partFlags ^ 0x00000004;
      this.raiseChanged('textEditable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether the user may group this part to be a member of a new {@link Group}.
   * The initial value is true.
   *
   * The grouping command is implemented by {@link CommandHandler#groupSelection}
   * and depends on {@link CommandHandler#archetypeGroupData} having been set to a node data object.
   * A Group can be ungrouped by the user if you set {@link Group#ungroupable} to true.
   * @see {@link #canGroup}
   */
  /** @export */ get groupable(): boolean { return (this._partFlags & 0x00000008) !== 0; }
  /** @export */ set groupable(value: boolean) {
    const old = (this._partFlags & 0x00000008) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'groupable');
      this._partFlags = this._partFlags ^ 0x00000008;
      this.raiseChanged('groupable', old, value);
    }
  }

  /**
   * Gets or sets whether the user may move this part.
   * The initial value is true.
   * @see {@link #canMove}
   * @see {@link #dragComputation}
   * @see {@link #maxLocation}
   * @see {@link #minLocation}
   */
  /** @export */ get movable(): boolean { return (this._partFlags & 0x00000010) !== 0; }
  /** @export */ set movable(value: boolean) {
    const old = (this._partFlags & 0x00000010) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'movable');
      this._partFlags = this._partFlags ^ 0x00000010;
      this.raiseChanged('movable', old, value);
    }
  }

  /**
   * Gets or sets whether a selection adornment is shown for this part when it is selected.
   * The initial value is true.
   * @see {@link #selectionChanged}
   * @see {@link #selectionAdornmentTemplate}
   */
  /** @export */ get selectionAdorned(): boolean { return (this._partFlags & 0x00000020) !== 0; }
  /** @export */ set selectionAdorned(value: boolean) {
    const old = (this._partFlags & 0x00000020) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'selectionAdorned');
      this._partFlags = this._partFlags ^ 0x00000020;
      this.raiseChanged('selectionAdorned', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether this Part is part of the document bounds.
   *
   * The initial value is true.
   * A value of false causes {@link Diagram#computeBounds} to ignore this part.
   * If the value is false, it is possible that user will not be able to scroll far enough to see this part,
   * if the part's {@link GraphObject#actualBounds} are outside of the {@link Diagram#documentBounds}.
   */
  /** @export */ get isInDocumentBounds(): boolean { return (this._partFlags & 0x00000040) !== 0; }
  /** @export */ set isInDocumentBounds(value: boolean) {
    const old = (this._partFlags & 0x00000040) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isInDocumentBounds');
      this._partFlags = this._partFlags ^ 0x00000040;
      const diagram = this.diagram;
      if (diagram !== null) diagram.invalidateDocumentBounds();
      this.raiseChanged('isInDocumentBounds', old, value);
    }
  }

  /**
   * Gets or sets whether a {@link Layout} positions this Node or routes this Link.
   * This property affects the value of {@link #canLayout}.
   *
   * The initial value is true, meaning that this part is laid out by the layout responsible for this Part.
   * If this part is a member of a {@link Group}, it is the {@link Group#layout},
   * otherwise it is the {@link Diagram#layout}.
   *
   * A value of false means that this part is not affected by and does not affect any automatic layout,
   * so the {@link #layoutConditions} property is ignored.
   * You will need to make sure that it has a real {@link #location} or {@link GraphObject#position} value,
   * or else the Part might not be visible anywhere in the diagram.
   *
   * Another way of controlling when layouts are invalidated is by setting
   * {@link Part#layoutConditions} or {@link Layout#isInitial} or {@link Layout#isOngoing}.
   */
  /** @export */ get isLayoutPositioned(): boolean { return (this._partFlags & 0x00000080) !== 0; }
  /** @export */ set isLayoutPositioned(value: boolean) {
    const old = (this._partFlags & 0x00000080) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isLayoutPositioned');
      this._partFlags = this._partFlags ^ 0x00000080;
      this.raiseChanged('isLayoutPositioned', old, value);
      this.invalidateLayout(value ? Part.LayoutShown : Part.LayoutHidden);  // no-op if isUndoingRedoing
    }
  }

  /**
   * Gets or sets whether the user may select this part.
   * The initial value is true.
   *
   * If you set this to true you may also want to set {@link #selectionObjectName}
   * to the {@link GraphObject#name}d element that you want to be adorned
   * when the Part is selected.
   * @see {@link #canSelect}
   * @see {@link #isSelected}
   * @see {@link #selectionAdorned}
   * @see {@link #selectionObjectName}
   * @see {@link #selectionAdornmentTemplate}
   * @see {@link #selectionChanged}
   */
  /** @export */ get selectable(): boolean { return (this._partFlags & 0x00000100) !== 0; }
  /** @export */ set selectable(value: boolean) {
    const old = (this._partFlags & 0x00000100) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'selectable');
      this._partFlags = this._partFlags ^ 0x00000100;
      this.raiseChanged('selectable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether the user may reshape this part.
   * The initial value is false.
   * @see {@link #canReshape}
   * @see {@link LinkReshapingTool}
   */
  /** @export */ get reshapable(): boolean { return (this._partFlags & 0x00000200) !== 0; }
  /** @export */ set reshapable(value: boolean) {
    const old = (this._partFlags & 0x00000200) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'reshapable');
      this._partFlags = this._partFlags ^ 0x00000200;
      this.raiseChanged('reshapable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether the user may resize this part.
   * The initial value is false.
   *
   * If you set this to true you may also want to set {@link #resizeObjectName}
   * to the {@link GraphObject#name}d element that you want the user to resize.
   * It is also commonplace to add a TwoWay {@link Binding} of that named element's
   * {@link GraphObject#desiredSize} in order to save to the model data the value that the user
   * set via the {@link ResizingTool}.
   * @see {@link #canResize}
   * @see {@link #resizeObjectName}
   * @see {@link #resizeCellSize}
   * @see {@link #resizeAdornmentTemplate}
   */
  /** @export */ get resizable(): boolean { return (this._partFlags & 0x00000400) !== 0; }
  /** @export */ set resizable(value: boolean) {
    const old = (this._partFlags & 0x00000400) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'resizable');
      this._partFlags = this._partFlags ^ 0x00000400;
      this.raiseChanged('resizable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether the user may rotate this part.
   * The initial value is false.
   *
   * If you set this to true you may also want to set {@link #rotateObjectName}
   * to the {@link GraphObject#name}d element that you want the user to rotate.
   * It is also commonplace to add a TwoWay {@link Binding} of that named element's
   * {@link GraphObject#angle} in order to save to the model data the value that the user
   * set via the {@link RotatingTool}.
   * @see {@link #canRotate}
   * @see {@link #rotateObjectName}
   * @see {@link #rotateAdornmentTemplate}
   */
  /** @export */ get rotatable(): boolean { return (this._partFlags & 0x00000800) !== 0; }
  /** @export */ set rotatable(value: boolean) {
    const old = (this._partFlags & 0x00000800) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'rotatable');
      this._partFlags = this._partFlags ^ 0x00000800;
      this.raiseChanged('rotatable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether this Part is selected.
   * The initial value is false.
   *
   * Selected parts typically are shown either with an {@link Adornment}
   * or with a different appearance by changing the brush or visibility of one or
   * more of the {@link GraphObject}s within the part.
   *
   * Changing this value does not by itself raise any "ChangingSelection" and "ChangedSelection" {@link DiagramEvent}s.
   * {@link Tool}s and the {@link CommandHandler} and methods such as {@link Diagram#select} do raise those DiagramEvents
   * because they want to surround changes to this property with a single "ChangingSelection" DiagramEvent beforehand and a single "ChangedSelection" afterwards.
   * @see {@link #selectable}
   * @see {@link #selectionAdorned}
   * @see {@link #selectionChanged}
   * @see {@link #selectionObjectName}
   * @see {@link #selectionAdornmentTemplate}
   */
  /** @export */ get isSelected(): boolean { return (this._partFlags & 0x00001000) !== 0; }
  /** @export */ set isSelected(value: boolean) {
    const old = (this._partFlags & 0x00001000) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isSelected');
      const diagram = this.diagram;
      // check to make sure it's OK to select
      if (value) {
        if (!this.canSelect()) return;
        if (diagram !== null) {
          const max = diagram.maxSelectionCount;
          const count = diagram.selection.count;
          if (count >= max) return;
        }
      }
      // OK: select it
      this._partFlags = this._partFlags ^ 0x00001000;
      let oldskip = false;
      if (diagram !== null) {
        oldskip = diagram.skipsUndoManager;
        diagram.skipsUndoManager = true;
        const coll = diagram.selection;
        coll.thaw();
        if (value) {
          coll.add(this);
        } else {
          coll.remove(this);
        }
        coll.freeze();
      }
      this.raiseChanged('isSelected', old, value);
      this.invalidateAdornments();
      const func = this.selectionChanged;
      if (func !== null) {
        func(this);
      }
      if (diagram !== null) {
        diagram.requestUpdate();
        diagram.skipsUndoManager = oldskip;
      }
    }
  }

  /**
   * Gets or sets whether this Part is highlighted.
   * The initial value is false.
   *
   * Highlighted parts may be shown with a different appearance by changing the brush or visibility of one or
   * more of the {@link GraphObject}s within the part.
   * One way of doing that is by using binding.
   * Consider part of the definition of a Node template:
   * ```js
   *   $(go.Shape,
   *     . . .,
   *     // Shape.fill is bound to Node.data.color
   *     new go.Binding("fill", "color"),
   *     // Shape.stroke is red when Node.isHighlighted is true, black otherwise
   *     new go.Binding("stroke", "isHighlighted",
   *                    function(h) { return h ? "red" : "black"; }).ofObject()),
   * ```
   * @see {@link #highlightedChanged}
   * @see {@link Diagram#highlighteds}
   * @see {@link Diagram#highlight}
   * @see {@link Diagram#highlightCollection}
   * @see {@link Diagram#clearHighlighteds}
   * @since 1.4
   */
  /** @export */ get isHighlighted(): boolean { return (this._partFlags & 0x00080000) !== 0; }
  /** @export */ set isHighlighted(value: boolean) {
    const old = (this._partFlags & 0x00080000) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isHighlighted');
      this._partFlags = this._partFlags ^ 0x00080000;
      const diagram = this.diagram;
      if (diagram !== null) {
        const coll = diagram.highlighteds;
        coll.thaw();
        if (value) {
          coll.add(this);
        } else {
          coll.remove(this);
        }
        coll.freeze();
      }
      this.raiseChanged('isHighlighted', old, value);
      this.invalidateDraw();
      const func = this.highlightedChanged;
      if (func !== null) {
        func(this);
      }
    }
  }

  /**
   * Gets or sets whether this part will draw shadows.
   * The initial value is false.
   *
   * By default, setting this property to true will attempt to draw shadows only
   * on the GraphObjects in this Part that appear to act as background objects, and
   * not on GraphObjects that appear to be in front of other GraphObjects in the Part.
   *
   * To finely control shadows, you may need to set {@link GraphObject#shadowVisible} on
   * elements of this Part, so that they explicitly do or do not get shadowed accordingly.
   * @see {@link #shadowOffset}
   * @see {@link #shadowColor}
   * @see {@link #shadowBlur}
   * @see {@link GraphObject#shadowVisible}
   */
  /** @export */ get isShadowed(): boolean { return (this._partFlags & 0x00002000) !== 0; }
  /** @export */ set isShadowed(value: boolean) {
    const old = (this._partFlags & 0x00002000) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isShadowed');
      this._partFlags = this._partFlags ^ 0x00002000;
      this.raiseChanged('isShadowed', old, value);
      this.invalidateDraw();
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public getInvalidAdornments(): boolean {
    return (this._partFlags & 0x00004000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {boolean} value
   */
  public setInvalidAdornments(value: boolean): void {
    if (value) {
      this._partFlags |= 0x00004000;
    } else {
      this._partFlags &= ~0x00004000;
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public getInvalidPosition(): boolean {
    return (this._partFlags & 0x00008000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {boolean} value
   */
  public setInvalidPosition(value: boolean): void {
    if (value) {
      this._partFlags |= 0x00008000;
    } else {
      this._partFlags &= ~0x00008000;
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public getNotifyLater(): boolean {
    return (this._partFlags & 0x00010000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {boolean} value
   */
  public setNotifyLater(value: boolean): void {
    if (value) {
      this._partFlags |= 0x00010000;
    } else {
      this._partFlags &= ~0x00010000;
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public getInView(): boolean {
    return (this._partFlags & 0x00020000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {boolean} value
   */
  public setInView(value: boolean): void {
    if (value) {
      this._partFlags |= 0x00020000;
    } else {
      this._partFlags &= ~0x00020000;
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public getIsCommonParent(): boolean {
    return (this._partFlags & 0x00100000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {boolean} value
   */
  public setIsCommonParent(value: boolean): void {
    if (value) {
      this._partFlags |= 0x00100000;
    } else {
      this._partFlags &= ~0x00100000;
    }
  }

  /**
   * Gets or sets whether this part may be animated.
   * The initial value is true.
   * @since 1.4
   */
  /** @export */ get isAnimated(): boolean { return (this._partFlags & 0x00040000) !== 0; }
  /** @export */ set isAnimated(value: boolean) {
    const old = (this._partFlags & 0x00040000) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Part, 'isAnimated');
      this._partFlags = this._partFlags ^ 0x00040000;
      this.raiseChanged('isAnimated', old, value);
    }
  }

  /**
   * Gets or sets the function to execute when this {@link #isHighlighted} changes.
   * It is typically used to modify the appearance of the part.
   * This function must not highlight or unhighlight any parts.
   *
   * If this property value is a function, it is called with one argument,
   * this {@link Part} that whose {@link #isHighlighted} value changed.
   * By default this property is null.
   * @see {@link #isHighlighted}
   * @since 1.7
   */
  /** @export */ get highlightedChanged(): ((thisPart: Part) => void) | null { return this._highlightedChanged; }
  /** @export */ set highlightedChanged(value: ((thisPart: Part) => void) | null) {
    const old = this._highlightedChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Part, 'highlightedChanged');
      this._highlightedChanged = value;
      this.raiseChanged('highlightedChanged', old, value);
    }
  }

  // Selecting

  /**
   * Gets or sets the name of the {@link GraphObject} that should get a selection handle
   * when this part is selected.
   * The value of this property affects the value of {@link #selectionObject}.
   * The initial value is an empty string, meaning the whole {@link Part} itself gets any selection handle.
   *
   * If no GraphObject has a {@link GraphObject#name} that is this name, {@link #selectionObject} returns the whole Part.
   * @see {@link #selectable}
   * @see {@link #selectionObject}
   */
  /** @export */ get selectionObjectName(): string { return this._selectionObjectName; }
  /** @export */ set selectionObjectName(value: string) {
    const old = this._selectionObjectName;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'selectionObjectName');
      this._selectionObjectName = value;
      this._selectionObject = null;  // clear any cached value
      this.raiseChanged('selectionObjectName', old, value);
    }
  }

  /**
   * Gets or sets the {@link Adornment} template used to create a selection handle for this Part.
   *
   * If this is null, depending on the class of this Part, the value of {@link Diagram#nodeSelectionAdornmentTemplate}, {@link Diagram#groupSelectionAdornmentTemplate},
   * or {@link Diagram#linkSelectionAdornmentTemplate} is used instead.
   *
   * It is commonplace to make use of a {@link Placeholder} in an {@link Adornment} for a {@link Node}, {@link Group}, or simple {@link Part}.
   * The Placeholder represents the {@link Adornment#adornedObject} of the adorned Part.
   * For {@link Link}s, the Adornment must be of {@link Panel#type} {@link Panel.Link}.
   * @see {@link #selectable}
   * @see {@link #selectionObjectName}
   */
  /** @export */ get selectionAdornmentTemplate(): Adornment | null { return this._selectionAdornmentTemplate; }
  /** @export */ set selectionAdornmentTemplate(value: Adornment | null) {
    const old = this._selectionAdornmentTemplate;
    if (old !== value) {
      if (Debug) Util.checkClass(value, Adornment, Part, 'selectionAdornmentTemplate');
      this._selectionAdornmentTemplate = value;
      this.raiseChanged('selectionAdornmentTemplate', old, value);
    }
  }

  /**
   * This read-only property returns the {@link GraphObject} that should get a selection handle when this part is selected.
   * The value will be in the visual tree of this Part and is usually named with
   * the value of {@link #selectionObjectName}.
   * When the {@link #selectionObjectName} is unspecified, this whole Part is used as the "selection object".
   * @see {@link #selectable}
   * @see {@link #selectionObjectName}
   */
  /** @export */ get selectionObject(): GraphObject {
    if (this._selectionObject === null) {
      const ename = this.selectionObjectName;
      if (ename !== null && ename !== '') {
        const elt = this.findObject(ename);
        if (elt !== null) {  // found it: cache it
          this._selectionObject = elt;
        } else {  // not yet found
          this._selectionObject = this;
        }
      } else {  // always use main visual child
        if (this instanceof Link) {  // but for Links use the path, if present
          const path = this.path;
          if (path !== null) {
            (this as Part)._selectionObject = path;
          } else {
            (this as Part)._selectionObject = this;
          }
        } else {
          this._selectionObject = this;
        }
      }
    }
    return this._selectionObject;
  }

  /**
   * Gets or sets the function to execute when this part is selected or deselected.
   * It is typically used to modify the appearance of the part.
   * This function must not select or deselect any parts.
   *
   * If this property value is a function, it is called with one argument,
   * this {@link Part} that was selected or that became unselected.
   * When it is called, the value of {@link Diagram#skipsUndoManager} is temporarily set to true.
   * By default this property is null.
   *
   * This function is called with {@link Diagram#skipsUndoManager} temporarily set to true,
   * so that any changes to {@link GraphObject}s are not recorded in the {@link UndoManager}.
   * You do not need to start and commit any transaction in this function.
   * @see {@link #isSelected}
   * @see {@link #selectionAdorned}
   * @see {@link #selectionAdornmentTemplate}
   */
  /** @export */ get selectionChanged(): ((thisPart: Part) => void) | null { return this._selectionChanged; }
  /** @export */ set selectionChanged(value: ((thisPart: Part) => void) | null) {
    const old = this._selectionChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Part, 'selectionChanged');
      this._selectionChanged = value;
      this.raiseChanged('selectionChanged', old, value);
    }
  }

  // Resizing

  /**
   * Gets or sets the adornment template used to create a resize handle {@link Adornment} for this part.
   * This is used by the {@link ResizingTool}, {@link ToolManager#resizingTool}.
   *
   * If an Adornment is supplied, it is normally a {@link Panel.Spot} panel that contains a {@link Placeholder}
   * with some number of resize handles at the four corners or at the four side midpoints.
   * @see {@link #resizable}
   * @see {@link #resizeObjectName}
   * @see {@link ResizingTool}
   */
  /** @export */ get resizeAdornmentTemplate(): Adornment | null { return this._resizeAdornmentTemplate; }
  /** @export */ set resizeAdornmentTemplate(value: Adornment | null) {
    const old = this._resizeAdornmentTemplate;
    if (old !== value) {
      if (Debug) Util.checkClass(value, Adornment, Part, 'resizeAdornmentTemplate');
      this._resizeAdornmentTemplate = value;
      this.raiseChanged('resizeAdornmentTemplate', old, value);
    }
  }

  /**
   * Gets or sets the name of the {@link GraphObject} that should get a resize handle
   * when this part is selected.
   * The value of this property affects the value of {@link #resizeObject}.
   * The initial value is an empty string, meaning the whole {@link Part} itself gets any resize handle.
   * @see {@link #resizable}
   * @see {@link #resizeObject}
   * @see {@link ResizingTool}
   */
  /** @export */ get resizeObjectName(): string { return this._resizeObjectName; }
  /** @export */ set resizeObjectName(value: string) {
    const old = this._resizeObjectName;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'resizeObjectName');
      this._resizeObjectName = value;
      this.raiseChanged('resizeObjectName', old, value);
    }
  }

  /**
   * This read-only property returns the {@link GraphObject} that should get resize handles when this part is selected.
   * The value will be in the visual tree of this Part and is usually named with
   * the value of {@link #resizeObjectName}.
   * @see {@link #resizable}
   * @see {@link #resizeObjectName}
   * @see {@link ResizingTool}
   */
  /** @export */ get resizeObject(): GraphObject {
    const ename = this.resizeObjectName;
    if (ename !== '') {
      const elt = this.findObject(ename);
      if (elt !== null) {  // found it
        return elt;
      }
    }
    // otherwise use the main visual
    return this;
  }

  /**
   * Gets or sets the width and height multiples used when resizing.
   * By default this property is the Size(NaN, NaN).
   * @see {@link #resizable}
   * @see {@link #resizeObjectName}
   * @see {@link ResizingTool}
   */
  /** @export */ get resizeCellSize(): Size { return this._resizeCellSize; }
  /** @export */ set resizeCellSize(value: Size) {
    const old = this._resizeCellSize;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Size, Part, 'resizeCellSize');
      value = value.copyFrozen();
      this._resizeCellSize = value;
      this.raiseChanged('resizeCellSize', old, value);
    }
  }

  // Rotating

  /**
   * Gets or sets the adornment template used to create a rotation handle {@link Adornment} for this part.
   * This is used by the {@link RotatingTool}, {@link ToolManager#rotatingTool}.
   *
   * This Adornment should not have a {@link Placeholder} in it, because the RotatingTool will position it away from the
   * {@link #rotateObject} at its {@link GraphObject#angle}.
   * @see {@link #rotatable}
   * @see {@link #rotateObjectName}
   * @see {@link RotatingTool}
   */
  /** @export */ get rotateAdornmentTemplate(): Adornment | null { return this._rotateAdornmentTemplate; }
  /** @export */ set rotateAdornmentTemplate(value: Adornment | null) {
    const old = this._rotateAdornmentTemplate;
    if (old !== value) {
      if (Debug) Util.checkClass(value, Adornment, Part, 'rotateAdornmentTemplate');
      this._rotateAdornmentTemplate = value;
      this.raiseChanged('rotateAdornmentTemplate', old, value);
    }
  }

  /**
   * Gets or sets the name of the {@link GraphObject} that should get a rotate handle
   * when this part is selected.
   * The value of this property affects the value of {@link #rotateObject}.
   * The initial value is an empty string, meaning the whole {@link Part} itself gets any rotate handle.
   * @see {@link #rotatable}
   * @see {@link #rotateObject}
   * @see {@link RotatingTool}
   */
  /** @export */ get rotateObjectName(): string { return this._rotateObjectName; }
  /** @export */ set rotateObjectName(value: string) {
    const old = this._rotateObjectName;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'rotateObjectName');
      this._rotateObjectName = value;
      this.raiseChanged('rotateObjectName', old, value);
    }
  }

  /**
   * This read-only property returns the {@link GraphObject} that should get rotate handles when this part is selected.
   * The value will be in the visual tree of this Part and is usually named with
   * the value of {@link #rotateObjectName}.
   * @see {@link #rotatable}
   * @see {@link #rotateObjectName}
   * @see {@link RotatingTool}
   */
  /** @export */ get rotateObject(): GraphObject {
    const ename = this.rotateObjectName;
    if (ename !== '') {
      const elt = this.findObject(ename);
      if (elt !== null) {  // found it
        return elt;
      }
    }
    // otherwise use main visual
    return this;
  }


  /**
   * Gets or sets the spot on the {@link #rotateObject}
   * that is used in rotating this part with the {@link RotatingTool}.
   *
   * Value must be of the type {@link Spot}.
   * The value must be a specific spot -- i.e. one for which {@link Spot#isSpot} is true, or else `Spot.Default`.
   *
   * If the value is `Spot.Default`, the {@link RotatingTool} uses
   * the locationSpot if the {@link #rotateObject} is equal to the {@link #locationObject}, otherwise
   * it uses `Spot.Center`.
   *
   * The initial value is `Spot.Default`.
   *
   * @since 2.0
   * @see {@link #rotateObjectName}
   */
  /** @export */ get rotationSpot(): Spot { return this._rotationSpot; }
  /** @export */ set rotationSpot(value: Spot) {
    const old = this._rotationSpot;
    if (!old.equals(value)) {
      if (Debug) {
        Util.checkClass(value, Spot, Part, 'rotationSpot');
        if (value !== Spot.Default && !value.isSpot()) Util.throwError('Part.rotationSpot must be a specific Spot value or Spot.Default, not: ' + value);
      }
      value = value.copyFrozen();
      this._rotationSpot = value;
      this.raiseChanged('rotationSpot', old, value);
    }
  }

  /**
   * Gets or sets a text string that is associated with this part.
   *
   * The initial value is an empty string.
   * This value is often used for sorting.
   */
  /** @export */ get text(): string { return this._text; }
  /** @export */ set text(value: string) {
    const old = this._text;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'text');
      this._text = value;
      this.raiseChanged('text', old, value);
    }
  }

  // Grouping

  /**
   * Gets or sets the {@link Group} of which this {@link Part} or {@link Node} is a member.
   * This will be null if this is a top-level part.
   *
   * You cannot set this property on a {@link Link}; it is set for you automatically
   * based on the group memberships of the connected nodes.
   * You cannot set this property on an {@link Adornment} at all.
   *
   * A template should not be a member of any group.
   * @see {@link Group#memberParts}
   * @see {@link #findTopLevelPart}
   * @see {@link #isTopLevel}
   */
  /** @export */ get containingGroup(): Group | null { return this._containingGroup; }
  /** @export */ set containingGroup(value: Group | null) {
    // don't allow this to be set on Links
    // (internally use setContainingGroup for Links, instead)
    if (this.isNodeLike()) {
      const old = this._containingGroup;
      if (old !== value) {
        if (Debug && value !== null) Util.checkClass(value, Group, Part, 'containingGroup');
        if (value !== null && (this as Part === value || value.isMemberOf(this))) {
          if (this as Part === value) Util.throwError('Cannot make a Group a member of itself: ' + this.toString());
          Util.throwError('Cannot make a Group indirectly contain itself: ' + this.toString() + ' already contains ' + value.toString());
        }

        // invalidate layout of old containing group or whole diagram
        this.invalidateLayout(Part.LayoutRemoved);  // no-op if isUndoingRedoing

        const diagram = this.diagram;
        if (old !== null) {
          old.removeMemberInternal(this);
        } else if (this instanceof Group) {
          if (diagram !== null) diagram._topLevelGroups.remove(this);
        }
        this._containingGroup = value;
        if (value !== null) {
          value.addMemberInternal(this);
        } else if (this instanceof Group) {
          if (diagram !== null) diagram._topLevelGroups.add(this);
        }

        // invalidate layout of new containing group or whole diagram
        this.invalidateLayout(Part.LayoutAdded);  // no-op if isUndoingRedoing

        // update model
        // ??? no-op if isUndoingRedoing
        if (diagram !== null) {
          const data = this.data;
          const model = diagram.model;
          if (data !== null && model.isGroupModel()) {
            // if VAL is null or if VAL.data is null, the node data's group key is set to undefined
            const gdata = (value !== null ? value.data : null);
            model.setGroupForData(data, model.getKeyForNodeData(gdata));
          }
        }

        const func = this.containingGroupChanged;
        if (func !== null) {
          let oldUpdatingModel = true;
          if (diagram !== null) {
            oldUpdatingModel = diagram.isUpdatingModel;
            diagram.isUpdatingModel = true;
          }
          func(this, old, value);
          if (diagram !== null) {
            diagram.isUpdatingModel = oldUpdatingModel;
          }
        }

        if (this instanceof Group) {
          // connected links need to have their containingGroup recalculated,
          // including links connected to member nodes and nested groups
          const coll = new Set<Part>();
          Part.gatherCollection(coll, this, true, 0, true);
          const it = coll.iterator;
          while (it.next()) {
            const p = it.value;
            if (p instanceof Node) {
              const lit = p.linksConnected;
              while (lit.next()) {
                const l = lit.value;
                // ?? can this be optimized not to bother with Links that are in COLL
                l.updateContainingGroup();
              }
            }
          }
        }
        if (this instanceof Node) {
          const lit = this.linksConnected;
          while (lit.next()) {
            const l2 = lit.value;
            l2.updateContainingGroup();
          }
          const ll = this.labeledLink;
          if (ll !== null) ll.updateContainingGroup();
        }

        this.raiseChanged('containingGroup', old, value);

        if (value !== null) value.sortZOrder();
      }
    } else {
      Util.throwError('cannot set the Part.containingGroup of a Link or Adornment');
      return;
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public invalidateContainingGroup(): void {
    const group = this.containingGroup;
    if (group !== null) {
      group.invalidateMeasure();
      if (group.placeholder !== null) {
        group.placeholder.invalidateMeasure();
      }
      group.invalidateConnectedLinks();
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public invalidateDraw(): void {
    const d = this.diagram;
    if (d !== null && !this.getInvalidArrange() &&
      !this.getInvalidPosition() && this.isVisible() &&
      this._actualBounds.isReal()) { // only if bounds are meaningful
      d.invalidateDraw(this.getDrawnSize(this._actualBounds));
    }
  }

  /**
   * @hidden @internal
   * @this {Part}
   */
  public invalidateMeasure(): void {
    super.invalidateMeasure();
    const diagram = this.diagram;
    if (diagram === null) return;
    diagram.invalidatePart(this);
    if (this instanceof Node) {
      if (this.labeledLink !== null) this.labeledLink.invalidateArrange();
    }
    // make sure an update is queued even if called during an update
    diagram.requestUpdate(true);
  }

  /**
   * @hidden @internal
   * Called by Layer.internalAdd, overridden on Node and Group and Link.
   * @this {Part}
   * @param {boolean} switchingLayers
   */
  public internalAddProtected(switchingLayers: boolean): void {
    if (!switchingLayers) {
      const old = this._containingGroup;
      if (old !== null) old.addMemberInternal(this);
    }
  }

  /**
   * @hidden @internal
   * Called by Layer.internalRemove, overridden on Node and Group and Link.
   * @this {Part}
   * @param {boolean} switchingLayers
   */
  public internalRemoveProtected(switchingLayers: boolean): void {
    if (!switchingLayers) {
      // do the following instead of: this.containingGroup = null;
      const old = this._containingGroup;
      if (old !== null) old.removeMemberInternal(this);
    }
  }

  /**
   * @hidden @internal
   * Overridden by Node and Group and Link.
   * @virtual
   * @this {Part}
   */
  public deleteCascade(): void {
    // this is not called for Links!
    const data = this.data;
    if (data !== null) {
      const diagram = this.diagram;
      if (diagram !== null) {
        const model = diagram.model;
        if (model !== null) {
          model.removeNodeData(data);
        }
      }
    }
  }

  /**
   * Gets or sets the function that is called after this Part has changed which {@link Group} it belongs to, if any.
   * It is typically used to modify the appearance of the part.
   * The first argument will be this Part.
   * The second argument will be the old Group, or null if it had been a top-level part.
   * The third argument will be the new Group, or null if it is now a top-level part.
   *
   * If the value is a function, that function must not modify the part's containing {@link Group}.
   * The containing Group has already been changed -- trying to change it again may produce undefined behavior.
   *
   * The initial value is null -- no function is called.
   */
  /** @export */ get containingGroupChanged(): ((thisPart: Part, oldGroup: Group | null, newGroup: Group | null) => void) | null { return this._containingGroupChanged; }
  /** @export */ set containingGroupChanged(value: ((thisPart: Part, oldGroup: Group | null, newGroup: Group | null) => void) | null) {
    const old = this._containingGroupChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Part, 'containingGroupChanged');
      this._containingGroupChanged = value;
      this.raiseChanged('containingGroupChanged', old, value);
    }
  }

  /**
   * Return how deep this part is in the hierarchy of nested {@link Group}s.
   * For parts that have no {@link #containingGroup} this returns zero.
   *
   * If you want to know how deep a Node is in a tree structure, call {@link Node#findTreeLevel}.
   * @export
   * @this {Part}
   * @return {number}
   * @since 1.5
   */
  public findSubGraphLevel(): number {
    return this._findSubGraphLevel(this);
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Part} p
   * @return {number}
   */
  protected _findSubGraphLevel(p: Part): number {
    const sg = p.containingGroup;
    if (sg !== null) return 1 + this._findSubGraphLevel(sg);
    if (p instanceof Node) {
      const ll = p.labeledLink;
      if (ll !== null) return this._findSubGraphLevel(ll); // don't add 1, only consider Groups when incrementing level
    }
    return 0;
  }

  /**
   * Gets the top-level Part for this part, which is itself when {@link #isTopLevel} is true.
   * If this Part is a member of a {@link Group}, this returns the top-level Part for that Group.
   * If this is a {@link Node} that is a label node for a labeled {@link Link}, this returns the top-level Part for that Link.
   *
   * If this is a Node and you are searching for the root of the tree that this node is in, use {@link Node#findTreeRoot}.
   * @export
   * @this {Part}
   * @return {Part} This will not return null.
   * @since 1.1
   */
  public findTopLevelPart(): Part {
    return this._findTopLevelPart(this);
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Part} p
   * @return {Part}
   */
  protected _findTopLevelPart(p: Part): Part {
    const sg = p.containingGroup;
    if (sg !== null) return this._findTopLevelPart(sg);
    if (p instanceof Node) {
      const ll = p.labeledLink;
      if (ll !== null) return this._findTopLevelPart(ll);
    }
    return p;
  }

  /**
   * This read-only property is true when this part is not member of any {@link Group} node
   * nor is it a label node for a {@link Link}.
   * @see {@link #containingGroup}
   * @see {@link Node#labeledLink}
   */
  /** @export */ get isTopLevel(): boolean {
    if (this.containingGroup !== null) return false;
    if (this instanceof Node) {
      if (this.labeledLink !== null) return false;
    }
    return true;
  }

  /**
   * This predicate is true if this part is a member of the given {@link Part}, perhaps indirectly.
   *
   * If the given part is a {@link Group} and this part is a member of the given group, this returns true.
   * If this part is a {@link Node} and it is a label node for the given link, this returns true.
   * Otherwise this searches recursively any {@link Part#containingGroup} of the given part.
   *
   * A part cannot be contained by itself.
   * A template should not be a member of any group.
   *
   * If this is a Node and you want to find whether it is in a subtree whose root is a given Node, use {@link Node#isInTreeOf}.
   * @export
   * @this {Part}
   * @param {Part} part
   * @return {boolean}
   */
  public isMemberOf(part: Part): boolean {
    if (!(part instanceof Group)) return false;
    return this._isMemberOf(this, part);
  }

  /**
   * @hidden @internal
   * @this {Part}
   * @param {Part} p
   * @param {Part} part
   * @return {boolean}
   */
  protected _isMemberOf(p: Part, part: Part): boolean {
    if (p === part || part === null) return false;
    const g = p.containingGroup;
    if (g !== null && (g === part || this._isMemberOf(g, part))) return true;
    if (p instanceof Node) {
      const ll = p.labeledLink;
      if (ll !== null) return this._isMemberOf(ll, part);
    }
    return false;
  }

  /**
   * Find the {@link Group} that perhaps indirectly contains both this part and another one.
   * If this is a Group and it contains the OTHER Part, return this.
   * If the OTHER Part is a Group and it contains this Part, return that OTHER Part.
   *
   * This returns null if the two parts are unrelated in the hierarchy of part membership.
   * If non-null, the result is a {@link Group}.
   *
   * If you want to find the {@link Node} that is the tree parent of two Nodes,
   * call {@link Node#findCommonTreeParent}.
   * @export
   * @this {Part}
   * @param {Part} other
   * @return {Group} may be null
   */
  public findCommonContainingGroup(other: Part): Group | null {
    if (other === null) return null;
    if (Debug) Util.checkClass(other, Part, Part, 'findCommonContainingGroup:other');
    if (this === other) return this.containingGroup;

    // mark this part and all of its containing groups
    let p = this as Part;
    while (p !== null) {
      if (p instanceof Group) p.setIsCommonParent(true);
      if (p instanceof Node) {
        const ll = p.labeledLink;
        if (ll !== null) p = ll;
      }
      p = p.containingGroup;
    }
    // starting at OTHER walk up all its containing groups, seeing if any are already marked
    let found: Group = null;
    p = other;
    while (p !== null) {
      if (p.getIsCommonParent()) {
        found = p as Group;
        break;
      }
      if (p instanceof Node) {
        const ll = p.labeledLink;
        if (ll !== null) p = ll;
      }
      p = p.containingGroup;
    }
    // clear this part and all of its containing groups
    p = this;
    while (p !== null) {
      if (p instanceof Group) p.setIsCommonParent(false);
      if (p instanceof Node) {
        const ll = p.labeledLink;
        if (ll !== null) p = ll;
      }
      p = p.containingGroup;
    }
    return found;
  }

  // Layout

  /**
   * Gets or sets flags that control when the {@link Layout} that is responsible for this Part is invalidated.
   * The initial value is {@link Part.LayoutStandard},
   * which causes the layout for this part to be invalidated when the part is added or removed or changes visibility or size.
   *
   * Individual layout conditions include: {@link Part.LayoutAdded}, {@link Part.LayoutRemoved},
   * {@link Part.LayoutShown}, {@link Part.LayoutHidden}, and {@link Part.LayoutNodeSized}.
   *
   * This property is ignored when {@link #isLayoutPositioned} is false -- no operation on this Part
   * will by itself cause the responsible Layout to be invalidated.
   *
   * You can also control when layouts are invalidated is by setting {@link Layout#isInitial} or {@link Layout#isOngoing}.
   */
  /** @export */ get layoutConditions(): number { return this._layoutConditions; }
  /** @export */ set layoutConditions(value: number) {
    const old = this._layoutConditions;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Part, 'layoutConditions');
      this._layoutConditions = value;
      this.raiseChanged('layoutConditions', old, value);
    }
  }

  /**
   * This predicate is called by Layout implementations to decide whether this Part
   * should be positioned and might affect the positioning of other Parts.
   *
   * This is false if {@link #isLayoutPositioned} is false, if {@link #isVisible} returns false,
   * or if the part is in a temporary {@link Layer}.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Part}
   * @return {boolean}
   */
  public canLayout(): boolean {
    if (!this.isLayoutPositioned) return false;
    if (!this.isVisible()) return false;
    // also consider whether this.layer.isTemporary
    const lay = this.layer;
    if (lay !== null && lay.isTemporary) return false;
    // link labels are always laid out by their Link, not by the Diagram or Group
    if (this instanceof Node) {
      if (this.isLinkLabel) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * This predicate is called by {@link #invalidateLayout} with a layout condition in order to decide whether
   * to invalidate the Layout responsible for this Part after a change identified by the condition argument.
   *
   * If the condition argument flag is not set in the value of {@link #layoutConditions}, this returns false.
   * In particular, a Part that becomes not visible may still cause a layout invalidation
   * if {@link #layoutConditions} includes {@link Part.LayoutHidden}, even though that Part
   * will not participate in the layout.
   *
   * This predicate is also false if {@link #isLayoutPositioned} is false
   * or if the part is in a temporary {@link Layer}.
   * @this {Part}
   * @param {number} condition the reason that the layout should be invalidated,
   *   a value such as {@link Part.LayoutAdded} or {@link Part.LayoutShown}.
   * @return {boolean}
   */
  public canInvalidateLayout(condition: number): boolean {
    if (!this.isLayoutPositioned) return false;
    // when invalidating, consider the layoutConditions
    if ((condition & this.layoutConditions) === 0) return false;
    // also consider whether this.layer.isTemporary
    const lay = this.layer;
    if (lay !== null && lay.isTemporary) return false;
    // link labels are always laid out by their Link, not by the Diagram or Group
    if (this instanceof Node) {
      if (this.isLinkLabel) return false;
    }
    const diagram = this.diagram;
    if (diagram !== null && diagram.undoManager.isUndoingRedoing) return false;
    return true;
  }

  /**
   * Invalidate the {@link Layout} that is responsible for positioning this Part.
   * If this part is in a {@link Group}, invalidate its {@link Group#layout}, if it has one.
   * Otherwise invalidate the {@link Diagram#layout}.
   *
   * But note that if {@link #isLayoutPositioned} is false, or if it is in a temporary {@link Layer},
   * or if it is not in a diagram or group, no layout is invalidated.
   * @export
   * @this {Part}
   * @param {number=} condition the reason that the layout should be invalidated;
   * if this argument is not supplied, any value of {@link #layoutConditions} other than {@link Part.LayoutNone}
   * will allow the layout to be invalidated.
   */
  public invalidateLayout(condition?: number): void {
    if (condition === undefined) condition = Part.LayoutAll;
    if (!this.canInvalidateLayout(condition)) return;
    const g = this._containingGroup;
    if (g !== null) {
      const lay = g.layout;
      if (lay !== null) {
        lay.invalidateLayout();
      } else {
        g.invalidateLayout(condition);  // go up chain of Groups
      }
    } else {
      const d = this.diagram;
      if (d !== null) {
        const lay = d.layout;
        if (lay !== null) {
          lay.invalidateLayout();
        }
      }
    }
  }

  /**
   * This predicate is called by Diagram's doucmentBounds computation and its internalAdd/Remove methods
   * @hidden @internal
   * @this {Part}
   * @return {boolean}
   */
  public canIncludeBounds(): boolean {
    if (!this.isVisible()) return false;
    const lay = this.layer;
    if (lay !== null && lay.isTemporary) return false;
    return true;
  }

  /**
   * @hidden @internal
   * Static function for collecting the member Parts of a Group,
   * or the tree children of a Node, or maybe even both.
   * This excludes any Adornments that might be encountered.
   * This will include the initial part P in the result collection,
   * unless the predicate is false for that part.
   * @param {Set.<Part>} coll the {@link Set} that is modified to hold the results.
   * @param {Part} p the {Part} from which to start the traversal.
   * @param {boolean} groupmembers whether to include member parts.
   * @param {number} treechildren int the number of levels of tree children to include (1 or less means just the node itself, no children).
   * @param {boolean | null} connectedlinks whether to include links connected with nodes in the collection.
   * @param {function(Part):boolean | null=} pred a predicate to execute on each part found,
   * if the predicate returns false, the part is not included in the result.
   */
  public static gatherCollection(coll: Set<Part>, p: Part, groupmembers: boolean, treechildren: number,
                                 connectedlinks: boolean | null, pred?: ((a: Part) => boolean) | null): void {
    if (pred === undefined) pred = null;
    if (coll.contains(p)) return;  // already seen
    if (pred !== null) {
      if (!pred(p)) return;  // filtered out
    }
    if (p instanceof Adornment) return;  // exclude Adornments
    coll.add(p);
    if (p instanceof Node) { // dont need to check for Part since
      if (groupmembers && p instanceof Group) {
        const mit = p.memberParts;
        while (mit.next()) {
          const m = mit.value;
          Part.gatherCollection(coll, m, groupmembers, treechildren, connectedlinks, pred);
        }
      }
      if (connectedlinks !== false) {  // true requires both connected nodes to be in collection; null only requires one
        const lit = p.linksConnected;
        while (lit.next()) {
          const l = lit.value;
          if (coll.contains(l)) continue;
          const from = l.fromNode;
          const to = l.toNode;
          const fromIncluded = from === null || coll.contains(from);
          const toIncluded = to === null || coll.contains(to);
          if (connectedlinks ? (fromIncluded && toIncluded) : (fromIncluded || toIncluded)) {
            Part.gatherCollection(coll, l, groupmembers, treechildren, connectedlinks, pred);
          }
        }
      }
      if (treechildren > 1) {
        const it = p.findTreeChildrenNodes();
        while (it.next()) {
          const c = it.value;
          Part.gatherCollection(coll, c, groupmembers, treechildren - 1, connectedlinks, pred);
        }
      }
    } else if (p instanceof Link) {
      const lit2 = p.labelNodes;
      while (lit2.next()) {
        const lab = lit2.value;
        Part.gatherCollection(coll, lab, groupmembers, treechildren, connectedlinks, pred);
      }
    }
    // include plain Parts, without additional traversal
  }

  /**
   * Gets or sets the function used to determine the location that this Part can be dragged to.
   * The first argument is a reference to the Part being dragged, the second argument is a Point describing
   * the proposed location, and the third argument is a snapped location, if one was determined during dragging.
   * It should return a Point that is the proposed new location.
   *
   * By default this function is null and the {@link DraggingTool} uses the snapped location,
   * if one was determined and if {@link DraggingTool#isGridSnapEnabled} is true,
   * or the proposed location (the second argument) if not snapping to a grid.
   *
   * In either case the {@link DraggingTool} will limit the proposed new location by {@link #minLocation} and {@link #maxLocation}.
   *
   * The function, if supplied, must not have any side-effects.
   *
   * An example that limits moving a Node to the current viewport:
   * ```js
   *   function stayInViewport(part, pt, gridpt) {
   *     var diagram = part.diagram;
   *     if (diagram === null) return pt;
   *     // compute the area inside the viewport
   *     var v = diagram.viewportBounds.copy();
   *     v.subtractMargin(diagram.padding);
   *     // get the bounds of the part being dragged
   *     var b = part.actualBounds;
   *     var loc = part.location;
   *     // now limit the location appropriately
   *     var x = Math.max(v.x+1, Math.min(pt.x, v.right-b.width-2)) + (loc.x-b.x);
   *     var y = Math.max(v.y+1, Math.min(pt.y, v.bottom-b.height-2)) + (loc.y-b.y);
   *     return new go.Point(x, y);
   *   }
   * ```
   * Note that for this functionality you will also probably want to set {@link Diagram#autoScrollRegion} to be a zero margin.
   * ```js
   *   myDiagram.nodeTemplate =
   *     $(go.Node, . . .,
   *       { dragComputation: stayInViewport },
   *       . . .
   *     );
   * ```
   * @see {@link #maxLocation}
   * @see {@link #minLocation}
   */
  /** @export */ get dragComputation(): ((thisPart: Part, newLoc: Point, snappedLoc: Point) => Point) | null { return this._dragComputation; }
  /** @export */ set dragComputation(value: ((thisPart: Part, newLoc: Point, snappedLoc: Point) => Point) | null) {
    const old = this._dragComputation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Part, 'dragComputation');
      this._dragComputation = value;
      this.raiseChanged('dragComputation', old, value);
    }
  }

  // Shadows

  /**
   * Gets or sets the X and Y offset of this part's shadow. This is only relevant if {@link #isShadowed} is true.
   * The initial value is (6, 6).
   * @see {@link #isShadowed}
   * @see {@link #shadowColor}
   * @see {@link #shadowBlur}
   */
  /** @export */ get shadowOffset(): Point { return this._shadowOffset; }
  /** @export */ set shadowOffset(value: Point) {
    const old = this._shadowOffset;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Point, Part, 'shadowOffset');
      value = value.copyFrozen();
      this._shadowOffset = value;
      this.invalidateDraw();
      this.raiseChanged('shadowOffset', old, value);
    }
  }

  /**
   * Gets or sets the CSS string that describes a shadow color. Default is 'gray'.
   * Brushes cannot be used for this property -- only strings.
   * @see {@link #isShadowed}
   * @see {@link #shadowOffset}
   * @see {@link #shadowBlur}
   */
  /** @export */ get shadowColor(): string { return this._shadowColor; }
  /** @export */ set shadowColor(value: string) {
    const old = this._shadowColor;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Part, 'shadowColor');
      this._shadowColor = value;
      this.invalidateDraw();
      this.raiseChanged('shadowColor', old, value);
    }
  }

  /**
   * Gets or sets the numerical value that describes the shadow's blur. Number must be a non-negative non-infinity float.
   * A value of 0 would mean the shadow does not blur and larger numbers represent increasingly more blur.
   * The total blur area is independent of the Part's area and can become quite large as this number is increased.
   *
   * This value is not affected by scale. Default value is 4.
   * @see {@link #isShadowed}
   * @see {@link #shadowOffset}
   * @see {@link #shadowColor}
   */
  /** @export */ get shadowBlur(): number { return this._shadowBlur; }
  /** @export */ set shadowBlur(value: number) {
    const old = this._shadowBlur;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Part, 'shadowBlur');
      this._shadowBlur = value;
      this.invalidateDraw();
      this.raiseChanged('shadowBlur', old, value);
    }
  }
} // #end Part
Part['className'] = 'Part';

Part['LayoutNone'] = Part.LayoutNone;
Part['LayoutAdded'] = Part.LayoutAdded;
Part['LayoutRemoved'] = Part.LayoutRemoved;
Part['LayoutShown'] = Part.LayoutShown;
Part['LayoutHidden'] = Part.LayoutHidden;
Part['LayoutNodeSized'] = Part.LayoutNodeSized;
Part['LayoutGroupLayout'] = Part.LayoutGroupLayout;
Part['LayoutNodeReplaced'] = Part.LayoutNodeReplaced;
Part['LayoutStandard'] = Part.LayoutStandard;
Part['LayoutAll'] = Part.LayoutAll;

/**
 * An Adornment is a special kind of {@link Part} that is associated with another Part,
 * the {@link Adornment#adornedPart}.
 *
 * Adornments are normally associated with a particular {@link GraphObject} in the adorned Part --
 * that is the value of {@link #adornedObject}.
 * However, the {@link #adornedObject} may be null, in which case the {@link #adornedPart} will also be null.
 *
 * The area occupied by the adorned object is represented in the Adornment's visual tree by a {@link Placeholder}.
 * The placeholder is always the {@link Part#locationObject},
 * although you may specify any {@link Spot} as the {@link Part#locationSpot}.
 * An adornment need not have a placeholder, but it may have at most one.
 *
 * Adornments can be distinguished by their {@link Part#category}.
 * This property can be an arbitrary string value determined by the code creating the adornment,
 * typically a tool that wants to be able to tell various adornments apart from each other.
 * Use the {@link Part#findAdornment} method to find an adornment for a part of a given category.
 *
 * For example, one of the Adornments created by {@link Part#updateAdornments} when the part
 * {@link Part#isSelected} has the {@link Part#category} of "Selection".
 * Those created by {@link ResizingTool#updateAdornments} have a category of "Resize"
 * and normally contain eight resize handles.
 *
 * Besides the selection Adornment and tool Adornments, Adornments are also used for
 * context menus and tooltips.
 * The {@link #adornedObject} in such cases refers to the {@link GraphObject} to which the
 * the context menu or tooltip applies.
 *
 * There cannot be any links connected to an Adornment,
 * nor can an Adornment have members or be a member of a group.
 *
 * An Adornment cannot have its own Adornments.
 * An Adornment cannot be selected.
 *
 * Adornments are not positioned by a {@link Layout} because they are normally positioned
 * according to the Part that they adorn.
 *
 * For more discussion and examples, see <a href="../../intro/selection.html">Selection</a>,
 * <a href="../../intro/toolTips.html">ToolTips</a>,
 * <a href="../../intro/contextMenus.html">Context Menus</a>, and
 * <a href="../../intro/tools.html">Tools</a>.
 * @extends Part
 * @unrestricted
 */
export class Adornment extends Part {
  /** @hidden @internal */ public static className = 'Adornment';
  /** @hidden @internal */ private _adornedObject: GraphObject | null;
  /** @hidden @internal */ private _placeholder: Placeholder;
  /** @hidden @internal */ public _oldPointsCount: number;
  /** @hidden @internal */ public _wasResegmentable: boolean;
  /** @hidden @internal */ public _childrenBounds: Array<Rect>;

  /**
   * @param {PanelLayout=} type if not supplied, the default Panel type is {@link Panel.Position}.
   */
  constructor(type?: PanelLayout) {
    super(type);

    this._partFlags &= ~0x00000100; // not selectable
    this._layerName = 'Adornment';

    this._adornedObject = null;

    // used in LinkReshapingTool
    this._oldPointsCount = 0;
    this._wasResegmentable = false;

    // used when adornments are of type Panel.Link
    this._childrenBounds = [];

    this._placeholder = null;
  }

  /**
   * @hidden @internal
   * @this {Adornment}
   * @return {string}
   */
  public toString(): string {
    const apart = this.adornedPart;
    return 'Adornment(' + this.category + ')' + (apart !== null ? apart.toString() : '');
  }

  /**
   * @hidden @internal
   * @this {Adornment}
   */
  public updateRelationshipsFromData(): void {
  }

  /**
   * @hidden @internal
   * Only for Link Adornments
   * @this {Adornment}
   * @param {boolean} force true if you want to re-create the geometry even if one exists and the route is valid
   */
  public makeShapeGeometry(force: boolean): void {
    const link = this.adornedObject.part;
    const shape = this.adornedObject;
    if (!(link instanceof Link)) return;
    if (!(shape instanceof Shape)) return;

    const path = link.path;
    let geom = path.geometry;
    // If the selection object does not have geometry we will make it now
    link.makeShapeGeometry(force);
    geom = path.geometry;

    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const e = children[i];
      // Take all shapes that have panelMain = true and give them a specific, shared geometry
      if (e.isPanelMain && e instanceof Shape) {
        e._geometry = geom;  // all must share same Geometry
      }
    }
  }

  /**
   * This read-only property returns a {@link Placeholder} that this Adornment may contain in its visual tree.
   * This may be null if there is no such placeholder object.
   */
  /** @export */ get placeholder(): Placeholder | null { return this._placeholder; }

  /**
   * Gets or sets the {@link GraphObject} that is adorned.
   * Setting this property updates the adorned object's Part by calling {@link Part#addAdornment}.
   * This may be null if the Adornment does not adorn a particular object.
   */
  /** @export */ get adornedObject(): GraphObject | null { return this._adornedObject; }
  /** @export */ set adornedObject(value: GraphObject | null) {
    if (Debug && value !== null) {
      Util.checkClass(value, GraphObject, Part, 'adornedObject:value');
    }
    const op = this.adornedPart;
    let part: Part = null;
    if (value !== null) part = value.part;
    if (op !== null && (value === null || op !== part)) {
      // remove from old map, if it exists
      op.removeAdornment(this.category);
    }
    this._adornedObject = value;
    if (part !== null) part.addAdornment(this.category, this);
  }

  /**
   * This read-only property returns the {@link Part} that contains the adorned object.
   * This will be null if the {@link #adornedObject} is null.
   */
  /** @export */ get adornedPart(): Part | null {
    const obj = this._adornedObject;
    if (obj !== null) return obj.part;
    return null;
  }

  /**
   * @hidden @internal
   * @this {Adornment}
   * @return {boolean}
   */
  public isReadyToMeasureArrange(): boolean {
    const obj = this._adornedObject;
    if (obj === null) return true;
    const adpart = obj.part;
    return (adpart === null || !adpart.getInvalidArrange());
  }

  /**
   * @hidden @internal
   * @this {Adornment}
   * @return {boolean}
   */
  public isNodeLike(): boolean {
    return false;
  }

  /**
   * @hidden @internal
   * This read-only property returns always returns null; it cannot be set.
   */
  /** @export */ get containingGroup(): Group | null { return null; }

  /**
   * @hidden @internal
   * Call this method to notify that the some state has changed.
   * This just calls the {@link #diagram}'s {@link Diagram#raiseChangedEvent} method.
   * @this {Adornment}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raisePartChangedEvent(change: EnumValue, propertyname: string, obj: Object, oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    // maintain Placeholder
    if (change === ChangedEvent.Insert && propertyname === 'elements') {
      if (newval instanceof Placeholder) {
        if (this._placeholder === null) {
          this._placeholder = newval;
        } else if (Debug && this._placeholder !== newval) {
          Util.throwError('Cannot insert a second Placeholder into the visual tree of an Adornment.');
        }
      } else if (newval instanceof Panel) {
        const plc = newval.findInVisualTree(function (g) { return g instanceof Placeholder; });
        if (plc instanceof Placeholder) {
          if (this._placeholder === null) {
            this._placeholder = plc;
          } else if (Debug && this._placeholder !== plc) {
            Util.throwError('Cannot insert a second Placeholder into the visual tree of an Adornment.');
          }
        }
      }
    } else if (change === ChangedEvent.Remove && propertyname === 'elements') {
      if (this._placeholder !== null) {
        if (oldval === this._placeholder) {
          this._placeholder = null;
        } else if (oldval instanceof Panel && this._placeholder.isContainedBy(oldval)) {
          this._placeholder = null;
        }
      }
    }
    super.raisePartChangedEvent(change, propertyname, obj, oldval, newval, oldparam, newparam);
  }

  /**
   * @hidden @internal
   * Adornments cannot have any of their own Adornments, so this method is always a no-op.
   * @this {Adornment}
   */
  public updateAdornments(): void {
  }

  /**
   * @hidden @internal
   * @this {Adornment}
   */
  public deleteCascade(): void {
  }

} // #end Adornment
Adornment['className'] = 'Adornment';

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Node
*/

/**
 * A Node is a {@link Part} that may connect to other nodes with {@link Link}s,
 * or that may be a member of a {@link Group}.
 *
 * {@link Group} inherits from Node,
 * enabling nodes to logically contain other nodes and links.
 * <p class="boxread">
 * For a more general discussion of how to define nodes, see <a href="../../intro/nodes.html">Introduction to Nodes</a>.
 *
 * Although you can create a Node and {@link Diagram#add} it to a Diagram, this does not update the Model.
 * It is more common to create a node by adding a node data object to the model
 * by calling {@link Model#addNodeData}. For example:
 * ```js
 *   myDiagram.startTransaction("make new node");
 *   myDiagram.model.addNodeData({ key: "Omega" });
 *   myDiagram.commitTransaction("make new node");
 * ```
 *
 * This will cause a Node or simple {@link Part} to be created (copying the template found in {@link Diagram#nodeTemplateMap}),
 * added to the Diagram in some {@link Layer} (based on {@link Part#layerName}), and bound to the node data
 * (resulting in {@link Panel#data} referring to that node data object).
 * If you do not keep a reference to that JavaScript object, as the above code does not,
 * you can retrieve it later by calling {@link Model#findNodeDataForKey}.
 *
 * It is very common to initialize a Diagram by setting {@link Model#nodeDataArray} to a JavaScript Array
 * of JavaScript objects holding the properties that you need in your model.
 * Nearly all of the samples do this kind of initialization.
 *
 * You can delete a Node by either calling {@link Diagram#remove} or by calling {@link Model#removeNodeData}.
 * The latter obviously will modify the Model; the former does so if the Node was created from model data.
 * Commands such as {@link CommandHandler#deleteSelection} call these methods within a transaction.
 *
 * You can find all of the {@link Link}s that are connected with a Node by calling {@link #findLinksConnected}.
 * Because links normally have a direction, you can find all of the links that have their {@link Link#toNode}
 * be a given Node by calling {@link #findLinksInto}.
 * Similarly, you can call {@link #findLinksOutOf} to find all of the links coming out from a node;
 * such links have their {@link Link#fromNode} be that node.
 * For tree-structured graphs, use {@link #findTreeChildrenLinks} or {@link #findTreeParentLink}.
 *
 * If you are not so interested in the links but are interested in the nodes at the other end of the links
 * connecting with a node, there are other methods that you can call.
 * {@link #findNodesConnected} returns all of the nodes that are at the other end of the links that
 * connect with a given node.
 * {@link #findNodesInto} and {@link #findNodesOutOf} return the subsets of those nodes considering
 * only those links that go into or come out of the given node.
 * For tree-structured graphs, use {@link #findTreeChildrenNodes} or {@link #findTreeParentNode}.
 *
 * For example, to operate on the data of all of the destination nodes:
 * ```js
 * var it = somenode.findNodesOutOf();
 * while (it.next()) {
 *   var child = it.value;
 *   if (child.data.text.indexOf("special") >= 0) { ... }
 * }
 * ```
 *
 * You can link two nodes by creating a new {@link Link}, setting its {@link Link#toNode} and {@link Link#fromNode}
 * (in either order), and {@link Diagram#add}ing it to the diagram.
 * But it is more common to add a link data object to the {@link Diagram#model} by calling {@link GraphLinksModel#addLinkData}.
 * Just creating and adding a {@link Link} will not update the model.
 *
 * Thus to add a link when using a {@link GraphLinksModel} you should do something like:
 * ```js
 *   myDiagram.startTransaction("make new link");
 *   myDiagram.model.addLinkData({ from: "Alpha", to: "Beta" });
 *   myDiagram.commitTransaction("make new link");
 * ```
 *
 * Where you would substitute the keys of the actual nodes that you want to connect with a link.
 * If you are using a {@link TreeModel}, there are no link data objects, so you just need to call
 * {@link TreeModel#setParentKeyForNodeData} to specify the "parent" node's key for a "child" node data.
 *
 * To find a {@link Link} given a link data object in the {@link GraphLinksModel},
 * call {@link Diagram#findLinkForData}.  When using a {@link TreeModel}, call either {@link Diagram#findNodeForData}
 * or {@link Diagram#findNodeForKey} to get a Node, and then call {@link #findTreeParentLink} to get the Link, if any exists.
 *
 * To find a link that connects two nodes, call {@link #findLinksTo} or {@link #findLinksBetween}.
 * With the former method, the direction matters; with the latter method it returns links in either direction.
 *
 * As links connect with a node or are disconnected, you may want to update the appearance of the node.
 * You can set the {@link #linkConnected} and {@link #linkDisconnected} properties to be functions that are called.
 * These functions must not modify any link relationships -- the properties just exist to update the appearance of the node.
 * A typical usage would be to change the color or figure of a shape.
 *
 * You can control whether the user may draw a new link or reconnect a link between a pair of Nodes
 * by affecting the result of {@link LinkingBaseTool#isValidLink}.
 * You can override that predicate on {@link LinkingTool} and {@link RelinkingTool},
 * but it is easier to set the {@link #linkValidation} or
 * {@link LinkingBaseTool#linkValidation} functional property.
 * <p class="boxread">
 * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 *
 * Nodes also support the ability to provide logical and physical distinctions in the connection points
 * that links use at a node. These connection objects are called "ports".
 * By default the port object will be the whole {@link Node}.
 * However, you can set the {@link GraphObject#portId} property on any {@link GraphObject} in the visual tree of a node
 * to cause that element to be treated as a "port".
 * The "port id" is just a string that ought to be unique amongst all of the port elements in the node.
 *
 * In the case of a node only having a single port, you should set the {@link GraphObject#portId} as an empty string.
 * When there is no such element declared as the default port, it uses the whole node.
 * You can use the {@link #port} property to get the only port element.
 *
 * When a node should have multiple ports,
 * i.e. multiple {@link GraphObject}s acting as separate connection points for links,
 * you should set each port's {@link GraphObject#portId} to a string value that is unique for the node.
 * When there may be multiple ports on a node, you can get a collection of elements representing ports
 * by using the {@link #ports} property.
 * Use the {@link #findPort} method to find a particular port element by name.
 *
 * Note: the only kind of model that can save port information, i.e. portIds that are not an empty string,
 * for links is a {@link GraphLinksModel} whose {@link GraphLinksModel#linkFromPortIdProperty} and
 * {@link GraphLinksModel#linkToPortIdProperty} have been set to name properties on the link data objects.
 * <p class="boxread">
 * For a more general discussion of ports, see <a href="../../intro/ports.html">Introduction to Ports</a>.
 *
 * All of the "findLinks..." and "findNodes..." methods mentioned above take an optional port id argument.
 * When no argument is passed, these methods consider all links connecting with the node.
 * When a port id argument is provided, these methods only consider links that connect with that port
 * in the given node.
 * Thus when navigating through the diagram, you can easily look at all of the nodes that links coming out of
 * a given node go to.  Or you can just look at those nodes at the ends of links coming out of a particular port.
 *
 * You can also control the default connecting behavior of {@link Link}s at each port.
 * Because a port can be any {@link GraphObject}, they are all properties on GraphObject.
 * The properties are duplicated so that you can guide the "from" ends of links differently from the "to" ends of links.
 * The properties include:
 *   - {@link GraphObject#fromSpot}, {@link GraphObject#toSpot}
 *   - {@link GraphObject#fromEndSegmentLength}, {@link GraphObject#toEndSegmentLength}
 *   - {@link GraphObject#fromShortLength}, {@link GraphObject#toShortLength}
 *   - {@link GraphObject#fromLinkable}, {@link GraphObject#toLinkable}
 *   - {@link GraphObject#fromLinkableDuplicates}, {@link GraphObject#toLinkableDuplicates}
 *   - {@link GraphObject#fromLinkableSelfNode}, {@link GraphObject#toLinkableSelfNode}
 *   - {@link GraphObject#fromMaxLinks}, {@link GraphObject#toMaxLinks}
 *
 * The "...Spot" and "...Length" and "...Direction" properties control the position and routing of links at a port.
 * The "...Linkable..." and "...MaxLinks" properties control whether or not users can draw a new link
 * or reconnect an existing link from or to a port.
 * (The "...Spot" and "...Length" and "...Direction" properties also exist on {@link Link}, to override for a particular
 * link the default values that come from a port element.)
 * <p class="boxread">
 * For a more general discussion of link points, see <a href="../../intro/connectionPoints.html">Introduction to Link Connection Points</a>.
 *
 * When the graph is tree-structured, you can use several functions for traversing the tree:
 *   - {@link #findTreeParentNode}
 *   - {@link #findTreeChildrenNodes}
 *   - {@link #findTreeParentLink}
 *   - {@link #findTreeChildrenLinks}
 *   - {@link #findTreeRoot}
 *   - {@link #findTreeParentChain}
 *   - {@link #findTreeParts}
 *   - {@link #findCommonTreeParent}
 *   - {@link #isInTreeOf}
 *   - {@link #findTreeLevel}
 *
 * Determining whether a tree grows from the root via links that go out to the children or vice-versa
 * is controlled for the whole diagram by the {@link Diagram#isTreePathToChildren} property.
 * However an individual link will be ignored by the above functions if {@link Link#isTreeLink} is false.
 *
 * The Node class also supports the notion of expanding and collapsing a subtree of nodes and links,
 * causing those nodes and links to be shown or hidden.
 * Principally this is a matter of setting {@link Node#isTreeExpanded}.
 * Of course if the diagram's graph is not tree-structured, these concepts and properties might not apply.
 *
 * If you want to change the appearance of the node you can do so in a function that you assign to
 * the {@link #treeExpandedChanged} property.
 * This function must not modify any link relationships or expand or collapse any subtrees -- the property just exists
 * to update the appearance of the node.
 *
 * There is an option for link routing to try to avoid crossing over nodes:
 * {@link Link#routing} = {@link Link.AvoidsNodes}.
 * You can control whether such links should avoid or ignore a node by setting {@link #avoidable}.
 * Set {@link #avoidableMargin} to control the area beyond the {@link GraphObject#actualBounds}
 * where AvoidsNodes links should not go.
 * <p class="boxread">
 * For more discussion and examples, see <a href="../../intro/nodes.html">Nodes</a>,
 * <a href="../../intro/ports.html">Ports</a>, and
 * <a href="../../intro/connectionPoints.html">Link Points</a>.
 * <p class="boxread">
 * For more about trees, see <a href="../../intro/trees.html">Trees</a>, and
 * <a href="../../intro/subtrees.html">SubTrees</a>.
 * <p class="boxread">
 * To customize user-resizing behavior, please read
 * <a href="../../intro/tools.html#ResizingTool">Introduction to the ResizingTool</a>.
 * To customize user-rotating behavior, please read
 * <a href="../../intro/tools.html#RotatingTool">Introduction to the RotatingTool</a>.
 *
 * Only Nodes that are in Diagrams can have connections via Links.
 * Templates should not be connected with Links, be labels of Links, be members of Groups, or have any Adornments.
 * @extends Part
 * @unrestricted
 */
export class Node extends Part {
  /** @hidden @internal */ public static className = 'Node';
  /** @hidden @internal */ private _linkConnected: ((thisNode: Node, newLink: Link, thisPort: GraphObject) => void) | null;
  /** @hidden @internal */ private _linkDisconnected: ((thisNode: Node, oldLink: Link, thisPort: GraphObject) => void) | null;
  /** @hidden @internal */ private _linkValidation: ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null;
  /** @hidden @internal */ private _labeledLink: Link | null;
  /** @hidden @internal */ private _treeExpandedChanged: ((thisNode: Node) => void) | null;
  /** @hidden @internal */ private _avoidableMargin: Margin;
  /** @hidden @internal */ private _bundles: Set<LinkBundle> | null;
  /** @hidden @internal */ private _portSpreading: EnumValue;
  /** @hidden @internal */ private _treeParentNode: Node | boolean | null;
  /** @hidden @internal */ public _nodeFlags: number;
  /** @hidden @internal */ public _ports: Map<string, GraphObject> | null;
  /** @hidden @internal */ public _linksConnected: List<Link>;

  /**
   * Constructs an empty Node.
   * The panel type must be one of the values permitted by {@link Panel#type}.
   * @param {PanelLayout=} type if not supplied, the default Panel type is {@link Panel.Position}.
   */
  constructor(type?: PanelLayout) {
    super(type);

    // BIT FLAGS for Node (and Group):
    // 0x00000001 this._isTreeExpanded = true;
    // 0x00000002 this._wasTreeExpanded = false;
    // 0x00000004 this._isTreeLeaf = true;
    // 0x00000008 this._avoidable = true;
    this._nodeFlags = 13;  // 0x00000001|0x00000004|0x00000008

    this._linksConnected = new List<Link>();  // related Links; must be List for filter support
    this._linkConnected = null;
    this._linkDisconnected = null;
    this._linkValidation = null;
    this._labeledLink = null;  // owning Link; non-null implies isLinkLabel===true

    this._treeExpandedChanged = null;

    this._avoidableMargin = Margin.TwoMargin;

    this._bundles = null;  // all LinkBundles connected to this node
    this._ports = null;  // must be in visual tree
    this._portSpreading = Node.SpreadingEvenly;

    this._treeParentNode = false; // null is a valid value, so default is false
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Node.
   * @this {Node}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);

    // BIT FLAGS for Node (and Group):
    // 0x00000001 copy._isTreeExpanded = this._isTreeExpanded;
    // 0x00000002 copy._wasTreeExpanded = this._wasTreeExpanded;
    // 0x00000004 copy._isTreeLeaf = this._isTreeLeaf;
    // 0x00000008 copy._avoidable = this._avoidable;
    copy._nodeFlags = this._nodeFlags;
    copy._nodeFlags = this._nodeFlags & ~0x00000010;

    // copy._linksConnected references should be handled by copier
    copy._linkConnected = this._linkConnected;
    copy._linkDisconnected = this._linkDisconnected;
    copy._linkValidation = this._linkValidation;
    // copy._labeledLink reference should be handled by copier

    copy._treeExpandedChanged = this._treeExpandedChanged;

    copy._avoidableMargin = this._avoidableMargin.copyFrozen();

    // copy._bundles should be rebuilt
    // copy._ports will be reconstructed upon demand
    copy._portSpreading = this._portSpreading;

    // copy._treeParentNode should be rebuilt since it is a reference
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can get a new visual tree
   * @this {Node}
   * @param {GraphObject} copy  The original Part whose category is being changed by displacing its visual tree with this Part's visual tree.
   */
  protected displaceProtected(copy: this): void {
    super.displaceProtected(copy);
    copy.invalidateConnectedLinks();
    copy._bundles = this._bundles;
    copy._ports = null;  // recalculate any ports for a Node
  }

  /**
   * This value for {@link Node#portSpreading} indicates that links connecting with a port
   * should all connect at a single point on the side(s) indicated by a Spot that is {@link Spot#isSide}.
   * @constant
   */
  public static SpreadingNone = new EnumValue(Node, 'SpreadingNone', 10);

  /**
   * This default value for {@link Node#portSpreading} indicates that links connecting with a port
   * should be distributed evenly along the side(s) indicated by a Spot that is {@link Spot#isSide}.
   * @constant
   */
  public static SpreadingEvenly = new EnumValue(Node, 'SpreadingEvenly', 11);

  /**
   * This value for {@link Node#portSpreading} indicates that links connecting with a port
   * should packed together based on the link's shape's width
   * on the side(s) indicated by a Spot that is {@link Spot#isSide}.
   * @constant
   */
  public static SpreadingPacked = new EnumValue(Node, 'SpreadingPacked', 12);

  /**
   * @hidden @internal
   * @this {Node}
   * @param {LinkBundle} bundle
   */
  public addBundle(bundle: LinkBundle): void {
    if (bundle === null) return;
    if (this._bundles === null) {  // create collection if needed
      this._bundles = new Set<LinkBundle>();
    }
    this._bundles.add(bundle);
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {LinkBundle} bundle
   */
  public removeBundle(bundle: LinkBundle): void {
    if (bundle === null) return;
    if (this._bundles !== null) {
      this._bundles.remove(bundle);
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {Node} other
   * @param {string} thisportid
   * @param {string} otherportid
   * @return {LinkBundle}
   */
  public findBundle(other: Node, thisportid: string, otherportid: string): LinkBundle | null {
    if (other === null) return null;
    if (this._bundles === null) return null;
    const it = this._bundles.iterator;
    while (it.next()) {
      const b = it.value;
      if ((b.node1 === this && b.node2 === other && b.portId1 === thisportid && b.portId2 === otherportid) ||
        (b.node1 === other && b.node2 === this && b.portId1 === otherportid && b.portId2 === thisportid)) {
        return b;
      }
    }
    return null;
  }

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Node}
   * @param {Node} other
   * @param {string=} thisportid
   * @param {string=} otherportid
   */
  public invalidateLinkBundle(other: Node, thisportid?: string, otherportid?: string): void {
    if (thisportid === undefined || thisportid === null) thisportid = '';
    if (otherportid === undefined || otherportid === null) otherportid = '';
    const bundle = this.findBundle(other, thisportid, otherportid);
    if (bundle !== null) bundle.invalidate();
  }

  /**
   * @hidden @internal
   * Call this method to notify that the some state has changed.
   * This just calls the {@link #diagram}'s {@link Diagram#raiseChangedEvent} method.
   * @this {Node}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raisePartChangedEvent(change: EnumValue, propertyname: string, obj: Object, oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    if (change === ChangedEvent.Insert && propertyname === 'elements') {
      this._ports = null;  // invalidate ports map
    } else if (change === ChangedEvent.Remove && propertyname === 'elements') {
      this._ports = null;  // invalidate ports map
    }
    super.raisePartChangedEvent(change, propertyname, obj, oldval, newval, oldparam, newparam);
  }

  //    * This invalidates the routes of all links connected to this node.
  //  * Basically this calls {@link Link#invalidateRoute} on all connected links
  //  * and invalidates any cached side-connection information.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Node}
   * @param {Set.<Part>=} ignore An optional set of {@link Link}s to ignore.
   */
  public invalidateConnectedLinks(ignore?: Set<Part>): void {
    if (ignore === undefined) ignore = null;

    const lit = this.linksConnected;
    while (lit.next()) {
      const link = lit.value;
      // maybe ignore this link:
      if (ignore !== null && ignore.contains(link)) continue;
      // remove any cached side connection info on either port of the link
      this.invalidatePortInfo(link.fromPort);
      this.invalidatePortInfo(link.toPort);
      // and invalidate the route of that link
      link.invalidateRoute();
    }
  }

  /**
   * @hidden @internal
   * For a given port, invalidate any connected links
   * @this {Node}
   * @param {GraphObject} port the port
   */
  public invalidatePortLinks(port: GraphObject): void {
    const lit = this.linksConnected;
    while (lit.next()) {
      const link = lit.value;
      if (link.fromPort !== port && link.toPort !== port) continue;
      // remove any cached side connection info on either port of the link
      this.invalidatePortInfo(link.fromPort);
      this.invalidatePortInfo(link.toPort);
      // and invalidate the route of that link
      link.invalidateRoute();
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {GraphObject} port
   */
  public invalidatePortInfo(port: GraphObject): void {
    if (port === null) return;
    const knot = port._portInfo;
    if (knot !== null) {
      knot.invalidate();
    }
    const g = this.containingGroup;
    if (g !== null && !g.isSubGraphExpanded) {
      g.invalidatePortInfo(g.port);
    }
  }

  /**
   * @hidden @internal
   * A node is ready to be measured/arranged if is not a link label,
   * or if it is a link label if its link has been arranged.
   * @this {Node}
   * @return {boolean}
   */
  public isReadyToMeasureArrange(): boolean {
    // Always returns true, though with link labels they may not get a correct position until later
    return true;
  }

  /**
   * Gets or sets how link points are computed when the port spot is a "side" spot.
   * The default value is {@link Node.SpreadingEvenly}.
   * @since 1.5
   */
  /** @export */ get portSpreading(): EnumValue { return this._portSpreading; }
  /** @export */ set portSpreading(value: EnumValue) {
    const old = this._portSpreading;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Node, Node, 'portSpreading');
      this._portSpreading = value;
      this.raiseChanged('portSpreading', old, value);
      const diagram = this.diagram;
      if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
      this.invalidateConnectedLinks();
    }
  }

  /**
   * Gets or sets whether this Node is to be avoided by {@link Link}s
   * whose {@link Link#routing} is {@link Link.AvoidsNodes}.
   *
   * The default value is true.
   */
  /** @export */ get avoidable(): boolean { return (this._nodeFlags & 0x00000008) !== 0; }
  /** @export */ set avoidable(value: boolean) {
    const old = (this._nodeFlags & 0x00000008) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Node, 'avoidable');
      this._nodeFlags = this._nodeFlags ^ 0x00000008;
      const diagram = this.diagram;
      if (diagram !== null) diagram.invalidatePositionArray(this);
      this.raiseChanged('avoidable', old, value);
    }
  }

  /**
   * Gets or sets the margin around this Node in which avoidable links will not be routed.
   *
   * You may need to increase the {@link #fromEndSegmentLength} and {@link #toEndSegmentLength} in
   * order to prevent link routes from turning within the avoidable area around the Node.
   *
   * Value must be of type Margin.
   * The default margin is Margin(2,2,2,2)
   */
  /** @export */ get avoidableMargin(): MarginLike { return this._avoidableMargin; }
  /** @export */ set avoidableMargin(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Node, 'avoidableMargin');
    }
    const old = this._avoidableMargin;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._avoidableMargin = value;
      const diagram = this.diagram;
      if (diagram !== null) diagram.invalidatePositionArray(this);
      this.raiseChanged('avoidableMargin', old, value);
    }
  }

  /**
   * @hidden @internal
   * This predicate determines whether this node should be avoided in the routing
   * of links whose {@link Link#routing} is {@link Link.AvoidsNodes}.
   * This is not called for Nodes in temporary layers (which includes Tools and Adornments).
   * @this {Node}
   * @return {boolean}
   */
  public canAvoid(): boolean {
    return this.avoidable && !this.isLinkLabel;
  }

  //    * Return the area to be avoided for this node -- the node's {@link GraphObject#actualBounds}
  //  * plus the {@link Node#avoidableMargin}.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Node}
   * @param {Rect} result  a {@link Rect} to be modified and returned.
   * @return {Rect} the area in document coordinates.
   */
  public getAvoidableRect(result: Rect): Rect {
    result.set(this.actualBounds);
    result.addMargin(this._avoidableMargin);
    return result;
  }

  //    * Starting with this node, walk up the chain of containingGroups to find a node that is visible.
  //  * This can be overridden to find a tree-parent/ancestor if the reason that this node
  //  * is not visible is because of a collapsed tree rather than a collapsed group.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Node}
   * @return {Node}
   */
  public findVisibleNode(): Node | null {
    let n = this as Node;
    while (n !== null && !n.isVisible()) {
      n = n.containingGroup;
    }
    return n;
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @return {boolean}
   */
  public isVisible(): boolean {
    if (!super.isVisible()) return false;
    let tochildren = true;
    let policy = Diagram.TreeParentCollapsed;
    const diagram = this.diagram;
    if (diagram !== null) {
      // Objects that are animating are forced to be visible?
      if (diagram.animationManager.objRemainsVisible(this)) return true;
      tochildren = diagram.isTreePathToChildren;
      policy = diagram.treeCollapsePolicy;
    }
    if (policy === Diagram.TreeParentCollapsed) {
      // check if this Node should be unseen because its tree parent is collapsed
      const parent = this.findTreeParentNode();
      if (parent !== null && !parent.isTreeExpanded) return false;
    } else if (policy === Diagram.AllParentsCollapsed) {
      // if all "parents" are collapsed, this node is not visible
      const parents = (tochildren ? this.findNodesInto() : this.findNodesOutOf());
      if (parents.count > 0 && parents.all(function (n) { return !n.isTreeExpanded; })) return false;
    } else if (policy === Diagram.AnyParentsCollapsed) {
      // if any "parent" is collapsed, this node is not visible
      const parents = (tochildren ? this.findNodesInto() : this.findNodesOutOf());
      if (parents.count > 0 && parents.any(function (n) { return !n.isTreeExpanded; })) return false;
    }
    // if it's a label node, we need to track down whether the labeled link is visible...
    const ll = this.labeledLink;
    if (ll !== null) {
      return ll.isVisible();
    }
    return true;
  }

  /**
   * @hidden @internal
   * Called when a part's visibility changes, from .visible or other indirect sources (like layer.visible)
   * @this {Node}
   * @param {boolean} val whether the part was just made visible (true) or just made invisible (false)
   */
  public madeVisible(val: boolean): void {
    super.madeVisible(val);
    // This updates all links connected to the node recursively,
    // validating routes if they are invalid, but does not invalidate any links.
    const lit = this.linksConnected;
    while (lit.next()) {
      const link = lit.value;
      link.madeVisible(val);
    }
  }

  /**
   * This read-only property returns an iterator over all of the {@link Link}s that are connected with this node.
   * This includes both links that are coming out of this node as well as links that are going into this node.
   * Setting {@link Link#fromNode} or {@link Link#toNode} to refer to this Node
   * will add that {@link Link} to this collection.
   *
   * Use the {@link #findLinksConnected}, {@link #findLinksOutOf}, or {@link #findLinksInto} methods
   * to get different subsets of the links, depending on direction or depending on connecting to a particular port.
   *
   * A template should not have any links connected with it.
   */
  /** @export */ get linksConnected(): Iterator<Link> { return this._linksConnected.iterator; }

  /**
   * Returns an iterator over all of the {@link Link}s that connect with this node in either direction,
   * perhaps limited to the given port id on this node.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Link>}
   */
  public findLinksConnected(pid?: string | null): Iterator<Link> {
    if (pid === undefined) pid = null;
    if (pid === null) return this._linksConnected.iterator;
    // if PID is a string, see if it matches what the link specifies
    if (Debug) Util.checkPrimitive(pid, 'string', Node, 'findLinksConnected:pid');
    const it = new ListIterator(this._linksConnected);
    const node = this;
    /**
     * @hidden @internal
     * @param {Link} l
     */
    it.predicate = function (l: Link) {
      return (l.fromNode === node && l.fromPortId === pid) ||
        (l.toNode === node && l.toPortId === pid);
    };

    return it;
  }

  /**
   * Returns an iterator over all of the {@link Link}s that come out of this node,
   * perhaps limited to the given port id on this node.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Link>}
   */
  public findLinksOutOf(pid?: string | null): Iterator<Link> {
    if (pid === undefined) pid = null;
    if (Debug && pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findLinksOutOf:pid');
    const it = new ListIterator(this._linksConnected);
    const node = this;
    /**
     * @hidden @internal
     * @param {Link} l
     */
    it.predicate = function (l: Link) {
      if (l.fromNode !== node) return false;
      if (pid === null) return true;
      return l.fromPortId === pid;
    };

    return it;
  }

  /**
   * Returns an iterator over all of the {@link Link}s that go into this node,
   * perhaps limited to the given port id on this node.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Link>}
   */
  public findLinksInto(pid?: string | null): Iterator<Link> {
    if (pid === undefined) pid = null;
    if (Debug && pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findLinksInto:pid');
    const it = new ListIterator(this._linksConnected);
    const node = this;
    /**
     * @hidden @internal
     * @param {Link} l
     */
    it.predicate = function (l: Link) {
      if (l.toNode !== node) return false;
      if (pid === null) return true;
      return l.toPortId === pid;
    };

    return it;
  }

  /**
   * Returns an iterator over the {@link Node}s that are connected with this node
   * in either direction, perhaps limited to the given port id on this node.
   *
   * The results may include this node itself if there is a reflexive link connecting this node with itself.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Node>}
   */
  public findNodesConnected(pid?: string | null): Iterator<Node> {
    if (pid === undefined) pid = null;
    if (Debug && pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findNodesConnected:pid');
    let nodes: Set<Node> = null;
    let firstnode: Node = null;
    const it = this._linksConnected.iterator;
    while (it.next()) {
      const link = it.value;
      if (link.fromNode === this) {
        if (pid === null || link.fromPortId === pid) {
          // found a Node
          const to = link.toNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(to);  // now more than two
          } else if (firstnode !== null && firstnode !== to) {  // found just one before
            nodes = new Set<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(to);
          } else {  // first one found
            firstnode = to;  // just one found
          }
        }
      } else if (link.toNode === this) {  // avoid case where toNode === null
        if (pid === null || link.toPortId === pid) {
          // found a Node
          const from = link.fromNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(from);  // now more than two
          } else if (firstnode !== null && firstnode !== from) {  // found just one before
            nodes = new Set<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(from);
          } else {  // first one found
            firstnode = from;  // just one found
          }
        }
      }
    }
    if (nodes !== null) return nodes.iterator;
    if (firstnode !== null) return new SingletonIterator<Node>(firstnode);
    return EmptyIterator.instance;
  }

  /**
   * Returns an iterator over the {@link Node}s that are connected with this node
   * by links coming out of this node, perhaps limited to the given port id on this node.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Node>}
   */
  public findNodesOutOf(pid?: string | null): Iterator<Node> {
    if (pid === undefined) pid = null;
    if (Debug && pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findNodesOutOf:pid');
    let nodes: Set<Node> = null;
    let firstnode: Node = null;
    const it = this._linksConnected.iterator;
    while (it.next()) {
      const link = it.value;
      if (link.fromNode === this) {
        if (pid === null || link.fromPortId === pid) {
          // found a Node
          const to = link.toNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(to);  // now more than two
          } else if (firstnode !== null && firstnode !== to) {  // found just one before
            nodes = new Set<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(to);
          } else {  // first one found
            firstnode = to;  // just one found
          }
        }
      }
    }
    if (nodes !== null) return nodes.iterator;
    if (firstnode !== null) return new SingletonIterator<Node>(firstnode);
    return EmptyIterator.instance;
  }

  /**
   * Returns an iterator over the {@link Node}s that are connected with this node
   * by links going into this node, perhaps limited to the given port id on this node.
   * @export
   * @this {Node}
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Node>}
   */
  public findNodesInto(pid?: string | null): Iterator<Node> {
    if (pid === undefined) pid = null;
    if (Debug && pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findNodesInto:pid');
    let nodes: Set<Node> = null;
    let firstnode: Node = null;
    const it = this._linksConnected.iterator;
    while (it.next()) {
      const link = it.value;
      if (link.toNode === this) {
        if (pid === null || link.toPortId === pid) {
          // found a Node
          const from = link.fromNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(from);  // now more than two
          } else if (firstnode !== null && firstnode !== from) {  // found just one before
            nodes = new Set<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(from);
          } else {  // first one found
            firstnode = from;  // just one found
          }
        }
      }
    }
    if (nodes !== null) return nodes.iterator;
    if (firstnode !== null) return new SingletonIterator<Node>(firstnode);
    return EmptyIterator.instance;
  }

  /**
   * Returns an iterator over all of the {@link Link}s that go from this node to another node or vice-versa,
   * perhaps limited to a given port id on this node and a port id on the other node.
   *
   * If you want all of the links between two nodes in just one direction, use {@link #findLinksTo}.
   * @export
   * @this {Node}
   * @param {Node} othernode
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @param {string|null=} otherpid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Link>}
   * @since 1.2
   */
  public findLinksBetween(othernode: Node, pid?: string | null, otherpid?: string | null): Iterator<Link> {
    if (pid === undefined) pid = null;
    if (otherpid === undefined) otherpid = null;
    if (Debug) {
      Util.checkClass(othernode, Node, Node, 'findLinksBetween:othernode');
      if (pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findLinksBetween:pid');
      if (otherpid !== null) Util.checkPrimitive(otherpid, 'string', Node, 'findLinksBetween:otherpid');
    }
    const it = new ListIterator(this._linksConnected);
    const node = this;
    /**
     * @hidden @internal
     * @param {Link} l
     */
    it.predicate = function (l: Link) {
      if (l.fromNode === node && l.toNode === othernode && (pid === null || l.fromPortId === pid) && (otherpid === null || l.toPortId === otherpid)) return true;
      if (l.fromNode === othernode && l.toNode === node && (otherpid === null || l.fromPortId === otherpid) && (pid === null || l.toPortId === pid)) return true;
      return false;
    };

    return it;
  }

  /**
   * Returns an iterator over all of the {@link Link}s that go from this node to another node,
   * perhaps limited to a given port id on this node and a port id on the other node.
   *
   * If you want all of the links between two nodes in both directions, use {@link #findLinksBetween}.
   * @export
   * @this {Node}
   * @param {Node} othernode
   * @param {string|null=} pid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @param {string|null=} otherpid A port identifier string; if null the link's portId is ignored and all links are included in the search.
   * @return {Iterator.<Link>}
   * @since 1.2
   */
  public findLinksTo(othernode: Node, pid?: string | null, otherpid?: string | null): Iterator<Link> {
    if (pid === undefined) pid = null;
    if (otherpid === undefined) otherpid = null;
    if (Debug) {
      Util.checkClass(othernode, Node, Node, 'findLinksTo:othernode');
      if (pid !== null) Util.checkPrimitive(pid, 'string', Node, 'findLinksTo:pid');
      if (otherpid !== null) Util.checkPrimitive(otherpid, 'string', Node, 'findLinksTo:otherpid');
    }
    const it = new ListIterator(this._linksConnected);
    const node = this;
    /**
     * @hidden @internal
     * @param {Link} l
     */
    it.predicate = function (l: Link) {
      if (l.fromNode !== node) return false;
      if (l.toNode !== othernode) return false;
      if (pid !== null && l.fromPortId !== pid) return false;
      if (otherpid !== null && l.toPortId !== otherpid) return false;
      return true;
    };

    return it;
  }

  /**
   * Gets or sets the function that is called after a {@link Link} has been connected with this Node.
   * It is typically used to modify the appearance of the node.
   * The first argument will be this Node.
   * The second argument will be a Link that is now connected with this node.
   * The third argument will be a GraphObject port indicating which port the link was connected with.
   *
   * If the value is a function, that function must not modify what this Node is connected with.
   * The Link has already been added -- trying to remove it or another link may produce undefined behavior.
   * However, the other end of the link may not yet have been connected with a node (and might never be),
   * so you cannot depend on looking at what the link connects with.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get linkConnected(): ((thisNode: Node, newLink: Link, thisPort: GraphObject) => void) | null { return this._linkConnected; }
  /** @export */ set linkConnected(value: ((thisNode: Node, newLink: Link, thisPort: GraphObject) => void) | null) {
    const old = this._linkConnected;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Node, 'linkConnected');
      this._linkConnected = value;
      this.raiseChanged('linkConnected', old, value);
    }
  }

  /**
   * Gets or sets the function that is called after a {@link Link} has been disconnected from this Node.
   * It is typically used to modify the appearance of the node.
   * The first argument will be this Node.
   * The second argument will be a Link that had been connected with this node.
   * The third argument will be a GraphObject port indicating which port the link had been connected with.
   *
   * If the value is a function, that function must not modify what this Node is connected with.
   * The Link has already been removed -- trying to add it or another link may produce undefined behavior.
   * The other end of the link may not yet have been disconnected from a node (and might never be),
   * so you cannot depend on looking at what the link connects with.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get linkDisconnected(): ((thisNode: Node, oldLink: Link, thisPort: GraphObject) => void) | null { return this._linkDisconnected; }
  /** @export */ set linkDisconnected(value: ((thisNode: Node, oldLink: Link, thisPort: GraphObject) => void) | null) {
    const old = this._linkDisconnected;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Node, 'linkDisconnected');
      this._linkDisconnected = value;
      this.raiseChanged('linkDisconnected', old, value);
    }
  }

  /**
   * Gets or sets a predicate that determines whether or not a Link may be connected with this node.
   * If this is non-null, the predicate is called in addition to the predicate that is {@link LinkingBaseTool#linkValidation}
   * on the {@link LinkingTool} and {@link RelinkingTool}.
   * See {@link LinkingBaseTool#isValidLink} for more details.
   *
   * The default predicate is null, which is equivalent to simply returning true.
   * The first argument will be the proposed "from" Node (may be null).
   * The second argument will be the proposed "from" GraphObject port (may be null).
   * The third argument will be the proposed "to" Node (may be null).
   * The fourth argument will be the proposed "to" GraphObject port (may be null).
   * The fifth argument may be null when asking about creating a new link,
   * or may be a Link when asking about reconnecting an existing link.
   *
   * The function, if supplied, must not have any side-effects.
   * @since 1.3
   */
  /** @export */ get linkValidation(): ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null { return this._linkValidation; }
  /** @export */ set linkValidation(value: ((fromNode: Node, fromPort: GraphObject, toNode: Node, toPort: GraphObject, link: Link) => boolean) | null) {
    const old = this._linkValidation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Node, 'linkValidation');
      this._linkValidation = value;
      this.raiseChanged('linkValidation', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {Link} link
   * @param {GraphObject} port
   */
  public addLinkInternal(link: Link, port: GraphObject): void {
    this.invalidatePortInfo(port);
    const present = this._linksConnected.contains(link);
    if (!present) this._linksConnected.add(link);
    // call .linkConnected event handler, and a second time for reflexive links
    // that are already present in the _linksConnected collection
    if (!present || link.fromNode === link.toNode) {
      const func = this.linkConnected;
      if (func !== null) {
        let oldUpdatingModel = true;
        const diagram = this.diagram;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, link, port);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }
    }
    if (!present) {
      if (!link.isTreeLink) return;  // non-tree-link cannot affect isTreeLeaf or treeParentNode
      const from = link.fromNode;
      const to = link.toNode;
      if (from !== null && to !== null && from !== to) {
        let treepathtochildren = true;
        const diagram = this.diagram;
        if (diagram !== null) treepathtochildren = diagram.isTreePathToChildren;
        const child = treepathtochildren ? to : from;
        const parent = treepathtochildren ? from : to;

        // update treeParentNode
        // if didn't know tree parent before or if there was no parent, set it;
        // otherwise don't change existing cached parent, in case of multiple "parent" links
        if (!child._treeParentNode) child._treeParentNode = parent;

        // update isTreeLeaf
        if (parent.isTreeLeaf) {
          if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
          if (treepathtochildren) {
            if (from === parent) parent.isTreeLeaf = false;
          } else {
            if (to === parent) parent.isTreeLeaf = false;
          }
        } else {
          // adding a link cannot change a non-leaf into a leaf
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {Link} link
   * @param {GraphObject} port
   */
  public removeLinkInternal(link: Link, port: GraphObject): void {
    this.invalidatePortInfo(port);
    const present = this._linksConnected.remove(link);
    let diagram = null;
    if (present || link.toNode === link.fromNode) {
      const func = this.linkDisconnected;
      diagram = this.diagram;
      if (func !== null) {
        let oldUpdatingModel = true;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, link, port);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }
    }
    if (present) {
      if (!link.isTreeLink) return;  // non-tree-link cannot affect isTreeLeaf or treeParentNode
      let treepathtochildren = true;
      if (diagram !== null) treepathtochildren = diagram.isTreePathToChildren;
      const child = treepathtochildren ? link.toNode : link.fromNode;
      const parent = treepathtochildren ? link.fromNode : link.toNode;

      // update treeParentNode
      if (child !== null) child._treeParentNode = false;

      // update isTreeLeaf
      if (parent !== null && !parent.isTreeLeaf) {
        const numlinks = parent._linksConnected.count;
        if (numlinks === 0) {
          parent._treeParentNode = null;  // definitely no parent
          if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
          parent.isTreeLeaf = true;  // definitely no children
        } else {
          parent.updateTreeLeaf();
        }
      } else {
        // removing a link cannot change it from a leaf to a non-leaf
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   */
  public updateTreeLeaf(): void {
    this._treeParentNode = false;  // clear cache
    if (this._linksConnected.count === 0) return;
    let treepathtochildren = true;
    const diagram = this.diagram;
    if (diagram !== null && diagram.undoManager.isUndoingRedoing) return;
    if (diagram !== null) treepathtochildren = diagram.isTreePathToChildren;
    const lit = this._linksConnected.iterator;
    while (lit.next()) {
      const l = lit.value;
      if (!l.isTreeLink) continue;
      if (treepathtochildren) {
        if (l.fromNode === this) {
          this.isTreeLeaf = false;
          return;
        }
      } else {
        if (l.toNode === this) {
          this.isTreeLeaf = false;
          return;
        }
      }
    }
    this.isTreeLeaf = true;
  }

  /**
   * @hidden @internal
   * @this {Node}
   */
  public updateRelationshipsFromData(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    diagram.partManager.updateRelationshipsFromData(this);
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {boolean} switchingLayers
   */
  public internalAddProtected(switchingLayers: boolean): void {
    super.internalAddProtected(switchingLayers);

    if (!switchingLayers) {
      this.updateTreeLeaf();

      const old = this._labeledLink;
      if (old !== null) old.addLabelNodeInternal(this);
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {boolean} switchingLayers
   */
  public internalRemoveProtected(switchingLayers: boolean): void {
    super.internalRemoveProtected(switchingLayers);

    if (!switchingLayers) {
      // do the following instead of: this.labeledLink = null;
      const old = this._labeledLink;
      if (old !== null) old.removeLabelNodeInternal(this);
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   */
  public deleteCascade(): void {
    // delete connected links
    if (this._linksConnected.count > 0) {
      const diagram = this.diagram;
      if (diagram !== null) {
        const del = (diagram.commandHandler !== null) ? diagram.commandHandler.deletesConnectedLinks : true;
        const links = this._linksConnected.copy();
        const lit = links.iterator;
        while (lit.next()) {
          const link = lit.value;
          if (del) {
            diagram.remove(link);
          } else {
            if (link.fromNode === this) link.fromNode = null;
            if (link.toNode === this) link.toNode = null;
          }
        }
      }
    }

    // delete from any link for which this might be a label node
    this.labeledLink = null;

    super.deleteCascade();
  }

  /**
   * This read-only property is true when this Node is a label node for a Link.
   *
   * If this is true, then `n.labeledLink` will be a Link and `n.labeledLink.isLabeledLink` will be true.
   * @see {@link #labeledLink}
   */
  /** @export */ get isLinkLabel(): boolean { return this._labeledLink !== null; }

  /**
   * Gets or sets the {@link Link} for which this Node is acting as a smart label.
   * Most nodes do not act as link labels, so this property will be null.
   *
   * A template should not be a label node for a link.
   * @see {@link #isLinkLabel}
   */
  /** @export */ get labeledLink(): Link | null { return this._labeledLink; }
  /** @export */ set labeledLink(value: Link | null) {
    const old = this._labeledLink;
    if (old !== value) {
      if (Debug && value !== null) Util.checkClass(value, Link, Node, 'labeledLink');
      const diagram = this.diagram;
      const ndata = this.data;
      if (old !== null) {
        old.removeLabelNodeInternal(this);
        // no-op if isUndoingRedoing
        if (diagram !== null && ndata !== null && !diagram.undoManager.isUndoingRedoing) {
          const ldata = old.data;
          const model = diagram.model;
          if (ldata !== null && model.isLinkDataModel()) {
            const key = model.getKeyForNodeData(ndata);
            if (key !== undefined) model.removeLabelKeyForLinkData(ldata, key);
          }
        }
        this.containingGroup = null;
      }
      this._labeledLink = value;
      if (value !== null) {
        value.addLabelNodeInternal(this);
        // no-op if isUndoingRedoing
        if (diagram !== null && ndata !== null && !diagram.undoManager.isUndoingRedoing) {
          const ldata = value.data;
          const model = diagram.model;
          if (ldata !== null && model.isLinkDataModel()) {
            const key = model.getKeyForNodeData(ndata);
            if (key !== undefined) model.addLabelKeyForLinkData(ldata, key);
          }
        }
        this.containingGroup = value.containingGroup;
      }
      this.invalidateArrange();
      this.raiseChanged('labeledLink', old, value);
    }
  }

  /**
   * Find a {@link GraphObject} with a given {@link GraphObject#portId}.
   * If no such GraphObject is found, search for one with the empty string as its port identifier.
   * Finally, when failing to find a port with either the given name or the empty string,
   * this method returns this whole node itself.
   * @export
   * @this {Node}
   * @param {string} pid
   * @return {GraphObject}
   */
  public findPort(pid: string): GraphObject {
    if (Debug) Util.checkPrimitive(pid, 'string', Node, 'findPort:pid');
    if (this._ports === null) {
      // if the ID is the default and there are no named ports, just return this Node
      if (pid === '' && this.hasNamedPorts === false) {
        return this;
      }
      // otherwise continue on as normal
      this._rebuildPorts();
    }
    // look in the cache
    let p = this._ports.getValue(pid);
    if (p !== null) return p;
    // if it wasn't already the empty string, look for one with the empty string,
    // which might be somewhere inside the Node
    if (pid !== '') {
      p = this._ports.getValue('');
      if (p !== null) return p;
    }
    // otherwise consider the whole node as the default port
    return this;
  }

  /**
   * This read-only property returns the primary {@link GraphObject} representing a port in this node.
   * If there is a GraphObject whose {@link GraphObject#portId} is the empty string, return it.
   * If there is no such element, just return this whole Node.
   */
  /** @export */ get port(): GraphObject { return this.findPort(''); }

  /**
   * This read-only property returns an iterator over all of the {@link GraphObject}s in this node that act as ports.
   */
  /** @export */ get ports(): Iterator<GraphObject> {
    if (this._ports === null) this._rebuildPorts();
    return this._ports.iteratorValues;  // _ports is a Map, but want .each to operate on port, not on KVP
  }

  /**
   * @hidden @internal
   * @this {Node}
   */
  private _rebuildPorts(): void {
    // if there's already a Map, clear it
    if (this._ports === null) {
      this._ports = new Map<string, GraphObject>();
    } else {
      this._ports.clear();
    }
    // collect all elements that have a non-null portId
    this.walkVisualTreeFrom(this, function (n, o) { (n as Node).internalAddPort(o); });
    // Didn't find any?  Then include this whole Node itself.
    if (this._ports.count === 0) {
      this._ports.add('', this);
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {GraphObject} obj
   */
  public internalAddPort(obj: GraphObject): void {
    const pid = obj.portId;
    if (pid !== null) {
      if (this._ports !== null) {
        this._ports.add(pid, obj);
      }  // otherwise any use of ._ports will rebuild it on demand
    }
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {GraphObject} obj
   * @param {boolean=} removelinks
   */
  public internalRemovePort(obj: GraphObject, removelinks?: boolean): void {
    const pid = obj.portId;
    if (pid !== null) {
      if (this._ports !== null) {
        this._ports.remove(pid);
      }
      const diagram = this.diagram;
      if (diagram !== null && removelinks) {
        let links: Array<Link> = null;
        const lit = this.findLinksConnected(pid);
        while (lit.next()) {
          const link = lit.value;
          if (links === null) links = Util.tempArray<Link>();
          links.push(link);
        }
        if (links !== null) {
          for (let i = 0; i < links.length; i++) {
            const link = links[i];
            diagram.remove(link);
          }
          Util.freeArray(links);
        }
      }
    }
  }

  /**
   * This predicate is true if this node is a child of the given {@link Node}, perhaps indirectly as a descendant.
   *
   * If this node is a child of the given node according to {@link Diagram#isTreePathToChildren}, this returns true.
   * Otherwise this searches recursively the chain of tree parents of this node,
   * ignoring links for which {@link Link#isTreeLink} is false.
   * A node cannot be in its own subtree.
   *
   * If you what to find out whether this Node is (perhaps indirectly) contained by a Group, use {@link Part#isMemberOf}.
   * @export
   * @this {Node}
   * @param {Node} node the Node that might be a parent or ancestor of this node.
   * @return {boolean}  true if the given node is an ancestor of this node, but false otherwise, including false if it is the same node.
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParentChain}
   */
  public isInTreeOf(node: Node): boolean {
    if (node === null || node === this) return false;
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    let child = this as Node;
    const ancestor = node;
    let parent: Node;
    if (tochildren) {
      while (child !== ancestor) {
        parent = null;
        const it = child._linksConnected.iterator;
        while (it.next()) {
          const link1 = it.value;
          if (!link1.isTreeLink) continue;
          parent = link1.fromNode;
          if (parent !== child && parent !== this) break;  // if reflexive link or cycle, keep going
        }
        if (parent === this) return false;  // break the cycle
        if (parent === null || parent === child) return false;  // no parent? must not be in subtree
        child = parent;  // go up the chain of parents
      }
    } else {
      while (child !== ancestor) {
        parent = null;
        const it = child._linksConnected.iterator;
        while (it.next()) {
          const link2 = it.value;
          if (!link2.isTreeLink) continue;
          parent = link2.toNode;
          if (parent !== child && parent !== this) break;  // if reflexive link or cycle, keep going
        }
        if (parent === this) return false;  // break the cycle
        if (parent === null || parent === child) return false;  // no parent? must not be in subtree
        child = parent;  // go up the chain of parents
      }
    }
    return true;
  }

  /**
   * Return the {@link Node} that is at the root of the tree that this node is in, perhaps this node itself.
   *
   * This node will be {@link #isInTreeOf} the resulting node, unless the resulting node is this node itself.
   * The graph traversal will ignore links for which {@link Link#isTreeLink} is false.
   *
   * If you want to search up the containment hierarchy of {@link Group}s, use {@link Part#findTopLevelPart}.
   * @export
   * @this {Node}
   * @return {Node} If this Node has no "tree parent", this returns itself.
   * @see {@link Diagram#isTreePathToChildren}
   * @see {@link #findTreeParentChain}
   * @since 1.2
   */
  public findTreeRoot(): Node {
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    let rootNode = this as Node;  // root might be itself if it has no parent
    let parent: Node;
    if (tochildren) {
      while (true) {
        parent = null;
        const it = rootNode._linksConnected.iterator;
        while (it.next()) {
          const link1 = it.value;
          if (!link1.isTreeLink) continue;
          parent = link1.fromNode;
          if (parent !== rootNode && parent !== this) break;  // but if reflexive link or cycle, keep going
        }
        if (parent === this) return this;  // break the cycle
        if (parent === null || parent === rootNode) return rootNode;  // no parent? got the root!
        rootNode = parent;  // go up the chain of parents
      }
    } else {
      while (true) {
        parent = null;
        const it = rootNode._linksConnected.iterator;
        while (it.next()) {
          const link2 = it.value;
          if (!link2.isTreeLink) continue;
          parent = link2.toNode;
          if (parent !== rootNode && parent !== this) break;  // but if reflexive link or cycle, keep going
        }
        if (parent === this) return this;  // break the cycle
        if (parent === null || parent === rootNode) return rootNode;  // no parent? got the root!
        rootNode = parent;  // go up the chain of parents
      }
    }
  }

  /**
   * Find the {@link Node} that is the perhaps indirect tree parent of both this node and another one,
   * or this node if it is an ancestor of the other node, or vice-versa.
   *
   * If you want to find the {@link Group} that contains two Parts, call {@link Part#findCommonContainingGroup}.
   * @export
   * @this {Node}
   * @param {Node} other
   * @return {Node} may be null if in different trees,
   *   or may be itself if the OTHER argument is THIS node,
   *   or may be itself if the OTHER node is a descendant of THIS node,
   *   or may be the OTHER node if THIS node is in the tree of the OTHER node.
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParentChain}
   * @since 1.5
   */
  public findCommonTreeParent(other: Node): Node | null {
    if (other === null) return null;
    if (Debug) Util.checkClass(other, Node, Node, 'findCommonTreeParent:other');
    if (this === other) return this;
    // mark this node and all of its parents
    let n = this as Node;
    while (n !== null) {
      n.setIsCommonParent(true);
      n = n.findTreeParentNode();
    }
    // starting at OTHER walk up all its parents, seeing if any are already marked
    let found = null;
    n = other;
    while (n !== null) {
      if (n.getIsCommonParent()) {
        found = n;
        break;
      }
      n = n.findTreeParentNode();
    }
    // clear this node and all of its parents
    n = this;
    while (n !== null) {
      n.setIsCommonParent(false);
      n = n.findTreeParentNode();
    }
    return found;
  }

  /**
   * Returns the {@link Link} that connects with the tree parent {@link Node} of this node if the graph is tree-structured,
   * if there is such a link and {@link Link#isTreeLink} is true.
   * @export
   * @this {Node}
   * @return {Link} The Link to the parent Node, or null if there is no parent node.
   * @see {@link #findTreeParentNode}
   * @see {@link #findTreeChildrenLinks}
   * @see {@link Diagram#isTreePathToChildren}
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParentChain}
   * @since 1.1
   */
  public findTreeParentLink(): Link | null {
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    const it = this._linksConnected.iterator;
    if (tochildren) {
      while (it.next()) {
        const link1 = it.value;
        if (!link1.isTreeLink) continue;
        if (link1.fromNode !== this) return link1;
      }
    } else {
      while (it.next()) {
        const link2 = it.value;
        if (!link2.isTreeLink) continue;
        if (link2.toNode !== this) return link2;
      }
    }
    return null;
  }

  /**
   * Returns the {@link Node} that is the tree parent of this node if the graph is tree-structured, if there is a parent.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   * @export
   * @this {Node}
   * @return {Node} The parent Node, or null if there is no parent node.
   * @see {@link #findTreeParentLink}
   * @see {@link #findTreeChildrenNodes}
   * @see {@link Diagram#isTreePathToChildren}
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParentChain}
   * @since 1.1
   */
  public findTreeParentNode(): Node | null {
    // second half of if is just to clear compiler issue with boolean part of Node|boolean|null
    const tpn = this._treeParentNode;
    if (tpn === null) return null;
    if (tpn instanceof Node) return tpn;
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    const links = this._linksConnected.iterator;
    if (tochildren) {
      while (links.next()) {
        const link = links.value;
        if (!link.isTreeLink) continue;
        const node = link.fromNode;
        if (node !== this) { this._treeParentNode = node; return node; }
      }
    } else {
      while (links.next()) {
        const link = links.value;
        if (!link.isTreeLink) continue;
        const node = link.toNode;
        if (node !== this) { this._treeParentNode = node; return node; }
      }
    }
    this._treeParentNode = null;
    return null;
  }

  /**
   * Return a collection of {@link Part}s including this {@link Node},
   * its tree parent link and node, and so on up the chain to the root node.
   *
   * This calls {@link #findTreeParentLink} and {@link #findTreeParentNode}.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   *
   * This may result in undefined behavior if there are cycles of Links that are {@link Link#isTreeLink}.
   *
   * The result will include this node and the "root" node and all nodes and links in between.
   * The root node is also accessible directly via {@link #findTreeRoot}.
   * If any of the nodes are {@link Group}s, their member parts are not included.
   * @export
   * @this {Node}
   * @return {Set.<Part>} A Set of Nodes and Links.
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParts}
   * @since 1.7
   */
  public findTreeParentChain(): Set<Part> {
    /**
     * @hidden @internal
     * @param {Node} n
     * @param {Set.<Part>} coll
     */
    function addParent(n: Node, coll: Set<Part>) {
      if (n === null) return;
      coll.add(n);
      const l = n.findTreeParentLink();
      if (l !== null) {
        coll.add(l);
        addParent(n.findTreeParentNode(), coll);
      }
    }

    const set = new Set<Part>();
    addParent(this, set);
    return set;
  }

  /**
   * Return how deep this node is in a tree structure.
   * For tree root nodes, this returns zero.
   * This calls {@link #findTreeParentNode} to find any tree parent node,
   * so this respects {@link Diagram#isTreePathToChildren} and {@link Link#isTreeLink}
   * to know which way to traverse links and to know to ignore non-tree links.
   *
   * This may result in undefined behavior if there are cycles of Links that are {@link Link#isTreeLink}.
   *
   * If you want to know how deep a Part is nested inside Groups, call {@link Part#findSubGraphLevel}.
   * @export
   * @this {Node}
   * @return {number}
   * @see {@link #findTreeRoot}
   * @see {@link #findTreeParentChain}
   * @since 1.5
   */
  public findTreeLevel(): number {
    return this._findTreeLevel(this);
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @param {Node} n
   * @return {number}
   */
  private _findTreeLevel(n: Node): number {
    const parent = n.findTreeParentNode();
    if (parent === null) return 0;
    return 1 + this._findTreeLevel(parent);
  }

  /**
   * Returns an {@link Iterator} for the collection of {@link Link}s that connect with the immediate tree children of this node.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   *
   * This basically returns either {@link #findLinksOutOf} or {@link #findLinksInto},
   * depending on {@link Diagram#isTreePathToChildren},
   * but the results excludes links for which {@link Link#isTreeLink} is false.
   * @export
   * @this {Node}
   * @return {Iterator.<Link>}
   * @see {@link #findTreeParentLink}
   * @see {@link #findTreeChildrenNodes}
   * @see {@link #findTreeParts}
   * @since 1.1
   */
  public findTreeChildrenLinks(): Iterator<Link> {
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    const it = new ListIterator(this._linksConnected);
    const node = this;
    if (tochildren) {
      it.predicate =
        /**
         * @hidden @internal
         * @param {Link} l
         */
        function (l: Link) {
          if (!l.isTreeLink) return false;
          if (l.fromNode !== node) return false;
          return true;
        };

    } else {
      it.predicate =
        /**
         * @hidden @internal
         * @param {Link} l
         */
        function (l: Link) {
          if (!l.isTreeLink) return false;
          if (l.toNode !== node) return false;
          return true;
        };

    }
    return it;
  }

  /**
   * Returns an {@link Iterator} for the collection of {@link Node}s that are the immediate tree children of this node.
   * Nodes only connected by links for which {@link Link#isTreeLink} is false are ignored.
   *
   * This basically returns either {@link #findNodesOutOf} or {@link #findNodesInto},
   * depending on {@link Diagram#isTreePathToChildren}.
   * @export
   * @this {Node}
   * @return {Iterator.<Node>}
   * @see {@link #findTreeParentNode}
   * @see {@link #findTreeChildrenLinks}
   * @see {@link #findTreeParts}
   * @since 1.1
   */
  public findTreeChildrenNodes(): Iterator<Node> {
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    let nodes: List<Node> = null;
    let firstnode: Node = null;
    const it = this._linksConnected.iterator;
    if (tochildren) {
      while (it.next()) {
        const tolink = it.value;
        if (!tolink.isTreeLink) continue;
        if (tolink.fromNode === this) {
          // found a Node
          const to = tolink.toNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(to);  // now more than two
          } else if (firstnode !== null && firstnode !== to) {  // found just one before
            nodes = new List<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(to);
          } else {  // first one found
            firstnode = to;  // just one found
          }
        }
      }
    } else {
      while (it.next()) {
        const fromlink = it.value;
        if (!fromlink.isTreeLink) continue;
        if (fromlink.toNode === this) {
          // found a Node
          const from = fromlink.fromNode;
          if (nodes !== null) {  // found at least two before
            nodes.add(from);  // now more than two
          } else if (firstnode !== null && firstnode !== from) {  // found just one before
            nodes = new List<Node>();  // found second one
            nodes.add(firstnode);
            nodes.add(from);
          } else {  // first one found
            firstnode = from;  // just one found
          }
        }
      }
    }
    if (nodes !== null) return nodes.iterator;
    if (firstnode !== null) return new SingletonIterator<Node>(firstnode);
    return EmptyIterator.instance;
  }

  /**
   * Return a collection of {@link Part}s including this {@link Node},
   * all of the {@link Link}s going to child {@link Node}s, and all of their
   * tree child nodes and links.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   *
   * Whether child nodes are found for a parent node by following links out of the parent node or
   * by links coming into the parent node is determined by the value of {@link Diagram#isTreePathToChildren}.
   *
   * The result will include this, the "root" node.
   * If any of the nodes are {@link Group}s, their member parts are not included.
   *
   * If you want to find the collection of Parts that are contained by a {@link Group}, use {@link Group#findSubGraphParts}.
   * @export
   * @this {Node}
   * @param {number=} level How many levels of the tree, starting at this node, to include;
   *    the default is Infinity, including all tree children of this node.  Values less than 1 are treated as 1.
   * @return {Set.<Part>} A Set of Nodes and Links.
   * @see {@link #findTreeChildrenNodes}
   * @see {@link #findTreeChildrenLinks}
   * @see {@link #findTreeParentChain}
   */
  public findTreeParts(level?: number): Set<Part> {
    if (level === undefined) level = Infinity;
    Util.checkPrimitive(level, 'number', Node, 'findTreeParts:level');
    const coll = new Set<Part>();
    Part.gatherCollection(coll, this, false, level, true);
    return coll;
  }

  /**
   * Hide each child node and the connecting link,
   * and recursively collapse each child node.
   * This changes the value of {@link Part#isVisible} of the whole subtree and the parts owned by
   * those nodes and links.
   * However, this root node's visibility is unchanged.
   *
   * Links are assumed to go from the parent node to the children nodes,
   * unless {@link Diagram#isTreePathToChildren} is false.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   *
   * This sets {@link #isTreeExpanded} to false on this node and on all of the children nodes.
   * For those child nodes that were expanded when they were collapsed,
   * {@link #wasTreeExpanded} is set to true.
   *
   * You can also pass in a number of levels to hide nodes beyond a certain level starting at this node.
   * If you want to make sure that all nodes are expanded up to a particular level, call {@link #expandTree}.
   * If you want to do both, call `expandTree` before calling `collapseTree` to
   * collapse nodes expanded due to the {@link #wasTreeExpanded} flag.
   *
   * This method does not perform a transaction.
   * You may want to call the {@link CommandHandler#collapseTree} command, which does perform a transaction
   * and raise a {@link DiagramEvent}.
   *
   * To collapse a {@link Group}'s subgraph of {@link Node}s and {@link Link}s, use {@link Group#collapseSubGraph}.
   * @export
   * @this {Node}
   * @param {number=} level How many levels of the tree, starting at this node, to keep expanded if already expanded;
   *    the default is 1, hiding all tree children of this node.  Values less than 1 are treated as 1.
   * @see {@link #expandTree}
   * @see {@link #findTreeParts}
   */
  public collapseTree(level?: number): void {
    if (level === undefined) level = 1;
    Util.checkRealNumber(level, Node, 'collapseTree:level');
    if (level < 1) level = 1;
    const diagram = this.diagram;
    if (diagram === null) return;
    if (diagram.isCollapsingExpanding) return;
    diagram.isCollapsingExpanding = true;
    const coll = new Set<Node>();
    coll.add(this);
    this._collapseTree2(coll, diagram.isTreePathToChildren, level, diagram, this, diagram.treeCollapsePolicy === Diagram.TreeParentCollapsed);
    diagram.isCollapsingExpanding = false;
  }

  /**
   * @hidden @internal
   * Recurse through the requested levels to be collapsed, to actually collapse only at that level, not before.
   * It's OK for nodes to remain collapsed at levels lower than requested.
   * @this {Node}
   * @param {Set.<Node>} coll
   * @param {boolean} tochildren
   * @param {number} level
   * @param {Diagram} diagram
   * @param {Node} expandroot
   * @param {boolean} singleparent
   */
  private _collapseTree2(coll: Set<Node>, tochildren: boolean, level: number, diagram: Diagram, expandroot: Node, singleparent: boolean): void {
    if (level > 1) {
      const cit = (tochildren ? this.findLinksOutOf() : this.findLinksInto());
      while (cit.next()) {
        const link = cit.value;
        if (!link.isTreeLink) continue;
        const child = link.getOtherNode(this);
        if (child !== null && child !== this && !coll.contains(child)) {
          coll.add(child);
          child._collapseTree2(coll, tochildren, level - 1, diagram, expandroot, singleparent);
        }
      }
    } else {
      this._collapseTree1(coll, tochildren, diagram, expandroot, singleparent);
    }
  }

  /**
   * @hidden @internal
   * This needs to handle non-tree-structured graphs.
   * @this {Node}
   * @param {Set.<Node>} coll
   * @param {boolean} tochildren
   * @param {Diagram} diagram
   * @param {Node} expandroot
   * @param {boolean} singleparent
   */
  private _collapseTree1(coll: Set<Node>, tochildren: boolean, diagram: Diagram, expandroot: Node, singleparent: boolean): void {
    const wasExpanded = (expandroot === this) ? true : this.isTreeExpanded;
    const cit = (tochildren ? this.findLinksOutOf() : this.findLinksInto());
    while (cit.next()) {
      const link = cit.value;
      if (!link.isTreeLink) continue;
      const child = link.getOtherNode(this);
      if (child !== null && child !== this) {
        const contains = coll.contains(child);
        if (!contains) coll.add(child);
        if (wasExpanded) {
          if (singleparent) diagram.addCollapsedNode(child, expandroot);
          child.invalidateContainingGroup();
          child.madeVisible(false);  // takes care of adornments
        }
        if (child.isTreeExpanded) {
          child.wasTreeExpanded = child.isTreeExpanded;
          if (!contains) child._collapseTree1(coll, tochildren, diagram, expandroot, singleparent);
        }
      }
    }
    this.isTreeExpanded = false;
  }

  /**
   * Show each child node and the connecting link,
   * and perhaps recursively expand their child nodes.
   * This may change the value of {@link Part#isVisible} of the whole subtree and the parts owned by
   * those nodes and links.
   * However, this root node's visibility is unchanged.
   *
   * This sets {@link #isTreeExpanded} to true on this node and on all of the children nodes.
   * Links are assumed to go from the parent node to the children nodes,
   * unless {@link Diagram#isTreePathToChildren} is false.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   *
   * This will expand a tree child node only if its {@link #wasTreeExpanded} property was true.
   *
   * You can also pass in a number of levels in order to be sure that all nodes
   * starting at this node and up through that number of levels are visible.
   * If you want to make sure that there are no nodes expanded after a particular level, call {@link #collapseTree}.
   * If you want to do both, call `expandTree` before calling `collapseTree` to
   * collapse nodes expanded due to the {@link #wasTreeExpanded} flag.
   *
   * This method does not perform a transaction.
   * You may want to call the {@link CommandHandler#expandTree} command, which does perform a transaction
   * and raise a {@link DiagramEvent}.
   *
   * To expand a {@link Group}'s subgraph of {@link Node}s and {@link Link}s, use {@link Group#expandSubGraph}.
   * @export
   * @this {Node}
   * @param {number=} level How many levels of the tree should be expanded;
   *    the default is 2, showing all tree children of this node and potentially more.
   *    Values less than 2 are treated as 2.
   * @see {@link #collapseTree}
   * @see {@link #findTreeParts}
   */
  public expandTree(level?: number): void {
    if (level === undefined) level = 2;
    Util.checkRealNumber(level, Node, 'expandTree:level');
    if (level < 2) level = 2;
    const diagram = this.diagram;
    if (diagram === null) return;
    if (diagram.isCollapsingExpanding) return;
    diagram.isCollapsingExpanding = true;
    const coll = new Set<Node>();
    coll.add(this);
    this._expandTree1(coll, diagram.isTreePathToChildren, level, diagram, this, diagram.treeCollapsePolicy === Diagram.TreeParentCollapsed);
    diagram.isCollapsingExpanding = false;
  }

  /**
   * @hidden @internal
   * This needs to handle non-tree-structured graphs.
   * @this {Node}
   * @param {Set.<Node>} coll
   * @param {boolean} tochildren
   * @param {number} level
   * @param {Diagram} diagram
   * @param {Node} expandroot
   * @param {boolean} singleparent
   */
  private _expandTree1(coll: Set<Node>, tochildren: boolean, level: number, diagram: Diagram, expandroot: Node, singleparent: boolean): void {
    const wasExpanded = (expandroot === this) ? false : this.isTreeExpanded;
    const cit = (tochildren ? this.findLinksOutOf() : this.findLinksInto());
    while (cit.next()) {
      const link = cit.value;
      if (!link.isTreeLink) continue;
      if (!wasExpanded) {
        // If it hasn't had a chance to get valid because it was invisible,
        // explicitly invalidate it again so that it can properly call invalidMeasure
        // and get added to the list of invalid objects
        if (!link.validRoute) link.invalidateRoute();
      }
      const child = link.getOtherNode(this);
      if (child !== null && child !== this && !coll.contains(child)) {
        coll.add(child);
        if (!wasExpanded) {
          child.madeVisible(true); // takes care of adornments
          child.invalidateContainingGroup();

          if (singleparent) diagram.addExpandedNode(child, expandroot);
        }
        if (level > 2 || child.wasTreeExpanded) {
          child.wasTreeExpanded = false;
          child._expandTree1(coll, tochildren, level - 1, diagram, expandroot, singleparent);
        }
      }
    }
    this.isTreeExpanded = true;
  }

  /**
   * Gets or sets whether the subtree graph starting at this node is expanded.
   * Changing this property's value will call {@link #collapseTree} or {@link #expandTree},
   * and also will call the value of {@link #treeExpandedChanged} if it is a function.
   *
   * The initial value is true -- "tree-child" nodes, and the links to them, are shown.
   *
   * There is an analogous property for expanded/collapsed {@link Group}s: {@link Group#isSubGraphExpanded}.
   */
  /** @export */ get isTreeExpanded(): boolean { return (this._nodeFlags & 0x00000001) !== 0; }
  /** @export */ set isTreeExpanded(value: boolean) {
    const old = (this._nodeFlags & 0x00000001) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Node, 'isTreeExpanded');
      this._nodeFlags = this._nodeFlags ^ 0x00000001;
      const diagram = this.diagram;
      this.raiseChanged('isTreeExpanded', old, value);
      const func = this.treeExpandedChanged;
      if (func !== null) {
        let oldUpdatingModel = true;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }

      if (diagram !== null && diagram.undoManager.isUndoingRedoing) {
        // also called in expandTree/collapseTree, which do not get called when undoing/redoing
        this.madeVisible(value);
        return;
      }
      if (value) {
        this.expandTree();
      } else {
        this.collapseTree();
      }
    }
  }

  /**
   * Gets or sets whether the subtree graph starting at this node
   * had been collapsed by a call to {@link #expandTree} on the parent node.
   * The initial value is false.
   * @see {@link #isTreeExpanded}
   */
  /** @export */ get wasTreeExpanded(): boolean { return (this._nodeFlags & 0x00000002) !== 0; }
  /** @export */ set wasTreeExpanded(value: boolean) {
    const old = (this._nodeFlags & 0x00000002) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Node, 'wasTreeExpanded');
      this._nodeFlags = this._nodeFlags ^ 0x00000002;
      this.raiseChanged('wasTreeExpanded', old, value);
    }
  }

  /**
   * Gets or sets the function that is called when {@link #isTreeExpanded} has changed value.
   * The argument to that function will be this Node.
   *
   * If the value is a function, that function must not expand or collapse any trees of nodes and links.
   * The Node has already been expanded or collapsed -- trying to change it again may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get treeExpandedChanged(): ((thisNode: Node) => void) | null { return this._treeExpandedChanged; }
  /** @export */ set treeExpandedChanged(value: ((thisNode: Node) => void) | null) {
    const old = this._treeExpandedChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Node, 'treeExpandedChanged');
      this._treeExpandedChanged = value;
      this.raiseChanged('treeExpandedChanged', old, value);
    }
  }

  /**
   * Gets whether this node has no tree children.
   *
   * The initial value is true, meaning that there are no links connected with child nodes
   * in the direction given by {@link Diagram#isTreePathToChildren}.
   * This value changes automatically as link connections are added to or removed from this node.
   * Links for which {@link Link#isTreeLink} is false are ignored.
   */
  /** @export */ get isTreeLeaf(): boolean { return (this._nodeFlags & 0x00000004) !== 0; }
  /** @export */ set isTreeLeaf(value: boolean) {
    const old = (this._nodeFlags & 0x00000004) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Node, 'isTreeLeaf');
      this._nodeFlags = this._nodeFlags ^ 0x00000004;
      this.raiseChanged('isTreeLeaf', old, value);
    }
  }
} // #end class node
Node['className'] = 'Node';

Node['SpreadingNone'] = Node.SpreadingNone;
Node['SpreadingEvenly'] = Node.SpreadingEvenly;
Node['SpreadingPacked'] = Node.SpreadingPacked;

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Group
Placeholder
*/

/**
 * A Group is a {@link Node} that can contain a subgraph of {@link Node}s and {@link Link}s,
 * which are members of the group.
 * <p class="box">
 * For more discussion, see <a href="../../intro/groups.html">Introduction to Groups</a>.
 * See samples that make use of Groups in the <a href="../../samples/index.html#groups">samples index</a>.
 *
 * Although you can create a Group and {@link Diagram#add} it to a Diagram, this does not update the Model.
 * It is more common to create a group by adding a node data object to the model
 * by calling {@link Model#addNodeData}. For example:
 * ```js
 *   myDiagram.startTransaction("make new group");
 *   myDiagram.model.addNodeData({ key: "Omega", isGroup: true });
 *   myDiagram.commitTransaction("make new group");
 * ```
 *
 * This will cause a Group to be created (copying the template found in {@link Diagram#groupTemplateMap}),
 * added to the Diagram in some {@link Layer} (based on {@link Part#layerName}), and bound to the group data
 * (resulting in {@link Panel#data} referring to that group data object).
 * Note that the JavaScript object includes setting `isGroup` to true,
 * to indicate that the object represents a Group rather than a regular Node or simple Part.
 *
 * The member Parts of a Group, which you can access as the {@link #memberParts} collection,
 * belong to the group but are not in the visual tree of the group.
 * All {@link Part}s are directly in {@link Layer}s -- they cannot be inside a {@link Panel}.
 * This allows group member parts to be in layers different from the group's layer.
 *
 * You can change the membership of a {@link Node} or a simple {@link Part} in a Group by setting
 * its {@link Part#containingGroup} property.
 * This is done automatically for you by the diagram if you initialize the `group` property on the node data
 * in the model to be the key of the containing group node data.
 * Thus you should do something like:
 * ```js
 *   myDiagram.startTransaction("add new member");
 *   myDiagram.model.addNodeData({ group: someexistinggroup.data.key, ... });
 *   myDiagram.commitTransaction("add new member");
 * ```
 *
 * where you would make sure the node data object included all of the properties you need.
 * You can also change the relationship dynamically by calling {@link GraphLinksModel#setGroupKeyForNodeData}.
 *
 * The membership of {@link Link}s is computed automatically for you by the diagram based on the membership of
 * the connected {@link Node}s.
 * For example, if the {@link Link#fromNode} is a top-level node but the {@link Link#toNode} is a member of a group,
 * the link is a top-level link.
 * If the two connected nodes both belong to the same group, the link is a member of that group.
 * If the two connected nodes belong to different groups, the link belongs to the common container group, if there is any.
 * Note that if a link connects a member of a group with the group itself, the link is a member of that group.
 *
 * All of the group-member relationships effectively form a tree structure.
 * These properties and methods are useful in navigating these relationships:
 *   - {@link Part#containingGroup}
 *   - {@link Part#isTopLevel}
 *   - {@link Part#findTopLevelPart}
 *   - {@link Part#findSubGraphLevel}
 *   - {@link Part#findCommonContainingGroup}
 *   - {@link #memberParts}
 *   - {@link #findSubGraphParts}
 *   - {@link #findExternalLinksConnected}
 *   - {@link #findExternalNodesConnected}
 *
 * As the membership of a group changes, you may want to update the appearance of the group.
 * You can set the {@link #memberAdded} and {@link #memberRemoved} properties to be functions that are called.
 * These functions must not modify any membership relationships -- these function properties just exist to update the appearance of the Group.
 *
 * You can control whether certain Nodes are added to a Group by {@link CommandHandler#groupSelection} or
 * {@link #addMembers} or {@link CommandHandler#addTopLevelParts} by affecting the result of {@link CommandHandler#isValidMember},
 * which is responsible for deciding whether it is OK to add a Node to a Group or to remove a Node from a Group to be a top-level node.
 * You can override that predicate on CommandHandler, but it is easier to set the {@link #memberValidation} or
 * {@link CommandHandler#memberValidation} functional property.
 * <p class="box">
 * For a more general discussion of validation, see <a href="../../intro/validation.html">Introduction to Validation</a>.
 *
 * The area occupied by the subgraph is represented in the group's visual tree by a {@link Placeholder}.
 * As the group {@link #placeholder} grows and shrinks based on the sizes and positions of the member nodes and links,
 * the group will grow and shrink accordingly.
 * The placeholder is always the {@link Part#locationObject},
 * although you may specify any {@link Spot} as the {@link Part#locationSpot}.
 * A Group need not have a placeholder, but it may have at most one.
 *
 * A group has its own {@link #layout} property that is used to position the member nodes and route the member links.
 *
 * The Group class also supports the notion of expanding and collapsing the subgraph,
 * causing the member nodes and links to be shown or hidden.
 * Principally this is a matter of setting {@link #isSubGraphExpanded}.
 * Changes to this property will result in calls to {@link #collapseSubGraph} or {@link #expandSubGraph}, as appropriate.
 *
 * If you want to change the appearance of the group you can do so in a function that you assign to
 * the {@link #subGraphExpandedChanged} property.
 * This function must not modify any member relationships or expand or collapse any groups -- the functional property just exists
 * to update the appearance of the Group.
 * <p class="box">
 * For more discussion and examples, see <a href="../../intro/subgraphs.html">SubGraphs</a>.
 *
 * If you want the user to be able to create a Group out of the currently
 * selected Parts using the {@link CommandHandler#groupSelection} command,
 * you need to first set the {@link CommandHandler#archetypeGroupData} property
 * to a data object with `isGroup` set to true.
 * If you want the user to be able to ungroup a Group,
 * using the {@link CommandHandler#ungroupSelection} command,
 * you need to set {@link #ungroupable} to true.
 * <p class="box">
 * For more discussion and examples, see <a href="../../intro/groups.html">Groups</a>,
 * <a href="../../intro/subgraphs.html">SubGraphs</a>, and
 * <a href="../../intro/sizedGroups.html">Sized Groups</a>.
 *
 * Only Groups that are in Diagrams can have member Parts or connections via Links.
 * Templates should not be connected with Links, be labels of Links, be members of Groups, have any member Parts, or have any Adornments.
 * @extends Node
 * @unrestricted
 */
export class Group extends Node {
  /** @hidden @internal */ public static className = 'Group';

  /** @hidden @internal */ private _memberParts: Set<Part>;
  /** @hidden @internal */ private _memberAdded: ((thisGroup: Group, newPart: Part) => void) | null;
  /** @hidden @internal */ private _memberRemoved: ((thisGroup: Group, oldPart: Part) => void) | null;
  /** @hidden @internal */ private _memberValidation: ((thisGroup: Group, part: Part) => boolean) | null;
  /** @hidden @internal */ private _subGraphExpandedChanged: ((thisGroup: Group) => void) | null;
  /** @hidden @internal */ private _placeholder: Placeholder | null;
  /** @hidden @internal */ private _layout: Layout | null;
  /** @hidden @internal */ public _nestedGroups: Set<Group>;

  /**
   * Constructs an empty Group with no visual elements and no member parts;
   * normally a Group will have some visual elements surrounding a {@link Placeholder}.
   * @param {PanelLayout=} type if not supplied, the default Panel type is {@link Panel.Position}.
   */
  constructor(type?: PanelLayout) {
    super(type);

    // BIT FLAGS for Group:
    // 0x00000100 this._ungroupable = false;
    // 0x00000200 this._isSubGraphExpanded = true;
    // 0x00000400 this._wasSubGraphExpanded = false;
    // 0x00000800 this._computesBoundsAfterDrag = false;
    // 0x00001000 this._computesBoundsIncludingLinks = true;
    // 0x00002000 this._computesBoundsIncludingLocation = false;
    // 0x00004000 this._handlesDragDropForMembers = false;
    // 0x00008000 this._needsLocation = false;
    // 0x00010000 this._hasInvalidatedConnectedRoutes = false;
    this._nodeFlags = this._nodeFlags | 4608;  // 0x00000200|0x00001000

    this._memberParts = new Set<Part>();  // owned Parts
    this._nestedGroups = new Set<Group>();
    this._memberAdded = null;
    this._memberRemoved = null;
    this._memberValidation = null;

    this._subGraphExpandedChanged = null;

    this._placeholder = null;

    this._layout = new Layout();
    this._layout.group = this;  // set backpointer
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Group.
   * @this {Group}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);

    // BIT FLAGS for Group:
    // 0x00000100 copy._ungroupable = this._ungroupable;
    // 0x00000200 copy._isSubGraphExpanded = this._isSubGraphExpanded;
    // 0x00000400 copy._wasSubGraphExpanded = this._wasSubGraphExpanded;
    // 0x00000800 copy._computesBoundsAfterDrag = this._computesBoundsAfterDrag;
    // 0x00001000 copy._computesBoundsIncludingLinks = this._computesBoundsIncludingLinks;
    // 0x00002000 copy._computesBoundsIncludingLocation = this._computesBoundsIncludingLocation;
    // 0x00004000 copy._handlesDragDropForMembers = this._handlesDragDropForMembers;
    // 0x00008000 don't copy this._needsLocation
    // _nodeFlags was copied by Node.cloneProtected; just reset not-copied flags for Groups
    this._nodeFlags = this._nodeFlags & ~0x00008000;

    // copy._memberParts references should be handled by copier
    // _nestedGroups will be updated automatically along with _memberParts
    copy._memberAdded = this._memberAdded;
    copy._memberRemoved = this._memberRemoved;
    copy._memberValidation = this._memberValidation;

    copy._subGraphExpandedChanged = this._subGraphExpandedChanged;

    const obj = copy.findInVisualTree(function (g) { return g instanceof Placeholder; });
    if (obj instanceof Placeholder) {
      copy._placeholder = obj;
    } else {
      copy._placeholder = null;
    }

    if (this._layout !== null) {
      copy._layout = this._layout.copy();
      copy._layout.group = copy;
    } else {
      if (copy._layout !== null) copy._layout.group = null;
      copy._layout = null;
    }
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can get a new visual tree
   * @this {Group}
   * @param {GraphObject} copy  The original Part whose category is being changed by displacing its visual tree with this Part's visual tree.
   */
  protected displaceProtected(copy: this): void {
    super.displaceProtected(copy);

    const allmembers = copy.findSubGraphParts();
    const cit = copy.memberParts;
    while (cit.next()) {
      const member = cit.value;
      member.invalidateMeasure();
      member.invalidateLayout(Part.LayoutHidden);
      member.clearAdornments();
      if (member instanceof Node) {
        member.invalidateConnectedLinks(allmembers);
      } else if (member instanceof Link) {
        const lit = member.labelNodes;
        while (lit.next()) {
          const lab = lit.value;
          lab.invalidateConnectedLinks(allmembers);
        }
      }
    }

    // nothing else special to do for Group
    // until _placeholder can be computed automatically
  }

  /**
   * @hidden @internal
   * Call this method to notify that the some state has changed.
   * This just calls the {@link #diagram}'s {@link Diagram#raiseChangedEvent} method.
   * @this {Group}
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values include {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string} propertyname names the property that was modified.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raisePartChangedEvent(change: EnumValue, propertyname: string, obj: Object, oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    // maintain Placeholder
    if (change === ChangedEvent.Insert && propertyname === 'elements') {
      if (newval instanceof Placeholder) {
        if (this._placeholder === null) {
          this._placeholder = newval;
        } else if (this._placeholder !== newval) {
          Util.throwError('Cannot insert a second Placeholder into the visual tree of a Group.');
        }
      } else if (newval instanceof Panel) {
        const plc = newval.findInVisualTree(function (g) { return g instanceof Placeholder; });
        if (plc instanceof Placeholder) {
          if (this._placeholder === null) {
            this._placeholder = plc;
          } else if (this._placeholder !== plc) {
            Util.throwError('Cannot insert a second Placeholder into the visual tree of a Group.');
          }
        }
      }
    } else if (change === ChangedEvent.Remove && propertyname === 'elements') {
      if (this._placeholder !== null) {
        if (oldval === this._placeholder) {
          this._placeholder = null;
        } else if (oldval instanceof Panel && this._placeholder.isContainedBy(oldval)) {
          this._placeholder = null;
        }
      }
    }
    super.raisePartChangedEvent(change, propertyname, obj, oldval, newval, oldparam, newparam);
  }

  // They inherit them from Panel and
  // Groups do not have their own measure and arrange protected methods

  /**
   * @hidden @internal
   * Causes a Group to arrange it's children.
   * @this {Group}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    this._locationObject = this._placeholder;
    super.arrangeProtected(fx, fy, fw, fh);
  }

  /**
   * @hidden @internal
   * @this {Group}
   * @return {boolean}
   */
  public isReadyToMeasureArrange(): boolean {
    if (!super.isReadyToMeasureArrange()) return false;
    const itr = this.memberParts;
    while (itr.next()) {
      const m = itr.value;
      if (m instanceof Node) {
        if (!m.isVisible()) continue;
        if (m.getInvalidArrange()) return false;
      } else if (m instanceof Link) {
        if (!m.isVisible()) continue;
        // OK if the link connects to the group itself
        if (m.getInvalidArrange() && m.fromNode !== this && m.toNode !== this) return false;
      }
    }
    return true;
  }

  /**
   * This read-only property returns a {@link Placeholder} that this group may contain in its visual tree.
   */
  /** @export */ get placeholder(): Placeholder | null { return this._placeholder; }

  /**
   * Gets or sets whether the size of the area of the {@link Group}'s {@link #placeholder}
   * should remain the same during a {@link DraggingTool} move until a drop occurs.
   * Groups within temporary layers (such as new Groups during a drag-copy) are unaffected by this property.
   *
   * In other words, when the value is true, re-computing the bounds of the
   * members is suspended until a drop occurs, at which time the border is recomputed,
   * perhaps not including some members that had been dragged out and reparented.
   * The initial value is false.
   */
  /** @export */ get computesBoundsAfterDrag(): boolean { return (this._nodeFlags & 0x00000800) !== 0; }
  /** @export */ set computesBoundsAfterDrag(value: boolean) {
    const old = (this._nodeFlags & 0x00000800) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Group, 'computesBoundsAfterDrag');
      this._nodeFlags = this._nodeFlags ^ 0x00000800;
      this.raiseChanged('computesBoundsAfterDrag', old, value);
    }
  }

  /**
   * Gets or sets whether a {@link #placeholder}'s bounds includes the bounds of member Links.
   * The default value is true.
   * If this is false, only non-Link member Parts are used to compute the Placeholder's bounds in document coordinates.
   */
  /** @export */ get computesBoundsIncludingLinks(): boolean { return (this._nodeFlags & 0x00001000) !== 0; }
  /** @export */ set computesBoundsIncludingLinks(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Group, 'computesBoundsIncludingLinks');
    const old = (this._nodeFlags & 0x00001000) !== 0;
    if (old !== value) {
      this._nodeFlags = this._nodeFlags ^ 0x00001000;
      this.raiseChanged('computesBoundsIncludingLinks', old, value);
    }
  }

  /**
   * Gets or sets whether a {@link #placeholder}'s bounds includes the previous Group.location.
   * The default value is false.
   */
  /** @export */ get computesBoundsIncludingLocation(): boolean { return (this._nodeFlags & 0x00002000) !== 0; }
  /** @export */ set computesBoundsIncludingLocation(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Group, 'computesBoundsIncludingLocation');
    const old = (this._nodeFlags & 0x00002000) !== 0;
    if (old !== value) {
      this._nodeFlags = this._nodeFlags ^ 0x00002000;
      this.raiseChanged('computesBoundsIncludingLocation', old, value);
    }
  }

  /**
   * Gets or sets whether drag-and-drop events may be bubbled up to this Group if not handled by member Parts.
   * The default value is false -- each Node or Link that is a member of the Group needs to define
   * its own {@link GraphObject#mouseDragEnter}, {@link GraphObject#mouseDragLeave}, and {@link GraphObject#mouseDrop}
   * event handlers if you want dragging/dropping on a member part to act as if the user were acting on the group.
   *
   * This is currently restricted to only call the mouseDragEnter, mouseDragLeave, and mouseDrop event handlers
   * defined on the whole Group, not on any element inside the Group's visual tree.
   * @since 1.5
   */
  /** @export */ get handlesDragDropForMembers(): boolean { return (this._nodeFlags & 0x00004000) !== 0; }
  /** @export */ set handlesDragDropForMembers(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Group, 'handlesDragDropForMembers');
    const old = (this._nodeFlags & 0x00004000) !== 0;
    if (old !== value) {
      this._nodeFlags = this._nodeFlags ^ 0x00004000;
      this.raiseChanged('handlesDragDropForMembers', old, value);
    }
  }

  /**
   * This read-only property returns an iterator over the member {@link Part}s of this Group.
   * Setting {@link Part#containingGroup} to refer to this Group
   * will add that part to this collection.
   * The Parts can be {@link Node}s, {@link Link}s, {@link Group}s, or simple {@link Part}s.
   *
   * A template should not have any member parts.
   */
  /** @export */ get memberParts(): Iterator<Part> { return this._memberParts.iterator; }

  /**
   * @hidden @internal
   * @this {Group}
   * @param {Part} part
   */
  public addMemberInternal(part: Part): void {
    if (this._memberParts.add(part)) {
      if (part instanceof Group) {
        this._nestedGroups.add(part);
      }
      const func = this.memberAdded;
      if (func !== null) {
        let oldUpdatingModel = true;
        const diagram = this.diagram;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, part);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }
      // If the group isn't visible or expanded, the newly added part is made invisible
      if (!this.isVisible() || !this.isSubGraphExpanded) part.madeVisible(false);
    }
    if (part instanceof Link) {
      if (!this.computesBoundsIncludingLinks) return;
    }
    let ph = this._placeholder as GraphObject;
    if (ph === null) ph = this;
    ph.invalidateMeasure();
  }

  /**
   * @hidden @internal
   * @this {Group}
   * @param {Part} part
   */
  public removeMemberInternal(part: Part): void {
    if (this._memberParts.remove(part)) {
      if (part instanceof Group) {
        this._nestedGroups.remove(part);
      }
      const func = this.memberRemoved;
      if (func !== null) {
        let oldUpdatingModel = true;
        const diagram = this.diagram;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, part);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }
      // If the group isn't visible or expanded, the newly removed part is made visible
      if (!this.isVisible() || !this.isSubGraphExpanded) part.madeVisible(true);
    }
    if (part instanceof Link) {
      if (!this.computesBoundsIncludingLinks) return;
    }
    let ph = this._placeholder as GraphObject;
    if (ph === null) ph = this;
    ph.invalidateMeasure();
  }

  /**
   * @hidden @internal
   * @this {Group}
   */
  public deleteCascade(): void {
    // delete all member parts
    if (this._memberParts.count > 0) {
      const diagram = this.diagram;
      if (diagram !== null) {
        const parts = this._memberParts.copy();
        const pit = parts.iterator;
        while (pit.next()) {
          const part = pit.value;
          diagram.remove(part);
        }
      }
    }
    super.deleteCascade();
  }

  /**
   * Gets or sets the {@link Layout} used to position all of the immediate member nodes and links in this group.
   * By default this property is an instance of {@link Layout} -- no special layout is used, which just makes
   * sure each member node has a valid location.
   */
  /** @export */ get layout(): Layout | null { return this._layout; }
  /** @export */ set layout(value: Layout | null) {
    const old = this._layout;
    if (old !== value) {
      if (value !== null) Util.checkClass(value, Layout, Group, 'layout');
      if (old !== null) {
        old.diagram = null;
        old.group = null;
      }
      this._layout = value;
      const diagram = this.diagram;
      if (value !== null) {
        value.diagram = diagram;
        value.group = this;
      }
      if (diagram !== null) {
        diagram._anyLayoutInvalid = true;
      }
      this.raiseChanged('layout', old, value);
      if (diagram !== null) {
        diagram.requestUpdate();
      }
    }
  }

  /**
   * Gets or sets the function that is called after a member {@link Part} has been added to this Group.
   * It is typically used to modify the appearance of the group.
   * The first argument will be this Group.
   * The second argument will be a Part, typically a Node, but may be a simple Part or a Link.
   *
   * If the value is a function, that function must not modify any membership relationships.
   * The member Part has already been added -- trying to remove it or adding or removing another member or the Group itself may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get memberAdded(): ((thisGroup: Group, newPart: Part) => void) | null { return this._memberAdded; }
  /** @export */ set memberAdded(value: ((thisGroup: Group, newPart: Part) => void) | null) {
    const old = this._memberAdded;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Group, 'memberAdded');
      this._memberAdded = value;
      this.raiseChanged('memberAdded', old, value);
    }
  }

  /**
   * Gets or sets the function that is called after a member {@link Part} has been removed from this Group.
   * It is typically used to modify the appearance of the group.
   * The first argument will be this Group.
   * The second argument will be a Part, typically a Node, but may be a simple Part or a Link.
   *
   * If the value is a function, that function must not modify any membership relationships.
   * The member Part has already been removed -- trying to add it or adding or removing another member or the Group itself may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get memberRemoved(): ((thisGroup: Group, oldPart: Part) => void) | null { return this._memberRemoved; }
  /** @export */ set memberRemoved(value: ((thisGroup: Group, oldPart: Part) => void) | null) {
    const old = this._memberRemoved;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Group, 'memberRemoved');
      this._memberRemoved = value;
      this.raiseChanged('memberRemoved', old, value);
    }
  }

  /**
   * Gets or sets the predicate that determines whether or not a Part may become a member of this group.
   * If this is non-null, the predicate is called in addition to any {@link CommandHandler#memberValidation} predicate.
   *
   * The default predicate is null, which is equivalent to simply returning true.
   * The first argument will be this Group.
   * The second argument will be a Part, typically a Node, but will not be a Link or an Adornment.
   *
   * The function, if supplied, must not have any side-effects.
   */
  /** @export */ get memberValidation(): ((thisGroup: Group, part: Part) => boolean) | null { return this._memberValidation; }
  /** @export */ set memberValidation(value: ((thisGroup: Group, part: Part) => boolean) | null) {
    const old = this._memberValidation;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Group, 'memberValidation');
      this._memberValidation = value;
      this.raiseChanged('memberValidation', old, value);
    }
  }

  /**
   * See if the given collection of {@link Part}s contains non-Links all for which
   * {@link CommandHandler#isValidMember} returns true.
   *
   * The CommandHandler predicate will use {@link CommandHandler#memberValidation}
   * and {@link #memberValidation}, if either or both are defined.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Group}
   * @param {Iterable.<Part>} coll
   * @return {boolean} true.
   */
  public canAddMembers(coll: Iterable<Part>): boolean {
    const diagram = this.diagram;
    if (diagram === null) return false;
    const cmd = diagram.commandHandler;
    // get all of the simple Parts and Nodes or Groups, but not Links,
    // that are not members of other parts in the given collection
    const unnested = cmd.findUnnestedNodes(coll);
    const it = unnested.iterator;
    while (it.next()) {
      const node = it.value;
      if (!cmd.isValidMember(this, node)) return false;
    }
    return true;
  }

  /**
   * Add the {@link Part}s in the given collection as members of this Group
   * for those Parts for which {@link CommandHandler#isValidMember} returns true.
   * If the check argument to this method is not supplied or false,
   * this will set {@link Part#containingGroup} on each part unconditionally, not calling {@link CommandHandler#isValidMember}.
   *
   * The CommandHandler predicate will use {@link CommandHandler#memberValidation}
   * and {@link #memberValidation}, if either or both are defined.
   *
   * At this time there is no "removeMembers" method.
   * If you want to make a collection of Parts to be top-level parts, not members of any Group but still in the Diagram, call {@link CommandHandler#addTopLevelParts}.
   * If you want to remove a collection of Parts not only from a Group but from the whole Diagram, call {@link Diagram#removeParts}.
   * @expose
   * @this {Group}
   * @param {Iterable.<Part>} coll
   * @param {boolean=} check whether to call {@link CommandHandler#isValidMember} to confirm that it is valid to add the Part to be a member of this Group.
   * @return {boolean} true if all non-Links were added to this Group; false if some Parts or Nodes were not able to be added.
   */
  public addMembers(coll: Iterable<Part>, check?: boolean): boolean {
    const diagram = this.diagram;
    if (diagram === null) return false;
    const cmd = diagram.commandHandler;
    // get all of the simple Parts and Nodes or Groups, but not Links,
    // that are not members of other parts in the given collection
    const unnested = cmd.findUnnestedNodes(coll);
    let ok = true;
    const it = unnested.iterator;
    while (it.next()) {
      const node = it.value;
      if (!check || cmd.isValidMember(this, node)) {
        node.containingGroup = this;
      } else {
        ok = false;
      }
    }
    return ok;
  }

  /**
   * Gets or sets whether the user may ungroup this group.
   * The initial value is false.
   * @see {@link #canUngroup}
   */
  /** @export */ get ungroupable(): boolean { return (this._nodeFlags & 0x00000100) !== 0; }
  /** @export */ set ungroupable(value: boolean) {
    const old = (this._nodeFlags & 0x00000100) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Group, 'ungroupable');
      this._nodeFlags = this._nodeFlags ^ 0x00000100;
      this.raiseChanged('ungroupable', old, value);
    }
  }

  /**
   * This predicate returns true if {@link #ungroupable} is true,
   * if the layer's {@link Layer#allowUngroup} is true, and
   * if the diagram's {@link Diagram#allowUngroup} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Group}
   * @return {boolean} true if the user may ungroup this object.
   */
  public canUngroup(): boolean {
    if (!this.ungroupable) return false;
    const layer = this.layer;
    if (layer !== null && !layer.allowUngroup) return false;
    const diagram = layer.diagram;
    if (diagram !== null && !diagram.allowUngroup) return false;
    return true;
  }

  /**
   * @hidden @internal
   * This invalidates all links connected to the node recursively
   * @this {Group}
   * @param {Set.<Part>=} ignore
   */
  public invalidateConnectedLinks(ignore?: Set<Part>): void {
    if (ignore === undefined) ignore = null;

    const invalidatedConnectedRoutes = this.getHasInvalidatedConnectedRoutes();
    // can't setHasInvalidateConnectedRoutes until after Node.invalidateConnectedLinks has done its thing
    // handle direct connections with this group, as with any node
    super.invalidateConnectedLinks(ignore);
    if (invalidatedConnectedRoutes) return;
    this.setHasInvalidatedConnectedRoutes(true);

    //     ?? this is wrong: what if a member of the group is not expanded!
    //     But this may cause slowness where there was none before in deeply nested groups
    //     with links going outside the group.
    //  if (this.isSubGraphExpanded) return;

    // also handle indirect connections with this group that are actually to non-visible member nodes
    const lit = this.findExternalLinksConnected();
    while (lit.next()) {
      const link = lit.value;
      // maybe ignore this link:
      if (ignore !== null && ignore.contains(link)) continue;
      // see if connected to some non-visible member -- not this Group
      const from = link.fromNode;
      if (from !== null && from !== this && from.isMemberOf(this) && !from.isVisible()) {
        from.invalidatePortInfo(link.fromPort);
        from.invalidatePortInfo(link.toPort);
        link.invalidateRoute();
      } else {
        const to = link.toNode;
        if (to !== null && to !== this && to.isMemberOf(this) && !to.isVisible()) {
          to.invalidatePortInfo(link.fromPort);
          to.invalidatePortInfo(link.toPort);
          link.invalidateRoute();
        }
      }
    }
  }

  /**
   * Returns an iterator over all of the {@link Link}s that connect with this group or any node contained by this group,
   * in either direction, but that are not internal to this group.
   *
   * Links that are contained by this group (even in nested groups) are not included in the result collection.
   * @export
   * @this {Group}
   * @return {Iterator.<Link>}
   * @see {@link Node#findLinksConnected}
   * @since 1.3
   */
  public findExternalLinksConnected(): Iterator<Link> {
    const allmembers = this.findSubGraphParts();
    allmembers.add(this);  // also iterate over the links connecting directly with this group
    const result = new Set<Link>();
    const nit = allmembers.iterator;
    while (nit.next()) {
      const part = nit.value;
      if (!(part instanceof Node)) continue;  // skip member Links
      const lit = part.linksConnected;
      while (lit.next()) {
        const link = lit.value;
        if (allmembers.contains(link)) continue;  // skip Links that are contained by this group
        result.add(link);
      }
    }
    return result.iterator;
  }

  /**
   * Returns an iterator over all of the {@link Node}s that are connected with this group or any node contained by this group,
   * by a link in either direction, but that are not internal to this group.
   *
   * Nodes that are contained by this group (even in nested groups) are not included in the result collection.
   * However this group itself might be in the results if there is a reflexive link connected to this group.
   * @export
   * @this {Group}
   * @return {Iterator.<Node>}
   * @see {@link Node#findNodesConnected}
   * @since 1.3
   */
  public findExternalNodesConnected(): Iterator<Node> {
    const allmembers = this.findSubGraphParts();
    allmembers.add(this);  // also iterate over the links connecting directly with this group
    const result = new Set<Node>();
    const nit = allmembers.iterator;
    while (nit.next()) {
      const part = nit.value;
      if (!(part instanceof Node)) continue;  // skip member Links
      const lit = part.linksConnected;
      while (lit.next()) {
        const link = lit.value;
        const from = link.fromNode;
        if (!allmembers.contains(from) || from === this) result.add(from);
        const to = link.toNode;
        if (!allmembers.contains(to) || to === this) result.add(to);
      }
    }
    return result.iterator;
  }

  /**
   * @hidden @internal
   * Return a collection of {@link Group}s that are all of the containing groups ({@link Part#containingGroup}) of this group.
   *
   * The resulting collection will include this Group itself.
   * @export
   * @this {Group}
   * @return {Set.<Group>}
   */
  public findContainingGroupChain(): Set<Group> {
    /**
     * @hidden @internal
     * @param {Group} n
     * @param {Set.<Group>} coll
     */
    function addParent(n: Group, coll: Set<Group>) {
      if (n === null) return;
      coll.add(n);
      addParent(n.containingGroup, coll);
    }

    const set = new Set<Group>();
    addParent(this, set);
    return set;
  }

  /**
   * Return a collection of {@link Part}s that are all of the nodes and links
   * that are members of this group, including inside nested groups,
   * but excluding this group itself.
   *
   * For member nodes that are {@link Group}s, this will include its members recursively.
   *
   * If you want only the immediate members of this group, use the {@link #memberParts} property.
   *
   * If you want to find the collection of Nodes and Links that are in the subtree of a given {@link Node}, use {@link Node#findTreeParts}.
   * @export
   * @this {Group}
   * @return {Set.<Part>}
   */
  public findSubGraphParts(): Set<Part> {
    const coll = new Set<Part>();
    Part.gatherCollection(coll, this, true, 0, true);
    coll.remove(this);
    return coll;
  }

  /**
   * @hidden @internal
   * Called when a part's visibility changes, from .visible or other indirect sources (like layer.visible)
   * @this {Group}
   * @param {boolean} val whether the part was just made visible (true) or just made invisible (false)
   */
  public madeVisible(val: boolean): void {
    super.madeVisible(val);
    // This updates all links connected to the node recursively,
    // validating routes if they are invalid, but does not invalidate any links.
    const cit = this.memberParts;
    while (cit.next()) {
      const member = cit.value;
      member.madeVisible(val);
    }
  }

  /**
   * Hide each of the member nodes and links of this group,
   * and recursively collapse any member groups.
   * This changes the value of {@link Part#isVisible} of the whole subgraph and the parts owned by
   * those member nodes and links.
   * However, this group's visibility is unchanged.
   *
   * This sets {@link #isSubGraphExpanded} to false on this group and on all of the nested {@link Group}s.
   * For those nested {@link Group}s that were expanded,
   * {@link #wasSubGraphExpanded} is set to true.
   *
   * To collapse trees made of {@link Node}s and {@link Link}s, use {@link Node#collapseTree}.
   * @export
   * @this {Group}
   */
  public collapseSubGraph(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    if (diagram.isCollapsingExpanding) return;
    diagram.isCollapsingExpanding = true;
    const allmembers = this.findSubGraphParts();
    this._collapseSubGraph1(allmembers, diagram, this);
    diagram.isCollapsingExpanding = false;
  }

  /**
   * @hidden @internal
   * @this {Group}
   * @param {Set.<Part>} allmembers
   * @param {Diagram} diagram
   * @param {Group} expandroot
   */
  private _collapseSubGraph1(allmembers: Set<Part>, diagram: Diagram, expandroot: Group): void {
    const cit = this.memberParts;
    while (cit.next()) {
      const member = cit.value;
      member.madeVisible(false);  // takes care of adornmentsgulp
      if (member instanceof Group) {
        if (member.isSubGraphExpanded) {
          member.wasSubGraphExpanded = member.isSubGraphExpanded;
          member._collapseSubGraph1(allmembers, diagram, expandroot);
        }
      }
      if (member instanceof Node) {
        member.invalidateConnectedLinks(allmembers);
        diagram.addCollapsedNode(member, expandroot);
      } else if (member instanceof Link) {
        const lit = member.labelNodes;
        while (lit.next()) {
          const lab = lit.value;
          lab.invalidateConnectedLinks(allmembers);
        }
      }
    }
    this.isSubGraphExpanded = false;
  }

  /**
   * Show each member node and link,
   * and perhaps recursively expand nested subgraphs.
   * This may change the value of {@link Part#isVisible} of the whole subgraph and the parts owned by
   * those member nodes and links.
   * However, this group's visibility is unchanged.
   *
   * This sets {@link #isSubGraphExpanded} to true on this group and on all of the nested {@link Group}s.
   * This will expand a nested group only if its {@link #wasSubGraphExpanded} property was true.
   *
   * To expand trees made of {@link Node}s and {@link Link}s, use {@link Node#expandTree}.
   * @export
   * @this {Group}
   */
  public expandSubGraph(): void {
    const diagram = this.diagram;
    if (diagram === null) return;
    if (diagram.isCollapsingExpanding) return;
    diagram.isCollapsingExpanding = true;
    const allmembers = this.findSubGraphParts();
    this._expandSubGraph1(allmembers, diagram, this);
    diagram.isCollapsingExpanding = false;
  }

  /**
   * @hidden @internal
   * @this {Group}
   * @param {Set.<Part>} allmembers
   * @param {Diagram} diagram
   * @param {Group} expandroot
   */
  private _expandSubGraph1(allmembers: Set<Part>, diagram: Diagram, expandroot: Group): void {
    const cit = this.memberParts;
    while (cit.next()) {
      const member = cit.value;
      member.madeVisible(true);  // takes care of adornments
      if (member instanceof Group) {
        if (member.wasSubGraphExpanded) {
          member.wasSubGraphExpanded = false;
          member._expandSubGraph1(allmembers, diagram, expandroot);
        }
      }
      if (member instanceof Node) {
        member.invalidateConnectedLinks(allmembers);
        diagram.addExpandedNode(member, expandroot);
      } else if (member instanceof Link) {
        const lit = member.labelNodes;
        while (lit.next()) {
          const lab = lit.value;
          lab.invalidateConnectedLinks(allmembers);
        }
      }
    }
    this.isSubGraphExpanded = true;
  }

  /**
   * Gets or sets whether the subgraph contained by this group is expanded.
   * Changing this property's value will call {@link #collapseSubGraph} or {@link #expandSubGraph},
   * and also will call the value of {@link #subGraphExpandedChanged} if it is a function.
   *
   * The initial value is true -- this group's member parts are shown.
   *
   * There is an analogous property for expanded/collapsed trees of {@link Node}s and {@link Link}s: {@link Node#isTreeExpanded}.
   */
  /** @export */ get isSubGraphExpanded(): boolean { return (this._nodeFlags & 0x00000200) !== 0; }
  /** @export */ set isSubGraphExpanded(value: boolean) {
    const old = (this._nodeFlags & 0x00000200) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Group, 'isSubGraphExpanded');
      this._nodeFlags = this._nodeFlags ^ 0x00000200;
      const diagram = this.diagram;
      this.raiseChanged('isSubGraphExpanded', old, value);
      const func = this.subGraphExpandedChanged;
      if (func !== null) {
        let oldUpdatingModel = true;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }

      if (diagram !== null && diagram.undoManager.isUndoingRedoing) {
        // When not undoing/redoing this is taken care of by madeVisible called from expandSubGraph/collapseSubGraph
        if (this._placeholder !== null) this._placeholder.invalidateMeasure();
        this.memberParts.each(function(p) {
          p.updateAdornments();
        });
        return;
      }
      if (value) {
        this.expandSubGraph();
      } else {
        this.collapseSubGraph();
      }
    }
  }

  /**
   * Gets or sets whether the subgraph starting at this group
   * had been collapsed by a call to {@link #expandSubGraph} on the containing {@link Group}.
   * The initial value is false.
   * @see {@link #isSubGraphExpanded}
   */
  /** @export */ get wasSubGraphExpanded(): boolean { return (this._nodeFlags & 0x00000400) !== 0; }
  /** @export */ set wasSubGraphExpanded(value: boolean) {
    const old = (this._nodeFlags & 0x00000400) !== 0;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Group, 'wasSubGraphExpanded');
      this._nodeFlags = this._nodeFlags ^ 0x00000400;
      this.raiseChanged('wasSubGraphExpanded', old, value);
    }
  }

  /**
   * Gets or sets the function that is called when {@link #isSubGraphExpanded} has changed value.
   * The argument to that function will be this Group.
   *
   * If the value is a function, that function must not expand or collapse any groups.
   * The Group has already been expanded or collapsed -- trying to change it again may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get subGraphExpandedChanged(): ((thisGroup: Group) => void) | null { return this._subGraphExpandedChanged; }
  /** @export */ set subGraphExpandedChanged(value: ((thisGroup: Group) => void) | null) {
    const old = this._subGraphExpandedChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Group, 'subGraphExpandedChanged');
      this._subGraphExpandedChanged = value;
      this.raiseChanged('subGraphExpandedChanged', old, value);
    }
  }

  /**
   * Move this Group and all of its member parts, recursively.
   * @this {Group}
   * @param {Point} newpos a new {@link Point} in document coordinates.
   * @param {boolean=} useLocation true if you want to set the {@link #location} instead of the position. False by default.
   */
  public move(newpos: Point, useLocation?: boolean): void {
    if (useLocation === undefined) useLocation = false;
    const old = useLocation ? this.location : this.position;
    const oldx = old.x;
    const oldy = old.y;
    const newx = newpos.x;
    const newy = newpos.y;
    // handle NaN values to treat NaN==NaN
    if ((oldx === newx || (isNaN(oldx) && isNaN(newx))) &&
        (oldy === newy || (isNaN(oldy) && isNaN(newy)))) return;
    const dx = newx - (isNaN(oldx) ? 0 : oldx);
    const dy = newy - (isNaN(oldy) ? 0 : oldy);
    const temp = Point.alloc();
    super.move(newpos, useLocation);
    const suspendeds: Set<Link> = new Set<Link>();
    // flatten all nested groups, so everything is moved once by the needed distance
    const mems = this.findSubGraphParts();
    const it = mems.iterator;
    while (it.next()) {
      const part = it.value;
      if (!(part instanceof Link)) continue;
      if (part.suspendsRouting) suspendeds.add(part);
      // let normal rerouting happen when Group was resized
      if (!part.validRoute && (part.fromNode === this || part.toNode === this)) continue;
      part.suspendsRouting = true;
    }
    it.reset();
    while (it.next()) {
      const part = it.value;
      if (part.isLinkLike()) continue;
      if (part instanceof Node) {
        if (part.isLinkLabel) continue;
      }
      // The part might not yet be measured if the group is collapsed
      // so offset either the position or location, whichever is real
      const oldp = part.position;
      const oldl = part.location;
      if (oldp.isReal()) {
        temp.x = oldp.x + dx;
        temp.y = oldp.y + dy;
        part.position = temp;
      } else if (oldl.isReal()) {
        temp.x = oldl.x + dx;
        temp.y = oldl.y + dy;
        part.location = temp;
      }
    }
    it.reset();
    while (it.next()) {
      const part = it.value;
      if (!(part instanceof Link)) continue;
      part.suspendsRouting = suspendeds.contains(part);
      // let normal rerouting happen when Group was resized
      if (!part.validRoute && (part.fromNode === this || part.toNode === this)) continue;
      const oldp = part.position; // offset so position or location does not matter
      temp.x = oldp.x + dx;
      temp.y = oldp.y + dy;
      if (temp.isReal()) {
        part.move(temp);  // call Link.move override to call movePoints and move any label nodes
      } else {
        part.invalidateRoute(); // data bad, possibly from a copy, just invalidate instead
      }
      // only need to re-route if the routing style is AvoidsNodes
      if (part.crossesOverNode()) part.invalidateRoute();
    }
    Point.free(temp);
  }

  // used so that Group invalidateConnectedLinks only happens once per group
  // Otherwise N members in a group would cause it to happen N times.

  /**
   * @hidden @internal
   * @this {Group}
   * @return {boolean}
   */
  public getHasInvalidatedConnectedRoutes(): boolean {
    return (this._nodeFlags & 0x00010000) !== 0;
  }

  /**
   * @hidden @internal
   * @this {Group}
   * @param {boolean} val
   */
  public setHasInvalidatedConnectedRoutes(val: boolean) {
    const old = (this._nodeFlags & 0x00010000) !== 0;
    if (old !== val) this._nodeFlags = this._nodeFlags ^ 0x00010000;
  }

  /**
   * @hidden @internal
   */
  /* noexport */get needsLocation() { return (this._nodeFlags & 0x00008000) !== 0; }
  set needsLocation(value) {
    const old = (this._nodeFlags & 0x00008000) !== 0;
    if (old !== value) {
      this._nodeFlags = this._nodeFlags ^ 0x00008000;
    }
  }

} // #end Group
Group['className'] = 'Group';

/**
 * If a Placeholder is in the visual tree of a {@link Group}, it represents the area of all of the member {@link Part}s of that Group.
 * If a Placeholder is in the visual tree of an {@link Adornment}, it represents the area of the {@link Adornment#adornedObject}.
 * Unlike other GraphObjects, the Placeholder measures itself based on these criteria, and not on its own {@link #desiredSize} or {@link #scale}.
 * It will set its own natural bounds and scale, and in the case of Groups will set the Group's location.
 *
 * It can only be used in the visual tree of a Group node or an Adornment.
 * There can be at most one Placeholder in a Group or an Adornment.
 *
 * See the documentation for {@link Group} and {@link Adornment} for descriptions of the Placeholder roles in each.
 * @extends GraphObject
 * @unrestricted
 */
export class Placeholder extends GraphObject {
  /** @hidden @internal */ public static className = 'Placeholder';
  /** @hidden @internal */ private _padding: Margin;
  /** @hidden @internal */ private _savedBounds: Rect;

  /**
   * This normally does not need any initialization,
   * except for maybe setting {@link #padding}.
   */
  constructor() {
    super();
    this._padding = Margin.ZeroMargin;
    this._savedBounds = new Rect(NaN, NaN, NaN, NaN);
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Placeholder.
   * @this {Placeholder}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._padding = this._padding.copyFrozen();
    copy._savedBounds = this._savedBounds.copy();
  }

  /**
   * @hidden @internal
   * We also need to consider the {@link GraphObject#background} and {@link GraphObject#areaBackground} --
   * if both are non-null this will return false.
   * Otherwise this just checks whether the given point is in the rectangle formed
   * by the {@link GraphObject#naturalBounds}.
   * @this {Placeholder}
   * @param {Point} p
   * @return {boolean}
   */
  protected containsPointProtected(p: Point): boolean {
    if (this.background === null && this.areaBackground === null) return false;
    // get the natural size because the point p is already transformed
    const sz = this.naturalBounds;
    return Rect.contains(0, 0, sz.width, sz.height, p.x, p.y);
  }

  /**
   * @hidden @internal
   * @this {Placeholder}
   * @param {number} width
   * @param {number} height
   * @param {number=} minw
   * @param {number=} minh
   */
  protected measureProtected(width: number, height: number, minw?: number, minh?: number): void {
    const part = this.part;
    if (part === null || (!(part instanceof Group) && !(part instanceof Adornment))) {
      Util.throwError('Placeholder is not inside a Group or Adornment.');
    }
    if (part instanceof Group) {
      const r = this.computeBorder(this._savedBounds);
      // New in 2.0, respect minSize:
      const minSize = this.minSize;
      const w = isFinite(minSize.width) ? Math.max(minSize.width, r.width) : r.width;
      const h = isFinite(minSize.height) ? Math.max(minSize.height, r.height) : r.height;

      const nb = this._naturalBounds;
      nb.assignWidthHeight(w || 0, h || 0);
      this._measureRect(0, 0, nb.width, nb.height);

      // don't assign Location if there aren't any visible member nodes
      const itr = part.memberParts;
      let visible = false;
      while (itr.next()) {
        const mem = itr.value;
        if (mem.isVisible()) {
          visible = true;
          break;
        }
      }

      const diagram = part.diagram;
      // can't set the location of the Group yet because the group is not yet done arranging
      // This gets set later on anyway by arrange
      if (visible && diagram !== null && !diagram.animationManager.isAnimatingPosition(part) &&
          !isNaN(r.x) && !isNaN(r.y)) {
        const p = Point.alloc();
        p.setRectSpot(r, part.locationSpot);
        if (!p.equals(part.location)) { // avoid pt allocation
          part.location = new Point(p.x, p.y);
        }
        Point.free(p);
      }/* else if (diagram !== null && diagram.animationManager.isAnimating) {
        diagram.animationManager.recomputeOnCompletion(this);
      }*/
    } else { // Adornment
      const ad = part as Adornment;
      const nb = this._naturalBounds;
      const pad = this._padding;
      const pw = pad.left + pad.right;
      const ph = pad.top + pad.bottom;
      // New in 2.0:
      const obj = ad.adornedObject;
      const angle = obj.getDocumentAngle();
      ad.angle = angle;

      let sw = 0.0;
      if (obj instanceof Shape) {
        sw = obj.strokeWidth;
      }
      const sc = obj.getDocumentScale();
      const ns = obj.naturalBounds;
      const desiredWidth = (ns.width + sw) * sc;
      const desiredHeight = (ns.height + sw) * sc;
      // Don't set this.desiredSize, since this won't measure again if you do. Calling measureRect  after assigning natural Bounds is enough.
      // Links, including adornment links, set their location in their measure phase
      if (ad.type !== Panel.Link) {
        // Selection adornments always use the top-left spot, but other adornments,
        // like resize, use the locationSpot of the adorned object
        const spot = (ad.category === 'Selection') ? Spot.TopLeft : ad.locationSpot;
        const loc =  obj.getDocumentPoint(spot, Point.alloc());
        ad.location = loc;
        Point.free(loc);
      }
      if (!isNaN(desiredWidth) && !isNaN(desiredHeight)) {
        nb.assignWidthHeight((desiredWidth + pw) || 0, (desiredHeight + ph) || 0);
        this._measureRect(-pad.left, -pad.top, nb.width, nb.height);
      } else {
        // Determine size based on adorned object
        const ao = ad.adornedObject;
        const p1 = ao.getDocumentPoint(Spot.TopLeft, Point.alloc());
        const r = Rect.allocAt(p1.x, p1.y, 0, 0);
        r.unionPoint(ao.getDocumentPoint(Spot.BottomRight, p1));
        r.unionPoint(ao.getDocumentPoint(Spot.TopRight, p1));
        r.unionPoint(ao.getDocumentPoint(Spot.BottomLeft, p1));
        nb.assignWidthHeight((r.width + pw) || 0, (r.height + ph) || 0);
        this._measureRect(-pad.left, -pad.top, nb.width, nb.height);

        Point.free(p1);
        Rect.free(r);
      }

    } // end adornment=specifics
  }

  /**
   * @hidden @internal
   * Just set the {@link GraphObject#actualBounds}.
   * @this {Placeholder}
   * @param {number} fx
   * @param {number} fy
   * @param {number} fw
   * @param {number} fh
   */
  protected arrangeProtected(fx: number, fy: number, fw: number, fh: number): void {
    this.actualBounds.assignTo(fx, fy, fw, fh);
  }

  /**
   * @hidden
   * This is only called when the Placeholder is inside a {@link Group}.
   * Normally this just returns the result of {@link #computeMemberBounds} expanded by the {@link #padding}.
   * However, if {@link Group#computesBoundsAfterDrag} is true,
   * and if the {@link Diagram#currentTool} is the {@link DraggingTool},
   * and if this {@link Group} is not being dragged,
   * this method returns the last value of {@link #computeBorder} before dragging began.
   * @expose
   * @this {Placeholder}
   * @param {Rect} result
   * @return {Rect} in document coordinates.
   */
  protected computeBorder(result: Rect): Rect {
    const part = this.part;
    const diagram = part.diagram;
    // Parts in the tool layer (such as groups being drag-copied) should always recompute, even if computesBoundsAfterDrag is true.
    if (diagram !== null && part instanceof Group && !part.layer.isTemporary) {
      if (part.computesBoundsAfterDrag && this._savedBounds.isReal()) {
        const tool = diagram.toolManager.findTool('Dragging') as any;
        if (tool === diagram.currentTool) {
          const border = tool['computeBorder'](part, this._savedBounds, result);
          if (border !== null) return border;
        }
      }
    }


    const temp = Rect.alloc();
    const r = this.computeMemberBounds(temp);
    const pad = this._padding;
    if (part instanceof Group && !part.isSubGraphExpanded) {
      result.assignTo(r.x - pad.left, r.y - pad.top, 0, 0); // zero size when collapsed
    } else {
      result.assignTo(r.x - pad.left, r.y - pad.top,
        Math.max(r.width + pad.left + pad.right, 0),
        Math.max(r.height + pad.top + pad.bottom, 0));
    }
    Rect.free(temp);

    if (part instanceof Group) {
      if (part.computesBoundsIncludingLocation && part.location.isReal()) {
        result.unionPoint(part.location);
      }
    }

    return result;
  }

  /**
   * @hidden
   * Compute the union of the Bounds of this Placeholder's parent's Group.memberParts.
   * If there are no members, this returns a {@link Rect} with Width and Height of zero
   * and an X and Y that are this panel's original location in document coordinates.
   * This is only called when the Placeholder is inside a {@link Group}.
   * @expose
   * @this {Placeholder}
   * @param {Rect} result
   * @return {Rect} in document coordinates.
   */
  protected computeMemberBounds(result: Rect): Rect {
    if (!(this.part instanceof Group)) {
      result.assignTo(0, 0, 0, 0);
      return result;
    }

    const sg = this.part;
    let minx = Infinity;
    let miny = Infinity;
    let maxx = -Infinity;
    let maxy = -Infinity;

    const itr = sg.memberParts;
    while (itr.next()) {
      const m = itr.value;
      if (!m.isVisible()) continue;
      if (m instanceof Link) {
        if (!sg.computesBoundsIncludingLinks) continue;
        if (m.getInvalidMeasure()) continue;  // ???
        // ignore links connecting with this Group
        if (m.fromNode === sg || m.toNode === sg) continue;
      } else {
        // support rotation of whole node
      }
      const b = m.actualBounds; // replaced GetElementBounds
      if (b.left < minx) minx = b.left;
      if (b.top < miny) miny = b.top;
      if (b.right > maxx) maxx = b.right;
      if (b.bottom > maxy) maxy = b.bottom;
    }

    if (!isFinite(minx) || !isFinite(miny)) {
      const loc = sg.location;
      result.assignTo(loc.x, loc.y, 0, 0);
    } else {
      result.assignTo(minx, miny, maxx - minx, maxy - miny);
    }
    return result;
  }

  /**
   * Gets or sets the padding around the members of the {@link Group} or around the {@link Adornment#adornedObject} {@link GraphObject}.
   * The initial value is a {@link Margin} of zero on all sides.
   */
  /** @export */ get padding(): MarginLike { return this._padding; }
  /** @export */ set padding(value: MarginLike) {
    if (typeof value === 'number') {
      value = new Margin(value);
    } else {
      Util.checkClass(value, Margin, Placeholder, 'padding');
    }
    const old = this._padding;
    if (!old.equals(value)) {
      value = value.copyFrozen();
      this._padding = value;
      this.raiseChanged('padding', old, value);
    }
  }
} // #end Placeholder
Placeholder['className'] = 'Placeholder';

/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Link
*/

/**
 * A Link is a {@link Part} that connects {@link Node}s.
 * The link relationship is directional, going from {@link Link#fromNode} to {@link Link#toNode}.
 * A link can connect to a specific port element in a node, as named by the {@link Link#fromPortId}
 * and {@link Link#toPortId} properties.
 * <p class="boxread">
 * For more discussion, see <a href="../../intro/links.html">Introduction to Links</a>.
 *
 * To add a Link to a Diagram when using a {@link GraphLinksModel} you should do something like:
 * ```js
 *   myDiagram.startTransaction("make new link");
 *   myDiagram.model.addLinkData({ from: "Alpha", to: "Beta" });
 *   myDiagram.commitTransaction("make new link");
 * ```
 *
 * where you would substitute the keys of the actual nodes that you want to connect with a link.
 * This will cause a Link to be created (copying the template found in {@link Diagram#linkTemplateMap}),
 * added to the Diagram in some {@link Layer} (based on {@link Part#layerName}), and bound to the link data
 * (resulting in {@link Panel#data} referring to that link data object).
 * Note that link data objects, unlike {@link Node} data, do not have their own unique keys or identifiers,
 * because other Parts do not refer to them.
 *
 * If you are using a {@link TreeModel}, there are no link data objects, so you just need to call
 * {@link TreeModel#setParentKeyForNodeData} to specify the "parent" node's key for a "child" node data.
 *
 * To find a {@link Link} given a link data object in the {@link GraphLinksModel},
 * call {@link Diagram#findLinkForData}.  When using a {@link TreeModel}, call either {@link Diagram#findNodeForData}
 * or {@link Diagram#findNodeForKey} to get a Node, and then call {@link Node#findTreeParentLink} to get the Link, if any exists.
 *
 * To find a link that connects two nodes, call {@link Node#findLinksTo} or {@link Node#findLinksBetween}.
 * With the former method, the direction matters; with the latter method it returns links in either direction.
 *
 * A link's position and size are determined by the two nodes that it connects.
 * Normally there should be a {@link Shape} as the main element in this Link.
 * This shape is what users will see as the "line" or "wire" --
 * you can set its {@link Shape#stroke} and other "stroke..." properties to control its appearance.
 *
 * The link will compute a route (a sequence of points) going from the {@link #fromNode}'s port element
 * to the {@link #toNode}'s port element.
 * That route is used to generate the path of the main shape.
 * Properties that affect the nature of the route and the geometry of the path include:
 *   - {@link #curve}
 *   - {@link #curviness}
 *   - {@link #corner}
 *   - {@link #routing}
 *   - {@link #smoothness}
 *   - {@link #adjusting}
 *
 * <p class="boxread">
 * For more discussion and examples, see <a href="../../intro/links.html">Links</a>.
 *
 * There are additional properties that affect how the end of the link connects to a port element of a node.
 * There are duplicate properties, ones for the "to" end and ones for the "from" end:
 *   - {@link #fromSpot}, {@link #toSpot}
 *   - {@link #fromEndSegmentLength}, {@link #toEndSegmentLength}
 *   - {@link #fromShortLength}, {@link #toShortLength}
 *
 * These properties normally have "default" values, causing the link's routing and path-geometry generating
 * to get the corresponding values from the connected port element.
 * This scheme permits an individual link to have its own specific connection to a port,
 * taking precedence over how the port normally expects links to connect to it.
 * For example, several of the {@link Layout} classes sets these properties on each Link
 * as part of their route computation for links.
 * <p class="boxread">
 * For more discussion and examples, see <a href="../../intro/connectionPoints.html">Link Points</a>.
 *
 * Elements other than the main {@link Shape} in the Link may act as decorations on the link, including arrowheads and labels.
 * You can control where they are located along the link route and how they are oriented.
 * Because these decorations may be any {@link GraphObject}, they are all properties of that class.
 * The properties include:
 *   - {@link GraphObject#segmentIndex}
 *   - {@link GraphObject#segmentFraction}
 *   - {@link GraphObject#segmentOffset}
 *   - {@link GraphObject#segmentOrientation}
 *
 * If you do not set the {@link GraphObject#segmentIndex} property, the object is positioned
 * to be at the middle of the link.
 * <p class="boxread">
 * For more discussion and examples, see <a href="../../intro/linkLabels.html">Link Labels</a>.
 *
 * **GoJS** makes it easy to add arrowheads to your link template.
 * Just add a {@link Shape} with the appearance properties that you want,
 * and also set the {@link Shape#toArrow} or {@link Shape#fromArrow} property
 * to the name of the kind of arrowhead that you want.
 * Doing so automatically sets the "segment..." properties that are appropriate for the chosen arrowhead.
 *
 * More than one shape may automatically get the route geometry.
 * This is useful when you want to have multiple link shapes with different thicknesses
 * to create a gradient effect across the path of the link or to produce parallel lines along the path.
 * Just set {@link GraphObject#isPanelMain} to true on each such {@link Shape}.
 *
 * If you want the user to be able to reconnect a link, using the {@link RelinkingTool},
 * you need to set one or both of {@link #relinkableFrom} and {@link #relinkableTo} to true.
 * The {@link RelinkingTool} shows a {@link RelinkingTool#fromHandleArchetype} and/or a
 * {@link RelinkingTool#toHandleArchetype} when the link is selected.
 * Such a relink handle can be dragged by the user to start a relinking operation.
 *
 * If you want the user to be able to change the path of the link, using the {@link LinkReshapingTool},
 * set {@link Part#reshapable} to true.
 * The {@link LinkReshapingTool} shows reshape handles that the user can drag to shift the position
 * of a point in the link's route.
 * The {@link LinkReshapingTool#handleArchetype} is copied for each reshape handle.
 *
 * Often if a Link is reshapable, you will want to save the route in the model so that it can be restored
 * upon load.  To save the route automatically, add a TwoWay Binding on the {@link #points} property:
 * `new go.Binding("points").makeTwoWay()`.  {@link Model#toJson} will automatically
 * convert the {@link List} of {@link Point}s into an Array of numbers in the JSON representation,
 * if the property is named "points".
 *
 * If a Link is {@link Part#reshapable}, it is also possible to allow the user to add and remove segments
 * from the link's route by setting {@link #resegmentable} to true.
 * This causes the {@link LinkReshapingTool} to add resegmenting handles at the midpoints of each segment.
 * The {@link LinkReshapingTool#midHandleArchetype} is copied for each resegment handle.
 * When the user drags such a resegmenting handle, a new segment is inserted into the route.
 * Also, when the user drags a reshape handle such that two adjacent segments end up in a straight line,
 * a segment is removed from the route.
 * <p class="boxread">
 * For more discussion and examples, see <a href="../../intro/links.html">Links</a>,
 * <a href="../../intro/linkLabels.html">Link Labels</a>, and
 * <a href="../../intro/connectionPoints.html">Link Points</a>.
 * <p class="boxread">
 * To control what links a user may draw or reconnect, please read about
 * <a href="../../intro/validation.html">Validation</a>.
 * <p class="boxread">
 * To customize linking and relinking behavior, please read
 * <a href="../../intro/tools.html#LinkingToolAndRelinkingTool">Introduction to the Linking Tools</a>
 * and <a href="../../intro/tools.html#RelinkingTool">Introduction to the RelinkingTool</a>.
 * For customizing the reshaping of Links,
 * see <a href="../../intro/tools.html#LinkReshapingTool">Introduction to the LinkReshapingTool</a>.
 *
 * Only Links that are in Diagrams can have connections with Nodes.
 * Templates should not be connected with Nodes, be members of Groups, or have any Adornments.
 * @extends Part
 * @unrestricted
 */
export class Link extends Part {
  /** @hidden @internal */ public static className = 'Link';
  /** @hidden @internal */ private _linkFlags: number;
  /** @hidden @internal */ private _fromNode: Node | null;
  /** @hidden @internal */ private _fromPortId: string;
  /** @hidden @internal */ private _fromPortChanged: ((link: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null;
  /** @hidden @internal */ private _toNode: Node | null;
  /** @hidden @internal */ private _toPortId: string;
  /** @hidden @internal */ private _toPortChanged: ((link: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null;
  /** @hidden @internal */ private _adjusting: EnumValue;
  /** @hidden @internal */ private _corner: number;
  /** @hidden @internal */ private _curve: EnumValue;
  /** @hidden @internal */ private _curviness: number;
  /** @hidden @internal */ private _routing: EnumValue;
  /** @hidden @internal */ private _smoothness: number;
  /** @hidden @internal */ private _labelNodes: Set<Node> | null;
  /** @hidden @internal */ private _points: List<Point>;
  /** @hidden @internal */ private _oldPoints: List<Point> | null;
  /** @hidden @internal */ public _savedRoute: List<Point> | null;
  /** @hidden @internal */ private _previousBounds: Rect;
  /** @hidden @internal */ private _effectiveBezier: boolean | null;
  /** @hidden @internal */ private _defaultFromPoint: Point;
  /** @hidden @internal */ private _defaultToPoint: Point;
  /** @hidden @internal */ private _tempOrigin: Point;
  /** @hidden @internal */ private _tempFromNode: Node | null;
  /** @hidden @internal */ private _tempToNode: Node | null;
  /** @hidden @internal */ private _flattenedLengths: Array<number> | null;
  /** @hidden @internal */ private _flattenedTotalLength: number;
  /** @hidden @internal */ public _geometry: Geometry;
  /** @hidden @internal */ public _invalidGeometry: boolean;
  /** @hidden @internal */ public _bundle: LinkBundle | null;
  /** @hidden @internal */ public _childrenBounds: Array<Rect>;

  /**
   * Constructs an empty link that does not connect any nodes.
   * If you want the link to be seen, you must provide a {@link Shape} as an element
   * to be used as the visual path for the link.
   */
  constructor() {
    super(Panel.Link);

    // BIT FLAGS for Link:
    // 0x00000001 this._relinkableFrom = false;
    // 0x00000002 this._relinkableTo = false;
    // 0x00000004 this._resegmentable = false;
    // 0x00000008 this._isTreeLink = true;
    // 0x00000010 this._validRoute = false;
    // 0x00000020 this._suspendsRouting = false;
    // 0x00000040 this._computingPoints = false;
    this._linkFlags = 0x00000008;

    this._fromNode = null;  // related Node
    this._fromPortId = '';
    this._fromPortChanged = null;

    this._toNode = null;  // related Node
    this._toPortId = '';
    this._toPortChanged = null;

    this._adjusting = Link.None;
    this._corner = 0;
    this._curve = Link.None;
    this._curviness = NaN;
    this._routing = Link.Normal;
    this._smoothness = 0.5;

    this._labelNodes = null;  // owned Nodes

    this._points = new List<Point>().freeze();
    this._oldPoints = null;
    this._savedRoute = null; // only used to save routes for animation
    this._previousBounds = new Rect();
    this._geometry = new Geometry();  // shared by "main" Shapes
    this._invalidGeometry = true;
    this._effectiveBezier = null;
    this._bundle = null;
    this._defaultFromPoint = null;
    this._defaultToPoint = null;

    this._childrenBounds = [];

    this._tempOrigin = new Point();

    // There is no guarantee that these are useful unless validRoute is true
    this._tempFromNode = null;
    this._tempToNode = null;

    this._flattenedLengths = null;
    this._flattenedTotalLength = NaN;

    this._linkSettings = null;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Link.
   * @this {Link}
   * @param {GraphObject} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);

    // BIT FLAGS for Link:
    // 0x00000001 copy._relinkableFrom = this._relinkableFrom;
    // 0x00000002 copy._relinkableTo = this._relinkableTo;
    // 0x00000004 copy._resegmentable = this._resegmentable;
    // 0x00000008 copy._isTreeLink = this._isTreeLink;
    // 0x00000010 NOT COPIED: this._validRoute = false;
    // 0x00000020 NOT COPIED: this._suspendsRouting = false;
    // 0x00000040 NOT COPIED: this._computingPoints = false;
    copy._linkFlags = this._linkFlags & ~(0x00000010 | 0x00000020 | 0x00000040);

    // copy._fromNode  The copier should fix this reference
    copy._fromPortId = this._fromPortId;
    copy._fromPortChanged = this._fromPortChanged;

    // copy._toNode  The copier should fix this reference
    copy._toPortId = this._toPortId;
    copy._toPortChanged = this._toPortChanged;

    copy._adjusting = this._adjusting;
    copy._corner = this._corner;
    copy._curve = this._curve;
    copy._curviness = this._curviness;
    copy._routing = this._routing;
    copy._smoothness = this._smoothness;

    // copy._labelNodes  The copier should fix this reference

    if (this._linkSettings !== null) {
      copy._linkSettings = this._linkSettings.copy();
    }

    // These all need to be recomputed on demand:
    // copy._points
    // copy._previousBounds
    // copy._geometry
    // copy._bundle
    // copy._tempFromNode
    // copy._tempToNode
    // copy._flattenedLengths
    // copy._flattenedTotalLength

    // These all should have their default values:
    // copy._oldPoints
    // copy._savedRoute
    // copy._effectiveBezier
    // copy._defaultFromPoint
    // copy._defaultToPoint
    // copy._childrenBounds
    // copy._tempOrigin
  }

  /**
   * @hidden @internal
   * Fix up references in the argument object so that the argument object can get a new visual tree
   * @this {Link}
   * @param {GraphObject} copy  The original Part whose category is being changed by displacing its visual tree with this Part's visual tree.
   */
  protected displaceProtected(copy: this): void {
    super.displaceProtected(copy);
    // hack: need to preserve the portIds
    // So we copy the portIds to the displacing Link so that cloneProtected copies them back
    this._fromPortId = copy._fromPortId;
    this._toPortId = copy._toPortId;

    copy._oldPoints = null;
    copy.invalidateRoute();
    copy._bundle = this._bundle;

    // invalidate portInfo since we need to recompute the route
    const fromPort = copy.fromPort;
    if (fromPort !== null) copy.fromNode.invalidatePortInfo(fromPort);
    const toPort = copy.toPort;
    if (toPort !== null) copy.toNode.invalidatePortInfo(toPort);
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Link}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Link) {
      if (Link.isOrthogonalRouting(arg)) {
        this.routing = arg;
      } else if (arg === Link.Bezier || arg === Link.JumpGap || arg === Link.JumpOver) {
        this.curve = arg;
      } else if (arg === Link.End || arg === Link.Scale || arg === Link.Stretch) {
        this.adjusting = arg;
      } else if (arg === Link.Normal || arg === Link.None) {
        // default values: don't set anything??
      } else {
        Util.throwError('Unknown Link enum value for a Link property: ' + arg);
      }
    } else {
      super.acceptEnum(arg);
    }
  }

  // for routing:
  /**
   * Used as the default value for {@link Link#routing}:
   * the route goes fairly straight between ports.
   * @constant
   */
  public static Normal = new EnumValue(Link, 'Normal', 1);

  /**
   * Used as a value for {@link Link#routing}:
   * each segment is horizontal or vertical.
   * @constant
   */
  public static Orthogonal = new EnumValue(Link, 'Orthogonal', 2);

  /**
   * Used as a value for {@link Link#routing}:
   * each segment is horizontal or vertical, but the route tries to avoid crossing over nodes.
   * @constant
   */
  public static AvoidsNodes = new EnumValue(Link, 'AvoidsNodes', 4 + 2);  // also Orthogonal

  /**
   * @hidden @internal
   * Used as a value for {@link Link#routing}:
   * each segment is horizontal or vertical, but the route tries to avoid crossing over nodes,
   * with the preference for going completely straight if the spot is a side spot.
   * @constant
   */
  public static AvoidsNodesStraight = new EnumValue(Link, 'AvoidsNodesStraight', 4 + 2 + 1);

  /**
   * This is the default value for {@link Link#curve} and {@link Link#adjusting},
   * to indicate that the path geometry consists of straight line segments and
   * to indicate that the link route computation does not depend on any previous route points;
   * this can also be used as a value for {@link GraphObject#segmentOrientation}
   * to indicate that the object is never rotated along the link route -- its angle is unchanged.
   * @constant
   */
  public static None = new EnumValue(Link, 'None', 0);

  // for curve:  None or:
  /**
   * Used as a value for {@link Link#curve}, to indicate that
   * the link path uses Bezier curve segments.
   * @constant
   */
  public static Bezier = new EnumValue(Link, 'Bezier', 9);

  /**
   * Used as a value for {@link Link#curve}, to indicate that
   * orthogonal link segments will be discontinuous where they cross over
   * other orthogonal link segments that have a {@link Link#curve} or JumpOver or JumpGap.
   * @constant
   */
  public static JumpGap = new EnumValue(Link, 'JumpGap', 10);

  /**
   * Used as a value for {@link Link#curve}, to indicate that
   * orthogonal link segments will veer around where they cross over
   * other orthogonal link segments that have a {@link Link#curve} or JumpOver or JumpGap.
   * @constant
   */
  public static JumpOver = new EnumValue(Link, 'JumpOver', 11);

  // for adjusting:  None or:
  /**
   * Used as a value for {@link Link#adjusting},
   * to indicate that the link route computation should keep the
   * intermediate points of the previous route, just modifying the first and/or last points;
   * if the routing is orthogonal, it will only modify the first two and/or last two points.
   * @constant
   */
  public static End = new EnumValue(Link, 'End', 17);

  /**
   * Used as a value for {@link Link#adjusting},
   * to indicate that the link route computation should scale and rotate the intermediate points
   * so that the link's shape looks approximately the same;
   * if the routing is orthogonal, this value is treated as if it were {@link Link.End}.
   * @constant
   */
  public static Scale = new EnumValue(Link, 'Scale', 18);

  /**
   * Used as a value for {@link Link#adjusting},
   * to indicate that the link route computation should linearly interpolate the intermediate points
   * so that the link's shape looks stretched;
   * if the routing is orthogonal, this value is treated as if it were {@link Link.End}.
   * @constant
   */
  public static Stretch = new EnumValue(Link, 'Stretch', 19);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject turned to have the same angle as the route:
   * the GraphObject's angle is always the same as the angle of the link's route
   * at the segment where the GraphObject is attached;
   * use this orientation for arrow heads.
   * @constant
   */
  public static OrientAlong = new EnumValue(Link, 'OrientAlong', 21);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject is turned clockwise to be perpendicular to the route:
   * the GraphObject's angle is always 90 degrees more than the angle of the link's route
   * at the segment where the GraphObject is attached.
   * @constant
   */
  public static OrientPlus90 = new EnumValue(Link, 'OrientPlus90', 22);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject being turned counter-clockwise to be perpendicular to the route:
   * the GraphObject's angle is always 90 degrees less than the angle of the link's route
   * at the segment where the GraphObject is attached.
   * @constant
   */
  public static OrientMinus90 = new EnumValue(Link, 'OrientMinus90', 23);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject's angle always being 180 degrees opposite from the angle of the link's route
   * at the segment where the GraphObject is attached.
   * @constant
   */
  public static OrientOpposite = new EnumValue(Link, 'OrientOpposite', 24);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject turned to have the same angle as the route, just like {@link Link.OrientAlong}, but is never upside down:
   * the GraphObject's angle always following the angle of the link's route
   * at the segment where the GraphObject is attached;
   * this is typically only used for {@link TextBlock}s or {@link Panel}s that contain text.
   * @constant
   */
  public static OrientUpright = new EnumValue(Link, 'OrientUpright', 25);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject turned clockwise to be perpendicular to the route, just like {@link Link.OrientPlus90}, but is never upside down:
   * the GraphObject's angle always being 90 degrees more than the angle of the link's route
   * at the segment where the GraphObject is attached;
   * this is typically only used for {@link TextBlock}s or {@link Panel}s that contain text.
   * @constant
   */
  public static OrientPlus90Upright = new EnumValue(Link, 'OrientPlus90Upright', 26);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject turned counter-clockwise to be perpendicular to the route, just like {@link Link.OrientMinus90}, but is never upside down:
   * the GraphObject's angle always being 90 degrees less than the angle of the link's route
   * at the segment where the GraphObject is attached;
   * this is typically only used for {@link TextBlock}s or {@link Panel}s that contain text.
   * @constant
   */
  public static OrientMinus90Upright = new EnumValue(Link, 'OrientMinus90Upright', 27);

  /**
   * This value for {@link GraphObject#segmentOrientation} results in
   * the GraphObject's angle always following the angle of the link's route
   * at the segment where the GraphObject is attached, but never upside down
   * and never angled more than +/- 45 degrees:
   * when the route's angle is within 45 degrees of vertical (90 or 270 degrees),
   * the GraphObject's angle is set to zero;
   * this is typically only used for {@link TextBlock}s or {@link Panel}s that contain text.
   * @constant
   */
  public static OrientUpright45 = new EnumValue(Link, 'OrientUpright45', 28);

  /**
   * @hidden @internal
   * @this {Link}
   */
  protected initLinkSettings(): void {
    if (this._linkSettings === null) {
      this._linkSettings = new LinkSettings();
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} e
   * @param {number} numpts int
   * @param {List.<Point>} pts
   */
  public _arrangeLabelNode(e: Node, numpts: number, pts: List<Point>): void {
    // get the properties controlling where the "label" should be along the route
    const index = e.segmentIndex;
    const frac = e.segmentFraction;
    let align = e.alignmentFocus;
    const orient = e.segmentOrientation;
    const offset = e.segmentOffset;

    let itemX = 0.0;  // document coordinates
    let itemY = 0.0;
    let labelangle = 0.0;
    let segangle = 0.0;
    if (index < -numpts || index >= numpts) {  // beyond range? assume at the MidPoint, with the MidAngle
      const mid = this.midPoint;  // in document coordinates
      segangle = this.midAngle;
      // maybe rotate the label
      if (orient !== Link.None) {
        labelangle = this.computeAngle(e, orient, segangle);
        e.angle = labelangle;
      }
      itemX = mid.x;
      itemY = mid.y;  // document coordinates
    } else {  // on a particular segment, given by Index, at a point given by Fraction
      // negative index means start from last point, going "backwards"
      let a: Point;
      let b: Point; // points
      let idx = 0;
      if (index >= 0) {
        a = pts._arr[index];
        b = (index < numpts - 1) ? pts._arr[index + 1] : a;
      } else {
        idx = numpts + index;  // remember that index is negative here
        a = pts._arr[idx];
        b = (idx > 0) ? pts._arr[idx - 1] : a;
      }
      // if both of these segment points are the same, check points on either side to determine a direction for the segment
      // this works similar to computeMidAngle
      if (a.equalsApprox(b)) {
        let c: Point; // previous point, if available; otherwise, just use a
        let d: Point; // next point, if available; otherwise, just use b
        if (index >= 0) {
          c = (index > 0) ? pts._arr[index - 1] : a;
          d = (index < numpts - 2) ? pts._arr[index + 2] : b;
        } else {
          c = (idx < numpts - 1) ? pts._arr[idx + 1] : a;
          d = (idx > 1) ? pts._arr[idx - 2] : b;
        }
        const d1 = c.distanceSquaredPoint(a);
        const d2 = b.distanceSquaredPoint(d);
        if (d1 > d2 + 10) {
          segangle = (index >= 0 ? c.directionPoint(a) : a.directionPoint(c));
        } else if (d2 > d1 + 10) {
          segangle = (index >= 0 ? b.directionPoint(d) : d.directionPoint(b));
        } else {
          segangle = (index >= 0 ? c.directionPoint(d) : d.directionPoint(c));
        }
      } else {
        segangle = (index >= 0 ? a.directionPoint(b) : b.directionPoint(a));
      }
      // maybe rotate the label
      if (orient !== Link.None) {
        labelangle = this.computeAngle(e, orient, segangle);
        e.angle = labelangle;
      }
      // compute the fractional point along the line, in document coordinates
      itemX = a.x + (b.x - a.x) * frac;
      itemY = a.y + (b.y - a.y) * frac;
    }

    if (align.isNone()) {  // === Spot.None
      // ignore alignmentFocus & segmentOffset
      e.location = new Point(itemX, itemY);
      return;
    } else if (align.isNoSpot()) {  // a Side Spot or some unspecific Spot
      align = Spot.Center;
    }

    // now determine the label/arrowhead's bounds
    const et = Transform.alloc();
    et.reset();
    et.scale(e.scale, e.scale);
    et.rotate(e.angle, 0, 0);

    // find the transformed alignment point inside the element
    const ns = e.naturalBounds;
    const nb = Rect.allocAt(0, 0, ns.width, ns.height);
    const alignpt = Point.alloc();
    alignpt.setRectSpot(nb, align);
    et.transformPoint(alignpt);
    let dx = -alignpt.x;
    let dy = -alignpt.y;

    // maybe the alignment point is away from the line
    const offsetpt = offset.copy();
    if (isNaN(offsetpt.x)) {
      if (index >= 0) {
        offsetpt.x = alignpt.x + 3;
      } else {
        offsetpt.x = -(alignpt.x + 3);
      }
    }
    if (isNaN(offsetpt.y)) {
      offsetpt.y = -(alignpt.y + 3);
    }
    offsetpt.rotate(segangle);
    itemX += offsetpt.x;
    itemY += offsetpt.y;

    // find the bounds of the transformed rectangle that the alignment spot is in
    et.transformRect(nb);
    dx += nb.x;
    dy += nb.y;
    const epos = Point.allocAt(itemX + dx, itemY + dy);
    e.move(epos);  // don't set position -- gotta call move in case it's a Group
    Point.free(epos);
    Point.free(alignpt);
    Rect.free(nb);
    Transform.free(et);
  }

  /**
   * @hidden @internal
   * A link is ready to be measure/arranged if both of its nodes have been arranged
   * @this {Link}
   * @return {boolean}
   */
  public isReadyToMeasureArrange(): boolean {
    let f = this.fromNode;
    if (f !== null) {
      const vis = f.findVisibleNode();
      if (vis !== null) f = vis;
      if (f.getInvalidArrange() || f.getInvalidPosition()) return false;
    }
    let t = this.toNode;
    if (t !== null) {
      const vis = t.findVisibleNode();
      if (vis !== null) t = vis;
      if (t.getInvalidArrange() || t.getInvalidPosition()) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * This is not called during animation, instead, it calls setPosInplace
   * @this {Link}
   * @param {Point} _val
   * @param {Point} _old
   * @return {boolean} true if successful, false if a no-op (only true calls raisechanged)
   */
  protected setPos(_val: Point, _old: Point): boolean {
    return false;
  }

  /**
   * @hidden @internal
   * Sets the flag for locLastSet, even if the position setter is a no-op
   * @this {Link}
   */
  protected setPositionFlag(): void {
    // no-op, like in GraphObject. The locLastSet flag should not apply to Links,
    // as they are always positioned by their location, which is determined by their points.
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @return {boolean}
   */
  public isNodeLike(): boolean {
    return false;
  }

  /**
   * @hidden @internal
   * Dynamically compute the desired angle of a GraphObject along a segment of the route.
   * This method is only called when the GraphObject's {@link GraphObject#segmentOrientation} property value is
   * not {@link Link.None}.
   * @expose
   * @this {Link}
   * @param {GraphObject} elt The {@link GraphObject} being rotated.
   * @param {EnumValue} orient A {@link Link} constant `Orient...` that indicates how the angle should be computed.
   * @param {number} angle The angle of the segment of the route where the GraphObject is attached.
   * @return {number} The intended angle for the GraphObject, in degrees.
   */
  public computeAngle(elt: GraphObject, orient: EnumValue, angle: number): number {
    return Link.computeAngle(orient, angle);
  }

  /**
   * @hidden @internal
   * Dynamically compute the desired angle of a GraphObject along a segment of the route.
   * This method is only called when the GraphObject's {@link GraphObject#segmentOrientation} property value is
   * not {@link Link.None}.
   * @expose
   * @param {EnumValue} orient A {@link Link} constant `Orient...` that indicates how the angle should be computed.
   * @param {number} angle The angle of the segment of the route where the GraphObject is attached.
   * @return {number} The intended angle for the GraphObject, in degrees.
   */
  public static computeAngle(orient: EnumValue, angle: number): number {
    let a = 0.0;
    switch (orient) {
      default:
      case Link.None: a = 0; break;
      case Link.OrientAlong: a = angle; break;
      case Link.OrientPlus90: a = angle + 90; break;
      case Link.OrientMinus90: a = angle - 90; break;
      case Link.OrientOpposite: a = angle + 180; break;
      case Link.OrientUpright:  // like Along
        a = Geo.normalizeAngle(angle);
        if (a > 90 && a < 270) a -= 180;  // make sure never upside-down
        break;
      case Link.OrientPlus90Upright:  // like Plus90
        a = Geo.normalizeAngle(angle + 90);
        if (a > 90 && a < 270) a -= 180;  // make sure never upside-down
        break;
      case Link.OrientMinus90Upright:  // like Minus90
        a = Geo.normalizeAngle(angle - 90);
        if (a > 90 && a < 270) a -= 180;  // make sure never upside-down
        break;
      case Link.OrientUpright45:  // like Along
        a = Geo.normalizeAngle(angle);
        if (a > 45 && a < 135) return 0;  // make sure never angled too much
        if (a > 225 && a < 315) return 0;
        if (a > 90 && a < 270) a -= 180;  // make sure never upside-down
        break;
    }
    return Geo.normalizeAngle(a);
  }

  /**
   * Gets or sets the {@link Node} that this link comes from.
   * The {@link #fromPortId} specifies which port the link comes from.
   * The default value is null -- this link is not coming from any node.
   *
   * A template should not be connected with any node.
   */
  /** @export */ get fromNode(): Node | null { return this._fromNode; }
  /** @export */ set fromNode(value: Node | null) {
    const old = this._fromNode;
    if (old !== value) {
      if (Debug && value !== null) Util.checkClass(value, Node, Link, 'fromNode');
      const oldport = this.fromPort;
      if (old !== null) {
        if (this._toNode !== old) old.removeLinkInternal(this, oldport);
        this.updateLinkBundleRemove();
        this.invalidateLayout(Part.LayoutRemoved);  // no-op if isUndoingRedoing
      }
      this._fromNode = value;
      if (value !== null) this.madeVisible(value.isVisible()); // maybe made it visible, or definitely made it invisible
      this._effectiveBezier = null;
      this.invalidateRoute();

      // update model
      // ??? no-op if isUndoingRedoing
      const diagram = this.diagram;
      if (diagram !== null) {
        // const data = this.data;
        // if (data !== null) {
          diagram.partManager.setFromNodeForLink(this, value, old);
        // }
      }

      const newport = this.fromPort;
      // different node, thus must be different port: must notify
      const func = this.fromPortChanged;
      if (func !== null) {
        let oldUpdatingModel = true;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, oldport, newport);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }

      if (value !== null) {
        if (this._toNode !== value) value.addLinkInternal(this, newport);
        this.updateLinkBundleAdd();
        this.invalidateLayout(Part.LayoutAdded);  // no-op if isUndoingRedoing
      }
      this.raiseChanged('fromNode', old, value);
      this.updateContainingGroup();
    }
  }

  /**
   * Gets or sets the identifier of the port that this link comes from.
   * The default value is the empty string.
   */
  /** @export */ get fromPortId(): string { return this._fromPortId; }
  /** @export */ set fromPortId(value: string) {
    const old = this._fromPortId;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Link, 'fromPortId');
      const oldport = this.fromPort;
      if (oldport !== null) this.fromNode.invalidatePortInfo(oldport);
      this.updateLinkBundleRemove();
      this._fromPortId = value;
      const newport = this.fromPort;
      if (newport !== null) this.fromNode.invalidatePortInfo(newport);

      // update model
      // ??? no-op if isUndoingRedoing
      const diagram = this.diagram;
      if (diagram !== null) {
        const data = this.data;
        const model = diagram.model;
        if (data !== null) {
          if (model.isLinkDataModel()) {
            model.setFromPortIdForLinkData(data, value);
          }  // do nothing if TreeModel
        }
      }

      // notify if port actually changed:
      if (oldport !== newport) {
        this._effectiveBezier = null;
        this.invalidateRoute();
        const func = this.fromPortChanged;
        if (func !== null) {
          let oldUpdatingModel = true;
          if (diagram !== null) {
            oldUpdatingModel = diagram.isUpdatingModel;
            diagram.isUpdatingModel = true;
          }
          func(this, oldport, newport);
          if (diagram !== null) {
            diagram.isUpdatingModel = oldUpdatingModel;
          }
        }
      }
      this.updateLinkBundleAdd();
      this.raiseChanged('fromPortId', old, value);
    }
  }

  /**
   * This read-only property returns a {@link GraphObject} that is the "from" port that this link is connected from.
   * The {@link #fromNode} provides the node that the link is coming from.
   * The {@link #fromPortId} provides the identifier for which port this link is coming from;
   * you can set that property in order to change the value of this property.
   * This method may return null.
   */
  /** @export */ get fromPort(): GraphObject | null {
    const node = this._fromNode;
    if (node === null) return null;
    const pid = this._fromPortId;
    return node.findPort(pid);
  }

  /**
   * Gets or sets the function that is called after this Link changes which Node or port it connects from.
   * The first argument will be this Link.
   * The second argument will be the old GraphObject port.
   * The third argument will be the new GraphObject port.
   *
   * If the value is a function, that function must not modify which nodes or links this link connects with.
   * The "from" node and/or port has already been changed -- trying to change it again may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get fromPortChanged(): ((thisLink: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null { return this._fromPortChanged; }
  /** @export */ set fromPortChanged(value: ((thisLink: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null) {
    const old = this._fromPortChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Link, 'fromPortChanged');
      this._fromPortChanged = value;
      this.raiseChanged('fromPortChanged', old, value);
    }
  }

  /**
   * Gets or sets the {@link Node} that this link goes to.
   * The {@link #toPortId} specifies which port the link goes to.
   * The default value is null -- this link is not going to any node.
   *
   * A template should not be connected with any node.
   */
  /** @export */ get toNode(): Node | null { return this._toNode; }
  /** @export */ set toNode(value: Node | null) {
    const old = this._toNode;
    if (old !== value) {
      if (Debug && value !== null) Util.checkClass(value, Node, Link, 'toNode');
      const oldport = this.toPort;
      if (old !== null) {
        if (this._fromNode !== old) old.removeLinkInternal(this, oldport);
        this.updateLinkBundleRemove();
        this.invalidateLayout(Part.LayoutRemoved);  // no-op if isUndoingRedoing
      }
      this._toNode = value;
      if (value !== null) this.madeVisible(value.isVisible()); // maybe made it visible, or definitely made it invisible
      this._effectiveBezier = null;
      this.invalidateRoute();

      // update model
      // ??? no-op if isUndoingRedoing
      const diagram = this.diagram;
      if (diagram !== null) {
        // const data = this.data;
        // if (data !== null) {
          diagram.partManager.setToNodeForLink(this, value, old);
        // }
      }

      const newport = this.toPort;
      // different node, thus must be different port: must notify
      const func = this.toPortChanged;
      if (func !== null) {
        let oldUpdatingModel = true;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, oldport, newport);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }

      if (value !== null) {
        if (this._fromNode !== value) value.addLinkInternal(this, newport);
        this.updateLinkBundleAdd();
        this.invalidateLayout(Part.LayoutAdded);  // no-op if isUndoingRedoing
      }
      this.raiseChanged('toNode', old, value);
      this.updateContainingGroup();
    }
  }

  /**
   * Gets or sets the identifier of the port that this link goes to.
   * The default value is the empty string.
   */
  /** @export */ get toPortId(): string { return this._toPortId; }
  /** @export */ set toPortId(value: string) {
    const old = this._toPortId;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'string', Link, 'toPortId');
      const oldport = this.toPort;
      if (oldport !== null) this.toNode.invalidatePortInfo(oldport);
      this.updateLinkBundleRemove();
      this._toPortId = value;
      const newport = this.toPort;
      if (newport !== null) this.toNode.invalidatePortInfo(newport);

      // update model
      // ??? no-op if isUndoingRedoing
      const diagram = this.diagram;
      if (diagram !== null) {
        const data = this.data;
        const model = diagram.model;
        if (data !== null) {
          if (model.isLinkDataModel()) {
            model.setToPortIdForLinkData(data, value);
          }  // do nothing if TreeModel
        }
      }

      // notify if port actually changed:
      if (oldport !== newport) {
        this._effectiveBezier = null;
        this.invalidateRoute();
        const func = this.toPortChanged;
        if (func !== null) {
          let oldUpdatingModel = true;
          if (diagram !== null) {
            oldUpdatingModel = diagram.isUpdatingModel;
            diagram.isUpdatingModel = true;
          }
          func(this, oldport, newport);
          if (diagram !== null) {
            diagram.isUpdatingModel = oldUpdatingModel;
          }

          //// call Node.linkDisconnected and/or Node.linkConnected
          //// if the reconnection is to a different port on the same node
          //// ??? problem: unnecessary events when reconnecting to a different node
          // if (newport !== null && oldport.part === newport.part) {
          //  const node = newport.part as Node;
          //  const func2 = node.linkDisconnected;
          //  if (func2 !== null) {
          //    let oldUpdatingModel = true;
          //    if (diagram !== null) {
          //      oldUpdatingModel = diagram.isUpdatingModel;
          //      diagram.isUpdatingModel = true;
          //    }
          //    func2(node, this, oldport);
          //    if (diagram !== null) {
          //      diagram.isUpdatingModel = oldUpdatingModel;
          //    }
          //  }
          //  const func3 = node.linkConnected;
          //  if (func3 !== null) {
          //    let oldUpdatingModel = true;
          //    if (diagram !== null) {
          //      oldUpdatingModel = diagram.isUpdatingModel;
          //      diagram.isUpdatingModel = true;
          //    }
          //    func3(node, this, newport);
          //    if (diagram !== null) {
          //      diagram.isUpdatingModel = oldUpdatingModel;
          //    }
          //  }
          // }

        }
      }
      this.updateLinkBundleAdd();
      this.raiseChanged('toPortId', old, value);
    }
  }

  /**
   * This read-only property returns a {@link GraphObject} that is the "to" port that this link is connected to.
   * The {@link #toNode} provides the node that the link is going to.
   * The {@link #toPortId} provides the identifier for which port this link is going to;
   * you can set that property in order to change the value of this property.
   * This method may return null.
   */
  /** @export */ get toPort(): GraphObject | null {
    const node = this._toNode;
    if (node === null) return null;
    const pid = this._toPortId;
    return node.findPort(pid);
  }

  /**
   * Gets or sets the function that is called after this Link changes which Node or port it connects to.
   * The first argument will be this Link.
   * The second argument will be the old GraphObject port.
   * The third argument will be the new GraphObject port.
   *
   * If the value is a function, that function must not modify which nodes or links this link connects with.
   * The "to" node and/or port has already been changed -- trying to change it again may produce undefined behavior.
   *
   * The default value is null -- no function is called.
   */
  /** @export */ get toPortChanged(): ((thisLink: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null { return this._toPortChanged; }
  /** @export */ set toPortChanged(value: ((thisLink: Link, oldPort: GraphObject, newPort: GraphObject) => void) | null) {
    const old = this._toPortChanged;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Link, 'toPortChanged');
      this._toPortChanged = value;
      this.raiseChanged('toPortChanged', old, value);
    }
  }

  /// Properties describing how this link should connect to the fromPort

  /**
   * Gets or sets where this link should connect at the {@link #fromPort}.
   * The default value is {@link Spot.Default}, meaning that the value
   * actually comes from the {@link GraphObject#fromSpot} property of the {@link #fromPort}.
   *
   * This property is set by some layouts, if {@link ForceDirectedLayout#setsPortSpots} or
   * {@link LayeredDigraphLayout#setsPortSpots} or {@link TreeLayout#setsPortSpot} or
   * {@link TreeLayout#setsChildPortSpot} is true.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * @see {@link #toSpot}
   * @see {@link GraphObject#fromSpot}
   * @see {@link #computeSpot}
   */
  /** @export */ get fromSpot(): Spot { return this._linkSettings !== null ? this._linkSettings._fromSpot : Spot.Default; }
  /** @export */ set fromSpot(value: Spot) {
    this.initLinkSettings();
    const old = this._linkSettings._fromSpot;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, Link, 'fromSpot');
      value = value.copyFrozen();
      this._linkSettings._fromSpot = value;
      this.raiseChanged('fromSpot', old, value);
      this.invalidateRoute();
    }
  }

  /**
   * Gets or sets the length of the first segment, when the computed "from spot" at the {@link #fromPort} is not {@link Spot.None}.
   * The default value is NaN, meaning that the value
   * actually comes from the {@link GraphObject#fromEndSegmentLength} property of the {@link #fromPort}.
   * This value also limits how short the {@link #fromShortLength} may be drawn.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * @see {@link #toEndSegmentLength}
   * @see {@link #computeEndSegmentLength}
   * @see {@link GraphObject#fromEndSegmentLength}
   * @see {@link #fromShortLength}
   */
  /** @export */ get fromEndSegmentLength(): number { return this._linkSettings !== null ? this._linkSettings._fromEndSegmentLength : NaN; }
  /** @export */ set fromEndSegmentLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._fromEndSegmentLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'fromEndSegmentLength');
      if (value < 0) Util.throwRangeError(value, '>= 0', Link, 'fromEndSegmentLength');
      this._linkSettings._fromEndSegmentLength = value;
      this.raiseChanged('fromEndSegmentLength', old, value);
      this.invalidateRoute();
    }
  }

  /**
   * Gets or sets how far the end segment stops short of the actual port.
   * Positive values are limited by the {@link #fromEndSegmentLength} or {@link GraphObject#fromEndSegmentLength}.
   * Negative values cause the link to extend into the port.
   * The default value is NaN -- the value actually comes from the
   * {@link GraphObject#fromShortLength} property of the {@link #fromPort}.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * If you want to move an arrowhead away from the start of the link route,
   * set the arrowhead Shape's {@link GraphObject#segmentOffset}.
   * @see {@link #toShortLength}
   * @see {@link GraphObject#fromShortLength}
   * @see {@link #fromEndSegmentLength}
   */
  /** @export */ get fromShortLength(): number { return this._linkSettings !== null ? this._linkSettings._fromShortLength : NaN; }
  /** @export */ set fromShortLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._fromShortLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'fromShortLength');
      this._linkSettings._fromShortLength = value;
      this.raiseChanged('fromShortLength', old, value);
      this.invalidateRoute();
      this.invalidateGeometry(); // so that even if the points remain the same, the geometry gets updated
    }
  }

  /// Properties describing how this link should connect to the toPort

  /**
   * Gets or sets where this link should connect at the {@link #toPort}.
   * The default value is {@link Spot.Default}, meaning that the value
   * actually comes from the {@link GraphObject#toSpot} property of the {@link #toPort}.
   *
   * This property is set by some layouts, if {@link ForceDirectedLayout#setsPortSpots} or
   * {@link LayeredDigraphLayout#setsPortSpots} or {@link TreeLayout#setsPortSpot} or
   * {@link TreeLayout#setsChildPortSpot} is true.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * @see {@link #fromSpot}
   * @see {@link GraphObject#toSpot}
   * @see {@link #computeSpot}
   */
  /** @export */ get toSpot(): Spot { return this._linkSettings !== null ? this._linkSettings._toSpot : Spot.Default; }
  /** @export */ set toSpot(value: Spot) {
    this.initLinkSettings();
    const old = this._linkSettings._toSpot;
    if (!old.equals(value)) {
      if (Debug) Util.checkClass(value, Spot, Link, 'toSpot');
      value = value.copyFrozen();
      this._linkSettings._toSpot = value;
      this.raiseChanged('toSpot', old, value);
      this.invalidateRoute();
    }
  }

  /**
   * Gets or sets the length of the last segment.
   * The default value is NaN, meaning that the value
   * actually comes from the {@link GraphObject#toEndSegmentLength} property of the {@link #toPort}.
   * This value also limits how short the {@link #toShortLength} may be drawn.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * @see {@link #fromEndSegmentLength}
   * @see {@link #computeEndSegmentLength}
   * @see {@link GraphObject#toEndSegmentLength}
   * @see {@link #toShortLength}
   */
  /** @export */ get toEndSegmentLength(): number { return this._linkSettings !== null ? this._linkSettings._toEndSegmentLength : NaN; }
  /** @export */ set toEndSegmentLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._toEndSegmentLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'toEndSegmentLength');
      if (value < 0) Util.throwRangeError(value, '>= 0', Link, 'toEndSegmentLength');
      this._linkSettings._toEndSegmentLength = value;
      this.raiseChanged('toEndSegmentLength', old, value);
      this.invalidateRoute();
    }
  }

  /**
   * Gets or sets how far the end segment stops short of the actual port.
   * Positive values are limited by the {@link #toEndSegmentLength} or {@link GraphObject#toEndSegmentLength}.
   * Negative values cause the link to extend into the port.
   * The default value is NaN -- the value actually comes from the
   * {@link GraphObject#toShortLength} property of the {@link #toPort}.
   *
   * For examples of how to use this property, see <a href="../../intro/connectionPoints.html">Link Connection Points</a>.
   * If you want to move an arrowhead away from the end of the link route,
   * set the arrowhead Shape's {@link GraphObject#segmentOffset}.
   * @see {@link #fromShortLength}
   * @see {@link GraphObject#toShortLength}
   * @see {@link #toEndSegmentLength}
   */
  /** @export */ get toShortLength(): number { return this._linkSettings !== null ? this._linkSettings._toShortLength : NaN; }
  /** @export */ set toShortLength(value: number) {
    this.initLinkSettings();
    const old = this._linkSettings._toShortLength;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'toShortLength');
      this._linkSettings._toShortLength = value;
      this.raiseChanged('toShortLength', old, value);
      this.invalidateRoute();
      this.invalidateGeometry(); // so that even if the points remain the same, the geometry gets updated
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public updateContainingGroup(): void {
    const fromnode = this.fromNode;
    const tonode = this.toNode;
    // Don't set the Part.containingGroup property:
    // avoid Link check and raiseChanged by calling the internal set method instead.
    let container = null;
    if (fromnode !== null) {
      if (tonode !== null) {  // both fromnode and tonode
        // treat a link connecting a Group with a member Node as belonging to that Group
        container = fromnode.findCommonContainingGroup(tonode);
      } else {  // fromnode but no tonode
        container = fromnode.containingGroup;
      }
    } else if (tonode !== null) {  // tonode but no fromnode
      container = tonode.containingGroup;
    } else {  // neither fromnode nor tonode
      container = null;
    }
    this.setContainingGroup(container);
    // new in 1.7.17: Group.memberParts includes label nodes
    if (this.isLabeledLink) {
      const lit = this.labelNodes;
      while (lit.next()) {
        const lab = lit.value;
        lab.containingGroup = container;
      }
    }
  }

  /**
   * @hidden @internal
   * Just used by Link.updateContainingGroup(), because Link.containingGroup is always computed, never set.
   * @this {Link}
   * @param {Group} val
   */
  public setContainingGroup(val: Group): void {
    const old = this._containingGroup;
    if (old !== val) {
      if (old !== null) old.removeMemberInternal(this);
      this._containingGroup = val;
      if (val !== null) val.addMemberInternal(this);
      const func = this.containingGroupChanged;
      if (func !== null) {
        let oldUpdatingModel = true;
        const diagram = this.diagram;
        if (diagram !== null) {
          oldUpdatingModel = diagram.isUpdatingModel;
          diagram.isUpdatingModel = true;
        }
        func(this, old, val);
        if (diagram !== null) {
          diagram.isUpdatingModel = oldUpdatingModel;
        }
      }
      // If the route was valid this link may have been pointing at its old group,
      // and that to/from node may no longer be a valid target
      if (this.validRoute && (this._tempFromNode === old || this._tempToNode === old)) {
        this.invalidateRoute();
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public invalidateContainingGroup(): void {
    // don't do the invalidation if this Link connects with the containing Group
    const group = this.containingGroup;
    if (group !== null && this.fromNode !== group && this.toNode !== group && group.computesBoundsIncludingLinks) {
      super.invalidateContainingGroup();
    }
  }

  /**
   * Given a {@link Node}, return the node at the other end of this link.
   * @export
   * @this {Link}
   * @param {Node} node
   * @return {Node} This may return the same node, if the link is reflexive.
   */
  public getOtherNode(node: Node): Node | null {
    if (Debug) Util.checkClass(node, Node, Link, 'getOtherNode:node');
    const from = this.fromNode;
    if (node === from) {
      return this.toNode;
    } else {
      return from;
    }
  }

  /**
   * Given a {@link GraphObject} that is a "port", return the port at the other end of this link.
   * @export
   * @this {Link}
   * @param {GraphObject} port
   * @return {GraphObject} This may return the same object, if the link is reflexive.
   */
  public getOtherPort(port: GraphObject): GraphObject | null {
    if (Debug) Util.checkClass(port, GraphObject, Link, 'getOtherPort:port');
    const from = this.fromPort;
    if (port === from) {
      return this.toPort;
    } else {
      return from;
    }
  }

  /**
   * This read-only property is true when this Link has any label Nodes, {@link Node}s that are owned by this Link
   * and are arranged along its path in the same manner as elements of the Link Panel.
   * @see {@link #labelNodes}
   * @since 1.1
   */
  /** @export */ get isLabeledLink(): boolean {
    if (this._labelNodes === null) return false;
    return this._labelNodes.count > 0;
  }

  /**
   * This read-only property returns an iterator over the {@link Node}s that act as labels on this Link.
   * Setting {@link Node#labeledLink} to refer to this Link
   * will add that {@link Node} to this collection.
   *
   * Do not confuse these {@link Node}s with the {@link GraphObject}s that
   * are part of this {@link Link} {@link Panel}.
   * Those objects can also be used as "labels", but not as objects to which
   * there can be link connections, because those panel elements cannot be {@link Node}s.
   * Those panel objects can be used for other purposes too, such as arrowheads.
   *
   * A template should not have any label nodes.
   * @see {@link #isLabeledLink}
   */
  /** @export */ get labelNodes(): Iterator<Node> {
    if (this._labelNodes === null) return EmptyIterator.instance;
    return this._labelNodes.iterator;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} node
   */
  public addLabelNodeInternal(node: Node): void {
    if (this._labelNodes === null) this._labelNodes = new Set<Node>();
    this._labelNodes.add(node);
    this.invalidateMeasure();
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} node
   */
  public removeLabelNodeInternal(node: Node): void {
    if (this._labelNodes === null) return;
    this._labelNodes.remove(node);
    this.invalidateMeasure();
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {boolean} switchingLayers
   */
  public internalAddProtected(switchingLayers: boolean): void {
    super.internalAddProtected(switchingLayers);

    if (this._jumpsOver()) {
      this.invalidateOtherJumpOvers(this.actualBounds);
    }

    if (!switchingLayers) {
      const from = this._fromNode;
      let fromport = null;
      if (from !== null) {
        fromport = this.fromPort;
        from.addLinkInternal(this, fromport);
      }
      const to = this._toNode;
      let toport = null;
      if (to !== null) {
        toport = this.toPort;
        // avoid duplicate calls to Node.linkConnected
        if (to !== from || toport !== fromport) {
          to.addLinkInternal(this, toport);
        }
      }
      this.updateLinkBundleAdd();
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {boolean} switchingLayers
   */
  public internalRemoveProtected(switchingLayers: boolean): void {
    super.internalRemoveProtected(switchingLayers);

    if (this._jumpsOver()) {
      this.invalidateOtherJumpOvers(this.actualBounds);
    }

    if (!switchingLayers) {
      const from = this._fromNode;
      let fromport = null;
      // do the following instead of: this.fromNode = null;
      if (from !== null) {
        fromport = this.fromPort;
        from.removeLinkInternal(this, fromport);
      }
      const to = this._toNode;
      let toport = null;
      // do the following instead of: this.toNode = null;
      if (to !== null) {
        toport = this.toPort;
        // avoid duplicate calls to Node.linkDisconnected
        if (to !== from || toport !== fromport) {
          to.removeLinkInternal(this, toport);
        }
      }
      this.updateLinkBundleRemove();
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public deleteCascade(): void {
    // pretend the route doesn't need to be recalculated, to avoid potential unnecessary
    // route recalculations which might produce bad results due to removed references
    this.validRoute = true;

    // delete any label nodes
    if (this._labelNodes !== null) {
      const diagram = this.diagram;
      if (diagram !== null) {
        const nodes = this._labelNodes.copy();
        const nit = nodes.iterator;
        while (nit.next()) {
          const node = nit.value;
          diagram.remove(node);
        }
      }
    }

    // don't call the base method!
    // super.deleteCascade();
    // handle link data differently:
    const data = this.data;
    if (data !== null) {
      const diagram = this.diagram;
      if (diagram !== null) {
        diagram.partManager.removeDataForLink(this);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public updateRelationshipsFromData(): void {
    // super.updateRelationshipsFromData();  // calling base method is not needed
    const data = this.data;
    if (data === null) return;
    const diagram = this.diagram;
    if (diagram === null) return;
    diagram.partManager.updateRelationshipsFromData(this);
  }

  /**
   * Move this link to a new position.
   * This also shifts all of the Points in the route accordingly.
   * This also moves any {@link #labelNodes}.
   * @this {Link}
   * @param {Point} newpos a new {@link Point} in document coordinates.
   * @param {boolean=} useLocation true if you want to set the {@link #location} instead of the position. False by default.
   */
  public move(newpos: Point, useLocation?: boolean): void {
    const oldpos = useLocation ? this.location : this.position;
    let oldx = oldpos.x;
    if (isNaN(oldx)) oldx = 0;
    let oldy = oldpos.y;
    if (isNaN(oldy)) oldy = 0;
    const dx = newpos.x - oldx;
    const dy = newpos.y - oldy;
    if (useLocation === true) {
      super.move(newpos, false);
    } else {
      // cannot set Link.location, make an offset position instead
      const newloc = Point.allocAt(oldpos.x + dx, oldpos.y + dy);
      super.move(newloc, false);
      Point.free(newloc);
    }
    this.movePoints(dx, dy);
    const lit = this.labelNodes;
    while (lit.next()) {
      const lab = lit.value;
      const oldp = lab.position;
      lab.moveTo(oldp.x + dx, oldp.y + dy);
    }
  }

  /**
   * Gets or sets whether the user may reconnect an existing link at the "from" end.
   * This affects the behavior of the {@link RelinkingTool}.
   *
   * The initial value is false.
   */
  /** @export */ get relinkableFrom(): boolean { return (this._linkFlags & 0x00000001) !== 0; }
  /** @export */ set relinkableFrom(value: boolean) {
    const old = (this._linkFlags & 0x00000001) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Link, 'relinkableFrom');
      this._linkFlags = this._linkFlags ^ 0x00000001;
      this.raiseChanged('relinkableFrom', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether the user may reconnect an existing link at the "to" end.
   * This affects the behavior of the {@link RelinkingTool}.
   *
   * The initial value is false.
   */
  /** @export */ get relinkableTo(): boolean { return (this._linkFlags & 0x00000002) !== 0; }
  /** @export */ set relinkableTo(value: boolean) {
    const old = (this._linkFlags & 0x00000002) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Link, 'relinkableTo');
      this._linkFlags = this._linkFlags ^ 0x00000002;
      this.raiseChanged('relinkableTo', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * This predicate returns true if {@link #relinkableFrom} is true,
   * if the layer's {@link Layer#allowRelink} is true, and
   * if the diagram's {@link Diagram#allowRelink} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Link}
   * @return {boolean} true if the user may relink the From end of this link.
   */
  public canRelinkFrom(): boolean {
    if (!this.relinkableFrom) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowRelink) return false;
    const diagram = layer.diagram;
    if (diagram !== null && !diagram.allowRelink) return false;
    return true;
  }

  /**
   * This predicate returns true if {@link #relinkableTo} is true,
   * if the layer's {@link Layer#allowRelink} is true, and
   * if the diagram's {@link Diagram#allowRelink} is true.
   *
   * This does not check {@link Diagram#isReadOnly} or {@link Model#isReadOnly},
   * but commands and tools should check those properties.
   * @expose
   * @this {Link}
   * @return {boolean} true if the user may relink the To end of this link.
   */
  public canRelinkTo(): boolean {
    if (!this.relinkableTo) return false;
    const layer = this.layer;
    if (layer === null) return true;
    if (!layer.allowRelink) return false;
    const diagram = layer.diagram;
    if (diagram !== null && !diagram.allowRelink) return false;
    return true;
  }

  /**
   * Gets or sets whether the user may change the number of segments in this Link,
   * if the link has straight segments.
   * This affects the behavior of the {@link LinkReshapingTool} when {@link #curve} is not {@link Link.Bezier}.
   *
   * The initial value is false.
   * @since 1.2
   */
  /** @export */ get resegmentable(): boolean { return (this._linkFlags & 0x00000004) !== 0; }
  /** @export */ set resegmentable(value: boolean) {
    const old = (this._linkFlags & 0x00000004) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Link, 'resegmentable');
      this._linkFlags = this._linkFlags ^ 0x00000004;
      this.raiseChanged('resegmentable', old, value);
      this.invalidateAdornments();
    }
  }

  /**
   * Gets or sets whether this Link is part of the tree for tree operations
   * such as {@link Node#findTreeChildrenNodes} or {@link Node#collapseTree}.
   *
   * The initial value is true.
   * @since 1.2
   */
  /** @export */ get isTreeLink(): boolean { return (this._linkFlags & 0x00000008) !== 0; }
  /** @export */ set isTreeLink(value: boolean) {
    const old = (this._linkFlags & 0x00000008) !== 0;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'boolean', Link, 'isTreeLink');
      this._linkFlags = this._linkFlags ^ 0x00000008;
      this.raiseChanged('isTreeLink', old, value);
      if (this.fromNode !== null) this.fromNode.updateTreeLeaf();
      if (this.toNode !== null) this.toNode.updateTreeLeaf();
    }
  }

  /**
   * This read-only property returns the main {@link Shape} representing the path of this Link.
   *
   * The value depends on there being an element in this Link that is
   * a {@link Shape} and that has {@link GraphObject#isPanelMain} set to true.
   * This shape is what displays the stroke going from one node to another node.
   *
   * This path shape's geometry is automatically generated based on the {@link #points}
   * of this link's route, depending on other properties such as {@link #curve}.
   * Note that the points are in document coordinates, but the path's {@link Shape#geometry}
   * will have points that are in local Shape coordinates.
   * This Shape's {@link GraphObject#scale} should be one and {@link GraphObject#angle} should be zero.
   *
   * If there is no Shape with {@link GraphObject#isPanelMain} set to true,
   * the Link will use the first Shape for its path.
   */
  /** @export */ get path(): Shape | null {
    const main = this.findMainElement();
    if (main instanceof Shape) return main;
    return null;
  }

  //   * This read-only property returns the bounds of the link geometry in document coordinates.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get routeBounds(): Rect {
    // ??? cache the result
    this.updateRoute();
    return this.computeRouteBounds(this._previousBounds);
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Rect} result
   * @return {Rect} in document coordinates
   */
  private computeRouteBounds(result: Rect): Rect {
    let bx = Infinity;
    let by = Infinity;
    const numpts = this.pointsCount;
    if (numpts === 0) {
      result.assignTo(NaN, NaN, 0, 0);  // no location means call UpdatePoints/ComputeRouteBounds again
      return result;
    } else if (numpts === 1) {
      const p = this.getPoint(0);
      bx = Math.min(p.x, bx);
      by = Math.min(p.y, by);
      result.assignTo(p.x, p.y, 0, 0);
    } else if (numpts === 2) {
      const p0 = this.getPoint(0);
      const p1 = this.getPoint(1);
      bx = Math.min(p0.x, p1.x);
      by = Math.min(p0.y, p1.y);
      result.assignTo(p0.x, p0.y, 0, 0);
      result.unionPoint(p1);
    } else {
      if (this.computeCurve() === Link.Bezier && numpts >= 3 && !this.isOrthogonal) {
        let start = this.getPoint(0);
        bx = start.x;
        by = start.y;
        result.assignTo(bx, by, 0, 0);
        if (numpts === 3) {
          const ctrl = this.getPoint(1);
          bx = Math.min(ctrl.x, bx);
          by = Math.min(ctrl.y, by);
          const end = this.getPoint(2);
          bx = Math.min(end.x, bx);
          by = Math.min(end.y, by);
          Geo.bezierBounds(start.x, start.y, ctrl.x, ctrl.y, ctrl.x, ctrl.y, end.x, end.y, 0.5, result);
        } else {
          for (let i = 3; i < numpts; i += 3) {
            const startControl = this.getPoint(i - 2);
            // if it's the last segment, use the last two points
            if (i + 3 >= numpts) {
              i = numpts - 1;
            }
            const endControl = this.getPoint(i - 1);
            const end = this.getPoint(i);
            Geo.bezierBounds(start.x, start.y, startControl.x, startControl.y, endControl.x, endControl.y, end.x, end.y, 0.5, result);  // ?? epsilon
            bx = Math.min(end.x, bx);
            by = Math.min(end.y, by);
            start = end;
          }
        }
      } else {
        const p0 = this.getPoint(0);
        const p1 = this.getPoint(1);
        bx = Math.min(p0.x, p1.x);
        by = Math.min(p0.y, p1.y);
        result.assignTo(p0.x, p0.y, 0, 0);
        result.unionPoint(p1);
        for (let i = 2; i < numpts; i++) {
          const pt = this.getPoint(i);
          bx = Math.min(pt.x, bx);
          by = Math.min(pt.y, by);
          result.unionPoint(pt);
        }

      }
    }
    this._tempOrigin.assignTo(bx - result.x, by - result.y);
    return result;
  }

  /**
   * This read-only property returns the point at the middle of the path, in document coordinates.
   */
  /** @export */ get midPoint(): Point {
    this.updateRoute();
    return this.computeMidPoint(new Point());  // ??? cache value
  }

  /**
   * @hidden @internal
   * Compute the point at the middle of the route.
   * @expose
   * @this {Link}
   * @param {Point} result
   * @return {Point} in document coordinates
   */
  public computeMidPoint(result: Point): Point {
    const numpts = this.pointsCount;
    if (numpts === 0) {  // no point means call UpdatePoints/ComputeMidPoint again
      result.assign(Point.NoPoint);
      return result;
    } else if (numpts === 1) {
      result.assign(this.getPoint(0));
      return result;
    } else if (numpts === 2) {
      const a = this.getPoint(0);
      const b = this.getPoint(1);
      result.assignTo((a.x + b.x) / 2, (a.y + b.y) / 2);
      return result;
    }

    if (this.isOrthogonal && (this.computeCorner() >= 15 || this.computeCurve() === Link.Bezier)) {
      // Use geometry to get point, then go from path geometry coordinates to document coodinates
      // We cannot use getDocumentPoint because this is called to find the coordinates of labels during the link measure
      this._geometry.getPointAlongPath(.5, result); // assigns result
      result.add(this.getPoint(0));
      const fig = this._geometry.figures.first();
      result.offset(-fig.startX, -fig.startY);
      return result;
    }
    if (this.computeCurve() === Link.Bezier) {
      if (numpts === 3) return this.getPoint(1);
      const numsegs = ((numpts - 1) / 3) | 0;
      const idx = ((numsegs / 2) | 0) * 3;
      if (numsegs % 2 === 1) {
        const a = this.getPoint(idx);
        const b = this.getPoint(idx + 1);
        const c = this.getPoint(idx + 2);
        const d = this.getPoint(idx + 3);
        Geo.bezierMidPoint(a.x, a.y, b.x, b.y, c.x, c.y, d.x, d.y, result);
        return result;
      } else {
        result.assign(this.getPoint(idx));
        return result;
      }
    }

    const segmentDists = this.flattenedLengths;
    const totalDist = this.flattenedTotalLength;

    let currentDistance = 0.0;
    let currentPointIndex = 0;
    let nextSegmentLength = 0.0;
    while (currentDistance < totalDist / 2 && currentPointIndex < numpts) {
      nextSegmentLength = segmentDists[currentPointIndex];
      if (currentDistance + nextSegmentLength > totalDist / 2) break;
      currentDistance += nextSegmentLength;
      currentPointIndex++;
    }
    const currentPoint = this.getPoint(currentPointIndex);
    const nextPoint = this.getPoint(currentPointIndex + 1);

    // Handle Orthogonal cases first
    if (Math.abs(currentPoint.x - nextPoint.x) < 1) {
      if (currentPoint.y > nextPoint.y) {
        result.assignTo(currentPoint.x, currentPoint.y - (totalDist / 2 - currentDistance));
      } else {
        result.assignTo(currentPoint.x, currentPoint.y + (totalDist / 2 - currentDistance));
      }
    } else if (Math.abs(currentPoint.y - nextPoint.y) < 1) {
      if (currentPoint.x > nextPoint.x) {
        result.assignTo(currentPoint.x - (totalDist / 2 - currentDistance), currentPoint.y);
      } else {
        result.assignTo(currentPoint.x + (totalDist / 2 - currentDistance), currentPoint.y);
      }
    } else { // Handle non-Orthogonal lines
      const similarTriangleRatio = (totalDist / 2 - currentDistance) / nextSegmentLength;
      const dx = similarTriangleRatio * (nextPoint.x - currentPoint.x);
      const dy = similarTriangleRatio * (nextPoint.y - currentPoint.y);

      result.assignTo(currentPoint.x + dx, currentPoint.y + dy);
    }
    return result;
  }

  /**
   * This read-only property returns the angle of the path at the {@link #midPoint}.
   */
  /** @export */ get midAngle(): number {
    this.updateRoute();
    return this.computeMidAngle();  // ??? cache value
  }

  /**
   * @hidden @internal
   * Compute the angle of the route at the {@link #midPoint}.
   * @expose
   * @this {Link}
   * @return {number} the angle in degrees.
   */
  public computeMidAngle(): number {
    const numpts = this.pointsCount;
    if (numpts < 2) {
      return NaN;  // no angle means call ComputeMidAngle again
    } else if (numpts === 2) {
      const a = this.getPoint(0);
      const b = this.getPoint(1);
      return a.directionPoint(b);
    }

    if (this.isOrthogonal && (this.computeCorner() >= 15 || this.computeCurve() === Link.Bezier)) {
      // use geometry to get angle at midpoint
      return this._geometry.getAngleAlongPath(.5);
    }
    if (this.computeCurve() === Link.Bezier && numpts >= 4) {
      const numsegs = ((numpts - 1) / 3) | 0;
      let idx = ((numsegs / 2) | 0) * 3;
      if (numsegs % 2 === 1) {
        idx = Math.floor(idx);
        const a = this.getPoint(idx);
        const b = this.getPoint(idx + 1);
        const c = this.getPoint(idx + 2);
        const d = this.getPoint(idx + 3);
        return Geo.bezierMidAngle(a.x, a.y, b.x, b.y, c.x, c.y, d.x, d.y);
      } else {
        if (idx > 0 && idx + 1 < numpts) {
          const a = this.getPoint(idx - 1);
          const b = this.getPoint(idx + 1);
          return a.directionPoint(b);
        }
        // else drop through and treat as a non-Bezier-curve
      }
    }

    const segmentDists = this.flattenedLengths;
    const totalDist = this.flattenedTotalLength;

    let currentDistance = 0.0;
    let currentPointIndex = 0;
    let nextSegmentLength = 0.0;
    while (currentDistance < totalDist / 2 && currentPointIndex < numpts) {
      nextSegmentLength = segmentDists[currentPointIndex];
      if (currentDistance + nextSegmentLength > totalDist / 2) break;
      currentDistance += nextSegmentLength;
      currentPointIndex++;
    }
    const currentPoint = this.getPoint(currentPointIndex);
    const nextPoint = this.getPoint(currentPointIndex + 1);

    // Handle Orthogonal cases first
    if (Math.abs(currentPoint.x - nextPoint.x) < 1 && Math.abs(currentPoint.y - nextPoint.y) < 1) {
      // points are approximately equal, look at surrounding points
      if (currentPointIndex > 0 && currentPointIndex + 2 < numpts) {
        const a = this.getPoint(currentPointIndex - 1);
        const b = this.getPoint(currentPointIndex + 2);
        return a.directionPoint(b);
      }
    } else if (Math.abs(currentPoint.x - nextPoint.x) < 1) {
      // vertical
      if (currentPoint.y > nextPoint.y) return 270;
      return 90;
    } else if (Math.abs(currentPoint.y - nextPoint.y) < 1) {
      // horizontal
      if (currentPoint.x > nextPoint.x) return 180;
      return 0;
    }
    // non-orthogonal lines
    return currentPoint.directionPoint(nextPoint);
  }

 /**
  * @hidden @internal
  * @this {Link}
  */
 private _maybeBuildLengths() {
   if (this._flattenedLengths === null) {
     if (!this.validRoute) {
       this._validateRoute();
     }
     const flattenedLengths: Array<number> = this._flattenedLengths = [];
     this._buildLengths(flattenedLengths);
   }
 }

  /**
   * @hidden @internal
   * This read-only property returns an Array segment lengths.
   */
  /** @export */ get flattenedLengths(): Array<number> {
    this._maybeBuildLengths();
    return this._flattenedLengths;
  }

  /**
   * @hidden @internal
   * This read-only property returns the approximate length of the path.
   */
  /** @export */ get flattenedTotalLength(): number {
    let len = this._flattenedTotalLength;
    if (isNaN(len)) {
      const allLengths = this.flattenedLengths;
      const ll = allLengths.length;
      len = 0.0;
      for (let i = 0; i < ll; i++) {
        len += allLengths[i];
      }
      this._flattenedTotalLength = len;
    }
    return len;
  }

  /**
   * Builds an array of segment lengths that make up the link path.
   * @hidden @internal
   * @this {Geometry}
   * @param {Array.<number>} allLengths
   */
  private _buildLengths(allLengths: Array<number>): void {
    const numpts = this.pointsCount;
    for (let i = 0; i < numpts - 1; i++) {
      let segmentDistance = 0.0;
      const a = this.getPoint(i);
      const b = this.getPoint(i + 1);

      // Handle computationally simple Orthogonal cases
      if (Geo.isApproxEqual(a.x, b.x)) {
        segmentDistance = b.y - a.y;
        if (segmentDistance < 0) segmentDistance = -segmentDistance;
        allLengths.push(segmentDistance);
      } else if (Geo.isApproxEqual(a.y, b.y)) {
        segmentDistance = b.x - a.x;
        if (segmentDistance < 0) segmentDistance = -segmentDistance;
        allLengths.push(segmentDistance);
      } else {
        // Handle non-Orthogonal lines
        segmentDistance = Math.sqrt(a.distanceSquaredPoint(b));
        allLengths.push(segmentDistance);
      }
    }
  }

  // * If you do want to modify the route without replacing the whole list of points,
  // * call {@link #startRoute} first, some number of calls to
  // * methods such as {@link #addPoint} or {@link #setPoint},
  // * followed by a call to {@link #commitRoute}.

  /**
   * Gets or sets the {@link List} of {@link Point}s in the route.
   * All of the Points must be {@link Point#isReal} -- no NaN or infinite values --
   * and all of the points are in document coordinates.
   * Although this list may be replaced by setting this property,
   * one must not modify the contents of the List directly.
   *
   * The setter also accepts an Array of alternating X,Y point values [1, 2, 3, 4, 5, 6],
   * or an Array of Objects each of which has (lowercase) 'x' and 'y' properties that are numbers
   * [{"x":1,"y":2}, {"x":3,"y":4}, {"x":5,"y":6}].
   *
   * Ownership of the {@link List} and all of its {@link Point}s that is provided to the setter
   * is transferred to this Link.
   */
  /** @export */ get points(): List<Point> { return this._points; }
  /** @export */ set points(value: List<Point>) {
    const old = this._points;
    if (old === value) return;
    let val2: List<Point> = null;
    if (Array.isArray(value)) {
      // we really check to see if this is an array of numbers or points
      const arr: Array<any> = value;
      // make sure there are an even number of values and that they are all numbers
      let allnums = arr.length % 2 === 0;
      if (allnums) {
        for (let j = 0; j < arr.length; j++) {
          if (typeof arr[j] !== 'number' || isNaN(arr[j])) {
            allnums = false;
            break;
          }
        }
      }
      if (allnums) {
        val2 = new List<Point>();
        for (let k = 0; k < arr.length / 2; k++) {
          const pt = new Point(arr[k * 2], arr[k * 2 + 1]).freeze();
          val2.add(pt);
        }
      } else {
        let allpts = true;
        for (let k = 0; k < arr.length; k++) {
          const pt2 = arr[k];
          if (!Util.isObject(pt2) ||
            typeof pt2['x'] !== 'number' || isNaN(pt2['x']) ||
            typeof pt2['y'] !== 'number' || isNaN(pt2['y'])) {
            allpts = false;
            break;
          }
        }
        if (allpts) {
          val2 = new List<Point>();
          for (let k = 0; k < arr.length; k++) {
            const pt = arr[k] as Point;
            val2.add(new Point(pt['x'], pt['y']).freeze());
          }
        } else if (Debug) {
          Util.throwError('Link.points array must contain only an even number of numbers or objects with x and y properties, not: ' + value);
        }
      }
    } else if (value instanceof List) { // list
      // gotta allocate a new List for undo/redo recognition
      // but the Points are not copied from the original VAL
      val2 = value.copy();
      // freeze the new List and each of its Points
      const pit = val2.iterator;
      while (pit.next()) {
        const p = pit.value;
        p.freeze();
      }
    } else {
      Util.throwError('Link.points value is not an instance of List or Array: ' + value);
    }

    val2.freeze();
    // caution: similar code is in the movePoints method!
    this._points = val2;
    this.invalidateGeometry();
    this.invalidateMeasure();
    this._validateRoute();

    const diagram = this.diagram;
    if (diagram !== null) {
      if (!diagram.isInitialLayoutCompleted && !diagram.undoManager.isUndoingRedoing) {
        diagram.loadedRouteLinks.add(this);
      }
      if (diagram.animationManager._pendingAnimation) this._savedRoute = val2;
    }
    this.raiseChanged('points', old, val2);
  }

  /**
   * This read-only property returns the number of points in the route.
   */
  /** @export */ get pointsCount(): number { return this._points.count; }

  /**
   * Gets a particular point of the route.
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the desired point.
   * @return {Point} in document coordinates
   */
  public getPoint(i: number): Point {
    return this._points._arr[i];
  }

  /**
   * Sets a particular point of the route; this may only be called within an override of {@link #computePoints}.
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the desired point.
   * @param {Point} p  The new point in document coordinates, which should not have infinite or NaN coordinate values, and which must not be modified afterwards.
   * @see {@link #getPoint}
   * @see {@link #insertPoint}
   * @see {@link #addPoint}
   * @see {@link #removePoint}
   * @since 1.6
   */
  public setPoint(i: number, p: Point): void {
    if (Debug) {
      Util.checkClass(p, Point, Link, 'setPoint');
      if (!p.isReal()) Util.throwError('Link.setPoint called with a Point that does not have real numbers: ' + p.toString());
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.setElt(i, p);
  }

  // * Sets a particular point of the route; this may only be called within an override of {@link #computePoints}.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the desired point.
   * @param {number} x  The new X, which should not be infinite or NaN, in document coordinates.
   * @param {number} y  The new Y, which should not be infinite or NaN, in document coordinates.
   */
  public setPointAt(i: number, x: number, y: number): void {
    if (Debug) {
      Util.checkRealNumber(x, Link, 'setPointAt:x');
      Util.checkRealNumber(y, Link, 'setPointAt:y');
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.setElt(i, new Point(x, y));
  }

  /**
   * Insert a point at a particular position in the route, without replacing an existing point; this may only be called within an override of {@link #computePoints}.
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the new point.
   * @param {Point} p  The new point in document coordinates, which should not have infinite or NaN coordinate values, and which must not be modified afterwards.
   * @see {@link #getPoint}
   * @see {@link #setPoint}
   * @see {@link #addPoint}
   * @see {@link #removePoint}
   * @since 1.6
   */
  public insertPoint(i: number, p: Point): void {
    if (Debug) {
      Util.checkClass(p, Point, Link, 'insertPoint');
      if (!p.isReal()) Util.throwError('Link.insertPoint called with a Point that does not have real numbers: ' + p.toString());
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.insertAt(i, p);
  }

  // * Insert a point at a particular position in the route, without replacing an existing point; this may only be called within an override of {@link #computePoints}.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the new point.
   * @param {number} x  The new X, which should not be infinite or NaN, in document coordinates.
   * @param {number} y  The new Y, which should not be infinite or NaN, in document coordinates.
   */
  public insertPointAt(i: number, x: number, y: number): void {
    if (Debug) {
      Util.checkRealNumber(x, Link, 'insertPointAt:x');
      Util.checkRealNumber(y, Link, 'insertPointAt:y');
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.insertAt(i, new Point(x, y));
  }

  /**
   * Add a point at the end of the route; this may only be called within an override of {@link #computePoints}.
   * @export
   * @this {Link}
   * @param {Point} p  The new point in document coordinates, which should not have infinite or NaN coordinate values, and which must not be modified afterwards.
   * @see {@link #getPoint}
   * @see {@link #setPoint}
   * @see {@link #insertPoint}
   * @see {@link #removePoint}
   * @since 1.6
   */
  public addPoint(p: Point): void {
    if (Debug) {
      Util.checkClass(p, Point, Link, 'addPoint');
      if (!p.isReal()) Util.throwError('Link.addPoint called with a Point that does not have real numbers: ' + p.toString());
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.add(p);
  }

  // * Add a point at the end of the route; this may only be called within an override of {@link #computePoints}.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Link}
   * @param {number} x  The new X, which should not be infinite or NaN, in document coordinates.
   * @param {number} y  The new Y, which should not be infinite or NaN, in document coordinates.
   */
  public addPointAt(x: number, y: number): void {
    if (Debug) {
      Util.checkRealNumber(x, Link, 'insertPointAt:x');
      Util.checkRealNumber(y, Link, 'insertPointAt:y');
    }
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.add(new Point(x, y));
  }

  /**
   * Remove a particular point from the route; this may only be called within an override of {@link #computePoints}.
   * @export
   * @this {Link}
   * @param {number} i int  The zero-based index of the point to extract.
   * @see {@link #getPoint}
   * @see {@link #setPoint}
   * @see {@link #insertPoint}
   * @see {@link #addPoint}
   * @see {@link #clearPoints}
   * @since 1.6
   */
  public removePoint(i: number): void {
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.removeAt(i);
  }

  /**
   * Remove all of the points from this link's route; this may only be called within an override of {@link #computePoints}.
   * @export
   * @this {Link}
   * @see {@link #getPoint}
   * @see {@link #setPoint}
   * @see {@link #insertPoint}
   * @see {@link #addPoint}
   * @see {@link #removePoint}
   * @since 1.6
   */
  public clearPoints(): void {
    if (Debug && this._oldPoints === null) Util.throwError('Call Link.startRoute before modifying the points of the route.');
    this._points.clear();
  }

  /**
   * @hidden @internal
   * Shift all of the {@link Point}s in the {@link #points} list by a given offsets.
   * @this {Link}
   * @param {number} dx in document coordinates
   * @param {number} dy in document coordinates
   */
  public movePoints(dx: number, dy: number): void {
    if (dx === 0 && dy === 0) return;
    const wasValid = this.validRoute;
    // caution: code copied from the 'points' property setter, to avoid duplicate List of Points!
    const newpts = new List<Point>();
    const pit = this._points.iterator;
    while (pit.next()) {
      const p = pit.value;
      newpts.add(new Point(p.x + dx, p.y + dy).freeze());
    }
    newpts.freeze();
    // caution: similar code is in the points setter!
    const old = this._points;
    this._points = newpts;

    // This used to be invalidate geometry, then invalidate measure (slow)
    // now we try to move the other Link information only, when possible
    if (isNaN(dx) || isNaN(dy) || this.diagram.animationManager._pendingAnimation) {
      this.invalidateMeasure();
    } else {
      this._location.assignTo(this._location.x + dx, this._location.y + dy);
      this._position.assignTo(this._position.x + dx, this._position.y + dy);
      this.invalidateArrange();
    }


    if (wasValid) this._validateRoute();
    const diagram = this.diagram;
    if (diagram !== null) {
      // don't need to add to diagram.loadedRouteLinks
      if (diagram.animationManager._pendingAnimation) this._savedRoute = newpts;
    }
    this.raiseChanged('points', old, newpts);
  }

  /**
   * @hidden
   * Allow calls to {@link #setPoint}, {@link #addPoint}, {@link #insertPoint},
   * {@link #removePoint}, and {@link #clearPoints}.
   * You must call {@link #commitRoute} when you are done modifying the route.
   * @export
   * @this {Link}
   */
  public startRoute(): void {
    if (this._oldPoints === null) {
      this._oldPoints = this._points;
      this._points = this._points.copy();  // now the current List is modifiable
    }
  }

  /**
   * @hidden
   * Call this method after a call to {@link #startRoute} and calls to methods that modify the route.
   * @export
   * @this {Link}
   */
  public commitRoute(): void {
    if (this._oldPoints === null) return;
    const oldpts = this._oldPoints;
    const pts = this._points;

    // find the old offset
    let ox = Infinity;
    let oy = Infinity;
    const oarr = oldpts._arr;
    const onum = oarr.length;
    for (let i = 0; i < onum; i++) {
      const p = oarr[i];
      ox = Math.min(p.x, ox);
      oy = Math.min(p.y, oy);
    }

    // find the new offset
    let nx = Infinity;
    let ny = Infinity;
    const narr = pts._arr;
    const nnum = narr.length;
    for (let i = 0; i < nnum; i++) {
      const p = narr[i];
      nx = Math.min(p.x, nx);
      ny = Math.min(p.y, ny);
      p.freeze();  // make sure all Points in the List are frozen
    }
    pts.freeze();  // and freeze the List itself

    // if the points count is the same and they all shifted
    // by an identical offset, do not invalidate the geometry
    if (nnum === onum) {
      for (let i = 0; i < nnum; i++) {
        const o = oarr[i];
        const n = narr[i];
        if (o.x - ox !== n.x - nx || o.y - oy !== n.y - ny) {
          this.invalidateMeasure();
          this.invalidateGeometry();
          break;
        }
      }
    } else {
      this.invalidateMeasure();
      this.invalidateGeometry();
    }

    // caution: similar code is in the movePoints method and points setter!
    this._oldPoints = null;

    const diagram = this.diagram;
    if (diagram !== null) {
      // don't need to add to diagram.loadedRouteLinks
      if (diagram.animationManager._pendingAnimation) this._savedRoute = pts;
    }

    this._validateRoute();
    this.raiseChanged('points', oldpts, pts);
  }

  /**
   * @hidden
   * Call this method instead of {@link #commitRoute} if you need to cancel changes to the route.
   * @export
   * @this {Link}
   */
  public rollbackRoute(): void {
    if (this._oldPoints === null) return;
    this._points = this._oldPoints;
    this._oldPoints = null;
  }

  /**
   * @hidden @internal
   * Used by animation to ensure custom link routing is saved
   *
   * ??? @@@ This is done on ALL links and should only be done on links pertaining to an animation, if efficient
   * @this {Link}
   */
  public restoreSavedRoute(): void {
    if (this._savedRoute === null) {
      // No route to restore, invalidate
      const stroke = this.path;
      if (stroke === null) return; // no stroke, no use checking further
      this.validRoute = false;
      this.invalidateMeasure();
      stroke.invalidateMeasure();
    } else {
      // restore route
      this.points = this._savedRoute;
      this._savedRoute = null;
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  private _validateRoute(): void {
    if (this._points.count === 0) {
      this.validRoute = false;
      return;
    }
    this.validRoute = true;
    this._flattenedLengths = null;
    this._flattenedTotalLength = NaN;
    this.defaultFromPoint = this.getPoint(0);
    this.defaultToPoint = this.getPoint(this.pointsCount - 1);
    this._invalidateOtherJumpOversMaybe(false);
  }

  /**
   * Declare that the route (the {@link #points}) of this Link need to be recomputed soon.
   * This causes {@link #updateRoute} to be called, which will call {@link #computePoints}
   * to perform the actual determination of the route.
   * @export
   * @this {Link}
   * @since 1.6
   */
  public invalidateRoute(): void {  // ??? needs optional argument to safely call invalidatePortInfo on connected ports
    if (this.suspendsRouting) return;
    const diagram = this.diagram;
    if (diagram) {
      if (diagram.loadedRouteLinks.contains(this)) return;
      if (diagram.undoManager.isUndoingRedoing) return;
      const AM = diagram.animationManager;
      // ??? interferes with canceled animations, even on 1.8:
      // line was commented out in Changeset 30526, but this breaks BeatPaths, so re-added for now
      if (AM.isTicking && !AM.isAnimating) return;
      if (this._savedRoute !== null && !AM.isTicking && AM.isAnimating) {
        this._savedRoute = null;
      }
    }
    const stroke = this.path;
    if (stroke === null) return; // no stroke, no use checking further
    this.validRoute = false;
    this.invalidateMeasure();
    stroke.invalidateMeasure();
  }

  /** @hidden @internal */
  /* noexport */get validRoute() { return (this._linkFlags & 0x00000010) !== 0; }
  set validRoute(value) {
    const old = (this._linkFlags & 0x00000010) !== 0;
    if (old !== value) {
      this._linkFlags = this._linkFlags ^ 0x00000010;
    }
  }

  /** @hidden @internal */
  /** @export */ get suspendsRouting() { return (this._linkFlags & 0x00000020) !== 0; }
  /** @export */ set suspendsRouting(value) {
    const old = (this._linkFlags & 0x00000020) !== 0;
    if (old !== value) {
      this._linkFlags = this._linkFlags ^ 0x00000020;
    }
  }

  /** @hidden @internal */
  /* noexport */get computingPoints() { return (this._linkFlags & 0x00000040) !== 0; }
  set computingPoints(value) {
    const old = (this._linkFlags & 0x00000040) !== 0;
    if (old !== value) {
      this._linkFlags = this._linkFlags ^ 0x00000040;
    }
  }

  /**
   * @hidden
   * Undocumented.
   */
  /** @export */ get defaultFromPoint() { return this._defaultFromPoint; }
  /** @export */ set defaultFromPoint(value) { this._defaultFromPoint = value.copy(); }

  /**
   * @hidden
   * Undocumented.
   */
  /** @export */ get defaultToPoint() { return this._defaultToPoint; }
  /** @export */ set defaultToPoint(value) { this._defaultToPoint = value.copy(); }

  /**
   * This method recomputes the route if the route is invalid,
   * to make sure the {@link #points} are up-to-date.
   * This method calls {@link #computePoints} in order to calculate a new route.
   *
   * The route may become invalid if either connected node is moved,
   * or if various properties are modified, such as {@link #routing}.
   * Such changes will call {@link #invalidateRoute}.
   *
   * It is unusual for you to have to call this method.
   * @export
   * @this {Link}
   * @since 1.6
   */
  public updateRoute(): void {
    if (this.validRoute) return;
    if (this.computingPoints) return;
    let calc = true;
    try {
      this.computingPoints = true;
      this.startRoute();
      calc = this.computePoints();
    } finally {
      this.computingPoints = false;
      if (calc) this.commitRoute();
      else this.rollbackRoute(); // in case the route is started but never finished
    }
  }

  /**
   * The code that constructs a new route by modifying the {@link #points}.
   * It is only called by {@link #updateRoute}, when needed.
   * All route points are always in document coordinates.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * Overrides of this method may call a number of "protected" methods in order to build the route:
   * {@link #clearPoints}, {@link #addPoint}, {@link #insertPoint}, {@link #removePoint}, {@link #setPoint}.
   * Overrides of this method must not set {@link #points}.
   *
   * Other methods that are commonly called and may be overridden are:
   * {@link #computeEndSegmentLength}, {@link #computeSpot}, {@link #computeOtherPoint},
   * {@link #computeCurve}, {@link #computeCurviness}, {@link #hasCurviness},
   * {@link #computeThickness},
   * {@link #makeGeometry}.
   * @expose
   * @this {Link}
   * @return {boolean} true if it computed a route of points
   * @since 1.6
   */
  public computePoints(): boolean {
    const diagram = this.diagram;
    if (diagram === null) return false;

    let fromnode = this.fromNode;
    let fromport: GraphObject = null;
    if (fromnode === null) {
      if (!diagram._routingFromNode) {
        diagram._routingFromPort = new Shape();
        diagram._routingFromPort.desiredSize = Size.ZeroSize;
        diagram._routingFromPort.strokeWidth = 0;  // must have zero actual size
        diagram._routingFromNode = new Node();
        diagram._routingFromNode.add(diagram._routingFromPort);
        diagram._routingFromNode.ensureBounds();
      }
      if (this.defaultFromPoint) {
        diagram._routingFromNode.position = diagram._routingFromNode.location = this.defaultFromPoint;
        diagram._routingFromNode.ensureBounds();
        fromnode = diagram._routingFromNode;
        fromport = diagram._routingFromPort;
      }
    } else {
      fromport = this.fromPort;
    }
    if (fromport !== null) {
      if (!fromnode.isVisible()) {  // fromport and its parents aren't visible
        const visfromnode = fromnode.findVisibleNode();  // is fromnode itself visible?
        if (visfromnode !== null && visfromnode !== fromnode) {  // no, found a containing group
          fromnode = visfromnode;  // use it instead
          fromport = visfromnode.port;  // use default port -- it wouldn't apply to a different node anyway
        } else {
          fromnode = visfromnode;  // might be null if everything is not visible
        }
      }
    }
    this._tempFromNode = fromnode;
    // if the fromport is not visible, even though the fromnode is visible, the port might not have a real bounds
    if (fromnode === null) return false;
    if (!fromnode.location.isReal()) return false;
    while (fromport !== null && (!fromport.actualBounds.isReal() || !fromport.isVisibleObject())) {
      fromport = fromport.panel;
    }
    if (fromport === null) return false;

    let tonode = this.toNode;
    let toport: GraphObject = null;
    if (tonode === null) {
      if (!diagram._routingToNode) {
        diagram._routingToPort = new Shape();
        diagram._routingToPort.desiredSize = Size.ZeroSize;
        diagram._routingToPort.strokeWidth = 0;  // must have zero actual size
        diagram._routingToNode = new Node();
        diagram._routingToNode.add(diagram._routingToPort);
        diagram._routingToNode.ensureBounds();
      }
      if (this.defaultToPoint) {
        diagram._routingToNode.position = diagram._routingToNode.location = this.defaultToPoint;
        diagram._routingToNode.ensureBounds();
        tonode = diagram._routingToNode;
        toport = diagram._routingToPort;
      }
    } else {
      toport = this.toPort;
    }
    if (toport !== null) {
      if (!tonode.isVisible()) {  // toport and its parents aren't visible
        const vistonode = tonode.findVisibleNode();  // is tonode itself visible?
        if (vistonode !== null && vistonode !== tonode) {  // no, found a containing group
          tonode = vistonode;  // use it instead
          toport = vistonode.port;  // use default port -- it wouldn't apply to a different node anyway
        } else {
          tonode = vistonode;  // might be null if everything is not visible
        }
      }
    }
    this._tempToNode = tonode;
    // if the toport is not visible, even though the tonode is visible, the port might not have a real bounds
    if (tonode === null) return false;
    if (!tonode.location.isReal()) return false;
    while (toport !== null && (!toport.actualBounds.isReal() || !toport.isVisibleObject())) {
      toport = toport.panel;
    }
    if (toport === null) return false;

    const num = this.pointsCount;

    const fromspot = this.computeSpot(true, fromport);  // link's Spot takes precedence, if defined
    const tospot = this.computeSpot(false, toport);  // link's Spot takes precedence, if defined
    const fromdirect = this.isSpotUndirected(fromspot);
    const todirect = this.isSpotUndirected(tospot);

    const selfloop = (fromport === toport && fromport !== null);
    const ortho = this.isOrthogonal;
    let bezier = (this.curve === Link.Bezier);
    if (selfloop && !ortho) {
      bezier = true;
      this._effectiveBezier = true;
    } else {
      this._effectiveBezier = false;
    }
    const calc = (this.adjusting === Link.None) || selfloop;

    if (!ortho && !selfloop && fromdirect && todirect) {
      let adjusted = false;
      if (!calc && num >= 3) {
        let pA = this.getLinkPoint(fromnode, fromport, fromspot, true, false, tonode, toport);  // must be newly allocated result
        let pB = this.getLinkPoint(tonode, toport, tospot, false, false, fromnode, fromport);  // must be newly allocated result
        adjusted = this.adjustPoints(0, pA, num - 1, pB);
        // improve routing of end points at Spot.None ports
        if (adjusted) {
          pA = this.getLinkPoint(fromnode, fromport, fromspot, true, false, tonode, toport);  // must be newly allocated result
          pB = this.getLinkPoint(tonode, toport, tospot, false, false, fromnode, fromport);  // must be newly allocated result
          this.adjustPoints(0, pA, num - 1, pB);
        }
      }
      if (!adjusted) {
        this.clearPoints();
        if (bezier) {
          this.calculateBezierNoSpot(fromnode, fromport, fromspot, tonode, toport, tospot);
        } else {
          this.calculateLineNoSpot(fromnode, fromport, fromspot, tonode, toport, tospot);
        }
      }
    } else {  // three segments (four points)
      const avoidsnodes = this.isAvoiding;
      if (calc && ((ortho && avoidsnodes) || selfloop)) this.clearPoints();
      const curviness = (selfloop ? this.computeCurviness() : 0);

      const fromPoint = this.getLinkPoint(fromnode, fromport, fromspot, true, ortho, tonode, toport);  // must be newly allocated result
      let Bx = 0.0;
      let By = 0.0;
      let fromDir = 0.0;
      if (ortho || !fromdirect || selfloop) {
        // get the length of the beginning segment
        let fromSeg = this.computeEndSegmentLength(fromnode, fromport, fromspot, true);

        // calculate the end point of the initial segment
        fromDir = this.getLinkDirection(fromnode, fromport, fromPoint, fromspot, true, ortho, tonode, toport);
        if (selfloop && (fromdirect || fromspot.equals(tospot) || (!ortho && fromspot.x + tospot.x === 1 && fromspot.y + tospot.y === 1))) {
          // add some curve, especially when to and from either the same or the opposite spot
          fromDir -= (ortho ? 90 : 30);
          if (curviness < 0) {
            fromDir -= 180;
          }
        }
        if (fromDir < 0) fromDir += 360;
        else if (fromDir >= 360) fromDir -= 360;

        if (selfloop) fromSeg += Math.abs(curviness) * (ortho ? 1 : 2);

        // handle common cases more quickly than by calling Cos and Sin
        if (fromDir === 0) {
          Bx = fromSeg;
        } else if (fromDir === 90) {
          By = fromSeg;
        } else if (fromDir === 180) {
          Bx = -fromSeg;
        } else if (fromDir === 270) {
          By = -fromSeg;
        } else {
          Bx = fromSeg * Math.cos(fromDir * Math.PI / 180);
          By = fromSeg * Math.sin(fromDir * Math.PI / 180);
        }

        // try to accomodate self-loops on nodes that have no spot
        if (fromspot.isNoSpot() && selfloop) {
          const ctr = fromport.getDocumentPoint(Spot.Center, Point.alloc());
          const temp = Point.allocAt(ctr.x + Bx * 1000, ctr.y + By * 1000);
          this.getLinkPointFromPoint(fromnode, fromport, ctr, temp, true, fromPoint);  // clobbers fromPoint
          Point.free(ctr);
          Point.free(temp);
        }
      }

      const toPoint = this.getLinkPoint(tonode, toport, tospot, false, ortho, fromnode, fromport);  // must be newly allocated result
      let Cx = 0.0;
      let Cy = 0.0;
      let toDir = 0.0;
      if (ortho || !todirect || selfloop) {
        // get the length of the ending segment
        let toSeg = this.computeEndSegmentLength(tonode, toport, tospot, false);

        // calculate the start point of the final segment
        toDir = this.getLinkDirection(tonode, toport, toPoint, tospot, false, ortho, fromnode, fromport);
        if (selfloop && (todirect || fromspot.equals(tospot) || (!ortho && fromspot.x + tospot.x === 1 && fromspot.y + tospot.y === 1))) {
          // add some curve, especially when to and from either the same or the opposite spot
          toDir += (ortho ? 0 : 30);
          if (curviness < 0) {
            toDir += 180;
          }
        }
        if (toDir < 0) toDir += 360;
        else if (toDir >= 360) toDir -= 360;

        if (selfloop) toSeg += Math.abs(curviness) * (ortho ? 1 : 2);

        // handle common cases more quickly than by calling Cos and Sin
        if (toDir === 0) {
          Cx = toSeg;
        } else if (toDir === 90) {
          Cy = toSeg;
        } else if (toDir === 180) {
          Cx = -toSeg;
        } else if (toDir === 270) {
          Cy = -toSeg;
        } else {
          Cx = toSeg * Math.cos(toDir * Math.PI / 180);
          Cy = toSeg * Math.sin(toDir * Math.PI / 180);
        }

        // try to accommodate self-loops on nodes that have no spot
        if (tospot.isNoSpot() && selfloop) {
          const ctr = toport.getDocumentPoint(Spot.Center, Point.alloc());
          const temp = Point.allocAt(ctr.x + Cx * 1000, ctr.y + Cy * 1000);
          this.getLinkPointFromPoint(tonode, toport, ctr, temp, false, toPoint);  // clobbers toPoint
          Point.free(ctr);
          Point.free(temp);
        }
      }

      // determine intermediate/control points
      let b = fromPoint;
      if (ortho || !fromdirect || selfloop) {
        b = new Point(fromPoint.x + Bx, fromPoint.y + By);
      }
      let c = toPoint;
      if (ortho || !todirect || selfloop) {
        c = new Point(toPoint.x + Cx, toPoint.y + Cy);
      }

      // only call AdjustPoints when AdjustingStyle is not Calculate, and if it returns true, it was successful
      if (!calc && !ortho && fromdirect && num > 3 && this.adjustPoints(0, fromPoint, num - 2, c)) {
        this.setPoint(num - 1, toPoint);
      } else if (!calc && !ortho && todirect && num > 3 && this.adjustPoints(1, b, num - 1, toPoint)) {
        this.setPoint(0, fromPoint);
      } else if (!calc && (ortho ? num >= 6 : num > 4) && this.adjustPoints(1, b, num - 2, c)) {
        this.setPoint(0, fromPoint);
        this.setPoint(num - 1, toPoint);
      } else {  // calculate standard four (non-ortho) or six (ortho) point stroke
        this.clearPoints();

        // add the initial point
        this.addPoint(fromPoint);

        if (ortho || !fromdirect || selfloop) {
          this.addPoint(b);
        }

        if (ortho) {
          this.addOrthoPoints(b, fromDir, c, toDir, fromnode, tonode);
        }

        if (ortho || !todirect || selfloop) {
          this.addPoint(c);
        }

        // add the final endpoint
        this.addPoint(toPoint);
      }
    }
    return true;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} c  the central point.
   * @param {Point} p  the orientation point; this is then modified and returned.
   * @return {Point}  this has either X or Y be a huge value.
   */
  private _orthoPointToward(c: Point, p: Point): Point {
    if (Math.abs(p.x - c.x) > Math.abs(p.y - c.y)) {
      if (p.x >= c.x) {
        p.x = c.x + 9e9;
      } else {
        p.x = c.x - 9e9;
      }
      p.y = c.y;
    } else {
      if (p.y >= c.y) {
        p.y = c.y + 9e9;
      } else {
        p.y = c.y - 9e9;
      }
      p.x = c.x;
    }
    return p;
  }

  /**
   * Compute the intersection point in document coordinates for the edge of a particular port GraphObject, given a point,
   * when no particular spot or side has been specified.
   * @expose
   * @this {Link}
   * @param {Node} node
   * @param {GraphObject} port the {@link GraphObject} representing a port on the node.
   * @param {Point} focus the point in document coordinates to/from which the link should point,
   *   normally the center of the port.
   * @param {Point} p often this point is far away from the node, to give a general direction,
   *   particularly an orthogonal one.
   * @param {boolean} from true if the link is coming out of the port; false if going to the port.
   * @param {Point=} result an optional Point that is modified and returned; otherwise it allocates and returns a new Point
   * @return {Point} the point in document coordinates of the intersection point on the edge of the port.
   * @since 1.2
   */
  public getLinkPointFromPoint(node: Node | null, port: GraphObject | null, focus: Point, p: Point, from: boolean, result?: Point): Point {
    if (result === undefined) result = new Point();
    if (node === null || port === null) {
      result.assign(focus);
      return result;
    }
    if (!node.isVisible()) {
      const vnode = node.findVisibleNode();
      if (vnode !== null && vnode !== node) {
        node = vnode;
        port = node.port;
      }
    }

    // ??? var portb = node.getDocumentBounds(port);
    // ??? if (!portb.containsPoint(focus)) focus = portb.center;
    // if the "distant" point is inside the port, just use the port's center, to reduce visual confusion
    // ??? if (portb.containsPoint(p)) { result.assign(focus); return result; }

    /*
    // otherwise get the closest edge point on the line from P to CTR
    var nodeLoc = node.location;
    // Convert p1/p2 to node coordinates
    var p1x = p.x - nodeLoc.x;
    var p1y = p.y - nodeLoc.y;
    var p2x = focus.x - nodeLoc.x;
    var p2y = focus.y - nodeLoc.y;
    */

    let p1x = 0.0;
    let p1y = 0.0;
    let p2x = 0.0;
    let p2y = 0.0;
    let tt: Transform = null;
    let portpanel = port.panel;
    if (portpanel !== null && !portpanel.trueTransform()) portpanel = portpanel.panel;
    if (portpanel === null) {
      p1x = p.x;
      p1y = p.y;
      p2x = focus.x;
      p2y = focus.y;
    } else {
      tt = portpanel.allTransforms;
      // inverted transform point:
      const d = 1 / (tt.m11 * tt.m22 - tt.m12 * tt.m21);
      const m0 = tt.m22 * d;
      const m1 = -tt.m12 * d;
      const m2 = -tt.m21 * d;
      const m3 = tt.m11 * d;
      const m4 = d * (tt.m21 * tt.dy - tt.m22 * tt.dx);
      const m5 = d * (tt.m12 * tt.dx - tt.m11 * tt.dy);
      p1x = p.x * m0 + p.y * m2 + m4;
      p1y = p.x * m1 + p.y * m3 + m5;

      p2x = focus.x * m0 + focus.y * m2 + m4;
      p2y = focus.x * m1 + focus.y * m3 + m5;
    }

    // get the element whose edge we'll use (which is normally the same element)
    port.getNearestIntersectionPoint2(p1x, p1y, p2x, p2y, result);
    if (tt !== null) result.transform(tt);
    return result;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {GraphObject} port
   * @param {Link} link
   * @return {LinkInfo}
   */
  private _findExistingLinkInfo(port: GraphObject, link: Link): LinkInfo {  // does not update
    const knot = port._portInfo;
    if (knot !== null) return knot.findLinkInfo(link);
    return null;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {GraphObject} port
   * @return {LinkInfo}
   */
  private _getLinkInfo(port: GraphObject): LinkInfo {  // updates if needed
    let knot = port._portInfo;
    if (knot === null) {
      knot = new Knot();
      knot.port = port;
      knot.node = port.part as Node;
      port._portInfo = knot;
    }
    return knot.findLinkInfo(this);
  }

  /**
   * Compute the point on a node/port in document coordinates at which the route of a link should end.
   * @expose
   * @this {Link}
   * @param {Node} node
   * @param {GraphObject} port the {@link GraphObject} representing a port on the node.
   * @param {Spot} spot a {@link Spot} value describing where the link should connect.
   * @param {boolean} from true if the link is coming out of the port; false if going to the port.
   * @param {boolean} ortho whether the link should have orthogonal segments.
   * @param {Node} othernode the node at the other end of the link.
   * @param {GraphObject} otherport the {@link GraphObject} port at the other end of the link.
   * @param {Point=} result an optional Point that is modified and returned; otherwise it allocates and returns a new Point
   * @return {Point} in document coordinates.
   * @since 1.2
   */
  public getLinkPoint(node: Node | null, port: GraphObject | null, spot: Spot, from: boolean, ortho: boolean,
                      othernode: Node | null, otherport: GraphObject | null, result?: Point): Point {
    if (result === undefined) result = new Point();
    if (spot.isSpot() && !this.isSpotUndirected(spot)) {
      port.getDocumentPoint(spot, result);
      return result;
    } else if (spot.isSide()) {
      const thisli = this._getLinkInfo(port);
      if (thisli !== null) {
        const ilp = thisli.linkPoint;  // specific value for that link
        result.assign(ilp);
        if (ortho && this.routing === Link.AvoidsNodesStraight) {
          const otherli = this._getLinkInfo(otherport);
          // Try to have the link go straight between two Nodes with spots that are isSide().
          // Assumption: OK to shift the resulting point from the ideal evenly-spread side points,
          // if there are fewer links at this side than there are at the other end of the link.
          if (otherli !== null && thisli.countOnSide < otherli.countOnSide) {  // required to avoid infinite loop
            const tempa = Point.alloc();
            const tempb = Point.alloc();
            const thisr = new Rect(port.getDocumentPoint(Spot.TopLeft, tempa),
              port.getDocumentPoint(Spot.BottomRight, tempb));
            const otherspot = this.computeSpot(!from, otherport);
            const otherpt = this.getLinkPoint(othernode, otherport, otherspot, !from, ortho, node, port, tempb);  // reuse TEMPB
            if ((spot.includesSide(Spot.LeftSide) || spot.includesSide(Spot.RightSide)) &&
              (otherpt.y >= thisr.y && otherpt.y <= thisr.y + thisr.height)) {
              result.y = otherpt.y;
            } else if ((spot.includesSide(Spot.TopSide) || spot.includesSide(Spot.BottomSide)) &&
              (otherpt.x >= thisr.x && otherpt.x <= thisr.x + thisr.width)) {
              result.x = otherpt.x;
            }
            Point.free(tempa);
            Point.free(tempb);
          }
        }
        return result;
      }
    }

    const focus = port.getDocumentPoint(this.computeFocusSpot(spot), Point.alloc());
    let temp: Point = null;
    let far: Point = null;
    // no defined spot or side -- need to calculate nearest intersection
    // of port's edge with line to focus point from some "far" point
    if (this.pointsCount > (ortho ? 6 : 2)) {  // try using the existing points in the route
      far = from ? this.getPoint(1) : this.getPoint(this.pointsCount - 2);
      if (ortho) {
        far = this._orthoPointToward(focus, far.copy());
      } else {
        // if the far point is inside the port, try another point "farther" along the route
        // ??? var b = port.getDocumentBounds();
        // ??? var i = 2;
        // ??? while (b.containsPoint(far) && i < this.pointsCount) {
        // ???   far = from ? this.getPoint(i) : this.getPoint(this.pointsCount-1-i);
        // ???   i++;
        // ??? }
        // ??? if (i >= this.pointsCount) far = otherport.getDocumentPoint(Spot.Center, ???);

        // ???far = from ? this.getPoint(this.pointsCount-1) : this.getPoint(0);  //???
      }
    } else {
      const otherspot = this.computeSpot(!from, otherport);
      temp = Point.alloc();
      far = otherport.getDocumentPoint(this.computeFocusSpot(otherspot), temp);
      if (ortho) {
        far = this._orthoPointToward(focus, far);
      }
      Point.free(temp);
    }

    this.getLinkPointFromPoint(node, port, focus, far, from, result);
    Point.free(focus);

    return result;
  }

  /**
   * Compute the direction in which a link should go from a given connection point.
   * @expose
   * @this {Link}
   * @param {Node} node
   * @param {GraphObject} port the {@link GraphObject} representing a port on the node.
   * @param {Point} linkpoint the connection point, in document coordinates.
   * @param {Spot} spot a {@link Spot} value describing where the link should connect.
   * @param {boolean} from true if the link is coming out of the port; false if going to the port.
   * @param {boolean} ortho whether the link should have orthogonal segments.
   * @param {Node} othernode the node at the other end of the link.
   * @param {GraphObject} otherport the {@link GraphObject} port at the other end of the link.
   * @return {number} the absolute angle, in degrees.
   * @since 1.2
   */
  public getLinkDirection(node: Node | null, port: GraphObject | null, linkpoint: Point, spot: Spot, from: boolean, ortho: boolean,
                          othernode: Node | null, otherport: GraphObject | null): number {
    let angle = this._getLinkDirection1(node, port, linkpoint, spot, from, ortho, othernode, otherport);
    if (spot.isNoSpot() && othernode.isMemberOf(node)) {
      angle += 180;
      if (angle >= 360) angle -= 360;
    }

    if (this.isSpotUndirected(spot)) return angle;
    const rot = port.getDocumentAngle();
    if (rot === 0) return angle;
    if (45 <= rot && rot < 135) angle += 90;
    else if (135 <= rot && rot < 225) angle += 180;
    else if (225 <= rot && rot < 315) angle += 270;
    if (angle >= 360) angle -= 360;
    return angle;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} node
   * @param {GraphObject} port the {@link GraphObject} representing a port on the node.
   * @param {Point} linkpoint the connection point, in document coordinates.
   * @param {Spot} spot a {@link Spot} value describing where the link should connect.
   * @param {boolean} from true if the link is coming out of the port; false if going to the port.
   * @param {boolean} ortho whether the link should have orthogonal segments.
   * @param {Node} othernode the node at the other end of the link.
   * @param {GraphObject} otherport the {@link GraphObject} port at the other end of the link.
   * @return {number} the absolute angle, in degrees.
   */
  private _getLinkDirection1(node: Node | null, port: GraphObject | null, linkpoint: Point, spot: Spot, from: boolean, ortho: boolean,
                             othernode: Node | null, otherport: GraphObject | null): number {
    if (spot.isSpot()) {  // specific spot is defined
      if (spot.x > spot.y) {
        if (spot.x > 1 - spot.y) {
          return 0;
        } else if (spot.x < 1 - spot.y) {
          return 270;
        } else {
          return 315;
        }
      } else if (spot.x < spot.y) {
        if (spot.x > 1 - spot.y) {
          return 90;
        } if (spot.x < 1 - spot.y) {
          return 180;
        } else {
          return 135;
        }
      } else {
        if (spot.x < 0.5) {
          return 225;
        } else if (spot.x > 0.5) {
          return 45;
        } else {
          return 0;
        }
      }
    }
    if (spot.isSide()) {
      const info = this._getLinkInfo(port);
      if (info !== null) {
        switch (info.side) {
          case Util.MTOP: return 270;
          case Util.MLEFT: return 180;
          default:
          case Util.MRIGHT: return 0;
          case Util.MBOTTOM: return 90;
        }
      }
    }
    // no defined spot -- guess direction based on other points
    // in the route, if any
    const c = port.getDocumentPoint(Spot.Center, Point.alloc());
    let temp: Point = null;
    let far: Point = null;
    if (this.pointsCount > (ortho ? 6 : 2)) {
      far = from ? this.getPoint(1) : this.getPoint(this.pointsCount - 2);
      if (ortho) {
        far = this._orthoPointToward(c, far.copy());
      } else {
        far = linkpoint;
      }
    } else {
      temp = Point.alloc();
      far = otherport.getDocumentPoint(Spot.Center, temp);
      Point.free(temp);
    }
    let result = 0.0;
    if (Math.abs(far.x - c.x) > Math.abs(far.y - c.y)) {
      if (far.x >= c.x) {
        result = 0;
      } else {
        result = 180;
      }
    } else {
      if (far.y >= c.y) {
        result = 90;
      } else {
        result = 270;
      }
    }
    Point.free(c);
    return result;
  }

  /**
   * Get the length of the end segment in document coordinates, typically a short distance, in document units.
   * For spot values that are {@link Spot#isSide}, this returns a computed value.
   * Depending on the `from` argument, this will return {@link #fromEndSegmentLength} or {@link #toEndSegmentLength}.
   * If the value is `NaN`, this will return the {@link #fromPort}'s {@link GraphObject#fromEndSegmentLength}
   * or the {@link #toPort}'s {@link GraphObject#toEndSegmentLength}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @param {Node} node
   * @param {GraphObject} port the {@link GraphObject} representing a port on the node.
   * @param {Spot} spot a {@link Spot} value describing where the link should connect.
   * @param {boolean} from true if the link is coming out of the port; false if going to the port.
   * @return {number} a distance in document coordinates; must be a real number, not NaN or infinity.
   * @since 1.6
   */
  public computeEndSegmentLength(node: Node | null, port: GraphObject | null, spot: Spot, from: boolean): number {
    if (port !== null && spot.isSide()) {
      const info = this._getLinkInfo(port);
      if (info !== null) return info.endSegmentLength;  // specific value for that link
    }
    let esl = NaN;
    if (from) {
      esl = this.fromEndSegmentLength;
    } else {
      esl = this.toEndSegmentLength;
    }
    if (port !== null && isNaN(esl)) {
      if (from) {
        esl = port.fromEndSegmentLength;
      } else {
        esl = port.toEndSegmentLength;
      }
    }
    if (isNaN(esl)) esl = 10;
    return esl;
  }

  /**
   * Get the {@link Spot} that describes how the end of the link should connect with the port.
   * Depending on the `from` argument, this will return {@link #fromSpot} or {@link #toSpot}.
   * If the value is {@link Spot#isDefault}, this will return the {@link #fromPort}'s {@link GraphObject#fromSpot}
   * or the {@link #toPort}'s {@link GraphObject#toSpot}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @param {boolean} from
   * @param {GraphObject=} port
   * @return {Spot} must not be {@link Spot.Default}, but may be {@link Spot.None}
   * @since 1.6
   */
  public computeSpot(from: boolean, port?: GraphObject | null): Spot {
    if (port === undefined) port = null;
    if (from) {
      return this._getFromSpot(port ? port : this.fromPort);
    } else {
      return this._getToSpot(port ? port : this.toPort);
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {GraphObject} port
   * @return {Spot}
   */
  private _getFromSpot(port: GraphObject | null): Spot {
    if (port === null) return Spot.Center;
    let s = this.fromSpot;
    if (s.isDefault()) {
      s = port.fromSpot;
    }
    if (s === Spot.Default) return Spot.None;
    return s;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {GraphObject} port
   * @return {Spot}
   */
  private _getToSpot(port: GraphObject | null): Spot {
    if (port === null) return Spot.Center;
    let s = this.toSpot;
    if (s.isDefault()) {
      s = port.toSpot;
    }
    if (s === Spot.Default) return Spot.None;
    return s;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Spot} s
   * @return {boolean}
   */
  private isSpotUndirected(s: Spot): boolean {
    return s === Spot.None || (s.x === 0.5 && s.y === 0.5); // 2nd condition new for 2.0
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Spot} s
   * @return {Spot}
   */
  private computeFocusSpot(s: Spot): Spot {
    if (s.x === 0.5 && s.y === 0.5) return s;
    return Spot.Center;
  }

  /**
   * Find the approximate point of the other end of the link in document coordinates.
   * This is useful when computing the connection point when there is no specific spot, to have an idea of which general direction the link should be going.
   * By default this will return the center of the other port.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @param {Node} othernode
   * @param {GraphObject} otherport
   * @return {Point} approximately where the other end of this link might end, in document coordinates
   * @since 1.6
   */
  public computeOtherPoint(othernode: Node, otherport: GraphObject): Point {
    let otherpoint = otherport.getDocumentPoint(Spot.Center);
    const otherinfo = this._findExistingLinkInfo(otherport, this);
    if (otherinfo !== null) {
      otherpoint = otherinfo.linkPoint;
    }
    return otherpoint;
  }

  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Link}
   * @param {boolean} from
   * @return {number} must be a real number, not NaN or infinity.
   */
  protected computeShortLength(from: boolean): number {
    if (from) {
      let s = this.fromShortLength;
      if (isNaN(s)) {
        const port = this.fromPort;
        if (port !== null) {
          s = port.fromShortLength;
        }
      }
      if (isNaN(s)) return 0;
      return s;
    } else {
      let s = this.toShortLength;
      if (isNaN(s)) {
        const port = this.toPort;
        if (port !== null) {
          s = port.toShortLength;
        }
      }
      if (isNaN(s)) return 0;
      return s;
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Rect} r A Rect in container coordinates.
   * @param {function(GraphObject):GraphObject | null} navig A function taking a GraphObject and
   * returning a GraphObject, defaulting to the identity.
   * @param {function(GraphObject):boolean | null} pred A function taking the GraphObject
   * returned by navig and returning true if that object should be returned,
   * defaulting to a predicate that always returns true.
   * @param {boolean} partialInclusion True if objects are to be included when they
   * intersect.
   * @param {List.<GraphObject>|Set.<GraphObject>} list A List (or Set) to add the results to.
   * @param {Transform=} txfm
   * @return {boolean} Whether or not this object is included.
   */
  public findObjectsIn<T extends GraphObject>(r: Rect,
                                              navig: ((a: GraphObject) => (T | null)) | null,
                                              pred: ((a: T) => boolean) | null,
                                              partialInclusion: boolean,
                                              list: List<T> | Set<T>,
                                              txfm?: Transform): boolean {
    if (this.pickable === false) return false;
    if (navig === undefined) navig = null;
    if (pred === undefined) pred = null;
    let transform = txfm;
    if (txfm === undefined) {
      transform = Transform.alloc();
      transform.reset();
    }

    // ??? how do we ensure this works when trueTransform is false?

    transform.multiply(this.transform);
    if (this.containedInRect(r, transform)) { // first 'yes'
      // add to the list "part" and everything inside of "part",
      // going down the object tree using a special function that only tests pred and nothing geometrical,c
      // because by everything inside is already completely within r
      this.addObjs(navig, pred, list);
      if (txfm === undefined) Transform.free(transform);
      return true;
    } else if (this.intersectsRect(r, transform)) { // first 'no'
      // 1. We have to go through all the children of "part", invoking "findObjectsIn" on each.
      // We do it regardless of partialInclusion or not, because "part" might have a child that is entirely included,
      // or it may not, and we have no way of knowing yet.
      let result = false;
      if (!this.isAtomic) {
        const arr = this._elements._arr;
        const l = arr.length;
        for (let i = l; i--; ) {
          const child = arr[i];
          if (!child.visible && child !== this.locationObject) continue;
          const cab = child.actualBounds;
          const ab = this.naturalBounds;
          if (cab.x > ab.width || cab.y > ab.height || cab.x + cab.width < 0 || cab.y + cab.height < 0) continue;

          const obj = child;
          const tt = Transform.alloc();
          tt.set(transform);
          if (obj instanceof Panel) {
            result = obj.findObjectsIn(r, navig, pred, partialInclusion, list, tt);
          } else {
            if (this.path === obj) {
              if (obj instanceof Shape) result = obj.findObjectInPath(r, partialInclusion, tt);
            } else {
              result = obj.findObjectIn(r, partialInclusion, tt);
            }
          }
          if (result) {
            let obj2 = obj as T;
            if (navig !== null) obj2 = navig(obj);
            if (obj2 && (pred === null || pred(obj2))) {
              list.add(obj2);
            }
          }
          Transform.free(tt);
        }
      }
      if (txfm === undefined) Transform.free(transform);
      return (result || this.background !== null || this.areaBackground !== null); // instead of partialInclusion
    }
    // second 'no'
    // nothing added to the list, we are done.
    if (txfm === undefined) Transform.free(transform);
    return false;
  }

  /**
   * This read-only property is true if {@link #routing} is a value that implies that the points
   * of the route should be orthogonal, such that each point shares a common X or a common Y value
   * with the immediately previous and next points.
   * This property is completely dependent on the {@link #routing} property.
   * Values of {@link Link.Orthogonal} and {@link Link.AvoidsNodes} causes this property to be true.
   *
   * The points in the route might not actually form an orthogonal route, but when the route is computed
   * the intent is to maintain orthogonality.
   * @since 1.2
   */
  /** @export */ get isOrthogonal(): boolean { return (this._routing.value & 2) === 2; }

  /**
   * @hidden @internal
   * An internal static predicate.
   * @param {EnumValue} v
   * @return {boolean}
   */
  public static isOrthogonalRouting(v: EnumValue) {
    return (v.value & 2) === 2;
  }

  /**
   * @hidden
   * This read-only property is true when the routing tries to be smart about not overlapping other parts.
   */
  /** @export */ get isAvoiding(): boolean { return (this._routing.value & 4) === 4; }

  /**
   * Returns the {@link #curve}, unless this link is supposed to pretend to be curved, as with reflexive links.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {EnumValue}
   * @since 1.6
   */
  public computeCurve(): EnumValue {
    if (this._effectiveBezier === null) {
      const fromport = this.fromPort;
      const selfloop = (fromport !== null && fromport === this.toPort);
      const ortho = this.isOrthogonal;
      this._effectiveBezier = (selfloop && !ortho);
    }
    if (this._effectiveBezier) return Link.Bezier;
    return this.curve;
  }

  //    * Returns the {@link #corner}, if it's a non-negative number, or else 10.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Link}
   * @return {number} must be a real number, not NaN or infinity
   */
  protected computeCorner(): number {
    if (this.curve === Link.Bezier) return 0;
    let c = this.corner;
    if (isNaN(c) || c < 0) c = 10;
    return c;
  }

  //    * Returns the first label that should be at the "middle" of the link, if there is any such label.
  //  * Elements that have .isPanelMain === true are ignored, including the {@link #path}.
  //  * Elements that have a {@link GraphObject#segmentIndex} set are ignored, such as arrowheads.
  //  * If there are no label objects within the Link Panel, consider any {@link #labelNodes}
  //  * whose segmentIndex has not been set.
  //  * This could be overridden to return a different label than the first one, or to always return null.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Link}
   * @return {GraphObject}
   */
  public findMidLabel(): GraphObject | null {
    const stroke = this.path;  // may be null
    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const e = children[i];
      if (e === stroke || e.isPanelMain) continue;
      if (e.segmentIndex === -Infinity || isNaN(e.segmentIndex)) return e;
    }
    // also check for label Nodes -- just return the first one that is "middle"
    const lit = this.labelNodes;
    while (lit.next()) {
      const labn = lit.value;
      if (labn.segmentIndex === -Infinity || isNaN(labn.segmentIndex)) return labn;
    }
    return null;
  }

  /**
   * Returns the expected spacing between this link and others that connect this link's fromPort and toPort.
   * This calls {@link #computeThickness} and also takes any "mid label"'s breadth into account.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {number} must be a real number, not NaN or infinity
   * @see {@link #computeCurviness}
   * @since 1.6
   */
  public computeSpacing(): number {
    if (!this.isVisible()) return 0;
    let spacing = 14.0;
    spacing = Math.max(spacing, this.computeThickness());
    const from = this.fromPort;
    const to = this.toPort;
    if (from !== null && to !== null) {
      const lab = this.findMidLabel();
      if (lab !== null) {
        const b = lab.naturalBounds;
        const m = lab.margin as Margin;
        const w = !isNaN(b.width) ? (b.width * lab.scale) + m.left + m.right : 30.0;
        const h = !isNaN(b.height) ? (b.height * lab.scale) + m.top + m.bottom : 14.0;
        // take the segmentOrientation into account
        const orient = lab.segmentOrientation;
        if (orient === Link.OrientAlong ||
          orient === Link.OrientUpright ||
          orient === Link.OrientOpposite) {
          spacing = Math.max(spacing, h);
        } else if (orient === Link.OrientMinus90 ||
          orient === Link.OrientMinus90Upright ||
          orient === Link.OrientPlus90 ||
          orient === Link.OrientPlus90Upright) {
          spacing = Math.max(spacing, w);
        } else {  // None, OrientUpright45?  Needed space depends on direction of link
          const dir = from.getDocumentPoint(Spot.Center).directionPoint(to.getDocumentPoint(Spot.Center));
          const ang = dir / 180 * Math.PI;
          spacing = Math.max(spacing, Math.abs(Math.sin(ang) * w) + Math.abs(Math.cos(ang) * h) + 1);
        }
        if (this.curve === Link.Bezier) spacing *= 1.333;  // ??? fudge for Bezier curves
      }
    }
    return spacing;
  }

  //  * Sort the Links between two ports and invalidate their routes.
  //  * Only one of the Links in the bundle will get this method called -- it is undetermined which Link is chosen.
  //  *
  //  * By default does not change the order of the links, but does invalidate all routes if REROUTE is true.
  //  * This method must not add Links to or remove Links from the argument Array, but may re-order the Links in the Array.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Link}
   * @param {Array.<Link>} links
   * @param {boolean} reroute
   */
  public arrangeBundledLinks(links: Array<Link>, reroute: boolean): void {
    if (reroute) {
      for (let i = 0; i < links.length; i++) {
        const l = links[i];
        if (l.adjusting === Link.None) l.invalidateRoute();
      }
    }
  }

  /**
   * Returns the {@link #curviness}, if it's a number,
   * or else a computed value based on how many links connect this pair of nodes/ports,
   * by calling {@link #computeSpacing} on each link.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {number} must be a real number, not NaN or infinity
   * @see {@link #computeSpacing}
   * @since 1.6
   */
  public computeCurviness(): number {
    let c = this.curviness;
    if (isNaN(c)) {
      c = 16.0;
      const bundle = this._bundle;
      if (bundle !== null) {
        const spacings = Util.tempArray<number>();
        let total = 0;
        const links = bundle.links;
        for (let i = 0; i < links.length; i++) {
          const l = links[i];
          const space = l.computeSpacing();  // inefficient if computeSpacing is slow
          spacings.push(space);
          total += space;
        }
        let sum = -total / 2;
        for (let i = 0; i < links.length; i++) {
          const l = links[i];
          if (l === this) { c = sum + spacings[i] / 2; break; }
          sum += spacings[i];
        }
        if (bundle.node1 === this.fromNode) c = -c;
        Util.freeArray(spacings);
      }
    }
    return c;
  }

  /**
   * Returns the thickness of this link.
   * By default it uses the strokeWidth of the main element, assuming it's a Shape.
   * This is called by {@link #computeSpacing}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {number} a non-negative real number
   * @since 1.6
   */
  public computeThickness(): number {
    if (!this.isVisible()) return 0;
    const shape = this.path;
    if (shape !== null) return Math.max(shape.strokeWidth, 1);
    return 1;
  }

  /**
   * Returns true if an extra or a different point is needed based on {@link #curviness}.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {boolean}
   * @since 1.6
   */
  public hasCurviness(): boolean {
    return !isNaN(this.curviness) || this._bundle !== null;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} fromnode
   * @param {GraphObject} fromport
   * @param {Spot} fromspot
   * @param {Node} tonode
   * @param {GraphObject} toport
   * @param {Spot} tospot
   */
  private calculateLineNoSpot(fromnode: Node, fromport: GraphObject, fromspot: Spot, tonode: Node, toport: GraphObject, tospot: Spot): void {
    const frompt = this.getLinkPoint(fromnode, fromport, fromspot, true, false, tonode, toport); // must be newly allocated Point
    const topt = this.getLinkPoint(tonode, toport, tospot, false, false, fromnode, fromport);  // must be newly allocated Point
    // dont need to clear points, caller already did
    if (this.hasCurviness()) {
      const Dx = topt.x - frompt.x;
      const Dy = topt.y - frompt.y;
      const rad = this.computeCurviness();
      const off = rad;

      const Mx = frompt.x + Dx / 2;
      const My = frompt.y + Dy / 2;
      let C1x = Mx;
      let C1y = My;
      if (Geo.isApprox(Dy, 0)) {
        if (Dx > 0) {
          C1y -= off;
        } else {
          C1y += off;
        }
      } else {
        const slope = -Dx / Dy;
        let E = Math.sqrt(off * off / (slope * slope + 1));
        if (rad < 0) E = -E;
        C1x = (Dy < 0 ? -1 : 1) * E + Mx;
        C1y = slope * (C1x - Mx) + My;
      }
      this.addPoint(frompt);
      this.addPointAt(C1x, C1y);
      this.addPoint(topt);
    } else {
      this.addPoint(frompt);
      this.addPoint(topt);
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Node} fromnode
   * @param {GraphObject} fromport
   * @param {Spot} fromspot
   * @param {Node} tonode
   * @param {GraphObject} toport
   * @param {Spot} tospot
   */
  private calculateBezierNoSpot(fromnode: Node, fromport: GraphObject, fromspot: Spot, tonode: Node, toport: GraphObject, tospot: Spot): void {
    const pA = this.getLinkPoint(fromnode, fromport, fromspot, true, false, tonode, toport);  // must be newly allocated Point
    const pB = this.getLinkPoint(tonode, toport, tospot, false, false, fromnode, fromport); // must be newly allocated Point

    const Dx = pB.x - pA.x;
    const Dy = pB.y - pA.y;
    const rad = this.computeCurviness();
    let slope = 0.0;  // can't calculate this now, if Dy is close to zero
    let E = 0.0;

    let Mx = pA.x + Dx / 3;
    let My = pA.y + Dy / 3;
    let C1x = Mx;
    let C1y = My;
    if (Geo.isApprox(Dy, 0)) {
      if (Dx > 0) {
        C1y -= rad;
      } else {
        C1y += rad;
      }
    } else {
      slope = -Dx / Dy;
      E = Math.sqrt(rad * rad / (slope * slope + 1));
      if (rad < 0) E = -E;
      C1x = (Dy < 0 ? -1 : 1) * E + Mx;
      C1y = slope * (C1x - Mx) + My;
    }

    Mx = pA.x + 2 * Dx / 3;
    My = pA.y + 2 * Dy / 3;
    let C2x = Mx;
    let C2y = My;
    if (Geo.isApprox(Dy, 0)) {
      if (Dx > 0) {
        C2y -= rad;
      } else {
        C2y += rad;
      }
    } else {
      // slope and E have been initialized earlier, when Dy !== 0
      C2x = (Dy < 0 ? -1 : 1) * E + Mx;
      C2y = slope * (C2x - Mx) + My;
    }

    this.clearPoints();
    this.addPoint(pA);
    this.addPointAt(C1x, C1y);
    this.addPointAt(C2x, C2y);
    this.addPoint(pB);
    // adjust end points so that they appear to come from the control points
    this.setPoint(0, this.getLinkPoint(fromnode, fromport, fromspot, true, false, tonode, toport));  // must be newly allocated Point
    this.setPoint(3, this.getLinkPoint(tonode, toport, tospot, false, false, fromnode, fromport)); // must be newly allocated Point
  }

  /**
   * @hidden @internal
   * Adjust all of the existing points in this link's stroke in an inclusive range
   * according to new first and last stroke points.
   *
   * This is primarily useful to help maintain intermediate inflection points
   * in a link when one or both ports moves.
   * By default this just calls {@link #rescalePoints}, {@link #stretchPoints},
   * or {@link #modifyEndPoints}.
   * This method is not called when there are no existing points to be adjusted
   * or when {@link #adjusting} is None.
   *
   * When this link is orthogonal, an {@link #adjusting} of
   * Scale will just return false to result in the standard orthogonal path.
   * An {@link #adjusting} of Stretch
   * for an orthogonal link is treated as if it were End.
   * @expose
   * @this {Link}
   * @param {number} startIndex int the zero-based index of the first point to be changed, to be
   * the value of the newFromPoint argument.
   * @param {Point} newFromPoint a Point in document coordinates.
   * @param {number} endIndex int the zero-based index of the last point to be changed, to be
   * the value of the newToPoint argument.
   * @param {Point} newToPoint a Point in document coordinates.
   * @return {boolean} This method should return true if the stroke points were adjusted.
   * Return false to tell {@link #computePoints} to plot the standard path.
   */
  public adjustPoints(startIndex: number, newFromPoint: Point, endIndex: number, newToPoint: Point): boolean {
    let s = this.adjusting;
    if (this.isOrthogonal) {
      if (s === Link.Scale) {
        return false;
      }
      if (s === Link.Stretch) {
        s = Link.End;
      }
    }
    switch (s) {
      case Link.Scale: return this.rescalePoints(startIndex, newFromPoint, endIndex, newToPoint);
      case Link.Stretch: return this.stretchPoints(startIndex, newFromPoint, endIndex, newToPoint);
      case Link.End: return this.modifyEndPoints(startIndex, newFromPoint, endIndex, newToPoint);
      default: return false;
    }
  }

  /**
   * @hidden @internal
   * Maintain the same shape for the stroke, but scale and rotate according to the
   * new point arguments newFromPoint and newToPoint.
   * @this {Link}
   * @param {number} startIndex int the zero-based index of the first point to be changed, to be
   * the value of the newFromPoint argument.
   * @param {Point} newFromPoint a Point in document coordinates.
   * @param {number} endIndex int the zero-based index of the last point to be changed, to be
   * the value of the newToPoint argument.
   * @param {Point} newToPoint a Point in document coordinates.
   * @return {boolean} This method should return true if the stroke points were adjusted.
   * Return false to tell {@link #computePoints} to plot the standard path.
   */
  private rescalePoints(startIndex: number, newFromPoint: Point, endIndex: number, newToPoint: Point): boolean {
    const oldFromPt = this.getPoint(startIndex);
    const oldToPt = this.getPoint(endIndex);
    if (oldFromPt.equalsApprox(newFromPoint) && oldToPt.equalsApprox(newToPoint)) return true;

    const Ax = oldFromPt.x;
    const Ay = oldFromPt.y;
    const Bx = oldToPt.x;
    const By = oldToPt.y;
    let Dx = Bx - Ax;
    let Dy = By - Ay;
    const oldDist = Math.sqrt(Dx * Dx + Dy * Dy);
    if (Geo.isApproxEqual(oldDist, 0)) {
      return true;
    }
    let oldAngle = 0.0;
    if (Geo.isApproxEqual(Dx, 0)) {
      if (Dy < 0) {
        oldAngle = -Math.PI / 2;
      } else {
        oldAngle = Math.PI / 2;
      }
    } else {
      oldAngle = Math.atan(Dy / Math.abs(Dx));
      if (Dx < 0) {
        oldAngle = Math.PI - oldAngle;
      }
    }

    const A2x = newFromPoint.x;
    const A2y = newFromPoint.y;
    const B2x = newToPoint.x;
    const B2y = newToPoint.y;
    const D2x = B2x - A2x;
    const D2y = B2y - A2y;
    const newDist = Math.sqrt(D2x * D2x + D2y * D2y);
    let newAngle = 0.0;
    if (Geo.isApproxEqual(D2x, 0)) {
      if (D2y < 0) {
        newAngle = -Math.PI / 2;
      } else {
        newAngle = Math.PI / 2;
      }
    } else {
      newAngle = Math.atan(D2y / Math.abs(D2x));
      if (D2x < 0) {
        newAngle = Math.PI - newAngle;
      }
    }

    const DistRatio = (newDist / oldDist);
    const AngleDiff = (newAngle - oldAngle);

    this.setPoint(startIndex, newFromPoint);
    for (let i = startIndex + 1; i < endIndex; i++) {
      const p = this.getPoint(i);
      Dx = p.x - Ax;
      Dy = p.y - Ay;
      const pDist = Math.sqrt(Dx * Dx + Dy * Dy);
      if (Geo.isApproxEqual(pDist, 0)) {
        continue;
      }
      let pAngle = 0.0;
      if (Geo.isApproxEqual(Dx, 0)) {
        if (Dy < 0) {
          pAngle = -Math.PI / 2;
        } else {
          pAngle = Math.PI / 2;
        }
      } else {
        pAngle = Math.atan(Dy / Math.abs(Dx));
        if (Dx < 0) {
          pAngle = Math.PI - pAngle;
        }
      }

      const p2Angle = pAngle + AngleDiff;
      const p2Dist = pDist * DistRatio;
      const P2x = A2x + p2Dist * Math.cos(p2Angle);
      const P2y = A2y + p2Dist * Math.sin(p2Angle);
      this.setPointAt(i, P2x, P2y);
    }
    this.setPoint(endIndex, newToPoint);
    return true;
  }

  /**
   * @hidden @internal
   * Stretch the points of this stroke by interpolating the points
   * from startIndex to endIndex between the
   * new point arguments newFromPoint and newToPoint.
   * @this {Link}
   * @param {number} startIndex int the zero-based index of the first point to be changed, to be
   * the value of the newFromPoint argument.
   * @param {Point} newFromPoint a Point in document coordinates.
   * @param {number} endIndex int the zero-based index of the last point to be changed, to be
   * the value of the newToPoint argument.
   * @param {Point} newToPoint a Point in document coordinates.
   * @return {boolean} This method should return true if the stroke points were adjusted.
   * Return false to tell {@link #computePoints} to plot the standard path.
   */
  private stretchPoints(startIndex: number, newFromPoint: Point, endIndex: number, newToPoint: Point): boolean {
    const a = this.getPoint(startIndex);
    const b = this.getPoint(endIndex);
    if (a.equalsApprox(newFromPoint) && b.equalsApprox(newToPoint)) return true;

    const Ax = a.x;
    const Ay = a.y;
    const Bx = b.x;
    const By = b.y;
    const L = ((Bx - Ax) * (Bx - Ax) + (By - Ay) * (By - Ay));

    const Cx = newFromPoint.x;
    const Cy = newFromPoint.y;
    const Dx = newToPoint.x;
    const Dy = newToPoint.y;
    let M = 0.0;
    let m2 = 1.0;
    if (Dx - Cx !== 0) {
      M = (Dy - Cy) / (Dx - Cx);
      m2 = Math.sqrt(1 + (1 / (M * M)));
    } else {
      M = 9e9;
    }

    this.setPoint(startIndex, newFromPoint);
    for (let i = startIndex + 1; i < endIndex; i++) {
      const p = this.getPoint(i);
      const Px = p.x;
      const Py = p.y;

      let Q = 0.5;
      if (L !== 0) {
        Q = ((Ax - Px) * (Ax - Bx) + (Ay - Py) * (Ay - By)) / L;
      }

      // find point on old line
      const Vx = Ax + Q * (Bx - Ax);
      const Vy = Ay + Q * (By - Ay);
      // distance from P to point V, on old line
      let dV = Math.sqrt((Px - Vx) * (Px - Vx) + (Py - Vy) * (Py - Vy));
      if (Py < M * (Px - Vx) + Vy) {
        dV = -dV;
      }
      if (M > 0) {
        dV = -dV;
      }

      // find point on new line
      const Wx = Cx + Q * (Dx - Cx);
      const Wy = Cy + Q * (Dy - Cy);

      if (M !== 0) {
        // compute new point for P off of new line, distance dV from W
        const x = Wx + dV / m2;
        const y = Wy - (x - Wx) / M;
        this.setPointAt(i, x, y);
      } else {
        this.setPointAt(i, Wx, Wy + dV);
      }
    }
    this.setPoint(endIndex, newToPoint);
    return true;
  }

  /**
   * @hidden @internal
   * Modify only the end points of this stroke to match any new
   * newFromPoint or newToPoint points;
   * intermediate points are not changed.
   *
   * The startIndex'th point should be set to the newFromPoint argument,
   * and the endIndex'th point should be set to the newToPoint argument,
   * and the intermediate points should be not be changed unless needed to maintain orthogonality.
   * @this {Link}
   * @param {number} startIndex int the zero-based index of the first point to be changed, to be
   * the value of the newFromPoint argument.
   * @param {Point} newFromPoint a Point in document coordinates.
   * @param {number} endIndex int the zero-based index of the last point to be changed, to be
   * the value of the newToPoint argument.
   * @param {Point} newToPoint a Point in document coordinates.
   * @return {boolean} This method should return true if the stroke points were adjusted.
   * Return false to tell {@link #computePoints} to plot the standard path.
   */
  private modifyEndPoints(startIndex: number, newFromPoint: Point, endIndex: number, newToPoint: Point): boolean {
    if (this.isOrthogonal) {
      let a = this.getPoint(startIndex);
      let b = this.getPoint(startIndex + 1);
      let c = this.getPoint(startIndex + 2);
      let fx = b.x;
      let fy = b.y;
      const b1x = fx;
      const b1y = fy;
      if (Geo.isApprox(a.y, b.y)) {
        if (Geo.isApprox(b.x, c.x)) {
          fy = newFromPoint.y;
        } else if (Geo.isApprox(b.y, c.y)) {
          fx = newFromPoint.x;
        }
      } else if (Geo.isApprox(a.x, b.x)) {
        if (Geo.isApprox(b.y, c.y)) {
          fx = newFromPoint.x;
        } else if (Geo.isApprox(b.x, c.x)) {
          fy = newFromPoint.y;
        }
      }
      this.setPointAt(startIndex + 1, fx, fy);

      a = this.getPoint(endIndex);
      b = this.getPoint(endIndex - 1);
      c = this.getPoint(endIndex - 2);
      let tx = b.x;
      let ty = b.y;
      const b2x = tx;
      const b2y = ty;
      if (Geo.isApprox(a.y, b.y)) {
        if (Geo.isApprox(b.x, c.x)) {
          ty = newToPoint.y;
        } else if (Geo.isApprox(b.y, c.y)) {
          tx = newToPoint.x;
        }
      } else if (Geo.isApprox(a.x, b.x)) {
        if (Geo.isApprox(b.y, c.y)) {
          tx = newToPoint.x;
        } else if (Geo.isApprox(b.x, c.x)) {
          ty = newToPoint.y;
        }
      }
      this.setPointAt(endIndex - 1, tx, ty);
      if (this.crossesOverNode()) {
        this.setPointAt(startIndex + 1, b1x, b1y);
        this.setPointAt(endIndex - 1, b2x, b2y);
        return false;
      }
    }
    this.setPoint(startIndex, newFromPoint);
    this.setPoint(endIndex, newToPoint);
    return true;
  }

  /**
   * @hidden
   * This method is called by {@link #computePoints} when the link is orthogonal
   * and at least one port has a link spot that is not {@link Spot#isNoSpot}.
   *
   * {@link #computePoints} is responsible for adding the first two
   * and the last two points of the stroke, including the startFrom and endTo values.
   * This method is responsible for adding any additional points in the middle of the stroke.
   * This method calls {@link #computeMidOrthoPosition} to determine the
   * distance of the middle segment between the two ports.
   * It also tries to avoid the source node and the destination node.
   * When the {@link #routing} property is not Normal,
   * this method uses another, more computationally expensive,
   * method for determining the proper path of the link, which may have many segments.
   * @expose
   * @this {Link}
   * @param {Point} startFrom this {@link Point} will already have been added to the stroke by {@link #computePoints} before calling this method.
   * @param {number} fromDir Normally 0, 90, 180, or 270 degrees.
   * @param {Point} endTo {@link #computePoints} will add this {@link Point} after calling this method.
   * @param {number} toDir Normally 0, 90, 180, or 270 degrees.
   * @param {Node} fromnode the {@link Node} that the link is coming from.
   * @param {Node} tonode the {@link Node} that the link is going to.
   */
  public addOrthoPoints(startFrom: Point, fromDir: number, endTo: Point, toDir: number, fromnode: Node, tonode: Node): void {
    const RIGHT = 0;
    const DOWN = 90;
    const LEFT = 180;
    const UP = 270;

    if (RIGHT - 45 <= fromDir && fromDir < RIGHT + 45) {
      fromDir = RIGHT;
    } else if (DOWN - 45 <= fromDir && fromDir < DOWN + 45) {
      fromDir = DOWN;
    } else if (LEFT - 45 <= fromDir && fromDir < LEFT + 45) {
      fromDir = LEFT;
    } else {
      fromDir = UP;
    }
    if (RIGHT - 45 <= toDir && toDir < RIGHT + 45) {
      toDir = RIGHT;
    } else if (DOWN - 45 <= toDir && toDir < DOWN + 45) {
      toDir = DOWN;
    } else if (LEFT - 45 <= toDir && toDir < LEFT + 45) {
      toDir = LEFT;
    } else {
      toDir = UP;
    }

    const s = startFrom;
    const t = endTo;

    const fromR = fromnode.actualBounds.copy();  // handle rotated nodes
    const toR = tonode.actualBounds.copy();
    if (!fromR.isReal() || !toR.isReal()) return;
    fromR.inflate(8, 8);
    toR.inflate(8, 8);
    fromR.unionPoint(startFrom);
    toR.unionPoint(endTo);

    // do normal routing
    let m2: Point;
    let m3: Point;
    if (fromDir === RIGHT) {
      if (t.x > s.x ||
        (toDir === UP && t.y < s.y && toR.right > s.x) ||
        (toDir === DOWN && t.y > s.y && toR.right > s.x)) {
        m2 = new Point(t.x, s.y);
        m3 = new Point(t.x, (s.y + t.y) / 2);
        if (toDir === LEFT) {
          m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          m3.x = m2.x;
          m3.y = t.y;
        } else if ((toDir === UP && t.y < s.y) || (toDir === DOWN && t.y > s.y)) {
          if (s.x < toR.left) {
            m2.x = this.computeMidOrthoPosition(s.x, s.y, toR.left, t.y, false);
          } else if (s.x < toR.right && ((toDir === UP && s.y < toR.top) || (toDir === DOWN && s.y > toR.bottom))) {
            m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          } else {
            m2.x = toR.right;
          }
          m3.x = m2.x;
          m3.y = t.y;
        } else if (toDir === RIGHT && s.x < toR.left && s.y > toR.top && s.y < toR.bottom) {
          m2.x = s.x;
          if (s.y < t.y) {
            m2.y = Math.min(t.y, toR.top);
          } else {
            m2.y = Math.max(t.y, toR.bottom);
          }
          m3.y = m2.y;
        }
      } else {
        m2 = new Point(s.x, t.y);
        m3 = new Point((s.x + t.x) / 2, t.y);
        if (toDir === LEFT ||
          (toDir === DOWN && t.y < fromR.top) ||
          (toDir === UP && t.y > fromR.bottom)) {
          if (toDir === LEFT && (toR.containsPoint(s) || fromR.containsPoint(t))) { // close to each other
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          } else if (t.y < s.y && (toDir === LEFT || toDir === DOWN)) {
            m2.y = this.computeMidOrthoPosition(s.x, fromR.top, t.x, Math.max(t.y, toR.bottom), true);
          } else if (t.y > s.y && (toDir === LEFT || toDir === UP)) {
            m2.y = this.computeMidOrthoPosition(s.x, fromR.bottom, t.x, Math.min(t.y, toR.top), true);
          }
          m3.x = t.x;
          m3.y = m2.y;
        }
        if (m2.y > fromR.top && m2.y < fromR.bottom) {
          if ((t.x >= fromR.left && t.x <= s.x) || (s.x <= toR.right && s.x >= t.x)) {
            if (toDir === DOWN || toDir === UP) {
              m2 = new Point(Math.max((s.x + t.x) / 2, s.x), s.y);
              m3 = new Point(m2.x, t.y);
            }
          } else {
            if (toDir === UP || ((toDir === RIGHT || toDir === LEFT) && t.y < s.y)) {
              m2.y = Math.min(t.y, ((toDir === RIGHT) ? fromR.top : Math.min(fromR.top, toR.top)));
            } else {
              m2.y = Math.max(t.y, ((toDir === RIGHT) ? fromR.bottom : Math.max(fromR.bottom, toR.bottom)));
            }
            m3.x = t.x;
            m3.y = m2.y;
          }
        }
      }
    } else if (fromDir === LEFT) {
      if (t.x < s.x ||
        (toDir === UP && t.y < s.y && toR.left < s.x) ||
        (toDir === DOWN && t.y > s.y && toR.left < s.x)) {
        m2 = new Point(t.x, s.y);
        m3 = new Point(t.x, (s.y + t.y) / 2);
        if (toDir === RIGHT) {
          m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          m3.x = m2.x;
          m3.y = t.y;
        } else if ((toDir === UP && t.y < s.y) || (toDir === DOWN && t.y > s.y)) {
          if (s.x > toR.right) {
            m2.x = this.computeMidOrthoPosition(s.x, s.y, toR.right, t.y, false);
          } else if (s.x > toR.left && ((toDir === UP && s.y < toR.top) || (toDir === DOWN && s.y > toR.bottom))) {
            m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          } else {
            m2.x = toR.left;
          }
          m3.x = m2.x;
          m3.y = t.y;
        } else if (toDir === LEFT && s.x > toR.right && s.y > toR.top && s.y < toR.bottom) {
          m2.x = s.x;
          if (s.y < t.y) {
            m2.y = Math.min(t.y, toR.top);
          } else {
            m2.y = Math.max(t.y, toR.bottom);
          }
          m3.y = m2.y;
        }
      } else {
        m2 = new Point(s.x, t.y);
        m3 = new Point((s.x + t.x) / 2, t.y);
        if (toDir === RIGHT ||
          (toDir === DOWN && t.y < fromR.top) ||
          (toDir === UP && t.y > fromR.bottom)) {
          if (toDir === RIGHT && (toR.containsPoint(s) || fromR.containsPoint(t))) { // close to each other
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          } else if (t.y < s.y && (toDir === RIGHT || toDir === DOWN)) {
            m2.y = this.computeMidOrthoPosition(s.x, fromR.top, t.x, Math.max(t.y, toR.bottom), true);
          } else if (t.y > s.y && (toDir === RIGHT || toDir === UP)) {
            m2.y = this.computeMidOrthoPosition(s.x, fromR.bottom, t.x, Math.min(t.y, toR.top), true);
          }
          m3.x = t.x;
          m3.y = m2.y;
        }
        if (m2.y > fromR.top && m2.y < fromR.bottom) {
          if ((t.x <= fromR.right && t.x >= s.x) || (s.x >= toR.left && s.x <= t.x)) {
            if (toDir === DOWN || toDir === UP) {
              m2 = new Point(Math.min((s.x + t.x) / 2, s.x), s.y);
              m3 = new Point(m2.x, t.y);
            }
          } else {
            if (toDir === UP || ((toDir === RIGHT || toDir === LEFT) && t.y < s.y)) {
              m2.y = Math.min(t.y, ((toDir === LEFT) ? fromR.top : Math.min(fromR.top, toR.top)));
            } else {
              m2.y = Math.max(t.y, ((toDir === LEFT) ? fromR.bottom : Math.max(fromR.bottom, toR.bottom)));
            }
            m3.x = t.x;
            m3.y = m2.y;
          }
        }
      }
    } else if (fromDir === DOWN) {
      if (t.y > s.y ||
        (toDir === LEFT && t.x < s.x && toR.bottom > s.y) ||
        (toDir === RIGHT && t.x > s.x && toR.bottom > s.y)) {
        m2 = new Point(s.x, t.y);
        m3 = new Point((s.x + t.x) / 2, t.y);
        if (toDir === UP) {
          m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          m3.x = t.x;
          m3.y = m2.y;
        } else if ((toDir === LEFT && t.x < s.x) || (toDir === RIGHT && t.x > s.x)) {
          if (s.y < toR.top) {
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, toR.top, true);
          } else if (s.y < toR.bottom && ((toDir === LEFT && s.x < toR.left) || (toDir === RIGHT && s.x > toR.right))) {
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          } else {
            m2.y = toR.bottom;
          }
          m3.x = t.x;
          m3.y = m2.y;
        } else if (toDir === DOWN && s.y < toR.top && s.x > toR.left && s.x < toR.right) {
          if (s.x < t.x) {
            m2.x = Math.min(t.x, toR.left);
          } else {
            m2.x = Math.max(t.x, toR.right);
          }
          m2.y = s.y;
          m3.x = m2.x;
        }
      } else {
        m2 = new Point(t.x, s.y);
        m3 = new Point(t.x, (s.y + t.y) / 2);
        if (toDir === UP ||
          (toDir === RIGHT && t.x < fromR.left) ||
          (toDir === LEFT && t.x > fromR.right)) {
          if (toDir === UP && (toR.containsPoint(s) || fromR.containsPoint(t))) { // close to each other
            m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          } else if (t.x < s.x && (toDir === UP || toDir === RIGHT)) {
            m2.x = this.computeMidOrthoPosition(fromR.left, s.y, Math.max(t.x, toR.right), t.y, false);
          } else if (t.x > s.x && (toDir === UP || toDir === LEFT)) {
            m2.x = this.computeMidOrthoPosition(fromR.right, s.y, Math.min(t.x, toR.left), t.y, false);
          }
          m3.x = m2.x;
          m3.y = t.y;
        }
        if (m2.x > fromR.left && m2.x < fromR.right) {
          if ((t.y >= fromR.top && t.y <= s.y) || (s.y <= toR.bottom && s.y >= t.y)) {
            if (toDir === RIGHT || toDir === LEFT) {
              m2 = new Point(s.x, Math.max((s.y + t.y) / 2, s.y));
              m3 = new Point(t.x, m2.y);
            }
          } else {
            if (toDir === LEFT || ((toDir === DOWN || toDir === UP) && t.x < s.x)) {
              m2.x = Math.min(t.x, ((toDir === DOWN) ? fromR.left : Math.min(fromR.left, toR.left)));
            } else {
              m2.x = Math.max(t.x, ((toDir === DOWN) ? fromR.right : Math.max(fromR.right, toR.right)));
            }
            m3.x = m2.x;
            m3.y = t.y;
          }
        }
      }
    } else {  // fromDir === UP
      if (t.y < s.y ||
        (toDir === LEFT && t.x < s.x && toR.top < s.y) ||
        (toDir === RIGHT && t.x > s.x && toR.top < s.y)) {
        m2 = new Point(s.x, t.y);
        m3 = new Point((s.x + t.x) / 2, t.y);
        if (toDir === DOWN) {
          m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          m3.x = t.x;
          m3.y = m2.y;
        } else if ((toDir === LEFT && t.x < s.x) || (toDir === RIGHT && t.x >= s.x)) {
          if (s.y > toR.bottom) {
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, toR.bottom, true);
          } else if (s.y > toR.top && ((toDir === LEFT && s.x < toR.left) || (toDir === RIGHT && s.x > toR.right))) {
            m2.y = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, true);
          } else {
            m2.y = toR.top;
          }
          m3.x = t.x;
          m3.y = m2.y;
        } else if (toDir === UP && s.y > toR.bottom && s.x > toR.left && s.x < toR.right) {
          if (s.x < t.x) {
            m2.x = Math.min(t.x, toR.left);
          } else {
            m2.x = Math.max(t.x, toR.right);
          }
          m2.y = s.y;
          m3.x = m2.x;
        }
      } else {
        m2 = new Point(t.x, s.y);
        m3 = new Point(t.x, (s.y + t.y) / 2);
        if (toDir === DOWN ||
          (toDir === RIGHT && t.x < fromR.left) ||
          (toDir === LEFT && t.x > fromR.right)) {
          if (toDir === DOWN && (toR.containsPoint(s) || fromR.containsPoint(t))) { // close to each other
            m2.x = this.computeMidOrthoPosition(s.x, s.y, t.x, t.y, false);
          } else if (t.x < s.x && (toDir === DOWN || toDir === RIGHT)) {
            m2.x = this.computeMidOrthoPosition(fromR.left, s.y, Math.max(t.x, toR.right), t.y, false);
          } else if (t.x > s.x && (toDir === DOWN || toDir === LEFT)) {
            m2.x = this.computeMidOrthoPosition(fromR.right, s.y, Math.min(t.x, toR.left), t.y, false);
          }
          m3.x = m2.x;
          m3.y = t.y;
        }
        if (m2.x > fromR.left && m2.x < fromR.right) {
          if ((t.y <= fromR.bottom && t.y >= s.y) || (s.y >= toR.top && s.y <= t.y)) {
            if (toDir === RIGHT || toDir === LEFT) {
              m2 = new Point(s.x, Math.min((s.y + t.y) / 2, s.y));
              m3 = new Point(t.x, m2.y);
            }
          } else {
            if (toDir === LEFT || ((toDir === DOWN || toDir === UP) && t.x < s.x)) {
              m2.x = Math.min(t.x, ((toDir === UP) ? fromR.left : Math.min(fromR.left, toR.left)));
            } else {
              m2.x = Math.max(t.x, ((toDir === UP) ? fromR.right : Math.max(fromR.right, toR.right)));
            }
            m3.x = m2.x;
            m3.y = t.y;
          }
        }
      }
    }

    // maybe do AvoidsNodes routing
    if (this.maybeAvoidsNodes(startFrom, fromDir, m2, m3, endTo, toDir, fromnode, tonode, fromR, toR)) return;
    // otherwise use the default ortho routing
    this.addPoint(m2);
    this.addPoint(m3);
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} startFrom this {@link Point} will already have been added to the stroke by {@link #computePoints} before calling this method.
   * @param {number} fromDir Normally 0, 90, 180, or 270 degrees.
   * @param {Point} m2 first default ortho midpoint.
   * @param {Point} m3 second default ortho midpoint.
   * @param {Point} endTo {@link #computePoints} will add this {@link Point} after calling this method.
   * @param {number} toDir Normally 0, 90, 180, or 270 degrees.
   * @param {Node} fromnode the {@link Node} that the link is coming from.
   * @param {Node} tonode the {@link Node} that the link is going to.
   * @param {Rect} fromR inflated fromnode.actualBounds.
   * @param {Rect} toR inflated tonode.actualBounds.
   * @return {boolean}
   */
  private maybeAvoidsNodes(startFrom: Point, fromDir: number, m2: Point, m3: Point, endTo: Point, toDir: number, fromnode: Node, tonode: Node, fromR: Rect, toR: Rect): boolean {
    if (!this.isAvoiding) return false;
    const diagram = this.diagram;
    if (diagram === null) return false;
    if (!diagram.canAvoidNodes()) return false;
    if (fromR.containsPoint(endTo) && !tonode.isMemberOf(fromnode)) return false;  // ignore AvoidsNodes if overlapping
    if (toR.containsPoint(startFrom) && !fromnode.isMemberOf(tonode)) return false;  // ignore AvoidsNodes if overlapping
    if (fromnode === tonode || this.layer.isTemporary) return false;  // ignore AvoidsNodes if self-loop or temporary
    // ??? if (this.delayedRouting) return false;  // or when rebuilding

    const positions = diagram.getPositions(true, this.containingGroup, null);

    // see if the current route crosses over any nodes
    if (// from startFrom to m2
      positions.isUnoccupied(Math.min(startFrom.x, m2.x), Math.min(startFrom.y, m2.y), Math.abs(startFrom.x - m2.x), Math.abs(startFrom.y - m2.y)) &&
      // from m2 to m3
      positions.isUnoccupied(Math.min(m2.x, m3.x), Math.min(m2.y, m3.y), Math.abs(m2.x - m3.x), Math.abs(m2.y - m3.y)) &&
      // from m3 to endTo
      positions.isUnoccupied(Math.min(m3.x, endTo.x), Math.min(m3.y, endTo.y), Math.abs(m3.x - endTo.x), Math.abs(m3.y - endTo.y))) {
      return false;
    }

    const origStart = startFrom;
    const origEnd = endTo;
    let dummyEnter: Point = null;
    let dummyLeave: Point = null;
    if (diagram.isVirtualized) {
      const r = positions.bounds.copy();
      r.inflate(-positions.cellWidth, -positions.cellHeight);
      const temp = Point.alloc();
      if (!positions.inBounds(startFrom.x, startFrom.y)) {
        if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, startFrom.x, startFrom.y, m2.x, m2.y, temp)) {
          dummyEnter = startFrom = temp.copy(); fromDir = temp.directionPoint(m2);
        } else if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, m2.x, m2.y, m3.x, m3.y, temp)) {
          dummyEnter = startFrom = temp.copy(); fromDir = temp.directionPoint(m3);
        } else if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, m3.x, m3.y, endTo.x, endTo.y, temp)) {
          dummyEnter = startFrom = temp.copy(); fromDir = temp.directionPoint(endTo);
        }
      }
      if (!positions.inBounds(endTo.x, endTo.y)) {
        if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, endTo.x, endTo.y, m3.x, m3.y, temp)) {
          dummyLeave = endTo = temp.copy(); toDir = m3.directionPoint(temp);
        } else if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, m3.x, m3.y, m2.x, m2.y, temp)) {
          dummyLeave = endTo = temp.copy(); toDir = m2.directionPoint(temp);
        } else if (Geo.getNearestIntersectionPoint(r.x, r.y, r.x + r.width, r.y + r.height, m2.x, m2.y, startFrom.x, startFrom.y, temp)) {
          dummyLeave = endTo = temp.copy(); toDir = startFrom.directionPoint(temp);
        }
      }
      Point.free(temp);
    }

    // first try to find a path that is within the bounds of the
    // rectangle formed by the two nodes--this should cover most cases
    // more efficiently than searching the whole document
    const minbounds = fromR.copy().unionRect(toR);
    let margin = positions.smallMargin;
    minbounds.inflate(positions.cellWidth * margin, positions.cellHeight * margin);
    positions.propagate(startFrom, fromDir, endTo, toDir, minbounds);  // may set .abort
    let endval = positions.getDistance(endTo.x, endTo.y);
    if (!positions.abort && endval >= PositionArray.MAX) {
      // try again with larger area, which should cover a lot of cases where
      // there is another node overlapping with either the source or the destination node
      positions.clearAllUnoccupied();
      margin = positions.largeMargin;
      minbounds.inflate(positions.cellWidth * margin, positions.cellHeight * margin);
      positions.propagate(startFrom, fromDir, endTo, toDir, minbounds);  // may set .abort
      endval = positions.getDistance(endTo.x, endTo.y);
    }
    if (!positions.abort && endval >= PositionArray.MAX && positions.wholeDocument) {
      // try again with larger area, the whole document
      positions.clearAllUnoccupied();
      positions.propagate(startFrom, fromDir, endTo, toDir, positions.bounds);  // may set .abort
      endval = positions.getDistance(endTo.x, endTo.y);
    }

    // if the propagation succeeded, get the points of the stroke
    if (!positions.abort && endval < PositionArray.MAX && !positions.isOccupied(endTo.x, endTo.y)) {
      this.traversePositions(positions, endTo.x, endTo.y, toDir, true);

      const RIGHT = 0;
      const DOWN = 90;
      const LEFT = 180;
      const UP = 270;

      // hook up the first segment with the grid-aligned points produced by TraversePositions
      const two = this.getPoint(2);
      if (this.pointsCount < 4) {
        if (fromDir === RIGHT || fromDir === LEFT) {
          two.x = startFrom.x;
          two.y = endTo.y;
        } else {
          two.x = endTo.x;
          two.y = startFrom.y;
        }
        this.setPointAt(2, two.x, two.y);
        this.insertPointAt(3, two.x, two.y);
      } else {
        const three = this.getPoint(3);
        if (fromDir === RIGHT || fromDir === LEFT) {
          if (Geo.isApprox(two.x, three.x)) {
            const newx = (fromDir === RIGHT) ? Math.max(two.x, startFrom.x) : Math.min(two.x, startFrom.x);
            this.setPointAt(2, newx, startFrom.y);
            this.setPointAt(3, newx, three.y);
          } else if (Geo.isApprox(two.y, three.y)) {
            if (Math.abs(startFrom.y - two.y) <= positions.cellHeight / 2) {
              this.setPointAt(2, two.x, startFrom.y);
              this.setPointAt(3, three.x, startFrom.y);
            }
            this.insertPointAt(2, two.x, startFrom.y);
          } else {
            this.setPointAt(2, startFrom.x, two.y);
          }
        } else if (fromDir === DOWN || fromDir === UP) {
          if (Geo.isApprox(two.y, three.y)) {
            const newy = (fromDir === DOWN) ? Math.max(two.y, startFrom.y) : Math.min(two.y, startFrom.y);
            this.setPointAt(2, startFrom.x, newy);
            this.setPointAt(3, three.x, newy);
          } else if (Geo.isApprox(two.x, three.x)) {
            if (Math.abs(startFrom.x - two.x) <= positions.cellWidth / 2) {
              this.setPointAt(2, startFrom.x, two.y);
              this.setPointAt(3, startFrom.x, three.y);
            }
            this.insertPointAt(2, startFrom.x, two.y);
          } else {
            this.setPointAt(2, two.x, startFrom.y);
          }
        }
      }

      if (dummyEnter !== null) {
        const p1 = this.getPoint(1);
        const p2 = this.getPoint(2);
        if (p1.x !== p2.x && p1.y !== p2.y) {
          if (fromDir === RIGHT || fromDir === LEFT) {
            this.insertPointAt(2, p1.x, p2.y);
          } else {
            this.insertPointAt(2, p2.x, p1.y);
          }
        } else if (fromDir === RIGHT || fromDir === LEFT) {
          this.insertPointAt(2, origStart.x, dummyEnter.y);
        } else {
          this.insertPointAt(2, dummyEnter.x, origStart.y);
        }
      }
      if (dummyLeave !== null) {
        if (toDir === RIGHT || toDir === LEFT) {
          this.addPointAt(origEnd.x, dummyLeave.y);
        } else {
          this.addPointAt(dummyLeave.x, origEnd.y);
        }
      }
      return true;  // success with special route
    }
    // otherwise, just depend on default behavior
    return false;
  }

  //  * This method is called by {@link #addOrthoPoints} to determine the distance
  //  * of the middle segment between the two ports.
  /**
   * @hidden
   * Undocumented
   * @expose
   * @this {Link}
   * @param {number} fromX The first point's X coordinate.
   * @param {number} fromY The first point's Y coordinate.
   * @param {number} toX The last point's X coordinate.
   * @param {number} toY The last point's Y coordinate.
   * @param {boolean} vertical Whether the mid-position is along the vertical axis or horizontal axis.
   * @return {number} must be a real number, not NaN or infinity
   */
  protected computeMidOrthoPosition(fromX: number, fromY: number, toX: number, toY: number, vertical: boolean): number {
    let c = 0.0;
    if (this.hasCurviness()) {
      c = this.computeCurviness();
    }
    if (vertical) {
      return (fromY + toY) / 2 + c;
    } else {
      return (fromX + toX) / 2 + c;
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @return {boolean}  false if not AvoidsNodes or if no non-end segment intersects with any avoidable Nodes
   */
  public crossesOverNode(): boolean {
    if (this.diagram === null) return false;
    if (!this.isAvoiding) return false;
    if (!this.diagram.canAvoidNodes()) return false;
    const a = this.points._arr;
    const len = a.length;
    if (len < 4) return false;  // assume AvoidsNodes routes always include 4+ points
    const positions = this.diagram.getPositions(true, this.containingGroup, null);
    for (let i = 1; i < len - 2; i++) {
      const p = a[i];
      const q = a[i + 1];
      if (!positions.isUnoccupied(Math.min(p.x, q.x), Math.min(p.y, q.y), Math.abs(p.x - q.x), Math.abs(p.y - q.y))) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {PositionArray} positions
   * @param {number} px
   * @param {number} py
   * @param {number} dir
   * @param {boolean} first
   */
  private traversePositions(positions: PositionArray, px: number, py: number, dir: number, first: boolean): void {
    const RIGHT = 0;
    const DOWN = 90;
    const LEFT = 180;
    const UP = 270;

    const cellw = positions.cellWidth;
    const cellh = positions.cellHeight;
    let val = positions.getDistance(px, py);
    let qx = px;
    let qy = py;

    // try going forward
    let fx = qx;
    let fy = qy;
    if (dir === RIGHT) {
      fx += cellw;
    } else if (dir === DOWN) {
      fy += cellh;
    } else if (dir === LEFT) {
      fx -= cellw;
    } else {
      fy -= cellh;
    }
    // keep going forward until we can't
    while (val > PositionArray.START) {
      const distf = positions.getDistance(fx, fy);
      if (distf !== val - 1 /*PositionArray.STEP*/) break;
      qx = fx;
      qy = fy;
      if (dir === RIGHT) {
        fx += cellw;
      } else if (dir === DOWN) {
        fy += cellh;
      } else if (dir === LEFT) {
        fx -= cellw;
      } else {
        fy -= cellh;
      }
      val -= 1 /*PositionArray.STEP*/;
    }

    // line up the points to be in the middle of the grid
    if (first) {
      if (val > PositionArray.START) {
        if (dir === LEFT || dir === RIGHT) {
          qx = Math.floor(qx / cellw) * cellw + cellw / 2;
        } else if (dir === DOWN || dir === UP) {
          qy = Math.floor(qy / cellh) * cellh + cellh / 2;
        }
      }
    } else {
      qx = Math.floor(qx / cellw) * cellw + cellw / 2;
      qy = Math.floor(qy / cellh) * cellh + cellh / 2;
    }

    if (val > PositionArray.START) {
      let newdir = dir;
      // try turning right
      let rx = qx;
      let ry = qy;
      if (dir === RIGHT) {
        newdir = DOWN;
        ry += cellh;
      } else if (dir === DOWN) {
        newdir = LEFT;
        rx -= cellw;
      } else if (dir === LEFT) {
        newdir = UP;
        ry -= cellh;
      } else if (dir === UP) {
        newdir = RIGHT;
        rx += cellw;
      }
      const distr = positions.getDistance(rx, ry);
      if (distr === val - 1 /*PositionArray.STEP*/) {
        this.traversePositions(positions, rx, ry, newdir, false);
      } else {
        // try turning left
        let lx = qx;
        let ly = qy;
        if (dir === RIGHT) {
          newdir = UP;
          ly -= cellh;
        } else if (dir === DOWN) {
          newdir = RIGHT;
          lx += cellw;
        } else if (dir === LEFT) {
          newdir = DOWN;
          ly += cellh;
        } else if (dir === UP) {
          newdir = LEFT;
          lx -= cellw;
        }
        const distl = positions.getDistance(lx, ly);
        if (distl === val - 1 /*PositionArray.STEP*/) {
          this.traversePositions(positions, lx, ly, newdir, false);
        }
      }
    }
    this.addPointAt(qx, qy);
  }

  /**
   * Find the index of the segment that is closest to a given point.
   * This assume the route only has straight line segments.
   * It ignores any jump-overs or jump-gaps.
   * @export
   * @this {Link}
   * @param {Point} p the Point, in document coordinates.
   * @return {number} int the index of the segment, from zero to the number of points minus 2.
   */
  public findClosestSegment(p: Point): number {
    if (Debug) Util.checkClass(p, Point, Link, 'findClosestSegment:p');
    const px = p.x;
    const py = p.y;
    let q = this.getPoint(0);
    let l = this.getPoint(1);
    let cdist = Point.distanceLineSegmentSquared(px, py, q.x, q.y, l.x, l.y);
    let ci = 0;  // closest index
    for (let i = 1; i < this.pointsCount - 1; i++) {
      q = this.getPoint(i + 1);
      const dist = Point.distanceLineSegmentSquared(px, py, l.x, l.y, q.x, q.y);
      l = q;
      if (dist < cdist) {
        ci = i;
        cdist = dist;
      }
    }
    return ci;
  }

  // the Geometry for the Route

  /**
   * @hidden @internal
   * @this {Link}
   */
  public invalidateGeometry(): void {
    // this._geometry = null;
    this._invalidGeometry = true;
  }

  /**
   * This read-only property returns the {@link Geometry} that is used by the {@link #path}, the link {@link Shape} based on the route points.
   *
   * This geometry is automatically generated using the route points and other properties
   * such as the {@link #curve}.
   * The points of the Geometry are in local coordinates, whereas the {@link #points} of the link route are in document coordinates.
   *
   * More than one {@link Shape} may share this geometry as its {@link Shape#geometry},
   * if there is more than one shape element in the Link with {@link GraphObject#isPanelMain} set to true.
   */
  /** @export */ get geometry(): Geometry {
    if (this._invalidGeometry) {
      this.updateRoute();
      this._geometry = this.makeGeometry();
    }
    return this._geometry;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {boolean} force true if you want to re-create the geometry even if one exists and the route is valid
   */
  public makeShapeGeometry(force: boolean): void {
    if (!force) {
      if (this.validRoute === false) return;
      const main = this.findMainElement() as Shape;
      if (/*this._geometry !== null*/ !this._invalidGeometry && (main === null || main.geometry !== null)) return;
    }

    this._geometry = this.makeGeometry();
    const stroke = this.path;  // may be null
    if (stroke === null) return; // possible there are no children of a link
    stroke._geometry = this._geometry;  // shared Geometry

    const children = this._elements._arr;
    const l = children.length;
    for (let i = 0; i < l; i++) {
      const e = children[i];
      if (e === stroke) continue;  // already measured the stroke, above
      // Take all shapes that have panelMain = true and give them a specific, shared geometry
      if (e.isPanelMain && e instanceof Shape) {
        e._geometry = this._geometry;  // all must share same Geometry
      }
    }
  }

  /**
   * Produce a {@link Geometry} given the points of this route,
   * depending on the value of {@link #curve} and {@link #corner} and perhaps other properties.
   * The points of the Geometry are in local coordinates, whereas the {@link #points} of the link route are in document coordinates.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Link}
   * @return {Geometry}
   * @since 1.6
   */
  public makeGeometry(): Geometry {
    const geo = this._geometry;
    const numpts = this.pointsCount;
    if (numpts < 2) {
      // return new Geometry(Geometry.Line);
      geo.type = Geometry.Line;
      this._invalidGeometry = false;
      return geo;
    }

    // if there are jumpovers, make a simplified geometry without them instead
    // // if its in _delayedGeometries, it will get created fully at draw time
    let makeSimplifiedGeo = false;
    const diagram = this.diagram;
    if (diagram !== null && this._jumpsOver() && diagram._delayedGeometries.contains(this) &&
       (this._previousBounds.width !== 0 || this._previousBounds.height !== 0)) {
      makeSimplifiedGeo = true;
    }

    let bx = 0.0;
    let by = 0.0;
    let p0 = this.getPoint(0).copy();
    const p = p0.copy();
    const ptsarr = this._points._arr;
    const curve = this.computeCurve();
    if (curve === Link.Bezier && numpts >= 3 && !Geo.isApproxEqual(this.smoothness, 0)) {
      if (numpts === 3) {
        let e = this.getPoint(1);
        bx = Math.min(p0.x, e.x);
        by = Math.min(p0.y, e.y);
        e = this.getPoint(2);
        bx = Math.min(bx, e.x);
        by = Math.min(by, e.y);
      } else {
        if (this.isOrthogonal) {
          for (let i = 0; i < numpts; i++) {
            const p2 = ptsarr[i];
            p.x = Math.min(p2.x, p.x);
            p.y = Math.min(p2.y, p.y);
          }
          bx = p.x;
          by = p.y;
        } else {
          for (let i = 3; i < numpts; i += 3) {
            // if it's the last segment, use the last two points
            if (i + 3 >= numpts) {
              i = numpts - 1;
            }
            const end = this.getPoint(i);
            p.x = Math.min(end.x, p.x);
            p.y = Math.min(end.y, p.y);
          }
          bx = p.x;
          by = p.y;
        }
      }
    } else {
      for (let i = 0; i < numpts; i++) {
        const p2 = ptsarr[i];
        p.x = Math.min(p2.x, p.x);
        p.y = Math.min(p2.y, p.y);
      }
      bx = p.x;
      by = p.y;
    }

    // in case the geometric origin is not actually a real point, because of a bezier/quadtratic
    bx -= this._tempOrigin.x;
    by -= this._tempOrigin.y;

    // need to shift all points back by bx/by, since the Link/LinkLayer will be translating them
    p0.x -= bx;
    p0.y -= by;

    if (numpts === 2 && !this._jumpsOver()) {  // can't jumpover or jumpgap using Geometry.Line
      let p1 = this.getPoint(1).copy();
      p1.x -= bx;
      p1.y -= by;
      if (this.computeShortLength(true) !== 0) p0 = this._shiftPointByShortLength(p0, true, p);
      if (this.computeShortLength(false) !== 0) p1 = this._shiftPointByShortLength(p1, false, p);
      // const geo = new Geometry(Geometry.Line);
      geo.type = Geometry.Line;
      geo.startX = p0.x;
      geo.startY = p0.y;
      geo.endX = p1.x;
      geo.endY = p1.y;
      this._invalidGeometry = false;
      return geo;
    } else {
      geo.type = Geometry.Path;
      const g = StreamGeometryContext.tempStreamGeometryContext(geo);

      if (this.computeShortLength(true) !== 0) p0 = this._shiftPointByShortLength(p0, true, p);
      g.beginFigure(p0.x, p0.y, false, false);
      if (curve === Link.Bezier && numpts >= 3 && !Geo.isApproxEqual(this.smoothness, 0)) {
        if (numpts === 3) {
          const c = this.getPoint(1);
          const cx = c.x - bx;
          const cy = c.y - by;
          let e = this.getPoint(2).copy();
          e.x -= bx;
          e.y -= by;
          if (this.computeShortLength(false) !== 0) e = this._shiftPointByShortLength(e, false, p);
          g.bezierTo(cx, cy, cx, cy, e.x, e.y /*, true, true*/);
        } else {
          if (this.isOrthogonal) {
            const cp10 = new Point(bx, by);
            const cp11 = this.getPoint(1).copy();
            const cp20 = new Point(bx, by);
            const cp21 = new Point(bx, by);
            let prev = this.getPoint(0);
            let curr: Point = null;
            // Precalculate this so it isn't calculated with every call of _getOrthogonalBezierControls
            const controlComponentLength = this.smoothness / 3;

            for (let i = 1; i < this.pointsCount - 1; i++) {
              curr = this.getPoint(i);

              const cp0 = prev;
              const cp1 = curr;
              const cp2 = this.getPoint(this._furthestPoint(curr, i, false));
              // If the curr lies on a straight line with the previous and next points, ignore it. One of these conditionals will be true if
              if (Geo.isApproxEqual(cp0.x, cp1.x) && Geo.isApproxEqual(cp1.x, cp2.x)) {
                continue;
              }
              if (Geo.isApproxEqual(cp0.y, cp1.y) && Geo.isApproxEqual(cp1.y, cp2.y)) {
                continue;
              }

              this._getOrthogonalBezierControls(cp0, cp1, cp2, controlComponentLength, cp20, cp21);

              if (i === 1) { // special case where there is no intermediate point between the start point and the first corner.
                cp11.x = (prev.x + curr.x) * 0.5;
                cp11.y = (prev.y + curr.y) * 0.5;
              } else if (i === 2 && // case where point 1 is on the line connecting point 2 and point 0, use midpoint of line as control point
                Geo.isApproxEqual(prev.x, this.getPoint(0).x) &&
                Geo.isApproxEqual(prev.y, this.getPoint(0).y)) {
                cp11.x = (prev.x + curr.x) * 0.5;
                cp11.y = (prev.y + curr.y) * 0.5;
              }

              // Account for control point overlap due to large values of Smoothness and/or small segments

              g.bezierTo(cp11.x - bx, cp11.y - by, cp20.x - bx, cp20.y - by, curr.x - bx, curr.y - by /*, true, false*/);

              cp10.set(cp20);
              cp11.set(cp21);
              prev = curr;
            }

            // Point endControl = this.getPoint(this.pointsCount-2).copy();
            let endControlX = prev.x;
            let endControlY = prev.y;
            let endPoint = this.getPoint(this.pointsCount - 1);
            if (this.computeShortLength(false) !== 0) endPoint = this._shiftPointByShortLength(endPoint.copy(), false, Point.Origin);
            // if (Geo.isApproxEqual(endControl, prev))
            endControlX = 0.5 * (endControlX + endPoint.x);
            endControlY = 0.5 * (endControlY + endPoint.y);

            g.bezierTo(cp21.x - bx, cp21.y - by, endControlX - bx, endControlY - by, endPoint.x - bx, endPoint.y - by /*, true, false*/);
          } else {
            for (let i = 3; i < numpts; i += 3) {
              const startControl2 = this.getPoint(i - 2);
              // if it's the last segment, use the last two points
              if (i + 3 >= numpts) {
                i = numpts - 1;
              }
              const endControl2 = this.getPoint(i - 1);
              let end2 = this.getPoint(i);
              if (i === numpts - 1 && this.computeShortLength(false) !== 0) end2 = this._shiftPointByShortLength(end2.copy(), false, Point.Origin);
              g.bezierTo(startControl2.x - bx, startControl2.y - by,
                endControl2.x - bx, endControl2.y - by,
                end2.x - bx, end2.y - by /*, true, true*/);
            }
          }
        }
      } else {
        const from = Point.alloc();
        from.assign(this.getPoint(0));
        let i = 1;
        let j = 0;
        while (i < numpts) {
          i = this._furthestPoint(from, i, (i > 1));  // skip over any intermediate points along the line
          let to = this.getPoint(i);
          if (i >= numpts - 1) {
            if (!from.equals(to)) {  // line actually goes somewhere
              if (this.computeShortLength(false) !== 0) to = this._shiftPointByShortLength(to.copy(), false, Point.Origin);
              this._addLine(g, -bx, -by, from, to, makeSimplifiedGeo);
            } else if (j === 0) {  // from==to && all points along one line
              i = 1;
              while (i < numpts) {  // don't short-circuit drawing
                to = this.getPoint(i++);
                this._addLine(g, -bx, -by, from, to, makeSimplifiedGeo);
                from.assign(to);
              }
            }
            break;
          }
          j = this._furthestPoint(to, i + 1, (i < numpts - 3));
          const next = this.getPoint(j);
          this._addLineAndCorner(g, -bx, -by, from, to, next, from, makeSimplifiedGeo);
          i = j;
        }
        Point.free(from);
      }
      // const geo = g.geo;
      StreamGeometryContext.freeStreamGeometryContext(g);
      this._invalidGeometry = false;
      return geo;
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {number} ax
   * @param {number} ay
   * @param {number} bx
   * @param {number} by
   * @return {number}
   */
  private _getDistanceBetweenLinkPoints(ax: number, ay: number, bx: number, by: number): number {
    let dx = bx - ax;
    if (isNaN(dx) || dx === Infinity || dx === -Infinity) return NaN;
    if (dx < 0) dx = -dx;
    let dy = by - ay;
    if (isNaN(dy) || dy === Infinity || dy === -Infinity) return NaN;
    if (dy < 0) dy = -dy;
    if (Geo.isApproxEqual(dx, 0)) return dy;
    if (Geo.isApproxEqual(dy, 0)) return dx;
    return Math.sqrt(dx * dx + dy * dy);
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} a
   * @param {Point} b
   * @param {Point} c
   * @param {number} ratio the control point distance ratio.
   * @param {Point} cp0 a Point to be modified to hold the results.
   * @param {Point} cp1 a Point to be modified to hold the results.
   */
  private _getOrthogonalBezierControls(a: Point, b: Point, c: Point, ratio: number, cp0: Point, cp1: Point): void {
    if (isNaN(ratio)) { // set to default value
      ratio = this.smoothness / 3;
    }

    const ax = a.x;
    const ay = a.y;
    const bx = b.x;
    const by = b.y;
    const cx = c.x;
    const cy = c.y;

    const length1 = ratio * this._getDistanceBetweenLinkPoints(ax, ay, bx, by);
    const length2 = ratio * this._getDistanceBetweenLinkPoints(bx, by, cx, cy);

    if (Geo.isApproxEqual(ay, by) && Geo.isApproxEqual(bx, cx)) {  // horizontal then vertical
      if (bx > ax) {  // right
        if (cy > by) {  // right then down
          cp0.x = bx - length1;
          cp0.y = by - length1;
          cp1.x = bx + length2;
          cp1.y = by + length2;
        } else {  // right then up
          cp0.x = bx - length1;
          cp0.y = by + length1;
          cp1.x = bx + length2;
          cp1.y = by - length2;
        }
      } else {  // left
        if (cy > by) {  // left then down
          cp0.x = bx + length1;
          cp0.y = by - length1;
          cp1.x = bx - length2;
          cp1.y = by + length2;
        } else {  // left then up
          cp0.x = bx + length1;
          cp0.y = by + length1;
          cp1.x = bx - length2;
          cp1.y = by - length2;
        }
      }
    }

    if (Geo.isApproxEqual(ax, bx) && Geo.isApproxEqual(by, cy)) {  // vertical then horizontal
      if (by > ay) {  // down
        if (cx > bx) {  // down then right
          cp0.x = bx - length1;
          cp0.y = by - length1;
          cp1.x = bx + length2;
          cp1.y = by + length2;
        } else {  // down then left
          cp0.x = bx + length1;
          cp0.y = by - length1;
          cp1.x = bx - length2;
          cp1.y = by + length2;
        }
      } else {  // up
        if (cx > bx) {  // up then right
          cp0.x = bx - length1;
          cp0.y = by + length1;
          cp1.x = bx + length2;
          cp1.y = by - length2;
        } else {  // up then left
          cp0.x = bx + length1;
          cp0.y = by + length1;
          cp1.x = bx - length2;
          cp1.y = by - length2;
        }
      }
    }

    if ((Geo.isApproxEqual(ax, bx) && Geo.isApproxEqual(bx, cx)) ||
      (Geo.isApproxEqual(ay, by) && Geo.isApproxEqual(by, cy))) {  // parallel segments
      const midpointX = 0.5 * (ax + cx);
      const midpointY = 0.5 * (ay + cy);
      cp0.x = midpointX;
      cp0.y = midpointY;
      cp1.x = midpointX;
      cp1.y = midpointY;
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} pt the Point to be modified with the result, in geometry coordinates.
   * @param {boolean} from
   * @param {Point} offset the translation from document to geometry coordinates.
   * @return {Point} the first argument
   */
  private _shiftPointByShortLength(pt: Point, from: boolean, offset: Point): Point {
    const numpts = this.pointsCount;
    if (numpts < 2) return pt;
    if (from) {
      const otherPt = this.getPoint(1);
      const otherX = otherPt.x - offset.x;
      const otherY = otherPt.y - offset.y;
      const segmentLength = this._getDistanceBetweenLinkPoints(pt.x, pt.y, otherX, otherY);
      if (segmentLength === 0) return pt;
      const max = (numpts === 2) ? segmentLength * 0.5 : segmentLength;
      let short = this.computeShortLength(true);
      if (short > max) short = max;
      const ox = short * (otherX - pt.x) / segmentLength;
      const oy = short * (otherY - pt.y) / segmentLength;
      pt.x += ox;
      pt.y += oy;
    } else {  // "to" end
      const otherPt = this.getPoint(numpts - 2);
      const otherX = otherPt.x - offset.x;
      const otherY = otherPt.y - offset.y;
      const segmentLength = this._getDistanceBetweenLinkPoints(pt.x, pt.y, otherX, otherY);
      if (segmentLength === 0) return pt;
      const max = (numpts === 2) ? segmentLength * 0.5 : segmentLength;
      let short = this.computeShortLength(false);
      if (short > max) short = max;
      const ox = short * (pt.x - otherX) / segmentLength;
      const oy = short * (pt.y - otherY) / segmentLength;
      pt.x -= ox;
      pt.y -= oy;
    }
    return pt;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} a
   * @param {number} i int
   * @param {boolean} oneway
   * @return {number} int
   */
  private _furthestPoint(a: Point, i: number, oneway: boolean): number {
    const numpts = this.pointsCount;
    // look for a different point, to give a direction from A to B
    let b = a;
    while (Geo.isApproxEqual(a.x, b.x) && Geo.isApproxEqual(a.y, b.y)) {
      if (i >= numpts) {
        return numpts - 1;
      }
      b = this.getPoint(i++);
    }
    // now A !== B, so we have a direction
    // make sure the direction is orthogonal
    if (!Geo.isApproxEqual(a.x, b.x) && !Geo.isApproxEqual(a.y, b.y)) {
      return i - 1;
    }
    // now a.x === b.x || a.y === b.y
    // keep going in the orthogonal direction
    let c = b;
    while ((Geo.isApproxEqual(a.x, b.x) && Geo.isApproxEqual(b.x, c.x) && (!oneway || (a.y >= b.y ? b.y >= c.y : b.y <= c.y))) ||
      (Geo.isApproxEqual(a.y, b.y) && Geo.isApproxEqual(b.y, c.y) && (!oneway || (a.x >= b.x ? b.x >= c.x : b.x <= c.x)))) {
      if (i >= numpts) {
        return numpts - 1;
      }
      c = this.getPoint(i++);
    }
    // now B !== C, so C is off the A-B line
    return i - 2;
  }

  // ?? NYI only perpendicular lines supported
  /**
   * @hidden @internal
   * @this {Link}
   * @param {StreamGeometryContext} g
   * @param {number} offx
   * @param {number} offy
   * @param {Point} a
   * @param {Point} b
   * @param {Point} c
   * @param {Point} end this is modified to be the end point of this operation
   * @param {boolean} makeSimplifiedGeo
   */
  private _addLineAndCorner(g: StreamGeometryContext, offx: number, offy: number, a: Point, b: Point, c: Point, end: Point, makeSimplifiedGeo: boolean): void {
    if (Geo.isApprox(a.y, b.y) && Geo.isApprox(b.x, c.x)) {  // horizontal then vertical
      const corner = this.computeCorner();
      let dx = Math.min(corner, Math.abs(b.x - a.x) / 2);
      const dy = Math.min(dx /* corner */, Math.abs(c.y - b.y) / 2);
      dx = dy;  // assume square corners
      if (Geo.isApprox(dx, 0)) {
        this._addLine(g, offx, offy, a, b, makeSimplifiedGeo);
        end.assign(b);
        return;
      }
      let arc1x = b.x;
      const arc1y = b.y;
      const arc2x = arc1x;
      let arc2y = arc1y;
      if (b.x > a.x) {  // right
        arc1x = b.x - dx;
        if (c.y > b.y) {  // right and down
          arc2y = b.y + dy;
        } else {  // right and up
          arc2y = b.y - dy;
        }
      } else {  // left
        arc1x = b.x + dx;
        if (c.y > b.y) {  // left and down
          arc2y = b.y + dy;
        } else {  // left and up
          arc2y = b.y - dy;
        }
      }
      this._addLine(g, offx, offy, a, new Point(arc1x, arc1y), makeSimplifiedGeo);
      g.quadraticBezierTo(b.x + offx, b.y + offy, arc2x + offx, arc2y + offy);
      end.assignTo(arc2x, arc2y);
    } else if (Geo.isApprox(a.x, b.x) && Geo.isApprox(b.y, c.y)) { // vertical then horizontal
      const corner = this.computeCorner();
      let dy = Math.min(corner, Math.abs(b.y - a.y) / 2);
      const dx = Math.min(dy /* corner */, Math.abs(c.x - b.x) / 2);
      dy = dx;  // assume square corners
      if (Geo.isApprox(dx, 0)) {
        this._addLine(g, offx, offy, a, b, makeSimplifiedGeo);
        end.assign(b);
        return;
      }
      const arc1x = b.x;
      let arc1y = b.y;
      let arc2x = arc1x;
      const arc2y = arc1y;
      if (b.y > a.y) {  // down
        arc1y = b.y - dy;
        if (c.x > b.x) {  // down and right
          arc2x = b.x + dx;
        } else {  // down and left
          arc2x = b.x - dx;
        }
      } else {  // up
        arc1y = b.y + dy;
        if (c.x > b.x) {  // up and right
          arc2x = b.x + dx;
        } else {  // up and left
          arc2x = b.x - dx;
        }
      }
      this._addLine(g, offx, offy, a, new Point(arc1x, arc1y), makeSimplifiedGeo);
      g.quadraticBezierTo(b.x + offx, b.y + offy, arc2x + offx, arc2y + offy);
      end.assignTo(arc2x, arc2y);
    } else {
      this._addLine(g, offx, offy, a, b, makeSimplifiedGeo);
      end.assign(b);
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {StreamGeometryContext} g
   * @param {number} offx
   * @param {number} offy
   * @param {Point} from
   * @param {Point} to
   * @param {boolean} makeSimplifiedGeo
   */
  private _addLine(g: StreamGeometryContext, offx: number, offy: number, from: Point, to: Point, makeSimplifiedGeo: boolean): void {
    if (makeSimplifiedGeo || !this._jumpsOver()) {
      g.lineTo(to.x + offx, to.y + offy /*, true, true*/);
    } else {
      const curve = 10.0;  // ??
      const curve2 = curve / 2;
      const vec: Array<number> = [];
      let numints = 0;
      if (this.isVisible()) numints = this._getIntersections(from, to, vec);
      let lastX = from.x;
      let lastY = from.y;
      if (numints > 0) {
        if (Geo.isApprox(from.y, to.y)) {
          // vec is sorted by increasing values
          if (from.x < to.x) {
            let j = 0;
            while (j < numints) {
              const closer = Math.max(from.x, Math.min(vec[j++] - curve2, to.x - curve));
              g.lineTo(closer + offx, to.y + offy /*, true, true*/);
              lastX = closer + offx;
              lastY = to.y + offy;
              let farther = Math.min(closer + curve, to.x);
              while (j < numints) {
                const next = vec[j];
                if (next < farther + curve) {
                  j++;
                  farther = Math.min(next + curve2, to.x);
                } else {
                  break;  // leave J as is for next lineTo call
                }
              }
              const controlY = to.y - curve + offy;
              const endX = farther + offx;
              const endY = to.y + offy;
              if (this.curve === Link.JumpGap) {
                g.beginFigure(endX, endY, false, false);
              } else { // this.curve === Link.JumpOver
                g.bezierTo(lastX, controlY, endX, controlY, endX, endY /*, true, true*/);
              }
              lastX = endX;
              lastY = endY;
            }
          } else {  // from.x > to.x
            let j = numints - 1;
            while (j >= 0) {
              const closer = Math.min(from.x, Math.max(vec[j--] + curve2, to.x + curve));
              g.lineTo(closer + offx, to.y + offy /*, true, true*/);
              lastX = closer + offx;
              lastY = to.y + offy;
              let farther = Math.max(closer - curve, to.x);
              while (j >= 0) {
                const next = vec[j];
                if (next > farther - curve) {
                  j--;
                  farther = Math.max(next - curve2, to.x);
                } else {
                  break;  // leave J as is for next lineTo call
                }
              }
              const controlY = to.y - curve + offy;
              const endX = farther + offx;
              const endY = to.y + offy;
              if (this.curve === Link.JumpGap) {
                g.beginFigure(endX, endY, false, false);
              } else { // this.curve === Link.JumpOver
                g.bezierTo(lastX, controlY, endX, controlY, endX, endY /*, true, true*/);
              }
              lastX = endX;
              lastY = endY;
            }
          }
        } else if (Geo.isApprox(from.x, to.x)) {
          // vec is sorted by increasing values
          if (from.y < to.y) {
            let j = 0;
            while (j < numints) {
              const closer = Math.max(from.y, Math.min(vec[j++] - curve2, to.y - curve));
              g.lineTo(to.x + offx, closer + offy /*, true, true*/);
              lastX = to.x + offx;
              lastY = closer + offy;
              let farther = Math.min(closer + curve, to.y);
              while (j < numints) {
                const next = vec[j];
                if (next < farther + curve) {
                  j++;
                  farther = Math.min(next + curve2, to.y);
                } else {
                  break;  // leave J as is for next lineTo call
                }
              }
              const controlX = to.x - curve + offx;
              const endX = to.x + offx;
              const endY = farther + offy;
              if (this.curve === Link.JumpGap) {
                g.beginFigure(endX, endY, false, false);
              } else { // this.curve === Link.JumpOver
                g.bezierTo(controlX, lastY, controlX, endY, endX, endY /*, true, true*/);
              }
              lastX = endX;
              lastY = endY;
            }
          } else {  // from.y > to.y
            let j = numints - 1;
            while (j >= 0) {
              const closer = Math.min(from.y, Math.max(vec[j--] + curve2, to.y + curve));
              g.lineTo(to.x + offx, closer + offy /*, true, true*/);
              lastX = to.x + offx;
              lastY = closer + offy;
              let farther = Math.max(closer - curve, to.y);
              while (j >= 0) {
                const next = vec[j];
                if (next > farther - curve) {
                  j--;
                  farther = Math.max(next - curve2, to.y);
                } else {
                  break;  // leave J as is for next lineTo call
                }
              }
              const controlX = to.x - curve + offx;
              const endX = to.x + offx;
              const endY = farther + offy;
              if (this.curve === Link.JumpGap) {
                g.beginFigure(endX, endY, false, false);
              } else { // this.curve === Link.JumpOver
                g.bezierTo(controlX, lastY, controlX, endY, endX, endY /*, true, true*/);
              }
              lastX = endX;
              lastY = endY;
            }
          }
        }
      }
      g.lineTo(to.x + offx, to.y + offy /*, true, true*/);
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} A
   * @param {Point} B
   * @param {Array.<number>} vec of numbers.
   * @return {number} int
   */
  private _getIntersections(A: Point, B: Point, vec: Array<number>): number {
    const diagram = this.diagram;
    if (diagram === null) return 0;

    if (A.equals(B)) return 0;

    const lit = diagram.layers;
    while (lit.next()) {
      const layer = lit.value;
      if (layer !== null && layer.visible) {
        const arr = layer.getOrthoLinksArray();
        const pitlen = arr.length;
        for (let pit = 0; pit < pitlen; pit++) {
          const otherpart = arr[pit];
          if (!(otherpart instanceof Link)) continue;
          // done searching for links underneath this one
          if (otherpart === this) {
            if (vec.length > 0) {
              vec.sort(
                /**
                 * @param {number} a
                 * @param {number} b
                 * @return {number}
                 */
                function (a: number, b: number): number { return a - b; });
            }
            return vec.length;
          }
          // ignore invisible links
          if (!otherpart.isVisible()) continue;
          // ignore links that don't also JumpOver or JumpGap
          if (!otherpart._jumpsOver()) continue;
          // ignore links that are far away
          const rb = otherpart.routeBounds;
          if (!rb.isReal()) continue;
          if (!this.routeBounds.intersectsRect(rb)) continue;
          // ignore links that connect to the same nodes as this link does
          if (this.usesSamePort(otherpart)) continue;
          // ignore link with no or invisible path
          const shape = otherpart.path;
          if (shape === null) continue;
          if (!shape.isVisibleObject()) continue;
          this._getIntersections2(A, B, vec, otherpart);
        }
      }
    }
    if (vec.length > 0) {
      vec.sort(
        /**
         * @param {number} a
         * @param {number} b
         * @return {number}
         */
        function (a: number, b: number): number { return a - b; });
    }
    return vec.length;
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} A
   * @param {Point} B known not be to equal to A
   * @param {Array.<number>} vec of numbers to hold intersections
   * @param {Link} route
   */
  private _getIntersections2(A: Point, B: Point, vec: Array<number>, route: Link): void {
    const horiz = Geo.isApprox(A.y, B.y);
    const numpts = route.pointsCount;
    let from = route.getPoint(0);
    const temp = Point.alloc();
    for (let i = 1; i < numpts; i++) {
      const to = route.getPoint(i);
      if (i < numpts - 1) {
        // handle case where consecutive segments are in a line going in the same direction
        const next = route.getPoint(i + 1);
        if (from.y === to.y && to.y === next.y) {
          if ((to.x > from.x && next.x >= to.x) || (to.x < from.x && next.x <= to.x)) {
            continue;
          }
        } else if (from.x === to.x && to.x === next.x) {
          if ((to.y > from.y && next.y >= to.y) || (to.y < from.y && next.y <= to.y)) {
            continue;
          }
        }
      }
      if (this._getOrthoSegmentIntersection(A, B, from, to, temp)) {
        if (horiz) {
          vec.push(temp.x);
        } else {
          vec.push(temp.y);
        }
      }
      from = to;
    }
    Point.free(temp);
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Point} A
   * @param {Point} B
   * @param {Point} C
   * @param {Point} D
   * @param {Point} result
   * @return {boolean}
   */
  private _getOrthoSegmentIntersection(A: Point, B: Point, C: Point, D: Point, result: Point): boolean {
    const Ax = A.x;
    const Ay = A.y;
    const Bx = B.x;
    const By = B.y;
    const Cx = C.x;
    const Cy = C.y;
    const Dx = D.x;
    const Dy = D.y;
    if (!Geo.isApprox(Ax, Bx)) {
      // AB is horizontal, so CD must be vertical
      if (Geo.isApprox(Ay, By) &&
        Geo.isApprox(Cx, Dx) &&
        Math.min(Ax, Bx) < Cx &&
        Math.max(Ax, Bx) > Cx &&
        Math.min(Cy, Dy) < Ay &&
        Math.max(Cy, Dy) > Ay &&
        !Geo.isApprox(Cy, Dy)) {
        result.x = Cx;
        result.y = Ay;
        return true;
      }
    } else {
      // A.x === B.x : AB is vertical, so CD must be horizontal
      if (!Geo.isApprox(Ay, By) &&
        Geo.isApprox(Cy, Dy) &&
        Math.min(Ay, By) < Cy &&
        Math.max(Ay, By) > Cy &&
        Math.min(Cx, Dx) < Ax &&
        Math.max(Cx, Dx) > Ax &&
        !Geo.isApprox(Cx, Dx)) {
        result.x = Ax;
        result.y = Cy;
        return true;
      }
    }
    result.x = 0;
    result.y = 0;
    return false;
  }

  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get firstPickIndex(): number {
    const n = this.pointsCount;
    if (n <= 2) return 0;
    if (this.isOrthogonal || !this.isSpotUndirected(this.computeSpot(true))) return 1;
    return 0;
  }

  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get lastPickIndex(): number {
    const n = this.pointsCount;
    if (n === 0) return 0;
    if (n <= 2) return n - 1;
    if (this.isOrthogonal || !this.isSpotUndirected(this.computeSpot(false))) return n - 2;
    return n - 1;
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public _jumpsOver(): boolean {
    const c = this.curve;
    return (c === Link.JumpOver || c === Link.JumpGap);
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {boolean} force
   */
  private _invalidateOtherJumpOversMaybe(force: boolean): void {
    if (force || this._jumpsOver()) {
      const diagram = this.diagram;
      if (diagram !== null &&  !diagram.animationManager.isTicking && /* don't do during animation or doCompletion, when its restoring routes */
        !diagram._delayedGeometries.contains(this) && (this._previousBounds.width !== 0 || this._previousBounds.height !== 0)) {
        diagram._delayedGeometries.add(this, this._previousBounds.copy());
      }
    }
  }

  // * Maybe invalidate the geometries of other links that jump or gap over this link

  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Link}
   * @param {Rect} prevbounds
   */
  public invalidateOtherJumpOvers(prevbounds: Rect): void {
    const thislayer = this.layer;
    if (thislayer === null || !thislayer.visible || thislayer.isTemporary) return;
    const diagram = thislayer.diagram;
    if (diagram === null) return;
    if (diagram.animationManager.isAnimating) return;
    let foundlayer = false;
    const lit = diagram.layers;
    while (lit.next()) {
      const layer = lit.value;
      // only consider visible LinkLayers
      if (!layer.visible) continue;
      // skip over layers before THISLAYER
      if (layer === thislayer) {
        foundlayer = true;
        let foundlink = false;
        const arr = layer.getOrthoLinksArray();
        const pitlen = arr.length;
        for (let pit = 0; pit < pitlen; pit++) {
          const part = arr[pit];
          if (!(part instanceof Link)) continue;
          // skip over links before THISLINK
          if (part === this) {
            foundlink = true;
          } else if (foundlink) {  // for any Link after THISLINK
            this._invalidateJumpOvers(part, prevbounds);
          }
        }
      } else if (foundlayer) {  // for any LinkLayer after THISLAYER
        const arr = layer.getOrthoLinksArray();
        const pitlen = arr.length;
        for (let pit = 0; pit < pitlen; pit++) {
          const part = arr[pit];
          if (!(part instanceof Link)) continue;
          this._invalidateJumpOvers(part, prevbounds);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @param {Link} otherlink
   * @param {Rect} prevbounds
   */
  private _invalidateJumpOvers(otherlink: Link, prevbounds: Rect): void {
    // OTHERLINK needs to have a JumpOver or JumpGap curve and intersecting bounds
    if (otherlink === null) return;
    if (otherlink._geometry === null) return;
    if (!otherlink._jumpsOver()) return;
    const rb = otherlink.routeBounds;
    if (!rb.isReal()) return;
    if (!this.routeBounds.intersectsRect(rb) && !prevbounds.intersectsRect(rb)) return;
    // if it uses the same port, don't jump-over
    if (this.usesSamePort(otherlink)) return;
    otherlink.invalidateGeometry();
  }

  /**
   * @hidden @internal
   * @expose
   * @this {Link}
   * @param {Link} otherlink
   * @return {boolean}
   */
  public usesSamePort(otherlink: Link): boolean {
    const numpts = this.pointsCount;
    const otherpts = otherlink.pointsCount;
    if (numpts > 0 && otherpts > 0) {
      // assume no jump-overs when links come from or go to the same point;
      // the Points are in document coordinates, not the local coordinates of geometries
      const start = this.getPoint(0);
      const ostart = otherlink.getPoint(0);
      if (start.equalsApprox(ostart)) return true;
      const end = this.getPoint(numpts - 1);
      const oend = otherlink.getPoint(otherpts - 1);
      if (end.equalsApprox(oend)) return true;
      // handle going in either direction
      if (start.equalsApprox(oend)) return true;
      if (end.equalsApprox(ostart)) return true;
    } else {
      // assume no jump-overs when crossing a link from/to the same node
      if (this.fromNode === otherlink.fromNode) return true;
      if (this.toNode === otherlink.toNode) return true;
      // handle going in either direction
      if (this.fromNode === otherlink.toNode) return true;
      if (this.toNode === otherlink.fromNode) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Node}
   * @return {boolean}
   */
  public isVisible(): boolean {
    if (!super.isVisible()) return false;
    const container = this.containingGroup;
    let tochildren = true;
    const diagram = this.diagram;
    if (diagram !== null) tochildren = diagram.isTreePathToChildren;
    // this allows for partly or completely disconnected Links that are visible:
    const from = this.fromNode;
    if (from !== null) {
      // handle collapsed trees
      if (this.isTreeLink && tochildren && !from.isTreeExpanded) return false;
      // handle member link connecting with containingGroup
      if (from === container) return true;
      // avoid infinite loops if .fromNode is also a label node on this link
      let lab = from;
      while (lab !== null) {
        if (lab.labeledLink === this) return true;
        lab = lab.containingGroup;
      }
      // handle collapsed subgraphs
      const visfrom = from.findVisibleNode();
      if (visfrom === null) return false;
      // links inside groups should not connect with their parent group
      if (visfrom === container) return false;
    }
    const to = this.toNode;
    if (to !== null) {
      // handle collapsed trees
      if (this.isTreeLink && !tochildren && !to.isTreeExpanded) return false;
      // handle member link connecting with containingGroup
      if (to === container) return true;
      // avoid infinite loops if .toNode is also a label node on this link
      let lab = to;
      while (lab !== null) {
        if (lab.labeledLink === this) return true;
        lab = lab.containingGroup;
      }
      // handle collapsed subgraphs
      const visto = to.findVisibleNode();
      if (visto === null) return false;
      // links inside groups should not connect with their parent group
      if (visto === container) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * Called when a part's visibility changes, from .visible or other indirect sources (like layer.visible)
   * @this {Link}
   * @param {boolean} val whether the part was just made visible (true) or just made invisible (false)
   */
  public madeVisible(val: boolean): void {
    super.madeVisible(val);
    if (this._bundle !== null) this._bundle.invalidate();
    // update any label nodes
    if (this._labelNodes !== null) {
      const nodes = this._labelNodes;
      const nit = nodes.iterator;
      while (nit.next()) {
        const node = nit.value;
        node.madeVisible(val);
      }
    }
  }

  /**
   * Gets or sets how the route is computed, including whether it uses the points of its old route to determine the new route.
   * The value must be one of {@link Link.None|None}, {@link Link.End|End}, {@link Link.Scale|Scale}, or {@link Link.Stretch|Stretch}.
   *
   * The default value is {@link Link.None|None} -- the route is completely recalculated each time.
   */
  /** @export */ get adjusting(): EnumValue { return this._adjusting; }
  /** @export */ set adjusting(value: EnumValue) {
    const old = this._adjusting;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Link, Link, 'adjusting');
      this._adjusting = value;
      this.raiseChanged('adjusting', old, value);
      // no immediate effect -- just used when reshaping
    }
  }

  /**
   * Gets or sets how rounded the corners are for adjacent line segments when the {@link #curve}
   * is {@link Link.None|None}, {@link Link.JumpGap|JumpGap}, or {@link Link.JumpOver|JumpOver} and
   * the two line segments are orthogonal to each other.
   *
   * The default value is zero -- there is no curve at a corner.
   */
  /** @export */ get corner(): number { return this._corner; }
  /** @export */ set corner(value: number) {
    const old = this._corner;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'corner');
      this._corner = value;
      this.invalidateGeometry();
      this.raiseChanged('corner', old, value);
    }
  }

  /**
   * Gets or sets the way the path is generated from the route's points.
   * The value must be one of {@link Link.None|None}, {@link Link.Bezier|Bezier}, {@link Link.JumpGap|JumpGap}, or {@link Link.JumpOver|JumpOver}.
   *
   * Setting this property to {@link Link.JumpOver|JumpOver} or {@link Link.JumpGap|JumpGap} requires the Diagram to do
   * considerable computation when calculating Link routes. Consider not using
   * a Jump... value with Diagrams that contain large numbers of Links
   * if you are targeting slow devices.
   *
   * The default value is {@link Link.None|None} -- each link segment is a straight line.
   * @see {@link #computeCurve}
   */
  /** @export */ get curve(): EnumValue { return this._curve; }
  /** @export */ set curve(value: EnumValue) {
    const old = this._curve;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Link, Link, 'curve');
      this._curve = value;
      this.invalidateRoute();
      this.invalidateGeometry();
      this._invalidateOtherJumpOversMaybe(old === Link.JumpGap || old === Link.JumpOver ||
        value === Link.JumpGap || value === Link.JumpOver);
      this.raiseChanged('curve', old, value);
    }
  }

  /**
   * Gets or sets how far the control points are offset when the {@link #curve} is {@link Link.Bezier|Bezier}
   * or when there are multiple links between the same two ports.
   *
   * The default value is NaN -- the actual curviness is computed based on how many links connect the same pair of ports.
   * @see {@link #computeCurviness}
   */
  /** @export */ get curviness(): number { return this._curviness; }
  /** @export */ set curviness(value: number) {
    const old = this._curviness;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'curviness');
      this._curviness = value;
      this.invalidateRoute();
      this.invalidateGeometry();
      this.raiseChanged('curviness', old, value);
    }
  }

  /**
   * Gets or sets whether the link's path tries to avoid other nodes.
   * The value must be one of {@link Link.Normal|Normal}, {@link Link.Orthogonal|Orthogonal}, or {@link Link.AvoidsNodes|AvoidsNodes}.
   *
   * Setting this property to {@link Link.AvoidsNodes|AvoidsNodes} requires the Diagram to do
   * considerable computation when calculating Link routes. Consider not using
   * the {@link Link.AvoidsNodes|AvoidsNodes} with Diagrams that contain large numbers of Nodes and Links
   * if you are targeting slow devices.
   * When using {@link Link.AvoidsNodes|AvoidsNodes}, dragging performance can be improved by setting {@link DraggingTool#isComplexRoutingRealtime} to false.
   *
   * The default value is {@link Link.Normal|Normal} -- the route is primarily a single straight segment,
   * with possible short end segments when connecting with ports that have a "spot" value
   * as the {@link #fromSpot} or {@link GraphObject#fromSpot} or {@link #toSpot} or {@link GraphObject#toSpot}.
   */
  /** @export */ get routing(): EnumValue { return this._routing; }
  /** @export */ set routing(value: EnumValue) {
    const old = this._routing;
    if (old !== value) {
      if (Debug) EnumValue.checkEnumValue(value, Link, Link, 'routing');
      this._routing = value;
      this._effectiveBezier = null;
      this.invalidateRoute();
      this._invalidateOtherJumpOversMaybe(Link.isOrthogonalRouting(old) || Link.isOrthogonalRouting(value));
      this.raiseChanged('routing', old, value);
    }
  }

  /**
   * Gets or sets how far the control points are from the points of the route
   * when {@link #routing} is {@link Link.Orthogonal|Orthogonal} and {@link #curve} is {@link Link.Bezier|Bezier}.
   *
   * The default value is 0.5.
   * Values of this property typically fall in the range of 0.0 to 1.0.
   * A value of 0.0 indicates that the control points of the curve are at the end points,
   * which will result in straight line segments.
   * A value of 1.0 indicates that the control points are one-third of the link's
   * length away from the end point along the direction of the link,
   * and the same distance away in a perpendicular direction.
   * The distance scales linearly with the value of this property,
   * even with negative values and values greater than 1.0.
   */
  /** @export */ get smoothness(): number { return this._smoothness; }
  /** @export */ set smoothness(value: number) {
    const old = this._smoothness;
    if (old !== value) {
      if (Debug) Util.checkPrimitive(value, 'number', Link, 'smoothness');
      this._smoothness = value;
      this.invalidateGeometry();
      this.raiseChanged('smoothness', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public updateLinkBundleAdd(): void {
    // ?? some way not to participate in link bundling?
    // ?? bundling all links between nodes, not specifically between pairs of ports
    const fromnode = this._fromNode;
    if (fromnode === null) return;
    const tonode = this._toNode;
    if (tonode === null) return;

    // see if there's an existing LinkBundle between the nodes/ports
    const fromportid = this._fromPortId;
    const toportid = this._toPortId;
    // if set to a list, we'll need a LinkBundle if one doesn't already exist
    let bundledlinks: Array<Link> = null;
    let firstlink: Link = null;

    const lit = fromnode._linksConnected._arr;
    const length = lit.length;
    for (let i = 0; i < length; i++) {
      const l = lit[i];
      if (!(l._fromNode === fromnode && l._fromPortId === fromportid && l._toNode === tonode && l._toPortId === toportid) &&
        !(l._fromNode === tonode && l._fromPortId === toportid && l._toNode === fromnode && l._toPortId === fromportid)) {
        continue;
      }
      if (firstlink === null) {  // found first link
        firstlink = l;
      } else {
        if (bundledlinks === null) {  // found second link: need bundle!
          bundledlinks = [];  // inefficient
          bundledlinks.push(firstlink);
        }
        // found any link after the first one
        bundledlinks.push(l);
      }
    }

    if (bundledlinks !== null) {  // need a LinkBundle
      let bundle = fromnode.findBundle(tonode, fromportid, toportid);
      if (bundle === null) {  // make sure one exists and the Nodes know about it
        bundle = new LinkBundle(fromnode, fromportid, tonode, toportid);
        fromnode.addBundle(bundle);
        tonode.addBundle(bundle);
      }
      bundle.links = bundledlinks;  // do this before calling invalidate()
      // setup backpointers
      for (let i = 0; i < bundledlinks.length; i++) {
        const l = bundledlinks[i];
        l._bundle = bundle;
      }
      bundle.invalidate();
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   */
  public updateLinkBundleRemove(): void {
    // see if there's an existing LinkBundle for this link
    const bundle = this._bundle;
    if (bundle !== null) {
      this._bundle = null;
      const idx = bundle.links.indexOf(this);
      if (idx >= 0) {  // if this is in the bundle, remove it and revalidate later
        Util.arrayRemoveAt(bundle.links, idx);
        bundle.invalidate();
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Link}
   * @return {boolean}
   */
  public isLinkLike(): this is Link {
    return true;
  }

  /**
   * This read-only property returns the Links's Model data key if it is in a Diagram and is backed by Model data.
   * Otherwise this returns undefined.
   *
   * If the link data is in a GraphLinksModel, the {@link GraphLinksModel#linkKeyProperty} must not be an empty string.
   * @since 1.8
   * @see {@link GraphLinksModel#getKeyForLinkData}
   */
  /** @export */ get key(): Key {
    const diagram = this.diagram;
    if (diagram === null || !diagram.model.isLinkDataModel()) return undefined;
    return diagram.model.getKeyForLinkData(this.data);
  }

} // #end Link
Link['className'] = 'Link';

Link['Normal'] = Link.Normal;
Link['Orthogonal'] = Link.Orthogonal;
Link['AvoidsNodes'] = Link.AvoidsNodes;
Link['AvoidsNodesStraight'] = Link.AvoidsNodesStraight;
Link['None'] = Link.None;
Link['Bezier'] = Link.Bezier;
Link['JumpGap'] = Link.JumpGap;
Link['JumpOver'] = Link.JumpOver;
Link['End'] = Link.End;
Link['Scale'] = Link.Scale;
Link['Stretch'] = Link.Stretch;
Link['OrientAlong'] = Link.OrientAlong;
Link['OrientPlus90'] = Link.OrientPlus90;
Link['OrientMinus90'] = Link.OrientMinus90;
Link['OrientOpposite'] = Link.OrientOpposite;
Link['OrientUpright'] = Link.OrientUpright;
Link['OrientPlus90Upright'] = Link.OrientPlus90Upright;
Link['OrientMinus90Upright'] = Link.OrientMinus90Upright;
Link['OrientUpright45'] = Link.OrientUpright45;

/**
 * @hidden @internal
 */
export class LinkBundle {
  /** @hidden @internal */ public static className = 'LinkBundle';
  /** @hidden @internal */ private _invalid: boolean;
  /** @hidden @internal */ private _isUndoingRedoing: boolean;
  /** @hidden @internal */ public node1: Node;
  /** @hidden @internal */ public portId1: string | null;
  /** @hidden @internal */ public node2: Node;
  /** @hidden @internal */ public portId2: string | null;
  /** @hidden @internal */ public links: Array<Link>;

  constructor(node1: Node, port1: string | null, node2: Node, port2: string | null) {
    Set.uniqueHash(this);
    this._invalid = false;
    this._isUndoingRedoing = false;
    this.node1 = node1;
    this.portId1 = port1;
    this.node2 = node2;
    this.portId2 = port2;
    this.links = [];
  }

  /**
   * @hidden @internal
   * @this {LinkBundle}
   */
  public invalidate(): void {
    if (!this._invalid) {
      const arr = this.links;
      if (arr.length > 0) {
        const diagram = arr[0].diagram;
        if (diagram !== null) {
          diagram._invalidLinkBundles.add(this);
          this._isUndoingRedoing = diagram.undoManager.isUndoingRedoing;
        }
      }
    }
    this._invalid = true;
  }

  /**
   * @hidden @internal
   * @this {LinkBundle}
   */
  public validate(): void {
    if (this._invalid) {
      this._invalid = false;
      const arr = this.links;
      if (arr.length > 0) {
        const aLink = arr[0];  // arbitrarily choose the Link on which to call arrangeBundledLinks
        const diagram = aLink.diagram;
        // don't reroute when undoing or redoing!
        const reroute = (diagram === null || (diagram.isInitialLayoutCompleted && !this._isUndoingRedoing));
        this._isUndoingRedoing = false;
        aLink.arrangeBundledLinks(arr, reroute);  // this may sort links and invalidate their routes

        // if there's only one link left in the LinkBundle, remove it
        if (arr.length === 1) {
          aLink._bundle = null;
          arr.length = 0;
        }
      }
      // if there are no links left, remove the bundle
      if (arr.length === 0) {
        this.node1.removeBundle(this);
        this.node2.removeBundle(this);
      }
    }
  }
} // #end LinkBundle
LinkBundle['className'] = 'LinkBundle';

/**
 * @hidden @internal
 */
export class PositionArray {
  /** @hidden @internal */ public static className = 'PositionArray';
  /** @hidden @internal */ private _minX: number;
  /** @hidden @internal */ private _minY: number;
  /** @hidden @internal */ private _maxX: number;
  /** @hidden @internal */ private _maxY: number;
  /** @hidden @internal */ private _cellX: number;
  /** @hidden @internal */ private _cellY: number;
  /** @hidden @internal */ private _array: Array<Array<number>>;
  /** @hidden @internal */ private _upperBoundX: number;
  /** @hidden @internal */ private _upperBoundY: number;
  /** @hidden @internal */ public group: Group | null;
  /** @hidden @internal */ public skip: Node | null;
  /** @hidden @internal */ public invalid: boolean;
  /** @hidden @internal */ public abort: boolean;
  /** @hidden @internal */ public wholeDocument: boolean;
  /** @hidden @internal */ public smallMargin: number;
  /** @hidden @internal */ public largeMargin: number;

  constructor() {
    Set.uniqueHash(this);
    this.group = null;
    this.skip = null;
    this.invalid = true;
    this.abort = false;
    this._minX = 1;
    this._minY = 1;
    this._maxX = -1;
    this._maxY = -1;
    this._cellX = 8;
    this._cellY = 8;
    this._array = [[]];
    this._upperBoundX = 0;
    this._upperBoundY = 0;
    this.wholeDocument = false;  // whether to search the whole document for a valid route
    this.smallMargin = 22;  // how many cells to inflate around the link's nodes' bounds to search
    this.largeMargin = 111;  // how many cells to inflate around the link's nodes' bounds to search
  }

  /** @hidden @internal @const @static */
  public static OCCUPIED = 0;
  /** @hidden @internal @const @static */
  public static START = 1;
  /// ** @hidden @internal @const @static */
  // static STEP = 1;
  /** @hidden @internal @const @static */
  public static MAX = 999999;
  /** @hidden @internal @const @static */
  public static UNOCCUPIED = PositionArray.MAX + 1;

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {Rect} rect
   */
  public initialize(rect: Rect): void {
    if (rect.width <= 0 || rect.height <= 0) return;
    const minx = rect.x;
    const miny = rect.y;
    const maxx = rect.x + rect.width;
    const maxy = rect.y + rect.height;
    // add a row and a column on all sides
    this._minX = Math.floor((minx - this._cellX) / this._cellX) * this._cellX;
    this._minY = Math.floor((miny - this._cellY) / this._cellY) * this._cellY;
    this._maxX = Math.ceil((maxx + 2 * this._cellX) / this._cellX) * this._cellX;
    this._maxY = Math.ceil((maxy + 2 * this._cellY) / this._cellY) * this._cellY;
    const xrange = 1 + (Math.ceil((this._maxX - this._minX) / this._cellX) | 0);
    const yrange = 1 + (Math.ceil((this._maxY - this._minY) / this._cellY) | 0);
    if (this._array === null || this._upperBoundX < xrange - 1 || this._upperBoundY < yrange - 1) {
      const a: Array<Array<number>> = [];
      for (let x = 0; x <= xrange; x++) {  // loop is unnecessary in .NET and Java
        a[x] = [];
      }
      this._array = a;
      this._upperBoundX = xrange - 1;
      this._upperBoundY = yrange - 1;
    }
    this.setAll(PositionArray.UNOCCUPIED);
  }
  /* noexport */get bounds(): Rect { return new Rect(this._minX, this._minY, this._maxX - this._minX, this._maxY - this._minY); }
  /* noexport */get cellWidth(): number { return this._cellX; }
  set cellWidth(value: number) {
    if (value > 0 && value !== this._cellX) {
      this._cellX = value;
      this.initialize(this.bounds);
    }
  }
  /* noexport */get cellHeight(): number { return this._cellY; }
  set cellHeight(value: number) {
    if (value > 0 && value !== this._cellY) {
      this._cellY = value;
      this.initialize(this.bounds);
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x  in document coordinates.
   * @param {number} y
   * @return {boolean}
   */
  public inBounds(x: number, y: number): boolean {
    return (this._minX <= x && x <= this._maxX && this._minY <= y && y <= this._maxY);
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x  in document coordinates.
   * @param {number} y
   * @return {number} int
   */
  public getDistance(x: number, y: number): number {
    if (!this.inBounds(x, y)) return PositionArray.UNOCCUPIED;
    x -= this._minX;
    x /= this._cellX;
    y -= this._minY;
    y /= this._cellY;
    const ix = x | 0;
    const iy = y | 0;
    return this._array[ix][iy];
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x  in document coordinates.
   * @param {number} y
   */
  public setOccupied(x: number, y: number): void {
    if (!this.inBounds(x, y)) return;
    x -= this._minX;
    x /= this._cellX;
    y -= this._minY;
    y /= this._cellY;
    const ix = x | 0;
    const iy = y | 0;
    this._array[ix][iy] = PositionArray.OCCUPIED;
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} v int
   */
  private setAll(v: number): void {
    if (this._array === null) return;
    for (let ix = 0; ix <= this._upperBoundX; ix++) {
      for (let iy = 0; iy <= this._upperBoundY; iy++) {
        this._array[ix][iy] = v;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   */
  public clearAllUnoccupied(): void {
    if (this._array === null) return;
    for (let ix = 0; ix <= this._upperBoundX; ix++) {
      for (let iy = 0; iy <= this._upperBoundY; iy++) {
        if (this._array[ix][iy] >= PositionArray.START) {
          this._array[ix][iy] = PositionArray.UNOCCUPIED;
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x  in document coordinates.
   * @param {number} y
   * @return {boolean}
   */
  public isOccupied(x: number, y: number): boolean {
    return this.getDistance(x, y) === PositionArray.OCCUPIED;
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x  in document coordinates.
   * @param {number} y
   * @param {number} w
   * @param {number} h
   * @return {boolean}
   */
  public isUnoccupied(x: number, y: number, w: number, h: number): boolean {
    if (x > this._maxX) return true;
    if (x + w < this._minX) return true;
    if (y > this._maxY) return true;
    if (y + h < this._minY) return true;
    let ix = ((x - this._minX) / this._cellX) | 0;
    let iy = ((y - this._minY) / this._cellY) | 0;
    let iw = ((Math.max(0, w) / this._cellX) + 1) | 0;
    let ih = ((Math.max(0, h) / this._cellY) + 1) | 0;
    if (ix < 0) {
      iw += ix;  // decrease width
      ix = 0;
    }
    if (iy < 0) {
      ih += iy;  // decrease height
      iy = 0;
    }
    if (iw < 0) return true;
    if (ih < 0) return true;
    const mx = Math.min(ix + iw - 1, this._upperBoundX) | 0;
    const my = Math.min(iy + ih - 1, this._upperBoundY) | 0;
    for (let i = ix; i <= mx; i++) {
      for (let j = iy; j <= my; j++) {
        if (this._array[i][j] === PositionArray.OCCUPIED) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x int
   * @param {number} y int
   * @param {number} inc int
   * @param {boolean} vert
   * @param {number} lowx int
   * @param {number} hix int
   * @param {number} lowy int
   * @param {number} hiy int
   * @return {number} int
   */
  private ray(x: number, y: number, inc: number, vert: boolean, lowx: number, hix: number, lowy: number, hiy: number): number {
    x = x | 0;
    y = y | 0;
    let val = this._array[x][y];
    if (val >= PositionArray.START && val < PositionArray.MAX) {
      if (vert) {
        y += inc;
      } else {
        x += inc;
      }
      val += 1 /*PositionArray.STEP*/;
      while (lowx <= x && x <= hix && lowy <= y && y <= hiy) {
        const oldval = this._array[x][y];
        if (val >= oldval) break;
        this._array[x][y] = val;
        val += 1 /*PositionArray.STEP*/;
        if (vert) {
          y += inc;
        } else {
          x += inc;
        }
      }
    }
    if (vert) {
      return y;
    } else {
      return x;
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x
   * @param {number} y
   * @param {number} inc
   * @param {boolean} vert
   * @param {number} lowx
   * @param {number} hix
   * @param {number} lowy
   * @param {number} hiy
   */
  private spread(x: number, y: number, inc: number, vert: boolean, lowx: number, hix: number, lowy: number, hiy: number): void {
    if (x < lowx || x > hix || y < lowy || y > hiy) return;
    const end = this.ray(x, y, inc, vert, lowx, hix, lowy, hiy);
    if (vert) {
      if (inc > 0) {
        for (let yy = y + inc; yy < end; yy += inc) {
          this.spread(x, yy, 1, !vert, lowx, hix, lowy, hiy);
          this.spread(x, yy, -1, !vert, lowx, hix, lowy, hiy);
        }
      } else {
        for (let yy = y + inc; yy > end; yy += inc) {
          this.spread(x, yy, 1, !vert, lowx, hix, lowy, hiy);
          this.spread(x, yy, -1, !vert, lowx, hix, lowy, hiy);
        }
      }
    } else {
      if (inc > 0) {
        for (let xx = x + inc; xx < end; xx += inc) {
          this.spread(xx, y, 1, !vert, lowx, hix, lowy, hiy);
          this.spread(xx, y, -1, !vert, lowx, hix, lowy, hiy);
        }
      } else {
        for (let xx = x + inc; xx > end; xx += inc) {
          this.spread(xx, y, 1, !vert, lowx, hix, lowy, hiy);
          this.spread(xx, y, -1, !vert, lowx, hix, lowy, hiy);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x1
   * @param {number} y1
   * @param {number} x2
   * @param {number} y2
   * @param {number} inc
   * @param {boolean} vert
   * @param {number} lowx
   * @param {number} hix
   * @param {number} lowy
   * @param {number} hiy
   * @return {number}
   */
  private breakOut(x1: number, y1: number, x2: number, y2: number, inc: number, vert: boolean, lowx: number, hix: number, lowy: number, hiy: number): number {
    let x = x1 | 0;
    let y = y1 | 0;
    let oldval = PositionArray.OCCUPIED;
    let val = PositionArray.START;
    this._array[x][y] = val;
    while (oldval === PositionArray.OCCUPIED && x > lowx && x < hix && y > lowy && y < hiy) {
      val += 1 /*PositionArray.STEP*/;
      this._array[x][y] = val;
      if (vert) {
        y += inc;
      } else {
        x += inc;
      }
      oldval = this._array[x][y];
    }
    if (vert) {
      return y;
    } else {
      return x;
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {number} x1
   * @param {number} y1
   * @param {number} x2
   * @param {number} y2
   * @param {number} inc
   * @param {boolean} vert
   * @param {number} lowx
   * @param {number} hix
   * @param {number} lowy
   * @param {number} hiy
   */
  private breakIn(x1: number, y1: number, x2: number, y2: number, inc: number, vert: boolean, lowx: number, hix: number, lowy: number, hiy: number): void {
    let x = x2 | 0;
    let y = y2 | 0;
    let oldval = PositionArray.OCCUPIED;
    const val = PositionArray.MAX;
    this._array[x][y] = val;
    while (oldval === PositionArray.OCCUPIED && x > lowx && x < hix && y > lowy && y < hiy) {
      // val -= 1 /*PositionArray.STEP*/;
      this._array[x][y] = val;
      if (vert) {
        y += inc;
      } else {
        x += inc;
      }
      oldval = this._array[x][y];
    }
  }

  /**
   * @hidden @internal
   * @this {PositionArray}
   * @param {Point} p1
   * @param {number} fromDir
   * @param {Point} p2
   * @param {number} toDir
   * @param {Rect} bounds
   */
  public propagate(p1: Point, fromDir: number, p2: Point, toDir: number, bounds: Rect): void {
    if (this._array === null) return;
    this.abort = false;

    let x1 = p1.x;
    let y1 = p1.y;
    if (!this.inBounds(x1, y1)) return;
    x1 -= this._minX;
    x1 /= this._cellX;
    y1 -= this._minY;
    y1 /= this._cellY;

    let x2 = p2.x;
    let y2 = p2.y;
    if (!this.inBounds(x2, y2)) return;
    x2 -= this._minX;
    x2 /= this._cellX;
    y2 -= this._minY;
    y2 /= this._cellY;

    if (Math.abs(x1 - x2) <= 1 && Math.abs(y1 - y2) <= 1) {
      // if start and end are at the same cell, never mind, just use the default algorithm
      this.abort = true;
      return;
    }

    let bx1 = bounds.x;
    let by1 = bounds.y;
    let bx2 = bounds.x + bounds.width;
    let by2 = bounds.y + bounds.height;
    bx1 -= this._minX;
    bx1 /= this._cellX;
    by1 -= this._minY;
    by1 /= this._cellY;
    bx2 -= this._minX;
    bx2 /= this._cellX;
    by2 -= this._minY;
    by2 /= this._cellY;

    const lowx = Math.max(0, Math.min(this._upperBoundX, bx1 | 0));
    const hix = Math.min(this._upperBoundX, Math.max(0, bx2 | 0));
    const lowy = Math.max(0, Math.min(this._upperBoundY, by1 | 0));
    const hiy = Math.min(this._upperBoundY, Math.max(0, by2 | 0));

    const RIGHT = 0;
    const DOWN = 90;
    // var LEFT = 180; // unused
    const UP = 270;

    const ix1 = x1 | 0;
    const iy1 = y1 | 0;
    const ix2 = x2 | 0;
    const iy2 = y2 | 0;
    const inc = ((fromDir === RIGHT || fromDir === DOWN) ? 1 : -1);
    const vert = (fromDir === DOWN || fromDir === UP);
    if (this._array[ix1][iy1] === PositionArray.OCCUPIED) {
      this.breakOut(ix1, iy1, ix2, iy2, inc, vert, lowx, hix, lowy, hiy);
      this.breakOut(ix1, iy1, ix2, iy2, 1, !vert, lowx, hix, lowy, hiy);
      this.breakOut(ix1, iy1, ix2, iy2, -1, !vert, lowx, hix, lowy, hiy);
    } else {
      this.breakOut(ix1, iy1, ix2, iy2, inc, vert, ix1, iy1, ix1, iy1);
    }

    if (this._array[ix2][iy2] === PositionArray.OCCUPIED) {
      this.breakIn(ix1, iy1, ix2, iy2, ((toDir === RIGHT || toDir === DOWN) ? 1 : -1), (toDir === DOWN || toDir === UP), lowx, hix, lowy, hiy);
      this.breakIn(ix1, iy1, ix2, iy2, 1, !(toDir === DOWN || toDir === UP), lowx, hix, lowy, hiy);
      this.breakIn(ix1, iy1, ix2, iy2, -1, !(toDir === DOWN || toDir === UP), lowx, hix, lowy, hiy);
    } else {
      this.breakIn(ix1, iy1, ix2, iy2, inc, vert, ix2, iy2, ix2, iy2);
    }

    if (this.abort) return;

    this.spread(ix1, iy1, 1, false, lowx, hix, lowy, hiy);
    this.spread(ix1, iy1, -1, false, lowx, hix, lowy, hiy);
    this.spread(ix1, iy1, 1, true, lowx, hix, lowy, hiy);
    this.spread(ix1, iy1, -1, true, lowx, hix, lowy, hiy);
  }

} // #end PositionArray
PositionArray['className'] = 'PositionArray';

/**
 * @hidden @internal
 */
export class Knot {
  /** @hidden @internal */ public static className = 'Knot';
  /** @hidden @internal */ public node: Node;
  /** @hidden @internal */ public port: GraphObject;
  /** @hidden @internal */ public mySortedLinks: Array<LinkInfo>;
  /** @hidden @internal */ public myRespreading: boolean;

  constructor() {  // also used by Parts
    Set.uniqueHash(this);
    this.node = null;
    this.port = null;

    this.mySortedLinks = [];
    this.myRespreading = false;
  }

  /**
   * @hidden @internal
   * @this {Knot}
   * @return {string}
   */
  public toString(): string {
    const arr = this.mySortedLinks;
    let msg = this.node.toString() + ' ' + arr.length.toString() + ':';
    for (let i = 0; i < arr.length; i++) {
      const info = arr[i];
      if (info === null) continue;
      msg += '\n  ' + info.toString();
    }
    return msg;
  }

  /**
   * @hidden @internal
   * @this {Knot}
   * @param {Rect} rect
   * @param {Spot} spot
   * @param {number} generalangle
   * @param {boolean} ortho
   * @return {number}
   */
  public getSideAngle(rect: Rect, spot: Spot, generalangle: number, ortho: boolean): number {
    const sides = spot.offsetY;
    switch (sides) {
      case Util.MBOTTOM: return 90;
      case Util.MLEFT: return 180;
      case Util.MTOP: return 270;
      case Util.MRIGHT: return 0;
    }
    let A = generalangle;
    switch (sides) {
      case Util.MBOTTOM | Util.MTOP: if (A > 180) return 270; else return 90;
      case Util.MLEFT | Util.MRIGHT: if (A > 90 && A <= 270) return 180; else return 0;
    }
    const R = Math.atan2(rect.height, rect.width) * 180 / Math.PI;
    switch (sides) {
      case Util.MLEFT | Util.MTOP: if (A > R && A <= 180 + R) return 180; else return 270;
      case Util.MTOP | Util.MRIGHT: if (A > 180 - R && A <= 360 - R) return 270; else return 0;
      case Util.MRIGHT | Util.MBOTTOM: if (A > R && A <= 180 + R) return 90; else return 0;
      case Util.MBOTTOM | Util.MLEFT: if (A > 180 - R && A <= 360 - R) return 180; else return 90;
      case Util.MLEFT | Util.MTOP | Util.MRIGHT: if (A > 90 && A <= 180 + R) return 180; else if (A > 180 + R && A <= 360 - R) return 270; else return 0;
      case Util.MTOP | Util.MRIGHT | Util.MBOTTOM: if (A > 180 && A <= 360 - R) return 270; else if (A > R && A <= 180) return 90; else return 0;
      case Util.MRIGHT | Util.MBOTTOM | Util.MLEFT: if (A > R && A <= 180 - R) return 90; else if (A > 180 - R && A <= 270) return 180; else return 0;
      case Util.MBOTTOM | Util.MLEFT | Util.MTOP: if (A > 180 - R && A <= 180 + R) return 180; else if (A > 180 + R) return 270; else return 90;
    }
    if (ortho && sides !== (Util.MLEFT | Util.MTOP | Util.MRIGHT | Util.MBOTTOM)) {
      A -= 15;
      if (A < 0) A += 360;
    }
    if (A > R && A < 180 - R) {
      return 90;
    } else if (A >= 180 - R && A <= 180 + R) {
      return 180;
    } else if (A > 180 + R && A < 360 - R) {
      return 270;
    } else {
      return 0;
    }
  }

  /**
   * @hidden @internal
   * @this {Knot}
   */
  public invalidate(): void {
    this.mySortedLinks.length = 0;
  }

  /**
   * @hidden @internal
   * First updates any needed data structures...
   * @this {Knot}
   * @param {Link} link
   * @return {LinkInfo}
   */
  public findLinkInfo(link: Link): LinkInfo | null {
    let arr = this.mySortedLinks;
    if (arr.length === 0) {
      this.getLinkInfos();
      arr = this.mySortedLinks;
    }
    for (let i = 0; i < arr.length; i++) {
      const info = arr[i];
      if (info !== null && info.link === link) return info;
    }
    return null;
  }

  // ??? what about self-loops
  /**
   * @hidden @internal
   * First updates any needed data structures...
   * @this {Knot}
   * @return {Array.<LinkInfo>}
   */
  public getLinkInfos(): Array<LinkInfo> {
    if (!this.myRespreading) {
      const oldRespreading = this.myRespreading;
      this.myRespreading = true;

      let connectedlinks: Iterator<Link>;
      let collgroup: Group = null;
      const node = this.node;
      const group: Group = node instanceof Group ? node : null;
      if (group !== null && !group.isSubGraphExpanded) {
        if (!group.actualBounds.isReal()) {
          this.myRespreading = oldRespreading;
          return this.mySortedLinks;
        }
        collgroup = group;
        connectedlinks = collgroup.findExternalLinksConnected();
      } else {
        connectedlinks = this.node.findLinksConnected(this.port.portId);
      }
      this.mySortedLinks.length = 0;
      let i = 0;

      const tlpt = this.port.getDocumentPoint(Spot.TopLeft, Point.alloc());
      const brpt = this.port.getDocumentPoint(Spot.BottomRight, Point.alloc());
      const thisrect = Rect.allocAt(tlpt.x, tlpt.y, 0, 0);
      thisrect.unionPoint(brpt);
      Point.free(tlpt);
      Point.free(brpt);

      const thisrectcenter = Point.allocAt(thisrect.x + thisrect.width / 2, thisrect.y + thisrect.height / 2);
      const portangle = this.port.getDocumentAngle();
      const lit = connectedlinks.iterator;
      while (lit.next()) {
        const l = lit.value;
        if (!l.isVisible()) continue;
        if (l.fromPort === l.toPort) continue;
        const from = (l.fromPort === this.port || (l.fromNode !== null && l.fromNode.isMemberOf(collgroup)));
        const spot = l.computeSpot(from, this.port);  // link's Spot takes precedence, if defined
        if (!spot.isSide()) continue;
        let otherport: GraphObject;
        if (from) {
          otherport = l.toPort;
        } else {
          otherport = l.fromPort;
        }
        if (otherport === null) continue;
        let othernode = otherport.part as Node;
        if (othernode === null) continue;
        const visnode = othernode.findVisibleNode();
        if (visnode !== null && visnode !== othernode) {
          othernode = visnode;
          otherport = othernode.port;
        }
        const otherpoint = l.computeOtherPoint(othernode, otherport);
        let angle = thisrectcenter.directionPoint(otherpoint);
        angle -= portangle;
        if (angle < 0) angle += 360;
        const dir = this.getSideAngle(thisrect, spot, angle, l.isOrthogonal);
        let side = 0;
        if (dir === 0) {
          side = Util.MRIGHT;
          if (angle > 180) angle -= 360;
        } else if (dir === 90) {
          side = Util.MBOTTOM;
        } else if (dir === 180) {
          side = Util.MLEFT;
        } else {
          side = Util.MTOP;
        }
        let info = this.mySortedLinks[i];
        if (info === undefined) {
          info = new LinkInfo(l, angle, side);
          this.mySortedLinks[i] = info;
        } else {
          info.link = l;
          info.angle = angle;
          info.side = side;
        }
        info.otherPoint.set(otherpoint);
        i++;
      }
      Point.free(thisrectcenter);

      this.sortLinkInfos();
      // now assign IndexOnSide for each set of LinkInfos with the same Side
      const numlinks = this.mySortedLinks.length;
      let currside = -1;
      let numonside = 0;
      for (i = 0; i < numlinks; i++) {
        const info = this.mySortedLinks[i];
        if (info === undefined) continue;
        if (info.side !== currside) {
          currside = info.side;
          numonside = 0;
        }
        info.indexOnSide = numonside;
        numonside++;
      }
      // the highest IndexOnSide now also indicates how many there are
      // with the same Side--assign CountOnSide correspondingly;
      // also compute the LinkPoint
      currside = -1;
      numonside = 0;
      for (i = numlinks - 1; i >= 0; i--) {  // backwards!
        const info = this.mySortedLinks[i];
        if (info === undefined) continue;
        if (info.side !== currside) {
          currside = info.side;
          numonside = info.indexOnSide + 1;
        }
        info.countOnSide = numonside;
      }
      this.assignLinkPoints(this.mySortedLinks);
      this.assignEndSegmentLengths(this.mySortedLinks);
      this.myRespreading = oldRespreading;
      Rect.free(thisrect);
    }
    return this.mySortedLinks;
  }

  /**
   * @hidden @internal
   * Sorting all of the link's other ports in circular order.
   * @this {Knot}
   * @param {LinkInfo} a
   * @param {LinkInfo} b
   * @return {number} int
   */
  public myComparer(a: LinkInfo, b: LinkInfo): number {
    if (a === b) return 0;
    if (a === null) return -1;
    if (b === null) return 1;
    if (a.side < b.side) {
      return -1;
    } else if (a.side > b.side) {
      return 1;
    } else {
      if (a.angle < b.angle) {
        return -1;
      } else if (a.angle > b.angle) {
        return 1;
      } else {
        return 0;
      }
    }
  }

  /**
   * @hidden @internal
   * Sort an array of angle and side information about the links connected to this port.
   * By default this just sorts by {@link LinkInfo#side} group, and by {@link LinkInfo#angle} for each side.
   * @this {Knot}
   */
  public sortLinkInfos(): void {
    this.mySortedLinks.sort(Knot.prototype.myComparer);
  }

  /**
   * @hidden @internal
   * Given a sorted array of angle and side information about the links connected to this port,
   * assign the actual {@link LinkInfo#linkPoint}.
   * By default this just spreads the link points evenly along each side.
   * @this {Knot}
   * @param {Array.<LinkInfo>} linkinfos an array of {@link LinkInfo} that are modified.
   */
  public assignLinkPoints(linkinfos: Array<LinkInfo>): void {
    const port = this.port;
    const spread = this.node.portSpreading;
    const tl = Point.alloc();
    const tr = Point.alloc();
    const br = Point.alloc();
    const bl = Point.alloc();
    port.getDocumentPoint(Spot.TopLeft, tl);
    port.getDocumentPoint(Spot.TopRight, tr);
    port.getDocumentPoint(Spot.BottomRight, br);
    port.getDocumentPoint(Spot.BottomLeft, bl);
    let thickt = 0.0;
    let thickr = 0.0;
    let thickb = 0.0;
    let thickl = 0.0;
    if (spread === Node.SpreadingPacked) {
      for (let i = 0; i < linkinfos.length; i++) {
        const info = linkinfos[i];
        if (info === null) continue;
        const thick = info.link.computeThickness();
        switch (info.side) {
          case Util.MBOTTOM: thickb += thick; break;
          case Util.MLEFT: thickl += thick; break;
          case Util.MTOP: thickt += thick; break;
          default:
          case Util.MRIGHT: thickr += thick; break;
        }
      }
    }
    let currside = 0;
    let currthick = 0.0;
    let thickfrac = 1.0;
    let a: Point;
    let b: Point;
    let dx = 0;
    let dy = 0;
    for (let i = 0; i < linkinfos.length; i++) {
      const info = linkinfos[i];
      if (info === null) continue;
      if (currside !== info.side) {
        currside = info.side;
        switch (currside) {
          case Util.MBOTTOM: a = br; b = bl; break;
          case Util.MLEFT: a = bl; b = tl; break;
          case Util.MTOP: a = tl; b = tr; break;
          default:
          case Util.MRIGHT: a = tr; b = br; break;
        }
        dx = b.x - a.x;
        dy = b.y - a.y;
        // if not enough space on the side to hold the thicknesses of all of the links, figure out how much to scale down
        switch (currside) {
          case Util.MBOTTOM: if (thickb > Math.abs(dx)) { thickfrac = Math.abs(dx) / thickb; thickb = Math.abs(dx); } else { thickfrac = 1; } break;
          case Util.MLEFT: if (thickl > Math.abs(dy)) { thickfrac = Math.abs(dy) / thickl; thickl = Math.abs(dy); } else { thickfrac = 1; } break;
          case Util.MTOP: if (thickt > Math.abs(dx)) { thickfrac = Math.abs(dx) / thickt; thickt = Math.abs(dx); } else { thickfrac = 1; } break;
          default:
          case Util.MRIGHT: if (thickr > Math.abs(dy)) { thickfrac = Math.abs(dy) / thickr; thickr = Math.abs(dy); } else { thickfrac = 1; } break;
        }
        currthick = 0;
      }
      const p = info.linkPoint;  // modify this result Point
      if (spread === Node.SpreadingPacked) {
        let thick = info.link.computeThickness();
        thick *= thickfrac;  // scale down if not enough room
        p.set(a);
        switch (currside) {
          case Util.MBOTTOM: p.x = a.x + dx / 2 + thickb / 2 - currthick - thick / 2; break;
          case Util.MLEFT: p.y = a.y + dy / 2 + thickl / 2 - currthick - thick / 2; break;
          case Util.MTOP: p.x = a.x + dx / 2 - thickt / 2 + currthick + thick / 2; break;
          default:
          case Util.MRIGHT: p.y = a.y + dy / 2 - thickr / 2 + currthick + thick / 2; break;
        }
        currthick += thick;
      } else {
        let frac = 0.5;
        if (spread === Node.SpreadingEvenly) frac = (info.indexOnSide + 1) / (info.countOnSide + 1);
        p.x = a.x + dx * frac;
        p.y = a.y + dy * frac;
      }
    }
    Point.free(tl);
    Point.free(tr);
    Point.free(br);
    Point.free(bl);
  }

  /**
   * @hidden @internal
   * Given a sorted array of angle, side, and link-point information about the links
   * connected to this port, assign the actual {@link LinkInfo#endSegmentLength}.
   * By default this just specifies shorter values for links at the ends of each side,
   * and longer values for links in the middle of each side.
   * @this {Knot}
   * @param {Array.<LinkInfo>} linkinfos an array of {@link LinkInfo} that are modified.
   */
  public assignEndSegmentLengths(linkinfos: Array<LinkInfo>): void {
    for (let i = 0; i < linkinfos.length; i++) {
      const info = linkinfos[i];
      if (info === null) continue;
      info.endSegmentLength = this.computeEndSegmentLength(info);
    }
  }

  /**
   * @hidden @internal
   * @this {Knot}
   * @param {LinkInfo} info
   * @return {number}
   */
  public computeEndSegmentLength(info: LinkInfo): number {
    const link = info.link;
    const from = link.fromPort === this.port;
    // get default EndSegmentLength
    const esl = link.computeEndSegmentLength(this.node, this.port, Spot.None, from);

    let idx = info.indexOnSide;
    if (idx < 0) return esl;  // if unused, don't extend length
    const count = info.countOnSide;
    if (count <= 1) return esl;  // if there's just one, don't bother

    const ortho = link.isOrthogonal;
    if (!ortho) return esl;

    const pc = info.otherPoint;
    const thisc = info.linkPoint;
    // indexed clockwise, not by absolute positions
    if (info.side === Util.MLEFT || info.side === Util.MBOTTOM) idx = count - 1 - idx;
    const step = 8;
    const horiz = (info.side === Util.MLEFT || info.side === Util.MRIGHT);
    if (horiz ? pc.y < thisc.y : pc.x < thisc.x) {
      return esl + idx * step;
    } else if (horiz ? pc.y === thisc.y : pc.x === thisc.x) {
      return esl;
    } else {
      return esl + (count - 1 - idx) * step;
    }
  }

} // #end Knot
Knot['className'] = 'Knot';

/**
 * @hidden @internal
 * This is used by Knot.SortLinkInfos, Knot.AssignLinkPoints, and Knot.AssignEndSegmentLengths
 * only when Knot.LinkPointsSpread is true.
 * This information is transient, just used when calculating link points.
 */
export class LinkInfo {
  /** @hidden @internal */ public static className = 'LinkInfo';
  /** @hidden @internal */ public link: Link;
  /** @hidden @internal */ public angle: number;
  /** @hidden @internal */ public side: number;
  /** @hidden @internal */ public otherPoint: Point;
  /** @hidden @internal */ public indexOnSide: number;
  /** @hidden @internal */ public countOnSide: number;
  /** @hidden @internal */ public linkPoint: Point;
  /** @hidden @internal */ public endSegmentLength: number;

  /**
   * @param {Link} l
   * @param {number} a
   * @param {number} s int
   */
  constructor(l: Link, a: number, s: number) {
    // The {@link Link} whose link point is being sorted around the port.
    this.link = l;

    // The effective angle at which the link connects with another node;
    // this value corresponds to the result of calling GetAngle.
    // The angle in degrees from this port to the other port.
    this.angle = a;

    // The side at which the link connects;
    // this value corresponds to the result of calling GetDirection.
    // One of Util.MiddleRight, Util.MiddleBottom, Util.MiddleLeft, Util.MiddleTop.
    this.side = s;

    this.otherPoint = new Point();

    // The index of this link on this side;
    // computed after calling SortLinkInfos and before calling AssignLinkPoints.
    this.indexOnSide = 0;

    // How many links are connected on this side;
    // computed after calling SortLinkInfos and before calling AssignLinkPoints.
    this.countOnSide = 0;

    // The document point at which the link should terminate;
    // should be set in AssignLinkPoints.
    this.linkPoint = new Point();

    // The value of length of the last segment for this link at this port;
    // should be set in AssignEndSegmentLengths.
    this.endSegmentLength = 0;
  }

  /**
   * @hidden @internal
   * @this {LinkInfo}
   * @return {string}
   */
  public toString(): string {
    return this.link.toString() + ' ' + this.angle.toString() + ' ' +
      this.side.toString() + ':' + this.indexOnSide.toString() + '/' + this.countOnSide.toString() + ' ' +
      this.linkPoint.toString() + ' ' + this.endSegmentLength.toString() + ' ' + this.otherPoint.toString();
  }

} // #end LinkInfo
LinkInfo['className'] = 'LinkInfo';

/**
 * @hidden @internal
 */
export class LinkSettings {
  /** @hidden @internal */ public static className = 'LinkSettings';
  /** @hidden @internal */ public _fromSpot: Spot;
  /** @hidden @internal */ public _toSpot: Spot;
  /** @hidden @internal */ public _fromEndSegmentLength: number;
  /** @hidden @internal */ public _toEndSegmentLength: number;
  /** @hidden @internal */ public _fromShortLength: number;
  /** @hidden @internal */ public _toShortLength: number;
  /** @hidden @internal */ public _fromLinkable: boolean | null;
  /** @hidden @internal */ public _toLinkable: boolean | null;
  /** @hidden @internal */ public _fromMaxLinks: number;
  /** @hidden @internal */ public _toMaxLinks: number;

  constructor() {
    this._fromSpot = Spot.Default;
    this._toSpot = Spot.Default;
    this._fromEndSegmentLength = NaN;
    this._toEndSegmentLength = NaN;
    this._fromShortLength = NaN;
    this._toShortLength = NaN;

    // Only in GraphObject, not in Link:
    this._fromLinkable = null;
    this._toLinkable = null;
    this._fromMaxLinks = Infinity;
    this._toMaxLinks = Infinity;
  }

  /**
   * @hidden @internal
   * @this {LinkSettings}
   * @return {LinkSettings}
   */
  public copy(): LinkSettings {
    const copy = new LinkSettings();
    copy._fromSpot = this._fromSpot.copyFrozen();
    copy._toSpot = this._toSpot.copyFrozen();
    copy._fromEndSegmentLength = this._fromEndSegmentLength;
    copy._toEndSegmentLength = this._toEndSegmentLength;
    copy._fromShortLength = this._fromShortLength;
    copy._toShortLength = this._toShortLength;

    copy._fromLinkable = this._fromLinkable;
    copy._toLinkable = this._toLinkable;
    copy._fromMaxLinks = this._fromMaxLinks;
    copy._toMaxLinks = this._toMaxLinks;
    return copy;
  }

} // #end LinkBundle
LinkSettings['className'] = 'LinkSettings';

// import { ChangedEvent } from './changedEvent';
// import { EmptyIterator, List, ListIterator, Map, Set, SingletonIterator } from './collections';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { StreamGeometryContext } from './figures';
// import { Geo } from './geo';
// import { Geometry } from './geometry';
// import { Layer } from './layer';
// import { Layout } from './layout'; // only needed because of Group
// import { Margin, MarginLike } from './margin';
// import { Key } from './model';
// import { PanelLayout } from './panelLayout';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Shape } from './shape';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Transform } from './transform';
// import { Util } from './util';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Layout
LayoutNetwork
LayoutVertex
LayoutEdge
MultiLayout
*/

// import { List, Map, Set } from './collections';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Margin } from './margin';
// import { Group, Node, Part } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData } from './root';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * This is the base class for all of the predefined diagram layout implementations.
 * They only arrange {@link Part}s (primarily {@link Node}s and {@link Link}s) in a {@link Diagram},
 * not to {@link GraphObject}s in {@link Panel}s (i.e. panel layout).
 *
 * The layout classes include {@link TreeLayout}, {@link ForceDirectedLayout},
 * {@link LayeredDigraphLayout}, {@link CircularLayout}, and {@link GridLayout}.
 * This base class is not abstract -- in fact an instance of this base class is the default
 * value for {@link Diagram#layout} and for {@link Group#layout}.
 *
 * An instance of a Layout class will be the value of {@link Diagram#layout}.
 * That layout positions the graph of top-level nodes and links.
 * Nodes and links that belong to a {@link Group} are laid out by that group's {@link Group#layout}.
 * The Diagram will automatically perform all nested group layouts before laying out the whole diagram.
 *
 * If you have position information for all of the nodes when you load a model,
 * you will typically have data bound the {@link Part#location} to some property on your node data.
 * In order to avoid an initial layout causing those saved node positions to be discarded,
 * you can either not set the {@link Diagram#layout} to a predefined layout or you can
 * set {@link #isInitial} to false.
 *
 * Because performing layouts can be expensive in space and time, automatic layouts
 * are performed only on "invalid" layouts, and only well after a layout has been invalidated.
 * This state is held by the {@link #isValidLayout} property.
 * Many standard operations, such as adding or removing nodes or links, will cause
 * the layout that is responsible for positioning those nodes or routing those links
 * to be invalidated.
 * Such invalidation is performed by calling {@link #invalidateLayout},
 * which not only clears the {@link #isValidLayout} state but also requests that the diagram
 * do an automatic layout soon.
 * You can avoid such invalidations by setting {@link #isOngoing} to false.
 *
 * Layouts will ignore parts that have {@link Part#isLayoutPositioned} set to false
 * or parts that are not {@link GraphObject#visible}.
 * Layouts will also ignore parts that are in layers that are {@link Layer#isTemporary}.
 *
 * Various operations on {@link Part}s will cause the responsible Layout to be invalidated.
 * This includes adding or removing parts, changing their visibility, and changing their size.
 * You can disable such automatic layout invalidations by setting {@link Part#layoutConditions}
 * to the combination of Part flags named "Layout..." that you want.
 *
 * But operations on parts are not the only way in which layouts become invalidated.
 * Setting most properties on the layouts, thereby changing their behavior, will invalidate that layout.
 * Replacing the {@link Diagram#layout} or {@link Group#layout} will automatically invalidate the new layout.
 * If {@link #isViewportSized} is true, when a diagram's {@link Diagram#viewportBounds} changes size,
 * the {@link Diagram#layout} is invalidated.
 * (This is normally only true for {@link GridLayout}s when its {@link GridLayout#wrappingWidth} is `NaN`.
 * Most layouts do not care about the size of the viewport.)
 *
 * You can also explicitly call {@link Diagram#layoutDiagram}, which can invalidate all layouts and
 * then perform them all.
 * But we recommend that you avoid doing so, to allow the normal updating process perform layouts as needed.
 *
 * If an automatic layout is the first time that a layout has been performed for the model,
 * the diagram first raises the {@link DiagramEvent} named "InitialLayoutCompleted".
 * Whenever a Diagram finishes an automatic layout, it raises the DiagramEvent named "LayoutCompleted".
 *
 * It is also possible to call {@link #doLayout} explicitly, but this is uncommon and only used with
 * instances of Layout that are not the {@link Diagram#layout} or {@link Group#layout}.
 * It should only be needed when you want to layout a collection of nodes and links that is not
 * the normal graph of top-level parts of a Diagram or a subgraph of a Group.
 *
 * More complicated layouts make use of a separate {@link LayoutNetwork}, consisting of {@link LayoutVertex}es and {@link LayoutEdge}s,
 * that normally holds a graph that is isomorphic to the graph consisting of {@link Node}s and {@link Link}s in the {@link Diagram} or {@link Group}.
 * The implementation of {@link #doLayout} will call {@link #makeNetwork} and remember the result as the {@link #network}.
 * {@link #makeNetwork} will call {@link #createNetwork} and initialize it by adding new instances of LayoutVertexes and LayoutEdges
 * corresponding to the given collection of Nodes and Links.
 *
 * When {@link #doLayout} is finished with its work it will call {@link #updateParts}, which will call {@link #commitLayout}
 * to set new node locations and route links. It then normally discards the {@link #network}.
 *
 * The LayoutVertex and LayoutEdge instances allow the layout to work with more information about each Node and Link without
 * actually modifying those Nodes and Links until {@link #commitLayout} is called to actually set the Node locations and route the Links.
 * The use of a LayoutNetwork also allows the Layout to work with a graph that is not isomorphic to the given collection of Nodes and Links.
 * This is useful when needing to use dummy vertexes and/or edges to achieve certain layout behaviors,
 * or when one wants to ignore certain vertexes or edges, without actually modifying or adding or removing the diagram's nodes or links.
 *
 * An instance of this base class provides a rudimentary default layout that will position
 * all of the parts that have no position (i.e. the {@link Part#location} is (NaN,NaN).
 * Parts that already have a position are ignored.
 * This primitive layout class does not make use of a {@link LayoutNetwork} because
 * it ignores all links.
 *
 * To implement your own custom layouts, you can inherit from either this class or
 * from one of the other predefined layout classes.
 * If you inherit from this base class, you will want to override the {@link #doLayout} method.
 * You can call the {@link Part#move} method to re-position a part, including whole groups.
 * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call a base method.
 * @category Layout
 */
export class Layout {
  /** @hidden @internal */ public static className = 'Layout';
  /** @hidden @internal */ private _diagram: Diagram | null;
  /** @hidden @internal */ private _group: Group | null;
  /** @hidden @internal */ private _isOngoing: boolean;
  /** @hidden @internal */ private _isInitial: boolean;
  /** @hidden @internal */ private _isViewportSized: boolean;
  /** @hidden @internal */ private _arrangementOrigin: Point;
  /** @hidden @internal */ private _isRouting: boolean;
  /** @hidden @internal */ private _isRealtime: boolean | null;
  /** @hidden @internal */ private _division: string;
  /** @hidden @internal */ private _boundsComputation: ((part: Part, lay: Layout, rect: Rect) => Rect) | null;
  /** @hidden @internal */ private _isValidLayout: boolean;
  /** @hidden @internal */ private _network: LayoutNetwork | null;

  /**
   * Create a minimal layout that only positions {@link Node}s that do not have a location.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(Layout);
    Set.uniqueHash(this);

    this._diagram = null;
    this._group = null;
    this._isOngoing = true;
    this._isInitial = true;
    this._isViewportSized = false;
    this._arrangementOrigin = new Point(0, 0).freeze();
    this._isRouting = true;
    this._isRealtime = null;
    this._division = '';
    this._boundsComputation = null;

    this._isValidLayout = false;
    this._network = null;
    // ??  this._multilayout = null;
  }

  /**
   * Copies properties from this object to the given object, which is of the same class.
   * This is called by {@link #copy} and should be overridden for each class that adds properties.
   * There are examples of such overrides in the samples.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   * @param {?} copy
   * @since 1.6
   */
  protected cloneProtected(copy: this): void {
    // new owner needs to set _diagram and _group
    copy._isOngoing = this._isOngoing;
    copy._isInitial = this._isInitial;
    copy._isViewportSized = this._isViewportSized;
    copy._arrangementOrigin.assign(this._arrangementOrigin);
    copy._isRouting = this._isRouting;
    copy._isRealtime = this._isRealtime;
    copy._division = this._division;

    copy._isValidLayout = true;  // let normal invalidations happen upon adding members to Group or node size changes
    // _network = null
  }

  /**
   * Creates a copy of this Layout and returns it.
   * When a {@link Group} is copied that has a {@link Group#layout}, the {@link Layout} must also be copied.
   * This calls {@link #cloneProtected} on a newly constructed Layout.
   * @expose
   * @this {Layout}
   * @return {Layout}
   */
  public copy(): this {
    const copy = new (this.constructor as any)();
    this.cloneProtected(copy);
    return copy;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Layout}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    Util.throwEnumPropertyValueError(this, arg);
  }

  /**
   * @hidden @internal
   * @this {Layout}
   * @return {string}
   */
  public toString(): string {
    let str = Util.getClassName(this.constructor);
    str += '('; // ?? + this.division;
    if (this.group !== null) str += ' in ' + this.group;
    if (this.diagram !== null) str += ' for ' + this.diagram;
    str += ')';
    return str;
  }

  /**
   * Gets the {@link Diagram} that owns this layout, if it is the value of {@link Diagram#layout}.
   *
   * If this property and {@link #group} are non-null, the Group should be in this Diagram.
   * @see {@link #group}
   */
  /** @export */ get diagram(): Diagram | null { return this._diagram; }
  /** @export */ set diagram(value: Diagram | null) {
    if (value !== null) Util.checkClass(value, Diagram, Layout, 'diagram');
    this._diagram = value;
  }

  /**
   * Gets the {@link Group} that uses this layout, if it is the value of a group's {@link Group#layout}.
   *
   * If this property is set to a Group, the {@link #diagram} is automatically set to be the Group's Diagram.
   * @see {@link #diagram}
   */
  /** @export */ get group(): Group | null { return this._group; }
  /** @export */ set group(value: Group | null) {
    if (this._group !== value) {
      if (value !== null) Util.checkClass(value, Group, Layout, 'group');
      this._group = value;
      if (value !== null) this._diagram = value.diagram;
    }
  }

  /**
   * Gets or sets whether this layout can be invalidated by {@link #invalidateLayout}.
   * Set this to false to prevent actions such as adding or removing Parts from invalidating this layout.
   * The default value is true.
   * Setting this property does not invalidate this layout.
   *
   * If you set both {@link #isInitial} and {@link #isOngoing} to false,
   * there will be no automatic layout invalidation, because {@link #invalidateLayout}
   * will not set {@link #isValidLayout} to false.
   * To get your nodes to appear, you will need to explicitly set or data-bind their {@link Part#location}
   * or {@link GraphObject#position} to real Point values, because automatic layout will not assign any positions.
   *
   * Another way of controlling when layouts are invalidated is by setting
   * {@link Part#isLayoutPositioned} or {@link Part#layoutConditions}.
   */
  /** @export */ get isOngoing(): boolean { return this._isOngoing; }
  /** @export */ set isOngoing(value: boolean) {
    if (this._isOngoing !== value) {
      Util.checkPrimitive(value, 'boolean', Layout, 'isOngoing');
      this._isOngoing = value;
    }
  }

  /**
   * Gets or sets whether this layout is performed on an initial layout.
   * The default value is true.
   * Setting this property to false causes {@link #isValidLayout} to be set to true
   * so that the diagram does not perform this layout.
   *
   * If you set both {@link #isInitial} and {@link #isOngoing} to false,
   * there will be no automatic layout invalidation, because {@link #invalidateLayout}
   * will not set {@link #isValidLayout} to false.
   * To get your nodes to appear, you will need to explicitly set or data-bind their {@link Part#location}
   * or {@link GraphObject#position} to real Point values, because automatic layout will not assign any positions.
   *
   * Another way of controlling when layouts are invalidated is by setting
   * {@link Part#isLayoutPositioned} or {@link Part#layoutConditions}.
   */
  /** @export */ get isInitial(): boolean { return this._isInitial; }
  /** @export */ set isInitial(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Layout, 'isInitial');
    this._isInitial = value;
    if (!value) {  // do this side-effect unconditionally
      this._isValidLayout = true;
    }
  }

  /**
   * Gets or sets whether this layout depends on the {@link Diagram#viewportBounds}'s size.
   * If set to true, the layout will invalidate when the Diagram's viewport changes size.
   * This only applies to diagram layouts, not to group layouts,
   * and only when {@link Diagram#autoScale} is set to {@link Diagram.None}.
   * The default value is false.
   * Setting this property to true will invalidate this layout.
   */
  /** @export */ get isViewportSized(): boolean { return this._isViewportSized; }
  /** @export */ set isViewportSized(value: boolean) {
    if (this._isViewportSized !== value) {
      Util.checkPrimitive(value, 'boolean', Layout, 'isViewportSized');
      this._isViewportSized = value;
      if (value) this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether this layout routes {@link Link}s.
   * The default value is true.
   * When false, this layout will not explicitly set the {@link Link#points},
   * and the default routing of each individual Link will take place after the Nodes are moved by {@link #commitLayout}.
   * Setting this property does not invalidate this layout.
   *
   * Some layouts ignore links, in which case this property is ignored.
   * @since 1.1
   */
  /** @export */ get isRouting(): boolean { return this._isRouting; }
  /** @export */ set isRouting(value: boolean) {
    if (this._isRouting !== value) {
      Util.checkPrimitive(value, 'boolean', Layout, 'isRouting');
      this._isRouting = value;
    }
  }

  /**
   * Gets or sets whether this layout be performed in real-time, before the end of a transaction.
   * All layouts that are invalidated will be performed at the end of a transaction.
   * The default value is null.
   * A null value is treated as true for a {@link Diagram#layout} but false for a {@link Group#layout}.
   * Setting this property does not invalidate this layout.
   * @since 1.2
   */
  /** @export */ get isRealtime(): boolean | null { return this._isRealtime; }
  /** @export */ set isRealtime(value: boolean | null) {
    if (this._isRealtime !== value) {
      if (value !== null) Util.checkPrimitive(value, 'boolean', Layout, 'isRealtime');
      this._isRealtime = value;
    }
  }

  /**
   * Gets or sets whether this layout needs to be performed again (if false).
   * Instead of setting this property directly, it is normal to set it to false by calling {@link #invalidateLayout},
   * since that also requests performing a layout in the near future.
   */
  /** @export */ get isValidLayout(): boolean { return this._isValidLayout; }
  /** @export */ set isValidLayout(value: boolean) {
    if (this._isValidLayout !== value) {
      Util.checkPrimitive(value, 'boolean', Layout, 'isValidLayout');
      this._isValidLayout = value;
      if (!value) {
        const diagram = this.diagram;
        if (diagram !== null) {
          diagram._anyLayoutInvalid = true;  // optimization for detecting whether any layout is invalid
        }
      }
    }
  }

  /**
   * If {@link #isOngoing} is true and if an initial layout has not yet been performed,
   * set the {@link #isValidLayout} property to false, and ask to perform another layout in the near future.
   * If {@link #isInitial} is true, this layout is invalidated only when the {@link Diagram#model} is replaced,
   * not under the normal circumstances such as when parts are added or removed or
   * due to other calls to {@link Layout#invalidateLayout}.
   *
   * If you set both {@link #isInitial} and {@link #isOngoing} to false,
   * there will be no automatic layout invalidation, because this method
   * will not set {@link #isValidLayout} to false.
   * However you can still set {@link #isValidLayout} explicitly.
   *
   * This is typically called when a layout property value has changed,
   * or when a Part is added or removed or changes visibility, if {@link Part#layoutConditions} includes the pertinent flags.
   * @export
   * @this {Layout}
   */
  public invalidateLayout(): void {
    if (this.isValidLayout) {
      const diagram = this.diagram;
      if (diagram === null) return;
      if (diagram.undoManager.isUndoingRedoing) return;

      // Layouts directly caused by animations must be ignored during animation
      // Layouts done during animation must stop the current animation
      const AM = diagram.animationManager;
      if (AM.isTicking) return;
      if (AM.defaultAnimation.isAnimating) AM.stopAnimation();

      if ((this.isOngoing && diagram.isInitialLayoutCompleted) ||
        (this.isInitial && !diagram.isInitialLayoutCompleted)) {
        this.isValidLayout = false;
        //      // invalidate container MultiLayouts
        //      var mlay = this._multilayout;
        //      while (mlay !== null) {
        //        mlay.isValidLayout = false;
        //        mlay = mlay._multilayout;
        //      }
        diagram.requestUpdate();
      }
    }
  }

  /**
   * Gets or sets the {@link LayoutNetwork} used by this Layout, if any.
   * The default value is null.
   * Setting this property does not invalidate this layout.
   * Not all kinds of layout make use of a {@link LayoutNetwork}.
   * Call {@link #createNetwork} or {@link #makeNetwork} to create a network.
   */
  /** @export */ get network(): LayoutNetwork | null { return this._network; }
  /** @export */ set network(value: LayoutNetwork | null) {
    const old = this._network;
    if (old !== value) {
      if (value !== null) Util.checkClass(value, LayoutNetwork, Layout, 'network');
      // if (old !== null) old.layout = null;
      this._network = value;
      if (value !== null) value.layout = this;
    }
  }

  /**
   * Create a new {@link LayoutNetwork} of {@link LayoutVertex}es and {@link LayoutEdge}s.
   * This may be overridden in Layout subclasses to create instances of subclasses of
   * {@link LayoutNetwork}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   * @return {LayoutNetwork} a new {@link LayoutNetwork}.
   */
  public createNetwork(): LayoutNetwork {
    return new LayoutNetwork(this);
  }

  /**
   * Create and initialize a {@link LayoutNetwork} with the given nodes and links.
   * This should be called by {@link #doLayout} when this layout uses a {@link #network}.
   * This method calls {@link #createNetwork} to allocate the network.
   * This may be overridden in Layout subclasses to customize the initialization.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   * @return {LayoutNetwork} normally the value of a call to {@link #createNetwork} initialized by vertexes and edges corresponding to the *coll* argument.
   */
  public makeNetwork(coll: Diagram | Group | Iterable<Part>): LayoutNetwork {
    const net = this.createNetwork();
    if (coll instanceof Diagram) {
      net.addParts(coll.nodes, true);
      net.addParts(coll.links, true);
    } else if (coll instanceof Group) {
      net.addParts(coll.memberParts);
    } else {
      net.addParts(coll.iterator);
    }
    return net;
  }

  /**
   * When using a {@link LayoutNetwork}, update the "physical" node positionings and link routings.
   * This should be called by {@link #doLayout} when this layout uses a {@link #network}.
   * This calls {@link #commitLayout} to actually set {@link Node} positions and route {@link Link}s.
   * This performs the changes within a transaction.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   */
  public updateParts(): void {
    let diagram = this.diagram;
    if (diagram === null && this.network !== null) {
      // look for a Node.diagram
      const it = this.network.vertexes.iterator;
      while (it.next()) {
        const v = it.value;
        const n = v.node;
        if (n !== null) {
          diagram = n.diagram;
          if (diagram !== null) break;
        }
      }
    }
    this.isValidLayout = true;  // prevent unintentional recursive layout
    try {
      if (diagram !== null) diagram.startTransaction('Layout');
      this.commitLayout();
    } finally {
      if (diagram !== null) diagram.commitTransaction('Layout');
    }
  }

  /**
   * When using a {@link LayoutNetwork}, commit changes to the diagram
   * by setting {@link Node} positions and by routing the {@link Link}s.
   * This is called by {@link #updateParts} within a transaction.
   *
   * You should not call this method -- it is a "protected virtual" method.
   * This may be overridden by subclasses of Layout.
   * By default this method is implemented as follows:
   * ```js
   * protected commitLayout() {
   *   if (this.network === null) return;
   *   var vit = this.network.vertexes.iterator;
   *   while (vit.next()) {
   *     var vert = vit.value;
   *     vert.commit();
   *   }
   *   if (this.isRouting) {
   *     var eit = this.network.edges.iterator;
   *     while (eit.next()) {
   *       var edge = eit.value;
   *       edge.commit();
   *     }
   *   }
   * }
   * ```
   *
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   */
  protected commitLayout(): void {
    if (this.network === null) return;
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const vert = vit.value;
      vert.commit();
    }
    if (this.isRouting) {
      const eit = this.network.edges.iterator;
      while (eit.next()) {
        const edge = eit.value;
        edge.commit();
      }
    }
  }

  /**
   * Position all of the nodes that do not have an assigned {@link Part#location}
   * in the manner of a simple rectangular array.
   * The default implementation ignores all {@link Group}s and {@link Link}s;
   * many subclasses of {@link Layout} ignore all instances of {@link Part}s
   * that are not {@link Node}s or {@link Link}s.
   *
   * You can override this method to do whatever node positioning and link routing
   * that you wish.
   *
   * When the layout makes use of a {@link LayoutNetwork},
   * this method should call {@link #makeNetwork} and {@link #updateParts}.
   * The {@link #updateParts} method will call {@link #commitLayout} within a transaction.
   * The outline of such an override should be like:
   * ```js
   * public doLayout(coll) {
   *   if (this.network === null) this.network = this.makeNetwork(coll);
   *   // assign LayoutVertex.bounds to all vertexes in the network:
   *   var vit = this.network.vertexes.iterator;
   *   while (vit.next()) {
   *     var v = vit.value;
   *     v.centerX = ...
   *     v.centerY = ...
   *   }
   *   this.updateParts();
   *   this.network = null;
   * }
   * ```
   * Override {@link #commitLayout} to actually position nodes and route links.
   *
   * When the layout does <em>not</em> make use of a {@link LayoutNetwork},
   * this method should make direct changes to {@link Node}s and {@link Link}s
   * within a transaction, and it should call {@link #collectParts} on the argument *coll* in order
   * to get the proper set of Nodes and Links to work on.
   * Examples are provided in the Extensions directory.
   *
   * If this Layout belongs to a Diagram, the argument must be either the same Diagram or must denote Parts that belong to the Diagram.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Layout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');
    const allparts = new Set<Part>();
    if (coll instanceof Diagram) {
      // include top-level nodes/groups/parts, but not links
      this.collectAllParts(allparts, coll.nodes, true, this.needsLocation, true, false, true);
      // skip Diagram.links
      this.collectAllParts(allparts, coll.parts, true, this.needsLocation, true, false, true);
    } else if (coll instanceof Group) {
      // include nodes/groups/parts, but not links
      this.collectAllParts(allparts, coll.memberParts, false, this.needsLocation, true, false, true);
    } else {
      // include all given parts
      allparts.addAll(coll.iterator);
    }
    const num = allparts.count;  // all non-links
    if (num > 0) {
      const diagram = this.diagram;
      if (diagram !== null) diagram.startTransaction('Layout');
      const sqrt = Math.ceil(Math.sqrt(num));
      this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);
      const originx = this.arrangementOrigin.x;
      const originy = this.arrangementOrigin.y;
      let x = originx;
      let y = originy;
      let i = 0;
      let maxh = 0.0;
      const idx = allparts.iterator;
      while (idx.next()) {
        const part = idx.value;
        this.maybeEnsureBounds(part);
        const b = part.measuredBounds;
        const w = b.width;
        const h = b.height;
        part.moveTo(x, y);
        if (part instanceof Group) part.needsLocation = false;
        x += Math.max(w, 50) + 20;
        maxh = Math.max(maxh, Math.max(h, 50));
        if (i >= sqrt - 1) {
          i = 0;
          x = originx;
          y += maxh + 20;
          maxh = 0;
        } else {
          i++;
        }
      }
      if (diagram !== null) diagram.commitTransaction('Layout');
    }
    this.isValidLayout = true;
  }

  /**
   * @hidden @internal
   * @this {Layout}
   * @param {Part} part
   * @return {boolean}
   */
  private needsLocation(part: Part): boolean {
    const loc = part.location;
    if (!loc.isReal()) return true;
    // hack: when a Group didn't have a real Location before it got laid out by Diagram.updateGroupLayouts,
    // it might get a Location anyway, but that would cause the above test to think that the Group doesn't need to be moved.
    if (part instanceof Group) {
      if (part.needsLocation) return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Layout}
   * @param {Set.<Part>} allparts
   * @param {Iterable.<Part>} parts
   * @param {boolean} toplevelonly
   * @param {function(Part):boolean | null} pred
   * @param {boolean} includenodes
   * @param {boolean} includelinks
   * @param {boolean} includeparts
   */
  private collectAllParts(allparts: Set<Part>, parts: Iterable<Part>, toplevelonly: boolean, pred: ((a: Part) => boolean) | null,
                          includenodes: boolean, includelinks: boolean, includeparts: boolean): void {
    const it = parts.iterator;
    while (it.next()) {
      const part = it.value;
      if (toplevelonly && !part.isTopLevel) continue;
      if (pred !== null && !pred(part)) continue;
      if (!part.canLayout()) continue;
      if (includenodes && part instanceof Node) {
        if (part.isLinkLabel) continue;
        // don't include a Group with no Group.layout, but do include all of its memberParts
        if (part instanceof Group) {
          if (part.layout === null) {
            this.collectAllParts(allparts, part.memberParts, false, pred, includenodes, includelinks, includeparts);
          } else { // a Group with a Group.layout is treated as an atomic node
            this.maybeEnsureBounds(part);
            allparts.add(part);
          }
        } else {  // plain Node
          this.maybeEnsureBounds(part);
          allparts.add(part);
        }
      } else if (includelinks && part instanceof Link) {
        allparts.add(part);
      } else if (includeparts && part.isNodeLike() && !(part instanceof Node)) {
        this.maybeEnsureBounds(part);
        allparts.add(part);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Layout}
   * @param {Part} part
   */
  private maybeEnsureBounds(part: Part): void {
    const b = part.actualBounds;
    if (b.width === 0 || b.height === 0 || isNaN(b.width) || isNaN(b.height)) {
      part.ensureBounds();
    }
  }

  /**
   * This method is called by layouts to determine the size and initial position of the nodes that it is laying out.
   * Normally this just returns the part's {@link GraphObject#actualBounds}.
   * However, if {@link #boundsComputation} has been set to a function, that function will be called
   * in order to return the bounds of the given Part in document coordinates that the layout should pretend it has.
   * @export
   * @this {Layout}
   * @param {Part} part  the Part being laid out
   * @param {Rect=} rect  an optional Rect that will be modified and returned
   * @return {Rect} a Rect in document coordinates
   * @since 2.0
   */
  public getLayoutBounds(part: Part, rect?: Rect): Rect {
    const func = this.boundsComputation;
    if (func !== null) {
      if (!rect) rect = new Rect();
      return func(part, this, rect);
    }
    if (!rect) return part.actualBounds;
    rect.set(part.actualBounds);
    return rect;
  }

  /**
   * Gets or sets a function that determines the initial size and position in document coordinates of a LayoutVertex corresponding to a Node.
   * This function is called by {@link #getLayoutBounds}.
   * The default value for this property is null, in which case the {@link GraphObject#actualBounds} of the Node is used.
   * Setting this property to a new value invalidates this layout.
   *
   * The non-null value must be a function that takes 3 arguments.
   * The first argument will be the Part whose bounds the Layout should use.
   * The second argument will be this Layout.
   * The third argument will be a Rect that must be modified and returned
   * The return value must be in document coordinates.  You may find it convenient to call {@link GraphObject#getDocumentBounds}
   * to get the bounds in document coordinates of an object within the node.
   * @since 2.0
   */
  /** @export */ get boundsComputation(): ((part: Part, lay: Layout, rect: Rect) => Rect) | null { return this._boundsComputation; }
  /** @export */ set boundsComputation(value: ((part: Part, lay: Layout, rect: Rect) => Rect) | null) {
    if (this._boundsComputation !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Layout, 'boundsComputation');
      this._boundsComputation = value;
      this.invalidateLayout();
    }
  }

  /**
   * A convenient way of converting the Diagram|Group|Iterable argument to doLayout to an actual collection of eligible Parts.
   * The resulting Set will not include any Nodes or Links for which {@link Part#canLayout} is false.
   * If the argument includes a Group for which {@link Group#layout} is null, the resulting Set
   * will include the member parts of that group rather than that group itself.
   * You will not need to call collectParts if you call {@link #makeNetwork},
   * because that method does effectively the same thing when building the {@link LayoutNetwork}.
   *
   * Typical usage:
   * ```js
   * public doLayout(coll) {
   *   // COLL might be a Diagram or a Group or some Iterable<Part>
   *   var it = this.collectParts(coll).iterator;
   *   while (it.next()) {
   *     var node = it.value;
   *     if (node instanceof go.Node) {
   *       . . . position the node . . .
   *     }
   *   }
   * }
   * ```
   * @export
   * @this {Layout}
   * @param {Diagram|Group|Iterable.<Part>} coll
   * @return {Set.<Part>}
   * @since 1.7
   */
  protected collectParts(coll: Diagram | Group | Iterable<Part>): Set<Part> {
    const allparts = new Set<Part>();
    if (coll instanceof Diagram) {
      // include top-level nodes/groups/parts
      this.collectAllParts(allparts, coll.nodes, true, null, true, true, true);
      this.collectAllParts(allparts, coll.links, true, null, true, true, true);
      this.collectAllParts(allparts, coll.parts, true, null, true, true, true);
    } else if (coll instanceof Group) {
      this.collectAllParts(allparts, coll.memberParts, false, null, true, true, true);
    } else {
      this.collectAllParts(allparts, coll.iterator, false, null, true, true, true);
    }
    return allparts;
  }

  /**
   * Gets or sets the top-left point for where the graph should be positioned when laid out.
   * The default value for this property is the Point(0, 0).
   * Setting this property to a new value invalidates this layout.
   * This property is likely to be set by many Layouts that belong to a Group when the layout is performed.
   */
  /** @export */ get arrangementOrigin(): Point { return this._arrangementOrigin; }
  /** @export */ set arrangementOrigin(value: Point) {
    Util.checkClass(value, Point, Layout, 'arrangementOrigin');
    if (!this._arrangementOrigin.equals(value)) {
      this._arrangementOrigin.assign(value);
      this.invalidateLayout();
    }
  }

  /**
   * Compute the desired value of {@link #arrangementOrigin} if this Layout is being performed for a Group.
   * This is typically called near the beginning of the implementation of {@link #doLayout}:
   * `this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);`
   * if the layout wants to respect the pre-layout location of the Group when deciding where to position its member nodes.
   * @expose
   * @this {Layout}
   * @param {Point} origin
   * @return {Point}
   */
  protected initialOrigin(origin: Point): Point {
    const group = this.group;
    if (group !== null) {
      let pos = group.position.copy();
      if (isNaN(pos.x) || isNaN(pos.y)) pos.set(origin);
      const gp = group.placeholder;
      if (gp !== null) {
        pos = gp.getDocumentPoint(Spot.TopLeft);
        if (isNaN(pos.x) || isNaN(pos.y)) pos.set(origin);
        const padding = gp.padding as Margin;
        pos.x += padding.left;
        pos.y += padding.top;
      }
      return pos;
    }
    return origin;
  }

} // #end Layout
Layout['className'] = 'Layout';

/**
 * This provides an abstract view of a diagram as a
 * network (graph) of vertexes and directed edges.
 * The network contains vertexes and edges corresponding to {@link Node}s and {@link Link}s.
 *
 * This class provides a framework for manipulating the
 * state of nodes and links without modifying the structure of the diagram.
 * Having a separate representation also permits adding or removing vertexes or edges
 * from the network so that the graph that is laid out is not isomorphic to the diagram's graph.
 *
 * For each kind of layout that uses a LayoutNetwork there is a subclass of {@link LayoutVertex} and a subclass of {@link LayoutEdge}:
 *   - {@link CircularLayout}: {@link CircularVertex} and {@link CircularEdge}
 *   - {@link ForceDirectedLayout}: {@link ForceDirectedVertex} and {@link ForceDirectedEdge}
 *   - {@link LayeredDigraphLayout}: {@link LayeredDigraphVertex} and {@link LayeredDigraphEdge}
 *   - {@link TreeLayout}: {@link TreeVertex} and {@link TreeEdge}
 *
 * Modifying a {@link LayoutNetwork} or a {@link LayoutVertex} or a {@link LayoutEdge} does not
 * invalidate the {@link Layout} or raise any changed events.
 * @unrestricted
 * @category Layout
 */
export class LayoutNetwork {
  /** @hidden @internal */ public static className = 'LayoutNetwork';
  /** @hidden @internal */ private _layout: Layout;
  /** @hidden @internal */ private _vertexes: Set<LayoutVertex>;
  /** @hidden @internal */ private _edges: Set<LayoutEdge>;
  /** @hidden @internal */ public nodeToVertexMap: Map<Node, LayoutVertex>;
  /** @hidden @internal */ public linkToEdgeMap: Map<Link, LayoutEdge>;

  /**
   * This constructs an empty network.
   * It is commonplace to call {@link #addParts} to build up the network
   * to have the same structure as a {@link Diagram}'s or {@link Group}'s nodes and links.
   */
  constructor(layout: Layout) {
    Set.uniqueHash(this);

    if (Debug && !layout) Util.throwError('LayoutNetwork constructor requires non-null Layout argument');

    this._layout = layout;

    this._vertexes = new Set<LayoutVertex>();

    this._edges = new Set<LayoutEdge>();

    // internal state
    this.nodeToVertexMap = new Map<Node, LayoutVertex>();
    this.linkToEdgeMap = new Map<Link, LayoutEdge>();
  }

  /**
   * @hidden @internal
   * Removes all vertexes and edges from the network, resulting in an empty network.
   * @this {LayoutNetwork}
   */
  public clear(): void {
    if (this._vertexes) {
      const vit = this._vertexes.iterator;
      while (vit.next()) {
        const v = vit.value;
        v.clear();
        // v.network = null;
      }
    }
    if (this._edges) {
      const eit = this._edges.iterator;
      while (eit.next()) {
        const e = eit.value;
        e.clear();
        // e.network = null;
      }
    }

    this._vertexes = new Set<LayoutVertex>();
    this._edges = new Set<LayoutEdge>();

    // internal state
    this.nodeToVertexMap = new Map<Node, LayoutVertex>();
    this.linkToEdgeMap = new Map<Link, LayoutEdge>();
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {number=} details int Optional: If greater than zero, this includes descriptions of each vertex and edge in the network.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    let str = 'LayoutNetwork' + (this.layout !== null ? '(' + this.layout.toString() + ')' : '');
    if (details <= 0) return str;
    str += ' vertexes: ' + this._vertexes.count + ' edges: ' + this._edges.count;
    if (details > 1) {
      const vit = this._vertexes.iterator;
      while (vit.next()) {
        const v = vit.value;
        str += '\n    ' + v.toString(details - 1);
      }
      const eit = this._edges.iterator;
      while (eit.next()) {
        const e = eit.value;
        str += '\n    ' + e.toString(details - 1);
      }
    }
    return str;
  }

  /**
   * Gets the {@link Layout} that uses this network
   * of {@link LayoutVertex}es and {@link LayoutEdge}s.
   * The default value is null.
   * It is set automatically for you when this setting {@link Layout#network}.
   */
  /** @export */ get layout(): Layout { return this._layout; }
  /** @export */ set layout(value: Layout) { if (value !== null) this._layout = value; }

  /**
   * Gets a collection of all of the {@link LayoutVertex}es in this network.
   * Do not modify this collection directly.
   * Instead, call {@link #addVertex}, {@link #deleteVertex}, {@link #addNode}, or {@link #deleteNode}.
   */
  /** @export */ get vertexes(): Set<LayoutVertex> { return this._vertexes; }

  /**
   * Gets a collection of all of the {@link LayoutEdge}s in this network.
   * Do not modify this collection directly.
   * Instead, call {@link #addEdge}, {@link #deleteEdge}, {@link #addLink}, or {@link #deleteLink}.
   */
  /** @export */ get edges(): Set<LayoutEdge> { return this._edges; }

  /**
   * Allocate a new instance of {@link LayoutVertex}.
   * This is overridden in LayoutNetwork subclasses to create instances of subclasses of {@link LayoutVertex}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayoutNetwork}
   * @return {LayoutVertex}
   */
  public createVertex(): LayoutVertex {
    return new LayoutVertex(this);
  }

  /**
   * Allocate a new instance of {@link LayoutEdge}.
   * This is overridden in LayoutNetwork subclasses to create instances of subclasses of {@link LayoutEdge}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayoutNetwork}
   * @return {LayoutEdge}
   */
  public createEdge(): LayoutEdge {
    return new LayoutEdge(this);
  }

  /**
   * Creates a network of {@link LayoutVertex}es and {@link LayoutEdge}s
   * corresponding to the given {@link Node}s and {@link Link}s.
   * This is usually more convenient than repeatedly calling {@link #addNode} and {@link #addLink}.
   * This method ignores all instances of {@link Part}s that are not Nodes or Links.
   * This will recurse through {@link Group}s that have no {@link Group#layout}.
   * @export
   * @this {LayoutNetwork}
   * @param {Iterable.<Part>} parts A collection of {@link Node}s or {@link Link}s.
   * @param {boolean=} toplevelonly whether to skip Parts in the given collection that are contained by Groups; default is false
   * @param {function(Part):boolean|null=} pred optional predicate to apply to each Part --
   *        if it returns false do not include Vertex or Edge in the network;
   *        default ignores link label nodes or links connecting with them
   */
  public addParts(parts: Iterable<Part>, toplevelonly?: boolean, pred?: ((a: Part) => boolean) | null): void {
    if (parts === null) return;
    if (toplevelonly === undefined) toplevelonly = false;
    Util.checkPrimitive(toplevelonly, 'boolean', LayoutNetwork, 'addParts:toplevelonly');
    if (pred === undefined) pred = null;

    // provide hook for easier filtering of Nodes or Links
    if (pred === null) {
      pred =
        /**
         * @hidden @internal
         * @param {Part} part
         */
        function (part: Part) {
          if (part instanceof Node) {
            return !part.isLinkLabel;
          } else if (part instanceof Link) {  // no links connecting with link label nodes
            const from = part.fromNode;
            if (from === null || from.isLinkLabel) return false;
            const to = part.toNode;
            if (to === null || to.isLinkLabel) return false;
            return true;
          }
          return false;  // no simple Parts
        };

    }

    // First, add all Nodes in the collection, that are not links,
    //  as vertexes to the network.
    const it = parts.iterator;
    while (it.next()) {
      const part = it.value;
      if (!(part instanceof Node)) continue;
      if (toplevelonly && !part.isTopLevel) continue;
      if (!part.canLayout()) continue;
      if (!pred(part)) continue;
      if (part instanceof Group && part.layout === null) {
        this.addParts(part.memberParts, false);
      } else {
        if (this.findVertex(part) !== null) continue;
        const pNetworkNode = this.createVertex();
        pNetworkNode.node = part;
        this.addVertex(pNetworkNode);
      }
    }

    // Second, add all node-level Links as links to the network.
    // Note: a link cannot be added to the network unless both of
    //  its to and from nodes are in the network.  Hence, it is possible
    //  to select a Link without selecting the Nodes at its
    //  endpoints, but the Link will not be added to the network.
    it.reset();
    while (it.next()) {
      const part = it.value;
      if (!(part instanceof Link)) continue;
      if (toplevelonly && !part.isTopLevel) continue;
      if (!part.canLayout()) continue;
      if (!pred(part)) continue;
      if (this.findEdge(part) !== null) continue;

      const fromNode = part.fromNode;
      const toNode = part.toNode;
      if (fromNode === null || toNode === null || fromNode === toNode) continue;

      // Verify that the top-level Vertexes for the from and to nodes
      //  are in the network and add a link between them.
      const fromVertex = this.findGroupVertex(fromNode);
      const toVertex = this.findGroupVertex(toNode);
      if (fromVertex !== null && toVertex !== null) {
        this.linkVertexes(fromVertex, toVertex, part);
      }
    }
  }

  /**
   * @hidden @internal
   * @expose
   * @this {LayoutNetwork}
   * @param {Node} n
   * @return {LayoutVertex}
   */
  public findGroupVertex(n: Node): LayoutVertex | null {
    if (n === null) return null;
    const visn = n.findVisibleNode();
    if (visn === null) return null;
    let v = this.findVertex(visn);
    if (v !== null) return v;
    let g = visn.containingGroup;
    while (g !== null) {
      v = this.findVertex(g);
      if (v !== null) return v;
      g = g.containingGroup;
    }
    return null;
  }

  /**
   * Adds a {@link LayoutVertex} to the network.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutVertex} vertex
   */
  public addVertex(vertex: LayoutVertex): void {
    if (vertex === null) return;
    if (Debug) Util.checkClass(vertex, LayoutVertex, LayoutNetwork, 'addVertex:vertex');
    this._vertexes.add(vertex);
    const node = vertex.node;
    if (node !== null) this.nodeToVertexMap.add(node, vertex);
    vertex.network = this;
  }

  /**
   * This convenience method makes sure there is a {@link LayoutVertex}
   * in this network corresponding to a {@link Node}.
   * @export
   * @this {LayoutNetwork}
   * @param {Node} node
   * @return {LayoutVertex} a vertex in this network.
   */
  public addNode(node: Node): LayoutVertex {
    if (node === null) return null;
    if (Debug) Util.checkClass(node, Node, LayoutNetwork, 'addNode:node');
    let lnode = this.findVertex(node);
    if (lnode === null) {
      lnode = this.createVertex();
      lnode.node = node;
      this.addVertex(lnode);
    }
    return lnode;
  }

  /**
   * Removes a {@link LayoutVertex} from the network.
   * This function also deletes all edges to or from the vertex.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutVertex} vertex
   */
  public deleteVertex(vertex: LayoutVertex): void {
    if (vertex === null) return;
    if (Debug) Util.checkClass(vertex, LayoutVertex, LayoutNetwork, 'deleteVertex:vertex');
    // Find the node in the network and remove.
    if (this._removeVertex(vertex)) {
      // Delete all links from predecessors of the node.
      let a = vertex.sourceEdgesList;
      for (let i = a.count - 1; i >= 0; i--) {
        const link = a.elt(i);
        this.deleteEdge(link);
      }

      // Delete all links to successors of the node.
      a = vertex.destinationEdgesList;
      for (let i = a.count - 1; i >= 0; i--) {
        const link = a.elt(i);
        this.deleteEdge(link);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {LayoutVertex} vertex
   * @return {boolean}
   */
  public _removeVertex(vertex: LayoutVertex): boolean {
    if (vertex === null) return false;
    const result = this._vertexes.remove(vertex);
    if (result) {
      const node = vertex.node;
      if (node !== null) this.nodeToVertexMap.remove(node);
      // vertex.network = null;
    }
    return result;
  }

  /**
   * This convenience method deletes any {@link LayoutVertex}
   * corresponding to a {@link Node}.
   * @export
   * @this {LayoutNetwork}
   * @param {Node} node
   */
  public deleteNode(node: Node): void {
    if (node === null) return;
    if (Debug) Util.checkClass(node, Node, LayoutNetwork, 'deleteNode:node');
    const lnode = this.findVertex(node);
    if (lnode !== null) {
      this.deleteVertex(lnode);
    }
  }

  /**
   * Returns the {@link LayoutVertex} that was constructed for the {@link Node}.
   * @export
   * @this {LayoutNetwork}
   * @param {Node} node
   * @return {LayoutVertex} a vertex in this network, or null if no such vertex exists.
   */
  public findVertex(node: Node): LayoutVertex | null {
    if (node === null) return null;
    if (Debug) Util.checkClass(node, Node, LayoutNetwork, 'findVertex:node');
    const vertex = this.nodeToVertexMap.getValue(node);
    return vertex;
  }

  /**
   * Adds a {@link LayoutEdge} to the network.
   * Although this method is provided for completeness,
   * {@link #linkVertexes} provides a more efficient manner of linking
   * vertexes already in the network.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutEdge} edge
   */
  public addEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutNetwork, 'addEdge:edge');
    this._addEdge(edge);

    const toNode = edge.toVertex;
    if (toNode !== null) {
      toNode.addSourceEdge(edge);
    }

    const fromNode = edge.fromVertex;
    if (fromNode !== null) {
      fromNode.addDestinationEdge(edge);
    }
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {LayoutEdge} edge
   */
  public _addEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    this._edges.add(edge);
    const link = edge.link;
    if (link !== null && this.findEdge(link) === null) {
      this.linkToEdgeMap.add(link, edge);
    }
    edge.network = this;
  }

  /**
   * This convenience method makes sure there is a {@link LayoutEdge}
   * in this network corresponding to a {@link Link}.
   * If {@link #findEdge} returns null, this method creates a new {@link LayoutEdge},
   * makes sure the {@link Link#fromNode} and {@link Link#toNode} have
   * corresponding {@link LayoutVertex}es in the network, and adds the edge itself
   * to the network.
   * @export
   * @this {LayoutNetwork}
   * @param {Link} link
   * @return {LayoutEdge} an edge in this network.
   */
  public addLink(link: Link): LayoutEdge {
    if (link === null) return null;
    if (Debug) Util.checkClass(link, Link, LayoutNetwork, 'addLink:link');
    const fromNode = link.fromNode;
    const toNode = link.toNode;
    let edge = this.findEdge(link);
    if (edge === null) {
      edge = this.createEdge();
      edge.link = link;

      if (fromNode !== null) {
        edge.fromVertex = this.addNode(fromNode);
      }

      if (toNode !== null) {
        edge.toVertex = this.addNode(toNode);
      }

      this.addEdge(edge);
    } else {
      if (fromNode !== null) {
        edge.fromVertex = this.addNode(fromNode);
      } else {
        edge.fromVertex = null;
      }

      if (toNode !== null) {
        edge.toVertex = this.addNode(toNode);
      } else {
        edge.toVertex = null;
      }
    }
    return edge;
  }

  /**
   * Removes a {@link LayoutEdge} from the network.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutEdge} edge
   */
  public deleteEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutNetwork, 'deleteEdge:edge');

    // Remove the link from its to node's predecessor list.
    const toNode = edge.toVertex;
    if (toNode !== null) {
      toNode.deleteSourceEdge(edge);
    }

    // Remove the link from its from node's successor list.
    const fromNode = edge.fromVertex;
    if (fromNode !== null) {
      fromNode.deleteDestinationEdge(edge);
    }

    this._removeEdge(edge);
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {LayoutEdge} edge
   * @return {boolean}
   */
  public _removeEdge(edge: LayoutEdge): boolean {
    if (edge === null) return false;
    // Find the link in the network and remove.
    const result = this._edges.remove(edge);
    if (result) {
      const link = edge.link;
      if (link !== null) this.linkToEdgeMap.remove(link);
      // edge.network = null;
    }
    return result;
  }

  /**
   * This convenience method deletes from this network any {@link LayoutEdge}
   * corresponding to a {@link Link}.
   * @export
   * @this {LayoutNetwork}
   * @param {Link} link
   */
  public deleteLink(link: Link): void {
    if (link === null) return;
    if (Debug) Util.checkClass(link, Link, LayoutNetwork, 'deleteLink:link');
    const llink = this.findEdge(link);
    if (llink !== null) {
      this.deleteEdge(llink);
    }
  }

  /**
   * Returns the {@link LayoutEdge} that was constructed for the {@link Link}.
   * @export
   * @this {LayoutNetwork}
   * @param {Link} link
   * @return {LayoutEdge} an edge in this network, or null if no such edge exists.
   */
  public findEdge(link: Link): LayoutEdge | null {
    if (link === null) return null;
    if (Debug) Util.checkClass(link, Link, LayoutNetwork, 'findEdge:link');
    const edge = this.linkToEdgeMap.getValue(link);
    return edge;
  }

  /**
   * Links two vertexes already in the network and returns the created {@link LayoutEdge}.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutVertex} fromVertex
   * @param {LayoutVertex} toVertex
   * @param {Link} link
   * @return {LayoutEdge} an edge in this network, or null if both vertexes don't exist.
   */
  public linkVertexes(fromVertex: LayoutVertex, toVertex: LayoutVertex, link: Link | null): LayoutEdge {
    if (fromVertex === null || toVertex === null) return null;
    if (Debug) {
      Util.checkClass(fromVertex, LayoutVertex, LayoutNetwork, 'linkVertexes:fromVertex');
      Util.checkClass(toVertex, LayoutVertex, LayoutNetwork, 'linkVertexes:toVertex');
      if (link !== null) Util.checkClass(link, Link, LayoutNetwork, 'linkVertexes:link');
    }
    // Verify that the from and to nodes are in the network.
    if (fromVertex.network === this && toVertex.network === this) {
      const pNetworkLink = this.createEdge();
      pNetworkLink.link = link;
      pNetworkLink.fromVertex = fromVertex;
      pNetworkLink.toVertex = toVertex;
      this.addEdge(pNetworkLink);
      return pNetworkLink;
    }
    return null;
  }

  /**
   * Reverses the direction of a {@link LayoutEdge} in the network.
   * @export
   * @this {LayoutNetwork}
   * @param {LayoutEdge} edge
   */
  public reverseEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutNetwork, 'reverseEdge:edge');
    const fromNode = edge.fromVertex;
    const toNode = edge.toVertex;

    if (fromNode === null || toNode === null) return;

    fromNode.deleteDestinationEdge(edge);
    toNode.deleteSourceEdge(edge);

    edge.reverseEdge();

    fromNode.addSourceEdge(edge);
    toNode.addDestinationEdge(edge);
  }

  /**
   * Deletes all {@link LayoutEdge}s whose "to vertex" and "from vertex" are the same vertex.
   * @export
   * @this {LayoutNetwork}
   */
  public deleteSelfEdges(): void {
    const deleteList = Util.tempArray<LayoutEdge>();
    const it = this._edges.iterator;
    while (it.next()) {
      const e = it.value;
      if (e.fromVertex === e.toVertex) {
        deleteList.push(e);
      }
    }
    const len = deleteList.length;
    for (let i = 0; i < len; i++) {
      this.deleteEdge(deleteList[i]);
    }
    Util.freeArray(deleteList);
  }

  /**
   * Deletes all vertexes and edges that have no {@link Part} associated with them.
   * @expose
   * @this {LayoutNetwork}
   */
  public deleteArtificialVertexes(): void {
    const vdummies = Util.tempArray<LayoutVertex>();
    const vit = this._vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      if (v.node === null) {
        vdummies.push(v);
      }
    }
    let len = vdummies.length;
    for (let i = 0; i < len; i++) {
      this.deleteVertex(vdummies[i]);
    }
    const edummies = Util.tempArray<LayoutEdge>();
    const eit = this._edges.iterator;
    while (eit.next()) {
      const e = eit.value;
      if (e.link === null) {
        edummies.push(e);
      }
    }
    len = edummies.length;
    for (let i = 0; i < len; i++) {
      this.deleteEdge(edummies[i]);
    }
    Util.freeArray(vdummies);
    Util.freeArray(edummies);
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   */
  public deleteUselessEdges(): void {
    const deleteList = Util.tempArray<LayoutEdge>();
    const it = this._edges.iterator;
    while (it.next()) {
      const e = it.value;
      if (e.fromVertex === null || e.toVertex === null) {
        deleteList.push(e);
      }
    }
    const len = deleteList.length;
    for (let i = 0; i < len; i++) {
      this.deleteEdge(deleteList[i]);
    }
    Util.freeArray(deleteList);
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {LayoutVertex} v
   * @return {boolean}
   */
  public isSingleton(v: LayoutVertex): boolean {
    if (v.sourceEdgesList.count > 0) return false;
    if (v.destinationEdgesList.count > 0) return false;
    return true;
  }

  /**
   * Modify this network by splitting it up into separate subnetworks,
   * each of which has all of its vertexes connected to each other, but not
   * to any vertexes in any other subnetworks.
   *
   * This method will first delete from this network all artificial vertexes and
   * all edges that do not connect two different vertexes.
   * Afterwards, this original network may be empty or may contain all of the
   * singleton vertexes, each of which had no edges connecting it to any other vertexes.
   * @export
   * @this {LayoutNetwork}
   * @param {boolean=} clean whether to delete artificial vertexes and edges, disconnected edges, and reflexive edges; default is true
   * @return {List.<LayoutNetwork>} a collection of {@link LayoutNetwork}s,
   * sorted in order of decreasing vertex count.
   */
  public splitIntoSubNetworks(clean?: boolean): List<LayoutNetwork> {
    if (clean === undefined) clean = true;
    if (clean) {
      this.deleteArtificialVertexes();
      this.deleteUselessEdges();
      this.deleteSelfEdges();
    }
    const nets = new List<LayoutNetwork>();
    let found = true;
    while (found) {
      found = false;
      const it = this._vertexes.iterator;
      while (it.next()) {
        const v = it.value;
        if (this.isSingleton(v)) continue;
        const net = this.layout.createNetwork();
        nets.add(net);
        this._traverseSubnet(net, v);
        found = true;  // start over again with all nodes
        break;
      }
    }
    nets.sort(function (a, b) {
      if (a === null || b === null || a === b) return 0;
      return b.vertexes.count - a.vertexes.count;
    });
    return nets;
  }

  /**
   * @hidden @internal
   * @this {LayoutNetwork}
   * @param {LayoutNetwork} net
   * @param {LayoutVertex} v
   */
  public _traverseSubnet(net: LayoutNetwork, v: LayoutVertex): void {
    if (v === null) return;
    if (v.network === net) return;
    this._removeVertex(v);
    net.addVertex(v);
    let it = v.sourceEdges;
    while (it.next()) {
      const se = it.value;
      if (se.network === net) continue;
      this._removeEdge(se);  // don't update any vertex's collections of edges
      net._addEdge(se);
      this._traverseSubnet(net, se.fromVertex);
    }
    it = v.destinationEdges;
    while (it.next()) {
      const de = it.value;
      if (de.network === net) continue;
      this._removeEdge(de);  // don't update the vertex's collections of edges
      net._addEdge(de);
      this._traverseSubnet(net, de.toVertex);
    }
  }

  /**
   * Retrieve all of the {@link Node}s and {@link Link}s from the
   * {@link LayoutVertex}es and {@link LayoutEdge}s that are in this network.
   * @export
   * @this {LayoutNetwork}
   * @return {Set.<Part>} A collection of {@link Part}s.
   * @since 1.1
   */
  public findAllParts(): Set<Part> {
    const coll = new Set<Part>();
    const vit = this._vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      coll.add(v.node);
    }
    const eit = this._edges.iterator;
    while (eit.next()) {
      const e = eit.value;
      coll.add(e.link);
    }
    return coll;
  }

} // #end LayoutNetwork
LayoutNetwork['className'] = 'LayoutNetwork';

/**
 * A vertex represents a {@link Node} in a {@link LayoutNetwork},
 * along with its {@link #bounds} and {@link #focus} and collections of {@link LayoutEdge}s that come into and go out of the vertex.
 * The {@link #node} property may be null for vertexes that represent "dummy" nodes,
 * when the layout wants to work with a network that is not isomorphic with the collection of Nodes and Links being laid out.
 *
 * This holds layout-specific data for the node.
 * For each kind of layout that uses a LayoutNetwork there is a subclass of {@link LayoutVertex} and a subclass of {@link LayoutEdge}:
 *   - {@link CircularLayout}: {@link CircularVertex} and {@link CircularEdge}
 *   - {@link ForceDirectedLayout}: {@link ForceDirectedVertex} and {@link ForceDirectedEdge}
 *   - {@link LayeredDigraphLayout}: {@link LayeredDigraphVertex} and {@link LayeredDigraphEdge}
 *   - {@link TreeLayout}: {@link TreeVertex} and {@link TreeEdge}
 *
 * Modifying a {@link LayoutNetwork} or a {@link LayoutVertex} or a {@link LayoutEdge} does not
 * invalidate the {@link Layout} or raise any changed events.
 * @unrestricted
 * @category Layout
 */
export class LayoutVertex {
  /** @hidden @internal */ public static className = 'LayoutVertex';
  /** @hidden @internal */ private _data: ObjectData | null;
  /** @hidden @internal */ private _node: Node |  null;
  /** @hidden @internal */ private _bounds: Rect;
  /** @hidden @internal */ private _focus: Point;
  /** @hidden @internal */ private _network: LayoutNetwork;
  /** @hidden @internal */ public sourceEdgesList: List<LayoutEdge>;
  /** @hidden @internal */ public destinationEdgesList: List<LayoutEdge>;

  /**
   * This constructs a vertex that does not know about any {@link Node}.
   */
  constructor(network: LayoutNetwork) {
    Set.uniqueHash(this);

    if (Debug && !network) Util.throwError('LayoutVertex constructor requires non-null LayoutNetwork argument');

    this._network = network;

    // these are not exposed because the property setters need to displace the coordinate values
    this._bounds = new Rect(0, 0, 10, 10).freeze();
    this._focus = new Point(5, 5).freeze();

    // clear:
    this._data = null;  // for virtualized layouts working on model data instead of a Node

    // this is not exposed because the property setter needs to initialize the other properties
    this._node = null;

    // internal state
    this.sourceEdgesList = new List<LayoutEdge>();
    this.destinationEdgesList = new List<LayoutEdge>();
  }

  /**
   * @hidden @internal
   * Remove all references to other objects, besides the owner network.
   * @this {LayoutVertex}
   */
  public clear(): void {
    this._data = null;  // for virtualized layouts working on model data instead of a Node
    // this is not exposed because the property setter needs to initialize the other properties
    this._node = null;
    // internal state
    this.sourceEdgesList = new List<LayoutEdge>();
    this.destinationEdgesList = new List<LayoutEdge>();
  }

  /**
   * @hidden @internal
   * @this {LayoutVertex}
   * @param {number=} details int Optional: If greater than zero, this includes more details about the vertex
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    let str = 'LayoutVertex#' + Set.hashId(this);
    if (details > 0) {
      str += (this.node !== null ? '(' + this.node.toString() + ')' : '');
      if (details > 1) {
        let srcs = '';
        let first = true;
        let eit = this.sourceEdgesList.iterator;
        while (eit.next()) {
          const se = eit.value;
          if (!first) {
            srcs += ',';
          } else {
            first = false;
          }
          srcs += se.toString(0);
        }
        let dsts = '';
        first = true;
        eit = this.destinationEdgesList.iterator;
        while (eit.next()) {
          const de = eit.value;
          if (!first) {
            dsts += ',';
          } else {
            first = false;
          }
          dsts += de.toString(0);
        }
        str += ' sources: ' + srcs + ' destinations: ' + dsts;
      }
    }
    return str;
  }

  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get sourceEdgesArrayAccess(): Array<LayoutEdge> { return this.sourceEdgesList._dataArray; }
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get destinationEdgesArrayAccess(): Array<LayoutEdge> { return this.destinationEdgesList._dataArray; }

  /**
   * Gets or sets the data object in the model that should be associated with this vertex, if any.
   * The value will be non-null if this vertex is virtualized,
   * not represented by a real {@link Node} in the {@link Diagram}.
   * Each data object should have a 'bounds' property that is a {@link Rect}
   * or any JavaScript Object with 'x', 'y', 'width', and 'height' properties that are real numbers,
   * and non-negative numbers for the width and the height.
   * Setting this data property will also set the {@link #bounds} and {@link #focus} properties.
   * @since 1.6
   */
  /** @export */ get data(): ObjectData | null { return this._data; }
  /** @export */ set data(value: ObjectData | null) {
    this._data = value;
    if (value !== null) {
      const b: { 'x': number, 'y': number, 'width': number, 'height': number } = value['bounds'];
      const x = b['x'];
      const y = b['y'];
      const w = b['width'];
      const h = b['height'];
      this._focus.assignTo(w / 2, h / 2);
      this._bounds.assignTo(x, y, w, h);
    }
  }

  /**
   * Gets or sets the {@link Node} associated with this vertex, if any.
   * The value may be null if this vertex is a "dummy" vertex,
   * not represented by a real {@link Node} in the {@link Diagram}.
   *
   * This property setter also sets the {@link #bounds} and {@link #focus} properties
   * according to the value of the Node's actualBounds and its {@link Part#locationObject}'s center point.
   */
  /** @export */ get node(): Node | null { return this._node; }
  /** @export */ set node(value: Node | null) {
    if (this._node !== value) {
      if (Debug && value !== null) Util.checkClass(value, Node, LayoutVertex, 'node');
      this._node = value;
      // make sure the width and height are known
      // was lay.maybeEsnureBounds(value), but caused problems in doubleCircle
      // because .location was never instantiated from NaN/NaN
      value.ensureBounds();
      const lay = this.network.layout;
      const temp = Rect.alloc();
      const ab = lay.getLayoutBounds(value, temp);
      let x = ab.x;
      let y = ab.y;
      const w = ab.width;
      const h = ab.height;
      // use zero if the position is not known
      if (isNaN(x)) x = 0;
      if (isNaN(y)) y = 0;
      this._bounds.assignTo(x, y, w, h);
      Rect.free(temp);

      // focus should be middle of locationObject, unless the node is a Group
      if (!(value instanceof Group)) {
        const loc = value.locationObject.getDocumentPoint(Spot.Center);
        if (loc.isReal()) {
          this._focus.assignTo(loc.x - x, loc.y - y);
          return;
        }
      }
      // but defaults to just the center of the node
      this._focus.assignTo(w / 2, h / 2);
    }
  }

  /**
   * Gets or sets the bounds of this vertex, in document coordinates.
   * Setting this property does not directly modify the position of any {@link Node},
   * but it will modify the value of {@link #centerX} and {@link #centerY},
   * as well as the subsidiary convenience properties {@link #x}, {@link #y}, {@link #width}, and {@link #height}.
   */
  /** @export */ get bounds(): Rect { return this._bounds; }
  /** @export */ set bounds(value: Rect) {
    if (!this._bounds.equals(value)) {
      if (Debug) Util.checkClass(value, Rect, LayoutVertex, 'bounds');
      this._bounds.assign(value);
    }
  }

  /**
   * Gets or sets the offset of the {@link #focusX} and {@link #focusY}
   * from the {@link #bounds} position.
   * Setting this property does not directly modify the position of any {@link Node},
   * but it does affect the value of {@link #centerX} and {@link #centerY}.
   *
   * The "focus" is normally the center of the Node's {@link Part#locationObject}.
   */
  /** @export */ get focus(): Point { return this._focus; }
  /** @export */ set focus(value: Point) {
    if (!this._focus.equals(value)) {
      if (Debug) Util.checkClass(value, Point, LayoutVertex, 'focus');
      this._focus.assign(value);
    }
  }

  /**
   * Gets or sets the center {@link Point#x} of this vertex, in document coordinates.
   * This property depends on both {@link #bounds} and {@link #focus}.
   * Setting this property does not modify the position of any {@link Node},
   * but it will modify the value of {@link #bounds}.
   */
  /** @export */ get centerX(): number { return this._bounds.x + this._focus.x; }
  /** @export */ set centerX(value: number) {
    const b = this._bounds;
    const old = b.x + this._focus.x;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, LayoutVertex, 'centerX');
      b.thaw();
      b.x = value - this._focus.x;
      b.freeze();
    }
  }

  /**
   * Gets or sets the center {@link Point#y} of this vertex, in document coordinates.
   * This property depends on both {@link #bounds} and {@link #focus}.
   * Setting this property does not modify the position of any {@link Node},
   * but it will modify the value of {@link #bounds}.
   */
  /** @export */ get centerY(): number { return this._bounds.y + this._focus.y; }
  /** @export */ set centerY(value: number) {
    const b = this._bounds;
    const old = b.y + this._focus.y;
    if (old !== value) {
      if (Debug) Util.checkRealNumber(value, LayoutVertex, 'centerY');
      b.thaw();
      b.y = value - this._focus.y;
      b.freeze();
    }
  }

  /**
   * Gets or sets the relative X position of the "center" point, the focus.
   */
  /** @export */ get focusX(): number { return this._focus.x; }
  /** @export */ set focusX(value: number) {
    const p = this._focus;
    const old = p.x;
    if (old !== value) {
      p.thaw();
      p.x = value;
      p.freeze();
    }
  }

  /**
   * Gets or sets the relative Y position of the "center" point, the focus.
   */
  /** @export */ get focusY(): number { return this._focus.y; }
  /** @export */ set focusY(value: number) {
    const p = this._focus;
    const old = p.y;
    if (old !== value) {
      p.thaw();
      p.y = value;
      p.freeze();
    }
  }

  /**
   * Gets or sets the left point of this vertex.
   */
  /** @export */ get x(): number { return this._bounds.x; }
  /** @export */ set x(value: number) {
    const b = this._bounds;
    const old = b.x;
    if (old !== value) {
      b.thaw();
      b.x = value;
      b.freeze();
    }
  }

  /**
   * Gets or sets the top point of this vertex.
   */
  /** @export */ get y(): number { return this._bounds.y; }
  /** @export */ set y(value: number) {
    const b = this._bounds;
    const old = b.y;
    if (old !== value) {
      b.thaw();
      b.y = value;
      b.freeze();
    }
  }

  /**
   * Gets or sets the width of this vertex.
   * This is always the horizontal distance reserved for the node.
   */
  /** @export */ get width(): number { return this._bounds.width; }
  /** @export */ set width(value: number) {
    const b = this._bounds;
    const old = b.width;
    if (old !== value) {
      b.thaw();
      b.width = value;
      b.freeze();
    }
  }

  /**
   * Gets or sets the height of this vertex.
   * This is always the vertical distance reserved for the node.
   */
  /** @export */ get height(): number { return this._bounds.height; }
  /** @export */ set height(value: number) {
    const b = this._bounds;
    const old = b.height;
    if (old !== value) {
      b.thaw();
      b.height = value;
      b.freeze();
    }
  }

  /**
   * Moves the {@link Node} corresponding to this vertex
   * so that its position is at the current {@link #bounds} point.
   *
   * To make the most common cases look right, the Node's {@link Part#locationObject}
   * is centered.  Thus iconic nodes will have the center of the icon be positioned
   * according to the center of this vertex, ignoring any labels.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayoutVertex}
   */
  public commit(): void {
    const data = this._data;
    if (data !== null) {
      // when virtualized, just set the 'bounds' on the data
      const vb = this.bounds;
      const nb: { 'x': number, 'y': number, 'width': number, 'height': number } = data['bounds'];
      if (Util.isObject(nb)) {
        nb['x'] = vb.x;
        nb['y'] = vb.y;
        nb['width'] = vb.width;
        nb['height'] = vb.height;
      } else {
        data['bounds'] = vb.copy();
      }
      return;
    }
    const node = this.node;
    if (node !== null) {
      const b = this.bounds;
      // focus should be middle of locationObject, unless the node is a Group
      if (!(node instanceof Group)) {
        const temp = Rect.alloc();
        const ab = this.network.layout.getLayoutBounds(node, temp);
        const loc = node.locationObject.getDocumentPoint(Spot.Center);
        if (ab.isReal() && loc.isReal()) {
          node.moveTo(b.x + this.focusX - (loc.x - ab.x), b.y + this.focusY - (loc.y - ab.y));
          Rect.free(temp);
          return;
        }
        Rect.free(temp);
      }
      node.moveTo(b.x, b.y);  /*?? , true */
    }
  }

  /**
   * Adds a {@link LayoutEdge} to the list of predecessors
   * (the edge will be coming into this vertex).
   * Calling this method does not modify the {@link LayoutEdge#toVertex} of the LayoutEdge.
   * @export
   * @this {LayoutVertex}
   * @param {LayoutEdge} edge
   */
  public addSourceEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutVertex, 'addSourceEdge:edge');
    if (!this.sourceEdgesList.contains(edge)) {
      this.sourceEdgesList.add(edge);
    }
  }

  /**
   * Deletes a {@link LayoutEdge} from the list of predecessors
   * (the edge was coming into this vertex).
   * Calling this method does not modify the {@link LayoutEdge#toVertex} of the LayoutEdge.
   * @export
   * @this {LayoutVertex}
   * @param {LayoutEdge} edge
   */
  public deleteSourceEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutVertex, 'deleteSourceEdge:edge');
    this.sourceEdgesList.remove(edge);
  }

  /**
   * Adds a {@link LayoutEdge} to the list of successors
   * (the edge will be going out from this vertex).
   * Calling this method does not modify the {@link LayoutEdge#fromVertex} of the LayoutEdge.
   * @export
   * @this {LayoutVertex}
   * @param {LayoutEdge} edge
   */
  public addDestinationEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutVertex, 'addDestinationEdge:edge');
    if (!this.destinationEdgesList.contains(edge)) {
      this.destinationEdgesList.add(edge);
    }
  }

  /**
   * Deletes a {@link LayoutEdge} from the list of successors
   * (the edge was going out from this vertex).
   * Calling this method does not modify the {@link LayoutEdge#fromVertex} of the LayoutEdge.
   * @export
   * @this {LayoutVertex}
   * @param {LayoutEdge} edge
   */
  public deleteDestinationEdge(edge: LayoutEdge): void {
    if (edge === null) return;
    if (Debug) Util.checkClass(edge, LayoutEdge, LayoutVertex, 'deleteDestinationEdge:edge');
    this.destinationEdgesList.remove(edge);
  }

  /**
   * Gets the {@link LayoutNetwork} that owns this vertex.
   * The default value is null.
   * It is set automatically for you in {@link LayoutNetwork#createVertex}.
   */
  /** @export */ get network(): LayoutNetwork { return this._network; }
  /** @export */ set network(value: LayoutNetwork) {
    if (Debug) Util.checkClass(value, LayoutNetwork, LayoutVertex, 'network');
    this._network = value;
  }

  /**
   * This read-only property returns an iterator for all of the vertexes that are connected with edges coming into this vertex.
   */
  /** @export */ get sourceVertexes(): Iterator<LayoutVertex> {
    const a = new Set<LayoutVertex>();
    const it = this.sourceEdges;
    while (it.next()) {
      const e = it.value;
      a.add(e.fromVertex);
    }
    return a.iterator;
  }

  /**
   * This read-only property returns an iterator for all of the vertexes that are connected with edges going out of this vertex.
   */
  /** @export */ get destinationVertexes(): Iterator<LayoutVertex> {
    const a = new Set<LayoutVertex>();
    const it = this.destinationEdges;
    while (it.next()) {
      const e = it.value;
      a.add(e.toVertex);
    }
    return a.iterator;
  }

  /**
   * This read-only property returns an iterator for all of the vertexes that are connected in either direction with this vertex.
   */
  /** @export */ get vertexes(): Iterator<LayoutVertex> {
    const a = new Set<LayoutVertex>();
    let it = this.sourceEdges;
    while (it.next()) {
      const se = it.value;
      a.add(se.fromVertex);
    }
    it = this.destinationEdges;
    while (it.next()) {
      const de = it.value;
      a.add(de.toVertex);
    }
    return a.iterator;
  }

  /**
   * This read-only property returns an iterator for all of the edges that come into this vertex.
   */
  /** @export */ get sourceEdges(): Iterator<LayoutEdge> { return this.sourceEdgesList.iterator; }

  /**
   * This read-only property returns an iterator for all of the edges that go out of this vertex.
   */
  /** @export */ get destinationEdges(): Iterator<LayoutEdge> { return this.destinationEdgesList.iterator; }

  /**
   * This read-only property returns an iterator for all of the edges that are connected with this vertex in either direction.
   */
  /** @export */ get edges(): Iterator<LayoutEdge> {
    const a = new List<LayoutEdge>();
    let it = this.sourceEdges;
    while (it.next()) {
      const se = it.value;
      a.add(se);
    }
    it = this.destinationEdges;
    while (it.next()) {
      const de = it.value;
      a.add(de);
    }
    return a.iterator;
  }

  /**
   * This read-only property returns the total number of edges that are connected with this vertex in either direction.
   */
  /** @export */ get edgesCount(): number {
    return this.sourceEdgesList.count + this.destinationEdgesList.count;
  }

  /**
   * This static function is used to compare the {@link Part#text} values of the {@link #node}s of
   * the argument {@link LayoutVertex}es.
   * This just does a simple case-sensitive string comparison.
   * Alternatively you may wish to use {@link LayoutVertex.smartComparer} or write your own comparison function.
   * @param {LayoutVertex} m
   * @param {LayoutVertex} n
   * @return {number} int -1, 0, or 1, depending on whether the comparison between M and N is less than, equal, or greater than.
   */
  public static standardComparer(m: LayoutVertex, n: LayoutVertex): number {
    if (Debug) Util.checkClass(m, LayoutVertex, LayoutVertex, 'standardComparer:m');
    if (Debug) Util.checkClass(n, LayoutVertex, LayoutVertex, 'standardComparer:n');
    const a = m._node;
    const b = n._node;
    if (a) {
      if (b) {
        const at = a.text;
        const bt = b.text;
        if (at < bt) return -1;
        if (at > bt) return 1;
        return 0;
      } else {
        return 1;
      }
    } else {
      if (b !== null) {
        return -1;
      } else {
        return 0;
      }
    }
  }

  /**
   * This static function is used to compare the {@link Part#text} values of the {@link #node}s of
   * the argument {@link LayoutVertex}es.
   * If the text ends in a number, the values are compared as numbers rather than lexicographically.
   * Naturally this comparison is significantly slower than {@link LayoutVertex.standardComparer}.
   * @param {LayoutVertex} m
   * @param {LayoutVertex} n
   * @return {number} int -1, 0, or 1, depending on whether the comparison between M and N is less than, equal, or greater than.
   */
  public static smartComparer(m: LayoutVertex, n: LayoutVertex): number {
    if (Debug) Util.checkClass(m, LayoutVertex, LayoutVertex, 'smartComparer:m');
    if (Debug) Util.checkClass(n, LayoutVertex, LayoutVertex, 'smartComparer:n');
    if (m !== null) {
      if (n !== null) {
        const a = m._node;
        const b = n._node;
        if (a !== null) {
          if (b !== null) {
            const aLower = a.text.toLocaleLowerCase();
            const bLower = b.text.toLocaleLowerCase();
            // split the string, separating numbers
            const aSplit = aLower.split(/([+\-]?[\.]?\d+(?:\.\d*)?(?:e[+\-]?\d+)?)/);
            const bSplit = bLower.split(/([+\-]?[\.]?\d+(?:\.\d*)?(?:e[+\-]?\d+)?)/);
            let i = 0;
            for (; i < aSplit.length; i++) {
              if (bSplit[i] !== '' && bSplit[i] !== undefined) {
                const aCurrentNum = parseFloat(aSplit[i]);
                const bCurrentNum = parseFloat(bSplit[i]);
                if (!isNaN(aCurrentNum)) {
                  if (!isNaN(bCurrentNum)) {
                    // both numbers
                    if (aCurrentNum - bCurrentNum !== 0) {
                      return aCurrentNum - bCurrentNum;
                    } else {
                      // same number, move to next element
                      continue;
                    }
                  } else {
                    // aSplit[i] is a number, bSplit[i] is NaN
                    return -1;
                  }
                } else {
                  if (!isNaN(bCurrentNum)) {
                    // aSplit[i] is NaN, bSplit[i] is a number
                    return 1;
                  } else {
                    // both strings
                    if (aSplit[i].localeCompare(bSplit[i]) !== 0) {
                      return aSplit[i].localeCompare(bSplit[i]);
                    } else {
                      // same string, move to next element
                      continue;
                    }
                  }
                }
              } else {
                // bSplit[i] is either "" or undefined
                if (aSplit[i] === '') continue;
                else return 1;
              }
            }
            // i is now aSplit.length
            if (bSplit[i] !== '' && bSplit[i] !== undefined) return -1;
            else return 0;
          } else {
            return 1;
          }
        } else {
          if (b !== null) {
            return -1;
          } else {
            return 0;
          }
        }
      } else {
        return 1;
      }
    } else {
      if (n !== null) {
        return -1;
      } else {
        return 0;
      }
    }
  }

} // #end LayoutVertex
LayoutVertex['className'] = 'LayoutVertex';
LayoutVertex['standardComparer'] = LayoutVertex.standardComparer;
LayoutVertex['smartComparer'] = LayoutVertex.smartComparer;

/**
 * An edge represents a Link in a {@link LayoutNetwork},
 * along with its {@link #fromVertex} and {@link #toVertex}.
 * The {@link #link} property may be null for edges that represent "dummy" links,
 * when the layout wants to work with a network that is not isomorphic with the collection of Nodes and Links being laid out.
 *
 * It holds layout-specific data for the link.
 * For each kind of layout that uses a LayoutNetwork there is a subclass of {@link LayoutVertex} and a subclass of {@link LayoutEdge}:
 *   - {@link CircularLayout}: {@link CircularVertex} and {@link CircularEdge}
 *   - {@link ForceDirectedLayout}: {@link ForceDirectedVertex} and {@link ForceDirectedEdge}
 *   - {@link LayeredDigraphLayout}: {@link LayeredDigraphVertex} and {@link LayeredDigraphEdge}
 *   - {@link TreeLayout}: {@link TreeVertex} and {@link TreeEdge}
 *
 * Modifying a {@link LayoutNetwork} or a {@link LayoutVertex} or a {@link LayoutEdge} does not
 * invalidate the {@link Layout} or raise any changed events.
 * @unrestricted
 * @category Layout
 */
export class LayoutEdge {
  /** @hidden @internal */ public static className = 'LayoutEdge';
  /** @hidden @internal */ private _network: LayoutNetwork;
  /** @hidden @internal */ private _data: ObjectData | null;
  /** @hidden @internal */ private _link: Link | null;
  /** @hidden @internal */ private _fromVertex: LayoutVertex | null;
  /** @hidden @internal */ private _toVertex: LayoutVertex | null;

  /**
   * This constructs an edge that does not know about any {@link Link}.
   */
  constructor(network: LayoutNetwork) {
    Set.uniqueHash(this);

    if (Debug && !network) Util.throwError('LayoutEdge constructor requires non-null LayoutNetwork argument');

    this._network = network;

    this._data = null;

    this._link = null;

    this._fromVertex = null;

    this._toVertex = null;
  }

  /**
   * @hidden @internal
   * Remove all references to other objects, besides the owner network.
   * @this {LayoutEdge}
   */
  public clear(): void {
    this._data = null;
    this._link = null;
    this._fromVertex = null;
    this._toVertex = null;
  }

  /**
   * @hidden @internal
   * @this {LayoutEdge}
   * @param {number=} details int Optional: If greater than zero, this includes more details about the edge
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    let str = 'LayoutEdge#' + Set.hashId(this);
    if (details > 0) {
      str += (this._link !== null ? '(' + this._link.toString() + ')' : '');
      if (details > 1) {
        str += ' ' + (this._fromVertex ? this._fromVertex.toString() : 'null') + ' --> ' + (this._toVertex ? this._toVertex.toString() : 'null');
      }
    }
    return str;
  }

  /**
   * @hidden @internal
   * @this {LayoutEdge}
   */
  public reverseEdge(): void {
    const temp = this._fromVertex;
    this._fromVertex = this._toVertex;
    this._toVertex = temp;
  }

  /**
   * Commits the route of this edge to the corresponding {@link Link}, if any.
   * By default this does nothing.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayoutEdge}
   */
  public commit(): void {
  }

  /**
   * Returns the edge's vertex at the other of this edge from the given vertex.
   * @export
   * @this {LayoutEdge}
   * @param {LayoutVertex} v
   */
  public getOtherVertex(v: LayoutVertex): LayoutVertex | null {
    if (Debug) Util.checkClass(v, LayoutVertex, LayoutEdge, 'getOtherVertex:v');
    if (this._toVertex === v) {
      return this._fromVertex;
    }
    if (this._fromVertex === v) {
      return this._toVertex;
    }
    return null;
  }

  /**
   * Gets the {@link LayoutNetwork} that owns this edge.
   * The default value is null.
   * It is set automatically for you in {@link LayoutNetwork#createEdge}.
   */
  /** @export */ get network(): LayoutNetwork { return this._network; }
  /** @export */ set network(value: LayoutNetwork) {
    if (Debug) Util.checkClass(value, LayoutNetwork, LayoutEdge, 'network');
    this._network = value;
  }

  /**
   * For virtualized layouts working on model data instead of a Link.
   * @since 1.6
   */
  /** @export */ get data(): ObjectData | null { return this._data; }
  /** @export */ set data(value: ObjectData | null) {
    if (this._data !== value) {
      if (Debug && value !== null) Util.checkPrimitive(value, 'object', LayoutEdge, 'data');
      this._data = value;
    }
  }

  /**
   * Gets or sets the {@link Link} associated with this edge, if any.
   * The value may be null if this edge is a "dummy" edge,
   * not represented by a real {@link Link} in the {@link Diagram}.
   */
  /** @export */ get link(): Link | null { return this._link; }
  /** @export */ set link(value: Link | null) {
    if (this._link !== value) {
      if (Debug && value !== null) Util.checkClass(value, Link, LayoutEdge, 'link');
      this._link = value;
    }
  }

  /**
   * Gets or sets the {@link LayoutVertex} that this edge comes from.
   * Setting this property does not change any {@link LayoutVertex#destinationEdges} collection.
   */
  /** @export */ get fromVertex(): LayoutVertex | null { return this._fromVertex; }
  /** @export */ set fromVertex(value: LayoutVertex | null) {
    if (this._fromVertex !== value) {
      if (Debug && value !== null) Util.checkClass(value, LayoutVertex, LayoutEdge, 'network');
      this._fromVertex = value;
    }
  }

  /**
   * Gets or sets the {@link LayoutVertex} that this edge goes to.
   * Setting this property does not change any {@link LayoutVertex#sourceEdges} collection.
   */
  /** @export */ get toVertex(): LayoutVertex | null { return this._toVertex; }
  /** @export */ set toVertex(value: LayoutVertex | null) {
    if (this._toVertex !== value) {
      if (Debug && value !== null) Util.checkClass(value, LayoutVertex, LayoutEdge, 'network');
      this._toVertex = value;
    }
  }

} // #end LayoutEdge
LayoutEdge['className'] = 'LayoutEdge';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
GridLayout
*/

// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Geo } from './geo';
// import { Layout } from './layout';
// import { Margin } from './margin';
// import { Group, Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Util } from './util';

/**
 * This simple layout places all of the Parts in a grid-like arrangement, ordered, spaced apart,
 * and wrapping as needed.  It ignores any Links connecting the Nodes being laid out.
 * There are many samples that use GridLayout.
 * Every {@link Palette} uses a GridLayout by default.
 * <p class="box">
 * If you want to experiment interactively with most of the properties, try the <a href="../../samples/gLayout.html">Grid Layout</a> sample.
 * See samples that make use of GridLayout in the <a href="../../samples/index.html#gridlayout">samples index</a>.
 *
 * By default this layout will sort all of the Parts alphabetically
 * (comparing {@link Part#text} values, respecting case)
 * and position them left-to-right, separated by {@link #spacing}`.width`,
 * until they do not fit in the current row.
 * At that time it starts a new row, separated from the previous row by {@link #spacing}`.height`.
 * There is a uniform cell size equal to the maximum Part width (plus spacing width)
 * and the maximum part height (plus spacing height).
 * At least one part is placed in each row, even if the part by itself is wider than the wrapping width.
 *
 * You can specify values for the {@link #cellSize} `width` and `height`.
 * If a part is wider than the cell size, it spans more than one cell in the row.
 * You can also specify a value for the {@link #wrappingWidth},
 * which will be used instead of the diagram's viewport width, to control when each row is considered "full".
 * The value of {@link Layout#isViewportSized} will be true when
 * the value of {@link #wrappingWidth} is `NaN`.
 * This causes the layout to be performed again automatically as the viewport changes size.
 *
 * You can also set {@link #wrappingColumn} to limit the number of items in each row.
 * Both the {@link #wrappingWidth} and the {@link #wrappingColumn} are respected when deciding when to wrap to the next row.
 *
 * This layout is sufficiently simple that it does not use a {@link LayoutNetwork}.
 * @extends Layout
 * @unrestricted
 * @category Layout
 */
export class GridLayout extends Layout {
  /** @hidden @internal */ public static className = 'GridLayout';
  /** @hidden @internal */ private _wrappingWidth: number;
  /** @hidden @internal */ private _wrappingColumn: number;
  /** @hidden @internal */ private _cellSize: Size;
  /** @hidden @internal */ private _spacing: Size;
  /** @hidden @internal */ private _alignment: EnumValue;
  /** @hidden @internal */ private _arrangement: EnumValue;
  /** @hidden @internal */ private _sorting: EnumValue;
  /** @hidden @internal */ private _comparer: ((a: Part, b: Part) => number);

  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(GridLayout);
    super();

    this.isViewportSized = true;  // because default value for wrappingWidth is NaN
    this._wrappingWidth = NaN;
    this._wrappingColumn = NaN;
    this._cellSize = new Size(NaN, NaN).freeze();
    this._spacing = new Size(10, 10).freeze();
    this._alignment = GridLayout.Location;
    this._arrangement = GridLayout.LeftToRight;
    this._sorting = GridLayout.Ascending;
    this._comparer = GridLayout.standardComparer;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Layout.
   * @this {GridLayout}
   * @param {?} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._wrappingWidth = this._wrappingWidth;
    copy._wrappingColumn = this._wrappingColumn;
    copy._cellSize.assign(this._cellSize);
    copy._spacing.assign(this._spacing);
    copy._alignment = this._alignment;
    copy._arrangement = this._arrangement;
    copy._sorting = this._sorting;
    copy._comparer = this._comparer;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {GridLayout}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === GridLayout) {
      if (arg === GridLayout.Ascending || arg === GridLayout.Descending ||
        arg === GridLayout.Forward || arg === GridLayout.Reverse) {
        this.sorting = arg;
      } else if (arg === GridLayout.LeftToRight || arg === GridLayout.RightToLeft) {
        this.arrangement = arg;
      } else if (arg === GridLayout.Location || arg === GridLayout.Position) {
        this.alignment = arg;
      } else {
        Util.throwError('Unknown enum value: ' + arg);
      }
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * Assign the positions of the parts, ignoring any links.
   * @this {GridLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');

    this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);

    // put parts in a set, getting rid of duplicates
    const allparts = this.collectParts(coll);
    const diagram = this.diagram;  // might be set by collectParts

    // copy allparts so even if a part is removed from allparts, it will still remove its memberParts
    const partlist = allparts.copy();
    const pit = partlist.iterator;
    while (pit.next()) {
      const p = pit.value;
      // ignore links unless they are completely disconnected
      if (p.isLinkLike()) {
        if (p.fromNode !== null || p.toNode !== null) {
          allparts.remove(p);
          continue;
        }
      }
      p.ensureBounds();  // make sure all parts have been measured
      // now we remove any parts contained by a Group
      if (p instanceof Group) {
        const mems = p.memberParts;
        while (mems.next()) {
          const m = mems.value;
          allparts.remove(m);
        }
      }
    }

    // order the parts, perhaps by sorting
    const partarray = allparts.toArray();
    if (partarray.length === 0) return;
    switch (this.sorting) {
      case GridLayout.Forward: break;
      case GridLayout.Reverse: partarray.reverse(); break;
      case GridLayout.Ascending: partarray.sort(this.comparer); break;
      case GridLayout.Descending: partarray.sort(this.comparer); partarray.reverse(); break;
    }

    let climit = this.wrappingColumn;
    if (isNaN(climit)) climit = 0;
    // calculate the wrapping width, which may depend on the panel's viewportBounds.width
    let xlimit = this.wrappingWidth;
    if (isNaN(xlimit) && diagram !== null) {
      const viewb = diagram.viewportBounds;
      const padding = diagram.padding as Margin;
      xlimit = Math.max(viewb.width - padding.left - padding.right, 0);
    } else {
      xlimit = Math.max(this.wrappingWidth, 0);
    }
    // must have some kind of limit
    if (climit <= 0 && xlimit <= 0) {
      climit = 1;
    }

    // get the minimum space between each part
    let spacex = this.spacing.width;
    if (!isFinite(spacex)) spacex = 0;
    let spacey = this.spacing.height;
    if (!isFinite(spacey)) spacey = 0;

    if (diagram !== null) diagram.startTransaction('Layout');
    const layerRects: Array<Rect> = [];
    switch (this.alignment) {
      case GridLayout.Position:
        this.doPositionLayout(partarray, xlimit, climit, spacex, spacey, layerRects);
        break;
      case GridLayout.Location:
        this.doLocationLayout(partarray, xlimit, climit, spacex, spacey, layerRects);
        break;
    }
    let minx = 0.0;
    let miny = 0.0;
    let maxx = 0.0;
    for (let i = 0; i < layerRects.length; i++) {
      const r = layerRects[i];
      minx = Math.min(minx, r.x);
      miny = Math.min(miny, r.y);
      maxx = Math.max(maxx, r.x + r.width);
    }
    if (this.arrangement === GridLayout.RightToLeft) {
      this.commitLayers(layerRects, new Point(minx + spacex / 2 - (maxx + minx), miny - spacey / 2));
    } else {
      this.commitLayers(layerRects, new Point(minx - spacex / 2, miny - spacey / 2));
    }
    if (diagram !== null) diagram.commitTransaction('Layout');
    this.isValidLayout = true;
  }

  /**
   * @hidden @internal
   * @this {GridLayout}
   * @param {Array.<Part>} partarray
   * @param {number} xlimit
   * @param {number} climit int
   * @param {number} spacex
   * @param {number} spacey
   * @param {Array.<Rect>} layerRects
   */
  private doPositionLayout(partarray: Array<Part>, xlimit: number, climit: number, spacex: number, spacey: number, layerRects: Array<Rect>): void {
    const temp = Rect.alloc();
    // calculate how wide each cell should be
    let cellW = Math.max(this.cellSize.width, 1);
    if (!isFinite(cellW)) {  // unknown -- calculate maximums
      cellW = 0;
      for (let i = 0; i < partarray.length; i++) {
        const p = partarray[i];
        const b = this.getLayoutBounds(p, temp);
        cellW = Math.max(cellW, b.width);
      }
    }
    cellW += spacex;
    cellW = Math.max(cellW, 1);

    // calculate how tall each cell should be
    let cellH = Math.max(this.cellSize.height, 1);
    if (!isFinite(cellH)) {
      cellH = 0;
      for (let i = 0; i < partarray.length; i++) {
        const p = partarray[i];
        const b = this.getLayoutBounds(p, temp);
        cellH = Math.max(cellH, b.height);
      }
    }
    cellH += spacey;
    cellH = Math.max(cellH, 1);

    // iterate through the parts, placing each one
    const arr = this.arrangement;
    const originx = this.arrangementOrigin.x;
    const originy = this.arrangementOrigin.y;
    let x = originx;
    let y = originy;
    let rowcount = 0;
    let rowheight = 0.0;
    for (let i = 0; i < partarray.length; i++) {
      const p = partarray[i];
      // figure out how many cells the part needs
      const b = this.getLayoutBounds(p, temp);
      const w = Math.ceil((b.width + spacex) / cellW);
      const h = Math.ceil((b.height + spacey) / cellH);
      const cszwidth = w * cellW;
      const cszheight = h * cellH;

      // maybe need to wrap to next row
      let needed = 0.0;
      switch (arr) {
        case GridLayout.RightToLeft: needed = Math.abs(x - b.width); break;
        default: needed = x + b.width; break;
      }
      // changed to account for location of origin
      if ((climit > 0 && rowcount > climit - 1) || (xlimit > 0 && rowcount > 0 && needed - originx > xlimit)) {
        layerRects.push(new Rect(0, y, xlimit + spacex, rowheight));
        rowcount = 0;
        x = originx;
        y += rowheight;
        rowheight = 0;
      }

      rowheight = Math.max(rowheight, cszheight);

      // assign new part Position
      let coff = 0.0;
      switch (arr) {
        case GridLayout.RightToLeft: coff = -b.width; break;
        default: coff = 0; break;
      }
      p.moveTo(x + coff, y);

      // advance to next column
      switch (arr) {
        case GridLayout.RightToLeft: x -= cszwidth; break;
        default: x += cszwidth; break;
      }
      rowcount++;
    }
    layerRects.push(new Rect(0, y, xlimit + spacex, rowheight));
    Rect.free(temp);
  }

  /**
   * @hidden @internal
   * @this {GridLayout}
   * @param {Array.<Part>} partarray
   * @param {number} xlimit
   * @param {number} climit int
   * @param {number} spacex
   * @param {number} spacey
   * @param {Array.<Rect>} layerRects
   */
  private doLocationLayout(partarray: Array<Part>, xlimit: number, climit: number, spacex: number, spacey: number, layerRects: Array<Rect>): void {
    const temp = Rect.alloc();
    // calculate how wide each cell should be
    let cellW = Math.max(this.cellSize.width, 1);
    let cellL = 0.0;
    let cellR = 0.0;
    let cellT = 0.0;
    const temploc = Point.alloc();
    for (let i = 0; i < partarray.length; i++) {
      const p = partarray[i];
      const b = this.getLayoutBounds(p, temp);
      // when AlignsLocation, accumulate separate maximums for left of the Location, and for right
      const relloc = p.getRelativePoint(p.locationObject, p.locationSpot, temploc);
      cellL = Math.max(cellL, relloc.x);
      cellR = Math.max(cellR, b.width - relloc.x);
      cellT = Math.max(cellT, relloc.y);
    }

    // account for space; which side depends on Arrangement
    const arr = this.arrangement;
    switch (arr) {
      case GridLayout.RightToLeft: cellL += spacex; break;
      default: cellR += spacex; break;
    }
    if (!isFinite(cellW)) {  // unknown -- calculate maximums
      cellW = Math.max(cellL + cellR, 1);
    } else {
      cellW = Math.max(cellW + spacex, 1);
    }

    // iterate through the parts, placing each part's Location at (X,Y)
    const originx = this.arrangementOrigin.x /* + cellL */;
    const originy = this.arrangementOrigin.y /* + cellT */;
    let x = originx;
    let y = originy;
    let rowcount = 0;

    if (xlimit >= cellL) xlimit -= cellL;   // remove space taken up at left edge beyond the location
    let xpos = 0.0;
    let minrelx = 0.0;

    // calculate how tall each cell should be
    const cellH = Math.max(this.cellSize.height, 1);
    cellT = 0.0;
    let cellB = 0.0;

    let firstRow = true; // We only layout the first row by strict locations, then the others we use position
    const temploc2 = Point.alloc();
    for (let i = 0; i < partarray.length; i++) {
      const p = partarray[i];
      // figure out how many cells the part needs
      const b = this.getLayoutBounds(p, temp);
      const relloc = p.getRelativePoint(p.locationObject, p.locationSpot, temploc);
      // assign new part Position
      if (rowcount > 0) {
        switch (arr) {
          case GridLayout.RightToLeft: {
            let tx = (x - originx - (b.width - relloc.x)) / cellW;
            if (Geo.isApproxEqual(Math.round(tx), tx)) tx = Math.round(tx);
            else tx = Math.floor(tx);
            x = tx * cellW + originx;
            break;
          }
          default: {
            let tx = (x - originx + relloc.x) / cellW;
            if (Geo.isApproxEqual(Math.round(tx), tx)) tx = Math.round(tx);
            else tx = Math.ceil(tx);
            x = tx * cellW + originx;
            break;
          }
        }
      } else {  // rowcount === 0
        switch (arr) {
          case GridLayout.RightToLeft: xpos = x + relloc.x + b.width; break;
          default: xpos = x - relloc.x; break;
        }
      }
      // maybe need to wrap to next row
      let needed = 0.0;
      // subtract the stuff sticking out beyond the end of the grid from the XLIMIT
      // changed fix problems with premature wrapping with RTL
      switch (arr) {
        case GridLayout.RightToLeft: needed = -(x + relloc.x) + xpos; break;
        default: needed = x + b.width - relloc.x - xpos; break;
      }

      // check the column limit or the width limit or both
      if ((climit > 0 && rowcount > climit - 1) || (xlimit > 0 && rowcount > 0 && needed > xlimit)) {
        // if firstRow, account for uneven edge on top
        layerRects.push(new Rect(0, (firstRow ? y - cellT : y), xlimit + spacex, cellB + cellT + spacey));

        for (let j = 0; j < rowcount; j++) {
          if (i === rowcount) break; // first row is set by location
          const prevp = partarray[i - rowcount + j];
          const prevrelloc = prevp.getRelativePoint(prevp.locationObject, prevp.locationSpot, temploc2);
          prevp.moveTo(prevp.position.x, prevp.position.y + cellT - prevrelloc.y);
        }

        // Done with the row, calculate the height of the cell bottom:
        // assume all space on bottom
        cellB += spacey;

        if (firstRow) {
          y += cellB;
        } else {
          y += cellB + cellT;
        }

        // Reset vals for next row
        cellT = 0;
        cellB = 0;
        rowcount = 0;
        x = originx;
        firstRow = false;
      }

      if (x === originx) {
        if (arr === GridLayout.RightToLeft) {
          minrelx = Math.max(minrelx, b.width - relloc.x);
        } else {
          minrelx = Math.min(minrelx, -relloc.x);
        }
      }

      // when AlignsLocation, accumulate separate maximums for above the Location, and for below
      cellT = Math.max(cellT, relloc.y);
      // consider the height of the part below the location point
      cellB = Math.max(cellB, b.height - relloc.y);
      if (isFinite(cellH)) cellB = Math.max(cellB, Math.max(b.height, cellH) - relloc.y);

      // the Position is just the Location minus the RelativeElementPoint
      if (firstRow) {
        p.moveTo(x - relloc.x, y - relloc.y);
      } else {
        p.moveTo(x - relloc.x, y);
      }

      // changed to account for space
      // advance to next column
      switch (arr) {
        case GridLayout.RightToLeft: x -= relloc.x + spacex; break;
        default: x += b.width - relloc.x + spacex; break;
      }
      rowcount++;
    }
    // last row
    layerRects.push(new Rect(0, y, xlimit + spacex, (firstRow ? cellB : cellB + cellT) + spacey));
    if (partarray.length !== rowcount) { // all in one row, first row is set by location
      for (let j = 0; j < rowcount; j++) {
        const prevp = partarray[partarray.length - rowcount + j];
        const prevrelloc = prevp.getRelativePoint(prevp.locationObject, prevp.locationSpot, temploc);
        prevp.moveTo(prevp.position.x, prevp.position.y + cellT - prevrelloc.y);
      }
    }
    Point.free(temploc);
    Point.free(temploc2);
    if (arr === GridLayout.RightToLeft) {
      // fix up layerRects to account for uneven edge on right side
      for (let i = 0; i < layerRects.length; i++) {
        const r = layerRects[i];
        r.width += minrelx;
        r.x -= minrelx;
      }
    } else {
      // fix up layerRects to account for uneven edge on left side
      for (let i = 0; i < layerRects.length; i++) {
        const r = layerRects[i];
        if (r.x > minrelx) {
          r.width += (r.x - minrelx);
          r.x = minrelx;
        }
      }
    }
    Rect.free(temp);
  }

  /**
   * @hidden @internal
   * This overridable method is called by {@link #doLayout}
   * to support custom arrangement of bands or labels across each layout layer.
   * By default this method does nothing.
   *
   * The coordinates used in the resulting Rects may need to be offset by the {@link Layout#arrangementOrigin}.
   * @expose
   * @this {GridLayout}
   * @param {Array.<Rect>} layerRects an Array of {@link Rect}s with the bounds of each of the "layers"
   * @param {Point} offset the position of the top-left corner of the banded area relative to the coordinates given by the layerRects
   * @since 1.4
   */
  public commitLayers(layerRects: Array<Rect>, offset: Point): void {
  }

  // ***************Properties***************

  /**
   * Gets or sets the wrapping width.
   *
   * The default is NaN, meaning to use the width of the diagram's panel's viewport.
   * Must be a value greater than 0.
   */
  /** @export */ get wrappingWidth(): number { return this._wrappingWidth; }
  /** @export */ set wrappingWidth(value: number) {
    if (this._wrappingWidth !== value) {
      Util.checkPrimitive(value, 'number', GridLayout, 'wrappingWidth');
      if (value > 0 || isNaN(value)) {
        this._wrappingWidth = value;
        this.isViewportSized = isNaN(value);
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the maximum number of columns.
   *
   * The default is NaN, meaning not to limit the number of columns.
   * 1 is a common value to produce a single column of parts.
   */
  /** @export */ get wrappingColumn(): number { return this._wrappingColumn; }
  /** @export */ set wrappingColumn(value: number) {
    if (this._wrappingColumn !== value) {
      Util.checkPrimitive(value, 'number', GridLayout, 'wrappingColumn');
      if (value > 0 || isNaN(value)) {
        this._wrappingColumn = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the minimum part size by which each part is positioned in the grid.
   *
   * The default value is NaN x NaN.
   * The units are in document coordinates.
   *
   * When the cell size is smaller than a part, the part will occupy more than one cell.
   * This allows parts to be positioned closer to each other, but then variations in part
   * sizes may cause them not to be aligned in perfect rows or columns.
   */
  /** @export */ get cellSize(): Size { return this._cellSize; }
  /** @export */ set cellSize(value: Size) {
    Util.checkClass(value, Size, GridLayout, 'cellSize');
    if (!this._cellSize.equals(value)) {
      this._cellSize.assign(value);
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the minimum horizontal and vertical space between parts.
   *
   * The default value is 10 x 10.
   * The units are in document coordinates.
   */
  /** @export */ get spacing(): Size { return this._spacing; }
  /** @export */ set spacing(value: Size) {
    Util.checkClass(value, Size, GridLayout, 'spacing');
    if (!this._spacing.equals(value)) {
      this._spacing.assign(value);
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the {@link Part#location} or the position should be used
   * to arrange each part.
   *
   * The default value is {@link GridLayout.Location} -- the {@link Part#location}s will
   * be aligned in a grid.
   */
  /** @export */ get alignment(): EnumValue { return this._alignment; }
  /** @export */ set alignment(value: EnumValue) {
    if (this._alignment !== value) {
      EnumValue.checkEnumValue(value, GridLayout, GridLayout, 'alignment');
      if (value === GridLayout.Location || value === GridLayout.Position) {
        this._alignment = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets how to arrange the parts.
   * Must be {@link GridLayout.LeftToRight} or {@link GridLayout.RightToLeft}.
   *
   * The default value is {@link GridLayout.LeftToRight}.
   */
  /** @export */ get arrangement(): EnumValue { return this._arrangement; }
  /** @export */ set arrangement(value: EnumValue) {
    if (this._arrangement !== value) {
      EnumValue.checkEnumValue(value, GridLayout, GridLayout, 'arrangement');
      if (value === GridLayout.LeftToRight || value === GridLayout.RightToLeft) {
        this._arrangement = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets what order to place the parts.
   * Must be {@link GridLayout.Forward}, {@link GridLayout.Reverse},
   * {@link GridLayout.Ascending}, or {@link GridLayout.Descending}.
   *
   * The default value is {@link GridLayout.Ascending}.
   */
  /** @export */ get sorting(): EnumValue { return this._sorting; }
  /** @export */ set sorting(value: EnumValue) {
    if (this._sorting !== value) {
      EnumValue.checkEnumValue(value, GridLayout, GridLayout, 'sorting');
      if (value === GridLayout.Forward || value === GridLayout.Reverse ||
        value === GridLayout.Ascending || value === GridLayout.Descending) {
        this._sorting = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the comparison function used to sort the parts.
   *
   * The default value is a case-insensitive alphabetic comparison
   * using the {@link Part#text} property of each part.
   * ```js
   *   $(go.GridLayout,
   *     {
   *       sorting: go.GridLayout.Ascending,
   *       comparer: function(pa, pb) {
   *         var da = pa.data;
   *         var db = pb.data;
   *         if (da.someProperty < db.someProperty) return -1;
   *         if (da.someProperty > db.someProperty) return 1;
   *         return 0;
   *       }
   *     }
   *   )
   * ```
   */
  /** @export */ get comparer(): ((a: Part, b: Part) => number) { return this._comparer; }
  /** @export */ set comparer(value: ((a: Part, b: Part) => number)) {
    if (this._comparer !== value) {
      Util.checkPrimitive(value, 'function', GridLayout, 'comparer');
      this._comparer = value;
      this.invalidateLayout();
    }
  }

  /**
   * @hidden @internal
   * This static function is used to compare the {@link Part#text} values of the argument {@link Part}s.
   * @param {Part} a
   * @param {Part} b
   * @return {number} int -1, 0, or 1, depending on whether the comparison between A and B is less than, equal, or greater than.
   */
  public static standardComparer(a: Part, b: Part): number {
    if (Debug) Util.checkClass(a, Part, GridLayout, 'standardComparer:a');
    if (Debug) Util.checkClass(b, Part, GridLayout, 'standardComparer:b');
    const at = a.text;
    const bt = b.text;
    if (at < bt) return -1;
    if (at > bt) return 1;
    return 0;
  }

  /**
   * @hidden @internal
   * This static function is used to compare the {@link Part#text} values of the argument {@link Part}s.
   * @param {Part} a
   * @param {Part} b
   * @return {number} int -1, 0, or 1, depending on whether the comparison between A and B is less than, equal, or greater than.
   */
  public static smartComparer(a: Part, b: Part): number {
    if (Debug) Util.checkClass(a, Part, GridLayout, 'standardComparer:a');
    if (Debug) Util.checkClass(b, Part, GridLayout, 'standardComparer:b');
    if (a !== null) {
      if (b !== null) {
        const aLower = a.text.toLocaleLowerCase();
        const bLower = b.text.toLocaleLowerCase();
        // split the string, separating numbers
        const aSplit = aLower.split(/([+\-]?[\.]?\d+(?:\.\d*)?(?:e[+\-]?\d+)?)/);
        const bSplit = bLower.split(/([+\-]?[\.]?\d+(?:\.\d*)?(?:e[+\-]?\d+)?)/);
        let i = 0;
        for (; i < aSplit.length; i++) {
          if (bSplit[i] !== '' && bSplit[i] !== undefined) {
            const aCurrentNum = parseFloat(aSplit[i]);
            const bCurrentNum = parseFloat(bSplit[i]);
            if (!isNaN(aCurrentNum)) {
              if (!isNaN(bCurrentNum)) {
                // both numbers
                if (aCurrentNum - bCurrentNum !== 0) {
                  return aCurrentNum - bCurrentNum;
                } else {
                  // same number, move to next element
                  continue;
                }
              } else {
                // aSplit[i] is a number, bSplit[i] is NaN
                return -1;
              }
            } else {
              if (!isNaN(bCurrentNum)) {
                // aSplit[i] is NaN, bSplit[i] is a number
                return 1;
              } else {
                // both strings
                if (aSplit[i].localeCompare(bSplit[i]) !== 0) {
                  return aSplit[i].localeCompare(bSplit[i]);
                } else {
                  // same string, move to next element
                  continue;
                }
              }
            }
          } else {
            // bSplit[i] is either "" or undefined
            if (aSplit[i] === '') continue;
            else return 1;
          }
        }
        // i is now aSplit.length
        if (bSplit[i] !== '' && bSplit[i] !== undefined) return -1;
        else return 0;
      } else {
        return 1;
      }
    } else {
      if (b !== null) return -1;
      else return 0;
    }
  }

  // ***************Enums***************

  // These values specify whether to postion each part by its location or by its position.
  /**
   * Position the top-left corner of each part at a grid point;
   * This value is used for {@link GridLayout#alignment}.
   * @constant
   */
  public static Position = new EnumValue(GridLayout, 'Position', 0);

  /**
   * Position the part's {@link Part#location} at a grid point;
   * This value is used for {@link GridLayout#alignment}.
   * @constant
   */
  public static Location = new EnumValue(GridLayout, 'Location', 1);

  // These values specify how to fill each row.
  /**
   * Fill each row from left to right;
   * This value is used for {@link GridLayout#arrangement}.
   * @constant
   */
  public static LeftToRight = new EnumValue(GridLayout, 'LeftToRight', 2);

  /**
   * Fill each row from right to left;
   * This value is used for {@link GridLayout#arrangement}.
   * @constant
   */
  public static RightToLeft = new EnumValue(GridLayout, 'RightToLeft', 3);

  // These values specify whether to sort the children of a part,
  // and in what order to position them.
  /**
   * Lay out each child in the order in which they were found;
   * This value is used for {@link GridLayout#sorting}.
   * @constant
   */
  public static Forward = new EnumValue(GridLayout, 'Forward', 4);

  /**
   * Lay out each child in reverse order from which they were found;
   * This value is used for {@link GridLayout#sorting}.
   * @constant
   */
  public static Reverse = new EnumValue(GridLayout, 'Reverse', 5);

  /**
   * Lay out each child according to the sort order given by {@link GridLayout#comparer}
   * This value is used for {@link GridLayout#sorting}.
   * @constant
   */
  public static Ascending = new EnumValue(GridLayout, 'Ascending', 6);

  /**
   * Lay out each child in reverse sort order given by {@link GridLayout#comparer}
   * This value is used for {@link GridLayout#sorting}.
   * @constant
   */
  public static Descending = new EnumValue(GridLayout, 'Descending', 7);

} // end class
GridLayout['className'] = 'GridLayout';

GridLayout['standardComparer'] = GridLayout.standardComparer;
GridLayout['smartComparer'] = GridLayout.smartComparer;

GridLayout['Position'] = GridLayout.Position;
GridLayout['Location'] = GridLayout.Location;
GridLayout['LeftToRight'] = GridLayout.LeftToRight;
GridLayout['RightToLeft'] = GridLayout.RightToLeft;
GridLayout['Forward'] = GridLayout.Forward;
GridLayout['Reverse'] = GridLayout.Reverse;
GridLayout['Ascending'] = GridLayout.Ascending;
GridLayout['Descending'] = GridLayout.Descending;


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
PartManager
*/

// import { ChangedEvent } from './changedEvent';
// import { Map, Set } from './collections';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { } from './globals';
// import { GraphObject } from './graphObject';
// import { Key } from './model';
// import { Panel } from './panel';
// import { Adornment, Group, Link, Node, Part } from './parts';
// import { Point } from './point';
// import { ObjectData } from './root';
// import { Util } from './util';

/**
 * @hidden @internal
 * The PartManager handles interaction between a {@link Diagram} and its {@link Model}.
 * It is responsible for creating and removing {@link Node}s, {@link Link}s, and {@link Part}s for the data in the model.
 *
 * Each {@link Diagram} has an instance of this class as its {@link Diagram#partManager} property.
 * This class maintains a collection of all diagram {@link Node}s, {@link Link}s, and {@link Part}s,
 * as well as dictionaries mapping model data to {@link Node}s and {@link Link}s.
 * You can find the {@link Part} for a model data object by calling {@link findNodeForData} and
 * {@link findLinkForData}.
 * @unrestricted
 * @since 2.0
 * @category Model
 */
export class PartManager {
  /** @hidden @internal */ public static className = 'PartManager';
  /** @hidden @internal */ private _nodes: Set<Node>;
  /** @hidden @internal */ private _links: Set<Link>;
  /** @hidden @internal */ private _parts: Set<Part>;
  /** @hidden @internal */ private _itemArrayPanelsMap: Map<Object, Array<Panel>>;
  /** @hidden @internal */ private _diagram: Diagram;
  /** @hidden @internal */ protected _addsToTemporaryLayer: boolean;
  /** @hidden @internal */ protected _dataPartMap: Map<Object, Part>;
  /** @hidden @internal */ protected _dataLinkMap: Map<Object, Link>;

  constructor() {
    this._nodes = new Set<Node>();
    this._links = new Set<Link>();
    this._parts = new Set<Part>();
    this._dataPartMap = new Map<Object, Part>();
    this._dataLinkMap = new Map<Object, Link>();
    this._itemArrayPanelsMap = new Map<Object, Array<Panel>>();  // Array of Panel
    this._diagram = null;
    // used by DraggingTool to add parts to the Tool layer
    // Necessary for an added part to be in the Tool layer before
    // Diagram.add is called so that the temporary part does not expand the document bounds
    this._addsToTemporaryLayer = false;
  }

  /**
   * Removes all Parts from the PartManager.
   * @this {PartManager}
   */
  public clear(): void {
    this._nodes.clear();
    this._links.clear();
    this._parts.clear();
    this._dataPartMap.clear();
    this._dataLinkMap.clear();
    this._itemArrayPanelsMap.clear();
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Diagram} val
   */
  public setDiagram(val: Diagram): void {
    if (Debug && val !== null) Util.checkClass(val, Diagram, PartManager, 'setDiagram');
    this._diagram = val;
  }

  /**
   * @hidden @internal
   * This is called from {@link Diagram#internalAdd} for bookkeeping.
   * @this {PartManager}
   * @param {Part} part
   */
  public internalAdd(part: Part): void {
    if (part instanceof Node) {
      this._nodes.add(part);
      if (part instanceof Group) {
        const cg = part.containingGroup;
        if (cg === null) {
          this._diagram._topLevelGroups.add(part);
        } else {
          cg._nestedGroups.add(part);
        }
        // update backpointer for Group.layout
        const layout = part.layout;
        if (layout !== null) layout.diagram = this._diagram;
      }
    } else if (part instanceof Link) {
      this._links.add(part);
    } else if (!(part instanceof Adornment)) {
      this._parts.add(part);
    }

    const data = part.data;
    if (data !== null) {
      if (!(part instanceof Adornment)) {
        if (part instanceof Link) {
          this.addDataLink(data, part);
        } else {
          this._dataPartMap.add(data, part);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * This is called from {@link Layer#internalRemove} for bookkeeping.
   * @this {PartManager}
   * @param {Part} part
   */
  public internalRemove(part: Part): void {
    part.clearAdornments();
    if (part instanceof Node) {
      this._nodes.remove(part);
      if (part instanceof Group) {
        const cg = part.containingGroup;
        if (cg === null) {
          this._diagram._topLevelGroups.remove(part);
        } else {
          cg._nestedGroups.remove(part);
        }
        // update backpointer for Group.layout
        const layout = part.layout;
        if (layout !== null) layout.diagram = null;
      }
    } else if (part instanceof Link) {
      this._links.remove(part);
    } else if (!(part instanceof Adornment)) {
      this._parts.remove(part);
    }

    const data = part.data;
    if (data !== null) {
      if (!(part instanceof Adornment)) {
        if (part instanceof Link) {
          this.removeDataLink(data, part);
        } else {
          this._dataPartMap.remove(data);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @param {Object} data
   * @param {Link} link
   */
  protected addDataLink(data: Object, link: Link): void {
    this._dataLinkMap.add(data, link);
  }

  /**
   * @hidden @internal
   * @param {Object} data
   * @param {Link} link
   */
  protected removeDataLink(data: Object, link: Link): void {
    this._dataLinkMap.remove(data);
  }

  /**
   * Remove all of the {@link Part}s created from model data
   * and then create them again.
   * This must be called after modifying or replacing any of the template maps
   * such as {@link Diagram#nodeTemplateMap}.
   * This re-selects all of the new Parts that were created from data of the original selected Parts.
   *
   * If you modify a template Map, there is no notification that the map has changed.
   * You will need to call rebuildParts explicitly.
   * If you are replacing the {@link Diagram#nodeTemplate} or the {@link Diagram#nodeTemplateMap} or
   * the corresponding properties for Groups or Links,
   * the Diagram property setters will automatically call rebuildParts.
   *
   * It is extremely wasteful to call this method after making some model data changes that you want to
   * be reflected in the diagram.  Instead, it is better call {@link Model#setDataProperty},
   * {@link Model#addNodeData}, {@link Model#removeNodeData}, or other model methods.
   * Not only do those methods update efficiently, they also preserve unbound state and support undo/redo.
   * @export
   * @this {PartManager}
   */
  public rebuildParts(): void {
    // make sure all template maps have proper kinds of templates in them
    const nit = this._diagram.nodeTemplateMap.iterator;
    while (nit.next()) {
      const part = nit.value;
      const name = nit.key;
      if (!part.isNodeLike() || part instanceof Group) {
        Util.throwError(`Invalid node template in Diagram.nodeTemplateMap: template for "${name}" must be a Node or a simple Part, not a Group or Link: ${part}`);
      }
    }
    const git = this._diagram.groupTemplateMap.iterator;
    while (git.next()) {
      const group = git.value;
      const gname = git.key;
      if (!(group instanceof Group)) {
        Util.throwError(`Invalid group template in Diagram.groupTemplateMap: template for "${gname}" must be a Group, not a normal Node or Link: ${group}`);
      }
    }
    const lit = this._diagram.linkTemplateMap.iterator;
    while (lit.next()) {
      const link = lit.value;
      const lname = lit.key;
      if (!(link instanceof Link)) {
        Util.throwError(`Invalid link template in Diagram.linkTemplateMap: template for "${lname}" must be a Link, not a normal Node or simple Part: ${link}`);
      }
    }

    // remember selected part data
    const seldata = Util.tempArray<Object>();
    const pit = this._diagram.selection.iterator;
    while (pit.next()) {
      const p1 = pit.value;
      const d = p1.data;
      if (d) seldata.push(d);
    }
    // remember highlighteds part data
    const highdata = Util.tempArray<Object>();
    const hit = this._diagram.highlighteds.iterator;
    while (hit.next()) {
      const p2 = hit.value;
      const d = p2.data;
      if (d) highdata.push(d);
    }
    // remember locations:
    const locdata = Util.tempArray<Object>(); // alternating data (object), location (Point)
    const nodes = this.nodes.iterator;
    while (nodes.next()) {
      const n = nodes.value;
      if (n.data !== null) {
        locdata.push(n.data);
        locdata.push(n.location);
      }
    }
    const links = this.links.iterator;
    while (links.next()) {
      const l = links.value;
      if (l.data !== null) {
        locdata.push(l.data);
        locdata.push(l.location);
      }
    }
    const parts = this.parts.iterator;
    while (parts.next()) {
      const p3 = parts.value;
      if (p3.data !== null) {
        locdata.push(p3.data);
        locdata.push(p3.location);
      }
    }

    this.removeAllModeledParts();
    this.addAllModeledParts();

    // restore selection
    for (let i = 0; i < seldata.length; i++) {
      const p = this.findPartForData(seldata[i]);
      if (p !== null) p.isSelected = true;
    }
    // restore highlights
    for (let i = 0; i < highdata.length; i++) {
      const p = this.findPartForData(highdata[i]);
      if (p !== null) p.isHighlighted = true;
    }
    // restore locations
    for (let i = 0; i < locdata.length; i += 2) {
      const p = this.findPartForData(locdata[i]);
      if (p !== null) p.location = locdata[i + 1] as Point;
    }

    Util.freeArray(seldata);
    Util.freeArray(highdata);
    Util.freeArray(locdata);
  }

  /**
   * Add {@link Part}s corresponding to all of the model data.
   *
   * By default, this will add all node data as {@link Node}s and then resolve any references.
   *
   * This method should be overridden by subclasses to handle any other
   * arrays of data that may be present in the respective {@link Model}.
   * Any override should call the super method to ensure node data is handled appropriately.
   * @expose
   * @this {PartManager}
   */
  public addAllModeledParts(): void {
    const model = this.diagram.model;
    this.addModeledParts(model.nodeDataArray);
  }

  /**
   * Add {@link Part}s corresponding to the given data array.
   *
   * This will also resolve references in the node data.
   *
   * This method should be overridden by subclasses to handle any other
   * arrays of data that may be present in the respective {@link Model}.
   * Any override should call the super method to ensure node data is handled appropriately.
   * @expose
   * @this {PartManager}
   * @param {Array<Object>} dataarray the array of node data to be added as Parts
   * @param {boolean=} invalidateLayouts whether or not to invalidate all the Diagram's layouts after adding parts
   */
  protected addModeledParts(dataarray: Array<Object>, invalidateLayouts?: boolean): void {
    const model = this.diagram.model;
    dataarray.forEach((data) => {
      if (model.containsNodeData(data)) {
        this.addModeledPart(data, false);
      }
    });
    // now resolve any references in the data
    dataarray.forEach((data) => {
      if (model.containsNodeData(data)) {
        this.resolveReferencesForData(data);
      }
    });
    if (invalidateLayouts !== false) this.diagram.invalidateAllLayouts(false);
  }

  /**
   * For a given data object, adds a {@link Part} to the {@link Diagram}
   * and optionally resolves any references for that data.
   * @this {PartManager}
   * @param {Object} data
   * @param {boolean=} resolve
   */
  protected addModeledPart(data: Object, resolve?: boolean): Part | null {
    if (data === undefined || data === null) return null;
    if (this.diagram.undoManager.isUndoingRedoing) return null;
    if (this._dataPartMap.contains(data)) return null;
    if (resolve === undefined) resolve = true;
    const p = this.addModeledNode(data);
    if (p !== null && resolve) this.resolveReferencesForData(data);
    return p;
  }

  /**
   * Add a {@link Node} to the {@link Diagram} for a given node data
   * @this {PartManager}
   * @param {Object} data
   * @return {Part}
   */
  protected addModeledNode(data: Object): Part | null {
    if (data === undefined || data === null) return null;
    if (this._diagram.undoManager.isUndoingRedoing) return null;
    if (this._dataPartMap.contains(data)) return null;

    const cat = this.getCategoryForNodeData(data);
    const template = this.findTemplateForNodeData(data, cat);
    if (template !== null) {
      template.freezeBindings();
      const part = template.copy();
      if (part !== null) {
        const oldskips = this.diagram.skipsModelSourceBindings;
        this.diagram.skipsModelSourceBindings = true;
        part._category = cat;  // don't call property setter
        part._data = data;  // don't call property setter
        if (this._addsToTemporaryLayer) part._layerName = 'Tool';
        this.diagram.add(part);
        part._data = null;  // don't call property setter
        part.data = data;  // now can call property setter and evaluate bindings
        this.diagram.skipsModelSourceBindings = oldskips;
        return part;
      }
    }
    return null;
  }

  /**
   * Make a copy of the {@link LinkingTool#archetypeLinkData}, set its node and port properties, and add it to the {@link GraphLinksModel}.
   * If the model is a {@link TreeModel}, set the parent key of the child's node data object.
   *
   * For GraphLinksModels, if {@link LinkingTool#archetypeLabelNodeData} is non-null, this method also adds that node data as an initial label node
   * for the new link data.
   *
   * For TreeModels, if {@link Diagram#isTreePathToChildren} is false, the roles of the *fromnode* and *tonode* arguments are exchanged.
   *
   * This method may be overridden.
   * Usually though it is easiest to just set {@link LinkingTool#archetypeLinkData} to control what kind of link is created
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {PartManager}
   * @param {Node} fromnode
   * @param {GraphObject} fromport the from-port, or null to use the node itself
   * @param {Node} tonode
   * @param {GraphObject} toport the to-port, or null to use the node itself
   * @return {Link} the newly created Link, or null if it failed.
   */
  public insertLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): Link | null {
    // a basic Model has no links, return null
    return null;
  }

  /**
   * Handle any references within a data object to other data objects in the model.
   * @expose
   * @this {PartManager}
   * @param {Object} data a data object in the model, potentially referencing other objects
   */
  protected resolveReferencesForData(data: Object) {
    // a basic Model has no references, do nothing
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Object} data a data for a Part/Node but not for a Link.
   * @return {string}
   */
  private getCategoryForNodeData(data: Object): string {
    const model = this._diagram.model;
    return model.getCategoryForNodeData(data);
  }

  /** @hidden @internal */
  public static _warnedAboutNoNodeTemplate = false;
  /** @hidden @internal */
  public static _warnedAboutNoGroupTemplate = false;

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Object} data a data for a Part/Node but not for a Link.
   * @param {string} cat the category.
   * @return {Part}
   */
  public findTemplateForNodeData(data: Object, cat: string): Part {
    const diagram = this._diagram;
    const model = diagram.model;
    const isgroup = model.isGroupModel() && model.isGroupForData(data);
    let template: Part = null;
    if (isgroup) {
      template = diagram.groupTemplateMap.getValue(cat) as Part;
      if (template === null) {
        template = diagram.groupTemplateMap.getValue('');
        if (template === null) {
          if (!PartManager._warnedAboutNoGroupTemplate) {
            PartManager._warnedAboutNoGroupTemplate = true;
            Util.trace('No Group template found for category \"' + cat + '\"');
            Util.trace('  Using default group template');
          }
          template = diagram._defaultGroupTemplate;
        }
      }
    } else {
      template = diagram.nodeTemplateMap.getValue(cat);
      if (template === null) {
        template = diagram.nodeTemplateMap.getValue('');
        if (template === null) {
          if (!PartManager._warnedAboutNoNodeTemplate) {
            PartManager._warnedAboutNoNodeTemplate = true;
            Util.trace('No Node template found for category \"' + cat + '\"');
            Util.trace('  Using default node template');
          }
          template = diagram._defaultNodeTemplate;
        }
      }
    }
    return template;
  }

  /**
   * Find the category of a given data representing a link, a string naming the link template
   * that the {@link Diagram} should use to represent the data.
   *
   * This should be overridden by subclasses that will be creating links from some model data.
   * @expose
   * @this {PartManager}
   * @param {Object} data a data for a Link (if GraphLinksModel) or for the child Node (if TreeModel)
   * @return {string}
   */
  public getLinkCategoryForData(data: Object): string {
    return ''; // a basic Model has no modeled links, return the empty string
  }

  /**
   * Change the category of a given data representing a link, a string naming the link template
   * that the {@link Diagram} should use to represent the data.
   *
   * This should be overridden by subclasses that will be creating links from some model data.
   * @expose
   * @this {PartManager}
   * @param {Object} data a data for a Link (if GraphLinksModel) or for the child Node (if TreeModel)
   */
  public setLinkCategoryForData(data: Object, value: string): void {
    // a basic Model has no modeled links, do nothing
  }

  /**
   * Change the node key that the given link references as the source of the link.
   *
   * This should be overridden by subclasses that will be creating links from some model data.
   * @expose
   * @this {PartManager}
   * @param {Link} link a Link
   * @param {Node} value the Node whose key will be referenced in the data
   * @param {Node} old an optional parameter holding the old value
   */
  public setFromNodeForLink(link: Link, value: Node, old?: Node): void {
    // a basic Model has no modeled links, do nothing
  }

  /**
   * Change the node key that the given data representing a link references as the
   * destination of the link.
   *
   * This should be overridden by subclasses that will be creating links from some model data.
   * @expose
   * @this {PartManager}
   * @param {Link} link a Link
   * @param {Node} value the Node whose key will be referenced in the data
   * @param {Node} old an optional parameter holding the old value
   */
  public setToNodeForLink(link: Link, value: Node, old?: Node): void {
    // a basic Model has no modeled links, do nothing
  }

  /** @hidden @internal */
  public static _warnedAboutNoLinkTemplate = false;

  /**
   * Looks up the Diagram's link template for a given category.
   * @expose
   * @this {PartManager}
   * @param {string} cat the category
   * @return {Link}
   */
  public findLinkTemplateForCategory(cat: string): Link {
    let template = this._diagram.linkTemplateMap.getValue(cat);
    if (template === null) {
      template = this._diagram.linkTemplateMap.getValue('');
      if (template === null) {
        if (!PartManager._warnedAboutNoLinkTemplate) {
          PartManager._warnedAboutNoLinkTemplate = true;
          Util.trace('No Link template found for category \"' + cat + '\"');
          Util.trace('  Using default link template');
        }
        template = this._diagram._defaultLinkTemplate;
      }
    }
    return template;
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Panel} part
   * @param {Object} olddata
   * @param {Object} newdata
   */
  public replaceBoundData(part: Panel, olddata: Object, newdata: Object): void {
    if (part instanceof Link) {
      if (olddata !== null) this.removeDataLink(olddata, part);
      if (newdata !== null) this.addDataLink(newdata, part);
    } else if (part instanceof Part) {
      if (olddata !== null) this._dataPartMap.remove(olddata);
      if (newdata !== null) this._dataPartMap.add(newdata, part);
    }
  }

  /**
   * Remove all of the {@link Part}s created from model data.
   * This will remove any part associated with the {@link Model#nodeDataArray}.
   *
   * This method can be overriden to remove parts associated with other model types like a
   * {@link GraphLinksModel#linkDataArray}.
   * @expose
   * @this {PartManager}
   */
  public removeAllModeledParts(): void {
    const m = this.diagram.model;
    this.removeModeledParts(m.nodeDataArray);
  }

  /**
   * Remove {@link Part}s associated with the given array of model data.
   *
   * You may override this method to handle other data arrays that a model subclass may need.
   * @export
   * @this {PartManager}
   * @param {Array<Object>} dataarray the array of data to be removed as Parts
   */
  protected removeModeledParts(dataarray: Array<Object>): void {
    dataarray.forEach((data) => {
      this.removeModeledPart(data);
    });
  }

  /**
   * Remove a {@link Part} given a data object.
   * @export
   * @this {PartManager}
   * @param {Object} data the node data to be removed as a Part
   */
  protected removeModeledPart(data: Object): void {
    const part = this.findPartForData(data);
    if (part !== null) {
      this.diagram.doRemove(part, false);
      this.unresolveReferencesForPart(part);
    }
  }

  /**
   * Remove any references that may exist for the given Part.
   *
   * By default, this does nothing.
   * This should be overridden by subclasses that will need to clear out references among data.
   * @expose
   * @this {PartManager}
   * @param {Part} part
   */
  protected unresolveReferencesForPart(part: Part): void {
    // a basic Model has no references, do nothing
  }

  /**
   * Remove model data associated with the given Link.
   *
   * By default, this does nothing.
   * This should be overridden by subclasses that will need to remove data for links that get deleted.
   * @expose
   * @this {PartManager}
   * @param {Link} link
   */
  public removeDataForLink(link: Link): void {
    // a basic Model has no modeled links, do nothing
  }

  /**
   * Look for a {@link Part} or {@link Node} or {@link Group} corresponding to a model's data object's unique key.
   *
   * This should be overridden by subclasses that may reference other {@link Part}s by a key.
   * @expose
   * @this {PartManager}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Part} null if a data with that key cannot be found in the model,
   * or if a corresponding {@link Part} cannot be found in the Diagram.
   */
  public findPartForKey(key: Key): Part {
    if (key === null || key === undefined) return null;
    const data = this._diagram.model.findNodeDataForKey(key);
    if (data !== null) {
      const part = this._dataPartMap.getValue(data);
      return part;
    }
    return null;
  }

  /**
   * Look for a {@link Node} or {@link Group} corresponding to a model's node data object's unique key.
   * @export
   * @this {PartManager}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Node} null if a node data with that key cannot be found in the model,
   * or if a corresponding {@link Node} or {@link Group} cannot be found in the Diagram,
   * or if what is found is just a {@link Part}.
   */
  public findNodeForKey(key: Key): Node {
    if (key === null || key === undefined) return null;
    const data = this._diagram.model.findNodeDataForKey(key);
    if (data === null) return null;
    const node = this._dataPartMap.getValue(data);
    if (node instanceof Node) return node;
    return null;
  }

  /**
   * Look for a {@link Link} corresponding to a model's link data object's unique key.
   *
   * This should be overridden by GraphLinksPartManager.
   * @expose
   * @this {PartManager}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Link} null if a link data with that key cannot be found in the model,
   * or if a corresponding {@link Link} cannot be found in the Diagram,
   * or if the model is not a {@link GraphLinksModel} with {@link GraphLinksModel#linkKeyProperty} set to a non-empty string.
   */
  public findLinkForKey(key: Key): Link {
    // a basic Model has no link keys, return null
    return null;
  }

  /**
   * Look for a {@link Part}, {@link Node}, {@link Group}, or {@link Link} corresponding to a {@link Model}'s data object.
   * We recommend that you call {@link #findNodeForData} or {@link #findLinkForData} if you are looking for a {@link Node} or a {@link Link}.
   * @export
   * @this {PartManager}
   * @param {Object} data a JavaScript object matched by reference identity
   * @return {Part} an existing {@link Part} in this Diagram that was
   * created because its `Part.data` was the data in the Diagram's {@link Model}.
   */
  public findPartForData(data: Object): Part {
    if (data === null) return null;
    let part = this._dataPartMap.getValue(data);
    if (part !== null) return part;
    part = this._dataLinkMap.getValue(data);
    return part;
  }

  /**
   * Look for a {@link Node} or {@link Group}  corresponding to a model's node data object.
   * @export
   * @this {PartManager}
   * @param {Object} nodedata a JavaScript object matched by reference identity;
   *   use {@link #findNodesByExample} if you want to find those Nodes whose data matches an example data object
   * @return {Node} an existing {@link Node} or {@link Group} in this Diagram that was
   * created because its `Part.data` was the node data in the Diagram's {@link Model}.
   * This will be null if there is no such part or if it's just a {@link Part} or {@link Link}.
   */
  public findNodeForData(nodedata: Object): Node {
    if (nodedata === null) return null;
    const node = this._dataPartMap.getValue(nodedata);
    if (node instanceof Node) return node;
    return null;
  }

  /**
   * Look for a {@link Link} corresponding to a {@link GraphLinksModel}'s link data object.
   * @export
   * @this {PartManager}
   * @param {Object} linkdata a JavaScript object matched by reference identity;
   *   use {@link #findLinksByExample} if you want to find those Links whose data matches an example data object
   * @param {Key=} key an optional key in case data can correspond to more than one link (GraphModel)
   * @return {Link} an existing {@link Link} in this Diagram that was
   * created because its `Part.data` was the link data in the Diagram's {@link Model}.
   */
  public findLinkForData(linkdata: Object, key?: Key): Link {
    if (linkdata === null) return null;
    const link = this._dataLinkMap.getValue(linkdata);
    return link;
  }

  /**
   * Return a collection of Nodes and Groups that are bound to data whose properties have values
   * that match those specified by the given example data.
   *
   * For example, calling this method with an argument object:
   * `{ sex: "M", name: /^Alex/i, age: function(n) { return n >= 18; } }`
   * will return an iterator of Nodes whose `Node.data` is a JavaScript object whose:
   *   - sex is "M" (a case-sensitive string comparison), and
   *   - name starts with the string "Alex" (using a case-insensitive match), and
   *   - age is greater than or equal to 18
   *
   * Here is how an example value can match the corresponding data value:
   *   - A string, number, or boolean is compared with the === operator.
   *   - A function is treated as a predicate and applied to the data value.
   *   - A regular expression (RexExp) is matched against the toString() of the data value.
   *     Common expressions include:
   *     - `/abc/` matches any string that includes exactly the substring "abc"
   *     - `/abc/i` matches any string that includes the substring "abc", but uses a case-insensitive comparison
   *     - `/^no/i` matches any string that starts with "no", ignoring case
   *     - `/ism$` matches any string that ends with "ism" exactly
   *     - `/(green|red) apple/` matches any string that includes either "green apple" or "red apple"
   *
   *     For more details read <a href="https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions">Regular Expressions (mozilla.org)</a>.
   *   - An Array requires the data value to also be an Array of equal or greater length.
   *     Each example array item that is not undefined is matched with the corresponding data array item.
   *   - An Object is recursively matched with the data value, which must also be an Object.
   *
   * All properties given by the argument example data must be present on the node data,
   * unless the example property value is undefined.
   * All other data properties are ignored when matching.
   *
   * When multiple argument objects are given, if any of the objects match the node's data,
   * the node is included in the results.
   * @export
   * @this {PartManager}
   * @param {...Object} examples one or more JavaScript Objects whose properties are either predicates to be
   *     applied or RegExps to be tested or values to be compared to the corresponding data property value
   * @return {Iterator.<Node>}
   * @see {@link #findLinksByExample}
   */
  public findNodesByExample(...examples: Array<Object>): Iterator<Node> {
    const result = new Set<Node>();
    const it = this._nodes.iterator;
    while (it.next()) {
      const n = it.value;
      const d = n.data;
      if (d === null) continue;
      for (let i = 0; i < arguments.length; i++) {
        const arg = arguments[i];
        if (Util.isObject(arg) && this.matchesData(d, arg)) {
          result.add(n);
          break;
        }
      }
    }
    return result.iterator;
  }

  /**
   * Return a collection of Links that are bound to data whose properties have values
   * that match those specified by the given example data.
   *
   * See the documentation of {@link #findNodesByExample} for how the example
   * data can match data of bound Parts.
   * @export
   * @this {PartManager}
   * @param {...Object} examples one or more JavaScript Objects whose properties are either predicates to be
   *     applied or RegExps to be tested or values to be compared to the corresponding data property value
   * @return {Iterator.<Link>}
   * @see {@link #findNodesByExample}
   */
  public findLinksByExample(...examples: Array<Object>): Iterator<Link> {
    const result = new Set<Link>();
    const it = this._links.iterator;
    while (it.next()) {
      const l = it.value;
      const d = l.data;
      if (d === null) continue;
      for (let i = 0; i < arguments.length; i++) {
        const arg = arguments[i];
        if (Util.isObject(arg) && this.matchesData(d, arg)) {
          result.add(l);
          break;
        }
      }
    }
    return result.iterator;
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Object} data
   * @param {Object} example
   * @return {boolean}
   */
  private matchesData(data: ObjectData, example: ObjectData): boolean {
    for (const p in example) {
      const dv = data[p];
      const ev = example[p];
      if (Util.isArray(ev)) {
        if (!Util.isArray(dv)) return false;
        const dvarr = dv;
        const evarr = ev;
        // just look for values at indexes specified by the example
        if (dvarr.length < evarr.length) return false;
        for (let i = 0; i < dvarr.length; i++) {
          const di = dvarr[i];
          const ei = evarr[i];
          if (ei === undefined) continue;  // if example item is undefined, don't try to match
          if (!this.matchesValue(di, ei)) return false;
        }
      } else {
        if (!this.matchesValue(dv, ev)) return false;
      }
    }
    return true;
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {*} dv
   * @param {*} ev
   * @return {boolean}
   */
  private matchesValue(dv: any, ev: any): boolean {
    if (typeof ev === 'function') {
      const pred: (v: any) => boolean = ev;
      if (!pred(dv)) return false;
    } else if (ev instanceof RegExp) {
      if (!dv || !ev.test(dv.toString())) return false;
    } else if (Util.isObject(dv) && Util.isObject(ev)) {
      if (!this.matchesData(dv, ev)) return false;
    } else {
      if (dv !== ev) return false;
    }
    return true;
  }

  /**
   * @hidden @internal
   * This is a temporary Model Changed event handler.
   * @this {Diagram}
   * @param {ChangedEvent} e
   */
  public doModelDataChanged(e: ChangedEvent): void {
    if (e.model !== this.diagram.model) return;
    if (!this.diagram.isUpdatingModel) return;
    this.diagram.isUpdatingModel = false;  // now Model is updating Diagram, not vice-versa
    try {
      const chg = e.change;
      const mname = e.modelChange;
      if (mname === '' && chg === ChangedEvent.Property) {
        this.updateDataBindings(e.object, e.propertyName);
      }
    } finally {
      this.diagram.isUpdatingModel = true;
    }
  }

  /**
   * This function is called when the Diagram's {@link #model} has changed.
   *
   * This function can be overridden by Model subclasses to handle any {@link ChangedEvent}s they might raise.
   * @expose
   * @this {PartManager}
   * @param {ChangedEvent} e
   */
  public doModelChanged(e: ChangedEvent): void {
    if (!this._diagram) return;
    const dia = this._diagram;
    if (e.model !== dia.model) return;

    const chg = e.change;

    dia.doModelChanged(e);

    if (!dia.isUpdatingModel) return;
    dia.isUpdatingModel = false;  // now Model is updating Diagram, not vice-versa
    try {
      const mname = e.modelChange;
      if (mname !== '') {
        // first consider all model data changes (not simple model property changes)
        if (chg === ChangedEvent.Property) {
          if (mname === 'nodeCategory') {
            const data = e.object;
            // if had been selected before, select it again
            const oldpart = this.findPartForData(data);
            const newcat = e.newValue;
            if (oldpart !== null && typeof newcat === 'string') {
              oldpart.category = newcat;
            }
          } else if (mname === 'nodeDataArray') {
            const olddata = e.oldValue as Array<Object>;
            this.removeModeledParts(olddata);
            const newdata = e.newValue as Array<Object>;
            this.addModeledParts(newdata);
          }
          dia.isModified = true;
          // end of ChangedEvent.Property
        } else if (chg === ChangedEvent.Insert) {  // named model-structure change
          const newdata = e.newValue;
          if (mname === 'nodeDataArray' && Util.isObject(newdata)) {  // model property
            this.addModeledPart(newdata);
          }
          dia.isModified = true;
          // end of ChangedEvent.Insert
        } else if (chg === ChangedEvent.Remove) {  // named model-structure change
          const olddata = e.oldValue;
          if (mname === 'nodeDataArray' && Util.isObject(olddata)) {  // model property
            this.removeModeledPart(olddata);
          }
          dia.isModified = true;
          // end of ChangedEvent.Remove
        } else if (chg === ChangedEvent.Transaction) {  // named model-structure change
          if (mname === 'SourceChanged') {  // for data bindings
            if (e.object !== null) {
              this.updateDataBindings(e.object, e.propertyName);
            } else {
              this.updateAllRelationshipsFromData();
              this.updateAllTargetBindings();
            }
          } else if (mname === 'ModelDisplaced') {  // for wholesale untracked model changes
            this.rebuildParts();
          }
          // end of ChangedEvent.Transaction
        }  // end of named model-structure change handling
      } else {  // mname === '', not a named model-structure change
        if (chg === ChangedEvent.Property) {
          const pname = e.propertyName;
          const data = e.object;
          const model = dia.model;
          if (data === model) {  // property change to model itself
            if (pname === 'nodeKeyProperty' || pname === 'nodeCategoryProperty') {
              if (!dia.undoManager.isUndoingRedoing) this.rebuildParts();
            }
          } else {  // property change to model data -- update the Part
            this.updateDataBindings(data, pname);
          }
          dia.isModified = true;
        } else if (chg === ChangedEvent.Insert || chg === ChangedEvent.Remove) {
          // changes to some data array, perhaps used as Panel.itemArray in multiple Panels
          this.modifyItemArray(e);
          dia.isModified = true;
        }
      }  // end of regular (not named model-structure) change
    } finally {
      dia.isUpdatingModel = true;
    }
  }

  /**
   * Update all of the data-bound properties of {@link Node}s and {@link Link}s in this diagram,
   * without having to call {@link Model#setDataProperty}.
   * This copies/converts model data properties to set properties on {@link Part}s.
   * This method does not conduct a transaction, so you need to start and commit one yourself.
   *
   * It is better to call {@link Model#setDataProperty} to modify data properties,
   * because that will both record changes for undo/redo and will update all bindings
   * that make depend on that property.
   * Simply modifying the data and calling an "update..." method will not be able to record
   * the previous value(s) of properties in the model data to support undo.
   *
   * If you know which model data objects have been modified,
   * it will be more efficient to update only the {@link Part}s that need it
   * by calling {@link Panel#updateTargetBindings}.
   *
   * To update relationships between nodes, call {@link #updateAllRelationshipsFromData}.
   * @expose
   * @this {PartManager}
   * @param {string=} srcprop An optional source data property name:
   *   when provided, only evaluates those {@link Binding}s that use that particular property;
   *   when not provided or when it is the empty string, all bindings are evaluated.
   * @see {@link #updateAllRelationshipsFromData}
   */
  public updateAllTargetBindings(srcprop?: string): void {
    if (srcprop === undefined) srcprop = '';
    let it = this.parts.iterator;
    while (it.next()) {
      const part = it.value;
      part.updateTargetBindings(srcprop);
    }
    it = this.nodes.iterator;
    while (it.next()) {
      const node = it.value;
      node.updateTargetBindings(srcprop);
    }
    it = this.links.iterator;
    while (it.next()) {
      const link = it.value as Link;
      link.updateTargetBindings(srcprop);
    }
  }

  /**
   * Add or remove any nodes or links according to additional or missing data objects in the model
   * and update all of the references to nodes, in case they had been modified in the model without
   * properly notifying the model by calling {@link Model#addNodeData} or
   * {@link GraphLinksModel#removeLinkData} or
   * {@link GraphLinksModel#setGroupKeyForNodeData} or
   * {@link GraphLinksModel#setToKeyForLinkData} or other similar methods.
   * This method does not conduct a transaction, so you need to start and commit one yourself.
   *
   * It is better to call {@link Model#addNodeData}, {@link Model#removeNodeData},
   * {@link GraphLinksModel#addLinkData}, {@link GraphLinksModel#removeLinkData},
   * {@link Model#setDataProperty}, and other model methods to add/remove/modify data,
   * because those methods will both record changes for undo/redo and will update all bindings
   * that make depend on that property.
   * Simply modifying the data and calling an "update..." method will not be able to record
   * the previous value(s) of properties in the model data to support undo.
   *
   * This only adds, removes, or updates the relationships between nodes and links,
   * to have them reflect what is now declared in the model data.
   * If you know which model data objects have been modified,
   * it will be more efficient to update only the {@link Part}s that need it
   * by calling {@link Part#updateRelationshipsFromData}.
   *
   * To update {@link GraphObject} properties that are data bound, call {@link Diagram#updateAllTargetBindings}.
   * @export
   * @this {PartManager}
   * @see {@link #updateAllTargetBindings}
   */
  public updateAllRelationshipsFromData(): void {
    const model = this._diagram.model;
    // add all new data bound parts/nodes/groups, and remember all existing and new ones
    const alldata = new Set<Object>();
    const narr = model.nodeDataArray;
    for (let i = 0; i < narr.length; i++) {
      const data = narr[i];
      alldata.add(data);
    }
    // remove all old nodes/groups
    const oldnodesarr: Array<Object> = [];
    this.nodes.each(function (n) {
      if (n.data !== null && !alldata.contains(n.data)) oldnodesarr.push(n.data);
    });
    this.parts.each(function (p) {
      if (p.data !== null && !alldata.contains(p.data)) oldnodesarr.push(p.data);
    });
    oldnodesarr.forEach(function (d) { model.doRemoveNodeData(d, false); });
    // add all new nodes/groups -- must do after removing old ones to avoid using new key due to apparent duplicate
    for (let i = 0; i < narr.length; i++) {
      const data = narr[i];
      const part = this.findPartForData(data);
      if (part === null) {
        model.doAddNodeData(data, false);
      }
    }
    this.refreshDataBoundLinks();

    let it = this.parts.iterator;
    while (it.next()) {
      const part = it.value;
      part.updateRelationshipsFromData();
    }
    it = this.nodes.iterator;
    while (it.next()) {
      const node = it.value;
      node.updateRelationshipsFromData();
    }
    it = this.links.iterator;
    while (it.next()) {
      const link = it.value as Link;
      link.updateRelationshipsFromData();
    }
  }

  /**
   * Remove any old data bound links and add all new data bound links.
   * By default, this does nothing.
   *
   * This should be overridden by subclasses that use a link data array.
   * @expose
   * @this {PartManager}
   */
  protected refreshDataBoundLinks(): void {
    // a basic Model has no link data, do nothing
  }

  /**
   * Given a Part, update any relationships based on its data.
   *
   * This should be overridden by subclasses that need to update Part relationships.
   * @expose
   * @param {Part} part
   */
  public updateRelationshipsFromData(part: Part): void {
    // a basic Model has no relationships, do nothing
  }

  /**
   * Re-evaluate a given binding on some Part associated with a data object.
   *
   * This method can be overridden to handle bindings that may not be stored in data specific to a Part.
   * An example would be a link binding stored on some node data in a TreeModel.
   * We recommend calling the super method in any override.
   * @expose
   * @this {PartManager}
   * @param {Object} data the data object for which the binding will be re-evaluated
   * @param {string|function(Object,*):*} pname the property name of the data to re-evaluate
   */
  protected updateDataBindings(data: Object, pname: string | ((a: Object, b: any) => any)): void {
    if (typeof pname === 'string') {
      const part = this.findPartForData(data);
      if (part !== null) {  // is node data or link data
        part.updateTargetBindings(pname);
      } else {  // maybe is itemArray data
        const panels = this.findPanelsForItem(data);
        if (panels !== null) {
          for (let i = 0; i < panels.length; i++) {
            const p = panels[i];
            p.updateTargetBindings(pname);
          }
          Util.freeArray(panels);
        }
      }
      // check for Model.modelData, which might also be a node data?
      if (data === this.diagram.model.modelData) {  // is modelData
        this.updateAllTargetBindings(pname);  // ?? this is inefficient
      }
    }
  }

  /**
   * @hidden @internal
   * Return an Array of item Panels that are bound to the given data.
   * There may be multiple container Panels whose .itemArray refers to the Array holding this data.
   * This assumes that a data object is not shared by different Arrays.
   * @this {Diagram}
   * @param {Object} data
   * @return {Array.<Panel>} null or a temporary Array that must be Util.freeArray()'d
   */
  private findPanelsForItem(data: Object): Array<Panel> | null {
    let results: Array<Panel> = null;
    const it = this._itemArrayPanelsMap.iterator;  // using this is slow!
    while (it.next()) {
      const panels = it.value;
      for (let i = 0; i < panels.length; i++) {
        const pan = panels[i];
        const p = pan.findItemPanelForData(data);
        if (p !== null) {
          if (results === null) results = Util.tempArray<Panel>();
          results.push(p);
        }
      }
      // assume the DATA is only in one ITEMARRAY, not shared by multiple Arrays;
      // but the item Array may be used by multiple container Panels,
      // so there can be multiple item Panels.
      if (results !== null) break;
    }
    return results;
  }

  /**
   * @hidden @internal
   * Support for ChangedEvent.Insert/Remove into/from item array for model data.
   * @this {Diagram}
   * @param {ChangedEvent} e
   */
  private modifyItemArray(e: ChangedEvent): void {
    const arr = e.object;
    const insert = e.change === ChangedEvent.Insert;
    const idx: number = insert ? e.newParam : e.oldParam;
    const val = insert ? e.newValue : e.oldValue;
    const panels = this._itemArrayPanelsMap.getValue(arr);
    // there may be multiple Panels whose Panel.itemArray refers to the modified Array
    if (Array.isArray(panels)) {
      for (let i = 0; i < panels.length; i++) {
        const p = panels[i];
        if (insert) {
          p.addItemElement(val, idx);
        } else {
          p.removeItemElement(idx);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Diagram}
   * @param {Panel} pan
   */
  public rememberItemArrayPanel(pan: Panel): void {
    const arr = pan._itemArray;
    if (Util.isArray(arr)) {
      const panels = this._itemArrayPanelsMap.getValue(arr);
      if (panels === null) {
        // if the array is unknown, remember the array and the one Panel
        const pans: Array<Panel> = [];
        pans.push(pan);
        this._itemArrayPanelsMap.add(arr, pans);
      } else {
        // is the Panel already known for the array, don't need to do anything
        for (let i = 0; i < panels.length; i++) {
          if (panels[i] === pan) return;
        }
        panels.push(pan); // otherwise add to the collection
      }
    }
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Panel} pan
   * @param {Diagram} diagram
   */
  public forgetItemArrayPanel(pan: Panel, diagram: Diagram): void {
    pan.walkPanels(function(p) { diagram.forgetPictures(p); });
    const arr = pan._itemArray;
    if (Util.isArray(arr)) {
      const panels = this._itemArrayPanelsMap.getValue(arr);
      if (panels !== null) {
        for (let i = 0; i < panels.length; i++) {
          if (panels[i] === pan) {
            panels.splice(i, 1);
            if (panels.length === 0) {
              this._itemArrayPanelsMap.remove(arr);
            }
            return;
          }
        }
      }
    }
  }

  // CommandHandler
  /**
   * @hidden @internal
   * This method makes a copy of a collection of {@link Part}s.
   * This is not a command and thus does not perform a transaction nor does it raise a {@link DiagramEvent}.
   * @this {PartManager}
   * @param {Iterable.<Part> | Array.<Part>} coll  A {@link List} or a {@link Set} of Parts, or an Array of Parts.
   * @param {Diagram} destdiagram  The destination diagram; if null, the copied parts are not added to this diagram.
   * @param {boolean} check  Whether to check {@link Part#canCopy} on each part.
   * @return {Map.<Part,Part>}
   */
  public copyParts(coll: Iterable<Part> | Array<Part>, destdiagram: Diagram | null, check: boolean): Map<Part, Part> {
    const dict = new Map<Part, Part>();
    if (Util.isArray(coll)) {
      for (let i = 0; i < coll.length; i++) {
        this.copyPart(coll[i], destdiagram, dict, check);
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        this.copyPart(it.value, destdiagram, dict, check);
      }
    }

    if (destdiagram !== null) {
      const model = destdiagram.model;
      const dragtool = destdiagram.toolManager.findTool('Dragging') as any;
      const dragsLink = dragtool !== null ? dragtool.dragOptions.dragsLink : destdiagram._defaultDragOptions.dragsLink;
      const discos = new Set<Link>();
      const plinks = new Map<Link, Link>();
      const dit = dict.iterator;
      while (dit.next()) {
        const part = dit.value;
        if (part instanceof Link) {
          // do not copy links that are partly or fully disconnected; but OK if dragsLink is true??
          if (!dragsLink && (part.fromNode === null || part.toNode === null)) {
            discos.add(part);
          }
        } else if (part instanceof Node && part.data !== null) {
          if (model.isTreeModel()) {
            const copiednode = part;
            // for a TreeModel, there is no link data to be copied for a Link --
            // so we have to look at the copied node data to see if we need to copy a parent Link
            const orignode = dit.key as Node;
            const origparent = orignode.findTreeParentNode();
            if (origparent !== null) {
              const copiedparent = dict.getValue(origparent) as Node;
              // modify the copied node data to have a reference to the copied parent data
              if (copiedparent !== null) {
                model.setParentKeyForNodeData(copiednode.data, model.getKeyForNodeData(copiedparent.data));
                // which should automatically create a parent Link in the target diagram
                // let's add the copied Link to the result Map for this copyParts method
                const copiedlink = destdiagram.findLinkForData(copiednode.data);
                const origlink = orignode.findTreeParentLink();
                if (origlink !== null && copiedlink !== null) plinks.add(origlink, copiedlink);
              } else {
                model.setParentKeyForNodeData(copiednode.data, undefined);
              }
            }
          } else if (model.isGraphModel()) {
            const copiednode = part;
            // for a GraphModel, there is no link data to be copied for a Link --
            // so we have to look at the copied node data to see if we need to copy incoming links
            const orignode = dit.key as Node;
            const origrels = model.isRelatedKeysPathTo ? orignode.findNodesOutOf() : orignode.findNodesInto();
            while (origrels.next()) {
              const origrel = origrels.value;
              const copiedrel = dict.getValue(origrel);
              // modify the copied node data to have a reference to the copied parent data
              if (copiedrel !== null) {
                model.addRelatedKeyForNodeData(copiednode.data, copiedrel.key);
                // which should automatically create a parent Link in the target diagram
                // let's add the copied Link to the result Map for this copyParts method
                const copiedlink = destdiagram.partManager.findLinkForData(copiednode.data, copiedrel.key);
                const origlink = this.findLinkForData(orignode.data, origrel.key);
                if (origlink !== null && copiedlink !== null) plinks.add(origlink, copiedlink);
              } else {
                model.removeRelatedKeyForNodeData(copiednode.data, origrel.key);
              }
            }
          }
        }
      }
      if (discos.count > 0) {
        destdiagram.removeParts(discos, false);
      }
      if (plinks.count > 0) {
        // add the copied parent Links to the resulting Map
        const pit = plinks.iterator;
        while (pit.next()) {
          const key = pit.key;
          const val = pit.value;
          dict.add(key, val);
        }
      }
    }

    // permit any second pass over the copied data
    if (destdiagram !== null && this._diagram !== null) {
      const dstmodel = destdiagram.model;
      const func = dstmodel.afterCopyFunction;
      if (func !== null) {
        const datadict = new Map<ObjectData, ObjectData>();
        dict.each(function (kvp) {
          if (kvp.key.data !== null) datadict.add(kvp.key.data, kvp.value.data);
        });
        const srcmodel = this._diagram.model;
        func(datadict, dstmodel, srcmodel);
      }
    }

    // update bindings in case any depend on source elements that were not available at the time
    // that the copy was added to the diagram
    const cit = dict.iterator;
    while (cit.next()) {
      const copy = cit.value;
      copy.updateTargetBindings();
    }
    return dict;
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Part} part
   * @param {Diagram} destdiagram  The destination diagram; if null, the copied parts are not added to this diagram.
   * @param {Map.<Part,Part>} dict
   * @param {boolean} check
   * @return {Part}  null if not copied.
   */
  protected copyPart(part: Part, destdiagram: Diagram, dict: Map<Part, Part>, check: boolean): Part {
    if (part === null) return null;
    if (check && !part.canCopy()) return null;

    // already there: just return it
    if (dict.contains(part)) {
      return dict.getValue(part);
    }

    // otherwise actually copy it:
    const copy = this.copyPartData(part, destdiagram);

    if (!(copy instanceof Part)) return null;
    copy.isSelected = false;  // ?? any other not-to-be-copied state
    copy.isHighlighted = false;
    dict.add(part, copy);
    if (part instanceof Node) { // Parts already taken care of by dict add
      // fix up all connected links
      const lit = part.linksConnected;
      while (lit.next()) {
        const origlink = lit.value;
        if (origlink.fromNode === part) {
          const newlink = dict.getValue(origlink) as Link;
          if (newlink !== null) {
            newlink.fromNode = copy as Node;
          }
        }
        if (origlink.toNode === part) {
          const newlink = dict.getValue(origlink) as Link;
          if (newlink !== null) {
            newlink.toNode = copy as Node;
          }
        }
      }
      // if it's a Group, copy its members
      if (part instanceof Group && copy instanceof Group) {
        const mit = part.memberParts;
        while (mit.next()) {
          const member = mit.value;
          // fix up container reference
          const memcopy = this.copyPart(member, destdiagram, dict, check);
          if (memcopy instanceof Link) {
            //  memcopy.setContainingGroup(copyg);  // not needed
          } else if (memcopy !== null) {
            memcopy.containingGroup = copy;
          }
        }
      }
    } else if (part instanceof Link && copy instanceof Link) {
      const origfrom = part.fromNode;
      if (origfrom !== null) {
        const newfrom = dict.getValue(origfrom);
        if (newfrom !== null) {
          copy.fromNode = newfrom as Node;
        }
      }
      const origto = part.toNode;
      if (origto !== null) {
        const newto = dict.getValue(origto);
        if (newto !== null) {
          copy.toNode = newto as Node;
        }
      }
      // if it has label nodes, copy them
      const labs = part.labelNodes;
      while (labs.next()) {
        const lab = labs.value;
        // fix up labeled Link reference
        const labcopy = this.copyPart(lab, destdiagram, dict, check);
        if (labcopy !== null && labcopy instanceof Node) {
          labcopy.labeledLink = copy;
        }
      }
    }

    return copy;
  }

  /**
   * This method handles the data copying portion of copying a {@link Part}.
   * This method is called while copying parts, and should not be called by the user.
   *
   * The method may be overridden to handle the copying of {@link Link}s or other special
   * data that may be stored in a {@link Model} subclass.
   * @expose
   * @this {PartManager}
   * @param {Part} part
   * @param {Diagram} destdiagram
   */
  protected copyPartData(part: Part, destdiagram: Diagram): Part | null {
    let copy = null;
    const data = part.data;
    if (data !== null && destdiagram !== null) {  // bound Part
      const model = destdiagram.model;
      if (!(part instanceof Link)) {
        const datacopy = model.copyNodeData(data);
        if (Util.isObject(datacopy)) {
          model.addNodeData(datacopy);
          copy = destdiagram.findPartForData(datacopy);
        }
      }
    } else {  // unbound Part
      part.freezeBindings();
      copy = part.copy();
      if (copy !== null) {
        const diagram = this._diagram;
        if (destdiagram !== null) {
          destdiagram.add(copy);
        } else if (data !== null && diagram !== null && diagram.commandHandler !== null && diagram.commandHandler.copiesClipboardData) {  // copy the bound data object too
          const model = diagram.model;  // don't have a target diagram, so use diagram
          let datacopy: Object = null;
          if (!(copy instanceof Link)) {
            datacopy = model.copyNodeData(data);
          }
          if (Util.isObject(datacopy)) {
            copy.data = datacopy;
          }
        }
      }
    }
    return copy;
  }

  /**
   * This read-only property returns a set of all {@link Node}s and {@link Group}s in the Diagram.
   *
   * This includes both data-bound and unbound nodes,
   * and both top-level nodes and nodes inside {@link Group}s.
   * All of the simple {@link Part}s are accessible via the {@link #parts} property.
   * @see {@link Diagram#findTopLevelGroups}
   * @see {@link #findTreeRoots}
   */
  /** @export */ get nodes(): Set<Node> { return this._nodes; }

  /**
   * This read-only property returns a set of all {@link Link}s in the Diagram.
   *
   * This includes both data-bound and unbound links,
   * and both top-level links and links inside {@link Group}s.
   */
  /** @export */ get links(): Set<Link> { return this._links; }

  /**
   * This read-only property returns a set of all {@link Part}s in the Diagram
   * that are not {@link Node}s or {@link Link}s or {@link Adornment}s.
   *
   * This includes both data-bound and unbound parts,
   * and both top-level parts and parts inside {@link Group}s.
   * Use the {@link #nodes} or {@link #links} properties for getting the
   * collection of all {@link Node}s or {@link Link}s in the diagram.
   */
  /** @export */ get parts(): Set<Part> { return this._parts; }

  /**
   * This read-only property returns the {@link Diagram} for which this class manages {@link Part}s.
   */
  /** @export */ get diagram(): Diagram { return this._diagram; }

  /**
   * Gets or sets whether newly added {@link Part}s should be in a temporary {@link Layer}.
   *
   * This is used by the {@link DraggingTool} to ensure bounds won't be improperly updated during drags.
   */
  /** @export */ get addsToTemporaryLayer(): boolean { return this._addsToTemporaryLayer; }
  /** @export */ set addsToTemporaryLayer(value: boolean) {
    Util.checkPrimitive(value, 'boolean', PartManager, 'addsToTemporaryLayer');
    this._addsToTemporaryLayer = value;
  }

}
PartManager['className'] = 'PartManager';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
GraphLinksPartManager
*/

// import { ChangedEvent } from './changedEvent';
// import { Set } from './collections';
// import { Diagram} from './diagram';
// import { GraphLinksModel } from './graphLinksModel';
// import { GraphObject } from './graphObject';
// import { Key } from './model';
// import { PartManager } from './partManager';
// import { Adornment, Group, Link, Node, Part } from './parts';

/**
 * @hidden @internal
 * The GraphLinksPartManager handles interaction between a {@link Diagram} and its {@link GraphLinksModel}.
 * It is responsible for creating and removing {@link Node}s, {@link Link}s, and {@link Part}s for the data in the model.
 *
 * This class extends the normal {@link PartManager} with more functionality related to links and groups.
 * @extends PartManager
 * @unrestricted
 * @since 2.0
 * @category Model
 */
export class GraphLinksPartManager extends PartManager {
  /** @hidden @internal */ public static className = 'GraphLinksPartManager';
  /**
   * Add {@link Part}s corresponding to all of the model data.
   *
   * This will add all node data as well as link data.
   * @export
   * @this {GraphLinksPartManager}
   */
  public addAllModeledParts() {
    const model = this.diagram.model as GraphLinksModel;
    this.addModeledParts(model.nodeDataArray); // first call the super which will add node data
    this.addModeledLinks(model.linkDataArray); // now add link data
  }

  /**
   * Add {@link Part}s corresponding to the given data array.
   *
   * This will also resolve references in the node data and update group membership for links.
   * @export
   * @this {GraphLinksPartManager}
   * @param {Array<Object>} dataarray the array of node data to be added as Parts
   */
  protected addModeledParts(dataarray: Array<Object>) {
    super.addModeledParts(dataarray, false); // first call the super which will add node data

    // nodes may have changed group membership -- need to update all existing links
    const lit = this.links.iterator;
    while (lit.next()) {
      const link = lit.value;
      link.updateContainingGroup();
    }
    this.diagram.invalidateAllLayouts(false);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksPartManager}
   * @param {Array<Object>} dataarray the array of link data to be added as Links
   */
  private addModeledLinks(dataarray: Array<Object>) {
    dataarray.forEach((data) => {
      this.addModeledLink(data);
    });
    this.diagram.invalidateAllLayouts(false);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksPartManager}
   * @param {Object} data
   */
  private addModeledLink(data: Object) {
    if (data === undefined || data === null) return null;
    if (this.diagram.undoManager.isUndoingRedoing) return null;
    if (this._dataLinkMap.contains(data)) return null;

    const cat = this.getLinkCategoryForData(data);
    const template = this.findLinkTemplateForCategory(cat);
    if (template !== null) {
      template.freezeBindings();
      const link = template.copy();
      if (link !== null) {
        const oldskips = this.diagram.skipsModelSourceBindings;
        this.diagram.skipsModelSourceBindings = true;
        link._category = cat;  // don't call property setter
        link._data = data;  // don't call property setter

        // try to resolve references to Nodes
        const model = this.diagram.model as GraphLinksModel;

        const frompid = model.getPortIdForLinkData(data, true);
        if (frompid !== '') {
          link.fromPortId = frompid;
        }
        const fromid = model.getNodeKeyForLinkData(data, true);
        if (fromid !== undefined) {
          const fromnode = this.findNodeForKey(fromid);
          if (fromnode instanceof Node) {
            link.fromNode = fromnode;
          }
        }

        const topid = model.getPortIdForLinkData(data, false);
        if (topid !== '') {
          link.toPortId = topid;
        }
        const toid = model.getNodeKeyForLinkData(data, false);
        if (toid !== undefined) {
          const tonode = this.findNodeForKey(toid);
          if (tonode instanceof Node) {
            link.toNode = tonode;
          }
        }

        // label nodes
        const labkeys = model.getLabelKeysForLinkData(data);
        if (Array.isArray(labkeys)) {
          labkeys.forEach((k) => {
            const node = this.findNodeForKey(k);
            if (node !== null) {
              node.labeledLink = link;
            }
          });
        }

        if (this._addsToTemporaryLayer) link._layerName = 'Tool';
        this.diagram.add(link);
        link._data = null;  // don't call property setter
        link.data = data;  // now we can call property setter and evaluate bindings
        this.diagram.skipsModelSourceBindings = oldskips;
        return link;
      }
    }
    return null;
  }

  /**
   * Remove all of the {@link Part}s created from model data.
   * This will remove any part associated with the {@link GraphLinksModel#linkDataArray} or {@link Model#nodeDataArray}.
   * @export
   * @this {GraphLinksPartManager}
   */
  public removeAllModeledParts() {
    const model = this.diagram.model as GraphLinksModel;
    this.removeModeledLinks(model.linkDataArray);
    this.removeModeledParts(model.nodeDataArray);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksPartManager}
   * @param {Array<Object>} dataarray
   */
  private removeModeledLinks(dataarray: Array<Object>) {
    dataarray.forEach((data) => {
      this.removeModeledPart(data);
    });
  }

  /**
   * Find the category of a given link data, a string naming the link template
   * that the {@link Diagram} should use to represent the link data.
   *
   * This calls {@link GraphLinksModel#getCategoryForLinkData}.
   * @this {GraphLinksPartManager}
   * @param {Object} data a JavaScript object represented by a link
   * @return {string}
   */
  public getLinkCategoryForData(data: Object): string {
    const model = this.diagram.model as GraphLinksModel;
    return model.getCategoryForLinkData(data);
  }

  /**
   * Change the category of a given link data, a string naming the link template
   * that the {@link Diagram} should use to represent the link data.
   *
   * This calls {@link GraphLinksModel#setCategoryForLinkData}.
   * @this {GraphLinksPartManager}
   * @param {Object} data a JavaScript object represented by a link
   * @param {string} value the category, must not be null
   */
  public setLinkCategoryForData(data: Object, value: string): void {
    const model = this.diagram.model as GraphLinksModel;
    return model.setCategoryForLinkData(data, value);
  }

  /**
   * Change the node key that the given link references as the source of the link.
   *
   * This calls {@link GraphLinksModel#setFromKeyForLinkData}.
   * @this {PartManager}
   * @param {Link} link a Link
   * @param {Node} value the Node whose key will be referenced in the data
   */
  public setFromNodeForLink(link: Link, value: Node): void {
    const model = this.diagram.model as GraphLinksModel;
    // if VAL is null or if VAL.data is null, the link data's node key is set to undefined
    const ndata = (value !== null ? value.data : null);
    model.setFromKeyForLinkData(link.data, model.getKeyForNodeData(ndata));
  }

  /**
   * Change the node key that the given link references as the destination of the link.
   *
   * This calls {@link GraphLinksModel#setToKeyForLinkData}.
   * @this {PartManager}
   * @param {Link} link a Link
   * @param {Node} value the Node whose key will be referenced in the data
   */
  public setToNodeForLink(link: Link, value: Node): void {
    const model = this.diagram.model as GraphLinksModel;
    // if VAL is null or if VAL.data is null, the link data's node key is set to undefined
    const ndata = (value !== null ? value.data : null);
    model.setToKeyForLinkData(link.data, model.getKeyForNodeData(ndata));
  }

  /**
   * Remove model data associated with the given Link.
   *
   * This calls {@link GraphLinksModel#removeLinkData}.
   * @this {GraphLinksPartManager}
   * @param {Link} link
   */
  public removeDataForLink(link: Link): void {
    const model = this.diagram.model as GraphLinksModel;
    model.removeLinkData(link.data);
  }

  /**
   * Look for a {@link Part} or {@link Node} or {@link Group} corresponding to a model's data object's unique key.
   * This will find a {@link Link} if the model is a {@link GraphLinksModel} that is maintaining a key on the link data objects.
   * @export
   * @this {GraphLinksPartManager}
   * @param {Key} key a string or number.
   * @return {Part} null if a data with that key cannot be found in the model,
   * or if a corresponding {@link Part} cannot be found in the Diagram.
   * This will not return a {@link Link} unless the model is a {@link GraphLinksModel} and
   * {@link GraphLinksModel#linkKeyProperty} has been set.
   * If the same key is used for both a node data object and a link data object, this will return a Node.
   */
  public findPartForKey(key: Key): Part {
    const part = super.findPartForKey(key);
    if (part === null) {
      const model = this.diagram.model as GraphLinksModel;
      const data = model.findLinkDataForKey(key);
      if (data !== null) {
        const link = this._dataLinkMap.getValue(data);
        return link;
      }
    }
    return part;
  }

  /**
   * Look for a {@link Link} corresponding to a model's link data object's unique key.
   * @export
   * @this {GraphLinksPartManager}
   * @param {(string|number|undefined)} key a string or number.
   * @return {Link} null if a link data with that key cannot be found in the model,
   * or if a corresponding {@link Link} cannot be found in the Diagram,
   * or if the model is not a {@link GraphLinksModel} with {@link GraphLinksModel#linkKeyProperty} set to a non-empty string.
   */
  public findLinkForKey(key: Key): Link {
    if (key === null || key === undefined) return null;
    const model = this.diagram.model as GraphLinksModel;
    const data = model.findLinkDataForKey(key);
    if (data !== null) {
      const link = this._dataLinkMap.getValue(data);
      return link;
    }
    return null;
  }

  /**
   * Handle changed events raised by a {@link GraphLinksModel}.
   * @param {ChangedEvent} e
   */
  public doModelChanged(e: ChangedEvent) {
    super.doModelChanged(e);

    if (!this.diagram) return;
    const dia = this.diagram;
    if (e.model !== dia.model) return;

    const chg = e.change;
    if (!dia.isUpdatingModel) return;
    dia.isUpdatingModel = false;  // now Model is updating Diagram, not vice-versa
    try {
      const mname = e.modelChange;
      if (mname !== '') {
        // first consider all model data changes (not simple model property changes)
        if (chg === ChangedEvent.Property) {
          if (mname === 'linkFromKey') {
            const data = e.object;
            const link = this.findLinkForData(data);
            if (link !== null) {
              const nid = e.newValue;
              const node = this.findNodeForKey(nid);
              link.fromNode = node;  // might be null
            }
          } else if (mname === 'linkToKey') {
            const data = e.object;
            const link = this.findLinkForData(data);
            if (link !== null) {
              const nid = e.newValue;
              const node = this.findNodeForKey(nid);
              link.toNode = node;  // might be null
            }
          } else if (mname === 'linkFromPortId') {
            const data = e.object;
            const link = this.findLinkForData(data);
            if (link !== null) {
              const pid = e.newValue;
              if (typeof pid === 'string') {
                link.fromPortId = pid;
              }
            }
          } else if (mname === 'linkToPortId') {
            const data = e.object;
            const link = this.findLinkForData(data);
            if (link !== null) {
              const pid = e.newValue;
              if (typeof pid === 'string') {
                link.toPortId = pid;
              }
            }
          } else if (mname === 'nodeGroupKey') {
            const data = e.object;
            const part = this.findPartForData(data);
            if (part !== null) {
              const gid = e.newValue;
              if (gid !== undefined) {
                const group = this.findNodeForKey(gid);
                if (group instanceof Group) {
                  part.containingGroup = group;
                } else {
                  part.containingGroup = null;
                }
              } else {
                part.containingGroup = null;
              }
            }
          } else if (mname === 'linkLabelKeys') {
            const data = e.object;
            const link = this.findLinkForData(data);
            if (link !== null) {
              const oldkeys = e.oldValue;
              const newkeys = e.newValue;
              if (Array.isArray(oldkeys)) {
                oldkeys.forEach((k) => {
                  const node = this.findNodeForKey(k);
                  if (node !== null) {
                    node.labeledLink = null;
                  }
                });
              }
              if (Array.isArray(newkeys)) {
                newkeys.forEach((k) => {
                  const node = this.findNodeForKey(k);
                  if (node !== null) {
                    node.labeledLink = link;
                  }
                });
              }
            }
          } else if (mname === 'linkCategory') {
            const data = e.object;
            const oldlink = this.findLinkForData(data);
            const newcat = e.newValue;
            if (oldlink !== null && typeof newcat === 'string') {
              oldlink.category = newcat;
            }
          } else if (mname === 'linkDataArray') {
            const olddata = e.oldValue as Array<Object>;
            this.removeModeledLinks(olddata);
            const newdata = e.newValue as Array<Object>;
            this.addModeledLinks(newdata);
          }
          dia.isModified = true;
          // end of ChangedEvent.Property
        } else if (chg === ChangedEvent.Insert) {  // named model-structure change
          const newdata = e.newValue;
          if (mname === 'linkDataArray' && typeof newdata === 'object' && newdata !== null) {  // model property
            this.addModeledLink(newdata);
          } else if (mname === 'linkLabelKeys' && e.model.isKeyType(newdata)) {  // model data change
            // presumably e.object is a link data object
            // and newdata is a key for a label node
            const linkdata = e.object;
            const link = this.findLinkForData(linkdata);
            const node = this.findNodeForKey(newdata);
            if (link !== null && node !== null) {
              node.labeledLink = link;
            }
          }
          dia.isModified = true;
          // end of ChangedEvent.Insert
        } else if (chg === ChangedEvent.Remove) {  // named model-structure change
          const olddata = e.oldValue;
          if (mname === 'linkDataArray' && typeof olddata === 'object' && olddata !== null) {  // model property
            this.removeModeledPart(olddata);
          } else if (mname === 'linkLabelKeys' && e.model.isKeyType(olddata)) {  // model data change
            // presumably olddata is a key for a label node
            const node = this.findNodeForKey(olddata);
            if (node !== null) {
              node.labeledLink = null;
            }
          }
          dia.isModified = true;
          // end of ChangedEvent.Remove
        }
      } else {  // mname === '', not a named model-structure change
        if (chg === ChangedEvent.Property) {
          const pname = e.propertyName;
          const data = e.object;
          const model = dia.model;
          if (data === model) {  // property change to model itself
            if (pname === 'linkFromKeyProperty' ||
              pname === 'linkToKeyProperty' ||
              pname === 'linkFromPortIdProperty' ||
              pname === 'linkToPortIdProperty' ||
              pname === 'linkLabelKeysProperty' ||
              pname === 'nodeIsGroupProperty' ||
              pname === 'nodeGroupKeyProperty' ||
              pname === 'linkCategoryProperty') {
              if (!dia.undoManager.isUndoingRedoing) this.rebuildParts();
            }
            // ignore the 'name', 'dataFormat' model properties;
            // 'nodeDataArray' and 'linkDataArray' are covered above
            // as named model-structure changes
          }
          dia.isModified = true;
        }
      }  // end of regular (not named model-structure) change
    } finally {
      dia.isUpdatingModel = true;
    }
  }

  /**
   * Remove any old data bound links and add all new data bound links.
   * This is called by {@link PartManager#updateAllRelationshipsFromData}.
   * @this {GraphLinksPartManager}
   */
  protected refreshDataBoundLinks(): void {
    const model = this.diagram.model as GraphLinksModel;
    // add all new data bound links, and remember all existing and new ones
    const alllinkdata = new Set<Object>();
    const larr = model.linkDataArray;
    larr.forEach((data) => {
      alllinkdata.add(data);
    });
    // remove all old links
    const oldlinksarr: Array<Object> = [];
    this.links.each((l) => {
      if (l.data !== null && !alllinkdata.contains(l.data)) oldlinksarr.push(l.data);
    });
    oldlinksarr.forEach((d) => { model.doRemoveLinkData(d, false); });
    // add all new links -- must do after removing old ones to avoid using new key due to apparent duplicate
    larr.forEach((data) => {
      const link = this.findLinkForData(data);
      if (link === null) {
        model.doAddLinkData(data, false);
      }
    });
  }

  /**
   * Handle any relationships that may have changed in a Part's data.
   *
   * This will update any group relationships or link relationships that may have changed.
   * @this {GraphLinksPartManager}
   * @param {Part} part
   */
  public updateRelationshipsFromData(part: Part) {
    const data = part.data;
    if (data === null) return;
    const diagram = part.diagram;
    if (diagram === null) return;
    const model = diagram.model as GraphLinksModel;

    if (part instanceof Link) {
      let key = model.getNodeKeyForLinkData(data, true);
      let node = diagram.findNodeForKey(key);
      part.fromNode = node;

      key = model.getNodeKeyForLinkData(data, false);
      node = diagram.findNodeForKey(key);
      part.toNode = node;

      // Handle changes to the Array holding the label node keys on link data
      // Optimization: don't do anything if both collections are empty,
      // or if both collections have only item in them and they are the same node,
      // or if there weren't any old keys removed or new keys added.
      const newkeys = model.getLabelKeysForLinkData(data);
      // don't do anything if there weren't or aren't any label nodes
      if (newkeys.length > 0 || part.labelNodes.count > 0) {
        // handle common case of only a single label nodeS
        if (newkeys.length === 1 && part.labelNodes.count === 1) {
          const newkey = newkeys[0];
          const oldlab = part.labelNodes.first();
          // if only one reference and if it didn't change, never mind....
          if (model.getKeyForNodeData(oldlab.data) === newkey) return;
        }
        const newset = new Set<Key>().addAll(newkeys);  // new keys in a Set, from the link data
        const oldset = new Set<Key>();  // old keys in a Set, as determined by part Link's data structures
        part.labelNodes.each(function (n) {
          if (n.data !== null) {
            const k = model.getKeyForNodeData(n.data);
            if (k !== undefined) oldset.add(k);
          }
        });
        const removeds = oldset.copy();  // which keys were removed from the oldset going to the newset
        removeds.removeAll(newset);
        const addeds = newset.copy();  // which keys were added to the oldset going to the newset
        addeds.removeAll(oldset);
        if (removeds.count > 0 || addeds.count > 0) {
          const link = part;
          removeds.each(function (k) {
            const n = diagram.findNodeForKey(k);
            if (n !== null && n.labeledLink === link) {
              n.labeledLink = null;
            }
          });
          addeds.each(function (k) {
            const n = diagram.findNodeForKey(k);
            if (n !== null && n.labeledLink !== link) {
              n.labeledLink = link;
            }
          });
        }
      }
    } else if (!(part instanceof Adornment)) { // regular Part/Node/Group
      const key = model.getGroupKeyForNodeData(data);
      const group = diagram.findPartForKey(key);  // might be null
      if (group === null || group instanceof Group) part.containingGroup = group as Group;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksPartManager}
   * @param {Object} data a non-link data.
   */
  protected resolveReferencesForData(data: Object) {
    const model = this.diagram.model as GraphLinksModel;
    const key = model.getKeyForNodeData(data);
    if (key !== undefined) {
      const refs = model.findDelayedReferences(key);
      const part = this.findPartForData(data);  // might be a Part or a Node or a Group
      if (refs !== null && part !== null) {
        // add references from other parts
        const refit = refs.iterator;
        while (refit.next()) {
          const ref = refit.value;
          if (model.containsNodeData(ref)) {  // the reference is from a node
            if (part instanceof Group && model.getGroupKeyForNodeData(ref) === key) {  // a member node?
              const mem = this.findPartForData(ref);
              if (mem !== null) {
                mem.containingGroup = part;
              }
            }
          } else {  // the reference is from a link
            const link = this.findLinkForData(ref);
            if (link !== null && part instanceof Node) {
              if (model.getNodeKeyForLinkData(ref, true) === key) {
                link.fromNode = part;
              }
              if (model.getNodeKeyForLinkData(ref, false) === key) {
                link.toNode = part;
              }
              // label nodes
              const lkeys = model.getLabelKeysForLinkData(ref);
              if (Array.isArray(lkeys)) {
                lkeys.some((k) => {
                  if (k === key) {
                    part.labeledLink = link;
                    return true;
                  }
                  return false;
                });
              }
            }
          }
        }
        // we have resolved all delayed/unresolved references
        model.forgetDelayedReference(key);
      }

      // this part might be part of a group
      const groupkey = model.getGroupKeyForNodeData(data);
      if (groupkey !== undefined) {
        const container = this.findNodeForKey(groupkey);
        if (container instanceof Group) {
          part.containingGroup = container;
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {PartManager}
   * @param {Part} part a Part/Node/Group, but not a Link or Adornment.
   */
  protected unresolveReferencesForPart(part: Part) {
    // no references in a plain Model, nor for a plain Part
    const model = this.diagram.model as GraphLinksModel;
    if (part instanceof Node) {
      const key = model.getKeyForNodeData(part.data);
      if (key !== undefined) {
        // remember references from all connected links
        const lit = part.linksConnected;
        while (lit.next()) {
          const l = lit.value;
          model.delayReference(key, l.data);
        }

        // for link label nodes, remember the reference from the labeled link
        if (part.isLinkLabel) {
          const llink = part.labeledLink;
          if (llink !== null) {
            model.delayReference(key, llink.data);
          }
        }

        // for groups, remember container references from member nodes
        if (part instanceof Group) {
          const mit = part.memberParts;
          while (mit.next()) {
            const m = mit.value;
            const memnodedata = m.data;
            if (model.containsNodeData(memnodedata)) {
              model.delayReference(key, memnodedata);
            }
          }
        }
      }
    }
  }

  /**
   * This method handles the data copying portion of copying a {@link Part}.
   *
   * This just calls the super method, then handles the data for any link being copied.
   * @this {GraphLinksPartManager}
   * @param {Part} part
   * @param {Diagram} destdiagram
   */
  protected copyPartData(part: Part, destdiagram: Diagram): Part | null {
    let copy = super.copyPartData(part, destdiagram);
    if (part instanceof Link) {
      const data = part.data;
      if (data !== null && destdiagram !== null) {  // bound Link
        const model = destdiagram.model as GraphLinksModel;
        const datacopy = model.copyLinkData(data);
        if (typeof datacopy === 'object' && datacopy !== null) {
          model.addLinkData(datacopy);
          copy = destdiagram.findLinkForData(datacopy);
        }
      } else if (copy !== null) {  // unbound Link
        const diagram = this.diagram;
        if (data !== null && diagram !== null && diagram.commandHandler !== null &&
            diagram.commandHandler.copiesClipboardData) {  // copy the bound data object too
          const model = diagram.model as GraphLinksModel;  // don't have a target diagram, so use diagram
          const datacopy = model.copyLinkData(data);
          if (typeof datacopy === 'object' && datacopy !== null) {
            copy.data = datacopy;
          }
        }
      }
    }
    return copy;
  }

  /**
   * Inserts a link between a from node and a to node.
   * Used by LinkingTool when creating links.
   * @this {GraphLinksPartManager}
   * @param {Node} fromnode
   * @param {GraphObject} fromport the from-port, or null to use the node itself
   * @param {Node} tonode
   * @param {GraphObject} toport the to-port, or null to use the node itself
   * @return {Link} the newly created Link, or null if it failed.
   */
  public insertLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): Link | null {
    const diagram = this.diagram;
    const model = diagram.model as GraphLinksModel;
    const linkingTool = diagram.toolManager.findTool('Linking') as any;
    let fromportid = '';
    if (fromnode !== null) {
      if (fromport === null) fromport = fromnode;
      fromportid = fromport.portId;
      if (fromportid === null) fromportid = '';
    }
    let toportid = '';
    if (tonode !== null) {
      if (toport === null) toport = tonode;
      toportid = toport.portId;
      if (toportid === null) toportid = '';
    }

    const arch = linkingTool.archetypeLinkData;
    if (arch instanceof Link) {
      arch.freezeBindings();
      const link = arch.copy();
      if (link !== null) {
        link.fromNode = fromnode;
        link.fromPortId = fromportid;
        link.toNode = tonode;
        link.toPortId = toportid;
        diagram.add(link);

        const archlab = linkingTool.archetypeLabelNodeData;
        if (archlab instanceof Node) {
          archlab.freezeBindings();
          const node = archlab.copy();
          if (node !== null) {
            node.labeledLink = link;
            diagram.add(node);
          }
        }

        return link;
      }
    } else if (arch !== null) {
      const data = model.copyLinkData(arch);
      if (typeof data === 'object' && data !== null) {
        if (fromnode !== null) {
          model.setNodeKeyForLinkData(data, model.getKeyForNodeData(fromnode.data), true);
        }
        model.setPortIdForLinkData(data, fromportid, true);
        if (tonode !== null) {
          model.setNodeKeyForLinkData(data, model.getKeyForNodeData(tonode.data), false);
        }
        model.setPortIdForLinkData(data, toportid, false);
        model.addLinkData(data);

        // maybe add a label node data for this new link data
        const archlab = linkingTool.archetypeLabelNodeData;
        if (archlab !== null && !(archlab instanceof Node)) {
          const labdata = model.copyNodeData(archlab);
          if (typeof labdata === 'object' && labdata !== null) {
            model.addNodeData(labdata);
            const labkey = model.getKeyForNodeData(labdata);
            if (labkey !== undefined) {
              model.addLabelKeyForLinkData(data, labkey);
            }
          }
        }

        const link = diagram.findLinkForData(data);
        return link;
      }
    }
    return null;
  }
}
GraphLinksPartManager['className'] = 'GraphLinksPartManager';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
TreePartManager
*/

// import { ChangedEvent } from './changedEvent';
// import { GraphObject } from './graphObject';
// import { LinkingTool } from './linkingTools';
// import { PartManager } from './partManager';
// import { Link, Node, Part } from './parts';
// import { TreeModel } from './treeModel';

/**
 * @hidden @internal
 * The TreePartManager handles interaction between a {@link Diagram} and its {@link TreeModel}.
 * It is responsible for creating and removing {@link Node}s, {@link Link}s, and {@link Part}s for the data in the model.
 *
 * This class extends the normal {@link PartManager} with more functionality related to trees.
 * @extends PartManager
 * @unrestricted
 * @since 2.0
 * @category Model
 */
export class TreePartManager extends PartManager {
  /** @hidden @internal */ public static className = 'TreePartManager';
  /** @hidden @internal */ private _reparentLink: Link = null;

  /**
   * Create a {@link Link} from a parent to a child,
   * assuming both {@link Node}s are data-bound in a {@link TreeModel}.
   * If {@link Diagram#isTreePathToChildren} is false, create the link in opposite direction.
   * @this {TreePartManager}
   * @param {Node} parent
   * @param {Node} child
   * @return {Link}
   */
  private addParentLink(parent: Node, child: Node): Link {
    if (parent === null || child === null) return null;
    const tool = this.diagram.toolManager.findTool('Linking') as LinkingTool;
    let from = parent;
    let to = child;
    if (this.diagram.isTreePathToChildren) {
      const lit = to.linksConnected;
      while (lit.next()) {
        const l = lit.value;
        if (l.toNode === to) return null;
      }
      if (tool !== null && tool._MakesDirectedCycleFast(from, to, null, true)) return null;
    } else {
      from = child;
      to = parent;
      const lit = from.linksConnected;
      while (lit.next()) {
        const l2 = lit.value;
        if (l2.fromNode === from) return null;
      }
      if (tool !== null && tool._MakesDirectedCycleFast(from, to, null, true)) return null;
    }

    const cat = this.getLinkCategoryForData(child.data);
    const template = this.findLinkTemplateForCategory(cat);
    if (template !== null) {
      template.freezeBindings();
      const link = template.copy();
      if (link !== null) {
        const oldskips = this.diagram.skipsModelSourceBindings;
        this.diagram.skipsModelSourceBindings = true;
        link._category = cat;  // don't call property setter
        link._data = child.data;  // don't call property setter
        link.fromNode = from;
        link.toNode = to;
        this.diagram.add(link);
        link._data = null;  // don't call property setter, yet
        link.data = child.data;  // no link data, so bind to child node data
        this.diagram.skipsModelSourceBindings = oldskips;
        return link;
      }
    }
    return null;
  }

  /**
   * Find the category for the parent link of a given child node data, a string naming the link template
   * that the {@link Diagram} should use to represent the link.
   *
   * This calls {@link TreeModel#getParentLinkCategoryForNodeData}.
   * @this {TreePartManager}
   * @param {Object} data a JavaScript object represented by a node data
   * @return {string}
   */
  public getLinkCategoryForData(data: Object): string {
    const model = this.diagram.model as TreeModel;
    return model.getParentLinkCategoryForNodeData(data);
  }

  /**
   * Change the category for the parent link of a given child node data, a string naming the link template
   * that the {@link Diagram} should use to represent the link
   *
   * This calls {@link TreeModel#setParentLinkCategoryForNodeData}.
   * @this {TreePartManager}
   * @param {Object} data a JavaScript object represented by a node data
   * @param {string} value the category, must not be null
   */
  public setLinkCategoryForData(data: Object, value: string): void {
    const model = this.diagram.model as TreeModel;
    model.setParentLinkCategoryForNodeData(data, value);
  }

  /**
   * Change the node key that the given link references as the source of the link.
   *
   * This calls {@link TreeModel#setParentKeyForNodeData}.
   * @this {TreePartManager}
   * @param {Link} link a link for which to change the fromNode
   * @param {Node} value the Node whose key will be referenced in the data
   * @param {Node=} old optional, the Node whose key will no longer be reference in the data
   */
  public setFromNodeForLink(link: Link, value: Node, old?: Node): void {
    const model = this.diagram.model as TreeModel;
    if (old === undefined) old = null;
    // data is now the child node data, not a link data
    // if VAL is null or if VAL.data is null, the parent key is set to undefined
    const ndata = (value !== null ? value.data : null);
    if (this.diagram.isTreePathToChildren) {
      model.setParentKeyForNodeData(link.data, model.getKeyForNodeData(ndata));
    } else {
      const oldReparent = this._reparentLink;
      this._reparentLink = link;
      if (old !== null) {
        model.setParentKeyForNodeData(old.data, undefined);
      }
      const pdata = (link.toNode !== null ? link.toNode.data : null);
      model.setParentKeyForNodeData(ndata, model.getKeyForNodeData(pdata));
      this._reparentLink = oldReparent;
    }
  }

  /**
   * Change the node key that the given link references as the destination of the link.
   *
   * This calls {@link TreeModel#setParentKeyForNodeData}.
   * @this {TreePartManager}
   * @param {Link} link a link for which to change the toNode
   * @param {Node} value the Node whose key will be referenced in the data
   * @param {Node=} old optional, the Node whose key will no longer be reference in the data
   */
  public setToNodeForLink(link: Link, value: Node, old?: Node): void {
    const model = this.diagram.model as TreeModel;
    if (old === undefined) old = null;
    // data is now the child node data, not a link data
    // if VAL is null or if VAL.data is null, the parent key is set to undefined
    const ndata = (value !== null ? value.data : null);
    if (this.diagram.isTreePathToChildren) {
      const oldReparent = this._reparentLink;
      this._reparentLink = link;
      if (old !== null) {
        model.setParentKeyForNodeData(old.data, undefined);
      }
      const pdata = (link.fromNode !== null ? link.fromNode.data : null);
      model.setParentKeyForNodeData(ndata, model.getKeyForNodeData(pdata));
      this._reparentLink = oldReparent;
    } else {
      model.setParentKeyForNodeData(link.data, model.getKeyForNodeData(ndata));
    }
  }

  /**
   * Remove model data associated with the given Link by clearing out parent relationships.
   *
   * This calls {@link TreeModel#setParentKeyForNodeData}.
   * @this {TreePartManager}
   * @param {Link} link
   */
  public removeDataForLink(link: Link): void {
    const model = this.diagram.model as TreeModel;
    model.setParentKeyForNodeData(link.data, undefined);
  }

  /**
   * Handle changed events raised by a {@link TreeModel}.
   * @param {ChangedEvent} e
   */
  public doModelChanged(e: ChangedEvent) {
    super.doModelChanged(e);

    if (!this.diagram) return;
    const dia = this.diagram;
    if (e.model !== dia.model) return;

    const chg = e.change;
    if (!dia.isUpdatingModel) return;
    dia.isUpdatingModel = false;  // now Model is updating Diagram, not vice-versa
    try {
      const mname = e.modelChange;
      if (mname !== '') {
        // first consider all model data changes (not simple model property changes)
        if (chg === ChangedEvent.Property) {
          if (mname === 'nodeParentKey') {
            const childdata = e.object;  // the child node data
            const parentkey = e.newValue;
            const parentnode = this.findNodeForKey(parentkey);
            const childnode = this.findNodeForData(childdata);
            if (this._reparentLink !== null) {
              if (parentnode !== null) {
                this._reparentLink.data = childdata;
                this._reparentLink.category = this.getLinkCategoryForData(childdata);
              }
            } else if (childnode !== null) {
                const parentlink = childnode.findTreeParentLink();
                // found an existing parent link
                if (parentlink !== null) {
                  // if no new parent node, delete the link
                  if (parentnode === null) {
                    dia.remove(parentlink);
                  } else {
                    // otherwise reconnect the existing parent link
                    if (dia.isTreePathToChildren) {
                      parentlink.fromNode = parentnode;
                    } else {
                      parentlink.toNode = parentnode;
                    }
                  }
                } else {  // not found -- create new link
                  this.addParentLink(parentnode, childnode);
                }
              }
          } else if (mname === 'parentLinkCategory') {
            const childdata = e.object;  // the child node data
            const childnode = this.findNodeForData(childdata);
            const newcat = e.newValue;
            if (childnode !== null && typeof newcat === 'string') {
              const parentlink = childnode.findTreeParentLink();
              if (parentlink !== null) {
                parentlink.category = newcat;
              }
            }
          }
          dia.isModified = true;
          // end of ChangedEvent.Property
        }
      } else {  // mname === '', not a named model-structure change
        if (chg === ChangedEvent.Property) {
          const pname = e.propertyName;
          const data = e.object;
          const model = dia.model;
          if (data === model) {  // property change to model itself
            if (pname === 'nodeParentKeyProperty') {
              if (!dia.undoManager.isUndoingRedoing) this.rebuildParts();
            }
          }
          dia.isModified = true;
        }
      }  // end of regular (not named model-structure) change
    } finally {
      dia.isUpdatingModel = true;
    }
  }

  /**
   * Handle any relationships that may have changed in a Node's data.
   * @this {TreePartManager}
   * @param {Part} part
   */
  public updateRelationshipsFromData(part: Part) {
    const data = part.data;
    if (data === null) return;
    const diagram = part.diagram;
    if (diagram === null) return;
    const model = diagram.model as TreeModel;

    if (part instanceof Node) {
      const key = model.getParentKeyForNodeData(data);
      const parent = diagram.findNodeForKey(key);
      // does parent key correspond to tree parent node?
      const currentparent = part.findTreeParentNode();
      if (parent !== currentparent) {
        const currentparentlink = part.findTreeParentLink();
        if (parent !== null) {
          if (currentparentlink !== null) {
            if (diagram.isTreePathToChildren) {
              currentparentlink.fromNode = parent;
            } else {
              currentparentlink.toNode = parent;
            }
          } else {
            this.addParentLink(parent, part);
          }
        } else {  // parent should be null
          if (currentparentlink !== null) {
            diagram.doRemove(currentparentlink, false);
          }
        }
      }
    }
  }

  /**
   * Re-evaluate a given binding on some Part associated with a data object
   * and potentially parent links associated with that data.
   * @this {TreePartManager}
   * @param {Object} data the data object for which the binding will be re-evaluated
   * @param {string|function(Object,*):*} pname the property name of the data to re-evaluate
   */
  protected updateDataBindings(data: Object, pname: string | ((a: Object, b: any) => any)): void {
    super.updateDataBindings(data, pname);
    if (typeof pname === 'string') {
      const part = this.findPartForData(data);
      if (part !== null) {  // is node data or link data
        const link = this.findLinkForData(data);  // handle parent links when TreeModel
        if (link !== null) {
          link.updateTargetBindings(pname);
        }
      }
    }
  }

  /**
   * Add any necessary links that may be referenced by a node data
   * @param {Object} data a node data object
   */
  protected resolveReferencesForData(data: Object) {
    const model = this.diagram.model as TreeModel;
    const key = model.getKeyForNodeData(data);
    if (key !== undefined) {
      const refs = model.findDelayedReferences(key);
      const part = this.findPartForData(data);  // might be a Part or a Node or a Group
      if (refs !== null && part !== null) {
        // add references from other parts
        const refit = refs.iterator;
        while (refit.next()) {
          const ref = refit.value;
          if (model.containsNodeData(ref) && part instanceof Node) {
            if (model.getParentKeyForNodeData(ref) === key) {
              const child = this.findNodeForData(ref);
              this.addParentLink(part, child);
            }
          }
        }
        // we have resolved all delayed/unresolved references
        model.forgetDelayedReference(key);
      }

      const parentkey = model.getParentKeyForNodeData(data);
      if (parentkey !== undefined && part instanceof Node) {
        const parent = this.findNodeForKey(parentkey);
        this.addParentLink(parent, part);
      }
    }
  }

  /**
   * Remove any links associated with a node that is being removed
   * @param {Part} part the part for which references should be removed
   */
  protected unresolveReferencesForPart(part: Part) {
    const model = this.diagram.model as TreeModel;
    if (part instanceof Node) {
      const key = model.getKeyForNodeData(part.data);
      // handle the implicitly defined parent Link that may exist for each Node
      const link = this.findLinkForData(part.data);
      if (link !== null) {
        // Don't call this.remove(link), because that will have side-effects on the model/data
        // Instead, do what removeBoundPart does for regular Parts:
        // make sure it's no longer selected
        link.isSelected = false;
        link.isHighlighted = false;
        // remove from layer only; it calls Diagram.internalRemove to finish cleaning up here
        const layer = link.layer;
        if (layer !== null) {
          const index = layer.internalRemove(-1, link, false);
          if (index >= 0) this.diagram.raiseChangedEvent(ChangedEvent.Remove, 'parts', layer, link, null, index, null);
          const func = link.layerChanged;
          if (func !== null) func(link, layer, null);
        }
      }

      // remember references from all tree children
      const tochildren = this.diagram.isTreePathToChildren;
      const it = part.linksConnected;
      while (it.next()) {
        const cl = it.value;
        const childnode = (tochildren ? cl.toNode : cl.fromNode);
        const childdata = childnode.data;
        if (model.containsNodeData(childdata)) {
          model.delayReference(key, childdata);
        }
      }
    }
  }

  /**
   * Inserts a link between a from node and a to node.
   * Used by LinkingTool when creating links.
   * @this {TreePartManager}
   * @param {Node} fromnode
   * @param {GraphObject} fromport the from-port, or null to use the node itself
   * @param {Node} tonode
   * @param {GraphObject} toport the to-port, or null to use the node itself
   * @return {Link} the newly created Link, or null if it failed.
   */
  public insertLink(fromnode: Node | null, fromport: GraphObject | null, tonode: Node | null, toport: GraphObject | null): Link | null {
    const model = this.diagram.model as TreeModel;
    let parentnode = fromnode;
    let childnode = tonode;
    if (!this.diagram.isTreePathToChildren) {
      parentnode = tonode;
      childnode = fromnode;
    }
    if (parentnode !== null && childnode !== null) {
      const parentdata = parentnode.data;
      const childdata = childnode.data;
      model.setParentKeyForNodeData(childdata, model.getKeyForNodeData(parentdata));
      return childnode.findTreeParentLink();
    }
    return null;
  }
}
TreePartManager['className'] = 'TreePartManager';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
Model
Binding
*/

// import { Brush } from './brush';
// import { ChangedEvent } from './changedEvent';
// import { List, Map, Set } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { EnumValue } from './enumValue';
// import { Geometry } from './geometry';
// import { GraphObject } from './graphObject';
// import { Margin } from './margin';
// import { Panel } from './panel';
// import { Point } from './point';
// import { Rect } from './rect';
// import { ObjectData, root } from './root';
// import { Size } from './size';
// import { Spot } from './spot';
// import { StringBuilder } from './stringBuilder';
// import { Transaction, UndoManager } from './undoManager';
// import { Util } from './util';

/**
 * The Key type is the same as `string | number | undefined`.
 * @category Type
 */
export type Key = string | number | undefined;

/** @hidden @internal */
export interface ITreeModel extends ILinkModel {
  nodeParentKeyProperty: string | ((a: ObjectData, b?: Key) => Key);
  copiesParentKeyOfNodeData: boolean;
  getParentKeyForNodeData(nodedata: ObjectData): Key;
  setParentKeyForNodeData(nodedata: ObjectData, key: Key): void;
  parentLinkCategoryProperty: string | ((a: ObjectData, b?: string) => string);
  getParentLinkCategoryForNodeData(childdata: ObjectData): string;
  setParentLinkCategoryForNodeData(childdata: ObjectData, cat: string): void;
  copyNodeData(nodedata: ObjectData): ObjectData;
  setDataProperty(data: ObjectData, propname: string, val: any): void;
}

/** @hidden @internal */
export interface IGraphLinksModel extends ILinkDataModel, IGroupModel {
  archetypeNodeData: ObjectData | null;
  copyLinkData(linkdata: ObjectData): ObjectData;
  copyNodeData(nodedata: ObjectData): ObjectData;
  setDataProperty(data: ObjectData, propname: string, val: any): void;
  copiesGroupKeyOfNodeData: boolean;
  nodeIsGroupProperty: string | ((a: ObjectData, b?: boolean) => boolean);
}

/** @hidden @internal */
export interface IGraphModel extends ILinkModel, IGroupModel {
  relatedKeysProperty: string | ((a: ObjectData, b?: Array<Key>) => Array<Key>);
  getRelatedKeysForNodeData(nodedata: ObjectData): Array<Key>;
  setRelatedKeysForNodeData(nodedata: ObjectData, arr: Array<Key>): void;
  addRelatedKeyForNodeData(nodedata: ObjectData, key: Key): void;
  removeRelatedKeyForNodeData(nodedata: ObjectData, key: Key): void;
  isRelatedKeysPathTo: boolean;
  copyNodeData(nodedata: ObjectData): ObjectData;
  setDataProperty(data: ObjectData, propname: string, val: any): void;
  copiesGroupKeyOfNodeData: boolean;
  nodeIsGroupProperty: string | ((a: ObjectData, b?: boolean) => boolean);
}

/**
 * @hidden @internal
 * This interface defines the properties/methods that must be implemented by models
 * that will include a link data array. Only models specifying a link data array may contain
 * port information and link labels.
 * @category Model
 * @since 2.0
 */
export interface ILinkDataModel extends ILinkModel {
  linkFromKeyProperty: string | ((a: ObjectData, b?: Key) => Key);
  getFromKeyForLinkData(linkdata: ObjectData): Key;
  setFromKeyForLinkData(linkdata: ObjectData, key: Key): void;
  linkToKeyProperty: string | ((a: ObjectData, b?: Key) => Key);
  getToKeyForLinkData(linkdata: ObjectData): Key;
  setToKeyForLinkData(linkdata: ObjectData, key: Key): void;
  linkFromPortIdProperty: string | ((a: ObjectData, b?: string) => string);
  getFromPortIdForLinkData(linkdata: ObjectData): string;
  setFromPortIdForLinkData(linkdata: ObjectData, portname: string): void;
  linkToPortIdProperty: string | ((a: ObjectData, b?: string) => string);
  getToPortIdForLinkData(linkdata: ObjectData): string;
  setToPortIdForLinkData(linkdata: ObjectData, portname: string): void;
  linkLabelKeysProperty: string | ((a: ObjectData, b?: Array<Key>) => Array<Key>);
  getLabelKeysForLinkData(linkdata: ObjectData): Array<Key>;
  setLabelKeysForLinkData(linkdata: ObjectData, arr: Array<Key>): void;
  addLabelKeyForLinkData(linkdata: ObjectData, key: Key): void;
  removeLabelKeyForLinkData(linkdata: ObjectData, key: Key): void;
  linkDataArray: Array<ObjectData>;
  linkKeyProperty: string | ((a: ObjectData, b?: Key) => Key);
  getKeyForLinkData(linkdata: ObjectData): Key;
  setKeyForLinkData(linkdata: ObjectData, key: Key): void;
  findLinkDataForKey(key: Key): ObjectData | null;
  makeLinkDataKeyUnique(linkdata: ObjectData): void;
  containsLinkData(linkdata: ObjectData): boolean;
  addLinkData(linkdata: ObjectData): void;
  addLinkDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void;
  removeLinkData(linkdata: ObjectData): void;
  removeLinkDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void;
  linkCategoryProperty: string | ((a: ObjectData, b?: string) => string);
  getCategoryForLinkData(linkdata: ObjectData): string;
  setCategoryForLinkData(linkdata: ObjectData, cat: string): void;
}

/**
 * @hidden @internal
 * This interface defines the properties/methods that must be implemented by models
 * that will include data representing {@link Link}s.
 * @category Model
 * @since 2.0
 */
export interface ILinkModel {
  linkCategoryProperty: string | ((a: ObjectData, b?: string) => string);
  getLinkCategoryForData(data: ObjectData): string;
  setLinkCategoryForData(data: ObjectData, cat: string): void;
}

/**
 * @hidden @internal
 * This interface defines the properties/methods that must be implemented by models
 * that will include data for {@link Group}s.
 * @category Model
 * @since 2.0
 */
export interface IGroupModel {
  isGroupForData(data: ObjectData): boolean;
  getGroupForData(data: ObjectData): Key;
  setGroupForData(data: ObjectData, key: Key): void;
}

/**
 * Interface for the output object of {@link Model#toIncrementalData}.
 * @category Model
 * @since 2.1
 */
export interface IncrementalData {
  /**
   * Object containing the modified {@link Model#modelData}.
   */
  modelData?: ObjectData;
  /**
   * Array of node keys added.
   */
  insertedNodeKeys?: Array<Key>;
  /**
   * Array of node data objects modified.
   */
  modifiedNodeData?: Array<ObjectData>;
  /**
   * Array of node keys deleted.
   */
  removedNodeKeys?: Array<Key>;
  /**
   * Array of link keys added.
   */
  insertedLinkKeys?: Array<Key>;
  /**
   * Array of link data objects modified.
   */
  modifiedLinkData?: Array<ObjectData>;
  /**
   * Array of link keys deleted.
   */
  removedLinkKeys?: Array<Key>;
}

/**
 * Models hold the essential data of a diagram, describing the basic entities and their properties and relationships
 * without specifying the appearance and behavior of the Nodes and Links and Groups that represent them visually.
 * Models tend to hold only relatively simple data, making them easy to persist by serialization as JSON or XML formatted text.
 *
 * Models hold simple data objects, not {@link Part}s such as {@link Node}s or {@link Link}s.
 * Node data is normally represented in a {@link Diagram} by instances of {@link Node},
 * but they could be represented by simple {@link Part}s or by {@link Group}s.
 * A Diagram constructs Parts for its {@link Diagram#model}'s data by copying templates.
 * Templates are {@link Panel}s of {@link GraphObject}s that get some property values from the model data,
 * accessible via the {@link Panel#data} property, using data {@link Binding}.
 * See <a href="../../intro/usingModels.html">Using Models</a> and <a href="../../intro/dataBinding.html">Data Binding</a> for an introduction.
 *
 * This Model class only supports holding an array of node data
 * and interpreting properties on that data to be able to refer to them
 * using unique key values.
 * To support simple tree-structured graphs, use a {@link TreeModel}, which inherits from this class.
 * To support links and grouping, use a {@link GraphLinksModel}.
 *
 * Each node data object is assumed to have a unique key value.
 * The {@link #nodeKeyProperty} property names the property on the node data whose value
 * is the unique key for that node data object.
 * The default value for this property is "key".
 * You should not have a TwoWay data binding on the node key property,
 * because that might cause the property value to be set to a duplicate key value.
 *
 * The key values must be either strings or numbers or undefined.
 * If the key is undefined, or if there are duplicate key values,
 * the model will automatically try to assign a new unique key value.
 * Caution: if your keys are numbers, do not try to use string representations of those numbers as keys.
 * Conversely, if your keys are strings that happen to have number syntax, do not try to use those number values.
 * Sometimes JavaScript will automatically convert from string to number or vice-versa, but sometimes it won't.
 *
 * For example, one can define a graph consisting of just two nodes:
 * ```js
 *  model.nodeDataArray = [
 *    { key: "Alpha" },
 *    { key: "Beta" }
 *  ];
 * ```
 *
 * This model cannot detect the modification of the {@link #nodeDataArray} array
 * or the modification of any node data object.
 * If you want to add or remove node data from the {@link #nodeDataArray},
 * call the {@link #addNodeData} or {@link #removeNodeData} methods.
 *
 * If you want to modify a node data object, it depends on whether the property you want to change
 * is a structural property that the model needs to know about, or whether it is a property
 * that is only used for data binding or other application-specific purposes.
 *
 * For the former case, call the appropriate method, such as
 * {@link #setKeyForNodeData}, {@link #setCategoryForNodeData},
 * {@link GraphLinksModel#setToKeyForLinkData}, or {@link GraphLinksModel#setGroupKeyForNodeData}.
 * These methods have names that start with "set", "add", "insert", or "remove".
 *
 * For the latter case, when setting an application-specific property, typically for data binding,
 * and to support undo/redo, call {@link #setDataProperty}.
 *
 * The {@link #copyNodeData} method can be called to make a shallow copy of a node data object.
 * However, if some of those property values are Arrays that want not to be shared but to be copied,
 * you can set {@link #copiesArrays} to true.  This is typically very useful when dealing
 * with data bound item arrays.
 * Furthermore if the items in those copied Arrays are in fact Objects that need to be copied,
 * you can also set {@link #copiesArrayObjects} to true, causing a copied Array to refer to
 * newly shallow-copied objects of the original array.
 *
 * Each model raises {@link ChangedEvent}s that you can follow by registering a listener via {@link #addChangedListener}.
 * Read more at the Introduction page: <a href="../../intro/changedEvents.html">Changed Events</a>.
 *
 * Each model comes with its own {@link UndoManager} that is initially not enabled.
 * You will need to set {@link UndoManager#isEnabled} to true in order for the
 * UndoManager to record model changes and for your users to perform undo and redo.
 *
 * You can temporarily turn off the recording of changes by setting {@link #skipsUndoManager} to true.
 * A number of places within the system do that routinely in order to avoid recording temporary changes,
 * so be sure to remember the original value beforehand and restore it afterwards.
 * Note that in a {@link ChangedEvent} listener you may want to ignore events that happen when {@link #skipsUndoManager} is true.
 *
 * One normally saves a diagram by just saving its model.
 * If you can use JSON-formatted text, this is easy to do -- just call {@link #toJson}
 * to get the string representation of the model, and save that string.
 * Load the diagram by replacing the {@link Diagram#model} with one created by calling
 * the static function {@link Model.fromJson}:
 * ```js
 *   myDiagram.model = go.Model.fromJson(loadedString);
 * ```
 * Note that JSON and other textual data formats cannot faithfully store all JavaScript functions.
 * {@link #toJson} and {@link Model.fromJson} do not try to save and load functional property values.
 * You should arrange that all such functions, including event handlers, are established by your app.
 * {@link #toJson} and {@link Model.fromJson} also cannot handle circular references; any sharing of
 * references will be lost too.
 * They also skip properties that are not enumerable, those whose names start with an underscore, and those whose values are undefined.
 *
 * Note that models also do not store the templates used by diagrams,
 * nor any transient or temporary parts such as {@link Adornment}s, nor any tools,
 * nor any {@link UndoManager} state, nor any event listeners.
 * These objects and all other properties of diagrams must be established by your app.
 *
 * You can add any number of properties to the {@link #modelData} object, which is serialized and deserialized
 * into JSON just like any other model data for nodes or links.
 * However {@link #modelData} is associated with the model as a whole and does not depend on the existence
 * of any node data or link data.
 *
 * It is also easy to save the changes that were recorded in the most recent transaction.
 * Call {@link #toIncrementalJson} to generate a JSON-format string that holds the current state of modified data
 * plus the keys of inserted or removed data.
 * That method requires as an argument a {@link ChangedEvent} that represents a transaction that completed or
 * an undo or a redo that just finished.
 *
 * It is also possible to use such "incremental" JSON to modify an existing model.
 * Call {@link #applyIncrementalJson}, giving it a string generated by {@link #toIncrementalJson},
 * to modify this model by making all of the changes recorded in the JSON text.
 * Note how this method is a regular instance method, whereas {@link Model.fromJson} is a static function.
 * @unrestricted
 * @category Model
 */
export class Model {
  /** @hidden @internal */ public static className = 'Model';
  /** @hidden @internal */ private _name: string;
  /** @hidden @internal */ private _dataFormat: string;
  /** @hidden @internal */ private _isReadOnly: boolean;
  /** @hidden @internal */ private _modelData: ObjectData; // since it allows arbitrary values
  /** @hidden @internal */ private _nodeDataArray: Array<ObjectData>;
  /** @hidden @internal */ private _makeUniqueKeyFunction: ((model: Model, data: ObjectData) => Key) | null;
  /** @hidden @internal */ private _copyNodeDataFunction: ((data: ObjectData, model: Model) => ObjectData) | null;
  /** @hidden @internal */ private _copiesArrays: boolean;
  /** @hidden @internal */ private _copiesArrayObjects: boolean;
  /** @hidden @internal */ private _copiesKey: boolean;
  /** @hidden @internal */ private _afterCopyFunction: ((map: Map<ObjectData, ObjectData>, destModel: Model, srcModel: Model) => void) | null;
  /** @hidden @internal */ private _nodeCategoryProperty: string | ((data: ObjectData, newCat: string) => string);
  /** @hidden @internal */ private _delayedNodeReferences: Map<(string | number), Set<ObjectData>>;
  /** @hidden @internal */ private _listeners: List<((a: ChangedEvent) => void)>;
  /** @hidden @internal */ private _skipsUndoManager: boolean;
  /** @hidden @internal */ private _undoManager: UndoManager;
  /** @hidden @internal */ protected _mapKeyNodeData: Map<(string | number), ObjectData>;
  /** @hidden @internal */ protected _nodeKeyProperty: string | ((a: ObjectData, b: Key) => Key);

  /**
   * You probably don't want to call this constructor, because this class
   * does not support links (relationships between nodes) or groups (nodes and links and subgraphs as nodes):
   * instead, create instances of a subclass such as {@link GraphLinksModel} or {@link TreeModel}.
   * @param {Array.<Object>=} nodedataarray an optional Array containing JavaScript objects to be represented by {@link Part}s.
   */
  constructor(nodedataarray?: Array<ObjectData>) {
    if (Debug && arguments.length > 1) {
      Util.throwError('Model constructor can only take one optional argument, the Array of node data.');
    }
    Set.uniqueHash(this);
    this._name = '';
    this._dataFormat = '';
    this._isReadOnly = false;
    this._modelData = {};

    this._nodeDataArray = [];
    this._mapKeyNodeData = new Map<(string | number), ObjectData>();
    this._nodeKeyProperty = 'key';
    this._makeUniqueKeyFunction = null;
    this._copyNodeDataFunction = null;
    this._copiesArrays = false;
    this._copiesArrayObjects = false;
    this._copiesKey = true;
    this._afterCopyFunction = null;
    this._nodeCategoryProperty = 'category';
    this._delayedNodeReferences = new Map<(string | number), Set<ObjectData>>();

    this._listeners = new List<((a: ChangedEvent) => void)>();

    this._skipsUndoManager = false;
    this._undoManager = null;
    // setup the UndoManager via property setter
    this.undoManager = new UndoManager();

    if (nodedataarray !== undefined) this.nodeDataArray = nodedataarray;
  }

  /**
   * Copies properties from this model to the given model, which must be of the same class.
   * This is called by {@link #copy}.
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {Model}
   * @param {Model} copy
   * @since 1.6
   */
  protected cloneProtected(copy: this): void {
    copy._name = this._name;
    copy._dataFormat = this._dataFormat;
    copy._isReadOnly = this._isReadOnly;
    // copy._modelData = this._modelData;  not copying data
    // copy._nodeDataArray = this._nodeDataArray;  not copying data
    // this._mapKeyNodeData should not be copied
    copy._nodeKeyProperty = this._nodeKeyProperty;
    copy._makeUniqueKeyFunction = this._makeUniqueKeyFunction;
    copy._copyNodeDataFunction = this._copyNodeDataFunction;
    copy._copiesArrays = this._copiesArrays;
    copy._copiesArrayObjects = this._copiesArrayObjects;
    copy._copiesKey = this._copiesKey;
    copy._afterCopyFunction = this._afterCopyFunction;
    copy._nodeCategoryProperty = this._nodeCategoryProperty;
    // this._delayedNodeReferences should not be copied
    // this._listeners should not be copied
    // this._skipsUndoManager should not be copied
    // this._undoManager should not be shared
  }

  /**
   * Creates a shallow copy of this Model and returns it.
   * <em>The data are not copied:</em> {@link #nodeDataArray}, {@link #modelData},
   * {@link GraphLinksModel#linkDataArray}, {@link GraphLinksModel#archetypeNodeData} are left empty.
   * Nor are any Changed listeners or the UndoManager copied.
   * @expose
   * @this {Model}
   * @return {Model} an empty copy of the model with the same properties, other than data
   * @since 1.6
   */
  public copy(): Model {
    const copy = new (this.constructor as any)();
    this.cloneProtected(copy);
    return copy;
  }

  /**
   * Clear out all references to any model data.
   * This also clears out the {@link UndoManager}, so this operation is not undoable.
   * This method is called by {@link Diagram#clear}; it does not notify any {@link Diagram}s or other listeners.
   * This method does not unregister any Changed event listeners.
   *
   * Instead of calling this method, you may prefer to set {@link #nodeDataArray} to an empty JavaScript Array.
   * If this model is a {@link GraphLinksModel}, you would also want to set {@link GraphLinksModel#linkDataArray} to a separate empty JavaScript Array.
   * @export
   * @this {Model}
   */
  public clear(): void {
    this._nodeDataArray = [];
    this._mapKeyNodeData.clear();
    this._delayedNodeReferences.clear();
    this.undoManager.clear();
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {number=} details Optional: If greater than zero, this includes descriptions of each node data.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    if (details > 1) {
      return this.toJson();
    } else {
      let s = (this.name !== '' ? this.name : '') + ' Model';
      if (details > 0) {
        s += '\n node data:';
        const arr = this.nodeDataArray;
        const len = Util.arrayLength(arr);
        for (let i = 0; i < len; i++) {
          const item = Util.arrayElt(arr, i);
          s += ' ' + this.getKeyForNodeData(item) + ':' + Util.toString(item);
        }
      }
      return s;
    }
  }

  /**
   * Produce an object representing the changes in the most recent {@link Transaction}.
   * The structure of the object follows the same format as the JSON output from {@link #toIncrementalJson}.
   *
   * Note that these incremental changes include the results of undo and redo operations.
   *
   * For {@link GraphLinksModel}s, this method requires that {@link GraphLinksModel#linkKeyProperty} is not an empty string.
   *
   * Any node or link data objects contained in the "modified..." properties will be deep copies of the data in the model.
   * The objects will contain proper copies of certain GoJS classes, and some common built-in objects such as Dates and RegExps.
   * Other classes will just be copied as plain Javascript objects, so it is best to avoid using special classes in one's data.
   *
   * This method is most commonly used when GoJS must communicate with some external data source
   * and maintain integrity between the two while avoiding serialization/deserialization.
   *
   * ```js
   *   myDiagram.addModelChangedListener(function(e) {
   *     if (e.isTransactionFinished) {
   *       var dataChanges = e.model.toIncrementalData(e);
   *       ... update React state/save to database ...
   *     }
   *   });
   * ```
   * @export
   * @this {Model}
   * @param {ChangedEvent} e a Transaction ChangedEvent for which {@link ChangedEvent#isTransactionFinished} is true
   * @return {IncrementalData} returns either null if no changes occured, or an object containing incremental model changes for the given Transaction
   * @see {@link #toIncrementalJson}
   * @since 2.1
   */
  public toIncrementalData(e: ChangedEvent): IncrementalData {
    Util.checkClass(e, ChangedEvent, Model, 'toIncrementalData:e');
    if (e.change !== ChangedEvent.Transaction) Util.throwError('Model.toIncrementalData argument is not a Transaction ChangedEvent:' + e.toString());
    const txn = e.object;
    if (!e.isTransactionFinished || !(txn instanceof Transaction)) return null;

    this.mapDataToPartData(txn);
    const obj = this.writeIncrementalModelData(txn, e.propertyName === 'FinishedUndo');
    this._NestedDataPartDataMap = null;

    return obj;
  }

  /**
   * @hidden @internal
   * For a given transaction, build an object with all data changes.
   * @this {Model}
   * @param {Transaction} txn
   * @param {boolean} undo
   * @return {IncrementalData}
   */
  protected writeIncrementalModelData(txn: Transaction, undo: boolean): IncrementalData {
    const model = this;
    let modelDataChanged = false;
    const newnodes = new Set<ObjectData>();
    const modnodes = new Set<ObjectData>();
    const oldnodes = new Set<ObjectData>();
    const nestedDataPartMap = this._NestedDataPartDataMap;
    txn.changes.each(function (e) {
      if (e.model !== model) return;
      if (e.modelChange === 'nodeDataArray') {
        if (e.change === ChangedEvent.Insert) {
          newnodes.add(e.newValue);
        } else if (e.change === ChangedEvent.Remove) {
          oldnodes.add(e.oldValue);
        }
      } else if (model.containsNodeData(e.object)) {
        modnodes.add(e.object);
      } else if (e.change === ChangedEvent.Property && (model.modelData === e.object || e.propertyName === 'modelData')) {
        modelDataChanged = true;
      } else if (e.object !== null) {  // some data object that isn't node data
        if (nestedDataPartMap && nestedDataPartMap.contains(e.object)) {
          const partData = nestedDataPartMap.get(e.object);
          partData.each(function(d) {
            if (model.containsNodeData(d)) {
              modnodes.add(d);
            }
          });
        } else {
          const partData = model.findDataInNodeDataArray(e.object);
          partData.each(function(d) {
            modnodes.add(d);
          });
        }
      }
    });
    const newnodekeys = new Set<Key>();  // just keys, not data objects
    newnodes.each(function (nd) {
      newnodekeys.add(model.getKeyForNodeData(nd));
      // make sure any new node data gets written out in full by including it in the array of modified data
      if (!undo) modnodes.add(nd);
    });
    const oldnodekeys = new Set<Key>();  // just keys, not data objects
    oldnodes.each(function (nd) {
      oldnodekeys.add(model.getKeyForNodeData(nd));
      // undoing a deletion is the same as adding, so make sure it's written out in full
      if (undo) modnodes.add(nd);
    });
    const modnodearr = model.cloneDeep(modnodes.toArray());

    let obj: IncrementalData = null;
    // use quoted string properties so they don't get minified
    if (modelDataChanged) {
      if (obj === null) obj = {};
      obj['modelData'] = this.cloneDeep(this.modelData);
    }
    if (newnodekeys.count > 0) {
      if (obj === null) obj = {};
      if (undo) {
        obj['removedNodeKeys'] = newnodekeys.toArray();
      } else {
        obj['insertedNodeKeys'] = newnodekeys.toArray();
      }
    }
    if (modnodearr.length > 0) {
      if (obj === null) obj = {};
      obj['modifiedNodeData'] = modnodearr;
    }
    if (oldnodekeys.count > 0) {
      if (obj === null) obj = {};
      if (undo) {
        obj['insertedNodeKeys'] = oldnodekeys.toArray();
      } else {
        obj['removedNodeKeys'] = oldnodekeys.toArray();
      }
    }

    return obj;
  }

  /**
   * Deeply copy an object/array and return the new object.
   * This is typically called on a nodeDataArray or linkDataArray or data objects within them.
   *
   * By default, this method will make deep clones of arrays and Javascript objects and maintain any shared and/or cyclic references.
   * It will properly copy any Date or RegExp object, and will call a `copy` function on any object where one exists.
   * It also handles certain GoJS classes: Point, Size, Rect, Margin, Spot, List, Set, and Map.
   *
   * This method may be overridden.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * Only override this method when the default behavior doesn't suit the data.
   * When cloning objects, we suggest skipping the `__gohashid` property.
   * @expose
   * @this {Model}
   * @param {T} obj
   * @return {T}
   */
  public cloneDeep<T>(obj: T): T {
    return this.cloneHelper(obj);
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {*} obj
   * @param {Map<*, *>=} copies
   * @return {*}
   */
  private cloneHelper(obj: any, copies?: Map<any, any>): any {
    // just return if not an object
    if (!Util.isObject(obj)) return obj;

    // if it's already been copied, maintain reference
    if (!copies) copies = new Map<any, any>();
    const copied = copies.get(obj);
    if (copied) return copied;

    let result;
    if (Array.isArray(obj)) {
      result = [] as Array<any>;
      copies.set(obj, result);
      for (let i = 0; i < obj.length; i++) {
        const val = obj[i];
        result.push(this.cloneHelper(val, copies));
      }
    } else {
      // special case for some GoJS classes
      if (obj instanceof Point || obj instanceof Size ||
          obj instanceof Rect || obj instanceof Margin ||
          obj instanceof Spot) {
        result = obj.copy();
        copies.set(obj, result);
      } else if (obj instanceof Brush || obj instanceof Geometry || obj instanceof EnumValue) { // Immutable when in use
        return obj;
      } else if (obj instanceof List) {
        result = new List<any>().addAll(this.cloneHelper(obj.toArray(), copies));
        copies.set(obj, result);
      } else if (obj instanceof Set) {
        result = new Set<any>().addAll(this.cloneHelper(obj.toArray(), copies));
        copies.set(obj, result);
      } else if (obj instanceof Map) {
        result = new Map<any, any>().addAll(this.cloneHelper(obj.toArray(), copies));
        copies.set(obj, result);
      } else if (obj instanceof Date) {  // handle some common built in object types
        result = new Date(obj.getTime());
        copies.set(obj, result);
      } else if (obj instanceof RegExp) {
        result = new RegExp(obj);
        copies.set(obj, result);
        result.lastIndex = obj.lastIndex;
      } else if (typeof obj.copy === 'function') {  // obj has a provided copy function
        result = obj.copy();
        copies.set(obj, result);
      } else {  // otherwise, treat it as a basic object
        result = {} as ObjectData;
        copies.set(obj, result);
        for (const key in obj) {
          if (key === '__gohashid') continue;
          const value = obj[key];
          result[key] = this.cloneHelper(value, copies);
        }
      }
    }
    return result;
  }

  /**
   * @hidden @internal
   * Write out the persistent properties of this model into a string using JSON syntax property pairs.
   * @this {Model}
   * @return {string}
   */
  protected writeJsonProtected(): string {
    let s = '';
    if (this.name !== '') s += ',\n  \"name\": ' + this.quote(this.name);
    if (this.dataFormat !== '') s += ',\n  \"dataFormat\": ' + this.quote(this.dataFormat);
    if (this.isReadOnly) s += ',\n  \"isReadOnly\": ' + this.isReadOnly;
    if (this.nodeKeyProperty !== 'key' && typeof this.nodeKeyProperty === 'string') s += ',\n  \"nodeKeyProperty\": ' + this.quote(this.nodeKeyProperty);
    if (this.copiesArrays) s += ',\n  \"copiesArrays\": true';
    if (this.copiesArrayObjects) s += ',\n  \"copiesArrayObjects\": true';
    if (!this.copiesKey) s += ',\n  \"copiesKey\": false';
    if (this.nodeCategoryProperty !== 'category' && typeof this.nodeCategoryProperty === 'string') s += ',\n  \"nodeCategoryProperty\": ' + this.quote(this.nodeCategoryProperty);
    return s;
  }

  /**
   * @hidden @internal
   * Find the persistent model properties of the given Object and assign them to this model.
   * @this {Model}
   * @param {Object} obj
   */
  protected readJsonProtected(obj: ObjectData): void {
    if (obj['name']) this.name = obj['name'];
    if (obj['dataFormat']) this.dataFormat = obj['dataFormat'];
    if (obj['isReadOnly']) this.isReadOnly = true;
    if (obj['nodeKeyProperty']) this.nodeKeyProperty = obj['nodeKeyProperty'];
    if (obj['copiesArrays']) this.copiesArrays = true;
    if (obj['copiesArrayObjects']) this.copiesArrayObjects = true;
    if (obj['copiesKey'] === false) this.copiesKey = false;
    if (obj['nodeCategoryProperty']) this.nodeCategoryProperty = obj['nodeCategoryProperty'];
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {string}
   */
  private writeSharedModelData(): string {
    const mdata = this.modelData;
    return ',\n  \"modelData\": ' + this.writeJsonObject(mdata);
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} obj
   */
  private readSharedModelData(obj: ObjectData): void {
    const mdata = obj['modelData'];
    if (Util.isObject(mdata)) {
      this.replaceJsonObjects(mdata);
      this.modelData = mdata;
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {string}
   */
  protected writeJsonData(): string {
    const mdata = this.modelData;
    let anyprops = false;
    for (const p in mdata) {
      if (this.skipsJsonObjectProperty(mdata, p, mdata[p])) continue;
      anyprops = true;
      break;
    }
    let s = '';
    if (anyprops) {
      s = this.writeSharedModelData();
    }
    return s + ',\n  \"nodeDataArray\": ' + this.writeJsonArray(this.nodeDataArray, true);
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} obj
   */
  protected readJsonData(obj: ObjectData): void {
    this.readSharedModelData(obj);
    const arr = obj['nodeDataArray'];
    if (Util.isArray(arr)) {
      this.replaceJsonObjects(arr);
      this.nodeDataArray = arr;
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} a
   * @param {Object} b
   * @param {Map.<*,*>} seen -- for detecting cycles or shared objects in structures
   * @return {boolean}
   */
  protected isEqualObject(a: ObjectData, b: ObjectData, seen: Map<ObjectData, ObjectData>): boolean {
    if (a === b) return true;
    if (typeof a !== typeof b) return false;
    if (typeof a === 'function' || typeof b === 'function') return false;
    if (Array.isArray(a) && Array.isArray(b)) {
      if (seen.getValue(a) === b) return true;
      seen.add(a, b);
      if (a.length !== b.length) return false;
      for (let i = 0; i < a.length; i++) {
        if (!this.isEqualObject(a[i], b[i], seen)) return false;
      }
      return true;
    } else if (Util.isObject(a) && Util.isObject(b)) {
      if (seen.getValue(a) === b) return true;
      seen.add(a, b);
      for (const p in a) {
        const av = a[p];
        if (this.skipsJsonObjectProperty(a, p, av)) continue;
        const bv = b[p];
        if (bv === undefined) return false;  // in A but not in B
        if (!this.isEqualObject(av, bv, seen)) return false;
      }
      for (const q in b) {
        const bv = b[q];
        if (this.skipsJsonObjectProperty(b, q, bv)) continue;
        const av = a[q];
        if (av === undefined) return false;  // in B but not in A
        if (!this.isEqualObject(av, bv, seen)) return false;
      }
      return true;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model} as Object
   * @param {Object} m2 second model
   * @param {string} propname
   */
  protected checkSameProperties(this: ObjectData, m2: ObjectData, propname: string): void {
    if (this[propname] !== m2[propname]) {
      Util.throwError(`Model.computeJsonDifference: Model.${propname} is not the same in both models: "${this[propname]}" and "${m2[propname]}"`);
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Model} newmodel
   * @return {string}
   */
  protected writeJsonDifference(newmodel: Model): string {
    this.checkSameProperties(newmodel, 'nodeKeyProperty');

    const newnodekeys = new Set<Key>();
    const modnodedatas = new Set<ObjectData>();
    const oldnodekeys = new Set<Key>().addAll(this._mapKeyNodeData.iteratorKeys); // JStoTS: Iterator not simple Iterable
    const seen = new Map<ObjectData, ObjectData>();
    const arr = newmodel.nodeDataArray;
    const len = Util.arrayLength(arr);
    for (let i = 0; i < len; i++) {
      const newobj = Util.arrayElt(arr, i);
      const key = newmodel.getKeyForNodeData(newobj);  // find the node's key, if any
      if (key !== undefined) {
        oldnodekeys.remove(key);  // can't be an existing key that is deleted
        const oldobj = this.findNodeDataForKey(key);
        if (oldobj === null) {    // not present in this model,
          newnodekeys.add(key);   // so it is being inserted
          modnodedatas.add(newobj);  // and remember details in "modified nodes" collection
        } else {  // existing object in this model
          if (!this.isEqualObject(oldobj, newobj, seen)) {
            modnodedatas.add(newobj);  // only add to "modified nodes" collection if it has changed properties
          }
        }
      } else {  // no key in new data object
        this.makeNodeDataKeyUnique(newobj);  // give it a unique key
        const newkey = this.getKeyForNodeData(newobj);
        newnodekeys.add(newkey);   // record it as a new key
        modnodedatas.add(newobj);  // and remember details for new node
      }
    }
    let s = '';
    if (!this.isEqualObject(this.modelData, newmodel.modelData, seen)) s += this.writeSharedModelData();
    if (newnodekeys.count > 0) s += this._InsertedNodesString + this.writeJsonArray(newnodekeys.toArray(), true);
    if (modnodedatas.count > 0) s += this._ModifiedNodesString + this.writeJsonArray(modnodedatas.toArray(), true);
    if (oldnodekeys.count > 0) s += this._RemovedNodesString + this.writeJsonArray(oldnodekeys.toArray(), true);
    return s;
  }

  // Produce a JSON-format string recording the differences between this Model and a given Model of the same class.
  // This produces the same kind of results as {@link #toIncrementalJson} does.
  // Those results could be applied to this model now to make it equivalent to the given model.
  //
  // Note that many of the "...Property" properties of {@link Model}, {@link GraphLinksModel}, or {@link TreeModel}
  // need to be the same in both this model and the given model, so that the references between nodes will be
  // consistent.
  //
  // The same restrictions on data property names and data property values applies to this method as it does to {@link #toJson}.
  /**
   * @hidden
   * Undocumented
   * @export
   * @this {Model}
   * @param {Model} newmodel
   * @param {string=} classname for the written model, defaults to the name of the class of the model
   * @return {string}
   */
  public computeJsonDifference(newmodel: Model, classname?: string): string {
    Util.checkClass(newmodel, Model, Model, 'computeJsonDifference:newmodel');
    if (classname === undefined) {
      classname = this.getClassName(this);
    }
    return '{ \"class\": ' + this.quote(classname) + ', \"incremental\": 1' +
      this.writeJsonProtected() +
      this.writeJsonDifference(newmodel) +
      '}';
  }

  /** @hidden @internal */
  protected _InsertedNodesString = ',\n  \"insertedNodeKeys\": ';
  /** @hidden @internal */
  protected _ModifiedNodesString = ',\n  \"modifiedNodeData\": ';
  /** @hidden @internal */
  protected _RemovedNodesString = ',\n  \"removedNodeKeys\": ';

  /** @hidden @internal */
  protected _NestedDataPartDataMap: Map<ObjectData, Set<ObjectData>> = null;

  /**
   * @hidden @internal
   * For a given transaction, use diagram changes to map nested data to top level data.
   * We do this by finding the panel for a given change, looking at its data, and getting the associated part's data.
   * @this {Model}
   * @param {Transaction} txn
   */
  private mapDataToPartData(txn: Transaction): void {
    let map = this._NestedDataPartDataMap;
    const addToMap = function(data: ObjectData, panel: Panel) {
      const topdata = panel.part.data;
      if (data === topdata) return;
      let set = map.get(data);
      if (set === null) {
        set = new Set<ObjectData>();
        set.add(topdata);
        map.add(data, set);
      } else {
        set.add(topdata);
      }
    };

    if (map === null) {
      map = new Map<ObjectData, Set<ObjectData>>();
      txn.changes.each(function (e) {
        if (e.diagram !== null) {
          const change = e.change;
          if (change === ChangedEvent.Property) {
            const panel = e.object.panel as Panel;
            if (panel) {
              const data = panel.data;
              if (data) {
                addToMap(data, panel);
              }
            }
          } else if (change === ChangedEvent.Insert || change === ChangedEvent.Remove) { // could be item array change
            const panel = e.object as Panel;
            const data = panel.itemArray;
            if (data) {
              addToMap(data, panel);
            }
          }
        }
      });
      this._NestedDataPartDataMap = map;
    }
  }

  /**
   * @hidden @internal
   * Search for a nested data item within the node data array.
   * This is used when we can't map nested data to node data via diagram panel changes.
   * @this {Model}
   * @param {ObjectData} data
   * @return {Set.<ObjectData>}
   */
  private findDataInNodeDataArray(data: ObjectData): Set<ObjectData> {
    const set = new Set<ObjectData>();
    for (let i = 0; i < this.nodeDataArray.length; i ++) {
      const nd = this.nodeDataArray[i];
      this.recurseData(data, nd, nd, set);
    }
    return set;
  }

  /**
   * @hidden @internal
   * Recurse into a data object in search of some nested data.
   * @param {ObjectData} data the data we're looking for
   * @param {ObjectData} parent the object we're searching in
   * @param {ObjectData} toplevel the top-level data we started on
   * @param {Set.<ObjectData>} set the Set of top level parts this nested data belongs to, which we modify if we find it
   * @return {boolean} whether the data was found
   */
  protected recurseData(data: ObjectData, parent: ObjectData, toplevel: ObjectData, set: Set<ObjectData>): boolean {
    if (Array.isArray(parent)) {
      for (let i = 0; i < parent.length; i++) {
        const item = parent[i];
        if (item === data) {
          set.add(toplevel);
          return true;
        }
        if (this.recurseData(data, item, toplevel, set)) return true;
      }
    } else if (Util.isObject(parent) && Object.getPrototypeOf(parent) === Object.prototype) {
      for (const p in parent) {
        const val = parent[p];
        if (val === data) {
          set.add(toplevel);
          return true;
        }
        if (this.recurseData(data, val, toplevel, set)) return true;
      }
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Transaction} txn
   * @param {boolean} undo
   * @return {string}
   */
  protected writeIncrementalJsonData(txn: Transaction, undo: boolean): string {
    const model = this;
    let modelDataChanged = false;
    const newnodes = new Set<ObjectData>();
    const modnodes = new Set<ObjectData>();
    const oldnodes = new Set<ObjectData>();
    const nestedDataPartMap = this._NestedDataPartDataMap;
    txn.changes.each(function (e) {
      if (e.model !== model) return;
      if (e.modelChange === 'nodeDataArray') {
        if (e.change === ChangedEvent.Insert) {
          newnodes.add(e.newValue);
        } else if (e.change === ChangedEvent.Remove) {
          oldnodes.add(e.oldValue);
        }
      } else if (model.containsNodeData(e.object)) {
        modnodes.add(e.object);
      } else if (e.change === ChangedEvent.Property && (model.modelData === e.object || e.propertyName === 'modelData')) {
        modelDataChanged = true;
      } else if (e.object !== null) {  // some data object that isn't node data
        if (nestedDataPartMap && nestedDataPartMap.contains(e.object)) {
          const partData = nestedDataPartMap.get(e.object);
          partData.each(function(d) {
            if (model.containsNodeData(d)) {
              modnodes.add(d);
            }
          });
        } else {
          const partData = model.findDataInNodeDataArray(e.object);
          partData.each(function(d) {
            modnodes.add(d);
          });
        }
      }
    });
    const newnodekeys = new Set<Key>();  // just keys, not data objects
    newnodes.each(function (nd) {
      newnodekeys.add(model.getKeyForNodeData(nd));
      // make sure any new node data gets written out in full by including it in the array of modified data
      if (!undo) modnodes.add(nd);
    });
    const oldnodekeys = new Set<Key>();  // just keys, not data objects
    oldnodes.each(function (nd) {
      oldnodekeys.add(model.getKeyForNodeData(nd));
      // undoing a deletion is the same as adding, so make sure it's written out in full
      if (undo) modnodes.add(nd);
    });
    let s = '';
    if (modelDataChanged) s += this.writeSharedModelData();
    if (newnodekeys.count > 0) s += (undo ? this._RemovedNodesString : this._InsertedNodesString) + this.writeJsonArray(newnodekeys.toArray(), true);
    if (modnodes.count > 0) s += this._ModifiedNodesString + this.writeJsonArray(modnodes.toArray(), true);
    if (oldnodekeys.count > 0) s += (undo ? this._InsertedNodesString : this._RemovedNodesString) + this.writeJsonArray(oldnodekeys.toArray(), true);
    return s;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} obj
   */
  protected readIncrementalJsonData(obj: ObjectData): void {
    // ensure Model properties are the same
    if (obj['name'] !== undefined && obj['name'] !== this.name ||
        obj['dataFormat'] !== undefined && obj['dataFormat'] !== this.dataFormat ||
        obj['isReadOnly'] !== undefined && obj['isReadOnly'] !== this.isReadOnly ||
        obj['nodeKeyProperty'] !== undefined && obj['nodeKeyProperty'] !== this.nodeKeyProperty ||
        obj['copiesArrays'] !== undefined && obj['copiesArrays'] !== this.copiesArrays ||
        obj['copiesArrayObjects'] !== undefined && obj['copiesArrayObjects'] !== this.copiesArrayObjects ||
        obj['copiesKey'] !== undefined && obj['copiesKey'] !== this.copiesKey ||
        obj['nodeCategoryProperty'] !== undefined && obj['nodeCategoryProperty'] !== this.nodeCategoryProperty ) {
      Util.throwError('applyIncrementalJson cannot change Model properties');
    }
    this.readSharedModelData(obj);
    const newkeys = obj['insertedNodeKeys'];
    const moddata = obj['modifiedNodeData'];
    const modmap = new Map();
    if (Array.isArray(moddata)) {
      for (let j = 0; j < moddata.length; j++) {
        const d = moddata[j];
        const k2 = this.getKeyForNodeData(d);
        if (k2 !== undefined && k2 !== null) modmap.set(k2, d);
      }
    }
    if (Array.isArray(newkeys)) {
      const len = newkeys.length;
      for (let i = 0; i < len; i++) {
        const k = newkeys[i];
        let nd = this.findNodeDataForKey(k);
        if (nd === null) {
          // if an existing node data object is given, use it,
          // to avoid having problems with trying to declare that a copied object is a group
          const newdata = modmap.get(k);
          nd = newdata ? newdata : this.copyNodeData({});
          this.setKeyForNodeData(nd, k);
          this.addNodeData(nd);
        }
      }
    }
    if (Array.isArray(moddata)) {
      const len = moddata.length;
      for (let i = 0; i < len; i++) {
        const d = moddata[i];
        const k2 = this.getKeyForNodeData(d);
        const nd = this.findNodeDataForKey(k2);
        if (nd !== null) {
          for (const p in d) {
            if (p === '__gohashid' || p === this.nodeKeyProperty || ((this.isGraphLinksModel() || this.isGraphModel()) && p === this.nodeIsGroupProperty)) continue;
            this.setDataProperty(nd, p, d[p]);
          }
        }
      }
    }
    const oldkeys = obj['removedNodeKeys'];
    if (Array.isArray(oldkeys)) {
      const len = oldkeys.length;
      for (let i = 0; i < len; i++) {
        const k = oldkeys[i];
        const nd = this.findNodeDataForKey(k);
        if (nd !== null) this.removeNodeData(nd);
      }
    }
  }

  /**
   * Produce a JSON-format string representing the changes in the most recent {@link Transaction}.
   * This writes out JSON for a model, but recording only changes in the given Transaction,
   * with the addition of the "incremental" property to mark it as different from a complete model.
   * Instead of the "nodeDataArray" property (and "linkDataArray" property for {@link GraphLinksModel}s),
   * this will have "inserted...", "modified...", and "removed..." properties that are non-empty Arrays.
   *
   * The "modifiedNodeData" Array holds JavaScript objects.
   * The "insertedNodeKeys" and "removedNodeKeys" Arrays hold keys (numbers or strings) of data,
   * not whole objects, that have been added and/or deleted.
   * The "modelData" property holds the {@link Model#modelData} object, if it was modified.
   *
   * Note that it is entirely plausible for the same object be in or referenced by all three Arrays,
   * because a single {@link Transaction} can include adding a node, modifying it, and removing it.
   *
   * The purpose of this method is to make it easier to send incremental changes to the server/database,
   * instead of sending the whole model.
   * Whereas it has always been easy to perform "batch" updates or "file saves":
   * ```js
   *   myDiagram.addModelChangedListener(function(e) {
   *     if (e.isTransactionFinished) {
   *       var json = e.model.toJson();
   *       // save the whole model upon each transaction completion or undo/redo
   *       ... send to server/database ...
   *     }
   *   });
   * ```
   * You can now easily send "incremental" updates:
   * ```js
   *   myDiagram.addModelChangedListener(function(e) {
   *     if (e.isTransactionFinished) {
   *       var json = e.model.toIncrementalJson(e);
   *       // record each Transaction as a JSON-format string
   *       ... send to server/database ...
   *     }
   *   });
   * ```
   * Note that these incremental changes include the results of undo and redo operations.
   * Also, when you might call {@link #applyIncrementalJson}, you will need to disable your Changed listener,
   * so that it does not send spurious changes to your database during the process of apply incremental changes from the database.
   *
   * For {@link GraphLinksModel}s, this method requires that {@link GraphLinksModel#linkKeyProperty} is not an empty string.
   * The incremental JSON for GraphLinksModels will include "modifiedLinkData", "insertedLinkKeys", and "removedLinkKeys"
   * properties that are non-empty Arrays.
   *
   * The same restrictions on data property names and data property values applies to this method as it does to {@link #toJson}.
   * @export
   * @this {Model}
   * @param {ChangedEvent} e a Transaction ChangedEvent for which {@link ChangedEvent#isTransactionFinished} is true
   * @param {string=} classname for the written model, defaults to the name of the class of the model
   * @return {string}
   * @since 1.6
   */
  public toIncrementalJson(e: ChangedEvent, classname?: string): string {
    Util.checkClass(e, ChangedEvent, Model, 'toIncrementalJson:e');
    if (e.change !== ChangedEvent.Transaction) Util.throwError('Model.toIncrementalJson argument is not a Transaction ChangedEvent:' + e.toString());
    const txn = e.object;
    if (!e.isTransactionFinished || !(txn instanceof Transaction)) return '{ \"incremental\": 0 }';
    if (classname === undefined) {
      classname = this.getClassName(this);
    }
    this.mapDataToPartData(txn);
    const incrementalJson = this.writeIncrementalJsonData(txn, e.propertyName === 'FinishedUndo');
    this._NestedDataPartDataMap = null;
    if (classname === '2') {  // temporary hack for new functionality???  Better to replace classname arg with options object
      return '{ \"incremental\": 2' +
        incrementalJson +
        '}';
    }
    return '{ \"class\": ' + this.quote(classname) + ', \"incremental\": 1' +
      this.writeJsonProtected() +
      incrementalJson +
      '}';
  }

  /**
   * @hidden @internal
   * @export
   * @this {Model}
   */
  public toIncrementalJSON(e: ChangedEvent, classname?: string): string {
    return this.toIncrementalJson(e, classname);
  }

  /**
   * Generate a string representation of the persistent data in this model, in JSON format,
   * that can be read in later with a call to {@link Model.fromJson}.
   *
   * Object properties that are not enumerable or whose names start with "_" are not written out.
   *
   * Functions are not able to be written in JSON format, so any properties that have function values
   * will not be saved in the JSON string.
   *
   * There must not be any circular references within the model data.
   * Any sharing of object references will be lost in the written JSON.
   *
   * Most object classes cannot be serialized into JSON without special knowledge and processing at both ends.
   * The {@link #toJson} and {@link Model.fromJson} methods automatically do such processing for numbers that are NaN
   * and for objects that are of class {@link Point}, {@link Size}, {@link Rect}, {@link Margin}, {@link Spot},
   * {@link Brush} (but not for brush patterns), and for {@link Geometry}.
   * However, we recommend that you use Binding converters (static functions named "parse" and "stringify")
   * to represent Points, Sizes, Rects, Margins, Spots, and Geometries as string values in your data, rather than as Objects.
   * This makes the JSON text smaller and simpler and easier to read.
   *
   * As a special case when serializing an object, if the property is named "points" and the
   * property value is a {@link List} of {@link Point}s, it will write an Array of numbers.
   *
   * Note that this is a method on the {@link Model} class.
   * It cannot render unmodeled Parts such as the background grid or any Parts that you have added directly to a Diagram.
   *
   * Typical usage:
   * ```js
   * var modelAsText = myDiagram.model.toJson();
   * // now save this text string by sending it to your database
   * ```
   * @export
   * @this {Model}
   * @param {string=} classname The optional name of the model class to use in the output;
   *     for the standard models, this is their class name prefixed with "go.".
   * @return {string} a String in JSON format containing all of the persistent properties of the model.
   */
  public toJson(classname?: string): string {
    if (classname === undefined) {
      classname = this.getClassName(this);
    }
    return '{ \"class\": ' + this.quote(classname) +
      this.writeJsonProtected() +
      this.writeJsonData() +
      '}';
  }

  /**
   * @hidden @internal
   * @export
   * @this {Model}
   */
  public toJSON(classname?: string): string {
    return this.toJson(classname);
  }

  /**
   * Modify this model by applying the changes given in an "incremental" model change in JSON format
   * generated by {@link #toIncrementalJson}.
   * The expected properties of the argument are described at {@link #toIncrementalJson}.
   * Incremental changes must be applied in the same order that the changes occurred in the original model.
   *
   * This requires the "incremental" property to be present and to be a number, as specified by {@link #toIncrementalJson}.
   * All of the top-level properties in the JSON, such as {@link #nodeKeyProperty}, must be the same as for this model.
   * Note that if the model is a {@link GraphLinksModel}, you will have to have set {@link GraphLinksModel#linkKeyProperty}
   * to the name of a property, the same both in the {@link Diagram#model} as well as in the data that you pass to this method.
   *
   * This conducts a transaction.
   * @export
   * @this {Model}
   * @param {string|ObjectData} s a String in JSON format containing modifications to be performed to the model,
   *    or a JavaScript Object parsed from such a string
   * @since 1.6
   */
  public applyIncrementalJson(s: string | ObjectData): void {
    let modeldata = null;
    if (typeof s === 'string') {
      try {
        modeldata = root.JSON.parse(s);
      } catch (ex) {
        if (Debug) Util.trace('JSON.parse error: ' + ex.toString());
      }
    } else if (typeof s === 'object') {
      modeldata = s;
    } else {
      Util.throwError('Unable to modify a Model from: ' + s);
    }

    const version = modeldata['incremental'];
    if (typeof version !== 'number') Util.throwError('Unable to apply non-incremental changes to Model: ' + s);
    if (version === 0) return;  // no model changes

    this.startTransaction('applyIncrementalJson');
    // don't modify Model properties (i.e. readJsonProtected)
    // ?? ought to check that Model properties are the same!
    this.readIncrementalJsonData(modeldata);
    this.commitTransaction('applyIncrementalJson');
  }

  /**
   * @hidden @internal
   * @export
   * @this {Model}
   */
  public applyIncrementalJSON(s: string | ObjectData): void {
    return this.applyIncrementalJson(s);
  }

  /**
   * This static function parses a string in JSON format that was written by {@link Model#toJson},
   * and then constructs, initializes, and returns a model with that information.
   *
   * Note that properties with values that are functions are not written out by {@link #toJson},
   * so reading in such a model will require constructing such a model, initializing its functional property values,
   * and explicitly passing it in as the second argument.
   *
   * As a special case when deserializing an object, if the property is named "points" and the
   * property value is an Array with an even number of numbers, it will substitute a {@link List} of {@link Point}s.
   *
   * Typical usage:
   * ```js
   * var modelAsText = ...;  // fetch the model in textual format from a database
   * myDiagram.model = go.Model.fromJson(modelAsText);
   * ```
   * @this {Model}
   * @param {string|ObjectData} s a String in JSON format containing all of the persistent properties of the model, or an Object already read from JSON text.
   * @param {Model=} model an optional model to be modified; if not supplied, it constructs and returns a new model whose name is specified by the "class" property.
   * @return {Model} the supplied or created model loaded with data from the given string.
   */
  public static fromJson(s: string | ObjectData, model?: Model): Model {
    if (model === undefined) model = null;
    if (model !== null) Util.checkClass(model, Model, Model, 'fromJson:model');

    let modeldata = null;
    if (typeof s === 'string') {
      try {
        modeldata = root.JSON.parse(s);
      } catch (ex) {
        if (Debug) Util.trace('JSON.parse error: ' + ex.toString());
      }
    } else if (typeof s === 'object') {
      modeldata = s;
    } else {
      Util.throwError('Unable to construct a Model from: ' + s);
    }

    // var existingmodel = model;
    if (model === null) {
      const m = Model.makeJsonClass(modeldata) as Model;
      if (m === null || m instanceof Model) {
        model = m;
      } else {
        Util.throwError('Unable to construct a Model of declared class: ' + modeldata['class']);
      }
    }
    if (model === null) {
      model = Model.constructGraphLinksModel();
    }
    model.readJsonProtected(modeldata);
    model.readJsonData(modeldata);
    // if (existingmodel !== null) {
    //  existingmodel.raiseModelChangedEvent('ModelDisplaced', ChangedEvent.Transaction, '', existingmodel, null, null);
    // }
    return model;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {string|ObjectData} s a String in JSON format containing all of the persistent properties of the model, or an Object already read from JSON text.
   * @param {Model=} model an optional model to be modified; if not supplied, it constructs and returns a new model whose name is specified by the "class" property.
   * @return {Model} the supplied or created model loaded with data from the given string.
   */
  public static fromJSON(s: string | ObjectData, model?: Model): Model {
    return this.fromJson(s, model);
  }

  /**
   * @hidden @internal
   * Undocumented
   * @expose
   * @return {Model} the supplied or created model loaded with data from the given string.
   */
  public static constructGraphLinksModel(): Model {
    if (Debug) Util.throwError('Unable to construct a Model. Provided JSON requires GraphLinksModel, which is not loaded.');
    return new Model();
  }

  /**
   * @hidden @internal
   * Create an object whose class name is given by the "class" property of the given data object.
   * @param {Object} data
   * @return {Object}
   */
  private static makeJsonClass(data: ObjectData): ObjectData {
    let result: ObjectData = null;
    let classname = data['class'];
    if (typeof classname === 'string') {
      try {
        let classfunc = null;
        if (classname.indexOf('go.') === 0) {  // if prefixed with "go.", only look in the GoJS namespace
          classname = classname.substr(3);
          classfunc = Model.getClassForString(classname);
        } else {  // otherwise try both "go" and "window" (root)
          classfunc = Model.getClassForString(classname);
          if (classfunc === undefined) classfunc = root[classname];
        }
        if (typeof classfunc === 'function') result = new classfunc();
      } catch (ex) {
        // empty
      }
    }
    return result;
  }

  /**
   * @hidden @internal
   * Walk a JavaScript tree of objects, finding simple JavaScript Objects
   * that should be replaced with instances of special classes.
   * @export
   * @this {Model}
   * @param {*} x
   * @return {*}
   */
  protected replaceJsonObjects(x: any): any {
    if (Util.isArray(x)) {
      // recurse through array items
      const len = Util.arrayLength(x);
      for (let i = 0; i < len; i++) {
        const d = Util.arrayElt(x, i);
        if (Util.isObject(d)) {
          Util.arraySetElt(x, i, this.replaceJsonObjects(d));
        }
      }
    } else if (Util.isObject(x)) {
      // recurse through object properties
      for (const p in x) {
        const d = x[p];
        if (Util.isObject(d)) {
          const replaced = this.replaceJsonObjects(d);
          x[p] = replaced;
          // handle special case of List of Points as an Array of x,y pairs
          if (p === 'points' && Array.isArray(replaced)) {  // not Util.isArray
            // make sure there are an even number of values and that they are all numbers
            let allpoints = replaced.length % 2 === 0;
            for (let j = 0; j < replaced.length; j++) {
              if (typeof replaced[j] !== 'number') {
                allpoints = false;
                break;
              }
            }
            if (allpoints) {
              const list = new List<Point>();
              for (let k = 0; k < replaced.length / 2; k++) {
                list.add(new Point(replaced[k * 2], replaced[k * 2 + 1]));
              }
              list.freeze();
              x[p] = list;
            }
          }
        }
      }
      // maybe replace the whole Object too
      return this.replaceJsonObject(x);
    }
    // otherwise just return the value
    return x;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {string} str
   * @return {string}
   */
  protected quote(str: string): string {
    let s = '';
    const len = str.length;
    for (let i = 0; i < len; i++) {
      const c = str[i];
      if (c === '\"' || c === '\\') {
        s += '\\' + c;
      } else if (c === '\b') {
        s += '\\b';
      } else if (c === '\f') {
        s += '\\f';
      } else if (c === '\n') {
        s += '\\n';
      } else if (c === '\r') {
        s += '\\r';
      } else if (c === '\t') {
        s += '\\t';
      } else {  // JSON does not support control characters nor \v
        const code = str.charCodeAt(i);
        if (code < 16) {
          s += '\\u000' + str.charCodeAt(i).toString(16);
        } else if (code < 32) {
          s += '\\u00' + str.charCodeAt(i).toString(16);
        } else if (code === 8232) {
          s += '\\u2028';
        } else if (code === 8233) {
          s += '\\u2029';
        } else {
          s += c;
        }
      }
    }
    return '\"' + s + '\"';
  }

  /**
   * @hidden @internal
   * @export
   * @this {Model}
   * @param {*} x
   * @return {string}
   */
  protected writeJsonValue(x: any): string {
    if (x === undefined) return 'undefined';  // shouldn't happen
    if (x === null) return 'null';
    if (x === true) return 'true';
    if (x === false) return 'false';
    if (typeof x === 'string') return this.quote(x);
    if (typeof x === 'number') {
      if (x === Infinity) return '9e9999';
      if (x === -Infinity) return '-9e9999';
      if (isNaN(x)) return '{\"class\":\"NaN\"}';
      return x.toString();
    }
    if (x instanceof Date) return '{\"class\":\"Date\", \"value\":\"' + x.toJSON() + '\"}';
    if (x instanceof Number) return this.writeJsonValue(x.valueOf());  // is this safe?
    if (Util.isArray(x)) return this.writeJsonArray(x);
    if (Util.isObject(x)) return this.writeJsonObject(x);
    if (typeof x === 'function') return 'null';  // ?? function
    return x.toString();  // ?? XML objects, DOM elements
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Array.<*>} arr
   * @param {boolean=} newlines
   * @return {string}
   */
  protected writeJsonArray(arr: Array<Key | ObjectData>, newlines?: boolean): string {
    if (newlines === undefined) newlines = false;
    const len = Util.arrayLength(arr);
    if (len <= 0) return '[]';
    const buf = new StringBuilder();
    buf.add('[ ');
    if (newlines && len > 1) buf.add('\n');
    for (let i = 0; i < len; i++) {
      const v = Util.arrayElt(arr, i);
      if (v === undefined) continue;
      if (i > 0) {
        buf.add(',');
        if (newlines) buf.add('\n');
      }
      buf.add(this.writeJsonValue(v));
    }
    if (newlines && len > 1) buf.add('\n');
    buf.add(' ]');
    return buf.toString();
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} data
   * @param {string} propname
   * @param {*} propval
   * @return {boolean}
   */
  private skipsJsonObjectProperty(data: ObjectData, propname: string, propval: any): boolean {
    if (propval === undefined) return true;
    if (propname === '__gohashid') return true;  // definitely want to skip this!
    if (propname[0] === '_') return true; // skip properties named this way
    if (typeof propval === 'function') return true;
    return false;
  }

  /**
   * @hidden @internal
   * When you know the value must be a number, use a JSON-safe representation.
   * @this {Model}
   * @param {number} n
   * @return {string|number}
   */
  private safeJsonNumber(n: number): string | number {
    if (isNaN(n)) return 'NaN';
    if (n === Infinity) return '9e9999';
    if (n === -Infinity) return '-9e9999';
    return n;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} data
   * @return {string}
   */
  private writeJsonObject(data: ObjectData): string {
    // maybe substitute a JSON-writable object for special classes
    // ?? someday generalize this
    const d = data;
    if (d instanceof Point) {
      data = { 'class': 'go.Point', 'x': this.safeJsonNumber(d.x), 'y': this.safeJsonNumber(d.y) };

    } else if (d instanceof Size) {
      data = { 'class': 'go.Size', 'width': this.safeJsonNumber(d.width), 'height': this.safeJsonNumber(d.height) };

    } else if (d instanceof Rect) {
      data = { 'class': 'go.Rect', 'x': this.safeJsonNumber(d.x), 'y': this.safeJsonNumber(d.y), 'width': this.safeJsonNumber(d.width), 'height': this.safeJsonNumber(d.height) };

    } else if (d instanceof Margin) {
      data = { 'class': 'go.Margin', 'top': this.safeJsonNumber(d.top), 'right': this.safeJsonNumber(d.right),
               'bottom': this.safeJsonNumber(d.bottom), 'left': this.safeJsonNumber(d.left) };

    } else if (d instanceof Spot) {
      if (!d.isSpot()) {  // if a special Spot value, write the name instead
        data = { 'class': 'go.Spot', 'enum': d.toString() };

      } else {
        data = { 'class': 'go.Spot', 'x': this.safeJsonNumber(d.x), 'y': this.safeJsonNumber(d.y),
                 'offsetX': this.safeJsonNumber(d.offsetX), 'offsetY': this.safeJsonNumber(d.offsetY) };

      }
    } else if (d instanceof Brush) {
      data = { 'class': 'go.Brush', type: d.type.name };

      if (d.type === Brush.Solid) {
        data['color'] = d.color;
      } else if (d.type === Brush.Linear || d.type === Brush.Radial) {
        data['start'] = d.start;
        data['end'] = d.end;
        if (d.type === Brush.Radial) {
          if (d.startRadius !== 0) data['startRadius'] = this.safeJsonNumber(d.startRadius);
          if (!isNaN(d.endRadius)) data['endRadius'] = this.safeJsonNumber(d.endRadius);
        }
        // ??? Brush.pattern
      }
      if (d.colorStops !== null) {
        const map: ObjectData = {};  // generating JSON
        const it = d.colorStops.iterator;
        while (it.next()) {  // JSON keys must be strings, not numbers
          const key = it.key;
          const val = it.value;
          map[key] = val;
        }
        data['colorStops'] = map;
      }
    } else if (d instanceof Geometry) {
      data = { 'class': 'go.Geometry', type: d.type.name };

      if (d.startX !== 0) data['startX'] = this.safeJsonNumber(d.startX);
      if (d.startY !== 0) data['startY'] = this.safeJsonNumber(d.startY);
      if (d.endX !== 0) data['endX'] = this.safeJsonNumber(d.endX);
      if (d.endY !== 0) data['endY'] = this.safeJsonNumber(d.endY);
      if (!d.spot1.equals(Spot.TopLeft)) data['spot1'] = d.spot1;
      if (!d.spot2.equals(Spot.BottomRight)) data['spot2'] = d.spot2;
      if (d.type === Geometry.Path) {
        data['path'] = Geometry.stringify(d);
      }
    } else if (d instanceof EnumValue) {
      data = { 'class': 'go.EnumValue', 'classType': this.getClassName(d.classType), 'name': d.name };
    } else {
        if (Debug && Model.getClassForString(this.getClassName(d)) !== null) {
        Util.trace('ERROR: trying to convert a GraphObject or Diagram or Model or Tool or Layout or UndoManager or other unknown data into JSON text: ' + d.toString());
        return '{}';  // for JSON
      }
    }

    // now we write out all of DATA's enumerable properties that we don't skip
    let str = '{';
    let first = true;
    for (const p in data) {
      const v = Model.safePropertyValue(data, p);
      if (this.skipsJsonObjectProperty(data, p, v)) continue;
      if (first) {
        first = false;
      } else {
        str += ', ';
      }
      str += '\"' + p + '\":';
      // handle special case of List of Points as an Array of x,y pairs
      if (p === 'points' && v instanceof List /*&& used to also test for: v._type === Point */) {
        const ap: List<Point> = v;
        let s = '[';
        const it2 = ap.iterator;
        while (it2.next()) {
          const pt = it2.value;
          if (s.length > 1) s += ',';
          s += this.writeJsonValue(pt.x);
          s += ',';
          s += this.writeJsonValue(pt.y);
        }
        s += ']';
        str += s;
      } else {
        str += this.writeJsonValue(v);
      }
    }
    str += '}';
    return str;
  }

  /**
   * @hidden @internal
   * Parse our JSON-safe representation of a number as a number.
   * @this {Model}
   * @param {number|string} n a number or a string.
   * @return {number}
   */
  private readSafeJsonNumber(n: number | string): number {
    if (typeof n === 'number') return n;
    if (n === 'NaN') return NaN;
    if (n === '9e9999') return Infinity;
    if (n === '-9e9999') return - Infinity;
    return parseFloat(n);
  }

  /**
   * @hidden @internal
   * Substitute a "real" object for a plain JavaScript object with 'class: "go.ClassName"'.
   * This assumes all subproperties have already been replaced if needed.
   * @this {Model}
   * @param {*} d
   * @return {*}
   */
  protected replaceJsonObject(d: any): any {
    if (typeof d !== 'object') return d;
    let obj = d;
    const cname = d['class'];
    if (cname === 'NaN') {
      obj = NaN;
    } else if (cname === 'Date') {
      obj = new Date(d['value']);
    } else if (cname === 'go.Point') {
      obj = new Point(this.readSafeJsonNumber(d['x']), this.readSafeJsonNumber(d['y']));
    } else if (cname === 'go.Size') {
      obj = new Size(this.readSafeJsonNumber(d['width']), this.readSafeJsonNumber(d['height']));
    } else if (cname === 'go.Rect') {
      obj = new Rect(this.readSafeJsonNumber(d['x']), this.readSafeJsonNumber(d['y']), this.readSafeJsonNumber(d['width']), this.readSafeJsonNumber(d['height']));
    } else if (cname === 'go.Margin') {
      obj = new Margin(this.readSafeJsonNumber(d['top']), this.readSafeJsonNumber(d['right']), this.readSafeJsonNumber(d['bottom']), this.readSafeJsonNumber(d['left']));
    } else if (cname === 'go.Spot') {
      if (typeof d['enum'] === 'string') {
        obj = Spot.parse(d['enum']);
      } else {
        obj = new Spot(this.readSafeJsonNumber(d['x']), this.readSafeJsonNumber(d['y']), this.readSafeJsonNumber(d['offsetX']), this.readSafeJsonNumber(d['offsetY']));
      }
    } else if (cname === 'go.Brush') {
      const b = new Brush();
      b.type = EnumValue.findName(Brush, d['type']);
      if (typeof d['color'] === 'string') b.color = d['color'];
      if (d['start'] instanceof Spot) b.start = d['start'];
      if (d['end'] instanceof Spot) b.end = d['end'];
      if (typeof d['startRadius'] === 'number') b.startRadius = this.readSafeJsonNumber(d['startRadius']);
      if (typeof d['endRadius'] === 'number') b.endRadius = this.readSafeJsonNumber(d['endRadius']);
      // ??? cannot restore any Brush.pattern
      const stops = d['colorStops'];
      if (Util.isObject(stops)) {
        for (const p in stops) {
          const f = parseFloat(p);
          b.addColorStop(f, stops[p]);
        }
      }
      obj = b;
    } else if (cname === 'go.Geometry') {
      let g: Geometry = null;
      if (typeof d['path'] === 'string') {
        g = Geometry.parse(d['path']);
      } else {
        g = new Geometry();
      }
      g.type = EnumValue.findName(Geometry, d['type']);
      if (typeof d['startX'] === 'number') g.startX = this.readSafeJsonNumber(d['startX']);
      if (typeof d['startY'] === 'number') g.startY = this.readSafeJsonNumber(d['startY']);
      if (typeof d['endX'] === 'number') g.endX = this.readSafeJsonNumber(d['endX']);
      if (typeof d['endY'] === 'number') g.endY = this.readSafeJsonNumber(d['endY']);
      if (d['spot1'] instanceof Spot) g.spot1 = d['spot1'];
      if (d['spot2'] instanceof Spot) g.spot2 = d['spot2'];
      obj = g;
    } else if (cname === 'go.EnumValue') {
      let ctn = d['classType'];
      if (ctn.indexOf('go.') === 0) {  // if prefixed with "go.", strip it
        ctn = ctn.substr(3);
      }
      const cls = Model.getClassForString(ctn);
      obj = EnumValue.findName(cls, d['name']);
    }
    return obj;
  }

  /**
   * Gets or sets the name of this model.
   * The initial name is an empty string.
   * The value must not be null.
   */
  /** @export */ get name(): string { return this._name; }
  /** @export */ set name(value: string) {
    const old = this._name;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', Model, 'name');
      this._name = value;
      this.raiseChanged('name', old, value);
    }
  }

  /**
   * Gets or sets the name of the format of the diagram data.
   * The default value is the empty string.
   * The value must not be null.
   * Use different values to prevent parts from one model to be copy/pasted or drag-and-dropped into another diagram/model.
   */
  /** @export */ get dataFormat(): string { return this._dataFormat; }
  /** @export */ set dataFormat(value: string) {
    const old = this._dataFormat;
    if (old !== value) {
      Util.checkPrimitive(value, 'string', Model, 'dataFormat');
      this._dataFormat = value;
      this.raiseChanged('dataFormat', old, value);
    }
  }

  /**
   * Gets or sets whether this model may be modified, such as adding nodes.
   * By default this value is false.
   * Setting the {@link #nodeDataArray} to something that is not a true Array of Objects
   * will cause this to be set to true.
   *
   * Model methods and property setters do not heed this property.
   * It is up to code that uses a model to check this property when it might want to
   * prevent changes to the model.
   */
  /** @export */ get isReadOnly(): boolean { return this._isReadOnly; }
  /** @export */ set isReadOnly(value: boolean) {
    const old = this._isReadOnly;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', Model, 'isReadOnly');
      this._isReadOnly = value;
      this.raiseChanged('isReadOnly', old, value);
    }
  }

  /**
   * Gets a JavaScript Object that can hold programmer-defined property values for the model as a whole,
   * rather than just for one node or one link.
   *
   * By default this an object with no properties.
   * Any properties that you add to this object will be written out by {@link #toJson}
   * and will be restored by {@link Model.fromJson}, if the following conditions are true:
   *   - the property is enumerable and its name does not start with an underscore ('_')
   *   - the property value is not undefined and is not a function
   *   - the model knows how to convert the property value to JSON format
   *   - property values that are Objects or Arrays form a tree structure -- no shared or cyclical references
   *
   * Most object classes cannot be serialized into JSON without special knowledge and processing at both ends.
   * The {@link #toJson} and {@link Model.fromJson} methods automatically do such processing for numbers that are NaN
   * and for objects that are of class {@link Point}, {@link Size}, {@link Rect}, {@link Margin}, {@link Spot},
   * {@link Brush} (but not for brush patterns), and for {@link Geometry}.
   *
   * At the current time one cannot have a {@link Diagram} as a binding target.
   * Calling {@link #setDataProperty} will work to change a property value, but there are no target bindings in any Diagrams to be updated.
   * Because the binding mechanism is unavailable for this object, we recommend that when you want to save a model
   * that you explicitly set properties on this object just before calling {@link #toJson}.
   * When loading a model, call {@link Model.fromJson} and explicitly get the properties that you want to set on a Diagram.
   */
  /** @export */ get modelData(): ObjectData { return this._modelData; }
  /** @export */ set modelData(value: ObjectData) {
    const old = this._modelData;
    if (old !== value) {
      Util.checkPrimitive(value, 'object', Model, 'modelData');
      this._modelData = value;
      this.raiseChanged('modelData', old, value);
      this.updateTargetBindings(value);
    }
  }

  // For ChangedEvents and notification:

  /**
   * Register an event handler that is called when there is a ChangedEvent.
   *
   * This registration does not raise a {@link ChangedEvent}.
   * In case a Diagram's Model may be replaced, you may prefer calling {@link Diagram#addModelChangedListener} instead of this method.
   *
   * Event listeners are not written out by {@link #toJson}.
   *
   * Do not add or remove Changed listeners during the execution of a Changed listener.
   * @export
   * @this {Model}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #removeChangedListener}
   * @see {@link Diagram#addModelChangedListener}
   */
  public addChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Model, 'addChangedListener:listener');
    this._listeners.add(listener);
  }

  /**
   * Unregister an event handler listener.
   *
   * This deregistration does not raise a {@link ChangedEvent}.
   * In case a Diagram's Model may be replaced, you may prefer calling {@link Diagram#removeModelChangedListener} instead of this method.
   * @export
   * @this {Model}
   * @param {function(ChangedEvent)} listener a function that takes a {@link ChangedEvent} as its argument.
   * @see {@link #addChangedListener}
   * @see {@link Diagram#removeModelChangedListener}
   */
  public removeChangedListener(listener: ((e: ChangedEvent) => void)): void {
    Util.checkPrimitive(listener, 'function', Model, 'removeChangedListener:listener');
    this._listeners.remove(listener);
  }

  /**
   * @hidden @internal
   * Notify any ChangedEvent listeners.
   * This is called by {@link #raiseModelChangedEvent}.
   * Each listener function registered with calls to {@link #addChangedListener} is called with the given event.
   * Unless {@link #skipsUndoManager} is true, this then calls {@link UndoManager#handleChanged}
   * so that the UndoManager gets the opportunity to record the change.
   * @this {Model}
   * @param {ChangedEvent} e
   */
  public callChangedListeners(e: ChangedEvent): void {
    if (!this.skipsUndoManager) {
      this.undoManager.handleChanged(e);
    }

    const list = this._listeners;
    const len = list.length;
    for (let i = 0; i < len; i++) {
      const func = list.elt(i);
      func(e);
    }
  }

  /**
   * Call this method to notify that the model or its objects have changed.
   * This constructs a {@link ChangedEvent} and calls all Changed listeners.
   * @export
   * @this {Model}
   * @param {EnumValue} change specifies the general nature of the change; typically the value is {@link ChangedEvent.Property}.
   * @param {string|function(ObjectData,?=):?} propertyname names the property that was modified, or a function that takes an Object and returns the property value.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}, {@link Diagram}, or a {@link Model}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam an optional value that helps describe the older value.
   * @param {*=} newparam an optional value that helps describe the newer value.
   */
  public raiseChangedEvent(change: EnumValue, propertyname: string | ((obj: ObjectData, val: any) => any), obj: ObjectData,
                           oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    this.raiseModelChangedEvent('', change, propertyname, obj, oldval, newval, oldparam, newparam);
  }

  /**
   * @hidden @internal
   * Call this method to notify about a model property having changed value.
   * This constructs a {@link ChangedEvent} and calls all Changed listeners..
   * @export
   * @this {Model}
   * @param {string|function(ObjectData,?=):?} propertyname the name of the property, or a function that takes an Object and returns the property value.
   * @param {*} oldval the previous or old value for the property.
   * @param {*} newval the next or new value for the property.
   * @param {*=} oldparam an optional value additionally describing the old value.
   * @param {*=} newparam an optional value additionally describing the new value.
   */
  protected raiseChanged(propertyname: string | ((obj: ObjectData, val: any) => any), oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    this.raiseModelChangedEvent('', ChangedEvent.Property, propertyname, this, oldval, newval, oldparam, newparam);
  }

  /**
   * Call this method to notify about a data property having changed value.
   * This constructs a {@link ChangedEvent} and calls all Changed listeners.
   *
   * You should call this method only if the property value actually changed.
   * This method is called by {@link #setDataProperty}.
   * @export
   * @this {Model}
   * @param {Object} data the data object whose property changed value.
   * @param {string|function(ObjectData,?=):?} propertyname the name of the property, or a function that takes an Object and returns the property value.
   * @param {*} oldval the previous or old value for the property.
   * @param {*} newval the next or new value for the property.
   * @param {*=} oldparam an optional value additionally describing the old value.
   * @param {*=} newparam an optional value additionally describing the new value.
   */
  public raiseDataChanged(data: ObjectData, propertyname: string | ((obj: ObjectData, val: any) => any), oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    this.raiseModelChangedEvent('', ChangedEvent.Property, propertyname, data, oldval, newval, oldparam, newparam);
  }

  /**
   * @hidden @internal
   * This is an internal method for notifying about changes to the model structure.
   * @this {Model}
   * @param {string} modelchangename the kind of model change;
   *   use the empty string for regular property changes that do not affect the model.
   * @param {EnumValue} change specifies the general nature of the change;
   * permissible values de {@link ChangedEvent.Property}, {@link ChangedEvent.Insert},
   * {@link ChangedEvent.Remove}, and {@link ChangedEvent.Transaction}.
   * @param {string|function(ObjectData,?=):?} propertyname names the property that was modified, or a function that takes an Object and returns the property value.
   * @param {Object} obj the object that was modified, typically a {@link GraphObject}, {@link Model}, or a {@link Layer}.
   * @param {*} oldval the previous or older value.
   * @param {*} newval the next or newer value.
   * @param {*=} oldparam helps describe the older value.
   * @param {*=} newparam helps describe the newer value.
   */
  protected raiseModelChangedEvent(modelchangename: string, change: EnumValue, propertyname: string | ((obj: ObjectData, val: any) => any), obj: ObjectData,
                                   oldval: any, newval: any, oldparam?: any, newparam?: any): void {
    if (oldparam === undefined) oldparam = null;
    if (newparam === undefined) newparam = null;
    // ChangedEvent property setters will check argument types

    const e = new ChangedEvent();  // ?? cache event objects
    e.model = this;
    e.change = change;
    e.modelChange = modelchangename;
    e.propertyName = propertyname;
    e.object = obj;
    e.oldValue = oldval;
    e.oldParam = oldparam;
    e.newValue = newval;
    e.newParam = newparam;

    this.callChangedListeners(e);
  }

  // for undo/redo:

  /**
   * Gets or sets the {@link UndoManager} for this Model.
   *
   * The default UndoManager has its {@link UndoManager#isEnabled} property set to false.
   * If you want users to undo and redo, you should set that property to true once you have initialized the Diagram or its Model.
   *
   * This property setter does not raise a {@link ChangedEvent}.
   */
  /** @export */ get undoManager(): UndoManager { return this._undoManager; }
  /** @export */ set undoManager(value: UndoManager) {
    const old = this._undoManager;
    if (old !== value) {
      Util.checkClass(value, UndoManager, Model, 'undoManager');
      if (old !== null) old.removeModel(this);
      this._undoManager = value;
      if (value !== null) value.addModel(this);
    }
  }

  /**
   * Gets or sets whether {@link ChangedEvent}s are not recorded by the {@link UndoManager}.
   * The initial and normal value is false.
   * WARNING: while this property is true do not perform any changes that cause any previous transactions
   * to become impossible to undo.
   *
   * When this property is true, changing the Model or any data object does not call {@link UndoManager#handleChanged}.
   * Even when this property is true,
   * transactions (such as calls to {@link #startTransaction}) and undo/redo (such as calls to {@link CommandHandler#undo})
   * are still delegated to the {@link #undoManager}.
   *
   * You should set this to true only temporarily, and you should remember its previous value before setting this to true.
   * When finishing the period for which you want the UndoManager to be disabled,
   * do not blindly set this property to false.
   * You should set this back to the value it had before you set it to true.
   * For more permanent disabling of the {@link UndoManager}, set {@link UndoManager#isEnabled} to false.
   *
   * This property is also set when setting {@link Diagram#skipsUndoManager}.
   * Setting this property does not raise a {@link ChangedEvent}.
   */
  /** @export */ get skipsUndoManager(): boolean { return this._skipsUndoManager; }
  /** @export */ set skipsUndoManager(value: boolean) {
    Util.checkPrimitive(value, 'boolean', Model, 'skipsUndoManager');
    this._skipsUndoManager = value;
  }

  /**
   * @hidden @internal
   * This is called during an undo or redo to modify the model or its objects.
   *
   * This does not raise a {@link ChangedEvent}.
   * @this {Model}
   * @param {ChangedEvent} e This describes the change that needs to be done.
   * @param {boolean} undo If true, this method should restore the older state, otherwise the newer state.
   */
  public changeState(e: ChangedEvent, undo: boolean): void {
    if (e === null) return;
    if (e.model !== this) return;  // ignore changes to Diagram or other Models
    if (e.change === ChangedEvent.Property) {
      const obj = e.object;
      const pn = e.propertyName;
      const v = e.getValue(undo);
      Model.safePropertySet(obj, pn, v);
    } else if (e.change === ChangedEvent.Insert) {
      const idx = e.newParam;
      if (e.modelChange === 'nodeDataArray') {
        const data = e.newValue;
        if (Util.isObject(data) && typeof idx === 'number') {
          const key = this.getKeyForNodeData(data);
          if (undo) {
            if (Util.arrayElt(this._nodeDataArray, idx) === data) {
              Util.arrayRemoveAt(this._nodeDataArray, idx);
            }
            if (key !== undefined) {
              this._mapKeyNodeData.remove(key);
            }
          } else {  // redo Insert
            if (Util.arrayElt(this._nodeDataArray, idx) !== data) {
              Util.arrayInsertAt(this._nodeDataArray, idx, data);
            }
            if (key !== undefined) {
              this._mapKeyNodeData.add(key, data);
            }
          }
        }  // no-op if idx is null
      } else if (e.modelChange === '') {  // inserting an item into an array that is a data property value
        let arr = e.object;
        if (arr && !Util.isArray(arr) && e.propertyName) arr = Model.safePropertyValue(e.object, e.propertyName);
        if (Util.isArray(arr) && typeof idx === 'number') {
          // duplicates are OK
          const item = e.newValue;
          if (undo) {
            Util.arrayRemoveAt(arr, idx);
          } else {
            Util.arrayInsertAt(arr, idx, item);
          }
        }  // no-op if idx is null
      } else {
        Util.throwError('unknown ChangedEvent.Insert modelChange: ' + e.toString());
      }
    } else if (e.change === ChangedEvent.Remove) {
      const idx = e.oldParam;
      if (e.modelChange === 'nodeDataArray') {
        const data = e.oldValue;
        if (Util.isObject(data) && typeof idx === 'number') {
          const key = this.getKeyForNodeData(data);
          if (undo) {
            if (Util.arrayElt(this._nodeDataArray, idx) !== data) {
              Util.arrayInsertAt(this._nodeDataArray, idx, data);
            }
            if (key !== undefined) {
              this._mapKeyNodeData.add(key, data);
            }
          } else {  // redo Remove
            if (Util.arrayElt(this._nodeDataArray, idx) === data) {
              Util.arrayRemoveAt(this._nodeDataArray, idx);
            }
            if (key !== undefined) {
              this._mapKeyNodeData.remove(key);
            }
          }
        }  // no-op if idx is null
      } else if (e.modelChange === '') {  // removing an item from an array that is a data property value
        let arr = e.object;
        if (arr && !Util.isArray(arr) && e.propertyName) arr = Model.safePropertyValue(e.object, e.propertyName);
        if (Util.isArray(arr) && typeof idx === 'number') {
          // duplicates are OK
          const item = e.oldValue;
          if (undo) {
            Util.arrayInsertAt(arr, idx, item);
          } else {
            Util.arrayRemoveAt(arr, idx);
          }
        }  // no-op if idx is null
      } else {
        Util.throwError('unknown ChangedEvent.Remove modelChange: ' + e.toString());
      }
    } else if (e.change === ChangedEvent.Transaction) {
      // don't do anything
    } else {
      Util.throwError('unknown ChangedEvent: ' + e.toString());
    }
  }

  /**
   * Begin a transaction, where the changes are held by a {@link Transaction} object
   * in the {@link UndoManager}.
   * This just calls {@link UndoManager#startTransaction}.
   * @export
   * @this {Model}
   * @param {string=} tname a descriptive name for the transaction.
   * @return {boolean} the value returned by {@link UndoManager#startTransaction}.
   * @see {@link #commit}
   */
  public startTransaction(tname?: string): boolean {
    return this.undoManager.startTransaction(tname);
  }

  /**
   * Commit the changes of the current transaction.
   * This just calls {@link UndoManager#commitTransaction}.
   * @export
   * @this {Model}
   * @param {string=} tname a descriptive name for the transaction.
   * @return {boolean} the value returned by {@link UndoManager#commitTransaction}.
   */
  public commitTransaction(tname?: string): boolean {
    return this.undoManager.commitTransaction(tname);
  }

  /**
   * Rollback the current transaction, undoing any recorded changes.
   * This just calls {@link UndoManager#rollbackTransaction}.
   * @export
   * @this {Model}
   * @return {boolean} the value returned by {@link UndoManager#rollbackTransaction}.
   */
  public rollbackTransaction(): boolean {
    return this.undoManager.rollbackTransaction();
  }

  /**
   * Starts a new transaction, calls the provided function, and commits the transaction.
   * Code is called within a try-finally loop.
   * If the function does not return normally, this rolls back the transaction rather than committing it.
   * Example usage:
   * ```js
   * model.commit(m => m.addNodeData({ counter: myCounter++ }), "Added Node");
   * ```
   * @export
   * @this {Model}
   * @param {Function} func the function to call as the transaction body
   * @param {(string|null)=} tname a descriptive name for the transaction, or null to temporarily set {@link #skipsUndoManager} to true;
   *        if no string transaction name is given, an empty string is used as the transaction name
   * @since 1.8
   */
  public commit(func: (m: Model) => void, tname?: string | null): void {
    let tn = tname;
    if (tn === undefined) tn = '';
    const oldskips = this.skipsUndoManager;
    if (tn === null) {
      this.skipsUndoManager = true;
      tn = '';
    }
    this.undoManager.startTransaction(tn);
    let success = false;
    try {
      func(this);
      success = true;
    } finally {
      if (success) {
        this.undoManager.commitTransaction(tn);
      } else {
        this.undoManager.rollbackTransaction();
      }
      this.skipsUndoManager = oldskips;
    }
  }

  /**
   * Find a {@link Part} corresponding to the given data and
   * call its {@link Panel#updateTargetBindings} method, in each {@link Diagram}
   * that uses this Model.
   *
   * Caution: setting a data property without calling {@link #setDataProperty}
   * and then calling this updateTargetBindings method will update GraphObjects that are bound to the property,
   * but such data settings will not be recorded in the UndoManager and therefore will not be undone/redone,
   * causing an inconsistency between the GraphObjects and the part data.
   * @export
   * @this {Model}
   * @param {Object} data The data object in this model that was modified.
   * @param {string=} srcpropname If not present or the empty string,
   *   update all bindings on the target {@link Part} or item {@link Panel}
   *   otherwise update only those bindings using this source property name.
   */
  public updateTargetBindings(data: ObjectData, srcpropname?: string): void {
    if (srcpropname === undefined) srcpropname = '';
    this.raiseModelChangedEvent('SourceChanged', ChangedEvent.Transaction, srcpropname, data, null, null);
  }

  /**
   * Gets or sets the name of the data property that returns a unique id number or string for each node data object.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return the unique key value;
   * if the second argument is supplied, the function should modify the node data object so that it has that new value as its unique key value.
   * The default value is the name 'key', meaning that it expects the data to have a property named 'key' if it has a key value.
   * The name must not be null or the empty string.
   * You must set this property before assigning the {@link #nodeDataArray}.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getKeyForNodeData}
   */
  /** @export */ get nodeKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._nodeKeyProperty; }
  /** @export */ set nodeKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._nodeKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, Model, 'nodeKeyProperty');
      if (value === '') Util.throwError('Model.nodeKeyProperty may not be the empty string');
      if (this._mapKeyNodeData.count > 0) Util.throwError('Cannot set Model.nodeKeyProperty when there is existing node data');
      this._nodeKeyProperty = value;
      this.raiseChanged('nodeKeyProperty', old, value);
      // ?? recompute relationships
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {*} val
   * @param {Function} cls
   * @param {string} member
   */
  protected checkStringOrFunction(val: any, cls: Function, member: string): void {
    if (typeof val !== 'string' && typeof val !== 'function') {
      Util.throwTypeError(val, 'string or function', cls, member);
    }
  }

  /**
   * Given a node data object return its unique key: a number or a string.
   * This returns undefined if there is no key value.
   *
   * It is possible to change the key for a node data object by calling {@link #setKeyForNodeData}.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {string|number|undefined}
   * @see {@link #nodeKeyProperty}
   * @see {@link #setKeyForNodeData}
   * @see {@link #findNodeDataForKey}
   */
  public getKeyForNodeData(nodedata: ObjectData): Key {
    if (nodedata === null) return undefined;
    const kname = this._nodeKeyProperty;
    if (kname === '') return undefined;  // ?? support direct references to node data
    const key = Model.safePropertyValue(nodedata, kname);
    if (key === undefined) return undefined;
    if (this.isKeyType(key)) return key;
    Util.throwError('Key value for node data ' + nodedata + ' is not a number or a string: ' + key);
    return undefined;
  }

  /**
   * Change the unique key of a given node data that is already in this model.
   * The new key value must be unique -- i.e. not in use by another node data object.
   * You can call {@link #findNodeDataForKey} to check if a proposed new key is already in use.
   *
   * This operation will check all data objects in the model and replace all references
   * using the old key value with the new one.
   *
   * If this is called on a node data object that is not (yet) in this model,
   * this unconditionally modifies the property to the new key value.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @param {string|number} key
   * @see {@link #nodeKeyProperty}
   * @see {@link #getKeyForNodeData}
   * @since 1.1
   */
  public setKeyForNodeData(nodedata: ObjectData, key: Key): void {
    if (key === undefined || key === null || !this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', Model, 'setKeyForNodeData:key');
    }
    if (nodedata === null) return;
    const kname = this._nodeKeyProperty;
    if (kname === '') return;
    if (!this.containsNodeData(nodedata)) {
      Model.safePropertySet(nodedata, kname, key);
      return;
    }
    const old = Model.safePropertyValue(nodedata, kname);
    if (old !== key) {
      if (this.findNodeDataForKey(key) !== null) {
        // Util.throwError('Model.setKeyForNodeData: trying to replace the key: ' + old + ' with a key used by another node data object: ' + key);
        return;
      }
      Model.safePropertySet(nodedata, kname, key);
      // update the key->data Map
      if (old !== undefined) this._mapKeyNodeData.remove(old);
      this._mapKeyNodeData.add(key, nodedata);
      this.raiseModelChangedEvent('nodeKey', ChangedEvent.Property, kname, nodedata, old, key);
      if (typeof kname === 'string') this.updateTargetBindings(nodedata, kname);
      // need to update all references using the old key so that all relationships are maintained
      this.replaceAllKeyReferences(old, key);
    }
  }

  /**
   * Gets or sets a function that returns a unique id number or string for a node data object.
   * This function is called by {@link #makeNodeDataKeyUnique}
   * when a node data object is added to the model, either as part of a new
   * {@link #nodeDataArray} or by a call to {@link #addNodeData}, to make sure the value of
   * {@link #getKeyForNodeData} is unique within the model.
   *
   * The value may be null in order to cause {@link #makeNodeDataKeyUnique} behave in the standard manner.
   * (The default value is null.)
   * You may want to supply a function here in order to make sure all of the automatically generated keys
   * are in a particular format.
   * Setting this property after setting {@link #nodeDataArray} has no real effect until there is a call
   * to {@link #addNodeData}.
   *
   * If you want to ensure that this function is called when copying data that already has a key,
   * set {@link #copiesKey} to false.
   * This is typically useful when copying a node from a Palette, where the key it has in the Palette's Model
   * happens to be unique within the target Diagram's Model.
   * Unless you set {@link #copiesKey} to false, the original key value will be copied and retained
   * if it is already unique within the target model.
   *
   * If a node data object is already in the model and you want to change its key value,
   * call {@link #setKeyForNodeData} with a new and unique key.
   */
  /** @export */ get makeUniqueKeyFunction(): ((model: Model, data: ObjectData) => Key) | null { return this._makeUniqueKeyFunction; }
  /** @export */ set makeUniqueKeyFunction(value: ((model: Model, data: ObjectData) => Key) | null) {
    const old = this._makeUniqueKeyFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Model, 'makeUniqueKeyFunction');
      this._makeUniqueKeyFunction = value;
      this.raiseChanged('makeUniqueKeyFunction', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {*} x
   * @return {boolean} true if the value might be used as a key identifying a node.
   */
  public isKeyType(x: any): boolean {
    return (typeof x === 'number' || typeof x === 'string');
  }

  /**
   * Decide if a given node data object is in this model, using reference equality.
   *
   * If you do not have a reference to the particular data object that is in the {@link #nodeDataArray},
   * you may need to search for it by iterating through that Array,
   * or by finding the desired {@link Node} or simple {@link Part} in a {@link Diagram} and getting that node's {@link Panel#data},
   * or most likely by calling {@link #findNodeDataForKey}.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {boolean} true if it is a node data object in this model; false otherwise.
   */
  public containsNodeData(nodedata: ObjectData): boolean {
    const key = this.getKeyForNodeData(nodedata);
    if (key === undefined) return false;
    return this._mapKeyNodeData.getValue(key) === nodedata;
  }

  /**
   * Given a number or string, find the node data object in this model
   * that uses the given value as its unique key.
   * @export
   * @this {Model}
   * @param {(string|number|undefined)} key a string or a number.
   * @return {Object} null if the key is not present in the model,
   * or if the key is null or undefined or not a string or number.
   * @see {@link #containsNodeData}
   * @see {@link #getKeyForNodeData}
   */
  public findNodeDataForKey(key: Key): ObjectData | null {
    if (key === null) Util.throwError('Model.findNodeDataForKey:key must not be null');
    if (key === undefined) return null;
    if (!this.isKeyType(key)) return null;
    return this._mapKeyNodeData.getValue(key);
  }

  // Node data

  /**
   * Gets or sets the array of node data objects that correspond to {@link Node}s,
   * {@link Group}s, or non-Link {@link Part}s in the {@link Diagram}.
   * The initial value is an empty Array.
   *
   * For each Object in the Array, {@link #getKeyForNodeData} should return a number or string
   * uniquely identifying the node data within the model.
   * If it returns undefined, this calls {@link #makeNodeDataKeyUnique},
   * to make sure the node data has a unique key.
   * These key values may be used by other objects to refer to that particular node data object.
   * If more than one node data object has the same key,
   * there may be some confusion about which object to reference.
   *
   * If you want to use a custom data property for holding the unique key value on a node data object,
   * you should set {@link #nodeKeyProperty} before you set this `nodeDataArray` property.
   *
   * Adding or removing data from this Array will not notify this model or the diagram
   * that there are any new nodes or that any nodes have been deleted.
   * Instead you should call {@link #addNodeData} or {@link #removeNodeData}.
   */
  /** @export */ get nodeDataArray(): Array<ObjectData> { return this._nodeDataArray; }
  /** @export */ set nodeDataArray(value: Array<ObjectData>) {
    const old = this._nodeDataArray;
    if (old !== value) {
      Util.checkArray(value, Model, 'nodeDataArray');

      // completely replacing all node data, so are replacing all key information
      this._mapKeyNodeData.clear();
      // This does not optimize cases where some of the same node data objects are in the new array.
      // (Such an optimization might not even be desired, in case the node data objects have been modified.)
      // Establish as unresolved all references in link data.
      this.unresolveAllReferencesToNodeData();

      // Assign unique numbers as keys for each node data.
      // The actual key values are needed for hashing in Diagram.
      const len = Util.arrayLength(value);
      for (let i = 0; i < len; i++) {
        const data = Util.arrayElt(value, i);
        if (!Util.isObject(data)) {
          Util.throwError('Model.nodeDataArray must only contain Objects, not: ' + data);
          return;
        }
        Set.hashIdUnique(data);
      }

      this._nodeDataArray = value;

      // Find the key for each node data, and remember in a Map.
      const delayed = new List<ObjectData>();
      for (let i = 0; i < len; i++) {
        const data = Util.arrayElt(value, i);
        const key = this.getKeyForNodeData(data);
        if (key === undefined) {
          delayed.add(data);
        } else if (this._mapKeyNodeData.getValue(key) !== null) {
          // duplicate keys: call makeNodeDataKeyUnique soon
          delayed.add(data);
        } else {
          this._mapKeyNodeData.add(key, data);
        }
      }

      // Some node data didn't have a key value;
      // generate a unique number by modifying the data, then remember it.
      // We delay the calls to makeNodeDataKeyUnique to avoid accidentally
      // generating a new key that happens to be the same as one that
      // is in a not-yet-seen node data.
      const dit = delayed.iterator;
      while (dit.next()) {
        const data = dit.value;
        this.makeNodeDataKeyUnique(data);
        const newkey = this.getKeyForNodeData(data);
        if (newkey !== undefined) this._mapKeyNodeData.add(newkey, data);
      }

      // Notify about the property change
      this.raiseModelChangedEvent('nodeDataArray', ChangedEvent.Property, 'nodeDataArray', this, old, value);

      // Now resolve references
      for (let i = 0; i < len; i++) {
        const data = Util.arrayElt(value, i);
        this.resolveReferencesToNodeData(data);
        this.resolveReferencesInNodeData(data);
      }

      // assume setting nodeDataArray means all nodes are known, at least for now
      this.clearUnresolvedReferences();

      if (!Util.isModifiableArray(value)) {  // not Util.isArray
        this.isReadOnly = true;
      }
    }
  }

  /**
   * This method is called when a node data object is added to the model to make sure that
   * {@link #getKeyForNodeData} returns a unique key value.
   *
   * The key value should be unique within the set of data managed by this model:
   * {@link #nodeDataArray}.
   * If the key is already in use, this will assign an unused number to the
   * {@link #nodeKeyProperty} property on the data.
   *
   * If you want to customize the way in which node data gets a unique key,
   * you can set the {@link #makeUniqueKeyFunction} functional property.
   *
   * If the node data object is already in the model and you want to change its key value,
   * call {@link #setKeyForNodeData} and give it a new unique key value.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   */
  public makeNodeDataKeyUnique(nodedata: ObjectData): void {
    if (nodedata === null) return;
    const kname = this._nodeKeyProperty;
    if (kname === '') return;  // not using key values
    // maybe already has a unique key?
    let key = this.getKeyForNodeData(nodedata);
    if (key !== undefined && !this._mapKeyNodeData.contains(key)) return;
    // try a programmer-supplied function for generating a key
    const func = this._makeUniqueKeyFunction;
    if (func !== null) {
      key = func(this, nodedata);
      // if the return value is unique, set the key on the node data, just to be sure
      if (key !== undefined && key !== null && !this._mapKeyNodeData.contains(key)) {
        Model.safePropertySet(nodedata, kname, key);
        return;
      }
      // otherwise drop through to really make it unique
    }
    // Note that this does NOT call setKeyForNodeData, because this should not notify
    // and should not modify any references using the duplicate key.
    // no key or not unique key
    if (typeof key === 'string') {
      let j = 2;
      while (this._mapKeyNodeData.contains(key + j)) j++;
      Model.safePropertySet(nodedata, kname, key + j);
    } else if (key === undefined || typeof key === 'number') {
      // use negatives to reduce chance of future conflicts
      let j = -this._mapKeyNodeData.count - 1;
      while (this._mapKeyNodeData.contains(j)) j--;
      Model.safePropertySet(nodedata, kname, j);
    } else if (Debug) {
      Util.throwError('Model.getKeyForNodeData returned something other than a string or a number: ' + key);
    }
  }

  /**
   * When you want to add a node or group to the diagram,
   * call this method with a new data object.
   * This will add that data to the {@link #nodeDataArray} and
   * notify all listeners that a new node data object has been inserted into the collection.
   *
   * To remove a node from the diagram, you can remove its data object by calling {@link #removeNodeData}.
   *
   * To add or remove an object or value from an item array, call {@link #insertArrayItem} or {@link #removeArrayItem}.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   */
  public addNodeData(nodedata: ObjectData): void {
    if (nodedata === null) return;
    Set.hashIdUnique(nodedata);
    if (!this.containsNodeData(nodedata)) {
      this.doAddNodeData(nodedata, true);
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} nodedata
   * @param {boolean} addtoarray whether to add to .nodeDataArray
   */
  public doAddNodeData(nodedata: ObjectData, addtoarray: boolean): void {
    // see if already in key->data Map
    let key = this.getKeyForNodeData(nodedata);
    if (key === undefined) {  // no key, need key to be unique
      this.makeNodeDataKeyUnique(nodedata);
      key = this.getKeyForNodeData(nodedata);
    } else {  // has key -- is this node data already in the Map?
      if (this._mapKeyNodeData.getValue(key) === nodedata) return;
      // if not, duplicate key!
      this.makeNodeDataKeyUnique(nodedata);
      key = this.getKeyForNodeData(nodedata);
    }
    if (key === undefined) {
      Util.throwError('Model.makeNodeDataKeyUnique failed on ' + nodedata + '.  Data not added to Model.');
      return;
    }
    // update the key->data Map
    this._mapKeyNodeData.add(key, nodedata);
    let idx = null;  // leave as null when not actually adding to the array
    if (addtoarray) {
      // modify the nodeDataArray
      idx = Util.arrayLength(this._nodeDataArray);
      Util.arrayInsertAt(this._nodeDataArray, idx, nodedata);
    }
    // notify
    this.raiseModelChangedEvent('nodeDataArray', ChangedEvent.Insert, 'nodeDataArray', this, null, nodedata, null, idx);
    // resolve some references?
    this.resolveReferencesToNodeData(nodedata);
    this.resolveReferencesInNodeData(nodedata);
  }

  /**
   * Add to this model all of the node data held in an Array or in an {@link Iterable} of node data objects.
   * @export
   * @this {Model}
   * @param {Iterable.<Object>|Array.<Object>} coll a collection of node data objects to add to the {@link #nodeDataArray}
   * @since 1.3
   */
  public addNodeDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void {
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        this.addNodeData(Util.arrayElt(coll, i));
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const data = it.value;
        this.addNodeData(data);
      }
    }
  }

  /**
   * When you want to remove a node or group from the diagram,
   * call this method with an existing data object.
   * This will remove that data from the {@link #nodeDataArray} and
   * notify all listeners that a node data object has been removed from the collection.
   *
   * If you do not have a reference to the particular data object that is in the {@link #nodeDataArray},
   * you may need to search for it by iterating through that Array,
   * or by finding the desired {@link Node} or simple {@link Part} in a {@link Diagram} and getting that node's {@link Panel#data},
   * or most likely by calling {@link #findNodeDataForKey}.
   *
   * Removing a node data from a model does not automatically remove
   * any connected link data from the model.
   * Removing a node data that represents a group does not automatically remove
   * any member node data or link data from the model.
   *
   * To add a node to the diagram, you can add its data object by calling {@link #addNodeData}.
   *
   * To add or remove an object or value from an item array, call {@link #insertArrayItem} or {@link #removeArrayItem}.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   */
  public removeNodeData(nodedata: ObjectData): void {
    if (nodedata === null) return;
    this.doRemoveNodeData(nodedata, true);
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {Object} nodedata
   * @param {boolean} removefromarray whether to remove from .nodeDataArray
   */
  public doRemoveNodeData(nodedata: ObjectData, removefromarray: boolean): void {
    // remove from key->data Map
    const key = this.getKeyForNodeData(nodedata);
    if (key !== undefined) {
      // update the key->data Map
      this._mapKeyNodeData.remove(key);
    }
    let idx = null;  // leave as null when not actually removing from the array
    if (removefromarray) {
      // see if already removed
      idx = Util.arrayIndexOf(this._nodeDataArray, nodedata);
      if (idx < 0) return;
      // modify nodeDataArray
      Util.arrayRemoveAt(this._nodeDataArray, idx);
    }
    // notify
    this.raiseModelChangedEvent('nodeDataArray', ChangedEvent.Remove, 'nodeDataArray', this, nodedata, null, idx, null);
    // ?? Diagrams will call-back to delayReference
    // unresolve some references?
    this.unresolveReferencesInNodeData(nodedata);
  }

  /**
   * Remove from this model all of the node data held in an Array or in an {@link Iterable} of node data objects.
   * @export
   * @this {Model}
   * @param {Iterable.<Object>|Array.<Object>} coll a collection of node data objects to remove from the {@link #nodeDataArray}
   * @since 1.3
   */
  public removeNodeDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void {
    if (Util.isArray(coll)) {
      const len = Util.arrayLength(coll);
      for (let i = 0; i < len; i++) {
        this.removeNodeData(Util.arrayElt(coll, i));
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const data = it.value;
        this.removeNodeData(data);
      }
    }
  }

  /**
   * Take an Array of node data objects and update {@link #nodeDataArray} without replacing
   * the Array and without replacing any existing node data objects that are identified by key.
   *
   * For node data objects that have the same key value, this makes calls to {@link #setDataProperty}
   * to update the existing node data object.
   * For new keys, this calls {@link #addNodeData} on a copy of the data to add a new node to the model.
   * For existing nodes that have keys that are not present in the given Array,
   * this calls {@link #removeNodeData} to remove the existing node from the model.
   *
   * This method is typically used when GoJS is being used within an application that is maintaining state
   * related to the diagram model. When state is updated, this method can be called to keep the GoJS model synchronized.
   *
   * This method does not conduct a transaction.
   * @export
   * @this {Model}
   * @param {Array.<ObjectData>} arr
   * @since 2.1
   */
  public mergeNodeDataArray(arr: Array<ObjectData>): void {
    if (Util.isArray(arr)) {
      // remember the set of original node keys
      const oldkeys = new Set<Key>(this._mapKeyNodeData.iteratorKeys);

      // for each object in the Array, merge its property values with any existing node data,
      // or else add it to this model
      const newkeys = new Set<Key>();
      const len = Util.arrayLength(arr);
      for (let i = 0; i < len; i++) {
        const obj = Util.arrayElt(arr, i);
        const key = this.getKeyForNodeData(obj);
        if (key !== undefined) {
          newkeys.add(key);
          const nodedata = this.findNodeDataForKey(key);
          if (nodedata !== null) {
            this.assignAllDataProperties(nodedata, obj);
          } else {
            const newdata = {};  // can't depend on copyNodeData due to its not copying references
            this.setKeyForNodeData(newdata, key);
            this.assignAllDataProperties(newdata, obj);
            this.addNodeData(newdata);
          }
        } else {
          this.addNodeData(obj);
          newkeys.add(this.getKeyForNodeData(obj));
        }
      }

      // now remove any node data objects from the model whose keys are not in the given Array
      const it = oldkeys.iterator;
      while (it.next()) {
        const k = it.value;
        if (!newkeys.contains(k)) {
          const olddata = this.findNodeDataForKey(k);
          if (olddata) this.removeNodeData(olddata);
        }
      }
    }
  }

  // These internal methods are overridden by subclasses

  /**
   * @hidden @internal
   * When a node data changes its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * The goal is to maintain all of the original relationships in all of the data maintained by the model.
   * @param {string|number|undefined} oldkey
   * @param {string|number} newkey
   * @this {Model}
   */
  protected replaceAllKeyReferences(oldkey: Key, newkey: Key): void {
    if (newkey === undefined) return;
    const coll = this.findDelayedReferences(oldkey);
    if (coll instanceof Set) {
      this._delayedNodeReferences.add(newkey, coll);
    }
  }

  /**
   * @hidden @internal
   * This is overridden by GraphLinksModel, since that is the only kind of model that
   * has data objects other than node data to have key references to node data.
   * @this {Model}
   */
  protected unresolveAllReferencesToNodeData(): void {
    // nothing to do, Model data do not support any kind of references to nodes
  }

  /**
   * @hidden @internal
   * Find any delayed references to the given node data by its key,
   * notify listeners that that reference now exists,
   * and finally forget that there were any delayed references to this data/key.
   * This is overridden by subclasses.
   * @this {Model}
   * @param {Object} nodedata
   */
  protected resolveReferencesToNodeData(nodedata: ObjectData): void {
    // nothing to do, Model data do not support any kind of references to nodes
  }

  /**
   * @hidden @internal
   * For each reference in the given node data to node data that does not (yet?) exist in this model,
   * delay that reference by remembering it, by calling {@link #delayReference}.
   * This is overridden by subclasses.
   * @this {Model}
   * @param {Object} nodedata
   */
  protected resolveReferencesInNodeData(nodedata: ObjectData): void {
    // nothing to do, Model data do not support any kind of references from nodes
  }

  /**
   * @hidden @internal
   * Forget any delayed references to node data that this node data has within it,
   * by calling {@link #forgetDelayedReference} on them.
   * This is overridden by subclasses.
   * @this {Model}
   * @param {Object} nodedata
   */
  protected unresolveReferencesInNodeData(nodedata: ObjectData): void {
    // nothing to do, Model data do not support any kind of references from nodes
  }

  /**
   * @hidden @internal
   * Remember that a given object has a node reference using a given key,
   * but that the node data is not presently in the model.
   * @this {Model}
   * @param {string|number|undefined} key
   * @param {Object} data Either node data or link data.
   * @see {@link #forgetDelayedReference}
   * @see {@link #findDelayedReferences}
   */
  public delayReference(key: Key, data: ObjectData): void {
    if (key === undefined) return;
    let coll = this._delayedNodeReferences.getValue(key);
    if (coll === null) {
      coll = new Set<ObjectData>();
      this._delayedNodeReferences.add(key, coll);
    }
    coll.add(data);
  }

  /**
   * @hidden @internal
   * Forget a data object that may have had a reference to a node using the given key.
   * This is called when the data is being removed from the model,
   * and we want to make sure there is no memory of that data object.
   * @this {Model}
   * @param {string|number|undefined} key
   * @param {*=} data Either node data or link data.
   * @see {@link #delayReference}
   * @see {@link #findDelayedReferences}
   */
  public forgetDelayedReference(key: Key, data?: ObjectData): void {
    if (key === undefined) return;
    const coll = this._delayedNodeReferences.getValue(key);
    if (coll instanceof Set) {
      if (data === undefined || data === null) {
        this._delayedNodeReferences.remove(key);
      } else {
        coll.remove(data);
        if (coll.count === 0) {
          this._delayedNodeReferences.remove(key);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {string|number|undefined} key
   * @return {Set.<Object>} a set of data objects that each have one or more references to the given key.
   * @see {@link #delayReference}
   * @see {@link #forgetDelayedReference}
   */
  public findDelayedReferences(key: Key): Set<ObjectData> | null {
    if (key === undefined) return null;
    const coll = this._delayedNodeReferences.getValue(key);
    if (coll instanceof Set) return coll;
    return null;
  }

  /**
   * @hidden @internal
   * @export
   * @this {Model}
   * @param {string|number=} key
   * @see {@link #delayReference}
   */
  public clearUnresolvedReferences(key?: Key): void {
    if (key === undefined) {
      this._delayedNodeReferences.clear();
    } else {
      this._delayedNodeReferences.remove(key);
    }
  }

  /**
   * Gets or sets a function that makes a copy of a node data object.
   *
   * You may need to set this property in order to ensure that a copied {@link Node} is bound
   * to data that does not share certain data structures between the original node data and the copied node data.
   * This property value may be null in order to cause {@link #copyNodeData} to make a shallow copy of a JavaScript Object.
   * The default value is null.
   *
   * The first argument to the function will be a node data object (potentially a {@link Part}'s {@link Panel#data}).
   * The second argument to the function will be this Model itself.
   *
   * It is common to implement a copying function when the node data has an Array of data
   * and that Array needs to be copied rather than shared.
   * Often the objects that are in the Array also need to be copied.
   */
  /** @export */ get copyNodeDataFunction(): ((data: ObjectData, model: Model) => ObjectData) | null { return this._copyNodeDataFunction; }
  /** @export */ set copyNodeDataFunction(value: ((data: ObjectData, model: Model) => ObjectData) | null) {
    const old = this._copyNodeDataFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Model, 'copyNodeDataFunction');
      this._copyNodeDataFunction = value;
      this.raiseChanged('copyNodeDataFunction', old, value);
    }
  }

  /**
   * Gets or sets whether the default behavior for {@link #copyNodeData} or {@link GraphLinksModel#copyLinkData}
   * makes copies of property values that are Arrays.
   * This only copies Arrays that are top-level property values in data objects, not for Arrays that are in nested objects.
   * Copying Arrays will also copy any array items that are Objects when {@link #copiesArrayObjects} is true.
   *
   * The default value is false.
   * It is commonplace to set {@link #copiesArrayObjects} to true when setting this property to true.
   * This property does not affect any behavior when the value of {@link #copyNodeDataFunction}
   * or {@link GraphLinksModel#copyLinkDataFunction} has been set to a function.
   *
   * Caution: if you want a copied data object to share some Arrays but not others,
   * you will need to provide your own copying function as {@link #copyNodeDataFunction} rather than setting
   * this property to true.
   *
   * Warning: there should not be any cyclical references within the model data.
   * @see {@link #copiesArrayObjects}
   * @see {@link #copyNodeDataFunction}
   * @since 1.5
   */
  /** @export */ get copiesArrays(): boolean { return this._copiesArrays; }
  /** @export */ set copiesArrays(value: boolean) {
    const old = this._copiesArrays;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'boolean', Model, 'copiesArrays');
      this._copiesArrays = value;
      this.raiseChanged('copiesArrays', old, value);
    }
  }

  /**
   * Gets or sets whether the default behavior for {@link #copyNodeData} or {@link GraphLinksModel#copyLinkData}
   * when copying Arrays also copies array items that are Objects.
   * This only covers copying Objects that are items in Arrays that are copied when {@link #copiesArrays} is true.
   * Copying an Object when this property is true also recursively copies any Arrays that are property values.
   * It also assumes that the object's constructor can be called with no arguments.
   *
   * The default value is false.
   * This property does not affect any behavior when the value of {@link #copyNodeDataFunction}
   * or {@link GraphLinksModel#copyLinkDataFunction} has been set to a function.
   * This property has no effect unless {@link #copiesArrays} is true.
   *
   * Caution: if you want a copied data object to share some references but not others,
   * you will need to provide your own copying function as {@link #copyNodeDataFunction} rather than setting
   * this property and {@link #copiesArrays} to true.
   *
   * Warning: there should not be any cyclical references within the model data.
   * @see {@link #copiesArrays}
   * @see {@link #copyNodeDataFunction}
   * @since 1.5
   */
  /** @export */ get copiesArrayObjects(): boolean { return this._copiesArrayObjects; }
  /** @export */ set copiesArrayObjects(value: boolean) {
    const old = this._copiesArrayObjects;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'boolean', Model, 'copiesArrayObjects');
      this._copiesArrayObjects = value;
      this.raiseChanged('copiesArrayObjects', old, value);
    }
  }

  /**
   * Gets or sets whether the default behavior for {@link #copyNodeData} or {@link GraphLinksModel#copyLinkData}
   * when copying properties of a data object also copies the key property value.
   * Set this to false in order to force a unique key generation for data copied from another Diagram, such as a Palette.
   *
   * The default value is true.
   * This property does not affect any behavior when the value of {@link #copyNodeDataFunction} has been set to a function.
   *
   * @see {@link #copiesArrays}
   * @see {@link #copyNodeDataFunction}
   * @since 2.0
   */
  /** @export */ get copiesKey(): boolean { return this._copiesKey; }
  /** @export */ set copiesKey(value: boolean) {
    const old = this._copiesKey;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'boolean', Model, 'copiesKey');
      this._copiesKey = value;
      this.raiseChanged('copiesKey', old, value);
    }
  }

  /**
   * Make a copy of a node data object.
   * This uses the value of {@link #copyNodeDataFunction} to actually perform the copy, unless that property is null.
   * When it is null the default behavior is to just make a shallow copy of the JavaScript Object.
   *
   * However when {@link #copiesArrays} is true, this will make a copy of property values that are JavaScript Arrays.
   * This is useful when you do not want the Arrays to be shared between the node data objects.
   * Note that if you want to copy some property values that are Arrays but not other properties that are Arrays,
   * you cannot use {@link #copiesArrays} but must implement your own {@link #copyNodeDataFunction}.
   *
   * In addition when {@link #copiesArrayObjects} is true, if items in the Array being copied are JavaScript Objects,
   * those objects are copied, recursively.
   * This is useful when the items in copied Arrays are themselves objects that need to be copied.
   * Note that if you want to share references to some but not all of the objects in copied Arrays you cannot use
   * {@link #copiesArrayObjects} but must implement your own {@link #copyNodeDataFunction}.
   *
   * This does not modify the model -- the returned data object is not added to this model.
   * This assumes that the data's constructor can be called with no arguments.
   *
   * Models should not have any references to Diagrams or GraphObjects or Tools or Layouts or other objects
   * that form a {@link Diagram}.
   *
   * Warning: there should not be any cyclical references within the model data, unless you
   * either do not turn on {@link #copiesArrays} or {@link #copiesArrayObjects} or unless you have
   * supplied your own {@link #copyNodeDataFunction} that can handle cyclical references.
   * @expose
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {Object}
   * @see {@link #addNodeData}
   */
  public copyNodeData(nodedata: ObjectData): ObjectData | null {
    if (nodedata === null) return null;
    let copy = null;
    // try a programmer-supplied function for copying the data
    const func = this._copyNodeDataFunction;
    if (func !== null) {
      copy = func(nodedata, this);
    } else {  // otherwise assume there's a constructor
      copy = this.internalCopyData(nodedata, true);
    }
    if (Util.isObject(copy)) {
      Set.uniqueHash(copy);
    }
    return copy;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @param {*} val
   * @param {boolean} top
   * @return {*}
   */
  protected internalCopyData(val: any, top: boolean): any {
    if (this.copiesArrays && Array.isArray(val)) {
      const arr = val;
      const arrcopy = [];
      for (let i = 0; i < arr.length; i++) {
        const aval = arr[i];
        const acopy = this.internalCopyData(aval, this.copiesArrayObjects);
        arrcopy.push(acopy);
      }
      Set.uniqueHash(arrcopy);
      return arrcopy;
    } else if (top && Util.isObject(val)) {
      const obj = val;
      const ctor = obj.constructor;
      const objcopy = ctor ? new ctor() : {};
      const keyname = (!this.copiesKey && typeof this.nodeKeyProperty === 'string') ? this.nodeKeyProperty : null;

      for (const p in obj) {
        if (p === '__gohashid') {
          objcopy['__gohashid'] = undefined;  // set same properties in same order
          continue;  // don't want to copy hash id; will set uniqueHash below
        }
        if (p === keyname) {
          objcopy[keyname] = undefined;  // set same properties in same order
          continue;  // don't copy key value
        }
        const pval = Model.safePropertyValue(obj, p);
        if (this.checkDataReference(obj, p, pval)) {  // don't copy Diagram or any piece of a Diagram
          Model.safePropertySet(objcopy, p, pval);
        } else {  // try copying Array or Object
          const pcopy = this.internalCopyData(pval, false);
          Model.safePropertySet(objcopy, p, pcopy);
        }
      }
      Set.uniqueHash(objcopy);
      return objcopy;
    } else if (val instanceof Point) {  // make real copies of basic geometric classes
      return val.copy();
    } else if (val instanceof Size) {
      return val.copy();
    } else if (val instanceof Rect) {
      return val.copy();
    } else if (val instanceof Spot) {
      return val.copy();
    } else if (val instanceof Margin) {
      return val.copy();
    } else {
      return val;
    }
  }

  /**
   * @hidden
   * (Undocumented)
   * This function (if not null) is called towards the end of {@link Diagram#copyParts}
   * in order to support custom data copying operations that depend on references between the parts.
   *
   * The first argument to the function is a {@link Map} mapping original data objects to copied data objects.
   * The second argument to the function is this Model, the destination model for the copied parts.
   * The third argument to the function is the source Model, of the original data objects.
   */
  /** @export */ get afterCopyFunction(): ((map: Map<ObjectData, ObjectData>, destModel: Model, srcModel: Model) => void) | null { return this._afterCopyFunction; }
  /** @export */ set afterCopyFunction(value: ((map: Map<ObjectData, ObjectData>, destModel: Model, srcModel: Model) => void) | null) {
    const old = this._afterCopyFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', Model, 'afterCopyFunction');
      this._afterCopyFunction = value;
      this.raiseChanged('afterCopyFunction', old, value);
    }
  }

  /**
   * @hidden @internal
   * Check that there aren't any references to Diagrams in model data.
   * @this {Model}
   * @param {Object} data
   * @param {string} prop
   * @param {*} val
   * @return {boolean} true if VAL is a Diagram or a part of a Diagram
   */
  protected checkDataReference(data: ObjectData, prop: string, val: any): boolean {
    const classname = this.getClassName(val);
    if (classname === 'GraphObject' || classname === 'Diagram' || classname === 'Layer' || classname === 'RowColumnDefinition' || classname === 'AnimationManager' ||
      classname === 'Tool' || classname === 'CommandHandler' || classname === 'Layout' || classname === 'InputEvent' || classname === 'DiagramEvent') {
      if (Debug && prop[0] !== '_') {
        // only warn on properties whose names don't begin with an underscore
        Util.trace('Warning: found GraphObject or Diagram reference when copying model data on property \"' + prop + '\" of data object: ' + data.toString() +
          '  \nModel data should not have any references to a Diagram or any part of a diagram, such as: ' + val.toString());
      }
      return true;  // don't copy
    } else if (val instanceof Model || val instanceof UndoManager || val instanceof Transaction || val instanceof ChangedEvent) {
      return true;  // don't copy
    }
    return false;
  }

  /** @hidden @internal */
  protected static _warnedAboutModifyingGraphObject = false;

  /**
   * Change the value of some property of a node data, a link data, an item data, or the {@link Model#modelData},
   * given a string naming the property and the new value,
   * in a manner that can be undone/redone and that automatically updates any bindings.
   *
   * This gets the old value of the property; if the value is the same as the new value, no side-effects occur.
   * This calls {@link #raiseDataChanged} to notify about the change.
   *
   * Note that it is insufficient to modify an item Array (for example by pushing a new item onto the Array) and
   * then calling `setDataProperty(data, "items", data.items)` because the value of
   * `data.items` is still the same reference.
   * Instead you will want to call {@link #insertArrayItem}, {@link #addArrayItem}, or {@link #removeArrayItem}.
   *
   * If you modify the property that is the {@link #nodeKeyProperty}, this will call {@link #setKeyForNodeData}.
   *
   * If you modify the property that is the {@link #nodeCategoryProperty} or the {@link GraphLinksModel#linkCategoryProperty},
   * this will call {@link #setCategoryForNodeData} or {@link GraphLinksModel#setCategoryForLinkData}.
   * But if the category might change, {@link Binding} sources should not be (or depend in a conversion function on) the category of the data,
   * because then some bindings might be evaluated before or after the category has been changed.
   * @expose
   * @this {Model}
   * @param {Object} data a JavaScript object typically the value of a {@link Panel#data} and represented by a Node, Link, Group, simple Part,
   *                      or item in a {@link Panel#itemArray}; or this model's {@link #modelData}.
   * @param {string} propname a string that is not null or the empty string.
   * @param {*} val the new value for the property.
   * @see {@link #set}
   */
  public setDataProperty(data: ObjectData, propname: string, val: any): void {
    if (Debug) {
      Util.checkPrimitive(data, 'object', Model, 'setDataProperty:data');
      Util.checkPrimitive(propname, 'string', Model, 'setDataProperty:propname');
      if (propname === '') Util.throwError('Model.setDataProperty: property name must not be an empty string when setting ' + data + ' to ' + val);
    }
    if (this.containsNodeData(data)) {
      if (propname === this.nodeKeyProperty) {
        this.setKeyForNodeData(data, val);
      } else if (propname === this.nodeCategoryProperty) {
        this.setCategoryForNodeData(data, val);
        return;
      }
    } else {
      if (!Model._warnedAboutModifyingGraphObject && data instanceof GraphObject) {
        Model._warnedAboutModifyingGraphObject = true;
        Util.trace('Model.setDataProperty is modifying a GraphObject, \"' + data.toString() + '\"');
        Util.trace('  Is that really your intent?');
      }
    }
    // set the property and raise a Property ChangedEvent with the old and new values
    const old = Model.safePropertyValue(data, propname);
    if (old !== val) {
      Model.safePropertySet(data, propname, val);
      this.raiseDataChanged(data, propname, old, val);
    }
  }


  /**
   * A synonym for {@link #setDataProperty}
   * @export
   * @this {Model}
   * @param {Object} data a JavaScript object typically the value of a {@link Panel#data} and represented by a Node, Link, Group, simple Part,
   *                      or item in a {@link Panel#itemArray}; or this model's {@link #modelData}.
   * @param {string} propname a string that is not null or the empty string.
   * @param {*} val the new value for the property.
   * @see {@link #setDataProperty}
   * @since 1.8
   */
  public set(data: ObjectData, propname: string, val: any): void {
    this.setDataProperty(data, propname, val);
  }

  /**
   * This is similar to <code>Object.assign</code>,
   * but safely calls {@link #setDataProperty} for each property other than a key property.
   * This does not delete any properties on the DATA object,
   * although properties may be set to undefined if they are set that way on the PROPS object.
   * @export
   * @this {Model}
   * @param data a data object
   * @param props an Object holding various properties whose values are to be assigned to the DATA object
   */
  public assignAllDataProperties(data: ObjectData, props: ObjectData): void {
    if (!props) return;
    const isnode = this.containsNodeData(data);
    for (const p in props) {
      if (p === '__gohashid') continue;
      if (isnode && p === this.nodeKeyProperty) continue;
      this.setDataProperty(data, p, props[p]);
    }
  }


  /**
   * Add an item at the end of a data array that may be data bound by a {@link Panel} as its {@link Panel#itemArray},
   * in a manner that can be undone/redone and that automatically updates any bindings.
   *
   * This also calls {@link #raiseChangedEvent} to notify all listeners about the {@link ChangedEvent.Insert}.
   *
   * If you want to add a new node or part to the diagram, call {@link #addNodeData}.
   * @export
   * @this {Model}
   * @param {Array.<*>} arr an Array that is the value of some {@link Panel}'s {@link Panel#itemArray}.
   * @param {*} val the new value to be pushed onto the array.
   * @see {@link #insertArrayItem}
   * @since 1.1
   */
  public addArrayItem(arr: Array<any>, val: any): void {
    this.insertArrayItem(arr, -1, val);
  }

  /**
   * Add an item to a data array that may be data bound by a {@link Panel} as its {@link Panel#itemArray},
   * given a new data value and the index at which to insert the new value, in a manner that can be undone/redone and that automatically updates any bindings.
   *
   * This also calls {@link #raiseChangedEvent} to notify all listeners about the {@link ChangedEvent.Insert}.
   *
   * If you want to add a new node or part to the diagram, call {@link #addNodeData}.
   * @export
   * @this {Model}
   * @param {Array.<*>} arr an Array that is the value of some {@link Panel}'s {@link Panel#itemArray}.
   * @param {number} idx the zero-based array index where the new value will be inserted; use -1 to push the new value on the end of the array.
   * @param {*} val the new value to be inserted into the array.
   * @see {@link #removeArrayItem}
   */
  public insertArrayItem(arr: Array<any>, idx: number, val: any): void {
    if (Debug) {
      Util.checkArray(arr, Model, 'insertArrayItem:arr');
      Util.checkRealNumber(idx, Model, 'insertArrayItem:idx');
      if (arr === this._nodeDataArray) {
        Util.throwError('Model.insertArrayItem or Model.addArrayItem should not be called on the Model.nodeDataArray');
      }
    }
    if (idx < 0) idx = Util.arrayLength(arr);
    Util.arrayInsertAt(arr, idx, val);
    this.raiseModelChangedEvent('', ChangedEvent.Insert, '', arr, null, val, null, idx);
  }

  /**
   * Remove an item from a data array that may be data bound by a {@link Panel} as its {@link Panel#itemArray},
   * given the index at which to remove a data value, in a manner that can be undone/redone and that automatically updates any bindings.
   *
   * This also calls {@link #raiseChangedEvent} to notify all listeners about the {@link ChangedEvent.Remove}.
   *
   * If you want to remove a node from the diagram, call {@link #removeNodeData}.
   *
   * Note that there is no version of this method that takes an item value instead of an index into the array.
   * Because item arrays may hold any JavaScript value, including numbers and strings, there may be duplicate entries with that value in the array.
   * To avoid ambiguity, removing an item from an array requires an index.
   * @export
   * @this {Model}
   * @param {Array.<*>} arr an Array that is the value of some {@link Panel}'s {@link Panel#itemArray}.
   * @param {number=} idx the zero-based array index of the data item to be removed from the array;
   *   if not supplied it will remove the last item of the array.
   * @see {@link #insertArrayItem}
   */
  public removeArrayItem(arr: Array<any>, idx?: number): void {
    if (idx === undefined) idx = -1;
    if (Debug) {
      Util.checkArray(arr, Model, 'removeArrayItem:arr');
      Util.checkRealNumber(idx, Model, 'removeArrayItem:idx');
    }
    if (arr === this._nodeDataArray) {
      Util.throwError('Model.removeArrayItem should not be called on the Model.nodeDataArray');
    }
    if (idx === -1) idx = Util.arrayLength(arr) - 1;
    const oldval = Util.arrayElt(arr, idx);
    Util.arrayRemoveAt(arr, idx);
    this.raiseModelChangedEvent('', ChangedEvent.Remove, '', arr, oldval, null, idx, null);
  }

  /**
   * Gets or sets the name of the node data property that returns a string naming that data's category.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return the category name;
   * if the second argument is supplied, the function should modify the node data object so that it has that new category name.
   * The default value is the string 'category', meaning that it expects the data to have a property named 'category' if it cares to name a category.
   * This is used by the diagram to distinguish between different kinds of nodes.
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getCategoryForNodeData} will return an empty string for all node data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getCategoryForNodeData}
   * @see {@link #setCategoryForNodeData}
   */
  /** @export */ get nodeCategoryProperty(): string | ((a: ObjectData, b?: string) => string) { return this._nodeCategoryProperty; }
  /** @export */ set nodeCategoryProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    const old = this._nodeCategoryProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, Model, 'nodeCategoryProperty');
      this._nodeCategoryProperty = value;
      this.raiseChanged('nodeCategoryProperty', old, value);
    }
  }

  /**
   * Find the category of a given node data, a string naming the node template
   * or group template or part template
   * that the {@link Diagram} should use to represent the node data.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {string}
   * @see {@link #nodeCategoryProperty}
   * @see {@link #setCategoryForNodeData}
   */
  public getCategoryForNodeData(nodedata: ObjectData): string {
    if (nodedata === null) return '';
    const cname = this._nodeCategoryProperty;
    if (cname === '') return '';
    const cat = Model.safePropertyValue(nodedata, cname);
    if (cat === undefined) return '';
    if (typeof cat === 'string') return cat;
    Util.throwError('getCategoryForNodeData found a non-string category for ' + nodedata + ': ' + cat);
    return '';
  }

  /**
   * Change the category of a given node data, a string naming the node template
   * or group template or part template
   * that the {@link Diagram} should use to represent the node data.
   *
   * Changing the node template for a node data will cause the existing {@link Node}, {@link Group}, or {@link Part}
   * to be replaced with a new instance of the same class
   * created by copying the new node template and applying any data-bindings.
   * That means that the templates in the {@link Diagram#nodeTemplateMap} or {@link Diagram#groupTemplateMap}
   * must be instances of the same class -- one cannot convert a {@link Node} into a {@link Group} or vice-versa
   * by setting the category.
   *
   * Binding sources should not be (or depend in a conversion function on) the category of the data
   * if you might be modifying the category, because then some bindings might be evaluated
   * before or after the category has been changed.
   * @export
   * @this {Model}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @param {string} cat Must not be null.
   * @see {@link #nodeCategoryProperty}
   * @see {@link #getCategoryForNodeData}
   */
  public setCategoryForNodeData(nodedata: ObjectData, cat: string): void {
    Util.checkPrimitive(cat, 'string', Model, 'setCategoryForNodeData:cat');
    if (nodedata === null) return;
    const cname = this._nodeCategoryProperty;
    if (cname === '') return;
    if (!this.containsNodeData(nodedata)) {
      Model.safePropertySet(nodedata, cname, cat);
      return;
    }
    let old = Model.safePropertyValue(nodedata, cname);
    if (old === undefined) old = '';
    if (old !== cat) {
      Model.safePropertySet(nodedata, cname, cat);
      this.raiseModelChangedEvent('nodeCategory', ChangedEvent.Property, cname, nodedata, old, cat);
    }
  }

  /**
   * @hidden @internal
   * Gets the string describing the type of model, used in {@link Diagram#addPartManager} to map a {@link Model} to
   * its {@link PartManager}. This just returns 'Model'.
   */
/** @export */   public static get type(): string { return 'Model'; }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isTreeModel(): this is ITreeModel {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isGraphLinksModel(): this is IGraphLinksModel {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isGraphModel(): this is IGraphModel {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isLinkDataModel(): this is ILinkDataModel {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isLinkModel(): this is ILinkModel {
    return false;
  }

  /**
   * @hidden @internal
   * @this {Model}
   * @return {boolean}
   */
  public isGroupModel(): this is IGroupModel {
    return false;
  }

  /**
   * @hidden @internal
   *
   * @return {Model}
   */
  public static initDiagramModel(): Model {
    return new Model();
  }

  /**
   * @hidden @internal
   * @param {function()|ObjectData} x
   * @return {string}
   */
  private getClassName(x: any): string {
    if (typeof x === 'function') {
      if (x.className) return x.className;
      if (x.name) return x.name;
    } else if ((typeof x === 'object' && x !== null) && x.constructor) {
      return this.getClassName(x.constructor);
    }
    return typeof x;
    // return null; ?
  }

  /**
   * @hidden @internal
   * @param {string} name
   * @return {string}
   */
  private static getClassForString(name: string): Function {
    if (Model.classMap[name]) return Model.classMap[name];
    return null;
  }

  /**
   * @hidden @internal
   * Registers a class so that Model can serialize it into JSON. This is required for subclasses of Models.
   * @param {Function} classFunc
   * @param {string} name
   */
  public static registerClass(classFunc: Function, name: string) {
    Model.classMap[name] = classFunc;
  }

  /** @hidden @internal */ public static classMap: { [index: string]: Function } = { };

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {string|function(ObjectData,?=):?} prop
   */
  public static safePropertyValue(obj: any, prop: string | ((a: ObjectData, b?: any) => any)): any {
    if (!obj || !prop) return null;
    let v;
    try {
      if (typeof prop === 'function') {
        v = prop(obj);
      } else if (typeof obj.getAttribute === 'function') {
        // handle XML DOM Element
        v = obj.getAttribute(prop);
        if (v === null) v = undefined;
      } else {
        v = obj[prop];
      }
    } catch (ex) {
      if (Debug) Util.trace('property get error: ' + ex.toString());
    }
    return v;
  }

  /**
   * @hidden @internal
   * @param {Object} obj
   * @param {string|function(ObjectData,?=):?} prop
   * @param {*} val
   */
  public static safePropertySet(obj: any, prop: string | Function, val: any): void {
    if (!obj || !prop) return;
    try {
      if (typeof prop === 'function') {
        prop(obj, val);
      } else if (typeof obj.setAttribute === 'function') {
        // handle XML DOM Element
        obj.setAttribute(prop, val);
      } else {
        obj[prop] = val;
      }
    } catch (ex) {
      if (Debug) Util.trace('property set error: ' + ex.toString());
    }
  }

  /**
   * @hidden @internal
   */
  public static initStatic(): void {
    Model.registerClass(Brush, 'Brush');
    Model.registerClass(ChangedEvent, 'ChangedEvent');
    Model.registerClass(Geometry, 'Geometry');
    Model.registerClass(GraphObject, 'GraphObject');
    Model.registerClass(Margin, 'Margin');
    Model.registerClass(Panel, 'Panel');
    Model.registerClass(Point, 'Point');
    Model.registerClass(Rect, 'Rect');
    Model.registerClass(Size, 'Size');
    Model.registerClass(Spot, 'Spot');
    Model.registerClass(Transaction, 'Transaction');
    Model.registerClass(UndoManager, 'UndoManager');
  }

} // #end Model
Model['className'] = 'Model';
Model['fromJSON'] = Model['fromJson'] = Model.fromJson;
Model['safePropertyValue'] = Model.safePropertyValue;
Model['safePropertySet'] = Model.safePropertySet;
Model.initStatic();

/**
 * A Binding describes how to automatically set a property on a {@link GraphObject}
 * to a value of a property of data in the model.
 * The target property name and the data source property name are strings.
 * All name matching is case-sensitive.
 *
 * Register bindings by calling {@link GraphObject#bind} with a new Binding.
 * Existing bindings become read-only, and no new bindings may be added,
 * when a template (a {@link Part}) is copied.
 * Bindings will be shared by all copies of the template's GraphObjects.
 *
 * For example, your node data might be like:
 * ```js
 * { key: 23, say: "hello!" }
 * ```
 *
 * Your simple node template might be like:
 * ```js
 *   var template = new go.Node(go.Panel.Auto);
 *   // . . . define the rest of the Node's visual tree . . .
 *   var txt = new go.TextBlock();
 *   txt.bind(new go.Binding("text", "say"));
 *   template.add(txt);
 *   myDiagram.nodeTemplate = template;
 * ```
 * Using {@link GraphObject.make} it might look like:
 * ```js
 *   var $ = go.GraphObject.make;
 *   myDiagram.nodeTemplate =
 *     $(go.Node, "Auto",
 *       . . .
 *       $(go.TextBlock, new go.Binding("text", "say"))
 *     )
 * ```
 *
 * The data binding causes the {@link TextBlock#text} property of
 * the TextBlock to be set to the value of the data's "say" property.
 * If the value of the "say" property of a particular data object is undefined,
 * the binding is not evaluated: the target property is not set.
 * If there is an error with the binding, you may see a message in the console log.
 * For this reason you may want to explicitly set the initial value for a property
 * when defining the GraphObject, since that value will remain as the default value
 * if the Binding is not evaluated.
 *
 * Bindings are not necessarily evaluated in any particular order.
 * Binding sources should not be (or depend in a conversion function on) the category of the data
 * if you might be modifying the category (e.g. by calling {@link Model#setCategoryForNodeData}),
 * because then some bindings might be evaluated before or after the category has been changed.
 * <h4>Conversions</h4>
 * Sometimes the data value needs to be modified or converted in order
 * to be used as the new value of a {@link GraphObject} property.
 * The most common conversion functions are provided for you --
 * they convert a string to a geometric class:
 * {@link Point.parse}, {@link Size.parse}, {@link Rect.parse},
 * {@link Margin.parse}, {@link Spot.parse}, and {@link Geometry.parse}.
 * But you can easily define your own conversion function.
 *
 * As an example of a conversion function, let's use a function that adds some
 * text prefixing the data property value:
 * ```js
 *   new go.Binding("text", "say", function(v) { return "I say: " + v; })
 * ```
 * Although simple conversions cover almost all binding cases, there are some infrequent uses
 * that are covered by "Advanced Conversions", discussed below.
 * Conversion functions must not have any side-effects.
 * Conversion functions may be called frequently, so they should be fast and avoid allocating memory.
 * The order in which conversion functions are called is not specified and may vary.
 * <h4>OneWay and TwoWay Bindings</h4>
 * By default bindings are {@link Binding.OneWay}.
 * OneWay bindings are evaluated when the {@link Panel#data} property is set
 * or when you call {@link Panel#updateTargetBindings} or {@link Model#setDataProperty}.
 * OneWay bindings only transfer values from the source to the target.
 *
 * TwoWay bindings are evaluated in the source-to-target direction just as OneWay
 * bindings are evaluated.
 * However when the {@link GraphObject} target property is set, the TwoWay
 * bindings are evaluated in the target-to-source direction.
 * There is no point in having a TwoWay binding on a GraphObject property that cannot be set.
 * For efficiency, avoid TwoWay bindings on GraphObject properties that do not change value in your app.
 *
 * You should not have a TwoWay binding with a source that is a node data object's key property,
 * i.e. on the data property whose name is the same as the value of {@link Model#nodeKeyProperty}.
 * Unintentionally changing the node key value to be the same as another node data's key value
 * may cause indeterminate behavior.
 * Furthermore, changing a node data key without changing any references to that node
 * using the key value will result in "dangling" references and inconsistent relationships.
 * You can make that change safely by calling {@link Model#setKeyForNodeData},
 * but not via a data binding.
 *
 * The target-to-source update can also go through a conversion function.
 * The most common back-conversion functions are provided for you.
 * They convert a geometric class to a string:
 * {@link Point.stringify}, {@link Size.stringify}, {@link Rect.stringify},
 * {@link Margin.stringify}, {@link Spot.stringify}, and {@link Geometry.stringify}.
 *
 * It is common to want to update some data properties based on changes to the diagram.
 * For example, as the user changes the {@link Part#location} by dragging a Node,
 * you can automatically keep the node's model data in sync using a TwoWay binding.
 * ```js
 *   new go.Binding("location", "loc", go.Point.parse).makeTwoWay(go.Point.stringify)
 * ```
 * The call to {@link Binding#makeTwoWay} changes the {@link Binding#mode}
 * to be {@link Binding.TwoWay} and specifies the {@link Binding#backConverter}
 * function to be the {@link Point.stringify} static function.
 *
 * Because the Binding is on the whole node (`template`),
 * the target object is the whole {@link Node} and the target property is "location".
 * The value of `data.loc` will be a string representation of the `Node.location` value.
 * <h4>Binding Sources</h4>
 * The target of a Binding is always a property of a {@link GraphObject} or a {@link RowColumnDefinition}.
 * The source of a Binding is normally a property of a data object in the model.
 * But it is also possible to have the source of a Binding be the shared JavaScript object that is the value of {@link Model#modelData}.
 * You can specify such a binding by calling {@link Binding#ofModel}, meaning "a binding of a source that is a property of the Model.modelData".
 *
 * As an example, you might want all Nodes to use the same color.
 * It would be possible but not natural to bind to a property on the node data object, because that property would have to be
 * duplicated on all of the node data objects in the model, and updating the property would mean calling {@link Model#setDataProperty}
 * on each node data object with the same new value.  Furthermore if there happened to be no nodes at all in the model,
 * there would be no place to save the data.  Hence using the shared {@link Model#modelData} object
 * would be the sensible place for that shared information.
 * ```js
 *   new go.Binding("stroke", "strokeColor").ofModel()
 * ```
 * and to set or modify that color one would just call, within a transaction:
 * ```js
 *   model.setDataProperty(model.modelData, "strokeColor", "red");
 * ```
 * That would cause all nodes with that model data binding to be re-evaluated.
 * It is not commonplace to have a TwoWay Binding on "ofModel" Bindings, but that should work.
 * Converters also work with "ofModel" Bindings.
 *
 * And it is also possible to have the source of a Binding be another GraphObject that is in the same Part.
 * You can enable such a binding by calling {@link Binding#ofObject}, meaning "a binding of a source that is a property of a GraphObject".
 * You just have to make sure that object has a unique {@link GraphObject#name} or is the Part itself.
 * The source property on the GraphObject has to be settable, and the Part must have a value for {@link Panel#data}.
 * (If the source property setter does not notify about property value changes, the binding mechanism will not be invoked.
 * Similarly, if there is no Panel.data, the binding mechanism is not active.)
 *
 * As a common kind of example of data binding between two properties of GraphObjects,
 * consider this Binding on a {@link Shape} which changes the color of the {@link Shape#stroke}
 * depending on whether the Node is selected ({@link Part#isSelected}):
 * ```js
 *   new go.Binding("stroke", "isSelected", function(s) { return s ? "dodgerblue" : "gray"; }).ofObject()
 * ```
 * Note the call to {@link Binding#ofObject}, which tells the Binding that it should use as the source
 * a GraphObject with a particular name.  However that name argument is optional -- supplying no name
 * (or supplying an empty string) will cause the binding to operate with the root GraphObject.
 * In this case that would be the Node itself.
 * Now with this binding whenever the value of {@link Part#isSelected} changes, this Shape's stroke changes color.
 * The conversion function is what changes the boolean "isSelected" value to a brush color specifier.
 * <h4>Advanced Conversions</h4>
 * The binding functionality also has more advanced features for less common situations.
 * The source property name may be an empty string, to convert the object as a whole.
 * Conversion functions may take a second argument that takes the object that is bound.
 * For source-to-target conversions, the second argument will be the {@link GraphObject} whose property is bound.
 * For target-to-source (back-)conversions, the second argument will be the source data object and
 * the third argument will be the {@link Model}.
 *
 * Here's an example of a two-way data-binding using two custom conversion functions working with two separate data properties.
 * First we define the two conversion functions.
 * ```js
 *   function toLocation(data, node) {
 *     return new go.Point(data.x, data.y);
 *   }
 *
 *   function fromLocation(loc, data, model) {
 *     model.setDataProperty(data, "x", loc.x);
 *     model.setDataProperty(data, "y", loc.y);
 *   }
 *
 * ```
 *
 * Then to data-bind the default template's {@link Part#location} property
 * to two separate data properties, "x" and "y":
 * ```js
 *   new go.Binding("location", "", toLocation).makeTwoWay(fromLocation)
 * ```
 *
 * An empty string argument for the *sourceprop* parameter indicates
 * that the whole data object should be passed to the `toLocation` function,
 * rather than the value of some property of that data.
 * The return value is used as the new value for the {@link Part#location} property.
 * In almost all cases the second argument is not used.
 * Caution: for efficiency reasons you should try to avoid using an empty source property name.
 * Such bindings will be evaluated much more frequently than ones whose source is a particular property name.
 *
 * The binding works normally for the source-to-target direction.
 * But when the target property is modified it is the source property that is
 * set with the back-converted property value from the target object.
 * Because in this example the source property name is the empty string,
 * and because one cannot replace the whole source data object,
 * any return value from the conversion function is ignored.
 * Instead the conversion function has to modify the data object directly,
 * as this example `fromLocation` function does.
 *
 * Note that because the source property name is the empty string, the binding system will not know
 * which properties are modified in the call to `fromLocation`.
 * Hence to support undo and redo, in order to make the data changes we have to call
 * {@link Model#setDataProperty} so that the {@link UndoManager} can record the change,
 * including the previous value.
 * <h4>Replacing Items in Arrays</h4>
 * However, although a TwoWay Binding cannot replace the node data object in the {@link Model#nodeDataArray},
 * it is possible to replace an item in an {@link Panel#itemArray}.
 * So if your node data were:
 *
 * ```js
 *   { key: 1, items: ["one", "two", "three"] }
 * ```
 *
 * And if your node template included something like:
 * ```js
 *     $(go.Panel, "Vertical",
 *       new go.Binding("itemArray", "items"),
 *       {
 *         itemTemplate:
 *           $(go.Panel,
 *             $(go.TextBlock, { editable: true },
 *               new go.Binding("text", "").makeTwoWay())
 *           )
 *       }
 *     )
 * ```
 *
 * Then the user would be able to edit any of the {@link TextBlock}s, causing the item Array to be modified,
 * for example resulting in this node data:
 * ```js
 *   { key: 1, items: ["one", "SOME NEW TEXT HERE", "three"] }
 * ```
 *
 * @unrestricted
 * @category Model
 */
export class Binding {
  /** @hidden @internal */ public static className = 'Binding';
  /** @hidden @internal */ private _isFrozen: boolean;
  /** @hidden @internal */ private _targetId: number;
  /** @hidden @internal */ private _targetProperty: string;
  /** @hidden @internal */ private _sourceName: string | null;
  /** @hidden @internal */ private _isToModel: boolean;
  /** @hidden @internal */ private _sourceProperty: string;
  /** @hidden @internal */ private _converter: ((val: any, targetObj: any) => void) | null;
  /** @hidden @internal */ private _mode: EnumValue;
  /** @hidden @internal */ private _backConverter: ((val: any, srcData: any, model: any) => void) | null;
  /** @hidden @internal */ private _elementsBeingEvaluated: Set<ObjectData>;
  /** @hidden @internal */ public _targetObject: ObjectData | null;
  /** @hidden @internal */ public _targetKind: number;
  /** @hidden @internal */ public _targetArg: number;

  /**
   * The constructor creates a one-way binding.
   * @param {string=} targetprop A string naming the target property on the target object.
   *   This should not be the empty string.
   * @param {string=} sourceprop A string naming the source property on the bound data object.
   *   If this is the empty string, the whole {@link Panel#data} object is used.
   *   If this argument is not supplied, the source property is assumed to be the same as the target property.
   * @param {function(*,*)|null=} conv A side-effect-free function converting the data property value to the value to set the target property.
   *   If the function is null or not supplied, no conversion takes place.
   */
  constructor(targetprop?: string, sourceprop?: string, conv?: ((val: any, targetObj: any) => void) | null) {
    Set.uniqueHash(this);
    this._isFrozen = false;
    if (targetprop === undefined) targetprop = '';
    else Util.checkPrimitive(targetprop, 'string', Binding, 'constructor:targetprop');
    if (sourceprop === undefined) sourceprop = targetprop;
    else Util.checkPrimitive(sourceprop, 'string', Binding, 'constructor:sourceprop');
    if (conv === undefined) conv = null;
    else if (conv !== null) Util.checkPrimitive(conv, 'function', Binding, 'constructor:conv');
    this._targetId = -1;
    this._targetObject = null;
    this._targetProperty = targetprop;
    this._targetKind = 0;  // 1 for column definitions, 2 for row definitions; only for 1st prop in path
    this._targetArg = 0;  // the column or row number, if targetKind is 1 or 2
    this._sourceName = null;  // ??? do we need a sourceId, like targetId, to uniquely identify source?
    this._isToModel = false;
    this._sourceProperty = sourceprop;
    this._converter = conv;
    this._mode = Binding.OneWay;
    this._backConverter = null;
    this._elementsBeingEvaluated = new Set<ObjectData>();
  }

  /**
   * Create a copy of this Binding, with the same property values.
   * @expose
   * @this {Binding}
   * @return {Binding}
   */
  public copy(): Binding {
    const copy = new Binding();
    // copy._isFrozen = false;
    // copy._targetId = -1;
    // copy._targetObject = null;
    copy._targetProperty = this._targetProperty;
    copy._targetKind = this._targetKind;
    copy._targetArg = this._targetArg;
    copy._sourceName = this._sourceName;
    copy._isToModel = this._isToModel;
    copy._sourceProperty = this._sourceProperty;
    copy._converter = this._converter;
    copy._mode = this._mode;
    copy._backConverter = this._backConverter;
    return copy;
  }

  /**
   * This value for {@link Binding#mode} uses data source values and sets {@link GraphObject} properties.
   * Bindings are evaluated when {@link Panel#updateTargetBindings} is called.
   * @constant
   */
  public static OneWay = new EnumValue(Binding, 'OneWay', 1);

  /**
   * This value for {@link Binding#mode} uses data source values and {@link GraphObject} properties and keeps them in sync.
   * When {@link Panel#updateTargetBindings} is called, the {@link GraphObject} properties are set.
   * When {@link GraphObject} properties are modified, the {@link Panel#data} properties are set.
   * @constant
   */
  public static TwoWay = new EnumValue(Binding, 'TwoWay', 2);

  /**
   * This static function can be used to create a function that parses
   * a string into an enumerated value, given the class that the enumeration values
   * are defined on and a default value if the string cannot be parsed successfully.
   *
   * The normal usage is to pass the result of this function as the conversion function of a {@link Binding}.
   * ```js
   * linktemplate.bind(new go.Binding('routing', 'dataPropName', go.Binding.parseEnum(go.Link, go.Link.Normal)));
   * ```
   * This binding will try to parse the string that is the value of the bound data's "dataPropName" property.
   * If it is a legitimate enumerated value defined on the {@link Link} class, the conversion
   * function will return that value.
   * If the bound data's "dataPropName" property is not present or has an unrecognized value,
   * the {@link Link#routing} property gets the default value, {@link Link.Normal}.
   * @param {function()} ctor the class constructor that defines the enumerated values that are being parsed.
   * @param {EnumValue} defval the default enumerated value to return if it fails to parse the given string.
   * @return {function(string):EnumValue} a function that takes a string and returns an enumerated value.
   */
  public static parseEnum(ctor: any, defval: EnumValue): ((name: string) => EnumValue) {
    Util.checkPrimitive(ctor, 'function', Binding, 'parseEnum:ctor');
    EnumValue.checkEnumValue(defval, ctor, Binding, 'parseEnum:defval');
    return function (str) {
      const v = EnumValue.findName(ctor, str);
      if (v === null) return defval;
      return v;
    };
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {Binding}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === Binding) {
      this.mode = arg;
    } else {
      Util.throwEnumPropertyValueError(this, arg);
    }
  }

  /**
   * This static function can be used to convert an object to a string,
   * looking for commonly defined data properties, such as "text", "name", "key", or "id".
   * If none are found, this just calls toString() on it.
   * @param {*=} val
   * @return {string}
   */
  public static toString(val?: any): string {
    return Util.toString(val);
  }

  /**
   * @hidden @internal
   * @this {Binding}
   * @return {string}
   */
  public toString(): string {
    return 'Binding(' + this.targetProperty + ':' + this.sourceProperty + (this.targetId !== -1 ? ' ' + this.targetId : '') + ' ' + this.mode.name + ')';
  }

  /**
   * @hidden @internal
   * Make this object 'frozen' or immutable.
   * Future attempts to change its state, for example by setting any of its properties,
   * may result in an exception, until it is thaw()'ed.
   * @this {Binding}
   * @return {Binding} this.
   */
  public freeze(): Binding {
    this._isFrozen = true;
    return this;
  }

  /**
   * @hidden @internal
   * Make this object 'thawed' or mutable.
   * @this {Binding}
   * @return {Binding}
   */
  public thaw(): Binding {
    this._isFrozen = false;
    return this;
  }

  /**
   * @hidden @internal
   * Gets or sets the id of the {@link GraphObject} whose property should be set
   * by this data binding.
   * Because Bindings may be shared, the targetId is needed to find in the Part
   * the actual object whose property is data-bound.
   * The default value is -1, which would match no GraphObject id.
   */
  /* noexport */get targetId(): number { return this._targetId; }
  set targetId(value: number) {
    if (this._isFrozen) Util.throwFrozen(this);
    Util.checkPrimitive(value, 'number', Binding, 'targetId');
    this._targetId = value;
  }

  /**
   * Gets or sets the name of the property to be set on the target {@link GraphObject}.
   * The default value is the empty string; you should set this to be the name of a property.
   */
  /** @export */ get targetProperty(): string { return this._targetProperty; }
  /** @export */ set targetProperty(value: string) {
    if (this._isFrozen) Util.throwFrozen(this);
    Util.checkPrimitive(value, 'string', Binding, 'targetProperty');
    this._targetProperty = value;
  }

  /**
   * Gets or sets the name of the {@link GraphObject} that should act as a source object
   * whose property should be gotten by this data binding.
   * The default value is null, which uses the bound {@link Panel#data} as the source.
   * If the value is a string, it should be the name of a {@link GraphObject} in the
   * visual tree of the {@link Panel} that is bound to the data.
   * Use the empty string to refer to the root panel, which is typically the whole {@link Node} or {@link Link},
   * but will be a {@link Panel} if used in a {@link Panel#itemTemplate}.
   * The name must not contain a period.
   *
   * Binding only works if the source property is settable, not on computed or read-only properties,
   * and if it supports notification.
   * The documentation for the GraphObject (or subclass of GraphObject) property will indicate if the property is settable
   * and if it does not notify.
   */
  /** @export */ get sourceName(): string | null { return this._sourceName; }
  /** @export */ set sourceName(value: string | null) {
    if (this._isFrozen) Util.throwFrozen(this);
    if (value !== null) Util.checkPrimitive(value, 'string', Binding, 'sourceName');
    this._sourceName = value;
    if (value !== null) this._isToModel = false;
  }

  /**
   * Gets or sets whether the source data is {@link Model#modelData} rather than a node data or link data object in the model.
   * The default value is false -- the source data object will not be the shared Model.modelData object.
   * @see {@link #ofModel}
   * @since 1.7
   */
  /** @export */ get isToModel(): boolean { return this._isToModel; }
  /** @export */ set isToModel(value: boolean) {
    if (this._isFrozen) Util.throwFrozen(this);
    Util.checkPrimitive(value, 'boolean', Binding, 'isToModel');
    this._isToModel = value;
  }

  /**
   * Gets or sets the name of the property to get from the bound data object,
   * the value of {@link Panel#data}.
   * The default value is the empty string, which results in setting the target
   * property to the whole data object, rather than to a property value of the data object.
   * If {@link #sourceName} is not null, then this property names the settable
   * property on the {@link GraphObject} or {@link RowColumnDefinition}
   * that acts as the source.
   */
  /** @export */ get sourceProperty(): string { return this._sourceProperty; }
  /** @export */ set sourceProperty(value: string) {
    if (this._isFrozen) Util.throwFrozen(this);
    Util.checkPrimitive(value, 'string', Binding, 'sourceProperty');
    this._sourceProperty = value;
  }

  /**
   * Gets or sets a converter function to apply to the data property value
   * in order to produce the value to set to the target property.
   * This conversion function is used in both OneWay and TwoWay bindings,
   * when transferring a value from the source to the target.
   * The default value is null -- no conversion takes place.
   * Otherwise the value should be a function that takes one or two arguments
   * and returns the desired value.
   * However, the return value is ignored when the {@link #targetProperty}
   * is the empty string.
   *
   * Conversion functions must not have any side-effects other than setting
   * the target property.  In particular you should not try to modify the
   * structure of the visual tree in the target GraphObject's Part's visual tree.
   *
   * The function is passed the value from the source
   * (the first argument) and the target {@link GraphObject} (the second argument).
   * If the {@link #targetProperty} is a property name, that property is set to
   * the function's return value.
   * If the {@link #targetProperty} is the empty string, the function should
   * set a property on the second argument, which will be the target GraphObject.
   */
  /** @export */ get converter(): ((val: any, targetObj: any) => void) | null { return this._converter; }
  /** @export */ set converter(value: ((val: any, targetObj: any) => void) | null) {
    if (this._isFrozen) Util.throwFrozen(this);
    if (value !== null) Util.checkPrimitive(value, 'function', Binding, 'converter');
    this._converter = value;
  }

  /**
   * Gets or sets a converter function to apply to the {@link GraphObject} property value
   * in order to produce the value to set to a data property.
   * This conversion function is only used in a TwoWay binding,
   * when transferring a value from the target to the source.
   * The default value is null -- no conversion takes place.
   * Otherwise the value should be a function that takes one or two arguments
   * and returns the desired value.
   * However, the return value is ignored when the {@link #sourceProperty}
   * is the empty string.
   *
   * Conversion functions must not have any side-effects other than setting
   * the source property.
   *
   * The function is passed the value from the target
   * (the first argument), the source {@link Panel#data} object (the second argument),
   * and the {@link Model} (the third argument).
   * If the {@link #sourceProperty} is a property name, that property is set to
   * the function's return value.
   * If the {@link #sourceProperty} is the empty string, the function should
   * modify the second argument, which will be the source data object.
   */
  /** @export */ get backConverter(): ((val: any, srcData: any, model: any) => void) | null { return this._backConverter; }
  /** @export */ set backConverter(value: ((val: any, srcData: any, model: any) => void) | null) {
    if (this._isFrozen) Util.throwFrozen(this);
    if (value !== null) Util.checkPrimitive(value, 'function', Binding, 'backConverter');
    this._backConverter = value;
  }

  /**
   * Gets or sets the directions and frequency in which the binding may be evaluated.
   * The default value is {@link Binding.OneWay}.
   * {@link Binding.TwoWay} is the other choice.
   *
   * Use OneWay bindings to initialize GraphObject properties based on model data,
   * or to modify GraphObject properties when the model data changes with a call to {@link Model#setDataProperty}.
   * Use TwoWay bindings to keep model data in sync with changes to GraphObject properties.
   * For efficiency, avoid TwoWay bindings on GraphObject properties that do not change value in your app.
   *
   * You should not have a TwoWay binding on a node data object's key property.
   */
  /** @export */ get mode(): EnumValue { return this._mode; }
  /** @export */ set mode(value: EnumValue) {
    if (this._isFrozen) Util.throwFrozen(this);
    EnumValue.checkEnumValue(value, Binding, Binding, 'mode');
    this._mode = value;
  }

  /**
   * Modify this Binding to set its {@link #mode} to be {@link Binding.TwoWay}, and
   * provide an optional conversion function to convert {@link GraphObject} property
   * values back to data values, as the value of {@link #backConverter}.
   *
   * Use TwoWay bindings to keep model data in sync with changes to GraphObject properties.
   * For efficiency, avoid TwoWay bindings on GraphObject properties that do not change value in your app.
   * It is typical only to use TwoWay bindings on properties that are modified by tools or commands.
   * Examples include {@link Part#location} by {@link DraggingTool} and
   * {@link TextBlock#text} by {@link TextEditingTool} (only if {@link TextBlock#editable} is true).
   *
   * You should not have a TwoWay binding on a node data object's key property.
   * @export
   * @this {Binding}
   * @param {function(*,*,*) | null=} backconv
   * @return {Binding} this two-way Binding.
   */
  public makeTwoWay(backconv?: ((val: any, srcData: any, model: any) => void) | null): Binding {
    if (backconv === undefined) backconv = null;
    if (backconv !== null) Util.checkPrimitive(backconv, 'function', Binding, 'makeTwoWay');
    this.mode = Binding.TwoWay;
    this.backConverter = backconv;
    return this;
  }

  /**
   * Modify this Binding to set its {@link #sourceName} property so as to identify
   * a {@link GraphObject} in the visual tree of the bound {@link Panel} as the data source,
   * instead of the {@link Panel#data} as the data source.
   *
   * This permits data binding on GraphObject properties, such as {@link Part#isSelected}.
   * Remember that you can reliably data bind only on settable properties, not on read-only or computed properties.
   * @export
   * @this {Binding}
   * @param {string=} srcname the {@link GraphObject#name} of an element in the visual tree of the bound {@link Panel};
   *   use an empty string to refer to the root panel of that visual tree, whose {@link Panel#data} is the bound data.
   * @return {Binding} this Binding to another GraphObject.
   * @see {@link #sourceName}
   * @see {@link #ofModel}
   */
  public ofObject(srcname?: string): Binding {
    if (srcname === undefined) srcname = '';
    if (Debug) Util.checkPrimitive(srcname, 'string', Binding, 'ofObject:srcname');
    this.sourceName = srcname;
    this.isToModel = false;
    return this;
  }

  /**
   * Modify this Binding so that the source is the {@link Model#modelData} object,
   * not a regular node data object or another {@link GraphObject} in the {@link Part}.
   * @export
   * @this {Binding}
   * @return {Binding} this Binding to the {@link Model#modelData} object.
   * @see {@link #isToModel}
   * @see {@link #ofObject}
   * @since 1.7
   */
  public ofModel(): Binding {
    this.sourceName = null;
    this.isToModel = true;
    return this;
  }

  /**
   * @hidden @internal
   * @this {Binding}
   * @param {Panel} binder
   * @param {GraphObject} obj
   * @return {GraphObject}
   */
  public findSourceGraphObject(binder: Panel, obj: GraphObject): GraphObject | null {
    const srcname = this.sourceName;
    let srcelt: GraphObject = null;
    if (srcname === null || srcname === '') srcelt = binder;
    else if (srcname === '/') srcelt = obj.part;
    else if (srcname === '.') srcelt = obj;
    else if (srcname === '..') srcelt = obj.panel;
    else srcelt = binder.findObject(srcname);
    return srcelt;
  }

  /**
   * @hidden @internal
   * @export
   * @this {Binding}
   * @param {Object} elt  the normal target element whose property will be modified.
   * @param {Object} data  the normal source data object that provides the property value.
   * @param {string=} srcprop  an optional property name on the DATA that should be copied;
   *   if supplied this will update the ELT only if the property name matches.
   */
  public updateTarget(elt: ObjectData, data: ObjectData, srcprop?: string): void {
    // might be limited to updating for a particular source property name
    const sprop = this._sourceProperty;
    if (srcprop !== undefined && sprop !== '' && sprop !== srcprop) return;

    // get the target property name
    const tprop = this._targetProperty;
    // if there's a converter but no target property name, no can do!
    const conv = this._converter;
    if (conv === null && tprop === '') {
      Util.trace('Binding error: target property is the empty string: ' + this.toString());
      return;  // can't replace GraphObject!
    }

    // check that the target property name is defined on the target object;
    // "name.prop" syntax is not supported
    if (Debug && typeof tprop === 'string') {
      if (typeof elt['setAttribute'] !== 'function' && tprop.length > 0 &&
        tprop[0] !== '_' && !Util.hasDefinedProperty(elt, tprop)) {
        Util.trace('Binding error: undefined target property: ' + tprop + ' on ' + elt.toString());
      } else if (tprop === 'name' && elt instanceof GraphObject) {
        Util.trace('Binding error: cannot modify GraphObject.name on ' + elt.toString());
      }
    }

    // find the property value on the data; use the DATA itself if the name is empty
    let val = data;
    if (sprop !== '') val = Model.safePropertyValue(data, sprop);
    if (val === undefined) return;  // ?? don't set property if new value is undefined

    // modify the target
    if (conv === null) {
      if (tprop !== '') {
        Model.safePropertySet(elt, tprop, val);
      }
    } else {  // apply conversion function
      try {  // call conversion function safely, with tracing information for debugging
        if (tprop !== '') {
          const cvtval = conv(val, elt);
          if (Debug && cvtval === undefined) {
            Util.trace('Binding warning: conversion function returned undefined when setting target property \"' + tprop + '\" on ' + elt.toString() + ', function is: ' + conv);
          }
          Model.safePropertySet(elt, tprop, cvtval);
        } else {
          conv(val, elt);
        }
      } catch (ex) {
        if (Debug) Util.trace('Binding error: ' + ex.toString() + ' setting target property \"' + tprop + '\" on ' + elt.toString() + ' with conversion function: ' + conv);
      }
    }
  }

  /**
   * @hidden @internal
   * Apply this Binding to given element in order to update
   * a property on a given source data object.
   * @export
   * @this {Binding}
   * @param {Object} elt  the normal target element whose property will be copied.
   * @param {Object} data  the normal source data object that will be modified.
   * @param {string} trgtprop  a property name on the ELT that should be copied;
   *   if supplied this will update the DATA only if the property name matches.
   * @param {Panel} binder
   */
  public updateSource(elt: ObjectData, data: ObjectData, trgtprop: string, binder: Panel): void {
    if (this._mode !== Binding.TwoWay) return;

    // get the target property name
    const tprop = this._targetProperty;
    // might be limited to updating for a particular target property name
    if (trgtprop !== undefined && tprop !== trgtprop) return;

    // if there's a back-converter but no source property name, no can do!
    const sprop = this._sourceProperty;
    const conv = this._backConverter;

    // find the property value on the element; use the ELT itself if the name is empty
    let val = elt;
    if (tprop !== '') val = Model.safePropertyValue(elt, tprop);
    if (val === undefined) return;  // ?? don't set property if new value is undefined

    if (this._elementsBeingEvaluated.contains(elt)) return;

    try {
      this._elementsBeingEvaluated.add(elt);

      const diagram = (binder !== null ? binder.diagram : null);
      const model = (diagram !== null ? diagram.model : null);
      // modify the source data
      if (conv === null) {
        if (sprop !== '') {
          if (model !== null) {
            if (Debug && model.nodeKeyProperty === sprop && model.containsNodeData(data)) {
              Util.trace('Binding error: cannot have TwoWay Binding on node data key property: ' + this.toString());
            }
            model.setDataProperty(data, sprop, val);
          } else {
            Model.safePropertySet(data, sprop, val);
          }
        } else {  // sprop === ''
          // might replace the item in the array
          // can't replace object of nodeDataArray or linkDataArray
          if (model !== null && binder !== null && binder.itemIndex >= 0 && binder.panel !== null && Array.isArray(binder.panel.itemArray)) {
            const idx = binder.itemIndex;
            const arr = binder.panel.itemArray;
            model.removeArrayItem(arr, idx);
            model.insertArrayItem(arr, idx, val);
          }
        }
      } else {  // apply conversion function
        try {  // call conversion function safely, with tracing information for debugging
          if (sprop !== '') {
            const cvtval = conv(val, data, model);
            if (model !== null) {
              if (Debug) {
                if (model.nodeKeyProperty === sprop && model.containsNodeData(data)) {
                  Util.trace('Binding error: cannot have TwoWay Binding on node data key property: ' + this.toString());
                }
                if (cvtval === undefined) {
                  Util.trace(`Binding warning: conversion function returned undefined when setting source property "${sprop}" on ${data.toString()}, function is: ${conv}`);
                }
              }
              model.setDataProperty(data, sprop, cvtval);
            } else {
              Model.safePropertySet(data, sprop, cvtval);
            }
          } else {  // sprop === ''
            const cvtval = conv(val, data, model);  // might have side-effects
            // might replace the item in the array
            // can't replace object of nodeDataArray or linkDataArray
            if (cvtval !== undefined && model !== null && binder !== null && binder.itemIndex >= 0 && binder.panel !== null && Array.isArray(binder.panel.itemArray)) {
              const idx = binder.itemIndex;
              const arr = binder.panel.itemArray;
              // ??? need replaceArrayItem
              model.removeArrayItem(arr, idx);
              model.insertArrayItem(arr, idx, cvtval);
            }
          }
        } catch (ex) {
          if (Debug) Util.trace('Binding error: ' + ex.toString() + ' setting source property \"' + sprop + '\" on ' + data.toString() + ' with conversion function: ' + conv);
        }
      }
    } finally {
      this._elementsBeingEvaluated.remove(elt);
    }
  }

} // #end Binding
Binding['className'] = 'Binding';

Binding['parseEnum'] = Binding.parseEnum;
Binding['toString'] = Util.toString;
Binding['OneWay'] = Binding.OneWay;
Binding['TwoWay'] = Binding.TwoWay;





/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
GraphLinksModel
*/

// import { Binding, IGraphLinksModel, IncrementalData, Key, Model } from './model';

/**
 * GraphLinksModels support links between nodes and grouping nodes and links into subgraphs.
 * GraphLinksModels hold node data and link data in separate arrays.
 * Node data is normally represented in a {@link Diagram} by instances of {@link Node},
 * but they could be represented by simple {@link Part}s or by {@link Group}s.
 * Link data should be represented by instances of {@link Link}.
 *
 * Each link data object is assumed to have two values, one referring to the node that the
 * link is coming from and one that the link is going to.
 * The {@link #linkFromKeyProperty} property names the property on the link data whose value
 * is the key of the "from" node.
 * The {@link #linkToKeyProperty} property names the property on the link data whose value
 * is the key of the "to" node.
 * The default values for these properties are "from" and "to" respectively.
 *
 * For example, one can define a graph consisting of two nodes with one link connecting them:
 * ```js
 *  model.nodeDataArray = [
 *    { key: "Alpha" },
 *    { key: "Beta" }
 *  ];
 *  model.linkDataArray = [
 *    { from: "Alpha", to: "Beta" }
 *  ];
 * ```
 *
 * If you want to have subgraphs in your diagram, where a group node contains some number of nodes and links,
 * you need to declare that some node data actually represent groups,
 * and you need to provide a reference from a member node data to its containing group node data.
 * The {@link #nodeIsGroupProperty} property names the property on a node data that is true
 * if that node data represents a group.
 * The {@link #nodeGroupKeyProperty} property names the property on a node data whose value
 * is the key of the containing group's node data.
 * The default values for these properties are "isGroup" and "group" respectively.
 *
 * For example, one can define a graph consisting of one group containing a subgraph of
 * two nodes connected by a link, with a second link from that group to a third node
 * that is not a member of that group:
 * ```js
 *  model.nodeDataArray = [
 *    { key: "Group1", isGroup: true},
 *    { key: "Alpha", group: "Group1" },
 *    { key: "Beta", group: "Group1" },
 *    { key: "Gamma" }
 *  ];
 *  model.linkDataArray = [
 *    { from: "Alpha", to: "Beta" },
 *    { from: "Group1", to: "Gamma" }
 *  ];
 * ```
 *
 * GraphLinksModels also support distinguishing the "port" element of a node to which
 * a link can connect, at either end of the link.
 * This identification is a string that names the "port" element in the node.
 * However, you need to set the {@link #linkFromPortIdProperty} and/or
 * {@link #linkToPortIdProperty} properties before the model is able to
 * get the "port id" information from the link data.
 *
 * For example, one can define a graph consisting of a "subtraction" node and two inputs and one output.
 * The "subtraction" node has two distinct inputs called "subtrahend" and "minuend";
 * the output is called "difference".
 * ```js
 *  model.linkFromPortIdProperty = "fromPort";  // necessary to remember portIds
 *  model.linkToPortIdProperty = "toPort";
 *  model.nodeDataArray = [
 *    { key: 1, constant: 5 },  // a constant input node
 *    { key: 2, constant: 2 },  // another constant node
 *    { key: 3, operation: "subtract" },
 *    { key: 4, value: 3 }  // the output node
 *  ];
 *  model.linkDataArray = [
 *    { from: 1, to: 3, toPort: "subtrahend" },
 *    { from: 2, to: 3, toPort: "minuend" },
 *    { from: 3, to: 4, fromPort: "difference" }
 *  ];
 * ```
 * In this case links connected to node 3 (which is the subtraction operation)
 * are distinguished by port id.
 * The connections to the other nodes do not have any port identification,
 * presumably because there is only one port on those nodes, representing the node value.
 *
 * Note that there is no requirement that the link data objects have any kind of unique identifier, unlike for node data.
 * There is no expectation that there be references to link data in the model, so there is no need for such an identifier.
 * When there are multiple links connecting two ports, the only way to distinguish the links in the model
 * is by reference to the particular link data object.
 * This is why there are two methods on the Diagram class for Nodes, {@link Diagram#findNodeForKey} and {@link Diagram#findNodeForData},
 * but there is only the one method for Links, {@link Diagram#findLinkForData}.
 *
 * However you may wish to have the model maintain string or number identifiers on the link data just as all models do for node data.
 * To get that behavior, so that you can call {@link #findLinkDataForKey}, you need to set {@link #linkKeyProperty} to be a non-empty string.
 * Just as with the assignment of node keys, you can customize the assignment of link keys by setting
 * {@link #makeUniqueLinkKeyFunction} to a function that returns a unique identifier.
 *
 * This model does not support the modification of whether a node data object is a group.
 *
 * This model cannot detect the modification of the {@link #linkDataArray} array
 * or the modification of any link data object.
 * If you want to add or remove link data from the {@link #linkDataArray},
 * call the {@link #addLinkData} or {@link #removeLinkData} methods.
 * If you want to modify the node a link connects to, call the
 * {@link #setFromKeyForLinkData} and/or {@link #setToKeyForLinkData} methods.
 * If you want to change the membership of a node data in a group,
 * call the {@link #setGroupKeyForNodeData} method.
 * @extends Model
 * @unrestricted
 * @category Model
 */
export class GraphLinksModel extends Model implements IGraphLinksModel {
  /** @hidden @internal */ public static className = 'GraphLinksModel';
  /** @hidden @internal */ private _linkDataArray: Array<ObjectData>;
  /** @hidden @internal */ private _linkDataSet: Set<ObjectData>;
  /** @hidden @internal */ private _mapKeyLinkData: Map<(string | number), ObjectData>;
  /** @hidden @internal */ private _linkKeyProperty: string | ((a: ObjectData, b: Key) => Key);
  /** @hidden @internal */ private _makeUniqueLinkKeyFunction: ((a: GraphLinksModel, b: ObjectData) => Key) | null;
  /** @hidden @internal */ private _copyLinkDataFunction: ((a: ObjectData, b: GraphLinksModel) => ObjectData) | null;
  /** @hidden @internal */ private _archetypeNodeData: ObjectData | null;
  /** @hidden @internal */ private _linkFromKeyProperty: string | ((a: ObjectData, b: Key) => Key);
  /** @hidden @internal */ private _linkToKeyProperty: string | ((a: ObjectData, b: Key) => Key);
  /** @hidden @internal */ private _linkFromPortIdProperty: string | ((a: ObjectData, b: string) => string);
  /** @hidden @internal */ private _linkToPortIdProperty: string | ((a: ObjectData, b: string) => string);
  /** @hidden @internal */ private _linkCategoryProperty: string | ((a: ObjectData, b: string) => string);
  /** @hidden @internal */ private _linkLabelKeysProperty: string | ((a: ObjectData, b: Array<Key>) => Array<Key>);
  /** @hidden @internal */ private _nodeIsGroupProperty: string | ((a: ObjectData, b: boolean) => boolean);
  /** @hidden @internal */ private _nodeGroupKeyProperty: string | ((a: ObjectData, b: Key) => Key);
  /** @hidden @internal */ private _copiesGroupKeyOfNodeData: boolean;

  /**
   * This constructs an empty GraphLinksModel unless one provides arguments as the initial data array values
   * for the {@link Model#nodeDataArray} and {@link GraphLinksModel#linkDataArray} properties.
   * @param {Array.<Object>=} nodedataarray an optional Array containing JavaScript objects to be represented by Nodes.
   * @param {Array.<Object>=} linkdataarray an optional Array containing JavaScript objects to be represented by Links.
   */
  constructor(nodedataarray?: Array<ObjectData>, linkdataarray?: Array<ObjectData>) {
    super();
    if (Debug && arguments.length > 2) {
      Util.throwError('GraphLinksModel constructor can only take two optional arguments, the Array of node data and the Array of link data.');
    }
    this._linkDataArray = [];
    this._linkDataSet = new Set<ObjectData>();  // can't depend on any link data key, so need to use this Set for containsLinkData
    this._mapKeyLinkData = new Map<(string | number), ObjectData>();
    this._linkKeyProperty = '';
    this._makeUniqueLinkKeyFunction = null;
    this._copyLinkDataFunction = null;
    this._archetypeNodeData = null;
    this._linkFromKeyProperty = 'from';
    this._linkToKeyProperty = 'to';
    this._linkFromPortIdProperty = '';
    this._linkToPortIdProperty = '';
    this._linkCategoryProperty = 'category';
    this._linkLabelKeysProperty = '';

    this._nodeIsGroupProperty = 'isGroup';
    this._nodeGroupKeyProperty = 'group';
    this._copiesGroupKeyOfNodeData = false;

    if (nodedataarray !== undefined) this.nodeDataArray = nodedataarray;
    if (linkdataarray !== undefined) this.linkDataArray = linkdataarray;
  }

  /**
   * @hidden @internal
   * Copies properties from this model to the given model, which must be of the same class.
   * This is called by {@link Model#copy}.
   * @expose
   * @this {GraphLinksModel}
   * @param {Model} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    // this._linkDataArray  not copying data
    // this._linkDataSet should not be copied
    // this._mapKeyLinkData should not be copied
    copy._linkKeyProperty = this._linkKeyProperty;
    copy._makeUniqueLinkKeyFunction = this._makeUniqueLinkKeyFunction;
    copy._copyLinkDataFunction = this._copyLinkDataFunction;
    // ??this._archetypeNodeData  not copying data
    copy._linkFromKeyProperty = this._linkFromKeyProperty;
    copy._linkToKeyProperty = this._linkToKeyProperty;
    copy._linkFromPortIdProperty = this._linkFromPortIdProperty;
    copy._linkToPortIdProperty = this._linkToPortIdProperty;
    copy._linkCategoryProperty = this._linkCategoryProperty;
    copy._linkLabelKeysProperty = this._linkLabelKeysProperty;

    copy._nodeIsGroupProperty = this._nodeIsGroupProperty;
    copy._nodeGroupKeyProperty = this._nodeGroupKeyProperty;
    copy._copiesGroupKeyOfNodeData = this._copiesGroupKeyOfNodeData;
  }

  /**
   * @hidden @internal
   * Clear out all references to any model data.
   * This also clears out the {@link UndoManager}.
   * @export
   * @this {GraphLinksModel}
   * @see {@link Model#clear}
   */
  public clear(): void {
    super.clear();
    this._linkDataArray = [];
    this._mapKeyLinkData.clear();
    this._linkDataSet.clear();
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {number=} details Optional: If greater than zero, this includes descriptions of each node data and link data.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    if (details >= 2) {
      return this.toJson();
    } else {
      let s = (this.name !== '' ? this.name : '') + ' GraphLinksModel';
      if (details > 0) {
        s += '\n node data:';
        let arr = this.nodeDataArray;
        let len = Util.arrayLength(arr);
        let i = 0;
        for (i = 0; i < len; i++) {
          const item = Util.arrayElt(arr, i);
          s += ' ' + this.getKeyForNodeData(item) + ':' + Binding.toString(item);
        }
        s += '\n link data:';
        arr = this.linkDataArray;
        len = Util.arrayLength(arr);
        for (i = 0; i < len; i++) {
          const item = Util.arrayElt(arr, i);
          s += ' ' + this.getNodeKeyForLinkData(item, true) + '-->' + this.getNodeKeyForLinkData(item, false);
        }
      }
      return s;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {Transaction} txn
   * @param {boolean} undo
   * @return {IncrementalData}
   */
  protected writeIncrementalModelData(txn: Transaction, undo: boolean): IncrementalData {
    if (Debug && this.linkKeyProperty === '') Util.throwError('GraphLinksModel.linkKeyProperty must not be an empty string for .toIncrementalData() to succeed.');
    let obj = super.writeIncrementalModelData(txn, undo);
    const model = this;
    const newlinks = new Set<ObjectData>();
    const modlinks = new Set<ObjectData>();
    const oldlinks = new Set<ObjectData>();
    const nestedDataPartMap = this._NestedDataPartDataMap;
    txn.changes.each(function (e) {
      if (e.model !== model) return;
      if (e.modelChange === 'linkDataArray') {
        if (e.change === ChangedEvent.Insert) {
          newlinks.add(e.newValue);
        } else if (e.change === ChangedEvent.Remove) {
          oldlinks.add(e.oldValue);
        }
      } else if (model.containsLinkData(e.object)) {
        modlinks.add(e.object);
      } else if (e.object !== null) {  // some data object that isn't link data
        if (e.object !== null && nestedDataPartMap && nestedDataPartMap.contains(e.object)) {
          const partData = nestedDataPartMap.get(e.object);
          partData.each(function(d) {
            if (model.containsLinkData(d)) {
              modlinks.add(d);
            }
          });
        } else {
          const partData = model.findDataInLinkDataArray(e.object);
          partData.each(function(d) {
            modlinks.add(d);
          });
        }
      }
    });
    const newlinkkeys = new Set<Key>();  // just keys, not data objects
    newlinks.each(function (ld) {
      newlinkkeys.add(model.getKeyForLinkData(ld));
      // make sure any new link data gets written out in full by including it in the array of modified data
      if (!undo) modlinks.add(ld);
    });
    const oldlinkkeys = new Set<Key>();  // just keys, not data objects
    oldlinks.each(function (ld) {
      oldlinkkeys.add(model.getKeyForLinkData(ld));
      // undoing a deletion is the same as adding, so make sure it's written out in full
      if (undo) modlinks.add(ld);
    });
    const modlinkarr = model.cloneDeep(modlinks.toArray());

    if (newlinkkeys.count > 0) {
      if (obj === null) obj = {};
      if (undo) {
        obj['removedLinkKeys'] = newlinkkeys.toArray();
      } else {
        obj['insertedLinkKeys'] = newlinkkeys.toArray();
      }
    }
    if (modlinkarr.length > 0) {
      if (obj === null) obj = {};
      obj['modifiedLinkData'] = modlinkarr;
    }
    if (oldlinkkeys.count > 0) {
      if (obj === null) obj = {};
      if (undo) {
        obj['insertedLinkKeys'] = oldlinkkeys.toArray();
      } else {
        obj['removedLinkKeys'] = oldlinkkeys.toArray();
      }
    }

    return obj;
  }

  /**
   * @hidden @internal
   * Write out the persistent properties of this model into a string using JSON syntax property pairs.
   * @this {GraphLinksModel}
   * @return {string}
   */
  protected writeJsonProtected(): string {
    const basestr = super.writeJsonProtected();
    let s = '';
    if (this.linkCategoryProperty !== 'category' && typeof this.linkCategoryProperty === 'string') s += ',\n  \"linkCategoryProperty\": ' + this.quote(this.linkCategoryProperty);
    if (this.linkKeyProperty !== '' && typeof this.linkKeyProperty === 'string') s += ',\n  \"linkKeyProperty\": ' + this.quote(this.linkKeyProperty);
    if (this.linkFromKeyProperty !== 'from' && typeof this.linkFromKeyProperty === 'string') s += ',\n  \"linkFromKeyProperty\": ' + this.quote(this.linkFromKeyProperty);
    if (this.linkToKeyProperty !== 'to' && typeof this.linkToKeyProperty === 'string') s += ',\n  \"linkToKeyProperty\": ' + this.quote(this.linkToKeyProperty);
    if (this.linkFromPortIdProperty !== '' && typeof this.linkFromPortIdProperty === 'string') s += ',\n  \"linkFromPortIdProperty\": ' + this.quote(this.linkFromPortIdProperty);
    if (this.linkToPortIdProperty !== '' && typeof this.linkToPortIdProperty === 'string') s += ',\n  \"linkToPortIdProperty\": ' + this.quote(this.linkToPortIdProperty);
    if (this.linkLabelKeysProperty !== '' && typeof this.linkLabelKeysProperty === 'string') s += ',\n  \"linkLabelKeysProperty\": ' + this.quote(this.linkLabelKeysProperty);
    if (this.nodeIsGroupProperty !== 'isGroup' && typeof this.nodeIsGroupProperty === 'string') s += ',\n  \"nodeIsGroupProperty\": ' + this.quote(this.nodeIsGroupProperty);
    if (this.nodeGroupKeyProperty !== 'group' && typeof this.nodeGroupKeyProperty === 'string') s += ',\n  \"nodeGroupKeyProperty\": ' + this.quote(this.nodeGroupKeyProperty);
    return basestr + s;
  }

  /**
   * @hidden @internal
   * Find the persistent model properties of the given Object and assign them to this model.
   * @this {GraphLinksModel}
   * @param {ObjectData} obj
   */
  protected readJsonProtected(obj: ObjectData): void {
    super.readJsonProtected(obj);
    if (obj['linkKeyProperty']) this.linkKeyProperty = obj['linkKeyProperty'];
    if (obj['linkFromKeyProperty']) this.linkFromKeyProperty = obj['linkFromKeyProperty'];
    if (obj['linkToKeyProperty']) this.linkToKeyProperty = obj['linkToKeyProperty'];
    if (obj['linkFromPortIdProperty']) this.linkFromPortIdProperty = obj['linkFromPortIdProperty'];
    if (obj['linkToPortIdProperty']) this.linkToPortIdProperty = obj['linkToPortIdProperty'];
    if (obj['linkCategoryProperty']) this.linkCategoryProperty = obj['linkCategoryProperty'];
    if (obj['linkLabelKeysProperty']) this.linkLabelKeysProperty = obj['linkLabelKeysProperty'];
    if (obj['nodeIsGroupProperty']) this.nodeIsGroupProperty = obj['nodeIsGroupProperty'];
    if (obj['nodeGroupKeyProperty']) this.nodeGroupKeyProperty = obj['nodeGroupKeyProperty'];
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @return {string}
   */
  protected writeJsonData(): string {
    const basestr = super.writeJsonData();
    const str = ',\n  \"linkDataArray\": ' + this.writeJsonArray(this.linkDataArray, true);
    return basestr + str;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} obj
   */
  protected readJsonData(obj: ObjectData): void {
    super.readJsonData(obj);
    const arr = obj['linkDataArray'];
    if (Array.isArray(arr)) {
      this.replaceJsonObjects(arr);
      this.linkDataArray = arr;
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {Model} newmodel
   * @return {string}
   */
  protected writeJsonDifference(newmodel: Model): string {
    if (!(newmodel instanceof GraphLinksModel)) {
      Util.throwError('Model.computeJsonDifference: newmodel must be a GraphLinksModel');
      return ''; // for TS
    }
    if (Debug && this.linkKeyProperty === '') Util.throwError('GraphLinksModel.linkKeyProperty must not be an empty string for .computeJsonDifference() to succeed.');
    const basestr = super.writeJsonDifference(newmodel);
    this.checkSameProperties(newmodel, 'linkKeyProperty');
    this.checkSameProperties(newmodel, 'linkFromKeyProperty');
    this.checkSameProperties(newmodel, 'linkToKeyProperty');
    this.checkSameProperties(newmodel, 'linkLabelKeysProperty');
    this.checkSameProperties(newmodel, 'nodeIsGroupProperty');
    this.checkSameProperties(newmodel, 'nodeGroupKeyProperty');

    const newlinkkeys = new Set<Key>();
    const modlinkdatas = new Set<ObjectData>();
    const oldlinkkeys = new Set<Key>().addAll(this._mapKeyLinkData.iteratorKeys); // JStoTS: Map not simple Iterable
    const seen = new Map<ObjectData, ObjectData>();
    const arr = newmodel.linkDataArray;
    const len = Util.arrayLength(arr);
    for (let i = 0; i < len; i++) {
      const newobj = Util.arrayElt(arr, i);
      const key = newmodel.getKeyForLinkData(newobj);  // find the link's key, if any
      if (key !== undefined) {
        oldlinkkeys.remove(key);  // can't be an existing key that is deleted
        const oldobj = this.findLinkDataForKey(key);
        if (oldobj === null) {    // not present in this model,
          newlinkkeys.add(key);   // so it is being inserted
          modlinkdatas.add(newobj);  // and remember details in "modified links" collection
        } else {  // existing object in this model
          if (!this.isEqualObject(oldobj, newobj, seen)) {
            modlinkdatas.add(newobj);  // only add to "modified links" collection if it has changed properties
          }
        }
      } else {  // no key in new data object
        this.makeLinkDataKeyUnique(newobj);  // give it a unique key
        const newkey = this.getKeyForLinkData(newobj);
        newlinkkeys.add(newkey);   // record it as a new key
        modlinkdatas.add(newobj);  // and remember details for new link
      }
    }
    let s = basestr;
    if (newlinkkeys.count > 0) s += this._InsertedLinksString + this.writeJsonArray(newlinkkeys.toArray(), true);
    if (modlinkdatas.count > 0) s += this._ModifiedLinksString + this.writeJsonArray(modlinkdatas.toArray(), true);
    if (oldlinkkeys.count > 0) s += this._RemovedLinksString + this.writeJsonArray(oldlinkkeys.toArray(), true);
    return s;
  }

  /** @hidden @internal */ protected _InsertedLinksString = ',\n  \"insertedLinkKeys\": ';
  /** @hidden @internal */ protected _ModifiedLinksString = ',\n  \"modifiedLinkData\": ';
  /** @hidden @internal */ protected _RemovedLinksString = ',\n  \"removedLinkKeys\": ';

  /**
   * @hidden @internal
   * Search for a nested data item within the link data array.
   * This is used when we can't map nested data to link data via diagram panel changes.
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   * @return {Set.<ObjectData>}
   */
  private findDataInLinkDataArray(data: ObjectData): Set<ObjectData> {
    const set = new Set<ObjectData>();
    for (let i = 0; i < this.linkDataArray.length; i ++) {
      const ld = this.linkDataArray[i];
      this.recurseData(data, ld, ld, set);
    }
    return set;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {Transaction} txn
   * @param {boolean} undo
   * @return {string}
   */
  protected writeIncrementalJsonData(txn: Transaction, undo: boolean): string {
    if (Debug && this.linkKeyProperty === '') Util.throwError('GraphLinksModel.linkKeyProperty must not be an empty string for .toIncrementalJson() to succeed.');
    const basestr = super.writeIncrementalJsonData(txn, undo);
    const model = this;
    const newlinks = new Set<ObjectData>();
    const modlinks = new Set<ObjectData>();
    const oldlinks = new Set<ObjectData>();
    const nestedDataPartMap = this._NestedDataPartDataMap;
    txn.changes.each(function (e) {
      if (e.model !== model) return;
      if (e.modelChange === 'linkDataArray') {
        if (e.change === ChangedEvent.Insert) {
          newlinks.add(e.newValue);
        } else if (e.change === ChangedEvent.Remove) {
          oldlinks.add(e.oldValue);
        }
      } else if (model.containsLinkData(e.object)) {
        modlinks.add(e.object);
      } else if (e.object !== null) {  // some data object that isn't link data
        if (e.object !== null && nestedDataPartMap && nestedDataPartMap.contains(e.object)) {
          const partData = nestedDataPartMap.get(e.object);
          partData.each(function(d) {
            if (model.containsLinkData(d)) {
              modlinks.add(d);
            }
          });
        } else {
          const partData = model.findDataInLinkDataArray(e.object);
          partData.each(function(d) {
            modlinks.add(d);
          });
        }
      }
    });
    const newlinkkeys = new Set<Key>();  // just keys, not data objects
    newlinks.each(function (ld) {
      newlinkkeys.add(model.getKeyForLinkData(ld));
      // make sure any new link data gets written out in full by including it in the array of modified data
      if (!undo) modlinks.add(ld);
    });
    const oldlinkkeys = new Set<Key>();  // just keys, not data objects
    oldlinks.each(function (ld) {
      oldlinkkeys.add(model.getKeyForLinkData(ld));
      // undoing a deletion is the same as adding, so make sure it's written out in full
      if (undo) modlinks.add(ld);
    });
    let s = basestr;
    if (newlinkkeys.count > 0) s += (undo ? this._RemovedLinksString : this._InsertedLinksString) + this.writeJsonArray(newlinkkeys.toArray(), true);
    if (modlinks.count > 0) s += this._ModifiedLinksString + this.writeJsonArray(modlinks.toArray(), true);
    if (oldlinkkeys.count > 0) s += (undo ? this._InsertedLinksString : this._RemovedLinksString) + this.writeJsonArray(oldlinkkeys.toArray(), true);
    return s;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} obj
   */
  protected readIncrementalJsonData(obj: ObjectData): void {
    // ensure Model properties are the same
    if (obj['linkCategoryProperty'] !== undefined && obj['linkCategoryProperty'] !== this.linkCategoryProperty ||
        obj['linkKeyProperty'] !== undefined && obj['linkKeyProperty'] !== this.linkKeyProperty ||
        obj['linkFromKeyProperty'] !== undefined && obj['linkFromKeyProperty'] !== this.linkFromKeyProperty ||
        obj['linkToKeyProperty'] !== undefined && obj['linkToKeyProperty'] !== this.linkToKeyProperty ||
        obj['linkFromPortIdProperty'] !== undefined && obj['linkFromPortIdProperty'] !== this.linkFromPortIdProperty ||
        obj['linkToPortIdProperty'] !== undefined && obj['linkToPortIdProperty'] !== this.linkToPortIdProperty ||
        obj['linkLabelKeysProperty'] !== undefined && obj['linkLabelKeysProperty'] !== this.linkLabelKeysProperty ||
        obj['nodeIsGroupProperty'] !== undefined && obj['nodeIsGroupProperty'] !== this.nodeIsGroupProperty ||
        obj['nodeGroupKeyProperty'] !== undefined && obj['nodeGroupKeyProperty'] !== this.nodeGroupKeyProperty) {
      Util.throwError('applyIncrementalJson cannot change Model properties');
    }
    super.readIncrementalJsonData(obj);
    const newkeys = obj['insertedLinkKeys'];
    if (Array.isArray(newkeys)) {
      const len = newkeys.length;
      for (let i = 0; i < len; i++) {
        const k = newkeys[i];
        let ld = this.findLinkDataForKey(k);
        if (ld === null) {
          ld = this.copyLinkData({});
          this.setKeyForLinkData(ld, k);
          this.addLinkData(ld);
        }
      }
    }
    const moddata = obj['modifiedLinkData'];
    if (Array.isArray(moddata)) {
      const len = moddata.length;
      for (let i = 0; i < len; i++) {
        const d = moddata[i];
        const k2 = this.getKeyForLinkData(d);
        const ld = this.findLinkDataForKey(k2);
        if (ld !== null) {
          for (const p in d) {
            if (p === '__gohashid' || p === this.linkKeyProperty) continue;
            this.setDataProperty(ld, p, d[p]);
          }
        }
      }
    }
    const oldkeys = obj['removedLinkKeys'];
    if (Array.isArray(oldkeys)) {
      const len = oldkeys.length;
      for (let i = 0; i < len; i++) {
        const k = oldkeys[i];
        const ld = this.findLinkDataForKey(k);
        if (ld !== null) this.removeLinkData(ld);
      }
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ChangedEvent} e This describes the change that needs to be done.
   * @param {boolean} undo If true, this method should restore the older state, otherwise the newer state.
   */
  public changeState(e: ChangedEvent, undo: boolean): void {
    if (e.change === ChangedEvent.Insert) {
      let idx = e.newParam;
      if (e.modelChange === 'linkDataArray') {
        const linkdata = e.newValue;
        if (Util.isObject(linkdata) && typeof idx === 'number') {
          const key = this.getKeyForLinkData(linkdata);
          if (undo) {
            this._linkDataSet.remove(linkdata);
            if (this._linkDataArray[idx] === linkdata) {
              this._linkDataArray.splice(idx, 1);
            }
            if (key !== undefined) {
              this._mapKeyLinkData.remove(key);
            }
          } else {  // redo Insert
            this._linkDataSet.add(linkdata);
            if (this._linkDataArray[idx] !== linkdata) {
              this._linkDataArray.splice(idx, 0, linkdata);
            }
            if (key !== undefined) {
              this._mapKeyLinkData.add(key, linkdata);
            }
          }
        }  // no-op if idx is null
        return;
      } else if (e.modelChange === 'linkLabelKeys') {
        const arr = this.getLabelKeysForLinkData(e.object);
        if (Array.isArray(arr) && typeof idx === 'number') {
          // avoid duplicates
          if (undo) {
            idx = arr.indexOf(e.newValue);
            if (idx >= 0) {
              arr.splice(idx, 1);
            }
          } else {  // redo Insert
            if (arr.indexOf(e.newValue) < 0) {
              arr.splice(idx, 0, e.newValue);
            }
          }
        }  // no-op if idx is null
        return;
      }  // call base method to handle other modelChange values
    } else if (e.change === ChangedEvent.Remove) {
      let idx = e.oldParam;
      if (e.modelChange === 'linkDataArray') {
        const linkdata = e.oldValue;
        if (Util.isObject(linkdata) && typeof idx === 'number') {
          const key = this.getKeyForLinkData(linkdata);
          if (undo) {
            this._linkDataSet.add(linkdata);
            if (this._linkDataArray[idx] !== linkdata) {
              this._linkDataArray.splice(idx, 0, linkdata);
            }
            if (key !== undefined) {
              this._mapKeyLinkData.add(key, linkdata);
            }
          } else {  // redo Remove
            this._linkDataSet.remove(linkdata);
            if (this._linkDataArray[idx] === linkdata) {
              this._linkDataArray.splice(idx, 1);
            }
            if (key !== undefined) {
              this._mapKeyLinkData.remove(key);
            }
          }
        }  // no-op if idx is null
        return;
      } else if (e.modelChange === 'linkLabelKeys') {
        const arr = this.getLabelKeysForLinkData(e.object);
        if (Array.isArray(arr) && typeof idx === 'number') {
          // avoid duplicates
          if (undo) {
            if (arr.indexOf(e.newValue) < 0) {
              arr.splice(idx, 0, e.newValue);
            }
          } else {  // redo Remove
            idx = arr.indexOf(e.newValue);
            if (idx >= 0) {
              arr.splice(idx, 1);
            }
          }
        }  // no-op if idx is null
        return;
      }  // call base method to handle other modelChange values
    }  // call base method to handle other change values
    // maybe the Model class can handle this change request
    super.changeState(e, undo);
  }

  /**
   * Gets or sets a data object that will be copied and added to the model as a new node data each time there
   * is a link reference (either the "to" or the "from" of a link data) to a node key that does not yet exist in the model.
   *
   * The default value is null -- node data is not automatically copied and added to the model
   * when there is an unresolved reference in a link data.
   * When adding or modifying a link data if there is a "from" or "to" key value for which {@link Model#findNodeDataForKey} returns null,
   * it will call {@link Model#copyNodeData} on this property value and {@link Model#addNodeData} on the result.
   * @since 1.1
   */
  /** @export */ get archetypeNodeData(): ObjectData | null { return this._archetypeNodeData; }
  /** @export */ set archetypeNodeData(value: ObjectData | null) {
    const old = this._archetypeNodeData;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'object', GraphLinksModel, 'archetypeNodeData');
      this._archetypeNodeData = value;
      this.raiseChanged('archetypeNodeData', old, value);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {string|number|undefined} key
   * @return {string|number|undefined}
   */
  protected maybeEnsureLinkReference(key: Key): Key {
    if (key === undefined) return undefined;
    const arch = this._archetypeNodeData;
    if (arch !== null) {
      let data = this.findNodeDataForKey(key);
      if (data === null) {
        data = this.copyNodeData(arch);
        Model.safePropertySet(data, this.nodeKeyProperty, key);
        this.addNodeData(data);
      }
    }
    return key;
  }

  /**
   * Gets or sets the name of the data property that returns
   * the key of the node data that the link data is coming from.
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the key of the link's source node;
   * if the second argument is supplied, the function should modify the link data object so that it has that new key
   * (which may be undefined to refer to no node) as the identifier to the "from" node.
   * The default value is the name 'from', meaning that it expects the data to have a property named 'from' to refer to the link's source node.
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getFromKeyForLinkData} will return undefined for all link data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getFromKeyForLinkData}
   * @see {@link #setFromKeyForLinkData}
   */
  /** @export */ get linkFromKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._linkFromKeyProperty; }
  /** @export */ set linkFromKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._linkFromKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkFromKeyProperty');
      this._linkFromKeyProperty = value;
      this.raiseChanged('linkFromKeyProperty', old, value);
    }
  }

  /**
   * From a link data retrieve a value uniquely identifying the node data
   * from which this link is connected.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {string|number|undefined}  This may return undefined if
   * the link is not coming from any node.
   * @see {@link #linkFromKeyProperty}
   * @see {@link #setFromKeyForLinkData}
   */
  public getFromKeyForLinkData(linkdata: ObjectData): Key {
    return this.getNodeKeyForLinkData(linkdata, true);
  }

  /**
   * Change the node key that the given link data references as the
   * source of the link.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string|number|undefined} key This may be undefined if
   * the link should no longer come from any node.
   * @see {@link #linkFromKeyProperty}
   * @see {@link #getFromKeyForLinkData}
   */
  public setFromKeyForLinkData(linkdata: ObjectData, key: Key): void {
    this.setNodeKeyForLinkData(linkdata, key, true);
  }

  /**
   * Gets or sets the name of the data property that returns
   * the key of the node data that the link data is going to,
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the key of the link's destination node;
   * if the second argument is supplied, the function should modify the link data object so that it has that new key
   * (which may be undefined to refer to no node) as the identifier to the "to" node.
   * The default value is the name 'to', meaning that it expects the data to have a property named 'to' to refer to the link's destination node.
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getToKeyForLinkData} will return undefined for all link data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getToKeyForLinkData}
   * @see {@link #setToKeyForLinkData}
   */
  /** @export */ get linkToKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._linkToKeyProperty; }
  /** @export */ set linkToKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._linkToKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkToKeyProperty');
      this._linkToKeyProperty = value;
      this.raiseChanged('linkToKeyProperty', old, value);
    }
  }

  /**
   * From a link data retrieve a value uniquely identifying the node data
   * to which this link is connected.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {string|number|undefined}  This may return undefined if
   * the link is not going to any node.
   * @see {@link #linkToKeyProperty}
   * @see {@link #setToKeyForLinkData}
   */
  public getToKeyForLinkData(linkdata: ObjectData): Key {
    return this.getNodeKeyForLinkData(linkdata, false);
  }

  /**
   * Change the node key that the given link data references as the
   * destination of the link.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string|number|undefined} key This may be undefined if
   * the link should no longer go to any node.
   * @see {@link #linkToKeyProperty}
   * @see {@link #getToKeyForLinkData}
   */
  public setToKeyForLinkData(linkdata: ObjectData, key: Key): void {
    this.setNodeKeyForLinkData(linkdata, key, false);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {boolean} isfrom
   * @return {Key}
   */
  public getNodeKeyForLinkData(linkdata: ObjectData, isfrom: boolean): Key {
    if (linkdata === null) return undefined;
    const kname = isfrom ? this._linkFromKeyProperty : this._linkToKeyProperty;
    if (kname === '') return undefined;
    const key = Model.safePropertyValue(linkdata, kname);
    if (key === undefined) return undefined;
    if (this.isKeyType(key)) return key;
    Util.throwError((isfrom ? 'FromKey' : 'ToKey') + ' value for link data ' + linkdata + ' is not a number or a string: ' + key);
    return undefined;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {Key} key
   * @param {boolean} isfrom
   */
  public setNodeKeyForLinkData(linkdata: ObjectData, key: Key, isfrom: boolean): void {
    if (key === null) key = undefined;
    if (key !== undefined && !this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', GraphLinksModel, isfrom ? 'setFromKeyForLinkData:key' : 'setToKeyForLinkData:key');
    }
    if (linkdata === null) return;
    const kname = isfrom ? this._linkFromKeyProperty : this._linkToKeyProperty;
    if (kname === '') return;
    key = this.maybeEnsureLinkReference(key);
    if (!this.containsLinkData(linkdata)) {
      Model.safePropertySet(linkdata, kname, key);
      return;
    }
    const old = Model.safePropertyValue(linkdata, kname);
    if (old !== key) {
      this.forgetDelayedReference(old, linkdata);
      Model.safePropertySet(linkdata, kname, key);
      if (this.findNodeDataForKey(key) === null) {
        this.delayReference(key, linkdata);
      }
      this.raiseModelChangedEvent(isfrom ? 'linkFromKey' : 'linkToKey', ChangedEvent.Property, kname, linkdata, old, key);
      if (typeof kname === 'string') this.updateTargetBindings(linkdata, kname);
    }
  }

  /**
   * Gets or sets the name of the data property that returns
   * the optional parameter naming a "port" element on the node that the link data is connected from.
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the string identifier of the link's source port;
   * if the second argument is supplied, the function should modify the link data object so that it has that string as the identifier to the "from" port.
   * The default value is the empty string indicating that one cannot distinguish
   * different logical connection points for any links.
   * The name must not be null nor the value of {@link #linkFromKeyProperty} or {@link #linkToKeyProperty}.
   * If the value is an empty string,
   * {@link #getFromPortIdForLinkData} will return an empty string for all link data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getFromPortIdForLinkData}
   * @see {@link #setFromPortIdForLinkData}
   */
  /** @export */ get linkFromPortIdProperty(): string | ((a: ObjectData, b?: string) => string) { return this._linkFromPortIdProperty; }
  /** @export */ set linkFromPortIdProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    const old = this._linkFromPortIdProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkFromPortIdProperty');
      if (Debug && (value === this.linkFromKeyProperty || value === this.linkToKeyProperty)) {
        Util.throwError('linkFromPortIdProperty name must not be the same as the GraphLinksModel.linkFromKeyProperty or linkToKeyProperty: ' + value);
      }
      this._linkFromPortIdProperty = value;
      this.raiseChanged('linkFromPortIdProperty', old, value);
    }
  }

  /**
   * From a link data retrieve a value identifying the port object of the node
   * from which this link is connected.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {string}  This may return the empty string if
   * there is no particular port parameter information.
   * @see {@link #linkFromPortIdProperty}
   * @see {@link #setFromPortIdForLinkData}
   */
  public getFromPortIdForLinkData(linkdata: ObjectData): string {
    return this.getPortIdForLinkData(linkdata, true);
  }

  /**
   * Change the information that the given link data uses to identify the
   * particular "port" that the link is coming from.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string} portname This may be the empty string if
   * the link should no longer be associated with any particular "port".
   * @see {@link #linkFromPortIdProperty}
   * @see {@link #getFromPortIdForLinkData}
   */
  public setFromPortIdForLinkData(linkdata: ObjectData, portname: string): void {
    this.setPortIdForLinkData(linkdata, portname, true);
  }

  /**
   * Gets or sets the name of the data property that returns
   * the optional parameter naming a "port" element on the node that the link data is connected to.
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the string identifier of the link's destination port;
   * if the second argument is supplied, the function should modify the link data object so that it has that string as the identifier to the "to" port.
   * The default value is the empty string indicating that one cannot distinguish
   * different logical connection points for any links.
   * The name must not be null nor the value of {@link #linkFromKeyProperty} or {@link #linkToKeyProperty}.
   * If the value is an empty string,
   * {@link #getToPortIdForLinkData} will return an empty string for all link data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getToPortIdForLinkData}
   * @see {@link #setToPortIdForLinkData}
   */
  /** @export */ get linkToPortIdProperty(): string | ((a: ObjectData, b?: string) => string) { return this._linkToPortIdProperty; }
  /** @export */ set linkToPortIdProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    const old = this._linkToPortIdProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkToPortIdProperty');
      if (Debug && (value === this.linkFromKeyProperty || value === this.linkToKeyProperty)) {
        Util.throwError('linkFromPortIdProperty name must not be the same as the GraphLinksModel.linkFromKeyProperty or linkToKeyProperty: ' + value);
      }
      this._linkToPortIdProperty = value;
      this.raiseChanged('linkToPortIdProperty', old, value);
    }
  }

  /**
   * From a link data retrieve a value identifying the port object of the node
   * to which this link is connected.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {string}  This may return the empty string if
   * there is no particular port parameter information.
   * @see {@link #linkToPortIdProperty}
   * @see {@link #setToPortIdForLinkData}
   */
  public getToPortIdForLinkData(linkdata: ObjectData): string {
    return this.getPortIdForLinkData(linkdata, false);
  }

  /**
   * Change the information that the given link data uses to identify the
   * particular "port" that the link is going to.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string} portname This may be the empty string if
   * the link should no longer be associated with any particular "port".
   * @see {@link #linkToPortIdProperty}
   * @see {@link #getToPortIdForLinkData}
   */
  public setToPortIdForLinkData(linkdata: ObjectData, portname: string): void {
    this.setPortIdForLinkData(linkdata, portname, false);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {boolean} isfrom
   * @return {string}
   */
  public getPortIdForLinkData(linkdata: ObjectData, isfrom: boolean): string {
    if (linkdata === null) return '';
    const pname = isfrom ? this._linkFromPortIdProperty : this._linkToPortIdProperty;
    if (pname === '') return '';  // not using PortId
    const id = Model.safePropertyValue(linkdata, pname);
    if (id === undefined) return '';
    return id;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {string} portname
   * @param {boolean} isfrom
   */
  public setPortIdForLinkData(linkdata: ObjectData, portname: string, isfrom: boolean): void {
    Util.checkPrimitive(portname, 'string', GraphLinksModel, isfrom ? 'setFromPortIdForLinkData:portname' : 'setToPortIdForLinkData:portname');
    if (linkdata === null) return;
    const pname = isfrom ? this._linkFromPortIdProperty : this._linkToPortIdProperty;
    if (pname === '') return;  // not using PortId
    if (!this.containsLinkData(linkdata)) {
      Model.safePropertySet(linkdata, pname, portname);
      return;
    }
    let old = Model.safePropertyValue(linkdata, pname);
    if (old === undefined) old = '';
    if (old !== portname) {
      Model.safePropertySet(linkdata, pname, portname);
      this.raiseModelChangedEvent(isfrom ? 'linkFromPortId' : 'linkToPortId', ChangedEvent.Property, pname, linkdata, old, portname);
      if (typeof pname === 'string') this.updateTargetBindings(linkdata, pname);
    }
  }

  /**
   * Gets or sets the name of the data property that returns
   * an array of keys of node data that are labels on that link data.
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the array of label node keys for the link;
   * if the second argument is supplied, the function should modify the link data object so that it holds that Array of node keys as references to label nodes.
   * The default value is the empty string: '', meaning that the model does not support links owning label nodes.
   *
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getLabelKeysForLinkData} will return an empty array for all link data objects.
   * You will need to set this property in order to support nodes as link labels.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getLabelKeysForLinkData}
   * @see {@link #setLabelKeysForLinkData}
   */
  /** @export */ get linkLabelKeysProperty(): string | ((a: ObjectData, b?: Array<Key>) => Array<Key>) { return this._linkLabelKeysProperty; }
  /** @export */ set linkLabelKeysProperty(value: string | ((a: ObjectData, b?: Array<Key>) => Array<Key>)) {
    const old = this._linkLabelKeysProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkLabelKeysProperty');
      this._linkLabelKeysProperty = value;
      this.raiseChanged('linkLabelKeysProperty', old, value);
    }
  }

  /**
   * Gets an Array of node key values that identify node data acting as labels on the given link data.
   *
   * This method only works if {@link #linkLabelKeysProperty} has been set to something other than an empty string.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {Array.<(string|number)>} an Array of node keys; an empty Array if the property was not present.
   * @see {@link #linkLabelKeysProperty}
   * @see {@link #setLabelKeysForLinkData}
   * @see {@link #addLabelKeyForLinkData}
   * @see {@link #removeLabelKeyForLinkData}
   */
  public getLabelKeysForLinkData(linkdata: ObjectData): Array<Key> {
    if (linkdata === null) return GraphLinksModel.EmptyArray;  // this is a constant, frozen JavaScript Array
    const kname = this._linkLabelKeysProperty;
    if (kname === '') return GraphLinksModel.EmptyArray;  // not using label nodes
    const keys = Model.safePropertyValue(linkdata, kname);
    if (keys === undefined) return GraphLinksModel.EmptyArray;
    return keys;
  }

  /**
   * Replaces an Array of node key values that identify node data acting as labels on the given link data.
   *
   * This method only works if {@link #linkLabelKeysProperty} has been set to something other than an empty string.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {Array.<(string|number)>} arr an Array of node keys; an empty Array if the property was not present.
   * @see {@link #linkLabelKeysProperty}
   * @see {@link #getLabelKeysForLinkData}
   * @see {@link #addLabelKeyForLinkData}
   * @see {@link #removeLabelKeyForLinkData}
   */
  public setLabelKeysForLinkData(linkdata: ObjectData, arr: Array<Key>): void {
    Util.checkArray(arr, GraphLinksModel, 'setLabelKeysForLinkData:arr');
    if (linkdata === null) return;
    const kname = this._linkLabelKeysProperty;
    if (kname === '') return;  // not using label nodes
    if (!this.containsLinkData(linkdata)) {
      Model.safePropertySet(linkdata, kname, arr);
      return;
    }
    let old = Model.safePropertyValue(linkdata, kname);
    if (old === undefined) old = GraphLinksModel.EmptyArray;
    if (old !== arr) {
      if (Array.isArray(old)) {
        const oldlen = old.length;
        for (let i = 0; i < oldlen; i++) {
          const key = old[i] as Key;
          this.forgetDelayedReference(key, linkdata);
        }
      }
      Model.safePropertySet(linkdata, kname, arr);
      const newlen = arr.length;
      for (let i = 0; i < newlen; i++) {
        const key = arr[i];
        if (this.findNodeDataForKey(key) === null) {
          this.delayReference(key, linkdata);
        }
      }
      this.raiseModelChangedEvent('linkLabelKeys', ChangedEvent.Property, kname, linkdata, old, arr);
      if (typeof kname === 'string') this.updateTargetBindings(linkdata, kname);
    }
  }

  /**
   * Adds a node key value that identifies a node data acting as a new label node on the given link data.
   *
   * This method only works if {@link #linkLabelKeysProperty} has been set to something other than an empty string.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string|number} key a number or string that is the key of the new label node.
   * @see {@link #removeLabelKeyForLinkData}
   * @see {@link #setLabelKeysForLinkData}
   */
  public addLabelKeyForLinkData(linkdata: ObjectData, key: Key): void {
    if (key === null || key === undefined) return;  // no-op
    if (!this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', GraphLinksModel, 'addLabelKeyForLinkData:key');
    }
    if (linkdata === null) return;
    const kname = this._linkLabelKeysProperty;
    if (kname === '') return;  // not using label nodes
    const keys = Model.safePropertyValue(linkdata, kname);
    if (keys === undefined) {
      const karr = [];
      karr.push(key);
      this.setLabelKeysForLinkData(linkdata, karr);
    } else if (Array.isArray(keys)) {
      // already there?
      let idx = keys.indexOf(key);
      if (idx >= 0) return;
      // modify Array
      idx = keys.length;
      keys.push(key);
      if (this.containsLinkData(linkdata)) {
        // handle any references
        if (this.findNodeDataForKey(key) === null) {
          this.delayReference(key, linkdata);
        }
        // notify
        this.raiseModelChangedEvent('linkLabelKeys', ChangedEvent.Insert, kname, linkdata, null, key, null, idx);
      }
    } else {
      Util.throwError(kname + ' property is not an Array; cannot addLabelKeyForLinkData: ' + linkdata);
    }
  }

  /**
   * Removes a node key value that identifies a node data acting as a former label node on the given link data.
   *
   * Removing a reference to a node data from the collection of link label keys
   * does not automatically remove any node data from the model.
   *
   * This method only works if {@link #linkLabelKeysProperty} has been set to something other than an empty string.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @param {string|number} key a number or string that is the key of the label node being removed from the link.
   * @see {@link #addLabelKeyForLinkData}
   * @see {@link #setLabelKeysForLinkData}
   */
  public removeLabelKeyForLinkData(linkdata: ObjectData, key: Key): void {
    if (key === null || key === undefined) return;  // no-op
    if (!this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', GraphLinksModel, 'removeLabelKeyForLinkData:key');
    }
    if (linkdata === null) return;
    const kname = this._linkLabelKeysProperty;
    if (kname === '') return;  // not using label nodes
    const keys = Model.safePropertyValue(linkdata, kname);
    if (Array.isArray(keys)) {
      // not there?
      const idx = keys.indexOf(key);
      if (idx < 0) return;
      // modify Array
      keys.splice(idx, 1);
      if (this.containsLinkData(linkdata)) {
        // handle any references
        this.forgetDelayedReference(key, linkdata);
        // notify
        this.raiseModelChangedEvent('linkLabelKeys', ChangedEvent.Remove, kname, linkdata, key, null, idx, null);
      }
    } else if (keys !== undefined) {
      Util.throwError(kname + ' property is not an Array; cannot removeLabelKeyforLinkData: ' + linkdata);
    }
  }

  // Link data

  /**
   * Gets or sets the array of link data objects that correspond to {@link Link}s in the {@link Diagram}.
   * The initial value is an empty Array.
   */
  /** @export */ get linkDataArray(): Array<ObjectData> { return this._linkDataArray; }
  /** @export */ set linkDataArray(value: Array<ObjectData>) {
    const old = this._linkDataArray;
    if (old !== value) {
      Util.checkArray(value, GraphLinksModel, 'linkDataArray');

      // completely replacing all link data, so are replacing all key information
      this._mapKeyLinkData.clear();
      // Blindly assign unique numbers as keys for each link data.
      // The actual key values are needed for hashing in Diagram.
      const len = value.length;
      for (let i = 0; i < len; i++) {
        const data = value[i];
        if (!(Util.isObject(data))) {
          Util.throwError('GraphLinksModel.linkDataArray must only contain Objects, not: ' + data);
          return;
        }
        Set.hashIdUnique(data);
      }

      this._linkDataArray = value;

      // Find the key for each link data, and remember in a Map.
      if (this.linkKeyProperty !== '') {
        const delayed = new List<ObjectData>();
        for (let i = 0; i < len; i++) {
          const data = value[i];
          const key = this.getKeyForLinkData(data);
          if (key === undefined) {
            delayed.add(data);
          } else if (this._mapKeyLinkData.getValue(key) !== null) {
            // duplicate keys: call makeLinkDataKeyUnique soon
            delayed.add(data);
          } else {
            this._mapKeyLinkData.add(key, data);
          }
        }

        // Some link data didn't have a key value;
        // generate a unique number by modify the data, then remember it.
        // We delay the calls to makeLinkDataKeyUnique to avoid accidentally
        // generating a new key that happens to be the same as one that
        // is in a not-yet-seen link data.
        const dit = delayed.iterator;
        while (dit.next()) {
          const data = dit.value;
          this.makeLinkDataKeyUnique(data);
          const newkey = this.getKeyForLinkData(data);
          if (newkey !== undefined) this._mapKeyLinkData.add(newkey, data);
        }
      }

      // also keep all link data in a Set for quicker implementation of containsLinkData
      const alllinks = new Set<ObjectData>();
      for (let i = 0; i < len; i++) {
        const data = value[i];
        alllinks.add(data);
      }
      this._linkDataSet = alllinks;  // this also discards the old Set

      // Notify about the property change
      this.raiseModelChangedEvent('linkDataArray', ChangedEvent.Property, 'linkDataArray', this, old, value);

      // maybe delay references
      for (let i = 0; i < len; i++) {
        const data = value[i];
        this.resolveReferencesInLinkData(data);
      }
    }
  }

  /**
   * Gets or sets the name of the data property that returns a unique id number or string for each link data object.
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the unique string or number key for that link data object;
   * if the second argument is supplied, the function should modify the link data object so that it has that string or number as the unique key for that link.
   * The default value is the empty string, which means the model will not maintain a key property value on link data objects.
   * The name must not be null.
   *
   * When this property has a value of an empty string (the default value),
   * {@link #getKeyForLinkData} will return undefined, and {@link #findLinkDataForKey} will always return null.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getKeyForLinkData}
   * @since 1.6
   */
  /** @export */ get linkKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._linkKeyProperty; }
  /** @export */ set linkKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._linkKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkKeyProperty');
      this._linkKeyProperty = value;
      // reinitialize the _mapKeyLinkData Map
      this._mapKeyLinkData.clear();
      const len = this.linkDataArray.length;
      for (let i = 0; i < len; i++) {
        const data = this.linkDataArray[i];
        let key = this.getKeyForLinkData(data);
        if (key === undefined) {
          this.makeLinkDataKeyUnique(data);
          key = this.getKeyForLinkData(data);
        }
        if (key !== undefined) this._mapKeyLinkData.add(key, data);
      }
      this.raiseChanged('linkKeyProperty', old, value);
    }
  }

  /**
   * Given a link data object return its unique key: a number or a string.
   * This returns undefined if there is no key value.
   * Unless {@link #linkKeyProperty} is set to a non-empty string, this model
   * will not automatically assign unique key values for link data objects.
   *
   * It is possible to change the key for a link data object by calling {@link #setKeyForLinkData}.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link
   * @return {string|number|undefined}
   * @see {@link #linkKeyProperty}
   * @see {@link #setKeyForLinkData}
   * @see {@link #findLinkDataForKey}
   * @since 1.6
   */
  public getKeyForLinkData(linkdata: ObjectData): Key {
    if (linkdata === null) return undefined;
    const kname = this._linkKeyProperty;
    if (kname === '') return undefined;
    const key = Model.safePropertyValue(linkdata, kname);
    if (key === undefined) return undefined;
    if (this.isKeyType(key)) return key;
    Util.throwError('Key value for link data ' + linkdata + ' is not a number or a string: ' + key);
    return undefined;
  }

  /**
   * Change the unique key of a given link data that is already in this model.
   * The new key value must be unique -- i.e. not in use by another link data object.
   * You can call {@link #findLinkDataForKey} to check if a proposed new key is already in use.
   *
   * If this is called when {@link #linkKeyProperty} is the empty string (i.e. its default value),
   * this method has no effect.
   * If this is called on a link data object that is not (yet) in this model,
   * this unconditionally modifies the property to the new key value.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link
   * @param {string|number} key
   * @see {@link #linkKeyProperty}
   * @see {@link #getKeyForLinkData}
   * @since 1.6
   */
  public setKeyForLinkData(linkdata: ObjectData, key: Key): void {
    if (key === undefined || key === null || !this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', GraphLinksModel, 'setKeyForLinkData:key');
    }
    if (linkdata === null) return;
    const kname = this._linkKeyProperty;
    if (kname === '') return;
    if (!this.containsLinkData(linkdata)) {
      Model.safePropertySet(linkdata, kname, key);
      return;
    }
    const old = Model.safePropertyValue(linkdata, kname);
    if (old !== key) {
      if (this.findLinkDataForKey(key) !== null) {
        // Util.throwError('GraphLinksModel.setKeyForLinkData: trying to replace the key: ' + old + ' with a key used by another link data object: ' + key);
        return;
      }
      Model.safePropertySet(linkdata, kname, key);
      // update the key->data Map
      if (old !== undefined) this._mapKeyLinkData.remove(old);
      this._mapKeyLinkData.add(key, linkdata);
      this.raiseModelChangedEvent('linkKey', ChangedEvent.Property, kname, linkdata, old, key);
      if (typeof kname === 'string') this.updateTargetBindings(linkdata, kname);
    }
  }

  /**
   * Gets or sets a function that returns a unique id number or string for a link data object.
   * This function is called by {@link #makeLinkDataKeyUnique}
   * when a link data object is added to the model, either as part of a new
   * {@link #linkDataArray} or by a call to {@link #addLinkData}, to make sure the value of
   * {@link #getKeyForLinkData} is unique within the model.
   * However it will not be called when {@link #linkKeyProperty} is the default value, an empty string.
   *
   * The value may be null in order to cause {@link #makeLinkDataKeyUnique} behave in the standard manner.
   * (The default value is null.)
   * You may want to supply a function here in order to make sure all of the automatically generated keys
   * are in a particular format.
   * Setting this property after setting {@link #linkDataArray} has no real effect until there is a call
   * to {@link #addLinkData}.
   *
   * If a link data object is already in the model and you want to change its key value,
   * call {@link #setKeyForLinkData} with a new and unique key.
   * @since 1.6
   */
  /** @export */ get makeUniqueLinkKeyFunction(): ((a: GraphLinksModel, b: ObjectData) => Key) | null { return this._makeUniqueLinkKeyFunction; }
  /** @export */ set makeUniqueLinkKeyFunction(value: ((a: GraphLinksModel, b: ObjectData) => Key) | null) {
    const old = this._makeUniqueLinkKeyFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphLinksModel, 'makeUniqueLinkKeyFunction');
      this._makeUniqueLinkKeyFunction = value;
      this.raiseChanged('makeUniqueLinkKeyFunction', old, value);
    }
  }

  /**
   * Given a number or string, find the link data object in this model
   * that uses the given value as its unique key.
   *
   * Unless {@link #linkKeyProperty} is set to a non-empty string, this model
   * will not automatically assign unique key values for link data objects,
   * and thus this method will always return null.
   * @export
   * @this {GraphLinksModel}
   * @param {(string|number|undefined)} key a string or a number.
   * @return {ObjectData} null if the key is not present in the model,
   * or if the key is null or undefined or not a string or number.
   * @see {@link #containsLinkData}
   * @see {@link #getKeyForLinkData}
   * @since 1.6
   */
  public findLinkDataForKey(key: Key): ObjectData | null {
    if (key === null) Util.throwError('GraphLinksModel.findLinkDataForKey:key must not be null');
    if (key === undefined) return null;
    if (!this.isKeyType(key)) return null;
    return this._mapKeyLinkData.getValue(key);
  }

  /**
   * This method is called when a link data object is added to the model to make sure that
   * {@link #getKeyForLinkData} returns a unique key value.
   *
   * The key value should be unique within the set of data managed by this model:
   * {@link #linkDataArray}.
   * If the key is already in use, this will assign an unused number to the
   * {@link #linkKeyProperty} property on the data.
   *
   * If you want to customize the way in which link data gets a unique key,
   * you can set the {@link #makeUniqueKeyFunction} functional property.
   *
   * If the link data object is already in the model and you want to change its key value,
   * call {@link #setKeyForLinkData} and give it a new unique key value.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link
   * @since 1.6
   */
  public makeLinkDataKeyUnique(linkdata: ObjectData): void {
    if (linkdata === null) return;
    const kname = this._linkKeyProperty;
    if (kname === '') return;  // not using key values
    // maybe already has a unique key?
    let key = this.getKeyForLinkData(linkdata);
    if (key !== undefined && !this._mapKeyLinkData.contains(key)) return;
    // try a programmer-supplied function for generating a key
    const func = this._makeUniqueLinkKeyFunction;
    if (func !== null) {
      key = func(this, linkdata);
      // if the return value is unique, set the key on the link data, just to be sure
      if (key !== undefined && key !== null && !this._mapKeyLinkData.contains(key)) {
        Model.safePropertySet(linkdata, kname, key);
        return;
      }
      // otherwise drop through to really make it unique
    }
    // Note that this does NOT call setKeyForLinkData, because this should not notify
    // and should not modify any references using the duplicate key.
    // no key or not unique key
    if (typeof key === 'string') {
      let j = 2;
      while (this._mapKeyLinkData.contains(key + j)) j++;
      Model.safePropertySet(linkdata, kname, key + j);
    } else if (key === undefined || typeof key === 'number') {
      // use negatives to reduce chance of future conflicts
      let j = -this._mapKeyLinkData.count - 1;
      while (this._mapKeyLinkData.contains(j)) j--;
      Model.safePropertySet(linkdata, kname, j);
    } else if (Debug) {
      Util.throwError('GraphLinksModel.getKeyForLinkData returned something other than a string or a number: ' + key);
    }
  }

  /**
   * Decide if a given link data object is in this model, using reference equality.
   *
   * If you do not have a reference to the particular data object that is in the {@link #linkDataArray},
   * you may need to search for it by iterating through that Array, or (more likely),
   * by finding the desired {@link Link} in a {@link Diagram} and getting that link's {@link Panel#data}.
   *
   * Note that because link data are not assumed to be have a unique key property
   * they cannot be found using an index that this model would maintain.
   * However you may choose to provide such a property on the link data objects
   * and maintain your own index.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {boolean}
   * @see {@link #addLinkData}
   * @see {@link #removeLinkData}
   */
  public containsLinkData(linkdata: ObjectData): boolean {
    if (linkdata === null) return false;
    return this._linkDataSet.contains(linkdata);
  }

  /**
   * When you want to add a link to the diagram, call this method with a new data object.
   * This will add that data to the {@link #linkDataArray} and
   * notify all listeners that a new link data object has been inserted into the collection.
   *
   * Presumably the link data object will already have its "from" and "to" node key references set,
   * but it is also possible to set them after the link data is in the model
   * by calling {@link #setFromKeyForLinkData} and {@link #setToKeyForLinkData}.
   *
   * This operation does nothing if the link data is already part of this model's {@link #linkDataArray}.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @see {@link #removeLinkData}
   */
  public addLinkData(linkdata: ObjectData): void {
    if (linkdata === null) return;
    Set.hashIdUnique(linkdata);
    if (!this.containsLinkData(linkdata)) {
      this.doAddLinkData(linkdata, true);
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {boolean} addtoarray  whether to add to the .linkDataArray
   */
  public doAddLinkData(linkdata: ObjectData, addtoarray: boolean): void {
    if (this.linkKeyProperty !== '') {
      let key = this.getKeyForLinkData(linkdata);
      if (key === undefined) {  // no key, need key to be unique
        this.makeLinkDataKeyUnique(linkdata);
        key = this.getKeyForLinkData(linkdata);
      } else {  // has key -- this this link data already in the Map?
        if (this._mapKeyLinkData.getValue(key) === linkdata) return;
        // if not, duplicate key!
        this.makeLinkDataKeyUnique(linkdata);
        key = this.getKeyForLinkData(linkdata);
      }
      if (key === undefined) {
        Util.throwError('GraphLinksModel.makeLinkDataKeyUnique failed on ' + linkdata + '. Data not added to model.');
        return;
      }
      // update the key->data Map
      this._mapKeyLinkData.add(key, linkdata);
    }
    this._linkDataSet.add(linkdata);
    let idx = null;  // leave index as null when not actually adding to the array
    if (addtoarray) {
      // modify linkDataArray
      idx = this._linkDataArray.length;
      this._linkDataArray.splice(idx, 0, linkdata);
    }
    // notify
    this.raiseModelChangedEvent('linkDataArray', ChangedEvent.Insert, 'linkDataArray', this, null, linkdata, null, idx);
    // maybe delay some references
    this.resolveReferencesInLinkData(linkdata);
  }

  /**
   * Add to this model all of the link data held in an Array or in an {@link Iterable} of link data objects.
   * @export
   * @this {GraphLinksModel}
   * @param {Iterable.<Object>|Array.<Object>} coll a collection of link data objects to add to the {@link #linkDataArray}
   * @since 1.3
   */
  public addLinkDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void {
    if (Array.isArray(coll)) {
      const len = coll.length;
      for (let i = 0; i < len; i++) {
        this.addLinkData(coll[i]);
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const data = it.value;
        this.addLinkData(data);
      }
    }
  }

  /**
   * When you want to remove a link from the diagram, call this method with an existing link data object.
   * This will remove that data object from the {@link #linkDataArray} and
   * notify all listeners that a link data object has been removed from the collection.
   *
   * If you do not have a reference to the particular data object that is in the {@link #linkDataArray},
   * you may need to search for it by iterating through that Array, or (more likely),
   * by finding the desired {@link Link} in a {@link Diagram} and getting that link's {@link Panel#data}.
   *
   * Removing a link data from a model does not automatically remove
   * any associated label node data from the model.
   *
   * This operation does nothing if the link data is not present in the {@link #linkDataArray}.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @see {@link #addLinkData}
   */
  public removeLinkData(linkdata: ObjectData): void {
    if (linkdata === null) return;
    this.doRemoveLinkData(linkdata, true);
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata
   * @param {boolean} removefromarray whether to remove from .linkDataArray
   */
  public doRemoveLinkData(linkdata: ObjectData, removefromarray: boolean): void {
    this._linkDataSet.remove(linkdata);
    // remove from key->data Map
    const key = this.getKeyForLinkData(linkdata);
    if (key !== undefined) {
      // update the key->data Map
      this._mapKeyLinkData.remove(key);
    }
    let idx = null;  // leave index as null when not actually removing from the array
    if (removefromarray) {
      // see if already removed
      idx = this._linkDataArray.indexOf(linkdata);
      if (idx < 0) return;
      // modify linkDataArray
      this._linkDataArray.splice(idx, 1);
    }
    // notify
    this.raiseModelChangedEvent('linkDataArray', ChangedEvent.Remove, 'linkDataArray', this, linkdata, null, idx, null);
    // maybe forget some delayed references
    this.unresolveReferencesInLinkData(linkdata);
  }

  /**
   * Remove from this model all of the link data held in an Array or in an {@link Iterable} of link data objects.
   * @export
   * @this {GraphLinksModel}
   * @param {Iterable.<Object>|Array.<Object>} coll a collection of link data objects to remove from the {@link #linkDataArray}
   * @since 1.3
   */
  public removeLinkDataCollection(coll: Iterable<ObjectData> | Array<ObjectData>): void {
    if (Array.isArray(coll)) {
      const len = coll.length;
      for (let i = 0; i < len; i++) {
        this.removeLinkData(coll[i]);
      }
    } else {
      const it = coll.iterator;
      while (it.next()) {
        const data = it.value;
        this.removeLinkData(data);
      }
    }
  }

  /**
   * Take an Array of link data objects and update {@link #linkDataArray} without replacing
   * the Array and without replacing any existing link data objects that are identified by key.
   * This depends on {@link #linkKeyProperty} being a non-empty string.
   *
   * For link data objects that have the same key value, this makes calls to {@link #setDataProperty}
   * to update the existing link data object.
   * For new keys, this calls {@link #addLinkData} on a copy of the data to add a new link to the model.
   * For existing links that have keys that are not present in the given Array,
   * this calls {@link #removeLinkData} to remove the existing link from the model.
   *
   * This method is typically used when GoJS is being used within an application that is maintaining state
   * related to the diagram model. When state is updated, this method can be called to keep the GoJS model synchronized.
   *
   * This method does not conduct a transaction.
   * @export
   * @this {GraphLinksModel}
   * @param {Array.<ObjectData>} arr
   * @since 2.1
   */
  public mergeLinkDataArray(arr: Array<ObjectData>): void {
    if (Debug && this.linkKeyProperty === '') Util.throwError('GraphLinksModel.linkKeyProperty must not be an empty string for .mergeLinkDataArray() to succeed.');
    if (Util.isArray(arr)) {
      // remember the set of original link keys
      const oldkeys = new Set<Key>(this._mapKeyLinkData.iteratorKeys);

      // for each object in the Array, merge its property values with any existing link data,
      // or else add it to this model
      const newkeys = new Set<Key>();
      const len = Util.arrayLength(arr);
      for (let i = 0; i < len; i++) {
        const obj = Util.arrayElt(arr, i);
        const key = this.getKeyForLinkData(obj);
        if (key !== undefined) {
          newkeys.add(key);
          const linkdata = this.findLinkDataForKey(key);
          if (linkdata !== null) {
            this.assignAllDataProperties(linkdata, obj);
          } else {
            const newdata = {};  // can't depend on copyLinkData due to its not copying references
            this.setKeyForLinkData(newdata, key);
            this.assignAllDataProperties(newdata, obj);
            this.addLinkData(newdata);
          }
        } else {
          this.addLinkData(obj);
          newkeys.add(this.getKeyForLinkData(obj));
        }
      }
      // now remove any link data objects from the model whose keys are not in the given Array
      const it = oldkeys.iterator;
      while (it.next()) {
        const k = it.value;
        if (!newkeys.contains(k)) {
          const olddata = this.findLinkDataForKey(k);
          if (olddata) this.removeLinkData(olddata);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   */
  protected resolveReferencesInLinkData(data: ObjectData): void {
    let key = this.getNodeKeyForLinkData(data, true);
    key = this.maybeEnsureLinkReference(key);
    if (this.findNodeDataForKey(key) === null) {
      this.delayReference(key, data);
    }
    key = this.getNodeKeyForLinkData(data, false);
    key = this.maybeEnsureLinkReference(key);
    if (this.findNodeDataForKey(key) === null) {
      this.delayReference(key, data);
    }
    const keyarray = this.getLabelKeysForLinkData(data);
    if (Array.isArray(keyarray)) {
      const len = keyarray.length;
      for (let j = 0; j < len; j++) {
        key = keyarray[j];
        if (this.findNodeDataForKey(key) === null) {
          this.delayReference(key, data);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   */
  protected unresolveReferencesInLinkData(data: ObjectData): void {
    let key = this.getNodeKeyForLinkData(data, true);
    this.forgetDelayedReference(key, data);
    key = this.getNodeKeyForLinkData(data, false);
    this.forgetDelayedReference(key, data);
    const keyarray = this.getLabelKeysForLinkData(data);
    if (Array.isArray(keyarray)) {
      const len = keyarray.length;
      for (let j = 0; j < len; j++) {
        key = keyarray[j];
        this.forgetDelayedReference(key, data);
      }
    }
  }

  /**
   * Gets or sets a function that makes a copy of a link data object.
   *
   * You may need to set this property in order to ensure that a copied {@link Link} is bound
   * to data that does not share certain data structures between the original link data and the copied link data.
   * This property value may be null in order to cause {@link #copyLinkData} to make a shallow copy of a JavaScript Object.
   * The default value is null.
   */
  /** @export */ get copyLinkDataFunction(): ((a: ObjectData, b: GraphLinksModel) => ObjectData) | null { return this._copyLinkDataFunction; }
  /** @export */ set copyLinkDataFunction(value: ((a: ObjectData, b: GraphLinksModel) => ObjectData) | null) {
    const old = this._copyLinkDataFunction;
    if (old !== value) {
      if (value !== null) Util.checkPrimitive(value, 'function', GraphLinksModel, 'copyLinkDataFunction');
      this._copyLinkDataFunction = value;
      this.raiseChanged('copyLinkDataFunction', old, value);
    }
  }

  /**
   * Make a copy of a link data object.
   * This uses the value of {@link #copyLinkDataFunction} to actually perform the copy,
   * unless it is null, in which case this method just makes a shallow copy of the JavaScript Object.
   *
   * This does not modify the model -- the returned data object is not added to this model.
   * This assumes that the data's constructor can be called with no arguments.
   * This also makes sure there is no reference to either the "from" or the "to" node of the original data.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {ObjectData}
   * @see {@link #addLinkData}
   */
  public copyLinkData(linkdata: ObjectData): ObjectData {
    if (linkdata === null) return null;
    let copy = null;
    // try a programmer-supplied function for copying the data
    const func = this._copyLinkDataFunction;
    if (func !== null) {
      copy = func(linkdata, this);
    } else {  // otherwise assume there's a constructor
      copy = this.internalCopyData(linkdata, true);
    }
    if (Util.isObject(copy)) {
      Set.uniqueHash(copy);
      if (this._linkFromKeyProperty !== '') {
        Model.safePropertySet(copy, this._linkFromKeyProperty, undefined);
      }
      if (this._linkToKeyProperty !== '') {
        Model.safePropertySet(copy, this._linkToKeyProperty, undefined);
      }
      // cannot share label keys array
      if (this._linkLabelKeysProperty !== '') {
        Model.safePropertySet(copy, this._linkLabelKeysProperty, []);
      }
    }
    return copy;
  }

  /**
   * Gets or sets the name of the boolean property on node data that indicates
   * whether the data should be represented as a group of nodes and links or as a simple node.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return true if the node data object should be represented by a Group and false otherwise.
   * At the current time the function will not be called to change whether the node is a group or not.
   * The default value is the name 'isGroup', meaning that it expects the data to have a property named 'isGroup' on those node data objects that should be represented by Groups.
   *
   * The value must not be null.
   * If the value is an empty string,
   * {@link #isGroupForNodeData} will return false for all node data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   */
  /** @export */ get nodeIsGroupProperty(): string | ((a: ObjectData, b?: boolean) => boolean) { return this._nodeIsGroupProperty; }
  /** @export */ set nodeIsGroupProperty(value: string | ((a: ObjectData, b?: boolean) => boolean)) {
    const old = this._nodeIsGroupProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'nodeIsGroupProperty');
      this._nodeIsGroupProperty = value;
      this.raiseChanged('nodeIsGroupProperty', old, value);
    }
  }

  /**
   * See if the given node data should be represented as a group or as a simple node.
   *
   * This value must not change as long as the node data is part of the model.
   * At the current time there is no `setIsGroupForNodeData` method.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {boolean}
   * @see {@link #nodeIsGroupProperty}
   */
  public isGroupForNodeData(nodedata: ObjectData): boolean {
    if (nodedata === null) return false;
    const gname = this._nodeIsGroupProperty;
    if (gname === '') return false;
    if (Model.safePropertyValue(nodedata, gname)) return true;
    return false;
  }

  /**
   * @hidden @internal
   * Implement IGroupModel.isGroupForData
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   * @return {boolean}
   */
  public isGroupForData(data: ObjectData): boolean {
    return this.isGroupForNodeData(data);
  }

  // ?? setIsGroupForNodeData

  /**
   * Gets or sets the name of the property on node data that specifies
   * the string or number key of the group data that "owns" that node data.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return the string or number key for the group data object of which the given data object is a member;
   * if the second argument is supplied, the function should modify the node data object so that it has that new key
   * (which may be undefined to refer to no node) as the containing group key for that node.
   * The default value is the name 'group', meaning that it expects the data to have a property named 'group' to refer to any containing group.
   *
   * The value must not be null.
   * If the value is an empty string,
   * {@link #getGroupKeyForNodeData} will return undefined for all node data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getGroupKeyForNodeData}
   * @see {@link #setGroupKeyForNodeData}
   */
  /** @export */ get nodeGroupKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._nodeGroupKeyProperty; }
  /** @export */ set nodeGroupKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._nodeGroupKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'nodeGroupKeyProperty');
      this._nodeGroupKeyProperty = value;
      this.raiseChanged('nodeGroupKeyProperty', old, value);
    }
  }

  /**
   * @hidden @internal
   * Gets or sets whether {@link #copyNodeData} copies the property that is the reference to the containing group data.
   *
   * The default value is false.
   */
  /* noexport */get copiesGroupKeyOfNodeData(): boolean { return this._copiesGroupKeyOfNodeData; }
  set copiesGroupKeyOfNodeData(value: boolean) {
    const old = this._copiesGroupKeyOfNodeData;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', GraphLinksModel, 'copiesGroupKeyOfNodeData');
      this._copiesGroupKeyOfNodeData = value;
      // ???this.raiseChanged('copiesGroupKeyOfNodeData', old, value);
    }
  }

  /**
   * If there is a container group for the given node data, return the group's key.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {string|number|undefined}  This returns undefined if there is no containing group data.
   * @see {@link #nodeGroupKeyProperty}
   * @see {@link #setGroupKeyForNodeData}
   */
  public getGroupKeyForNodeData(nodedata: ObjectData): Key {
    if (nodedata === null) return undefined;
    const kname = this._nodeGroupKeyProperty;
    if (kname === '') return undefined;
    const key = Model.safePropertyValue(nodedata, kname);
    if (key === undefined) return undefined;
    if (this.isKeyType(key)) return key;
    Util.throwError('GroupKey value for node data ' + nodedata + ' is not a number or a string: ' + key);
    return undefined;
  }

  /**
   * @hidden @internal
   * Implement IGroupModel.getGroupForData
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   * @return {Key}
   */
  public getGroupForData(data: ObjectData): Key {
    return this.getGroupKeyForNodeData(data);
  }

  /**
   * Change the container group for the given node data, given a key for the new group.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata a JavaScript object represented by a node, group, or non-link.
   * @param {string|number|undefined} key This may be undefined if there should be no containing group data.
   * @see {@link #nodeGroupKeyProperty}
   * @see {@link #getGroupKeyForNodeData}
   */
  public setGroupKeyForNodeData(nodedata: ObjectData, key: Key): void {
    if (key === null) key = undefined;
    if (key !== undefined && !this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', GraphLinksModel, 'setGroupKeyForNodeData:key');
    }
    if (nodedata === null) return;
    const kname = this._nodeGroupKeyProperty;
    if (kname === '') return;
    if (!this.containsNodeData(nodedata)) {
      Model.safePropertySet(nodedata, kname, key);
      return;
    }
    const old = Model.safePropertyValue(nodedata, kname);
    if (old !== key) {
      this.forgetDelayedReference(old, nodedata);
      Model.safePropertySet(nodedata, kname, key);
      if (this.findNodeDataForKey(key) === null) {
        this.delayReference(key, nodedata);
      }
      this.raiseModelChangedEvent('nodeGroupKey', ChangedEvent.Property, kname, nodedata, old, key);
      if (typeof kname === 'string') this.updateTargetBindings(nodedata, kname);
    }
  }

  /**
   * @hidden @internal
   * Implement IGroupModel.setGroupForData
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   * @param {Key} key
   */
  public setGroupForData(data: ObjectData, key: Key): void {
    this.setGroupKeyForNodeData(data, key);
  }

  /**
   * This override also makes sure any copied node data does not have a reference to the containing group.
   * @expose
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {ObjectData}
   * @see {@link Model#copyNodeData}
   */
  public copyNodeData(nodedata: ObjectData): ObjectData | null {
    if (nodedata === null) return null;
    const copy = super.copyNodeData(nodedata);
    if (!this.copiesGroupKeyOfNodeData && this._nodeGroupKeyProperty !== ''
      && Model.safePropertyValue(copy, this._nodeGroupKeyProperty) !== undefined) {
      Model.safePropertySet(copy, this._nodeGroupKeyProperty, undefined);
    }
    return copy;
  }

  /**
   * This override changes the value of some property of a node data, a link data, or an item data, given a string naming the property
   * and the new value, in a manner that can be undone/redone and that automatically updates any bindings.
   * This override handles link data as well as node data.
   *
   * This gets the old value of the property; if the value is the same as the new value, no side-effects occur.
   * @expose
   * @this {GraphLinksModel}
   * @param {ObjectData} data a JavaScript object typically the value of a {@link Panel#data} and represented by a Node, Link, Group, simple Part,
   *                      or item in a {@link Panel#itemArray}; or this model's {@link #modelData}.
   * @param {string} propname a string that is not null or the empty string.
   * @param {*} val the new value for the property.
   * @see {@link Model#setDataProperty}
   */
  public setDataProperty(data: ObjectData, propname: string, val: any): void {
    if (Debug) {
      Util.checkPrimitive(data, 'object', GraphLinksModel, 'setDataProperty:data');
      Util.checkPrimitive(propname, 'string', GraphLinksModel, 'setDataProperty:propname');
      if (propname === '') Util.throwError('GraphLinksModel.setDataProperty: property name must not be an empty string when setting ' + data + ' to ' + val);
    }
    if (this.containsNodeData(data)) {
      if (propname === this.nodeKeyProperty) {
        this.setKeyForNodeData(data, val);
      } else if (propname === this.nodeCategoryProperty) {
        this.setCategoryForNodeData(data, val);
        return;
      } else if (propname === this.nodeGroupKeyProperty) {
        this.setGroupKeyForNodeData(data, val);
        return;
      } else if (Debug && propname === this.nodeIsGroupProperty) {
        Util.throwError('GraphLinksModel.setDataProperty: property name must not be the nodeIsGroupProperty: ' + propname);
      }
    } else if (this.containsLinkData(data)) {
      if (propname === this.linkFromKeyProperty) {
        this.setNodeKeyForLinkData(data, val, true);
        return;
      } else if (propname === this.linkToKeyProperty) {
        this.setNodeKeyForLinkData(data, val, false);
        return;
      } else if (propname === this.linkFromPortIdProperty) {
        this.setPortIdForLinkData(data, val, true);
        return;
      } else if (propname === this.linkToPortIdProperty) {
        this.setPortIdForLinkData(data, val, false);
        return;
      } else if (propname === this.linkKeyProperty) {
        this.setKeyForLinkData(data, val);
        return;
      } else if (propname === this.linkCategoryProperty) {
        this.setCategoryForLinkData(data, val);
        return;
      } else if (propname === this.linkLabelKeysProperty) {
        this.setLabelKeysForLinkData(data, val);
        return;
      }
    } else {
      // ???: warn if modifying a GraphObject
      // if (!Model._warnedAboutModifyingGraphObject && data instanceof GraphObject) {
      //   Model._warnedAboutModifyingGraphObject = true;
      //   Util.trace('GraphLinksModel.setDataProperty is modifying a GraphObject, \"' + data.toString() + '\"');
      //   Util.trace('  Is that really your intent?');
      // }
    }
    // set the property and raise a Property ChangedEvent with the old and new values
    const old = Model.safePropertyValue(data, propname);
    if (old !== val) {
      Model.safePropertySet(data, propname, val);
      this.raiseDataChanged(data, propname, old, val);
    }
  }

  /**
   * @hidden
   * This override is similar to <code>Object.assign</code>,
   * but safely calls {@link #setDataProperty} for each property other than a key property.
   * @export
   * @this {GraphLinksModel}
   * @param data a data object
   * @param props an Object holding various properties whose values are to be assigned to the DATA object
   */
  public assignAllDataProperties(data: ObjectData, props: ObjectData): void {
    if (!props) return;
    const isnode = this.containsNodeData(data);
    const islink = this.containsLinkData(data);
    for (const p in props) {
      if (p === '__gohashid') continue;
      if (isnode && p === this.nodeKeyProperty) continue;
      else if (islink && p === this.linkKeyProperty) continue;
      this.setDataProperty(data, p, props[p]);
    }
  }

  // ?? nodeMemberKeysName, getMemberKeysForNodeData, setMemberKeysForNodeData
  //   addMemberKeyForNodeData, removeMemberKeyForNodeData

  /**
   * @hidden @internal
   * When a node data changes its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * The goal is to maintain all of the original relationships in all of the data maintained by the model.
   * @this {GraphLinksModel}
   * @param {string|number|undefined} oldkey
   * @param {string|number} newkey
   */
  protected replaceAllKeyReferences(oldkey: Key, newkey: Key): void {
    super.replaceAllKeyReferences(oldkey, newkey);
    // ?? this is very slow, but it might not be worth it to remember and maintain back-pointers
    const nit = this._mapKeyNodeData.iterator;
    while (nit.next()) {
      const ndata = nit.value;
      this.replaceKeyReferencesInNodeData(ndata, oldkey, newkey);
    }
    const lit = this._linkDataSet.iterator;
    while (lit.next()) {
      const ldata = lit.value;
      this.replaceKeyReferencesInLinkData(ldata, oldkey, newkey);
    }
  }

  /**
   * @hidden @internal
   * When a node data changes its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * @this {GraphLinksModel}
   * @param {ObjectData} d a node data object.
   * @param {string|number=} oldkey
   * @param {string|number=} newkey
   */
  protected replaceKeyReferencesInNodeData(d: ObjectData, oldkey?: Key, newkey?: Key): void {
    const gkey = this.getGroupKeyForNodeData(d);
    if (gkey === oldkey) {
      const kname = this._nodeGroupKeyProperty;
      Model.safePropertySet(d, kname, newkey);
      this.raiseModelChangedEvent('nodeGroupKey', ChangedEvent.Property, kname, d, oldkey, newkey);
      if (typeof kname === 'string') this.updateTargetBindings(d, kname);
    }
  }

  /**
   * @hidden @internal
   * When a node data changes its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * @this {GraphLinksModel}
   * @param {ObjectData} d a link data object.
   * @param {string|number=} oldkey
   * @param {string|number=} newkey
   */
  protected replaceKeyReferencesInLinkData(d: ObjectData, oldkey?: Key, newkey?: Key): void {
    const fkey = this.getNodeKeyForLinkData(d, true);
    if (fkey === oldkey) {
      const kname = this._linkFromKeyProperty;
      Model.safePropertySet(d, kname, newkey);
      this.raiseModelChangedEvent('linkFromKey', ChangedEvent.Property, kname, d, oldkey, newkey);
      if (typeof kname === 'string') this.updateTargetBindings(d, kname);
    }
    const tkey = this.getNodeKeyForLinkData(d, false);
    if (tkey === oldkey) {
      const kname = this._linkToKeyProperty;
      Model.safePropertySet(d, kname, newkey);
      this.raiseModelChangedEvent('linkToKey', ChangedEvent.Property, kname, d, oldkey, newkey);
      if (typeof kname === 'string') this.updateTargetBindings(d, kname);
    }
    const lkeys = this.getLabelKeysForLinkData(d);
    if (Array.isArray(lkeys)) {
      const len = lkeys.length;
      const kname2 = this._linkLabelKeysProperty;
      for (let j = 0; j < len; j++) {
        const lkey = lkeys[j];
        if (lkey === oldkey) {
          lkeys[j] = newkey;
          this.raiseModelChangedEvent('linkLabelKeys', ChangedEvent.Insert, kname2, d, oldkey, newkey, j, j);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   */
  protected unresolveAllReferencesToNodeData(): void {
    super.unresolveAllReferencesToNodeData();
    const array = this.linkDataArray;
    const len = array.length;
    for (let i = 0; i < len; i++) {
      const data = array[i];
      this.resolveReferencesInLinkData(data);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata
   */
  protected resolveReferencesToNodeData(nodedata: ObjectData): void {
    super.resolveReferencesToNodeData(nodedata);
    const key = this.getKeyForNodeData(nodedata);
    const refdata = this.findDelayedReferences(key);
    if (refdata !== null) {
      const removes = Util.tempArray<ObjectData>();
      const dit = refdata.iterator;
      while (dit.next()) {
        const d = dit.value;
        if (this.containsNodeData(d)) {  // D is a node data
          const gkey = this.getGroupKeyForNodeData(d);
          if (gkey === key) {
            const kname = this._nodeGroupKeyProperty;
            this.raiseModelChangedEvent('nodeGroupKey', ChangedEvent.Property, kname, d, key, key);
            if (typeof kname === 'string') this.updateTargetBindings(d, kname);
            removes.push(d);
          }
        } else {  // D is a link data
          const fkey = this.getNodeKeyForLinkData(d, true);
          if (fkey === key) {
            const kname = this._linkFromKeyProperty;
            this.raiseModelChangedEvent('linkFromKey', ChangedEvent.Property, kname, d, key, key);
            if (typeof kname === 'string') this.updateTargetBindings(d, kname);
            removes.push(d);
          }
          const tkey = this.getNodeKeyForLinkData(d, false);
          if (tkey === key) {
            const kname = this._linkToKeyProperty;
            this.raiseModelChangedEvent('linkToKey', ChangedEvent.Property, kname, d, key, key);
            if (typeof kname === 'string') this.updateTargetBindings(d, kname);
            removes.push(d);
          }
          const lkeys = this.getLabelKeysForLinkData(d);
          if (Array.isArray(lkeys)) {
            const len = lkeys.length;
            const kname2 = this._linkLabelKeysProperty;
            for (let j = 0; j < len; j++) {
              const lkey = lkeys[j];
              if (lkey === key) {
                this.raiseModelChangedEvent('linkLabelKeys', ChangedEvent.Insert, kname2, d, key, key, j, j);
                removes.push(d);
              }
            }
          }
        }
      }
      for (let i = 0; i < removes.length; i++) {
        this.forgetDelayedReference(key, removes[i]);
      }
      Util.freeArray(removes);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata
   */
  protected resolveReferencesInNodeData(nodedata: ObjectData): void {
    super.resolveReferencesInNodeData(nodedata);
    const gkey = this.getGroupKeyForNodeData(nodedata);
    if (this.findNodeDataForKey(gkey) === null) {
      this.delayReference(gkey, nodedata);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {GraphLinksModel}
   * @param {ObjectData} nodedata
   */
  protected unresolveReferencesInNodeData(nodedata: ObjectData): void {
    super.unresolveReferencesInNodeData(nodedata);
    const gkey = this.getGroupKeyForNodeData(nodedata);
    this.forgetDelayedReference(gkey, nodedata);
  }

  /**
   * Gets or sets the name of the data property that returns a string naming that data's category,
   * The value may also be a function taking two arguments, where the first argument will be a link data object.
   * If the second argument is not supplied, the function should return the category name;
   * if the second argument is supplied, the function should modify the link data object so that it has that new category name.
   * The default value is the name 'category', meaning that it expects the data to have a property named 'category' if it cares to name the category for the Link.
   * This is used by the diagram to distinguish between different kinds of links.
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getCategoryForLinkData} will return an empty string for all link data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getCategoryForLinkData}
   * @see {@link #setCategoryForLinkData}
   */
  /** @export */ get linkCategoryProperty(): string | ((a: ObjectData, b?: string) => string) { return this._linkCategoryProperty; }
  /** @export */ set linkCategoryProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    const old = this._linkCategoryProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, GraphLinksModel, 'linkCategoryProperty');
      this._linkCategoryProperty = value;
      this.raiseChanged('linkCategoryProperty', old, value);
    }
  }

  /**
   * Find the category of a given link data, a string naming the link template
   * that the {@link Diagram} should use to represent the link data.
   * @export
   * @this {GraphLinksModel}
   * @param {ObjectData} linkdata a JavaScript object represented by a link.
   * @return {string}
   * @see {@link #linkCategoryProperty}
   * @see {@link #setCategoryForLinkData}
   */
  public getCategoryForLinkData(linkdata: ObjectData): string {
    if (linkdata === null) return '';
    const cname = this._linkCategoryProperty;
    if (cname === '') return '';
    const cat = Model.safePropertyValue(linkdata, cname);
    if (cat === undefined) return '';
    if (typeof cat === 'string') return cat;
    Util.throwError('getCategoryForLinkData found a non-string category for ' + linkdata + ': ' + cat);
    return '';
  }

  /**
   * @hidden @internal
   * Implement ILinkModel.getLinkCategoryForData
   * @this {GraphLinksModel}
   * @param {ObjectData} data
   * @return {string}
   */
  public getLinkCategoryForData(data: ObjectData): string {
    return this.getCategoryForLinkData(data);
  }

  /**
   * Change the category of a given link data, a string naming the link template
   * that the {@link Diagram} should use to represent the link data.
   *
   * Changing the link template for a link data will cause the existing {@link Link}
   * to be removed from the {@link Diagram} and be replaced with a new {@link Link}
   * created by copying the new link template and applying any data-bindings.
   * Note that the new template must be an instance of the same class as the original link.
   * Thus one cannot change the category of a link from an instance of {@link Link}
   * to an instance of a subclass of Link, nor vice-versa.
   * @export
   * @this {GraphLinksModel}
   * @param {Object} linkdata a JavaScript object represented by a link.
   * @param {string} cat Must not be null.
   * @see {@link #linkCategoryProperty}
   * @see {@link #getCategoryForLinkData}
   */
  public setCategoryForLinkData(linkdata: ObjectData, cat: string): void {
    Util.checkPrimitive(cat, 'string', GraphLinksModel, 'setCategoryForLinkData:cat');
    if (linkdata === null) return;
    const cname = this._linkCategoryProperty;
    if (cname === '') return;
    if (!this.containsLinkData(linkdata)) {
      Model.safePropertySet(linkdata, cname, cat);
      return;
    }
    let old = Model.safePropertyValue(linkdata, cname);
    if (old === undefined) old = '';
    if (old !== cat) {
      Model.safePropertySet(linkdata, cname, cat);
      this.raiseModelChangedEvent('linkCategory', ChangedEvent.Property, cname, linkdata, old, cat);
      if (typeof cname === 'string') this.updateTargetBindings(linkdata, cname);
    }
  }

  /**
   * @hidden @internal
   * Implement ILinkModel.setLinkCategoryForData
   * @this {GraphLinksModel}
   * @param {Object} data
   * @param {string} cat
   */
  public setLinkCategoryForData(data: ObjectData, cat: string): void {
    this.setCategoryForLinkData(data, cat);
  }

  /**
   * @hidden @internal
   * Gets the string string describing the type of model, used in {@link Diagram#addPartManager} to map a {@link Model} to
   * its {@link PartManager}. This just returns 'GraphLinksModel'.
   */
/** @export */   public static get type(): string { return 'GraphLinksModel'; }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @return {boolean}
   */
  public isGraphLinksModel(): this is IGraphLinksModel {
    return true;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @return {boolean}
   */
  public isLinkDataModel(): boolean {
    return true;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @return {boolean}
   */
  public isLinkModel(): boolean {
    return true;
  }

  /**
   * @hidden @internal
   * @this {GraphLinksModel}
   * @return {boolean}
   */
  public isGroupModel(): boolean {
    return true;
  }

  /** @hidden @internal @const */
  public static EmptyArray = Object.freeze([]) as Array<never>;

} // #end GraphLinksModel
GraphLinksModel['className'] = 'GraphLinksModel';

Model.registerClass(GraphLinksModel, 'GraphLinksModel');

Model['constructGraphLinksModel'] = Model.constructGraphLinksModel = function () {
  return new GraphLinksModel();
};

Model['initDiagramModel'] = Model.initDiagramModel = function () {
  return new GraphLinksModel();
};




// import { ChangedEvent } from './changedEvent';
// import { List, Map, Set } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { ObjectData } from './root';
// import { Transaction } from './undoManager';
// import { Util } from './util';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
TreeModel
*/

// import { Binding, ITreeModel, Key, Model } from './model';

/**
 * TreeModels support tree-structured graphs of nodes and links.
 * Each node can have at most one "tree parent"; cycles are not permitted.
 * The reference to the parent node's key is a property of the child node data.
 *
 * TreeModels, unlike {@link GraphLinksModel}s, do not support arbitrary link relationships between nodes,
 * nor is there a separate link data object for each parent-child relationship.
 * Furthermore there is no support for grouping or label nodes.
 *
 * The {@link #nodeParentKeyProperty} property names the property on the node data whose value
 * is the key of the "tree parent" node.
 * The default value for this property is "parent".
 *
 * For example, one can define a graph consisting of one parent node with two child nodes:
 * ```js
 *  model.nodeDataArray = [
 *    { key: "Alpha" },
 *    { key: "Beta", parent: "Alpha" },
 *    { key: "Gamma", parent: "Alpha" }
 *  ];
 * ```
 *
 * If you need to show a more complicated graph structure than a tree, use a {@link GraphLinksModel}.
 * If you want to have multiple links between the same pair of nodes,
 * or if you want to associate more information with each link and cannot put the information on the child node,
 * you will need to have a separate link data object for each link,
 * and that would require the use of {@link GraphLinksModel}.
 * @extends Model
 * @unrestricted
 * @since 1.1
 * @category Model
 */
export class TreeModel extends Model implements ITreeModel {
  /** @hidden @internal */ public static className = 'TreeModel';
  /** @hidden @internal */ private _nodeParentKeyProperty: string | ((a: ObjectData, b: Key) => Key);
  /** @hidden @internal */ private _copiesParentKeyOfNodeData: boolean;
  /** @hidden @internal */ private _parentLinkCategoryProperty: string | ((a: ObjectData, b: string) => string);

  /**
   * This constructs an empty TreeModel unless one provides arguments as the initial data array values
   * for the {@link Model#nodeDataArray} property.
   * @param {Array.<Object>=} nodedataarray an optional Array containing JavaScript objects to be represented by Nodes.
   */
  constructor(nodedataarray?: Array<ObjectData>) {
    if (Debug && arguments.length > 1) {
      Util.throwError('TreeModel constructor can only take one optional argument, the Array of node data.');
    }
    super();

    this._nodeParentKeyProperty = 'parent';
    this._copiesParentKeyOfNodeData = false;
    this._parentLinkCategoryProperty = 'parentLinkCategory';

    if (nodedataarray !== undefined) this.nodeDataArray = nodedataarray;
  }

  /**
   * @hidden @internal
   * Copies properties from this model to the given model, which must be of the same class.
   * This is called by {@link Model#copy}.
   * @expose
   * @this {TreeModel}
   * @param {Model} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._nodeParentKeyProperty = this._nodeParentKeyProperty;
    copy._copiesParentKeyOfNodeData = this._copiesParentKeyOfNodeData;
    copy._parentLinkCategoryProperty = this._parentLinkCategoryProperty;
  }

  /**
   * @hidden @internal
   * @this {TreeModel}
   * @param {number=} details Optional: If greater than zero, this includes descriptions of each node data and link data.
   * @return {string}
   */
  public toString(details?: number): string {
    if (details === undefined) details = 0;
    if (details >= 2) {
      return this.toJson();
    } else {
      let s = (this.name !== '' ? this.name : '') + ' TreeModel';
      if (details > 0) {
        s += '\n node data:';
        const arr = this.nodeDataArray;
        const len = arr.length;
        for (let i = 0; i < len; i++) {
          const item = arr[i];
          s += ' ' + this.getKeyForNodeData(item) + ':' + Binding.toString(item);
        }
      }
      return s;
    }
  }

  /**
   * @hidden @internal
   * Write out the persistent properties of this model into a string using JSON syntax property pairs.
   * @this {TreeModel}
   * @return {string}
   */
  protected writeJsonProtected(): string {
    const basestr = super.writeJsonProtected();
    let s = '';
    if (this.nodeParentKeyProperty !== 'parent' && typeof this.nodeParentKeyProperty === 'string') s += ',\n  \"nodeParentKeyProperty\": ' + this.quote(this.nodeParentKeyProperty);
    return basestr + s;
  }

  /**
   * @hidden @internal
   * Find the persistent model properties of the given Object and assign them to this model.
   * @this {TreeModel}
   * @param {Object} obj
   */
  protected readJsonProtected(obj: ObjectData): void {
    super.readJsonProtected(obj);
    if (obj['nodeParentKeyProperty']) this.nodeParentKeyProperty = obj['nodeParentKeyProperty'];
  }

  /**
   * @hidden @internal
   * @this {TreeModel}
   * @param {TreeModel} newmodel
   * @return {string}
   */
  protected writeJsonDifference(newmodel: TreeModel): string {
    this.checkSameProperties(newmodel, 'nodeParentKeyProperty');
    return super.writeJsonDifference(newmodel);
  }

  /**
   * @hidden @internal
   * @this {TreeModel}
   * @param {ObjectData} obj
   */
  protected readIncrementalJsonData(obj: ObjectData): void {
    // ensure Model properties are the same
    if (obj['nodeParentKeyProperty'] !== undefined && obj['nodeParentKeyProperty'] !== this.nodeParentKeyProperty) {
      Util.throwError('applyIncrementalJson cannot change Model properties');
    }
    super.readIncrementalJsonData(obj);
  }

  /**
   * @hidden @internal
   * Not really needed, but maybe some day there might be a use....
   * @this {TreeModel}
   * @param {string|number|undefined} key
   * @return {string|number|undefined}
   */
  protected maybeEnsureLinkReference(key: Key): Key {
    return key;
  }

  /**
   * Gets or sets the name of the property on node data that specifies
   * the string or number key of the node data that acts as the "parent" for this "child" node data.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return the string or number key for the parent node data object of which the given data object is a child;
   * if the second argument is supplied, the function should modify the node data object so that it has that new key
   * (which may be undefined to refer to no node) as the parent key for that node..
   * The default value is the name 'parent', meaning that it expects the data to have a property named 'parent' if the node wants to refer to the parent node by its key.
   * The value must not be null nor an empty string.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getParentKeyForNodeData}
   * @see {@link #setParentKeyForNodeData}
   */
  /** @export */ get nodeParentKeyProperty(): string | ((a: ObjectData, b?: Key) => Key) { return this._nodeParentKeyProperty; }
  /** @export */ set nodeParentKeyProperty(value: string | ((a: ObjectData, b?: Key) => Key)) {
    const old = this._nodeParentKeyProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, TreeModel, 'nodeParentKeyProperty');
      this._nodeParentKeyProperty = value;
      this.raiseChanged('nodeParentKeyProperty', old, value);
    }
  }

  /**
   * @hidden @internal
   * Gets or sets whether {@link #copyNodeData} copies the property that is the reference to the tree-parent node data.
   *
   * The default value is false.
   */
  /* noexport */get copiesParentKeyOfNodeData(): boolean { return this._copiesParentKeyOfNodeData; }
  set copiesParentKeyOfNodeData(value: boolean) {
    const old = this._copiesParentKeyOfNodeData;
    if (old !== value) {
      Util.checkPrimitive(value, 'boolean', TreeModel, 'copiesParentKeyOfNodeData');
      this._copiesParentKeyOfNodeData = value;
      // ???this.raiseChanged('copiesParentKeyOfNodeData', old, value);
    }
  }

  /**
   * If there is a parent node for the given node data, return the parent's key.
   * @export
   * @this {TreeModel}
   * @param {Object} nodedata a JavaScript object represented by a node.
   * @return {string|number|undefined}  This returns undefined if there is no parent node data object.
   * @see {@link #nodeParentKeyProperty}
   * @see {@link #setParentKeyForNodeData}
   */
  public getParentKeyForNodeData(nodedata: ObjectData): Key {
    if (nodedata === null) return undefined;
    const kname = this._nodeParentKeyProperty;
    if (kname === '') return undefined;
    const key = Model.safePropertyValue(nodedata, kname);
    if (key === undefined) return undefined;
    if (this.isKeyType(key)) return key;
    Util.throwError('ParentKey value for node data ' + nodedata + ' is not a number or a string: ' + key);
    return undefined;
  }

  /**
   * Change the parent node for the given node data, given a key for the new parent, or undefined if there should be no parent.
   * @export
   * @this {TreeModel}
   * @param {Object} nodedata a JavaScript object represented by a node.
   * @param {string|number|undefined} key This may be undefined if there should be no parent node data.
   * @see {@link #nodeParentKeyProperty}
   * @see {@link #getParentKeyForNodeData}
   */
  public setParentKeyForNodeData(nodedata: ObjectData, key: Key): void {
    if (key === null) key = undefined;
    if (key !== undefined && !this.isKeyType(key)) {
      Util.throwTypeError(key, 'number or string', TreeModel, 'setParentKeyForNodeData:key');
    }
    if (nodedata === null) return;
    const kname = this._nodeParentKeyProperty;
    if (kname === '') return;
    key = this.maybeEnsureLinkReference(key);
    if (!this.containsNodeData(nodedata)) {
      Model.safePropertySet(nodedata, kname, key);
      return;
    }
    const old = Model.safePropertyValue(nodedata, kname);
    if (old !== key) {
      this.forgetDelayedReference(old, nodedata);
      Model.safePropertySet(nodedata, kname, key);
      if (this.findNodeDataForKey(key) === null) {
        this.delayReference(key, nodedata);
      }
      this.raiseModelChangedEvent('nodeParentKey', ChangedEvent.Property, kname, nodedata, old, key);
      if (typeof kname === 'string') this.updateTargetBindings(nodedata, kname);
    }
  }

  /**
   * Gets or sets the name of the data property that returns a string describing that node data's parent link's category.
   * The value may also be a function taking two arguments, where the first argument will be a node data object.
   * If the second argument is not supplied, the function should return the category name for any parent link;
   * if the second argument is supplied, the function should modify the node data object so that its parent link has that new category name.
   * The default value is the name 'parentLinkCategory'.
   * This is used by the diagram to distinguish between different kinds of links.
   * The name must not be null.
   * If the value is an empty string,
   * {@link #getParentLinkCategoryForNodeData} will return an empty string for all node data objects.
   *
   * If you want to set this property you must do so before using the model, and especially before you assign {@link Diagram#model}.
   * Note that functions cannot be serialized into JSON-formatted text, so if you are using {@link #toJson} and {@link Model.fromJson},
   * and if you want this property to be a function, you will need to assign this property to your desired function
   * immediately after creating the model, including when it is created by {@link Model.fromJson}.
   * @see {@link #getParentLinkCategoryForNodeData}
   * @see {@link #setParentLinkCategoryForNodeData}
   */
  /** @export */ get parentLinkCategoryProperty(): string | ((a: ObjectData, b?: string) => string) { return this._parentLinkCategoryProperty; }
  /** @export */ set parentLinkCategoryProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    const old = this._parentLinkCategoryProperty;
    if (old !== value) {
      this.checkStringOrFunction(value, TreeModel, 'parentLinkCategoryProperty');
      this._parentLinkCategoryProperty = value;
      this.raiseChanged('parentLinkCategoryProperty', old, value);
    }
  }

  /**
   * @hidden @internal
   * Implement ILinkModel.linkCategoryProperty
   */
  /** @export */ get linkCategoryProperty(): string | ((a: ObjectData, b?: string) => string) { return this.parentLinkCategoryProperty; }
  /** @export */ set linkCategoryProperty(value: string | ((a: ObjectData, b?: string) => string)) {
    this.parentLinkCategoryProperty = value;
  }

  /**
   * Find the category for the parent link of a given child node data, a string naming the link template
   * that the {@link Diagram} should use to represent the link.
   * @export
   * @this {TreeModel}
   * @param {Object} childdata a JavaScript object represented by a node data.
   * @return {string}
   * @see {@link #parentLinkCategoryProperty}
   * @see {@link #setParentLinkCategoryForNodeData}
   */
  public getParentLinkCategoryForNodeData(childdata: ObjectData): string {
    if (childdata === null) return '';
    const cname = this._parentLinkCategoryProperty;
    if (cname === '') return '';
    const cat = Model.safePropertyValue(childdata, cname);
    if (cat === undefined) return '';
    if (typeof cat === 'string') return cat;
    Util.throwError('getParentLinkCategoryForNodeData found a non-string category for ' + childdata + ': ' + cat);
    return '';
  }

  /**
   * @hidden @internal
   * Implement ILinkModel.getLinkCategoryForData
   * @this {TreeModel}
   * @param {Object} data
   * @return {string}
   */
  public getLinkCategoryForData(data: ObjectData): string {
    return this.getParentLinkCategoryForNodeData(data);
  }

  /**
   * Change the category for the parent link of a given child node data, a string naming the link template
   * that the {@link Diagram} should use to represent the link.
   *
   * Changing the link template will cause any existing {@link Link}
   * to be removed from the {@link Diagram} and replaced with a new {@link Link}
   * created by copying the new link template and applying any data-bindings.
   * Note that the new template must be an instance of the same class as the original link.
   * Thus one cannot change the category of a link from an instance of {@link Link}
   * to an instance of a subclass of Link, nor vice-versa.
   * @export
   * @this {TreeModel}
   * @param {Object} childdata a JavaScript object represented by a node data.
   * @param {string} cat Must not be null.
   * @see {@link #parentLinkCategoryProperty}
   * @see {@link #getParentLinkCategoryForNodeData}
   */
  public setParentLinkCategoryForNodeData(childdata: ObjectData, cat: string): void {
    Util.checkPrimitive(cat, 'string', TreeModel, 'setParentLinkCategoryForNodeData:cat');
    if (childdata === null) return;
    const cname = this._parentLinkCategoryProperty;
    if (cname === '') return;
    if (!this.containsNodeData(childdata)) {
      Model.safePropertySet(childdata, cname, cat);
      return;
    }
    let old = Model.safePropertyValue(childdata, cname);
    if (old === undefined) old = '';
    if (old !== cat) {
      Model.safePropertySet(childdata, cname, cat);
      this.raiseModelChangedEvent('parentLinkCategory', ChangedEvent.Property, cname, childdata, old, cat);
      if (typeof cname === 'string') this.updateTargetBindings(childdata, cname);
    }
  }

  /**
   * @hidden @internal
   * Implement ILinkModel.setLinkCategoryForData
   * @this {TreeModel}
   * @param {Object} data
   * @param {string} cat
   */
  public setLinkCategoryForData(data: ObjectData, cat: string): void {
    this.setParentLinkCategoryForNodeData(data, cat);
  }

  /**
   * This override also makes sure any copied node data does not have a reference to a parent node.
   * @expose
   * @this {TreeModel}
   * @param {Object} nodedata a JavaScript object represented by a node, group, or non-link.
   * @return {Object}
   * @see {@link Model#copyNodeData}
   */
  public copyNodeData(nodedata: ObjectData): ObjectData | null {
    if (nodedata === null) return null;
    const copy = super.copyNodeData(nodedata);
    if (!this.copiesParentKeyOfNodeData && this._nodeParentKeyProperty !== ''
      && Model.safePropertyValue(copy, this._nodeParentKeyProperty) !== undefined) {
      Model.safePropertySet(copy, this._nodeParentKeyProperty, undefined);
    }
    return copy;
  }

  /**
   * This override changes the value of some property of a node data or an item data, given a string naming the property
   * and the new value, in a manner that can be undone/redone and that automatically updates any bindings.
   *
   * This gets the old value of the property; if the value is the same as the new value, no side-effects occur.
   * @expose
   * @this {TreeModel}
   * @param {Object} data a JavaScript object typically the value of a {@link Panel#data} and represented by a Node, Link, Group, simple Part,
   *                      or item in a {@link Panel#itemArray}; or this model's {@link #modelData}.
   * @param {string} propname a string that is not null or the empty string.
   * @param {*} val the new value for the property.
   * @see {@link Model#setDataProperty}
   */
  public setDataProperty(data: ObjectData, propname: string, val: any): void {
    if (Debug) {
      Util.checkPrimitive(data, 'object', TreeModel, 'setDataProperty:data');
      Util.checkPrimitive(propname, 'string', TreeModel, 'setDataProperty:propname');
      if (propname === '') Util.throwError('TreeModel.setDataProperty: property name must not be an empty string when setting ' + data + ' to ' + val);
    }
    if (this.containsNodeData(data)) {
      if (propname === this.nodeKeyProperty) {
        this.setKeyForNodeData(data, val);
      } else if (propname === this.nodeCategoryProperty) {
        this.setCategoryForNodeData(data, val);
        return;
      } else if (propname === this.nodeParentKeyProperty) {
        this.setParentKeyForNodeData(data, val);
        return;
      }
    } else {
      // TODO: warn if modifying a GraphObject
      // if (!Model._warnedAboutModifyingGraphObject && data instanceof GraphObject) {
      //   Model._warnedAboutModifyingGraphObject = true;
      //   Util.trace('TreeModel.setDataProperty is modifying a GraphObject, \"' + data.toString() + '\"');
      //   Util.trace('  Is that really your intent?');
      // }
    }
    // set the property and raise a Property ChangedEvent with the old and new values
    const old = Model.safePropertyValue(data, propname);
    if (old !== val) {
      Model.safePropertySet(data, propname, val);
      this.raiseDataChanged(data, propname, old, val);
    }
  }

  /**
   * @hidden @internal
   * When a node data changes its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * The goal is to maintain all of the original relationships in all of the data maintained by the model.
   * @this {TreeModel}
   * @param {string|number|undefined} oldkey
   * @param {string|number} newkey
   */
  protected replaceAllKeyReferences(oldkey: Key, newkey: Key): void {
    super.replaceAllKeyReferences(oldkey, newkey);
    // ?? this is very slow, but it might not be worth it to remember and maintain back-pointers
    const it = this._mapKeyNodeData.iterator;
    while (it.next()) {
      const data = it.value;
      this.replaceKeyReferencesInNodeData(data, oldkey, newkey);
    }
  }

  /**
   * @hidden @internal
   * When a node key has changed its key, we need to update all data references to that node data
   * that are using the old key value by replacing them with the new key value.
   * @this {TreeModel}
   * @param {Object} d
   * @param {string|number=} oldkey
   * @param {string|number=} newkey
   */
  protected replaceKeyReferencesInNodeData(d: ObjectData, oldkey?: Key, newkey?: Key): void {
    const pkey = this.getParentKeyForNodeData(d);
    if (pkey === oldkey) {
      const kname = this._nodeParentKeyProperty;
      Model.safePropertySet(d, kname, newkey);
      this.raiseModelChangedEvent('nodeParentKey', ChangedEvent.Property, kname, d, oldkey, newkey);
      if (typeof kname === 'string') this.updateTargetBindings(d, kname);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {TreeModel}
   * @param {Object} nodedata
   */
  protected resolveReferencesToNodeData(nodedata: ObjectData): void {
    super.resolveReferencesToNodeData(nodedata);
    const key = this.getKeyForNodeData(nodedata);
    const refdata = this.findDelayedReferences(key);
    if (refdata !== null) {
      const removes = Util.tempArray<ObjectData>();
      const dit = refdata.iterator;
      while (dit.next()) {
        const d = dit.value;
        if (this.containsNodeData(d)) {  // D is a node data
          const pkey = this.getParentKeyForNodeData(d);
          if (pkey === key) {
            const kname = this._nodeParentKeyProperty;
            this.raiseModelChangedEvent('nodeParentKey', ChangedEvent.Property, kname, d, key, key);
            if (typeof kname === 'string') this.updateTargetBindings(d, kname);
            removes.push(d);
          }
        }
      }
      for (let i = 0; i < removes.length; i++) {
        this.forgetDelayedReference(key, removes[i]);
      }
      Util.freeArray(removes);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {TreeModel}
   * @param {Object} nodedata
   */
  protected resolveReferencesInNodeData(nodedata: ObjectData): void {
    super.resolveReferencesInNodeData(nodedata);
    let pkey = this.getParentKeyForNodeData(nodedata);
    pkey = this.maybeEnsureLinkReference(pkey);
    if (this.findNodeDataForKey(pkey) === null) {
      this.delayReference(pkey, nodedata);
    }
  }

  /**
   * @hidden @internal
   * Overridden.
   * @this {TreeModel}
   * @param {Object} nodedata
   */
  protected unresolveReferencesInNodeData(nodedata: ObjectData): void {
    super.unresolveReferencesInNodeData(nodedata);
    const pkey = this.getParentKeyForNodeData(nodedata);
    this.forgetDelayedReference(pkey, nodedata);
  }

  /**
   * @hidden @internal
   * Gets the string describing the type of model, used in {@link Diagram#addPartManager} to map a {@link Model} to
   * its {@link PartManager}. This just returns 'TreeModel'.
   */
/** @export */   public static get type(): string { return 'TreeModel'; }

  /**
   * @hidden @internal
   * @this {TreeModel}
   * @return {boolean}
   */
  public isTreeModel(): this is ITreeModel {
    return true;
  }

  /**
   * @hidden @internal
   * @this {TreeModel}
   * @return {boolean}
   */
  public isLinkModel(): boolean {
    return true;
  }
} // #end TreeModel
TreeModel['className'] = 'TreeModel';

Model.registerClass(TreeModel, 'TreeModel');

// import { ChangedEvent } from './changedEvent';
// import { Debug } from './debug';
// import { ObjectData } from './root';
// import { Util } from './util';



/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
CircularLayout
CircularNetwork
CircularVertex
CircularEdge
*/

// import { List, Set } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Layout, LayoutEdge, LayoutNetwork, LayoutVertex } from './layout';
// import { Group, Part } from './parts';
// import { Point } from './point';
// import { Util } from './util';

// ***************CircularLayout***************

/**
 * This layout positions nodes in a circular arrangement.
 * There are several samples that use CircularLayout.
 * The layout cannot guarantee that it provides optimal positioning of nodes when trying to minimize link crossings.
 * <p class="boxrun">
 * If you want to experiment interactively with most of the properties, try the <a href="../../samples/cLayout.html">Circular Layout</a> sample.
 * See samples that make use of CircularLayout in the <a href="../../samples/index.html#circularlayout">samples index</a>.
 *
 * This layout makes use of a {@link LayoutNetwork} of
 * {@link CircularVertex}es and {@link CircularEdge}s that normally
 * correspond to the {@link Node}s and {@link Link}s of the {@link Diagram}.
 * @extends Layout
 * @unrestricted
 * @category Layout
 */
export class CircularLayout extends Layout {
  /** @hidden @internal */ public static className = 'CircularLayout';
  /** @hidden @internal */ private _eRadius: number;
  /** @hidden @internal */ private _eAspectRatio: number;
  /** @hidden @internal */ private _eStartAngle: number;
  /** @hidden @internal */ private _eSweepAngle: number;
  /** @hidden @internal */ private _eArrangement: EnumValue;
  /** @hidden @internal */ private _eSpacing: number;
  /** @hidden @internal */ private _actualCenter: Point;
  /** @hidden @internal */ private _eNodeDiameterFormula: EnumValue;
  /** @hidden @internal */ private _totalElementSizes: number;
  /** @hidden @internal */ private _yRadius: number;
  /** @hidden @internal */ private _constDist: number;
  /** @hidden @internal */ private _vertexArrangement: VertexArrangement;
  /** @hidden @internal */ private _packedIters: number;
  /** @hidden @internal */ private _correctLastAngle: number;
  /** @hidden @internal */ private _riemannnum: number;
  /** @hidden @internal */ private _radius: number;
  /** @hidden @internal */ private _aspectRatio: number;
  /** @hidden @internal */ private _startAngle: number;
  /** @hidden @internal */ private _sweepAngle: number;
  /** @hidden @internal */ private _arrangement: EnumValue;
  /** @hidden @internal */ private _direction: EnumValue;
  /** @hidden @internal */ private _sorting: EnumValue;
  /** @hidden @internal */ private _comparer: ((a: CircularVertex, b: CircularVertex) => number);
  /** @hidden @internal */ private _spacing: number;
  /** @hidden @internal */ private _nodeDiameterFormula: EnumValue;

  /**
   * Constructs a {@link CircularLayout} with no {@link Layout#network}
   * and with no owning {@link Layout#diagram}.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(CircularLayout);
    super();

    // private variables
    this._eRadius = 0;
    this._eAspectRatio = 0;
    this._eStartAngle = 0;
    this._eSweepAngle = 360;
    this._eArrangement = CircularLayout.ConstantSpacing;
    this._eSpacing = 0;
    this._actualCenter = new Point();
    this._eNodeDiameterFormula = CircularLayout.ConstantSpacing;
    this._totalElementSizes = 0;  // the sum of the diameters of all the elements of the layout
    this._yRadius = 0;
    this._constDist = 0;  // for CircularLayout.ConstantDistance or ConstantAngle
    this._vertexArrangement = new VertexArrangement();
    this._packedIters = 0;
    this._correctLastAngle = 0;
    this._riemannnum = 600;

    // properties
    this._radius = NaN;
    this._aspectRatio = 1;
    this._startAngle = 0;
    this._sweepAngle = 360;
    this._arrangement = CircularLayout.ConstantSpacing;
    this._direction = CircularLayout.Clockwise;
    this._sorting = CircularLayout.Optimized;
    this._comparer = LayoutVertex.standardComparer;
    this._spacing = 6;
    this._nodeDiameterFormula = CircularLayout.Pythagorean;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Layout.
   * @this {CircularLayout}
   * @param {?} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    // these properties are computed:
    // this._eRadius
    // this._eAspectRatio
    // this._eStartAngle
    // this._eSweepAngle
    // this._eArrangement
    // this._eSpacing
    // this._actualCenter
    // this._totalElementSizes
    // this._yRadius
    // this._constDist
    // this._vertexArrangement
    // this._packedIters
    // this._correctLastAngle
    // this._riemannnum
    copy._radius = this._radius;
    copy._aspectRatio = this._aspectRatio;
    copy._startAngle = this._startAngle;
    copy._sweepAngle = this._sweepAngle;
    copy._arrangement = this._arrangement;
    copy._direction = this._direction;
    copy._sorting = this._sorting;
    copy._comparer = this._comparer;
    copy._spacing = this._spacing;
    copy._nodeDiameterFormula = this._nodeDiameterFormula;
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {CircularLayout}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === CircularLayout) {
      if (arg === CircularLayout.Ascending || arg === CircularLayout.Descending ||
        arg === CircularLayout.Forwards || arg === CircularLayout.Reverse || arg === CircularLayout.Optimized) {
        this.sorting = arg;
      } else if (arg === CircularLayout.BidirectionalLeft || arg === CircularLayout.BidirectionalRight ||
        arg === CircularLayout.Clockwise || arg === CircularLayout.Counterclockwise) {
        this.direction = arg;
      } else if (arg === CircularLayout.ConstantAngle || arg === CircularLayout.ConstantDistance ||
        arg === CircularLayout.ConstantSpacing || arg === CircularLayout.Packed) {
        this.arrangement = arg;
      } else if (arg === CircularLayout.Circular || arg === CircularLayout.Pythagorean) {
        this.nodeDiameterFormula = arg;
      }
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * Create a new {@link LayoutNetwork} of {@link CircularVertex}es and {@link CircularEdge}s.
   * @this {CircularLayout}
   * @return {CircularNetwork} a new {@link LayoutNetwork}.
   */
  public createNetwork(): CircularNetwork {
    return new CircularNetwork(this);
  }

  /**
   * @hidden @internal
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {List.<CircularVertex>} evens
   * @param {List.<CircularVertex>} odds
   * @return {List.<CircularVertex>} modified or new List of vertices
   */
  private setEffectiveValues(vertices: List<CircularVertex>, evens: List<CircularVertex>, odds: List<CircularVertex>): List<CircularVertex> {
    vertices = this.sort(vertices);

    // get the private variables, and we will set them at the end
    let eArrangement = this._eArrangement;
    let eNodeDiameterFormula = this._eNodeDiameterFormula;
    let eRadius = this._eRadius;
    let eAspectRatio = this._eAspectRatio;
    let eStartAngle = this._eStartAngle;
    let eSweepAngle = this._eSweepAngle;
    let eSpacing = this._eSpacing;
    let totalElementSizes = this._totalElementSizes;
    let yRadius = this._yRadius;
    let constDist = this._constDist;

    // fetch local copies of property values; if invalid, use a valid value instead

    eArrangement = this.arrangement;
    eNodeDiameterFormula = this.nodeDiameterFormula;

    eRadius = this.radius;
    if (!isFinite(eRadius) || eRadius <= 0) {
      eRadius = NaN;
    }

    eAspectRatio = this.aspectRatio;
    if (!isFinite(eAspectRatio) || eAspectRatio <= 0) {
      eAspectRatio = 1;
    }

    eStartAngle = this.startAngle;
    if (!isFinite(eStartAngle)) {
      eStartAngle = 0;
    }

    eSweepAngle = this.sweepAngle;
    if (!isFinite(eSweepAngle) || eSweepAngle > 360 || eSweepAngle < 1) {
      eSweepAngle = 360;
    }

    eSpacing = this.spacing;
    if (!isFinite(eSpacing)) {
      eSpacing = NaN;
    }

    if (eArrangement === CircularLayout.Packed && eNodeDiameterFormula === CircularLayout.Circular) {
      // Packing circular nodes is equivalent to what CircularLayout.ConstantSpacing does
      eArrangement = CircularLayout.ConstantSpacing;
    } else if (eArrangement === CircularLayout.Packed && eNodeDiameterFormula !== CircularLayout.Circular) {
      eNodeDiameterFormula = CircularLayout.Circular;
      eArrangement = this.arrangement;
    }

    if ((this.direction === CircularLayout.BidirectionalLeft || this.direction === CircularLayout.BidirectionalRight)
      && this.sorting !== CircularLayout.Optimized) {
      // bidirectional
      for (let i = 0; ; i += 2) {
        if (i >= vertices.length) {
          break;
        }
        evens.add(vertices.elt(i));
        if (i + 1 >= vertices.length) {
          break;
        }
        odds.add(vertices.elt(i + 1));
      }
      if (this.direction === CircularLayout.BidirectionalLeft) {
        // reverse evens if Packed, append odds
        if (this.arrangement === CircularLayout.Packed) {
          evens.reverse();
        }
        vertices = new List<CircularVertex>();
        vertices.addAll(evens);
        vertices.addAll(odds);
      } else {
        // reverse odds if Packed, append evens
        if (this.arrangement === CircularLayout.Packed) {
          odds.reverse();
        }
        vertices = new List<CircularVertex>();
        vertices.addAll(odds);
        vertices.addAll(evens);
      }
    }

    const num = vertices.length;
    totalElementSizes = 0;
    let n = 0;
    for (let i = 0; i < vertices.length; i++) {
      const angle = eStartAngle + (eSweepAngle * n * (this.direction === CircularLayout.Clockwise ? 1 : -1)) / num;
      let add = vertices.elt(i).diameter;
      // var add = vertices.elt(i).computeDiameter(angle);
      if (isNaN(add)) add = vertices.elt(i).computeDiameter(angle);
      if (eSweepAngle < 360 && (i === 0 || i === vertices.length - 1)) {
        add /= 2;
      }
      totalElementSizes += add;
      n++;
    }

    if (isNaN(eRadius) || eArrangement === CircularLayout.Packed) {  // radius not set
      // spacing not set
      if (isNaN(eSpacing)) {
        eSpacing = 6;
      }

      // set radius based on spacing
      if (eArrangement !== CircularLayout.ConstantSpacing && eArrangement !== CircularLayout.Packed) {
        let maxdist = -Infinity;
        for (let i = 0; i < num; i++) {
          const curr = vertices.elt(i);
          const next = vertices.elt(i === num - 1 ? 0 : i + 1);
          if (isNaN(curr.diameter)) curr.computeDiameter(0);
          if (isNaN(next.diameter)) next.computeDiameter(0);
          maxdist = Math.max(maxdist, (curr.diameter + next.diameter) / 2);
        }
        constDist = maxdist + eSpacing;
        if (eArrangement === CircularLayout.ConstantAngle) {
          const angle = 2 * Math.PI / num;
          const minrad = (maxdist + eSpacing) / angle;
          if (eAspectRatio > 1) {
            eRadius = minrad;
            yRadius = eRadius * eAspectRatio;
          } else {
            yRadius = minrad;
            eRadius = yRadius;
          }
        } else {
          eRadius = this.inverseEllipsePerim((constDist * (eSweepAngle >= 360 ? num : num - 1)), eAspectRatio, eStartAngle * Math.PI / 180, eSweepAngle * Math.PI / 180);
        }
      } else {
        // constant spacing
        eRadius = this.inverseEllipsePerim((totalElementSizes + (eSweepAngle >= 360 ? num : num - 1) * (eArrangement !== CircularLayout.Packed ? eSpacing : eSpacing * 1.6)),
                                           eAspectRatio, eStartAngle * Math.PI / 180, eSweepAngle * Math.PI / 180);
      }
      yRadius = eRadius * eAspectRatio;
    } else {
      // radius is a number
      yRadius = eRadius * eAspectRatio;
      const circ = this.ellipsePerim(eRadius, yRadius, eStartAngle * Math.PI / 180, eSweepAngle * Math.PI / 180);
      if (isNaN(eSpacing)) {
        // spacing is NaN
        if (eArrangement === CircularLayout.ConstantSpacing || eArrangement === CircularLayout.Packed) {
          eSpacing = (circ - totalElementSizes) / (eSweepAngle >= 360 ? num : num - 1);
        }
        // else eSpacing doesn't matter
      } else {  // spacing is set
        if (eArrangement === CircularLayout.ConstantSpacing || eArrangement === CircularLayout.Packed) {
          // spacing is set, const spacing
          const currentspacing = (circ - totalElementSizes) / (eSweepAngle >= 360 ? num : num - 1);

          if (currentspacing < eSpacing) {
            eRadius = this.inverseEllipsePerim((totalElementSizes + eSpacing * (eSweepAngle >= 360 ? num : num - 1)),
                                               eAspectRatio, eStartAngle * Math.PI / 180, eSweepAngle * Math.PI / 180);
            yRadius = eRadius * eAspectRatio;
          } else {
            eSpacing = currentspacing;
          }
        } else {
          // constant spacing (perhaps)
          let maxdist = -Infinity;
          for (let i = 0; i < num; i++) {
            const curr = vertices.elt(i);
            const next = vertices.elt(i === num - 1 ? 0 : i + 1);
            if (isNaN(curr.diameter)) curr.computeDiameter(0);
            if (isNaN(next.diameter)) next.computeDiameter(0);
            maxdist = Math.max(maxdist, (curr.diameter + next.diameter) / 2);
          }
          const minConstDist = maxdist + eSpacing;
          const minNeededRad = this.inverseEllipsePerim((minConstDist * (eSweepAngle >= 360 ? num : num - 1)),
                                                        eAspectRatio, eStartAngle * Math.PI / 180, eSweepAngle * Math.PI / 180);
          if (minNeededRad > eRadius) {
            eRadius = minNeededRad;
            yRadius = eRadius * eAspectRatio;
            constDist = minConstDist;
          } else {
            constDist = circ / (eSweepAngle >= 360 ? num : num - 1); // may be unnecessary
          }
        }
      }
    }
    this._eArrangement = eArrangement;
    this._eNodeDiameterFormula = eNodeDiameterFormula;
    this._eRadius = eRadius;
    this._eAspectRatio = eAspectRatio;
    this._eStartAngle = eStartAngle;
    this._eSweepAngle = eSweepAngle;
    this._eSpacing = eSpacing;
    this._totalElementSizes = totalElementSizes;
    this._yRadius = yRadius;
    this._constDist = constDist;

    return vertices;
  }

  /**
   * Assign the positions of the vertexes in the network.
   * @this {CircularLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');
    // create network
    if (this.network === null) {
      this.network = this.makeNetwork(coll);
    }

    this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);

    // if there is only one node, put it at the origin
    const vertexes = this.network.vertexes as any as Set<CircularVertex>;
    if (vertexes.count <= 1) {
      if (vertexes.count === 1) {
        const cv = vertexes.first();
        cv.centerX = 0;
        cv.centerY = 0;
      }
      this.updateParts();
      this.network = null;
      this.isValidLayout = true;
      return;
    }

    let vertices = new List<CircularVertex>();
    vertices.addAll(vertexes.iterator);
    const evens = new List<CircularVertex>();
    const odds = new List<CircularVertex>();

    vertices = this.setEffectiveValues(vertices, evens, odds);

    const eArrangement = this._eArrangement;
    const eRadius = this._eRadius;
    const eStartAngle = this._eStartAngle;
    const eSweepAngle = this._eSweepAngle;
    const eSpacing = this._eSpacing;
    const yRadius = this._yRadius;
    const constDist = this._constDist;

    if ((this.direction === CircularLayout.BidirectionalLeft || this.direction === CircularLayout.BidirectionalRight) &&
      eArrangement === CircularLayout.Packed) {
      // if Arrangement === Packed and Direction is bidirectional, this is achieved by using a rearrangement of the vertices, and shifting the StartAngle
      this.packedLayout(vertices, eSweepAngle, eStartAngle - eSweepAngle / 2, CircularLayout.Clockwise);
    } else if (this.direction === CircularLayout.BidirectionalLeft || this.direction === CircularLayout.BidirectionalRight) {
      let dsa = 0.0; // change in start angle
      // bidirectionals are achieved by laying out the even nodes in one direction, then laying out the odds in the opposite direction.
      // StartAngles must be slightly different.  The switch below computes the difference
      switch (eArrangement) {
        case CircularLayout.ConstantDistance:
          dsa = this.ellipseAngle(eRadius, yRadius, eStartAngle, constDist) * 180 / Math.PI;
          break;
        case CircularLayout.ConstantSpacing: {
          let evendia = 0.0;
          let odddia = 0.0;
          const firsteven = evens.first();
          if (firsteven !== null) {
            evendia = firsteven.computeDiameter(Math.PI / 2);
          }
          const firstodd = odds.first();
          if (firstodd !== null) {
            odddia = firstodd.computeDiameter(Math.PI / 2);
          }
          dsa = this.ellipseAngle(eRadius, yRadius, eStartAngle, eSpacing + (evendia + odddia) / 2) * 180 / Math.PI;
          break;
        }
        case CircularLayout.ConstantAngle:
          dsa = eSweepAngle / vertices.length;
          break;
      }
      // layout evens, then odds
      if (this.direction === CircularLayout.BidirectionalLeft) {
        switch (eArrangement) {
          case CircularLayout.ConstantDistance:
            this.distanceLayout(evens, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
          case CircularLayout.ConstantSpacing:
            this.spacingLayout(evens, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
          case CircularLayout.ConstantAngle:
            this.angleLayout(evens, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
        }
        switch (eArrangement) {
          case CircularLayout.ConstantDistance:
            this.distanceLayout(odds, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
          case CircularLayout.ConstantSpacing:
            this.spacingLayout(odds, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
          case CircularLayout.ConstantAngle:
            this.angleLayout(odds, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
        }
      } else {
        switch (eArrangement) {
          case CircularLayout.ConstantDistance:
            this.distanceLayout(odds, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
          case CircularLayout.ConstantSpacing:
            this.spacingLayout(odds, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
          case CircularLayout.ConstantAngle:
            this.angleLayout(odds, eSweepAngle / 2, eStartAngle, CircularLayout.Counterclockwise);
            break;
        }
        switch (eArrangement) {
          case CircularLayout.ConstantDistance:
            this.distanceLayout(evens, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
          case CircularLayout.ConstantSpacing:
            this.spacingLayout(evens, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
          case CircularLayout.ConstantAngle:
            this.angleLayout(evens, eSweepAngle / 2, eStartAngle + dsa, CircularLayout.Clockwise);
            break;
        }
      }
    } else {
      // if Direction isn't bidirectional, the nodes are laid out by calling one of these functions.
      switch (eArrangement) {
        case CircularLayout.ConstantDistance:
          this.distanceLayout(vertices, eSweepAngle, eStartAngle, this.direction);
          break;
        case CircularLayout.ConstantSpacing:
          this.spacingLayout(vertices, eSweepAngle, eStartAngle, this.direction);
          break;
        case CircularLayout.ConstantAngle:
          this.angleLayout(vertices, eSweepAngle, eStartAngle, this.direction);
          break;
        case CircularLayout.Packed:
          this.packedLayout(vertices, eSweepAngle, eStartAngle, this.direction);
          break;
      }
    }

    // Update the "physical" positions of the nodes and links.
    this.updateParts();
    this.network = null;
    this.isValidLayout = true;
  }

  /**
   * @hidden @internal
   * Arranges the items so the angle between any item and
   * an adjacent item is the same.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private angleLayout(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    const eSweepAngle = this._eSweepAngle;
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;
    const theta0 = start * Math.PI / 180;
    const sweepingAngle = sweep * Math.PI / 180;
    // var e = eRadius > yRadius ? Math.sqrt(eRadius * eRadius - yRadius * yRadius) / eRadius : Math.Ssrt(yRadius * yRadius - eRadius * eRadius) / yRadius;
    const num = vertices.length;
    for (let i = 0; i < num; i++) {
      const theta = theta0 + (dir === CircularLayout.Clockwise ? (i * sweepingAngle) / (eSweepAngle >= 360 ? num : num - 1) : -(i * sweepingAngle) / num);
      const cv = vertices.elt(i);
      const w = eRadius * Math.tan(theta) / yRadius;
      const r = Math.sqrt((eRadius * eRadius + yRadius * yRadius * w * w) / (1 + w * w));
      cv.centerX = r * Math.cos(theta);
      cv.centerY = r * Math.sin(theta);
      cv.actualAngle = theta * 180 / Math.PI;
    }
  }

  /**
   * @hidden @internal
   * Arranges the items so the spacing between any item and
   * an adjacent item is the same.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private spacingLayout(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;
    const eSpacing = this._eSpacing;
    let theta = start * Math.PI / 180;
    const num = vertices.length;
    for (let i = 0; i < num; i++) {
      const curr = vertices.elt(i);
      const next = vertices.elt(i === num - 1 ? 0 : i + 1);
      const x = eRadius * Math.cos(theta);
      const y = yRadius * Math.sin(theta);
      curr.centerX = x;
      curr.centerY = y;
      curr.actualAngle = theta * 180 / Math.PI;
      if (isNaN(curr.diameter)) curr.computeDiameter(0);
      if (isNaN(next.diameter)) next.computeDiameter(0);
      const meandiam = (curr.diameter + next.diameter) / 2;
      const thetachange = this.ellipseAngle(eRadius, yRadius, dir === CircularLayout.Clockwise ? theta : -theta, meandiam + eSpacing);
      theta += dir === CircularLayout.Clockwise ? thetachange : -thetachange;
    }
  }

  /**
   * @hidden @internal
   * Arranges the items so the distance between any item and
   * an adjacent item is the same.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private distanceLayout(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;
    const constDist = this._constDist;
    let theta = start * Math.PI / 180;
    const num = vertices.length;
    for (let i = 0; i < num; i++) {
      const cv = vertices.elt(i);
      cv.centerX = eRadius * Math.cos(theta);
      cv.centerY = yRadius * Math.sin(theta);
      cv.actualAngle = theta * 180 / Math.PI;
      const thetachange = this.ellipseAngle(eRadius, yRadius, dir === CircularLayout.Clockwise ? theta : -theta, constDist);
      theta += dir === CircularLayout.Clockwise ? thetachange : -thetachange;
    }
  }

  /**
   * @hidden @internal
   * Arranges the items so the spacing between any item and
   * an adjacent item is the same, but takes into account the assumption
   * that the nodes are rectangular.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private packedLayout(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    let correctLastAngle = this._correctLastAngle;
    const eSweepAngle = this._eSweepAngle;

    this._packedIters = 0;
    this._vertexArrangement = new VertexArrangement();

    if (sweep < 360) {
      correctLastAngle = start + (dir === CircularLayout.Clockwise ? eSweepAngle : -eSweepAngle);
      while (correctLastAngle < 0) { correctLastAngle += 360; }
      correctLastAngle %= 360;
      if (correctLastAngle > 180) { correctLastAngle -= 360; }
      correctLastAngle *= Math.PI / 180;
      this._correctLastAngle = correctLastAngle;
      this.packedLayoutSemi(vertices, sweep, start, dir);
    } else {
      this.packedLayoutFull(vertices, sweep, start, dir);
    }
    this._vertexArrangement.commit(vertices);
    return;
  }

  /**
   * @hidden @internal
   * Assume a 360 degree sweep.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private packedLayoutFull(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;
    const eSpacing = this._eSpacing;
    const eAspectRatio = this._eAspectRatio;

    // do layout
    let x = eRadius * Math.cos(start * Math.PI / 180);
    let y = yRadius * Math.sin(start * Math.PI / 180);
    const verts = vertices.toArray();

    if (verts.length === 3) {
      verts[0].centerX = eRadius;
      verts[0].centerY = 0;
      verts[1].centerX = verts[0].centerX - verts[0].width / 2 - verts[1].width / 2 - eSpacing;
      verts[1].y = verts[0].y;
      verts[2].centerX = (verts[0].centerX + verts[1].centerX) / 2;
      verts[2].y = verts[0].y - verts[2].height - eSpacing;
      return;
    } else if (verts.length === 4) {
      verts[0].centerX = eRadius;
      verts[0].centerY = 0;
      verts[2].centerX = -verts[0].centerX;
      verts[2].centerY = verts[0].centerY;
      verts[1].centerX = 0;
      verts[1].y = Math.min(verts[0].y, verts[2].y) - verts[1].height - eSpacing;
      verts[3].centerX = 0;
      verts[3].y = Math.max(verts[0].y + verts[0].height + eSpacing, verts[2].y + verts[2].height + eSpacing);
      return;
    }

    const temp = Point.alloc();
    for (let v = 0; v < verts.length; v++) {
      verts[v].centerX = x;
      verts[v].centerY = y;

      if (v >= verts.length - 1) break; // avoid index out of bounds
      if (!this.nexth(x, y, verts, v, dir, temp)) {
        this.nextv(x, y, verts, v, dir, temp);
      }
      x = temp.x;
      y = temp.y;
    }
    Point.free(temp);

    // check for gap, find gap sizes
    this._packedIters++;
    if (this._packedIters > 23) return;
    const locfx = verts[0].centerX;
    const locfy = verts[0].centerY;
    const loclx = verts[verts.length - 1].centerX;
    const locly = verts[verts.length - 1].centerY;
    let gapx = Math.abs(locfx - loclx) - ((verts[0].width + verts[verts.length - 1].width) / 2 + eSpacing);
    const gapy = Math.abs(locfy - locly) - ((verts[0].height + verts[verts.length - 1].height) / 2 + eSpacing);

    // compute gap
    let gap = 0.0;
    if (Math.abs(gapy) < 1) {
      const hgap = Math.abs(locfx - loclx);
      const max = (verts[0].width + verts[verts.length - 1].width) / 2;
      if (hgap < max) gap = 0.0;
      gapx = hgap - max;
    } else {
      // gapy not 0
      if (gapy > 0) {
        gap = gapy;
      } else {
        if (Math.abs(gapx) < 1) {
          gap = 0.0;
        } else {
          gap = gapx;
        }
      }
    }

    // check for excessive overlap
    let overlap = false;
    if (Math.abs(loclx) > Math.abs(locly)) {
      // use x
      overlap = (loclx > 0) !== (locfy > locly);
    } else {
      // use y
      overlap = (locly > 0) !== (locfx < loclx);
    }
    overlap = dir === CircularLayout.Clockwise ? overlap : !overlap;

    // if there's an excessive overlap, gap must be <0
    if (overlap) {
      gap = -Math.abs(gap);

      gap = Math.min(gap, -verts[verts.length - 1].width);
      gap = Math.min(gap, -verts[verts.length - 1].height);
    }

    // record this
    this._vertexArrangement.compare(gap, verts);

    // change radii accordingly
    if (Math.abs(gap) > 1) {
      if (this._packedIters < 8) {
        this._eRadius -= gap / (2 * Math.PI);
      } else {
        if (verts.length < 5 && gap > 10) {
          this._eRadius /= 2;
        } else {
          this._eRadius -= (gap > 0 ? 1.7 : -2.3);
        }
      }
      this._yRadius = this._eRadius * eAspectRatio;
      this.packedLayoutFull(vertices, sweep, start, dir);
    }
  }

  /**
   * @hidden @internal
   * Assume less than 360 degree sweep.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @param {number} sweep
   * @param {number} start
   * @param {EnumValue} dir
   */
  private packedLayoutSemi(vertices: List<CircularVertex>, sweep: number, start: number, dir: EnumValue): void {
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;
    const eAspectRatio = this._eAspectRatio;

    // do layout
    let x = eRadius * Math.cos(start * Math.PI / 180);
    let y = yRadius * Math.sin(start * Math.PI / 180);
    const temp = Point.alloc();
    const verts = vertices.toArray();
    for (let v = 0; v < verts.length; v++) {
      verts[v].centerX = x;
      verts[v].centerY = y;

      if (v >= verts.length - 1) break;  // avoid index out of bounds

      if (!this.nexth(x, y, verts, v, dir, temp)) {
        this.nextv(x, y, verts, v, dir, temp);
      }
      x = temp.x;
      y = temp.y;
    }
    Point.free(temp);

    // check for gap, find gap sizes
    this._packedIters++;
    if (this._packedIters > 23) return;

    // compute gap
    const lastangle = Math.atan2(y, x);
    let diff = dir === CircularLayout.Clockwise ? this._correctLastAngle - lastangle : lastangle - this._correctLastAngle;
    diff = Math.abs(diff) < Math.abs(diff - 2 * Math.PI) ? diff : diff - 2 * Math.PI;
    const gap = diff * (eRadius + yRadius) / 2;

    // record this
    this._vertexArrangement.compare2(gap, verts);

    // change radii accordingly
    if (Math.abs(gap) > 1) {
      if (this._packedIters < 8) {
        this._eRadius -= gap / (2 * Math.PI);
      } else {
        this._eRadius -= (gap > 0 ? 1.7 : -2.3);
      }
      this._yRadius = this._eRadius * eAspectRatio;
      this.packedLayoutSemi(vertices, sweep, start, dir);
    }
  }

  /**
   * @hidden @internal
   * Finds the coords. of the next item in the PackedLayout if it's
   * added horizontally.
   * @this {CircularLayout}
   * @param {number} x
   * @param {number} y
   * @param {Array.<CircularVertex>} verts
   * @param {number} v int
   * @param {EnumValue} dir
   * @param {Point} result @out
   * @return {boolean}
   */
  private nexth(x: number, y: number, verts: Array<CircularVertex>, v: number, dir: EnumValue, result: Point): boolean {
    const eSpacing = this._eSpacing;
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;

    let newx = 0.0;
    let newy = 0.0;
    const delta = (verts[v].width + verts[v + 1].width) / 2 + eSpacing;
    let inv = false;
    if ((y >= 0) !== (dir === CircularLayout.Clockwise)) {
      // x increases
      newx = x + delta;
      if (newx > eRadius) {
        // x decreases
        newx = x - delta;
        if (newx < -eRadius) {
          result.x = newx;
          result.y = newy;
          return false;
        }
        inv = true;
      }
    } else {
      // x decreases
      newx = x - delta;
      if (newx < -eRadius) {
        newx = x + delta;
        if (newx > eRadius) {
          result.x = newx;
          result.y = newy;
          return false;
        }
        inv = true;
      }
    }

    newy = Math.sqrt(1 - Math.min(1, newx * newx / (eRadius * eRadius))) * yRadius;
    if ((y < 0) !== inv) {
      newy = -newy;
    }
    if (Math.abs(y - newy) > (verts[v].height + verts[v + 1].height) / 2) {
      result.x = newx;
      result.y = newy;
      return false;
    } else {
      result.x = newx;
      result.y = newy;
      return true;
    }
  }

  /**
   * @hidden @internal
   * Finds the coords. of the next item in the PackedLayout if it's
   * added vertically.
   * @this {CircularLayout}
   * @param {number} x
   * @param {number} y
   * @param {Array.<CircularVertex>} verts
   * @param {number} v int
   * @param {EnumValue} dir
   * @param {Point} result @out
   * @return {boolean}
   */
  private nextv(x: number, y: number, verts: Array<CircularVertex>, v: number, dir: EnumValue, result: Point): boolean {
    const eSpacing = this._eSpacing;
    const eRadius = this._eRadius;
    const yRadius = this._yRadius;

    let newx = 0.0;
    let newy = 0.0;
    const delta = (verts[v].height + verts[v + 1].height) / 2 + eSpacing;
    let inv = false; // true if it changes sides
    if ((x >= 0) !== (dir === CircularLayout.Clockwise)) {
      // y decreases
      newy = y - delta;
      if (newy < -yRadius) {
        // y increases
        newy = y + delta;
        if (newy > yRadius) {
          result.x = newx;
          result.y = newy;
          return false;
        }
        inv = true;
      }
    } else {
      // y increases
      newy = y + delta;
      if (newy > yRadius) {
        // y decreases
        newy = y - delta;
        if (newy < -yRadius) {
          result.x = newx;
          result.y = newy;
          return false;
        }
        inv = true;
      }
    }

    newx = Math.sqrt(1 - Math.min(1, newy * newy / (yRadius * yRadius))) * eRadius;
    if ((x < 0) !== inv) {
      newx = -newx;
    }
    if (Math.abs(x - newx) > (verts[v].width + verts[v + 1].width) / 2) {
      result.x = newx;
      result.y = newy;
      return false;
    } else {
      result.x = newx;
      result.y = newy;
      return true;
    }
  }

  /**
   * Position each {@link Node} according to the Vertex position, and then position the {@link Link}s.
   *
   * You should not call this method -- it is a "protected virtual" method.
   * @this {CircularLayout}
   */
  protected commitLayout(): void {
    this.commitNodes();
    if (this.isRouting) this.commitLinks();
  }

  /**
   * Commit the position of all vertex nodes.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CircularLayout}
   */
  protected commitNodes(): void {
    const ingroup = (this.group !== null && this.group.placeholder !== null && this.group.isSubGraphExpanded);
    const oldloc = ingroup ? this.group.location.copy() : null;

    let center = this.actualCenter;
    if (ingroup) {  // nodes and group moved later
      center = new Point(0, 0);
    } else {  // go ahead and commit at final positions
      center.x = this.arrangementOrigin.x + this._eRadius;
      center.y = this.arrangementOrigin.y + this._yRadius;
    }

    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      v.x += center.x;
      v.y += center.y;
      v.commit();
    }

    if (ingroup) {
      this.group.ensureBounds();
      const temppos = this.group.position.copy();
      const temploc = this.group.location.copy();
      const newpos = oldloc.subtract(temploc.subtract(temppos));
      this.group.move(newpos);
      this._actualCenter = newpos.subtract(temppos);
    }
  }

  /**
   * Commit the position and routing of all edge links.
   * This is called by {@link #commitLayout}.
   * This is only called if {@link Layout#isRouting} is true.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {CircularLayout}
   */
  protected commitLinks(): void {
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const e = eit.value;
      e.commit();
    }
  }

  /**
   * @hidden @internal
   * Finds the perimeter of an ellipse with radii a and b.
   * @this {CircularLayout}
   * @param {number} a
   * @param {number} b
   * @param {number=} theta0
   * @param {number=} sweep
   * @return {number}
   */
  private ellipsePerim(a: number, b: number, theta0?: number, sweep?: number): number {
    const riemannnum = this._riemannnum;
    // circular case
    if (Math.abs(this._eAspectRatio - 1) < .001) {
      if (theta0 !== undefined && sweep !== undefined) return sweep * a;
      return 2 * Math.PI * a;
    }

    const e = a > b ? Math.sqrt(a * a - b * b) / a : Math.sqrt(b * b - a * a) / b; // eccentricity
    let integral = 0.0;
    let wid = 0.0;
    if (theta0 !== undefined && sweep !== undefined) {
      wid = sweep / (riemannnum + 1);
    } else {
      wid = Math.PI / (2 * (riemannnum + 1));
    }
    let theta = 0.0;
    for (let i = 0; i <= riemannnum; i++) {
      if (theta0 !== undefined && sweep !== undefined) {
        theta = theta0 + i * sweep / riemannnum;
      } else {
        theta = i * Math.PI / (2 * riemannnum);
      }
      const sin = Math.sin(theta);
      integral += Math.sqrt(1 - e * e * sin * sin) * wid;
    }
    if (theta0 !== undefined && sweep !== undefined) return (a > b ? a : b) * integral;
    return 4 * (a > b ? a : b) * integral;
  }

  /**
   * @hidden @internal
   * @this {CircularLayout}
   * @param {number} circ
   * @param {number} asprat
   * @param {number} theta0
   * @param {number} sweep
   * @return {number}
   */
  private inverseEllipsePerim(circ: number, asprat: number, theta0: number, sweep: number): number {
    let perim1 = 0.0;
    if (theta0 !== undefined && sweep !== undefined) {
      perim1 = this.ellipsePerim(1, asprat, theta0, sweep);
    } else {
      perim1 = this.ellipsePerim(1, asprat);
    }
    return circ / perim1;
  }

  /**
   * @hidden @internal
   * Returns the angle of an elliptical arc of specified length.
   * @this {CircularLayout}
   * @param {number} a
   * @param {number} b
   * @param {number} theta0
   * @param {number} length
   * @return {number}
   */
  private ellipseAngle(a: number, b: number, theta0: number, length: number): number {
    if (Math.abs(this._eAspectRatio - 1) < .001) return length / a; // just added *****

    const e = a > b ? Math.sqrt(a * a - b * b) / a : Math.sqrt(b * b - a * a) / b; // eccentricity
    let integral = 0.0;
    const wid = (2 * Math.PI) / (this.network.vertexes.count * 700);
    if (a > b) { theta0 += Math.PI / 2; }
    for (let i = 0; ; i++) {
      const theta = theta0 + i * wid;
      const sin = Math.sin(theta);
      integral += (a > b ? a : b) * Math.sqrt(1 - e * e * sin * sin) * wid;
      if (integral >= length) {
        return i * wid;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} vertices
   * @return {List.<CircularVertex>}
   */
  private sort(vertices: List<CircularVertex>): List<CircularVertex> {
    switch (this.sorting) {
      case CircularLayout.Forwards: break;
      case CircularLayout.Reverse: vertices.reverse(); break;
      case CircularLayout.Ascending: vertices.sort(this.comparer); break;
      case CircularLayout.Descending: vertices.sort(this.comparer); vertices.reverse(); break;
      case CircularLayout.Optimized: return this.optimize(this.connectivityArrange(vertices));
      default: Util.throwError('Invalid sorting type.');
    }
    return vertices;
  }

  /**
   * @hidden @internal
   * Sort the nodes based on connectivity. The first node is the one with the highest connectivity, and
   * each subsequent node is the one with the greatest number of links to nodes already in the list.
   * This is used for cross reduction before optimize is called to make cross reduction more effective.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} l
   * @return {List.<CircularVertex>}
   */
  private connectivityArrange(l: List<CircularVertex>): List<CircularVertex> {
    const conns: Array<number> = [];
    for (let i = 0; i < l.length; i++) {
      conns.push(0);
    }
    const res = new List<CircularVertex>();
    for (let i = 0; i < l.length; i++) {
      let maxconn = -1;
      let maxind = -1;
      if (i === 0) {
        for (let j = 0; j < l.length; j++) {
          const vtx = l.elt(j);
          // var nconn = vtx.sourceEdges.count + vtx.destinationEdges.count;
          const nconn = vtx.edgesCount;
          if (nconn > maxconn) {
            maxconn = nconn;
            maxind = j;
          }
        }
      } else {
        for (let j = 0; j < l.length; j++) {
          const nconn2 = conns[j];
          if (nconn2 > maxconn) {
            maxconn = nconn2;
            maxind = j;
          }
        }
      }
      res.add(l.elt(maxind));
      conns[maxind] = -1;

      const v = l.elt(maxind);
      let ind = 0;
      const seit = v.sourceEdges;
      while (seit.next()) {
        const se = seit.value;
        const f = se.fromVertex as CircularVertex;
        ind = l.indexOf(f);
        if (ind < 0) continue;
        if (conns[ind] >= 0) conns[ind]++;
      }
      const deit = v.destinationEdges;
      while (deit.next()) {
        const de = deit.value;
        const t = de.toVertex as CircularVertex;
        ind = l.indexOf(t);
        if (ind < 0) continue;
        if (conns[ind] >= 0) conns[ind]++;
      }
    }
    return res;
  }

  /**
   * @hidden @internal
   * Sorts the nodes to avoid crossing links.
   * @this {CircularLayout}
   * @param {List.<CircularVertex>} v
   * @return {List.<CircularVertex>}
   */
  private optimize(v: List<CircularVertex>): List<CircularVertex> {
    const linkswith: Array<Array<number>> = [];
    for (let i = 0; i < v.length; i++) {
      const n = v.elt(i);
      linkswith[i] = [];
      let ind = 0;
      const deit = n.destinationEdges;
      while (deit.next()) {
        const dv = deit.value.toVertex as CircularVertex;
        ind = v.indexOf(dv);
        if (ind !== i && linkswith[i].indexOf(ind) < 0) linkswith[i].push(ind);
      }
      const seit = n.sourceEdges;
      while (seit.next()) {
        const sv = seit.value.fromVertex as CircularVertex;
        ind = v.indexOf(sv);
        if (ind !== i && linkswith[i].indexOf(ind) < 0) linkswith[i].push(ind);
      }
    }

    const indexes: Array<number> = [];
    for (let i = 0; i < linkswith.length; i++) {
      indexes[i] = 0;
    }
    const opt: Array<number> = [];
    const links1: Array<number> = [];
    const links2: Array<number> = [];
    const monoconnected: Array<number> = [];
    const disconnected = new List<CircularVertex>();

    let j = 0;
    for (let i = 0; i < linkswith.length; i++) {
      const conns = linkswith[i].length;
      if (conns === 1) {
        monoconnected.push(i);
        continue;
      }
      if (conns === 0) {
        disconnected.add(v.elt(i));
        continue;
      }

      if (j === 0) {
        opt.push(i);
        j++;
        continue;
      }

      let mincrosses = Infinity;
      let mindist = Infinity;
      let minpos = -1;
      const newposses: Array<number> = [];
      for (let k = 0; k < opt.length; k++) {
        if (linkswith[opt[k]].indexOf(opt[k === opt.length - 1 ? 0 : k + 1]) < 0) {
          newposses.push(k === opt.length - 1 ? 0 : k + 1);
        }
      }
      if (newposses.length === 0) {
        for (let k = 0; k < opt.length; k++) {
          newposses.push(k);
        }
      }
      for (let k = 0; k < newposses.length; k++) {
        const newpos = newposses[k];
        const crosses = this.crossings(linkswith[i], links1, links2, indexes, newpos, opt);

        let dist = 0;
        for (let l = 0; l < linkswith[i].length; l++) {
          const conn = linkswith[i][l];
          const ind = opt.indexOf(conn);
          if (ind >= 0) {
            const absdist = Math.abs(newpos - (ind >= newpos ? ind + 1 : ind));
            dist += absdist < opt.length + 1 - absdist ? absdist : opt.length + 1 - absdist;
          }
        }
        for (let l = 0; l < links1.length; l++) {
          let a = indexes[links1[l]];
          let b = indexes[links2[l]];  // a < b
          if (a >= newpos) a++;
          if (b >= newpos) b++;
          if (a > b) {
            const c = b;
            b = a;
            a = c;

          }
          if ((b - a < (opt.length + 2) / 2) === (a < newpos && newpos <= b)) {
            dist++;
          }
        }

        if (crosses < mincrosses || (crosses === mincrosses && dist < mindist)) {
          mincrosses = crosses;
          mindist = dist;
          minpos = newpos;
        }
      }
      opt.splice(minpos, 0, i);

      // update indexes
      for (let k = 0; k < opt.length; k++) {
        indexes[opt[k]] = k;
      }

      // update links
      for (let k = 0; k < linkswith[i].length; k++) {
        const link = linkswith[i][k];
        if (opt.indexOf(link) >= 0) {
          links1.push(i);
          links2.push(link);
        }
      }
      j++;
    }

    // add monoconnected
    let done = false;
    const cnt = opt.length;
    while (true) {
      done = true;
      for (let n2 = 0; n2 < monoconnected.length; n2++) {
        const i2 = monoconnected[n2];
        const tonode = linkswith[i2][0];
        const ind = opt.indexOf(tonode);
        if (ind >= 0) {
          // insert either before or after
          let low = 0;  // links to node at lower index
          for (let jj = 0; jj < linkswith[tonode].length; jj++) {
            const link = linkswith[tonode][jj];
            const linkind = opt.indexOf(link);
            if (linkind < 0 || linkind === ind) continue;
            const d1 = linkind > ind ? linkind - ind : ind - linkind;
            const d2 = cnt - d1;
            low += ((linkind < ind) !== (d1 > d2)) ? 1 : -1;
          }
          opt.splice(low < 0 ? ind : ind + 1, 0, i2);
          monoconnected.splice(n2, 1);
          n2--;
        } else {
          done = false;
        }
      }
      if (done) {
        break;
      } else {
        opt.push(monoconnected[0]);
        monoconnected.splice(0, 1);
      }
    }

    for (let i = 0; i < opt.length; i++) {
      const ind2 = opt[i];
      disconnected.add(v.elt(ind2));
    }
    return disconnected;
  }

  /**
   * @hidden @internal
   * Finds the number of link crossings added when a node is inserted at a specified index.
   * @this {CircularLayout}
   * @param {Array.<number>} newlinks List<int>
   * @param {Array.<number>} links1 List<int>
   * @param {Array.<number>} links2 List<int>
   * @param {Array.<number>} indexes List<int>
   * @param {number} node int
   * @param {Array.<number>} opt List<int>
   * @return {number} int
   */
  private crossings(newlinks: Array<number>, links1: Array<number>, links2: Array<number>, indexes: Array<number>, node: number, opt: Array<number>): number {
    let count = 0;
    for (let i = 0; i < links1.length; i++) {
      const l1 = links1[i];
      const l2 = links2[i];
      const l1pos = indexes[l1];
      const l2pos = indexes[l2];
      let c = 0;
      let d = 0;
      if (l1pos < l2pos) {
        c = l1pos;
        d = l2pos;
      } else {
        c = l2pos;
        d = l1pos;
      }
      // c < d

      if (c < node && node <= d) {
        // node between c and d
        for (let j = 0; j < newlinks.length; j++) {
          const link = newlinks[j];
          if (opt.indexOf(link) < 0) continue;
          if (!((c < indexes[link] && indexes[link] < d) || c === indexes[link] || d === indexes[link])) count++;
        }
      } else {
        // node not between c and d
        for (let j = 0; j < newlinks.length; j++) {
          const link = newlinks[j];
          if (opt.indexOf(link) < 0) continue;
          if (!(!(c < indexes[link] && indexes[link] < d) || c === indexes[link] || d === indexes[link])) count++;
        }
      }
    }
    return count;
  }

  // ***************Properties***************

  /**
   * Gets or sets the horizontal radius of the elliptical arrangement.
   *
   * The default value is NaN.
   * NaN indicates that the {@link #spacing} will determine the size of the ring.
   * If {@link #spacing} is also NaN, the effective spacing will be 6.
   * If {@link #spacing} is a number, the effective radius will be > radius if and only if
   * the spacing between elements would otherwise be less than spacing.
   * The specified value for {@link #radius} will be ignored if {@link #arrangement} === {@link CircularLayout.Packed}.
   * This property must always be positive or NaN.
   */
  /** @export */ get radius(): number { return this._radius; }
  /** @export */ set radius(value: number) {
    if (this._radius !== value) {
      Util.checkPrimitive(value, 'number', CircularLayout, 'radius');
      if (value > 0 || isNaN(value)) {
        this._radius = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the ratio of the arrangement's height to its width
   * (1 for a circle, &gt;1 for a vertically elongated ellipse).
   *
   * This is 1 by default.
   * The value must be a positive number.
   *
   * Modifying this value changes the height, but keeps the width and the {@link #radius} constant.
   */
  /** @export */ get aspectRatio(): number { return this._aspectRatio; }
  /** @export */ set aspectRatio(value: number) {
    if (this._aspectRatio !== value) {
      Util.checkPrimitive(value, 'number', CircularLayout, 'aspectRatio');
      if (value > 0) {
        this._aspectRatio = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the angle (in degrees, clockwise from the positive side of the X axis) of the first element.
   *
   * The default value is 0.
   */
  /** @export */ get startAngle(): number { return this._startAngle; }
  /** @export */ set startAngle(value: number) {
    if (this._startAngle !== value) {
      Util.checkPrimitive(value, 'number', CircularLayout, 'startAngle');
      this._startAngle = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the absolute angle (in degrees) between the first and last node.
   *
   * The default value is 360.
   * The value must be greater than zero and less than or equal to 360.
   * If it is not in this range, it will be automatically set to 360.
   *
   * Whether the arrangement is clockwise or counterclockwise does not depend on the sign of this value.
   * The direction can be controlled by setting {@link #direction}.
   * If 360 is the specified value, the actual value will be less to keep the first and last
   * elements from overlapping, and the spacing between the first and last nodes will be determined
   * the same way as for all other adjacent nodes.
   */
  /** @export */ get sweepAngle(): number { return this._sweepAngle; }
  /** @export */ set sweepAngle(value: number) {
    if (this._sweepAngle !== value) {
      Util.checkPrimitive(value, 'number', CircularLayout, 'sweepAngle');
      if (value > 0 && value <= 360) {
        this._sweepAngle = value;
      } else {
        this._sweepAngle = 360;
      }
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets how the nodes are spaced.
   * If {@link #arrangement} === {@link CircularLayout.Packed},
   * the specified {@link #radius} will be ignored.
   *
   * The default value is {@link CircularLayout.ConstantSpacing}.
   */
  /** @export */ get arrangement(): EnumValue { return this._arrangement; }
  /** @export */ set arrangement(value: EnumValue) {
    if (this._arrangement !== value) {
      EnumValue.checkEnumValue(value, CircularLayout, CircularLayout, 'arrangement');
      if (value === CircularLayout.Packed || value === CircularLayout.ConstantSpacing ||
        value === CircularLayout.ConstantDistance || value === CircularLayout.ConstantAngle) {
        this._arrangement = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets whether the nodes are arranged clockwise or counterclockwise.
   *
   * The default value is {@link CircularLayout.Clockwise}.
   */
  /** @export */ get direction(): EnumValue { return this._direction; }
  /** @export */ set direction(value: EnumValue) {
    if (this._direction !== value) {
      EnumValue.checkEnumValue(value, CircularLayout, CircularLayout, 'direction');
      if (value === CircularLayout.Clockwise || value === CircularLayout.Counterclockwise ||
        value === CircularLayout.BidirectionalLeft || value === CircularLayout.BidirectionalRight) {
        this._direction = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets if and how the nodes are sorted.
   *
   * {@link CircularLayout.Forwards} indicates that the nodes are arranged in the order the layout gets them.
   * {@link CircularLayout.Reverse} indicates that the nodes are arranged in the reverse order that the layout gets them.
   * {@link CircularLayout.Ascending} and {@link CircularLayout.Descending} indicate that the nodes
   * will be sorted using the {@link #comparer}.
   * {@link CircularLayout.Optimized} indicates that the nodes will be arranged to minimize link crossings.
   *
   * The default value is {@link CircularLayout.Optimized}.
   */
  /** @export */ get sorting(): EnumValue { return this._sorting; }
  /** @export */ set sorting(value: EnumValue) {
    if (this._sorting !== value) {
      EnumValue.checkEnumValue(value, CircularLayout, CircularLayout, 'sorting');
      if (value === CircularLayout.Forwards || value === CircularLayout.Reverse ||
        value === CircularLayout.Ascending || CircularLayout.Descending ||
        value === CircularLayout.Optimized) {
        this._sorting = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the comparer which sorts the data when {@link #sorting} is
   * set to {@link CircularLayout.Ascending} or {@link CircularLayout.Descending}.
   *
   * The default function compares the {@link Part#text} values of the vertexes' {@link LayoutVertex#node}s.
   */
  /** @export */ get comparer(): ((a: CircularVertex, b: CircularVertex) => number) { return this._comparer; }
  /** @export */ set comparer(value: ((a: CircularVertex, b: CircularVertex) => number)) {
    if (this._comparer !== value) {
      Util.checkPrimitive(value, 'function', CircularLayout, 'comparer');
      this._comparer = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the distance between nodes (if {@link #radius} is NaN)
   * or the minimum distance between nodes (if {@link #radius} is a number).
   *
   * The default value is 6.
   * The value may be NaN.
   *
   * If {@link #spacing} is NaN, there is no minimum spacing, allowing nodes to overlap,
   * unless {@link #radius} is NaN,
   * in which case the effective spacing will be 6 to determine an effective radius.
   * If {@link #spacing} is a number but {@link #radius} isn't,
   * the effective spacing will be spacing, and this will determine the effective radius.
   * If both {@link #spacing} and {@link #radius} are numbers,
   * the effective radius will be at least {@link #radius},
   * but may be larger so that the minimum spacing between nodes is {@link #spacing}.
   */
  /** @export */ get spacing(): number { return this._spacing; }
  /** @export */ set spacing(value: number) {
    if (this._spacing !== value) {
      Util.checkPrimitive(value, 'number', CircularLayout, 'spacing');
      this._spacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Specifies how the diameter of nodes will be calculated.
   * When a node is not circular, it is not clear what its diameter is.
   *
   * The default is {@link CircularLayout.Pythagorean}.
   */
  /** @export */ get nodeDiameterFormula(): EnumValue { return this._nodeDiameterFormula; }
  /** @export */ set nodeDiameterFormula(value: EnumValue) {
    if (this._nodeDiameterFormula !== value) {
      EnumValue.checkEnumValue(value, CircularLayout, CircularLayout, 'nodeDiameterFormula');
      if (value === CircularLayout.Pythagorean || value === CircularLayout.Circular) {
        this._nodeDiameterFormula = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * This read-only property is the effective X radius that may have been calculated by the layout.
   */
  /** @export */ get actualXRadius(): number { return this._eRadius; }

  /**
   * This read-only property is the effective Y radius that may have been calculated by the layout.
   */
  /** @export */ get actualYRadius(): number { return this._yRadius; }

  /**
   * This read-only property is the effective spacing that may have been calculated by the layout.
   */
  /** @export */ get actualSpacing(): number { return this._eSpacing; }

  /**
   * This read-only property is the coordinates of the center of the laid-out ellipse immediately after the layout.
   */
  /** @export */ get actualCenter(): Point { return this._actualCenter; }

  // ***************Enums***************

  // These values specify how nodes should be spaced in the ring.
  /**
   * The spacing between the idealized boundaries of the nodes is constant;
   * This value is used for {@link CircularLayout#arrangement}.
   * @constant
   */
  public static ConstantSpacing = new EnumValue(CircularLayout, 'ConstantSpacing', 0);

  /**
   * The distance between the centers of the nodes is constant;
   * This value is used for {@link CircularLayout#arrangement}.
   * @constant
   */
  public static ConstantDistance = new EnumValue(CircularLayout, 'ConstantDistance', 1);

  /**
   * The angular distance between the nodes is constant;
   * This value is used for {@link CircularLayout#arrangement}.
   * @constant
   */
  public static ConstantAngle = new EnumValue(CircularLayout, 'ConstantAngle', 2);

  /**
   * The vertices are arranged as close together as possible considering the {@link CircularLayout#spacing},
   * assuming the nodes are rectangular;
   * This value is used for {@link CircularLayout#arrangement}.
   * @constant
   */
  public static Packed = new EnumValue(CircularLayout, 'Packed', 3);

  // These values represent the direction in which the nodes fill the ring.
  /**
   * Rings are filled clockwise;
   * This value is used for {@link CircularLayout#direction}.
   * @constant
   */
  public static Clockwise = new EnumValue(CircularLayout, 'Clockwise', 4);

  /**
   * Rings are filled counterclockwise;
   * This value is used for {@link CircularLayout#direction}.
   * @constant
   */
  public static Counterclockwise = new EnumValue(CircularLayout, 'Counterclockwise', 5);

  /**
   * The ring is filled by alternating sides; the second node is counterclockwise from the first node;
   * This value is used for {@link CircularLayout#direction}.
   * @constant
   */
  public static BidirectionalLeft = new EnumValue(CircularLayout, 'BidirectionalLeft', 6);

  /**
   * The ring is filled by alternating sides; the second node is clockwise from the first node;
   * This value is used for {@link CircularLayout#direction}.
   * @constant
   */
  public static BidirectionalRight = new EnumValue(CircularLayout, 'BidirectionalRight', 7);

  // These values specify how to sort the nodes.
  /**
   * Nodes are arranged in the order given;
   * This value is used for {@link CircularLayout#sorting}.
   * @constant
   */
  public static Forwards = new EnumValue(CircularLayout, 'Forwards', 8);

  /**
   * Nodes are arranged in the reverse of the order given;
   * This value is used for {@link CircularLayout#sorting}.
   * @constant
   */
  public static Reverse = new EnumValue(CircularLayout, 'Reverse', 9);

  /**
   * Nodes are sorted using the {@link #comparer}, in ascending order;
   * This value is used for {@link CircularLayout#sorting}.
   * @constant
   */
  public static Ascending = new EnumValue(CircularLayout, 'Ascending', 10);

  /**
   * Nodes are sorted using the {@link #comparer}, in reverse ascending (descending) order;
   * This value is used for {@link CircularLayout#sorting}.
   * @constant
   */
  public static Descending = new EnumValue(CircularLayout, 'Descending', 11);

  /**
   * Nodes are ordered to reduce link crossings;
   * This value is used for {@link CircularLayout#sorting}.
   * @constant
   */
  public static Optimized = new EnumValue(CircularLayout, 'Optimized', 12);

  // These values specify a method for finding the size of a node.
  /**
   * The effective diameter is sqrt(width^2+height^2);
   * The corners of square nodes will touch at 45 degrees when {@link CircularLayout#spacing} is 0;
   * This value is used for {@link CircularLayout#nodeDiameterFormula}.
   * @constant
   */
  public static Pythagorean = new EnumValue(CircularLayout, 'Pythagorean', 13);

  /**
   * The effective diameter is either the width or height of the node, whichever is larger;
   * This will cause circular nodes to touch when {@link CircularLayout#spacing} is 0;
   * This is ideal when the nodes are circular.
   * This value is used for {@link CircularLayout#nodeDiameterFormula}.
   * @constant
   */
  public static Circular = new EnumValue(CircularLayout, 'Circular', 14);

} // #end CircularLayout
CircularLayout['className'] = 'CircularLayout';

CircularLayout['ConstantSpacing'] = CircularLayout.ConstantSpacing;
CircularLayout['ConstantDistance'] = CircularLayout.ConstantDistance;
CircularLayout['ConstantAngle'] = CircularLayout.ConstantAngle;
CircularLayout['Packed'] = CircularLayout.Packed;
CircularLayout['Clockwise'] = CircularLayout.Clockwise;
CircularLayout['Counterclockwise'] = CircularLayout.Counterclockwise;
CircularLayout['BidirectionalLeft'] = CircularLayout.BidirectionalLeft;
CircularLayout['BidirectionalRight'] = CircularLayout.BidirectionalRight;
CircularLayout['Forwards'] = CircularLayout.Forwards;
CircularLayout['Reverse'] = CircularLayout.Reverse;
CircularLayout['Ascending'] = CircularLayout.Ascending;
CircularLayout['Descending'] = CircularLayout.Descending;
CircularLayout['Optimized'] = CircularLayout.Optimized;
CircularLayout['Pythagorean'] = CircularLayout.Pythagorean;
CircularLayout['Circular'] = CircularLayout.Circular;

/**
 * @hidden @internal
 * @unrestricted
 */
export class VertexArrangement {
  /** @hidden @internal */ public static className = 'VertexArrangement';
  /** @hidden @internal */ private gap: number;
  /** @hidden @internal */ private xcoords: Array<number>;
  /** @hidden @internal */ private ycoords: Array<number>;

  constructor() {
    this.gap = -Infinity;
    this.xcoords = null;
    this.ycoords = null;
  }

  /**
   * @hidden @internal
   * Compares a new gap with the current one.  If it's better, it updates gap, xcoords, and ycoords.
   * This is for a full sweep, where overshooting (gap less than 0) is worse than undershooting (gap > 0)
   * because it will cause nodes to overlap.
   * @this {VertexArrangement}
   * @param {number} gap
   * @param {Array.<CircularVertex>} verts
   */
  public compare(gap: number, verts: Array<CircularVertex>): void {
    if ((gap > 0 && this.gap < 0) || (Math.abs(gap) < Math.abs(this.gap) && !(gap < 0 && this.gap > 0))) {
      this.gap = gap;
      this.xcoords = [];
      this.ycoords = [];
      for (let i = 0; i < verts.length; i++) {
        this.xcoords[i] = verts[i].bounds.x;
        this.ycoords[i] = verts[i].bounds.y;
      }
    }
  }

  /**
   * @hidden @internal
   * Compares a new gap with the current one.  If it's better, it updates gap, xcoords, and ycoords.
   * This is for a partial sweep, where overshooting (gap less than 0) is no worse than undershooting (gap > 0).
   * @this {VertexArrangement}
   * @param {number} gap
   * @param {Array.<CircularVertex>} verts
   */
  public compare2(gap: number, verts: Array<CircularVertex>): void {
    if (Math.abs(gap) < Math.abs(this.gap)) {
      this.gap = gap;
      this.xcoords = [];
      this.ycoords = [];
      for (let i = 0; i < verts.length; i++) {
        this.xcoords[i] = verts[i].bounds.x;
        this.ycoords[i] = verts[i].bounds.y;
      }
    }
  }

  /**
   * @hidden @internal
   * Set the vertices in verts to the correct positions.
   * @this {VertexArrangement}
   * @param {List.<CircularVertex>} verts
   */
  public commit(verts: List<CircularVertex>): void {
    if (this.xcoords === null || this.ycoords === null) return;
    for (let i = 0; i < this.xcoords.length; i++) {
      // verts.elt(i).position = new Point(this.xcoords[i], this.ycoords[i]);
      const v = verts.elt(i);
      v.x = this.xcoords[i];
      v.y = this.ycoords[i];
    }
  }

} // #end VertexArrangement
VertexArrangement['className'] = 'VertexArrangement';

// ***************CircularNetwork***************

/**
 * This class represents an abstract graph of {@link CircularVertex}es and {@link CircularEdge}s
 * that can be constructed based on the {@link Node}s and {@link Link}s of a {@link Diagram}
 * so that the {@link CircularLayout} can operate independently of the diagram until it
 * is time to commit any node positioning or link routing.
 *
 * This class inherits from {@link LayoutNetwork}.
 * @extends LayoutNetwork
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class CircularNetwork extends LayoutNetwork {
  /** @hidden @internal */ public static className = 'CircularNetwork';

  constructor(layout: CircularLayout) {
    super(layout);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link CircularVertex}.
   * @this {CircularNetwork}
   * @return {CircularVertex}
   */
  public createVertex(): CircularVertex {
    return new CircularVertex(this);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link CircularEdge}.
   * @this {CircularNetwork}
   * @return {CircularEdge}
   */
  public createEdge(): CircularEdge {
    return new CircularEdge(this);
  }

} // #end CircularNetwork
CircularNetwork['className'] = 'CircularNetwork';

/**
 * This holds {@link CircularLayout}-specific information about {@link Node}s.
 *
 * This class inherits from {@link LayoutVertex}.
 * @extends LayoutVertex
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class CircularVertex extends LayoutVertex {
  /** @hidden @internal */ public static className = 'CircularVertex';
  /** @hidden @internal */ private _diameter: number;
  /** @hidden @internal */ private _actualAngle: number;

  constructor(network: CircularNetwork) {
    super(network);

    // circular-specific properties:
    this._diameter = NaN;

    this._actualAngle = NaN;
  }

  /**
   * @hidden @internal
   * Finds the effective diameter of this CircularVertex, which may depend on the angle
   * at which the ellipse is being filled.
   * @this {CircularVertex}
   * @param {number} angle
   * @return {number}
   */
  public computeDiameter(angle: number): number {
    const network = this.network;
    if (network === null) return NaN;
    const layout = network.layout as CircularLayout;
    if (layout === null) return NaN;
    if (layout.arrangement === CircularLayout.Packed) {
      if (layout.nodeDiameterFormula === CircularLayout.Circular) {
        this._diameter = Math.max(this.width, this.height);
        return this._diameter;
      } else {
        const sin = Math.abs(Math.sin(angle));
        const cos = Math.abs(Math.cos(angle));
        if (sin === 0) return this.width;
        if (cos === 0) return this.height;
        this._diameter = Math.min(this.height / sin, this.width / cos);
        return this._diameter;
      }
    } else {
      if (layout.nodeDiameterFormula === CircularLayout.Circular) {
        this._diameter = Math.max(this.width, this.height);
        return this._diameter;
      } else {
        this._diameter = Math.sqrt(this.width * this.width + this.height * this.height);
        return this._diameter;
      }
    }
  }

  /**
   * Gets or sets the value used as the vertex's diameter.
   *
   * By default the value depends on the {@link CircularLayout#arrangement} property.
   * Any computed value is cached, to avoid unnecessary expensive computations.
   */
  /** @export */ get diameter(): number { return this._diameter; }
  /** @export */ set diameter(value: number) {
    if (this._diameter !== value) {
      Util.checkPrimitive(value, 'number', CircularVertex, 'diameter');
      this._diameter = value;
    }
  }

  /**
   * Gets or sets the value used as the vertex's angle.
   */
  /** @export */ get actualAngle(): number { return this._actualAngle; }
  /** @export */ set actualAngle(value: number) {
    if (this._actualAngle !== value) {
      Util.checkPrimitive(value, 'number', CircularVertex, 'actualAngle');
      this._actualAngle = value;
    }
  }

} // #end CircularVertex
CircularVertex['className'] = 'CircularVertex';

/**
 * This holds {@link CircularLayout}-specific information about {@link Link}s.
 *
 * This class inherits from {@link LayoutEdge}.
 * @extends LayoutEdge
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class CircularEdge extends LayoutEdge {
  /** @hidden @internal */ public static className = 'CircularEdge';
  constructor(network: CircularNetwork) {
    super(network);
    // no circular-specific properties
  }
} // #end CircularEdge
CircularEdge['className'] = 'CircularEdge';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
ForceDirectedLayout
ForceDirectedNetwork
ForceDirectedVertex
ForceDirectedEdge
*/

// import { List, Set } from './collections';
// import { Iterable, Iterator } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { Geo } from './geo';
// import { Layout, LayoutEdge, LayoutNetwork, LayoutVertex } from './layout';
// import { Group } from './parts';
// import { Part } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * Force-directed layout treats the graph as if it were a system of physical bodies
 * with repulsive electrical, attractional gravitational, and spring forces acting on them and between them.
 *
 * Electrical forces come both from the field at the vertex's location as well as from neighboring vertexes and are quadratic by distance.
 * Gravitational forces come from the field at the vertex's location and are constant.
 * Spring forces are only exerted between two different vertexes that are connected by an edge and are linear by distance.
 *
 * The electrical forces on a vertex are the sum of the electrical charge times the electrical field at that location
 * ({@link #electricalCharge}, {@link #electricalFieldX}, {@link #electricalFieldY})
 * and the electrical forces of all nearby vertexes divided by the square of the distance between them.
 * You can easily assign the electrical charge for all vertexes by assigning {@link #defaultElectricalCharge}.
 * By default there is no electrical field, so all forces are due to nearby charged vertexes.
 * For efficiency, {@link #infinityDistance} determines a cut-off distance between vertexes for which to consider any influence.
 *
 * The gravitational forces on a vertex are the sum of the gravitational mass times the gravitational field at that location
 * ({@link #gravitationalMass}, {@link #gravitationalFieldX}, {@link #gravitationalFieldY}).
 * You can easily assign the gravitational mass for all vertexes by assigning {@link #defaultGravitationalMass}.
 * By default there is no gravitational field.
 *
 * The spring forces on a vertex are only exerted by the edges connecting it with other vertexes.
 * The force along an edge is the stiffness of the spring times the difference of the distance between the vertexes and the nominal length of the spring
 * ({@link #springStiffness}, {@link #springLength}) divided by the distance between the vertexes.
 * When the distance is less than the nominal length, the force pushes the vertexes apart; when the distance is greater, the force pulls them together.
 * You can easily assign the spring length and stiffness for all edges by assigning {@link #defaultSpringLength} and {@link #defaultSpringStiffness}.
 *
 * When the distance between two vertexes is less than one unit, this uses a random number generator to decide which direction the forces should go.
 * For layouts that start with all of the vertexes at the same location, this results in potentially dramatically different results.
 * Set {@link #randomNumberGenerator} to null in order to produce reproducible results given the same initial vertex locations.
 *
 * The algorithm seeks a configuration of the bodies with locally minimal energy,
 * i.e. vertex positions such that the sum of the forces on each vertex is zero.
 * This is achieved by repeatedly computing the forces on each vertex, moving them, and repeating.
 * Computations stop when no vertex moves more than {@link #epsilonDistance} or when {@link #maxIterations} have happened.
 *
 * The layout cannot guarantee that it provides optimal positioning of nodes.
 * Nodes will normally not overlap each other, but when there is a dense interconnectivity overlaps might not be avoidable.
 * <p class="boxrun">
 * If you want to experiment interactively with most of the properties, try the <a href="../../samples/fdLayout.html">Force Directed Layout</a> sample.
 * See samples that make use of ForceDirectedLayout in the <a href="../../samples/index.html#forcedirectedlayout">samples index</a>.
 *
 * This layout makes use of a {@link LayoutNetwork} of
 * {@link ForceDirectedVertex}es and {@link ForceDirectedEdge}s that normally
 * correspond to the {@link Node}s and {@link Link}s of the {@link Diagram}.
 * @extends Layout
 * @unrestricted
 * @category Layout
 */
export class ForceDirectedLayout extends Layout {
  /** @hidden @internal */ public static className = 'ForceDirectedLayout';
  /** @hidden @internal */ private _vertexArray: Array<ForceDirectedVertex> | null;
  /** @hidden @internal */ private _iteration: number;
  /** @hidden @internal */ private _arrangementSpacing: Size;
  /** @hidden @internal */ private _arrangesToOrigin: boolean;
  /** @hidden @internal */ private _setsPortSpots: boolean;
  /** @hidden @internal */ private _comments: boolean;
  /** @hidden @internal */ private _maxIterations: number;
  /** @hidden @internal */ private _epsilonDistance: number;
  /** @hidden @internal */ private _infinityDistance: number;
  /** @hidden @internal */ private _moveLimit: number;
  /** @hidden @internal */ private _randomNumberGenerator: ({ random: () => number } | null);
  /** @hidden @internal */ private _defaultSpringStiffness: number;
  /** @hidden @internal */ private _defaultSpringLength: number;
  /** @hidden @internal */ private _defaultElectricalCharge: number;
  /** @hidden @internal */ private _defaultGravitationalMass: number;
  /** @hidden @internal */ private _defaultCommentSpringLength: number;
  /** @hidden @internal */ private _defaultCommentElectricalCharge: number;

  /**
   * Constructs a {@link ForceDirectedLayout} with no {@link Layout#network}
   * and with no owning {@link Layout#diagram}.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(ForceDirectedLayout);
    super();

    this._vertexArray = null;
    this._iteration = 0;
    this._arrangementSpacing = new Size(100, 100).freeze();
    this._arrangesToOrigin = false;
    this._setsPortSpots = true;
    this._comments = false;
    this._maxIterations = 100;
    this._epsilonDistance = 1;
    this._infinityDistance = 1000;
    this._moveLimit = 10;
    this._randomNumberGenerator = Math;
    this._defaultSpringStiffness = 0.05;
    this._defaultSpringLength = 50;
    this._defaultElectricalCharge = 150;
    this._defaultGravitationalMass = 0;
    this._defaultCommentSpringLength = 10;
    this._defaultCommentElectricalCharge = 5;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Layout.
   * @this {ForceDirectedLayout}
   * @param {?} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    // these are computed:
    // this._vertexArray
    // this._iteration
    copy._arrangementSpacing.assign(this._arrangementSpacing);
    copy._arrangesToOrigin = this._arrangesToOrigin;
    copy._setsPortSpots = this._setsPortSpots;
    copy._comments = this._comments;
    copy._maxIterations = this._maxIterations;
    copy._epsilonDistance = this._epsilonDistance;
    copy._infinityDistance = this._infinityDistance;
    copy._moveLimit = this._moveLimit;
    copy._randomNumberGenerator = this._randomNumberGenerator;
    copy._defaultSpringStiffness = this._defaultSpringStiffness;
    copy._defaultSpringLength = this._defaultSpringLength;
    copy._defaultElectricalCharge = this._defaultElectricalCharge;
    copy._defaultGravitationalMass = this._defaultGravitationalMass;
    copy._defaultCommentSpringLength = this._defaultCommentSpringLength;
    copy._defaultCommentElectricalCharge = this._defaultCommentElectricalCharge;
  }

  /**
   * Create a new {@link LayoutNetwork} of {@link ForceDirectedVertex}es and {@link ForceDirectedEdge}s.
   * @this {ForceDirectedLayout}
   * @return {ForceDirectedNetwork} a new {@link LayoutNetwork}.
   */
  public createNetwork(): ForceDirectedNetwork {
    return new ForceDirectedNetwork(this);
  }

  /**
   * Perform the force-directed layout.
   *
   * If there is no {@link Layout#network}, this calls {@link #makeNetwork} to create a {@link LayoutNetwork} from the given collection of Parts.
   * This removes any reflexive edges in the network, since they should be ignored.
   *
   * For each vertex this calls and remembers the result of {@link #electricalCharge} as the {@link ForceDirectedVertex#charge}
   * and the result of {@link #gravitationalMass} as the {@link ForceDirectedVertex#mass}.
   *
   * For each edge this calls and remembers the result of {@link #springStiffness} as the {@link ForceDirectedEdge#stiffness}
   * and the result of {@link #springLength} as the {@link ForceDirectedEdge#length}.
   *
   * This then iterates, updating the position of each vertex according to the forces upon it,
   * until reaching {@link #maxIterations} or until no vertex moves more than about {@link #epsilonDistance}.
   *
   * Finally this calls {@link Layout#updateParts} to commit the Node positions from the vertex positions.
   * {@link Layout#updateParts} calls {@link #commitLayout} within a transaction.
   * @this {ForceDirectedLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');
    if (this.network === null) {
      this.network = this.makeNetwork(coll);
    }

    const maxiter = this.maxIterations;
    if (this.network.vertexes.count > 0) {
      // Remove self-links from the input network.
      this.network.deleteSelfEdges();

      // Augment each node in the input network with
      //  ForceDirected auxilary data.
      const vit = this.network.vertexes.iterator;
      while (vit.next()) {
        const vertex = vit.value as ForceDirectedVertex;
        // Record the charge of the node.
        vertex.charge = this.electricalCharge(vertex);
        // Record the mass of the node.
        vertex.mass = this.gravitationalMass(vertex);
      }

      // Augment each link in the input network with
      //  ForceDirected auxilary data.
      const eit = this.network.edges.iterator;
      while (eit.next()) {
        const edge = eit.value as ForceDirectedEdge;
        // Record the stiffness of the link.
        edge.stiffness = this.springStiffness(edge);
        // Record the length of the link.
        edge.length = this.springLength(edge);
      }

      // maybe look for "Comment" nodes to drag along with a vertex/Node
      this.annotateAll();

      this._iteration = 0;
      if (this.needsClusterLayout()) {
        // split into subnetworks, plus the original network holding any singleton nodes
        const net = this.network as ForceDirectedNetwork;
        const subnets = net.splitIntoSubNetworks().iterator as Iterator<ForceDirectedNetwork>;
        // layout each subnetwork independently
        while (subnets.next()) {
          const sn = subnets.value;
          this.network = sn;
          this.initializeClustering();
          this.layoutClusters(0, maxiter);
        }
        this.network = net;
        // layout each subnetwork so that they don't overlap and are separated by InfinityDistance
        subnets.reset();
        this.arrangeConnectedGraphs(subnets, net);
        // merge back into original network
        subnets.reset();
        while (subnets.next()) {
          const subnet = subnets.value;
          const svit = subnet.vertexes.iterator;
          while (svit.next()) {
            const sv = svit.value;
            net.addVertex(sv);
          }
          const seit = subnet.edges.iterator;
          while (seit.next()) {
            const se = seit.value;
            net.addEdge(se);
          }
        }
      }
      // do normal force-directed calculations
      this.performIterations(maxiter);

      // Update the "physical" positions of the nodes and links.
      this.updateParts();
    }

    // restore in case of future re-use
    this._maxIterations = maxiter;
    this.network = null;
    this.isValidLayout = true;
  }

  /**
   * @hidden @internal
   * Determines whether a clustering layout should be done before the regular force-directed layout.
   *
   * A clustering layout is useful when the graph is somewhat tree-like.
   * @expose
   * @this {ForceDirectedLayout}
   * @return {boolean}
   */
  public needsClusterLayout(): boolean {
    if (this.network.vertexes.count < 3) return false;
    let count = 0;
    let i = 0;
    const r = this.network.vertexes.first().bounds;
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value;
      const b = vertex.bounds;
      if (b.intersectsRect(r)) {
        count++;
        if (count > 2) return true;
      }
      if (i > 10) return false;  // ?? nothing intersects, so far
      i++;
    }
    return false;
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedNetwork} net
   * @param {Rect} result
   * @return {Rect}
   */
  private computeBounds(net: ForceDirectedNetwork, result: Rect): Rect {
    let first = true;
    const vit = net.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value;
      if (first) {
        first = false;
        result.set(vertex.bounds);
      } else {
        result.unionRect(vertex.bounds);
      }
    }
    return result;
  }

  /**
   * @hidden @internal
   * Determine how many additional force-directed layout iterations should occur for this clustered network.
   * @this {ForceDirectedLayout}
   * @param {number} level int the recursion depth for clustered layout.
   * @param {number} maxiter int
   * @return {number} int a non-negative number that will be temporarily added to {@link #maxIterations}.
   */
  private computeClusterLayoutIterations(level: number, maxiter: number): number {
    if (Debug) {
      Util.checkRealNumber(level, ForceDirectedLayout, 'computeClusterLayoutIterations:level');
      Util.checkRealNumber(maxiter, ForceDirectedLayout, 'computeClusterLayoutIterations:maxiter');
    }
    return Math.max(Math.min(this.network.vertexes.count, maxiter * (level + 1) / 11.0), 10);  // ??
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   */
  private initializeClustering(): void {
    // assume self-links have already been removed
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value as ForceDirectedVertex;
      vertex.numConnections = vertex.vertexes.count;  // cache number of nodes connected to this node
      vertex.numInCluster = 1;
      vertex.clustereds = null;  // clear any internal state
      vertex.savedState = null;
    }
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {number} level int the recursion depth for clustered layout.
   * @param {number} maxiter int
   */
  private layoutClusters(level: number, maxiter: number): void {
    if (Debug) {
      Util.checkRealNumber(level, ForceDirectedLayout, 'layoutClusters:level');
      Util.checkRealNumber(maxiter, ForceDirectedLayout, 'layoutClusters:maxiter');
    }
    if (this.hasClusters(level)) {  // also sorts by # connections
      const oldInf = this._infinityDistance;
      this._infinityDistance *= (1.0 + 1.0 / (level + 1));  // ??

      const oldnet = this.pushSubNetwork(level);
      const subiter = Math.max(0, this.computeClusterLayoutIterations(level, maxiter));
      this.maxIterations += subiter;
      this.layoutClusters(level + 1, maxiter);
      this.performIterations(subiter);
      this.popNetwork(oldnet, level);

      const varr = (oldnet.vertexes as any as Set<ForceDirectedVertex>).toArray();
      varr.sort(function (a, b) {
        if (a === null || b === null || a === b) return 0;
        return b.numConnections - a.numConnections;
      });
      for (let vit = 0; vit < varr.length; vit++) {
        this.surroundNode(varr[vit], level);
      }

      this._infinityDistance = oldInf;
    }
  }

  /**
   * @hidden @internal
   * Determines if the a layer has clusters.
   * A layer has clusters if it has more than one vertex with vertex.numConnections > 1.
   * @this {ForceDirectedLayout}
   * @param {number} level int the recursion depth for clustered layout.
   * @return {boolean}
   */
  private hasClusters(level: number): boolean {
    if (Debug) Util.checkRealNumber(level, ForceDirectedLayout, 'hasClusters:level');
    if (level > 10) return false;
    if (this.network.vertexes.count < 3) return false;  // redundant? since needsClusterLayout checks this
    this._vertexArray = (this.network.vertexes as any as Set<ForceDirectedVertex>).toArray();

    const varr = this._vertexArray;
    varr.sort(function (a, b) {
      if (a === null || b === null || a === b) return 0;
      return b.numConnections - a.numConnections;
    });
    let i = varr.length - 1;
    // the vertexes are now sorted in descending order by number of connections
    while (i >= 0 && varr[i].numConnections <= 1) i--;
    const leaves = varr.length - i;  // leaves represents the number of vertexes for which numConnections > 1.
    return leaves > 1;
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {number} level int the recursion depth for clustered layout.
   * @return {ForceDirectedNetwork}
   */
  private pushSubNetwork(level: number): ForceDirectedNetwork {
    if (Debug) Util.checkRealNumber(level, ForceDirectedLayout, 'pushSubNetwork:level');
    const oldnet = this.network as ForceDirectedNetwork;
    const newnet = new ForceDirectedNetwork(this);
    // reparent multiply-connected nodes and their links
    // use this._vertexArray, the sorted List of the networks vertexes.
    for (let vit = 0; vit < this._vertexArray.length; vit++) {
      const vertex = this._vertexArray[vit];
      if (vertex.numConnections > 1) {
        newnet.addVertex(vertex);  // move to new network
        const info = new ForceDirectedSubnet();  // save state
        info.NumConnections = vertex.numConnections;
        info.Width = vertex.width;
        info.Height = vertex.height;
        info.FocusX = vertex.focus.x;
        info.FocusY = vertex.focus.y;
        if (vertex.savedState === null) vertex.savedState = new List<ForceDirectedSubnet>();
        vertex.savedState.add(info);
        vertex.savedStateIndex = vertex.savedState.count - 1;
      } else {
        break;  // assume sorted (in hasClusters) by decreasing # connections
      }
    }
    let eit = oldnet.edges.iterator;
    while (eit.next()) {
      const edge = eit.value as ForceDirectedEdge;
      const fromVert = edge.fromVertex as ForceDirectedVertex;
      const toVert = edge.toVertex as ForceDirectedVertex;
      if (fromVert.network === newnet && toVert.network === newnet) {
        newnet.addEdge(edge);  // move to new network
      } else if (fromVert.network === newnet) {
        let clus = fromVert.clustereds;
        if (clus === null) {
          clus = new List<ForceDirectedVertex>();
          fromVert.clustereds = clus;
        }
        clus.add(toVert);  // add to other node's cluster
        fromVert.numConnections--;
        fromVert.numInCluster += toVert.numInCluster;
      } else if (toVert.network === newnet) {
        let clus = toVert.clustereds;
        if (clus === null) {
          clus = new List<ForceDirectedVertex>();
          toVert.clustereds = clus;
        }
        clus.add(fromVert);  // add to other node's cluster
        toVert.numConnections--;
        toVert.numInCluster += fromVert.numInCluster;
      }
    }
    // relax the links connecting clustered nodes
    const eit2 = newnet.edges.iterator;
    while (eit2.next()) {
      const edge2 = eit2.value as ForceDirectedEdge;
      const fromVert2 = edge2.fromVertex as ForceDirectedVertex;
      const toVert2 = edge2.toVertex as ForceDirectedVertex;
      const num = fromVert2.numInCluster + toVert2.numInCluster;
      edge2.length *= Math.max(1, Geo.sqrt(num / (4.0 * level + 1)));  // ???
    }
    // pretend clustered nodes are larger
    const vit2 = newnet.vertexes.iterator;
    while (vit2.next()) {
      const vertex2 = vit2.value as ForceDirectedVertex;
      // update state
      const clus = vertex2.clustereds;
      if (clus !== null && clus.count > 0) {
        const info = vertex2.savedState.elt(vertex2.savedState.count - 1);
        const cnt = info.NumConnections - vertex2.numConnections;
        if (cnt <= 0) continue;
        let newarea = 0.0;
        let linklen = 0.0;
        for (let i = clus.count - cnt; i < clus.count; i++) {
          const v = clus.elt(i);
          // just check nodes connected directly to this node
          let connedge: ForceDirectedEdge = null;
          eit = v.edges.iterator;
          while (eit.next()) {
            const e = eit.value as ForceDirectedEdge;
            if (e.getOtherVertex(v) === vertex2) {
              connedge = e;
              break;
            }
          }
          if (connedge !== null) {
            // if (cnt > 4) connedge.length *= Geo.sqrt(cnt/4.0);  //??
            linklen += connedge.length;
            newarea += v.width * v.height;
          }
        }
        // update Bounds and Focus (maintaining Center)
        const ptx = vertex2.centerX;
        const pty = vertex2.centerY;
        const szw = vertex2.width;
        const szh = vertex2.height;
        const foc = vertex2.focus;
        let area = szw * szh;
        if (area < 1) area = 1;
        const comp = Geo.sqrt((newarea + area + ((linklen * linklen * 4) / (cnt * cnt))) / area);  // ??
        const fw = (comp - 1) * szw / 2;
        const fh = (comp - 1) * szh / 2;
        vertex2.bounds = new Rect(ptx - foc.x - fw, pty - foc.y - fh, szw + fw * 2, szh + fh * 2);
        vertex2.focus = new Point(foc.x + fw, foc.y + fh);
      }
    }
    this.network = newnet;
    return oldnet;
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedNetwork} oldnet
   * @param {number} level int the recursion depth for clustered layout.
   */
  private popNetwork(oldnet: ForceDirectedNetwork, level: number): void {
    if (Debug) {
      Util.checkClass(oldnet, ForceDirectedNetwork, ForceDirectedLayout, 'popNetwork:oldnet');
      Util.checkRealNumber(level, ForceDirectedLayout, 'popNetwork:level');
    }
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value as ForceDirectedVertex;
      vertex.network = oldnet;  // reparent node to old network; NOT: oldnet.addNode(node)
      // pop NumConnections & Bounds/Focus
      if (vertex.savedState !== null) {
        const idx = vertex.savedStateIndex;
        const info = vertex.savedState.elt(idx);
        vertex.numConnections = info.NumConnections;
        const w = info.Width;
        const h = info.Height;
        const focx = info.FocusX;
        const focy = info.FocusY;
        const ptx = vertex.centerX;
        const pty = vertex.centerY;
        // restore Size and Focus, adapting to new Center (if it had been moved)
        vertex.bounds = new Rect(ptx - focx, pty - focy, w, h);
        vertex.focus = new Point(focx, focy);
        vertex.savedStateIndex--;  // pop stack without losing StateInfos
      }
    }
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const edge = eit.value;
      edge.network = oldnet;  // reparent link to old network; NOT: oldnet.addLink(link);
      // ?? restore edge.length
    }
    this.network = oldnet;
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} vertex
   * @param {number} level int the recursion depth for clustered layout.
   */
  private surroundNode(vertex: ForceDirectedVertex, level: number): void {
    if (Debug) {
      Util.checkClass(vertex, ForceDirectedVertex, ForceDirectedLayout, 'surroundNode:oldnet');
      Util.checkRealNumber(level, ForceDirectedLayout, 'surroundNode:level');
    }
    const clus = vertex.clustereds;
    if (clus === null || clus.count === 0) return;  // ignore non-clusters
    const centerX = vertex.centerX;
    const centerY = vertex.centerY;
    let width = vertex.width;
    let height = vertex.height;
    if (vertex.savedState !== null && vertex.savedState.count > 0) {
      const info = vertex.savedState.elt(0);
      width = info.Width;
      height = info.Height;
    }
    const radius = Geo.sqrt(width * width + height * height) / 2;
    // where's it coming from? look at connected non-leaf nodes
    let external = false;
    let angle = 0.0;
    let count = 0;
    let externalcount = 0;
    const vit = vertex.vertexes.iterator;
    while (vit.next()) {
      const vt = vit.value as ForceDirectedVertex;
      if (vt.numConnections <= 1) {
        count++;
      } else {
        external = true;
        externalcount++;
        angle += Math.atan2(vertex.centerY - vt.centerY, vertex.centerX - vt.centerX);
      }
    }
    if (count === 0) return;
    // average the angles to multi-connected node
    if (externalcount > 0) {
      angle /= externalcount;
    }
    // over what angle should they be laid out around the node
    let ang = 0.0;
    let offsetang = 0.0;
    if (external) {
      ang = Math.PI * 2 / (count + 1);
    } else {
      ang = Math.PI * 2 / count;
    }
    if (count % 2 === 0) offsetang = ang / 2;
    // now position the leaf nodes around the node
    // reorder according to node size
    if (clus.count > 1) {
      clus.sort(function (a, b) {
        if (a === null || b === null || a === b) return 0;
        return b.width * b.height - a.width * a.height;
      });
    }
    let i = (count % 2 === 0 ? 0 : 1);
    const vit2 = clus.iterator;
    while (vit2.next()) {
      const v = vit2.value;
      if (v.numConnections > 1) continue;  // shouldn't happen?
      if (this.isFixed(v)) continue;
      // position satellite node according to distance given by link.length plus radii
      let edge: ForceDirectedEdge = null;
      const eit = v.edges.iterator;
      while (eit.next()) {
        const e = eit.value as ForceDirectedEdge;
        edge = e;
        break;
      }
      const w = v.width;
      const h = v.height;
      const r = Geo.sqrt(w * w + h * h) / 2;
      const d = radius + edge.length + r;
      const a = angle + (ang * ((i / 2) >> 1) + offsetang) * (i % 2 === 0 ? 1 : -1);
      v.centerX = centerX + d * Math.cos(a);
      v.centerY = centerY + d * Math.sin(a);
      i++;
    }
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {Iterator.<ForceDirectedNetwork>} subnets
   * @param {ForceDirectedNetwork} singletons
   */
  private arrangeConnectedGraphs(subnets: Iterator<ForceDirectedNetwork>, singletons: ForceDirectedNetwork): void {
    if (Debug) {
      Util.checkClass(singletons, ForceDirectedNetwork, ForceDirectedLayout, 'arrangeConnectedGraphs:singletons');
    }
    const spacing = this.arrangementSpacing;
    // how many subnetworks are there?
    const subnetcount = subnets.count;

    let first = true;
    let centerX = 0.0;  // first subnetwork's Center
    let centerY = 0.0;
    const fringe = Util.tempArray<Point>();
    for (let i = 0; i < subnetcount + singletons.vertexes.count + 2; i++) fringe[i] = null;
    let fringelen = 0.0;  // how many FRINGE points are valid
    subnets.reset();
    const bounds = Rect.alloc();
    let vit;
    while (subnets.next()) {
      const subnet = subnets.value;
      this.computeBounds(subnet, bounds);
      if (first) {
        first = false;
        centerX = bounds.x + bounds.width / 2;
        centerY = bounds.y + bounds.height / 2;
        fringe[0] = new Point(bounds.x + bounds.width + spacing.width, bounds.y);
        fringe[1] = new Point(bounds.x, bounds.y + bounds.height + spacing.height);
        fringelen = 2;
      } else {
        // place subnetwork somewhere closest to first subnetwork's center
        const idx = this.closestFringePoint(fringe, fringelen, centerX, centerY, bounds.width, bounds.height, spacing);
        const old = fringe[idx];
        const tr = new Point(old.x + bounds.width + spacing.width, old.y);
        const bl = new Point(old.x, old.y + bounds.height + spacing.height);
        // make room for additional fringe point
        if (idx + 1 < fringelen) fringe.splice(idx + 1, 0, null);
        fringe[idx] = tr;
        fringe[idx + 1] = bl;
        fringelen++;
        // shift subnetwork to fit at OLD point
        const offx = old.x - bounds.x;
        const offy = old.y - bounds.y;
        vit = subnet.vertexes.iterator;
        while (vit.next()) {
          const vt = vit.value;
          vt.centerX += offx;
          vt.centerY += offy;
        }
      }
    }
    Rect.free(bounds);

    // place singletons
    vit = singletons.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value;
      const vb = vertex.bounds;
      if (fringelen < 2) {
        centerX = vb.x + vb.width / 2;
        centerY = vb.y + vb.height / 2;
        fringe[0] = new Point(vb.x + vb.width + spacing.width, vb.y);
        fringe[1] = new Point(vb.x, vb.y + vb.height + spacing.height);
        fringelen = 2;
        continue;
      }
      const idx = this.closestFringePoint(fringe, fringelen, centerX, centerY, vb.width, vb.height, spacing);
      const old = fringe[idx];
      const tr = new Point(old.x + vb.width + spacing.width, old.y);
      const bl = new Point(old.x, old.y + vb.height + spacing.height);
      // make room for additional fringe point
      if (idx + 1 < fringelen) fringe.splice(idx + 1, 0, null);
      fringe[idx] = tr;
      fringe[idx + 1] = bl;
      fringelen++;
      vertex.centerX = old.x + vertex.width / 2;
      vertex.centerY = old.y + vertex.height / 2;
    }
    Util.freeArray(fringe);
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {Array.<Point>} fringe of Point
   * @param {number} fringelen
   * @param {number} centerx
   * @param {number} centery
   * @param {number} sizew
   * @param {number} sizeh
   * @param {Size} spacing
   * @return {number}
   */
  private closestFringePoint(fringe: Array<Point>, fringelen: number, centerx: number, centery: number, sizew: number, sizeh: number, spacing: Size): number {
    let mindist = 9e19;
    let idx = -1;
    ANOTHER:
    for (let i = 0; i < fringelen; i++) {
      const p = fringe[i];
      const x = p.x - centerx;
      const y = p.y - centery;
      const sq = x * x + y * y;
      if (sq < mindist) {
        for (let j = i - 1; j >= 0; j--) {
          if (fringe[j].y > p.y && fringe[j].x - p.x < sizew + spacing.width) continue ANOTHER;
        }
        for (let j = i + 1; j < fringelen; j++) {
          if (fringe[j].x > p.x && fringe[j].y - p.y < sizeh + spacing.height) continue ANOTHER;
        }
        idx = i;
        mindist = sq;
      }
    }
    return idx;
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   */
  private annotateAll(): void {
    if (!this.comments) return;
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value as ForceDirectedVertex;
      this.addComments(v);
    }
  }

  /**
   * Find any associated objects to be positioned along with the {@link LayoutVertex#node}.
   *
   * This method is called for each vertex in the network, when {@link #comments} is true.
   * The standard behavior is to look for {@link Node}s whose {@link Part#category}
   * is "Comment" and that refer to the {@link LayoutVertex#node}.
   * By default this method will not be called unless you set {@link #comments} to true.
   *
   * You may want to override this method in order to customize how any
   * associated objects are found and how a new {@link ForceDirectedVertex}
   * and {@link ForceDirectedEdge}
   * may be added to the network to represent the (balloon?) comment.
   * This method sets the new vertex's {@link ForceDirectedVertex#charge}
   * to the value of {@link #defaultCommentElectricalCharge},
   * and sets the new edge's {@link ForceDirectedEdge#length}
   * to the value of {@link #defaultCommentSpringLength}.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} v
   * @since 1.3
   */
  public addComments(v: ForceDirectedVertex): void {
    const node = v.node;
    if (node !== null) {
      const nit = node.findNodesConnected();
      while (nit.next()) {
        const comment = nit.value;
        if (comment.category !== 'Comment') continue;
        if (!comment.isVisible()) continue;

        // is there already a Vertex for the comment Node?
        let cv = this.network.findVertex(comment) as ForceDirectedVertex;
        if (cv === null) {  // if not, add it
          cv = this.network.addNode(comment) as ForceDirectedVertex;
        }
        cv.charge = this.defaultCommentElectricalCharge;

        // is there already an Edge connecting the comment Vertex with V?
        let edge: ForceDirectedEdge = null;
        const deit = cv.destinationEdges;
        while (deit.next()) {
          const te = deit.value as ForceDirectedEdge;
          if (te.toVertex === v) {
            edge = te;
            break;
          }
        }
        if (edge === null) {
          const seit = cv.sourceEdges;
          while (seit.next()) {
            const fe = seit.value as ForceDirectedEdge;
            if (fe.fromVertex === v) {
              edge = fe;
              break;
            }
          }
        }
        if (edge === null) {  // if not, add a dummy edge
          edge = this.network.linkVertexes(v, cv, null) as ForceDirectedEdge;
        }
        edge.length = this.defaultCommentSpringLength;
      }
    }
  }

  /**
   * @hidden @internal
   * Returns the distance between two vertexes.
   * The default implementation considers the shortest distance between the two nodes.
   * If the vertexes correspond to top-level parts, the width and height of the
   * part is factored into the distance between the vertexes.
   * If the vertexes do not correspond to parts, then the vertexes implicitly
   * have a width and height of zero, and the distance is calculated.
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} vertexA
   * @param {ForceDirectedVertex} vertexB
   * @return {number} the approximate distance between the two vertexes.
   */
  private getNodeDistance(vertexA: ForceDirectedVertex, vertexB: ForceDirectedVertex): number {
    if (Debug) {
      Util.checkClass(vertexA, ForceDirectedVertex, ForceDirectedLayout, 'getNodeDistance:vertexA');
      Util.checkClass(vertexB, ForceDirectedVertex, ForceDirectedLayout, 'getNodeDistance:vertexB');
    }
    const rA = vertexA.bounds;
    const rectAx = rA.x;
    const rectAy = rA.y;
    const rectAw = rA.width;
    const rectAh = rA.height;

    const rB = vertexB.bounds;
    const rectBx = rB.x;
    const rectBy = rB.y;
    const rectBw = rB.width;
    const rectBh = rB.height;

    if ((rectAx + rectAw) < rectBx) {
      if (rectAy > (rectBy + rectBh)) {
        const dx = rectAx + rectAw - rectBx;
        const dy = rectAy - rectBy - rectBh;
        return Geo.sqrt(dx * dx + dy * dy);
      } else if ((rectAy + rectAh) < rectBy) {
        const dx = rectAx + rectAw - rectBx;
        const dy = rectAy + rectAh - rectBy;
        return Geo.sqrt(dx * dx + dy * dy);
      } else {
        return rectBx - (rectAx + rectAw);
      }
    } else if (rectAx > (rectBx + rectBw)) {
      if (rectAy > (rectBy + rectBh)) {
        const dx = rectAx - rectBx - rectBw;
        const dy = rectAy - rectBy - rectBh;
        return Geo.sqrt(dx * dx + dy * dy);
      } else if ((rectAy + rectAh) < rectBy) {
        const dx = rectAx - rectBx - rectBw;
        const dy = rectAy + rectAh - rectBy;
        return Geo.sqrt(dx * dx + dy * dy);
      } else {
        return rectAx - (rectBx + rectBw);
      }
    } else {
      if (rectAy > (rectBy + rectBh)) {
        return rectAy - (rectBy + rectBh);
      } else if ((rectAy + rectAh) < rectBy) {
        return rectBy - (rectAy + rectAh);
      } else {
        // The two rectangles intersect.
        // Technically, could return 0.0, but that causes division
        //  by zero errors.  Instead, we return 0.1, which is smaller
        //  than any real distance between integer-coordinate points.
        //  Since the centers of the two rectangles are probably not
        //  the same, the forces between the nodes will point in some
        //  direction separating the nodes.
        return 0.1;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   * @param {number} num int
   */
  private performIterations(num: number): void {
    if (Debug) Util.checkRealNumber(num, ForceDirectedLayout, 'performIterations:num');
    this._vertexArray = null;
    const end = this._iteration + num;
    while (this._iteration < end) {
      this._iteration++;
      // If no change occurred, solution found.
      if (!this.updatePositions()) break;
    }
    this._vertexArray = null;
  }

  /**
   * @hidden @internal
   * Performs one iteration of the ForceDirected algorithm and updates the positions
   * of the vertexes, but not the diagram's nodes.
   * @this {ForceDirectedLayout}
   * @return {boolean} returns true if additional iterations are needed to find a solution.
   */
  private updatePositions(): boolean {
    if (this._vertexArray === null) {
      this._vertexArray = (this.network.vertexes as any as Set<ForceDirectedVertex>).toArray();
    }

    // Every node initially has no forces acting upon it.
    const vertexarray = this._vertexArray;
    if (vertexarray.length <= 0) return false;

    const firstnode = vertexarray[0];
    firstnode.forceX = 0;
    firstnode.forceY = 0;
    let minx = firstnode.centerX;
    let maxx = minx;
    let miny = firstnode.centerY;
    let maxy = miny;
    for (let ni = 1; ni < vertexarray.length; ni++) {
      const vertex = vertexarray[ni];
      vertex.forceX = 0;
      vertex.forceY = 0;
      const px = vertex.centerX;
      const py = vertex.centerY;
      minx = Math.min(minx, px);
      maxx = Math.max(maxx, px);
      miny = Math.min(miny, py);
      maxy = Math.max(maxy, py);
    }
    //  If there is a larger horizontal seperation than a vertical one,
    // sort the vertexes by horizontal position, else sort them by their
    // vertical positions.
    const sortedx = (maxx - minx > maxy - miny);
    if (sortedx) {
      vertexarray.sort(function (a, b) {
        if (a === null || b === null || a === b) return 0;
        return a.centerX - b.centerX;
      });
    } else {
      vertexarray.sort(function (a, b) {
        if (a === null || b === null || a === b) return 0;
        return a.centerY - b.centerY;
      });
    }

    const infinity = this._infinityDistance;
    let force = 0.0;
    let forceX = 0.0;
    let forceY = 0.0;
    // Calculate gravitational and electrical forces.
    for (let ni = 0; ni < vertexarray.length; ni++) {
      const vertexU = vertexarray[ni];
      const vUb = vertexU.bounds;
      const vUf = vertexU.focus;
      const NodeCenterUx = vUb.x + vUf.x; // vertexU.centerX;
      const NodeCenterUy = vUb.y + vUf.y; // vertexU.centerY;

      // Calculate the electrical forces in the X and Y directions.
      forceX = vertexU.charge * this.electricalFieldX(NodeCenterUx, NodeCenterUy);
      forceY = vertexU.charge * this.electricalFieldY(NodeCenterUx, NodeCenterUy);

      // Calculate the gravitational forces in the X and Y directions.
      forceX += vertexU.mass * this.gravitationalFieldX(NodeCenterUx, NodeCenterUy);
      forceY += vertexU.mass * this.gravitationalFieldY(NodeCenterUx, NodeCenterUy);

      // Update the total forces acting upon the node.
      vertexU.forceX += forceX;
      vertexU.forceY += forceY;

      // Scan through remaining nodes in the network.
      for (let posV = ni + 1; posV < vertexarray.length; posV++) {
        const vertexV = vertexarray[posV];
        if (vertexV === vertexU) continue;
        const vVb = vertexV.bounds;
        const vVf = vertexV.focus;
        const NodeCenterVx = vVb.x + vVf.x; // vertexV.centerX;
        const NodeCenterVy = vVb.y + vVf.y; // vertexV.centerY;

        if (NodeCenterUx - NodeCenterVx > infinity || NodeCenterVx - NodeCenterUx > infinity) { // if (Math.abs(NodeCenterUx - NodeCenterVx) > infinity) {
          if (sortedx) break;
          continue;
        }
        if (NodeCenterUy - NodeCenterVy > infinity || NodeCenterVy - NodeCenterUy > infinity) { // if (Math.abs(NodeCenterUy - NodeCenterVy) > infinity) {
          if (!sortedx) break;
          continue;
        }

        const dist = this.getNodeDistance(vertexU, vertexV);

        if (dist < 1) {
          let rand = this.randomNumberGenerator;
          if (rand === null) this.randomNumberGenerator = rand = new RandomNumberGenerator(0);
          const randnumx = rand['random']();
          const randnumy = rand['random']();
          // overlapping nodes
          if (NodeCenterUx > NodeCenterVx) {
            forceX = Math.abs(vertexV.bounds.right - vertexU.bounds.x);
            forceX = (1 + forceX) * randnumx;
          } else if (NodeCenterUx < NodeCenterVx) {
            forceX = Math.abs(vertexV.bounds.x - vertexU.bounds.right);
            forceX = -(1 + forceX) * randnumx;
          } else {
            const maxw = Math.max(vertexV.width, vertexU.width);
            forceX = (1 + maxw) * randnumx - maxw / 2;
          }
          if (NodeCenterUy > NodeCenterVy) {
            forceY = Math.abs(vertexV.bounds.bottom - vertexU.bounds.y);
            forceY = (1 + forceY) * randnumy;
          } else if (NodeCenterUx < NodeCenterVx) {
            forceY = Math.abs(vertexV.bounds.y - vertexU.bounds.bottom);
            forceY = -(1 + forceY) * randnumy;
          } else {
            const maxh = Math.max(vertexV.height, vertexU.height);
            forceY = (1 + maxh) * randnumy - maxh / 2;
          }
        } else {
          // Calculate the electrical force acting on the nodes.
          // There is no equivalent for gravitational force between vertexes because
          // among charged particles, the gravitational force is insignificant
          // when compared to the electrical force.
          force = -(vertexU.charge * vertexV.charge) / (dist * dist);

          // Calculate the electrical forces acting on the nodes in the X and Y directions.
          forceX = force * ((NodeCenterVx - NodeCenterUx) / dist);
          forceY = force * ((NodeCenterVy - NodeCenterUy) / dist);
        }

        // Update the total forces acting upon the nodes.
        vertexU.forceX += forceX;
        vertexU.forceY += forceY;
        vertexV.forceX -= forceX;
        vertexV.forceY -= forceY;
      }
    }

    // Calculate spring forces.
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const edge = eit.value as ForceDirectedEdge;
      const vertexU = edge.fromVertex as ForceDirectedVertex;
      const vertexV = edge.toVertex as ForceDirectedVertex;
      const vUb = vertexU.bounds;
      const vUf = vertexU.focus;
      const NodeCenterUx = vUb.x + vUf.x; // vertexU.centerX;
      const NodeCenterUy = vUb.y + vUf.y; // vertexU.centerY;
      const vVb = vertexV.bounds;
      const vVf = vertexV.focus;
      const NodeCenterVx = vVb.x + vVf.x; // vertexV.centerX;
      const NodeCenterVy = vVb.y + vVf.y; // vertexV.centerY;

      const dist = this.getNodeDistance(vertexU, vertexV);

      if (dist < 1) {
        let rand = this.randomNumberGenerator;
        if (rand === null) this.randomNumberGenerator = rand = new RandomNumberGenerator(0);
        const randnumx = rand['random']();
        const randnumy = rand['random']();
        forceX = (NodeCenterUx > NodeCenterVx ? 1 : -1) * (1 + ((vertexV.width > vertexU.width) ? vertexV.width : vertexU.width)) * randnumx;
        forceY = (NodeCenterUy > NodeCenterVy ? 1 : -1) * (1 + ((vertexV.height > vertexU.height) ? vertexV.height : vertexU.height)) * randnumy;
      } else {
        // Calculate the spring force acting on the nodes.
        force = edge.stiffness * (dist - edge.length);

        // Calculate the spring forces acting on the nodes in the X and Y directions.
        forceX = force * ((NodeCenterVx - NodeCenterUx) / dist);
        forceY = force * ((NodeCenterVy - NodeCenterUy) / dist);
      }

      // Update the total forces acting upon the nodes.
      vertexU.forceX += forceX;
      vertexU.forceY += forceY;
      vertexV.forceX -= forceX;
      vertexV.forceY -= forceY;
    }

    let maxMove = 0.0;
    const moveLimit = this.moveLimit;

    // Update the positions of all nodes in the network.
    for (let i = 0; i < vertexarray.length; i++) {
      const vertex = vertexarray[i];
      // Only move the node if it is not fixed.
      if (this.isFixed(vertex)) {
        this.moveFixedVertex(vertex);
      } else {
        // Calculate the distance to move the node based on the forces acting upon it.
        let dX = vertex.forceX;
        let dY = vertex.forceY;

        // Do not allow a node to move more than some number of units.
        if (dX < -moveLimit) {
          dX = -moveLimit;
        } else if (dX > moveLimit) {
          dX = moveLimit;
        }

        if (dY < -moveLimit) {
          dY = -moveLimit;
        } else if (dY > moveLimit) {
          dY = moveLimit;
        }

        // Set the center of the node.
        vertex.centerX += dX;
        vertex.centerY += dY;

        // Check to see if any node has moved significantly.
        maxMove = Math.max(maxMove, dX * dX + dY * dY);
      }
    }

    return (maxMove > this.epsilonDistance * this.epsilonDistance);
  }

  /**
   * Maybe move a vertex that {@link #isFixed}.
   * This is called each iteration on each such vertex.
   * By default this does nothing.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} v
   */
  public moveFixedVertex(v: ForceDirectedVertex): void {
  }

  /**
   * Set the fromSpot and toSpot on each Link, position each Node according
   * to the vertex position, and then position/route the Links.
   *
   * This calls the {@link #commitNodes} and {@link #commitLinks} methods, the latter only if {@link #isRouting} is true.
   * You should not call this method -- it is a "protected virtual" method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @this {ForceDirectedLayout}
   */
  protected commitLayout(): void {
    this.setPortSpotsAll();
    this.commitNodes();
    if (this.isRouting) this.commitLinks();
  }

  /**
   * @hidden @internal
   * @this {ForceDirectedLayout}
   */
  private setPortSpotsAll(): void {
    if (!this.setsPortSpots) return;
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const edge = eit.value;
      const link = edge.link;
      if (link !== null) {
        link.fromSpot = Spot.Default;
        link.toSpot = Spot.Default;
      }
    }
  }

  /**
   * Commit the position of all nodes.
   *
   * This is called by {@link #commitLayout}.
   * See also {@link #commitLinks}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   */
  protected commitNodes(): void {
    let offsetx = 0.0;
    let offsety = 0.0;
    if (this.arrangesToOrigin) {
      const b = Rect.alloc();
      this.computeBounds(this.network as ForceDirectedNetwork, b);
      const orig = this.arrangementOrigin;
      offsetx = orig.x - b.x;
      offsety = orig.y - b.y;
      Rect.free(b);
    }
    const r = Rect.alloc();
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const vertex = vit.value;
      if (offsetx !== 0 || offsety !== 0) {
        r.assign(vertex.bounds);
        r.x += offsetx;
        r.y += offsety;
        vertex.bounds = r;
      }
      vertex.commit();
    }
    Rect.free(r);
  }

  /**
   * Routes the links.
   *
   * This is called by {@link #commitLayout}.
   * This is only called if {@link Layout#isRouting} is true.
   * See also {@link #commitNodes}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   */
  protected commitLinks(): void {
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const edge = eit.value;
      edge.commit();
    }
  }


  /**
   * Returns the stiffness of the spring representing an edge.
   *
   * The spring force between two vertexes connected by an edge is linearly proportional by distance
   * to the difference between the {@link #springLength} and the distance.
   * When the distance is greater than the length, the force pulls the vertexes closer to each other.
   * When the distance is less than the length, the force pushes them apart.
   *
   * The two vertexes connected by the edge E are acted upon by a force of proportional to
   * `springStiffness(E) * (getNodeDistance(E.fromVertex, E.toVertex) - springLength(E))` divided by the distance between the vertexes.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedEdge} e
   * @return {number}
   * Returns the stiffness of the edge representing a link,
   * the value of {@link ForceDirectedEdge#stiffness} if it's a number,
   * or else the value of {@link #defaultSpringStiffness}.
   */
  public springStiffness(e: ForceDirectedEdge): number {
    const s = e.stiffness;
    if (!isNaN(s)) return s;
    return this._defaultSpringStiffness;
  }

  /**
   * Returns the length of the spring representing an edge.
   * The two vertexes connected by the edge E are acted upon by a force of proportional to
   * `springStiffness(E) * (getNodeDistance(E.fromVertex, E.toVertex) - springLength(E))` divided by the distance between the vertexes.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedEdge} e
   * @return {number}
   * Returns the length of the edge representing a link,
   * the value of {@link ForceDirectedEdge#length} if it's a number,
   * or else the value of {@link #defaultSpringLength}.
   */
  public springLength(e: ForceDirectedEdge): number {
    const s = e.length;
    if (!isNaN(s)) return s;
    return this._defaultSpringLength;
  }


  /**
   * Returns the charge of the vertex,
   * the value of {@link ForceDirectedVertex#charge} if it's a number,
   * or else the value of {@link #defaultElectricalCharge}.
   *
   * The electrical forces between two vertexes decrease by the square of the distance between them.
   * Vertexes that are more than {@link #infinityDistance} apart are assumed to have no electrical charge effect on each other.
   *
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} v
   * @return {number}
   */
  public electricalCharge(v: ForceDirectedVertex): number {
    const c = v.charge;
    if (!isNaN(c)) return c;
    return this._defaultElectricalCharge;
  }

  /**
   * Returns the electrical field in the X direction acting on a vertex at the given point.
   * By default there is no electrical field at any location.
   *
   * Used to define an external electrical field at a point independent of the vertex charges.
   * A vertex L is acted upon by a force in the X direction of proportional to
   * `electricalFieldX(L.center.x, L.center.y) * electricalCharge(L)`.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {number} x
   * @param {number} y
   * @return {number} the default implementation returns zero.
   */
  public electricalFieldX(x: number, y: number): number {
    return 0;
  }

  /**
   * Returns the electrical field in the Y direction acting on a vertex at the given point.
   * By default there is no electrical field at any location.
   *
   * Used to define an external electrical field at a point independent of the vertex charges.
   * A vertex L is acted upon by a force in the Y direction of proportional to
   * `electricalFieldY(L.center.x, L.center.y) * electricalCharge(L)`.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {number} x
   * @param {number} y
   * @return {number} the default implementation returns zero.
   */
  public electricalFieldY(x: number, y: number): number {
    return 0;
  }


  /**
   * Returns the mass of the vertex,
   * the value of {@link ForceDirectedVertex#mass} if it's a number,
   * or else the value of {@link #defaultGravitationalMass}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} v
   * @return {number}
   */
  public gravitationalMass(v: ForceDirectedVertex): number {
    const m = v.mass;
    if (!isNaN(m)) return m;
    return this._defaultGravitationalMass;
  }

  /**
   * This returns the gravitational field in the X direction acting on a vertex at the given point.
   * By default there is no gravitational field at any location.
   *
   * Used to define an external gravitational field at a point independent of the vertex masses.
   * A vertex L is acted upon by a force in the X direction of proportional to
   * `gravitationalFieldX(L.center.x, L.center.y) * gravitationalMass(L)`.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {number} x
   * @param {number} y
   * @return {number} the default implementation returns zero.
   */
  public gravitationalFieldX(x: number, y: number): number {
    return 0;
  }

  /**
   * This returns the gravitational field in the Y direction acting on a vertex at the given point.
   * By default there is no gravitational field at any location.
   *
   * Used to define an external gravitational field at a point independent of the vertex masses.
   * A vertex L is acted upon by a force in the Y direction of proportional to
   * `gravitationalFieldY(L.center.x, L.center.y) * gravitationalMass(L)`.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {number} x
   * @param {number} y
   * @return {number} the default implementation returns zero.
   */
  public gravitationalFieldY(x: number, y: number): number {
    return 0;
  }


  /**
   * This predicate returns true if the vertex should not be moved
   * by the layout algorithm but still have an effect on nearby and connected vertexes.
   * The default implementation returns {@link ForceDirectedVertex#isFixed}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {ForceDirectedLayout}
   * @param {ForceDirectedVertex} v
   * @return {boolean} returns true if the node should not be moved by the layout algorithm.
   */
  public isFixed(v: ForceDirectedVertex): boolean {
    return v.isFixed;
  }


  /**
   * This read-only property returns the current iteration count, valid during a call to {@link #doLayout}.
   */
  /** @export */ get currentIteration(): number { return this._iteration; }

  /**
   * Gets or sets the space between which the layout
   * will position the connected graphs that together compose the network.
   * This defaults to Size(100, 100).
   * These distances are used during a clustered layout;
   * afterwards the normal force-directed layout
   * will likely cause the size of any space between connected graphs to change,
   * perhaps considerably.
   */
  /** @export */ get arrangementSpacing(): Size { return this._arrangementSpacing; }
  /** @export */ set arrangementSpacing(value: Size) {
    Util.checkClass(value, Size, ForceDirectedLayout, 'arrangementSpacing');
    if (!this._arrangementSpacing.equals(value)) {
      this._arrangementSpacing.assign(value);
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether {@link #commitNodes} should move all of the
   * nodes so that the nodes all fit with the top-left corner at the
   * {@link Layout#arrangementOrigin}.
   * By default this is false -- the {@link Layout#arrangementOrigin} is ignored.
   * When this is true, nodes are moved even though {@link #isFixed} was true.
   */
  /** @export */ get arrangesToOrigin(): boolean { return this._arrangesToOrigin; }
  /** @export */ set arrangesToOrigin(value: boolean) {
    if (this._arrangesToOrigin !== value) {
      Util.checkPrimitive(value, 'boolean', ForceDirectedLayout, 'arrangesToOrigin');
      this._arrangesToOrigin = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the fromSpot and the toSpot of every {@link Link}
   * should be set to {@link Spot.Default}.
   * The default value is true.
   */
  /** @export */ get setsPortSpots(): boolean { return this._setsPortSpots; }
  /** @export */ set setsPortSpots(value: boolean) {
    if (this._setsPortSpots !== value) {
      Util.checkPrimitive(value, 'boolean', ForceDirectedLayout, 'setsPortSpots');
      this._setsPortSpots = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether this layout should find all {@link Node}s
   * whose category is "Comment" and
   * whose anchors are nodes represented in the network,
   * and add {@link ForceDirectedVertex}es representing those balloon comments
   * as nodes in the network.
   * The default value is false.
   */
  /** @export */ get comments(): boolean { return this._comments; }
  /** @export */ set comments(value: boolean) {
    if (this._comments !== value) {
      Util.checkPrimitive(value, 'boolean', ForceDirectedLayout, 'comments');
      this._comments = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the maximum number of iterations to perform when doing the
   * force-directed auto layout.
   * The value must be non-negative.
   * The default value is 100.
   */
  /** @export */ get maxIterations(): number { return this._maxIterations; }
  /** @export */ set maxIterations(value: number) {
    if (this._maxIterations !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'maxIterations');
      if (value >= 0) {
        this._maxIterations = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets approximately how far a node must move in order for the iterations to continue.
   * The default value is 1.
   * The value must be larger than zero.
   */
  /** @export */ get epsilonDistance(): number { return this._epsilonDistance; }
  /** @export */ set epsilonDistance(value: number) {
    if (this._epsilonDistance !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'epsilonDistance');
      if (value > 0) {
        this._epsilonDistance = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets a threshold for the distance beyond which the electrical charge forces may be ignored.
   * The default value is 1000.
   * The value must be larger than 1.
   */
  /** @export */ get infinityDistance(): number { return this._infinityDistance; }
  /** @export */ set infinityDistance(value: number) {
    if (this._infinityDistance !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'infinityDistance');
      if (value > 1) {
        this._infinityDistance = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets how far a vertex may be moved in an iteration.
   * The default value is 10.
   * The value must be larger than 1.
   * @since 1.8
   */
  /** @export */ get moveLimit(): number { return this._moveLimit; }
  /** @export */ set moveLimit(value: number) {
    if (this._moveLimit !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'moveLimit');
      if (value > 1) {
        this._moveLimit = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets a random number generator.
   * The default value is Math, which results in calling Math.random().
   * Change this to null in order to use an instance of an internal repeatable pseudo-random number generator,
   * which will become the new value of this property.
   *
   * The new value must be either null or an Object with a method named "random" taking zero arguments
   * and returning a random number between zero (inclusive) and one (exclusive).
   * @since 1.5
   */
  /** @export */ get randomNumberGenerator(): ({ random: () => number } | null) { return this._randomNumberGenerator; }
  /** @export */ set randomNumberGenerator(value: ({ random: () => number } | null)) {
    if (this._randomNumberGenerator !== value) {
      if (value !== null && typeof (value['random']) !== 'function') Util.throwError('ForceDirectedLayout.randomNumberGenerator must have a \"random()\" function on it: ' + value);
      this._randomNumberGenerator = value;
    }
  }

  /**
   * Gets or sets the default value computed by {@link #springStiffness}.
   * The initial value is 0.05.
   */
  /** @export */ get defaultSpringStiffness(): number { return this._defaultSpringStiffness; }
  /** @export */ set defaultSpringStiffness(value: number) {
    if (this._defaultSpringStiffness !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultSpringStiffness');
      this._defaultSpringStiffness = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default value computed by {@link #springLength}.
   * The initial value is 50.
   */
  /** @export */ get defaultSpringLength(): number { return this._defaultSpringLength; }
  /** @export */ set defaultSpringLength(value: number) {
    if (this._defaultSpringLength !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultSpringLength');
      this._defaultSpringLength = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default value computed by {@link #electricalCharge}.
   * The initial value is 150.
   */
  /** @export */ get defaultElectricalCharge(): number { return this._defaultElectricalCharge; }
  /** @export */ set defaultElectricalCharge(value: number) {
    if (this._defaultElectricalCharge !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultElectricalCharge');
      this._defaultElectricalCharge = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default value computed by {@link #gravitationalMass}.
   * The initial value is zero.
   */
  /** @export */ get defaultGravitationalMass(): number { return this._defaultGravitationalMass; }
  /** @export */ set defaultGravitationalMass(value: number) {
    if (this._defaultGravitationalMass !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultGravitationalMass');
      this._defaultGravitationalMass = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default value computed by {@link #springLength}.
   * The initial value is 10.
   */
  /** @export */ get defaultCommentSpringLength(): number { return this._defaultCommentSpringLength; }
  /** @export */ set defaultCommentSpringLength(value: number) {
    if (this._defaultCommentSpringLength !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultCommentSpringLength');
      this._defaultCommentSpringLength = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default value computed by {@link #electricalCharge}.
   * The initial value is 5.
   */
  /** @export */ get defaultCommentElectricalCharge(): number { return this._defaultCommentElectricalCharge; }
  /** @export */ set defaultCommentElectricalCharge(value: number) {
    if (this._defaultCommentElectricalCharge !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedLayout, 'defaultCommentElectricalCharge');
      this._defaultCommentElectricalCharge = value;
      this.invalidateLayout();
    }
  }
} // #end ForceDirectedLayout
ForceDirectedLayout['className'] = 'ForceDirectedLayout';

/**
 * @hidden @internal
 * @unrestricted
 */
export class ForceDirectedSubnet {
  /** @hidden @internal */ public static className = 'ForceDirectedSubnet';
  /** @hidden @internal */ public NumConnections: number;
  /** @hidden @internal */ public Width: number;
  /** @hidden @internal */ public Height: number;
  /** @hidden @internal */ public FocusX: number;
  /** @hidden @internal */ public FocusY: number;

  constructor() {
    this.NumConnections = 0;
    this.Width = 0;
    this.Height = 0;
    this.FocusX = 0;
    this.FocusY = 0;
  }
} // #end ForceDirectedSubnet
ForceDirectedSubnet['className'] = 'ForceDirectedSubnet';

/**
 * This class represents an abstract graph of {@link ForceDirectedVertex}es and {@link ForceDirectedEdge}s
 * that can be constructed based on the {@link Node}s and {@link Link}s of a {@link Diagram}
 * so that the {@link ForceDirectedLayout} can operate independently of the diagram until it
 * is time to commit any node positioning or link routing.
 *
 * This class inherits from {@link LayoutNetwork}.
 * @extends LayoutNetwork
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class ForceDirectedNetwork extends LayoutNetwork {
  /** @hidden @internal */ public static className = 'ForceDirectedNetwork';
  constructor(layout: ForceDirectedLayout) {
    super(layout);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link ForceDirectedVertex}.
   * @this {ForceDirectedNetwork}
   * @return {ForceDirectedVertex}
   */
  public createVertex(): ForceDirectedVertex {
    return new ForceDirectedVertex(this);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link ForceDirectedEdge}.
   * @this {ForceDirectedNetwork}
   * @return {ForceDirectedEdge}
   */
  public createEdge(): ForceDirectedEdge {
    return new ForceDirectedEdge(this);
  }

} // #end ForceDirectedNetwork
ForceDirectedNetwork['className'] = 'ForceDirectedNetwork';

/**
 * This holds {@link ForceDirectedLayout}-specific information about {@link Node}s.
 *
 * This class inherits from {@link LayoutVertex}.
 * @extends LayoutVertex
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class ForceDirectedVertex extends LayoutVertex {
  /** @hidden @internal */ public static className = 'ForceDirectedVertex';
  /** @hidden @internal */ private _isFixed: boolean;
  /** @hidden @internal */ private _charge: number;
  /** @hidden @internal */ private _mass: number;
  /** @hidden @internal */ private _forceX: number;
  /** @hidden @internal */ private _forceY: number;
  /** @hidden @internal */ public numConnections: number;
  /** @hidden @internal */ public numInCluster: number;
  /** @hidden @internal */ public clustereds: List<ForceDirectedVertex> | null;
  /** @hidden @internal */ public savedState: List<ForceDirectedSubnet> | null;
  /** @hidden @internal */ public savedStateIndex: number;

  constructor(network: ForceDirectedNetwork) {
    super(network);

    // force-specific properties:
    this._isFixed = false;

    this._charge = NaN;

    this._mass = NaN;

    this._forceX = 0;

    this._forceY = 0;

    // internal state
    this.numConnections = 0;
    this.numInCluster = 0;
    this.clustereds = null;
    this.savedState = null;
    this.savedStateIndex = 0;
  }

  /**
   * Gets or sets whether the vertex may be moved by any forces.
   * The default value is false.
   */
  /** @export */ get isFixed(): boolean { return this._isFixed; }
  /** @export */ set isFixed(value: boolean) {
    if (this._isFixed !== value) {
      Util.checkPrimitive(value, 'boolean', ForceDirectedVertex, 'isFixed');
      this._isFixed = value;
    }
  }

  /**
   * Gets or sets the electrical charge for this vertex.
   * The default value is NaN.
   */
  /** @export */ get charge(): number { return this._charge; }
  /** @export */ set charge(value: number) {
    if (this._charge !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedVertex, 'charge');
      this._charge = value;
    }
  }

  /**
   * Gets or sets the gravitational mass for this vertex.
   * The default value is NaN.
   */
  /** @export */ get mass(): number { return this._mass; }
  /** @export */ set mass(value: number) {
    if (this._mass !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedVertex, 'mass');
      this._mass = value;
    }
  }

  /**
   * Gets or sets the cumulative force on the vertex in the X direction.
   * The default value is zero.
   */
  /** @export */ get forceX(): number { return this._forceX; }
  /** @export */ set forceX(value: number) {
    if (this._forceX !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedVertex, 'forceX');
      this._forceX = value;
    }
  }

  /**
   * Gets or sets the cumulative force on the vertex in the Y direction.
   * The default value is zero.
   */
  /** @export */ get forceY(): number { return this._forceY; }
  /** @export */ set forceY(value: number) {
    if (this._forceY !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedVertex, 'forceY');
      this._forceY = value;
    }
  }
} // #end ForceDirectedVertex
ForceDirectedVertex['className'] = 'ForceDirectedVertex';

/**
 * This holds {@link ForceDirectedLayout}-specific information about {@link Link}s.
 *
 * This class inherits from {@link LayoutEdge}.
 * @extends LayoutEdge
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class ForceDirectedEdge extends LayoutEdge {
  /** @hidden @internal */ public static className = 'ForceDirectedEdge';
  /** @hidden @internal */ private _stiffness: number;
  /** @hidden @internal */ private _length: number;

  constructor(network: ForceDirectedNetwork) {
    super(network);

    // force-specific properties:
    this._stiffness = NaN;

    this._length = NaN;
  }

  /**
   * Gets or sets this edge's stiffness or resistance to compression or stretching.
   * The default value is NaN.
   */
  /** @export */ get stiffness(): number { return this._stiffness; }
  /** @export */ set stiffness(value: number) {
    if (this._stiffness !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedEdge, 'stiffness');
      this._stiffness = value;
    }
  }

  /**
   * Gets or sets the length of this edge.
   * The default value is NaN.
   */
  /** @export */ get length(): number { return this._length; }
  /** @export */ set length(value: number) {
    if (this._length !== value) {
      Util.checkPrimitive(value, 'number', ForceDirectedEdge, 'length');
      this._length = value;
    }
  }
} // #end ForceDirectedEdge
ForceDirectedEdge['className'] = 'ForceDirectedEdge';

/**
 * @hidden @internal
 * A minimal, not very random, seeded/predictable random number generator.
 * @unrestricted
 */
export class RandomNumberGenerator {
  /** @hidden @internal */ public static className = 'RandomNumberGenerator';
  /** @hidden @internal */ private I: number;
  /** @hidden @internal */ private R: number;
  /** @hidden @internal */ private Q: number;
  /** @hidden @internal */ private M: number;
  /** @hidden @internal */ private A: number;
  /** @hidden @internal */ private seed: number;

  /**
   * @param {number=} seed
   */
  constructor(seed?: number) {
    if (seed === undefined) seed = 42;
    this.seed = seed;
    this.A = 48271;
    this.M = 2147483647;
    this.Q = this.M / this.A;
    this.R = this.M % this.A;
    this.I = 1.0 / this.M;
    this.random();
  }

  /**
   * @hidden @internal
   * @this {RandomNumberGenerator}
   * @return {number}
   */
  public random(): number {
    const hi = this.seed / this.Q;
    const lo = this.seed % this.Q;
    const test = this.A * lo - this.R * hi;
    if (test > 0) {
      this.seed = test;
    } else {
      this.seed = test + this.M;
    }
    return (this.seed * this.I);
  }

} // #end RandomNumberGenerator
RandomNumberGenerator['className'] = 'RandomNumberGenerator';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
LayeredDigraphLayout
LayeredDigraphNetwork
LayeredDigraphVertex
LayeredDigraphEdge
*/

// import { List } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Layout, LayoutEdge, LayoutNetwork, LayoutVertex } from './layout';
// import { Group, Part } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Spot } from './spot';
// import { Util } from './util';

/**
 * This arranges nodes of directed graphs into layers (rows or columns).
 * There are many samples that use LayeredDigraphLayout.
 * <p class="boxrun">
 * If you want to experiment interactively with most of the properties, try the <a href="../../samples/ldLayout.html">Layered Digraph Layout</a> sample.
 * See samples that make use of LayeredDigraphLayout in the <a href="../../samples/index.html#layereddigraphlayout">samples index</a>.
 *
 * The {@link #layerSpacing} property controls the distance between layers.
 * The {@link #columnSpacing} property controls the breadth of each "column" --
 * this affects the distance between nodes within a layer, although the exact distance also depends on the breadth of each node.
 * The {@link #layeringOption} property determines whether nodes without links coming in or without links going out are
 * lined up at the edge of the graph, or whether they are positioned close to their connected nodes.
 *
 * By default the layout will route the links in a manner that is consistent with the {@link #direction}.
 * So, for example, if the {@link #direction} is 90 degrees (i.e. downward), the links are expected to go from the top towards the bottom.
 * That means the links should come out from the bottom of the ports and should go into the top of the ports.
 * Basically the layout will set {@link Link#fromSpot} to `Spot.Bottom` and {@link Link#toSpot} to `Spot.Top`.
 *
 * If you want to the links to use the spots that are given by the ports or by the links themselves, you will need to set
 * {@link #setsPortSpots} to false to prevent this layout from setting the spots on the links.
 * For example, if each node only has one port that is the whole node, and if you want the links to be spread out along the sides
 * of the nodes, then you should set {@link #setsPortSpots} to false and set the node's {@link GraphObject#fromSpot} to
 * `Spot.BottomSide` and {@link GraphObject#toSpot} to `Spot.TopSide`.
 *
 * This layout handles links that form cycles better than {@link TreeLayout} does.
 * The normal routing behavior for "backwards" links is to route them "around" the source node and "around" the destination node,
 * so that all links come in one side and go out the other side.
 * However if you want "backwards" links to go more directly between nodes, set {@link #setsPortSpots} to false and
 * the node's {@link GraphObject#fromSpot} and {@link GraphObject#toSpot} both to `Spot.TopBottomSides`.
 * (Of course if the {@link #direction} is zero or 180, you'll want to use `Spot.LeftRightSides`.
 *
 * If the diagram is structured in a tree-like fashion,
 * it may be better to use {@link TreeLayout},
 * which has more options specific to trees.
 * {@link TreeLayout} is much faster than LayeredDigraphLayout,
 * and can handle a limited number of links that would prevent the
 * graph structure from being a true tree (i.e. some nodes having multiple parents).
 *
 * This layout makes use of a {@link LayoutNetwork} of
 * {@link LayeredDigraphVertex}es and {@link LayeredDigraphEdge}s that normally
 * correspond to the {@link Node}s and {@link Link}s of the {@link Diagram}.
 *
 * The layout algorithm consists of four-major steps: Cycle Removal,
 * Layer Assignment, Crossing Reduction, and Straightening and Packing.
 * The layout cannot guarantee that it provides optimal positioning of nodes or routing of links.
 * @extends Layout
 * @unrestricted
 * @category Layout
 */
export class LayeredDigraphLayout extends Layout {
  /** @hidden @internal */ public static className = 'LayeredDigraphLayout';
  /** @hidden @internal */ private _layerSpacing: number;
  /** @hidden @internal */ private _columnSpacing: number;
  /** @hidden @internal */ private _direction: number;
  /** @hidden @internal */ private _cycleRemoveOption: EnumValue;
  /** @hidden @internal */ private _layeringOption: EnumValue;
  /** @hidden @internal */ private _initializeOption: EnumValue;
  /** @hidden @internal */ private _iterations: number;
  /** @hidden @internal */ private _aggressiveOption: EnumValue;
  /** @hidden @internal */ private _packOption: number;
  /** @hidden @internal */ private _setsPortSpots: boolean;
  /** @hidden @internal */ private _linkSpacing: number;
  /** @hidden @internal */ private _maxLayer: number;
  /** @hidden @internal */ private _maxIndex: number;
  /** @hidden @internal */ private _maxColumn: number;
  /** @hidden @internal */ private _minIndexLayer: number;
  /** @hidden @internal */ private _maxIndexLayer: number;
  /** @hidden @internal */ private _indices: Array<number>;
  /** @hidden @internal */ private _layerSpaceLeft: Array<number>;
  /** @hidden @internal */ private _layerSpaceRight: Array<number>;
  /** @hidden @internal */ private _layerPositions: Array<number>;
  /** @hidden @internal */ private _layerSpaces: Array<number>;
  /** @hidden @internal */ private _myDepthFirstSearchCycleRemovalTime: number;
  /** @hidden @internal */ private _mySavedLayout: Array<number>;
  /** @hidden @internal */ private _myCrossings: Array<number>;
  /** @hidden @internal */ private _myComponent: number;
  /** @hidden @internal */ private _myLayers: Array<Array<LayeredDigraphVertex>>;
  /** @hidden @internal */ private _arrangementOffset: Point;
  /** @hidden @internal */ private _myCachedNodeArrayLists: Array<Array<LayeredDigraphVertex>>;

  /**
   * Constructs a {@link LayeredDigraphLayout} with no {@link Layout#network}
   * and with no owning {@link Layout#diagram}.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(LayeredDigraphLayout);
    super();

    this._layerSpacing = 25.0;
    this._columnSpacing = 25.0;
    this._direction = 0.0;
    this._cycleRemoveOption = LayeredDigraphLayout.CycleDepthFirst;
    this._layeringOption = LayeredDigraphLayout.LayerOptimalLinkLength;
    this._initializeOption = LayeredDigraphLayout.InitDepthFirstOut;
    this._iterations = 4;
    this._aggressiveOption = LayeredDigraphLayout.AggressiveLess;
    this._packOption = LayeredDigraphLayout.PackAll; // should be bitwise combinable enum.
    this._setsPortSpots = true;
    this._linkSpacing = 4;

    this._maxLayer = -1;
    this._maxIndex = -1;
    this._maxColumn = -1;
    this._minIndexLayer = 0;
    this._maxIndexLayer = 0;
    this._indices = null;
    this._layerSpaceLeft = null;
    this._layerSpaceRight = null;
    this._layerPositions = null;
    this._layerSpaces = null;
    this._myDepthFirstSearchCycleRemovalTime = 0;
    this._mySavedLayout = null;
    // this._myMedians = null; //???
    // this._myColumnsPS = null; //???
    this._myCrossings = null;
    this._myComponent = 0;
    this._myLayers = null;
    this._arrangementOffset = new Point();
    this._myCachedNodeArrayLists = [];
    this._myCachedNodeArrayLists.length = 100;
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Layout.
   * @this {LayeredDigraphLayout}
   * @param {?} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    copy._layerSpacing = this._layerSpacing;
    copy._columnSpacing = this._columnSpacing;
    copy._direction = this._direction;
    copy._cycleRemoveOption = this._cycleRemoveOption;
    copy._layeringOption = this._layeringOption;
    copy._initializeOption = this._initializeOption;
    copy._iterations = this._iterations;
    copy._aggressiveOption = this._aggressiveOption;
    copy._packOption = this._packOption;
    copy._setsPortSpots = this._setsPortSpots;
    copy._linkSpacing = this._linkSpacing;
    // these are computed and thus do not need to be copied:
    // this._maxLayer
    // this._maxIndex
    // this._maxColumn
    // this._minIndexLayer
    // this._maxIndexLayer
    // this._indices
    // this._layerSpaceLeft
    // this._layerSpaceRight
    // this._layerPositions
    // this._layerSpaces
    // this._myDepthFirstSearchCycleRemovalTime
    // this._mySavedLayout
    // this._myCrossings
    // this._myComponent
    // this._myLayers
    // this._arrangementOffset
    // this._myCachedNodeArrayLists
    // this._myCachedNodeArrayLists.length
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {LayeredDigraphLayout}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === LayeredDigraphLayout) {
      if (arg.name.indexOf('Aggressive') === 0) {
        this.aggressiveOption = arg;
      } else if (arg.name.indexOf('Cycle') === 0) {
        this.cycleRemoveOption = arg;
      } else if (arg.name.indexOf('Init') === 0) {
        this.initializeOption = arg;
      } else if (arg.name.indexOf('Layer') === 0) {
        this.layeringOption = arg;
      } else {
        Util.throwError('Unknown enum value: ' + arg);
      }
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * Create a new {@link LayoutNetwork} of {@link LayeredDigraphVertex}es and {@link LayeredDigraphEdge}s.
   * @this {LayeredDigraphLayout}
   * @return {LayeredDigraphNetwork} a new {@link LayoutNetwork}.
   */
  public createNetwork(): LayeredDigraphNetwork {
    return new LayeredDigraphNetwork(this);
  }

  /**
   * Perform the layered digraph layout.
   *
   * If there is no {@link Layout#network}, this calls {@link #makeNetwork} to create a {@link LayoutNetwork} from the given collection of Parts.
   * This removes any reflexive edges in the network, since they should be ignored.
   *
   * In order to influence a vertex's layer, you can override {@link #assignLayers}.
   *
   * Finally this calls {@link Layout#updateParts} to commit the Node positions from the vertex positions.
   * {@link Layout#updateParts} calls {@link #commitLayout} within a transaction.
   * @this {LayeredDigraphLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');
    if (this.network === null) {
      this.network = this.makeNetwork(coll);
    }

    this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);

    this._clearCaches();
    if (this.network.vertexes.count > 0) {
      // Remove self-links from the input network.
      this.network.deleteSelfEdges();

      // Removes cycles from the input network by reversing some number of links.
      if (this.cycleRemoveOption !== LayeredDigraphLayout.CycleFromLayers) this.removeCycles();

      // Assigns every node in the input network to a layer.
      // The layering satisfies the following:
      //  if L is a link from node U to node V,
      //  then U.layer > V.layer;
      //  further, U.layer - V.layer >= LinkMinLength(L).
      // In addition, the nodes will satisfy
      //  for all nodes U, U.layer >= 0, with equality in at least one case,
      //  for all nodes U, U.layer <= maxLayer, with equality in at least one case.
      this._assignLayersInternal();

      // Just in this one case, depend on assignLayers to implicitly determine which edges are going in "reverse".
      if (this.cycleRemoveOption === LayeredDigraphLayout.CycleFromLayers) this.removeCycles();

      // Converts the input network into a proper digraph; i.e., artificial nodes and links
      //  are introduced into the network such that every link is between nodes in adjacent
      //  layers.  This has the effect of breaking up long links into a sequence of artificial
      //  nodes.
      this.makeProper();

      // Assigns every node in the input network an index number,
      //  such that nodes in the same layer will be labeled with
      //  consecutive indices in left to right order.
      // All consecutive layout operations will preserve or update
      //  the indices.
      this._initializeIndicesInternal();

      // Assigns every node in the input network a column number,
      //  such that nodes in the same layer will be labeled with
      //  increasing indices in left to right order.
      // In addition, a node U is assigned to a column such that
      //  2 * minColumnSpace(U) + 1 columns are "allocated" to node U,
      //  and no two nodes have overlapping "allocations" of columns.
      // All consecutive layout operations will preserve or update
      //  the columns.
      this.initializeColumns();

      // Reorders nodes within layers to reduce the total number of link
      //  crossings in the network.
      this.reduceCrossings();

      // Adjusts the columns of nodes in the network to produce a layout which reduces
      //  the number of bends and is tightly packed.
      this.straightenAndPack();

      // Update the "physical" positions of the nodes and links.
      this.updateParts();
    }
    this.network = null;
    this.isValidLayout = true;
  }

  /**
   * @hidden
   * (Undocumented)
   * The function LinkMinLength returns the minimum length of the
   * link represented by the LayeredDigraphEdge link.
   * The default implementation gives links a minimum length of 1.
   * This function can be overridden to provide "fine-tuning" of the layout.
   * @expose
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphEdge} edge
   * @return {number} Returns the minimum length of the link represented by link.
   */
  protected linkMinLength(edge: LayeredDigraphEdge): number {
    // var fromVertex = edge.fromVertex;
    // var toVertex = edge.toVertex;

    // var links = 0;
    // var succEdgeIter = fromVertex.destinationEdges;
    // while (succEdgeIter.next()) {
    //  var e = succEdgeIter.value;
    //  if (e.toVertex === toVertex) {
    //    links++;
    //  }
    // }
    // if (links > 1) return 2;  // NO -- depend on default routing to separate the links between nodes in adjacent layers
    return 1;
  }

  /**
   * @hidden @internal
   * The function LinkLengthWeight returns the weight of the link
   * represented by the LayeredDigraphEdge link.  This weight is used by
   * OptimalLinkLengthLayering to minimize weighted link lengths.
   * The default implementation gives all links a length weight of 1.
   * This function can be overridden to provide "fine-tuning" of the layout.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphEdge} edge
   * @return {number} Returns the weight of the link represented by link.
   */
  protected linkLengthWeight(edge: LayeredDigraphEdge): number {
    return 1;
  }

  /**
   * @hidden @internal
   * The function linkStraightenWeight returns the weight of the link
   * represented by the LayeredDigraphEdge link.  This weight is used by
   * the straightening methods to give priority straightening to those links
   * with higher weights.
   * The default implementation gives links between two "real" nodes a weight of 1,
   * links between a "real" node and an "artificial" node a weight of 4,
   * and links between two "artificial" nodes a weight of 8.
   * This function can be overridden to provide "fine-tuning" of the layout.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphEdge} edge
   * @return {number} Returns the weight of the link represented by link.
   */
  protected linkStraightenWeight(edge: LayeredDigraphEdge): number {
    const fromNode = edge.fromVertex.node;
    const toNode = edge.toVertex.node;

    if ((fromNode === null) && (toNode === null)) {
      return 8;
    }

    if ((fromNode === null) || (toNode === null)) {
      return 4;
    }

    return 1;
  }

  /**
   * @hidden
   * (Undocumented)
   * This function returns the minimum space reserved for this node from the center point
   * for the "depth" of the layer that it is in.
   *
   * The default implementation returns 0 for nodes that do not
   * correspond to top-level Go objects.  For nodes that do correspond
   * to top-level Go objects, the layer space is determined by the
   * width or height of the object depending on the {@link #direction}.
   * By default this adds 10 to the space, to account for port end segment lengths.
   *
   * @this {LayeredDigraphLayout}
   * @expose
   * @param {LayeredDigraphVertex} v
   * @param {boolean} topleft whether to return the distance from the vertex's position
   * @return {number} Returns the minimum space reserved above and below this node, in document coordinates.
   */
  protected nodeMinLayerSpace(v: LayeredDigraphVertex, topleft: boolean): number {
    if (v.node === null) return 0;
    // Don't use copy() function for performance reasons
    if (this._direction === 90 || this._direction === 270) {
      if (topleft) {
        return v.focus.y + 10;
      } else {
        return v.bounds.height - v.focus.y + 10;
      }
    } else {
      if (topleft) {
        return v.focus.x + 10;
      } else {
        return v.bounds.width - v.focus.x + 10;
      }
    }
  }

  /**
   * @hidden
   * (Undocumented)
   * The function nodeMinColumnSpace returns the minimum space
   * reserved to either side of this node.
   *
   * The default implementation returns 0 for nodes that do not
   * correspond to top-level Go objects.  For nodes that do correspond
   * to top-level Go objects, the column space is determined by the
   * width and height of the object divided by the {@link #columnSpcacing}.
   * Note: all sub-classes that override this method should ensure that
   * nodes that do not correspond to top-level Go objects have a minimum
   * column space of 0.
   * This function can be overridden to provide "fine-tuning" of the layout.
   * @expose
   * @param {LayeredDigraphVertex} v
   * @param {boolean} topleft whether to return the distance from the vertex's position
   * @return {number} Returns the minimum space reserved to either side of the center of this node, in units of {@link #columnSpacing}.
   */
  protected nodeMinColumnSpace(v: LayeredDigraphVertex, topleft: boolean): number {
    if (v.node === null) return 0;
    const s = (topleft ? v.nodeMinColumnSpaceTopLeft : v.nodeMinColumnSpaceBottomRight);
    if (s !== null) return s;
    // Don't use copy() function for performance reasons
    const direction = this._direction;
    if (direction === 90 || direction === 270) {
      if (topleft) {
        return v.nodeMinColumnSpaceTopLeft = ((v.focus.x / this._columnSpacing) + 1) | 0;
      } else {
        return v.nodeMinColumnSpaceBottomRight = (((v.bounds.width - v.focus.x) / this._columnSpacing) + 1) | 0;
      }
    } else {
      if (topleft) {
        return v.nodeMinColumnSpaceTopLeft = ((v.focus.y / this._columnSpacing) + 1) | 0;
      } else {
        return v.nodeMinColumnSpaceBottomRight = (((v.bounds.height - v.focus.y) / this._columnSpacing) + 1) | 0;
      }
    }
  }

  // public dumpNetwork() {
  //  var s = "";
  //  var it = this.network.vertexes.iterator;
  //  while (it.next()) {
  //    var v = it.value;
  //    var n = v.node;
  //    if (s !== "") s += "\n";
  //    s += (n !== null ? n.data.key : "(dummy)");
  //    s += " " + v.layer + " " + v.column + " " + v.index;
  //  }
  //  return s;
  // }

  /**
   * @hidden @internal
   * The function SaveLayout stores the layer, column, and index of all
   * nodes in an array of integers.
   *
   * See also {@link #restoreLayout}
   * @this {LayeredDigraphLayout}
   * @return {Array.<number>} Returns an integer array representation of the current layout.
   */
  private saveLayout(): Array<number> {
    if (this._mySavedLayout === null) {
      this._mySavedLayout = [];
    }
    let i = 0;

    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      // ??? is saving/restoring the vertex.layer unnecessary?
      this._mySavedLayout[i] = vertex.layer;
      i++;
      this._mySavedLayout[i] = vertex.column;
      i++;
      this._mySavedLayout[i] = vertex.index;
      i++;
    }
    return this._mySavedLayout;
  }

  /**
   * @hidden @internal
   * The function restoreLayout restores the layer, column, and index of all
   * nodes from an array of integers.
   *
   * See also {@link #saveLayout}
   * @this {LayeredDigraphLayout}
   * @param {Array.<number>} layout
   */
  private restoreLayout(layout: Array<number>): void {
    let i = 0;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      // ??? is saving/restoring the vertex.layer unnecessary?
      vertex.layer = layout[i];
      i++;
      vertex.column = layout[i];
      i++;
      vertex.index = layout[i];
      i++;
    }
  }

  // ********************************************************************************************
  // Crossings

  /**
   * @hidden @internal
   * Computes the crossing matrix between the unfixedLayer and its adjacent layers.
   * The direction argument indicates which adjacent layers should be taken into
   * consideration when computing the crossing matrix:
   * direction === 0  --  use unfixedLayer - 1 and unfixedLayer + 1
   * direction > 0  --  use unfixedLayer - 1 (sweeping away from layer 0)
   * direction < 0  --  use unfixedLayer + 1 (sweeping towards layer 0)
   * The resulting integer array can be used as follows:
   * if index1 and index2 are the indices corresponding to two nodes on the
   * unfixedLayer and crossmat is the crossing matrix, then
   * crossmat[index1 * indices[unfixedLayer] + index2]
   * is the number of crossing that occur if the node corresponding to index1 is
   * placed to the left of the node corresponding to index2.  If index1 === index2,
   * then crossmat[index1 * indices[unfixedLayer] + index2] is the number of crossings
   * between links to and from the node corresponding to index1.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {Array.<number>} Returns the total number of crossings in the network.
   */
  private crossingMatrix(unfixedLayer: number, direction: number): Array<number> {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'crossingMatrix:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'crossingMatrix:direction');
    }
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    if (this._myCrossings === null || this._myCrossings.length < num * num) {
      this._myCrossings = [];
    }
    const crossmat = this._myCrossings;

    // Process all pairs of nodes in the unfixed layer.
    for (let A = 0; A < num; A++) {
      // First, compute the crossings between links to and from unfixedLayerNodes[A].

      // Initially, no crossings discovered.
      let crossings = 0;

      // count unrelated nodes separating a node from the node it is supposed to be Near
      const vertex = unfixedLayerNodes[A];
      const near = vertex.near;
      let B = 0;
      let v: LayeredDigraphVertex;
      if (near !== null && near.layer === vertex.layer) {
        B = near.index;
        if (B > A) {
          for (let i = A + 1; i < B; i++) {
            v = unfixedLayerNodes[i];
            if (v.near !== near || v.artificialType !== near.artificialType) crossings++;
          }
        } else {
          for (let i = A - 1; i > B; i--) {
            v = unfixedLayerNodes[i];
            if (v.near !== near || v.artificialType !== near.artificialType) crossings++;
          }
        }
      }
      let links: Array<LayeredDigraphEdge>;
      let posA = 0;
      let edgeA: LayeredDigraphEdge;
      let indexA = 0;
      let portToPosA = 0;
      let portFromPosA = 0;
      let linksA: Array<LayeredDigraphEdge>;
      let posB = 0;
      let edgeB: LayeredDigraphEdge;
      let indexB = 0;
      let portToPosB = 0;
      let portFromPosB = 0;
      let linksB: Array<LayeredDigraphEdge>;

      // Compute crossings in positive direction.
      if (direction >= 0) {
        // The first link.
        links = unfixedLayerNodes[A].sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;
        for (posA = 0; posA < links.length; posA++) {
          edgeA = links[posA];
          if (edgeA.valid && (edgeA.fromVertex.layer !== unfixedLayer)) {
            // Record the index of the node the link originates from.
            indexA = edgeA.fromVertex.index;

            // Record the portToPos and portFromPos for the link.
            portToPosA = edgeA.portToPos;
            portFromPosA = edgeA.portFromPos;

            // The second link.
            for (posB = posA + 1; posB < links.length; posB++) {
              edgeB = links[posB];
              if (edgeB.valid && (edgeB.fromVertex.layer !== unfixedLayer)) {
                // Record the index of the node the link originates from.
                indexB = edgeB.fromVertex.index;

                // Record the portToPos and portFromPos for the link.
                portToPosB = edgeB.portToPos;
                portFromPosB = edgeB.portFromPos;

                // If the to position of the first link is
                //  to the left of the to position of the second link
                //  and the from position of the first link is
                //  to the right of the from position of the second link,
                //  then a crossing occurs.
                if ((portToPosA < portToPosB) &&
                  ((indexA > indexB) || ((indexA === indexB) && (portFromPosA > portFromPosB)))) {
                  crossings++;
                }
                // If the to position of the second link is
                //  to the left of the to position of the first link
                //  and the from position of the second link is
                //  to the right of the from position of the first link,
                //  then a crossing occurs.
                if ((portToPosB < portToPosA) &&
                  ((indexB > indexA) || ((indexB === indexA) && (portFromPosB > portFromPosA)))) {
                  crossings++;
                }
              }
            }
          }
        }
      }

      if (direction <= 0) {
        // The first link.
        links = unfixedLayerNodes[A].destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;
        for (posA = 0; posA < links.length; posA++) {
          edgeA = links[posA];
          if (edgeA.valid && (edgeA.toVertex.layer !== unfixedLayer)) {
            // Record the index of the node the link goes to.
            indexA = edgeA.toVertex.index;

            // Record the portToPos and portFromPos for the link.
            portToPosA = edgeA.portToPos;
            portFromPosA = edgeA.portFromPos;

            // The second link.
            for (posB = posA + 1; posB < links.length; posB++) {
              edgeB = links[posB];
              if (edgeB.valid && (edgeB.toVertex.layer !== unfixedLayer)) {
                // Record the index of the node the link goes to.
                indexB = edgeB.toVertex.index;

                // Record the portToPos and portFromPos for the link.
                portToPosB = edgeB.portToPos;
                portFromPosB = edgeB.portFromPos;

                // If the from position of the first link is
                //  to the left of the from position of the second link
                //  and the to position of the first link is
                //  to the right of the to position of the second link,
                //  then a crossing occurs.
                if ((portFromPosA < portFromPosB) &&
                  ((indexA > indexB) || ((indexA === indexB) && (portToPosA > portToPosB)))) {
                  crossings++;
                }
                // If the from position of the second link is
                //  to the left of the from position of the first link
                //  and the to position of the second link is
                //  to the right of the to position of the first link,
                //  then a crossing occurs.
                if ((portFromPosB < portFromPosA) &&
                  ((indexB > indexA) || ((indexB === indexA) && (portToPosB > portToPosA)))) {
                  crossings++;
                }
              }
            }
          }
        }
      }

      // Set the crossings of unfixedLayerNodes[A] with itself.
      crossmat[A * num + A] = crossings;

      // Now, compute the crossings between links to and from unfixedLayerNodes[A]
      //  and unfixedLayerNodes[B] for B > A.
      for (B = A + 1; B < num; B++) {
        // Initially, no crossings discovered.
        // The variable crossingAB records the number of crossings if
        //  unfixedLayerNodes[A] (the A node) is placed to the left
        //  of unfixedLayerNodes[B] (the B node).
        // The variable crossingBA records the number of crossings if
        //  unfixedLayerNodes[A] (the A node) is placed to the right
        //  of unfixedLayerNodes[B] (the B node).
        let crossingAB = 0;
        let crossingBA = 0;

        // Compute crossings in positive direction.
        if (direction >= 0) {
          linksA = unfixedLayerNodes[A].sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;
          linksB = unfixedLayerNodes[B].sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;

          // The first link.
          for (posA = 0; posA < linksA.length; posA++) {
            edgeA = linksA[posA];
            if (edgeA.valid && (edgeA.fromVertex.layer !== unfixedLayer)) {
              // Record the index of the node the link originates from.
              indexA = edgeA.fromVertex.index;

              // Record the portToPos and portFromPos for the link.
              portToPosA = edgeA.portToPos;
              portFromPosA = edgeA.portFromPos;

              // The second link.
              for (posB = 0; posB < linksB.length; posB++) {
                edgeB = linksB[posB];
                if (edgeB.valid && (edgeB.fromVertex.layer !== unfixedLayer)) {
                  // Record the index of the node the link originates from.
                  indexB = edgeB.fromVertex.index;

                  // Record the portToPos and portFromPos for the link.
                  portToPosB = edgeB.portToPos;
                  portFromPosB = edgeB.portFromPos;

                  // If the from position of the first link is
                  //  to the right of the from position of the second link,
                  //  then a crossing occurs if the first node is to the right
                  //  of the second node.
                  if ((indexA < indexB) || ((indexA === indexB) && (portFromPosA < portFromPosB))) {
                    crossingBA++;
                  }
                  // If the from position of the second link is
                  //  to the right of the from position of the first link,
                  //  then a crossing occurs if the first node is to the left
                  //  of the second node.
                  if ((indexB < indexA) || ((indexB === indexA) && (portFromPosB < portFromPosA))) {
                    crossingAB++;
                  }
                }
              }
            }
          }
        }

        // Compute crossings in negative direction.
        if (direction <= 0) {
          linksA = unfixedLayerNodes[A].destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;
          linksB = unfixedLayerNodes[B].destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;

          // The first link.
          for (posA = 0; posA < linksA.length; posA++) {
            edgeA = linksA[posA];
            if (edgeA.valid && (edgeA.toVertex.layer !== unfixedLayer)) {
              // Record the index of the node the link goes to.
              indexA = edgeA.toVertex.index;

              // Record the portToPos and portFromPos for the link.
              portToPosA = edgeA.portToPos;
              portFromPosA = edgeA.portFromPos;

              // The second link.
              for (posB = 0; posB < linksB.length; posB++) {
                edgeB = linksB[posB];
                if (edgeB.valid && (edgeB.toVertex.layer !== unfixedLayer)) {
                  // Record the index of the node the link goes to.
                  indexB = edgeB.toVertex.index;

                  // Record the portToPos and portFromPos for the link.
                  portToPosB = edgeB.portToPos;
                  portFromPosB = edgeB.portFromPos;

                  // If the to position of the first link is
                  //  to the right of the to position of the second link,
                  //  then a crossing occurs if the first node is to the right
                  if ((indexA < indexB) || ((indexA === indexB) && (portToPosA < portToPosB))) {
                    crossingBA++;
                  }
                  // If the to position of the second link is
                  //  to the right of the to position of the first link,
                  //  then a crossing occurs if the first node is to the left
                  //  of the second node.
                  if ((indexB < indexA) || ((indexB === indexA) && (portToPosB < portToPosA))) {
                    crossingAB++;
                  }
                }
              }
            }
          }
        }

        // Set the crossings of unfixedLayerNode[A] with unfixedLayerNode[B].
        crossmat[A * num + B] = crossingAB;
        crossmat[B * num + A] = crossingBA;
      }
    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return crossmat;
  }

  /**
   * @hidden @internal
   * (Undocumented)
   * Returns the total number of crossings in the network.
   * Internal method used by {@link #reduceCrossings}.
   * @expose
   * @this {LayeredDigraphLayout}
   * @return {number} Returns the total number of crossings in the network.
   */
  protected countCrossings(): number {
    let crossings = 0;
    // Sweep from layer 0 to layer maxLayer.
    for (let layer = 0; layer <= this._maxLayer; layer++) {
      const crossmat = this.crossingMatrix(layer, 1);
      // Since index i <= index j,
      //  summing the entries in the crossing matrix
      //  yields the total number of crossings in the layer.
      const num = this._indices[layer];
      for (let i = 0; i < num; i++) {
        for (let j = i; j < num; j++) {
          crossings += crossmat[i * num + j];
        }
      }
    }
    return crossings;
  }

  // ********************************************************************************************
  // Bends

  /**
   * @hidden @internal
   * Computes the bends between the unfixedLayer and its adjacent layers.
   * The "bend" between a node U and a node V connected by a link L is calcluated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset))
   * The "weighted bend" between a node U and a node V connected by link L is calculated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset)) * linkStraightenWeight(L)
   * The linkStraightenWeight attempts to give higher priority to links between
   * "artificial" nodes; i.e., long links in the final layout will be straighter.
   * The direction argument indicates which adjacent layers should be taken into
   * consideration when computing the crossing matrix:
   * direction === 0  --  use unfixedLayer - 1 and unfixedLayer + 1
   * direction > 0  --  use unfixedLayer - 1 (sweeping away from layer 0)
   * direction < 0  --  use unfixedLayer + 1 (sweeping towards layer 0)
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @param {boolean} weighted
   * @return {number} Returns the bends between the unfixedLayer and its adjacent layers.
   */
  private bends(unfixedLayer: number, direction: number, weighted: boolean): number {
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    // The total bends.
    let bends = 0;

    for (let index = 0; index < num; index++) {
      // Calculate the predecessor link list, based on direction.
      let LinksP: Array<LayeredDigraphEdge> = null;
      if (direction <= 0) {
        LinksP = unfixedLayerNodes[index].sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }
      // Calculate the successor link list, based on direction.
      let LinksS: Array<LayeredDigraphEdge> = null;
      if (direction >= 0) {
        LinksS = unfixedLayerNodes[index].destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }

      let edge: LayeredDigraphEdge;
      let fromColumn = 0;
      let toColumn = 0;
      if (LinksP !== null) {
        // Contribution to bends from predecessors not on same layer.
        for (let posP = 0; posP < LinksP.length; posP++) {
          edge = LinksP[posP] as LayeredDigraphEdge;
          if (edge.valid && (edge.fromVertex.layer !== unfixedLayer)) {
            fromColumn = edge.fromVertex.column + edge.portFromColOffset;
            toColumn = edge.toVertex.column + edge.portToColOffset;
            if (weighted) {
              bends += Math.abs(fromColumn - toColumn) * this.linkStraightenWeight(edge);
            } else {
              bends += Math.abs(fromColumn - toColumn);
            }
          }
        }
      }

      if (LinksS !== null) {
        // Contribution to bends from successors not on same layer.
        for (let posS = 0; posS < LinksS.length; posS++) {
          edge = LinksS[posS] as LayeredDigraphEdge;
          if (edge.valid && (edge.toVertex.layer !== unfixedLayer)) {
            fromColumn = edge.fromVertex.column + edge.portFromColOffset;
            toColumn = edge.toVertex.column + edge.portToColOffset;
            if (weighted) {
              bends += (Math.abs(fromColumn - toColumn) + 1) * this.linkStraightenWeight(edge);
            } else {
              bends += Math.abs(fromColumn - toColumn);
            }
          }
        }
      }

    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return bends;
  }

  /**
   * @hidden @internal
   * Returns the total number of bends in the network.
   * The "bend" between a node U and a node V connected by a link L is calcluated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset))
   * The "weighted bend" between a node U// and a node V connected by link L is calculated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset)) * linkStraightenWeight(L)
   * The linkStraightenWeight attempts to give higher priority to links between
   * "artificial" nodes; i.e., long links in the final layout will be straighter.
   * @this {LayeredDigraphLayout}
   * @param {boolean} weighted
   * @return {number} Returns the total number of bends in the network.
   */
  private countBends(weighted: boolean): number {
    let bends = 0;
    for (let layer = 0; layer <= this._maxLayer; layer++) {
      bends += this.bends(layer, 1, weighted);
    }
    return bends;
  }

  // ********************************************************************************************
  // Normalize

  /**
   * @hidden @internal
   * Adjusts the columns of all nodes such that the leftmost column will be
   * column 0 and maxColumn is updated appropriately.
   * @this {LayeredDigraphLayout}
   */
  private normalize(): void {
    let minColumn = Infinity;
    this._maxColumn = -1;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      minColumn = Math.min(minColumn, vertex.column - this.nodeMinColumnSpace(vertex, true));
      this._maxColumn = Math.max(this._maxColumn, vertex.column + this.nodeMinColumnSpace(vertex, false));
    }

    vertIter.reset();
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.column -= minColumn;
    }

    this._maxColumn -= minColumn;
  }

  // ********************************************************************************************
  // Barycenters and Medians

  /**
   * @hidden @internal
   * Computes the array of barycenters (average) columns for the nodes in the
   * unfixedLayer based on the columns of predecessors (direction < 0),
   * successors (direction > 0), or both predecessors and successors (direction === 0)
   * Elements without a defined barycenter will have an entry of -1.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {Array.<number>} Returns the array of barycenters (average) columns for the nodes in
   * the unfixedLayer.
   */
  private barycenters(unfixedLayer: number, direction: number): Array<number> {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'barycenters:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'barycenters:direction');
    }
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    const barycentersArr: Array<number> = [];

    for (let index = 0; index < num; index++) {
      const vertex = unfixedLayerNodes[index];

      // Calculate the predecessor link list, based on direction.
      let linksP: Array< LayeredDigraphEdge > = null;
      if (direction <= 0) {
        linksP = vertex.sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }
      // Calculate the successor link list, based on direction.
      let linksS: Array< LayeredDigraphEdge > = null;
      if (direction >= 0) {
        linksS = vertex.destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }

      // Compute the barycenter.
      // The variable total counts the total column position.
      // The variable N counts the number of predecessors or successors,
      // and any node in the same layer that this node is near.
      let total = 0;
      let N = 0;

      // regardless of direction, use the column of the node that should be Near to this node
      const near = vertex.near;
      if (near !== null && near.layer === vertex.layer) {
        total += near.column - 1;  // ??? -1
        N++;
      }
      let edge: LayeredDigraphEdge;
      if (linksP !== null) {
        // Contribution to barycenter from predecessors not on same layer.
        for (let posP = 0; posP < linksP.length; posP++) {
          edge = linksP[posP] as LayeredDigraphEdge;
          const fromVertex = edge.fromVertex;
          if (edge.valid && !edge.rev && (fromVertex.layer !== unfixedLayer)) {
            total += fromVertex.column; // + edge.portFromColOffset;  CAUSED INFINITE LOOP
            N++;
          }
        }
      }

      // INFINITE LOOP in bendStraighten with multiports:
      // nodes A (with output ports A0, A1), B (with input ports B0, B1)
      // links A0 - B0, A0 - B1
      // Due to non-monotonic improvements with bends

      if (linksS !== null) {
        // Contribution to barycenter from successors not on same layer.
        for (let posS = 0; posS < linksS.length; posS++) {
          edge = linksS[posS] as LayeredDigraphEdge;
          const toVertex = edge.toVertex;
          if (edge.valid && !edge.rev && (toVertex.layer !== unfixedLayer)) {
            total += toVertex.column; // + edge.portToColOffset;  CAUSED INFINITE LOOP
            N++;
          }
        }
      }

      // Calculate the barycenter.
      if (N === 0) {
        barycentersArr[index] = -1;
      } else {
        barycentersArr[index] = total / N;
      }
    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return barycentersArr;
  }

  /**
   * @hidden @internal
   * Computes the array of median columns for the nodes in the
   * unfixedLayer based on the columns of predecessors (direction < 0),
   * successors (direction > 0), or both predecessors and successors (direction === 0).
   * Elements without a defined median will have an entry of -1.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {Array.<number>} Returns the array of median columns for the nodes in
   * the unfixedLayer.
   */
  private medians(unfixedLayer: number, direction: number): Array<number> {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'medians:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'medians:direction');
    }
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    const mediansArr: Array<number> = [];

    for (let index = 0; index < num; index++) {
      const vertex = unfixedLayerNodes[index];

      // Calculate the predecessor link list, based on direction.
      let linksP: Array<LayeredDigraphEdge> = null;
      if (direction <= 0) {
        linksP = vertex.sourceEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }

      // Calculate the successor link list, based on direction.
      let linksS: Array<LayeredDigraphEdge> = null;
      if (direction >= 0) {
        linksS = vertex.destinationEdgesArrayAccess as Array<LayeredDigraphEdge>;
      }

      // Compute the median.
      // The variable N counts the number of predecessors or successors not on same layer,
      // and any node in the same layer that this node is near.
      let N = 0;

      // Allocate an array of integers to hold the columns of the predecessors
      //  and successors.
      const myColumnsPS: Array<number> = [];

      const near = vertex.near;
      if (near !== null && near.layer === vertex.layer) {
        myColumnsPS[N] = near.column - 1; // ??? -1
        N++;
      }

      // Count the total number of predecessors and successors not on same layer.
      // And fill in the myColumnsPS array with the columns of the predecessors and successors not on same layer.
      let edge: LayeredDigraphEdge;
      if (linksP !== null) {
        for (let posP = 0; posP < linksP.length; posP++) {
          edge = linksP[posP] as LayeredDigraphEdge;
          const fromVertex = edge.fromVertex;
          if (edge.valid && !edge.rev && (fromVertex.layer !== unfixedLayer)) {
            myColumnsPS[N] = fromVertex.column + edge.portFromColOffset;
            N++;
          }
        }
      }

      if (linksS !== null) {
        for (let posS = 0; posS < linksS.length; posS++) {
          edge = linksS[posS] as LayeredDigraphEdge;
          const toVertex = edge.toVertex;
          if (edge.valid && !edge.rev && (toVertex.layer !== unfixedLayer)) {
            myColumnsPS[N] = toVertex.column + edge.portToColOffset;
            N++;
          }
        }
      }

      if (N === 0) {
        mediansArr[index] = -1;
      } else {
        // sort the array of columns
        myColumnsPS.sort(function (a, b) { return a - b; });

        // Calculate the position of the median.
        const M = N >> 1;
        // If there is a distinct median, use it.
        if ((N & 1) !== 0) {
          mediansArr[index] = myColumnsPS[M];
          // If there are two medians, take their average.
        } else {
          mediansArr[index] = (myColumnsPS[M - 1] + myColumnsPS[M]) >> 1;
        }
      }
    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return mediansArr;
  }

  // ********************************************************************************************
  // Component methods

  /**
   * @hidden @internal
   * Uses a depth first search algorithm to set the component of all nodes in a component.
   * Tight** functions only set the component and recurse on nodes which are "tight",
   * in the sense that the nodes are separated by a link which corresponds to the
   * minimum link length of the link between the two nodes.
   * **Unset functions only set the component and recurse on nodes whose component is currently
   * set to the unset value.
   * The forward and backward bools indicate the direction to use for a
   * directed depth first search from node.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   * @param {number} component
   * @param {number} unset
   * @param {boolean} forward
   * @param {boolean} backward
   */
  private tightComponentUnset(v: LayeredDigraphVertex, component: number, unset: number, forward: boolean, backward: boolean): void {
    if (v.component === unset) {
      v.component = component;
      let length = 0;
      let minLength = 0;
      if (forward) {
        const succEdgeIter = v.destinationEdges;
        while (succEdgeIter.next()) {
          const se = succEdgeIter.value as LayeredDigraphEdge;
          const succVertex = se.toVertex as LayeredDigraphVertex;
          length = v.layer - succVertex.layer;
          minLength = this.linkMinLength(se);
          if (length === minLength) {
            this.tightComponentUnset(succVertex, component, unset, forward, backward);
          }
        }
      }

      if (backward) {
        const predEdgeIter = v.sourceEdges;
        while (predEdgeIter.next()) {
          const pe = predEdgeIter.value as LayeredDigraphEdge;
          const predVertex = pe.fromVertex as LayeredDigraphVertex;
          length = predVertex.layer - v.layer;
          minLength = this.linkMinLength(pe);
          if (length === minLength) {
            this.tightComponentUnset(predVertex, component, unset, forward, backward);
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Uses a depth first search algorithm to set the component of all nodes in a component.
   * Unset functions only set the component and recurse on nodes whose component is currently
   * set to the unset value.
   * The forward and backward bools indicate the direction to use for a
   * directed depth first search from node.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   * @param {number} component
   * @param {number} unset
   * @param {boolean} forward
   * @param {boolean} backward
   */
  private componentUnset(v: LayeredDigraphVertex, component: number, unset: number, forward: boolean, backward: boolean): void {
    if (v.component === unset) {
      v.component = component;
      if (forward) {
        const succEdgeIter = v.destinationEdges;
        while (succEdgeIter.next()) {
          const se = succEdgeIter.value;
          const succVertex = se.toVertex as LayeredDigraphVertex;
          this.componentUnset(succVertex, component, unset, forward, backward);
        }
      }

      if (backward) {
        const predEdgeIter = v.sourceEdges;
        while (predEdgeIter.next()) {
          const pe = predEdgeIter.value;
          const predVertex = pe.fromVertex as LayeredDigraphVertex;
          this.componentUnset(predVertex, component, unset, forward, backward);
        }
      }
    }
  }


  /**
   * @hidden
   * (Undocumented)
   * Removes cycles from the input network by reversing some number of links.
   * By default, this just calls {@link #greedyCycleRemoval} or {@link #depthFirstCycleRemoval},
   * as appropriate, give the value of {@link #cycleRemoveOption}.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected removeCycles(): void {
    const iter = this.network.edges.iterator;
    while (iter.next()) {
      const edge = iter.value as LayeredDigraphEdge;
      edge.rev = false;
    }
    switch (this._cycleRemoveOption) {
      default:
      case LayeredDigraphLayout.CycleGreedy:
        this.greedyCycleRemoval();
        break;
      case LayeredDigraphLayout.CycleDepthFirst:
        this.depthFirstSearchCycleRemoval();
        break;
      case LayeredDigraphLayout.CycleFromLayers:
        this.fromLayersCycleRemoval();
        break;
    }
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   */
  private fromLayersCycleRemoval(): void {
    const net = this.network as LayeredDigraphNetwork;
    const vit = net.vertexes.iterator;

    // First we have to normalize all of the vertex layer numbers
    // Minimum must be zero.
    let min = Infinity;
    while (vit.next()) {
      const v = vit.value as LayeredDigraphVertex;
      min = Math.min(min, v.layer);  // assume it's an integer
    }
    if (min < Infinity) {
      // make sure minimum layer is zero
      if (min < 0) {
        vit.reset();
        while (vit.next()) {
          const v = vit.value as LayeredDigraphVertex;
          v.layer -= min;
        }
      }

      // now collect all vertexes by layer
      const arr = [] as Array<Array<LayeredDigraphVertex>>;
      vit.reset();
      while (vit.next()) {
        const v = vit.value as LayeredDigraphVertex;
        const vs = arr[v.layer];
        if (vs === undefined) {
          arr[v.layer] = [v];
        } else {
          vs.push(v);
        }
      }

      // decrement layer for each vertex above any empty layer
      let decr = 0;
      for (let l = 0; l < arr.length; l++) {
        const a = arr[l];
        if (!a || a.length === 0) {
          decr++;
        } else if (l > 0) {
          for (let i = 0; i < a.length; i++) {
            const v = a[i];
            v.layer -= decr;
          }
        }
      }

      // reverse the edges according to the layers of each edge's two vertexes
      const eit = net.edges.iterator;
      while (eit.next()) {
        const edge = eit.value as LayeredDigraphEdge;
        const fromVert = edge.fromVertex as LayeredDigraphVertex;
        const toVert = edge.toVertex as LayeredDigraphVertex;
        if (fromVert.layer < toVert.layer) {
          net.reverseEdge(edge);
          edge.rev = true;
        }
      }
    }
  }

  // ********************************************************************************************
  // Greedy Cycle Removal

  /**
   * @hidden @internal
   * Removes cycles from the input network using a Greedy-Cycle-Removal algorithm.
   * The idea is to induce an order on all nodes
   * in the network (U1, U2, U3, ..., Uk) such that for the majority of links L = (Ui, Uj)
   * it is true that i < j.  All links L = (Ui, Uj) such that i > j are reversed.
   * @this {LayeredDigraphLayout}
   */
  private greedyCycleRemoval(): void {
    const net = this.network as LayeredDigraphNetwork;
    // The array nodes will represent the ordering of the nodes.
    let l = 0;
    let r = net.vertexes.count - 1;
    const nodes: Array<LayeredDigraphVertex> = [];
    nodes.length = r + 1;

    // Every node is initially valid -- i.e., available to be added to the ordering.
    const vertIter = net.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.valid = true;
    }

    // While some node is still valid...
    while (this.greedyCycleRemovalFindNode(net) !== null) {
      let vertex: LayeredDigraphVertex;
      // Add all sink nodes to the ordering at the rightmost available position
      //  and invalidate the node.
      vertex = this.greedyCycleRemovalFindSink(net);
      while (vertex !== null) {
        nodes[r] = vertex;
        r--;
        vertex.valid = false;
        vertex = this.greedyCycleRemovalFindSink(net);
      }
      // Add all source nodes to the ordering at the leftmost available position,
      //  and invalidate them.
      vertex = this.greedyCycleRemovalFindSource(net);
      while (vertex !== null) {
        nodes[l] = vertex;
        l++;
        vertex.valid = false;
        vertex = this.greedyCycleRemovalFindSource(net);
      }
      // Add a node which maximizes outdeg - indeg at the leftmost available position,
      //  and invalidate the node.
      // This is the greedy step, which locally maximizes the number of "good" links
      //  for a given number of "bad" links.
      vertex = this.greedyCycleRemovalFindNodeMaxDegDiff(net);
      if (vertex !== null) {
        nodes[l] = vertex;
        l++;
        vertex.valid = false;
      }
    }

    // Set the index of each node to the appropriate order position.
    for (let nodeIndex = 0; nodeIndex < net.vertexes.count; nodeIndex++) {
      nodes[nodeIndex].index = nodeIndex;
    }

    // Reverse the necessary links.
    const edgeIter = net.edges.iterator;
    while (edgeIter.next()) {
      const edge = edgeIter.value as LayeredDigraphEdge;
      const fromVert = edge.fromVertex as LayeredDigraphVertex;
      const toVert = edge.toVertex as LayeredDigraphVertex;
      if (fromVert.index > toVert.index) {
        net.reverseEdge(edge);
        edge.rev = true;
      }
    }
  }

  /**
   * @hidden @internal
   * Finds a valid node in the network.
   * Returns null if no valid node exists.
   * Used by {@link #greedyCycleRemoval}.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphNetwork} network
   * @return {LayeredDigraphVertex} Returns a valid node in the network or null if no valid nodes exist.
   */
  private greedyCycleRemovalFindNode(network: LayeredDigraphNetwork): LayeredDigraphVertex {
    const iter = network.vertexes.iterator;
    while (iter.next()) {
      const vertex = iter.value as LayeredDigraphVertex;
      if (vertex.valid) {
        return vertex;
      }
    }
    return null;
  }

  /**
   * @hidden @internal
   * Finds a sink node in the network.  A node is considered a sink node if it is
   * valid and all of its successors are invalid.  A valid node with no successors
   * is vacuously a sink.
   * Returns null if no valid sink node exists.
   * Used by {@link #greedyCycleRemoval}.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphNetwork} network
   * @return {LayeredDigraphVertex} Returns a sink node in the network, or null if no valid sink node exists.
   */
  private greedyCycleRemovalFindSink(network: LayeredDigraphNetwork): LayeredDigraphVertex {
    const vertIter = network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      if (vertex.valid) {
        let sink = true;
        const edgeIter = vertex.destinationEdges;
        while (edgeIter.next()) {
          const toVert = edgeIter.value.toVertex as LayeredDigraphVertex;
          if (toVert.valid) {
            sink = false;
            break;
          }
        }
        if (sink) return vertex;
      }
    }
    return null;
  }

  /**
   * @hidden @internal
   * Finds a source node in the network.  A node is considered a source node if it is
   * valid and all of its predecessors are invalid.  A valid node with no predecessors
   * is vacuously a source.
   * Returns null if no valid source node exists.
   * Used by {@link #greedyCycleRemoval}.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphNetwork} network
   * @return {LayeredDigraphVertex} Returns a source node in the network, or null if no valid source node exists.
   */
  private greedyCycleRemovalFindSource(network: LayeredDigraphNetwork): LayeredDigraphVertex {
    const vertIter = network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      if (vertex.valid) {
        let source = true;
        const edgeIter = vertex.sourceEdges;
        while (edgeIter.next()) {
          const fromVert = edgeIter.value.fromVertex as LayeredDigraphVertex;
          if (fromVert.valid) {
            source = false;
            break;
          }
        }
        if (source) return vertex;
      }
    }
    return null;
  }

  /**
   * @hidden @internal
   * Finds a valid node in the network that maximizes outdeg - indeg.
   * The degree difference is computed using valid successors and predecessors.
   * Returns null if no valid node exists.
   * Used by {@link #greedyCycleRemoval}.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphNetwork} network
   * @return {LayeredDigraphVertex} Returns a valid node in the network that maximized outdeg-indeg or null
   * if no valid node exists.
   */
  private greedyCycleRemovalFindNodeMaxDegDiff(network: LayeredDigraphNetwork): LayeredDigraphVertex {
    let vertexMax: LayeredDigraphVertex = null;
    let max = 0;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      if (vertex.valid) {
        let outdeg = 0;
        const succEdgeIter = vertex.destinationEdges;
        while (succEdgeIter.next()) {
          const toVert = succEdgeIter.value.toVertex as LayeredDigraphVertex;
          if (toVert.valid) {
            outdeg++;
          }
        }
        let indeg = 0;
        const predEdgeIter = vertex.sourceEdges;
        while (predEdgeIter.next()) {
          const fromVert = predEdgeIter.value.fromVertex as LayeredDigraphVertex;
          if (fromVert.valid) {
            indeg++;
          }
        }
        if ((vertexMax === null) || (max < (outdeg - indeg))) {
          vertexMax = vertex;
          max = outdeg - indeg;
        }
      }
    }
    return vertexMax;
  }

  // ********************************************************************************************
  // DepthFirst Cycle Removal

  /**
   * @hidden @internal
   * Removes cycles from the input network using a depth first search.
   * A link not in the depth first forest is reversed if the from-node was discovered
   * and finished by the depth first search after the to-node was discovered but before
   * the to-node was finished.
   * @this {LayeredDigraphLayout}
   */
  private depthFirstSearchCycleRemoval(): void {
    // Every node is initially undiscovered and unfinished.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.discover = -1;
      v.finish = -1;
    }
    // Every link is initially not in the depth first forest.
    const edgeIter = this.network.edges.iterator;
    while (edgeIter.next()) {
      const e = edgeIter.value as LayeredDigraphEdge;
      e.forest = false;
    }

    this._myDepthFirstSearchCycleRemovalTime = 0;

    vertIter.reset();
    while (vertIter.next()) {
      const v1 = vertIter.value as LayeredDigraphVertex;
      if (v1.sourceEdges.count === 0) {
        this.depthFirstSearchCycleRemovalVisit(v1);
      }
    }

    vertIter.reset();
    while (vertIter.next()) {
      const v2 = vertIter.value as LayeredDigraphVertex;
      if (v2.discover === -1) {
        this.depthFirstSearchCycleRemovalVisit(v2);
      }
    }

    // Scan through all non-forest links and reverse the appropriate links.
    edgeIter.reset();
    while (edgeIter.next()) {
      const edge = edgeIter.value as LayeredDigraphEdge;
      if (!(edge.forest)) {
        const fromVertex = edge.fromVertex as LayeredDigraphVertex;
        const fDiscover = fromVertex.discover;
        const fFinish = fromVertex.finish;

        const toVertex = edge.toVertex as LayeredDigraphVertex;
        const tDiscover = toVertex.discover;
        const tFinish = toVertex.finish;

        if ((tDiscover < fDiscover) && (fFinish < tFinish)) {
          this.network.reverseEdge(edge);
          edge.rev = true;
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Performs the recursive step of the depth first search on node.
   * Updates the discover and finish time of node.
   * Updates the forest flag of followed links.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private depthFirstSearchCycleRemovalVisit(v: LayeredDigraphVertex): void {
    v.discover = this._myDepthFirstSearchCycleRemovalTime;
    this._myDepthFirstSearchCycleRemovalTime++;
    const edgeIter = v.destinationEdges;
    while (edgeIter.next()) {
      const edge = edgeIter.value as LayeredDigraphEdge;
      const vertexS = edge.toVertex as LayeredDigraphVertex;
      if (vertexS.discover === -1) {
        edge.forest = true;
        this.depthFirstSearchCycleRemovalVisit(vertexS);
      }
    }
    v.finish = this._myDepthFirstSearchCycleRemovalTime;
    this._myDepthFirstSearchCycleRemovalTime++;
  }

  // ********************************************************************************************
  // Layer Assignment

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   */
  private _assignLayersInternal(): void {
    // Every node is initially unlayered.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      vertex.layer = -1;
    }
    this._maxLayer = -1;
    this.assignLayers();
    // Determine the actual maximum
    vertIter.reset();
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      this._maxLayer = Math.max(this._maxLayer, v.layer);
    }
  }

  /**
   * Assigns every vertex in the input network to a layer.
   * The layer is a non-negative integer describing which row of vertexes each vertex belongs in.
   * (Do not confuse this concept of "layer" with {@link Layer}s that control the Z-ordering of Parts.)
   *
   * The layering satisfies the following relationship:
   * if L is a link from node U to node V, then U.layer > V.layer.
   *
   * This method can be overridden to customize how nodes are assigned layers.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * By default, this does the appropriate assignments given the value of {@link #layeringOption}.
   * @expose
   * @this {LayeredDigraphLayout}
   * @since 1.1
   */
  protected assignLayers(): void {
    switch (this._layeringOption) {
      case LayeredDigraphLayout.LayerLongestPathSink:
        this.longestPathSinkLayering();
        break;
      case LayeredDigraphLayout.LayerLongestPathSource:
        this.longestPathSourceLayering();
        break;
      default:
      case LayeredDigraphLayout.LayerOptimalLinkLength:
        this.optimalLinkLengthLayering();
        break;
    }
  }

  // ********************************************************************************************
  // Longest Path Sink Layering

  /**
   * @hidden @internal
   * Assigns every node in the input network to a layer.
   * In addition to the requirements described in {@link #assignLayers},
   * LongestPathSinkLayering ensures that every sink appears in layer 0
   * and every node is as close to a sink as possible.
   * @this {LayeredDigraphLayout}
   */
  private longestPathSinkLayering(): void {
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const layer = this.longestPathSinkLayeringLength(vertex);
      this._maxLayer = Math.max(layer, this._maxLayer);
    }
  }

  /**
   * @hidden @internal
   * Computes the length of the longest path from node to a sink node and sets the
   * layer of node to that length.
   * Returns the length of the longest path from node to a sink node.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   * @return {number} Returns the length of the longest path from node to a sink node.
   */
  private longestPathSinkLayeringLength(v: LayeredDigraphVertex): number {
    let length = 0;
    if (v.layer === -1) {
      const edgeIter = v.destinationEdges;
      while (edgeIter.next()) {
        const edge = edgeIter.value as LayeredDigraphEdge;
        const succVertex = edge.toVertex as LayeredDigraphVertex;
        const minLength = this.linkMinLength(edge);
        length = Math.max(length, this.longestPathSinkLayeringLength(succVertex) + minLength);
      }
      v.layer = length;
    } else {
      length = v.layer;
    }

    return length;
  }

  // ********************************************************************************************
  // Longest Path Source Layering

  /**
   * @hidden @internal
   * Assigns every node in the input network to a layer.
   * In addition to the requirements described in {@link #assignLayers},
   * LongestPathSourceLayering ensures that every source appears in layer maxlayer
   * and every node is as close to a source as possible.
   * @this {LayeredDigraphLayout}
   */
  private longestPathSourceLayering(): void {
    let layer = 0;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      layer = this.longestPathSourceLayeringLength(v);
      this._maxLayer = Math.max(layer, this._maxLayer);
    }

    // Restore the correct orientation of the network,
    //  since LongestPathSourceLayeringLength assigns layers "backwards".
    vertIter.reset();
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      vertex.layer = this._maxLayer - vertex.layer;
    }
  }

  /**
   * @hidden @internal
   * Computes the length of the longest path from node to a source node and sets the
   * layer of node to that length.
   * Returns the length of the longest path from node to a source node.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   * @return {number} Returns the length of the longest path from node to a source node.
   */
  private longestPathSourceLayeringLength(v: LayeredDigraphVertex): number {
    let length = 0;
    if (v.layer === -1) {
      const edgeIter = v.sourceEdges;
      while (edgeIter.next()) {
        const edge = edgeIter.value as LayeredDigraphEdge;
        const predVertex = edge.fromVertex as LayeredDigraphVertex;
        const minLength = this.linkMinLength(edge);
        length = Math.max(length, this.longestPathSourceLayeringLength(predVertex) + minLength);
      }
      v.layer = length;
    } else {
      length = v.layer;
    }

    return length;
  }

  // ********************************************************************************************
  // Optimal Link Length Layering

  /**
   * @hidden @internal
   * Assigns every node in the input network to a layer.
   * In addition to the requirements described in AssignLayers(),
   * OptimalLinkLengthLayering ensures that nodes are set in layers
   * to minimize the total weighted link length.
   * Hence, OptimalLinkLengthLayering minimizes the sum
   * (U.layer - V.layer) * LinkLengthWeight(L)
   * over all links L = (U,V).
   * @this {LayeredDigraphLayout}
   */
  private optimalLinkLengthLayering(): void {
    // Begin by performing a LongestPathSinkLayering.
    // This ensures that all sinks are in layer 0 and every node
    //  is as close to a sink as possible.
    this.longestPathSinkLayering();

    // Every node is initially invalid -- i.e., it hasn't been reached by
    //  the depth first search.
    // This is mainly an optimization -- depth first searching through a
    //  portion of the network that is already valid will not change
    //  the layering, but it will spend the time calculating that nothing needs
    //  to be done.  By only examining invalid nodes, the running time improves.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.valid = false;
    }

    // Perform a depth first search from the sources of the network.
    // At the conclusion of this depth first search, all nodes have been
    //  moved into their appropriate layers relative to their decendents.
    vertIter.reset();
    while (vertIter.next()) {
      const vert = vertIter.value as LayeredDigraphVertex;
      if (vert.sourceEdges.count === 0) {
        this.optimalLinkLengthLayeringDepthFirstSearch(vert);
      }
    }

    // However, the entire network may have been shifted.
    // Find the minimum layer.
    let minLayer = Infinity;
    vertIter.reset();
    while (vertIter.next()) {
      const v2 = vertIter.value as LayeredDigraphVertex;
      minLayer = Math.min(minLayer, v2.layer);
    }
    // And move evey node so that the minimum layer is now layer 0.
    // Also recompute the maxLayer, since nodes may have been pushed.
    this._maxLayer = -1;
    vertIter.reset();
    while (vertIter.next()) {
      const vert2 = vertIter.value as LayeredDigraphVertex;
      vert2.layer -= minLayer;
      this._maxLayer = Math.max(this._maxLayer, vert2.layer);
    }
  }

  /**
   * @hidden @internal
   * Performs the depth first search of the network.
   * After traversing all descendents, the node is "pull"-ed into
   * the appropriate layer.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private optimalLinkLengthLayeringDepthFirstSearch(v: LayeredDigraphVertex): void {
    if (!(v.valid)) {
      v.valid = true;
      const destEdgeIter = v.destinationEdges;
      while (destEdgeIter.next()) {
        const e = destEdgeIter.value;
        const succVertex = e.toVertex as LayeredDigraphVertex;
        this.optimalLinkLengthLayeringDepthFirstSearch(succVertex);
      }
      this.optimalLinkLengthLayeringPull(v);
      this.optimalLinkLengthLayeringPush(v);
    }
  }

  /**
   * @hidden @internal
   * Attempts to move node and its tight component to a higher layer.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private optimalLinkLengthLayeringPull(v: LayeredDigraphVertex): void {
    // Every node is initially unset -- i.e., it hasn't been assigned to
    //  a component.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vert = vertIter.value as LayeredDigraphVertex;
      vert.component = -1;
    }

    // Constants for parent and child components.
    /** @hidden @internal @constant */
    const parentComponent = 0;
    /** @hidden @internal @constant */
    const childComponent = 1;

    // Form the tight forward component of node's non-tight parents.
    // The idea is to exclude from node's component all those nodes which
    //  are already tight with respect to the parents of node, since
    //  node will move towards it's non-tight parents.
    const predEdgeIter = v.sourceEdgesArrayAccess;
    let l = predEdgeIter.length;
    for (let i = 0; i < l; i++) {
      const predEdge = predEdgeIter[i] as LayeredDigraphEdge;
      const fromVert = predEdge.fromVertex as LayeredDigraphVertex;
      const toVert = predEdge.toVertex as LayeredDigraphVertex;
      const minLength = this.linkMinLength(predEdge);
      const length = fromVert.layer - toVert.layer;
      if (length > minLength) {
        this.tightComponentUnset(fromVert, parentComponent, -1, true, false);
      }
    }

    // Form the tight component of node which excludes nodes in the
    //  parent component.
    // The tight component of node is the set of nodes which
    //  are decendents or ancestors of node and have a path of "tight"
    //  links from or to node.
    // A link L from node U to node V is tight if
    //  U.layer - V.layer === LinkMinLength(L)
    // Hence, if node can be shifted to a higher layer, all of these
    //  nodes must also be shifted.
    this.tightComponentUnset(v, childComponent, -1, true, true);

    // If node is in it's parent component, then it cannot be pulled
    //  (it would violate some LinkMinLength),
    //  or it has been cast out of the childComponent and a pull
    //  would not reduce the weighted sum of the links in the network.
    while (v.component !== parentComponent) {
      // Compute the difference in the weights of links into the childComponent and
      //  links out of the childCcomponent.
      let totalDiff = 0;
      // Keep track of the maximum amount by which the component can be shifted.
      let shift = Infinity;
      // Also compute the node in the childComponent which satisfies the following:
      //  the node has no children in the childComponent
      //  the node minimizes the weighted difference of links into the node
      //  and links out of the node (i.e., the node is "best" in the
      //  sense that it will most increase the totalDiff next time through
      //  the loop.
      let minDiff = 0;
      let minDiffVertex: LayeredDigraphVertex = null;

      // Step through all nodes in the network.
      const childCompVertexIter = this.network.vertexes.iterator;
      while (childCompVertexIter.next()) {
        const childCompVertex = childCompVertexIter.value as LayeredDigraphVertex;
        if (childCompVertex.component === childComponent) {
          // Compute the weighted difference of links into the node
          //  and links out of the node.
          // Also record if the node has no children in the childComponent.
          let diff = 0;
          let children = false;

          // Links into the node.
          const childCompPredEdgeIter = childCompVertex.sourceEdgesArrayAccess;
          l = childCompPredEdgeIter.length;
          for (let i = 0; i < l; i++) {
            const predEdge = childCompPredEdgeIter[i] as LayeredDigraphEdge;
            const predVertex = predEdge.fromVertex as LayeredDigraphVertex;
            // Update the node difference.
            diff += this.linkLengthWeight(predEdge);
            if (predVertex.component !== childComponent) {
              // Update the total difference.
              totalDiff += this.linkLengthWeight(predEdge);
              // Update shift if the link has the limiting length.
              const childCompLength = predVertex.layer - childCompVertex.layer;
              const childCompMinLength = this.linkMinLength(predEdge);
              shift = Math.min(shift, childCompLength - childCompMinLength);
            }
          }

          // Links out of the node.
          const childCompSuccEdgeIter = childCompVertex.destinationEdgesArrayAccess;
          l = childCompSuccEdgeIter.length;
          for (let i = 0; i < l; i++) {
            const succEdge = childCompSuccEdgeIter[i] as LayeredDigraphEdge;
            const succVertex = succEdge.toVertex as LayeredDigraphVertex;
            // Update the node difference.
            diff -= this.linkLengthWeight(succEdge);
            if (succVertex.component !== childComponent) {
              // Update the total difference.
              totalDiff -= this.linkLengthWeight(succEdge);
            } else {
              // Update the children.
              children = true;
            }
          }

          // Update the minimum node difference.
          if (((minDiffVertex === null) || (diff < minDiff)) && (!children)) {
            minDiffVertex = childCompVertex;
            minDiff = diff;
          }
        }
      }

      // If the sum of the weights of links into the component is greater than
      //  the sum of the weights out of the component, then shifting the
      //  component to a higher level will decrease the weighted sum of the
      //  links in the network.
      // If not, cast out the MinDiffNode (which will have the greatest effect
      //  in increasing the totalDiff next time through) and recompute.
      if (totalDiff > 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const vertex = vertIter.value as LayeredDigraphVertex;
          if (vertex.component === childComponent) {
            vertex.layer += shift;
            // Since the node was shifted, invalidate it so that additional shifts can be made.
            // pNode.valid = false;
          }
        }
        // Break the while loop.
        v.component = parentComponent;
      } else {
        minDiffVertex.component = parentComponent;
      }
    }
  }

  /**
   * @hidden @internal
   * Attempts to move node and its tight component to a lower layer.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private optimalLinkLengthLayeringPush(v: LayeredDigraphVertex): void {
    // Every node is initially unset -- i.e., it hasn't been assigned to
    //  a component.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vert = vertIter.value as LayeredDigraphVertex;
      vert.component = -1;
    }

    // Constants for parent and child components.
    /** @hidden @internal @constant */
    const parentComponent = 0;
    /** @hidden @internal @constant */
    const childComponent = 1;

    // Form the tight forward component of node.
    // The tight forward component of node is the set of nodes which
    //  are decendents of node and have a path of "tight"
    //  links from or to node.
    // A link L from node U to node V is tight if
    //  U.layer - V.layer === LinkMinLength(L)
    // Hence, if node can be shifted to a lower layer, all of these
    //  nodes must also be shifted.
    this.tightComponentUnset(v, childComponent, -1, true, false);

    // If node is in it's parent component, then it has been cast out of the
    //  childComponent and a push would not reduce the weighted sum of the
    //  links in the network.
    while (v.component !== parentComponent) {
      // Compute the difference in the weights of links into the childComponent and
      //  links out of the childCcomponent.
      let totalDiff = 0;
      // Keep track of the maximum amount by which the component can be shifted.
      let shift = Infinity;
      // Also compute the node in the childComponent which satisfies the following:
      //  the node has no parents in the childComponent
      //  the node maximizes the weighted difference of links into the node
      //   and links out of the node (i.e., the node is "best" in the
      //   sense that it will most decrease the totalDiff next time through
      //   the loop.
      let maxDiff = 0;
      let maxDiffVertex: LayeredDigraphVertex = null;
      // Step through all nodes in the network.
      const childCompVertexIter = this.network.vertexes.iterator;
      while (childCompVertexIter.next()) {
        const childCompVertex = childCompVertexIter.value as LayeredDigraphVertex;
        if (childCompVertex.component === childComponent) {
          // Compute the weighted difference of links into the node
          //  and links out of the node.
          // Also record if the node has no children in the childComponent.
          let diff = 0;
          let parents = false;

          // Links into the node.
          const predEdgeIter = childCompVertex.sourceEdgesArrayAccess;
          let l = predEdgeIter.length;
          for (let i = 0; i < l; i++) {
            const predEdge = predEdgeIter[i] as LayeredDigraphEdge;
            const predVertex = predEdge.fromVertex as LayeredDigraphVertex;
            // Update the node difference.
            diff += this.linkLengthWeight(predEdge);
            if (predVertex.component !== childComponent) {
              // Update the total difference.
              totalDiff += this.linkLengthWeight(predEdge);
            } else {
              // Update the parents.
              parents = true;
            }
          }
          // Links out of the node.
          const succEdgeIter = childCompVertex.destinationEdgesArrayAccess;
          l = succEdgeIter.length;
          for (let i = 0; i < l; i++) {
            const succEdge = succEdgeIter[i] as LayeredDigraphEdge;
            const succVertex = succEdge.toVertex as LayeredDigraphVertex;
            // Update the node difference.
            diff -= this.linkLengthWeight(succEdge);
            if (succVertex.component !== childComponent) {
              // Update the total difference.
              totalDiff -= this.linkLengthWeight(succEdge);
              // Update shift if the link has the limiting length.
              const childCompLength = childCompVertex.layer - succVertex.layer;
              const childCompMinLength = this.linkMinLength(succEdge);
              shift = Math.min(shift, childCompLength - childCompMinLength);
            }
          }

          // Update the maximum difference node.
          if (((maxDiffVertex === null) || (diff > maxDiff)) && (!parents)) {
            maxDiffVertex = childCompVertex;
            maxDiff = diff;
          }
        }
      }

      // If the sum of the weights of links into the component is greater than
      //  the sum of the weights out of the component, then shifting the
      //  component to a higher level will decrease the weighted sum of the
      //  links in the network.
      // If not, cast out the pMinDiffNode (which will have the greatest effect
      //  in increasing the totalDiff next time through) and recompute.
      if (totalDiff < 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const vertex = vertIter.value as LayeredDigraphVertex;
          if (vertex.component === childComponent) {
            vertex.layer -= shift;
            // Since the node was shifted, invalidate it so that additional shifts can be made.
            // node.valid = false;
          }
        }
        // Break the while loop.
        v.component = parentComponent;
      } else {
        maxDiffVertex.component = parentComponent;
      }
    }
  }

  // ********************************************************************************************
  // Make Proper methods

  /**
   * @hidden @internal
   * Converts the input network into a proper digraph; i.e., artificial nodes and links
   * are introduced into the network such that every link is between nodes in adjacent
   * layers.  This has the effect of breaking up long links into a sequence of artificial
   * nodes.
   * This method must not change the layer of any existing node, nor add or remove any layers.
   * @this {LayeredDigraphLayout}
   */
  private makeProper(): void {
    const net = this.network as LayeredDigraphNetwork;
    // Initially, all links are invalid.
    // As links are examined, they are either validated or a sequence of
    //  valid links is added to represent long list.
    const edgeArray: Array<LayeredDigraphEdge> = [];
    const edgeIter = net.edges.iterator;
    while (edgeIter.next()) {
      const edge = edgeIter.value as LayeredDigraphEdge;
      edge.valid = false;
      // The array is being constructed this way rather than using List.toArray since
      // the list is already being iterated through.
      edgeArray.push(edge);
    }

    // Since we are adding edges to the network as we examine links,
    //  only calculate ports on invalid links.  This will be the first
    //  set of links, which correspond to "real" links in the network.
    for (let i = 0; i < edgeArray.length; i++) {
      const edge = edgeArray[i] as LayeredDigraphEdge;
      let vertexFrom = edge.fromVertex as LayeredDigraphVertex;
      const vertexTo = edge.toVertex as LayeredDigraphVertex;
      if (edge.valid) continue;
      if ((vertexFrom.node === null || vertexTo.node === null) && vertexFrom.layer === vertexTo.layer) {
        continue;
      }
      // Set the portFromColOffset and portToColOffset and portFromPos and portToPos.
      // This allows the crossing matrix to correctly calculate the crossings
      //  for nodes with multiple ports and assists in straightening.
      let portFromColOffset = 0;
      let portToColOffset = 0;
      let portFromPos = 0;
      let portToPos = 0;

      if (edge.link !== null) {
        const reallink = edge.link;
        if (reallink === null) continue;
        const realFromNode = vertexFrom.node;
        const realToNode = vertexTo.node;
        if (realFromNode === null || realToNode === null) continue;
        let fromNode = reallink.fromNode;
        let toNode = reallink.toNode;
        let fromPort = reallink.fromPort;
        let toPort = reallink.toPort;
        if (edge.rev) {
          const dummy = fromNode;
          const dummyelt = fromPort;
          fromNode = toNode;
          fromPort = toPort;
          toNode = dummy;
          toPort = dummyelt;
        }
        const nodeFromPt = vertexFrom.focus;
        const nodeToPt = vertexTo.focus;

        const fromR = (edge.rev) ? vertexTo.bounds : vertexFrom.bounds;
        const portFromPt = Point.alloc();
        // if connected node (fromNode) is a member of a group (realFromNode) and it is visible
        if (realFromNode !== fromNode) {
          if (fromR.isReal() && fromNode.isVisible()) {
            if (fromNode.actualBounds.isReal()) {
              fromNode.getRelativePoint(fromPort, Spot.Center, portFromPt);
              portFromPt.x += (fromNode.actualBounds.x - fromR.x);
              portFromPt.y += (fromNode.actualBounds.y - fromR.y);
            } else {
              fromNode.getRelativePoint(fromPort, Spot.Center, portFromPt);
              if (!portFromPt.isReal()) portFromPt.assign(nodeFromPt);
            }
          } else {
            portFromPt.assign(nodeFromPt);
          }
        } else {  // just the relative position of the port on the node
          if (fromR.isReal()) {
            fromNode.getRelativePoint(fromPort, Spot.Center, portFromPt);
            if (!portFromPt.isReal()) portFromPt.assign(nodeFromPt);
          } else {
            portFromPt.assign(nodeFromPt);
          }
        }

        const toR = (edge.rev) ? vertexFrom.bounds : vertexTo.bounds;
        const portToPt = Point.alloc();
        // if connected node (toNode) is a member of a group (realToNode) and it is visible
        if (realToNode !== toNode) {
          if (toR.isReal() && toNode.isVisible()) {
            if (toNode.actualBounds.isReal()) {
              toNode.getRelativePoint(toPort, Spot.Center, portToPt);
              portToPt.x += (toNode.actualBounds.x - toR.x);
              portToPt.y += (toNode.actualBounds.y - toR.y);
            } else {
              toNode.getRelativePoint(toPort, Spot.Center, portToPt);
              if (!portToPt.isReal()) portToPt.assign(nodeToPt);
            }
          } else {
            portToPt.assign(nodeToPt);
          }
        } else {  // just the relative position of the port on the node
          if (toR.isReal()) {
            toNode.getRelativePoint(toPort, Spot.Center, portToPt);
            if (!portToPt.isReal()) portToPt.assign(nodeToPt);
          } else {
            portToPt.assign(nodeToPt);
          }
        }

        if (this._direction === 90 || this._direction === 270) {
          portFromColOffset = Math.round((portFromPt.x - nodeFromPt.x) / this._columnSpacing);
          portFromPos = portFromPt.x;
          portToColOffset = Math.round((portToPt.x - nodeToPt.x) / this._columnSpacing);
          portToPos = portToPt.x;
        } else {
          portFromColOffset = Math.round((portFromPt.y - nodeFromPt.y) / this._columnSpacing);
          portFromPos = portFromPt.y;
          portToColOffset = Math.round((portToPt.y - nodeToPt.y) / this._columnSpacing);
          portToPos = portToPt.y;
        }
        Point.free(portFromPt);
        Point.free(portToPt);
        edge.portFromColOffset = portFromColOffset;
        edge.portFromPos = portFromPos;
        edge.portToColOffset = portToColOffset;
        edge.portToPos = portToPos;
      } else {
        edge.portFromColOffset = 0;
        edge.portFromPos = 0;
        edge.portToColOffset = 0;
        edge.portToPos = 0;
      }

      let layerFrom = vertexFrom.layer;
      const layerTo = vertexTo.layer;
      const aroundNodes = this.goesAround(edge);
      const aroundfromnode = (aroundNodes === 1 || aroundNodes === 3) ? true : false;
      const aroundtonode = (aroundNodes === 2 || aroundNodes === 3) ? true : false;
      // Pointers to the "artifical" nodes and links added.
      let edgeA: LayeredDigraphEdge;
      let vertexA: LayeredDigraphVertex;
      // add an artifical node in the to-node's layer when a reversed link is coming back around
      if (aroundtonode) {
        //// Add an artificial node at the appropriate layer.
        vertexA = net.createVertex();
        vertexA.node = null;
        vertexA.artificialType = 1;
        vertexA.layer = layerFrom;
        vertexA.near = vertexFrom;
        net.addVertex(vertexA);
        // Link up the artifical node to its parent.
        edgeA = net.linkVertexes(vertexFrom, vertexA, edge.link) as LayeredDigraphEdge;
        // not between adjacent layers
        edgeA.valid = false;
        // Inherit the reversed status of the "real" link.
        edgeA.rev = edge.rev;
        edgeA.portFromColOffset = portFromColOffset;
        edgeA.portToColOffset = 0;
        edgeA.portFromPos = portFromPos;
        edgeA.portToPos = 0;
        vertexFrom = vertexA;
      }
      let loopback = 1;
      if (aroundfromnode) loopback--;
      if (layerFrom - layerTo > loopback && layerFrom > 0) {
        // Invalidate the "real" link.  The remainder of the layout algorithm will
        //  only work on "valid" links, which will be links between adjacent layers.
        edge.valid = false;
        // Add the first artificial node at the appropriate layer.
        vertexA = net.createVertex();
        vertexA.node = null;
        vertexA.artificialType = 2;
        vertexA.layer = layerFrom - 1;
        net.addVertex(vertexA);
        // Link up the artifical node to its parent.
        edgeA = net.linkVertexes(vertexFrom, vertexA, edge.link) as LayeredDigraphEdge;
        // Validate the new link, since it is between adjacent layers.
        edgeA.valid = true;
        // Inherit the reversed status of the "real" link.
        edgeA.rev = edge.rev;
        // The portFromColOffset is inherited from the "real" link,
        //  but portToColOffset is to an artifical node, so is 0.
        edgeA.portFromColOffset = (aroundtonode ? 0 : portFromColOffset);
        edgeA.portToColOffset = 0;
        edgeA.portFromPos = (aroundtonode ? 0 : portFromPos);
        edgeA.portToPos = 0;
        // Make the new artifical node the from node and loop.
        vertexFrom = vertexA;
        layerFrom--;
        // Add the intermediate artifical nodes.
        while (layerFrom - layerTo > loopback && layerFrom > 0) {
          // Add an artificial node at the appropriate layer.
          vertexA = net.createVertex();
          vertexA.node = null;
          vertexA.artificialType = 3;
          vertexA.layer = layerFrom - 1;
          net.addVertex(vertexA);
          // Link up the artifical node to its parent.
          edgeA = net.linkVertexes(vertexFrom, vertexA, edge.link) as LayeredDigraphEdge;
          // Validate the new link, since it is between adjacent layers.
          edgeA.valid = true;
          // Inherit the reversed status of the "real" link.
          edgeA.rev = edge.rev;
          // The portFromColOffset and portToColOffset are to artificial nodes,
          //  so are 0.
          edgeA.portFromColOffset = 0;
          edgeA.portToColOffset = 0;
          edgeA.portFromPos = 0;
          edgeA.portToPos = 0;
          // Make the new artifical node the from node and loop.
          vertexFrom = vertexA;
          layerFrom--;
        }

        // Link up the last artifical node to the "real" to node.
        edgeA = net.linkVertexes(vertexA, vertexTo, edge.link) as LayeredDigraphEdge;
        // Validate the new link, since it may be between adjacent layers.
        edgeA.valid = !aroundfromnode;
        if (aroundfromnode) vertexA.near = vertexTo;
        // Inherit the reversed status of the "real" link.
        edgeA.rev = edge.rev;
        // The portToColOffset is inherited from the "real" link,
        //  but portFromColOffset is to an artifical node, so is 0.
        edgeA.portFromColOffset = 0;
        edgeA.portToColOffset = portToColOffset;
        edgeA.portFromPos = 0;
        edgeA.portToPos = portToPos;
      } else {
        // The "real" link is between adjacent layers, so validate it.
        edge.valid = true;
      }
    }
  }

  /**
   * @hidden @internal
   * aroundfromnode and aroundtonode are outs.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphEdge} edge
   * @return {number}
   */
  private goesAround(edge: LayeredDigraphEdge): number {
    let around = 0;
    const pLink = edge.link;
    if (pLink !== null) {
      const gofromport = pLink.fromPort;
      const gotoport = pLink.toPort;
      if (gofromport !== null && gotoport !== null) {
        const gofromnode = pLink.fromNode;
        const gotonode = pLink.toNode;
        if (gofromnode !== null && gotonode !== null) {
          const deffromspot = this._getDefaultSpot(true);
          const deftospot = this._getDefaultSpot(false);
          const fromspot = this.setsPortSpots ? deffromspot : pLink.computeSpot(true, gofromport);
          const tospot = this.setsPortSpots ? deftospot : pLink.computeSpot(false, gotoport);
          const ortho = pLink.isOrthogonal;

          if (fromspot.isSide() && fromspot.includesSide(deftospot) &&
              tospot.isSide() && tospot.includesSide(deffromspot)) {
            return 0;
          }

          const frompt = pLink.getLinkPoint(gofromnode, gofromport, fromspot, true, ortho, gotonode, gotoport, Point.alloc());
          // account for node rotation
          const fromdir = pLink.getLinkDirection(gofromnode, gofromport, frompt, fromspot, true, ortho, gotonode, gotoport);
          Point.free(frompt);
          // need to go around the "from" node?
          if (!fromspot.isNone() && fromdir === this._oppositeSpotDir(edge, true)) {
            around += 1;
          } else if (this.setsPortSpots && gofromnode !== null && gofromnode.ports.count === 1) {
            if (edge.rev) {
              around += 1;
            }
          }

          const topt = pLink.getLinkPoint(gotonode, gotoport, tospot, false, ortho, gofromnode, gofromport, Point.alloc());
          // account for node rotation
          const todir = pLink.getLinkDirection(gotonode, gotoport, topt, tospot, false, ortho, gofromnode, gofromport);
          Point.free(topt);
          // need to go around the "to" node?
          if (!tospot.isNone() && todir === this._oppositeSpotDir(edge, false)) {
            around += 2;
          } else if (this.setsPortSpots && gotonode !== null && gotonode.ports.count === 1) {
            if (edge.rev) {
              around += 2;
            }
          }
        }
      }
    }
    return around;
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphEdge} edge
   * @param {boolean} from
   * @return {number}
   */
  private _oppositeSpotDir(edge: LayeredDigraphEdge, from: boolean): number {
    if (this._direction === 90) {
      if (from && !edge.rev || !from && edge.rev) {
        return 270;
      } else {
        return 90;
      }
    } else if (this._direction === 180) {
      if (from && !edge.rev || !from && edge.rev) {
        return 0;
      } else {
        return 180;
      }
    } else if (this._direction === 270) {
      if (from && !edge.rev || !from && edge.rev) {
        return 90;
      } else {
        return 270;
      }
    } else {
      if (from && !edge.rev || !from && edge.rev) {
        return 180;
      } else {
        return 0;
      }
    }
  }

  // ********************************************************************************************
  // Initialize Indices methods

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   */
  private _initializeIndicesInternal(): void {
    const indices: Array<number> = this._indices = [];
    for (let i = 0; i <= this._maxLayer; i++) {
      indices[i] = 0;
    }
    // Every node is initially unset -- i.e., it hasn't been assigned an index.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.index = -1;
    }

    this.initializeIndices();

    this._maxIndex = -1;
    this._minIndexLayer = 0;
    this._maxIndexLayer = 0;
    for (let i = 0; i <= this._maxLayer; i++) {
      if (indices[i] > indices[this._maxIndexLayer]) {
        this._maxIndex = indices[i] - 1;
        this._maxIndexLayer = i;
      }
      if (indices[i] < indices[this._minIndexLayer]) {
        this._minIndexLayer = i;
      }
    }

    // organize all nodes by layer, for faster access.
    this._myLayers = [];
    for (let i = 0; i < indices.length; i++) {
      this._myLayers[i] = [];
    }

    vertIter.reset();
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const layer = vertex.layer;
      const v2 = this._myLayers[layer];
      v2[vertex.index] = vertex;
    }
  }

  /**
   * @hidden
   * (Undocumented)
   * Assigns every node in the input network an index number,
   * such that nodes in the same layer will be labeled with
   * consecutive indices in left to right order.
   * All consecutive layout operations will preserve or update
   * the indices.
   *
   * In addition, the indices array is initialized such that
   * indices[layer] indicates the number of nodes in the layer.
   * Finally, the variables minIndexLayer and maxIndexLayer record
   * the layers that correspond to the minimum and maximum nodes
   * in a layer.
   * By default this will just call {@link #naiveInitializeIndices},
   * {@link #depthfirstOutInitializeIndices}, or
   * {@link #depthfirstInInitializeIndices}, as appropriate
   * given the value of {@link #initializeOption}.
   * This method must not change the layer of any existing node, nor add or remove any layers.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected initializeIndices(): void {
    switch (this._initializeOption) {
      default:
      case LayeredDigraphLayout.InitNaive:
        this.naiveInitializeIndices();
        break;
      case LayeredDigraphLayout.InitDepthFirstOut:
        this.depthFirstOutInitializeIndices();
        break;
      case LayeredDigraphLayout.InitDepthFirstIn:
        this.depthFirstInInitializeIndices();
        break;
    }
  }

  // ********************************************************************************************
  // InitNaive index initialization

  /**
   * @hidden @internal
   * Assigns every node in the input network an index number,
   * such that nodes in the same layer will be labeled with
   * consecutive indices in left to right order.
   * Uses a naive implementation that assigns indices to nodes as they
   * are encountered in a sweep of the network.  Because of the way
   * networks are stored, this has the effect of initially placing all
   * "artificial" nodes to the right of all "real" nodes.
   * @this {LayeredDigraphLayout}
   */
  private naiveInitializeIndices(): void {
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const layer = vertex.layer;
      vertex.index = this._indices[layer];
      this._indices[layer]++;
    }
  }

  // ********************************************************************************************
  // InitDepthFirstOut index initialization

  /**
   * @hidden @internal
   * Assigns every node in the input network an index number,
   * such that nodes in the same layer will be labeled with
   * consecutive indices in left to right order.
   * Uses a depth first "outward" (i.e., following links from "from-node" to "to-node")
   * traversal of the network, assigning indices to nodes as they are discovered.
   * @this {LayeredDigraphLayout}
   */
  private depthFirstOutInitializeIndices(): void {
    // Begin at maxLayer and sweep towards 0.
    // This ensures that the sources are discovered in the correct order.
    const vertIter = this.network.vertexes.iterator;
    for (let layer = this._maxLayer; layer >= 0; layer--) {
      vertIter.reset();
      while (vertIter.next()) {
        const vertex = vertIter.value as LayeredDigraphVertex;
        if ((vertex.layer === layer) && (vertex.index === -1)) {
          this.depthFirstOutInitializeIndicesVisit(vertex);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Assigns node the appropriate index and updates the indices array.
   * Implements the recursive portion of a depth first search.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private depthFirstOutInitializeIndicesVisit(v: LayeredDigraphVertex): void {
    const layer = v.layer;
    v.index = this._indices[layer];
    this._indices[layer]++;

    // Copy the destinationLinksList into succLinksArray and sort on portFromPos.
    // This ensures that multi-port nodes initialize children in the correct order.
    const succLinksArray = v.destinationEdgesArrayAccess;

    let exchange = true;
    while (exchange) {
      exchange = false;
      for (let indexS = 0; indexS < (succLinksArray.length - 1); indexS++) {
        const edgeL = succLinksArray[indexS] as LayeredDigraphEdge;
        const edgeR = succLinksArray[indexS + 1] as LayeredDigraphEdge;
        if (edgeL.portFromColOffset > edgeR.portFromColOffset) {
          exchange = true;
          succLinksArray[indexS] = edgeR;
          succLinksArray[indexS + 1] = edgeL;
        }
      }
    }

    for (let indexS = 0; indexS < succLinksArray.length; indexS++) {
      const edge = succLinksArray[indexS] as LayeredDigraphEdge;
      if (edge.valid) {
        const vertexS = edge.toVertex as LayeredDigraphVertex;
        if (vertexS.index === -1) {
          this.depthFirstOutInitializeIndicesVisit(vertexS);
        }
      }
    }
  }

  // ********************************************************************************************
  // InitDepthFirstIn index initialization

  /**
   * @hidden @internal
   * Assigns every node in the input network an index number,
   * such that nodes in the same layer will be labeled with
   * consecutive indices in left to right order.
   * Uses a depth first "inward" (i.e., following links from "to-node" to "from-node")
   * traversal of the network, assigning indices to nodes as they are discovered.
   * @this {LayeredDigraphLayout}
   */
  private depthFirstInInitializeIndices(): void {
    // Begin at 0 and sweep towards maxLayer.
    // This ensures that the sinks are discovered in the correct order.
    const vertIter = this.network.vertexes.iterator;
    for (let layer = 0; layer <= this._maxLayer; layer++) {
      vertIter.reset();
      while (vertIter.next()) {
        const vertex = vertIter.value as LayeredDigraphVertex;
        if ((vertex.layer === layer) && (vertex.index === -1)) {
          this.depthFirstInInitializeIndicesVisit(vertex);
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Assigns node the appropriate index and updates the indices array.
   * Implements the recursive portion of a depth first search.
   * @this {LayeredDigraphLayout}
   * @param {LayeredDigraphVertex} v
   */
  private depthFirstInInitializeIndicesVisit(v: LayeredDigraphVertex): void {
    const layer = v.layer;
    v.index = this._indices[layer];
    this._indices[layer]++;

    // Copy the sourceLinksList into predLinksArray and sort on portToPos.
    // This ensures that multi-port nodes initialize children in the correct order.
    const predLinksArray = v.sourceEdgesArrayAccess;

    let exchange = true;
    let indexP = 0;
    while (exchange) {
      exchange = false;
      for (indexP = 0; indexP < (predLinksArray.length - 1); indexP++) {
        const edgeL = predLinksArray[indexP] as LayeredDigraphEdge;
        const edgeR = predLinksArray[indexP + 1] as LayeredDigraphEdge;
        if (edgeL.portToColOffset > edgeR.portToColOffset) {
          exchange = true;
          predLinksArray[indexP] = edgeR;
          predLinksArray[indexP + 1] = edgeL;
        }
      }
    }

    let vertexP: LayeredDigraphVertex;
    let edge: LayeredDigraphEdge;
    for (indexP = 0; indexP < predLinksArray.length; indexP++) {
      edge = predLinksArray[indexP] as LayeredDigraphEdge;
      if (edge.valid) {
        vertexP = edge.fromVertex as LayeredDigraphVertex;
        if (vertexP.index === -1) {
          this.depthFirstInInitializeIndicesVisit(vertexP);
        }
      }
    }
  }

  // ********************************************************************************************
  // Initialize columns methods

  /**
   * @hidden @internal
   * Assigns every node in the input network a column number,
   * such that nodes in the same layer will be labeled with
   * increasing indices in left to right order.
   *
   * In addition, a node U is assigned to a column such that
   * 2 * MinColumnSpace(U) + 1 columns are "allocated" to node U,
   * and no two nodes have overlapping "allocations" of columns.
   * All consecutive layout operations will preserve or update
   * the columns.
   * This method can be overridden to customize the layout algorithm.
   * This method must not change the layer of any existing node, nor add or remove any layers.
   * @this {LayeredDigraphLayout}
   */
  private initializeColumns(): void {
    this._maxColumn = -1;
    for (let layer = 0; layer <= this._maxLayer; layer++) {
      // Fill unfixedLayerNodes with nodes in the layer, ordered by index.
      const unfixedLayerNodes = this._getCachedNodeArrayList(layer);

      // Assign columns to the unfixedLayerNodes such that "allocated"
      //  columns do not overlap.
      let column = 0;
      const layerIndex = this._indices[layer];
      for (let index = 0; index < layerIndex; index++) {
        const vertex = unfixedLayerNodes[index];
        column += this.nodeMinColumnSpace(vertex, true);
        vertex.column = column;
        column += 1;
        column += this.nodeMinColumnSpace(vertex, false);
      }

      this._maxColumn = Math.max(this._maxColumn, column - 1);
      this._freeCachedNodeArrayList(layer, unfixedLayerNodes);
    }
  }

  // ********************************************************************************************
  // Reduce Crossings methods

  /**
   * @hidden
   * (Undocumented)
   * Reorders nodes within layers to reduce the total number of link
   * crossings in the network.
   *
   * There are many, many possible implementations of this function.
   * Basically, some iteration of MedianBarycenterCrossingReduction
   * and AdjacentExchangeCrossingReductionBendStraighten
   * sweeping back and forth over the layers is needed.
   * The default implementation has performed favorably on a large number
   * of networks, but other options are available.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected reduceCrossings(): void {
    // Record the current number of crossings and the current layout.
    // These variables will be updated with the "best" crossings and
    //  layouts produced.
    let bestcrossings = this.countCrossings();
    let bestlayout = this.saveLayout();

    let count = 0;
    let layer = 0;
    let crossings = 0;

    // The first sequence of sweeps.
    for (count = 0; count < this._iterations; count++) {
      // Sweep from layer 0 to layer maxLayer,
      //  performing directed medianBarycenter and adjacentExchange
      //  crossing reductions on each layer.
      for (layer = 0; layer <= this._maxLayer; layer++) {
        this.medianBarycenterCrossingReduction(layer, 1);
        this.adjacentExchangeCrossingReductionBendStraighten(layer, 1, false, 1);
      }
      // Update the bestcrossings and bestlayout if an improvement was made.
      crossings = this.countCrossings();
      if (crossings < bestcrossings) {
        bestcrossings = crossings;
        bestlayout = this.saveLayout();
      }
      // Sweep from layer maxLayer to layer 0,
      //  performing directed medianBarycenter and adjacentExchange
      //  crossing reductions on each layer.
      for (layer = this._maxLayer; layer >= 0; layer--) {
        this.medianBarycenterCrossingReduction(layer, -1);
        this.adjacentExchangeCrossingReductionBendStraighten(layer, -1, false, -1);
      }
      // Update the bestcrossings and bestlayout if an improvement was made.
      crossings = this.countCrossings();
      if (crossings < bestcrossings) {
        bestcrossings = crossings;
        bestlayout = this.saveLayout();
      }
    }
    // Restore the bestlayout for subsequent operations.
    this.restoreLayout(bestlayout);

    // The second sequence of sweeps.
    for (count = 0; count < this._iterations; count++) {
      // Sweep from layer 0 to layer maxLayer,
      //  performing undirected medianBarycenter and adjacentExchange
      //  crossing reductions on each layer.
      for (layer = 0; layer <= this._maxLayer; layer++) {
        this.medianBarycenterCrossingReduction(layer, 0);
        this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0);
      }
      // Update the bestcrossings and bestlayout if an improvement was made.
      crossings = this.countCrossings();
      if (crossings < bestcrossings) {
        bestcrossings = crossings;
        bestlayout = this.saveLayout();
      }
      // Sweep from layer maxLayer to layer 0,
      //  performing undirected medianBarycenter and adjacentExchange
      //  crossing reductions on each layer.
      for (layer = this._maxLayer; layer >= 0; layer--) {
        this.medianBarycenterCrossingReduction(layer, 0);
        this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0);
      }
      // Update the bestcrossings and bestlayout if an improvement was made.
      crossings = this.countCrossings();
      if (crossings < bestcrossings) {
        bestcrossings = crossings;
        bestlayout = this.saveLayout();
      }
    }
    // Restore the bestlayout for subsequent operations.
    this.restoreLayout(bestlayout);

    // A final sequence of operations.
    // The idea is that sometimes the bestlayout was found early in
    //  the iterations, and not every operation got a chance to permute it.
    // This final sequence takes the bestlayout and uses all the operations
    //  to move it towards a local minimum number of crossings.
    // If the aggressiveOption is set, then every sub-interval of layers
    //  is examined to reduce the crossings in the whole graph.  The idea
    //  is that some "twisted links" just need to be permuted in the right
    //  direction for the right number of layers.  By considering all of the
    //  sub-intervals, we run accross each one.  Of course, with N layers,
    //  that's N(N+1)/2 sub-intervals to consider, and we keep considering
    //  these sub-intervals until no crossings can be reduced.  Hence, we
    //  leave it as a option.
    // If the aggressiveOption is not set, then only the full interval of layers
    //  is examined.
    let change = false;
    let topLayer = 0;
    let botLayer = 0;
    let lastcrossings = 0;
    let recent = 0;
    switch (this._aggressiveOption) {
      case LayeredDigraphLayout.AggressiveNone: break;
      case LayeredDigraphLayout.AggressiveMore:
        lastcrossings = bestcrossings + 1;
        while ((recent = this.countCrossings()) < lastcrossings) {
          lastcrossings = recent;
          for (topLayer = this._maxLayer; topLayer >= 0; topLayer--) {
            for (botLayer = 0; botLayer <= topLayer; botLayer++) {
              change = true;
              while (change) {
                change = false;
                for (layer = topLayer; layer >= botLayer; layer--) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, -1, false, -1) || change;
                }
              }
              crossings = this.countCrossings();
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }

              change = true;
              while (change) {
                change = false;
                for (layer = topLayer; layer >= botLayer; layer--) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 1, false, 1) || change;
                }
              }
              crossings = this.countCrossings();
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }

              change = true;
              while (change) {
                change = false;
                for (layer = botLayer; layer <= topLayer; layer++) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 1, false, 1) || change;
                }
              }
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }

              change = true;
              while (change) {
                change = false;
                for (layer = botLayer; layer <= topLayer; layer++) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, -1, false, -1) || change;
                }
              }
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }

              change = true;
              while (change) {
                change = false;
                for (layer = topLayer; layer >= botLayer; layer--) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0) || change;
                }
              }
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }

              change = true;
              while (change) {
                change = false;
                for (layer = botLayer; layer <= topLayer; layer++) {
                  change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0) || change;
                }
              }
              if (crossings >= bestcrossings) {
                this.restoreLayout(bestlayout);
              } else {
                bestcrossings = crossings;
                bestlayout = this.saveLayout();
              }
            }
          }
        }
        break;
      default:
      case LayeredDigraphLayout.AggressiveLess:
        topLayer = this._maxLayer;
        botLayer = 0;
        lastcrossings = bestcrossings + 1;
        while ((recent = this.countCrossings()) < lastcrossings) {
          lastcrossings = recent;
          change = true;
          while (change) {
            change = false;
            for (layer = topLayer; layer >= botLayer; layer--) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, -1, false, -1) || change;
            }
          }
          crossings = this.countCrossings();
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }
          change = true;
          while (change) {
            change = false;
            for (layer = topLayer; layer >= botLayer; layer--) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 1, false, 1) || change;
            }
          }
          crossings = this.countCrossings();
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }

          change = true;
          while (change) {
            change = false;
            for (layer = botLayer; layer <= topLayer; layer++) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 1, false, 1) || change;
            }
          }
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }
          change = true;
          while (change) {
            change = false;
            for (layer = botLayer; layer <= topLayer; layer++) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, -1, false, -1) || change;
            }
          }
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }
          change = true;
          while (change) {
            change = false;
            for (layer = topLayer; layer >= botLayer; layer--) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0) || change;
            }
          }
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }
          change = true;
          while (change) {
            change = false;
            for (layer = botLayer; layer <= topLayer; layer++) {
              change = this.adjacentExchangeCrossingReductionBendStraighten(layer, 0, false, 0) || change;
            }
          }
          if (crossings >= bestcrossings) {
            this.restoreLayout(bestlayout);
          } else {
            bestcrossings = crossings;
            bestlayout = this.saveLayout();
          }
        }
        break;
    }
    // Restore the bestlayout.
    this.restoreLayout(bestlayout);
  }

  /**
   * @hidden @internal
   * Reorders nodes within the unfixedLayer to reduce the number of link crossings between
   * the unfixedLayer and its adjacent layers.  The direction argument indicates which of the
   * adjacent layers should be taken into consideration when reducing the number of crossings.
   * direction === 0  --  use unfixedLayer - 1 and unfixedLayer + 1
   * direction > 0  --  use unfixedLayer - 1 (sweeping away from layer 0)
   * direction < 0  --  use unfixedLayer + 1 (sweeping towards layer 0)
   * The idea is to calculate the median and barycenter for each node in the unfixedLayer,
   * and to sort the nodes in the unfixedLayer by their median and barycenter values.
   * Returns true if some change was made to the layer.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {boolean} Returns true if some change was made to the layer and false otherwise.
   */
  private medianBarycenterCrossingReduction(unfixedLayer: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'medianBarycenterCrossingReduction:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'medianBarycenterCrossingReduction:direction');
    }
    let index = 0;
    let layerChange = false;

    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    // Initialize the medians and barycenters.
    const medians = this.medians(unfixedLayer, direction);
    const barycenters = this.barycenters(unfixedLayer, direction);

    // A median or barycenter value of -1 indicates that the node does not
    //  have a defined median or barycenter (i.e., the node has no neighbors
    //  in the appropriate adjacent layer).  In that case, use the current column
    //  value as the median or barycenter.
    for (index = 0; index < num; index++) {
      if (barycenters[index] === -1) {
        barycenters[index] = unfixedLayerNodes[index].column;
      }
      if (medians[index] === -1) {
        medians[index] = unfixedLayerNodes[index].column;
      }
    }

    // Sort nodes on medians and barycenters.
    // Nodes are permuted within the unfixedLayerNodes.
    let change = true;
    let vertex: LayeredDigraphVertex;
    while (change) {
      change = false;
      for (index = 0; index < (num - 1); index++) {
        if ((medians[index + 1] < medians[index]) ||
          ((medians[index + 1] === medians[index]) && (barycenters[index + 1] < barycenters[index]))) {
          layerChange = true;
          change = true;
          const medianTemp = medians[index];
          medians[index] = medians[index + 1];
          medians[index + 1] = medianTemp;
          const barycenterTemp = barycenters[index];
          barycenters[index] = barycenters[index + 1];
          barycenters[index + 1] = barycenterTemp;
          vertex = unfixedLayerNodes[index];
          unfixedLayerNodes[index] = unfixedLayerNodes[index + 1];
          unfixedLayerNodes[index + 1] = vertex;
        }
      }
    }

    // Update the index for each node in the unfixedLayer.
    // Update the column for each node in the unfixedLayer.
    let column = 0;
    for (index = 0; index < num; index++) {
      vertex = unfixedLayerNodes[index];
      vertex.index = index;
      column += this.nodeMinColumnSpace(vertex, true);
      vertex.column = column;
      column += 1;
      column += this.nodeMinColumnSpace(vertex, false);
    }

    // Return status of the unfixedLayer.
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return layerChange;
  }

  /**
   * @hidden @internal
   * Adjusts the columns of nodes within the unfixedLayer to simultaneously reduce the
   * number of link crossings and the number of "bends" between the unfixedLayer and its
   * adjacent layers.
   * The directionCR argument indicates which of the adjacent layers should be taken
   * into consideration when reducing the number of link crossings.
   * direction === 0  --  use unfixedLayer - 1 and unfixedLayer + 1
   * direction > 0  --  use unfixedLayer - 1 (sweeping away from layer 0)
   * direction < 0  --  use unfixedLayer + 1 (sweeping towards layer 0)
   * The directionBS argument indicates which of the adjacent layers should be taken
   * into consideration when reducing the number of bends.
   * direction === 0  --  use unfixedLayer - 1 and unfixedLayer + 1
   * direction > 0  --  use unfixedLayer - 1 (sweeping away from layer 0)
   * direction < 0  --  use unfixedLayer + 1 (sweeping towards layer 0)
   * The "weighted bend" between a node U and a node V connected by link L is calculated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset)) * LinkStraightenWeight(L)
   * The LinkStraightenWeight attempts to give higher priority to links between
   * "artificial" nodes; i.e., long links in the final layout will be straighter.
   * The idea is to use a bubble-sort technique to exchange adjacent nodes whenever
   * doing so reduces the number of link crossings or the number of bends.
   * This function is used in both crossing reduction and bend straightening.
   * Returns true if some change was made to the layer.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer The layer to be reordered.
   * @param {number} directionCR Indicates which adjacent layers should be taken into consideration when calculating the crossing matrix.
   * @param {boolean} straighten Indicates whether or not to reorder to nodes to straighten links.
   * @param {number} directionBS Indicates which adjacent layers should be taken into consideration when calculating the bends of a link.
   * @return {boolean} Returns true if some change was made to the layer and false otherwise.
   */
  private adjacentExchangeCrossingReductionBendStraighten(unfixedLayer: number, directionCR: number, straighten: boolean, directionBS: number): boolean {
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    // Initialize the crossing matrix such that
    //  if index1 and index2 are the indices corresponding to two nodes on the
    //  unfixedLayer and crossmat is the crossing matrix, then
    //   crossmat[index1 * indices[unfixedLayer] + index2]
    //  is the number of crossing that occur if the node corresponding to index1 is
    //  placed to the left of the node corresponding to index2.
    const crossmat = this.crossingMatrix(unfixedLayer, directionCR);
    let index = 0;
    // Initialize the barycenters from the negative direction.
    let barycentersNeg: Array<number>;
    // If no straightening or direction > 0,
    //  then set barycenters from the negative direction to -1.
    if ((!straighten) || (directionBS > 0)) {
      barycentersNeg = [];
      for (index = 0; index < num; index++) {
        barycentersNeg[index] = -1;
      }
    } else {
      barycentersNeg = this.barycenters(unfixedLayer, -1);
    }
    // Initialize the barycenters from the positive direction.
    let barycentersPos: Array<number>;
    // If no straightening or direction < 0,
    //  then set barycenters from the positive direction to -1.
    if ((!straighten) || (directionBS < 0)) {
      barycentersPos = [];
      for (index = 0; index < num; index++) {
        barycentersPos[index] = -1;
      }
    } else {
      barycentersPos = this.barycenters(unfixedLayer, 1);
    }

    let layerChange = false;
    let exchange = true;
    while (exchange) {
      exchange = false;
      for (index = 0; index < num - 1; index++) {
        const crossingLR = crossmat[unfixedLayerNodes[index].index * num + unfixedLayerNodes[index + 1].index];
        const crossingRL = crossmat[unfixedLayerNodes[index + 1].index * num + unfixedLayerNodes[index].index];
        // Calculate the crossing numbers of nodes on the same layer.
        let layerCrossingLR = 0;
        let layerCrossingRL = 0;

        // Column of the node to the left.
        const columnL = unfixedLayerNodes[index].column;
        // Column of the node to the right.
        const columnR = unfixedLayerNodes[index + 1].column;
        const indexSpaceLeft = this.nodeMinColumnSpace(unfixedLayerNodes[index], true);
        const indexSpaceRight = this.nodeMinColumnSpace(unfixedLayerNodes[index], false);
        const indexPlusOneSpaceLeft = this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true);
        const indexPlusOneSpaceRight = this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], false);
        // Column of the node to the left after an exchange (takes into account nodeMinColumSpace of the two nodes).
        const columnLX = columnL - indexSpaceLeft + indexPlusOneSpaceLeft;
        // Column of the node to the right after an exchange (takes into account nodeMinColumSpace of the two nodes).
        const columnRX = columnR - indexSpaceRight + indexPlusOneSpaceRight;
        // Calculate the bend numbers if the left node (L) and the right node (R) are
        //  order L-R  or R-L.
        let bendLR = 0.0;
        let bendRL = 0.0;
        let weight = 0.0;
        let portFromColOffset = 0;
        let portToColOffset = 0;
        let fromCol = 0;
        let toCol = 0;
        let fromVertex: LayeredDigraphVertex;
        let indexFrom = 0;
        let toVertex: LayeredDigraphVertex;
        let indexTo = 0;
        // Calculate the predecessor link list of the left node, based on direction
        const sourceEdgeIter = unfixedLayerNodes[index].sourceEdges.iterator;
        if (straighten && (directionBS <= 0)) {
          while (sourceEdgeIter.next()) {
            const se = sourceEdgeIter.value as LayeredDigraphEdge;
            fromVertex = se.fromVertex as LayeredDigraphVertex;
            if (se.valid && (fromVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(se);
              portFromColOffset = se.portFromColOffset;
              portToColOffset = se.portToColOffset;
              fromCol = fromVertex.column;
              bendLR += (Math.abs((columnL + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
              bendRL += (Math.abs((columnRX + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
            }
          }
        }

        // Compute "crossings" of nodes on same layer.
        sourceEdgeIter.reset();
        while (sourceEdgeIter.next()) {
          const se2 = sourceEdgeIter.value as LayeredDigraphEdge;
          fromVertex = se2.fromVertex as LayeredDigraphVertex;
          if (se2.valid && (fromVertex.layer === unfixedLayer)) {
            indexFrom = 0;
            while (unfixedLayerNodes[indexFrom] !== fromVertex) indexFrom++;
            // Depending on where the from node falls,
            //  we increase the appropriate crossing value.
            // The different values indicate how "strongly" that particular
            //  ordering is desired.
            if (indexFrom < index) {
              layerCrossingLR += 2 * (index - indexFrom);
              layerCrossingRL += 2 * ((index + 1) - indexFrom);
            }
            if (indexFrom === index + 1) {
              layerCrossingLR += 1;
            }
            if (indexFrom > index + 1) {
              layerCrossingLR += 4 * (indexFrom - index);
              layerCrossingRL += 4 * (indexFrom - (index + 1));
            }
          }
        }

        // Calculate the successor link list of the left node, based on direction
        const destinationEdgeIter = unfixedLayerNodes[index].destinationEdges.iterator;
        if (straighten && (directionBS >= 0)) {
          while (destinationEdgeIter.next()) {
            const de = destinationEdgeIter.value as LayeredDigraphEdge;
            toVertex = de.toVertex as LayeredDigraphVertex;
            if (de.valid && (toVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(de);
              portFromColOffset = de.portFromColOffset;
              portToColOffset = de.portToColOffset;
              toCol = toVertex.column;
              bendLR += (Math.abs((columnL + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
              bendRL += (Math.abs((columnRX + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
            }
          }
        }

        // Compute "crossings" of nodes on same layer.
        destinationEdgeIter.reset();
        while (destinationEdgeIter.next()) {
          const de2 = destinationEdgeIter.value as LayeredDigraphEdge;
          toVertex = de2.toVertex as LayeredDigraphVertex;
          if (de2.valid && (toVertex.layer === unfixedLayer)) {
            indexTo = 0;
            while (unfixedLayerNodes[indexTo] !== toVertex) indexTo++;
            // Depending on where the to node falls,
            //  we increase the appropriate crossing value.
            // The different values indicate how "strongly" that particular
            //  ordering is desired.
            if (indexTo === index + 1) {
              layerCrossingRL += 1;
            }
          }
        }

        // Calculate the predecessor link list of the right node, based on direction
        const rightSourceEdgeIter = unfixedLayerNodes[index + 1].sourceEdges.iterator;
        if (straighten && (directionBS <= 0)) {
          while (rightSourceEdgeIter.next()) {
            const rse = rightSourceEdgeIter.value as LayeredDigraphEdge;
            fromVertex = rse.fromVertex as LayeredDigraphVertex;
            if (rse.valid && (fromVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(rse);
              portFromColOffset = rse.portFromColOffset;
              portToColOffset = rse.portToColOffset;
              fromCol = fromVertex.column;
              bendLR += (Math.abs((columnR + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
              bendRL += (Math.abs((columnLX + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
            }
          }
        }

        // Compute "crossings" of nodes on same layer.
        rightSourceEdgeIter.reset();
        while (rightSourceEdgeIter.next()) {
          const rse2 = rightSourceEdgeIter.value as LayeredDigraphEdge;
          fromVertex = rse2.fromVertex as LayeredDigraphVertex;
          if (rse2.valid && (fromVertex.layer === unfixedLayer)) {
            indexFrom = 0;
            while (unfixedLayerNodes[indexFrom] !== fromVertex) indexFrom++;
            // Depending on where the from node falls,
            //  we increase the appropriate crossing value.
            // The different values indicate how "strongly" that particular
            //  ordering is desired.
            if (indexFrom < index) {
              layerCrossingLR += 2 * ((index + 1) - indexFrom);
              layerCrossingRL += 2 * (index - indexFrom);
            }
            if (indexFrom === index) {
              layerCrossingRL += 1;
            }
            if (indexFrom > index + 1) {
              layerCrossingLR += 4 * (indexFrom - (index + 1));
              layerCrossingRL += 4 * (indexFrom - index);
            }
          }
        }

        // Calculate the successor link list of the right node, based on direction
        const rightDestinationEdgeIter = unfixedLayerNodes[index + 1].destinationEdges.iterator;
        if (straighten && (directionBS >= 0)) {
          while (rightDestinationEdgeIter.next()) {
            const rde = rightDestinationEdgeIter.value as LayeredDigraphEdge;
            toVertex = rde.toVertex as LayeredDigraphVertex;
            if (rde.valid && (toVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(rde);
              portFromColOffset = rde.portFromColOffset;
              portToColOffset = rde.portToColOffset;
              toCol = toVertex.column;
              bendLR += (Math.abs((columnR + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
              bendRL += (Math.abs((columnLX + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
            }
          }
        }

        // Compute "crossings" of nodes on same layer.
        rightDestinationEdgeIter.reset();
        while (rightDestinationEdgeIter.next()) {
          const rde2 = rightDestinationEdgeIter.value as LayeredDigraphEdge;
          toVertex = rde2.toVertex as LayeredDigraphVertex;
          if (rde2.valid && (toVertex.layer === unfixedLayer)) {
            indexTo = 0;
            while (unfixedLayerNodes[indexTo] !== toVertex) indexTo++;
            // Depending on where the to node falls,
            //  we increase the appropriate crossing value.
            // The different values indicate how "strongly" that particular
            //  ordering is desired.
            if (indexTo === index) {
              layerCrossingLR += 1;
            }
          }
        }

        // Calculate the bend numbers for the left and right nodes,
        //  based on their barycenter values.
        // The idea is that if neither a shift to the left or the right reduces
        //  the actual number of bends, moving towards the barycenter will
        //  improve the centering and balance of the layout.
        let BbendLR = 0;
        let BbendRL = 0;
        const barycenterNegL = barycentersNeg[unfixedLayerNodes[index].index];
        const barycenterPosL = barycentersPos[unfixedLayerNodes[index].index];
        const barycenterNegR = barycentersNeg[unfixedLayerNodes[index + 1].index];
        const barycenterPosR = barycentersPos[unfixedLayerNodes[index + 1].index];

        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterNegL !== -1) {
          BbendLR += Math.abs(barycenterNegL - columnL);
          BbendRL += Math.abs(barycenterNegL - columnRX);
        }
        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterPosL !== -1) {
          BbendLR += Math.abs(barycenterPosL - columnL);
          BbendRL += Math.abs(barycenterPosL - columnRX);
        }
        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterNegR !== -1) {
          BbendLR += Math.abs(barycenterNegR - columnR);
          BbendRL += Math.abs(barycenterNegR - columnLX);
        }
        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterPosR !== -1) {
          BbendLR += Math.abs(barycenterPosR - columnR);
          BbendRL += Math.abs(barycenterPosR - columnLX);
        }
        // Peform an exchange if it improves the number of link crossings
        //  or the number of bends.
        if ((layerCrossingRL < layerCrossingLR - 0.5) ||
          ((layerCrossingRL === layerCrossingLR) && (crossingRL < crossingLR - 0.5)) ||
          ((layerCrossingRL === layerCrossingLR) && (crossingRL === crossingLR) && (bendRL < bendLR - 0.5)) ||
          ((layerCrossingRL === layerCrossingLR) && (crossingRL === crossingLR) && (bendRL === bendLR) && (BbendRL < BbendLR - 0.5))) {
          layerChange = true;
          exchange = true;
          unfixedLayerNodes[index].column = columnRX;
          unfixedLayerNodes[index + 1].column = columnLX;
          const vertex = unfixedLayerNodes[index];
          unfixedLayerNodes[index] = unfixedLayerNodes[index + 1];
          unfixedLayerNodes[index + 1] = vertex;
        }
      }
    }
    // Update the index for each node in the unfixedLayer.
    for (index = 0; index < num; index++) {
      unfixedLayerNodes[index].index = index;
    }
    // Return status of the unfixedLayer.
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    return layerChange;
  }

  // ********************************************************************************************
  // StraightenAndPack

  /**
   * @hidden
   * (Undocumented)
   * Adjusts the columns of nodes in the network to produce a layout which reduces
   * the number of bends and is tightly packed.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected straightenAndPack(): void {
    let layer = 0;
    let change = false;

    let expand = ((this._packOption & LayeredDigraphLayout.PackExpand) !== 0);
    const forceexpand = (this._packOption === LayeredDigraphLayout.PackAll);
    if (this.network.edges.count > 1000 && !forceexpand) expand = false;
    if (expand) {
      // In order to give the straightening methods more "room",
      //  widen the spaces between each node.
      // First, compute the maximum column used by each layer.
      // This will be used to center the nodes in each layer.
      // The idea is that centered nodes are more likely to align
      //  with their parents and children.
      const columns: Array<number> = [];
      let index = 0;
      for (index = 0; index <= this._maxLayer; index++) {
        columns[index] = 0;
      }
      let column = 0;
      const vertIter = this.network.vertexes.iterator;
      while (vertIter.next()) {
        const vertex = vertIter.value as LayeredDigraphVertex;
        layer = vertex.layer;
        column = vertex.column;
        const minColumnSpace = this.nodeMinColumnSpace(vertex, false);
        columns[layer] = Math.max(columns[layer], column + minColumnSpace);
      }

      // Next, modify the column of each node so that it is spaced and centered.
      const MULT = 8;
      vertIter.reset();
      while (vertIter.next()) {
        const vert = vertIter.value as LayeredDigraphVertex;
        layer = vert.layer;
        column = vert.column;
        vert.column = ((((this._maxColumn - columns[layer]) * MULT) >> 1) + column * MULT);
      }
      this._maxColumn *= MULT;
    }

    if ((this._packOption & LayeredDigraphLayout.PackStraighten) !== 0) {
      // Use the bendStraighten method, sweeping away from the maxIndexLayer,
      //  iterating until no changes occur.
      // The idea is that the layer with the most nodes will have the greatest
      //  influence on the remainder of the nodes.
      change = true;
      while (change) {
        change = false;
        for (layer = this._maxIndexLayer + 1; layer <= this._maxLayer; layer++) {
          change = this.bendStraighten(layer, 1) || change;
        }
        for (layer = this._maxIndexLayer - 1; layer >= 0; layer--) {
          change = this.bendStraighten(layer, -1) || change;
        }
        change = this.bendStraighten(this._maxIndexLayer, 0) || change;
      }
    }

    if ((this._packOption & LayeredDigraphLayout.PackMedian) !== 0) {
      // Use the medianStraighten method, sweeping away from the maxIndexLayer.
      // The idea is that the layer with the most nodes will have the greatest
      //  influence on the remainder of the nodes.
      for (layer = this._maxIndexLayer + 1; layer <= this._maxLayer; layer++) {
        this.medianStraighten(layer, 1);
      }
      for (layer = this._maxIndexLayer - 1; layer >= 0; layer--) {
        this.medianStraighten(layer, -1);
      }
      this.medianStraighten(this._maxIndexLayer, 0);
    }

    // Iterate compactification and straightening until bends are not reduced.
    // Since componentPack and bendStraighten both are guaranteed not to
    //  increase the number of "bends", this loop will terminate,
    //  usually only after two or three iterations.

    if (expand) {
      // Use the componentPack method to tightly pack the nodes.
      this.componentPack(-1);
      this.componentPack(1);
    }

    if ((this._packOption & LayeredDigraphLayout.PackStraighten) !== 0) {
      // Use the bendStraighten method, sweeping away from the maxIndexLayer,
      //  iterating until no changes occur.
      // The idea is that the layer with the most nodes will have the greatest
      //  influence on the remainder of the nodes.
      change = true;
      while (change) {
        change = false;
        change = this.bendStraighten(this._maxIndexLayer, 0) || change;
        for (layer = this._maxIndexLayer + 1; layer <= this._maxLayer; layer++) {
          change = this.bendStraighten(layer, 0) || change;
        }
        for (layer = this._maxIndexLayer - 1; layer >= 0; layer--) {
          change = this.bendStraighten(layer, 0) || change;
        }
      }
    }
  }

  // ********************************************************************************************
  // Straighten methods

  /**
   * @hidden @internal
   * Adjusts the columns of nodes within the unfixedLayer to reduce the number of "bends"
   * between the unfixedLayer and its adjacent layers.  The direction argument indicates
   * which of the adjacent layers should be taken into consideration when reducing the number
   * of bends.
   * The "weighted bend" between a node U and a node V connected by link L is calculated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset)) * LinkStraightenWeight(L)
   * The LinkStraightenWeight attempts to give higher priority to links between
   * "artificial" nodes; i.e., long links in the final layout will be straighter.
   * The idea is to iterate the shiftBendStraighten and adjacentExchangeBendStraighten
   * methods until no improvements are made.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {boolean} Returns true if an improvement was made.
   */
  private bendStraighten(unfixedLayer: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'bendStraighten:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'bendStraighten:direction');
    }
    let layerChange = false;
    while (this.shiftBendStraighten(unfixedLayer, direction)) {
      layerChange = true;
    }
    return layerChange;
  }

  /**
   * @hidden @internal
   * Adjusts the columns of nodes within the unfixedLayer to reduce the number of "bends"
   * between the unfixedLayer and its adjacent layers.  The direction argument indicates
   * which of the adjacent layers should be taken into consideration when reducing the number
   * of bends.
   * The "weighted bend" between a node U and a node V connected by link L is calculated by
   * abs((U.column + L.portFromColOffset) - (V.column + L.portToColOffset)) * linkStraightenWeight(L)
   * The linkStraightenWeight attempts to give higher priority to links between
   * "artificial" nodes; i.e., long links in the final layout will be straighter.
   * The idea is shift nodes to the left and to the right to reduce the bends
   * ensuring that no two nodes have overlapping "allocations" of columns.
   * Return true if some change was made to the layer.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {boolean} Returns true if some change was made to the layer and false otherwise.
   */
  private shiftBendStraighten(unfixedLayer: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'shiftbendStraighten:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'shiftbendStraighten:direction');
    }
    let index = 0;
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index.
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    // Initialize the barycenters from the negative direction.
    const barycentersNeg = this.barycenters(unfixedLayer, -1);
    // If direction > 0, the set barycenters from the negative direction to -1.
    if (direction > 0) {
      for (index = 0; index < num; index++) {
        barycentersNeg[index] = -1;
      }
    }
    // Initialize the barycenters from the positive direction.
    const barycentersPos = this.barycenters(unfixedLayer, 1);
    // If direction < 0, the set barycenters from the positive direction to -1.
    if (direction < 0) {
      for (index = 0; index < num; index++) {
        barycentersPos[index] = -1;
      }
    }

    let layerChange = false;
    let shift = true;
    while (shift) {
      shift = false;

      for (index = 0; index < num; index++) {
        // Current column of unfixedLayerNodes[index].
        const columnC = unfixedLayerNodes[index].column;
        // Minimum column space for unfixedLayerNodes[index].
        const minColumnSpaceLeft = this.nodeMinColumnSpace(unfixedLayerNodes[index], true);
        const minColumnSpaceRight = this.nodeMinColumnSpace(unfixedLayerNodes[index], false);

        // Calculate the column of a shift to the left.
        // If there are no nodes to the left, or there is enough space between
        //  unfixedLayerNodes[index] and unfixedLayerNodes[index-1],
        //  then columnL = columnC - 1.
        // Otherwise, columnL = columnC and no shift will be taken
        //  (since the current bends will equal the bends after a shift into columnL).
        let columnL = 0;
        if ((index - 1 < 0) ||
          ((columnC - unfixedLayerNodes[index - 1].column - 1) > (minColumnSpaceLeft + this.nodeMinColumnSpace(unfixedLayerNodes[index - 1], false)))) {
          columnL = columnC - 1;
        } else {
          columnL = columnC;
        }

        // Calculate the column of a shift to the right.
        // If there are no nodes to the right, or there is enough space between
        //  unfixedLayerNodes[index] and unfixedLayerNodes[index+1],
        //  then columnR = columnC + 1.
        // Otherwise, columnR = columnC and no shift will be taken
        //  (since the current bends will equal the bends after a shift into columnR).
        let columnR = 0;
        if ((index + 1 >= num) ||
          ((unfixedLayerNodes[index + 1].column - columnC - 1) > (minColumnSpaceRight + this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true)))) {
          columnR = columnC + 1;
        } else {
          columnR = columnC;
        }

        // Calculate the bend numbers for the current column, left column, and right column.
        let bendC = 0.0;
        let bendL = 0.0;
        let bendR = 0.0;

        let weight = 0.0;
        let portFromColOffset = 0;
        let portToColOffset = 0;

        // Calculate the predecessor link list, based on direction.
        if (direction <= 0) {
          const sourceEdgeIter = unfixedLayerNodes[index].sourceEdges.iterator;
          while (sourceEdgeIter.next()) {
            const se = sourceEdgeIter.value as LayeredDigraphEdge;
            const fromVertex = se.fromVertex as LayeredDigraphVertex;
            if (se.valid && (fromVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(se);
              portFromColOffset = se.portFromColOffset;
              portToColOffset = se.portToColOffset;
              const fromCol = fromVertex.column;
              bendC += (Math.abs((columnC + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
              bendL += (Math.abs((columnL + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
              bendR += (Math.abs((columnR + portToColOffset) - (fromCol + portFromColOffset)) + 1) * weight;
            }
          }
        }

        // Calculate the successor link list, based on direction.
        if (direction >= 0) {
          const destinationEdgeIter = unfixedLayerNodes[index].destinationEdges.iterator;
          while (destinationEdgeIter.next()) {
            const de = destinationEdgeIter.value as LayeredDigraphEdge;
            const toVertex = de.toVertex as LayeredDigraphVertex;
            if (de.valid && (toVertex.layer !== unfixedLayer)) {
              // Update the bends.
              weight = this.linkStraightenWeight(de);
              portFromColOffset = de.portFromColOffset;
              portToColOffset = de.portToColOffset;
              const toCol = toVertex.column;
              bendC += (Math.abs((columnC + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
              bendL += (Math.abs((columnL + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
              bendR += (Math.abs((columnR + portFromColOffset) - (toCol + portToColOffset)) + 1) * weight;
            }
          }
        }

        // Calculate the bend numbers for the current column, left column,
        //  and right column, based on the barycenter values.
        // The idea is that if neither a shift to the left or the right reduces
        //  the actual number of bends, moving towards the barycenter will
        //  improve the centering and balance of the layout.
        let BbendC = 0;
        let BbendL = 0;
        let BbendR = 0;
        const barycenterNegC = barycentersNeg[unfixedLayerNodes[index].index];
        const barycenterPosC = barycentersPos[unfixedLayerNodes[index].index];
        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterNegC !== -1) {
          BbendC += Math.abs(barycenterNegC - columnC);
          BbendL += Math.abs(barycenterNegC - columnL);
          BbendR += Math.abs(barycenterNegC - columnR);
        }
        // A barycenter of -1 indicates that the barycenter does not exist
        //  for that direction; hence no bends exist.
        if (barycenterPosC !== -1) {
          BbendC += Math.abs(barycenterPosC - columnC);
          BbendL += Math.abs(barycenterPosC - columnL);
          BbendR += Math.abs(barycenterPosC - columnR);
        }
        // Peform a shift if it improves the bends.
        if ((bendL < bendC) || ((bendL === bendC) && (BbendL < BbendC))) {
          layerChange = true;
          shift = true;
          unfixedLayerNodes[index].column = columnL;
        } else if ((bendR < bendC) || ((bendR === bendC) && (BbendR < BbendC))) {
          layerChange = true;
          shift = true;
          unfixedLayerNodes[index].column = columnR;
        }
      }
    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    // Some nodes may have been moved to the left of column 0 or to the right
    //  of column maxColumn, so normalize the nodes; i.e., the leftmost column will
    //  be column 0 and maxColumn will be updated appropriately.
    this.normalize();

    // Return status of the unfixedLayer.
    return layerChange;
  }

  /**
   * @hidden @internal
   * Adjusts the columns of nodes within the unfixedLayer in order to move nodes towards
   * their median columns.  The direction argument indicates which of the adjacent layers
   * should be taken into consideration when computing the median column.
   * The idea is shift nodes to the left and to the right to move nodes towards their
   * median columns, ensuring that no two nodes have overlapping "allocations" of columns.
   * Returns true if some change was made to the layer.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {number} direction
   * @return {boolean} Returns true if some change was made to the layer and false otherwise.
   */
  private medianStraighten(unfixedLayer: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(unfixedLayer, LayeredDigraphLayout, 'medianStraighten:unfixedLayer');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'medianStraighten:direction');
    }
    let index = 0;
    // Fill unfixedLayerNodes with nodes in the unfixedLayer, ordered by index
    const unfixedLayerNodes = this._getCachedNodeArrayList(unfixedLayer);
    const num = this._indices[unfixedLayer];

    // Calculate the median value of each node and convert to integer.
    const dmedians = this.medians(unfixedLayer, direction);
    const medians: Array<number> = [];
    for (index = 0; index < num; index++) {
      medians[index] = dmedians[index];
    }
    let layerChange = false;
    let change = true;
    while (change) {
      change = false;
      for (index = 0; index < num; index++) {
        // The current column of unfixedLayerNodes[index].
        const column = unfixedLayerNodes[index].column;
        const minColumnSpaceLeft = this.nodeMinColumnSpace(unfixedLayerNodes[index], true);
        const minColumnSpaceRight = this.nodeMinColumnSpace(unfixedLayerNodes[index], false);
        // The new column of unfixedLayerNodes[index].
        let ncolumn = 0;
        let rcolumn = 0;
        let lcolumn = 0;
        let nrcolumn = 0;
        let nlcolumn = 0;
        // A median value of -1 indicates that the node does not
        //  have a defined median (i.e., the node has no neighbors
        //  in the appropriate adjacent layer).
        // In that case, treat these nodes specially, since they can be
        //  shifted "out of the way" of nodes that have a defined median.
        if (medians[index] === -1) {
          // If index === 0 and index === num - 1,
          //  then there is only one node on the layer
          //  and the node does not move.
          if ((index === 0) && (index === num - 1)) {
            ncolumn = column;
            // If index === 0,
            //  then check if the node is "adjacent" (in the sense of nodeMinColumnSpace values)
            //  to its rightmost neighbor.  If so, move the node to the left.
            // The idea is that the rightmost neighbor may be "moving left",
            //  so move the node out of the way.
          } else if (index === 0) {
            rcolumn = unfixedLayerNodes[index + 1].column;
            if (rcolumn - column === (minColumnSpaceRight + this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true))) {
              ncolumn = column - 1;
            } else {
              ncolumn = column;
            }

            // If index === num - 1,
            //  then check if the node is "adjacent" (in the sense of nodeMinColumnSpace values)
            //  to its leftmost neighbor.  If so, move the node to the right.
            // The idea is that the leftmost neighbor may be "moving right",
            //  so move the node out of the way.
          } else if (index === num - 1) {
            lcolumn = unfixedLayerNodes[index - 1].column;
            if (column - lcolumn === (minColumnSpaceLeft + this.nodeMinColumnSpace(unfixedLayerNodes[index - 1], false))) {
              ncolumn = column + 1;
            } else {
              ncolumn = column;
            }

            // Otherwise, position the node between its left and right neighbors.
          } else {
            lcolumn = unfixedLayerNodes[index - 1].column;
            nlcolumn = lcolumn + this.nodeMinColumnSpace(unfixedLayerNodes[index - 1], false) + minColumnSpaceLeft + 1;
            rcolumn = unfixedLayerNodes[index + 1].column;
            nrcolumn = rcolumn - this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true) - minColumnSpaceRight - 1;
            ncolumn = ((nlcolumn + nrcolumn) / 2) | 0;
          }

          // Otherwise, the node has a defined median.
        } else {
          // If index === 0 and index === num - 1,
          //  then there is only one node on the layer
          //  and the node can be moved to its median position.
          if ((index === 0) && (index === num - 1)) {
            ncolumn = medians[index];
            // If index === 0,
            //  then move the node to the position closest to its median that does
            //  not overlap with its rightmost neighbor.
          } else if (index === 0) {
            rcolumn = unfixedLayerNodes[index + 1].column;
            nrcolumn = rcolumn - this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true) - minColumnSpaceRight - 1;
            ncolumn = Math.min(medians[index], nrcolumn);
            // If index === num - 1,
            //  then move the node to the position closest to its median that does
            //  not overlap with its leftmost neighbor.
          } else if (index === num - 1) {
            lcolumn = unfixedLayerNodes[index - 1].column;
            nlcolumn = lcolumn + this.nodeMinColumnSpace(unfixedLayerNodes[index - 1], false) + minColumnSpaceLeft + 1;
            ncolumn = Math.max(medians[index], nlcolumn);
            // Otherwise, move the node to the position closest to its median that does
            //  not overlap with either its left or right neighbors.
          } else {
            lcolumn = unfixedLayerNodes[index - 1].column;
            nlcolumn = lcolumn + this.nodeMinColumnSpace(unfixedLayerNodes[index - 1], false) + minColumnSpaceLeft + 1;
            rcolumn = unfixedLayerNodes[index + 1].column;
            nrcolumn = rcolumn - this.nodeMinColumnSpace(unfixedLayerNodes[index + 1], true) - minColumnSpaceRight - 1;
            if ((nlcolumn < medians[index]) && (medians[index] < nrcolumn)) {
              ncolumn = medians[index];
            } else if (nlcolumn >= medians[index]) {
              ncolumn = nlcolumn;
            } else if (nrcolumn <= medians[index]) {
              ncolumn = nrcolumn;
            }
          }
        }

        // If the node changed position, make the appropriate updates.
        if (ncolumn !== column) {
          layerChange = true;
          change = true;
          unfixedLayerNodes[index].column = ncolumn;
        }
      }
    }
    this._freeCachedNodeArrayList(unfixedLayer, unfixedLayerNodes);
    // Some nodes may have been moved to the left of column 0 or to the right
    //  of column maxColumn, so normalize the nodes; i.e., the leftmost column will
    //  be column 0 and maxColumn will be updated appropriately.
    this.normalize();
    // Return status of the unfixedLayer.
    return layerChange;
  }

  // ********************************************************************************************
  // Pack methods

  /**
   * @hidden @internal
   * Adjusts the columns of nodes in the network to produce a layout which is tightly packed.
   * The idea is that columns which are "un-allocated" through all layers can be eliminated
   * and the nodes can be shifted into that space.
   * @this {LayeredDigraphLayout}
   */
  private pack(): void {
    // Sweep the network from column 0 to column maxColumn, eliminating empty columns.
    for (let column = 0; column <= this._maxColumn; column++) {
      while (this.packAux(column, 1)) {
        // empty
      }
    }
    // Some columns may have been eliminated, so normalize the nodes;
    //  i.e., the leftmost column will be column 0 and maxColumn will
    //  be updated appropriately.
    this.normalize();
  }

  /**
   * @hidden @internal
   * Attempts to remove the argument column by shifting columns into from the argument direction.
   * direction > 0 -- columns > argument column are shifted
   * direction < 0 -- columns < argument column are shifted
   * Returns true if the argument column was removed.
   * @this {LayeredDigraphLayout}
   * @param {number} column
   * @param {number} direction
   * @return {boolean} Returns true if the argument column was removed and false otherwise.
   */
  private packAux(column: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(column, LayeredDigraphLayout, 'packAux:column');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'packAux:direction');
    }
    // Scan all nodes in the network.
    // A shift is possible if no node has been "allocated" the argument column.
    let doshift = true;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const minColumnSpaceLeft = this.nodeMinColumnSpace(vertex, true);
      const minColumnSpaceRight = this.nodeMinColumnSpace(vertex, false);
      if ((vertex.column - minColumnSpaceLeft <= column) && (vertex.column + minColumnSpaceRight >= column)) {
        doshift = false;
        break;
      }
    }

    // If no node has been "allocated" the argument column,
    //  shift nodes into the eliminated column from the appropriate direction.
    let change = false;
    if (doshift) {
      if (direction > 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const v = vertIter.value as LayeredDigraphVertex;
          if (v.column > column) {
            v.column -= 1;
            change = true;
          }
        }
      }
      if (direction < 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const v2 = vertIter.value as LayeredDigraphVertex;
          if (v2.column < column) {
            v2.column += 1;
            change = true;
          }
        }
      }
    }

    // Return status of the column.
    return change;
  }

  /**
   * @hidden @internal
   * Adjusts the columns of nodes in the network to produce a layout which is tightly packed.
   * The idea is that two adjacent columns can be "merged" if each layer has at most one of
   * the two columns "allocated" to a node.
   * @this {LayeredDigraphLayout}
   */
  private tightPack(): void {
    // Begin by using the pack method.
    // This (faster) method eliminates some columns and reduces the number of columns
    //  that need to be examined.
    this.pack();
    // Sweep the network from column 0 to column maxColumn, merging columns.
    for (let column = 0; column < this._maxColumn; column++) {
      while (this.tightPackAux(column, 1)) {
        // empty
      }
    }
    // Some columns may have been eliminated, so normalize the nodes;
    //  i.e., the leftmost column will be column 0 and maxColumn will
    //  be updated appropriately.
    this.normalize();
  }

  /**
   * @hidden @internal
   * Attempts to augment the argument column by merging columns into from the argument direction.
   * direction > 0 -- columns > the argument column are shifted
   * direction < 0 -- columns < the argument column are shifted
   * Returns true if the argument column was changed.
   * @this {LayeredDigraphLayout}
   * @param {number} column
   * @param {number} direction
   * @return {boolean} Returns true if the argument column was changed and false otherwise.
   */
  private tightPackAux(column: number, direction: number): boolean {
    if (Debug) {
      Util.checkRealNumber(column, LayeredDigraphLayout, 'tightPackAux:column');
      Util.checkRealNumber(direction, LayeredDigraphLayout, 'tightPackAux:direction');
    }
    // Set the next column to the appropriate value based on direction.
    let nextcolumn = column;
    if (direction > 0) {
      nextcolumn = column + 1;
    }
    if (direction < 0) {
      nextcolumn = column - 1;
    }

    // The curCol and nextCol arrays indicate which layers in the current column and
    //  the next column have been "allocated" to nodes.
    // Initially, both the curCol and nextCol are empty.
    let index = 0;
    const curCol: Array<boolean> = [];
    const nextCol: Array<boolean> = [];
    for (index = 0; index <= this._maxLayer; index++) {
      curCol[index] = false;
      nextCol[index] = false;
    }

    // Scan all nodes in the network.
    // If a node has been "allocated" either the current column or the next column,
    //  update the appropriate entry in curCol and nextCol.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const leftColumn = vertex.column - this.nodeMinColumnSpace(vertex, true);
      const rightColumn = vertex.column + this.nodeMinColumnSpace(vertex, false);
      if ((leftColumn <= column) && (rightColumn >= column)) {
        curCol[vertex.layer] = true;
      }
      if ((leftColumn <= nextcolumn) && (rightColumn >= nextcolumn)) {
        nextCol[vertex.layer] = true;
      }
    }

    // A shift is possible if each layer has at most one column "allocated" between
    //  the current column and the next column. A logical NAND verifies that the columns
    //  can be merged at a particular layer.
    let doshift = true;
    let change = false;
    for (index = 0; index <= this._maxLayer; index++) {
      doshift = doshift && (!(curCol[index] && nextCol[index]));
    }

    // If each layer has at most one column "allocated" between
    //  the current column and the next column shift nodes into the merged
    //  column from the appropriate direction.
    if (doshift) {
      if (direction > 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const v1 = vertIter.value as LayeredDigraphVertex;
          if (v1.column > column) {
            v1.column -= 1;
            change = true;
          }
        }
      }
      if (direction < 0) {
        vertIter.reset();
        while (vertIter.next()) {
          const v2 = vertIter.value as LayeredDigraphVertex;
          if (v2.column < column) {
            v2.column += 1;
            change = true;
          }
        }
      }
    }

    // Return status of the column.
    return change;
  }

  /**
   * @hidden @internal
   * Adjusts the columns of nodes in the network to produce a layout which is tightly packed.
   * The idea is that the network can be fragmented from a given column in the following way:
   * all nodes "behind" the column are placed into a single component, and the remainder of
   * the network is divided into connected components.  Each of these new components can be
   * examined, and those that can be merged with the given column do so.
   * @this {LayeredDigraphLayout}
   * @param {number} direction
   */
  private componentPack(direction: number): void {
    if (Debug) Util.checkRealNumber(direction, LayeredDigraphLayout, 'componentPack:direction');
    // Begin by using the tightPack method.
    // This (faster) method eliminates some columns and reduces the number of columns
    //  that need to be examined.
    this.tightPack();

    let column = 0;
    let bestlayout: Array<number>;
    let bestbends = 0.0;
    let lastbends = 0.0;
    let bends = 0.0;
    if (direction > 0) {
      // Sweep the network from column 0 to column maxColumn, merging components.
      for (column = 0; column <= this._maxColumn; column++) {
        // Continue merging components into column
        //  while it reduces the number of "bends".
        bestlayout = this.saveLayout();
        bestbends = this.countBends(true);
        lastbends = bestbends + 1;
        while (bestbends < lastbends) {
          lastbends = bestbends;
          this.componentPackAux(column, 1);
          bends = this.countBends(true);
          if (bends > bestbends) {
            this.restoreLayout(bestlayout);
          } else if (bends < bestbends) {
            bestbends = bends;
            bestlayout = this.saveLayout();
          }
        }
      }
    }
    if (direction < 0) {
      // Sweep the network from column maxColumn to column 0, merging components.
      for (column = this._maxColumn; column >= 0; column--) {
        // Continue merging components into column
        //  while it reduces the number of "bends".
        bestlayout = this.saveLayout();
        bestbends = this.countBends(true);
        lastbends = bestbends + 1;
        while (bestbends < lastbends) {
          lastbends = bestbends;
          this.componentPackAux(column, -1);
          bends = this.countBends(true);
          if (bends > bestbends) {
            this.restoreLayout(bestlayout);
          } else if (bends < bestbends) {
            bestbends = bends;
            bestlayout = this.saveLayout();
          }
        }
      }
    }

    // Some columns may have been eliminated, so normalize the nodes;
    //  i.e., the leftmost column will be column 0 and maxColumn will
    //  be updated appropriately.
    this.normalize();
  }

  /**
   * @hidden @internal
   * Attempts to augment the argument column by merging components into from the argument direction.
   * direction > 0 -- columns > the argument column are shifted
   * direction < 0 -- columns < the argument column are shifted
   * Returns true if the argument column was changed.
   * @this {LayeredDigraphLayout}
   * @param {number} column
   * @param {number} direction
   * @return {boolean} Returns true if the argument column was changed and false otherwise.
   */
  private componentPackAux(column: number, direction: number): boolean {
    // Every node is initially unset -- i.e., it hasn't been assigned to
    //  a component.

    // Add nodes "behind" the argument column to component 0.
    this._myComponent = 0;
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      v.component = -1;
    }
    if (direction > 0) {
      vertIter.reset();
      while (vertIter.next()) {
        const v1 = vertIter.value as LayeredDigraphVertex;
        if (v1.column - this.nodeMinColumnSpace(v1, true) <= column) {
          v1.component = this._myComponent;
        }
      }
    }
    if (direction < 0) {
      vertIter.reset();
      while (vertIter.next()) {
        const v2 = vertIter.value as LayeredDigraphVertex;
        if (v2.column + this.nodeMinColumnSpace(v2, false) >= column) {
          v2.component = this._myComponent;
        }
      }
    }

    // The remaining nodes are placed into successive "compact" components.
    this._myComponent++;
    vertIter.reset();
    while (vertIter.next()) {
      const v3 = vertIter.value as LayeredDigraphVertex;
      if (v3.component === -1) {
        this.componentUnset(v3, this._myComponent, -1, true, true);
        this._myComponent++;
      }
    }

    // Create the interference matrix.
    // If interference[component1 * component + component2] is true,
    //  then component1 must be shifted in order for component2 to be shifted.
    // This allows components that must be moved together to be detected.
    let index = 0;
    const interference: Array<boolean> = [];
    for (index = 0; index < this._myComponent * this._myComponent; index++) {
      interference[index] = false;
    }

    // Fill in the interference matrix.
    // Begin by filling in a matrix representing the current distribution
    //  of components among layers and columns.
    const components: Array<number> = [];
    for (index = 0; index < (this._maxLayer + 1) * (this._maxColumn + 1); index++) {
      components[index] = -1;
    }

    // Scan all nodes in the network.
    // If a node has been "allocated" to some columns,
    //  update the appropriate entries in components.
    vertIter.reset();
    while (vertIter.next()) {
      const v4 = vertIter.value as LayeredDigraphVertex;
      const layer = v4.layer;
      const startCol = Math.max(0, v4.column - this.nodeMinColumnSpace(v4, true));
      const endCol = Math.min(this._maxColumn, v4.column + this.nodeMinColumnSpace(v4, false));
      for (let col = startCol; col <= endCol; col++) {
        components[layer * (this._maxColumn + 1) + col] = v4.component;
      }
    }

    // A component interferes with another component if two different components
    //  in the current column and the "next" column are allocated to the same layer.
    // The "next" column is dependent on the direction.
    for (let layerIndex = 0; layerIndex <= this._maxLayer; layerIndex++) {
      // In the positive direction, the "next" column is col + 1.
      if (direction > 0) {
        for (let col = 0; col < this._maxColumn; col++) {
          if ((components[layerIndex * (this._maxColumn + 1) + col] !== -1) &&
            (components[layerIndex * (this._maxColumn + 1) + col + 1] !== -1) &&
            (components[layerIndex * (this._maxColumn + 1) + col] !== components[layerIndex * (this._maxColumn + 1) + col + 1])) {
            interference[components[layerIndex * (this._maxColumn + 1) + col] * this._myComponent + components[layerIndex * (this._maxColumn + 1) + col + 1]] = true;
          }
        }
      }
      // In the negative direction, the "next" column is col - 1.
      if (direction < 0) {
        for (let col = this._maxColumn; col > 0; col--) {
          if ((components[layerIndex * (this._maxColumn + 1) + col] !== -1) &&
            (components[layerIndex * (this._maxColumn + 1) + col - 1] !== -1) &&
            (components[layerIndex * (this._maxColumn + 1) + col] !== components[layerIndex * (this._maxColumn + 1) + col - 1])) {
            interference[components[layerIndex * (this._maxColumn + 1) + col] * this._myComponent + components[layerIndex * (this._maxColumn + 1) + col - 1]] = true;
          }
        }
      }
    }

    // Determine which components can be shifted.
    // Initially, all components can be shifted.
    // Using the interference matrix, dependencies are detected and marked.
    // In particular, component 0 can not be shifted,
    //  and it dictates which other components cannot be shifted.
    const doshift: Array<boolean> = [];
    for (index = 0; index < this._myComponent; index++) {
      doshift[index] = true;
    }
    // Use a queue to keep track of which components cannot be shifted.
    const Queue: Array<number> = [];
    // Initially, only component 0 cannot be shifted.
    Queue.push(0);
    let head = 0;
    while (Queue.length !== 0) {
      // Dequeue.
      head = Queue[Queue.length - 1];
      Queue.pop();
      // If this is the first time encountering this component,
      //  mark it as unshiftable and queue the components it interferes with.
      if (doshift[head]) {
        doshift[head] = false;
        for (index = 0; index < this._myComponent; index++) {
          if (interference[head * this._myComponent + index]) {
            Queue.splice(0, 0, index);
          }
        }
      }
    }

    // Components which can be shifted are moved in the appropriate direction.
    let change = false;
    if (direction > 0) {
      vertIter.reset();
      while (vertIter.next()) {
        const v5 = vertIter.value as LayeredDigraphVertex;
        if (doshift[v5.component]) {
          v5.column -= 1;
          change = true;
        }
      }
    }
    if (direction < 0) {
      vertIter.reset();
      while (vertIter.next()) {
        const v6 = vertIter.value as LayeredDigraphVertex;
        if (doshift[v6.component]) {
          v6.column += 1;
          change = true;
        }
      }
    }

    // Return status of the column.
    return change;
  }

  // ********************************************************************************************
  // Layout Nodes and Links

  /**
   * Set the fromSpot and toSpot on each Link, position each Node according
   * to the vertex position, and then position/route the Links.
   *
   * This calls the {@link #commitNodes} and {@link #commitLinks} methods, the latter only if {@link #isRouting} is true.
   * You should not call this method -- it is a "protected virtual" method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @this {LayeredDigraphLayout}
   */
  protected commitLayout(): void {
    this._setPortSpotsAll();
    this.commitNodes();
    this.commitLayersInternal();
    if (this.isRouting) this.commitLinks();
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   */
  private _setPortSpotsAll(): void {
    if (!this.setsPortSpots) return;
    const fromspot = this._getDefaultSpot(true);
    const tospot = this._getDefaultSpot(false);
    const edgeIter = this.network.edges.iterator;
    while (edgeIter.next()) {
      const edge = edgeIter.value;
      const link = edge.link;
      if (link !== null) {
        link.fromSpot = fromspot;
        link.toSpot = tospot;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {boolean} from
   * @return {Spot}
   */
  private _getDefaultSpot(from: boolean): Spot {
    if (this._direction === 270) {
      return (from ? Spot.MiddleTop : Spot.MiddleBottom);
    } else if (this._direction === 90) {
      return (from ? Spot.MiddleBottom : Spot.MiddleTop);
    } else if (this._direction === 180) {
      return (from ? Spot.MiddleLeft : Spot.MiddleRight);
    } else {
      return (from ? Spot.MiddleRight : Spot.MiddleLeft);
    }
  }

  /**
   * Commit the position of all nodes.
   *
   * This is called by {@link #commitLayout}.
   * See also {@link #commitLinks}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected commitNodes(): void {
    // The layerSpace array records the maximum layer spacing needed
    //  by a node in each layer, in document units.

    this._layerSpaceLeft = [];
    this._layerSpaceRight = [];
    this._layerPositions = [];
    this._layerSpaces = [];

    for (let i = 0; i <= this._maxLayer; i++) {
      this._layerSpaceLeft[i] = 0;
      this._layerSpaceRight[i] = 0;
      this._layerPositions[i] = 0;
      this._layerSpaces[i] = 0;
    }

    // Scan all nodes in the network, updating the layerSpace array.
    const vertIter = this.network.vertexes.iterator;
    while (vertIter.next()) {
      const v = vertIter.value as LayeredDigraphVertex;
      const layer = v.layer;
      this._layerSpaceLeft[layer] = Math.max(this._layerSpaceLeft[layer], this.nodeMinLayerSpace(v, true));
      this._layerSpaceRight[layer] = Math.max(this._layerSpaceRight[layer], this.nodeMinLayerSpace(v, false));
    }

    // The layerPositions array records the position in document units for each layer.
    // Each layer receives layerSpace[layer] space "above" and "below" it,
    // to accomodate the largest node in the layer, plus LayerSpacing between layers
    let layerPosition = 0;
    const space = this._layerSpacing;
    for (let layerInd = 0; layerInd <= this._maxLayer; layerInd++) {
      let s = space;
      // don't reserve LayerSpacing around layers consisting entirely of artificial nodes
      if (this._layerSpaceLeft[layerInd] + this._layerSpaceRight[layerInd] <= 0) s = 0;
      if (layerInd > 0) layerPosition += s / 2;  // avoid a margin: don't add half a LayerSpacing for first layer
      if (this._direction === 90 || this._direction === 0) {
        layerPosition += this._layerSpaceRight[layerInd];
        this._layerPositions[layerInd] = layerPosition;  // middle of layer
        layerPosition += this._layerSpaceLeft[layerInd];
      } else {
        layerPosition += this._layerSpaceLeft[layerInd];
        this._layerPositions[layerInd] = layerPosition;  // middle of layer
        layerPosition += this._layerSpaceRight[layerInd];
      }
      if (layerInd < this._maxLayer) layerPosition += s / 2;  // space between layers, except when beyond last layer
      this._layerSpaces[layerInd] = layerPosition;  // middle of space between layers
    }
    const totalDepth = layerPosition;  // this is now the total

    // position everything at arrangementOrigin
    const orig = this.arrangementOrigin;

    // convert layerPositions and layerSpaces arrays to use actual document coordinates, not just document units
    for (let layerInd = 0; layerInd <= this._maxLayer; layerInd++) {
      if (this._direction === 270) {
        this._layerPositions[layerInd] = orig.y + this._layerPositions[layerInd];
      } else if (this._direction === 90) {
        this._layerPositions[layerInd] = orig.y + totalDepth - this._layerPositions[layerInd];
        this._layerSpaces[layerInd] = totalDepth - this._layerSpaces[layerInd];
      } else if (this._direction === 180) {
        this._layerPositions[layerInd] = orig.x + this._layerPositions[layerInd];
      } else {
        this._layerPositions[layerInd] = orig.x + totalDepth - this._layerPositions[layerInd];
        this._layerSpaces[layerInd] = totalDepth - this._layerSpaces[layerInd];
      }
    }

    // Scan all nodes in the network (both "real" and "artificial"),
    //  and set their locations.
    vertIter.reset();
    let smallestX = Infinity;
    let smallestY = Infinity;
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      const lay = vertex.layer;
      const column = vertex.column | 0;
      // Place the center of the node depending on the direction of the layout.
      let x = 0.0;
      let y = 0.0;
      if (this._direction === 270 || this._direction === 90) {
        x = orig.x + this._columnSpacing * column;
        y = this._layerPositions[lay];
      } else {
        x = this._layerPositions[lay];
        y = orig.y + this._columnSpacing * column;
      }
      // Set the center position, update the position in the network,
      //  and commit the position to the physical location of the node.
      vertex.centerX = x;
      vertex.centerY = y;
      smallestX = Math.min(vertex.x, smallestX);
      smallestY = Math.min(vertex.y, smallestY);
    }

    // shift everything to account for arrangmentOrigin not lining up with columns,
    // depending on the size of the nodes, but only for layouts within a Group
    const offX = orig.x - smallestX;
    const offY = orig.y - smallestY;
    this._arrangementOffset = new Point(offX, offY);  // remember for calling commitLayers
    vertIter.reset();
    while (vertIter.next()) {
      const vertex = vertIter.value as LayeredDigraphVertex;
      vertex.x += offX;
      vertex.y += offY;
      vertex.commit();
    }
  }

  /**
   * @hidden @internal
   * Call {@link #commitLayers}, but only if all vertexes will be completely aligned in layers across the whole tree.
   * @this {LayeredDigraphLayout}
   */
  private commitLayersInternal(): void {
    let totalDepth = 0;
    const space = this._layerSpacing;
    for (let layerInd = 0; layerInd <= this._maxLayer; layerInd++) {
      totalDepth += this._layerSpaceLeft[layerInd];
      totalDepth += this._layerSpaceRight[layerInd];
    }
    totalDepth += this._maxLayer * space;

    // record where the layers are
    const layerRects: Array<Rect> = [];
    const columnsBreadth = this._columnSpacing * this._maxColumn;
    for (let i = this.maxLayer; i >= 0; i--) {
      if (this._direction === 270) {
        if (i === 0) {
          layerRects.push(new Rect(0, 0, columnsBreadth, Math.abs(this._layerSpaces[0])));
        } else {
          layerRects.push(new Rect(0, this._layerSpaces[i - 1], columnsBreadth, Math.abs(this._layerSpaces[i - 1] - this._layerSpaces[i])));
        }
      } else if (this._direction === 90) {
        if (i === 0) {
          layerRects.push(new Rect(0, this._layerSpaces[0], columnsBreadth, Math.abs(this._layerSpaces[0] - totalDepth)));
        } else {
          layerRects.push(new Rect(0, this._layerSpaces[i], columnsBreadth, Math.abs(this._layerSpaces[i - 1] - this._layerSpaces[i])));
        }
      } else if (this._direction === 180) {
        if (i === 0) {
          layerRects.push(new Rect(0, 0, Math.abs(this._layerSpaces[0]), columnsBreadth));
        } else {
          layerRects.push(new Rect(this._layerSpaces[i - 1], 0, Math.abs(this._layerSpaces[i - 1] - this._layerSpaces[i]), columnsBreadth));
        }
      } else {
        if (i === 0) {
          layerRects.push(new Rect(this._layerSpaces[0], 0, Math.abs(this._layerSpaces[0] - totalDepth), columnsBreadth));
        } else {
          layerRects.push(new Rect(this._layerSpaces[i], 0, Math.abs(this._layerSpaces[i - 1] - this._layerSpaces[i]), columnsBreadth));
        }
      }
    }
    this.commitLayers(layerRects, this._arrangementOffset);
  }

  /**
   * This overridable method is called by {@link #commitLayout}
   * to support custom arrangement of bands or labels across each layout layer.
   * By default this method does nothing.
   *
   * The coordinates used in the resulting Rects may need to be offset by the {@link Layout#arrangementOrigin}.
   * @expose
   * @this {LayeredDigraphLayout}
   * @param {Array.<Rect>} layerRects an Array of {@link Rect}s with the bounds of each of the "layers"
   * @param {Point} offset the position of the top-left corner of the banded area relative to the coordinates given by the layerRects
   * @since 1.4
   */
  protected commitLayers(layerRects: Array<Rect>, offset: Point): void {
  }

  /**
   * Routes the links.
   *
   * This is called by {@link #commitLayout}.
   * This is only called if {@link Layout#isRouting} is true.
   * See also {@link #commitNodes}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected commitLinks(): void {
    // first, clear out all the existing link points
    const edgeIter = this.network.edges.iterator;
    let pLink: Link;
    while (edgeIter.next()) {
      const e = edgeIter.value;
      pLink = e.link;
      if (pLink !== null) {
        pLink.startRoute();
        pLink.clearPoints();
        pLink.commitRoute();
      }
    }
    // second, call Link.updateRoute on each link, just once per Link
    // NB: multiple LayeredDigraphEdges can refer to the same Link
    edgeIter.reset();
    while (edgeIter.next()) {
      const e1 = edgeIter.value;
      pLink = e1.link;
      if (pLink !== null) pLink.updateRoute();
    }
    // Scan all links in the network (both "real" and "artificial"),
    //  and set their locations.
    edgeIter.reset();
    while (edgeIter.next()) {
      const edge = edgeIter.value as LayeredDigraphEdge;
      pLink = edge.link;
      if (pLink === null) continue;
      pLink.startRoute();

      // The "real" physical link and "real" physical ports.
      const golink = pLink;
      let gofromnode = golink.fromNode;
      let gotonode = golink.toNode;
      let gofromport = golink.fromPort;
      let gotoport = golink.toPort;
      if (gofromnode !== null) {
        const visfromnode = gofromnode.findVisibleNode();  // is fromnode itself visible?
        if (visfromnode !== null && visfromnode !== gofromnode) {  // no, found a containing group
          gofromnode = visfromnode;  // use it instead
          gofromport = visfromnode.port;  // no parameter -- it wouldn't apply to a different node anyway
        }
      }
      if (gotonode !== null) {
        const vistonode = gotonode.findVisibleNode();  // is tonode itself visible?
        if (vistonode !== null && vistonode !== gotonode) {  // no, found a containing group
          gotonode = vistonode;  // use it instead
          gotoport = vistonode.port;  // no parameter -- it wouldn't apply to a different node anyway
        }
      }
      const gofromspot = pLink.computeSpot(true, gofromport);
      const gotospot = pLink.computeSpot(false, gotoport);

      // get the from and to nodes.
      let vertexFrom = edge.fromVertex as LayeredDigraphVertex;
      let vertexTo = edge.toVertex as LayeredDigraphVertex;

      // A valid link requires no bends, unless maybe if it's Bezier, to avoid crossing other nodes in the same layers
      if (edge.valid) {
        // it's a "real" link connecting two "real" nodes, not needing any intermediate points
        if (pLink.curve === Link.Bezier && pLink.pointsCount === 4) {
          // when same column, assume straight
          if (vertexFrom.column === vertexTo.column) {
            const p0 = pLink.getLinkPoint(gofromnode, gofromport, gofromspot, true, false, gotonode, gotoport);
            const p3 = pLink.getLinkPoint(gotonode, gotoport, gotospot, false, false, gofromnode, gofromport);
            // handle cases where the port is not visible, so has NaN in its bounds
            if (!p0.isReal()) p0.set(gofromnode.actualBounds.center);
            if (!p3.isReal()) p3.set(gotonode.actualBounds.center);
            pLink.clearPoints();
            pLink.addPointAt(p0.x, p0.y);
            pLink.addPointAt((2 * p0.x + p3.x) / 3, (2 * p0.y + p3.y) / 3);
            pLink.addPointAt((p0.x + 2 * p3.x) / 3, (p0.y + 2 * p3.y) / 3);
            pLink.addPointAt(p3.x, p3.y);
          } else {
            // Not needed when the ports have spots -- check each port separately
            let fromSpotNone = false;
            let toSpotNone = false;
            if (gofromport !== null && gofromspot === Spot.None) fromSpotNone = true;
            if (gotoport !== null && gotospot === Spot.None) toSpotNone = true;
            if (fromSpotNone || toSpotNone) {
              const p0x = pLink.getPoint(0).x;
              const p0y = pLink.getPoint(0).y;
              let c1x = pLink.getPoint(1).x;
              let c1y = pLink.getPoint(1).y;
              let c2x = pLink.getPoint(2).x;
              let c2y = pLink.getPoint(2).y;
              const p3x = pLink.getPoint(3).x;
              const p3y = pLink.getPoint(3).y;
              if (fromSpotNone) {
                if (this._direction === 90 || this._direction === 270) {
                  c1x = p0x;
                  c1y = (p0y + p3y) / 2;
                } else {
                  c1x = (p0x + p3x) / 2;
                  c1y = p0y;
                }
                pLink.setPointAt(1, c1x, c1y);
                const endpt = pLink.getLinkPoint(gofromnode, gofromport, gofromspot, true, false, gotonode, gotoport);
                // handle cases where the port is not visible, so has NaN in its bounds
                if (!endpt.isReal()) endpt.set(gofromnode.actualBounds.center);
                pLink.setPointAt(0, endpt.x, endpt.y);
              }
              if (toSpotNone) {
                if (this._direction === 90 || this._direction === 270) {
                  c2x = p3x;
                  c2y = (p0y + p3y) / 2;
                } else {
                  c2x = (p0x + p3x) / 2;
                  c2y = p3y;
                }
                pLink.setPointAt(2, c2x, c2y);
                const endpt = pLink.getLinkPoint(gotonode, gotoport, gotospot, false, false, gofromnode, gofromport);
                // handle cases where the port is not visible, so has NaN in its bounds
                if (!endpt.isReal()) endpt.set(gotonode.actualBounds.center);
                pLink.setPointAt(3, endpt.x, endpt.y);
              }
            }
          }
        }
        pLink.commitRoute();
        continue;
      }

      // skip links connecting nodes within the same layer, including dummy links
      // inserted to support loopbacks
      if (vertexFrom.layer === vertexTo.layer) {
        pLink.commitRoute();
        continue;
      }

      // now dealing with a !valid link -- find and route all dummy links
      // corresponding to this link that connect artificial nodes with the
      // real from and to nodes.
      let bezier = false;
      let ortho = false;
      let numpts = 0;
      let si = pLink.firstPickIndex + 1;  // insertion index for intermediate stroke points
      if (pLink.isOrthogonal) {
        ortho = true;
        // keep first and last two points
        numpts = pLink.pointsCount;
        if (numpts > 4) {
          pLink.points.removeRange(2, numpts - 3);
        }
      } else if (pLink.curve === Link.Bezier) {
        bezier = true;
        // keep first and last two points
        numpts = pLink.pointsCount;
        if (numpts > 4) {
          pLink.points.removeRange(2, numpts - 3);
        }
        if (numpts === 4) si = 2;  // insert triplets of bezier points after first control point
      } else {
        // keep the first and last points only
        numpts = pLink.pointsCount;
        const nofromspot = (gofromspot === Spot.None);
        const notospot = (gotospot === Spot.None);
        if (numpts > 2 && nofromspot && notospot) {
          pLink.points.removeRange(1, numpts - 2);
        } else if (numpts > 3 && nofromspot && !notospot) {
          pLink.points.removeRange(1, numpts - 3);
        } else if (numpts > 3 && !nofromspot && notospot) {
          pLink.points.removeRange(2, numpts - 2);
        } else if (numpts > 4 && !nofromspot && !notospot) {
          pLink.points.removeRange(2, numpts - 3);
        }
      }

      let vertexA: LayeredDigraphVertex;
      let prevVertexA: LayeredDigraphVertex;

      // An invalid link requires bends.
      // Add bend points at the positions of the "artificial" nodes.
      // Be smarter about Bezier style and about Orthogonal links.
      let prevx = 0.0;
      let prevy = 0.0;
      let nextx = 0.0;
      let nexty = 0.0;
      let mid = 0.0;
      let depthL = 0.0;
      let depthR = 0.0;
      if (!edge.rev) {  // i.e. normal direction
        // Walk the sequence of "artificial" nodes to set the bend points of the links.
        while (vertexFrom !== null && vertexFrom !== vertexTo) {
          // Find the "artificial" node that corresponds to the next bend in the link.
          vertexA = null;
          prevVertexA = null;
          // All links between "artificial" nodes have their to and from vertexes set to the
          // those of the link they are replacing.  Hence,
          // it suffices to find the successor link of the current node
          // which is the "real" physical link.
          const destinationEdgeIter = vertexFrom.destinationEdges.iterator;
          while (destinationEdgeIter.next()) {
            const dedgeA = destinationEdgeIter.value as LayeredDigraphEdge;
            if (dedgeA.link === edge.link) {
              vertexA = dedgeA.toVertex;
              prevVertexA = dedgeA.fromVertex;
              if (prevVertexA.node !== null) prevVertexA = null;
              if (vertexA.node === null) break;
            }
          }
          if (vertexA !== vertexTo) {  // not yet end of iteration?
            prevx = pLink.getPoint(si - 1).x;
            prevy = pLink.getPoint(si - 1).y;
            nextx = vertexA.centerX;
            nexty = vertexA.centerY;
            if (ortho) {
              if (this._direction === 180 || this._direction === 0) {
                const py = (prevVertexA !== null ? prevVertexA.centerY : prevy);
                if (py !== nexty) {
                  mid = this._layerSpaces[vertexA.layer] + this._arrangementOffset.x;
                  if (si === pLink.firstPickIndex + 1) {
                    if (this._direction === 0) {
                      mid = Math.max(mid, prevx);
                    } else {
                      mid = Math.min(mid, prevx);
                    }
                  }
                  pLink.insertPointAt(si++, mid, prevy);
                  pLink.insertPointAt(si++, mid, nexty);
                }
              } else {
                const px = (prevVertexA !== null ? prevVertexA.centerX : prevx);
                if (px !== nextx) {
                  mid = this._layerSpaces[vertexA.layer] + this._arrangementOffset.y;
                  if (si === pLink.firstPickIndex + 1) {
                    if (this._direction === 90) {
                      mid = Math.max(mid, prevy);
                    } else {
                      mid = Math.min(mid, prevy);
                    }
                  }
                  pLink.insertPointAt(si++, prevx, mid);
                  pLink.insertPointAt(si++, nextx, mid);
                }
              }
            } else {  // when style is line or bezier
              depthL = Math.max(10, this._layerSpaceLeft[vertexA.layer]);
              depthR = Math.max(10, this._layerSpaceRight[vertexA.layer]);
              if (this._direction === 180) {
                pLink.insertPointAt(si++, nextx + depthR, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty);
                pLink.insertPointAt(si++, nextx - depthL, nexty);
              } else if (this._direction === 90) {
                pLink.insertPointAt(si++, nextx, nexty - depthL);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty);
                pLink.insertPointAt(si++, nextx, nexty + depthR);
              } else if (this._direction === 270) {
                pLink.insertPointAt(si++, nextx, nexty + depthR);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty);
                pLink.insertPointAt(si++, nextx, nexty - depthL);
              } else {
                pLink.insertPointAt(si++, nextx - depthL, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty);
                pLink.insertPointAt(si++, nextx + depthR, nexty);
              }
            }
          }
          // Update the from node and iterate.
          vertexFrom = vertexA;
        }
        if (ortho) {  // do the last zig-zag
          prevx = pLink.getPoint(si - 1).x;
          prevy = pLink.getPoint(si - 1).y;
          nextx = pLink.getPoint(si).x;
          nexty = pLink.getPoint(si).y;
          if (this._direction === 180 || this._direction === 0) {
            if (prevy !== nexty) {
              if (this._direction === 0) {
                mid = Math.min(Math.max((nextx + prevx) / 2, this._layerSpaces[vertexTo.layer] + this._arrangementOffset.x), nextx);
              } else {
                mid = Math.max(Math.min((nextx + prevx) / 2, this._layerSpaces[vertexTo.layer] + this._arrangementOffset.x), nextx);
              }
              pLink.insertPointAt(si++, mid, prevy);
              pLink.insertPointAt(si++, mid, nexty);
            }
          } else {
            if (prevx !== nextx) {
              // Diagram.Debug(((nexty+prevy)/2).toString() + "  " + this._layerSpaces[vertexTo.layer] + this._arrangementOffset.y.toString());
              if (this._direction === 90) {
                mid = Math.min(Math.max((nexty + prevy) / 2, this._layerSpaces[vertexTo.layer] + this._arrangementOffset.y), nexty);
              } else {
                mid = Math.max(Math.min((nexty + prevy) / 2, this._layerSpaces[vertexTo.layer] + this._arrangementOffset.y), nexty);
              }
              pLink.insertPointAt(si++, prevx, mid);
              pLink.insertPointAt(si++, nextx, mid);
            }
          }
        }
      } else {  // reversed
        // Walk the sequence of "artificial" nodes to set the bend points of the links.
        let f = 0.0;
        while (vertexTo !== null && vertexFrom !== vertexTo) {
          // Find the "artificial" node that corresponds to the next bend in the link.
          vertexA = null;
          prevVertexA = null;
          // All links between "artificial" nodes have their GoObject
          //  set to the GoObject of the link they are replacing.  Hence,
          //  it suffices to find the predecessor link of the current node
          //  which has a GoObject equal to the "real" physical link.
          const sourceEdgeIter = vertexTo.sourceEdges.iterator;
          while (sourceEdgeIter.next()) {
            const sedgeA = sourceEdgeIter.value as LayeredDigraphEdge;
            if (sedgeA.link === edge.link) {
              vertexA = sedgeA.fromVertex;
              prevVertexA = sedgeA.toVertex;
              if (vertexA.node === null) break;
            }
          }
          if (vertexA !== vertexFrom) {  // not yet end of iteration?
            prevx = pLink.getPoint(si - 1).x;
            prevy = pLink.getPoint(si - 1).y;
            nextx = vertexA.centerX;
            nexty = vertexA.centerY;
            if (ortho) {
              if (this._direction === 180 || this._direction === 0) {
                if (si === pLink.firstPickIndex + 1) {  // from real node to first artificial node
                  pLink.insertPointAt(si++, prevx, prevy);
                  pLink.insertPointAt(si++, prevx, nexty);
                } else {
                  const py = (prevVertexA !== null ? prevVertexA.centerY : prevy);
                  if (py !== nexty) {
                    mid = this._layerSpaces[vertexA.layer - 1] + this._arrangementOffset.x;
                    pLink.insertPointAt(si++, mid, prevy);
                    pLink.insertPointAt(si++, mid, nexty);
                  }
                }
              } else {
                if (si === pLink.firstPickIndex + 1) {  // from real node to first artificial node
                  pLink.insertPointAt(si++, prevx, prevy);
                  pLink.insertPointAt(si++, nextx, prevy);
                } else {
                  const px = (prevVertexA !== null ? prevVertexA.centerX : prevx);
                  if (px !== nextx) {
                    mid = this._layerSpaces[vertexA.layer - 1] + this._arrangementOffset.y;
                    pLink.insertPointAt(si++, prevx, mid);
                    pLink.insertPointAt(si++, nextx, mid);
                  }
                }
              }
            } else if (si === pLink.firstPickIndex + 1) {
              depthL = Math.max(10, this._layerSpaceLeft[vertexTo.layer]);
              depthR = Math.max(10, this._layerSpaceRight[vertexTo.layer]);
              if (bezier) {  // when style is bezier, for first segment
                if (this._direction === 180) {
                  if (nextx <= vertexTo.bounds.x) {
                    f = vertexTo.bounds.x;
                    pLink.insertPointAt(si++, f - depthL, nexty);
                    pLink.insertPointAt(si++, f, nexty);
                    pLink.insertPointAt(si++, f + depthR, nexty);
                  } else {
                    pLink.insertPointAt(si++, nextx - depthL, nexty);
                    pLink.insertPointAt(si++, nextx, nexty);
                    pLink.insertPointAt(si++, nextx + depthR, nexty);
                  }
                } else if (this._direction === 90) {
                  if (nexty >= vertexTo.bounds.bottom) {
                    f = vertexTo.bounds.y + vertexTo.bounds.height;
                    pLink.insertPointAt(si++, nextx, f + depthR);
                    pLink.insertPointAt(si++, nextx, f);
                    pLink.insertPointAt(si++, nextx, f - depthL);
                  } else {
                    pLink.insertPointAt(si++, nextx, nexty + depthR);
                    pLink.insertPointAt(si++, nextx, nexty);
                    pLink.insertPointAt(si++, nextx, nexty - depthL);
                  }
                } else if (this._direction === 270) {
                  if (nexty <= vertexTo.bounds.y) {
                    f = vertexTo.bounds.y;
                    pLink.insertPointAt(si++, nextx, f - depthL);
                    pLink.insertPointAt(si++, nextx, f);
                    pLink.insertPointAt(si++, nextx, f + depthR);
                  } else {
                    pLink.insertPointAt(si++, nextx, nexty - depthL);
                    pLink.insertPointAt(si++, nextx, nexty);
                    pLink.insertPointAt(si++, nextx, nexty + depthR);
                  }
                } else if (this._direction === 0) {
                  if (nextx >= vertexTo.bounds.right) {
                    f = vertexTo.bounds.x + vertexTo.bounds.width;
                    pLink.insertPointAt(si++, f + depthR, nexty);
                    pLink.insertPointAt(si++, f, nexty);
                    pLink.insertPointAt(si++, f - depthL, nexty);
                  } else {
                    pLink.insertPointAt(si++, nextx + depthR, nexty);
                    pLink.insertPointAt(si++, nextx, nexty);
                    pLink.insertPointAt(si++, nextx - depthL, nexty);
                  }
                }
              } else {  // for first segment, regular lines
                pLink.insertPointAt(si++, prevx, prevy);
                // make "orthogonal", even though not needed
                let offset = 0;
                if (this._direction === 180 || this._direction === 0) {
                  if (this._direction === 180 ? nextx >= vertexTo.bounds.right : nextx <= vertexTo.bounds.x) {
                    offset = (this._direction === 0 ? -depthL : depthR) / 2;
                  }
                  pLink.insertPointAt(si++, prevx + offset, nexty);
                } else {
                  if (this._direction === 270 ? nexty >= vertexTo.bounds.bottom : nexty <= vertexTo.bounds.y) {
                    offset = (this._direction === 90 ? -depthL : depthR) / 2;
                  }
                  pLink.insertPointAt(si++, nextx, prevy + offset);
                }
                pLink.insertPointAt(si++, nextx, nexty);
              }
            } else {  // when style is line or bezier, for intermediate segments
              depthL = Math.max(10, this._layerSpaceLeft[vertexA.layer]);
              depthR = Math.max(10, this._layerSpaceRight[vertexA.layer]);
              if (this._direction === 180) {
                if (bezier) pLink.insertPointAt(si++, nextx - depthL, nexty);
                pLink.insertPointAt(si++, nextx, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx + depthR, nexty);
              } else if (this._direction === 90) {
                if (bezier) pLink.insertPointAt(si++, nextx, nexty + depthR);
                pLink.insertPointAt(si++, nextx, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty - depthL);
              } else if (this._direction === 270) {
                if (bezier) pLink.insertPointAt(si++, nextx, nexty - depthL);
                pLink.insertPointAt(si++, nextx, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx, nexty + depthR);
              } else {
                if (bezier) pLink.insertPointAt(si++, nextx + depthR, nexty);
                pLink.insertPointAt(si++, nextx, nexty);
                if (bezier) pLink.insertPointAt(si++, nextx - depthL, nexty);
              }
            }
          }
          // Update the to node and iterate.
          vertexTo = vertexA;
        }  // now coming from pNodeFrom
        if (gotoport === null || gofromspot !== Spot.None || ortho) {
          prevx = pLink.getPoint(si - 1).x;
          prevy = pLink.getPoint(si - 1).y;
          nextx = pLink.getPoint(si).x;
          nexty = pLink.getPoint(si).y;
          // make adjustments for reverse links coming "around" the node back into it
          if (ortho) {  // when style is ortho for last segment
            const depth2 = this._layerSpaceRight[vertexFrom.layer];
            let s = 0.0;
            if (this._direction === 180 || this._direction === 0) {
              s = prevy;

              // insertPoints if there is an artificial node in the to-node's layer for reverse links
              if (s >= vertexFrom.bounds.y && s <= vertexFrom.bounds.bottom &&
                (this._direction === 180 ? nextx >= vertexFrom.bounds.x : nextx <= vertexFrom.bounds.right)) {
                f = vertexFrom.centerX + (this._direction === 180 ? -depth2 : depth2);
                if (s < (vertexFrom.bounds.y + vertexFrom.bounds.height / 2)) {
                  s = vertexFrom.bounds.y - this._columnSpacing / 2;
                } else {
                  s = vertexFrom.bounds.bottom + this._columnSpacing / 2;
                }
                pLink.insertPointAt(si++, f, prevy);
                pLink.insertPointAt(si++, f, s);
              }

              pLink.insertPointAt(si++, nextx, s);
              pLink.insertPointAt(si++, nextx, nexty);
            } else {
              s = prevx;

              // insertPoints if there is an artificial node in the to-node's layer for reverse links
              if (s >= vertexFrom.bounds.x && s <= vertexFrom.bounds.right &&
                (this._direction === 270 ? nexty >= vertexFrom.bounds.y : nexty <= vertexFrom.bounds.bottom)) {
                f = vertexFrom.centerY + (this._direction === 270 ? -depth2 : depth2);
                if (s < (vertexFrom.bounds.x + vertexFrom.bounds.width / 2)) {
                  s = vertexFrom.bounds.x - this._columnSpacing / 2;
                } else {
                  s = vertexFrom.bounds.right + this._columnSpacing / 2;
                }
                pLink.insertPointAt(si++, prevx, f);
                pLink.insertPointAt(si++, s, f);
              }

              pLink.insertPointAt(si++, s, nexty);
              pLink.insertPointAt(si++, nextx, nexty);
            }
          } else if (bezier) {  // when style is bezier for last segment
            depthL = Math.max(10, this._layerSpaceLeft[vertexFrom.layer]);
            depthR = Math.max(10, this._layerSpaceRight[vertexFrom.layer]);
            if (this._direction === 180 && nextx >= vertexFrom.bounds.x) {  // loop around
              f = vertexFrom.bounds.x + vertexFrom.bounds.width;
              pLink.setPointAt(si - 2, f, prevy);
              pLink.setPointAt(si - 1, f + depthR, prevy);
            } else if (this._direction === 90 && nexty <= vertexFrom.bounds.bottom) {
              f = vertexFrom.bounds.y;
              pLink.setPointAt(si - 2, prevx, f);
              pLink.setPointAt(si - 1, prevx, f - depthL);
            } else if (this._direction === 270 && nexty >= vertexFrom.bounds.y) {
              f = vertexFrom.bounds.y + vertexFrom.bounds.height;
              pLink.setPointAt(si - 2, prevx, f);
              pLink.setPointAt(si - 1, prevx, f + depthR);
            } else if (this._direction === 0 && nextx <= vertexFrom.bounds.right) {
              f = vertexFrom.bounds.x;
              pLink.setPointAt(si - 2, f, prevy);
              pLink.setPointAt(si - 1, f - depthL, prevy);
            }
          } else {  // when style is line for last segment
            depthL = Math.max(10, this._layerSpaceLeft[vertexFrom.layer]);
            depthR = Math.max(10, this._layerSpaceRight[vertexFrom.layer]);
            // make "orthogonal", even though not needed
            let offset = 0;
            if (this._direction === 180 || this._direction === 0) {
              if (this._direction === 180 ? nextx <= vertexFrom.bounds.x : nextx >= vertexFrom.bounds.right) {
                offset = (this._direction === 0 ? depthR : -depthL) / 2;
              }
              pLink.insertPointAt(si++, nextx + offset, prevy);
            } else {
              if (this._direction === 270 ? nexty <= vertexFrom.bounds.y : nexty >= vertexFrom.bounds.bottom) {
                offset = (this._direction === 90 ? depthR : -depthL) / 2;
              }
              pLink.insertPointAt(si++, prevx, nexty + offset);
            }
            pLink.insertPointAt(si++, nextx, nexty);
          }
        }
      }  // end reversed

      // adjust end points for Bezier links, to look better at the node
      if (golink !== null && bezier) {
        if (gofromport !== null) {
          if (gofromspot === Spot.None) {
            const end = pLink.getPoint(0);
            const ctrl = pLink.getPoint(2);
            if (!end.equals(ctrl)) pLink.setPointAt(1, (end.x + ctrl.x) / 2, (end.y + ctrl.y) / 2);
          }
          const endpt = pLink.getLinkPoint(gofromnode, gofromport, Spot.None, true, false, gotonode, gotoport);
          // handle cases where the port is not visible, so has NaN in its bounds
          if (!endpt.isReal()) endpt.set(gofromnode.actualBounds.center);
          pLink.setPointAt(0, endpt.x, endpt.y);
        }
        if (gotoport !== null) {
          if (gotospot === Spot.None) {
            const end = pLink.getPoint(pLink.pointsCount - 1);
            const ctrl = pLink.getPoint(pLink.pointsCount - 3);
            if (!end.equals(ctrl)) pLink.setPointAt(pLink.pointsCount - 2, (end.x + ctrl.x) / 2, (end.y + ctrl.y) / 2);
          }
          const endpt = pLink.getLinkPoint(gotonode, gotoport, Spot.None, false, false, gofromnode, gofromport);
          // handle cases where the port is not visible, so has NaN in its bounds
          if (!endpt.isReal()) endpt.set(gotonode.actualBounds.center);
          pLink.setPointAt(pLink.pointsCount - 1, endpt.x, endpt.y);
        }
      }
      pLink.commitRoute();
      edge.commit();
    }

    this.avoidOrthogonalOverlaps();
  }

  /**
   * @hidden @internal
   * Try to avoid overlapping segments of Orthogonal links.
   *
   * This is called as part of {@link #commitLinks}.
   * @expose
   * @this {LayeredDigraphLayout}
   */
  protected avoidOrthogonalOverlaps(): void {
    const orthos = new List<Link>();
    const edgeIter = this.network.edges.iterator;
    while (edgeIter.next()) {
      const edge = edgeIter.value;
      const pLink = edge.link;
      if (pLink !== null && pLink.isOrthogonal && !orthos.contains(pLink)) {
        orthos.add(pLink);
      }
    }
    if (orthos.count > 0) {
      if (this._direction === 90 || this._direction === 270) {
        this.adjustOverlapsH(orthos);
      } else {
        this.adjustOverlapsV(orthos);
      }
    }
  }

  /**
   * @hidden @internal
   * @param {*} a
   * @param {*} b
   * @return {number} The result of the comparison.
   */
  private segInfoComparerDefault(a: any, b: any): number {
    if (!(a instanceof SegInfo) || !(b instanceof SegInfo) || (a === b)) {
      return 0;
    }
    if (a.layer < b.layer) return -1;
    if (a.layer > b.layer) return 1;
    if (a.columnMin < b.columnMin) return -1;
    if (a.columnMin > b.columnMin) return 1;
    if (a.columnMax < b.columnMax) return -1;
    if (a.columnMax > b.columnMax) return 1;
    return 0;
  }

  /**
   * @hidden @internal
   * @param {*} a
   * @param {*} b
   * @return {number} The result of the comparison.
   */
  private segInfoComparer2DefaultFirst(a: any, b: any): number {
    if (!(a instanceof SegInfo) || !(b instanceof SegInfo) || (a === b)) {
      return 0;
    }
    if (a.first < b.first) return -1;
    if (a.first > b.first) return 1;
    if (a.turns < b.turns) return 1;
    if (a.turns > b.turns) return -1;
    // if ((a.columnMax-a.columnMin) < (b.columnMax-b.columnMin)) return -1;
    // if ((a.columnMax-a.columnMin) > (b.columnMax-b.columnMin)) return 1;
    if (a.columnMin < b.columnMin) return -1;
    if (a.columnMin > b.columnMin) return 1;
    if (a.columnMax < b.columnMax) return -1;
    if (a.columnMax > b.columnMax) return 1;
    return 0;
  }

  /**
   * @hidden @internal
   * @param {*} a
   * @param {*} b
   * @return {number} The result of the comparison.
   */
  private segInfoComparer2DefaultLast(a: any, b: any): number {
    if (!(a instanceof SegInfo) || !(b instanceof SegInfo) || (a === b)) {
      return 0;
    }
    if (a.last < b.last) return -1;
    if (a.last > b.last) return 1;
    if (a.turns < b.turns) return 1;
    if (a.turns > b.turns) return -1;
    // if ((a.columnMax-a.columnMin) < (b.columnMax-b.columnMin)) return -1;
    // if ((a.columnMax-a.columnMin) > (b.columnMax-b.columnMin)) return 1;
    if (a.columnMin < b.columnMin) return -1;
    if (a.columnMin > b.columnMin) return 1;
    if (a.columnMax < b.columnMax) return -1;
    if (a.columnMax > b.columnMax) return 1;
    return 0;
  }

  /**
   * @hidden @internal
   * @param {number} a
   * @param {number} b
   * @return {boolean}
   */
  private isApprox(a: number, b: number): boolean {
    if (Debug) {
      Util.checkRealNumber(a, LayeredDigraphLayout, 'isApprox:a');
      Util.checkRealNumber(b, LayeredDigraphLayout, 'isApprox:b');
    }
    const d = a - b;
    return (d > -1.0 && d < 1.0);
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {List.<Link>} routables
   * @return {number}
   */
  private adjustOverlapsH(routables: List<Link>): number {
    const layersize = 2;  // smaller than for document
    let overlaps = 0;
    const arr: Array<SegInfo> = [];
    let p: Point;
    let q: Point;
    const linkIter = routables.iterator;
    while (linkIter.next()) {
      const l = linkIter.value;
      if (l !== null && l.isOrthogonal) {
        for (let i = 2; i < l.pointsCount - 3; i++) {
          p = l.getPoint(i);
          q = l.getPoint(i + 1);
          if (this.isApprox(p.y, q.y) && !this.isApprox(p.x, q.x)) {  // assuming Vertical orientation for whole layout
            const info = new SegInfo();
            info.layer = Math.floor(p.y / layersize);
            const p0 = l.getPoint(0);
            const pn = l.getPoint(l.pointsCount - 1);
            info.first = p0.x * p0.x + p0.y;
            info.last = pn.x * pn.x + pn.y;
            info.columnMin = Math.min(p.x, q.x);
            info.columnMax = Math.max(p.x, q.x);
            info.index = i;
            info.link = l;
            if (i + 2 < l.pointsCount) {
              const b = l.getPoint(i - 1);
              const a = l.getPoint(i + 2);
              let t = 0;
              if (b.y < p.y) {
                if (a.y < p.y) t = 3;  // down & up
                else if (p.x < q.x) t = 2;  // crosses segment upward, and jags right
                else t = 1;  // crosses, and jags left
              } else if (b.y > p.y) {
                if (a.y > p.y) t = 0;  // up & down
                else if (q.x < p.x) t = 2;  // crosses segment downward, and jags left
                else t = 1;  // crosses, and jags right
              }
              info.turns = t;
            }
            arr.push(info);
          }
        }
      }
    }
    if (arr.length > 1) {
      arr.sort(this.segInfoComparerDefault);
      let firstlayer = 0;
      while (firstlayer < arr.length) {
        // find all link segments at a layer
        const curlayer = (arr[firstlayer]).layer;
        let nextlayer = firstlayer + 1;
        while (nextlayer < arr.length && (arr[nextlayer]).layer === curlayer) nextlayer++;
        if (nextlayer - firstlayer > 1) {
          // now all SegInfos from firstlayer to nextlayer-1 have layer === curlayer
          let firstgroup = firstlayer;
          while (firstgroup < nextlayer) {
            // find groups of overlapping link segments
            let curmax = (arr[firstgroup]).columnMax;
            let nextgroup = firstlayer + 1;
            while (nextgroup < nextlayer && (arr[nextgroup]).columnMin < curmax) {
              curmax = Math.max(curmax, (arr[nextgroup]).columnMax);
              nextgroup++;
            }
            const groupcount = nextgroup - firstgroup;
            if (groupcount > 1) {
              // now all SegInfos from firstgroup to nextgroup-1 have some overlap with some other
              // link segment in the same group
              // for (i = firstgroup; i < nextgroup; i++) {
              //  var found = false;
              //  for (var j = firstgroup; j < nextgroup; j++) {
              //    if (j === i) continue;
              //    var a = arr[i];
              //    var b = arr[j];
              //    if (a.columnMax >= b.columnMin || b.columnMax >= a.columnMin) {
              //      found = true;
              //      break;
              //    }
              //  }
              // }

              // sort all of the overlapping links by their Last position and Turns number
              this.sortRange(arr, this.segInfoComparer2DefaultLast, firstgroup, firstgroup + groupcount);
              // count how many overlapping links connect to the same port (actually, same port column)
              let samelastportcount = 1;
              let prevportpos = (arr[firstgroup]).last;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if (si.last !== prevportpos) {
                  samelastportcount++;
                  prevportpos = si.last;
                }
              }
              // sort all of the overlapping links by their First position and Turns number
              this.sortRange(arr, this.segInfoComparer2DefaultFirst, firstgroup, firstgroup + groupcount);
              // count how many overlapping links connect to the same port (actually, same port column)
              let samefirstportcount = 1;
              prevportpos = (arr[firstgroup]).first;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if (si.first !== prevportpos) {
                  samefirstportcount++;
                  prevportpos = si.first;
                }
              }
              let useFirst = true;
              let sameportcount = samefirstportcount;
              if (samelastportcount < samefirstportcount) {
                useFirst = false;
                sameportcount = samelastportcount;
                prevportpos = (arr[firstgroup]).last;
                // resort by Last position
                this.sortRange(arr, this.segInfoComparer2DefaultLast, firstgroup, firstgroup + groupcount);
              } else {
                prevportpos = (arr[firstgroup]).first;
                // already sorted by First position
              }

              // now spread out overlapping links
              // ??? need to do bin packing
              let lay = 0;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if ((useFirst ? si.first : si.last) !== prevportpos) {
                  lay++;
                  prevportpos = (useFirst ? si.first : si.last);
                }
                const l = si.link;
                p = l.getPoint(si.index);
                q = l.getPoint(si.index + 1);
                const dy = this.linkSpacing * (lay - (sameportcount - 1) / 2.0);  // ??? 4
                if (!l.isAvoiding || this.isUnoccupied2(p.x, p.y + dy, q.x, q.y + dy)) {
                  overlaps++;
                  l.startRoute();
                  l.setPointAt(si.index, p.x, p.y + dy);
                  l.setPointAt(si.index + 1, q.x, q.y + dy);
                  l.commitRoute();
                }
              }
            }
            firstgroup = nextgroup;
          }
        }
        firstlayer = nextlayer;
      }
    }
    return overlaps;
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {Array<T>} arr
   * @param {((a: T, b: T) => number)} sortfunc
   * @param {number} from
   * @param {number} to
   */
  private sortRange<T>(arr: Array<T>, sortfunc: ((a: T, b: T) => number), from: number, to: number) {
    const len = arr.length;
    const num = to - from;
    if (num <= 1) return;

    // FROM can't be at or beyond the last item, because NUM >= 2
    if (from < 0 || from >= len - 1) Util.throwError('not in range 0 <= from < length: ' + from);

    if (num === 2) {  // handle special case of just two items in the range
      const first = arr[from];
      const second = arr[from + 1];
      if (sortfunc(first, second) > 0) {
        arr[from] = second;
        arr[from + 1] = first;
      }
      return;
    }

    // now NUM >= 3
    if (from === 0) {
      if (to >= len) {  // sort the whole array
        arr.sort(sortfunc);
      } else {  // not the end of the array
        const temp = arr.slice(0, to);
        temp.sort(sortfunc);
        for (let i = 0; i < to; i++) arr[i] = temp[i];
      }
    } else {
      if (to >= len) { // not the beginning of the array, but all the rest
        const temp = arr.slice(from);
        temp.sort(sortfunc);
        for (let i = from; i < len; i++) arr[i] = temp[i - from];
      } else {  // neither beginning nor the end of the array
        const temp = arr.slice(from, to);
        temp.sort(sortfunc);
        for (let i = from; i < to; i++) arr[i] = temp[i - from];
      }
    }
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {List.<Link>} routables
   * @return {number}
   */
  private adjustOverlapsV(routables: List<Link>): number {
    const layersize = 2;  // smaller than for document
    let overlaps = 0;
    const arr: Array<SegInfo> = [];
    let p: Point;
    let q: Point;
    const linkIter = routables.iterator;
    while (linkIter.next()) {
      const l = linkIter.value;
      if (l !== null && l.isOrthogonal) {
        for (let i = 2; i < l.pointsCount - 3; i++) {
          p = l.getPoint(i);
          q = l.getPoint(i + 1);
          if (this.isApprox(p.x, q.x) && !this.isApprox(p.y, q.y)) {  // assuming Vertical orientation for whole layout
            const info = new SegInfo();
            info.layer = Math.floor(p.x / layersize);
            const p0 = l.getPoint(0);
            const pn = l.getPoint(l.pointsCount - 1);
            info.first = p0.x + p0.y * p0.y;
            info.last = pn.x + pn.y * pn.y;
            info.columnMin = Math.min(p.y, q.y);
            info.columnMax = Math.max(p.y, q.y);
            info.index = i;
            info.link = l;
            if (i + 2 < l.pointsCount) {
              const b = l.getPoint(i - 1);
              const a = l.getPoint(i + 2);
              let t = 0;
              if (b.x < p.x) {
                if (a.x < p.x) t = 3;  // down & up
                else if (p.y < q.y) t = 2;  // crosses segment upward, and jags right
                else t = 1;  // crosses, and jags left
              } else if (b.x > p.x) {
                if (a.x > p.x) t = 0;  // up & down
                else if (q.y < p.y) t = 2;  // crosses segment downward, and jags left
                else t = 1;  // crosses, and jags right
              }
              info.turns = t;
            }
            arr.push(info);
          }
        }
      }
    }
    if (arr.length > 1) {
      arr.sort(this.segInfoComparerDefault);
      let firstlayer = 0;
      while (firstlayer < arr.length) {
        // find all link segments at a layer
        const curlayer = (arr[firstlayer]).layer;
        let nextlayer = firstlayer + 1;
        while (nextlayer < arr.length && (arr[nextlayer]).layer === curlayer) nextlayer++;
        if (nextlayer - firstlayer > 1) {
          // now all SegInfos from firstlayer to nextlayer-1 have layer === curlayer
          let firstgroup = firstlayer;
          while (firstgroup < nextlayer) {
            // find groups of overlapping link segments
            let curmax = (arr[firstgroup]).columnMax;
            let nextgroup = firstlayer + 1;
            while (nextgroup < nextlayer && (arr[nextgroup]).columnMin < curmax) {
              curmax = Math.max(curmax, (arr[nextgroup]).columnMax);
              nextgroup++;
            }
            const groupcount = nextgroup - firstgroup;
            if (groupcount > 1) {
              // now all SegInfos from firstgroup to nextgroup-1 have some overlap with some other
              // link segment in the same group
              // for (i = firstgroup; i < nextgroup; i++) {
              //  var found = false;
              //  for (var j = firstgroup; j < nextgroup; j++) {
              //    if (j === i) continue;
              //    var a = arr[i];
              //    var b = arr[j];
              //    if (a.columnMax >= b.columnMin || b.columnMax >= a.columnMin) {
              //      found = true;
              //      break;
              //    }
              //  }
              // }

              // sort all of the overlapping links by their Last position and Turns number
              this.sortRange(arr, this.segInfoComparer2DefaultLast, firstgroup, firstgroup + groupcount);
              // count how many overlapping links connect to the same port (actually, same port column)
              let samelastportcount = 1;
              let prevportpos = (arr[firstgroup]).last;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if (si.last !== prevportpos) {
                  samelastportcount++;
                  prevportpos = si.last;
                }
              }
              // sort all of the overlapping links by their First position and Turns number
              this.sortRange(arr, this.segInfoComparer2DefaultFirst, firstgroup, firstgroup + groupcount);
              // count how many overlapping links connect to the same port (actually, same port column)
              let samefirstportcount = 1;
              prevportpos = (arr[firstgroup]).first;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if (si.first !== prevportpos) {
                  samefirstportcount++;
                  prevportpos = si.first;
                }
              }
              let useFirst = true;
              let sameportcount = samefirstportcount;
              if (samelastportcount < samefirstportcount) {
                useFirst = false;
                sameportcount = samelastportcount;
                prevportpos = (arr[firstgroup]).last;
                // resort by Last position
                this.sortRange(arr, this.segInfoComparer2DefaultLast, firstgroup, firstgroup + groupcount);
              } else {
                prevportpos = (arr[firstgroup]).first;
                // already sorted by First position
              }

              // now spread out overlapping links
              // ??? need to do bin packing
              let lay = 0;
              for (let i = firstgroup; i < nextgroup; i++) {
                const si = arr[i];
                if ((useFirst ? si.first : si.last) !== prevportpos) {
                  lay++;
                  prevportpos = (useFirst ? si.first : si.last);
                }
                const l = si.link;
                p = l.getPoint(si.index);
                q = l.getPoint(si.index + 1);
                const dx = this.linkSpacing * (lay - (sameportcount - 1) / 2.0);  // ??? 4
                if (!l.isAvoiding || this.isUnoccupied2(p.x + dx, p.y, q.x + dx, q.y)) {
                  overlaps++;
                  l.startRoute();
                  l.setPointAt(si.index, p.x + dx, p.y);
                  l.setPointAt(si.index + 1, q.x + dx, q.y);
                  l.commitRoute();
                }
              }
            }
            firstgroup = nextgroup;
          }
        }
        firstlayer = nextlayer;
      }
    }
    return overlaps;
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {number} px
   * @param {number} py
   * @param {number} qx
   * @param {number} qy
   * @return {boolean}
   */
  private isUnoccupied2(px: number, py: number, qx: number, qy: number): boolean {  // TODO implement when functionality is available
    if (Debug) {
      Util.checkRealNumber(px, LayeredDigraphLayout, 'isUnoccupied2:px');
      Util.checkRealNumber(py, LayeredDigraphLayout, 'isUnoccupied2:py');
      Util.checkRealNumber(qx, LayeredDigraphLayout, 'isUnoccupied2:qx');
      Util.checkRealNumber(qy, LayeredDigraphLayout, 'isUnoccupied2:qy');
    }
    return true;
    //  if (this.diagram === null) return true;
    //  var panel = this.diagram.panel;
    //  if (panel === null) return true;
    //  var minx = Math.min(px, qx);
    //  var miny = Math.min(py, qy);
    //  var maxx = Math.max(px, qx);
    //  var maxy = Math.max(py, qy);
    //  return panel.isUnoccupied(new Rect(minx, miny, maxx-minx, maxy-miny), null);

  }

  /**
   * @hidden @internal
   * Resets several of the private variables in LayeredDigraphLayout.
   * @this {LayeredDigraphLayout}
   */
  private _clearCaches(): void {
    this._maxIndex = -1;
    this._minIndexLayer = 0;
    this._maxIndexLayer = 0;
    this._mySavedLayout = null;
    this._myCrossings = null;
    this._myLayers = null;
    for (let i = 0; i < this._myCachedNodeArrayLists.length; i++) {
      this._myCachedNodeArrayLists[i] = null;
    }
  }

  /**
   * @hidden @internal
   * Gets the cached node array list.
   * Tries to reuse some of the temporary node arrays.
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer the layer from which nodes are drawn to fill {@link #unfixedLayerNodes}.
   * @return {Array.<LayeredDigraphVertex>} the cached node array list.
   */
  private _getCachedNodeArrayList(unfixedLayer: number): Array<LayeredDigraphVertex> {
    let unfixedLayerNodes: Array<LayeredDigraphVertex>;
    const num = this._indices[unfixedLayer];

    if (num >= this._myCachedNodeArrayLists.length) {
      const temp: Array<Array<LayeredDigraphVertex>> = [];
      for (let i = 0; i < this._myCachedNodeArrayLists.length; i++) {
        temp[i] = this._myCachedNodeArrayLists[i];
      }
      this._myCachedNodeArrayLists = temp;
    }
    if (this._myCachedNodeArrayLists[num] === undefined || this._myCachedNodeArrayLists[num] === null) {
      unfixedLayerNodes = [];
    } else {
      unfixedLayerNodes = this._myCachedNodeArrayLists[num];
      this._myCachedNodeArrayLists[num] = null;
    }

    const layer = this._myLayers[unfixedLayer];
    for (let i = 0; i < layer.length; i++) {
      const vertex = layer[i];
      unfixedLayerNodes[vertex.index] = vertex;
    }
    return unfixedLayerNodes;
  }

  /**
   * @hidden @internal
   * @this {LayeredDigraphLayout}
   * @param {number} unfixedLayer
   * @param {Array.<LayeredDigraphVertex>} nodes
   */
  private _freeCachedNodeArrayList(unfixedLayer: number, nodes: Array<LayeredDigraphVertex>) {
    this._myCachedNodeArrayLists[this._indices[unfixedLayer]] = nodes;
  }

  // ***************************************************************
  // Property Definitions

  /**
   * Gets or sets the space between each layer.
   * This value must be positive and it defaults to 25.
   */
  /** @export */ get layerSpacing(): number { return this._layerSpacing; }
  /** @export */ set layerSpacing(value: number) {
    if (this._layerSpacing !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphLayout, 'layerSpacing');
      if (value >= 0) {
        this._layerSpacing = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the size of each column.
   * This value must be positive and it defaults to 25.
   */
  /** @export */ get columnSpacing(): number { return this._columnSpacing; }
  /** @export */ set columnSpacing(value: number) {
    if (this._columnSpacing !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphLayout, 'columnSpacing');
      if (value > 0) {
        this._columnSpacing = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the direction the graph grows towards.
   * 0 is towards the right, 90 is downwards, 180 is towards the left, and 270 is upwards.
   * The default value is 0.
   */
  /** @export */ get direction(): number { return this._direction; }
  /** @export */ set direction(value: number) {
    if (this._direction !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphLayout, 'direction');
      if (value === 0 || value === 90 || value === 180 || value === 270) {
        this._direction = value;
        this.invalidateLayout();
      } else {
        Util.throwError('LayeredDigraphLayout.direction must be 0, 90, 180, or 270');
      }
    }
  }

  /**
   * Gets or set which cycle removal option is used.
   * The default value is {@link LayeredDigraphLayout.CycleDepthFirst}.
   */
  /** @export */ get cycleRemoveOption(): EnumValue { return this._cycleRemoveOption; }
  /** @export */ set cycleRemoveOption(value: EnumValue) {
    if (this._cycleRemoveOption !== value) {
      EnumValue.checkEnumValue(value, LayeredDigraphLayout, LayeredDigraphLayout, 'cycleRemoveOption');
      if (value === LayeredDigraphLayout.CycleGreedy || value === LayeredDigraphLayout.CycleDepthFirst || value === LayeredDigraphLayout.CycleFromLayers) {
        this._cycleRemoveOption = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets which layering option is being used.
   * The default value is {@link LayeredDigraphLayout.LayerOptimalLinkLength}.
   */
  /** @export */ get layeringOption(): EnumValue { return this._layeringOption; }
  /** @export */ set layeringOption(value: EnumValue) {
    if (this._layeringOption !== value) {
      EnumValue.checkEnumValue(value, LayeredDigraphLayout, LayeredDigraphLayout, 'layeringOption');
      if (value === LayeredDigraphLayout.LayerOptimalLinkLength || value === LayeredDigraphLayout.LayerLongestPathSink || value === LayeredDigraphLayout.LayerLongestPathSource) {
        this._layeringOption = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets which indices initialization option is being used.
   * The default value is {@link LayeredDigraphLayout.InitDepthFirstOut}.
   */
  /** @export */ get initializeOption(): EnumValue { return this._initializeOption; }
  /** @export */ set initializeOption(value: EnumValue) {
    if (this._initializeOption !== value) {
      EnumValue.checkEnumValue(value, LayeredDigraphLayout, LayeredDigraphLayout, 'initializeOption');
      if (value === LayeredDigraphLayout.InitDepthFirstOut || value === LayeredDigraphLayout.InitDepthFirstIn || value === LayeredDigraphLayout.InitNaive) {
        this._initializeOption = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the number of iterations to be done.
   * The value must be non-negative.  The default value is 4.
   */
  /** @export */ get iterations(): number { return this._iterations; }
  /** @export */ set iterations(value: number) {
    if (this._iterations !== value) {
      Util.checkRealNumber(value, LayeredDigraphNetwork, 'iterations');
      if (value >= 0) {
        this._iterations = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets which aggressive option is being used to look for link crossings.
   * The default value is {@link LayeredDigraphLayout.AggressiveLess}.
   */
  /** @export */ get aggressiveOption(): EnumValue { return this._aggressiveOption; }
  /** @export */ set aggressiveOption(value: EnumValue) {
    if (this._aggressiveOption !== value) {
      EnumValue.checkEnumValue(value, LayeredDigraphLayout, LayeredDigraphLayout, 'aggressiveOption');
      if (value === LayeredDigraphLayout.AggressiveNone || value === LayeredDigraphLayout.AggressiveLess || value === LayeredDigraphLayout.AggressiveMore) {
        this._aggressiveOption = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the options used by the straighten and pack function,
   * The default value is {@link LayeredDigraphLayout.PackAll}.
   */
  /** @export */ get packOption(): number { return this._packOption; }
  /** @export */ set packOption(value: number) {
    if (this._packOption !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphLayout, 'packOption');
      if (value >= 0 && value < 8) {
        this._packOption = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets whether the FromSpot and ToSpot of each link should be set
   * to values appropriate for the given value of {@link LayeredDigraphLayout#direction}.
   * The default value is true.
   *
   * If you set this to false, the spot values of the links and port objects will be used.
   * If you do not set the spot values to sensible values matching the {@link #direction},
   * the routing results may be poor and they may cross over nodes.
   */
  /** @export */ get setsPortSpots(): boolean { return this._setsPortSpots; }
  /** @export */ set setsPortSpots(value: boolean) {
    if (this._setsPortSpots !== value) {
      Util.checkPrimitive(value, 'boolean', LayeredDigraphLayout, 'setsPortSpots');
      this._setsPortSpots = value;
      this.invalidateLayout();
    }
  }

  //    * Gets or sets the distance between link segments for unrelated links that would otherwise be overlapping.
  //  * The default value is 4.  The value must be positive.
  //  *
  //  * If there are a lot of possibly overlapping links and the {@link #layerSpacing} is not large enough,
  //  * some links may cross over nodes.
  //  * You will either need to increase the layerSpacing or decrease this linkSpacing property.
  /**
   * @hidden
   * Undocumented
   */
  /** @export */ get linkSpacing(): number { return this._linkSpacing; }
  /** @export */ set linkSpacing(value: number) {
    if (this._linkSpacing !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphLayout, 'linkSpacing');
      if (value >= 0) {
        this._linkSpacing = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * This read-only property returns the largest layer value.
   */
  /** @export */ get maxLayer(): number { return this._maxLayer; }

  /**
   * This read-only property returns the largest index value.
   */
  /** @export */ get maxIndex(): number { return this._maxIndex; }

  /**
   * This read-only property returns the largest column value.
   */
  /** @export */ get maxColumn(): number { return this._maxColumn; }

  /**
   * This read-only property returns the smallest index layer.
   */
  /** @export */ get minIndexLayer(): number { return this._minIndexLayer; }

  /**
   * This read-only property returns the larges index layer.
   */
  /** @export */ get maxIndexLayer(): number { return this._maxIndexLayer; }

  // *****************************************************************************************
  // Enums

  /**
   * Remove cycles using depth first cycle removal;
   * a valid value of {@link LayeredDigraphLayout#cycleRemoveOption}.
   * @constant
   */
  public static CycleDepthFirst = new EnumValue(LayeredDigraphLayout, 'CycleDepthFirst', 0);

  /**
   * Remove cycles using greedy cycle removal;
   * a valid value of {@link LayeredDigraphLayout#cycleRemoveOption}.
   * @constant
   */
  public static CycleGreedy = new EnumValue(LayeredDigraphLayout, 'CycleGreedy', 1);

  /**
   * Remove cycles depending on {@link #assignLayers} determining entirely which links are backwards links;
   * a valid value of {@link LayeredDigraphLayout#cycleRemoveOption}.
   * @constant
   */
  public static CycleFromLayers = new EnumValue(LayeredDigraphLayout, 'CycleFromLayers', 2);

  /**
   * Assign layers using optimal link length layering;
   * A valid value for {@link LayeredDigraphLayout#layeringOption}.
   * @constant
   */
  public static LayerOptimalLinkLength = new EnumValue(LayeredDigraphLayout, 'LayerOptimalLinkLength', 0);

  /**
   * Assign layers using longest path sink layering;
   * a valid value for {@link LayeredDigraphLayout#layeringOption}.
   * @constant
   */
  public static LayerLongestPathSink = new EnumValue(LayeredDigraphLayout, 'LayerLongestPathSink', 1);

  /**
   * Assign layers using longest path source layering;
   * a valid value for {@link LayeredDigraphLayout#layeringOption}.
   * @constant
   */
  public static LayerLongestPathSource = new EnumValue(LayeredDigraphLayout, 'LayerLongestPathSource', 2);

  /**
   * Initialize using depth first out initialization;
   * a valid value for {@link LayeredDigraphLayout#initializeOption}.
   * @constant
   */
  public static InitDepthFirstOut = new EnumValue(LayeredDigraphLayout, 'InitDepthFirstOut', 0);

  /**
   * Initialize using depth first in initialization;
   * a valid value for {@link LayeredDigraphLayout#initializeOption}.
   * @constant
   */
  public static InitDepthFirstIn = new EnumValue(LayeredDigraphLayout, 'InitDepthFirstIn', 1);

  /**
   * Initialize using naive initialization;
   * a valid value for {@link LayeredDigraphLayout#initializeOption}.
   * @constant
   */
  public static InitNaive = new EnumValue(LayeredDigraphLayout, 'InitNaive', 2);

  /**
   * The fastest, but poorest, crossing reduction algorithm;
   * a valid value for {@link LayeredDigraphLayout#aggressiveOption}.
   * @constant
   */
  public static AggressiveNone = new EnumValue(LayeredDigraphLayout, 'AggressiveNone', 0);

  /**
   * The faster, less aggressive, crossing reduction algorithm;
   * a valid value for {@link LayeredDigraphLayout#aggressiveOption}.
   * @constant
   */
  public static AggressiveLess = new EnumValue(LayeredDigraphLayout, 'AggressiveLess', 1);

  /**
   * The slower, more aggressive, crossing reduction algorithm,
   * a valid value for {@link LayeredDigraphLayout#aggressiveOption}.
   * @constant
   */
  public static AggressiveMore = new EnumValue(LayeredDigraphLayout, 'AggressiveMore', 2);

  /**
   * Does minimal work in packing the nodes;
   * a valid value for {@link LayeredDigraphLayout#packOption}.
   * @constant
   */
  public static PackNone = 0;

  /**
   * This option gives more chances for the packing algorithm to improve the network,
   * but is very expensive in time for large networks;
   * a valid value for {@link LayeredDigraphLayout#packOption}.
   * @constant
   */
  public static PackExpand = 1;

  /**
   * This option tries to have the packing algorithm straighten many of the
   * links that cross layers,
   * a valid value for {@link LayeredDigraphLayout#packOption}.
   * @constant
   */
  public static PackStraighten = 2;

  /**
   * This option tries to have the packing algorithm center groups of nodes
   * based on their relationships with nodes in other layers,
   * a valid value for {@link LayeredDigraphLayout#packOption}.
   * @constant
   */
  public static PackMedian = 4;

  /**
   * Enable all options for the {@link LayeredDigraphLayout#packOption} property;
   * See also {@link LayeredDigraphLayout.PackExpand}, {@link LayeredDigraphLayout.PackStraighten},
   * and {@link LayeredDigraphLayout.PackMedian}.
   * @constant
   */
  public static PackAll = 7;
} // #end LayeredDigraphLayout
LayeredDigraphLayout['className'] = 'LayeredDigraphLayout';

LayeredDigraphLayout['CycleDepthFirst'] = LayeredDigraphLayout.CycleDepthFirst;
LayeredDigraphLayout['CycleGreedy'] = LayeredDigraphLayout.CycleGreedy;
LayeredDigraphLayout['CycleFromLayers'] = LayeredDigraphLayout.CycleFromLayers;
LayeredDigraphLayout['LayerOptimalLinkLength'] = LayeredDigraphLayout.LayerOptimalLinkLength;
LayeredDigraphLayout['LayerLongestPathSink'] = LayeredDigraphLayout.LayerLongestPathSink;
LayeredDigraphLayout['LayerLongestPathSource'] = LayeredDigraphLayout.LayerLongestPathSource;
LayeredDigraphLayout['InitDepthFirstOut'] = LayeredDigraphLayout.InitDepthFirstOut;
LayeredDigraphLayout['InitDepthFirstIn'] = LayeredDigraphLayout.InitDepthFirstIn;
LayeredDigraphLayout['InitNaive'] = LayeredDigraphLayout.InitNaive;
LayeredDigraphLayout['AggressiveNone'] = LayeredDigraphLayout.AggressiveNone;
LayeredDigraphLayout['AggressiveLess'] = LayeredDigraphLayout.AggressiveLess;
LayeredDigraphLayout['AggressiveMore'] = LayeredDigraphLayout.AggressiveMore;

LayeredDigraphLayout['PackNone'] = LayeredDigraphLayout.PackNone;
LayeredDigraphLayout['PackExpand'] = LayeredDigraphLayout.PackExpand;
LayeredDigraphLayout['PackStraighten'] = LayeredDigraphLayout.PackStraighten;
LayeredDigraphLayout['PackMedian'] = LayeredDigraphLayout.PackMedian;
LayeredDigraphLayout['PackAll'] = LayeredDigraphLayout.PackAll;

/**
 * @hidden @internal
 * @unrestricted
 */
export class SegInfo {
  /** @hidden @internal */ public static className = 'SegInfo';
  /** @hidden @internal */ public layer: number;
  /** @hidden @internal */ public first: number;
  /** @hidden @internal */ public last: number;
  /** @hidden @internal */ public columnMin: number;
  /** @hidden @internal */ public columnMax: number;
  /** @hidden @internal */ public index: number;
  /** @hidden @internal */ public link: Link | null;
  /** @hidden @internal */ public turns: number;

  constructor() {
    this.layer = 0;
    this.first = 0;
    this.last = 0;
    this.columnMin = 0;
    this.columnMax = 0;
    this.index = 0;
    this.link = null;
    this.turns = 0;
  }
} // #end SegInfo
SegInfo['className'] = 'SegInfo';

// ****************************************************************************************
// Network, Vertex, and Edge

/**
 * This class represents an abstract graph of {@link LayeredDigraphVertex}es and {@link LayeredDigraphEdge}s
 * that can be constructed based on the {@link Node}s and {@link Link}s of a {@link Diagram}
 * so that the {@link LayeredDigraphLayout} can operate independently of the diagram until it
 * is time to commit any node positioning or link routing.
 *
 * This class inherits from {@link LayoutNetwork}.
 * @extends LayoutNetwork
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class LayeredDigraphNetwork extends LayoutNetwork {
  /** @hidden @internal */ public static className = 'LayeredDigraphNetwork';
  constructor(layout: LayeredDigraphLayout) {
    super(layout);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link LayeredDigraphVertex}.
   * @this {LayeredDigraphNetwork}
   * @return {LayeredDigraphVertex}
   */
  public createVertex(): LayeredDigraphVertex {
    return new LayeredDigraphVertex(this);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link LayeredDigraphEdge}.
   * @this {LayeredDigraphNetwork}
   * @return {LayeredDigraphEdge}
   */
  public createEdge(): LayeredDigraphEdge {
    return new LayeredDigraphEdge(this);
  }

} // #end LayeredDigraphNetwork
LayeredDigraphNetwork['className'] = 'LayeredDigraphNetwork';

/**
 * This holds {@link LayeredDigraphLayout}-specific information about {@link Node}s.
 *
 * This class inherits from {@link LayoutVertex}.
 * @extends LayoutVertex
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class LayeredDigraphVertex extends LayoutVertex {
  /** @hidden @internal */ public static className = 'LayeredDigraphVertex';
  /** @hidden @internal */ private _layer: number;
  /** @hidden @internal */ private _column: number;
  /** @hidden @internal */ private _index: number;
  /** @hidden @internal */ private _component: number;
  /** @hidden @internal */ private _near: LayeredDigraphVertex | null;
  /** @hidden @internal */ public valid: boolean;
  /** @hidden @internal */ public discover: number;
  /** @hidden @internal */ public finish: number;
  /** @hidden @internal */ public artificialType: number;
  /** @hidden @internal */ public nodeMinColumnSpaceTopLeft: number | null;
  /** @hidden @internal */ public nodeMinColumnSpaceBottomRight: number | null;

  /** @hidden @internal */ // public sourceEdgesList: List<LayeredDigraphEdge>;
  /** @hidden @internal */ // public destinationEdgesList: List<LayeredDigraphEdge>;

  constructor(network: LayeredDigraphNetwork) {
    super(network);

    // layered-specific properties:
    this._layer = -1;

    this._column = -1;

    this._index = -1;

    this._component = NaN;

    this._near = null;

    // internal state
    this.valid = false;
    this.discover = NaN;
    this.finish = NaN;

    this.artificialType = 0;
    this.nodeMinColumnSpaceTopLeft = null;
    this.nodeMinColumnSpaceBottomRight = null;
  }

  /**
   * Gets or sets the layer to which the node is assigned.
   * The default value is -1.
   */
  /** @export */ get layer(): number { return this._layer; }
  /** @export */ set layer(value: number) {
    if (this._layer !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphVertex, 'layer');
      this._layer = value;
    }
  }

  /**
   * Gets or sets the column to which the node is assigned.
   * The default value is -1.
   */
  /** @export */ get column(): number { return this._column; }
  /** @export */ set column(value: number) {
    if (this._column !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphVertex, 'column');
      this._column = value;
    }
  }

  /**
   * Gets or sets the index to which the node is assigned.
   * The default value is -1.
   */
  /** @export */ get index(): number { return this._index; }
  /** @export */ set index(value: number) {
    if (this._index !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphVertex, 'index');
      this._index = value;
    }
  }

  /**
   * Gets or sets the connected component to which the node is assigned.
   * The default value is NaN.
   */
  /** @export */ get component(): number { return this._component; }
  /** @export */ set component(value: number) {
    if (this._component !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphVertex, 'component');
      this._component = value;
    }
  }

  /**
   * Gets or sets another LayeredDigraphVertex in the same layer that this node should be near.
   * The default value is null.
   */
  /** @export */ get near(): LayeredDigraphVertex { return this._near; }
  /** @export */ set near(value: LayeredDigraphVertex) {
    if (this._near !== value) {
      if (Debug && value !== null) Util.checkClass(value, LayeredDigraphVertex, LayeredDigraphVertex, 'near');
      this._near = value;
    }
  }
} // #end LayeredDigraphVertex
LayeredDigraphVertex['className'] = 'LayeredDigraphVertex';

/**
 * This holds {@link LayeredDigraphLayout}-specific information about {@link Link}s.
 *
 * This class inherits from {@link LayoutEdge}.
 * @extends LayoutEdge
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class LayeredDigraphEdge extends LayoutEdge {
  /** @hidden @internal */ public static className = 'LayeredDigraphEdge';
  /** @hidden @internal */ private _valid: boolean;
  /** @hidden @internal */ private _rev: boolean;
  /** @hidden @internal */ private _forest: boolean;
  /** @hidden @internal */ private _portFromPos: number;
  /** @hidden @internal */ private _portToPos: number;
  /** @hidden @internal */ private _portFromColOffset: number;
  /** @hidden @internal */ private _portToColOffset: number;

  /** @hidden @internal */ public fromVertex: LayeredDigraphVertex | null;
  /** @hidden @internal */ public toVertex: LayeredDigraphVertex | null;

  constructor(network: LayeredDigraphNetwork) {
    super(network);

    // layered-specific properties:
    this._valid = false;

    this._rev = false;

    this._forest = false;

    this._portFromPos = NaN;

    this._portToPos = NaN;

    this._portFromColOffset = 0;

    this._portToColOffset = 0;
  }

  /**
   * True if the link is part of the proper digraph.
   * The default value is false.
   */
  /** @export */ get valid(): boolean { return this._valid; }
  /** @export */ set valid(value: boolean) {
    if (this._valid !== value) {
      Util.checkPrimitive(value, 'boolean', LayeredDigraphEdge, 'valid');
      this._valid = value;
    }
  }

  /**
   * True if the link was reversed during cycle removal.
   * The default value is false.
   */
  /** @export */ get rev(): boolean { return this._rev; }
  /** @export */ set rev(value: boolean) {
    if (this._rev !== value) {
      Util.checkPrimitive(value, 'boolean', LayeredDigraphEdge, 'rev');
      this._rev = value;
    }
  }

  /**
   * True if the link is part of the depth first forest.
   * The default value is false.
   */
  /** @export */ get forest(): boolean { return this._forest; }
  /** @export */ set forest(value: boolean) {
    if (this._forest !== value) {
      Util.checkPrimitive(value, 'boolean', LayeredDigraphEdge, 'forest');
      this._forest = value;
    }
  }

  /**
   * Location of the port at the from node of the link.
   * Allows the crossing matrix to correctly calculate the crossings for nodes with multiple ports.
   * The default value is NaN.
   */
  /** @export */ get portFromPos(): number { return this._portFromPos; }
  /** @export */ set portFromPos(value: number) {
    if (this._portFromPos !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphEdge, 'portFromPos');
      this._portFromPos = value;
    }
  }

  /**
   * Location of the port at the to node of the link.
   * Allows the crossing matrix to correctly calculate the crossings for nodes with multiple ports.
   * The default value is NaN.
   */
  /** @export */ get portToPos(): number { return this._portToPos; }
  /** @export */ set portToPos(value: number) {
    if (this._portToPos !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphEdge, 'portToPos');
      this._portToPos = value;
    }
  }

  /**
   * Approximate column offset of the from port of the link from the from node column used in straightening.
   * The default value is 0.
   */
  /** @export */ get portFromColOffset(): number { return this._portFromColOffset; }
  /** @export */ set portFromColOffset(value: number) {
    if (this._portFromColOffset !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphEdge, 'portFromColOffset');
      this._portFromColOffset = value;
    }
  }

  /**
   * Approximate column offset of the to port of the link from the to node column used in straightening.
   * The default value is 0.
   */
  /** @export */ get portToColOffset(): number { return this._portToColOffset; }
  /** @export */ set portToColOffset(value: number) {
    if (this._portToColOffset !== value) {
      Util.checkPrimitive(value, 'number', LayeredDigraphEdge, 'portToColOffset');
      this._portToColOffset = value;
    }
  }
} // #end LayeredDigraphEdge
LayeredDigraphEdge['className'] = 'LayeredDigraphEdge';


/*
*  Copyright (C) 1998-2019 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

/*
TreeLayout
TreeNetwork
TreeVertex
TreeEdge
*/

// import { List, Set } from './collections';
// import { Iterable } from './collectionsInterfaces';
// import { Debug } from './debug';
// import { Diagram } from './diagram';
// import { EnumValue } from './enumValue';
// import { Layout, LayoutEdge, LayoutNetwork, LayoutVertex } from './layout';
// import { Group, Part } from './parts';
// import { Node } from './parts';
// import { Link } from './parts';
// import { Point } from './point';
// import { Rect } from './rect';
// import { Size } from './size';
// import { Spot } from './spot';
// import { Util } from './util';


/**
 * This layout positions nodes of a tree-structured graph in layers (rows or columns).
 * <p class="boxrun">
 * For a discussion and examples of the most commonly used properties, see <a href="../../intro/trees.html">Trees</a> page in the Introduction.
 * If you want to experiment interactively with most of the properties, try the <a href="../../samples/tLayout.html">Tree Layout</a> sample.
 * See samples that make use of TreeLayout in the <a href="../../samples/index.html#treelayout">samples index</a>.
 *
 * This layout makes use of a {@link LayoutNetwork} of
 * {@link TreeVertex}es and {@link TreeEdge}s that normally
 * correspond to the {@link Node}s and {@link Link}s of the {@link Diagram}.
 *
 * The most commonly set properties for controlling the results of a TreeLayout are:
 *   - {@link #angle}: the direction in which the tree grows, from parent to child;
 *     the default value of zero means that the tree grows towards the right, with the children of a node arranged in a layer that is a column.
 *     An angle of 0 or 180 means that children form vertical layers -- breadth is height and depth is width;
 *     an angle of 90 or 270 means that children form horizontal layers -- breadth is width and depth is height.
 *   - {@link #layerSpacing}: the distance between layers -- between a parent node and its child nodes.
 *   - {@link #nodeSpacing}: the distance between nodes within a layer -- between siblings.
 *   - {@link #alignment}: the relative position of a parent node with its children.
 *   - {@link #sorting} and {@link #comparer}: specify the order of the immediate children of a parent node.
 *   - {@link #compaction}: whether subtrees should be packed closer together if there is room.
 *   - {@link #layerStyle}: whether the children of one node are aligned with the children of a sibling node.
 *   - {@link #setsPortSpot}, {@link #portSpot}, {@link #setsChildPortSpot}, and {@link #childPortSpot}: this controls whether to set the {@link Link#fromSpot} and
 *     {@link Link#toSpot} to be sensible for the {@link #angle}.
 *   - {@link #nodeIndent} and {@link #nodeIndentPastParent}: if the {@link #alignment} is {@link TreeLayout.AlignmentStart} or {@link TreeLayout.AlignmentEnd},
 *     control how much extra space the first child is given when positioned.
 *   - {@link #breadthLimit}, {@link #rowSpacing}: try to limit the total breadth of a subtree to a certain distance;
 *     when there are too many children or when they are too broad, this puts children into additional rows (or columns, depending on the angle)
 *     thereby limiting the breadth while increasing the depth of the tree.
 *
 * When you set one of the TreeLayout properties listed above, that property normally applies to all of the nodes in the tree.
 * What if you want {@link #alignment} to be {@link TreeLayout.AlignmentCenterChildren} for the root node but {@link TreeLayout.AlignmentBus} for the other nodes in the tree?
 * Or what if you want want {@link #layerSpacing} to be 50 for all layers except for the layer separating "leaf" nodes from their parent?
 *
 * One common solution is to set {@link #treeStyle}.
 * For the former scenario, you could set {@link #treeStyle} to {@link TreeLayout.StyleRootOnly}; the value of {@link #alignment} would only apply to the root node.
 * For the latter scenario, you could set it to {@link TreeLayout.StyleLastParents};
 * the value of {@link #layerSpacing} would apply to all nodes except those that have children but that do not have grandchildren.
 * How do you then set the alignment or layerSpacing for the other nodes?
 * By setting the TreeLayout properties whose names start with "alternate...".
 * In these cases that would mean setting {@link #alternateAlignment} or {@link #alternateLayerSpacing}.
 *
 * These TreeLayout properties actually apply to the {@link TreeVertex} that the TreeLayout uses to represent a {@link Node} within the {@link LayoutNetwork}.
 * All of those TreeLayout properties are actually stored in {@link #rootDefaults}; all of the "alternate..." properties are stored in {@link #alternateDefaults}.
 * Depending on the value of {@link #treeStyle}, the actual TreeVertex properties for each Node are copied appropriately from either rootDefaults or alternateDefaults.
 * In the default case where treeStyle is {@link TreeLayout.StyleLayered}, the alternateDefaults are ignored.
 * (Note that treeStyle, and a few other properties such as {@link #path} and {@link #arrangement}, apply to the whole layout, not to an individual node/vertex.)
 *
 * The use of {@link #treeStyle} and "alternate..." TreeLayout properties will cover a lot of common needs for tree layout customization.
 * However, there may be times when that is not enough.
 * Imagine a situation where you want a special {@link TreeVertex} property value for a particular {@link Node}.
 * The solution is to override {@link #assignTreeVertexValues}, where you can examine the given {@link TreeVertex},
 * including its corresponding {@link LayoutVertex#node}, to decide what TreeVertex property values should apply.
 * @extends Layout
 * @unrestricted
 * @category Layout
 */
export class TreeLayout extends Layout {
  /** @hidden @internal */ public static className = 'TreeLayout';
  /** @hidden @internal */ private _roots: Set<(TreeVertex | Node)>;
  /** @hidden @internal */ private _path: EnumValue;
  /** @hidden @internal */ private _treeStyle: EnumValue;
  /** @hidden @internal */ private _layerStyle: EnumValue;
  /** @hidden @internal */ private _allDepths: Array<number>;
  /** @hidden @internal */ private _comments: boolean;
  /** @hidden @internal */ private _arrangement: EnumValue;
  /** @hidden @internal */ private _arrangementSpacing: Size;
  /** @hidden @internal */ private _rootDefaults: TreeVertex;
  /** @hidden @internal */ private _alternateDefaults: TreeVertex;
  /** @hidden @internal */ private _tempArrays: Array<Array<Array<Point>>>;
  /** @hidden @internal */ public _effectivePath: EnumValue;

  /**
   * Constructs a {@link TreeLayout} with no {@link Layout#network}
   * and with no owning {@link Layout#diagram}.
   */
  constructor() {
    if (arguments.length > 0) Util.throwConstructorArgumentsError(TreeLayout);
    super();

    this._roots = new Set<TreeVertex | Node>();
    this._path = TreeLayout.PathDefault;
    this._effectivePath = TreeLayout.PathDestination;
    this._treeStyle = TreeLayout.StyleLayered;
    this._layerStyle = TreeLayout.LayerIndividual;
    this._allDepths = [];
    this._comments = true;
    this._arrangement = TreeLayout.ArrangementVertical;
    this._arrangementSpacing = new Size(10, 10).freeze();
    const net = new TreeNetwork(this);
    this._rootDefaults = new TreeVertex(net);
    this._alternateDefaults = new TreeVertex(net);

    this._tempArrays = [];
  }

  /**
   * @hidden @internal
   * Copies properties to a cloned Layout.
   * @this {TreeLayout}
   * @param {?} copy
   */
  protected cloneProtected(copy: this): void {
    super.cloneProtected(copy);
    // _roots gets new empty List
    copy._path = this._path;
    copy._treeStyle = this._treeStyle;
    copy._layerStyle = this._layerStyle;
    copy._comments = this._comments;
    copy._arrangement = this._arrangement;
    copy._arrangementSpacing.assign(this._arrangementSpacing);
    copy._rootDefaults.copyInheritedPropertiesFrom(this._rootDefaults);
    copy._alternateDefaults.copyInheritedPropertiesFrom(this._alternateDefaults);
  }

  /**
   * @hidden @internal
   * Called by GraphObject.processMakeEnum
   * @this {TreeLayout}
   * @param {EnumValue} arg
   */
  public acceptEnum(arg: EnumValue): void {
    if (arg.classType === TreeLayout) {
      if (arg.name.indexOf('Alignment') === 0) {
        this.alignment = arg;
      } else if (arg.name.indexOf('Arrangement') === 0) {
        this.arrangement = arg;
      } else if (arg.name.indexOf('Compaction') === 0) {
        this.compaction = arg;
      } else if (arg.name.indexOf('Path') === 0) {
        this.path = arg;
      } else if (arg.name.indexOf('Sorting') === 0) {
        this.sorting = arg;
      } else if (arg.name.indexOf('Style') === 0) {
        this.treeStyle = arg;
      } else {
        Util.throwError('Unknown enum value: ' + arg);
      }
    } else {
      super.acceptEnum(arg);
    }
  }

  /**
   * Create a new {@link LayoutNetwork} of {@link TreeVertex}es and {@link TreeEdge}s.
   * @this {TreeLayout}
   * @return {TreeNetwork} a new {@link LayoutNetwork}.
   */
  public createNetwork(): TreeNetwork {
    return new TreeNetwork(this);
  }

  /**
   * Create and initialize a {@link LayoutNetwork} with the given nodes and links.
   *
   * This does not include any nodes of category "Comment".
   * Comment nodes are added by the {@link #addComments} method.
   * @this {TreeLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   * @return {LayoutNetwork}
   */
  public makeNetwork(coll: Diagram | Group | Iterable<Part>): LayoutNetwork {
    const net = this.createNetwork();
    const pred =
      /**
       * @hidden @internal
       * @param {Part} part
       */
      function (part: Part) {
        if (part instanceof Node) {
          return !part.isLinkLabel && part.category !== 'Comment';
        } else if (part instanceof Link) {  // no links connecting with link label nodes
          const from = part.fromNode;
          if (from === null || from.isLinkLabel || from.category === 'Comment') return false;
          const to = part.toNode;
          if (to === null || to.isLinkLabel || to.category === 'Comment') return false;
          return true;
        }
        return false;  // no simple Parts
      };

    if (coll instanceof Diagram) {
      net.addParts(coll.nodes, true, pred);
      net.addParts(coll.links, true, pred);
    } else if (coll instanceof Group) {
      net.addParts(coll.memberParts, false, pred);
    } else {
      net.addParts(coll.iterator, false, pred);
    }
    return net;
  }

  /**
   * Perform the tree layout.
   *
   * If there is no {@link Layout#network}, this calls {@link #makeNetwork} to create a {@link LayoutNetwork} from the given collection of Parts.
   *
   * If there are no {@link TreeVertex} {@link #roots} specified, this finds all roots in the {@link Layout#network}.
   *
   * This initializes all of the {@link TreeVertex}es and {@link TreeEdge}s, calling {@link #initializeTreeVertexValues} on each vertex,
   * supporting inheritance of vertex values.
   * Then it calls {@link #assignTreeVertexValues} on each one, to allow for node/vertex-specific customizations.
   * Next it sorts all of the child vertexes for each parent vertex, if needed.
   *
   * This also calls {@link #addComments} of each vertex, in order to find any comment nodes associated with each vertex,
   * so that they can be accommodated by the layout.
   *
   * This then actually does the "layout" of the vertexes and optionally the routing of the edges of each tree in the network.
   * To deal with multiple roots/trees this also calls {@link #arrangeTrees} to position each separate tree relative to each other.
   *
   * Finally this calls {@link Layout#updateParts} to commit the Node positions from the vertex positions and the Link routes from the edges.
   * {@link Layout#updateParts} calls {@link #commitLayout} within a transaction.
   * @this {TreeLayout}
   * @param {Diagram|Group|Iterable.<Part>} coll A {@link Diagram} or a {@link Group} or a collection of {@link Part}s.
   */
  public doLayout(coll: Diagram | Group | Iterable<Part>): void {
    if (Debug && coll === null) Util.throwError('Layout.doLayout(collection) argument must not be null but a Diagram, a Group, or an Iterable of Parts');
    if (this.network === null) {
      this.network = this.makeNetwork(coll);
    }

    if (this.arrangement !== TreeLayout.ArrangementFixedRoots) {
      this.arrangementOrigin = this.initialOrigin(this.arrangementOrigin);
    }

    let diagram = this.diagram;
    if (diagram === null && coll instanceof Diagram) diagram = coll;
    if (this.path === TreeLayout.PathDefault && diagram !== null) {
      this._effectivePath = (diagram.isTreePathToChildren ? TreeLayout.PathDestination : TreeLayout.PathSource);
    } else {
      this._effectivePath = (this.path === TreeLayout.PathDefault ? TreeLayout.PathDestination : this.path);
    }

    if (this.network.vertexes.count > 0) {
      this.createTrees();  // calls recursive walkTree
      this.initializeAll();  // calls non-recursive initializeTreeVertexValues
      this.assignAll();  // calls non-recursive assignTreeVertexValues
      this.sortAll();  // calls non-recursive sortTreeVertexChildren
      this.annotateAll();  // calls non-recursive addComments
      this.layoutAllTrees();  // calls recursive layoutTree
      this.arrangeTrees();

      // Update the "physical" positions of the nodes and links.
      this.updateParts();
    }

    // restore in case of future re-use
    this.network = null;
    this._roots = new Set<(TreeVertex | Node)>();

    this.isValidLayout = true;
  }

  /**
   * @hidden @internal
   * This method is responsible for initializing all of the {@link TreeVertex}es,
   * setting {@link TreeVertex#initialized}, {@link TreeVertex#level},
   * {@link TreeVertex#parent}, and {@link TreeVertex#children}, and making sure roots
   * has at least one suitable {@link TreeVertex} in it.
   *
   * This will call findRoots if roots is empty.
   * Then it will iterate over the roots, calling walkTree on each one.
   * @this {TreeLayout}
   */
  private createTrees(): void {
    // ??? support incremental layouts
    this.network.deleteSelfEdges();
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value as TreeVertex;
      v.initialized = false;
      v.level = 0;
      v.parent = null;
      v.children = [];
    }
    // for compatibility, substitute TreeVertexes for all Nodes in _roots collection
    if (this._roots.count > 0) {
      const newroots = new Set<(TreeVertex | Node)>();
      const rit = this._roots.iterator;
      while (rit.next()) {
        const r = rit.value;  // either Node or TreeVertex
        if (r instanceof Node) {
          const rv = this.network.findVertex(r) as TreeVertex;
          if (rv !== null) {
            newroots.add(rv);
          }
        } else if (r instanceof TreeVertex) {
          newroots.add(r);
        }
      }
      this._roots = newroots;
    }
    if (this._roots.count === 0) {
      this.findRoots();  // this can be overridden
    }
    // iterate over a copy of the .roots Set because walkTree might remove vertexes from that collection
    const it = this._roots.copy().iterator;
    while (it.next()) {
      const rt = it.value as TreeVertex;
      if (!rt.initialized) {
        // walkTree will remove objects in the roots collection that are child nodes
        rt.initialized = true;
        this.walkTree(rt);
      }
    }

    // Maybe there are some rings, of the structure A --> B --> C --> A.
    // If there is at least one root found (some vertex other than A, B, or C),
    // findRoots would not normally choose any of A,B,C because they all have input edges.
    // Hence we need to see if there are any vertexes that had not been walked by walkTree above.
    // From those we need to choose an extra root and try walkTree again.
    let remainingVertexes = this.network.vertexes as any as Set<TreeVertex>;
    let unwalked: Set<TreeVertex> = null;
    while (unwalked = this.findUnwalkedVertexes(remainingVertexes), unwalked.count > 0) {
      const bestroot = this.chooseRoot(unwalked);
      if (bestroot !== null) this._roots.add(bestroot);

      // walkTree will remove objects in the roots collection that are child nodes
      bestroot.initialized = true;
      this.walkTree(bestroot);

      // search for any remaining unwalked vertexes from the most recent set of known unwalked
      remainingVertexes = unwalked;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Set.<TreeVertex>} vertexes
   * @return {Set.<TreeVertex>}
   */
  private findUnwalkedVertexes(vertexes: Set<TreeVertex>): Set<TreeVertex> {
    const set = new Set<TreeVertex>();
    const vit = vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      if (!v.initialized) set.add(v);
    }
    return set;
  }

  /**
   * @hidden @internal
   * This method is responsible for finding all of the root nodes.
   *
   * When you have not already added the roots to the roots collection,
   * this will choose a root node.
   * The choice may not be what you want, so we recommend that you specify the roots.
   * @expose
   * @this {TreeLayout}
   */
  protected findRoots(): void {
    const vertexes = this.network.vertexes as any as Set<TreeVertex>;
    const vit = vertexes.iterator;
    while (vit.next()) {
      const v = vit.value as TreeVertex;
      switch (this._effectivePath) {
        case TreeLayout.PathDestination:
          if (v.sourceEdges.count === 0) this._roots.add(v);
          break;
        case TreeLayout.PathSource:
          if (v.destinationEdges.count === 0) this._roots.add(v);
          break;
        default:
          Util.throwError('Unhandled path value ' + this._effectivePath.toString());
          break;
      }
    }
    if (this._roots.count === 0) {
      const bestroot = this.chooseRoot(vertexes);
      if (bestroot !== null) this._roots.add(bestroot);
    }
  }

  /**
   * @hidden @internal
   * This method is responsible for picking a root node if none were found.
   * @this {TreeLayout}
   * @param {Set.<TreeVertex>} vertexes
   * @return {TreeVertex}
   */
  private chooseRoot(vertexes: Set<TreeVertex>): TreeVertex {
    let mincount = 999999;
    let bestroot: TreeVertex = null;
    const vit = vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      switch (this._effectivePath) {
        case TreeLayout.PathDestination:
          if (v.sourceEdges.count < mincount) { mincount = v.sourceEdges.count; bestroot = v; }
          break;
        case TreeLayout.PathSource:
          if (v.destinationEdges.count < mincount) { mincount = v.destinationEdges.count; bestroot = v; }
          break;
        default:
          Util.throwError('Unhandled path value ' + this._effectivePath.toString());
          break;
      }
    }
    return bestroot;
  }

  /**
   * @hidden @internal
   * Traverse the {@link Network} and assign the {@link TreeVertex#parent},
   * {@link TreeVertex#level}, and {@link TreeVertex#children} properties.
   *
   * This method should walk the tree recursively.
   * The standard implementation gracefully handles shared nodes, including cyclical references.
   * However, there can only be one {@link TreeVertex#parent} per node.
   * If a node is declared to be the child of multiple nodes, it is not guaranteed which
   * node wins as the parent node.
   * If any of the children of this node are in the roots collection,
   * they are removed from that collection.
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private walkTree(v: TreeVertex): void {  // ?? this could be optimized
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'walkTree:v');
    // handle DAGs gracefully, and ignore cycles
    switch (this._effectivePath) {
      case TreeLayout.PathDestination:
        if (v.destinationEdges.count > 0) {
          const list = new List<TreeVertex>();
          const vit = v.destinationVertexes;
          while (vit.next()) {
            const dc = vit.value as TreeVertex;
            if (this.walkOK(v, dc)) list.add(dc);
          }
          if (list.count > 0) v.children = list.toArray();
        }
        break;
      case TreeLayout.PathSource:
        if (v.sourceEdges.count > 0) {
          const list = new List<TreeVertex>();
          const vit = v.sourceVertexes;
          while (vit.next()) {
            const sc = vit.value as TreeVertex;
            if (this.walkOK(v, sc)) list.add(sc);
          }
          if (list.count > 0) v.children = list.toArray();
        }
        break;
      default:
        Util.throwError('Unhandled path value' + this._effectivePath.toString());
    }
    const children = v.children;
    const childrenLength = children.length;
    // set Initialized before calling WalkTree
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      c.initialized = true;
      c.level = v.level + 1;
      c.parent = v;
      this._roots.remove(c);
    }
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.walkTree(c);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {TreeVertex} c
   * @return {boolean}
   */
  private walkOK(v: TreeVertex, c: TreeVertex): boolean {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'walkOK:v');
    if (Debug) Util.checkClass(c, TreeVertex, TreeLayout, 'walkOK:c');
    if (c.initialized) {
      if (this.isAncestor(c, v)) return false;
      if (c.level > v.level) return false;
      this.removeChild(c.parent, c);
      // new level will be assigned just before calling walkTree on c
      return true;
    } else {
      return true;
    }
  }

  /**
   * @hidden @internal
   * Determines if vertex a is an ancestor of vertex b.
   * @this {TreeLayout}
   * @param {TreeVertex} a
   * @param {TreeVertex} b
   * @return {boolean}
   */
  private isAncestor(a: TreeVertex, b: TreeVertex): boolean {
    if (b === null) return false;
    if (Debug) Util.checkClass(a, TreeVertex, TreeLayout, 'isAncestor:a');
    if (Debug) Util.checkClass(b, TreeVertex, TreeLayout, 'isAncestor:b');
    let x = b.parent;
    while (x !== null && x !== a) x = x.parent;
    return (x === a);
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} p
   * @param {TreeVertex} c
   */
  private removeChild(p: TreeVertex, c: TreeVertex): void {
    if (p === null) return;
    if (c === null) return;
    if (Debug) Util.checkClass(p, TreeVertex, TreeLayout, 'removeChild:p');
    if (Debug) Util.checkClass(c, TreeVertex, TreeLayout, 'removeChild:c');
    const v = p.children;
    let num = 0;
    for (let i = 0; i < v.length; i++) if (v[i] === c) num++;
    if (num > 0) {
      const r = new Array(v.length - num);
      let j = 0;
      for (let i = 0; i < v.length; i++) if (v[i] !== c) r[j++] = v[i];
      p.children = r;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private initializeAll(): void {
    const rit = this._roots.iterator;
    while (rit.next()) {
      const v = rit.value;
      if (v instanceof TreeVertex) this.initializeTree(v);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private initializeTree(v: TreeVertex): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'initializeTree:v');
    this.initializeTreeVertexValues(v); // inherit property values
    if (v.alignment === TreeLayout.AlignmentBusBranching) {
      this.sortTreeVertexChildren(v); // maybe modify order of v.children
    }
    let count = 0;
    let maxchildren = v.childrenCount;
    let maxgeneration = 0;
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.initializeTree(c);
      count += c.descendantCount + 1;
      maxchildren = Math.max(maxchildren, c.maxChildrenCount);
      maxgeneration = Math.max(maxgeneration, c.maxGenerationCount);
    }
    v.descendantCount = count;
    v.maxChildrenCount = maxchildren;
    v.maxGenerationCount = (maxchildren > 0 ? maxgeneration + 1 : 0);
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @return {TreeVertex}
   */
  private mom(v: TreeVertex): TreeVertex {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'mom:v');
    switch (this._treeStyle) {
      default:
      case TreeLayout.StyleLayered:
        if (v.parent !== null) return v.parent;
        return this._rootDefaults;
      case TreeLayout.StyleRootOnly:
        if (v.parent === null) return this._rootDefaults;
        if (v.parent.parent === null) return this._alternateDefaults;
        return v.parent;
      case TreeLayout.StyleAlternating:
        if (v.parent !== null) {
          if (v.parent.parent !== null) return v.parent.parent;
          return this._alternateDefaults;
        }
        return this._rootDefaults;
      case TreeLayout.StyleLastParents: {
        let lastparent = true;
        if (v.childrenCount === 0) {
          lastparent = false;
        } else {
          const children = v.children;
          const childrenLength = children.length;
          for (let i = 0; i < childrenLength; i++) {
            const c = children[i];
            if (c.childrenCount > 0) {
              lastparent = false;
              break;
            }
          }
        }
        if (lastparent && v.parent !== null) return this._alternateDefaults;
        if (v.parent !== null) return v.parent;
        return this._rootDefaults;
      }
    }
  }

  /**
   * Assign initial property values for a {@link TreeVertex}.
   *
   * The values may be inherited, so this method is called while
   * propagating values from the root nodes.
   * This method should not walk the tree, since it is called for each
   * {@link TreeVertex} in a depth-first manner starting at a root.
   *
   * You probably do not need to override this method,
   * but if you do you should call first either the base method
   * or {@link TreeVertex#copyInheritedPropertiesFrom}, since they
   * assign most of the {@link TreeVertex} property values
   * used to influence the layout.
   * Informational properties such as {@link TreeVertex#descendantCount}
   * and {@link TreeVertex#maxGenerationCount} will not yet have been initialized
   * by the time this method is called.
   * It is more common to override {@link #assignTreeVertexValues} in order to
   * modify a property or two to customize the layout at that node.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * When the {@link TreeVertex#alignment} is {@link TreeLayout.AlignmentBusBranching},
   * this will modify the {@link TreeVertex#angle} appropriately depending on which
   * side of the bus the vertex is on.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @since 1.2
   */
  public initializeTreeVertexValues(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'initializeTreeVertexValues:v');
    const mom = this.mom(v);
    v.copyInheritedPropertiesFrom(mom);

    // caution: this depends on the parent having been sorted first
    if (v.parent !== null && v.parent.alignment === TreeLayout.AlignmentBusBranching) {
      let a = v.angle;
      const pchildren = v.parent.children;
      let i = 0;
      while (i < pchildren.length) {
        if (v === pchildren[i]) break;
        i++;
      }
      if (i % 2 === 0) {  // even child: turn
        // but last child keeps going in same direction, the rest turn
        if (i !== pchildren.length - 1) {
          if (a === 90) a = 180;
          else if (a === 180) a = 270;
          else if (a === 270) a = 180;
          else a = 270;
        }
      } else {  // odd child: turn the other way
        if (a === 90) a = 0;
        else if (a === 180) a = 90;
        else if (a === 270) a = 0;
        else a = 90;
      }
      v.angle = a;
    }

    v.initialized = true;  // should already be true, but make sure
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private assignAll(): void {
    const rit = this._roots.iterator;
    while (rit.next()) {
      const v = rit.value;
      if (v instanceof TreeVertex) this.assignTree(v);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private assignTree(v: TreeVertex): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'assignTree:v');
    this.assignTreeVertexValues(v);
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.assignTree(c);
    }
  }

  /**
   * Assign final property values for a {@link TreeVertex}.
   *
   * This method is commonly overridden in order to provide
   * tree layout properties for particular nodes.
   * This method is called after values have been inherited from other
   * {@link TreeVertex}es, so you can examine and modify the
   * values of related tree nodes.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   *
   * However, when {@link TreeVertex#alignment} is {@link TreeLayout.AlignmentBusBranching},
   * changing the {@link TreeVertex#sorting} or
   * {@link TreeVertex#comparer} properties in this method will have no effect.
   *
   * This method should not walk the tree, since it is called for each
   * {@link TreeVertex} in a depth-first manner starting at a root.
   *
   * Here is an example where the children are squeezed together if there are many of them,
   * but only on nodes that have no grandchildren.  This makes use of two {@link TreeVertex}
   * properties that are automatically computed for you, {@link TreeVertex#childrenCount} and {@link TreeVertex#descendantCount}.
   * ```js
   *   function SqueezingTreeLayout() {
   *     go.TreeLayout.call(this);
   *   }
   *   go.Diagram.inherit(SqueezingTreeLayout, go.TreeLayout);
   *
   *   public assignTreeVertexValues(v) {
   *     if (v.childrenCount > 6 && v.childrenCount === v.descendantCount) {
   *       v.alignment = go.TreeLayout.AlignmentBottomRightBus;
   *       v.layerSpacing = 10;
   *       v.rowSpacing = 0;
   *     }
   *   }
   * ```
   *
   * If you need to assign TreeVertex values and also have them be "inherited" by the child vertexes,
   * you should override {@link #initializeTreeVertexValues} instead.
   * However at the time that method is called, the computed properties of TreeVertex will not be available.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @since 1.1
   */
  public assignTreeVertexValues(v: TreeVertex): void {
    // if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'assignTreeVertexValues:v');
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private sortAll(): void {
    const rit = this._roots.iterator;
    while (rit.next()) {
      const v = rit.value;
      if (v instanceof TreeVertex) this.sortTree(v);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private sortTree(v: TreeVertex): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'sortTree:v');
    if (v.alignment !== TreeLayout.AlignmentBusBranching) {
      this.sortTreeVertexChildren(v);
    }
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.sortTree(c);
    }
  }

  /**
   * @hidden @internal
   * Sort the {@link TreeVertex#children} of a node.
   * The default behavior depends on the value of {@link TreeVertex#sorting}.
   *
   * This method should not walk the tree, since it is called for each
   * {@link TreeVertex} in a depth-first manner starting at a root.
   * Normally this is called after all of the vertexes have had their values
   * set by {@link #assignTreeVertexValues}.
   *
   * However, if {@link TreeVertex#alignment} is {@link TreeLayout.AlignmentBusBranching},
   * this will be called before {@link #assignTreeVertexValues}.
   * The "BusBranching" alignment depends on inheriting the {@link TreeVertex#angle},
   * which in turn depends on the sort order, hence sorting must be performed first.
   * The difference is that when "BusBranching" the initialization pass has not yet finished,
   * so informational properties have not yet been computed when this method is called.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  public sortTreeVertexChildren(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'sortTreeVertexChildren:v');
    switch (v.sorting) {
      case TreeLayout.SortingForwards:  // nothing needed
        break;
      case TreeLayout.SortingReverse:  // just reverse the array
        v.children.reverse();
        break;
      case TreeLayout.SortingAscending:  // sort using comparer
        v.children.sort(v.comparer);
        break;
      case TreeLayout.SortingDescending:  // sort, then reverse
        v.children.sort(v.comparer);
        v.children.reverse();
        break;
      default:
        Util.throwError('Unhandled sorting value ' + v.sorting.toString());
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private annotateAll(): void {
    if (!this.comments) return;
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value as TreeVertex;
      this.addComments(v);
    }
  }

  /**
   * Find any associated objects to be positioned along with the {@link LayoutVertex#node}.
   *
   * This looks for visible {@link Node}'s whose category is "Comment" and
   * that refer to the tree vertex's {@link Node}.
   * This method is only called when {@link #comments} is true.
   *
   * You may want to override this method in order to customize how
   * any associated objects are found and how the node's {@link LayoutVertex#bounds}
   * are set to reserve space for those associated objects.
   * This method should not walk the tree, since it is called for each
   * {@link TreeVertex} in an indeterminate order.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @since 1.2
   */
  public addComments(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'addComments:v');
    const a = v.angle;
    const parent = v.parent;
    let pa = 0.0;
    let palign = TreeLayout.AlignmentCenterChildren;
    let inbus = false;
    if (parent !== null) {
      pa = parent.angle;
      palign = parent.alignment;
      inbus = this.isBusAlignment(palign);
    }
    const vertical = (a === 90 || a === 270);
    const pvertical = (pa === 90 || pa === 270);
    const leaf = (v.childrenCount === 0);
    let extraWidth = 0.0;
    let extraHeight = 0.0;
    let spacing = 0.0;
    const commentSpacing = v.commentSpacing;
    if (v.node !== null) {
      const nit = v.node.findNodesConnected();
      while (nit.next()) {
        const comment = nit.value;
        if (comment.category !== 'Comment') continue;
        if (!comment.canLayout()) continue;
        if (v.comments === null) v.comments = [];
        v.comments.push(comment);
        comment.ensureBounds();
        const cb = comment.measuredBounds;
        if ((vertical && !leaf) || (!inbus && !pvertical && leaf) || (inbus && pvertical && leaf)) {
          // if vertical tree, leave room for comments on the side, unless it's a leaf
          extraWidth = Math.max(extraWidth, cb.width);
          extraHeight += cb.height + Math.abs(spacing);
        } else {
          extraWidth += cb.width + Math.abs(spacing);
          extraHeight = Math.max(extraHeight, cb.height);
        }
        spacing = commentSpacing;
      }
    }
    if (v.comments !== null) {
      if ((vertical && !leaf) || (!inbus && !pvertical && leaf) || (inbus && pvertical && leaf)) {
        extraWidth += Math.abs(v.commentMargin);
        extraHeight = Math.max(0, extraHeight - v.height);
      } else {
        extraHeight += Math.abs(v.commentMargin);
        extraWidth = Math.max(0, extraWidth - v.width);
      }
      const crect = Rect.allocAt(0, 0, v.bounds.width + extraWidth, v.bounds.height + extraHeight);
      v.bounds = crect;
      Rect.free(crect);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {EnumValue} a
   * @return {boolean}
   */
  public isBusAlignment(a: EnumValue): boolean {
    return a === TreeLayout.AlignmentBus || a === TreeLayout.AlignmentBusBranching ||
      a === TreeLayout.AlignmentTopLeftBus || a === TreeLayout.AlignmentBottomRightBus;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {EnumValue} a
   * @return {boolean}
   */
  private isBusTwoSided(a: EnumValue): boolean {
    return a === TreeLayout.AlignmentBus || a === TreeLayout.AlignmentBusBranching;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @return {boolean}
   */
  private isLeftSideBus(v: TreeVertex): boolean {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'isLeftSideBus:v');
    const parent = v.parent;
    if (parent !== null) {
      const palign = parent.alignment;
      if (this.isBusAlignment(palign)) {
        if (this.isBusTwoSided(palign)) {
          const children = parent.children;
          let i = 0;
          while (i < children.length) {
            if (v === children[i]) break;
            i++;
          }
          return (i % 2 === 0);
        } else if (palign === TreeLayout.AlignmentTopLeftBus) {
          return true;
        } else {
          return false;
        }
      }
    }
    return false;
  }

  /**
   * Position and {@link TreeVertex#comments} around the vertex.
   *
   * This method should not walk the tree, since it is called for each
   * {@link TreeVertex} in an indeterminate order.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @since 1.2
   */
  public layoutComments(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutComments:v');
    if (v.comments !== null) {
      const node = v.node;
      const nodebounds = node.measuredBounds;
      const parent = v.parent;
      const a = v.angle;
      let pa = 0.0;
      let palign = TreeLayout.AlignmentCenterChildren;
      let inbus = false;
      if (parent !== null) {
        pa = parent.angle;
        palign = parent.alignment;
        inbus = this.isBusAlignment(palign);
      }
      const vertical = (a === 90 || a === 270);
      const pvertical = (pa === 90 || pa === 270);
      const leaf = (v.childrenCount === 0);
      const left = this.isLeftSideBus(v);
      let f = 0.0;
      const comments = v.comments;
      const commentsLength = comments.length;
      const cloc = Point.alloc();
      for (let i = 0; i < commentsLength; i++) {
        const c = comments[i];
        const cbounds = c.measuredBounds;
        if ((vertical && !leaf) || (!inbus && !pvertical && leaf) || (inbus && pvertical && leaf)) {
          if ((pa > 135 && !inbus) || (pvertical && left)) {
            if (v.commentMargin >= 0) {
              cloc.assignTo(v.bounds.x - v.commentMargin - cbounds.width, v.bounds.y + f);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l1 = lit.value;
                l1.fromSpot = Spot.MiddleLeft;
                l1.toSpot = Spot.MiddleRight;
              }
            } else {
              cloc.assignTo(v.bounds.x + v.focus.x * 2 - v.commentMargin, v.bounds.y + f);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l2 = lit.value;
                l2.fromSpot = Spot.MiddleRight;
                l2.toSpot = Spot.MiddleLeft;
              }
            }
          } else {
            if (v.commentMargin >= 0) {
              cloc.assignTo(v.bounds.x + v.focus.x * 2 + v.commentMargin, v.bounds.y + f);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l3 = lit.value;
                l3.fromSpot = Spot.MiddleRight;
                l3.toSpot = Spot.MiddleLeft;
              }
            } else {
              cloc.assignTo(v.bounds.x + v.commentMargin - cbounds.width, v.bounds.y + f);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l4 = lit.value;
                l4.fromSpot = Spot.MiddleLeft;
                l4.toSpot = Spot.MiddleRight;
              }
            }
          }
          if (v.commentSpacing >= 0) {
            f += cbounds.height + v.commentSpacing;
          } else {
            f += v.commentSpacing - cbounds.height;
          }
        } else {
          if ((pa > 135 && !inbus) || (!pvertical && left)) {
            if (v.commentMargin >= 0) {
              cloc.assignTo(v.bounds.x + f, v.bounds.y - v.commentMargin - cbounds.height);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l5 = lit.value;
                l5.fromSpot = Spot.MiddleTop;
                l5.toSpot = Spot.MiddleBottom;
              }
            } else {
              cloc.assignTo(v.bounds.x + f, v.bounds.y + v.focus.y * 2 - v.commentMargin);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l6 = lit.value;
                l6.fromSpot = Spot.MiddleBottom;
                l6.toSpot = Spot.MiddleTop;
              }
            }
          } else {
            if (v.commentMargin >= 0) {
              cloc.assignTo(v.bounds.x + f, v.bounds.y + v.focus.y * 2 + v.commentMargin);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l7 = lit.value;
                l7.fromSpot = Spot.MiddleBottom;
                l7.toSpot = Spot.MiddleTop;
              }
            } else {
              cloc.assignTo(v.bounds.x + f, v.bounds.y + v.commentMargin - cbounds.height);
              c.move(cloc);
              const lit = c.findLinksInto();
              while (lit.next()) {
                const l8 = lit.value;
                l8.fromSpot = Spot.MiddleTop;
                l8.toSpot = Spot.MiddleBottom;
              }
            }
          }
          if (v.commentSpacing >= 0) {
            f += cbounds.width + v.commentSpacing;
          } else {
            f += v.commentSpacing - cbounds.width;
          }
        }
      }
      Point.free(cloc);
      // if the comments stick down beyound the bottom of the node,
      // make sure the child links avoid the comments
      const diff = f - v.commentSpacing - (vertical ? nodebounds.height : nodebounds.width);
      if (this._effectivePath === TreeLayout.PathDestination) {
        const destEdges = v.destinationEdges;
        while (destEdges.next()) {
          const dedge = destEdges.value;
          const link = dedge.link;
          if (link !== null && !link.isAvoiding) {
            link.fromEndSegmentLength = (diff > 0 ? diff : NaN);
          }
        }
      } else { // the path is TreeLayout.PathSource
        const sourceEdges = v.sourceEdges;
        while (sourceEdges.next()) {
          const sedge = sourceEdges.value;
          const link = sedge.link;
          if (link !== null && !link.isAvoiding) {
            link.toEndSegmentLength = (diff > 0 ? diff : NaN);
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private layoutAllTrees(): void {
    if (this.layerStyle === TreeLayout.LayerUniform) {
      // this array remembers the needed depth for each layer of nodes
      const alldepths: Array<number> = [];
      // find out how deep each layer must be
      let vit = this.network.vertexes.iterator;
      while (vit.next()) {
        const t = vit.value as TreeVertex;
        let parent = t.parent;
        if (parent === null) parent = t;
        const horiz = (parent.angle === 0 || parent.angle === 180);
        let d = alldepths[t.level];
        if (d === undefined) d = 0;
        alldepths[t.level] = Math.max(d, (horiz ? t.width : t.height));
      }
      for (let i = 0; i < alldepths.length; i++) if (alldepths[i] === undefined) alldepths[i] = 0;
      this._allDepths = alldepths;
      // assign all TreeVertex.width or .height values
      vit = this.network.vertexes.iterator;
      while (vit.next()) {
        const t1 = vit.value as TreeVertex;
        let parent = t1.parent;
        if (parent === null) parent = t1;
        if (parent.angle === 0 || parent.angle === 180) {
          if (parent.angle === 180) t1.focusX += (alldepths[t1.level] - t1.width);
          t1.width = alldepths[t1.level];
        } else {
          if (parent.angle === 270) t1.focusY += (alldepths[t1.level] - t1.height);
          t1.height = alldepths[t1.level];
        }
      }
    } else if (this.layerStyle === TreeLayout.LayerSiblings) {
      const vit = this.network.vertexes.iterator;
      while (vit.next()) {
        const st = vit.value as TreeVertex;
        const horiz = (st.angle === 0 || st.angle === 180);
        let depth = -1;
        for (let i = 0; i < st.children.length; i++) {
          const child = st.children[i];
          depth = Math.max(depth, (horiz ? child.width : child.height));
        }
        if (depth >= 0) {
          for (let i = 0; i < st.children.length; i++) {
            const child = st.children[i];
            if (horiz) {
              if (st.angle === 180) child.focusX += (depth - child.width);
              child.width = depth;
            } else {
              if (st.angle === 270) child.focusY += (depth - child.height);
              child.height = depth;
            }
          }
        }
      }
    }
    const rit = this._roots.iterator;
    while (rit.next()) {
      const v = rit.value;
      if (v instanceof TreeVertex) this.layoutTree(v);
    }
  }

  /**
   * @hidden @internal
   * Recursively lay out a subtree starting with the given parent node.
   *
   * This method should walk the tree recursively.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @since 1.4
   */
  public layoutTree(v: TreeVertex): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutTree:v');
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      this.layoutTree(child);
    }
    switch (v.compaction) {
      case TreeLayout.CompactionNone:
        this.layoutTreeNone(v);
        break;
      case TreeLayout.CompactionBlock:
        if (v.alignment === TreeLayout.AlignmentBusBranching) {
          this.layoutTreeNone(v); // simpler, faster, and less buggy
        } else {
          this.layoutTreeBlock(v);
        }
        break;
      default:
        Util.throwError('Unhandled compaction value ' + v.compaction.toString());
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private layoutTreeNone(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutTreeNone:v');
    if (v.childrenCount === 0) {
      let inbus = false;
      let pa = 0.0;
      let palign = TreeLayout.AlignmentCenterChildren;
      if (v.parent !== null) {
        pa = v.parent.angle;
        palign = v.parent.alignment;
        inbus = this.isBusAlignment(palign);
      }
      const left = this.isLeftSideBus(v);
      v._relativePosition.assignTo(0, 0);
      v._subtreeSize.assignTo(v.width, v.height);
      // if a comment goes in a negative direction, offset the node to account for the extra space needed for the subwidth/subheight
      if (v.parent !== null && v.comments !== null && (((pa === 180 || pa === 270) && !inbus) || left)) {
        if ((pa === 180 && !inbus) || ((pa === 90 || pa === 270) && left)) {
          v._subtreeOffset.assignTo(v.width - v.focus.x * 2, 0);
        } else {
          v._subtreeOffset.assignTo(0, v.height - v.focus.y * 2);
        }
      } else {
        v._subtreeOffset.assignTo(0, 0);
      }
      return;
    }
    const angle = this.orthoAngle(v);
    const vertical = (angle === 90 || angle === 270);
    let maxsubbreadth = 0.0;
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      maxsubbreadth = Math.max(maxsubbreadth, (vertical ? child._subtreeSize.width : child._subtreeSize.height));
    }
    const align = v.alignment;
    const alignstart = (align === TreeLayout.AlignmentStart);
    const alignend = (align === TreeLayout.AlignmentEnd);
    const alignbus = this.isBusAlignment(align);
    const limit = Math.max(0, v.breadthLimit);
    const layerspacing = this.computeLayerSpacing(v);  // limit layerspacing to be no more negative than the depth of the node
    const nodespacing = v.nodeSpacing;
    const nodeindent = this.computeNodeIndent(v);
    const initialspacing = ((alignstart || alignend) ? 0 : nodeindent / 2);
    const rowspacing = v.rowSpacing;
    let rowindent = 0.0;
    if (alignstart || alignend || v.routeAroundCentered || (v.routeAroundLastParent && v.maxGenerationCount === 1)) {
      rowindent = Math.max(0, v.rowIndent);
    }
    const vWidth = v.width;
    const vHeight = v.height;
    let subwidth = 0.0;
    let subheight = 0.0;
    let olddepth = 0.0;
    let x = 0.0;
    let y = 0.0;
    let row = 0;
    let col = 0;
    let firstidx = 0;
    let maxsubhalf = 0.0;
    let maxsubhalf2 = 0.0;
    if (alignbus && !this.isBusTwoSided(align)) {
      if (angle > 135) children.reverse();
    }
    if (this.isBusTwoSided(align)) {
      if (childrenLength > 1) {
        for (let i = 0; i < childrenLength; i++) {
          const child = children[i];
          const childSubSize = child._subtreeSize;
          if (i % 2 === 0 && i !== childrenLength - 1) {
            // for a bus, remember the subwidth of one half so we know where to offset the parent
            // we don't include the last odd child because it will be in the middle
            maxsubhalf = Math.max(maxsubhalf,
              (vertical ? childSubSize.width : childSubSize.height)
              + this.computeBusNodeSpacing(child) - nodespacing);
          } else if (i % 2 !== 0) {
            maxsubhalf2 = Math.max(maxsubhalf2,
              (vertical ? childSubSize.width : childSubSize.height)
              + this.computeBusNodeSpacing(child) - nodespacing);
          }
        }
      } else if (childrenLength === 1) {
        maxsubhalf = (vertical ? children[0]._subtreeSize.width : children[0]._subtreeSize.height);
      }
    }
    // lay out nodes in a bus
    if (alignbus) {
      switch (align) {
        case TreeLayout.AlignmentBus:
        case TreeLayout.AlignmentBusBranching:
          let rvalues: Rect;  // fake passing by reference by returning an array
          if (angle < 135) rvalues = this.layoutBusChildrenPosDir(v, children, maxsubhalf, maxsubhalf2, subwidth, subheight);
          else rvalues = this.layoutBusChildrenNegDir(v, children, maxsubhalf, maxsubhalf2, subwidth, subheight);
          maxsubhalf = rvalues.x;
          subwidth = rvalues.width;
          subheight = rvalues.height;
          break;
        case TreeLayout.AlignmentTopLeftBus:
          for (let i = 0; i < childrenLength; i++) {
            const child = children[i];
            const childSubSize = child._subtreeSize;
            const rspace = (col === 0 ? 0 : rowspacing);
            if (vertical) {
              // align the right edges of the vertexes so the distance to the parent's center is constant
              child._relativePosition.assignTo(maxsubbreadth - childSubSize.width, y + rspace);
              subwidth = Math.max(subwidth, childSubSize.width);
              subheight = Math.max(subheight, y + rspace + childSubSize.height);
              y += rspace + childSubSize.height;
            } else {
              // align the bottom edges of the vertexes so the distance to the parent's center is constant
              child._relativePosition.assignTo(x + rspace, maxsubbreadth - childSubSize.height);
              subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
              subheight = Math.max(subheight, childSubSize.height);
              x += rspace + childSubSize.width;
            }
            col++;
          }
          break;
        case TreeLayout.AlignmentBottomRightBus:
          for (let l = 0; l < childrenLength; l++) {
            const child = children[l];
            const childSubSize = child._subtreeSize;
            const rspace = (col === 0 ? 0 : rowspacing);
            if (vertical) {
              // set the child's RelativePosition offset from the parent, because we don't want a negative offset later
              child._relativePosition.assignTo(nodespacing / 2 + v.focus.x, y + rspace);
              subwidth = Math.max(subwidth, childSubSize.width);
              subheight = Math.max(subheight, y + rspace + childSubSize.height);
              y += rspace + childSubSize.height;
            } else {
              child._relativePosition.assignTo(x + rspace, nodespacing / 2 + v.focus.y);
              subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
              subheight = Math.max(subheight, childSubSize.height);
              x += rspace + childSubSize.width;
            }
            col++;
          }
          break;
      }
    } else {  // children not in a bus
      for (let idx = 0; idx < childrenLength; idx++) {
        const child = children[idx];
        const childSubSize = child._subtreeSize;
        if (vertical) {
          if (limit > 0 && col > 0 && x + nodespacing + childSubSize.width > limit) {  // need new row
            if (x < maxsubbreadth) {  // align the previous row
              this.shiftRelPosAlign(v, align, maxsubbreadth - x, 0, firstidx, idx - 1);
            }
            row++;
            col = 0;  // start new row
            firstidx = idx;  // remember index of first child in this row
            olddepth = subheight;  // remember current depth for this row
            x = 0;
            y = (angle > 135 ? -subheight - rowspacing : subheight + rowspacing);
          }
          const nspace = (col === 0 ? initialspacing : nodespacing);
          this.recordMidPoints(child, 0, y);  // for links being routed around rows
          child._relativePosition.assignTo(x + nspace, y);
          subwidth = Math.max(subwidth, x + nspace + childSubSize.width);
          subheight = Math.max(subheight, olddepth + (row === 0 ? 0 : rowspacing) + childSubSize.height);
          x += nspace + childSubSize.width;
        } else {  // horizontal growth direction
          if (limit > 0 && col > 0 && y + nodespacing + childSubSize.height > limit) {  // need new row
            if (y < maxsubbreadth) {  // align the previous row
              this.shiftRelPosAlign(v, align, 0, maxsubbreadth - y, firstidx, idx - 1);
            }
            row++;
            col = 0;  // start new row
            firstidx = idx;
            olddepth = subwidth;  // remember current depth for this row
            y = 0;
            x = (angle > 135 ? -subwidth - rowspacing : subwidth + rowspacing);
          }
          const nspace = (col === 0 ? initialspacing : nodespacing);
          this.recordMidPoints(child, x, 0);
          child._relativePosition.assignTo(x, y + nspace);
          subheight = Math.max(subheight, y + nspace + childSubSize.height);
          subwidth = Math.max(subwidth, olddepth + (row === 0 ? 0 : rowspacing) + childSubSize.width);
          y += nspace + childSubSize.height;
        }
        col++;
      }
    }
    if (row > 0) { // if there were multiple rows
      if (vertical) {
        subheight += Math.max(0, layerspacing);  // leave extra room past last row
        if (x < subwidth) {  // align the last row
          this.shiftRelPosAlign(v, align, subwidth - x, 0, firstidx, childrenLength - 1);
        }
        if (rowindent > 0) {
          if (!alignend) this.shiftRelPos(v, rowindent, 0, 0, childrenLength - 1);
          subwidth += rowindent;
        }
      } else {
        subwidth += Math.max(0, layerspacing);  // leave extra room past last row
        if (y < subheight) {  // align the last row
          this.shiftRelPosAlign(v, align, 0, subheight - y, firstidx, childrenLength - 1);
        }
        if (rowindent > 0) {
          if (!alignend) this.shiftRelPos(v, 0, rowindent, 0, childrenLength - 1);
          subheight += rowindent;
        }
      }
    }
    // align this parent node relative to all of its children/descendents
    let offsetx = 0.0;
    let offsety = 0.0;
    switch (align) {
      case TreeLayout.AlignmentCenterSubtrees:
        if (vertical) {
          offsetx += subwidth / 2 - v.focus.x - nodeindent / 2;
        } else {
          offsety += subheight / 2 - v.focus.y - nodeindent / 2;
        }
        break;
      case TreeLayout.AlignmentCenterChildren:
        if (row > 0) {  // same as CenterSubtrees when there's multiple rows
          if (vertical) {
            offsetx += subwidth / 2 - v.focus.x - nodeindent / 2;
          } else {
            offsety += subheight / 2 - v.focus.y - nodeindent / 2;
          }
        } else {
          const num = childrenLength;
          if (vertical) {
            const first = children[0]._relativePosition.x + children[0]._subtreeOffset.x;
            const last = children[num - 1]._relativePosition.x + children[num - 1]._subtreeOffset.x + children[num - 1].focus.x * 2;
            offsetx += first + (last - first) / 2 - v.focus.x - nodeindent / 2;
          } else {
            const first = children[0]._relativePosition.y + children[0]._subtreeOffset.y;
            const last = children[num - 1]._relativePosition.y + children[num - 1]._subtreeOffset.y + children[num - 1].focus.y * 2;
            offsety += first + (last - first) / 2 - v.focus.y - nodeindent / 2;
          }
        }
        break;
      case TreeLayout.AlignmentStart:
        if (vertical) {
          offsetx -= nodeindent;
          subwidth += nodeindent;
        } else {
          offsety -= nodeindent;
          subheight += nodeindent;
        }
        break;
      case TreeLayout.AlignmentEnd:
        if (vertical) {
          offsetx += (subwidth - v.width) + nodeindent;
          subwidth += nodeindent;
        } else {
          offsety += (subheight - v.height) + nodeindent;
          subheight += nodeindent;
        }
        break;
      case TreeLayout.AlignmentBus:
      case TreeLayout.AlignmentBusBranching:
        if (vertical) {
          if (childrenLength > 1) {
            offsetx += maxsubhalf + nodespacing / 2 - v.focus.x;
          } else { // with only one child, offset just like CenterChildren
            offsetx += children[0].focus.x - v.focus.x + children[0]._subtreeOffset.x;
          }
        } else {
          if (childrenLength > 1) {
            offsety += maxsubhalf + nodespacing / 2 - v.focus.y;
          } else {
            offsety += children[0].focus.y - v.focus.y + children[0]._subtreeOffset.y;
          }
        }
        break;
      case TreeLayout.AlignmentTopLeftBus:
        if (vertical) {
          offsetx += subwidth + nodespacing / 2 - v.focus.x;
        } else {
          offsety += subheight + nodespacing / 2 - v.focus.y;
        }
        break;
      case TreeLayout.AlignmentBottomRightBus:
        break;
      default:
        Util.throwError('Unhandled alignment value ' + align.toString());
    }
    // fix up all children's RelativePositions relative to the desired position of this parent node
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      if (vertical) {
        child._relativePosition.assignTo(
          child._relativePosition.x + child._subtreeOffset.x - offsetx,
          child._relativePosition.y + ((angle > 135) ?
            ((alignbus ? -subheight : -child._subtreeSize.height) + child._subtreeOffset.y - layerspacing) :
            (vHeight + layerspacing + child._subtreeOffset.y)));
      } else {
        child._relativePosition.assignTo(
          child._relativePosition.x + ((angle > 135) ?
            ((alignbus ? -subwidth : -child._subtreeSize.width) + child._subtreeOffset.x - layerspacing) :
            (vWidth + layerspacing + child._subtreeOffset.x)),
          child._relativePosition.y + child._subtreeOffset.y - offsety);
      }
    }
    // include the NODE itself in the subtree size, and adjust the offset if it is negative
    if (vertical) {
      subwidth = this.calculateSubwidth(v, subwidth, offsetx);
      // maybe the alignment caused the parent node to stick out
      if (offsetx < 0) offsetx = 0; // subwidth has already been compensated
      if (angle > 135) offsety += subheight + layerspacing;
      subheight += vHeight + layerspacing;
    } else {
      if (angle > 135) offsetx += subwidth + layerspacing;
      subwidth += vWidth + layerspacing;
      subheight = this.calculateSubheight(v, subheight, offsety);
      // maybe the alignment caused the parent node to stick out
      if (offsety < 0) offsety = 0; // subheight has already been compensated
    }
    // save results
    v._subtreeOffset.assignTo(offsetx, offsety);
    v._subtreeSize.assignTo(subwidth, subheight);
  }

  /**
   * @hidden @internal
   * For a positive direction, we will set up the first half of the children,
   * then the second half, and finally the last odd child, if there is one.
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {Array.<TreeVertex>} children
   * @param {number} maxsubhalf
   * @param {number} maxsubhalf2
   * @param {number} subwidth
   * @param {number} subheight
   * @return {Rect}
   */
  private layoutBusChildrenPosDir(v: TreeVertex, children: Array<TreeVertex>, maxsubhalf: number, maxsubhalf2: number, subwidth: number, subheight: number): Rect {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutBusChildrenPosDir:v');
    const childrenLength = children.length;
    if (childrenLength === 0) {
      return new Rect(maxsubhalf, 0, subwidth, subheight);
    }
    if (childrenLength === 1) {
      const child = children[0];
      subwidth = child._subtreeSize.width;
      subheight = child._subtreeSize.height;
      return new Rect(maxsubhalf, 0, subwidth, subheight);
    }
    const nodespacing = v.nodeSpacing;
    const rowspacing = v.rowSpacing;
    const angle = this.orthoAngle(v);
    const vertical = (angle === 90);
    let col = 0;
    let x = 0.0;
    let y = 0.0;
    // keep track of the area taken by the first half of the children so we know where
    // to place the last odd child later
    // o-|-o
    // | |-o
    // o |
    // o-|
    // for instance, above, ysave would be a number greater than y after setting up the
    // second half of the children, so we will use it when positioning the last odd child
    // set RelativePositions for the first half of the children
    for (let i = 0; i < childrenLength; i++) {
      // if odd number of children, set up the last one later
      if (i % 2 !== 0 || (childrenLength > 1 && i === childrenLength - 1)) continue;
      const child = children[i];
      const childSubSize = child._subtreeSize;
      const rspace = (col === 0 ? 0 : rowspacing);
      if (vertical) {
        // align the right edges of the vertexes so the distances between the two sides of the bus are constant
        const dx = this.computeBusNodeSpacing(child) - nodespacing;
        child._relativePosition.assignTo(maxsubhalf - (childSubSize.width + dx), y + rspace);
        subwidth = Math.max(subwidth, childSubSize.width + dx);
        subheight = Math.max(subheight, y + rspace + childSubSize.height);
        y += rspace + childSubSize.height;
      } else {
        // align the bottom edges of the vertexes so the distances between the two sides of the bus are constant
        const dy = this.computeBusNodeSpacing(child) - nodespacing;
        child._relativePosition.assignTo(x + rspace, maxsubhalf - (childSubSize.height + dy));
        subheight = Math.max(subheight, childSubSize.height + dy);
        subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
        x += rspace + childSubSize.width;
      }
      col++;
    }
    // now adjust x and y for the second half of the children
    col = 0;
    const xsave = x;  // save x and y for placing the last odd child
    const ysave = y;
    if (vertical) {
      x = maxsubhalf + nodespacing;
      y = 0;
    } else {
      x = 0;
      y = maxsubhalf + nodespacing;
    }
    for (let i = 0; i < childrenLength; i++) {
      if (i % 2 === 0) continue;
      const child = children[i];
      const childSubSize = child._subtreeSize;
      const rspace = (col === 0 ? 0 : rowspacing);
      if (vertical) {
        const dx = this.computeBusNodeSpacing(child) - nodespacing;
        child._relativePosition.assignTo(x + dx, y + rspace);
        subwidth = Math.max(subwidth, x + childSubSize.width + dx);
        subheight = Math.max(subheight, y + rspace + childSubSize.height);
        y += rspace + childSubSize.height;
      } else {
        const dy = this.computeBusNodeSpacing(child) - nodespacing;
        child._relativePosition.assignTo(x + rspace, y + dy);
        subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
        subheight = Math.max(subheight, y + childSubSize.height + dy);
        x += rspace + childSubSize.width;
      }
      col++;
    }
    // set up the last odd child
    if (childrenLength > 1 && childrenLength % 2 === 1) {
      const child = children[childrenLength - 1];
      const childSubSize = child._subtreeSize;
      const rowsp = this.computeBusLastRowSpacing(child, vertical ? Math.max(Math.abs(ysave), Math.abs(y)) : Math.max(Math.abs(xsave), Math.abs(x)));
      // newmsh is used when we have a last odd child that extends beyond maxsubhalf
      //  o-|-o
      // o--o
      // in this case, maxsubhalf is not correct because it only takes into account the first half of the children
      // and doesn't include the last odd child
      if (vertical) {
        child._relativePosition.assignTo(maxsubhalf + nodespacing / 2 - child.focus.x - child._subtreeOffset.x, subheight + rowsp);
        const min = maxsubhalf + nodespacing / 2 - child.focus.x - child._subtreeOffset.x;
        subwidth = Math.max(subwidth, min + childSubSize.width);
        if (min < 0) subwidth -= min;
        subheight = Math.max(subheight, Math.max(ysave, y) + rowsp + childSubSize.height);
        // if the child wants a negative RelativePosition, move all children and make it 0
        if (child._relativePosition.x < 0) maxsubhalf = this.fixRelativePositions(v, child._relativePosition.x, false, maxsubhalf, nodespacing);
      } else {
        child._relativePosition.assignTo(subwidth + rowsp, maxsubhalf + nodespacing / 2 - child.focus.y - child._subtreeOffset.y);
        subwidth = Math.max(subwidth, Math.max(xsave, x) + rowsp + childSubSize.width);
        const min = maxsubhalf + nodespacing / 2 - child.focus.y - child._subtreeOffset.y;
        subheight = Math.max(subheight, min + childSubSize.height);
        if (min < 0) subheight -= min;
        // if the child wants a negative RelativePosition, move all the children and make it 0
        if (child._relativePosition.y < 0) maxsubhalf = this.fixRelativePositions(v, child._relativePosition.y, true, maxsubhalf, nodespacing);
      }
    }
    return new Rect(maxsubhalf, 0, subwidth, subheight);
  }

  /**
   * @hidden @internal
   * For a negative direction, we will set up the first half of the children,
   * then the second half, and finally the last odd child, if there is one.
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {Array.<TreeVertex>} children
   * @param {number} maxsubhalf
   * @param {number} maxsubhalf2
   * @param {number} subwidth
   * @param {number} subheight
   * @return {Rect}
   */
  private layoutBusChildrenNegDir(v: TreeVertex, children: Array<TreeVertex>, maxsubhalf: number, maxsubhalf2: number, subwidth: number, subheight: number): Rect {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutBusChildrenNegDir:v');
    const childrenLength = children.length;
    if (childrenLength === 0) {
      return new Rect(maxsubhalf, 0, subwidth, subheight);
    }
    if (childrenLength === 1) {
      const child = children[0];
      subwidth = child._subtreeSize.width;
      subheight = child._subtreeSize.height;
      return new Rect(maxsubhalf, 0, subwidth, subheight);
    }
    const nodespacing = v.nodeSpacing;
    const rowspacing = v.rowSpacing;
    const angle = this.orthoAngle(v);
    const vertical = (angle === 270);
    let col = 0;
    let x = 0.0;
    let y = 0.0;
    // keep track of the area taken by the first half of the children so we know where
    // to place the last odd child later
    // o-|-o
    // | |-o
    // o |
    // o-|
    // for instance, above, ysave would be a number greater than y after setting up the
    // second half of the children, so we will use it when positioning the last odd child
    // set RelativePositions for the first half of the children
    for (let i = 0; i < childrenLength; i++) {
      // if odd number of children, set up the last one later
      if (i % 2 !== 0 || (childrenLength > 1 && i === childrenLength - 1)) continue;
      const child = children[i];
      const childSubSize = child._subtreeSize;
      const rspace = (col === 0 ? 0 : rowspacing);
      if (vertical) {
        // align the right edges of the vertexes so the distances between two sides of the bus are constant
        const dx = this.computeBusNodeSpacing(child) - nodespacing;
        y -= rspace + childSubSize.height;
        child._relativePosition.assignTo(maxsubhalf - (childSubSize.width + dx), y);
        subwidth = Math.max(subwidth, childSubSize.width + dx);
        subheight = Math.max(subheight, Math.abs(y));
      } else {
        // align the bottom edges of the vertexes so the distance between the two sides of the bus are constant
        const dy = this.computeBusNodeSpacing(child) - nodespacing;
        x -= rspace + childSubSize.width;
        child._relativePosition.assignTo(x, maxsubhalf - (childSubSize.height + dy));
        subheight = Math.max(subheight, childSubSize.height + dy);
        subwidth = Math.max(subwidth, Math.abs(x));
      }
      col++;
    }
    // now adjust x and y for the second half of the children
    col = 0;
    const xsave = x;  // save x and y for placing the last odd childe
    const ysave = y;
    if (vertical) {
      x = maxsubhalf + nodespacing;
      y = 0;
    } else {
      x = 0;
      y = maxsubhalf + nodespacing;
    }
    for (let i = 0; i < childrenLength; i++) {
      if (i % 2 === 0) continue;
      const child = children[i];
      const childSubSize = child._subtreeSize;
      const rspace = (col === 0 ? 0 : rowspacing);
      if (vertical) {
        const dx = this.computeBusNodeSpacing(child) - nodespacing;
        y -= rspace + childSubSize.height;
        child._relativePosition.assignTo(x + dx, y);
        subwidth = Math.max(subwidth, x + childSubSize.width + dx);
        subheight = Math.max(subheight, Math.abs(y));
      } else {
        const dy = this.computeBusNodeSpacing(child) - nodespacing;
        x -= rspace + childSubSize.width;
        child._relativePosition.assignTo(x, y + dy);
        subheight = Math.max(subheight, y + childSubSize.height + dy);
        subwidth = Math.max(subwidth, Math.abs(x));
      }
      col++;
    }
    // set up the last odd child
    if (childrenLength > 1 && childrenLength % 2 === 1) {
      const child = children[childrenLength - 1];
      const childSubSize = child._subtreeSize;
      const rowsp = this.computeBusLastRowSpacing(child, vertical ? Math.max(Math.abs(ysave), Math.abs(y)) : Math.max(Math.abs(xsave), Math.abs(x)));
      // newmsh is used when we have a last odd child that extends beyond maxsubhalf
      //  o-|-o
      // o--o
      // in this case, maxsubhalf is not correct because it only takes into account the first half of the children
      // and doesn't include the last odd child
      if (vertical) {
        child._relativePosition.assignTo(maxsubhalf + nodespacing / 2 - child.focus.x - child._subtreeOffset.x, -subheight - childSubSize.height - rowsp);
        const min = maxsubhalf + nodespacing / 2 - child.focus.x - child._subtreeOffset.x;
        subwidth = Math.max(subwidth, min + childSubSize.width);
        if (min < 0) subwidth -= min;
        subheight = Math.max(subheight, Math.abs(Math.min(ysave, y)) + rowsp + childSubSize.height);
        // if the child wants a negative RelativePosition, move all the children to make it 0
        if (child._relativePosition.x < 0) maxsubhalf = this.fixRelativePositions(v, child._relativePosition.x, false, maxsubhalf, nodespacing);
      } else {
        child._relativePosition.assignTo(-subwidth - childSubSize.width - rowsp, maxsubhalf + nodespacing / 2 - child.focus.y - child._subtreeOffset.y);
        subwidth = Math.max(subwidth, Math.abs(Math.min(xsave, x)) + rowsp + childSubSize.width);
        const min = maxsubhalf + nodespacing / 2 - child.focus.y - child._subtreeOffset.y;
        subheight = Math.max(subheight, min + childSubSize.height);
        if (min < 0) subheight -= min;
        // if the child wants a negative RelativePosition, move all the children to make it 0
        if (child._relativePosition.y < 0) maxsubhalf = this.fixRelativePositions(v, child._relativePosition.y, true, maxsubhalf, nodespacing);
      }
    }
    // now normalize all RelativePositions to be non-negative
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      if (vertical) {
        child._relativePosition.assignTo(child._relativePosition.x, child._relativePosition.y + subheight);
      } else {
        child._relativePosition.assignTo(child._relativePosition.x + subwidth, child._relativePosition.y);
      }
    }
    return new Rect(maxsubhalf, 0, subwidth, subheight);
  }

  /**
   * @hidden @internal
   * Get the amount of space to insert between a child vertex and its parent spine
   * when the parent's {@link TreeVertex#alignment} is Bus or BusBranching.
   *
   * This function is called on each child vertex to determine the initial distance of a child
   * from the spine that connects to its parent vertex.
   * This function is not called on the last odd child
   * because that vertex is placed at the end of the spine rather than to either side.
   * @this {TreeLayout}
   * @param {TreeVertex} child
   * @return {number} By default this returns the parent's {@link TreeVertex#nodeSpacing};
   * a small value may result in unpleasant routing of links,
   * a negative value may cause the node overlaps.
   */
  private computeBusNodeSpacing(child: TreeVertex): number {
    if (Debug) Util.checkClass(child, TreeVertex, TreeLayout, 'fixRelativePostions:child');
    if (child.parent === null) return 0;
    return child.parent.nodeSpacing;
  }

  /**
   * @hidden @internal
   * Get the amound of space to leave between the sibling child vertexes and this last odd vertex
   * when the parent's {@link TreeVertex#alignment} is Bus or BusBranching.
   *
   * This function is called only when there is an odd number of child vertexes.
   * This function is not called when the parent vertex only has one child.
   * In that case there is no spine separating the two columns of child vertexes.
   * @this {TreeLayout}
   * @param {TreeVertex} lastchild
   * @param {number} dist the length of the central line (spine) separating the two sides,
   * so far excluding the initial {@link TreeVertex#nodeSpacing}
   * that extends form the parent's spine.
   * @return {number} By default this returns the parent's {@link TreeVertex#rowSpacing};
   * a negative value may cause node overlaps.
   */
  private computeBusLastRowSpacing(lastchild: TreeVertex, dist: number): number {
    if (Debug) Util.checkClass(lastchild, TreeVertex, TreeLayout, 'fixRelativePostions:lastchild');
    if (lastchild.parent === null) return 0;
    return lastchild.parent.rowSpacing;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} offset
   * @param {boolean} vertical
   * @param {number} maxsubhalf
   * @param {number} nodespacing
   * @return {number}
   */
  private fixRelativePositions(v: TreeVertex, offset: number, vertical: boolean, maxsubhalf: number, nodespacing: number): number {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'fixRelativePostions:v');
    const children = v.children;
    const childrenLength = children.length;
    // move all children by the offset so none have negative RelativePosition values
    for (let i = 0; i < childrenLength; i++) {
      if (vertical) {
        children[i]._relativePosition.assignTo(children[i]._relativePosition.x, children[i]._relativePosition.y - offset);
      } else {
        children[i]._relativePosition.assignTo(children[i]._relativePosition.x - offset, children[i]._relativePosition.y);
      }
    }
    const lastchild = children[childrenLength - 1];
    return Math.max(maxsubhalf, (vertical ? lastchild._subtreeOffset.y + lastchild.focus.y - nodespacing / 2 :
      lastchild._subtreeOffset.x + lastchild.focus.x - nodespacing / 2));
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} subwidth
   * @param {number} offset
   * @return {number}
   */
  private calculateSubwidth(v: TreeVertex, subwidth: number, offset: number): number {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'calculateSubwidth:v');
    // when comments are involved, make sure they are accounted for in subwidth if they extend beyond the edges of the subtree
    switch (v.alignment) {
      case TreeLayout.AlignmentCenterChildren:
      case TreeLayout.AlignmentCenterSubtrees:
        let w = subwidth;
        if (offset + v.width > w) w = offset + v.width;
        if (offset < 0) w -= offset; // increases w
        return w;
      case TreeLayout.AlignmentStart:
        if (v.width > subwidth) return v.width;
        return subwidth;
      case TreeLayout.AlignmentEnd:
        if (v.focus.x * 2 > subwidth) return v.width;
        return subwidth + v.width - v.focus.x * 2;
      case TreeLayout.AlignmentBus:
      case TreeLayout.AlignmentBusBranching:
        const min = Math.min(0, offset);
        const max = Math.max(subwidth, offset + v.width);
        return Math.max(v.width, max - min);
      case TreeLayout.AlignmentTopLeftBus:
        return v.width - v.focus.x + v.nodeSpacing / 2 + subwidth;
      case TreeLayout.AlignmentBottomRightBus:
        return Math.max(v.width, v.focus.x + v.nodeSpacing / 2 + subwidth);
      default:
        return subwidth;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} subheight
   * @param {number} offset
   * @return {number}
   */
  private calculateSubheight(v: TreeVertex, subheight: number, offset: number): number {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'calculateSubheight:v');
    // when comments are involved, make sure they are accounted for in subheight if they extend beyond the edges of the subtree
    switch (v.alignment) {
      case TreeLayout.AlignmentCenterChildren:
      case TreeLayout.AlignmentCenterSubtrees:
        let h = subheight;
        if (offset + v.height > h) h = offset + v.height;
        if (offset < 0) h -= offset; // increases h
        return h;
      case TreeLayout.AlignmentStart:
        if (v.height > subheight) return v.height;
        return subheight;
      case TreeLayout.AlignmentEnd:
        if (v.focus.y * 2 > subheight) return v.height;
        return subheight + v.height - v.focus.y * 2;
      case TreeLayout.AlignmentBus:
      case TreeLayout.AlignmentBusBranching:
        const min = Math.min(0, offset);
        const max = Math.max(subheight, offset + v.height);
        return Math.max(v.height, max - min);
      case TreeLayout.AlignmentTopLeftBus:
        return v.height - v.focus.y + v.nodeSpacing / 2 + subheight;
      case TreeLayout.AlignmentBottomRightBus:
        return Math.max(v.height, v.focus.y + v.nodeSpacing / 2 + subheight);
      default:
        return subheight;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {EnumValue} align
   * @param {number} offwidth
   * @param {number} offheight
   * @return {Point}
   */
  private alignOffset(align: EnumValue, offwidth: number, offheight: number): Point {
    if (Debug) Util.checkClass(align, EnumValue, TreeLayout, 'alignOffset:align');
    switch (align) {
      case TreeLayout.AlignmentCenterSubtrees: offwidth /= 2; offheight /= 2; break;
      case TreeLayout.AlignmentCenterChildren: offwidth /= 2; offheight /= 2; break;
      case TreeLayout.AlignmentStart: offwidth = 0; offheight = 0; break;
      case TreeLayout.AlignmentEnd: break;
      default: Util.throwError('Unhandled alignment value ' + align.toString());
    }
    return new Point(offwidth, offheight);
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {EnumValue} align
   * @param {number} offwidth
   * @param {number} offheight
   * @param {number} first
   * @param {number} last
   */
  private shiftRelPosAlign(v: TreeVertex, align: EnumValue, offwidth: number, offheight: number, first: number, last: number): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'shiftRelPosAlign:v');
    if (Debug) Util.checkClass(align, EnumValue, TreeLayout, 'shiftRelPosAlign:align');
    const shift = this.alignOffset(align, offwidth, offheight); // offwidth stored in shift[0], offheight in shift[1]
    this.shiftRelPos(v, shift.x, shift.y, first, last);
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} offwidth
   * @param {number} offheight
   * @param {number} first
   * @param {number} last
   */
  private shiftRelPos(v: TreeVertex, offwidth: number, offheight: number, first: number, last: number): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'shiftRelPos:v');
    if (offwidth === 0 && offheight === 0) return;
    const children = v.children;
    for (let i = first; i <= last; i++) {
      const c = children[i];
      const p = c._relativePosition;  // OK to clobber Point
      p.x += offwidth;
      p.y += offheight;
    }
  }

  /**
   * @hidden @internal
   * Tell TreeEdges their parent node relative coordinates
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} x
   * @param {number} y
   */
  private recordMidPoints(v: TreeVertex, x: number, y: number): void {
    if (Debug) {
      Util.checkClass(v, TreeVertex, TreeLayout, 'recordMidPoints:v');
      Util.checkPrimitive(x, 'number', TreeLayout, 'recordMidPoints:x');
      Util.checkPrimitive(y, 'number', TreeLayout, 'recordMidPoints:y');
    }
    const parent = v.parent;
    switch (this._effectivePath) {
      case TreeLayout.PathDestination: {
        const eit = v.sourceEdges;
        while (eit.next()) {
          const se = eit.value as TreeEdge;
          if (se.fromVertex === parent) {
            se.relativePoint.assignTo(x, y);
          }
        }
        break;
      }
      case TreeLayout.PathSource: {
        const eit = v.destinationEdges;
        while (eit.next()) {
          const de = eit.value as TreeEdge;
          if (de.toVertex === parent) {
            de.relativePoint.assignTo(x, y);
          }
        }
        break;
      }
      default:
        Util.throwError('Unhandled path value ' + this._effectivePath.toString());
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private layoutTreeBlock(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'layoutTreeBlock:v');
    if (v.childrenCount === 0) {
      const parent = v.parent;
      let inbus = false;
      let pa = 0.0;
      let palign = TreeLayout.AlignmentCenterChildren;
      if (parent !== null) {
        pa = parent.angle;
        palign = parent.alignment;
        inbus = this.isBusAlignment(palign);
      }
      const left = this.isLeftSideBus(v);
      v._relativePosition.assignTo(0, 0);
      v._subtreeSize.assignTo(v.width, v.height);
      // if a comment goes in a negative direction, offset the node to account for the extra space needed for the subwidth/subheight
      if (v.parent !== null && v.comments !== null && (((pa === 180 || pa === 270) && !inbus) || left)) {
        if ((pa === 180 && !inbus) || ((pa === 90 || pa === 270) && left)) {
          v._subtreeOffset.assignTo(v.width - v.focus.x * 2, 0);
        } else {
          v._subtreeOffset.assignTo(0, v.height - v.focus.y * 2);
        }
      } else {
        v._subtreeOffset.assignTo(0, 0);
      }
      v.leftFringe = null;
      v.rightFringe = null;
      return;
    }
    const angle = this.orthoAngle(v);
    const vertical = (angle === 90 || angle === 270);
    let maxsubbreadth = 0.0;
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      maxsubbreadth = Math.max(maxsubbreadth, (vertical ? child._subtreeSize.width : child._subtreeSize.height));
    }
    const align = v.alignment;
    const alignstart = (align === TreeLayout.AlignmentStart);
    const alignend = (align === TreeLayout.AlignmentEnd);
    const alignbus = this.isBusAlignment(align);
    const limit = Math.max(0, v.breadthLimit);
    const layerspacing = this.computeLayerSpacing(v);  // limit layerspacing to be no more negative than the depth of the node
    const nodespacing = v.nodeSpacing;
    const nodeindent = this.computeNodeIndent(v);
    const rowspacing = v.rowSpacing;
    let rowindent = 0.0;
    if (alignstart || alignend || v.routeAroundCentered || (v.routeAroundLastParent && v.maxGenerationCount === 1)) {
      rowindent = Math.max(0, v.rowIndent);
    }
    const vWidth = v.width;
    const vHeight = v.height;
    let subwidth = 0.0;
    let subheight = 0.0;
    let olddepth = 0.0;
    let lfringe: Array<Point> = null;
    let rfringe: Array<Point> = null;
    let lsizewidth = 0.0;
    let lsizeheight = 0.0;
    let x = 0.0;
    let y = 0.0;
    let row = 0;
    let col = 0;
    let firstidx = 0;
    let maxsubhalf = 0.0;
    let maxsubhalf2 = 0.0;
    if (alignbus && !this.isBusTwoSided(align)) {
      if (angle > 135) children.reverse();
    }
    if (this.isBusTwoSided(align)) {
      if (childrenLength > 1) {
        for (let i = 0; i < childrenLength; i++) {
          if (i % 2 === 0 && i !== childrenLength - 1) {
            // for a bus, remember the subwidth of one half so we know where to offset the parent
            // we don't include the last odd child because it will be in the middle
            maxsubhalf = Math.max(maxsubhalf, (vertical ? children[i]._subtreeSize.width : children[i]._subtreeSize.height));
          } else if (i % 2 !== 0) {
            maxsubhalf2 = Math.max(maxsubhalf2, (vertical ? children[i]._subtreeSize.width : children[i]._subtreeSize.height));
          }
        }
      } else if (childrenLength === 1) {
        maxsubhalf = (vertical ? children[0]._subtreeSize.width : children[0]._subtreeSize.height);
      }
    }
    // lay out nodes in a bus
    if (alignbus) {
      switch (align) {
        case TreeLayout.AlignmentBus:
          let rvalues: Rect;  // fake passing by reference by returning an array
          if (angle < 135) rvalues = this.layoutBusChildrenPosDir(v, children, maxsubhalf, maxsubhalf2, subwidth, subheight);
          else rvalues = this.layoutBusChildrenNegDir(v, children, maxsubhalf, maxsubhalf2, subwidth, subheight);
          maxsubhalf = rvalues.x;
          subwidth = rvalues.width;
          subheight = rvalues.height;
          break;
        case TreeLayout.AlignmentTopLeftBus:
          for (let i = 0; i < childrenLength; i++) {
            const child = children[i];
            const childSubSize = child._subtreeSize;
            const rspace = (col === 0 ? 0 : rowspacing);
            if (vertical) {
              // align the right edges of the vertexes so the distance to the parent's center is constant
              child._relativePosition.assignTo(maxsubbreadth - childSubSize.width, y + rspace);
              subwidth = Math.max(subwidth, childSubSize.width);
              subheight = Math.max(subheight, y + rspace + childSubSize.height);
              y += rspace + childSubSize.height;
            } else {
              // align the bottom edges of the vertexes so the distance to the parent's center is constant
              child._relativePosition.assignTo(x + rspace, maxsubbreadth - childSubSize.height);
              subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
              subheight = Math.max(subheight, childSubSize.height);
              x += rspace + childSubSize.width;
            }
            col++;
          }
          break;
        case TreeLayout.AlignmentBottomRightBus:
          for (let i = 0; i < childrenLength; i++) {
            const child = children[i];
            const childSubSize = child._subtreeSize;
            const rspace = (col === 0 ? 0 : rowspacing);
            if (vertical) {
              // set the child's RelativePosition offset from the parent, because we don't want a negative offset later
              child._relativePosition.assignTo(nodespacing / 2 + v.focus.x, y + rspace);
              subwidth = Math.max(subwidth, childSubSize.width);
              subheight = Math.max(subheight, y + rspace + childSubSize.height);
              y += rspace + childSubSize.height;
            } else {
              child._relativePosition.assignTo(x + rspace, nodespacing / 2 + v.focus.y);
              subwidth = Math.max(subwidth, x + rspace + childSubSize.width);
              subheight = Math.max(subheight, childSubSize.height);
              x += rspace + childSubSize.width;
            }
            col++;
          }
          break;
      }
      // set up fringes for the bus
      lfringe = this.allocTempPointArray(2);
      rfringe = this.allocTempPointArray(2);
      if (vertical) {
        lfringe[0].assignTo(0, 0);
        lfringe[1].assignTo(0, subheight);
        rfringe[0].assignTo(subwidth, 0);
        rfringe[1].assignTo(subwidth, subheight);
      } else {
        lfringe[0].assignTo(0, 0);
        lfringe[1].assignTo(subwidth, 0);
        rfringe[0].assignTo(0, subheight);
        rfringe[1].assignTo(subwidth, subheight);
      }
    } else {  // children not in a bus
      for (let idx = 0; idx < childrenLength; idx++) {
        const child = children[idx];
        const childSubSize = child._subtreeSize;
        if (vertical) {
          if (limit > 0 && col > 0 && x + nodespacing + childSubSize.width > limit) {  // need new row
            if (x < maxsubbreadth) {  // align the previous row
              this.shiftRelPosAlign(v, align, maxsubbreadth - x, 0, firstidx, idx - 1);
            }
            row++;  // start new row
            col = 0;  // number of nodes in this row
            firstidx = idx;  // remember index of first child in this row
            olddepth = subheight;  // remember current depth for this row
            x = 0;  // total width so far, excluding any indents
            y = (angle > 135 ? -subheight - rowspacing : subheight + rowspacing);
          }
          this.recordMidPoints(child, 0, y);  // for links being routed around rows
          let cpos = 0.0;
          if (col === 0) {  // get or create first set of fringes for these children
            lfringe = child.leftFringe;
            rfringe = child.rightFringe;
            lsizewidth = childSubSize.width;
            lsizeheight = childSubSize.height;
            // if CHILD doesn't have fringes, or we can't use them, create them
            if (lfringe === null || rfringe === null || angle !== this.orthoAngle(child)) {
              lfringe = this.allocTempPointArray(2);
              rfringe = this.allocTempPointArray(2);
              lfringe[0].assignTo(0, 0);
              lfringe[1].assignTo(0, lsizeheight);
              rfringe[0].assignTo(lsizewidth, 0);
              rfringe[1].assignTo(lsizewidth, lsizeheight);
            }
          } else {
            const resultFringes = Util.tempArray<Array<Point>>();
            const rvalues2 = this.mergeFringes(v, child, lfringe, rfringe, lsizewidth, lsizeheight, resultFringes);
            cpos = rvalues2.x;
            lfringe = resultFringes[0];
            rfringe = resultFringes[1];
            lsizewidth = rvalues2.width;
            lsizeheight = rvalues2.height;
            Util.freeArray(resultFringes);
            // if CHILD's subtree after merge extends "left" beyond original fringe,
            // we need to adjust all of the previous children's RelativePositions (starting at FIRSTIDX)
            // and normalize the fringes
            if (x < childSubSize.width && cpos < 0) {
              this.shiftRelPos(v, -cpos, 0, firstidx, idx - 1);
              this.shiftFringe(lfringe, -cpos, 0);
              this.shiftFringe(rfringe, -cpos, 0);
              cpos = 0;
            }
          }
          child._relativePosition.assignTo(cpos, y);
          subwidth = Math.max(subwidth, lsizewidth);
          subheight = Math.max(subheight, olddepth + (row === 0 ? 0 : rowspacing) + childSubSize.height);
          x = lsizewidth;
        } else {  // horizontal growth direction
          if (limit > 0 && col > 0 && y + nodespacing + childSubSize.height > limit) {  // need new row
            if (y < maxsubbreadth) {  // align the previous row
              this.shiftRelPosAlign(v, align, 0, maxsubbreadth - y, firstidx, idx - 1);
            }
            row++;  // start new row
            col = 0;  // number of nodes in this row
            firstidx = idx;  // remember index of first child in this row
            olddepth = subwidth;  // remember current depth for this row
            y = 0;  // total height so far, excluding any indents
            x = (angle > 135 ? -subwidth - rowspacing : subwidth + rowspacing);
          }
          this.recordMidPoints(child, x, 0);  // for links being routed around rows
          let cpos = 0.0;
          if (col === 0) {  // get or create first set of fringes for these children
            lfringe = child.leftFringe;
            rfringe = child.rightFringe;
            lsizewidth = childSubSize.width;
            lsizeheight = childSubSize.height;
            // if CHILD doesn't have fringes, or we can't use them, create them
            if (lfringe === null || rfringe === null || angle !== this.orthoAngle(child)) {
              lfringe = this.allocTempPointArray(2);
              rfringe = this.allocTempPointArray(2);
              lfringe[0].assignTo(0, 0);
              lfringe[1].assignTo(lsizewidth, 0);
              rfringe[0].assignTo(0, lsizeheight);
              rfringe[1].assignTo(lsizewidth, lsizeheight);
            }
          } else {
            const resultFringes = Util.tempArray<Array<Point>>();
            const rvalues3 = this.mergeFringes(v, child, lfringe, rfringe, lsizewidth, lsizeheight, resultFringes);
            cpos = rvalues3.x;
            lfringe = resultFringes[0];
            rfringe = resultFringes[1];
            lsizewidth = rvalues3.width;
            lsizeheight = rvalues3.height;
            Util.freeArray(resultFringes);
            // if CHILD's subtree after merge extends "above" beyond original fringe,
            // we need to adjust all of the previous children's RelativePositions (starting at FIRSTIDX)
            // and normalize the fringes
            if (y < childSubSize.height && cpos < 0) {
              this.shiftRelPos(v, 0, -cpos, firstidx, idx - 1);
              this.shiftFringe(lfringe, 0, -cpos);
              this.shiftFringe(rfringe, 0, -cpos);
              cpos = 0;
            }
          }
          child._relativePosition.assignTo(x, cpos);
          subheight = Math.max(subheight, lsizeheight);
          subwidth = Math.max(subwidth, olddepth + (row === 0 ? 0 : rowspacing) + childSubSize.width);
          y = lsizeheight;
        }
        col++;
      }
    }
    if (row > 0) { // if there were multiple rows...
      if (vertical) {
        subheight += Math.max(0, layerspacing);  // leave extra room past last row
        if (x < subwidth) {  // align the last row
          this.shiftRelPosAlign(v, align, subwidth - x, 0, firstidx, childrenLength - 1);
        }
        if (rowindent > 0) {
          if (!alignend) this.shiftRelPos(v, rowindent, 0, 0, childrenLength - 1);
          subwidth += rowindent;
        }
      } else {
        subwidth += Math.max(0, layerspacing);  // leave extra room past last row
        if (y < subheight) {  // align the last row
          this.shiftRelPosAlign(v, align, 0, subheight - y, firstidx, childrenLength - 1);
        }
        if (rowindent > 0) {
          if (!alignend) this.shiftRelPos(v, 0, rowindent, 0, childrenLength - 1);
          subheight += rowindent;
        }
      }
    }
    // align this parent node relative to all of its children/descendents
    let offsetx = 0.0;
    let offsety = 0.0;
    switch (align) {
      case TreeLayout.AlignmentCenterSubtrees:
        if (vertical) {
          offsetx += subwidth / 2 - v.focus.x - nodeindent / 2;
        } else {
          offsety += subheight / 2 - v.focus.y - nodeindent / 2;
        }
        break;
      case TreeLayout.AlignmentCenterChildren:
        if (row > 0) {  // same as CenterSubtrees when there's multiple rows
          if (vertical) {
            offsetx += subwidth / 2 - v.focus.x - nodeindent / 2;
          } else {
            offsety += subheight / 2 - v.focus.y - nodeindent / 2;
          }
        } else {
          const num = childrenLength;
          if (vertical) {
            const first = children[0]._relativePosition.x + children[0]._subtreeOffset.x;
            const last = children[num - 1]._relativePosition.x + children[num - 1]._subtreeOffset.x + children[num - 1].focus.x * 2;
            offsetx += first + (last - first) / 2 - v.focus.x - nodeindent / 2;
          } else {
            const first = children[0]._relativePosition.y + children[0]._subtreeOffset.y;
            const last = children[num - 1]._relativePosition.y + children[num - 1]._subtreeOffset.y + children[num - 1].focus.y * 2;
            offsety += first + (last - first) / 2 - v.focus.y - nodeindent / 2;
          }
        }
        break;
      case TreeLayout.AlignmentStart:
        if (vertical) {
          offsetx -= nodeindent;
          subwidth += nodeindent;
        } else {
          offsety -= nodeindent;
          subheight += nodeindent;
        }
        break;
      case TreeLayout.AlignmentEnd:
        if (vertical) {
          offsetx += (subwidth - v.width) + nodeindent;
          subwidth += nodeindent;
        } else {
          offsety += (subheight - v.height) + nodeindent;
          subheight += nodeindent;
        }
        break;
      case TreeLayout.AlignmentBus:
        if (vertical) {
          if (childrenLength > 1) {
            offsetx += maxsubhalf + nodespacing / 2 - v.focus.x;
          } else { // with only one child, offset just like CenterChildren
            offsetx += children[0].focus.x - v.focus.x + children[0]._subtreeOffset.x;
          }
        } else {
          if (childrenLength > 1) {
            offsety += maxsubhalf + nodespacing / 2 - v.focus.y;
          } else { // with only one child, offset just like CenterChildren
            offsety += children[0].focus.y - v.focus.y + children[0]._subtreeOffset.y;
          }
        }
        break;
      case TreeLayout.AlignmentTopLeftBus:
        if (vertical) {
          offsetx += subwidth + nodespacing / 2 - v.focus.x;
        } else {
          offsety += subheight + nodespacing / 2 - v.focus.y;
        }
        break;
      case TreeLayout.AlignmentBottomRightBus:
        break;
      default:
        Util.throwError('Unhandled alignment value ' + align.toString());
    }
    // fix up all children's RelativePositions relative to the desired position of this parent node
    for (let i = 0; i < childrenLength; i++) {
      const child = children[i];
      if (vertical) {
        child._relativePosition.assignTo(
          child._relativePosition.x + child._subtreeOffset.x - offsetx,
          child._relativePosition.y + ((angle > 135) ?
            ((alignbus ? -subheight : -child._subtreeSize.height) + child._subtreeOffset.y - layerspacing) :
            (vHeight + layerspacing + child._subtreeOffset.y)));
      } else {
        child._relativePosition.assignTo(
          child._relativePosition.x + ((angle > 135) ?
            ((alignbus ? -subwidth : -child._subtreeSize.width) + child._subtreeOffset.x - layerspacing) :
            (vWidth + layerspacing + child._subtreeOffset.x)),
          child._relativePosition.y + child._subtreeOffset.y - offsety);
      }
    }
    // include the NODE itself in the subtree size, and adjust the offset if the direction is negative
    let fringeoffwidth = 0.0;
    let fringeoffheight = 0.0;
    if (alignbus) {
      if (vertical) {
        subwidth = this.calculateSubwidth(v, subwidth, offsetx);
        // maybe the alignment caused the parent node to stick out
        if (offsetx < 0) offsetx = 0;  // subwidth has already been compensated
        if (angle > 135) offsety += subheight + layerspacing;
        subheight += vHeight + layerspacing;
        if (align === TreeLayout.AlignmentBottomRightBus) fringeoffwidth += nodespacing / 2 + v.focus.x;
        fringeoffheight += vHeight + layerspacing;
      } else {
        if (angle > 135) offsetx += subwidth + layerspacing;
        subwidth += vWidth + layerspacing;
        subheight = this.calculateSubheight(v, subheight, offsety);
        // maybe the alignment caused the parent node to stick out
        if (offsety < 0) offsety = 0; // subheight has already been compensated
        if (align === TreeLayout.AlignmentBottomRightBus) fringeoffheight += nodespacing / 2 + v.focus.y;
        fringeoffwidth += vWidth + layerspacing;
      }
    } else {
      if (vertical) {
        if (v.comments === null) {
          if (vWidth > subwidth) {
            const fringeoffarr = this.alignOffset(align, vWidth - subwidth, 0);
            fringeoffwidth = fringeoffarr.x;
            fringeoffheight = fringeoffarr.y;
            subwidth = vWidth;
            offsetx = 0;
          }
        } else {
          subwidth = this.calculateSubwidth(v, subwidth, offsetx);
        }
        if (offsetx < 0) {
          fringeoffwidth -= offsetx;  // increase fringe offset
          offsetx = 0;
        }
        if (angle > 135) offsety += subheight + layerspacing;
        subheight = Math.max(Math.max(subheight, vHeight), subheight + vHeight + layerspacing);
        fringeoffheight += vHeight + layerspacing;
      } else {
        if (angle > 135) offsetx += subwidth + layerspacing;
        subwidth = Math.max(Math.max(subwidth, vWidth), subwidth + vWidth + layerspacing);
        if (v.comments === null) {
          if (vHeight > subheight) {
            const fringeoffarr = this.alignOffset(align, 0, vHeight - subheight);
            fringeoffwidth = fringeoffarr.x;
            fringeoffheight = fringeoffarr.y;
            subheight = vHeight;
            offsety = 0;
          }
        } else {
          subheight = this.calculateSubheight(v, subheight, offsety);
        }
        if (offsety < 0) {
          fringeoffheight -= offsety;  // increase fringe offset
          offsety = 0;
        }
        fringeoffwidth += vWidth + layerspacing;
      }
    }
    // create new fringes for the whole subtree
    let newleft: Array<Point>;
    let newright: Array<Point>;
    if (row > 0) {  // when there are multiple rows, just assume everything's a big block
      newleft = this.allocTempPointArray(4);
      newright = this.allocTempPointArray(4);
      if (vertical) {
        newleft[2].assignTo(0, vHeight + layerspacing);
        newleft[3].assignTo(newleft[2].x, subheight);
        newright[2].assignTo(subwidth, newleft[2].y);
        newright[3].assignTo(newright[2].x, newleft[3].y);
      } else {
        newleft[2].assignTo(vWidth + layerspacing, 0);
        newleft[3].assignTo(subwidth, newleft[2].y);
        newright[2].assignTo(newleft[2].x, subheight);
        newright[3].assignTo(newleft[3].x, newright[2].y);
      }
    } else {
      newleft = this.allocTempPointArray(lfringe.length + 2);
      newright = this.allocTempPointArray(rfringe.length + 2);
      for (let i = 0; i < lfringe.length; i++) {
        const p = lfringe[i];
        newleft[i + 2].assignTo(p.x + fringeoffwidth, p.y + fringeoffheight);
      }
      for (let i = 0; i < rfringe.length; i++) {
        const p = rfringe[i];
        newright[i + 2].assignTo(p.x + fringeoffwidth, p.y + fringeoffheight);
      }
    }
    // add the parent node to the fringe
    if (vertical) {
      newleft[0].assignTo(offsetx, 0);
      newleft[1].assignTo(newleft[0].x, vHeight);
      // handle negative LayerSpacing causing overlap in breadth
      if (newleft[2].y < newleft[1].y) {
        if (newleft[2].x > newleft[0].x) {
          newleft[2].assign(newleft[1]);  // inside parent node: use parent corner
        } else {
          newleft[1].assign(newleft[2]);  // outside parent node: extend exterior
        }
      }
      if (newleft[3].y < newleft[2].y) {
        if (newleft[3].x > newleft[0].x) {
          newleft[3].assign(newleft[2]);
        } else {
          newleft[2].assign(newleft[3]);
        }
      }
      newright[0].assignTo(offsetx + vWidth, 0);
      newright[1].assignTo(newright[0].x, vHeight);
      // handle negative LayerSpacing causing overlap in breadth
      if (newright[2].y < newright[1].y) {
        if (newright[2].x < newright[0].x) {
          newright[2].assign(newright[1]);  // inside parent node: use parent corner
        } else {
          newright[1].assign(newright[2]);  // outside parent node: extend exterior
        }
      }
      if (newright[3].y < newright[2].y) {
        if (newright[3].x < newright[0].x) {
          newright[3].assign(newright[2]);
        } else {
          newright[2].assign(newright[3]);
        }
      }
      // extend the fringe vertically so that orthogonal links don't extend outside
      newleft[2].y -= layerspacing / 2;
      newright[2].y -= layerspacing / 2;
    } else {
      newleft[0].assignTo(0, offsety);
      newleft[1].assignTo(vWidth, newleft[0].y);
      // handle negative LayerSpacing causing overlap in breadth
      if (newleft[2].x < newleft[1].x) {
        if (newleft[2].y > newleft[0].y) {
          newleft[2].assign(newleft[1]);  // inside parent node: use parent corner
        } else {
          newleft[1].assign(newleft[2]);  // outside parent node: extend exterior
        }
      }
      if (newleft[3].x < newleft[2].x) {
        if (newleft[3].y > newleft[0].y) {
          newleft[3].assign(newleft[2]);
        } else {
          newleft[2].assign(newleft[3]);
        }
      }
      newright[0].assignTo(0, offsety + vHeight);
      newright[1].assignTo(vWidth, newright[0].y);
      // handle negative LayerSpacing causing overlap in breadth
      if (newright[2].x < newright[1].x) {
        if (newright[2].y < newright[0].y) {
          newright[2].assign(newright[1]);  // inside parent node: use parent corner
        } else {
          newright[1].assign(newright[2]);  // outside parent node: extend exterior
        }
      }
      if (newright[3].x < newright[2].x) {
        if (newright[3].y < newright[0].y) {
          newright[3].assign(newright[2]);
        } else {
          newright[2].assign(newright[3]);
        }
      }
      // extend the fringe horizontally so that orthogonal links don't extend outside
      newleft[2].x -= layerspacing / 2;
      newright[2].x -= layerspacing / 2;
    }
    this.freeTempPointArray(lfringe);
    this.freeTempPointArray(rfringe);
    // save results
    v.leftFringe = newleft;
    v.rightFringe = newright;
    v._subtreeOffset.assignTo(offsetx, offsety);
    v._subtreeSize.assignTo(subwidth, subheight);
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} fringe
   * @param {number} offsetwidth
   * @param {number} offsetheight
   */
  private shiftFringe(fringe: Array<Point>, offsetwidth: number, offsetheight: number): void {
    for (let i = 0; i < fringe.length; i++) {
      const p = fringe[i];  // OK to clobber?
      p.x += offsetwidth;
      p.y += offsetheight;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} parent
   * @param {TreeVertex} child
   * @param {Array.<Point>} lfringe
   * @param {Array.<Point>} rfringe
   * @param {number} lsizewidth
   * @param {number} lsizeheight
   * @param {Array.<Array.<Point>>} resultFringes
   * @return {Rect}
   */
  private mergeFringes(parent: TreeVertex, child: TreeVertex, lfringe: Array<Point>, rfringe: Array<Point>,
                       lsizewidth: number, lsizeheight: number, resultFringes: Array<Array<Point>>): Rect {
    if (Debug) Util.checkClass(parent, TreeVertex, TreeLayout, 'mergeFringes:parent');
    if (Debug) Util.checkClass(child, TreeVertex, TreeLayout, 'mergeFringes:child');
    const angle = this.orthoAngle(parent);
    const vertical = (angle === 90 || angle === 270);
    const nodespacing = parent.nodeSpacing;
    const al = lfringe;
    const ar = rfringe;
    const aszwidth = lsizewidth;
    const aszheight = lsizeheight;
    let bl = child.leftFringe;
    let br = child.rightFringe;
    const bsz = child._subtreeSize;
    const d = (vertical ? Math.max(aszheight, bsz.height) : Math.max(aszwidth, bsz.width));
    if (bl === null || angle !== this.orthoAngle(child)) {
      bl = this.allocTempPointArray(2);
      br = this.allocTempPointArray(2);
      if (vertical) {
        bl[0].assignTo(0, 0);
        bl[1].assignTo(0, bsz.height);
        br[0].assignTo(bsz.width, 0);
        br[1].assignTo(br[0].x, bl[1].y);
      } else {
        bl[0].assignTo(0, 0);
        bl[1].assignTo(bsz.width, 0);
        br[0].assignTo(0, bsz.height);
        br[1].assignTo(bl[1].x, br[0].y);
      }
    }
    // handle general case
    if (vertical) {
      const abreadth = aszwidth;
      let m = abreadth - this.fringeDistanceX(ar, bl, abreadth);
      m += nodespacing;
      lfringe = this.fringeUnionLeftX(al, bl, m);
      rfringe = this.fringeUnionRightX(ar, br, m);
      lsizewidth = Math.max(0, m) + bsz.width;
      lsizeheight = d;
      this.freeTempPointArray(al);
      this.freeTempPointArray(bl);
      this.freeTempPointArray(ar);
      this.freeTempPointArray(br);
      resultFringes[0] = lfringe;
      resultFringes[1] = rfringe;
      return new Rect(m, 0, lsizewidth, lsizeheight);
    } else {
      const abreadth = aszheight;
      let m = abreadth - this.fringeDistanceY(ar, bl, abreadth);
      m += nodespacing;
      lfringe = this.fringeUnionLeftY(al, bl, m);
      rfringe = this.fringeUnionRightY(ar, br, m);
      lsizewidth = d;
      lsizeheight = Math.max(0, m) + bsz.height;
      this.freeTempPointArray(al);
      this.freeTempPointArray(bl);
      this.freeTempPointArray(ar);
      this.freeTempPointArray(br);
      resultFringes[0] = lfringe;
      resultFringes[1] = rfringe;
      return new Rect(m, 0, lsizewidth, lsizeheight);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {Array.<Point>}
   */
  private fringeUnionLeftY(av: Array<Point>, bv: Array<Point>, offset: number): Array<Point> {
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return null;  // not enough points!
    const cv = this.allocTempPointArray(av.length + bv.length);
    let i = 0;
    let j = 0;
    let k = 0;
    // copy any leading points of BV, shifted by OFFSET
    while (j < bv.length && bv[j].x < av[0].x) {
      const c = bv[j++];
      cv[k++].assignTo(c.x, c.y + offset);
    }
    // copy all of AV
    while (i < av.length) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // skip the part of BV that is covered by AV
    const lastX = av[av.length - 1].x;
    while (j < bv.length && bv[j].x <= lastX) {
      j++;
    }
    // copy the trailing points of BV, shifted by OFFSET
    while (j < bv.length && bv[j].x > lastX) {
      const c = bv[j++];
      cv[k++].assignTo(c.x, c.y + offset);
    }
    // minimize size of result array
    const r = this.allocTempPointArray(k);
    for (i = 0; i < k; i++) r[i].assign(cv[i]);
    this.freeTempPointArray(cv);
    return r;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {Array.<Point>}
   */
  private fringeUnionLeftX(av: Array<Point>, bv: Array<Point>, offset: number): Array<Point> {
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return null;  // not enough points!
    const cv = this.allocTempPointArray(av.length + bv.length);
    let i = 0;
    let j = 0;
    let k = 0;
    // copy any leading points of BV, shifted by OFFSET
    while (j < bv.length && bv[j].y < av[0].y) {
      const c = bv[j++];
      cv[k++].assignTo(c.x + offset, c.y);
    }
    // copy all of AV
    while (i < av.length) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // skip the part of BV that is covered by AV
    const lastY = av[av.length - 1].y;
    while (j < bv.length && bv[j].y <= lastY) {
      j++;
    }
    // copy the trailing points of BV, shifted by OFFSET
    while (j < bv.length && bv[j].y > lastY) {
      const c = bv[j++];
      cv[k++].assignTo(c.x + offset, c.y);
    }
    // minimize size of result array
    const r = this.allocTempPointArray(k);
    for (i = 0; i < k; i++) r[i].assign(cv[i]);
    this.freeTempPointArray(cv);
    return r;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {Array.<Point>}
   */
  private fringeUnionRightY(av: Array<Point>, bv: Array<Point>, offset: number): Array<Point> {
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return null;  // not enough points!
    const cv = this.allocTempPointArray(av.length + bv.length);
    let i = 0;
    let j = 0;
    let k = 0;
    // copy any leading points of AV
    while (i < av.length && av[i].x < bv[0].x) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // copy all of BV, shifted by OFFSET
    while (j < bv.length) {
      const c = bv[j++];
      cv[k++].assignTo(c.x, c.y + offset);
    }
    // skip the part of AV that is covered by BV
    const lastX = bv[bv.length - 1].x;
    while (i < av.length && av[i].x <= lastX) {
      i++;
    }
    // copy the trailing points of AV
    while (i < av.length && av[i].x > lastX) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // minimize size of result array
    const r = this.allocTempPointArray(k);
    for (i = 0; i < k; i++) r[i].assign(cv[i]);
    this.freeTempPointArray(cv);
    return r;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {Array.<Point>}
   */
  private fringeUnionRightX(av: Array<Point>, bv: Array<Point>, offset: number): Array<Point> {
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return null;  // not enough points!
    const cv = this.allocTempPointArray(av.length + bv.length);
    let i = 0;
    let j = 0;
    let k = 0;
    // copy any leading points of AV
    while (i < av.length && av[i].y < bv[0].y) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // copy all of BV, shifted by OFFSET
    while (j < bv.length) {
      const c = bv[j++];
      cv[k++].assignTo(c.x + offset, c.y);
    }
    // skip the part of AV that is covered by BV
    const lastY = bv[bv.length - 1].y;
    while (i < av.length && av[i].y <= lastY) {
      i++;
    }
    // copy the trailing points of AV
    while (i < av.length && av[i].y > lastY) {
      const c = av[i++];
      cv[k++].assignTo(c.x, c.y);
    }
    // minimize size of result array
    const r = this.allocTempPointArray(k);
    for (i = 0; i < k; i++) r[i].assign(cv[i]);
    this.freeTempPointArray(cv);
    return r;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {number}
   */
  private fringeDistanceY(av: Array<Point>, bv: Array<Point>, offset: number): number {
    let min = 9999999;
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return min;  // not enough points!
    let i = 0;
    let j = 0;
    while (i < av.length && j < bv.length) {
      const ai = av[i];
      const bj = bv[j]; const bjx = bj.x; let bjy = bj.y;
      bjy += offset;
      let an = ai;
      if (i + 1 < av.length) {
        an = av[i + 1];
      }
      let bn = bj; let bnx = bn.x; let bny = bn.y;
      if (j + 1 < bv.length) {
        bn = bv[j + 1]; bnx = bn.x; bny = bn.y;
        bny += offset;
      }
      let diff = min;
      if (ai.x === bjx) {  // same X's, so just compare Y's
        diff = bjy - ai.y;
      } else if (ai.x > bjx && ai.x < bnx) {  // ai is between bj and bn
        diff = (bjy + ((ai.x - bjx) / (bnx - bjx)) * (bny - bjy)) - ai.y;
      } else if (bjx > ai.x && bjx < an.x) {  // bj is between ai and an
        diff = bjy - (ai.y + ((bjx - ai.x) / (an.x - ai.x)) * (an.y - ai.y));
      }
      //    Maybe use the below if copy() is too expensive??
      //    bjy -= offset;
      //    if (j + 1 < bv.length) {
      //      bny -= offset;  // set bnx (or bv[j+1].x) back to its original value for the next pass
      //    }
      if (diff < min) {
        min = diff;
      }
      // pick next point
      if (an.x <= ai.x) {
        i++;
      } else if (bnx <= bjx) {
        j++;
      } else {
        // both indices increment if both next points have the same X
        if (an.x <= bnx) i++;
        if (bnx <= an.x) j++;
      }
    }
    return min;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} av
   * @param {Array.<Point>} bv
   * @param {number} offset
   * @return {number}
   */
  private fringeDistanceX(av: Array<Point>, bv: Array<Point>, offset: number): number {
    let min = 9999999;
    if (av === null || av.length < 2 || bv === null || bv.length < 2) return min;  // not enough points!
    let i = 0;
    let j = 0;
    while (i < av.length && j < bv.length) {
      const ai = av[i];
      const bj = bv[j]; let bjx = bj.x; const bjy = bj.y;
      bjx += offset;
      let an = ai;
      if (i + 1 < av.length) {
        an = av[i + 1];
      }
      let bn = bj; let bnx = bn.x; let bny = bn.y;
      if (j + 1 < bv.length) {
        bn = bv[j + 1]; bnx = bn.x; bny = bn.y;
        bnx += offset;
      }
      let diff = min;
      if (ai.y === bjy) {  // same Y's, so just compare X's
        diff = bjx - ai.x;
      } else if (ai.y > bjy && ai.y < bny) {  // ai is between bj and bn
        diff = (bjx + ((ai.y - bjy) / (bny - bjy)) * (bnx - bjx)) - ai.x;
      } else if (bjy > ai.y && bjy < an.y) {  // bj is between ai and an
        diff = bjx - (ai.x + ((bjy - ai.y) / (an.y - ai.y)) * (an.x - ai.x));
      }
      //    Maybe use the below if copy() is too expensive??
      //    bjx -= offset;
      //    if (j + 1 < bv.length) {
      //      bnx -= offset;  // set bnx (or bv[j+1].x) back to its original value for the next pass
      //    }
      if (diff < min) {
        min = diff;
      }
      // pick next point
      if (an.y <= ai.y) {
        i++;
      } else if (bny <= bjy) {
        j++;
      } else {
        // both indices increment if both next points have the same Y
        if (an.y <= bny) i++;
        if (bny <= an.y) j++;
      }
    }
    return min;
  }

  /**
   * @hidden @internal
   * Arrays of tree fringe points are now pre-allocated with Points, so that they can be re-used too.
   * @this {TreeLayout}
   * @param {number} len
   * @return {Array.<Point>} of uninitialized Points
   */
  private allocTempPointArray(len: number): Array<Point> {
    // see if there's a cache (actually an Array) of Arrays of the requested length
    const cache = this._tempArrays[len];
    if (cache !== undefined) {
      // see if there's any Array in the cache
      const oldarr = cache.pop();
      if (oldarr !== undefined) return oldarr;
    }
    // allocate a new Array holding that number of Points
    const newarr: Array<Point> = [];
    for (let i = 0; i < len; i++) newarr[i] = new Point();
    return newarr;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {Array.<Point>} a
   */
  private freeTempPointArray(a: Array<Point>): void {
    const len = a.length;
    let cache = this._tempArrays[len];
    if (cache === undefined) {
      cache = [];
      this._tempArrays[len] = cache;
    }
    cache.push(a);
  }

  /**
   * Position each separate tree.
   *
   * This is called after each tree has been laid out and thus each subtree
   * bounds are known.
   * The {@link #arrangement} and {@link #arrangementSpacing} and {@link Layout#arrangementOrigin}
   * properties affect this method's behavior.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   * @since 1.2
   */
  public arrangeTrees(): void {
    if (this._arrangement === TreeLayout.ArrangementFixedRoots) {
      const rit = this._roots.iterator;
      while (rit.next()) {
        const rootVal = rit.value;
        if (!(rootVal instanceof TreeVertex)) continue;
        const r = rootVal.node;
        if (r === null) continue;
        const pos = r.position;
        let posx = pos.x;
        let posy = pos.y;
        if (!isFinite(posx)) posx = 0;
        if (!isFinite(posy)) posy = 0;
        this.assignAbsolutePositions(rootVal, posx, posy);
      }
    } else {
      const arr = [];
      const rit = this._roots.iterator;
      while (rit.next()) {
        const rootVal = rit.value;
        if (!(rootVal instanceof TreeVertex)) continue;
        arr.push(rootVal);
      }
      switch (this.sorting) {
        case TreeLayout.SortingForwards:  // nothing needed
          break;
        case TreeLayout.SortingReverse:  // just reverse the array
          arr.reverse();
          break;
        case TreeLayout.SortingAscending:  // sort using comparer
          arr.sort(this.comparer);
          break;
        case TreeLayout.SortingDescending:  // sort, then reverse
          arr.sort(this.comparer);
          arr.reverse();
          break;
        default:
          Util.throwError('Unhandled sorting value ' + this.sorting.toString());
      }

      const p = this.arrangementOrigin;
      let px = p.x;
      let py = p.y;
      for (let i = 0; i < arr.length; i++) {
        const rt = arr[i];
        this.assignAbsolutePositions(rt, px + rt._subtreeOffset.x, py + rt._subtreeOffset.y);
        switch (this._arrangement) {
          case TreeLayout.ArrangementVertical:
            py += rt._subtreeSize.height + this._arrangementSpacing.height;
            break;
          case TreeLayout.ArrangementHorizontal:
            px += rt._subtreeSize.width + this._arrangementSpacing.width;
            break;
          default:
            Util.throwError('Unhandled arrangement value ' + this._arrangement.toString());
        }
      }
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {number} px
   * @param {number} py
   */
  private assignAbsolutePositions(v: TreeVertex, px: number, py: number): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'assignAbsolutePositions:v');
    v.x = px;
    v.y = py;
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.assignAbsolutePositions(c, px + c._relativePosition.x, py + c._relativePosition.y);
    }
  }

  /**
   * Set the fromSpot and toSpot on each Link, position each Node according
   * to the vertex position, and then position/route the Links.
   *
   * This calls the {@link #commitNodes} and {@link #commitLinks} methods, the latter only if {@link #isRouting} is true.
   * You should not call this method -- it is a "protected virtual" method.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @this {TreeLayout}
   */
  protected commitLayout(): void {
    this.setPortSpotsAll();
    this.commitNodes();
    this.commitLayersInternal();
    if (this.isRouting) this.commitLinks();
  }

  /**
   * Commit the position of all nodes.
   *
   * This is called by {@link #commitLayout}.
   * See also {@link #commitLinks}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   */
  protected commitNodes(): void {
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value;
      v.commit();
    }
    vit.reset();
    while (vit.next()) {
      const c = vit.value as TreeVertex;
      this.layoutComments(c);
    }
  }

  /**
   * @hidden @internal
   * Call {@link #commitLayers}, but only if all vertexes will be completely aligned in layers across the whole tree.
   * @this {TreeLayout}
   */
  private commitLayersInternal(): void {
    if (this.layerStyle !== TreeLayout.LayerUniform) return;
    const depths = this._allDepths;
    const spaces: Array<number> = [];
    let totalBounds: Rect = null;
    const vit = this.network.vertexes.iterator;
    while (vit.next()) {
      const v = vit.value as TreeVertex;
      if (totalBounds === null) {
        totalBounds = v.bounds.copy();
      } else {
        totalBounds.unionRect(v.bounds);
      }
      let space = spaces[v.level];
      if (space === undefined) {
        space = this.computeLayerSpacing(v);
      } else {
        space = Math.max(space, this.computeLayerSpacing(v));
      }
      spaces[v.level] = space;
    }
    // make sure there's a number for each layer
    for (let i = 0; i < spaces.length; i++) if (spaces[i] === undefined) spaces[i] = 0;
    let offset: Point;
    if (this.angle === 90 || this.angle === 270) {
      totalBounds.inflate(this.nodeSpacing / 2, this.layerSpacing);
      offset = new Point(-this.nodeSpacing / 2, -this.layerSpacing / 2);
    } else {
      totalBounds.inflate(this.layerSpacing, this.nodeSpacing / 2);
      offset = new Point(-this.layerSpacing / 2, -this.nodeSpacing / 2);
    }

    const layerRects: Array<Rect> = [];
    const columnsBreadth = (this.angle === 90 || this.angle === 270) ? totalBounds.width : totalBounds.height;
    let d = 0.0;
    if (this.angle === 180 || this.angle === 270) {
      for (let i = 0; i < depths.length; i++) {
        d += depths[i] + spaces[i];
      }
    }
    for (let i = 0; i < depths.length; i++) {
      const depth = depths[i] + spaces[i];
      if (this.angle === 270) {
        d -= depth;
        layerRects.push(new Rect(0, d, columnsBreadth, depth));
      } else if (this.angle === 90) {
        layerRects.push(new Rect(0, d, columnsBreadth, depth));
        d += depth;
      } else if (this.angle === 180) {
        d -= depth;
        layerRects.push(new Rect(d, 0, depth, columnsBreadth));
      } else {
        layerRects.push(new Rect(d, 0, depth, columnsBreadth));
        d += depth;
      }
    }
    this.commitLayers(layerRects, offset);
  }

  /**
   * This overridable method is called by {@link #commitLayout}
   * if {@link #layerStyle} is LayerUniform
   * to support custom arrangement of bands or labels across each layout layer.
   * By default this method does nothing.
   *
   * The coordinates used in the resulting Rects may need to be offset by the {@link Layout#arrangementOrigin}
   * and/or by the arrangement of subtrees done by {@link #arrangeTrees}.
   * @expose
   * @this {TreeLayout}
   * @param {Array.<Rect>} layerRects an Array of {@link Rect}s with the bounds of each of the "layers"
   * @param {Point} offset the position of the top-left corner of the banded area relative to the coordinates given by the layerRects
   * @since 1.4
   */
  protected commitLayers(layerRects: Array<Rect>, offset: Point): void {
  }

  /**
   * Routes the links.
   *
   * This is called by {@link #commitLayout}.
   * This is only called if {@link Layout#isRouting} is true.
   * See also {@link #commitNodes}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   */
  protected commitLinks(): void {
    const eit = this.network.edges.iterator;
    while (eit.next()) {
      const e = eit.value;
      e.commit();
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   */
  private setPortSpotsAll(): void {
    const rit = this._roots.iterator;
    while (rit.next()) {
      const v = rit.value;
      if (v instanceof TreeVertex) this.setPortSpotsTree(v);
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  private setPortSpotsTree(v: TreeVertex): void {
    if (v === null) return;
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'setPortSpotsTree:v');
    this.setPortSpots(v);
    const children = v.children;
    const childrenLength = children.length;
    for (let i = 0; i < childrenLength; i++) {
      const c = children[i];
      this.setPortSpotsTree(c);
    }
  }

  /**
   * @hidden @internal
   * Assign port spots for single-port nodes, if {@link TreeVertex#setsPortSpot} or
   * {@link TreeVertex#setsChildPortSpot} is true, according to the value of {@link #portSpot}
   * or {@link #childPortSpot}.
   *
   * This iterates over all of the {@link TreeEdge}s in the network,
   * and assigns their fromSpot and/or toSpot property if the
   * {@link TreeVertex#setsPortSpot} and/or {@link TreeVertex#setsChildPortSpot} properties
   * are true, and if the port's node only has a single port.
   *
   * The spot values are given by {@link #portSpot}
   * and {@link #childPortSpot} respectively, unless the value is {@link Spot.Default}.
   * In the latter case the actual spot is determined by the {@link TreeVertex#angle}
   * of the parent node (the fromVertex).
   * For example, when the angle is zero, the tree grows deeper toward the right.
   * So the fromSpot will be set to {@link Spot.MiddleRight}, and the
   * toSpot will be set to {@link Spot.MiddleLeft}.
   * Please read the Introduction page on <a href="../../intro/extensions.html">Extensions</a> for how to override methods and how to call this base method.
   * @expose
   * @this {TreeLayout}
   * @param {TreeVertex} v
   */
  public setPortSpots(v: TreeVertex): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'setPortSpots:v');
    const align = v.alignment;
    if (this.isBusAlignment(align)) {
      this.setPortSpotsBus(v, align);
    } else {
      const angle = this.orthoAngle(v);
      if (this._effectivePath === TreeLayout.PathDestination) {
        const eit = v.destinationEdges;
        while (eit.next()) {
          const dedge = eit.value;
          const link = dedge.link;
          if (link !== null) {
            if (v.setsPortSpot) {
              if (v.portSpot.isDefault()) {
                switch (angle) {
                  case 0: link.fromSpot = Spot.MiddleRight; break;
                  case 90: link.fromSpot = Spot.MiddleBottom; break;
                  case 180: link.fromSpot = Spot.MiddleLeft; break;
                  default: link.fromSpot = Spot.MiddleTop; break;
                }
              } else {
                link.fromSpot = v.portSpot;
              }
            }
            if (v.setsChildPortSpot) {
              if (v.childPortSpot.isDefault()) {
                switch (angle) {
                  case 0: link.toSpot = Spot.MiddleLeft; break;
                  case 90: link.toSpot = Spot.MiddleTop; break;
                  case 180: link.toSpot = Spot.MiddleRight; break;
                  default: link.toSpot = Spot.MiddleBottom; break;
                }
              } else {
                link.toSpot = v.childPortSpot;
              }
            }
          }
        }
      } else {  // tree path === source
        const eit = v.sourceEdges;
        while (eit.next()) {
          const sedge = eit.value;
          const link = sedge.link;
          if (link !== null) {
            if (v.setsPortSpot) {
              if (v.portSpot.isDefault()) {
                switch (angle) {
                  case 0: link.toSpot = Spot.MiddleRight; break;
                  case 90: link.toSpot = Spot.MiddleBottom; break;
                  case 180: link.toSpot = Spot.MiddleLeft; break;
                  default: link.toSpot = Spot.MiddleTop; break;
                }
              } else {
                link.toSpot = v.portSpot;
              }
            }
            if (v.setsChildPortSpot) {
              if (v.childPortSpot.isDefault()) {
                switch (angle) {
                  case 0: link.fromSpot = Spot.MiddleLeft; break;
                  case 90: link.fromSpot = Spot.MiddleTop; break;
                  case 180: link.fromSpot = Spot.MiddleRight; break;
                  default: link.fromSpot = Spot.MiddleBottom; break;
                }
              } else {
                link.fromSpot = v.childPortSpot;
              }
            }
          }
        }
      }
    }
  }

  /**
   * @hidden @internal
   * Sets the PortSpots for vertexes with Bus alignments
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @param {EnumValue} align
   */
  private setPortSpotsBus(v: TreeVertex, align: EnumValue): void {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'setPortSpotsBus:v');
    if (Debug) Util.checkClass(align, EnumValue, TreeLayout, 'setPortSpots:align');
    const dest = (this._effectivePath === TreeLayout.PathDestination);
    const angle = this.orthoAngle(v);
    let parentspot: Spot;
    switch (angle) {
      case 0: parentspot = Spot.MiddleRight; break;
      case 90: parentspot = Spot.MiddleBottom; break;
      case 180: parentspot = Spot.MiddleLeft; break;
      default: parentspot = Spot.MiddleTop; break;
    }
    const children = v.children;
    const childrenLength = children.length;
    switch (align) {
      case TreeLayout.AlignmentBus:
      case TreeLayout.AlignmentBusBranching: {
        for (let i = 0; i < childrenLength; i++) {
          const c = children[i];
          const edge = (dest ? c.sourceEdges : c.destinationEdges).first();
          if (edge === null) continue;
          const l = edge.link;
          if (l !== null) {
            let childspot = (angle === 90 || angle === 270 ? Spot.MiddleLeft : Spot.MiddleTop);
            if (childrenLength === 1 || (i === childrenLength - 1 && childrenLength % 2 === 1)) {
              switch (angle) {
                case 0: childspot = Spot.MiddleLeft; break;
                case 90: childspot = Spot.MiddleTop; break;
                case 180: childspot = Spot.MiddleRight; break;
                default: childspot = Spot.MiddleBottom; break;
              }
            } else {
              if (i % 2 === 0) {
                childspot = (angle === 90 || angle === 270 ? Spot.MiddleRight : Spot.MiddleBottom);
              }
            }
            if (dest) {
              if (v.setsPortSpot) l.fromSpot = parentspot;
              if (v.setsChildPortSpot) l.toSpot = childspot;
            } else {
              if (v.setsPortSpot) l.fromSpot = childspot;
              if (v.setsChildPortSpot) l.toSpot = parentspot;
            }
          }
        }
        break;
      }
      case TreeLayout.AlignmentTopLeftBus: {
        const childspot = (angle === 90 || angle === 270 ? Spot.MiddleRight : Spot.MiddleBottom);
        const eit = (dest ? v.destinationEdges : v.sourceEdges);
        while (eit.next()) {
          const tledge = eit.value;
          const l = tledge.link;
          if (l !== null) {
            if (dest) {
              if (v.setsPortSpot) l.fromSpot = parentspot;
              if (v.setsChildPortSpot) l.toSpot = childspot;
            } else {
              if (v.setsPortSpot) l.fromSpot = childspot;
              if (v.setsChildPortSpot) l.toSpot = parentspot;
            }
          }
        }
        break;
      }
      case TreeLayout.AlignmentBottomRightBus: {
        const childspot = (angle === 90 || angle === 270 ? Spot.MiddleLeft : Spot.MiddleTop);
        const eit = (dest ? v.destinationEdges : v.sourceEdges);
        while (eit.next()) {
          const bredge = eit.value;
          const l = bredge.link;
          if (l !== null) {
            if (dest) {
              if (v.setsPortSpot) l.fromSpot = parentspot;
              if (v.setsChildPortSpot) l.toSpot = childspot;
            } else {
              if (v.setsPortSpot) l.fromSpot = childspot;
              if (v.setsChildPortSpot) l.toSpot = parentspot;
            }
          }
        }
        break;
      }
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @return {number}
   */
  public orthoAngle(v: TreeVertex): number {
    const angle = v.angle;
    if (angle <= 45) {
      return 0;
    } else if (angle <= 135) {
      return 90;
    } else if (angle <= 225) {
      return 180;
    } else if (angle <= 315) {
      return 270;
    } else {
      return 0;
    }
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @return {number}
   */
  public computeLayerSpacing(v: TreeVertex): number {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'computeLayerSpacing:v');
    const angle = this.orthoAngle(v);
    const vertical = (angle === 90 || angle === 270);
    let layerspacing = v.layerSpacing;  // limit layerspacing to be no more negative than the depth of the node
    if (v.layerSpacingParentOverlap > 0) {
      const overlap = Math.min(1, v.layerSpacingParentOverlap);
      layerspacing -= (vertical ? v.height * overlap : v.width * overlap);
    }
    if (layerspacing < (vertical ? -v.height : -v.width)) {
      layerspacing = (vertical ? -v.height : -v.width);
    }
    return layerspacing;
  }

  /**
   * @hidden @internal
   * @this {TreeLayout}
   * @param {TreeVertex} v
   * @return {number}
   */
  public computeNodeIndent(v: TreeVertex): number {
    if (Debug) Util.checkClass(v, TreeVertex, TreeLayout, 'computeNodeIndent:v');
    const angle = this.orthoAngle(v);
    const vertical = (angle === 90 || angle === 270);
    let nodeindent = v.nodeIndent;
    if (v.nodeIndentPastParent > 0) {
      const ind = Math.min(1, v.nodeIndentPastParent);
      nodeindent += (vertical ? v.width * ind : v.height * ind);
    }
    nodeindent = Math.max(0, nodeindent);
    return nodeindent;
  }

  // ***********************************Properties********************************

  /**
   * Gets or sets the collection of root vertexes.
   *
   * Initially this will be an empty {@link Set}.
   *
   * If the {@link #path} is either {@link TreeLayout.PathDestination|PathDestination} or
   * {@link TreeLayout.PathSource|PathSource}, this layout can easily
   * determine all of the tree roots by searching the whole network.
   * Otherwise, you should explicitly initialize this collection
   * with one or more {@link TreeVertex}es.
   */
  /** @export */ get roots(): Set<(TreeVertex | Node)> { return this._roots; }
  /** @export */ set roots(value: Set<(TreeVertex | Node)>) {
    if (this._roots !== value) {
      Util.checkClass(value, Set, TreeLayout, 'roots');
      this._roots = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets how the tree should be constructed from the
   * {@link TreeEdge}s connecting {@link TreeVertex}es.
   * Must be {@link TreeLayout.PathDestination} or {@link TreeLayout.PathSource} or {@link TreeLayout.PathDefault}.
   *
   * The default value is {@link TreeLayout.PathDefault},
   * where the value of {@link Diagram#isTreePathToChildren} determines the effective value of this property.
   */
  /** @export */ get path(): EnumValue { return this._path; }
  /** @export */ set path(value: EnumValue) {
    if (this._path !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'path');
      this._path = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the Style for the resulting trees.
   * Must be {@link TreeLayout.StyleLayered}, {@link TreeLayout.StyleAlternating},
   * {@link TreeLayout.StyleLastParents}, or {@link TreeLayout.StyleRootOnly}.
   *
   * The default value is {@link TreeLayout.StyleLayered}.
   */
  /** @export */ get treeStyle(): EnumValue { return this._treeStyle; }
  /** @export */ set treeStyle(value: EnumValue) {
    if (this._arrangement !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'treeStyle');
      if (value === TreeLayout.StyleLayered || value === TreeLayout.StyleAlternating ||
        value === TreeLayout.StyleLastParents || value === TreeLayout.StyleRootOnly) {
        this._treeStyle = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the manner in which the nodes are aligned in layers.
   * Must be {@link TreeLayout.LayerIndividual}, {@link TreeLayout.LayerSiblings},
   * or {@link TreeLayout.LayerUniform}.
   *
   * The default value is {@link TreeLayout.LayerIndividual}.
   * @since 1.4
   */
  /** @export */ get layerStyle(): EnumValue { return this._layerStyle; }
  /** @export */ set layerStyle(value: EnumValue) {
    if (this._arrangement !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'layerStyle');
      if (value === TreeLayout.LayerIndividual || value === TreeLayout.LayerSiblings || value === TreeLayout.LayerUniform) {
        this._layerStyle = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets whether this layout should find all {@link Node}s
   * whose category is "Comment" and
   * whose anchors are nodes represented in the network,
   * and increase the size of the corresponding TreeVertex to make room for the comment nodes.
   * The default value is true.
   * @since 1.3
   */
  /** @export */ get comments(): boolean { return this._comments; }
  /** @export */ set comments(value: boolean) {
    if (this._comments !== value) {
      Util.checkPrimitive(value, 'boolean', TreeLayout, 'comments');
      this._comments = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets how {@link #arrangeTrees} should lay out the separate trees.
   * Must be {@link TreeLayout.ArrangementVertical}, {@link TreeLayout.ArrangementHorizontal},
   * or {@link TreeLayout.ArrangementFixedRoots}.
   *
   * The default value is {@link TreeLayout.ArrangementVertical}
   */
  /** @export */ get arrangement(): EnumValue { return this._arrangement; }
  /** @export */ set arrangement(value: EnumValue) {
    if (this._arrangement !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'arrangement');
      if (value === TreeLayout.ArrangementVertical || value === TreeLayout.ArrangementHorizontal || value === TreeLayout.ArrangementFixedRoots) {
        this._arrangement = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the space between which {@link #arrangeTrees} will position the trees.
   *
   * This defaults to the Size(10, 10).
   *
   * This property is ignored if {@link #arrangement} is {@link TreeLayout.ArrangementFixedRoots}.
   */
  /** @export */ get arrangementSpacing(): Size { return this._arrangementSpacing; }
  /** @export */ set arrangementSpacing(value: Size) {
    Util.checkClass(value, Size, TreeLayout, 'arrangementSpacing');
    if (!this._arrangementSpacing.equals(value)) {
      this._arrangementSpacing.assign(value);
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the object holding the default values for root {@link TreeVertex}es.
   *
   * The values for the following inheritable properties are actually stored in this object:
   * {@link #sorting}, {@link #comparer}, {@link #angle},
   * {@link #alignment}, {@link #nodeIndent}, {@link #nodeIndentPastParent},
   * {@link #nodeSpacing}, {@link #layerSpacing}, {@link #layerSpacingParentOverlap},
   * {@link #compaction}, {@link #breadthLimit}, {@link #rowSpacing}, {@link #rowIndent},
   * {@link #commentSpacing}, {@link #commentMargin},
   * {@link #setsPortSpot}, {@link #portSpot}, {@link #setsChildPortSpot}, {@link #childPortSpot}.
   * The other properties of this {@link TreeVertex} are ignored.
   */
  /** @export */ get rootDefaults(): TreeVertex { return this._rootDefaults; }
  /** @export */ set rootDefaults(value: TreeVertex) {
    if (this._rootDefaults !== value) {
      Util.checkClass(value, TreeVertex, TreeLayout, 'rootDefaults');
      this._rootDefaults = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the object holding the default values for alternate layer {@link TreeVertex}es,
   * used when the {@link #treeStyle} is {@link TreeLayout.StyleAlternating|StyleAlternating} or {@link TreeLayout.StyleLastParents|StyleLastParents}.
   *
   * See the list of inheritable properties in the documentation for {@link #rootDefaults}.
   * The other properties of this {@link TreeVertex} are ignored.
   */
  /** @export */ get alternateDefaults(): TreeVertex { return this._alternateDefaults; }
  /** @export */ set alternateDefaults(value: TreeVertex) {
    if (this._alternateDefaults !== value) {
      Util.checkClass(value, TreeVertex, TreeLayout, 'alternateDefaults');
      this._alternateDefaults = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default sorting policy for ordering the immediate children of a vertex.
   * Must be {@link TreeLayout.SortingForwards}, {@link TreeLayout.SortingReverse},
   * {@link TreeLayout.SortingAscending}, or {@link TreeLayout.SortingDescending}.
   *
   * The default value is {@link TreeLayout.SortingForwards}.
   *
   * This sets the {@link #rootDefaults}' property of the same name.
   * The sort order is determined by {@link #comparer}.
   */
  /** @export */ get sorting(): EnumValue { return this._rootDefaults.sorting; }
  /** @export */ set sorting(value: EnumValue) {
    if (this._rootDefaults.sorting !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'sorting');
      if (value === TreeLayout.SortingForwards || value === TreeLayout.SortingReverse ||
        value === TreeLayout.SortingAscending || TreeLayout.SortingDescending) {
        this._rootDefaults.sorting = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the default comparison function used for sorting the immediate children of a vertex.
   *
   * The default comparer compares the {@link LayoutVertex#node} Text values.
   *
   * This sets the {@link #rootDefaults}' property of the same name.
   * Whether this comparison function is used is determined by the value of {@link #sorting}.
   * ```js
   *   $(go.TreeLayout,
   *     {
   *       sorting: go.TreeLayout.SortingAscending,
   *       comparer: function(va, vb) {
   *         var da = va.node.data;
   *         var db = vb.node.data;
   *         if (da.someProperty < db.someProperty) return -1;
   *         if (da.someProperty > db.someProperty) return 1;
   *         return 0;
   *       }
   *     }
   *   )
   * ```
   */
  /** @export */ get comparer(): ((a: TreeVertex, b: TreeVertex) => number) { return this._rootDefaults.comparer; }
  /** @export */ set comparer(value: ((a: TreeVertex, b: TreeVertex) => number)) {
    if (this._rootDefaults.comparer !== value) {
      Util.checkPrimitive(value, 'function', TreeLayout, 'comparer');
      this._rootDefaults.comparer = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default direction for tree growth.
   *
   * The default value is 0; the value must be one of: 0, 90, 180, 270.
   *
   * These values are in degrees, where 0 is along the positive X axis,
   * and where 90 is along the positive Y axis.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get angle(): number { return this._rootDefaults.angle; }
  /** @export */ set angle(value: number) {
    if (this._rootDefaults.angle !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'angle');
      if (value === 0 || value === 90 || value === 180 || value === 270) {
        this._rootDefaults.angle = value;
        this.invalidateLayout();
      } else {
        Util.throwError('TreeLayout.angle must be 0, 90, 180, or 270');
      }
    }
  }

  /**
   * Gets or sets the default alignment of parents relative to their children.
   * Must be a static constant of {@link TreeLayout} whose name starts with "Alignment".
   *
   * The default value is {@link TreeLayout.AlignmentCenterChildren}.
   *
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get alignment(): EnumValue { return this._rootDefaults.alignment; }
  /** @export */ set alignment(value: EnumValue) {
    if (this._rootDefaults.alignment !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'alignment');
      this._rootDefaults.alignment = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default indentation of the first child.
   *
   * The default value is zero.  The value should be non-negative.
   *
   * This property is only sensible when the {@link #alignment}
   * is {@link TreeLayout.AlignmentStart|AlignmentStart} or {@link TreeLayout.AlignmentEnd|AlignmentEnd}.
   * Having a positive value is useful if you want to reserve space
   * at the start of the row of children for some reason.
   * For example, if you want to pretend the parent node is infinitely deep,
   * you can set this to be the breadth of the parent node.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get nodeIndent(): number { return this._rootDefaults.nodeIndent; }
  /** @export */ set nodeIndent(value: number) {
    if (this._rootDefaults.nodeIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'nodeIndent');
      if (value >= 0) {
        this._rootDefaults.nodeIndent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the fraction of this node's breadth is added to {@link #nodeIndent}
   * to determine any spacing at the start of the children.
   *
   * The default value is 0.0 -- the only indentation is specified by {@link #nodeIndent}.
   * When the value is 1.0, the children will be indented past the breadth of the parent node.
   *
   * This property is only sensible when the {@link #alignment}
   * is {@link TreeLayout.AlignmentStart|AlignmentStart} or {@link TreeLayout.AlignmentEnd|AlignmentEnd}.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get nodeIndentPastParent(): number { return this._rootDefaults.nodeIndentPastParent; }
  /** @export */ set nodeIndentPastParent(value: number) {
    if (this._rootDefaults.nodeIndentPastParent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'nodeIndentPastParent');
      if (value >= 0 && value <= 1) {
        this._rootDefaults.nodeIndentPastParent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the distance between child nodes.
   *
   * The default value is 20.
   *
   * A negative value causes sibling nodes to overlap.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get nodeSpacing(): number { return this._rootDefaults.nodeSpacing; }
  /** @export */ set nodeSpacing(value: number) {
    if (this._rootDefaults.nodeSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'nodeSpacing');
      this._rootDefaults.nodeSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the distance between a parent node and its children.
   *
   * The default value is 50.
   *
   * This is the distance between a parent node and its first row
   * of children, in case there are multiple rows of its children.
   * The {@link #rowSpacing} property determines the distance
   * between rows of children.
   * Negative values may cause children to overlap with the parent.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get layerSpacing(): number { return this._rootDefaults.layerSpacing; }
  /** @export */ set layerSpacing(value: number) {
    if (this._rootDefaults.layerSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'layerSpacing');
      this._rootDefaults.layerSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the fraction of the node's depth for which the children's layer starts overlapped with the parent's layer.
   *
   * The default value is 0.0 -- there is overlap between layers only if {@link #layerSpacing} is negative.
   * A value of 1.0 and a zero {@link #layerSpacing} will cause child nodes to completely overlap the parent.
   *
   * A value greater than zero may still cause overlap between layers,
   * unless the value of {@link #layerSpacing} is large enough.
   * A value of zero might still allow overlap between layers,
   * if {@link #layerSpacing} is negative.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get layerSpacingParentOverlap(): number { return this._rootDefaults.layerSpacingParentOverlap; }
  /** @export */ set layerSpacingParentOverlap(value: number) {
    if (this._rootDefaults.layerSpacingParentOverlap !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'layerSpacingParentOverlap');
      if (value >= 0 && value <= 1) {
        this._rootDefaults.layerSpacingParentOverlap = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets how closely to pack the child nodes of a subtree.
   * Must be either {@link TreeLayout.CompactionBlock} or {@link TreeLayout.CompactionNone}.
   *
   * The default value is {@link TreeLayout.CompactionBlock|CompactionBlock}.
   *
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get compaction(): EnumValue { return this._rootDefaults.compaction; }
  /** @export */ set compaction(value: EnumValue) {
    if (this._rootDefaults.compaction !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'compaction');
      if (value === TreeLayout.CompactionNone || value === TreeLayout.CompactionBlock) {
        this._rootDefaults.compaction = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets a limit on how broad a tree should be.
   *
   * A value of zero (the default) means there is no limit;
   * a positive value specifies a limit.
   * The default value is zero.
   *
   * This property is just a suggested constraint on how
   * broadly the tree will be laid out.
   * When there isn't enough breadth for all of the children of a node,
   *  the children are placed in as many rows as needed to try to stay
   * within the given breadth limit.
   * If the value is too small, since this layout algorithm
   * does not modify the size or shape of any node, the nodes will
   * just be laid out in a line, one per row, and the breadth is
   * determined by the broadest node.
   * The distance between rows is specified by {@link #rowSpacing}.
   * To make room for the links that go around earlier rows to get to
   * later rows, when the alignment is not a "center" alignment, the
   * {@link #rowIndent} property specifies that space at the
   * start of each row.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get breadthLimit(): number { return this._rootDefaults.breadthLimit; }
  /** @export */ set breadthLimit(value: number) {
    if (this._rootDefaults.breadthLimit !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'breadthLimit');
      if (value >= 0) {
        this._rootDefaults.breadthLimit = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the distance between rows of children.
   *
   * The default value is 25.
   *
   * This property is only used when there is more than one
   * row of children for a given parent node.
   * {@link #layerSpacing} determines the distance between
   * the parent node and its first row of child nodes.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get rowSpacing(): number { return this._rootDefaults.rowSpacing; }
  /** @export */ set rowSpacing(value: number) {
    if (this._rootDefaults.rowSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'rowSpacing');
      this._rootDefaults.rowSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the default indentation of the first child of each row,
   * if the {@link #alignment} is not a "Center" alignment.
   *
   * The default value is 10.  The value should be non-negative.
   *
   * This is used to leave room for the links that connect a parent node
   * with the child nodes that are in additional rows.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get rowIndent(): number { return this._rootDefaults.rowIndent; }
  /** @export */ set rowIndent(value: number) {
    if (this._rootDefaults.rowIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'rowIndent');
      if (value >= 0) {
        this._rootDefaults.rowIndent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the distance between comments.
   *
   * The default value is 10.
   *
   * This is used by {@link #addComments} and {@link #layoutComments}.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get commentSpacing(): number { return this._rootDefaults.commentSpacing; }
  /** @export */ set commentSpacing(value: number) {
    if (this._rootDefaults.commentSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'commentSpacing');
      this._rootDefaults.commentSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the distance between a node and its comments.
   *
   * The default value is 20.
   *
   * This is used by {@link #addComments} and {@link #layoutComments}.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get commentMargin(): number { return this._rootDefaults.commentMargin; }
  /** @export */ set commentMargin(value: number) {
    if (this._rootDefaults.commentMargin !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'commentMargin');
      this._rootDefaults.commentMargin = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the
   * FromSpot for this parent node port.
   *
   * The default value is true -- this may modify the spot of the port of this node, the parent,
   * if the node has only a single port.
   *
   * The spot used depends on the value of {@link #portSpot}.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get setsPortSpot(): boolean { return this._rootDefaults.setsPortSpot; }
  /** @export */ set setsPortSpot(value: boolean) {
    if (this._rootDefaults.setsPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeLayout, 'setsPortSpot');
      this._rootDefaults.setsPortSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the spot that this node's port gets as its FromSpot.
   *
   * The default value is {@link Spot.Default}.
   *
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a FromSpot based on the parent node's {@link TreeVertex#angle}.
   * If the value is other than NoSpot, it is just assigned.
   * When {@link #path} is {@link TreeLayout.PathSource|PathSource},
   * the port's ToSpot is set instead of the FromSpot.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get portSpot(): Spot { return this._rootDefaults.portSpot; }
  /** @export */ set portSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeLayout, 'portSpot');
    if (!this._rootDefaults.portSpot.equals(value)) {
      this._rootDefaults.portSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the
   * ToSpot for each child node port.
   *
   * The default value is true -- this may modify the spot of the ports of the children nodes,
   * if the node has only a single port.
   *
   * The spot used depends on the value of {@link #childPortSpot}.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get setsChildPortSpot(): boolean { return this._rootDefaults.setsChildPortSpot; }
  /** @export */ set setsChildPortSpot(value: boolean) {
    if (this._rootDefaults.setsChildPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeLayout, 'setsChildPortSpot');
      this._rootDefaults.setsChildPortSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the spot that children nodes' ports get as their ToSpot.
   *
   * The default value is {@link Spot.Default}.
   *
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a ToSpot based on the parent node's {@link TreeVertex#angle}.
   * If the value is other than NoSpot, it is just assigned.
   * When {@link #path} is {@link TreeLayout.PathSource|PathSource},
   * the port's FromSpot is set instead of the ToSpot.
   * This sets the {@link #rootDefaults}' property of the same name.
   */
  /** @export */ get childPortSpot(): Spot { return this._rootDefaults.childPortSpot; }
  /** @export */ set childPortSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeLayout, 'childPortSpot');
    if (!this._rootDefaults.childPortSpot.equals(value)) {
      this._rootDefaults.childPortSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate sorting policy for ordering the immediate children of a vertex.
   * Must be {@link TreeLayout.SortingForwards}, {@link TreeLayout.SortingReverse},
   * {@link TreeLayout.SortingAscending}, or {@link TreeLayout.SortingDescending}.
   *
   * The default value is {@link TreeLayout.SortingForwards|SortingForwards}.
   *
   * This sets the {@link #alternateDefaults}' property of the same name.
   * The sort order is determined by {@link #alternateComparer}.
   */
  /** @export */ get alternateSorting(): EnumValue { return this._alternateDefaults.sorting; }
  /** @export */ set alternateSorting(value: EnumValue) {
    if (this._alternateDefaults.sorting !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'alternateSorting');
      if (value === TreeLayout.SortingForwards || value === TreeLayout.SortingReverse ||
        value === TreeLayout.SortingAscending || TreeLayout.SortingDescending) {
        this._alternateDefaults.sorting = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the alternate comparison function used for sorting the immediate children of a vertex.
   *
   * The default comparer compares the {@link LayoutVertex#node} Text values.
   *
   * This sets the {@link #alternateDefaults}' property of the same name.
   * Whether this comparison function is used is determined by the value of {@link #alternateSorting}.
   */
  /** @export */ get alternateComparer(): ((a: TreeVertex, b: TreeVertex) => number) { return this._alternateDefaults.comparer; }
  /** @export */ set alternateComparer(value: ((a: TreeVertex, b: TreeVertex) => number)) {
    if (this._alternateDefaults.comparer !== value) {
      Util.checkPrimitive(value, 'function', TreeLayout, 'alternateComparer');
      this._alternateDefaults.comparer = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate direction for tree growth.
   *
   * The default value is 0; the value must be one of: 0, 90, 180, 270.
   *
   * These values are in degrees, where 0 is along the positive X axis,
   * and where 90 is along the positive Y axis.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateAngle(): number { return this._alternateDefaults.angle; }
  /** @export */ set alternateAngle(value: number) {
    if (this._alternateDefaults.angle !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateAngle');
      if (value === 0 || value === 90 || value === 180 || value === 270) {
        this._alternateDefaults.angle = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the alternate alignment of parents relative to their children.
   * Must be a static constant of {@link TreeLayout} whose name starts with "Alignment".
   *
   * The default value is {@link TreeLayout.AlignmentCenterChildren}.
   *
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateAlignment(): EnumValue { return this._alternateDefaults.alignment; }
  /** @export */ set alternateAlignment(value: EnumValue) {
    if (this._alternateDefaults.alignment !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'alternateAlignment');
      this._alternateDefaults.alignment = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate indentation of the first child.
   *
   * The default value is zero.  The value should be non-negative.
   *
   * This property is only sensible when the {@link #alignment}
   * is {@link TreeLayout.AlignmentStart|AlignmentStart} or {@link TreeLayout.AlignmentEnd|AlignmentEnd}.
   * Having a positive value is useful if you want to reserve space
   * at the start of the row of children for some reason.
   * For example, if you want to pretend the parent node is infinitely deep,
   * you can set this to be the breadth of the parent node.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateNodeIndent(): number { return this._alternateDefaults.nodeIndent; }
  /** @export */ set alternateNodeIndent(value: number) {
    if (this._alternateDefaults.nodeIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateNodeIndent');
      if (value >= 0) {
        this._alternateDefaults.nodeIndent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the fraction of this node's breadth is added to {@link #nodeIndent}
   * to determine any spacing at the start of the children.
   *
   * The default value is 0.0 -- the only indentation is specified by {@link #nodeIndent}.
   * When the value is 1.0, the children will be indented past the breadth of the parent node.
   *
   * This property is only sensible when the {@link #alignment}
   * is {@link TreeLayout.AlignmentStart|AlignmentStart} or {@link TreeLayout.AlignmentEnd|AlignmentEnd}.
   */
  /** @export */ get alternateNodeIndentPastParent(): number { return this._alternateDefaults.nodeIndentPastParent; }
  /** @export */ set alternateNodeIndentPastParent(value: number) {
    if (this._alternateDefaults.nodeIndentPastParent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateNodeIndentPastParent');
      if (value >= 0 && value <= 1) {
        this._alternateDefaults.nodeIndentPastParent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the alternate distance between child nodes.
   *
   * The default value is 20.
   *
   * A negative value causes sibling nodes to overlap.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateNodeSpacing(): number { return this._alternateDefaults.nodeSpacing; }
  /** @export */ set alternateNodeSpacing(value: number) {
    if (this._alternateDefaults.nodeSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateNodeSpacing');
      this._alternateDefaults.nodeSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate distance between a parent node and its children.
   *
   * The default value is 50.
   *
   * This is the distance between a parent node and its first row
   * of children, in case there are multiple rows of its children.
   * The {@link #rowSpacing} property determines the distance
   * between rows of children.
   * Negative values may cause children to overlap with the parent.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateLayerSpacing(): number { return this._alternateDefaults.layerSpacing; }
  /** @export */ set alternateLayerSpacing(value: number) {
    if (this._alternateDefaults.layerSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateLayerSpacing');
      this._alternateDefaults.layerSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate fraction of the node's depth for which the children's layer starts overlapped with the parent's layer.
   *
   * The default value is 0.0 -- there is overlap between layers only if {@link #layerSpacing} is negative.
   * A value of 1.0 and a zero {@link #layerSpacing} will cause child nodes to completely overlap the parent.
   *
   * A value greater than zero may still cause overlap between layers,
   * unless the value of {@link #layerSpacing} is large enough.
   * A value of zero might still allow overlap between layers,
   * if {@link #layerSpacing} is negative.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateLayerSpacingParentOverlap(): number { return this._alternateDefaults.layerSpacingParentOverlap; }
  /** @export */ set alternateLayerSpacingParentOverlap(value: number) {
    if (this._alternateDefaults.layerSpacingParentOverlap !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateLayerSpacingParentOverlap');
      if (value >= 0 && value <= 1) {
        this._alternateDefaults.layerSpacingParentOverlap = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets how closely to pack the child nodes of a subtree.
   * Must be either {@link TreeLayout.CompactionBlock} or {@link TreeLayout.CompactionNone}.
   *
   * The default value is {@link TreeLayout.CompactionBlock}.
   *
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateCompaction(): EnumValue { return this._alternateDefaults.compaction; }
  /** @export */ set alternateCompaction(value: EnumValue) {
    if (this._alternateDefaults.compaction !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeLayout, 'alternateCompaction');
      if (value === TreeLayout.CompactionNone || value === TreeLayout.CompactionBlock) {
        this._alternateDefaults.compaction = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets an alternate limit on how broad a tree should be.
   *
   * A value of zero (the default) means there is no limit;
   * a positive value specifies a limit.
   * The default value is zero.
   *
   * This property is just a suggested constraint on how
   * broadly the tree will be laid out.
   * When there isn't enough breadth for all of the children of a node,
   *  the children are placed in as many rows as needed to try to stay
   * within the given breadth limit.
   * If the value is too small, since this layout algorithm
   * does not modify the size or shape of any node, the nodes will
   * just be laid out in a line, one per row, and the breadth is
   * determined by the broadest node.
   * The distance between rows is specified by {@link #rowSpacing}.
   * To make room for the links that go around earlier rows to get to
   * later rows, when the alignment is not a "center" alignment, the
   * {@link #rowIndent} property specifies that space at the
   * start of each row.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateBreadthLimit(): number { return this._alternateDefaults.breadthLimit; }
  /** @export */ set alternateBreadthLimit(value: number) {
    if (this._alternateDefaults.breadthLimit !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateBreadthLimit');
      if (value >= 0) {
        this._alternateDefaults.breadthLimit = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the alternate distance between rows of children.
   *
   * The default value is 25.
   *
   * This property is only used when there is more than one
   * row of children for a given parent node.
   * {@link #layerSpacing} determines the distance between
   * the parent node and its first row of child nodes.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateRowSpacing(): number { return this._alternateDefaults.rowSpacing; }
  /** @export */ set alternateRowSpacing(value: number) {
    if (this._alternateDefaults.rowSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateRowSpacing');
      this._alternateDefaults.rowSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate indentation of the first child of each row,
   * if the {@link #alignment} is not a "Center" alignment.
   *
   * The default value is 10.  The value should be non-negative.
   *
   * This is used to leave room for the links that connect a parent node
   * with the child nodes that are in additional rows.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateRowIndent(): number { return this._alternateDefaults.rowIndent; }
  /** @export */ set alternateRowIndent(value: number) {
    if (this._alternateDefaults.rowIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateRowIndent');
      if (value >= 0) {
        this._alternateDefaults.rowIndent = value;
        this.invalidateLayout();
      }
    }
  }

  /**
   * Gets or sets the alternate distance between comments.
   *
   * The default value is 10.
   *
   * This is used by {@link #addComments} and {@link #layoutComments}.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateCommentSpacing(): number { return this._alternateDefaults.commentSpacing; }
  /** @export */ set alternateCommentSpacing(value: number) {
    if (this._alternateDefaults.commentSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateCommentSpacing');
      this._alternateDefaults.commentSpacing = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate distance between a node and its comments.
   *
   * The default value is 20.
   *
   * This is used by {@link #addComments} and {@link #layoutComments}.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateCommentMargin(): number { return this._alternateDefaults.commentMargin; }
  /** @export */ set alternateCommentMargin(value: number) {
    if (this._alternateDefaults.commentMargin !== value) {
      Util.checkPrimitive(value, 'number', TreeLayout, 'alternateCommentMargin');
      this._alternateDefaults.commentMargin = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the
   * FromSpot for this parent node port.
   *
   * The default value is true -- this may modify the spot of the port of this node, the parent,
   * if the node has only a single port.
   *
   * The spot used depends on the value of {@link #portSpot}.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateSetsPortSpot(): boolean { return this._alternateDefaults.setsPortSpot; }
  /** @export */ set alternateSetsPortSpot(value: boolean) {
    if (this._alternateDefaults.setsPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeLayout, 'alternateSetsPortSpot');
      this._alternateDefaults.setsPortSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate spot that this node's port gets as its FromSpot.
   *
   * The default value is {@link Spot.Default}.
   *
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a FromSpot based on the parent node's {@link TreeVertex#angle}.
   * If the value is other than NoSpot, it is just assigned.
   * When {@link #path} is {@link TreeLayout.PathSource|PathSource},
   * the port's ToSpot is set instead of the FromSpot.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternatePortSpot(): Spot { return this._alternateDefaults.portSpot; }
  /** @export */ set alternatePortSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeLayout, 'alternatePortSpot');
    if (!this._alternateDefaults.portSpot.equals(value)) {
      this._alternateDefaults.portSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the
   * ToSpot for each child node port.
   *
   * The default value is true -- this may modify the spot of the ports of the children nodes,
   * if the node has only a single port.
   *
   * The spot used depends on the value of {@link #childPortSpot}.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateSetsChildPortSpot(): boolean { return this._alternateDefaults.setsChildPortSpot; }
  /** @export */ set alternateSetsChildPortSpot(value: boolean) {
    if (this._alternateDefaults.setsChildPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeLayout, 'alternateSetsChildPortSpot');
      this._alternateDefaults.setsChildPortSpot = value;
      this.invalidateLayout();
    }
  }

  /**
   * Gets or sets the alternate spot that children nodes' ports get as their ToSpot
   * The default value is {@link Spot.Default}.
   *
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a ToSpot based on the parent node's {@link TreeVertex#angle}.
   *
   * If the value is other than NoSpot, it is just assigned.
   * When {@link #path} is {@link TreeLayout.PathSource|PathSource},
   * the port's FromSpot is set instead of the ToSpot.
   * This sets the {@link #alternateDefaults}' property of the same name.
   */
  /** @export */ get alternateChildPortSpot(): Spot { return this._alternateDefaults.childPortSpot; }
  /** @export */ set alternateChildPortSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeLayout, 'alternateChildPortSpot');
    if (!this._alternateDefaults.childPortSpot.equals(value)) {
      this._alternateDefaults.childPortSpot = value;
      this.invalidateLayout();
    }
  }

  // ***********************************Enums*************************************

  // These values specify how to build a tree from the TreeNetwork
  /**
   * This value for {@link TreeLayout#path} causes the value of {@link Diagram#isTreePathToChildren}
   * to effectively choose either {@link TreeLayout.PathDestination} (if true) or {@link TreeLayout.PathSource} (if false).
   * @constant
   */
  public static PathDefault = new EnumValue(TreeLayout, 'PathDefault', -1);

  /**
   * The children of a {@link TreeVertex} are its {@link LayoutVertex#destinationVertexes},
   * the collection of connected {@link LayoutEdge#toVertex}es;
   * this value is used for {@link TreeLayout#path}.
   *
   * The tree roots are those {@link TreeVertex}es that have zero source edges.
   * @constant
   */
  public static PathDestination = new EnumValue(TreeLayout, 'PathDestination', 0);

  /**
   * The children of a {@link TreeVertex} are its {@link LayoutVertex#sourceVertexes},
   * the collection of connected {@link LayoutEdge#fromVertex}es;
   * this value is used for {@link TreeLayout#path}.
   *
   * The tree roots are those {@link TreeVertex}es that have zero destination edges.
   * @constant
   */
  public static PathSource = new EnumValue(TreeLayout, 'PathSource', 1);

  // These values specify whether to sort the children of a node,
  // and in what order to position them.
  /**
   * Lay out each child in the order in which they were found;
   * this value is used for {@link TreeLayout#sorting} or {@link TreeLayout#alternateSorting}.
   * @constant
   */
  public static SortingForwards = new EnumValue(TreeLayout, 'SortingForwards', 10);

  /**
   * Lay out each child in reverse order from which they were found;
   * this value is used for {@link TreeLayout#sorting} or {@link TreeLayout#alternateSorting}.
   * @constant
   */
  public static SortingReverse = new EnumValue(TreeLayout, 'SortingReverse', 11);

  /**
   * Lay out each child according to the sort order given by {@link TreeVertex#comparer};
   * this value is used for {@link TreeLayout#sorting} or {@link TreeLayout#alternateSorting}.
   * @constant
   */
  public static SortingAscending = new EnumValue(TreeLayout, 'SortingAscending', 12);

  /**
   * Lay out each child in reverse sort order given by {@link TreeVertex#comparer};
   * this value is used for {@link TreeLayout#sorting} or {@link TreeLayout#alternateSorting}.
   * @constant
   */
  public static SortingDescending = new EnumValue(TreeLayout, 'SortingDescending', 13);

  // These values specify how to position a parent {@link TreeVertex}
  // relative to its children, or how to align children in a bus.
  /**
   * The parent is centered at the middle of the range of its child subtrees;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * When there is a breadth limit that causes there to be multiple rows,
   * the links that extend from the parent to those children in rows past
   * the first one may cross over the nodes that are in earlier rows.
   * @constant
   */
  public static AlignmentCenterSubtrees = new EnumValue(TreeLayout, 'AlignmentCenterSubtrees', 20);

  /**
   * The parent is centered at the middle of the range of its immediate child nodes;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * When there is a breadth limit that causes there to be multiple rows,
   * the links that extend from the parent to those children in rows past
   * the first one may cross over the nodes that are in earlier rows.
   * @constant
   */
  public static AlignmentCenterChildren = new EnumValue(TreeLayout, 'AlignmentCenterChildren', 21);

  /**
   * The parent is positioned near the first of its children;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   * @constant
   */
  public static AlignmentStart = new EnumValue(TreeLayout, 'AlignmentStart', 22);

  /**
   * The parent is positioned near the last of its children;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   * @constant
   */
  public static AlignmentEnd = new EnumValue(TreeLayout, 'AlignmentEnd', 23);

  /**
   * The children are positioned in a bus on both sides of an "aisle" where the links to them go,
   * with the last odd child (if any) placed at the end of the aisle in the middle;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * The children, if they are themselves parents, continue at the same inherited angle;
   * use {@link TreeLayout.AlignmentBusBranching} if you want grandchildren to proceed growing in
   * the different angle as determined by the side.
   *
   * A bus does not take {@link TreeVertex#breadthLimit} into account.
   * @constant
   */
  public static AlignmentBus = new EnumValue(TreeLayout, 'AlignmentBus', 24);

  /**
   * Like {@link TreeLayout.AlignmentBus} with the children arranged on both sides of an "aisle"
   * with any last odd child placed at the end of the aisle,
   * but the children get an {@link TreeVertex#angle} that depends on which side of the aisle
   * they were placed;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * This only works well when the {@link TreeLayout#treeStyle} is {@link TreeLayout.StyleLayered}.
   *
   * A bus does not take {@link TreeVertex#breadthLimit} into account.
   * @constant
   */
  public static AlignmentBusBranching = new EnumValue(TreeLayout, 'AlignmentBusBranching', 25);

  /**
   * The children are positioned in a bus, only on the top or left side of the parent;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * A bus does not take {@link TreeVertex#breadthLimit} into account.
   * @constant
   */
  public static AlignmentTopLeftBus = new EnumValue(TreeLayout, 'AlignmentTopLeftBus', 26);

  /**
   * The children are positioned in a bus, only on the bottom or right side of the parent;
   * this value is used for {@link TreeLayout#alignment} or {@link TreeLayout#alternateAlignment}.
   *
   * A bus does not take {@link TreeVertex#breadthLimit} into account.
   * @constant
   */
  public static AlignmentBottomRightBus = new EnumValue(TreeLayout, 'AlignmentBottomRightBus', 27);

  // These values specify how closely packed the children of a node should be.
  /**
   * Only simple placement of children next to each other, as determined by their subtree breadth;
   * this value is used for {@link TreeLayout#compaction} or {@link TreeLayout#alternateCompaction}.
   *
   * For any node, there will not be another node at any depth occupying the same breadth position,
   * unless there are multiple rows;  In other words, if there is no breadth limit resulting in
   * multiple rows, with this compaction mode it is as if every node were infinitely deep.
   * @constant
   */
  public static CompactionNone = new EnumValue(TreeLayout, 'CompactionNone', 30);

  /**
   * A simple fitting of subtrees;
   * this value is used for {@link TreeLayout#compaction} or {@link TreeLayout#alternateCompaction}.
   *
   * This mode produces more compact trees -- often nicer looking too;
   * Nodes will not overlap each other, unless you have negative values
   * for some of the spacing properties;
   * However it is possible when the links are orthogonally styled that
   * occasionally the subtrees will be placed so close together that some
   * links may overlap the links or even the nodes of other subtrees.
   * @constant
   */
  public static CompactionBlock = new EnumValue(TreeLayout, 'CompactionBlock', 31);

  // These values specify the general style in which the nodes are laid out.
  /**
   * The normal tree style, where all of the children of each {@link TreeVertex} are lined up
   * horizontally or vertically;
   * this value is used for {@link TreeLayout#treeStyle}.
   *
   * Each {@link TreeVertex} gets its properties from its parent node;
   * {@link TreeLayout#rootDefaults} is used for all default {@link TreeVertex} property values;
   * {@link TreeLayout#alternateDefaults} is ignored.
   * @constant
   */
  public static StyleLayered = new EnumValue(TreeLayout, 'StyleLayered', 40);

  /**
   * Just like the standard layered tree style, except that the nodes with children but no grandchildren
   * have alternate properties;
   * this value is used for {@link TreeLayout#treeStyle}.
   *
   * Each {@link TreeVertex} gets its properties from its parent node;
   * However, for those nodes whose {@link TreeVertex#maxGenerationCount} is 1,
   * in other words when it has children but no grandchildren,
   * the properties are copied from {@link TreeLayout#alternateDefaults}.
   *
   * If the tree only has two levels, the root node gets the {@link TreeLayout#rootDefaults}.
   * @constant
   */
  public static StyleLastParents = new EnumValue(TreeLayout, 'StyleLastParents', 41);

  /**
   * Alternate layers of the tree have different properties, typically including the angle;
   * this value is used for {@link TreeLayout#treeStyle}.
   *
   * Each {@link TreeVertex} gets its properties from its grandparent node;
   * The root nodes get their defaults from {@link TreeLayout#rootDefaults}.
   *
   * The immediate children of root nodes get their defaults from {@link TreeLayout#alternateDefaults}.
   *
   * Depending on the properties used, it is possible for some link routes to cross over nodes.
   * @constant
   */
  public static StyleAlternating = new EnumValue(TreeLayout, 'StyleAlternating', 42);

  /**
   * All of the nodes get the alternate properties, except the root node gets the default properties;
   * this value is used for {@link TreeLayout#treeStyle}.
   *
   * The root node gets the {@link TreeLayout#rootDefaults} properties,
   * the root node's children get the {@link TreeLayout#alternateDefaults} properties,
   * and all the rest of the {@link TreeVertex}es get their properties from their parent node.
   * @constant
   */
  public static StyleRootOnly = new EnumValue(TreeLayout, 'StyleRootOnly', 43);

  // These values specify how to position the resulting trees in the document.
  /**
   * Position each tree in a non-overlapping fashion by increasing Y coordinates,
   * starting at the {@link Layout#arrangementOrigin};
   * this value is used for {@link TreeLayout#arrangement}.
   * @constant
   */
  public static ArrangementVertical = new EnumValue(TreeLayout, 'ArrangementVertical', 50);

  /**
   * Position each tree in a non-overlapping fashion by increasing X coordinates,
   * starting at the {@link Layout#arrangementOrigin};
   * this value is used for {@link TreeLayout#arrangement}.
   * @constant
   */
  public static ArrangementHorizontal = new EnumValue(TreeLayout, 'ArrangementHorizontal', 51);

  /**
   * Do not move each root node, but position all of their descendants relative to their root;
   * this value is used for {@link TreeLayout#arrangement}.
   * @constant
   */
  public static ArrangementFixedRoots = new EnumValue(TreeLayout, 'ArrangementFixedRoots', 52);

  // These values specify the manner in which nodes are placed into layers.
  /**
   * The normal layer style, where each node takes up only the depth that it needs;
   * this value is used for {@link TreeLayout#layerStyle}.
   * @constant
   * @since 1.4
   */
  public static LayerIndividual = new EnumValue(TreeLayout, 'LayerIndividual', 60);

  /**
   * A layer style where all of the children of a parent node take up the same amount of depth --
   * this typically causes all cousins to be aligned;
   * this value is used for {@link TreeLayout#layerStyle}.
   * @constant
   * @since 1.4
   */
  public static LayerSiblings = new EnumValue(TreeLayout, 'LayerSiblings', 61);

  /**
   * A layer style where all nodes with the same {@link TreeVertex#level} throughout the tree take up the same amount of depth --
   * if the {@link TreeVertex#angle} is the same for all nodes, this will result in all nodes in the same layer to be aligned;
   * this value is used for {@link TreeLayout#layerStyle}.
   * @constant
   * @since 1.4
   */
  public static LayerUniform = new EnumValue(TreeLayout, 'LayerUniform', 62);
} // #end TreeLayout
TreeLayout['className'] = 'TreeLayout';

TreeLayout['PathDefault'] = TreeLayout.PathDefault;
TreeLayout['PathDestination'] = TreeLayout.PathDestination;
TreeLayout['PathSource'] = TreeLayout.PathSource;
TreeLayout['SortingForwards'] = TreeLayout.SortingForwards;
TreeLayout['SortingReverse'] = TreeLayout.SortingReverse;
TreeLayout['SortingAscending'] = TreeLayout.SortingAscending;
TreeLayout['SortingDescending'] = TreeLayout.SortingDescending;
TreeLayout['AlignmentCenterSubtrees'] = TreeLayout.AlignmentCenterSubtrees;
TreeLayout['AlignmentCenterChildren'] = TreeLayout.AlignmentCenterChildren;
TreeLayout['AlignmentStart'] = TreeLayout.AlignmentStart;
TreeLayout['AlignmentEnd'] = TreeLayout.AlignmentEnd;
TreeLayout['AlignmentBus'] = TreeLayout.AlignmentBus;
TreeLayout['AlignmentBusBranching'] = TreeLayout.AlignmentBusBranching;
TreeLayout['AlignmentTopLeftBus'] = TreeLayout.AlignmentTopLeftBus;
TreeLayout['AlignmentBottomRightBus'] = TreeLayout.AlignmentBottomRightBus;
TreeLayout['CompactionNone'] = TreeLayout.CompactionNone;
TreeLayout['CompactionBlock'] = TreeLayout.CompactionBlock;
TreeLayout['StyleLayered'] = TreeLayout.StyleLayered;
TreeLayout['StyleLastParents'] = TreeLayout.StyleLastParents;
TreeLayout['StyleAlternating'] = TreeLayout.StyleAlternating;
TreeLayout['StyleRootOnly'] = TreeLayout.StyleRootOnly;
TreeLayout['ArrangementVertical'] = TreeLayout.ArrangementVertical;
TreeLayout['ArrangementHorizontal'] = TreeLayout.ArrangementHorizontal;
TreeLayout['ArrangementFixedRoots'] = TreeLayout.ArrangementFixedRoots;
TreeLayout['LayerIndividual'] = TreeLayout.LayerIndividual;
TreeLayout['LayerSiblings'] = TreeLayout.LayerSiblings;
TreeLayout['LayerUniform'] = TreeLayout.LayerUniform;

// ***********************************TreeNetwork*******************************

/**
 * This class represents an abstract graph of {@link TreeVertex}es and {@link TreeEdge}s
 * that can be constructed based on the {@link Node}s and {@link Link}s of a {@link Diagram}
 * so that the {@link TreeLayout} can operate independently of the diagram until it
 * is time to commit any node positioning or link routing.
 *
 * This class inherits from {@link LayoutNetwork}.
 * @extends LayoutNetwork
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class TreeNetwork extends LayoutNetwork {
  /** @hidden @internal */ public static className = 'TreeNetwork';
  constructor(layout: TreeLayout) {
    super(layout);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link TreeVertex}.
   * @this {TreeNetwork}
   * @return {TreeVertex}
   */
  public createVertex(): TreeVertex {
    return new TreeVertex(this);
  }

  /**
   * @hidden @internal
   * Allocate a new instance of {@link TreeEdge}.
   * @this {TreeNetwork}
   * @return {TreeEdge}
   */
  public createEdge(): TreeEdge {
    return new TreeEdge(this);
  }

} // #end TreeNetwork
TreeNetwork['className'] = 'TreeNetwork';

// ***********************************TreeVertex********************************

/**
 * This holds {@link TreeLayout}-specific information about {@link Node}s.
 *
 * This class inherits from {@link LayoutVertex}.
 * @extends LayoutVertex
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class TreeVertex extends LayoutVertex {
  /** @hidden @internal */ public static className = 'TreeVertex';
  /** @hidden @internal */ public _relativePosition: Point;
  /** @hidden @internal */ public _subtreeSize: Size;
  /** @hidden @internal */ public _subtreeOffset: Point;
  /** @hidden @internal */ private _initialized: boolean;
  /** @hidden @internal */ private _parent: TreeVertex | null;
  /** @hidden @internal */ private _children: Array<TreeVertex>;
  /** @hidden @internal */ private _level: number;
  /** @hidden @internal */ private _descendantCount: number;
  /** @hidden @internal */ private _maxChildrenCount: number;
  /** @hidden @internal */ private _maxGenerationCount: number;
  /** @hidden @internal */ private _comments: Array<Node> | null;
  /** @hidden @internal */ public routeFirstRow: boolean;
  /** @hidden @internal */ public routeAroundCentered: boolean;
  /** @hidden @internal */ public routeAroundLastParent: boolean;
  /** @hidden @internal */ public leftFringe: Array<Point> | null;
  /** @hidden @internal */ public rightFringe: Array<Point> | null;
  /** @hidden @internal */ private _sorting: EnumValue;
  /** @hidden @internal */ private _comparer: ((a: TreeVertex, b: TreeVertex) => number);
  /** @hidden @internal */ private _angle: number;
  /** @hidden @internal */ private _alignment: EnumValue;
  /** @hidden @internal */ private _nodeIndent: number;
  /** @hidden @internal */ private _nodeIndentPastParent: number;
  /** @hidden @internal */ private _nodeSpacing: number;
  /** @hidden @internal */ private _layerSpacing: number;
  /** @hidden @internal */ private _layerSpacingParentOverlap: number;
  /** @hidden @internal */ private _compaction: EnumValue;
  /** @hidden @internal */ private _breadthLimit: number;
  /** @hidden @internal */ private _rowSpacing: number;
  /** @hidden @internal */ private _rowIndent: number;
  /** @hidden @internal */ private _commentSpacing: number;
  /** @hidden @internal */ private _commentMargin: number;
  /** @hidden @internal */ private _setsPortSpot: boolean;
  /** @hidden @internal */ private _portSpot: Spot;
  /** @hidden @internal */ private _setsChildPortSpot: boolean;
  /** @hidden @internal */ private _childPortSpot: Spot;

  constructor(network: TreeNetwork) {
    super(network);

    // tree specific properties:
    this._initialized = false;

    this._parent = null;

    this._children = [];

    this._level = 0;

    this._descendantCount = 0;

    this._maxChildrenCount = 0;

    this._maxGenerationCount = 0;

    this._comments = null;

    // these three Point or Size objects get clobbered a lot, and we don't bother freezing them
    // they are not exposed because the property setters displace the values
    this._relativePosition = new Point(0, 0);
    this._subtreeSize = new Size(0, 0);
    this._subtreeOffset = new Point(0, 0);

    // unexposed properties
    this.routeFirstRow = false;
    this.routeAroundCentered = false;
    this.routeAroundLastParent = false;

    // internal state
    this.leftFringe = null;
    this.rightFringe = null;

    // inherited properties:
    this._sorting = TreeLayout.SortingForwards;

    this._comparer = LayoutVertex.standardComparer;

    this._angle = 0;

    this._alignment = TreeLayout.AlignmentCenterChildren;

    this._nodeIndent = 0;

    this._nodeIndentPastParent = 0;

    this._nodeSpacing = 20;

    this._layerSpacing = 50;

    this._layerSpacingParentOverlap = 0;

    this._compaction = TreeLayout.CompactionBlock;

    this._breadthLimit = 0;

    this._rowSpacing = 25;

    this._rowIndent = 10;

    this._commentSpacing = 10;

    this._commentMargin = 20;

    this._setsPortSpot = true;

    this._portSpot = Spot.Default;

    this._setsChildPortSpot = true;

    this._childPortSpot = Spot.Default;
  }

  /**
   * Copy inherited properties from another TreeVertex to this one.
   * @expose
   * @this {TreeVertex}
   * @param {TreeVertex} copy
   */
  public copyInheritedPropertiesFrom(copy: TreeVertex): void {
    if (copy === null) return;
    this._sorting = copy.sorting;
    this._comparer = copy.comparer;
    this._angle = copy.angle;
    this._alignment = copy.alignment;
    this._nodeIndent = copy.nodeIndent;
    this._nodeIndentPastParent = copy.nodeIndentPastParent;
    this._nodeSpacing = copy.nodeSpacing;
    this._layerSpacing = copy.layerSpacing;
    this._layerSpacingParentOverlap = copy.layerSpacingParentOverlap;
    this._compaction = copy.compaction;
    this._breadthLimit = copy.breadthLimit;
    this._rowSpacing = copy.rowSpacing;
    this._rowIndent = copy.rowIndent;
    this._commentSpacing = copy.commentSpacing;
    this._commentMargin = copy.commentMargin;
    this._setsPortSpot = copy.setsPortSpot;
    this._portSpot = copy.portSpot;
    this._setsChildPortSpot = copy.setsChildPortSpot;
    this._childPortSpot = copy.childPortSpot;
  }

  // ***********************************Properties********************************

  /**
   * Gets or sets whether this node has been initialized as part of {@link TreeLayout#doLayout} when building the tree structures.
   *
   * The default value is false.
   */
  /** @export */ get initialized(): boolean { return this._initialized; }
  /** @export */ set initialized(value: boolean) {
    if (this._initialized !== value) {
      Util.checkPrimitive(value, 'boolean', TreeVertex, 'initialized');
      this._initialized = value;
    }
  }

  /**
   * Gets or sets the logical parent for this node.
   *
   * The default value is null.
   *
   * This structural property is computed in {@link TreeLayout#doLayout} when building the tree structures.
   * You should probably not be setting this property.
   */
  /** @export */ get parent(): TreeVertex { return this._parent; }
  /** @export */ set parent(value: TreeVertex) {
    if (this._parent !== value) {
      if (Debug && value !== null) Util.checkClass(value, TreeVertex, TreeVertex, 'parent');
      this._parent = value;
    }
  }

  /**
   * Gets or sets the logical children for this node.
   *
   * The default value is an empty array.
   *
   * This structural property is computed in {@link TreeLayout#doLayout} when building the tree structures.
   * You should probably not be setting this property.
   */
  /** @export */ get children(): Array<TreeVertex> { return this._children; }
  /** @export */ set children(value: Array<TreeVertex>) {
    if (this._children !== value) {
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', TreeVertex, 'children:value');
      if (value !== null) {
        const l = value.length;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (Debug) Util.checkClass(n, TreeVertex, TreeVertex, 'children');
        }
      }
      this._children = value;
    }
  }

  /**
   * Gets or sets the number of single-parent ancestors this node has.
   *
   * This could also be interpreted as which layer this node is in.
   * A root node will have a value of zero.
   *
   * This informational property is computed in {@link TreeLayout#doLayout} when building the tree structures.
   * You should probably not be setting this property.
   */
  /** @export */ get level(): number { return this._level; }
  /** @export */ set level(value: number) {
    if (this._level !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'level');
      this._level = value;
    }
  }

  /**
   * Gets or sets the number of descendants this node has.
   *
   * For a leaf node, this will be zero.
   *
   * This informational property is computed as part of the {@link TreeLayout#initializeTreeVertexValues} pass.
   * You should probably not be setting this property.
   */
  /** @export */ get descendantCount(): number { return this._descendantCount; }
  /** @export */ set descendantCount(value: number) {
    if (this._descendantCount !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'descendantCount');
      this._descendantCount = value;
    }
  }

  /**
   * Gets or sets the maximum number of children of any descendant of this node.
   *
   * For a leaf node, this will be zero.
   *
   * This informational property is computed as part of the {@link TreeLayout#initializeTreeVertexValues} pass.
   * You should probably not be setting this property.
   */
  /** @export */ get maxChildrenCount(): number { return this._maxChildrenCount; }
  /** @export */ set maxChildrenCount(value: number) {
    if (this._maxChildrenCount !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'maxChildrenCount');
      this._maxChildrenCount = value;
    }
  }

  /**
   * Gets or sets the maximum depth of the subtrees below this node.
   *
   * For a leaf node, this will be zero.
   *
   * This informational property is computed as part of the {@link TreeLayout#initializeTreeVertexValues} pass.
   * You should probably not be setting this value.
   */
  /** @export */ get maxGenerationCount(): number { return this._maxGenerationCount; }
  /** @export */ set maxGenerationCount(value: number) {
    if (this._maxGenerationCount !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'maxGenerationCount');
      this._maxGenerationCount = value;
    }
  }

  /**
   * Gets or sets an array of {@link Node}s that will be positioned near this node.
   *
   * The default value is null.
   *
   * These objects should not have their own {@link TreeVertex}es to be laid out.
   * Typically these will be {@link Node}s whose Category is "Comment".
   * This array should be allocated and initialized in {@link TreeLayout#addComments}.
   */
  /** @export */ get comments(): Array<Node> { return this._comments; }
  /** @export */ set comments(value: Array<Node>) {
    if (this._comments !== value) {
      if (value !== null && !Array.isArray(value)) Util.throwTypeError(value, 'Array', TreeVertex, 'comments:value');
      if (value !== null) {
        const l = value.length;
        for (let i = 0; i < l; i++) {
          const n = value[i];
          if (Debug) Util.checkClass(n, Node, TreeVertex, 'comments');
        }
      }
      this._comments = value;
    }
  }

  /**
   * Gets or sets whether and in what order the children should be sorted.
   *
   * The default value is {@link TreeLayout.SortingForwards}.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get sorting(): EnumValue { return this._sorting; }
  /** @export */ set sorting(value: EnumValue) {
    if (this._sorting !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeVertex, 'sorting');
      this._sorting = value;
    }
  }

  /**
   * Gets or sets how the children should be sorted.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get comparer(): ((a: TreeVertex, b: TreeVertex) => number) { return this._comparer; }
  /** @export */ set comparer(value: ((a: TreeVertex, b: TreeVertex) => number)) {
    if (this._comparer !== value) {
      Util.checkPrimitive(value, 'function', TreeVertex, 'comparer');
      this._comparer = value;
    }
  }

  /**
   * Gets or sets the absolute angle at which this subtree should grow from this vertex.
   *
   * The default value is zero, meaning that general tree growth should proceed rightwards along the X axis.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * By default this is the same as the Angle of the parent {@link TreeVertex}.
   * However, after the initial propagation of property values, perhaps in
   * an override of {@link TreeLayout#assignTreeVertexValues},
   * you could just set this property to specify the angle at which this node grows its subtrees.
   */
  /** @export */ get angle(): number { return this._angle; }
  /** @export */ set angle(value: number) {
    if (this._angle !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'angle');
      this._angle = value;
    }
  }

  /**
   * Gets or sets how this parent node should be aligned relative to its children.
   *
   * The default value is {@link TreeLayout.AlignmentCenterChildren}.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get alignment(): EnumValue { return this._alignment; }
  /** @export */ set alignment(value: EnumValue) {
    if (this._alignment !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeVertex, 'alignment');
      this._alignment = value;
    }
  }

  /**
   * Gets or sets the distance the first child should be indented.
   *
   * The default value is zero. The value should be non-negative.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * This property is only sensible when the {@link #alignment} is
   * {@link TreeLayout.AlignmentStart} or {@link TreeLayout.AlignmentEnd}.
   */
  /** @export */ get nodeIndent(): number { return this._nodeIndent; }
  /** @export */ set nodeIndent(value: number) {
    if (this._nodeIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'nodeIndent');
      this._nodeIndent = value;
    }
  }

  /**
   * Gets or sets whether the first child should be indented past the parent node's breadth.
   *
   * The default value is 0.0 -- the only start or end spacing is provided by {@link #nodeIndent}.
   * Values must range from 0.0 to 1.0, where 1.0 means the full breadth of this node.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * This property is only sensible when the {@link #alignment} is
   * {@link TreeLayout.AlignmentStart} or {@link TreeLayout.AlignmentEnd}.
   */
  /** @export */ get nodeIndentPastParent(): number { return this._nodeIndentPastParent; }
  /** @export */ set nodeIndentPastParent(value: number) {
    if (this._nodeIndentPastParent !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'nodeIndentPastParent');
      this._nodeIndentPastParent = value;
    }
  }

  /**
   * Gets or sets the distance between child nodes.
   *
   * The default value is 20.
   *
   * A negative value will cause sibling nodes to overlap.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get nodeSpacing(): number { return this._nodeSpacing; }
  /** @export */ set nodeSpacing(value: number) {
    if (this._nodeSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'nodeSpacing');
      this._nodeSpacing = value;
    }
  }

  /**
   * Gets or sets the distance between this node and its children.
   *
   * The default value is 50.
   *
   * Negative values may cause children to overlap with the parent.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get layerSpacing(): number { return this._layerSpacing; }
  /** @export */ set layerSpacing(value: number) {
    if (this._layerSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'layerSpacing');
      this._layerSpacing = value;
    }
  }

  /**
   * Gets or sets the fraction of this node's depth that may overlap with the children's layer.
   *
   * The default value is 0.0 -- there is overlap only if {@link #layerSpacing} is negative.
   * Values must range from 0.0 to 1.0, where 1.0 means the full depth of this node.
   *
   * When this value is greater than 0.0, there might not be overlap if {@link #layerSpacing}
   * is larger than the depth of this node times this fraction.
   * Even when this value is 0.0, there may be overlap when {@link #layerSpacing} is negative.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get layerSpacingParentOverlap(): number { return this._layerSpacingParentOverlap; }
  /** @export */ set layerSpacingParentOverlap(value: number) {
    if (this._layerSpacingParentOverlap !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'layerSpacingParentOverlap');
      this._layerSpacingParentOverlap = value;
    }
  }

  /**
   * Gets or sets how the children of this node should be packed together.
   *
   * The default value is {@link TreeLayout.CompactionBlock}.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get compaction(): EnumValue { return this._compaction; }
  /** @export */ set compaction(value: EnumValue) {
    if (this._compaction !== value) {
      EnumValue.checkEnumValue(value, TreeLayout, TreeVertex, 'compaction');
      this._compaction = value;
    }
  }

  /**
   * Gets or sets how broad a node and its descendants should be.
   *
   * By default this is zero.  A value of zero imposes no limit;
   * a positive value will specify a limit for the total width of this subtree.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get breadthLimit(): number { return this._breadthLimit; }
  /** @export */ set breadthLimit(value: number) {
    if (this._breadthLimit !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'breadthLimit');
      this._breadthLimit = value;
    }
  }

  /**
   * Gets or sets the distance between rows within one layer, all sharing the same parent.
   *
   * The default value is 25.
   *
   * Negative values may cause nodes to overlap.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get rowSpacing(): number { return this._rowSpacing; }
  /** @export */ set rowSpacing(value: number) {
    if (this._rowSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'rowSpacing');
      this._rowSpacing = value;
    }
  }

  /**
   * Gets or sets the distance the first child of each row should be indented.
   *
   * The default value is 10.  The value should be non-negative.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * This property is only used when the {@link #breadthLimit} is positive,
   * and some initial space needs to be reserved in each row of nodes for the links
   * that are routed around those rows.
   */
  /** @export */ get rowIndent(): number { return this._rowIndent; }
  /** @export */ set rowIndent(value: number) {
    if (this._rowIndent !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'rowIndent');
      this._rowIndent = value;
    }
  }

  /**
   * Gets or sets the space to leave between consecutive comments.
   *
   * The default value is 10.
   *
   * Negative values may cause comments to overlap.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get commentSpacing(): number { return this._commentSpacing; }
  /** @export */ set commentSpacing(value: number) {
    if (this._commentSpacing !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'commentSpacing');
      this._commentSpacing = value;
    }
  }

  /**
   * Gets or sets the space to leave between the node and the comments.
   *
   * The default value is 20.
   *
   * Negative values may cause comments to overlap with the node.
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   */
  /** @export */ get commentMargin(): number { return this._commentMargin; }
  /** @export */ set commentMargin(value: number) {
    if (this._commentMargin !== value) {
      Util.checkPrimitive(value, 'number', TreeVertex, 'commentMargin');
      this._commentMargin = value;
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the FromSpot for this parent node port.
   *
   * The default value is true -- this may modify the spot of the port of this node, the parent,
   * if the node has only a single port.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * The spot used depends on the value of {@link #portSpot}.
   */
  /** @export */ get setsPortSpot(): boolean { return this._setsPortSpot; }
  /** @export */ set setsPortSpot(value: boolean) {
    if (this._setsPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeVertex, 'setsPortSpot');
      this._setsPortSpot = value;
    }
  }

  /**
   * Gets or sets the spot that this node's port gets as its FromSpot,
   * if {@link #setsPortSpot} is true and the node has only a single port.
   *
   * The default value is {@link Spot.Default}.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a FromSpot based on the parent node's {@link TreeVertex#angle}.
   * If the value is other than NoSpot, it is just assigned.
   * When {@link TreeLayout#path} is {@link TreeLayout.PathSource}, the port's ToSpot is set instead of the FromSpot.
   */
  /** @export */ get portSpot(): Spot { return this._portSpot; }
  /** @export */ set portSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeVertex, 'portSpot');
    if (!this._portSpot.equals(value)) {
      this._portSpot = value;
    }
  }

  /**
   * Gets or sets whether the TreeLayout should set the ToSpot for each child node port.
   *
   * The default value is true -- this may modify the spots of the ports of the children nodes,
   * if the node has only a single port.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * The spot used depends on the value of {@link #childPortSpot}.
   */
  /** @export */ get setsChildPortSpot(): boolean { return this._setsChildPortSpot; }
  /** @export */ set setsChildPortSpot(value: boolean) {
    if (this._setsChildPortSpot !== value) {
      Util.checkPrimitive(value, 'boolean', TreeVertex, 'setsChildPortSpot');
      this._setsChildPortSpot = value;
    }
  }

  /**
   * Gets or sets the spot that children nodes' ports get as their ToSpot,
   * if {@link #setsChildPortSpot} is true and the node has only a single port.
   *
   * The default value is {@link Spot.Default}.
   *
   * This inherited property is initialized in the {@link TreeLayout#initializeTreeVertexValues} pass.
   * A value of {@link Spot.Default} will cause the TreeLayout
   * to assign a ToSpot based on the parent node's {@link TreeVertex#angle}.
   * If the value is other than NoSpot, it is just assigned.
   * When {@link TreeLayout#path} is {@link TreeLayout.PathSource}, the port's FromSpot is set instead of the ToSpot.
   */
  /** @export */ get childPortSpot(): Spot { return this._childPortSpot; }
  /** @export */ set childPortSpot(value: Spot) {
    Util.checkClass(value, Spot, TreeVertex, 'childPortSpot');
    if (!this._childPortSpot.equals(value)) {
      this._childPortSpot = value;
    }
  }

  /**
   * This read-only property returns the number of immediate children this node has.
   *
   * The default value is zero.
   */
  /** @export */ get childrenCount(): number { return this.children.length; }

  /**
   * Gets or sets the position of this node relative to its parent node.
   *
   * This informational property is computed during {@link TreeLayout#doLayout}.
   * You should probably not be setting this property.
   */
  /** @export */ get relativePosition(): Point { return this._relativePosition; }
  /** @export */ set relativePosition(value: Point) { this._relativePosition.set(value); }

  /**
   * Gets or sets the size of the subtree (including all descendants) parented by this node.
   *
   * This informational property is computed during {@link TreeLayout#doLayout}.
   * Of course if there are no children, this is just the same as Size.
   * You should probably not be setting this property.
   */
  /** @export */ get subtreeSize(): Size { return this._subtreeSize; }
  /** @export */ set subtreeSize(value: Size) { this._subtreeSize.set(value); }

  /**
   * Gets or sets the offset of this parent node relative to its whole subtree.
   *
   * This informational property is computed during {@link TreeLayout#doLayout}.
   * Of course if there are no children, this is just (0, 0).
   * You should probably not be setting this property.
   */
  /** @export */ get subtreeOffset(): Point { return this._subtreeOffset; }
  /** @export */ set subtreeOffset(value: Point) { this._subtreeOffset.set(value); }
} // #end TreeVertex
TreeVertex['className'] = 'TreeVertex';

// ***********************************TreeEdge**********************************

/**
 * This holds {@link TreeLayout}-specific information about {@link Link}s.
 *
 * This class inherits from {@link LayoutEdge}.
 * @extends LayoutEdge
 * @unrestricted
 * @category Layout
 * @unindexed
 */
export class TreeEdge extends LayoutEdge {
  /** @hidden @internal */ public static className = 'TreeEdge';
  /** @hidden @internal */ private _relativePoint: Point;

  /** @hidden @internal */ public fromVertex: TreeVertex;
  /** @hidden @internal */ public toVertex: TreeVertex;

  constructor(network: TreeNetwork) {
    super(network);

    // tree specific properties:

    // this gets clobbered, but we don't bother to freeze it
    // this is not exposed because the property setter displaces the X,Y values
    this._relativePoint = new Point(0, 0);
  }

  /**
   * Commits the position of the Link and routes it.
   * @this {TreeEdge}
   */
  public commit(): void {
    const link = this.link;
    if (link === null) return;
    if (link.isAvoiding) return;

    const layout = this.network.layout as TreeLayout;
    let parent: TreeVertex = null;
    let child: TreeVertex = null;
    switch (layout._effectivePath) {
      case TreeLayout.PathDestination: parent = this.fromVertex; child = this.toVertex; break;
      case TreeLayout.PathSource: parent = this.toVertex; child = this.fromVertex; break;
      default: Util.throwError('Unhandled path value ' + layout._effectivePath.toString());
    }
    if (parent === null || child === null) return;

    const p = this._relativePoint;
    if (p.x === 0 && p.y === 0 && !parent.routeFirstRow) { // no rows
      this.adjustRouteForAngleChange(parent, child);
      return;
    }
    const vertexbounds = parent.bounds;
    const angle = layout.orthoAngle(parent);
    const layerspacing = layout.computeLayerSpacing(parent);
    const rowspacing = parent.rowSpacing;
    link.updateRoute();
    const bezier = (link.curve === Link.Bezier);
    const ortho = link.isOrthogonal;
    let idx = 0;
    let prev: Point;
    let next: Point;
    link.startRoute();
    if (ortho || bezier) {
      idx = 2;
      while (link.pointsCount > 4) link.removePoint(2);
      prev = link.getPoint(1);
      next = link.getPoint(2);
    } else {
      idx = 1;
      while (link.pointsCount > 3) link.removePoint(1);
      prev = link.getPoint(0);
      next = link.getPoint(link.pointsCount - 1);
    }
    const last = link.getPoint(link.pointsCount - 1);
    let c = 0.0;
    if (angle === 0) {
      if (parent.alignment === TreeLayout.AlignmentEnd) {
        // route around at Y coordinate relative to the bottom of the parent node
        c = vertexbounds.bottom + p.y;
        // try to keep the links straight from the parent node -- consider room from RowIndent and NodeIndent
        if (p.y === 0 && prev.y > last.y + parent.rowIndent) {
          c = Math.min(c, Math.max(prev.y, c - layout.computeNodeIndent(parent)));
        }
      } else if (parent.alignment === TreeLayout.AlignmentStart) {
        // route around at Y coordinate relative to the top of the parent node
        c = vertexbounds.top + p.y;
        // try to keep the links straight from the parent node -- consider room from RowIndent and NodeIndent
        if (p.y === 0 && prev.y < last.y - parent.rowIndent) {
          c = Math.max(c, Math.min(prev.y, c + layout.computeNodeIndent(parent)));
        }
      } else if (parent.routeAroundCentered || (parent.routeAroundLastParent && parent.maxGenerationCount === 1)) {
        c = vertexbounds.top - parent._subtreeOffset.y + p.y;
      } else {
        c = vertexbounds.y + vertexbounds.height / 2 + p.y;
      }
      if (bezier) {  // curved segments
        // add a straight curve at Y-coord C
        link.insertPointAt(idx, prev.x, c); idx++;
        link.insertPointAt(idx, vertexbounds.right + layerspacing, c); idx++;
        link.insertPointAt(idx, vertexbounds.right + layerspacing + (p.x - rowspacing) / 3, c); idx++;
        link.insertPointAt(idx, vertexbounds.right + layerspacing + (p.x - rowspacing) * 2 / 3, c); idx++;
        link.insertPointAt(idx, vertexbounds.right + layerspacing + (p.x - rowspacing), c); idx++;
        link.insertPointAt(idx, next.x, c); idx++;
      } else {  // straight line segments
        if (ortho) { link.insertPointAt(idx, vertexbounds.right + layerspacing / 2, prev.y); idx++; }
        link.insertPointAt(idx, vertexbounds.right + layerspacing / 2, c); idx++;
        link.insertPointAt(idx, vertexbounds.right + layerspacing + p.x - (ortho ? rowspacing / 2 : rowspacing), c); idx++;
        if (ortho) { link.insertPointAt(idx, link.getPoint(idx - 1).x, next.y); idx++; }
      }
    } else if (angle === 90) {
      if (parent.alignment === TreeLayout.AlignmentEnd) {
        c = vertexbounds.right + p.x;
        if (p.x === 0 && prev.x > last.x + parent.rowIndent) {
          c = Math.min(c, Math.max(prev.x, c - layout.computeNodeIndent(parent)));
        }
      } else if (parent.alignment === TreeLayout.AlignmentStart) {
        c = vertexbounds.left + p.x;
        if (p.x === 0 && prev.x < last.x - parent.rowIndent) {
          c = Math.max(c, Math.min(prev.x, c + layout.computeNodeIndent(parent)));
        }
      } else if (parent.routeAroundCentered || (parent.routeAroundLastParent && parent.maxGenerationCount === 1)) {
        c = vertexbounds.left - parent._subtreeOffset.x + p.x;
      } else {
        c = vertexbounds.x + vertexbounds.width / 2 + p.x;
      }
      if (bezier) {
        link.insertPointAt(idx, c, prev.y); idx++;
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing); idx++;
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing + (p.y - rowspacing) / 3); idx++;
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing + (p.y - rowspacing) * 2 / 3); idx++;
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing + (p.y - rowspacing)); idx++;
        link.insertPointAt(idx, c, next.y); idx++;
      } else {
        if (ortho) { link.insertPointAt(idx, prev.x, vertexbounds.bottom + layerspacing / 2); idx++; }
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing / 2); idx++;
        link.insertPointAt(idx, c, vertexbounds.bottom + layerspacing + p.y - (ortho ? rowspacing / 2 : rowspacing)); idx++;
        if (ortho) { link.insertPointAt(idx, next.x, link.getPoint(idx - 1).y); idx++; }
      }
    } else if (angle === 180) {
      if (parent.alignment === TreeLayout.AlignmentEnd) {
        c = vertexbounds.bottom + p.y;
        if (p.y === 0 && prev.y > last.y + parent.rowIndent) {
          c = Math.min(c, Math.max(prev.y, c - layout.computeNodeIndent(parent)));
        }
      } else if (parent.alignment === TreeLayout.AlignmentStart) {
        c = vertexbounds.top + p.y;
        if (p.y === 0 && prev.y < last.y - parent.rowIndent) {
          c = Math.max(c, Math.min(prev.y, c + layout.computeNodeIndent(parent)));
        }
      } else if (parent.routeAroundCentered || (parent.routeAroundLastParent && parent.maxGenerationCount === 1)) {
        c = vertexbounds.top - parent._subtreeOffset.y + p.y;
      } else {
        c = vertexbounds.y + vertexbounds.height / 2 + p.y;
      }
      if (bezier) {
        link.insertPointAt(idx, prev.x, c); idx++;
        link.insertPointAt(idx, vertexbounds.left - layerspacing, c); idx++;
        link.insertPointAt(idx, vertexbounds.left - layerspacing + (p.x + rowspacing) / 3, c); idx++;
        link.insertPointAt(idx, vertexbounds.left - layerspacing + (p.x + rowspacing) * 2 / 3, c); idx++;
        link.insertPointAt(idx, vertexbounds.left - layerspacing + (p.x + rowspacing), c); idx++;
        link.insertPointAt(idx, next.x, c); idx++;
      } else {
        if (ortho) { link.insertPointAt(idx, vertexbounds.left - layerspacing / 2, prev.y); idx++; }
        link.insertPointAt(idx, vertexbounds.left - layerspacing / 2, c); idx++;
        link.insertPointAt(idx, vertexbounds.left - layerspacing + p.x + (ortho ? rowspacing / 2 : rowspacing), c); idx++;
        if (ortho) { link.insertPointAt(idx, link.getPoint(idx - 1).x, next.y); idx++; }
      }
    } else if (angle === 270) {
      if (parent.alignment === TreeLayout.AlignmentEnd) {
        c = vertexbounds.right + p.x;
        if (p.x === 0 && prev.x > last.x + parent.rowIndent) {
          c = Math.min(c, Math.max(prev.x, c - layout.computeNodeIndent(parent)));
        }
      } else if (parent.alignment === TreeLayout.AlignmentStart) {
        c = vertexbounds.left + p.x;
        if (p.x === 0 && prev.x < last.x - parent.rowIndent) {
          c = Math.max(c, Math.min(prev.x, c + layout.computeNodeIndent(parent)));
        }
      } else if (parent.routeAroundCentered || (parent.routeAroundLastParent && parent.maxGenerationCount === 1)) {
        c = vertexbounds.left - parent._subtreeOffset.x + p.x;
      } else {
        c = vertexbounds.x + vertexbounds.width / 2 + p.x;
      }
      if (bezier) {
        link.insertPointAt(idx, c, prev.y); idx++;
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing); idx++;
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing + (p.y + rowspacing) / 3); idx++;
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing + (p.y + rowspacing) * 2 / 3); idx++;
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing + (p.y + rowspacing)); idx++;
        link.insertPointAt(idx, c, next.y); idx++;
      } else {
        if (ortho) { link.insertPointAt(idx, prev.x, vertexbounds.top - layerspacing / 2); idx++; }
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing / 2); idx++;
        link.insertPointAt(idx, c, vertexbounds.top - layerspacing + p.y + (ortho ? rowspacing / 2 : rowspacing)); idx++;
        if (ortho) { link.insertPointAt(idx, next.x, link.getPoint(idx - 1).y); idx++; }
      }
    } else {
      Util.throwError('Invalid angle ' + angle);
    }
    link.commitRoute();
  }

  /**
   * @hidden @internal
   * @this {TreeEdge}
   * @param {TreeVertex} parent
   * @param {TreeVertex} child
   */
  private adjustRouteForAngleChange(parent: TreeVertex, child: TreeVertex): void {
    if (Debug) Util.checkClass(parent, TreeVertex, TreeEdge, 'adjustRouteForAngleChange:parent');
    if (Debug) Util.checkClass(child, TreeVertex, TreeEdge, 'adjustRouteForAngleChange:child');
    const link = this.link;
    const layout = this.network.layout as TreeLayout;
    const angle = layout.orthoAngle(parent);
    const childangle = layout.orthoAngle(child);
    if (angle === childangle) return;
    const layerspacing = layout.computeLayerSpacing(parent);
    const pb = parent.bounds;
    const cb = child.bounds;
    // but maybe an Angle change causes the child node not to be adjacent to the layer
    // separating it with the parent node; only need to do anything if that's the case
    if ((angle === 0 && cb.left - pb.right < layerspacing + 1) ||
      (angle === 90 && cb.top - pb.bottom < layerspacing + 1) ||
      (angle === 180 && pb.left - cb.right < layerspacing + 1) ||
      (angle === 270 && pb.top - cb.bottom < layerspacing + 1)) {
      return;
    }
    link.updateRoute();
    const bezier = (link.curve === Link.Bezier);
    const ortho = link.isOrthogonal;
    const fromVertex = this.fromVertex as TreeVertex;
    const alignbus = layout.isBusAlignment(fromVertex.alignment);
    link.startRoute();
    if (angle === 0) {
      const x = pb.right + layerspacing / 2;
      if (bezier) {  // curved segments
        if (link.pointsCount === 4) {
          const y = link.getPoint(3).y;
          link.setPointAt(1, x - 20, link.getPoint(1).y);
          link.insertPointAt(2, x - 20, y);
          link.insertPointAt(3, x, y);
          link.insertPointAt(4, x + 20, y);
          link.setPointAt(5, link.getPoint(5).x, y);
        }
      } else if (ortho) {
        if (alignbus) {
          link.setPointAt(3, link.getPoint(2).x, link.getPoint(4).y);
        } else if (link.pointsCount === 6) {
          link.setPointAt(2, x, link.getPoint(2).y);
          link.setPointAt(3, x, link.getPoint(3).y);
        }
      } else {
        if (link.pointsCount === 4) {
          link.insertPointAt(2, x, link.getPoint(2).y);
        } else if (link.pointsCount === 3) {
          link.setPointAt(1, x, link.getPoint(2).y);
        } else if (link.pointsCount === 2) {
          link.insertPointAt(1, x, link.getPoint(1).y);
        }
      }
    } else if (angle === 90) {
      const y = pb.bottom + layerspacing / 2;
      if (bezier) {  // curved segments
        if (link.pointsCount === 4) {
          const x = link.getPoint(3).x;
          link.setPointAt(1, link.getPoint(1).x, y - 20);
          link.insertPointAt(2, x, y - 20);
          link.insertPointAt(3, x, y);
          link.insertPointAt(4, x, y + 20);
          link.setPointAt(5, x, link.getPoint(5).y);
        }
      } else if (ortho) {
        if (alignbus) {
          link.setPointAt(3, link.getPoint(2).x, link.getPoint(4).y);
        } else if (link.pointsCount === 6) {
          link.setPointAt(2, link.getPoint(2).x, y);
          link.setPointAt(3, link.getPoint(3).x, y);
        }
      } else {
        if (link.pointsCount === 4) {
          link.insertPointAt(2, link.getPoint(2).x, y);
        } else if (link.pointsCount === 3) {
          link.setPointAt(1, link.getPoint(2).x, y);
        } else if (link.pointsCount === 2) {
          link.insertPointAt(1, link.getPoint(1).x, y);
        }
      }
    } else if (angle === 180) {
      const x = pb.left - layerspacing / 2;
      if (bezier) {  // curved segments
        if (link.pointsCount === 4) {
          const y = link.getPoint(3).y;
          link.setPointAt(1, x + 20, link.getPoint(1).y);
          link.insertPointAt(2, x + 20, y);
          link.insertPointAt(3, x, y);
          link.insertPointAt(4, x - 20, y);
          link.setPointAt(5, link.getPoint(5).x, y);
        }
      } else if (ortho) {
        if (alignbus) {
          link.setPointAt(3, link.getPoint(2).x, link.getPoint(4).y);
        } else if (link.pointsCount === 6) {
          link.setPointAt(2, x, link.getPoint(2).y);
          link.setPointAt(3, x, link.getPoint(3).y);
        }
      } else {
        if (link.pointsCount === 4) {
          link.insertPointAt(2, x, link.getPoint(2).y);
        } else if (link.pointsCount === 3) {
          link.setPointAt(1, x, link.getPoint(2).y);
        } else if (link.pointsCount === 2) {
          link.insertPointAt(1, x, link.getPoint(1).y);
        }
      }
    } else if (angle === 270) {
      const y = pb.top - layerspacing / 2;
      if (bezier) {  // curved segments
        if (link.pointsCount === 4) {
          const x = link.getPoint(3).x;
          link.setPointAt(1, link.getPoint(1).x, y + 20);
          link.insertPointAt(2, x, y + 20);
          link.insertPointAt(3, x, y);
          link.insertPointAt(4, x, y - 20);
          link.setPointAt(5, x, link.getPoint(5).y);
        }
      } else if (ortho) {
        if (alignbus) {
          link.setPointAt(3, link.getPoint(2).x, link.getPoint(4).y);
        } else if (link.pointsCount === 6) {
          link.setPointAt(2, link.getPoint(2).x, y);
          link.setPointAt(3, link.getPoint(3).x, y);
        }
      } else {
        if (link.pointsCount === 4) {
          link.insertPointAt(2, link.getPoint(2).x, y);
        } else if (link.pointsCount === 3) {
          link.setPointAt(1, link.getPoint(2).x, y);
        } else if (link.pointsCount === 2) {
          link.insertPointAt(1, link.getPoint(1).x, y);
        }
      }
    }
    link.commitRoute();
  }

  // ***********************************Properties********************************

  /**
   * Gets or sets a Point, relative to the parent node,
   * that may be useful in routing this link.
   */
  /** @export */ get relativePoint(): Point { return this._relativePoint; }
  /** @export */ set relativePoint(value: Point) { this._relativePoint.set(value); }

} // #end TreeEdge
TreeEdge['className'] = 'TreeEdge';

/* tslint:disable:ordered-imports */
// // import { Util } from './util';
// import { EnumValue } from './enumValue';
// import { Map, Set, List } from './collections';
// import { Point } from './point';
// import { Size } from './size';
// import { Rect } from './rect';
// import { Margin } from './margin';
// import { Spot } from './spot';
// import { PathSegment, PathFigure, Geometry } from './geometry';
// import { DiagramEvent, InputEvent } from './inputEvent';
// import { ChangedEvent } from './changedEvent';
// import { Binding, Model } from './model';
// import { GraphLinksModel } from './graphLinksModel';
// import { TreeModel } from './treeModel';
// import { UndoManager, Transaction } from './undoManager';
// import { CommandHandler } from './commandHandler';
// import { Tool } from './tool';
// import { DraggingTool } from './draggingTool';
// import { RelinkingTool, LinkingTool, LinkingBaseTool } from './linkingTools';
// import { LinkReshapingTool } from './linkReshapingTool';
// import { ResizingTool } from './resizingTool';
// import { RotatingTool } from './rotatingTool';
// import { PanningTool, DragSelectingTool, ClickCreatingTool, ActionTool, ClickSelectingTool } from './selectingTools';
// import { TextEditingTool } from './textEditingTool';
// import { ToolManager } from './toolManager';
// import { Animation, AnimationManager, AnimationTrigger } from './animationManager';
// import { Layer } from './layer';
// import { Diagram, DraggingInfo } from './diagram';
// import { Palette } from './palette';
// import { Overview } from './overview';
// import { Brush } from './brush';
// import { GraphObject } from './graphObject';
// import { Panel } from './panel';
// import { RowColumnDefinition } from './rowColumnDefinition';
// import { Shape } from './shape';
// import { TextBlock, TextBlockMetrics } from './textBlock';
// import { Picture } from './picture';
// import { Adornment, Part } from './parts';
// import { Node } from './parts';
// import { Link } from './parts';
// import { Group, Placeholder } from './parts';
// import { LayoutEdge, LayoutVertex, LayoutNetwork, Layout } from './layout';
// import { GridLayout } from './gridLayout';
// import { GraphLinksPartManager } from './graphLinksPartManager';
// import { TreePartManager } from './treePartManager';

// import { CircularEdge, CircularVertex, CircularNetwork, CircularLayout } from './circularLayout';
// import { ForceDirectedEdge, ForceDirectedVertex, ForceDirectedNetwork, ForceDirectedLayout } from './forceDirectedLayout';
// import { LayeredDigraphEdge, LayeredDigraphVertex, LayeredDigraphNetwork, LayeredDigraphLayout } from './layeredDigraphLayout';
// import { TreeEdge, TreeVertex, TreeNetwork, TreeLayout } from './treeLayout';
// import { root } from './root';
// import { HTMLInfo } from './htmlInfo';
// import { ContextMenuTool } from './contextMenuTool';
// import { PanelLayout } from './panelLayout';
// import { PanelLayoutHorizontal } from './panelLayoutHorizontal';
// import { PanelLayoutSpot } from './panelLayoutSpot';
// import { PanelLayoutTable, PanelLayoutTableRow, PanelLayoutTableColumn } from './panelLayoutTable';
// import { PanelLayoutViewbox } from './panelLayoutViewbox';
// import { PanelLayoutGraduated } from './panelLayoutGraduated';
// import { PanelLayoutGrid } from './panelLayoutGrid';
// import { Util } from './util';

// Already imported in Panel
// // import { PanelLayoutPosition } from './panelLayoutPosition';
// // import { PanelLayoutVertical } from './panelLayoutVertical';
// // import { PanelLayoutAuto } from './panelLayoutAuto';
// // import { PanelLayoutLink } from './panelLayoutLink';
// // import { PanelLayoutGrid } from './panelLayoutGrid';

/**
 * @suppress {duplicate}
 * @this {ToolManager}
 */
ToolManager.prototype.initializeStandardTools = function(this: ToolManager): void {
  // For each of these lists, the tools need to be assigned in a specific order,
  // so that they are added to the list in the correct order.

  // mouse-down tools:
  this.replaceStandardTool('Action', new ActionTool(), this.mouseDownTools);
  this.replaceStandardTool('Relinking', new RelinkingTool(), this.mouseDownTools);
  this.replaceStandardTool('LinkReshaping', new LinkReshapingTool(), this.mouseDownTools);
  this.replaceStandardTool('Rotating', new RotatingTool(), this.mouseDownTools);
  this.replaceStandardTool('Resizing', new ResizingTool(), this.mouseDownTools);

  // mouse-move tools:
  this.replaceStandardTool('Linking', new LinkingTool(), this.mouseMoveTools);
  this.replaceStandardTool('Dragging', new DraggingTool(), this.mouseMoveTools);
  this.replaceStandardTool('DragSelecting', new DragSelectingTool(), this.mouseMoveTools);
  this.replaceStandardTool('Panning', new PanningTool(), this.mouseMoveTools);

  // mouse-up tools:
  this.replaceStandardTool('ContextMenu', new ContextMenuTool(), this.mouseUpTools);
  this.replaceStandardTool('TextEditing', new TextEditingTool(), this.mouseUpTools);
  this.replaceStandardTool('ClickCreating', new ClickCreatingTool(), this.mouseUpTools);
  this.replaceStandardTool('ClickSelecting', new ClickSelectingTool(), this.mouseUpTools);
};

// By default, these are added to the library in index.ts. When building from source, you must add them manually.
Panel.definePanelLayout('Horizontal', new PanelLayoutHorizontal());
Panel.definePanelLayout('Spot', new PanelLayoutSpot());
Panel.definePanelLayout('Table', new PanelLayoutTable());
Panel.definePanelLayout('Viewbox', new PanelLayoutViewbox());
Panel.definePanelLayout('TableRow', new PanelLayoutTableRow());
Panel.definePanelLayout('TableColumn', new PanelLayoutTableColumn());
Panel.definePanelLayout('Graduated', new PanelLayoutGraduated());
Panel.definePanelLayout('Grid', new PanelLayoutGrid());
// These PanelLayouts are essential to GoJs and are added in panel.ts, so we don't add them here:
// Panel.definePanelLayout('Position', new PanelLayoutPosition());
// Panel.definePanelLayout('Vertical', new PanelLayoutVertical());
// Panel.definePanelLayout('Auto', new PanelLayoutAuto());
// Panel.definePanelLayout('Link', new PanelLayoutLink());


// Add PartManagers for model subclasses.
Diagram.addPartManager(GraphLinksModel.type, GraphLinksPartManager);
Diagram.addPartManager(TreeModel.type, TreePartManager);

